#ifndef __APM_H__
#define __APM_H__

#include "genlib.h"

//////////////////////////////////////////////////////////////////////
// APM Class -- Adaptive Precision Math
//

class APM
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//
public:
  enum AccuracyType { atGlobal, atLocal, atExact, atMPFUN, atMPFUNFilter, atMPFUNNormalized, atSize };

//////////////////////////////////////////////////
// Initialization
//
public:
  APM();
  virtual ~APM();
  virtual VOID Init();
  virtual VOID Uninit() = 0;

//////////////////////////////////////////////////
// Polynomial Solvers
//
public:
  // c1 x + c0 = 0
  static VOID SolveLinear(DOUBLE *pdSolutions, UINT_32 &uiSolutions, DOUBLE *pdCoefs, UINT_32 uiCoefs = 2);
  // c2 x^2 + c1 x + c0 = 0
  static VOID SolveQuadratic(DOUBLE *pdSolutions, UINT_32 &uiSolutions, DOUBLE *pdCoefs, UINT_32 uiCoefs = 3);
  // c3 x^3 + c2 x^2 + c1 x + c0 = 0
  static VOID SolveCubic(DOUBLE *pdSolutions, UINT_32 &uiSolutions, DOUBLE *pdCoefs, UINT_32 uiCoefs = 4);

//////////////////////////////////////////////////
// Accessor Functions
//
public:

//////////////////////////////////////////////////
// Class Accessor Functions
//
public:
  static inline AccuracyType GetAccuracyType();
  static inline VOID SetAccuracyType(AccuracyType eAccuracyType);

//////////////////////////////////////////////////
// Class Variables
//
protected:
  static AccuracyType m_eAccuracyType;

//////////////////////////////////////////////////
// Member Variables
//
protected:
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline APM::AccuracyType APM::GetAccuracyType()
{
  return m_eAccuracyType;
}

inline VOID APM::SetAccuracyType(APM::AccuracyType eAccuracyType)
{
  m_eAccuracyType = eAccuracyType;
}

#endif // __APM_H__
