#ifndef __APM3D_H__
#define __APM3D_H__

#include "apm.h"

//////////////////////////////////////////////////////////////////////
// APM2D Class -- Adaptive Precision Math 3D
//

class APM3D
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//
public:

//////////////////////////////////////////////////
// Initialization
//
public:
  APM3D();
  virtual ~APM3D();
  virtual VOID Init();
  virtual VOID Uninit() = 0;

//////////////////////////////////////////////////
// Orientation
//
public:
  static DOUBLE OrientationValueVVVV(DOUBLE dV0_x, DOUBLE dV0_y, DOUBLE dV0_z,
				     DOUBLE dV1_x, DOUBLE dV1_y, DOUBLE dV1_z,
				     DOUBLE dV2_x, DOUBLE dV2_y, DOUBLE dV2_z,
				     DOUBLE dV3_x, DOUBLE dV3_y, DOUBLE dV3_z);
    
//////////////////////////////////////////////////
// Inscribed Circle Center
//
public:
  static BOOL InscribedCircleCenterVVV(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdC_z, DOUBLE *pdR, 
				       DOUBLE dV0_x, DOUBLE dV0_y, DOUBLE dV0_z,
				       DOUBLE dV1_x, DOUBLE dV1_y, DOUBLE dV1_z,
				       DOUBLE dV2_x, DOUBLE dV2_y, DOUBLE dV2_z);

//////////////////////////////////////////////////
// Inscribed Sphere Center
//
public:
  static BOOL InscribedSphereCenterVVVV(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdC_z, DOUBLE *pdR, 
					DOUBLE dV0_x, DOUBLE dV0_y, DOUBLE dV0_z,
					DOUBLE dV1_x, DOUBLE dV1_y, DOUBLE dV1_z,
					DOUBLE dV2_x, DOUBLE dV2_y, DOUBLE dV2_z,
					DOUBLE dV3_x, DOUBLE dV3_y, DOUBLE dV3_z);
  static BOOL InscribedSphereCenterVVVSA(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdC_z, DOUBLE *pdR, 
					 DOUBLE dV0_x, DOUBLE dV0_y, DOUBLE dV0_z,
					 DOUBLE dV1_x, DOUBLE dV1_y, DOUBLE dV1_z,
					 DOUBLE dV2_x, DOUBLE dV2_y, DOUBLE dV2_z,
					 DOUBLE dSolidAngle);
  static BOOL SolidAngle(DOUBLE *pdSolidAngle, 
			 DOUBLE dV0_x, DOUBLE dV0_y, DOUBLE dV0_z,
			 DOUBLE dV1_x, DOUBLE dV1_y, DOUBLE dV1_z,
			 DOUBLE dV2_x, DOUBLE dV2_y, DOUBLE dV2_z,
			 DOUBLE dSphereR);

//////////////////////////////////////////////////
// Accessor Functions
//
public:

//////////////////////////////////////////////////
// Class Accessor Functions
//
public:
  static VOID DisplayMPFUNFilterCounters(FILE *fpOutput);

//////////////////////////////////////////////////
// Class Variables
//
public:

//////////////////////////////////////////////////
// Member Variables
//
protected:
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

#endif // __APM3D_H__
