#include "bitmap.h"
#include "gif.h"
#include <stdarg.h>

#ifdef TIFFANDJPEG
#include "slidetiff.h"
#include "slidejpeg.h"
#endif

static int ReadFileType(char *pcTextureFile);
static int ReadPicFile(int iFileType, char *pcFile, 
		       FILE *fpFile, BitMap *pBitMap, char* pcTextureFile);

int LoadBitMap(char *pcFile, char *pcTextureFile, BitMap *pBitMap)
{
  FILE *fpFile;
  int iResult, iFileType;

#ifdef WIN32
  fpFile = fopen(pcTextureFile, "rb");
#else
  fpFile = fopen(pcTextureFile, "r");
#endif 

  if ( fpFile == NULL )
    return 0;

  assert( fpFile != NULL && pBitMap != NULL );

  iFileType = ReadFileType(pcTextureFile);

  iResult = ReadPicFile(iFileType, pcFile, fpFile, pBitMap, pcTextureFile);

  fclose(fpFile);
  return iResult;
}

//////////////////////////////////////////////////
// ReadFileType
//
static int ReadFileType(char *pcTextureFile)
{
  int iFileType=FT_UNKNOWN;
  int uiLength = strlen(pcTextureFile);

  if(strcmp(pcTextureFile + uiLength - 4, ".gif")==0||
     strcmp(pcTextureFile + uiLength - 4, ".GIF")==0)
    iFileType = FT_GIF;
#ifdef TIFFANDJPEG
  if(strcmp(pcTextureFile + uiLength - 4, ".tif")==0||
     strcmp(pcTextureFile + uiLength - 4, ".TIF")==0)
    iFileType = FT_TIFF;
  if(strcmp(pcTextureFile + uiLength - 4, ".jpg")==0||
     strcmp(pcTextureFile + uiLength - 4, ".JPG")==0)
    iFileType = FT_JFIF;  
#endif
  return iFileType;
}

//////////////////////////////////////////////////
// ReadPicFile
//
static int ReadPicFile(int iFileType, char *pcFile, 
		       FILE *fpFile, BitMap *pBitMap, char *pcTextureFile)
{
  int iResult;

  iResult = 0;

  assert( pcFile != NULL );
  assert( fpFile != NULL );
  assert( pBitMap != NULL );

  switch ( iFileType ) 
    {
    case FT_GIF:     
      iResult = LoadGIF(pcFile, fpFile, pBitMap);
      break;
#ifdef TIFFANDJPEG
    case FT_TIFF:
      iResult = LoadTIFF(pcFile, pcTextureFile, pBitMap);
      break;
    case FT_JFIF:
      iResult = LoadJPEG(pcFile, pcTextureFile, pBitMap);
      break;
#endif
    }

  return iResult;
}

// if compiling for windows, define rindex
#ifdef WIN32
char *rindex (char* path, char tag)
{
	int pos = -1;
	
	for (unsigned int i=0; i < strlen(path); i++)
	{
		if (path[i] == tag)
		{
			pos = i;
		}
		
	}
	if (pos >= 0)
		return (path + pos);
	else
		return path;
}
#endif


char *BaseName(char *fname)
{
  char *basname;

  /* given a complete path name ('/foo/bar/weenie.gif'), returns just the 
     'simple' name ('weenie.gif').  Note that it does not make a copy of
     the name, so don't be modifying it... */

  basname = (char *) rindex(fname, '/');
  if (!basname) basname = fname;
  else basname++;

  return basname;
}

void SetISTR(int /*stnum*/, char *pcFormat, ...)
{
  va_list vaArgs;

  va_start(vaArgs, pcFormat);
  vfprintf(stderr, pcFormat, vaArgs);
  va_end(vaArgs);  
}

/*
#if defined(__STDC__) && !defined(NOSTDHDRS)
void SetISTR(int stnum, ...)
{
  va_list args;
  char     *fmt;

  va_start(args, stnum);
#else
//VARARGS0
void SetISTR(va_alist)
va_dcl
{
  va_list args;
  char    *fmt;
  int     stnum;

  va_start(args);

  stnum = va_arg(args, int);
#endif

  if (stnum>=0 && stnum < NISTR) {
    fmt = va_arg(args, char *);
    if (fmt) vsprintf(istrs[stnum], fmt, args);
    else istrs[stnum][0] = '\0';
  }
  va_end(args);
  
  if (stnum == ISTR_COLOR) {
    sprintf(istrs[ISTR_INFO], "%s  %s  %s", formatStr, 
	    (picType==PIC8) ? "8-bit mode." : "24-bit mode.",
	    istrs[ISTR_COLOR]);
  }

  if (infoUp) {
    redrawString(stnum);
    if (stnum == ISTR_COLOR) redrawString(ISTR_INFO);
    XFlush(theDisp);
  }

  if (ctrlUp && (stnum == ISTR_INFO || stnum == ISTR_WARNING || 
		 stnum == ISTR_COLOR)) {
    DrawCtrlStr();
    XFlush(theDisp);
  }

  if (anyBrowUp && (stnum == ISTR_WARNING || stnum == ISTR_INFO) 
      && strlen(istrs[stnum])) {
    SetBrowStr(istrs[stnum]);
    XFlush(theDisp);
  }

  if (stnum == ISTR_WARNING && !ctrlUp && !infoUp && !anyBrowUp && 
      strlen(istrs[stnum])) {
    OpenAlert(istrs[stnum]);
    sleep(3);
    CloseAlert();
  }
}
*/
