#include "bsplinemath.h"

////////////////////////////////////////////////////////////////////////////
// Curves

VOID BlossomAt(Point &pt, UINT_32 uiOrder, const Point *cpDeBoorPoint, 
	       const FLOAT *cpfKnot, const FLOAT t)
{
  ASSERT( uiOrder >= 2 );
  ASSERT( cpDeBoorPoint != NULL );
  ASSERT( cpfKnot != NULL );
  //ASSERT( t );

  UINT_32 i, uiLevel, uiDegree;
  FLOAT fBeta, fAlpha;
  Point *pBlossomPoint;

  uiDegree = uiOrder-1;
  pBlossomPoint = new Point[uiOrder];

  for( i = 0; i < uiOrder; i++ )
    {
      pBlossomPoint[i] = cpDeBoorPoint[i];
    }

  for ( uiLevel = 0 ; uiLevel < uiDegree; uiLevel++ ) 
    {
      for( i = 0; i < uiDegree - uiLevel; i++ ) 
	{
	  fBeta = (cpfKnot[uiDegree+i] - cpfKnot[uiLevel+i]);
	  
	  if( fBeta != 0.0 )
	    {  
	      // Make sure no division by zero
	      fBeta = (t - cpfKnot[uiLevel+i]) / fBeta;
	    }
	  fAlpha = 1 - fBeta;

	  pBlossomPoint[i] = pBlossomPoint[i] * fAlpha + pBlossomPoint[i+1] * fBeta;
      }
  }
  
  pt = pBlossomPoint[0];
  delete[] pBlossomPoint;
}

VOID BlossomAt(Point &pt, UINT_32 uiOrder, const Point *cpDeBoorPoint, 
	       const FLOAT *cpfKnot, const FLOAT *cpfBlossomLabel)
{
  ASSERT( uiOrder >= 2 );
  ASSERT( cpDeBoorPoint != NULL );
  ASSERT( cpfKnot != NULL );
  ASSERT( cpfBlossomLabel != NULL );

  UINT_32 i, uiLevel, uiDegree;
  FLOAT fBeta, fAlpha;
  Point *pBlossomPoint;

  uiDegree = uiOrder-1;
  pBlossomPoint = new Point[uiOrder];

  for( i = 0; i < uiOrder; i++ )
    {
      pBlossomPoint[i] = cpDeBoorPoint[i];
    }
  
  for ( uiLevel = 0 ; uiLevel < uiDegree; uiLevel++ ) 
    {
      for( i = 0; i < uiDegree - uiLevel; i++ ) 
	{
	  fBeta = (cpfKnot[uiDegree+i] - cpfKnot[uiLevel+i]);
	  
	  if( fBeta != 0.0 )
	    {  
	      // Make sure no division by zero
	      fBeta = (cpfBlossomLabel[uiLevel] - cpfKnot[uiLevel+i]) / fBeta;
	    }
	  fAlpha = 1 - fBeta;
	  
	  pBlossomPoint[i] = pBlossomPoint[i] * fAlpha +  pBlossomPoint[i+1] * fBeta;
      }
  }

  pt = pBlossomPoint[0];
  delete[] pBlossomPoint;
}

VOID DerivativeAt(Vector &v, UINT_32 uiOrder, Point **ppDerivativeTable, 
		  FLOAT **ppfDeBoorTable, UINT_32 uiDerivative, FLOAT t, 
		  UINT_32 uiDelta)
{
  ASSERT( uiDelta >= (uiOrder - 1 - uiDerivative) );
  ASSERT( ppDerivativeTable != NULL );
  ASSERT( ppfDeBoorTable != NULL );

  DerivativesAt(&v, uiOrder, ppDerivativeTable, ppfDeBoorTable, 
		uiDerivative, uiDerivative, t, uiDelta);
}

VOID DerivativesAt(Vector *pv, UINT_32 uiOrder, Point **ppDerivativeTable,
		   FLOAT **ppfDeBoorTable, UINT_32 uiStart, UINT_32 uiEnd,
		   FLOAT /*t*/, UINT_32 uiDelta)
{
  ASSERT( pv != NULL );
  ASSERT( ppDerivativeTable != NULL );
  ASSERT( ppfDeBoorTable != NULL );
  ASSERT( uiStart <= uiEnd );
  ASSERT( uiDelta >= (uiOrder - 1 - uiStart) );
  ASSERT( uiDelta >= (uiOrder - 1 - uiEnd) );

  UINT_32 i, j, uiDerivative, uiKnotBase, uiStop;
  Vector v;

  j = 0;
  for ( uiDerivative = uiStart; uiDerivative <= uiEnd; uiDerivative++ )
    {
      pv[j].Null();
      j++;
    }

  uiStop = MIN(uiEnd+1, uiOrder);
  j = 0;
  for ( uiDerivative = uiStart; uiDerivative < uiStop; uiDerivative++ )
    {
      uiKnotBase = uiDelta - (uiOrder - 1 - uiDerivative);

      for ( i = 0; i < uiOrder - uiDerivative; i++ )
	{
	  v = (ppDerivativeTable[uiKnotBase + i][uiDerivative]*
	       ppfDeBoorTable[uiDerivative+i][(uiOrder-1)-uiDerivative]);
	  
	  pv[j] = pv[j] + v;
	}
      j++;
    }
}

VOID WDerivativesAt(FLOAT *pf, UINT_32 uiOrder, FLOAT **ppfDerivativeTable,
		    FLOAT **ppfDeBoorTable, UINT_32 uiStart, UINT_32 uiEnd,
		    FLOAT /*t*/, UINT_32 uiDelta)
{
  ASSERT( pf != NULL );
  ASSERT( ppfDerivativeTable != NULL );
  ASSERT( ppfDeBoorTable != NULL );
  ASSERT( uiStart <= uiEnd );
  ASSERT( uiDelta >= (uiOrder - 1 - uiStart) );
  ASSERT( uiDelta >= (uiOrder - 1 - uiEnd) );

  UINT_32 i, j, uiDerivative, uiKnotBase, uiStop;
  FLOAT f;

  j = 0;
  for( uiDerivative = uiStart; uiDerivative <= uiEnd; uiDerivative++ )
    {
      pf[j] = 0;      j++;
    }

  uiStop = MIN(uiEnd+1, uiOrder);
  j = 0;
  for( uiDerivative = uiStart; uiDerivative < uiStop; uiDerivative++ )
    {
      uiKnotBase = uiDelta - (uiOrder - 1 - uiDerivative);

      for ( i = 0; i < uiOrder - uiDerivative; i++ )
	{
	  f = (ppfDerivativeTable[uiKnotBase + i][uiDerivative]*
	       ppfDeBoorTable[uiDerivative+i][(uiOrder-1)-uiDerivative]);
	  
	  pf[j] = pf[j] + f;
	}
      j++;
    }
}

VOID CalculateDeBoorTable(UINT_32 uiOrder, FLOAT *pfKnot, FLOAT **&ppfDeBoorTable,
			  FLOAT t, UINT_32 uiDelta)
{
  UINT_32 i, j, uiKnotBase, uiKnot;
  FLOAT fPrev;

  ASSERT( uiDelta >= (uiOrder - 1) );

  uiKnotBase = uiDelta - (uiOrder - 1);

  // Calculate the Order = 1 Basis Functions
  for ( i = 0; i < (2*uiOrder - 1); i++ )
    {
      uiKnot = uiKnotBase+i;
      if ( pfKnot[uiKnot] <= t && t < pfKnot[uiKnot+1] )
	{
	  ppfDeBoorTable[i][0] = 1.0;
	}
      else
	{
	  ppfDeBoorTable[i][0] = 0.0;
	}
    }

  for ( j = 1; j < uiOrder; j++ )
    {
      fPrev = pfKnot[uiKnotBase + j] - pfKnot[uiKnotBase];
      if ( fPrev != 0.0 )
	{
	  fPrev = ppfDeBoorTable[0][j-1] / fPrev;
	}
      for ( i = 0; i < (2*uiOrder - j - 1); i++ )
	{
	  uiKnot = uiKnotBase+i;
	  ppfDeBoorTable[i][j] = fPrev * (t - pfKnot[uiKnot]);

	  fPrev = pfKnot[uiKnot+1+j] - pfKnot[uiKnot+1];
	  if ( fPrev != 0.0 )
	    {
	      fPrev = ppfDeBoorTable[i+1][j-1] / fPrev;
	      ppfDeBoorTable[i][j] += fPrev * (pfKnot[uiKnot+1+j] - t);
	    }
	}
    }
}

VOID CalculateDerivativeTable(UINT_32 uiControlPoints, UINT_32 uiDerivatives, 
			      UINT_32 uiOrder, FLOAT *pfKnot, Point *pControlPoint,
			      Point **&ppDerivativeTable)
{
  UINT_32 uiDerivative, i;
  FLOAT fExponent, fUDelta;

  uiDerivative = 0;
  for ( i = uiDerivative; i < uiControlPoints; i++ )
    {
      ppDerivativeTable[i][uiDerivative] = pControlPoint[i];
    }

  for ( uiDerivative = 1; uiDerivative < uiDerivatives; uiDerivative++ )
    {
      //ASSERT( uiDerivative < uiOrder );
      fExponent = uiOrder - uiDerivative;

      for ( i = uiDerivative; i < uiControlPoints; i++ )
	{
	  fUDelta = pfKnot[i+uiOrder-uiDerivative] - pfKnot[i];
	  if ( fUDelta == 0.0 )
	    {
	      ppDerivativeTable[i][uiDerivative].Null();
	    }
	  else
	    {
	      ppDerivativeTable[i][uiDerivative] = 
		(ppDerivativeTable[i][uiDerivative-1] + 
		 (ppDerivativeTable[i-1][uiDerivative-1]*(-1.0))) * 
		(fExponent / fUDelta);
	    }
	}
    }  
}

VOID CalculateWDerivativeTable(UINT_32 uiControlPoints, UINT_32 uiDerivatives, 
			       UINT_32 uiOrder, FLOAT *pfKnot, FLOAT *pfWeight,
			       FLOAT **&ppfWDerivativeTable)
{
  UINT_32 uiDerivative, i;
  FLOAT fExponent, fUDelta;

  uiDerivative = 0;
  for ( i = uiDerivative; i < uiControlPoints; i++ )
    {
      ppfWDerivativeTable[i][uiDerivative] = pfWeight[i];
    }

  for ( uiDerivative = 1; uiDerivative < uiDerivatives; uiDerivative++ )
    {
      //ASSERT( uiDerivative < uiOrder );
      fExponent = uiOrder - uiDerivative;

      for ( i = uiDerivative; i < uiControlPoints; i++ )
	{
	  fUDelta = pfKnot[i+uiOrder-uiDerivative] - pfKnot[i];
	  if ( fUDelta == 0.0 )
	    {
	      ppfWDerivativeTable[i][uiDerivative] = 0.0;
	    }
	  else
	    {
	      ppfWDerivativeTable[i][uiDerivative] = 
		(ppfWDerivativeTable[i][uiDerivative-1] + 
		 (ppfWDerivativeTable[i-1][uiDerivative-1]*(-1.0))) * 
		(fExponent / fUDelta);
	    }
	}
    }  
}





