#include "cda.h"
#include "cdacell.h"

//////////////////////////////////////////////////////////////////////
// CDA Class
//

CDA::CDA()
{
}

CDA::~CDA()
{
  Uninit();
}

VOID CDA::Init()
{
  m_listCDACell.Init();
}

VOID CDA::Init(CListInt *plistCDAVertex)
{
  CDACell *pCDACell;

  ASSERT( plistCDAVertex != NULL );

  Init();

  pCDACell = AllocateCDACell();
  ASSERT( pCDACell != NULL );
  pCDACell->Init(plistCDAVertex); 

  AddCell(pCDACell);
}

VOID CDA::Uninit()
{
}

BOOL CDA::Initialized()
{
  return TRUE;
}

BOOL CDA::Valid()
{
  return TRUE;
}

CDA *CDA::Copy()
{
  CDA *pCDACopy;
  CListIterInt iterCDACell;
  CDACell *pCDACell;

  pCDACopy = AllocateCDA();
  ASSERT( pCDACopy != NULL );
  pCDACopy->Init();

  ////////////////////////////////////////////////////////////
  // Copy the Cells
  iterCDACell.Init(&m_listCDACell);
  for ( pCDACell = (CDACell*)(iterCDACell.PeekFirst());
	pCDACell != NULL;
	pCDACell = (CDACell*)(iterCDACell.PeekNext()) )
    {
      pCDACopy->AddCell(pCDACell->Copy());
    }

  return pCDACopy;
}

VOID CDA::Transform(const Matrix &mGroot_obj, const Matrix &mGobj_root)
{
  CListIterInt iterCDACell;
  CDACell *pCDACell;

  iterCDACell.Init(&m_listCDACell);
  for ( pCDACell = (CDACell*)(iterCDACell.PeekFirst());
	pCDACell != NULL;
	pCDACell = (CDACell*)(iterCDACell.PeekNext()) )
    {
      pCDACell->Transform(mGroot_obj, mGobj_root);
    }
}

CDA *CDA::Union(CDA *pCDA)
{
  CListIterInt iterCDACell;
  CDA *pCDAResult;
  CDACell *pCDACell;

  ASSERT( pCDA != NULL );

  pCDAResult = AllocateCDA();
  ASSERT( pCDAResult != NULL );
  pCDAResult->Init();

  ////////////////////////////////////////////////////////////
  // Make a new CDA which contains all the cells of the two operands

  iterCDACell.Init(&(this->m_listCDACell));
  for ( pCDACell = (CDACell*)(iterCDACell.PeekFirst());
	pCDACell != NULL;
	pCDACell = (CDACell*)(iterCDACell.PeekNext()) )
    {
      pCDAResult->AddCell(pCDACell);
    }

  iterCDACell.Init(&(pCDA->m_listCDACell));
  for ( pCDACell = (CDACell*)(iterCDACell.PeekFirst());
	pCDACell != NULL;
	pCDACell = (CDACell*)(iterCDACell.PeekNext()) )
    {
      pCDAResult->AddCell(pCDACell);
    }

  if ( pCDAResult->m_listCDACell.GetLength() == 0 )
    {
      DeallocateCDA(pCDAResult);
      pCDAResult = NULL;
    }

  return pCDAResult;
}

CDA *CDA::Intersect(CDA *pCDA)
{
  CListIterInt piterCDACell[2];
  CDA *pCDAResult;
  CDACell *ppCDACell[3];

  ASSERT( pCDA != NULL );

  pCDAResult = AllocateCDA();
  ASSERT( pCDAResult != NULL );
  pCDAResult->Init();

  ////////////////////////////////////////////////////////////
  // Calculate the intersections of all pairings of cells

  piterCDACell[0].Init(&(this->m_listCDACell));
  piterCDACell[1].Init(&(pCDA->m_listCDACell));
  for ( ppCDACell[0] = (CDACell*)(piterCDACell[0].PeekFirst());
	ppCDACell[0] != NULL;
	ppCDACell[0] = (CDACell*)(piterCDACell[0].PeekNext()) )
    {
      for ( ppCDACell[1] = (CDACell*)(piterCDACell[1].PeekFirst());
	    ppCDACell[1] != NULL;
	    ppCDACell[1] = (CDACell*)(piterCDACell[1].PeekNext()) )
	{
	  ppCDACell[2] = ppCDACell[0]->Intersect(ppCDACell[1]);
	  if ( ppCDACell[2] != NULL )
	    {
	      pCDAResult->AddCell(ppCDACell[2]);
	    }
	}
    }

  if ( pCDAResult->m_listCDACell.GetLength() == 0 )
    {
      DeallocateCDA(pCDAResult);
      pCDAResult = NULL;
    }

  return pCDAResult;
}

CDA *CDA::Minus(CDA *pCDA)
{
  CListIterInt piterCDACell[2];
  CDA *pCDAResult;
  CDACell *ppCDACell[2];

  ASSERT( pCDA != NULL );

  pCDAResult = AllocateCDA();
  ASSERT( pCDAResult != NULL );
  pCDAResult->Init();

  ////////////////////////////////////////////////////////////
  // Calculate the difference of all pairings of cells

  piterCDACell[0].Init(&(this->m_listCDACell));
  piterCDACell[1].Init(&(pCDA->m_listCDACell));
  for ( ppCDACell[0] = (CDACell*)(piterCDACell[0].PeekFirst());
	ppCDACell[0] != NULL;
	ppCDACell[0] = (CDACell*)(piterCDACell[0].PeekNext()) )
    {
      ppCDACell[0] = ppCDACell[0]->Copy();
      pCDAResult->m_listCDACell.InsertLast(ppCDACell[0], 0);

      for ( ppCDACell[1] = (CDACell*)(piterCDACell[1].PeekFirst());
	    ppCDACell[1] != NULL;
	    ppCDACell[1] = (CDACell*)(piterCDACell[1].PeekNext()) )
	{
	  ppCDACell[0]->Minus(&(pCDAResult->m_listCDACell), ppCDACell[1]);
	}
    }

  if ( pCDAResult->m_listCDACell.GetLength() == 0 )
    {
      DeallocateCDA(pCDAResult);
      pCDAResult = NULL;
    }

  return pCDAResult;
}

VOID CDA::CalculateBound()
{
/*
*/
}

VOID CDA::Render()
{
  CListIterInt iter;
  CDACell *pCDACell;

  // Set up OpenGL to use the Stencil Buffer

  iter.Init(&m_listCDACell);
  
  for ( pCDACell = (CDACell*)(iter.PeekFirst());
	pCDACell != NULL;
	pCDACell = (CDACell*)(iter.PeekNext()) )
    {
      pCDACell->Render();
    }

  // Stop OpenGL from using the Stencil Buffer
}

VOID CDA::RenderNormals()
{
  CListIterInt iter;
  CDACell *pCDACell;

  iter.Init(&m_listCDACell);
  
  for ( pCDACell = (CDACell*)(iter.PeekFirst());
	pCDACell != NULL;
	pCDACell = (CDACell*)(iter.PeekNext()) )
    {
      pCDACell->RenderNormals();
    }
}

VOID CDA::Display()
{
  CListIterInt iter;
  CDACell *pCDACell;

  printf("START: CDA %p\n", this);

  iter.Init(&m_listCDACell);
  
  for ( pCDACell = (CDACell*)(iter.PeekFirst());
	pCDACell != NULL;
	pCDACell = (CDACell*)(iter.PeekNext()) )
    {
      pCDACell->Display();
    }

  printf("END  : CDA %p\n", this);
}
