#include "cdacell.h"
#include "cdapolyhedron.h"

UINT_32 CDACell::m_uiObjectCount = 0;
CDACell *CDACell::m_pCDACellTemp = NULL;

//////////////////////////////////////////////////////////////////////
// CDACell Class
//

CDACell::CDACell()
{
  m_uiObjectCount++;
  if ( m_uiObjectCount == 1 )
    {
      m_pCDACellTemp = new CDACell;
      ASSERT( m_pCDACellTemp != NULL );
      m_pCDACellTemp->Init();
    }

  m_pCDAPolyhedronPos = NULL;
}

CDACell::~CDACell()
{
  Uninit();

  m_uiObjectCount--;
  if ( m_uiObjectCount == 0 )
    {
      if ( m_pCDACellTemp != NULL )
	{
	  delete m_pCDACellTemp;
	}
    }
}

VOID CDACell::Init()
{
  m_pCDAPolyhedronPos = NULL;

  m_listCDAPolyhedronNeg.Init();
  m_listCDAPolyhedronZero.Init();
}

VOID CDACell::Init(CListInt *plistCDAVertex)
{
  Init();

  m_pCDAPolyhedronPos = AllocateCDAPolyhedron();
  ASSERT( m_pCDAPolyhedronPos != NULL );
  m_pCDAPolyhedronPos->Init(plistCDAVertex); 
}

VOID CDACell::Uninit()
{
  if ( m_pCDAPolyhedronPos != NULL )
    {
      DeallocateCDAPolyhedron(m_pCDAPolyhedronPos);
      m_pCDAPolyhedronPos = NULL;
    }
}

BOOL CDACell::Initialized()
{
  return (m_pCDAPolyhedronPos != NULL);
}

BOOL CDACell::Valid()
{
  return (m_pCDAPolyhedronPos != NULL);
}

CDACell *CDACell::Copy()
{
  CDACell *pCDACellCopy;
  CListIterInt iterCDAPolyhedron;
  CDAPolyhedron *pCDAPolyhedron;

  pCDACellCopy = AllocateCDACell();
  ASSERT( pCDACellCopy != NULL );
  pCDACellCopy->Init();

  ////////////////////////////////////////////////////////////
  // Copy the Positive Polyhedron
  ASSERT( m_pCDAPolyhedronPos != NULL );
  pCDACellCopy->SetPositive(m_pCDAPolyhedronPos->Copy());

  ////////////////////////////////////////////////////////////
  // Copy the Negative Polyhedra
  iterCDAPolyhedron.Init(&m_listCDAPolyhedronNeg);
  for ( pCDAPolyhedron = (CDAPolyhedron*)(iterCDAPolyhedron.PeekFirst());
	pCDAPolyhedron != NULL;
	pCDAPolyhedron = (CDAPolyhedron*)(iterCDAPolyhedron.PeekNext()) )
    {
      pCDACellCopy->AddNegative(pCDAPolyhedron->Copy());
    }

  ////////////////////////////////////////////////////////////
  // Copy the Zero Polyhedra
  iterCDAPolyhedron.Init(&m_listCDAPolyhedronZero);
  for ( pCDAPolyhedron = (CDAPolyhedron*)(iterCDAPolyhedron.PeekFirst());
	pCDAPolyhedron != NULL;
	pCDAPolyhedron = (CDAPolyhedron*)(iterCDAPolyhedron.PeekNext()) )
    {
      pCDACellCopy->AddZero(pCDAPolyhedron->Copy());
    }

  return pCDACellCopy;
}

VOID CDACell::Transform(const Matrix &mGroot_obj, const Matrix &mGobj_root)
{
  CListIterInt iterCDAPolyhedron;
  CDAPolyhedron *pCDAPolyhedron;

  m_pCDAPolyhedronPos->Transform(mGroot_obj, mGobj_root);

  iterCDAPolyhedron.Init(&m_listCDAPolyhedronNeg);
  for ( pCDAPolyhedron = (CDAPolyhedron*)(iterCDAPolyhedron.PeekFirst());
	pCDAPolyhedron != NULL;
	pCDAPolyhedron = (CDAPolyhedron*)(iterCDAPolyhedron.PeekNext()) )
    {
      pCDAPolyhedron->Transform(mGroot_obj, mGobj_root);
    }

  iterCDAPolyhedron.Init(&m_listCDAPolyhedronZero);
  for ( pCDAPolyhedron = (CDAPolyhedron*)(iterCDAPolyhedron.PeekFirst());
	pCDAPolyhedron != NULL;
	pCDAPolyhedron = (CDAPolyhedron*)(iterCDAPolyhedron.PeekNext()) )
    {
      pCDAPolyhedron->Transform(mGroot_obj, mGobj_root);
    }
}

CDACell *CDACell::Intersect(CDACell *pCDACell)
{
  CDACell *pCDACellResult;
  CListIterInt piterCDAPolyhedron[2];
  CDAPolyhedron *ppCDAPolyhedron[3];

  ASSERT( pCDACell != NULL );
/*
  if ( pCDACell == NULL )
    {
      return NULL;
    }
*/

  ////////////////////////////////////////////////////////////
  // Calculate the Positive Cell

  ppCDAPolyhedron[0] = this->GetPositive();
  ppCDAPolyhedron[1] = pCDACell->GetPositive();

  ppCDAPolyhedron[2] = ppCDAPolyhedron[0]->Intersect(ppCDAPolyhedron[1]);
  if ( ppCDAPolyhedron[2] == NULL )
    {
      // There is no intersection
      return NULL;
    }

  pCDACellResult = AllocateCDACell();
  ASSERT( pCDACellResult != NULL );
  pCDACellResult->Init();
  
  pCDACellResult->SetPositive(ppCDAPolyhedron[2]);
  
  ////////////////////////////////////////////////////////////
  // Calculate the Negative Cells
  
  ppCDAPolyhedron[0] = ppCDAPolyhedron[2];
  
  piterCDAPolyhedron[0].Init(&(this->m_listCDAPolyhedronNeg));
  for ( ppCDAPolyhedron[1] = (CDAPolyhedron*)(piterCDAPolyhedron[0].PeekFirst());
	ppCDAPolyhedron[1] != NULL;
	ppCDAPolyhedron[1] = (CDAPolyhedron*)(piterCDAPolyhedron[0].PeekNext()) )
    {
      ppCDAPolyhedron[2] = ppCDAPolyhedron[0]->Intersect(ppCDAPolyhedron[1]);
      if ( ppCDAPolyhedron[2] != NULL )
	{
	  ppCDAPolyhedron[2]->Bind(FALSE);
	  pCDACellResult->AddNegative(ppCDAPolyhedron[2]);
	}
    }
	  
  piterCDAPolyhedron[0].Init(&(pCDACell->m_listCDAPolyhedronNeg));
  for ( ppCDAPolyhedron[1] = (CDAPolyhedron*)(piterCDAPolyhedron[0].PeekFirst());
	ppCDAPolyhedron[1] != NULL;
	ppCDAPolyhedron[1] = (CDAPolyhedron*)(piterCDAPolyhedron[0].PeekNext()) )
    {
      ppCDAPolyhedron[2] = ppCDAPolyhedron[0]->Intersect(ppCDAPolyhedron[1]);
      if ( ppCDAPolyhedron[2] != NULL )
	{
	  ppCDAPolyhedron[2]->Bind(TRUE);
	  pCDACellResult->AddNegative(ppCDAPolyhedron[2]);
	}
    }

  ////////////////////////////////////////////////////////////
  // Calculate the Zero Cells

  piterCDAPolyhedron[0].Init(&(pCDACellResult->m_listCDAPolyhedronNeg));
  piterCDAPolyhedron[1].Init(&(pCDACellResult->m_listCDAPolyhedronNeg));
  for ( ppCDAPolyhedron[0] = (CDAPolyhedron*)(piterCDAPolyhedron[0].PeekFirst());
	ppCDAPolyhedron[0] != NULL;
	ppCDAPolyhedron[0] = (CDAPolyhedron*)(piterCDAPolyhedron[0].PeekNext()) )
    {
      for ( ppCDAPolyhedron[1] = (CDAPolyhedron*)(piterCDAPolyhedron[1].PeekFirst());
	    ppCDAPolyhedron[1] != NULL;
	    ppCDAPolyhedron[1] = (CDAPolyhedron*)(piterCDAPolyhedron[1].PeekNext()) )
	{
	  if ( ppCDAPolyhedron[0] != ppCDAPolyhedron[1] )
	    {
	      ppCDAPolyhedron[2] = ppCDAPolyhedron[0]->Intersect(ppCDAPolyhedron[1]);
	      if ( ppCDAPolyhedron[2] != NULL )
		{
		  ppCDAPolyhedron[2]->Bind(TRUE);
		  pCDACellResult->AddZero(ppCDAPolyhedron[2]);
		}
	    }
	}
    }

  return pCDACellResult;
}

VOID CDACell::Minus(CListInt *plistCDACellResult, CDACell *pCDACell)
{
  CListIterInt iterCDAPolyhedron;
  CDA *pCDAResult;
  CDACell *ppCDACell[3];
  CDAPolyhedron *pCDAPolyhedron;
  CDAPolyhedron *ppCDAPolyhedron[3];

  ASSERT( plistCDACellResult != NULL );
  ASSERT( pCDACell != NULL );

  ppCDACell[0] = this;
  ppCDACell[1] = pCDACell;

  ppCDAPolyhedron[0] = ppCDACell[0]->GetPositive()->Intersect(ppCDACell[1]->GetPositive());
  if ( ppCDAPolyhedron[0] == NULL )
    {
      return;
    }

  //////////////////////////////////////////////////
  // Intersect the negative polyhedrons with the cell
  ASSERT( m_pCDACellTemp->m_pCDAPolyhedronPos == NULL );
  iterCDAPolyhedron.Init(&(ppCDACell[1]->m_listCDAPolyhedronNeg));
  for ( pCDAPolyhedron = (CDAPolyhedron*)(iterCDAPolyhedron.PeekFirst());
	pCDAPolyhedron != NULL;
	pCDAPolyhedron = (CDAPolyhedron*)(iterCDAPolyhedron.PeekNext()) )
    {
      m_pCDACellTemp->m_pCDAPolyhedronPos = pCDAPolyhedron;
      
      ppCDACell[2] = ppCDACell[0]->Intersect(m_pCDACellTemp);
      if ( ppCDACell[2] != NULL )
	{
	  plistCDACellResult->InsertLast(ppCDACell[2], 0);
	}
    }
  m_pCDACellTemp->m_pCDAPolyhedronPos = NULL;
  
  //////////////////////////////////////////////////
  // Add the extra negative polyhedron to the cell
  //
  ppCDAPolyhedron[0]->Bind(FALSE);
  ppCDACell[0]->AddNegative(ppCDAPolyhedron[0]);

  iterCDAPolyhedron.Init(&(ppCDACell[0]->m_listCDAPolyhedronNeg));
  for ( ppCDAPolyhedron[1] = (CDAPolyhedron*)(iterCDAPolyhedron.PeekFirst());
	ppCDAPolyhedron[1] != NULL;
	ppCDAPolyhedron[1] = (CDAPolyhedron*)(iterCDAPolyhedron.PeekNext()) )
    {
      if ( ppCDAPolyhedron[0] != ppCDAPolyhedron[1] )
	{
	  ppCDAPolyhedron[2] = ppCDAPolyhedron[0]->Intersect(ppCDAPolyhedron[1]);
	  if ( ppCDAPolyhedron[2] != NULL )
	    {
	      ppCDAPolyhedron[2]->Bind(TRUE);
	      ppCDACell[0]->AddZero(ppCDAPolyhedron[2]);
	    }
	}  
    }
}

VOID CDACell::CalculateBound()
{
/*
  CListIterInt iter;
  CDAPolyhedron *pCDAPolyhedron;
  BOOL bBoundDirty;

  CBound bound;
  
  iter.Init(&m_listCDAPolyhedron);
  
  for ( pCDAPolyhedron = (CDAPolyhedron*)iter.PeekFirst();
	pCDAPolyhedron != NULL;
	pCDAPolyhedron = (CDAPolyhedron*)iter.PeekNext() )
    {
      pCDAPolyhedron->InsertIntoBound(bound);
    }

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);
    }
*/
}

VOID CDACell::Render()
{
  CListIterInt iter;
  CDAPolyhedron *pCDAPolyhedronNeg;

  // Set up OpenGL to use the Stencil Buffer

  m_pCDAPolyhedronPos->Render(TRUE);

  iter.Init(&m_listCDAPolyhedronNeg);
  
  for ( pCDAPolyhedronNeg = (CDAPolyhedron*)iter.PeekFirst();
	pCDAPolyhedronNeg != NULL;
	pCDAPolyhedronNeg = (CDAPolyhedron*)iter.PeekNext() )
    {
      pCDAPolyhedronNeg->Render(FALSE);
    }

  // Stop OpenGL from using the Stencil Buffer
  
}

VOID CDACell::RenderNormals()
{
  CListIterInt iter;
  CDAPolyhedron *pCDAPolyhedronNeg;

  m_pCDAPolyhedronPos->RenderNormals(TRUE);

  iter.Init(&m_listCDAPolyhedronNeg);
  
  for ( pCDAPolyhedronNeg = (CDAPolyhedron*)iter.PeekFirst();
	pCDAPolyhedronNeg != NULL;
	pCDAPolyhedronNeg = (CDAPolyhedron*)iter.PeekNext() )
    {
      pCDAPolyhedronNeg->RenderNormals(FALSE);
    }
}

VOID CDACell::Display()
{
  CListIterInt iterCDAPolyhedron;
  CDAPolyhedron *pCDAPolyhedron;

  printf("  START: CDACell %p\n", this);
  printf("    START: Positive\n");
  m_pCDAPolyhedronPos->Display();
  printf("    END  : Positive\n");

  printf("    START: Negatives\n");
  iterCDAPolyhedron.Init(&m_listCDAPolyhedronNeg);
  for ( pCDAPolyhedron = (CDAPolyhedron*)(iterCDAPolyhedron.PeekFirst());
	pCDAPolyhedron != NULL;
	pCDAPolyhedron = (CDAPolyhedron*)(iterCDAPolyhedron.PeekNext()) )
    {
      pCDAPolyhedron->Display();
    }
  printf("    END  : Negatives\n");

  printf("    START: Zeroes\n");
  iterCDAPolyhedron.Init(&m_listCDAPolyhedronZero);
  for ( pCDAPolyhedron = (CDAPolyhedron*)(iterCDAPolyhedron.PeekFirst());
	pCDAPolyhedron != NULL;
	pCDAPolyhedron = (CDAPolyhedron*)(iterCDAPolyhedron.PeekNext()) )
    {
      pCDAPolyhedron->Display();
    }
  printf("    END  : Zeroes\n");

  printf("  END  : CDACell %p\n", this);
}
