#ifndef __CDACELL_H__
#define __CDACELL_H__

#include "cdamem.h"

//////////////////////////////////////////////////////////////////////
// CDACell Class
//

class CDACell
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CDACell();
  virtual ~CDACell();
  virtual VOID Init();
  virtual VOID Init(CListInt *plistCDAVertex);
  virtual VOID Uninit();

  virtual BOOL Initialized();
  virtual BOOL Valid();

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  CDACell *Copy();
  VOID Transform(const Matrix &mGroot_obj, const Matrix &mGobj_root);

  CDACell *Intersect(CDACell *pCDACell);
  VOID Minus(CListInt *plistCDACellResult, CDACell *pCDACell);

protected:
  virtual VOID CalculateBound();

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID Render();
  virtual VOID RenderNormals();

//////////////////////////////////////////////////
// Output Functions
//
public:
  VOID Display();

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  inline CDAPolyhedron *GetPositive();
  inline VOID SetPositive(CDAPolyhedron *pCDAPolyhedron);
  inline VOID AddNegative(CDAPolyhedron *pCDAPolyhedronNeg);
  inline VOID AddZero(CDAPolyhedron *pCDAPolyhedronZero);

//////////////////////////////////////////////////
// Class Variables
//
protected:
  static UINT_32 m_uiObjectCount;
  static CDACell *m_pCDACellTemp;

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CDAPolyhedron *m_pCDAPolyhedronPos;
  CListInt m_listCDAPolyhedronNeg;
  CListInt m_listCDAPolyhedronZero;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline CDAPolyhedron *CDACell::GetPositive()
{
  return m_pCDAPolyhedronPos;
}

inline VOID CDACell::SetPositive(CDAPolyhedron *pCDAPolyhedronPos)
{
  ASSERT( pCDAPolyhedronPos != NULL );
  ASSERT( m_pCDAPolyhedronPos == NULL );

  m_pCDAPolyhedronPos = pCDAPolyhedronPos;
}

inline VOID CDACell::AddNegative(CDAPolyhedron *pCDAPolyhedronNeg)
{
  ASSERT( pCDAPolyhedronNeg != NULL );
  m_listCDAPolyhedronNeg.InsertLast(pCDAPolyhedronNeg, 0);
}

inline VOID CDACell::AddZero(CDAPolyhedron *pCDAPolyhedronZero)
{
  ASSERT( pCDAPolyhedronZero != NULL );
  m_listCDAPolyhedronZero.InsertLast(pCDAPolyhedronZero, 0);
}

#endif // __CDACELL_H__
