#ifndef __CDAEDGE_H__
#define __CDAEDGE_H__

#include "cdamem.h"

#include "cdavertex.h"
class CDAVertex;
class CDAFace;

//////////////////////////////////////////////////////////////////////
// CDAEdge Class
//

class CDAEdge
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CDAEdge();
  virtual ~CDAEdge();
  virtual VOID Init(CDAVertex *pCDAVertex0, CDAVertex *pCDAVertex1);
  virtual VOID Uninit();

  virtual BOOL Initialized();
  virtual BOOL Valid();

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  CDAEdge *Copy();

  VOID Preprocess();

protected:

//////////////////////////////////////////////////
// Rendering
//
public:

//////////////////////////////////////////////////
// Output Functions
//
public:
  VOID Display();

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  inline VOID AddFace(CDAFace *pCDAFace);
  inline CDAFace *GetFaceNeighbor(CDAFace *pCDAFace);
  inline VOID SetFaceNeighbor(CDAFace *pCDAFace, 
			      CDAFace *pCDAFaceNeighbor);
  inline VOID SetVertexNeighbor(CDAVertex *pCDAVertex, 
				CDAVertex *pCDAVertexNeighbor);

  inline VOID Orient(CDAVertex *pCDAVertexFrom, CDAVertex *pCDAVertexTo);
  inline CDAVertex *GetVertexFrom();
  inline CDAVertex *GetVertexTo();
  inline VOID SetVerticesOnHull();

  inline BOOL GetOnHull();
  inline VOID SetOnHull(BOOL bOnHull);

  inline CDAEdge *GetCopy();

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CDAVertex *m_ppCDAVertex[2];
  CDAFace *m_ppCDAFace[2];

  BOOL m_bOnHull;

  CDAEdge *m_pCDAEdgeCopy;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CDAEdge::AddFace(CDAFace *pCDAFace)
{
  ASSERT( pCDAFace != NULL );
  ASSERT( m_ppCDAFace[0] == NULL || m_ppCDAFace[1] == NULL );

  if ( m_ppCDAFace[0] == NULL )
    {
      m_ppCDAFace[0] = pCDAFace;
    }
  else
    {
      ASSERT( m_ppCDAFace[1] == NULL );
      m_ppCDAFace[1] = pCDAFace;
    }
}

inline CDAFace *CDAEdge::GetFaceNeighbor(CDAFace *pCDAFace)
{
  ASSERT( m_ppCDAFace[0] == pCDAFace || m_ppCDAFace[1] == pCDAFace );

  if ( m_ppCDAFace[0] == pCDAFace )
    {
      return m_ppCDAFace[1];
    }
  else
    {
      ASSERT( m_ppCDAFace[1] == pCDAFace );
      return m_ppCDAFace[0];
    }
}

inline VOID CDAEdge::SetFaceNeighbor(CDAFace *pCDAFace, CDAFace *pCDAFaceNeighbor)
{
  ASSERT( m_ppCDAFace[0] == pCDAFace || m_ppCDAFace[1] == pCDAFace );

  if ( m_ppCDAFace[0] == pCDAFace )
    {
      m_ppCDAFace[1] = pCDAFaceNeighbor;
    }
  else
    {
      ASSERT( m_ppCDAFace[1] == pCDAFace );
      m_ppCDAFace[0] = pCDAFaceNeighbor;
    }  
}

inline VOID CDAEdge::SetVertexNeighbor(CDAVertex *pCDAVertex, 
				       CDAVertex *pCDAVertexNeighbor)
{
  ASSERT( m_ppCDAVertex[0] == pCDAVertex || 
	  m_ppCDAVertex[1] == pCDAVertex );

  if ( m_ppCDAVertex[0] == pCDAVertex )
    {
      m_ppCDAVertex[1] = pCDAVertexNeighbor;
    }
  else
    {
      ASSERT( m_ppCDAVertex[1] == pCDAVertex );
      m_ppCDAVertex[0] = pCDAVertexNeighbor;
    }  
}

inline VOID CDAEdge::Orient(CDAVertex *pCDAVertexFrom, CDAVertex *pCDAVertexTo)
{
  ASSERT( pCDAVertexFrom != NULL );
  ASSERT( pCDAVertexTo != NULL );
  ASSERT( (m_ppCDAVertex[0] == pCDAVertexFrom && 
	   m_ppCDAVertex[1] == pCDAVertexTo) ||
	  (m_ppCDAVertex[1] == pCDAVertexFrom && 
	   m_ppCDAVertex[0] == pCDAVertexTo) );

  if ( m_ppCDAVertex[1] == pCDAVertexFrom )
    {
      // We must swap the order
      ASSERT( m_ppCDAVertex[0] == pCDAVertexTo );
      m_ppCDAVertex[0] = pCDAVertexFrom; 
      m_ppCDAVertex[1] = pCDAVertexTo;
    }
  else
    {
      ASSERT( m_ppCDAVertex[0] == pCDAVertexFrom && 
	      m_ppCDAVertex[1] == pCDAVertexTo );
    }
}

inline CDAVertex *CDAEdge::GetVertexFrom()
{
  return m_ppCDAVertex[0];
}

inline CDAVertex *CDAEdge::GetVertexTo()
{
  return m_ppCDAVertex[1];
}

inline VOID CDAEdge::SetVerticesOnHull()
{
  ASSERT( m_ppCDAVertex[0] != NULL );
  ASSERT( m_ppCDAVertex[1] != NULL );

  m_ppCDAVertex[0]->SetOnHull(TRUE);
  m_ppCDAVertex[1]->SetOnHull(TRUE);
}

inline BOOL CDAEdge::GetOnHull()
{
  return m_bOnHull;
}

inline VOID CDAEdge::SetOnHull(BOOL bOnHull)
{
  m_bOnHull = bOnHull;
}

inline CDAEdge *CDAEdge::GetCopy()
{
  return m_pCDAEdgeCopy;
}

#endif // __CDAEDGE_H__

