#include "cdaface.h"

#include "cdavertex.h"
#include "cdaedge.h"

//////////////////////////////////////////////////////////////////////
// CDAFace Class
//

CDAFace::CDAFace()
{
  m_pCDAFaceUp = NULL;
  m_listCDAFaceDown.Init();
  m_listCDAVertex.Init();
  m_listCDAEdge.Init();

  m_heapCDAVertex.Init();
  m_pCHeapEntry = NULL;

  m_ppCDAFaceOperand[0] = NULL;
  m_ppCDAFaceOperand[1] = NULL;

  m_bOnHull = TRUE;

  m_pCDAFaceCopy = NULL;
}

CDAFace::~CDAFace()
{
  Uninit();
}

VOID CDAFace::Init()
{
  m_pCDAFaceUp = NULL;
  ASSERT( m_listCDAFaceDown.GetLength() == 0 );
  ASSERT( m_listCDAVertex.GetLength() == 0 );
  ASSERT( m_listCDAEdge.GetLength() == 0 );

  m_pCHeapEntry = NULL;

  m_ptCentroid.Null();
  m_vNormal.Null();
  m_ptCentroidRaw.Null();
  m_vNormalRaw.Null();

  m_ppCDAFaceOperand[0] = NULL;
  m_ppCDAFaceOperand[1] = NULL;

  m_bOnHull = TRUE;

  m_pCDAFaceCopy = NULL;
}

VOID CDAFace::Init(UINT_32 uiSize, CDAVertex **ppCDAVertex,
		   CDAEdge **ppCDAEdge, CListInt *plistCDAVertex)
{
  UINT_32 i;

  ASSERT( uiSize == 3 );
  ASSERT( ppCDAVertex != NULL );
  ASSERT( ppCDAEdge != NULL );
  ASSERT( plistCDAVertex != NULL );

  Init();

  for ( i = 0; i < uiSize; i++ )
    {
      ASSERT( ppCDAVertex[i] != NULL );
      ASSERT( ppCDAEdge[i] != NULL );

      m_listCDAVertex.InsertLast(ppCDAVertex[i], 0);
      m_listCDAEdge.InsertLast(ppCDAEdge[i], 0);
      ppCDAEdge[i]->AddFace(this);
    }

  Preprocess();

  //////////////////////////////////////////////////
  // Remove the vertices which make this face from the Vertex List
  // Remove the vertices outside this face from the Vertex List and
  // add them to the Face's Outside List

  CListIterInt iter;
  CDAVertex *pCDAVertexCurr;

  iter.Init(plistCDAVertex);

  for ( pCDAVertexCurr = (CDAVertex*)(iter.PeekFirst());
	pCDAVertexCurr != NULL; )
    {
      if ( CheckVertex(pCDAVertexCurr) )
	{
	  // We do not want to consider the Vertices which make up
	  // this face any more
	  pCDAVertexCurr = (CDAVertex*)(iter.PeekNextRemoveCurr());
	}
/*
      // We are going to put in the set of the face which it is furthest outside of
      else if ( CheckOutside(pCDAVertexCurr) )
	{
	  // Add it to our Outside Set
	  GetOutsideVertexList()->InsertLast(pCDAVertexCurr, 0);
	  pCDAVertexCurr = (CDAVertex*)(iter.PeekNextRemoveCurr());
	}
*/
      else
	{
	  // Put it back in the Vertex Pool to let other Face's examine it
	  pCDAVertexCurr = (CDAVertex*)(iter.PeekNext());
	}
    }
}

VOID CDAFace::Uninit()
{
  m_pCDAFaceUp = NULL;
  m_listCDAFaceDown.RemoveAll();
  m_listCDAVertex.RemoveAll();
  m_listCDAEdge.RemoveAll();

  ASSERT( m_heapCDAVertex.GetSize() == 0 );
  m_pCHeapEntry = NULL;

  m_ppCDAFaceOperand[0] = NULL;
  m_ppCDAFaceOperand[1] = NULL;

  m_bOnHull = TRUE;

  m_pCDAFaceCopy = NULL;
}

BOOL CDAFace::Initialized()
{
  return TRUE;
}

BOOL CDAFace::Valid()
{
  return TRUE;
}

CDAFace *CDAFace::Copy()
{
  CListIterInt iterCDAVertex, iterCDAEdge;
  CDAVertex *pCDAVertex;
  CDAEdge *pCDAEdge;

  m_pCDAFaceCopy = AllocateCDAFace();
  ASSERT( m_pCDAFaceCopy != NULL );
  m_pCDAFaceCopy->Init();

  iterCDAVertex.Init(&m_listCDAVertex);
  iterCDAEdge.Init(&m_listCDAEdge);
  for ( pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekFirst()),
	  pCDAEdge = (CDAEdge*)(iterCDAEdge.PeekFirst());
	pCDAVertex != NULL && pCDAEdge != NULL;
	pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekNext()),
	  pCDAEdge = (CDAEdge*)(iterCDAEdge.PeekNext()) )
    { 
      m_pCDAFaceCopy->m_listCDAVertex.InsertLast(pCDAVertex->GetCopy(), 0);
      m_pCDAFaceCopy->m_listCDAEdge.InsertLast(pCDAEdge->GetCopy(), 0);
      pCDAEdge->GetCopy()->AddFace(m_pCDAFaceCopy);
    }
  ASSERT( pCDAVertex == NULL && pCDAEdge == NULL );

  m_pCDAFaceCopy->Preprocess();

  ////////////////////////////////////////////////////////////
  // Also set up the bindings correctly

  if ( m_pCDAFaceUp != NULL )
    {
      m_pCDAFaceUp->GetCopy()->AddBoundDownFace(m_pCDAFaceCopy);
    }

  return m_pCDAFaceCopy;  
}

VOID CDAFace::Transform(const Matrix &/*mGroot_obj*/, const Matrix &/*mGobj_root*/)
{
  /*
  Vector vNormal;
  FLOAT fMag;

  m_ptCentroid = mGroot_obj * m_ptCentroid;
  m_vNormal = m_vNormal * mGobj_root;

  vNormal = m_vNormal;
  vNormal[3] = 0.0;
  fMag = vNormal.Magnitude();
  ASSERT( fMag > 0.0 );

  m_vNormal = m_vNormal / fMag;
  */
  Preprocess();
}

VOID CDAFace::Preprocess()
{
  CalculateNormal();
  CalculateCentroid();
  
  // Calculate the D coefficient of the Tangent Plane defined by the normal
  m_vNormal.CompletePlaneEquation(m_ptCentroid);
/*
  if ( m_listCDAVertex.GetLength() > 3 )
    {
      TestCoplanar();
    }
*/
}

FLOAT CDAFace::CalculateDistance(const Point &pt)
{
  Vector v;
  
  v = pt;
  v[W] = 1.0;
  
  return v.Dot(m_vNormal);
}

// In this function we calculate the plane equation which would be created if the
// vertex were inserted in the correct place in the contour list.  This is done
// by incrementally changing the Newell's Formula plane equation.  We calculate 
// the absolute projected areas, and then rely on the partial polygon's plane 
// equation to get the orientations (sign of components).  Then we plug in all 
// points into the plane equation and get the largest absolute distance from the 
// plane, if this distance is less than epsilon then the point is coplanar.

BOOL CDAFace::CheckCoplanar(CDAVertex *pCDAVertex)
{
  UINT_32 i;
  CListIterInt iterCDAVertex;
  CDAVertex *pCDAVertexStart, *pCDAVertexPrev, *pCDAVertexCurr;
  Vector v, v0, v1, vProj0, vProj1, vNormal;
  Point ptCentroid;
  FLOAT fDistance, fDistanceMax;

  ASSERT( pCDAVertex != NULL );

  ptCentroid = m_ptCentroidRaw + pCDAVertex->GetPoint();

  // The Raw Normal contains the signed projected areas of the face onto the 
  // coordinate planes
  vNormal[0] = ABS(m_vNormalRaw[0]);
  vNormal[1] = ABS(m_vNormalRaw[1]);
  vNormal[2] = ABS(m_vNormalRaw[2]);

  ASSERT( m_listCDAVertex.GetLength() >= 3 );

  iterCDAVertex.Init(&m_listCDAVertex);
  pCDAVertexPrev = (CDAVertex*)(iterCDAVertex.PeekLast());
  ASSERT( pCDAVertexPrev != NULL );
  for ( pCDAVertexCurr = (CDAVertex*)(iterCDAVertex.PeekFirst());
	pCDAVertexCurr != NULL;
	pCDAVertexCurr = (CDAVertex*)(iterCDAVertex.PeekNext()) )
    {
      v0 = (pCDAVertexPrev->GetPoint() - pCDAVertex->GetPoint());
      v1 = (pCDAVertexCurr->GetPoint() - pCDAVertex->GetPoint());
      
      for ( i = 0; i < 3; i++ )
	{
	  vProj0 = v0;
	  vProj1 = v1;

	  vProj0[i] = 0.0;
	  vProj1[i] = 0.0;

	  // The magnitude of the cross product is the area of the 
	  // parallelogram which is twice
	  // the area of the triangle.  But in the Newell's equation I removed 
	  // the division by 2, so I am not doing it here either.
	  v = vProj0 * vProj1;

	  vNormal[i] += v.Magnitude();
	  //vNormal[i] += (v.Magnitude() / 2.0);
	}

      pCDAVertexPrev = pCDAVertexCurr;
    }

  // Normalize our Normal and Centroid
  ptCentroid /= (m_listCDAVertex.GetLength() + 1);
  ptCentroid[W] = 1.0;

  for ( i = 0; i < 3; i++ )
    {
      if ( m_vNormalRaw[i] < 0.0 )
	{
	  vNormal[i] = -vNormal[i];
	}
    }
  if ( !vNormal.Normalize() )
    {
      return FALSE;
    }

  vNormal.CompletePlaneEquation(ptCentroid);

  // Check to see if all the vertices are close enough to the Newell best fit 
  // plane equation.
  v = pCDAVertex->GetPoint();
  v[W] = 1.0;
  fDistanceMax = ABS(vNormal.Dot(v));

  for ( pCDAVertexCurr = (CDAVertex*)(iterCDAVertex.PeekFirst());
	pCDAVertexCurr != NULL;
	pCDAVertexCurr = (CDAVertex*)(iterCDAVertex.PeekNext()) )
    {
      v = pCDAVertex->GetPoint();
      v[W] = 1.0;
      fDistance = ABS(vNormal.Dot(v));
      
      fDistanceMax = MAX(fDistance, fDistanceMax);
    }
/*
printf("JORDAN: Face %p: Vertex %p: %3.15f\n", this, pCDAVertex, fDistanceMax);
printf("  Normal Before: ");
m_vNormal.Display();
printf("  Normal After : ");
vNormal.Display();
*/
  return ( fDistanceMax <= CDA_EPSILON );
}

// Testing to see if the CheckCoplanar Function really works
VOID CDAFace::TestCoplanar()
{
  UINT_32 i;
  CListIterInt iterCDAVertex;
  CDAVertex *pCDAVertexStart, *pCDAVertexLast, *pCDAVertexPrev, *pCDAVertexCurr;
  Vector v, v0, v1, vProj0, vProj1, vNormal, vNormalRaw;
  Point ptCentroid, ptCentroidRaw;
  FLOAT fDistance, fDistanceMax;

  //////////////////////////////////////////////////////////////////////
  // START: Calculate Normal for the first n-1 vertices by Newell's Equation

  {
    CListIterInt iter;
    CDAVertex *pCDAVertex;
    Point ptStart, ptPrev, ptCurr;
    
    vNormal.Null();
    
    iter.Init(&m_listCDAVertex);
    
    pCDAVertexLast = (CDAVertex*)iter.PeekLast(); 
    ASSERT( pCDAVertexLast != NULL );

    pCDAVertex = (CDAVertex*)iter.PeekFirst(); 
    ASSERT( pCDAVertex != NULL );

    ptStart = pCDAVertex->GetPoint();
    ptPrev = ptStart;
    
    for ( pCDAVertex = (CDAVertex*)iter.PeekNext(); 
	  pCDAVertex != pCDAVertexLast;
	  pCDAVertex = (CDAVertex*)iter.PeekNext() )
      {
	ptCurr = pCDAVertex->GetPoint();
	
	vNormal[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
	vNormal[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
	vNormal[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);
	
	ptPrev = ptCurr;
      }
    ASSERT( (CDAVertex*)iter.PeekNext() == NULL );

    ptCurr = ptStart;
    vNormal[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
    vNormal[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
    vNormal[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);
    
    vNormalRaw = vNormal;
    
    if ( !vNormal.Normalize() )
      {
	printf("CDA Warning: Null normal because of zero area face");
      }
  }

  // END: Calculate Normal for the first n-1 vertices by Newell's Equation
  //////////////////////////////////////////////////////////////////////

  //////////////////////////////////////////////////////////////////////
  // START: Calculate Centroid for the first n-1 vertices

  {
    CListIterInt iter;
    UINT_32 i;
    CDAVertex *pCDAVertex;
    
    iter.Init(&m_listCDAVertex);
    
    pCDAVertexLast = (CDAVertex*)iter.PeekLast(); 
    ASSERT( pCDAVertexLast != NULL );   

    ptCentroid.Null();
    ptCentroid[W] = 0.0;
    
    i = 0;
    for ( pCDAVertex = (CDAVertex*)iter.PeekFirst();
	  pCDAVertex != pCDAVertexLast;
	  pCDAVertex = (CDAVertex*)iter.PeekNext() )
      {
	ptCentroid = ptCentroid + pCDAVertex->GetPoint();
	i++;
      }
    ASSERT( i == m_listCDAVertex.GetLength()-1 );
    ASSERT( (CDAVertex*)iter.PeekNext() == NULL );    
    
    ptCentroidRaw = ptCentroid;
  }

  // END: Calculate Centroid for the first n-1 vertices
  //////////////////////////////////////////////////////////////////////

  ptCentroid = ptCentroidRaw + pCDAVertexLast->GetPoint();

  // The Raw Normal contains the signed projected areas of the face onto the 
  // coordinate planes
  vNormal[0] = ABS(vNormalRaw[0]);
  vNormal[1] = ABS(vNormalRaw[1]);
  vNormal[2] = ABS(vNormalRaw[2]);

  ASSERT( m_listCDAVertex.GetLength() > 3 );

  iterCDAVertex.Init(&m_listCDAVertex);
  pCDAVertexLast = (CDAVertex*)(iterCDAVertex.PeekLast());
  ASSERT( pCDAVertexLast != NULL );
  pCDAVertexPrev = (CDAVertex*)(iterCDAVertex.PeekPrev());
  ASSERT( pCDAVertexPrev != NULL );
  for ( pCDAVertexCurr = (CDAVertex*)(iterCDAVertex.PeekFirst());
	pCDAVertexCurr != pCDAVertexLast;
	pCDAVertexCurr = (CDAVertex*)(iterCDAVertex.PeekNext()) )
    {
      v0 = (pCDAVertexPrev->GetPoint() - pCDAVertexLast->GetPoint());
      v1 = (pCDAVertexCurr->GetPoint() - pCDAVertexLast->GetPoint());
      
      for ( i = 0; i < 3; i++ )
	{
	  vProj0 = v0;
	  vProj1 = v1;

	  vProj0[i] = 0.0;
	  vProj1[i] = 0.0;

	  // The magnitude of the cross product is the area of the 
	  // parallelogram which is twice
	  // the area of the triangle.  But in the Newell's equation I removed 
	  // the division by 2, so I am not doing it here either.
	  v = vProj0 * vProj1;

	  vNormal[i] += v.Magnitude();
	  //vNormal[i] += (v.Magnitude() / 2.0);
	}

      pCDAVertexPrev = pCDAVertexCurr;
    }

  // Normalize our Normal and Centroid
  ptCentroidRaw = ptCentroid;
  ptCentroid /= m_listCDAVertex.GetLength();
  ptCentroid[W] = 1.0;

  for ( i = 0; i < 3; i++ )
    {
      if ( vNormalRaw[i] < 0.0 )
	{
	  vNormal[i] = -vNormal[i];
	}
    }

  vNormalRaw = vNormal / 2.0;

  ASSERT( vNormal.Normalize() );

  vNormal.CompletePlaneEquation(ptCentroid);

  printf("START: TestCoplanar\n");

  printf("  CentroidRaw: ");
  m_ptCentroidRaw.Display();
  printf("               ");
  ptCentroidRaw.Display();

  printf("  Centroid: ");
  m_ptCentroid.Display();
  printf("            ");
  ptCentroid.Display();

  printf("  NormalRaw: ");
  m_vNormalRaw.Display();
  printf("             ");
  vNormalRaw.Display();

  printf("  Normal: ");
  m_vNormal.Display();
  printf("          ");
  vNormal.Display();

  printf("END  : TestCoplanar\n");  
}

BOOL CDAFace::CheckOutside(CDAVertex *pCDAVertex)
{
  FLOAT fDistance, fMag;

  ASSERT( pCDAVertex != NULL );

  fDistance = CalculateDistance(pCDAVertex->GetPoint());
  // I don't think that this is a good idea
  //fMag = m_vNormal.Magnitude();
  //ASSERT( fMag >= 1.0 );
  //fDistance /= fMag;
/*
  // This is the bad coplanarity test
  if ( (fDistance >= 0.0 - CDA_EPSILON) &&
       (fDistance <= 0.0 + CDA_EPSILON) )
*/
  if ( CheckCoplanar(pCDAVertex) )
    {
      // This point may be the same as one of the vertices making up
      // this face so also store the dual information in one vertex
      // structure
      CListIterInt iter;
      CDAVertex *pCDAVertexCurr;
      Vector v, vCurr;
      FLOAT fMag, fMagCurr, fDot;

      v = pCDAVertex->GetPoint();
      fMag = v.Dot(v);

      iter.Init(&m_listCDAVertex);
      
      for ( pCDAVertexCurr = (CDAVertex*)iter.PeekFirst();
	    pCDAVertexCurr != NULL;
	    pCDAVertexCurr = (CDAVertex*)iter.PeekNext() )
	{
	  vCurr = pCDAVertexCurr->GetPoint();
	  fMagCurr = vCurr.Dot(vCurr);
	  fDot = v.Dot(vCurr);
	  if ( fDot == fMag && fDot == fMagCurr )
	    {
	      ASSERT( ( pCDAVertex->GetDual(0) == NULL &&
			pCDAVertex->GetDual(1) == NULL ) ||
		      ( pCDAVertex->GetDual(0) != NULL &&
			pCDAVertex->GetDual(1) == NULL ) ||
		      ( pCDAVertex->GetDual(0) == NULL &&
			pCDAVertex->GetDual(1) != NULL ) );

	      if ( pCDAVertex->GetDual(0) != NULL )
		{
		  pCDAVertexCurr->SetDual(0, pCDAVertex->GetDual(0));
		}
	      if ( pCDAVertex->GetDual(1) != NULL )
		{
		  pCDAVertexCurr->SetDual(1, pCDAVertex->GetDual(1));
		}
	      break;
	    }
	}

      return FALSE;
    }
/*
printf("CheckOutside: Face %p: Vertex %p: Distance = %3.15f\n", this, 
       pCDAVertex, fDistance);
*/
  return fDistance > 0.0; 
}

// Check if the vertex is one of the ones which define the face
// this is to combat any floating point errors we might have
BOOL CDAFace::CheckVertex(CDAVertex *pCDAVertex)
{
  CListIterInt iter;
  CDAVertex *pCDAVertexCurr;

  ASSERT( pCDAVertex != NULL );

  iter.Init(&m_listCDAVertex);

  for ( pCDAVertexCurr = (CDAVertex*)iter.PeekFirst();
	pCDAVertexCurr != NULL;
	pCDAVertexCurr = (CDAVertex*)iter.PeekNext() )
    {
      if ( pCDAVertex == pCDAVertexCurr )
	{
	  return TRUE;
	}
    }

  return FALSE;
}

VOID CDAFace::CalculateHorizon(CDAVertex *pCDAVertexEye, 
			       CHeap<VOID*, FLOAT> *pheapCDAFace,
			       CListInt *plistCDAVertex,
			       CListInt *plistCDAEdge,
			       CDAEdge *pCDAEdgeCrossed)
{
  CDAVertex *pCDAVertex, *pCDAVertexPrev;
  CDAEdge *pCDAEdge;
  CDAFace *pCDAFace;
  CListIterInt iterCDAEdge, iterCDAVertex;

  ASSERT( pCDAVertexEye != NULL );
  ASSERT( pheapCDAFace != NULL );
  ASSERT( plistCDAVertex != NULL );
  ASSERT( plistCDAEdge != NULL );

  if ( !GetOnHull() )
    {
      // We have already processed this face and it is marked for
      // removal

      // The first Face must still be on the Hull at this point
      ASSERT( pCDAEdgeCrossed != NULL );
      pCDAEdgeCrossed->SetOnHull(FALSE);

      return;
    }

  // Check if this face is back facing to the eye point
  if ( !CheckOutside(pCDAVertexEye) )
    {
      // The first Face must be visible
      ASSERT( pCDAEdgeCrossed != NULL );

      // The Crossed Edge is part of the Horizon Contour
      plistCDAEdge->InsertLast(pCDAEdgeCrossed, 0);

      // Set the neighbor pointer in the Crossed Edge to be NULL
      // because this face is temporarily unattached on that side
      pCDAEdgeCrossed->SetFaceNeighbor(this, NULL);

      // Orient the Edge to make building the cone to the horizon
      // easier later
      iterCDAVertex.Init(&m_listCDAVertex);
      iterCDAEdge.Init(&m_listCDAEdge);

      pCDAVertexPrev = (CDAVertex*)(iterCDAVertex.PeekFirst());
      pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekLast());

      for ( pCDAEdge = (CDAEdge*)(iterCDAEdge.PeekLast());
	    pCDAEdge != NULL;
	    pCDAEdge = (CDAEdge*)(iterCDAEdge.PeekPrev()) )
	{
	  ASSERT( pCDAVertex != NULL && pCDAVertexPrev != NULL );
	  ASSERT( pCDAVertex != pCDAVertexPrev );

	  if ( pCDAEdge == pCDAEdgeCrossed )
	    {
	      pCDAEdgeCrossed->Orient(pCDAVertexPrev, pCDAVertex);
	      break;
	    }

	  pCDAVertexPrev = pCDAVertex;
	  pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekPrev());
	}

      ASSERT( pCDAEdge != NULL );

      return;
    }

  //////////////////////////////////////////////////
  // The Face is visible to the Eye
  //

  // We are not on the Convex Hull so we should note that for the face
  // and its vertices ( NOTE: some of its vertices might still be on 
  // the convex Hull, but they will be caught later while processing the
  // horizon edges )
  SetOnHull(FALSE);
  if ( GetPriority() != NULL )
    {
      CDAFace *pCDAFace;

      pCDAFace = (CDAFace*)(pheapCDAFace->Remove(GetPriority()));
      ASSERT( pCDAFace == this );
      pCDAFace->SetPriority(NULL);
    }
  else
    {
      ASSERT( m_heapCDAVertex.GetSize() == 0 );
    }

  iterCDAVertex.Init(&m_listCDAVertex);
  for ( pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekFirst());
	pCDAVertex != NULL;
	pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekNext()))
    {
      pCDAVertex->SetOnHull(FALSE);
    }
   
  // Release our Outside Vertices
  for ( pCDAVertex = (CDAVertex*)(m_heapCDAVertex.RemoveLast());
	pCDAVertex != NULL;
	pCDAVertex = (CDAVertex*)(m_heapCDAVertex.RemoveLast()) )
    {
      plistCDAVertex->InsertLast(pCDAVertex, 0);
    }
   
  // Continue the Depth First Search through the faces by visiting
  // our neighbors
  
  iterCDAEdge.Init(&m_listCDAEdge);

  if ( pCDAEdgeCrossed == NULL )
    {
      // This is the initial visible face so we need to cheat
      pCDAEdgeCrossed = (CDAEdge*)(iterCDAEdge.PeekFirst());
      ASSERT( pCDAEdgeCrossed != NULL );
      ASSERT( pCDAEdgeCrossed->GetOnHull() );

      pCDAFace = pCDAEdgeCrossed->GetFaceNeighbor(this);
      pCDAFace->CalculateHorizon(pCDAVertexEye, 
				 pheapCDAFace,
				 plistCDAVertex,
				 plistCDAEdge,
				 pCDAEdgeCrossed);
    }
  else
    {
      // Set the Crossed Edge to not be on the Hull
      pCDAEdgeCrossed->SetOnHull(FALSE);

      // Find the Edge next to the Crossed Edge in CCW ordering of the Face 
      for ( pCDAEdge = (CDAEdge*)(iterCDAEdge.PeekFirst());
	    pCDAEdge != NULL;
	    pCDAEdge = (CDAEdge*)(iterCDAEdge.PeekNext()) )
	{
	  if ( pCDAEdge == pCDAEdgeCrossed )
	    {
	      break;
	    }
	}
    }

  // Iterate through the other edges
  for ( pCDAEdge = (CDAEdge*)(iterCDAEdge.PeekNext());
	pCDAEdge != NULL;
	pCDAEdge = (CDAEdge*)(iterCDAEdge.PeekNext()) )
    {
      ASSERT( pCDAEdge != pCDAEdgeCrossed );

      if ( pCDAEdge->GetOnHull() )
	{
	  pCDAFace = pCDAEdge->GetFaceNeighbor(this);
	  pCDAFace->CalculateHorizon(pCDAVertexEye, 
				     pheapCDAFace,
				     plistCDAVertex,
				     plistCDAEdge,
				     pCDAEdge);
	}
    }

  // Loop to the beginning and continue until we see the Crossed Edge
  // again
  for ( pCDAEdge = (CDAEdge*)(iterCDAEdge.PeekFirst());
	pCDAEdge != pCDAEdgeCrossed;
	pCDAEdge = (CDAEdge*)(iterCDAEdge.PeekNext()) )
    {
      ASSERT( pCDAEdge != NULL );

      if ( pCDAEdge->GetOnHull() )
	{
	  pCDAFace = pCDAEdge->GetFaceNeighbor(this);
	  pCDAFace->CalculateHorizon(pCDAVertexEye,
				     pheapCDAFace,
				     plistCDAVertex,
				     plistCDAEdge,
				     pCDAEdge);
	}
    }
}

// Calculate the Dual around some point ptCenter
// vTrans = ( -ptCenter[X], -ptCenter[X], -ptCenter[X], 1 )
// so we can take the dot product with our m_vNormal ( plane equation of 
// the face ) and it will give us the plane equation that we will then
// actually dualize.

Point CDAFace::CalculateDual(const Vector &vTrans)
{
  FLOAT fD;
  Point pt;
  Vector vNormal;

  // Translate the plane so that the Origin is at the center of Dualization
/*
  fD = vTrans.Dot(m_vNormal);

  fD = m_vNormal[3] - fD;

  ASSERT( fD != 0.0 );

  pt.Set(m_vNormal[X]/fD, m_vNormal[Y]/fD, m_vNormal[Z]/fD);
*/
  pt = m_ptCentroid + vTrans;
  vNormal = m_vNormalRaw;

  vNormal.CompletePlaneEquation(pt);
  pt.Set(vNormal[X]/vNormal[W], m_vNormal[Y]/vNormal[W], m_vNormal[Z]/vNormal[W]);

  return pt;
}

VOID CDAFace::ReplaceEdge(CDAEdge *pCDAEdge, UINT_32 uiEdgesNew,
			  CDAVertex **ppCDAVerticesNew,
			  CDAEdge **ppCDAEdgesNew,
			  CListInt *plistCDAVertex)
{
  CListIterInt iterCDAVertex, iterCDAEdge;
  UINT_32 i;
  Vector vEdge, vEdgeNew;
  CDAVertex *pCDAVertexLast, *pCDAVertexPrev, *pCDAVertexCurr, *pCDAVertexNext;
  CDAEdge *pCDAEdgePrev, *pCDAEdgeCurr, *pCDAEdgeNext;
  BOOL bFoundIt;
  FLOAT fDot;

  ASSERT( pCDAEdge != NULL );
  ASSERT( uiEdgesNew == 2 );
  ASSERT( ppCDAVerticesNew != NULL );
  ASSERT( ppCDAEdgesNew != NULL );
  ASSERT( plistCDAVertex != NULL );

  bFoundIt = FALSE;

  iterCDAVertex.Init(&m_listCDAVertex);
  iterCDAEdge.Init(&m_listCDAEdge);

  for ( pCDAVertexCurr = (CDAVertex*)(iterCDAVertex.PeekFirst()),
	  pCDAEdgeCurr = (CDAEdge*)(iterCDAEdge.PeekFirst());
	pCDAVertexCurr != NULL && pCDAEdgeCurr != NULL;
	pCDAVertexCurr = (CDAVertex*)(iterCDAVertex.PeekNext()),
	  pCDAEdgeCurr = (CDAEdge*)(iterCDAEdge.PeekNext()) )
    {
      if ( pCDAEdgeCurr == pCDAEdge )
	{
	  bFoundIt = TRUE;

	  // This is the one we are replacing

	  ASSERT( ppCDAVerticesNew[0] == pCDAVertexCurr ); 

	  // Temporarily remove the current vertex and edge
	  pCDAVertexPrev = (CDAVertex*)(iterCDAVertex.PeekPrevWrapRemoveCurr());
	  ASSERT( pCDAVertexPrev != NULL );
	  pCDAEdgePrev = (CDAEdge*)(iterCDAEdge.PeekPrevWrapRemoveCurr());
	  ASSERT( pCDAVertexPrev != NULL );
	  
	  // Check to see if the first new edge is colinear to edge it
	  // is adjacent to
	  vEdge = pCDAVertexCurr->GetPoint() - 
	    pCDAVertexPrev->GetPoint();
	  VERIFY( vEdge.Normalize() );
	  vEdgeNew = ppCDAVerticesNew[1]->GetPoint() - 
	    ppCDAVerticesNew[0]->GetPoint();
	  VERIFY( vEdgeNew.Normalize() );

	  fDot = vEdge.Dot(vEdgeNew);

	  //if ( fDot >= 1.0 )
	  if ( fDot == 1.0 )
	    {
	      // The edges are colinear
	      
	      // Remove the current vertex
	      pCDAVertexCurr->SetOnHull(FALSE);
	      pCDAEdgePrev->SetVertexNeighbor(pCDAVertexPrev, 
					      ppCDAVerticesNew[1]);
	      ppCDAEdgesNew[0]->SetOnHull(FALSE);
	    }
	  else
	    {
	      ASSERT( fDot < 1.0 );

	      // The Normal Case

	      // Insert the first new vertex and iterate to it
	      iterCDAVertex.InsertNext(pCDAVertexCurr, 0);
	      pCDAVertexCurr = (CDAVertex*)(iterCDAVertex.PeekNextWrap());
	      ASSERT( pCDAVertexCurr != NULL );
	      ASSERT( pCDAVertexCurr == ppCDAVerticesNew[0] );

	      iterCDAEdge.InsertNext(ppCDAEdgesNew[0], 0);
	      ppCDAEdgesNew[0]->AddFace(this);
	      pCDAEdgeCurr = (CDAEdge*)(iterCDAEdge.PeekNextWrap());
	      ASSERT( pCDAEdgeCurr != NULL );
	      ASSERT( pCDAEdgeCurr == ppCDAEdgesNew[0] );
	    }

	  pCDAVertexCurr = (CDAVertex*)(iterCDAVertex.PeekNextWrap());
	  ASSERT( pCDAVertexCurr != NULL );
	  ASSERT( pCDAVertexCurr == ppCDAVerticesNew[2] );
	  pCDAEdgeCurr = (CDAEdge*)(iterCDAEdge.PeekNextWrap());
	  ASSERT( pCDAEdgeCurr != NULL );

	  // Temporarily remove the next vertex and edge
	  pCDAVertexNext = (CDAVertex*)(iterCDAVertex.PeekNextWrapRemoveCurr());
	  ASSERT( pCDAVertexNext != NULL );
	  pCDAEdgeNext = (CDAEdge*)(iterCDAEdge.PeekNextWrapRemoveCurr());
	  ASSERT( pCDAEdgeNext != NULL );

	  // Check to see if the last new edge is colinear to edge it
	  // is adjacent to
	  vEdge = pCDAVertexNext->GetPoint() - 
	    pCDAVertexCurr->GetPoint();
	  VERIFY( vEdge.Normalize() );
	  vEdgeNew = ppCDAVerticesNew[2]->GetPoint() - 
	    ppCDAVerticesNew[1]->GetPoint();
	  VERIFY( vEdgeNew.Normalize() );

	  fDot = vEdge.Dot(vEdgeNew);

	  //if ( fDot >= 1.0 )
	  if ( fDot == 1.0 )
	    {
	      // The edges are colinear
	      
	      // Remove the current vertex
	      ppCDAEdgesNew[1]->SetOnHull(FALSE);
	      pCDAVertexCurr->SetOnHull(FALSE);

	      // Add in the new vertex and reinsert and modify the existing edge
	      iterCDAVertex.InsertPrev(ppCDAVerticesNew[1], 0);
	      iterCDAEdge.InsertPrev(pCDAEdgeCurr, 0);
	      pCDAEdgeCurr->SetVertexNeighbor(pCDAVertexNext, 
					      ppCDAVerticesNew[1]);

	    }
	  else
	    {
	      ASSERT( fDot < 1.0 );

	      // The Normal Case
	      iterCDAVertex.InsertPrev(ppCDAVerticesNew[1], 0);
	      iterCDAEdge.InsertPrev(ppCDAEdgesNew[1], 0);
	      ppCDAEdgesNew[1]->AddFace(this);

	      iterCDAVertex.InsertPrev(pCDAVertexCurr, 0);
	      iterCDAEdge.InsertPrev(pCDAEdgeCurr, 0);
	    }

	  // We are done looping around this face
	  break;
	}
    }
  
  ASSERT( bFoundIt );

  // Recalculate the centroid and plane equation
  Preprocess();

  iterCDAVertex.Init(plistCDAVertex);

  for ( pCDAVertexCurr = (CDAVertex*)(iterCDAVertex.PeekFirst());
	pCDAVertexCurr != NULL; )
    {
      if ( CheckVertex(pCDAVertexCurr) )
	{
	  // We do not want to consider the Vertices which make up
	  // this face any more
	  pCDAVertexCurr = (CDAVertex*)(iterCDAVertex.PeekNextRemoveCurr());
	}
/*
      // We are going to put in the set of the face which it is furthest outside of
      else if ( CheckOutside(pCDAVertexCurr) )
	{
	  // Add it to our Outside Set
	  GetOutsideVertexList()->InsertLast(pCDAVertexCurr, 0);
	  pCDAVertexCurr = (CDAVertex*)(iterCDAVertex.PeekNextRemoveCurr());
	}
*/
      else
	{
	  // Put it back in the Vertex Pool to let other Face's examine it
	  pCDAVertexCurr = (CDAVertex*)(iterCDAVertex.PeekNext());
	}
    }

  //////////////////////////////////////////////////
  // Put all points in this face's outside set on the unassigned list

  for ( pCDAVertexCurr = (CDAVertex*)(m_heapCDAVertex.RemoveLast());
	pCDAVertexCurr != NULL;
	pCDAVertexCurr = (CDAVertex*)(m_heapCDAVertex.RemoveLast()) )
    {
      // Put it back in the Vertex Pool to let other Face's examine it
      plistCDAVertex->InsertLast(pCDAVertexCurr, 0);
    }

}

VOID CDAFace::RecheckOutsideSet()
{
  /*
  CListIterInt iterCDAEdge;
  CDAVertex *pCDAVertexLast, *pCDAVertexPrev, *pCDAVertexCurr;
  CDAEdge *pCDAEdge;
  CDAFace *pCDAFace;

  iterCDAEdge.Init(&m_listCDAEdge);
  
  pCDAVertexLast = (CDAVertex*)(GetOutsideVertexList()->PeekLast());
  pCDAVertexPrev = NULL;
  for ( pCDAVertexCurr = (CDAVertex*)(GetOutsideVertexList()->PeekFirst());
	pCDAVertexPrev != pCDAVertexLast;
	pCDAVertexCurr = (CDAVertex*)(GetOutsideVertexList()->PeekFirst()) )
    {
      pCDAVertexCurr = (CDAVertex*)(GetOutsideVertexList()->RemoveFirst());

      if ( !CheckOutside(pCDAVertexCurr) )
	{
	  // This Vertex is no longer outside so let's try to put it
	  // in one of our neighbor's outside sets, if not then it is
	  // not on the convex hull
	  pCDAVertexCurr->SetOnHull(FALSE);
	  for ( pCDAEdge = (CDAEdge*)(iterCDAEdge.PeekFirst());
		pCDAEdge != NULL;
		pCDAEdge = (CDAEdge*)(iterCDAEdge.PeekNext()) )
	    {
	      pCDAFace = pCDAEdge->GetFaceNeighbor(this);
	      ASSERT( pCDAFace != NULL );
	      if ( pCDAFace->CheckOutside(pCDAVertexCurr) )
		{
		  pCDAVertexCurr->SetOnHull(TRUE);
		  pCDAFace->GetOutsideVertexList()->InsertLast(pCDAVertexCurr, 0);
		  break;
		}
	    }
	}
      else
	{
	  // It is still in the outside set
	  GetOutsideVertexList()->InsertLast(pCDAVertexCurr, 0);
	}

      pCDAVertexPrev = pCDAVertexCurr;
    }
    */
}

VOID CDAFace::Bind(const Vector &vTrans)
{
  CListIterInt iterCDAVertex, iterCDAVertexDual;
  CDAVertex *pCDAVertex, *ppCDAVertexDual[3];
  CDAFace *pCDAFace, *pCDAFaceTemp;
  UINT_32 uiCommonVertices;
/*
  printf("START: Face Bind %p: ", this);
  m_vNormal.Display();
*/
  ppCDAVertexDual[0] = NULL;
  ppCDAVertexDual[1] = NULL;

  // Find the Dual space vertex which corresponds to this face
  iterCDAVertex.Init(&m_listCDAVertex);
  pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekFirst());
  ASSERT( pCDAVertex != NULL );
  ASSERT( pCDAVertex->GetDual(1) == NULL );
  pCDAFace = pCDAVertex->GetDual(0);
  ASSERT( pCDAFace != NULL );
/*
  printf("    Vertex %p: ", pCDAVertex);
  pCDAVertex->GetPoint().Display();
  printf("      Dual Face %p: ", pCDAFace);
  pCDAFace->GetNormal().Display();
*/
  pCDAFaceTemp = pCDAFace;


//  printf("Iterating face's vertex count = %ld\n", pCDAFace->m_listCDAVertex.GetLength());
  iterCDAVertexDual.Init(&(pCDAFace->m_listCDAVertex));

  pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekNext());
  ASSERT( pCDAVertex != NULL );
  ASSERT( pCDAVertex->GetDual(1) == NULL );
  pCDAFace = pCDAVertex->GetDual(0);
  ASSERT( pCDAFace != NULL );
/*
  printf("    Vertex %p: ", pCDAVertex);
  pCDAVertex->GetPoint().Display();
  printf("      Dual Face %p: ", pCDAFace);
  pCDAFace->GetNormal().Display();
*/
  ASSERT( pCDAFace != pCDAFaceTemp );

  // Find the Vertices which these first two faces have in common
  // it should not be more than two

//printf("Search for Common Vertices: \n");

  uiCommonVertices = 0;
  for ( ppCDAVertexDual[uiCommonVertices] = (CDAVertex*)(iterCDAVertexDual.PeekFirst());
	ppCDAVertexDual[uiCommonVertices] != NULL;
	ppCDAVertexDual[uiCommonVertices] = (CDAVertex*)(iterCDAVertexDual.PeekNext()) )
    {
      if ( pCDAFace->CheckVertex(ppCDAVertexDual[uiCommonVertices]) )
	{
	  uiCommonVertices++;
	  if ( !( uiCommonVertices <= 2 ) )
	    {
/*
	      printf("Too many common vertices found between face %p and face %p\n", pCDAFaceTemp, pCDAFace);
	      printf("There must be colinear edges which are not being detected.\n");

CListIterInt iter;
CDAVertex *pCDAVertex;

printf("\n********************************\n");

printf("\nFace %p: ", pCDAFaceTemp);
pCDAFaceTemp->GetNormal().Display();
iter.Init(&(pCDAFaceTemp->m_listCDAVertex));
for ( pCDAVertex = (CDAVertex*)(iter.PeekFirst());
      pCDAVertex != NULL;
      pCDAVertex = (CDAVertex*)(iter.PeekNext()) )
  {
    printf("  Vertex %p: ", pCDAVertex);
    pCDAVertex->GetPoint().Display();
  }

printf("\nFace %p: ", pCDAFace);
pCDAFace->GetNormal().Display();
iter.Init(&(pCDAFace->m_listCDAVertex));
for ( pCDAVertex = (CDAVertex*)(iter.PeekFirst());
      pCDAVertex != NULL;
      pCDAVertex = (CDAVertex*)(iter.PeekNext()) )
  {
    printf("  Vertex %p: ", pCDAVertex);
    pCDAVertex->GetPoint().Display();
  }

printf("********************************\n");
*/
	    }
	  ASSERT( uiCommonVertices <= 2 );
	  /*
	  if ( uiCommonVertices == 2 )
	    {
	      break;
	    }
	    */
	}
    }
  if ( !( uiCommonVertices > 0 && uiCommonVertices < 3 ) )
    {
      printf("Bind Problem: Common Vertices = %ld\n", uiCommonVertices);
    }
  ASSERT( uiCommonVertices > 0 && uiCommonVertices < 3 );

  if ( uiCommonVertices == 2 )
    {
      // We need to look at the next face to narrow the selection down
      // to one vertex

      pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekNext());
      ASSERT( pCDAVertex != NULL );
      ASSERT( pCDAVertex->GetDual(1) == NULL );
      pCDAFace = pCDAVertex->GetDual(0);
      ASSERT( pCDAFace != NULL );
      
      if ( !pCDAFace->CheckVertex(ppCDAVertexDual[0]) )
	{
	  // It must be the other one
	  ASSERT( pCDAFace->CheckVertex(ppCDAVertexDual[1]) );
	  ppCDAVertexDual[0] = ppCDAVertexDual[1];
	}

      ppCDAVertexDual[1] = NULL;
      uiCommonVertices = 1;
    }
  ASSERT( ppCDAVertexDual[0] != NULL );
  ASSERT( uiCommonVertices == 1 );

  // Let's Make sure this vertex is in all the faces which are the
  // duals of the verices of this face
  for ( pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekNext());
	pCDAVertex != NULL;
	pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekNext()) )
    {
      ASSERT( pCDAVertex->GetDual(1) == NULL );

      pCDAFace = pCDAVertex->GetDual(0);
      ASSERT( pCDAFace != NULL );
/*
      printf("    Vertex %p: ", pCDAVertex);
      pCDAVertex->GetPoint().Display();
      printf("      Dual Face %p: ", pCDAFace);
      pCDAFace->GetNormal().Display();
*/
      ASSERT( pCDAFace->CheckVertex(ppCDAVertexDual[0]) );
    }

  // Ok now we have found the Dual of this face in the Dual space 
  // intersection polyhedron (the hard way), so now we can read off
  // the face binds from the vertex
  ASSERT( GetOperandFace(0) == NULL );
  SetOperandFace(0, ppCDAVertexDual[0]->GetDual(0));
  ASSERT( GetOperandFace(1) == NULL );
  SetOperandFace(1, ppCDAVertexDual[0]->GetDual(1));
/*
  printf("  Dual Vertex %p: ", ppCDAVertexDual[0]);
  ppCDAVertexDual[0]->GetPoint().Display();

  printf("  END: Face Bind %p: ", this);
  m_vNormal.Display();
  */
}

// This uses Newell's Algorithm for finding the best fit planar equation
// for a set of points in 3 space (pg. 476 Foley and van Dam)
// I modified it so that the normal would be positive for counter-clockwise
// polygons, and I removed the 1/2 coefficient because we are concerned 
// with relative areas, not absolute areas.

VOID CDAFace::CalculateNormal()
{
  CListIterInt iter;
  CDAVertex *pCDAVertex;
  Point ptStart, ptPrev, ptCurr;

  m_vNormal.Null();

  iter.Init(&m_listCDAVertex);

  pCDAVertex = (CDAVertex*)iter.PeekFirst(); 
  if ( pCDAVertex == NULL )
    {
      printf("CDA Warning: Null normal because no points for face");
      return;
    }
  ptStart = pCDAVertex->GetPoint();
  ptPrev = ptStart;

  for ( pCDAVertex = (CDAVertex*)iter.PeekNext(); 
	pCDAVertex != NULL;
	pCDAVertex = (CDAVertex*)iter.PeekNext() )
    {
      ptCurr = pCDAVertex->GetPoint();

      m_vNormal[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
      m_vNormal[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
      m_vNormal[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);

      ptPrev = ptCurr;
    }

  ptCurr = ptStart;
  m_vNormal[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
  m_vNormal[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
  m_vNormal[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);
  
  m_vNormalRaw = m_vNormal;

  if ( !m_vNormal.Normalize() )
    {
      printf("CDA Warning: Null normal because of zero area face");
    }
}

VOID CDAFace::CalculateCentroid()
{
  CListIterInt iter;
  UINT_32 i;
  CDAVertex *pCDAVertex;

  iter.Init(&m_listCDAVertex);

  m_ptCentroid.Null();
  m_ptCentroid[W] = 0.0;

  i = 0;
  for ( pCDAVertex = (CDAVertex*)iter.PeekFirst();
	pCDAVertex != NULL;
	pCDAVertex = (CDAVertex*)iter.PeekNext() )
    {
      m_ptCentroid = m_ptCentroid + pCDAVertex->GetPoint();
      i++;
    }

  m_ptCentroidRaw = m_ptCentroid;

  if ( i > 0 )
    {
      m_ptCentroid = m_ptCentroid / (FLOAT)i;
    }

  m_ptCentroid[W] = 1.0;
}

VOID CDAFace::CalculateBound()
{
/*
  CListIterInt iter;
  CDAPolyhedron *pCDAPolyhedron;
  BOOL bBoundDirty;

  CBound bound;
  
  iter.Init(&m_listCDAPolyhedron);
  
  for ( pCDAPolyhedron = (CDAPolyhedron*)iter.PeekFirst();
	pCDAPolyhedron != NULL;
	pCDAPolyhedron = (CDAPolyhedron*)iter.PeekNext() )
    {
      pCDAPolyhedron->InsertIntoBound(bound);
    }

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);
    }
*/
}

VOID CDAFace::Render(BOOL bPositive)
{
  CListIterInt iter;
  CDAFace *pCDAFaceDown;

  // Check if Bound Up, if so it is a hole in a different face so don't draw
  if ( m_pCDAFaceUp != NULL )
    {
      return;
    }

  if ( m_listCDAFaceDown.GetLength() > 0 )
    {
      gilEnable(GL_STENCIL_TEST); 

      // Clear the Stencil Buffer
      gilClearStencil(0x0);
      gilClear(GL_STENCIL_BUFFER_BIT);

      // Set OpenGL to draw to the Stencil Buffer
      
      gilColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
      gilDepthMask(GL_FALSE);
      
      gilStencilFunc(GL_ALWAYS, 0x1, 0x1);
      gilStencilOp(GL_REPLACE, GL_REPLACE, GL_REPLACE);
      
      iter.Init(&m_listCDAFaceDown);
      
      for ( pCDAFaceDown = (CDAFace*)iter.PeekFirst();
	    pCDAFaceDown != NULL;
	    pCDAFaceDown = (CDAFace*)iter.PeekNext() )
	{
	  if ( bPositive )
	    {
	      pCDAFaceDown->RenderGeometry();
	    }
	  else
	    {
	      pCDAFaceDown->RenderGeometryNegative();
	    }
	}
      
      gilStencilFunc(GL_NOTEQUAL, 0x1, 0x1);
      gilStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
      
      // Restore OpenGL to drawing to the Frame Buffer
      gilColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
      gilDepthMask(GL_TRUE);
    }

  if ( bPositive )
    {
      RenderGeometry();
    }
  else
    {
      RenderGeometryNegative();
    }

  gilDisable(GL_STENCIL_TEST); 
}

VOID CDAFace::RenderGeometry()
{
  CListIterInt iter;
  CDAVertex *pCDAVertex;

  gilNormal3fv(m_vNormal.Get());
  gilBegin(GL_POLYGON);
  {
    iter.Init(&m_listCDAVertex);
    
    for ( pCDAVertex = (CDAVertex*)iter.PeekFirst();
	  pCDAVertex != NULL;
	  pCDAVertex = (CDAVertex*)iter.PeekNext() )
      {
	pCDAVertex->Render();
      }
  }
  gilEnd();
}

VOID CDAFace::RenderGeometryNegative()
{
  CListIterInt iter;
  CDAVertex *pCDAVertex;
  Vector vNormal;

  vNormal = -m_vNormal;
  gilNormal3fv(vNormal.Get());
  gilBegin(GL_POLYGON);
  {
    iter.Init(&m_listCDAVertex);
    
    for ( pCDAVertex = (CDAVertex*)iter.PeekLast();
	  pCDAVertex != NULL;
	  pCDAVertex = (CDAVertex*)iter.PeekPrev() )
      {
	pCDAVertex->Render();
      }
  }
  gilEnd();
}

VOID CDAFace::RenderNormals(BOOL bPositive)
{
  Vector vNormal;
  Point pt;

  // Check if Bound Up, if so it is a hole in a different face so don't draw
  if ( m_pCDAFaceUp != NULL )
    {
      return;
    }

  if ( bPositive )
    {
      vNormal = m_vNormal;
    }
  else
    {
      vNormal = -m_vNormal;
    }
  vNormal[3] = 0.0;

  gilBegin(GL_LINES);
  {    
    gilVertex3fv(m_ptCentroid.Get());
    pt = m_ptCentroid + vNormal;
    gilVertex3fv(pt.Get());
  }
  gilEnd();
  
}

VOID CDAFace::Display()
{
  CListIterInt iter;
  CDAVertex *pCDAVertex;
  CDAEdge *pCDAEdge;
  CDAFace *pCDAFace;

  printf("    START: Face %p: ", this);
  m_vNormal.Display();

  printf("      START: Vertices\n");
  iter.Init(&m_listCDAVertex);
  
  for ( pCDAVertex = (CDAVertex*)iter.PeekFirst();
	pCDAVertex != NULL;
	pCDAVertex = (CDAVertex*)iter.PeekNext() )
    {
      printf("    ");
      pCDAVertex->Display();
    }
  printf("      END:   Vertices\n");

  printf("      START: Neighboring Faces\n");
  iter.Init(&m_listCDAEdge);

  for ( pCDAEdge = (CDAEdge*)iter.PeekFirst();
	pCDAEdge != NULL;
	pCDAEdge = (CDAEdge*)iter.PeekNext() )
    {
      printf("        Edge %p => ", pCDAEdge);
      pCDAFace = pCDAEdge->GetFaceNeighbor(this);
      if ( pCDAFace == NULL )
	{
	  printf("NULL\n");
	}
      else
	{
	  printf("Face %p: ", pCDAFace);
	  pCDAFace->GetNormal().Display();
	}
    }
  printf("      END:   Neighboring Faces\n");

  printf("      START: Face Bindings\n");
  iter.Init(&m_listCDAFaceDown);

  if ( m_pCDAFaceUp == NULL )
    {
      printf("        UP => NULL\n");
    }
  else
    {
      printf("        UP => %p: ", m_pCDAFaceUp);
      m_pCDAFaceUp->GetNormal().Display();
    }

  for ( pCDAFace = (CDAFace*)iter.PeekFirst();
	pCDAFace != NULL;
	pCDAFace = (CDAFace*)iter.PeekNext() )
    {
      printf("        DOWN => %p: ", pCDAFace);
      pCDAFace->GetNormal().Display();
    }
  printf("      END  : Face Bindings\n");

  printf("    END  : Face %p: ", this);
  m_vNormal.Display();
}





