#ifndef __CDAFACE_H__
#define __CDAFACE_H__

#include "heap.h"
#include "cdamem.h"

//#define CDA_EPSILON 0.000000000
//#define CDA_EPSILON 0.0000000001

//#define CDA_EPSILON 0.000000001
// This one caused a problem with the face's outside sets
//#define CDA_EPSILON 0.00000001

//#define CDA_EPSILON 0.0000001
// 0.0810 Degrees
//#define CDA_EPSILON 0.000001
// 0.1403 Degrees -- this fixes the duality bug 
//#define CDA_EPSILON 0.000003
// ???? 0.1403 Degrees -- this fixes the duality bug 
//#define CDA_EPSILON 0.00005
#define CDA_EPSILON 0.00025

// 0.2562 Degrees
//#define CDA_EPSILON 0.00001
//#define CDA_EPSILON 0.001
//#define CDA_EPSILON 0.23

class CDAVertex;
class CDAEdge;
class CDAPolyhedron;

//////////////////////////////////////////////////////////////////////
// CDAFace Class
//

class CDAFace
{
//////////////////////////////////////////////////
// Friends
//
public:
  friend CDAPolyhedron;

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CDAFace();
  virtual ~CDAFace();
  virtual VOID Init();
  VOID Init(UINT_32 uiSize, CDAVertex **ppCDAVertex,
	    CDAEdge **ppCDAEdge, CListInt *plistCDAVertex);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  CDAFace *Copy();
  VOID Transform(const Matrix &mGroot_obj, const Matrix &mGobj_root);

  VOID Preprocess();
  FLOAT CalculateDistance(const Point &pt);
  BOOL CheckCoplanar(CDAVertex *pCDAVertex);
  VOID TestCoplanar();
  BOOL CheckOutside(CDAVertex *pCDAVertex);
  BOOL CheckVertex(CDAVertex *pCDAVertex);
  VOID CalculateHorizon(CDAVertex *pCDAVertexEye, 
                        CHeap<VOID*, FLOAT> *pheapCDAFace,
                        CListInt *plistCDAVertex,
			CListInt *plistCDAEdge,
			CDAEdge *pCDAEdgeCrossed);
  Point CalculateDual(const Vector &vTrans);

  VOID ReplaceEdge(CDAEdge *pCDAEdge, UINT_32 uiEdgesNew,
		   CDAVertex **ppCDAVerticesNew, 
		   CDAEdge **ppCDAEdgesNew,
		   CListInt *plistCDAVertex);

  VOID RecheckOutsideSet();

  // This is to bind a face of an intersection Polyhedron to the face
  // of the operand polyhedron which created it
  VOID Bind(const Vector &vTrans);

protected:
  VOID CalculateNormal();
  VOID CalculateCentroid();
  virtual VOID CalculateBound();

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID Render(BOOL bPositive);
  VOID RenderGeometry();
  VOID RenderGeometryNegative();
  virtual VOID RenderNormals(BOOL bPositive);

//////////////////////////////////////////////////
// Output Functions
//
public:
  VOID Display();

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  inline VOID AddVertex(CDAVertex *pCDAVertex);
  inline VOID SetBoundUpFace(CDAFace *pCDAFace);
  inline VOID AddBoundDownFace(CDAFace *pCDAFace);
  inline BOOL GetOnHull();
  inline VOID SetOnHull(BOOL bOnHull);

  inline const Vector &GetNormal();
  inline const Point &GetCentroid();

  inline CDAFace *GetOperandFace(UINT_32 uiOperand);
  inline VOID SetOperandFace(UINT_32 uiOperand, CDAFace *pCDAFace);

  inline CDAFace *GetCopy();


//  inline CListInt *GetOutsideVertexList();
  inline VOID AddOutsideVertex(FLOAT fDistance, CDAVertex *pCDAVertex);
  inline CDAVertex *PeekFurthestOutsideVertex(FLOAT *pfDistance = NULL);
  inline CDAVertex *RemoveFurthestOutsideVertex(FLOAT *pfDistance = NULL);

  inline CHeapEntry<VOID*, FLOAT> *GetPriority();
  inline VOID SetPriority(CHeapEntry<VOID*, FLOAT> *pCHeapEntry);

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CDAFace *m_pCDAFaceUp;
  CListInt m_listCDAFaceDown; // This is also used as the Outside Set of Vertices
			      // in the Convex Hull Procedure, which can only
			      // be called while initializing

  CListInt m_listCDAVertex;
  CListInt m_listCDAEdge;

  CHeap<VOID*, FLOAT> m_heapCDAVertex;
  CHeapEntry<VOID*, FLOAT> *m_pCHeapEntry;

  Point m_ptCentroid;
  Vector m_vNormal;
  Point m_ptCentroidRaw;
  Vector m_vNormalRaw;

  CDAFace *m_ppCDAFaceOperand[2];

  BOOL m_bOnHull;

  CDAFace *m_pCDAFaceCopy;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CDAFace::AddVertex(CDAVertex *pCDAVertex)
{
  ASSERT( pCDAVertex != NULL );
  m_listCDAVertex.InsertLast(pCDAVertex, 0);
}

inline VOID CDAFace::SetBoundUpFace(CDAFace *pCDAFace)
{
  ASSERT( pCDAFace != NULL );
  ASSERT( m_pCDAFaceUp == NULL );

  m_pCDAFaceUp = pCDAFace;
}

inline VOID CDAFace::AddBoundDownFace(CDAFace *pCDAFace)
{
  ASSERT( pCDAFace != NULL );

  pCDAFace->SetBoundUpFace(this);

  m_listCDAFaceDown.InsertLast(pCDAFace, 0);
}
/*
inline CListInt *CDAFace::GetOutsideVertexList()
{
  return &m_listCDAFaceDown;
}
*/
inline VOID CDAFace::AddOutsideVertex(FLOAT fDistance, CDAVertex *pCDAVertex)
{
  ASSERT( fDistance > 0.0 );
  ASSERT( pCDAVertex != NULL );

  m_heapCDAVertex.Insert(fDistance, pCDAVertex);
}

inline CDAVertex *CDAFace::PeekFurthestOutsideVertex(FLOAT *pfDistance/* = NULL*/)
{
  return (CDAVertex*)(m_heapCDAVertex.PeekTop(pfDistance));
}

inline CDAVertex *CDAFace::RemoveFurthestOutsideVertex(FLOAT *pfDistance/* = NULL*/)
{
  return (CDAVertex*)(m_heapCDAVertex.RemoveTop(pfDistance));
}

inline CHeapEntry<VOID*, FLOAT> *CDAFace::GetPriority()
{
  return m_pCHeapEntry;
}

inline VOID CDAFace::SetPriority(CHeapEntry<VOID*, FLOAT> *pCHeapEntry)
{
  m_pCHeapEntry = pCHeapEntry;
}

inline BOOL CDAFace::GetOnHull()
{
  return m_bOnHull;
}

inline VOID CDAFace::SetOnHull(BOOL bOnHull)
{
  m_bOnHull = bOnHull;
}

inline const Vector &CDAFace::GetNormal()
{
  return m_vNormal;
}

inline const Point &CDAFace::GetCentroid()
{
  return m_ptCentroid;
}

inline CDAFace *CDAFace::GetOperandFace(UINT_32 uiOperand)
{
  ASSERT( uiOperand < 2 );
  return m_ppCDAFaceOperand[uiOperand];
}

inline VOID CDAFace::SetOperandFace(UINT_32 uiOperand, CDAFace *pCDAFace)
{
  ASSERT( uiOperand < 2 );
  ASSERT( m_ppCDAFaceOperand[uiOperand] == NULL || 
	  m_ppCDAFaceOperand[uiOperand] == pCDAFace );
  m_ppCDAFaceOperand[uiOperand] = pCDAFace;
}

inline CDAFace *CDAFace::GetCopy()
{
  return m_pCDAFaceCopy;
}

#endif // __CDAFACE_H__
