#include "cdamem.h"
#include "cdavertex.h"
#include "cdaedge.h"
#include "cdaface.h"
#include "cdapolyhedron.h"
#include "cdacell.h"
#include "cda.h"

static UINT_32 s_uiCDAObjectCount = 0;

static UINT_32 s_uiNewedCDAVertices = 0;
static UINT_32 s_uiAllocatedCDAVertices = 0;
static CListInt *s_plistCDAVerticesFree = NULL;

static UINT_32 s_uiNewedCDAEdges = 0;
static UINT_32 s_uiAllocatedCDAEdges = 0;
static CListInt *s_plistCDAEdgesFree = NULL;

static UINT_32 s_uiNewedCDAFaces = 0;
static UINT_32 s_uiAllocatedCDAFaces = 0;
static CListInt *s_plistCDAFacesFree = NULL;

static UINT_32 s_uiNewedCDAPolyhedra = 0;
static UINT_32 s_uiAllocatedCDAPolyhedra = 0;
static CListInt *s_plistCDAPolyhedraFree = NULL;

static UINT_32 s_uiNewedCDACells = 0;
static UINT_32 s_uiAllocatedCDACells = 0;
static CListInt *s_plistCDACellsFree = NULL;

static UINT_32 s_uiNewedCDAs = 0;
static UINT_32 s_uiAllocatedCDAs = 0;
static CListInt *s_plistCDAsFree = NULL;

static VOID InitCDAMemory()
{
  ASSERT( s_uiCDAObjectCount == 0 );

  ASSERT( s_plistCDAVerticesFree == NULL );
  s_plistCDAVerticesFree = new CListInt;
  ASSERT( s_plistCDAVerticesFree != NULL );
  s_plistCDAVerticesFree->Init();
  
  ASSERT( s_plistCDAEdgesFree == NULL );
  s_plistCDAEdgesFree = new CListInt;
  ASSERT( s_plistCDAEdgesFree != NULL );
  s_plistCDAEdgesFree->Init();
  
  ASSERT( s_plistCDAFacesFree == NULL );
  s_plistCDAFacesFree = new CListInt;
  ASSERT( s_plistCDAFacesFree != NULL );
  s_plistCDAFacesFree->Init();
  
  ASSERT( s_plistCDAPolyhedraFree == NULL );
  s_plistCDAPolyhedraFree = new CListInt;
  ASSERT( s_plistCDAPolyhedraFree != NULL );
  s_plistCDAPolyhedraFree->Init();

  ASSERT( s_plistCDACellsFree == NULL );
  s_plistCDACellsFree = new CListInt;
  ASSERT( s_plistCDACellsFree != NULL );
  s_plistCDACellsFree->Init();
  
  ASSERT( s_plistCDAsFree == NULL );
  s_plistCDAsFree = new CListInt;
  ASSERT( s_plistCDAsFree != NULL );
  s_plistCDAsFree->Init();
}

static VOID UninitCDAMemory()
{
  ASSERT( s_uiCDAObjectCount == 0 );

  // Clean up our free list of CDA Vertices
  ASSERT( s_uiAllocatedCDAVertices == 0 );
  if ( s_plistCDAVerticesFree != NULL )
    {
      CDAVertex *pCDAVertex;
      
      for ( pCDAVertex = (CDAVertex*)(s_plistCDAVerticesFree->RemoveFirst());
	    pCDAVertex != NULL;
	    pCDAVertex = (CDAVertex*)(s_plistCDAVerticesFree->RemoveFirst()) )
	{
	  delete pCDAVertex;
	}

      delete s_plistCDAVerticesFree;
      s_plistCDAVerticesFree = NULL;
    }

  // Clean up our free list of CDA Edges
  ASSERT( s_uiAllocatedCDAEdges == 0 );
  if ( s_plistCDAEdgesFree != NULL )
    {
      CDAEdge *pCDAEdge;
      
      for ( pCDAEdge = (CDAEdge*)(s_plistCDAEdgesFree->RemoveFirst());
	    pCDAEdge != NULL;
	    pCDAEdge = (CDAEdge*)(s_plistCDAEdgesFree->RemoveFirst()) )
	{
	  delete pCDAEdge;
	}

      delete s_plistCDAEdgesFree;
      s_plistCDAEdgesFree = NULL;
    }

  // Clean up our free list of Faces
  ASSERT( s_uiAllocatedCDAFaces == 0 );
  if ( s_plistCDAFacesFree != NULL )
    {
      CDAFace *pCDAFace;
      
      for ( pCDAFace = (CDAFace*)(s_plistCDAFacesFree->RemoveFirst());
	    pCDAFace != NULL;
	    pCDAFace = (CDAFace*)(s_plistCDAFacesFree->RemoveFirst()) )
	{
	  delete pCDAFace;
	}

      delete s_plistCDAFacesFree;
      s_plistCDAFacesFree = NULL;
    }
 
  // Clean up our free list of CDA Polyhedra
  ASSERT( s_uiAllocatedCDAPolyhedra == 0 );
  if ( s_plistCDAPolyhedraFree != NULL )
    {
      CDAPolyhedron *pCDAPolyhedron;
      
      for ( pCDAPolyhedron = (CDAPolyhedron*)(s_plistCDAPolyhedraFree->RemoveFirst());
	    pCDAPolyhedron != NULL;
	    pCDAPolyhedron = (CDAPolyhedron*)(s_plistCDAPolyhedraFree->RemoveFirst()) )
	{
	  delete pCDAPolyhedron;
	}

      delete s_plistCDAPolyhedraFree;
      s_plistCDAPolyhedraFree = NULL;
    }

  // Clean up our free list of CDA Cells
  ASSERT( s_uiAllocatedCDACells == 0 );
  if ( s_plistCDACellsFree != NULL )
    {
      CDACell *pCDACell;
      
      for ( pCDACell = (CDACell*)(s_plistCDACellsFree->RemoveFirst());
	    pCDACell != NULL;
	    pCDACell = (CDACell*)(s_plistCDACellsFree->RemoveFirst()) )
	{
	  delete pCDACell;
	}

      delete s_plistCDACellsFree;
      s_plistCDACellsFree = NULL;
    }

  // Clean up our free list of CDAs
  ASSERT( s_uiAllocatedCDAs == 0 );
  if ( s_plistCDAsFree != NULL )
    {
      CDA *pCDA;
      
      for ( pCDA = (CDA*)(s_plistCDAsFree->RemoveFirst());
	    pCDA != NULL;
	    pCDA = (CDA*)(s_plistCDAsFree->RemoveFirst()) )
	{
	  delete pCDA;
	}

      delete s_plistCDAsFree;
      s_plistCDAsFree = NULL;
    }
}      

CDAVertex *AllocateCDAVertex()
{
  CDAVertex *pCDAVertex;

  if ( s_uiCDAObjectCount == 0 )
    {
      InitCDAMemory();
    }
  s_uiCDAObjectCount++;

  ASSERT( s_plistCDAVerticesFree != NULL );

  pCDAVertex = (CDAVertex*)(s_plistCDAVerticesFree->RemoveFirst());
  if ( pCDAVertex == NULL )
    {
      ASSERT( s_uiNewedCDAVertices == s_uiAllocatedCDAVertices );
      s_uiNewedCDAVertices++;
      pCDAVertex = new CDAVertex;
      ASSERT( pCDAVertex != NULL );
    }
  ASSERT( s_uiAllocatedCDAVertices < s_uiNewedCDAVertices );

  s_uiAllocatedCDAVertices++;

  return pCDAVertex;
}

VOID DeallocateCDAVertex(CDAVertex *pCDAVertex)
{
  ASSERT( pCDAVertex != NULL );
  ASSERT( s_plistCDAVerticesFree != NULL );
  
  ASSERT( s_uiAllocatedCDAVertices > 0 );

  pCDAVertex->Uninit();
  s_plistCDAVerticesFree->InsertFirst(pCDAVertex, 0);
  s_uiAllocatedCDAVertices--;
  ASSERT( s_uiAllocatedCDAVertices < s_uiNewedCDAVertices );

  s_uiCDAObjectCount--;
  if ( s_uiCDAObjectCount == 0 )
    {
      UninitCDAMemory();
    }
}

CDAEdge *AllocateCDAEdge()
{
  CDAEdge *pCDAEdge;

  if ( s_uiCDAObjectCount == 0 )
    {
      InitCDAMemory();
    }
  s_uiCDAObjectCount++;

  ASSERT( s_plistCDAEdgesFree != NULL );

  pCDAEdge = (CDAEdge*)(s_plistCDAEdgesFree->RemoveFirst());
  if ( pCDAEdge == NULL )
    {
      ASSERT( s_uiNewedCDAEdges == s_uiAllocatedCDAEdges );
      s_uiNewedCDAEdges++;
      pCDAEdge = new CDAEdge;
      ASSERT( pCDAEdge != NULL );
    }
  ASSERT( s_uiAllocatedCDAEdges < s_uiNewedCDAEdges );

  s_uiAllocatedCDAEdges++;

  return pCDAEdge;
}

VOID DeallocateCDAEdge(CDAEdge *pCDAEdge)
{
  ASSERT( pCDAEdge != NULL );
  ASSERT( s_plistCDAEdgesFree != NULL );
  
  ASSERT( s_uiAllocatedCDAEdges > 0 );

  pCDAEdge->Uninit();
  s_plistCDAEdgesFree->InsertFirst(pCDAEdge, 0);
  s_uiAllocatedCDAEdges--;
  ASSERT( s_uiAllocatedCDAEdges < s_uiNewedCDAEdges );

  s_uiCDAObjectCount--;
  if ( s_uiCDAObjectCount == 0 )
    {
      UninitCDAMemory();
    }
}

CDAFace *AllocateCDAFace()
{
  CDAFace *pCDAFace;

  if ( s_uiCDAObjectCount == 0 )
    {
      InitCDAMemory();
    }
  s_uiCDAObjectCount++;

  ASSERT( s_plistCDAFacesFree != NULL );

  pCDAFace = (CDAFace*)(s_plistCDAFacesFree->RemoveFirst());
  if ( pCDAFace == NULL )
    {
      ASSERT( s_uiNewedCDAFaces == s_uiAllocatedCDAFaces );
      s_uiNewedCDAFaces++;
      pCDAFace = new CDAFace;
      ASSERT( pCDAFace != NULL );
    }
  ASSERT( s_uiAllocatedCDAFaces < s_uiNewedCDAFaces );

  s_uiAllocatedCDAFaces++;

  return pCDAFace;
}

VOID DeallocateCDAFace(CDAFace *pCDAFace)
{
  ASSERT( pCDAFace != NULL );
  ASSERT( s_plistCDAFacesFree != NULL );
  
  ASSERT( s_uiAllocatedCDAFaces > 0 );

  pCDAFace->Uninit();
  s_plistCDAFacesFree->InsertFirst(pCDAFace, 0);
  s_uiAllocatedCDAFaces--;
  ASSERT( s_uiAllocatedCDAFaces < s_uiNewedCDAFaces );

  s_uiCDAObjectCount--;
  if ( s_uiCDAObjectCount == 0 )
    {
      UninitCDAMemory();
    }
}

CDAPolyhedron *AllocateCDAPolyhedron()
{
  CDAPolyhedron *pCDAPolyhedron;

  if ( s_uiCDAObjectCount == 0 )
    {
      InitCDAMemory();
    }
  s_uiCDAObjectCount++;

  ASSERT( s_plistCDAPolyhedraFree != NULL );

  pCDAPolyhedron = (CDAPolyhedron*)(s_plistCDAPolyhedraFree->RemoveFirst());
  if ( pCDAPolyhedron == NULL )
    {
      ASSERT( s_uiNewedCDAPolyhedra == s_uiAllocatedCDAPolyhedra );
      s_uiNewedCDAPolyhedra++;
      pCDAPolyhedron = new CDAPolyhedron;
      ASSERT( pCDAPolyhedron != NULL );
    }
  ASSERT( s_uiAllocatedCDAPolyhedra < s_uiNewedCDAPolyhedra );

  s_uiAllocatedCDAPolyhedra++;

  return pCDAPolyhedron;
}

VOID DeallocateCDAPolyhedron(CDAPolyhedron *pCDAPolyhedron)
{
  ASSERT( pCDAPolyhedron != NULL );
  ASSERT( s_plistCDAPolyhedraFree != NULL );
  
  ASSERT( s_uiAllocatedCDAPolyhedra > 0 );

  pCDAPolyhedron->Uninit();
  s_plistCDAPolyhedraFree->InsertFirst(pCDAPolyhedron, 0);
  s_uiAllocatedCDAPolyhedra--;
  ASSERT( s_uiAllocatedCDAPolyhedra < s_uiNewedCDAPolyhedra );

  s_uiCDAObjectCount--;
  if ( s_uiCDAObjectCount == 0 )
    {
      UninitCDAMemory();
    }
}

CDACell *AllocateCDACell()
{
  CDACell *pCDACell;

  if ( s_uiCDAObjectCount == 0 )
    {
      InitCDAMemory();
    }
  s_uiCDAObjectCount++;

  ASSERT( s_plistCDACellsFree != NULL );

  pCDACell = (CDACell*)(s_plistCDACellsFree->RemoveFirst());
  if ( pCDACell == NULL )
    {
      ASSERT( s_uiNewedCDACells == s_uiAllocatedCDACells );
      s_uiNewedCDACells++;
      pCDACell = new CDACell;
      ASSERT( pCDACell != NULL );
    }
  ASSERT( s_uiAllocatedCDACells < s_uiNewedCDACells );

  s_uiAllocatedCDACells++;

  return pCDACell;
}

VOID DeallocateCDACell(CDACell *pCDACell)
{
  ASSERT( pCDACell != NULL );
  ASSERT( s_plistCDACellsFree != NULL );
  
  ASSERT( s_uiAllocatedCDACells > 0 );

  pCDACell->Uninit();
  s_plistCDACellsFree->InsertFirst(pCDACell, 0);
  s_uiAllocatedCDACells--;
  ASSERT( s_uiAllocatedCDACells < s_uiNewedCDACells );

  s_uiCDAObjectCount--;
  if ( s_uiCDAObjectCount == 0 )
    {
      UninitCDAMemory();
    }
}

CDA *AllocateCDA()
{
  CDA *pCDA;

  if ( s_uiCDAObjectCount == 0 )
    {
      InitCDAMemory();
    }
  s_uiCDAObjectCount++;

  ASSERT( s_plistCDAsFree != NULL );

  pCDA = (CDA*)(s_plistCDAsFree->RemoveFirst());
  if ( pCDA == NULL )
    {
      ASSERT( s_uiNewedCDAs == s_uiAllocatedCDAs );
      s_uiNewedCDAs++;
      pCDA = new CDA;
      ASSERT( pCDA != NULL );
    }
  ASSERT( s_uiAllocatedCDAs < s_uiNewedCDAs );

  s_uiAllocatedCDAs++;

  return pCDA;
}

VOID DeallocateCDA(CDA *pCDA)
{
  ASSERT( pCDA != NULL );
  ASSERT( s_plistCDAsFree != NULL );
  
  ASSERT( s_uiAllocatedCDAs > 0 );

  pCDA->Uninit();
  s_plistCDAsFree->InsertFirst(pCDA, 0);
  s_uiAllocatedCDAs--;
  ASSERT( s_uiAllocatedCDAs < s_uiNewedCDAs );

  s_uiCDAObjectCount--;
  if ( s_uiCDAObjectCount == 0 )
    {
      UninitCDAMemory();
    }
}



