#include "cdapolyhedron.h"

#include "cdavertex.h"
#include "cdaedge.h"
#include "cdaface.h"

#include "lpsolver.h"

UINT_32 CDAPolyhedron::m_uiObjectCount = 0;
LPSolver *CDAPolyhedron::m_pLPSolver = NULL;
CHeap<VOID*, FLOAT> *CDAPolyhedron::m_pheapCDAFace = NULL;
CListInt *CDAPolyhedron::m_plistCDAFaceCone = NULL;

//////////////////////////////////////////////////////////////////////
// CDAPolyhedron Class
//

CDAPolyhedron::CDAPolyhedron()
{
  if ( m_uiObjectCount == 0 )
    {
      m_pLPSolver = new LPSolver;
      ASSERT( m_pLPSolver != NULL );
      m_pLPSolver->Init();

      m_pheapCDAFace = new CHeap<VOID*, FLOAT>;
      ASSERT( m_pheapCDAFace != NULL);
      m_pheapCDAFace->Init();

      m_plistCDAFaceCone = new CListInt;
      ASSERT( m_plistCDAFaceCone != NULL);
      m_plistCDAFaceCone->Init();
    }
  m_uiObjectCount++;

  m_listCDAVertex.Init();
  m_listCDAEdge.Init();
  m_listCDAFace.Init();
}

CDAPolyhedron::~CDAPolyhedron()
{
  Uninit();

  m_uiObjectCount--;
  if ( m_uiObjectCount == 0 )
    {
      // LP Solver
      if ( m_pLPSolver != NULL )
	{
	  delete m_pLPSolver;
	  m_pLPSolver = NULL;
	}

      // Priority Queue of Faces with sets of outside Points
      if ( m_pheapCDAFace != NULL)
	{
	  ASSERT( m_pheapCDAFace->GetSize() == 0 );
	  delete m_pheapCDAFace;
	  m_pheapCDAFace = NULL;
        }

      // Cone Faces
      if ( m_plistCDAFaceCone != NULL)
	{
	  ASSERT( m_plistCDAFaceCone->GetLength() == 0 );
	  delete m_plistCDAFaceCone;
	  m_plistCDAFaceCone = NULL;
        }
    }
}

POINT4D gppt[4] = {
  {1.0, 1.0, 0.0, 1.0},
  {-1.0, 1.0, 0.0, 1.0},
  {-1.0, -1.0, 0.0, 1.0},
  {1.0, -1.0, 0.0, 1.0},
};

/*
VOID CDAPolyhedron::Init()
{
  m_listCDAVertex.Init();
  m_listCDAEdge.Init();
  m_listCDAFace.Init();

  //////////////////////////////
  // START: Jordan Fix This!!!!
  //

  UINT_32 i, j, k;
  Point pt;
  CDAVertex *pppCDAVertex[2][4];
  CDAVertex *pCDAVertexPrev;
  CDAEdge *pCDAEdge;
  CDAFace *pppCDAFace[2][2];

  // Create Points for the Face
  for ( i = 0; i < 4; i++ )
    {
      pppCDAVertex[0][i] = AllocateCDAVertex();
      ASSERT( pppCDAVertex[0][i] != NULL );
      pppCDAVertex[0][i]->Init();
      
      pt.Set(gppt[i]);
      pppCDAVertex[0][i]->SetPoint(pt);

      m_listCDAVertex.InsertLast(pppCDAVertex[0][i], 0);
    }

  // Create Points for the Hole
  for ( i = 0; i < 4; i++ )
    {
      pppCDAVertex[1][i] = AllocateCDAVertex();
      ASSERT( pppCDAVertex[1][i] != NULL );
      pppCDAVertex[1][i]->Init();
      
      pt.Set(gppt[i]);
      pt = pt / 2.0;
      pt[W] = 1.0;
      pppCDAVertex[1][i]->SetPoint(pt);

      m_listCDAVertex.InsertLast(pppCDAVertex[1][i], 0);
    }

  // Create the front Faces
  for ( j = 0; j < 2; j++ )
    {
      for ( i = 0; i < 2; i++ )
	{
	  pppCDAFace[j][i] = AllocateCDAFace();
	  ASSERT( pppCDAFace[j][i] != NULL );
	  pppCDAFace[j][i]->Init();
	  
	  for ( k = 0; k < 4; k++ )
	    {
	      pppCDAFace[j][i]->AddVertex(pppCDAVertex[j][(i==0) ? (k) : (3-k)]);
	    }
	  
	  m_listCDAFace.InsertLast(pppCDAFace[j][i], 0);
	}
    }

  pppCDAFace[0][0]->AddBoundDownFace(pppCDAFace[1][0]);
  pppCDAFace[0][1]->AddBoundDownFace(pppCDAFace[1][1]);

  // Create all the Edges
  for ( j = 0; j < 2; j++ )
    {
      pCDAVertexPrev = pppCDAVertex[j][3];

      for ( i = 0; i < 4; i++ )
	{
	  pCDAEdge = AllocateCDAEdge();
	  ASSERT( pCDAEdge != NULL );
	  pCDAEdge->Init(pCDAVertexPrev, pppCDAVertex[j][i],
			   pppCDAFace[j][0], pppCDAFace[j][0]);

	  m_listCDAEdge.InsertLast(pCDAEdge, 0);
	}
    }
  //
  // END: Jordan Fix This!!!!
  //////////////////////////////
}
*/

VOID CDAPolyhedron::Init(CListInt *plistCDAVertex)
{
  ConvexHull(plistCDAVertex);
}

VOID CDAPolyhedron::Uninit()
{
  CDAVertex *pCDAVertex;
  for ( pCDAVertex = (CDAVertex*)(m_listCDAVertex.RemoveFirst());
	pCDAVertex != NULL;
	pCDAVertex = (CDAVertex*)(m_listCDAVertex.RemoveFirst()) )
    {
      DeallocateCDAVertex(pCDAVertex);
    }

  CDAEdge *pCDAEdge;
  for ( pCDAEdge = (CDAEdge*)(m_listCDAEdge.RemoveFirst());
	pCDAEdge != NULL;
	pCDAEdge = (CDAEdge*)(m_listCDAEdge.RemoveFirst()) )
    {
      DeallocateCDAEdge(pCDAEdge);
    }

  CDAFace *pCDAFace;
  for ( pCDAFace = (CDAFace*)(m_listCDAFace.RemoveFirst());
	pCDAFace != NULL;
	pCDAFace = (CDAFace*)(m_listCDAFace.RemoveFirst()) )
    {
      DeallocateCDAFace(pCDAFace);
    }
}

BOOL CDAPolyhedron::Initialized()
{
  return TRUE;
}

BOOL CDAPolyhedron::Valid()
{
  return TRUE;
}
/*
VOID CDAPolyhedron::Preprocess()
{
  CListIterInt iter;

  CDAVertex *pCDAVertex;
  iter.Init(&m_listCDAVertex);
  for ( pCDAVertex = (CDAVertex*)(iter.PeekFirst());
	pCDAVertex != NULL;
	pCDAVertex = (CDAVertex*)(iter.PeekNext()) )
    {
      pCDAVertex->Preprocess();
    }

  CDAEdge *pCDAEdge;
  iter.Init(&m_listCDAEdge);
  for ( pCDAEdge = (CDAEdge*)(iter.PeekFirst());
	pCDAEdge != NULL;
	pCDAEdge = (CDAEdge*)(iter.PeekNext()) )
    {
      pCDAEdge->Preprocess();
    }

  CDAFace *pCDAFace;
  iter.Init(&m_listCDAFace);
  for ( pCDAFace = (CDAFace*)(iter.PeekFirst());
	pCDAFace != NULL;
	pCDAFace = (CDAFace*)(iter.PeekNext()) )
    {
      pCDAFace->Preprocess();
    }  
}
*/

CDAPolyhedron *CDAPolyhedron::Copy()
{
  CDAPolyhedron *pCDAPolyhedronCopy;
  CListIterInt iterCDAVertex, iterCDAEdge, iterCDAFace;
  CDAVertex *pCDAVertex;
  CDAEdge *pCDAEdge;
  CDAFace *pCDAFace;

  pCDAPolyhedronCopy = AllocateCDAPolyhedron();
  ASSERT( pCDAPolyhedronCopy != NULL );
  //pCDAPolyhedronCopy->Init();

  ////////////////////////////////////////////////////////////
  // Copy the Vertices
  iterCDAVertex.Init(&m_listCDAVertex);
  for ( pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekFirst());
	pCDAVertex != NULL;
	pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekNext()) )
    { 
      pCDAPolyhedronCopy->m_listCDAVertex.InsertLast(pCDAVertex->Copy(), 0);
    }

  ////////////////////////////////////////////////////////////
  // Copy the Edges
  iterCDAEdge.Init(&m_listCDAEdge);
  for ( pCDAEdge = (CDAEdge*)(iterCDAEdge.PeekFirst());
	pCDAEdge != NULL;
	pCDAEdge = (CDAEdge*)(iterCDAEdge.PeekNext()) )
    { 
      pCDAPolyhedronCopy->m_listCDAEdge.InsertLast(pCDAEdge->Copy(), 0);
    }

  ////////////////////////////////////////////////////////////
  // Copy the Faces
  iterCDAFace.Init(&m_listCDAFace);
  for ( pCDAFace = (CDAFace*)(iterCDAFace.PeekFirst());
	pCDAFace != NULL;
	pCDAFace = (CDAFace*)(iterCDAFace.PeekNext()) )
    { 
      pCDAPolyhedronCopy->m_listCDAFace.InsertLast(pCDAFace->Copy(), 0);
    }

  return pCDAPolyhedronCopy;
}

VOID CDAPolyhedron::Transform(const Matrix &mGroot_obj, const Matrix &mGobj_root)
{
  CListIterInt iterCDAVertex;
  CDAVertex *pCDAVertex;
  CListIterInt iterCDAFace;
  CDAFace *pCDAFace;

  iterCDAVertex.Init(&m_listCDAVertex);
  for ( pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekFirst());
	pCDAVertex != NULL;
	pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekNext()) )
    {
      pCDAVertex->Transform(mGroot_obj, mGobj_root);
    }

  iterCDAFace.Init(&m_listCDAFace);
  for ( pCDAFace = (CDAFace*)(iterCDAFace.PeekFirst());
	pCDAFace != NULL;
	pCDAFace = (CDAFace*)(iterCDAFace.PeekNext()) )
    {
      pCDAFace->Transform(mGroot_obj, mGobj_root);
    }
}

static const UINT_32 cppuiVertexSimplexIndex[4][3] = {
  {0, 1, 2},
  {1, 0, 3},
  {2, 1, 3},
  {0, 2, 3}
};
static const UINT_32 cppuiEdgeSimplexIndex[4][3] = {
  {0, 3, 1},
  {0, 2, 4},
  {3, 4, 5},
  {1, 5, 2}
};

// This uses the QuickHull algorithm for computing the convex hull of
// a point set ( URL == http://www.geom.umn.edu/software/qhull/ ).
//
// Return value is the Dimension of the Convex Hull (3, 2, 1, or 0)
UINT_32 CDAPolyhedron::ConvexHull(CListInt *plistCDAVertex)
{
  UINT_32 i, j, k, uiDimension;
  CDAVertex *pCDAVertex, *pCDAVertexNew;
  CDAVertex *pCDAVertexEye, *pCDAVertexCurr;
  CDAEdge *pCDAEdgePrev, *pCDAEdgeCurr;
  CDAEdge *pCDAEdgeConePrev, *pCDAEdgeConeCurr, *pCDAEdgeConeStart;
  CDAFace *pCDAFace, *pCDAFaceEye, *pCDAFaceNew;
  CDAFace *pCDAFaceCurr;
  CDAVertex *ppCDAVertexSimplex[4], *ppCDAVertex[3];
  CDAEdge *ppCDAEdgeSimplex[6], *ppCDAEdge[3];
  Vector vL, vV, vNormal;
  FLOAT fDistance, fDistanceMax, fDot;

  CListInt listCDAVertexUnassigned;
  CListInt listCDAEdgeHorizon;
  CListIterInt iterCDAVertex;
  CListIterInt iterCDAEdge;
  CListIterInt iterCDAFace;

  BOOL bDone;

  ASSERT( plistCDAVertex != NULL );
  ASSERT( m_listCDAVertex.GetLength() == 0 );
  ASSERT( m_pheapCDAFace->GetSize() == 0 );

  // d == 3

  uiDimension = 3;

  listCDAVertexUnassigned.Init();
  listCDAEdgeHorizon.Init();

  //////////////////////////////////////////////////
  // Copy the List of Vertices

  iterCDAVertex.Init(plistCDAVertex);
  for ( pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekFirst());
	pCDAVertex != NULL;
	pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekNext()) )
    {
      pCDAVertexNew = AllocateCDAVertex();
      ASSERT( pCDAVertexNew != NULL );
      pCDAVertexNew->Init(pCDAVertex->GetPoint());

      pCDAVertexNew->SetDual(0, pCDAVertex->GetDual(0));
      pCDAVertexNew->SetDual(1, pCDAVertex->GetDual(1));

      m_listCDAVertex.InsertLast(pCDAVertexNew, 0);
      listCDAVertexUnassigned.InsertLast(pCDAVertexNew, 0);
    }

  //////////////////////////////////////////////////
  // Create a simplex of d+1 points

  for ( i = 0; i <= uiDimension; i++ )
    {
      ppCDAVertexSimplex[i] = NULL;
    }

  //////////////////////////////
  // Find the extreme points from the YZ-plane
  // i.e. The hopefully 2 extreme points along the X-axis
  iterCDAVertex.Init(&listCDAVertexUnassigned);
  pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekFirst());

  if ( pCDAVertex == NULL )
    {
      // We have the empty set which is basically -1D
      ASSERT( FALSE );
      return 0;
    }

  ppCDAVertexSimplex[0] = pCDAVertex;
  ppCDAVertexSimplex[1] = pCDAVertex;
 
  for ( pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekNext());
	pCDAVertex != NULL;
	pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekNext()) )
    {
      if ( pCDAVertex->GetPoint()(X) < ppCDAVertexSimplex[0]->GetPoint()(X) )
	{
	  ppCDAVertexSimplex[0] = pCDAVertex;
	}
      if ( pCDAVertex->GetPoint()(X) > ppCDAVertexSimplex[1]->GetPoint()(X) )
	{
	  ppCDAVertexSimplex[1] = pCDAVertex;
	}
    }

  if ( ppCDAVertexSimplex[0] == ppCDAVertexSimplex[1] )
    {
      // Our Convex Hull is at most 2D
      ASSERT( FALSE );
      // There is more work to be done
    }

  UINT_32 uiConvexHullCounter;

  uiConvexHullCounter = 0;
/*
  printf("ConvexHull[%ld] = ", uiConvexHullCounter++);
  ppCDAVertexSimplex[0]->GetPoint().Display();
  printf("ConvexHull[%ld] = ", uiConvexHullCounter++);
  ppCDAVertexSimplex[1]->GetPoint().Display();
*/

  // Find a Third Point which is furthest from the line between the first two
  vL = ppCDAVertexSimplex[1]->GetPoint() - ppCDAVertexSimplex[0]->GetPoint();
  VERIFY( vL.Normalize() );

  fDistanceMax = 0.0;
  ppCDAVertexSimplex[2] = NULL;

  for ( pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekFirst());
	pCDAVertex != NULL;
	pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekNext()) )
    {
      if ( ppCDAVertexSimplex[0] != pCDAVertex && ppCDAVertexSimplex[1] != pCDAVertex )
	{
	  vV = pCDAVertex->GetPoint() - ppCDAVertexSimplex[0]->GetPoint();
	  vNormal = vV * vL;
	  fDistance = vNormal.Magnitude();

	  if ( fDistance > fDistanceMax )
	    {
	      fDistanceMax = fDistance;
	      ppCDAVertexSimplex[2] = pCDAVertex;
	    }
	}
    }

  if ( ppCDAVertexSimplex[2] == NULL )
    {
      // Our Point set is 1D along the line from Left to Right
      ASSERT( FALSE );

      return 1;
    }
/*
  printf("ConvexHull[%ld] = ", uiConvexHullCounter++);
  ppCDAVertexSimplex[2]->GetPoint().Display();
*/

  vV = ppCDAVertexSimplex[2]->GetPoint() - ppCDAVertexSimplex[0]->GetPoint();
  vNormal = vL * vV;
  VERIFY( vNormal.Normalize() );
  vNormal.CompletePlaneEquation(ppCDAVertexSimplex[0]->GetPoint());

  ppCDAVertexSimplex[3] = NULL;
  fDistanceMax = 0.0;

  for ( pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekFirst());
	pCDAVertex != NULL;
	pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekNext()) )
    {
      if ( ppCDAVertexSimplex[0] != pCDAVertex && 
	   ppCDAVertexSimplex[1] != pCDAVertex &&
	   ppCDAVertexSimplex[2] != pCDAVertex )
	{
	  vV = pCDAVertex->GetPoint();
	  vV[W] = 1.0;
	  fDistance = vNormal.Dot(vV);

	  if ( fabs(fDistance) > fabs(fDistanceMax) )
	    {
	      fDistanceMax = fDistance;
	      ppCDAVertexSimplex[3] = pCDAVertex;
	    }
	}
    }

  if ( ppCDAVertexSimplex[3] == NULL )
    {
      // Our Point set is 2D in the plane we have defined
      ASSERT( FALSE );

      return 2;
    }
/*
  printf("ConvexHull[%ld] = ", uiConvexHullCounter++);
  ppCDAVertexSimplex[3]->GetPoint().Display();
*/
  if ( fDistanceMax > 0.0 )
    {
      // Our Triangle is oriented incorrectly because the fourth
      // vertex of our tetrahedron was above it.
      pCDAVertex = ppCDAVertexSimplex[1];
      ppCDAVertexSimplex[1] = ppCDAVertexSimplex[2];
      ppCDAVertexSimplex[2] = pCDAVertex;
    }

  // We have 4 non-degenerate points, so now make our tetrahedron

  // Create the edges of our tetrahedron

  k = 0;
  for ( i = 0; i < 3; i++ )
    {
      for ( j = i+1; j < 4; j++ )
	{
	  // Create the Face
	  ppCDAEdgeSimplex[k] = AllocateCDAEdge();
	  ASSERT( ppCDAEdgeSimplex[k] != NULL );
	  ppCDAEdgeSimplex[k]->Init(ppCDAVertexSimplex[i], ppCDAVertexSimplex[j]);
	  
	  m_listCDAEdge.InsertLast(ppCDAEdgeSimplex[k], 0);

	  k++;
	}
    }

  //////////////////////////////////////////////////
  // Create the faces of our tetrahedron
  for ( i = 0; i < 4; i++ )
    {
//      printf("START: Creating Tetra Face %ld:\n", i);
      for ( j = 0; j < 3; j++ )
	{
	  ppCDAVertex[j] = ppCDAVertexSimplex[cppuiVertexSimplexIndex[i][j]];
	  ppCDAEdge[j] = ppCDAEdgeSimplex[cppuiEdgeSimplexIndex[i][j]];
/*
printf("START: Making Vertex: %ld\n", j);
ppCDAVertex[j]->GetPoint().Display();
printf("END: Making Vertex: %ld\n", j);

printf("START: Making Edge: %ld\n", j);
    printf("From: ");
    ppCDAEdge[j]->GetVertexFrom()->GetPoint().Display();
    printf("To: ");
    ppCDAEdge[j]->GetVertexTo()->GetPoint().Display();
printf("END: Making Edge: %ld\n", j);
*/
	}

      // Create the Face
      pCDAFace = AllocateCDAFace();
      ASSERT( pCDAFace != NULL );
      pCDAFace->Init(3, ppCDAVertex, ppCDAEdge, &listCDAVertexUnassigned);

      m_listCDAFace.InsertLast(pCDAFace, 0);
//      printf("END: Creating Tetra Face %ld:\n", i);
    }

  //////////////////////////////////////////////////
  // Initial Partitioning of Point Set to Outside Sets
  // Points should go to the face which they are furthest outside of 
  PartitionVertices(&listCDAVertexUnassigned, &m_listCDAFace);
/*
printf("START: The Vertices of the Tetrahedron:\n");
ppCDAVertexSimplex[0]->GetPoint().Display();
ppCDAVertexSimplex[1]->GetPoint().Display();
ppCDAVertexSimplex[2]->GetPoint().Display();
ppCDAVertexSimplex[3]->GetPoint().Display();
printf("END: The Vertices of the Tetrahedron:\n");

printf("The Tetrahedron Simplex:\n");
Display();
*/
  //////////////////////////////////////////////////
  // Do the Main Loop


  iterCDAFace.Init(&m_listCDAFace);

  for ( pCDAFaceEye = (CDAFace*)(m_pheapCDAFace->PeekTop());
	pCDAFaceEye != NULL;
	pCDAFaceEye = (CDAFace*)(m_pheapCDAFace->PeekTop()) )
	
    {
      // Verify that our scratch lists have been emptied
      ASSERT( listCDAVertexUnassigned.GetLength() == 0 );
      ASSERT( listCDAEdgeHorizon.GetLength() == 0 );

      // Verify that the current face is valid
      ASSERT( pCDAFaceEye->GetOnHull() );

/*
  // I am going to want to check this later
	  if ( !pCDAFace->GetOnHull() )
	    {
	      // Let's verify that this Face's Outside Points have been assigned to 
	      // another Face
	      ASSERT( pCDAFace->GetOutsideVertexList() != NULL );
	      ASSERT( pCDAFace->GetOutsideVertexList()->GetLength() == 0 );
	      
	      continue;
	    }
*/

      pCDAVertexEye = pCDAFaceEye->RemoveFurthestOutsideVertex();
      ASSERT( pCDAVertexEye != NULL );
      ASSERT( pCDAVertexEye->GetOnHull() );
/*  
printf("START: The Eye Point:\n");
printf("ConvexHull[%ld] = ", uiConvexHullCounter++);
pCDAVertexEye->GetPoint().Display();
printf("END: The Eye Point:\n");
*/
      //////////////////////////////////////////////////
      // Initialize the Visible Set of Faces
      // Find the Horizon Edges
      // Recover all the Vertices in the outside sets of the 
      // visible Faces

      ASSERT( listCDAEdgeHorizon.GetLength() == 0 );
      pCDAFaceEye->CalculateHorizon(pCDAVertexEye,
				    m_pheapCDAFace,
				    &listCDAVertexUnassigned, 
				    &listCDAEdgeHorizon, 
				    NULL);
      
      // Check to make sure the Horizon was correctly computed
      ASSERT( listCDAEdgeHorizon.GetLength() >= 3 );

      //////////////////////////////////////////////////
      // Create the Cone of new Faces which connect the 
      // Eye Vertex to the Horizon Contour, also put 
      // the unassigned Vertices into the correct 
      // Outside sets

      iterCDAEdge.Init(&listCDAEdgeHorizon);
      
      // We must seed the loop
      pCDAEdgePrev = (CDAEdge*)(iterCDAEdge.PeekFirst()); 

      pCDAEdgeConePrev = AllocateCDAEdge();
      ASSERT( pCDAEdgeConePrev != NULL );
      pCDAEdgeConePrev->Init(pCDAVertexEye, pCDAEdgePrev->GetVertexFrom());
      m_listCDAEdge.InsertLast(pCDAEdgeConePrev, 0);

      pCDAEdgeConeStart = pCDAEdgeConePrev;

      pCDAEdgePrev->SetVerticesOnHull();

      for ( pCDAEdgeCurr = (CDAEdge*)(iterCDAEdge.PeekNext()); 
	    pCDAEdgeCurr != NULL;
	    pCDAEdgeCurr = (CDAEdge*)(iterCDAEdge.PeekNext()) )
	{
	  pCDAEdgeCurr->SetVerticesOnHull();

	  pCDAEdgeConeCurr = AllocateCDAEdge();
	  ASSERT( pCDAEdgeConeCurr != NULL );
	  pCDAEdgeConeCurr->Init(pCDAVertexEye, pCDAEdgeCurr->GetVertexFrom());
	  m_listCDAEdge.InsertLast(pCDAEdgeConeCurr, 0);

	  ppCDAVertex[0] = pCDAEdgePrev->GetVertexTo();
	  ppCDAVertex[1] = pCDAVertexEye;
	  ppCDAVertex[2] = pCDAEdgePrev->GetVertexFrom();

	  ppCDAEdge[0] = pCDAEdgeConeCurr;
	  ppCDAEdge[1] = pCDAEdgeConePrev;
	  ppCDAEdge[2] = pCDAEdgePrev;

	  // Check to see if the triangle we are about to create is
	  // coplanar to the face on the other side of the horizon.
	  // If so then we should add two sides to that face instead
	  // of creating a new face.

	  // Check if they are coplanar
	  pCDAFaceNew = pCDAEdgePrev->GetFaceNeighbor(NULL);
	  ASSERT( pCDAFaceNew != NULL );

	  if ( pCDAFaceNew->CheckCoplanar(pCDAVertexEye) )
	    {  
//	      printf("Replacing edge of Faces on plane = ");
//	      pCDAFaceNew->GetNormal().Display();
	      // They are coplanar so just adjust the existing face
	      pCDAFaceNew->ReplaceEdge(pCDAEdgePrev, 2, 
				       ppCDAVertex, ppCDAEdge, 
				       &listCDAVertexUnassigned);
	      pCDAEdgePrev->SetOnHull(FALSE);

	      m_plistCDAFaceCone->InsertLast(pCDAFaceNew, 0);
	    }
	  else
	    {
	      pCDAFaceNew = AllocateCDAFace();
	      ASSERT( pCDAFaceNew != NULL );
	      pCDAFaceNew->Init(3, ppCDAVertex, ppCDAEdge, 
				&listCDAVertexUnassigned); 
	      m_listCDAFace.InsertLast(pCDAFaceNew, 0);
	      m_plistCDAFaceCone->InsertLast(pCDAFaceNew, 0);
	    }

	  pCDAEdgeConePrev = pCDAEdgeConeCurr;
	  pCDAEdgePrev = pCDAEdgeCurr;
	}

      // There is one more face we need to make
      ppCDAVertex[0] = pCDAEdgePrev->GetVertexTo();
      ppCDAVertex[1] = pCDAVertexEye;
      ppCDAVertex[2] = pCDAEdgePrev->GetVertexFrom();
      
      ppCDAEdge[0] = pCDAEdgeConeStart;
      ppCDAEdge[1] = pCDAEdgeConePrev;
      ppCDAEdge[2] = pCDAEdgePrev;
      
      // Check to see if the triangle we are about to create is
      // coplanar to the face on the other side of the horizon.
      // If so then we should add two sides to that face instead
      // of creating a new face.

      // Check if they are coplanar
      pCDAFaceNew = pCDAEdgePrev->GetFaceNeighbor(NULL);
      ASSERT( pCDAFaceNew != NULL );

      if ( pCDAFaceNew->CheckCoplanar(pCDAVertexEye) )
	{  
//	  printf("Replacing edge of Faces on plane = ");
//	  pCDAFaceNew->GetNormal().Display();
	  // They are coplanar so just adjust the existing face
	  pCDAFaceNew->ReplaceEdge(pCDAEdgePrev, 2, 
				   ppCDAVertex, ppCDAEdge, 
				   &listCDAVertexUnassigned);
	  pCDAEdgePrev->SetOnHull(FALSE);
	  
	  if ( pCDAFaceNew->GetPriority() != NULL )
	    {
	      CDAFace *pCDAFace;
	      pCDAFace = (CDAFace*)(m_pheapCDAFace->Remove(pCDAFaceNew->GetPriority()));
	      ASSERT( pCDAFace = pCDAFaceNew );
	      pCDAFaceNew->SetPriority(NULL);
	    }
	  m_plistCDAFaceCone->InsertLast(pCDAFaceNew, 0);
	}
      else
	{
	  pCDAFaceNew = AllocateCDAFace();
	  ASSERT( pCDAFaceNew != NULL );
	  pCDAFaceNew->Init(3, ppCDAVertex, ppCDAEdge, 
			    &listCDAVertexUnassigned); 
	  m_listCDAFace.InsertLast(pCDAFaceNew, 0);
	  m_plistCDAFaceCone->InsertLast(pCDAFaceNew, 0);
	}

      //////////////////////////////////////////////////
      // Partitioning of Point Set to Outside Sets
      // Points should go to the face which they are furthest outside of 
      PartitionVertices(&listCDAVertexUnassigned, m_plistCDAFaceCone);
			
      // Reset the Cone Face list
      m_plistCDAFaceCone->RemoveAll();

      // Reset the Horizon Edge List to empty for the next iteration
      listCDAEdgeHorizon.RemoveAll();
    }

  //////////////////////////////////////////////////
  // Clean up the list of Convex Hull Faces, Edges, and Vertices
  // We must remove the ones which are not on the Convex Hull any more

  iterCDAVertex.Init(&m_listCDAVertex);
  for ( pCDAVertexCurr = (CDAVertex*)(iterCDAVertex.PeekFirst());
	pCDAVertexCurr != NULL; )
    {
      if ( pCDAVertexCurr->GetOnHull() )
	{
	  // This is actually on the Convex Hull so keep it
	  pCDAVertexCurr = (CDAVertex*)(iterCDAVertex.PeekNext());
	}
      else
	{
	  // This was engulfed by later geometry so we must deallocate it
	  DeallocateCDAVertex(pCDAVertexCurr);
	  pCDAVertexCurr = (CDAVertex*)(iterCDAVertex.PeekNextRemoveCurr());
	}
    }

  iterCDAEdge.Init(&m_listCDAEdge);
  for ( pCDAEdgeCurr = (CDAEdge*)(iterCDAEdge.PeekFirst());
	pCDAEdgeCurr != NULL; )
    {
      if ( pCDAEdgeCurr->GetOnHull() )
	{
	  // This is actually on the Convex Hull so keep it
	  pCDAEdgeCurr = (CDAEdge*)(iterCDAEdge.PeekNext());
	}
      else
	{
	  // This was engulfed by later geometry so we must deallocate it
	  DeallocateCDAEdge(pCDAEdgeCurr);
	  pCDAEdgeCurr = (CDAEdge*)(iterCDAEdge.PeekNextRemoveCurr());
	}
    }

  iterCDAFace.Init(&m_listCDAFace);
  for ( pCDAFaceCurr = (CDAFace*)(iterCDAFace.PeekFirst());
	pCDAFaceCurr != NULL; )
    {
      if ( pCDAFaceCurr->GetOnHull() )
	{
	  // This is actually on the Convex Hull so keep it
	  pCDAFaceCurr = (CDAFace*)(iterCDAFace.PeekNext());
	}
      else
	{
	  // This was engulfed by later geometry so we must deallocate it
	  DeallocateCDAFace(pCDAFaceCurr);
	  pCDAFaceCurr = (CDAFace*)(iterCDAFace.PeekNextRemoveCurr());
	}
    }
  
  return uiDimension;
}

VOID CDAPolyhedron::PartitionVertices(CListInt *plistCDAVertex, 
				      CListInt *plistCDAFace)
{
  CListIterInt iterCDAFace;
  CDAFace *pCDAFace, *pCDAFaceMax;
  CDAVertex *pCDAVertex;
  FLOAT fDistance, fDistanceMax;
  CHeapEntry<VOID*, FLOAT> *pCHeapEntry;

  ASSERT( plistCDAVertex != NULL );
  ASSERT( plistCDAFace != NULL );

  iterCDAFace.Init(plistCDAFace);
  for ( pCDAVertex = (CDAVertex*)(plistCDAVertex->RemoveFirst());
	pCDAVertex != NULL;
	pCDAVertex = (CDAVertex*)(plistCDAVertex->RemoveFirst()) )
    {
      pCDAFaceMax = (CDAFace*)(iterCDAFace.PeekFirst());
      ASSERT( pCDAFaceMax != NULL );

      fDistanceMax = pCDAFaceMax->CalculateDistance(pCDAVertex->GetPoint());

      for ( pCDAFace = (CDAFace*)(iterCDAFace.PeekNext());
	    pCDAFace != NULL;
	    pCDAFace = (CDAFace*)(iterCDAFace.PeekNext()) )
	{
	  fDistance = pCDAFace->CalculateDistance(pCDAVertex->GetPoint());
	  if ( fDistance > fDistanceMax )
	    {
	      pCDAFaceMax = pCDAFace;
	      fDistanceMax = fDistance;
	    }
	}

      if ( fDistanceMax > 0.0 && !pCDAFaceMax->CheckCoplanar(pCDAVertex) )
	{
	  pCDAFaceMax->AddOutsideVertex(fDistanceMax, pCDAVertex);
	}
      else
	{
	  pCDAVertex->SetOnHull(FALSE);
	}
    }

  // Add the faces which have outside vertices to the priority queue of 
  // unprocessed faces
  for ( pCDAFace = (CDAFace*)(iterCDAFace.PeekFirst());
	pCDAFace != NULL;
	pCDAFace = (CDAFace*)(iterCDAFace.PeekNext()) )
    {
      pCDAVertex = pCDAFace->PeekFurthestOutsideVertex(&fDistance);
      if ( pCDAVertex != NULL && pCDAFace->GetPriority() == NULL )
	{
	  pCHeapEntry = m_pheapCDAFace->Insert(fDistance, pCDAFace);
	  ASSERT( pCHeapEntry != NULL );
	  pCDAFace->SetPriority(pCHeapEntry);
	}
    }
}

static const VECTOR4D s_cpvObjective[6] =
{
  { 1, 0, 0, 0 },
  { -1, 0, 0, 0 },
  { 0, 1, 0, 0 },
  { 0, -1, 0, 0 },
  { 0, 0, 1, 0 },
  { 0, 0, -1, 0 }
};

CDAPolyhedron *CDAPolyhedron::Intersect(CDAPolyhedron *pCDAPolyhedron)
{
  CDAPolyhedron *pCDAPolyhedronResult, *pCDAPolyhedronDual;
  CDAVertex *pCDAVertex;
  CListInt listCDAVertex;
  CListIterInt iterCDAVertex;
  Vector vOrigin, vTrans;

  ASSERT( pCDAPolyhedron != NULL );

  listCDAVertex.Init();
  
  // START: Jordan Fix this!!!

  // We must find a point in the intersection around which we will 
  // dualize our polyhedra
  
  vTrans.Null();
  if ( !m_pLPSolver->SolveCDAIntersection(&m_listCDAFace, 
					  &(pCDAPolyhedron->m_listCDAFace),
					  vTrans) )
    {
      // There is no intersection
      return NULL;
    }
  ASSERT( vTrans[3] == 1.0 );
  vTrans[3] = 0.0;

  vTrans = -vTrans;
  
  //   END: Jordan Fix this!!!

  // Calculate the Dual Vertices which are the duals of the faces, and 
  // put them in one list

  this->CalculateDualVertices(0, &listCDAVertex, vTrans);
  pCDAPolyhedron->CalculateDualVertices(1, &listCDAVertex, vTrans);

  // Create a polyhedron which is the convex hull (in dual space) of the
  // duals of the two operand polyhedra.  This polyhedron is the dual of 
  // the intersection of our two operand polyhedra
  pCDAPolyhedronDual = AllocateCDAPolyhedron();
  ASSERT( pCDAPolyhedronDual != NULL );
  pCDAPolyhedronDual->Init(&listCDAVertex);

  // Deallocate the dual space vertices
  for ( pCDAVertex = (CDAVertex*)(listCDAVertex.RemoveFirst());
	pCDAVertex != NULL;
	pCDAVertex = (CDAVertex*)(listCDAVertex.RemoveFirst()) )
    {
      DeallocateCDAVertex(pCDAVertex);
    }
/*
  printf("Dual Space Polyhedron:\n");
  pCDAPolyhedronDual->Display();
*/
  // Finally we need to redualize this convex hull polyhedron to get
  // intersection in real space, remember to translate the points back
  vOrigin.Set(0.0, 0.0, 0.0, 0.0);
  pCDAPolyhedronDual->CalculateDualVertices(0, &listCDAVertex, vOrigin);

  vTrans[3] = 0.0;
  vTrans = -vTrans;

  iterCDAVertex.Init(&listCDAVertex);
  for ( pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekFirst());
	pCDAVertex != NULL;
	pCDAVertex = (CDAVertex*)(iterCDAVertex.PeekNext()) )
    {
      pCDAVertex->SetPoint(pCDAVertex->GetPoint() + vTrans);
    }

  // Now Create the intersection polyhedron
  pCDAPolyhedronResult = AllocateCDAPolyhedron();
  ASSERT( pCDAPolyhedronResult != NULL );
  pCDAPolyhedronResult->Init(&listCDAVertex);

  // Deallocate the real space intersection vertices
  for ( pCDAVertex = (CDAVertex*)(listCDAVertex.RemoveFirst());
	pCDAVertex != NULL;
	pCDAVertex = (CDAVertex*)(listCDAVertex.RemoveFirst()) )
    {
      DeallocateCDAVertex(pCDAVertex);
    }
/*
  printf("Real Space Polyhedron:\n");
  pCDAPolyhedronResult->Display();
*/
  // Link the faces of the intersection polyhedron to the faces of
  // the two polyhedra which created it
  pCDAPolyhedronResult->BindFaces(vOrigin);

  // We are now done with the Dual Space convex hull polyhedron
  DeallocateCDAPolyhedron(pCDAPolyhedronDual);

  return pCDAPolyhedronResult;
}

// This Function calculates the dual of polyhedron's faces and 
// appends these dual space vertices to the end of the list.
// The vertices are allocated by this function and are the 
// responsibility of the caller to deallocate.
VOID CDAPolyhedron::CalculateDualVertices(UINT_32 uiOperand, 
					  CListInt *plistCDAVertex, 
					  const Vector &vTrans)
{
  CListIterInt iterCDAFace;
  CDAFace *pCDAFace;
  CDAVertex *pCDAVertex;

  ASSERT( uiOperand < 2 );
  ASSERT( plistCDAVertex != NULL );

  iterCDAFace.Init(&m_listCDAFace);
  for ( pCDAFace = (CDAFace*)(iterCDAFace.PeekFirst());
	pCDAFace != NULL;
	pCDAFace = (CDAFace*)(iterCDAFace.PeekNext()) )
    {
      pCDAVertex = AllocateCDAVertex();
      ASSERT( pCDAVertex != NULL );
      pCDAVertex->Init(pCDAFace->CalculateDual(vTrans));
      pCDAVertex->SetDual(uiOperand, pCDAFace);

      plistCDAVertex->InsertLast(pCDAVertex, 0);
    }
}

VOID CDAPolyhedron::BindFaces(const Vector &vTrans)
{
  CListIterInt iterCDAFace;
  CDAFace *pCDAFace;

  iterCDAFace.Init(&m_listCDAFace);
  for ( pCDAFace = (CDAFace*)(iterCDAFace.PeekFirst());
	pCDAFace != NULL;
	pCDAFace = (CDAFace*)(iterCDAFace.PeekNext()) )
    {
      pCDAFace->Bind(vTrans);
    }
}
  
VOID CDAPolyhedron::Bind(BOOL bZero)
{
  CListIterInt iterCDAFace;
  CDAFace *pCDAFace, *pCDAFaceBind;

  //printf("Bind CDAPolyhedron %p: %ld\n", this, bZero);

  if ( !bZero )
    {
      iterCDAFace.Init(&m_listCDAFace);
      for ( pCDAFace = (CDAFace*)(iterCDAFace.PeekFirst());
	    pCDAFace != NULL;
	    pCDAFace = (CDAFace*)(iterCDAFace.PeekNext()) )
	{
	  pCDAFaceBind = pCDAFace->GetOperandFace(0);
	  if ( pCDAFaceBind != NULL )
	    {
	      pCDAFaceBind->AddBoundDownFace(pCDAFace);
	    }
	}
    }
  else
    {
      Vector pvNormal[3];

      iterCDAFace.Init(&m_listCDAFace);
      for ( pCDAFace = (CDAFace*)(iterCDAFace.PeekFirst());
	    pCDAFace != NULL;
	    pCDAFace = (CDAFace*)(iterCDAFace.PeekNext()) )
	{
	  if ( pCDAFace->GetOperandFace(0) != NULL &&
	       pCDAFace->GetOperandFace(1) != NULL )
	    {
	      pvNormal[0] = pCDAFace->GetOperandFace(0)->GetNormal();
	      pvNormal[0][3] = 0.0;
	      pvNormal[1] = pCDAFace->GetOperandFace(1)->GetNormal();
	      pvNormal[1][3] = 0.0;
	      if ( pvNormal[0].Dot(pvNormal[1]) < 0.0 )
		{
		  // The faces point in opposite directions so we must
		  // make holes in both of them

		  // This face will make a hole in the one it points in 
		  // the same direction as, and there will be another
		  // face which will create a hole in the other one
		  pvNormal[2] = pCDAFace->GetNormal();
		  pvNormal[2][3] = 0.0;
		  if ( pvNormal[0].Dot(pvNormal[2]) > 0.0 )
		    {
		      pCDAFace->GetOperandFace(0)->AddBoundDownFace(pCDAFace);
		    }
		  else
		    {
		      pCDAFace->GetOperandFace(1)->AddBoundDownFace(pCDAFace);

		    }
		}
	    }
	  else if ( pCDAFace->GetOperandFace(0) != NULL )
	    {
	      pCDAFace->GetOperandFace(0)->AddBoundDownFace(pCDAFace);
	    }
	  else if ( pCDAFace->GetOperandFace(1) != NULL )
	    {
	      pCDAFace->GetOperandFace(1)->AddBoundDownFace(pCDAFace);
	    }
	}
    }
}

VOID CDAPolyhedron::CalculateBound()
{
/*
  CListIterInt iter;
  CDAPolyhedron *pCDAPolyhedron;
  BOOL bBoundDirty;

  CBound bound;
  
  iter.Init(&m_listCDAPolyhedron);
  
  for ( pCDAPolyhedron = (CDAPolyhedron*)iter.PeekFirst();
	pCDAPolyhedron != NULL;
	pCDAPolyhedron = (CDAPolyhedron*)iter.PeekNext() )
    {
      pCDAPolyhedron->InsertIntoBound(bound);
    }

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);
    }
*/
}

VOID CDAPolyhedron::Render(BOOL bPositive/* = TRUE*/)
{
  CListIterInt iter;
  CDAFace *pCDAFace;

  //gilEnable(GL_STENCIL_TEST); 

  iter.Init(&m_listCDAFace);
  
  for ( pCDAFace = (CDAFace*)iter.PeekFirst();
	pCDAFace != NULL;
	pCDAFace = (CDAFace*)iter.PeekNext() )
    {
      pCDAFace->Render(bPositive);
    }

  //gilDisable(GL_STENCIL_TEST); 
}

VOID CDAPolyhedron::RenderNormals(BOOL bPositive/* = TRUE*/)
{
  CListIterInt iter;
  CDAFace *pCDAFace;

  iter.Init(&m_listCDAFace);
  
  for ( pCDAFace = (CDAFace*)iter.PeekFirst();
	pCDAFace != NULL;
	pCDAFace = (CDAFace*)iter.PeekNext() )
    {
      pCDAFace->RenderNormals(bPositive);
    }
}

VOID CDAPolyhedron::Display()
{
  CListIterInt iter;
  CDAVertex *pCDAVertex;
  CDAFace *pCDAFace;

  printf("START: Polyhedron %p\n", this);

  printf("  START: Vertices\n");

  iter.Init(&m_listCDAVertex);
  
  for ( pCDAVertex = (CDAVertex*)iter.PeekFirst();
	pCDAVertex != NULL;
	pCDAVertex = (CDAVertex*)iter.PeekNext() )
    {
      pCDAVertex->Display();
    }

  printf("  END  : Vertices\n");

  printf("  START: Faces\n");

  iter.Init(&m_listCDAFace);
  
  for ( pCDAFace = (CDAFace*)iter.PeekFirst();
	pCDAFace != NULL;
	pCDAFace = (CDAFace*)iter.PeekNext() )
    {
      pCDAFace->Display();
    }

  printf("  END  : Faces\n");

  printf("END  : Polyhedron %p\n", this);
}


