#ifndef __CDAPOLYHEDRON_H__
#define __CDAPOLYHEDRON_H__

#include "cdamem.h"
#include "heap.h"

class CDAVertex;
class CDAEdge;
class CDAFace;

class LPSolver;

//////////////////////////////////////////////////////////////////////
// CDAPolyhedron Class
//

class CDAPolyhedron
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CDAPolyhedron();
  virtual ~CDAPolyhedron();
  //virtual VOID Init();
  VOID Init(CListInt *plistCDAVertex);
  virtual VOID Uninit();

  virtual BOOL Initialized();
  virtual BOOL Valid();

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  CDAPolyhedron *Copy();
  VOID Transform(const Matrix &mGroot_obj, const Matrix &mGobj_root);

  //VOID Preprocess();
  UINT_32 ConvexHull(CListInt *plistCDAVertex);
protected:
  VOID PartitionVertices(CListInt *plistCDAVertex, CListInt *plistCDAFace);

public:
  CDAPolyhedron *Intersect(CDAPolyhedron *pCDAPolyhedron);
  VOID CalculateDualVertices(UINT_32 uiOperand, 
			     CListInt *plistCDAVertex,
			     const Vector &vTrans);
  VOID BindFaces(const Vector &vTrans);
  
  // Actually bind the polyhedron to the correct operand
  VOID Bind(BOOL bZero);

protected:
  virtual VOID CalculateBound();

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID Render(BOOL bPositive = TRUE);
  virtual VOID RenderNormals(BOOL bPositive = TRUE);

//////////////////////////////////////////////////
// Output Functions
//
public:
  VOID Display();

//////////////////////////////////////////////////
// Accessor Functions
//
public:

//////////////////////////////////////////////////
// Class Functions
//
public:

//////////////////////////////////////////////////
// Class Variables
//
protected:
  static UINT_32 m_uiObjectCount;

  static LPSolver *m_pLPSolver;
  static CHeap<VOID*, FLOAT> *m_pheapCDAFace;
  static CListInt *m_plistCDAFaceCone;

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CListInt m_listCDAVertex;
  CListInt m_listCDAEdge;
  CListInt m_listCDAFace;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

#endif // __CDAPOLYHEDRON_H__
