#ifndef __CDAVERTEX_H__
#define __CDAVERTEX_H__

#include "cdamem.h"

class CDAEdge;
class CDAFace;

//////////////////////////////////////////////////////////////////////
// CDAVertex Class
//

class CDAVertex
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CDAVertex();
  virtual ~CDAVertex();
  virtual VOID Init(const Point &pt);
  virtual VOID Uninit();

  virtual BOOL Initialized();
  virtual BOOL Valid();

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  CDAVertex *Copy();
  VOID Transform(const Matrix &mGroot_obj, const Matrix &mGobj_root);

  VOID Preprocess();

protected:

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID Render();

//////////////////////////////////////////////////
// Output Functions
//
public:
  VOID Display();

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  inline const Point &GetPoint();
  inline VOID SetPoint(const Point &pt);
  inline BOOL GetOnHull();
  inline VOID SetOnHull(BOOL bOnHull);
  inline CDAFace *GetDual(UINT_32 uiOperand);
  inline VOID SetDual(UINT_32 uiOperand, CDAFace *pCDAFace);

  inline CDAVertex *GetCopy();

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  Point m_pt;

  CDAFace *m_ppCDAFaceDual[2];

  BOOL m_bOnHull;

  CDAVertex *m_pCDAVertexCopy;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline const Point &CDAVertex::GetPoint()
{
  return m_pt;
}

inline VOID CDAVertex::SetPoint(const Point &pt)
{
  m_pt = pt;
}

inline BOOL CDAVertex::GetOnHull()
{
  return m_bOnHull;
}

inline VOID CDAVertex::SetOnHull(BOOL bOnHull)
{
  m_bOnHull = bOnHull;
}

inline CDAFace *CDAVertex::GetDual(UINT_32 uiOperand)
{
  ASSERT( uiOperand < 2 );
  return m_ppCDAFaceDual[uiOperand];
}

inline VOID CDAVertex::SetDual(UINT_32 uiOperand, CDAFace *pCDAFace)
{
  ASSERT( uiOperand < 2 );
  if ( !( m_ppCDAFaceDual[uiOperand] == NULL || 
	  m_ppCDAFaceDual[uiOperand] == pCDAFace ) )
    {
      printf("Vertex Dual Problem\n");
    }
  ASSERT( m_ppCDAFaceDual[uiOperand] == NULL || 
	  m_ppCDAFaceDual[uiOperand] == pCDAFace );
  m_ppCDAFaceDual[uiOperand] = pCDAFace;
}

inline CDAVertex *CDAVertex::GetCopy()
{
  return m_pCDAVertexCopy;
}

#endif // __CDAVERTEX_H__



