#include "dynamics.h"

#include "slideprimitive.h"
#include "slideworld.h"
#include <tcl.h>

BOOL DynEval(FLOAT *pfValue, CHAR **ppcDynExpr, UINT_32 uiComponents)
{
  UINT_32 i;
  INT_32 iCode;
  FLOAT fNew;
  BOOL bChanged;

  bChanged = FALSE;

  ASSERT( pfValue != NULL );
  ASSERT( ppcDynExpr != NULL );

  for ( i = 0; i < uiComponents; i++ )
    {
      if ( ppcDynExpr[i] != NULL )
	{
	  iCode = Tcl_GlobalEval(CSLIDEWorld::GetInterp(), ppcDynExpr[i]);
	  if ( iCode != TCL_OK )
	    {
	      fprintf(CSLIDEPrimitive::GetErrorFile(), 
		      "Error: Tcl returned error code \"%s\" for the block:\n%s\n",
		      Tcl_GetStringResult(CSLIDEWorld::GetInterp()), ppcDynExpr[i]);
	      exit(1);
	    }
	  else
	    {
              fNew = (FLOAT)atof(Tcl_GetStringResult(CSLIDEWorld::GetInterp()));
	      bChanged = ( bChanged || ( fNew != pfValue[i] ) );
	      pfValue[i] = fNew;
	    }
	}      
    }

  return bChanged;
}

BOOL DynEvalInt(INT_32 *piValue, CHAR **ppcDynExpr, UINT_32 uiComponents)
{
  UINT_32 i;
  INT_32 iCode;
  INT_32 iNew;
  BOOL bChanged;

  bChanged = FALSE;

  ASSERT( piValue != NULL );
  ASSERT( ppcDynExpr != NULL );

  for ( i = 0; i < uiComponents; i++ )
    {
      if ( ppcDynExpr[i] != NULL )
	{
	  iCode = Tcl_GlobalEval(CSLIDEWorld::GetInterp(), ppcDynExpr[i]);
	  if ( iCode != TCL_OK )
	    {
	      fprintf(CSLIDEPrimitive::GetErrorFile(), 
		      "Error: Tcl returned error code \"%s\" for the block:\n%s",
		      Tcl_GetStringResult(CSLIDEWorld::GetInterp()), ppcDynExpr[i]);
	      exit(1);
	    }
	  else
	    {
	      iNew = atol(Tcl_GetStringResult(CSLIDEWorld::GetInterp()));
	      bChanged = ( bChanged || ( iNew != piValue[i] ) );
	      piValue[i] = iNew;
	    }
	}      
    }

  return bChanged;
}

BOOL DynEvalObj(FLOAT *pfValue, CHAR **ppcDynExpr, Tcl_Obj **pptoDynExpr, 
		UINT_32 uiComponents)
{
  UINT_32 i;
  INT_32 iCode;
  FLOAT fNew;
  BOOL bChanged;

  bChanged = FALSE;

  ASSERT( pfValue != NULL );
  ASSERT( ppcDynExpr != NULL );
  ASSERT( pptoDynExpr != NULL );

  for ( i = 0; i < uiComponents; i++ )
    {
      if ( ppcDynExpr[i] != NULL )
	{
	  if ( pptoDynExpr[i] == NULL )
	    {
	      pptoDynExpr[i] = Tcl_NewStringObj(ppcDynExpr[i], -1);
	    }
	  ASSERT( pptoDynExpr[i] != NULL );

	  iCode = Tcl_GlobalEvalObj(CSLIDEWorld::GetInterp(), pptoDynExpr[i]);
	  if ( iCode != TCL_OK )
	    {
	      fprintf(CSLIDEPrimitive::GetErrorFile(), 
		      "Error: Tcl returned error code \"%s\" for the block:\n%s\n",
		      Tcl_GetStringResult(CSLIDEWorld::GetInterp()), ppcDynExpr[i]);
	      exit(1);
	    }
	  else
	    {
              fNew = (FLOAT)atof(Tcl_GetStringResult(CSLIDEWorld::GetInterp()));
	      bChanged = ( bChanged || ( fNew != pfValue[i] ) );
	      pfValue[i] = fNew;
	    }
	}      
    }

  return bChanged;
}

BOOL DynEvalObjInt(INT_32 *piValue, CHAR **ppcDynExpr, Tcl_Obj **pptoDynExpr, 
		   UINT_32 uiComponents)
{
  UINT_32 i;
  INT_32 iCode;
  INT_32 iNew;
  BOOL bChanged;

  bChanged = FALSE;

  ASSERT( piValue != NULL );
  ASSERT( ppcDynExpr != NULL );

  for ( i = 0; i < uiComponents; i++ )
    {
      if ( ppcDynExpr[i] != NULL )
	{
	  if ( pptoDynExpr[i] == NULL )
	    {
	      pptoDynExpr[i] = Tcl_NewStringObj(ppcDynExpr[i], -1);
	    }
	  ASSERT( pptoDynExpr[i] != NULL );

	  iCode = Tcl_GlobalEvalObj(CSLIDEWorld::GetInterp(), pptoDynExpr[i]);
	  if ( iCode != TCL_OK )
	    {
	      fprintf(CSLIDEPrimitive::GetErrorFile(), 
		      "Error: Tcl returned error code \"%s\" for the block:\n%s",
		      Tcl_GetStringResult(CSLIDEWorld::GetInterp()), ppcDynExpr[i]);
	      exit(1);
	    }
	  else
	    {
	      iNew = atol(Tcl_GetStringResult(CSLIDEWorld::GetInterp()));
	      bChanged = ( bChanged || ( iNew != piValue[i] ) );
	      piValue[i] = iNew;
	    }
	}      
    }

  return bChanged;
}

//////////////////////////////////////////////////////////////////////
// DynFloat Class
//

DynFloat::DynFloat() 
{
  m_fValue = 0.0;
  m_pcDynExpr = NULL;
  m_ptoDynExpr = NULL;
  m_bIsMacro = FALSE;
}

DynFloat::~DynFloat()
{
  m_fValue = 0.0;
  m_pcDynExpr = NULL;
  if ( m_ptoDynExpr != NULL )
    {
      Tcl_DecrRefCount(m_ptoDynExpr);
      m_ptoDynExpr = NULL;
    }
}

BOOL DynFloat::Update()
{
  BOOL bChanged;

  bChanged = DynEvalObj(&m_fValue, &m_pcDynExpr, &m_ptoDynExpr, 1);
  
  // if is a macro make it non-dynamic after 1 update
  if (m_bIsMacro)
    {
      m_pcDynExpr = NULL;
    }

  return bChanged;
}

//////////////////////////////////////////////////////////////////////
// DynInt Class
//

DynInt::DynInt()
{
  m_iValue = 0;
  m_pcDynExpr = NULL;
  m_ptoDynExpr = NULL;
  m_bIsMacro = FALSE;
}

DynInt::~DynInt()
{
  m_iValue = 0;
  m_pcDynExpr = NULL;
  if ( m_ptoDynExpr != NULL )
    {
      Tcl_DecrRefCount(m_ptoDynExpr);
      m_ptoDynExpr = NULL;
    }
}

BOOL DynInt::Update()
{
  BOOL bChanged;

  bChanged = DynEvalObjInt(&m_iValue, &m_pcDynExpr, &m_ptoDynExpr, 1);

  // if is a macro make it non-dynamic after 1 update
  if (m_bIsMacro)
    {
      m_pcDynExpr = NULL;
    }

  return bChanged;
}




