#ifndef __DYNPOINT_H__
#define __DYNPOINT_H__

#include "point.h"
#include "dynamics.h"

//////////////////////////////////////////////////////////////////////
// DynPoint Class
//

class DynPoint : public Point
{
public:
  DynPoint();
  DynPoint(FLOAT x, FLOAT y, FLOAT z, FLOAT w = 1.0f);
  virtual ~DynPoint();

public:
  inline VOID SetAll(DynPoint *pdynpt);
  inline VOID SetAll(DynFloat *pdynfX, DynFloat *pdynfY = NULL, 
		     DynFloat *pdynfZ = NULL, DynFloat *pdynfW = NULL);
  inline VOID SetDynExprs(CHAR *pcX, CHAR *pcY = NULL, 
                          CHAR *pcZ = NULL, CHAR *pcW = NULL);
  virtual inline BOOL GetDynamic();

public:
  virtual BOOL Update();

protected:
  CHAR *m_ppcDynExpr[4];
  Tcl_Obj *m_pptoDynExpr[4];
  BOOL m_pbIsMacro[4];
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID DynPoint::SetAll(DynPoint *pdynpt)
{
  ASSERT( pdynpt != NULL );

  memcpy(this, pdynpt, sizeof(DynPoint));
}

inline VOID DynPoint::SetAll(DynFloat *pdynfX, DynFloat *pdynfY/* = NULL*/, 
			  DynFloat *pdynfZ/* = NULL*/, 
			  DynFloat *pdynfW/* = NULL*/)
{
  Null();
  if ( pdynfX != NULL )
    {
      m_pt[0] = pdynfX->GetValue();
      m_ppcDynExpr[0] = pdynfX->GetDynExpr();
	  m_pbIsMacro[0] = pdynfX->GetIsMacro();
    }
  if ( pdynfY != NULL )
    {
      m_pt[1] = pdynfY->GetValue();
      m_ppcDynExpr[1] = pdynfY->GetDynExpr();
	  m_pbIsMacro[1] = pdynfY->GetIsMacro();
    }
  if ( pdynfZ != NULL )
    {
      m_pt[2] = pdynfZ->GetValue();
      m_ppcDynExpr[2] = pdynfZ->GetDynExpr();
	  m_pbIsMacro[2] = pdynfZ->GetIsMacro();
    }
  if ( pdynfW != NULL )
    {
      m_pt[3] = pdynfW->GetValue();
      m_ppcDynExpr[3] = pdynfW->GetDynExpr();
	  m_pbIsMacro[3] = pdynfW->GetIsMacro();
    }
}

inline VOID DynPoint::SetDynExprs(CHAR *pcX, CHAR *pcY/* = NULL*/, 
				  CHAR *pcZ/* = NULL*/, 
				  CHAR *pcW/* = NULL*/)
{
  m_ppcDynExpr[0] = pcX;
  m_ppcDynExpr[1] = pcY;
  m_ppcDynExpr[2] = pcZ;
  m_ppcDynExpr[3] = pcW;
}

inline BOOL DynPoint::GetDynamic()
{
  return ( m_ppcDynExpr[0] != NULL ||
	   m_ppcDynExpr[1] != NULL ||
	   m_ppcDynExpr[2] != NULL ||
	   m_ppcDynExpr[3] != NULL );
}

#endif // __DYNPOINT_H__

