#ifndef __DYNVECTOR_H__
#define __DYNVECTOR_H__

#include "vector.h"
#include "dynamics.h"

//////////////////////////////////////////////////////////////////////
// DynVector Class
//

class DynVector : public Vector
{
public:
  DynVector();
  DynVector(FLOAT x, FLOAT y, FLOAT z, FLOAT w = 0.0f);
  virtual ~DynVector();

public:
  inline VOID SetAll(DynVector *pdynv);
  inline VOID SetAll(DynFloat *pdynfX, DynFloat *pdynfY = NULL, 
		     DynFloat *pdynfZ = NULL, DynFloat *pdynfW = NULL);
  inline VOID SetDynExprs(CHAR *pcX, CHAR *pcY = NULL, 
                          CHAR *pcZ = NULL, CHAR *pcW = NULL);
  inline CHAR *GetDynExpr(UINT_32 uiIndex) const;
  virtual inline BOOL GetDynamic();

public:
  virtual BOOL Update();

public:
  DynVector &operator=(const Vector &v);

protected:
  CHAR *m_ppcDynExpr[4];
  Tcl_Obj *m_pptoDynExpr[4];
  BOOL m_pbIsMacro[4];
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID DynVector::SetAll(DynVector *pdynv)
{
  ASSERT( pdynv != NULL );

  memcpy(this, pdynv, sizeof(DynVector));
}

inline VOID DynVector::SetAll(DynFloat *pdynfX, 
			      DynFloat *pdynfY/* = NULL*/,
			      DynFloat *pdynfZ/* = NULL*/, 
			      DynFloat *pdynfW/* = NULL*/)
{
  Null();
  if ( pdynfX != NULL )
    {
      m_v[0] = pdynfX->GetValue();
      m_ppcDynExpr[0] = pdynfX->GetDynExpr();
	  m_pbIsMacro[0] = pdynfX->GetIsMacro();
    }
  if ( pdynfY != NULL )
    {
      m_v[1] = pdynfY->GetValue();
      m_ppcDynExpr[1] = pdynfY->GetDynExpr();
	  m_pbIsMacro[1] = pdynfY->GetIsMacro();
    }
  if ( pdynfZ != NULL )
    {
      m_v[2] = pdynfZ->GetValue();
      m_ppcDynExpr[2] = pdynfZ->GetDynExpr();
	  m_pbIsMacro[2] = pdynfZ->GetIsMacro();
    }
  if ( pdynfW != NULL )
    {
      m_v[3] = pdynfW->GetValue();
      m_ppcDynExpr[3] = pdynfW->GetDynExpr();
	  m_pbIsMacro[3] = pdynfW->GetIsMacro();
    }
}

inline VOID DynVector::SetDynExprs(CHAR *pcX, 
				   CHAR *pcY/* = NULL*/, 
                                   CHAR *pcZ/* = NULL*/, 
				   CHAR *pcW/* = NULL*/)
{
  m_ppcDynExpr[0] = pcX;
  m_ppcDynExpr[1] = pcY;
  m_ppcDynExpr[2] = pcZ;
  m_ppcDynExpr[3] = pcW;
}

inline CHAR *DynVector::GetDynExpr(UINT_32 uiIndex) const
{
  ASSERT( uiIndex < 4 );

  return m_ppcDynExpr[uiIndex];
}

inline BOOL DynVector::GetDynamic()
{
  return ( m_ppcDynExpr[0] != NULL ||
	   m_ppcDynExpr[1] != NULL ||
	   m_ppcDynExpr[2] != NULL ||
	   m_ppcDynExpr[3] != NULL );
}

#endif // __DYNVECTOR_H__

