#include "graphics.h"
#include "mathlib.h"

CGraphics gCGraphics;

CGraphics::CGraphics()
{
  Init();
}

CGraphics::~CGraphics()
{
  Uninit();
}

VOID CGraphics::Init()
{
  m_eRenderMode = GL_RENDER;

  m_pcOutputFile[0] = '\0';
  m_fpOutputFile = stdout;

  m_eGeometricPrimitiveMode = GL_TRIANGLES;
}

VOID CGraphics::Uninit()
{
}

GLint CGraphics::gilRenderMode(GLenum mode)
{
  GLint iResult;

  iResult = 0;

  if ( m_eRenderMode == GL_WRITE_STL ||
       m_eRenderMode == GL_WRITE_RIB )
    {
      WriteFileTail(m_eRenderMode);
      CloseOutputFile();
    }

  switch( mode )
  {
  case GL_RENDER:
  case GL_SELECT:
  case GL_FEEDBACK:
    iResult = glRenderMode(mode);
    break;
  case GL_WRITE_STL:
  case GL_WRITE_RIB:
    iResult = glRenderMode(GL_RENDER);
    OpenOutputFile();
    WriteFileHeader(mode);    
    break;
  default:
    ASSERT( FALSE );
    break;
  }

  m_eRenderMode = mode;

  return iResult;
}

void CGraphics::gilBegin(GLenum mode)
{
  switch( m_eRenderMode )
  {
  case GL_WRITE_STL:
  case GL_WRITE_RIB:
/*
    switch( mode )
    {
    case GL_POINTS:
      break;
    case GL_LINES:
      break;
    case GL_POLYGON:
      break;
    case GL_TRIANGLES:
      break;
    case GL_QUADS:
      break;
    case GL_LINE_STRIP:
      break;
    case GL_LINE_LOOP:
      break;
    case GL_TRIANGLE_STRIP:
      break;
    case GL_TRIANGLE_FAN:
      break;
    case GL_QUAD_STRIP:
      break;
    default:
      ASSERT( FALSE );
      break;
    }
*/
    m_uiVertexCounter = 0;
    m_uiVerticesBuffered = 0;
    break;
  default:
    glBegin(mode);
    break;
  }

  m_eGeometricPrimitiveMode = mode;
}

void CGraphics::gilEnd()
{
  switch( m_eRenderMode )
  {
  case GL_WRITE_STL:
  case GL_WRITE_RIB:
    break;
  default:
    glEnd();
    break;
  }
}

void CGraphics::gilVertex3f(GLfloat x, GLfloat y, GLfloat z)
{
  gilVertex4f(x, y, z, 1.0);
}

void CGraphics::gilVertex3fv(const GLfloat *v)
{
  gilVertex4f(v[X], v[Y], v[Z], 1.0);
}

void CGraphics::gilVertex4f(GLfloat x, GLfloat y, GLfloat z, GLfloat w)
{

  VECTOR4D vGL;

  vGL[X] = x;
  vGL[Y] = y;
  vGL[Z] = z;
  vGL[W] = w;

#ifdef WIN32
  gilVertex4dv(vGL);
#else
  gilVertex4fv(vGL);
#endif
  
}

void CGraphics::gilVertex4fv(const GLfloat *v)
{
  switch( m_eRenderMode )
  {
  case GL_WRITE_STL:
  case GL_WRITE_RIB:
    {
      memcpy(m_pvVertices[m_uiVerticesBuffered], v, sizeof(VECTOR4D));
      m_uiVertexCounter++;
      m_uiVerticesBuffered++;

      switch( m_eGeometricPrimitiveMode )
      {
      case GL_POINTS:
      case GL_LINES:
      case GL_LINE_STRIP:
      case GL_LINE_LOOP:
        m_uiVerticesBuffered = 0;
        break;
      case GL_POLYGON:
        {
          if ( m_uiVerticesBuffered == 3 )
            {
              WriteTriangle(m_pvVertices[0], m_pvVertices[1], m_pvVertices[2]);

              memcpy(m_pvVertices[1], m_pvVertices[2], sizeof(VECTOR4D));
              m_uiVerticesBuffered = 2;
            }
        }
        break;
      case GL_TRIANGLES:
        {
          if ( m_uiVerticesBuffered == 3 )
            {
              WriteTriangle(m_pvVertices[0], m_pvVertices[1], m_pvVertices[2]);
              m_uiVerticesBuffered = 0;
            }
        }
        break;
      case GL_QUADS:
        {
          if ( m_uiVerticesBuffered == 4 )
            {
              WriteTriangle(m_pvVertices[0], m_pvVertices[1], m_pvVertices[2]);
              WriteTriangle(m_pvVertices[0], m_pvVertices[2], m_pvVertices[3]);
              m_uiVerticesBuffered = 0;
            }
        }
        break;
      case GL_TRIANGLE_STRIP:
        {
          if ( m_uiVerticesBuffered == 3 )
            {
              if ( ODD(m_uiVertexCounter) )
                {
                  WriteTriangle(m_pvVertices[0], m_pvVertices[1], m_pvVertices[2]);
                }
              else
                {
                  WriteTriangle(m_pvVertices[2], m_pvVertices[1], m_pvVertices[0]);
                }

              memcpy(m_pvVertices[0], m_pvVertices[1], 2*sizeof(VECTOR4D));
              m_uiVerticesBuffered = 2;
            }
        }
        break;
      case GL_TRIANGLE_FAN:
        {
          if ( m_uiVerticesBuffered == 3 )
            {
              WriteTriangle(m_pvVertices[0], m_pvVertices[1], m_pvVertices[2]);

              memcpy(m_pvVertices[1], m_pvVertices[2], sizeof(VECTOR4D));
              m_uiVerticesBuffered = 2;
            }
        }
        break;
      case GL_QUAD_STRIP:
        {
          if ( m_uiVerticesBuffered == 3 )
            {
              if ( ODD(m_uiVertexCounter) )
                {
                  WriteTriangle(m_pvVertices[0], m_pvVertices[1], m_pvVertices[2]);
                }
              else
                {
                  WriteTriangle(m_pvVertices[2], m_pvVertices[1], m_pvVertices[0]);
                }

              memcpy(m_pvVertices[0], m_pvVertices[1], 2*sizeof(VECTOR4D));
              m_uiVerticesBuffered = 2;
            }
        }
        break;
      default:
        ASSERT( FALSE );
        break;
      }
    }
    break;
  default:
    glVertex4fv(v);
    break;
  }
}

VOID CGraphics::gilSetOutputFile(CHAR *pcOutputFile)
{
  if ( pcOutputFile == NULL )
    {
      m_pcOutputFile[0] = '\0';
    }
  else
    {
      strcpy(m_pcOutputFile, pcOutputFile);
    }
}

VOID CGraphics::OpenOutputFile()
{
  if ( m_pcOutputFile[0] == '\0' )
    {
      m_fpOutputFile = stdout;
      fprintf(m_fpOutputFile, "*************** START OUTPUT ***************\n");
    }
  else
    {
      ASSERT( FALSE );

      // Make a check to see if the file exists and we are then
      // blowing it away
      m_fpOutputFile = fopen(m_pcOutputFile, "w");
      ASSERT( m_fpOutputFile != NULL );
    }
}

VOID CGraphics::CloseOutputFile()
{
  if ( m_fpOutputFile == stdout )
    {
      fprintf(m_fpOutputFile, "*************** END OUTPUT ***************\n");
    }
  else
    {
      fclose(m_fpOutputFile);
    }

  m_fpOutputFile = stdout;
}

VOID CGraphics::WriteFileHeader(GLenum mode)
{
  switch( mode )
  {
  case GL_WRITE_STL:
    fprintf(m_fpOutputFile, "solid ascii\n");
    break;
  case GL_WRITE_RIB:
    fprintf(m_fpOutputFile, "// This is a RenderMan .RIB part description\n\n");    
    break;
  default:
    ASSERT( FALSE );
    break;
  }
}

VOID CGraphics::WriteFileTail(GLenum mode)
{
  switch( mode )
  {
  case GL_WRITE_STL:
    fprintf(m_fpOutputFile, "end solid\n");
    break;
  case GL_WRITE_RIB:
    fprintf(m_fpOutputFile, "// This is the end of a RenderMan .RIB part description\n\n");    
    break;
  default:
    ASSERT( FALSE );
    break;
  }
}

VOID CGraphics::WriteTriangle(VECTOR4D v1, VECTOR v2, VECTOR v3)
{
  switch( m_eRenderMode )
  {
  case GL_WRITE_STL:
    {
      Vector vV1, vV2, vV3;

      vV1.Set(v1);
      vV2.Set(v2);
      vV3.Set(v3);

      WriteSTLTriangle(vV1, vV2, vV3);    
    }
    break;
  case GL_WRITE_RIB:
    break;
  default:
    ASSERT(FALSE);
    break;
  }
}

VOID CGraphics::WriteSTLTriangle(Vector &v1, Vector &v2, Vector &v3)
{
  Vector vU, vV, vNormal;

  vU = v2 - v1;
  vV = v3 - v2;

  vNormal = vU * vV;
  VERIFY( vNormal.Normalize() );

  fprintf(m_fpOutputFile, "  facet normal %f %f %f\n", vNormal[X], vNormal[Y], vNormal[Z]);
  fprintf(m_fpOutputFile, "    outer loop\n");
  fprintf(m_fpOutputFile, "      vertex %f %f %f\n", v1[X], v1[Y], v1[Z]);
  fprintf(m_fpOutputFile, "      vertex %f %f %f\n", v2[X], v2[Y], v2[Z]);
  fprintf(m_fpOutputFile, "      vertex %f %f %f\n", v3[X], v3[Y], v3[Z]);
  fprintf(m_fpOutputFile, "    endloop\n");
  fprintf(m_fpOutputFile, "  endfacet\n");
}

void CGraphics::giluNurbsCallback(GLUnurbs* nurb, GLenum which, GLvoid(*CallBackFunc)())
{
#ifndef WIN32
  gluNurbsCallback(nurb, which, CallBackFunc);
#else
  //gluNurbsCallback(nurb, which, (__stdcall *)CallBackFunc);
#endif
}

void CGraphics::giluQuadricCallback(GLUquadric* quad, GLenum which, GLvoid(*CallBackFunc)())
{
#ifndef WIN32
  gluQuadricCallback(quad, which, CallBackFunc);
#else
  //gluQuadricCallback(quad, which, (__stdcall *)CallBackFunc);
#endif
}

void CGraphics::giluTessCallback(GLUtesselator* tess, GLenum which, GLvoid(*CallBackFunc)())
{
#ifndef WIN32
  gluTessCallback(tess, which, CallBackFunc);
#else
  //gluTessCallback(tess, which, (__stdcall *)CallBackFunc);
#endif
}

