#ifndef __GRAPHICS_H__
#define __GRAPHICS_H__

///////////////////////////////////////////////////////////////////////////
// Include OpenGL to get all the types and constants
//

#ifdef WIN32
#include <wtypes.h>
#include <wingdi.h>
#endif

#include "genlib.h"
#include "mathlib.h"
#include <GL/gl.h>

////////////////////
// RenderingMode
//#define GL_RENDER                           0x1C00
//#define GL_FEEDBACK                         0x1C01
//#define GL_SELECT                           0x1C02
#define GL_RENDER_SOFTWARE                    0x1C03
#define GL_WRITE_SIF                          0x1C04
#define GL_WRITE_STL                          0x1C05
#define GL_WRITE_RIB                          0x1C06

#include <GL/glu.h>

#ifndef WIN32
#include <GL/glx.h>
#endif

///////////////////////////////////////////////////////////////////////////
// CGraphics Class
//

class CGraphics
{
public:
  CGraphics();
  ~CGraphics();
  VOID Init();
  VOID Uninit();

//////////////////////////////////////////////////
// OpenGL Functions which we are overriding
//
public:
  GLint gilRenderMode(GLenum mode);
  void gilBegin(GLenum mode);
  void gilEnd(void);
  void gilVertex3f(GLfloat x, GLfloat y, GLfloat z);
  void gilVertex3fv(const GLfloat *v);
  void gilVertex4f(GLfloat x, GLfloat y, GLfloat z, GLfloat w);
  void gilVertex4fv(const GLfloat *v);

//////////////////////////////////////////////////
// Additional Interface Functions which we need
//
public:
  VOID gilSetOutputFile(CHAR *pcOutputFile);

//////////////////////////////////////////////////
// Functions which implement our enhanced OpenGL interface functions
//
protected:
  // Helpers for -- GLint gilRenderMode(GLenum mode)
  VOID OpenOutputFile();
  VOID CloseOutputFile();
  VOID WriteFileHeader(GLenum mode);
  VOID WriteFileTail(GLenum mode);

  // Helpers for -- void gilBegin(GLenum mode)
  VOID WriteTriangle(VECTOR4D v1, VECTOR v2, VECTOR v3);
  VOID WriteSTLTriangle(Vector &v1, Vector &v2, Vector &v3);

//////////////////////////////////////////////////
// Member Variables
//
protected:
  // Variables for -- GLint gilRenderMode(GLenum mode)
  GLenum m_eRenderMode;

  CHAR m_pcOutputFile[256];
  FILE *m_fpOutputFile;

  // Variables for -- void gilBegin(GLenum mode)
  GLenum m_eGeometricPrimitiveMode;

  UINT_32 m_uiVertexCounter;
  UINT_32 m_uiVerticesBuffered;
  VECTOR4D m_pvVertices[4];

//////////////////////////////////////////////////
// OpenGL Interface -- "GL/gl.h"
//
public:
  void gilAccum(GLenum op, GLfloat value);
  void gilAlphaFunc(GLenum func, GLclampf ref);
  GLboolean gilAreTexturesResidentEXT(GLsizei n, const GLuint *textures, GLboolean *residences);
  void gilArrayElementEXT(GLint i);
  void gilBindTextureEXT(GLenum target, GLuint texture);
  void gilBitmap(GLsizei width, GLsizei height, GLfloat xorig, GLfloat yorig, GLfloat xmove, GLfloat ymove, const GLubyte *bitmap);
  void gilBlendColorEXT(GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha);
  void gilBlendEquationEXT(GLenum mode);
  void gilBlendFunc(GLenum sfactor, GLenum dfactor);
  void gilCallList(GLuint list);
  void gilCallLists(GLsizei n, GLenum type, const GLvoid *lists);
  void gilClear(GLbitfield mask);
  void gilClearAccum(GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha);
  void gilClearColor(GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha);
  void gilClearDepth(GLclampd depth);
  void gilClearIndex(GLfloat c);
  void gilClearStencil(GLint s);
  void gilClipPlane(GLenum plane, const GLdouble *equation);
  void gilColor3b(GLbyte red, GLbyte green, GLbyte blue);
  void gilColor3bv(const GLbyte *v);
  void gilColor3d(GLdouble red, GLdouble green, GLdouble blue);
  void gilColor3dv(const GLdouble *v);
  void gilColor3f(GLfloat red, GLfloat green, GLfloat blue);
  void gilColor3fv(const GLfloat *v);
  void gilColor3i(GLint red, GLint green, GLint blue);
  void gilColor3iv(const GLint *v);
  void gilColor3s(GLshort red, GLshort green, GLshort blue);
  void gilColor3sv(const GLshort *v);
  void gilColor3ub(GLubyte red, GLubyte green, GLubyte blue);
  void gilColor3ubv(const GLubyte *v);
  void gilColor3ui(GLuint red, GLuint green, GLuint blue);
  void gilColor3uiv(const GLuint *v);
  void gilColor3us(GLushort red, GLushort green, GLushort blue);
  void gilColor3usv(const GLushort *v);
  void gilColor4b(GLbyte red, GLbyte green, GLbyte blue, GLbyte alpha);
  void gilColor4bv(const GLbyte *v);
  void gilColor4d(GLdouble red, GLdouble green, GLdouble blue, GLdouble alpha);
  void gilColor4dv(const GLdouble *v);
  void gilColor4f(GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha);
  void gilColor4fv(const GLfloat *v);
  void gilColor4i(GLint red, GLint green, GLint blue, GLint alpha);
  void gilColor4iv(const GLint *v);
  void gilColor4s(GLshort red, GLshort green, GLshort blue, GLshort alpha);
  void gilColor4sv(const GLshort *v);
  void gilColor4ub(GLubyte red, GLubyte green, GLubyte blue, GLubyte alpha);
  void gilColor4ubv(const GLubyte *v);
  void gilColor4ui(GLuint red, GLuint green, GLuint blue, GLuint alpha);
  void gilColor4uiv(const GLuint *v);
  void gilColor4us(GLushort red, GLushort green, GLushort blue, GLushort alpha);
  void gilColor4usv(const GLushort *v);
  void gilColorMask(GLboolean red, GLboolean green, GLboolean blue, GLboolean alpha);
  void gilColorMaterial(GLenum face, GLenum mode);
  void gilColorPointerEXT(GLint size, GLenum type, GLsizei stride, GLsizei count, const GLvoid *pointer);
  void gilColorTableParameterfvSGI(GLenum target, GLenum pname, const GLfloat *params);
  void gilColorTableParameterivSGI(GLenum target, GLenum pname, const GLint *params);
  void gilColorTableSGI(GLenum target, GLenum internalformat, GLsizei width, GLenum format, GLenum type, const GLvoid *table);
  void gilConvolutionFilter1DEXT(GLenum target, GLenum internalformat, GLsizei width, GLenum format, GLenum type, const GLvoid *image);
  void gilConvolutionFilter2DEXT(GLenum target, GLenum internalformat, GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *image);
  void gilConvolutionParameterfEXT(GLenum target, GLenum pname, GLfloat params);
  void gilConvolutionParameterfvEXT(GLenum target, GLenum pname, const GLfloat *params);
  void gilConvolutionParameteriEXT(GLenum target, GLenum pname, GLint params);
  void gilConvolutionParameterivEXT(GLenum target, GLenum pname, const GLint *params);
  void gilCopyColorTableSGI(GLenum target, GLenum internalformat, GLint x, GLint y, GLsizei width);
  void gilCopyConvolutionFilter1DEXT(GLenum target, GLenum internalformat, GLint x, GLint y, GLsizei width);
  void gilCopyConvolutionFilter2DEXT(GLenum target, GLenum internalformat, GLint x, GLint y, GLsizei width, GLsizei height);
  void gilCopyPixels(GLint x, GLint y, GLsizei width, GLsizei height, GLenum type);
  void gilCopyTexImage1DEXT(GLenum target, GLint level, GLenum internalformat, GLint x, GLint y, GLsizei width, GLint border);
  void gilCopyTexImage2DEXT(GLenum target, GLint level, GLenum internalformat, GLint x, GLint y, GLsizei width, GLsizei height, GLint border);
  void gilCopyTexSubImage1DEXT(GLenum target, GLint level, GLint xoffset, GLint x, GLint y, GLsizei width);
  void gilCopyTexSubImage2DEXT(GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint x, GLint y, GLsizei width, GLsizei height);
  void gilCopyTexSubImage3DEXT(GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint zoffset, GLint x, GLint y, GLsizei width, GLsizei height);
  void gilCullFace(GLenum mode);
  void gilDeformSGIX(GLbitfield mask);
  void gilDeformationMap3dSGIX(GLenum target, GLdouble u1, GLdouble u2, GLint ustride, GLint uorder, GLdouble v1, GLdouble v2, GLint vstride, GLint vorder, GLdouble w1, GLdouble w2, GLint wstride, GLint worder, const GLdouble *points);
  void gilDeformationMap3fSGIX(GLenum target, GLfloat u1, GLfloat u2, GLint ustride, GLint uorder, GLfloat v1, GLfloat v2, GLint vstride, GLint vorder, GLfloat w1, GLfloat w2, GLint wstride, GLint worder, const GLfloat *points);
  void gilDeleteLists(GLuint list, GLsizei range);
  void gilDeleteTexturesEXT(GLsizei n, const GLuint *textures);
  void gilDepthFunc(GLenum func);
  void gilDepthMask(GLboolean flag);
  void gilDepthRange(GLclampd near, GLclampd far);
  void gilDetailTexFuncSGIS(GLenum target, GLsizei n, const GLfloat *points);
  void gilDisable(GLenum cap);
  void gilDrawArraysEXT(GLenum mode, GLint first, GLsizei count);
  void gilDrawBuffer(GLenum mode);
  void gilDrawPixels(GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *pixels);
  void gilEdgeFlag(GLboolean flag);
  void gilEdgeFlagPointerEXT(GLsizei stride, GLsizei count, const GLboolean *pointer);
  void gilEdgeFlagv(const GLboolean *flag);
  void gilEnable(GLenum cap);
  void gilEndList(void);
  void gilEvalCoord1d(GLdouble u);
  void gilEvalCoord1dv(const GLdouble *u);
  void gilEvalCoord1f(GLfloat u);
  void gilEvalCoord1fv(const GLfloat *u);
  void gilEvalCoord2d(GLdouble u, GLdouble v);
  void gilEvalCoord2dv(const GLdouble *u);
  void gilEvalCoord2f(GLfloat u, GLfloat v);
  void gilEvalCoord2fv(const GLfloat *u);
  void gilEvalMesh1(GLenum mode, GLint i1, GLint i2);
  void gilEvalMesh2(GLenum mode, GLint i1, GLint i2, GLint j1, GLint j2);
  void gilEvalPoint1(GLint i);
  void gilEvalPoint2(GLint i, GLint j);
  void gilFeedbackBuffer(GLsizei size, GLenum type, GLfloat *buffer);
  void gilFinish(void);
  void gilFlush(void);
  void gilFlushRasterSGIX(void);
  void gilFogFuncSGIS(GLsizei n, const GLfloat *points);
  void gilFogf(GLenum pname, GLfloat param);
  void gilFogfv(GLenum pname, const GLfloat *params);
  void gilFogi(GLenum pname, GLint param);
  void gilFogiv(GLenum pname, const GLint *params);
  void gilFrameZoomSGIX(GLint factor);
  void gilFrontFace(GLenum mode);
  void gilFrustum(GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble near, GLdouble far);
  GLuint gilGenLists(GLsizei range);
  void gilGenTexturesEXT(GLsizei n, GLuint *textures);
  void gilGetBooleanv(GLenum pname, GLboolean *params);
  void gilGetClipPlane(GLenum plane, GLdouble *equation);
  void gilGetColorTableParameterfvSGI(GLenum target, GLenum pname, GLfloat *params);
  void gilGetColorTableParameterivSGI(GLenum target, GLenum pname, GLint *params);
  void gilGetColorTableSGI(GLenum target, GLenum format, GLenum type, GLvoid *table);
  void gilGetConvolutionFilterEXT(GLenum target, GLenum format, GLenum type, GLvoid *image);
  void gilGetConvolutionParameterfvEXT(GLenum target, GLenum pname, GLfloat *params);
  void gilGetConvolutionParameterivEXT(GLenum target, GLenum pname, GLint *params);
  void gilGetDetailTexFuncSGIS(GLenum target, GLfloat *points);
  void gilGetDoublev(GLenum pname, GLdouble *params);
  GLenum gilGetError(void);
  void gilGetFloatv(GLenum pname, GLfloat *params);
  void gilGetHistogramEXT(GLenum target, GLboolean reset, GLenum format, GLenum type, GLvoid *values);
  void gilGetHistogramParameterfvEXT(GLenum target, GLenum pname, GLfloat *params);
  void gilGetHistogramParameterivEXT(GLenum target, GLenum pname, GLint *params);
  GLint gilGetInstrumentsSGIX(void);
  void gilGetIntegerv(GLenum pname, GLint *params);
  void gilGetLightfv(GLenum light, GLenum pname, GLfloat *params);
  void gilGetLightiv(GLenum light, GLenum pname, GLint *params);
  void gilGetMapdv(GLenum target, GLenum query, GLdouble *v);
  void gilGetMapfv(GLenum target, GLenum query, GLfloat *v);
  void gilGetMapiv(GLenum target, GLenum query, GLint *v);
  void gilGetMaterialfv(GLenum face, GLenum pname, GLfloat *params);
  void gilGetMaterialiv(GLenum face, GLenum pname, GLint *params);
  void gilGetMinmaxEXT(GLenum target, GLboolean reset, GLenum format, GLenum type, GLvoid *values);
  void gilGetMinmaxParameterfvEXT(GLenum target, GLenum pname, GLfloat *params);
  void gilGetMinmaxParameterivEXT(GLenum target, GLenum pname, GLint *params);
  void gilGetPixelMapfv(GLenum map, GLfloat *values);
  void gilGetPixelMapuiv(GLenum map, GLuint *values);
  void gilGetPixelMapusv(GLenum map, GLushort *values);
  void gilGetPointervEXT(GLenum pname, GLvoid* *params);
  void gilGetPolygonStipple(GLubyte *mask);
  void gilGetSeparableFilterEXT(GLenum target, GLenum format, GLenum type, GLvoid *row, GLvoid *column, GLvoid *span);
  void gilGetSharpenTexFuncSGIS(GLenum target, GLfloat *points);
  const GLubyte *gilGetString(GLenum name);
  void gilGetTexEnvfv(GLenum target, GLenum pname, GLfloat *params);
  void gilGetTexEnviv(GLenum target, GLenum pname, GLint *params);
  void gilGetTexFilterFuncSGIS(GLenum target, GLenum filter, GLfloat *weights);
  void gilGetTexGendv(GLenum coord, GLenum pname, GLdouble *params);
  void gilGetTexGenfv(GLenum coord, GLenum pname, GLfloat *params);
  void gilGetTexGeniv(GLenum coord, GLenum pname, GLint *params);
  void gilGetTexImage(GLenum target, GLint level, GLenum format, GLenum type, GLvoid *pixels);
  void gilGetTexLevelParameterfv(GLenum target, GLint level, GLenum pname, GLfloat *params);
  void gilGetTexLevelParameteriv(GLenum target, GLint level, GLenum pname, GLint *params);
  void gilGetTexParameterfv(GLenum target, GLenum pname, GLfloat *params);
  void gilGetTexParameteriv(GLenum target, GLenum pname, GLint *params);
  void gilHint(GLenum target, GLenum mode);
  void gilHistogramEXT(GLenum target, GLsizei width, GLenum internalformat, GLboolean sink);
  void gilIndexMask(GLuint mask);
  void gilIndexPointerEXT(GLenum type, GLsizei stride, GLsizei count, const GLvoid *pointer);
  void gilIndexd(GLdouble c);
  void gilIndexdv(const GLdouble *c);
  void gilIndexf(GLfloat c);
  void gilIndexfv(const GLfloat *c);
  void gilIndexi(GLint c);
  void gilIndexiv(const GLint *c);
  void gilIndexs(GLshort c);
  void gilIndexsv(const GLshort *c);
  void gilInitNames(void);
  GLboolean gilIsEnabled(GLenum cap);
  GLboolean gilIsList(GLuint list);
  GLboolean gilIsTextureEXT(GLuint texture);
  void gilLightModelf(GLenum pname, GLfloat param);
  void gilLightModelfv(GLenum pname, const GLfloat *params);
  void gilLightModeli(GLenum pname, GLint param);
  void gilLightModeliv(GLenum pname, const GLint *params);
  void gilLightf(GLenum light, GLenum pname, GLfloat param);
  void gilLightfv(GLenum light, GLenum pname, const GLfloat *params);
  void gilLighti(GLenum light, GLenum pname, GLint param);
  void gilLightiv(GLenum light, GLenum pname, const GLint *params);
  void gilLineStipple(GLint factor, GLushort pattern);
  void gilLineWidth(GLfloat width);
  void gilListBase(GLuint base);
  void gilLoadIdentity(void);
  void gilLoadIdentityDeformationMapSGIX(GLbitfield mask);
  void gilLoadMatrixd(const GLdouble *m);
  void gilLoadMatrixf(const GLfloat *m);
  void gilLoadName(GLuint name);
  void gilLogicOp(GLenum opcode);
  void gilMap1d(GLenum target, GLdouble u1, GLdouble u2, GLint stride, GLint order, const GLdouble *points);
  void gilMap1f(GLenum target, GLfloat u1, GLfloat u2, GLint stride, GLint order, const GLfloat *points);
  void gilMap2d(GLenum target, GLdouble u1, GLdouble u2, GLint ustride, GLint uorder, GLdouble v1, GLdouble v2, GLint vstride, GLint vorder, const GLdouble *points);
  void gilMap2f(GLenum target, GLfloat u1, GLfloat u2, GLint ustride, GLint uorder, GLfloat v1, GLfloat v2, GLint vstride, GLint vorder, const GLfloat *points);
  void gilMapGrid1d(GLint un, GLdouble u1, GLdouble u2);
  void gilMapGrid1f(GLint un, GLfloat u1, GLfloat u2);
  void gilMapGrid2d(GLint un, GLdouble u1, GLdouble u2, GLint vn, GLdouble v1, GLdouble v2);
  void gilMapGrid2f(GLint un, GLfloat u1, GLfloat u2, GLint vn, GLfloat v1, GLfloat v2);
  void gilMaterialf(GLenum face, GLenum pname, GLfloat param);
  void gilMaterialfv(GLenum face, GLenum pname, const GLfloat *params);
  void gilMateriali(GLenum face, GLenum pname, GLint param);
  void gilMaterialiv(GLenum face, GLenum pname, const GLint *params);
  void gilMatrixMode(GLenum mode);
  void gilMinmaxEXT(GLenum target, GLenum internalformat, GLboolean sink);
  void gilMultMatrixd(const GLdouble *m);
  void gilMultMatrixf(const GLfloat *m);
  void gilNewList(GLuint list, GLenum mode);
  void gilNormal3b(GLbyte nx, GLbyte ny, GLbyte nz);
  void gilNormal3bv(const GLbyte *v);
  void gilNormal3d(GLdouble nx, GLdouble ny, GLdouble nz);
  void gilNormal3dv(const GLdouble *v);
  void gilNormal3f(GLfloat nx, GLfloat ny, GLfloat nz);
  void gilNormal3fv(const GLfloat *v);
  void gilNormal3i(GLint nx, GLint ny, GLint nz);
  void gilNormal3iv(const GLint *v);
  void gilNormal3s(GLshort nx, GLshort ny, GLshort nz);
  void gilNormal3sv(const GLshort *v);
  void gilNormalPointerEXT(GLenum type, GLsizei stride, GLsizei count, const GLvoid *pointer);
  void gilOrtho(GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble near, GLdouble far);
  void gilPassThrough(GLfloat token);
  void gilPipelineInstrumentsBufferSGIX(GLsizei size, GLfloat *buffer);
  void gilPixelMapfv(GLenum map, GLint mapsize, const GLfloat *values);
  void gilPixelMapuiv(GLenum map, GLint mapsize, const GLuint *values);
  void gilPixelMapusv(GLenum map, GLint mapsize, const GLushort *values);
  void gilPixelStoref(GLenum pname, GLfloat param);
  void gilPixelStorei(GLenum pname, GLint param);
  void gilPixelTexGenSGIX(GLenum mode);
  void gilPixelTransferf(GLenum pname, GLfloat param);
  void gilPixelTransferi(GLenum pname, GLint param);
  void gilPixelZoom(GLfloat xfactor, GLfloat yfactor);
  void gilPointParameterfSGIS(GLenum pname, GLfloat param);
  void gilPointParameterfvSGIS(GLenum pname, const GLfloat *params);
  void gilPointSize(GLfloat size);
  GLint gilPollInstrumentsSGIX(GLint *marker_p);
  void gilPolygonMode(GLenum face, GLenum mode);
  void gilPolygonOffsetEXT(GLfloat factor, GLfloat bias);
  void gilPolygonStipple(const GLubyte *mask);
  void gilPopAttrib(void);
  void gilPopMatrix(void);
  void gilPopName(void);
  void gilPrioritizeTexturesEXT(GLsizei n, const GLuint *textures, const GLclampf *priorities);
  void gilPushAttrib(GLbitfield mask);
  void gilPushMatrix(void);
  void gilPushName(GLuint name);
  void gilRasterPos2d(GLdouble x, GLdouble y);
  void gilRasterPos2dv(const GLdouble *v);
  void gilRasterPos2f(GLfloat x, GLfloat y);
  void gilRasterPos2fv(const GLfloat *v);
  void gilRasterPos2i(GLint x, GLint y);
  void gilRasterPos2iv(const GLint *v);
  void gilRasterPos2s(GLshort x, GLshort y);
  void gilRasterPos2sv(const GLshort *v);
  void gilRasterPos3d(GLdouble x, GLdouble y, GLdouble z);
  void gilRasterPos3dv(const GLdouble *v);
  void gilRasterPos3f(GLfloat x, GLfloat y, GLfloat z);
  void gilRasterPos3fv(const GLfloat *v);
  void gilRasterPos3i(GLint x, GLint y, GLint z);
  void gilRasterPos3iv(const GLint *v);
  void gilRasterPos3s(GLshort x, GLshort y, GLshort z);
  void gilRasterPos3sv(const GLshort *v);
  void gilRasterPos4d(GLdouble x, GLdouble y, GLdouble z, GLdouble w);
  void gilRasterPos4dv(const GLdouble *v);
  void gilRasterPos4f(GLfloat x, GLfloat y, GLfloat z, GLfloat w);
  void gilRasterPos4fv(const GLfloat *v);
  void gilRasterPos4i(GLint x, GLint y, GLint z, GLint w);
  void gilRasterPos4iv(const GLint *v);
  void gilRasterPos4s(GLshort x, GLshort y, GLshort z, GLshort w);
  void gilRasterPos4sv(const GLshort *v);
  void gilReadBuffer(GLenum mode);
  void gilReadPixels(GLint x, GLint y, GLsizei width, GLsizei height, GLenum format, GLenum type, GLvoid *pixels);
  void gilRectd(GLdouble x1, GLdouble y1, GLdouble x2, GLdouble y2);
  void gilRectdv(const GLdouble *v1, const GLdouble *v2);
  void gilRectf(GLfloat x1, GLfloat y1, GLfloat x2, GLfloat y2);
  void gilRectfv(const GLfloat *v1, const GLfloat *v2);
  void gilRecti(GLint x1, GLint y1, GLint x2, GLint y2);
  void gilRectiv(const GLint *v1, const GLint *v2);
  void gilRects(GLshort x1, GLshort y1, GLshort x2, GLshort y2);
  void gilRectsv(const GLshort *v1, const GLshort *v2);
  void gilReferencePlaneSGIX(const GLdouble *equation);
  void gilResetHistogramEXT(GLenum target);
  void gilResetMinmaxEXT(GLenum target);
  void gilRotated(GLdouble angle, GLdouble x, GLdouble y, GLdouble z);
  void gilRotatef(GLfloat angle, GLfloat x, GLfloat y, GLfloat z);
  void gilSampleMaskSGIS(GLclampf value, GLboolean invert);
  void gilSamplePatternSGIS(GLenum pattern);
  void gilScaled(GLdouble x, GLdouble y, GLdouble z);
  void gilScalef(GLfloat x, GLfloat y, GLfloat z);
  void gilScissor(GLint x, GLint y, GLsizei width, GLsizei height);
  void gilSelectBuffer(GLsizei size, GLuint *buffer);
  void gilSeparableFilter2DEXT(GLenum target, GLenum internalformat, GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *row, const GLvoid *column);
  void gilShadeModel(GLenum mode);
  void gilSharpenTexFuncSGIS(GLenum target, GLsizei n, const GLfloat *points);
  void gilSpriteParameterfSGIX(GLenum pname, GLfloat param);
  void gilSpriteParameterfvSGIX(GLenum pname, const GLfloat *params);
  void gilSpriteParameteriSGIX(GLenum pname, GLint param);
  void gilSpriteParameterivSGIX(GLenum pname, const GLint *params);
  void gilStartInstrumentsSGIX(void);
  void gilStencilFunc(GLenum func, GLint ref, GLuint mask);
  void gilStencilMask(GLuint mask);
  void gilStencilOp(GLenum fail, GLenum zfail, GLenum zpass);
  void gilStopInstrumentsSGIX(GLint marker);
  void gilTagSampleBufferSGIX(void);
  void gilTexCoord1d(GLdouble s);
  void gilTexCoord1dv(const GLdouble *v);
  void gilTexCoord1f(GLfloat s);
  void gilTexCoord1fv(const GLfloat *v);
  void gilTexCoord1i(GLint s);
  void gilTexCoord1iv(const GLint *v);
  void gilTexCoord1s(GLshort s);
  void gilTexCoord1sv(const GLshort *v);
  void gilTexCoord2d(GLdouble s, GLdouble t);
  void gilTexCoord2dv(const GLdouble *v);
  void gilTexCoord2f(GLfloat s, GLfloat t);
  void gilTexCoord2fv(const GLfloat *v);
  void gilTexCoord2i(GLint s, GLint t);
  void gilTexCoord2iv(const GLint *v);
  void gilTexCoord2s(GLshort s, GLshort t);
  void gilTexCoord2sv(const GLshort *v);
  void gilTexCoord3d(GLdouble s, GLdouble t, GLdouble r);
  void gilTexCoord3dv(const GLdouble *v);
  void gilTexCoord3f(GLfloat s, GLfloat t, GLfloat r);
  void gilTexCoord3fv(const GLfloat *v);
  void gilTexCoord3i(GLint s, GLint t, GLint r);
  void gilTexCoord3iv(const GLint *v);
  void gilTexCoord3s(GLshort s, GLshort t, GLshort r);
  void gilTexCoord3sv(const GLshort *v);
  void gilTexCoord4d(GLdouble s, GLdouble t, GLdouble r, GLdouble q);
  void gilTexCoord4dv(const GLdouble *v);
  void gilTexCoord4f(GLfloat s, GLfloat t, GLfloat r, GLfloat q);
  void gilTexCoord4fv(const GLfloat *v);
  void gilTexCoord4i(GLint s, GLint t, GLint r, GLint q);
  void gilTexCoord4iv(const GLint *v);
  void gilTexCoord4s(GLshort s, GLshort t, GLshort r, GLshort q);
  void gilTexCoord4sv(const GLshort *v);
  void gilTexCoordPointerEXT(GLint size, GLenum type, GLsizei stride, GLsizei count, const GLvoid *pointer);
  void gilTexEnvf(GLenum target, GLenum pname, GLfloat param);
  void gilTexEnvfv(GLenum target, GLenum pname, const GLfloat *params);
  void gilTexEnvi(GLenum target, GLenum pname, GLint param);
  void gilTexEnviv(GLenum target, GLenum pname, const GLint *params);
  void gilTexFilterFuncSGIS(GLenum target, GLenum filter, GLsizei n, const GLfloat *weights);
  void gilTexGend(GLenum coord, GLenum pname, GLdouble param);
  void gilTexGendv(GLenum coord, GLenum pname, const GLdouble *params);
  void gilTexGenf(GLenum coord, GLenum pname, GLfloat param);
  void gilTexGenfv(GLenum coord, GLenum pname, const GLfloat *params);
  void gilTexGeni(GLenum coord, GLenum pname, GLint param);
  void gilTexGeniv(GLenum coord, GLenum pname, const GLint *params);
  void gilTexImage1D(GLenum target, GLint level, GLint components, GLsizei width, GLint border, GLenum format, GLenum type, const GLvoid *pixels);
  void gilTexImage2D(GLenum target, GLint level, GLint components, GLsizei width, GLsizei height, GLint border, GLenum format, GLenum type, const GLvoid *pixels);
  void gilTexImage3DEXT(GLenum target, GLint level, GLenum internalformat, GLsizei width, GLsizei height, GLsizei depth, GLint border, GLenum format, GLenum type, const GLvoid *pixels);
  void gilTexImage4DSGIS(GLenum target, GLint level, GLenum internalformat, GLsizei width, GLsizei height, GLsizei depth, GLsizei size4d, GLint border, GLenum format, GLenum type, const GLvoid *pixels);
  void gilTexParameterf(GLenum target, GLenum pname, GLfloat param);
  void gilTexParameterfv(GLenum target, GLenum pname, const GLfloat *params);
  void gilTexParameteri(GLenum target, GLenum pname, GLint param);
  void gilTexParameteriv(GLenum target, GLenum pname, const GLint *params);
  void gilTexSubImage1DEXT(GLenum target, GLint level, GLint xoffset, GLsizei width, GLenum format, GLenum type, const GLvoid *pixels);
  void gilTexSubImage2DEXT(GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *pixels);
  void gilTexSubImage3DEXT(GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint zoffset, GLsizei width, GLsizei height, GLsizei depth, GLenum format, GLenum type, const GLvoid *pixels);
  void gilTexSubImage4DSGIS(GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint zoffset, GLint woffset, GLsizei width, GLsizei height, GLsizei depth, GLsizei size4d, GLenum format, GLenum type, const GLvoid *pixels);
  void gilTranslated(GLdouble x, GLdouble y, GLdouble z);
  void gilTranslatef(GLfloat x, GLfloat y, GLfloat z);
  void gilVertex2d(GLdouble x, GLdouble y);
  void gilVertex2dv(const GLdouble *v);
  void gilVertex2f(GLfloat x, GLfloat y);
  void gilVertex2fv(const GLfloat *v);
  void gilVertex2i(GLint x, GLint y);
  void gilVertex2iv(const GLint *v);
  void gilVertex2s(GLshort x, GLshort y);
  void gilVertex2sv(const GLshort *v);
  void gilVertex3d(GLdouble x, GLdouble y, GLdouble z);
  void gilVertex3dv(const GLdouble *v);
  void gilVertex3i(GLint x, GLint y, GLint z);
  void gilVertex3iv(const GLint *v);
  void gilVertex3s(GLshort x, GLshort y, GLshort z);
  void gilVertex3sv(const GLshort *v);
  void gilVertex4d(GLdouble x, GLdouble y, GLdouble z, GLdouble w);
  void gilVertex4dv(const GLdouble *v);
  void gilVertex4i(GLint x, GLint y, GLint z, GLint w);
  void gilVertex4iv(const GLint *v);
  void gilVertex4s(GLshort x, GLshort y, GLshort z, GLshort w);
  void gilVertex4sv(const GLshort *v);
  void gilVertexPointerEXT(GLint size, GLenum type, GLsizei stride, GLsizei count, const GLvoid *pointer);
  void gilViewport(GLint x, GLint y, GLsizei width, GLsizei height);

//////////////////////////////////////////////////
// OpenGL Interface -- "GL/glu.h"
//
public:
  void giluDisk(GLUquadric* quad, GLdouble inner, GLdouble outer, GLint slices, GLint loops);
  void giluEndCurve(GLUnurbs* nurb);
  void giluEndPolygon(GLUtesselator* tess);
  void giluEndSurface(GLUnurbs* nurb);
  void giluEndTrim(GLUnurbs* nurb);
  const GLubyte* giluErrorString(GLenum error);
  void giluGetNurbsProperty(GLUnurbs* nurb, GLenum property, GLfloat* data);
  const GLubyte* giluGetString(GLenum name);
  void giluGetTessProperty(GLUtesselator* tess, GLenum which, GLdouble* data);
  void giluLoadSamplingMatrices(GLUnurbs* nurb, const GLfloat* model, const GLfloat* perspective, const GLint* view);
  void giluLookAt(GLdouble eyeX, GLdouble eyeY, GLdouble eyeZ, GLdouble centerX, GLdouble centerY, GLdouble centerZ, GLdouble upX, GLdouble upY, GLdouble upZ);
  GLUnurbs* giluNewNurbsRenderer();
  GLUquadric* giluNewQuadric();
  GLUtesselator* giluNewTess();
  void giluNextContour(GLUtesselator* tess, GLenum type);
  void giluNurbsCallback(GLUnurbs* nurb, GLenum which, GLvoid(*CallBackFunc)());
  void giluNurbsCurve(GLUnurbs* nurb, GLint knotCount, GLfloat* knots, GLint stride, GLfloat* control, GLint order, GLenum type);
  void giluNurbsProperty(GLUnurbs* nurb, GLenum property, GLfloat value);
  void giluNurbsSurface(GLUnurbs* nurb, GLint sKnotCount, GLfloat* sKnots, GLint tKnotCount, GLfloat* tKnots, GLint sStride, GLint tStride, GLfloat* control, GLint sOrder, GLint tOrder, GLenum type);
  void giluOrtho2D(GLdouble left, GLdouble right, GLdouble bottom, GLdouble top);
  void giluPartialDisk(GLUquadric* quad, GLdouble inner, GLdouble outer, GLint slices, GLint loops, GLdouble start, GLdouble sweep);
  void giluPerspective(GLdouble fovy, GLdouble aspect, GLdouble zNear, GLdouble zFar);
  void giluPickMatrix(GLdouble x, GLdouble y, GLdouble delX, GLdouble delY, GLint* viewport);
  GLint giluProject(GLdouble objX, GLdouble objY, GLdouble objZ, const GLdouble* model, const GLdouble* proj, const GLint* view, GLdouble* winX, GLdouble* winY, GLdouble* winZ);
  void giluPwlCurve(GLUnurbs* nurb, GLint count, GLfloat* data, GLint stride, GLenum type);
  void giluQuadricCallback(GLUquadric* quad, GLenum which, GLvoid(*CallBackFunc)());
  void giluQuadricDrawStyle(GLUquadric* quad, GLenum draw);
  void giluQuadricNormals(GLUquadric* quad, GLenum normal);
  void giluQuadricOrientation(GLUquadric* quad, GLenum orientation);
  void giluQuadricTexture(GLUquadric* quad, GLboolean texture);
  GLint giluScaleImage(GLenum format, GLsizei wIn, GLsizei hIn, GLenum typeIn, const void* dataIn, GLsizei wOut, GLsizei hOut, GLenum typeOut, GLvoid* dataOut);
  void giluSphere(GLUquadric* quad, GLdouble radius, GLint slices, GLint stacks);
  void giluTessBeginContour(GLUtesselator* tess);
  void giluTessBeginPolygon(GLUtesselator* tess, GLvoid* data);
  void giluTessCallback(GLUtesselator* tess, GLenum which, GLvoid(*CallBackFunc)());
  void giluTessEndContour(GLUtesselator* tess);
  void giluTessEndPolygon(GLUtesselator* tess);
  void giluTessNormal(GLUtesselator* tess, GLdouble valueX, GLdouble valueY, GLdouble valueZ);
  void giluTessProperty(GLUtesselator* tess, GLenum which, GLdouble data);
  void giluTessVertex(GLUtesselator* tess, GLdouble* location, GLvoid* data);
  GLint giluUnProject(GLdouble winX, GLdouble winY, GLdouble winZ, const GLdouble* model, const GLdouble* proj, const GLint* view, GLdouble* objX, GLdouble* objY, GLdouble* objZ);
};



// fix for systems like win32 without the EXT versions of the gl funcs
#ifdef WIN32

void donothing (...)
{
	printf ("Unsupported GL function called!\n");
};

int donothing2 (...)
{
	printf ("Unsupported GL function called!");
	return 0;
};

#define glAreTexturesResidentEXT glAreTexturesResident
#define glArrayElementEXT glArrayElement 
#define glBindTextureEXT glBindTexture
#define glBlendColorEXT donothing
#define glBlendEquationEXT donothing
#define glColorPointerEXT donothing
#define glColorTableParameterfvSGI donothing
#define glColorTableParameterivSGI donothing
#define glColorTableSGI donothing
#define glConvolutionFilter1DEXT donothing
#define glConvolutionFilter2DEXT donothing
#define glConvolutionParameterfEXT donothing
#define glConvolutionParameterfvEXT donothing
#define glConvolutionParameteriEXT donothing
#define glConvolutionParameterivEXT  donothing
#define glCopyColorTableSGI donothing
#define glCopyConvolutionFilter1DEXT donothing
#define glCopyConvolutionFilter2DEXT donothing
#define glCopyTexImage1DEXT glCopyTexImage1D
#define glCopyTexImage2DEXT glCopyTexImage2D
#define glCopyTexSubImage1DEXT donothing
#define glCopyTexSubImage2DEXT glCopyTexSubImage2D
#define glCopyTexSubImage3DEXT donothing
#define glDeformSGIX donothing
#define glDeformationMap3dSGIX donothing
#define glDeformationMap3fSGIX donothing
#define glDeleteTexturesEXT glDeleteTextures
#define glDepthRange donothing
#define glDetailTexFuncSGIS donothing
#define glDrawArraysEXT glDrawArrays
#define glEdgeFlagPointerEXT donothing
#define glFlushRasterSGIX donothing
#define glFogFuncSGIS donothing
#define glFrameZoomSGIX donothing
#define glGenTexturesEXT glGenTextures
#define glGetColorTableParameterfvSGI donothing
#define glGetColorTableParameterivSGI donothing
#define glGetColorTableSGI donothing
#define glGetConvolutionFilterEXT donothing
#define glGetConvolutionParameterfvEXT donothing
#define glGetConvolutionParameterivEXT donothing
#define glGetDetailTexFuncSGIS donothing
#define glGetHistogramEXT donothing
#define glGetHistogramParameterfvEXT donothing
#define glGetHistogramParameterivEXT donothing
#define glGetInstrumentsSGIX donothing2
#define glGetMinmaxEXT donothing
#define glGetMinmaxParameterfvEXT donothing
#define glGetMinmaxParameterivEXT donothing
#define glGetPointervEXT glGetPointerv
#define glGetSeparableFilterEXT donothing
#define glGetSharpenTexFuncSGIS donothing
#define glGetTexFilterFuncSGIS donothing
#define glHistogramEXT donothing
#define glIndexPointerEXT donothing
#define glIsTextureEXT glIsTexture
#define glLoadIdentityDeformationMapSGIX donothing
#define glMinmaxEXT donothing
#define glNormalPointerEXT donothing
#define glPipelineInstrumentsBufferSGIX donothing
#define glPixelTexGenSGIX donothing
#define glPointParameterfSGIS donothing
#define glPointParameterfvSGIS donothing
#define glPollInstrumentsSGIX donothing2
#define glPolygonOffsetEXT glPolygonOffset
#define glPrioritizeTexturesEXT glPrioritizeTextures
#define glReferencePlaneSGIX donothing
#define glResetHistogramEXT donothing
#define glResetMinmaxEXT donothing
#define glSampleMaskSGIS donothing
#define glSamplePatternSGIS donothing
#define glSeparableFilter2DEXT donothing
#define glSharpenTexFuncSGIS donothing
#define glSpriteParameterfSGIX donothing
#define glSpriteParameterfvSGIX donothing
#define glSpriteParameteriSGIX donothing
#define glSpriteParameterivSGIX donothing
#define glStartInstrumentsSGIX donothing
#define glStopInstrumentsSGIX donothing
#define glTagSampleBufferSGIX donothing
#define glTexCoordPointerEXT donothing
#define glTexFilterFuncSGIS donothing
#define glTexImage3DEXT donothing
#define glTexImage4DSGIS donothing
#define glTexSubImage1DEXT glTexSubImage1D
#define glTexSubImage2DEXT glTexSubImage2D
#define glTexSubImage3DEXT donothing
#define glTexSubImage4DSGIS donothing
#define glVertexPointerEXT donothing

#endif // win32



///////////////////////////////////////////////////////////////////////////
// Inlined Functions
//

//////////////////////////////////////////////////
// OpenGL Interface -- "GL/gl.h"
//

inline void CGraphics::gilAccum(GLenum op, GLfloat value)
{
  glAccum(op, value);
}

inline void CGraphics::gilAlphaFunc(GLenum func, GLclampf ref)
{
  glAlphaFunc(func, ref);
}

inline GLboolean CGraphics::gilAreTexturesResidentEXT(GLsizei n, const GLuint *textures, GLboolean *residences)
{
  return glAreTexturesResidentEXT(n, textures, residences);
}

inline void CGraphics::gilArrayElementEXT(GLint i)
{
  glArrayElementEXT(i);
}

inline void CGraphics::gilBindTextureEXT(GLenum target, GLuint texture)
{
  glBindTextureEXT(target, texture);
}

inline void CGraphics::gilBitmap(GLsizei width, GLsizei height, GLfloat xorig, GLfloat yorig, GLfloat xmove, GLfloat ymove, const GLubyte *bitmap)
{
  glBitmap(width, height, xorig, yorig, xmove, ymove, bitmap);
}

inline void CGraphics::gilBlendColorEXT(GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha)
{
  glBlendColorEXT(red, green, blue, alpha);
}

inline void CGraphics::gilBlendEquationEXT(GLenum mode)
{
  glBlendEquationEXT(mode);
}

inline void CGraphics::gilBlendFunc(GLenum sfactor, GLenum dfactor)
{
  glBlendFunc(sfactor, dfactor);
}

inline void CGraphics::gilCallList(GLuint list)
{
  glCallList(list);
}

inline void CGraphics::gilCallLists(GLsizei n, GLenum type, const GLvoid *lists)
{
  glCallLists(n, type, lists);
}

inline void CGraphics::gilClear(GLbitfield mask)
{
  glClear(mask);
}

inline void CGraphics::gilClearAccum(GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha)
{
  glClearAccum(red, green, blue, alpha);
}

inline void CGraphics::gilClearColor(GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha)
{
  glClearColor(red, green, blue, alpha);
}

inline void CGraphics::gilClearDepth(GLclampd depth)
{
  glClearDepth(depth);
}

inline void CGraphics::gilClearIndex(GLfloat c)
{
  glClearIndex(c);
}

inline void CGraphics::gilClearStencil(GLint s)
{
  glClearStencil(s);
}

inline void CGraphics::gilClipPlane(GLenum plane, const GLdouble *equation)
{
  glClipPlane(plane, equation);
}

inline void CGraphics::gilColor3b(GLbyte red, GLbyte green, GLbyte blue)
{
  glColor3b(red, green, blue);
}

inline void CGraphics::gilColor3bv(const GLbyte *v)
{
  glColor3bv(v);
}

inline void CGraphics::gilColor3d(GLdouble red, GLdouble green, GLdouble blue)
{
  glColor3d(red, green, blue);
}

inline void CGraphics::gilColor3dv(const GLdouble *v)
{
  glColor3dv(v);
}

inline void CGraphics::gilColor3f(GLfloat red, GLfloat green, GLfloat blue)
{
  glColor3f(red, green, blue);
}

inline void CGraphics::gilColor3fv(const GLfloat *v)
{
  glColor3fv(v);
}

inline void CGraphics::gilColor3i(GLint red, GLint green, GLint blue)
{
  glColor3i(red, green, blue);
}

inline void CGraphics::gilColor3iv(const GLint *v)
{
  glColor3iv(v);
}

inline void CGraphics::gilColor3s(GLshort red, GLshort green, GLshort blue)
{
  glColor3s(red, green, blue);
}

inline void CGraphics::gilColor3sv(const GLshort *v)
{
  glColor3sv(v);
}

inline void CGraphics::gilColor3ub(GLubyte red, GLubyte green, GLubyte blue)
{
  glColor3ub(red, green, blue);
}

inline void CGraphics::gilColor3ubv(const GLubyte *v)
{
  glColor3ubv(v);
}

inline void CGraphics::gilColor3ui(GLuint red, GLuint green, GLuint blue)
{
  glColor3ui(red, green, blue);
}

inline void CGraphics::gilColor3uiv(const GLuint *v)
{
  glColor3uiv(v);
}

inline void CGraphics::gilColor3us(GLushort red, GLushort green, GLushort blue)
{
  glColor3us(red, green, blue);
}

inline void CGraphics::gilColor3usv(const GLushort *v)
{
  glColor3usv(v);
}

inline void CGraphics::gilColor4b(GLbyte red, GLbyte green, GLbyte blue, GLbyte alpha)
{
  glColor4b(red, green, blue, alpha);
}

inline void CGraphics::gilColor4bv(const GLbyte *v)
{
  glColor4bv(v);
}

inline void CGraphics::gilColor4d(GLdouble red, GLdouble green, GLdouble blue, GLdouble alpha)
{
  glColor4d(red, green, blue, alpha);
}

inline void CGraphics::gilColor4dv(const GLdouble *v)
{
  glColor4dv(v);
}

inline void CGraphics::gilColor4f(GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha)
{
  glColor4f(red, green, blue, alpha);
}

inline void CGraphics::gilColor4fv(const GLfloat *v)
{
  glColor4fv(v);
}

inline void CGraphics::gilColor4i(GLint red, GLint green, GLint blue, GLint alpha)
{
  glColor4i(red, green, blue, alpha);
}

inline void CGraphics::gilColor4iv(const GLint *v)
{
  glColor4iv(v);
}

inline void CGraphics::gilColor4s(GLshort red, GLshort green, GLshort blue, GLshort alpha)
{
  glColor4s(red, green, blue, alpha);
}

inline void CGraphics::gilColor4sv(const GLshort *v)
{
  glColor4sv(v);
}

inline void CGraphics::gilColor4ub(GLubyte red, GLubyte green, GLubyte blue, GLubyte alpha)
{
  glColor4ub(red, green, blue, alpha);
}

inline void CGraphics::gilColor4ubv(const GLubyte *v)
{
  glColor4ubv(v);
}

inline void CGraphics::gilColor4ui(GLuint red, GLuint green, GLuint blue, GLuint alpha)
{
  glColor4ui(red, green, blue, alpha);
}

inline void CGraphics::gilColor4uiv(const GLuint *v)
{
  glColor4uiv(v);
}

inline void CGraphics::gilColor4us(GLushort red, GLushort green, GLushort blue, GLushort alpha)
{
  glColor4us(red, green, blue, alpha);
}

inline void CGraphics::gilColor4usv(const GLushort *v)
{
  glColor4usv(v);
}

inline void CGraphics::gilColorMask(GLboolean red, GLboolean green, GLboolean blue, GLboolean alpha)
{
  glColorMask(red, green, blue, alpha);
}

inline void CGraphics::gilColorMaterial(GLenum face, GLenum mode)
{
  glColorMaterial(face, mode);
}

inline void CGraphics::gilColorPointerEXT(GLint size, GLenum type, GLsizei stride, GLsizei count, const GLvoid *pointer)
{
  glColorPointerEXT(size, type, stride, count, pointer);
}

inline void CGraphics::gilColorTableParameterfvSGI(GLenum target, GLenum pname, const GLfloat *params)
{
  glColorTableParameterfvSGI(target, pname, params);
}

inline void CGraphics::gilColorTableParameterivSGI(GLenum target, GLenum pname, const GLint *params)
{
  glColorTableParameterivSGI(target, pname, params);
}

inline void CGraphics::gilColorTableSGI(GLenum target, GLenum internalformat, GLsizei width, GLenum format, GLenum type, const GLvoid *table)
{
  glColorTableSGI(target, internalformat, width, format, type, table);
}

inline void CGraphics::gilConvolutionFilter1DEXT(GLenum target, GLenum internalformat, GLsizei width, GLenum format, GLenum type, const GLvoid *image)
{
  glConvolutionFilter1DEXT(target, internalformat, width, format, type, image);
}

inline void CGraphics::gilConvolutionFilter2DEXT(GLenum target, GLenum internalformat, GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *image)
{
  glConvolutionFilter2DEXT(target, internalformat, width, height, format, type, image);
}

inline void CGraphics::gilConvolutionParameterfEXT(GLenum target, GLenum pname, GLfloat params)
{
  glConvolutionParameterfEXT(target, pname, params);
}

inline void CGraphics::gilConvolutionParameterfvEXT(GLenum target, GLenum pname, const GLfloat *params)
{
  glConvolutionParameterfvEXT(target, pname, params);
}

inline void CGraphics::gilConvolutionParameteriEXT(GLenum target, GLenum pname, GLint params)
{
  glConvolutionParameteriEXT(target, pname, params);
}

inline void CGraphics::gilConvolutionParameterivEXT(GLenum target, GLenum pname, const GLint *params)
{
  glConvolutionParameterivEXT(target, pname, params);
}

inline void CGraphics::gilCopyColorTableSGI(GLenum target, GLenum internalformat, GLint x, GLint y, GLsizei width)
{
  glCopyColorTableSGI(target, internalformat, x, y, width);
}

inline void CGraphics::gilCopyConvolutionFilter1DEXT(GLenum target, GLenum internalformat, GLint x, GLint y, GLsizei width)
{
  glCopyConvolutionFilter1DEXT(target, internalformat, x, y, width);
}

inline void CGraphics::gilCopyConvolutionFilter2DEXT(GLenum target, GLenum internalformat, GLint x, GLint y, GLsizei width, GLsizei height)
{
  glCopyConvolutionFilter2DEXT(target, internalformat, x, y, width, height);
}

inline void CGraphics::gilCopyPixels(GLint x, GLint y, GLsizei width, GLsizei height, GLenum type)
{
  glCopyPixels(x, y, width, height, type);
}

inline void CGraphics::gilCopyTexImage1DEXT(GLenum target, GLint level, GLenum internalformat, GLint x, GLint y, GLsizei width, GLint border)
{
  glCopyTexImage1DEXT(target, level, internalformat, x, y, width, border);
}

inline void CGraphics::gilCopyTexImage2DEXT(GLenum target, GLint level, GLenum internalformat, GLint x, GLint y, GLsizei width, GLsizei height, GLint border)
{
  glCopyTexImage2DEXT(target, level, internalformat, x, y, width, height, border);
}

inline void CGraphics::gilCopyTexSubImage1DEXT(GLenum target, GLint level, GLint xoffset, GLint x, GLint y, GLsizei width)
{
  glCopyTexSubImage1DEXT(target, level, xoffset, x, y, width);
}

inline void CGraphics::gilCopyTexSubImage2DEXT(GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint x, GLint y, GLsizei width, GLsizei height)
{
  glCopyTexSubImage2DEXT(target, level, xoffset, yoffset, x, y, width, height);
}

inline void CGraphics::gilCopyTexSubImage3DEXT(GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint zoffset, GLint x, GLint y, GLsizei width, GLsizei height)
{
  glCopyTexSubImage3DEXT(target, level, xoffset, yoffset, zoffset, x, y, width, height);
}

inline void CGraphics::gilCullFace(GLenum mode)
{
  glCullFace(mode);
}

inline void CGraphics::gilDeformSGIX(GLbitfield mask)
{
  glDeformSGIX(mask);
}

inline void CGraphics::gilDeformationMap3dSGIX(GLenum target, GLdouble u1, GLdouble u2, GLint ustride, GLint uorder, GLdouble v1, GLdouble v2, GLint vstride, GLint vorder, GLdouble w1, GLdouble w2, GLint wstride, GLint worder, const GLdouble *points)
{
  glDeformationMap3dSGIX(target, u1, u2, ustride, uorder, v1, v2, vstride, vorder, w1, w2, wstride, worder, points);
}

inline void CGraphics::gilDeformationMap3fSGIX(GLenum target, GLfloat u1, GLfloat u2, GLint ustride, GLint uorder, GLfloat v1, GLfloat v2, GLint vstride, GLint vorder, GLfloat w1, GLfloat w2, GLint wstride, GLint worder, const GLfloat *points)
{
  glDeformationMap3fSGIX(target, u1, u2, ustride, uorder, v1, v2, vstride, vorder, w1, w2, wstride, worder, points);
}

inline void CGraphics::gilDeleteLists(GLuint list, GLsizei range)
{
  glDeleteLists(list, range);
}

inline void CGraphics::gilDeleteTexturesEXT(GLsizei n, const GLuint *textures)
{
  glDeleteTexturesEXT(n, textures);
}

inline void CGraphics::gilDepthFunc(GLenum func)
{
  glDepthFunc(func);
}

inline void CGraphics::gilDepthMask(GLboolean flag)
{
  glDepthMask(flag);
}

inline void CGraphics::gilDepthRange(GLclampd near, GLclampd far)
{
#ifndef WIN32
  glDepthRange(near, far);
#endif
}

inline void CGraphics::gilDetailTexFuncSGIS(GLenum target, GLsizei n, const GLfloat *points)
{
  glDetailTexFuncSGIS(target, n, points);
}

inline void CGraphics::gilDisable(GLenum cap)
{
  glDisable(cap);
}

inline void CGraphics::gilDrawArraysEXT(GLenum mode, GLint first, GLsizei count)
{
  glDrawArraysEXT(mode, first, count);
}

inline void CGraphics::gilDrawBuffer(GLenum mode)
{
  glDrawBuffer(mode);
}

inline void CGraphics::gilDrawPixels(GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *pixels)
{
  glDrawPixels(width, height, format, type, pixels);
}

inline void CGraphics::gilEdgeFlag(GLboolean flag)
{
  glEdgeFlag(flag);
}

inline void CGraphics::gilEdgeFlagPointerEXT(GLsizei stride, GLsizei count, const GLboolean *pointer)
{
  glEdgeFlagPointerEXT(stride, count, pointer);
}

inline void CGraphics::gilEdgeFlagv(const GLboolean *flag)
{
  glEdgeFlagv(flag);
}

inline void CGraphics::gilEnable(GLenum cap)
{
  glEnable(cap);
}

inline void CGraphics::gilEndList()
{
  glEndList();
}

inline void CGraphics::gilEvalCoord1d(GLdouble u)
{
  glEvalCoord1d(u);
}

inline void CGraphics::gilEvalCoord1dv(const GLdouble *u)
{
  glEvalCoord1dv(u);
}

inline void CGraphics::gilEvalCoord1f(GLfloat u)
{
  glEvalCoord1f(u);
}

inline void CGraphics::gilEvalCoord1fv(const GLfloat *u)
{
  glEvalCoord1fv(u);
}

inline void CGraphics::gilEvalCoord2d(GLdouble u, GLdouble v)
{
  glEvalCoord2d(u, v);
}

inline void CGraphics::gilEvalCoord2dv(const GLdouble *u)
{
  glEvalCoord2dv(u);
}

inline void CGraphics::gilEvalCoord2f(GLfloat u, GLfloat v)
{
  glEvalCoord2f(u, v);
}

inline void CGraphics::gilEvalCoord2fv(const GLfloat *u)
{
  glEvalCoord2fv(u);
}

inline void CGraphics::gilEvalMesh1(GLenum mode, GLint i1, GLint i2)
{
  glEvalMesh1(mode, i1, i2);
}

inline void CGraphics::gilEvalMesh2(GLenum mode, GLint i1, GLint i2, GLint j1, GLint j2)
{
  glEvalMesh2(mode, i1, i2, j1, j2);
}

inline void CGraphics::gilEvalPoint1(GLint i)
{
  glEvalPoint1(i);
}

inline void CGraphics::gilEvalPoint2(GLint i, GLint j)
{
  glEvalPoint2(i, j);
}

inline void CGraphics::gilFeedbackBuffer(GLsizei size, GLenum type, GLfloat *buffer)
{
  glFeedbackBuffer(size, type, buffer);
}

inline void CGraphics::gilFinish()
{
  glFinish();
}

inline void CGraphics::gilFlush()
{
  glFlush();
}

inline void CGraphics::gilFlushRasterSGIX()
{
  glFlushRasterSGIX();
}

inline void CGraphics::gilFogFuncSGIS(GLsizei n, const GLfloat *points)
{
  glFogFuncSGIS(n, points);
}

inline void CGraphics::gilFogf(GLenum pname, GLfloat param)
{
  glFogf(pname, param);
}

inline void CGraphics::gilFogfv(GLenum pname, const GLfloat *params)
{
  glFogfv(pname, params);
}

inline void CGraphics::gilFogi(GLenum pname, GLint param)
{
  glFogi(pname, param);
}

inline void CGraphics::gilFogiv(GLenum pname, const GLint *params)
{
  glFogiv(pname, params);
}

inline void CGraphics::gilFrameZoomSGIX(GLint factor)
{
  glFrameZoomSGIX(factor);
}

inline void CGraphics::gilFrontFace(GLenum mode)
{
  glFrontFace(mode);
}

inline void CGraphics::gilFrustum(GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble near, GLdouble far)
{
#ifndef WIN32
  glFrustum(left, right, bottom, top, near, far);
#endif
}

inline GLuint CGraphics::gilGenLists(GLsizei range)
{
  return glGenLists(range);
}

inline void CGraphics::gilGenTexturesEXT(GLsizei n, GLuint *textures)
{
  glGenTexturesEXT(n, textures);
}

inline void CGraphics::gilGetBooleanv(GLenum pname, GLboolean *params)
{
  glGetBooleanv(pname, params);
}

inline void CGraphics::gilGetClipPlane(GLenum plane, GLdouble *equation)
{
  glGetClipPlane(plane, equation);
}

inline void CGraphics::gilGetColorTableParameterfvSGI(GLenum target, GLenum pname, GLfloat *params)
{
  glGetColorTableParameterfvSGI(target, pname, params);
}

inline void CGraphics::gilGetColorTableParameterivSGI(GLenum target, GLenum pname, GLint *params)
{
  glGetColorTableParameterivSGI(target, pname, params);
}

inline void CGraphics::gilGetColorTableSGI(GLenum target, GLenum format, GLenum type, GLvoid *table)
{
  glGetColorTableSGI(target, format, type, table);
}

inline void CGraphics::gilGetConvolutionFilterEXT(GLenum target, GLenum format, GLenum type, GLvoid *image)
{
  glGetConvolutionFilterEXT(target, format, type, image);
}

inline void CGraphics::gilGetConvolutionParameterfvEXT(GLenum target, GLenum pname, GLfloat *params)
{
  glGetConvolutionParameterfvEXT(target, pname, params);
}

inline void CGraphics::gilGetConvolutionParameterivEXT(GLenum target, GLenum pname, GLint *params)
{
  glGetConvolutionParameterivEXT(target, pname, params);
}

inline void CGraphics::gilGetDetailTexFuncSGIS(GLenum target, GLfloat *points)
{
  glGetDetailTexFuncSGIS(target, points);
}

inline void CGraphics::gilGetDoublev(GLenum pname, GLdouble *params)
{
  glGetDoublev(pname, params);
}

inline GLenum CGraphics::gilGetError()
{
  return glGetError();
}

inline void CGraphics::gilGetFloatv(GLenum pname, GLfloat *params)
{
  glGetFloatv(pname, params);
}

inline void CGraphics::gilGetHistogramEXT(GLenum target, GLboolean reset, GLenum format, GLenum type, GLvoid *values)
{
  glGetHistogramEXT(target, reset, format, type, values);
}

inline void CGraphics::gilGetHistogramParameterfvEXT(GLenum target, GLenum pname, GLfloat *params)
{
  glGetHistogramParameterfvEXT(target, pname, params);
}

inline void CGraphics::gilGetHistogramParameterivEXT(GLenum target, GLenum pname, GLint *params)
{
  glGetHistogramParameterivEXT(target, pname, params);
}

inline GLint CGraphics::gilGetInstrumentsSGIX()
{
  return glGetInstrumentsSGIX();
}

inline void CGraphics::gilGetIntegerv(GLenum pname, GLint *params)
{
  glGetIntegerv(pname, params);
}

inline void CGraphics::gilGetLightfv(GLenum light, GLenum pname, GLfloat *params)
{
  glGetLightfv(light, pname, params);
}

inline void CGraphics::gilGetLightiv(GLenum light, GLenum pname, GLint *params)
{
  glGetLightiv(light, pname, params);
}

inline void CGraphics::gilGetMapdv(GLenum target, GLenum query, GLdouble *v)
{
  glGetMapdv(target, query, v);
}

inline void CGraphics::gilGetMapfv(GLenum target, GLenum query, GLfloat *v)
{
  glGetMapfv(target, query, v);
}

inline void CGraphics::gilGetMapiv(GLenum target, GLenum query, GLint *v)
{
  glGetMapiv(target, query, v);
}

inline void CGraphics::gilGetMaterialfv(GLenum face, GLenum pname, GLfloat *params)
{
  glGetMaterialfv(face, pname, params);
}

inline void CGraphics::gilGetMaterialiv(GLenum face, GLenum pname, GLint *params)
{
  glGetMaterialiv(face, pname, params);
}

inline void CGraphics::gilGetMinmaxEXT(GLenum target, GLboolean reset, GLenum format, GLenum type, GLvoid *values)
{
  glGetMinmaxEXT(target, reset, format, type, values);
}

inline void CGraphics::gilGetMinmaxParameterfvEXT(GLenum target, GLenum pname, GLfloat *params)
{
  glGetMinmaxParameterfvEXT(target, pname, params);
}

inline void CGraphics::gilGetMinmaxParameterivEXT(GLenum target, GLenum pname, GLint *params)
{
  glGetMinmaxParameterivEXT(target, pname, params);
}

inline void CGraphics::gilGetPixelMapfv(GLenum map, GLfloat *values)
{
  glGetPixelMapfv(map, values);
}

inline void CGraphics::gilGetPixelMapuiv(GLenum map, GLuint *values)
{
  glGetPixelMapuiv(map, values);
}

inline void CGraphics::gilGetPixelMapusv(GLenum map, GLushort *values)
{
  glGetPixelMapusv(map, values);
}

inline void CGraphics::gilGetPointervEXT(GLenum pname, GLvoid* *params)
{
  glGetPointervEXT(pname, params);
}

inline void CGraphics::gilGetPolygonStipple(GLubyte *mask)
{
  glGetPolygonStipple(mask);
}

inline void CGraphics::gilGetSeparableFilterEXT(GLenum target, GLenum format, GLenum type, GLvoid *row, GLvoid *column, GLvoid *span)
{
  glGetSeparableFilterEXT(target, format, type, row, column, span);
}

inline void CGraphics::gilGetSharpenTexFuncSGIS(GLenum target, GLfloat *points)
{
  glGetSharpenTexFuncSGIS(target, points);
}

inline const GLubyte *CGraphics::gilGetString(GLenum name)
{
  return glGetString(name);
}

inline void CGraphics::gilGetTexEnvfv(GLenum target, GLenum pname, GLfloat *params)
{
  glGetTexEnvfv(target, pname, params);
}

inline void CGraphics::gilGetTexEnviv(GLenum target, GLenum pname, GLint *params)
{
  glGetTexEnviv(target, pname, params);
}

inline void CGraphics::gilGetTexFilterFuncSGIS(GLenum target, GLenum filter, GLfloat *weights)
{
  glGetTexFilterFuncSGIS(target, filter, weights);
}

inline void CGraphics::gilGetTexGendv(GLenum coord, GLenum pname, GLdouble *params)
{
  glGetTexGendv(coord, pname, params);
}

inline void CGraphics::gilGetTexGenfv(GLenum coord, GLenum pname, GLfloat *params)
{
  glGetTexGenfv(coord, pname, params);
}

inline void CGraphics::gilGetTexGeniv(GLenum coord, GLenum pname, GLint *params)
{
  glGetTexGeniv(coord, pname, params);
}

inline void CGraphics::gilGetTexImage(GLenum target, GLint level, GLenum format, GLenum type, GLvoid *pixels)
{
  glGetTexImage(target, level, format, type, pixels);
}

inline void CGraphics::gilGetTexLevelParameterfv(GLenum target, GLint level, GLenum pname, GLfloat *params)
{
  glGetTexLevelParameterfv(target, level, pname, params);
}

inline void CGraphics::gilGetTexLevelParameteriv(GLenum target, GLint level, GLenum pname, GLint *params)
{
  glGetTexLevelParameteriv(target, level, pname, params);
}

inline void CGraphics::gilGetTexParameterfv(GLenum target, GLenum pname, GLfloat *params)
{
  glGetTexParameterfv(target, pname, params);
}

inline void CGraphics::gilGetTexParameteriv(GLenum target, GLenum pname, GLint *params)
{
  glGetTexParameteriv(target, pname, params);
}

inline void CGraphics::gilHint(GLenum target, GLenum mode)
{
  glHint(target, mode);
}

inline void CGraphics::gilHistogramEXT(GLenum target, GLsizei width, GLenum internalformat, GLboolean sink)
{
  glHistogramEXT(target, width, internalformat, sink);
}

inline void CGraphics::gilIndexMask(GLuint mask)
{
  glIndexMask(mask);
}

inline void CGraphics::gilIndexPointerEXT(GLenum type, GLsizei stride, GLsizei count, const GLvoid *pointer)
{
  glIndexPointerEXT(type, stride, count, pointer);
}

inline void CGraphics::gilIndexd(GLdouble c)
{
  glIndexd(c);
}

inline void CGraphics::gilIndexdv(const GLdouble *c)
{
  glIndexdv(c);
}

inline void CGraphics::gilIndexf(GLfloat c)
{
  glIndexf(c);
}

inline void CGraphics::gilIndexfv(const GLfloat *c)
{
  glIndexfv(c);
}

inline void CGraphics::gilIndexi(GLint c)
{
  glIndexi(c);
}

inline void CGraphics::gilIndexiv(const GLint *c)
{
  glIndexiv(c);
}

inline void CGraphics::gilIndexs(GLshort c)
{
  glIndexs(c);
}

inline void CGraphics::gilIndexsv(const GLshort *c)
{
  glIndexsv(c);
}

inline void CGraphics::gilInitNames()
{
  glInitNames();
}

inline GLboolean CGraphics::gilIsEnabled(GLenum cap)
{
  return glIsEnabled(cap);
}

inline GLboolean CGraphics::gilIsList(GLuint list)
{
  return glIsList(list);
}

inline GLboolean CGraphics::gilIsTextureEXT(GLuint texture)
{
  return glIsTextureEXT(texture);
}

inline void CGraphics::gilLightModelf(GLenum pname, GLfloat param)
{
  glLightModelf(pname, param);
}

inline void CGraphics::gilLightModelfv(GLenum pname, const GLfloat *params)
{
  glLightModelfv(pname, params);
}

inline void CGraphics::gilLightModeli(GLenum pname, GLint param)
{
  glLightModeli(pname, param);
}

inline void CGraphics::gilLightModeliv(GLenum pname, const GLint *params)
{
  glLightModeliv(pname, params);
}

inline void CGraphics::gilLightf(GLenum light, GLenum pname, GLfloat param)
{
  glLightf(light, pname, param);
}

inline void CGraphics::gilLightfv(GLenum light, GLenum pname, const GLfloat *params)
{
  glLightfv(light, pname, params);
}

inline void CGraphics::gilLighti(GLenum light, GLenum pname, GLint param)
{
  glLighti(light, pname, param);
}

inline void CGraphics::gilLightiv(GLenum light, GLenum pname, const GLint *params)
{
  glLightiv(light, pname, params);
}

inline void CGraphics::gilLineStipple(GLint factor, GLushort pattern)
{
  glLineStipple(factor, pattern);
}

inline void CGraphics::gilLineWidth(GLfloat width)
{
  glLineWidth(width);
}

inline void CGraphics::gilListBase(GLuint base)
{
  glListBase(base);
}

inline void CGraphics::gilLoadIdentity()
{
  glLoadIdentity();
}

inline void CGraphics::gilLoadIdentityDeformationMapSGIX(GLbitfield mask)
{
  glLoadIdentityDeformationMapSGIX(mask);
}

inline void CGraphics::gilLoadMatrixd(const GLdouble *m)
{
  glLoadMatrixd(m);
}

inline void CGraphics::gilLoadMatrixf(const GLfloat *m)
{
  glLoadMatrixf(m);
}

inline void CGraphics::gilLoadName(GLuint name)
{
  glLoadName(name);
}

inline void CGraphics::gilLogicOp(GLenum opcode)
{
  glLogicOp(opcode);
}

inline void CGraphics::gilMap1d(GLenum target, GLdouble u1, GLdouble u2, GLint stride, GLint order, const GLdouble *points)
{
  glMap1d(target, u1, u2, stride, order, points);
}

inline void CGraphics::gilMap1f(GLenum target, GLfloat u1, GLfloat u2, GLint stride, GLint order, const GLfloat *points)
{
  glMap1f(target, u1, u2, stride, order, points);
}

inline void CGraphics::gilMap2d(GLenum target, GLdouble u1, GLdouble u2, GLint ustride, GLint uorder, GLdouble v1, GLdouble v2, GLint vstride, GLint vorder, const GLdouble *points)
{
  glMap2d(target, u1, u2, ustride, uorder, v1, v2, vstride, vorder, points);
}

inline void CGraphics::gilMap2f(GLenum target, GLfloat u1, GLfloat u2, GLint ustride, GLint uorder, GLfloat v1, GLfloat v2, GLint vstride, GLint vorder, const GLfloat *points)
{
  glMap2f(target, u1, u2, ustride, uorder, v1, v2, vstride, vorder, points);
}

inline void CGraphics::gilMapGrid1d(GLint un, GLdouble u1, GLdouble u2)
{
  glMapGrid1d(un, u1, u2);
}

inline void CGraphics::gilMapGrid1f(GLint un, GLfloat u1, GLfloat u2)
{
  glMapGrid1f(un, u1, u2);
}

inline void CGraphics::gilMapGrid2d(GLint un, GLdouble u1, GLdouble u2, GLint vn, GLdouble v1, GLdouble v2)
{
  glMapGrid2d(un, u1, u2, vn, v1, v2);
}

inline void CGraphics::gilMapGrid2f(GLint un, GLfloat u1, GLfloat u2, GLint vn, GLfloat v1, GLfloat v2)
{
  glMapGrid2f(un, u1, u2, vn, v1, v2);
}

inline void CGraphics::gilMaterialf(GLenum face, GLenum pname, GLfloat param)
{
  glMaterialf(face, pname, param);
}

inline void CGraphics::gilMaterialfv(GLenum face, GLenum pname, const GLfloat *params)
{
  glMaterialfv(face, pname, params);
}

inline void CGraphics::gilMateriali(GLenum face, GLenum pname, GLint param)
{
  glMateriali(face, pname, param);
}

inline void CGraphics::gilMaterialiv(GLenum face, GLenum pname, const GLint *params)
{
  glMaterialiv(face, pname, params);
}

inline void CGraphics::gilMatrixMode(GLenum mode)
{
  glMatrixMode(mode);
}

inline void CGraphics::gilMinmaxEXT(GLenum target, GLenum internalformat, GLboolean sink)
{
  glMinmaxEXT(target, internalformat, sink);
}

inline void CGraphics::gilMultMatrixd(const GLdouble *m)
{
  glMultMatrixd(m);
}

inline void CGraphics::gilMultMatrixf(const GLfloat *m)
{
  glMultMatrixf(m);
}

inline void CGraphics::gilNewList(GLuint list, GLenum mode)
{
  glNewList(list, mode);
}

inline void CGraphics::gilNormal3b(GLbyte nx, GLbyte ny, GLbyte nz)
{
  glNormal3b(nx, ny, nz);
}

inline void CGraphics::gilNormal3bv(const GLbyte *v)
{
  glNormal3bv(v);
}

inline void CGraphics::gilNormal3d(GLdouble nx, GLdouble ny, GLdouble nz)
{
  glNormal3d(nx, ny, nz);
}

inline void CGraphics::gilNormal3dv(const GLdouble *v)
{
  glNormal3dv(v);
}

inline void CGraphics::gilNormal3f(GLfloat nx, GLfloat ny, GLfloat nz)
{
  glNormal3f(nx, ny, nz);
}

inline void CGraphics::gilNormal3fv(const GLfloat *v)
{
  glNormal3fv(v);
}

inline void CGraphics::gilNormal3i(GLint nx, GLint ny, GLint nz)
{
  glNormal3i(nx, ny, nz);
}

inline void CGraphics::gilNormal3iv(const GLint *v)
{
  glNormal3iv(v);
}

inline void CGraphics::gilNormal3s(GLshort nx, GLshort ny, GLshort nz)
{
  glNormal3s(nx, ny, nz);
}

inline void CGraphics::gilNormal3sv(const GLshort *v)
{
  glNormal3sv(v);
}

inline void CGraphics::gilNormalPointerEXT(GLenum type, GLsizei stride, GLsizei count, const GLvoid *pointer)
{
  glNormalPointerEXT(type, stride, count, pointer);
}

inline void CGraphics::gilOrtho(GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble near, GLdouble far)
{
#ifndef WIN32
  glOrtho(left, right, bottom, top, near, far);
#endif
}

inline void CGraphics::gilPassThrough(GLfloat token)
{
  glPassThrough(token);
}

inline void CGraphics::gilPipelineInstrumentsBufferSGIX(GLsizei size, GLfloat *buffer)
{
  glPipelineInstrumentsBufferSGIX(size, buffer);
}

inline void CGraphics::gilPixelMapfv(GLenum map, GLint mapsize, const GLfloat *values)
{
  glPixelMapfv(map, mapsize, values);
}

inline void CGraphics::gilPixelMapuiv(GLenum map, GLint mapsize, const GLuint *values)
{
  glPixelMapuiv(map, mapsize, values);
}

inline void CGraphics::gilPixelMapusv(GLenum map, GLint mapsize, const GLushort *values)
{
  glPixelMapusv(map, mapsize, values);
}

inline void CGraphics::gilPixelStoref(GLenum pname, GLfloat param)
{
  glPixelStoref(pname, param);
}

inline void CGraphics::gilPixelStorei(GLenum pname, GLint param)
{
  glPixelStorei(pname, param);
}

inline void CGraphics::gilPixelTexGenSGIX(GLenum mode)
{
  glPixelTexGenSGIX(mode);
}

inline void CGraphics::gilPixelTransferf(GLenum pname, GLfloat param)
{
  glPixelTransferf(pname, param);
}

inline void CGraphics::gilPixelTransferi(GLenum pname, GLint param)
{
  glPixelTransferi(pname, param);
}

inline void CGraphics::gilPixelZoom(GLfloat xfactor, GLfloat yfactor)
{
  glPixelZoom(xfactor, yfactor);
}

inline void CGraphics::gilPointParameterfSGIS(GLenum pname, GLfloat param)
{
  glPointParameterfSGIS(pname, param);
}

inline void CGraphics::gilPointParameterfvSGIS(GLenum pname, const GLfloat *params)
{
  glPointParameterfvSGIS(pname, params);
}

inline void CGraphics::gilPointSize(GLfloat size)
{
  glPointSize(size);
}

inline GLint CGraphics::gilPollInstrumentsSGIX(GLint *marker_p)
{
  return glPollInstrumentsSGIX(marker_p);
}

inline void CGraphics::gilPolygonMode(GLenum face, GLenum mode)
{
  glPolygonMode(face, mode);
}

inline void CGraphics::gilPolygonOffsetEXT(GLfloat factor, GLfloat bias)
{
  glPolygonOffsetEXT(factor, bias);
}

inline void CGraphics::gilPolygonStipple(const GLubyte *mask)
{
  glPolygonStipple(mask);
}

inline void CGraphics::gilPopAttrib()
{
  glPopAttrib();
}

inline void CGraphics::gilPopMatrix()
{
  glPopMatrix();
}

inline void CGraphics::gilPopName()
{
  glPopName();
}

inline void CGraphics::gilPrioritizeTexturesEXT(GLsizei n, const GLuint *textures, const GLclampf *priorities)
{
  glPrioritizeTexturesEXT(n, textures, priorities);
}

inline void CGraphics::gilPushAttrib(GLbitfield mask)
{
  glPushAttrib(mask);
}

inline void CGraphics::gilPushMatrix()
{
  glPushMatrix();
}

inline void CGraphics::gilPushName(GLuint name)
{
  glPushName(name);
}

inline void CGraphics::gilRasterPos2d(GLdouble x, GLdouble y)
{
  glRasterPos2d(x, y);
}

inline void CGraphics::gilRasterPos2dv(const GLdouble *v)
{
  glRasterPos2dv(v);
}

inline void CGraphics::gilRasterPos2f(GLfloat x, GLfloat y)
{
  glRasterPos2f(x, y);
}

inline void CGraphics::gilRasterPos2fv(const GLfloat *v)
{
  glRasterPos2fv(v);
}

inline void CGraphics::gilRasterPos2i(GLint x, GLint y)
{
  glRasterPos2i(x, y);
}

inline void CGraphics::gilRasterPos2iv(const GLint *v)
{
  glRasterPos2iv(v);
}

inline void CGraphics::gilRasterPos2s(GLshort x, GLshort y)
{
  glRasterPos2s(x, y);
}

inline void CGraphics::gilRasterPos2sv(const GLshort *v)
{
  glRasterPos2sv(v);
}

inline void CGraphics::gilRasterPos3d(GLdouble x, GLdouble y, GLdouble z)
{
  glRasterPos3d(x, y, z);
}

inline void CGraphics::gilRasterPos3dv(const GLdouble *v)
{
  glRasterPos3dv(v);
}

inline void CGraphics::gilRasterPos3f(GLfloat x, GLfloat y, GLfloat z)
{
  glRasterPos3f(x, y, z);
}

inline void CGraphics::gilRasterPos3fv(const GLfloat *v)
{
  glRasterPos3fv(v);
}

inline void CGraphics::gilRasterPos3i(GLint x, GLint y, GLint z)
{
  glRasterPos3i(x, y, z);
}

inline void CGraphics::gilRasterPos3iv(const GLint *v)
{
  glRasterPos3iv(v);
}

inline void CGraphics::gilRasterPos3s(GLshort x, GLshort y, GLshort z)
{
  glRasterPos3s(x, y, z);
}

inline void CGraphics::gilRasterPos3sv(const GLshort *v)
{
  glRasterPos3sv(v);
}

inline void CGraphics::gilRasterPos4d(GLdouble x, GLdouble y, GLdouble z, GLdouble w)
{
  glRasterPos4d(x, y, z, w);
}

inline void CGraphics::gilRasterPos4dv(const GLdouble *v)
{
  glRasterPos4dv(v);
}

inline void CGraphics::gilRasterPos4f(GLfloat x, GLfloat y, GLfloat z, GLfloat w)
{
  glRasterPos4f(x, y, z, w);
}

inline void CGraphics::gilRasterPos4fv(const GLfloat *v)
{
  glRasterPos4fv(v);
}

inline void CGraphics::gilRasterPos4i(GLint x, GLint y, GLint z, GLint w)
{
  glRasterPos4i(x, y, z, w);
}

inline void CGraphics::gilRasterPos4iv(const GLint *v)
{
  glRasterPos4iv(v);
}

inline void CGraphics::gilRasterPos4s(GLshort x, GLshort y, GLshort z, GLshort w)
{
  glRasterPos4s(x, y, z, w);
}

inline void CGraphics::gilRasterPos4sv(const GLshort *v)
{
  glRasterPos4sv(v);
}

inline void CGraphics::gilReadBuffer(GLenum mode)
{
  glReadBuffer(mode);
}

inline void CGraphics::gilReadPixels(GLint x, GLint y, GLsizei width, GLsizei height, GLenum format, GLenum type, GLvoid *pixels)
{
  glReadPixels(x, y, width, height, format, type, pixels);
}

inline void CGraphics::gilRectd(GLdouble x1, GLdouble y1, GLdouble x2, GLdouble y2)
{
  glRectd(x1, y1, x2, y2);
}

inline void CGraphics::gilRectdv(const GLdouble *v1, const GLdouble *v2)
{
  glRectdv(v1, v2);
}

inline void CGraphics::gilRectf(GLfloat x1, GLfloat y1, GLfloat x2, GLfloat y2)
{
  glRectf(x1, y1, x2, y2);
}

inline void CGraphics::gilRectfv(const GLfloat *v1, const GLfloat *v2)
{
  glRectfv(v1, v2);
}

inline void CGraphics::gilRecti(GLint x1, GLint y1, GLint x2, GLint y2)
{
  glRecti(x1, y1, x2, y2);
}

inline void CGraphics::gilRectiv(const GLint *v1, const GLint *v2)
{
  glRectiv(v1, v2);
}

inline void CGraphics::gilRects(GLshort x1, GLshort y1, GLshort x2, GLshort y2)
{
  glRects(x1, y1, x2, y2);
}

inline void CGraphics::gilRectsv(const GLshort *v1, const GLshort *v2)
{
  glRectsv(v1, v2);
}

inline void CGraphics::gilReferencePlaneSGIX(const GLdouble *equation)
{
  glReferencePlaneSGIX(equation);
}

inline void CGraphics::gilResetHistogramEXT(GLenum target)
{
  glResetHistogramEXT(target);
}

inline void CGraphics::gilResetMinmaxEXT(GLenum target)
{
  glResetMinmaxEXT(target);
}

inline void CGraphics::gilRotated(GLdouble angle, GLdouble x, GLdouble y, GLdouble z)
{
  glRotated(angle, x, y, z);
}

inline void CGraphics::gilRotatef(GLfloat angle, GLfloat x, GLfloat y, GLfloat z)
{
  glRotatef(angle, x, y, z);
}

inline void CGraphics::gilSampleMaskSGIS(GLclampf value, GLboolean invert)
{
  glSampleMaskSGIS(value, invert);
}

inline void CGraphics::gilSamplePatternSGIS(GLenum pattern)
{
  glSamplePatternSGIS(pattern);
}

inline void CGraphics::gilScaled(GLdouble x, GLdouble y, GLdouble z)
{
  glScaled(x, y, z);
}

inline void CGraphics::gilScalef(GLfloat x, GLfloat y, GLfloat z)
{
  glScalef(x, y, z);
}

inline void CGraphics::gilScissor(GLint x, GLint y, GLsizei width, GLsizei height)
{
  glScissor(x, y, width, height);
}

inline void CGraphics::gilSelectBuffer(GLsizei size, GLuint *buffer)
{
  glSelectBuffer(size, buffer);
}

inline void CGraphics::gilSeparableFilter2DEXT(GLenum target, GLenum internalformat, GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *row, const GLvoid *column)
{
  glSeparableFilter2DEXT(target, internalformat, width, height, format, type, row, column);
}

inline void CGraphics::gilShadeModel(GLenum mode)
{
  glShadeModel(mode);
}

inline void CGraphics::gilSharpenTexFuncSGIS(GLenum target, GLsizei n, const GLfloat *points)
{
  glSharpenTexFuncSGIS(target, n, points);
}

inline void CGraphics::gilSpriteParameterfSGIX(GLenum pname, GLfloat param)
{
  glSpriteParameterfSGIX(pname, param);
}

inline void CGraphics::gilSpriteParameterfvSGIX(GLenum pname, const GLfloat *params)
{
  glSpriteParameterfvSGIX(pname, params);
}

inline void CGraphics::gilSpriteParameteriSGIX(GLenum pname, GLint param)
{
  glSpriteParameteriSGIX(pname, param);
}

inline void CGraphics::gilSpriteParameterivSGIX(GLenum pname, const GLint *params)
{
  glSpriteParameterivSGIX(pname, params);
}

inline void CGraphics::gilStartInstrumentsSGIX()
{
  glStartInstrumentsSGIX();
}

inline void CGraphics::gilStencilFunc(GLenum func, GLint ref, GLuint mask)
{
  glStencilFunc(func, ref, mask);
}

inline void CGraphics::gilStencilMask(GLuint mask)
{
  glStencilMask(mask);
}

inline void CGraphics::gilStencilOp(GLenum fail, GLenum zfail, GLenum zpass)
{
  glStencilOp(fail, zfail, zpass);
}

inline void CGraphics::gilStopInstrumentsSGIX(GLint marker)
{
  glStopInstrumentsSGIX(marker);
}

inline void CGraphics::gilTagSampleBufferSGIX()
{
  glTagSampleBufferSGIX();
}

inline void CGraphics::gilTexCoord1d(GLdouble s)
{
  glTexCoord1d(s);
}

inline void CGraphics::gilTexCoord1dv(const GLdouble *v)
{
  glTexCoord1dv(v);
}

inline void CGraphics::gilTexCoord1f(GLfloat s)
{
  glTexCoord1f(s);
}

inline void CGraphics::gilTexCoord1fv(const GLfloat *v)
{
  glTexCoord1fv(v);
}

inline void CGraphics::gilTexCoord1i(GLint s)
{
  glTexCoord1i(s);
}

inline void CGraphics::gilTexCoord1iv(const GLint *v)
{
  glTexCoord1iv(v);
}

inline void CGraphics::gilTexCoord1s(GLshort s)
{
  glTexCoord1s(s);
}

inline void CGraphics::gilTexCoord1sv(const GLshort *v)
{
  glTexCoord1sv(v);
}

inline void CGraphics::gilTexCoord2d(GLdouble s, GLdouble t)
{
  glTexCoord2d(s, t);
}

inline void CGraphics::gilTexCoord2dv(const GLdouble *v)
{
  glTexCoord2dv(v);
}

inline void CGraphics::gilTexCoord2f(GLfloat s, GLfloat t)
{
  glTexCoord2f(s, t);
}

inline void CGraphics::gilTexCoord2fv(const GLfloat *v)
{
  glTexCoord2fv(v);
}

inline void CGraphics::gilTexCoord2i(GLint s, GLint t)
{
  glTexCoord2i(s, t);
}

inline void CGraphics::gilTexCoord2iv(const GLint *v)
{
  glTexCoord2iv(v);
}

inline void CGraphics::gilTexCoord2s(GLshort s, GLshort t)
{
  glTexCoord2s(s, t);
}

inline void CGraphics::gilTexCoord2sv(const GLshort *v)
{
  glTexCoord2sv(v);
}

inline void CGraphics::gilTexCoord3d(GLdouble s, GLdouble t, GLdouble r)
{
  glTexCoord3d(s, t, r);
}

inline void CGraphics::gilTexCoord3dv(const GLdouble *v)
{
  glTexCoord3dv(v);
}

inline void CGraphics::gilTexCoord3f(GLfloat s, GLfloat t, GLfloat r)
{
  glTexCoord3f(s, t, r);
}

inline void CGraphics::gilTexCoord3fv(const GLfloat *v)
{
  glTexCoord3fv(v);
}

inline void CGraphics::gilTexCoord3i(GLint s, GLint t, GLint r)
{
  glTexCoord3i(s, t, r);
}

inline void CGraphics::gilTexCoord3iv(const GLint *v)
{
  glTexCoord3iv(v);
}

inline void CGraphics::gilTexCoord3s(GLshort s, GLshort t, GLshort r)
{
  glTexCoord3s(s, t, r);
}

inline void CGraphics::gilTexCoord3sv(const GLshort *v)
{
  glTexCoord3sv(v);
}

inline void CGraphics::gilTexCoord4d(GLdouble s, GLdouble t, GLdouble r, GLdouble q)
{
  glTexCoord4d(s, t, r, q);
}

inline void CGraphics::gilTexCoord4dv(const GLdouble *v)
{
  glTexCoord4dv(v);
}

inline void CGraphics::gilTexCoord4f(GLfloat s, GLfloat t, GLfloat r, GLfloat q)
{
  glTexCoord4f(s, t, r, q);
}

inline void CGraphics::gilTexCoord4fv(const GLfloat *v)
{
  glTexCoord4fv(v);
}

inline void CGraphics::gilTexCoord4i(GLint s, GLint t, GLint r, GLint q)
{
  glTexCoord4i(s, t, r, q);
}

inline void CGraphics::gilTexCoord4iv(const GLint *v)
{
  glTexCoord4iv(v);
}

inline void CGraphics::gilTexCoord4s(GLshort s, GLshort t, GLshort r, GLshort q)
{
  glTexCoord4s(s, t, r, q);
}

inline void CGraphics::gilTexCoord4sv(const GLshort *v)
{
  glTexCoord4sv(v);
}

inline void CGraphics::gilTexCoordPointerEXT(GLint size, GLenum type, GLsizei stride, GLsizei count, const GLvoid *pointer)
{
  glTexCoordPointerEXT(size, type, stride, count, pointer);
}

inline void CGraphics::gilTexEnvf(GLenum target, GLenum pname, GLfloat param)
{
  glTexEnvf(target, pname, param);
}

inline void CGraphics::gilTexEnvfv(GLenum target, GLenum pname, const GLfloat *params)
{
  glTexEnvfv(target, pname, params);
}

inline void CGraphics::gilTexEnvi(GLenum target, GLenum pname, GLint param)
{
  glTexEnvi(target, pname, param);
}

inline void CGraphics::gilTexEnviv(GLenum target, GLenum pname, const GLint *params)
{
  glTexEnviv(target, pname, params);
}

inline void CGraphics::gilTexFilterFuncSGIS(GLenum target, GLenum filter, GLsizei n, const GLfloat *weights)
{
  glTexFilterFuncSGIS(target, filter, n, weights);
}

inline void CGraphics::gilTexGend(GLenum coord, GLenum pname, GLdouble param)
{
  glTexGend(coord, pname, param);
}

inline void CGraphics::gilTexGendv(GLenum coord, GLenum pname, const GLdouble *params)
{
  glTexGendv(coord, pname, params);
}

inline void CGraphics::gilTexGenf(GLenum coord, GLenum pname, GLfloat param)
{
  glTexGenf(coord, pname, param);
}

inline void CGraphics::gilTexGenfv(GLenum coord, GLenum pname, const GLfloat *params)
{
  glTexGenfv(coord, pname, params);
}

inline void CGraphics::gilTexGeni(GLenum coord, GLenum pname, GLint param)
{
  glTexGeni(coord, pname, param);
}

inline void CGraphics::gilTexGeniv(GLenum coord, GLenum pname, const GLint *params)
{
  glTexGeniv(coord, pname, params);
}

inline void CGraphics::gilTexImage1D(GLenum target, GLint level, GLint components, GLsizei width, GLint border, GLenum format, GLenum type, const GLvoid *pixels)
{
  glTexImage1D(target, level, components, width, border, format, type, pixels);
}

inline void CGraphics::gilTexImage2D(GLenum target, GLint level, GLint components, GLsizei width, GLsizei height, GLint border, GLenum format, GLenum type, const GLvoid *pixels)
{
  glTexImage2D(target, level, components, width, height, border, format, type, pixels);
}

inline void CGraphics::gilTexImage3DEXT(GLenum target, GLint level, GLenum internalformat, GLsizei width, GLsizei height, GLsizei depth, GLint border, GLenum format, GLenum type, const GLvoid *pixels)
{
  glTexImage3DEXT(target, level, internalformat, width, height, depth, border, format, type, pixels);
}

inline void CGraphics::gilTexImage4DSGIS(GLenum target, GLint level, GLenum internalformat, GLsizei width, GLsizei height, GLsizei depth, GLsizei size4d, GLint border, GLenum format, GLenum type, const GLvoid *pixels)
{
  glTexImage4DSGIS(target, level, internalformat, width, height, depth, size4d, border, format, type, pixels);
}

inline void CGraphics::gilTexParameterf(GLenum target, GLenum pname, GLfloat param)
{
  glTexParameterf(target, pname, param);
}

inline void CGraphics::gilTexParameterfv(GLenum target, GLenum pname, const GLfloat *params)
{
  glTexParameterfv(target, pname, params);
}

inline void CGraphics::gilTexParameteri(GLenum target, GLenum pname, GLint param)
{
  glTexParameteri(target, pname, param);
}

inline void CGraphics::gilTexParameteriv(GLenum target, GLenum pname, const GLint *params)
{
  glTexParameteriv(target, pname, params);
}

inline void CGraphics::gilTexSubImage1DEXT(GLenum target, GLint level, GLint xoffset, GLsizei width, GLenum format, GLenum type, const GLvoid *pixels)
{
  glTexSubImage1DEXT(target, level, xoffset, width, format, type, pixels);
}

inline void CGraphics::gilTexSubImage2DEXT(GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *pixels)
{
  glTexSubImage2DEXT(target, level, xoffset, yoffset, width, height, format, type, pixels);
}

inline void CGraphics::gilTexSubImage3DEXT(GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint zoffset, GLsizei width, GLsizei height, GLsizei depth, GLenum format, GLenum type, const GLvoid *pixels)
{
  glTexSubImage3DEXT(target, level, xoffset, yoffset, zoffset, width, height, depth, format, type, pixels);
}

inline void CGraphics::gilTexSubImage4DSGIS(GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint zoffset, GLint woffset, GLsizei width, GLsizei height, GLsizei depth, GLsizei size4d, GLenum format, GLenum type, const GLvoid *pixels)
{
  glTexSubImage4DSGIS(target, level, xoffset, yoffset, zoffset, woffset, width, height, depth, size4d, format, type, pixels);
}

inline void CGraphics::gilTranslated(GLdouble x, GLdouble y, GLdouble z)
{
  glTranslated(x, y, z);
}

inline void CGraphics::gilTranslatef(GLfloat x, GLfloat y, GLfloat z)
{
  glTranslatef(x, y, z);
}

inline void CGraphics::gilVertex2d(GLdouble x, GLdouble y)
{
  glVertex2d(x, y);
}

inline void CGraphics::gilVertex2dv(const GLdouble *v)
{
  glVertex2dv(v);
}

inline void CGraphics::gilVertex2f(GLfloat x, GLfloat y)
{
  glVertex2f(x, y);
}

inline void CGraphics::gilVertex2fv(const GLfloat *v)
{
  glVertex2fv(v);
}

inline void CGraphics::gilVertex2i(GLint x, GLint y)
{
  glVertex2i(x, y);
}

inline void CGraphics::gilVertex2iv(const GLint *v)
{
  glVertex2iv(v);
}

inline void CGraphics::gilVertex2s(GLshort x, GLshort y)
{
  glVertex2s(x, y);
}

inline void CGraphics::gilVertex2sv(const GLshort *v)
{
  glVertex2sv(v);
}

inline void CGraphics::gilVertex3d(GLdouble x, GLdouble y, GLdouble z)
{
  glVertex3d(x, y, z);
}

inline void CGraphics::gilVertex3dv(const GLdouble *v)
{
  glVertex3dv(v);
}

inline void CGraphics::gilVertex3i(GLint x, GLint y, GLint z)
{
  glVertex3i(x, y, z);
}

inline void CGraphics::gilVertex3iv(const GLint *v)
{
  glVertex3iv(v);
}

inline void CGraphics::gilVertex3s(GLshort x, GLshort y, GLshort z)
{
  glVertex3s(x, y, z);
}

inline void CGraphics::gilVertex3sv(const GLshort *v)
{
  glVertex3sv(v);
}

inline void CGraphics::gilVertex4d(GLdouble x, GLdouble y, GLdouble z, GLdouble w)
{
  glVertex4d(x, y, z, w);
}

inline void CGraphics::gilVertex4dv(const GLdouble *v)
{
  glVertex4dv(v);
}

inline void CGraphics::gilVertex4i(GLint x, GLint y, GLint z, GLint w)
{
  glVertex4i(x, y, z, w);
}

inline void CGraphics::gilVertex4iv(const GLint *v)
{
  glVertex4iv(v);
}

inline void CGraphics::gilVertex4s(GLshort x, GLshort y, GLshort z, GLshort w)
{
  glVertex4s(x, y, z, w);
}

inline void CGraphics::gilVertex4sv(const GLshort *v)
{
  glVertex4sv(v);
}

inline void CGraphics::gilVertexPointerEXT(GLint size, GLenum type, GLsizei stride, GLsizei count, const GLvoid *pointer)
{
  glVertexPointerEXT(size, type, stride, count, pointer);
}

inline void CGraphics::gilViewport(GLint x, GLint y, GLsizei width, GLsizei height)
{
  glViewport(x, y, width, height);
}

//////////////////////////////////////////////////
// OpenGL Interface -- "GL/glu.h"
//
inline void CGraphics::giluDisk(GLUquadric* quad, GLdouble inner, GLdouble outer, GLint slices, GLint loops)
{
  gluDisk(quad, inner, outer, slices, loops);
}

inline void CGraphics::giluEndCurve(GLUnurbs* nurb)
{
  gluEndCurve(nurb);
}

inline void CGraphics::giluEndPolygon(GLUtesselator* tess)
{
  gluEndPolygon(tess);
}

inline void CGraphics::giluEndSurface(GLUnurbs* nurb)
{
  gluEndSurface(nurb);
}

inline void CGraphics::giluEndTrim(GLUnurbs* nurb)
{
  gluEndTrim(nurb);
}

inline const GLubyte* CGraphics::giluErrorString(GLenum error)
{
  return gluErrorString(error);
}

inline void CGraphics::giluGetNurbsProperty(GLUnurbs* nurb, GLenum property, GLfloat* data)
{
  gluGetNurbsProperty(nurb, property, data);
}

inline const GLubyte* CGraphics::giluGetString(GLenum name)
{
  return gluGetString(name);
}

inline void CGraphics::giluGetTessProperty(GLUtesselator* tess, GLenum which, GLdouble* data)
{
  gluGetTessProperty(tess, which, data);
}

inline void CGraphics::giluLoadSamplingMatrices(GLUnurbs* nurb, const GLfloat* model, const GLfloat* perspective, const GLint* view)
{
  gluLoadSamplingMatrices(nurb, model, perspective, view);
}

inline void CGraphics::giluLookAt(GLdouble eyeX, GLdouble eyeY, GLdouble eyeZ, GLdouble centerX, GLdouble centerY, GLdouble centerZ, GLdouble upX, GLdouble upY, GLdouble upZ)
{
  gluLookAt(eyeX, eyeY, eyeZ, centerX, centerY, centerZ, upX, upY, upZ);
}

inline GLUnurbs* CGraphics::giluNewNurbsRenderer()
{
  return gluNewNurbsRenderer();
}

inline GLUquadric* CGraphics::giluNewQuadric()
{
  return gluNewQuadric();
}

inline GLUtesselator* CGraphics::giluNewTess()
{
  return gluNewTess();
}

inline void CGraphics::giluNextContour(GLUtesselator* tess, GLenum type)
{
  gluNextContour(tess, type);
}

inline void CGraphics::giluNurbsCurve(GLUnurbs* nurb, GLint knotCount, GLfloat* knots, GLint stride, GLfloat* control, GLint order, GLenum type)
{
  gluNurbsCurve(nurb, knotCount, knots, stride, control, order, type);
}

inline void CGraphics::giluNurbsProperty(GLUnurbs* nurb, GLenum property, GLfloat value)
{
  gluNurbsProperty(nurb, property, value);
}

inline void CGraphics::giluNurbsSurface(GLUnurbs* nurb, GLint sKnotCount, GLfloat* sKnots, GLint tKnotCount, GLfloat* tKnots, GLint sStride, GLint tStride, GLfloat* control, GLint sOrder, GLint tOrder, GLenum type)
{
  gluNurbsSurface(nurb, sKnotCount, sKnots, tKnotCount, tKnots, sStride, tStride, control, sOrder, tOrder, type);
}

inline void CGraphics::giluOrtho2D(GLdouble left, GLdouble right, GLdouble bottom, GLdouble top)
{
  gluOrtho2D(left, right, bottom, top);
}

inline void CGraphics::giluPartialDisk(GLUquadric* quad, GLdouble inner, GLdouble outer, GLint slices, GLint loops, GLdouble start, GLdouble sweep)
{
  gluPartialDisk(quad, inner, outer, slices, loops, start, sweep);
}

inline void CGraphics::giluPerspective(GLdouble fovy, GLdouble aspect, GLdouble zNear, GLdouble zFar)
{
  gluPerspective(fovy, aspect, zNear, zFar);
}

inline void CGraphics::giluPickMatrix(GLdouble x, GLdouble y, GLdouble delX, GLdouble delY, GLint* viewport)
{
  gluPickMatrix(x, y, delX, delY, viewport);
}

inline GLint CGraphics::giluProject(GLdouble objX, GLdouble objY, GLdouble objZ, const GLdouble* model, const GLdouble* proj, const GLint* view, GLdouble* winX, GLdouble* winY, GLdouble* winZ)
{
  return gluProject(objX, objY, objZ, model, proj, view, winX, winY, winZ);
}

inline void CGraphics::giluPwlCurve(GLUnurbs* nurb, GLint count, GLfloat* data, GLint stride, GLenum type)
{
  gluPwlCurve(nurb, count, data, stride, type);
}

inline void CGraphics::giluQuadricDrawStyle(GLUquadric* quad, GLenum draw)
{
  gluQuadricDrawStyle(quad, draw);
}

inline void CGraphics::giluQuadricNormals(GLUquadric* quad, GLenum normal)
{
  gluQuadricNormals(quad, normal);
}

inline void CGraphics::giluQuadricOrientation(GLUquadric* quad, GLenum orientation)
{
  gluQuadricOrientation(quad, orientation);
}

inline void CGraphics::giluQuadricTexture(GLUquadric* quad, GLboolean texture)
{
  gluQuadricTexture(quad, texture);
}

inline GLint CGraphics::giluScaleImage(GLenum format, GLsizei wIn, GLsizei hIn, GLenum typeIn, const void* dataIn, GLsizei wOut, GLsizei hOut, GLenum typeOut, GLvoid* dataOut)
{
  return gluScaleImage(format, wIn, hIn, typeIn, dataIn, wOut, hOut, typeOut, dataOut);
}

inline void CGraphics::giluSphere(GLUquadric* quad, GLdouble radius, GLint slices, GLint stacks)
{
  gluSphere(quad, radius, slices, stacks);
}

inline void CGraphics::giluTessBeginContour(GLUtesselator* tess)
{
  gluTessBeginContour(tess);
}

inline void CGraphics::giluTessBeginPolygon(GLUtesselator* tess, GLvoid* data)
{
  gluTessBeginPolygon(tess, data);
}

inline void CGraphics::giluTessEndContour(GLUtesselator* tess)
{
  gluTessEndContour(tess);
}

inline void CGraphics::giluTessEndPolygon(GLUtesselator* tess)
{
  gluTessEndPolygon(tess);
}

inline void CGraphics::giluTessNormal(GLUtesselator* tess, GLdouble valueX, GLdouble valueY, GLdouble valueZ)
{
  gluTessNormal(tess, valueX, valueY, valueZ);
}

inline void CGraphics::giluTessProperty(GLUtesselator* tess, GLenum which, GLdouble data)
{
  gluTessProperty(tess, which, data);
}

inline void CGraphics::giluTessVertex(GLUtesselator* tess, GLdouble* location, GLvoid* data)
{
  gluTessVertex(tess, location, data);
}

inline GLint CGraphics::giluUnProject(GLdouble winX, GLdouble winY, GLdouble winZ, const GLdouble* model, const GLdouble* proj, const GLint* view, GLdouble* objX, GLdouble* objY, GLdouble* objZ)
{
  return gluUnProject(winX, winY, winZ, model, proj, view, objX, objY, objZ);
}

//////////////////////////////////////////////////
// Our Global Graphics Object
//

extern CGraphics gCGraphics;

#endif // __GRAPHICS_H__
