#ifndef __HEAP_H__
#define __HEAP_H__

#include "genlib.h"
#include "list.h"

//////////////////////////////////////////////////////////////////////
// CHeapEntry Class
//

template <class TPData, class TKey>
class CHeapEntry
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  inline CHeapEntry();
  inline ~CHeapEntry();
  inline VOID Init();
  inline VOID Init(UINT_32 uiIndex);
  inline VOID Init(UINT_32 uiIndex, TKey key, TPData pvData);
  inline VOID Uninit();

  inline BOOL Initialized();
  inline BOOL Valid();

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Output Functions
//
public:
  inline VOID Display();

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  inline UINT_32 GetIndex();
  inline VOID SetIndex(UINT_32 uiIndex);

  inline TKey GetKey();
  inline VOID SetKey(TKey key);
  inline TPData GetData();
  inline VOID SetData(TPData pvData);

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  UINT_32 m_uiIndex;

  TKey m_key;
  TPData m_pvData;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

template <class TPData, class TKey>
inline UINT_32 CHeapEntry<TPData, TKey>::GetIndex()
{
  return m_uiIndex;
}

template <class TPData, class TKey>
inline VOID CHeapEntry<TPData, TKey>::SetIndex(UINT_32 uiIndex)
{
  m_uiIndex = uiIndex;
}

template <class TPData, class TKey>
inline TKey CHeapEntry<TPData, TKey>::GetKey()
{
  return m_key;
}

template <class TPData, class TKey>
inline VOID CHeapEntry<TPData, TKey>::SetKey(TKey key)
{
  m_key = key;
}

template <class TPData, class TKey>
inline TPData CHeapEntry<TPData, TKey>::GetData()
{
  return m_pvData;
}

template <class TPData, class TKey>
inline VOID CHeapEntry<TPData, TKey>::SetData(TPData pvData)
{
  m_pvData = pvData;
}

//////////////////////////////////////////////////////////////////////
// CHeap Class
//

template <class TPData, class TKey>
class CHeap
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  inline CHeap();
  inline ~CHeap();
  inline VOID Init(UINT_32 uiSizeMax = 1);
  inline VOID Uninit();

  inline BOOL Initialized();
  inline BOOL Valid();

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Operations
//
public:
  CHeapEntry<TPData, TKey> *Insert(TKey key, TPData pvData);
  TPData PeekTop(TKey *pTKey = NULL);
  VOID RemoveAll();
  TPData RemoveTop(TKey *pTKey = NULL);
  TPData RemoveLast(TKey *pTKey = NULL);
  TPData Remove(CHeapEntry<TPData, TKey> *pCHeapEntry, TKey *pTKey = NULL);
  VOID Update(CHeapEntry<TPData, TKey> *pCHeapEntry, TKey keyNew);

protected:
  VOID Swap(UINT_32 uiIndex0, UINT_32 uiIndex1);
  VOID HeapifyUp(UINT_32 uiIndex);
  VOID HeapifyDown(UINT_32 uiIndex);

  // These functions automatically adjust m_uiSize
  CHeapEntry<TPData, TKey> *AllocateCHeapEntry();
  VOID DeallocateCHeapEntry(CHeapEntry<TPData, TKey> *pCHeapEntry);

//////////////////////////////////////////////////
// Output Functions
//
public:
  VOID Display();

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  inline BOOL GetBiggestFirst();
  inline VOID SetBiggestFirst(BOOL bBiggestFirst);

  inline UINT_32 GetSize();
  inline UINT_32 GetParent(UINT_32 uiIndex);
  inline UINT_32 GetLeftChild(UINT_32 uiIndex);
  inline UINT_32 GetRightChild(UINT_32 uiIndex);
  inline UINT_32 CheckRoot(UINT_32 uiIndex);
  inline UINT_32 CheckLeaf(UINT_32 uiIndex);

protected:
  inline VOID SetEntry(UINT_32 uiIndex, CHeapEntry<TPData, TKey> *pCHeapEntry);

//////////////////////////////////////////////////
// Class Variables
//
protected:
/*
  static UINT_32 m_uiObjectCount;
  static UINT_32 m_uiNewedEntries;
  static UINT_32 m_uiAllocatedEntries;
  static CList<CHeapEntry<TPData, TKey>*, UINT_32> *m_plistFreeEntries;
  */

  UINT_32 m_uiNewedEntries;
  UINT_32 m_uiAllocatedEntries;
  CList<CHeapEntry<TPData, TKey>*, UINT_32> *m_plistFreeEntries;

//////////////////////////////////////////////////
// Member Variables
//
protected:
  // I implemented this with a sentinel node so there are size+1 array entries,
  // but only size CHeapEntries
  BOOL m_bBiggestFirst;
  UINT_32 m_uiSize;
  UINT_32 m_uiSizeMax;

  CHeapEntry<TPData, TKey> **m_ppCHeapEntry;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

template <class TPData, class TKey>
inline BOOL CHeap<TPData, TKey>::GetBiggestFirst()
{
  return m_bBiggestFirst;
}

template <class TPData, class TKey>
inline VOID CHeap<TPData, TKey>::SetBiggestFirst(BOOL bBiggestFirst)
{
  m_bBiggestFirst = bBiggestFirst;
}

template <class TPData, class TKey>
inline UINT_32 CHeap<TPData, TKey>::GetSize()
{
  return m_uiSize;
}

template <class TPData, class TKey>
inline UINT_32 CHeap<TPData, TKey>::GetParent(UINT_32 uiIndex)
{
  return uiIndex/2;
}

template <class TPData, class TKey>
inline UINT_32 CHeap<TPData, TKey>::GetLeftChild(UINT_32 uiIndex)
{
  return 2*uiIndex;
}

template <class TPData, class TKey>
inline UINT_32 CHeap<TPData, TKey>::GetRightChild(UINT_32 uiIndex)
{
  return 2*uiIndex + 1;
}

template <class TPData, class TKey>
inline UINT_32 CHeap<TPData, TKey>::CheckRoot(UINT_32 uiIndex)
{
  return uiIndex == 1;
}

template <class TPData, class TKey>
inline UINT_32 CHeap<TPData, TKey>::CheckLeaf(UINT_32 uiIndex)
{
  return (2*uiIndex) > m_uiSize;
}

template <class TPData, class TKey>
inline VOID CHeap<TPData, TKey>::SetEntry(UINT_32 uiIndex, CHeapEntry<TPData, TKey> *pCHeapEntry)
{
  ASSERT( uiIndex <= m_uiSize );
  ASSERT( pCHeapEntry != NULL );

  m_ppCHeapEntry[uiIndex] = pCHeapEntry;
  m_ppCHeapEntry[uiIndex]->SetIndex(uiIndex);
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// Class Implementations
//

//////////////////////////////////////////////////////////////////////
// CHeapEntry Class
//

template <class TPData, class TKey>
CHeapEntry<TPData, TKey>::CHeapEntry()
{
  m_uiIndex = 0;

  ::SetMinValue(m_key);
  m_pvData = NULL;
}

template <class TPData, class TKey>
CHeapEntry<TPData, TKey>::~CHeapEntry()
{
  Uninit();
}

template <class TPData, class TKey>
VOID CHeapEntry<TPData, TKey>::Init()
{
  m_uiIndex = 0;

  ::SetMinValue(m_key);
  m_pvData = NULL;
}

template <class TPData, class TKey>
VOID CHeapEntry<TPData, TKey>::Init(UINT_32 uiIndex)
{
  Init();

  m_uiIndex = uiIndex;
}

template <class TPData, class TKey>
VOID CHeapEntry<TPData, TKey>::Init(UINT_32 uiIndex, TKey key, TPData pvData)
{
  Init(uiIndex);

  m_key = key;
  m_pvData = pvData;
}

template <class TPData, class TKey>
VOID CHeapEntry<TPData, TKey>::Uninit()
{
  m_uiIndex = 0;

  ::SetMinValue(m_key);
  m_pvData = NULL;
}

template <class TPData, class TKey>
BOOL CHeapEntry<TPData, TKey>::Initialized()
{
  return TRUE;
}

template <class TPData, class TKey>
BOOL CHeapEntry<TPData, TKey>::Valid()
{
//  return ( m_key == -INFINITY && m_pvData == NULL ) ||
//         ( m_pvData != NULL );
  return TRUE;
}

template <class TPData, class TKey>
VOID CHeapEntry<TPData, TKey>::Display()
{
  printf("(%ld: ", m_uiIndex);
  ::DisplayValue(m_key);
  printf(", %p)", m_pvData);
}

//////////////////////////////////////////////////////////////////////
// CHeap Class
//

/*
template <class TPData, class TKey>
UINT_32 CHeap<TPData, TKey>::m_uiObjectCount = 0;
template <class TPData, class TKey>
UINT_32 CHeap<TPData, TKey>::m_uiNewedEntries = 0;
template <class TPData, class TKey>
UINT_32 CHeap<TPData, TKey>::m_uiAllocatedEntries = 0;
template <class TPData, class TKey>
CList<CHeapEntry<TPData, TKey>*, UINT_32> *CHeap<TPData, TKey>::m_plistFreeEntries = NULL;
*/

template <class TPData, class TKey>
CHeap<TPData, TKey>::CHeap()
{
//  if ( m_uiObjectCount == 0 )
//    {
      m_plistFreeEntries = new CList<CHeapEntry<TPData, TKey>*, UINT_32>;
      ASSERT( m_plistFreeEntries != NULL );
      m_plistFreeEntries->Init();

      m_uiNewedEntries = 0;
      m_uiAllocatedEntries = 0;
//    }
//  m_uiObjectCount++;

  m_bBiggestFirst = TRUE;
  m_uiSizeMax = 0;

  m_ppCHeapEntry = NULL;
}

template <class TPData, class TKey>
CHeap<TPData, TKey>::~CHeap()
{
  Uninit();

//  m_uiObjectCount--;
//  if ( m_uiObjectCount == 0 )
//    {
      // Clean up our free list of Entries
      ASSERT( m_uiAllocatedEntries == 0 );
      if ( m_plistFreeEntries != NULL )
	{
	  CHeapEntry<TPData, TKey> *pCHeapEntry;

	  for ( pCHeapEntry = (CHeapEntry<TPData, TKey>*)m_plistFreeEntries->RemoveFirst();
		pCHeapEntry != NULL;
		pCHeapEntry = (CHeapEntry<TPData, TKey>*)m_plistFreeEntries->RemoveFirst() )
	    {
	      delete pCHeapEntry;
	    }

	  delete m_plistFreeEntries;
	  m_plistFreeEntries = NULL;
	}
      m_uiNewedEntries = 0;
      m_uiAllocatedEntries = 0;
//    }
}

template <class TPData, class TKey>
VOID CHeap<TPData, TKey>::Init(UINT_32 uiSizeMax/* = 1*/)
{

  ASSERT( uiSizeMax > 0 );
  
  m_bBiggestFirst = TRUE;
  m_uiSize = 0;
  m_uiSizeMax = uiSizeMax;

  m_ppCHeapEntry = new CHeapEntry<TPData, TKey>*[m_uiSizeMax + 1];
  ASSERT( m_ppCHeapEntry != NULL );
  memset(m_ppCHeapEntry, 0, (m_uiSizeMax+1)*sizeof(CHeapEntry<TPData, TKey>*));
}

template <class TPData, class TKey>
VOID CHeap<TPData, TKey>::Uninit()
{
  RemoveAll();

  m_bBiggestFirst = TRUE;
  m_uiSize = 0;
  m_uiSizeMax = 0;
  
  if ( m_ppCHeapEntry != NULL )
    {
      delete[] m_ppCHeapEntry;
      m_ppCHeapEntry = NULL;
    }
}

template <class TPData, class TKey>
BOOL CHeap<TPData, TKey>::Initialized()
{
  return m_ppCHeapEntry != NULL;
}

template <class TPData, class TKey>
BOOL CHeap<TPData, TKey>::Valid()
{
  return Initialized();
}

template <class TPData, class TKey>
CHeapEntry<TPData, TKey> *CHeap<TPData, TKey>::Insert(TKey key, TPData pvData)
{
  CHeapEntry<TPData, TKey> *pCHeapEntry;
/*
  for ( UINT_32 i = 1; i <= m_uiSize; i++ )
    {
      ASSERT( m_ppCHeapEntry[i]->GetData() != pvData );
    }
*/
  pCHeapEntry = AllocateCHeapEntry();
  pCHeapEntry->Init(m_uiSize, key, pvData);

  m_ppCHeapEntry[m_uiSize] = pCHeapEntry;

  HeapifyUp(m_uiSize);

  return pCHeapEntry;
}

template <class TPData, class TKey>
TPData CHeap<TPData, TKey>::PeekTop(TKey *pTKey/* = NULL*/)
{
  if ( m_uiSize == 0 )
    {
      return NULL;
    }

  if ( pTKey != NULL )
    {
      *pTKey = m_ppCHeapEntry[1]->GetKey();
    }

  return m_ppCHeapEntry[1]->GetData();
}

template <class TPData, class TKey>
VOID CHeap<TPData, TKey>::RemoveAll()
{
  CHeapEntry<TPData, TKey> *pCHeapEntry;

  while ( m_uiSize > 0 )
    {
      pCHeapEntry = m_ppCHeapEntry[m_uiSize];
      m_ppCHeapEntry[m_uiSize] = NULL;
      DeallocateCHeapEntry(pCHeapEntry);
    }
}

template <class TPData, class TKey>
TPData CHeap<TPData, TKey>::RemoveTop(TKey *pTKey/* = NULL*/)
{
  if ( m_uiSize == 0 )
    {
      return NULL;
    }

  return Remove(m_ppCHeapEntry[1], pTKey);
}

template <class TPData, class TKey>
TPData CHeap<TPData, TKey>::RemoveLast(TKey *pTKey/* = NULL*/)
{
  if ( m_uiSize == 0 )
    {
      return NULL;
    }

  return Remove(m_ppCHeapEntry[m_uiSize], pTKey);
}

template <class TPData, class TKey>
TPData CHeap<TPData, TKey>::Remove(CHeapEntry<TPData, TKey> *pCHeapEntry, TKey *pTKey/* = NULL*/)
{
  UINT_32 uiIndex;
  TPData pvData;

  ASSERT( pCHeapEntry != NULL );

  uiIndex = pCHeapEntry->GetIndex();
  ASSERT( m_ppCHeapEntry[uiIndex] == pCHeapEntry );

  pvData = pCHeapEntry->GetData();
  if ( pTKey != NULL )
    {
      *pTKey = pCHeapEntry->GetKey();
    }

  if ( uiIndex != m_uiSize )
    {
      // Swap the last array element into its position
      SetEntry(uiIndex, m_ppCHeapEntry[m_uiSize]);
      m_ppCHeapEntry[m_uiSize] = NULL;

      DeallocateCHeapEntry(pCHeapEntry);

      // Reheapify
      HeapifyDown(uiIndex);
    }
  else
    {
      DeallocateCHeapEntry(pCHeapEntry);
    }

  return pvData;
}

template <class TPData, class TKey>
VOID CHeap<TPData, TKey>::Update(CHeapEntry<TPData, TKey> *pCHeapEntry, TKey keyNew)
{
  UINT_32 uiIndex;
  TKey keyOld;

  ASSERT( pCHeapEntry != NULL );

  uiIndex = pCHeapEntry->GetIndex();
  ASSERT( m_ppCHeapEntry[uiIndex] == pCHeapEntry );

  keyOld = pCHeapEntry->GetKey();
  pCHeapEntry->SetKey(keyNew);

  if ( ( m_bBiggestFirst && 
	 ( keyNew < keyOld ) ) ||
       ( !m_bBiggestFirst && 
	 ( keyNew > keyOld ) ) )
    {
      HeapifyDown(uiIndex);
    }
  else
    {
      HeapifyUp(uiIndex);
    }
}

template <class TPData, class TKey>
VOID CHeap<TPData, TKey>::Swap(UINT_32 uiIndex0, UINT_32 uiIndex1)
{
  CHeapEntry<TPData, TKey> *pCHeapEntry;

  ASSERT( uiIndex0 <= m_uiSize );
  ASSERT( uiIndex1 <= m_uiSize );
  ASSERT( m_ppCHeapEntry[uiIndex0]->GetIndex() == uiIndex0 );
  ASSERT( m_ppCHeapEntry[uiIndex1]->GetIndex() == uiIndex1 );

  pCHeapEntry = m_ppCHeapEntry[uiIndex0];
  m_ppCHeapEntry[uiIndex0] = m_ppCHeapEntry[uiIndex1];
  m_ppCHeapEntry[uiIndex1] = pCHeapEntry;

  m_ppCHeapEntry[uiIndex0]->SetIndex(uiIndex0);
  m_ppCHeapEntry[uiIndex1]->SetIndex(uiIndex1);
}

template <class TPData, class TKey>
VOID CHeap<TPData, TKey>::HeapifyUp(UINT_32 uiIndex)
{
  UINT_32 uiParentIndex;

  if ( CheckRoot(uiIndex) )
    {
      return;
    }

  uiParentIndex = GetParent(uiIndex);

  if ( ( m_bBiggestFirst && 
	 ( m_ppCHeapEntry[uiIndex]->GetKey() > 
	   m_ppCHeapEntry[uiParentIndex]->GetKey() ) ) ||
       ( !m_bBiggestFirst && 
	 ( m_ppCHeapEntry[uiIndex]->GetKey() < 
	   m_ppCHeapEntry[uiParentIndex]->GetKey() ) ) )
    {
      Swap(uiParentIndex, uiIndex);

      HeapifyUp(uiParentIndex);
    }
}

template <class TPData, class TKey>
VOID CHeap<TPData, TKey>::HeapifyDown(UINT_32 uiIndex)
{
  UINT_32 uiChildIndex, uiLeftChildIndex, uiRightChildIndex;

  uiLeftChildIndex = GetLeftChild(uiIndex);
  uiRightChildIndex = GetRightChild(uiIndex);

  if ( uiLeftChildIndex > m_uiSize )
    {
      // This is a leaf node
      return;
    }
  else if ( uiRightChildIndex > m_uiSize )
    {
      // There is only a left child
      uiChildIndex = uiLeftChildIndex;
    }
  else
    {
      // There are two children so chose the maximum one
      if ( ( m_bBiggestFirst && 
	     ( m_ppCHeapEntry[uiRightChildIndex]->GetKey() > 
	       m_ppCHeapEntry[uiLeftChildIndex]->GetKey() ) ) ||
	   ( !m_bBiggestFirst && 
	     ( m_ppCHeapEntry[uiRightChildIndex]->GetKey() < 
	       m_ppCHeapEntry[uiLeftChildIndex]->GetKey() ) ) )
	{
	  uiChildIndex = uiRightChildIndex;
	}
      else
	{
	  uiChildIndex = uiLeftChildIndex;
	}
    }

  if ( ( m_bBiggestFirst && 
	 ( m_ppCHeapEntry[uiChildIndex]->GetKey() > 
	   m_ppCHeapEntry[uiIndex]->GetKey() ) ) ||
       ( !m_bBiggestFirst && 
	 ( m_ppCHeapEntry[uiChildIndex]->GetKey() < 
	   m_ppCHeapEntry[uiIndex]->GetKey() ) ) )
    {
      Swap(uiIndex, uiChildIndex);

      HeapifyDown(uiChildIndex);
    }
}

template <class TPData, class TKey>
CHeapEntry<TPData, TKey> *CHeap<TPData, TKey>::AllocateCHeapEntry()
{
  CHeapEntry<TPData, TKey> *pCHeapEntry;

  ASSERT( m_plistFreeEntries != NULL );
  ASSERT( m_uiSize <= m_uiSizeMax );
  ASSERT( m_ppCHeapEntry != NULL );

  if ( m_uiSize == m_uiSizeMax )
    {
      CHeapEntry<TPData, TKey> **ppCHeapEntry;
      UINT_32 uiSizeMax;

      uiSizeMax = m_uiSizeMax;
      ppCHeapEntry = m_ppCHeapEntry;

      // Allocate a new array
      m_uiSizeMax = 2*(m_uiSizeMax + 1) - 1;
      m_ppCHeapEntry = new CHeapEntry<TPData, TKey>*[m_uiSizeMax+1];
      ASSERT( m_ppCHeapEntry != NULL );
      memset(m_ppCHeapEntry, 0, (m_uiSizeMax+1)*sizeof(CHeapEntry<TPData, TKey>*));

      // Copy the current Heap into the new array
      memcpy(m_ppCHeapEntry, ppCHeapEntry, (uiSizeMax+1)*sizeof(CHeapEntry<TPData, TKey>*));

      // Deallocate the old array
      delete[] ppCHeapEntry;
    }

  m_uiSize++;

  // Allocate the Entry
  pCHeapEntry = (CHeapEntry<TPData, TKey>*)m_plistFreeEntries->RemoveFirst();
  if ( pCHeapEntry == NULL )
    {
      ASSERT( m_uiNewedEntries == m_uiAllocatedEntries );
      m_uiNewedEntries++;
      pCHeapEntry = new CHeapEntry<TPData, TKey>;
      ASSERT( pCHeapEntry != NULL );
    }
  ASSERT( m_uiAllocatedEntries < m_uiNewedEntries );

  m_uiAllocatedEntries++;

  ASSERT( m_uiSize <= m_uiSizeMax );

  return pCHeapEntry;
}

template <class TPData, class TKey>
VOID CHeap<TPData, TKey>::DeallocateCHeapEntry(CHeapEntry<TPData, TKey> *pCHeapEntry)
{
  ASSERT( pCHeapEntry != NULL );
  ASSERT( m_uiSize > 0 );

  m_uiSize--;

  ASSERT( m_plistFreeEntries != NULL );
  
  ASSERT( m_uiAllocatedEntries > 0 );

  m_plistFreeEntries->InsertFirst(pCHeapEntry, 0);
  m_uiAllocatedEntries--;
  ASSERT( m_uiAllocatedEntries < m_uiNewedEntries );
}

template <class TPData, class TKey>
VOID CHeap<TPData, TKey>::Display()
{
  UINT_32 i;

  printf("\nHeap: size = %ld max = %ld\n", m_uiSize, m_uiSizeMax); 

  for ( i = 1; i <= m_uiSize; i++ )
    {
      m_ppCHeapEntry[i]->Display();

      printf(" ");
      if ( (i & (i+1)) == 0x0 )
	{
	  printf("\n");
	}
    }

  printf("\n");
}
#endif // __HEAP_H__



