#include "genlib.h"

int main()
{
  UINT_32 uiDim, uiCoincidences;
  UINT_32 *m_puiCoincidentHalfPlanes;
  UINT_32 uiSetSize, 
  
  uiDim = 3;
  uiCoincidences = 5;

  m_puiCoincidentHalfPlanes = new UINT_32[uiCoincidences];
  ASSERT( m_puiCoincidentHalfPlanes != NULL );

  bInterior = FALSE;
  for ( uiMask = (0x1 << uiDim) - 1; 
	uiMask < (0x1 << uiCoincidences) && !bInterior; 
	uiMask++ )
    {
      // Build the inward facing vector
      vIn.Init(uiDim, m_pfNormalNeg);
      vIn.Null();
      uiBit = 0x1;
      uiSetSize = 0;
      for ( i = 0; i < uiCoincidences; i++ )
	{
	  if ( uiBit & uiMask )
	    {
	      // Add on to the vector
	      vNormal.Init(uiDim, GetValues(uiDim, i));
	      vIn += vNormal;
	      uiSetSize++;
	    }
	  uiBit = (uiBit << 1);
	}

      if ( uiSetSize >= uiDim )
	{
	  // We have the minimum number of half planes necessary so
	  // try to find an interior point
	  if ( !vIn.Normalize() )
	    {
	      continue;
	    }
	  vIn *= -1.0;
	  vIn.Init(uiDim+1, m_pfNormalNeg);
	  vIn[uiDim] = 0.0;
	  
	  printf("vIn = ");
	  vIn.Display();
	  
	  // So now shoot a stabbing line from the solution point 
	  // along vIn and find the first intersection with a 
	  // Hyperplane 
	  
	  fTMin = INFINITY;
	  for ( j = uiCoincidences; j < uiHalfSpaces; j++ )
	    {
	      vNormal.Init(uiDim+1, GetValues(uiDim, j));      
	      ASSERT( vSol.Dot(vNormal) != 0.0 );
	      fT = vNormal.Dot(vIn);
	      if ( fT != 0.0 )
		{
		  fT = -vNormal.Dot(vSol) / fT;
		  if ( fT > 0.0 && fT < fTMin )
		    {
		      fTMin = fT;
		    }
		}
	    }
	  
	  if ( fTMin == INFINITY )
	    {
	      continue;
	    }
	  
	  // Compute the interior point as the midpoint from 
	  // the line equation
	  fTMin = fTMin / 2.0;
	  for ( j = 0; j < uiDim; j++ )
	    {
	      vIn[j] = vSol[j] + fTMin*vIn[j];
	    }
	  vIn[3] = 1.0;
	  
	  printf("Interior Point = ");
	  vIn.Display();
	  
	  // Verify that the Interior point really lies within 
	  // all of the half spaces, and not on them
	  bInterior = TRUE;
	  for ( j = 0; j < uiHalfSpaces; j++ )
	    {
	      vNormal.Init(uiDim+1, GetValues(uiDim, j));
	      if ( vIn.Dot(vNormal) >= 0.0 )
		{
		  bInterior = FALSE;
		  break;
		}
	    }
	}
    }

  return 0;
}

// The inner loop
{
  for ( i = 0; i < uiSetSize; i++ )
    {
      printf("%ld ", m_puiCoincidentHalfPlane[i]);
    }
  printf("\n");
}
