#ifndef __LIST_H__
#define __LIST_H__

#include "genlib.h"

//////////////////////////////////////////////////////////////////////////////
// CLink Class
//

template <class TPData, class TKey>
class CLink
{
public:
  inline CLink();
  inline ~CLink();
  inline BOOL Init(TKey key, TPData pData);

protected:
  inline VOID Uninit();

public:
  inline VOID Display();

  // Inlined Functions
public:
  inline CLink<TPData, TKey> *GetNext();
  inline VOID SetNext(CLink *pNext);
  inline CLink<TPData, TKey> *GetPrev();
  inline VOID SetPrev(CLink *pPrev);
  inline TKey GetKey();
  inline TPData GetData(TKey *pTKey = NULL);

protected:
  CLink *m_pNext;
  CLink *m_pPrev;

  TKey m_key;
  TPData m_pData;
};

//////////////////////
// Inlined Functions
//

template <class TPData, class TKey>
inline CLink<TPData, TKey> *CLink<TPData, TKey>::GetNext()
{
  return m_pNext;
}

template <class TPData, class TKey>
inline VOID CLink<TPData, TKey>::SetNext(CLink *pNext)
{
  m_pNext = pNext;
}

template <class TPData, class TKey>
inline CLink<TPData, TKey> *CLink<TPData, TKey>::GetPrev()
{
  return m_pPrev;
}

template <class TPData, class TKey>
inline VOID CLink<TPData, TKey>::SetPrev(CLink *pPrev)
{
  m_pPrev = pPrev;
}

template <class TPData, class TKey>
inline TKey CLink<TPData, TKey>::GetKey()
{
  return m_key;
}

template <class TPData, class TKey>
inline TPData CLink<TPData, TKey>::GetData(TKey *pTKey/* = NULL*/)
{
  if ( pTKey != NULL )
    {
      *pTKey = m_key;
    }
  
  return m_pData;
}

#define CLINKBUFFER
#ifdef CLINKBUFFER

//////////////////////////////////////////////////////////////////////////////
// CLinkBuffer Class
//

template <class TPData, class TKey>
class CLinkBuffer
{
public:
  inline CLinkBuffer();
  inline ~CLinkBuffer();
  inline BOOL Init(UINT_32 uiSize = 2);

protected:
  inline VOID Uninit();

  // Inlined Functions
public:
  inline CLinkBuffer<TPData, TKey> *GetNext();
  inline VOID SetNext(CLinkBuffer *pNext);
  inline CLink<TPData, TKey> *GetBuffer();

protected:
  CLinkBuffer *m_pNext;

  UINT_32 m_uiCLinks;
  CLink<TPData, TKey> *m_pCLinks;
};

//////////////////////////////////////////////////
// Inlined Functions
//

template <class TPData, class TKey>
inline CLinkBuffer<TPData, TKey> *CLinkBuffer<TPData, TKey>::GetNext()
{
  return m_pNext;
}

template <class TPData, class TKey>
inline VOID CLinkBuffer<TPData, TKey>::SetNext(CLinkBuffer *pNext)
{
  m_pNext = pNext;
}

template <class TPData, class TKey>
inline CLink<TPData, TKey> *CLinkBuffer<TPData, TKey>::GetBuffer()
{
  return m_pCLinks;
}

#endif // CLINKBUFFER

//////////////////////////////////////////////////////////////////////////////
// CList Class
//

template <class TPData, class TKey>
class CListIter;

template <class TPData, class TKey>
class CList
{
public:
  friend class CListIter<TPData, TKey>;

public:
  inline CList();
  inline ~CList();
  inline BOOL Init();

protected:
  inline VOID Uninit();
  inline BOOL Initialized() const;
  inline BOOL Valid() const;

public:
  // Insertion Functions
  inline TPData Insert(TPData pData, TKey key);
  inline TPData InsertForward(TPData pData, TKey key);
  inline TPData InsertBackward(TPData pData, TKey key);
  inline TPData InsertFirst(TPData pData, TKey key);
  inline TPData InsertLast(TPData pData, TKey key);

  // Deletion Functions
  inline TPData Remove(TKey key);
  inline TPData RemoveForward(TKey key);
  inline TPData RemoveBackward(TKey key);
  inline TPData RemoveFirst(TKey *pTKey = NULL);
  inline TPData RemoveLast(TKey *pTKey = NULL);
  inline VOID RemoveAll();
  inline TPData RemoveUnordered(TKey key);

  // Look up Functions
  inline TPData Peek(TKey key) const;
  inline TPData PeekForward(TKey key) const;
  inline TPData PeekBackward(TKey key) const;
  inline TPData PeekFirst(TKey *pTKey = NULL) const;
  inline TPData PeekLast(TKey *pTKey = NULL) const;
  inline TPData PeekUnordered(TKey key) const;

#define SORT_ENABLED
#ifdef SORT_ENABLED
  // Sorting Functions
  inline BOOL Sort(); // This preforms a Merge Sort
#endif // SORT_ENABLED

  // Display Functions
  inline VOID Display();

  //Inlined Functions
public:
  inline UINT_32 GetLength();

protected:
  // CLink Allocation
  inline CLink<TPData, TKey> *AllocateLink();
  inline VOID FreeLink(CLink<TPData, TKey> *pCLink);

  // Insertion and Deletion
  inline VOID InsertNext(CLink<TPData, TKey> *pCurr, CLink<TPData, TKey> *pCLink);
  inline CLink<TPData, TKey> *RemoveNext(CLink<TPData, TKey> *pCurr);
  inline VOID InsertPrev(CLink<TPData, TKey> *pCurr, CLink<TPData, TKey> *pCLink);
  inline CLink<TPData, TKey> *RemovePrev(CLink<TPData, TKey> *pCurr);

#ifdef SORT_ENABLED
  // Merge Sort Functions
  inline CLink<TPData, TKey> *Divide(CLink<TPData, TKey> *pCLink, UINT_32 uiLength);
  inline CLink<TPData, TKey> *Merge(CLink<TPData, TKey> *pLeft, CLink<TPData, TKey> *pRight);
  inline VOID MergeAdd(CLink<TPData, TKey>** ppTail, CLink<TPData, TKey>** ppCLink);
#endif // SORT_ENABLED

protected:
  CLink<TPData, TKey> *m_pHead;
  CLink<TPData, TKey> *m_pTail; // This is the same Sentinal Node as the Head 

  UINT_32 m_uiAllocations;
  UINT_32 m_uiLength;

#ifdef CLINKBUFFER
  CLink<TPData, TKey> *m_pFree;
  CLinkBuffer<TPData, TKey> *m_pCLinkBuffer;
#endif // CLINKBUFFER
};

//////////////////////////////////////////////////
// Inlined Functions
//

template <class TPData, class TKey>
inline BOOL CList<TPData, TKey>::Initialized() const
{
  return (m_pHead != NULL) &&
         (m_pTail != NULL) &&
         (m_uiLength >= 1) &&
         (m_uiAllocations >= 1);
}

template <class TPData, class TKey>
inline BOOL CList<TPData, TKey>::Valid() const
{
  return Initialized() && 
         (m_pHead == m_pTail) &&
         (m_uiAllocations == m_uiLength);
}

template <class TPData, class TKey>
inline TPData CList<TPData, TKey>::Insert(TPData pData, TKey key)
{
  return InsertForward(pData, key);
}

template <class TPData, class TKey>
inline TPData CList<TPData, TKey>::Remove(TKey key)
{
  return RemoveForward(key);
}

template <class TPData, class TKey>
inline TPData CList<TPData, TKey>::Peek(TKey key) const
{
  return PeekForward(key);
}

template <class TPData, class TKey>
inline UINT_32 CList<TPData, TKey>::GetLength()
{
  return m_uiLength - 1;
}

//////////////////////////////////////////////////////////////////////////////
// CListIter Class
//

template <class TPData, class TKey>
class CListIter
{
public:
  inline CListIter();
  inline ~CListIter();
  inline BOOL Init(CList<TPData, TKey> *pCList);

protected:
  inline VOID Uninit();
  inline BOOL Valid();

public:
  // Iteration Functions
  inline TPData Peek(TKey key);
  inline TPData PeekForward(TKey key);
  inline TPData PeekBackward(TKey key);
  inline TPData PeekFirst(TKey *pTKey = NULL);
  inline TPData PeekNext(TKey *pTKey = NULL);
  inline TPData PeekLast(TKey *pTKey = NULL);
  inline TPData PeekPrev(TKey *pTKey = NULL);

  // Modification and Iteration Functions
  inline TPData PeekNextRemoveCurr(TPData *ppDataRemoved = NULL, TKey *pTKeyRemoved = NULL,
				   TKey *pTKey = NULL);
  inline TPData PeekPrevRemoveCurr(TPData *ppDataRemoved = NULL, TKey *pTKeyRemoved = NULL,
				   TKey *pTKey = NULL);
  inline TPData InsertNext(TPData pData, TKey key);
  inline TPData InsertPrev(TPData pData, TKey key);

  // Cyclic iteration functions
  inline TPData PeekNextWrap(TKey *pTKey = NULL);
  inline TPData PeekPrevWrap(TKey *pTKey = NULL);
  inline TPData PeekNextWrapRemoveCurr(TPData *ppDataRemoved = NULL, TKey *pTKeyRemoved = NULL,
				   TKey *pTKey = NULL);
  inline TPData PeekPrevWrapRemoveCurr(TPData *ppDataRemoved = NULL, TKey *pTKeyRemoved = NULL,
				   TKey *pTKey = NULL);

protected:
  CList<TPData, TKey> *m_pCList;
  CLink<TPData, TKey> *m_pCurr;
};

//////////////////////////////////////////////////
// Inlined Functions
//

template <class TPData, class TKey>
inline TPData CListIter<TPData, TKey>::Peek(TKey key)
{
  return PeekForward(key);
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// Class Implementations
//

//////////////////////////////////////////////////////////////////////////////
// CLink Class
//

template <class TPData, class TKey>
inline CLink<TPData, TKey>::CLink()
{
  m_pNext = NULL;
  m_pPrev = NULL;

  ::SetNull(m_key);
  m_pData = NULL;
}

template <class TPData, class TKey>
inline CLink<TPData, TKey>::~CLink()
{
  Uninit();
}

template <class TPData, class TKey>
inline BOOL CLink<TPData, TKey>::Init(TKey key, TPData pData)
{
  m_pNext = NULL;
  m_pPrev = NULL;

  m_key = key;
  m_pData = pData;

  return TRUE;
}

template <class TPData, class TKey>
inline VOID CLink<TPData, TKey>::Uninit()
{
  m_pNext = NULL;

  ::SetNull(m_key);
  m_pData = NULL;
}

template <class TPData, class TKey>
inline VOID CLink<TPData, TKey>::Display()
{
  printf("( k = ");
  ::DisplayValue(m_key);
  printf(", d = %p)", m_pData);
}

//////////////////////////////////////////////////////////////////////////////
// CLinkBuffer Class
//

template <class TPData, class TKey>
inline CLinkBuffer<TPData, TKey>::CLinkBuffer()
{
  m_pNext = NULL;

  m_uiCLinks = 0;
  m_pCLinks = NULL;
}

template <class TPData, class TKey>
inline CLinkBuffer<TPData, TKey>::~CLinkBuffer()
{
  Uninit();
}

template <class TPData, class TKey>
inline BOOL CLinkBuffer<TPData, TKey>::Init(UINT_32 uiSize/* = 20*/)
{
  UINT_32 i;

  ASSERT( uiSize != 0 );

  m_pNext = NULL;

  m_uiCLinks = uiSize;

  m_pCLinks = new CLink<TPData, TKey>[m_uiCLinks];
  ASSERT( m_pCLinks != NULL );

  for ( i = 1; i < m_uiCLinks; i++ )
    {
      m_pCLinks[i-1].SetNext(&(m_pCLinks[i]));
    }

  return TRUE;
}

template <class TPData, class TKey>
inline VOID CLinkBuffer<TPData, TKey>::Uninit()
{
  m_pNext = NULL;

  m_uiCLinks = 0;
  if ( m_pCLinks != NULL )
    {
      delete []m_pCLinks;
      m_pCLinks = NULL;
    }
}

//////////////////////////////////////////////////////////////////////////////
// CList Class
//

template <class TPData, class TKey>
inline CList<TPData, TKey>::CList()
{
  m_pHead = NULL;
  m_pTail = NULL;

  m_uiAllocations = 0;
  m_uiLength = 0;

#ifdef CLINKBUFFER
  m_pFree = NULL;
  m_pCLinkBuffer = NULL;
#endif // CLINKBUFFER
}

template <class TPData, class TKey>
inline CList<TPData, TKey>::~CList()
{
  Uninit();
}

template <class TPData, class TKey>
inline BOOL CList<TPData, TKey>::Init()
{
  ASSERT( m_pHead == NULL &&
	  m_pTail == NULL &&
	  m_uiLength == 0 );

  m_pHead = AllocateLink();
  ASSERT( m_pHead != NULL );
  m_pHead->SetNext(m_pHead);
  m_pHead->SetPrev(m_pHead);
  m_pTail = m_pHead;

  m_uiLength = 1;
  ASSERT( Valid() );

  return TRUE;
}

template <class TPData, class TKey>
inline VOID CList<TPData, TKey>::Uninit()
{
  if ( m_pHead != NULL )
    {
      RemoveAll();
      ASSERT( m_pHead != NULL &&
	      m_pHead == m_pTail &&
	      m_uiLength == 1 );      
      
      m_uiLength = 0;
      FreeLink(m_pHead);
      ASSERT( m_uiAllocations == 0 );
    }

  m_pHead = NULL;
  m_pTail = NULL;
  m_uiAllocations = 0;
  m_uiLength = 0;

#ifdef CLINKBUFFER
  while ( m_pCLinkBuffer != NULL )
    {
      CLinkBuffer<TPData, TKey> *pCLinkBuffer;

      pCLinkBuffer = m_pCLinkBuffer;
      m_pCLinkBuffer = m_pCLinkBuffer->GetNext();
      delete pCLinkBuffer;
    }

  m_pFree = NULL;
#endif // CLINKBUFFER
}

template <class TPData, class TKey>
inline TPData CList<TPData, TKey>::InsertForward(TPData pData, TKey key)
{
  ASSERT( Valid() );

  CLink<TPData, TKey> *pCurr;
  CLink<TPData, TKey> *pCLink;

  // Allocate and Initialize a CLink
  pCLink = AllocateLink();
  ASSERT( pCLink != NULL );
  pCLink->Init(key, pData);

  // Find the correct position in the CList for the CLink
  for ( pCurr = m_pHead; pCurr->GetNext() != m_pTail; pCurr = pCurr->GetNext() )
    {
      if ( pCurr->GetNext()->GetKey() > key )
	{
	  break;
	} 
    }
 
  // Insert the CLink at the current position in the CList
  InsertNext(pCurr, pCLink);
  
  return pData;
}

template <class TPData, class TKey>
inline TPData CList<TPData, TKey>::InsertBackward(TPData pData, TKey key)
{
  ASSERT( Valid() );

  CLink<TPData, TKey> *pCurr;
  CLink<TPData, TKey> *pCLink;

  // Allocate and Initialize a CLink
  pCLink = AllocateLink();
  ASSERT( pCLink != NULL );
  pCLink->Init(key, pData);

  // Find the correct position in the CList for the CLink
  for ( pCurr = m_pTail; pCurr->GetPrev() != m_pHead; pCurr = pCurr->GetPrev() )
    {
      if ( pCurr->GetPrev()->GetKey() < key )
	{
	  break;
	} 
    }
 
  // Insert the CLink at the current position in the CList
  InsertPrev(pCurr, pCLink);
  
  return pData;
}

template <class TPData, class TKey>
inline TPData CList<TPData, TKey>::InsertFirst(TPData pData, TKey key)
{
  ASSERT( Valid() );

  CLink<TPData, TKey> *pCLink;

  // Allocate and Initialize a CLink
  pCLink = AllocateLink();
  ASSERT( pCLink != NULL );
  pCLink->Init(key, pData);

  // Insert the CLink at the beginning of the CList
  InsertNext(m_pHead, pCLink);
  
  return pData;
}

template <class TPData, class TKey>
inline TPData CList<TPData, TKey>::InsertLast(TPData pData, TKey key)
{
  ASSERT( Valid() );

  CLink<TPData, TKey> *pCLink;

  // Allocate and Initialize a CLink
  pCLink = AllocateLink();
  ASSERT( pCLink != NULL );
  pCLink->Init(key, pData);

  // Insert the CLink at the end of the CList
  InsertPrev(m_pTail, pCLink);
  
  return pData;
}

template <class TPData, class TKey>
inline TPData CList<TPData, TKey>::RemoveForward(TKey key)
{
  ASSERT( Valid() );

  CLink<TPData, TKey> *pCurr;
  CLink<TPData, TKey> *pCLink;
  TPData pData;

  pData = NULL;

  // Find the correct position in the CList for the CLink
  for ( pCurr = m_pHead; pCurr->GetNext() != m_pTail; pCurr = pCurr->GetNext() )
    {
      if ( pCurr->GetNext()->GetKey() < key )
	{
	  // Not there yet
	  continue;
	}
      else if ( pCurr->GetNext()->GetKey() == key )
	{
	  // We found it

	  // Remove this next CLink
	  pCLink = RemoveNext(pCurr);
	  ASSERT( pCLink != NULL );
	  
	  // Pull out the Data from the CLink
	  pData = pCLink->GetData(NULL);

	  // Deallocate the CLink
	  FreeLink(pCLink);

	  break;
	}
      else
	{
	  // We passed it
	  break;
	}
    }

  return pData;
}

template <class TPData, class TKey>
inline TPData CList<TPData, TKey>::RemoveBackward(TKey key)
{
  ASSERT( Valid() );

  CLink<TPData, TKey> *pCurr;
  CLink<TPData, TKey> *pCLink;
  TPData pData;

  pData = NULL;

  // Find the correct position in the CList for the CLink
  for ( pCurr = m_pTail; pCurr->GetPrev() != m_pHead; pCurr = pCurr->GetPrev() )
    {
      if ( pCurr->GetPrev()->GetKey() > key )
	{
	  // Not there yet
	  continue;
	}
      else if ( pCurr->GetPrev()->GetKey() == key )
	{
	  // We found it

	  // Remove this prev CLink
	  pCLink = RemovePrev(pCurr);
	  ASSERT( pCLink != NULL );
	  
	  // Pull out the Data from the CLink
	  pData = pCLink->GetData(NULL);

	  // Deallocate the CLink
	  FreeLink(pCLink);

	  break;
	}
      else
	{
	  // We passed it
	  break;
	}
    }

  return pData;
}

template <class TPData, class TKey>
inline TPData CList<TPData, TKey>::RemoveFirst(TKey *pTKey/* = NULL*/)
{
  ASSERT( Valid() );

  CLink<TPData, TKey> *pCLink;
  TPData pData;

  // Remove the CLink at the beginning of the CList
  pCLink = RemoveNext(m_pHead);
  if ( pCLink == NULL )
    {
      return NULL;
    }

  // Pull out the Data from the CLink
  pData = pCLink->GetData(pTKey);

  // Deallocate the CLink
  FreeLink(pCLink);

  return pData;
}

template <class TPData, class TKey>
inline TPData CList<TPData, TKey>::RemoveLast(TKey *pTKey/* = NULL*/)
{
  ASSERT( Valid() );

  CLink<TPData, TKey> *pCLink;
  TPData pData;

  // Remove the CLink at the end of the CList
  pCLink = RemovePrev(m_pTail);
  if ( pCLink == NULL )
    {
      return NULL;
    }

  // Pull out the Data from the CLink
  pData = pCLink->GetData(pTKey);

  // Deallocate the CLink
  FreeLink(pCLink);

  return pData;
}

template <class TPData, class TKey>
inline VOID CList<TPData, TKey>::RemoveAll()
{
  ASSERT( Valid() );

  CLink<TPData, TKey> *pCLink;

  pCLink = RemoveNext(m_pHead);
  while ( pCLink != NULL )
    {
      FreeLink(pCLink);
      pCLink = RemoveNext(m_pHead);
    }
}

template <class TPData, class TKey>
inline TPData CList<TPData, TKey>::RemoveUnordered(TKey key)
{
  ASSERT( Valid() );

  CLink<TPData, TKey> *pCurr;
  CLink<TPData, TKey> *pCLink;
  TPData pData;

  pData = NULL;

  // Find the correct position in the CList for the CLink
  for ( pCurr = m_pHead; pCurr->GetNext() != m_pTail; pCurr = pCurr->GetNext() )
    {
      if ( pCurr->GetNext()->GetKey() == key )
	{
	  // We found it

	  // Remove this next CLink
	  pCLink = RemoveNext(pCurr);
	  ASSERT( pCLink != NULL );
	  
	  // Pull out the Data from the CLink
	  pData = pCLink->GetData(NULL);

	  // Deallocate the CLink
	  FreeLink(pCLink);

	  break;
	}
    }

  if ( pCurr->GetNext()->GetKey() == key )
    {
      // We found it
      
      // Remove this next CLink
      pCLink = RemoveNext(pCurr);
      ASSERT( pCLink != NULL );
      
      // Pull out the Data from the CLink
      pData = pCLink->GetData(NULL);
      
      // Deallocate the CLink
      FreeLink(pCLink);
    }

  return pData;
}

template <class TPData, class TKey>
inline TPData CList<TPData, TKey>::PeekForward(TKey key) const
{
  ASSERT( Valid() ); 

  CLink<TPData, TKey> *pCurr;
  TPData pData;

  pData = NULL;

  // Find the correct position in the CList for the CLink
  for ( pCurr = m_pHead->GetNext(); pCurr != m_pTail; 
	pCurr = pCurr->GetNext() )
    {
      if ( pCurr->GetKey() < key )
	{
	  // Not there yet
	  continue;
	}
      else if ( pCurr->GetKey() == key )
	{
	  // We found it

	  // Pull out the Data from the CLink
	  pData = pCurr->GetData(NULL);
	  
	  break;
	}
      else
	{
	  // We passed it
	  break;
	}
    }

  return pData;
}

template <class TPData, class TKey>
inline TPData CList<TPData, TKey>::PeekBackward(TKey key) const
{
  ASSERT( Valid() ); 

  CLink<TPData, TKey> *pCurr;
  TPData pData;

  pData = NULL;

  // Find the correct position in the CList for the CLink
  for ( pCurr = m_pTail->GetPrev(); pCurr != m_pHead; 
	pCurr = pCurr->GetPrev() )
    {
      if ( pCurr->GetKey() > key )
	{
	  // Not there yet
	  continue;
	}
      else if ( pCurr->GetKey() == key )
	{
	  // We found it

	  // Pull out the Data from the CLink
	  pData = pCurr->GetData(NULL);
	  
	  break;
	}
      else
	{
	  // We passed it
	  break;
	}
    }

  return pData;
}

template <class TPData, class TKey>
inline TPData CList<TPData, TKey>::PeekFirst(TKey *pTKey/* = NULL */) const
{
  CLink<TPData, TKey> *pCurr;

  ASSERT( Valid() ); 

  pCurr = m_pHead->GetNext();
  if ( pCurr == m_pTail )
    {
      pCurr = NULL;
      return NULL;
    }

  return pCurr->GetData(pTKey);
}

template <class TPData, class TKey>
inline TPData CList<TPData, TKey>::PeekLast(TKey *pTKey/* = NULL */) const
{
  CLink<TPData, TKey> *pCurr;

  ASSERT( Valid() );

  pCurr = m_pTail->GetPrev();
  if ( pCurr == m_pHead )
    {
      pCurr = NULL;
      return NULL;
    }

  return pCurr->GetData(pTKey);
}

template <class TPData, class TKey>
inline TPData CList<TPData, TKey>::PeekUnordered(TKey key) const
{
  ASSERT( Valid() );

  CLink<TPData, TKey> *pCurr;
  TPData pData;

  pData = NULL;

  // Find the correct position in the CList for the CLink
  for ( pCurr = m_pHead; pCurr->GetNext() != m_pTail; pCurr = pCurr->GetNext() )
    {
      if ( pCurr->GetKey() == key )
	{
	  // We found it

	  // Pull out the Data from the CLink
	  pData = pCurr->GetData(NULL);

	  break;
	}
    }

  if ( pCurr->GetKey() == key )
    {
      // We found it
      
      // Pull out the Data from the CLink
      pData = pCurr->GetData(NULL);
    }

  return pData;
}

template <class TPData, class TKey>
inline VOID CList<TPData, TKey>::Display()
{
  ASSERT( Valid() );
  
  CLink<TPData, TKey> *pCurr;

  for ( pCurr = m_pHead; pCurr->GetNext() != m_pTail; pCurr = pCurr->GetNext() )
    {
      if ( pCurr == m_pHead )
	{
	  printf("(HEAD)->");
	}
      pCurr->GetNext()->Display();
      printf("->");
      if ( m_pTail == pCurr->GetNext() )
	{
	  printf("<-(TAIL)");
	}
    }
  printf("\n");
}

#ifdef CLINKBUFFER

template <class TPData, class TKey>
inline CLink<TPData, TKey> *CList<TPData, TKey>::AllocateLink()
{
  CLink<TPData, TKey> *pCLink;  

  ASSERT( m_uiAllocations == m_uiLength );

  if ( m_pFree == NULL )
    {
//printf("*** Allocating a new link Buffer ***\n");
      // The free list is empty so
      // dynamically allocate a new buffer
      CLinkBuffer<TPData, TKey> *pCLinkBuffer;  

      pCLinkBuffer = new CLinkBuffer<TPData, TKey>;
      ASSERT( pCLinkBuffer != NULL );

      pCLinkBuffer->Init();

      // Insert the new buffer at the beginning of the buffer list
      // so that we can deallocate it later
      pCLinkBuffer->SetNext(m_pCLinkBuffer);
      m_pCLinkBuffer = pCLinkBuffer;

      // Set the free list to be the list contained in the buffer
      m_pFree = m_pCLinkBuffer->GetBuffer();
      ASSERT( m_pFree != NULL );
    }

  ASSERT( m_pFree != NULL );
  // Take the first free link off the free list
  pCLink = m_pFree;
  m_pFree = m_pFree->GetNext();
  pCLink->SetNext(NULL);
  m_uiAllocations++;

  return pCLink;
}

template <class TPData, class TKey>
inline VOID CList<TPData, TKey>::FreeLink(CLink<TPData, TKey> *pCLink)
{
  ASSERT( pCLink != NULL );

  // Insert the newly freed link at the beginning of the free list to be reused
  pCLink->SetNext(m_pFree);
  m_pFree = pCLink;
  m_uiAllocations--;

  ASSERT( m_uiAllocations == m_uiLength );
}

#else // CLINKBUFFER

template <class TPData, class TKey>
inline CLink<TPData, TKey> *CList<TPData, TKey>::AllocateLink()
{
  CLink<TPData, TKey> *pCLink;  

  ASSERT( Valid() );

  pCLink = new CLink<TPData, TKey>;
  ASSERT( pCLink != NULL );
  m_uiAllocations++;

  return pCLink;
}

template <class TPData, class TKey>
inline VOID CList<TPData, TKey>::FreeLink(CLink<TPData, TKey> *pCLink)
{
  ASSERT( Initialized() );
  ASSERT( pCLink != NULL );

  delete pCLink;
  m_uiAllocations--;

  ASSERT( Valid() );
}

#endif // CLINKBUFFER

template <class TPData, class TKey>
inline VOID CList<TPData, TKey>::InsertNext(CLink<TPData, TKey> *pCurr, CLink<TPData, TKey> *pCLink)
{
  ASSERT( Initialized() ); 
  ASSERT( pCurr != NULL && pCLink != NULL );

  pCLink->SetNext(pCurr->GetNext());
  pCLink->SetPrev(pCurr);
  pCurr->GetNext()->SetPrev(pCLink);
  pCurr->SetNext(pCLink);
  m_uiLength++;

  ASSERT( Valid() );
}

template <class TPData, class TKey>
inline CLink<TPData, TKey> *CList<TPData, TKey>::RemoveNext(CLink<TPData, TKey> *pCurr)
{
  ASSERT( Valid() ); 
  ASSERT( pCurr != NULL );

  CLink<TPData, TKey> *pCLink;

  pCLink = pCurr->GetNext();
  if ( pCLink == m_pTail )
    {
      return NULL;
    }
  
  pCurr->SetNext(pCLink->GetNext());
  pCurr->GetNext()->SetPrev(pCurr);
  m_uiLength--;

  return pCLink;
}

template <class TPData, class TKey>
inline VOID CList<TPData, TKey>::InsertPrev(CLink<TPData, TKey> *pCurr, CLink<TPData, TKey> *pCLink)
{
  ASSERT( Initialized() ); 
  ASSERT( pCurr != NULL && pCLink != NULL );

  pCLink->SetPrev(pCurr->GetPrev());
  pCLink->SetNext(pCurr);
  pCurr->GetPrev()->SetNext(pCLink);
  pCurr->SetPrev(pCLink);
  m_uiLength++;

  ASSERT( Valid() );
}

template <class TPData, class TKey>
inline CLink<TPData, TKey> *CList<TPData, TKey>::RemovePrev(CLink<TPData, TKey> *pCurr)
{
  ASSERT( Valid() ); 
  ASSERT( pCurr != NULL );

  CLink<TPData, TKey> *pCLink;

  pCLink = pCurr->GetPrev();
  if ( pCLink == m_pTail )
    {
      return NULL;
    }
  
  pCurr->SetPrev(pCLink->GetPrev());
  pCurr->GetPrev()->SetNext(pCurr);
  m_uiLength--;

  return pCLink;
}

#ifdef SORT_ENABLED

template <class TPData, class TKey>
inline BOOL CList<TPData, TKey>::Sort()
{
  ASSERT( Valid() );

  UINT_32 i;
  CLink<TPData, TKey> *pCLink;

  if ( GetLength() == 0 )
    {
      return TRUE;
    }

  pCLink = Divide(m_pHead->GetNext(), GetLength());
  ASSERT( pCLink != NULL );

  m_pHead->SetNext(pCLink);
  i = 0;
  for ( pCLink = m_pHead; pCLink->GetNext() != NULL; pCLink = pCLink->GetNext() )
    {
      pCLink->GetNext()->SetPrev(pCLink);
      i++;
    }
  ASSERT( i == GetLength() );
  pCLink->SetNext(m_pTail);
  m_pTail->SetPrev(pCLink);

  return TRUE;
}

template <class TPData, class TKey>
inline CLink<TPData, TKey> *CList<TPData, TKey>::Divide(CLink<TPData, TKey> *pCLink, UINT_32 uiLength)
{
  ASSERT( m_pHead != NULL && m_pTail != NULL );
  ASSERT( uiLength > 0 && pCLink != NULL );

  CLink<TPData, TKey> *pCurr, *pLeft, *pRight;
  UINT_32 uiLeft, uiRight;
  UINT_32 i;

  if ( uiLength == 1 )
    {
      pCLink->SetNext(NULL);
      return pCLink;
    }

  uiLeft = uiLength/2;
  uiRight = uiLength - uiLeft;

  pCurr = pCLink;
  for ( i = 1; i < uiLeft; i++ )
    {
      ASSERT( pCurr != NULL );
      pCurr = pCurr->GetNext();
    }

  pLeft = pCLink;
  pRight = pCurr->GetNext();
  ASSERT( pRight != NULL );
  pCurr->SetNext(NULL);
  
  return Merge(Divide(pLeft, uiLeft), Divide(pRight, uiRight));
}

template <class TPData, class TKey>
inline CLink<TPData, TKey> *CList<TPData, TKey>::Merge(CLink<TPData, TKey> *pLeft, CLink<TPData, TKey> *pRight)
{
  ASSERT( m_pHead != NULL && m_pTail != NULL );
  ASSERT( pLeft != NULL && pRight != NULL );

  CLink<TPData, TKey> *pHead, *pTail;
  
//  if ( pLeft->KeyCompare(pRight->GetKeyInteger()) < 0 )
  if ( pLeft->GetKey() < pRight->GetKey() )
    {
      pHead = pLeft;
      pLeft = pLeft->GetNext();
    }
  else
    {
      pHead = pRight;
      pRight = pRight->GetNext();
    }

  pTail = pHead;

  while ( pLeft != NULL && pRight != NULL )
    {
      if ( pLeft->GetKey() < pRight->GetKey() )
	{
	  MergeAdd(&pTail, &pLeft);
	}
      else
	{
	  MergeAdd(&pTail, &pRight);
	}
    }

  if ( pLeft != NULL )
    {
      ASSERT( pRight == NULL );
      MergeAdd(&pTail, &pLeft);
    }
  if ( pRight != NULL )
    {
      ASSERT( pLeft == NULL );
      MergeAdd(&pTail, &pRight);
    }

  return pHead;
}

template <class TPData, class TKey>
inline VOID CList<TPData, TKey>::MergeAdd(CLink<TPData, TKey>** ppTail, CLink<TPData, TKey>** ppCLink)
{
  ASSERT( m_pHead != NULL && m_pTail != NULL );
  ASSERT( ppTail != NULL && ppCLink != NULL );
  ASSERT( *ppTail != NULL && *ppCLink != NULL );
  
  (*ppTail)->SetNext(*ppCLink);
  *ppTail = (*ppTail)->GetNext();
  *ppCLink = (*ppCLink)->GetNext();
}

#endif // SORT_ENABLED

//////////////////////////////////////////////////////////////////////////////
// CListIter Class
//

template <class TPData, class TKey>
inline CListIter<TPData, TKey>::CListIter()
{
  m_pCList = NULL;
  m_pCurr = NULL;
}

template <class TPData, class TKey>
inline CListIter<TPData, TKey>::~CListIter()
{
  Uninit();
}

template <class TPData, class TKey>
inline BOOL CListIter<TPData, TKey>::Init(CList<TPData, TKey> *pCList)
{
  ASSERT( pCList != NULL );

  Uninit();

  m_pCList = pCList;
  m_pCurr = NULL;

  return TRUE;
}

template <class TPData, class TKey>
inline VOID CListIter<TPData, TKey>::Uninit()
{
  m_pCList = NULL;
  m_pCurr = NULL;
}

template <class TPData, class TKey>
inline BOOL CListIter<TPData, TKey>::Valid()
{
  return (m_pCList != NULL);
}

template <class TPData, class TKey>
inline TPData CListIter<TPData, TKey>::PeekForward(TKey key)
{
  ASSERT( Valid() ); 

  TPData pData;

  pData = NULL;

  // Find the correct position in the CList for the CLink
  for ( m_pCurr = m_pCList->m_pHead->GetNext(); m_pCurr != m_pCList->m_pTail; 
	m_pCurr = m_pCurr->GetNext() )
    {
      if ( m_pCurr->GetKey() < key )
	{
	  // Not there yet
	  continue;
	}
      else if ( m_pCurr->GetKey() == key )
	{
	  // We found it

	  // Pull out the Data from the CLink
	  pData = m_pCurr->GetData(NULL);
	  
	  break;
	}
      else
	{
	  // We passed it
	  break;
	}
    }

  return pData;
}

template <class TPData, class TKey>
inline TPData CListIter<TPData, TKey>::PeekBackward(TKey key)
{
  ASSERT( Valid() ); 

  TPData pData;

  pData = NULL;

  // Find the correct position in the CList for the CLink
  for ( m_pCurr = m_pCList->m_pTail->GetPrev(); m_pCurr != m_pCList->m_pHead; 
	m_pCurr = m_pCurr->GetPrev() )
    {
      if ( m_pCurr->GetKey() > key )
	{
	  // Not there yet
	  continue;
	}
      else if ( m_pCurr->GetKey() == key )
	{
	  // We found it

	  // Pull out the Data from the CLink
	  pData = m_pCurr->GetData(NULL);
	  
	  break;
	}
      else
	{
	  // We passed it
	  break;
	}
    }

  return pData;
}

template <class TPData, class TKey>
inline TPData CListIter<TPData, TKey>::PeekFirst(TKey *pTKey/* = NULL */)
{
  ASSERT( Valid() ); 

  m_pCurr = m_pCList->m_pHead->GetNext();
  if ( m_pCurr == m_pCList->m_pTail )
    {
      m_pCurr = NULL;
      return NULL;
    }

  return m_pCurr->GetData(pTKey);
}

template <class TPData, class TKey>
inline TPData CListIter<TPData, TKey>::PeekNext(TKey *pTKey/* = NULL*/)
{
  ASSERT( Valid() ); 

  if ( m_pCurr == NULL )
    {
      return NULL;
    }

  m_pCurr = m_pCurr->GetNext();
  if ( m_pCurr == m_pCList->m_pTail )
    {
      m_pCurr = NULL;
      return NULL;
    }

  return m_pCurr->GetData(pTKey);
}

template <class TPData, class TKey>
inline TPData CListIter<TPData, TKey>::PeekLast(TKey *pTKey/* = NULL */)
{
  ASSERT( Valid() );

  m_pCurr = m_pCList->m_pTail->GetPrev();
  if ( m_pCurr == m_pCList->m_pHead )
    {
      m_pCurr = NULL;
      return NULL;
    }

  return m_pCurr->GetData(pTKey);
}

template <class TPData, class TKey>
inline TPData CListIter<TPData, TKey>::PeekPrev(TKey *pTKey/* = NULL*/)
{
  ASSERT( Valid() );

  if ( m_pCurr == NULL )
    {
      return NULL;
    }

  m_pCurr = m_pCurr->GetPrev();
  if ( m_pCurr == m_pCList->m_pHead )
    {
      m_pCurr = NULL;
      return NULL;
    }

  return m_pCurr->GetData(pTKey);
}

template <class TPData, class TKey>
inline TPData CListIter<TPData, TKey>::PeekNextRemoveCurr(TPData *ppDataRemoved/* = NULL*/,
							  TKey *pTKeyRemoved/* = NULL*/,
							  TKey *pTKey/* = NULL*/)
{
  ASSERT( Valid() ); 

  CLink<TPData, TKey> *pCLink;

  if ( m_pCurr == NULL )
    {
      return NULL;
    }

  // Advance the current pointer
  m_pCurr = m_pCurr->GetNext();

  // Remove the CLink at the end of the CList
  pCLink = m_pCList->RemovePrev(m_pCurr);
  if ( pCLink == NULL )
    {
      return NULL;
    }

  // Pull out the Data from the CLink if desired
  if ( ppDataRemoved != NULL )
    {
      *ppDataRemoved = pCLink->GetData(pTKeyRemoved);
    }

  // Deallocate the CLink
  m_pCList->FreeLink(pCLink);

  // Check if this is the end of the list iteration
  if ( m_pCurr == m_pCList->m_pTail )
    {
      m_pCurr = NULL;
      return NULL;
    }

  return m_pCurr->GetData(pTKey);
}

template <class TPData, class TKey>
inline TPData CListIter<TPData, TKey>::PeekPrevRemoveCurr(TPData *ppDataRemoved/* = NULL*/,
							  TKey *pTKeyRemoved/* = NULL*/,
							  TKey *pTKey/* = NULL*/)
{
  ASSERT( Valid() ); 

  CLink<TPData, TKey> *pCLink;

  if ( m_pCurr == NULL )
    {
      return NULL;
    }

  // Advance the current pointer
  m_pCurr = m_pCurr->GetPrev();

  // Remove the CLink at the end of the CList
  pCLink = m_pCList->RemoveNext(m_pCurr);
  if ( pCLink == NULL )
    {
      return NULL;
    }

  // Pull out the Data from the CLink if desired
  if ( ppDataRemoved != NULL )
    {
      *ppDataRemoved = pCLink->GetData(pTKeyRemoved);
    }

  // Deallocate the CLink
  m_pCList->FreeLink(pCLink);

  // Check if this is the end of the list iteration
  if ( m_pCurr == m_pCList->m_pHead )
    {
      m_pCurr = NULL;
      return NULL;
    }

  return m_pCurr->GetData(pTKey);
}

template <class TPData, class TKey>
inline TPData CListIter<TPData, TKey>::InsertNext(TPData pData, TKey key)
{
  ASSERT( Valid() );

  CLink<TPData, TKey> *pCLink;

  // Allocate and Initialize a CLink
  pCLink = m_pCList->AllocateLink();
  ASSERT( pCLink != NULL );
  pCLink->Init(key, pData);

  // Insert the CLink to the right of the current link
  m_pCList->InsertNext(m_pCurr, pCLink);
  
  return pData;
}

template <class TPData, class TKey>
inline TPData CListIter<TPData, TKey>::InsertPrev(TPData pData, TKey key)
{
  ASSERT( Valid() );

  CLink<TPData, TKey> *pCLink;

  // Allocate and Initialize a CLink
  pCLink = m_pCList->AllocateLink();
  ASSERT( pCLink != NULL );
  pCLink->Init(key, pData);

  // Insert the CLink to the right of the current link
  m_pCList->InsertPrev(m_pCurr, pCLink);
  
  return pData;
}

template <class TPData, class TKey>
inline TPData CListIter<TPData, TKey>::PeekNextWrap(TKey *pTKey/* = NULL*/)
{
  ASSERT( Valid() ); 

  if ( m_pCurr == NULL )
    {
      return NULL;
    }

  m_pCurr = m_pCurr->GetNext();
  if ( m_pCurr == m_pCList->m_pTail )
    {
      m_pCurr = m_pCurr->GetNext();
      if ( m_pCurr == m_pCList->m_pTail )
	{
	  m_pCurr = NULL;
	  return NULL;
	}
    }

  return m_pCurr->GetData(pTKey);
}

template <class TPData, class TKey>
inline TPData CListIter<TPData, TKey>::PeekPrevWrap(TKey *pTKey/* = NULL*/)
{
  ASSERT( Valid() ); 

  if ( m_pCurr == NULL )
    {
      return NULL;
    }

  m_pCurr = m_pCurr->GetPrev();
  if ( m_pCurr == m_pCList->m_pHead )
    {
      m_pCurr = m_pCurr->GetPrev();
      if ( m_pCurr == m_pCList->m_pHead )
	{
	  m_pCurr = NULL;
	  return NULL;
	}
    }

  return m_pCurr->GetData(pTKey);
}

template <class TPData, class TKey>
inline TPData CListIter<TPData, TKey>::PeekNextWrapRemoveCurr(TPData *ppDataRemoved/* = NULL*/,
							  TKey *pTKeyRemoved/* = NULL*/,
							  TKey *pTKey/* = NULL*/)
{
  ASSERT( Valid() ); 

  CLink<TPData, TKey> *pCLink;

  if ( m_pCurr == NULL )
    {
      return NULL;
    }

  // Advance the current pointer
  m_pCurr = m_pCurr->GetNext();

  // Remove the CLink at the end of the CList
  pCLink = m_pCList->RemovePrev(m_pCurr);
  if ( pCLink == NULL )
    {
      return NULL;
    }

  // Pull out the Data from the CLink if desired
  if ( ppDataRemoved != NULL )
    {
      *ppDataRemoved = pCLink->GetData(pTKeyRemoved);
    }

  // Deallocate the CLink
  m_pCList->FreeLink(pCLink);

  // Check if this is the end of the list iteration
  if ( m_pCurr == m_pCList->m_pTail )
    {
      m_pCurr = m_pCurr->GetNext();
      if ( m_pCurr == m_pCList->m_pTail )
	{
	  m_pCurr = NULL;
	  return NULL;
	}
    }

  return m_pCurr->GetData(pTKey);
}

template <class TPData, class TKey>
inline TPData CListIter<TPData, TKey>::PeekPrevWrapRemoveCurr(TPData *ppDataRemoved/* = NULL*/,
							  TKey *pTKeyRemoved/* = NULL*/,
							  TKey *pTKey/* = NULL*/)
{
  ASSERT( Valid() ); 

  CLink<TPData, TKey> *pCLink;

  if ( m_pCurr == NULL )
    {
      return NULL;
    }

  // Advance the current pointer
  m_pCurr = m_pCurr->GetPrev();

  // Remove the CLink at the end of the CList
  pCLink = m_pCList->RemoveNext(m_pCurr);
  if ( pCLink == NULL )
    {
      return NULL;
    }

  // Pull out the Data from the CLink if desired
  if ( ppDataRemoved != NULL )
    {
      *ppDataRemoved = pCLink->GetData(pTKeyRemoved);
    }

  // Deallocate the CLink
  m_pCList->FreeLink(pCLink);

  // Check if this is the end of the list iteration
  if ( m_pCurr == m_pCList->m_pHead )
    {
      m_pCurr = m_pCurr->GetPrev();
      if ( m_pCurr == m_pCList->m_pHead )
	{
	  m_pCurr = NULL;
	  return NULL;
	}
    }

  return m_pCurr->GetData(pTKey);
}

#endif // __LIST_H__








