#ifndef __LPSOLVER_H__
#define __LPSOLVER_H__

#include "lpvector.h"
#define __INTEGRATED_WITH_CDA__
#ifdef __INTEGRATED_WITH_CDA__
#include "slideprimitive.h"
#endif // __INTEGRATED_WITH_CDA__

//////////////////////////////////////////////////////////////////////
// LPSolver Class
//

class LPSolver
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  LPSolver();
  ~LPSolver();
  VOID Init();
  VOID Uninit();

  BOOL Initialized();
  BOOL Valid();

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  VOID Run(); // Simple for Debugging
#ifdef __INTEGRATED_WITH_CDA__
  BOOL SolveCDAIntersection(CListInt *plistCDAFace0, 
			    CListInt *plistCDAFace1, 
			    Vector &vSolution);
#endif // __INTEGRATED_WITH_CDA__

protected:
  BOOL BoundedLP(UINT_32 uiDim, UINT_32 uiHalfSpaces, 
		 const LPVector &vC, LPVector &vSolution);
  BOOL UnboundedLP(UINT_32 uiDim, UINT_32 uiHalfSpaces, 
		   const LPVector &vC, LPVector &vSolution);

protected:
  VOID RandomPermutation(UINT_32 uiSize, UINT_32 *puiPermutation);
  VOID AllocateWorkSpace(UINT_32 uiDim, UINT_32 uiSize);
  VOID DeallocateWorkSpace();

//////////////////////////////////////////////////
// Output Functions
//
public:
  VOID Display();

//////////////////////////////////////////////////
// Accessor Functions
//
public:

protected:
  inline UINT_32 GetIndex(UINT_32 uiDim, UINT_32 uiHalfSpace);
  inline FLOAT *GetValues(UINT_32 uiDim, UINT_32 uiHalfSpace);
  inline FLOAT *GetC(UINT_32 uiDim);
  inline VOID Swap(UINT_32 uiDim, UINT_32 uiHalfSpace0, 
			     UINT_32 uiHalfSpace1);

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  UINT_32 m_uiDimMax;
  UINT_32 m_uiHalfSpacesMax;

  UINT_32 m_uiWorkSpaceSize;
  FLOAT *m_pfWorkSpace;
  UINT_32 m_uiSwapSpaceSize;
  FLOAT *m_pfSwapSpace;
  FLOAT *m_pfCSpace;
  UINT_32 m_uiDimSpaceSize;
  FLOAT *m_pfNormalNeg;
  FLOAT *m_pfSolution;
  UINT_32 m_uiHalfSpacesSize;
  UINT_32 *m_puiRandomPermutation;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline UINT_32 LPSolver::GetIndex(UINT_32 uiDim, UINT_32 uiHalfSpace)
{
  ASSERT( 0 < uiDim && uiDim <= m_uiDimMax );
  ASSERT( uiHalfSpace < m_uiHalfSpacesMax );

  return ((uiDim-1) + (uiHalfSpace*m_uiDimMax))*(m_uiDimMax+1);
}

inline FLOAT *LPSolver::GetValues(UINT_32 uiDim, UINT_32 uiHalfSpace)
{
  return &(m_pfWorkSpace[GetIndex(uiDim, uiHalfSpace)]);
}

inline FLOAT *LPSolver::GetC(UINT_32 uiDim)
{
  ASSERT( uiDim > 0 );

  return &(m_pfCSpace[(uiDim-1)*(m_uiDimMax+1)]);  
}

inline VOID LPSolver::Swap(UINT_32 uiDim, UINT_32 uiHalfSpace0, 
			   UINT_32 uiHalfSpace1)
{
  if ( uiHalfSpace0 == uiHalfSpace1 )
    {
      return;
    }

  memcpy(m_pfSwapSpace, 
	 GetValues(uiDim, uiHalfSpace0),
	 (m_uiDimMax-uiDim+1)*(m_uiDimMax+1)*sizeof(FLOAT));
  memcpy(GetValues(uiDim, uiHalfSpace0),
	 GetValues(uiDim, uiHalfSpace1),
	 (m_uiDimMax-uiDim+1)*(m_uiDimMax+1)*sizeof(FLOAT));
  memcpy(GetValues(uiDim, uiHalfSpace1),
	 m_pfSwapSpace, 
	 (m_uiDimMax-uiDim+1)*(m_uiDimMax+1)*sizeof(FLOAT));
}

#endif // __LPSOLVER_H__
