#include "window.h"
#include "slideparser.h"
#include "slideworld.h"
#include "slide.h"

#include <time.h>
#include "itcl.h"
#include "itk.h"

EXTERN int		TkConsoleInit(Tcl_Interp *interp);
extern BOOL gbPreprocessDone;
extern BOOL gbCheckSemanticsDone;

//////////////////////////////////////////////////////////////////////
// Tcl/Tk Stuff
//

#ifdef WIN32

#define matherr _matherr
#else

// The following variable is a special hack that is needed in order for
// Sun shared libraries to be used for Tcl.
extern "C"
{
extern int matherr();
};

#endif //WIN32

int *tclDummyMathPtr = (int *) matherr;



// Called by Tk_Main() to let me init the modules (Togl) I will need.
int Tcl_AppInit(Tcl_Interp *pInterp);
VOID InitSLIDETCLGlobals(Tcl_Interp *pInterp);

//
// Tcl/Tk Stuff
//////////////////////////////////////////////////////////////////////

BOOL ParseCommandLine(UINT_32 &uiArgc, CHAR **ppcArgv);
VOID ConsumeArgument(UINT_32 *puiArgc, CHAR **ppcArgv, 
		     UINT_32 i);
VOID ConsumeArguments(UINT_32 *puiArgc, CHAR **ppcArgv, 
		      UINT_32 i, UINT_32 uiArgs);
VOID SkipArgument(UINT_32 *pi, CHAR ***pppc);
VOID PrintUsage();
BOOL LoadWorld();
BOOL LoadWorldGUI();
VOID UnloadWorld();

VOID OpenGUI(Tcl_Interp *pInterp);
VOID ClearGUI(Tcl_Interp *pInterp);
VOID CloseGUI(Tcl_Interp *pInterp);

// Main Design Studio Call Backs
int HandleMode(    ClientData clientData, Tcl_Interp *interp, int argc, char **argv);
int HandleOpen(    ClientData clientData, Tcl_Interp *interp, int argc, char **argv);

int HandleNew(     ClientData clientData, Tcl_Interp *interp, int argc, char **argv);
int HandleClose(   ClientData clientData, Tcl_Interp *interp, int argc, char **argv);
int HandleExit(    ClientData clientData, Tcl_Interp *interp, int argc, char **argv);

int HandleWrite(   ClientData clientData, Tcl_Interp *interp, int argc, char **argv);
int HandleWriteDOT(ClientData clientData, Tcl_Interp *interp, int argc, char **argv);
int HandleWriteRIB(ClientData clientData, Tcl_Interp *interp, int argc, char **argv);
int HandleWritePOV(ClientData clientData, Tcl_Interp *interp, int argc, char **argv);
int HandleWriteWRLZ(ClientData clientData, Tcl_Interp *interp, int argc, char **argv);
int HandleWriteSTL(ClientData clientData, Tcl_Interp *interp, int argc, char **argv);
int HandleWriteSIF(ClientData clientData, Tcl_Interp *interp, int argc, char **argv);
int HandleWriteLSIF(ClientData clientData, Tcl_Interp *interp, int argc, char **argv);

Tcl_Command gtclcmdHandleMode = NULL;
Tcl_Command gtclcmdHandleOpen = NULL;
Tcl_Command gtclcmdHandleNew = NULL;
Tcl_Command gtclcmdHandleClose = NULL;
Tcl_Command gtclcmdHandleExit = NULL;
Tcl_Command gtclcmdHandleWrite = NULL;
Tcl_Command gtclcmdHandleWriteDOT = NULL;
Tcl_Command gtclcmdHandleWriteRIB = NULL;
Tcl_Command gtclcmdHandleWritePOV = NULL;
Tcl_Command gtclcmdHandleWriteWRLZ = NULL;
Tcl_Command gtclcmdHandleWriteSTL = NULL;
Tcl_Command gtclcmdHandleWriteSIF = NULL;
Tcl_Command gtclcmdHandleWriteLSIF = NULL;

enum Mode { VIEWER, EDITOR };
Mode UIMode = EDITOR;

static CHAR gpcBuffer[200];
static CHAR *gpcProgramName = NULL;
static CHAR *gpcInputFile = NULL;
static CSLIDEParser *gpCSLIDEParser = NULL;

CHAR *GetFileDirPath (CHAR *pcFullPath)
{
  INT_32 iLastslash;
  iLastslash = -1;
  CHAR *temp;

  // Grab the directory name out of this fully-qualified filename
  // Is there a premade function to do this?  maybe...
  // 
  for (UINT_32 i=0; i< strlen (pcFullPath); i++)
  {
	  if (pcFullPath[i] == '/')
	  {
		  iLastslash = i;
	  }
  }
  if (iLastslash >= 0)
  {
	  // if not in current directory, CD there to be in the right
	  // directory.
	  //
	  temp = new char[iLastslash + 3];
	  strncpy ( temp, pcFullPath, iLastslash+1);
	  temp[iLastslash+1] = '\0';
  }
  else
  {
	  temp = new char[1];
	  temp[0] = '\0';
  }

  return temp;
}
  
const UINT_32 cuiSearchPathDirectoriesMax = 20;
CHAR *gppcSearchPathDirectories[cuiSearchPathDirectoriesMax];
UINT_32 guiSearchPathDirectories = 0;

int main(int argc, char **argv)
{
  UINT_32 uiArgc;
  CHAR **ppcArgv;


  // Make a copy of argc and argv so that we can take our 
  // arguments from it and modify it
  uiArgc = (UINT_32)argc;
  ppcArgv = new CHAR*[uiArgc];
  ASSERT( ppcArgv != NULL );
  memcpy(ppcArgv, argv, uiArgc*sizeof(CHAR*));

  // Parse the command line to find the input file name
  if ( !ParseCommandLine(uiArgc, ppcArgv) )
    {
      PrintUsage();
      exit(1);
    }

  // This calls the event loop handler and only returns once the 
  // program is exited
  Tk_Main( (int)uiArgc, (char**)ppcArgv, Tcl_AppInit );

  return 0;
}

BOOL ParseCommandLine(UINT_32 &uiArgc, CHAR **ppcArgv)
{
  BOOL bCorrect;
  UINT_32 i, uiArgsMatched;
  CHAR **ppc;
  CHAR *pc;

  ASSERT( uiArgc > 0 );
  ASSERT( ppcArgv != NULL );

  bCorrect = TRUE;
  i = 0;
  ppc = ppcArgv;

  gpcProgramName = new CHAR[strlen(ppcArgv[i])+1];
  ASSERT( gpcProgramName != NULL );
  strcpy(gpcProgramName, ppcArgv[i]);
  //ConsumeArguments(&uiArgc, ppcArgv, 0, 1);
  i++;
  
  while ( i < uiArgc )
    {
      if ( *ppc[i] == '-' )
	{
	  pc = (ppc[i]) + 1;

	  switch ( *pc )
	    {
	    case 'I':
	      pc++;
	      // Search Path Directory

	      if ( i+1 >= uiArgc )
		{
		  fprintf(stderr, "Missing argument for option -I\n");
		  return FALSE;
		}
	      
	      pc = ppc[i+1];
	      gppcSearchPathDirectories[guiSearchPathDirectories] = new CHAR[strlen(pc) + 1];
	      ASSERT( gppcSearchPathDirectories[guiSearchPathDirectories] != NULL );
	      strcpy(gppcSearchPathDirectories[guiSearchPathDirectories], pc);
	      guiSearchPathDirectories++;
	      uiArgsMatched = 2;
	      break;

	    default:
	      fprintf(stderr, "Unknown option: %s\n", ppc[i]);
	      return FALSE;
	    }

	  ASSERT( uiArgsMatched != 0 );
	  //if ( uiArgsMatched == 0 )
	  //  {
	  //    bCorrect = FALSE;
	  //    SkipArgument(&i, &ppc);
	  //  }
	  //else
	  //  {
	  ConsumeArguments(&uiArgc, ppcArgv, i, uiArgsMatched);
	  //  }
	}
      else
	{
	  break;
	}
    }

  if ( i == uiArgc )
    {
      return TRUE;
    }

  if ( !( i+1 == uiArgc) )
    {
      return FALSE;
    }

  gpcInputFile = new CHAR[strlen(ppcArgv[i])+1];
  ASSERT( gpcInputFile != NULL );
  strcpy(gpcInputFile, ppcArgv[i]);

  ConsumeArguments(&uiArgc, ppcArgv, i, 1);
	
  return bCorrect;
}

VOID ConsumeArgument(UINT_32 *puiArgc, CHAR **ppcArgv, UINT_32 i)
{
  ASSERT( puiArgc != NULL );
  ASSERT( ppcArgv != NULL );
  ASSERT( i+1 <= *puiArgc );

  memcpy((ppcArgv + i), (ppcArgv + i + 1), 
	 (*puiArgc - i - 1)*sizeof(CHAR*));
  (*puiArgc)--;
}

VOID ConsumeArguments(UINT_32 *puiArgc, CHAR **ppcArgv, 
		      UINT_32 i, UINT_32 uiArgs)
{
  ASSERT( puiArgc != NULL );
  ASSERT( ppcArgv != NULL );
  ASSERT( uiArgs > 0 );
  ASSERT( i+uiArgs <= *puiArgc );

  memcpy((ppcArgv + i), (ppcArgv + i + uiArgs), 
	 (*puiArgc - i - uiArgs)*sizeof(CHAR*));
  (*puiArgc) -= uiArgs;
}

VOID SkipArgument(UINT_32 *pi, CHAR ***pppc)
{
  ASSERT( pi != NULL );
  ASSERT( pppc != NULL );

  (*pi)++;
  (*pppc)++;
}
VOID PrintUsage()
{
  fprintf(stderr, "Usage: %s [-I search_path_directory]* input.slf\n", gpcProgramName);
}

BOOL LoadWorld()
{
  CSLIDEWorld *pCSLIDEWorld;

  UnloadWorld();

  ASSERT( gpcInputFile != NULL );
  gpCSLIDEParser = slide_InitParser(new CSLIDEParser, gpcInputFile,
				    (guiSearchPathDirectories == 0) ? ((CHAR**)NULL) : (gppcSearchPathDirectories), 
				    guiSearchPathDirectories);
  ASSERT( gpCSLIDEParser != NULL );

  fprintf(stderr, "Parsing...\n");
  pCSLIDEWorld = (CSLIDEWorld*)(gpCSLIDEParser->Parse());
  CSLIDEWorld::SetWorld((CSLIDEWorld*)(gpCSLIDEParser->GetAST()));
  if ( pCSLIDEWorld == NULL )
    {
      fprintf(stderr, "Compilation halted because of syntax errors.\n");
      return FALSE;
    }

  fprintf(stderr, "Checking Semantics...\n");
  if ( !pCSLIDEWorld->CheckSemantics() )
    {
      fprintf(stderr, "Semantic Checking Done: %ld Errors and %ld Warnings.\n",
	      pCSLIDEWorld->GetErrorCount(), 
	      pCSLIDEWorld->GetWarningCount());
      fprintf(stderr, "Compilation halted because of semantic errors.\n");
      return FALSE;
    }
  fprintf(stderr, "Semantic Checking Done: %ld Errors and %ld Warnings.\n",
	  pCSLIDEWorld->GetErrorCount(), pCSLIDEWorld->GetWarningCount());
  fprintf(stderr, "Compilation was successful.\n");

  CSLIDEWorld::GetWorld()->Preprocess();

  return TRUE;
}

VOID UnloadWorld()
{
  if ( CSLIDEWorld::GetWorld() != NULL )
    {
      CSLIDEWorld::GetWorld()->UninitTcl();
    }

  // Flush the Event Queue
  while ( Tk_DoOneEvent(TK_ALL_EVENTS|TK_DONT_WAIT) )
    {
    }

  if ( CSLIDEWorld::GetWorld() != NULL )
    {
      delete CSLIDEWorld::GetWorld();
      CSLIDEWorld::SetWorld(NULL);
    }

  if ( gpCSLIDEParser != NULL )
    {
      delete gpCSLIDEParser;
      gpCSLIDEParser = NULL;
    }
}

BOOL LoadWorldGUI()
{
  CSLIDEWorld *pCSLIDEWorld;

  UnloadWorld();

  ASSERT( gpcInputFile != NULL );
  gpCSLIDEParser = slide_InitParser(new CSLIDEParser, gpcInputFile,
				    (guiSearchPathDirectories == 0) ? ((CHAR**)NULL) : (gppcSearchPathDirectories), 
				    guiSearchPathDirectories);
  ASSERT( gpCSLIDEParser != NULL );

  fprintf(stderr, "Parsing...\n");
  pCSLIDEWorld = (CSLIDEWorld*)(gpCSLIDEParser->Parse());
  CSLIDEWorld::SetWorld((CSLIDEWorld*)(gpCSLIDEParser->GetAST()));
  if ( pCSLIDEWorld == NULL )
    {
      fprintf(stderr, "Compilation halted because of syntax errors.\n");
      return FALSE;
    }

  fprintf(stderr, "Checking Semantics...\n");
  if ( !pCSLIDEWorld->CheckSemantics() )
    {
      fprintf(stderr, "Semantic Checking Done: %ld Errors and %ld Warnings.\n",
	      pCSLIDEWorld->GetErrorCount(), 
	      pCSLIDEWorld->GetWarningCount());
      fprintf(stderr, "Compilation halted because of semantic errors.\n");
      return FALSE;
    }
  fprintf(stderr, "Semantic Checking Done: %ld Errors and %ld Warnings.\n",
	  pCSLIDEWorld->GetErrorCount(), pCSLIDEWorld->GetWarningCount());
  fprintf(stderr, "Compilation was successful.\n");

  CSLIDEWorld::GetWorld()->PreprocessGUI();

  return TRUE;
}

int Tcl_AppInit(Tcl_Interp *pInterp)
{
  INT_32 iCode;

  CSLIDEWorld::SetInterp(pInterp);

  // Initialize Tcl, Tk, and Oot.
  if (Tcl_Init(pInterp) == TCL_ERROR)
    {
      return TCL_ERROR;
    }
  if (Tk_Init(pInterp) == TCL_ERROR) 
    {
      return TCL_ERROR;
    }
  if (Itcl_Init(pInterp) == TCL_ERROR) 
    {
      return TCL_ERROR;
    }
  if (Itk_Init(pInterp) == TCL_ERROR) 
    {
      return TCL_ERROR;
    }
  Tcl_StaticPackage(pInterp, "Itcl", Itcl_Init, Itcl_SafeInit);
  Tcl_StaticPackage(pInterp, "Itk", Itk_Init, (Tcl_PackageInitProc *) NULL);
  
  /*
   *  This is itkwish, so import all [incr Tcl] commands by
   *  default into the global namespace.  Fix up the autoloader
   *  to do the same.
   */
  if (Tcl_Import(pInterp, Tcl_GetGlobalNamespace(pInterp),
		 "::itk::*", /* allowOverwrite */ 1) != TCL_OK) {
    return TCL_ERROR;
  }
  
  if (Tcl_Import(pInterp, Tcl_GetGlobalNamespace(pInterp),
		 "::itcl::*", /* allowOverwrite */ 1) != TCL_OK) {
    return TCL_ERROR;
  }

  if (Tcl_Eval(pInterp, "auto_mkindex_parser::slavehook { _%@namespace import -force ::itcl::* ::itk::* }") != TCL_OK) {
    return TCL_ERROR;
  }
  
  if (Togl_Init(pInterp) == TCL_ERROR) 
    {
      return TCL_ERROR;
    }

#ifdef WIN32
  if (TkConsoleInit(pInterp) == TCL_ERROR) 
  {
    return TCL_ERROR;
  }
#endif // WIN32

  if (Oot_Init(pInterp) == TCL_ERROR) 
    {
      return TCL_ERROR;
    }

  if ( CWindow::CreateOotClass(pInterp) == TCL_ERROR )
    {
       return TCL_ERROR;
    }
  Togl_CreateFunc( CWindow::CBHandleToglCreate );
  Togl_DisplayFunc( CWindow::CBHandleToglDisplay );
  Togl_ReshapeFunc( CWindow::CBHandleToglReshape );
  Togl_DestroyFunc( CWindow::CBHandleToglDestroy );


  if ( CSLIDEParser::CreateOotClass(pInterp) == TCL_ERROR )
    {
       return TCL_ERROR;
    }

  InitSLIDETCLGlobals(pInterp);

  strcpy(gpcBuffer, "slideMode");
  gtclcmdHandleMode = Tcl_CreateCommand(pInterp, gpcBuffer, HandleMode, NULL, NULL);
  if ( gtclcmdHandleMode == NULL )
    {
      return TCL_ERROR;
    }
  strcpy(gpcBuffer, "slideOpen");
  gtclcmdHandleOpen = Tcl_CreateCommand(pInterp, gpcBuffer, HandleOpen, NULL, NULL);
  if ( gtclcmdHandleOpen == NULL )
    {
      return TCL_ERROR;
    }
  strcpy(gpcBuffer, "slideNew");
  gtclcmdHandleNew = Tcl_CreateCommand(pInterp, gpcBuffer, HandleNew, NULL, NULL);
  if ( gtclcmdHandleNew == NULL )
    {
      return TCL_ERROR;
    }
  strcpy(gpcBuffer, "slideClose");
  gtclcmdHandleClose = Tcl_CreateCommand(pInterp, gpcBuffer, HandleClose, NULL, NULL);
  if ( gtclcmdHandleClose == NULL )
    {
      return TCL_ERROR;
    }
  strcpy(gpcBuffer, "slideExit");
  gtclcmdHandleExit = Tcl_CreateCommand(pInterp, gpcBuffer, HandleExit, NULL, NULL);
  if ( gtclcmdHandleExit == NULL )
    {
      return TCL_ERROR;
    }

  strcpy(gpcBuffer, "slideWrite");
  gtclcmdHandleWrite = Tcl_CreateCommand(pInterp, gpcBuffer, HandleWrite, NULL, NULL);
  if ( gtclcmdHandleWrite == NULL )
    {
      return TCL_ERROR;
    }
  strcpy(gpcBuffer, "slideWriteDOT");
  gtclcmdHandleWriteDOT = Tcl_CreateCommand(pInterp, gpcBuffer, HandleWriteDOT, NULL, NULL);
  if ( gtclcmdHandleWriteDOT == NULL )
    {
      return TCL_ERROR;
    }
  strcpy(gpcBuffer, "slideWriteRIB");
  gtclcmdHandleWriteRIB = Tcl_CreateCommand(pInterp, gpcBuffer, HandleWriteRIB, NULL, NULL);
  if ( gtclcmdHandleWriteRIB == NULL )
    {
      return TCL_ERROR;
    }
  strcpy(gpcBuffer, "slideWritePOV");
  gtclcmdHandleWritePOV = Tcl_CreateCommand(pInterp, gpcBuffer, HandleWritePOV, NULL, NULL);
  if ( gtclcmdHandleWritePOV == NULL )
    {
      return TCL_ERROR;
    }
  strcpy(gpcBuffer, "slideWriteWRLZ");
  gtclcmdHandleWriteWRLZ = Tcl_CreateCommand(pInterp, gpcBuffer, HandleWriteWRLZ, NULL, NULL);
  if ( gtclcmdHandleWriteWRLZ == NULL )
    {
      return TCL_ERROR;
    }
  strcpy(gpcBuffer, "slideWriteSTL");
  gtclcmdHandleWriteSTL = Tcl_CreateCommand(pInterp, gpcBuffer, HandleWriteSTL, NULL, NULL);
  if ( gtclcmdHandleWriteSTL == NULL )
    {
      return TCL_ERROR;
    }
  strcpy(gpcBuffer, "slideWriteSIF");
  gtclcmdHandleWriteSIF = Tcl_CreateCommand(pInterp, gpcBuffer, HandleWriteSIF, NULL, NULL);
  if ( gtclcmdHandleWriteSIF == NULL )
    {
      return TCL_ERROR;
    }
  strcpy(gpcBuffer, "slideWriteLSIF");
  gtclcmdHandleWriteLSIF = Tcl_CreateCommand(pInterp, gpcBuffer, HandleWriteLSIF, NULL, NULL);
  if ( gtclcmdHandleWriteLSIF == NULL )
    {
      return TCL_ERROR;
    }

  //////////////////////////////////////////////////
  // START: Source main.tcl window.tcl gui.tcl
  sprintf(gpcBuffer, "if { [lsearch [array names env] SLIDE_LIBRARY] != -1 } {set auto_path [concat $env(SLIDE_LIBRARY) $auto_path]}");
  iCode = Tcl_Eval(pInterp, gpcBuffer); 
  if ( iCode != TCL_OK )
    {
      fprintf(stderr, "Tcl Error: \"%s\"\n", pInterp->result);
      ASSERT( FALSE );
    }

  sprintf(gpcBuffer, "package require slide");
  iCode = Tcl_Eval(pInterp, gpcBuffer); 
  if ( iCode != TCL_OK )
    {
      fprintf(stderr, "Tcl Error: \"%s\"\n", pInterp->result);
      ASSERT( FALSE );
    }
  // END  : Source main.tcl window.tcl gui.tcl
  //////////////////////////////////////////////////

  if ( gpcInputFile != NULL )
    {
      // Enable Reload
      sprintf(gpcBuffer, "Main:Reload");
      iCode = Tcl_Eval(pInterp, gpcBuffer); 
      ASSERT( iCode == TCL_OK );  

      // Get entire file path
      CHAR pcWorkingDirectory[200];
      CHAR* pcInputFileName;

      sprintf(gpcBuffer, "pwd");
      ASSERT( Tcl_Eval(pInterp, gpcBuffer) == TCL_OK );
      ASSERT( pInterp->result != NULL );
      sprintf(pcWorkingDirectory, "%s", pInterp->result);

#ifdef WIN32
  {
    pcInputFileName = gpcInputFile;      
    gpcInputFile = new CHAR[strlen(pcWorkingDirectory)+strlen(pcInputFileName)+1];
    ASSERT( gpcInputFile != NULL );

    sprintf(gpcInputFile, "%s\\\\%s", pcWorkingDirectory, pcInputFileName);

    delete pcInputFileName;
  }
#else // WIN32
  {

    // If the user gave a relative path, find the absolute path
    if( gpcInputFile[0] != '/' )
      {
	pcInputFileName = gpcInputFile;      
	gpcInputFile = new CHAR[strlen(pcWorkingDirectory)+strlen(pcInputFileName)+1];
	ASSERT( gpcInputFile != NULL );
	
	sprintf(gpcInputFile, "%s/%s", pcWorkingDirectory, pcInputFileName);

	delete pcInputFileName;
      }
  }
#endif // WIN32


      // Change directory to file directory before world is loaded
      CHAR *pDirname;
      pDirname = GetFileDirPath( gpcInputFile );
      sprintf(gpcBuffer, "cd %s", pDirname);
      iCode = Tcl_Eval(pInterp, gpcBuffer); 

      if( LoadWorld() )
	{
	  // Enable Main Menu commands
	  sprintf(gpcBuffer, "Main:Opened");
	  iCode = Tcl_Eval(pInterp, gpcBuffer); 
	  ASSERT( iCode == TCL_OK );
	}
    }

/*
  // Load the world from the file into memory
  if ( !LoadWorld() )
    {
      exit(2);
    }

  // This should Allocate all the windows for the application
  CSLIDEWorld::GetWorld()->Preprocess();
*/
  return TCL_OK;
}

VOID InitSLIDETCLGlobals(Tcl_Interp *pInterp)
{
  UINT_32 i;

  ASSERT( pInterp != NULL );

  // Set up SLIDE Globals like SLF_TIME, SLF_MOUSE_X, SLF_MOUSE_Y
  Tcl_SetVar(pInterp, "SLF_PI", "3.1415927", TCL_GLOBAL_ONLY);
  Tcl_SetVar(pInterp, "SLF_TIME", "0", TCL_GLOBAL_ONLY);
  Tcl_SetVar(pInterp, "SLF_FRAME", "0", TCL_GLOBAL_ONLY);
  Tcl_SetVar(pInterp, "SLF_START_FRAME", "0", TCL_GLOBAL_ONLY);
  Tcl_SetVar(pInterp, "SLF_END_FRAME", "0", TCL_GLOBAL_ONLY);

  // On Off flags
  for ( i = 0; i < slideoftSize; i++ )
    {
      sprintf(gpcBuffer, "%ld", i);
      Tcl_SetVar(pInterp, gcppcSLIDEOnOffReservedWords[i], 
		 gpcBuffer, TCL_GLOBAL_ONLY);
    }

  // Solid flags
  for ( i = 0; i < slidestSize; i++ )
    {
      sprintf(gpcBuffer, "%ld", i);
      Tcl_SetVar(pInterp, gcppcSLIDESolidReservedWords[i], 
		 gpcBuffer, TCL_GLOBAL_ONLY);
    }

  // Shading flags
  gcppcSLIDEShadingReservedWords[slideshtSLF_INHERIT] = 
    gcppcSLIDEReservedWords[sliderwtSLF_INHERIT];
  for ( i = 0; i < slideshtSize; i++ )
    {
      sprintf(gpcBuffer, "%ld", i);
      Tcl_SetVar(pInterp, gcppcSLIDEShadingReservedWords[i], 
		 gpcBuffer, TCL_GLOBAL_ONLY);
    }

  // LOD flags
  gcppcSLIDELODReservedWords[slidelodtSLF_OFF] = 
    gcppcSLIDEOnOffReservedWords[slideoftSLF_OFF];
  for ( i = 0; i < slidelodtSize; i++ )
    {
      sprintf(gpcBuffer, "%ld", i);
      Tcl_SetVar(pInterp, gcppcSLIDELODReservedWords[i], 
		 gpcBuffer, TCL_GLOBAL_ONLY);
    }

  // Projection flags
  for ( i = 0; i < slidepjtSize; i++ )
    {
      sprintf(gpcBuffer, "%ld", i);
      Tcl_SetVar(pInterp, gcppcSLIDEProjectionReservedWords[i], 
		 gpcBuffer, TCL_GLOBAL_ONLY);
    }

  // Lighting Type flags
  for ( i = 0; i < slideltSize; i++ )
    {
      sprintf(gpcBuffer, "%ld", i);
      Tcl_SetVar(pInterp, gcppcSLIDELightReservedWords[i], 
		 gpcBuffer, TCL_GLOBAL_ONLY);
    }

  // Fog Type flags
  for ( i = 0; i < slidefogtSize; i++ )
    {
      sprintf(gpcBuffer, "%ld", i);
      Tcl_SetVar(pInterp, gcppcSLIDEFogReservedWords[i], 
		 gpcBuffer, TCL_GLOBAL_ONLY);
    }

  // Stencil flags
  for ( i = 0; i < slidesttSize; i++ )
    {
      sprintf(gpcBuffer, "%ld", i);
      Tcl_SetVar(pInterp, gcppcSLIDEStencilReservedWords[i], 
		 gpcBuffer, TCL_GLOBAL_ONLY);
    }

  // Aspect Ratio flags
  for ( i = 0; i < slideartSize; i++ )
    {
      sprintf(gpcBuffer, "%ld", i);
      Tcl_SetVar(pInterp, gcppcSLIDEAspectRatioReservedWords[i], 
		 gpcBuffer, TCL_GLOBAL_ONLY);
    }

  // CameraUI flags
  for ( i = 0; i < slidecuitSize; i++ )
    {
      sprintf(gpcBuffer, "%ld", i);
      Tcl_SetVar(pInterp, gcppcSLIDECameraUserInterfaceReservedWords[i], 
		 gpcBuffer, TCL_GLOBAL_ONLY);
    }

  // Subdivision flags
  for ( i = 0; i < slidesubtSize; i++ )
    {
      sprintf(gpcBuffer, "%ld", i);
      Tcl_SetVar(pInterp, gcppcSLIDESubdivisionReservedWords[i], 
		 gpcBuffer, TCL_GLOBAL_ONLY);
    }

  // Offset flags
  for ( i = 0; i < slideofftSize; i++ )
    {
      sprintf(gpcBuffer, "%ld", i);
      Tcl_SetVar(pInterp, gcppcSLIDEOffsetReservedWords[i], 
		 gpcBuffer, TCL_GLOBAL_ONLY);
    }

  // Terrain Type flags
  for ( i = 0; i < slidetertSize; i++ )
    {
      sprintf(gpcBuffer, "%ld", i);
      Tcl_SetVar(pInterp, gcppcSLIDETerrainReservedWords[i], 
		 gpcBuffer, TCL_GLOBAL_ONLY);
    }

  // Voronoi Type flags
  for ( i = 0; i < slidevtSize; i++ )
    {
      sprintf(gpcBuffer, "%ld", i);
      Tcl_SetVar(pInterp, gcppcSLIDEVoronoiReservedWords[i], 
		 gpcBuffer, TCL_GLOBAL_ONLY);
    }

  // Accuracy flags
  for ( i = 0; i < slideatSize; i++ )
    {
      sprintf(gpcBuffer, "%ld", i);
      Tcl_SetVar(pInterp, gcppcSLIDEAccuracyReservedWords[i], 
		 gpcBuffer, TCL_GLOBAL_ONLY);
    }
}

/////////////////////////////////////////////////////////////////////////////////////////

VOID OpenGUI(Tcl_Interp *pInterp)
{
  INT_32 iCode;

  ASSERT( !CSLIDEWorld::GetTclGUIExists() ); 

  // Create GUI
  time_t tt1, tt2;

  sprintf(gpcBuffer, "GUI:Create");
  tt1 = time(NULL);
  iCode = Tcl_Eval(pInterp, gpcBuffer);
  tt2 = time(NULL);
  printf( "GUI:Create time = %d seconds\n", tt2-tt1);

  if ( iCode != TCL_OK )
    {
      printf("Error %s \n", pInterp->result);
    }

  CSLIDEWorld::SetTclGUIExists(TRUE);
}

VOID ClearGUI(Tcl_Interp *pInterp)
{
  INT_32 iCode;

  ASSERT( CSLIDEWorld::GetTclGUIExists() ); 

  // Clear GUI
  sprintf(gpcBuffer, "GUI:Clear");
  iCode = Tcl_Eval(pInterp, gpcBuffer);
  if ( iCode != TCL_OK )
    {
      printf("Error %s \n", pInterp->result);
    }
}

VOID CloseGUI(Tcl_Interp *pInterp)
{
  INT_32 iCode;

  ASSERT( CSLIDEWorld::GetTclGUIExists() ); 

  // Destroy GUI
  sprintf(gpcBuffer, "GUI:Destroy");
  iCode = Tcl_Eval(pInterp, gpcBuffer);
  if ( iCode != TCL_OK )
    {
      printf("Error %s \n", pInterp->result);
    }
  
  CSLIDEWorld::SetTclGUIExists(FALSE);
}

int HandleMode(ClientData clientData, Tcl_Interp *pInterp, 
	       int argc, char **argv)
{
  Mode prevMode;
  INT_32 iCode;

  ASSERT( argc == 2 );

  prevMode = UIMode;

  if( !strcmp("Viewer", argv[1]) )
    {
      UIMode = VIEWER;
    }
  else
    {
      UIMode = EDITOR;
    }

  if ( (prevMode != UIMode) && (CSLIDEWorld::GetWorld() != NULL) )
    {
      // Resize window for GUI - toggles between gui size and user specified size
      sprintf(gpcBuffer, "ResizeWindow");
      iCode = Tcl_Eval(pInterp, gpcBuffer); 
      if ( iCode != TCL_OK )
	{
	  fprintf(stderr, "Tcl Error: %s: \"%s\"\n", gpcBuffer, pInterp->result);
	  ASSERT( FALSE );
	}

      if ( UIMode == VIEWER )
	{
	  CloseGUI(pInterp);
	}
      else
	{
	  OpenGUI(pInterp);
	  CSLIDEWorld::GetWorld()->DrawGUIGraph();
	}
    }

  return TCL_OK;
}

int HandleOpen(ClientData /*clientData*/, Tcl_Interp *pInterp, 
	       int argc, char **argv)
{
  ASSERT( (argc == 2) && (argv[1] != NULL) );

  INT_32 iCode;
  UINT_32 uiLength;
  CHAR *pDirname;

  if( !strcmp("reload", argv[1] ) )
    {
      // Reload current file
      ASSERT( gpcInputFile != NULL );
    }
  else  
    {
      // Open SLIDE file
      if ( gpcInputFile != NULL )
	{
	  delete gpcInputFile;
	  gpcInputFile = NULL;
	}
  
      sprintf(gpcBuffer, "tk_getOpenFile -defaultextension .slf -filetypes {{\"SLIDE Files\" {.slf}}}");
      iCode = Tcl_Eval(pInterp, gpcBuffer); 
      ASSERT( iCode == TCL_OK );

      uiLength = strlen(pInterp->result);
      if ( uiLength == 0 )
	{
	  // The user hit Cancel
	  return TCL_OK;
	}
      
      ASSERT( pInterp->result != NULL );
      gpcInputFile = new CHAR[uiLength + 1];
      ASSERT( gpcInputFile != NULL );
      strcpy(gpcInputFile, pInterp->result);

      // Change directory

      // Warning: If there is a space in the directory name (in Windows or NT) then
      //          cd won't work
      pDirname = GetFileDirPath ( gpcInputFile );
      sprintf(gpcBuffer, "cd %s", pDirname);
      iCode = Tcl_Eval(pInterp, gpcBuffer); 
      if ( iCode != TCL_OK )
	{
	  sprintf(gpcBuffer, "tk_messageBox -icon error -type ok -default ok -message \"Error could not cd to directory \\\"%s\\\"\\nMake sure there are no spaces in the directory name\\nExiting program.\"", pDirname);
	  iCode = Tcl_Eval(pInterp, gpcBuffer); 
	  ASSERT( iCode == TCL_OK );
	}
      delete[] pDirname;
    }

  // Enable Main Menu commands
  sprintf(gpcBuffer, "Main:Opened");
  iCode = Tcl_Eval(pInterp, gpcBuffer); 
  ASSERT( iCode == TCL_OK );

  // Graph and Editor windows
  if( UIMode == EDITOR )
    {
      if( CSLIDEWorld::GetTclGUIExists() )
	{
	  ClearGUI(pInterp);
	}
      else
	{
	  OpenGUI(pInterp);
	}
    }

  gbPreprocessDone = FALSE;
  gbCheckSemanticsDone = FALSE;

  if( UIMode == VIEWER )
    {
      LoadWorld();
    }
  else
    {
      LoadWorldGUI();

      // Resize window for GUI	  
      sprintf(gpcBuffer, "update idletasks");   // Update to make sure togl stuff is packed
      iCode = Tcl_Eval(pInterp, gpcBuffer); 
      ASSERT( iCode == TCL_OK );
      
      sprintf(gpcBuffer, "ResizeWindow");
      iCode = Tcl_Eval(pInterp, gpcBuffer); 
      ASSERT( iCode == TCL_OK );
    }

  return TCL_OK;
}

int HandleNew(ClientData /*clientData*/, Tcl_Interp *pInterp, 
	      int /*argc*/, char ** /*argv*/)
{
  if ( gpcInputFile != NULL )
    {
      delete gpcInputFile;
      gpcInputFile = NULL;
    }

  INT_32 iCode;
  UINT_32 uiLength;
  CHAR *pDirname;

  // Make sure in Editor Mode
  UIMode = EDITOR;
  sprintf(gpcBuffer, "SetMode Editor");
  iCode = Tcl_Eval(pInterp, gpcBuffer); 
  if( iCode != TCL_OK )
    {      
      fprintf(stderr, "Tcl Error: %s: \"%s\"\n", gpcBuffer, pInterp->result);
      ASSERT( FALSE );
    }

  // Open Default SLIDE file
  sprintf(gpcBuffer, "Main:GetDefaultInputFile");
  iCode = Tcl_Eval(pInterp, gpcBuffer); 
  if( iCode != TCL_OK )
    {      
      fprintf(stderr, "Tcl Error: %s: \"%s\"\n", gpcBuffer, pInterp->result);
      ASSERT( FALSE );
    }
  uiLength = strlen(pInterp->result);
  gpcInputFile = new CHAR[uiLength + 1];
  ASSERT( gpcInputFile != NULL );
  strcpy(gpcInputFile, pInterp->result);
  
  // Graph and Editor windows
  if( CSLIDEWorld::GetTclGUIExists() )
    {
      ClearGUI(pInterp);
    }
  else
    {
      OpenGUI(pInterp);
    }

  // Run it through SLIDE
  LoadWorldGUI();

  // Resize window for GUI	
  sprintf(gpcBuffer, "update idletasks");   // Update to make sure togl stuff is packed
  iCode = Tcl_Eval(pInterp, gpcBuffer); 
  ASSERT( iCode == TCL_OK );

  sprintf(gpcBuffer, "ResizeWindow");
  iCode = Tcl_Eval(pInterp, gpcBuffer); 
  ASSERT( iCode == TCL_OK );

  return TCL_OK;
}

int HandleClose(ClientData /*clientData*/, Tcl_Interp *pInterp, 
		int /*argc*/, char ** /*argv*/)
{
  UnloadWorld();

  if( CSLIDEWorld::GetTclGUIExists() )
    {
      CloseGUI(pInterp);
    }

  return TCL_OK;
}

int HandleExit(ClientData /*clientData*/, Tcl_Interp * /*interp*/, 
	       int /*argc*/, char ** /*argv*/)
{
  exit(0);

  return TCL_OK;
}

//////////////////////////////////////////////////////////////////////////////////////
// Write functions

int HandleWrite(ClientData /*clientData*/, Tcl_Interp *pInterp, 
	       int /*argc*/, char ** /*argv*/)
{
  if ( CSLIDEWorld::GetWorld() != NULL )
    {
      INT_32 iCode;
      UINT_32 uiLength;
      FILE *fpOutputFile;
      CHAR *pcOutputFile;
     
      sprintf(gpcBuffer, "tk_getSaveFile -defaultextension .slf -filetypes {{\"SLIDE Files\" {.slf}}}");
      iCode = Tcl_Eval(pInterp, gpcBuffer); 
      ASSERT( iCode == TCL_OK );
      
      uiLength = strlen(Tcl_GetStringResult(pInterp));
      if ( uiLength == 0 )
	{
	  // The user hit Cancel
	  return TCL_OK;
	}
  
      pcOutputFile = new CHAR[strlen(Tcl_GetStringResult(pInterp)) + 1];
      ASSERT( pcOutputFile != NULL );
      strcpy(pcOutputFile, Tcl_GetStringResult(pInterp));
      fpOutputFile = NULL;

      while ( fpOutputFile == NULL )
	{
	  fpOutputFile = fopen(pcOutputFile, "w");
	  if ( fpOutputFile == NULL )
	    {
	      sprintf(gpcBuffer, "tk_messageBox -icon error -message \"File \\\"%s\\\" cannot be written.\" -type retrycancel -default retry", pcOutputFile);
	      iCode = Tcl_Eval(pInterp, gpcBuffer); 
	      ASSERT( iCode == TCL_OK );
	      
	      if ( strcmp(Tcl_GetStringResult(pInterp), "cancel") == 0 )
		{
		  // Cancel
		  goto cancel;
		}
	    }
	}

      CSLIDEWorld::GetWorld()->HandleWriteSLIDE(pcOutputFile, fpOutputFile);

    cancel:
      if ( pcOutputFile != NULL )
	{
	  delete []pcOutputFile;
	}
      
      if ( fpOutputFile != stdout && fpOutputFile != NULL )
	{
	  ASSERT( fpOutputFile != NULL );
	  fclose(fpOutputFile);
	}
      
      fpOutputFile = NULL;
    }
  
  return TCL_OK;
}

int HandleWriteDOT(ClientData /*clientData*/, Tcl_Interp *pInterp, 
		   int /*argc*/, char ** /*argv*/)
{
  if ( CSLIDEWorld::GetWorld() != NULL )
    {
      INT_32 iCode;
      UINT_32 uiLength;
      FILE *fpOutputFile;
      CHAR *pcOutputFile;
     
      sprintf(gpcBuffer, "tk_getSaveFile -defaultextension .dot -filetypes {{\"DOT Files\" {.dot}}}");
      iCode = Tcl_Eval(pInterp, gpcBuffer); 
      ASSERT( iCode == TCL_OK );
      
      uiLength = strlen(Tcl_GetStringResult(pInterp));
      if ( uiLength == 0 )
	{
	  // The user hit Cancel
	  return TCL_OK;
	}
  
      pcOutputFile = new CHAR[strlen(Tcl_GetStringResult(pInterp)) + 1];
      ASSERT( pcOutputFile != NULL );
      strcpy(pcOutputFile, Tcl_GetStringResult(pInterp));
      fpOutputFile = NULL;

      while ( fpOutputFile == NULL )
	{
	  fpOutputFile = fopen(pcOutputFile, "w");
	  if ( fpOutputFile == NULL )
	    {
	      sprintf(gpcBuffer, "tk_messageBox -icon error -message \"File \\\"%s\\\" cannot be written.\" -type retrycancel -default retry", pcOutputFile);
	      iCode = Tcl_Eval(pInterp, gpcBuffer); 
	      ASSERT( iCode == TCL_OK );
	      
	      if ( strcmp(Tcl_GetStringResult(pInterp), "cancel") == 0 )
		{
		  // Cancel
		  goto cancel;
		}
	    }
	}

      CSLIDEWorld::GetWorld()->HandleWriteDOT(pcOutputFile, fpOutputFile);

    cancel:
      if ( pcOutputFile != NULL )
	{
	  delete []pcOutputFile;
	}
      
      if ( fpOutputFile != stdout && fpOutputFile != NULL )
	{
	  ASSERT( fpOutputFile != NULL );
	  fclose(fpOutputFile);
	}
      
      fpOutputFile = NULL;
    }
 
  return TCL_OK;
}

int HandleWriteRIB(ClientData /*clientData*/, Tcl_Interp *pInterp, 
		   int /*argc*/, char ** /*argv*/)
{
 if ( CSLIDEWorld::GetWorld() != NULL )
    {
      INT_32 iCode;
      UINT_32 uiLength;
      FILE *fpOutputFile;
      CHAR *pcOutputFile;
     
      sprintf(gpcBuffer, "tk_getSaveFile -defaultextension .rib -filetypes {{\"RIB Files\" {.rib}}}");
      iCode = Tcl_Eval(pInterp, gpcBuffer); 
      ASSERT( iCode == TCL_OK );
      
      uiLength = strlen(Tcl_GetStringResult(pInterp));
      if ( uiLength == 0 )
	{
	  // The user hit Cancel
	  return TCL_OK;
	}
  
      pcOutputFile = new CHAR[strlen(Tcl_GetStringResult(pInterp)) + 1];
      ASSERT( pcOutputFile != NULL );
      strcpy(pcOutputFile, Tcl_GetStringResult(pInterp));
      fpOutputFile = NULL;

      while ( fpOutputFile == NULL )
	{
	  fpOutputFile = fopen(pcOutputFile, "w");
	  if ( fpOutputFile == NULL )
	    {
	      sprintf(gpcBuffer, "tk_messageBox -icon error -message \"File \\\"%s\\\" cannot be written.\" -type retrycancel -default retry", pcOutputFile);
	      iCode = Tcl_Eval(pInterp, gpcBuffer); 
	      ASSERT( iCode == TCL_OK );
	      
	      if ( strcmp(Tcl_GetStringResult(pInterp), "cancel") == 0 )
		{
		  // Cancel
		  goto cancel;
		}
	    }
	}

      CSLIDEWorld::GetWorld()->HandleWriteRIB(pcOutputFile, fpOutputFile);  

    cancel:
      if ( pcOutputFile != NULL )
	{
	  delete []pcOutputFile;
	}
      
      if ( fpOutputFile != stdout && fpOutputFile != NULL )
	{
	  ASSERT( fpOutputFile != NULL );
	  fclose(fpOutputFile);
	}
      
      fpOutputFile = NULL;
    }
 
  return TCL_OK;
}

int HandleWritePOV(ClientData /*clientData*/, Tcl_Interp *pInterp, 
		   int /*argc*/, char ** /*argv*/)
{
 if ( CSLIDEWorld::GetWorld() != NULL )
    {
      INT_32 iCode;
      UINT_32 uiLength;
      FILE *fpOutputFile;
      CHAR *pcOutputFile;
     
      sprintf(gpcBuffer, "tk_getSaveFile -defaultextension .pov -filetypes {{\"POV Files\" {.pov}}}");
      iCode = Tcl_Eval(pInterp, gpcBuffer); 
      ASSERT( iCode == TCL_OK );
      
      uiLength = strlen(Tcl_GetStringResult(pInterp));
      if ( uiLength == 0 )
	{
	  // The user hit Cancel
	  return TCL_OK;
	}
  
      pcOutputFile = new CHAR[strlen(Tcl_GetStringResult(pInterp)) + 1];
      ASSERT( pcOutputFile != NULL );
      strcpy(pcOutputFile, Tcl_GetStringResult(pInterp));
      fpOutputFile = NULL;

      while ( fpOutputFile == NULL )
	{
	  fpOutputFile = fopen(pcOutputFile, "w");
	  if ( fpOutputFile == NULL )
	    {
	      sprintf(gpcBuffer, "tk_messageBox -icon error -message \"File \\\"%s\\\" cannot be written.\" -type retrycancel -default retry", pcOutputFile);
	      iCode = Tcl_Eval(pInterp, gpcBuffer); 
	      ASSERT( iCode == TCL_OK );
	      
	      if ( strcmp(Tcl_GetStringResult(pInterp), "cancel") == 0 )
		{
		  // Cancel
		  goto cancel;
		}
	    }
	}

      CSLIDEWorld::GetWorld()->HandleWritePOV(pcOutputFile, fpOutputFile);  

    cancel:
      if ( pcOutputFile != NULL )
	{
	  delete []pcOutputFile;
	}
      
      if ( fpOutputFile != stdout && fpOutputFile != NULL )
	{
	  ASSERT( fpOutputFile != NULL );
	  fclose(fpOutputFile);
	}
      
      fpOutputFile = NULL;
    }
 
  return TCL_OK;
}

int HandleWriteWRLZ(ClientData /*clientData*/, Tcl_Interp *pInterp, 
		    int /*argc*/, char ** /*argv*/)
{
 if ( CSLIDEWorld::GetWorld() != NULL )
    {
      INT_32 iCode;
      UINT_32 uiLength;
      FILE *fpOutputFile;
      CHAR *pcOutputFile;
     
      sprintf(gpcBuffer, "tk_getSaveFile -defaultextension .wrl -filetypes {{\"VRML Files\" {.wrl}}}");
      iCode = Tcl_Eval(pInterp, gpcBuffer); 
      ASSERT( iCode == TCL_OK );
      
      uiLength = strlen(Tcl_GetStringResult(pInterp));
      if ( uiLength == 0 )
	{
	  // The user hit Cancel
	  return TCL_OK;
	}
  
      pcOutputFile = new CHAR[strlen(Tcl_GetStringResult(pInterp)) + 1];
      ASSERT( pcOutputFile != NULL );
      strcpy(pcOutputFile, Tcl_GetStringResult(pInterp));
      fpOutputFile = NULL;

      while ( fpOutputFile == NULL )
	{
	  fpOutputFile = fopen(pcOutputFile, "w");
	  if ( fpOutputFile == NULL )
	    {
	      sprintf(gpcBuffer, "tk_messageBox -icon error -message \"File \\\"%s\\\" cannot be written.\" -type retrycancel -default retry", pcOutputFile);
	      iCode = Tcl_Eval(pInterp, gpcBuffer); 
	      ASSERT( iCode == TCL_OK );
	      
	      if ( strcmp(Tcl_GetStringResult(pInterp), "cancel") == 0 )
		{
		  // Cancel
		  goto cancel;
		}
	    }
	}

      CSLIDEWorld::GetWorld()->HandleWriteWRLZ(pcOutputFile, fpOutputFile);  

    cancel:
      if ( pcOutputFile != NULL )
	{
	  delete []pcOutputFile;
	}
      
      if ( fpOutputFile != stdout && fpOutputFile != NULL )
	{
	  ASSERT( fpOutputFile != NULL );
	  fclose(fpOutputFile);
	}
      
      fpOutputFile = NULL;
    }
 
  return TCL_OK;
}

int HandleWriteSTL(ClientData /*clientData*/, Tcl_Interp *pInterp, 
		   int /*argc*/, char ** /*argv*/)
{
 if ( CSLIDEWorld::GetWorld() != NULL )
    {
      INT_32 iCode;
      UINT_32 uiLength;
      FILE *fpOutputFile;
      CHAR *pcOutputFile;
     
      sprintf(gpcBuffer, "tk_getSaveFile -defaultextension .stl -filetypes {{\"STL Files\" {.stl}}}");
      iCode = Tcl_Eval(pInterp, gpcBuffer); 
      ASSERT( iCode == TCL_OK );
      
      uiLength = strlen(Tcl_GetStringResult(pInterp));
      if ( uiLength == 0 )
	{
	  // The user hit Cancel
	  return TCL_OK;
	}
  
      pcOutputFile = new CHAR[strlen(Tcl_GetStringResult(pInterp)) + 1];
      ASSERT( pcOutputFile != NULL );
      strcpy(pcOutputFile, Tcl_GetStringResult(pInterp));
      fpOutputFile = NULL;

      while ( fpOutputFile == NULL )
	{
	  fpOutputFile = fopen(pcOutputFile, "w");
	  if ( fpOutputFile == NULL )
	    {
	      sprintf(gpcBuffer, "tk_messageBox -icon error -message \"File \\\"%s\\\" cannot be written.\" -type retrycancel -default retry", pcOutputFile);
	      iCode = Tcl_Eval(pInterp, gpcBuffer); 
	      ASSERT( iCode == TCL_OK );
	      
	      if ( strcmp(Tcl_GetStringResult(pInterp), "cancel") == 0 )
		{
		  // Cancel
		  goto cancel;
		}
	    }
	}

      CSLIDEWorld::GetWorld()->HandleWriteSTL(pcOutputFile, fpOutputFile);  

    cancel:
      if ( pcOutputFile != NULL )
	{
	  delete []pcOutputFile;
	}
      
      if ( fpOutputFile != stdout && fpOutputFile != NULL )
	{
	  ASSERT( fpOutputFile != NULL );
	  fclose(fpOutputFile);
	}
      
      fpOutputFile = NULL;
    }
 
  return TCL_OK;
}

int HandleWriteSIF(ClientData /*clientData*/, Tcl_Interp *pInterp, 
		   int /*argc*/, char ** /*argv*/)
{
 if ( CSLIDEWorld::GetWorld() != NULL )
    {
      INT_32 iCode;
      UINT_32 uiLength;
      FILE *fpOutputFile;
      CHAR *pcOutputFile;
     
      sprintf(gpcBuffer, "tk_getSaveFile -defaultextension .sif -filetypes {{\"SIF Files\" {.sif}}}");
      iCode = Tcl_Eval(pInterp, gpcBuffer); 
      ASSERT( iCode == TCL_OK );
      
      uiLength = strlen(Tcl_GetStringResult(pInterp));
      if ( uiLength == 0 )
	{
	  // The user hit Cancel
	  return TCL_OK;
	}
  
      pcOutputFile = new CHAR[strlen(Tcl_GetStringResult(pInterp)) + 1];
      ASSERT( pcOutputFile != NULL );
      strcpy(pcOutputFile, Tcl_GetStringResult(pInterp));
      fpOutputFile = NULL;

      while ( fpOutputFile == NULL )
	{
	  fpOutputFile = fopen(pcOutputFile, "w");
	  if ( fpOutputFile == NULL )
	    {
	      sprintf(gpcBuffer, "tk_messageBox -icon error -message \"File \\\"%s\\\" cannot be written.\" -type retrycancel -default retry", pcOutputFile);
	      iCode = Tcl_Eval(pInterp, gpcBuffer); 
	      ASSERT( iCode == TCL_OK );
	      
	      if ( strcmp(Tcl_GetStringResult(pInterp), "cancel") == 0 )
		{
		  // Cancel
		  goto cancel;
		}
	    }
	}

      CSLIDEWorld::GetWorld()->HandleWriteSIF(pcOutputFile, fpOutputFile);  

    cancel:
      if ( pcOutputFile != NULL )
	{
	  delete []pcOutputFile;
	}
      
      if ( fpOutputFile != stdout && fpOutputFile != NULL )
	{
	  ASSERT( fpOutputFile != NULL );
	  fclose(fpOutputFile);
	}
      
      fpOutputFile = NULL;
    }
 
  return TCL_OK;
}

int HandleWriteLSIF(ClientData /*clientData*/, Tcl_Interp *pInterp, 
		   int argc, char ** argv)
{
 if ( CSLIDEWorld::GetWorld() != NULL )
    {
      if ( argc != 2 )
	{
	  sprintf(pInterp->result, "ERROR WriteLSIF: wrong number of arguments %d != 2\n",
		  argc);
	  return TCL_ERROR;
	}
      INT_32 iCode;
      UINT_32 uiLength;
      FILE *fpOutputFile;
      CHAR *pcOutputFile;
     
      pcOutputFile = argv[1];
      fpOutputFile = NULL;
      fpOutputFile = fopen(pcOutputFile, "w");
      if ( fpOutputFile == NULL )
	{
	  sprintf(pInterp->result, "ERROR WriteLSIF: can not open \"%s\" for writing\n",
		  pcOutputFile);
	  return TCL_ERROR;
	}

      CSLIDEWorld::GetWorld()->HandleWriteLSIF(pcOutputFile, fpOutputFile);  

      if ( fpOutputFile != stdout && fpOutputFile != NULL )
	{
	  ASSERT( fpOutputFile != NULL );
	  fclose(fpOutputFile);
	}
      
      fpOutputFile = NULL;
    }
 
  return TCL_OK;
}

