#ifndef __OOT_H__
#define __OOT_H__

#include <tcl.h>
#include <tk.h>
#include "genlib.h"
#include "datalib.h"

typedef struct _OotClass OotClass;
typedef struct _OotStaticClass OotStaticClass;
typedef struct _OotObject OotObject;
typedef struct _OotMethodEntry OotMethodEntry;

typedef int (Oot_Constructor) (Tcl_Interp *pInterp, OotClass *pClass, 
			       int argc, char *argv[]);

#define OOT_PARAM_TYPE_VOID   0
#define OOT_PARAM_TYPE_BOOL   1
#define OOT_PARAM_TYPE_INT    2
#define OOT_PARAM_TYPE_FLOAT  3
#define OOT_PARAM_TYPE_DOUBLE 4
#define OOT_PARAM_TYPE_OTHER  5
#define OOT_PARAM_TYPE_INT_2  6
#define OOT_PARAM_TYPE_INT_3  7
#define OOT_PARAM_TYPE_INT_4  8
#define OOT_PARAM_TYPE_FLOAT_3  9
#define OOT_PARAM_TYPE_FLOAT_4  10
#define OOT_PARAM_TYPE_KEYBOARD 11
#define OOT_PARAM_TYPE_PCHAR    12
#define OOT_PARAM_TYPE_MAX    13

typedef int (Oot_Method_Void) (Tcl_Interp *pInterp, OotObject *pObject);
typedef int (Oot_Method_Bool) (Tcl_Interp *pInterp, OotObject *pObject, BOOL b);
typedef int (Oot_Method_Int) (Tcl_Interp *pInterp, OotObject *pObject, INT_32 i);
typedef int (Oot_Method_Float) (Tcl_Interp *pInterp, OotObject *pObject, FLOAT f);
typedef int (Oot_Method_Double) (Tcl_Interp *pInterp, OotObject *pObject, DOUBLE d);
typedef int (Oot_Method_Other) (Tcl_Interp *pInterp, OotObject *pObject, int argc, char *argv[]);
typedef int (Oot_Method_Int_2) (Tcl_Interp *pInterp, OotObject *pObject, 
				INT_32 i0, INT_32 i1);
typedef int (Oot_Method_Int_3) (Tcl_Interp *pInterp, OotObject *pObject, 
				INT_32 i0, INT_32 i1, INT_32 i2);
typedef int (Oot_Method_Int_4) (Tcl_Interp *pInterp, OotObject *pObject, 
				INT_32 i0, INT_32 i1, INT_32 i2, INT_32 i3);
typedef int (Oot_Method_Float_3) (Tcl_Interp *pInterp, OotObject *pObject, 
				  FLOAT f0, FLOAT f1, FLOAT f2);
typedef int (Oot_Method_Float_4) (Tcl_Interp *pInterp, OotObject *pObject, 
				  FLOAT f0, FLOAT f1, FLOAT f2, FLOAT f3);
typedef int (Oot_Method_Keyboard) (Tcl_Interp *pInterp, OotObject *pObject, 
				   CHAR *pcKeyCode, INT_32 iState, INT_32 iX, INT_32 iY);
typedef int (Oot_Method_pChar) (Tcl_Interp *pInterp, OotObject *pObject, CHAR *pc);

struct _OotClass
{
  CHAR *pcName;
  Tcl_Interp *pInterp;
  Oot_Constructor *fpConstructor;
  CHashString *pCHashObjects;
  CHashString *pCHashMethods;
  OotMethodEntry *pMethodEntries;
  UINT_32 uiMethodEntries;
  const CHAR **cppcFields;
  UINT_32 uiFields;
};


struct _OotStaticClass
{
  CHAR *pcName;
  Tcl_Interp *pInterp;
  OotMethodEntry *pMethodEntries;
  UINT_32 uiMethodEntries;
  
};

struct _OotObject
{
  CHAR *pcName;
  OotClass *pClass;
  ClientData pvClientData;
};

struct _OotMethodEntry
{
  CHAR *pcName;
  UINT_32 uiParamType;
  VOID *fpMethod;
};

typedef struct _OotScaleEntry
{
  FLOAT fMin;
  FLOAT fMax;
  FLOAT fResolution;
} OotScaleEntry;

int Oot_Init(Tcl_Interp *pInterp);

OotClass *Oot_CreateClass(Tcl_Interp *pInterp, const CHAR *pcName,
			  Oot_Constructor *fpConstructor, 
			  OotMethodEntry *pMethods, UINT_32 uiMethods,
			  const CHAR **cppcFields, UINT_32 uiFields);
OotStaticClass *Oot_CreateStaticClass(Tcl_Interp *pInterp, const CHAR *pcName,
			  OotMethodEntry *pMethods, UINT_32 uiMethods);
OotObject *Oot_CreateObject(Tcl_Interp *pInterp, OotClass *pClass, 
			    const CHAR *pcName, ClientData pvClientData);
VOID Oot_DestroyObject(Tcl_Interp *pInterp, OotClass *pClass, 
		       const CHAR *pcName);
OotObject *Oot_GetObject(OotClass *pClass, const CHAR *pcName);

VOID Oot_SetObjectField(OotObject *pObject, const CHAR *pcField, BOOL b);
VOID Oot_SetObjectField(OotObject *pObject, const CHAR *pcField, INT_32 i);
VOID Oot_SetObjectField(OotObject *pObject, const CHAR *pcField, FLOAT f);
VOID Oot_SetObjectField(OotObject *pObject, const CHAR *pcField, DOUBLE d);
VOID Oot_SetObjectField(OotObject *pObject, const CHAR *pcField, const CHAR *pc);

#endif // __OOT_H__
