#include "point.h"

MATHLIB_API Point::Point()
{
  Null();
}

MATHLIB_API Point::Point(FLOAT x, FLOAT y, FLOAT z, FLOAT w/* = 1.0f*/)
{
  m_pt[X] = x;
  m_pt[Y] = y;
  m_pt[Z] = z;
  m_pt[W] = w;
}

MATHLIB_API Point::~Point()
{
}

MATHLIB_API VOID Point::Null()
{
  m_pt[X] = 0.0f;
  m_pt[Y] = 0.0f;
  m_pt[Z] = 0.0f;
  m_pt[W] = 1.0f;
}

MATHLIB_API BOOL Point::Homogenize()
{
  if ( m_pt[W] == 0.0f )
    {
      return FALSE;
    }

  m_pt[X] = m_pt[X] / m_pt[W];
  m_pt[Y] = m_pt[Y] / m_pt[W];
  m_pt[Z] = m_pt[Z] / m_pt[W];
  m_pt[W] = 1.0f;

  return TRUE;
}

MATHLIB_API Point &Point::operator=(const Point &pt)
{
  m_pt[0] = pt(0);
  m_pt[1] = pt(1);
  m_pt[2] = pt(2);
  m_pt[3] = pt(3);

  return *this;
}

MATHLIB_API Point &Point::operator=(const Vector &v)
{
  m_pt[0] = v(0);
  m_pt[1] = v(1);
  m_pt[2] = v(2);
  m_pt[3] = v(3) + (FLOAT)1.0;

  return *this;
}

MATHLIB_API Point &Point::operator+=(const Vector &v)
{
  m_pt[0] = m_pt[0] + v(0);
  m_pt[1] = m_pt[1] + v(1);
  m_pt[2] = m_pt[2] + v(2);
  m_pt[3] = m_pt[3] + v(3);

  return *this;
}

MATHLIB_API Point &Point::operator-=(const Vector &v)
{
  m_pt[0] = m_pt[0] - v(0);
  m_pt[1] = m_pt[1] - v(1);
  m_pt[2] = m_pt[2] - v(2);
  m_pt[3] = m_pt[3] - v(3);

  return *this;
}

MATHLIB_API Point &Point::operator+=(const Point &pt)
{
  m_pt[0] = m_pt[0] + pt(0);
  m_pt[1] = m_pt[1] + pt(1);
  m_pt[2] = m_pt[2] + pt(2);
  m_pt[3] = m_pt[3] + pt(3);

  return *this;
}

MATHLIB_API Point &Point::operator*=(const FLOAT f)
{
  m_pt[0] = m_pt[0] * f;
  m_pt[1] = m_pt[1] * f;
  m_pt[2] = m_pt[2] * f;
  m_pt[3] = m_pt[3] * f;

  return *this;
}

MATHLIB_API Point &Point::operator/=(const FLOAT f)
{
  m_pt[0] = m_pt[0] / f;
  m_pt[1] = m_pt[1] / f;
  m_pt[2] = m_pt[2] / f;
  m_pt[3] = m_pt[3] / f;

  return *this;
}

MATHLIB_API BOOL Point::operator==(const Point &pt) const
{
  FLOAT fDot, fDotV, fVDotV;

  fDot = this->Dot(*this);
  fDotV = this->Dot(pt);
  fVDotV = pt.Dot(pt);

  return ((fDot > (fDotV - EPSILON)) && (fDot < (fDotV + EPSILON))) &&
         ((fDot > (fVDotV - EPSILON)) && (fDot < (fVDotV + EPSILON)));
}

MATHLIB_API BOOL Point::operator!=(const Point &pt) const
{
  return !(*this == pt);
}

MATHLIB_API FLOAT Point::Dot(const Vector &v) const
{
  return (((*this)(0)*v(0)) + ((*this)(1)*v(1)) + 
          ((*this)(2)*v(2)) + ((*this)(3)*v(3))); 
}

MATHLIB_API BOOL Point::operator<(const Point &pt) const
{
/*
  if ( (*this)(Y) == pt(Y) )
    {
      (*this)(X) < pt(X);
    }
  return (*this)(Y) < pt(Y);
 */
  return ( (*this)(X) < pt(X) &&
	   (*this)(Y) < pt(Y) &&
	   (*this)(Z) < pt(Z) );
}

MATHLIB_API BOOL Point::operator<=(const Point &pt) const
{
  return ( (*this)(X) <= pt(X) &&
	   (*this)(Y) <= pt(Y) &&
	   (*this)(Z) <= pt(Z) );
}

MATHLIB_API BOOL Point::operator>(const Point &pt) const
{
/*
  if ( (*this)(Y) == pt(Y) )
    {
      (*this)(X) > pt(X);
    }
  return (*this)(Y) > pt(Y);
*/
  return ( (*this)(X) > pt(X) &&
	   (*this)(Y) > pt(Y) &&
	   (*this)(Z) > pt(Z) );
}

MATHLIB_API BOOL Point::operator>=(const Point &pt) const
{
  return ( (*this)(X) >= pt(X) &&
	   (*this)(Y) >= pt(Y) &&
	   (*this)(Z) >= pt(Z) );
}

MATHLIB_API FLOAT Point::Dot(const Point &pt) const
{
  return (((*this)(0)*pt(0)) + ((*this)(1)*pt(1)) + 
          ((*this)(2)*pt(2)) + ((*this)(3)*pt(3))); 
}

MATHLIB_API VOID Point::Display(FILE *fpOutput/* = stdout*/) const
{
  UINT_32 i;

  fprintf(fpOutput, "< ");
  for ( i = 0; i < 4; i++)
    {
      fprintf(fpOutput, "%3.3f ", (*this)(i));
    }
  fprintf(fpOutput, ">\n");
}

