#include "qeedge.h"

#include "qevertex.h"
#include "qewedge.h"
#include "qefacet.h"
#include "qemesh.h"

//////////////////////////////////////////////////////////////////////
// QEEdge Class
//

const FLOAT cfKnotInterval = 0.0;

QEEdge::QEEdge()
{
  UINT_32 i, j;

  m_pQEVertex = NULL;
  m_pQEWedge = NULL;
  m_pQEFacet = NULL;
  m_pQEEdge = NULL;
  for ( i = 0; i < 2; i++ )
    {
      for ( j = 0; j < 2; j++ )
	{
	  m_pppQEEdge[i][j] = NULL;
	}
    }

  m_fKnotInterval = cfKnotInterval;
  m_eVertexType = vtStart;

  m_peForward = NULL;
  m_pvxForward = NULL;

  m_phe = NULL;

  for ( i = 0; i < 2; i++ )
    {
      m_peLink[i] = NULL;
    }
}

QEEdge::~QEEdge()
{
  Uninit();
}

VOID QEEdge::Init()
{
  UINT_32 i, j;

  m_pQEVertex = NULL;
  m_pQEWedge = NULL;
  m_pQEFacet = NULL;
  m_pQEEdge = NULL;
  for ( i = 0; i < 2; i++ )
    {
      for ( j = 0; j < 2; j++ )
	{
	  m_pppQEEdge[i][j] = NULL;
	}
    }

  m_fKnotInterval = cfKnotInterval;
  m_eVertexType = vtStart;

  m_peForward = NULL;
  m_pvxForward = NULL;

  m_phe = NULL;

  for ( i = 0; i < 2; i++ )
    {
      ASSERT( m_peLink[i] == NULL );
    }
}

VOID QEEdge::Uninit()
{
  UINT_32 i, j;

  m_pQEVertex = NULL;
  m_pQEWedge = NULL;
  m_pQEFacet = NULL;
  m_pQEEdge = NULL;
  for ( i = 0; i < 2; i++ )
    {
      for ( j = 0; j < 2; j++ )
	{
	  m_pppQEEdge[i][j] = NULL;
	}
    }

  m_fKnotInterval = cfKnotInterval;
  m_eVertexType = vtStart;

  m_peForward = NULL;
  m_pvxForward = NULL;

  m_phe = NULL;

  for ( i = 0; i < 2; i++ )
    {
      m_peLink[i] = NULL;
    }
}

BOOL QEEdge::Initialized()
{
  return TRUE;
}

BOOL QEEdge::Valid()
{
  return TRUE;
}

VOID QEEdge::Preprocess()
{
}

QEEdge::VertexType QEEdge::CalculateVertexType()
{
  QEEdge *peCurr, *pePrev, *peNext;

  peCurr = this;

  pePrev = peCurr->GetEdgePrev(0);
  ASSERT( pePrev != NULL );
  peNext = peCurr->GetEdge();
  ASSERT( peNext != NULL );
  ASSERT( pePrev != peNext );

  if ( (*(peCurr->GetVertex())) > (*(pePrev->GetVertex())) &&
       (*(peCurr->GetVertex())) > (*(peNext->GetVertex())) )
    {
      // Start Vertex or a Split Vertex
      if ( QEMesh::Counterclockwise(pePrev->GetVertex()->GetPoint(),
				    peCurr->GetVertex()->GetPoint(),
				    peNext->GetVertex()->GetPoint()) )
	{
	  // Start Vertex
	  return vtStart;
	}
      else
	{
	  // Split Vertex
	  return vtSplit;
	}
    }
  else if ( (*(peCurr->GetVertex())) < (*(pePrev->GetVertex())) &&
	    (*(peCurr->GetVertex())) < (*(peNext->GetVertex())) )
    {
      // End Vertex or a Merge Vertex
      if ( QEMesh::Counterclockwise(pePrev->GetVertex()->GetPoint(),
				    peCurr->GetVertex()->GetPoint(),
				    peNext->GetVertex()->GetPoint()) )
	{
	  // End Vertex
	  return vtEnd;
	}
      else
	{
	  // Merge Vertex
	  return vtMerge;
	}
    }
  else 
    {
      // Regular Vertex
      if ( (*(pePrev->GetVertex())) < (*(peNext->GetVertex())) )
	{
	  // Right Side Edge -- Material is to the left
	  return vtRight;
	}
      else
	{
	  ASSERT( (*(pePrev->GetVertex())) > (*(peNext->GetVertex())) );
	  // Left Side Edge -- Material is to the right
	  return vtLeft;
	}
    }
}

FLOAT QEEdge::CalculateScanX(FLOAT fY)
{
  Point pt0, pt1;
  Vector v;
  FLOAT fDeltaY;

  pt0 = GetVertex()->GetPoint();
  pt1 = GetEdge()->GetVertex()->GetPoint();
  v = pt1 - pt0;
  fDeltaY = fY - pt0(Y);
    
  if ( v(Y) == 0.0 )
    {
      return pt0(X);
    }

  return pt0(X) + (v(X) * fDeltaY / v(Y));
}

Point QEEdge::CalculateMidpointCatmullClark()
{
  FLOAT pfWeight[2], fWeights;
  Point pt, ppt[2];
  QEEdge *ppe[2];

  ppe[0] = this;
  ppe[1] = ppe[0]->GetEdge();

  // P[0]
  ppt[0] = ppe[0]->GetVertex()->GetPoint();
  // d 1,0,0
  pfWeight[0] = ppe[1]->GetKnotInterval();
  // d 1,0,2
  pfWeight[0] += ppe[1]->GetEdgeNext(1)->GetEdgeNext(1)->GetKnotInterval();
  // d 1,0,-2
  pfWeight[0] += ppe[1]->GetEdgePrev(1)->GetEdgePrev(1)->GetKnotInterval();

  // P[1]
  ppt[1] = ppe[1]->GetVertex()->GetPoint();
  // d 0,1,0
  pfWeight[1] = ppe[0]->GetKnotInterval();
  // d 0,1,2
  pfWeight[1] += ppe[0]->GetEdgeNext(1)->GetEdgeNext(1)->GetKnotInterval();
  // d 0,1,-2
  pfWeight[1] += ppe[0]->GetEdgePrev(1)->GetEdgePrev(1)->GetKnotInterval();

  fWeights = pfWeight[0] + pfWeight[1];
  if ( fWeights == 0.0 )
    {
      pt.Null();
    }
  else
    {
      pt = ( pfWeight[0] * ppt[0] + pfWeight[1] * ppt[1] ) / fWeights;
      pt[3] = 1.0;
    }

  return pt;
}

VOID QEEdge::Display()
{
  fprintf(stderr, "    E %p:  V[0] = %p  <==>  V[1] = %p: ",
	  this, GetVertex(), GetNeighborVertex());

  if ( GetKnotInterval() == INFINITY )
    {
      fprintf(stderr, "INFINTY\n");
    }
  else
    {
      fprintf(stderr, "%f\n", GetKnotInterval());
    }
}
