#include "qemem.h"

#include "qeedgecollapse.h"
#include "qenode.h"
#include "qevertex.h"
#include "qeedge.h"
#include "qewedge.h"
#include "qefacet.h"
#include "qemesh.h"

static UINT_32 s_uiQEObjectCount = 0;

static UINT_32 s_uiNewedQEEdgeCollapses = 0;
static UINT_32 s_uiAllocatedQEEdgeCollapses = 0;
static CObjectList<QEEdgeCollapse> *s_plistQEEdgeCollapsesFree = NULL;

static UINT_32 s_uiNewedQEEdgeNodes = 0;
static UINT_32 s_uiAllocatedQEEdgeNodes = 0;
static CObjectList<QEEdgeNode> *s_plistQEEdgeNodesFree = NULL;

static UINT_32 s_uiNewedQEFacetNodes = 0;
static UINT_32 s_uiAllocatedQEFacetNodes = 0;
static CObjectList<QEFacetNode> *s_plistQEFacetNodesFree = NULL;

static UINT_32 s_uiNewedQEVertices = 0;
static UINT_32 s_uiAllocatedQEVertices = 0;
static CObjectList<QEVertex> *s_plistQEVerticesFree;

static UINT_32 s_uiNewedQEEdges = 0;
static UINT_32 s_uiAllocatedQEEdges = 0;
static CObjectList<QEEdge> *s_plistQEEdgesFree = NULL;

static UINT_32 s_uiNewedQEWedges = 0;
static UINT_32 s_uiAllocatedQEWedges = 0;
static CObjectList<QEWedge> *s_plistQEWedgesFree = NULL;

static UINT_32 s_uiNewedQEFacets = 0;
static UINT_32 s_uiAllocatedQEFacets = 0;
static CObjectList<QEFacet> *s_plistQEFacetsFree = NULL;

static UINT_32 s_uiNewedQEMeshs = 0;
static UINT_32 s_uiAllocatedQEMeshs = 0;
static CObjectList<QEMesh> *s_plistQEMeshsFree = NULL;

static VOID InitQEMemory()
{
  ASSERT( s_uiQEObjectCount == 0 );

  ASSERT( s_plistQEEdgeCollapsesFree == NULL );
  s_plistQEEdgeCollapsesFree = new CObjectList<QEEdgeCollapse>;
  ASSERT( s_plistQEEdgeCollapsesFree != NULL );
  s_plistQEEdgeCollapsesFree->Init();
  
  ASSERT( s_plistQEEdgeNodesFree == NULL );
  s_plistQEEdgeNodesFree = new CObjectList<QEEdgeNode>;
  ASSERT( s_plistQEEdgeNodesFree != NULL );
  s_plistQEEdgeNodesFree->Init();
  
  ASSERT( s_plistQEFacetNodesFree == NULL );
  s_plistQEFacetNodesFree = new CObjectList<QEFacetNode>;
  ASSERT( s_plistQEFacetNodesFree != NULL );
  s_plistQEFacetNodesFree->Init();
  
  ASSERT( s_plistQEVerticesFree == NULL );
  s_plistQEVerticesFree = new CObjectList<QEVertex>;
  ASSERT( s_plistQEVerticesFree != NULL );
  s_plistQEVerticesFree->Init();
  
  ASSERT( s_plistQEEdgesFree == NULL );
  s_plistQEEdgesFree = new CObjectList<QEEdge>;
  ASSERT( s_plistQEEdgesFree != NULL );
  s_plistQEEdgesFree->Init();
  
  ASSERT( s_plistQEWedgesFree == NULL );
  s_plistQEWedgesFree = new CObjectList<QEWedge>;
  ASSERT( s_plistQEWedgesFree != NULL );
  s_plistQEWedgesFree->Init();

  ASSERT( s_plistQEFacetsFree == NULL );
  s_plistQEFacetsFree = new CObjectList<QEFacet>;
  ASSERT( s_plistQEFacetsFree != NULL );
  s_plistQEFacetsFree->Init();

  ASSERT( s_plistQEMeshsFree == NULL );
  s_plistQEMeshsFree = new CObjectList<QEMesh>;
  ASSERT( s_plistQEMeshsFree != NULL );
  s_plistQEMeshsFree->Init();
}

static VOID UninitQEMemory()
{
  ASSERT( s_uiQEObjectCount == 0 );

  // Clean up our free list of QE EdgeCollapses
  ASSERT( s_uiAllocatedQEEdgeCollapses == 0 );
  if ( s_plistQEEdgeCollapsesFree != NULL )
    {
      QEEdgeCollapse *pQEEdgeCollapse;
      
      for ( pQEEdgeCollapse = s_plistQEEdgeCollapsesFree->RemoveFirst();
	    pQEEdgeCollapse != NULL;
	    pQEEdgeCollapse = s_plistQEEdgeCollapsesFree->RemoveFirst() )
	{
	  delete pQEEdgeCollapse;
	}

      delete s_plistQEEdgeCollapsesFree;
      s_plistQEEdgeCollapsesFree = NULL;
      s_uiNewedQEEdgeCollapses = 0;
    }

  // Clean up our free list of QE EdgeNodes
  ASSERT( s_uiAllocatedQEEdgeNodes == 0 );
  if ( s_plistQEEdgeNodesFree != NULL )
    {
      QEEdgeNode *pQEEdgeNode;
      
      for ( pQEEdgeNode = s_plistQEEdgeNodesFree->RemoveFirst();
	    pQEEdgeNode != NULL;
	    pQEEdgeNode = s_plistQEEdgeNodesFree->RemoveFirst() )
	{
	  delete pQEEdgeNode;
	}

      delete s_plistQEEdgeNodesFree;
      s_plistQEEdgeNodesFree = NULL;
      s_uiNewedQEEdgeNodes = 0;
    }

  // Clean up our free list of QE FacetNodes
  ASSERT( s_uiAllocatedQEFacetNodes == 0 );
  if ( s_plistQEFacetNodesFree != NULL )
    {
      QEFacetNode *pQEFacetNode;
      
      for ( pQEFacetNode = s_plistQEFacetNodesFree->RemoveFirst();
	    pQEFacetNode != NULL;
	    pQEFacetNode = s_plistQEFacetNodesFree->RemoveFirst() )
	{
	  delete pQEFacetNode;
	}

      delete s_plistQEFacetNodesFree;
      s_plistQEFacetNodesFree = NULL;
      s_uiNewedQEFacetNodes = 0;
    }

  // Clean up our free list of QE Vertices
  ASSERT( s_uiAllocatedQEVertices == 0 );
  if ( s_plistQEVerticesFree != NULL )
    {
      QEVertex *pQEVertex;
      
      for ( pQEVertex = s_plistQEVerticesFree->RemoveFirst();
	    pQEVertex != NULL;
	    pQEVertex = s_plistQEVerticesFree->RemoveFirst() )
	{
	  delete pQEVertex;
	}

      delete s_plistQEVerticesFree;
      s_plistQEVerticesFree = NULL;
      s_uiNewedQEVertices = 0;
    }

  // Clean up our free list of QE Edges
  ASSERT( s_uiAllocatedQEEdges == 0 );
  if ( s_plistQEEdgesFree != NULL )
    {
      QEEdge *pQEEdge;
      
      for ( pQEEdge = s_plistQEEdgesFree->RemoveFirst();
	    pQEEdge != NULL;
	    pQEEdge = s_plistQEEdgesFree->RemoveFirst() )
	{
	  delete pQEEdge;
	}

      delete s_plistQEEdgesFree;
      s_plistQEEdgesFree = NULL;
      s_uiNewedQEEdges = 0;
    }

  // Clean up our free list of QE Wedges
  ASSERT( s_uiAllocatedQEWedges == 0 );
  if ( s_plistQEWedgesFree != NULL )
    {
      QEWedge *pQEWedge;
      
      for ( pQEWedge = s_plistQEWedgesFree->RemoveFirst();
	    pQEWedge != NULL;
	    pQEWedge = s_plistQEWedgesFree->RemoveFirst() )
	{
	  delete pQEWedge;
	}

      delete s_plistQEWedgesFree;
      s_plistQEWedgesFree = NULL;
      s_uiNewedQEWedges = 0;
    }

  // Clean up our free list of Facets
  ASSERT( s_uiAllocatedQEFacets == 0 );
  if ( s_plistQEFacetsFree != NULL )
    {
      QEFacet *pQEFacet;
      
      for ( pQEFacet = s_plistQEFacetsFree->RemoveFirst();
	    pQEFacet != NULL;
	    pQEFacet = s_plistQEFacetsFree->RemoveFirst() )
	{
	  delete pQEFacet;
	}

      delete s_plistQEFacetsFree;
      s_plistQEFacetsFree = NULL;
      s_uiNewedQEFacets = 0;
    }

  // Clean up our free list of Facets
  ASSERT( s_uiAllocatedQEMeshs == 0 );
  if ( s_plistQEMeshsFree != NULL )
    {
      QEMesh *pQEMesh;
      
      for ( pQEMesh = s_plistQEMeshsFree->RemoveFirst();
	    pQEMesh != NULL;
	    pQEMesh = s_plistQEMeshsFree->RemoveFirst() )
	{
	  delete pQEMesh;
	}

      delete s_plistQEMeshsFree;
      s_plistQEMeshsFree = NULL;
      s_uiNewedQEMeshs = 0;
    }
}      

QEEdgeCollapse *AllocateQEEdgeCollapse()
{
  QEEdgeCollapse *pQEEdgeCollapse;

  if ( s_uiQEObjectCount == 0 )
    {
      InitQEMemory();
    }
  s_uiQEObjectCount++;

  ASSERT( s_plistQEEdgeCollapsesFree != NULL );

  pQEEdgeCollapse = s_plistQEEdgeCollapsesFree->RemoveFirst();
  if ( pQEEdgeCollapse == NULL )
    {
      ASSERT( s_uiNewedQEEdgeCollapses == s_uiAllocatedQEEdgeCollapses );
      s_uiNewedQEEdgeCollapses++;
      pQEEdgeCollapse = new QEEdgeCollapse;
      ASSERT( pQEEdgeCollapse != NULL );
    }
  ASSERT( s_uiAllocatedQEEdgeCollapses < s_uiNewedQEEdgeCollapses );

  s_uiAllocatedQEEdgeCollapses++;

  return pQEEdgeCollapse;
}

VOID DeallocateQEEdgeCollapse(QEEdgeCollapse *pQEEdgeCollapse)
{
  ASSERT( pQEEdgeCollapse != NULL );
  ASSERT( s_plistQEEdgeCollapsesFree != NULL );
  
  ASSERT( s_uiAllocatedQEEdgeCollapses > 0 );

  pQEEdgeCollapse->Uninit();
  s_plistQEEdgeCollapsesFree->InsertFirst(pQEEdgeCollapse);
  s_uiAllocatedQEEdgeCollapses--;
  ASSERT( s_uiAllocatedQEEdgeCollapses < s_uiNewedQEEdgeCollapses );

  s_uiQEObjectCount--;
  if ( s_uiQEObjectCount == 0 )
    {
      UninitQEMemory();
    }
}

QEEdgeNode *AllocateQEEdgeNode()
{
  QEEdgeNode *pQEEdgeNode;

  if ( s_uiQEObjectCount == 0 )
    {
      InitQEMemory();
    }
  s_uiQEObjectCount++;

  ASSERT( s_plistQEEdgeNodesFree != NULL );

  pQEEdgeNode = s_plistQEEdgeNodesFree->RemoveFirst();
  if ( pQEEdgeNode == NULL )
    {
      ASSERT( s_uiNewedQEEdgeNodes == s_uiAllocatedQEEdgeNodes );
      s_uiNewedQEEdgeNodes++;
      pQEEdgeNode = new QEEdgeNode;
      ASSERT( pQEEdgeNode != NULL );
    }
  ASSERT( s_uiAllocatedQEEdgeNodes < s_uiNewedQEEdgeNodes );

  s_uiAllocatedQEEdgeNodes++;

  return pQEEdgeNode;
}

VOID DeallocateQEEdgeNode(QEEdgeNode *pQEEdgeNode)
{
  ASSERT( pQEEdgeNode != NULL );
  ASSERT( s_plistQEEdgeNodesFree != NULL );
  
  ASSERT( s_uiAllocatedQEEdgeNodes > 0 );

  pQEEdgeNode->Uninit();
  s_plistQEEdgeNodesFree->InsertFirst(pQEEdgeNode);
  s_uiAllocatedQEEdgeNodes--;
  ASSERT( s_uiAllocatedQEEdgeNodes < s_uiNewedQEEdgeNodes );

  s_uiQEObjectCount--;
  if ( s_uiQEObjectCount == 0 )
    {
      UninitQEMemory();
    }
}

QEFacetNode *AllocateQEFacetNode()
{
  QEFacetNode *pQEFacetNode;

  if ( s_uiQEObjectCount == 0 )
    {
      InitQEMemory();
    }
  s_uiQEObjectCount++;

  ASSERT( s_plistQEFacetNodesFree != NULL );

  pQEFacetNode = s_plistQEFacetNodesFree->RemoveFirst();
  if ( pQEFacetNode == NULL )
    {
      ASSERT( s_uiNewedQEFacetNodes == s_uiAllocatedQEFacetNodes );
      s_uiNewedQEFacetNodes++;
      pQEFacetNode = new QEFacetNode;
      ASSERT( pQEFacetNode != NULL );
    }
  ASSERT( s_uiAllocatedQEFacetNodes < s_uiNewedQEFacetNodes );

  s_uiAllocatedQEFacetNodes++;

  return pQEFacetNode;
}

VOID DeallocateQEFacetNode(QEFacetNode *pQEFacetNode)
{
  ASSERT( pQEFacetNode != NULL );
  ASSERT( s_plistQEFacetNodesFree != NULL );
  
  ASSERT( s_uiAllocatedQEFacetNodes > 0 );

  pQEFacetNode->Uninit();
  s_plistQEFacetNodesFree->InsertFirst(pQEFacetNode);
  s_uiAllocatedQEFacetNodes--;
  ASSERT( s_uiAllocatedQEFacetNodes < s_uiNewedQEFacetNodes );

  s_uiQEObjectCount--;
  if ( s_uiQEObjectCount == 0 )
    {
      UninitQEMemory();
    }
}

QEVertex *AllocateQEVertex()
{
  QEVertex *pQEVertex;

  if ( s_uiQEObjectCount == 0 )
    {
      InitQEMemory();
    }
  s_uiQEObjectCount++;

  ASSERT( s_plistQEVerticesFree != NULL );

  pQEVertex = s_plistQEVerticesFree->RemoveFirst();
  if ( pQEVertex == NULL )
    {
      ASSERT( s_uiNewedQEVertices == s_uiAllocatedQEVertices );
      s_uiNewedQEVertices++;
      pQEVertex = new QEVertex;
      ASSERT( pQEVertex != NULL );
    }
  ASSERT( s_uiAllocatedQEVertices < s_uiNewedQEVertices );

  s_uiAllocatedQEVertices++;

  return pQEVertex;
}

VOID DeallocateQEVertex(QEVertex *pQEVertex)
{
  ASSERT( pQEVertex != NULL );
  ASSERT( s_plistQEVerticesFree != NULL );
  
  ASSERT( s_uiAllocatedQEVertices > 0 );

  pQEVertex->Uninit();
  s_plistQEVerticesFree->InsertFirst(pQEVertex);
  s_uiAllocatedQEVertices--;
  ASSERT( s_uiAllocatedQEVertices < s_uiNewedQEVertices );

  s_uiQEObjectCount--;
  if ( s_uiQEObjectCount == 0 )
    {
      UninitQEMemory();
    }
}

QEEdge *AllocateQEEdge()
{
  QEEdge *pQEEdge;

  if ( s_uiQEObjectCount == 0 )
    {
      InitQEMemory();
    }
  s_uiQEObjectCount++;

  ASSERT( s_plistQEEdgesFree != NULL );

  pQEEdge = s_plistQEEdgesFree->RemoveFirst();
  if ( pQEEdge == NULL )
    {
      ASSERT( s_uiNewedQEEdges == s_uiAllocatedQEEdges );
      s_uiNewedQEEdges++;
      pQEEdge = new QEEdge;
      ASSERT( pQEEdge != NULL );
    }
  ASSERT( s_uiAllocatedQEEdges < s_uiNewedQEEdges );

  s_uiAllocatedQEEdges++;

  return pQEEdge;
}

VOID DeallocateQEEdge(QEEdge *pQEEdge)
{
  ASSERT( pQEEdge != NULL );
  ASSERT( s_plistQEEdgesFree != NULL );
  
  ASSERT( s_uiAllocatedQEEdges > 0 );

  pQEEdge->Uninit();
  s_plistQEEdgesFree->InsertFirst(pQEEdge);
  s_uiAllocatedQEEdges--;
  ASSERT( s_uiAllocatedQEEdges < s_uiNewedQEEdges );

  s_uiQEObjectCount--;
  if ( s_uiQEObjectCount == 0 )
    {
      UninitQEMemory();
    }
}

QEWedge *AllocateQEWedge()
{
  QEWedge *pQEWedge;

  if ( s_uiQEObjectCount == 0 )
    {
      InitQEMemory();
    }
  s_uiQEObjectCount++;

  ASSERT( s_plistQEWedgesFree != NULL );

  pQEWedge = s_plistQEWedgesFree->RemoveFirst();
  if ( pQEWedge == NULL )
    {
      ASSERT( s_uiNewedQEWedges == s_uiAllocatedQEWedges );
      s_uiNewedQEWedges++;
      pQEWedge = new QEWedge;
      ASSERT( pQEWedge != NULL );
    }
  ASSERT( s_uiAllocatedQEWedges < s_uiNewedQEWedges );

  s_uiAllocatedQEWedges++;

  return pQEWedge;
}

VOID DeallocateQEWedge(QEWedge *pQEWedge)
{
  ASSERT( pQEWedge != NULL );
  ASSERT( s_plistQEWedgesFree != NULL );
  
  ASSERT( s_uiAllocatedQEWedges > 0 );

  pQEWedge->Uninit();
  s_plistQEWedgesFree->InsertFirst(pQEWedge);
  s_uiAllocatedQEWedges--;
  ASSERT( s_uiAllocatedQEWedges < s_uiNewedQEWedges );

  s_uiQEObjectCount--;
  if ( s_uiQEObjectCount == 0 )
    {
      UninitQEMemory();
    }
}

QEFacet *AllocateQEFacet()
{
  QEFacet *pQEFacet;

  if ( s_uiQEObjectCount == 0 )
    {
      InitQEMemory();
    }
  s_uiQEObjectCount++;

  ASSERT( s_plistQEFacetsFree != NULL );

  pQEFacet = s_plistQEFacetsFree->RemoveFirst();
  if ( pQEFacet == NULL )
    {
      ASSERT( s_uiNewedQEFacets == s_uiAllocatedQEFacets );
      s_uiNewedQEFacets++;
      pQEFacet = new QEFacet;
      ASSERT( pQEFacet != NULL );
    }
  ASSERT( s_uiAllocatedQEFacets < s_uiNewedQEFacets );

  s_uiAllocatedQEFacets++;

  return pQEFacet;
}

VOID DeallocateQEFacet(QEFacet *pQEFacet)
{
  ASSERT( pQEFacet != NULL );
/*
  switch ( pQEFacet->GetType() )
    {
    case QEFacet::ftFacet:
      break;
    case QEFacet::ftTriangle:
      DeallocateQETriangle((QETriangle*)pQEFacet);
      return;
      break;
    default:
      ASSERT( FALSE );
      break;
    }
*/
  ASSERT( s_plistQEFacetsFree != NULL );
  
  ASSERT( s_uiAllocatedQEFacets > 0 );

  pQEFacet->Uninit();
  s_plistQEFacetsFree->InsertFirst(pQEFacet);
  s_uiAllocatedQEFacets--;
  ASSERT( s_uiAllocatedQEFacets < s_uiNewedQEFacets );

  s_uiQEObjectCount--;
  if ( s_uiQEObjectCount == 0 )
    {
      UninitQEMemory();
    }
}

QEMesh *AllocateQEMesh()
{
  QEMesh *pQEMesh;

  if ( s_uiQEObjectCount == 0 )
    {
      InitQEMemory();
    }
  s_uiQEObjectCount++;

  ASSERT( s_plistQEMeshsFree != NULL );

  pQEMesh = s_plistQEMeshsFree->RemoveFirst();
  if ( pQEMesh == NULL )
    {
      ASSERT( s_uiNewedQEMeshs == s_uiAllocatedQEMeshs );
      s_uiNewedQEMeshs++;
      pQEMesh = new QEMesh;
      ASSERT( pQEMesh != NULL );
    }
  ASSERT( s_uiAllocatedQEMeshs < s_uiNewedQEMeshs );

  s_uiAllocatedQEMeshs++;

  return pQEMesh;
}

VOID DeallocateQEMesh(QEMesh *pQEMesh)
{
  ASSERT( pQEMesh != NULL );
  ASSERT( s_plistQEMeshsFree != NULL );
  
  ASSERT( s_uiAllocatedQEMeshs > 0 );

  pQEMesh->Uninit();
  s_plistQEMeshsFree->InsertFirst(pQEMesh);
  s_uiAllocatedQEMeshs--;
  ASSERT( s_uiAllocatedQEMeshs < s_uiNewedQEMeshs );

  s_uiQEObjectCount--;
  if ( s_uiQEObjectCount == 0 )
    {
      UninitQEMemory();
    }
}
