#ifndef __QEMESH_H__
#define __QEMESH_H__

#include "qe.h"
#include "qemem.h"

#include "qevertex.h"
#include "qewedge.h"
#include "qeedge.h"
#include "qefacet.h"

#include "apm2d.h"

class QEEdgeCollapse;
class QEEdgeNode;

class QEMeshIter;

//////////////////////////////////////////////////////////////////////
// QEMesh Class
//

class QEMesh : public QE
{
//////////////////////////////////////////////////
// Friends
//
public:
  friend class QEMeshIter;

//////////////////////////////////////////////////
// Enumerated Types
//
public:
/*
  enum SubdivisionType { stCornerCutting, stCornerRounding, stTriangle, 
			 stDooSabin, stCatmullClark, 
			 stTesselate, stLoop, stButterfly, stZorin, stSphere, 
			 stSize }; 
  enum DelaunayType { dtIncremental, dtIncrementalSlow, dtIncrementalNoBound, 
		      dtDivideAndConquer, dtDivideAndConquerAlternating, 
		      dtConvexHull, dtSize };
  enum VoronoiType { vtDelaunay, vtDivideAndConquer, vtDivideAndConquerAlternating, 
                     vtCones, vtPlanes, vtSize };
  enum OffsetType { otStart, otSplit, otClose, otConnect, otSize };
  enum AccuracyType { atFloatNaive, atFloat, atExact, atSize };

  enum ErrorMetricType { emtBestFitPlane, emtLimitDistance, emtSize };
  */
//////////////////////////////////////////////////
// Initialization
//
public:
  QEMesh();
  virtual ~QEMesh();
  virtual VOID Init();
  virtual VOID Uninit();

  virtual BOOL Initialized();

  virtual BOOL Valid();

  QEMesh *Copy();
  QEMesh *CopyVoronoiPolygon();
  QEMesh *CopyVoronoiMountainMesh();
  QEMesh *CreateDual();
  QEMesh *CreateMirror();
  VOID Transform(const Matrix &m, const Matrix &mInv);
  VOID CopyBoundaryEdges(QEMesh *pmeSrc);
  VOID ExtractComponents(CObjectList<QEMesh> *plistQEMesh);
protected:
  VOID ExtractComponent(QEMesh *pmeComponent, 
			UINT_32 &uiVertexID, UINT_32 &uiFacetID, 
			QEFacet *pfaCurr);
public:
  VOID Preprocess();

  VOID RemoveAllVertices(CObjectList<QEVertex> *plistQEVertex);
  // Add an externally allocated vertex
  VOID AddVertex(QEVertex *pQEVertex);
  // Allocate and add a vertex locally
  QEVertex *CreateVertex(const Point &pt);
  QEFacet *CreateFacet();
  QEEdge *CreateEdge();

protected:
  // Allocate and add elements locally
  QEEdgeCollapse *CreateEdgeCollapse();
  QEEdgeNode *CreateEdgeNode(const Point &ptFrom, const Point &ptTo);
  QEFacetNode *CreateFacetNode();

  // Add externally allocated elements
  VOID AddWedge(QEWedge *pwe);
  VOID AddEdge(QEEdge *pe);
  VOID AddFacet(QEFacet *pfa);
  VOID AddEdgeCollapse(QEEdgeCollapse *pec);
  VOID AddEdgeNode(QEEdgeNode *pen);
  VOID AddFacetNode(QEFacetNode *pfn);

  // Remove and deallocate elements
  VOID DestroyVertex(QEVertex *pvx);
  VOID DestroyWedge(QEWedge *pwe);
  VOID DestroyEdge(QEEdge *pe);
  VOID DestroyFacet(QEFacet *pfa);
  VOID DestroyEdgeCollapse(QEEdgeCollapse *pec);
  VOID DestroyEdgeNode(QEEdgeNode *pen);
  VOID DestroyFacetNode(QEFacetNode *pfn);

  VOID CollapseEdge(QEEdge *pe); // Destroy the edge and its vertex

  // Remove elements but do not deallocate them
  VOID RemoveVertex(QEVertex *pvx);
  VOID RemoveWedge(QEWedge *pwe);
  VOID RemoveEdge(QEEdge *pe);
  VOID RemoveFacet(QEFacet *pfa);
  VOID RemoveEdgeCollapse(QEEdgeCollapse *pec);
  VOID RemoveEdgeNode(QEEdgeNode *pen);
  VOID RemoveFacetNode(QEFacetNode *pfn);

  // Contour building routines
  QEEdge *CreateContour(QEFacet *pfa, QEVertex *pvx);
  QEEdge *AddContourVertex(QEEdge *peContour, QEVertex *pvx);


//////////////////////////////////////////////////
// Topological Functions
//
public:
  static VOID RandomizeVertexList(CObjectList<QEVertex> *plistQEVertex);
  static VOID RandomizeArray(UINT_32 uiSize, VOID **ppv);
  static VOID Swap(VOID **ppv0, VOID **ppv1);
  VOID Splice(UINT_32 uiDual, QEEdge *pe0, QEEdge *pe1);

protected:
  QEEdge *InsertDiagonal(QEEdge *pe0, QEEdge *pe1);
  VOID InsertDiagonal(QEEdge *pe0, QEEdge *pe1, QEEdge *pe);
  VOID RemoveDiagonal(QEEdge *pe);

//////////////////////////////////////////////////
// Triangulation Functions
//
public:
  VOID CreateContours(CListInt *plistContours);
  VOID CreateTriangulatedPolygon(CListInt *plistContours);
protected:
  VOID TriangulatePolygon(CListInt *plistEdges);
public:
  VOID TriangulateYMonotone(QEEdge *peContour, CListInt *plistFlipCandidateEdges);
  VOID ConstrainedDelaunayEdgeFlipping(CListInt *plistFlipCandidateEdges);
  VOID CreateVoronoiPolygon(VoronoiType eType, QEMesh *pmeCDT, BOOL bCleanUpContour);
  VOID CreateVoronoiPolygonHeld(CObjectList<QEVertex> *plistQEVertex, BOOL bCleanUpContour);

  static FLOAT CalculateContourArea(CObjectList<QEVertex> *plistQEVertex);
  VOID MergeVoronoiPolygonsHeld(QEMesh *pmeVoronoi);
protected:
  QEFacet *PointLocateVoronoiPolygonHeld(const Point &pt);
  QEFacet *PointLocateVoronoiPolygonHeldNaive(const Point &pt);
  QEFacet *PointLocateVoronoiPolygonHeldGradientDescent(const Point &pt);
  VOID MergeVoronoiPolygonsHeld(QEEdge *peLeftEnd, QEEdge *&peLeft,
				QEEdge *&peBisector,
				QEEdge *&peRight, QEEdge *peRightEnd, FLOAT fW);
public:
  VOID CreateVoronoiPolygonShewchuk(QEMesh *pmeCDT, BOOL bCleanUpContour);
  VOID CreateVoronoiPolygonShewchuk(CObjectList<QEVertex> *plistQEVertex, BOOL bCleanUpContour);
  VOID MergeVoronoiPolygonsShewchuk(QEMesh *pmeVoronoi);
protected:
  VOID MergeVoronoiPolygonsDegenerateShewchuk(QEVertex *pvxBase, 
					      QEMesh *pmeVoronoi, QEVertex *pvxVoronoi);
  VOID MergeVoronoiPolygonsVertexZoneShewchuk(QEEdge *peLeftUp, QEEdge *peLeftDown, 
					      QEEdge *peRightUp, QEEdge *peRightDown);
  VOID VerifyVoronoiPolygonsShewchuk();
public:
  VOID WriteDebugVoronoiPolygonShewchuk(FILE *pfOutput);

protected:
  VOID CreateVoronoiPolygonShewchuk(UINT_32 uiSize, QEEdge **ppeContourLeft, 
				    QEEdge **ppeContourRight, FLOAT fW, BOOL bConvex, BOOL bInside);
  VOID MergeVoronoiPolygonShewchuk(QEEdge *peLeftEnd, QEEdge *peLeftMiddle, 
				   QEEdge *peRightMiddle, QEEdge *peRightEnd, FLOAT fW);
  static VOID CalculateInscribedCircleCenterSSSShewchuk(Point *pptResults, UINT_32 &uiResults, 
							QEFacet *pfa0, 
							QEFacet *pfa1, 
							QEFacet *pfa2, 
							BOOL bRadiusInZ);
  static BOOL CounterclockwiseSSSShewchuk(QEFacet *pfa0, QEFacet *pfa1, QEFacet *pfa2);
  static APM2D::InscribedCircleCenterType InscribedCircleCenterSSSShewchuk(Point &ptCenter, 
									   QEFacet *pfa0, 
									   QEFacet *pfa1, 
									   QEFacet *pfa2, 
									   BOOL bRadiusInZ);

  static BOOL ReplaceCircleCenterShewchuk(APM2D::InscribedCircleCenterType eICCT,
					  QEFacet *pfa0, QEFacet *pfa1, QEFacet *pfa2,
					  Point &ptCenter,
					  QEFacet *pfa3);
  static APM2D::CircleTestType InCircleSSS_SShewchuk(APM2D::InscribedCircleCenterType eICCT,
						     QEFacet *pfa0, 
						     QEFacet *pfa1, 
						     QEFacet *pfa2, 
						     Point &ptCenter,
						     QEFacet *pfa3);
  static APM2D::CircleTestType InCircleSSS_VShewchuk(APM2D::InscribedCircleCenterType eICCT,
						     QEFacet *pfa0, 
						     QEFacet *pfa1, 
						     QEFacet *pfa2, 
						     Point &ptCenter,
						     QEFacet *pfa3, QEVertex *pvx3);
  static APM2D::CircleTestType InCircleSSS_EShewchuk(APM2D::InscribedCircleCenterType eICCT,
						     QEFacet *pfa0, 
						     QEFacet *pfa1, 
						     QEFacet *pfa2, 
						     Point &ptCenter,
						     QEFacet *pfa3, QEEdge *pe3);

  static APM2D::EdgeZoneType EdgeZone(const Point &pt, QEEdge *pe, 
				      QEVertex **ppvxEndpoint);
  static INT_32 DistanceCompareS_SS(QEFacet *pfaTest, 
				    QEVertex **ppvxEndpointTest_0,
				    QEVertex **ppvxEndpointTest_1,
				    QEFacet *pfa0, QEVertex **ppvxEndpoint0,
				    QEFacet *pfa1, QEVertex **ppvxEndpoint1);
  static INT_32 DistanceCompareV_SS(QEVertex *pvxTest,
				    QEFacet *pfa0, QEVertex **ppvxEndpoint0,
				    QEFacet *pfa1, QEVertex **ppvxEndpoint1);
  static INT_32 DistanceCompareE_SS(QEEdge *peTest,
				    QEVertex **ppvxEndpointTest_0,
				    QEVertex **ppvxEndpointTest_1,
				    QEFacet *pfa0, QEVertex **ppvxEndpoint0,
				    QEFacet *pfa1, QEVertex **ppvxEndpoint1);

  QEFacet *PointLocateVoronoiPolygonShewchuk(QEFacet *pfaTest, QEVertex *pvxTest);
  QEFacet *PointLocateVoronoiPolygonShewchuk(QEFacet *pfaTest, QEEdge *peStart);
  QEFacet *PointLocateVoronoiPolygonShewchukNaive(QEFacet *pfaTest);
  QEFacet *PointLocateVoronoiPolygonShewchukGradientDescent(QEFacet *pfaTest, 
							    QEVertex *pvxTest);
  QEFacet *PointLocateVoronoiPolygonShewchukGradientDescent(QEFacet *pfaTest, 
							    QEEdge *peStart);
  INT_32 SideOfEdgeShewchuk(QEEdge *pe, QEFacet *pfaTest); 
  VOID MergeVoronoiPolygonsShewchuk(QEEdge *&peLeftEnd, QEEdge *&peLeft,
				    QEEdge *&peBisector, QEEdge *&peBisectorEnd,
				    QEEdge *&peRight, QEEdge *&peRightEnd, FLOAT fW);


public:
  VOID ZMonotoneVoronoiPolygon();
  VOID CreateOffsetPolygon(QEMesh *pmeCDT, QEMesh *pmeVoronoi, FLOAT fOffset);
  VOID CreateMillingContour(QEMesh *pmeCDT, QEMesh *pmeVoronoi, FLOAT fToolRadius);
  VOID DisplayMinChannelVertices();

  // This does not use the vertices
  VOID CreateTerrain(DelaunayType eType, CObjectList<QEVertex> *plistQEVertex);
  // This uses the vertices
  VOID CreateDelaunay(DelaunayType eType, CObjectList<QEVertex> *plistQEVertex);
  VOID CreateVoronoi(VoronoiType eType, QEMesh *pmeDelaunay);

  VOID CreateTetrahedron(const Point &pt0, const Point &pt1, const Point &pt2, 
			 const Point &pt3);
  VOID CreateTetrahedron(QEVertex *pvx0, QEVertex *pvx1, 
			 QEVertex *pvx2, QEVertex *pvx3);
  UINT_32 CreateConvexHull(CObjectList<QEVertex> *plistQEVertex);
  VOID PartitionVertices(CHeap<QEFacet*, FLOAT> *pheapQEFacet,
			 CObjectList<QEFacet> *plistQEFacet,
			 CObjectList<QEVertex> *plistQEVertex);
  VOID CalculateHorizon(QEEdge *peContour, QEVertex *pvxEye, 
			CHeap<QEFacet*, FLOAT> *pheapQEFacet, 
			CObjectList<QEVertex> *plistQEVertexUnassigned, 
			CObjectList<QEEdge> *plistQEEdgeRemoved,
			BOOL bStart);

  QEMesh *CreateOffsetMeshAngular(FLOAT fOffset);
  QEMesh *CreateOffsetMeshRounded(FLOAT fOffset);
  QEMesh *CreateOffsetMeshPerforated(FLOAT fOffsetNormal, FLOAT fOffsetTangential);

protected:
   QEEdge* CreateOffsetMeshBisector(QEMesh* new_mesh, QEEdge* old_edge, float offset);

protected:
  VOID InsertVertex(QEFacet *pfa, QEVertex *pvx);
  VOID InsertVertex(QEFacet *pfa, QEVertex *pvx, CListInt *plistQEEdgeQueue);
  VOID InsertVertexUpdateSearch(QEFacet *pfa, QEVertex *pvx, CListInt *plistQEEdgeQueue);
  VOID InsertVertex(QEEdge *pe, QEVertex *pvx);
  VOID InsertVertex(QEEdge *pe, QEVertex *pvx, CListInt *plistQEEdgeQueue);
  VOID InsertVertexUpdateSearch(QEEdge *pe, QEVertex *pvx, CListInt *plistQEEdgeQueue);
  VOID DelaunayFlip(QEEdge *pe);
  VOID DelaunayFlip(QEEdge *pe, CListInt *plistQEEdgeQueue);
  VOID DelaunayFlipIncremental(QEEdge *pe, CListInt *plistQEEdgeQueue);
  VOID DelaunayFlipUpdateSearch(QEEdge *pe, CListInt *plistQEEdgeQueue);

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  static Point CalculateCircumcenter(const Point &pt0, const Point &pt1, const Point &pt2, BOOL bRadiusInZ = FALSE);
  static Point CalculateInscribedCircumcenter(const Point &pt0, const Point &pt1, const Point &pt2, BOOL bRadiusInZ = FALSE);

  static BOOL PointInEdgeZone(const Point &pt, const Point &ptE0, const Point &ptE1);
  static FLOAT DistanceFromEdgeZone(const Point &pt, const Point &ptE0, const Point &ptE1);
  static Point CalculatePointProjectionOnLine(const Point &pt, 
					      const Point &ptE0, const Point &ptE1, 
					      BOOL bRadiusInZ = FALSE);
  static VOID CalculatePointProjectionOnLineSegment(Point &ptResult, UINT_32 &uiResults,
						    const Point &pt, 
						    const Point &ptE0, const Point &ptE1, 
						    BOOL bRadiusInZ = FALSE);
  // General Voronoi
  static Vector CalculateBisectorDirection(QEFacet *pfa0, BOOL bInsideContour0, 
					   QEFacet *pfa1, BOOL bInsideContour1, 
					   BOOL bBelow);
  static BOOL Counterclockwise(QEFacet *pfa0, QEFacet *pfa1, QEFacet *pfa2);
  static BOOL Counterclockwise(QEFacet *pfa0, BOOL bInsideContour0, 
			       QEFacet *pfa1, BOOL bInsideContour1, 
			       QEFacet *pfa2, BOOL bInsideContour2);
  static BOOL Counterclockwise(QEVertex *pppvx[3][2], 
			       Point pppt[3][2], UINT_32 uiEdges);
  static FLOAT CalculateDistance(QEFacet *pfa, const Point &pt, QEVertex **ppvxEndpoint);
  static VOID CalculateVoronoiPoint(Point *pptResults, UINT_32 &uiResults, 
				    QEFacet *pfa0, QEFacet *pfa1, QEFacet *pfa2,
				    BOOL bRadiusInZ);
  static VOID CalculateVoronoiPoint(Point *pptResults, UINT_32 &uiResults, 
				    QEFacet *pfa0, BOOL bInsideContour0, 
				    QEFacet *pfa1, BOOL bInsideContour1, 
				    QEFacet *pfa2, BOOL bInsideContour2, 
				    BOOL bRadiusInZ);
  // Finds the closest point to V0, V1, and V2
  static VOID CalculateVoronoiPointVVV(Point *pptResults, UINT_32 &uiResults,
				       const Point &ptV0, 
				       const Point &ptV1, 
				       const Point &ptV2, 
				       BOOL bRadiusInZ = FALSE);
  // Finds the closest point to V0, E1_0, and E1_0-E1_1
  static VOID CalculateVoronoiPointVE(Point *pptResults, UINT_32 &uiResults, 
				       const Point &ptV0, 
				       const Point &ptE1_0, const Point &ptE1_1, 
				       BOOL bRadiusInZ = FALSE);
  // Finds the closest point to V0, V1, and E1_0-E1_1
  static VOID CalculateVoronoiPointVVE(Point *pptResults, UINT_32 &uiResults, 
				       const Point &ptV0_0, const Point &ptV0, const Point &ptV0_1, 
				       const Point &ptV1_0, const Point &ptV1, const Point &ptV1_1, 
				       const Point &ptE2_0, const Point &ptE2_1, 
				       BOOL bRadiusInZ = FALSE);
  // Finds the closest point to V0, E0_0-E0_1, and E1_0-E1_1
  static VOID CalculateVoronoiPointVEE(Point *pptResults, UINT_32 &uiResults, 
				       const Point &ptV0_0, const Point &ptV0, const Point &ptV0_1, 
				       const Point &ptE1_0, const Point &ptE1_1, 
				       const Point &ptE2_0, const Point &ptE2_1, 
				       BOOL bRadiusInZ = FALSE);
  // Finds the closest point to E0_0, E0_0-E0_1, and E1_0-E1_1
  static VOID CalculateVoronoiPointEE(Point *pptResults, UINT_32 &uiResults, 
				      const Point &ptE0_0, const Point &ptE0_1, 
				      const Point &ptE1_0, const Point &ptE1_1, 
				      BOOL bRadiusInZ = FALSE, 
				      BOOL bCounterclockwise = TRUE);
  // Finds the closest point to E0_0-E1_0, E1_0-E1_1, and E2_0-E2_1
  static VOID CalculateVoronoiPointEvEE(Point *pptResults, UINT_32 &uiResults,
					const Point &ptE0_0,
					const Point &ptE1_0, const Point &ptE1_1, 
					const Point &ptE2_0, const Point &ptE2_1, 
					BOOL bRadiusInZ = FALSE);
  // Finds the closest point to E0_0-E0_1, E1_0-E1_1, and E2_0-E2_1
  static VOID CalculateVoronoiPointEEE(Point *pptResults, UINT_32 &uiResults, 
				       const Point &ptE0_0, const Point &ptE0_1, 
				       const Point &ptE1_0, const Point &ptE1_1, 
				       const Point &ptE2_0, const Point &ptE2_1, 
				       BOOL bRadiusInZ = FALSE);

  static VOID CalculateOffsetPoint(Point *pptResults, UINT_32 &uiResults, 
				   QEEdge *peVoronoi, QEFacet *pfa0, QEFacet *pfa1, 
				   FLOAT fOffset);

protected:
  virtual VOID CalculateBound();

  BOOL ConstrainedDelaunayEdge(QEEdge *pe);
  BOOL DelaunayEdge(QEEdge *pe);
  BOOL InCircumcircle(const Point &ptTri0, const Point &ptTri1, const Point &ptTri2, 
		      const Point &pt);

public:
  static BOOL Counterclockwise(const Point &ptTri0, const Point &ptTri1, 
			       const Point &ptTri2);
  static DOUBLE CounterclockwiseValue(const Point &ptTri0, const Point &ptTri1, 
				      const Point &ptTri2);

protected:
  BOOL RightOfEdge(QEEdge *pe, const Point &pt);
  BOOL LeftOfEdge(QEEdge *pe, const Point &pt);
  QEFacet *FindFacet(const Point &pt);
  QEEdge *FindFacetEdge(const Point &pt);

  VOID TriangulateDelaunayIncremental();
  VOID TriangulateDelaunayIncrementalSlow();
  VOID RemoveDelaunayBoundingBox();
  VOID TriangulateDelaunayIncrementalNoBound();
  VOID TriangulateDelaunayDivideAndConquer();
  VOID TriangulateDelaunayDivideAndConquer(UINT_32 uiSize, QEVertex **ppvx, 
					   QEEdge **ppeLeftAndRight);
  VOID TriangulateDelaunayDivideAndConquerAlternating();
  VOID TriangulateDelaunayDivideAndConquerAlternating(QEVertex::ComparisonType eDirection, 
						      UINT_32 uiSize, QEVertex **ppvx, 
						      QEEdge *pppeLeftAndRight[2][2]);
  VOID TriangulateDelaunayConvexHull();

  VOID CreateVoronoiDelaunay(QEMesh *pmeDelaunay);
  VOID CreateVoronoiDivideAndConquer(QEMesh *pmeDelaunay);
  VOID CreateVoronoiDivideAndConquer(UINT_32 uiSize, QEVertex **ppvx, 
				     QEEdge **ppeLeftAndRight);
  VOID CreateVoronoiDivideAndConquerAlternating(QEMesh *pmeDelaunay);

  VOID SortAlternating(QEVertex::ComparisonType eDirection, 
		       UINT_32 uiSize, QEVertex **ppvx);
  VOID QuickSelect(UINT_32 uiSize, QEVertex **ppvx, UINT_32 uiPos);
  UINT_32 QuickPartition(UINT_32 uiSize, QEVertex **ppvx);
  VOID Swap(QEVertex **ppvx0, QEVertex **ppvx1);

  VOID SortAlternatingBubble(QEVertex::ComparisonType eDirection, 
			     UINT_32 uiSize, QEVertex **ppvx);

  VOID CreateVoronoiPolygonChordalAxis(QEMesh *pmeCDT);
  VOID CreateVoronoiPolygonCDT(QEMesh *pmeCDT);
  VOID SubdivideCDTBoundary();
  VOID CreateVoronoiPolygonDivideAndConquer(QEMesh *pmeCDT);
  VOID CreateVoronoiPolygonDivideAndConquer(UINT_32 uiSize, QEEdge **ppeContour, 
					    QEEdge **ppeLeftAndRight);
  VOID CreateVoronoiPolygonHeld(QEMesh *pmeCDT, BOOL bCleanUpContour);
  VOID CreateVoronoiPolygonHeld(UINT_32 uiSize, QEEdge **ppeContourLeft, 
				QEEdge **ppeContourRight, FLOAT fW, BOOL bConvex, BOOL bInside);
  VOID MergeVoronoiPolygonHeld(QEEdge *peLeftEnd, QEEdge *peLeftMiddle, 
			       QEEdge *peRightMiddle, QEEdge *peRightEnd, FLOAT fW);

//////////////////////////////////////////////////
// Subdivision
//
public:
  // Limit Positions and Normals
  VOID CalculateSubdivisionLimitPositionsAndNormalsDefault();
  VOID CalculateSubdivisionLimitPositionsAndNormalsLoop();
  VOID CalculateSubdivisionLimitPositionsAndNormalsSharpLoop();
  VOID CalculateSubdivisionLimitPositionsAndNormalsCatmullClark();
  VOID CalculateSubdivisionLimitPositionsAndNormalsCatmullClarkC2();
  VOID CalculateSubdivisionLimitPositionsAndNormalsSharpCatmullClark();
  VOID CalculateSubdivisionLimitPositionsAndNormalsDooSabin();
  VOID CalculateSubdivisionLimitPositionsAndNormalsSqrt3();

  VOID CalculateSubdivisionSharpVertexTypesSharpLoop();
  VOID CalculateSubdivisionBoundaryEdgesSharpnessesSharpCatmullClark();
  VOID CalculateSubdivisionSharpVertexTypesSharpCatmullClark();

protected:
  VOID CalculateSubdivisionLimitPositionsAndNormals(SubdivisionType eSubdivisionType);
  SharpVertexType CalculateSubdivisionSharpVertexTypeSharpLoop(QEVertex *pvx);
  SharpVertexType CalculateSubdivisionSharpVertexTypeSharpCatmullClark(QEVertex *pvx);

public:
  // Subdivision
  QEMesh *SubdivideCornerCutting();

  QEMesh *SubdivideCornerRoundingTesselate();
  QEMesh *SubdivideCornerRounding();
  QEMesh *SubdivideCornerRoundingLoop(); // More general form of Loop's scheme
  QEMesh *SubdivideCornerRoundingSharpLoop(); // Hoppe and DeRose sharpness scheme 
  QEMesh *SubdivideCornerRoundingButterfly();
  QEMesh *SubdivideCornerRoundingZorin();

  QEMesh *SubdivideTriangulate(); // Turn any non-triangular facets into triangles
  QEMesh *SubdivideTriangleTesselate();
  QEMesh *SubdivideTriangleLoop();
  QEMesh *SubdivideTriangleLoopEdgeRule();
  QEMesh *SubdivideTriangleButterfly();
  QEMesh *SubdivideTriangleZorin();
  QEMesh *SubdivideTriangleSphere();

public:
  QEMesh *SubdivideGeometricTesselate();
  QEMesh *SubdivideGeometricBiarc();
protected:
  QEMesh *SubdivideGeometric(SubdivisionType eSubdivisionType);
  Point CalculateVertexSubdivisionGeometric(SubdivisionType eSubdivisionType, QEVertex *pvx, Vector &vNormal);
  Point CalculateEdgeSubdivisionGeometric(SubdivisionType eSubdivisionType, QEEdge *pe, Vector &vNormal);
  static Point CalculateVertexSubdivisionGeometricTesselate(QEVertex *pvx, Vector &vNormal);
  static Point CalculateEdgeSubdivisionGeometricTesselate(QEEdge *pe, Vector &vNormal);
  static Point CalculateVertexSubdivisionGeometricBiarc(QEVertex *pvx, Vector &vNormal);
  static Point CalculateEdgeSubdivisionGeometricBiarc(QEEdge *pe, Vector &vNormal);

public:
  QEMesh *SubdivideSqrt3();
  QEMesh *SubdivideSqrt3SphereBarycenter();
  QEMesh *SubdivideSqrt3SphereCenter();
  QEMesh *SubdivideSqrt3SphereRadius();
  QEMesh *SubdivideSqrt3SphereCapArea();
  QEMesh *SubdivideSqrt3SphereTriangleArea();
  QEMesh *SubdivideSqrt3SphereTangent();
protected:
  QEMesh *SubdivideSqrt3(SubdivisionType eSubdivisionType);
public:
  QEMesh *SubdivideSqrt4();

  QEMesh *SubdivideDooSabin();
  QEMesh *SubdivideCatmullClark(); // Vertex Centric
  QEMesh *SubdivideSharpCatmullClark(); // Vertex Centric
  QEMesh *SubdivideCatmullClarkFacetCentric(); // Facet Centric
  QEMesh *SubdivideCatmullClarkTesselate();
  QEMesh *SubdivideCatmullClarkInterpolate();
  QEMesh *SubdivideCatmullClarkC2(INT_32 iGen); // Vertex Centric

  QEMesh *SubdivideHeterogeneous(); // Facet Centric
  QEMesh *SubdivideHeterogeneousStamLoop(); // Facet Centric
  QEMesh *SubdivideHeterogeneousOld(); // Facet Centric
  QEMesh *SubdivideHeterogeneousTesselate(); // Facet Centric

  QEMesh *SubdividePentagon(); // Facet Centric
  QEMesh *SubdividePentagonTesselate(); // Facet Centric

  //////////////////////////////////////////////////
  // Selective Schemes

  //////////////////////////////
  // Loop
  QEMesh *SubdivideCornerRoundingSelective(ErrorMetricType eErrorMetric, 
                                           FLOAT fTolerance, UINT_32 uiFacetsMax);
  VOID ReSubdivideCornerRoundingSelective(ErrorMetricType eErrorMetric,
                                          FLOAT fTolerance, UINT_32 uiFacetsMax);
  VOID UnSubdivideCornerRoundingSelective(ErrorMetricType eErrorMetric,
                                          FLOAT fTolerance, UINT_32 uiFacetsMax);

  QEMesh *SubdivideCornerRoundingSelectiveLimit(ErrorMetricType eErrorMetric, 
                                                FLOAT fTolerance, UINT_32 uiFacetsMax);
  VOID ReSubdivideCornerRoundingSelectiveLimit(ErrorMetricType eErrorMetric,
                                               FLOAT fTolerance, UINT_32 uiFacetsMax);
  VOID UnSubdivideCornerRoundingSelectiveLimit(ErrorMetricType eErrorMetric,
                                               FLOAT fTolerance, UINT_32 uiFacetsMax);

  QEMesh *SubdivideCornerRoundingSharpSelectiveLimit(ErrorMetricType eErrorMetric, 
                                                     FLOAT fTolerance, UINT_32 uiFacetsMax);
  VOID PreSubdivideCornerRoundingSharpSelectiveLimit();
  VOID ReSubdivideCornerRoundingSharpSelectiveLimit(ErrorMetricType eErrorMetric,
                                                    FLOAT fTolerance, UINT_32 uiFacetsMax);
  VOID UnSubdivideCornerRoundingSharpSelectiveLimit(ErrorMetricType eErrorMetric,
                                                    FLOAT fTolerance, UINT_32 uiFacetsMax);

#define LOOP_FACET_CENTRIC
#ifdef LOOP_FACET_CENTRIC
  //////////////////////////////
  // Loop
  QEMesh *SubdivideCornerRoundingFacetSharpSelectiveLimit(ErrorMetricType eErrorMetric, 
                                                     FLOAT fTolerance, UINT_32 uiFacetsMax);
  VOID PreSubdivideCornerRoundingFacetSharpSelectiveLimit();
  VOID ReSubdivideCornerRoundingFacetSharpSelectiveLimit(ErrorMetricType eErrorMetric,
                                                    FLOAT fTolerance, UINT_32 uiFacetsMax);
  VOID UnSubdivideCornerRoundingFacetSharpSelectiveLimit(ErrorMetricType eErrorMetric,
                                                    FLOAT fTolerance, UINT_32 uiFacetsMax);
#endif // LOOP_FACET_CENTRIC

  //////////////////////////////
  // Doo-Sabin
  QEMesh *SubdivideDooSabinSelective(ErrorMetricType eErrorMetric,
                                     FLOAT fTolerance, UINT_32 uiFacetsMax);
  VOID ReSubdivideDooSabinSelective(ErrorMetricType eErrorMetric,
                                    FLOAT fTolerance, UINT_32 uiFacetsMax);
  VOID UnSubdivideDooSabinSelective(ErrorMetricType eErrorMetric,
                                    FLOAT fTolerance, UINT_32 uiFacetsMax);

  QEMesh *SubdivideDooSabinSelectiveLimit(ErrorMetricType eErrorMetric,
                                          FLOAT fTolerance, UINT_32 uiFacetsMax);
  VOID ReSubdivideDooSabinSelectiveLimit(ErrorMetricType eErrorMetric,
                                         FLOAT fTolerance, UINT_32 uiFacetsMax);

  //////////////////////////////
  // Catmull-Clark
  QEMesh *SubdivideCatmullClarkSelective(ErrorMetricType eErrorMetric,
                                         FLOAT fTolerance, UINT_32 uiFacetsMax);
  VOID ReSubdivideCatmullClarkSelective(ErrorMetricType eErrorMetric,
                                        FLOAT fTolerance, UINT_32 uiFacetsMax);

  //////////////////////////////
  // Sqrt4
  QEMesh *SubdivideSqrt4Selective(ErrorMetricType eErrorMetric,
                                  FLOAT fTolerance, UINT_32 uiFacetsMax);
  VOID ReSubdivideSqrt4Selective(ErrorMetricType eErrorMetric,
                                 FLOAT fTolerance, UINT_32 uiFacetsMax);

  //////////////////////////////
  // Sqrt3
  QEMesh *SubdivideSqrt3Selective(ErrorMetricType eErrorMetric,
                                  FLOAT fTolerance, UINT_32 uiFacetsMax);
  VOID ReSubdivideSqrt3Selective(ErrorMetricType eErrorMetric,
                                 FLOAT fTolerance, UINT_32 uiFacetsMax);

  QEMesh *SubdivideSqrt3SelectiveFacet(ErrorMetricType eErrorMetric,
                                       FLOAT fTolerance, UINT_32 uiFacetsMax);
  VOID ReSubdivideSqrt3SelectiveFacet(ErrorMetricType eErrorMetric,
                                      FLOAT fTolerance, UINT_32 uiFacetsMax);

protected:
  QEMesh *SubdivideCornerRounding(SubdivisionType eSubdivisionType);
  QEMesh *SubdivideCornerRoundingSharp(SubdivisionType eSubdivisionType);
  QEMesh *SubdivideTriangle(SubdivisionType eSubdivisionType);
  QEMesh *SubdivideCatmullClarkFacetCentric(SubdivisionType eSubdivisionType);
  QEMesh *SubdivideHeterogeneous(SubdivisionType eSubdivisionType);
  QEMesh *SubdividePentagon(SubdivisionType eSubdivisionType);

  Point CalculateVertexSubdivisionCornerRounding(SubdivisionType eSubdivisionType, QEVertex *pvx);
  Point CalculateVertexLimitSubdivisionCornerRounding(SubdivisionType eSubdivisionType, QEVertex *pvx);
  Vector CalculateVertexNormalSubdivisionCornerRounding(SubdivisionType eSubdivisionType, QEVertex *pvx);
  Point CalculateEdgeSubdivisionCornerRounding(SubdivisionType eSubdivisionType, QEEdge *pe);
  Point CalculateVertexSubdivisionCornerRounding(QEVertex *pvx);
  Point CalculateEdgeSubdivisionCornerRounding(QEEdge *pe);
  Point CalculateVertexSubdivisionCornerRoundingButterfly(QEVertex *pvx);
  Point CalculateEdgeSubdivisionCornerRoundingButterfly(QEEdge *pe);
  Point CalculateVertexSubdivisionCornerRoundingZorin(QEVertex *pvx);
  Point CalculateEdgeSubdivisionCornerRoundingZorin(QEEdge *pe);

  Point CalculateVertexSubdivisionTriangle(SubdivisionType eSubdivisionType, QEVertex *pvx);
  Point CalculateEdgeSubdivisionTriangle(SubdivisionType eSubdivisionType, QEEdge *pe);

  static Point CalculateFacetSubdivisionTesselate(QEFacet *pfa);
  static Point CalculateVertexSubdivisionTesselate(QEVertex *pvx);
  static Point CalculateEdgeSubdivisionTesselate(QEEdge *pe);
  static Point CalculateVertexSubdivisionLoop(QEVertex *pvx);
  static Point CalculateVertexLimitSubdivisionLoop(QEVertex *pvx);
  static Vector CalculateVertexNormalSubdivisionLoop(QEVertex *pvx);
  static Point CalculateEdgeSubdivisionLoop(QEEdge *pe);
  static Point CalculateVertexSubdivisionSharpLoop(QEVertex *pvx);
  static Point CalculateVertexLimitSubdivisionSharpLoop(QEVertex *pvx);
  static Vector CalculateVertexNormalSubdivisionSharpLoop(QEVertex *pvx);
public:
  static Vector CalculateVertexNormalSubdivisionSharpLoop(QEEdge *peCrease);
protected:
  static Point CalculateEdgeSubdivisionSharpLoop(QEEdge *pe);
  static Point CalculateVertexSubdivisionButterfly(QEVertex *pvx);
  static Point CalculateEdgeSubdivisionButterfly(QEEdge *pe);
  static Point CalculateVertexSubdivisionZorin(QEVertex *pvx);
  static Point CalculateEdgeSubdivisionZorin(QEEdge *pe);
  static Point CalculateVertexSubdivisionSphere(QEVertex *pvx);
  static Point CalculateEdgeSubdivisionSphere(QEEdge *pe);

  static Point CalculateVertexSubdivisionHermite(QEVertex *pvx);
  static Point CalculateEdgeSubdivisionHermite(QEEdge *pe);

  static Point CalculateVertexSubdivisionLoopEdgeRule(QEVertex *pvx);
  static Point CalculateEdgeSubdivisionLoopEdgeRule(QEEdge *pe);

  static Point CalculateVertexSubdivisionLoopExperimental(QEVertex *pvx);
  static Point CalculateEdgeSubdivisionLoopExperimental(QEEdge *pe);


  Point CalculateFacetSubdivisionSqrt3(SubdivisionType eSubdivisionType, QEFacet *pfa);
  Point CalculateVertexSubdivisionSqrt3(SubdivisionType eSubdivisionType, QEVertex *pvx);

  Point CalculateFacetSubdivisionSqrt3(QEFacet *pfa);
  Point CalculateVertexSubdivisionSqrt3(QEVertex *pvx);
  Point CalculateVertexLimitSubdivisionSqrt3(QEVertex *pvx);
  Vector CalculateVertexNormalSubdivisionSqrt3(QEVertex *pvx);
  Point CalculateFacetSubdivisionSqrt3SphereBarycenter(QEFacet *pfa);
  Point CalculateVertexSubdivisionSqrt3SphereBarycenter(QEVertex *pvx);
  Point CalculateFacetSubdivisionSqrt3SphereCenter(QEFacet *pfa);
  Point CalculateVertexSubdivisionSqrt3SphereCenter(QEVertex *pvx);
  Point CalculateFacetSubdivisionSqrt3SphereRadius(QEFacet *pfa);
  Point CalculateVertexSubdivisionSqrt3SphereRadius(QEVertex *pvx);
  Point CalculateFacetSubdivisionSqrt3SphereCapArea(QEFacet *pfa);
  Point CalculateVertexSubdivisionSqrt3SphereCapArea(QEVertex *pvx);
  Point CalculateFacetSubdivisionSqrt3SphereTriangleArea(QEFacet *pfa);
  Point CalculateVertexSubdivisionSqrt3SphereTriangleArea(QEVertex *pvx);
  Point CalculateFacetSubdivisionSqrt3SphereTangent(QEFacet *pfa);
  Point CalculateVertexSubdivisionSqrt3SphereTangent(QEVertex *pvx);

  Point CalculateFacetSubdivisionSqrt4(QEFacet *pfa);
  Point CalculateVertexSubdivisionSqrt4(QEVertex *pvx);

  Point CalculateFacetSubdivisionCatmullClark(SubdivisionType eSubdivisionType, QEFacet *pfa);
  Point CalculateVertexSubdivisionCatmullClark(SubdivisionType eSubdivisionType, QEVertex *pvx);
  Point CalculateEdgeSubdivisionCatmullClark(SubdivisionType eSubdivisionType, QEEdge *pe);

  Point CalculateFacetSubdivisionCatmullClark(QEFacet *pfa);
  Point CalculateVertexSubdivisionCatmullClark(QEVertex *pvx);
  Point CalculateVertexLimitSubdivisionCatmullClark(QEVertex *pvx);
  Vector CalculateVertexNormalSubdivisionCatmullClark(QEVertex *pvx);
  Point CalculateEdgeSubdivisionCatmullClark(QEEdge *pe);

  Point CalculateFacetSubdivisionCatmullClarkC2(QEFacet *pfa);
  Point CalculateVertexSubdivisionCatmullClarkC2(QEVertex *pvx);
  Point CalculateVertexLimitSubdivisionCatmullClarkC2(QEVertex *pvx);
  Vector CalculateVertexNormalSubdivisionCatmullClarkC2(QEVertex *pvx);
  Point CalculateEdgeSubdivisionCatmullClarkC2(QEEdge *pe);

  Point CalculateFacetSubdivisionCatmullClarkButterfly(QEFacet *pfa);
  Point CalculateVertexSubdivisionCatmullClarkButterfly(QEVertex *pvx);
  Point CalculateEdgeSubdivisionCatmullClarkButterfly(QEEdge *pe);
  Point CalculateVirtualEdgeSubdivisionCatmullClarkButterfly(QEEdge *pe);

  static Point CalculateFacetSubdivisionSharpCatmullClark(QEFacet *pfa);
  static Point CalculateVertexSubdivisionSharpCatmullClark(QEVertex *pvx);
  static Point CalculateVertexLimitSubdivisionSharpCatmullClark(QEVertex *pvx);
  static Vector CalculateVertexNormalSubdivisionSharpCatmullClark(QEVertex *pvx);
  static Vector CalculateVertexNormalSubdivisionSharpCatmullClark(QEEdge *pe);
  static Point CalculateEdgeSubdivisionSharpCatmullClark(QEEdge *peCrease);
/*
  Point CalculateFacetSubdivisionDooSabin(QEVertex *pvx);
  Point CalculateVertexSubdivisionDooSabin(QEVertex *pvx);
  Point CalculateEdgeSubdivisionDooSabin(QEEdge *pe);
*/
  Point CalculateVertexFacetSubdivisionDooSabin(QEEdge *pe);
  Point CalculateVertexLimitSubdivisionDooSabin(QEVertex *pvx);
  Vector CalculateVertexNormalSubdivisionDooSabin(QEVertex *pvx);

  Point CalculateFacetSubdivisionHeterogeneous(SubdivisionType eSubdivisionType, QEFacet *pfa);
  Point CalculateVertexSubdivisionHeterogeneous(SubdivisionType eSubdivisionType, QEVertex *pvx);
  Point CalculateEdgeSubdivisionHeterogeneous(SubdivisionType eSubdivisionType, QEEdge *pe);

  static Point CalculateFacetSubdivisionHeterogeneous(QEFacet *pfa);
  static Point CalculateVertexSubdivisionHeterogeneous(QEVertex *pvx);
  static Point CalculateVertexLimitSubdivisionHeterogeneous(QEVertex *pvx);
  static Vector CalculateVertexNormalSubdivisionHeterogeneous(QEVertex *pvx);
  static Vector CalculateVertexNormalSubdivisionHeterogeneous(QEEdge *pe);
  static Point CalculateEdgeSubdivisionHeterogeneous(QEEdge *peCrease);

  static Point CalculateFacetSubdivisionHeterogeneousStamLoop(QEFacet *pfa);
  static Point CalculateVertexSubdivisionHeterogeneousStamLoop(QEVertex *pvx);
  static Point CalculateVertexLimitSubdivisionHeterogeneousStamLoop(QEVertex *pvx);
  static Vector CalculateVertexNormalSubdivisionHeterogeneousStamLoop(QEVertex *pvx);
  static Vector CalculateVertexNormalSubdivisionHeterogeneousStamLoop(QEEdge *pe);
  static Point CalculateEdgeSubdivisionHeterogeneousStamLoop(QEEdge *peCrease);

  static Point CalculateFacetSubdivisionHeterogeneousOld(QEFacet *pfa);
  static Point CalculateVertexSubdivisionHeterogeneousOld(QEVertex *pvx);
  static Point CalculateVertexLimitSubdivisionHeterogeneousOld(QEVertex *pvx);
  static Vector CalculateVertexNormalSubdivisionHeterogeneousOld(QEVertex *pvx);
  static Vector CalculateVertexNormalSubdivisionHeterogeneousOld(QEEdge *pe);
  static Point CalculateEdgeSubdivisionHeterogeneousOld(QEEdge *peCrease);

  Point CalculateFacetSubdivisionPentagon(SubdivisionType eSubdivisionType, QEEdge *pe);
  Point CalculateVertexSubdivisionPentagon(SubdivisionType eSubdivisionType, QEVertex *pvx);
  Point CalculateEdgeSubdivisionPentagon(SubdivisionType eSubdivisionType, QEEdge *pe);

  static Point CalculateFacetSubdivisionPentagonTesselate(QEEdge *pe);
  static Point CalculateFacetSubdivisionPentagon(QEEdge *pe);
  static Point CalculateVertexSubdivisionPentagon(QEVertex *pvx);
  //static Point CalculateVertexLimitSubdivisionPentagon(QEVertex *pvx);
  //static Vector CalculateVertexNormalSubdivisionPentagon(QEVertex *pvx);
  //static Vector CalculateVertexNormalSubdivisionPentagon(QEEdge *pe);
  static Point CalculateEdgeSubdivisionPentagon(QEEdge *pe);

  //////////////////////////////////////////////////
  // Selective Schemes

  //////////////////////////////
  // Loop Vertex Centric
  VOID SubdivideCornerRoundingSelective(QEVertex *pvx);
  VOID UnSubdivideCornerRoundingSelective(QEVertex *pvx);
  static FLOAT CalculateVertexErrorSubdivideCornerRoundingSelective(QEVertex *pvx);

  VOID SubdivideCornerRoundingSelectiveLimit(QEVertex *pvx);
  VOID UnSubdivideCornerRoundingSelectiveLimit(QEVertex *pvx);
  static FLOAT CalculateVertexErrorSubdivideCornerRoundingSelectiveLimit(QEVertex *pvx);

  VOID SubdivideCornerRoundingSharpSelectiveLimit(QEVertex *pvx, BOOL bPresubdivide = FALSE);
  VOID UnSubdivideCornerRoundingSharpSelectiveLimit(QEVertex *pvx);
  static FLOAT CalculateVertexErrorSubdivideCornerRoundingSharpSelectiveLimit(ErrorMetricType eErrorMetric, QEVertex *pvx);
  static FLOAT CalculateVertexErrorSubdivideCornerRoundingSharpSelectiveLimitLimitDistance(QEVertex *pvx);
  static FLOAT CalculateVertexErrorSubdivideCornerRoundingSharpSelectiveLimitTangentPlane(QEVertex *pvx);

  static QEEdge *FindSubdivideCornerRoundingSelectiveCenterVertex(QEEdge *pe);
  static Point CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(QEVertex *pvx, UINT_32 uiGeneration);

  static Point CalculateVertexSubdivisionCornerRoundingSelective(QEVertex *pvx);
  static Point CalculateVertexHelperSubdivisionCornerRoundingSelective(QEVertex *pvx);
  static Vector CalculateGenerationVectorSubdivisionCornerRoundingSelective(QEVertex *pvx);
  static Point CalculateVertexLimitSubdivisionCornerRoundingSelective(QEVertex *pvx);
  static Vector CalculateVertexNormalSubdivisionCornerRoundingSelective(QEVertex *pvx);
  static Point CalculateEdgeSubdivisionCornerRoundingSelective(QEEdge *pe);

  static Point CalculateVertexSubdivisionCornerRoundingSharpSelective(QEVertex *pvx);
  static Point CalculateVertexHelperSubdivisionCornerRoundingSharpSelective(QEVertex *pvx);
  static Vector CalculateGenerationVectorSubdivisionCornerRoundingSharpSelective(QEVertex *pvx);
  static Point CalculateVertexLimitSubdivisionCornerRoundingSharpSelective(QEVertex *pvx);
  static Vector CalculateVertexNormalSubdivisionCornerRoundingSharpSelective(QEVertex *pvx);
public:
  static Vector CalculateVertexNormalSubdivisionCornerRoundingSharpSelective(QEEdge *peCrease);
protected:
  static Point CalculateEdgeSubdivisionCornerRoundingSharpSelective(QEEdge *pe);

#ifdef LOOP_FACET_CENTRIC
  //////////////////////////////
  // Loop Facet Centric
  VOID SubdivideCornerRoundingFacetSharpSelectiveLimit(QEFacet *pfa, BOOL bPresubdivide = FALSE);
  VOID UnSubdivideCornerRoundingFacetSharpSelectiveLimit(QEFacet *pfa);
  static FLOAT CalculateFacetErrorSubdivideCornerRoundingFacetSharpSelectiveLimit(ErrorMetricType eErrorMetric, QEFacet *pfa);
  static FLOAT CalculateFacetErrorSubdivideCornerRoundingFacetSharpSelectiveLimitMidpoint(QEFacet *pfa);

  static Point CalculateVertexSubdivisionCornerRoundingFacetSharpSelective(QEVertex *pvx);
  //static Point CalculateVertexHelperSubdivisionCornerRoundingFacetSharpSelective(QEVertex *pvx);
  //static Vector CalculateGenerationVectorSubdivisionCornerRoundingFacetSharpSelective(QEVertex *pvx);
  static Point CalculateVertexLimitSubdivisionCornerRoundingFacetSharpSelective(QEVertex *pvx);
  static Vector CalculateVertexNormalSubdivisionCornerRoundingFacetSharpSelective(QEVertex *pvx);
public:
  //static Vector CalculateVertexNormalSubdivisionCornerRoundingFacetSharpSelective(QEEdge *peCrease);
protected:
  static Point CalculateEdgeSubdivisionCornerRoundingFacetSharpSelective(QEEdge *pe);
#endif // LOOP_FACET_CENTRIC

  //////////////////////////////
  // Doo-Sabin
  VOID SubdivideDooSabinSelective(QEVertex *pvx);
  VOID UnSubdivideDooSabinSelective(QEVertex *pvx);
  FLOAT CalculateVertexErrorSubdivideDooSabinSelective(QEVertex *pvx);
  VOID SubdivideDooSabinSelectiveLimit(QEVertex *pvx);
  FLOAT CalculateVertexErrorSubdivideDooSabinSelectiveLimit(QEVertex *pvx);
  Point CalculateEdgeSubdivisionDooSabinSelective(QEEdge *pe);
  Point CalculateVertexLimitSubdivisionDooSabinSelective(QEEdge *pe);
  Vector CalculateVertexNormalSubdivisionDooSabinSelective(QEEdge *pe);

  //////////////////////////////
  // Catmull-Clark
  VOID SubdivideCatmullClarkSelective(QEVertex *pvx);
  FLOAT CalculateVertexErrorSubdivideCatmullClarkSelective(QEVertex *pvx);
  Point CalculateFacetSubdivisionCatmullClarkSelective(QEFacet *pfa);
  Point CalculateVertexSubdivisionCatmullClarkSelective(QEVertex *pvx);
  Point CalculateEdgeSubdivisionCatmullClarkSelective(QEEdge *pe);

  //////////////////////////////
  // Sqrt4
  VOID SubdivideSqrt4Selective(QEVertex *pvx);
  FLOAT CalculateVertexErrorSubdivideSqrt4Selective(QEVertex *pvx);
  Point CalculateVertexSubdivisionSqrt4Selective(QEVertex *pvx);
  Point CalculateFacetSubdivisionSqrt4Selective(QEFacet *pfa);

  //////////////////////////////
  // Sqrt3
  VOID SubdivideSqrt3Selective(QEVertex *pvx);
  FLOAT CalculateVertexErrorSubdivideSqrt3Selective(QEVertex *pvx);
  Point CalculateVertexSubdivisionSqrt3Selective(QEVertex *pvx);
  Point CalculateFacetSubdivisionSqrt3Selective(QEFacet *pfa);

  VOID SubdivideSqrt3SelectiveFacet(ErrorMetricType eErrorMetric, QEFacet *pfa);
  QEVertex *SubdivideSqrt3SelectiveFacetSplit(ErrorMetricType eErrorMetric, QEFacet *pfa);
  VOID SubdivideSqrt3SelectiveFacetSwap(ErrorMetricType eErrorMetric, QEEdge *pe);
  FLOAT CalculateFacetErrorSubdivideSqrt3SelectiveFacet(ErrorMetricType eErrorMetric,
                                                        QEFacet *pfa);
  FLOAT CalculateVertexErrorSubdivideSqrt3SelectiveFacet(ErrorMetricType eErrorMetric, 
                                                         UINT_32 uiGeneration, 
                                                         QEVertex *pvx);
  FLOAT CalculateVertexErrorSubdivideSqrt3SelectiveFacetBestFitPlane(QEVertex *pvx);
  FLOAT CalculateVertexErrorSubdivideSqrt3SelectiveFacetLimitDistance(UINT_32 uiGeneration, 
                                                                      QEVertex *pvx);
  Point CalculateVertexSubdivisionSqrt3SelectiveFacet(QEVertex *pvx);
  Point CalculateVertexSubdivisionSqrt3SelectiveFacetGeneration(QEVertex *pvx, UINT_32 uiGeneration);
  Point CalculateFacetSubdivisionSqrt3SelectiveFacet(QEFacet *pfa);

//////////////////////////////////////////////////
// Progressive Meshes
//
public:
  QEMesh *CreateProgressiveMesh(ErrorMetricType eErrorMetric, UINT_32 uiFacetsMin = 4);
  VOID AdaptProgressiveMesh(UINT_32 uiFacetsMax);

protected:
  FLOAT CalculateEdgeErrorProgressiveMesh(ErrorMetricType eErrorMetric,
                                          QEEdge *pe);
  FLOAT CalculateVertexErrorProgressiveMesh(ErrorMetricType eErrorMetric, 
                                            QEVertex *pvx);
  FLOAT CalculateVertexErrorProgressiveMeshBestFitPlane(QEVertex *pvx);
  VOID CreateProgressiveCollapseEdge(ErrorMetricType eErrorMetric, QEEdge *pe);
  VOID ProgressiveCollapseEdge(QEEdgeCollapse *pec);
  VOID ProgressiveSplitVertex(QEEdgeCollapse *pec);

//////////////////////////////////////////////////
// Evolver for surface optimization
//   this works with 2 mesh with forward pointers between the vertices
//
public:
  QEMesh *EvolverCopy();
  FLOAT EvolverEnergy();
  FLOAT EvolverOptimizeStep(FLOAT fGradientPercent = 0.25);
  Point EvolverMoveVertex(QEVertex *pvx, FLOAT fGradientPercent);
  FLOAT EvolverVertexEnergy(QEVertex *pvx);

//////////////////////////////////////////////////
// Forward Pointers
//
protected:
  VOID ClearForwardPointers();

//////////////////////////////////////////////////
// Merging
//
public:
  // Special Merge function for SLIDEMesh object
  VOID MergeLocalFacet(QEFacet *pfa);

  // More general merging facilities
public:
  BOOL EqualEdge (QEEdge *pe, QEEdge *peResult);
  VOID MergeFacet(CObjectList<QEVertex> *plistQEVertex);
  VOID MergeFacet(CObjectList<QEVertex> *plistQEVertex, CSLIDESurface *pCSLIDESurface);
  VOID MergeTriangleStrip(CObjectList<QEVertex> *plistQEVertex);
  VOID MergeTriangleStrip(CObjectList<QEVertex> *plistQEVertex, 
			  CSLIDESurface *pCSLIDESurface);
  VOID MergeTriangleFan(CObjectList<QEVertex> *plistQEVertex);
  VOID MergeQuadStrip(CObjectList<QEVertex> *plistQEVertex);
  VOID MergeQuadStrip(CObjectList<QEVertex> *plistQEVertex, CSLIDESurface *pCSLIDESurface);
  VOID SpliceMesh(QEMesh *pme);
  VOID MergeMesh(QEMesh *pme);
  VOID MergeMesh(QEMesh *pme, CSLIDESurface *pCSLIDESurface);
  QEMesh *CreateCleanMesh();

protected:
  VOID MergeEdge(QEEdge *pe, QEEdge *peMerge, QEMesh *pme);
  VOID MergeMesh(UINT_32 uiSearchType, QEMesh *pme);
  QEEdge *SearchEdge(UINT_32 uiSearchType, QEEdge *pe, FLOAT fEpsilon);


//////////////////////////////////////////////////
// Rendering
//
public:
  //virtual VOID Render();
  //virtual VOID RenderGeometry();
  //virtual VOID RenderNormals();

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID Display();
  virtual VOID DisplayEuler();

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  inline UINT_32 GetVertexCount();
  inline UINT_32 GetFacetCount();
  inline UINT_32 GetEdgeCount();
  inline UINT_32 GetBoundaryEdgeCount();

  inline QEMesh *GetPrevLink();
  inline VOID SetPrevLink(QEMesh *pme);
  inline QEMesh *GetNextLink();
  inline VOID SetNextLink(QEMesh *pme);

//////////////////////////////////////////////////
// Class Accessor Functions
//
public:
  static inline QEFacet *GetNullFacet();
  static inline BOOL GetRandomize();
  static inline VOID SetRandomize(BOOL bRandomize);
  static inline AccuracyType GetAccuracyType();
  static inline VOID SetAccuracyType(AccuracyType eAccuracy);
  static inline FLOAT GetMergeEpsilon();
  static inline VOID SetMergeEpsilon(FLOAT fMergeEpsilon);

//////////////////////////////////////////////////
// Class Variables
//
protected:
  static QEFacet m_faNULL;
  static QEFacet *m_pfaNULL;

  static BOOL m_bRandomize;
  static AccuracyType m_eAccuracy;

  static const FLOAT m_cfEpsilon;
  static FLOAT m_fMergeEpsilon;

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CObjectList<QEVertex> *m_plistQEVertex;
  CObjectList<QEWedge> *m_plistQEWedge;
  CObjectList<QEEdge> *m_plistQEEdge;
  CObjectList<QEEdge> *m_plistQEEdgeBoundary;
  CObjectList<QEFacet> *m_plistQEFacet;

  CObjectList<QEEdgeCollapse> *m_plistQEEdgeCollapse;
  QEEdgeCollapse *m_pecCurr;

  CObjectList<QEEdgeNode> *m_plistQEEdgeNode;
  CObjectList<QEFacetNode> *m_plistQEFacetNode;

  CHeap<VOID*, FLOAT> *m_pheapSelective;
  CHeap<VOID*, FLOAT> *m_pheapSelectiveReverse;
  CListInt *m_plistSelective;
  QEEdgeNode *m_penRoot;

  // Listing on a Free List
  QEMesh *m_pmeLink[2];
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline UINT_32 QEMesh::GetVertexCount()
{
  return m_plistQEVertex->GetLength();
}

inline UINT_32 QEMesh::GetFacetCount()
{
  return m_plistQEFacet->GetLength();
}

inline UINT_32 QEMesh::GetEdgeCount()
{
  return m_plistQEEdge->GetLength();
}

inline UINT_32 QEMesh::GetBoundaryEdgeCount()
{
  return m_plistQEEdgeBoundary->GetLength();
}

inline QEMesh *QEMesh::GetPrevLink()
{
  return m_pmeLink[0];
}

inline VOID QEMesh::SetPrevLink(QEMesh *pme)
{
  m_pmeLink[0] = pme;
}

inline QEMesh *QEMesh::GetNextLink()
{
  return m_pmeLink[1];
}

inline VOID QEMesh::SetNextLink(QEMesh *pme)
{
  m_pmeLink[1] = pme;
}

inline QEFacet *QEMesh::GetNullFacet()
{
  return m_pfaNULL;
}

inline BOOL QEMesh::GetRandomize()
{
  return m_bRandomize;
}

inline VOID QEMesh::SetRandomize(BOOL bRandomize)
{
  m_bRandomize = bRandomize;
}

inline QEMesh::AccuracyType QEMesh::GetAccuracyType()
{
  return m_eAccuracy;
}

inline VOID QEMesh::SetAccuracyType(AccuracyType eAccuracy)
{
  m_eAccuracy = eAccuracy;
}

inline FLOAT QEMesh::GetMergeEpsilon()
{
  return m_fMergeEpsilon;
}

inline VOID QEMesh::SetMergeEpsilon(FLOAT fMergeEpsilon)
{
  m_fMergeEpsilon = fMergeEpsilon;
}

//////////////////////////////////////////////////////////////////////
// QEMeshIter Class
//

class QEMeshIter
{
//////////////////////////////////////////////////
// Initialization
//
public:
  QEMeshIter();
  ~QEMeshIter();
  VOID Init(QEMesh *pQEMesh);
  VOID Uninit();

  BOOL Initialized();

  BOOL Valid();

//////////////////////////////////////////////////
// Iteration Functions
//
public:
  QEVertex *PeekFirstVertex();
  QEVertex *PeekNextVertex();
  QEVertex *PeekLastVertex();
  QEVertex *PeekPrevVertex();

  QEWedge *PeekFirstWedge();
  QEWedge *PeekNextWedge();
  QEWedge *PeekLastWedge();
  QEWedge *PeekPrevWedge();

  QEEdge *PeekFirstEdge();
  QEEdge *PeekNextEdge();
  QEEdge *PeekLastEdge();
  QEEdge *PeekPrevEdge();

  QEEdge *PeekFirstEdgeBoundary();
  QEEdge *PeekNextEdgeBoundary();
  QEEdge *PeekLastEdgeBoundary();
  QEEdge *PeekPrevEdgeBoundary();

  QEFacet *PeekFirstFacet();
  QEFacet *PeekNextFacet();
  QEFacet *PeekLastFacet();
  QEFacet *PeekPrevFacet();

//////////////////////////////////////////////////
// Accessor Functions
//
public:

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  QEMesh *m_pQEMesh;

  CObjectListIter<QEVertex> m_iterQEVertex;
  CObjectListIter<QEWedge> m_iterQEWedge;
  CObjectListIter<QEEdge> m_iterQEEdge;
  CObjectListIter<QEEdge> m_iterQEEdgeBoundary;
  CObjectListIter<QEFacet> m_iterQEFacet;
};

#endif // __QEMESH_H__




