#include "qemesh.h"

QEMesh *QEMesh::EvolverCopy()
{
  QEMeshIter iterQEMesh;
  QEMesh *pme;
  QEVertex *pvx, *pvxNew;
  QEFacet *pfa, *pfaNew;
  QEEdge *pe, *peNew, *peCurr;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  ClearForwardPointers();

  iterQEMesh.Init(this);

  // Copy Vertices
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pvxNew = pme->CreateVertex(pvx->GetPoint());
      pvxNew->SetLimitPosition(pvx->GetLimitPosition());
      pvxNew->SetNormal(pvx->GetNormal());
      pvxNew->SetTexture(pvx->GetTexture());
      pvxNew->SetSharpVertexType(pvx->GetSharpVertexType());
      pvxNew->SetForwardVertex(pvx);
      pvx->SetForwardVertex(pvxNew);
    }

  // Copy Facets
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      pfaNew = pme->CreateFacet();

      pfa->SetForwardFacet(pfaNew);
    }

  // Copy Edges
  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      if ( pe->GetForwardEdge() == NULL )
	{
	  peNew = pme->CreateEdge();
	  peNew->SetKnotInterval(pe->GetKnotInterval());
	  peNew->GetEdge()->SetKnotInterval(pe->GetEdge()->GetKnotInterval());

	  pe->SetForwardEdge(peNew);
	  pe->GetEdge()->SetForwardEdge(peNew->GetEdge());

	  if ( pe->GetFacet() == m_pfaNULL )
	    {
	      peNew->SetFacet(m_pfaNULL);
	      pme->m_plistQEEdge->Remove(peNew);
	      pme->m_plistQEEdgeBoundary->InsertLast(peNew);
	    }
	  else
	    {
	      peNew->SetFacet(pe->GetFacet()->GetForwardFacet());
	    }

	  if ( pe->GetEdge()->GetFacet() == m_pfaNULL )
	    {
	      peNew->GetEdge()->SetFacet(m_pfaNULL);
	      pme->m_plistQEEdge->Remove(peNew->GetEdge());
	      pme->m_plistQEEdgeBoundary->InsertLast(peNew->GetEdge());
	    }
	  else
	    {
	      peNew->GetEdge()->SetFacet(pe->GetEdge()->GetFacet()->GetForwardFacet());
	    }

	  peNew->SetVertex(pe->GetVertex()->GetForwardVertex());
	  peNew->GetEdge()->SetVertex(pe->GetEdge()->GetVertex()->GetForwardVertex());

	  for ( peCurr = pe->GetEdgePrev(1);
		peCurr != pe;
		peCurr = peCurr->GetEdgePrev(1) )
	    {
	      if ( peCurr->GetForwardEdge() != NULL )
		{
		  Splice(1, peNew, peCurr->GetForwardEdge());
		  break;
		}
	    }

	  for ( peCurr = pe->GetEdge()->GetEdgePrev(1);
		peCurr != pe->GetEdge();
		peCurr = peCurr->GetEdgePrev(1) )
	    {
	      if ( peCurr->GetForwardEdge() != NULL )
		{
		  Splice(1, peNew->GetEdge(), peCurr->GetForwardEdge());
		  break;
		}
	    }
	}
    }

  //pme->Preprocess();

  return pme;
}

FLOAT QEMesh::EvolverEnergy()
{
  QEMeshIter iterQEMesh;
  QEVertex *pvx;

  FLOAT fEnergy;

  //////////////////////////////////////////////////
  // Calculate the energy of the mesh

  iterQEMesh.Init(this);

  fEnergy = 0.0;
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      fEnergy += EvolverVertexEnergy(pvx);
    }

  return fEnergy;
}

#define LOOP_CURVATURE_GRADIENT
//#define LOOP_LIMIT_TRIANGLE_GRADIENT

#ifdef LOOP_CURVATURE_GRADIENT

FLOAT QEMesh::EvolverOptimizeStep(FLOAT fGradientPercent/* = 0.25*/)
{
  QEMeshIter iterQEMesh;
  QEVertex *pvx;

  Point pt;
  FLOAT fEnergyNew;

  //////////////////////////////////////////////////
  // Move the vertex positions to be more relaxed

  iterQEMesh.Init(this);

  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pt = EvolverMoveVertex(pvx->GetForwardVertex(), fGradientPercent);
      pvx->SetPoint(pt);
    }

  //CalculateSubdivisionLimitPositionsAndNormalsSharpLoop();
  fEnergyNew = EvolverEnergy();
  /*
  printf("Energy Old: %f > %f :Energy New, Delta = %f\n", fEnergyOld, fEnergyNew, fEnergyOld - fEnergyNew);
  */
  //Preprocess();

  return fEnergyNew;
}

Point QEMesh::EvolverMoveVertex(QEVertex *pvx, FLOAT fGradientPercent)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  Point pt, ptV;
  Vector vSu, vSv, vSuu, vSuv, vSvv, vN;
  FLOAT fAuu, fAuv, fAvv, fTheta, fThetaDelta, pfI[3];
  Vector v, vGradient;
  FLOAT fDen, fSharpness;
  UINT_32 uiN, uiSharpEdges, i;
  UINT_32 uiEdge;
  QEEdge *peStart;

  ptV = pvx->GetPoint();
  ptV[3] = 1.0;
  if ( pvx->GetSharpVertexType() == svtCorner )
    {
      return ptV;
    }

  iterQEVertex.Init(pvx);

  uiSharpEdges = 0;
  for ( pe = iterQEVertex.PeekFirstEdge(), uiN = 0;
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge(), uiN++ )
    {
      fSharpness = pe->GetKnotInterval();
      if ( fSharpness == INFINITY || fSharpness >= 1.0 )
	{
	  // Infinitely sharp for this step
	  uiSharpEdges++;
	}
    }

  if ( uiSharpEdges > 0 )
    {
      return ptV;
    }

  // Debugging: Check if the u,v directions matter to k1^2 + k2^2
  for ( peStart = iterQEVertex.PeekFirstEdge(), uiEdge = 0;
	peStart != NULL;
	peStart = iterQEVertex.PeekNextEdge(), uiEdge++ )
    {

  //////////////////////////////////////////////////
  // Iterate around the neighbor vertices and set up the system Ax = b
  //   where A is the symmetric matrix
  //   | d0 c2 c1 |
  //   | c2 d1 c0 |
  //   | c1 c0 d2 |

  // gradV = 2 ( ( (avv f^2 - 2 auv f g + auu g^2) Suu + 
  //             -2 (avv e f - auv (e g + f^2) + auu f g) Suv + 
  //             (avv e^2 - 2 auv e f + auu f^2) Svv) . normal ) normal;

  //if ( pe->GetEdge()->GetVertex()->GetSharpVertexType() == svtSmooth )

  fAuu = -2.0;
  fAuv = -1.0;
  fAvv = -2.0;

  v = ptV;
  vSu.Null();
  vSv.Null();
  vSuu = (fAuu*v);
  vSuv = (fAuv*v);
  vSvv = (fAvv*v);
  fThetaDelta = 2.0 * PI / uiN;
  fTheta = -fThetaDelta;
  
  //for ( pe = iterQEVertex.PeekFirstEdge(), i = 0;
  //pe != NULL;
  //pe = iterQEVertex.PeekNextEdge(), i++ )
  i = 0;
  pe = peStart;
  do
    {
      pt = pe->GetEdge()->GetVertex()->GetPoint();
      v = pt;
      vSv += ((cos(fTheta)/3.0)*v);
      fTheta += fThetaDelta;
      vSu += ((cos(fTheta)/3.0)*v);

      if ( i == 0 )
	{
	  vSuu += v;
	}
      if ( i == 1 )
	{
	  vSvv += v;
	}

      if ( uiN & 0x1 )
	{
	  // Odd

	  if ( i == 0 || i == 1 || i == uiN/2 || i == uiN/2 + 1 )
	    {
	      vSuv += 0.5*v;
	    }
	  if ( i == uiN/4 + 1 || i == (3*uiN)/4 + 1 )
	    {
	      vSuv -= 0.5*v;
	    }

	  if ( i == uiN/2 || i == uiN/2 + 1 )
	    {
	      vSuu += 0.5*v;
	    }
	  if ( i == uiN/2 + 1 ||  i == uiN/2 + 2 )
	    {
	      vSvv += 0.5*v;
	    }
	}
      else
	{
	  // Even

	  if ( i == 0 || i == 1 || i == uiN/2 || i == uiN/2 + 1 )
	    {
	      vSuv += 0.5*v;
	    }
	  if ( i == uiN/4 + 1 || i == (3*uiN)/4 + 1 )
	    {
	      vSuv -= 0.5*v;
	    }

	  if ( i == uiN/2 )
	    {
	      vSuu += v;
	    }
	  if ( i == uiN/2 + 1 )
	    {
	      vSvv += v;
	    }
	}

      i++;
      pe = pe->GetEdgeNext(1);
    } while ( pe != peStart );

  vSu[3] = 0.0;
  vSv[3] = 0.0;
  vSuu[3] = 0.0;
  vSuv[3] = 0.0;
  vSvv[3] = 0.0;

  vN = vSu * vSv;
  VERIFY( vN.Normalize() );

  pfI[0] = vSu.Dot(vSu); // E
  pfI[1] = vSu.Dot(vSv); // F
  pfI[2] = vSv.Dot(vSv); // G

  vGradient = 2*((fAvv*pfI[1]*pfI[1] - 2*fAuv*pfI[1]*pfI[2] + fAuu*pfI[2]*pfI[2])*vSuu - 
		 2*(fAvv*pfI[0]*pfI[1] - fAuv*(pfI[0]*pfI[2] + pfI[1]*pfI[1]) + fAuu*pfI[1]*pfI[2])*vSuv + 
		 (fAvv*pfI[0]*pfI[0] - 2*fAuv*pfI[0]*pfI[1] + fAuu*pfI[1]*pfI[1])*vSvv);
  vGradient = (-(vGradient.Dot(vN)))*vN;
  vGradient[3] = 0.0;

  fDen = (pfI[0]*pfI[2] - pfI[1]*pfI[1]);
  fDen *= fDen;

  if ( fDen == 0.0 )
    {
      pt = ptV;
      pt[3] = 1.0;
    }
  else
    {
      vGradient /= fDen;

      pt = ptV + (fGradientPercent * vGradient);
      pt[3] = 1.0;
    }

  if ( uiN == 6 )
    {
      return pt;
    }

  /*
    printf("ptV (%f, %f, %f) -> pt (%f, %f, %f)\n",
    ptV(0), ptV(1), ptV(2),
    pt(0), pt(1), pt(2));
  */
  printf("%ld: ptV (%f, %f, %f) -> v %f (%f, %f, %f)\n",
	 uiEdge, ptV(0), ptV(1), ptV(2), vGradient.Magnitude(),
	 vGradient(0), vGradient(1), vGradient(2));

    }

  return pt;
}

FLOAT QEMesh::EvolverVertexEnergy(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  Point pt;
  Vector v, vN, vSu, vSv, vSuu, vSuv, vSvv;
  FLOAT pfI[3], pfII[3];
  FLOAT fTheta, fThetaDelta;
  FLOAT fEnergy, fGauss, fMean, fDen;
  UINT_32 i, uiEdge, uiN;
  QEEdge *peStart;

  if ( pvx->GetSharpVertexType() != svtSmooth )
    {
      return 0.0;
    }

  iterQEVertex.Init(pvx);

  uiN = 0;
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      uiN++;
    }

  // Debugging: Check if initial edge effects k1^2 + k2^2
  for ( peStart = iterQEVertex.PeekFirstEdge(), uiEdge = 0;
	peStart != NULL;
	peStart = iterQEVertex.PeekNextEdge(), uiEdge++ )
    {

  pt = pvx->GetPoint();
  v = pt;

  vSu.Null();
  vSv.Null();
  vSuu = (-2.0*v);
  vSuv = -v;
  vSvv = (-2.0*v);
  fThetaDelta = 2.0 * PI / uiN;
  fTheta = -fThetaDelta;
  //for ( pe = iterQEVertex.PeekFirstEdge(), i = 0;
  //pe != NULL;
  //pe = iterQEVertex.PeekNextEdge(), i++ )
  i = 0;
  pe = peStart;
  do
    {
      pt = pe->GetEdge()->GetVertex()->GetPoint();
      v = pt;
      vSv += ((cos(fTheta)/3.0)*v);
      fTheta += fThetaDelta;
      vSu += ((cos(fTheta)/3.0)*v);

      if ( i == 0 )
	{
	  vSuu += v;
	}
      if ( i == 1 )
	{
	  vSvv += v;
	}

      if ( uiN & 0x1 )
	{
	  // Odd

	  if ( i == 0 || i == 1 || i == uiN/2 || i == uiN/2 + 1 )
	    {
	      vSuv += 0.5*v;
	    }
	  if ( i == uiN/4 + 1 || i == (3*uiN)/4 + 1 )
	    {
	      vSuv -= 0.5*v;
	    }

	  if ( i == uiN/2 || i == uiN/2 + 1 )
	    {
	      vSuu += 0.5*v;
	    }
	  if ( i == uiN/2 + 1 ||  i == uiN/2 + 2 )
	    {
	      vSvv += 0.5*v;
	    }
	}
      else
	{
	  // Even

	  if ( i == 0 || i == 1 || i == uiN/2 || i == uiN/2 + 1 )
	    {
	      vSuv += 0.5*v;
	    }
	  if ( i == uiN/4 + 1 || i == (3*uiN)/4 + 1 )
	    {
	      vSuv -= 0.5*v;
	    }

	  if ( i == uiN/2 )
	    {
	      vSuu += v;
	    }
	  if ( i == uiN/2 + 1 )
	    {
	      vSvv += v;
	    }
	}

      i++;
      pe = pe->GetEdgeNext(1);
    } while ( pe != peStart );
  vSu[3] = 0.0;
  vSv[3] = 0.0;
  vSuu[3] = 0.0;
  vSuv[3] = 0.0;
  vSvv[3] = 0.0;

  vN = vSu * vSv;
  VERIFY( vN.Normalize() );

  pfI[0] = vSu.Dot(vSu); // E
  pfI[1] = vSu.Dot(vSv); // F
  pfI[2] = vSv.Dot(vSv); // G

  pfII[0] = vSuu.Dot(vN); // L
  pfII[1] = vSuv.Dot(vN); // M
  pfII[2] = vSvv.Dot(vN); // N

  // Gaussian Curvature = K =       k1 k2 = (L N - M^2) / (E G - F^2)
  // Mean Curvature =     H = (k1 + k2)/2 = (E N - 2 F M + G L) / (2 (E G - F^2))
  // Bending Energy =         k1^2 + k2^2 = (2 H)^2 - 2 K =
  //   ((E N - 2 F M + G L)^2 - 2 (L N - M^2) (E G - F^2)) / (E G - F^2)^2

  fDen = (pfI[0] * pfI[2] - pfI[1] * pfI[1]);

  fGauss = (pfII[0] * pfII[2] - pfII[1] * pfII[1]);

  fMean = (pfI[0] * pfII[2] - 2.0 * pfI[1] * pfII[1] + pfI[2] * pfII[0]);

  fEnergy = (pfI[0] * pfII[2] - 2 * pfI[1] * pfII[1] + pfI[2] * pfII[0]);
  fEnergy = fEnergy * fEnergy - 2 * (pfII[0] * pfII[2] - pfII[1] * pfII[1]) * fDen;

  if ( fDen == 0.0 )
    {
      fGauss = INFINITY;
      fMean = INFINITY;
      fEnergy = INFINITY;
    }
  else
    {
      fGauss /= fDen;
      fMean /= (2.0*fDen);
      fEnergy /= (fDen*fDen);
    }

  ASSERT( fEnergy >= 0.0 );

  if ( uiN == 6 )
    {
      return fEnergy;
    } 

  printf("%ld: Kb = %f : %f : Kg %f : Km %f : Den %f\n", uiEdge, 
	 fEnergy, 4.0*fMean*fMean - 2.0*fGauss, fGauss, fMean, fDen);

    }
  printf("\n");

  return fEnergy;
}

#else // LOOP_CURVATURE_GRADIENT

#ifdef LOOP_LIMIT_TRIANGLE_GRADIENT

FLOAT QEMesh::EvolverOptimizeStep(FLOAT fGradientPercent/* = 0.25*/)
{
  QEMeshIter iterQEMesh;
  QEVertex *pvx;

  Point pt;
  FLOAT fEnergyNew;

  //////////////////////////////////////////////////
  // Move the vertex positions to be more relaxed

  iterQEMesh.Init(this);

  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pt = EvolverMoveVertex(pvx->GetForwardVertex(), fGradientPercent);
      pvx->SetPoint(pt);
    }

  CalculateSubdivisionLimitPositionsAndNormalsSharpLoop();
  fEnergyNew = EvolverEnergy();
  /*
  printf("Energy Old: %f > %f :Energy New, Delta = %f\n", fEnergyOld, fEnergyNew, fEnergyOld - fEnergyNew);
  */

  //Preprocess();

  return fEnergyNew;
}

Point QEMesh::EvolverMoveVertex(QEVertex *pvx, FLOAT fGradientPercent)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peCurr;
  Point pt, ptV, ptE, ppt[3], ptStart, ptB, ptBSum;
  Vector v, vD, vC, vDSum, vCSum, vDInv, vCInv, vGradient;
  Vector pv[3], pvU[2], vN;
  FLOAT fDen, fSharpness;
  FLOAT fA, fEpsilon;
  FLOAT pfGamma[3], fGammaStart;
  UINT_32 uiN, uiSharpEdges, i;

  ptV = pvx->GetPoint();
  ptV[3] = 1.0;
  if ( pvx->GetSharpVertexType() == svtCorner )
    {
      return ptV;
    }

  iterQEVertex.Init(pvx);

  ptE.Null();
  uiSharpEdges = 0;
  for ( pe = iterQEVertex.PeekFirstEdge(), uiN = 0;
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge(), uiN++ )
    {
      ptE += pe->GetEdge()->GetVertex()->GetPoint();

      fSharpness = pe->GetKnotInterval();
      if ( fSharpness == INFINITY || fSharpness >= 1.0 )
	{
	  // Infinitely sharp for this step
	  uiSharpEdges++;
	}
    }

  if ( uiSharpEdges > 0 )
    {
      return ptV;
    }

  ptE /= uiN;

  // Loop limit position Vinf = (e(n) V0 + Sum Vi)/(e(n) + n)
  fA = (3.0 + 2.0*cos(2.0*PI/(FLOAT)uiN));
  fA = fA * fA;
  fA = (5.0 / 8.0) - (fA / 64.0);
  fEpsilon = (3.0 * (FLOAT)uiN) / (8.0 * fA);

  // Gamma is the coefficient on V for the limit position of Vi
  pfGamma[0] = fEpsilon / (fEpsilon + uiN);

  // ppt[i] is the limit position Vi^inf - gamma V
  ppt[0] = pvx->GetLimitPosition();
  ppt[0] += (-pfGamma[0] * ptV);


  //////////////////////////////////////////////////
  // Iterate around the neighbor vertices and set up the system Ax = b
  //   where A is the symmetric matrix
  //   | d0 c2 c1 |
  //   | c2 d1 c0 |
  //   | c1 c0 d2 |

  ptBSum.Null();
  vDSum.Null();
  vCSum.Null();

  pe = iterQEVertex.PeekFirstEdge();
  ASSERT( pe != NULL );

  ppt[2] = pe->GetEdge()->GetVertex()->GetLimitPosition();
  if ( pe->GetEdge()->GetVertex()->GetSharpVertexType() == svtSmooth )
    {
      // Smooth vertex

      for ( peCurr = pe->GetEdge()->GetEdgeNext(1), uiN = 1;
	    peCurr != pe->GetEdge();
	    peCurr = peCurr->GetEdgeNext(1), uiN++ )
	{
	}

      // Loop limit position Vinf = (e(n) V0 + Sum Vi)/(e(n) + n)
      fA = (3.0 + 2.0*cos(2.0*PI/(FLOAT)uiN));
      fA = fA * fA;
      fA = (5.0 / 8.0) - (fA / 64.0);
      fEpsilon = (3.0 * (FLOAT)uiN) / (8.0 * fA);
      
      // Gamma is the coefficient on V for the limit position of Vi
      pfGamma[2] = 1.0 / (fEpsilon + uiN);
      
      // ppt[i] is the limit position Vi^inf - gamma V
      ppt[2] += (-pfGamma[2] * ptV);
    }
  else
    {
      pfGamma[2] = 0.0;
    }

  fGammaStart = pfGamma[2];
  ptStart = ppt[2];

  for ( pe = iterQEVertex.PeekNextEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ppt[1] = ppt[2];
      pfGamma[1] = pfGamma[2];

      ppt[2] = pe->GetEdge()->GetVertex()->GetLimitPosition();
      if ( pe->GetEdge()->GetVertex()->GetSharpVertexType() == svtSmooth )
	{
	  // Smooth vertex
	  
	  for ( peCurr = pe->GetEdge()->GetEdgeNext(1), uiN = 1;
		peCurr != pe->GetEdge();
		peCurr = peCurr->GetEdgeNext(1), uiN++ )
	    {
	    }

	  // Loop limit position Vinf = (e(n) V0 + Sum Vi)/(e(n) + n)
	  fA = (3.0 + 2.0*cos(2.0*PI/(FLOAT)uiN));
	  fA = fA * fA;
	  fA = (5.0 / 8.0) - (fA / 64.0);
	  fEpsilon = (3.0 * (FLOAT)uiN) / (8.0 * fA);
	  
	  // Gamma is the coefficient on V for the limit position of Vi
	  pfGamma[2] = 1.0 / (fEpsilon + uiN);
      
	  // ppt[i] is the limit position Vi^inf - gamma V
	  ppt[2] += (-pfGamma[2] * ptV);
	}
      else
	{
	  pfGamma[2] = 0.0;
	}

      pv[0] = ppt[2] - ppt[1];
      pv[1] = ppt[0] - ppt[2];
      pv[2] = ppt[1] - ppt[0];

      pvU[0] = ppt[1] - ppt[0];
      pvU[1] = ppt[2] - ppt[0];
      vN = pvU[0] * pvU[1];

      v.Null();
      for ( i = 0; i < 3; i++ )
	{
	  v += (pfGamma[i]*pv[i]);
	}

      vD.Set(v(1)*v(1) + v(2)*v(2), v(2)*v(2) + v(0)*v(0), v(0)*v(0) + v(1)*v(1));
      vDSum += vD;
      vC.Set(-v(1)*v(2), -v(2)*v(0), -v(0)*v(1));
      vCSum += vC;

      ptB.Set(v(1)*vN(2) - v(2)*vN(1),
	      v(2)*vN(0) - v(0)*vN(2),
	      v(0)*vN(1) - v(1)*vN(0));

      ptBSum += ptB;
    }

  ppt[1] = ppt[2];
  pfGamma[1] = pfGamma[2];
  ppt[2] = ptStart;
  pfGamma[2] = fGammaStart;

  pv[0] = ppt[2] - ppt[1];
  pv[1] = ppt[0] - ppt[2];
  pv[2] = ppt[1] - ppt[0];
  
  pvU[0] = ppt[1] - ppt[0];
  pvU[1] = ppt[2] - ppt[0];
  vN = pvU[0] * pvU[1];
  
  v.Null();
  for ( i = 0; i < 3; i++ )
    {
      v += (pfGamma[i]*pv[i]);
    }
  
  vD.Set(v(1)*v(1) + v(2)*v(2), v(2)*v(2) + v(0)*v(0), v(0)*v(0) + v(1)*v(1));
  vDSum += vD;
  vC.Set(-v(1)*v(2), -v(2)*v(0), -v(0)*v(1));
  vCSum += vC;
  
  ptB.Set(v(1)*vN(2) - v(2)*vN(1),
	  v(2)*vN(0) - v(0)*vN(2),
	  v(0)*vN(1) - v(1)*vN(0));
  
  ptBSum += ptB;

  //////////////////////////////////////////////////
  // Solve for the area minimizing point
  fDen = (2.0*vCSum(0)*vCSum(1)*vCSum(2) + 
	  vDSum(0)*vDSum(1)*vDSum(2) - 
	  vCSum(0)*vCSum(0)*vDSum(0) - 
	  vCSum(1)*vCSum(1)*vDSum(1) - 
	  vCSum(2)*vCSum(2)*vDSum(2));

  //printf("fDen = %1.10f\n", fDen);
  //if ( fabs(fDen) < 0.000000001 )
  if ( fDen == 0.0 )
    {
      pt = ptE;
    }
  else
    {
      vDInv.Set(vDSum(1)*vDSum(2) - vCSum(0)*vCSum(0), 
		vDSum(2)*vDSum(0) - vCSum(1)*vCSum(1), 
		vDSum(0)*vDSum(1) - vCSum(2)*vCSum(2)); 
      vCInv.Set(vCSum(1)*vCSum(2) - vCSum(0)*vDSum(0), 
		vCSum(2)*vCSum(0) - vCSum(1)*vDSum(1), 
		vCSum(0)*vCSum(1) - vCSum(2)*vDSum(2)); 
      
      pt.Set(vDInv(0)*ptBSum(0) + vCInv(2)*ptBSum(1) + vCInv(1)*ptBSum(2),
		  vCInv(2)*ptBSum(0) + vDInv(1)*ptBSum(1) + vCInv(0)*ptBSum(2),
		  vCInv(1)*ptBSum(0) + vCInv(0)*ptBSum(1) + vDInv(2)*ptBSum(2));
      pt /= fDen;

      /*
      printf("ptE (%f, %f, %f) - pt (%f, %f, %f) = (%f, %f, %f)\n",
	     ptE(0), ptE(1), ptE(2),
	     pt(0), pt(1), pt(2),
	     ptE(0) - pt(0), ptE(1) - pt(1), ptE(2) - pt(2));
      */
    }

  vGradient = pt - ptV;

  pt = ptV + (fGradientPercent * vGradient);
  pt[3] = 1.0;
  /*
  printf("ptV (%f, %f, %f) -> pt (%f, %f, %f)\n",
	 ptV(0), ptV(1), ptV(2),
	 pt(0), pt(1), pt(2));
  */
  return pt;
}

FLOAT QEMesh::EvolverVertexEnergy(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peStart;
  Point ptV, ppt[2];
  Vector pv[2], v;
  FLOAT fArea;
  UINT_32 uiN;

  ptV = pvx->GetLimitPosition();

  iterQEVertex.Init(pvx);

  uiN = 0;
  fArea = 0.0;
  pe = iterQEVertex.PeekFirstEdge();
  ASSERT( pe != NULL );
  peStart = pe;

  ppt[1] = pe->GetEdge()->GetVertex()->GetLimitPosition();

  for ( pe = iterQEVertex.PeekNextEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ppt[0] = ppt[1];
      ppt[1] = pe->GetEdge()->GetVertex()->GetLimitPosition();

      if ( pe->GetFacet() != GetNullFacet() )
	{
	  uiN++;

	  pv[0] = ppt[0] - ptV;
          pv[0][3] = 0.0;
	  pv[1] = ppt[1] - ppt[0];
          pv[1][3] = 0.0;

	  v = pv[0] * pv[1];

	  fArea += sqrt(v.Dot(v));
	}
    }

  pe = peStart;

  ppt[0] = ppt[1];
  ppt[1] = pe->GetEdge()->GetVertex()->GetLimitPosition();
  
  if ( pe->GetFacet() != GetNullFacet() )
    {
      uiN++;

      pv[0] = ppt[0] - ptV;
      pv[0][3] = 0.0;
      pv[1] = ppt[1] - ppt[0];
      pv[1][3] = 0.0;
      
      v = pv[0] * pv[1];
      
      fArea += sqrt(v.Dot(v));
    }

  // The factor of 3 is because the area of each triangle should be counted
  //   1/3 for each of its vertices
  // The factor of 2 is because the cross product computes the area of the 
  //   parallelogram which is twice the area of the triangle
  fArea /= (3 * 2 * uiN);

  return fArea;
}

#else // LOOP_LIMIT_TRIANGLE_GRADIENT

FLOAT QEMesh::EvolverOptimizeStep(FLOAT fGradientPercent/* = 0.25*/)
{
  QEMeshIter iterQEMesh;
  QEVertex *pvx;

  Point pt;
  FLOAT fEnergyNew;

  //////////////////////////////////////////////////
  // Move the vertex positions to be more relaxed

  iterQEMesh.Init(this);

  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pt = EvolverMoveVertex(pvx->GetForwardVertex(), fGradientPercent);
      pvx->SetPoint(pt);
    }

  //CalculateSubdivisionLimitPositionsAndNormalsSharpLoop();
  fEnergyNew = EvolverEnergy();
  /*
  printf("Energy Old: %f > %f :Energy New, Delta = %f\n", fEnergyOld, fEnergyNew, fEnergyOld - fEnergyNew);
  */

  //Preprocess();

  return fEnergyNew;
}

Point QEMesh::EvolverMoveVertex(QEVertex *pvx, FLOAT fGradientPercent)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peStart;
  Point pt, ptV, ptE, ppt[2], ptB, ptBSum;
  Vector v, vD, vC, vDSum, vCSum, vDInv, vCInv, vGradient;
  FLOAT fDen, fSharpness;
  FLOAT fA, fEpsilon;
  UINT_32 uiN, uiSharpEdges;

  ptV = pvx->GetPoint();
  ptV[3] = 1.0;
  if ( pvx->GetSharpVertexType() == svtCorner )
    {
      return ptV;
    }

  ptE.Null();
  ptBSum.Null();
  vDSum.Null();
  vCSum.Null();

  iterQEVertex.Init(pvx);

  pe = iterQEVertex.PeekFirstEdge();
  ASSERT( pe != NULL );
  peStart = pe;

  uiSharpEdges = 0;
  fSharpness = pe->GetKnotInterval();
  if ( fSharpness == INFINITY || fSharpness >= 1.0 )
    {
      // Infinitely sharp for this step
      uiSharpEdges++;
    }

  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();

  for ( pe = iterQEVertex.PeekNextEdge(), uiN = 1;
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge(), uiN++ )
    {
      fSharpness = pe->GetKnotInterval();
      if ( fSharpness == INFINITY || fSharpness >= 1.0 )
	{
	  // Infinitely sharp for this step
	  uiSharpEdges++;
	}

      ppt[0] = ppt[1];
      ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
      ptE += ppt[1];

      v = ppt[1] - ppt[0];

      vD.Set(v(1)*v(1) + v(2)*v(2), v(2)*v(2) + v(0)*v(0), v(0)*v(0) + v(1)*v(1));
      vDSum += vD;
      vC.Set(-v(1)*v(2), -v(2)*v(0), -v(0)*v(1));
      vCSum += vC;

      ptB.Set(vD(0)*ppt[0](0) + vC(2)*ppt[0](1) + vC(1)*ppt[0](2),
	      vC(2)*ppt[0](0) + vD(1)*ppt[0](1) + vC(0)*ppt[0](2),
	      vC(1)*ppt[0](0) + vC(0)*ppt[0](1) + vD(2)*ppt[0](2));

      ptBSum += ptB;
    }

  pe = peStart;
  ppt[0] = ppt[1];
  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
  ptE += ppt[1];
  
  v = ppt[1] - ppt[0];
  
  vD.Set(v(1)*v(1) + v(2)*v(2), v(2)*v(2) + v(0)*v(0), v(0)*v(0) + v(1)*v(1));
  vDSum += vD;
  vC.Set(-v(1)*v(2), -v(2)*v(0), -v(0)*v(1));
  vCSum += vC;
  
  ptB.Set(vD(0)*ppt[0](0) + vC(2)*ppt[0](1) + vC(1)*ppt[0](2),
	  vC(2)*ppt[0](0) + vD(1)*ppt[0](1) + vC(0)*ppt[0](2),
	  vC(1)*ppt[0](0) + vC(0)*ppt[0](1) + vD(2)*ppt[0](2));
  
  ptBSum += ptB;

  if ( uiSharpEdges > 0 )
    {
      return ptV;
    }

  //////////////////////////////////////////////////
  // Solve for the area minimizing point
  fDen = (2.0*vCSum(0)*vCSum(1)*vCSum(2) + 
	  vDSum(0)*vDSum(1)*vDSum(2) - 
	  vCSum(0)*vCSum(0)*vDSum(0) - 
	  vCSum(1)*vCSum(1)*vDSum(1) - 
	  vCSum(2)*vCSum(2)*vDSum(2));

  ptE /= uiN;
  //printf("fDen = %1.10f\n", fDen);
  //if ( fabs(fDen) < 0.000000001 )
  if ( fDen == 0.0 )
    {
      pt = ptE;
    }
  else
    {
      vDInv.Set(vDSum(1)*vDSum(2) - vCSum(0)*vCSum(0), 
		vDSum(2)*vDSum(0) - vCSum(1)*vCSum(1), 
		vDSum(0)*vDSum(1) - vCSum(2)*vCSum(2)); 
      vCInv.Set(vCSum(1)*vCSum(2) - vCSum(0)*vDSum(0), 
		vCSum(2)*vCSum(0) - vCSum(1)*vDSum(1), 
		vCSum(0)*vCSum(1) - vCSum(2)*vDSum(2)); 
      
      pt.Set(vDInv(0)*ptBSum(0) + vCInv(2)*ptBSum(1) + vCInv(1)*ptBSum(2),
		  vCInv(2)*ptBSum(0) + vDInv(1)*ptBSum(1) + vCInv(0)*ptBSum(2),
		  vCInv(1)*ptBSum(0) + vCInv(0)*ptBSum(1) + vDInv(2)*ptBSum(2));
      pt /= fDen;

      /*
      printf("ptE (%f, %f, %f) - pt (%f, %f, %f) = (%f, %f, %f)\n",
	     ptE(0), ptE(1), ptE(2),
	     pt(0), pt(1), pt(2),
	     ptE(0) - pt(0), ptE(1) - pt(1), ptE(2) - pt(2));
      */
    }

  vGradient = pt - ptV;

  pt = ptV + (fGradientPercent * vGradient);
  /*
  // Derive the appropriate placement for the control point to achieve the 
  // desired limit position
  fA = (3.0 + 2.0*cos(2.0*PI/(FLOAT)uiN));
  fA = fA * fA;
  fA = (5.0 / 8.0) - (fA / 64.0);
  fEpsilon = (3.0 * (FLOAT)uiN) / (8.0 * fA);
  pt = (ptE + (fEpsilon + (FLOAT)uiN) * pt) / (fEpsilon);
  pt[3] = 1.0;
  */
  /*
  printf("ptV (%f, %f, %f) -> pt (%f, %f, %f)\n",
	 ptV(0), ptV(1), ptV(2),
	 pt(0), pt(1), pt(2));
  */
  return pt;
}

FLOAT QEMesh::EvolverVertexEnergy(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peStart;
  Point ptV, ppt[2];
  Vector pv[2], v;
  FLOAT fArea;
  UINT_32 uiN;

  ptV = pvx->GetPoint();

  iterQEVertex.Init(pvx);

  uiN = 0;
  fArea = 0.0;
  pe = iterQEVertex.PeekFirstEdge();
  ASSERT( pe != NULL );
  peStart = pe;

  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();

  for ( pe = iterQEVertex.PeekNextEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ppt[0] = ppt[1];
      ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();

      if ( pe->GetFacet() != GetNullFacet() )
	{
	  uiN++;

	  pv[0] = ppt[0] - ptV;
          pv[0][3] = 0.0;
	  pv[1] = ppt[1] - ppt[0];
          pv[1][3] = 0.0;

	  v = pv[0] * pv[1];

	  fArea += sqrt(v.Dot(v));
	}
    }

  pe = peStart;

  ppt[0] = ppt[1];
  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
  
  if ( pe->GetFacet() != GetNullFacet() )
    {
      uiN++;

      pv[0] = ppt[0] - ptV;
      pv[0][3] = 0.0;
      pv[1] = ppt[1] - ppt[0];
      pv[1][3] = 0.0;
      
      v = pv[0] * pv[1];
      
      fArea += sqrt(v.Dot(v));
    }

  // The factor of 3 is because the area of each triangle should be counted
  //   1/3 for each of its vertices
  // The factor of 2 is because the cross product computes the area of the 
  //   parallelogram which is twice the area of the triangle
  fArea /= (3 * 2 * uiN);

  return fArea;
}

#endif // LOOP_LIMIT_TRIANGLE_GRADIENT

#endif // LOOP_CURVATURE_GRADIENT
