#ifndef __QENODE_H__
#define __QENODE_H__

#include "qemem.h"

#include "qevertex.h"
class QEVertex;
class QEFacet;

//////////////////////////////////////////////////////////////////////
// QENode Class
//

class QENode
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  QENode();
  virtual ~QENode();
  virtual VOID Init();
  virtual VOID Uninit();

  virtual BOOL Initialized();
  virtual BOOL Valid();

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual QEFacet *FindFacet(const Point &pt) = 0;

protected:

//////////////////////////////////////////////////
// Rendering
//
public:

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID Display();

//////////////////////////////////////////////////
// Accessor Functions
//
public:

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

//////////////////////////////////////////////////////////////////////
// QEEdgeNode Class
//

class QEEdgeNode : public QENode
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  QEEdgeNode();
  virtual ~QEEdgeNode();
  virtual VOID Init(const Point &pt0, const Point &pt1);
  virtual VOID Uninit();

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual QEFacet *FindFacet(const Point &pt);

protected:

//////////////////////////////////////////////////
// Rendering
//
public:

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID Display();

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  inline QENode *GetChild(UINT_32 uiIndex);
  inline VOID SetChild(UINT_32 uiIndex, QENode *pQENode);

  inline QEEdgeNode *GetPrevLink();
  inline VOID SetPrevLink(QEEdgeNode *pen);
  inline QEEdgeNode *GetNextLink();
  inline VOID SetNextLink(QEEdgeNode *pen);

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  Point m_ppt[2];

  QENode *m_ppQENode[2];

  // Listing in a QEMesh or on a Free List
  QEEdgeNode *m_penLink[2];
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline QENode *QEEdgeNode::GetChild(UINT_32 uiIndex)
{
  ASSERT( uiIndex < 2 );

  return m_ppQENode[uiIndex];
}

inline VOID QEEdgeNode::SetChild(UINT_32 uiIndex, QENode *pQENode)
{
  ASSERT( uiIndex < 2 );

  m_ppQENode[uiIndex] = pQENode;
}

inline QEEdgeNode *QEEdgeNode::GetPrevLink()
{
  return m_penLink[0];
}

inline VOID QEEdgeNode::SetPrevLink(QEEdgeNode *pen)
{
  m_penLink[0] = pen;
}

inline QEEdgeNode *QEEdgeNode::GetNextLink()
{
  return m_penLink[1];
}

inline VOID QEEdgeNode::SetNextLink(QEEdgeNode *pen)
{
  m_penLink[1] = pen;
}

//////////////////////////////////////////////////////////////////////
// QEFacetNode Class
//

class QEFacetNode : public QENode
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  QEFacetNode();
  virtual ~QEFacetNode();
  virtual VOID Init();
  virtual VOID Uninit();

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual QEFacet *FindFacet(const Point &pt);

protected:

//////////////////////////////////////////////////
// Rendering
//
public:

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID Display();

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  inline QEEdgeNode *GetParent(UINT_32 uiIndex);
  inline VOID SetParents(QEEdgeNode *pParent0, QEEdgeNode *pParent1);
  inline QEFacet *GetFacet();
  inline VOID SetFacet(QEFacet *pQEFacet);

  inline QEFacetNode *GetPrevLink();
  inline VOID SetPrevLink(QEFacetNode *pfn);
  inline QEFacetNode *GetNextLink();
  inline VOID SetNextLink(QEFacetNode *pfn);

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  QEEdgeNode *m_ppParent[2];
  QEFacet *m_pQEFacet;

  // Listing in a QEMesh or on a Free List
  QEFacetNode *m_pfnLink[2];
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline QEEdgeNode *QEFacetNode::GetParent(UINT_32 uiIndex)
{
  ASSERT( uiIndex < 2 );

  return m_ppParent[uiIndex];
}

inline VOID QEFacetNode::SetParents(QEEdgeNode *pParent0, QEEdgeNode *pParent1)
{
  m_ppParent[0] = pParent0;
  m_ppParent[1] = pParent1;
}

inline QEFacet *QEFacetNode::GetFacet()
{
  return m_pQEFacet;
}

inline VOID QEFacetNode::SetFacet(QEFacet *pQEFacet)
{
  m_pQEFacet = pQEFacet;
}

inline QEFacetNode *QEFacetNode::GetPrevLink()
{
  return m_pfnLink[0];
}

inline VOID QEFacetNode::SetPrevLink(QEFacetNode *pfn)
{
  m_pfnLink[0] = pfn;
}

inline QEFacetNode *QEFacetNode::GetNextLink()
{
  return m_pfnLink[1];
}

inline VOID QEFacetNode::SetNextLink(QEFacetNode *pfn)
{
  m_pfnLink[1] = pfn;
}

#endif // __QENODE_H__

