#include "qevertex.h"

#include "qeedge.h"
#include "qewedge.h"
#include "qemesh.h"

QEVertex::ComparisonType QEVertex::m_eComparison = QEVertex::ctYXZ;

//////////////////////////////////////////////////////////////////////
// QEVertex Class
//

QEVertex::QEVertex()
{
  UINT_32 i;

  m_uiID = 0;
  m_uiValence = 0;
  m_uiGeneration = 0;
  m_uiBirth = 0;
  m_uiSharpVertexType = QE::svtSmooth;

  m_eICCT = APM2D::icctNull;

  m_pQEWedge = NULL;
  m_pQEEdge = NULL;

  m_pvxForward = NULL;
  m_pfaForward = NULL;

  for ( i = 0; i < 2; i++ )
    {
      m_pphe[i] = NULL;
      m_pvxLink[i] = NULL;
    }
}

QEVertex::~QEVertex()
{
  Uninit();
}

VOID QEVertex::Init(const Point &pt)
{
  UINT_32 i;
  static UINT_32 suiID = 0;

  m_uiID = suiID++;
  m_uiValence = 0;
  m_uiGeneration = 0;
  m_uiBirth = 0;
  m_uiSharpVertexType = QE::svtSmooth;

  m_eICCT = APM2D::icctNull;
  
  m_pt = pt;
  m_ptLimit.Null();
  m_vNormal.Null();

  ASSERT( m_pQEWedge == NULL );
  ASSERT( m_pQEEdge == NULL );

  m_pvxForward = NULL;
  m_pfaForward = NULL;

  for ( i = 0; i < 2; i++ )
    {
      m_pphe[i] = NULL;
      ASSERT( m_pvxLink[i] == NULL );
    }
}

VOID QEVertex::Uninit()
{
  UINT_32 i;
  
  m_pQEWedge = NULL;
  m_pQEEdge = NULL;

  m_pvxForward = NULL;
  m_pfaForward = NULL;

  for ( i = 0; i < 2; i++ )
    {
      m_pphe[i] = NULL;
      m_pvxLink[i] = NULL;
    }
}

BOOL QEVertex::Initialized()
{
  return TRUE;
}

BOOL QEVertex::Valid()
{
  return TRUE;
}

VOID QEVertex::Preprocess()
{
}

Vector QEVertex::CalculateAveragedFacetNormal()
{
  QEVertexIter iter;
  QEFacet *pfa;
  Vector vNormal;
  
  vNormal.Null();
  iter.Init((QEVertex*)this);
  for ( pfa = iter.PeekFirstFacet();
	pfa != NULL;
	pfa = iter.PeekNextFacet() )
    {
      if ( pfa != QEMesh::GetNullFacet() )
	{
	  vNormal += pfa->GetNormal();
	}
    }

  if ( !vNormal.Normalize() )
    {
      fprintf(stderr, "QEVertex Warning: Null normal because neighboring facet normals cancelled.\n");
    }

  // Calculate the D coefficient of the Tangent Plane defined by the normal
  vNormal.CompletePlaneEquation(m_pt);
  
  return vNormal;
}

QEVertex &QEVertex::operator=(const QEVertex &vx)
{
  m_pt = vx.m_pt;

  return *this;
}

BOOL QEVertex::operator<(const QEVertex &vx) const
{
  switch ( m_eComparison )
    {
    case ctXYZ:
      {
	if ( m_pt(X) == vx.m_pt(X) )
	  {
	    return m_pt(Y) < vx.m_pt(Y);
	  }
	return m_pt(X) < vx.m_pt(X);
      }
      break;
    case ctYXZ:
      {
	if ( m_pt(Y) == vx.m_pt(Y) )
	  {
	    return m_pt(X) < vx.m_pt(X);
	  }
	return m_pt(Y) < vx.m_pt(Y);
      }
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  return FALSE;
}

BOOL QEVertex::operator>(const QEVertex &vx) const
{
  switch ( m_eComparison )
    {
    case ctXYZ:
      {
	if ( m_pt(X) == vx.m_pt(X) )
	  {
	    return m_pt(Y) > vx.m_pt(Y);
	  }
	return m_pt(X) > vx.m_pt(X);
      }
      break;
    case ctYXZ:
      {
	if ( m_pt(Y) == vx.m_pt(Y) )
	  {
	    return m_pt(X) > vx.m_pt(X);
	  }
	return m_pt(Y) > vx.m_pt(Y);
      }
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  return FALSE;
}

BOOL QEVertex::operator==(const QEVertex &vx) const
{
  return (m_pt(X) == vx.m_pt(X)) && (m_pt(Y) == vx.m_pt(Y)) && (m_pt(Z) == vx.m_pt(Z));
}

BOOL QEVertex::operator!=(const QEVertex &vx) const
{
  return !(*this == vx);
}

BOOL QEVertex::operator<=(const QEVertex &vx) const
{
  return ((*this) < vx) ||  ((*this) == vx);
}

BOOL QEVertex::operator>=(const QEVertex &vx) const
{
  return ((*this) > vx) ||  ((*this) == vx);
}

VOID QEVertex::Display()
{
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  UINT_32 uiValence;

  iterQEVertex.Init(this);
  uiValence = 0;
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      uiValence++;
    }

  fprintf(stderr, "    Vertex %p: val %ld: ", this, uiValence);
  m_pt.Display(stderr);
}

//////////////////////////////////////////////////////////////////////
// QEVertexIter Class
//

QEVertexIter::QEVertexIter()
{
  m_pQEVertex = NULL;
  m_peHead = NULL;
  m_peCurr = NULL;
}

QEVertexIter::~QEVertexIter()
{
  Uninit();
}

VOID QEVertexIter::Init(QEVertex *pQEVertex)
{
  ASSERT( pQEVertex != NULL );

  m_pQEVertex = pQEVertex;

  m_peHead = NULL;
  m_peCurr = NULL;
}

VOID QEVertexIter::Uninit()
{
  m_pQEVertex = NULL;
  m_peHead = NULL;
  m_peCurr = NULL;
}

BOOL QEVertexIter::Initialized()
{
  return m_pQEVertex != NULL;
}

BOOL QEVertexIter::Valid()
{
  return Initialized();
}

QEFacet *QEVertexIter::PeekFirstFacet()
{
  PeekFirstEdge();

  if ( m_peCurr == NULL )
    {
      return NULL;
    }

  return m_peCurr->GetFacet();
}

QEFacet *QEVertexIter::PeekNextFacet()
{
  PeekNextEdge();

  if ( m_peCurr == NULL )
    {
      return NULL;
    }

  return m_peCurr->GetFacet();
}

QEFacet *QEVertexIter::PeekNextWrapFacet()
{
  PeekNextWrapEdge();

  if ( m_peCurr == NULL )
    {
      return NULL;
    }

  return m_peCurr->GetFacet();
}

QEFacet *QEVertexIter::PeekLastFacet()
{
  PeekLastEdge();

  if ( m_peCurr == NULL )
    {
      return NULL;
    }

  return m_peCurr->GetFacet();
}

QEFacet *QEVertexIter::PeekPrevFacet()
{
  PeekPrevEdge();

  if ( m_peCurr == NULL )
    {
      return NULL;
    }

  return m_peCurr->GetFacet();
}

QEFacet *QEVertexIter::PeekPrevWrapFacet()
{
  PeekPrevWrapEdge();

  if ( m_peCurr == NULL )
    {
      return NULL;
    }

  return m_peCurr->GetFacet();
}

QEEdge *QEVertexIter::PeekFirstEdge()
{
  m_peHead = m_pQEVertex->m_pQEEdge;
  m_peCurr = m_peHead;

  return m_peCurr;
}

QEEdge *QEVertexIter::PeekNextEdge()
{
  if ( m_peCurr == NULL )
    {
      return NULL;
    }

  m_peCurr = m_peCurr->GetEdgeNext(1);
  if ( m_peCurr == m_peHead )
    {
      m_peCurr = NULL;
    }

  return m_peCurr;
}

QEEdge *QEVertexIter::PeekNextWrapEdge()
{
  m_peCurr = m_peCurr->GetEdgeNext(1);

  return m_peCurr;
}

QEEdge *QEVertexIter::PeekLastEdge()
{
  m_peHead = m_pQEVertex->m_pQEEdge;
  m_peCurr = m_peHead->GetEdgePrev(1);

  return m_peCurr;
}

QEEdge *QEVertexIter::PeekPrevEdge()
{
  if ( m_peCurr == NULL )
    {
      return NULL;
    }

  if ( m_peCurr == m_peHead )
    {
      m_peCurr = NULL;
      return NULL;
    }
  m_peCurr = m_peCurr->GetEdgePrev(1);

  return m_peCurr;
}

QEEdge *QEVertexIter::PeekPrevWrapEdge()
{
  m_peCurr = m_peCurr->GetEdgePrev(1);

  return m_peCurr;
}


