#include "qewrite.h"

#include <ri.h>
#include "stl.h"
#include "sif.h"
#include "slidesurface.h"

VOID WriteSLIDEMesh(QEMesh *pme, FILE *fpOutputFile)
{
  QEMeshIter iterQEMesh;
  QEFacetIter iterQEFacet;
  QEVertex *pvx;
  QEFacet *pfa;
  Point pt;
  UINT_32 i;

  iterQEMesh.Init(pme);

  fprintf(fpOutputFile, "mesh oMesh\n");

  for ( pvx = iterQEMesh.PeekFirstVertex(), i = 0;
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex(), i++ )
    {
      pvx->SetID(i);
      pt = pvx->GetPoint();
      fprintf(fpOutputFile, "  vertex v%ld ( %f %f %f ) endvertex\n", 
	      pvx->GetID(), pt(0), pt(1), pt(2));
    }

  for ( pfa = iterQEMesh.PeekFirstFacet(), i = 0;
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet(), i++ )
    {
      iterQEFacet.Init(pfa);
      fprintf(fpOutputFile, "  facet f%ld (", i); 
      for ( pvx = iterQEFacet.PeekFirstVertex();
	    pvx != NULL;
	    pvx = iterQEFacet.PeekNextVertex() )
	{
	  fprintf(fpOutputFile, " v%ld", pvx->GetID());
	}
      fprintf(fpOutputFile, " ) endfacet\n"); 
    }

  fprintf(fpOutputFile, "endmesh\n");
}

VOID WriteRIB(QEMesh *pme, FILE *fpOutputFile, const Matrix &m)
{
  QEMeshIter iterQEMesh;
  QEFacet *pfa;
  QEFacetIter iterQEFacet;
  QEVertex *pvx;
  Point ppt[3];

  iterQEMesh.Init(pme);
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      iterQEFacet.Init(pfa);

      pvx = iterQEFacet.PeekFirstVertex();
      ASSERT( pvx != NULL );
      ppt[0] = pvx->GetPoint();
      ppt[0] = m * ppt[0];

      pvx = iterQEFacet.PeekNextVertex();
      ASSERT( pvx != NULL );
      ppt[1] = pvx->GetPoint();
      ppt[1] = m * ppt[1];

      for ( pvx = iterQEFacet.PeekNextVertex();
	    pvx != NULL;
	    pvx = iterQEFacet.PeekNextVertex() )
	{
	  ppt[2] = pvx->GetPoint();
	  ppt[2] = m * ppt[2];


	  RiArchiveRecord(RI_COMMENT, "\nPolygon \"P\" [%.6f %.6f %.6f %.6f %.6f %.6f %.6f %.6f %.6f]\n",
			  ppt[0][X], ppt[0][Y], ppt[0][Z], 
			  ppt[1][X], ppt[1][Y], ppt[1][Z], 
			  ppt[2][X], ppt[2][Y], ppt[2][Z]);

	  ppt[1] = ppt[2];
	}
    }
}

VOID WriteSTL(QEMesh *pme, FILE *fpOutputFile, const Matrix &m)
{
  QEMeshIter iterQEMesh;
  QEFacet *pfa;
  QEFacetIter iterQEFacet;
  QEVertex *pvx;
  Point ppt[3];

  iterQEMesh.Init(pme);
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      iterQEFacet.Init(pfa);

      pvx = iterQEFacet.PeekFirstVertex();
      ASSERT( pvx != NULL );
      ppt[0] = pvx->GetPoint();
      if ( ppt[0](3) != 1.0 )
	{
	  ppt[0].Display();
	  //ASSERT( FALSE );
	}
      ppt[0] = m * ppt[0];

      pvx = iterQEFacet.PeekNextVertex();
      ASSERT( pvx != NULL );
      ppt[1] = pvx->GetPoint();
      if ( ppt[1](3) != 1.0 )
	{
	  ppt[1].Display();
	  //ASSERT( FALSE );
	}
      ppt[1] = m * ppt[1];

      for ( pvx = iterQEFacet.PeekNextVertex();
	    pvx != NULL;
	    pvx = iterQEFacet.PeekNextVertex() )
	{
	  ppt[2] = pvx->GetPoint();
	  if ( ppt[2](3) != 1.0 )
	    {
	      ppt[2].Display();
	      //ASSERT( FALSE );
	    }
	  ppt[2] = m * ppt[2];

	  ::WriteSTLTriangle(fpOutputFile, ppt[0], ppt[1], ppt[2]);

	  ppt[1] = ppt[2];
	}
    }
}

VOID WriteSIFShell(QEMesh *pme, FILE *fpOutputFile, UINT_32 uiIndentLevel)
{
  QEMeshIter iterQEMesh;
  QEFacetIter iterQEFacet;
  QEFacet *pfa;
  QEEdge *pe, *peCurr;
  QEVertex *pvx, *ppvx[3];
  UINT_32 uiID;
  UINT_32 uiVertices, uiEdges, uiTriangles, uiGenus;
/*
  if ( pme->GetBoundaryEdgeCount() > 0 )
    {
      fprintf(stderr, "ERROR: QEMesh::WriteSIF:  The mesh is not a closed shell:  boundary edges = %ld\n", m_plistQEEdgeBoundary->GetLength());
      return;
    }
*/
  iterQEMesh.Init(pme);

  // Set the vertex ID's
  for ( pvx = iterQEMesh.PeekFirstVertex(), uiID = 0;
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex(), uiID++ )
    {
      pvx->SetID(uiID);
    }

  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "(%s\n", gcppcSIFReservedWords[sifrwtShell]);
  uiIndentLevel++;
  
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "(*\n");
  uiIndentLevel++;
  
  uiEdges = ( pme->GetEdgeCount() + pme->GetBoundaryEdgeCount() ) / 2;
  uiTriangles = 0;
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      iterQEFacet.Init(pfa);
      uiVertices = 0;
      for ( pvx = iterQEFacet.PeekFirstVertex();
	    pvx != NULL;
	    pvx = iterQEFacet.PeekNextVertex() )
	{
	  uiVertices++;
	}
      
      ASSERT( uiVertices > 2 );
      uiTriangles += uiVertices - 2;
      uiEdges += uiVertices - 3;
    }
  uiVertices = pme->GetVertexCount();
  uiGenus = ( ( 2 + uiEdges ) - ( uiVertices + uiTriangles ) ) / 2;
  
  if ( pme->GetBoundaryEdgeCount() > 0 )
    {
      ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
      fprintf(fpOutputFile, "ERROR: Unclosed shell: unmatched edges = %ld\n", 
	      pme->GetBoundaryEdgeCount());
      for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
	    pe != NULL;
	    pe = iterQEMesh.PeekNextEdgeBoundary() )
	{
	  pe->SetForwardEdge(NULL);
	}
      for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
	    pe != NULL;
	    pe = iterQEMesh.PeekNextEdgeBoundary() )
	{
	  if ( pe->GetForwardEdge() == NULL )
	    {
              ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
	      fprintf(fpOutputFile, "(loop");
	      peCurr = pe;
	      do
		{
		  peCurr->SetForwardEdge(peCurr);
		  fprintf(fpOutputFile, " %ld", peCurr->GetVertex()->GetID());
		  peCurr = peCurr->GetEdgeNext(0);
		}
	      while ( peCurr != pe );
	      fprintf(fpOutputFile, ")\n");	      
	    }
	}
    }
  else
    {
      ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
      fprintf(fpOutputFile, "Euler's Formula:  V - E + F = 2 * ( 1 - G )\n");
      ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
      fprintf(fpOutputFile, "                  %ld - %ld + %ld = 2 * ( 1 - %ld )\n",
	      uiVertices, uiEdges, uiTriangles, uiGenus);
      ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
      fprintf(fpOutputFile, "Vertices = %ld\n", uiVertices);
      ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
      fprintf(fpOutputFile, "Edges = %ld\n", uiEdges);
      ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
      fprintf(fpOutputFile, "Triangles = %ld\n", uiTriangles);
      ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
      fprintf(fpOutputFile, "Genus = %ld\n", uiGenus);
    }
  
  uiIndentLevel--;
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "*)\n");
  
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "(%s %ld\n", gcppcSIFReservedWords[sifrwtVertices],
	  uiVertices);
  uiIndentLevel++;
  
  for ( pvx = iterQEMesh.PeekFirstVertex(), uiID = 0;
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex(), uiID++ )
    {
      ASSERT( pvx->GetID() == uiID );
      ::WriteSIFVertex(fpOutputFile, uiIndentLevel, pvx->GetPoint());
    }

  uiIndentLevel--;
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, ")\n");
  
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "(%s %ld\n", gcppcSIFReservedWords[sifrwtTriangles],
	  uiTriangles);
  uiIndentLevel++;
  
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      iterQEFacet.Init(pfa);
      
      ppvx[0] = iterQEFacet.PeekFirstVertex();
      ASSERT( ppvx[0] != NULL );
      
      ppvx[1] = iterQEFacet.PeekNextVertex();
      ASSERT( ppvx[1] != NULL );
      
      for ( ppvx[2] = iterQEFacet.PeekNextVertex();
	    ppvx[2] != NULL;
	    ppvx[2] = iterQEFacet.PeekNextVertex() )
	{
	  ::WriteSIFTriangle(fpOutputFile, uiIndentLevel,
			     ppvx[0]->GetID(),
			     ppvx[1]->GetID(),
			     ppvx[2]->GetID());
	  ppvx[1] = ppvx[2];
	}
    }
  
  uiIndentLevel--;
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, ")\n");

  uiIndentLevel--;
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, ")\n");
}

/*
VOID QEMesh::WriteSIF(FILE *fpOutputFile, UINT_32 uiIndentLevel)
{
  CObjectList<QEVertex> listQEVertex;
  CObjectList<QEFacet> listQEFacet;
  CObjectList<QEVertex> listQEVertexShell;
  CObjectList<QEFacet> listQEFacetShell;
  CObjectListIter<QEFacet> iterQEFacetShell;

  QEMeshIter iterQEMesh;
  QEFacetIter iterQEFacet;
  QEFacet *pfa, *pfaShell;
  QEVertex *pvx, *pvxShell, *ppvx[3];
  UINT_32 uiVertexID, uiFacetID;
  UINT_32 uiVertices, uiEdges, uiTriangles, uiGenus;

  if ( m_plistQEEdgeBoundary->GetLength() > 0 )
    {
      fprintf(stderr, "ERROR: QEMesh::WriteSIF:  The mesh is not closed:  boundary edges = %ld\n", m_plistQEEdgeBoundary->GetLength());
      return;
    }

  // Initialize the vertex and facet ID's to 0
  iterQEMesh.Init(pme);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pvx->SetID(0);
    }

  iterQEMesh.Init(pme);
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      pfa->SetID(0);
    }

  listQEVertex.Init();
  listQEFacet.Init();
  listQEVertexShell.Init();
  listQEFacetShell.Init();

  listQEVertex.Splice(m_plistQEVertex);
  listQEFacet.Splice(m_plistQEFacet);

  for ( pfa = listQEFacet.PeekFirst();
	pfa != NULL;
	pfa = listQEFacet.PeekFirst() )
    {
      uiVertexID = 0;
      uiFacetID = 0;

      WriteSIFShell(fpOutputFile, uiIndentLevel+1, &listQEVertex, &listQEFacet, 
		    &listQEVertexShell, &listQEFacetShell, uiVertexID, uiFacetID, pfa);

      ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
      fprintf(fpOutputFile, "(%s\n", gcppcSIFReservedWords[sifrwtShell]);
      uiIndentLevel++;

      ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
      fprintf(fpOutputFile, "(*");
      uiIndentLevel++;

      uiTriangles = 0;
      uiEdges = 0;
      iterQEFacetShell.Init(&listQEFacetShell);
      for ( pfaShell = iterQEFacetShell.PeekFirst();
	    pfaShell != NULL;
	    pfaShell = iterQEFacetShell.PeekNext() )
	{
	  iterQEFacet.Init(pfaShell);
	  uiVertices = 0;
	  for ( pvx = iterQEFacet.PeekFirstVertex();
		pvx != NULL;
		pvx = iterQEFacet.PeekNextVertex() )
	    {
	      uiVertices++;
	    }

	  ASSERT( uiVertices > 2 );
	  uiTriangles += uiVertices - 2;
	  uiEdges += uiVertices - 2;
	}

      uiVertices = listQEVertexShell.GetLength();
      uiGenus = ( ( 2 + uiEdges ) - ( uiVertices + uiTriangles ) ) / 2;

      ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
      fprintf(fpOutputFile, "Euler's Formula:  V - E + F = 2 * ( 1 - G )\n");
      ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
      fprintf(fpOutputFile, "                 %ld - %ld + %ld = 2 * ( 1 - %ld )\n",
	      uiVertices, uiEdges, uiTriangles, uiGenus);
      ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
      fprintf(fpOutputFile, "Vertices = %ld\n", uiVertices);
      ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
      fprintf(fpOutputFile, "Edges = %ld\n", uiEdges);
      ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
      fprintf(fpOutputFile, "Triangles = %ld\n", uiTriangles);
      ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
      fprintf(fpOutputFile, "Genus = %ld\n", uiGenus);

      uiIndentLevel--;
      ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
      fprintf(fpOutputFile, "*)\n");

      ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
      fprintf(fpOutputFile, "(%s %ld\n", gcppcSIFReservedWords[sifrwtVertices],
	      uiVertices);
      uiIndentLevel++;

      for ( pvxShell = listQEVertexShell.RemoveFirst(), uiVertexID = 1;
	    pvxShell != NULL;
	    pvxShell = listQEVertexShell.RemoveFirst(), uiVertexID++ )
	{
	  ASSERT( pvxShell->GetID() == uiVertexID );
	  m_plistQEVertex->InsertLast(pvxShell);
	  ::WriteSIFVertex(fpOutputFile, uiIndentLevel, pvxShell->GetPoint());
	}
	
      uiIndentLevel--;
      ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
      fprintf(fpOutputFile, ")\n");

      ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
      fprintf(fpOutputFile, "(%s %ld\n", gcppcSIFReservedWords[sifrwtTriangles],
	      uiTriangles);
      uiIndentLevel++;

      for ( pfaShell = listQEFacetShell.RemoveFirst();
	    pfaShell != NULL;
	    pfaShell = listQEFacetShell.RemoveFirst() )
	{
	  m_plistQEFacet->InsertLast(pfaShell);

	  iterQEFacet.Init(pfaShell);

	  ppvx[0] = iterQEFacet.PeekFirstVertex();
	  ASSERT( ppvx[0] != NULL );

	  ppvx[1] = iterQEFacet.PeekNextVertex();
	  ASSERT( ppvx[1] != NULL );

	  for ( ppvx[2] = iterQEFacet.PeekNextVertex();
		ppvx[2] != NULL;
		ppvx[2] = iterQEFacet.PeekNextVertex() )
	    {
	      ::WriteSIFTriangle(fpOutputFile, uiIndentLevel,
				 ppvx[0]->GetID() - 1,
				 ppvx[1]->GetID() - 1,
				 ppvx[2]->GetID() - 1);
	      ppvx[1] = ppvx[2];
	    }
	}

      uiIndentLevel--;
      ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
      fprintf(fpOutputFile, ")\n");

      uiIndentLevel--;
      ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
      fprintf(fpOutputFile, ")\n");
    }
}

VOID QEMesh::WriteSIFShell(FILE *fpOutputFile, UINT_32 uiIndentLevel, 
			   CObjectList<QEVertex> *plistQEVertex,
			   CObjectList<QEFacet> *plistQEFacet,
			   CObjectList<QEVertex> *plistQEVertexShell,
			   CObjectList<QEFacet> *plistQEFacetShell,
			   UINT_32 &uiVertexID, UINT_32 &uiFacetID,
			   QEFacet *pfaShell)
{
  QEFacetIter iterQEFacet;
  QEFacet *pfa;
  QEEdge *pe;
  QEVertex *pvx;

  uiFacetID++;
  pfaShell->SetID(uiFacetID);
  plistQEFacet->Remove(pfaShell);
  plistQEFacetShell->InsertLast(pfaShell);	  

  iterQEFacet.Init(pfaShell);
  
  for ( pe = iterQEFacet.PeekFirstEdge();
	pe != NULL;
	pe = iterQEFacet.PeekNextEdge() )
    {
      pvx = pe->GetVertex();
      if ( pvx->GetID() == 0 )
	{
	  uiVertexID++;
	  pvx->SetID(uiVertexID);
	  plistQEVertex->Remove(pvx);
	  plistQEVertexShell->InsertLast(pvx);	  
	}
      
      pfa = pe->GetEdge()->GetFacet();
      ASSERT( pfa != m_pfaNULL );
      if ( pfa->GetID() == 0 )
	{
	  WriteSIFShell(fpOutputFile, uiIndentLevel, plistQEVertex, plistQEFacet,
			plistQEVertexShell, plistQEFacetShell, uiVertexID, uiFacetID,
			pfa);
	}
    }
}
*/

VOID WriteELE(QEMesh *pme, FILE *fpOutputFile, UINT_32 uiStart /*= 0*/)
{
  QEMeshIter iterQEMesh;
  QEFacet *pfa;
  QEFacetIter iterQEFacet;
  QEVertex *pvx;
  UINT_32 i, j;
  
  fprintf(fpOutputFile, "%ld 3 0\n", pme->GetFacetCount());
  
  iterQEMesh.Init(pme);
  for ( i = uiStart, pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	i++, pfa = iterQEMesh.PeekNextFacet() )
    {
      iterQEFacet.Init(pfa);

      fprintf(fpOutputFile, "%ld", i);
      for ( j = 0, pvx = iterQEFacet.PeekFirstVertex();
	    pvx != NULL;
	    j++, pvx = iterQEFacet.PeekNextVertex() )
	{
	  fprintf(fpOutputFile, " %ld", pvx->GetID());
	}
      ASSERT( j == 3 );

      fprintf(fpOutputFile, "\n");
    }
}

VOID WriteVoronoi(QEMesh *pme, FILE *fpNode, FILE *fpEdge, UINT_32 uiStart /*= 0*/)
{
  QEMeshIter iterQEMesh;
  QEFacet *pfa;
  QEEdge *pe;
  Point ptCenter, pt;
  Vector v, vZ;
  UINT_32 i;
  
  vZ.Set(0.0, 0.0, 1.0);

  iterQEMesh.Init(pme);

  fprintf(fpNode, "%ld 2 0 0\n", pme->GetFacetCount());
  for ( i = uiStart, pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	i++, pfa = iterQEMesh.PeekNextFacet() )
    {
      pfa->SetID(i);
      pt = pfa->CalculateCircumcenter();
      fprintf(fpNode, "%ld %f %f\n", pfa->GetID(), pt(0), pt(1));
    }

  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      pe->SetForwardEdge(NULL);
    }

  i = uiStart;
  fprintf(fpEdge, "%ld 0\n", 
	  (pme->GetEdgeCount() + pme->GetBoundaryEdgeCount()) / 2);
  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      if ( pe->GetForwardEdge() == NULL )
	{
	  pe->SetForwardEdge(pe);
	  pe->GetEdge()->SetForwardEdge(pe->GetEdge());

	  if ( pe->GetFacet() == QEMesh::GetNullFacet() )
	    {
	      pe = pe->GetEdge();
	      ASSERT( pe->GetFacet() != QEMesh::GetNullFacet() );
	    }

	  fprintf(fpEdge, "%ld %ld", i, pe->GetFacet()->GetID());
	  
	  if ( pe->GetEdge()->GetFacet() == QEMesh::GetNullFacet() )
	    {
	      /*
	      ptCenter = pe->GetFacet()->CalculateCircumcenter();

	      pt = ( pe->GetVertex()->GetPoint() + 
		     pe->GetNeighborVertex()->GetPoint() ) / 2.0; 
	      v = pt - ptCenter;
	      v.Normalize();
	      if ( RightOfEdge(pe, ptCenter) )
		{
		  v = -v;
		}
		*/
	      
	      v = pe->GetNeighborVertex()->GetPoint() - pe->GetVertex()->GetPoint();
	      if ( !v.Normalize() )
		{
		  fprintf(stderr, "Write Voronoi: Facet %ld: Zero length edge\n", pe->GetFacet()->GetID());
		}
	      v = v * vZ;
	      v.Normalize();
	      if ( !v.Normalize() )
		{
		  fprintf(stderr, "Write Voronoi: Facet %ld: Zero length direction after cross product\n", pe->GetFacet()->GetID());
		}

	      fprintf(fpEdge, " -1 %f %f\n", v(0), v(1));
	    }
	  else
	    {
	      fprintf(fpEdge, " %ld\n", pe->GetEdge()->GetFacet()->GetID());
	    }

	  i++;
	}
    }

  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      pe->SetForwardEdge(NULL);
    }
}

VOID WriteWRLZ(QEMesh *pme, FILE *fpOutputFile, const Matrix &m)
{
  ASSERT(fpOutputFile != NULL);

  UINT_32 uiIndentLevel;
  QEMeshIter iterQEMesh;
  QEFacetIter iterQEFacet;
  QEFacet *pfa;
  QEVertex *pvx, *ppvx[3];
  UINT_32 uiID;
  UINT_32 uiVertices, uiEdges, uiTriangles, uiGenus;
  Point pt, p1, p2, p3;
  Vector v, v1, v2;

  uiIndentLevel = 1;

  // Set the vertex ID's
  iterQEMesh.Init(pme);
  for ( pvx = iterQEMesh.PeekFirstVertex(), uiID = 0;
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex(), uiID++ )
    {
      pvx->SetID(uiID);
    }

  // Count 
  uiEdges = ( pme->GetEdgeCount() + pme->GetBoundaryEdgeCount() ) / 2;
  uiTriangles = 0;
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      iterQEFacet.Init(pfa);
      uiVertices = 0;
      for ( pvx = iterQEFacet.PeekFirstVertex();
	    pvx != NULL;
	    pvx = iterQEFacet.PeekNextVertex() )
	{
	  uiVertices++;
	}
      
      ASSERT( uiVertices > 2 );
      uiTriangles += uiVertices - 2;
      uiEdges += uiVertices - 3;
    }
  uiVertices = pme->GetVertexCount();
  uiGenus = ( ( 2 + uiEdges ) - ( uiVertices + uiTriangles ) ) / 2;

  // Header
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "geometry IndexedFaceSet {\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "ccw    TRUE\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "convex TRUE\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "solid  FALSE\n\n");
  
  // Triangles
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "# %lu triangles\n", uiTriangles);
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "coordIndex [");

  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      iterQEFacet.Init(pfa);
      
      ppvx[0] = iterQEFacet.PeekFirstVertex();
      ASSERT( ppvx[0] != NULL );
      
      ppvx[1] = iterQEFacet.PeekNextVertex();
      ASSERT( ppvx[1] != NULL );
      
      for ( ppvx[2] = iterQEFacet.PeekNextVertex();
	    ppvx[2] != NULL;
	    ppvx[2] = iterQEFacet.PeekNextVertex() )
	{
	  fprintf(fpOutputFile, "%lu,%lu,%lu,-1, ", 
		  ppvx[0]->GetID(),
		  ppvx[1]->GetID(),
		  ppvx[2]->GetID());

	  ppvx[1] = ppvx[2];
	}
    }
  fprintf(fpOutputFile, "] \n\n");

  // Points
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "# %lu points\n", uiVertices);
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "coord Coordinate {\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+2);
  fprintf(fpOutputFile, "point [");

  for ( pvx = iterQEMesh.PeekFirstVertex(), uiID = 0;
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex(), uiID++ )
    {
      ASSERT( pvx->GetID() == uiID );
      pt = m * pvx->GetPoint();
      
      fprintf(fpOutputFile, "%.6f %.6f %.6f, ", pt[X], pt[Y], pt[Z]);
    }
  fprintf(fpOutputFile, "]\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "}\n\n");

  // Normals
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "normal Normal {\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+2);
  fprintf(fpOutputFile, "vector [");
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      iterQEFacet.Init(pfa);

      ppvx[0] = iterQEFacet.PeekFirstVertex();
      ASSERT( ppvx[0] != NULL );
      
      ppvx[1] = iterQEFacet.PeekNextVertex();
      ASSERT( ppvx[1] != NULL );

      for ( ppvx[2] = iterQEFacet.PeekNextVertex();
	    ppvx[2] != NULL;
	    ppvx[2] = iterQEFacet.PeekNextVertex() )
	{
	  p1 = m * ppvx[0]->GetPoint();
	  p2 = m * ppvx[1]->GetPoint();
	  p3 = m * ppvx[2]->GetPoint();

	  v1 = p3 - p2; 
	  v2 = p1 - p2; 
	  v = v1 * v2;
	  v.Normalize();

	  fprintf(fpOutputFile, "%.6f %.6f %.6f, ", v[X], v[Y], v[Z]);

	  ppvx[1] = ppvx[2];
	}
    }
  fprintf(fpOutputFile, "]\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "}\n\n");

  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "normalPerVertex FALSE\n\n");

  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "} # end geometry\n");
}

VOID WriteWRLZ(QEMesh *pme, FILE *fpOutputFile, const Matrix &m, CSLIDESurface *pCSLIDESurface)
{
  ASSERT(fpOutputFile != NULL);

  UINT_32 i, uiIndentLevel;
  QEMeshIter iterQEMesh;
  QEFacetIter iterQEFacet;
  QEFacet *pfa;
  QEVertex *pvx, *ppvx[3];
  UINT_32 uiID;
  UINT_32 uiVertices, uiEdges, uiTriangles, uiGenus;
  Point pt;
  Vector v;
  CSLIDESurface *pCSLIDEFaceSurface;

  uiIndentLevel = 1;

  // Set the vertex ID's
  iterQEMesh.Init(pme);
  for ( pvx = iterQEMesh.PeekFirstVertex(), uiID = 0;
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex(), uiID++ )
    {
      pvx->SetID(uiID);
    }

  // Count 
  uiEdges = ( pme->GetEdgeCount() + pme->GetBoundaryEdgeCount() ) / 2;
  uiTriangles = 0;
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      iterQEFacet.Init(pfa);
      uiVertices = 0;
      for ( pvx = iterQEFacet.PeekFirstVertex();
	    pvx != NULL;
	    pvx = iterQEFacet.PeekNextVertex() )
	{
	  uiVertices++;
	}
      
      ASSERT( uiVertices > 2 );
      uiTriangles += uiVertices - 2;
      uiEdges += uiVertices - 3;
    }
  uiVertices = pme->GetVertexCount();
  uiGenus = ( ( 2 + uiEdges ) - ( uiVertices + uiTriangles ) ) / 2;

  // Header
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "geometry IndexedFaceSet {\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "ccw    TRUE\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "convex TRUE\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "solid  FALSE\n\n");
  
  // Triangles
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "# %lu triangles\n", uiTriangles);
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "coordIndex [");

  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      iterQEFacet.Init(pfa);
      
      ppvx[0] = iterQEFacet.PeekFirstVertex();
      ASSERT( ppvx[0] != NULL );
      
      ppvx[1] = iterQEFacet.PeekNextVertex();
      ASSERT( ppvx[1] != NULL );
      
      for ( ppvx[2] = iterQEFacet.PeekNextVertex();
	    ppvx[2] != NULL;
	    ppvx[2] = iterQEFacet.PeekNextVertex() )
	{
	  fprintf(fpOutputFile, "%lu,%lu,%lu,-1, ", 
		  ppvx[0]->GetID(),
		  ppvx[1]->GetID(),
		  ppvx[2]->GetID());

	  ppvx[1] = ppvx[2];
	}
    }
  fprintf(fpOutputFile, "] \n\n");

  // Points
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "# %lu points\n", uiVertices);
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "coord Coordinate {\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+2);
  fprintf(fpOutputFile, "point [");

  for ( pvx = iterQEMesh.PeekFirstVertex(), uiID = 0;
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex(), uiID++ )
    {
      ASSERT( pvx->GetID() == uiID );
      pt = m * pvx->GetPoint();
      
      fprintf(fpOutputFile, "%.6f %.6f %.6f, ", pt[X], pt[Y], pt[Z]);
    }
  fprintf(fpOutputFile, "]\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "}\n\n");

  // Normals
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "normal Normal {\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+2);
  fprintf(fpOutputFile, "vector [");
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      pfa->Preprocess(); // Calculate Normal
      v = pfa->GetNormal();

      iterQEFacet.Init(pfa);

      ppvx[0] = iterQEFacet.PeekFirstVertex();
      ASSERT( ppvx[0] != NULL );
      
      ppvx[1] = iterQEFacet.PeekNextVertex();
      ASSERT( ppvx[1] != NULL );

      for ( ppvx[2] = iterQEFacet.PeekNextVertex();
	    ppvx[2] != NULL;
	    ppvx[2] = iterQEFacet.PeekNextVertex() )
	{
	  fprintf(fpOutputFile, "%.6f %.6f %.6f, ", v[X], v[Y], v[Z]);
	}
    }
  fprintf(fpOutputFile, "]\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "}\n\n");

  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "normalPerVertex FALSE\n\n");

  // Color
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "color Color {\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+2);
  fprintf(fpOutputFile, "color [");
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      pCSLIDEFaceSurface = pfa->GetSurface();

      iterQEFacet.Init(pfa);

      ppvx[0] = iterQEFacet.PeekFirstVertex();
      ASSERT( ppvx[0] != NULL );
      
      ppvx[1] = iterQEFacet.PeekNextVertex();
      ASSERT( ppvx[1] != NULL );

      for ( ppvx[2] = iterQEFacet.PeekNextVertex();
	    ppvx[2] != NULL;
	    ppvx[2] = iterQEFacet.PeekNextVertex() )
	{
	  if( pCSLIDEFaceSurface == CSLIDESurface::m_pCSLIDESurfaceDefault
	      || pCSLIDEFaceSurface == NULL )
	    {
	      pCSLIDESurface->WriteWRLZColor(fpOutputFile);
	    }
	  else
	    {
	      pCSLIDEFaceSurface->WriteWRLZColor(fpOutputFile);
	    }
	}
    }
  fprintf(fpOutputFile, "]\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "}\n\n");

  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "colorIndex [");
  for ( pfa = iterQEMesh.PeekFirstFacet(), i=0;
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      iterQEFacet.Init(pfa);

      ppvx[0] = iterQEFacet.PeekFirstVertex();
      ASSERT( ppvx[0] != NULL );
      
      ppvx[1] = iterQEFacet.PeekNextVertex();
      ASSERT( ppvx[1] != NULL );

      for ( ppvx[2] = iterQEFacet.PeekNextVertex();
	    ppvx[2] != NULL;
	    ppvx[2] = iterQEFacet.PeekNextVertex(), i++ )
	{
	  fprintf(fpOutputFile, "%lu, ", i);
	}
    }
  fprintf(fpOutputFile, "]\n\n");

  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "colorPerVertex FALSE\n\n");


  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "} # end geometry\n");
}

VOID WriteWRLZQuads(QEMesh *pme, FILE *fpOutputFile, const Matrix &m)
{
  ASSERT(fpOutputFile != NULL);

  UINT_32 uiIndentLevel;
  QEMeshIter iterQEMesh;
  QEFacetIter iterQEFacet;
  QEFacet *pfa;
  QEVertex *pvx, *ppvx[4];
  UINT_32 uiID;
  UINT_32 uiVertices, uiQuads, uiTriangles, uiFacets, uiFacet;
  Point pt, p1, p2, p3;
  Vector v, v1, v2;
  UINT_32 *puiQuads;

  uiIndentLevel = 1;

  // Set the vertex ID's
  iterQEMesh.Init(pme);
  for ( pvx = iterQEMesh.PeekFirstVertex(), uiID = 0;
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex(), uiID++ )
    {
      pvx->SetID(uiID);
    }

  // Count 
  uiTriangles = 0;
  uiFacets = pme->GetFacetCount();
  puiQuads = new UINT_32[uiFacets];

  for ( pfa = iterQEMesh.PeekFirstFacet(), uiFacet=0;
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet(), uiFacet++ )
    {
      iterQEFacet.Init(pfa);
      uiVertices = 0;
      for ( pvx = iterQEFacet.PeekFirstVertex();
	    pvx != NULL;
	    pvx = iterQEFacet.PeekNextVertex() )
	{
	  uiVertices++;
	}
      
      ASSERT( uiVertices > 2 );

      if( uiVertices == 4)
	{
	  uiQuads++; 
	  puiQuads[uiFacet] = 1;
	}
      else
	{
	  uiTriangles += uiVertices - 2;
	  puiQuads[uiFacet] = 0;
	}
    }
  uiVertices = pme->GetVertexCount();

  // Header
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "geometry IndexedFaceSet {\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "ccw    TRUE\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "convex TRUE\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "solid  FALSE\n\n");
  
  // Quads
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "# %lu quads  %lu triangles\n", uiQuads, uiTriangles);
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "coordIndex [");

  for ( pfa = iterQEMesh.PeekFirstFacet(), uiFacet=0;
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet(), uiFacet++ )
    {
      iterQEFacet.Init(pfa);

      if( puiQuads[uiFacet] == 1 ) // Quad
	{
	  ppvx[0] = iterQEFacet.PeekFirstVertex();
	  ASSERT( ppvx[0] != NULL );    
	  ppvx[1] = iterQEFacet.PeekNextVertex();
	  ASSERT( ppvx[1] != NULL );
	  ppvx[2] = iterQEFacet.PeekNextVertex();
	  ASSERT( ppvx[2] != NULL );
	  ppvx[3] = iterQEFacet.PeekNextVertex();
	  ASSERT( ppvx[3] != NULL );

	  fprintf(fpOutputFile, "%lu,%lu,%lu,%lu,-1, ", 
		  ppvx[0]->GetID(),
		  ppvx[1]->GetID(),
		  ppvx[2]->GetID(),
		  ppvx[3]->GetID());
	}
      else // Triangles
	{
	  ppvx[0] = iterQEFacet.PeekFirstVertex();
	  ASSERT( ppvx[0] != NULL );
	  
	  ppvx[1] = iterQEFacet.PeekNextVertex();
	  ASSERT( ppvx[1] != NULL );
	  
	  for ( ppvx[2] = iterQEFacet.PeekNextVertex();
		ppvx[2] != NULL;
		ppvx[2] = iterQEFacet.PeekNextVertex() )
	    {
	      fprintf(fpOutputFile, "%lu,%lu,%lu,-1, ", 
		      ppvx[0]->GetID(),
		      ppvx[1]->GetID(),
		      ppvx[2]->GetID());
	      
	      ppvx[1] = ppvx[2];
	    }
	}
    }
  fprintf(fpOutputFile, "] \n\n");

  // Points
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "# %lu points\n", uiVertices);
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "coord Coordinate {\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+2);
  fprintf(fpOutputFile, "point [");

  for ( pvx = iterQEMesh.PeekFirstVertex(), uiID = 0;
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex(), uiID++ )
    {
      ASSERT( pvx->GetID() == uiID );
      pt = m * pvx->GetPoint();
      
      fprintf(fpOutputFile, "%.6f %.6f %.6f, ", pt[X], pt[Y], pt[Z]);
    }
  fprintf(fpOutputFile, "]\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "}\n\n");

  // No Normals

  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "} # end geometry\n");

  delete[] puiQuads;
}

VOID WriteWRLZQuads(QEMesh *pme, FILE *fpOutputFile, const Matrix &m, 
		    CSLIDESurface *pCSLIDESurface)
{
  ASSERT(fpOutputFile != NULL);

  UINT_32 uiIndentLevel, i;
  QEMeshIter iterQEMesh;
  QEFacetIter iterQEFacet;
  QEFacet *pfa;
  QEVertex *pvx, *ppvx[4];
  UINT_32 uiID;
  UINT_32 uiVertices, uiQuads, uiTriangles, uiFacets, uiFacet;
  Point pt, p1, p2, p3;
  Vector v, v1, v2;
  CSLIDESurface *pCSLIDEFaceSurface;
  UINT_32 *puiQuads;

  uiIndentLevel = 1;

  // Set the vertex ID's
  iterQEMesh.Init(pme);
  for ( pvx = iterQEMesh.PeekFirstVertex(), uiID = 0;
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex(), uiID++ )
    {
      pvx->SetID(uiID);
    }

  // Count 
  uiTriangles = 0;
  uiFacets = pme->GetFacetCount();
  puiQuads = new UINT_32[uiFacets];

  for ( pfa = iterQEMesh.PeekFirstFacet(), uiFacet=0;
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet(), uiFacet++ )
    {
      iterQEFacet.Init(pfa);
      uiVertices = 0;
      for ( pvx = iterQEFacet.PeekFirstVertex();
	    pvx != NULL;
	    pvx = iterQEFacet.PeekNextVertex() )
	{
	  uiVertices++;
	}
      
      ASSERT( uiVertices > 2 );

      if( uiVertices == 4)
	{
	  uiQuads++; 
	  puiQuads[uiFacet] = 1;
	}
      else
	{
	  uiTriangles += uiVertices - 2;
	  puiQuads[uiFacet] = 0;
	}
    }
  uiVertices = pme->GetVertexCount();

  // Header
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "geometry IndexedFaceSet {\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "ccw    TRUE\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "convex TRUE\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "solid  FALSE\n\n");
  
  // Quads
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "# %lu quads  %lu triangles\n", uiQuads, uiTriangles);
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "coordIndex [");

  for ( pfa = iterQEMesh.PeekFirstFacet(), uiFacet=0;
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet(), uiFacet++ )
    {
      iterQEFacet.Init(pfa);

      if( puiQuads[uiFacet] == 1 ) // Quad
	{
	  ppvx[0] = iterQEFacet.PeekFirstVertex();
	  ASSERT( ppvx[0] != NULL );    
	  ppvx[1] = iterQEFacet.PeekNextVertex();
	  ASSERT( ppvx[1] != NULL );
	  ppvx[2] = iterQEFacet.PeekNextVertex();
	  ASSERT( ppvx[2] != NULL );
	  ppvx[3] = iterQEFacet.PeekNextVertex();
	  ASSERT( ppvx[3] != NULL );

	  fprintf(fpOutputFile, "%lu,%lu,%lu,%lu,-1, ", 
		  ppvx[0]->GetID(),
		  ppvx[1]->GetID(),
		  ppvx[2]->GetID(),
		  ppvx[3]->GetID());
	}
      else // Triangles
	{
	  ppvx[0] = iterQEFacet.PeekFirstVertex();
	  ASSERT( ppvx[0] != NULL );
	  
	  ppvx[1] = iterQEFacet.PeekNextVertex();
	  ASSERT( ppvx[1] != NULL );
	  
	  for ( ppvx[2] = iterQEFacet.PeekNextVertex();
		ppvx[2] != NULL;
		ppvx[2] = iterQEFacet.PeekNextVertex() )
	    {
	      fprintf(fpOutputFile, "%lu,%lu,%lu,-1, ", 
		      ppvx[0]->GetID(),
		      ppvx[1]->GetID(),
		      ppvx[2]->GetID());
	      
	      ppvx[1] = ppvx[2];
	    }
	}
    }
  fprintf(fpOutputFile, "] \n\n");

  // Points
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "# %lu points\n", uiVertices);
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "coord Coordinate {\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+2);
  fprintf(fpOutputFile, "point [");

  for ( pvx = iterQEMesh.PeekFirstVertex(), uiID = 0;
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex(), uiID++ )
    {
      ASSERT( pvx->GetID() == uiID );
      pt = m * pvx->GetPoint();
      
      fprintf(fpOutputFile, "%.6f %.6f %.6f, ", pt[X], pt[Y], pt[Z]);
    }
  fprintf(fpOutputFile, "]\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "}\n\n");

  // No Normals

  // Color
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "color Color {\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+2);
  fprintf(fpOutputFile, "color [");
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      pCSLIDEFaceSurface = pfa->GetSurface();

      iterQEFacet.Init(pfa);

      ppvx[0] = iterQEFacet.PeekFirstVertex();
      ASSERT( ppvx[0] != NULL );
      
      ppvx[1] = iterQEFacet.PeekNextVertex();
      ASSERT( ppvx[1] != NULL );

      for ( ppvx[2] = iterQEFacet.PeekNextVertex();
	    ppvx[2] != NULL;
	    ppvx[2] = iterQEFacet.PeekNextVertex() )
	{
	  if( pCSLIDEFaceSurface == CSLIDESurface::m_pCSLIDESurfaceDefault
	      || pCSLIDEFaceSurface == NULL )
	    {
	      pCSLIDESurface->WriteWRLZColor(fpOutputFile);
	    }
	  else
	    {
	      pCSLIDEFaceSurface->WriteWRLZColor(fpOutputFile);
	    }
	}
    }
  fprintf(fpOutputFile, "]\n");
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "}\n\n");

  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "colorIndex [");
  for ( pfa = iterQEMesh.PeekFirstFacet(), uiFacet=0, i=0;
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet(), uiFacet++ )
    {
      if( puiQuads[uiFacet] == 1 ) // Quad
	{
	  fprintf(fpOutputFile, "%lu, ", i);
	  i++;
	}
      else // triangles
	{
	  iterQEFacet.Init(pfa);	  
	  ppvx[0] = iterQEFacet.PeekFirstVertex();
	  ASSERT( ppvx[0] != NULL );	  
	  ppvx[1] = iterQEFacet.PeekNextVertex();
	  ASSERT( ppvx[1] != NULL );

	  for ( ppvx[2] = iterQEFacet.PeekNextVertex();
		ppvx[2] != NULL;
		ppvx[2] = iterQEFacet.PeekNextVertex() )
	    {
	      fprintf(fpOutputFile, "%lu, ", i);
	      i++;
	    }
	}
    }
  fprintf(fpOutputFile, "]\n\n");

  ::WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "colorPerVertex FALSE\n\n");

  // end
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "} # end geometry\n");

  delete[] puiQuads;
}

