#include "qewritelsif.h"

// The number of line segments you would like to use to approximate a unit circle
static const UINT_32 s_cuiThetaSlicesMax = 30;
static const FLOAT s_cfThetaDelta = DEGTORAD(2.0);

VOID WriteLSIFOffsetContours(QEMesh *pmeOffset, FILE *fpOutputFile, UINT_32 /*uiIndentLevel*/)
{
  ASSERT( pmeOffset != NULL );

  QEMeshIter iterQEMesh;
  QEEdge *peContour, *pe;
  Point ptCenter, ppt[2], pt;
  Vector pv[2], v, vAxis;
  Matrix mRot;
  FLOAT fDot, fMag, fTheta, fThetaDelta, fR, fEpsilon;
  UINT_32 i, uiSlices, uiVertex;

  fEpsilon = 0.001;

  iterQEMesh.Init(pmeOffset);
  for ( peContour = iterQEMesh.PeekFirstEdge();
	peContour != NULL;
	peContour = iterQEMesh.PeekNextEdge() )
    {
      if ( peContour->GetFacet() == QEMesh::GetNullFacet() )
	{
	  peContour->SetForwardEdge(NULL);
	}
    }

  for ( peContour = iterQEMesh.PeekFirstEdge();
	peContour != NULL;
	peContour = iterQEMesh.PeekNextEdge() )
    {
      if ( peContour->GetFacet() != QEMesh::GetNullFacet() &&
	   peContour->GetEdge()->GetForwardEdge() == NULL )
	{
	  fprintf(fpOutputFile, "      (contour\n");
	  uiVertex = 0;
	  pe = peContour;
	  do
	    {
	      pe->GetEdge()->SetForwardEdge(pe->GetEdge());

	      if ( pe->GetForwardEdge() == NULL )
		{
		  ppt[0] = pe->GetEdge()->GetVertex()->GetPoint();
		  ppt[1] = pe->GetVertex()->GetPoint();
	      
		  fprintf(fpOutputFile, "        (v %ld %f %f)\n", 
			  uiVertex, ppt[1](0), ppt[1](1));
		  uiVertex++;
		}
	      else
		{
		  //////////////////////////////////////////////////
		  // Circular Arc
	      
		  ptCenter = pe->GetForwardEdge()->GetVertex()->GetPoint();
		  ppt[0] = pe->GetEdge()->GetVertex()->GetPoint();
		  ppt[1] = pe->GetVertex()->GetPoint();
		  pv[0] = ppt[0] - ptCenter;
		  pv[0][2] = 0.0;
		  pv[0][3] = 0.0;
		  fR = pv[0].Magnitude();
	      
		  pv[1] = ppt[1] - ptCenter;
		  pv[1][2] = 0.0;
		  pv[1][3] = 0.0;
	      
		  fDot = pv[0].Dot(pv[1]);
		  vAxis = pv[0] * pv[1];
		  fMag = vAxis.Magnitude();
		  vAxis.Normalize();
	      
		  fTheta = atan2(fMag, fDot);
	      
		  fThetaDelta = DEGTORAD(10.0);
		  if ( fR != 0.0 )
		    {
		      fR = 1.0 - (fEpsilon/fR);
		      if ( fR > 0.0 )
			{
			  fR = 2.0*acos(fR);
			  if ( fThetaDelta > fR )
			    {
			      fThetaDelta = fR;
			    }
			}
		    }
		  uiSlices = (UINT_32)(1.0 + fTheta/fThetaDelta);
		  uiSlices = MIN(uiSlices, s_cuiThetaSlicesMax);
		  //printf("Circle Slices = %ld\n", uiSlices);
	      
		  fThetaDelta = fTheta / (FLOAT)uiSlices;
	      
		  mRot.SetRotate(vAxis, fThetaDelta);
		
		  v = pv[0];
	      
		  v[2] = ppt[0][2]; 
	      
		  pt = ptCenter + v;
		  for ( i = 0; i < uiSlices; i++ )
		    {
		      fprintf(fpOutputFile, "        (v %ld %f %f)\n", 
			      uiVertex, pt(0), pt(1));
		      uiVertex++;
		  
		      v = mRot * v;
		  
		      pt = ptCenter + v;
		    }
		}

	      pe = pe->GetEdgePrev(0);
	    }
	  while ( pe != peContour );

	  fprintf(fpOutputFile, "\n        ");
	  for ( i = 0; i < uiVertex; i++ )
	    {
	      fprintf(fpOutputFile, "%ld ", i);
	    }
	  
	  fprintf(fpOutputFile, "\n      )\n");
	}
    }

  for ( peContour = iterQEMesh.PeekFirstEdge();
	peContour != NULL;
	peContour = iterQEMesh.PeekNextEdge() )
    {
      if ( peContour->GetFacet() == QEMesh::GetNullFacet() )
	{
	  peContour->SetForwardEdge(NULL);
	}
    }
}
