%{
// Hack because stdio.h is not defining fileno for us for some reason!!!
#define YY_NEVER_INTERACTIVE 1

#include "slideparser.h"
#include "slide_.tab.h"

#define yylval slide_lval
extern YYSTYPE yylval;    // holds start line and column of token.
#define yylloc slide_lloc
extern YYLTYPE yylloc;    // holds start line and column of token.
CScanner *slide_gpCScanner = NULL;

#undef yywrap
#define yywrap() slide_gpCScanner->PopFile()

#define FLEX_DEBUG 1

	// Define macro to return token and update column number.
#define ReturnToken(x) \
	slide_gpCScanner->ConsumeToken(); \
	return (x)

#define ReturnTokenInt() \
	slide_gpCScanner->ScanInteger(yytext); \
	return tokCONST_INT

#define ReturnTokenFloat() \
	slide_gpCScanner->ScanFloat(yytext); \
	return tokCONST_FLOAT

#define ReturnTokenString() \
	slide_gpCScanner->ScanString(); \
	return tokCONST_STRING

#define ReturnTokenTclBlock() \
	slide_gpCScanner->ScanTclBlock(); \
	return tokTCL_BLOCK

#define ReturnTokenTclMacroBlock() \
	slide_gpCScanner->ScanTclBlock(); \
	return tokTCL_MACRO_BLOCK

#define ReturnTokenIdentifier() \
	slide_gpCScanner->ScanIdentifier(yytext); \
	return tokIDENTIFIER

#define TABSIZE 8	// Size of tab.

%}

OCT     [0-7]
DIGIT	[0-9]
HEX     [0-9A-Fa-f]
LETTER	[A-Za-z_]
IDTAIL	[A-Za-z0-9_]
ID	{LETTER}{IDTAIL}*
STR     [^\0]|"\\\""
SPACE	[
 \t\f\v\a]

INT	"-"?{DIGIT}+
FLOAT	"-"?({DIGIT}*\.)?{DIGIT}+([eE][-+]?{DIGIT}+)?

ESC_SEQ ("\\x"{HEX}{1,2})|("\\"{OCT}{1,3})|("\\".)


%x COMMENT_STATE
%x STRING_STATE
%x TCL_BLOCK_STATE
%x INCLUDE_STATE

%%

<COMMENT_STATE>"(*"	{slide_gpCScanner->CommentStart();}
<COMMENT_STATE>"*)"	{
			  if ( slide_gpCScanner->CommentEnd() )
			    {
			      BEGIN 0;
			    }
			}
<COMMENT_STATE>"\n"	{slide_gpCScanner->ConsumeNewLine();}
<COMMENT_STATE>[*(]	{slide_gpCScanner->ConsumeToken();}
<COMMENT_STATE>[^*(\n]+	{slide_gpCScanner->ConsumeToken();}

<STRING_STATE>"\""	{slide_gpCScanner->StringEnd(); BEGIN 0; ReturnTokenString();}
<STRING_STATE>"\n"	{slide_gpCScanner->StringAppendNewLine();}
<STRING_STATE>"\\\n"	{slide_gpCScanner->ConsumeNewLine();}
<STRING_STATE>{ESC_SEQ}	{slide_gpCScanner->StringAppendEscapeSequence(yytext);}
<STRING_STATE>[^"\\\n]+	{slide_gpCScanner->StringAppend(yytext);}

<TCL_BLOCK_STATE>"%{"	{slide_gpCScanner->TclBlockStart();}
<TCL_BLOCK_STATE>"}%"	{
			  if ( slide_gpCScanner->TclBlockEnd() )
			    {
			      BEGIN 0;
			      ReturnTokenTclMacroBlock();
			    }
			}

<TCL_BLOCK_STATE>"{"	{slide_gpCScanner->TclBlockStart();}
<TCL_BLOCK_STATE>"}"	{
			  if ( slide_gpCScanner->TclBlockEnd() )
			    {
			      BEGIN 0;
			      ReturnTokenTclBlock();
			    }
			}
<TCL_BLOCK_STATE>"\n"	{slide_gpCScanner->TclBlockAppendNewLine();}
<TCL_BLOCK_STATE>"\\\n"	{slide_gpCScanner->ConsumeNewLine();}
<TCL_BLOCK_STATE>"\\{"	{slide_gpCScanner->TclBlockAppend(yytext);}
<TCL_BLOCK_STATE>"\\}"	{slide_gpCScanner->TclBlockAppend(yytext);}
<TCL_BLOCK_STATE>[^{}\\\n]+	{slide_gpCScanner->TclBlockAppend(yytext);}

<INCLUDE_STATE>"\""	{
			  slide_gpCScanner->FileNameEnd(); 
			  BEGIN 0; 
			  slide_gpCScanner->PushFile(slide_gpCScanner->ScanFileName());
			}
<INCLUDE_STATE>"\n"	        {slide_gpCScanner->FileNameAppendNewLine();}
<INCLUDE_STATE>"\\\n"	        {slide_gpCScanner->ConsumeNewLine();}
<INCLUDE_STATE>[^"\\\n]+	{slide_gpCScanner->FileNameAppend(yytext);}

\n		{slide_gpCScanner->ConsumeNewLine();}
{SPACE}+	{slide_gpCScanner->ConsumeToken();}

"tclinit"	{ReturnToken(rwTCL_INIT);}
"tclupdate"	{ReturnToken(rwTCL_UPDATE);}
"tcltag"	{ReturnToken(rwTCL_TAG);}
"mergeepsilon"	{ReturnToken(rwMERGE_EPSILON);}
"ribbegin"	{ReturnToken(rwRIB_BEGIN);}
"ribend"	{ReturnToken(rwRIB_END);}
"graphposition"  {ReturnToken(rwGRAPH_POSITION);}
"window"       	{ReturnToken(rwWINDOW);}
"endwindow"	{ReturnToken(rwEND_WINDOW);}
"display"	{ReturnToken(rwDISPLAY);}
"viewport"	{ReturnToken(rwVIEWPORT);}
"endviewport"	{ReturnToken(rwEND_VIEWPORT);}
"origin"	{ReturnToken(rwORIGIN);}
"size"		{ReturnToken(rwSIZE);}
"resize"	{ReturnToken(rwRESIZE);}
"ratio"		{ReturnToken(rwASPECT_RATIO);}
"SLF_FIXED"	{ReturnToken(rwSLF_FIXED);}
"SLF_ANY"	{ReturnToken(rwSLF_ANY);}
"SLF_FULL_SCREEN"	{ReturnToken(rwSLF_FULL_SCREEN);}
"background"	{ReturnToken(rwBACKGROUND);}
"camera"	{ReturnToken(rwCAMERA);}
"endcamera"	{ReturnToken(rwEND_CAMERA);}
"projection"	{ReturnToken(rwPROJECTION);}
"frustum"	{ReturnToken(rwFRUSTUM);}
"light"       	{ReturnToken(rwLIGHT);}
"endlight"	{ReturnToken(rwEND_LIGHT);}
"type"       	{ReturnToken(rwTYPE);}
"deaddistance"	{ReturnToken(rwDEAD_DISTANCE);}
"falloff"	{ReturnToken(rwFALL_OFF);}
"angularfalloff"       	{ReturnToken(rwANGULAR_FALL_OFF);}
"SLF_AMBIENT"	{ReturnToken(rwSLF_AMBIENT);}
"SLF_DIRECTIONAL"	{ReturnToken(rwSLF_DIRECTIONAL);}
"SLF_POINT"	{ReturnToken(rwSLF_POINT);}
"SLF_SPOT"	{ReturnToken(rwSLF_SPOT);}
"render"       	{ReturnToken(rwRENDER);}
"endrender"	{ReturnToken(rwEND_RENDER);}
"arealight"	{ReturnToken(rwAREA_LIGHT);}
"endarealight"	{ReturnToken(rwEND_AREA_LIGHT);}
"samples"	{ReturnToken(rwSAMPLES);}
"group"		{ReturnToken(rwGROUP);}
"endgroup"	{ReturnToken(rwEND_GROUP);}
"instance"	{ReturnToken(rwINSTANCE);}
"endinstance"	{ReturnToken(rwEND_INSTANCE);}
"seam"		{ReturnToken(rwSEAM);}
"endseam"	{ReturnToken(rwEND_SEAM);}
"contour"	{ReturnToken(rwCONTOUR);}
"endcontour"	{ReturnToken(rwEND_CONTOUR);}
"segment"	{ReturnToken(rwSEGMENT);}
"endsegment"	{ReturnToken(rwEND_SEGMENT);}
"side"		{ReturnToken(rwSIDE);}
"reverse"		{ReturnToken(rwREVERSE);}
"camerainstance"	{ReturnToken(rwINSTANCE);}
"endcamerainstance"	{ReturnToken(rwEND_INSTANCE);}
"lightinstance"		{ReturnToken(rwINSTANCE);}
"endlightinstance"	{ReturnToken(rwEND_INSTANCE);}
"transform"	{ReturnToken(rwTRANSFORM);}
"endtransform"	{ReturnToken(rwEND_TRANSFORM);}
"translate"	{ReturnToken(rwTRANSLATE);}
"scale"		{ReturnToken(rwSCALE);}
"rotate"	{ReturnToken(rwROTATE);}
"lookat"	{ReturnToken(rwLOOK_AT);}
"endlookat"	{ReturnToken(rwEND_LOOK_AT);}
"eye"		{ReturnToken(rwEYE);}
"target"	{ReturnToken(rwTARGET);}
"up"		{ReturnToken(rwUP);}
"object"	{ReturnToken(rwOBJECT);}
"endobject"	{ReturnToken(rwEND_OBJECT);}
"face"		{ReturnToken(rwFACE);}
"endface"	{ReturnToken(rwEND_FACE);}
"point"		{ReturnToken(rwPOINT);}
"endpoint"	{ReturnToken(rwEND_POINT);}
"controlpoint"  {ReturnToken(rwCONTROLPOINT);}
"endcontrolpoint"	{ReturnToken(rwEND_CONTROLPOINT);}
"surface"	{ReturnToken(rwSURFACE);}
"endsurface"	{ReturnToken(rwEND_SURFACE);}
"color"		{ReturnToken(rwCOLOR);}
"reflectivity"	{ReturnToken(rwREFLECTIVITY);}
"exponent"	{ReturnToken(rwEXPONENT);}
"metallic"	{ReturnToken(rwMETALLIC);}
"SLF_INHERIT"	{ReturnToken(rwSLF_INHERIT);}
"fog"		{ReturnToken(rwFOG);}
"endfog"	{ReturnToken(rwEND_FOG);}
"density"		{ReturnToken(rwDENSITY);}
"start"		{ReturnToken(rwSTART);}
"end"		{ReturnToken(rwEND);}
"SLF_LINEAR"		{ReturnToken(rwSLF_LINEAR);}
"SLF_EXP"		{ReturnToken(rwSLF_EXP);}
"SLF_EXP2"		{ReturnToken(rwSLF_EXP2);}
"id"		{ReturnToken(rwID);}
"SLF_OFF"	{ReturnToken(rwSLF_OFF);}
"SLF_ON"	{ReturnToken(rwSLF_ON);}
"solid"		{ReturnToken(rwSOLID);}
"SLF_SOLID"	{ReturnToken(rwSLF_SOLID);}
"SLF_HOLLOW"	{ReturnToken(rwSLF_HOLLOW);}
"shading"	{ReturnToken(rwSHADING);}
"SLF_WIRE"	{ReturnToken(rwSLF_WIRE);}
"SLF_FLAT"	{ReturnToken(rwSLF_FLAT);}
"SLF_GOURAUD"	{ReturnToken(rwSLF_GOURAUD);}
"SLF_PHONG"	{ReturnToken(rwSLF_PHONG);}
"lod"		{ReturnToken(rwLOD);}
"SLF_BOUND"	{ReturnToken(rwSLF_BOUND);}
"SLF_EDGES"	{ReturnToken(rwSLF_EDGES);}
"SLF_FULL"	{ReturnToken(rwSLF_FULL);}
"stencil"	{ReturnToken(rwSTENCIL);}
"SLF_BOTH"	{ReturnToken(rwSLF_BOTH);}
"SLF_ODD"	{ReturnToken(rwSLF_ODD);}
"SLF_EVEN"	{ReturnToken(rwSLF_EVEN);}
"SLF_PARALLEL"	{ReturnToken(rwSLF_PARALLEL);}
"SLF_PERSPECTIVE"	{ReturnToken(rwSLF_PERSPECTIVE);}
"normal"	{ReturnToken(rwNORMAL);}
"SLF_ENVMAP"    {ReturnToken(rwSLF_ENVMAP);}
"SLF_TEXMAP"    {ReturnToken(rwSLF_TEXMAP);}
"bitmap"	{ReturnToken(rwBITMAP);}
"texture"       {ReturnToken(rwTEXTURE);}

"mesh"		{ReturnToken(rwMESH);}
"endmesh"	{ReturnToken(rwEND_MESH);}
"vertex"	{ReturnToken(rwVERTEX);}
"endvertex"	{ReturnToken(rwEND_VERTEX);}
"edge"		{ReturnToken(rwEDGE);}
"endedge"	{ReturnToken(rwEND_EDGE);}
"facet"		{ReturnToken(rwFACET);}
"endfacet"	{ReturnToken(rwEND_FACET);}
"edgeproperty"	{ReturnToken(rwEDGE_PROPERTY);}
"endedgeproperty"	{ReturnToken(rwEND_EDGE_PROPERTY);}
"sharpness"	{ReturnToken(rwSHARPNESS);}
"SLF_INFINITY"	{ReturnToken(rwSLF_INFINITY);}
"vertexproperty"	{ReturnToken(rwVERTEX_PROPERTY);}
"endvertexproperty"	{ReturnToken(rwEND_VERTEX_PROPERTY);}
"SLF_VERTEX_SMOOTH"	{ReturnToken(rwSLF_VERTEX_SMOOTH);}
"SLF_VERTEX_CORNER"	{ReturnToken(rwSLF_VERTEX_CORNER);}

"convexhull"	{ReturnToken(rwCONVEX_HULL);}
"endconvexhull"	{ReturnToken(rwEND_CONVEX_HULL);}
"generalface"	{ReturnToken(rwGENERAL_FACE);}
"endgeneralface"	{ReturnToken(rwEND_GENERAL_FACE);}
"cleanupcontour"	{ReturnToken(rwCLEAN_UP_CONTOUR);}
"millingcontour"	{ReturnToken(rwMILLING_CONTOUR);}
"offsets"	{ReturnToken(rwOFFSETS);}
"offsetmin"	{ReturnToken(rwOFFSET_MIN);}
"offsetmax"	{ReturnToken(rwOFFSET_MAX);}
"individualvoronoimin"	{ReturnToken(rwINDIVIDUAL_VORONOI_MIN);}
"individualvoronoimax"	{ReturnToken(rwINDIVIDUAL_VORONOI_MAX);}
"terrain"	{ReturnToken(rwTERRAIN);}
"endterrain"	{ReturnToken(rwEND_TERRAIN);}
"voronoi"	{ReturnToken(rwVORONOI);}
"accuracy"	{ReturnToken(rwACCURACY);}
"random"	{ReturnToken(rwRANDOM);}
"drawdelaunay"	{ReturnToken(rwDRAW_DELAUNAY);}
"drawvertices"	{ReturnToken(rwDRAW_VERTICES);}
"drawcontours"	{ReturnToken(rwDRAW_CONTOURS);}
"drawCDT"	{ReturnToken(rwDRAW_CDT);}
"drawindividualvoronoi"	{ReturnToken(rwDRAW_INDIVIDUAL_VORONOI);}
"drawvoronoi"	{ReturnToken(rwDRAW_VORONOI);}
"drawheight"	{ReturnToken(rwDRAW_HEIGHT);}
"drawoffset"	{ReturnToken(rwDRAW_OFFSET);}
"drawmillingcontour"	{ReturnToken(rwDRAW_MILLING_CONTOUR);}
"SLF_TERRAIN_DELAUNAY_INCREMENTAL"	{ReturnToken(rwSLF_TERRAIN_DELAUNAY_INCREMENTAL);}
"SLF_TERRAIN_DELAUNAY_INCREMENTAL_SLOW"	{ReturnToken(rwSLF_TERRAIN_DELAUNAY_INCREMENTAL_SLOW);}
"SLF_TERRAIN_DELAUNAY_INCREMENTAL_NO_BOUND"	{ReturnToken(rwSLF_TERRAIN_DELAUNAY_INCREMENTAL_NO_BOUND);}
"SLF_TERRAIN_DELAUNAY_DIVIDE_AND_CONQUER"	{ReturnToken(rwSLF_TERRAIN_DELAUNAY_DIVIDE_AND_CONQUER);}
"SLF_TERRAIN_DELAUNAY_DIVIDE_AND_CONQUER_ALTERNATING"	{ReturnToken(rwSLF_TERRAIN_DELAUNAY_DIVIDE_AND_CONQUER_ALTERNATING);}
"SLF_TERRAIN_DELAUNAY_CONVEX_HULL"	{ReturnToken(rwSLF_TERRAIN_DELAUNAY_CONVEX_HULL);}
"SLF_VORONOI_OFF"	{ReturnToken(rwSLF_VORONOI_OFF);}
"SLF_VORONOI_DELAUNAY"	{ReturnToken(rwSLF_VORONOI_DELAUNAY);}
"SLF_VORONOI_DIVIDE_AND_CONQUER"	{ReturnToken(rwSLF_VORONOI_DIVIDE_AND_CONQUER);}
"SLF_VORONOI_DIVIDE_AND_CONQUER_ALTERNATING"	{ReturnToken(rwSLF_VORONOI_DIVIDE_AND_CONQUER_ALTERNATING);}
"SLF_VORONOI_CONES"	{ReturnToken(rwSLF_VORONOI_CONES);}
"SLF_VORONOI_PLANES"	{ReturnToken(rwSLF_VORONOI_PLANES);}
"SLF_FLOAT_NAIVE"	{ReturnToken(rwSLF_FLOAT_NAIVE);}
"SLF_FLOAT"	{ReturnToken(rwSLF_FLOAT);}
"SLF_EXACT"	{ReturnToken(rwSLF_EXACT);}
"sphere"	{ReturnToken(rwSPHERE);}
"endsphere"	{ReturnToken(rwEND_SPHERE);}
"cylinder"	{ReturnToken(rwCYLINDER);}
"endcylinder"	{ReturnToken(rwEND_CYLINDER);}
"cone"		{ReturnToken(rwCONE);}
"endcone"	{ReturnToken(rwEND_CONE);}
"torus"		{ReturnToken(rwTORUS);}
"endtorus"	{ReturnToken(rwEND_TORUS);}
"cyclide"	{ReturnToken(rwCYCLIDE);}
"endcyclide"	{ReturnToken(rwEND_CYCLIDE);}
"drawellipse"	{ReturnToken(rwDRAW_ELLIPSE);}
"drawhyperbola"	{ReturnToken(rwDRAW_HYPERBOLA);}
"height"	{ReturnToken(rwHEIGHT);}
"width"		{ReturnToken(rwWIDTH);}
"radius"	{ReturnToken(rwRADIUS);}
"majorradius"	{ReturnToken(rwMAJOR_RADIUS);}
"minorradius"	{ReturnToken(rwMINOR_RADIUS);}
"zmin"		{ReturnToken(rwZ_MIN);}
"zmax"		{ReturnToken(rwZ_MAX);}
"phimin"	{ReturnToken(rwPHI_MIN);}
"phimax"	{ReturnToken(rwPHI_MAX);}
"thetamin"	{ReturnToken(rwTHETA_MIN);}
"thetamax"	{ReturnToken(rwTHETA_MAX);}
"zslices"	{ReturnToken(rwZ_SLICES);}
"phislices"	{ReturnToken(rwPHI_SLICES);}
"thetaslices"	{ReturnToken(rwTHETA_SLICES);}
"stringlength"  {ReturnToken(rwSTRING_LENGTH);}

"heightfield"   {ReturnToken(rwHEIGHTFIELD);}
"endheightfield"   {ReturnToken(rwEND_HEIGHTFIELD);}
"func"          {ReturnToken(rwHEIGHTFIELD_FUNC);}
"umin"          {ReturnToken(rwU_MIN);}
"umax"          {ReturnToken(rwU_MAX);}
"vmin"          {ReturnToken(rwV_MIN);}
"vmax"          {ReturnToken(rwV_MAX);}

"bsplinepatch"	{ReturnToken(rwBSPLINEPATCH);}
"endbsplinepatch"	{ReturnToken(rwEND_BSPLINEPATCH);}
"nurbpatch"	{ReturnToken(rwNURBPATCH);}
"endnurbpatch"	{ReturnToken(rwEND_NURBPATCH);}
"bezierpatch"	{ReturnToken(rwBEZIERPATCH);}
"endbezierpatch"	{ReturnToken(rwEND_BEZIERPATCH);}
"uorder"       {ReturnToken(rwU_ORDER);}
"uknots"       {ReturnToken(rwU_KNOTS);}
"uknotlist"    {ReturnToken(rwU_KNOT_LIST);}
"vorder"       {ReturnToken(rwV_ORDER);}
"vknots"       {ReturnToken(rwV_KNOTS);}
"vknotlist"    {ReturnToken(rwV_KNOT_LIST);}
"controlpointlist"     {ReturnToken(rwCONTROL_POINT_LIST);}
"uslices"     {ReturnToken(rwU_SLICES);}
"vslices"     {ReturnToken(rwV_SLICES);}
"drawcontrols"	{ReturnToken(rwDRAW_CONTROLS);}

"bsplinecurve" {ReturnToken(rwBSPLINECURVE);}
"endbsplinecurve" {ReturnToken(rwEND_BSPLINECURVE);}
"beziercurve" {ReturnToken(rwBEZIERCURVE);}
"endbeziercurve" {ReturnToken(rwEND_BEZIERCURVE);}
"order"       {ReturnToken(rwORDER);}
"knots"       {ReturnToken(rwKNOTS);}
"knotlist"    {ReturnToken(rwKNOT_LIST);}
"slices"     {ReturnToken(rwSLICES);}
"minimizetorsion"	{ReturnToken(rwMINIMIZE_TORSION);}
"closed"	{ReturnToken(rwCLOSED);}
"azimuth"	{ReturnToken(rwAZIMUTH);}
"twist"	        {ReturnToken(rwTWIST);}

"polyline"	      {ReturnToken(rwPOLYLINE);}
"endpolyline"    {ReturnToken(rwEND_POLYLINE);}
"pointlist"     {ReturnToken(rwPOINT_LIST);}
"drawpoints"     {ReturnToken(rwDRAW_POINTS);}

"sweepmorph"	{ReturnToken(rwSWEEP_MORPH);}
"endsweepmorph"  {ReturnToken(rwEND_SWEEP_MORPH);}
"sweep"	      {ReturnToken(rwSWEEP);}
"endsweep"    {ReturnToken(rwEND_SWEEP);}
"path"       {ReturnToken(rwPATH);}
"endpath"    {ReturnToken(rwEND_PATH);}
"crosssection"       {ReturnToken(rwCROSS_SECTION);}
"endcrosssection"    {ReturnToken(rwEND_CROSS_SECTION);}
"drawpath"	{ReturnToken(rwDRAW_PATH);}
"drawsweep"	{ReturnToken(rwDRAW_SWEEP);}
"circle"     {ReturnToken(rwCIRCLE);}
"cresent"     {ReturnToken(rwCRESENT);}
"star"     {ReturnToken(rwSTAR);}
"meshtype"     {ReturnToken(rwMESH_TYPE);}
"SLF_TRIANGLES"     {ReturnToken(rwSLF_TRIANGLES);}
"SLF_QUADRILATERALS"     {ReturnToken(rwSLF_QUADRILATERALS);}

"thickness"     {ReturnToken(rwTHICKNESS);}
"theta"     {ReturnToken(rwTHETA);}
"innerradius"     {ReturnToken(rwINNER_RADIUS);}
"outerradius"     {ReturnToken(rwOUTER_RADIUS);}
"symmetry"     {ReturnToken(rwSYMMETRY);}
"orientation"     {ReturnToken(rwORIENTATION);}
"endcap"     {ReturnToken(rwEND_CAP);}
"begincap"     {ReturnToken(rwBEGIN_CAP);}

"subdivision"		{ReturnToken(rwSUBDIVISION);}
"endsubdivision"	{ReturnToken(rwEND_SUBDIVISION);}
"usedisplaylists"	{ReturnToken(rwUSE_DISPLAY_LISTS);}
"drawcurrent"		{ReturnToken(rwDRAW_CURRENT);}
"SLF_SUBDIVISION_DOO_SABIN"		{ReturnToken(rwSLF_SUBDIVISION_DOO_SABIN);}
"SLF_SUBDIVISION_DOO_SABIN_SELECTIVE"	{ReturnToken(rwSLF_SUBDIVISION_DOO_SABIN_SELECTIVE);}
"SLF_SUBDIVISION_DOO_SABIN_SELECTIVE_LIMIT"	{ReturnToken(rwSLF_SUBDIVISION_DOO_SABIN_SELECTIVE_LIMIT);}
"SLF_SUBDIVISION_CATMULL_CLARK"		{ReturnToken(rwSLF_SUBDIVISION_CATMULL_CLARK);}
"SLF_SUBDIVISION_CATMULL_CLARK_TESSELATE"	{ReturnToken(rwSLF_SUBDIVISION_CATMULL_CLARK_TESSELATE);}
"SLF_SUBDIVISION_CATMULL_CLARK_INTERPOLATE"	{ReturnToken(rwSLF_SUBDIVISION_CATMULL_CLARK_INTERPOLATE);}
"SLF_SUBDIVISION_CATMULL_CLARK_C2"		{ReturnToken(rwSLF_SUBDIVISION_CATMULL_CLARK_C2);}
"SLF_SUBDIVISION_CATMULL_CLARK_SELECTIVE"	{ReturnToken(rwSLF_SUBDIVISION_CATMULL_CLARK_SELECTIVE);}
"SLF_SUBDIVISION_CATMULL_CLARK_OLD"	{ReturnToken(rwSLF_SUBDIVISION_CATMULL_CLARK_OLD);}
"SLF_SUBDIVISION_CORNER_CUTTING"	{ReturnToken(rwSLF_SUBDIVISION_CORNER_CUTTING);}
"SLF_SUBDIVISION_CORNER_ROUNDING_TESSELATE"	{ReturnToken(rwSLF_SUBDIVISION_CORNER_ROUNDING_TESSELATE);}
"SLF_SUBDIVISION_CORNER_ROUNDING"	{ReturnToken(rwSLF_SUBDIVISION_CORNER_ROUNDING);}
"SLF_SUBDIVISION_CORNER_ROUNDING_LOOP"	{ReturnToken(rwSLF_SUBDIVISION_CORNER_ROUNDING_LOOP);}
"SLF_SUBDIVISION_CORNER_ROUNDING_SHARP_LOOP"	{ReturnToken(rwSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_LOOP);}
"SLF_SUBDIVISION_CORNER_ROUNDING_BUTTERFLY"	{ReturnToken(rwSLF_SUBDIVISION_CORNER_ROUNDING_BUTTERFLY);}
"SLF_SUBDIVISION_CORNER_ROUNDING_ZORIN"	{ReturnToken(rwSLF_SUBDIVISION_CORNER_ROUNDING_ZORIN);}
"SLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE"	{ReturnToken(rwSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE);}
"SLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE_LIMIT"	{ReturnToken(rwSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE_LIMIT);}
"SLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT"	{ReturnToken(rwSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT);}
"SLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT"	{ReturnToken(rwSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT);}
"SLF_SUBDIVISION_TRIANGLE_TESSELATE"	{ReturnToken(rwSLF_SUBDIVISION_TRIANGLE_TESSELATE);}
"SLF_SUBDIVISION_TRIANGLE_LOOP"		{ReturnToken(rwSLF_SUBDIVISION_TRIANGLE_LOOP);}
"SLF_SUBDIVISION_TRIANGLE_LOOP_EDGE_RULE"	{ReturnToken(rwSLF_SUBDIVISION_TRIANGLE_LOOP_EDGE_RULE);}
"SLF_SUBDIVISION_TRIANGLE_BUTTERFLY"	{ReturnToken(rwSLF_SUBDIVISION_TRIANGLE_BUTTERFLY);}
"SLF_SUBDIVISION_TRIANGLE_ZORIN"	{ReturnToken(rwSLF_SUBDIVISION_TRIANGLE_ZORIN);}
"SLF_SUBDIVISION_TRIANGLE_SPHERE"	{ReturnToken(rwSLF_SUBDIVISION_TRIANGLE_SPHERE);}
"SLF_SUBDIVISION_SQRT_3"	{ReturnToken(rwSLF_SUBDIVISION_SQRT_3);}
"SLF_SUBDIVISION_SQRT_3_SPHERE_BARYCENTER"	{ReturnToken(rwSLF_SUBDIVISION_SQRT_3_SPHERE_BARYCENTER);}
"SLF_SUBDIVISION_SQRT_3_SPHERE_CENTER"	{ReturnToken(rwSLF_SUBDIVISION_SQRT_3_SPHERE_CENTER);}
"SLF_SUBDIVISION_SQRT_3_SPHERE_RADIUS"	{ReturnToken(rwSLF_SUBDIVISION_SQRT_3_SPHERE_RADIUS);}
"SLF_SUBDIVISION_SQRT_3_SPHERE_CAP_AREA"	{ReturnToken(rwSLF_SUBDIVISION_SQRT_3_SPHERE_CAP_AREA);}
"SLF_SUBDIVISION_SQRT_3_SPHERE_TRIANGLE_AREA"	{ReturnToken(rwSLF_SUBDIVISION_SQRT_3_SPHERE_TRIANGLE_AREA);}
"SLF_SUBDIVISION_SQRT_3_SPHERE_TANGENT"	{ReturnToken(rwSLF_SUBDIVISION_SQRT_3_SPHERE_TANGENT);}
"SLF_SUBDIVISION_SQRT_3_SELECTIVE"	{ReturnToken(rwSLF_SUBDIVISION_SQRT_3_SELECTIVE);}
"SLF_SUBDIVISION_SQRT_3_SELECTIVE_FACET"	{ReturnToken(rwSLF_SUBDIVISION_SQRT_3_SELECTIVE_FACET);}
"SLF_SUBDIVISION_SQRT_4"	{ReturnToken(rwSLF_SUBDIVISION_SQRT_4);}
"SLF_SUBDIVISION_SQRT_4_SELECTIVE"	{ReturnToken(rwSLF_SUBDIVISION_SQRT_4_SELECTIVE);}
"SLF_SUBDIVISION_PROGRESSIVE"	{ReturnToken(rwSLF_SUBDIVISION_PROGRESSIVE);}
"SLF_ERROR_METRIC_BEST_FIT_PLANE"	{ReturnToken(rwSLF_ERROR_METRIC_BEST_FIT_PLANE);}
"SLF_ERROR_METRIC_LIMIT_DISTANCE"	{ReturnToken(rwSLF_ERROR_METRIC_LIMIT_DISTANCE);}
"SLF_SUBDIVISION_HETEROGENEOUS"		{ReturnToken(rwSLF_SUBDIVISION_HETEROGENEOUS);}
"SLF_SUBDIVISION_HETEROGENEOUS_STAM_LOOP"		{ReturnToken(rwSLF_SUBDIVISION_HETEROGENEOUS_STAM_LOOP);}
"SLF_SUBDIVISION_HETEROGENEOUS_OLD"		{ReturnToken(rwSLF_SUBDIVISION_HETEROGENEOUS_OLD);}
"SLF_SUBDIVISION_HETEROGENEOUS_TESSELATE"	{ReturnToken(rwSLF_SUBDIVISION_HETEROGENEOUS_TESSELATE);}
"SLF_SUBDIVISION_PENTAGON"		{ReturnToken(rwSLF_SUBDIVISION_PENTAGON);}
"SLF_SUBDIVISION_PENTAGON_TESSELATE"	{ReturnToken(rwSLF_SUBDIVISION_PENTAGON_TESSELATE);}
"SLF_SUBDIVISION_GEOMETRIC"		{ReturnToken(rwSLF_SUBDIVISION_GEOMETRIC);}
"SLF_SUBDIVISION_GEOMETRIC_TESSELATE"	{ReturnToken(rwSLF_SUBDIVISION_GEOMETRIC_TESSELATE);}
"subdivisions"		{ReturnToken(rwSUBDIVISIONS);}
"errormetric"		{ReturnToken(rwERROR_METRIC);}
"tolerance"		{ReturnToken(rwTOLERANCE);}
"facetsmax"		{ReturnToken(rwFACETS_MAX);}
"offset"	{ReturnToken(rwOFFSET);}
"endoffset"	{ReturnToken(rwEND_OFFSET);}
"SLF_OFFSET_ANGULAR"      {ReturnToken(rwSLF_OFFSET_ANGULAR);}
"SLF_OFFSET_ROUNDED"      {ReturnToken(rwSLF_OFFSET_ROUNDED);}
"SLF_OFFSET_PERFORATED"      {ReturnToken(rwSLF_OFFSET_PERFORATED);}
"dual"		{ReturnToken(rwDUAL);}
"enddual"	{ReturnToken(rwEND_DUAL);}
"evolver"		{ReturnToken(rwEVOLVER);}
"endevolver"	{ReturnToken(rwEND_EVOLVER);}
"mirror"	{ReturnToken(rwMIRROR);}
"endmirror"	{ReturnToken(rwEND_MIRROR);}
"geodesic"	{ReturnToken(rwGEODESIC);}
"endgeodesic"	{ReturnToken(rwEND_GEODESIC);}

"union"		{ReturnToken(rwUNION);}
"endunion"	{ReturnToken(rwEND_UNION);}
"intersection"	{ReturnToken(rwINTERSECTION);}
"endintersection"	{ReturnToken(rwEND_INTERSECTION);}
"difference"	{ReturnToken(rwDIFFERENCE);}
"enddifference"	{ReturnToken(rwEND_DIFFERENCE);}
"sgroot"	{ReturnToken(rwSGROOT);}
"endsgroot"	{ReturnToken(rwEND_SGROOT);}

"cameraui"	{ReturnToken(rwCAMERAUI);}
"SLF_RESET_CUI"	{ReturnToken(rwSLF_RESET_CUI);}
"SLF_PAUSE_CUI"	{ReturnToken(rwSLF_PAUSE_CUI);}
"SLF_CAD_CUI"	{ReturnToken(rwSLF_CAD_CUI);}
"SLF_FLY_CUI"	{ReturnToken(rwSLF_FLY_CUI);}
"SLF_EXPLORE_CUI"	{ReturnToken(rwSLF_EXPLORE_CUI);}
"camerauispeed" {ReturnToken(rwCAMERAUISPEED);}

"wormwarp"	{ReturnToken(rwWORMWARP);}
"pathonoff"     {ReturnToken(rwPATHONOFF);}
"endwormwarp"	{ReturnToken(rwEND_WORMWARP);}

"("		{ReturnToken('(');}
")"		{ReturnToken(')');}
"."		{ReturnToken('.');}

"(*"		{slide_gpCScanner->CommentStart(); BEGIN COMMENT_STATE;}
"*)"		{
		  slide_gpCScanner->HandleError(0, 0, "Mismatched comment parentheses");
		  slide_gpCScanner->ConsumeToken();
		}
"#".*		{slide_gpCScanner->ConsumeToken();}

"\""		{slide_gpCScanner->StringStart(); BEGIN STRING_STATE;}

"{"		{slide_gpCScanner->TclBlockStart(); BEGIN TCL_BLOCK_STATE;}
"}"		{
		  slide_gpCScanner->HandleError(0, 0, "Mismatched brace");
		  slide_gpCScanner->ConsumeToken();
		}

"%{"	{slide_gpCScanner->TclBlockStart(); BEGIN TCL_BLOCK_STATE;}
"}%"	{
		  slide_gpCScanner->HandleError(0, 0, "Mismatched macro brace");
		  slide_gpCScanner->ConsumeToken();
		}


"include"{SPACE}+"\"" {slide_gpCScanner->FileNameStart(); BEGIN INCLUDE_STATE;}

{INT}		{ReturnTokenInt();}
{FLOAT}		{ReturnTokenFloat();}

{ID}		{ReturnTokenIdentifier();}

.		{
		  slide_gpCScanner->HandleError(0, 0, "Unknown character \"%s\"", 
		                               yylloc.text);
		  slide_gpCScanner->ConsumeToken();
		}

%%

//"general"	{ReturnToken(rwGENERAL);}

VOID slide_BeginInitial()
{
  BEGIN 0;
}

YY_BUFFER_STATE slide_GetCurrentBuffer()
{
  return YY_CURRENT_BUFFER;
}

YY_BUFFER_STATE slide_CreateBuffer(FILE *fpFile)
{
  return yy_create_buffer(fpFile, YY_BUF_SIZE);
}

VOID slide_DeleteBuffer(YY_BUFFER_STATE yyBufferState)
{
  yy_delete_buffer(yyBufferState);
}

VOID slide_SwitchToBuffer(YY_BUFFER_STATE yyBufferState)
{
  yy_switch_to_buffer(yyBufferState);
}

VOID slide_SetValueInt(int i)
{
  yylval.i = i;
}

VOID slide_SetValueFloat(FLOAT f)
{
  yylval.f = f;
}

VOID slide_SetValueString(CHAR *pc)
{
  yylval.str = pc;
}








