#include "slidearealight.h"

#include "slideworld.h"
#include "slidesurface.h"
#include "slideinstance.h"
#include "slidelight.h"

static const VECTOR4D gcvColor = { 1.0, 1.0, 1.0, 0.0 };
static const UINT_32 cuiSamples = 30;

//////////////////////////////////////////////////////////////////////
// CSLIDEAreaLight Class
//

CSLIDEAreaLight::CSLIDEAreaLight()
{
  m_vColor.Set(gcvColor);
  m_dyniSamples.SetValue(cuiSamples);
}

CSLIDEAreaLight::~CSLIDEAreaLight()
{
}

VOID CSLIDEAreaLight::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		       CHAR *pcID)
{
  CSLIDEGroup::Init(pcFile, uiLine, uiColumn, pcID);

  m_vColor.Set(gcvColor);
  m_dyniSamples.SetValue(cuiSamples);
}

VOID CSLIDEAreaLight::Uninit()
{
  CSLIDEGroup::Uninit();
}

BOOL CSLIDEAreaLight::Initialized()
{
  if ( !CSLIDEGroup::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

BOOL CSLIDEAreaLight::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  if ( !CSLIDEGroup::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  return TRUE;
}

BOOL CSLIDEAreaLight::Valid()
{
  if ( !CSLIDEGroup::Valid() )
    {
      return FALSE;
    }

  return TRUE;
}

CSLIDECamera *CSLIDEAreaLight::FindCamera(CListIterInt *piterIDs, 
				       CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtCamera],
	      gcppcSLIDEReservedWords[sliderwtAreaLight], GetID());

  return NULL;
}

CSLIDELight *CSLIDEAreaLight::FindLight(CListIterInt *piterIDs, 
				     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtLight],
	      gcppcSLIDEReservedWords[sliderwtAreaLight], GetID());

  return NULL;
}

CSLIDEAreaLight *CSLIDEAreaLight::FindAreaLight(CListIterInt *piterIDs, 
				    CListInt *plistCSLIDEInstance)
{
  CHAR *pcID;
  CSLIDEAreaLight *pCSLIDEAreaLight;

  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  pCSLIDEAreaLight = this;

  pcID = (CHAR*)piterIDs->PeekNext();
  if ( pcID == NULL )
    {
      pcID = (CHAR*)piterIDs->PeekFirst();
      if ( pcID != GetID() )
	{
	  HandleError("Bad %s path: %s id \"%s\" not specified", 
		      gcppcSLIDEReservedWords[sliderwtAreaLight],
		      gcppcSLIDEReservedWords[sliderwtAreaLight], GetID());
	  pCSLIDEAreaLight = NULL;
	}
    }
  else if ( pcID != GetID() )
    {
      HandleError("Bad %s path: path id \"%s\" != %s id \"%s\"", 
		  gcppcSLIDEReservedWords[sliderwtAreaLight], pcID,
		  gcppcSLIDEReservedWords[sliderwtAreaLight], GetID());
      pCSLIDEAreaLight = NULL;
    }

  return pCSLIDEAreaLight;
}

#ifdef JORDAN_IS_SO_COOL
VOID CSLIDEAreaLight::Preprocess()
{
  CSLIDEGroup::Preprocess();
}
#endif // JORDAN_IS_SO_COOL

VOID CSLIDEAreaLight::CheckDynamic()
{
  CSLIDEGroup::CheckDynamic();

  SetDynamic(GetDynamic() || 
	     m_dyniSamples.GetDynamic());
}

#ifdef JORDAN_IS_SO_COOL
VOID CSLIDEAreaLight::PreprocessGraph()
{
  if ( !GetVisited() )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;

      BOOL bGeometryEmpty;
      BOOL bGeometryDynamic;
      BOOL bNormalsEmpty;
      BOOL bNormalsDynamic;
      BOOL bBoundsEmpty;
      BOOL bBoundsDynamic;
      BOOL bBoundEmpty;
      BOOL bBoundDynamic;

      bGeometryEmpty = TRUE;
      bGeometryDynamic = FALSE;
      bNormalsEmpty = TRUE;
      bNormalsDynamic = FALSE;
      bBoundsEmpty = TRUE;
      bBoundsDynamic = FALSE;

      bBoundEmpty = TRUE;
      bBoundDynamic = FALSE;

      //////////////////////////////////////////////////
      // Visit Children Nodes and get starting values from them
      iter.Init(&m_listCSLIDEInstance);
      
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pCSLIDEInstance->PreprocessGraph();

	  bGeometryEmpty = ( bGeometryEmpty && 
			     pCSLIDEInstance->GetGeometryEmpty() );
	  bGeometryDynamic = ( bGeometryDynamic ||
			       pCSLIDEInstance->GetGeometryDynamic() );
	  bNormalsEmpty = ( bNormalsEmpty && 
			    pCSLIDEInstance->GetNormalsEmpty() );
	  bNormalsDynamic = ( bNormalsDynamic ||
			      pCSLIDEInstance->GetNormalsDynamic() );
	  bBoundsEmpty = ( bBoundsEmpty && 
			   pCSLIDEInstance->GetBoundsEmpty() );
	  bBoundsDynamic = ( bBoundsDynamic ||
			     pCSLIDEInstance->GetBoundsDynamic() );
	  bBoundEmpty = ( bBoundEmpty && 
			  pCSLIDEInstance->GetBoundEmpty() );
	  bBoundDynamic = ( bBoundDynamic ||
			    pCSLIDEInstance->GetBoundDynamic() );
	}

      //////////////////////////////////////////////////
      // Look at Local flags

      // Update whether we are Empty based on our LOD flag
      if ( !m_dyniLOD.GetDynamic() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      // This node is always off
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	      bBoundsEmpty = TRUE;
	      bBoundEmpty = TRUE;
	    }
	  else if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	    {
	      // This node is always just a bounding box
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	    }
	}

      // Check what is really dynamic

      bGeometryDynamic = ( bGeometryDynamic ||
			   m_pCSLIDESurface->GetDynamic() ||
			   m_dyniShading.GetDynamic() ||
			   m_dyniLOD.GetDynamic() );
      bNormalsDynamic = ( bNormalsDynamic ||
			  m_dyniShading.GetDynamic() ||
			  m_dyniLOD.GetDynamic() );
      bBoundsDynamic = ( bBoundsDynamic || 
			 m_dyniLOD.GetDynamic() );
      bBoundDynamic = ( bBoundDynamic || 
			m_dyniLOD.GetDynamic() );

      // If anything is now Empty then it will always be empty so
      // it is not dynamic

      bGeometryDynamic = ( bGeometryDynamic && !bGeometryEmpty );
      bNormalsDynamic = ( bNormalsDynamic && !bNormalsEmpty );
      bBoundsDynamic = ( bBoundsDynamic && !bBoundsEmpty );
      bBoundDynamic = ( bBoundDynamic && !bBoundEmpty );

      //////////////////////////////////////////////////
      // Now save these correct flag values

      SetGeometryEmpty(bGeometryEmpty);
      SetGeometryDynamic(bGeometryDynamic);
      SetNormalsEmpty(bNormalsEmpty);
      SetNormalsDynamic(bNormalsDynamic);
      SetBoundsEmpty(bBoundsEmpty);
      SetBoundsDynamic(bBoundsDynamic);
      SetBoundEmpty(bBoundEmpty);
      SetBoundDynamic(bBoundDynamic);
      
      //////////////////////////////
      // START: Jordan Fix This!!!
      //

      SetGeometryEmpty(FALSE);
      SetGeometryDynamic(TRUE);
      SetNormalsEmpty(FALSE);
      SetNormalsDynamic(TRUE);
      SetBoundsEmpty(FALSE);
      SetBoundsDynamic(TRUE);
      SetBoundEmpty(FALSE);
      SetBoundDynamic(TRUE);
      
      //
      // END: Jordan Fix This!!!
      //////////////////////////////

      //////////////////////////////////////////////////
      // Calculate the Bounding Box (if Empty this will do nothing)
      if ( GetBoundEmpty() )
	{
	  SetBoundValid(TRUE);
	}
      else
	{
	  CalculateBound();
	}
      SetBoundChanged(FALSE);

      //////////////////////////////////////////////////
      // Done Preprocessing this Node in the Graph
      SetVisited(TRUE);
    }
}

VOID CSLIDEAreaLight::Update()
{
  SetBoundChanged(FALSE);
  if ( GetDynamic() )
    {
      UpdateState();
    }

  if ( GetBoundsDynamic() )
    {
      CSLIDEBoundedNode::InvalidateQEMesh();
    }

  if ( GetBoundDynamic() )
    {
      SetBoundValid(FALSE);
    }

  SetVisited(FALSE);
}
#endif // JORDAN_IS_SO_COOL

VOID CSLIDEAreaLight::UpdateState()
{
  CSLIDEGroup::UpdateState();

  m_dyniSamples.Update();
}

#ifdef JORDAN_IS_SO_COOL
VOID CSLIDEAreaLight::UpdateGraph()
{
  if ( !GetVisited() )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;
      BOOL bBoundChanged;

      if ( !GetBoundValid() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      if ( !m_bound.GetEmpty() )
		{
		  SetBoundChanged(TRUE);
		  m_bound.Null();
		}
	      return;
	    }
	  
	  bBoundChanged = FALSE;
	  
	  iter.Init(&m_listCSLIDEInstance);
	  
	  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
		pCSLIDEInstance != NULL;
		pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	    {
	      pCSLIDEInstance->UpdateGraph();
	      bBoundChanged = bBoundChanged || 
		pCSLIDEInstance->GetBoundChanged();
	    }
	  
	  if ( bBoundChanged )
	    {
	      CalculateBound();
	    }
	  
	  SetBoundValid(TRUE);
	}

      SetVisited(TRUE);
    }
}
#endif // JORDAN_IS_SO_COOL

GLbitfield CSLIDEAreaLight::CalculateAttributesMask(SLIDEShadingType eShading)
{
  return ( CSLIDEGeometryNode::CalculateAttributesMask(eShading) |
	   GL_LIGHTING_BIT );
}

VOID CSLIDEAreaLight::Place()
{
  SLIDELightType eType;
  GLfloat v[4] = {0.0, 0.0, 0.0, 1.0};

  BOOL m_bLocalViewer;

  m_bLocalViewer = TRUE;
  eType = slideltSLF_SPOT;
  //eType = (SLIDELightType)m_dyniType.GetValue();
  ASSERT( eType < slideltSize );

  if ( eType == slideltSLF_AMBIENT )
    {
      UINT_32 i;

      // Add to the global ambient light

      gilGetFloatv(GL_LIGHT_MODEL_AMBIENT, v);
      for ( i = 0; i < 4; i++ )
	{
	  v[i] += m_vColor[i];
	}
      gilLightModelfv(GL_LIGHT_MODEL_AMBIENT, v);
    }
  else
    {
      GLenum eLightID;
      FLOAT d0, n1;

      if ( eType != slideltSLF_DIRECTIONAL && !m_bLocalViewer )
	{
	  m_bLocalViewer = TRUE;
	  gilLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
	}

      // Allocate a GL Light Structure (there are only 8)
      if ( CSLIDELight::m_uiGLLightsAllocated >= cuiGLLights )
	{
	  HandleWarning("No more GL Lights could not enable light \"%s\"",
			GetID());
	  return;
	}

      eLightID = CSLIDELight::m_peGLLightID[CSLIDELight::m_uiGLLightsAllocated];
      CSLIDELight::m_uiGLLightsAllocated++;

      // Enable this GL Light
      gilEnable(eLightID);

      // Set all the values of the light

      v[0] = 0.0;
      v[1] = 0.0;
      v[2] = 0.0;
      v[3] = 1.0;
      gilLightfv(eLightID, GL_AMBIENT, v);
      //m_vColor.Get(v);
      v[0] = m_vColor[0];
      v[1] = m_vColor[1];
      v[2] = m_vColor[2];
      v[3] = 1.0;
      gilLightfv(eLightID, GL_DIFFUSE, v);
      gilLightfv(eLightID, GL_SPECULAR, v);

      // Distance Attenuation
      //d0 = m_dynfDeadDistance.GetValue();
      //n1 = m_dynfFallOff.GetValue();
      d0 = (FLOAT)0.1;
      n1 = 1.0;
      n1 = MIN(n1, (FLOAT)2.0); // OpenGL can not handle any exponents higher than this
      gilLightf(eLightID, GL_CONSTANT_ATTENUATION, (FLOAT)pow(d0, n1));
      gilLightf(eLightID, GL_LINEAR_ATTENUATION, n1*(FLOAT)pow(d0, MAX(n1 - 1.0, 0.0)));
      if ( n1 <= 1.0 )
	{
	  gilLightf(eLightID, GL_QUADRATIC_ATTENUATION, 0.0);
	}
      else
	{	  
	  //gilLightf(eLightID, GL_QUADRATIC_ATTENUATION, (n1*(n1 - 1.0)/2.0)*pow(d0, MAX(n1 - 2.0, 0.0)));
          // This simplifies to the following for all valid values of n1
	  gilLightf(eLightID, GL_QUADRATIC_ATTENUATION, (FLOAT)(n1*(n1 - 1.0)/2.0));
	}

      if ( eType == slideltSLF_DIRECTIONAL )
	{
	  v[0] = 0.0;
	  v[1] = 0.0;
	  v[2] = 1.0;
	  v[3] = 0.0;
	  gilLightfv(eLightID, GL_POSITION, v);
	  gilLightf(eLightID, GL_SPOT_EXPONENT, 0.0);
	  gilLightf(eLightID, GL_SPOT_CUTOFF, 180.0);
	}
      else if ( eType == slideltSLF_POINT )
	{
	  v[0] = 0.0;
	  v[1] = 0.0;
	  v[2] = 0.0;
	  v[3] = 1.0;
	  gilLightfv(eLightID, GL_POSITION, v);
	  gilLightf(eLightID, GL_SPOT_EXPONENT, 0.0);
	  gilLightf(eLightID, GL_SPOT_CUTOFF, 180.0);
	}
      else
	{
	  FLOAT n2;

	  ASSERT( eType == slideltSLF_SPOT );

	  v[0] = 0.0;
	  v[1] = 0.0;
	  v[2] = 0.0;
	  v[3] = 1.0;
	  gilLightfv(eLightID, GL_POSITION, v);
	  //n2 = m_dynfAngularFallOff.GetValue();
	  n2 = 1.0;
	  gilLightf(eLightID, GL_SPOT_EXPONENT, n2);
	  gilLightf(eLightID, GL_SPOT_CUTOFF, 90.0);
	}

      v[0] = 0.0;
      v[1] = 0.0;
      v[2] = -1.0;
      v[3] = 0.0;
      gilLightfv(eLightID, GL_SPOT_DIRECTION, v);      
    }
}

VOID CSLIDEAreaLight::PrerenderGeometry(SLIDEShadingType eShading)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  SLIDEShadingType eShade;

  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  eShade = (SLIDEShadingType)m_dyniShading.GetValue();

  GLbitfield GLbAttributesMask;
  
  GLbAttributesMask = CalculateAttributesMask(eShade);
  
  PushAttributes(GLbAttributesMask);
  {
    PushShadingAttributes(eShade);
    {
      m_pCSLIDESurface->Push();
      {
	GLfloat v[4];
	v[0] = m_vColor[0];
	v[1] = m_vColor[1];
	v[2] = m_vColor[2];
	v[3] = 1.0;
	gilMaterialfv(GL_FRONT, GL_EMISSION, v);

	m_uiPrerenderAttributesStackDepthMax = 0;

	iter.Init(&m_listCSLIDEInstance);
	
	for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	      pCSLIDEInstance != NULL;
	      pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	  {
	    pCSLIDEInstance->PrerenderGeometry(eShading);
	    m_uiPrerenderAttributesStackDepthMax = 
	      MAX(m_uiPrerenderAttributesStackDepthMax,
		  pCSLIDEInstance->GetPrerenderAttributesStackDepthMax());
	  }

	if ( GLbAttributesMask != 0x0 )
	  {
	    m_uiPrerenderAttributesStackDepthMax++;
	  }
      }
      m_pCSLIDESurface->Pop();
    }
    PopShadingAttributes(eShade);
  }
  PopAttributes(GLbAttributesMask);
}

#ifdef JORDAN_IS_SO_COOL
VOID CSLIDEAreaLight::Prerender()
{
  if ( !(!m_dyniLOD.GetDynamic() && 
	 ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
	   m_dyniLOD.GetValue() == slidelodtSLF_BOUND )) )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;
      
      iter.Init(&m_listCSLIDEInstance);
      
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pCSLIDEInstance->Prerender();
	}
    }
      
  CSLIDEBoundedNode::Prerender();
}

VOID CSLIDEAreaLight::RenderGeometry(const Matrix &mGproj_obj, 
					CBound *pCBoundClip, 
					SLIDEShadingType /*eShading*/,
					SLIDELODType eLOD)
{

  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  // For Debugging
  if ( GetRenderCDAOperandsEnabled() )
    {
      gilPushAttrib(GL_ALL_ATTRIB_BITS);
      {
	PushAuxillaryGeometryAttributes();
	{
	  PushWireFrameAttributes();
	  {
	    iter.Init(&m_listCSLIDEInstance);
	    
	    for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
		pCSLIDEInstance != NULL;
		  pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	      {
		pCSLIDEInstance->Render(mGproj_obj, pCBoundClip, slideshtSLF_WIRE, eLOD);
	      }
	  }
	  PopWireFrameAttributes();
	}
	PopAuxillaryGeometryAttributes();
      }
      gilPopAttrib();
    }
}

VOID CSLIDEAreaLight::RenderNormals(SLIDEShadingType /*eShading*/)
{
  if ( m_pCDA != NULL )
    {
      PushAuxillaryGeometryAttributes();
      {
	gilColor3f(1.0, 0.0, 1.0);
	
	m_pCDA->RenderNormals();
      }
      PopAuxillaryGeometryAttributes();
    }
}

VOID CSLIDEAreaLight::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtAreaLight], m_pcGUIID);

  WriteSLIDEFlags(fpOutputFile, 0);

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->WriteSLIDE(fpOutputFile);    
    }

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndAreaLight]);
}
#endif // JORDAN_IS_SO_COOL

VOID CSLIDEAreaLight::WriteRIBAreaLight(FILE *fpOutputFile,
					const Matrix &mGvrc_light, 
					CBound *pCBoundClip, 
					SLIDEShadingType eShading,
					SLIDELODType eLOD)
{
  ModifyShadingFlag(eShading);
  ModifyLODFlag(eLOD);

  RtLightHandle rtlh;
  RtFloat rtfIntensity;
  RtColor rtColor;

  WriteRIBPushAttributes(fpOutputFile, eShading);
  {
    RiTransformBegin();
    {
      WriteRIBMultMatrix(fpOutputFile, mGvrc_light);

      RiArchiveRecord(RI_COMMENT, "\nAttribute \"light\" \"shadows\" \"on\"");
      RiArchiveRecord(RI_COMMENT, "\nAttribute \"light\" \"nsamples\" [%ld]", m_dyniSamples.GetValue());
      RiArchiveRecord(RI_COMMENT, "\nAttribute \"trace\" \"numarealightsamples\" [%ld]", m_dyniSamples.GetValue());
      rtfIntensity = MAX((FLOAT)1.0, MAX(m_vColor[0], MAX(m_vColor[1], m_vColor[2])));
      rtColor[0] = m_vColor[0] / rtfIntensity;
      rtColor[1] = m_vColor[1] / rtfIntensity;
      rtColor[2] = m_vColor[2] / rtfIntensity;
      rtlh = RiAreaLightSource("arealight", 
			       RI_INTENSITY, &rtfIntensity, 
			       RI_LIGHTCOLOR, rtColor,
			       RI_NULL);

      // Write the Area Light Geometry
      //rtColor[0] = m_vColor[0];
      //rtColor[1] = m_vColor[1];
      //rtColor[2] = m_vColor[2];

      RiColor(rtColor);
      RiSurface(RI_CONSTANT, RI_NULL);

      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;
      
      ASSERT( fpOutputFile != NULL );
      
      iter.Init(&m_listCSLIDEInstance);
      
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pCSLIDEInstance->WriteRIB(fpOutputFile,
				    mGvrc_light, pCBoundClip, eShading, eLOD);    
	}
    }
    RiTransformEnd();
  }
  WriteRIBPopAttributes(fpOutputFile, eShading);

  RiIlluminate(rtlh, TRUE);
}

VOID CSLIDEAreaLight::WritePOVAreaLight(FILE *fpOutputFile, 
					CListInt *plistTransformIDs,
					CListInt *plistColors) 
{
  ASSERT( fpOutputFile != NULL ); 

  UINT_32 iuSample;
  
  iuSample = (UINT_32)floor(sqrt(m_dyniSamples.GetValue()));
  fprintf(fpOutputFile, "area_light x y %d %d \n", iuSample, iuSample);
}

VOID CSLIDEAreaLight::WriteRIB(FILE * /*fpOutputFile*/,
				 const Matrix &/*mGproj_obj*/, 
				 CBound * /*pCBoundClip*/, 
				 SLIDEShadingType /*eShading*/,
				 SLIDELODType /*eLOD*/)
{
/*
  CBound::BoundIntersectionType eIntersect;

  ModifyShadingFlag(eShading);
  ModifyLODFlag(eLOD);

  if ( eLOD == slidelodtSLF_OFF )
    {
      // This branch has been turned off by the LOD flags
      return;
    }

  if ( eLOD > slidelodtSLF_BOUND )
    {
      WriteRIBPushAttributes(fpOutputFile, eShading);
      {
	WriteRIBGeometry(fpOutputFile, 
			 mGproj_obj, pCBoundClip, eShading, eLOD);
      }
      WriteRIBPopAttributes(fpOutputFile, eShading);
    }
*/
}

VOID CSLIDEAreaLight::WriteRIBGeometry(FILE * /*fpOutputFile*/,
				       const Matrix &/*mGproj_obj*/, 
				       CBound * /*pCBoundClip*/, 
				       SLIDEShadingType /*eShading*/,
				       SLIDELODType /*eLOD*/)
{
/*
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( fpOutputFile != NULL );

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->WriteRIB(fpOutputFile,
				mGproj_obj, pCBoundClip, eShading, eLOD);    
    }
*/
}

VOID CSLIDEAreaLight::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE AREALIGHT: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);
}

VOID CSLIDEAreaLight::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE AREALIGHT: %s", m_pcGUIID);
}

//////////////////////////////////////////////////////////////////////
// CSLIDEAreaLightPath Class
//

CSLIDEAreaLightPath::CSLIDEAreaLightPath()
{
  m_mGlight_vrc.SetIdentity();
  m_mGvrc_light.SetIdentity();
}

CSLIDEAreaLightPath::~CSLIDEAreaLightPath()
{
}

VOID CSLIDEAreaLightPath::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			   CListInt *plistPathIDs, DynInt *pdyniOnOff)
{
  CSLIDEPath::Init(pcFile, uiLine, uiColumn, plistPathIDs);

  m_dyniOnOff.SetValue(slideoftSLF_ON);
  if ( pdyniOnOff != NULL )
    {
      m_dyniOnOff.Set(pdyniOnOff);
    }

  m_mGlight_vrc.SetIdentity();
  m_mGvrc_light.SetIdentity();
}

VOID CSLIDEAreaLightPath::Uninit()
{
  CSLIDEPath::Uninit();
}

CSLIDENode *CSLIDEAreaLightPath::FindNode(CSLIDENode *pRoot, CListIterInt *piterIDs, 
				       CListInt *plistCSLIDEInstance)
{
  ASSERT( pRoot != NULL );
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  return pRoot->FindAreaLight(piterIDs, plistCSLIDEInstance);
}

VOID CSLIDEAreaLightPath::CheckDynamic()
{
  CSLIDEPath::CheckDynamic();

  SetDynamic(GetDynamic() || m_dyniOnOff.GetDynamic());
}

VOID CSLIDEAreaLightPath::UpdateState()
{
  CSLIDEPath::UpdateState();

  m_dyniOnOff.Update();
  if ( !( 0 <= m_dyniOnOff.GetValue() &&
	  m_dyniOnOff.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal OnOff flag value:  %ld", 
		  m_dyniOnOff.GetValue());
      m_dyniOnOff.SetValue(slideoftSLF_ON);
    }
}
