#ifndef __SLIDEAREALIGHT_H__
#define __SLIDEAREALIGHT_H__

#include "slidegroup.h"
#include "slidepath.h"

//////////////////////////////////////////////////////////////////////
// CSLIDEArealight Class
//

class CSLIDEAreaLight : public CSLIDEGroup
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEAreaLight();
  virtual ~CSLIDEAreaLight();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  inline VOID SetColor(DynVector *pdynvColor);
  inline VOID SetSamples(DynInt *pdyniSamples);

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual CSLIDECamera *FindCamera(CListIterInt *piterIDs, 
				   CListInt *plistCSLIDEInstance);
  virtual CSLIDELight *FindLight(CListIterInt *piterIDs, 
				 CListInt *plistCSLIDEInstance);
  virtual CSLIDEAreaLight *FindAreaLight(CListIterInt *piterIDs, 
					 CListInt *plistCSLIDEInstance);

/*
  virtual VOID Preprocess();
  virtual VOID PreprocessGraph();

  virtual VOID Update();
  virtual VOID UpdateGraph();
*/

protected:
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();

//////////////////////////////////////////////////
// Rendering
//
public:
  VOID Place();

  virtual VOID PrerenderGeometry(SLIDEShadingType eShading);
/*
  virtual VOID Prerender();

  virtual VOID RenderGeometry(const Matrix &mGproj_obj,
			      CBound *pCBoundClip, 
			      SLIDEShadingType eShading,
			      SLIDELODType eLOD);
  virtual VOID RenderNormals(SLIDEShadingType eShading);
*/
  virtual GLbitfield CalculateAttributesMask(SLIDEShadingType eShading);

//////////////////////////////////////////////////
// Output Functions
//
public:
//  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  // The Light
  virtual VOID WriteRIBAreaLight(FILE *fpOutputFile,
				 const Matrix &mGproj_obj,
				 CBound *pCBoundClip, 
				 SLIDEShadingType eShading,
				 SLIDELODType eLOD);
  virtual VOID WritePOVAreaLight(FILE *fpOutputFile, 
				 CListInt *plistTransformIDs,
				 CListInt *plistColors);

  // The Geometry
  virtual VOID WriteRIB(FILE *fpOutputFile,
			const Matrix &mGproj_obj,
			CBound *pCBoundClip, 
			SLIDEShadingType eShading,
			SLIDELODType eLOD);
  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);
 
//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline CHAR *GetTypeString();
  virtual inline SLIDEReservedWordType GetType();

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  DynVector m_vColor;
  DynInt m_dyniSamples;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDEAreaLight::SetColor(DynVector *pdynvColor)
{
  m_vColor.SetAll(pdynvColor);
}

inline VOID CSLIDEAreaLight::SetSamples(DynInt *pdyniSamples)
{
  m_dyniSamples.Set(pdyniSamples);
}

inline CHAR *CSLIDEAreaLight::GetTypeString()
{
  return gcppcSLIDEReservedWords[sliderwtAreaLight];
}

inline SLIDEReservedWordType CSLIDEAreaLight::GetType()
{
  return sliderwtAreaLight;
}

//////////////////////////////////////////////////////////////////////
// CSLIDEAreaLightPath Class
//

class CSLIDEAreaLightPath : public CSLIDEPath
{
//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEAreaLightPath();
  virtual ~CSLIDEAreaLightPath();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CListInt *plistPathIDs, DynInt *pdyniOnOff);
  virtual VOID Uninit();

  virtual CSLIDENode *FindNode(CSLIDENode *pRoot, CListIterInt *piterIDs, 
			       CListInt *plistCSLIDEInstance);

//////////////////////////////////////////////////
// Geometric Computations
//
public:

protected:
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  inline CSLIDEAreaLight *GetAreaLight();

  inline SLIDEOnOffType GetOnOff();

  inline const Matrix &GetVRCToLight() const;
  inline Matrix &SetVRCToLight();
  inline const Matrix &GetLightToVRC() const;
  inline Matrix &SetLightToVRC();

//////////////////////////////////////////////////
// Member Variables
//
protected:
  DynInt m_dyniOnOff;

  Matrix m_mGlight_vrc; // VRC to Light coordinates
  Matrix m_mGvrc_light; // Light to VRC coordinates
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline CSLIDEAreaLight *CSLIDEAreaLightPath::GetAreaLight()
{
  return (CSLIDEAreaLight*)m_pNode;
}

inline SLIDEOnOffType CSLIDEAreaLightPath::GetOnOff()
{
  return (SLIDEOnOffType)m_dyniOnOff.GetValue();
}

inline const Matrix &CSLIDEAreaLightPath::GetVRCToLight() const
{
  return m_mGlight_vrc;
}

inline Matrix &CSLIDEAreaLightPath::SetVRCToLight()
{
  return m_mGlight_vrc;
}

inline const Matrix &CSLIDEAreaLightPath::GetLightToVRC() const
{
  return m_mGvrc_light;
}

inline Matrix &CSLIDEAreaLightPath::SetLightToVRC()
{
  return m_mGvrc_light;
}

#endif // __SLIDEAREALIGHT_H__
