#include "slideatomnode.h"

#include "slidesurface.h"

//////////////////////////////////////////////////////////////////////
// CSLIDEAtomNode Class
//

CSLIDEAtomNode::CSLIDEAtomNode()
{
  m_dyniSolid.SetValue(slidestSLF_SOLID);
}

CSLIDEAtomNode::~CSLIDEAtomNode()
{
}

VOID CSLIDEAtomNode::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			CHAR *pcID)
{
  CSLIDEBoundedNode::Init(pcFile, uiLine, uiColumn, pcID);

  m_dyniSolid.SetValue(slidestSLF_SOLID);
}

VOID CSLIDEAtomNode::Uninit()
{
  CSLIDEBoundedNode::Uninit();
}

BOOL CSLIDEAtomNode::Initialized()
{
  if ( !CSLIDEBoundedNode::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

BOOL CSLIDEAtomNode::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;

  if ( !CSLIDEBoundedNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  // Check Solid Type

  ASSERT( 0 <= m_dyniSolid.GetValue() &&
	  m_dyniSolid.GetValue() < slidestSize );

  return bCorrect;
}

BOOL CSLIDEAtomNode::Valid()
{
  if ( !CSLIDEBoundedNode::Valid() )
    {
      return FALSE;
    }

  return ( ( 0 <= m_dyniSolid.GetValue() &&
	     m_dyniSolid.GetValue() < slidestSize ) );
}

CSLIDECamera *CSLIDEAtomNode::FindCamera(CListIterInt *piterIDs, 
				       CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtCamera],
	      gcppcSLIDEReservedWords[sliderwtObject], GetID());

  return NULL;
}

CSLIDELight *CSLIDEAtomNode::FindLight(CListIterInt *piterIDs, 
				     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtLight],
	      gcppcSLIDEReservedWords[sliderwtObject], GetID());

  return NULL;
}

CSLIDEAreaLight *CSLIDEAtomNode::FindAreaLight(CListIterInt *piterIDs, 
					       CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtAreaLight],
	      gcppcSLIDEReservedWords[sliderwtObject], GetID());

  return NULL;
}

VOID CSLIDEAtomNode::GetSide(CListInt *plistPoint, CListInt *plistNormal, UINT_32 uiSide, 
			     const Matrix &mTransform, const Matrix &mInverse, 
			     BOOL bReverse)
{
  ASSERT( plistPoint != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtAreaLight],
	      gcppcSLIDEReservedWords[sliderwtObject], GetID());
}

VOID CSLIDEAtomNode::Preprocess()
{
  CSLIDEBoundedNode::Preprocess();

  CheckDynamic();
  UpdateState();
}

VOID CSLIDEAtomNode::CheckDynamic()
{
  CSLIDEBoundedNode::CheckDynamic();

  // Are there variables which I must call TCL to update?
  SetDynamic(GetDynamic() ||
	     m_dyniSolid.GetDynamic());
}

VOID CSLIDEAtomNode::PreprocessGraph()
{
  if ( !GetVisited() )
    {
      BOOL bGeometryEmpty;
      BOOL bGeometryDynamic;
      BOOL bNormalsEmpty;
      BOOL bNormalsDynamic;
      BOOL bBoundsEmpty;
      BOOL bBoundsDynamic;

      BOOL bBoundEmpty;
      BOOL bBoundDynamic;

      PreprocessGraphLocal();

      bGeometryEmpty = GetGeometryEmpty();
      bGeometryDynamic = GetGeometryDynamic();
      bNormalsEmpty = GetNormalsEmpty();
      bNormalsDynamic = GetNormalsDynamic();
      bBoundsEmpty = GetBoundsEmpty();
      bBoundsDynamic = GetBoundsDynamic();
      bBoundEmpty = GetBoundEmpty();
      bBoundDynamic = GetBoundDynamic();

      //////////////////////////////////////////////////
      // Look at Local flags

      // Update whether we are Empty based on our LOD flag
      if ( !m_dyniLOD.GetDynamic() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      // This node is always off
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	      bBoundsEmpty = TRUE;
	      bBoundEmpty = TRUE;
	    }
	  else if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	    {
	      // This node is always just a bounding box
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	    }
	}

      // Check what is really dynamic

      bGeometryDynamic = ( bGeometryDynamic ||
			   m_pCSLIDESurface->GetDynamic() ||
			   m_dyniShading.GetDynamic() ||
			   m_dyniLOD.GetDynamic() ||
			   m_dyniSolid.GetDynamic() );
      bNormalsDynamic = ( bNormalsDynamic ||
			  m_dyniShading.GetDynamic() ||
			  m_dyniLOD.GetDynamic() );
      bBoundsDynamic = ( bBoundsDynamic || 
			 m_dyniLOD.GetDynamic() );
      bBoundDynamic = ( bBoundDynamic || 
			m_dyniLOD.GetDynamic() );

      // If anything is now Empty then it will always be empty so
      // it is not dynamic

      bGeometryDynamic = ( bGeometryDynamic && !bGeometryEmpty );
      bNormalsDynamic = ( bNormalsDynamic && !bNormalsEmpty );
      bBoundsDynamic = ( bBoundsDynamic && !bBoundsEmpty );
      bBoundDynamic = ( bBoundDynamic && !bBoundEmpty );
      
      //////////////////////////////////////////////////
      // For the OpenGL Display Lists to be really useful 
      // it is all or nothing so we enforce that here

      bGeometryDynamic = ( bGeometryDynamic ||
			   bNormalsDynamic ||
			   bBoundsDynamic );
      bNormalsDynamic = bGeometryDynamic;
      bBoundsDynamic = bGeometryDynamic;

      //////////////////////////////////////////////////
      // Now save these correct flag values

      SetGeometryEmpty(bGeometryEmpty);
      SetGeometryDynamic(bGeometryDynamic);
      SetNormalsEmpty(bNormalsEmpty);
      SetNormalsDynamic(bNormalsDynamic);
      SetBoundsEmpty(bBoundsEmpty);
      SetBoundsDynamic(bBoundsDynamic);
      SetBoundEmpty(bBoundEmpty);
      SetBoundDynamic(bBoundDynamic);
      
      //////////////////////////////////////////////////
      // Calculate the Bounding Box (if Empty this will do nothing)
      if ( GetBoundEmpty() )
	{
	  SetBoundValid(TRUE);
	}
      else
	{
	  CalculateBound();
	}
      SetBoundChanged(FALSE);

      //////////////////////////////////////////////////
      // Done Preprocessing this Node in the Graph
      SetVisited(TRUE);
    }
}

VOID CSLIDEAtomNode::PreprocessGraphLocal()
{
  SetGeometryEmpty(TRUE);
  SetGeometryDynamic(FALSE);
  SetNormalsEmpty(TRUE);
  SetNormalsDynamic(FALSE);
  SetBoundsEmpty(TRUE);
  SetBoundsDynamic(FALSE);
  SetBoundEmpty(TRUE);
  SetBoundDynamic(FALSE);
}

VOID CSLIDEAtomNode::Update()
{
  SetBoundChanged(FALSE);
  if ( GetDynamic() )
    {
      UpdateState();
    }
  
  if ( GetBoundsDynamic() )
    {
      CSLIDEBoundedNode::InvalidateQEMesh();
    }

  if ( GetBoundDynamic() )
    {
      SetBoundValid(FALSE);
    }

  SetVisited(FALSE);
}

VOID CSLIDEAtomNode::UpdateState()
{
  CSLIDEBoundedNode::UpdateState();

  m_dyniSolid.Update();
  if ( !( 0 <= m_dyniSolid.GetValue() &&
	  m_dyniSolid.GetValue() < slidestSize ) )
    {
      HandleError("Illegal solid flag value:  %ld", 
		  m_dyniSolid.GetValue());
      m_dyniSolid.SetValue(slidestSLF_HOLLOW);
    }
}

VOID CSLIDEAtomNode::UpdateGraph()
{
  if ( !GetVisited() )
    {
      if ( !GetBoundValid() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      if ( !m_bound.GetEmpty() )
		{
		  SetBoundChanged(TRUE);
		  m_bound.Null();
		}
	      return;
	    }
	  
	  CalculateBound();
	  SetBoundValid(TRUE);
	}

      SetVisited(TRUE);
    }
}

VOID CSLIDEAtomNode::CalculateBound()
{
}

VOID CSLIDEAtomNode::PrerenderGeometry(SLIDEShadingType eShading)
{
  SLIDEShadingType eShade;
  
  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  // This flag tells us whether to use Point or Face Normals
  ModifyShadingFlag(eShading);

  // This one set the shading attributes which in many cases will 
  // be SLF_INHERIT and so it will not set anything and leave up
  // to the caller above
  eShade = (SLIDEShadingType)m_dyniShading.GetValue();

  GLbitfield GLbAttributesMask;
  
  GLbAttributesMask = CalculateAttributesMask(eShade);
  
  PushAttributes(GLbAttributesMask);
  {
    PushShadingAttributes(eShade);
    {
      m_pCSLIDESurface->Push();
      {
	PushSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
	{
	  m_uiPrerenderAttributesStackDepthMax = 0;

	  if (eShading == slideshtSLF_GOURAUD ||
	      eShading == slideshtSLF_PHONG)
	    {
	      RenderGeometrySmooth();
	    }
	  else
	    {
	      RenderGeometryFlat();
	    }

	  if ( GLbAttributesMask != 0x0 )
	    {
	      m_uiPrerenderAttributesStackDepthMax++;
	    }
	}
	PopSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
      }
      m_pCSLIDESurface->Pop();
    }
    PopShadingAttributes(eShade);
  }
  PopAttributes(GLbAttributesMask);
}

VOID CSLIDEAtomNode::PrerenderNormals(SLIDEShadingType eShading)
{
  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  ModifyShadingFlag(eShading);

  if (eShading == slideshtSLF_GOURAUD ||
      eShading == slideshtSLF_PHONG)
    {
      RenderNormalsSmooth();
    }
  else
    {
      RenderNormalsFlat();
    }
}

GLbitfield CSLIDEAtomNode::CalculateAttributesMask(SLIDEShadingType eShading)
{
  return ( CSLIDEGeometryNode::CalculateAttributesMask(eShading) |
	   CalculateSolidAttributesMask((SLIDESolidType)m_dyniSolid.GetValue()) );
}

VOID CSLIDEAtomNode::RenderGeometry(const Matrix & /*mGproj_obj*/, 
				    CBound * /*pCBoundClip*/, 
				    SLIDEShadingType eShading,
				    SLIDELODType /* eLOD */,
				    CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/ )
{
  PushSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
  {
    if (eShading == slideshtSLF_GOURAUD ||
	eShading == slideshtSLF_PHONG)
      {
	RenderGeometrySmooth(pCSLIDEWarp);
      }
    else
      {
	RenderGeometryFlat(pCSLIDEWarp);
      }
  }
  PopSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
}

VOID CSLIDEAtomNode::RenderGeometrySelect(const Matrix & /*mGproj_obj*/, 
					  CBound * /*pCBoundClip*/, 
					  SLIDEShadingType eShading,
					  SLIDELODType /* eLOD */ )
{
  if ( !GetGeometryDynamic() ) // Use the Display list if possible
    {
      CheckAttributesStackDepth(GetPrerenderAttributesStackDepthMax());
      RecallGeometry(eShading);
    }
  else
    {
      RenderGeometryFlat();
    }
}

VOID CSLIDEAtomNode::RenderNormals(SLIDEShadingType eShading)
{
  PushAuxillaryGeometryAttributes();
  {
    gilColor3f(1.0, 0.0, 1.0);

    if (eShading == slideshtSLF_GOURAUD ||
	eShading == slideshtSLF_PHONG)
      {
	RenderNormalsSmooth();
      }
    else
      {
	RenderNormalsFlat();
      }
  }
  PopAuxillaryGeometryAttributes();
}

VOID CSLIDEAtomNode::RenderGeometrySmooth(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{
}

VOID CSLIDEAtomNode::RenderGeometryFlat(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{
}

VOID CSLIDEAtomNode::RenderNormalsSmooth()
{
}

VOID CSLIDEAtomNode::RenderNormalsFlat()
{
}

VOID CSLIDEAtomNode::WriteSLIDEFlags(FILE *fpOutputFile, UINT_32 uiIndent)
{
  CSLIDEBoundedNode::WriteSLIDEFlags(fpOutputFile, uiIndent);

  if ( m_dyniSolid.GetValue() != slidestSLF_SOLID )
    {
      WriteIndent(fpOutputFile, uiIndent);
      fprintf(fpOutputFile, "  %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtSolid], 
	      gcppcSLIDESolidReservedWords[m_dyniSolid.GetValue()]);
    }
}

VOID CSLIDEAtomNode::WriteRIBGeometry(FILE * /*fpOutputFile */ ,
				      const Matrix & /*mGproj_obj*/, 
				      CBound * /*pCBoundClip*/, 
				      SLIDEShadingType /* eShading */,
				      SLIDELODType /* eLOD */)
{
}

VOID CSLIDEAtomNode::WriteRIBPushAttributes(FILE *fpOutputFile, 
					    SLIDEShadingType eShading)
{
  CSLIDEGeometryNode::WriteRIBPushAttributes(fpOutputFile, eShading);

  WriteRIBPushSolidAttributes(fpOutputFile,
			      (SLIDESolidType)m_dyniSolid.GetValue());
}

VOID CSLIDEAtomNode::WriteRIBPopAttributes(FILE *fpOutputFile, 
					   SLIDEShadingType eShading)
{
  WriteRIBPopSolidAttributes(fpOutputFile, 
			     (SLIDESolidType)m_dyniSolid.GetValue());

  CSLIDEGeometryNode::WriteRIBPopAttributes(fpOutputFile, eShading);
}

VOID CSLIDEAtomNode::WritePOVGeometry(FILE * /*fpOutputFile */, CListInt * /*plistTransformIDs*/, CListInt * /*plistColors*/)
{
}

VOID CSLIDEAtomNode::WriteWRLZGeometry(FILE * /*fpOutputFile */, 
				       const Matrix &/*mTransform*/, 
				       CSLIDESurface */*pCSLIDESurface*/)
{
}

VOID CSLIDEAtomNode::WriteSTLGeometry(FILE * /*fpOutputFile */, const Matrix &/*mTransform*/)
{
}

QEMesh *CSLIDEAtomNode::WriteSIF(FILE * /*fpOutputFile*/, UINT_32 /*uiIndentLevel*/, 
				 const Matrix &/*mGroot_obj*/, const Matrix &/*mGobj_root*/,
				 QEMesh * /*pmeBoundaryEdges*/)
{
  return GetQEMesh();
}

UINT_32 CSLIDEAtomNode::SetParameter(CSLIDEWorld *pCSLIDEWorld, 
				  SLIDEReservedWordType uiParameter, CHAR **argv)
{
  UINT_32 uiValue;

  switch(uiParameter) 
    {
    case sliderwtSolid:
      for( uiValue=0; uiValue < slidestSize; uiValue++ )
	{
	  if( !strcmp(argv[0], gcppcSLIDEGUISolidReservedWords[uiValue]) )
	    {
	      break;
	    }
	}
      m_dyniSolid.SetValue((SLIDESolidType)uiValue);
      return 1;
      break;
    default:
      return CSLIDEBoundedNode::SetParameter(pCSLIDEWorld, uiParameter, argv);
      break;
    }
}
