#ifndef __SLIDEATOMNODE_H__
#define __SLIDEATOMNODE_H__

#include "slideboundednode.h"

//////////////////////////////////////////////////////////////////////////////
// CSLIDEAtomNode Class
//

class CSLIDEAtomNode : public CSLIDEBoundedNode
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEAtomNode();
  virtual ~CSLIDEAtomNode();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  inline VOID SetSolid(DynInt *pdyniSolid);

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual CSLIDECamera *FindCamera(CListIterInt *piterIDs, 
				   CListInt *plistCSLIDEInstance);
  virtual CSLIDELight *FindLight(CListIterInt *piterIDs, 
				 CListInt *plistCSLIDEInstance);
  virtual CSLIDEAreaLight *FindAreaLight(CListIterInt *piterIDs, 
					 CListInt *plistCSLIDEInstance);

  virtual VOID Preprocess();
  virtual VOID PreprocessGraph();

  virtual VOID Update();
  virtual VOID UpdateGraph();

  virtual VOID GetSide(CListInt *plistPoint, CListInt *plistNormal, UINT_32 uiSide, 
		       const Matrix &mTransform, const Matrix &mInverse, BOOL bReverse);

  virtual UINT_32 SetParameter(CSLIDEWorld *pCSLIDEWorld, 
			       SLIDEReservedWordType uiParameter, CHAR **argv);

protected:
  virtual VOID CheckDynamic();
  virtual VOID PreprocessGraphLocal();
  virtual VOID UpdateState();
  virtual VOID CalculateBound();

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID PrerenderGeometry(SLIDEShadingType eShading);
  virtual VOID PrerenderNormals(SLIDEShadingType eShading);

  virtual GLbitfield CalculateAttributesMask(SLIDEShadingType eShading);
  virtual VOID RenderGeometry(const Matrix &mGproj_obj,
			      CBound *pCBoundClip, 
			      SLIDEShadingType eShading,
			      SLIDELODType eLOD,
			      CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderGeometrySelect(const Matrix &mGproj_obj,
				    CBound *pCBoundClip, 
				    SLIDEShadingType eShading,
				    SLIDELODType eLOD);
  virtual VOID RenderNormals(SLIDEShadingType eShading);

protected:
  virtual VOID RenderGeometrySmooth(CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderGeometryFlat(CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderNormalsSmooth();
  virtual VOID RenderNormalsFlat();

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDEFlags(FILE *fpOutputFile, UINT_32 uiIndent);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);

  virtual VOID WriteRIBPushAttributes(FILE *fpOutputFile,
				      SLIDEShadingType eShading);
  virtual VOID WriteRIBPopAttributes(FILE *fpOutputFile,
				     SLIDEShadingType eShading);

  virtual VOID WritePOVGeometry(FILE *fpOutputFile, 
				CListInt *plistTransformIDs,
				CListInt *plistColors);
  virtual VOID WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				 CSLIDESurface *pCSLIDESurface);

  virtual VOID WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform);
  virtual QEMesh *WriteSIF(FILE *fpOutputFile, UINT_32 uiIndentLevel, 
			   const Matrix &mGroot_obj, const Matrix &mGobj_root,
			   QEMesh *pmeBoundaryEdges);

//////////////////////////////////////////////////
// Accessor Functions
//
public:

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  DynInt m_dyniSolid;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDEAtomNode::SetSolid(DynInt *pdyniSolid)
{
  m_dyniSolid.Set(pdyniSolid);
}

#endif // __SLIDEATOMNODE_H__
