#include "slidebezierpatch.h"
#include "slidepoint.h"
#include "slideworld.h"
#include "slidesurface.h"

//////////////////////////////////////////////////////////////////////
// CSLIDEBezierPatch Class
//

CSLIDEBezierPatch::CSLIDEBezierPatch()
{
}

CSLIDEBezierPatch::~CSLIDEBezierPatch()
{
}

VOID CSLIDEBezierPatch::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID)
{
  CSLIDEBSplinePatch::Init(pcFile, uiLine, uiColumn, pcID);
}

VOID CSLIDEBezierPatch::InitKnots()
{
  UINT_32 i;

  for(i = 0; i < m_uiUKnots; i++)
    {
      m_pfUKnot[i] = !(i < m_uiUKnots/2); 
    }

  for(i = 0; i < m_uiVKnots; i++)
    {
      m_pfVKnot[i] = !(i < m_uiVKnots/2);
    }
}

VOID CSLIDEBezierPatch::Uninit()
{
  CSLIDEBSplinePatch::Uninit();
}

BOOL CSLIDEBezierPatch::Initialized()
{
  if ( !CSLIDEBSplinePatch::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

BOOL CSLIDEBezierPatch::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  m_uiUKnots = 2*m_uiUOrder;
  m_uiVKnots = 2*m_uiVOrder;
  ReallocateUKnots();
  ReallocateVKnots();
  InitKnots();

  if ( !CSLIDEBSplinePatch::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  return TRUE;
}

BOOL CSLIDEBezierPatch::Valid()
{
  if ( !CSLIDEBSplinePatch::Valid() )
    {
      return FALSE;
    }

  return TRUE;
}

VOID CSLIDEBezierPatch::CheckDynamic()
{
  CSLIDEBSplinePatch::CheckDynamic();
}

VOID CSLIDEBezierPatch::UpdateState()
{
  CSLIDEBSplinePatch::UpdateState();
}

#ifdef __USE_CDA__
VOID CSLIDEBezierPatch::CalculateCDAVertices(CListInt *plistCDAVertex, 
					      const Matrix &mGroot_obj)
{

}
#endif // __USE_CDA__

VOID CSLIDEBezierPatch::WriteSLIDE(FILE *fpOutputFile)
{
  UINT_32 i, j;
  CListIterInt iter;
  CHAR *pcID;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtBezierPatch], m_pcGUIID);

  WriteSLIDEFlags(fpOutputFile, 0);

  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtUOrder],
	  m_uiUOrder);
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtVOrder],
	  m_uiVOrder);

  if( m_plistControlPointIDs != NULL )
    {
      fprintf(fpOutputFile, "  %s (",
	      gcppcSLIDEReservedWords[sliderwtControlPointList]);

      iter.Init(m_plistControlPointIDs);
      
      for ( pcID = (CHAR*)(iter.PeekFirst()), i = 0;
	    pcID != NULL;
	    pcID = (CHAR*)(iter.PeekNext()) )
	{
	  if( i == m_uiUControlPoints )
	    {
	      fprintf(fpOutputFile, "\n\t\t    " );
	      i = 0;
	    }
	  fprintf(fpOutputFile, " %s", pcID );
	  i++;
	}
    }
  else
    {
      const FLOAT* pt;

      fprintf(fpOutputFile, "# Default Control Points:\n");
      fprintf(fpOutputFile, "# %s (\n#\t",
	      gcppcSLIDEReservedWords[sliderwtControlPointList]);
      
      for( j = 0; j < m_uiVControlPoints; j++ )
	{
	  for( i = 0; i < m_uiUControlPoints; i++ )
	    {
	      pt = m_ppControlPoint[i][m_uiVControlPoints-1-j].Get();
	      fprintf(fpOutputFile, "(%.3f %.3f %.3f)", pt[X], pt[Y], pt[Z] );

	      if( (j != m_uiVControlPoints-1) && ( i == m_uiUControlPoints-1) )
		{
		  fprintf(fpOutputFile, "\n#\t" );
		}
	    }
	}
    }
  fprintf(fpOutputFile, " )\n");

  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtUSlices],
	  m_dyniUSlices.GetValue());
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtVSlices],
	  m_dyniVSlices.GetValue());
  fprintf(fpOutputFile, "  %s (%.3f %.3f) (%.3f %.3f) (%.3f %.3f) (%.3f %.3f)\n",
	  gcppcSLIDEReservedWords[sliderwtTexture],
	  m_pvTexture[0][X], m_pvTexture[0][Y],
	  m_pvTexture[1][X], m_pvTexture[1][Y],
	  m_pvTexture[2][X], m_pvTexture[2][Y],
	  m_pvTexture[3][X], m_pvTexture[3][Y]);
  fprintf(fpOutputFile, "  %s %.3f %ld\n",
	  gcppcSLIDEReservedWords[sliderwtOffset],
	  m_dynfOffset.GetValue(), m_dyniOffsetConnected.GetValue());
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtDrawControls],
	  m_dyniDrawControls.GetValue());

  
  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndBezierPatch]);
}

VOID CSLIDEBezierPatch::WriteRIBGeometry(FILE *fpOutputFile,
				    const Matrix &/*mGproj_obj*/, 
				    CBound * /*pCBoundClip*/, 
				    SLIDEShadingType eShading,
				    SLIDELODType eLOD)
{
  UINT_32 i, j;

  if( (m_uiUOrder == 4) && (m_uiVOrder == 4) )
    {
      const FLOAT *pt;
      char *pointIds = new char[800];
      char *thisPoint = new char[50];

      RiArchiveRecord(RI_COMMENT, "\nBasis \"bezier\" 3 \"bezier\" 3\n");

      for( j = 0; j < m_uiVControlPoints; j++ )
	{
	  for( i = 0; i < m_uiUControlPoints; i++ )
      	    {
	      pt = m_ppControlPoint[i][j].Get();
	      sprintf( thisPoint, " %.6f %.6f %.6f", pt[X], pt[Y], pt[Z] );
	      strcat( pointIds, thisPoint );
	    }
	}

      RiArchiveRecord(RI_COMMENT, "\nPatch \"bicubic\" \"P\" [%s ] \"st\" [%.6f %.6f %.6f %.6f %.6f %.6f %.6f %.6f]\n", 
      		      pointIds,
		      m_pvTexture[3][X], m_pvTexture[3][Y], 
		      m_pvTexture[2][X], m_pvTexture[2][Y],
		      m_pvTexture[0][X], m_pvTexture[0][Y], 
		      m_pvTexture[1][X], m_pvTexture[1][Y]);

      delete[] pointIds;
      delete[] thisPoint;
    }
  else 
    {
      const FLOAT *p1, *p2, *p3, *p4, *v1, *v2, *v3, *v4;
      Vector pvVDelta[2];
      Vector pvUDelta[2];
      Vector ppvU[2][2];
      Vector pvV[2];

      pvUDelta[0] = (m_pvTexture[2] - m_pvTexture[3]) / (FLOAT)m_uiUSlices;
      pvUDelta[1] = (m_pvTexture[1] - m_pvTexture[0]) / (FLOAT)m_uiUSlices;
      
      ppvU[0][0] = m_pvTexture[3];
      ppvU[0][1] = m_pvTexture[0];
      ppvU[1][0] = ppvU[0][0] + pvUDelta[0];
      ppvU[1][1] = ppvU[0][1] + pvUDelta[1];

      for( i = 0; i < m_uiUSlices; i++ )
	{
	  pvVDelta[0] = (ppvU[0][1] - ppvU[0][0]) / (FLOAT)m_uiVSlices;
	  pvVDelta[1] = (ppvU[1][1] - ppvU[1][0]) / (FLOAT)m_uiVSlices;
	  pvV[0] = ppvU[0][0];
	  pvV[1] = ppvU[1][0];

	  for( j = 0; j < m_uiVSlices; j++)
	    {
	      p1 = m_ppSurfacePoint[i][j].Get();
	      p2 = m_ppSurfacePoint[i+1][j].Get();
	      p3 = m_ppSurfacePoint[i][j+1].Get();
	      p4 = m_ppSurfacePoint[i+1][j+1].Get();
	      
	      if( (eShading == slideshtSLF_GOURAUD) || (eShading == slideshtSLF_PHONG) )
		{
		  v1 = m_ppvSurfaceNormal[i][j].Get();
		  v2 = m_ppvSurfaceNormal[i+1][j].Get();
		  v3 = m_ppvSurfaceNormal[i][j+1].Get();
		  v4 = m_ppvSurfaceNormal[i+1][j+1].Get();

		  RiArchiveRecord(RI_COMMENT, "\nPolygon \"P\" [%.6f %.6f %.6f %.6f %.6f %.6f %.6f %.6f %.6f] \"N\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"st\" [%.6f %.6f %.6f %.6f %.6f %.6f]\n",
				  p1[X], p1[Y], p1[Z], p2[X], p2[Y], p2[Z], p3[X], p3[Y], p3[Z],
				  v1[X], v1[Y], v1[Z], v2[X], v2[Y], v2[Z], v3[X], v3[Y], v3[Z],
				  pvV[0][0], pvV[0][1], pvV[1][0], pvV[1][1], 
				  (pvV[0] + pvVDelta[0])[0], (pvV[0] + pvVDelta[0])[1]);

		  pvV[0] += pvVDelta[0];

		  RiArchiveRecord(RI_COMMENT, "\nPolygon \"P\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"N\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"st\" [%.6f %.6f %.6f %.6f %.6f %.6f]\n",
				  p3[X], p3[Y], p3[Z], p2[X], p2[Y], p2[Z], p4[X], p4[Y], p4[Z],
				  v3[X], v3[Y], v3[Z], v2[X], v2[Y], v2[Z], v4[X], v4[Y], v4[Z],
				  pvV[0][0], pvV[0][1], pvV[1][0], pvV[1][1],
				  (pvV[1] + pvVDelta[1])[0], (pvV[1] + pvVDelta[1])[1]);
		  pvV[1] += pvVDelta[1];
		}
	      else
		{
		  RiArchiveRecord(RI_COMMENT, "\nPolygon \"P\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"st\" [%.6f %.6f %.6f %.6f %.6f %.6f]\n",
				  p1[X], p1[Y], p1[Z], p2[X], p2[Y], p2[Z], p3[X], p3[Y], p3[Z],
				  pvV[0][0], pvV[0][1], pvV[1][0], pvV[1][1], 
				  (pvV[0] + pvVDelta[0])[0], (pvV[0] + pvVDelta[0])[1]);
		  pvV[0] += pvVDelta[0];

		  RiArchiveRecord(RI_COMMENT, "\nPolygon \"P\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"st\" [%.6f %.6f %.6f %.6f %.6f %.6f]\n",
				  p3[X], p3[Y], p3[Z], p2[X], p2[Y], p2[Z], p4[X], p4[Y], p4[Z],
				   pvV[0][0], pvV[0][1], pvV[1][0], pvV[1][1],
				  (pvV[1] + pvVDelta[1])[0], (pvV[1] + pvVDelta[1])[1]);
		  pvV[1] += pvVDelta[1];
		}
	    }
	  ppvU[0][0] += pvUDelta[0];
	  ppvU[0][1] += pvUDelta[1];
	  ppvU[1][0] += pvUDelta[0];
	  ppvU[1][1] += pvUDelta[1];
	}
    }
}

VOID CSLIDEBezierPatch::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE BezierPatch: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);
}

VOID CSLIDEBezierPatch::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE BezierPatch: %s", m_pcGUIID);
}

