#include "slideboundednode.h"
#include "slidesurface.h"
#include "slidecylinder.h"
#include "slideinstance.h"

#include "qemesh.h"

//////////////////////////////////////////////////////////////////////
// CSLIDEBoundedNode Class
//

GLuint CSLIDEBoundedNode::m_displayListCount = 0;

CSLIDEBoundedNode::CSLIDEBoundedNode()
{
  m_pQEMesh = NULL;
}

CSLIDEBoundedNode::~CSLIDEBoundedNode()
{
}

VOID CSLIDEBoundedNode::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID)
{
  CSLIDEGeometryNode::Init(pcFile, uiLine, uiColumn, pcID);

#ifdef __USE_CDA__
  m_pCDA = NULL;
  m_bCDAVisited = FALSE;
#endif // __USE_CDA__

  m_bPrerenderVisited = FALSE;

  m_GLuiFlatGeometry = 0;
  m_GLuiSmoothGeometry = 0;
  m_GLuiFlatNormals = 0;
  m_GLuiSmoothNormals = 0;
  m_GLuiBounds = 0;
}

VOID CSLIDEBoundedNode::Uninit()
{
  CSLIDEGeometryNode::Uninit();

#ifdef __USE_CDA__
  if ( m_pCDA != NULL )
    {
      DeallocateCDA(m_pCDA);
      m_pCDA = NULL;
    }
#endif // __USE_CDA__


  if ( m_pQEMesh != NULL )
    {
      DeallocateQEMesh(m_pQEMesh);
      m_pQEMesh = NULL;
    }

  // FIX THIS!!!
  // This will only destroy the display lists from one of the drawing contexts
  // we need to remove them from all of the drawing contexts which have them
  if ( gilIsList(m_GLuiFlatGeometry) )
    {
      gilDeleteLists(m_GLuiFlatGeometry, 1);
    }
  if ( m_GLuiFlatGeometry != m_GLuiSmoothGeometry && 
       gilIsList(m_GLuiSmoothGeometry) )      
    {
      gilDeleteLists(m_GLuiSmoothGeometry, 1);
    }
  m_GLuiFlatGeometry = 0;
  m_GLuiSmoothGeometry = 0;

  if ( gilIsList(m_GLuiFlatNormals) )
    {
      gilDeleteLists(m_GLuiFlatNormals, 1);
    }
  if ( m_GLuiFlatNormals != m_GLuiSmoothNormals && 
       gilIsList(m_GLuiSmoothNormals) )      
    {
      gilDeleteLists(m_GLuiSmoothNormals, 1);
    }
  m_GLuiFlatNormals = 0;
  m_GLuiSmoothNormals = 0;

  if ( m_GLuiBounds != 0 )
    {
      gilDeleteLists(m_GLuiBounds, 1);
      m_GLuiBounds = 0;
    }
}

BOOL CSLIDEBoundedNode::Initialized()
{
  if ( !CSLIDEGeometryNode::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

BOOL CSLIDEBoundedNode::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  if ( !CSLIDEGeometryNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  return TRUE;
}

BOOL CSLIDEBoundedNode::Valid()
{
  if ( !CSLIDEGeometryNode::Valid() )
    {
      return FALSE;
    }

  return TRUE;
}

CSLIDECamera *CSLIDEBoundedNode::FindCamera(CListIterInt *piterIDs, 
					    CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtCamera],
	      gcppcSLIDEReservedWords[sliderwtObject], GetID());

  return NULL;
}

CSLIDELight *CSLIDEBoundedNode::FindLight(CListIterInt *piterIDs, 
					  CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtLight],
	      gcppcSLIDEReservedWords[sliderwtObject], GetID());

  return NULL;
}

CSLIDEAreaLight *CSLIDEBoundedNode::FindAreaLight(CListIterInt *piterIDs, 
					  CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtAreaLight],
	      gcppcSLIDEReservedWords[sliderwtObject], GetID());

  return NULL;
}

VOID CSLIDEBoundedNode::GetSide(CListInt *plistPoint,CListInt *plistNormal, UINT_32 uiSide, 
				const Matrix &mTransform,  const Matrix &mInverse, 
				BOOL bReverse)
{
  ASSERT( plistPoint != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtAreaLight],
	      gcppcSLIDEReservedWords[sliderwtObject], GetID());
}

VOID CSLIDEBoundedNode::Preprocess()
{
  CSLIDEGeometryNode::Preprocess();

  m_bound.Null();
}

VOID CSLIDEBoundedNode::SetUnvisited()
{
  if( !GetVisited() )
    {
      return;
    }
  
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  
  SetVisited(FALSE);
  
  iter.Init(&m_listCSLIDEParentInstance);
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->SetUnvisited();
    }
}

VOID CSLIDEBoundedNode::InsertIntoBound(const Matrix &m, CBound &bound)
{
  bound.Insert(m, m_bound);
}


VOID CSLIDEBoundedNode::CheckDynamic()
{
  CSLIDEGeometryNode::CheckDynamic();
}

VOID CSLIDEBoundedNode::UpdateState()
{
  CSLIDEGeometryNode::UpdateState();
}

VOID CSLIDEBoundedNode::CalculateBound()
{
}

#ifdef __USE_CDA__
VOID CSLIDEBoundedNode::CalculateCDA()
{
  if ( m_bCDAVisited )
    {
      return;
    }
  m_bCDAVisited = TRUE;

  CListInt listCDAVertex;
  Matrix m;

  ASSERT( m_pCDA == NULL );

  listCDAVertex.Init();
  m.SetIdentity();

  CalculateCDAVertices(&listCDAVertex, m);

  if ( listCDAVertex.GetLength() > 0 )
    {
      CDAVertex *pCDAVertex;

      m_pCDA = AllocateCDA();
      ASSERT( m_pCDA != NULL );
      m_pCDA->Init(&listCDAVertex);

      // Clean up the Vertices
      for ( pCDAVertex = (CDAVertex*)(listCDAVertex.RemoveFirst());
	    pCDAVertex != NULL;
	    pCDAVertex = (CDAVertex*)(listCDAVertex.RemoveFirst()) )
	{
	  DeallocateCDAVertex(pCDAVertex);
	}   
    }
}

VOID CSLIDEBoundedNode::CalculateCDAVertices(CListInt * /*plistCDAVertex*/, 
					     const Matrix & /*mGroot_obj*/)
{
}
#endif // __USE_CDA__

VOID CSLIDEBoundedNode::SetPrerenderUnvisited()
{
  if( GetDynamic() || !GetPrerenderVisited() )
    {
      return;
    }
  
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  
  SetPrerenderVisited(FALSE);
  
  iter.Init(&m_listCSLIDEParentInstance);
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->SetPrerenderUnvisited();
    }

  // Delete Display lists so they get remade
  if ( gilIsList(m_GLuiFlatGeometry) )
    {
      gilDeleteLists(m_GLuiFlatGeometry, 1);
    }
  if ( m_GLuiFlatGeometry != m_GLuiSmoothGeometry && gilIsList(m_GLuiSmoothGeometry) )      
    {
      gilDeleteLists(m_GLuiSmoothGeometry, 1);
    }
  if ( gilIsList(m_GLuiFlatNormals) )
    {
      gilDeleteLists(m_GLuiFlatNormals, 1);
    }
  if ( m_GLuiFlatNormals != m_GLuiSmoothNormals && gilIsList(m_GLuiSmoothNormals) )      
    {
      gilDeleteLists(m_GLuiSmoothNormals, 1);
    }
  if ( m_GLuiBounds != 0 && gilIsList(m_GLuiBounds) )
    {
      gilDeleteLists(m_GLuiBounds, 1);
    }
}

VOID CSLIDEBoundedNode::Prerender()
{
  SLIDEShadingType eShading;
  GLuint uiList;

  // We should only Prerender once per node
  if ( m_bPrerenderVisited )
    {
      return;
    }

  //printf("BoundedNode::Prerender %s\n", this->GetID());


  m_bPrerenderVisited = TRUE;

  eShading = (SLIDEShadingType)m_dyniShading.GetValue();

  if ( !GetGeometryEmpty() && !GetGeometryDynamic() )
    {
      if ( eShading == slideshtSLF_INHERIT )
	{
	  // If the display list has never been initialized, then 
	  // grab a new global over contexts name 
	  if ( m_GLuiFlatGeometry == 0 )
	    {
	      m_GLuiFlatGeometry = GenLists();
	    }

	  // If the Display list does not exist, then this is the first time
	  // this name has been used in this drawing context so create it
	  if ( !gilIsList(m_GLuiFlatGeometry) )
	    {
	    /*
	      //////////////////////////////////////////////////
	      // Test to see if this context can support another Display List
	      uiList = gilGenLists(1);
	      ASSERT( gilIsList(uiList) );
	      // Should gracefully stop using Display Lists when we run out, but
	      // that is too complicated for right now!!! FIX THIS!!!!!
	      if ( !gilIsList(uiList) )
		{
		  goto error;
		}
	      gilDeleteLists(uiList, 1);
	      ASSERT( !gilIsList(uiList) );
	      // Test to see if this context can support another Display List
	      //////////////////////////////////////////////////
	      */

	      gilNewList(m_GLuiFlatGeometry, GL_COMPILE);
	      {
		PrerenderGeometry(slideshtSLF_INHERIT);
	      }
	      gilEndList();

	      ASSERT( gilIsList(m_GLuiFlatGeometry) );
	    }

	  // If the display list has never been initialized, then 
	  // grab a new global over contexts name 
	  if ( m_GLuiSmoothGeometry == 0 )
	    {
	      m_GLuiSmoothGeometry = GenLists();
	    }

	  // If the Display list does not exist, then this is the first time
	  // this name has been used in this drawing context so create it
	  if ( !gilIsList(m_GLuiSmoothGeometry) )
	    {
	      /*
	      //////////////////////////////////////////////////
	      // Test to see if this context can support another Display List
	      uiList = gilGenLists(1);
	      ASSERT( gilIsList(uiList) );
	      // Should gracefully stop using Display Lists when we run out, but
	      // that is too complicated for right now!!! FIX THIS!!!!!
	      if ( !gilIsList(uiList) )
		{
		  goto error;
		}
	      gilDeleteLists(uiList, 1);
	      ASSERT( !gilIsList(uiList) );
	      // Test to see if this context can support another Display List
	      //////////////////////////////////////////////////
	      */

	      gilNewList(m_GLuiSmoothGeometry, GL_COMPILE);
	      {
		PrerenderGeometry(slideshtSLF_GOURAUD);
	      }
	      gilEndList();

	      ASSERT( gilIsList(m_GLuiSmoothGeometry) );
	    }
	}
      else
	{
	  // If the display list has never been initialized, then 
	  // grab a new global over contexts name 
	  if ( m_GLuiFlatGeometry == 0 )
	    {
	      m_GLuiFlatGeometry = GenLists();
	    }

	  // If the Display list does not exist, then this is the first time
	  // this name has been used in this drawing context so create it
	  if ( !gilIsList(m_GLuiFlatGeometry) )
	    {
	      /*
	      //////////////////////////////////////////////////
	      // Test to see if this context can support another Display List
	      uiList = gilGenLists(1);
	      ASSERT( gilIsList(uiList) );
	      // Should gracefully stop using Display Lists when we run out, but
	      // that is too complicated for right now!!! FIX THIS!!!!!
	      if ( !gilIsList(uiList) )
		{
		  goto error;
		}
	      gilDeleteLists(uiList, 1);
	      ASSERT( !gilIsList(uiList) );
	      // Test to see if this context can support another Display List
	      //////////////////////////////////////////////////
	      */
	      gilNewList(m_GLuiFlatGeometry, GL_COMPILE);
	      {
		PrerenderGeometry((SLIDEShadingType)m_dyniShading.GetValue());
	      }
	      gilEndList();

	      ASSERT( gilIsList(m_GLuiFlatGeometry) );
	    }

	  // Requests for Smooth or Flat always get the same result
	  m_GLuiSmoothGeometry = m_GLuiFlatGeometry;
	}
    }

  if ( !GetNormalsEmpty() && !GetNormalsDynamic() )
    {
      if ( eShading == slideshtSLF_INHERIT )
	{
	  // If the display list has never been initialized, then 
	  // grab a new global over contexts name 
	  if ( m_GLuiFlatNormals == 0 )
	    {
	      m_GLuiFlatNormals = GenLists();
	    }

	  // If the Display list does not exist, then this is the first time
	  // this name has been used in this drawing context so create it
	  if ( !gilIsList(m_GLuiFlatNormals) )
	    {
	      /*
	      //////////////////////////////////////////////////
	      // Test to see if this context can support another Display List
	      uiList = gilGenLists(1);
	      ASSERT( gilIsList(uiList) );
	      // Should gracefully stop using Display Lists when we run out, but
	      // that is too complicated for right now!!! FIX THIS!!!!!
	      if ( !gilIsList(uiList) )
		{
		  goto error;
		}
	      gilDeleteLists(uiList, 1);
	      ASSERT( !gilIsList(uiList) );
	      // Test to see if this context can support another Display List
	      //////////////////////////////////////////////////
	      */
	      gilNewList(m_GLuiFlatNormals, GL_COMPILE);
	      {
		PrerenderNormals(slideshtSLF_FLAT);
	      }
	      gilEndList();

	      ASSERT( gilIsList(m_GLuiFlatNormals) );
	    }

	  // If the display list has never been initialized, then 
	  // grab a new global over contexts name 
	  if ( m_GLuiSmoothNormals == 0 )
	    {
	      m_GLuiSmoothNormals = GenLists();
	    }

	  // If the Display list does not exist, then this is the first time
	  // this name has been used in this drawing context so create it
	  if ( !gilIsList(m_GLuiSmoothNormals) )
	    {
	      /*
	      //////////////////////////////////////////////////
	      // Test to see if this context can support another Display List
	      uiList = gilGenLists(1);
	      ASSERT( gilIsList(uiList) );
	      // Should gracefully stop using Display Lists when we run out, but
	      // that is too complicated for right now!!! FIX THIS!!!!!
	      if ( !gilIsList(uiList) )
		{
		  goto error;
		}
	      gilDeleteLists(uiList, 1);
	      ASSERT( !gilIsList(uiList) );
	      // Test to see if this context can support another Display List
	      //////////////////////////////////////////////////
	      */
	      gilNewList(m_GLuiSmoothNormals, GL_COMPILE);
	      {
		PrerenderNormals(slideshtSLF_GOURAUD);
	      }
	      gilEndList();

	      ASSERT( gilIsList(m_GLuiSmoothNormals) );
	    }
	}
      else
	{
	  // If the display list has never been initialized, then 
	  // grab a new global over contexts name 
	  if ( m_GLuiFlatNormals == 0 )
	    {
	      m_GLuiFlatNormals = GenLists();
	    }

	  // If the Display list does not exist, then this is the first time
	  // this name has been used in this drawing context so create it
	  if ( !gilIsList(m_GLuiFlatNormals) )
	    {
	      /*
	      //////////////////////////////////////////////////
	      // Test to see if this context can support another Display List
	      uiList = gilGenLists(1);
	      ASSERT( gilIsList(uiList) );
	      // Should gracefully stop using Display Lists when we run out, but
	      // that is too complicated for right now!!! FIX THIS!!!!!
	      if ( !gilIsList(uiList) )
		{
		  goto error;
		}
	      gilDeleteLists(uiList, 1);
	      ASSERT( !gilIsList(uiList) );
	      // Test to see if this context can support another Display List
	      //////////////////////////////////////////////////
	      */

	      gilNewList(m_GLuiFlatNormals, GL_COMPILE);
	      {
		PrerenderNormals((SLIDEShadingType)m_dyniShading.GetValue());
	      }
	      gilEndList();

	      ASSERT( gilIsList(m_GLuiFlatNormals) );
	    }

	  // Requests for Smooth or Flat always get the same result
	  m_GLuiSmoothNormals = m_GLuiFlatNormals;
	}
    }

  if ( !GetBoundsEmpty() && !GetBoundsDynamic() )
    {
      // If the display list has never been initialized, then 
      // grab a new global over contexts name 
      if ( m_GLuiBounds == 0 )
	{
	  m_GLuiBounds = GenLists();
	}
      
      // If the Display list does not exist, then this is the first time
      // this name has been used in this drawing context so create it
      if ( !gilIsList(m_GLuiBounds) )
	{
	  /*
	  //////////////////////////////////////////////////
	  // Test to see if this context can support another Display List
	  uiList = gilGenLists(1);
	  ASSERT( gilIsList(uiList) );
	  // Should gracefully stop using Display Lists when we run out, but
	  // that is too complicated for right now!!! FIX THIS!!!!!
	  if ( !gilIsList(uiList) )
	    {
	      goto error;
	    }
	  gilDeleteLists(uiList, 1);
	  ASSERT( !gilIsList(uiList) );
	  // Test to see if this context can support another Display List
	  //////////////////////////////////////////////////
	  */

	  gilNewList(m_GLuiBounds, GL_COMPILE);
	  {
	    PrerenderBounds();
	  }
	  gilEndList();
	  
	  ASSERT( gilIsList(m_GLuiBounds) );
	}
    }

  return;

error:
  AbortDisplayLists();
}

VOID CSLIDEBoundedNode::PrerenderBounds()
{
  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
    {
      return;
    }

  m_bound.Render();
}

VOID CSLIDEBoundedNode::AbortDisplayLists()
{
  // FIX THIS!!!
  // This will not work for multiple contexts correctly!!!!
  ASSERT( FALSE );

  SetGeometryDynamic(TRUE);
  SetNormalsDynamic(TRUE);
  SetBoundsDynamic(TRUE);

  if ( gilIsList(m_GLuiFlatGeometry) )
    {
      gilDeleteLists(m_GLuiFlatGeometry, 1);
    }
  if ( m_GLuiFlatGeometry != m_GLuiSmoothGeometry && 
       gilIsList(m_GLuiSmoothGeometry) )      
    {
      gilDeleteLists(m_GLuiSmoothGeometry, 1);
    }
  m_GLuiFlatGeometry = 0;
  m_GLuiSmoothGeometry = 0;

  if ( gilIsList(m_GLuiFlatNormals) )
    {
      gilDeleteLists(m_GLuiFlatNormals, 1);
    }
  if ( m_GLuiFlatNormals != m_GLuiSmoothNormals && 
       gilIsList(m_GLuiSmoothNormals) )      
    {
      gilDeleteLists(m_GLuiSmoothNormals, 1);
    }
  m_GLuiFlatNormals = 0;
  m_GLuiSmoothNormals = 0;

  if ( m_GLuiBounds != 0 )
    {
      gilDeleteLists(m_GLuiBounds, 1);
      m_GLuiBounds = 0;
    }  
}

VOID CSLIDEBoundedNode::RecallGeometry(SLIDEShadingType eShading)
{
  //ASSERT( !GetGeometryDynamic() );
 
  if ( GetGeometryEmpty() )
    {
      return;
    }

  switch ( eShading )
    {
    case slideshtSLF_INHERIT:
    case slideshtSLF_WIRE:
    case slideshtSLF_FLAT:
      {
	ASSERT( gilIsList(m_GLuiFlatGeometry) );
	gilCallList(m_GLuiFlatGeometry);      
      }
      break;
    case slideshtSLF_GOURAUD:
    case slideshtSLF_PHONG:
      {
	ASSERT( gilIsList(m_GLuiSmoothGeometry) );
	gilCallList(m_GLuiSmoothGeometry);      
      }
      break;
    default:
      {
	ASSERT( FALSE );
      }
      break;
    }
}

VOID CSLIDEBoundedNode::RecallNormals(SLIDEShadingType eShading)
{
  ASSERT( !GetNormalsDynamic() );
 
  if ( GetNormalsEmpty() )
    {
      return;
    }

  switch ( eShading )
    {
    case slideshtSLF_INHERIT:
    case slideshtSLF_FLAT:
      {
	ASSERT( gilIsList(m_GLuiFlatNormals) );
	gilCallList(m_GLuiFlatNormals);      
      }
      break;
    case slideshtSLF_WIRE:
    case slideshtSLF_GOURAUD:
    case slideshtSLF_PHONG:
      {
	ASSERT( gilIsList(m_GLuiSmoothNormals) );
	gilCallList(m_GLuiSmoothNormals);      
      }
      break;
    default:
      {
	ASSERT( FALSE );
      }
      break;
    }
}

VOID CSLIDEBoundedNode::RecallBounds()
{
  ASSERT( !GetBoundsDynamic() );
 
  if ( GetBoundsEmpty() )
    {
      return;
    }

  ASSERT( gilIsList(m_GLuiBounds) );
  gilCallList(m_GLuiBounds);      
}

VOID CSLIDEBoundedNode::Render(const Matrix &mGproj_obj, 
			       CBound *pCBoundClip, 
			       SLIDEShadingType eShading,
			       SLIDELODType eLOD)
{
  CBound::BoundIntersectionType eIntersect;

  ModifyShadingFlag(eShading);
  ModifyLODFlag(eLOD);

  if ( eLOD == slidelodtSLF_OFF )
    {
      // This branch has been turned off by the LOD flags
      return;
    }

  // Check for Bound Rejection
  if ( GetBoundRejectEnabled() && pCBoundClip != NULL )
    {
      // We have not been trivially accepted so we must check our bound
      // against the clip bound
      eIntersect = CheckBoundReject(mGproj_obj, pCBoundClip);

      if ( eIntersect == CBound::bitDisjoint )
	{
	  // We are entirely outside of the clipping region
	  // Trivially Reject
	  eIntersect = CheckBoundReject(mGproj_obj, pCBoundClip);
	  return;
	}
      else if ( eIntersect == CBound::bitEnclosed )
	{
	  // We are entirely outside of the clipping region
	  // Trivially Accept
	  pCBoundClip = NULL;
	}
    }

  if ( eLOD > slidelodtSLF_BOUND )
    {
      //if ( !GetGeometryDynamic() )
      if ( gilIsList(m_GLuiFlatGeometry) ||
	   gilIsList(m_GLuiSmoothGeometry) )
	{
	  CheckAttributesStackDepth(GetPrerenderAttributesStackDepthMax());
	  RecallGeometry(eShading);
	}
      else
	{
	  SLIDEShadingType eShade;

	  eShade = (SLIDEShadingType)m_dyniShading.GetValue();

	  GLbitfield GLbAttributesMask;
	  
	  GLbAttributesMask = CalculateAttributesMask(eShade);
	  
	  PushAttributes(GLbAttributesMask);
	  {
	    PushShadingAttributes(eShade);
	    {
	      m_pCSLIDESurface->Push();
	      {
		RenderGeometry(mGproj_obj, pCBoundClip, eShading, eLOD);
	      }
	      m_pCSLIDESurface->Pop();
	    }
	    PopShadingAttributes(eShade);
	  }
	  PopAttributes(GLbAttributesMask);
	}

      // Render Normals if necessary
      if ( GetRenderNormalsEnabled() )
	{
	  if ( !GetNormalsDynamic() )
	    {
	      PushAuxillaryGeometryAttributes();
	      {
		gilColor3f(1.0, 0.0, 1.0);
		RecallNormals(eShading);
	      }
	      PopAuxillaryGeometryAttributes();
	    }
	  else
	    {
	      RenderNormals(eShading);
	    }
	}
    }

  // Render Bounding Box if necessary
  if ( GetRenderBoundsEnabled() && !GetBoundsDynamic() )
    {
      PushAuxillaryGeometryAttributes();
      {
	gilColor3f(1.0, 1.0, 1.0);
	RecallBounds();
      }
      PopAuxillaryGeometryAttributes();
    }
  else if ( GetRenderBoundsEnabled() || eLOD == slidelodtSLF_BOUND )
    {
      RenderBound();
    }
}

VOID CSLIDEBoundedNode::RenderSelect(const Matrix &mGproj_obj, 
				     CBound *pCBoundClip, 
				     SLIDEShadingType eShading,
				     SLIDELODType eLOD)
{
  CBound::BoundIntersectionType eIntersect;

  ModifyShadingFlag(eShading);
  ModifyLODFlag(eLOD);

  if ( eLOD == slidelodtSLF_OFF )
    {
      // This branch has been turned off by the LOD flags
      return;
    }

  // Check for Bound Rejection
  if ( GetBoundRejectEnabled() && pCBoundClip != NULL )
    {
      // We have not been trivially accepted so we must check our bound
      // against the clip bound
      eIntersect = CheckBoundReject(mGproj_obj, pCBoundClip);

      if ( eIntersect == CBound::bitDisjoint )
	{
	  // We are entirely outside of the clipping region
	  // Trivially Reject
	  eIntersect = CheckBoundReject(mGproj_obj, pCBoundClip);
	  return;
	}
      else if ( eIntersect == CBound::bitEnclosed )
	{
	  // We are entirely outside of the clipping region
	  // Trivially Accept
	  pCBoundClip = NULL;
	}
    }

  if ( eLOD > slidelodtSLF_BOUND )
    {
      RenderGeometrySelect(mGproj_obj, pCBoundClip, eShading, eLOD);
    }
}

CBound::BoundIntersectionType CSLIDEBoundedNode::CheckBoundReject(const Matrix &m,
						        CBound *pCBoundClip)
{
   ASSERT( pCBoundClip != NULL );

  return pCBoundClip->Intersect(m, m_bound);
}

VOID CSLIDEBoundedNode::RenderBound()
{
  PushAuxillaryGeometryAttributes();
  {
    if ( GetBoundChanged() && GetColorChangedBounds() )
      {
	// Orange for Changed Bounding Boxes
	gilColor3f(1.0, 0.373, 0.0);
      }
    else if ( GetBoundDynamic() && GetColorDynamicBounds() )
      {
	// Red for Dynamic Bounding Boxes
	gilColor3f(1.0, 0.0, 0.0);
      }
    else
      {
        // White for Static Bounding Boxes
	gilColor3f(1.0, 1.0, 1.0);
      }
    m_bound.Render();
  }
  PopAuxillaryGeometryAttributes();
}

VOID CSLIDEBoundedNode::RenderGraphSelectedBound()
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  
  iter.Init(&m_listCSLIDEParentInstance);
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->RenderGUISelectedBound();
    }
}

VOID CSLIDEBoundedNode::RenderGUISelectedBound()
{
  PushAuxillaryGeometryAttributes();
  {
    gilColor3f(1.0, 1.0, 0.0);
    m_bound.Render();
  }
  PopAuxillaryGeometryAttributes();
}

VOID CSLIDEBoundedNode::RenderGUISelectedAxes(BOOL bSelected, BOOL bXAxisSelected,
					      BOOL bYAxisSelected, BOOL bZAxisSelected)
{
  MATRIX4D m1;  
  Matrix m;
  Vector v;
  Point pMin, pMax;
  FLOAT f, fBoundRadius;
  
  pMin = m_bound.Get(CBound::bbtMin);
  pMax = m_bound.Get(CBound::bbtMax);
  gilGetFloatv(GL_MODELVIEW_MATRIX, m1);
  m.Set(m1);

  pMin = m * pMin;
  pMax = m * pMax;
  fBoundRadius = (pMax - pMin).Magnitude() * 0.5;

  gilPushMatrix();
  {
    gilLoadIdentity();

    // Normalize the current matrix
    v.Set(m1[0], m1[1], m1[2], m1[3]);
    f = v.Normalize();
    m1[0] = v[0];
    m1[1] = v[1];
    m1[2] = v[2];
    m1[3] = v[3];
    v.Set(m1[4], m1[5], m1[6], m1[7]);
    f = v.Normalize();
    m1[4] = v[0];
    m1[5] = v[1];
    m1[6] = v[2];
    m1[7] = v[3];
    v.Set(m1[8], m1[9], m1[10], m1[11]);
    f = v.Normalize();
    m1[8] = v[0];
    m1[9] = v[1];
    m1[10] = v[2];
    m1[11] = v[3];
    
    gilMultMatrixf(m1);

    PushAuxillaryGeometryAttributes();
    {
      m_pCSLIDECylinderAxis->SetRadiusValue(fBoundRadius/8.0);
      m_pCSLIDECylinderAxis->SetZMaxValue(20*fBoundRadius);
      m_pCSLIDECylinderAxis->SetZSlicesValue(10);
      m_pCSLIDECylinderAxis->SetThetaSlicesValue(4);

      // z-axis
      gilPushName(Z);
      {
	if( bSelected && bZAxisSelected )
	  gilColor3f(0.3, 0.3, 1.0);
	else
	  gilColor3f(0.0, 0.0, 1.0);

	m_pCSLIDECylinderAxis->RenderGeometryFlat();
      }
      gilPopName();

      // x-axis
      gilPushName(X);
      {
	if( bSelected && bXAxisSelected )
	  gilColor3f(1.0, 0.3, 0.3);
	else
	  gilColor3f(1.0, 0.0, 0.0);

	m.SetRotateY(DEGTORAD(90));
	gilMultMatrixf(m.Get());
	m_pCSLIDECylinderAxis->RenderGeometryFlat();
      }
      gilPopName();

      // y-axis
      gilPushName(Y);
      {
	if( bSelected && bYAxisSelected )
	  gilColor3f(0.4, 1.0, 0.4);
	else
	  gilColor3f(0.0, 1.0, 0.0);

	m.SetRotateX(DEGTORAD(-90));
	gilMultMatrixf(m.Get());
	m_pCSLIDECylinderAxis->RenderGeometryFlat();
      }
      gilPopName();
    }
    PopAuxillaryGeometryAttributes();
  }
  gilPopMatrix();
}

VOID CSLIDEBoundedNode::WriteRIB(FILE *fpOutputFile,
				 const Matrix &mGproj_obj, 
				 CBound *pCBoundClip, 
				 SLIDEShadingType eShading,
				 SLIDELODType eLOD)
{
  //CBound::BoundIntersectionType eIntersect;

  ModifyShadingFlag(eShading);
  ModifyLODFlag(eLOD);

  if ( eLOD == slidelodtSLF_OFF )
    {
      // This branch has been turned off by the LOD flags
      return;
    }

  if ( eLOD > slidelodtSLF_BOUND )
    {
      WriteRIBPushAttributes(fpOutputFile, eShading);
      {
	WriteRIBGeometry(fpOutputFile, 
			 mGproj_obj, pCBoundClip, eShading, eLOD);
      }
      WriteRIBPopAttributes(fpOutputFile, eShading);
    }

}

VOID CSLIDEBoundedNode::WritePOV(FILE *fpOutputFile, 
				 CListInt *plistTransformIDs,
				 CListInt *plistColors) 
{
  WritePOVGeometry(fpOutputFile, plistTransformIDs, plistColors);
}

VOID CSLIDEBoundedNode::WriteWRLZ(FILE *fpOutputFile, const Matrix &mTransform,
				  CSLIDESurface *pCSLIDESurface)
{
  WriteWRLZGeometry(fpOutputFile, mTransform, pCSLIDESurface);
}

VOID CSLIDEBoundedNode::WriteSTL(FILE *fpOutputFile, const Matrix &mTransform)
{
  WriteSTLGeometry(fpOutputFile, mTransform);
}

VOID CSLIDEBoundedNode::WriteRIBPushAttributes(FILE *fpOutputFile, 
					       SLIDEShadingType eShading)
{
  CSLIDEGeometryNode::WriteRIBPushAttributes(fpOutputFile, eShading);

  // This says that the following geometry lies within this bound
  WriteRIBPushBound(fpOutputFile);
}

VOID CSLIDEBoundedNode::WriteRIBPopAttributes(FILE *fpOutputFile, 
					      SLIDEShadingType eShading)
{
  WriteRIBPopBound(fpOutputFile);

  CSLIDEGeometryNode::WriteRIBPopAttributes(fpOutputFile, eShading);
}

VOID CSLIDEBoundedNode::WriteRIBPushBound(FILE * /*fpOutputFile */ )
{
  RtBound rtBound;

  if ( !m_bound.GetEmpty() )
    {
      RiAttributeBegin();

      rtBound[0] = m_bound.Get(CBound::bbtMin)(0);
      rtBound[1] = m_bound.Get(CBound::bbtMax)(0);
      rtBound[2] = m_bound.Get(CBound::bbtMin)(1);
      rtBound[3] = m_bound.Get(CBound::bbtMax)(1);
      rtBound[4] = m_bound.Get(CBound::bbtMin)(2);
      rtBound[5] = m_bound.Get(CBound::bbtMax)(2);

      RiBound(rtBound);
    }
}

VOID CSLIDEBoundedNode::WriteRIBPopBound(FILE * /*fpOutputFile */ )
{
  if ( !m_bound.GetEmpty() )
    {
      RiAttributeEnd();
    }
}

QEMesh *CSLIDEBoundedNode::GetQEMesh()
{
  if ( m_pQEMesh == NULL )
    {
      CalculateQEMesh();
      
      if ( m_pQEMesh == NULL )
	{
	  return NULL;
	}
    }

  return m_pQEMesh->Copy();
}

VOID CSLIDEBoundedNode::CalculateQEMesh()
{
}


VOID CSLIDEBoundedNode::InvalidateQEMesh()
{
  if ( m_pQEMesh != NULL && 
       GetBoundsDynamic() )
    {
      DeallocateQEMesh(m_pQEMesh);
      m_pQEMesh = NULL;
    }
}

VOID CSLIDEBoundedNode::CalculateLeafQEMesh()
{
}

VOID CSLIDEBoundedNode::DeallocateLeafQEMesh()
{
}







