#ifndef __SLIDEBOUNDEDNODE_H__
#define __SLIDEBOUNDEDNODE_H__

#include "slidegeometrynode.h"

//////////////////////////////////////////////////////////////////////
// CSLIDEBoundedNode Class
//

class CSLIDEBoundedNode : public CSLIDEGeometryNode
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//
public:

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEBoundedNode();
  virtual ~CSLIDEBoundedNode();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual CSLIDECamera *FindCamera(CListIterInt *piterIDs, 
				   CListInt *plistCSLIDEInstance);
  virtual CSLIDELight *FindLight(CListIterInt *piterIDs, 
				 CListInt *plistCSLIDEInstance);
  virtual CSLIDEAreaLight *FindAreaLight(CListIterInt *piterIDs, 
				 CListInt *plistCSLIDEInstance);
  virtual inline CBound GetBound();
  virtual VOID GetSide(CListInt *plistPoint, CListInt *plistNormal, UINT_32 uiSide, 
		       const Matrix &mTransform, const Matrix &mInverse, BOOL bReverse);
  virtual VOID Preprocess();
  virtual VOID SetUnvisited();

  virtual VOID InsertIntoBound(const Matrix &m, CBound &bound);

#ifdef __USE_CDA__
  virtual inline CDA *GetCDA();
  virtual VOID CalculateCDA();
  virtual VOID CalculateCDAVertices(CListInt *plistCDAVertex, 
				    const Matrix &mGroot_obj);
#endif // __USE_CDA__

  virtual QEMesh *GetQEMesh();
  virtual VOID CalculateQEMesh();
  virtual VOID InvalidateQEMesh();

  virtual VOID CalculateLeafQEMesh();  // for WriteWRLZ
  virtual VOID DeallocateLeafQEMesh();  // for WriteWRLZ

protected:
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();
  virtual VOID CalculateBound();

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID SetPrerenderUnvisited();
  virtual VOID Prerender();
  virtual VOID PrerenderBounds();
  virtual VOID AbortDisplayLists();

  virtual VOID RecallGeometry(SLIDEShadingType eShading);
  virtual VOID RecallNormals(SLIDEShadingType eShading);
  virtual VOID RecallBounds();

  virtual VOID Render(const Matrix &mGproj_obj,
		      CBound *pCBoundClip, 
		      SLIDEShadingType eShading,
		      SLIDELODType eLOD);
  virtual VOID RenderSelect(const Matrix &mGproj_obj,
			    CBound *pCBoundClip, 
			    SLIDEShadingType eShading,
			    SLIDELODType eLOD);
  virtual CBound::BoundIntersectionType CheckBoundReject(const Matrix &m, CBound *pCBound);

  virtual VOID RenderBound();
  virtual VOID RenderGraphSelectedBound();
  virtual VOID RenderGUISelectedBound();
  virtual VOID RenderGUISelectedAxes(BOOL bSelected, BOOL bXAxisSelected,
				     BOOL bYAxisSelected, BOOL bZAxisSelected);

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteRIB(FILE *fpOutputFile,
			const Matrix &mGproj_obj,
			CBound *pCBoundClip, 
			SLIDEShadingType eShading,
			SLIDELODType eLOD);

  virtual VOID WriteRIBPushAttributes(FILE *fpOutputFile,
				      SLIDEShadingType eShading);
  virtual VOID WriteRIBPopAttributes(FILE *fpOutputFile,
				     SLIDEShadingType eShading);

  VOID WriteRIBPushBound(FILE *fpOutputFile);
  VOID WriteRIBPopBound(FILE *fpOutputFile);

  virtual VOID WritePOV(FILE *fpOutputFile, 
			CListInt *plistTransformIDs,
			CListInt *plistColors); 
  virtual VOID WriteWRLZ(FILE *fpOutputFile, const Matrix &mTransform,
			 CSLIDESurface *pCSLIDESurface);
  virtual VOID WriteSTL(FILE *fpOutputFile, const Matrix &mTransform);

//////////////////////////////////////////////////
// Class Accessor Functions
//
public:
  static inline GLuint GenLists();

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline BOOL GetPrerenderVisited();
  virtual inline VOID SetPrerenderVisited(BOOL bPrerenderVisited);

//////////////////////////////////////////////////
// Class Variables
//
protected:
  static GLuint m_displayListCount;

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CBound m_bound;
#ifdef __USE_CDA__
  CDA *m_pCDA;
  BOOL m_bCDAVisited;
#endif // __USE_CDA__
  
  QEMesh *m_pQEMesh;

  BOOL m_bPrerenderVisited;

  GLuint m_GLuiFlatGeometry;
  GLuint m_GLuiFlatNormals;
  GLuint m_GLuiSmoothGeometry;
  GLuint m_GLuiSmoothNormals;
  GLuint m_GLuiBounds;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

#ifdef __USE_CDA__
inline CDA *CSLIDEBoundedNode::GetCDA()
{
  if ( m_pCDA == NULL )
    {
      return NULL;
    }

  return m_pCDA->Copy();
}
#endif // __USE_CDA__

inline GLuint CSLIDEBoundedNode::GenLists()
{
  m_displayListCount++;
  return m_displayListCount;
}

inline BOOL CSLIDEBoundedNode::GetPrerenderVisited()
{
  return m_bPrerenderVisited;
}

inline VOID CSLIDEBoundedNode::SetPrerenderVisited(BOOL bPrerenderVisited)
{
  m_bPrerenderVisited = bPrerenderVisited;
}

inline CBound CSLIDEBoundedNode::GetBound()
{
  return m_bound;
}

#endif // __SLIDEBOUNDEDNODE_H__
