#ifndef __SLIDEBSPLINEPATCH_H__
#define __SLIDEBSPLINEPATCH_H__

#include "slideatomnode.h"

//////////////////////////////////////////////////////////////////////////////
// CSLIDEBSplinePatch Class
//

class CSLIDEBSplinePatch : public CSLIDEAtomNode
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEBSplinePatch();
  virtual ~CSLIDEBSplinePatch();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  virtual VOID InitKnots();
  virtual VOID InitControlPoints();
  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);
  virtual BOOL Valid();

  virtual inline VOID SetUOrder(DynInt *pdyniUOrder);
  virtual inline VOID SetUKnots(DynInt *pdyniUKnots);
  virtual inline VOID SetUKnotList(CListInt *plistUKnot);
  virtual inline VOID SetVOrder(DynInt *pdyniVOrder);
  virtual inline VOID SetVKnots(DynInt *pdyniVKnots);
  virtual inline VOID SetVKnotList(CListInt *plistVKnot);
  virtual inline VOID SetControlPointList(CListInt *plistControlPoint);
  virtual inline VOID SetUSlices(DynInt *pdyniUSlices);
  virtual inline VOID SetVSlices(DynInt *pdyniVSlices);
  virtual inline VOID SetDrawControls(DynInt *pdyniDrawControls);
  virtual inline VOID SetTexture(DynVector *pdynv0, DynVector *pdynv1, DynVector *pdynv2, 
				 DynVector *pdynv3);
  virtual inline VOID SetOffset(DynFloat *pdynfOffset, DynInt *pdyniOffsetConnected);

protected:

  virtual VOID ReallocateUKnots();
  virtual VOID DeallocateUKnots();
  virtual VOID ReallocateVKnots();
  virtual VOID DeallocateVKnots();
  virtual VOID ReallocateControlPoints();
  virtual VOID DeallocateControlPoints();
  virtual VOID ReallocateSurfacePoints();
  virtual VOID DeallocateSurfacePoints();
  virtual VOID ReallocateSurfaceNormals();
  virtual VOID DeallocateSurfaceNormals();

  virtual VOID ReallocateDerivativeCurve();
  virtual VOID DeallocateDerivativeCurve();
  virtual VOID ReallocateCurveDerivativeTable();
  virtual VOID DeallocateCurveDerivativeTable();
  virtual VOID ReallocateCurveDeBoorTable();
  virtual VOID DeallocateCurveDeBoorTable();

//////////////////////////////////////////////////
// Geometric Computations
//
public:

#ifdef __USE_CDA__
  virtual VOID CalculateCDAVertices(CListInt *plistCDAVertex, const Matrix &mGroot_obj);
#endif // __USE_CDA__

protected:
  virtual VOID CheckDynamic();
  virtual VOID PreprocessGraphLocal();
  virtual VOID UpdateState();
  virtual VOID CalculateBound();
  virtual VOID CalculateQEMesh();
  virtual VOID CalculateLeafQEMesh();  // for WriteWRLZ
  virtual VOID DeallocateLeafQEMesh();  // for WriteWRLZ

public:
  virtual UINT_32 FindKnotInterval(FLOAT t, BOOL bUDirection);
  virtual UINT_32 FindBlossomKnotInterval(FLOAT t, BOOL bUDirection);

  virtual VOID PointAt(Point &pt, FLOAT fU, FLOAT fV);
  virtual VOID ControlCurveAt(Point *pt, FLOAT t, BOOL bUDirection);

protected:
  virtual VOID NormalAt(Vector &v, FLOAT fU, FLOAT fV);
  virtual VOID CurveTangentAt(Vector &vT, FLOAT fU, FLOAT fV, BOOL bUDirection);

  virtual VOID CalculateSurface();
  virtual VOID CalculateNormals();

//////////////////////////////////////////////////
// Rendering
//
public:

protected:
  virtual VOID RenderGeometrySmooth(CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderOffsetGeometrySmooth();
  virtual VOID RenderOffsetGeometryConnectedSmooth(); 
  virtual VOID RenderGeometryFlat(CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderOffsetGeometryFlat();
  virtual VOID RenderOffsetGeometryConnectedFlat(); 
  virtual VOID RenderNormalsSmooth();
  virtual VOID RenderOffsetNormalsSmooth();
  virtual VOID RenderNormalsFlat();
  virtual VOID RenderOffsetNormalsFlat();
  virtual VOID RenderControls();

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

  virtual VOID WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				 CSLIDESurface *pCSLIDESurface);
  virtual VOID WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline CHAR *GetTypeString();
  virtual inline SLIDEReservedWordType GetType();

  virtual VOID GetSide(CListInt *plistPoint, CListInt *plistNormal, UINT_32 uiSide, 
		       const Matrix &mTransform, const Matrix &mInverse, BOOL bReverse);

  virtual inline FLOAT GetTu();
  virtual inline FLOAT GetTuDelta();
  virtual inline FLOAT GetTv();
  virtual inline FLOAT GetTvDelta();

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  FLOAT m_fNormalLength;

  BOOL m_bControlPointsDynamic;
  BOOL m_bSlicesChanged;

  DynInt m_dyniDrawControls;
  DynInt m_dyniUSlices;
  UINT_32 m_uiUSlices;
  DynInt m_dyniVSlices;
  UINT_32 m_uiVSlices;
  UINT_32 m_uiUSlicesPrev;

  UINT_32 m_uiUOrderPrev;
  UINT_32 m_uiUOrder;
  UINT_32 m_uiUKnots;
  FLOAT *m_pfUKnot;
  CListInt *m_plistUKnot;

  UINT_32 m_uiVOrderPrev;
  UINT_32 m_uiVOrder;
  UINT_32 m_uiVKnots;
  FLOAT *m_pfVKnot;
  CListInt *m_plistVKnot;

  UINT_32 m_uiUControlPoints;
  UINT_32 m_uiVControlPoints;
  UINT_32 m_uiUControlPointsPrev;
  Point **m_ppControlPoint;
  CListInt *m_plistControlPointIDs;
  CListInt m_listCSLIDEControlPoint;

  Point **m_ppSurfacePoint;
  Vector **m_ppvSurfaceNormal;

  DynVector m_pvTexture[4];
  DynFloat m_dynfOffset;
  DynInt m_dyniOffsetConnected;

  // For calculating derivatives one control curve at a time
  Point *m_pDerivativeCurve;
  UINT_32 m_uiCurveDerivatives;
  UINT_32 m_uiCurveDerivativeControlPoints;
  Point **m_ppCurveDerivativeTable;
  FLOAT **m_ppfCurveDeBoorTable;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDEBSplinePatch::SetUOrder(DynInt *pdyniUOrder)
{
  ASSERT( pdyniUOrder->GetValue() >= 0 );
  m_uiUOrder = (UINT_32)(pdyniUOrder->GetValue());
}

inline VOID CSLIDEBSplinePatch::SetUKnots(DynInt *pdyniUKnots)
{
  ASSERT( pdyniUKnots->GetValue() >= 0 );
  m_uiUKnots = (UINT_32)(pdyniUKnots->GetValue());
}

inline VOID CSLIDEBSplinePatch::SetUKnotList(CListInt *plistUKnot)
{
  m_plistUKnot = plistUKnot;
}

inline VOID CSLIDEBSplinePatch::SetVOrder(DynInt *pdyniVOrder)
{
  ASSERT( pdyniVOrder->GetValue() >= 0 );
  m_uiVOrder = (UINT_32)(pdyniVOrder->GetValue());
}

inline VOID CSLIDEBSplinePatch::SetVKnots(DynInt *pdyniVKnots)
{
  ASSERT( pdyniVKnots->GetValue() >= 0 );
  m_uiVKnots = (UINT_32)(pdyniVKnots->GetValue());
}

inline VOID CSLIDEBSplinePatch::SetVKnotList(CListInt *plistVKnot)
{
  m_plistVKnot = plistVKnot;
}

inline VOID CSLIDEBSplinePatch::SetControlPointList(CListInt *plistControlPoint)
{
  m_plistControlPointIDs = plistControlPoint;
}

inline VOID CSLIDEBSplinePatch::SetUSlices(DynInt *pdyniUSlices)
{
  m_dyniUSlices.Set(pdyniUSlices);
}

inline VOID CSLIDEBSplinePatch::SetVSlices(DynInt *pdyniVSlices)
{
  m_dyniVSlices.Set(pdyniVSlices);
}

inline VOID CSLIDEBSplinePatch::SetDrawControls(DynInt *pdyniDrawControls)
{
  m_dyniDrawControls.Set(pdyniDrawControls);
}

inline VOID CSLIDEBSplinePatch::SetTexture(DynVector *pdynv0, DynVector *pdynv1, 
					   DynVector *pdynv2, DynVector *pdynv3)
{
  m_pvTexture[0].SetAll(pdynv0);
  m_pvTexture[1].SetAll(pdynv1);
  m_pvTexture[2].SetAll(pdynv2);
  m_pvTexture[3].SetAll(pdynv3);
}

inline VOID CSLIDEBSplinePatch::SetOffset(DynFloat *pdynfOffset, DynInt *pdyniOffsetConnected)
{
  m_dynfOffset.Set(pdynfOffset);
  m_dyniOffsetConnected.Set(pdyniOffsetConnected);
}

inline CHAR *CSLIDEBSplinePatch::GetTypeString()
{
  return gcppcSLIDEReservedWords[sliderwtBSplinePatch];
}

inline SLIDEReservedWordType CSLIDEBSplinePatch::GetType()
{
  return sliderwtBSplinePatch;
}

inline FLOAT CSLIDEBSplinePatch::GetTu() 
{
  return m_pfUKnot[m_uiUOrder-1];
}

inline FLOAT CSLIDEBSplinePatch::GetTuDelta() 
{
  return m_pfUKnot[m_uiUKnots - m_uiUOrder] - m_pfUKnot[m_uiUOrder-1];
}

inline FLOAT CSLIDEBSplinePatch::GetTv() 
{
  return m_pfVKnot[m_uiVOrder-1];
}

inline FLOAT CSLIDEBSplinePatch::GetTvDelta() 
{
  return m_pfVKnot[m_uiVKnots - m_uiVOrder] - m_pfVKnot[m_uiVOrder-1];
}

#endif // __SLIDEBSPLINEPATCH_H__
