#include "slidecontour.h"

#include "slideworld.h"
#include "slidesurface.h"

//////////////////////////////////////////////////////////////////////
// CSLIDEContour Class
//

CSLIDEContour::CSLIDEContour()
{
  m_pcSurfaceID = NULL;
  m_pCSLIDESurface = NULL;
  m_plistSegmentIDs = NULL;
  m_plistSegmentSides = NULL;
  m_plistSegmentReverse = NULL;
  m_plistPoint = NULL;
  m_plistNormal = NULL;

  m_pvTexture[0].Set(0.0, 0.0, 0.0);
  m_pvTexture[1].Set(1.0, 0.0, 0.0);
  m_pvTexture[2].Set(1.0, 1.0, 0.0);
  m_pvTexture[3].Set(0.0, 1.0, 0.0);
}

CSLIDEContour::~CSLIDEContour()
{
}

VOID CSLIDEContour::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID)
{
  CSLIDEAtomNode::Init(pcFile, uiLine, uiColumn, pcID);

  m_plistSegmentIDs = new CListInt();
  m_plistSegmentIDs->Init();
  m_plistSegmentSides = new CListInt();
  m_plistSegmentSides->Init();
  m_plistSegmentReverse = new CListInt();
  m_plistSegmentReverse->Init();
  m_plistPoint = new CListInt();
  m_plistPoint->Init();
  m_plistNormal = new CListInt();
  m_plistNormal->Init();

  m_listCSLIDENode.Init();

  m_pvTexture[0].Set(0.0, 0.0, 0.0);
  m_pvTexture[1].Set(1.0, 0.0, 0.0);
  m_pvTexture[2].Set(1.0, 1.0, 0.0);
  m_pvTexture[3].Set(0.0, 1.0, 0.0);
}

VOID CSLIDEContour::Uninit()
{
  CSLIDEAtomNode::Uninit();

  if ( m_plistSegmentIDs != NULL )
    {
      delete m_plistSegmentIDs;
      m_plistSegmentIDs = NULL;
    }

  if ( m_plistSegmentSides != NULL )
    {
      delete m_plistSegmentSides;
      m_plistSegmentSides = NULL;
    }

  if ( m_plistSegmentReverse != NULL )
    {
      delete m_plistSegmentReverse;
      m_plistSegmentReverse = NULL;
    }

  if ( m_plistPoint != NULL )
    {
      delete m_plistPoint;
      m_plistPoint = NULL;
    }

  if ( m_plistNormal != NULL )
    {
      delete m_plistNormal;
      m_plistNormal = NULL;
    }
}

BOOL CSLIDEContour::Initialized()
{
  if ( !CSLIDEAtomNode::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

VOID CSLIDEContour::AddSegment(CHAR *pcID, DynInt *pdyniSide, BOOL bReverse)
{
  ASSERT( (pcID != NULL) && (pdyniSide != NULL) );

  BOOL *pbReverse;

  pbReverse = new BOOL;
  *pbReverse = bReverse;

  m_plistSegmentIDs->InsertLast(pcID, 0);
  m_plistSegmentSides->InsertLast(pdyniSide, 0);
  m_plistSegmentReverse->InsertLast(pbReverse, 0);
}

BOOL CSLIDEContour::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;
  CListIterInt iter;
  CHAR *pcID;
  CSLIDENode *pCSLIDENode;

  if ( !CSLIDEAtomNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  // Check Segment References
  iter.Init(m_plistSegmentIDs);
  
  for ( pcID = (CHAR*)(iter.PeekFirst());
	pcID != NULL;
	pcID = (CHAR*)(iter.PeekNext()) )
    {
      pCSLIDENode = pCSLIDEWorld->PeekNode(pcID);
      if ( pCSLIDENode == NULL )
	{
	  HandleError("Undefined Group \"%s\"", pcID);
	  bCorrect = FALSE;
	}
      m_listCSLIDENode.InsertLast(pCSLIDENode, 0);
    }
  
  return bCorrect;
}

BOOL CSLIDEContour::Valid()
{
  if ( !CSLIDEAtomNode::Valid() )
    {
      return FALSE;
    }

  return TRUE;
  /*
  return ( ( m_dynfYMinorB.GetValue() <= m_dynfXMajorA.GetValue() ) &&
	   ( 0.0 < m_dynfYMinorB.GetValue() ) )
	   */
}

UINT_32 CSLIDEContour::GetPointCount()
{
  UINT_32 uiPoints;
  CListIterInt iter;
  Point *pt;
  Point p0, p1;

  iter.Init(m_plistPoint);
  uiPoints = 0;
  for ( pt = (Point *)(iter.PeekFirst()); 
	pt != NULL;
	pt = (Point *)(iter.PeekNext()) )
    {
      p0 = *pt;

      if( (uiPoints == 0) || (p0 != p1) )
	{
	  uiPoints++;
	  p1 = p0;
	}
    }
  
  return uiPoints;
}

VOID CSLIDEContour::GetPoints(Point *points)
{
  UINT_32 i;
  CListIterInt iter;
  Point *pt;
  Point p0, p1;

  iter.Init(m_plistPoint);

  for ( pt = (Point *)(iter.PeekFirst()), i = 0; 
	pt != NULL;
	pt = (Point *)(iter.PeekNext()) )
    {
      p0 = *pt;

      if( (i == 0) || (p0 != p1) )
	{
	  points[i] = *pt;
	  i++;
	  p1 = p0;
	}
    }
}

VOID CSLIDEContour::GetNormals(Vector *normals)
{
  UINT_32 i;
  CListIterInt iter;
  Vector *v;

  iter.Init(m_plistNormal);

  for ( v = (Vector *)(iter.PeekFirst()), i = 0; 
	v != NULL;
	v = (Vector *)(iter.PeekNext()), i++ )
    {
      normals[i] = *v;
    }
}

VOID CSLIDEContour::CheckDynamic()
{
  CSLIDEAtomNode::CheckDynamic();

  SetDynamic(GetDynamic() ||
	     m_pvTexture[0].GetDynamic() ||
	     m_pvTexture[1].GetDynamic() ||
	     m_pvTexture[2].GetDynamic() ||
	     m_pvTexture[3].GetDynamic());
}

VOID CSLIDEContour::PreprocessGraph()
{
  // Like slidegroup
  
  if ( !GetVisited() )
    {
      CListIterInt iter;
      CSLIDENode *pCSLIDENode;

      BOOL bGeometryEmpty;
      BOOL bGeometryDynamic;
      BOOL bNormalsEmpty;
      BOOL bNormalsDynamic;
      BOOL bBoundsEmpty;
      BOOL bBoundsDynamic;
      BOOL bBoundEmpty;
      BOOL bBoundDynamic;

      bGeometryEmpty = TRUE;
      bGeometryDynamic = FALSE;
      bNormalsEmpty = TRUE;
      bNormalsDynamic = FALSE;
      bBoundsEmpty = TRUE;
      bBoundsDynamic = FALSE;
      bBoundEmpty = TRUE;
      bBoundDynamic = FALSE;

      //////////////////////////////////////////////////
      // Visit Children Nodes and get starting values from them
      
      iter.Init(&m_listCSLIDENode);
      
      for ( pCSLIDENode = (CSLIDENode*)iter.PeekFirst(); 
	    pCSLIDENode != NULL;
	    pCSLIDENode = (CSLIDENode*)iter.PeekNext() )
	{
	  pCSLIDENode->PreprocessGraph();
	  
	  bGeometryEmpty = ( bGeometryEmpty && 
			     pCSLIDENode->GetGeometryEmpty() );
	  bGeometryDynamic = ( bGeometryDynamic ||
			       pCSLIDENode->GetGeometryDynamic() );
	  bNormalsEmpty = ( bNormalsEmpty && 
			    pCSLIDENode->GetNormalsEmpty() );
	  bNormalsDynamic = ( bNormalsDynamic ||
			      pCSLIDENode->GetNormalsDynamic() );
	  bBoundsEmpty = ( bBoundsEmpty && 
			   pCSLIDENode->GetBoundsEmpty() );
	  bBoundsDynamic = ( bBoundsDynamic ||
			     pCSLIDENode->GetBoundsDynamic() );
	  bBoundEmpty = ( bBoundEmpty && 
			  pCSLIDENode->GetBoundEmpty() );
	  bBoundDynamic = ( bBoundDynamic ||
			    pCSLIDENode->GetBoundDynamic() ||
			    pCSLIDENode->GetGeometryDynamic() ); // must add this
	}

      //////////////////////////////////////////////////
      // Look at Local flags

      // Update whether we are Empty based on our LOD flag
      if ( !m_dyniLOD.GetDynamic() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      // This node is always off
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	      bBoundsEmpty = TRUE;
	      bBoundEmpty = TRUE;
	    }
	  else if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	    {
	      // This node is always just a bounding box
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	    }
	}
      
      // Check what is really dynamic
      bGeometryDynamic = ( bGeometryDynamic ||
			   m_pCSLIDESurface->GetDynamic() ||
			   m_dyniShading.GetDynamic() ||
			   m_dyniLOD.GetDynamic() );
      bNormalsDynamic = ( bNormalsDynamic ||
			  m_dyniShading.GetDynamic() ||
			  m_dyniLOD.GetDynamic() );
      bBoundsDynamic = ( bBoundsDynamic || 
			 m_dyniLOD.GetDynamic() );
      bBoundDynamic = ( bBoundDynamic || 
			m_dyniLOD.GetDynamic() );

      // If anything is now Empty then it will always be empty so
      // it is not dynamic
      bGeometryDynamic = ( bGeometryDynamic && !bGeometryEmpty );
      bNormalsDynamic = ( bNormalsDynamic && !bNormalsEmpty );
      bBoundsDynamic = ( bBoundsDynamic && !bBoundsEmpty );
      bBoundDynamic = ( bBoundDynamic && !bBoundEmpty );

      //////////////////////////////////////////////////
      // Now save these correct flag values
      SetGeometryEmpty(bGeometryEmpty);
      SetGeometryDynamic(bGeometryDynamic);
      SetNormalsEmpty(bNormalsEmpty);
      SetNormalsDynamic(bNormalsDynamic);
      SetBoundsEmpty(bBoundsEmpty);
      SetBoundsDynamic(bBoundsDynamic);
      SetBoundEmpty(bBoundEmpty);
      SetBoundDynamic(bBoundDynamic);
      
      SetDynamic(bGeometryDynamic); // must add this

      //////////////////////////////////////////////////
      // Calculate the Bounding Box (if Empty this will do nothing)
      if ( GetBoundEmpty() )
	{
	  SetBoundValid(TRUE);
	}
      else
	{
	  CalculateBound();
	}
      SetBoundChanged(FALSE);

      //////////////////////////////////////////////////
      // Done Preprocessing this Node in the Graph
      SetVisited(TRUE);
    }
}

VOID CSLIDEContour::UpdateState()
{
  CSLIDEAtomNode::UpdateState();

  UINT_32 uiSide, i;
  CListIterInt iter, iter2, iter3;
  DynInt *pDyniSide;
  CSLIDENode *pCSLIDENode;
  Matrix mT, mI;
  BOOL *pbReverse;
  
  if( m_plistPoint != NULL )
    {
      delete m_plistPoint;
      delete m_plistNormal;
    }
  m_plistPoint = new CListInt;
  m_plistPoint->Init();
  m_plistNormal = new CListInt;
  m_plistNormal->Init();
  
  iter.Init(&m_listCSLIDENode);
  iter2.Init(m_plistSegmentSides);
  iter3.Init(m_plistSegmentReverse);
  
  for ( pCSLIDENode = (CSLIDENode*)(iter.PeekFirst()),
	  pDyniSide = (DynInt*)(iter2.PeekFirst()),
	  pbReverse = (BOOL*)(iter3.PeekFirst());
	pCSLIDENode != NULL;
	pCSLIDENode = (CSLIDENode*)(iter.PeekNext()),
	      pDyniSide = (DynInt*)(iter2.PeekNext()),
	      pbReverse = (BOOL*)(iter3.PeekNext()) )
    {
      ASSERT( pDyniSide != NULL );
      
      uiSide = pDyniSide->GetValue();
      if( uiSide > 3 )
	{
	  HandleError("Contour side (%ld) > 3", uiSide, *pbReverse);
	}
      else
	{
	  pCSLIDENode->GetSide(m_plistPoint, m_plistNormal, uiSide, mT, mI, *pbReverse);
	}
    }

  for ( i = 0; i < 4; i++ )
    {
      m_pvTexture[i].Update();
    }
}

VOID CSLIDEContour::CalculateBound()
{
  CBound bound;
  CListIterInt iter;
  Point *pt;

  iter.Init(m_plistPoint);

  for ( pt = (Point *)(iter.PeekFirst()); 
	pt != NULL;
	pt = (Point *)(iter.PeekNext()) )
    {
      bound.Insert(*pt);
    }

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);
    }
}

#ifdef __USE_CDA__
VOID CSLIDEContour::CalculateCDAVertices(CListInt *plistCDAVertex, 
				       const Matrix &mGroot_obj)
{

}
#endif // __USE_CDA__

VOID CSLIDEContour::RenderGeometrySmooth()
{
  RenderGeometryFlat();
}

VOID CSLIDEContour::RenderGeometryFlat()
{
  CListIterInt iter;
  Point *pt;

  gilBegin(GL_LINE_STRIP);
  {	
    iter.Init(m_plistPoint);
    
    for( pt = (Point*)iter.PeekFirst(); 
	 pt != NULL;
	 pt = (Point*)iter.PeekNext() )
      {
	gilVertex3fv(pt->Get());
      }
  }
  gilEnd();
}

VOID CSLIDEContour::WriteSLIDE(FILE *fpOutputFile)
{
  /*
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtContour], m_pcGUIID);

  WriteSLIDEFlags(fpOutputFile, 0);

  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtXMajorA],
	  m_dynfXMajorA.GetValue());
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtYMinorB],
	  m_dynfYMinorB.GetValue());
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtStringLength],
	  m_dynfStringLength.GetValue());
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtPhiMin],
	  RADTODEG(m_dynfPhiMin.GetValue()));
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtPhiMax],
	  RADTODEG(m_dynfPhiMax.GetValue()));
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtThetaMin],
	  RADTODEG(m_dynfThetaMin.GetValue()));
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtThetaMax],
	  RADTODEG(m_dynfThetaMax.GetValue()));
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtPhiSlices],
	  m_dyniPhiSlices.GetValue());
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtThetaSlices],
	  m_dyniThetaSlices.GetValue());
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtDrawEllipse],
	  m_dyniDrawEllipse.GetValue());
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtDrawHyperbola],
	  m_dyniDrawHyperbola.GetValue());

  
  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndContour]);
	  */
}

VOID CSLIDEContour::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE Contour: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);
}

VOID CSLIDEContour::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE Contour: %s", m_pcGUIID);
}

VOID CSLIDEContour::WriteDotID(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "contour_%s", GetGUIID());
}

VOID CSLIDEContour::WriteDotNode(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "  ");
  WriteDotID(fpOutputFile);
  fprintf(fpOutputFile, " [ label = \"%s\", shape = ellipse, peripheries = 2", GetGUIID());
  if ( GetBoundDynamic() )
    {
      fprintf(fpOutputFile, ", color = red, fontcolor = red");
    }
  else if ( GetGeometryDynamic() )
    {
      fprintf(fpOutputFile, ", color = blue, fontcolor = blue");
    }
  fprintf(fpOutputFile, " ];\n");
}

VOID CSLIDEContour::WriteDotEdges(FILE * /*fpOutputFile*/)
{
}
