#ifndef __SLIDECONTOUR_H__
#define __SLIDECONTOUR_H__

#include "slideatomnode.h"

//////////////////////////////////////////////////////////////////////////////
// CSLIDEContour Class
//

class CSLIDEContour : public CSLIDEAtomNode
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEContour();
  virtual ~CSLIDEContour();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);

  virtual VOID Uninit();

  virtual BOOL Initialized();

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual VOID AddSegment(CHAR *pcID, DynInt *pdyniSide, BOOL bReverse);
  inline VOID SetTexture(DynVector *pdynv0, DynVector *pdynv1, DynVector *pdynv2, DynVector *pdynv3);
  virtual BOOL Valid();

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual VOID PreprocessGraph();

#ifdef __USE_CDA__
  virtual VOID CalculateCDAVertices(CListInt *plistCDAVertex, 
				    const Matrix &mGroot_obj);
#endif // __USE_CDA__

protected:
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();
  virtual VOID CalculateBound();

//////////////////////////////////////////////////
// Rendering
//
public:

protected:
  virtual VOID RenderGeometrySmooth();
  virtual VOID RenderGeometryFlat();

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

  virtual VOID WriteDotID(FILE *fpOutputFile);
  virtual VOID WriteDotNode(FILE *fpOutputFile);
  virtual VOID WriteDotEdges(FILE *fpOutputFile);
//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual UINT_32 GetPointCount();
  virtual VOID GetPoints(Point *points);
  virtual VOID GetNormals(Vector *normals);

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CHAR *m_pcSurfaceID;
  CSLIDESurface *m_pCSLIDESurface;

  CListInt *m_plistSegmentIDs;
  CListInt m_listCSLIDENode;
  CListInt *m_plistSegmentSides;
  CListInt *m_plistSegmentReverse;

  CListInt *m_plistPoint;
  CListInt *m_plistNormal;

  DynVector m_pvTexture[4];
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//



inline VOID CSLIDEContour::SetTexture(DynVector *pdynv0, DynVector *pdynv1, 
				      DynVector *pdynv2, DynVector *pdynv3)
{
  m_pvTexture[0].SetAll(pdynv0);
  m_pvTexture[1].SetAll(pdynv1);
  m_pvTexture[2].SetAll(pdynv2);
  m_pvTexture[3].SetAll(pdynv3);
}

#endif // __SLIDECONTOUR_H__
