#include "slidecontrolpoint.h"
#include "slideworld.h"
#include "slidecrosssection.h"

//////////////////////////////////////////////////////////////////////////////
// CSLIDEControlPoint Class
//

CSLIDEControlPoint::CSLIDEControlPoint()
{
  m_bBoundDynamic = FALSE;
  m_bGeometryDynamic = FALSE;

  m_pcPointID = NULL;
  m_pCSLIDEPoint = NULL;

  m_pcCrossSectionID = NULL;
  m_pCSLIDECrossSection = NULL;

  m_pdvScale = new DynVector();
  m_pdvScale->Set(1.0, 1.0, 1.0);
  m_pdvRotate = new DynVector();
  m_pdvRotate->Set(0.0, 0.0, 0.0);
}

CSLIDEControlPoint::~CSLIDEControlPoint()
{
}

VOID CSLIDEControlPoint::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			      CHAR *pcID)
{
  CSLIDEPrimitive::Init(pcFile, uiLine, uiColumn, pcID);

  m_bBoundDynamic = FALSE;
  m_bGeometryDynamic = FALSE;

  m_pcPointID = NULL;
  m_pCSLIDEPoint = NULL;

  m_pcCrossSectionID = NULL;
  m_pCSLIDECrossSection = NULL;
}

VOID CSLIDEControlPoint::Uninit()
{
  CSLIDEPrimitive::Uninit();
}

BOOL CSLIDEControlPoint::Initialized()
{
  if ( !CSLIDEPrimitive::Initialized() )
    {
      return FALSE;
    }

  return m_pcPointID != NULL; 
}

BOOL CSLIDEControlPoint::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;

  if ( !CSLIDEPrimitive::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  // Check Point Reference
  m_pCSLIDEPoint = pCSLIDEWorld->PeekPoint(m_pcPointID);
  if ( m_pCSLIDEPoint == NULL )
    {
      HandleError("Undefined %s \"%s\"", 
		  gcppcSLIDEReservedWords[sliderwtPoint],
		  m_pcPointID);
      bCorrect = FALSE;
    }

  // Check CrossSection Reference
  if( m_pcCrossSectionID != NULL )
    {
      m_pCSLIDECrossSection = (CSLIDECrossSection*)pCSLIDEWorld->PeekNode(m_pcCrossSectionID);
      if ( m_pCSLIDECrossSection == NULL )
	{
	  HandleError("Undefined %s \"%s\"", 
		      gcppcSLIDEReservedWords[sliderwtCrossSection],
		      m_pcCrossSectionID);
	  bCorrect = FALSE;
	}
    }

  return bCorrect;
}

BOOL CSLIDEControlPoint::Valid()
{
  if ( !CSLIDEPrimitive::Valid() )
    {
      return FALSE;
    }

  return TRUE;
}

VOID CSLIDEControlPoint::CheckDynamic()
{
  SetDynamic(m_pCSLIDEPoint->GetDynamic() ||
	     m_pdvScale->GetDynamic() ||
	     m_pdvRotate->GetDynamic());

  if( !GetDynamic() && (m_pCSLIDECrossSection != NULL) )
    {
      SetDynamic( m_pCSLIDECrossSection->GetDynamic() );
    }

  SetBoundDynamic(GetDynamic());
  SetGeometryDynamic(GetDynamic());
}

BOOL CSLIDEControlPoint::GetCrossSectionDynamic()
{
  if( m_pCSLIDECrossSection == NULL )
    {
      return FALSE;
    }
  else
    {
      return (m_pCSLIDECrossSection->GetDynamic());
    }
}

VOID CSLIDEControlPoint::PreprocessGraphLocalCrossSection()
{
  if( m_pCSLIDECrossSection != NULL )
    {
      m_pCSLIDECrossSection->PreprocessGraphLocal();
    }
}

VOID CSLIDEControlPoint::UpdateState()
{
  // Don't update, sweeppath will update
  //m_pdvScale->Update();
  //m_pdvRotate->Update();
}

VOID CSLIDEControlPoint::WriteSLIDE(FILE *fpOutputFile)
{
  Vector v;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtControlPoint], m_pcGUIID);

  fprintf(fpOutputFile, "  %s %s\n",
	  gcppcSLIDEReservedWords[sliderwtPoint], 
	  m_pcPointID);

  if( m_pcCrossSectionID != NULL )
    {
      fprintf(fpOutputFile, "  %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtCrossSection], 
	      m_pcCrossSectionID);
    }

  fprintf(fpOutputFile, "  %s ( %.3f %.3f %.3f )\n",
	  gcppcSLIDEReservedWords[sliderwtScale], 
	  (m_pdvScale->Get())[X], (m_pdvScale->Get())[Y], (m_pdvScale->Get())[Z]); 

  fprintf(fpOutputFile, "  %s ( %.3f %.3f %.3f )\n",
	  gcppcSLIDEReservedWords[sliderwtRotate], 
	  (m_pdvRotate->Get())[X], (m_pdvRotate->Get())[Y], (m_pdvRotate->Get())[Z]); 

  CSLIDEPrimitive::WriteSLIDEFlags(fpOutputFile, 0);

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndControlPoint]); 
}
