#ifndef __CONTROLPOINT_H__
#define __CONTROLPOINT_H__

#include "slideprimitive.h"
#include "slidepoint.h"

class CSLIDECrossSection;

//////////////////////////////////////////////////////////////////////
// CSLIDEControlPoint Class
//

class CSLIDEControlPoint : public CSLIDEPrimitive
{
//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEControlPoint();
  virtual ~CSLIDEControlPoint();

  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  inline VOID SetPointID(CHAR *pcPointID);
  inline VOID SetCrossSectionID(CHAR *pcCrossSectionID);
  inline VOID SetScale(DynVector *pdynvScale);
  inline VOID SetRotate(DynVector *pdynvRotate);

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  BOOL GetCrossSectionDynamic();

  VOID CheckDynamic();
  VOID UpdateState();

  VOID PreprocessGraphLocalCrossSection();

//////////////////////////////////////////////////
// Rendering
//
public:

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  inline CHAR* GetPointID();
  inline CSLIDEPoint* GetPoint();

  inline CSLIDECrossSection* GetCrossSection();
  inline CHAR* GetCrossSectionID();
  inline DynVector* GetScale();
  inline DynVector* GetRotate();

  inline BOOL GetBoundDynamic();
  inline VOID SetBoundDynamic(BOOL bBoundDynamic);
  inline BOOL GetGeometryDynamic();
  inline VOID SetGeometryDynamic(BOOL bGeometryDynamic);

//////////////////////////////////////////////////
// Class Variables
//
public:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  BOOL m_bBoundDynamic;
  BOOL m_bGeometryDynamic;

  CHAR *m_pcPointID;
  CSLIDEPoint *m_pCSLIDEPoint;

  CHAR *m_pcCrossSectionID;
  CSLIDECrossSection *m_pCSLIDECrossSection;

  DynVector *m_pdvScale;
  DynVector *m_pdvRotate;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDEControlPoint::SetPointID(CHAR *pcPointID)
{
  m_pcPointID = pcPointID;
}

inline CHAR* CSLIDEControlPoint::GetPointID()
{
  return m_pcPointID;
}

inline CSLIDEPoint* CSLIDEControlPoint::GetPoint()
{
  return m_pCSLIDEPoint;
}

inline CSLIDECrossSection* CSLIDEControlPoint::GetCrossSection()
{
  return m_pCSLIDECrossSection;
}

inline CHAR* CSLIDEControlPoint::GetCrossSectionID()
{
  return m_pcCrossSectionID;
}

inline VOID CSLIDEControlPoint::SetCrossSectionID(CHAR *pcCrossSectionID)
{
  m_pcCrossSectionID = pcCrossSectionID;
}

inline VOID CSLIDEControlPoint::SetScale(DynVector *pdynvScale)
{
  m_pdvScale->SetAll( pdynvScale );
}

inline VOID CSLIDEControlPoint::SetRotate(DynVector *pdynvRotate)
{
  m_pdvRotate->SetAll( pdynvRotate );
}

inline DynVector* CSLIDEControlPoint::GetScale()
{
  return m_pdvScale;
}

inline DynVector* CSLIDEControlPoint::GetRotate()
{
  return m_pdvRotate;
}

inline BOOL CSLIDEControlPoint::GetBoundDynamic()
{
  return m_bBoundDynamic;
}

inline VOID CSLIDEControlPoint::SetBoundDynamic(BOOL bBoundDynamic)
{
  m_bBoundDynamic = bBoundDynamic;
}

inline BOOL CSLIDEControlPoint::GetGeometryDynamic()
{
  return m_bGeometryDynamic;
}

inline VOID CSLIDEControlPoint::SetGeometryDynamic(BOOL bGeometryDynamic)
{
  m_bGeometryDynamic = bGeometryDynamic;
}


#endif // __CONTROLPOINT_H__


