#include "slideconvexhull.h"

#include "slideworld.h"
#include "slidesurface.h"

#include "slidepoint.h"

#include "qemesh.h"
#include "qerender.h"
#include "qewrite.h"

//////////////////////////////////////////////////////////////////////
// CSLIDEConvexHull Class
//

CSLIDEConvexHull::CSLIDEConvexHull()
{
  m_dyniSolid.SetValue(slidestSLF_SOLID);
  m_plistPointIDs = NULL;
}

CSLIDEConvexHull::~CSLIDEConvexHull()
{
}

VOID CSLIDEConvexHull::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			 CHAR *pcID,
			 CListInt *plistPointIDs)
{
  CSLIDEBoundedNode::Init(pcFile, uiLine, uiColumn, pcID);

  m_dyniSolid.SetValue(slidestSLF_SOLID);

  m_plistPointIDs = plistPointIDs;
  m_listCSLIDEPoint.Init();
}

VOID CSLIDEConvexHull::Uninit()
{
  CSLIDEBoundedNode::Uninit();

  if ( m_plistPointIDs != NULL )
    {
      delete m_plistPointIDs;
      m_plistPointIDs = NULL;
    }
}

BOOL CSLIDEConvexHull::Initialized()
{
  if ( !CSLIDEBoundedNode::Initialized() )
    {
      return FALSE;
    }

  return (m_plistPointIDs != NULL);
}

BOOL CSLIDEConvexHull::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;
  CListIterInt iter;
  CHAR *pcID;
  CSLIDEPoint *pCSLIDEPoint;

  if ( !CSLIDEBoundedNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  // Check Solid Type

  ASSERT( 0 <= m_dyniSolid.GetValue() &&
	  m_dyniSolid.GetValue() < slidestSize );

  // Check Point References

  iter.Init(m_plistPointIDs);

  for ( pcID = (CHAR*)(iter.PeekFirst());
	pcID != NULL;
	pcID = (CHAR*)(iter.PeekNext()) )
    {
      pCSLIDEPoint = pCSLIDEWorld->PeekPoint(pcID);
      if ( pCSLIDEPoint == NULL )
	{
	  HandleError("Undefined %s \"%s\"", gcppcSLIDEReservedWords[sliderwtPoint], pcID);
	  bCorrect = FALSE;
	}
      m_listCSLIDEPoint.InsertLast(pCSLIDEPoint, 0);
    }

  return bCorrect;
}

BOOL CSLIDEConvexHull::Valid()
{
  if ( !CSLIDEBoundedNode::Valid() )
    {
      return FALSE;
    }

  return ( ( 0 <= m_dyniSolid.GetValue() &&
	     m_dyniSolid.GetValue() < slidestSize ) );
	   //GetPointCount() > 0 );
}

CSLIDECamera *CSLIDEConvexHull::FindCamera(CListIterInt *piterIDs, 
				       CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtCamera],
	      gcppcSLIDEReservedWords[sliderwtConvexHull], GetID());

  return NULL;
}

CSLIDELight *CSLIDEConvexHull::FindLight(CListIterInt *piterIDs, 
				     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtLight],
	      gcppcSLIDEReservedWords[sliderwtConvexHull], GetID());

  return NULL;
}

CSLIDEAreaLight *CSLIDEConvexHull::FindAreaLight(CListIterInt *piterIDs, 
					     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtAreaLight],
	      gcppcSLIDEReservedWords[sliderwtConvexHull], GetID());

  return NULL;
}

VOID CSLIDEConvexHull::Preprocess()
{
  CSLIDEBoundedNode::Preprocess();

  CheckDynamic();
  UpdateState();
}

VOID CSLIDEConvexHull::CheckDynamic()
{
  CListIterInt iter;
  //CSLIDEPoint *pCSLIDEPoint;
  //BOOL bAutoNormals;

  // Are there variables which I must call TCL to update?
  SetDynamic(m_dyniShading.GetDynamic() ||
	     m_dyniLOD.GetDynamic() ||
	     m_dyniSolid.GetDynamic());
/*
  bAutoNormals = !(!m_dyniShading.GetDynamic() && 
		   (m_dyniShading.GetValue() == 
		    (INT_32)slideshtSLF_WIRE ||
		    m_dyniShading.GetValue() == 
		    (INT_32)slideshtSLF_FLAT));
      
  iter.Init(&m_listCSLIDEPoint);
  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      pCSLIDEPoint->SetAutoNormals(bAutoNormals || pCSLIDEPoint->GetAutoNormals());
    }
*/
}

VOID CSLIDEConvexHull::PreprocessGraph()
{
  if ( !GetVisited() )
    {
      CListIterInt iter;
      CSLIDEPoint *pCSLIDEPoint;
      
      BOOL bGeometryEmpty;
      BOOL bGeometryDynamic;
      BOOL bNormalsEmpty;
      BOOL bNormalsDynamic;
      BOOL bBoundsEmpty;
      BOOL bBoundsDynamic;

      BOOL bBoundEmpty;
      BOOL bBoundDynamic;

      bGeometryEmpty = TRUE;
      bGeometryDynamic = FALSE;
      bNormalsEmpty = TRUE;
      bNormalsDynamic = FALSE;
      bBoundsEmpty = TRUE;
      bBoundsDynamic = FALSE;

      bBoundEmpty = TRUE;
      bBoundDynamic = FALSE;

      //////////////////////////////////////////////////
      // Look at Points

      iter.Init(&m_listCSLIDEPoint);
      
      for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	    pCSLIDEPoint != NULL;
	    pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
	{
	  bBoundEmpty = FALSE;
	  bBoundDynamic = bBoundDynamic || 
	                  pCSLIDEPoint->GetBoundDynamic();
	  bGeometryDynamic = bGeometryDynamic || 
	                     pCSLIDEPoint->GetGeometryDynamic();
	  bNormalsDynamic = bNormalsDynamic ||
	                    pCSLIDEPoint->GetNormalDynamic();
	}

      bGeometryEmpty = bBoundEmpty;
      bNormalsEmpty = bBoundEmpty;
      bBoundsEmpty = bBoundEmpty;
      bBoundsDynamic = bBoundDynamic;

      //////////////////////////////////////////////////
      // Look at Local flags

      // Update whether we are Empty based on our LOD flag
      if ( !m_dyniLOD.GetDynamic() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      // This node is always off
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	      bBoundsEmpty = TRUE;
	      bBoundEmpty = TRUE;
	    }
	  else if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	    {
	      // This node is always just a bounding box
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	    }
	}

      // Check what is really dynamic

      bGeometryDynamic = ( bGeometryDynamic ||
			   m_pCSLIDESurface->GetDynamic() ||
			   m_dyniShading.GetDynamic() ||
			   m_dyniLOD.GetDynamic() ||
			   m_dyniSolid.GetDynamic() );
      bNormalsDynamic = ( bNormalsDynamic ||
			  m_dyniShading.GetDynamic() ||
			  m_dyniLOD.GetDynamic() );
      bBoundsDynamic = ( bBoundsDynamic || 
			 m_dyniLOD.GetDynamic() );
      bBoundDynamic = ( bBoundDynamic || 
			m_dyniLOD.GetDynamic() );

      // If anything is now Empty then it will always be empty so
      // it is not dynamic

      bGeometryDynamic = ( bGeometryDynamic && !bGeometryEmpty );
      bNormalsDynamic = ( bNormalsDynamic && !bNormalsEmpty );
      bBoundsDynamic = ( bBoundsDynamic && !bBoundsEmpty );
      bBoundDynamic = ( bBoundDynamic && !bBoundEmpty );
      
      //////////////////////////////////////////////////
      // For the OpenGL Display Lists to be really useful 
      // it is all or nothing so we enforce that here

      bGeometryDynamic = ( bGeometryDynamic ||
			   bNormalsDynamic ||
			   bBoundsDynamic );
      bNormalsDynamic = bGeometryDynamic;
      bBoundsDynamic = bGeometryDynamic;

      //////////////////////////////////////////////////
      // Now save these correct flag values

      SetGeometryEmpty(bGeometryEmpty);
      SetGeometryDynamic(bGeometryDynamic);
      SetNormalsEmpty(bNormalsEmpty);
      SetNormalsDynamic(bNormalsDynamic);
      SetBoundsEmpty(bBoundsEmpty);
      SetBoundsDynamic(bBoundsDynamic);
      SetBoundEmpty(bBoundEmpty);
      SetBoundDynamic(bBoundDynamic);
      
      //////////////////////////////////////////////////
      // Calculate the Bounding Box (if Empty this will do nothing)
      if ( GetBoundEmpty() )
	{
	  SetBoundValid(TRUE);
	}
      else
	{
	  CalculateBound();
	}
      SetBoundChanged(FALSE);

      //////////////////////////////////////////////////
      // Done Preproces(FLOAT)sing this Node in the Graph
      SetVisited(TRUE);
    }
}

VOID CSLIDEConvexHull::Update()
{
  SetBoundChanged(FALSE);
  if ( GetDynamic() ||
       GetBoundsDynamic() )
    {
      if ( GetBoundsDynamic() )
	{
	  CSLIDEBoundedNode::InvalidateQEMesh();
	}
      UpdateState();
    }

  if ( GetBoundDynamic() )
    {
      SetBoundValid(FALSE);
    }

  SetVisited(FALSE);
}

VOID CSLIDEConvexHull::UpdateState()
{
  UINT_32 i;

  CSLIDEBoundedNode::UpdateState();

  m_dyniSolid.Update();
  if ( !( 0 <= m_dyniSolid.GetValue() &&
	  m_dyniSolid.GetValue() < slidestSize ) )
    {
      HandleError("Illegal solid flag value:  %ld", 
		  m_dyniSolid.GetValue());
      m_dyniSolid.SetValue(slidestSLF_HOLLOW);
    }

  if ( m_pQEMesh == NULL )
    {
      CalculateQEMesh();
    }
}

VOID CSLIDEConvexHull::UpdateGraph()
{
  if ( !GetVisited() )
    {
      if ( !GetBoundValid() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      if ( !m_bound.GetEmpty() )
		{
		  SetBoundChanged(TRUE);
		  m_bound.Null();
		}
	      return;
	    }
	  
	  CalculateBound();
	  SetBoundValid(TRUE);
	}

      SetVisited(TRUE);
    }
}

VOID CSLIDEConvexHull::CalculateBound()
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;

  CBound bound;

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      pCSLIDEPoint->InsertIntoBound(&bound);
    }

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);
    }
}

#ifdef __USE_CDA__
VOID CSLIDEConvexHull::CalculateCDAVertices(CListInt *plistCDAVertex, 
					const Matrix &mGroot_obj)
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      pCSLIDEPoint->CalculateCDAVertices(plistCDAVertex, mGroot_obj);
    }
}
#endif // __USE_CDA__

VOID CSLIDEConvexHull::CalculateQEMesh()
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;
  CObjectList<QEVertex> listQEVertex;
  QEVertex *pvx;
  
  m_pQEMesh = AllocateQEMesh();
  ASSERT( m_pQEMesh != NULL );
  m_pQEMesh->Init();
  
  iter.Init(&m_listCSLIDEPoint);
  
  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      pvx = AllocateQEVertex();
      ASSERT( pvx != NULL );
      pvx->Init(pCSLIDEPoint->GetPoint());
      
      listQEVertex.InsertLast(pvx);
    }
  
  m_pQEMesh->CreateConvexHull(&listQEVertex);
}

VOID CSLIDEConvexHull::PrerenderGeometry(SLIDEShadingType eShading)
{
  SLIDEShadingType eShade;
  
  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  // This flag tells us whether to use Point or Point Normals
  ModifyShadingFlag(eShading);

  // This one set the shading attributes which in many cases will 
  // be SLF_INHERIT and so it will not set anything and leave up
  // to the caller above
  eShade = (SLIDEShadingType)m_dyniShading.GetValue();
  
  GLbitfield GLbAttributesMask;
  
  GLbAttributesMask = CalculateAttributesMask(eShade);
  
  PushAttributes(GLbAttributesMask);
  {
    PushShadingAttributes(eShade);
    {
      m_pCSLIDESurface->Push();
      {
	PushSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
	{
	  m_uiPrerenderAttributesStackDepthMax = 0;

	  ::RenderGeometry(m_pQEMesh);

	  if ( GLbAttributesMask != 0x0 )
	    {
	      m_uiPrerenderAttributesStackDepthMax++;
	    }
	}
	PopSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
      }
      m_pCSLIDESurface->Pop();
    }
    PopShadingAttributes(eShade);
  }
  PopAttributes(GLbAttributesMask);
}

VOID CSLIDEConvexHull::PrerenderNormals(SLIDEShadingType eShading)
{
  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  ModifyShadingFlag(eShading);

  ::RenderNormals(m_pQEMesh);
}

VOID CSLIDEConvexHull::RenderGeometry(const Matrix &/*mGproj_obj*/, 
				  CBound * /*pCBoundClip*/, 
				  SLIDEShadingType eShading,
				  SLIDELODType /* eLOD */ )
{
  PushSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
  {
    ::RenderGeometry(m_pQEMesh);
  }
  PopSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
}

VOID CSLIDEConvexHull::RenderNormals(SLIDEShadingType eShading)
{
  PushAuxillaryGeometryAttributes();
  {
    gilColor3f(1.0, 0.0, 1.0);

    ::RenderNormals(m_pQEMesh);
  }
  PopAuxillaryGeometryAttributes();
}

VOID CSLIDEConvexHull::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s ( ", 
	  gcppcSLIDEReservedWords[sliderwtConvexHull], m_pcGUIID);

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      fprintf(fpOutputFile, "%s ", pCSLIDEPoint->GetGUIID());      
    }

  fprintf(fpOutputFile, ")\n");

  WriteSLIDEFlags(fpOutputFile, 0);

  if ( m_dyniSolid.GetValue() != slidestSLF_SOLID )
    {
      fprintf(fpOutputFile, "  %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtSolid], 
	      gcppcSLIDESolidReservedWords[m_dyniSolid.GetValue()]);
    }

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndConvexHull]);
}

VOID CSLIDEConvexHull::WriteRIBGeometry(FILE *fpOutputFile,
				    const Matrix &/*mGproj_obj*/, 
				    CBound * /*pCBoundClip*/, 
				    SLIDEShadingType eShading,
				    SLIDELODType eLOD)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBPushSolidAttributes(fpOutputFile,
			      (SLIDESolidType)m_dyniSolid.GetValue());
  {
    WriteRIBGeometryPolygonal(fpOutputFile, eShading, eLOD);
  }
  WriteRIBPopSolidAttributes(fpOutputFile, 
			     (SLIDESolidType)m_dyniSolid.GetValue());
}

BOOL CSLIDEConvexHull::CheckRIBPrimitive()
{
  BOOL bResult;
  //UINT_32 i;

  bResult = FALSE;
/*
  for ( i = 0; i < (UINT_32)slideribotSize; i++ )
    {
      bResult = bResult || (m_pcGUIID == gcppcSLIDERIBConvexHullReservedWords[i]);  
    }
*/
  return bResult;
}

VOID CSLIDEConvexHull::WriteRIBGeometryPolygonal(FILE *fpOutputFile,
					     SLIDEShadingType eShading,
					     SLIDELODType /* eLOD */)
{
  //  CListIterInt iter;
  //CSLIDEPoint *pCSLIDEPoint;

  ASSERT( fpOutputFile != NULL );
/*
  iter.Init(&m_listCSLIDEPoint);
  
  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      pCSLIDEPoint->WriteRIB(fpOutputFile, eShading);
    }
*/
}

VOID CSLIDEConvexHull::WriteRIBGeometryPrimitive(FILE *fpOutputFile,
					     SLIDEShadingType /* eShading */,
					     SLIDELODType /* eLOD */)
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDEConvexHull::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE ConvexHull: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);
}

VOID CSLIDEConvexHull::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE ConvexHull: %s", m_pcGUIID);
}

VOID CSLIDEConvexHull::WritePOVGeometry(FILE *fpOutputFile, 
					CListInt *plistTransformIDs,
					CListInt *plistColors) 
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDEConvexHull::WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &m,
					 CSLIDESurface *pCSLIDESurface)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "Shape { # triangle mesh: CONVEXHULL\n\n");

  // Color Inheritance
  if( m_pCSLIDESurface == CSLIDESurface::GetSLF_INHERIT() )
    {
      pCSLIDESurface->WriteWRLZ(fpOutputFile);
    }
  else
    {
      ASSERT( m_pCSLIDESurface != NULL );
      m_pCSLIDESurface->WriteWRLZ(fpOutputFile);
    }

  // Mesh
  ASSERT(m_pQEMesh != NULL);
  ::WriteWRLZ(m_pQEMesh, fpOutputFile, m);

  fprintf(fpOutputFile, "} # end triangle mesh\n\n");
}

VOID CSLIDEConvexHull::WriteSTLGeometry(FILE *fpOutputFile, const Matrix &m)
{
  ASSERT( fpOutputFile != NULL );

  ::WriteSTL(m_pQEMesh, fpOutputFile, m);
}

VOID CSLIDEConvexHull::WriteDotID(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "ConvexHull_%s", GetGUIID());
}

VOID CSLIDEConvexHull::WriteDotNode(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "  ");
  WriteDotID(fpOutputFile);
  fprintf(fpOutputFile, " [ label = \"%s\", shape = ellipse, peripheries = 2", GetGUIID());
  if ( GetBoundDynamic() )
    {
      fprintf(fpOutputFile, ", color = red, fontcolor = red");
    }
  else if ( GetGeometryDynamic() )
    {
      fprintf(fpOutputFile, ", color = blue, fontcolor = blue");
    }
  fprintf(fpOutputFile, " ];\n");
}

VOID CSLIDEConvexHull::WriteDotEdges(FILE * /*fpOutputFile*/)
{
}

UINT_32 CSLIDEConvexHull::GetPointCount()
{
  return m_listCSLIDEPoint.GetLength();
}

