#include "slidecrosssection.h"
#include "slidesweepcrosssection.h"
#include "slidesweepmorph.h"
#include "slidebsplinecurve.h"
#include "slideworld.h"
#include "slidesurface.h"

extern BOOL gbCheckSemanticsDone;

static const UINT_32 cuiType = slidecstCircle;
static const FLOAT cfRadius = 1.0;
static const FLOAT cfInnerRadius = 0.75;
static const FLOAT cfThickness = 1.0;
static const FLOAT cfTheta = 90.0;
static const UINT_32 cuiSlices = 5;

static const UINT_32 cuiOrientation = 0;

//////////////////////////////////////////////////////////////////////
// CSLIDECrossSection Class
//

CSLIDECrossSection::CSLIDECrossSection()
{
  m_uiType = (SLIDECrossSectionType)cuiType;

  m_pcCrossSectionID = NULL;
  m_pCSLIDEPolyline = NULL;
  m_pCSLIDEBSplineCurve = NULL;

  m_dyniSlices.SetValue(cuiSlices);
  m_uiSlices = cuiSlices;
  m_uiPoints = cuiSlices + 1;

  m_uiOrientation = cuiOrientation;
  m_dynfRadius.SetValue(cfRadius);             // Default is circle
  m_dynfInnerRadius.SetValue(cfInnerRadius);
  m_dynfThickness.SetValue(cfThickness);
  m_dynfTheta.SetValue(cfTheta);

  m_pPoint = NULL;
  m_pvNormal = NULL;
  m_plistPointIDs = NULL;
  m_bCrossSectionChanged = FALSE;
}

CSLIDECrossSection::~CSLIDECrossSection()
{
}

VOID CSLIDECrossSection::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID)
{
  CSLIDEAtomNode::Init(pcFile, uiLine, uiColumn, pcID);

  m_uiType = (SLIDECrossSectionType)cuiType;

  m_pcCrossSectionID = NULL;
  m_pCSLIDEPolyline = NULL;
  m_pCSLIDEBSplineCurve = NULL;

  m_dyniSlices.SetValue(cuiSlices);
  m_uiSlices = cuiSlices;
  m_uiPoints = cuiSlices + 1;

  m_dynfRadius.SetValue(cfRadius);             // Default is circle
  m_dynfInnerRadius.SetValue(cfInnerRadius);
  m_dynfThickness.SetValue(cfThickness);
  m_dynfTheta.SetValue(cfTheta);

  // Initialize
  m_listCSLIDESweepCrossSectionParent.Init();
  m_listCSLIDESweepMorphParent.Init();
  m_listCSLIDEPoint.Init();

  ReallocatePoints();
  ReallocateNormals();

  CalculateCrossSection(cuiOrientation);
}

VOID CSLIDECrossSection::Uninit()
{
  CSLIDEAtomNode::Uninit();

  if( m_plistPointIDs != NULL )
    {
      delete m_plistPointIDs;
      m_plistPointIDs = NULL;
    }
  
  DeallocatePoints();
  DeallocateNormals();
}

BOOL CSLIDECrossSection::Initialized()
{
  if ( !CSLIDEAtomNode::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

VOID CSLIDECrossSection::InsertSweepCrossSectionParent(CSLIDESweepCrossSection *pCSLIDESweepCS)
{
  ASSERT( pCSLIDESweepCS != NULL );
  m_listCSLIDESweepCrossSectionParent.InsertLast(pCSLIDESweepCS, (UINT_32)pCSLIDESweepCS);
}

VOID CSLIDECrossSection::RemoveSweepCrossSectionParent(CSLIDESweepCrossSection *pCSLIDESweepCSToRemove)
{
  ASSERT( pCSLIDESweepCSToRemove != NULL );

  CListIterInt iter;
  CSLIDESweepCrossSection *pCSLIDESweepCS;
  CHAR *pcID;

  pcID = pCSLIDESweepCSToRemove->GetID();

  iter.Init(&m_listCSLIDESweepCrossSectionParent);
  for ( pCSLIDESweepCS = (CSLIDESweepCrossSection*)iter.PeekFirst();
	pCSLIDESweepCS != NULL; )
    {
      if( (pCSLIDESweepCS->GetID()) == pcID )
	{
	  pCSLIDESweepCS = (CSLIDESweepCrossSection*)iter.PeekNextRemoveCurr();
	}
      else
	{
	  pCSLIDESweepCS = (CSLIDESweepCrossSection*)iter.PeekNext();
	}
    }
}

VOID CSLIDECrossSection::InsertSweepMorphParent(CSLIDESweepMorph *pCSLIDESweepMorph)
{
  ASSERT( pCSLIDESweepMorph != NULL );
  m_listCSLIDESweepMorphParent.InsertLast(pCSLIDESweepMorph, (UINT_32)pCSLIDESweepMorph);
}

VOID CSLIDECrossSection::RemoveSweepMorphParent(CSLIDESweepMorph *pCSLIDESweepMorphToRemove)
{
  ASSERT( pCSLIDESweepMorphToRemove != NULL );

  CListIterInt iter;
  CSLIDESweepMorph *pCSLIDESweepMorph;
  CHAR *pcID;

  pcID = pCSLIDESweepMorphToRemove->GetID();

  iter.Init(&m_listCSLIDESweepMorphParent);
  for ( pCSLIDESweepMorph = (CSLIDESweepMorph*)iter.PeekFirst();
	pCSLIDESweepMorph != NULL; )
    {
      if( (pCSLIDESweepMorph->GetID()) == pcID )
	{
	  pCSLIDESweepMorph = (CSLIDESweepMorph*)iter.PeekNextRemoveCurr();
	}
      else
	{
	  pCSLIDESweepMorph = (CSLIDESweepMorph*)iter.PeekNext();
	}
    }
}

VOID CSLIDECrossSection::DeallocatePoints()
{
  if ( m_pPoint != NULL )
    {
      delete[] m_pPoint;
      m_pPoint = NULL;
    }
}

VOID CSLIDECrossSection::ReallocatePoints()
{
  if( m_pPoint != NULL )
    {
      DeallocatePoints();
    }

  m_pPoint = new Point[m_uiPoints];
  ASSERT(m_pPoint != NULL);
}

VOID CSLIDECrossSection::DeallocateNormals()
{
  if ( m_pvNormal != NULL )
    {
      delete[] m_pvNormal;
      m_pvNormal = NULL;
    }
}

VOID CSLIDECrossSection::ReallocateNormals()
{
  if( m_pvNormal != NULL )
    {
      DeallocateNormals();
    }
  
  m_pvNormal = new Vector[m_uiPoints];
  ASSERT(m_pvNormal != NULL);
}

BOOL CSLIDECrossSection::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  if ( !CSLIDEAtomNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  BOOL bCorrect;
  CListIterInt iter, iter2;

  bCorrect = TRUE;

  // Check Cross Section Reference: Polyline, Bezier, BSpline
  if( (m_uiType == slidecstPolyline) || 
      (m_uiType == slidecstBezierCurve) || 
      (m_uiType == slidecstBSplineCurve) )
    {
      if( m_pcCrossSectionID != NULL )
	{
	  if( m_uiType == slidecstPolyline )
	    {
	      m_pCSLIDEPolyline = 
		(CSLIDEPolyline *)pCSLIDEWorld->PeekNode(m_pcCrossSectionID);
	    }
	  else if(( m_uiType == slidecstBSplineCurve) || (m_uiType == slidecstBezierCurve) )
	    {
	      m_pCSLIDEBSplineCurve = 
		(CSLIDEBSplineCurve *)pCSLIDEWorld->PeekNode(m_pcCrossSectionID);
	    }
	  else
	    {
	      ASSERT(FALSE);
	    }
	
	  if( (m_pCSLIDEPolyline == NULL) && (m_pCSLIDEBSplineCurve == NULL) )
	    {
	      if(m_uiType == slidecstPolyline)
		{
		  HandleError("Undefined %s \"%s\"", 
			      gcppcSLIDEReservedWords[sliderwtPolyline],
			      m_pcCrossSectionID);
		}
	      else if( m_uiType == slidecstBSplineCurve)
		{
		  HandleError("Undefined %s \"%s\"", 
			      gcppcSLIDEReservedWords[sliderwtBSplineCurve],
			      m_pcCrossSectionID);
		}	      
	      else if( m_uiType == slidecstBezierCurve)
		{
		  HandleError("Undefined %s \"%s\"", 
			      gcppcSLIDEReservedWords[sliderwtBezierCurve],
			      m_pcCrossSectionID);
		}	      
	      bCorrect = FALSE;
	    }
	  else
	    {
	      if( !gbCheckSemanticsDone )
		{
		  if(m_uiType == slidecstPolyline)
		    {
		      m_pCSLIDEPolyline->InsertCrossSectionParent(this);
		    }
		  else if( (m_uiType == slidecstBSplineCurve) ||
			   (m_uiType == slidecstBezierCurve) )
		    {
		      m_pCSLIDEBSplineCurve->InsertCrossSectionParent(this);
		    }
		}
	    }
	}
      else
	{
	  HandleError("%s %s has no specified %s id", 
		      gcppcSLIDEReservedWords[sliderwtCrossSection],
		      m_pcGUIID,
		      gcppcSLIDEReservedWords[m_uiType]);
	  bCorrect = FALSE;
	}
    }

  if( (1 > m_uiSlices) )
    {
      HandleError("CrossSection slices must be >= 1" );
      bCorrect = FALSE;
    }

  return bCorrect;
}

BOOL CSLIDECrossSection::Valid()
{
  if ( !CSLIDEAtomNode::Valid() )
    {
      return FALSE;
    }

  return ( ( m_dyniSlices.GetValue() >= 0 ) &&
	   ( m_dynfRadius.GetValue() >= 0 ) );
}

VOID CSLIDECrossSection::CheckDynamic()
{
  CSLIDEAtomNode::CheckDynamic();

  if( !GetDynamic() )
    {
      SetDynamic(m_dyniSlices.GetDynamic() ||
		 m_dynfRadius.GetDynamic() ||
		 m_dynfInnerRadius.GetDynamic() ||
		 m_dynfThickness.GetDynamic() ||
		 m_dynfTheta.GetDynamic() ||
		 m_pCSLIDESurface->GetDynamic());
    }

  if( !GetDynamic() )
    {
      if( m_uiType == slidecstPolyline )
	{
	  ASSERT(m_pCSLIDEPolyline != NULL);
	  if( m_pCSLIDEPolyline->GetDynamic() )
	    {
	      SetDynamic(TRUE);
	    }
	}
      else if( (m_uiType == slidecstBSplineCurve) || (m_uiType == slidecstBezierCurve) )
	{
	  ASSERT(m_pCSLIDEBSplineCurve != NULL);
	  if( m_pCSLIDEBSplineCurve->GetDynamic() )
	    {
	      SetDynamic(TRUE);
	    }
	}
    }
}

VOID CSLIDECrossSection::PreprocessGraphLocal()
{
  CListIterInt iter;
  CSLIDESweepCrossSection *pCSLIDESweepCS;
  CSLIDESweepMorph *pCSLIDESweepMorph;
  BOOL bRealSpaceDynamic;
  BOOL bParameterSpaceDynamic;
  BOOL bTextureSpaceDynamic;
  BOOL bBoundDynamic;
  BOOL bGeometryDynamic;
  BOOL bNormalsDynamic;

  // These may not be instanced so call preprocessgraph
  if( m_uiType == slidecstPolyline )
    {
      ASSERT(m_pCSLIDEPolyline != NULL);
      m_pCSLIDEPolyline->PreprocessGraphLocal();
    }
  else if( (m_uiType == slidecstBSplineCurve) || (m_uiType == slidecstBezierCurve) )
    {
      ASSERT(m_pCSLIDEBSplineCurve != NULL);
      m_pCSLIDEBSplineCurve->PreprocessGraphLocal();
    }

  // If type = polyline or bsplinecurve then these will be set appropriately
  bBoundDynamic = GetBoundDynamic();
  bGeometryDynamic = GetGeometryDynamic();
  bNormalsDynamic = GetNormalsDynamic();


  bRealSpaceDynamic = FALSE;
  bTextureSpaceDynamic = FALSE;
  bParameterSpaceDynamic = ( m_dyniSlices.GetDynamic() ||
			     m_dynfRadius.GetDynamic() ||
			     m_dynfInnerRadius.GetDynamic() ||
			     m_dynfThickness.GetDynamic() ||
			     m_dynfTheta.GetDynamic() );

  SetGeometryEmpty(FALSE);
  SetGeometryDynamic(bRealSpaceDynamic || 
		     bParameterSpaceDynamic ||
		     bTextureSpaceDynamic ||
		     bGeometryDynamic ||
		     ((m_pCSLIDESurface != NULL) && m_pCSLIDESurface->GetDynamic()) );
  SetNormalsEmpty(FALSE);
  SetNormalsDynamic(bRealSpaceDynamic ||		    
		    bParameterSpaceDynamic ||
		    bNormalsDynamic );
  SetBoundsEmpty(FALSE);
  SetBoundsDynamic(bRealSpaceDynamic ||
		   bParameterSpaceDynamic ||
		   bBoundDynamic );
  SetBoundEmpty(FALSE);
  SetBoundDynamic(bRealSpaceDynamic ||
		  bParameterSpaceDynamic ||
		  bBoundDynamic );


  //////////////////////////////////////////////////////////////////////////////////////

  iter.Init(&m_listCSLIDESweepCrossSectionParent);  
  for ( pCSLIDESweepCS = (CSLIDESweepCrossSection*)iter.PeekFirst(); 
	pCSLIDESweepCS != NULL;
	pCSLIDESweepCS = (CSLIDESweepCrossSection*)iter.PeekNext() )
    {
      if( GetBoundDynamic() )
	pCSLIDESweepCS->SetBoundDynamic(TRUE);

      if( GetGeometryDynamic() )
	pCSLIDESweepCS->SetGeometryDynamic(TRUE);

      if( GetNormalsDynamic() )
	pCSLIDESweepCS->SetNormalsDynamic(TRUE);
    }

  iter.Init(&m_listCSLIDESweepMorphParent);
  for ( pCSLIDESweepMorph = (CSLIDESweepMorph*)iter.PeekFirst(); 
	pCSLIDESweepMorph != NULL;
	pCSLIDESweepMorph = (CSLIDESweepMorph*)iter.PeekNext() )
    {
      if( GetBoundDynamic() )
	pCSLIDESweepMorph->SetBoundDynamic(TRUE);

      if( GetGeometryDynamic() )
	pCSLIDESweepMorph->SetGeometryDynamic(TRUE);

      if( GetNormalsDynamic() )
	pCSLIDESweepMorph->SetNormalsDynamic(TRUE);
    }

  if( m_listCSLIDESweepMorphParent.GetLength() > 0 )
    {
      if( m_uiType == slidecstPolyline )
	{
	  ASSERT(m_pCSLIDEPolyline != NULL);

	  if( GetBoundDynamic() )
	    m_pCSLIDEPolyline->SetBoundDynamic(TRUE);
	  
	  if( GetGeometryDynamic() )
	    m_pCSLIDEPolyline->SetGeometryDynamic(TRUE);
	  
	  if( GetNormalsDynamic() )
	    m_pCSLIDEPolyline->SetNormalsDynamic(TRUE);
	}
      else if( (m_uiType == slidecstBSplineCurve) || (m_uiType == slidecstBezierCurve) )
	{
	  ASSERT(m_pCSLIDEBSplineCurve != NULL);

	  if( GetBoundDynamic() )
	    m_pCSLIDEBSplineCurve->SetBoundDynamic(TRUE);
	  
	  if( GetGeometryDynamic() )
	    m_pCSLIDEBSplineCurve->SetGeometryDynamic(TRUE);
	  
	  if( GetNormalsDynamic() )
	    m_pCSLIDEBSplineCurve->SetNormalsDynamic(TRUE);
	}
    }
}

VOID CSLIDECrossSection::UpdateSweepCrossSectionParents()
{
  CListIterInt iter;
  CSLIDESweepCrossSection *pCSLIDESweepCS;
  CSLIDESweepMorph *pCSLIDESweepMorph;

  iter.Init(&m_listCSLIDESweepCrossSectionParent);
  for ( pCSLIDESweepCS = (CSLIDESweepCrossSection*)iter.PeekFirst(); 
	pCSLIDESweepCS != NULL;
	pCSLIDESweepCS = (CSLIDESweepCrossSection*)iter.PeekNext() )
    {
      pCSLIDESweepCS->CalculateEndCaps();
    }

  iter.Init(&m_listCSLIDESweepMorphParent);
  for ( pCSLIDESweepMorph = (CSLIDESweepMorph*)iter.PeekFirst(); 
	pCSLIDESweepMorph != NULL;
	pCSLIDESweepMorph = (CSLIDESweepMorph*)iter.PeekNext() )
    {
      pCSLIDESweepMorph->UpdateCrossSectionPoints();
    }
}

VOID CSLIDECrossSection::UpdateCrossSectionPoints(UINT_32 uiPoints, Point *pPoints)
{
  // Called by polyline and bsplinecurve when they change

  ASSERT( pPoints != NULL );
  UINT_32 i;
  
  m_uiPoints = uiPoints;
  m_uiSlices = m_uiPoints-1;

  ReallocatePoints();
  ReallocateNormals();

  // Copy the points in reverse order to maintain correct CrossSection orientation
  for( i = 0; i < uiPoints; i++ )
    {
      m_pPoint[uiPoints-1-i] = pPoints[i];
    }

  if( GetDynamic() )
    {
      m_bCrossSectionChanged = TRUE;
    }
  else
    {
      CalculateCrossSection(cuiOrientation);
      UpdateSweepCrossSectionParents();
    }
}

VOID CSLIDECrossSection::UpdateState()
{
  CSLIDEAtomNode::UpdateState();

  Point pt;
  FLOAT f;
  BOOL bCrossSectionChanged, bCrossSectionSlicesChanged;

  bCrossSectionChanged = m_bCrossSectionChanged;
  bCrossSectionSlicesChanged = FALSE;

  if( (m_uiType == slidecstPolyline) ||
      (m_uiType == slidecstBSplineCurve) ||
      (m_uiType == slidecstBezierCurve) ) // Polyline, BSpline, Bezier
    {
      // This is updated by the curve which calls UpdateCrossSectionPoints
    }
  else if( m_uiType == slidecstCircle ) // Circle
    {
      f = m_dynfRadius.GetValue();
      m_dynfRadius.Update();
      if ( m_dynfRadius.GetValue() < 0.0 )
	{
	  HandleError("CrossSection radius must be between >= 0.0" );
	  m_dynfRadius.SetValue(cfRadius);
	}
      bCrossSectionChanged = ((f != m_dynfRadius.GetValue()) || bCrossSectionChanged);
      
      m_dyniSlices.Update();
      if ((3 >  m_dyniSlices.GetValue()) )
	{
	  HandleError("CrossSection circle crosssection slices must be >= 3" );
	  m_dyniSlices.SetValue(cuiSlices);
	}
      bCrossSectionSlicesChanged = (m_bCrossSectionChanged ||
	    (m_uiSlices != (UINT_32)m_dyniSlices.GetValue()) );

      m_uiSlices = m_dyniSlices.GetValue();
      m_uiPoints = m_uiSlices+1;
    }
  else if( m_uiType == slidecstCresent ) // Cresent
    {
      f = m_dynfRadius.GetValue();
      m_dynfRadius.Update();
      if ( m_dynfRadius.GetValue() < 0.0 )
	{
	  HandleError("CrossSection radius must be >= 0.0" );
	  m_dynfRadius.SetValue(cfRadius);
	}
      bCrossSectionChanged = ((f != m_dynfRadius.GetValue()) || bCrossSectionChanged);

      f = m_dynfThickness.GetValue();
      m_dynfThickness.Update();
      if ( m_dynfThickness.GetValue() < 0.0 )
	{
	  HandleError("CrossSection thickness must be >= 0.0" );
	  m_dynfThickness.SetValue(cfThickness);
	}
      bCrossSectionChanged = ((f != m_dynfThickness.GetValue()) || bCrossSectionChanged);

      f = m_dynfTheta.GetValue();
      m_dynfTheta.Update();
      if ( m_dynfTheta.GetValue() < 0.0 )
	{
	  HandleError("CrossSection theta must be between >= 0.0" );
	  m_dynfTheta.SetValue(cfTheta);
	}
      bCrossSectionChanged = ((f != m_dynfTheta.GetValue()) || bCrossSectionChanged);
      
      m_dyniSlices.Update();
      if ((1 >  m_dyniSlices.GetValue()) )
	{
	  HandleError("CrossSection cresent crosssection slices must be >= 1" );
	  m_dyniSlices.SetValue(cuiSlices);
	}
      bCrossSectionSlicesChanged = (m_bCrossSectionChanged ||
		 (m_uiSlices != (UINT_32)m_dyniSlices.GetValue()));

      m_uiSlices = m_dyniSlices.GetValue();
      m_uiPoints = 2 * (m_uiSlices+1) + 1;
    }
  else if( m_uiType == slidecstStar ) // Star
    {
      f = m_dynfRadius.GetValue();
      m_dynfRadius.Update();
      if ( m_dynfRadius.GetValue() < 0.0 )
	{
	  HandleError("CrossSection radius must be between >= 0.0" );
	  m_dynfRadius.SetValue(cfRadius);
	}
      bCrossSectionChanged = ((f != m_dynfRadius.GetValue()) || bCrossSectionChanged);

      f = m_dynfInnerRadius.GetValue();
      m_dynfInnerRadius.Update();
      if ( m_dynfInnerRadius.GetValue() < 0.0 )
	{
	  HandleError("CrossSection innerradius must be between >= 0.0" );
	  m_dynfInnerRadius.SetValue(cfInnerRadius);
	}
      bCrossSectionChanged = ((f != m_dynfInnerRadius.GetValue()) || bCrossSectionChanged);

      m_dyniSlices.Update();
      if ((2 >  m_dyniSlices.GetValue()) )
	{
	  HandleError("CrossSection star crosssection slices must be >= 2" );
	  m_dyniSlices.SetValue(cuiSlices);
	}
      bCrossSectionSlicesChanged = (m_bCrossSectionChanged ||
				    (m_uiSlices != (UINT_32)m_dyniSlices.GetValue()));

      m_uiSlices = m_dyniSlices.GetValue();
      m_uiPoints = 2*m_uiSlices + 1;
    }

  // Check to see if the cross section geometry has to be recalculated
  if( bCrossSectionSlicesChanged )
    {
      ReallocatePoints();
      ReallocateNormals();

      CalculateCrossSection(cuiOrientation);
      UpdateSweepCrossSectionParents();
    }
  else if( bCrossSectionChanged ) 
    {
      CalculateCrossSection(cuiOrientation);
      UpdateSweepCrossSectionParents();
    }

  m_bCrossSectionChanged = FALSE;
}

VOID CSLIDECrossSection::CalculateBound()
{
  UINT_32 uiSlice;
  Point pt;
  CBound bound;

 for( uiSlice = 0; uiSlice < m_uiPoints; uiSlice++)
   {
     pt = m_pPoint[uiSlice];
     bound.Insert(pt);
   }

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);
    }
}

////////////////////////////////////////////////////////////////////////////////////////////

VOID CSLIDECrossSection::CalculateCrossSection(UINT_32 uiOrientation)
{
  UINT_32 uiCSSlice;

  if( (m_uiType == slidecstPolyline) ||   
      (m_uiType == slidecstBSplineCurve) ||   
      (m_uiType == slidecstBezierCurve) )   // Polyline, Bezier, BSpline
    {
      UINT_32 i, uiSlice;
      Vector vPrev, vPost, v;

      if( m_uiOrientation != uiOrientation )
	{
	  Point *pPoint;
	  pPoint = new Point[m_uiPoints];

	  // Copy in reverse order
	  for( i = 0; i < m_uiPoints; i++ )
	    {
	      pPoint[m_uiPoints-1-i] = m_pPoint[i];
	    }

	  for( i = 0; i < m_uiPoints; i++ )
	    {
	      m_pPoint[i] = pPoint[i];
	    }
	  delete[] pPoint;
	}

      // Calculate Smooth normals

      // End Points      
      if( m_pPoint[0] == m_pPoint[m_uiPoints-1] ) // closed
	{
	  // End Points (first(0) and last(m_uiSlices) point are the same)
	  v = m_pPoint[0] - m_pPoint[m_uiPoints-2];
	  vPrev.Set( -v[Y], v[X], 0, 0 );
	  vPrev.Normalize();
	  
	  v = m_pPoint[1] - m_pPoint[0];
	  vPost.Set( -v[Y], v[X], 0, 0 );
	  vPost.Normalize();
	  
	  v = (vPrev + vPost) / 2.0;
	  v.Normalize();
	  m_pvNormal[0] = v;
	  m_pvNormal[m_uiPoints-1] = v;
	}
      else // not closed
	{
	  // First
	  v = m_pPoint[0] - m_pPoint[m_uiPoints-1];
	  vPrev.Set( -v[Y], v[X], 0, 0 );
	  vPrev.Normalize();
	  
	  v = m_pPoint[1] - m_pPoint[0];
	  vPost.Set( -v[Y], v[X], 0, 0 );
	  vPost.Normalize();
	  
	  v = (vPrev + vPost) / 2.0;
	  v.Normalize();
	  m_pvNormal[0] = v;

	  // Last
	  v = m_pPoint[m_uiPoints-1] - m_pPoint[m_uiPoints-2];
	  vPrev.Set( -v[Y], v[X], 0, 0 );
	  vPrev.Normalize();
	  
	  v = m_pPoint[0] - m_pPoint[m_uiPoints-1];
	  vPost.Set( -v[Y], v[X], 0, 0 );
	  vPost.Normalize();
	  
	  v = (vPrev + vPost) / 2.0;
	  v.Normalize();
	  m_pvNormal[m_uiPoints-1] = v;
	}

      for( uiSlice = 1; uiSlice < m_uiPoints-1; uiSlice++ )
	{
	  v = m_pPoint[uiSlice] - m_pPoint[uiSlice-1];
	  vPrev.Set( -v[Y], v[X], 0, 0 );
	  vPrev.Normalize();

	  v = m_pPoint[uiSlice+1] - m_pPoint[uiSlice];
	  vPost.Set( -v[Y], v[X], 0, 0 );
	  vPost.Normalize();	  

	  v = (vPrev + vPost) / 2.0;
	  v.Normalize();	  
	  m_pvNormal[uiSlice] = v;
	}
    }
  else if( m_uiType == slidecstCircle ) // Circle
    {
      FLOAT fTheta, fThetaDelta, fRadius;
      Vector v;
      
      // Construct the cross section at the world origin     
      fRadius = m_dynfRadius.GetValue();
      fThetaDelta = -2.0*PI/(FLOAT)m_uiSlices;

      if( uiOrientation == 1 )
	{
	  fThetaDelta *= -1;
	}
      
      fTheta = 0.0;
      for( uiCSSlice = 0; uiCSSlice < m_uiSlices+1; uiCSSlice++)
	{
	  m_pPoint[uiCSSlice].Set(fRadius*cos(fTheta), fRadius*sin(fTheta),0.0);
	  v.Set( cos(fTheta), sin(fTheta), 0.0);
	  v.Normalize();
	  if( uiOrientation == 1 )
	    {
	      v = -v;
	    }
	  m_pvNormal[uiCSSlice].Set(v.Get());
	  
	  fTheta += fThetaDelta;
	}
    }
  else if( m_uiType == slidecstCresent ) // Cresent
    {
      UINT_32 uiCSOther;
      FLOAT fThetaDelta, fRadius, fThickness, fTheta, fX1, fX2, fY;
      Vector v;
      
      // Construct the cross section at the world origin     
      fRadius = m_dynfRadius.GetValue();
      fThickness = m_dynfThickness.GetValue();
      fTheta = -DEGTORAD(m_dynfTheta.GetValue()) / 2.0;      

      fThetaDelta = -2.0*fTheta/(FLOAT)m_uiSlices;

      if( uiOrientation == 0 )
	{
	  for( uiCSSlice = 0; uiCSSlice < m_uiSlices+1; uiCSSlice++)
	    {	  
	      uiCSOther = m_uiPoints-uiCSSlice-2;
	      
	      fX1 = fRadius*cos(fTheta)-fRadius-fThickness/2.0;
	      fX2 = fRadius*cos(fTheta)-fRadius+fThickness/2.0;
	      fY  = fRadius*sin(fTheta);
	      
	      m_pPoint[uiCSSlice].Set( fX1, fY, 0.0 );
	      m_pPoint[uiCSOther].Set( fX2, fY, 0.0 );
	      
	      v.Set( -cos(fTheta), -sin(fTheta), 0.0 );
	      v.Normalize();
	      if( uiOrientation == 1 )
		{
		  v = -v;
		}
	      m_pvNormal[uiCSSlice].Set(v.Get());


	      v.Set( cos(fTheta), sin(fTheta), 0.0 );
	      v.Normalize();
	      if( uiOrientation == 1 )
		{
		  v = -v;
		}
	      m_pvNormal[uiCSOther].Set(v.Get());
	      
	      fTheta += fThetaDelta;
	    }
	}
      else  // inward orientation
	{
	  for( uiCSSlice = 0; uiCSSlice < m_uiSlices+1; uiCSSlice++)
	    {	  
	      uiCSOther = m_uiPoints-uiCSSlice-2;
	      
	      fX1 = fRadius*cos(fTheta)-fRadius-fThickness/2.0;
	      fX2 = fRadius*cos(fTheta)-fRadius+fThickness/2.0;
	      fY  = fRadius*sin(fTheta);
	      
	      m_pPoint[uiCSOther].Set( fX1, fY, 0.0 );
	      m_pPoint[uiCSSlice].Set( fX2, fY, 0.0 );
	      
	      v.Set( -cos(fTheta), -sin(fTheta), 0.0 );
	      v.Normalize();
	      if( uiOrientation == 1 )
		{
		  v = -v;
		}
	      m_pvNormal[uiCSOther].Set(v.Get());

	      v.Set( cos(fTheta), sin(fTheta), 0.0 );
	      v.Normalize();
	      if( uiOrientation == 1 )
		{
		  v = -v;
		}
	      m_pvNormal[uiCSSlice].Set(v.Get());
	      
	      fTheta += fThetaDelta;
	    }
	}
      m_pPoint[m_uiPoints-1] = m_pPoint[0];
      m_pvNormal[m_uiPoints-1] = m_pvNormal[0];
    }
  else if( m_uiType == slidecstStar ) // Star
    {
      FLOAT fTheta, fThetaDelta, fRadius, fInnerRadius;
      Vector v;
      
      // Construct the cross section at the world origin     
      fRadius = m_dynfRadius.GetValue();
      fInnerRadius = m_dynfInnerRadius.GetValue();

      fThetaDelta = -PI/(FLOAT)m_uiSlices;
      if( uiOrientation == 1 )
	{
	  fThetaDelta *= -1;
	}
      
      fTheta = 0.0;
      for( uiCSSlice = 0; uiCSSlice < m_uiSlices; uiCSSlice++)
	{
	  // Inner point
	  m_pPoint[2*uiCSSlice+1].Set(fInnerRadius*cos(fTheta), 
						fInnerRadius*sin(fTheta), 0.0);
	  v.Set( cos(fTheta), sin(fTheta), 0.0);
	  v.Normalize();
	  if( uiOrientation == 1 )
	    {
	      v = -v;
	    }
	  m_pvNormal[2*uiCSSlice+1].Set(v.Get());
	  fTheta += fThetaDelta;

	  // Outer point
	  m_pPoint[2*uiCSSlice+2].Set(fRadius*cos(fTheta), 
						  fRadius*sin(fTheta), 0.0);
	  v.Set( cos(fTheta), sin(fTheta), 0.0);
	  v.Normalize();
	  if( uiOrientation == 1 )
	    {
	      v = -v;
	    }
	  m_pvNormal[2*uiCSSlice+2].Set(v.Get());
	  fTheta += fThetaDelta;
	}

      m_pPoint[0] = m_pPoint[m_uiPoints-1];
      m_pvNormal[0] = m_pvNormal[m_uiPoints-1];
    }

  m_uiOrientation = uiOrientation; // save current orientation 
}

////////////////////////////////////////////////////////////////////////////////////////////

VOID CSLIDECrossSection::RenderGeometrySmooth(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{  
  RenderGeometryFlat(pCSLIDEWarp);
}

VOID CSLIDECrossSection::RenderGeometryFlat(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{
  UINT_32 uiSlice;
  Point pt;
  
  gilBegin(GL_LINE_STRIP);
  {
    for( uiSlice = 0; uiSlice < m_uiPoints; uiSlice++)
      {
	pt = m_pPoint[uiSlice];
	
	gilVertex3fv( pt.Get() );
      }
  }
  gilEnd();
}

VOID CSLIDECrossSection::WriteSLIDE(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtCrossSection], m_pcGUIID);
  WriteSLIDEFlags(fpOutputFile, 0);

  if( m_uiType == slidecstPolyline )
    {
      fprintf(fpOutputFile, "  %s %s %s\n", 
	      gcppcSLIDEReservedWords[sliderwtType], 
	      gcppcSLIDEReservedWords[sliderwtPolyline], 
	      m_pcCrossSectionID);
    }
  else if( m_uiType == slidecstBSplineCurve )
    {
      fprintf(fpOutputFile, "  %s %s %s\n", 
	      gcppcSLIDEReservedWords[sliderwtType], 
	      gcppcSLIDEReservedWords[sliderwtBSplineCurve], 
	      m_pcCrossSectionID);
    }
  else if( m_uiType == slidecstBezierCurve )
    {
      fprintf(fpOutputFile, "  %s %s %s\n", 
	      gcppcSLIDEReservedWords[sliderwtType], 
	      gcppcSLIDEReservedWords[sliderwtBezierCurve], 
	      m_pcCrossSectionID);
    }
  else if( m_uiType == slidecstCircle ) // Circle
    {
      fprintf(fpOutputFile, "  %s %s\n", 
	      gcppcSLIDEReservedWords[sliderwtType],
	      gcppcSLIDEReservedWords[sliderwtCircle]);

      fprintf(fpOutputFile, "  %s %.3f\n",
	      gcppcSLIDEReservedWords[sliderwtRadius],
	      m_dynfRadius.GetValue());
      fprintf(fpOutputFile, "  %s %ld\n",
	      gcppcSLIDEReservedWords[sliderwtSlices],
	      m_dyniSlices.GetValue());
    }
  else if( m_uiType == slidecstCresent ) // Cresent
    {
      fprintf(fpOutputFile, "  %s %s\n", 
	      gcppcSLIDEReservedWords[sliderwtType],
	      gcppcSLIDEReservedWords[sliderwtCresent]);

      fprintf(fpOutputFile, "  %s %.3f\n",
	      gcppcSLIDEReservedWords[sliderwtRadius],
	      m_dynfRadius.GetValue());
      fprintf(fpOutputFile, "  %s %.3f\n",
	      gcppcSLIDEReservedWords[sliderwtThickness],
	      m_dynfThickness.GetValue());
      fprintf(fpOutputFile, "  %s %.3f\n",
	      gcppcSLIDEReservedWords[sliderwtTheta],
	      m_dynfTheta.GetValue());
      fprintf(fpOutputFile, "  %s %ld\n",
	      gcppcSLIDEReservedWords[sliderwtSlices],
	      m_dyniSlices.GetValue());
    }
  else if( m_uiType == slidecstStar ) // Star
    {
      fprintf(fpOutputFile, "  %s %s\n", 
	      gcppcSLIDEReservedWords[sliderwtType],
	      gcppcSLIDEReservedWords[sliderwtStar]);

      fprintf(fpOutputFile, "  %s %.3f\n",
	      gcppcSLIDEReservedWords[sliderwtOuterRadius],
	      m_dynfRadius.GetValue());
      fprintf(fpOutputFile, "  %s %.3f\n",
	      gcppcSLIDEReservedWords[sliderwtInnerRadius],
	      m_dynfInnerRadius.GetValue());
      fprintf(fpOutputFile, "  %s %ld\n",
	      gcppcSLIDEReservedWords[sliderwtSlices],
	      m_dyniSlices.GetValue());
    }

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndCrossSection]);
}

VOID CSLIDECrossSection::WriteRIBGeometry(FILE */*fpOutputFile*/,
					  const Matrix &mGproj_obj,
					  CBound *pCBoundClip, 
					  SLIDEShadingType eShading,
					  SLIDELODType eLOD)
{
}

VOID CSLIDECrossSection::WriteRIBPushSLIDEName(FILE */*fpOutputFile*/)
{
}

VOID CSLIDECrossSection::WriteRIBPopSLIDEName(FILE */*fpOutputFile*/)
{
}






