#ifndef __SLIDECROSSSECTION_H__
#define __SLIDECROSSSECTION_H__

#include "slideatomnode.h"
#include "slidepolyline.h"
#include "slidebsplinecurve.h"

class CSLIDESweepCrossSection;

enum SLIDECrossSectionType
{
  slidecstPolyline,
  slidecstBSplineCurve,
  slidecstBezierCurve,
  slidecstCircle,
  slidecstCresent,
  slidecstStar,
  slidecstSize
};

//////////////////////////////////////////////////////////////////////////////
// CSLIDECrossSection Class
//

class CSLIDECrossSection : public CSLIDEAtomNode
{
//////////////////////////////////////////////////
// Friends
//
public: 
  friend class CSLIDESweepCrossSection;
  friend class CSLIDEControlPoint;

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDECrossSection();
  virtual ~CSLIDECrossSection();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID Uninit();
  virtual BOOL Initialized();
  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);
  virtual BOOL Valid();

  inline VOID SetType(UINT_32 uiType);
  inline VOID SetID(CHAR *pcID);
  inline VOID SetRadius(DynFloat *pdynfRadius);
  inline VOID SetInnerRadius(DynFloat *pdynfInnerRadius);
  inline VOID SetThickness(DynFloat *pdynfThickness);
  inline VOID SetTheta(DynFloat *pdynfTheta);
  inline VOID SetSlices(DynInt *pdyniCrossSectionSlices);

  VOID InsertSweepCrossSectionParent(CSLIDESweepCrossSection *pCSLIDESweepCS);
  VOID RemoveSweepCrossSectionParent(CSLIDESweepCrossSection *pCSLIDESweepCSToRemove);

  VOID InsertSweepMorphParent(CSLIDESweepMorph *pCSLIDESweepMorph);
  VOID RemoveSweepMorphParent(CSLIDESweepMorph *pCSLIDESweepMorphToRemove);

protected:
  virtual VOID ReallocatePoints();
  virtual VOID DeallocatePoints();
  virtual VOID ReallocateNormals();
  virtual VOID DeallocateNormals();

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  VOID UpdateSweepCrossSectionParents();
  VOID UpdateCrossSectionPoints(UINT_32 uiPoints, Point *pPoints);

protected:
  virtual VOID CheckDynamic();
  virtual VOID PreprocessGraphLocal();
  virtual VOID UpdateState();
  virtual VOID CalculateBound();

protected:
  virtual VOID CalculateCrossSection(UINT_32 uiOrientation);

//////////////////////////////////////////////////
// Rendering
//
public:

protected:
  virtual VOID RenderGeometrySmooth(CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderGeometryFlat(CSLIDEWormwarp *pCSLIDEWarp = NULL);

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline BOOL GetChildrenBoundsDynamic();
  inline VOID SetChildrenBoundsDynamic(BOOL bChildrenBoundsDynamic);

  inline UINT_32 GetPoints();
  inline Point* GetPoint();
  inline Vector* GetNormal();
  
  inline UINT_32 GetEndCapPoints();
  inline CSLIDESurface *GetSurface();

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:

  // Sweep
  CListInt m_listCSLIDESweepCrossSectionParent;
  CListInt m_listCSLIDESweepMorphParent;

  // Cross Section  [polyline, bsplinecurve, beziercurve, circle, cresent, star]
  SLIDECrossSectionType m_uiType;  

  CHAR *m_pcCrossSectionID;               
  CSLIDEPolyline *m_pCSLIDEPolyline;
  CSLIDEBSplineCurve *m_pCSLIDEBSplineCurve;

  UINT_32 m_uiOrientation;     // 0 = outward, 1 = inward
  DynFloat m_dynfRadius;       // = outerradius
  DynFloat m_dynfInnerRadius;
  DynFloat m_dynfThickness;
  DynFloat m_dynfTheta;
  DynInt m_dyniSlices;

  UINT_32 m_uiSlices;  // User parameter
  UINT_32 m_uiPoints;  // Actual number of points allocated

  Point *m_pPoint;
  Vector *m_pvNormal;
  CListInt *m_plistPointIDs;
  CListInt m_listCSLIDEPoint;
  BOOL m_bCrossSectionChanged;

  BOOL m_bChildrenBoundsDynamic;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDECrossSection::SetType(UINT_32 uiType)
{
  m_uiType = (SLIDECrossSectionType)uiType;
  m_bCrossSectionChanged = TRUE;
}

inline VOID CSLIDECrossSection::SetID(CHAR *pcID)
{
  m_pcCrossSectionID = pcID;
}

inline VOID CSLIDECrossSection::SetRadius(DynFloat *pdynfRadius)
{
  m_dynfRadius.Set(pdynfRadius);
}

inline VOID CSLIDECrossSection::SetInnerRadius(DynFloat *pdynfInnerRadius)
{
  m_dynfInnerRadius.Set(pdynfInnerRadius);
}

inline VOID CSLIDECrossSection::SetThickness(DynFloat *pdynfThickness)
{
  m_dynfThickness.Set(pdynfThickness);
}

inline VOID CSLIDECrossSection::SetTheta(DynFloat *pdynfTheta)
{
  m_dynfTheta.Set(pdynfTheta);
}

inline VOID CSLIDECrossSection::SetSlices(DynInt *pdyniSlices)
{
  m_dyniSlices.Set(pdyniSlices);
}

inline BOOL CSLIDECrossSection::GetChildrenBoundsDynamic()
{
  return m_bChildrenBoundsDynamic;
}

inline VOID CSLIDECrossSection::SetChildrenBoundsDynamic(BOOL bChildrenBoundsDynamic)
{
  m_bChildrenBoundsDynamic = bChildrenBoundsDynamic;
}

inline UINT_32 CSLIDECrossSection::GetPoints()
{
  return m_uiPoints;
}

inline Point* CSLIDECrossSection::GetPoint()
{
  return m_pPoint;
}

inline Vector* CSLIDECrossSection::GetNormal()
{
  return m_pvNormal;
}

inline UINT_32 CSLIDECrossSection::GetEndCapPoints()
{
  if( ((m_uiType == slidecstPolyline) || 
       (m_uiType == slidecstBSplineCurve) || 
       (m_uiType == slidecstBezierCurve) ) &&
      ( m_pPoint[m_uiPoints-1] != m_pPoint[0]) )
    {
      return m_uiPoints;
    }
  else
    {
      return m_uiPoints-1;
    }
}

inline CSLIDESurface* CSLIDECrossSection::GetSurface()
{
  return m_pCSLIDESurface;
}

#endif // __SLIDECROSSSECTION_H__
