#ifndef __SLIDECYCLIDE_H__
#define __SLIDECYCLIDE_H__

#include "slideatomnode.h"

//////////////////////////////////////////////////////////////////////////////
// CSLIDECyclide Class
//

class CSLIDECyclide : public CSLIDEAtomNode
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDECyclide();
  virtual ~CSLIDECyclide();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  inline VOID SetXMajorA(DynFloat *pdynfXMajorA);
  inline VOID SetYMinorB(DynFloat *pdynfYMinorB);
  inline VOID SetStringLength(DynFloat *pdynfStringLength);
  inline VOID SetPhiMin(DynFloat *pdynfPhiMin);
  inline VOID SetPhiMax(DynFloat *pdynfPhiMax);
  inline VOID SetThetaMin(DynFloat *pdynfThetaMin);
  inline VOID SetThetaMax(DynFloat *pdynfThetaMax);
  inline VOID SetPhiSlices(DynInt *pdyniPhiSlices);
  inline VOID SetThetaSlices(DynInt *pdyniThetaSlices);
  inline VOID SetDrawEllipse(DynInt *pdyniDrawEllipse);
  inline VOID SetDrawHyperbola(DynInt *pdyniDrawHyperbola);
  inline VOID SetTexture(DynVector *pdynv0, DynVector *pdynv1, DynVector *pdynv2, DynVector *pdynv3);
    
  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Geometric Computations
//
public:

#ifdef __USE_CDA__
  virtual VOID CalculateCDAVertices(CListInt *plistCDAVertex, 
				    const Matrix &mGroot_obj);
#endif // __USE_CDA__

protected:
  virtual VOID CheckDynamic();
  virtual VOID PreprocessGraphLocal();
  virtual VOID UpdateState();
  virtual VOID CalculateBound();
  virtual VOID CalculateQEMesh();
  virtual VOID CalculateLeafQEMesh();  // for WriteWRLZ
  virtual VOID DeallocateLeafQEMesh();  // for WriteWRLZ

//////////////////////////////////////////////////
// Rendering
//
public:

protected:
  VOID CalculateZ(FLOAT fTheta, FLOAT fPhi, Point &pZ, Vector &vPQ );

  virtual VOID RenderGeometrySmooth(CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderGeometryFlat(CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderNormalsSmooth();
  virtual VOID RenderNormalsFlat();

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);

  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

  virtual VOID WritePOVGeometry(FILE *fpOutputFile,
				CListInt *plistTransformIDs,
				CListInt *plistColors);  

  virtual VOID WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				 CSLIDESurface *pCSLIDESurface);
  virtual VOID WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline CHAR *GetTypeString();
  virtual inline SLIDEReservedWordType GetType();

  virtual VOID GetSide(CListInt *plistPoint, CListInt *plistNormal, UINT_32 uiSide, 
		       const Matrix &mTransform, const Matrix &mInverse, BOOL bReverse);

  virtual VOID GetParameters(Tcl_Interp *pInterp);
  virtual UINT_32 SetParameter(CSLIDEWorld *pCSLIDEWorld,
			       SLIDEReservedWordType uiParameter, CHAR **argv);

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  DynFloat m_dynfXMajorA;
  DynFloat m_dynfYMinorB;
  DynFloat m_dynfStringLength;
  DynFloat m_dynfPhiMin;
  DynFloat m_dynfPhiMax;
  DynFloat m_dynfThetaMin;
  DynFloat m_dynfThetaMax;

  DynInt m_dyniPhiSlices;
  DynInt m_dyniThetaSlices;

  DynInt m_dyniDrawEllipse;
  DynInt m_dyniDrawHyperbola;

  DynVector m_pvTexture[4];
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDECyclide::SetXMajorA(DynFloat *pdynfXMajorA)
{
  m_dynfXMajorA.Set(pdynfXMajorA);
}

inline VOID CSLIDECyclide::SetYMinorB(DynFloat *pdynfYMinorB)
{
  m_dynfYMinorB.Set(pdynfYMinorB);
}

inline VOID CSLIDECyclide::SetStringLength(DynFloat *pdynfStringLength)
{
  m_dynfStringLength.Set(pdynfStringLength);
}

inline VOID CSLIDECyclide::SetPhiMin(DynFloat *pdynfPhiMin)
{
  m_dynfPhiMin.Set(pdynfPhiMin);
}

inline VOID CSLIDECyclide::SetPhiMax(DynFloat *pdynfPhiMax)
{
  m_dynfPhiMax.Set(pdynfPhiMax);
}

inline VOID CSLIDECyclide::SetThetaMin(DynFloat *pdynfThetaMin)
{
  m_dynfThetaMin.Set(pdynfThetaMin);
}

inline VOID CSLIDECyclide::SetThetaMax(DynFloat *pdynfThetaMax)
{
  m_dynfThetaMax.Set(pdynfThetaMax);
}

inline VOID CSLIDECyclide::SetPhiSlices(DynInt *pdyniPhiSlices)
{
  m_dyniPhiSlices.Set(pdyniPhiSlices);
}

inline VOID CSLIDECyclide::SetThetaSlices(DynInt *pdyniThetaSlices)
{
  m_dyniThetaSlices.Set(pdyniThetaSlices);
}

inline VOID CSLIDECyclide::SetDrawEllipse(DynInt *pdyniDrawEllipse)
{
  m_dyniDrawEllipse.Set(pdyniDrawEllipse);
}

inline VOID CSLIDECyclide::SetDrawHyperbola(DynInt *pdyniDrawHyperbola)
{
  m_dyniDrawHyperbola.Set(pdyniDrawHyperbola);
}

inline VOID CSLIDECyclide::SetTexture(DynVector *pdynv0, DynVector *pdynv1, 
				      DynVector *pdynv2, DynVector *pdynv3)
{
  m_pvTexture[0].SetAll(pdynv0);
  m_pvTexture[1].SetAll(pdynv1);
  m_pvTexture[2].SetAll(pdynv2);
  m_pvTexture[3].SetAll(pdynv3);
}

inline CHAR *CSLIDECyclide::GetTypeString()
{
  return gcppcSLIDEReservedWords[sliderwtCyclide];
}

inline SLIDEReservedWordType CSLIDECyclide::GetType()
{
  return sliderwtCyclide;
}

#endif // __SLIDECYCLIDE_H__
