#ifndef __SLIDECYLINDER_H__
#define __SLIDECYLINDER_H__

#include "slideatomnode.h"

//////////////////////////////////////////////////////////////////////////////
// CSLIDECylinder Class
//

class CSLIDECylinder : public CSLIDEAtomNode
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDECylinder();
  virtual ~CSLIDECylinder();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  inline VOID SetRadius(DynFloat *pdynfRadius);
  inline VOID SetZMin(DynFloat *pdynfZMin);
  inline VOID SetZMax(DynFloat *pdynfZMax);
  inline VOID SetThetaMax(DynFloat *pdynfThetaMax);
  inline VOID SetZSlices(DynInt *pdyniZSlices);
  inline VOID SetEndCap(DynInt *pdyniEndCap);
  inline VOID SetBeginCap(DynInt *pdyniBeginCap);
  inline VOID SetThetaSlices(DynInt *pdyniThetaSlices);

  inline VOID SetTexture(DynVector *pdynv0, DynVector *pdynv1, DynVector *pdynv2, DynVector *pdynv3);

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Geometric Computations
//
public:

#ifdef __USE_CDA__
  virtual VOID CalculateCDAVertices(CListInt *plistCDAVertex, 
				    const Matrix &mGroot_obj);
#endif // __USE_CDA__

protected:
  virtual VOID CheckDynamic();
  virtual VOID PreprocessGraphLocal();
  virtual VOID UpdateState();
  virtual VOID CalculateBound();
  virtual VOID CalculateQEMesh();
  virtual VOID CalculateLeafQEMesh();  // for WriteWRLZ
  virtual VOID DeallocateLeafQEMesh();  // for WriteWRLZ

//////////////////////////////////////////////////
// Rendering
//
public:

  virtual VOID RenderGeometrySmooth(CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderGeometryFlat(CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderNormalsSmooth();
  virtual VOID RenderNormalsFlat();

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

  virtual VOID WritePOVGeometry(FILE *fpOutputFile, 
				CListInt *plistTransformIDs,
				CListInt *plistColors); 

  virtual VOID WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				 CSLIDESurface *pCSLIDESurface);
  virtual VOID WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform);

  virtual VOID WriteDotID(FILE *fpOutputFile);
  virtual VOID WriteDotNode(FILE *fpOutputFile);
  virtual VOID WriteDotEdges(FILE *fpOutputFile);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline CHAR *GetTypeString();
  virtual inline SLIDEReservedWordType GetType();

  inline VOID SetRadiusValue(FLOAT fRadius);
  inline VOID SetZMaxValue(FLOAT fZNax);
  inline VOID SetZSlicesValue(UINT_32 uiZSlices);
  inline VOID SetThetaSlicesValue(UINT_32 uiThetaSlices);

  virtual VOID GetSide(CListInt *plistPoint, CListInt *plistNormal, UINT_32 uiSide, 
		       const Matrix &mTransform, const Matrix &mInverse, BOOL bReverse);

  virtual VOID GetParameters(Tcl_Interp *pInterp);
  virtual UINT_32 SetParameter(CSLIDEWorld *pCSLIDEWorld,
			       SLIDEReservedWordType uiParameter, CHAR **argv);

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  DynFloat m_dynfRadius;
  DynFloat m_dynfZMin;
  DynFloat m_dynfZMax;
  DynFloat m_dynfThetaMax;

  DynInt m_dyniZSlices;
  DynInt m_dyniThetaSlices;  
  DynInt m_dyniEndCap;
  DynInt m_dyniBeginCap;

  DynVector m_pvTexture[4];
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDECylinder::SetRadius(DynFloat *pdynfRadius)
{
  m_dynfRadius.Set(pdynfRadius);
}

inline VOID CSLIDECylinder::SetZMin(DynFloat *pdynfZMin)
{
  m_dynfZMin.Set(pdynfZMin);
}

inline VOID CSLIDECylinder::SetZMax(DynFloat *pdynfZMax)
{
  m_dynfZMax.Set(pdynfZMax);
}

inline VOID CSLIDECylinder::SetThetaMax(DynFloat *pdynfThetaMax)
{
  m_dynfThetaMax.Set(pdynfThetaMax);
}

inline VOID CSLIDECylinder::SetZSlices(DynInt *pdyniZSlices)
{
  m_dyniZSlices.Set(pdyniZSlices);
}

inline VOID CSLIDECylinder::SetThetaSlices(DynInt *pdyniThetaSlices)
{
  m_dyniThetaSlices.Set(pdyniThetaSlices);
}

inline VOID CSLIDECylinder::SetEndCap(DynInt *pdyniEndCap)
{
  m_dyniEndCap.Set(pdyniEndCap);
}

inline VOID CSLIDECylinder::SetBeginCap(DynInt *pdyniBeginCap)
{
  m_dyniBeginCap.Set(pdyniBeginCap);
}

inline VOID CSLIDECylinder::SetTexture(DynVector *pdynv0, DynVector *pdynv1, 
				       DynVector *pdynv2, DynVector *pdynv3)
{
  m_pvTexture[0].SetAll(pdynv0);
  m_pvTexture[1].SetAll(pdynv1);
  m_pvTexture[2].SetAll(pdynv2);
  m_pvTexture[3].SetAll(pdynv3);
}

///

inline VOID CSLIDECylinder::SetRadiusValue(FLOAT fRadius)
{
  m_dynfRadius.SetValue(fRadius);
}
 
inline VOID CSLIDECylinder::SetZMaxValue(FLOAT fZMax)
{
  m_dynfZMax.SetValue(fZMax);
}

inline VOID CSLIDECylinder::SetZSlicesValue(UINT_32 uiZSlices)
{
  m_dyniZSlices.SetValue(uiZSlices);
}

inline VOID CSLIDECylinder::SetThetaSlicesValue(UINT_32 uiThetaSlices)
{
  m_dyniThetaSlices.SetValue(uiThetaSlices);
}

inline CHAR *CSLIDECylinder::GetTypeString()
{
  return gcppcSLIDEReservedWords[sliderwtCylinder];
}

inline SLIDEReservedWordType CSLIDECylinder::GetType()
{
  return sliderwtCylinder;
}

#endif // __SLIDECYLINDER_H__
