#include "slidedifference.h"

#include "slideworld.h"
#include "slidesurface.h"
#include "slideinstance.h"
#include "cdapolyhedron.h"
#include "cdavertex.h"

#include "qemesh.h"
#include "qewrite.h"
#include "sif.h"

//////////////////////////////////////////////////////////////////////
// CSLIDEDifference Class
//

CSLIDEDifference::CSLIDEDifference()
{
}

CSLIDEDifference::~CSLIDEDifference()
{
}

VOID CSLIDEDifference::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		       CHAR *pcID)
{
  CSLIDEGroup::Init(pcFile, uiLine, uiColumn, pcID);

  m_listCSLIDEInstance.Init();
}

VOID CSLIDEDifference::Uninit()
{
  CSLIDEGroup::Uninit();
}

BOOL CSLIDEDifference::Initialized()
{
  if ( !CSLIDEGroup::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

VOID CSLIDEDifference::InsertInstance(CSLIDEInstance *pCSLIDEInstance)
{
  ASSERT( pCSLIDEInstance != NULL );

  m_listCSLIDEInstance.InsertLast(pCSLIDEInstance, 0);
}

BOOL CSLIDEDifference::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  CSLIDEInstance *pCSLIDEInstance;

  if ( !CSLIDEGroup::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  pCSLIDEInstance = (CSLIDEInstance*)(m_listCSLIDEInstance.PeekFirst());
  if ( pCSLIDEInstance != NULL )
    {
      switch ( pCSLIDEInstance->GetNode()->GetType() )
	{
	case sliderwtIntersection:
	case sliderwtUnion:
	case sliderwtDifference:
	  break;
	default:
	  HandleError("The first argument to a %s node must be a %s, %s, or %s", 
		      gcppcSLIDEReservedWords[sliderwtDifference],
		      gcppcSLIDEReservedWords[sliderwtDifference],
		      gcppcSLIDEReservedWords[sliderwtUnion],
		      gcppcSLIDEReservedWords[sliderwtIntersection]);
	  return FALSE;
	  break;
	}
    }

  return TRUE;
}

BOOL CSLIDEDifference::Valid()
{
  if ( !CSLIDEGroup::Valid() )
    {
      return FALSE;
    }

  return TRUE;
}

CSLIDECamera *CSLIDEDifference::FindCamera(CListIterInt *piterIDs, 
					     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtCamera],
	      gcppcSLIDEReservedWords[sliderwtDifference], GetID());

  return NULL;
}

CSLIDELight *CSLIDEDifference::FindLight(CListIterInt *piterIDs, 
				     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtLight],
	      gcppcSLIDEReservedWords[sliderwtDifference], GetID());

  return NULL;
}

CSLIDEAreaLight *CSLIDEDifference::FindAreaLight(CListIterInt *piterIDs, 
				     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtAreaLight],
	      gcppcSLIDEReservedWords[sliderwtDifference], GetID());

  return NULL;
}

VOID CSLIDEDifference::Preprocess()
{
  CSLIDEGroup::Preprocess();

  CheckDynamic();
  UpdateState();
}

VOID CSLIDEDifference::CheckDynamic()
{
  SetDynamic(m_dyniShading.GetDynamic() || 
	     m_dyniLOD.GetDynamic());
}

VOID CSLIDEDifference::PreprocessGraph()
{
  if ( !GetVisited() )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;

      BOOL bGeometryEmpty;
      BOOL bGeometryDynamic;
      BOOL bNormalsEmpty;
      BOOL bNormalsDynamic;
      BOOL bBoundsEmpty;
      BOOL bBoundsDynamic;
      BOOL bBoundEmpty;
      BOOL bBoundDynamic;

      bGeometryEmpty = TRUE;
      bGeometryDynamic = FALSE;
      bNormalsEmpty = TRUE;
      bNormalsDynamic = FALSE;
      bBoundsEmpty = TRUE;
      bBoundsDynamic = FALSE;

      bBoundEmpty = TRUE;
      bBoundDynamic = FALSE;

      //////////////////////////////////////////////////
      // Visit Children Nodes and get starting values from them
      iter.Init(&m_listCSLIDEInstance);
      
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pCSLIDEInstance->PreprocessGraph();

	  bGeometryEmpty = ( bGeometryEmpty && 
			     pCSLIDEInstance->GetGeometryEmpty() );
	  bGeometryDynamic = ( bGeometryDynamic ||
			       pCSLIDEInstance->GetGeometryDynamic() );
	  bNormalsEmpty = ( bNormalsEmpty && 
			    pCSLIDEInstance->GetNormalsEmpty() );
	  bNormalsDynamic = ( bNormalsDynamic ||
			      pCSLIDEInstance->GetNormalsDynamic() );
	  bBoundsEmpty = ( bBoundsEmpty && 
			   pCSLIDEInstance->GetBoundsEmpty() );
	  bBoundsDynamic = ( bBoundsDynamic ||
			     pCSLIDEInstance->GetBoundsDynamic() );
	  bBoundEmpty = ( bBoundEmpty && 
			  pCSLIDEInstance->GetBoundEmpty() );
	  bBoundDynamic = ( bBoundDynamic ||
			    pCSLIDEInstance->GetBoundDynamic() );
	}

      //////////////////////////////////////////////////
      // Look at Local flags

      // Update whether we are Empty based on our LOD flag
      if ( !m_dyniLOD.GetDynamic() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      // This node is always off
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	      bBoundsEmpty = TRUE;
	      bBoundEmpty = TRUE;
	    }
	  else if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	    {
	      // This node is always just a bounding box
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	    }
	}

      // Check what is really dynamic

      bGeometryDynamic = ( bGeometryDynamic ||
			   m_pCSLIDESurface->GetDynamic() ||
			   m_dyniShading.GetDynamic() ||
			   m_dyniLOD.GetDynamic() );
      bNormalsDynamic = ( bNormalsDynamic ||
			  m_dyniShading.GetDynamic() ||
			  m_dyniLOD.GetDynamic() );
      bBoundsDynamic = ( bBoundsDynamic || 
			 m_dyniLOD.GetDynamic() );
      bBoundDynamic = ( bBoundDynamic || 
			m_dyniLOD.GetDynamic() );

      // If anything is now Empty then it will always be empty so
      // it is not dynamic

      bGeometryDynamic = ( bGeometryDynamic && !bGeometryEmpty );
      bNormalsDynamic = ( bNormalsDynamic && !bNormalsEmpty );
      bBoundsDynamic = ( bBoundsDynamic && !bBoundsEmpty );
      bBoundDynamic = ( bBoundDynamic && !bBoundEmpty );

      //////////////////////////////////////////////////
      // Now save these correct flag values

      SetGeometryEmpty(bGeometryEmpty);
      SetGeometryDynamic(bGeometryDynamic);
      SetNormalsEmpty(bNormalsEmpty);
      SetNormalsDynamic(bNormalsDynamic);
      SetBoundsEmpty(bBoundsEmpty);
      SetBoundsDynamic(bBoundsDynamic);
      SetBoundEmpty(bBoundEmpty);
      SetBoundDynamic(bBoundDynamic);
      
      //////////////////////////////
      // START: Jordan Fix This!!!
      //

      SetGeometryEmpty(FALSE);
      SetGeometryDynamic(TRUE);
      SetNormalsEmpty(FALSE);
      SetNormalsDynamic(TRUE);
      SetBoundsEmpty(FALSE);
      SetBoundsDynamic(TRUE);
      SetBoundEmpty(FALSE);
      SetBoundDynamic(TRUE);
      
      //
      // END: Jordan Fix This!!!
      //////////////////////////////

      //////////////////////////////////////////////////
      // Calculate the Bounding Box (if Empty this will do nothing)
      if ( GetBoundEmpty() )
	{
	  SetBoundValid(TRUE);
	}
      else
	{
	  CalculateBound();
	}
      SetBoundChanged(FALSE);

      //////////////////////////////////////////////////
      // Done Preprocessing this Node in the Graph
      SetVisited(TRUE);
    }
}

VOID CSLIDEDifference::Update()
{
  SetBoundChanged(FALSE);
  if ( GetDynamic() )
    {
      UpdateState();
    }

  if ( GetBoundsDynamic() )
    {
      CSLIDEBoundedNode::InvalidateQEMesh();
    }

  if ( GetBoundDynamic() )
    {
      SetBoundValid(FALSE);
    }

  SetVisited(FALSE);
}

VOID CSLIDEDifference::UpdateState()
{
  CSLIDEGroup::UpdateState();
}

VOID CSLIDEDifference::UpdateGraph()
{
  if ( !GetVisited() )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;
      BOOL bBoundChanged;

      if ( !GetBoundValid() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      if ( !m_bound.GetEmpty() )
		{
		  SetBoundChanged(TRUE);
		  m_bound.Null();
		}
	      return;
	    }
	  
	  bBoundChanged = FALSE;
	  
	  iter.Init(&m_listCSLIDEInstance);
	  
	  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
		pCSLIDEInstance != NULL;
		pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	    {
	      pCSLIDEInstance->UpdateGraph();
	      bBoundChanged = bBoundChanged || 
		pCSLIDEInstance->GetBoundChanged();
	    }
	  
	  if ( bBoundChanged )
	    {
	      CalculateBound();
	    }
	  
	  SetBoundValid(TRUE);
	}

      SetVisited(TRUE);
    }
}

VOID CSLIDEDifference::CalculateBound()
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  CBound bound;
  
  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->InsertIntoBound(bound);
    }

  ///////////////////////////////////////
  // START: Jordan Fix This!!!
  //

  Point pt;
  pt.Set(-1, -1, -1);
  bound.Insert(pt);
  pt.Set(1, 1, 1);
  bound.Insert(pt);

  //
  // END: Jordan Fix This!!!
  ///////////////////////////////////////

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);
    }
}

VOID CSLIDEDifference::CalculateCDA()
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  CDA *ppCDA[2];

  // HACK to disable CDA code
  m_bCDAVisited = TRUE;

  if ( m_bCDAVisited )
    {
      return;
    }
  m_bCDAVisited = TRUE;

  ASSERT( m_pCDA == NULL );

  iter.Init(&m_listCSLIDEInstance);
  
  // Create the Positive CDA
  pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
  if ( pCSLIDEInstance != NULL )
    {
      pCSLIDEInstance->CalculateCDA();
      m_pCDA = pCSLIDEInstance->GetCDA();
    }

  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext();
	pCSLIDEInstance != NULL && m_pCDA != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->CalculateCDA();
      
      ppCDA[1] = pCSLIDEInstance->GetCDA();
      if ( ppCDA[1] != NULL )
	{
	  ppCDA[0] = m_pCDA;
	  m_pCDA = ppCDA[0]->Minus(ppCDA[1]);
	  if ( m_pCDA == NULL )
	    {
	      //DeallocateCDA(ppCDA[0]);
	    }
	  //DeallocateCDA(ppCDA[1]);
	}
    }

  m_pCDA->Display();
}

VOID CSLIDEDifference::PrerenderGeometry(SLIDEShadingType eShading)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  PushShadingAttributes(eShading);
  {
    m_pCSLIDESurface->Push();
    {
      iter.Init(&m_listCSLIDEInstance);
      
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pCSLIDEInstance->PrerenderGeometry(eShading);
	}
    }
    m_pCSLIDESurface->Pop();
  }
  PopShadingAttributes(eShading);
}

VOID CSLIDEDifference::PrerenderNormals(SLIDEShadingType eShading)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->PrerenderNormals(eShading);
    }
}

VOID CSLIDEDifference::PrerenderBounds()
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
    {
    }
  else 
    {
      if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	{
	}
      else
	{
	  iter.Init(&m_listCSLIDEInstance);
	  
	  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
		pCSLIDEInstance != NULL;
		pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	    {
	      pCSLIDEInstance->PrerenderBounds();
	    }
	}

      m_bound.Render();
    }
}

VOID CSLIDEDifference::Prerender()
{
  if ( !(!m_dyniLOD.GetDynamic() && 
	 ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
	   m_dyniLOD.GetValue() == slidelodtSLF_BOUND )) )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;
      
      iter.Init(&m_listCSLIDEInstance);
      
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pCSLIDEInstance->Prerender();
	}
    }
      
  CSLIDEBoundedNode::Prerender();
}

VOID CSLIDEDifference::RenderGeometry(const Matrix &mGproj_obj, 
				      CBound *pCBoundClip, 
				      SLIDEShadingType eShading,
				      SLIDELODType eLOD,
				      CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{

  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  // For Debugging
  if ( GetRenderCDAOperandsEnabled() )
    {
      gilPushAttrib(GL_ALL_ATTRIB_BITS);
      {
	PushAuxillaryGeometryAttributes();
	{
	  PushWireFrameAttributes();
	  {
	    iter.Init(&m_listCSLIDEInstance);
	    
	    for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
		pCSLIDEInstance != NULL;
		  pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	      {
		pCSLIDEInstance->Render(mGproj_obj, pCBoundClip, slideshtSLF_WIRE, eLOD);
	      }
	  }
	  PopWireFrameAttributes();
	}
	PopAuxillaryGeometryAttributes();
      }
      gilPopAttrib();
    }
      
  CalculateCDA();
  if ( m_pCDA != NULL )
    {
      m_pCDA->Render();
    }

  // While CDA's don't work
  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->Render(mGproj_obj, pCBoundClip, eShading, eLOD);
    }
}

VOID CSLIDEDifference::RenderNormals(SLIDEShadingType /*eShading*/)
{
  if ( m_pCDA != NULL )
    {
      PushAuxillaryGeometryAttributes();
      {
	gilColor3f(1.0, 0.0, 1.0);
	
	m_pCDA->RenderNormals();
      }
      PopAuxillaryGeometryAttributes();
    }
}

VOID CSLIDEDifference::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtDifference], m_pcGUIID);

  WriteSLIDEFlags(fpOutputFile, 0);

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->WriteSLIDE(fpOutputFile);    
    }

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndDifference]);
}

VOID CSLIDEDifference::WriteRIBGeometry(FILE *fpOutputFile,
				   const Matrix &mGproj_obj, 
				   CBound *pCBoundClip, 
				   SLIDEShadingType eShading,
				   SLIDELODType eLOD)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( fpOutputFile != NULL );

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->WriteRIB(fpOutputFile,
				mGproj_obj, pCBoundClip, eShading, eLOD);    
    }
}

VOID CSLIDEDifference::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE DIFFERENCE: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);
}

VOID CSLIDEDifference::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE DIFFERENCE: %s", m_pcGUIID);
}

VOID CSLIDEDifference::WritePOVGeometry(FILE *fpOutputFile, 
					CListInt *plistTransformIDs,
					CListInt *plistColors) 
{
  ASSERT( fpOutputFile != NULL );

  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  fprintf(fpOutputFile, "difference { \n");

  iter.Init(&m_listCSLIDEInstance);
 
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->WritePOV(fpOutputFile, plistTransformIDs, plistColors);
    }
  fprintf(fpOutputFile, "}\n\n");
}

QEMesh *CSLIDEDifference::WriteSIF(FILE *fpOutputFile, UINT_32 uiIndentLevel, 
				   const Matrix &mGroot_obj, const Matrix &mGobj_root,
				   QEMesh *pmeBoundaryEdges)
{
  ASSERT( fpOutputFile != NULL );
  ASSERT( pmeBoundaryEdges != NULL );

  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  QEMesh *pme;
  
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "(%s\n", gcppcSIFReservedWords[sifrwtDifference]);
  uiIndentLevel++;

  iter.Init(&m_listCSLIDEInstance);
  
  if ( m_pmeSIF != NULL )
    {
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pme = pCSLIDEInstance->WriteSIF(fpOutputFile, uiIndentLevel, 
					  mGroot_obj, mGobj_root, pmeBoundaryEdges);
	  if ( pme != NULL )
	    {
	      DeallocateQEMesh(pme);
	    }
	}
    }
  else
    {
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pme = pCSLIDEInstance->WriteSIF(fpOutputFile, uiIndentLevel, 
					  mGroot_obj, mGobj_root, pmeBoundaryEdges);
	  if ( m_pmeSIF == NULL )
	    {
	      m_pmeSIF = pme;
	    }
	  else
	    {
	      if ( pme != NULL )
		{
		  // This will automatically destroy pme
		  m_pmeSIF->MergeMesh(pme);
		}
	    }
	}
    }

  if ( m_pmeSIF != NULL )
    {
      CObjectList<QEMesh> listQEMesh;

      listQEMesh.Init();

      pme = m_pmeSIF->Copy();
      pme->Transform(mGroot_obj, mGobj_root);
      pmeBoundaryEdges->CopyBoundaryEdges(pme);
      pme->ExtractComponents(&listQEMesh);
      DeallocateQEMesh(pme);

      for ( pme = listQEMesh.RemoveFirst();
	    pme != NULL;
	    pme = listQEMesh.RemoveFirst() )
	{
	  ::WriteSIFShell(pme, fpOutputFile, uiIndentLevel);
	  DeallocateQEMesh(pme);
	}
    }

  uiIndentLevel--;
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, ")\n");

  return NULL;
}

UINT_32 CSLIDEDifference::GetInstanceCount()
{
  return m_listCSLIDEInstance.GetLength();
}

