#ifndef __SLIDEDIFFERENCE_H__
#define __SLIDEDIFFERENCE_H__

#include "slidegroup.h"

class CSLIDEWorld;
class CSLIDEInstance;

//////////////////////////////////////////////////////////////////////
// CSLIDEDifference Class
//

class CSLIDEDifference : public CSLIDEGroup
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEDifference();
  virtual ~CSLIDEDifference();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  virtual VOID InsertInstance(CSLIDEInstance *pCSLIDEInstance);

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual CSLIDECamera *FindCamera(CListIterInt *piterIDs, 
				   CListInt *plistCSLIDEInstance);
  virtual CSLIDELight *FindLight(CListIterInt *piterIDs, 
				 CListInt *plistCSLIDEInstance);
  virtual CSLIDEAreaLight *FindAreaLight(CListIterInt *piterIDs, 
					 CListInt *plistCSLIDEInstance);

  virtual VOID Preprocess();
  virtual VOID PreprocessGraph();

  virtual VOID Update();
  virtual VOID UpdateGraph();

  virtual VOID CalculateCDA();

protected:
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();
  virtual VOID CalculateBound();

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID PrerenderGeometry(SLIDEShadingType eShading);
  virtual VOID PrerenderNormals(SLIDEShadingType eShading);
  virtual VOID PrerenderBounds();

  virtual VOID Prerender();

  virtual VOID RenderGeometry(const Matrix &mGproj_obj,
			      CBound *pCBoundClip, 
			      SLIDEShadingType eShading,
			      SLIDELODType eLOD,
			      CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderNormals(SLIDEShadingType eShading);

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

  virtual VOID WritePOVGeometry(FILE *fpOutputFile, 
				CListInt *plistTransformIDs,
				CListInt *plistColors); 

  virtual QEMesh *WriteSIF(FILE *fpOutputFile, UINT_32 uiIndentLevel, 
			   const Matrix &mGroot_obj, const Matrix &mGobj_root,
			   QEMesh *pmeBoundaryEdges);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline SLIDEReservedWordType GetType();

  UINT_32 GetInstanceCount();

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CListInt m_listCSLIDEInstance;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline SLIDEReservedWordType CSLIDEDifference::GetType()
{
  return sliderwtDifference;
}

#endif // __SLIDEDIFFERENCE_H__
