#include "slidedual.h"

#include "slidesurface.h"
#include "slideinstance.h"

#include "qemesh.h"
#include "qerender.h"
#include "qewrite.h"

extern BOOL gbPreprocessDone;

static const UINT_32 cuiDrawControlMesh = FALSE;

//////////////////////////////////////////////////////////////////////
// CSLIDEDual Class
//

CSLIDEDual::CSLIDEDual()
{
  m_dyniDrawControlMesh.SetValue(cuiDrawControlMesh);
  m_pmeDual = NULL;
}

CSLIDEDual::~CSLIDEDual()
{
}

VOID CSLIDEDual::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			     CHAR *pcID)
{
  CSLIDEGroup::Init(pcFile, uiLine, uiColumn, pcID);

  m_dyniDrawControlMesh.SetValue(cuiDrawControlMesh);
  m_pmeDual = NULL;
}

VOID CSLIDEDual::Uninit()
{
  CSLIDEGroup::Uninit();

  if ( m_pmeDual != NULL )
    {
      DeallocateQEMesh(m_pmeDual);
      m_pmeDual = NULL;
    }
}

BOOL CSLIDEDual::Initialized()
{
  if ( !CSLIDEGroup::Initialized() )
    {
      return FALSE;
    }

  return m_pmeDual != NULL;
}

BOOL CSLIDEDual::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;

  if ( !CSLIDEGroup::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  if ( !m_dyniDrawControlMesh.GetDynamic() &&
       !( m_dyniDrawControlMesh.GetValue() == TRUE || 
	  m_dyniDrawControlMesh.GetValue() == FALSE ) )
    {
      HandleError("Illegal drawcontrolmesh value:  %ld", 
      	  m_dyniDrawControlMesh.GetValue());
      m_dyniDrawControlMesh.SetValue(cuiDrawControlMesh);
      bCorrect = FALSE;
    }

  return bCorrect;
}

BOOL CSLIDEDual::Valid()
{
  if ( !CSLIDEGroup::Valid() )
    {
      return FALSE;
    }

  return TRUE;
}

CSLIDECamera *CSLIDEDual::FindCamera(CListIterInt *piterIDs, 
					     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtCamera],
	      gcppcSLIDEReservedWords[sliderwtDual], GetID());

  return NULL;
}

CSLIDELight *CSLIDEDual::FindLight(CListIterInt *piterIDs, 
				     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtLight],
	      gcppcSLIDEReservedWords[sliderwtDual], GetID());

  return NULL;
}

CSLIDEAreaLight *CSLIDEDual::FindAreaLight(CListIterInt *piterIDs, 
					    CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtAreaLight],
	      gcppcSLIDEReservedWords[sliderwtDual], GetID());

  return NULL;
}

VOID CSLIDEDual::Preprocess()
{
  CSLIDEGroup::Preprocess();
}

VOID CSLIDEDual::CheckDynamic()
{
  SetDynamic(m_dyniShading.GetDynamic() || 
	     m_dyniLOD.GetDynamic() ||
             m_dyniDrawControlMesh.GetDynamic());
}

VOID CSLIDEDual::PreprocessGraph()
{
  if ( !GetVisited() )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;

      BOOL bGeometryEmpty;
      BOOL bGeometryDynamic;
      BOOL bNormalsEmpty;
      BOOL bNormalsDynamic;
      BOOL bBoundsEmpty;
      BOOL bBoundsDynamic;
      BOOL bBoundEmpty;
      BOOL bBoundDynamic;

      bGeometryEmpty = TRUE;
      bGeometryDynamic = FALSE;
      bNormalsEmpty = TRUE;
      bNormalsDynamic = FALSE;
      bBoundsEmpty = TRUE;
      bBoundsDynamic = FALSE;

      bBoundEmpty = TRUE;
      bBoundDynamic = FALSE;

      //////////////////////////////////////////////////
      // Visit Children Nodes and get starting values from them
      iter.Init(&m_listCSLIDEInstance);
      
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pCSLIDEInstance->PreprocessGraph();

	  bGeometryEmpty = ( bGeometryEmpty && 
			     pCSLIDEInstance->GetGeometryEmpty() );
	  bGeometryDynamic = ( bGeometryDynamic ||
			       pCSLIDEInstance->GetGeometryDynamic() );
	  bNormalsEmpty = ( bNormalsEmpty && 
			    pCSLIDEInstance->GetNormalsEmpty() );
	  bNormalsDynamic = ( bNormalsDynamic ||
			      pCSLIDEInstance->GetNormalsDynamic() );
	  bBoundsEmpty = ( bBoundsEmpty && 
			   pCSLIDEInstance->GetBoundsEmpty() );
	  bBoundsDynamic = ( bBoundsDynamic ||
			     pCSLIDEInstance->GetBoundsDynamic() );
	  bBoundEmpty = ( bBoundEmpty && 
			  pCSLIDEInstance->GetBoundEmpty() );
	  bBoundDynamic = ( bBoundDynamic ||
			    pCSLIDEInstance->GetBoundDynamic() );
	}

      //////////////////////////////////////////////////
      // Look at Local flags

      // Update whether we are Empty based on our LOD flag
      if ( !m_dyniLOD.GetDynamic() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      // This node is always off
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	      bBoundsEmpty = TRUE;
	      bBoundEmpty = TRUE;
	    }
	  else if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	    {
	      // This node is always just a bounding box
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	    }
	}

      // Check what is really dynamic

      bGeometryDynamic = ( bGeometryDynamic ||
			   m_pCSLIDESurface->GetDynamic() ||
			   m_dyniShading.GetDynamic() ||
			   m_dyniLOD.GetDynamic() );
      bNormalsDynamic = ( bNormalsDynamic ||
			  m_dyniShading.GetDynamic() ||
			  m_dyniLOD.GetDynamic() );
      bBoundsDynamic = ( bBoundsDynamic || 
			 m_dyniLOD.GetDynamic() );
      bBoundDynamic = ( bBoundDynamic || 
			m_dyniLOD.GetDynamic() );

      // If anything is now Empty then it will always be empty so
      // it is not dynamic

      bGeometryDynamic = ( bGeometryDynamic && !bGeometryEmpty );
      bNormalsDynamic = ( bNormalsDynamic && !bNormalsEmpty );
      bBoundsDynamic = ( bBoundsDynamic && !bBoundsEmpty );
      bBoundDynamic = ( bBoundDynamic && !bBoundEmpty );

      //////////////////////////////////////////////////
      // Now save these correct flag values

      SetGeometryEmpty(bGeometryEmpty);
      SetGeometryDynamic(bGeometryDynamic);
      SetNormalsEmpty(bNormalsEmpty);
      SetNormalsDynamic(bNormalsDynamic);
      SetBoundsEmpty(bBoundsEmpty);
      SetBoundsDynamic(bBoundsDynamic);
      SetBoundEmpty(bBoundEmpty);
      SetBoundDynamic(bBoundDynamic);
      
      //////////////////////////////
      // START: Jordan Fix This!!!
      //

      SetGeometryEmpty(FALSE);
      SetGeometryDynamic(TRUE);
      SetNormalsEmpty(FALSE);
      SetNormalsDynamic(TRUE);
      SetBoundsEmpty(FALSE);
      SetBoundsDynamic(TRUE);
      SetBoundEmpty(FALSE);
      SetBoundDynamic(TRUE);
      
      UpdateQEMesh();
      UpdateDual();

      //
      // END: Jordan Fix This!!!
      //////////////////////////////

      //////////////////////////////////////////////////
      // Calculate the Bounding Box (if Empty this will do nothing)
      if ( GetBoundEmpty() )
	{
	  SetBoundValid(TRUE);
	}
      else
	{
	  CalculateBound();
	}
      SetBoundChanged(FALSE);

      //////////////////////////////////////////////////
      // Done Preprocessing this Node in the Graph
      SetVisited(TRUE);
    }
}

VOID CSLIDEDual::Update()
{
  SetBoundChanged(FALSE);
  if ( GetDynamic() ||
       GetBoundsDynamic() )
    {
      if ( GetBoundsDynamic() )
	{
	  CSLIDEBoundedNode::InvalidateQEMesh();
	}
      UpdateState();
    }

  if ( GetBoundDynamic() )
    {
      SetBoundValid(FALSE);
    }

  SetVisited(FALSE);
}

VOID CSLIDEDual::UpdateState()
{
  CSLIDEGroup::UpdateState();

  m_dyniDrawControlMesh.Update();

  //////////////////////////////////////////////////
  // Destroy any meshes that need to regenerated
  if ( m_pQEMesh == NULL )
    {
      if ( m_pmeDual != NULL )
	{
	  DeallocateQEMesh(m_pmeDual);
	  m_pmeDual = NULL;
	}
    }
}

VOID CSLIDEDual::UpdateQEMesh()
{
  if ( m_pQEMesh == NULL )
    {
      ASSERT( m_pmeDual == NULL );
      CalculateQEMesh();
    }
  ASSERT( m_pQEMesh != NULL );
}

VOID CSLIDEDual::UpdateDual()
{
  ASSERT( m_pQEMesh != NULL );
  if ( m_pmeDual == NULL )
    {
      if ( m_pQEMesh->GetBoundaryEdgeCount() > 0 )
	{
	  m_pmeDual = m_pQEMesh->Copy();
	}
      else
	{
	  m_pmeDual = m_pQEMesh->CreateDual();
	}
    }
  ASSERT( m_pmeDual != NULL );
}

VOID CSLIDEDual::UpdateGraph()
{
  if ( !GetVisited() )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;
      BOOL bBoundChanged;

      if ( !GetBoundValid() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      if ( !m_bound.GetEmpty() )
		{
		  SetBoundChanged(TRUE);
		  m_bound.Null();
		}
	      return;
	    }
	  
	  bBoundChanged = FALSE;
	  
	  iter.Init(&m_listCSLIDEInstance);
	  
	  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
		pCSLIDEInstance != NULL;
		pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	    {
	      pCSLIDEInstance->UpdateGraph();
	      bBoundChanged = bBoundChanged || 
		pCSLIDEInstance->GetBoundChanged();
	    }
	  
	  UpdateQEMesh();
	  UpdateDual();

	  if ( bBoundChanged )
	    {
	      CalculateBound();
	    }
	  
	  SetBoundValid(TRUE);
	}

      SetVisited(TRUE);
    }
}

VOID CSLIDEDual::CalculateBound()
{
  CBound bound;

/*
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  
  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->InsertIntoBound(bound);
    }
*/
  ///////////////////////////////////////
  // START: Jordan Fix This!!!
  //
  {
    QEMeshIter iterQEMesh;
    QEVertex *pvx;
    Point pt;

    iterQEMesh.Init(m_pmeDual);
    for ( pvx = iterQEMesh.PeekFirstVertex();
	  pvx != NULL;
	  pvx = iterQEMesh.PeekNextVertex() )
      {
	pt = pvx->GetPoint();
	bound.Insert(pt);
      }
  }
  //
  // END: Jordan Fix This!!!
  ///////////////////////////////////////

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);
    }
}

VOID CSLIDEDual::PrerenderGeometry(SLIDEShadingType eShading)
{
  CListIterInt iter;

  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  PushShadingAttributes(eShading);
  {
    m_pCSLIDESurface->Push();
    {
      ::RenderGeometry(m_pmeDual);
    }
    m_pCSLIDESurface->Pop();
  }
  PopShadingAttributes(eShading);

  if ( m_dyniDrawControlMesh.GetValue() )
    {
      QEMeshIter iterQEMesh;
      QEEdge *pe;

      iterQEMesh.Init(m_pQEMesh);

      PushAuxillaryGeometryAttributes();
      {
	gilColor3f(1.0, 1.0, 0.0);
	gilBegin(GL_LINES);
	{
	  for ( pe = iterQEMesh.PeekFirstEdge();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdge() )
	    {
	      gilVertex3fv(pe->GetVertex()->GetPoint().Get());
	      gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
	    }
	  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdgeBoundary() )
	    {
	      gilVertex3fv(pe->GetVertex()->GetPoint().Get());
	      gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
	    }
	}
	gilEnd();
      }
      PopAuxillaryGeometryAttributes();
    }

  if ( m_pQEMesh->GetBoundaryEdgeCount() > 0 )
    {
      QEMeshIter iterQEMesh;
      QEEdge *pe;

      iterQEMesh.Init(m_pQEMesh);

      PushAuxillaryGeometryAttributes();
      {
	gilColor3f(1.0, 0.0, 0.0);
	gilBegin(GL_LINES);
	{
	  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdgeBoundary() )
	    {
	      gilVertex3fv(pe->GetVertex()->GetPoint().Get());
	      gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
	    }
	}
	gilEnd();
      }
      PopAuxillaryGeometryAttributes();
    }
}

VOID CSLIDEDual::PrerenderNormals(SLIDEShadingType eShading)
{
  CListIterInt iter;

  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  ::RenderNormals(m_pmeDual);
}

VOID CSLIDEDual::PrerenderBounds()
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
    {
    }
  else 
    {
      if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	{
	}
      else
	{
	  iter.Init(&m_listCSLIDEInstance);
	  
	  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
		pCSLIDEInstance != NULL;
		pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	    {
	      pCSLIDEInstance->PrerenderBounds();
	    }
	}

      m_bound.Render();
    }
}

VOID CSLIDEDual::Prerender()
{
  if ( !(!m_dyniLOD.GetDynamic() && 
	 ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
	   m_dyniLOD.GetValue() == slidelodtSLF_BOUND )) )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;
      
      iter.Init(&m_listCSLIDEInstance);
      
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pCSLIDEInstance->Prerender();
	}
    }
      
  CSLIDEBoundedNode::Prerender();
}

VOID CSLIDEDual::RenderGeometry(const Matrix &mGproj_obj, 
				CBound *pCBoundClip, 
				SLIDEShadingType /*eShading*/,
				SLIDELODType eLOD,
				CSLIDEWormwarp *pCSLIDEWarp/* = NULL*/)
{

  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  // For Debugging
  if ( GetRenderCDAOperandsEnabled() )
    {
      gilPushAttrib(GL_ALL_ATTRIB_BITS);
      {
	PushAuxillaryGeometryAttributes();
	{
	  PushWireFrameAttributes();
	  {
	    iter.Init(&m_listCSLIDEInstance);
	    
	    for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
		pCSLIDEInstance != NULL;
		  pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	      {
		pCSLIDEInstance->Render(mGproj_obj, pCBoundClip, slideshtSLF_WIRE, eLOD);
	      }
	  }
	  PopWireFrameAttributes();
	}
	PopAuxillaryGeometryAttributes();
      }
      gilPopAttrib();
    }
      
  ::RenderGeometry(m_pmeDual);

  if ( m_dyniDrawControlMesh.GetValue() )
    {
      QEMeshIter iterQEMesh;
      QEEdge *pe;

      iterQEMesh.Init(m_pQEMesh);

      PushAuxillaryGeometryAttributes();
      {
	gilColor3f(1.0, 1.0, 0.0);
	gilBegin(GL_LINES);
	{
	  for ( pe = iterQEMesh.PeekFirstEdge();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdge() )
	    {
	      gilVertex3fv(pe->GetVertex()->GetPoint().Get());
	      gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
	    }
	  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdgeBoundary() )
	    {
	      gilVertex3fv(pe->GetVertex()->GetPoint().Get());
	      gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
	    }
	}
	gilEnd();
      }
      PopAuxillaryGeometryAttributes();
    }

  if ( m_pQEMesh->GetBoundaryEdgeCount() > 0 )
    {
      QEMeshIter iterQEMesh;
      QEEdge *pe;

      iterQEMesh.Init(m_pQEMesh);

      PushAuxillaryGeometryAttributes();
      {
	gilColor3f(1.0, 0.0, 0.0);
	gilBegin(GL_LINES);
	{
	  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdgeBoundary() )
	    {
	      gilVertex3fv(pe->GetVertex()->GetPoint().Get());
	      gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
	    }
	}
	gilEnd();
      }
      PopAuxillaryGeometryAttributes();
    }
}

VOID CSLIDEDual::RenderNormals(SLIDEShadingType /*eShading*/)
{
  if ( m_pCDA != NULL )
    {
      PushAuxillaryGeometryAttributes();
      {
	gilColor3f(1.0, 0.0, 1.0);
	
	::RenderNormals(m_pmeDual);
      }
      PopAuxillaryGeometryAttributes();
    }
}

VOID CSLIDEDual::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtDual], m_pcID);

  WriteSLIDEFlags(fpOutputFile, 0);

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->WriteSLIDE(fpOutputFile);    
    }

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndDual]);
}

VOID CSLIDEDual::WriteRIBGeometry(FILE *fpOutputFile,
				   const Matrix &mGproj_obj, 
				   CBound *pCBoundClip, 
				   SLIDEShadingType eShading,
				   SLIDELODType eLOD)
{
/*
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( fpOutputFile != NULL );

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->WriteRIB(fpOutputFile,
				mGproj_obj, pCBoundClip, eShading, eLOD);    
    }
    */
}

VOID CSLIDEDual::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );
  /*
  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE DUAL: %s", m_pcID);
  WriteRIBBegin(fpOutputFile);
  */
}

VOID CSLIDEDual::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );
  /*
  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE DUAL: %s", m_pcID);
  */
}

VOID CSLIDEDual::WriteWRLZ(FILE *fpOutputFile, const Matrix &m, 
			     CSLIDESurface *pCSLIDESurface)
{
  ::WriteWRLZ(m_pmeDual, fpOutputFile, m);
}

VOID CSLIDEDual::WriteSTL(FILE *fpOutputFile, const Matrix &m)
{
  ::WriteSTL(m_pmeDual, fpOutputFile, m);
}






