#ifndef __SLIDEDUAL_H__
#define __SLIDEDUAL_H__

#include "slidegroup.h"

class CSLIDEWorld;

class QEMesh;

//////////////////////////////////////////////////////////////////////
// CSLIDEDual Class
//

class CSLIDEDual : public CSLIDEGroup
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEDual();
  virtual ~CSLIDEDual();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  inline VOID SetDrawControlMesh(DynInt *pdyniDrawControlMesh);

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual CSLIDECamera *FindCamera(CListIterInt *piterIDs, 
				   CListInt *plistCSLIDEInstance);
  virtual CSLIDELight *FindLight(CListIterInt *piterIDs, 
				 CListInt *plistCSLIDEInstance);
  virtual CSLIDEAreaLight *FindAreaLight(CListIterInt *piterIDs, 
					 CListInt *plistCSLIDEInstance);

  virtual VOID Preprocess();
  virtual VOID PreprocessGraph();

  virtual VOID Update();
  virtual VOID UpdateGraph();

protected:
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();
  virtual VOID UpdateQEMesh();
  virtual VOID UpdateDual();
  virtual VOID CalculateBound();

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID PrerenderGeometry(SLIDEShadingType eShading);
  virtual VOID PrerenderNormals(SLIDEShadingType eShading);
  virtual VOID PrerenderBounds();

  virtual VOID Prerender();

  virtual VOID RenderGeometry(const Matrix &mGproj_obj,
			      CBound *pCBoundClip, 
			      SLIDEShadingType eShading,
			      SLIDELODType eLOD,
			      CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderNormals(SLIDEShadingType eShading);

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

  virtual VOID WriteWRLZ(FILE *fpOutputFile, const Matrix &mTransform,
			 CSLIDESurface *pCSLIDESurface);
  virtual VOID WriteSTL(FILE *fpOutputFile, const Matrix &mTransform);

//////////////////////////////////////////////////
// Accessor Functions
//
public:

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  DynInt m_dyniDrawControlMesh;

  QEMesh *m_pmeDual;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDEDual::SetDrawControlMesh(DynInt *pdyniDrawControlMesh)
{
  m_dyniDrawControlMesh.Set(pdyniDrawControlMesh);
}

#endif // __SLIDEDUAL_H__
