#include "slideevolver.h"

#include "slidesurface.h"
#include "slideinstance.h"

#include "qemesh.h"
#include "qerender.h"
#include "qewrite.h"

extern BOOL gbPreprocessDone;

static const UINT_32 cuiDrawControlMesh = FALSE;
static const UINT_32 cuiDrawCurrentMesh = FALSE;
static const UINT_32 cuiSubdivisionLevel = 0;
static const UINT_32 cuiSubdivisionLevelSize = 6;
static const FLOAT cfGradientPercent = 0.5 ;

//////////////////////////////////////////////////////////////////////
// CSLIDEEvolver Class
//

CSLIDEEvolver::CSLIDEEvolver()
{
  m_dyniDrawControlMesh.SetValue(cuiDrawControlMesh);
  m_dyniDrawCurrentMesh.SetValue(cuiDrawCurrentMesh);
  m_dyniSubdivisionLevel.SetValue(cuiSubdivisionLevel);

  m_uiSubdivisionLevelSize = cuiSubdivisionLevelSize;
  m_ppQEMeshSubdivision = NULL;

  m_fGradientPercent = cfGradientPercent;
}

CSLIDEEvolver::~CSLIDEEvolver()
{
}

VOID CSLIDEEvolver::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			 CHAR *pcID)
{
  UINT_32 i;

  CSLIDEGroup::Init(pcFile, uiLine, uiColumn, pcID);

  m_dyniDrawControlMesh.SetValue(cuiDrawControlMesh);
  m_dyniDrawCurrentMesh.SetValue(cuiDrawCurrentMesh);
  m_dyniSubdivisionLevel.SetValue(cuiSubdivisionLevel);

  m_ppQEMeshSubdivision = new QEMesh*[m_uiSubdivisionLevelSize];
  ASSERT( m_ppQEMeshSubdivision != NULL );
  for ( i = 0; i < m_uiSubdivisionLevelSize; i++ )
    {
      m_ppQEMeshSubdivision[i] = NULL;
    }

  m_fGradientPercent = cfGradientPercent;
}

VOID CSLIDEEvolver::Uninit()
{
  UINT_32 i;

  m_pQEMesh = NULL;

  CSLIDEGroup::Uninit();

  if ( m_ppQEMeshSubdivision != NULL )
    {
      for ( i = 0; i < m_uiSubdivisionLevelSize; i++ )
	{
	  if ( m_ppQEMeshSubdivision[i] != NULL )
	    {
	      DeallocateQEMesh(m_ppQEMeshSubdivision[i]);
	      m_ppQEMeshSubdivision[i] = NULL;
	    }
	}
      delete []m_ppQEMeshSubdivision;
      m_ppQEMeshSubdivision = NULL;
    }
}

BOOL CSLIDEEvolver::Initialized()
{
  if ( !CSLIDEGroup::Initialized() )
    {
      return FALSE;
    }

  return m_ppQEMeshSubdivision != NULL;
}

BOOL CSLIDEEvolver::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;

  if ( !CSLIDEGroup::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  if ( !m_dyniDrawControlMesh.GetDynamic() &&
       !( m_dyniDrawControlMesh.GetValue() == TRUE || 
	  m_dyniDrawControlMesh.GetValue() == FALSE ) )
    {
      HandleError("Illegal drawcontrolmesh value:  %ld", 
      	  m_dyniDrawControlMesh.GetValue());
      m_dyniDrawControlMesh.SetValue(cuiDrawControlMesh);
      bCorrect = FALSE;
    }

  if ( !m_dyniDrawCurrentMesh.GetDynamic() &&
       !( m_dyniDrawCurrentMesh.GetValue() == TRUE || 
	  m_dyniDrawCurrentMesh.GetValue() == FALSE ) )
    {
      HandleError("Illegal drawcurrentmesh value:  %ld", 
      	  m_dyniDrawCurrentMesh.GetValue());
      m_dyniDrawCurrentMesh.SetValue(cuiDrawCurrentMesh);
      bCorrect = FALSE;
    }

  if ( !m_dyniSubdivisionLevel.GetDynamic() &&
       !( 0 <= m_dyniSubdivisionLevel.GetValue() ) )
    {
      HandleError("Illegal SubdivisionLevel value:  %ld", 
      	  m_dyniSubdivisionLevel.GetValue());
      m_dyniSubdivisionLevel.SetValue(cuiSubdivisionLevel);
      bCorrect = FALSE;
    }

  return bCorrect;
}

BOOL CSLIDEEvolver::Valid()
{
  if ( !CSLIDEGroup::Valid() )
    {
      return FALSE;
    }

  return TRUE;
}

CSLIDECamera *CSLIDEEvolver::FindCamera(CListIterInt *piterIDs, 
					     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtCamera],
	      gcppcSLIDEReservedWords[sliderwtEvolver], GetID());

  return NULL;
}

CSLIDELight *CSLIDEEvolver::FindLight(CListIterInt *piterIDs, 
				     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtLight],
	      gcppcSLIDEReservedWords[sliderwtEvolver], GetID());

  return NULL;
}

CSLIDEAreaLight *CSLIDEEvolver::FindAreaLight(CListIterInt *piterIDs, 
					    CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtAreaLight],
	      gcppcSLIDEReservedWords[sliderwtEvolver], GetID());

  return NULL;
}

VOID CSLIDEEvolver::Preprocess()
{
  CSLIDEGroup::Preprocess();
}

VOID CSLIDEEvolver::CheckDynamic()
{
  SetDynamic(m_dyniShading.GetDynamic() || 
	     m_dyniLOD.GetDynamic() ||
             m_dyniDrawControlMesh.GetDynamic() ||
             m_dyniDrawCurrentMesh.GetDynamic() ||
             m_dyniSubdivisionLevel.GetDynamic());
}

VOID CSLIDEEvolver::PreprocessGraph()
{
  if ( !GetVisited() )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;

      BOOL bGeometryEmpty;
      BOOL bGeometryDynamic;
      BOOL bNormalsEmpty;
      BOOL bNormalsDynamic;
      BOOL bBoundsEmpty;
      BOOL bBoundsDynamic;
      BOOL bBoundEmpty;
      BOOL bBoundDynamic;

      bGeometryEmpty = TRUE;
      bGeometryDynamic = FALSE;
      bNormalsEmpty = TRUE;
      bNormalsDynamic = FALSE;
      bBoundsEmpty = TRUE;
      bBoundsDynamic = FALSE;

      bBoundEmpty = TRUE;
      bBoundDynamic = FALSE;

      //////////////////////////////////////////////////
      // Visit Children Nodes and get starting values from them
      iter.Init(&m_listCSLIDEInstance);
      
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pCSLIDEInstance->PreprocessGraph();

	  bGeometryEmpty = ( bGeometryEmpty && 
			     pCSLIDEInstance->GetGeometryEmpty() );
	  bGeometryDynamic = ( bGeometryDynamic ||
			       pCSLIDEInstance->GetGeometryDynamic() );
	  bNormalsEmpty = ( bNormalsEmpty && 
			    pCSLIDEInstance->GetNormalsEmpty() );
	  bNormalsDynamic = ( bNormalsDynamic ||
			      pCSLIDEInstance->GetNormalsDynamic() );
	  bBoundsEmpty = ( bBoundsEmpty && 
			   pCSLIDEInstance->GetBoundsEmpty() );
	  bBoundsDynamic = ( bBoundsDynamic ||
			     pCSLIDEInstance->GetBoundsDynamic() );
	  bBoundEmpty = ( bBoundEmpty && 
			  pCSLIDEInstance->GetBoundEmpty() );
	  bBoundDynamic = ( bBoundDynamic ||
			    pCSLIDEInstance->GetBoundDynamic() );
	}

      //////////////////////////////////////////////////
      // Look at Local flags

      // Update whether we are Empty based on our LOD flag
      if ( !m_dyniLOD.GetDynamic() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      // This node is always off
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	      bBoundsEmpty = TRUE;
	      bBoundEmpty = TRUE;
	    }
	  else if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	    {
	      // This node is always just a bounding box
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	    }
	}

      // Check what is really dynamic

      bGeometryDynamic = ( bGeometryDynamic ||
			   m_pCSLIDESurface->GetDynamic() ||
			   m_dyniShading.GetDynamic() ||
			   m_dyniLOD.GetDynamic() );
      bNormalsDynamic = ( bNormalsDynamic ||
			  m_dyniShading.GetDynamic() ||
			  m_dyniLOD.GetDynamic() );
      bBoundsDynamic = ( bBoundsDynamic || 
			 m_dyniLOD.GetDynamic() ||
			 m_dyniSubdivisionLevel.GetDynamic() );
      bBoundDynamic = ( bBoundDynamic || 
			m_dyniLOD.GetDynamic() ||
			m_dyniSubdivisionLevel.GetDynamic() );

      // If anything is now Empty then it will always be empty so
      // it is not dynamic

      bGeometryDynamic = ( bGeometryDynamic && !bGeometryEmpty );
      bNormalsDynamic = ( bNormalsDynamic && !bNormalsEmpty );
      bBoundsDynamic = ( bBoundsDynamic && !bBoundsEmpty );
      bBoundDynamic = ( bBoundDynamic && !bBoundEmpty );

      //////////////////////////////////////////////////
      // Now save these correct flag values

      SetGeometryEmpty(bGeometryEmpty);
      SetGeometryDynamic(bGeometryDynamic);
      SetNormalsEmpty(bNormalsEmpty);
      SetNormalsDynamic(bNormalsDynamic);
      SetBoundsEmpty(bBoundsEmpty);
      SetBoundsDynamic(bBoundsDynamic);
      SetBoundEmpty(bBoundEmpty);
      SetBoundDynamic(bBoundDynamic);
      
      //////////////////////////////
      // START: Jordan Fix This!!!
      //

      SetGeometryEmpty(FALSE);
      SetGeometryDynamic(TRUE);
      SetNormalsEmpty(FALSE);
      SetNormalsDynamic(TRUE);
      SetBoundsEmpty(FALSE);
      SetBoundsDynamic(TRUE);
      SetBoundEmpty(FALSE);
      SetBoundDynamic(TRUE);
      
      UpdateQEMesh();
      UpdateEvolver();

      //
      // END: Jordan Fix This!!!
      //////////////////////////////

      //////////////////////////////////////////////////
      // Calculate the Bounding Box (if Empty this will do nothing)
      if ( GetBoundEmpty() )
	{
	  SetBoundValid(TRUE);
	}
      else
	{
	  CalculateBound();
	}
      SetBoundChanged(FALSE);

      //////////////////////////////////////////////////
      // Done Preprocessing this Node in the Graph
      SetVisited(TRUE);
    }
}

VOID CSLIDEEvolver::Update()
{
  SetBoundChanged(FALSE);
  if ( GetDynamic() ||
       GetBoundsDynamic() )
    {
      UpdateState();
    }

  if ( GetBoundDynamic() )
    {
      SetBoundValid(FALSE);
    }

  SetVisited(FALSE);
}

VOID CSLIDEEvolver::UpdateState()
{
  UINT_32 i;
  BOOL b;

  CSLIDEGroup::UpdateState();

  m_dyniDrawControlMesh.Update();

  m_dyniDrawCurrentMesh.Update();

  b = m_dyniSubdivisionLevel.Update();
  m_bChanged = ( m_bChanged || b );
  if ( !( 0 <= m_dyniSubdivisionLevel.GetValue() ) )
    {
      HandleError("Illegal subdivisionlevel value:  %ld", 
		  m_dyniSubdivisionLevel.GetValue());
      m_dyniSubdivisionLevel.SetValue(cuiSubdivisionLevel);
    }

  //////////////////////////////////////////////////
  // See if we need to grow the mesh array
  if ( (UINT_32)m_dyniSubdivisionLevel.GetValue() + 1 > m_uiSubdivisionLevelSize )
    {
      QEMesh **ppQEMesh;

      ppQEMesh = m_ppQEMeshSubdivision;
      m_ppQEMeshSubdivision = new QEMesh*[m_dyniSubdivisionLevel.GetValue() + 1];
      ASSERT( m_ppQEMeshSubdivision != NULL );
      for ( i = 0; i < m_uiSubdivisionLevelSize; i++ )
	{
	  m_ppQEMeshSubdivision[i] = ppQEMesh[i];
	}
      m_uiSubdivisionLevelSize = m_dyniSubdivisionLevel.GetValue() + 1;
      for ( ; i < m_uiSubdivisionLevelSize; i++ )
	{
	  m_ppQEMeshSubdivision[i] = NULL;	  
	}

      delete []ppQEMesh;
    }
}

VOID CSLIDEEvolver::UpdateQEMesh()
{
  if ( m_ppQEMeshSubdivision[0] == NULL )
    {
      if ( m_pQEMesh == NULL )
	{
	  CalculateQEMesh();
	}
      ASSERT( m_pQEMesh != NULL );
      m_ppQEMeshSubdivision[0] = m_pQEMesh;
      m_ppQEMeshSubdivision[0]->CalculateSubdivisionBoundaryEdgesSharpnessesSharpCatmullClark();
      m_ppQEMeshSubdivision[0]->CalculateSubdivisionLimitPositionsAndNormalsSharpLoop();
    }
  ASSERT( m_ppQEMeshSubdivision[0] != NULL );
  ASSERT( m_pQEMesh != NULL );
}

VOID CSLIDEEvolver::UpdateEvolver()
{
  UINT_32 i, j;
  FLOAT fEnergyOld, fEnergyNew;
  QEMesh *pmeSwap, *ppme[2];

  const UINT_32 cuiMicroSteps = 1000;
  const FLOAT cfMinGradientPercent = 1.0/((FLOAT)(0x1 << 10));

  ASSERT( m_ppQEMeshSubdivision[0] != NULL );
  ASSERT( m_pQEMesh != NULL );

  for ( i = 1; i <= (UINT_32)m_dyniSubdivisionLevel.GetValue(); i++ )
    {
      if ( m_ppQEMeshSubdivision[i] == NULL )
	{
	  m_fGradientPercent = cfGradientPercent;

	  //////////////////////////////////////////////////
	  // Make two copies so that they can ping pong which is the current saved state
	  if ( i == 1 )
	    {
	      ppme[0] = m_ppQEMeshSubdivision[i-1]->EvolverCopy();
	    }
	  else
	    {
	      ppme[0] = m_ppQEMeshSubdivision[i-1]->SubdivideCornerRoundingSharpLoop();
	    }
	  ppme[0]->CalculateSubdivisionLimitPositionsAndNormalsSharpLoop();
	  ppme[1] = ppme[0]->EvolverCopy();

	  fEnergyOld = ppme[0]->EvolverEnergy();
	  //////////////////////////////////////////////////
	  //  We are recording the mesh at every nth iteration for debugging
	  for ( j = 0; j < cuiMicroSteps; j++ )
	    {
	      //////////////////////////////////////////////////
	      // Try to reduce the energy in the mesh
	      fEnergyNew = ppme[1]->EvolverOptimizeStep(m_fGradientPercent);
	      printf("Enew = %f > %f = Eold\n", fEnergyNew, fEnergyOld); 
	      while ( fEnergyNew >= fEnergyOld )
		{
		  m_fGradientPercent /= 2.0;
		  if ( m_fGradientPercent < cfMinGradientPercent )
		    {
		      ////////////////////////////////////////
		      // The mesh is not improving fast enough any more
		      break;
		    }
		  fEnergyNew = ppme[1]->EvolverOptimizeStep(m_fGradientPercent);
		}

	      //////////////////////////////////////////////////
	      // Check if the energy actually reduced
	      if ( m_fGradientPercent < cfMinGradientPercent )
		{
		  ////////////////////////////////////////
		  // The mesh is not improving fast enough any more
		  break;
		}
	      else
		{
		  printf("j = %ld : Grad = %f : Eold = %f > %f = Enew : Delta = %f\n", 
			 j, m_fGradientPercent, 
			 fEnergyOld, fEnergyNew, fEnergyOld - fEnergyNew);
		  pmeSwap = ppme[0];
		  ppme[0] = ppme[1];
		  ppme[1] = pmeSwap;
		  
		  fEnergyOld = fEnergyNew;
		}
	    }

	  ppme[0]->CalculateSubdivisionLimitPositionsAndNormalsSharpLoop();
	  m_ppQEMeshSubdivision[i] = ppme[0];
	  m_ppQEMeshSubdivision[i]->Preprocess();
	  //////////////////////////////
	  // Deallocate the extra mesh
	  DeallocateQEMesh(ppme[1]);
	}
    }
  m_pQEMesh = m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()];
  ASSERT( m_pQEMesh != NULL );
}

VOID CSLIDEEvolver::UpdateGraph()
{
  if ( !GetVisited() )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;
      BOOL bBoundChanged;

      if ( !GetBoundValid() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      if ( !m_bound.GetEmpty() )
		{
		  SetBoundChanged(TRUE);
		  m_bound.Null();
		}
	      return;
	    }
	  
	  bBoundChanged = FALSE;
	  
	  iter.Init(&m_listCSLIDEInstance);
	  
	  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
		pCSLIDEInstance != NULL;
		pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	    {
	      pCSLIDEInstance->UpdateGraph();
	      bBoundChanged = bBoundChanged || 
		pCSLIDEInstance->GetBoundChanged();
	    }
	  
	  UpdateQEMesh();
	  UpdateEvolver();

	  if ( bBoundChanged )
	    {
	      CalculateBound();
	    }
	  
	  SetBoundValid(TRUE);
	}

      SetVisited(TRUE);
    }
}

VOID CSLIDEEvolver::CalculateBound()
{
  CBound bound;

/*
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  
  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->InsertIntoBound(bound);
    }
*/
  ///////////////////////////////////////
  // START: Jordan Fix This!!!
  //
  {
    QEMeshIter iterQEMesh;
    QEVertex *pvx;
    Point pt;

    iterQEMesh.Init(m_pQEMesh);
    for ( pvx = iterQEMesh.PeekFirstVertex();
	  pvx != NULL;
	  pvx = iterQEMesh.PeekNextVertex() )
      {
	pt = pvx->GetPoint();
	bound.Insert(pt);
      }
  }
  //
  // END: Jordan Fix This!!!
  ///////////////////////////////////////

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);
    }
}

VOID CSLIDEEvolver::PrerenderGeometry(SLIDEShadingType eShading)
{
  CListIterInt iter;
  BOOL bFaceted, bLimitSurface, bCalculateNormals, bSelective;

  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  PushShadingAttributes(eShading);
  {
    m_pCSLIDESurface->Push();
    {
      if ( m_dyniSubdivisionLevel.GetValue() == 0 )
	{
	  ::RenderGeometry(m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()]);
	}
      else
	{
	  bLimitSurface = TRUE;
	  bCalculateNormals = TRUE;
	  bSelective = FALSE;
	  bFaceted = ( !(eShading == slideshtSLF_GOURAUD ||
			 eShading == slideshtSLF_PHONG) );
	  ::RenderGeometry(m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()], 
			   bFaceted, bLimitSurface, bCalculateNormals, bSelective);
	}
    }
    m_pCSLIDESurface->Pop();
  }
  PopShadingAttributes(eShading);

  if ( m_dyniDrawControlMesh.GetValue() )
    {
      // Initial Control Mesh
      QEMeshIter iterQEMesh;
      QEEdge *pe;

      PushAuxillaryGeometryAttributes();
      {	
	iterQEMesh.Init(m_ppQEMeshSubdivision[0]);
	gilBegin(GL_LINES);
	{
	  FLOAT fSharpness;

	  for ( pe = iterQEMesh.PeekFirstEdge();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdge() )
	    {
	      fSharpness = pe->GetKnotInterval();
	      if ( fSharpness == -1.0 )
		{
		  fSharpness = INFINITY;
		}
	      
	      if ( fSharpness == 0.0 )
		{
		  gilColor3f(1.0, 1.0, 0.0);
		}
	      else if ( fSharpness == INFINITY )
		{
		  gilColor3f(1.0, 0.0, 0.0);
		}
	      else
		{
		  gilColor3f(0.0, 1.0, 0.0);
		}
	      gilVertex3fv(pe->GetVertex()->GetPoint().Get());
	      gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
	    }
	  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdgeBoundary() )
	    {
	      fSharpness = pe->GetKnotInterval();
	      if ( fSharpness == -1.0 )
		{
		  fSharpness = INFINITY;
		}
	      
	      if ( fSharpness == 0.0 )
		{
		  gilColor3f(1.0, 1.0, 0.0);
		    }
	      else if ( fSharpness == INFINITY )
		{
		  gilColor3f(1.0, 0.0, 0.0);
		}
	      else
		{
		  gilColor3f(0.0, 1.0, 0.0);
		}
	      gilVertex3fv(pe->GetVertex()->GetPoint().Get());
	      gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
	    }
	}
	gilEnd();
      }
      PopAuxillaryGeometryAttributes();
    }

  if ( m_dyniDrawCurrentMesh.GetValue() )
    {
      // Current Control Mesh
      QEMeshIter iterQEMesh;
      QEEdge *pe;

      PushAuxillaryGeometryAttributes();
      {	
	iterQEMesh.Init(m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()]);
	gilBegin(GL_LINES);
	{
	  FLOAT fSharpness;

	  for ( pe = iterQEMesh.PeekFirstEdge();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdge() )
	    {
	      fSharpness = pe->GetKnotInterval();
	      if ( fSharpness == -1.0 )
		{
		  fSharpness = INFINITY;
		}

	      if ( fSharpness == 0.0 )
		{
		  gilColor3f(0.0, 1.0, 1.0);
		}
	      else if ( fSharpness == INFINITY )
		{
		  gilColor3f(1.0, 0.0, 0.0);
		}
	      else
		{
		  gilColor3f(0.0, 1.0, 0.0);
		}
	      gilVertex3fv(pe->GetVertex()->GetPoint().Get());
	      gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
	    }
	  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdgeBoundary() )
	    {
	      fSharpness = pe->GetKnotInterval();
	      if ( fSharpness == -1.0 )
		{
		  fSharpness = INFINITY;
		}
	      
	      if ( fSharpness == 0.0 )
		{
		  gilColor3f(1.0, 1.0, 0.0);
		}
	      else if ( fSharpness == INFINITY )
		{
		  gilColor3f(1.0, 0.0, 0.0);
		}
	      else
		{
		  gilColor3f(0.0, 1.0, 0.0);
		}
	      gilVertex3fv(pe->GetVertex()->GetPoint().Get());
	      gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
	    }
	}
	gilEnd();
      }
      PopAuxillaryGeometryAttributes();
    }
}

VOID CSLIDEEvolver::PrerenderNormals(SLIDEShadingType eShading)
{
  CListIterInt iter;

  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  ::RenderNormals(m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()]);
}

VOID CSLIDEEvolver::PrerenderBounds()
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
    {
    }
  else 
    {
      if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	{
	}
      else
	{
	  iter.Init(&m_listCSLIDEInstance);
	  
	  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
		pCSLIDEInstance != NULL;
		pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	    {
	      pCSLIDEInstance->PrerenderBounds();
	    }
	}

      m_bound.Render();
    }
}

VOID CSLIDEEvolver::Prerender()
{
  if ( !(!m_dyniLOD.GetDynamic() && 
	 ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
	   m_dyniLOD.GetValue() == slidelodtSLF_BOUND )) )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;
      
      iter.Init(&m_listCSLIDEInstance);
      
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pCSLIDEInstance->Prerender();
	}
    }
      
  CSLIDEBoundedNode::Prerender();
}

VOID CSLIDEEvolver::RenderGeometry(const Matrix &mGproj_obj, 
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD,
				CSLIDEWormwarp *pCSLIDEWarp/* = NULL*/)
{

  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  BOOL bFaceted, bLimitSurface, bCalculateNormals, bSelective;

  // For Debugging
  if ( GetRenderCDAOperandsEnabled() )
    {
      gilPushAttrib(GL_ALL_ATTRIB_BITS);
      {
	PushAuxillaryGeometryAttributes();
	{
	  PushWireFrameAttributes();
	  {
	    iter.Init(&m_listCSLIDEInstance);
	    
	    for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
		pCSLIDEInstance != NULL;
		  pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	      {
		pCSLIDEInstance->Render(mGproj_obj, pCBoundClip, slideshtSLF_WIRE, eLOD);
	      }
	  }
	  PopWireFrameAttributes();
	}
	PopAuxillaryGeometryAttributes();
      }
      gilPopAttrib();
    }
      
  //if ( m_dyniSubdivisionLevel.GetValue() == 0 )
  //{
  //  ::RenderGeometry(m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()]);
  //}
  //else
  //{
      bLimitSurface = TRUE;
      bCalculateNormals = TRUE;
      bSelective = FALSE;
      bFaceted = ( !(eShading == slideshtSLF_GOURAUD ||
		     eShading == slideshtSLF_PHONG) );
      ::RenderGeometry(m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()], 
		       bFaceted, bLimitSurface, bCalculateNormals, bSelective);
  //}
  
  if ( m_dyniDrawControlMesh.GetValue() )
    {
      // Initial Control Mesh
      QEMeshIter iterQEMesh;
      QEEdge *pe;

      PushAuxillaryGeometryAttributes();
      {	
	iterQEMesh.Init(m_ppQEMeshSubdivision[0]);
	gilBegin(GL_LINES);
	{
	  FLOAT fSharpness;

	  for ( pe = iterQEMesh.PeekFirstEdge();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdge() )
	    {
	      fSharpness = pe->GetKnotInterval();
	      if ( fSharpness == -1.0 )
		{
		  fSharpness = INFINITY;
		}
	      
	      if ( fSharpness == 0.0 )
		{
		  gilColor3f(1.0, 1.0, 0.0);
		}
	      else if ( fSharpness == INFINITY )
		{
		  gilColor3f(1.0, 0.0, 0.0);
		}
	      else
		{
		  gilColor3f(0.0, 1.0, 0.0);
		}
	      gilVertex3fv(pe->GetVertex()->GetPoint().Get());
	      gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
	    }
	  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdgeBoundary() )
	    {
	      fSharpness = pe->GetKnotInterval();
	      if ( fSharpness == -1.0 )
		{
		  fSharpness = INFINITY;
		}
	      
	      if ( fSharpness == 0.0 )
		{
		  gilColor3f(1.0, 1.0, 0.0);
		    }
	      else if ( fSharpness == INFINITY )
		{
		  gilColor3f(1.0, 0.0, 0.0);
		}
	      else
		{
		  gilColor3f(0.0, 1.0, 0.0);
		}
	      gilVertex3fv(pe->GetVertex()->GetPoint().Get());
	      gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
	    }
	}
	gilEnd();
      }
      PopAuxillaryGeometryAttributes();
    }

  if ( m_dyniDrawCurrentMesh.GetValue() )
    {
      // Current Control Mesh
      QEMeshIter iterQEMesh;
      QEEdge *pe;

      PushAuxillaryGeometryAttributes();
      {	
	iterQEMesh.Init(m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()]);
	gilBegin(GL_LINES);
	{
	  FLOAT fSharpness;

	  for ( pe = iterQEMesh.PeekFirstEdge();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdge() )
	    {
	      fSharpness = pe->GetKnotInterval();
	      if ( fSharpness == -1.0 )
		{
		  fSharpness = INFINITY;
		}

	      if ( fSharpness == 0.0 )
		{
		  gilColor3f(0.0, 1.0, 1.0);
		}
	      else if ( fSharpness == INFINITY )
		{
		  gilColor3f(1.0, 0.0, 0.0);
		}
	      else
		{
		  gilColor3f(0.0, 1.0, 0.0);
		}
	      gilVertex3fv(pe->GetVertex()->GetPoint().Get());
	      gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
	    }
	  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdgeBoundary() )
	    {
	      fSharpness = pe->GetKnotInterval();
	      if ( fSharpness == -1.0 )
		{
		  fSharpness = INFINITY;
		}
	      
	      if ( fSharpness == 0.0 )
		{
		  gilColor3f(1.0, 1.0, 0.0);
		}
	      else if ( fSharpness == INFINITY )
		{
		  gilColor3f(1.0, 0.0, 0.0);
		}
	      else
		{
		  gilColor3f(0.0, 1.0, 0.0);
		}
	      gilVertex3fv(pe->GetVertex()->GetPoint().Get());
	      gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
	    }
	}
	gilEnd();
      }
      PopAuxillaryGeometryAttributes();
    }
}

VOID CSLIDEEvolver::RenderNormals(SLIDEShadingType eShading)
{
  BOOL bFaceted, bLimitSurface, bCalculateNormals, bSelective;

  PushAuxillaryGeometryAttributes();
  {
    gilColor3f(1.0, 0.0, 1.0);
    
    bLimitSurface = TRUE;
    bCalculateNormals = TRUE;
    bSelective = FALSE;
    bFaceted = ( !(eShading == slideshtSLF_GOURAUD ||
		   eShading == slideshtSLF_PHONG) );
    ::RenderNormals(m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()],
		    bFaceted, bLimitSurface, bCalculateNormals, bSelective);
  }
  PopAuxillaryGeometryAttributes();
}

VOID CSLIDEEvolver::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtEvolver], m_pcID);

  WriteSLIDEFlags(fpOutputFile, 0);

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->WriteSLIDE(fpOutputFile);    
    }

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndEvolver]);
}

VOID CSLIDEEvolver::WriteRIBGeometry(FILE *fpOutputFile,
				   const Matrix &mGproj_obj, 
				   CBound *pCBoundClip, 
				   SLIDEShadingType eShading,
				   SLIDELODType eLOD)
{
/*
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( fpOutputFile != NULL );

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->WriteRIB(fpOutputFile,
				mGproj_obj, pCBoundClip, eShading, eLOD);    
    }
    */
}

VOID CSLIDEEvolver::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );
  /*
  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE EVOLVER: %s", m_pcID);
  WriteRIBBegin(fpOutputFile);
  */
}

VOID CSLIDEEvolver::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );
  /*
  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE EVOLVER: %s", m_pcID);
  */
}

VOID CSLIDEEvolver::WriteWRLZ(FILE *fpOutputFile, const Matrix &m, 
			     CSLIDESurface *pCSLIDESurface)
{
  ::WriteWRLZ(m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()], fpOutputFile, m);
}

VOID CSLIDEEvolver::WriteSTL(FILE *fpOutputFile, const Matrix &m)
{
  ::WriteSTL(m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()], fpOutputFile, m);
}






