#include "slideface.h"
#include "slideobject.h"
#include "slideparser.h"

#include "slideworld.h"
#include "slidepoint.h"
#include "slidesurface.h"
#include "stl.h"

#ifdef DT_TRIANGULATE_TEST
#include "dtvertex.h"
//#include "dtface.h"
#include "dttriangle.h"
#endif // DT_TRIANGULATE_TEST

//////////////////////////////////////////////////////////////////////////////
// CSLIDEFace Class
//

CSLIDEFace::CSLIDEFace()
{
  m_pcSurfaceID = NULL;
  m_pCSLIDESurface = NULL;
  m_plistPointIDs = NULL;

  m_listCSLIDEParentObject.Init();

  m_uiPrerenderAttributesStackDepthMax = 0;

#ifdef DT_TRIANGULATE_TEST
  //  m_pDTFace = NULL;
#endif // DT_TRIANGULATE_TEST
}

CSLIDEFace::~CSLIDEFace()
{
}

VOID CSLIDEFace::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID, 
		      CListInt *plistPointIDs)
{
  CSLIDEPrimitive::Init(pcFile, uiLine, uiColumn, pcID);

  ASSERT( plistPointIDs != NULL );

  m_pcSurfaceID = gcppcSLIDEReservedWords[sliderwtSLF_INHERIT];
  m_pCSLIDESurface = NULL;

  m_plistPointIDs = plistPointIDs;
  m_listCSLIDEPoint.Init();

  SetAutoNormals(FALSE);
  SetAutoNormalsDynamic(FALSE);

  m_uiPrerenderAttributesStackDepthMax = 0;

#ifdef DT_TRIANGULATE_TEST
  /*
  m_pDTFace = AllocateDTFace();
  ASSERT( m_pDTFace != NULL );
  m_pDTFace->Init();
  m_listDTTriangles.Init();
  */
#endif // DT_TRIANGULATE_TEST
}

VOID CSLIDEFace::Uninit()
{
  CSLIDEPrimitive::Uninit();

  if ( m_plistPointIDs != NULL )
    {
      delete m_plistPointIDs;
      m_plistPointIDs = NULL;
    }
}

BOOL CSLIDEFace::Initialized()
{
  if ( !CSLIDEPrimitive::Initialized() )
    {
      return FALSE;
    }

  return ( m_pcSurfaceID != NULL &&
	   m_plistPointIDs != NULL );
}

VOID CSLIDEFace::InsertParentObject(CSLIDEObject *pCSLIDEObject)
{
  ASSERT( pCSLIDEObject != NULL );

  m_listCSLIDEParentObject.InsertLast(pCSLIDEObject, (UINT_32)pCSLIDEObject);
}

VOID CSLIDEFace::RemoveParentObject(CSLIDEObject *pCSLIDEObjectToRemove)
{
  ASSERT( pCSLIDEObjectToRemove != NULL );

  CListIterInt iter;
  CSLIDEObject *pCSLIDEObject;
  CHAR *pcID;

  pcID = pCSLIDEObjectToRemove->GetID();

  iter.Init(&m_listCSLIDEParentObject);
  for ( pCSLIDEObject = (CSLIDEObject*)iter.PeekFirst();
	pCSLIDEObject != NULL; )
    {
      if( (pCSLIDEObject->GetID()) == pcID )
	{
	  pCSLIDEObject = (CSLIDEObject*)iter.PeekNextRemoveCurr();
	}
      else
	{
	  pCSLIDEObject = (CSLIDEObject*)iter.PeekNext();
	}
    }
}

BOOL CSLIDEFace::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;
  CListIterInt iter;
  CHAR *pcID;
  CSLIDEPoint *pCSLIDEPoint;

  if ( !CSLIDEPrimitive::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  // Check Surface Reference

  m_pCSLIDESurface = pCSLIDEWorld->PeekSurface(m_pcSurfaceID);
  if ( m_pCSLIDESurface == NULL )
    {
      HandleError("Undefined %s \"%s\"", 
		  gcppcSLIDEReservedWords[sliderwtSurface],
		  m_pcSurfaceID);
      bCorrect = FALSE;
    }

  // Check Point References
  iter.Init(m_plistPointIDs);
  for ( pcID = (CHAR*)(iter.PeekFirst());
	pcID != NULL;
	pcID = (CHAR*)(iter.PeekNext()) )
    {
      pCSLIDEPoint = pCSLIDEWorld->PeekPoint(pcID);
      if ( pCSLIDEPoint == NULL )
	{
	  HandleError("Undefined %s \"%s\"", gcppcSLIDEReservedWords[sliderwtPoint], 
		      pcID);
	  bCorrect = FALSE;
	}
      else
	{
	  m_listCSLIDEPoint.InsertLast(pCSLIDEPoint, 0);
	  pCSLIDEPoint->InsertParentFace(this);
	}
    }

  return bCorrect;
}

BOOL CSLIDEFace::Valid()
{
  if ( !CSLIDEPrimitive::Valid() )
    {
      return FALSE;
    }

  return TRUE;
  //return (GetPointCount() >= 3);
}

VOID CSLIDEFace::Remove(CSLIDEWorld *pCSLIDEWorld)
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;
  CSLIDEObject *pCSLIDEObject;
  
  //Remove from parent
  iter.Init(&m_listCSLIDEParentObject);
  for ( pCSLIDEObject = (CSLIDEObject*)iter.PeekFirst();
	pCSLIDEObject != NULL;
	pCSLIDEObject = (CSLIDEObject*)iter.PeekNext() )
    {
      pCSLIDEObject->RemoveFace(this);
    }

  // Remove from child
  iter.Init(&m_listCSLIDEPoint);
  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      pCSLIDEPoint->RemoveParentFace(this);
    }

  // Remove from world
  pCSLIDEWorld->RemoveFace(this);

  // This deletes the points in the faces
  //Uninit();
}

VOID CSLIDEFace::RemovePoint(CSLIDEPoint *pCSLIDEPointToRemove)
{
  CListIterInt iter;
  CHAR *pcID, *pcPointIDToRemove;
  CSLIDEPoint *pCSLIDEPoint;

  pcPointIDToRemove = pCSLIDEPointToRemove->GetID();

  iter.Init(m_plistPointIDs);
  for ( pcID = (CHAR*)(iter.PeekFirst());
	pcID != NULL;
	pcID = (CHAR*)(iter.PeekNext()) )
    {
      if( pcID == pcPointIDToRemove )
	{
	  iter.PeekNextRemoveCurr();
	  break;
	}
    }

  iter.Init(&m_listCSLIDEPoint);
  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      if( (pCSLIDEPoint->GetID()) == pcPointIDToRemove )
	{
	  iter.PeekNextRemoveCurr();
	  break;
	}
    }
}

VOID CSLIDEFace::InsertIntoBound(CBound *pCBound)
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;

  ASSERT( pCBound != NULL );

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      pCSLIDEPoint->InsertIntoBound(pCBound);
    }
}

VOID CSLIDEFace::Preprocess()
{
  CheckDynamic();
  UpdateState();

  // All faces must calculate their normal at least once

  CalculateNormal();
  CalculateCentroid();
  
  // Calculate the D coefficient of the Tangent Plane defined by the normal
  m_vNormal.CompletePlaneEquation(m_ptCentroid);

#ifdef DT_TRIANGULATE_TEST
/*
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;
  DTVertex *pvx;
  
  iter.Init(&m_listCSLIDEPoint);
      
  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      pvx = AllocateDTVertex();
      ASSERT( pvx != NULL );
      pvx->Init(pCSLIDEPoint->GetPoint());
      m_pDTFace->AddVertex(pvx);
    }
  m_pDTFace->Preprocess();
  //m_pDTFace->Triangulate(&m_listDTTriangles);
*/
#endif // DT_TRIANGULATE_TEST
}

// This should only be called by objects which can be smooth shaded
VOID CSLIDEFace::PreprocessAutoNormals()
{
  if ( !GetAutoNormals() )
    {
      // We do not contribute to any vertex AutoNormals because we are not
      // part of any smooth shaded objects
      return;
    }

  // Calculate Point Normals
  if ( m_listCSLIDEPoint.GetLength() >= 3 )
    {
      CListIterInt iter;
      CSLIDEPoint *pgptPrev, *pgptCurr, *pgptNext;
      Vector u, v;
      FLOAT fTheta;
      
      iter.Init(&m_listCSLIDEPoint);
	  
      pgptPrev = (CSLIDEPoint*)m_listCSLIDEPoint.PeekLast();
      ASSERT( pgptPrev != NULL );
      pgptCurr = (CSLIDEPoint*)iter.PeekFirst();
      ASSERT( pgptCurr != NULL );
	  
      for ( pgptNext = (CSLIDEPoint*)iter.PeekNext(); 
	    pgptNext != NULL;
	    pgptNext = (CSLIDEPoint*)iter.PeekNext() )
	{
	  // Check if the point had its normal specified in the SLIDE file
	  if ( pgptCurr->GetAutoNormal() )
	    {
	      // Calculate the angle at the vertex
	      u = pgptNext->GetPoint() - pgptCurr->GetPoint();
	      v = pgptPrev->GetPoint() - pgptCurr->GetPoint();
	      
	      fTheta = (FLOAT)AngleBetween(u, v);
	      pgptCurr->AddFaceNormal(m_vNormal, fTheta);
	      if ( pgptPrev->GetBoundDynamic() ||
		   pgptCurr->GetBoundDynamic() ||
		   pgptNext->GetBoundDynamic() )
		{
		  pgptCurr->SetAutoNormalDynamic(TRUE);
		}
	    }
	  
	  pgptPrev = pgptCurr;
	  pgptCurr = pgptNext;
	}
      
      pgptNext = (CSLIDEPoint*)m_listCSLIDEPoint.PeekFirst();
      ASSERT( pgptNext != NULL );

      if ( pgptCurr->GetAutoNormal() )
	{
	  // Calculate the angle at the vertex
	  u = pgptNext->GetPoint() - pgptCurr->GetPoint();
	  v = pgptPrev->GetPoint() - pgptCurr->GetPoint();
	  
	  fTheta = (FLOAT)AngleBetween(u, v);
	  pgptCurr->AddFaceNormal(m_vNormal, fTheta);
	  if ( pgptPrev->GetBoundDynamic() ||
	       pgptCurr->GetBoundDynamic() ||
	       pgptNext->GetBoundDynamic() )
	    {
	      pgptCurr->SetAutoNormalDynamic(TRUE);
	    }
	}
    }
}

// This should only be called by objects which can be smooth shaded
// Check to see if any of the face's points has dynamic auto normals
// in which case we must remember to resend our normal to it on every update
VOID CSLIDEFace::CheckAutoNormalsDynamic()
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      if ( pCSLIDEPoint->GetAutoNormalDynamic() )
	{
	  SetAutoNormalsDynamic(TRUE);
	  SetNormalsDynamic(TRUE);
	  SetDynamic(TRUE);
	  break;
	}
    }
}

VOID CSLIDEFace::SetBoundUnvisited()
{
  CListIterInt iter;
  CSLIDEObject *pCSLIDEObject;

  iter.Init(&m_listCSLIDEParentObject);
  for ( pCSLIDEObject = (CSLIDEObject*)iter.PeekFirst();
	pCSLIDEObject != NULL;
	pCSLIDEObject = (CSLIDEObject*)iter.PeekNext() )
    {
      pCSLIDEObject->SetBoundUnvisited(); 
    }
}

VOID CSLIDEFace::SetBoundUnchanged()
{
  CListIterInt iter;
  CSLIDEObject *pCSLIDEObject;

  iter.Init(&m_listCSLIDEParentObject);
  for ( pCSLIDEObject = (CSLIDEObject*)iter.PeekFirst();
	pCSLIDEObject != NULL;
	pCSLIDEObject = (CSLIDEObject*)iter.PeekNext() )
    {
      pCSLIDEObject->SetBoundUnchanged();
    }
}

VOID CSLIDEFace::SetPrerenderUnvisited()
{
  CListIterInt iter;
  CSLIDEObject *pCSLIDEObject;

  iter.Init(&m_listCSLIDEParentObject);
  for ( pCSLIDEObject = (CSLIDEObject*)iter.PeekFirst();
	pCSLIDEObject != NULL;
	pCSLIDEObject = (CSLIDEObject*)iter.PeekNext() )
    {
      pCSLIDEObject->SetPrerenderUnvisited(); 
    }
}

#ifdef __USE_CDA__
VOID CSLIDEFace::CalculateCDAVertices(CListInt *plistCDAVertex, 
				      const Matrix &mGroot_obj)
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      pCSLIDEPoint->CalculateCDAVertices(plistCDAVertex, mGroot_obj);
    }
}
#endif // __USE_CDA__

VOID CSLIDEFace::CheckDynamic()
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;
  BOOL bBoundEmpty;
  BOOL bBoundDynamic;
  BOOL bGeometryDynamic;
  BOOL bNormalsDynamic;

  bBoundEmpty = TRUE;
  bBoundDynamic = FALSE;
  bGeometryDynamic = FALSE;
  bNormalsDynamic = FALSE;

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      bBoundEmpty = FALSE;
      bBoundDynamic = bBoundDynamic || pCSLIDEPoint->GetBoundDynamic();
      bGeometryDynamic = bGeometryDynamic || pCSLIDEPoint->GetGeometryDynamic();
      bNormalsDynamic = bNormalsDynamic || pCSLIDEPoint->GetNormalDynamic();
    }

  SetBoundEmpty(bBoundEmpty);
  SetBoundDynamic(bBoundDynamic);
  SetGeometryDynamic(bGeometryDynamic ||
		     m_pCSLIDESurface->GetDynamic());
  SetNormalsDynamic(bNormalsDynamic);
  SetDynamic(bBoundDynamic);
  SetAutoNormalsDynamic(FALSE);
}

VOID CSLIDEFace::DrawGUIGraphPointReferences(Tcl_Interp *pInterp)
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;
  CHAR pcBuffer[200];
  INT_32 iCode;

  iter.Init(&m_listCSLIDEPoint);
  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      sprintf(pcBuffer, "Graph:CreateReference %s %s", GetID(), pCSLIDEPoint->GetID());
      iCode = Tcl_Eval(pInterp, pcBuffer);
      ASSERT( iCode == TCL_OK );
    }
}

VOID CSLIDEFace::UpdateState()
{
  if ( GetBoundDynamic() )
    {
      CalculateNormal();
      CalculateCentroid();

      // Calculate the D coefficient of the Tangent Plane defined by the normal
      m_vNormal.CompletePlaneEquation(m_ptCentroid);

#ifdef DT_TRIANGULATE_TEST
/*
      CListIterInt iter;
      CSLIDEPoint *pCSLIDEPoint;
      DTVertex *pvx;
      
      iter.Init(&m_listCSLIDEPoint);
      
      for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	    pCSLIDEPoint != NULL;
	    pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
	{
	  pvx = AllocateDTVertex();
	  ASSERT( pvx != NULL );
	  pvx->Init(pCSLIDEPoint->GetPoint());
	  m_pDTFace->AddVertex(pvx);
	}
      m_pDTFace->Preprocess();
      //m_pDTFace->Triangulate(&m_listDTTriangles);
      */
#endif // DT_TRIANGULATE_TEST
    }

  if ( GetAutoNormalsDynamic() )
    {
      // Calculate Point Normals
      if ( m_listCSLIDEPoint.GetLength() >= 3 )
	{
	  CListIterInt iter;
	  CSLIDEPoint *pgptPrev, *pgptCurr, *pgptNext;
	  Vector u, v;
	  FLOAT fTheta;
	  
	  iter.Init(&m_listCSLIDEPoint);
	  
	  pgptPrev = (CSLIDEPoint*)m_listCSLIDEPoint.PeekLast();
	  ASSERT( pgptPrev != NULL );
	  pgptCurr = (CSLIDEPoint*)iter.PeekFirst();
	  ASSERT( pgptCurr != NULL );
	  
	  for ( pgptNext = (CSLIDEPoint*)iter.PeekNext(); 
		pgptNext != NULL;
		pgptNext = (CSLIDEPoint*)iter.PeekNext() )
	    {
	      if ( pgptCurr->GetAutoNormalDynamic() )
		{
		  // Calculate the angle at the vertex
		  u = pgptNext->GetPoint() - pgptCurr->GetPoint();
		  v = pgptPrev->GetPoint() - pgptCurr->GetPoint();
		  
		  fTheta = (FLOAT)AngleBetween(u, v);
		  pgptCurr->AddFaceNormal(m_vNormal, fTheta);
		}
		  
	      pgptPrev = pgptCurr;
	      pgptCurr = pgptNext;
	    }

	  pgptNext = (CSLIDEPoint*)m_listCSLIDEPoint.PeekFirst();
	  ASSERT( pgptNext != NULL );

	  if ( pgptCurr->GetAutoNormalDynamic() )
	    {
	      // Calculate the angle at the vertex
	      u = pgptNext->GetPoint() - pgptCurr->GetPoint();
	      v = pgptPrev->GetPoint() - pgptCurr->GetPoint();
	      
	      fTheta = (FLOAT)AngleBetween(u, v);
	      pgptCurr->AddFaceNormal(m_vNormal, fTheta);
	    }
	}
    }
}

// This uses Newell's Algorithm for finding the best fit planar equation
// for a set of points in 3 space (pg. 476 Foley and van Dam)
// I modified it so that the normal would be positive for counter-clockwise
// polygons, and I removed the 1/2 coefficient because we are concerned 
// with relative areas, not absolute areas.

VOID CSLIDEFace::CalculateNormal()
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;
  Point ptStart, ptPrev, ptCurr;

  m_vNormal.Null();

  iter.Init(&m_listCSLIDEPoint);

  pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
  if ( pCSLIDEPoint == NULL )
    {
      HandleWarning("Null normal because no points for face");
      return;
    }
  ptStart = pCSLIDEPoint->GetPoint();
  ptPrev = ptStart;

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      ptCurr = pCSLIDEPoint->GetPoint();

      m_vNormal[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
      m_vNormal[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
      m_vNormal[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);

      ptPrev = ptCurr;
    }

  ptCurr = ptStart;
  m_vNormal[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
  m_vNormal[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
  m_vNormal[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);
  
  if ( !m_vNormal.Normalize() )
    {
      HandleWarning("Null normal because of zero area face");
    }
}

VOID CSLIDEFace::CalculateCentroid()
{
  CListIterInt iter;
  UINT_32 i;
  CSLIDEPoint *pCSLIDEPoint;

  iter.Init(&m_listCSLIDEPoint);

  m_ptCentroid.Null();

  i = 0;
  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst();
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      m_ptCentroid = m_ptCentroid + pCSLIDEPoint->GetPoint();
      i++;
    }

  if ( i > 0 )
    {
      m_ptCentroid = m_ptCentroid / (FLOAT)i;
    }

  m_ptCentroid[W] = 1.0;
}

VOID CSLIDEFace::Prerender(SLIDEShadingType eShading)
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;

  GLbitfield GLbAttributesMask;
  
  GLbAttributesMask = m_pCSLIDESurface->CalculateAttributesMask();
  
  PushAttributes(GLbAttributesMask);
  {
    m_pCSLIDESurface->Push();
    {

#ifndef DT_TRIANGULATE_TEST

      gilLineWidth(2.0);
      gilBegin(GL_POLYGON);
      {
	gilNormal3fv(m_vNormal.Get());
	
	m_uiPrerenderAttributesStackDepthMax = 0;

	iter.Init(&m_listCSLIDEPoint);
	
	for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	      pCSLIDEPoint != NULL;
	      pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
	  {
	    pCSLIDEPoint->Render(eShading == slideshtSLF_GOURAUD ||
				 eShading == slideshtSLF_PHONG);
	    m_uiPrerenderAttributesStackDepthMax = 
	      MAX(m_uiPrerenderAttributesStackDepthMax,
		  pCSLIDEPoint->GetPrerenderAttributesStackDepthMax());
	  }

	if ( GLbAttributesMask != 0x0 )
	  {
	    m_uiPrerenderAttributesStackDepthMax++;
	  }
      }
      gilEnd();
      gilLineWidth(1.0);

#else // DT_TRIANGULATE_TEST
      /*
      //m_pDTFace->Render();
      DTTriangle *pDTTriangle;
      iter.Init(&m_listDTTriangles);
      for ( pDTTriangle = (DTTriangle*)(iter.PeekFirst());
	    pDTTriangle != NULL;
	    pDTTriangle = (DTTriangle*)(iter.PeekNext()) )
	{
	  pDTTriangle->Render();
	}
	*/
#endif // DT_TRIANGULATE_TEST

    }
    m_pCSLIDESurface->Pop();
  }
  PopAttributes(GLbAttributesMask);
}

VOID CSLIDEFace::Render(SLIDEShadingType eShading, 
			CSLIDEWormwarp * pCSLIDEWarp /*= NULL*/)
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;

  GLbitfield GLbAttributesMask;
  
  GLbAttributesMask = m_pCSLIDESurface->CalculateAttributesMask();
  
  PushAttributes(GLbAttributesMask);
  {
    m_pCSLIDESurface->Push();
    {

#ifndef DT_TRIANGULATE_TEST

      gilBegin(GL_POLYGON);
      {
	if (NULL == pCSLIDEWarp)
	  {
	    gilNormal3fv(m_vNormal.Get());
	  }
	else
	  {
	    UINT_32 i;
	    Vector v1, v2, vNorm;
	    Point pTri[3];

	    iter.Init(&m_listCSLIDEPoint);
	
	    for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(), i = 0;
		  pCSLIDEPoint != NULL;
		  pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext(), i++ )
	      {
		pCSLIDEPoint->CalculatePointWarped( pCSLIDEWarp );
		if ( 3 > i )
		  {
		    pTri[i] = pCSLIDEPoint->GetPointWarped();
		  }
	      }
	    v1.Set( pTri[1](X) - pTri[0](X),
		    pTri[1](Y) - pTri[0](Y),
		    pTri[1](Z) - pTri[0](Z));
	    v2.Set( pTri[2](X) - pTri[0](X),
		    pTri[2](Y) - pTri[0](Y),
		    pTri[2](Z) - pTri[0](Z));
	    vNorm = v1.Cross(v2);
	    gilNormal3fv(vNorm.Get());
	  }
	
	iter.Init(&m_listCSLIDEPoint);
	
	for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	      pCSLIDEPoint != NULL;
	      pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
	  {
	    pCSLIDEPoint->Render(eShading == slideshtSLF_GOURAUD ||
				 eShading == slideshtSLF_PHONG,
				 pCSLIDEWarp);
	  }
      }
      gilEnd();

#else // DT_TRIANGULATE_TEST
      /*
      //m_pDTFace->Render();
      DTTriangle *pDTTriangle;
      iter.Init(&m_listDTTriangles);
      for ( pDTTriangle = (DTTriangle*)(iter.PeekFirst());
	    pDTTriangle != NULL;
	    pDTTriangle = (DTTriangle*)(iter.PeekNext()) )
	{
	  pDTTriangle->Render();
	}
	*/
#endif // DT_TRIANGULATE_TEST
    }
    m_pCSLIDESurface->Pop();
  }
  PopAttributes(GLbAttributesMask);
}

VOID CSLIDEFace::RenderSelect()
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;

#ifndef DT_TRIANGULATE_TEST
  
  gilBegin(GL_POLYGON);
  {
    gilNormal3fv(m_vNormal.Get());
    
    iter.Init(&m_listCSLIDEPoint);
    
    for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	  pCSLIDEPoint != NULL;
	  pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
      {
	pCSLIDEPoint->RenderSelect();
      }
  }
  gilEnd();
  
#else // DT_TRIANGULATE_TEST
  /*
    //m_pDTFace->Render();
    DTTriangle *pDTTriangle;
    iter.Init(&m_listDTTriangles);
    for ( pDTTriangle = (DTTriangle*)(iter.PeekFirst());
    pDTTriangle != NULL;
    pDTTriangle = (DTTriangle*)(iter.PeekNext()) )
    {
    pDTTriangle->Render();
    }
    */
#endif // DT_TRIANGULATE_TEST
}

VOID CSLIDEFace::RenderNormals(SLIDEShadingType eShading)
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;

  switch ( eShading )
    {
    case slideshtSLF_INHERIT:
    case slideshtSLF_FLAT:
      { 
	gilBegin(GL_LINES);
	{
	  Vector v;
	  Point pt;
	  
	  gilVertex3fv(m_ptCentroid.Get());
	  v = m_vNormal;
	  v[W] = 0.0;
	  pt = m_ptCentroid + v;
	  gilVertex3fv(pt.Get());
	}
	gilEnd();
      }
      break;

    case slideshtSLF_WIRE:
    case slideshtSLF_GOURAUD:
    case slideshtSLF_PHONG:
      {
	gilBegin(GL_LINES);
	{
	  iter.Init(&m_listCSLIDEPoint);
	  
	  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
		pCSLIDEPoint != NULL;
		pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
	    {
	      pCSLIDEPoint->RenderNormal();
	    }
	}
	gilEnd();
      }
      break;

    default:
      ASSERT( FALSE );
      break;
    }
}

VOID CSLIDEFace::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s ( ", gcppcSLIDEReservedWords[sliderwtFace], 
	  m_pcGUIID);

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      fprintf(fpOutputFile, "%s ", pCSLIDEPoint->GetGUIID());      
    }

  fprintf(fpOutputFile, ")\n");

  if ( m_pcSurfaceID != gcppcSLIDEReservedWords[sliderwtSLF_INHERIT] )
    {
      fprintf(fpOutputFile, "  %s %s\n", 
	      gcppcSLIDEReservedWords[sliderwtSurface], 
	      m_pcSurfaceID);
    } 
  CSLIDEPrimitive::WriteSLIDEFlags(fpOutputFile, 0);

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndFace]);
}

VOID CSLIDEFace::WriteRIB(FILE *fpOutputFile, SLIDEShadingType eShading)
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;

  UINT_32 i, uiPoints;
  RtPoint *prtPoints, *prtNormals, *prtColors, rtPlaneNormal;
  RtFloat *prtTextureCoords;

  ASSERT( fpOutputFile != NULL );

  uiPoints = m_listCSLIDEPoint.GetLength();
  prtPoints = new RtPoint[uiPoints];
  ASSERT( prtPoints != NULL );
  prtNormals = new RtPoint[uiPoints];
  ASSERT( prtNormals != NULL );
  prtColors = new RtPoint[uiPoints];
  ASSERT( prtColors != NULL );
  prtTextureCoords = new RtFloat[2*uiPoints];
  ASSERT( prtTextureCoords != NULL );

  rtPlaneNormal[0] = m_vNormal[0];
  rtPlaneNormal[1] = m_vNormal[1];
  rtPlaneNormal[2] = m_vNormal[2];

  iter.Init(&m_listCSLIDEPoint);

  for ( i = 0; i < uiPoints; i++ )
    {
      prtColors[i][0] = m_pCSLIDESurface->GetColor()(0);
      prtColors[i][1] = m_pCSLIDESurface->GetColor()(1);
      prtColors[i][2] = m_pCSLIDESurface->GetColor()(2);
    }

  i = 0;
  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      ASSERT( i < uiPoints );

      pCSLIDEPoint->WriteRIB(&prtPoints[i], 
			     &prtNormals[i], 
			     &prtColors[i], 
			     &prtTextureCoords[2*i]);

      if ( !(eShading == slideshtSLF_GOURAUD ||
	     eShading == slideshtSLF_PHONG) )
	{
	  memcpy(prtNormals[i], rtPlaneNormal, sizeof(RtPoint));
	}

      i++;
    }

  if ( eShading != slideshtSLF_GOURAUD &&
       eShading != slideshtSLF_PHONG )
    {
      for ( i = 0; i < uiPoints; i++ )
	{
	  prtColors[i][0] = m_pCSLIDESurface->GetColor()(0);
	  prtColors[i][1] = m_pCSLIDESurface->GetColor()(1);
	  prtColors[i][2] = m_pCSLIDESurface->GetColor()(2);
	}
    }

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE FACE: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);

  m_pCSLIDESurface->WriteRIBPush(fpOutputFile);
  {    
    RiPolygon(uiPoints, 
	      RI_P, prtPoints, 
	      RI_N, prtNormals, 
	      //RI_CS, prtColors, 
	      RI_ST, prtTextureCoords, 
	      //RI_NP, rtPlaneNormal, // Okan says no!!!
	      RI_NULL);
  }
  m_pCSLIDESurface->WriteRIBPop(fpOutputFile);
  
  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE FACE: %s", m_pcGUIID);

  delete[] prtPoints;
  delete[] prtNormals;
  delete[] prtColors;
  delete[] prtTextureCoords;
}

VOID CSLIDEFace::WritePOV(FILE *fpOutputFile, 
			  CListInt *plistTransformIDs,
			  CListInt *plistColors) 
{
  ASSERT( fpOutputFile != NULL );

  UINT_32 uiPoints;
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;
  Point pCenter, p0, p1;

  // Assume face is convex
  uiPoints = m_listCSLIDEPoint.GetLength();

  iter.Init(&m_listCSLIDEPoint);
  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      pCenter += pCSLIDEPoint->GetPoint();
    }
  pCenter /= uiPoints;

  p0 = ((CSLIDEPoint*)iter.PeekFirst())->GetPoint();
  
  fprintf(fpOutputFile, "polygon {\n");
  fprintf(fpOutputFile, "%ld,\n", uiPoints+1);
  fprintf(fpOutputFile, "<%f, %f, %f>,", (p0.Get())[X], (p0.Get())[Y], (p0.Get())[Z]);
  
  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      p1 = pCSLIDEPoint->GetPoint();	  
      fprintf(fpOutputFile, "<%f, %f, %f>,", (p1.Get())[X], (p1.Get())[Y], (p1.Get())[Z]);
    }
  fprintf(fpOutputFile, "<%f, %f, %f>\n", (p0.Get())[X], (p0.Get())[Y], (p0.Get())[Z]);

  //color
  
  ASSERT(m_pCSLIDESurface != NULL);

  if (strcmp(m_pcSurfaceID, "SLF_INHERIT"))
    {
      fprintf(fpOutputFile, " pigment { ");
      m_pCSLIDESurface->WritePOV(fpOutputFile);
      fprintf(fpOutputFile, " }\n");
    }

  fprintf(fpOutputFile, "}\n");
}

VOID CSLIDEFace::WriteWRLZ(FILE *fpOutputFile, const Matrix &mTransform,
			   CSLIDESurface *pCSLIDESurface)
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDEFace::WriteSTL(FILE *fpOutputFile, const Matrix &mTransform)
{
  ASSERT( fpOutputFile != NULL );

  UINT_32 uiPoints;
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;
  Point pCenter, p0, p1, p2;

  // Assume face is convex
  uiPoints = m_listCSLIDEPoint.GetLength();

  iter.Init(&m_listCSLIDEPoint);
  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      pCenter += mTransform * pCSLIDEPoint->GetPoint();
    }
  pCenter /= uiPoints;

  if( uiPoints == 3 ) // Triangle
    { 
      p0 = mTransform * ((CSLIDEPoint*)iter.PeekFirst())->GetPoint();
      p1 = mTransform * ((CSLIDEPoint*)iter.PeekNext())->GetPoint();
      p2 = mTransform * ((CSLIDEPoint*)iter.PeekNext())->GetPoint();

      if( mTransform.Determinant() > 0 ) // Don't Mirror
	{
	  WriteSTLTriangle(fpOutputFile, p0, p1, p2);
	}
      else // Mirror
	{
	  WriteSTLTriangle(fpOutputFile, p2, p1, p0);
	}
    }
  else // Polygon
    {
      p0 = mTransform * ((CSLIDEPoint*)iter.PeekFirst())->GetPoint();
      p1 = p0;

      if( mTransform.Determinant() > 0 ) // Don't Mirror
	{
	  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext(); 
		pCSLIDEPoint != NULL;
		pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
	    {
	      p2 = mTransform * pCSLIDEPoint->GetPoint();

	      WriteSTLTriangle(fpOutputFile, pCenter, p1, p2);
	      
	      p1 = p2;
	    }
	  
	  WriteSTLTriangle(fpOutputFile, pCenter, p1, p0);
	}
      else // Mirror
	{
	  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext(); 
		pCSLIDEPoint != NULL;
		pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
	    {
	      p2 = mTransform * pCSLIDEPoint->GetPoint();
	      
	      WriteSTLTriangle(fpOutputFile, p2, p1, pCenter);
	      
	      p1 = p2;
	    }
	  
	  WriteSTLTriangle(fpOutputFile, p0, p1, pCenter);
	}
    }
}

UINT_32 CSLIDEFace::GetPointCount()
{
  return m_listCSLIDEPoint.GetLength();
}

