#ifndef __SLIDEFACE_H__
#define __SLIDEFACE_H__

#include "slideprimitive.h"

class CSLIDEWorld;
class CSLIDEPoint;
class CSLIDEObject;
class CSLIDESurface;

#ifdef DT_TRIANGULATE_TEST
//class DTFace;
#endif // DT_TRIANGULATE_TEST

//////////////////////////////////////////////////////////////////////
// CSLIDEFace Class
//

class CSLIDEFace : public CSLIDEPrimitive
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEFace();
  virtual ~CSLIDEFace();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID, 
		    CListInt *plistPointIDs);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  VOID InsertParentObject(CSLIDEObject *pCSLIDEObject);
  VOID RemoveParentObject(CSLIDEObject *pCSLIDEObject);

  inline VOID SetSurfaceID(CHAR *pcSurfaceID);

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

  virtual VOID Remove(CSLIDEWorld *pCSLIDEWorld);
  VOID RemovePoint(CSLIDEPoint *pCSLIDEPoint);

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  VOID InsertIntoBound(CBound *pCBound);
  virtual VOID Preprocess();
  VOID PreprocessAutoNormals();
  VOID CheckAutoNormalsDynamic();

  virtual VOID SetBoundUnvisited();
  virtual VOID SetBoundUnchanged();
  virtual VOID SetPrerenderUnvisited();

  VOID DrawGUIGraphPointReferences(Tcl_Interp *pInterp);

#ifdef __USE_CDA__
  VOID CalculateCDAVertices(CListInt *plistCDAVertex, 
			    const Matrix &mGroot_obj);
#endif // __USE_CDA__

protected:
  VOID CheckDynamic();
  VOID UpdateState();
  VOID CalculateNormal();
  VOID CalculateCentroid();

//////////////////////////////////////////////////
// Rendering
//
public:
  VOID Prerender(SLIDEShadingType eShading);
  VOID Render(SLIDEShadingType eShading, CSLIDEWormwarp *pCSLIDEWarp = NULL);
  VOID RenderNormals(SLIDEShadingType eShading);

  VOID RenderSelect();

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);
  virtual VOID WriteRIB(FILE *fpOutputFile, SLIDEShadingType eShading);
  virtual VOID WritePOV(FILE *fpOutputFile, 
			CListInt *plistTransformIDs,
			CListInt *plistColors); 
  virtual VOID WriteWRLZ(FILE *fpOutputFile, const Matrix &mTransform,
			 CSLIDESurface *pCSLIDESurface);
  virtual VOID WriteSTL(FILE *fpOutputFile, const Matrix &mTransform);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline SLIDEReservedWordType GetType();

  virtual UINT_32 GetPointCount();
  inline const Vector &GetNormal() const;
  inline CSLIDESurface *GetSurface();

  virtual inline BOOL GetBoundEmpty();
  virtual inline VOID SetBoundEmpty(BOOL bBoundEmpty);
  virtual inline BOOL GetBoundDynamic();
  virtual inline VOID SetBoundDynamic(BOOL bBoundDynamic);
  virtual inline BOOL GetGeometryDynamic();
  virtual inline VOID SetGeometryDynamic(BOOL bGeometryDynamic);
  virtual inline BOOL GetNormalsDynamic();
  virtual inline VOID SetNormalsDynamic(BOOL bNormalsDynamic);
  virtual inline BOOL GetAutoNormals();
  virtual inline VOID SetAutoNormals(BOOL bAutoNormals);
  virtual inline BOOL GetAutoNormalsDynamic();
  virtual inline VOID SetAutoNormalsDynamic(BOOL bAutoNormalsDynamic);

  inline UINT_32 GetPrerenderAttributesStackDepthMax();

  inline CListInt *GetPointList();
#ifdef DT_TRIANGULATE_TEST
//  inline DTFace *GetDTFace();
#endif // DT_TRIANGULATE_TEST

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CHAR *m_pcSurfaceID;
  CSLIDESurface *m_pCSLIDESurface;
  CListInt *m_plistPointIDs;
  CListInt m_listCSLIDEPoint;

  CListInt m_listCSLIDEParentObject;

  BOOL m_bBoundEmpty;
  BOOL m_bBoundDynamic;
  BOOL m_bGeometryDynamic;
  BOOL m_bNormalsDynamic;

  BOOL m_bAutoNormals;
  BOOL m_bAutoNormalsDynamic;

  Vector m_vNormal;
  Point m_ptCentroid;

  UINT_32 m_uiPrerenderAttributesStackDepthMax;

#ifdef DT_TRIANGULATE_TEST
//  DTFace *m_pDTFace;
  CListInt m_listDTTriangles;
#endif // DT_TRIANGULATE_TEST
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline SLIDEReservedWordType CSLIDEFace::GetType()
{
  return sliderwtFace;
}

inline const Vector &CSLIDEFace::GetNormal() const
{
  return m_vNormal;
}

inline CSLIDESurface *CSLIDEFace::GetSurface()
{
  return m_pCSLIDESurface;
}

inline VOID CSLIDEFace::SetSurfaceID(CHAR *pcSurfaceID)
{
  m_pcSurfaceID = pcSurfaceID;
}

inline BOOL CSLIDEFace::GetBoundEmpty()
{
  return m_bBoundEmpty;
}

inline VOID CSLIDEFace::SetBoundEmpty(BOOL bBoundEmpty)
{
  m_bBoundEmpty = bBoundEmpty;
}

inline BOOL CSLIDEFace::GetBoundDynamic()
{
  return m_bBoundDynamic;
}

inline VOID CSLIDEFace::SetBoundDynamic(BOOL bBoundDynamic)
{
  m_bBoundDynamic = bBoundDynamic;
}

inline BOOL CSLIDEFace::GetGeometryDynamic()
{
  return m_bGeometryDynamic;
}

inline VOID CSLIDEFace::SetGeometryDynamic(BOOL bGeometryDynamic)
{
  m_bGeometryDynamic = bGeometryDynamic;
}

inline BOOL CSLIDEFace::GetNormalsDynamic()
{
  return m_bNormalsDynamic;
}

inline VOID CSLIDEFace::SetNormalsDynamic(BOOL bNormalsDynamic)
{
  m_bNormalsDynamic = bNormalsDynamic;
}

inline BOOL CSLIDEFace::GetAutoNormals()
{
  return m_bAutoNormals;
}

inline VOID CSLIDEFace::SetAutoNormals(BOOL bAutoNormals)
{
  m_bAutoNormals = bAutoNormals;
}

inline BOOL CSLIDEFace::GetAutoNormalsDynamic()
{
  return m_bAutoNormalsDynamic;
}

inline VOID CSLIDEFace::SetAutoNormalsDynamic(BOOL bAutoNormalsDynamic)
{
  m_bAutoNormalsDynamic = bAutoNormalsDynamic;
}

inline UINT_32 CSLIDEFace::GetPrerenderAttributesStackDepthMax()
{
  return m_uiPrerenderAttributesStackDepthMax;
}

inline CListInt *CSLIDEFace::GetPointList()
{
  return &m_listCSLIDEPoint;
}

#ifdef DT_TRIANGULATE_TEST
/*
inline DTFace *CSLIDEFace::GetDTFace()
{
  return m_pDTFace;
}
*/
#endif // DT_TRIANGULATE_TEST

#endif // __SLIDEFACE_H__




