#include "slidegeneralface.h"

#include "slideworld.h"
#include "slidesurface.h"

#include "slidepoint.h"

#include "qemesh.h"
#include "qerender.h"
#include "qewritelsif.h"

#define NO_TRANSFORMS

static const SLIDETerrainType ceType = slidetertSLF_TERRAIN_DELAUNAY_INCREMENTAL;
static const SLIDEVoronoiType ceVoronoi = slidevtSLF_VORONOI_OFF;
static const SLIDEAccuracyType ceAccuracy = slideatSLF_FLOAT;
static const SLIDEOnOffType ceRandomize = slideoftSLF_OFF;
static const SLIDEOnOffType ceCleanUpContour = slideoftSLF_OFF;

static const SLIDEOnOffType ceMillingContour = slideoftSLF_OFF;
static const INT_32 ciOffsets = 0;
static const FLOAT cfOffsetMin = 0.0;
static const FLOAT cfOffsetMax = 0.0;

static const INT_32 ciIndividualVoronoiMin = 0;
static const INT_32 ciIndividualVoronoiMax = 0;

static const SLIDEOnOffType ceDrawContours = slideoftSLF_OFF;
static const SLIDEOnOffType ceDrawVertices = slideoftSLF_OFF;
static const SLIDEOnOffType ceDrawCDT = slideoftSLF_ON;
static const SLIDEOnOffType ceDrawIndividualVoronoi = slideoftSLF_OFF;
static const SLIDEOnOffType ceDrawVoronoi = slideoftSLF_OFF;
static const SLIDEOnOffType ceDrawHeight = slideoftSLF_OFF;
static const SLIDEOnOffType ceDrawOffset = slideoftSLF_OFF;
static const SLIDEOnOffType ceDrawMillingContour = slideoftSLF_OFF;

static const INT_32 ciSubdivisionLevel = 0;
static const FLOAT cfTolerance = 0.001;

//////////////////////////////////////////////////////////////////////
// CSLIDEGeneralFace Class
//

CSLIDEGeneralFace::CSLIDEGeneralFace()
{
  m_dyniType.SetValue(ceType);
  m_dyniVoronoi.SetValue(ceVoronoi);
  m_dyniAccuracy.SetValue(ceAccuracy);
  m_dyniRandomize.SetValue(ceRandomize);
  m_dyniCleanUpContour.SetValue(ceCleanUpContour);

  m_dyniMillingContour.SetValue(ceMillingContour);
  m_dyniOffsets.SetValue(ciOffsets);
  m_dynfOffsetMin.SetValue(cfOffsetMin);
  m_dynfOffsetMax.SetValue(cfOffsetMax);

  m_dyniIndividualVoronoiMin.SetValue(ciIndividualVoronoiMin);
  m_dyniIndividualVoronoiMax.SetValue(ciIndividualVoronoiMax);

  m_dyniDrawContours.SetValue(ceDrawContours);
  m_dyniDrawVertices.SetValue(ceDrawVertices);
  m_dyniDrawCDT.SetValue(ceDrawCDT);
  m_dyniDrawIndividualVoronoi.SetValue(ceDrawIndividualVoronoi);
  m_dyniDrawVoronoi.SetValue(ceDrawVoronoi);
  m_dyniDrawHeight.SetValue(ceDrawHeight);
  m_dyniDrawOffset.SetValue(ceDrawOffset);
  m_dyniDrawMillingContour.SetValue(ceDrawMillingContour);

  m_dyniSubdivisionLevel.SetValue(ciSubdivisionLevel);
  m_dynfTolerance.SetValue(cfTolerance);

  m_ppmeContourVoronoi = NULL;
  m_pmeVoronoi = NULL;
  m_pmeOffset = NULL;
  m_pmeMillingContour = NULL;
}

CSLIDEGeneralFace::~CSLIDEGeneralFace()
{
  UninitLocal();
}

VOID CSLIDEGeneralFace::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			 CHAR *pcID,
			 CListInt *plistPointIDs)
{
  CSLIDEBoundedNode::Init(pcFile, uiLine, uiColumn, pcID);

  m_listPointIDLists.Init();
  m_listPointIDLists.InsertLast(plistPointIDs, 0);
  m_listCSLIDEPointLists.Init();
  m_listQEVertexLists.Init();

  m_dyniType.SetValue(ceType);
  m_dyniVoronoi.SetValue(ceVoronoi);
  m_dyniAccuracy.SetValue(ceAccuracy);
  m_dyniRandomize.SetValue(ceRandomize);
  m_dyniCleanUpContour.SetValue(ceCleanUpContour);

  m_dyniMillingContour.SetValue(ceMillingContour);
  m_dyniOffsets.SetValue(ciOffsets);
  m_dynfOffsetMin.SetValue(cfOffsetMin);
  m_dynfOffsetMax.SetValue(cfOffsetMax);

  m_dyniIndividualVoronoiMin.SetValue(ciIndividualVoronoiMin);
  m_dyniIndividualVoronoiMax.SetValue(ciIndividualVoronoiMax);

  m_dyniDrawContours.SetValue(ceDrawContours);
  m_dyniDrawVertices.SetValue(ceDrawVertices);
  m_dyniDrawCDT.SetValue(ceDrawCDT);
  m_dyniDrawIndividualVoronoi.SetValue(ceDrawIndividualVoronoi);
  m_dyniDrawVoronoi.SetValue(ceDrawVoronoi);
  m_dyniDrawHeight.SetValue(ceDrawHeight);
  m_dyniDrawOffset.SetValue(ceDrawOffset);
  m_dyniDrawMillingContour.SetValue(ceDrawMillingContour);

  m_dyniSubdivisionLevel.SetValue(ciSubdivisionLevel);
  m_dynfTolerance.SetValue(cfTolerance);

  m_ppmeContourVoronoi = NULL;
  m_pmeVoronoi = NULL;
  m_pmeOffset = NULL;
  m_pmeMillingContour = NULL;
}

VOID CSLIDEGeneralFace::Uninit()
{
  UninitLocal();

  CSLIDEBoundedNode::Uninit();
}

VOID CSLIDEGeneralFace::UninitLocal()
{
  CListInt *plist;
  CObjectList<QEVertex>* plistQEVertex; 
  QEVertex *pvx;
  UINT_32 i;

  for ( plist = (CListInt*)(m_listPointIDLists.RemoveLast());
	plist != NULL;
	plist = (CListInt*)(m_listPointIDLists.RemoveLast()) )
    {
      delete plist;
    }

  for ( plist = (CListInt*)(m_listCSLIDEPointLists.RemoveLast());
	plist != NULL;
	plist = (CListInt*)(m_listCSLIDEPointLists.RemoveLast()) )
    {
      delete plist;
    }

  if ( m_pmeVoronoi != NULL )
    {
      DeallocateQEMesh(m_pmeVoronoi);
      m_pmeVoronoi = NULL;
    }
  if ( m_pmeOffset != NULL )
    {
      DeallocateQEMesh(m_pmeOffset);
      m_pmeOffset = NULL;
    }
  if ( m_pmeMillingContour != NULL )
    {
      DeallocateQEMesh(m_pmeMillingContour);
      m_pmeMillingContour = NULL;
    }

  if ( m_ppmeContourVoronoi != NULL )
    {
      for ( i = 0; i < m_listQEVertexLists.GetLength(); i++ )
	{
	  if ( m_ppmeContourVoronoi[i] != NULL )
	    {
	      DeallocateQEMesh(m_ppmeContourVoronoi[i]);
	      m_ppmeContourVoronoi[i] = NULL;
	    }
	}

      delete []m_ppmeContourVoronoi;
      m_ppmeContourVoronoi = NULL;
    }
  
  for ( plistQEVertex = (CObjectList<QEVertex>*)(m_listQEVertexLists.RemoveLast());
	plistQEVertex != NULL;
	plistQEVertex = (CObjectList<QEVertex>*)(m_listQEVertexLists.RemoveLast()) )
    {
      for ( pvx = plistQEVertex->RemoveLast();
	    pvx != NULL;
	    pvx = plistQEVertex->RemoveLast() )
	{
	  DeallocateQEVertex(pvx);
	}

      delete plistQEVertex;
    }
}

VOID CSLIDEGeneralFace::InsertContour(CListInt *plistPointIDs)
{
  ASSERT( plistPointIDs != NULL );
  m_listPointIDLists.InsertLast(plistPointIDs, 0);
}

BOOL CSLIDEGeneralFace::Initialized()
{
  if ( !CSLIDEBoundedNode::Initialized() )
    {
      return FALSE;
    }

  return (m_listPointIDLists.PeekFirst() != NULL);
}

BOOL CSLIDEGeneralFace::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;
  CListIterInt iterLists;
  CListIterInt iter;
  CListInt *plistIDs, *plist;
  CHAR *pcID;
  CSLIDEPoint *pCSLIDEPoint;

  CObjectList<QEVertex> *plistQEVertex;
  QEVertex *pvx;
  Point pt;
  UINT_32 i;

  if ( !CSLIDEBoundedNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  // Check Solid Type

  ASSERT( 0 <= m_dyniSolid.GetValue() &&
	  m_dyniSolid.GetValue() < slidestSize );

  // Check Terrain Type
  m_dyniType.Update();
  if ( !( 0 <= m_dyniType.GetValue() &&
	  m_dyniType.GetValue() < slidetertSize ) )
    {
      HandleError("Illegal Terrain Type flag value:  %ld", 
		  m_dyniType.GetValue());
      m_dyniType.SetValue(ceType);
    }

  // Check Accuracy Type
  m_dyniAccuracy.Update();
  if ( !( 0 <= m_dyniAccuracy.GetValue() &&
	  m_dyniAccuracy.GetValue() < slideatSize ) )
    {
      HandleError("Illegal GeneralFace Accuracy flag value:  %ld", 
		  m_dyniAccuracy.GetValue());
      m_dyniAccuracy.SetValue(ceAccuracy);
    }

  // Check Randomize
  m_dyniRandomize.Update();
  if ( !( 0 <= m_dyniRandomize.GetValue() &&
	  m_dyniRandomize.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal GeneralFace Random flag value:  %ld", 
		  m_dyniRandomize.GetValue());
      m_dyniRandomize.SetValue(ceRandomize);
    }

  // Check Draw Contours
  m_dyniDrawContours.Update();
  if ( !( 0 <= m_dyniDrawContours.GetValue() &&
	  m_dyniDrawContours.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal GeneralFace Draw Contours flag value:  %ld", 
		  m_dyniDrawContours.GetValue());
      m_dyniDrawContours.SetValue(ceDrawContours);
    }

  // Check Draw Vertices
  m_dyniDrawVertices.Update();
  if ( !( 0 <= m_dyniDrawVertices.GetValue() &&
	  m_dyniDrawVertices.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal Terrain Draw Vertices flag value:  %ld", 
		  m_dyniDrawVertices.GetValue());
      m_dyniDrawVertices.SetValue(ceDrawVertices);
    }

  // Check Draw CDT
  m_dyniDrawCDT.Update();
  if ( !( 0 <= m_dyniDrawCDT.GetValue() &&
	  m_dyniDrawCDT.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal Terrain Draw CDT flag value:  %ld", 
		  m_dyniDrawCDT.GetValue());
      m_dyniDrawCDT.SetValue(ceDrawCDT);
    }

  // Check Draw Individual Voronoi
  m_dyniDrawIndividualVoronoi.Update();
  if ( !( 0 <= m_dyniDrawIndividualVoronoi.GetValue() &&
	  m_dyniDrawIndividualVoronoi.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal Terrain Draw Individual Voronoi flag value:  %ld", 
		  m_dyniDrawIndividualVoronoi.GetValue());
      m_dyniDrawIndividualVoronoi.SetValue(ceDrawIndividualVoronoi);
    }

  // Check Draw Voronoi
  m_dyniDrawVoronoi.Update();
  if ( !( 0 <= m_dyniDrawVoronoi.GetValue() &&
	  m_dyniDrawVoronoi.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal Terrain Draw Voronoi flag value:  %ld", 
		  m_dyniDrawVoronoi.GetValue());
      m_dyniDrawVoronoi.SetValue(ceDrawVoronoi);
    }

  // Check Draw Height
  m_dyniDrawHeight.Update();
  if ( !( 0 <= m_dyniDrawHeight.GetValue() &&
	  m_dyniDrawHeight.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal Terrain Draw Height flag value:  %ld", 
		  m_dyniDrawHeight.GetValue());
      m_dyniDrawHeight.SetValue(ceDrawHeight);
    }

  // Check Voronoi
  m_dyniVoronoi.Update();
  if ( !( 0 <= m_dyniVoronoi.GetValue() &&
	  m_dyniVoronoi.GetValue() < slidevtSize ) )
    {
      HandleError("Illegal GeneralFace Voronoi flag value:  %ld", 
		  m_dyniVoronoi.GetValue());
      m_dyniVoronoi.SetValue(ceVoronoi);
    }

  // Check Draw Offset
  m_dyniDrawOffset.Update();
  if ( !( 0 <= m_dyniDrawOffset.GetValue() &&
	  m_dyniDrawOffset.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal GeneralFace Draw Offset flag value:  %ld", 
		  m_dyniDrawOffset.GetValue());
      m_dyniDrawOffset.SetValue(ceDrawOffset);
    }

  // Check Draw Milling Contour
  m_dyniDrawMillingContour.Update();
  if ( !( 0 <= m_dyniDrawMillingContour.GetValue() &&
	  m_dyniDrawMillingContour.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal GeneralFace Draw Milling Contour flag value:  %ld", 
		  m_dyniDrawMillingContour.GetValue());
      m_dyniDrawMillingContour.SetValue(ceDrawMillingContour);
    }

  // Check  Milling Contour
  m_dyniMillingContour.Update();
  if ( !( 0 <= m_dyniMillingContour.GetValue() &&
	  m_dyniMillingContour.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal GeneralFace Milling Contour flag value:  %ld", 
		  m_dyniMillingContour.GetValue());
      m_dyniMillingContour.SetValue(ceMillingContour);
    }

  // Check Clean Up Contour
  m_dyniCleanUpContour.Update();
  if ( !( 0 <= m_dyniCleanUpContour.GetValue() &&
	  m_dyniCleanUpContour.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal GeneralFace Clean Up Contour flag value:  %ld", 
		  m_dyniCleanUpContour.GetValue());
      m_dyniCleanUpContour.SetValue(ceCleanUpContour);
    }

  // Check Offsets
  m_dyniOffsets.Update();
  if ( m_dyniOffsets.GetValue() < 0 )
    {
      HandleError("Illegal GeneralFace Offsets value:  %ld", 
		  m_dyniOffsets.GetValue());
      m_dyniOffsets.SetValue(ciOffsets);
    }

  // Check Offset Min
  m_dynfOffsetMin.Update();

  // Check Offset Max
  m_dynfOffsetMax.Update();

  // Check Individual Voronoi Min
  m_dyniIndividualVoronoiMin.Update();

  // Check Individual Voronoi Max
  m_dyniIndividualVoronoiMax.Update();

  // Check SubdivisionLevel
  m_dyniSubdivisionLevel.Update();
  if ( m_dyniSubdivisionLevel.GetValue() < 0 )
    {
      HandleError("Illegal GeneralFace SubdivisionLevel value:  %ld", 
		  m_dyniSubdivisionLevel.GetValue());
      m_dyniSubdivisionLevel.SetValue(ciSubdivisionLevel);
    }

  // Check Tolerance
  m_dynfTolerance.Update();
  if ( m_dynfTolerance.GetValue() < 0.0 )
    {
      HandleError("Illegal GeneralFace Tolerance value:  %f", 
		  m_dynfTolerance.GetValue());
      m_dynfTolerance.SetValue(cfTolerance);
    }

  // Check Point References

  iterLists.Init(&m_listPointIDLists);

  for ( plistIDs = (CListInt*)(iterLists.PeekFirst());
	plistIDs != NULL;
	plistIDs = (CListInt*)(iterLists.PeekNext()) )
    {
      plist = new CListInt;
      ASSERT( plist != NULL );
      plist->Init();
      m_listCSLIDEPointLists.InsertLast(plist, 0);

      plistQEVertex = new CObjectList<QEVertex>;
      ASSERT( plistQEVertex != NULL );
      plistQEVertex->Init();
      m_listQEVertexLists.InsertLast(plistQEVertex, 0);

      pt.Null();

      iter.Init(plistIDs);

      for ( pcID = (CHAR*)(iter.PeekFirst());
	    pcID != NULL;
	    pcID = (CHAR*)(iter.PeekNext()) )
	{
	  pCSLIDEPoint = pCSLIDEWorld->PeekPoint(pcID);
	  if ( pCSLIDEPoint == NULL )
	    {
	      HandleError("Undefined %s \"%s\"", gcppcSLIDEReservedWords[sliderwtPoint], pcID);
	      bCorrect = FALSE;
	    }
	  plist->InsertLast(pCSLIDEPoint, 0);

	  pvx = AllocateQEVertex();
	  ASSERT( pvx != NULL );
	  pvx->Init(pt);
	  plistQEVertex->InsertLast(pvx);
	}
    }

  if ( bCorrect )
    {
      m_ppmeContourVoronoi = new QEMesh*[m_listQEVertexLists.GetLength()];
      ASSERT( m_ppmeContourVoronoi != NULL );
      for ( i = 0; i < m_listQEVertexLists.GetLength(); i++ )
	{
	  m_ppmeContourVoronoi[i] = NULL;
	}
    }

  return bCorrect;
}

BOOL CSLIDEGeneralFace::Valid()
{
  if ( !CSLIDEBoundedNode::Valid() )
    {
      return FALSE;
    }

  return ( ( 0 <= m_dyniSolid.GetValue() &&
	     m_dyniSolid.GetValue() < slidestSize ) );
	   //GetPointCount() > 0 );
}

CSLIDECamera *CSLIDEGeneralFace::FindCamera(CListIterInt *piterIDs, 
				       CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtCamera],
	      gcppcSLIDEReservedWords[sliderwtGeneralFace], GetID());

  return NULL;
}

CSLIDELight *CSLIDEGeneralFace::FindLight(CListIterInt *piterIDs, 
				     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtLight],
	      gcppcSLIDEReservedWords[sliderwtGeneralFace], GetID());

  return NULL;
}

CSLIDEAreaLight *CSLIDEGeneralFace::FindAreaLight(CListIterInt *piterIDs, 
					     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtAreaLight],
	      gcppcSLIDEReservedWords[sliderwtGeneralFace], GetID());

  return NULL;
}

VOID CSLIDEGeneralFace::Preprocess()
{
  CSLIDEBoundedNode::Preprocess();

  CheckDynamic();
  UpdateState();
}

VOID CSLIDEGeneralFace::CheckDynamic()
{
  CListIterInt iter;
  //CSLIDEPoint *pCSLIDEPoint;
  //BOOL bAutoNormals;

  // Are there variables which I must call TCL to update?
  SetDynamic(m_dyniShading.GetDynamic() ||
	     m_dyniLOD.GetDynamic() ||
	     m_dyniSolid.GetDynamic() ||
	     m_dyniType.GetDynamic() ||
	     m_dyniVoronoi.GetDynamic() ||
	     m_dyniAccuracy.GetDynamic() ||
	     m_dyniRandomize.GetDynamic() ||
	     m_dyniCleanUpContour.GetDynamic() ||
	     m_dyniMillingContour.GetDynamic() ||
	     m_dyniOffsets.GetDynamic() ||
	     m_dynfOffsetMin.GetDynamic() ||
	     m_dynfOffsetMax.GetDynamic() ||
	     m_dyniIndividualVoronoiMin.GetDynamic() ||
	     m_dyniIndividualVoronoiMax.GetDynamic() ||
	     m_dyniDrawContours.GetDynamic() ||
	     m_dyniDrawVertices.GetDynamic() ||
	     m_dyniDrawCDT.GetDynamic() ||
	     m_dyniDrawIndividualVoronoi.GetDynamic() ||
	     m_dyniDrawVoronoi.GetDynamic() ||
	     m_dyniDrawHeight.GetDynamic() ||
	     m_dyniDrawOffset.GetDynamic() ||
	     m_dyniDrawMillingContour.GetDynamic() ||
	     m_dyniSubdivisionLevel.GetDynamic() ||
	     m_dynfTolerance.GetDynamic());
/*
  bAutoNormals = !(!m_dyniShading.GetDynamic() && 
		   (m_dyniShading.GetValue() == 
		    (INT_32)slideshtSLF_WIRE ||
		    m_dyniShading.GetValue() == 
		    (INT_32)slideshtSLF_FLAT));
      
  iter.Init(&m_listCSLIDEPoint);
  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      pCSLIDEPoint->SetAutoNormals(bAutoNormals || pCSLIDEPoint->GetAutoNormals());
    }
*/
}

VOID CSLIDEGeneralFace::PreprocessGraph()
{
  if ( !GetVisited() )
    {
      CListIterInt iterLists;
      CListIterInt iter;
      CListInt *plist;
      CSLIDEPoint *pCSLIDEPoint;
      
      BOOL bGeometryEmpty;
      BOOL bGeometryDynamic;
      BOOL bNormalsEmpty;
      BOOL bNormalsDynamic;
      BOOL bBoundsEmpty;
      BOOL bBoundsDynamic;

      BOOL bBoundEmpty;
      BOOL bBoundDynamic;

      BOOL bChildrenBoundsDynamic;

      bGeometryEmpty = TRUE;
      bGeometryDynamic = FALSE;
      bNormalsEmpty = TRUE;
      bNormalsDynamic = FALSE;
      bBoundsEmpty = TRUE;
      bBoundsDynamic = FALSE;

      bBoundEmpty = TRUE;
      bBoundDynamic = FALSE;

      //////////////////////////////////////////////////
      // Look at Points

      iterLists.Init(&m_listCSLIDEPointLists);
      
      for ( plist = (CListInt*)iterLists.PeekFirst(); 
	    plist != NULL;
	    plist = (CListInt*)iterLists.PeekNext() )
	{
	  iter.Init(plist);

	  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
		pCSLIDEPoint != NULL;
		pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
	    {
	      bBoundEmpty = FALSE;
	      bBoundDynamic = bBoundDynamic || 
		pCSLIDEPoint->GetBoundDynamic();
	      bGeometryDynamic = bGeometryDynamic || 
		pCSLIDEPoint->GetGeometryDynamic();
	      bNormalsDynamic = bNormalsDynamic ||
		pCSLIDEPoint->GetNormalDynamic();
	    }
	}

      bGeometryEmpty = bBoundEmpty;
      bNormalsEmpty = bBoundEmpty;
      bBoundsEmpty = bBoundEmpty;
      bBoundsDynamic = bBoundDynamic;

      //////////////////////////////////////////////////
      // Look at Local flags

      // Update whether we are Empty based on our LOD flag
      if ( !m_dyniLOD.GetDynamic() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      // This node is always off
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	      bBoundsEmpty = TRUE;
	      bBoundEmpty = TRUE;
	    }
	  else if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	    {
	      // This node is always just a bounding box
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	    }
	}

      // Check what is really dynamic

      bChildrenBoundsDynamic = bBoundsDynamic;

      bGeometryDynamic = ( bGeometryDynamic ||
			   m_pCSLIDESurface->GetDynamic() ||
			   m_dyniShading.GetDynamic() ||
			   m_dyniLOD.GetDynamic() ||
			   m_dyniSolid.GetDynamic() );
      bNormalsDynamic = ( bNormalsDynamic ||
			  m_dyniShading.GetDynamic() ||
			  m_dyniLOD.GetDynamic() );
      bBoundsDynamic = ( bBoundsDynamic || 
			 m_dyniLOD.GetDynamic() );
      bBoundDynamic = ( bBoundDynamic || 
			m_dyniLOD.GetDynamic() );

      // If anything is now Empty then it will always be empty so
      // it is not dynamic

      bChildrenBoundsDynamic = ( bChildrenBoundsDynamic && !bBoundsEmpty );

      bGeometryDynamic = ( bGeometryDynamic && !bGeometryEmpty );
      bNormalsDynamic = ( bNormalsDynamic && !bNormalsEmpty );
      bBoundsDynamic = ( bBoundsDynamic && !bBoundsEmpty );
      bBoundDynamic = ( bBoundDynamic && !bBoundEmpty );
      
      //////////////////////////////////////////////////
      // For the OpenGL Display Lists to be really useful 
      // it is all or nothing so we enforce that here

      bGeometryDynamic = ( bGeometryDynamic ||
			   bNormalsDynamic ||
			   bBoundsDynamic );
      bNormalsDynamic = bGeometryDynamic;
      bBoundsDynamic = bGeometryDynamic;

      //////////////////////////////////////////////////
      // Now save these correct flag values

      SetChildrenBoundsDynamic(bChildrenBoundsDynamic);

      SetGeometryEmpty(bGeometryEmpty);
      SetGeometryDynamic(bGeometryDynamic);
      SetNormalsEmpty(bNormalsEmpty);
      SetNormalsDynamic(bNormalsDynamic);
      SetBoundsEmpty(bBoundsEmpty);
      SetBoundsDynamic(bBoundsDynamic);
      SetBoundEmpty(bBoundEmpty);
      SetBoundDynamic(bBoundDynamic);
      
      //////////////////////////////////////////////////
      // Calculate the Bounding Box (if Empty this will do nothing)
      if ( GetBoundEmpty() )
	{
	  SetBoundValid(TRUE);
	}
      else
	{
	  CalculateBound();
	}
      SetBoundChanged(FALSE);

      //////////////////////////////////////////////////
      // Done Preprocessing this Node in the Graph
      SetVisited(TRUE);
    }
}

VOID CSLIDEGeneralFace::Update()
{
  SetBoundChanged(FALSE);
  if ( GetDynamic() ||
       GetChildrenBoundsDynamic() )
    {
      if ( GetChildrenBoundsDynamic() )
	{
	  CSLIDEBoundedNode::InvalidateQEMesh();
	}
      UpdateState();
    }

  if ( GetBoundDynamic() )
    {
      SetBoundValid(FALSE);
    }

  SetVisited(FALSE);
}

VOID CSLIDEGeneralFace::UpdateState()
{
  UINT_32 uiType, uiVoronoi, uiAccuracy, uiRandomize, uiCleanUpContour, uiMillingContour;
  INT_32 iOffsets;
  FLOAT fOffsetMin, fOffsetMax, fToolRadius;
  UINT_32 i;

  CSLIDEBoundedNode::UpdateState();

  m_dyniSolid.Update();
  if ( !( 0 <= m_dyniSolid.GetValue() &&
	  m_dyniSolid.GetValue() < slidestSize ) )
    {
      HandleError("Illegal solid flag value:  %ld", 
		  m_dyniSolid.GetValue());
      m_dyniSolid.SetValue(slidestSLF_HOLLOW);
    }

  uiType = m_dyniType.GetValue();
  m_dyniType.Update();
  if ( !( 0 <= m_dyniType.GetValue() &&
	  m_dyniType.GetValue() < slidetertSize ) )
    {
      HandleError("Illegal Terrain Type flag value:  %ld", 
		  m_dyniType.GetValue());
      m_dyniType.SetValue(ceType);
    }

  uiVoronoi = m_dyniVoronoi.GetValue();
  m_dyniVoronoi.Update();
  if ( !( 0 <= m_dyniVoronoi.GetValue() &&
	  m_dyniVoronoi.GetValue() < slidevtSize ) )
    {
      HandleError("Illegal GeneralFace Voronoi flag value:  %ld", 
		  m_dyniVoronoi.GetValue());
      m_dyniVoronoi.SetValue(ceVoronoi);
    }

  uiAccuracy = m_dyniAccuracy.GetValue();
  m_dyniAccuracy.Update();
  if ( !( 0 <= m_dyniAccuracy.GetValue() &&
	  m_dyniAccuracy.GetValue() < slideatSize ) )
    {
      HandleError("Illegal GeneralFace Accuracy flag value:  %ld", 
		  m_dyniAccuracy.GetValue());
      m_dyniAccuracy.SetValue(ceAccuracy);
    }

  uiRandomize = m_dyniRandomize.GetValue();
  m_dyniRandomize.Update();
  if ( !( 0 <= m_dyniRandomize.GetValue() &&
	  m_dyniRandomize.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal GeneralFace Random flag value:  %ld", 
		  m_dyniRandomize.GetValue());
      m_dyniRandomize.SetValue(ceRandomize);
    }

  m_dyniDrawContours.Update();
  if ( !( 0 <= m_dyniDrawContours.GetValue() &&
	  m_dyniDrawContours.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal GeneralFace Draw Contours flag value:  %ld", 
		  m_dyniDrawContours.GetValue());
      m_dyniDrawContours.SetValue(ceDrawContours);
    }

  m_dyniDrawVertices.Update();
  if ( !( 0 <= m_dyniDrawVertices.GetValue() &&
	  m_dyniDrawVertices.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal GeneralFace Draw Vertices flag value:  %ld", 
		  m_dyniDrawVertices.GetValue());
      m_dyniDrawVertices.SetValue(ceDrawVertices);
    }

  m_dyniDrawCDT.Update();
  if ( !( 0 <= m_dyniDrawCDT.GetValue() &&
	  m_dyniDrawCDT.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal GeneralFace Draw CDT flag value:  %ld", 
		  m_dyniDrawCDT.GetValue());
      m_dyniDrawCDT.SetValue(ceDrawCDT);
    }

  m_dyniDrawIndividualVoronoi.Update();
  if ( !( 0 <= m_dyniDrawIndividualVoronoi.GetValue() &&
	  m_dyniDrawIndividualVoronoi.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal GeneralFace Draw Individual Voronoi flag value:  %ld", 
		  m_dyniDrawIndividualVoronoi.GetValue());
      m_dyniDrawIndividualVoronoi.SetValue(ceDrawIndividualVoronoi);
    }

  m_dyniDrawVoronoi.Update();
  if ( !( 0 <= m_dyniDrawVoronoi.GetValue() &&
	  m_dyniDrawVoronoi.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal GeneralFace Draw Voronoi flag value:  %ld", 
		  m_dyniDrawVoronoi.GetValue());
      m_dyniDrawVoronoi.SetValue(ceDrawVoronoi);
    }

  m_dyniDrawHeight.Update();
  if ( !( 0 <= m_dyniDrawHeight.GetValue() &&
	  m_dyniDrawHeight.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal GeneralFace Draw Height flag value:  %ld", 
		  m_dyniDrawHeight.GetValue());
      m_dyniDrawHeight.SetValue(ceDrawHeight);
    }

  m_dyniDrawOffset.Update();
  if ( !( 0 <= m_dyniDrawOffset.GetValue() &&
	  m_dyniDrawOffset.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal GeneralFace Draw Offset flag value:  %ld", 
		  m_dyniDrawOffset.GetValue());
      m_dyniDrawOffset.SetValue(ceDrawOffset);
    }

  m_dyniDrawMillingContour.Update();
  if ( !( 0 <= m_dyniDrawMillingContour.GetValue() &&
	  m_dyniDrawMillingContour.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal GeneralFace Draw Milling Contour flag value:  %ld", 
		  m_dyniDrawMillingContour.GetValue());
      m_dyniDrawMillingContour.SetValue(ceDrawMillingContour);
    }

  uiMillingContour = m_dyniMillingContour.GetValue();
  m_dyniMillingContour.Update();
  if ( !( 0 <= m_dyniMillingContour.GetValue() &&
	  m_dyniMillingContour.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal GeneralFace Milling Contour flag value:  %ld", 
		  m_dyniMillingContour.GetValue());
      m_dyniMillingContour.SetValue(ceMillingContour);
    }

  uiCleanUpContour = m_dyniCleanUpContour.GetValue();
  m_dyniCleanUpContour.Update();
  if ( !( 0 <= m_dyniCleanUpContour.GetValue() &&
	  m_dyniCleanUpContour.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal GeneralFace Clean Up Contour flag value:  %ld", 
		  m_dyniCleanUpContour.GetValue());
      m_dyniCleanUpContour.SetValue(ceCleanUpContour);
    }

  m_dyniSubdivisionLevel.Update();
  if ( m_dyniSubdivisionLevel.GetValue() < 0 )
    {
      HandleError("Illegal GeneralFace SubdivisionLevel value:  %ld", 
		  m_dyniSubdivisionLevel.GetValue());
      m_dyniSubdivisionLevel.SetValue(ciSubdivisionLevel);
    }

  m_dynfTolerance.Update();
  if ( m_dynfTolerance.GetValue() < 0.0 )
    {
      HandleError("Illegal GeneralFace Tolerance value:  %f", 
		  m_dynfTolerance.GetValue());
      m_dynfTolerance.SetValue(cfTolerance);
    }

  iOffsets = m_dyniOffsets.GetValue();
  m_dyniOffsets.Update();
  if ( m_dyniOffsets.GetValue() < 0 )
    {
      HandleError("Illegal GeneralFace Offsets value:  %ld", 
		  m_dyniOffsets.GetValue());
      m_dyniOffsets.SetValue(ciOffsets);
    }

  fOffsetMin = m_dynfOffsetMin.GetValue();
  m_dynfOffsetMin.Update();

  fOffsetMax = m_dynfOffsetMax.GetValue();
  m_dynfOffsetMax.Update();

  m_dyniIndividualVoronoiMin.Update();

  m_dyniIndividualVoronoiMax.Update();

  if ( uiType != (UINT_32)m_dyniType.GetValue() || 
       uiAccuracy != (UINT_32)m_dyniAccuracy.GetValue() || 
       uiRandomize != (UINT_32)m_dyniRandomize.GetValue() ||
       m_pQEMesh == NULL )
    {
      if ( m_pQEMesh != NULL )
	{
	  DeallocateQEMesh(m_pQEMesh);
	  m_pQEMesh = NULL;
	}

      m_pQEMesh = AllocateQEMesh();
      ASSERT( m_pQEMesh != NULL );
      m_pQEMesh->Init();

      CListIterInt iterQEVertexLists;
      CObjectListIter<QEVertex> iterQEVertex;
      CObjectList<QEVertex> *plistQEVertex;
      QEVertex *pvx;
      CListIterInt iterLists;
      CListIterInt iter;
      CListInt *plist;
      CSLIDEPoint *pCSLIDEPoint;
      Point pt, ptPrev;

      iterLists.Init(&m_listCSLIDEPointLists);

      //////////////////////////////////////////////////
      // Calculate the Gobj_xy transform
      {
	Vector vZ, vAxis, vTrans;
	Point ptOrigin;
	FLOAT fSin, fCos, fTheta;
	Matrix mTrans, mTransInv, mRot, mRotInv;

	ptOrigin.Set(0.0, 0.0, 0.0);
	vZ.Set(0.0, 0.0, 1.0);

	plist = (CListInt*)iterLists.PeekFirst(); 
	
	m_vNormal = CalculateNormal(plist);
	m_ptCentroid = CalculateCentroid(plist);  
	
	vTrans = ptOrigin - m_ptCentroid;
	mTrans.SetTranslate(vTrans);
	vTrans = -vTrans;
	mTransInv.SetTranslate(vTrans);

	vAxis = m_vNormal * vZ;
	fCos = m_vNormal.Dot(vZ);
	fSin = vAxis.Magnitude();
	if ( !vAxis.Normalize() )
	  {
	    vAxis.Set(1.0, 0.0, 0.0);
	    if ( fCos < 0.0 )
	      {
		fTheta = PI;
	      }
	    else
	      {
		fTheta = 0.0;
	      }
	  }
	else
	  {
	    fTheta = atan2(fSin, fCos);
	  }
	mRot.SetRotate(vAxis, fTheta);
	mRotInv.SetRotate(vAxis, -fTheta);	

	m_mGxy_obj = mRot * mTrans;
	m_mGobj_xy = mTransInv * mRotInv;
      }
      // Calculate the Gobj_xy transform
      //////////////////////////////////////////////////
    
      iterQEVertexLists.Init(&m_listQEVertexLists);
      for ( plist = (CListInt*)iterLists.PeekFirst(),
	      plistQEVertex = (CObjectList<QEVertex>*)iterQEVertexLists.PeekFirst(); 
	    plist != NULL && plistQEVertex != NULL;
	    plist = (CListInt*)iterLists.PeekNext(),
	      plistQEVertex = (CObjectList<QEVertex>*)iterQEVertexLists.PeekNext() )
	{
	  iter.Init(plist);
  	  iterQEVertex.Init(plistQEVertex);

	  pCSLIDEPoint = (CSLIDEPoint*)iter.PeekLast();
#ifdef NO_TRANSFORMS
	  ptPrev = pCSLIDEPoint->GetPoint();
#else // NO_TRANSFORMS
	  ptPrev = m_mGxy_obj * pCSLIDEPoint->GetPoint();
#endif // NO_TRANSFORMS
	  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(), 
		  pvx = iterQEVertex.PeekFirst();
		pCSLIDEPoint != NULL && pvx != NULL;
		pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext(),
		  pvx = iterQEVertex.PeekNext() )
	    {
#ifdef NO_TRANSFORMS
	      pt = pCSLIDEPoint->GetPoint();
#else // NO_TRANSFORMS
	      pt = m_mGxy_obj * pCSLIDEPoint->GetPoint();
#endif // NO_TRANSFORMS
	      pt[2] = 0.0;
	      pt[3] = 1.0;

	      // There used to be code here that checked for zero length edges
	      pvx->SetPoint(pt);

	      ptPrev = pt;
	    }
	  ASSERT( pCSLIDEPoint == NULL && pvx == NULL );
	}
      ASSERT( plist == NULL && plistQEVertex == NULL );

      if ( m_dyniRandomize.GetValue() )
	{
	  QEMesh::SetRandomize(TRUE);
	}
      else
	{
	  QEMesh::SetRandomize(FALSE);
	}

      switch ( m_dyniAccuracy.GetValue() )
	{
	case slideatSLF_FLOAT_NAIVE:
	  QEMesh::SetAccuracyType(QEMesh::atFloatNaive);
	  break;
	case slideatSLF_FLOAT:
	  QEMesh::SetAccuracyType(QEMesh::atFloat);
	  break;
	case slideatSLF_EXACT:
	  QEMesh::SetAccuracyType(QEMesh::atExact);
	  break;
	default:
	  ASSERT( FALSE );
	  break;
	}

      switch ( m_dyniType.GetValue() )
	{
	case slidetertSLF_TERRAIN_DELAUNAY_CONVEX_HULL:
	  m_pQEMesh->CreateContours(&m_listQEVertexLists);
	  break;
	default:
	  m_pQEMesh->CreateTriangulatedPolygon(&m_listQEVertexLists);
	  break;
	}
    }

  //////////////////////////////////////////////////
  // Voronoi Diagram
  //
  if ( uiVoronoi != (UINT_32)m_dyniVoronoi.GetValue() ||
       uiCleanUpContour != (UINT_32)m_dyniCleanUpContour.GetValue() )
    {
      if ( m_pmeVoronoi != NULL )
	{
	  DeallocateQEMesh(m_pmeVoronoi);
	  m_pmeVoronoi = NULL;
	}
      if ( m_ppmeContourVoronoi != NULL )
	{
	  for ( i = 0; i < m_listQEVertexLists.GetLength(); i++ )
	    {
	      if ( m_ppmeContourVoronoi[i] != NULL )
		{
		  DeallocateQEMesh(m_ppmeContourVoronoi[i]);
		  m_ppmeContourVoronoi[i] = NULL;
		}
	    }
	}
    }
  if ( (m_dyniVoronoi.GetValue() != slidevtSLF_VORONOI_OFF) &&
       (m_pmeVoronoi == NULL) )
    {
      if ( m_dyniRandomize.GetValue() )
	{
	  QEMesh::SetRandomize(TRUE);
	}
      else
	{
	  QEMesh::SetRandomize(FALSE);
	}

      switch ( m_dyniAccuracy.GetValue() )
	{
	case slideatSLF_FLOAT_NAIVE:
	  QEMesh::SetAccuracyType(QEMesh::atFloatNaive);
	  break;
	case slideatSLF_FLOAT:
	  QEMesh::SetAccuracyType(QEMesh::atFloat);
	  break;
	case slideatSLF_EXACT:
	  QEMesh::SetAccuracyType(QEMesh::atExact);
	  break;
	default:
	  ASSERT( FALSE );
	  break;
	}

      switch ( m_dyniVoronoi.GetValue() )
	{
	case slidevtSLF_VORONOI_DELAUNAY: 
	  m_pmeVoronoi = AllocateQEMesh();
	  ASSERT( m_pmeVoronoi != NULL );
	  m_pmeVoronoi->Init();

	  m_pmeVoronoi->CreateVoronoiPolygon(QEMesh::vtDelaunay, m_pQEMesh, m_dyniCleanUpContour.GetValue());
	  break;
	case slidevtSLF_VORONOI_DIVIDE_AND_CONQUER: 
	  {
	    CListIterInt iterQEVertexLists;
	    CObjectList<QEVertex> *plistQEVertex;
	    CObjectListIter<QEVertex> iterListQEVertex;
	    CHeap<QEMesh*, QEVertex> heapQEMesh;
	    QEMesh *pme;
	    QEVertex *pvx, *pvxMin;
	    UINT_32 i;

	    heapQEMesh.Init();
	    heapQEMesh.SetBiggestFirst(FALSE);

	    iterQEVertexLists.Init(&m_listQEVertexLists);
	    for ( plistQEVertex = (CObjectList<QEVertex>*)iterQEVertexLists.PeekFirst(); 
		  plistQEVertex != NULL;
		  plistQEVertex = (CObjectList<QEVertex>*)iterQEVertexLists.PeekNext() )
	      {
		iterListQEVertex.Init(plistQEVertex);
		pvxMin = iterListQEVertex.PeekFirst();
		for ( pvx = iterListQEVertex.PeekNext();
		      pvx != NULL;
		      pvx = iterListQEVertex.PeekNext() )
		  {
		    if ( *pvx < *pvxMin )
		      {
			pvxMin = pvx;
		      }
		  }

		pme = AllocateQEMesh();
		ASSERT( pme != NULL );
		pme->Init();
		
		pme->CreateVoronoiPolygonShewchuk(plistQEVertex, m_dyniCleanUpContour.GetValue());
		heapQEMesh.Insert(*pvxMin, pme);
	      }

	    for ( i = 0, pme = heapQEMesh.RemoveTop();
		  pme != NULL;
		  i++, pme = heapQEMesh.RemoveTop() )
	      {
		m_ppmeContourVoronoi[i] = pme;
	      }
	    ASSERT( i == m_listQEVertexLists.GetLength() );

	    //////////////////////////////////////////////////
	    // Merge Individual Voronoi Diagrams
	    m_pmeVoronoi = m_ppmeContourVoronoi[0]->CopyVoronoiPolygon();
	    ASSERT( m_pmeVoronoi != NULL );
	    for ( i = 1; i < m_listQEVertexLists.GetLength(); i++ )
	      {
		//m_pmeVoronoi->MergeVoronoiPolygonsHeld(m_ppmeContourVoronoi[i]->CopyVoronoiPolygon());
		m_pmeVoronoi->MergeVoronoiPolygonsShewchuk(m_ppmeContourVoronoi[i]->CopyVoronoiPolygon());
	      }
	    m_pmeVoronoi->ZMonotoneVoronoiPolygon();
	    m_pmeVoronoi->DisplayMinChannelVertices();
	  }
	  break;
	case slidevtSLF_VORONOI_DIVIDE_AND_CONQUER_ALTERNATING: 
	  m_pmeVoronoi = AllocateQEMesh();
	  ASSERT( m_pmeVoronoi != NULL );
	  m_pmeVoronoi->Init();

	  //m_pmeVoronoi->CreateVoronoiPolygon(QEMesh::vtDivideAndConquerAlternating, m_pQEMesh, m_dyniCleanUpContour.GetValue());
	  //m_pmeVoronoi->CreateVoronoiPolygonHeld(m_pQEMesh, m_dyniCleanUpContour.GetValue());
	  m_pmeVoronoi->CreateVoronoiPolygonShewchuk(m_pQEMesh, m_dyniCleanUpContour.GetValue());
	  m_pmeVoronoi->ZMonotoneVoronoiPolygon();
	  //m_pmeVoronoi->DisplayMinChannelVertices();
	  break;
	case slidevtSLF_VORONOI_CONES: 
	  m_pmeVoronoi = AllocateQEMesh();
	  ASSERT( m_pmeVoronoi != NULL );
	  m_pmeVoronoi->Init();

	  m_pmeVoronoi->CreateVoronoiPolygon(QEMesh::vtCones, m_pQEMesh, m_dyniCleanUpContour.GetValue());
	  break;
	case slidevtSLF_VORONOI_PLANES: 
	  m_pmeVoronoi = AllocateQEMesh();
	  ASSERT( m_pmeVoronoi != NULL );
	  m_pmeVoronoi->Init();

	  m_pmeVoronoi->CreateVoronoiPolygon(QEMesh::vtPlanes, m_pQEMesh, m_dyniCleanUpContour.GetValue());
	  break;
	default:
	  ASSERT(FALSE);
	  break;
	} 
    }

  //////////////////////////////////////////////////
  // Offset Contours
  //
  if ( !( iOffsets == m_dyniOffsets.GetValue() &&
	  fOffsetMin == m_dynfOffsetMin.GetValue() &&
	  fOffsetMax == m_dynfOffsetMax.GetValue() ) )
    {
      if ( m_pmeOffset != NULL )
	{
	  DeallocateQEMesh(m_pmeOffset);
	  m_pmeOffset = NULL;
	}
    }
  if ( (m_dyniDrawOffset.GetValue() != slideoftSLF_OFF) &&
       ((m_dyniVoronoi.GetValue() == slidevtSLF_VORONOI_DIVIDE_AND_CONQUER ||
	 m_dyniVoronoi.GetValue() == slidevtSLF_VORONOI_DIVIDE_AND_CONQUER_ALTERNATING) &&
	m_pmeVoronoi != NULL) &&
       (m_dyniOffsets.GetValue() > 0 ) &&
       (m_pmeOffset == NULL) )
    {
      if ( m_dyniRandomize.GetValue() )
	{
	  QEMesh::SetRandomize(TRUE);
	}
      else
	{
	  QEMesh::SetRandomize(FALSE);
	}

      switch ( m_dyniAccuracy.GetValue() )
	{
	case slideatSLF_FLOAT_NAIVE:
	  QEMesh::SetAccuracyType(QEMesh::atFloatNaive);
	  break;
	case slideatSLF_FLOAT:
	  QEMesh::SetAccuracyType(QEMesh::atFloat);
	  break;
	case slideatSLF_EXACT:
	  QEMesh::SetAccuracyType(QEMesh::atExact);
	  break;
	default:
	  ASSERT( FALSE );
	  break;
	}

      m_pmeOffset = AllocateQEMesh();
      ASSERT( m_pmeOffset != NULL );
      m_pmeOffset->Init();

      {
	UINT_32 i, uiOffsets;
	FLOAT fOffset, fOffsetDelta;

	uiOffsets = (UINT_32)m_dyniOffsets.GetValue();
	fOffsetDelta = m_dynfOffsetMax.GetValue() - m_dynfOffsetMin.GetValue();
	if ( iOffsets > 1 )
	  {
	    fOffsetDelta /= (FLOAT)(iOffsets-1);
	  }
	fOffset = m_dynfOffsetMin.GetValue();
	for ( i = 0; i < uiOffsets; i++ )
	  {
	    m_pmeOffset->CreateOffsetPolygon(m_pQEMesh, m_pmeVoronoi, fOffset);
	    fOffset += fOffsetDelta;
	  }
      }
    }

  //////////////////////////////////////////////////
  // Milling Contour
  //
  if ( !( uiMillingContour == (UINT_32)(m_dyniMillingContour.GetValue()) &&
	  fOffsetMin == m_dynfOffsetMin.GetValue() ) )
    {
      if ( m_pmeMillingContour != NULL )
	{
	  DeallocateQEMesh(m_pmeMillingContour);
	  m_pmeMillingContour = NULL;
	}
    }
  if ( (m_dyniDrawMillingContour.GetValue() != slideoftSLF_OFF) &&
       ((m_dyniVoronoi.GetValue() == slidevtSLF_VORONOI_DIVIDE_AND_CONQUER ||
	 m_dyniVoronoi.GetValue() == slidevtSLF_VORONOI_DIVIDE_AND_CONQUER_ALTERNATING) &&
	m_pmeVoronoi != NULL) &&
       m_dyniMillingContour.GetValue() &&
       (m_pmeMillingContour == NULL) )
    {
      if ( m_dyniRandomize.GetValue() )
	{
	  QEMesh::SetRandomize(TRUE);
	}
      else
	{
	  QEMesh::SetRandomize(FALSE);
	}

      switch ( m_dyniAccuracy.GetValue() )
	{
	case slideatSLF_FLOAT_NAIVE:
	  QEMesh::SetAccuracyType(QEMesh::atFloatNaive);
	  break;
	case slideatSLF_FLOAT:
	  QEMesh::SetAccuracyType(QEMesh::atFloat);
	  break;
	case slideatSLF_EXACT:
	  QEMesh::SetAccuracyType(QEMesh::atExact);
	  break;
	default:
	  ASSERT( FALSE );
	  break;
	}

      m_pmeMillingContour = AllocateQEMesh();
      ASSERT( m_pmeMillingContour != NULL );
      m_pmeMillingContour->Init();

      fToolRadius = m_dynfOffsetMin.GetValue();
      if ( fToolRadius < 0.0 )
	{
	  fToolRadius = 0.0;
	}
      m_pmeMillingContour->CreateMillingContour(m_pQEMesh, m_pmeVoronoi, fToolRadius);
    }
}

VOID CSLIDEGeneralFace::UpdateGraph()
{
  if ( !GetVisited() )
    {
      if ( !GetBoundValid() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      if ( !m_bound.GetEmpty() )
		{
		  SetBoundChanged(TRUE);
		  m_bound.Null();
		}
	      return;
	    }
	  
	  CalculateBound();
	  SetBoundValid(TRUE);
	}

      SetVisited(TRUE);
    }
}

// This uses Newell's Algorithm for finding the best fit planar equation
// for a set of points in 3 space (pg. 476 Foley and van Dam)
// I modified it so that the normal would be positive for counter-clockwise
// polygons, and I removed the 1/2 coefficient because we are concerned 
// with relative areas, not absolute areas.

Vector CSLIDEGeneralFace::CalculateNormal(CListInt *plistCSLIDEPoint)
{
  ASSERT( plistCSLIDEPoint != NULL );

  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;
  Point ptStart, ptPrev, ptCurr;
  Vector vNormal;

  vNormal.Null();

  iter.Init(plistCSLIDEPoint);

  pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
  if ( pCSLIDEPoint == NULL )
    {
      HandleWarning("Null normal because no points for face");
      vNormal.Set(0.0, 0.0, 1.0);
      return vNormal;
    }
  ptStart = pCSLIDEPoint->GetPoint();
  ptPrev = ptStart;

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      ptCurr = pCSLIDEPoint->GetPoint();

      vNormal[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
      vNormal[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
      vNormal[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);

      ptPrev = ptCurr;
    }

  ptCurr = ptStart;
  vNormal[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
  vNormal[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
  vNormal[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);
  
  if ( !vNormal.Normalize() )
    {
      HandleWarning("Null normal because of zero area face");
      vNormal.Set(0.0, 0.0, 1.0);
    }

  return vNormal;
}

Point CSLIDEGeneralFace::CalculateCentroid(CListInt *plistCSLIDEPoint)
{
  ASSERT( plistCSLIDEPoint != NULL );

  CListIterInt iter;
  UINT_32 i;
  CSLIDEPoint *pCSLIDEPoint;
  Point ptCentroid;

  iter.Init(plistCSLIDEPoint);

  ptCentroid.Null();

  i = 0;
  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst();
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      ptCentroid = ptCentroid + pCSLIDEPoint->GetPoint();
      i++;
    }

  if ( i > 0 )
    {
      ptCentroid = ptCentroid / (FLOAT)i;
    }

  ptCentroid[W] = 1.0;

  return ptCentroid;
}

VOID CSLIDEGeneralFace::CalculateBound()
{
  CListIterInt iterLists;
  CListIterInt iter;
  CListInt *plist;
  CSLIDEPoint *pCSLIDEPoint;

  CBound bound;

  iterLists.Init(&m_listCSLIDEPointLists);

  for ( plist = (CListInt*)iterLists.PeekFirst(); 
	plist != NULL;
	plist = (CListInt*)iterLists.PeekNext() )
    {
      iter.Init(plist);
      
      for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	    pCSLIDEPoint != NULL;
	    pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
	{
	  pCSLIDEPoint->InsertIntoBound(&bound);
	}
    }

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);
    }
}

#ifdef __USE_CDA__
VOID CSLIDEGeneralFace::CalculateCDAVertices(CListInt *plistCDAVertex, 
					const Matrix &mGroot_obj)
{
  CListIterInt iterLists;
  CListIterInt iter;
  CListInt *plist;
  CSLIDEPoint *pCSLIDEPoint;

  iterLists.Init(&m_listCSLIDEPointLists);

  for ( plist = (CListInt*)iterLists.PeekFirst(); 
	plist != NULL;
	plist = (CListInt*)iterLists.PeekNext() )
    {
      iter.Init(plist);
      
      for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	    pCSLIDEPoint != NULL;
	    pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
	{
	  pCSLIDEPoint->CalculateCDAVertices(plistCDAVertex, mGroot_obj);
	}
    }
}
#endif // __USE_CDA__

QEMesh *CSLIDEGeneralFace::GetQEMesh()
{
  QEMesh *pme;

  if ( m_pQEMesh == NULL )
    {
      CalculateQEMesh();
      
      if ( m_pQEMesh == NULL )
	{
	  return NULL;
	}
    }

  pme = m_pQEMesh->Copy();
#ifdef NO_TRANSFORMS
#else // NO_TRANSFORMS
  pme->Transform(m_mGobj_xy, m_mGxy_obj);
#endif // NO_TRANSFORMS

  return pme;
}

VOID CSLIDEGeneralFace::PrerenderGeometry(SLIDEShadingType eShading)
{
  SLIDEShadingType eShade;
  CListIterInt iterLists;
  CListIterInt iter;
  CListInt *plist;
  CSLIDEPoint *pCSLIDEPoint;
  QEMeshIter iterQEMesh;
  Point pt;
  
  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  // This flag tells us whether to use Point or Point Normals
  ModifyShadingFlag(eShading);

  // This one set the shading attributes which in many cases will 
  // be SLF_INHERIT and so it will not set anything and leave up
  // to the caller above
  eShade = (SLIDEShadingType)m_dyniShading.GetValue();
  
  GLbitfield GLbAttributesMask;
  
  GLbAttributesMask = CalculateAttributesMask(eShade);
  
  PushAttributes(GLbAttributesMask);
  {
    PushShadingAttributes(eShade);
    {
      m_pCSLIDESurface->Push();
      {
	gilPushMatrix();
	{
#ifdef NO_TRANSFORMS
#else // NO_TRANSFORMS
	  gilMultMatrixf(m_mGobj_xy.Get());
#endif // NO_TRANSFORMS

	  PushSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
	  {
	    m_uiPrerenderAttributesStackDepthMax = 0;
	    
	    if ( m_dyniDrawCDT.GetValue() )
	      {
		::RenderGeometry(m_pQEMesh);
	      }
	    
	    if ( GLbAttributesMask != 0x0 )
	      {
		m_uiPrerenderAttributesStackDepthMax++;
	      }
	  }
	  PopSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
	  
	  PushAuxillaryGeometryAttributes();
	  {
	    gilLineWidth(1.0);
	    if ( m_dyniDrawIndividualVoronoi.GetValue() )
	      {
		gilColor3f(0.0, 1.0, 0.0);
		switch ( m_dyniVoronoi.GetValue() )
		  {
		  case slidevtSLF_VORONOI_OFF: 
		  case slidevtSLF_VORONOI_DELAUNAY:
		  case slidevtSLF_VORONOI_DIVIDE_AND_CONQUER_ALTERNATING: 
		  case slidevtSLF_VORONOI_CONES: 
		  case slidevtSLF_VORONOI_PLANES: 
		    break;
		  case slidevtSLF_VORONOI_DIVIDE_AND_CONQUER: 
		    {
		      UINT_32 i, uiLength, uiMin, uiMax;
		      FLOAT fBlue, fBlueDelta;
		      
		      uiMin = m_dyniIndividualVoronoiMin.GetValue();
		      uiMax = m_dyniIndividualVoronoiMax.GetValue();
		      
		      uiLength = m_listQEVertexLists.GetLength();
		      if ( uiLength == 1 )
			{
			  fBlueDelta = 0.0;
			}
		      else
			{
			  fBlueDelta = 1.0 / (FLOAT)(uiLength - 1);
			}
		      
		      for ( i = 0, fBlue = 0.0; i < uiLength; i++, fBlue += fBlueDelta )
			{
			  if ( uiMin <= i && i <= uiMax )
			    { 
			      gilColor3f(1.0, 0.0, fBlue);
			      ASSERT( m_ppmeContourVoronoi[i] != NULL );
			      ::RenderVoronoiPolygon(m_pQEMesh, m_ppmeContourVoronoi[i], 
						     QE::vtDivideAndConquer, 
						     m_dyniSubdivisionLevel.GetValue(),
						     m_dynfTolerance.GetValue(),
						     m_dyniDrawHeight.GetValue());
			    }
			}
		    }
		  break;
		  default:
		    ASSERT(FALSE);
		    break;
		  } 
	      }
	    
	    if ( m_dyniDrawVoronoi.GetValue() )
	      {
		gilColor3f(1.0, 1.0, 0.0);
		switch ( m_dyniVoronoi.GetValue() )
		  {
		  case slidevtSLF_VORONOI_OFF: 
		    break;
		  case slidevtSLF_VORONOI_DELAUNAY:
		    ::RenderVoronoiPolygon(m_pQEMesh, m_pmeVoronoi, 
					   QE::vtDelaunay, 
					   m_dyniSubdivisionLevel.GetValue(),
					   m_dynfTolerance.GetValue(),
					   m_dyniDrawHeight.GetValue());
		    break;
		  case slidevtSLF_VORONOI_DIVIDE_AND_CONQUER: 
		    ::RenderVoronoiPolygon(m_pQEMesh, m_pmeVoronoi, 
					   QE::vtDivideAndConquer, 
					   m_dyniSubdivisionLevel.GetValue(),
					   m_dynfTolerance.GetValue(),
					   m_dyniDrawHeight.GetValue());
		    break;
		  case slidevtSLF_VORONOI_DIVIDE_AND_CONQUER_ALTERNATING: 
		    ::RenderVoronoiPolygon(m_pQEMesh, m_pmeVoronoi, 
					   QE::vtDivideAndConquerAlternating, 
					   m_dyniSubdivisionLevel.GetValue(),
					   m_dynfTolerance.GetValue(),
					   m_dyniDrawHeight.GetValue());
		    break;
		  case slidevtSLF_VORONOI_CONES: 
		    ::RenderVoronoiPolygon(m_pQEMesh, m_pmeVoronoi, 
					   QE::vtCones, 
					   m_dyniSubdivisionLevel.GetValue(),
					   m_dynfTolerance.GetValue(),
					   m_dyniDrawHeight.GetValue());
		    break;
		  case slidevtSLF_VORONOI_PLANES: 
		    ::RenderVoronoiPolygon(m_pQEMesh, m_pmeVoronoi, 
					   QE::vtPlanes, 
					   m_dyniSubdivisionLevel.GetValue(),
					   m_dynfTolerance.GetValue(),
					   m_dyniDrawHeight.GetValue());
		    break;
		  default:
		    ASSERT(FALSE);
		    break;
		  } 
	      }
	    
	    if ( m_dyniDrawOffset.GetValue() )
	      {
		if ( m_pmeOffset != NULL )
		  {
		    gilColor3f(1.0, 0.0, 1.0);
		    ::RenderOffsetContour(m_pmeOffset, m_dyniDrawHeight.GetValue());
		  }
	      }

	    if ( m_dyniDrawMillingContour.GetValue() )
	      {
		if ( m_pmeMillingContour != NULL )
		  {
		    gilColor3f(0.0, 1.0, 1.0);
		    ::RenderOffsetContour(m_pmeMillingContour, FALSE);
		  }
	      }
	    gilLineWidth(1.0);
	  }
	  PopAuxillaryGeometryAttributes();
	}
      }
      gilPopMatrix();
      m_pCSLIDESurface->Pop();


      PushAuxillaryGeometryAttributes();
      {
	iterLists.Init(&m_listCSLIDEPointLists);
	    
	if ( m_dyniDrawContours.GetValue() )
	  {
	    gilColor3f(0.0, 0.0, 1.0);
	    gilLineWidth(3.0);
	    for ( plist = (CListInt*)iterLists.PeekFirst(); 
		  plist != NULL;
		  plist = (CListInt*)iterLists.PeekNext() )
	      {
		iter.Init(plist);
		    
		gilBegin(GL_LINE_LOOP);
		{
		  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
			pCSLIDEPoint != NULL;
			pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
		    {
		      pt = pCSLIDEPoint->GetPoint();
		      gilVertex3fv(pt.Get());	
		    }
		}
		gilEnd();
	      }
	  }
	    
	if ( m_dyniDrawVertices.GetValue() )
	  {
	    gilPointSize(6.0);
	    gilColor3f(0.0, 1.0, 0.0);
	    gilBegin(GL_POINTS);
	    {
	      for ( plist = (CListInt*)iterLists.PeekFirst(); 
		    plist != NULL;
		    plist = (CListInt*)iterLists.PeekNext() )
		{
		  iter.Init(plist);
		      
		  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
			pCSLIDEPoint != NULL;
			pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
		    {
		      pt = pCSLIDEPoint->GetPoint();
		      gilVertex3fv(pt.Get());	
		    }
		}
	    }
	    gilEnd();
	  }

	gilLineWidth(1.0);
      }
      PopAuxillaryGeometryAttributes();
	    
    }
    PopShadingAttributes(eShade);
  }
  PopAttributes(GLbAttributesMask);
}

VOID CSLIDEGeneralFace::PrerenderNormals(SLIDEShadingType eShading)
{
  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  ModifyShadingFlag(eShading);

  gilPushMatrix();
  {
#ifdef NO_TRANSFORMS
#else // NO_TRANSFORMS
    gilMultMatrixf(m_mGobj_xy.Get());
#endif // NO_TRANSFORMS

  //::RenderNormals(m_pQEMesh);
  }
  gilPopMatrix();
}

VOID CSLIDEGeneralFace::RenderGeometry(const Matrix &/*mGproj_obj*/, 
				       CBound * /*pCBoundClip*/, 
				       SLIDEShadingType eShading,
				       SLIDELODType /* eLOD */,
				       CSLIDEWormwarp */*pCSLIDEWarp*//* = NULL*/)

{
  CListIterInt iterLists;
  CListIterInt iter;
  CListInt *plist;
  CSLIDEPoint *pCSLIDEPoint;
  QEMeshIter iterQEMesh;
  Point pt;

  gilPushMatrix();
  {
#ifdef NO_TRANSFORMS
#else // NO_TRANSFORMS
    gilMultMatrixf(m_mGobj_xy.Get());
#endif // NO_TRANSFORMS

    PushSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
    {
      if ( m_dyniDrawCDT.GetValue() )
	{
	  ::RenderGeometry(m_pQEMesh);
	}
    }
    PopSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());

    PushAuxillaryGeometryAttributes();
    {
      gilLineWidth(1.0);
      if ( m_dyniDrawIndividualVoronoi.GetValue() )
	{
	  gilColor3f(0.0, 1.0, 0.0);
	  switch ( m_dyniVoronoi.GetValue() )
	    {
	    case slidevtSLF_VORONOI_OFF: 
	    case slidevtSLF_VORONOI_DELAUNAY:
	    case slidevtSLF_VORONOI_DIVIDE_AND_CONQUER_ALTERNATING: 
	    case slidevtSLF_VORONOI_CONES: 
	    case slidevtSLF_VORONOI_PLANES: 
	      break;
	    case slidevtSLF_VORONOI_DIVIDE_AND_CONQUER: 
	      {
		UINT_32 i, uiLength, uiMin, uiMax;
		FLOAT fBlue, fBlueDelta;

		uiMin = m_dyniIndividualVoronoiMin.GetValue();
		uiMax = m_dyniIndividualVoronoiMax.GetValue();

		uiLength = m_listQEVertexLists.GetLength();
		if ( uiLength == 1 )
		  {
		    fBlueDelta = 0.0;
		  }
		else
		  {
		    fBlueDelta = 1.0 / (FLOAT)(uiLength - 1);
		  }

		for ( i = 0, fBlue = 0.0; i < uiLength; i++, fBlue += fBlueDelta )
		  {
		    if ( uiMin <= i && i <= uiMax )
		      { 
			gilColor3f(1.0, 0.0, fBlue);
			ASSERT( m_ppmeContourVoronoi[i] != NULL );
			::RenderVoronoiPolygon(m_pQEMesh, m_ppmeContourVoronoi[i], 
					       QE::vtDivideAndConquer, 
					       m_dyniSubdivisionLevel.GetValue(),
					       m_dynfTolerance.GetValue(),
					       m_dyniDrawHeight.GetValue());
		      }
		  }
	      }
	    break;
	    default:
	      ASSERT(FALSE);
	      break;
	    } 
	}
	    
      if ( m_dyniDrawVoronoi.GetValue() )
	{
	  gilColor3f(1.0, 1.0, 0.0);
	  switch ( m_dyniVoronoi.GetValue() )
	    {
	    case slidevtSLF_VORONOI_OFF: 
	      break;
	    case slidevtSLF_VORONOI_DELAUNAY:
	      ::RenderVoronoiPolygon(m_pQEMesh, m_pmeVoronoi, 
				     QE::vtDelaunay, 
				     m_dyniSubdivisionLevel.GetValue(),
				     m_dynfTolerance.GetValue(),
				     m_dyniDrawHeight.GetValue());
	      break;
	    case slidevtSLF_VORONOI_DIVIDE_AND_CONQUER: 
	      ::RenderVoronoiPolygon(m_pQEMesh, m_pmeVoronoi, 
				     QE::vtDivideAndConquer, 
				     m_dyniSubdivisionLevel.GetValue(),
				     m_dynfTolerance.GetValue(),
				     m_dyniDrawHeight.GetValue());
	      break;
	    case slidevtSLF_VORONOI_DIVIDE_AND_CONQUER_ALTERNATING: 
	      ::RenderVoronoiPolygon(m_pQEMesh, m_pmeVoronoi, 
				     QE::vtDivideAndConquerAlternating, 
				     m_dyniSubdivisionLevel.GetValue(),
				     m_dynfTolerance.GetValue(),
				     m_dyniDrawHeight.GetValue());
	      break;
	    case slidevtSLF_VORONOI_CONES: 
	      ::RenderVoronoiPolygon(m_pQEMesh, m_pmeVoronoi, 
				     QE::vtCones, 
				     m_dyniSubdivisionLevel.GetValue(),
				     m_dynfTolerance.GetValue(),
				     m_dyniDrawHeight.GetValue());
	      break;
	    case slidevtSLF_VORONOI_PLANES: 
	      ::RenderVoronoiPolygon(m_pQEMesh, m_pmeVoronoi, 
				     QE::vtPlanes, 
				     m_dyniSubdivisionLevel.GetValue(),
				     m_dynfTolerance.GetValue(),
				     m_dyniDrawHeight.GetValue());
	      break;
	    default:
	      ASSERT(FALSE);
	      break;
	    } 
	}

      if ( m_dyniDrawOffset.GetValue() )
	{
	  if ( m_pmeOffset != NULL )
	    {
	      gilColor3f(1.0, 0.0, 1.0);
	      gilLineWidth(2.0);
	      ::RenderOffsetContour(m_pmeOffset, m_dyniDrawHeight.GetValue());
	    }
	}

      if ( m_dyniDrawMillingContour.GetValue() )
	{
	  if ( m_pmeMillingContour != NULL )
	    {
	      gilColor3f(0.0, 1.0, 1.0);
	      gilLineWidth(3.0);
	      ::RenderOffsetContour(m_pmeMillingContour, FALSE);
	    }
	}
      gilLineWidth(1.0);
    }
    PopAuxillaryGeometryAttributes();
  }
  gilPopMatrix();

  PushAuxillaryGeometryAttributes();
  {
    iterLists.Init(&m_listCSLIDEPointLists);
	  
    if ( m_dyniDrawContours.GetValue() )
      {
	gilColor3f(0.0, 0.0, 1.0);
	gilLineWidth(3.0);
	for ( plist = (CListInt*)iterLists.PeekFirst(); 
	      plist != NULL;
	      plist = (CListInt*)iterLists.PeekNext() )
	  {
	    iter.Init(plist);
		  
	    gilBegin(GL_LINE_LOOP);
	    {
	      for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
		    pCSLIDEPoint != NULL;
		    pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
		{
		  pt = pCSLIDEPoint->GetPoint();
		  gilVertex3fv(pt.Get());	
		}
	    }
	    gilEnd();
	  }
      }

    if ( m_dyniDrawVertices.GetValue() )
      {
	gilPointSize(6.0);
	gilColor3f(0.0, 1.0, 0.0);
	gilBegin(GL_POINTS);
	{
	  for ( plist = (CListInt*)iterLists.PeekFirst(); 
		plist != NULL;
		plist = (CListInt*)iterLists.PeekNext() )
	    {
	      iter.Init(plist);
		    
	      for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
		    pCSLIDEPoint != NULL;
		    pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
		{
		  pt = pCSLIDEPoint->GetPoint();
		  gilVertex3fv(pt.Get());	
		}
	    }
	}
	gilEnd();
      }
    gilLineWidth(1.0);
  }
  PopAuxillaryGeometryAttributes();

//#define JORDAN_IS_DEBUGGING_THE_POINT_NEAR_VORONOI_SITES_CODE
#ifdef JORDAN_IS_DEBUGGING_THE_POINT_NEAR_VORONOI_SITES_CODE
  // Debugging Equal Distance Point Calculations
  PushAuxillaryGeometryAttributes();
  {
    CListIterInt iter;
    CListInt *plist;
    CSLIDEPoint *pCSLIDEPoint;
    Point ppptE[3][2], pptResults[2], ptProj;
    UINT_32 uiResults;
    UINT_32 i, j;
    BOOL bLift;

    bLift = (m_dyniVoronoi.GetValue() == slidevtSLF_VORONOI_CONES);

    plist = (CListInt*)(m_listCSLIDEPointLists.PeekFirst());
    ASSERT( plist != NULL );
	  
    if ( plist->GetLength() >= 3 )
      {
	iter.Init(plist);

	gilPointSize(3.0);

	if ( plist->GetLength() >= 6 )
	  {
	    pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	    for ( i = 0; i < 3; i++ )
	      {
		for ( j = 0; j < 2; j++ )
		  {
		    ASSERT( pCSLIDEPoint != NULL );

		    ppptE[i][j] = pCSLIDEPoint->GetPoint();
		    
		    pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext();
		  }
	      }
	  }
	else
	  {
	    j = 2;
	    for ( i = 0, pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
		  i < 3 && pCSLIDEPoint != NULL;
		  i++, pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
	      {
		ppptE[i][0] = pCSLIDEPoint->GetPoint();
		ppptE[j][1] = pCSLIDEPoint->GetPoint();
		j = i;
	      }
	    ASSERT( i == 3 );
	  }


	//////////////////////////////////////////////////
	// VVV
	// 

	if ( m_dyniType.GetValue() == (INT_32)slidetertSLF_TERRAIN_DELAUNAY_INCREMENTAL ||
	     m_dyniType.GetValue() == (INT_32)slidetertSLF_TERRAIN_DELAUNAY_DIVIDE_AND_CONQUER_ALTERNATING )
	  {
	    gilColor3f(0.0, 1.0, 0.0);
	    
	    QEMesh::CalculateVoronoiPointVVV(pptResults[0], ppptE[0][0], ppptE[1][0], ppptE[2][0], bLift);

	    gilBegin(GL_POINTS);
	    {
	      gilVertex3fv(pptResults[0].Get());
	    }
	    gilEnd();
	    
	    gilBegin(GL_LINES);
	    {
	      for ( i = 0; i < 3; i++ )
		{
		  gilVertex3fv(pptResults[0].Get());
		  gilVertex3fv(ppptE[i][0].Get());
		}
	    }
	    gilEnd();
	  }
	
	//////////////////////////////////////////////////
	// VVE
	// 

	if ( m_dyniType.GetValue() == (INT_32)slidetertSLF_TERRAIN_DELAUNAY_INCREMENTAL_SLOW ||
	     m_dyniType.GetValue() == (INT_32)slidetertSLF_TERRAIN_DELAUNAY_DIVIDE_AND_CONQUER_ALTERNATING )
	  {
	    gilColor3f(1.0, 1.0, 0.0);
	    
	    QEMesh::CalculateVoronoiPointVVE(pptResults, uiResults, 
					     ppptE[0][0], 
					     ppptE[1][0], 
					     ppptE[2][0], ppptE[2][1], bLift);
	    
	    for ( j = 0; j < uiResults; j++ )
	      {
		gilBegin(GL_POINTS);
		{
		  gilVertex3fv(pptResults[j].Get());
		}
		gilEnd();
		
		gilBegin(GL_LINES);
		{
		  for ( i = 0; i < 2; i++ )
		    {
		      gilVertex3fv(pptResults[j].Get());
		      gilVertex3fv(ppptE[i][0].Get());
		    }
		  
		  gilVertex3fv(pptResults[j].Get());
		  ptProj = QEMesh::CalculatePointProjectionOnLine(pptResults[j], 
								  ppptE[i][0], ppptE[i][1]);
		  gilVertex3fv(ptProj.Get());
		}
		gilEnd();
	      }
	  }

	//////////////////////////////////////////////////
	// VEE
	// 

	if ( m_dyniType.GetValue() == (INT_32)slidetertSLF_TERRAIN_DELAUNAY_INCREMENTAL_NO_BOUND ||
	     m_dyniType.GetValue() == (INT_32)slidetertSLF_TERRAIN_DELAUNAY_DIVIDE_AND_CONQUER_ALTERNATING )
	  {
	    gilColor3f(1.0, 0.0, 0.0);
	    
	    QEMesh::CalculateVoronoiPointVEE(pptResults, uiResults, 
					     ppptE[0][0], 
					     ppptE[1][0], ppptE[1][1], 
					     ppptE[2][0], ppptE[2][1], bLift);
	    
	    for ( j = 0; j < uiResults; j++ )
	      {
		gilBegin(GL_POINTS);
		{
		  gilVertex3fv(pptResults[j].Get());
		}
		gilEnd();
		
		gilBegin(GL_LINES);
		{
		  i = 0;
		  gilVertex3fv(pptResults[j].Get());
		  gilVertex3fv(ppptE[i][0].Get());
		  
		  for ( i = 1; i < 3; i++ )
		    {
		      gilVertex3fv(pptResults[j].Get());
		      ptProj = QEMesh::CalculatePointProjectionOnLine(pptResults[j], 
								      ppptE[i][0], ppptE[i][1]);
		      gilVertex3fv(ptProj.Get());
		    }
		}
		gilEnd();
	      }
	  }

	//////////////////////////////////////////////////
	// EEE
	// 

	if ( m_dyniType.GetValue() == (INT_32)slidetertSLF_TERRAIN_DELAUNAY_DIVIDE_AND_CONQUER ||
	     m_dyniType.GetValue() == (INT_32)slidetertSLF_TERRAIN_DELAUNAY_DIVIDE_AND_CONQUER_ALTERNATING )
	  {
	    gilColor3f(1.0, 0.0, 1.0);
	    
	    QEMesh::CalculateVoronoiPointEEE(pptResults[0], ppptE[0][0], ppptE[0][1], ppptE[1][0], ppptE[1][1], ppptE[2][0], ppptE[2][1], bLift);
	    
	    gilBegin(GL_POINTS);
	    {
	      gilVertex3fv(pptResults[0].Get());
	    }
	    gilEnd();
	    
	    gilBegin(GL_LINES);
	    {
	      for ( i = 0; i < 3; i++ )
		{
		  gilVertex3fv(pptResults[0].Get());
		  ptProj = QEMesh::CalculatePointProjectionOnLine(pptResults[0], 
								  ppptE[i][0], ppptE[i][1]);
		  gilVertex3fv(ptProj.Get());
		}
	    }
	    gilEnd();
	  } 
      }
  }
  PopAuxillaryGeometryAttributes();

#endif // JORDAN_IS_DEBUGGING_THE_POINT_NEAR_VORONOI_SITES_CODE
}

VOID CSLIDEGeneralFace::RenderNormals(SLIDEShadingType eShading)
{
  PushAuxillaryGeometryAttributes();
  {
    gilPushMatrix();
    {
#ifdef NO_TRANSFORMS
#else // NO_TRANSFORMS
      gilMultMatrixf(m_mGobj_xy.Get());
#endif // NO_TRANSFORMS

      gilColor3f(1.0, 0.0, 1.0);
      
      //::RenderNormals(m_pQEMesh);
    }
    gilPopMatrix();
  }
  PopAuxillaryGeometryAttributes();
}

VOID CSLIDEGeneralFace::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iterLists;
  CListIterInt iter;
  CListInt *plist;
  CSLIDEPoint *pCSLIDEPoint;
  
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s ( ", 
	  gcppcSLIDEReservedWords[sliderwtGeneralFace], m_pcID);

  iterLists.Init(&m_listCSLIDEPointLists);
  
  plist = (CListInt*)iterLists.PeekFirst();
  ASSERT( plist != NULL );
  iter.Init(plist);
  
  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      fprintf(fpOutputFile, "%s ", pCSLIDEPoint->GetID());      
    }
  fprintf(fpOutputFile, ")\n");

  for ( plist = (CListInt*)iterLists.PeekNext(); 
	plist != NULL;
	plist = (CListInt*)iterLists.PeekNext() )
    {
      fprintf(fpOutputFile, "  %s ( ", 
	      gcppcSLIDEReservedWords[sliderwtContour]);
      
      iter.Init(plist);
      
      for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	    pCSLIDEPoint != NULL;
	    pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
	{
	  fprintf(fpOutputFile, "%s ", pCSLIDEPoint->GetID());      
	}
      
      fprintf(fpOutputFile, ")\n");
    }

  WriteSLIDEFlags(fpOutputFile, 0);

  if ( m_dyniSolid.GetValue() != slidestSLF_SOLID )
    {
      fprintf(fpOutputFile, "  %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtSolid], 
	      gcppcSLIDESolidReservedWords[m_dyniSolid.GetValue()]);
    }

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndGeneralFace]);
}

VOID CSLIDEGeneralFace::WriteLSIF(FILE *fpOutputFile, UINT_32 uiIndentLevel)
{
  CListIterInt iterLists;
  CListIterInt iter;
  CListInt *plist;
  CSLIDEPoint *pCSLIDEPoint;
  UINT_32 uiVertex, i;
  Point pt;

  ASSERT( fpOutputFile != NULL );

  UpdateState();

  iterLists.Init(&m_listCSLIDEPointLists);
  for ( plist = (CListInt*)iterLists.PeekFirst(); 
	plist != NULL;
	plist = (CListInt*)iterLists.PeekNext() )
    {
      fprintf(fpOutputFile, "      (contour\n");

      uiVertex = 0;
      iter.Init(plist);
      
      for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	    pCSLIDEPoint != NULL;
	    pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
	{
	  pt = pCSLIDEPoint->GetPoint();
	  fprintf(fpOutputFile, "        (v %ld %f %f)\n", uiVertex, pt(0), pt(1));
	  uiVertex++;
	}
      
      fprintf(fpOutputFile, "\n        ");
      for ( i = 0; i < uiVertex; i++ )
	{
	  fprintf(fpOutputFile, "%ld ", i);
	}
      
      fprintf(fpOutputFile, "\n      )\n");
    }
}

VOID CSLIDEGeneralFace::WriteLSIFOffset(FILE *fpOutputFile, UINT_32 uiIndentLevel)
{
  CListIterInt iterLists;
  CListIterInt iter;
  CListInt *plist;
  //CSLIDEPoint *pCSLIDEPoint;
  //UINT_32 uiVertex, i;
  Point pt;
  QEMesh *pme;

  ASSERT( fpOutputFile != NULL );

  if ( m_pmeOffset != NULL )
    {
      pme = m_pmeOffset->Copy();
#ifdef NO_TRANSFORMS
#else // NO_TRANSFORMS
      pme->Transform(m_mGobj_xy, m_mGxy_obj);
#endif // NO_TRANSFORMS
    }

  iterLists.Init(&m_listCSLIDEPointLists);
  for ( plist = (CListInt*)iterLists.PeekFirst(); 
	plist != NULL;
	plist = (CListInt*)iterLists.PeekNext() )
    {
      //////////////////////////////////////////////////
      // Offset Contours
      if ( pme != NULL )
	{
	  ::WriteLSIFOffsetContours(pme, fpOutputFile, uiIndentLevel);
	}
    }

  if ( pme != NULL )
    {
      DeallocateQEMesh(pme);
    }
}

VOID CSLIDEGeneralFace::WriteRIBGeometry(FILE *fpOutputFile,
				    const Matrix &/*mGproj_obj*/, 
				    CBound * /*pCBoundClip*/, 
				    SLIDEShadingType eShading,
				    SLIDELODType eLOD)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBPushSolidAttributes(fpOutputFile,
			      (SLIDESolidType)m_dyniSolid.GetValue());
  {
    WriteRIBGeometryPolygonal(fpOutputFile, eShading, eLOD);
  }
  WriteRIBPopSolidAttributes(fpOutputFile, 
			     (SLIDESolidType)m_dyniSolid.GetValue());
}

BOOL CSLIDEGeneralFace::CheckRIBPrimitive()
{
  BOOL bResult;
  //UINT_32 i;

  bResult = FALSE;
/*
  for ( i = 0; i < (UINT_32)slideribotSize; i++ )
    {
      bResult = bResult || (m_pcID == gcppcSLIDERIBGeneralFaceReservedWords[i]);  
    }
*/
  return bResult;
}

VOID CSLIDEGeneralFace::WriteRIBGeometryPolygonal(FILE *fpOutputFile,
					     SLIDEShadingType eShading,
					     SLIDELODType /* eLOD */)
{
  //  CListIterInt iter;
  //CSLIDEPoint *pCSLIDEPoint;

  ASSERT( fpOutputFile != NULL );
/*
  iter.Init(&m_listCSLIDEPoint);
  
  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      pCSLIDEPoint->WriteRIB(fpOutputFile, eShading);
    }
*/
}

VOID CSLIDEGeneralFace::WriteRIBGeometryPrimitive(FILE *fpOutputFile,
					     SLIDEShadingType /* eShading */,
					     SLIDELODType /* eLOD */)
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDEGeneralFace::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE GENERAL FACE: %s", m_pcID);
  WriteRIBBegin(fpOutputFile);
}

VOID CSLIDEGeneralFace::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE GENERAL FACE: %s", m_pcID);
}

VOID CSLIDEGeneralFace::WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &/*mTransform*/,
					  CSLIDESurface *pCSLIDESurface)
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDEGeneralFace::WriteSTLGeometry(FILE *fpOutputFile, const Matrix &/*mTransform*/)
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDEGeneralFace::WriteDotID(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "generalface_%s", GetID());
}

VOID CSLIDEGeneralFace::WriteDotNode(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "  ");
  WriteDotID(fpOutputFile);
  fprintf(fpOutputFile, " [ label = \"%s\", shape = ellipse, peripheries = 2", GetID());
  if ( GetBoundDynamic() )
    {
      fprintf(fpOutputFile, ", color = red, fontcolor = red");
    }
  else if ( GetGeometryDynamic() )
    {
      fprintf(fpOutputFile, ", color = blue, fontcolor = blue");
    }
  fprintf(fpOutputFile, " ];\n");
}

VOID CSLIDEGeneralFace::WriteDotEdges(FILE * /*fpOutputFile*/)
{
}



