#ifndef __SLIDEGENERALFACE_H__
#define __SLIDEGENERALFACE_H__

#include "slideatomnode.h"

class CSLIDEPoint;

//////////////////////////////////////////////////////////////////////////////
// CSLIDEGeneralFace Class
//

class CSLIDEGeneralFace : public CSLIDEAtomNode
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEGeneralFace();
  virtual ~CSLIDEGeneralFace();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID, 
		    CListInt *plistPointIDs);
  virtual VOID Uninit();
protected:
  VOID UninitLocal();

public:
  virtual BOOL Initialized();

  VOID InsertContour(CListInt *plistPointIDs);
  inline VOID SetType(DynInt *pdyniType);
  inline VOID SetVoronoi(DynInt *pdyniVoronoi);
  inline VOID SetAccuracy(DynInt *pdyniAccuracy);
  inline VOID SetRandomize(DynInt *pdyniRandomize);
  inline VOID SetCleanUpContour(DynInt *pdyniCleanUpContour);

  inline VOID SetMillingContour(DynInt *pdyniMillingContour);
  inline VOID SetOffsets(DynInt *pdyniOffsets);
  inline VOID SetOffsetMin(DynFloat *pdynfOffsetMin);
  inline VOID SetOffsetMax(DynFloat *pdynfOffsetMax);

  inline VOID SetIndividualVoronoiMin(DynInt *pdyniIndividualVoronoiMin);
  inline VOID SetIndividualVoronoiMax(DynInt *pdyniIndividualVoronoiMax);

  inline VOID SetDrawContours(DynInt *pdyniDrawContours);
  inline VOID SetDrawVertices(DynInt *pdyniDrawVertices);
  inline VOID SetDrawCDT(DynInt *pdyniDrawCDT);
  inline VOID SetDrawIndividualVoronoi(DynInt *pdyniDrawIndividualVoronoi);
  inline VOID SetDrawVoronoi(DynInt *pdyniDrawVoronoi);
  inline VOID SetDrawHeight(DynInt *pdyniDrawHeight);
  inline VOID SetDrawOffset(DynInt *pdyniDrawOffset);
  inline VOID SetDrawMillingContour(DynInt *pdyniDrawMillingContour);

  inline VOID SetSubdivisionLevel(DynInt *pdyniSubdivisionLevel);
  inline VOID SetTolerance(DynFloat *pdynfTolerance);

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual CSLIDECamera *FindCamera(CListIterInt *piterIDs, 
				   CListInt *plistCSLIDEInstance);
  virtual CSLIDELight *FindLight(CListIterInt *piterIDs, 
				 CListInt *plistCSLIDEInstance);
  virtual CSLIDEAreaLight *FindAreaLight(CListIterInt *piterIDs, 
					 CListInt *plistCSLIDEInstance);

  virtual VOID Preprocess();
  virtual VOID PreprocessGraph();

  virtual VOID Update();
  virtual VOID UpdateGraph();

#ifdef __USE_CDA__
  virtual VOID CalculateCDAVertices(CListInt *plistCDAVertex, 
				    const Matrix &mGroot_obj);
#endif // __USE_CDA__

  virtual QEMesh *GetQEMesh();

protected:
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();
  virtual Vector CalculateNormal(CListInt *plistCSLIDEPoint);
  virtual Point CalculateCentroid(CListInt *plistCSLIDEPoint);
  virtual VOID CalculateBound();

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID PrerenderGeometry(SLIDEShadingType eShading);
  virtual VOID PrerenderNormals(SLIDEShadingType eShading);

  virtual VOID RenderGeometry(const Matrix &mGproj_obj,
			      CBound *pCBoundClip, 
			      SLIDEShadingType eShading,
			      SLIDELODType eLOD,
			      CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderNormals(SLIDEShadingType eShading);

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);
  virtual VOID WriteLSIF(FILE *fpOutputFile, UINT_32 uiIndentLevel);
  virtual VOID WriteLSIFOffset(FILE *fpOutputFile, UINT_32 uiIndentLevel);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);

  BOOL CheckRIBPrimitive();
  VOID WriteRIBGeometryPolygonal(FILE *fpOutputFile,
				 SLIDEShadingType eShading,
				 SLIDELODType eLOD);
  VOID WriteRIBGeometryPrimitive(FILE *fpOutputFile,
				 SLIDEShadingType eShading,
				 SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

  virtual VOID WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				 CSLIDESurface *pCSLIDESurface);
  virtual VOID WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform);

  virtual VOID WriteDotID(FILE *fpOutputFile);
  virtual VOID WriteDotNode(FILE *fpOutputFile);
  virtual VOID WriteDotEdges(FILE *fpOutputFile);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline BOOL GetChildrenBoundsDynamic();
  inline VOID SetChildrenBoundsDynamic(BOOL bChildrenBoundsDynamic);

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CListInt m_listPointIDLists;
  CListInt m_listCSLIDEPointLists;
  CListInt m_listQEVertexLists;

  DynInt m_dyniType;
  DynInt m_dyniVoronoi;
  DynInt m_dyniAccuracy;
  DynInt m_dyniRandomize;
  DynInt m_dyniCleanUpContour;

  DynInt m_dyniMillingContour;
  DynInt m_dyniOffsets;
  DynFloat m_dynfOffsetMin;
  DynFloat m_dynfOffsetMax;

  DynInt m_dyniIndividualVoronoiMin;
  DynInt m_dyniIndividualVoronoiMax;

  DynInt m_dyniDrawContours;
  DynInt m_dyniDrawVertices;
  DynInt m_dyniDrawCDT;
  DynInt m_dyniDrawIndividualVoronoi;
  DynInt m_dyniDrawVoronoi;
  DynInt m_dyniDrawHeight;
  DynInt m_dyniDrawOffset;
  DynInt m_dyniDrawMillingContour;

  DynInt m_dyniSubdivisionLevel;
  DynFloat m_dynfTolerance;

  QEMesh **m_ppmeContourVoronoi;
  QEMesh *m_pmeVoronoi;
  QEMesh *m_pmeOffset;
  QEMesh *m_pmeMillingContour;

  Matrix m_mGxy_obj;
  Matrix m_mGobj_xy;
  Vector m_vNormal;
  Point m_ptCentroid;

  BOOL m_bChildrenBoundsDynamic;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDEGeneralFace::SetType(DynInt *pdyniType)
{
  m_dyniType.Set(pdyniType);
}

inline VOID CSLIDEGeneralFace::SetVoronoi(DynInt *pdyniVoronoi)
{
  m_dyniVoronoi.Set(pdyniVoronoi);
}

inline VOID CSLIDEGeneralFace::SetAccuracy(DynInt *pdyniAccuracy)
{
  m_dyniAccuracy.Set(pdyniAccuracy);
}

inline VOID CSLIDEGeneralFace::SetRandomize(DynInt *pdyniRandomize)
{
  m_dyniRandomize.Set(pdyniRandomize);
}

inline VOID CSLIDEGeneralFace::SetCleanUpContour(DynInt *pdyniCleanUpContour)
{
  m_dyniCleanUpContour.Set(pdyniCleanUpContour);
}

inline VOID CSLIDEGeneralFace::SetMillingContour(DynInt *pdyniMillingContour)
{
  m_dyniMillingContour.Set(pdyniMillingContour);
}

inline VOID CSLIDEGeneralFace::SetOffsets(DynInt *pdyniOffsets)
{
  m_dyniOffsets.Set(pdyniOffsets);
}

inline VOID CSLIDEGeneralFace::SetOffsetMin(DynFloat *pdynfOffsetMin)
{
  m_dynfOffsetMin.Set(pdynfOffsetMin);
}

inline VOID CSLIDEGeneralFace::SetOffsetMax(DynFloat *pdynfOffsetMax)
{
  m_dynfOffsetMax.Set(pdynfOffsetMax);
}

inline VOID CSLIDEGeneralFace::SetIndividualVoronoiMin(DynInt *pdyniIndividualVoronoiMin)
{
  m_dyniIndividualVoronoiMin.Set(pdyniIndividualVoronoiMin);
}

inline VOID CSLIDEGeneralFace::SetIndividualVoronoiMax(DynInt *pdyniIndividualVoronoiMax)
{
  m_dyniIndividualVoronoiMax.Set(pdyniIndividualVoronoiMax);
}

inline VOID CSLIDEGeneralFace::SetDrawContours(DynInt *pdyniDrawContours)
{
  m_dyniDrawContours.Set(pdyniDrawContours);
}

inline VOID CSLIDEGeneralFace::SetDrawVertices(DynInt *pdyniDrawVertices)
{
  m_dyniDrawVertices.Set(pdyniDrawVertices);
}

inline VOID CSLIDEGeneralFace::SetDrawCDT(DynInt *pdyniDrawCDT)
{
  m_dyniDrawCDT.Set(pdyniDrawCDT);
}

inline VOID CSLIDEGeneralFace::SetDrawIndividualVoronoi(DynInt *pdyniDrawIndividualVoronoi)
{
  m_dyniDrawIndividualVoronoi.Set(pdyniDrawIndividualVoronoi);
}

inline VOID CSLIDEGeneralFace::SetDrawVoronoi(DynInt *pdyniDrawVoronoi)
{
  m_dyniDrawVoronoi.Set(pdyniDrawVoronoi);
}

inline VOID CSLIDEGeneralFace::SetDrawHeight(DynInt *pdyniDrawHeight)
{
  m_dyniDrawHeight.Set(pdyniDrawHeight);
}

inline VOID CSLIDEGeneralFace::SetDrawOffset(DynInt *pdyniDrawOffset)
{
  m_dyniDrawOffset.Set(pdyniDrawOffset);
}

inline VOID CSLIDEGeneralFace::SetDrawMillingContour(DynInt *pdyniDrawMillingContour)
{
  m_dyniDrawMillingContour.Set(pdyniDrawMillingContour);
}

inline VOID CSLIDEGeneralFace::SetSubdivisionLevel(DynInt *pdyniSubdivisionLevel)
{
  m_dyniSubdivisionLevel.Set(pdyniSubdivisionLevel);
}

inline VOID CSLIDEGeneralFace::SetTolerance(DynFloat *pdynfTolerance)
{
  m_dynfTolerance.Set(pdynfTolerance);
}

inline BOOL CSLIDEGeneralFace::GetChildrenBoundsDynamic()
{
  return m_bChildrenBoundsDynamic;
}

inline VOID CSLIDEGeneralFace::SetChildrenBoundsDynamic(BOOL bChildrenBoundsDynamic)
{
  m_bChildrenBoundsDynamic = bChildrenBoundsDynamic;
}

#endif // __SLIDEGENERALFACE_H__
