#include "slidegeodesic.h"

#include "slidesurface.h"
#include "slideinstance.h"

#include "qemesh.h"
#include "qerender.h"
#include "qewrite.h"

extern BOOL gbPreprocessDone;

static const UINT_32 cuiDrawControlMesh = FALSE;
static const FLOAT cfTheta = 0.0;
static const UINT_32 cuiIntersections = 0;

//////////////////////////////////////////////////////////////////////
// CSLIDEGeodesic Class
//

CSLIDEGeodesic::CSLIDEGeodesic()
{
  m_dyniDrawControlMesh.SetValue(cuiDrawControlMesh);
  m_dynfTheta.SetValue(cfTheta);
  m_dyniIntersections.SetValue(cuiIntersections);
  m_pmeGeodesic = NULL;
}

CSLIDEGeodesic::~CSLIDEGeodesic()
{
}

VOID CSLIDEGeodesic::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			     CHAR *pcID)
{
  CSLIDEGroup::Init(pcFile, uiLine, uiColumn, pcID);

  m_dyniDrawControlMesh.SetValue(cuiDrawControlMesh);
  m_dynfTheta.SetValue(cfTheta);
  m_dyniIntersections.SetValue(cuiIntersections);
  m_pmeGeodesic = NULL;
}

VOID CSLIDEGeodesic::Uninit()
{
  CSLIDEGroup::Uninit();

  if ( m_pmeGeodesic != NULL )
    {
      DeallocateQEMesh(m_pmeGeodesic);
      m_pmeGeodesic = NULL;
    }
}

BOOL CSLIDEGeodesic::Initialized()
{
  if ( !CSLIDEGroup::Initialized() )
    {
      return FALSE;
    }

  return m_pmeGeodesic != NULL;
}

BOOL CSLIDEGeodesic::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;

  if ( !CSLIDEGroup::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  if ( !m_dyniDrawControlMesh.GetDynamic() &&
       !( m_dyniDrawControlMesh.GetValue() == TRUE || 
	  m_dyniDrawControlMesh.GetValue() == FALSE ) )
    {
      HandleError("Illegal drawcontrolmesh value:  %ld", 
      	  m_dyniDrawControlMesh.GetValue());
      m_dyniDrawControlMesh.SetValue(cuiDrawControlMesh);
      bCorrect = FALSE;
    }

  m_dyniIntersections.Update();
  if ( !( 0 <= m_dyniIntersections.GetValue() ) )
    {
      HandleError("Illegal intersections flag value:  %ld", 
		  m_dyniIntersections.GetValue());
      m_dyniIntersections.SetValue(cuiIntersections);
    }

  return bCorrect;
}

BOOL CSLIDEGeodesic::Valid()
{
  if ( !CSLIDEGroup::Valid() )
    {
      return FALSE;
    }

  return TRUE;
}

CSLIDECamera *CSLIDEGeodesic::FindCamera(CListIterInt *piterIDs, 
					     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtCamera],
	      gcppcSLIDEReservedWords[sliderwtGeodesic], GetID());

  return NULL;
}

CSLIDELight *CSLIDEGeodesic::FindLight(CListIterInt *piterIDs, 
				     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtLight],
	      gcppcSLIDEReservedWords[sliderwtGeodesic], GetID());

  return NULL;
}

CSLIDEAreaLight *CSLIDEGeodesic::FindAreaLight(CListIterInt *piterIDs, 
					    CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtAreaLight],
	      gcppcSLIDEReservedWords[sliderwtGeodesic], GetID());

  return NULL;
}

VOID CSLIDEGeodesic::Preprocess()
{
  CSLIDEGroup::Preprocess();
}

VOID CSLIDEGeodesic::CheckDynamic()
{
  SetDynamic(m_dyniShading.GetDynamic() || 
	     m_dyniLOD.GetDynamic() ||
             m_dyniDrawControlMesh.GetDynamic() ||
             m_dyniIntersections.GetDynamic() ||
             m_dynfTheta.GetDynamic());
}

VOID CSLIDEGeodesic::PreprocessGraph()
{
  if ( !GetVisited() )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;

      BOOL bGeometryEmpty;
      BOOL bGeometryDynamic;
      BOOL bNormalsEmpty;
      BOOL bNormalsDynamic;
      BOOL bBoundsEmpty;
      BOOL bBoundsDynamic;
      BOOL bBoundEmpty;
      BOOL bBoundDynamic;

      BOOL bChildrenBoundsDynamic;

      bGeometryEmpty = TRUE;
      bGeometryDynamic = FALSE;
      bNormalsEmpty = TRUE;
      bNormalsDynamic = FALSE;
      bBoundsEmpty = TRUE;
      bBoundsDynamic = FALSE;

      bBoundEmpty = TRUE;
      bBoundDynamic = FALSE;

      //////////////////////////////////////////////////
      // Visit Children Nodes and get starting values from them
      iter.Init(&m_listCSLIDEInstance);
      
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pCSLIDEInstance->PreprocessGraph();

	  bGeometryEmpty = ( bGeometryEmpty && 
			     pCSLIDEInstance->GetGeometryEmpty() );
	  bGeometryDynamic = ( bGeometryDynamic ||
			       pCSLIDEInstance->GetGeometryDynamic() );
	  bNormalsEmpty = ( bNormalsEmpty && 
			    pCSLIDEInstance->GetNormalsEmpty() );
	  bNormalsDynamic = ( bNormalsDynamic ||
			      pCSLIDEInstance->GetNormalsDynamic() );
	  bBoundsEmpty = ( bBoundsEmpty && 
			   pCSLIDEInstance->GetBoundsEmpty() );
	  bBoundsDynamic = ( bBoundsDynamic ||
			     pCSLIDEInstance->GetBoundsDynamic() );
	  bBoundEmpty = ( bBoundEmpty && 
			  pCSLIDEInstance->GetBoundEmpty() );
	  bBoundDynamic = ( bBoundDynamic ||
			    pCSLIDEInstance->GetBoundDynamic() );
	}

      //////////////////////////////////////////////////
      // Look at Local flags

      // Update whether we are Empty based on our LOD flag
      if ( !m_dyniLOD.GetDynamic() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      // This node is always off
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	      bBoundsEmpty = TRUE;
	      bBoundEmpty = TRUE;
	    }
	  else if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	    {
	      // This node is always just a bounding box
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	    }
	}

      // Check what is really dynamic

      bChildrenBoundsDynamic = bBoundsDynamic;

      bGeometryDynamic = ( bGeometryDynamic ||
			   m_pCSLIDESurface->GetDynamic() ||
			   m_dyniShading.GetDynamic() ||
			   m_dyniLOD.GetDynamic() );
      bNormalsDynamic = ( bNormalsDynamic ||
			  m_dyniShading.GetDynamic() ||
			  m_dyniLOD.GetDynamic() );
      bBoundsDynamic = ( bBoundsDynamic || 
			 m_dyniLOD.GetDynamic() );
      bBoundDynamic = ( bBoundDynamic || 
			m_dyniLOD.GetDynamic() );

      // If anything is now Empty then it will always be empty so
      // it is not dynamic

      bChildrenBoundsDynamic = ( bChildrenBoundsDynamic && !bBoundsEmpty );

      bGeometryDynamic = ( bGeometryDynamic && !bGeometryEmpty );
      bNormalsDynamic = ( bNormalsDynamic && !bNormalsEmpty );
      bBoundsDynamic = ( bBoundsDynamic && !bBoundsEmpty );
      bBoundDynamic = ( bBoundDynamic && !bBoundEmpty );

      //////////////////////////////////////////////////
      // Now save these correct flag values

      SetChildrenBoundsDynamic(bChildrenBoundsDynamic);

      SetGeometryEmpty(bGeometryEmpty);
      SetGeometryDynamic(bGeometryDynamic);
      SetNormalsEmpty(bNormalsEmpty);
      SetNormalsDynamic(bNormalsDynamic);
      SetBoundsEmpty(bBoundsEmpty);
      SetBoundsDynamic(bBoundsDynamic);
      SetBoundEmpty(bBoundEmpty);
      SetBoundDynamic(bBoundDynamic);
      
      //////////////////////////////
      // START: Jordan Fix This!!!
      //

      SetGeometryEmpty(FALSE);
      SetGeometryDynamic(TRUE);
      SetNormalsEmpty(FALSE);
      SetNormalsDynamic(TRUE);
      SetBoundsEmpty(FALSE);
      SetBoundsDynamic(TRUE);
      SetBoundEmpty(FALSE);
      SetBoundDynamic(TRUE);
      
      CalculateQEMesh();
      
      m_pmeGeodesic = m_pQEMesh->Copy();

      //
      // END: Jordan Fix This!!!
      //////////////////////////////

      //////////////////////////////////////////////////
      // Calculate the Bounding Box (if Empty this will do nothing)
      if ( GetBoundEmpty() )
	{
	  SetBoundValid(TRUE);
	}
      else
	{
	  CalculateBound();
	}
      SetBoundChanged(FALSE);

      //////////////////////////////////////////////////
      // Done Preprocessing this Node in the Graph
      SetVisited(TRUE);
    }
}

VOID CSLIDEGeodesic::Update()
{
  SetBoundChanged(FALSE);
  if ( GetDynamic() ||
       GetChildrenBoundsDynamic() )
    {
      if ( GetChildrenBoundsDynamic() )
	{
	  CSLIDEBoundedNode::InvalidateQEMesh();
	}
      UpdateState();
    }

  if ( GetBoundDynamic() )
    {
      SetBoundValid(FALSE);
    }

  SetVisited(FALSE);
}

VOID CSLIDEGeodesic::UpdateState()
{
  BOOL bMeshChanged, b;

  CSLIDEGroup::UpdateState();

  bMeshChanged = FALSE;

  m_dyniDrawControlMesh.Update();

  b = m_dynfTheta.Update();
  bMeshChanged = ( bMeshChanged || b );

  m_dyniIntersections.Update();
  bMeshChanged = ( bMeshChanged || b );
  if ( !( 0 <= m_dyniIntersections.GetValue() ) )
    {
      HandleError("Illegal intersections value:  %ld", 
		  m_dyniIntersections.GetValue());
      m_dyniIntersections.SetValue(cuiIntersections);
    }

  if ( !gbPreprocessDone )
    {
      return;
    }

  if ( m_pQEMesh == NULL )
    {
      CalculateQEMesh();
      bMeshChanged = TRUE;
    }

  ASSERT( m_pQEMesh != NULL );
  if ( bMeshChanged ||
       m_pmeGeodesic == NULL )
    {
      if ( m_pmeGeodesic != NULL )
	{
	  DeallocateQEMesh(m_pmeGeodesic);
	  m_pmeGeodesic = NULL;
	}

      m_pmeGeodesic = m_pQEMesh->Copy();
    }
  ASSERT( m_pmeGeodesic != NULL );
}

VOID CSLIDEGeodesic::UpdateGraph()
{
  if ( !GetVisited() )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;
      BOOL bBoundChanged;

      if ( !GetBoundValid() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      if ( !m_bound.GetEmpty() )
		{
		  SetBoundChanged(TRUE);
		  m_bound.Null();
		}
	      return;
	    }
	  
	  bBoundChanged = FALSE;
	  
	  iter.Init(&m_listCSLIDEInstance);
	  
	  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
		pCSLIDEInstance != NULL;
		pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	    {
	      pCSLIDEInstance->UpdateGraph();
	      bBoundChanged = bBoundChanged || 
		pCSLIDEInstance->GetBoundChanged();
	    }
	  
	  if ( bBoundChanged )
	    {
	      CalculateBound();
	    }
	  
	  SetBoundValid(TRUE);
	}

      SetVisited(TRUE);
    }
}

VOID CSLIDEGeodesic::CalculateBound()
{
  CBound bound;

/*
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  
  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->InsertIntoBound(bound);
    }
*/
  ///////////////////////////////////////
  // START: Jordan Fix This!!!
  //
  {
    QEMeshIter iterQEMesh;
    QEVertex *pvx;
    Point pt;

    iterQEMesh.Init(m_pmeGeodesic);
    for ( pvx = iterQEMesh.PeekFirstVertex();
	  pvx != NULL;
	  pvx = iterQEMesh.PeekNextVertex() )
      {
	pt = pvx->GetPoint();
	bound.Insert(pt);
      }
  }
  //
  // END: Jordan Fix This!!!
  ///////////////////////////////////////

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);
    }
}

VOID CSLIDEGeodesic::PrerenderGeometry(SLIDEShadingType eShading)
{
  CListIterInt iter;

  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  PushShadingAttributes(eShading);
  {
    m_pCSLIDESurface->Push();
    {
      ::RenderGeometry(m_pmeGeodesic);
    }
    m_pCSLIDESurface->Pop();
  }
  PopShadingAttributes(eShading);

  if ( m_dyniDrawControlMesh.GetValue() )
    {
      QEMeshIter iterQEMesh;
      QEEdge *pe;

      iterQEMesh.Init(m_pQEMesh);

      PushAuxillaryGeometryAttributes();
      {
	gilColor3f(1.0, 1.0, 0.0);
	gilBegin(GL_LINES);
	{
	  for ( pe = iterQEMesh.PeekFirstEdge();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdge() )
	    {
	      gilVertex3fv(pe->GetVertex()->GetPoint().Get());
	      gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
	    }
	  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdgeBoundary() )
	    {
	      gilVertex3fv(pe->GetVertex()->GetPoint().Get());
	      gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
	    }
	}
	gilEnd();
      }
      PopAuxillaryGeometryAttributes();
    }
}

VOID CSLIDEGeodesic::PrerenderNormals(SLIDEShadingType eShading)
{
  CListIterInt iter;

  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  ::RenderNormals(m_pmeGeodesic);
}

VOID CSLIDEGeodesic::PrerenderBounds()
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
    {
    }
  else 
    {
      if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	{
	}
      else
	{
	  iter.Init(&m_listCSLIDEInstance);
	  
	  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
		pCSLIDEInstance != NULL;
		pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	    {
	      pCSLIDEInstance->PrerenderBounds();
	    }
	}

      m_bound.Render();
    }
}

VOID CSLIDEGeodesic::Prerender()
{
  if ( !(!m_dyniLOD.GetDynamic() && 
	 ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
	   m_dyniLOD.GetValue() == slidelodtSLF_BOUND )) )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;
      
      iter.Init(&m_listCSLIDEInstance);
      
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pCSLIDEInstance->Prerender();
	}
    }
      
  CSLIDEBoundedNode::Prerender();
}

VOID CSLIDEGeodesic::RenderGeometry(const Matrix &mGproj_obj, 
				  CBound *pCBoundClip, 
				  SLIDEShadingType /*eShading*/,
				  SLIDELODType eLOD,
				  CSLIDEWormwarp *pCSLIDEWarp/* = NULL*/)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  // For Debugging
  if ( GetRenderCDAOperandsEnabled() )
    {
      gilPushAttrib(GL_ALL_ATTRIB_BITS);
      {
	PushAuxillaryGeometryAttributes();
	{
	  PushWireFrameAttributes();
	  {
	    iter.Init(&m_listCSLIDEInstance);
	    
	    for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
		pCSLIDEInstance != NULL;
		  pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	      {
		pCSLIDEInstance->Render(mGproj_obj, pCBoundClip, slideshtSLF_WIRE, eLOD);
	      }
	  }
	  PopWireFrameAttributes();
	}
	PopAuxillaryGeometryAttributes();
      }
      gilPopAttrib();
    }
      
  ::RenderGeometry(m_pmeGeodesic);

  if ( m_dyniDrawControlMesh.GetValue() )
    {
      QEMeshIter iterQEMesh;
      QEEdge *pe;

      iterQEMesh.Init(m_pQEMesh);

      PushAuxillaryGeometryAttributes();
      {
	gilColor3f(1.0, 1.0, 0.0);
	gilBegin(GL_LINES);
	{
	  for ( pe = iterQEMesh.PeekFirstEdge();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdge() )
	    {
	      gilVertex3fv(pe->GetVertex()->GetPoint().Get());
	      gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
	    }
	  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdgeBoundary() )
	    {
	      gilVertex3fv(pe->GetVertex()->GetPoint().Get());
	      gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
	    }
	}
	gilEnd();
      }
      PopAuxillaryGeometryAttributes();
    }
}

VOID CSLIDEGeodesic::RenderNormals(SLIDEShadingType /*eShading*/)
{
  if ( m_pCDA != NULL )
    {
      PushAuxillaryGeometryAttributes();
      {
	gilColor3f(1.0, 0.0, 1.0);
	
	::RenderNormals(m_pmeGeodesic);
      }
      PopAuxillaryGeometryAttributes();
    }
}

VOID CSLIDEGeodesic::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtGeodesic], m_pcID);

  WriteSLIDEFlags(fpOutputFile, 0);

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->WriteSLIDE(fpOutputFile);    
    }

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndGeodesic]);
}

VOID CSLIDEGeodesic::WriteRIBGeometry(FILE *fpOutputFile,
				      const Matrix &mGproj_obj, 
				      CBound *pCBoundClip, 
				      SLIDEShadingType eShading,
				      SLIDELODType eLOD)
{
/*
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( fpOutputFile != NULL );

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->WriteRIB(fpOutputFile,
				mGproj_obj, pCBoundClip, eShading, eLOD);    
    }
    */
}

VOID CSLIDEGeodesic::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );
  /*
  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE GEODESIC: %s", m_pcID);
  WriteRIBBegin(fpOutputFile);
  */
}

VOID CSLIDEGeodesic::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );
  /*
  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE GEODESIC: %s", m_pcID);
  */
}

VOID CSLIDEGeodesic::WriteWRLZ(FILE *fpOutputFile, const Matrix &m, 
			       CSLIDESurface *pCSLIDESurface)
{
  ::WriteWRLZ(m_pmeGeodesic, fpOutputFile, m);
}

VOID CSLIDEGeodesic::WriteSTL(FILE *fpOutputFile, const Matrix &m)
{
  ::WriteSTL(m_pmeGeodesic, fpOutputFile, m);
}






