#include "slidegeometrynode.h"

#include "slideworld.h"
#include "slidesurface.h"
#include "slideinstance.h"

//////////////////////////////////////////////////////////////////////
// CSLIDEGeometryNode Class
//

CSLIDEGeometryNode::CSLIDEGeometryNode()
{
  m_pcSurfaceID = NULL;
  m_pCSLIDESurface = NULL;

  m_bGUIDeletedChild = FALSE;

  m_uiPrerenderAttributesStackDepthMax = 0;
}

CSLIDEGeometryNode::~CSLIDEGeometryNode()
{
}

VOID CSLIDEGeometryNode::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID)
{
  CSLIDENode::Init(pcFile, uiLine, uiColumn, pcID);

  m_pcSurfaceID = gcppcSLIDEReservedWords[sliderwtSLF_INHERIT];
  m_pCSLIDESurface = NULL;
  m_dyniShading.SetValue(slideshtSLF_INHERIT);
  m_dyniLOD.SetValue(slidelodtSLF_FULL);

  m_uiPrerenderAttributesStackDepthMax = 0;
}

VOID CSLIDEGeometryNode::Uninit()
{
  CSLIDENode::Uninit();
}

BOOL CSLIDEGeometryNode::Initialized()
{
  if ( !CSLIDENode::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

BOOL CSLIDEGeometryNode::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  if ( !CSLIDENode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  // Check Surface Reference

  m_pCSLIDESurface = pCSLIDEWorld->PeekSurface(m_pcSurfaceID);
  if ( m_pCSLIDESurface == NULL )
    {
      HandleError("Undefined %s \"%s\"", 
		  gcppcSLIDEReservedWords[sliderwtSurface], 
		  m_pcSurfaceID);
      return FALSE;
    }

  // Check Shading Type

  ASSERT( 0 <= m_dyniShading.GetValue() &&
	  m_dyniShading.GetValue() < slideshtSize );

  // Check LOD Type

  ASSERT( 0 <= m_dyniLOD.GetValue() &&
	  m_dyniLOD.GetValue() < slidelodtSize );

  return TRUE;
}

BOOL CSLIDEGeometryNode::Valid()
{
  if ( !CSLIDENode::Valid() )
    {
      return FALSE;
    }

  return ( ( 0 <= m_dyniShading.GetValue() &&
	     m_dyniShading.GetValue() < slideshtSize ) &&
	   ( 0 <= m_dyniLOD.GetValue() &&
	     m_dyniLOD.GetValue() < slidelodtSize ) );
}

VOID CSLIDEGeometryNode::Preprocess()
{
  SetVisited(FALSE);
  SetGeometryEmpty(TRUE);
  SetGeometryDynamic(FALSE);
  SetNormalsEmpty(TRUE);
  SetNormalsDynamic(FALSE);
  SetBoundsEmpty(TRUE);
  SetBoundsDynamic(FALSE);
  SetBoundEmpty(TRUE);
  SetBoundDynamic(FALSE);
  SetBoundValid(FALSE);
  SetBoundChanged(FALSE);
}

VOID CSLIDEGeometryNode::SetBoundUnvisited()
{
  if( !GetVisited() )
    {
      return;
    }
  
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  SetVisited(FALSE);
  SetBoundValid(FALSE);
  
  iter.Init(&m_listCSLIDEParentInstance);
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->SetBoundUnvisited();
    }
}

VOID CSLIDEGeometryNode::SetBoundUnchanged()
{
  if( !GetBoundChanged() )
    {
      return;
    }
  
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  SetBoundChanged(FALSE);
  m_bGUIDeletedChild = FALSE;
    
  iter.Init(&m_listCSLIDEParentInstance);
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->SetBoundUnchanged();
    }
}

VOID CSLIDEGeometryNode::CheckDynamic()
{
  // Are there variables which I must call TCL to update?
  SetDynamic(GetDynamic() || 
	     m_dyniShading.GetDynamic() ||
	     m_dyniLOD.GetDynamic());
}

VOID CSLIDEGeometryNode::UpdateState()
{
  BOOL b;

  m_bChanged = FALSE;

  b = m_dyniShading.Update();
  m_bChanged = ( m_bChanged || b );
  if ( !( 0 <= m_dyniShading.GetValue() &&
	  m_dyniShading.GetValue() < slideshtSize ) )
    {
      HandleError("Illegal Shading flag value:  %ld", 
		  m_dyniShading.GetValue());
      m_dyniShading.SetValue(slideshtSLF_FLAT);
    }

  b = m_dyniLOD.Update();
  m_bChanged = ( m_bChanged || b );
  if ( !( 0 <= m_dyniLOD.GetValue() &&
	  m_dyniLOD.GetValue() < slideshtSize ) )
    {
      HandleError("Illegal LOD flag value:  %ld", 
		  m_dyniLOD.GetValue());
      m_dyniLOD.SetValue(slidelodtSLF_FULL);
    }
}

GLbitfield CSLIDEGeometryNode::CalculateAttributesMask(SLIDEShadingType eShading)
{
  return ( CalculateShadingAttributesMask(eShading) |
	   m_pCSLIDESurface->CalculateAttributesMask() );
}

VOID CSLIDEGeometryNode::ModifyShadingFlag(SLIDEShadingType &eShading)
{
  if ( m_dyniShading.GetValue() != slideshtSLF_INHERIT )
    {
      eShading = (SLIDEShadingType)m_dyniShading.GetValue();
    }
}

VOID CSLIDEGeometryNode::ModifyLODFlag(SLIDELODType &eLOD)
{
  eLOD = (SLIDELODType)MIN( (INT_32)eLOD, m_dyniLOD.GetValue());
}

CBound::BoundIntersectionType CSLIDEGeometryNode::CheckBoundReject(const Matrix &/*m*/,
						        CBound * /*pCBoundClip*/)
{
  // Default to not doing any rejections
  return CBound::bitIntersect;
}

VOID CSLIDEGeometryNode::RenderNormals(SLIDEShadingType /*eShading*/)
{
}

VOID CSLIDEGeometryNode::WriteSLIDEFlags(FILE *fpOutputFile, UINT_32 uiIndent)
{
  ASSERT( fpOutputFile != NULL );

  CSLIDEPrimitive::WriteSLIDEFlags(fpOutputFile, uiIndent);

  if ( m_pcSurfaceID != gcppcSLIDEReservedWords[sliderwtSLF_INHERIT] )
    {
      WriteIndent(fpOutputFile, uiIndent);
      fprintf(fpOutputFile, "  %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtSurface], 
	      m_pcSurfaceID);
    } 
  if ( m_dyniShading.GetValue() != slideshtSLF_INHERIT )
    {
      WriteIndent(fpOutputFile, uiIndent);
      fprintf(fpOutputFile, "  %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtShading], 
	      gcppcSLIDEShadingReservedWords[m_dyniShading.GetValue()]);
    }
  if ( m_dyniLOD.GetValue() != slidelodtSLF_FULL )
    {
      WriteIndent(fpOutputFile, uiIndent);
      fprintf(fpOutputFile, "  %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtLOD], 
	      gcppcSLIDELODReservedWords[m_dyniLOD.GetValue()]);
    }
}

VOID CSLIDEGeometryNode::WriteRIBPushAttributes(FILE *fpOutputFile, 
						SLIDEShadingType eShading)
{
  WriteRIBPushSLIDEName(fpOutputFile);
  WriteRIBPushShadingAttributes(fpOutputFile, eShading);
  m_pCSLIDESurface->WriteRIBPush(fpOutputFile);
}

VOID CSLIDEGeometryNode::WriteRIBPopAttributes(FILE *fpOutputFile, 
					       SLIDEShadingType eShading)
{
  m_pCSLIDESurface->WriteRIBPop(fpOutputFile);
  WriteRIBPopShadingAttributes(fpOutputFile, eShading);
  WriteRIBPopSLIDEName(fpOutputFile);
}

UINT_32 CSLIDEGeometryNode::SetParameter(CSLIDEWorld *pCSLIDEWorld,
					 SLIDEReservedWordType uiParameter, CHAR **argv)
{
  UINT_32 uiValue;

  switch(uiParameter) 
    {
    case sliderwtSurface:
      CSLIDESurface *pCSLIDESurface;
      pCSLIDESurface = pCSLIDEWorld->PeekSurface(argv[0]);
      ASSERT(pCSLIDESurface != NULL);
      m_pCSLIDESurface = pCSLIDESurface;
      return 1;
      break;
    case sliderwtShading:
      for( uiValue=0; uiValue < slideshtSize; uiValue++ )
	{
	  if( !strcmp(argv[0], gcppcSLIDEGUIShadingReservedWords[uiValue]) )
	    {
	      break;
	    }
	}
      m_dyniShading.SetValue((SLIDEShadingType)uiValue);
      return 1;
      break;
    case sliderwtLOD:
      for( uiValue=0; uiValue < slideshtSize; uiValue++ )
	{
	  if( !strcmp(argv[0], gcppcSLIDEGUILODReservedWords[uiValue]) )
	    {
	      break;
	    }
	}
      m_dyniLOD.SetValue((SLIDELODType)uiValue);
      return 1;
      break;
    default:
      return CSLIDENode::SetParameter(pCSLIDEWorld, uiParameter, argv);      
      break;
    }
}

