#ifndef __SLIDEGEOMETRYNODE_H__
#define __SLIDEGEOMETRYNODE_H__

#include "slidenode.h"

class CSLIDESurface;

//////////////////////////////////////////////////////////////////////
// CSLIDEGeometryNode Class
//

class CSLIDEGeometryNode : public CSLIDENode
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//
public:

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEGeometryNode();
  virtual ~CSLIDEGeometryNode();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  //modified by WM
  inline CHAR* GetSurfaceID(); 

  inline VOID SetSurfaceID(CHAR *pcSurfaceID);
  inline VOID SetShading(DynInt *pdyniShading);
  inline VOID SetLOD(DynInt *pdyniLOD);

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual VOID Preprocess();
  virtual VOID SetBoundUnvisited();
  virtual VOID SetBoundUnchanged();

protected:
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();

public:
  virtual VOID CalculateLeafQEMesh() = 0;  // for WriteWRLZ
  virtual VOID DeallocateLeafQEMesh() = 0;  // for WriteWRLZ

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID PrerenderGeometry(SLIDEShadingType eShading) = 0;
  virtual VOID PrerenderNormals(SLIDEShadingType eShading) = 0;
  virtual VOID PrerenderBounds() = 0;

  virtual VOID RecallGeometry(SLIDEShadingType eShading) = 0;
  virtual VOID RecallNormals(SLIDEShadingType eShading) = 0;
  virtual VOID RecallBounds() = 0;

  virtual VOID Render(const Matrix &mGproj_obj,
		      CBound *pCBoundClip, 
		      SLIDEShadingType eShading,
		      SLIDELODType eLOD) = 0;

  virtual VOID RenderSelect(const Matrix &mGproj_obj,
			    CBound *pCBoundClip, 
			    SLIDEShadingType eShading,
			    SLIDELODType eLOD) = 0;

  virtual GLbitfield CalculateAttributesMask(SLIDEShadingType eShading);
  VOID ModifyShadingFlag(SLIDEShadingType &eShading);
  VOID ModifyLODFlag(SLIDELODType &eLOD);
  virtual CBound::BoundIntersectionType CheckBoundReject(const Matrix &m,
						        CBound *pCBound);

  virtual VOID RenderGeometry(const Matrix &mGproj_obj,
			      CBound *pCBoundClip, 
			      SLIDEShadingType eShading,
			      SLIDELODType eLOD,
			      CSLIDEWormwarp *pCSLIDEWarp = NULL) = 0;

  virtual VOID RenderGeometrySelect(const Matrix &mGproj_obj,
				    CBound *pCBoundClip, 
				    SLIDEShadingType eShading,
				    SLIDELODType eLOD) = 0;

  virtual VOID RenderNormals(SLIDEShadingType eShading);
  virtual VOID RenderBound() = 0;



//////////////////////////////////////////////////
// Output Functions
//
protected:
  virtual VOID WriteSLIDEFlags(FILE *fpOutputFile, UINT_32 uiIndent);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD) = 0;
  virtual VOID WriteRIBPushAttributes(FILE *fpOutputFile, 
				      SLIDEShadingType eShading);
  virtual VOID WriteRIBPopAttributes(FILE *fpOutputFile, 
				     SLIDEShadingType eShading);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile) = 0;
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile) = 0;

  virtual VOID WritePOVGeometry(FILE *fpOutputFile, 
				CListInt *plistTransformIDs,
				CListInt *plistColors) = 0;

  virtual VOID WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				 CSLIDESurface *pCSLIDESurface) = 0;

  virtual VOID WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform) = 0;

//////////////////////////////////////////////////
// Class Accessor Functions
//
public:

//////////////////////////////////////////////////
// Accessor Functions
//
public:

  inline CSLIDESurface *GetSurface();

  virtual inline BOOL GetVisited();
  virtual inline VOID SetVisited(BOOL bVisited);

  virtual inline BOOL GetGeometryEmpty();
  inline VOID SetGeometryEmpty(BOOL bGeometryEmpty);
  virtual inline BOOL GetGeometryDynamic();
  inline VOID SetGeometryDynamic(BOOL bGeometryDynamic);
  virtual inline BOOL GetNormalsEmpty();
  inline VOID SetNormalsEmpty(BOOL bNormalsEmpty);
  virtual inline BOOL GetNormalsDynamic();
  inline VOID SetNormalsDynamic(BOOL bNormalsDynamic);
  virtual inline BOOL GetBoundsEmpty();
  inline VOID SetBoundsEmpty(BOOL bBoundsEmpty);
  virtual inline BOOL GetBoundsDynamic();
  inline VOID SetBoundsDynamic(BOOL bBoundsDynamic);

  virtual inline BOOL GetBoundEmpty();
  inline VOID SetBoundEmpty(BOOL bBoundEmpty);
  virtual inline BOOL GetBoundDynamic();
  inline VOID SetBoundDynamic(BOOL bBoundDynamic);
  virtual inline BOOL GetBoundValid();
  inline VOID SetBoundValid(BOOL bBoundValid);
  virtual inline BOOL GetBoundChanged();
  inline VOID SetBoundChanged(BOOL bBoundChanged);

  virtual inline UINT_32 GetPrerenderAttributesStackDepthMax();

  virtual UINT_32 SetParameter(CSLIDEWorld *pCSLIDEWorld,
			       SLIDEReservedWordType uiParameter, CHAR **argv);

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CHAR *m_pcSurfaceID;
  CSLIDESurface *m_pCSLIDESurface;
  DynInt m_dyniShading;
  DynInt m_dyniLOD;

  BOOL m_bVisited;

  // Prerendering Flags
  BOOL m_bGeometryEmpty;
  BOOL m_bGeometryDynamic;

  BOOL m_bNormalsEmpty;
  BOOL m_bNormalsDynamic;

  BOOL m_bBoundsEmpty;
  BOOL m_bBoundsDynamic;

  // Bounding Box Flags
  BOOL m_bBoundEmpty;
  BOOL m_bBoundDynamic;
  BOOL m_bBoundValid;
  BOOL m_bBoundChanged;

  BOOL m_bGUIDeletedChild;

  // Attribute Stack Stuff
  UINT_32 m_uiPrerenderAttributesStackDepthMax;

  BOOL m_bChanged;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline CHAR* CSLIDEGeometryNode::GetSurfaceID() 
{
  return m_pcSurfaceID;
}

inline VOID CSLIDEGeometryNode::SetSurfaceID(CHAR *pcSurfaceID)
{
  m_pcSurfaceID = pcSurfaceID;
}

inline VOID CSLIDEGeometryNode::SetShading(DynInt *pdyniShading)
{
  m_dyniShading.Set(pdyniShading);
}

inline VOID CSLIDEGeometryNode::SetLOD(DynInt *pdyniLOD)
{
  m_dyniLOD.Set(pdyniLOD);
}

inline CSLIDESurface *CSLIDEGeometryNode::GetSurface()
{
  return m_pCSLIDESurface;
}

inline BOOL CSLIDEGeometryNode::GetVisited()
{
  return m_bVisited;
}

inline VOID CSLIDEGeometryNode::SetVisited(BOOL bVisited)
{
  m_bVisited = bVisited;
}

inline BOOL CSLIDEGeometryNode::GetGeometryEmpty()
{
  return m_bGeometryEmpty;
}

inline VOID CSLIDEGeometryNode::SetGeometryEmpty(BOOL bGeometryEmpty)
{
  m_bGeometryEmpty = bGeometryEmpty;
}

inline BOOL CSLIDEGeometryNode::GetGeometryDynamic()
{
  return m_bGeometryDynamic;
}

inline VOID CSLIDEGeometryNode::SetGeometryDynamic(BOOL bGeometryDynamic)
{
  m_bGeometryDynamic = bGeometryDynamic;
}

inline BOOL CSLIDEGeometryNode::GetNormalsEmpty()
{
  return m_bNormalsEmpty;
}

inline VOID CSLIDEGeometryNode::SetNormalsEmpty(BOOL bNormalsEmpty)
{
  m_bNormalsEmpty = bNormalsEmpty;
}

inline BOOL CSLIDEGeometryNode::GetNormalsDynamic()
{
  return m_bNormalsDynamic;
}

inline VOID CSLIDEGeometryNode::SetNormalsDynamic(BOOL bNormalsDynamic)
{
  m_bNormalsDynamic = bNormalsDynamic;
}

inline BOOL CSLIDEGeometryNode::GetBoundsEmpty()
{
  return m_bBoundsEmpty;
}

inline VOID CSLIDEGeometryNode::SetBoundsEmpty(BOOL bBoundsEmpty)
{
  m_bBoundsEmpty = bBoundsEmpty;
}

inline BOOL CSLIDEGeometryNode::GetBoundsDynamic()
{
  return m_bBoundsDynamic;
}

inline VOID CSLIDEGeometryNode::SetBoundsDynamic(BOOL bBoundsDynamic)
{
  m_bBoundsDynamic = bBoundsDynamic;
}

inline BOOL CSLIDEGeometryNode::GetBoundEmpty()
{
  return m_bBoundEmpty;
}

inline VOID CSLIDEGeometryNode::SetBoundEmpty(BOOL bBoundEmpty)
{
  m_bBoundEmpty = bBoundEmpty;
}

inline BOOL CSLIDEGeometryNode::GetBoundDynamic()
{
  return m_bBoundDynamic;
}

inline VOID CSLIDEGeometryNode::SetBoundDynamic(BOOL bBoundDynamic)
{
  m_bBoundDynamic = bBoundDynamic;
}

inline BOOL CSLIDEGeometryNode::GetBoundValid()
{
  return m_bBoundValid;
}

inline VOID CSLIDEGeometryNode::SetBoundValid(BOOL bBoundValid)
{
  m_bBoundValid = bBoundValid;
}

inline BOOL CSLIDEGeometryNode::GetBoundChanged()
{
  return m_bBoundChanged;
}

inline VOID CSLIDEGeometryNode::SetBoundChanged(BOOL bBoundChanged)
{
  m_bBoundChanged = bBoundChanged;
}

inline UINT_32 CSLIDEGeometryNode::GetPrerenderAttributesStackDepthMax()
{
  return m_uiPrerenderAttributesStackDepthMax;
}

#endif // __SLIDEGEOMETRYNODE_H__

