#include "slidegroup.h"

#include "slideworld.h"
#include "slidesurface.h"
#include "slideinstance.h"

#include "qemesh.h"

CSLIDESurface *m_pCSLIDESurfaceCurr = NULL;

//////////////////////////////////////////////////////////////////////
// CSLIDEGroup Class
//

CSLIDEGroup::CSLIDEGroup()
{
  m_pmeSIF = NULL;
}

CSLIDEGroup::~CSLIDEGroup()
{
}

VOID CSLIDEGroup::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		       CHAR *pcID)
{
  CSLIDEBoundedNode::Init(pcFile, uiLine, uiColumn, pcID);

  m_listCSLIDEInstance.Init();
  m_pmeSIF = NULL;
}

VOID CSLIDEGroup::Uninit()
{
  CSLIDEBoundedNode::Uninit();

  if ( m_pmeSIF != NULL )
    {
      DeallocateQEMesh(m_pmeSIF);
      m_pmeSIF = NULL;
    }  
}

BOOL CSLIDEGroup::Initialized()
{
  if ( !CSLIDEBoundedNode::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

VOID CSLIDEGroup::InsertInstance(CSLIDEInstance *pCSLIDEInstance)
{
  ASSERT( pCSLIDEInstance != NULL );

  m_listCSLIDEInstance.InsertLast(pCSLIDEInstance, (UINT_32)pCSLIDEInstance);
}

VOID CSLIDEGroup::RemoveChildInstance(CSLIDEInstance *pCSLIDEInstance)
{
  m_bGUIDeletedChild = TRUE;
  m_listCSLIDEInstance.RemoveUnordered((UINT_32)pCSLIDEInstance);
}

BOOL CSLIDEGroup::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  if ( !CSLIDEBoundedNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  return TRUE;
}

BOOL CSLIDEGroup::Valid()
{
  if ( !CSLIDEBoundedNode::Valid() )
    {
      return FALSE;
    }

  return TRUE;
}

VOID CSLIDEGroup::Remove(CSLIDEWorld *pCSLIDEWorld)
{
  CSLIDEPrimitive::Remove(pCSLIDEWorld);

  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  iter.Init(&m_listCSLIDEInstance);
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->RemoveFromChild(pCSLIDEWorld);
      pCSLIDEWorld->RemoveInstance(pCSLIDEInstance);
    }

  pCSLIDEWorld->RemoveNode(this);
  Uninit();
}

CSLIDECamera *CSLIDEGroup::FindCamera(CListIterInt *piterIDs, 
				      CListInt *plistCSLIDEInstance)
{
  CHAR *pcID;
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );
  
  pcID = (CHAR*)piterIDs->PeekNext();
  if ( pcID == NULL )
    {
      HandleError("Bad %s path: path ends in %s \"%s\"",
		  gcppcSLIDEReservedWords[sliderwtCamera],
		  gcppcSLIDEReservedWords[sliderwtGroup], GetID());
      return NULL;
    }
  
  iter.Init(&m_listCSLIDEInstance);
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      if ( pCSLIDEInstance->GetID() == pcID )
	{
	  plistCSLIDEInstance->InsertLast(pCSLIDEInstance, 0);
	  return pCSLIDEInstance->FindCamera(piterIDs, plistCSLIDEInstance);
	}
    }
  
  HandleError("Bad %s path: %s \"%s\" does not contain an %s \"%s\"",
	      gcppcSLIDEReservedWords[sliderwtCamera],
	      gcppcSLIDEReservedWords[sliderwtGroup], GetID(),
	      gcppcSLIDEReservedWords[sliderwtInstance], pcID);
  return NULL;
}

CSLIDELight *CSLIDEGroup::FindLight(CListIterInt *piterIDs, 
				       CListInt *plistCSLIDEInstance)
{
  CHAR *pcID;
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  pcID = (CHAR*)piterIDs->PeekNext();
  if ( pcID == NULL )
    {
      HandleError("Bad %s path: path ends in %s \"%s\"",
		  gcppcSLIDEReservedWords[sliderwtLight],
		  gcppcSLIDEReservedWords[sliderwtGroup], GetID());
      return NULL;
    }

  iter.Init(&m_listCSLIDEInstance);
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      if ( pCSLIDEInstance->GetID() == pcID )
	{
	  plistCSLIDEInstance->InsertLast(pCSLIDEInstance, 0);
	  return pCSLIDEInstance->FindLight(piterIDs, plistCSLIDEInstance);
	}
    }

  HandleError("Bad %s path: %s \"%s\" does not contain an %s \"%s\"",
	      gcppcSLIDEReservedWords[sliderwtLight],
	      gcppcSLIDEReservedWords[sliderwtGroup], GetID(),
	      gcppcSLIDEReservedWords[sliderwtInstance], pcID);
  return NULL;
}

CSLIDEAreaLight *CSLIDEGroup::FindAreaLight(CListIterInt *piterIDs, 
				       CListInt *plistCSLIDEInstance)
{
  CHAR *pcID;
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  pcID = (CHAR*)piterIDs->PeekNext();
  if ( pcID == NULL )
    {
      HandleError("Bad %s path: path ends in %s \"%s\"",
		  gcppcSLIDEReservedWords[sliderwtAreaLight],
		  gcppcSLIDEReservedWords[sliderwtGroup], GetID());
      return NULL;
    }

  iter.Init(&m_listCSLIDEInstance);
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      if ( pCSLIDEInstance->GetID() == pcID )
	{
	  plistCSLIDEInstance->InsertLast(pCSLIDEInstance, 0);
	  return pCSLIDEInstance->FindAreaLight(piterIDs, plistCSLIDEInstance);
	}
    }

  HandleError("Bad %s path: %s \"%s\" does not contain an %s \"%s\"",
	      gcppcSLIDEReservedWords[sliderwtAreaLight],
	      gcppcSLIDEReservedWords[sliderwtGroup], GetID(),
	      gcppcSLIDEReservedWords[sliderwtInstance], pcID);
  return NULL;
}

VOID CSLIDEGroup::GetSide(CListInt *plistPoint, CListInt *plistNormal, UINT_32 uiSide, 
			  const Matrix &mTransform, const Matrix &mInverse, BOOL bReverse)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  
  ASSERT( plistPoint != NULL );
  
  iter.Init(&m_listCSLIDEInstance);
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->GetSide(plistPoint, plistNormal, uiSide, mTransform, mInverse,
			       bReverse);
    }
}

VOID CSLIDEGroup::Preprocess()
{
  CSLIDEBoundedNode::Preprocess();

  CheckDynamic();
  UpdateState();
}

VOID CSLIDEGroup::CheckDynamic()
{
  SetDynamic(m_dyniShading.GetDynamic() || 
	     m_dyniLOD.GetDynamic());
}

VOID CSLIDEGroup::PreprocessGraph()
{
  if ( !GetVisited() )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;

      BOOL bGeometryEmpty;
      BOOL bGeometryDynamic;
      BOOL bNormalsEmpty;
      BOOL bNormalsDynamic;
      BOOL bBoundsEmpty;
      BOOL bBoundsDynamic;
      BOOL bBoundEmpty;
      BOOL bBoundDynamic;

      bGeometryEmpty = TRUE;
      bGeometryDynamic = FALSE;
      bNormalsEmpty = TRUE;
      bNormalsDynamic = FALSE;
      bBoundsEmpty = TRUE;
      bBoundsDynamic = FALSE;

      bBoundEmpty = TRUE;
      bBoundDynamic = FALSE;

      //////////////////////////////////////////////////
      // Visit Children Nodes and get starting values from them
      iter.Init(&m_listCSLIDEInstance);
      
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pCSLIDEInstance->PreprocessGraph();

	  bGeometryEmpty = ( bGeometryEmpty && 
			     pCSLIDEInstance->GetGeometryEmpty() );
	  bGeometryDynamic = ( bGeometryDynamic ||
			       pCSLIDEInstance->GetGeometryDynamic() );
	  bNormalsEmpty = ( bNormalsEmpty && 
			    pCSLIDEInstance->GetNormalsEmpty() );
	  bNormalsDynamic = ( bNormalsDynamic ||
			      pCSLIDEInstance->GetNormalsDynamic() );
	  bBoundsEmpty = ( bBoundsEmpty && 
			   pCSLIDEInstance->GetBoundsEmpty() );
	  bBoundsDynamic = ( bBoundsDynamic ||
			     pCSLIDEInstance->GetBoundsDynamic() );
	  bBoundEmpty = ( bBoundEmpty && 
			  pCSLIDEInstance->GetBoundEmpty() );
	  bBoundDynamic = ( bBoundDynamic ||
			    pCSLIDEInstance->GetBoundDynamic() );
	}

      //////////////////////////////////////////////////
      // Look at Local flags

      // Update whether we are Empty based on our LOD flag
      if ( !m_dyniLOD.GetDynamic() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      // This node is always off
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	      bBoundsEmpty = TRUE;
	      bBoundEmpty = TRUE;
	    }
	  else if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	    {
	      // This node is always just a bounding box
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	    }
	}

      // Check what is really dynamic

      bGeometryDynamic = ( bGeometryDynamic ||
			   m_pCSLIDESurface->GetDynamic() ||
			   m_dyniShading.GetDynamic() ||
			   m_dyniLOD.GetDynamic() );
      bNormalsDynamic = ( bNormalsDynamic ||
			  m_dyniShading.GetDynamic() ||
			  m_dyniLOD.GetDynamic() );
      bBoundsDynamic = ( bBoundsDynamic || 
			 m_dyniLOD.GetDynamic() );
      bBoundDynamic = ( bBoundDynamic || 
			m_dyniLOD.GetDynamic() );

      // If anything is now Empty then it will always be empty so
      // it is not dynamic

      bGeometryDynamic = ( bGeometryDynamic && !bGeometryEmpty );
      bNormalsDynamic = ( bNormalsDynamic && !bNormalsEmpty );
      bBoundsDynamic = ( bBoundsDynamic && !bBoundsEmpty );
      bBoundDynamic = ( bBoundDynamic && !bBoundEmpty );

      //////////////////////////////////////////////////
      // Now save these correct flag values

      SetGeometryEmpty(bGeometryEmpty);
      SetGeometryDynamic(bGeometryDynamic);
      SetNormalsEmpty(bNormalsEmpty);
      SetNormalsDynamic(bNormalsDynamic);
      SetBoundsEmpty(bBoundsEmpty);
      SetBoundsDynamic(bBoundsDynamic);
      SetBoundEmpty(bBoundEmpty);
      SetBoundDynamic(bBoundDynamic);
      
      //////////////////////////////////////////////////
      // Calculate the Bounding Box (if Empty this will do nothing)
      if ( GetBoundEmpty() )
	{
	  SetBoundValid(TRUE);
	}
      else
	{
	  CalculateBound();
	}
      SetBoundChanged(FALSE);

      //////////////////////////////////////////////////
      // Done Preprocessing this Node in the Graph
      SetVisited(TRUE);
    }
}

VOID CSLIDEGroup::Update()
{
  SetBoundChanged(FALSE);
  if ( GetDynamic() )
    {
      UpdateState();
    }

  if ( GetBoundsDynamic() )
    {
      CSLIDEBoundedNode::InvalidateQEMesh();

      if ( m_pmeSIF != NULL )
	{
	  DeallocateQEMesh(m_pmeSIF);
	  m_pmeSIF = NULL;
	}
    }

  if ( GetBoundDynamic() )
    {
      SetBoundValid(FALSE);
    }

  SetVisited(FALSE);
}

VOID CSLIDEGroup::UpdateState()
{
  CSLIDEBoundedNode::UpdateState();
}

VOID CSLIDEGroup::UpdateGraph()
{
  if ( !GetVisited() )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;
      BOOL bBoundChanged;

      if ( !GetBoundValid() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      if ( !m_bound.GetEmpty() )
		{
		  SetBoundChanged(TRUE);
		  m_bound.Null();
		}
	      return;
	    }
	  
	  bBoundChanged = FALSE;
	  
	  iter.Init(&m_listCSLIDEInstance);
	  
	  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
		pCSLIDEInstance != NULL;
		pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	    {
	      pCSLIDEInstance->UpdateGraph();
	      bBoundChanged = bBoundChanged || 
		pCSLIDEInstance->GetBoundChanged() ||
		(m_bound.GetEmpty() && !pCSLIDEInstance->GetBoundEmpty());
	    }
	  
	  if ( bBoundChanged || m_bGUIDeletedChild )
	    {
	      CalculateBound();
	    }
	  
	  SetBoundValid(TRUE);
	}

      SetVisited(TRUE);
    }
}

VOID CSLIDEGroup::CalculateBound()
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  CBound bound;
  
  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->InsertIntoBound(bound);
    }

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);
    }
}

#ifdef __USE_CDA__
VOID CSLIDEGroup::CalculateCDAVertices(CListInt *plistCDAVertex, 
				       const Matrix &mGroot_obj)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->CalculateCDAVertices(plistCDAVertex, mGroot_obj);
    }
}
#endif // __USE_CDA__

VOID CSLIDEGroup::CalculateQEMesh()
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  QEMesh *pme;
  
  if ( m_pQEMesh != NULL )
    {
      DeallocateQEMesh(m_pQEMesh);
      m_pQEMesh = NULL;
    }

  iter.Init(&m_listCSLIDEInstance);
  
  pme = NULL;
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pme = pCSLIDEInstance->GetQEMesh();
      if ( m_pQEMesh == NULL )
	{
	  m_pQEMesh = pme;
	}
      else
	{
	  if ( pme != NULL )
	    {
	      // This will automatically destroy pme
	      m_pQEMesh->MergeMesh(pme);
	    }
	}
    }

  if ( m_pQEMesh != NULL )
    {
      m_pQEMesh->Preprocess();
    }
}

VOID CSLIDEGroup::CalculateLeafQEMesh()
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  iter.Init(&m_listCSLIDEInstance);  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->CalculateLeafQEMesh();
    }
}

VOID CSLIDEGroup::DeallocateLeafQEMesh()
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  iter.Init(&m_listCSLIDEInstance);  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->DeallocateLeafQEMesh();
    }
}

VOID CSLIDEGroup::PrerenderGeometry(SLIDEShadingType eShading)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  SLIDEShadingType eShade;

  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  eShade = (SLIDEShadingType)m_dyniShading.GetValue();

  GLbitfield GLbAttributesMask;
  
  GLbAttributesMask = CalculateAttributesMask(eShade);
  
  PushAttributes(GLbAttributesMask);
  {
    PushShadingAttributes(eShade);
    {
      m_pCSLIDESurface->Push();
      {
	m_uiPrerenderAttributesStackDepthMax = 0;

	iter.Init(&m_listCSLIDEInstance);
	
	for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	      pCSLIDEInstance != NULL;
	      pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	  {
	    pCSLIDEInstance->PrerenderGeometry(eShading);
	    m_uiPrerenderAttributesStackDepthMax = 
	      MAX(m_uiPrerenderAttributesStackDepthMax,
		  pCSLIDEInstance->GetPrerenderAttributesStackDepthMax());
	  }

	if ( GLbAttributesMask != 0x0 )
	  {
	    m_uiPrerenderAttributesStackDepthMax++;
	  }
      }
      m_pCSLIDESurface->Pop();
    }
    PopShadingAttributes(eShade);
  }
  PopAttributes(GLbAttributesMask);
}

VOID CSLIDEGroup::PrerenderNormals(SLIDEShadingType eShading)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->PrerenderNormals(eShading);
    }
}

VOID CSLIDEGroup::PrerenderBounds()
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
    {
    }
  else 
    {
      if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	{
	}
      else
	{
	  iter.Init(&m_listCSLIDEInstance);
	  
	  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
		pCSLIDEInstance != NULL;
		pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	    {
	      pCSLIDEInstance->PrerenderBounds();
	    }
	}

      m_bound.Render();
    }
}

VOID CSLIDEGroup::Prerender()
{
  if ( !(!m_dyniLOD.GetDynamic() && 
	 ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
	   m_dyniLOD.GetValue() == slidelodtSLF_BOUND )) )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;
      
      iter.Init(&m_listCSLIDEInstance);
      
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pCSLIDEInstance->Prerender();
	}
    }
      
  CSLIDEBoundedNode::Prerender();
}

VOID CSLIDEGroup::RenderGeometry(const Matrix &mGproj_obj, 
				 CBound *pCBoundClip, 
				 SLIDEShadingType eShading,
				 SLIDELODType eLOD,
				 CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      if (NULL == pCSLIDEWarp)
	{
	  pCSLIDEInstance->Render(mGproj_obj, pCBoundClip, eShading, eLOD);
	}
      else
	{
	  pCSLIDEInstance->RenderGeometry(mGproj_obj, pCBoundClip, eShading, eLOD, pCSLIDEWarp);
	}
    }
}

VOID CSLIDEGroup::RenderGeometrySelect(const Matrix &mGproj_obj, 
				       CBound *pCBoundClip, 
				       SLIDEShadingType eShading,
				       SLIDELODType eLOD)
{
  UINT_32 i;
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  iter.Init(&m_listCSLIDEInstance);
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst(), i=0;
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext(), i++ )
    {
      gilPushName(i);
      {
	pCSLIDEInstance->RenderSelect(mGproj_obj, pCBoundClip, eShading, eLOD);
      }
      gilPopName();
    }
}

CSLIDEInstance *CSLIDEGroup::FindSelection(GLuint *puiID, int iLength)
{
  ASSERT( iLength > 0);

  UINT_32 uiInstance, i;
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  uiInstance = puiID[0];

  iter.Init(&m_listCSLIDEInstance);
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst(), i=0;
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext(), i++ )
    {
      if( uiInstance == i )
	{  
	  return pCSLIDEInstance->FindSelection(&(puiID[1]), iLength-1);
	}
    }

  ASSERT( FALSE );
  return pCSLIDEInstance;
}

VOID CSLIDEGroup::RenderGUISelected(GLuint *puiID, int iLength,
				    Matrix &mChildFrame, Matrix &mParentFrame,
				    BOOL bDrawBBox, BOOL bDrawAxes,
				    BOOL bChildAxesSelected, BOOL bXAxisSelected,
				    BOOL bYAxisSelected, BOOL bZAxisSelected)
{
  ASSERT( iLength > 0 );

  UINT_32 uiInstance, i;
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  uiInstance = puiID[0];

  iter.Init(&m_listCSLIDEInstance);
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst(), i=0;
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext(), i++ )
    {
      if( uiInstance == i )
	{
	  pCSLIDEInstance->RenderGUISelected(&(puiID[1]), iLength-1,
					     mChildFrame, mParentFrame, 
					     bDrawBBox, bDrawAxes,
					     bChildAxesSelected, bXAxisSelected,
					     bYAxisSelected, bZAxisSelected);
	  return;
	}
    }

  ASSERT( FALSE );
}

VOID CSLIDEGroup::RenderGraphSelectedBound()
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  
  iter.Init(&m_listCSLIDEInstance);
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->RenderGUISelectedBound();
    }
}

VOID CSLIDEGroup::RenderGUISelectedBound()
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  
  iter.Init(&m_listCSLIDEInstance);
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->RenderGUISelectedBound();
    }
}

VOID CSLIDEGroup::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtGroup], m_pcGUIID);

  WriteSLIDEFlags(fpOutputFile, 0);

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->WriteSLIDE(fpOutputFile);    
    }

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndGroup]);
}

VOID CSLIDEGroup::WriteRIBGeometry(FILE *fpOutputFile,
				   const Matrix &mGproj_obj, 
				   CBound *pCBoundClip, 
				   SLIDEShadingType eShading,
				   SLIDELODType eLOD)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( fpOutputFile != NULL );

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->WriteRIB(fpOutputFile,
				mGproj_obj, pCBoundClip, eShading, eLOD);    
    }
}

VOID CSLIDEGroup::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE GROUP: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);
}

VOID CSLIDEGroup::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE GROUP: %s", m_pcGUIID);
}

VOID CSLIDEGroup::WritePOVGeometry(FILE *fpOutputFile, 
				   CListInt *plistTransformIDs,
				   CListInt *plistColors) 
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( fpOutputFile != NULL );

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->WritePOV(fpOutputFile, plistTransformIDs, plistColors);
    }
}

VOID CSLIDEGroup::WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				    CSLIDESurface *pCSLIDESurface)
{
  ASSERT( fpOutputFile != NULL );

  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  CSLIDESurface *pCSLIDESurfacePush;

  if( m_pCSLIDESurface == CSLIDESurface::GetSLF_INHERIT() )
    {
      pCSLIDESurfacePush = pCSLIDESurface;
    }
  else
    {
      pCSLIDESurfacePush = m_pCSLIDESurface;
    }

  iter.Init(&m_listCSLIDEInstance);  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->WriteWRLZ(fpOutputFile, mTransform, pCSLIDESurfacePush);
    }
}

VOID CSLIDEGroup::WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( fpOutputFile != NULL );

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->WriteSTL(fpOutputFile, mTransform);
    }
}

QEMesh *CSLIDEGroup::WriteSIF(FILE *fpOutputFile, UINT_32 uiIndentLevel, 
			      const Matrix &mGroot_obj, const Matrix &mGobj_root,
			      QEMesh *pmeBoundaryEdges)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  QEMesh *pme;
  
  iter.Init(&m_listCSLIDEInstance);
  
  if ( m_pmeSIF != NULL )
    {
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pme = pCSLIDEInstance->WriteSIF(fpOutputFile, uiIndentLevel, 
					  mGroot_obj, mGobj_root, pmeBoundaryEdges);
	  if ( pme != NULL )
	    {
	      DeallocateQEMesh(pme);
	    }
	}
    }
  else
    {
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pme = pCSLIDEInstance->WriteSIF(fpOutputFile, uiIndentLevel, 
					  mGroot_obj, mGobj_root, pmeBoundaryEdges);
	  if ( m_pmeSIF == NULL )
	    {
	      m_pmeSIF = pme;
	    }
	  else
	    {
	      if ( pme != NULL )
		{
		  // This will automatically destroy pme
		  m_pmeSIF->MergeMesh(pme);
		}
	    }
	}
    }

  if ( m_pmeSIF != NULL )
    {
      return m_pmeSIF->Copy();
    }
  else
    {
      return NULL;
    }
}

VOID CSLIDEGroup::WriteLSIF(FILE *fpOutputFile, UINT_32 uiIndentLevel)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( fpOutputFile != NULL );

  if ( m_pcRIBBegin != NULL &&
       strncmp(m_pcRIBBegin, "  (layer", 8) == 0 )
    {
      if ( m_pcRIBBegin != NULL )
	{
	  fprintf(fpOutputFile, "%s", m_pcRIBBegin);
	  uiIndentLevel++;
	}

      iter.Init(&m_listCSLIDEInstance);
  
      fprintf(fpOutputFile, "    (contours\n");
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pCSLIDEInstance->WriteLSIF(fpOutputFile, uiIndentLevel);
	}
      fprintf(fpOutputFile, "    )\n");

      fprintf(fpOutputFile, "    (offsets\n");
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pCSLIDEInstance->WriteLSIFOffset(fpOutputFile, uiIndentLevel);
	}
      fprintf(fpOutputFile, "    )\n");

      if ( m_pcRIBEnd != NULL )
	{
	  fprintf(fpOutputFile, "%s", m_pcRIBEnd);
	}
    }
  else
    {
      if ( m_pcRIBBegin != NULL )
	{
	  fprintf(fpOutputFile, "%s", m_pcRIBBegin);
	  uiIndentLevel++;
	}

      iter.Init(&m_listCSLIDEInstance);
  
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pCSLIDEInstance->WriteLSIF(fpOutputFile, uiIndentLevel);
	}

      if ( m_pcRIBEnd != NULL )
	{
	  fprintf(fpOutputFile, "%s", m_pcRIBEnd);
	}
    }
}

VOID CSLIDEGroup::WriteDotID(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "group_%s", GetGUIID());
}

VOID CSLIDEGroup::WriteDotNode(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "  ");
  WriteDotID(fpOutputFile);
  fprintf(fpOutputFile, " [ label = \"%s\", shape = ellipse", GetGUIID());
  if ( GetBoundDynamic() )
    {
      fprintf(fpOutputFile, ", color = purple4, fontcolor = purple4");
    }
  else if ( GetGeometryDynamic() )
    {
      fprintf(fpOutputFile, ", color = blue, fontcolor = blue");
    }
  fprintf(fpOutputFile, " ];\n");

  iter.Init(&m_listCSLIDEInstance);

  for ( pCSLIDEInstance = (CSLIDEInstance*)(iter.PeekFirst());
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)(iter.PeekNext()) )
    {
      pCSLIDEInstance->WriteDotNode(fpOutputFile);
    }  
}

VOID CSLIDEGroup::WriteDotEdges(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( fpOutputFile != NULL );

  iter.Init(&m_listCSLIDEInstance);

  for ( pCSLIDEInstance = (CSLIDEInstance*)(iter.PeekFirst());
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)(iter.PeekNext()) )
    {
      fprintf(fpOutputFile, "  ");
      WriteDotID(fpOutputFile);
      fprintf(fpOutputFile, " -> ");
      pCSLIDEInstance->WriteDotID(fpOutputFile);
      fprintf(fpOutputFile, "\n");
      pCSLIDEInstance->WriteDotEdges(fpOutputFile);
    }  
}

UINT_32 CSLIDEGroup::GetInstanceCount()
{
  return m_listCSLIDEInstance.GetLength();
}

VOID CSLIDEGroup::GetParameters(Tcl_Interp *pInterp)
{
  INT_32 iCode;
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  CHAR pcBuffer[5000];
  CHAR pcInstances[1000];
  CHAR pcSurface[35];

  if( !strcmp(m_pCSLIDESurface->GetGUIID(), "SLF_INHERIT" ) )
      sprintf(pcSurface, "inherit");
  else 
      sprintf(pcSurface, m_pCSLIDESurface->GetGUIID());

  pcInstances[0] = '\0';
  iter.Init(&m_listCSLIDEInstance);
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      strcat(pcInstances, " ");
      strcat(pcInstances, pCSLIDEInstance->GetGUIID());
    }

  sprintf(pcBuffer, "Editor:SetParameterValues Group {{surface %s} {shading %s} {lod %s} {ribbegin %s} {ribend %s} {instances {%s}}}",
	  pcSurface,
	  gcppcSLIDEGUIShadingReservedWords[m_dyniShading.GetValue()],
	  gcppcSLIDEGUILODReservedWords[m_dyniLOD.GetValue()],
	  m_pcRIBBegin,
	  m_pcRIBEnd,
	  pcInstances);
	  
  iCode = Tcl_Eval(pInterp, pcBuffer);
  //ASSERT( iCode == TCL_OK );
  if ( iCode != TCL_OK )
    {
      printf( "%s\n", pcBuffer );
      printf("Error %s \n", pInterp->result);
    }
}
