#ifndef __SLIDEGROUP_H__
#define __SLIDEGROUP_H__

#include "slideboundednode.h"

class CSLIDEWorld;
class CSLIDESurface;
class CSLIDEInstance;

//////////////////////////////////////////////////////////////////////
// CSLIDEGroup Class
//

class CSLIDEGroup : public CSLIDEBoundedNode
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEGroup();
  virtual ~CSLIDEGroup();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  virtual VOID InsertInstance(CSLIDEInstance *pCSLIDEInstance);

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

  virtual VOID Remove(CSLIDEWorld *pCSLIDEWorld);
  virtual VOID RemoveChildInstance(CSLIDEInstance *pCSLIDEInstance);

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual CSLIDECamera *FindCamera(CListIterInt *piterIDs, 
				   CListInt *plistCSLIDEInstance);
  virtual CSLIDELight *FindLight(CListIterInt *piterIDs, 
				 CListInt *plistCSLIDEInstance);
  virtual CSLIDEAreaLight *FindAreaLight(CListIterInt *piterIDs, 
					 CListInt *plistCSLIDEInstance);
  virtual VOID Preprocess();
  virtual VOID PreprocessGraph();

  virtual VOID Update();
  virtual VOID UpdateGraph();

#ifdef __USE_CDA__
  virtual VOID CalculateCDAVertices(CListInt *plistCDAVertex, 
				    const Matrix &mGroot_obj);
#endif // __USE_CDA__

  virtual VOID CalculateQEMesh();
  virtual VOID CalculateLeafQEMesh();  // for WriteWRLZ
  virtual VOID DeallocateLeafQEMesh();  // for WriteWRLZ

protected:
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();
  virtual VOID CalculateBound();

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID PrerenderGeometry(SLIDEShadingType eShading);
  virtual VOID PrerenderNormals(SLIDEShadingType eShading);
  virtual VOID PrerenderBounds();

  virtual VOID Prerender();

  virtual VOID RenderGeometry(const Matrix &mGproj_obj,
			      CBound *pCBoundClip, 
			      SLIDEShadingType eShading,
			      SLIDELODType eLOD,
			      CSLIDEWormwarp *pCSLIDEWarp = NULL);

  virtual VOID RenderGeometrySelect(const Matrix &mGproj_obj,
				    CBound *pCBoundClip, 
				    SLIDEShadingType eShading,
				    SLIDELODType eLOD);
  virtual CSLIDEInstance *FindSelection(GLuint *puiID, int iLength);  
  virtual VOID RenderGUISelected(GLuint *puiID, int iLength,
				 Matrix &mChildFrame, Matrix &mParentFrame, 
				 BOOL bDrawBBox, BOOL bDrawAxes,
				 BOOL bChildAxesSelected, BOOL bXAxisSelected,
				 BOOL bYAxisSelected, BOOL bZAxisSelected);  
  virtual VOID RenderGraphSelectedBound();
  virtual VOID RenderGUISelectedBound();

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

  virtual VOID WritePOVGeometry(FILE *fpOutputFile, 
				CListInt *plistTransformIDs,
				CListInt *plistColors); 
  virtual VOID WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				 CSLIDESurface *pCSLIDESurface);

  virtual VOID WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform);

  virtual QEMesh *WriteSIF(FILE *fpOutputFile, UINT_32 uiIndentLevel, 
			   const Matrix &mGroot_obj, const Matrix &mGobj_root,
			   QEMesh *pmeBoundaryEdges);
  virtual VOID WriteLSIF(FILE *fpOutputFile, UINT_32 uiIndentLevel); 

  virtual VOID WriteDotID(FILE *fpOutputFile);
  virtual VOID WriteDotNode(FILE *fpOutputFile);
  virtual VOID WriteDotEdges(FILE *fpOutputFile);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline CHAR *GetTypeString();
  virtual inline SLIDEReservedWordType GetType();

  UINT_32 GetInstanceCount();

  virtual VOID GetSide(CListInt *plistPoint, CListInt *plistNormal, UINT_32 uiSide, 
		       const Matrix &mTransform, const Matrix &mInverse, BOOL bReverse);

  virtual VOID GetParameters(Tcl_Interp *pInterp);

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CListInt m_listCSLIDEInstance;

  QEMesh *m_pmeSIF;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline CHAR *CSLIDEGroup::GetTypeString()
{
  return gcppcSLIDEReservedWords[sliderwtGroup];
}

inline SLIDEReservedWordType CSLIDEGroup::GetType()
{
  return sliderwtGroup;
}

#endif // __SLIDEGROUP_H__
