#ifndef __SLIDEHEIGHTFIELD_H__
#define __SLIDEHEIGHTFIELD_H__

#include "slideatomnode.h"

//////////////////////////////////////////////////////////////////////////////
// CSLIDEHeightfield Class
//

class CSLIDEHeightfield : public CSLIDEAtomNode
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEHeightfield();
  virtual ~CSLIDEHeightfield();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);
  virtual BOOL Valid();

  virtual inline VOID SetUMin(DynFloat *pdynfUMin);
  virtual inline VOID SetVMin(DynFloat *pdynfVMin);
  virtual inline VOID SetUMax(DynFloat *pdynfUMax);
  virtual inline VOID SetVMax(DynFloat *pdynfVMax);
  virtual inline VOID SetUSlices(DynInt *pdyniUSlices);
  virtual inline VOID SetVSlices(DynInt *pdyniVSlices);
  virtual inline VOID SetFunc(DynVector *pdynvFunc);

protected:

  virtual VOID ReallocateSurfacePoints();
  virtual VOID DeallocateSurfacePoints();

//////////////////////////////////////////////////
// Geometric Computations
//
public:

#ifdef __USE_CDA__
  virtual VOID CalculateCDAVertices(CListInt *plistCDAVertex, const Matrix &mGroot_obj);
#endif // __USE_CDA__

protected:
  virtual VOID CheckDynamic();
  virtual VOID PreprocessGraphLocal();
  virtual VOID UpdateState();
  virtual VOID CalculateBound();
  virtual VOID CalculateQEMesh();

public:

protected:
  virtual VOID CalculateSurface();

//////////////////////////////////////////////////
// Rendering
//
public:

protected:
  virtual VOID RenderGeometrySmooth();
  virtual VOID RenderGeometryFlat();
  virtual VOID RenderNormalsSmooth();
  virtual VOID RenderNormalsFlat();

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

  virtual VOID WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				 CSLIDESurface *pCSLIDESurface);
  virtual VOID WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline CHAR *GetTypeString();
  virtual inline SLIDEReservedWordType GetType();

//////////////////////////////////////////////////
// Member Variables
//

protected:
  DynFloat m_dynfUMin;
  FLOAT m_fUMin;
  DynFloat m_dynfVMin;
  FLOAT m_fVMin;
  DynFloat m_dynfUMax;
  FLOAT m_fUMax;
  DynFloat m_dynfVMax;
  FLOAT m_fVMax;

  DynInt m_dyniUSlices;
  UINT_32 m_uiUSlices;
  DynInt m_dyniVSlices;
  UINT_32 m_uiVSlices;
  UINT_32 m_uiUSlicesPrev;

  DynVector m_dynvFunc;
  Point **m_ppSurfacePoint;

};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDEHeightfield::SetUMin(DynFloat *pdynfUMin)
{
  m_dynfUMin.Set(pdynfUMin);
}

inline VOID CSLIDEHeightfield::SetVMin(DynFloat *pdynfVMin)
{
  m_dynfVMin.Set(pdynfVMin);
}

inline VOID CSLIDEHeightfield::SetUMax(DynFloat *pdynfUMax)
{
  m_dynfUMax.Set(pdynfUMax);
}

inline VOID CSLIDEHeightfield::SetVMax(DynFloat *pdynfVMax)
{
  m_dynfVMax.Set(pdynfVMax);
}

inline VOID CSLIDEHeightfield::SetUSlices(DynInt *pdyniUSlices)
{
  m_dyniUSlices.Set(pdyniUSlices);
}

inline VOID CSLIDEHeightfield::SetVSlices(DynInt *pdyniVSlices)
{
  m_dyniVSlices.Set(pdyniVSlices);
}

inline VOID CSLIDEHeightfield::SetFunc(DynVector *pdynvFunc)
{
  m_dynvFunc.SetAll(pdynvFunc);
}

inline CHAR *CSLIDEHeightfield::GetTypeString()
{
  return gcppcSLIDEReservedWords[sliderwtHeightfield];
}

inline SLIDEReservedWordType CSLIDEHeightfield::GetType()
{
  return sliderwtHeightfield;
}

#endif // __SLIDEHEIGHTFIELD_H__
