#include "slideinstance.h"

#include "slideworld.h"
#include "slidetransform.h"
#include "slidetranslate.h"
#include "sliderotate.h"
#include "slidescale.h"
#include "slidesurface.h"
#include "slidecontour.h"
#include "slideseam.h"

#include "slidewarp.h"
#include "qemesh.h"

//////////////////////////////////////////////////////////////////////
// CSLIDEInstance Class
//

CSLIDEInstance::CSLIDEInstance()
{
  m_pcChildID = NULL;
  m_pcParentID = NULL;
  m_pCSLIDENode = NULL;

  m_pCSLIDETransformLastAdded = NULL;

  m_uiDotID = 0;
}

CSLIDEInstance::~CSLIDEInstance()
{
}

VOID CSLIDEInstance::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			  CHAR *pcChildID, CHAR *pcParentID)
{
  CSLIDEGeometryNode::Init(pcFile, uiLine, uiColumn, NULL);

  ASSERT( pcChildID != NULL );

  m_pcChildID = pcChildID;
  m_pcParentID = pcParentID;
  m_pCSLIDENode = NULL;
  
  m_listCSLIDETransform.Init();
  m_mTransform.SetIdentity();
  m_mInverse.SetIdentity();

  m_uiDotID = 0;
}

VOID CSLIDEInstance::Uninit()
{
  CSLIDEGeometryNode::Uninit();
}

BOOL CSLIDEInstance::Initialized()
{
  if ( !CSLIDEGeometryNode::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

VOID CSLIDEInstance::InsertTransform(CSLIDETransform *pCSLIDETransform)
{
  ASSERT( pCSLIDETransform != NULL );

  m_listCSLIDETransform.InsertLast(pCSLIDETransform, 0);
}

VOID CSLIDEInstance::InsertTurtleWalkTransform(CSLIDETransform *pCSLIDETransformNew,
					       BOOL bNewTransform)
{
  ASSERT( pCSLIDETransformNew != NULL );

  CListIterInt iter;
  CSLIDETransform *pCSLIDETransformFirst, *pCSLIDETransform;

  m_bGUIChangedTransforms = TRUE;

  iter.Init(&m_listCSLIDETransform);
  pCSLIDETransformFirst = (CSLIDETransform*)iter.PeekFirst(); 

  // Check if this transform can be composed with the first transform
  if( (pCSLIDETransformFirst != NULL) && 
      (pCSLIDETransformNew->GetType() == pCSLIDETransformFirst->GetType()) &&
      (!strcmp(pCSLIDETransformFirst->GetFileName(), "[Manipulation]")) &&
      !bNewTransform )
    {
      pCSLIDETransformFirst->Set(pCSLIDETransformNew);
      /*
      if( bNewTransform ) 
	{
	  pCSLIDETransformFirst->ComposeWith(pCSLIDETransformNew);
	}
      else
	{
	  pCSLIDETransformFirst->Modify(m_pCSLIDETransformLastAdded, pCSLIDETransformNew);
	}

      if( m_pCSLIDETransformLastAdded != NULL )
	{
	  delete m_pCSLIDETransformLastAdded;
	  m_pCSLIDETransformLastAdded = NULL;
	}
      m_pCSLIDETransformLastAdded = pCSLIDETransformNew;
      */
    }
  else
    {
      m_listCSLIDETransform.InsertFirst(pCSLIDETransformNew, 0);
    }

  // Update Transforms
  m_mTransform.SetIdentity(); 
  m_mInverse.SetIdentity(); 

  iter.Init(&m_listCSLIDETransform);
  for ( pCSLIDETransform = (CSLIDETransform*)iter.PeekFirst(); 
	pCSLIDETransform != NULL;
	pCSLIDETransform = (CSLIDETransform*)iter.PeekNext() )
    {
      m_mTransform = pCSLIDETransform->GetMatrix() * m_mTransform;      
      m_mInverse = m_mInverse * pCSLIDETransform->GetInverse();
    }
}

VOID CSLIDEInstance::InsertCommandPostTransform(CSLIDETransform *pCSLIDETransformNew,
						BOOL bNewTransform)
{
  ASSERT( pCSLIDETransformNew != NULL );

  CListIterInt iter;
  CSLIDETransform *pCSLIDETransformLast, *pCSLIDETransform;

  m_bGUIChangedTransforms = TRUE;

  iter.Init(&m_listCSLIDETransform);
  pCSLIDETransformLast = (CSLIDETransform*)iter.PeekLast(); 

  // Check if this transform can be composed with the last transform
  if( (pCSLIDETransformLast != NULL) && 
      (pCSLIDETransformNew->GetType() == pCSLIDETransformLast->GetType()) &&
      (!strcmp(pCSLIDETransformLast->GetFileName(), "[Manipulation]")) &&
      !bNewTransform )
    {
      pCSLIDETransformLast->Set(pCSLIDETransformNew);

      /*
      if( bNewTransform ) 
	{
	  pCSLIDETransformLast->ComposeWith(pCSLIDETransformNew);
	}
      else
	{
	  pCSLIDETransformLast->Modify(m_pCSLIDETransformLastAdded, pCSLIDETransformNew);
	}

      if( m_pCSLIDETransformLastAdded != NULL )
	{
	  delete m_pCSLIDETransformLastAdded;
	  m_pCSLIDETransformLastAdded = NULL;
	}
      m_pCSLIDETransformLastAdded = pCSLIDETransformNew;
      */
    }
  else
    {
      m_listCSLIDETransform.InsertLast(pCSLIDETransformNew, 0);
    }

  // Update Transforms
  m_mTransform.SetIdentity(); 
  m_mInverse.SetIdentity(); 

  iter.Init(&m_listCSLIDETransform);
  for ( pCSLIDETransform = (CSLIDETransform*)iter.PeekFirst(); 
	pCSLIDETransform != NULL;
	pCSLIDETransform = (CSLIDETransform*)iter.PeekNext() )
    {
      m_mTransform = pCSLIDETransform->GetMatrix() * m_mTransform;      
      m_mInverse = m_mInverse * pCSLIDETransform->GetInverse();
    }
}

BOOL CSLIDEInstance::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  if ( !CSLIDEGeometryNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  m_pCSLIDENode = pCSLIDEWorld->PeekNode(m_pcChildID);
  if ( m_pCSLIDENode == NULL )
    {
      HandleError("Undefined %s \"%s\"", 
		  gcppcSLIDEReservedWords[sliderwtObject], m_pcChildID);
      return FALSE;
    }

  m_pCSLIDENodeParent = pCSLIDEWorld->PeekNode(m_pcParentID);
  ASSERT( m_pCSLIDENode != NULL );

  return TRUE;
}

BOOL CSLIDEInstance::Valid()
{
  if ( !CSLIDEGeometryNode::Valid() )
    {
      return FALSE;
    }

  return TRUE;
}

VOID CSLIDEInstance::Remove(CSLIDEWorld *pCSLIDEWorld)
{
  //printf("Instance::Remove %s\n", this->GetID());

  m_pCSLIDENodeParent->RemoveChildInstance(this);
  m_pCSLIDENode->RemoveParentInstance(this);
  pCSLIDEWorld->RemoveInstance(this);
  Uninit();
}

VOID CSLIDEInstance::RemoveFromParent(CSLIDEWorld *pCSLIDEWorld)
{
  //printf("Instance::Remove %s\n", this->GetID());

  m_pCSLIDENodeParent->RemoveChildInstance(this);
}

VOID CSLIDEInstance::RemoveFromChild(CSLIDEWorld *pCSLIDEWorld)
{
  //printf("Instance::Remove %s\n", this->GetID());

  m_pCSLIDENode->RemoveParentInstance(this);
}

CSLIDECamera *CSLIDEInstance::FindCamera(CListIterInt *piterIDs, 
					 CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  return m_pCSLIDENode->FindCamera(piterIDs, plistCSLIDEInstance);
}

CSLIDELight *CSLIDEInstance::FindLight(CListIterInt *piterIDs, 
				       CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  return m_pCSLIDENode->FindLight(piterIDs, plistCSLIDEInstance);
}

CSLIDEAreaLight *CSLIDEInstance::FindAreaLight(CListIterInt *piterIDs, 
					       CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  return m_pCSLIDENode->FindAreaLight(piterIDs, plistCSLIDEInstance);
}

VOID CSLIDEInstance::GetSide(CListInt *plistPoint, CListInt *plistNormal, UINT_32 uiSide, 
			     const Matrix &mTransform, const Matrix &mInverse, 
			     BOOL bReverse)
{
  ASSERT( plistPoint != NULL );

  Matrix mT, mI;
  mT = mTransform * m_mTransform;
  mI = m_mInverse * mInverse;

  m_pCSLIDENode->GetSide(plistPoint, plistNormal, uiSide, mT, mI, bReverse);
}

VOID CSLIDEInstance::Preprocess()
{
  CSLIDEGeometryNode::Preprocess();

  CheckDynamic();
  UpdateState();
}

VOID CSLIDEInstance::SetUnvisited()
{
  SetVisited(FALSE);

  m_pCSLIDENodeParent->SetUnvisited();
}

VOID CSLIDEInstance::SetBoundUnvisited()
{
  SetVisited(FALSE);
  SetBoundValid(FALSE);
  
  m_pCSLIDENodeParent->SetBoundUnvisited();
}

VOID CSLIDEInstance::SetBoundUnchanged()
{
  SetBoundChanged(FALSE);
  m_bGUIChangedTransforms = FALSE;
  
  m_pCSLIDENodeParent->SetBoundUnchanged();
}

VOID CSLIDEInstance::CheckDynamic()
{
  CListIterInt iter;
  CSLIDETransform *pCSLIDETransform;
  BOOL bTransformDynamic;

  // Check if the NodeID is dynamic
  SetDynamic(m_dyniShading.GetDynamic() || 
	     m_dyniLOD.GetDynamic());

  bTransformDynamic = FALSE;
  iter.Init(&m_listCSLIDETransform);

  for ( pCSLIDETransform = (CSLIDETransform*)iter.PeekFirst(); 
	pCSLIDETransform != NULL;
	pCSLIDETransform = (CSLIDETransform*)iter.PeekNext() )
    {
      if ( pCSLIDETransform->GetDynamic() )
	{ 
	  bTransformDynamic = TRUE;
	  break;
	}
    }

  SetTransformDynamic(bTransformDynamic);
}

VOID CSLIDEInstance::PreprocessGraph()
{
  if ( !GetVisited() )
    {
      BOOL bGeometryEmpty;
      BOOL bGeometryDynamic;
      BOOL bNormalsEmpty;
      BOOL bNormalsDynamic;
      BOOL bBoundsEmpty;
      BOOL bBoundsDynamic;
      BOOL bBoundEmpty;
      BOOL bBoundDynamic;

      m_pCSLIDENode->InsertParentInstance(this);
      m_pCSLIDENode->PreprocessGraph();

      //////////////////////////////////////////////////
      // Getting starting values from the child node

      bGeometryEmpty = m_pCSLIDENode->GetGeometryEmpty();
      bGeometryDynamic = m_pCSLIDENode->GetGeometryDynamic();
      bNormalsEmpty = m_pCSLIDENode->GetNormalsEmpty();
      bNormalsDynamic = m_pCSLIDENode->GetNormalsDynamic();
      bBoundsEmpty = m_pCSLIDENode->GetBoundsEmpty();
      bBoundsDynamic = m_pCSLIDENode->GetBoundsDynamic();
      bBoundEmpty = m_pCSLIDENode->GetBoundEmpty();
      bBoundDynamic = m_pCSLIDENode->GetBoundDynamic();

      //////////////////////////////////////////////////
      // Look at Local flags

      // Update whether we are Empty based on our LOD flag
      if ( !m_dyniLOD.GetDynamic() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      // This node is always off
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	      bBoundsEmpty = TRUE;
	      bBoundEmpty = TRUE;
	    }
	  else if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	    {
	      // This node is always just a bounding box
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	    }
	}

      // Check what is really dynamic

      bGeometryDynamic = ( bGeometryDynamic ||
			   m_pCSLIDESurface->GetDynamic() ||
			   m_dyniShading.GetDynamic() ||
			   m_dyniLOD.GetDynamic() ||
			   GetTransformDynamic() );
      bNormalsDynamic = ( bNormalsDynamic ||
			  m_dyniShading.GetDynamic() ||
			  m_dyniLOD.GetDynamic() ||
			  GetTransformDynamic() );
      bBoundsDynamic = ( bBoundsDynamic || 
			 m_dyniLOD.GetDynamic() ||
			 GetTransformDynamic() );
      bBoundDynamic = ( bBoundDynamic || 
			m_dyniLOD.GetDynamic() ||
			GetTransformDynamic() );

      // If anything is now Empty then it will always be empty so
      // it is not dynamic

      bGeometryDynamic = ( bGeometryDynamic && !bGeometryEmpty );
      bNormalsDynamic = ( bNormalsDynamic && !bNormalsEmpty );
      bBoundsDynamic = ( bBoundsDynamic && !bBoundsEmpty );
      bBoundDynamic = ( bBoundDynamic && !bBoundEmpty );

      //////////////////////////////////////////////////
      // Now save these correct flag values

      SetGeometryEmpty(bGeometryEmpty);
      SetGeometryDynamic(bGeometryDynamic);
      SetNormalsEmpty(bNormalsEmpty);
      SetNormalsDynamic(bNormalsDynamic);
      SetBoundsEmpty(bBoundsEmpty);
      SetBoundsDynamic(bBoundsDynamic);
      SetBoundEmpty(bBoundEmpty);
      SetBoundDynamic(bBoundDynamic);
      
      //////////////////////////////////////////////////
      // Calculate the Bounding Box (if Empty this will do nothing)
      if ( GetBoundEmpty() )
	{
	  SetBoundValid(TRUE);
	}
      else
	{
	  CalculateBound();
	}
      SetBoundChanged(FALSE);

      //////////////////////////////////////////////////
      // Done Preprocessing this Node in the Graph
      SetVisited(TRUE);
    }
}

VOID CSLIDEInstance::Update()
{
  SetBoundChanged(FALSE);
  if ( GetDynamic() || GetTransformDynamic() )
    {
      UpdateState();
    }

  if ( GetBoundDynamic() )
    {
      SetBoundValid(FALSE);
    }

  SetVisited(FALSE);
}

VOID CSLIDEInstance::UpdateState()
{
  CSLIDEGeometryNode::UpdateState();

  // Smash updated Transform list into a single Matrix

  CListIterInt iter;
  CSLIDETransform *pCSLIDETransform;

  m_mTransform.SetIdentity(); 
  m_mInverse.SetIdentity(); 

  iter.Init(&m_listCSLIDETransform);

  for ( pCSLIDETransform = (CSLIDETransform*)iter.PeekFirst(); 
	pCSLIDETransform != NULL;
	pCSLIDETransform = (CSLIDETransform*)iter.PeekNext() )
    {
      m_mTransform = pCSLIDETransform->GetMatrix() * m_mTransform;      
      m_mInverse = m_mInverse * pCSLIDETransform->GetInverse();
    }
  /*
printf("***** Start Instance Update *****\n");
printf("Transform:\n");
m_mTransform.Display();
printf("Inverse:\n");
m_mInverse.Display();
printf("Identity:\n");
Matrix mJordan;
mJordan = m_mTransform * m_mInverse;
mJordan.Display();
printf("***** End Instance Update *****\n");
*/
}

VOID CSLIDEInstance::UpdateGraph()
{
  if ( !GetVisited() )
    {
      if ( !GetBoundValid() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      if ( !GetBoundEmpty() )
		{
		  SetBoundChanged(TRUE);
		  SetBoundEmpty(TRUE);
		}
	      return;
	    }
	  
	  m_pCSLIDENode->UpdateGraph();
	  if ( m_pCSLIDENode->GetBoundChanged() || 
	       m_bGUIChangedTransforms || 
	       (GetTransformDynamic() && !m_pCSLIDENode->GetBoundEmpty()) ||
	       (GetBoundEmpty() && !m_pCSLIDENode->GetBoundEmpty()) )
	    {
	      CalculateBound();
	    }

	  SetBoundValid(TRUE);
	}

      SetVisited(TRUE);
    }
}

VOID CSLIDEInstance::InsertIntoBound(CBound &bound)
{
  if ( !GetBoundEmpty() )
    {
      m_pCSLIDENode->InsertIntoBound(m_mTransform, bound);
    }
}

VOID CSLIDEInstance::CalculateBound()
{
  // If this was called it means that the group above it
  // in the scene graph must recompute its bounding box
  SetBoundEmpty(m_pCSLIDENode->GetBoundEmpty());
  SetBoundChanged(TRUE);
}

#ifdef __USE_CDA__
VOID CSLIDEInstance::CalculateCDA()
{
  m_pCSLIDENode->CalculateCDA();
}

VOID CSLIDEInstance::CalculateCDAVertices(CListInt *plistCDAVertex, 
					  const Matrix &mGroot_obj)
{
  Matrix m;

  m = mGroot_obj * m_mTransform;

  m_pCSLIDENode->CalculateCDAVertices(plistCDAVertex, m);  
}
#endif // __USE_CDA__

VOID CSLIDEInstance::PrerenderGeometry(SLIDEShadingType eShading)
{
  SLIDEShadingType eShade;

  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  ModifyShadingFlag(eShading);

  eShade = (SLIDEShadingType)m_dyniShading.GetValue();

  GLbitfield GLbAttributesMask;
  
  GLbAttributesMask = CalculateAttributesMask(eShade);
  
  PushAttributes(GLbAttributesMask);
  {
    PushShadingAttributes(eShade);
    {
      m_pCSLIDESurface->Push();
      {
	gilPushMatrix();
	{
	  gilMultMatrixf(m_mTransform.Get());
	  
	  m_pCSLIDENode->RecallGeometry(eShading);
	  
	  m_uiPrerenderAttributesStackDepthMax = 
	    m_pCSLIDENode->GetPrerenderAttributesStackDepthMax();
	  
	  if ( GLbAttributesMask != 0x0 )
	    {
	      m_uiPrerenderAttributesStackDepthMax++;
	    }
	}
	gilPopMatrix();
      }
      m_pCSLIDESurface->Pop();
    }
    PopShadingAttributes(eShade);
  }
  PopAttributes(GLbAttributesMask);
}

VOID CSLIDEInstance::PrerenderNormals(SLIDEShadingType eShading)
{
  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  ModifyShadingFlag(eShading);

  gilPushMatrix();
  {
    gilMultMatrixf(m_mTransform.Get());
    
    m_pCSLIDENode->RecallNormals(eShading);
  }
  gilPopMatrix();
}

VOID CSLIDEInstance::PrerenderBounds()
{
  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
    {
    }
  else if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      RenderBound();
    }
  else
    {
      gilPushMatrix();
      {
	gilMultMatrixf(m_mTransform.Get());
	
	m_pCSLIDENode->RecallBounds();
      }
      gilPopMatrix();
    }
}

VOID CSLIDEInstance::Prerender()
{
  if ( !(!m_dyniLOD.GetDynamic() && 
	 ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
	   m_dyniLOD.GetValue() == slidelodtSLF_BOUND )) )
    {
      //printf("Instance::Prerender %s\n", this->GetID());

      m_pCSLIDENode->Prerender();
    }
}

VOID CSLIDEInstance::SetPrerenderUnvisited()
{
  if( GetDynamic() )
    {
      return;
    }

  SetPrerenderVisited(FALSE);
  
  m_pCSLIDENodeParent->SetPrerenderUnvisited();
}

VOID CSLIDEInstance::RecallGeometry(SLIDEShadingType /*eShading*/)
{
  ASSERT( FALSE );
}

VOID CSLIDEInstance::RecallNormals(SLIDEShadingType /*eShading*/)
{
  ASSERT( FALSE );
}

VOID CSLIDEInstance::RecallBounds()
{
  ASSERT( FALSE );
}

VOID CSLIDEInstance::Render(const Matrix &mGproj_obj, 
			    CBound *pCBoundClip, 
			    SLIDEShadingType eShading,
			    SLIDELODType eLOD)
{
  ModifyShadingFlag(eShading);
  ModifyLODFlag(eLOD);

  if ( eLOD == slidelodtSLF_OFF )
    {
      // This branch has been turned off by the LOD flags
      return;
    }

  if ( eLOD > slidelodtSLF_BOUND )
    {
      SLIDEShadingType eShade;
      
      eShade = (SLIDEShadingType)m_dyniShading.GetValue();

      GLbitfield GLbAttributesMask;
      
      GLbAttributesMask = CalculateAttributesMask(eShade);
      
      PushAttributes(GLbAttributesMask);
      {
	PushShadingAttributes(eShade);
	{
	  m_pCSLIDESurface->Push();
	  {
	    RenderGeometry(mGproj_obj, pCBoundClip, eShading, eLOD);
	  }
	  m_pCSLIDESurface->Pop();
	}
	PopShadingAttributes(eShade);
      }
      PopAttributes(GLbAttributesMask);

      // Render Normals if necessary
      if ( GetRenderNormalsEnabled() )
	{
	  RenderNormals(eShading);
	}
    }

  // Render Bounding Box if necessary
  if ( GetRenderBoundsEnabled() || eLOD == slidelodtSLF_BOUND )
    {
      RenderBound();
    }
}

VOID CSLIDEInstance::RenderSelect(const Matrix &mGproj_obj, 
				  CBound *pCBoundClip, 
				  SLIDEShadingType eShading,
				  SLIDELODType eLOD)
{
  ModifyShadingFlag(eShading);
  ModifyLODFlag(eLOD);

  if ( eLOD == slidelodtSLF_OFF )
    {
      // This branch has been turned off by the LOD flags
      return;
    }

  if ( eLOD > slidelodtSLF_BOUND )
    {
      RenderGeometrySelect(mGproj_obj, pCBoundClip, eShading, eLOD);
    }
}

VOID CSLIDEInstance::RenderGeometry(const Matrix &mGproj_obj, 
				    CBound *pCBoundClip, 
				    SLIDEShadingType eShading, 
				    SLIDELODType eLOD,
				    CSLIDEWormwarp *pCSLIDEWarp /*= NULL */)
{
  Matrix m;

  if (NULL==pCSLIDEWarp)
    {
      gilPushMatrix();
      {
	gilMultMatrixf(m_mTransform.Get());
	
	m = mGproj_obj * m_mTransform;
	
	m_pCSLIDENode->Render(m, pCBoundClip, eShading, eLOD);
      }
      gilPopMatrix();
    }
  else
    {
      pCSLIDEWarp->PushMatrix();
      pCSLIDEWarp->MultMatrix(m_mTransform);
      m = mGproj_obj * m_mTransform;
      m_pCSLIDENode->RenderGeometry(m, pCBoundClip, eShading, eLOD, pCSLIDEWarp);
      pCSLIDEWarp->PopMatrix();
    }
}

VOID CSLIDEInstance::RenderGeometrySelect(const Matrix &mGproj_obj, 
					  CBound *pCBoundClip, 
					  SLIDEShadingType eShading, 
					  SLIDELODType eLOD)
{
  Matrix m;

  gilPushMatrix();
  {
    gilMultMatrixf(m_mTransform.Get());
    
    m = mGproj_obj * m_mTransform;

    m_pCSLIDENode->RenderSelect(m, pCBoundClip, eShading, eLOD);
  }
  gilPopMatrix();
}

CSLIDEInstance *CSLIDEInstance::FindSelection(GLuint *puiID, int iLength)
{
  ASSERT( iLength >= 0);

  INT_32 iCode;
  CHAR pcBuffer[200];
  Tcl_Interp *pInterp;

  if( CSLIDEWorld::GetTclGUIExists() )
    {
      pInterp = CSLIDEWorld::GetInterp();
      sprintf(pcBuffer, "Graph:SelectedPathAdd %s", m_pcID );
      iCode = Tcl_Eval(pInterp, pcBuffer);
      //ASSERT( iCode == TCL_OK );
      if ( iCode != TCL_OK )
	{
	  printf( "%s\n", pcBuffer );
	  printf("Error %s \n", pInterp->result);
	}
    }

  if( iLength == 0 )
    {
      return this;
    }
  else
    {
      return m_pCSLIDENode->FindSelection(puiID, iLength);
    }
}

VOID CSLIDEInstance::RenderGUISelected(GLuint *puiID, int iLength,
				       Matrix &mChildFrame, Matrix &mParentFrame,
				       BOOL bDrawBBox, BOOL bDrawAxes,
				       BOOL bChildAxesSelected, BOOL bXAxisSelected,
				       BOOL bYAxisSelected, BOOL bZAxisSelected)
{
  ASSERT( iLength >= 0);

  gilPushMatrix();
  {
    if( iLength == 0 )
      {
	MATRIX4D m;

	// Parent
	if( bDrawAxes )
	  {
	    gilGetFloatv(GL_MODELVIEW_MATRIX, m);
	    mParentFrame.Set(m);
	    gilPushName(0);
	    {
	      m_pCSLIDENodeParent->RenderGUISelectedAxes(!bChildAxesSelected,bXAxisSelected,
							 bYAxisSelected, bZAxisSelected);
	    }
	    gilPopName();

	    // Child
	    gilMultMatrixf(m_mTransform.Get());

	    gilGetFloatv(GL_MODELVIEW_MATRIX, m);
	    mChildFrame.Set(m);

	    gilPushName(1);
	    {
	      m_pCSLIDENode->RenderGUISelectedAxes(bChildAxesSelected, bXAxisSelected,
						   bYAxisSelected, bZAxisSelected);
	    }
	    gilPopName();
	  }

	if( bDrawBBox )
	  {
	    // Child
	    m_pCSLIDENode->RenderGUISelectedBound();
	  }
      }
    else
      {
	gilMultMatrixf(m_mTransform.Get());
	m_pCSLIDENode->RenderGUISelected(puiID, iLength, mChildFrame, mParentFrame,
					 bDrawBBox, bDrawAxes, 
					 bChildAxesSelected, bXAxisSelected,
					 bYAxisSelected, bZAxisSelected);
      }
  }
  gilPopMatrix();
}

VOID CSLIDEInstance::RenderGraphSelectedBound()
{
  gilPushMatrix();
  {
    gilMultMatrixf(m_mTransform.Get());	
    m_pCSLIDENode->RenderGUISelectedBound();
  }
  gilPopMatrix();
}

VOID CSLIDEInstance::RenderGUISelectedBound()
{
  gilPushMatrix();
  {
    gilMultMatrixf(m_mTransform.Get());	
    m_pCSLIDENode->RenderGUISelectedBound();
  }
  gilPopMatrix();
}

VOID CSLIDEInstance::RenderBound()
{
  if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      // We do not store a bound at the instance level, so we must render
      // its child's bound transformed

      gilPushMatrix();
      {
	gilMultMatrixf(m_mTransform.Get());
	
	m_pCSLIDENode->RenderBound();
      }
      gilPopMatrix();
    }
}

VOID CSLIDEInstance::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDETransform *pCSLIDETransform;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "  %s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtInstance], 
	  m_pCSLIDENode->GetGUIID());

  if ( m_pcGUIID != NULL )
    {
      fprintf(fpOutputFile, "    %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtID], 
	      m_pcGUIID);
    } 
  if ( m_pcSurfaceID != gcppcSLIDEReservedWords[sliderwtSLF_INHERIT] )
    {
      fprintf(fpOutputFile, "    %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtSurface], 
	      m_pcSurfaceID);
    } 
  if ( m_dyniShading.GetValue() != slideshtSLF_INHERIT )
    {
      fprintf(fpOutputFile, "    %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtShading], 
	      gcppcSLIDEShadingReservedWords[m_dyniShading.GetValue()]);
    }
  if ( m_dyniLOD.GetValue() != slidelodtSLF_FULL )
    {
      fprintf(fpOutputFile, "    %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtLOD], 
	      gcppcSLIDELODReservedWords[m_dyniLOD.GetValue()]);
    }

  iter.Init(&m_listCSLIDETransform);

  for ( pCSLIDETransform = (CSLIDETransform*)iter.PeekFirst(); 
	pCSLIDETransform != NULL;
	pCSLIDETransform = (CSLIDETransform*)iter.PeekNext() )
    {
      pCSLIDETransform->WriteSLIDE(fpOutputFile);      
    }

  fprintf(fpOutputFile, "  %s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndInstance]);
}

VOID CSLIDEInstance::WriteRIB(FILE *fpOutputFile,
			      const Matrix &mGproj_obj,
			      CBound *pCBoundClip, 
			      SLIDEShadingType eShading,
			      SLIDELODType eLOD)
{
  ASSERT( fpOutputFile != NULL );

  ModifyShadingFlag(eShading);
  ModifyLODFlag(eLOD);

  if ( eLOD == slidelodtSLF_OFF )
    {
      // This branch has been turned off by the LOD flags
      return;
    }

  if ( eLOD > slidelodtSLF_BOUND )
    {
      WriteRIBPushSLIDEName(fpOutputFile);
      {
	WriteRIBPushShadingAttributes(fpOutputFile, eShading);
	{
	  m_pCSLIDESurface->WriteRIBPush(fpOutputFile);
	  {
	    WriteRIBGeometry(fpOutputFile, 
			     mGproj_obj, pCBoundClip, eShading, eLOD);
	  }
	  m_pCSLIDESurface->WriteRIBPop(fpOutputFile);
	}
	WriteRIBPopShadingAttributes(fpOutputFile, eShading);
      }
      WriteRIBPopSLIDEName(fpOutputFile);
    }
}

VOID CSLIDEInstance::WriteRIBGeometry(FILE *fpOutputFile,
				      const Matrix &mGproj_obj,
				      CBound *pCBoundClip, 
				      SLIDEShadingType eShading,
				      SLIDELODType eLOD)
{
  CListIterInt iter;
  CSLIDETransform *pCSLIDETransform;
  Matrix m;

  ASSERT( fpOutputFile != NULL );

  RiTransformBegin();
  {
    iter.Init(&m_listCSLIDETransform);
    
    // RenderMan is based on row vectors so we must list our 
    // column based operations in reverse order based on the effect of
    // transposing
    for ( pCSLIDETransform = (CSLIDETransform*)iter.PeekLast(); 
	  pCSLIDETransform != NULL;
	  pCSLIDETransform = (CSLIDETransform*)iter.PeekPrev() )
      {
	pCSLIDETransform->WriteRIB(fpOutputFile);      
      }

    m = mGproj_obj * m_mTransform;

    m_pCSLIDENode->WriteRIB(fpOutputFile, m, pCBoundClip, eShading, eLOD);
  }
  RiTransformEnd();
}

VOID CSLIDEInstance::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  if ( m_pcGUIID != NULL )
    {
      RiArchiveRecord(RI_COMMENT, "BEGIN SLIDE INSTANCE: %s\n", m_pcGUIID);
    }
  WriteRIBBegin(fpOutputFile);
}

VOID CSLIDEInstance::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBEnd(fpOutputFile);
  if ( m_pcGUIID != NULL )
    {
      RiArchiveRecord(RI_COMMENT, "  END SLIDE INSTANCE: %s\n", m_pcGUIID);
    }
}

VOID CSLIDEInstance::WritePOV(FILE *fpOutputFile, 
			      CListInt *plistTransformIDs,
			      CListInt *plistColors) 
{
  ASSERT( fpOutputFile != NULL );
  CHAR pcTransformID[100];

  if (m_listCSLIDETransform.PeekFirst() == NULL)
    WritePOVGeometry(fpOutputFile, plistTransformIDs, plistColors);
  else
    {
      ASSERT( GetID() != NULL );
      sprintf(pcTransformID, "Trans_%s", GetID());
      
      plistTransformIDs->InsertFirst(pcTransformID, 0);   // Push on id for transform      
      WritePOVGeometry(fpOutputFile, plistTransformIDs, plistColors);
      plistTransformIDs->RemoveFirst();                   // Pop
    }
}

VOID CSLIDEInstance::WritePOVGeometry(FILE *fpOutputFile, 
				      CListInt *plistTransformIDs,
				      CListInt *plistColors) 
{
  ASSERT( fpOutputFile != NULL );

  CSLIDETransform *pCSLIDETransform;
  CListIterInt iter;
  CHAR pcColorID[100];  

  iter.Init(&m_listCSLIDETransform);

  if( (CSLIDETransform*)iter.PeekFirst() != NULL )
    {      
      fprintf(fpOutputFile, "#declare Trans_%s = transform { \n", GetID());
            
      for ( pCSLIDETransform = (CSLIDETransform*)iter.PeekFirst(); 
	    pCSLIDETransform != NULL;
	    pCSLIDETransform = (CSLIDETransform*)iter.PeekNext() )
	{
	  pCSLIDETransform->WritePOV(fpOutputFile);
	}
      fprintf(fpOutputFile, "} \n\n");
      
      //output transform identifier
    }

  if (strcmp(GetSurfaceID(), "SLF_INHERIT")) {
    fprintf(fpOutputFile, "#declare color_%s = ", GetSurfaceID());
    m_pCSLIDESurface->WritePOV(fpOutputFile);
    fprintf(fpOutputFile, "; \n");
    //output color identifier
  }
  sprintf(pcColorID, "color_%s", GetSurfaceID());

  plistColors->InsertFirst(pcColorID, 0);
  m_pCSLIDENode->WritePOV(fpOutputFile, plistTransformIDs, plistColors);
  plistColors->RemoveFirst();
}

VOID CSLIDEInstance::WriteWRLZ(FILE *fpOutputFile, const Matrix &mTransform, 
			       CSLIDESurface *pCSLIDESurface)
{
  ASSERT( fpOutputFile != NULL );

  WriteWRLZGeometry(fpOutputFile, mTransform, pCSLIDESurface);
}

VOID CSLIDEInstance::WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				       CSLIDESurface *pCSLIDESurface)
{
  ASSERT( fpOutputFile != NULL );

  Matrix m;
  CSLIDESurface *pCSLIDESurfacePush;

  m = mTransform * m_mTransform;
  
  if( m_pCSLIDESurface == CSLIDESurface::GetSLF_INHERIT() )
    {
      pCSLIDESurfacePush = pCSLIDESurface;
    }
  else
    {
      pCSLIDESurfacePush = m_pCSLIDESurface;
    }

  m_pCSLIDENode->WriteWRLZ(fpOutputFile, m, pCSLIDESurfacePush);
}

VOID CSLIDEInstance::WriteSTL(FILE *fpOutputFile, const Matrix &mTransform)
{
  ASSERT( fpOutputFile != NULL );

  WriteSTLGeometry(fpOutputFile, mTransform);
}

VOID CSLIDEInstance::WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform)
{
  ASSERT( fpOutputFile != NULL );

  Matrix m;
  //m = m_mTransform * mTransform;
  m = mTransform * m_mTransform;

  m_pCSLIDENode->WriteSTL(fpOutputFile, m);
}

QEMesh *CSLIDEInstance::WriteSIF(FILE *fpOutputFile, UINT_32 uiIndentLevel, 
				 const Matrix &mGroot_obj, const Matrix &mGobj_root,
				 QEMesh *pmeBoundaryEdges)
{
  QEMesh *pme;
  Matrix m, mInv;

  m = mGroot_obj * m_mTransform;
  mInv = m_mInverse * mGobj_root;

  pme = m_pCSLIDENode->WriteSIF(fpOutputFile, uiIndentLevel, m, mInv, pmeBoundaryEdges);
  if ( pme != NULL )
    {
      pme->Transform(m_mTransform, m_mInverse);
    }

  return pme;
}

VOID CSLIDEInstance::WriteLSIF(FILE *fpOutputFile, UINT_32 uiIndentLevel)
{
  m_pCSLIDENode->WriteLSIF(fpOutputFile, uiIndentLevel);
}

VOID CSLIDEInstance::WriteLSIFOffset(FILE *fpOutputFile, UINT_32 uiIndentLevel)
{
  m_pCSLIDENode->WriteLSIFOffset(fpOutputFile, uiIndentLevel);
}

VOID CSLIDEInstance::WriteDotID(FILE *fpOutputFile)
{
  static UINT_32 suiDotID = 0;

  ASSERT( fpOutputFile != NULL );

  if ( m_uiDotID == 0 )
    {
      suiDotID++;
      m_uiDotID = suiDotID;
    }

  fprintf(fpOutputFile, "instance_%ld", m_uiDotID);
}

VOID CSLIDEInstance::WriteDotNode(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDETransform *pCSLIDETransform;
  UINT_32 i;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "  "); 
  WriteDotID(fpOutputFile);

  fprintf(fpOutputFile, "[ shape = record"); 
  if ( GetTransformDynamic() )
    {
      fprintf(fpOutputFile, ", color = red, fontcolor = red");
    }
  else if ( GetBoundDynamic() )
    {
      fprintf(fpOutputFile, ", color = purple4, fontcolor = purple4");
    }
  else if ( GetGeometryDynamic() )
    {
      fprintf(fpOutputFile, ", color = blue, fontcolor = blue");
    }
  fprintf(fpOutputFile, ", label = \"{"); 

  if ( GetGUIID() != NULL )
    {
      fprintf(fpOutputFile, "%s", GetGUIID());
    }
  else
    {
      fprintf(fpOutputFile, "Instance %ld", m_uiDotID);
    }
  fprintf(fpOutputFile, " | {");

  i = 0;
  iter.Init(&m_listCSLIDETransform);

  for ( pCSLIDETransform = (CSLIDETransform*)iter.PeekFirst(); 
	pCSLIDETransform != NULL;
	pCSLIDETransform = (CSLIDETransform*)iter.PeekNext() )
    {
      if ( i > 0 )
	{
	  fprintf(fpOutputFile, "|");
	}
      pCSLIDETransform->WriteDotNode(fpOutputFile);      
      i++;
    }
  fprintf(fpOutputFile, "}}\"];\n");
}

VOID CSLIDEInstance::WriteDotEdges(FILE *fpOutputFile)
{
  CListIterInt iter;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "  ");
  WriteDotID(fpOutputFile);
  fprintf(fpOutputFile, " -> ");
  m_pCSLIDENode->WriteDotID(fpOutputFile);
  fprintf(fpOutputFile, "\n");
}

UINT_32 CSLIDEInstance::GetTransformCount()
{
  return m_listCSLIDETransform.GetLength();
}

inline QEMesh *CSLIDEInstance::GetQEMesh()
{
  QEMesh *pQEMesh;

  pQEMesh = m_pCSLIDENode->GetQEMesh();
  if ( pQEMesh != NULL )
    {
      pQEMesh->Transform(m_mTransform, m_mInverse);
    }

  return pQEMesh;
}

VOID CSLIDEInstance::CalculateLeafQEMesh()
{
  m_pCSLIDENode->CalculateLeafQEMesh();
}

VOID CSLIDEInstance::DeallocateLeafQEMesh()
{
  m_pCSLIDENode->DeallocateLeafQEMesh();
}

VOID CSLIDEInstance::GetParameters(Tcl_Interp *pInterp)
{
  INT_32 iCode;
  CHAR pcBuffer[1000];
  CHAR pcBuffer2[500];
  CHAR pcSurface[35];

  if( !strcmp(m_pCSLIDESurface->GetGUIID(), "SLF_INHERIT" ) )
      sprintf(pcSurface, "inherit");
  else 
      sprintf(pcSurface, m_pCSLIDESurface->GetGUIID());

  sprintf(pcBuffer, "Editor:SetParameterValues Instance {{surface %s} {shading %s} {lod %s} {ribbegin %s} {ribend %s}}",
	  pcSurface,
	  gcppcSLIDEGUIShadingReservedWords[m_dyniShading.GetValue()],
	  gcppcSLIDEGUILODReservedWords[m_dyniLOD.GetValue()],
	  m_pcRIBBegin,
	  m_pcRIBEnd);
	  
  iCode = Tcl_Eval(pInterp, pcBuffer);
  //ASSERT( iCode == TCL_OK );
  if ( iCode != TCL_OK )
    {
      printf( "%s\n", pcBuffer );
      printf("Error %s \n", pInterp->result);
    }

  CListIterInt iter;
  CSLIDETransform *pCSLIDETransform;
  FLOAT x, y, z, theta;
  BOOL bTransforms;

  bTransforms = FALSE;
  x = y = z = theta = 0;
  sprintf(pcBuffer, "InstanceMenu:SetTransforms {");

  iter.Init(&m_listCSLIDETransform);
  for ( pCSLIDETransform = (CSLIDETransform*)iter.PeekFirst(); 
	pCSLIDETransform != NULL;
	pCSLIDETransform = (CSLIDETransform*)iter.PeekNext() )
    {
      if( !strcmp(pCSLIDETransform->GetTypeString(), "scale") )
	{
	  x = ((CSLIDEScale*)pCSLIDETransform)->GetScale()[X];
	  y = ((CSLIDEScale*)pCSLIDETransform)->GetScale()[Y];
	  z = ((CSLIDEScale*)pCSLIDETransform)->GetScale()[Z];
	}
      else if( !strcmp(pCSLIDETransform->GetTypeString(), "translate" ) )
	{
	  x = ((CSLIDETranslate*)pCSLIDETransform)->GetTranslate()[X];
	  y = ((CSLIDETranslate*)pCSLIDETransform)->GetTranslate()[Y];
	  z = ((CSLIDETranslate*)pCSLIDETransform)->GetTranslate()[Z];
	}
      else if( !strcmp(pCSLIDETransform->GetTypeString(), "rotate" ) )
	{
	  x = ((CSLIDERotate*)pCSLIDETransform)->GetAxis()[X];
	  y = ((CSLIDERotate*)pCSLIDETransform)->GetAxis()[Y];
	  z = ((CSLIDERotate*)pCSLIDETransform)->GetAxis()[Z];
	  theta = ((CSLIDERotate*)pCSLIDETransform)->GetTheta().GetValue();
	}
      else if( !strcmp(pCSLIDETransform->GetTypeString(), "lookat" ) )
	{
	  ASSERT(FALSE);
	}
      else
	{
	  ASSERT(FALSE);
	}

      sprintf(pcBuffer2," {%s %f %f %f %f}", 
	      pCSLIDETransform->GetTypeString(), x, y, z, RADTODEG(theta));
      strcat(pcBuffer, pcBuffer2);

      bTransforms = TRUE;
    }

  if( bTransforms ) 
    {
      strcat(pcBuffer, "} 1");      
    }
  else 
    {
      sprintf(pcBuffer, "InstanceMenu:SetTransforms {{none}} 1");
    }

  iCode = Tcl_Eval(pInterp, pcBuffer);
  //ASSERT( iCode == TCL_OK );
  if ( iCode != TCL_OK )
    {
      printf( "%s\n", pcBuffer );
      printf("Error %s \n", pInterp->result);
    }
}

UINT_32 CSLIDEInstance::SetParameter(CSLIDEWorld *pCSLIDEWorld,
				     SLIDEReservedWordType uiParameter, CHAR **argv)
{
  UINT_32 i, uiTransforms;
  CHAR *pcType;
  FLOAT fX, fY, fZ, fT;
  CListIterInt iter;
  CSLIDETransform *pCSLIDETransform;
  CSLIDETranslate *pCSLIDETranslate;
  CSLIDEScale *pCSLIDEScale;
  CSLIDERotate *pCSLIDERotate;
  DynVector *pdynv;
  DynFloat *pdynf;
   
  switch(uiParameter) 
    {
    case sliderwtTransform:

      m_bGUIChangedTransforms = TRUE;

      // Delete old transforms
      iter.Init(&m_listCSLIDETransform);
      for ( pCSLIDETransform = (CSLIDETransform*)iter.PeekFirst(); 
	    pCSLIDETransform != NULL;
	    pCSLIDETransform = (CSLIDETransform*)iter.PeekNext() )
	{
	  pCSLIDEWorld->RemoveTransform(pCSLIDETransform);
	  delete pCSLIDETransform;
	}      
      m_listCSLIDETransform.RemoveAll();

      // Create new transforms
      uiTransforms = atoi(argv[0]);
      for( i = 0; i < uiTransforms; i++ ) {

	pcType =  argv[(i*5)+1];
	fX = atof(argv[(i*5)+2]);
	fY = atof(argv[(i*5)+3]);
	fZ = atof(argv[(i*5)+4]);
	fT = atof(argv[(i*5)+5]);

	pdynv = new DynVector;

	if( !strcmp(pcType, "translate") )
	  {
	    pCSLIDETranslate = new CSLIDETranslate();
	    ASSERT( pCSLIDETranslate != NULL );
	    pdynv->Set(fX, fY, fZ);
	    pCSLIDETranslate->Init("[GUI generated Translate]", 0, 0, pdynv);
	    VERIFY(pCSLIDEWorld->InsertTransform(pCSLIDETranslate));
	    InsertTransform(pCSLIDETranslate);
	  }
	else if( !strcmp(pcType, "scale") )
	  {
	    pCSLIDEScale = new CSLIDEScale();
	    ASSERT( pCSLIDEScale != NULL );
	    pdynv->Set(fX, fY, fZ);
	    pCSLIDEScale->Init("[GUI generated Scale]", 0, 0, pdynv);
	    VERIFY(pCSLIDEWorld->InsertTransform(pCSLIDEScale));
	    InsertTransform(pCSLIDEScale);
	  }
	else if( !strcmp(pcType, "rotate") )
	  {
	    pdynf = new DynFloat;
	    pCSLIDERotate = new CSLIDERotate();
	    ASSERT( pCSLIDERotate != NULL );
	    pdynv->Set(fX, fY, fZ);
	    pdynf->SetValue(DEGTORAD(fT));
	    pCSLIDERotate->Init("[GUI generated Rotate]", 0, 0, pdynv, pdynf);
	    delete pdynf;
	    VERIFY(pCSLIDEWorld->InsertTransform(pCSLIDERotate));
	    InsertTransform(pCSLIDERotate);
	  }
	else
	  {
	    ASSERT(FALSE);
	  }
	delete pdynv;
      }
      
      // Recalculate this instance's Transform
      m_mTransform.SetIdentity(); 
      m_mInverse.SetIdentity(); 
      
      iter.Init(&m_listCSLIDETransform);
      for ( pCSLIDETransform = (CSLIDETransform*)iter.PeekFirst(); 
	    pCSLIDETransform != NULL;
	    pCSLIDETransform = (CSLIDETransform*)iter.PeekNext() )
	{
	  m_mTransform = pCSLIDETransform->GetMatrix() * m_mTransform;      
	  m_mInverse = m_mInverse * pCSLIDETransform->GetInverse();
	}
      return (5*uiTransforms + 1);      
      break;
    default:
      return CSLIDEGeometryNode::SetParameter(pCSLIDEWorld, uiParameter, argv);
      break;
    }
}
