#ifndef __SLIDEINSTANCE_H__
#define __SLIDEINSTANCE_H__

#include "slidegeometrynode.h"

class CSLIDETransform;

//////////////////////////////////////////////////////////////////////////////
// CSLIDEInstance Class
//

class CSLIDEInstance : public CSLIDEGeometryNode
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEInstance();
  virtual ~CSLIDEInstance();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcChildID, CHAR *pcParentID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  inline VOID SetID(CHAR *pcID);
  VOID InsertTransform(CSLIDETransform *pCSLIDETransform);
  VOID InsertTurtleWalkTransform(CSLIDETransform *pCSLIDETransform, BOOL bNewTransform);
  VOID InsertCommandPostTransform(CSLIDETransform *pCSLIDETransform,  BOOL bNewTransform);

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

  virtual VOID Remove(CSLIDEWorld *pCSLIDEWorld);
  virtual VOID RemoveFromParent(CSLIDEWorld *pCSLIDEWorld);
  virtual VOID RemoveFromChild(CSLIDEWorld *pCSLIDEWorld);

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual CSLIDECamera *FindCamera(CListIterInt *piterIDs, 
				   CListInt *plistCSLIDEInstance);
  virtual CSLIDELight *FindLight(CListIterInt *piterIDs, 
				 CListInt *plistCSLIDEInstance);
  virtual CSLIDEAreaLight *FindAreaLight(CListIterInt *piterIDs, 
					 CListInt *plistCSLIDEInstance);

  virtual VOID PreprocessGraph();
  virtual VOID Preprocess();
  virtual VOID SetUnvisited();
  virtual VOID SetBoundUnvisited();
  virtual VOID SetBoundUnchanged();

  virtual VOID Update();
  virtual VOID UpdateGraph();
  virtual VOID InsertIntoBound(CBound &bound);

#ifdef __USE_CDA__
  virtual inline CDA *GetCDA();
  virtual VOID CalculateCDA();
  virtual VOID CalculateCDAVertices(CListInt *plistCDAVertex, 
				    const Matrix &mGroot_obj);
#endif // __USE_CDA__

  virtual QEMesh *GetQEMesh();
  virtual VOID CalculateLeafQEMesh();  // for WriteWRLZ
  virtual VOID DeallocateLeafQEMesh();  // for WriteWRLZ

protected:
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();
  virtual VOID CalculateBound();

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID PrerenderGeometry(SLIDEShadingType eShading);
  virtual VOID PrerenderNormals(SLIDEShadingType eShading);
  virtual VOID PrerenderBounds();

  virtual VOID Prerender();
  virtual VOID SetPrerenderUnvisited();

  virtual VOID RecallGeometry(SLIDEShadingType eShading);
  virtual VOID RecallNormals(SLIDEShadingType eShading);
  virtual VOID RecallBounds();

  virtual VOID Render(const Matrix &mGproj_obj,
		      CBound *pCBoundClip, 
		      SLIDEShadingType eShading,
		      SLIDELODType eLOD);

  virtual VOID RenderSelect(const Matrix &mGproj_obj,
			    CBound *pCBoundClip, 
			    SLIDEShadingType eShading,
			    SLIDELODType eLOD);

  virtual VOID RenderGeometry(const Matrix &mGproj_obj,
			      CBound *pCBoundClip, 
			      SLIDEShadingType eShading,
			      SLIDELODType eLOD,
			      CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderBound();

  virtual VOID RenderGeometrySelect(const Matrix &mGproj_obj,
				    CBound *pCBoundClip, 
				    SLIDEShadingType eShading,
				    SLIDELODType eLOD);
  virtual CSLIDEInstance *FindSelection(GLuint *puiID, int iLength);  
  virtual VOID RenderGUISelected(GLuint *puiID, int iLength,
				 Matrix &mChildFrame, Matrix &mParentFrame, 
				 BOOL bDrawBBox, BOOL bDrawAxes,
				 BOOL bChildAxesSelected, BOOL bXAxisSelected,
				 BOOL bYAxisSelected, BOOL bZAxisSelected);
  virtual VOID RenderGraphSelectedBound();
  virtual VOID RenderGUISelectedBound();

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIB(FILE *fpOutputFile,
			const Matrix &mGproj_obj,
			CBound *pCBoundClip, 
			SLIDEShadingType eShading,
			SLIDELODType eLOD);
  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

  virtual VOID WritePOV(FILE *fpOutputFile, 
			CListInt *plistTransformIDs,
			CListInt *plistColors); 
  virtual VOID WritePOVGeometry(FILE *fpOutputFile, 
				CListInt *plistTransformIDs,
				CListInt *plistColors); 

  virtual VOID WriteWRLZ(FILE *fpOutputFile, const Matrix &mTransform,
			 CSLIDESurface *pCSLIDESurface);
  virtual VOID WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				 CSLIDESurface *pCSLIDESurface);

  virtual VOID WriteSTL(FILE *fpOutputFile, const Matrix &mTransform);
  virtual VOID WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform);

  virtual QEMesh *WriteSIF(FILE *fpOutputFile, UINT_32 uiIndentLevel, 
			   const Matrix &mGroot_obj, const Matrix &mGobj_root,
			   QEMesh *pmeBoundaryEdges);
  virtual VOID WriteLSIF(FILE *fpOutputFile, UINT_32 uiIndentLevel); 
  virtual VOID WriteLSIFOffset(FILE *fpOutputFile, UINT_32 uiIndentLevel); 

  virtual VOID WriteDotID(FILE *fpOutputFile);
  virtual VOID WriteDotNode(FILE *fpOutputFile);
  virtual VOID WriteDotEdges(FILE *fpOutputFile);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline CHAR *GetChildID();
  virtual inline CHAR *GetParentID();
  virtual inline CSLIDENode *GetNode();
  virtual inline CSLIDENode *GetNodeParent();

  virtual inline const Matrix &GetTransform();
  virtual inline const Matrix &GetInverse();

  virtual inline BOOL GetTransformDynamic();
  virtual inline VOID SetTransformDynamic(BOOL bTransformDynamic);
  virtual inline BOOL GetBoundEmpty();
  virtual inline VOID SetBoundEmpty(BOOL bBoundEmpty);

  virtual UINT_32 GetTransformCount();

  virtual VOID GetSide(CListInt *plistPoint, CListInt *plistNormal, UINT_32 uiSide, 
		       const Matrix &mTransform, const Matrix &mInverse, BOOL bReverse);

  virtual VOID GetParameters(Tcl_Interp *pInterp);
  virtual UINT_32 SetParameter(CSLIDEWorld *pCSLIDEWorld,
			       SLIDEReservedWordType uiParameter, CHAR **argv);

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CHAR *m_pcChildID;
  CHAR *m_pcParentID;
  CSLIDENode *m_pCSLIDENode;
  CSLIDENode *m_pCSLIDENodeParent;
  
  CListInt m_listCSLIDETransform;
  Matrix m_mTransform;
  Matrix m_mInverse;

  BOOL m_bTransformDynamic;
  BOOL m_bBoundEmpty;
  BOOL m_bGUIChangedTransforms;

  CSLIDETransform *m_pCSLIDETransformLastAdded;

  UINT_32 m_uiDotID;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDEInstance::SetID(CHAR *pcID)
{
  m_pcID = pcID;

  m_pcGUIID = new CHAR[strlen(pcID)+1];
  ASSERT(m_pcGUIID != NULL);
  strcpy(m_pcGUIID, pcID);
}

inline CHAR *CSLIDEInstance::GetChildID()
{
  return m_pcChildID;
}

inline CHAR *CSLIDEInstance::GetParentID()
{
  return m_pcParentID;
}

inline CSLIDENode *CSLIDEInstance::GetNode()
{
  return m_pCSLIDENode;
}

inline CSLIDENode *CSLIDEInstance::GetNodeParent()
{
  return m_pCSLIDENodeParent;
}

inline const Matrix &CSLIDEInstance::GetTransform()
{
  return m_mTransform;
}

inline const Matrix &CSLIDEInstance::GetInverse()
{
  return m_mInverse;
}

inline BOOL CSLIDEInstance::GetTransformDynamic()
{
  return m_bTransformDynamic;
}

inline VOID CSLIDEInstance::SetTransformDynamic(BOOL bTransformDynamic)
{
  m_bTransformDynamic = bTransformDynamic;
}

inline BOOL CSLIDEInstance::GetBoundEmpty()
{
  return m_bBoundEmpty;
}

inline VOID CSLIDEInstance::SetBoundEmpty(BOOL bBoundEmpty)
{
  m_bBoundEmpty = bBoundEmpty;
}

#ifdef __USE_CDA__
inline CDA *CSLIDEInstance::GetCDA()
{
  CDA *pCDA;

  pCDA = m_pCSLIDENode->GetCDA();
  if ( pCDA != NULL )
    {
      pCDA->Transform(m_mTransform, m_mInverse);
    }

  return pCDA;
}
#endif // __USE_CDA__

#endif // __SLIDEINSTANCE_H__
