#include "slideintersection.h"

#include "slideworld.h"
#include "slidesurface.h"
#include "slideinstance.h"
#include "cdapolyhedron.h"
#include "cdavertex.h"

#include "qemesh.h"
#include "qewrite.h"
#include "sif.h"

//////////////////////////////////////////////////////////////////////
// CSLIDEIntersection Class
//

CSLIDEIntersection::CSLIDEIntersection()
{
}

CSLIDEIntersection::~CSLIDEIntersection()
{
}

const POINT4D cpptCube0[8] = {
  {-3, -3, -3, 1},
  {-3, -3, 2, 1},
  {-3, 2, -3, 1},
  {-3, 2, 2, 1},
  {2, -3, -3, 1},
  {2, -3, 2, 1},
  {2, 2, -3, 1},
  {2, 2, 2, 1}
};

const POINT4D cpptCube1[8] = {
  {1, 1, 1, 1},
  {1, 1, 3, 1},
  {1, 3, 1, 1},
  {1, 3, 3, 1},
  {3, 1, 1, 1},
  {3, 1, 3, 1},
  {3, 3, 1, 1},
  {3, 3, 3, 1}
};

const POINT4D cpptCube[8] = {
  {-1, -1, -1, 1},
  {-1, -1, 1, 1},
  {-1, 1, -1, 1},
  {-1, 1, 1, 1},
  {1, -1, -1, 1},
  {1, -1, 1, 1},
  {1, 1, -1, 1},
  {1, 1, 1, 1}
};

const POINT4D cpptOctahedron[6] = {
  {1, 0, 0, 1},
  {-1, 0, 0, 1},
  {0, 1, 0, 1},
  {0, -1, 0, 1},
  {0, 0, 1, 1},
  {0, 0, -1, 1}
};

const POINT4D cpptEasy[5] = {
  {1, 0, 0, 1},
  {-1, 0, 0, 1},
  {0, 1, 0, 1},
  {0, 0.5, 0.5, 1},
  {0, 0.5, -0.5, 1}
};

const POINT4D cpptTetrahedron[4] = {
  {1, 0, 0, 1},
  {-1, 0, 0, 1},
  {0, 1, 0, 1},
  {0, 0.5, 0.5, 1},
};

VOID CSLIDEIntersection::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		       CHAR *pcID)
{
  CSLIDEGroup::Init(pcFile, uiLine, uiColumn, pcID);

  m_listCSLIDEInstance.Init();

  ///////////////////////////////////////
  // START: Jordan Fix This!!!
  //
//#define INTERSECTION_TEST
#ifdef INTERSECTION_TEST
  Point pt;
  UINT_32 i;
  CListInt listCDAVertex;
  CDAVertex *pCDAVertex;
  
  listCDAVertex.Init();
#endif // INTERSECTION_TEST

//#define CONVEX_HULL_TEST
#ifdef CONVEX_HULL_TEST
  // Cube Test
  for ( i = 0; i < 8; i++ )
    {
      pt.Set(cpptCube[i]);
  /*
  // Octahedron Test
  for ( i = 0; i < 6; i++ )
    {
      pt.Set(cpptOctahedron[i]);
      */
  /*
  // Easy Test
  for ( i = 0; i < 5; i++ )
    {
      pt.Set(cpptEasy[i]);
      */
  /*
  // Tetrahedron Test
  for ( i = 0; i < 4; i++ )
    {
      pt.Set(cpptTetrahedron[i]);
      */
      pCDAVertex = new CDAVertex;
      ASSERT( pCDAVertex != NULL );
      pCDAVertex->Init(pt);
      
      listCDAVertex.InsertLast(pCDAVertex, 0);
    }

  m_pCDAPolyhedron = new CDAPolyhedron;
  ASSERT( m_pCDAPolyhedron != NULL ); 
  m_pCDAPolyhedron->Init(&listCDAVertex); 
  
  for ( pCDAVertex = (CDAVertex*)(listCDAVertex.RemoveFirst());
	pCDAVertex != NULL;
	pCDAVertex = (CDAVertex*)(listCDAVertex.RemoveFirst()) )
    {
      delete pCDAVertex;
    }
 
#endif // CONVEX_HULL_TEST
  
#ifdef INTERSECTION_TEST
  // This is a Hack to get the free lists set up

  // Tetrahedron Test
  for ( i = 0; i < 4; i++ )
    {
      pt.Set(cpptTetrahedron[i]);
      pCDAVertex = new CDAVertex;
      ASSERT( pCDAVertex != NULL );
      pCDAVertex->Init(pt);
      
      listCDAVertex.InsertLast(pCDAVertex, 0);
    }

  m_ppCDAPolyhedron[2] = new CDAPolyhedron;
  ASSERT( m_ppCDAPolyhedron[2] != NULL ); 
  m_ppCDAPolyhedron[2]->Init(&listCDAVertex); 
  
  for ( pCDAVertex = (CDAVertex*)(listCDAVertex.RemoveFirst());
	pCDAVertex != NULL;
	pCDAVertex = (CDAVertex*)(listCDAVertex.RemoveFirst()) )
    {
      delete pCDAVertex;
    }

  //////////////////////////////////////////////////
  // Cube0 Test
  for ( i = 0; i < 8; i++ )
    {
      pt.Set(cpptCube0[i]);

      pCDAVertex = new CDAVertex;
      ASSERT( pCDAVertex != NULL );
      pCDAVertex->Init(pt);
      
      listCDAVertex.InsertLast(pCDAVertex, 0);
    }

  m_ppCDAPolyhedron[0] = AllocateCDAPolyhedron();
  ASSERT( m_ppCDAPolyhedron[0] != NULL ); 
  m_ppCDAPolyhedron[0]->Init(&listCDAVertex); 
  
  for ( pCDAVertex = (CDAVertex*)(listCDAVertex.RemoveFirst());
	pCDAVertex != NULL;
	pCDAVertex = (CDAVertex*)(listCDAVertex.RemoveFirst()) )
    {
      delete pCDAVertex;
    }
 
  //////////////////////////////////////////////////
  // Cube1 Test
  for ( i = 0; i < 8; i++ )
    {
      pt.Set(cpptCube1[i]);

      pCDAVertex = new CDAVertex;
      ASSERT( pCDAVertex != NULL );
      pCDAVertex->Init(pt);
      
      listCDAVertex.InsertLast(pCDAVertex, 0);
    }

  m_ppCDAPolyhedron[1] = AllocateCDAPolyhedron();
  ASSERT( m_ppCDAPolyhedron[1] != NULL ); 
  m_ppCDAPolyhedron[1]->Init(&listCDAVertex); 
  
  for ( pCDAVertex = (CDAVertex*)(listCDAVertex.RemoveFirst());
	pCDAVertex != NULL;
	pCDAVertex = (CDAVertex*)(listCDAVertex.RemoveFirst()) )
    {
      delete pCDAVertex;
    }
 
  // Hack Continued
  delete m_ppCDAPolyhedron[2];

  // Take their intersection

  m_ppCDAPolyhedron[2] = m_ppCDAPolyhedron[0]->Intersect(m_ppCDAPolyhedron[1]);
  ASSERT( m_ppCDAPolyhedron[2] != NULL );
  m_ppCDAPolyhedron[2]->Bind(1);
#endif // INTERSECTION_TEST

  //
  // END: Jordan Fix This!!!
  ///////////////////////////////////////
}

VOID CSLIDEIntersection::Uninit()
{

  CSLIDEGroup::Uninit();

#ifdef INTERSECTION_TEST
  UINT_32 i;

  for ( i = 0; i < 3; i++ )
    {
      if ( m_ppCDAPolyhedron[i] != NULL )
	{
	  DeallocateCDAPolyhedron(m_ppCDAPolyhedron[i]);
	}
    }
#endif // INTERSECTION_TEST
}

BOOL CSLIDEIntersection::Initialized()
{
  if ( !CSLIDEGroup::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

VOID CSLIDEIntersection::InsertInstance(CSLIDEInstance *pCSLIDEInstance)
{
  ASSERT( pCSLIDEInstance != NULL );

  m_listCSLIDEInstance.InsertLast(pCSLIDEInstance, 0);
}

BOOL CSLIDEIntersection::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  if ( !CSLIDEGroup::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  return TRUE;
}

BOOL CSLIDEIntersection::Valid()
{
  if ( !CSLIDEGroup::Valid() )
    {
      return FALSE;
    }

  return TRUE;
}

CSLIDECamera *CSLIDEIntersection::FindCamera(CListIterInt *piterIDs, 
					     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtCamera],
	      gcppcSLIDEReservedWords[sliderwtIntersection], GetID());

  return NULL;
}

CSLIDELight *CSLIDEIntersection::FindLight(CListIterInt *piterIDs, 
				     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtLight],
	      gcppcSLIDEReservedWords[sliderwtIntersection], GetID());

  return NULL;
}

CSLIDEAreaLight *CSLIDEIntersection::FindAreaLight(CListIterInt *piterIDs, 
						   CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtAreaLight],
	      gcppcSLIDEReservedWords[sliderwtIntersection], GetID());

  return NULL;
}

VOID CSLIDEIntersection::Preprocess()
{
  CSLIDEGroup::Preprocess();

  CheckDynamic();
  UpdateState();
}

VOID CSLIDEIntersection::CheckDynamic()
{
  SetDynamic(m_dyniShading.GetDynamic() || 
	     m_dyniLOD.GetDynamic());
}

VOID CSLIDEIntersection::PreprocessGraph()
{
  if ( !GetVisited() )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;

      BOOL bGeometryEmpty;
      BOOL bGeometryDynamic;
      BOOL bNormalsEmpty;
      BOOL bNormalsDynamic;
      BOOL bBoundsEmpty;
      BOOL bBoundsDynamic;
      BOOL bBoundEmpty;
      BOOL bBoundDynamic;

      bGeometryEmpty = TRUE;
      bGeometryDynamic = FALSE;
      bNormalsEmpty = TRUE;
      bNormalsDynamic = FALSE;
      bBoundsEmpty = TRUE;
      bBoundsDynamic = FALSE;

      bBoundEmpty = TRUE;
      bBoundDynamic = FALSE;

      //////////////////////////////////////////////////
      // Visit Children Nodes and get starting values from them
      iter.Init(&m_listCSLIDEInstance);
      
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pCSLIDEInstance->PreprocessGraph();

	  bGeometryEmpty = ( bGeometryEmpty && 
			     pCSLIDEInstance->GetGeometryEmpty() );
	  bGeometryDynamic = ( bGeometryDynamic ||
			       pCSLIDEInstance->GetGeometryDynamic() );
	  bNormalsEmpty = ( bNormalsEmpty && 
			    pCSLIDEInstance->GetNormalsEmpty() );
	  bNormalsDynamic = ( bNormalsDynamic ||
			      pCSLIDEInstance->GetNormalsDynamic() );
	  bBoundsEmpty = ( bBoundsEmpty && 
			   pCSLIDEInstance->GetBoundsEmpty() );
	  bBoundsDynamic = ( bBoundsDynamic ||
			     pCSLIDEInstance->GetBoundsDynamic() );
	  bBoundEmpty = ( bBoundEmpty && 
			  pCSLIDEInstance->GetBoundEmpty() );
	  bBoundDynamic = ( bBoundDynamic ||
			    pCSLIDEInstance->GetBoundDynamic() );
	}

      //////////////////////////////////////////////////
      // Look at Local flags

      // Update whether we are Empty based on our LOD flag
      if ( !m_dyniLOD.GetDynamic() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      // This node is always off
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	      bBoundsEmpty = TRUE;
	      bBoundEmpty = TRUE;
	    }
	  else if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	    {
	      // This node is always just a bounding box
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	    }
	}

      // Check what is really dynamic

      bGeometryDynamic = ( bGeometryDynamic ||
			   m_pCSLIDESurface->GetDynamic() ||
			   m_dyniShading.GetDynamic() ||
			   m_dyniLOD.GetDynamic() );
      bNormalsDynamic = ( bNormalsDynamic ||
			  m_dyniShading.GetDynamic() ||
			  m_dyniLOD.GetDynamic() );
      bBoundsDynamic = ( bBoundsDynamic || 
			 m_dyniLOD.GetDynamic() );
      bBoundDynamic = ( bBoundDynamic || 
			m_dyniLOD.GetDynamic() );

      // If anything is now Empty then it will always be empty so
      // it is not dynamic

      bGeometryDynamic = ( bGeometryDynamic && !bGeometryEmpty );
      bNormalsDynamic = ( bNormalsDynamic && !bNormalsEmpty );
      bBoundsDynamic = ( bBoundsDynamic && !bBoundsEmpty );
      bBoundDynamic = ( bBoundDynamic && !bBoundEmpty );

      //////////////////////////////////////////////////
      // Now save these correct flag values

      SetGeometryEmpty(bGeometryEmpty);
      SetGeometryDynamic(bGeometryDynamic);
      SetNormalsEmpty(bNormalsEmpty);
      SetNormalsDynamic(bNormalsDynamic);
      SetBoundsEmpty(bBoundsEmpty);
      SetBoundsDynamic(bBoundsDynamic);
      SetBoundEmpty(bBoundEmpty);
      SetBoundDynamic(bBoundDynamic);
      
      //////////////////////////////
      // START: Jordan Fix This!!!
      //

      SetGeometryEmpty(FALSE);
      SetGeometryDynamic(TRUE);
      SetNormalsEmpty(FALSE);
      SetNormalsDynamic(TRUE);
      SetBoundsEmpty(FALSE);
      SetBoundsDynamic(TRUE);
      SetBoundEmpty(FALSE);
      SetBoundDynamic(TRUE);
      
      //
      // END: Jordan Fix This!!!
      //////////////////////////////

      //////////////////////////////////////////////////
      // Calculate the Bounding Box (if Empty this will do nothing)
      if ( GetBoundEmpty() )
	{
	  SetBoundValid(TRUE);
	}
      else
	{
	  CalculateBound();
	}
      SetBoundChanged(FALSE);

      //////////////////////////////////////////////////
      // Done Preprocessing this Node in the Graph
      SetVisited(TRUE);
    }
}

VOID CSLIDEIntersection::Update()
{
  SetBoundChanged(FALSE);
  if ( GetDynamic() )
    {
      UpdateState();
    }

  if ( GetBoundsDynamic() )
    {
      CSLIDEBoundedNode::InvalidateQEMesh();
    }

  if ( GetBoundDynamic() )
    {
      SetBoundValid(FALSE);
    }

  SetVisited(FALSE);
}

VOID CSLIDEIntersection::UpdateState()
{
  CSLIDEGroup::UpdateState();
}

VOID CSLIDEIntersection::UpdateGraph()
{
  if ( !GetVisited() )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;
      BOOL bBoundChanged;

      if ( !GetBoundValid() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      if ( !m_bound.GetEmpty() )
		{
		  SetBoundChanged(TRUE);
		  m_bound.Null();
		}
	      return;
	    }
	  
	  bBoundChanged = FALSE;
	  
	  iter.Init(&m_listCSLIDEInstance);
	  
	  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
		pCSLIDEInstance != NULL;
		pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	    {
	      pCSLIDEInstance->UpdateGraph();
	      bBoundChanged = bBoundChanged || 
		pCSLIDEInstance->GetBoundChanged();
	    }
	  
	  if ( bBoundChanged )
	    {
	      CalculateBound();
	    }
	  
	  SetBoundValid(TRUE);
	}

      SetVisited(TRUE);
    }
}

VOID CSLIDEIntersection::CalculateBound()
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;


  CBound bound;
  
  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->InsertIntoBound(bound);
    }

  ///////////////////////////////////////
  // START: Jordan Fix This!!!
  //

  Point pt;
  pt.Set(-1, -1, -1);
  bound.Insert(pt);
  pt.Set(1, 1, 1);
  bound.Insert(pt);

  //
  // END: Jordan Fix This!!!
  ///////////////////////////////////////

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);
    }
}

VOID CSLIDEIntersection::CalculateCDA()
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  CDA *ppCDA[2];

  // HACK to disable CDA code
  m_bCDAVisited = TRUE;

  if ( m_bCDAVisited )
    {
      return;
    }
  m_bCDAVisited = TRUE;

  ASSERT( m_pCDA == NULL );

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->CalculateCDA();
      
      if ( m_pCDA == NULL )
	{
	  m_pCDA = pCSLIDEInstance->GetCDA();
	}
      else
	{
	  ppCDA[1] = pCSLIDEInstance->GetCDA();
	  if ( ppCDA[1] != NULL )
	    {
	      ppCDA[0] = m_pCDA;
	      m_pCDA = ppCDA[0]->Intersect(ppCDA[1]);
	      //DeallocateCDA(ppCDA[0]);
	      //DeallocateCDA(ppCDA[1]);
	    }
	}
    }

  m_pCDA->Display();
}

VOID CSLIDEIntersection::PrerenderGeometry(SLIDEShadingType eShading)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  PushShadingAttributes(eShading);
  {
    m_pCSLIDESurface->Push();
    {
      iter.Init(&m_listCSLIDEInstance);
      
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pCSLIDEInstance->PrerenderGeometry(eShading);
	}
    }
    m_pCSLIDESurface->Pop();
  }
  PopShadingAttributes(eShading);
}

VOID CSLIDEIntersection::PrerenderNormals(SLIDEShadingType eShading)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->PrerenderNormals(eShading);
    }
}

VOID CSLIDEIntersection::PrerenderBounds()
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
    {
    }
  else 
    {
      if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	{
	}
      else
	{
	  iter.Init(&m_listCSLIDEInstance);
	  
	  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
		pCSLIDEInstance != NULL;
		pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	    {
	      pCSLIDEInstance->PrerenderBounds();
	    }
	}

      m_bound.Render();
    }
}

VOID CSLIDEIntersection::Prerender()
{
  if ( !(!m_dyniLOD.GetDynamic() && 
	 ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
	   m_dyniLOD.GetValue() == slidelodtSLF_BOUND )) )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;
      
      iter.Init(&m_listCSLIDEInstance);
      
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pCSLIDEInstance->Prerender();
	}
    }
      
#ifdef INTERSECTION_TEST
  UINT_32 i;
  for ( i = 1; i < 3; i++ )
    {
      m_ppCDAPolyhedron[i]->Render(i&0x1);
    }
  //m_ppCDAPolyhedron[0]->Render();
#endif // INTERSECTION_TEST

  CSLIDEBoundedNode::Prerender();
}

VOID CSLIDEIntersection::RenderGeometry(const Matrix &mGproj_obj, 
					CBound *pCBoundClip, 
					SLIDEShadingType eShading,
					SLIDELODType eLOD,
					CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{

  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  // For Debugging
  if ( GetRenderCDAOperandsEnabled() )
    {
      gilPushAttrib(GL_ALL_ATTRIB_BITS);
      {
	PushAuxillaryGeometryAttributes();
	{
	  PushWireFrameAttributes();
	  {
	    iter.Init(&m_listCSLIDEInstance);
	    
	    for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
		  pCSLIDEInstance != NULL;
		  pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	      {
		pCSLIDEInstance->Render(mGproj_obj, pCBoundClip, slideshtSLF_WIRE, eLOD);
	      }
	  }
	  PopWireFrameAttributes();
	}
	PopAuxillaryGeometryAttributes();
      }
      gilPopAttrib();
    }
      
#ifdef INTERSECTION_TEST
  UINT_32 i;
  for ( i = 1; i < 3; i++ )
    {
      m_ppCDAPolyhedron[i]->Render(i&0x1);
    }
  //m_ppCDAPolyhedron[0]->Render();
#endif // INTERSECTION_TEST

  CalculateCDA();
  if ( m_pCDA != NULL )
    {
      m_pCDA->Render();
    }

  // While CDA's don't work
  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->Render(mGproj_obj, pCBoundClip, eShading, eLOD);
    }
}

VOID CSLIDEIntersection::RenderNormals(SLIDEShadingType /*eShading*/)
{
  if ( m_pCDA != NULL )
    {
      PushAuxillaryGeometryAttributes();
      {
	gilColor3f(1.0, 0.0, 1.0);
	
	m_pCDA->RenderNormals();
      }
      PopAuxillaryGeometryAttributes();
    }
}

VOID CSLIDEIntersection::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtIntersection], m_pcGUIID);

  WriteSLIDEFlags(fpOutputFile, 0);

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->WriteSLIDE(fpOutputFile);    
    }

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndIntersection]);
}

VOID CSLIDEIntersection::WriteRIBGeometry(FILE *fpOutputFile,
				   const Matrix &mGproj_obj, 
				   CBound *pCBoundClip, 
				   SLIDEShadingType eShading,
				   SLIDELODType eLOD)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( fpOutputFile != NULL );

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->WriteRIB(fpOutputFile,
				mGproj_obj, pCBoundClip, eShading, eLOD);    
    }
}

VOID CSLIDEIntersection::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE INTERSECTION: %s", m_pcGUIID);
}

VOID CSLIDEIntersection::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, "   END SLIDE INTERSECTION: %s", m_pcGUIID);
}

VOID CSLIDEIntersection::WritePOVGeometry(FILE *fpOutputFile, 
					  CListInt *plistTransformIDs,
					  CListInt *plistColors) 
{
  ASSERT( fpOutputFile != NULL );

  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  fprintf(fpOutputFile, "intersection { \n");

  iter.Init(&m_listCSLIDEInstance);
 
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->WritePOV(fpOutputFile, plistTransformIDs, plistColors);
    }
  fprintf(fpOutputFile, "}\n\n");
}

QEMesh *CSLIDEIntersection::WriteSIF(FILE *fpOutputFile, UINT_32 uiIndentLevel, 
				     const Matrix &mGroot_obj, const Matrix &mGobj_root,
				     QEMesh *pmeBoundaryEdges)
{
  ASSERT( fpOutputFile != NULL );
  ASSERT( pmeBoundaryEdges != NULL );

  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  QEMesh *pme;
  
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "(%s\n", gcppcSIFReservedWords[sifrwtIntersection]);
  uiIndentLevel++;

  iter.Init(&m_listCSLIDEInstance);
  
  if ( m_pmeSIF != NULL )
    {
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pme = pCSLIDEInstance->WriteSIF(fpOutputFile, uiIndentLevel, 
					  mGroot_obj, mGobj_root, pmeBoundaryEdges);
	  if ( pme != NULL )
	    {
	      DeallocateQEMesh(pme);
	    }
	}
    }
  else
    {
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pme = pCSLIDEInstance->WriteSIF(fpOutputFile, uiIndentLevel, 
					  mGroot_obj, mGobj_root, pmeBoundaryEdges);
	  if ( m_pmeSIF == NULL )
	    {
	      m_pmeSIF = pme;
	    }
	  else
	    {
	      if ( pme != NULL )
		{
		  // This will automatically destroy pme
		  m_pmeSIF->MergeMesh(pme);
		}
	    }
	}
    }

  if ( m_pmeSIF != NULL )
    {
      CObjectList<QEMesh> listQEMesh;

      listQEMesh.Init();

      pme = m_pmeSIF->Copy();
      pme->Transform(mGroot_obj, mGobj_root);
      pmeBoundaryEdges->CopyBoundaryEdges(pme);
      pme->ExtractComponents(&listQEMesh);
      DeallocateQEMesh(pme);

      for ( pme = listQEMesh.RemoveFirst();
	    pme != NULL;
	    pme = listQEMesh.RemoveFirst() )
	{
	  ::WriteSIFShell(pme, fpOutputFile, uiIndentLevel);
	  DeallocateQEMesh(pme);
	}
    }

  uiIndentLevel--;
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, ")\n");

  return NULL;
}

UINT_32 CSLIDEIntersection::GetInstanceCount()
{
  return m_listCSLIDEInstance.GetLength();
}







