#include "slidelight.h"
#include "slidecamera.h"
#include "slidesurface.h"

GLenum CSLIDELight::m_peGLLightID[cuiGLLights] =
{
  GL_LIGHT0,
  GL_LIGHT1,
  GL_LIGHT2,
  GL_LIGHT3,
  GL_LIGHT4,
  GL_LIGHT5,
  GL_LIGHT6,
  GL_LIGHT7
};
UINT_32 CSLIDELight::m_uiSLIDELights = 0;
UINT_32 CSLIDELight::m_uiGLLightsAllocated = 0;
BOOL CSLIDELight::m_bLocalViewer = FALSE;

static const VECTOR4D gcvColor = { 1.0, 1.0, 1.0, 0.0 };
static const FLOAT gcfDeadDistance = (FLOAT)0.1;
static const FLOAT gcfFallOff = 1.0;
static const FLOAT gcfAngularFallOff = 1.0;

//////////////////////////////////////////////////////////////////////
// CSLIDELight Class
//

CSLIDELight::CSLIDELight()
{
  m_dyniType.SetValue(slideltSLF_AMBIENT);
  m_vColor.Set(gcvColor);
  m_dynfDeadDistance.SetValue(gcfDeadDistance);
  m_dynfFallOff.SetValue(gcfFallOff);
  m_dynfAngularFallOff.SetValue(gcfAngularFallOff);
  /*
  m_mGlight_vrc.SetIdentity();
  m_mGvrc_light.SetIdentity();
  */
}

CSLIDELight::~CSLIDELight()
{
}

VOID CSLIDELight::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			CHAR *pcID)
{
  CSLIDEPrimitive::Init(pcFile, uiLine, uiColumn, pcID);

  m_dyniType.SetValue(slideltSLF_AMBIENT);
  m_vColor.Set(gcvColor);
  m_dynfDeadDistance.SetValue(gcfDeadDistance);
  m_dynfFallOff.SetValue(gcfFallOff);
  m_dynfAngularFallOff.SetValue(gcfAngularFallOff);
  /*
  m_mGlight_vrc.SetIdentity();
  m_mGvrc_light.SetIdentity();
  */
}

VOID CSLIDELight::Uninit()
{
  CSLIDEPrimitive::Uninit();
}

BOOL CSLIDELight::Initialized()
{
  return CSLIDEPrimitive::Initialized();
}

BOOL CSLIDELight::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  return CSLIDEPrimitive::CheckSemantics(pCSLIDEWorld);
}

BOOL CSLIDELight::Valid()
{
  if ( !CSLIDEPrimitive::Valid() )
    {
      return FALSE;
    }

  return TRUE;
}

CSLIDECamera *CSLIDELight::FindCamera(CListIterInt *piterIDs, 
				       CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtCamera],
	      gcppcSLIDEReservedWords[sliderwtLight], GetID());

  return NULL;
}

CSLIDELight *CSLIDELight::FindLight(CListIterInt *piterIDs, 
				    CListInt *plistCSLIDEInstance)
{
  CHAR *pcID;
  CSLIDELight *pCSLIDELight;

  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  pCSLIDELight = this;

  pcID = (CHAR*)piterIDs->PeekNext();
  if ( pcID == NULL )
    {
      pcID = (CHAR*)piterIDs->PeekFirst();
      if ( pcID != GetID() )
	{
	  HandleError("Bad %s path: %s id \"%s\" not specified", 
		      gcppcSLIDEReservedWords[sliderwtLight],
		      gcppcSLIDEReservedWords[sliderwtLight], GetID());
	  pCSLIDELight = NULL;
	}
    }
  else if ( pcID != GetID() )
    {
      HandleError("Bad %s path: path id \"%s\" != %s id \"%s\"", 
		  gcppcSLIDEReservedWords[sliderwtLight], pcID,
		  gcppcSLIDEReservedWords[sliderwtLight], GetID());
      pCSLIDELight = NULL;
    }

  return pCSLIDELight;
}

CSLIDEAreaLight *CSLIDELight::FindAreaLight(CListIterInt *piterIDs, 
				       CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtAreaLight],
	      gcppcSLIDEReservedWords[sliderwtLight], GetID());

  return NULL;
}

VOID CSLIDELight::ResetGLLights()
{
  UINT_32 i;
  //VECTOR4D v = {0.0, 0.0, 0.0, 1.0};
  GLfloat v[4] = {0.0, 0.0, 0.0, 1.0};

  // Assume that all light sources are either ambient or directional
  // until we encounter a point or spot light
  m_bLocalViewer = FALSE;
  gilLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_FALSE);

  // Begin with zero global ambient light
  gilLightModelfv(GL_LIGHT_MODEL_AMBIENT, v);

  for ( i = 0; i < m_uiGLLightsAllocated; i++ )
    {
      gilDisable(m_peGLLightID[i]);
    }

  m_uiGLLightsAllocated = 0;
}

VOID CSLIDELight::CheckDynamic()
{
  SetDynamic(m_dyniType.GetDynamic() ||
	     m_vColor.GetDynamic() ||
	     m_dynfDeadDistance.GetDynamic() ||
	     m_dynfFallOff.GetDynamic() ||
	     m_dynfAngularFallOff.GetDynamic());
}

VOID CSLIDELight::UpdateState()
{
  m_dyniType.Update();
  if ( !( 0 <= m_dyniType.GetValue() &&
	  m_dyniType.GetValue() < slideltSize ) )
    {
      HandleError("Illegal Light Type flag value:  %ld", 
		  m_dyniType.GetValue());
      m_dyniType.SetValue(slideltSLF_AMBIENT);
    }

  m_vColor.Update();

  m_dynfDeadDistance.Update();
  if ( !( 0.0 < m_dynfDeadDistance.GetValue() ) )
    {
      HandleError("Illegal deaddistance value:  %.3f", 
		  m_dynfDeadDistance.GetValue());
      m_dynfDeadDistance.SetValue(gcfDeadDistance);
    }

  m_dynfFallOff.Update();
  if ( !( 0.0 <= m_dynfFallOff.GetValue() ) )
    {
      HandleError("Illegal falloff value:  %.3f", 
		  m_dynfFallOff.GetValue());
      m_dynfFallOff.SetValue(gcfFallOff);
    }

  m_dynfAngularFallOff.Update();
  if ( !( 0.0 <= m_dynfAngularFallOff.GetValue() ) )
    {
      HandleError("Illegal angularfalloff value:  %.3f", 
		  m_dynfAngularFallOff.GetValue());
      m_dynfAngularFallOff.SetValue(gcfAngularFallOff);
    }
}

VOID CSLIDELight::Place()
{
  SLIDELightType eType;
  GLfloat v[4] = {0.0, 0.0, 0.0, 1.0};

  eType = (SLIDELightType)m_dyniType.GetValue();
  ASSERT( eType < slideltSize );

  //gilPushMatrix();
  //gilMultMatrixf(m_mGvrc_light.Get());

  if ( eType == slideltSLF_AMBIENT )
    {
      UINT_32 i;

      // Add to the global ambient light

      gilGetFloatv(GL_LIGHT_MODEL_AMBIENT, v);
      for ( i = 0; i < 4; i++ )
	{
	  v[i] += m_vColor[i];
	}
      gilLightModelfv(GL_LIGHT_MODEL_AMBIENT, v);
    }
  else
    {
      GLenum eLightID;
      FLOAT d0, n1;

      if ( eType != slideltSLF_DIRECTIONAL && !m_bLocalViewer )
	{
	  m_bLocalViewer = TRUE;
	  gilLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
	}

      // Allocate a GL Light Structure (there are only 8)
      if ( m_uiGLLightsAllocated >= cuiGLLights )
	{
	  HandleWarning("No more GL Lights could not enable light \"%s\"",
			GetID());
	  return;
	}

      eLightID = m_peGLLightID[m_uiGLLightsAllocated];
      m_uiGLLightsAllocated++;

      // Enable this GL Light
      gilEnable(eLightID);

      // Set all the values of the light

      v[0] = 0.0;
      v[1] = 0.0;
      v[2] = 0.0;
      v[3] = 1.0;
      gilLightfv(eLightID, GL_AMBIENT, v);
      //m_vColor.Get(v);
      v[0] = m_vColor[0];
      v[1] = m_vColor[1];
      v[2] = m_vColor[2];
      v[3] = 1.0;
      gilLightfv(eLightID, GL_DIFFUSE, v);
      gilLightfv(eLightID, GL_SPECULAR, v);

      // Distance Attenuation
      d0 = m_dynfDeadDistance.GetValue();
      n1 = m_dynfFallOff.GetValue();
      n1 = MIN(n1, (FLOAT)2.0); // OpenGL can not handle any exponents higher than this
      gilLightf(eLightID, GL_CONSTANT_ATTENUATION, (FLOAT)pow(d0, n1));
      gilLightf(eLightID, GL_LINEAR_ATTENUATION, n1*(FLOAT)pow(d0, MAX(n1 - 1.0, 0.0)));
      if ( n1 <= 1.0 )
	{
	  gilLightf(eLightID, GL_QUADRATIC_ATTENUATION, 0.0);
	}
      else
	{	  
	  //gilLightf(eLightID, GL_QUADRATIC_ATTENUATION, (n1*(n1 - 1.0)/2.0)*pow(d0, MAX(n1 - 2.0, 0.0)));
          // This simplifies to the following for all valid values of n1
	  gilLightf(eLightID, GL_QUADRATIC_ATTENUATION, (FLOAT)(n1*(n1 - 1.0)/2.0));
	}

      if ( eType == slideltSLF_DIRECTIONAL )
	{
	  v[0] = 0.0;
	  v[1] = 0.0;
	  v[2] = 1.0;
	  v[3] = 0.0;
	  gilLightfv(eLightID, GL_POSITION, v);
	  gilLightf(eLightID, GL_SPOT_EXPONENT, 0.0);
	  gilLightf(eLightID, GL_SPOT_CUTOFF, 180.0);
	}
      else if ( eType == slideltSLF_POINT )
	{
	  v[0] = 0.0;
	  v[1] = 0.0;
	  v[2] = 0.0;
	  v[3] = 1.0;
	  gilLightfv(eLightID, GL_POSITION, v);
	  gilLightf(eLightID, GL_SPOT_EXPONENT, 0.0);
	  gilLightf(eLightID, GL_SPOT_CUTOFF, 180.0);
	}
      else
	{
	  ASSERT( eType == slideltSLF_SPOT );

	  v[0] = 0.0;
	  v[1] = 0.0;
	  v[2] = 0.0;
	  v[3] = 1.0;
	  gilLightfv(eLightID, GL_POSITION, v);
	  gilLightf(eLightID, GL_SPOT_EXPONENT, m_dynfAngularFallOff.GetValue());
	  gilLightf(eLightID, GL_SPOT_CUTOFF, 90.0);
	}

      v[0] = 0.0;
      v[1] = 0.0;
      v[2] = -1.0;
      v[3] = 0.0;
      gilLightfv(eLightID, GL_SPOT_DIRECTION, v);      
    }

  //gilPopMatrix();
}

Vector CSLIDELight::CalculateColor(const Point &pt, const Vector &vNormal, 
				   CSLIDESurface *pCSLIDESurface, 
				   const Matrix &mGlight_vrc, 
				   const Matrix &mGvrc_light)
{
  // pt and vNormal are both in the light's local coordinate system
  // Remember to drop the W coord and normalize the normal

  SLIDELightType eType;
  Vector vColor, vK, vColorDiffuse, vColorLight, vColorDiffuseLight;

  ASSERT( pCSLIDESurface != NULL );

  eType = (SLIDELightType)m_dyniType.GetValue();
  ASSERT( eType < slideltSize );

  vK = pCSLIDESurface->GetReflectivity();
  vColorDiffuse = pCSLIDESurface->GetColor();
  vColorLight = m_vColor;
  vColorDiffuseLight = ComponentMultiply(vColorDiffuse, vColorLight); 

  if ( eType == slideltSLF_AMBIENT )
    {
      vColor = vK[0]*vColorDiffuseLight;
    }
  else
    {
      FLOAT fSpec, fNphong;
      Point ptOrigin, ptLocal, ptEye;
      Vector vColorSpec, vN, vV, vL, vR;

      fSpec = pCSLIDESurface->GetMetallic();
      fNphong = pCSLIDESurface->GetExponent();
      vColorSpec = (fSpec*vColorDiffuseLight) + (((FLOAT)1.0 - fSpec)*vColorLight);

      ptLocal = mGlight_vrc * pt;
      ASSERT( m_pCSLIDECameraCurr != NULL );
      if ( m_pCSLIDECameraCurr->GetProjection() == slidepjtSLF_PERSPECTIVE )
	{
	  ptEye = mGlight_vrc * ptOrigin;
	  vV = ptEye - ptLocal;
	}
      else
	{
	  Vector vZ;

	  vZ.Set(0.0, 0.0, 1.0);
	  vV = mGlight_vrc * vZ;
	}

      vN = vNormal * mGvrc_light;
      vN[W] = 0.0;
      vN.Normalize();

      if ( eType == slideltSLF_DIRECTIONAL )
	{
	  vL.Set(0.0, 0.0, 1.0);
	  vR = ((FLOAT)2.0*(vN.Dot(vL)))*vN - vL;
	  
	  vColor = (vK[1]*(vN.Dot(vL)))*vColorDiffuseLight + 
	           (vK[2]*(FLOAT)pow(MAX(vR.Dot(vV), 0.0), fNphong))*vColorSpec;
	}
      else
	{
	  FLOAT fDistance, fDeadDistance, fFallOff;

	  fDeadDistance = m_dynfDeadDistance.GetValue();
	  fFallOff = m_dynfFallOff.GetValue();

	  vL = ptOrigin - ptLocal;
	  fDistance = vL.Magnitude();
	  vL.Normalize();

	  vR = ((FLOAT)(2.0*(vN.Dot(vL)))*vN) - vL;
	  vR.Normalize();

	  vColor = (vK[1]*(vN.Dot(vL)))*vColorDiffuseLight + 
	           (vK[2]*(FLOAT)pow(MAX(vR.Dot(vV), 0.0), fNphong))*vColorSpec;

	  /*
	  printf("***** Start: Calculate Point Light Color *****\n");
	  printf("V:\n");
	  vV.Display();
	  printf("N:\n");
	  vN.Display();
	  printf("L:\n");
	  vL.Display();
	  printf("R:\n");
	  vR.Display();
	  printf("|L| == %f\n", fDistance);
	  printf("L * N == %f\n", vN.Dot(vL));
	  printf("R * V == %f\n", vR.Dot(vV));
	  printf("Color:\n");
	  vColor.Display();
	  printf("***** End: Calculate Point Light Color *****\n");
	  */

	  if ( eType == slideltSLF_POINT )
	    {
	      vColor = vColor / (FLOAT)pow(fDeadDistance + fDistance, fFallOff);
	    }
	  else
	    {
	      ASSERT( eType == slideltSLF_SPOT );

	      Vector vD;
	      FLOAT fAngularFallOff;

	      fAngularFallOff = m_dynfAngularFallOff.GetValue();

	      vD.Set(0.0, 0.0, -1.0);

	      vColor = vColor * ((FLOAT)pow(-(vD.Dot(vL)), fAngularFallOff) / 
				 (FLOAT)pow(fDeadDistance + fDistance, fFallOff));
	    }
	}
    }

  return vColor;
}

VOID CSLIDELight::WriteSLIDE(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );
  
  fprintf(fpOutputFile, "%s %s\n",
	  gcppcSLIDEReservedWords[sliderwtLight], m_pcGUIID);
  fprintf(fpOutputFile, "  %s %s\n",
	  gcppcSLIDEReservedWords[sliderwtType], 
	  gcppcSLIDELightReservedWords[m_dyniType.GetValue()] );
  fprintf(fpOutputFile, "  %s ( %.3f %.3f %.3f )\n",
	  gcppcSLIDEReservedWords[sliderwtColor],
	  m_vColor[X], m_vColor[Y], m_vColor[Z]);
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtDeadDistance],
	  m_dynfDeadDistance.GetValue());
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtFallOff],
	  m_dynfFallOff.GetValue());
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtAngularFallOff],
	  m_dynfAngularFallOff.GetValue()); 
  CSLIDEPrimitive::WriteSLIDEFlags(fpOutputFile, 0);

  fprintf(fpOutputFile, "%s\n",
	  gcppcSLIDEReservedWords[sliderwtEndLight]);
}

VOID CSLIDELight::WriteRIB(FILE *fpOutputFile, const Matrix &mGvrc_light)
{
  FLOAT fIntensity;
  RtColor rtColor;
  RtPoint rtFrom, rtTo;

  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE LIGHT: %s", m_pcGUIID);
  RiTransformBegin();
  {
    WriteRIBMultMatrix(fpOutputFile, mGvrc_light);

    fIntensity = 1.0;
    rtColor[0] = m_vColor[0];
    rtColor[1] = m_vColor[1];
    rtColor[2] = m_vColor[2];

    rtFrom[0] = 0.0;
    rtFrom[1] = 0.0;
    rtFrom[2] = 0.0;

    rtTo[0] = 0.0;
    rtTo[1] = 0.0;
    rtTo[2] = -1.0;

    switch ( m_dyniType.GetValue() )
      {
      case slideltSLF_AMBIENT:
	{
	  RiLightSource(RI_AMBIENTLIGHT,
			RI_INTENSITY, &fIntensity,
			RI_LIGHTCOLOR, rtColor,
			RI_NULL);
	  break;
	}
      case slideltSLF_DIRECTIONAL:
	{
	  RiLightSource(RI_DISTANTLIGHT,
			RI_INTENSITY, &fIntensity,
			RI_LIGHTCOLOR, rtColor,
			RI_FROM, rtFrom,
			RI_TO, rtTo,
			RI_NULL);
	  break;
	}
      case slideltSLF_POINT:
	{
	  RiLightSource(RI_POINTLIGHT,
			RI_INTENSITY, &fIntensity,
			RI_LIGHTCOLOR, rtColor,
			RI_FROM, rtFrom,
			RI_NULL);
	  break;
	}
      case slideltSLF_SPOT:
	{
	  FLOAT fConeAngle, fConeDeltaAngle, fAngularFallOff;

	  fConeAngle = 180.0;
	  fConeDeltaAngle = 0.0;
	  fAngularFallOff = m_dynfAngularFallOff.GetValue();

	  RiLightSource(RI_SPOTLIGHT,
			RI_INTENSITY, &fIntensity,
			RI_LIGHTCOLOR, rtColor,
			RI_FROM, rtFrom,
			RI_TO, rtTo,
			RI_CONEANGLE, &fConeAngle,
			RI_CONEDELTAANGLE, &fConeDeltaAngle,
			RI_BEAMDISTRIBUTION, &fAngularFallOff, 
			RI_NULL);
	  break;
	}
      default:
	{
	  ASSERT( FALSE );
	  break;
	}
      }
  }
  RiTransformEnd();
  RiArchiveRecord(RI_COMMENT, "   END SLIDE LIGHT: %s", m_pcGUIID);
}

VOID CSLIDELight::WritePOVLight(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  switch ( m_dyniType.GetValue() )
    {
    case slideltSLF_AMBIENT:
      {
	fprintf(fpOutputFile, "global_settings { ambient_light ");
	fprintf(fpOutputFile, "color rgb < %.3f, %.3f, %.3f > ", 
		m_vColor[0], m_vColor[1], m_vColor[2]);
	break;
      }
    case slideltSLF_DIRECTIONAL:
      {
	fprintf(fpOutputFile, "light_source { \n");
	fprintf(fpOutputFile, " < 0, 0, 0 > \n");
	fprintf(fpOutputFile, " color rgb < %.3f, %.3f, %.3f > \n",
		m_vColor[0], m_vColor[1], m_vColor[2]);
	break;
      }
    case slideltSLF_POINT:
      {
	fprintf(fpOutputFile, "light_source %s{ \n", GetID());
	fprintf(fpOutputFile, " < 0, 0, 0 >");
	fprintf(fpOutputFile, " color rgb < %.3f, %.3f, %.3f > \n",
		m_vColor[0], m_vColor[1], m_vColor[2]);
	break;
      }
    case slideltSLF_SPOT:
      {
	FLOAT fConeAngle, fConeDeltaAngle, fAngularFallOff;
	
	fConeAngle = 180.0;
	fConeDeltaAngle = 0.0;
	fAngularFallOff = m_dynfAngularFallOff.GetValue();
	
	fprintf(fpOutputFile, "light_source { \n");
	fprintf(fpOutputFile, " spotlight \n");
	fprintf(fpOutputFile, " < 0, 0, 0 > \n");
	fprintf(fpOutputFile, " color rgb <%.3f, %.3f, %.3f > \n",
		m_vColor[0], m_vColor[1], m_vColor[2]);
	fprintf(fpOutputFile, " radius %.3f", fConeAngle);
	fprintf(fpOutputFile, " falloff %.3f", fAngularFallOff);
	fprintf(fpOutputFile, " tightness 10.000");
	fprintf(fpOutputFile, " point_at < 0, 0, -1 > \n");
	/*
	RiLightSource(RI_SPOTLIGHT,
		      RI_INTENSITY, &fIntensity,
		      RI_LIGHTCOLOR, rtColor,
		      RI_FROM, rtFrom,
		      RI_TO, rtTo,
		      RI_CONEANGLE, &fConeAngle,
		      RI_CONEDELTAANGLE, &fConeDeltaAngle,
		      RI_BEAMDISTRIBUTION, &fAngularFallOff, 
		      RI_NULL);*/
	break;
      }
    default:
      {
	ASSERT( FALSE );
	break;
      }
    }
  fprintf(fpOutputFile, "}\n\n");
}
VOID CSLIDELight::WritePOV(FILE *fpOutputFile, 
			   CListInt *plistTransformIDs,
			   CListInt *plistColors) 
{
  ASSERT( fpOutputFile != NULL );

  CListIterInt iter;
  CHAR *pcTransformID;

  switch ( m_dyniType.GetValue() )
    {
    case slideltSLF_AMBIENT:
      {
	fprintf(fpOutputFile, "global_settings { ambient_light ");
	fprintf(fpOutputFile, "color rgb < %.3f, %.3f, %.3f > ", 
		m_vColor[0], m_vColor[1], m_vColor[2]);
	break;
      }
    case slideltSLF_DIRECTIONAL:
      {
	fprintf(fpOutputFile, "light_source { \n");
	fprintf(fpOutputFile, " < 0, 0, 0 > \n");
	fprintf(fpOutputFile, " color rgb < %.3f, %.3f, %.3f > \n",
		m_vColor[0], m_vColor[1], m_vColor[2]);
	break;
      }
    case slideltSLF_POINT:
      {
	fprintf(fpOutputFile, "light_source %s{ \n", GetID());
	fprintf(fpOutputFile, " < 0, 0, 0 > \n");
	fprintf(fpOutputFile, " color rgb < %.3f, %.3f, %.3f > \n",
		m_vColor[0], m_vColor[1], m_vColor[2]);
	break;
      }
    case slideltSLF_SPOT:
      {
	FLOAT fConeAngle, fConeDeltaAngle, fAngularFallOff;
	
	fConeAngle = 180.0;
	fConeDeltaAngle = 0.0;
	fAngularFallOff = m_dynfAngularFallOff.GetValue();
	
	fprintf(fpOutputFile, "light_source { \n");
	fprintf(fpOutputFile, " spotlight \n");
	fprintf(fpOutputFile, " < 0, 0, 0 > \n");
	fprintf(fpOutputFile, " color rgb <%.3f, %.3f, %.3f > \n",
		m_vColor[0], m_vColor[1], m_vColor[2]);
	fprintf(fpOutputFile, " radius %.3f", fConeAngle);
	fprintf(fpOutputFile, " falloff %.3f", fAngularFallOff);
	fprintf(fpOutputFile, " tightness 10.000");
	fprintf(fpOutputFile, " point_at < 0, 0, -1 > \n");
	break;
      }
    default:
      {
	ASSERT( FALSE );
	break;
      }
    }
  iter.Init(plistTransformIDs);
  for ( pcTransformID = (CHAR*)iter.PeekFirst(); 
	pcTransformID != NULL;
	pcTransformID = (CHAR*)iter.PeekNext() )
    {
      fprintf(fpOutputFile, " transform %s\n", pcTransformID);
    }
  fprintf(fpOutputFile, "}\n\n");
}


VOID CSLIDELight::WriteDotID(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "light_%s", GetGUIID());
}

VOID CSLIDELight::WriteDotNode(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "  ");
  WriteDotID(fpOutputFile);
  fprintf(fpOutputFile, " [ label = \"%s\", shape = diamond ];\n", GetGUIID());
}

VOID CSLIDELight::WriteDotEdges(FILE * /*fpOutputFile*/)
{
}

//////////////////////////////////////////////////////////////////////
// CSLIDELightPath Class
//

CSLIDELightPath::CSLIDELightPath()
{
  m_mGlight_vrc.SetIdentity();
  m_mGvrc_light.SetIdentity();
}

CSLIDELightPath::~CSLIDELightPath()
{
}

VOID CSLIDELightPath::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			   CListInt *plistPathIDs, DynInt *pdyniOnOff)
{
  CSLIDEPath::Init(pcFile, uiLine, uiColumn, plistPathIDs);

  m_dyniOnOff.SetValue(slideoftSLF_ON);
  if ( pdyniOnOff != NULL )
    {
      m_dyniOnOff.Set(pdyniOnOff);
    }

  m_mGlight_vrc.SetIdentity();
  m_mGvrc_light.SetIdentity();
}

VOID CSLIDELightPath::Uninit()
{
  CSLIDEPath::Uninit();
}

CSLIDENode *CSLIDELightPath::FindNode(CSLIDENode *pRoot, CListIterInt *piterIDs, 
				       CListInt *plistCSLIDEInstance)
{
  ASSERT( pRoot != NULL );
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  return pRoot->FindLight(piterIDs, plistCSLIDEInstance);
}

VOID CSLIDELightPath::CheckDynamic()
{
  CSLIDEPath::CheckDynamic();

  SetDynamic(GetDynamic() || m_dyniOnOff.GetDynamic());
}

VOID CSLIDELightPath::UpdateState()
{
  CSLIDEPath::UpdateState();

  m_dyniOnOff.Update();
  if ( !( 0 <= m_dyniOnOff.GetValue() &&
	  m_dyniOnOff.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal OnOff flag value:  %ld", 
		  m_dyniOnOff.GetValue());
      m_dyniOnOff.SetValue(slideoftSLF_ON);
    }
}
