#ifndef __SLIDELIGHT_H__
#define __SLIDELIGHT_H__

#include "slidenode.h"
#include "slidepath.h"

//////////////////////////////////////////////////////////////////////
// CSLIDELight Class
//

const UINT_32 cuiGLLights = 8;

class CSLIDELight : public CSLIDENode
{
//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDELight();
  virtual ~CSLIDELight();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  inline VOID SetType(DynInt *pdyniType);
  inline VOID SetColor(DynVector *pdynvColor);
  inline VOID SetDeadDistance(DynFloat *pdynfDeadDistance);
  inline VOID SetFallOff(DynFloat *pdynfFallOff);
  inline VOID SetAngularFallOff(DynFloat *pdynfAngularFallOff);

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual CSLIDECamera *FindCamera(CListIterInt *piterIDs, 
				   CListInt *plistCSLIDEInstance);
  virtual CSLIDELight *FindLight(CListIterInt *piterIDs, 
				 CListInt *plistCSLIDEInstance);
  virtual CSLIDEAreaLight *FindAreaLight(CListIterInt *piterIDs, 
				 CListInt *plistCSLIDEInstance);

  static VOID ResetGLLights();

protected:
  VOID CheckDynamic();
  VOID UpdateState();

//////////////////////////////////////////////////
// Rendering
//
public:
  Vector CalculateColor(const Point &pt, const Vector &vNormal, 
			CSLIDESurface *pCSLIDESurface,
			const Matrix &mGlight_vrc, 
			const Matrix &mGvrc_light);
  VOID Place();

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);
  virtual VOID WriteRIB(FILE *fpOutputFile, const Matrix &mGvrc_light);
  virtual VOID WritePOVLight(FILE *fpOutputFile);
  virtual VOID WritePOV(FILE *fpOutputFile, 
			CListInt *plistTransformIDs,
			CListInt *plistColors); 
  virtual VOID WriteDotID(FILE *fpOutputFile);
  virtual VOID WriteDotNode(FILE *fpOutputFile);
  virtual VOID WriteDotEdges(FILE *fpOutputFile);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline CHAR *GetTypeString();
  virtual inline SLIDEReservedWordType GetType();

/*
  inline const Matrix &GetVRCToLight() const;
  inline Matrix &SetVRCToLight();
  inline const Matrix &GetLightToVRC() const;
  inline Matrix &SetLightToVRC();
  */

//////////////////////////////////////////////////
// Class Variables
//
public:
  // OpenGL has 8 complicated light objects
  // We can pack a SLIDE Ambient light and another SLIDE Light into a 
  // single OpenGL light.
  //
  // We also have to make considerations 
  static GLenum m_peGLLightID[cuiGLLights];

  static UINT_32 m_uiSLIDELights;
  static UINT_32 m_uiGLLightsAllocated;
  static BOOL m_bLocalViewer;

//////////////////////////////////////////////////
// Member Variables
//
protected:
  DynInt m_dyniType;
  DynVector m_vColor;
  DynFloat m_dynfDeadDistance;
  DynFloat m_dynfFallOff;
  DynFloat m_dynfAngularFallOff;
  /*
  Matrix m_mGlight_vrc; // VRC to Light coordinates
  Matrix m_mGvrc_light; // Light to VRC coordinates
  */
};

//////////////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDELight::SetType(DynInt *pdyniType)
{
  m_dyniType.Set(pdyniType);
}

inline VOID CSLIDELight::SetColor(DynVector *pdynvColor)
{
  m_vColor.SetAll(pdynvColor);
}

inline VOID CSLIDELight::SetDeadDistance(DynFloat *pdynfDeadDistance)
{
  m_dynfDeadDistance.Set(pdynfDeadDistance);
}

inline VOID CSLIDELight::SetFallOff(DynFloat *pdynfFallOff)
{
  m_dynfFallOff.Set(pdynfFallOff);
}

inline VOID CSLIDELight::SetAngularFallOff(DynFloat *pdynfAngularFallOff)
{
  m_dynfAngularFallOff.Set(pdynfAngularFallOff);
}
/*
inline const Matrix &CSLIDELight::GetVRCToLight() const
{
  return m_mGlight_vrc;
}

inline Matrix &CSLIDELight::SetVRCToLight()
{
  return m_mGlight_vrc;
}

inline const Matrix &CSLIDELight::GetLightToVRC() const
{
  return m_mGvrc_light;
}

inline Matrix &CSLIDELight::SetLightToVRC()
{
  return m_mGvrc_light;
}
*/

inline CHAR *CSLIDELight::GetTypeString()
{
  return gcppcSLIDEReservedWords[sliderwtLight];
}

inline SLIDEReservedWordType CSLIDELight::GetType()
{
  return sliderwtLight;
}

//////////////////////////////////////////////////////////////////////
// CSLIDELightPath Class
//

class CSLIDELightPath : public CSLIDEPath
{
//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDELightPath();
  virtual ~CSLIDELightPath();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CListInt *plistPathIDs, DynInt *pdyniOnOff);
  virtual VOID Uninit();

  virtual CSLIDENode *FindNode(CSLIDENode *pRoot, CListIterInt *piterIDs, 
			       CListInt *plistCSLIDEInstance);

//////////////////////////////////////////////////
// Geometric Computations
//
public:

protected:
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  inline CSLIDELight *GetLight();

  inline SLIDEOnOffType GetOnOff();

  inline const Matrix &GetVRCToLight() const;
  inline Matrix &SetVRCToLight();
  inline const Matrix &GetLightToVRC() const;
  inline Matrix &SetLightToVRC();

//////////////////////////////////////////////////
// Member Variables
//
protected:
  DynInt m_dyniOnOff;

  Matrix m_mGlight_vrc; // VRC to Light coordinates
  Matrix m_mGvrc_light; // Light to VRC coordinates
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline CSLIDELight *CSLIDELightPath::GetLight()
{
  return (CSLIDELight*)m_pNode;
}

inline SLIDEOnOffType CSLIDELightPath::GetOnOff()
{
  return (SLIDEOnOffType)m_dyniOnOff.GetValue();
}

inline const Matrix &CSLIDELightPath::GetVRCToLight() const
{
  return m_mGlight_vrc;
}

inline Matrix &CSLIDELightPath::SetVRCToLight()
{
  return m_mGlight_vrc;
}

inline const Matrix &CSLIDELightPath::GetLightToVRC() const
{
  return m_mGvrc_light;
}

inline Matrix &CSLIDELightPath::SetLightToVRC()
{
  return m_mGvrc_light;
}

#endif // __SLIDELIGHT_H__


