#include "slidemesh.h"

#include "slideworld.h"
#include "slidesurface.h"

#include "slidemeshvertex.h"
#include "slidemeshvertexproperty.h"
#include "slidemeshfacet.h"
#include "slidemeshedge.h"
#include "slidemeshedgeproperty.h"

#include "qemesh.h"
#include "qerender.h"
#include "qewrite.h"

//////////////////////////////////////////////////////////////////////
// CSLIDEMesh Class
//

CSLIDEMesh::CSLIDEMesh()
{
}

CSLIDEMesh::~CSLIDEMesh()
{
}

VOID CSLIDEMesh::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		      CHAR *pcID)
{
  CSLIDEAtomNode::Init(pcFile, uiLine, uiColumn, pcID);

  m_listCSLIDEMeshVertexProperty.Init();
  m_hashCSLIDEMeshVertexProperty.Init();
  m_listCSLIDEMeshEdgeProperty.Init();
  m_hashCSLIDEMeshEdgeProperty.Init();
  m_listCSLIDEMeshVertex.Init();
  m_hashCSLIDEMeshVertex.Init();
  m_listCSLIDEMeshFacet.Init();
  m_hashCSLIDEMeshFacet.Init();
  m_listCSLIDEMeshEdge.Init();
  m_hashCSLIDEMeshEdge.Init();

  m_listDynamics.Init();

  // Default Edge Property
  ASSERT( CSLIDEMeshVertexProperty::GetSLF_INHERIT() != NULL );
  InsertMeshVertexProperty(CSLIDEMeshVertexProperty::GetSLF_INHERIT());
  ASSERT( CSLIDEMeshEdgeProperty::GetSLF_INHERIT() != NULL );
  InsertMeshEdgeProperty(CSLIDEMeshEdgeProperty::GetSLF_INHERIT());
}

VOID CSLIDEMesh::Uninit()
{
  CSLIDEMeshVertex *pCSLIDEMeshVertex;
  CSLIDEMeshFacet *pCSLIDEMeshFacet;
  CSLIDEMeshEdge *pCSLIDEMeshEdge;
  CSLIDEMeshVertexProperty *pCSLIDEMeshVertexProperty;
  CSLIDEMeshEdgeProperty *pCSLIDEMeshEdgeProperty;

  CSLIDEAtomNode::Uninit();
  
  for ( pCSLIDEMeshVertexProperty = (CSLIDEMeshVertexProperty*)m_listCSLIDEMeshVertexProperty.RemoveFirst();
	pCSLIDEMeshVertexProperty != NULL;
	pCSLIDEMeshVertexProperty = (CSLIDEMeshVertexProperty*)m_listCSLIDEMeshVertexProperty.RemoveFirst() )
    {
      if ( pCSLIDEMeshVertexProperty != CSLIDEMeshVertexProperty::GetSLF_INHERIT() )
	{
	  delete pCSLIDEMeshVertexProperty;
	}
    }

  for ( pCSLIDEMeshEdgeProperty = (CSLIDEMeshEdgeProperty*)m_listCSLIDEMeshEdgeProperty.RemoveFirst();
	pCSLIDEMeshEdgeProperty != NULL;
	pCSLIDEMeshEdgeProperty = (CSLIDEMeshEdgeProperty*)m_listCSLIDEMeshEdgeProperty.RemoveFirst() )
    {
      if ( pCSLIDEMeshEdgeProperty != CSLIDEMeshEdgeProperty::GetSLF_INHERIT() )
	{
	  delete pCSLIDEMeshEdgeProperty;
	}
    }

  for ( pCSLIDEMeshVertex = (CSLIDEMeshVertex*)m_listCSLIDEMeshVertex.RemoveFirst();
	pCSLIDEMeshVertex != NULL;
	pCSLIDEMeshVertex = (CSLIDEMeshVertex*)m_listCSLIDEMeshVertex.RemoveFirst() )
    {
      delete pCSLIDEMeshVertex;
    }

  for ( pCSLIDEMeshFacet = (CSLIDEMeshFacet*)m_listCSLIDEMeshFacet.RemoveFirst();
	pCSLIDEMeshFacet != NULL;
	pCSLIDEMeshFacet = (CSLIDEMeshFacet*)m_listCSLIDEMeshFacet.RemoveFirst() )
    {
      delete pCSLIDEMeshFacet;
    }

  for ( pCSLIDEMeshEdge = (CSLIDEMeshEdge*)m_listCSLIDEMeshEdge.RemoveFirst();
	pCSLIDEMeshEdge != NULL;
	pCSLIDEMeshEdge = (CSLIDEMeshEdge*)m_listCSLIDEMeshEdge.RemoveFirst() )
    {
      delete pCSLIDEMeshEdge;
    }
}

BOOL CSLIDEMesh::Initialized()
{
  if ( !CSLIDEAtomNode::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

BOOL CSLIDEMesh::InsertMeshVertexProperty(CSLIDEMeshVertexProperty *pCSLIDEMeshVertexProperty)
{
  CHAR *pcID;

  ASSERT( pCSLIDEMeshVertexProperty != NULL );

  pcID = pCSLIDEMeshVertexProperty->GetID();

  m_hashCSLIDEMeshVertexProperty.Insert(pCSLIDEMeshVertexProperty, (UINT_32)pcID);
  m_listCSLIDEMeshVertexProperty.InsertLast(pCSLIDEMeshVertexProperty, (UINT_32)pCSLIDEMeshVertexProperty);

  return TRUE;  
}

CSLIDEMeshVertexProperty *CSLIDEMesh::PeekMeshVertexProperty(CHAR *pcID)
{
  ASSERT( pcID != NULL );

  return (CSLIDEMeshVertexProperty*)(m_hashCSLIDEMeshVertexProperty.Peek((UINT_32)pcID));
}

BOOL CSLIDEMesh::InsertMeshEdgeProperty(CSLIDEMeshEdgeProperty *pCSLIDEMeshEdgeProperty)
{
  CHAR *pcID;

  ASSERT( pCSLIDEMeshEdgeProperty != NULL );

  pcID = pCSLIDEMeshEdgeProperty->GetID();

  m_hashCSLIDEMeshEdgeProperty.Insert(pCSLIDEMeshEdgeProperty, (UINT_32)pcID);
  m_listCSLIDEMeshEdgeProperty.InsertLast(pCSLIDEMeshEdgeProperty, (UINT_32)pCSLIDEMeshEdgeProperty);

  return TRUE;  
}

CSLIDEMeshEdgeProperty *CSLIDEMesh::PeekMeshEdgeProperty(CHAR *pcID)
{
  ASSERT( pcID != NULL );

  return (CSLIDEMeshEdgeProperty*)(m_hashCSLIDEMeshEdgeProperty.Peek((UINT_32)pcID));
}

BOOL CSLIDEMesh::InsertMeshVertex(CSLIDEMeshVertex *pCSLIDEMeshVertex)
{
  CHAR *pcID;

  ASSERT( pCSLIDEMeshVertex != NULL );

  pcID = pCSLIDEMeshVertex->GetID();

  m_hashCSLIDEMeshVertex.Insert(pCSLIDEMeshVertex, (UINT_32)pcID);
  m_listCSLIDEMeshVertex.InsertLast(pCSLIDEMeshVertex, (UINT_32)pCSLIDEMeshVertex);

  return TRUE;  
}

CSLIDEMeshVertex *CSLIDEMesh::PeekMeshVertex(CHAR *pcID)
{
  ASSERT( pcID != NULL );

  return (CSLIDEMeshVertex*)(m_hashCSLIDEMeshVertex.Peek((UINT_32)pcID));
}

BOOL CSLIDEMesh::InsertMeshFacet(CSLIDEMeshFacet *pCSLIDEMeshFacet)
{
  CHAR *pcID;

  ASSERT( pCSLIDEMeshFacet != NULL );

  pcID = pCSLIDEMeshFacet->GetID();

  m_hashCSLIDEMeshFacet.Insert(pCSLIDEMeshFacet, (UINT_32)pcID);
  m_listCSLIDEMeshFacet.InsertLast(pCSLIDEMeshFacet, (UINT_32)pCSLIDEMeshFacet);

  return TRUE;  
}

CSLIDEMeshFacet *CSLIDEMesh::PeekMeshFacet(CHAR *pcID)
{
  ASSERT( pcID != NULL );

  return (CSLIDEMeshFacet*)(m_hashCSLIDEMeshFacet.Peek((UINT_32)pcID));
}

BOOL CSLIDEMesh::InsertMeshEdge(CSLIDEMeshEdge *pCSLIDEMeshEdge)
{
  CHAR *pcID;

  ASSERT( pCSLIDEMeshEdge != NULL );

  pcID = pCSLIDEMeshEdge->GetID();

  m_hashCSLIDEMeshEdge.Insert(pCSLIDEMeshEdge, (UINT_32)pcID);
  m_listCSLIDEMeshEdge.InsertLast(pCSLIDEMeshEdge, (UINT_32)pCSLIDEMeshEdge);

  return TRUE;  
}

CSLIDEMeshEdge *CSLIDEMesh::PeekMeshEdge(CHAR *pcID)
{
  ASSERT( pcID != NULL );

  return (CSLIDEMeshEdge*)(m_hashCSLIDEMeshEdge.Peek((UINT_32)pcID));
}

BOOL CSLIDEMesh::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL b, bCorrect, bConnectivity;
  CListIterInt iter;
  QEMeshIter iterQEMesh;
  QEVertexIter iterQEVertex;
  CSLIDEMeshVertex *pCSLIDEMeshVertex;
  CSLIDEMeshFacet *pCSLIDEMeshFacet;
  CSLIDEMeshEdge *pCSLIDEMeshEdge;
  CSLIDEMeshVertexProperty *pCSLIDEMeshVertexProperty;
  CSLIDEMeshEdgeProperty *pCSLIDEMeshEdgeProperty;
  FLOAT fSharpness;
  QEEdge *pe, *peCurr;
  QEVertex *pvx, *pvxCurr;
  BOOL bVertexMatch;
  UINT_32 uiVertexMultiple, i;
  CSLIDEMeshVertex *psv;

  if ( !CSLIDEAtomNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  ASSERT( m_pQEMesh == NULL );
  
  m_pQEMesh = AllocateQEMesh();
  ASSERT( m_pQEMesh != NULL );
  m_pQEMesh->Init();

  bCorrect = TRUE;
  bConnectivity = TRUE;

  iter.Init(&m_listCSLIDEMeshVertexProperty);
  for ( pCSLIDEMeshVertexProperty = (CSLIDEMeshVertexProperty*)(iter.PeekFirst());
	pCSLIDEMeshVertexProperty != NULL;
	pCSLIDEMeshVertexProperty = (CSLIDEMeshVertexProperty*)(iter.PeekNext()) )
    {
      b = pCSLIDEMeshVertexProperty->CheckSemantics(pCSLIDEWorld);
      bCorrect = bCorrect && b;
    }

  iter.Init(&m_listCSLIDEMeshEdgeProperty);
  for ( pCSLIDEMeshEdgeProperty = (CSLIDEMeshEdgeProperty*)(iter.PeekFirst());
	pCSLIDEMeshEdgeProperty != NULL;
	pCSLIDEMeshEdgeProperty = (CSLIDEMeshEdgeProperty*)(iter.PeekNext()) )
    {
      b = pCSLIDEMeshEdgeProperty->CheckSemantics(pCSLIDEWorld);
      bCorrect = bCorrect && b;
    }

  iter.Init(&m_listCSLIDEMeshVertex);
  for ( pCSLIDEMeshVertex = (CSLIDEMeshVertex*)(iter.PeekFirst());
	pCSLIDEMeshVertex != NULL;
	pCSLIDEMeshVertex = (CSLIDEMeshVertex*)(iter.PeekNext()) )
    {
      b = pCSLIDEMeshVertex->CheckSemantics(pCSLIDEWorld, this, m_pQEMesh);
      bCorrect = bCorrect && b;
    }

  iter.Init(&m_listCSLIDEMeshFacet);
  for ( pCSLIDEMeshFacet = (CSLIDEMeshFacet*)(iter.PeekFirst());
	pCSLIDEMeshFacet != NULL;
	pCSLIDEMeshFacet = (CSLIDEMeshFacet*)(iter.PeekNext()) )
    {
      b = pCSLIDEMeshFacet->CheckSemantics(pCSLIDEWorld, this, m_pQEMesh);
      bCorrect = bCorrect && b;
    }

  iter.Init(&m_listCSLIDEMeshVertex);
  for ( pCSLIDEMeshVertex = (CSLIDEMeshVertex*)(iter.PeekFirst());
	pCSLIDEMeshVertex != NULL;
	pCSLIDEMeshVertex = (CSLIDEMeshVertex*)(iter.PeekNext()) )
    {
      b = pCSLIDEMeshVertex->CheckValence();
      bConnectivity = bConnectivity && b;
    }

  fSharpness = CSLIDEMeshEdgeProperty::GetSLF_INHERIT()->GetSharpness();
  iterQEMesh.Init(m_pQEMesh);
  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      pe->SetKnotInterval(fSharpness);

      // Check if the edge is in the ring of edges around the vertex
      bVertexMatch = FALSE;
      uiVertexMultiple = 0;
      iterQEVertex.Init(pe->GetVertex());
      pvx = pe->GetEdge()->GetVertex();
      for ( peCurr = iterQEVertex.PeekFirstEdge();
	    peCurr != NULL;
	    peCurr = iterQEVertex.PeekNextEdge() )
	{
	  if ( pe == peCurr )
	    {
	      bVertexMatch = TRUE;
	    }
	  pvxCurr = peCurr->GetEdge()->GetVertex();
	  if ( pvx == pvxCurr ) 
	    {
	      uiVertexMultiple++;
	    }
	}
      if ( !bVertexMatch )
	{
	  pCSLIDEMeshVertex = (CSLIDEMeshVertex*)((VOID*)(pe->GetVertex()->GetForwardFacet()));
	  ASSERT( pCSLIDEMeshVertex != NULL );
	  pCSLIDEMeshVertex->HandleError("vertex \"%s\" has multiple edge rings", 
					 pCSLIDEMeshVertex->GetID());
	  bConnectivity = FALSE;
	}
      if ( uiVertexMultiple > 1 )
	{
	  pCSLIDEMeshVertex = (CSLIDEMeshVertex*)((VOID*)(pe->GetVertex()->GetForwardFacet()));
	  psv = (CSLIDEMeshVertex*)((VOID*)(pvx->GetForwardFacet()));
	  ASSERT( pCSLIDEMeshVertex != NULL );
	  pCSLIDEMeshVertex->HandleError("vertex \"%s\" has the same vertex neighbor %ld times: vertex \"%s\"", 
					 pCSLIDEMeshVertex->GetID(),
					 uiVertexMultiple, psv->GetID());
	  bConnectivity = FALSE;
	}
    }
  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdgeBoundary() )
    {
      pe->SetKnotInterval(fSharpness);

      // Check if the edge is in the ring of edges around the vertex
      bVertexMatch = FALSE;
      uiVertexMultiple = 0;
      iterQEVertex.Init(pe->GetVertex());
      pvx = pe->GetEdge()->GetVertex();
      for ( peCurr = iterQEVertex.PeekFirstEdge();
	    peCurr != NULL;
	    peCurr = iterQEVertex.PeekNextEdge() )
	{
	  if ( pe == peCurr )
	    {
	      bVertexMatch = TRUE;
	    }
	  pvxCurr = peCurr->GetEdge()->GetVertex();
	  if ( pvx == pvxCurr ) 
	    {
	      uiVertexMultiple++;
	    }
	}
      if ( !bVertexMatch )
	{
	  pCSLIDEMeshVertex = (CSLIDEMeshVertex*)((VOID*)(pe->GetVertex()->GetForwardFacet()));
	  ASSERT( pCSLIDEMeshVertex != NULL );
	  pCSLIDEMeshVertex->HandleError("vertex \"%s\" has multiple edge rings", 
					 pCSLIDEMeshVertex->GetID());
	  bConnectivity = FALSE;
	}
      if ( uiVertexMultiple > 1 )
	{
	  pCSLIDEMeshVertex = (CSLIDEMeshVertex*)((VOID*)(pe->GetVertex()->GetForwardFacet()));
	  psv = (CSLIDEMeshVertex*)((VOID*)(pvx->GetForwardFacet()));
	  ASSERT( pCSLIDEMeshVertex != NULL );
	  pCSLIDEMeshVertex->HandleError("vertex \"%s\" has the same vertex neighbor %ld times: vertex \"%s\"", 
					 pCSLIDEMeshVertex->GetID(),
					 uiVertexMultiple, psv->GetID());
	  bConnectivity = FALSE;
	}
    }

  if ( bCorrect && !bConnectivity )
    {
      QEMesh *pmeClean;

      pmeClean = m_pQEMesh->CreateCleanMesh();
      ASSERT( pmeClean != NULL );
      fprintf(stderr, "START: cleaner version of mesh %s\n", m_pcID);
      ::WriteSLIDEMesh(pmeClean, stderr);
      fprintf(stderr, "(*\n");
      fprintf(stderr, "  START: possible other facets to make mesh closed\n");
      iterQEMesh.Init(pmeClean);
      for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
	    pe != NULL;
	    pe = iterQEMesh.PeekNextEdgeBoundary() )
	{
	  pe->SetForwardEdge(NULL);
	}
      for ( pe = iterQEMesh.PeekFirstEdgeBoundary(), i = 0;
	    pe != NULL;
	    pe = iterQEMesh.PeekNextEdgeBoundary(), i++ )
	{
	  if ( pe->GetForwardEdge() == NULL )
	    {
	      fprintf(stderr, "  facet fHole%ld (", i);
	      peCurr = pe;
	      do
		{
		  peCurr->SetForwardEdge(peCurr);
		  fprintf(stderr, " v%ld", peCurr->GetVertex()->GetID());
		  peCurr = peCurr->GetEdgeNext(0);
		}
	      while ( peCurr != pe );
	      fprintf(stderr, ") endfacet\n");	      
	    }
	}
      fprintf(stderr, "  END  : possible other facets to make mesh closed\n");
      fprintf(stderr, "*)\n");
      fprintf(stderr, "END  : cleaner version of mesh %s\n", m_pcID);
      DeallocateQEMesh(pmeClean);
    }
  bCorrect = ( bCorrect && bConnectivity );

  iter.Init(&m_listCSLIDEMeshEdge);
  for ( pCSLIDEMeshEdge = (CSLIDEMeshEdge*)(iter.PeekFirst());
	pCSLIDEMeshEdge != NULL;
	pCSLIDEMeshEdge = (CSLIDEMeshEdge*)(iter.PeekNext()) )
    {
      b = pCSLIDEMeshEdge->CheckSemantics(pCSLIDEWorld, this, m_pQEMesh);
      bCorrect = bCorrect && b;
    }

  return bCorrect;
}

BOOL CSLIDEMesh::Valid()
{
  if ( !CSLIDEAtomNode::Valid() )
    {
      return FALSE;
    }

  return TRUE;
}

CSLIDECamera *CSLIDEMesh::FindCamera(CListIterInt *piterIDs, 
				       CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtCamera],
	      gcppcSLIDEReservedWords[sliderwtMesh], GetID());

  return NULL;
}

CSLIDELight *CSLIDEMesh::FindLight(CListIterInt *piterIDs, 
				     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtLight],
	      gcppcSLIDEReservedWords[sliderwtMesh], GetID());

  return NULL;
}

CSLIDEAreaLight *CSLIDEMesh::FindAreaLight(CListIterInt *piterIDs, 
					     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtAreaLight],
	      gcppcSLIDEReservedWords[sliderwtMesh], GetID());

  return NULL;
}

VOID CSLIDEMesh::Preprocess()
{
  CSLIDEAtomNode::Preprocess();

  CListIterInt iter;
  CSLIDEMeshVertex *pCSLIDEMeshVertex;
  CSLIDEMeshFacet *pCSLIDEMeshFacet;
  CSLIDEMeshEdge *pCSLIDEMeshEdge;
  CSLIDEMeshEdgeProperty *pCSLIDEMeshEdgeProperty;
  CSLIDEMeshVertexProperty *pCSLIDEMeshVertexProperty;

  iter.Init(&m_listCSLIDEMeshVertexProperty);
  for ( pCSLIDEMeshVertexProperty = (CSLIDEMeshVertexProperty*)iter.PeekFirst(); 
	pCSLIDEMeshVertexProperty != NULL;
	pCSLIDEMeshVertexProperty = (CSLIDEMeshVertexProperty*)iter.PeekNext() )
    {
      pCSLIDEMeshVertexProperty->Preprocess();
    }

  iter.Init(&m_listCSLIDEMeshEdgeProperty);
  for ( pCSLIDEMeshEdgeProperty = (CSLIDEMeshEdgeProperty*)iter.PeekFirst(); 
	pCSLIDEMeshEdgeProperty != NULL;
	pCSLIDEMeshEdgeProperty = (CSLIDEMeshEdgeProperty*)iter.PeekNext() )
    {
      pCSLIDEMeshEdgeProperty->Preprocess();
    }

  iter.Init(&m_listCSLIDEMeshVertex);
  for ( pCSLIDEMeshVertex = (CSLIDEMeshVertex*)iter.PeekFirst(); 
	pCSLIDEMeshVertex != NULL;
	pCSLIDEMeshVertex = (CSLIDEMeshVertex*)iter.PeekNext() )
    {
      pCSLIDEMeshVertex->Preprocess();
    }

  iter.Init(&m_listCSLIDEMeshFacet);
  for ( pCSLIDEMeshFacet = (CSLIDEMeshFacet*)iter.PeekFirst(); 
	pCSLIDEMeshFacet != NULL;
	pCSLIDEMeshFacet = (CSLIDEMeshFacet*)iter.PeekNext() )
    {
      pCSLIDEMeshFacet->Preprocess();
    }

  iter.Init(&m_listCSLIDEMeshEdge);
  for ( pCSLIDEMeshEdge = (CSLIDEMeshEdge*)iter.PeekFirst(); 
	pCSLIDEMeshEdge != NULL;
	pCSLIDEMeshEdge = (CSLIDEMeshEdge*)iter.PeekNext() )
    {
      pCSLIDEMeshEdge->Preprocess();
    }

  CreateDynamicsList();

  CheckDynamic();
  UpdateState();
}

VOID CSLIDEMesh::CreateDynamicsList()
{
  CListIterInt iter;
  CSLIDEMeshVertex *pCSLIDEMeshVertex;
  CSLIDEMeshFacet *pCSLIDEMeshFacet;
  CSLIDEMeshEdge *pCSLIDEMeshEdge;
  CSLIDEMeshEdgeProperty *pCSLIDEMeshEdgeProperty;
  CSLIDEMeshVertexProperty *pCSLIDEMeshVertexProperty;

  iter.Init(&m_listCSLIDEMeshVertexProperty);
  for ( pCSLIDEMeshVertexProperty = (CSLIDEMeshVertexProperty*)iter.PeekFirst(); 
	pCSLIDEMeshVertexProperty != NULL;
	pCSLIDEMeshVertexProperty = (CSLIDEMeshVertexProperty*)iter.PeekNext() )
    {
      if ( pCSLIDEMeshVertexProperty->GetDynamic() )
	{
	  m_listDynamics.InsertLast(pCSLIDEMeshVertexProperty, (UINT_32)pCSLIDEMeshVertexProperty);
	}
    }

  iter.Init(&m_listCSLIDEMeshEdgeProperty);
  for ( pCSLIDEMeshEdgeProperty = (CSLIDEMeshEdgeProperty*)iter.PeekFirst(); 
	pCSLIDEMeshEdgeProperty != NULL;
	pCSLIDEMeshEdgeProperty = (CSLIDEMeshEdgeProperty*)iter.PeekNext() )
    {
      if ( pCSLIDEMeshEdgeProperty->GetDynamic() )
	{
	  m_listDynamics.InsertLast(pCSLIDEMeshEdgeProperty, (UINT_32)pCSLIDEMeshEdgeProperty);
	}
    }

  iter.Init(&m_listCSLIDEMeshVertex);
  for ( pCSLIDEMeshVertex = (CSLIDEMeshVertex*)iter.PeekFirst(); 
	pCSLIDEMeshVertex != NULL;
	pCSLIDEMeshVertex = (CSLIDEMeshVertex*)iter.PeekNext() )
    {
      if ( pCSLIDEMeshVertex->GetDynamic() )
	{
	  m_listDynamics.InsertLast(pCSLIDEMeshVertex, (UINT_32)pCSLIDEMeshVertex);
	}
    }

  iter.Init(&m_listCSLIDEMeshFacet);
  for ( pCSLIDEMeshFacet = (CSLIDEMeshFacet*)iter.PeekFirst(); 
	pCSLIDEMeshFacet != NULL;
	pCSLIDEMeshFacet = (CSLIDEMeshFacet*)iter.PeekNext() )
    {
      if ( pCSLIDEMeshFacet->GetDynamic() )
	{
	  m_listDynamics.InsertLast(pCSLIDEMeshFacet, (UINT_32)pCSLIDEMeshFacet);
	}
    }

  iter.Init(&m_listCSLIDEMeshEdge);
  for ( pCSLIDEMeshEdge = (CSLIDEMeshEdge*)iter.PeekFirst(); 
	pCSLIDEMeshEdge != NULL;
	pCSLIDEMeshEdge = (CSLIDEMeshEdge*)iter.PeekNext() )
    {
      if ( pCSLIDEMeshEdge->GetDynamic() )
	{
	  m_listDynamics.InsertLast(pCSLIDEMeshEdge, (UINT_32)pCSLIDEMeshEdge);
	}
    }
}

VOID CSLIDEMesh::CheckDynamic()
{
  // Are there variables which I must call TCL to update?
  SetDynamic(m_dyniShading.GetDynamic() ||
	     m_dyniLOD.GetDynamic() ||
	     m_dyniSolid.GetDynamic() ||
	     m_listDynamics.GetLength() > 0 );
}

VOID CSLIDEMesh::PreprocessGraph()
{
  if ( !GetVisited() )
    {
      CListIterInt iter;
      CSLIDEMeshVertex *pCSLIDEMeshVertex;
      CSLIDEMeshEdge *pCSLIDEMeshEdge;
      
      BOOL bGeometryEmpty;
      BOOL bGeometryDynamic;
      BOOL bNormalsEmpty;
      BOOL bNormalsDynamic;
      BOOL bBoundsEmpty;
      BOOL bBoundsDynamic;

      BOOL bBoundEmpty;
      BOOL bBoundDynamic;

      bGeometryEmpty = TRUE;
      bGeometryDynamic = FALSE;
      bNormalsEmpty = TRUE;
      bNormalsDynamic = FALSE;
      bBoundsEmpty = TRUE;
      bBoundsDynamic = FALSE;

      bBoundEmpty = TRUE;
      bBoundDynamic = FALSE;

      //////////////////////////////////////////////////
      // Look at the Vertices

      iter.Init(&m_listCSLIDEMeshVertex);
      for ( pCSLIDEMeshVertex = (CSLIDEMeshVertex*)iter.PeekFirst(); 
	    pCSLIDEMeshVertex != NULL;
	    pCSLIDEMeshVertex = (CSLIDEMeshVertex*)iter.PeekNext() )
	{
	  bBoundEmpty = FALSE;
	  bBoundDynamic = bBoundDynamic || 
	                  pCSLIDEMeshVertex->GetBoundDynamic();
	  bGeometryDynamic = bGeometryDynamic || 
	                     pCSLIDEMeshVertex->GetGeometryDynamic();
	  bNormalsDynamic = bNormalsDynamic ||
	                    pCSLIDEMeshVertex->GetNormalDynamic();
	}

      //////////////////////////////////////////////////
      // Look at the Edges

      iter.Init(&m_listCSLIDEMeshVertex);
      for ( pCSLIDEMeshVertex = (CSLIDEMeshVertex*)iter.PeekFirst(); 
	    pCSLIDEMeshVertex != NULL;
	    pCSLIDEMeshVertex = (CSLIDEMeshVertex*)iter.PeekNext() )
	{
	  bGeometryDynamic = bGeometryDynamic || 
	                     pCSLIDEMeshVertex->GetVertexProperty()->GetDynamic();
	}

      iter.Init(&m_listCSLIDEMeshEdge);
      for ( pCSLIDEMeshEdge = (CSLIDEMeshEdge*)iter.PeekFirst(); 
	    pCSLIDEMeshEdge != NULL;
	    pCSLIDEMeshEdge = (CSLIDEMeshEdge*)iter.PeekNext() )
	{
	  bGeometryDynamic = bGeometryDynamic || 
	                     pCSLIDEMeshEdge->GetEdgeProperty()->GetDynamic();
	}

      bGeometryEmpty = bBoundEmpty;
      bNormalsEmpty = bBoundEmpty;
      bBoundsEmpty = bBoundEmpty;
      bBoundsDynamic = bBoundDynamic;

      //////////////////////////////////////////////////
      // Look at Local flags

      // Update whether we are Empty based on our LOD flag
      if ( !m_dyniLOD.GetDynamic() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      // This node is always off
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	      bBoundsEmpty = TRUE;
	      bBoundEmpty = TRUE;
	    }
	  else if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	    {
	      // This node is always just a bounding box
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	    }
	}

      // Check what is really dynamic

      bGeometryDynamic = ( bGeometryDynamic ||
			   m_pCSLIDESurface->GetDynamic() ||
			   m_dyniShading.GetDynamic() ||
			   m_dyniLOD.GetDynamic() ||
			   m_dyniSolid.GetDynamic() );
      bNormalsDynamic = ( bNormalsDynamic ||
			  m_dyniShading.GetDynamic() ||
			  m_dyniLOD.GetDynamic() );
      bBoundsDynamic = ( bBoundsDynamic || 
			 m_dyniLOD.GetDynamic() );
      bBoundDynamic = ( bBoundDynamic || 
			m_dyniLOD.GetDynamic() );

      // If anything is now Empty then it will always be empty so
      // it is not dynamic

      bGeometryDynamic = ( bGeometryDynamic && !bGeometryEmpty );
      bNormalsDynamic = ( bNormalsDynamic && !bNormalsEmpty );
      bBoundsDynamic = ( bBoundsDynamic && !bBoundsEmpty );
      bBoundDynamic = ( bBoundDynamic && !bBoundEmpty );
      
      //////////////////////////////////////////////////
      // For the OpenGL Display Lists to be really useful 
      // it is all or nothing so we enforce that here

      bGeometryDynamic = ( bGeometryDynamic ||
			   bNormalsDynamic ||
			   bBoundsDynamic );
      bNormalsDynamic = bGeometryDynamic;
      bBoundsDynamic = bGeometryDynamic;

      //////////////////////////////////////////////////
      // Now save these correct flag values

      SetGeometryEmpty(bGeometryEmpty);
      SetGeometryDynamic(bGeometryDynamic);
      SetNormalsEmpty(bNormalsEmpty);
      SetNormalsDynamic(bNormalsDynamic);
      SetBoundsEmpty(bBoundsEmpty);
      SetBoundsDynamic(bBoundsDynamic);
      SetBoundEmpty(bBoundEmpty);
      SetBoundDynamic(bBoundDynamic);
      
      //////////////////////////////////////////////////
      // Calculate the Bounding Box (if Empty this will do nothing)
      ASSERT( m_pQEMesh != NULL );
      m_pQEMesh->Preprocess();

      if ( GetBoundEmpty() )
	{
	  SetBoundValid(TRUE);
	}
      else
	{
	  CalculateBound();
	}
      SetBoundChanged(FALSE);

      //////////////////////////////////////////////////
      // Done Preprocessing this Node in the Graph
      SetVisited(TRUE);
    }
}

VOID CSLIDEMesh::Update()
{
  SetBoundChanged(FALSE);
  if ( GetDynamic() ||
       GetBoundsDynamic() )
    {
      UpdateState();
    }

  if ( GetBoundDynamic() )
    {
      SetBoundValid(FALSE);
    }

  SetVisited(FALSE);
}

VOID CSLIDEMesh::UpdateState()
{
  CListIterInt iter;
  CSLIDEPrimitive *pCSLIDEPrimitive;

  CSLIDEAtomNode::UpdateState();

  iter.Init(&m_listDynamics);
  for ( pCSLIDEPrimitive = (CSLIDEPrimitive*)(iter.PeekFirst());
	pCSLIDEPrimitive != NULL;
	pCSLIDEPrimitive = (CSLIDEPrimitive*)(iter.PeekNext()) )
    {
      pCSLIDEPrimitive->Update();
    }
}

VOID CSLIDEMesh::UpdateGraph()
{
  if ( !GetVisited() )
    {
      if ( !GetBoundValid() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      if ( !m_bound.GetEmpty() )
		{
		  SetBoundChanged(TRUE);
		  m_bound.Null();
		}
	      return;
	    }
	  
	  CalculateBound();
	  SetBoundValid(TRUE);
	}

      ASSERT( m_pQEMesh != NULL );
      m_pQEMesh->Preprocess();

      SetVisited(TRUE);
    }
}

VOID CSLIDEMesh::CalculateBound()
{
  CListIterInt iter;
  CSLIDEMeshVertex *pCSLIDEMeshVertex;

  CBound bound;

  iter.Init(&m_listCSLIDEMeshVertex);

  for ( pCSLIDEMeshVertex = (CSLIDEMeshVertex*)iter.PeekFirst(); 
	pCSLIDEMeshVertex != NULL;
	pCSLIDEMeshVertex = (CSLIDEMeshVertex*)iter.PeekNext() )
    {
      pCSLIDEMeshVertex->InsertIntoBound(&bound);
    }

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);
    }
}

#ifdef __USE_CDA__
VOID CSLIDEMesh::CalculateCDAVertices(CListInt *plistCDAVertex, 
					const Matrix &mGroot_obj)
{
  CListIterInt iter;
  CSLIDEMeshVertex *pCSLIDEMeshVertex;

  iter.Init(&m_listCSLIDEMeshVertex);
  for ( pCSLIDEMeshVertex = (CSLIDEMeshVertex*)iter.PeekFirst(); 
	pCSLIDEMeshVertex != NULL;
	pCSLIDEMeshVertex = (CSLIDEMeshVertex*)iter.PeekNext() )
    {
      pCSLIDEMeshVertex->CalculateCDAVertices(plistCDAVertex, mGroot_obj);
    }
}
#endif // __USE_CDA__

VOID CSLIDEMesh::CalculateQEMesh()
{
  ASSERT( m_pQEMesh != NULL );
}

VOID CSLIDEMesh::PrerenderGeometry(SLIDEShadingType eShading)
{
  SLIDEShadingType eShade;
  
  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  // This flag tells us whether to use Point or Point Normals
  ModifyShadingFlag(eShading);

  // This one set the shading attributes which in many cases will 
  // be SLF_INHERIT and so it will not set anything and leave up
  // to the caller above
  eShade = (SLIDEShadingType)m_dyniShading.GetValue();
  
  GLbitfield GLbAttributesMask;
  
  GLbAttributesMask = CalculateAttributesMask(eShade);
  
  PushAttributes(GLbAttributesMask);
  {
    PushShadingAttributes(eShade);
    {
      m_pCSLIDESurface->Push();
      {
	PushSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
	{
	  m_uiPrerenderAttributesStackDepthMax = 0;

	  ::RenderGeometry(m_pQEMesh);

	  if ( GLbAttributesMask != 0x0 )
	    {
	      m_uiPrerenderAttributesStackDepthMax++;
	    }
	}
	PopSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
      }
      m_pCSLIDESurface->Pop();
    }
    PopShadingAttributes(eShade);
  }
  PopAttributes(GLbAttributesMask);
}

VOID CSLIDEMesh::PrerenderNormals(SLIDEShadingType eShading)
{
  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  ModifyShadingFlag(eShading);

  ::RenderNormals(m_pQEMesh);
}

VOID CSLIDEMesh::RenderGeometry(const Matrix &/*mGproj_obj*/, 
				CBound * /*pCBoundClip*/, 
				SLIDEShadingType eShading,
				SLIDELODType /* eLOD */,
				CSLIDEWormwarp * /* pCSLIDEWarp *//* = NULL */)
{
  PushSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
  {
    ::RenderGeometry(m_pQEMesh);
  }
  PopSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
}

VOID CSLIDEMesh::RenderNormals(SLIDEShadingType eShading)
{
  PushAuxillaryGeometryAttributes();
  {
    gilColor3f(1.0, 0.0, 1.0);

    ::RenderNormals(m_pQEMesh);
  }
  PopAuxillaryGeometryAttributes();
}

VOID CSLIDEMesh::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDEMeshVertexProperty *pCSLIDEMeshVertexProperty;
  CSLIDEMeshEdgeProperty *pCSLIDEMeshEdgeProperty;
  CSLIDEMeshVertex *pCSLIDEMeshVertex;
  CSLIDEMeshFacet *pCSLIDEMeshFacet;
  CSLIDEMeshEdge *pCSLIDEMeshEdge;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtMesh], m_pcGUIID);

  WriteSLIDEFlags(fpOutputFile, 0);

  if ( m_dyniSolid.GetValue() != slidestSLF_SOLID )
    {
      fprintf(fpOutputFile, "  %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtSolid], 
	      gcppcSLIDESolidReservedWords[m_dyniSolid.GetValue()]);
    }

  iter.Init(&m_listCSLIDEMeshVertexProperty);
  for ( pCSLIDEMeshVertexProperty = (CSLIDEMeshVertexProperty*)iter.PeekFirst(); 
	pCSLIDEMeshVertexProperty != NULL;
	pCSLIDEMeshVertexProperty = (CSLIDEMeshVertexProperty*)iter.PeekNext() )
    {
      pCSLIDEMeshVertexProperty->WriteSLIDE(fpOutputFile);
    }

  iter.Init(&m_listCSLIDEMeshEdgeProperty);
  for ( pCSLIDEMeshEdgeProperty = (CSLIDEMeshEdgeProperty*)iter.PeekFirst(); 
	pCSLIDEMeshEdgeProperty != NULL;
	pCSLIDEMeshEdgeProperty = (CSLIDEMeshEdgeProperty*)iter.PeekNext() )
    {
      pCSLIDEMeshEdgeProperty->WriteSLIDE(fpOutputFile);
    }

  iter.Init(&m_listCSLIDEMeshVertex);
  for ( pCSLIDEMeshVertex = (CSLIDEMeshVertex*)iter.PeekFirst(); 
	pCSLIDEMeshVertex != NULL;
	pCSLIDEMeshVertex = (CSLIDEMeshVertex*)iter.PeekNext() )
    {
      pCSLIDEMeshVertex->WriteSLIDE(fpOutputFile);
    }

  iter.Init(&m_listCSLIDEMeshFacet);
  for ( pCSLIDEMeshFacet = (CSLIDEMeshFacet*)iter.PeekFirst(); 
	pCSLIDEMeshFacet != NULL;
	pCSLIDEMeshFacet = (CSLIDEMeshFacet*)iter.PeekNext() )
    {
      pCSLIDEMeshFacet->WriteSLIDE(fpOutputFile);
    }

  iter.Init(&m_listCSLIDEMeshEdge);
  for ( pCSLIDEMeshEdge = (CSLIDEMeshEdge*)iter.PeekFirst(); 
	pCSLIDEMeshEdge != NULL;
	pCSLIDEMeshEdge = (CSLIDEMeshEdge*)iter.PeekNext() )
    {
      pCSLIDEMeshEdge->WriteSLIDE(fpOutputFile);
    }

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndMesh]);
}

VOID CSLIDEMesh::WriteRIBGeometry(FILE *fpOutputFile,
				    const Matrix &/*mGproj_obj*/, 
				    CBound * /*pCBoundClip*/, 
				    SLIDEShadingType eShading,
				    SLIDELODType eLOD)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBPushSolidAttributes(fpOutputFile,
			      (SLIDESolidType)m_dyniSolid.GetValue());
  {
    WriteRIBGeometryPolygonal(fpOutputFile, eShading, eLOD);
  }
  WriteRIBPopSolidAttributes(fpOutputFile, 
			     (SLIDESolidType)m_dyniSolid.GetValue());
}

BOOL CSLIDEMesh::CheckRIBPrimitive()
{
  BOOL bResult;

  bResult = FALSE;

  return bResult;
}

VOID CSLIDEMesh::WriteRIBGeometryPolygonal(FILE *fpOutputFile,
					   SLIDEShadingType eShading,
					   SLIDELODType /* eLOD */)
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDEMesh::WriteRIBGeometryPrimitive(FILE *fpOutputFile,
					     SLIDEShadingType /* eShading */,
					     SLIDELODType /* eLOD */)
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDEMesh::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE MESH: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);
}

VOID CSLIDEMesh::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE MESH: %s", m_pcGUIID);
}

VOID CSLIDEMesh::WritePOVGeometry(FILE *fpOutputFile, CListInt *plistTransformIDs)
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDEMesh::WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &m,
					 CSLIDESurface *pCSLIDESurface)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "Shape { # triangle mesh: MESH\n\n");

  // Mesh
  ASSERT(m_pQEMesh != NULL);
  ::WriteWRLZ(m_pQEMesh, fpOutputFile, m);

  fprintf(fpOutputFile, "} # end triangle mesh\n\n");
}

VOID CSLIDEMesh::WriteSTLGeometry(FILE *fpOutputFile, const Matrix &m)
{
  ASSERT( fpOutputFile != NULL );

  ::WriteSTL(m_pQEMesh, fpOutputFile, m);
}

VOID CSLIDEMesh::WriteDotID(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "Mesh_%s", GetGUIID());
}

VOID CSLIDEMesh::WriteDotNode(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "  ");
  WriteDotID(fpOutputFile);
  fprintf(fpOutputFile, " [ label = \"%s\", shape = ellipse, peripheries = 2", GetGUIID());
  if ( GetBoundDynamic() )
    {
      fprintf(fpOutputFile, ", color = red, fontcolor = red");
    }
  else if ( GetGeometryDynamic() )
    {
      fprintf(fpOutputFile, ", color = blue, fontcolor = blue");
    }
  fprintf(fpOutputFile, " ];\n");
}

VOID CSLIDEMesh::WriteDotEdges(FILE * /*fpOutputFile*/)
{
}
