#ifndef __SLIDEMESH_H__
#define __SLIDEMESH_H__

#include "slideatomnode.h"

class CSLIDEMeshVertex;
class CSLIDEMeshVertexProperty;
class CSLIDEMeshFacet;
class CSLIDEMeshEdge;
class CSLIDEMeshEdgeProperty;
class QEMesh;

//////////////////////////////////////////////////////////////////////////////
// CSLIDEMesh Class
//

class CSLIDEMesh : public CSLIDEAtomNode
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEMesh();
  virtual ~CSLIDEMesh();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  BOOL InsertMeshVertexProperty(CSLIDEMeshVertexProperty *pCSLIDEMeshVertexProperty);
  CSLIDEMeshVertexProperty *PeekMeshVertexProperty(CHAR *pcID);
  BOOL InsertMeshEdgeProperty(CSLIDEMeshEdgeProperty *pCSLIDEMeshEdgeProperty);
  CSLIDEMeshEdgeProperty *PeekMeshEdgeProperty(CHAR *pcID);
  BOOL InsertMeshVertex(CSLIDEMeshVertex *pCSLIDEMeshVertex);
  CSLIDEMeshVertex *PeekMeshVertex(CHAR *pcID);
  BOOL InsertMeshFacet(CSLIDEMeshFacet *pCSLIDEMeshFacet);
  CSLIDEMeshFacet *PeekMeshFacet(CHAR *pcID);
  BOOL InsertMeshEdge(CSLIDEMeshEdge *pCSLIDEMeshEdge);
  CSLIDEMeshEdge *PeekMeshEdge(CHAR *pcID);

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual CSLIDECamera *FindCamera(CListIterInt *piterIDs, 
				   CListInt *plistCSLIDEInstance);
  virtual CSLIDELight *FindLight(CListIterInt *piterIDs, 
				 CListInt *plistCSLIDEInstance);
  virtual CSLIDEAreaLight *FindAreaLight(CListIterInt *piterIDs, 
					 CListInt *plistCSLIDEInstance);

  virtual VOID Preprocess();
  virtual VOID PreprocessGraph();

  virtual VOID Update();
  virtual VOID UpdateGraph();

#ifdef __USE_CDA__
  virtual VOID CalculateCDAVertices(CListInt *plistCDAVertex, 
				    const Matrix &mGroot_obj);
#endif // __USE_CDA__

  virtual VOID CalculateQEMesh();

protected:
  virtual VOID CreateDynamicsList();
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();
  virtual VOID CalculateBound();

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID PrerenderGeometry(SLIDEShadingType eShading);
  virtual VOID PrerenderNormals(SLIDEShadingType eShading);

  virtual VOID RenderGeometry(const Matrix &mGproj_obj,
			      CBound *pCBoundClip, 
			      SLIDEShadingType eShading,
			      SLIDELODType eLOD,
			      CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderNormals(SLIDEShadingType eShading);

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);

  BOOL CheckRIBPrimitive();
  VOID WriteRIBGeometryPolygonal(FILE *fpOutputFile,
				 SLIDEShadingType eShading,
				 SLIDELODType eLOD);
  VOID WriteRIBGeometryPrimitive(FILE *fpOutputFile,
				 SLIDEShadingType eShading,
				 SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

  virtual VOID WritePOVGeometry(FILE *fpOutputFile, CListInt *plistTransformIDs);

  virtual VOID WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				 CSLIDESurface *pCSLIDESurface);
  virtual VOID WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform);

  virtual VOID WriteDotID(FILE *fpOutputFile);
  virtual VOID WriteDotNode(FILE *fpOutputFile);
  virtual VOID WriteDotEdges(FILE *fpOutputFile);

//////////////////////////////////////////////////
// Accessor Functions
//
public:

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CListInt m_listCSLIDEMeshVertexProperty;
  CHashInt m_hashCSLIDEMeshVertexProperty;
  CListInt m_listCSLIDEMeshEdgeProperty;
  CHashInt m_hashCSLIDEMeshEdgeProperty;
  CListInt m_listCSLIDEMeshVertex;
  CHashInt m_hashCSLIDEMeshVertex;
  CListInt m_listCSLIDEMeshFacet;
  CHashInt m_hashCSLIDEMeshFacet;
  CListInt m_listCSLIDEMeshEdge;
  CHashInt m_hashCSLIDEMeshEdge;

  CListInt m_listDynamics;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

#endif // __SLIDEMESH_H__
