#include "slidemeshedge.h"

#include "slidemesh.h"
#include "slidemeshvertex.h"
#include "slidemeshedgeproperty.h"
#include "qemesh.h"

//////////////////////////////////////////////////////////////////////////////
// CSLIDEMeshEdge Class
//

CSLIDEMeshEdge::CSLIDEMeshEdge()
{
  m_ppcMeshVertexIDs[0] = NULL;
  m_ppcMeshVertexIDs[1] = NULL;
  m_ppCSLIDEMeshVertex[0] = NULL;
  m_ppCSLIDEMeshVertex[1] = NULL;
  m_pcMeshEdgePropertyID = NULL;
  m_pCSLIDEMeshEdgeProperty = NULL;
  m_pQEEdge = NULL;
}

CSLIDEMeshEdge::~CSLIDEMeshEdge()
{
}

VOID CSLIDEMeshEdge::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID, 
			  CHAR *pcVertexID0, CHAR *pcVertexID1)
{
  CSLIDEPrimitive::Init(pcFile, uiLine, uiColumn, pcID);

  ASSERT( pcVertexID0 != NULL );
  ASSERT( pcVertexID1 != NULL );

  m_ppcMeshVertexIDs[0] = pcVertexID0;
  m_ppcMeshVertexIDs[1] = pcVertexID1;
  m_ppCSLIDEMeshVertex[0] = NULL;
  m_ppCSLIDEMeshVertex[1] = NULL;
  m_pcMeshEdgePropertyID = gcppcSLIDEReservedWords[sliderwtSLF_INHERIT];
  m_pCSLIDEMeshEdgeProperty = NULL;
  m_pQEEdge = NULL;
}

VOID CSLIDEMeshEdge::Uninit()
{
  CSLIDEPrimitive::Uninit();
}

BOOL CSLIDEMeshEdge::Initialized()
{
  if ( !CSLIDEPrimitive::Initialized() )
    {
      return FALSE;
    }

  return ( m_ppcMeshVertexIDs[0] != NULL &&
	   m_ppcMeshVertexIDs[1] != NULL );
}

BOOL CSLIDEMeshEdge::CheckSemantics(CSLIDEWorld *pCSLIDEWorld, CSLIDEMesh *pCSLIDEMesh,
				    QEMesh *pQEMesh)
{
  BOOL bCorrect;
  UINT_32 i;
  CListIterInt iter;
  QEVertexIter iterQEVertex;
  QEVertex *ppvx[2];
  QEEdge *pe;

  if ( !CSLIDEPrimitive::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  // Check Mesh Edge Property Reference
  m_pCSLIDEMeshEdgeProperty = pCSLIDEMesh->PeekMeshEdgeProperty(m_pcMeshEdgePropertyID);
  if ( m_pCSLIDEMeshEdgeProperty == NULL )
    {
      HandleError("Undefined %s \"%s\"", gcppcSLIDEReservedWords[sliderwtMeshEdgeProperty], 
		  m_pcMeshEdgePropertyID);
      bCorrect = FALSE;
    }

  // Check Mesh Vertex References
  for ( i = 0; i < 2; i++ )
    {
      m_ppCSLIDEMeshVertex[i] = pCSLIDEMesh->PeekMeshVertex(m_ppcMeshVertexIDs[i]);
      if ( m_ppCSLIDEMeshVertex[i] == NULL )
	{
	  HandleError("Undefined %s \"%s\"", gcppcSLIDEReservedWords[sliderwtMeshVertex], 
		      m_ppcMeshVertexIDs[i]);
	  bCorrect = FALSE;
	}
    }

  if ( !bCorrect )
    {
      return FALSE;
    }

  // Find the edge in the QEMesh

  ppvx[0] = m_ppCSLIDEMeshVertex[0]->GetQEVertex();
  ppvx[1] = m_ppCSLIDEMeshVertex[1]->GetQEVertex();
  
  m_pQEEdge = NULL;
  iterQEVertex.Init(ppvx[0]);
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      if ( pe->GetEdge()->GetVertex() == ppvx[1] )
	{
	  m_pQEEdge = pe;
	  break;
	}
    }

  if ( m_pQEEdge == NULL )
    {
      HandleError("%s \"%s\" from %s \"%s\" to %s \"%s\" does not exist in the %s", 
		  gcppcSLIDEReservedWords[sliderwtMeshEdge], GetID(),
		  gcppcSLIDEReservedWords[sliderwtMeshVertex], m_ppcMeshVertexIDs[0],
		  gcppcSLIDEReservedWords[sliderwtMeshVertex], m_ppcMeshVertexIDs[1],
		  gcppcSLIDEReservedWords[sliderwtMesh]);
      bCorrect = FALSE;
    }
  
  return bCorrect;
}

BOOL CSLIDEMeshEdge::Valid()
{
  if ( !CSLIDEPrimitive::Valid() )
    {
      return FALSE;
    }

  return ( m_ppcMeshVertexIDs[0] != NULL &&
	   m_ppcMeshVertexIDs[1] != NULL );
}

VOID CSLIDEMeshEdge::Preprocess()
{
  CheckDynamic();
  UpdateState();
}

VOID CSLIDEMeshEdge::CheckDynamic()
{
}

VOID CSLIDEMeshEdge::UpdateState()
{
  FLOAT fSharpness;

  fSharpness = m_pCSLIDEMeshEdgeProperty->GetSharpness();
  m_pQEEdge->SetKnotInterval(fSharpness);
  m_pQEEdge->GetEdge()->SetKnotInterval(fSharpness);
}

VOID CSLIDEMeshEdge::WriteSLIDE(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "  %s %s ( %s %s )\n", 
	  gcppcSLIDEReservedWords[sliderwtMeshEdge], 
	  m_pcGUIID,
	  m_ppcMeshVertexIDs[0], m_ppcMeshVertexIDs[1]);

  if ( m_pCSLIDEMeshEdgeProperty != CSLIDEMeshEdgeProperty::GetSLF_INHERIT() )
    {
      fprintf(fpOutputFile, "    %s %s\n", 
	      gcppcSLIDEReservedWords[sliderwtMeshEdgeProperty], 
	      m_pcMeshEdgePropertyID);
    }

  fprintf(fpOutputFile, "  %s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndMeshEdge]);
}
