#include "slidemeshfacet.h"

#include "slidemesh.h"
#include "slidemeshvertex.h"
#include "qemesh.h"

//////////////////////////////////////////////////////////////////////////////
// CSLIDEMeshFacet Class
//

CSLIDEMeshFacet::CSLIDEMeshFacet()
{
  m_pQEFacet = NULL;
}

CSLIDEMeshFacet::~CSLIDEMeshFacet()
{
}

VOID CSLIDEMeshFacet::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID, 
		      CListInt *plistMeshVertexIDs)
{
  CSLIDEPrimitive::Init(pcFile, uiLine, uiColumn, pcID);

  ASSERT( plistMeshVertexIDs != NULL );

  m_plistMeshVertexIDs = plistMeshVertexIDs;
  m_listCSLIDEMeshVertex.Init();
}

VOID CSLIDEMeshFacet::Uninit()
{
  CSLIDEPrimitive::Uninit();

  if ( m_plistMeshVertexIDs != NULL )
    {
      delete m_plistMeshVertexIDs;
      m_plistMeshVertexIDs = NULL;
    }
}

BOOL CSLIDEMeshFacet::Initialized()
{
  if ( !CSLIDEPrimitive::Initialized() )
    {
      return FALSE;
    }

  return ( m_plistMeshVertexIDs != NULL );
}

BOOL CSLIDEMeshFacet::CheckSemantics(CSLIDEWorld *pCSLIDEWorld, CSLIDEMesh *pCSLIDEMesh,
				     QEMesh *pQEMesh)
{
  BOOL bCorrect;
  CListIterInt iter;
  CHAR *pcID;
  CSLIDEMeshVertex *pCSLIDEMeshVertex;

  if ( !CSLIDEPrimitive::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  // Check Mesh Vertex References
  iter.Init(m_plistMeshVertexIDs);
  for ( pcID = (CHAR*)(iter.PeekFirst());
	pcID != NULL;
	pcID = (CHAR*)(iter.PeekNext()) )
    {
      pCSLIDEMeshVertex = pCSLIDEMesh->PeekMeshVertex(pcID);
      if ( pCSLIDEMeshVertex == NULL )
	{
	  HandleError("Undefined %s \"%s\"", gcppcSLIDEReservedWords[sliderwtMeshVertex], 
		      pcID);
	  bCorrect = FALSE;
	}
      else
	{
	  m_listCSLIDEMeshVertex.InsertLast(pCSLIDEMeshVertex, 0);
	}
    }

  if ( m_plistMeshVertexIDs->GetLength() < 3 )
    {
      HandleError("%s requires at least 3 vertices", 
		  gcppcSLIDEReservedWords[sliderwtMeshFacet]);
      bCorrect = FALSE;
    }

  if ( !bCorrect )
    {
      return FALSE;
    }

  QEVertex *pvx;
  QEEdge *peStart, *pePrev, *peCurr;

  m_pQEFacet = pQEMesh->CreateFacet();
  ASSERT( m_pQEFacet != NULL );

  iter.Init(&m_listCSLIDEMeshVertex);
  pCSLIDEMeshVertex = (CSLIDEMeshVertex*)iter.PeekFirst();
  ASSERT( pCSLIDEMeshVertex != NULL );
  pvx = pCSLIDEMeshVertex->GetQEVertex();
  ASSERT( pvx != NULL );
  peCurr = pQEMesh->CreateEdge();
  ASSERT( peCurr != NULL );

  peCurr->SetVertex(pvx);
  peCurr->SetFacet(m_pQEFacet);
  peCurr->GetEdge()->SetFacet(pQEMesh->GetNullFacet());

  peStart = peCurr;
  pePrev = peCurr;

  for ( pCSLIDEMeshVertex = (CSLIDEMeshVertex*)iter.PeekNext(); 
	pCSLIDEMeshVertex != NULL;
	pCSLIDEMeshVertex = (CSLIDEMeshVertex*)iter.PeekNext() )
    {
      pvx = pCSLIDEMeshVertex->GetQEVertex();
      ASSERT( pvx != NULL );
      peCurr = pQEMesh->CreateEdge();
      ASSERT( peCurr != NULL );
      
      pePrev->GetEdge()->SetVertex(pvx);
      peCurr->SetVertex(pvx);
      peCurr->SetFacet(m_pQEFacet);
      peCurr->GetEdge()->SetFacet(pQEMesh->GetNullFacet());

      pQEMesh->Splice(1, pePrev->GetEdge(), peCurr);

      pePrev = peCurr;
    }
  pePrev->GetEdge()->SetVertex(peStart->GetVertex());
  pQEMesh->Splice(1, pePrev->GetEdge(), peStart);

  pQEMesh->MergeLocalFacet(m_pQEFacet);

  return bCorrect;
}

BOOL CSLIDEMeshFacet::Valid()
{
  if ( !CSLIDEPrimitive::Valid() )
    {
      return FALSE;
    }

  return ( m_listCSLIDEMeshVertex.GetLength() >= 3);
}

VOID CSLIDEMeshFacet::Preprocess()
{
  CheckDynamic();
  UpdateState();
}

VOID CSLIDEMeshFacet::CheckDynamic()
{
}

VOID CSLIDEMeshFacet::UpdateState()
{
}

VOID CSLIDEMeshFacet::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDEMeshVertex *pCSLIDEMeshVertex;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "  %s %s ( ", gcppcSLIDEReservedWords[sliderwtMeshFacet], 
	  m_pcGUIID);

  iter.Init(&m_listCSLIDEMeshVertex);
  for ( pCSLIDEMeshVertex = (CSLIDEMeshVertex*)iter.PeekFirst(); 
	pCSLIDEMeshVertex != NULL;
	pCSLIDEMeshVertex = (CSLIDEMeshVertex*)iter.PeekNext() )
    {
      fprintf(fpOutputFile, "%s ", pCSLIDEMeshVertex->GetGUIID());      
    }

  fprintf(fpOutputFile, ")\n");

  fprintf(fpOutputFile, "  %s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndMeshFacet]);
}
