#include "slidemeshvertex.h"
#include "slidemeshvertexproperty.h"
#include "slidemesh.h"

#include "qemesh.h"

//////////////////////////////////////////////////////////////////////////////
// CSLIDEMeshVertex Class
//

CSLIDEMeshVertex::CSLIDEMeshVertex()
{
  m_pcMeshVertexPropertyID = NULL;
  m_pCSLIDEMeshVertexProperty = NULL;
  m_pt.Null();
  m_vNormal.Null();
  m_ptTexture.Null();

  m_pQEVertex = NULL;
}

CSLIDEMeshVertex::~CSLIDEMeshVertex()
{
}

VOID CSLIDEMeshVertex::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		       CHAR *pcID, 
		       DynPoint *pdynpt)
{
  ASSERT( pdynpt != NULL );

  CSLIDEPrimitive::Init(pcFile, uiLine, uiColumn, pcID);

  m_pcMeshVertexPropertyID = gcppcSLIDEReservedWords[sliderwtSLF_INHERIT];
  m_pCSLIDEMeshVertexProperty = NULL;
  m_pt.SetAll(pdynpt);
  m_vNormal.Null();
  m_ptTexture.Null();

  SetBoundDynamic(FALSE);
}

VOID CSLIDEMeshVertex::Uninit()
{
  CSLIDEPrimitive::Uninit();
}

BOOL CSLIDEMeshVertex::Initialized()
{
  if ( !CSLIDEPrimitive::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

BOOL CSLIDEMeshVertex::CheckSemantics(CSLIDEWorld *pCSLIDEWorld, CSLIDEMesh *pCSLIDEMesh, 
				      QEMesh *pQEMesh)
{
  Vector v;
  BOOL bCorrect;

  bCorrect = TRUE;

  if ( !CSLIDEPrimitive::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  ASSERT( m_pQEVertex == NULL );

  m_pQEVertex = pQEMesh->CreateVertex(m_pt);
  ASSERT( m_pQEVertex != NULL );

  // Check Mesh Vertex Property Reference
  m_pCSLIDEMeshVertexProperty = pCSLIDEMesh->PeekMeshVertexProperty(m_pcMeshVertexPropertyID);
  if ( m_pCSLIDEMeshVertexProperty == NULL )
    {
      HandleError("Undefined %s \"%s\"", gcppcSLIDEReservedWords[sliderwtMeshVertexProperty], 
		  m_pcMeshVertexPropertyID);
      bCorrect = FALSE;
    }
  else if ( m_pCSLIDEMeshVertexProperty->GetType() == slidevxtSLF_VERTEX_CORNER )
    {
      m_pQEVertex->SetSharpVertexType(QE::svtCorner);
    }
  m_pQEVertex->SetNormal(m_vNormal);
  v = m_ptTexture;
  m_pQEVertex->SetTexture(v);
  m_pQEVertex->SetForwardFacet((QEFacet*)((VOID*)this));

  UpdateState();

  return bCorrect;
}

BOOL CSLIDEMeshVertex::CheckValence()
{
  ASSERT( m_pQEVertex != NULL );

  QEVertexIter iterQEVertex;
  QEEdge *pe;
  UINT_32 uiValence, uiBoundaryEdges;
  BOOL bResult;

  bResult = TRUE;

  uiBoundaryEdges = 0;
  iterQEVertex.Init(m_pQEVertex);
  for ( pe = iterQEVertex.PeekFirstEdge(), uiValence = 0;
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge(), uiValence++ )
    {
      if ( pe->GetFacet() == QEMesh::GetNullFacet() || 
	   pe->GetEdge()->GetFacet() == QEMesh::GetNullFacet() )
	{
	  uiBoundaryEdges++;
	}
    }

  switch ( uiValence )
    {
    case 0:
      //HandleError("%s \"%s\" is not referenced by any %s's", 
      HandleWarning("%s \"%s\" is not referenced by any %s's", 
		    gcppcSLIDEReservedWords[sliderwtMeshVertex], GetID(), 
		    gcppcSLIDEReservedWords[sliderwtMeshFacet]);
      //bResult = FALSE;
      break;
    case 1:
      ASSERT( FALSE );
      break;
    case 2:
      if ( uiBoundaryEdges < 2 )
	{
	  HandleError("%s \"%s\" is a smooth interior vertex of valence 2", 
		      gcppcSLIDEReservedWords[sliderwtMeshVertex], GetID());
	  bResult = FALSE;
	}
      break;
    default:
      break;
    }

  return bResult;
}

BOOL CSLIDEMeshVertex::Valid()
{
  if ( !CSLIDEPrimitive::Valid() )
    {
      return FALSE;
    }

  return m_pt[W] != 0.0;
}

VOID CSLIDEMeshVertex::Preprocess()
{
  UpdateState();
  CheckDynamic();
}

VOID CSLIDEMeshVertex::InsertIntoBound(CBound *pCBound)
{
  ASSERT( pCBound != NULL );

  pCBound->Insert(m_pt);
}

#ifdef __USE_CDA__
VOID CSLIDEMeshVertex::CalculateCDAVertices(CListInt *plistCDAVertex, 
					    const Matrix &mGroot_obj)
{
  CDAVertex *pCDAVertex;

  ASSERT( plistCDAVertex != NULL );

  pCDAVertex = AllocateCDAVertex();
  ASSERT( pCDAVertex != NULL );
  pCDAVertex->Init(mGroot_obj*m_pt);
  plistCDAVertex->InsertLast(pCDAVertex, 0);
}
#endif // __USE_CDA__

VOID CSLIDEMeshVertex::CheckDynamic()
{
  SetDynamic(m_pt.GetDynamic() ||
	     m_vNormal.GetDynamic() ||
	     m_ptTexture.GetDynamic());
  SetBoundDynamic(m_pt.GetDynamic());
  SetGeometryDynamic(GetDynamic());
  SetNormalDynamic(m_vNormal.GetDynamic());
}

VOID CSLIDEMeshVertex::UpdateState()
{
  Vector v;

  switch ( m_pCSLIDEMeshVertexProperty->GetType() )
    {
    case slidevxtSLF_VERTEX_CORNER:
      m_pQEVertex->SetSharpVertexType(QE::svtCorner);
      break;
    default:
      break;
    }
  m_pt.Update();

  if ( m_vNormal.GetDynamic() )
    {
      m_vNormal.Update();
      m_vNormal.CompletePlaneEquation(m_pt);
    }

  m_ptTexture.Update();

  m_pQEVertex->SetPoint(m_pt);
  m_pQEVertex->SetNormal(m_vNormal);
  v = m_ptTexture;
  m_pQEVertex->SetTexture(v);
}

VOID CSLIDEMeshVertex::WriteSLIDE(FILE *fpOutputFile)
{
  Vector v;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s ( %.3f %.3f %.3f )\n", 
	  gcppcSLIDEReservedWords[sliderwtMeshVertex], m_pcGUIID, 
	  m_pt[X], m_pt[Y], m_pt[Z]);

  if ( m_vNormal.Magnitude() != 0.0 )
    {
      fprintf(fpOutputFile, "  %s ( %.3f %.3f %.3f )\n",
	      gcppcSLIDEReservedWords[sliderwtNormal], 
	      m_vNormal[X], m_vNormal[Y], m_vNormal[Z]); 
    }
  v = m_ptTexture;
  if ( v.Magnitude() != 0.0 )
    {
      fprintf(fpOutputFile, "  %s ( %.3f %.3f %.3f )\n",
	      gcppcSLIDEReservedWords[sliderwtTexture], 
	      m_ptTexture[0], m_ptTexture[1], m_ptTexture[2]); 
    }
  CSLIDEPrimitive::WriteSLIDEFlags(fpOutputFile, 0);

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndMeshVertex]); 
}

