#ifndef __SLIDEMESHVERTEX_H__
#define __SLIDEMESHVERTEX_H__

#include "slideprimitive.h"

class CSLIDEMesh;
class CSLIDEMeshVertexProperty;
class QEMesh;
class QEVertex;

//////////////////////////////////////////////////////////////////////
// CSLIDEMeshVertex Class
//

class CSLIDEMeshVertex : public CSLIDEPrimitive
{
//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEMeshVertex();
  virtual ~CSLIDEMeshVertex();

  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID, 
		    DynPoint *pdynpt);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  inline VOID SetVertexPropertyID(CHAR *pcVertexPropertyID);
  inline VOID SetPoint(DynPoint *pdynpt);
  inline VOID SetNormal(DynVector *pdynvNormal);
  inline VOID SetTexture(DynPoint *pdynptTexture);

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld, CSLIDEMesh *pCSLIDEMesh, 
			      QEMesh *pQEMesh);
  virtual BOOL CheckValence();

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  VOID Preprocess();
  VOID InsertIntoBound(CBound *pCBound);

#ifdef __USE_CDA__
  VOID CalculateCDAVertices(CListInt *plistCDAVertex, 
			    const Matrix &mGroot_obj);
#endif // __USE_CDA__

protected:
  VOID CheckDynamic();
  VOID UpdateState();

//////////////////////////////////////////////////
// Rendering
//
public:

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline SLIDEReservedWordType GetType();

  inline CSLIDEMeshVertexProperty *GetVertexProperty();
  inline const Point &GetPoint();
  inline const Vector &GetNormal();
  inline const Point &GetTexture();

  inline QEVertex *GetQEVertex();

  inline BOOL GetBoundDynamic();
  inline VOID SetBoundDynamic(BOOL bBoundDynamic);
  inline BOOL GetGeometryDynamic();
  inline VOID SetGeometryDynamic(BOOL bGeometryDynamic);
  inline BOOL GetNormalDynamic();
  inline VOID SetNormalDynamic(BOOL bNormalDynamic);

//////////////////////////////////////////////////
// Class Variables
//
public:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CHAR *m_pcMeshVertexPropertyID;
  CSLIDEMeshVertexProperty *m_pCSLIDEMeshVertexProperty;

  DynPoint m_pt;
  DynVector m_vNormal;
  DynPoint m_ptTexture;

  BOOL m_bBoundDynamic;
  BOOL m_bGeometryDynamic;
  BOOL m_bNormalDynamic;

  QEVertex *m_pQEVertex;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline SLIDEReservedWordType CSLIDEMeshVertex::GetType()
{
  return sliderwtMeshVertex;
}

inline VOID CSLIDEMeshVertex::SetVertexPropertyID(CHAR *pcVertexPropertyID)
{
  m_pcMeshVertexPropertyID = pcVertexPropertyID;
}

inline CSLIDEMeshVertexProperty *CSLIDEMeshVertex::GetVertexProperty()
{
  return m_pCSLIDEMeshVertexProperty;
}

inline VOID CSLIDEMeshVertex::SetPoint(DynPoint *pdynpt)
{
  m_pt.SetAll(pdynpt);
}

inline VOID CSLIDEMeshVertex::SetNormal(DynVector *pdynvNormal)
{
  m_vNormal.SetAll(pdynvNormal);
}

inline VOID CSLIDEMeshVertex::SetTexture(DynPoint *pdynptTexture)
{
  m_ptTexture.SetAll(pdynptTexture);
}

inline const Point &CSLIDEMeshVertex::GetPoint()
{
  return m_pt;
}

inline const Vector &CSLIDEMeshVertex::GetNormal()
{
  return m_vNormal;
}

inline const Point &CSLIDEMeshVertex::GetTexture()
{
  return m_ptTexture;
}

inline QEVertex *CSLIDEMeshVertex::GetQEVertex()
{
  return m_pQEVertex;
}

inline BOOL CSLIDEMeshVertex::GetBoundDynamic()
{
  return m_bBoundDynamic;
}

inline VOID CSLIDEMeshVertex::SetBoundDynamic(BOOL bBoundDynamic)
{
  m_bBoundDynamic = bBoundDynamic;
}

inline BOOL CSLIDEMeshVertex::GetGeometryDynamic()
{
  return m_bGeometryDynamic;
}

inline VOID CSLIDEMeshVertex::SetGeometryDynamic(BOOL bGeometryDynamic)
{
  m_bGeometryDynamic = bGeometryDynamic;
}

inline BOOL CSLIDEMeshVertex::GetNormalDynamic()
{
  return m_bNormalDynamic;
}

inline VOID CSLIDEMeshVertex::SetNormalDynamic(BOOL bNormalDynamic)
{
  m_bNormalDynamic = bNormalDynamic;
}

#endif // __SLIDEMESHVERTEX_H__


