#ifndef __SLIDENODE_H__
#define __SLIDENODE_H__

#include "slideprimitive.h"

class CSLIDEWorld;
class CSLIDECamera;
class CSLIDELight;
class CSLIDEAreaLight;

class QEMesh;

//////////////////////////////////////////////////////////////////////
// CSLIDENode Class
//

class CSLIDENode : public CSLIDEPrimitive
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//
public:

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDENode();
  virtual ~CSLIDENode();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

  virtual VOID Remove(CSLIDEWorld *pCSLIDEWorld);

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual CSLIDECamera *FindCamera(CListIterInt *piterIDs, 
				   CListInt *plistCSLIDEInstance) = 0;
  virtual CSLIDELight *FindLight(CListIterInt *piterIDs, 
				 CListInt *plistCSLIDEInstance) = 0;
  virtual CSLIDEAreaLight *FindAreaLight(CListIterInt *piterIDs, 
					 CListInt *plistCSLIDEInstance) = 0;
  virtual VOID GetSide(CListInt *plistPoint, CListInt *plistNormal, UINT_32 uiSide, 
		       const Matrix &mTransform, const Matrix &mInverse, BOOL bReverse);

  virtual VOID PreprocessGraph(); // Runs through the Scene Graph once to
				  // establish which subtrees are dynamic
  virtual VOID UpdateGraph(); // Runs through Scene Graph
                              // and recalculates any changed bounds
  virtual VOID InsertIntoBound(const Matrix &m, CBound &bound);

#ifdef __USE_CDA__
  virtual inline CDA *GetCDA();
  virtual VOID CalculateCDA();
  virtual VOID CalculateCDAVertices(CListInt *plistCDAVertex, 
				    const Matrix &mGroot_obj);
#endif // __USE_CDA__

  virtual QEMesh *GetQEMesh();
  virtual VOID CalculateLeafQEMesh();  // for WriteWRLZ
  virtual VOID DeallocateLeafQEMesh();  // for WriteWRLZ

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID Prerender();

  virtual VOID RecallGeometry(SLIDEShadingType eShading);
  virtual VOID RecallNormals(SLIDEShadingType eShading);
  virtual VOID RecallBounds();

  virtual VOID Render(const Matrix &mGproj_obj,
		      CBound *pCBoundClip, 
		      SLIDEShadingType eShading, 
		      SLIDELODType eLOD);

  virtual VOID RenderGeometry(const Matrix &mGproj_obj, 
				CBound * pCBoundClip, 
				SLIDEShadingType eShading, 
				SLIDELODType eLOD,
				CSLIDEWormwarp *pCSLIDEWarp = NULL);

  virtual VOID RenderSelect(const Matrix &mGproj_obj,
			    CBound *pCBoundClip, 
			    SLIDEShadingType eShading, 
			    SLIDELODType eLOD);

  virtual VOID RenderBound();
  virtual VOID RenderGUISelectedBound();
  virtual VOID RenderGraphSelectedBound();
  virtual VOID RenderGUISelectedAxes(BOOL bSelected, BOOL bXAxisSelected,
				     BOOL bYAxisSelected, BOOL bZAxisSelected);

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile) = 0;
  virtual VOID WriteRIB(FILE *fpOutputFile,
			const Matrix &mGproj_obj,
			CBound *pCBoundClip, 
			SLIDEShadingType eShading,
			SLIDELODType eLOD);
  virtual VOID WritePOV(FILE *fpOutputFile, 
			CListInt *plistTransformIDs,
			CListInt *plistColors); 
  virtual VOID WriteWRLZ(FILE *fpOutputFile, const Matrix &mTransform,
			 CSLIDESurface *pCSLIDESurface);

  virtual VOID WriteSTL(FILE *fpOutputFile, const Matrix &mTransform);
  virtual QEMesh *WriteSIF(FILE *fpOutputFile, UINT_32 uiIndentLevel, 
			   const Matrix &mGroot_obj, const Matrix &mGobj_root,
			   QEMesh *pmeBoundaryEdges);
  virtual VOID WriteLSIF(FILE *fpOutputFile, UINT_32 uiIndentLevel); 
  virtual VOID WriteLSIFOffset(FILE *fpOutputFile, UINT_32 uiIndentLevel); 

//////////////////////////////////////////////////
// Class Accessor Functions
//
public:

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline BOOL GetVisited();
  virtual inline VOID SetVisited(BOOL bVisited);
  virtual inline BOOL GetPrerenderVisited();
  virtual inline VOID SetPrerenderVisited(BOOL bPrerenderVisited);

  virtual inline BOOL GetGeometryEmpty();
  virtual inline BOOL GetGeometryDynamic();
  virtual inline BOOL GetNormalsEmpty();
  virtual inline BOOL GetNormalsDynamic();
  virtual inline BOOL GetBoundsEmpty();
  virtual inline BOOL GetBoundsDynamic();

  virtual inline BOOL GetBoundEmpty();
  virtual inline BOOL GetBoundDynamic();
  virtual inline BOOL GetBoundValid();
  virtual inline BOOL GetBoundChanged();

  virtual inline UINT_32 GetPrerenderAttributesStackDepthMax();

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline BOOL CSLIDENode::GetVisited()
{
  return TRUE;
}

inline VOID CSLIDENode::SetVisited(BOOL /*bVisited*/)
{
}

inline BOOL CSLIDENode::GetPrerenderVisited()
{
  ASSERT( FALSE );
  return TRUE;
}

inline VOID CSLIDENode::SetPrerenderVisited(BOOL /*bPrerenderVisited*/)
{
}

inline BOOL CSLIDENode::GetGeometryEmpty()
{
  return TRUE;
}

inline BOOL CSLIDENode::GetGeometryDynamic()
{
  return FALSE;
}

inline BOOL CSLIDENode::GetNormalsEmpty()
{
  return TRUE;
}

inline BOOL CSLIDENode::GetNormalsDynamic()
{
  return FALSE;
}

inline BOOL CSLIDENode::GetBoundsEmpty()
{
  return TRUE;
}

inline BOOL CSLIDENode::GetBoundsDynamic()
{
  return FALSE;
}

inline BOOL CSLIDENode::GetBoundEmpty()
{
  return TRUE;
}

inline BOOL CSLIDENode::GetBoundDynamic()
{
  return FALSE;
}

inline BOOL CSLIDENode::GetBoundValid()
{
  return TRUE;
}

inline BOOL CSLIDENode::GetBoundChanged()
{
  return FALSE;
}

inline UINT_32 CSLIDENode::GetPrerenderAttributesStackDepthMax()
{
  return 0;
}

#ifdef __USE_CDA__
inline CDA *CSLIDENode::GetCDA()
{
  return NULL;
}
#endif // __USE_CDA__

#endif // __SLIDENODE_H__




