#ifndef __SLIDENURBPATCH_H__
#define __SLIDENURBPATCH_H__

#include "slidebsplinepatch.h"

//////////////////////////////////////////////////////////////////////////////
// CSLIDENurbPatch Class
//

class CSLIDENurbPatch : public CSLIDEBSplinePatch
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDENurbPatch();
  virtual ~CSLIDENurbPatch();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();    
  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);
  virtual BOOL Valid();

protected:
  VOID ReallocateWDerivativeCurve();
  VOID DeallocateWDerivativeCurve();
  VOID ReallocateCurveWDerivativeTable();
  VOID DeallocateCurveWDerivativeTable();

//////////////////////////////////////////////////
// Geometric Computations
//
public:

#ifdef __USE_CDA__
  virtual VOID CalculateCDAVertices(CListInt *plistCDAVertex, const Matrix &mGroot_obj);
#endif // __USE_CDA__

protected:
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();

//////////////////////////////////////////////////
// Rendering
//
protected:
  virtual VOID RenderControls();

//////////////////////////////////////////////////
// Geometric Computations
//

public:
  virtual VOID PointAt(Point &pt, FLOAT fU, FLOAT fV);
  VOID ControlWeightAt();

protected:
  virtual VOID CalculateBound();
  virtual VOID CalculateSurface();

  virtual VOID NormalAt(Vector &v, FLOAT fU, FLOAT fV);
  virtual VOID CurveTangentAt(Vector &vT, FLOAT fU, FLOAT fV, BOOL bUDirection);

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

//////////////////////////////////////////////////
// Accessor Functions
//
public:

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:

  FLOAT *m_pfWDerivativeCurve;
  FLOAT **m_ppfCurveWDerivativeTable;
  UINT_32 m_uiCurveWDerivativeControlPoints;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

#endif // __SLIDENURBPATCH_H__
