#ifndef __SLIDEOBJECT_H__
#define __SLIDEOBJECT_H__

#include "slideatomnode.h"

class CSLIDEFace;

//////////////////////////////////////////////////////////////////////////////
// CSLIDEObject Class
//

class CSLIDEObject : public CSLIDEAtomNode
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEObject();
  virtual ~CSLIDEObject();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID, 
		    CListInt *plistFaceIDs);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  //inline VOID SetSolid(DynInt *pdyniSolid);

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

  virtual VOID Remove(CSLIDEWorld *pCSLIDEWorld);
  VOID RemoveFace(CSLIDEFace *pCSLIDEFace);

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual CSLIDECamera *FindCamera(CListIterInt *piterIDs, 
				   CListInt *plistCSLIDEInstance);
  virtual CSLIDELight *FindLight(CListIterInt *piterIDs, 
				 CListInt *plistCSLIDEInstance);
  virtual CSLIDEAreaLight *FindAreaLight(CListIterInt *piterIDs, 
					 CListInt *plistCSLIDEInstance);

  virtual VOID Preprocess();
  virtual VOID PreprocessGraph();
  VOID DrawGUIGraphFaceReferences(Tcl_Interp *pInterp);

  virtual VOID Update();
  virtual VOID UpdateGraph();

#ifdef __USE_CDA__
  virtual VOID CalculateCDAVertices(CListInt *plistCDAVertex, 
				    const Matrix &mGroot_obj);
#endif // __USE_CDA__

  virtual VOID CalculateQEMesh();
  virtual VOID CalculateLeafQEMesh();  // for WriteWRLZ
  virtual VOID DeallocateLeafQEMesh();  // for WriteWRLZ

protected:
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();
  virtual VOID CalculateBound();

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID PrerenderGeometry(SLIDEShadingType eShading);
  virtual VOID PrerenderNormals(SLIDEShadingType eShading);

  virtual VOID RenderGeometry(const Matrix &mGproj_obj,
			      CBound *pCBoundClip, 
			      SLIDEShadingType eShading,
			      SLIDELODType eLOD,
			      CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderGeometrySelect(const Matrix &mGproj_obj,
				    CBound *pCBoundClip, 
				    SLIDEShadingType eShading,
				    SLIDELODType eLOD);
  virtual VOID RenderNormals(SLIDEShadingType eShading);

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);

  BOOL CheckRIBPrimitive();
  VOID WriteRIBGeometryPolygonal(FILE *fpOutputFile,
				 SLIDEShadingType eShading,
				 SLIDELODType eLOD);
  VOID WriteRIBGeometryPrimitive(FILE *fpOutputFile,
				 SLIDEShadingType eShading,
				 SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

  virtual VOID WritePOVGeometry(FILE *fpOutputFile, 
				CListInt *plistTransformIDs,
				CListInt *plistColors); 

  virtual VOID WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				 CSLIDESurface *pCSLIDESurface);
  virtual VOID WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform);

  virtual VOID WriteDotID(FILE *fpOutputFile);
  virtual VOID WriteDotNode(FILE *fpOutputFile);
  virtual VOID WriteDotEdges(FILE *fpOutputFile);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline CHAR *GetTypeString();
  virtual inline SLIDEReservedWordType GetType();

  UINT_32 GetFaceCount();

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  //DynInt m_dyniSolid;
  CListInt *m_plistFaceIDs;
  CListInt m_listCSLIDEFace;

#ifdef DT_TRIANGULATE_TEST
  CListInt m_listDTVertex;
  CListInt m_listDTEdges;
  CListInt m_listDTContours;
  CListInt m_listDTTriangles;
#endif // DT_TRIANGULATE_TEST
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline SLIDEReservedWordType CSLIDEObject::GetType()
{
  return sliderwtObject;
}

inline CHAR *CSLIDEObject::GetTypeString()
{
  return gcppcSLIDEReservedWords[sliderwtObject];
}

/*
inline VOID CSLIDEObject::SetSolid(DynInt *pdyniSolid)
{
  m_dyniSolid.Set(pdyniSolid);
}
*/
#endif // __SLIDEOBJECT_H__
