#ifndef __SLIDEOFFSET_H__
#define __SLIDEOFFSET_H__

#include "slidegroup.h"

class CSLIDEWorld;

class QEMesh;

//////////////////////////////////////////////////////////////////////
// CSLIDEOffset Class
//

class CSLIDEOffset : public CSLIDEGroup
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEOffset();
  virtual ~CSLIDEOffset();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  inline VOID SetDrawControlMesh(DynInt *pdyniDrawControlMesh);
  inline VOID SetType(DynInt *pdyniType);
  inline VOID SetOffset(DynFloat *pdynfOffset);
  inline VOID SetWidth(DynFloat *pdynfWidth);

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual CSLIDECamera *FindCamera(CListIterInt *piterIDs, 
				   CListInt *plistCSLIDEInstance);
  virtual CSLIDELight *FindLight(CListIterInt *piterIDs, 
				 CListInt *plistCSLIDEInstance);
  virtual CSLIDEAreaLight *FindAreaLight(CListIterInt *piterIDs, 
					 CListInt *plistCSLIDEInstance);

  virtual VOID Preprocess();
  virtual VOID PreprocessGraph();

  virtual VOID Update();
  virtual VOID UpdateGraph();

  virtual QEMesh *GetQEMesh();

protected:
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();
  virtual VOID UpdateQEMesh();
  virtual VOID UpdateOffset();
  virtual VOID CalculateBound();

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID PrerenderGeometry(SLIDEShadingType eShading);
  virtual VOID PrerenderNormals(SLIDEShadingType eShading);
  virtual VOID PrerenderBounds();

  virtual VOID Prerender();

  virtual VOID RenderGeometry(const Matrix &mGproj_obj,
			      CBound *pCBoundClip, 
			      SLIDEShadingType eShading,
			      SLIDELODType eLOD,
			      CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderNormals(SLIDEShadingType eShading);

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

  virtual VOID WriteWRLZ(FILE *fpOutputFile, const Matrix &mTransform,
			 CSLIDESurface *pCSLIDESurface);
  virtual VOID WriteSTL(FILE *fpOutputFile, const Matrix &mTransform);

  virtual QEMesh *WriteSIF(FILE *fpOutputFile, UINT_32 uiIndentLevel, 
			   const Matrix &mGroot_obj, const Matrix &mGobj_root,
			   QEMesh *pmeBoundaryEdges);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline BOOL GetChildrenBoundsDynamic();
  inline VOID SetChildrenBoundsDynamic(BOOL bChildrenBoundsDynamic);

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  DynInt m_dyniDrawControlMesh;
  DynInt m_dyniType;
  DynFloat m_dynfOffset;
  DynFloat m_dynfWidth;

  QEMesh *m_pmeOffset;

  BOOL m_bChildrenBoundsDynamic;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDEOffset::SetDrawControlMesh(DynInt *pdyniDrawControlMesh)
{
  m_dyniDrawControlMesh.Set(pdyniDrawControlMesh);
}

inline VOID CSLIDEOffset::SetType(DynInt *pdyniType)
{
  m_dyniType.Set(pdyniType);
}

inline VOID CSLIDEOffset::SetOffset(DynFloat *pdynfOffset)
{
  m_dynfOffset.Set(pdynfOffset);
}

inline VOID CSLIDEOffset::SetWidth(DynFloat *pdynfWidth)
{
  m_dynfWidth.Set(pdynfWidth);
}

inline BOOL CSLIDEOffset::GetChildrenBoundsDynamic()
{
  return m_bChildrenBoundsDynamic;
}

inline VOID CSLIDEOffset::SetChildrenBoundsDynamic(BOOL bChildrenBoundsDynamic)
{
  m_bChildrenBoundsDynamic = bChildrenBoundsDynamic;
}

#endif // __SLIDEOFFSET_H__
