#ifndef __SLIDEPARSER_H__
#define __SLIDEPARSER_H__

#include "parser.h"
#include "dynamics.h"
#include "mathlib.h"
#include "slide.h"
#include "oot.h"
#include "hashstring.h"

class CSLIDEWorld;
class CSLIDEPrimitive;
class CSLIDERender;
class CWindow;
class CSLIDEViewport;
class CSLIDELight;
class CSLIDELightPath;
class CSLIDECamera;
class CSLIDECameraPath;
class CSLIDEGeometryNode;
class CSLIDEAtomNode;
class CSLIDEAreaLight;
class CSLIDEAreaLightPath;
class CSLIDEGroup;
class CSLIDESeam;
class CSLIDEContour;
class CSLIDESubdivision;
class CSLIDEOffset;
class CSLIDEDual;
class CSLIDEEvolver;
class CSLIDEMirror;
class CSLIDEGeodesic;
class CSLIDEUnion;
class CSLIDEIntersection;
class CSLIDEDifference;
class CSLIDESGRoot;
class CSLIDEInstance;
class CSLIDETransform;
class CSLIDETranslate;
class CSLIDEScale;
class CSLIDERotate;
class CSLIDELookAt;
class CSLIDEObject;
class CSLIDEMesh;
class CSLIDEMeshVertexProperty;
class CSLIDEMeshEdgeProperty;
class CSLIDEMeshVertex;
class CSLIDEMeshEdge;
class CSLIDEMeshFacet;
class CSLIDEConvexHull;
class CSLIDEGeneralFace;
class CSLIDETerrain;
class CSLIDESphere;
class CSLIDECylinder;
class CSLIDECone;
class CSLIDETorus;
class CSLIDECyclide;
class CSLIDEPolyline;
class CSLIDEBSplinePatch;
class CSLIDEHeightfield;
class CSLIDENurbPatch;
class CSLIDEBezierPatch;
class CSLIDEBSplineCurve;
class CSLIDEPolyline;
class CSLIDECrossSection;
class CSLIDESweep;
class CSLIDESweepMorph;
class CSLIDESweepCrossSection;
class CSLIDEWormwarp;
class CSLIDEFace;
class CSLIDEPoint;
class CSLIDEControlPoint;
class CSLIDESurface;
class CSLIDEFog;
class Matrix;

//////////////////////////////////////////////////////////////////////////////
// CSLIDEParser Class
//


/////////////////////////////////////////////////////////////////////
// SLIDE API:   "slide command [args]"

const int cuiCSLIDEParserOotMethods = 4;

#define CSLIDEPARSER_METHODS \
  {"slide", OOT_PARAM_TYPE_OTHER, CSLIDEParser::CBHandleSLIDECmd}, \
  {"slideUnselect", OOT_PARAM_TYPE_OTHER, CSLIDEParser::CBHandleGUIUnselect}, \
  {"slideGetParameters", OOT_PARAM_TYPE_OTHER, CSLIDEParser::CBHandleGUIGetParameters}, \
  {"slideSetParameters", OOT_PARAM_TYPE_OTHER, CSLIDEParser::CBHandleGUISetParameters}

  // slide api
const int cuiCSLIDEAPICommands = 3;

#define CSLIDEAPI_COMMANDS \
  "create", \
  "delete",\
  "modify"

#define CSLIDEAPI_CMDMETHODS \
  &CSLIDEParser::HandleCreate, \
  &CSLIDEParser::HandleDelete,\
  &CSLIDEParser::HandleModify


  // slide create
  const int cuiCSLIDECreateCommands = 29;

#define CSLIDECREATE_COMMANDS \
  "point", \
  "face",  \
  "object", \
  "group", \
  "instance", \
  "surface", \
  "convexhull", \
  "generalface", \
  "contour", \
  "terrain", \
  "sgroot", \
  "sphere", \
  "cone", \
  "cylinder",  \
  "torus", \
  "cyclide", \
  "polyline",\
  "bsplinecurve",\
  "beziercurve",\
  "sweep",\
  "bsplinepatch",\
  "bezierpatch", \
  "render", \
  "camera", \
  "light", \
  "arealight", \
  "window", \
  "viewport", \
  "fog"

#define CSLIDECREATE_CMDMETHODS \
  &CSLIDEParser::HandleAddPoint, \
  &CSLIDEParser::HandleAddFace,  \
  &CSLIDEParser::HandleAddObject, \
  &CSLIDEParser::HandleAddGroup, \
  &CSLIDEParser::HandleAddInstance, \
  &CSLIDEParser::HandleAddSurface, \
  &CSLIDEParser::HandleAddConvexHull, \
  &CSLIDEParser::HandleAddGeneralFace, \
  &CSLIDEParser::HandleAddContour, \
  &CSLIDEParser::HandleAddTerrain, \
  &CSLIDEParser::HandleAddSGRoot, \
  &CSLIDEParser::HandleAddSphere, \
  &CSLIDEParser::HandleAddCone, \
  &CSLIDEParser::HandleAddCylinder, \
  &CSLIDEParser::HandleAddTorus, \
  &CSLIDEParser::HandleAddCyclide, \
  &CSLIDEParser::HandleAddPolyline, \
  &CSLIDEParser::HandleAddBSplineCurve, \
  &CSLIDEParser::HandleAddBezierCurve, \
  &CSLIDEParser::HandleAddSweep, \
  &CSLIDEParser::HandleAddBSplinePatch, \
  &CSLIDEParser::HandleAddBezierPatch, \
  &CSLIDEParser::HandleAddRender, \
  &CSLIDEParser::HandleAddCamera, \
  &CSLIDEParser::HandleAddLight, \
  &CSLIDEParser::HandleAddAreaLight, \
  &CSLIDEParser::HandleAddWindow, \
  &CSLIDEParser::HandleAddViewport, \
  &CSLIDEParser::HandleAddFog


///////////////////////////
// slide modify
const int cuiCSLIDEModifyCommands = 2;

#define CSLIDEMODIFY_COMMANDS \
  "point", \
  "polyline"

#define CSLIDEMODIFY_CMDMETHODS \
  &CSLIDEParser::HandleModifyPoint, \
  &CSLIDEParser::HandleModifyPolyline


// SLIDE API
/////////////////////////////////////////////////////////////////////

class CSLIDEParser : public CParser
{
//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEParser();
  virtual ~CSLIDEParser();
  virtual VOID Init(CHAR *pcFile, CHAR **ppcDirectories, UINT_32 uiDirectories, 
		    CSLIDEParser **ppyy_CSLIDEParser, 
		    YYLTYPE *pyylloc, YY_PARSE_TYPE *pfyy_Parse,
		    CScanner **ppyy_CScanner,
		    int *pyyleng,
		    YY_BEGIN_INITIAL_TYPE *pfyy_BeginInitial,
		    YY_GET_CURRENT_BUFFER_TYPE *pfyy_GetCurrentBuffer,
		    YY_CREATE_BUFFER_TYPE *pfyy_CreateBuffer,
		    YY_DELETE_BUFFER_TYPE *pfyy_DeleteBuffer,
		    YY_SWITCH_TO_BUFFER_TYPE *pfyy_SwitchToBuffer,
		    YY_SET_VALUE_INT_TYPE *pfyy_SetValueInt,
		    YY_SET_VALUE_FLOAT_TYPE *pfyy_SetValueFloat,
		    YY_SET_VALUE_STRING_TYPE *pfyy_SetValueString);
  virtual VOID Uninit();

//////////////////////////////////////////////////
// AST Functions
//
public:
  virtual VOID SetPrimitiveRIBBegin(CHAR *pcRIBBegin);
  virtual VOID SetPrimitiveRIBEnd(CHAR *pcRIBEnd);
  virtual VOID SetPrimitiveGraphPosition(DynInt *pdyniX, DynInt *pdyniY);

  virtual VOID AddRender(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			 CHAR *pcViewportID, CListInt *plistCameraPath, 
			 CHAR *pcRootID);
  virtual VOID EndRender();
  virtual VOID SetRenderTclTag(CHAR *pcTclTag);
  virtual VOID SetRenderMergeEpsilon(DynFloat *pdynfMergeEpsilon);
  virtual VOID SetRenderLOD(DynInt *pdyniLOD);
  virtual VOID SetRenderCUIType(DynInt *pdyniCUIType);
  virtual VOID SetRenderCUISpeed(DynFloat *pdynfCUISpeed);
  virtual VOID SetRenderStencil(DynInt *pdyniStencil);
  virtual VOID SetRenderFogID(CHAR *pcSurfaceID, DynInt *pdyniOnOff);
  virtual VOID SetBlendFlag(DynInt *pdyniOnOff);
  virtual VOID AddRenderLightPath(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
				  CListInt *plistLightPath, DynInt *pdyniOnOff);
  virtual VOID AddRenderAreaLightPath(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
				      CListInt *plistAreaLightPath, DynInt *pdyniOnOff);
  virtual VOID AddWindow(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			  CHAR *pcID);
  virtual VOID EndWindow();
  virtual VOID SetWindowOrigin(FLOAT fX, FLOAT fY);
  virtual VOID SetWindowSize(FLOAT fX, FLOAT fY);
  virtual VOID SetWindowResize(INT_32 iResize);
  virtual VOID SetWindowAspectRatio(INT_32 iAspectRatio);
  virtual VOID SetWindowBackground(DynVector *pdynv);
  virtual VOID SetWindowDisplay(CHAR *pcDisplayID);
  virtual VOID AddViewport(CHAR *pcFile, UINT_32 uiLine, 
			   UINT_32 uiColumn,
			   CHAR *pcID, CHAR *pcWindowID);
  virtual VOID EndViewport();
  virtual VOID SetViewportOrigin(DynVector *pdynv);
  virtual VOID SetViewportSize(DynVector *pdynv);
  virtual VOID AddCamera(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			  CHAR *pcID);
  virtual VOID EndCamera();
  virtual VOID SetCameraProjection(DynInt *pdyniProjection);
  virtual VOID SetCameraFrustum(DynVector *pdynvMin, 
				DynVector *pdynvMax);
  virtual VOID AddLight(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			CHAR *pcID);
  virtual VOID EndLight();
  virtual VOID SetLightType(DynInt *pdyniType);
  virtual VOID SetLightColor(DynVector *pdynvColor);
  virtual VOID SetLightDeadDistance(DynFloat *pdynfDeadDistance);
  virtual VOID SetLightFallOff(DynFloat *pdynfFallOff);
  virtual VOID SetLightAngularFallOff(DynFloat *pdynfAngularFallOff);
  virtual VOID SetNodeSurfaceID(CHAR *pcSurfaceID);
  virtual VOID SetNodeShading(DynInt *pdyniShading);
  virtual VOID SetNodeLOD(DynInt *pdyniLOD);
  virtual VOID SetAtomNodeSolid(DynInt *pdyniSolid);
  virtual VOID AddAreaLight(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID EndAreaLight();
  virtual VOID SetAreaLightColor(DynVector *pdynvColor);
  virtual VOID SetAreaLightSamples(DynInt *pdyniSamples);
  virtual VOID AddGroup(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,	CHAR *pcID);
  virtual VOID EndGroup();
  virtual VOID SelectGroup(CHAR *pcID);
  virtual VOID AddSeam(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID,
		       CHAR *pcContour1ID, CHAR *pcContour2ID);
  virtual VOID EndSeam();
  virtual VOID AddContour(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcChildID);
  virtual VOID EndContour();
  virtual VOID AddContourSegment(CHAR *pcID, DynInt *pdyniSide, BOOL bReverse);
  virtual VOID AddSubdivision(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			      CHAR *pcID);
  virtual VOID EndSubdivision();
  virtual VOID SetSubdivisionUseDisplayLists(DynInt *pdyniUseDisplayLists);
  virtual VOID SetSubdivisionDrawControlMesh(DynInt *pdyniDrawControlMesh);
  virtual VOID SetSubdivisionDrawCurrentMesh(DynInt *pdyniDrawCurrentMesh);
  virtual VOID SetSubdivisionDrawVertices(DynInt *pdyniDrawVertices);
  virtual VOID SetSubdivisionType(DynInt *pdyniType);
  virtual VOID SetSubdivisionErrorMetric(DynInt *pdyniErrorMetric);
  virtual VOID SetSubdivisionSubdivisionLevel(DynInt *pdyniSubdivisionLevel);
  virtual VOID SetSubdivisionFacetsMax(DynInt *pdyniFacetsMax);
  virtual VOID SetSubdivisionWeightedEdges(DynInt *pdyniWeightedEdges);
  virtual VOID SetSubdivisionWeight(DynFloat *pdynfWeight);
  virtual VOID SetSubdivisionTolerance(DynFloat *pdynfTolerance);
  virtual VOID SetSubdivisionA(DynInt *pdyniA);
  virtual VOID SetSubdivisionB(DynInt *pdyniB);
  virtual VOID AddOffset(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			 CHAR *pcID);
  virtual VOID EndOffset();
  virtual VOID SetOffsetDrawControlMesh(DynInt *pdyniDrawControlMesh);
  virtual VOID SetOffsetType(DynInt *pdyniType);
  virtual VOID SetOffsetRadius(DynFloat *pdynfRadius);
  virtual VOID SetOffsetWidth(DynFloat *pdynfWidth);
  virtual VOID AddDual(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			 CHAR *pcID);
  virtual VOID EndDual();
  virtual VOID SetDualDrawControlMesh(DynInt *pdyniDrawControlMesh);
  virtual VOID AddEvolver(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			 CHAR *pcID);
  virtual VOID EndEvolver();
  virtual VOID SetEvolverDrawControlMesh(DynInt *pdyniDrawControlMesh);
  virtual VOID SetEvolverDrawCurrentMesh(DynInt *pdyniDrawCurrentMesh);
  virtual VOID SetEvolverSubdivisionLevel(DynInt *pdyniSubdivisionLevel);
  virtual VOID AddMirror(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			 CHAR *pcID);
  virtual VOID EndMirror();
  virtual VOID SetMirrorDrawControlMesh(DynInt *pdyniDrawControlMesh);
  virtual VOID AddGeodesic(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			   CHAR *pcID);
  virtual VOID EndGeodesic();
  virtual VOID SetGeodesicDrawControlMesh(DynInt *pdyniDrawControlMesh);
  virtual VOID SetGeodesicTheta(DynFloat *pdynfTheta);
  virtual VOID SetGeodesicIntersections(DynInt *pdyniIntersections);
  virtual VOID AddUnion(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			CHAR *pcID);
  virtual VOID EndUnion();
  virtual VOID AddIntersection(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			       CHAR *pcID);
  virtual VOID EndIntersection();
  virtual VOID AddDifference(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			     CHAR *pcID);
  virtual VOID EndDifference();
  virtual VOID AddSGRoot(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			 CHAR *pcID, CHAR *pcSGRoot);
  virtual VOID EndSGRoot();
  virtual VOID AddInstance(CHAR *pcFile, UINT_32 uiLine, 
			   UINT_32 uiColumn, CHAR *pcChildID);
  virtual VOID EndInstance();
  virtual VOID SetInstanceID(CHAR *pcID);
  virtual VOID AddTranslate(CHAR *pcFile, UINT_32 uiLine, 
			    UINT_32 uiColumn,
			    DynVector *pdynv);
  virtual VOID AddScale(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			DynVector *pdynv);
  virtual VOID AddRotate(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			 DynVector *pdynvAxis, DynFloat *pdynfAngle);
  virtual VOID AddLookAt(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn);
  virtual VOID EndLookAt();
  virtual VOID SetLookAtEye(DynVector *pdynvEye);
  virtual VOID SetLookAtTarget(DynVector *pdynvTarget);
  virtual VOID SetLookAtUp(DynVector *pdynvUp);
/*
  virtual VOID AddGeneral(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			  FLOAT f0, FLOAT f1, FLOAT f2, FLOAT f3, 
			  FLOAT f4, FLOAT f5, FLOAT f6, FLOAT f7, 
			  FLOAT f8, FLOAT f9, FLOAT f10, FLOAT f11, 
			  FLOAT f12, FLOAT f13, FLOAT f14, FLOAT f15);
*/
  virtual VOID AddObject(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			 CHAR *pcID, 
			 CListInt *plistFaceIDs);
  virtual VOID EndObject();

  virtual VOID AddMesh(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID EndMesh();
  virtual VOID AddMeshVertexProperty(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
				   CHAR *pcID);
  virtual VOID EndMeshVertexProperty();
  virtual VOID SetMeshVertexPropertyType(DynInt *pdyni);
  virtual VOID AddMeshEdgeProperty(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
				   CHAR *pcID);
  virtual VOID EndMeshEdgeProperty();
  virtual VOID SetMeshEdgePropertySharpness(DynFloat *pdynf);
  virtual VOID AddMeshVertex(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			     CHAR *pcID, DynPoint *pdynpt);
  virtual VOID EndMeshVertex();
  virtual VOID SetMeshVertexVertexPropertyID(CHAR *pcVertexPropertyID);
  virtual VOID SetMeshVertexNormal(DynVector *pdynv);
  virtual VOID SetMeshVertexTexture(DynPoint *pdynpt);
  virtual VOID AddMeshFacet(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			     CHAR *pcID,
			     CListInt *plistMeshVertexIDs);
  virtual VOID EndMeshFacet();
  virtual VOID AddMeshEdge(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			     CHAR *pcID,
			     CHAR *pcMeshVertexID0, CHAR *pcMeshVertexID1);
  virtual VOID EndMeshEdge();
  virtual VOID SetMeshEdgeEdgePropertyID(CHAR *pcEdgePropertyID);

  virtual VOID AddConvexHull(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			     CHAR *pcID,
			     CListInt *plistPointIDs);
  virtual VOID EndConvexHull();
  virtual VOID SelectGeneralFace(CHAR *pcID);
  virtual VOID AddGeneralFace(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			  CHAR *pcID,
			  CListInt *plistPointIDs);
  virtual VOID EndGeneralFace();
  virtual VOID InsertGeneralFaceContour(CListInt *plistPointIDs);
  virtual VOID SetGeneralFaceType(DynInt *pdyniType);
  virtual VOID SetGeneralFaceVoronoi(DynInt *pdyniVoronoi);
  virtual VOID SetGeneralFaceAccuracy(DynInt *pdyniAccuracy);
  virtual VOID SetGeneralFaceRandom(DynInt *pdyniRandom);
  virtual VOID SetGeneralFaceDrawContours(DynInt *pdyniDrawContours);
  virtual VOID SetGeneralFaceDrawVertices(DynInt *pdyniDrawVertices);
  virtual VOID SetGeneralFaceDrawCDT(DynInt *pdyniDrawCDT);
  virtual VOID SetGeneralFaceDrawIndividualVoronoi(DynInt *pdyniDrawIndividualVoronoi);
  virtual VOID SetGeneralFaceDrawVoronoi(DynInt *pdyniDrawVoronoi);
  virtual VOID SetGeneralFaceDrawHeight(DynInt *pdyniDrawHeight);
  virtual VOID SetGeneralFaceDrawOffset(DynInt *pdyniDrawOffset);
  virtual VOID SetGeneralFaceDrawMillingContour(DynInt *pdyniDrawMillingContour);
  virtual VOID SetGeneralFaceMillingContour(DynInt *pdyniMillingContour);
  virtual VOID SetGeneralFaceCleanUpContour(DynInt *pdyniCleanUpContour);
  virtual VOID SetGeneralFaceOffsets(DynInt *pdyniOffsets);
  virtual VOID SetGeneralFaceOffsetMin(DynFloat *pdynfOffsetMin);
  virtual VOID SetGeneralFaceOffsetMax(DynFloat *pdynfOffsetMax);
  virtual VOID SetGeneralFaceIndividualVoronoiMin(DynInt *pdyniIndividualVoronoiMin);
  virtual VOID SetGeneralFaceIndividualVoronoiMax(DynInt *pdyniIndividualVoronoiMax);
  virtual VOID SetGeneralFaceSubdivisionLevel(DynInt *pdyniSubdivisionLevel);
  virtual VOID SetGeneralFaceTolerance(DynFloat *pdynfTolerance);
  virtual VOID AddTerrain(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			  CHAR *pcID,
			  CListInt *plistPointIDs);
  virtual VOID EndTerrain();
  virtual VOID SetTerrainType(DynInt *pdyniType);
  virtual VOID SetTerrainVoronoi(DynInt *pdyniVoronoi);
  virtual VOID SetTerrainAccuracy(DynInt *pdyniAccuracy);
  virtual VOID SetTerrainRandom(DynInt *pdyniRandom);
  virtual VOID SetTerrainDrawDelaunay(DynInt *pdyniDrawDelaunay);
  virtual VOID SetTerrainDrawVertices(DynInt *pdyniDrawVertices);
  virtual VOID AddSphere(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			 CHAR *pcID);
  virtual VOID EndSphere();
  virtual VOID SetSphereRadius(DynFloat *pdynfRadius);
  virtual VOID SetSphereZMin(DynFloat *pdynfZMin);
  virtual VOID SetSphereZMax(DynFloat *pdynfZMax);
  virtual VOID SetSphereThetaMax(DynFloat *pdynfThetaMax);
  virtual VOID SetSphereZSlices(DynInt *pdyniZSlices);
  virtual VOID SetSphereThetaSlices(DynInt *pdyniThetaSlices);
  virtual VOID SetSphereTexture(DynVector *pdynv0, DynVector *pdynv1, DynVector *pdynv2, DynVector *pdynv3);

  virtual VOID AddCylinder(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID EndCylinder();
  virtual VOID SetCylinderRadius(DynFloat *pdynfRadius);
  virtual VOID SetCylinderZMin(DynFloat *pdynfZMin);
  virtual VOID SetCylinderZMax(DynFloat *pdynfZMax);
  virtual VOID SetCylinderThetaMax(DynFloat *pdynfThetaMax);
  virtual VOID SetCylinderZSlices(DynInt *pdyniZSlices);
  virtual VOID SetCylinderThetaSlices(DynInt *pdyniThetaSlices);
  virtual VOID SetCylinderEndCap(DynInt *pdyniEndCap);
  virtual VOID SetCylinderBeginCap(DynInt *pdyniBeginCap);
  virtual VOID SetCylinderTexture(DynVector *pdynv0, DynVector *pdynv1, DynVector *pdynv2, DynVector *pdynv3);
  virtual VOID AddCone(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
		       CHAR *pcID);
  virtual VOID EndCone();
  virtual VOID SetConeHeight(DynFloat *pdynfHeight);
  virtual VOID SetConeZMin(DynFloat *pdynfZMin);
  virtual VOID SetConeZMax(DynFloat *pdynfZMax);
  virtual VOID SetConeRadius(DynFloat *pdynfRadius);
  virtual VOID SetConeThetaMax(DynFloat *pdynfThetaMax);
  virtual VOID SetConeZSlices(DynInt *pdyniZSlices);
  virtual VOID SetConeThetaSlices(DynInt *pdyniThetaSlices);
  virtual VOID SetConeBeginCap(DynInt *pdyniBeginCap);
  virtual VOID SetConeTexture(DynVector *pdynv0, DynVector *pdynv1, DynVector *pdynv2, DynVector *pdynv3);
  virtual VOID AddTorus(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			CHAR *pcID);
  virtual VOID EndTorus();
  virtual VOID SetTorusMajorRadius(DynFloat *pdynfMajorRadius);
  virtual VOID SetTorusMinorRadius(DynFloat *pdynfMinorRadius);
  virtual VOID SetTorusPhiMin(DynFloat *pdynfPhiMin);
  virtual VOID SetTorusPhiMax(DynFloat *pdynfPhiMax);
  virtual VOID SetTorusThetaMax(DynFloat *pdynfThetaMax);
  virtual VOID SetTorusPhiSlices(DynInt *pdyniPhiSlices);
  virtual VOID SetTorusThetaSlices(DynInt *pdyniThetaSlices);
  virtual VOID SetTorusTexture(DynVector *pdynv0, DynVector *pdynv1, DynVector *pdynv2, DynVector *pdynv3);

  virtual VOID AddCyclide(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID EndCyclide();
  virtual VOID SetCyclideXMajorA(DynFloat *pdynfXMajorA);
  virtual VOID SetCyclideYMinorB(DynFloat *pdynfYMinorB);
  virtual VOID SetCyclideStringLength(DynFloat *pdynfStringLength);    
  virtual VOID SetCyclidePhiMin(DynFloat *pdynfPhiMin);
  virtual VOID SetCyclidePhiMax(DynFloat *pdynfPhiMax);
  virtual VOID SetCyclideThetaMin(DynFloat *pdynfThetaMax);
  virtual VOID SetCyclideThetaMax(DynFloat *pdynfThetaMax);
  virtual VOID SetCyclidePhiSlices(DynInt *pdyniPhiSlices);
  virtual VOID SetCyclideThetaSlices(DynInt *pdyniThetaSlices);
  virtual VOID SetCyclideDrawEllipse(DynInt *pdyniDrawEllipse);
  virtual VOID SetCyclideDrawHyperbola(DynInt *pdyniDrawHyperbola);   
  virtual VOID SetCyclideTexture(DynVector *pdynv0, DynVector *pdynv1, DynVector *pdynv2, DynVector *pdynv3);

  virtual VOID AddBSplinePatch(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID EndBSplinePatch();
  virtual VOID SetBSplinePatchUOrder(DynInt *pdyniUOrder);
  virtual VOID SetBSplinePatchUKnots(DynInt *pdyniUKnots);
  virtual VOID SetBSplinePatchUKnotList(CListInt *plistUKnot);
  virtual VOID SetBSplinePatchVOrder(DynInt *pdyniVOrder);
  virtual VOID SetBSplinePatchVKnots(DynInt *pdyniVKnots);
  virtual VOID SetBSplinePatchVKnotList(CListInt *plistVKnot);
  virtual VOID SetBSplinePatchControlPointList(CListInt *plistControlPoint);
  virtual VOID SetBSplinePatchUSlices(DynInt *pdyniUSlices);
  virtual VOID SetBSplinePatchVSlices(DynInt *pdyniVSlices);
  virtual VOID SetBSplinePatchDrawControls(DynInt *pdyniDrawHyperbola);   
  virtual VOID SetBSplinePatchTexture(DynVector *pdynv0, DynVector *pdynv1, DynVector *pdynv2, DynVector *pdynv3);
  virtual VOID SetBSplinePatchOffset(DynFloat *pdynfOffset, DynInt *pdyniOffsetConnected);

  virtual VOID AddHeightfield(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID EndHeightfield();
  virtual VOID SetHeightfieldUMin(DynFloat *pdynfUMin);
  virtual VOID SetHeightfieldUMax(DynFloat *pdynfUMax);
  virtual VOID SetHeightfieldVMin(DynFloat *pdynfVMin);
  virtual VOID SetHeightfieldVMax(DynFloat *pdynfVMax);
  virtual VOID SetHeightfieldUSlices(DynInt *pdyniUSlices);
  virtual VOID SetHeightfieldVSlices(DynInt *pdyniVSlices);
  virtual VOID SetHeightfieldFunc(DynVector *pdynvFunc);

  virtual VOID AddNurbPatch(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID EndNurbPatch();
  virtual VOID SetNurbPatchUOrder(DynInt *pdyniUOrder);
  virtual VOID SetNurbPatchUKnots(DynInt *pdyniUKnots);
  virtual VOID SetNurbPatchUKnotList(CListInt *plistUKnot);
  virtual VOID SetNurbPatchVOrder(DynInt *pdyniVOrder);
  virtual VOID SetNurbPatchVKnots(DynInt *pdyniVKnots);
  virtual VOID SetNurbPatchVKnotList(CListInt *plistVKnot);
  virtual VOID SetNurbPatchControlPointList(CListInt *plistControlPoint);
  virtual VOID SetNurbPatchUSlices(DynInt *pdyniUSlices);
  virtual VOID SetNurbPatchVSlices(DynInt *pdyniVSlices);
  virtual VOID SetNurbPatchDrawControls(DynInt *pdyniDrawHyperbola);   
  virtual VOID SetNurbPatchTexture(DynVector *pdynv0, DynVector *pdynv1, DynVector *pdynv2, DynVector *pdynv3);
  virtual VOID SetNurbPatchOffset(DynFloat *pdynfOffset, DynInt *pdyniOffsetConnected);

  virtual VOID AddBezierPatch(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID EndBezierPatch();
  virtual VOID SetBezierPatchUOrder(DynInt *pdyniUOrder);
  virtual VOID SetBezierPatchVOrder(DynInt *pdyniVOrder);
  virtual VOID SetBezierPatchControlPointList(CListInt *plistControlPoint);
  virtual VOID SetBezierPatchUSlices(DynInt *pdyniUSlices);
  virtual VOID SetBezierPatchVSlices(DynInt *pdyniVSlices);
  virtual VOID SetBezierPatchDrawControls(DynInt *pdyniDrawHyperbola);   
  virtual VOID SetBezierPatchTexture(DynVector *pdynv0, DynVector *pdynv1, DynVector *pdynv2, DynVector *pdynv3);
  virtual VOID SetBezierPatchOffset(DynFloat *pdynfOffset, DynInt *pdyniOffsetConnected);

  virtual VOID AddBSplineCurve(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID EndBSplineCurve();
  virtual VOID SetBSplineCurveType(UINT_32 uiType);
  virtual VOID SetBSplineCurveOrder(DynInt *pdyniOrder);
  virtual VOID SetBSplineCurveKnotList(CListInt *plistKnot);
  virtual VOID SetBSplineCurveControlPointList(CListInt *plistControlPoint);
  virtual VOID SetBSplineCurveBezierControlPointList(CListInt *plistControlPoint);
  virtual VOID SetBSplineCurveClosed(DynInt *pdyniClosed);
  virtual VOID SetBSplineCurveSlices(DynInt *pdyniSlices);
  virtual VOID SetBSplineCurveDrawControls(DynInt *pdyniDrawControls);   

  virtual VOID AddPolyline(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID EndPolyline();
  virtual VOID SetPolylinePointList(CListInt *plistPoint);
  virtual VOID SetPolylineDrawPoints(DynInt *pdyniDrawPoints);

  virtual VOID AddCrossSection(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID EndCrossSection();
  virtual VOID SetCrossSectionSurfaceID(CHAR *pcSurfaceID);
  virtual VOID SetCrossSectionType(UINT_32 uiType);
  virtual VOID SetCrossSectionID(CHAR *pcID);
  virtual VOID SetCrossSectionSlices(DynInt *pdyniCrossSectionSlices);
  virtual VOID SetCrossSectionRadius(DynFloat *pdynfRadius);
  virtual VOID SetCrossSectionInnerRadius(DynFloat *pdynfInnerRadius);
  virtual VOID SetCrossSectionThickness(DynFloat *pdynfThickness);
  virtual VOID SetCrossSectionTheta(DynFloat *pdynfTheta);

  virtual VOID AddSweep(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID EndSweep();
  virtual VOID EndSweepCrossSection();
  virtual VOID SetSweepPathID(CHAR *pcID);
  virtual VOID SetSweepPathClosed(DynInt *pdyniClosed);
  virtual VOID SetSweepPathSymmetry(DynInt *pdyniSymmetry);
  virtual VOID SetSweepPathMinimizeTorsion(DynInt *pdyniMinimizeTorsion);
  virtual VOID SetSweepPathAzimuth(DynFloat *pdynfAzimuth);
  virtual VOID SetSweepPathTwist(DynFloat *pdynfTwist);
  virtual VOID AddSweepCrossSectionID(CHAR *pcID);
  virtual VOID SetSweepCrossSectionSurfaceID(CHAR *pcID);
  virtual VOID SetSweepCrossSectionTexture(DynVector *pdynv0, DynVector *pdynv1, 
					   DynVector *pdynv2, DynVector *pdynv3);
  virtual VOID SetSweepCrossSectionEndCap(DynInt *pdyniEndCap);
  virtual VOID SetSweepCrossSectionBeginCap(DynInt *pdyniBeginCap);
  virtual VOID SetSweepCrossSectionOrientation(DynInt *pdyniOrientation);
  virtual VOID AddSweepCrossSectionTranslate(CHAR *pcFile, UINT_32 uiLine, 
					     UINT_32 uiColumn,
					     DynVector *pdynv);
  virtual VOID AddSweepCrossSectionScale(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
					 DynVector *pdynv);
  virtual VOID AddSweepCrossSectionRotate(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
					  DynVector *pdynvAxis, DynFloat *pdynfAngle);
  virtual VOID SetSweepDrawSweep(DynInt *pdyniDrawSweep);   
  virtual VOID SetSweepDrawPath(DynInt *pdyniDrawPath);
  virtual VOID SetSweepMeshType(DynInt *pdyniMeshType);

  virtual VOID AddSweepMorph(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID EndSweepMorph();
  virtual VOID SetSweepMorphPathID(CHAR *pcID);
  virtual VOID SetSweepMorphPathClosed(DynInt *pdyniClosed);
  virtual VOID SetSweepMorphPathSymmetry(DynInt *pdyniSymmetry);
  virtual VOID SetSweepMorphPathMinimizeTorsion(DynInt *pdyniMinimizeTorsion);
  virtual VOID SetSweepMorphPathAzimuth(DynFloat *pdynfAzimuth);
  virtual VOID SetSweepMorphPathTwist(DynFloat *pdynfTwist);
  virtual VOID SetSweepMorphMeshType(DynInt *pdyniMeshType);
  virtual VOID SetSweepMorphDrawSweep(DynInt *pdyniDrawSweepMorph);   
  virtual VOID SetSweepMorphDrawPath(DynInt *pdyniDrawPath);
  virtual VOID SetSweepMorphEndCap(DynInt *pdyniEndCap);
  virtual VOID SetSweepMorphBeginCap(DynInt *pdyniBeginCap);

  virtual VOID AddFace(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
		       CHAR *pcID, CListInt *plistPointIDs);
  virtual VOID EndFace();
  virtual VOID SetFaceSurfaceID(CHAR *pcSurfaceID);
  virtual VOID AddPoint(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			CHAR *pcID, DynPoint *pdynpt);
  virtual VOID EndPoint();
  virtual VOID SetPointSurfaceID(CHAR *pcSurfaceID);
  virtual VOID SetPointNormal(DynVector *pdynv);
  virtual VOID SetPointTexture(DynPoint *pdynpt);

  virtual VOID AddControlPoint(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			CHAR *pcID);
  virtual VOID EndControlPoint();
  virtual VOID SetControlPointPointID(CHAR *pcPointID);
  virtual VOID SetControlPointCrossSectionID(CHAR *pcCrossSectionID);
  virtual VOID SetControlPointScale(DynVector *pdynv);
  virtual VOID SetControlPointRotate(DynVector *pdynv);

  virtual VOID AddSurface(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			  CHAR *pcID);
  virtual VOID EndSurface();
  virtual VOID SetSurfaceColor(DynVector *pdynv);
  virtual VOID SetSurfaceReflectivity(DynVector *pdynv);
  virtual VOID SetSurfaceExponent(DynFloat *pdynfExponent);
  virtual VOID SetSurfaceMetallic(DynFloat *pdynfMetallic);
  virtual VOID SetSurfaceTexture(CHAR *pcTextureFile, DynInt *pdyniTexture);
  virtual VOID AddWormwarp(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			  CHAR *pcID);
  virtual VOID EndWormwarp();
  virtual VOID SetWormwarpGroupID(CHAR *pcGroupID);
  virtual VOID SetWormwarpPathID(CHAR *pcPathID);
  virtual VOID SetWormwarpInstances(DynFloat *pdynfInstances);
  virtual VOID SetWormwarpType(DynInt *pdyniType);
  virtual VOID SetWormwarpPathOnOff(DynInt *pdyniPathOnOff);
  virtual VOID AddFog(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
			  CHAR *pcID);
  virtual VOID EndFog();
  virtual VOID SetFogType(DynInt *pdyniType);
  virtual VOID SetFogDensity(DynFloat *pdynfDensity);
  virtual VOID SetFogStart(DynFloat *pdynfStart);
  virtual VOID SetFogEnd(DynFloat *pdynfEnd);
  virtual VOID SetFogColor(DynVector *pdynv);
  virtual VOID AddGlobalInitTclBlock(CHAR *pcTclBlock);
  virtual VOID AddGlobalUpdateTclBlock(CHAR *pcTclBlock);
  virtual CListInt *CreateIDList();
  virtual CListInt *InsertIDList(CListInt *pCListInt, CHAR *pcID);
  virtual CListInt *CreateDynvalueList();
  virtual CListInt *InsertDynvalueList(CListInt *pCListInt, DynFloat *pdfValue);

  virtual DynInt *TranslateOnOffFlag(CHAR *pcFlag);
  virtual DynInt *TranslateSolidFlag(CHAR *pcFlag);
  virtual DynInt *TranslateShadingFlag(CHAR *pcFlag);
  virtual DynInt *TranslateLODFlag(CHAR *pcFlag);
  virtual DynInt *TranslateTextureFlag(CHAR *pcFlag);
  virtual DynInt *TranslateProjectionFlag(CHAR *pcFlag);
  virtual DynInt *TranslateLightFlag(CHAR *pcFlag);
  virtual DynInt *TranslateFogFlag(CHAR *pcFlag);
  virtual DynInt *TranslateCUIType(CHAR *pcCUIT);
  virtual DynInt *TranslateTerrainTypeFlag(CHAR *pcFlag);
  virtual DynInt *TranslateVoronoiFlag(CHAR *pcFlag);
  virtual DynInt *TranslateAccuracyFlag(CHAR *pcFlag);

protected:
  virtual CSLIDEWorld *CreateWorld();
  virtual CSLIDERender *CreateRender(CHAR *pcFile, UINT_32 uiLine, 
				     UINT_32 uiColumn,
				     CHAR *pcViewportID, 
				     CListInt *plistCameraPath, 
				     CHAR *pcRootID);
  virtual CSLIDELightPath *CreateLightPath(CHAR *pcFile, UINT_32 uiLine, 
					   UINT_32 uiColumn,
					   CListInt *plistLightPath,
					   DynInt *pdyniOnOff); 
  virtual CSLIDEAreaLightPath *CreateAreaLightPath(CHAR *pcFile, UINT_32 uiLine, 
					   UINT_32 uiColumn,
					   CListInt *plistAreaLightPath,
					   DynInt *pdyniOnOff); 

// Cheap WIN32 hack?
#ifdef CreateWindowA
#undef CreateWindowA
#endif
#ifdef CreateWindow
#undef CreateWindow
#endif



  virtual CWindow *CreateWindow(CHAR *pcFile, UINT_32 uiLine, 
				UINT_32 uiColumn,
				CHAR *pcID);
  virtual CSLIDEViewport *CreateViewport(CHAR *pcFile, UINT_32 uiLine, 
					 UINT_32 uiColumn,
					 CHAR *pcID, CHAR *pcWindowID);
  virtual CSLIDECamera *CreateCamera(CHAR *pcFile, UINT_32 uiLine, 
				     UINT_32 uiColumn,
				     CHAR *pcID);
  virtual CSLIDELight *CreateLight(CHAR *pcFile, UINT_32 uiLine, 
				   UINT_32 uiColumn,
				   CHAR *pcID);
  virtual CSLIDEAreaLight *CreateAreaLight(CHAR *pcFile, UINT_32 uiLine, 
					   UINT_32 uiColumn,
					   CHAR *pcID);
  virtual CSLIDEGroup *CreateGroup(CHAR *pcFile, UINT_32 uiLine, 
				   UINT_32 uiColumn,
				   CHAR *pcID);
  virtual CSLIDESeam *CreateSeam(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
				 CHAR *pcID, CHAR *pcContour1ID, CHAR *pcContour2ID);
  virtual CSLIDEContour *CreateContour(CHAR *pcFile, UINT_32 uiLine, 
				       UINT_32 uiColumn,
				       CHAR *pcID);
  virtual CSLIDESubdivision *CreateSubdivision(CHAR *pcFile, UINT_32 uiLine, 
					       UINT_32 uiColumn,
					       CHAR *pcID);
  virtual CSLIDEOffset *CreateOffset(CHAR *pcFile, UINT_32 uiLine, 
				     UINT_32 uiColumn,
				     CHAR *pcID);
  virtual CSLIDEDual *CreateDual(CHAR *pcFile, UINT_32 uiLine, 
				 UINT_32 uiColumn,
				 CHAR *pcID);
  virtual CSLIDEEvolver *CreateEvolver(CHAR *pcFile, UINT_32 uiLine, 
				 UINT_32 uiColumn,
				 CHAR *pcID);
  virtual CSLIDEMirror *CreateMirror(CHAR *pcFile, UINT_32 uiLine, 
				 UINT_32 uiColumn,
				 CHAR *pcID);
  virtual CSLIDEGeodesic *CreateGeodesic(CHAR *pcFile, UINT_32 uiLine, 
					 UINT_32 uiColumn,
					 CHAR *pcID);
  virtual CSLIDEUnion *CreateUnion(CHAR *pcFile, UINT_32 uiLine, 
				   UINT_32 uiColumn,
				   CHAR *pcID);
  virtual CSLIDEIntersection *CreateIntersection(CHAR *pcFile, UINT_32 uiLine, 
				   UINT_32 uiColumn,
				   CHAR *pcID);
  virtual CSLIDEDifference *CreateDifference(CHAR *pcFile, UINT_32 uiLine, 
					     UINT_32 uiColumn,
					     CHAR *pcID);
  virtual CSLIDESGRoot *CreateSGRoot(CHAR *pcFile, UINT_32 uiLine, 
				     UINT_32 uiColumn,
				     CHAR *pcID, CHAR *pcSGRoot);
  virtual CSLIDEInstance *CreateInstance(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
					 CHAR *pcChildID, CHAR *pcParentID);
  virtual CSLIDETranslate *CreateTranslate(CHAR *pcFile, UINT_32 uiLine, 
					   UINT_32 uiColumn,
					   DynVector *pdynv);
  virtual CSLIDEScale *CreateScale(CHAR *pcFile, UINT_32 uiLine, 
				   UINT_32 uiColumn,
				   DynVector *pdynv);
  virtual CSLIDERotate *CreateRotate(CHAR *pcFile, UINT_32 uiLine, 
				     UINT_32 uiColumn,
				     DynVector *pdynvAxis,
				     DynFloat *pdynfAngle);
  virtual CSLIDELookAt *CreateLookAt(CHAR *pcFile, UINT_32 uiLine, 
				     UINT_32 uiColumn);
/*
  virtual CSLIDETransform *CreateGeneral(CHAR *pcFile, UINT_32 uiLine, 
					 UINT_32 uiColumn,
					 const Matrix &m);
*/
  virtual CSLIDEObject *CreateObject(CHAR *pcFile, UINT_32 uiLine, 
				     UINT_32 uiColumn,
				     CHAR *pcID, 
				     CListInt *plistFaceIDs);
  virtual CSLIDEMesh *CreateMesh(CHAR *pcFile, UINT_32 uiLine, 
				 UINT_32 uiColumn,
				 CHAR *pcID);
  virtual CSLIDEMeshVertexProperty *CreateMeshVertexProperty(CHAR *pcFile, UINT_32 uiLine, 
							     UINT_32 uiColumn,
							     CHAR *pcID);
  virtual CSLIDEMeshEdgeProperty *CreateMeshEdgeProperty(CHAR *pcFile, UINT_32 uiLine, 
							 UINT_32 uiColumn,
							 CHAR *pcID);
  virtual CSLIDEMeshVertex *CreateMeshVertex(CHAR *pcFile, UINT_32 uiLine, 
					     UINT_32 uiColumn,
					     CHAR *pcID, 
					     DynPoint *pdynpt);
  virtual CSLIDEMeshFacet *CreateMeshFacet(CHAR *pcFile, UINT_32 uiLine, 
					   UINT_32 uiColumn,
					   CHAR *pcID, 
					   CListInt *plistMeshVertexIDs);
  virtual CSLIDEMeshEdge *CreateMeshEdge(CHAR *pcFile, UINT_32 uiLine, 
					 UINT_32 uiColumn,
					 CHAR *pcID, 
					 CHAR *pcMeshVertexID0, CHAR *pcMeshVertexID1);
  virtual CSLIDEConvexHull *CreateConvexHull(CHAR *pcFile, UINT_32 uiLine, 
					     UINT_32 uiColumn,
					     CHAR *pcID, 
					     CListInt *plistPointIDs);
  virtual CSLIDEGeneralFace *CreateGeneralFace(CHAR *pcFile, UINT_32 uiLine, 
				       UINT_32 uiColumn,
				       CHAR *pcID, 
				     CListInt *plistPointIDs);
  virtual CSLIDETerrain *CreateTerrain(CHAR *pcFile, UINT_32 uiLine, 
				       UINT_32 uiColumn,
				       CHAR *pcID, 
				     CListInt *plistPointIDs);
  virtual CSLIDESphere *CreateSphere(CHAR *pcFile, UINT_32 uiLine, 
				     UINT_32 uiColumn,
				     CHAR *pcID);
  virtual CSLIDECylinder *CreateCylinder(CHAR *pcFile, UINT_32 uiLine, 
					 UINT_32 uiColumn,
					 CHAR *pcID);
  virtual CSLIDECone *CreateCone(CHAR *pcFile, UINT_32 uiLine, 
				 UINT_32 uiColumn,
				 CHAR *pcID);
  virtual CSLIDETorus *CreateTorus(CHAR *pcFile, UINT_32 uiLine, 
				   UINT_32 uiColumn,
				   CHAR *pcID);
  virtual CSLIDECyclide *CreateCyclide(CHAR *pcFile, UINT_32 uiLine, 
				       UINT_32 uiColumn,
				       CHAR *pcID);
  virtual CSLIDEBSplinePatch *CreateBSplinePatch(CHAR *pcFile, UINT_32 uiLine, 
						 UINT_32 uiColumn,
						 CHAR *pcID);
  virtual CSLIDEHeightfield *CreateHeightfield(CHAR *pcFile, UINT_32 uiLine, 
						 UINT_32 uiColumn,
						 CHAR *pcID);
  virtual CSLIDENurbPatch *CreateNurbPatch(CHAR *pcFile, UINT_32 uiLine, 
					   UINT_32 uiColumn,
					   CHAR *pcID);
  virtual CSLIDEBezierPatch *CreateBezierPatch(CHAR *pcFile, UINT_32 uiLine, 
					       UINT_32 uiColumn,
					       CHAR *pcID);
  virtual CSLIDEBSplineCurve *CreateBSplineCurve(CHAR *pcFile, UINT_32 uiLine, 
						 UINT_32 uiColumn,
						 CHAR *pcID);
  virtual CSLIDEPolyline *CreatePolyline(CHAR *pcFile, UINT_32 uiLine, 
						 UINT_32 uiColumn,
						 CHAR *pcID);
  virtual CSLIDECrossSection *CreateCrossSection(CHAR *pcFile, UINT_32 uiLine, 
						 UINT_32 uiColumn,
						 CHAR *pcID);
  virtual CSLIDESweep *CreateSweep(CHAR *pcFile, UINT_32 uiLine, 
				   UINT_32 uiColumn,
				   CHAR *pcID);
  virtual CSLIDESweepMorph *CreateSweepMorph(CHAR *pcFile, UINT_32 uiLine, 
					     UINT_32 uiColumn,
					     CHAR *pcID);
  virtual CSLIDESweepCrossSection *CreateSweepCrossSection(CHAR *pcFile, UINT_32 uiLine, 
							   UINT_32 uiColumn,
							   CHAR *pcID);
  virtual CSLIDEFace *CreateFace(CHAR *pcFile, UINT_32 uiLine, 
				 UINT_32 uiColumn,
				 CHAR *pcID, 
				 CListInt *plistPointIDs);
  virtual CSLIDEPoint *CreatePoint(CHAR *pcFile, UINT_32 uiLine, 
				   UINT_32 uiColumn,
				   CHAR *pcID, 
				   DynPoint *pdynpt);
  virtual CSLIDEControlPoint *CreateControlPoint(CHAR *pcFile, UINT_32 uiLine, 
						 UINT_32 uiColumn,
						 CHAR *pcID);
  virtual CSLIDESurface *CreateSurface(CHAR *pcFile, UINT_32 uiLine, 
				       UINT_32 uiColumn,
				       CHAR *pcID);
  virtual CSLIDEWormwarp *CreateWormwarp(CHAR *pcFile, UINT_32 uiLine, 
				       UINT_32 uiColumn,
				       CHAR *pcID);
  virtual CSLIDEFog *CreateFog(CHAR *pcFile, UINT_32 uiLine, 
			       UINT_32 uiColumn,
			       CHAR *pcID);
  CListInt *CreateList();
  CListInt *InsertList(CListInt *pCListInt, VOID *pvData);

  CListString *CreateStringList();
  CListString *InsertStringList(CListString *pCListString, VOID *pvData);

//////////////////////////////////////////////////
// Handler Methods
//
public:
  int HandleCreate(int argc, char **argv);
  int HandleDelete(int argc, char **argv);
  int HandleModify(int argc, char **argv);

  int HandleUnselect();
  int HandleGetParameters(CHAR *pcID, CHAR *pcType);
  int HandleSetParameters(CHAR *pcID, CHAR *pcType, BOOL bPrerender, int argc, CHAR **argv);

  // Geometry
  int HandleAddPoint(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddFace(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddObject(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddConvexHull(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddGeneralFace(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddContour(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddTerrain(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddGroup(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddInstance(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddSGRoot(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddSphere(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddCylinder(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddCone(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddTorus(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddCyclide(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddPolyline(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddBSplineCurve(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddBezierCurve(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddSweep(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddBSplinePatch(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddHeightfield(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddBezierPatch(VOID* &pCreatedObject, int argc, char **argv);
  // Non-Geometry
  int HandleAddFog(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddSurface(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddCamera(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddLight(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddAreaLight(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddViewport(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddRender(VOID* &pCreatedObject, int argc, char **argv);
  int HandleAddWindow(VOID* &pCreatedObject, int argc, char **argv);


  int HandleModifyPoint(VOID* &pModifiedObject, int argc, char **argv);
  int HandleModifyPolyline(VOID* &pModifiedObject, int argc, char **argv);

//////////////////////////////////////////////////
// Callback Functions
//
public:
  static int CreateOotClass(Tcl_Interp *pInterp);

  // Richard Sun
  static int CBHandleSLIDECmd(ClientData clientData, Tcl_Interp *pInterp,  
			       int argc, char **argv);

  // Jane Yen
  static int CBHandleGUIUnselect(ClientData clientData, Tcl_Interp *pInterp,  
				 int argc, char **argv);
  static int CBHandleGUIGetParameters(ClientData clientData, Tcl_Interp *pInterp,  
				      int argc, char **argv);
  static int CBHandleGUISetParameters(ClientData clientData, Tcl_Interp *pInterp,  
				      int argc, char **argv);

//////////////////////////////////////////////////
// Accessor Functions
//
public:

//////////////////////////////////////////////////
// Class Variables
//
protected:
  static CSLIDEParser *c_pCCurrParser;

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CSLIDEWorld *m_pCSLIDEWorld;
  CSLIDEPrimitive *m_pCSLIDEPrimitiveCurr;
  CSLIDERender *m_pCSLIDERenderCurr;
  CWindow *m_pCWindowCurr;
  CSLIDEViewport *m_pCSLIDEViewportCurr;
  CSLIDECamera *m_pCSLIDECameraCurr;
  CSLIDELight *m_pCSLIDELightCurr;
  CSLIDEGeometryNode *m_pCSLIDENodeCurr;
  CSLIDEAtomNode *m_pCSLIDEAtomNodeCurr;
  CSLIDEAreaLight *m_pCSLIDEAreaLightCurr;
  CSLIDEGroup *m_pCSLIDEGroupCurr;
  CSLIDESubdivision *m_pCSLIDESubdivisionCurr;
  CSLIDEOffset *m_pCSLIDEOffsetCurr;
  CSLIDEDual *m_pCSLIDEDualCurr;
  CSLIDEEvolver *m_pCSLIDEEvolverCurr;
  CSLIDEMirror *m_pCSLIDEMirrorCurr;
  CSLIDEGeodesic *m_pCSLIDEGeodesicCurr;
  CSLIDEInstance *m_pCSLIDEInstanceCurr;
  CSLIDESeam *m_pCSLIDESeamCurr;
  CSLIDEContour *m_pCSLIDEContourCurr;
  CSLIDEObject *m_pCSLIDEObjectCurr;
  CSLIDEMesh *m_pCSLIDEMeshCurr;
  CSLIDEMeshVertexProperty *m_pCSLIDEMeshVertexPropertyCurr;
  CSLIDEMeshEdgeProperty *m_pCSLIDEMeshEdgePropertyCurr;
  CSLIDEMeshVertex *m_pCSLIDEMeshVertexCurr;
  CSLIDEMeshFacet *m_pCSLIDEMeshFacetCurr;
  CSLIDEMeshEdge *m_pCSLIDEMeshEdgeCurr;
  CSLIDEConvexHull *m_pCSLIDEConvexHullCurr;
  CSLIDEGeneralFace *m_pCSLIDEGeneralFaceCurr;
  CSLIDETerrain *m_pCSLIDETerrainCurr;
  CSLIDESGRoot *m_pCSLIDESGRootCurr;
  CSLIDESphere *m_pCSLIDESphereCurr;
  CSLIDECylinder *m_pCSLIDECylinderCurr;
  CSLIDECone *m_pCSLIDEConeCurr;
  CSLIDETorus *m_pCSLIDETorusCurr;
  CSLIDECyclide *m_pCSLIDECyclideCurr;
  CSLIDEBSplinePatch *m_pCSLIDEBSplinePatchCurr;    
  CSLIDEHeightfield *m_pCSLIDEHeightfieldCurr;
  CSLIDENurbPatch *m_pCSLIDENurbPatchCurr;    
  CSLIDEBezierPatch *m_pCSLIDEBezierPatchCurr;    
  CSLIDEBSplineCurve *m_pCSLIDEBSplineCurveCurr;    
  CSLIDEPolyline *m_pCSLIDEPolylineCurr;    
  CSLIDECrossSection *m_pCSLIDECrossSectionCurr;    
  CSLIDESweep *m_pCSLIDESweepCurr;    
  CSLIDESweepMorph *m_pCSLIDESweepMorphCurr;    
  CSLIDESweepCrossSection *m_pCSLIDESweepCrossSectionCurr;    
  CSLIDEFace *m_pCSLIDEFaceCurr;
  CSLIDEPoint *m_pCSLIDEPointCurr;
  CSLIDEControlPoint *m_pCSLIDEControlPointCurr;
  CSLIDESurface *m_pCSLIDESurfaceCurr;
  CSLIDEWormwarp *m_pCSLIDEWormwarpCurr;
  CSLIDEFog *m_pCSLIDEFogCurr;
  CSLIDELookAt *m_pCSLIDELookAtCurr;

  CHAR *m_pcOotName;

  static Tcl_Interp *m_pInterp;

  static const int m_cuiOotMethods;
  static OotMethodEntry m_pOotMethods[cuiCSLIDEParserOotMethods];

  static const CHAR *m_cpcOotClassName;
  static OotStaticClass *m_pOotClass;  

  CHashString m_hashOotAPIFuncMap;
  CHashString m_hashOotCreateFuncMap;
  CHashString m_hashOotModifyFuncMap;

  UINT_32 m_uiInstances;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

//////////////////////////////////////////////////////////////////////
// Globals
//

extern CSLIDEParser *slide_InitParser(CSLIDEParser *pCSLIDEParser, 
				      CHAR *pcInput, 
				      CHAR **ppcDirectories, 
				      UINT_32 uiDirectories);

#endif // __SLIDEPARSER_H__
