#include "slidepath.h"
#include "slideworld.h"
#include "slideinstance.h"

//////////////////////////////////////////////////////////////////////////////
// CSLIDEPath Class
//

CSLIDEPath::CSLIDEPath()
{
  m_plistPathIDs = NULL;
  
  m_pRoot = NULL;
  m_pNode = NULL;
}

CSLIDEPath::~CSLIDEPath()
{
}

VOID CSLIDEPath::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
		       CListInt *plistPathIDs)
{
  ASSERT( plistPathIDs != NULL );

  CSLIDEPrimitive::Init(pcFile, uiLine, uiColumn, NULL);

  m_plistPathIDs = plistPathIDs;
  
  m_pRoot = NULL;
  m_listCSLIDEInstance.Init();
  m_pNode = NULL;
}

VOID CSLIDEPath::Uninit()
{
  CSLIDEPrimitive::Uninit();

  if ( m_plistPathIDs != NULL )
    {
      delete m_plistPathIDs;
      m_plistPathIDs = NULL;
    }

  m_pRoot = NULL;
  m_pNode = NULL;
}

BOOL CSLIDEPath::Initialized()
{
  if ( !CSLIDEPrimitive::Initialized() )
    {
      return FALSE;
    }

  return m_plistPathIDs != NULL; 
}

BOOL CSLIDEPath::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;
  CListIterInt iter;
  CHAR *pcID;

  if ( !CSLIDEPrimitive::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  // Check Path References

  iter.Init(m_plistPathIDs);

  pcID = (CHAR*)iter.PeekFirst();
  ASSERT( pcID != NULL );
  m_pRoot = pCSLIDEWorld->PeekNode(pcID);
  if ( m_pRoot == NULL )
    {
      HandleError("Undefined %s \"%s\" in path", 
		  gcppcSLIDEReservedWords[sliderwtObject],
		  pcID);
      bCorrect = FALSE;
    }
  else
    {
      m_pNode = FindNode(m_pRoot, &iter, &m_listCSLIDEInstance);
      if ( m_pNode == NULL )
	{
	  HandleError("Bad path");
	  bCorrect = FALSE;
	}
    } 

  return bCorrect;
}

BOOL CSLIDEPath::Valid()
{
  if ( !CSLIDEPrimitive::Valid() )
    {
      return FALSE;
    }

  return ( m_pRoot != NULL && 
	   m_pNode != NULL );
}

VOID CSLIDEPath::CheckDynamic()
{
  BOOL bDynamic;
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  bDynamic = FALSE;

  iter.Init(&m_listCSLIDEInstance);
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      if ( pCSLIDEInstance->GetTransformDynamic() )
	{
	  bDynamic = TRUE;
	  break;
	}
    }

  SetDynamic(bDynamic);
}

VOID CSLIDEPath::UpdateState()
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  m_mGroot_obj.SetIdentity();
  m_mGobj_root.SetIdentity();

  iter.Init(&m_listCSLIDEInstance);
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      m_mGroot_obj = m_mGroot_obj * pCSLIDEInstance->GetTransform();
      m_mGobj_root = pCSLIDEInstance->GetInverse() * m_mGobj_root;
    }  
  /*
printf("***** Start Path Update *****\n");
printf("Groot_obj:\n");
m_mGroot_obj.Display();
printf("Gobj_root:\n");
m_mGobj_root.Display();
printf("Identity:\n");
Matrix mJordan;
mJordan = m_mGobj_root * m_mGroot_obj;
mJordan.Display();
printf("***** End Path Update *****\n");
*/
}

/*
VOID CSLIDEPath::Render(const Matrix &m)
{
}
*/
VOID CSLIDEPath::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CHAR *pcID;

  ASSERT( fpOutputFile != NULL );
  
  iter.Init(m_plistPathIDs);

  pcID = (CHAR*)iter.PeekFirst();
  ASSERT( pcID != NULL );

  fprintf(fpOutputFile, "%s", pcID);

  for ( pcID = (CHAR*)iter.PeekNext();
	pcID != NULL;
	pcID = (CHAR*)iter.PeekNext() )
    {
      fprintf(fpOutputFile, ".%s", pcID);
    }
}

VOID CSLIDEPath::WriteDotEdges(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( fpOutputFile != NULL );

  if ( m_pRoot != m_pNode )
    {
      CSLIDENode *pCSLIDENode;
      
      fprintf(fpOutputFile, " -> ");
      m_pRoot->WriteDotID(fpOutputFile);
      fprintf(fpOutputFile, " [dir = back, style = dotted];\n");

      pCSLIDENode = m_pRoot;

      iter.Init(&m_listCSLIDEInstance);
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  fprintf(fpOutputFile, "  ");
	  pCSLIDENode->WriteDotID(fpOutputFile);

	  pCSLIDENode = pCSLIDEInstance->GetNode();
	  ASSERT( pCSLIDENode != NULL );

	  fprintf(fpOutputFile, " -> ");
	  pCSLIDEInstance->WriteDotID(fpOutputFile);
	  fprintf(fpOutputFile, " [dir = back, style = dotted];\n");
	  
	  fprintf(fpOutputFile, "  ");
	  pCSLIDEInstance->WriteDotID(fpOutputFile);
	  fprintf(fpOutputFile, " -> ");
	  pCSLIDENode->WriteDotID(fpOutputFile);
	  fprintf(fpOutputFile, " [dir = back, style = dotted];\n");
	}  
    }
  else
    {
      fprintf(fpOutputFile, " -> ");
      m_pNode->WriteDotID(fpOutputFile);
      fprintf(fpOutputFile, " [dir = back, style = dotted];\n");
    }
}
