#include "slidepoint.h"
#include "slideworld.h"
#include "slidesurface.h"
#include "slideface.h"
#include "slidewarp.h"

//////////////////////////////////////////////////////////////////////////////
// CSLIDEPoint Class
//

CSLIDEPoint::CSLIDEPoint()
{
  m_pcSurfaceID = NULL;
  m_pCSLIDESurface = NULL;

  m_pt.Null();
  m_vNormal.Null();
  m_ptTexture.Null();

  m_listCSLIDEParentFace.Init();
}

CSLIDEPoint::~CSLIDEPoint()
{
}

VOID CSLIDEPoint::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		       CHAR *pcID, 
		       DynPoint *pdynpt)
{
  ASSERT( pdynpt != NULL );

  CSLIDEPrimitive::Init(pcFile, uiLine, uiColumn, pcID);

  m_pcSurfaceID = gcppcSLIDEReservedWords[sliderwtSLF_INHERIT];
  m_pCSLIDESurface = NULL;

  m_pt.SetAll(pdynpt);
  m_vNormal.Null();
  m_ptTexture.Null();

  SetBoundDynamic(FALSE);
  SetAutoNormal(TRUE);
  SetAutoNormalDynamic(FALSE);

  m_uiFaceCountMax = 0xffffffff;
  m_uiFaceCount = 0;
}

VOID CSLIDEPoint::Uninit()
{
  CSLIDEPrimitive::Uninit();
}

BOOL CSLIDEPoint::Initialized()
{
  if ( !CSLIDEPrimitive::Initialized() )
    {
      return FALSE;
    }

  return m_pcSurfaceID != NULL; 
}

VOID CSLIDEPoint::InsertParentFace(CSLIDEFace *pCSLIDEFace)
{
  ASSERT( pCSLIDEFace != NULL );

  m_listCSLIDEParentFace.InsertLast(pCSLIDEFace, (UINT_32)pCSLIDEFace);
}

VOID CSLIDEPoint::RemoveParentFace(CSLIDEFace *pCSLIDEFaceToRemove)
{
  ASSERT( pCSLIDEFaceToRemove != NULL );

  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;
  CHAR *pcID;

  pcID = pCSLIDEFaceToRemove->GetID();

  iter.Init(&m_listCSLIDEParentFace);
  for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst();
	pCSLIDEFace != NULL; )
    {
      if( (pCSLIDEFace->GetID()) == pcID )
	{
	  pCSLIDEFace = (CSLIDEFace*)iter.PeekNextRemoveCurr();
	}
      else
	{
	  pCSLIDEFace = (CSLIDEFace*)iter.PeekNext();
	}
    }
}

BOOL CSLIDEPoint::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;

  if ( !CSLIDEPrimitive::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  // Check Surface Reference
  m_pCSLIDESurface = pCSLIDEWorld->PeekSurface(m_pcSurfaceID);
  if ( m_pCSLIDESurface == NULL )
    {
      HandleError("Undefined %s \"%s\"", 
		  gcppcSLIDEReservedWords[sliderwtSurface],
		  m_pcSurfaceID);
      bCorrect = FALSE;
    }

  return bCorrect;
}

BOOL CSLIDEPoint::Valid()
{
  if ( !CSLIDEPrimitive::Valid() )
    {
      return FALSE;
    }

  return m_pt[W] != 0.0;
}

VOID CSLIDEPoint::Remove(CSLIDEWorld *pCSLIDEWorld)
{
  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;

  // Remove from parent
  iter.Init(&m_listCSLIDEParentFace);
  for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst();
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
    {
      pCSLIDEFace->RemovePoint(this);
    }

  // Remove from world
  pCSLIDEWorld->RemovePoint(this);
  Uninit();
}

VOID CSLIDEPoint::RecalculateParent()
{
  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;

  iter.Init(&m_listCSLIDEParentFace);
  for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst();
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
    {
      pCSLIDEFace->Preprocess();
    }
}

VOID CSLIDEPoint::InsertIntoBound(CBound *pCBound)
{
  ASSERT( pCBound != NULL );

  pCBound->Insert(m_pt);
}

VOID CSLIDEPoint::AddFaceNormal(const Vector &vNormal, FLOAT fWeight)
{
  m_uiFaceCount++;

  m_vNormal += (vNormal * fWeight);

  if ( m_uiFaceCount == m_uiFaceCountMax )
    {
      m_vNormal.CompletePlaneEquation(m_pt);
    }
}

VOID CSLIDEPoint::PreprocessNormal()
{
  m_uiFaceCountMax = m_uiFaceCount;

  m_vNormal.CompletePlaneEquation(m_pt);
}

VOID CSLIDEPoint::SetBoundUnvisited()
{
  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;

  iter.Init(&m_listCSLIDEParentFace);
  for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst();
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
    {
      pCSLIDEFace->SetBoundUnvisited();
    }
}

VOID CSLIDEPoint::SetBoundUnchanged()
{
  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;

  iter.Init(&m_listCSLIDEParentFace);
  for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst();
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
    {
      pCSLIDEFace->SetBoundUnchanged();
    }
}

VOID CSLIDEPoint::SetPrerenderUnvisited()
{
  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;

  iter.Init(&m_listCSLIDEParentFace);
  for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst();
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
    {
      pCSLIDEFace->SetPrerenderUnvisited();
    }
}

VOID CSLIDEPoint::CalculatePointWarped( CSLIDEWormwarp *pCSLIDEWarp )
{
  m_pWarpLast = pCSLIDEWarp;
  pCSLIDEWarp->CalcWarpPoint(m_pt, m_ptWarped);
}

#ifdef __USE_CDA__
VOID CSLIDEPoint::CalculateCDAVertices(CListInt *plistCDAVertex, 
				       const Matrix &mGroot_obj)
{
  CDAVertex *pCDAVertex;

  ASSERT( plistCDAVertex != NULL );

  pCDAVertex = AllocateCDAVertex();
  ASSERT( pCDAVertex != NULL );
  pCDAVertex->Init(mGroot_obj*m_pt);
  plistCDAVertex->InsertLast(pCDAVertex, 0);
}
#endif // __USE_CDA__

VOID CSLIDEPoint::CheckDynamic()
{
  SetDynamic(m_pt.GetDynamic() ||
	     m_vNormal.GetDynamic() ||
	     m_ptTexture.GetDynamic());
  SetBoundDynamic(m_pt.GetDynamic());
  SetNormalDynamic(m_pt.GetDynamic() ||
		   m_vNormal.GetDynamic());
  SetGeometryDynamic(GetDynamic() || 
		     m_pCSLIDESurface->GetDynamic());
}

VOID CSLIDEPoint::UpdateState()
{
  m_pt.Update();

  m_uiFaceCount = 0;
  if ( GetAutoNormalDynamic() )
    {
      m_vNormal.Null();
    }
  else
    {
      if ( m_vNormal.GetDynamic() )
	{
	  m_vNormal.Update();
	  m_vNormal.CompletePlaneEquation(m_pt);
	}
    }

  m_ptTexture.Update();
}

VOID CSLIDEPoint::Render(BOOL bUseLocalSurface, CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{
  Point ptNew;
  ptNew = m_pt;
  if (NULL != pCSLIDEWarp)
    {
      ptNew = m_ptWarped;
      //      pCSLIDEWarp->CalcWarpPoint(m_pt, ptNew);
    }

  if ( bUseLocalSurface )
    {
      GLbitfield GLbAttributesMask;
      
      GLbAttributesMask = m_pCSLIDESurface->CalculateAttributesMask();
      
      PushAttributes(GLbAttributesMask);
      {
	m_pCSLIDESurface->Push();
	{
	  gilTexCoord2fv(m_ptTexture.Get());
	  if (NULL == pCSLIDEWarp)
	    {
	      gilNormal3fv(m_vNormal.Get());
	    }
	  gilVertex3fv(ptNew.Get());
	}
	m_pCSLIDESurface->Pop();
      }
      PopAttributes(GLbAttributesMask);
    }
  else
    {
      gilTexCoord2fv(m_ptTexture.Get());
      gilVertex3fv(ptNew.Get());
    }

}

VOID CSLIDEPoint::RenderSelect()
{
  gilVertex3fv(m_pt.Get());
}

VOID CSLIDEPoint::RenderNormal()
{
  Vector v;
  Point pt;
  
  gilVertex3fv(m_pt.Get());
  v = m_vNormal;
  v[W] = 0.0;
  pt = m_pt + v;
  gilVertex3fv(pt.Get());
}

VOID CSLIDEPoint::WriteSLIDE(FILE *fpOutputFile)
{
  Vector v;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s ( %.3f %.3f %.3f )\n", 
	  gcppcSLIDEReservedWords[sliderwtPoint], m_pcGUIID, 
	  m_pt[X], m_pt[Y], m_pt[Z]);

  if ( m_pcSurfaceID != gcppcSLIDEReservedWords[sliderwtSLF_INHERIT] )
    {
      fprintf(fpOutputFile, "  %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtSurface], 
	      m_pcSurfaceID);
    } 
  if ( m_vNormal.Magnitude() != 0.0 )
    {
      fprintf(fpOutputFile, "  %s ( %.3f %.3f %.3f )\n",
	      gcppcSLIDEReservedWords[sliderwtNormal], 
	      m_vNormal[X], m_vNormal[Y], m_vNormal[Z]); 
    }
  v = m_ptTexture;
  if ( v.Magnitude() != 0.0 )
    {
      fprintf(fpOutputFile, "  %s ( %.3f %.3f %.3f )\n",
	      gcppcSLIDEReservedWords[sliderwtTexture], 
	      m_ptTexture[0], m_ptTexture[1], m_ptTexture[2]); 
    }
  CSLIDEPrimitive::WriteSLIDEFlags(fpOutputFile, 0);

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndPoint]); 
}

VOID CSLIDEPoint::WriteRIB(RtPoint *prtPoint, 
			   RtPoint *prtNormal, 
			   RtPoint *prtColor, 
			   RtFloat *prtTexture)
{
  ASSERT( prtPoint != NULL );
  ASSERT( prtNormal != NULL );
  ASSERT( prtColor != NULL );
  ASSERT( prtTexture != NULL );

  (*prtPoint)[0] = m_pt[0];
  (*prtPoint)[1] = m_pt[1];
  (*prtPoint)[2] = m_pt[2];

  (*prtNormal)[0] = m_vNormal[0];
  (*prtNormal)[1] = m_vNormal[1];
  (*prtNormal)[2] = m_vNormal[2];

  prtTexture[0] = m_ptTexture[0];
  prtTexture[1] = m_ptTexture[1];
  //(*prtTexture)[0] = m_ptTexture[0];
  //(*prtTexture)[1] = m_ptTexture[1];
  //(*prtTexture)[2] = m_ptTexture[2];

  if ( m_pCSLIDESurface != CSLIDESurface::GetSLF_INHERIT() ) 
    {
      (*prtColor)[0] = m_pCSLIDESurface->GetColor()(0);
      (*prtColor)[1] = m_pCSLIDESurface->GetColor()(1);
      (*prtColor)[2] = m_pCSLIDESurface->GetColor()(2);
    }
}

UINT_32 CSLIDEPoint::GetPrerenderAttributesStackDepthMax()
{
  if ( m_pCSLIDESurface->CalculateAttributesMask() != 0x0 )
    {
      return 1;
    }
  else
    {
      return 0;
    }
}

