#ifndef __POINT4D_H__
#define __POINT4D_H__

#include "slideprimitive.h"

class CSLIDESurface;
class CSLIDEFace;

//////////////////////////////////////////////////////////////////////
// CSLIDEPoint Class
//

class CSLIDEPoint : public CSLIDEPrimitive
{
//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEPoint();
  virtual ~CSLIDEPoint();

  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID, 
		    DynPoint *pdynpt);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  VOID InsertParentFace(CSLIDEFace *pCSLIDEFace);
  VOID RemoveParentFace(CSLIDEFace *pCSLIDEFace);

  inline VOID SetPoint(DynPoint *pdynpt);
  inline VOID SetSurfaceID(CHAR *pcSurfaceID);
  inline VOID SetNormal(DynVector *pdynvNormal);
  inline VOID SetTexture(DynPoint *pdynptTexture);

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

  virtual VOID Remove(CSLIDEWorld *pCSLIDEWorld);
  VOID RecalculateParent();

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  VOID InsertIntoBound(CBound *pCBound);
  VOID AddFaceNormal(const Vector &vNormal, FLOAT fWeight);

  VOID PreprocessNormal();
  virtual VOID SetBoundUnvisited();
  virtual VOID SetBoundUnchanged();
  virtual VOID SetPrerenderUnvisited();

  virtual VOID CalculatePointWarped( CSLIDEWormwarp *pCSLIDEWarp );

#ifdef __USE_CDA__
  VOID CalculateCDAVertices(CListInt *plistCDAVertex, 
			    const Matrix &mGroot_obj);
#endif // __USE_CDA__

protected:
  VOID CheckDynamic();
  VOID UpdateState();

//////////////////////////////////////////////////
// Rendering
//
public:
  VOID Render(BOOL bUseLocalSurface, CSLIDEWormwarp *pCSLIDEWarp = NULL);
  VOID RenderSelect();
  VOID RenderNormal();

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);
  VOID WriteRIB(RtPoint *prtPoint, 
		RtPoint *prtNormal, 
		RtPoint *prtColor, 
		RtFloat *prtTexture);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline SLIDEReservedWordType GetType();

  inline CSLIDESurface *GetSurface();
  inline const Point &GetPoint();
  inline const Vector &GetNormal();
  inline const Point &GetTexture();
  inline const Point &GetPointWarped();

  inline BOOL GetBoundDynamic();
  inline VOID SetBoundDynamic(BOOL bBoundDynamic);
  inline BOOL GetGeometryDynamic();
  inline VOID SetGeometryDynamic(BOOL bGeometryDynamic);
  inline BOOL GetNormalDynamic();
  inline VOID SetNormalDynamic(BOOL bNormalDynamic);
  inline BOOL GetAutoNormal();
  inline VOID SetAutoNormal(BOOL bAutoNormal);
  inline BOOL GetAutoNormalDynamic();
  inline VOID SetAutoNormalDynamic(BOOL bAutoNormalDynamic);

  UINT_32 GetPrerenderAttributesStackDepthMax();

//////////////////////////////////////////////////
// Class Variables
//
public:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CHAR *m_pcSurfaceID;
  CSLIDESurface *m_pCSLIDESurface;

  DynPoint m_pt;
  DynVector m_vNormal;
  DynPoint m_ptTexture;
  Point m_ptWarped;
  CSLIDEWormwarp *m_pWarpLast;

  BOOL m_bBoundDynamic;
  BOOL m_bGeometryDynamic;
  BOOL m_bNormalDynamic;
  BOOL m_bAutoNormal;
  BOOL m_bAutoNormalDynamic;

  UINT_32 m_uiFaceCountMax;
  UINT_32 m_uiFaceCount;

  CListInt m_listCSLIDEParentFace;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline SLIDEReservedWordType CSLIDEPoint::GetType()
{
  return sliderwtPoint;
}

inline VOID CSLIDEPoint::SetPoint(DynPoint *pdynpt)
{
  m_pt.SetAll(pdynpt);
}

inline VOID CSLIDEPoint::SetSurfaceID(CHAR *pcSurfaceID)
{
  m_pcSurfaceID = pcSurfaceID;
}

inline VOID CSLIDEPoint::SetNormal(DynVector *pdynvNormal)
{
  SetAutoNormal(FALSE);
  m_vNormal.SetAll(pdynvNormal);
}

inline VOID CSLIDEPoint::SetTexture(DynPoint *pdynptTexture)
{
  m_ptTexture.SetAll(pdynptTexture);
}

inline CSLIDESurface *CSLIDEPoint::GetSurface()
{
  return m_pCSLIDESurface;
}

inline const Point &CSLIDEPoint::GetPoint()
{
  return m_pt;
}

inline const Vector &CSLIDEPoint::GetNormal()
{
  return m_vNormal;
}

inline const Point &CSLIDEPoint::GetTexture()
{
  return m_ptTexture;
}

inline const Point &CSLIDEPoint::GetPointWarped()
{
  return m_ptWarped;
}

inline BOOL CSLIDEPoint::GetBoundDynamic()
{
  return m_bBoundDynamic;
}

inline VOID CSLIDEPoint::SetBoundDynamic(BOOL bBoundDynamic)
{
  m_bBoundDynamic = bBoundDynamic;
}

inline BOOL CSLIDEPoint::GetGeometryDynamic()
{
  return m_bGeometryDynamic;
}

inline VOID CSLIDEPoint::SetGeometryDynamic(BOOL bGeometryDynamic)
{
  m_bGeometryDynamic = bGeometryDynamic;
}

inline BOOL CSLIDEPoint::GetNormalDynamic()
{
  return m_bNormalDynamic;
}

inline VOID CSLIDEPoint::SetNormalDynamic(BOOL bNormalDynamic)
{
  m_bNormalDynamic = bNormalDynamic;
}

inline BOOL CSLIDEPoint::GetAutoNormal()
{
  return m_bAutoNormal;
}

inline VOID CSLIDEPoint::SetAutoNormal(BOOL bAutoNormal)
{
  m_bAutoNormal = bAutoNormal;
}

inline BOOL CSLIDEPoint::GetAutoNormalDynamic()
{
  return m_bAutoNormalDynamic;
}

inline VOID CSLIDEPoint::SetAutoNormalDynamic(BOOL bAutoNormalDynamic)
{
  m_bAutoNormalDynamic = bAutoNormalDynamic;
  m_bDynamic = m_bDynamic || bAutoNormalDynamic;
  m_bGeometryDynamic = m_bGeometryDynamic || bAutoNormalDynamic;
  m_bNormalDynamic = m_bNormalDynamic || bAutoNormalDynamic;
}

#endif // __POINT4D_H__


