#include "slidepolyline.h"
#include "slidesweep.h"
#include "slidesweepmorph.h"
#include "slidecrosssection.h"
#include "slidepoint.h"
#include "slidecontrolpoint.h"
#include "slideworld.h"
#include "slidesurface.h"

static const UINT_32 cuiSlices = 2;
static const UINT_32 cuiDrawPoints = 0;

static const FLOAT cfNormalLength = 1.0;
static const FLOAT cfNormalRatio = 1.0 / 6.0;

//////////////////////////////////////////////////////////////////////
// CSLIDEPolyline Class
//

CSLIDEPolyline::CSLIDEPolyline()
{
  m_fNormalLength = cfNormalLength;
  m_bSlicesChanged = FALSE;

  m_dyniDrawPoints.SetValue(cuiDrawPoints);

  // Warp
  m_bWarpDynamic = TRUE;
  m_pWarpScale  = NULL;
  m_pWarpRotate = NULL;
  m_pScaledWarp = NULL;

  // Curve
  m_bPointsDynamic = TRUE;
  m_plistPointIDs = NULL;
  m_plistCSLIDEControlPoint = new CListInt();

  // Frenet Frame
  m_pvNormal = NULL;
  m_pvBinormal = NULL;
  m_pvTangent = NULL;
  m_pvNormalTorsionReduced = NULL;
  m_pvBinormalTorsionReduced = NULL;
  m_pvTangentTorsionReduced = NULL;
  m_pmSweepFrenetFrame = NULL;
}

CSLIDEPolyline::~CSLIDEPolyline()
{
}

VOID CSLIDEPolyline::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID)
{
  CSLIDESweepPath::Init(pcFile, uiLine, uiColumn, pcID);

  m_fNormalLength = cfNormalLength;
  m_bSlicesChanged = FALSE;

  m_dyniDrawPoints.SetValue(cuiDrawPoints);

  // Warp
  m_bWarpDynamic = TRUE;
  m_listWarpScale.Init();
  m_listWarpRotate.Init();

  // Curve
  m_bPointsDynamic = TRUE;
  m_listCSLIDEPoint.Init();
  m_plistCSLIDEControlPoint->Init();

  ReallocatePoints();
  ReallocateScaledWarps();
  ReallocateWarps();

  InitPoints();

  ReallocateNormals();
  ReallocateBinormals();
  ReallocateTangents();
  ReallocateNormalsTorsionReduced();
  ReallocateBinormalsTorsionReduced();
  ReallocateTangentsTorsionReduced();
  ReallocateSweepFrenetFrames();

  // Initialize
  CalculateFrenetFrames();
  CalculateScaledWarp();
  CalculateTorsion();
}

VOID CSLIDEPolyline::Uninit()
{
  CSLIDESweepPath::Uninit();

  if( m_plistPointIDs != NULL )
    {
      delete m_plistPointIDs;
      m_plistPointIDs = NULL;
    }

  DeallocatePoints();
  DeallocateScaledWarps();
  DeallocateWarps();

  DeallocateNormals();
  DeallocateBinormals();
  DeallocateTangents();
  DeallocateNormalsTorsionReduced();
  DeallocateBinormalsTorsionReduced();
  DeallocateTangentsTorsionReduced();
  DeallocateSweepFrenetFrames();
}

BOOL CSLIDEPolyline::Initialized()
{
  if ( !CSLIDESweepPath::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

VOID CSLIDEPolyline::InitPoints()
{
  // Default Points
  /*
  m_pPoint[0].Set(-2, -1, 0);
  m_pPoint[1].Set(-2,  1, 0);
  m_pPoint[2].Set( 0,  1, 0);
  m_pPoint[3].Set( 0, -1, 0);
  m_pPoint[4].Set( 2, -1, 0);
  m_pPoint[5].Set( 2,  1, 0);
  */

  m_pPoint[0].Set(-1, 0, 0);
  m_pPoint[1].Set( 0, 0, 0);
  m_pPoint[2].Set( 1, 0, 0);
}

VOID CSLIDEPolyline::InitWarps()
{
  UINT_32 i;

  for( i = 0; i < m_uiSlices+1; i++ )
    {
      m_pWarpScale[i].Set(1.0, 1.0, 1.0);
      m_pWarpRotate[i].Set(0.0, 0.0, 0.0);

      m_pScaledWarp[i].Set(1.0, 1.0, 0.0);
    }
}

VOID CSLIDEPolyline::DeallocateWarps()
{
  if ( m_pWarpScale != NULL )
    {
      delete[] m_pWarpScale;
      m_pWarpScale = NULL;
    }

  if ( m_pWarpRotate != NULL )
    {
      delete[] m_pWarpRotate;
      m_pWarpRotate = NULL;
    }
}

VOID CSLIDEPolyline::ReallocateWarps()
{
  if( (m_pWarpScale != NULL) || (m_pWarpRotate != NULL) )
    {
      DeallocateWarps();
    }

  m_pWarpScale = new Point[m_uiSlices+1];
  ASSERT(m_pWarpScale != NULL);

  m_pWarpRotate = new Point[m_uiSlices+1];
  ASSERT(m_pWarpRotate != NULL);

  InitWarps();
}

VOID CSLIDEPolyline::DeallocateScaledWarps()
{
  if ( m_pScaledWarp != NULL )
    {
      delete[] m_pScaledWarp;
      m_pScaledWarp = NULL;
    }
}

VOID CSLIDEPolyline::ReallocateScaledWarps()
{
  if( m_pScaledWarp != NULL )
    {
      DeallocateScaledWarps();
    }

  m_pScaledWarp = new Point[m_uiSlices+1];
  ASSERT(m_pScaledWarp != NULL);
}

VOID CSLIDEPolyline::DeallocatePoints()
{
  if ( m_pPoint != NULL )
    {
      delete[] m_pPoint;
      m_pPoint = NULL;
    }
}

VOID CSLIDEPolyline::ReallocatePoints()
{
  if( m_pPoint != NULL )
    {
      DeallocatePoints();
    }

  m_pPoint = new Point[m_uiSlices+1];
  ASSERT(m_pPoint != NULL);
}

VOID CSLIDEPolyline::DeallocateNormals()
{
  if ( m_pvNormal != NULL )
    {
      delete[] m_pvNormal;
      m_pvNormal = NULL;
    }
}

VOID CSLIDEPolyline::ReallocateNormals()
{
  if( m_pvNormal != NULL )
    {
      DeallocateNormals();
    }

  m_pvNormal = new Vector[m_uiSlices+1];
  ASSERT(m_pvNormal != NULL);
}

VOID CSLIDEPolyline::DeallocateBinormals()
{
  if ( m_pvBinormal != NULL )
    {
      delete[] m_pvBinormal;
      m_pvBinormal = NULL;
    }
}

VOID CSLIDEPolyline::ReallocateBinormals()
{
  if( m_pvBinormal != NULL )
    {
      DeallocateBinormals();
    }

  m_pvBinormal = new Vector[m_uiSlices+1];
  ASSERT(m_pvBinormal != NULL);
}

VOID CSLIDEPolyline::DeallocateTangents()
{
  if ( m_pvTangent != NULL )
    {
      delete[] m_pvTangent;
      m_pvTangent = NULL;
    }
}

VOID CSLIDEPolyline::ReallocateTangents()
{
  if( m_pvTangent != NULL )
    {
      DeallocateTangents();
    }

  m_pvTangent = new Vector[m_uiSlices+1];
  ASSERT(m_pvTangent != NULL);
}

VOID CSLIDEPolyline::DeallocateNormalsTorsionReduced()
{
  if ( m_pvNormalTorsionReduced != NULL )
    {
      delete[] m_pvNormalTorsionReduced;
      m_pvNormalTorsionReduced = NULL;
    }
}

VOID CSLIDEPolyline::ReallocateNormalsTorsionReduced()
{
  if( m_pvNormalTorsionReduced != NULL )
    {
      DeallocateNormalsTorsionReduced();
    }

  m_pvNormalTorsionReduced = new Vector[m_uiSlices+1];
  ASSERT(m_pvNormalTorsionReduced != NULL);
}

VOID CSLIDEPolyline::DeallocateBinormalsTorsionReduced()
{
  if ( m_pvBinormalTorsionReduced != NULL )
    {
      delete[] m_pvBinormalTorsionReduced;
      m_pvBinormalTorsionReduced = NULL;
    }
}

VOID CSLIDEPolyline::ReallocateBinormalsTorsionReduced()
{
  if( m_pvBinormalTorsionReduced != NULL )
    {
      DeallocateBinormalsTorsionReduced();
    }

  m_pvBinormalTorsionReduced = new Vector[m_uiSlices+1];
  ASSERT(m_pvBinormalTorsionReduced != NULL);
}

VOID CSLIDEPolyline::DeallocateTangentsTorsionReduced()
{
  if ( m_pvTangentTorsionReduced != NULL )
    {
      delete[] m_pvTangentTorsionReduced;
      m_pvTangentTorsionReduced = NULL;
    }
}

VOID CSLIDEPolyline::ReallocateTangentsTorsionReduced()
{
  if( m_pvTangentTorsionReduced != NULL )
    {
      DeallocateTangentsTorsionReduced();
    }

  m_pvTangentTorsionReduced = new Vector[m_uiSlices+1];
  ASSERT(m_pvTangentTorsionReduced != NULL);
}

VOID CSLIDEPolyline::DeallocateSweepFrenetFrames()
{
  if ( m_pmSweepFrenetFrame != NULL )
    {
      delete[] m_pmSweepFrenetFrame;
      m_pmSweepFrenetFrame = NULL;
    }
}

VOID CSLIDEPolyline::ReallocateSweepFrenetFrames()
{
  if( m_pmSweepFrenetFrame != NULL )
    {
      DeallocateSweepFrenetFrames();
    }

  m_pmSweepFrenetFrame = new Matrix[m_uiSlices+1];
  ASSERT(m_pmSweepFrenetFrame != NULL);
}

BOOL CSLIDEPolyline::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  if ( !CSLIDESweepPath::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }
  BOOL bCorrect;
  CListIterInt iter;
  CHAR *pcID;
  CSLIDEPoint *pCSLIDEPoint;
  CSLIDEControlPoint *pCSLIDEControlPoint, *pCSLIDEControlPointHolder;

  bCorrect = TRUE;

  // Check Point References
  if( m_plistPointIDs != NULL )
    {
      m_uiSlices = m_plistPointIDs->GetLength()-1;

      if( m_uiSlices < 1 )
	{
	  HandleError("polyline %s must have atleast two points", m_pcGUIID );
	  bCorrect = FALSE;
	}
      else
	{
	  if( m_plistPointIDs->GetLength() != cuiSlices )
	    {
	      ReallocatePoints();
	      ReallocateScaledWarps();
	      ReallocateWarps();

	      m_bSlicesChanged = TRUE;
	    }

	  m_listCSLIDEPoint.RemoveAll();
	  m_plistCSLIDEControlPoint->RemoveAll();
	  
	  iter.Init(m_plistPointIDs);	  
	  for ( pcID = (CHAR*)(iter.PeekFirst());
		pcID != NULL;
		pcID = (CHAR*)(iter.PeekNext()) )
	    {
	      pCSLIDEPoint = pCSLIDEWorld->PeekPoint(pcID);
	      pCSLIDEControlPoint = pCSLIDEWorld->PeekControlPoint(pcID);

	      if( pCSLIDEPoint != NULL )
		{
		  pCSLIDEControlPointHolder = new CSLIDEControlPoint();
		  pCSLIDEControlPointHolder->SetPointID(pcID);

		  m_listCSLIDEPoint.InsertLast(pCSLIDEPoint, 0);
		  m_plistCSLIDEControlPoint->InsertLast(pCSLIDEControlPointHolder, 0);

		  m_listWarpScale.InsertLast( pCSLIDEControlPointHolder->GetScale(), 0 );
		  m_listWarpRotate.InsertLast( pCSLIDEControlPointHolder->GetRotate(), 0 );
		}
	      else if( pCSLIDEControlPoint != NULL )
		{
		  pCSLIDEPoint = pCSLIDEWorld->PeekPoint(pCSLIDEControlPoint->GetPointID());
		  
		  if ( pCSLIDEPoint == NULL )
		    {
		      HandleError("Undefined %s \"%s\" in %s \"%s\"", 
				  gcppcSLIDEReservedWords[sliderwtPoint], 
				  pCSLIDEControlPoint->GetPointID(),
				  gcppcSLIDEReservedWords[sliderwtControlPoint], 
				  pcID);
		    }
		  m_listCSLIDEPoint.InsertLast(pCSLIDEPoint, 0);
		  m_plistCSLIDEControlPoint->InsertLast(pCSLIDEControlPoint, 0);

		  m_listWarpScale.InsertLast( pCSLIDEControlPoint->GetScale(), 0 );
		  m_listWarpRotate.InsertLast( pCSLIDEControlPoint->GetRotate(), 0 );
		}
	      else
		{
		  HandleError("Undefined %s \"%s\"", 
			      gcppcSLIDEReservedWords[sliderwtPoint], pcID);
		  HandleError("Undefined %s \"%s\"", 
			      gcppcSLIDEReservedWords[sliderwtControlPoint], pcID);
		  bCorrect = FALSE;
		}
	    }
	  m_bPointsDynamic = TRUE;
	}
    }

  return bCorrect;
}

BOOL CSLIDEPolyline::Valid()
{
  if ( !CSLIDESweepPath::Valid() )
    {
      return FALSE;
    }

  return TRUE;
}

VOID CSLIDEPolyline::CheckSweepMorphCrossSectionDynamic()
{
  CListIterInt iter;
  CSLIDEControlPoint *pCSLIDEControlPoint;

  iter.Init(m_plistCSLIDEControlPoint);
  for ( pCSLIDEControlPoint = (CSLIDEControlPoint*)iter.PeekFirst(); 
	pCSLIDEControlPoint != NULL;
	pCSLIDEControlPoint = (CSLIDEControlPoint*)iter.PeekNext() )
    {
      if( pCSLIDEControlPoint->GetCrossSectionDynamic() )
	{
	  m_bSweepMorphCrossSectionDynamic = TRUE;
	  break;
	}
    }
}

VOID CSLIDEPolyline::CheckDynamic()
{
  CSLIDESweepPath::CheckDynamic();

  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;
  CSLIDEControlPoint *pCSLIDEControlPoint;
      
  if( !GetDynamic() )
    {
      SetDynamic(m_dyniMinimizeTorsion.GetDynamic() ||
		 m_dyniClosed.GetDynamic() ||
		 m_dyniDrawPoints.GetDynamic() ||
		 m_dynfAzimuth.GetDynamic() ||
		 m_dynfTwist.GetDynamic() ||
		 m_dyniSymmetry.GetDynamic() );
    }

  // Check if points are dynamic
  if( !GetDynamic() )
    {
      if( m_plistPointIDs != NULL )
	{
	  iter.Init(&m_listCSLIDEPoint);
	  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
		pCSLIDEPoint != NULL;
		pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
	    {
	      if( pCSLIDEPoint->GetBoundDynamic() )
		{
		  SetDynamic(TRUE);
		  break;
		}
	    }
	}     
    }

  // Check if control points are dynamic (scale, warp)
  if( !GetDynamic() )
    {
      iter.Init(m_plistCSLIDEControlPoint);
      for ( pCSLIDEControlPoint = (CSLIDEControlPoint*)iter.PeekFirst(); 
	    pCSLIDEControlPoint != NULL;
	    pCSLIDEControlPoint = (CSLIDEControlPoint*)iter.PeekNext() )
	{
	  if( pCSLIDEControlPoint->GetBoundDynamic() )
	    {
	      SetDynamic(TRUE);
	      break;
	    }
	}
    }
}

VOID CSLIDEPolyline::PreprocessGraphLocal()
{
  DynVector *pdynv;
  BOOL bRealSpaceDynamic;
  BOOL bParameterSpaceDynamic;

  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;
  CSLIDEControlPoint *pCSLIDEControlPoint;
  BOOL bBoundEmpty;
  BOOL bBoundDynamic;
  BOOL bGeometryDynamic;
  BOOL bNormalsDynamic;

  bBoundEmpty = FALSE;
  bBoundDynamic = FALSE;
  bGeometryDynamic = FALSE;
  bNormalsDynamic = FALSE;

  // My points (path points and controlpoints)
  m_bPointsDynamic = FALSE;
  if( m_plistPointIDs != NULL )
    {
      bBoundEmpty = TRUE;
      bBoundDynamic = FALSE;
      bGeometryDynamic = FALSE;
      bNormalsDynamic = FALSE;
      
      iter.Init(&m_listCSLIDEPoint);
      for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	    pCSLIDEPoint != NULL;
	    pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
	{
	  bBoundEmpty = FALSE;
	  bBoundDynamic = bBoundDynamic || pCSLIDEPoint->GetBoundDynamic();
	  bGeometryDynamic = bGeometryDynamic || pCSLIDEPoint->GetGeometryDynamic();
	  bNormalsDynamic = bNormalsDynamic || pCSLIDEPoint->GetNormalDynamic();

	  if(  pCSLIDEPoint->GetGeometryDynamic() )
	    {
	      m_bPointsDynamic = TRUE;
	    }
	}

      // sweepmorph crosssection
      if( m_listCSLIDESweepMorphParent.GetLength() > 0 )
	{
	  iter.Init(m_plistCSLIDEControlPoint);
	  for ( pCSLIDEControlPoint = (CSLIDEControlPoint*)iter.PeekFirst(); 
		pCSLIDEControlPoint != NULL;
		pCSLIDEControlPoint = (CSLIDEControlPoint*)iter.PeekNext() )
	    {
	      pCSLIDEControlPoint->PreprocessGraphLocalCrossSection();

	      bBoundEmpty = FALSE;
	      bBoundDynamic =bBoundDynamic || pCSLIDEControlPoint->GetCrossSectionDynamic();
	      bNormalsDynamic = bNormalsDynamic || pCSLIDEControlPoint->GetCrossSectionDynamic();
	      bGeometryDynamic = bGeometryDynamic || pCSLIDEControlPoint->GetCrossSectionDynamic();
	    }
	}
    }

  // Control Points scale and rotate
  m_bWarpDynamic = FALSE;
  iter.Init(&m_listWarpScale);
  for ( pdynv = (DynVector*)(iter.PeekFirst());
	pdynv != NULL;
	pdynv = (DynVector*)(iter.PeekNext()) )
    {    
      bBoundEmpty = FALSE;
      bBoundDynamic = bBoundDynamic || pdynv->GetDynamic();
      bGeometryDynamic = bGeometryDynamic || pdynv->GetDynamic();
      bNormalsDynamic = bNormalsDynamic || pdynv->GetDynamic();

      if( pdynv->GetDynamic() )
	{
	  m_bWarpDynamic = TRUE;
	}
    }

  iter.Init(&m_listWarpRotate);
  for ( pdynv = (DynVector*)(iter.PeekFirst());
	pdynv != NULL;
	pdynv = (DynVector*)(iter.PeekNext()) )
    {    
      bBoundEmpty = FALSE;
      bBoundDynamic = bBoundDynamic || pdynv->GetDynamic();
      bGeometryDynamic = bGeometryDynamic || pdynv->GetDynamic();
      bNormalsDynamic = bNormalsDynamic || pdynv->GetDynamic();

      if( pdynv->GetDynamic() )
	{
	  m_bWarpDynamic = TRUE;
	}
    }
  
  bRealSpaceDynamic = ( m_dyniDrawPoints.GetDynamic() );
  bParameterSpaceDynamic = ( m_dyniSlices.GetDynamic() ||
			     m_dyniMinimizeTorsion.GetDynamic() ||
			     m_dyniClosed.GetDynamic() ||
			     m_dynfAzimuth.GetDynamic() ||
			     m_dynfTwist.GetDynamic() ||
			     m_dyniSymmetry.GetDynamic() );


  SetGeometryEmpty(FALSE);
  SetGeometryDynamic(bRealSpaceDynamic || 
		     bParameterSpaceDynamic ||
		     bGeometryDynamic ||
		     ((m_pCSLIDESurface != NULL) && m_pCSLIDESurface->GetDynamic()) );
  SetNormalsEmpty(FALSE);
  SetNormalsDynamic(bRealSpaceDynamic || 
		    bParameterSpaceDynamic || 
		    bNormalsDynamic );
  SetBoundsEmpty(FALSE);
  SetBoundsDynamic(bRealSpaceDynamic ||
		   bParameterSpaceDynamic ||
		   bBoundDynamic);
  SetBoundEmpty(bBoundEmpty);
  SetBoundDynamic(bRealSpaceDynamic || 
		  bParameterSpaceDynamic ||
		  bBoundDynamic);


  ////////////////////////////////////////////////////////////////////////////////////
  CSLIDESweep *pCSLIDESweep;
  CSLIDESweepMorph *pCSLIDESweepMorph;
  CSLIDECrossSection *pCSLIDECrossSection;

  iter.Init(&m_listCSLIDECrossSectionParent);  
  for ( pCSLIDECrossSection = (CSLIDECrossSection*)iter.PeekFirst(); 
	pCSLIDECrossSection != NULL;
	pCSLIDECrossSection = (CSLIDECrossSection*)iter.PeekNext() )
    {
      if( GetBoundDynamic() )
	pCSLIDECrossSection->SetBoundDynamic(TRUE);

      if( GetGeometryDynamic() )
	pCSLIDECrossSection->SetGeometryDynamic(TRUE);

      if( GetNormalsDynamic() )
	pCSLIDECrossSection->SetNormalsDynamic(TRUE);
    }

  iter.Init(&m_listCSLIDESweepParent);  
  for ( pCSLIDESweep = (CSLIDESweep*)iter.PeekFirst(); 
	pCSLIDESweep != NULL;
	pCSLIDESweep = (CSLIDESweep*)iter.PeekNext() )
    {
      if( GetBoundDynamic() )
	pCSLIDESweep->SetBoundDynamic(TRUE);

      if( GetGeometryDynamic() )
	pCSLIDESweep->SetGeometryDynamic(TRUE);

      if( GetNormalsDynamic() )
	pCSLIDESweep->SetNormalsDynamic(TRUE);
    }

  iter.Init(&m_listCSLIDESweepMorphParent);  
  for ( pCSLIDESweepMorph = (CSLIDESweepMorph*)iter.PeekFirst(); 
	pCSLIDESweepMorph != NULL;
	pCSLIDESweepMorph = (CSLIDESweepMorph*)iter.PeekNext() )
    {
      if( GetBoundDynamic() )
	pCSLIDESweepMorph->SetBoundDynamic(TRUE);

      if( GetGeometryDynamic() )
	pCSLIDESweepMorph->SetGeometryDynamic(TRUE);

      if( GetNormalsDynamic() )
	pCSLIDESweepMorph->SetNormalsDynamic(TRUE);
    }
}

VOID CSLIDEPolyline::UpdateCrossSectionParents()
{
  CListIterInt iter;
  CSLIDECrossSection *pCSLIDECrossSection;

  iter.Init(&m_listCSLIDECrossSectionParent);
  for ( pCSLIDECrossSection = (CSLIDECrossSection*)iter.PeekFirst(); 
	pCSLIDECrossSection != NULL;
	pCSLIDECrossSection = (CSLIDECrossSection*)iter.PeekNext() )
    {
      pCSLIDECrossSection->UpdateCrossSectionPoints(m_uiSlices+1, m_pPoint);
    }
}

VOID CSLIDEPolyline::UpdateSweepPathParents()
{
  CListIterInt iter;
  CSLIDESweep *pCSLIDESweep;
  CSLIDESweepMorph *pCSLIDESweepMorph;

  iter.Init(&m_listCSLIDESweepParent);
  for ( pCSLIDESweep = (CSLIDESweep*)iter.PeekFirst(); 
	pCSLIDESweep != NULL;
	pCSLIDESweep = (CSLIDESweep*)iter.PeekNext() )
    {
      pCSLIDESweep->UpdatePathPoints();
    }

  iter.Init(&m_listCSLIDESweepMorphParent);
  for ( pCSLIDESweepMorph = (CSLIDESweepMorph*)iter.PeekFirst(); 
	pCSLIDESweepMorph != NULL;
	pCSLIDESweepMorph = (CSLIDESweepMorph*)iter.PeekNext() )
    {
      pCSLIDESweepMorph->UpdatePathPoints();
    }
}

VOID CSLIDEPolyline::UpdateState()
{
  CSLIDESweepPath::UpdateState();

  UINT_32 i;
  FLOAT f;
  Point pt;
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;
  BOOL bPointsChanged, bWarpChanged, bTorsionChanged;

  bPointsChanged = FALSE;
  bWarpChanged = FALSE;
  bTorsionChanged = FALSE;

  if( (m_plistPointIDs != NULL) && (m_bPointsDynamic == TRUE) )
    {
      iter.Init(&m_listCSLIDEPoint);

      for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(), i = 0; 
	    pCSLIDEPoint != NULL;
	    pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext(), i++ )
	{
	  ASSERT( i <= m_uiSlices );
	  if( m_pPoint[i] != pCSLIDEPoint->GetPoint() )
	    {
	      m_pPoint[i] = pCSLIDEPoint->GetPoint();
	      bPointsChanged = TRUE;
	    }
	}
    }

  // Update all sweeps that use me
  if( bPointsChanged )
    {
      UpdateCrossSectionParents();
    }

  /*
  if( bPointsChanged )
    {
      Point pt1, pt2;

      f = 0;
      for( i = 0; i < m_uiSlices; i++)
	{
	  pt1 = m_pPoint[i];
	  pt2 = m_pPoint[i+1];

	  f += (pt2 - pt1).Magnitude();
	}
      printf( "polyline \"%s\" has a total length of %f\n", m_pcGUIID, f);
    }
    */

  if( m_bWarpDynamic )
    {
      FLOAT fx, fy, fz, fSx, fSy, fSz, fRx, fRy, fRz;
      DynVector *pdynv;

      iter.Init(&m_listWarpScale);
      for ( pdynv = (DynVector*)(iter.PeekFirst()), i = 0;
	    pdynv != NULL;
	    pdynv = (DynVector*)(iter.PeekNext()), i++ )
	{  
	  fSx = m_pWarpScale[i][0];
	  fSy = m_pWarpScale[i][1];
	  fSz = m_pWarpScale[i][2];

	  pdynv->Update();

	  fx = (pdynv->Get())[X];
	  fy = (pdynv->Get())[Y];
	  fz = (pdynv->Get())[Z];

	  if( (fSx != fx) || (fSy != fy) || (fSz != fz) )
	    {
	      m_pWarpScale[i].Set( fx, fy, fz );
	      bWarpChanged = TRUE;
	    }
	}
      
      iter.Init(&m_listWarpRotate);
      for ( pdynv = (DynVector*)(iter.PeekFirst()), i = 0;
	    pdynv != NULL;
	    pdynv = (DynVector*)(iter.PeekNext()), i++ )
	{  
	  fRx = m_pWarpRotate[i][0];
	  fRy = m_pWarpRotate[i][1];
	  fRz = m_pWarpRotate[i][2];

	  pdynv->Update();

	  fx = (pdynv->Get())[X];
	  fy = (pdynv->Get())[Y];
	  fz = (pdynv->Get())[Z];

	  if( (fRx != fx) || (fRy != fy) || (fRz != fz) )
	    {
	      m_pWarpRotate[i].Set( DEGTORAD(fx), DEGTORAD(fy), DEGTORAD(fz) );
	      bWarpChanged = TRUE;
	    }
	}
    }
  
  i = m_dyniMinimizeTorsion.GetValue();
  m_dyniMinimizeTorsion.Update();
  bTorsionChanged = (i != (UINT_32)m_dyniMinimizeTorsion.GetValue());

  i = m_dyniClosed.GetValue();
  m_dyniClosed.Update();
  bTorsionChanged = (i != (UINT_32)m_dyniClosed.GetValue()) || bTorsionChanged;

  f = m_dynfAzimuth.GetValue();
  m_dynfAzimuth.Update();
  bTorsionChanged = (f != (FLOAT)m_dynfAzimuth.GetValue()) || bTorsionChanged;

  f = m_dynfTwist.GetValue();
  m_dynfTwist.Update();
  bTorsionChanged = (f != (FLOAT)m_dynfTwist.GetValue()) || bTorsionChanged;

  i = m_dyniSymmetry.GetValue();
  m_dyniSymmetry.Update();
  bTorsionChanged = (i != (UINT_32)m_dyniSymmetry.GetValue()) || bTorsionChanged;

  m_dyniDrawPoints.Update();

  if( m_bSlicesChanged )
    {      
      m_bSlicesChanged = FALSE;
      
      ReallocateNormals();
      ReallocateBinormals();
      ReallocateTangents();
      ReallocateNormalsTorsionReduced();
      ReallocateBinormalsTorsionReduced();
      ReallocateTangentsTorsionReduced();
      ReallocateSweepFrenetFrames();

      CalculateFrenetFrames();
      CalculateScaledWarp();
      CalculateTorsion();

      UpdateSweepPathParents();
    }
  else if( bPointsChanged )
    {
      CalculateFrenetFrames();
      CalculateScaledWarp();
      CalculateTorsion();

      UpdateSweepPathParents();
    }
  else if( bWarpChanged || bTorsionChanged )
    {
      CalculateScaledWarp();
      CalculateTorsion();

      UpdateSweepPathParents();
    }
}

VOID CSLIDEPolyline::CalculateBound()
{
  UINT_32 uiSlice;
  CBound bound;

  for( uiSlice = 0; uiSlice < m_uiSlices+1; uiSlice++)
    {
      bound.Insert(m_pPoint[uiSlice]);
    }

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);

      // Calculate normal length
      m_fNormalLength = bound.GetRadius() * cfNormalRatio;
    }
}

////////////////////////////////////////////////////////////////////////////////////////////
VOID CSLIDEPolyline::CalculateFrenetFrame(Vector &vT, Vector &vB, Vector &vN,
					  Point pt0, Point pt1, Point pt2, UINT_32 uiSlice)
{
  Vector v;

  vB = ((pt2-pt1) * (pt0-pt1));
  
  // collinear points
  if( vB.Magnitude() <= EPSILON ) 
    {	  
      // Tangent
      vT = pt2 - pt1;
      vT.Normalize();
      
      // check for zero tangent => repeated point
      if( vT.Magnitude() <= EPSILON )
	{
	  vT = m_pvTangent[uiSlice-1];

	  //printf( "Warning: Polyline %s has consecutive points with (x y z) = (%.3f %.3f %.3f).\n", m_pcID, pt1[X], pt1[Y], pt1[Z]);
	}
      
      // Binormal
      v.Set(0, 1, 0);
      vB = v - (v.Dot(vT)) * vT;
      vB.Normalize();
      
      // if vT = (0 1 0) then pick (1 0 0)
      if( vB.Magnitude() <= EPSILON )
	{
	  v.Set(1, 0, 0);
	  vB = v - (v.Dot(vT)) * vT;
	  vB.Normalize();
	}
      
      // Normal
      vN = vB * vT;
      vN.Normalize();
    }
  else // non-collinear
    {
      // Binormal
      vB.Normalize();
      
      // Normal
      v = pt0 - pt1;
      v.Normalize();
      vN = pt2 - pt1;	  
      vN.Normalize();
      vN = (v + vN) / 2.0;
      vN.Normalize();
	  
      // Tangent
      vT = vN * vB;
      vT.Normalize();
    }
}

VOID CSLIDEPolyline::CalculateFrenetFrames()
{
  UINT_32 uiSlice;
  Vector vT, vB, vN, v;
 
  if( m_uiSlices == 1 ) // Two Points
    {
      CalculateFrenetFrame(vT, vB, vN, m_pPoint[0], m_pPoint[0], m_pPoint[1], 0);
      m_pvTangent[0]  = vT;
      m_pvBinormal[0] = vB;	  
      m_pvNormal[0]   = vN;

      m_pvTangent[1]  = vT;
      m_pvBinormal[1] = vB;	  
      m_pvNormal[1]   = vN;
      
      return;
    }

  
  // All Frames except first and last
  for( uiSlice = 1; uiSlice < m_uiSlices; uiSlice++ )
    {
      CalculateFrenetFrame(vT, vB, vN, m_pPoint[uiSlice-1], m_pPoint[uiSlice], 
			   m_pPoint[uiSlice+1], uiSlice);
      
      m_pvTangent[uiSlice]  = vT;
      m_pvBinormal[uiSlice] = vB;	  
      m_pvNormal[uiSlice]   = vN;
    }

  // First and Last Frame
  if( m_pPoint[0] == m_pPoint[m_uiSlices] ) // closed => endpoint binormals should align
    {
      CalculateFrenetFrame(vT, vB, vN, m_pPoint[m_uiSlices-1], m_pPoint[0], m_pPoint[1],
			   uiSlice);

      m_pvTangent[0]  = vT;
      m_pvBinormal[0] = vB;
      m_pvNormal[0]   = vN;

      m_pvTangent[m_uiSlices]  = vT;
      m_pvBinormal[m_uiSlices] = vB;
      m_pvNormal[m_uiSlices]   = vN;
    }
  else // not closed
    {
      // First frame
      vT = m_pPoint[1] - m_pPoint[0];
      vT.Normalize();

      // check for zero tangent => repeated point
      if( vT.Magnitude() <= EPSILON )
        {
          vT = m_pvTangent[1];
        }
      m_pvTangent[0] = vT;
      
      vB = m_pvBinormal[1];
      m_pvBinormal[0] = vB; 
      
      vN = vB * vT;
      vN.Normalize();
      m_pvNormal[0] = vN;
      
      // Last frame
      vT = m_pPoint[m_uiSlices] - m_pPoint[m_uiSlices-1];
      vT.Normalize();

      // check for zero tangent => repeated point
      if( vT.Magnitude() <= EPSILON )
        {
          vT = m_pvTangent[m_uiSlices-1];
        }
      m_pvTangent[m_uiSlices] = vT;
      
      vB = m_pvBinormal[m_uiSlices-1];
      m_pvBinormal[m_uiSlices] = vB; 
      
      vN = vB * vT;
      vN.Normalize();
      m_pvNormal[m_uiSlices] = vN;
    }
}

VOID CSLIDEPolyline::CalculateScaledWarp()
{
  // Adjust the scale values so swept objects will have uniform size

  if( m_uiSlices >= 2 )
    {      
      UINT_32 uiSlice;
      Vector vPrev, vPost, v;
      DOUBLE dTheta, s;

      // All points except first and last
      for( uiSlice = 1; uiSlice < m_uiSlices; uiSlice++ )
	{
	  vPrev = m_pPoint[uiSlice-1] - m_pPoint[uiSlice];
	  vPost = m_pPoint[uiSlice+1] - m_pPoint[uiSlice];

	  v = vPrev*vPost;
	  // Make sure points aren't colinear
	  if( !(v.Magnitude() <= EPSILON) )
	    {
	      dTheta = AngleBetween( vPrev, vPost ) / 2.0;
	      
	      if( sin(dTheta) != 0.0 )
		{
		  s = 1 / sin(dTheta);
		  m_pScaledWarp[uiSlice][0] = s;
		}
	      else
		{
		  m_pScaledWarp[uiSlice][0] = 1.0;
		}
	    }
	}

      // If closed => scale endpoints
      if( m_pPoint[0] == m_pPoint[m_uiSlices] ) 
	{
	  vPrev = m_pPoint[m_uiSlices-1] - m_pPoint[0];
	  vPost = m_pPoint[1] - m_pPoint[0];
	  
	  v = vPrev*vPost;
	  // Make sure points aren't colinear
	  if( !(v.Magnitude() <= EPSILON) )
	    {
	      dTheta = AngleBetween( vPrev, vPost ) / 2.0;
	      
	      if( sin(dTheta) != 0.0 )
		{
		  s = 1 / sin(dTheta);
		  m_pScaledWarp[0][0] = s;
		  m_pScaledWarp[m_uiSlices][0] = s;
		}
	      else
		{
		  m_pScaledWarp[0][0] = 1.0;
		  m_pScaledWarp[m_uiSlices][0] = 1.0;
		}
	    }
	}
    }
}

VOID CSLIDEPolyline::CalculateTorsion()
{
  UINT_32 uiSlice;
  DOUBLE dTheta, dSymmetryAngle, dPhi, t;
  Point  p1, p2, c1, c2;
  Vector v, v1, v2, vT1, vT2, vB1, vB2, vN1, vN2, vPlane;
  Matrix m, m1, m2, m3, m4, m5;
  MATRIX4D mm;

  for( uiSlice = 0; uiSlice < m_uiSlices + 1; uiSlice++ )
    {
      m_pvNormalTorsionReduced[uiSlice] = m_pvNormal[uiSlice];
      m_pvBinormalTorsionReduced[uiSlice] = m_pvBinormal[uiSlice];
      m_pvTangentTorsionReduced[uiSlice] = m_pvTangent[uiSlice];
    }

  // Minimize Torsion
  if( m_dyniMinimizeTorsion.GetValue() != 0 )
    {
      // better mintorsion scheme
      dPhi = 0.0;
      
      for( uiSlice = 0; uiSlice < m_uiSlices; uiSlice++ )
	{
	  p1 = m_pPoint[uiSlice];
	  p2 = m_pPoint[uiSlice+1];

	  if( p1 != p2 )
	    {
	      vT2 = m_pvTangentTorsionReduced[uiSlice+1];
	      vN1 = m_pvNormalTorsionReduced[uiSlice];
	      vN2 = m_pvNormalTorsionReduced[uiSlice+1];
	      
	      v  = p2 - p1;
	      vN1.Normalize();
	      c1 = p1 + vN1;
	      
	      vPlane = vT2;
	      vPlane.CompletePlaneEquation(p2);
	      
	      t = -((vPlane[X]*c1[X] + vPlane[Y]*c1[Y] + vPlane[Z]*c1[Z] + vPlane[W]) /
		    (vPlane[X]*v[X] + vPlane[Y]*v[Y] + vPlane[Z]*v[Z]));
	      
	      c2 = c1 + t*v;
	      v2 = c2 - p2;
	      
	      dTheta = AngleBetween( vN2, v2 );
	      v = vN2 * v2;
	      if( v.Dot(vT2) < 0.0 )
		{
		  dTheta *= -1.0;
		}
	      m.SetRotate( vT2, dTheta );
	      
	      m_pvNormalTorsionReduced[uiSlice+1] = m * m_pvNormalTorsionReduced[uiSlice+1];
	      m_pvBinormalTorsionReduced[uiSlice+1] = m * m_pvBinormalTorsionReduced[uiSlice+1];
	    }
	  dPhi += dTheta;
	}

      /*
      // Original min torsion scheme
      dPhi = 0.0;

      for( uiSlice = 0; uiSlice < m_uiSlices; uiSlice++ )
	{
	  vT1 = m_pvTangent[uiSlice];
	  vT2 = m_pvTangent[uiSlice+1];
	  vB1 = m_pvBinormalTorsionReduced[uiSlice];
	  vB2 = m_pvBinormal[uiSlice+1];
	  
	  // Align Tangents  1->2
	  v = vT1*vT2;
	  dTheta = AngleBetween( vT1, vT2 );
	  m.SetRotate( v, dTheta );
	  vB1 = m * vB1;
	  
	  // Find angle between binormals 2->1
	  v = vB2 * vB1;
	  dTheta = AngleBetween( vB1, vB2 );
	  if( v.Dot(vT2) < 0.0 )
	    {
	      dTheta *= -1.0;
	    }
	  m.SetRotate( vT2, dTheta );
	  
	  m_pvNormalTorsionReduced[uiSlice+1] = m * m_pvNormalTorsionReduced[uiSlice+1];
	  m_pvBinormalTorsionReduced[uiSlice+1] = m * m_pvBinormalTorsionReduced[uiSlice+1];

	  dPhi += dTheta;
	}
      printf( "polyline \"%s\" has an intrisic twist of %f degrees\n", m_pcGUIID, RADTODEG(dPhi));
      */
    }

  // Closed Curve - Calculate overall torsion
  if( m_dyniClosed.GetValue() != 0 )
    {
      // Assumes that first and last points are the same
      vN1 = m_pvNormalTorsionReduced[0];
      vN2 = m_pvNormalTorsionReduced[m_uiSlices];

      // Find angle between normals 2->1
      v = vN2 * vN1;
      dTheta = AngleBetween( vN1, vN2 );
      
      // Use Symmetry to close with minimal twisting
      DOUBLE ds = m_dyniSymmetry.GetValue();
      if( ds < 1.0 )  // After CheckSemantics sym => 0 ???
	{	 
	  ds = 1.0;
	}

      dSymmetryAngle = 2.0 * PI / ds;
      dTheta = fmod(dTheta, dSymmetryAngle);

      if( dTheta != 0.0 )
	{
	  if( dTheta > (dSymmetryAngle/2.0) )
	    {
	      dTheta = dTheta - dSymmetryAngle;
	    }

	  if( v.Dot(vT2) < 0.0 ) // make sure orientation is correct
	    {
	      dTheta *= -1.0;
	    }
	  dTheta = dTheta / (DOUBLE)m_uiSlices;
	  
	  for( uiSlice = 1; uiSlice < m_uiSlices + 1; uiSlice++ )
	    {
	      m.SetRotate( m_pvTangentTorsionReduced[uiSlice], (DOUBLE)uiSlice*dTheta);
	      
	      m_pvNormalTorsionReduced[uiSlice] = m * m_pvNormalTorsionReduced[uiSlice];
	      m_pvBinormalTorsionReduced[uiSlice] = m * m_pvBinormalTorsionReduced[uiSlice];
	    }
	}
    }
      
  // Azimuth
  if( m_dynfAzimuth.GetValue() != 0.0 )
    {
      dTheta = DEGTORAD(m_dynfAzimuth.GetValue());
      
      for( uiSlice = 0; uiSlice < m_uiSlices+1; uiSlice++ )
	{
	  m.SetRotate( m_pvTangentTorsionReduced[uiSlice], dTheta );
	  
	  m_pvNormalTorsionReduced[uiSlice] = m * m_pvNormalTorsionReduced[uiSlice];
	  m_pvBinormalTorsionReduced[uiSlice] = m * m_pvBinormalTorsionReduced[uiSlice];
	}
    }

  // Twist
  if( m_dynfTwist.GetValue() != 0.0 )
    {
      dTheta = DEGTORAD(m_dynfTwist.GetValue());
      dTheta = dTheta / (DOUBLE)m_uiSlices;

      for( uiSlice = 1; uiSlice < m_uiSlices + 1; uiSlice++ )
	{
	  m.SetRotate( m_pvTangentTorsionReduced[uiSlice], (DOUBLE)uiSlice*dTheta);

	  m_pvNormalTorsionReduced[uiSlice] = m * m_pvNormalTorsionReduced[uiSlice];
	  m_pvBinormalTorsionReduced[uiSlice] = m * m_pvBinormalTorsionReduced[uiSlice];
	}
    }

  // User specified rotation
  for( uiSlice = 0; uiSlice < m_uiSlices+1; uiSlice++ )
    {
      m.SetRotate( m_pvNormalTorsionReduced[uiSlice], m_pWarpRotate[uiSlice][X] );
      m_pvBinormalTorsionReduced[uiSlice] = m * m_pvBinormalTorsionReduced[uiSlice];
      m_pvTangentTorsionReduced[uiSlice] = m * m_pvTangentTorsionReduced[uiSlice];

      m.SetRotate( m_pvBinormalTorsionReduced[uiSlice], m_pWarpRotate[uiSlice][Y] );
      m_pvNormalTorsionReduced[uiSlice] = m * m_pvNormalTorsionReduced[uiSlice];
      m_pvTangentTorsionReduced[uiSlice] = m * m_pvTangentTorsionReduced[uiSlice];

      m.SetRotate( m_pvTangentTorsionReduced[uiSlice], m_pWarpRotate[uiSlice][Z] );
      m_pvNormalTorsionReduced[uiSlice] = m * m_pvNormalTorsionReduced[uiSlice];
      m_pvBinormalTorsionReduced[uiSlice] = m * m_pvBinormalTorsionReduced[uiSlice];
    }

  // Set Sweep Frenet Frames
  for( uiSlice = 0; uiSlice < m_uiSlices+1; uiSlice++)
    {
      // Scale specified by the user in Torsion changed Frenet Frame
      v.Set(m_pWarpScale[uiSlice][X], m_pWarpScale[uiSlice][Y], m_pWarpScale[uiSlice][Z]);
      m1.SetScale(v);
      
      // Scale in direction that counteracts torsion reduction
      v1 = m_pvNormal[uiSlice];
      v2 = m_pvNormalTorsionReduced[uiSlice];
      dTheta = AngleBetween( v2, v1 );
      v = v2 * v1;
      if( v.Dot(m_pvTangentTorsionReduced[uiSlice]) < 0.0 ) // make sure orientation correct
	{
	  dTheta *= -1.0;
	}
      v.Set(0, 0, 1);
      m2.SetRotate(v, -dTheta);

      // Scale to maintain crosssection shape in intrinstic Frenet Frame
      v.Set(m_pScaledWarp[uiSlice][X], m_pScaledWarp[uiSlice][Y], 1.0);
      m3.SetScale(v);

      v.Set(0, 0, 1);
      m4.SetRotate(v, dTheta);

      // Set Change of Basis Matrix to torsion reduced Frenet Frame
      mm[0]  = m_pvNormalTorsionReduced[uiSlice][X];
      mm[1]  = m_pvNormalTorsionReduced[uiSlice][Y];
      mm[2]  = m_pvNormalTorsionReduced[uiSlice][Z];
      mm[3]  = m_pvNormalTorsionReduced[uiSlice][W];
      mm[4]  = m_pvBinormalTorsionReduced[uiSlice][X];
      mm[5]  = m_pvBinormalTorsionReduced[uiSlice][Y];
      mm[6]  = m_pvBinormalTorsionReduced[uiSlice][Z];
      mm[7]  = m_pvBinormalTorsionReduced[uiSlice][W];
      mm[8]  = m_pvTangentTorsionReduced[uiSlice][X];
      mm[9]  = m_pvTangentTorsionReduced[uiSlice][Y];
      mm[10] = m_pvTangentTorsionReduced[uiSlice][Z];
      mm[11] = m_pvTangentTorsionReduced[uiSlice][W];
      mm[12] = 0;
      mm[13] = 0;
      mm[14] = 0;
      mm[15] = 1;
      m5.Set(mm);

      m_pmSweepFrenetFrame[uiSlice] = m5 * m4 * m3 * m2 * m1;
    }
}

////////////////////////////////////////////////////////////////////////////////////////////

VOID CSLIDEPolyline::RenderGeometrySmooth(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{
  UINT_32 i;

  glPushAttrib(GL_LINE_BIT | GL_COLOR_BUFFER_BIT);
  {
    gilLineWidth(2.0);
    
    // Lines
    gilBegin(GL_LINE_STRIP);
    {
      for( i = 0; i < m_uiSlices + 1; i++ )
	{
	  gilVertex3fv(m_pPoint[i].Get());
	}
    }
    gilEnd();
  }
  PopAuxillaryGeometryAttributes();
  
  // Points
  if( m_dyniDrawPoints.GetValue() != 0 )
    {
      GLfloat v[4] = {0.0, 0.0, 0.0, 1.0};
      gilGetFloatv(GL_POINT_SIZE, v);

      gilPointSize(3.0);
      {	  
	gilColor3f(1, 1, 0);

	gilBegin(GL_POINTS);
	{
	  for( i = 0; i < m_uiSlices + 1; i++ )
	    {
	      gilVertex3fv(m_pPoint[i].Get());
	    }
	}
	gilEnd();
      }
      gilPointSize(v[0]);
    }
}


VOID CSLIDEPolyline::RenderGeometryFlat(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{
  RenderGeometrySmooth();
}

VOID CSLIDEPolyline::RenderNormalsSmooth()
{
  UINT_32 i;
  Point pt;
  
  // Render Frenet Frame
  PushAuxillaryGeometryAttributes();
  {
    gilLineWidth(2.0);

    // Tangent
    gilColor3f(0, 0, 1);  // Blue    
    gilBegin(GL_LINES);
    {
      for(i = 0; i < m_uiSlices + 1; i++ ) 
	{
	  gilVertex3fv(m_pPoint[i].Get());
	  pt = m_pPoint[i] + m_pvTangentTorsionReduced[i] * m_fNormalLength;
	  gilVertex3fv(pt.Get());
	}
    }
    gilEnd();

    // Normal
    gilColor3f(1, 0, 0);  // Red    
    gilBegin(GL_LINES);
    {
      for(i = 0; i < m_uiSlices + 1; i++ ) 
	{
	  gilVertex3fv(m_pPoint[i].Get());
	  pt = m_pPoint[i] + m_pvNormalTorsionReduced[i] * m_fNormalLength;
	  gilVertex3fv(pt.Get());
	}
    }
    gilEnd();

    // Binormal
    gilColor3f(0, 1, 0);  // Green
    gilBegin(GL_LINES);
    {
      for(i = 0; i < m_uiSlices + 1; i++ ) 
	{
	  gilVertex3fv(m_pPoint[i].Get());
	  pt = m_pPoint[i] + m_pvBinormalTorsionReduced[i] * m_fNormalLength;
	  gilVertex3fv(pt.Get());
	}
    }
    gilEnd();

  }
  PopAuxillaryGeometryAttributes();
}

VOID CSLIDEPolyline::RenderNormalsFlat()
{
  RenderNormalsSmooth();
}

VOID CSLIDEPolyline::WriteSLIDE(FILE *fpOutputFile)
{
  UINT_32 i;
  CListIterInt iter;
  CHAR *pcID;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", gcppcSLIDEReservedWords[sliderwtPolyline], m_pcGUIID);
  WriteSLIDEFlags(fpOutputFile, 0);

  if( m_plistPointIDs != NULL )
    {
      fprintf(fpOutputFile, "  %s (",
	      gcppcSLIDEReservedWords[sliderwtPointList]);

      iter.Init(m_plistPointIDs);
      
      for ( pcID = (CHAR*)(iter.PeekFirst()), i = 0;
	    pcID != NULL;
	    pcID = (CHAR*)(iter.PeekNext()), i++ )
	{
	  fprintf(fpOutputFile, " %s", pcID );
	}
    }
  else
    {
      const FLOAT* pt;

      fprintf(fpOutputFile, "# Default Control Points:\n");
      fprintf(fpOutputFile, "# %s (\n#\t",
	      gcppcSLIDEReservedWords[sliderwtPointList]);

      for( i = 0; i < m_uiSlices+1; i++ )
	{
	  pt = m_pPoint[i].Get();
	  fprintf(fpOutputFile, "(%.3f %.3f %.3f)", pt[X], pt[Y], pt[Z] );
	}
    }
  fprintf(fpOutputFile, " )\n");

  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtDrawPoints],
	  m_dyniDrawPoints.GetValue());

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndPolyline]);
}

VOID CSLIDEPolyline::WriteRIBGeometry(FILE *fpOutputFile,
				    const Matrix &/*mGproj_obj*/, 
				    CBound * /*pCBoundClip*/, 
				    SLIDEShadingType eShading,
				    SLIDELODType eLOD)
{
}

VOID CSLIDEPolyline::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE Polyline: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);
}

VOID CSLIDEPolyline::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE Polyline: %s", m_pcGUIID);
}

VOID CSLIDEPolyline::WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				       CSLIDESurface *pCSLIDESurface)
{
}

VOID CSLIDEPolyline::WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform)
{
}

