#ifndef __SLIDEPOLYLINE_H__
#define __SLIDEPOLYLINE_H__

#include "slidesweeppath.h"

//////////////////////////////////////////////////////////////////////////////
// CSLIDEPolyline Class
//

class CSLIDEPolyline : public CSLIDESweepPath
{
//////////////////////////////////////////////////
// Friends
//
public: friend class CSLIDECrossSection;

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEPolyline();
  virtual ~CSLIDEPolyline();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID Uninit();
  virtual BOOL Initialized();

  virtual VOID InitPoints();
  virtual VOID InitWarps();
  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);
  virtual BOOL Valid();

  virtual inline VOID SetPointList(CListInt *plistPoint);
  virtual inline VOID SetDrawPoints(DynInt *pdyniDrawPoints);

protected:

  virtual VOID ReallocatePoints();
  virtual VOID DeallocatePoints();
  virtual VOID ReallocateWarps();
  virtual VOID DeallocateWarps();
  virtual VOID ReallocateScaledWarps();
  virtual VOID DeallocateScaledWarps();

  virtual VOID ReallocateNormals();
  virtual VOID DeallocateNormals();
  virtual VOID ReallocateBinormals();
  virtual VOID DeallocateBinormals();
  virtual VOID ReallocateTangents();
  virtual VOID DeallocateTangents();
  virtual VOID ReallocateNormalsTorsionReduced();
  virtual VOID DeallocateNormalsTorsionReduced();
  virtual VOID ReallocateBinormalsTorsionReduced();
  virtual VOID DeallocateBinormalsTorsionReduced();
  virtual VOID ReallocateTangentsTorsionReduced();
  virtual VOID DeallocateTangentsTorsionReduced();
  virtual VOID ReallocateSweepFrenetFrames();
  virtual VOID DeallocateSweepFrenetFrames();

//////////////////////////////////////////////////
// Geometric Computations
//
protected:
  virtual VOID CheckSweepMorphCrossSectionDynamic();

  virtual VOID CheckDynamic();
  virtual VOID PreprocessGraphLocal();
  virtual VOID UpdateState();
  virtual VOID CalculateBound();

  VOID UpdateCrossSectionParents();
  VOID UpdateSweepPathParents();

protected:
  virtual VOID CalculateFrenetFrame(Vector &vT, Vector &vB, Vector &vN, 
				    Point pt0, Point pt1, Point pt2, UINT_32 uiSlice);
  virtual VOID CalculateFrenetFrames();
  virtual VOID CalculateTorsion();
  virtual VOID CalculateScaledWarp();

//////////////////////////////////////////////////
// Rendering
//
protected:
  virtual VOID RenderGeometrySmooth(CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderGeometryFlat(CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderNormalsSmooth();
  virtual VOID RenderNormalsFlat();

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

  virtual VOID WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				 CSLIDESurface *pCSLIDESurface);
  virtual VOID WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline CHAR *GetTypeString();
  virtual inline SLIDEReservedWordType GetType();
  
  virtual inline UINT_32 GetControlPoints();
  virtual inline Matrix *GetFrenetMatrix(); 

  virtual inline CListInt *GetControlPointList();
  virtual inline CListInt *GetControlPointIDList();

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  FLOAT m_fNormalLength;
  DynInt m_dyniDrawPoints;

  BOOL m_bSlicesChanged;

  // Warp
  BOOL m_bWarpDynamic;
  CListInt m_listWarpScale;   // DynVectors
  CListInt m_listWarpRotate;  // DynVectors
  Point *m_pWarpScale;        // [Sx, Sy, Sz]
  Point *m_pWarpRotate;       // [Rx, Ry, Rz]
  Point *m_pScaledWarp;

  // Curve
  BOOL m_bPointsDynamic;
  CListInt *m_plistPointIDs;
  CListInt m_listCSLIDEPoint;
  CListInt *m_plistCSLIDEControlPoint;

  // Frenet Frame
  Vector *m_pvNormal;
  Vector *m_pvBinormal;
  Vector *m_pvTangent;
  Vector *m_pvNormalTorsionReduced;
  Vector *m_pvBinormalTorsionReduced;
  Vector *m_pvTangentTorsionReduced;  // Add for rotate (x, y, z)
  Matrix *m_pmSweepFrenetFrame;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDEPolyline::SetPointList(CListInt *plistPoint)
{
  m_plistPointIDs = plistPoint;
}

inline VOID CSLIDEPolyline::SetDrawPoints(DynInt *pdyniDrawPoints)
{
  m_dyniDrawPoints.Set(pdyniDrawPoints);
}

inline CHAR *CSLIDEPolyline::GetTypeString()
{
  return gcppcSLIDEReservedWords[sliderwtPolyline];
}

inline SLIDEReservedWordType CSLIDEPolyline::GetType()
{
  return sliderwtPolyline;
}

inline UINT_32 CSLIDEPolyline::GetControlPoints() 
{
  return m_uiSlices+1;
}

inline Matrix *CSLIDEPolyline::GetFrenetMatrix() 
{
  return m_pmSweepFrenetFrame;
}

inline CListInt *CSLIDEPolyline::GetControlPointList()
{
  return m_plistCSLIDEControlPoint;
}

inline CListInt *CSLIDEPolyline::GetControlPointIDList()
{
  return m_plistPointIDs;
}

#endif // __SLIDEPOLYLINE_H__
