#include "slideprimitive.h"
#include "slideinstance.h"
#include "slideworld.h"
#include "slidecylinder.h"


FILE *CSLIDEPrimitive::m_fpErrorFile = stderr;
UINT_32 CSLIDEPrimitive::m_uiErrorCount = 0;
UINT_32 CSLIDEPrimitive::m_uiWarningCount = 0;

BOOL CSLIDEPrimitive::m_bRenderClipBoundEnabled = FALSE;
BOOL CSLIDEPrimitive::m_bFogEnabled = TRUE;
BOOL CSLIDEPrimitive::m_bLightingEnabled = FALSE;
BOOL CSLIDEPrimitive::m_bTextureMappingEnabled = FALSE;
BOOL CSLIDEPrimitive::m_bCrystalBallEnabled = TRUE;
int CSLIDEPrimitive::m_iWindowCUIType = 3;
FLOAT CSLIDEPrimitive::m_fWindowCUISpeed = 1.0;
BOOL CSLIDEPrimitive::m_bSelectionEnabled = FALSE;

BOOL CSLIDEPrimitive::m_bRenderBoundsEnabled = FALSE;
BOOL CSLIDEPrimitive::m_bColorDynamicBounds = FALSE;
BOOL CSLIDEPrimitive::m_bColorChangedBounds = FALSE;
BOOL CSLIDEPrimitive::m_bCheckBoundChanged = TRUE;
BOOL CSLIDEPrimitive::m_bBoundRejectEnabled = TRUE;

BOOL CSLIDEPrimitive::m_bBackFaceCullingEnabled = TRUE;
BOOL CSLIDEPrimitive::m_bClipEnabled = TRUE;
BOOL CSLIDEPrimitive::m_bRenderNormalsEnabled = FALSE;
FLOAT CSLIDEPrimitive::m_fNormalsRenderLength = (FLOAT)0.2;

BOOL CSLIDEPrimitive::m_bRenderCDAOperandsEnabled = FALSE;

CSLIDECamera *CSLIDEPrimitive::m_pCSLIDECameraCurr = NULL;

UINT_32 CSLIDEPrimitive::m_uiAttributeDepthMax = 16;
UINT_32 CSLIDEPrimitive::m_uiAttributeDepthCurr = 0;

CSLIDECylinder *CSLIDEPrimitive::m_pCSLIDECylinderAxis = new CSLIDECylinder();

CSLIDEPrimitive::CSLIDEPrimitive()
{
  m_pcFile = NULL;
  m_uiLine = 0;
  m_uiColumn = 0;

  m_pcID = NULL;
  m_pcGUIID = NULL;
  m_bDynamic = FALSE;

  m_pcRIBBegin = NULL;
  m_pcRIBEnd = NULL;

  m_listCSLIDEParentInstance.Init();

  m_dyniGraphPositionX.SetValue(10);
  m_dyniGraphPositionY.SetValue(10);
}

CSLIDEPrimitive::~CSLIDEPrimitive()
{
  Uninit();
}

VOID CSLIDEPrimitive::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID)
{
  ASSERT( pcFile != NULL );

  m_pcFile = pcFile;
  m_uiLine = uiLine;
  m_uiColumn = uiColumn;

  m_pcID = pcID;

  if( pcID != NULL )
    {
      m_pcGUIID = new CHAR[strlen(pcID)+1];
      ASSERT(m_pcGUIID != NULL);
      strcpy(m_pcGUIID, pcID);
    }
}

VOID CSLIDEPrimitive::Uninit()
{
  m_pcID = NULL;
  m_pcGUIID = NULL;
}

BOOL CSLIDEPrimitive::Initialized()
{
  return m_pcFile != NULL;
}

VOID CSLIDEPrimitive::InsertParentInstance(CSLIDEInstance *pCSLIDEInstance)
{
  ASSERT( pCSLIDEInstance != NULL );

  m_listCSLIDEParentInstance.InsertLast(pCSLIDEInstance, (UINT_32)pCSLIDEInstance);
}

VOID CSLIDEPrimitive:: RemoveParentInstance(CSLIDEInstance *pCSLIDEInstance)
{
  m_listCSLIDEParentInstance.RemoveUnordered((UINT_32)pCSLIDEInstance);
}

VOID CSLIDEPrimitive::GetParameters(Tcl_Interp *pInterp)
{
  ASSERT(FALSE);
}

UINT_32 CSLIDEPrimitive::SetParameter(CSLIDEWorld *pCSLIDEWorld,
				      SLIDEReservedWordType uiParameter, CHAR **argv)
{
  switch(uiParameter) 
    {
    case sliderwtRIBBegin:
      if( m_pcRIBBegin != NULL )
	{
	  delete m_pcRIBBegin;
	}
      m_pcRIBBegin = new CHAR[strlen(argv[0])+1];
      ASSERT(m_pcRIBBegin != NULL);
      strcpy(m_pcRIBBegin, argv[0]);
      return 1;
      break;
    case sliderwtRIBEnd:
      if( m_pcRIBEnd != NULL )
	{
	  delete m_pcRIBEnd;
	}
      m_pcRIBEnd = new CHAR[strlen(argv[0])+1];
      ASSERT(m_pcRIBEnd != NULL);
      strcpy(m_pcRIBEnd, argv[0]);
      return 1;
      break;
    case sliderwtGraphPosition:
      ASSERT( (argv[0] != NULL) && (argv[1] != NULL) );
      SetGraphPosition( atoi(argv[0]), atoi(argv[1]) );
      return 2;
      break;
    case sliderwtGUIID:
      if( m_pcGUIID != NULL )
	{
	  delete m_pcGUIID;
	}
      m_pcGUIID = new CHAR[strlen(argv[0])+1];
      ASSERT(m_pcGUIID != NULL);
      strcpy(m_pcGUIID, argv[0]);
      return 1;
      break;
    default:
      ASSERT(FALSE);
      return 0;
      break;
    }
}

BOOL CSLIDEPrimitive::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  ASSERT( pCSLIDEWorld != NULL );

  return TRUE;
}

BOOL CSLIDEPrimitive::Valid()
{
  return Initialized();
}

VOID CSLIDEPrimitive::Remove(CSLIDEWorld *pCSLIDEWorld)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  iter.Init(&m_listCSLIDEParentInstance);
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->RemoveFromParent(pCSLIDEWorld);
      pCSLIDEWorld->RemoveInstance(pCSLIDEInstance);
    }

  m_listCSLIDEParentInstance.RemoveAll();
}

VOID CSLIDEPrimitive::RemoveChildInstance(CSLIDEInstance */*pCSLIDEInstance*/)
{
}

VOID CSLIDEPrimitive::RecalculateParent()
{
}

VOID CSLIDEPrimitive::Preprocess()
{
  UpdateState();
  CheckDynamic();
}

VOID CSLIDEPrimitive::SetUnvisited()
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  
  iter.Init(&m_listCSLIDEParentInstance);
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->SetUnvisited();
    }
}

VOID CSLIDEPrimitive::SetBoundUnvisited()
{
  ASSERT(FALSE);
}

VOID CSLIDEPrimitive::SetBoundUnchanged()
{
  ASSERT(FALSE);
}

VOID CSLIDEPrimitive::Update()
{
  if ( GetDynamic() )
    {
      UpdateState();
    }
}

VOID CSLIDEPrimitive::SetPrerenderUnvisited()
{
  ASSERT(FALSE);
}

VOID CSLIDEPrimitive::CheckDynamic()
{
  SetDynamic(FALSE);
}

VOID CSLIDEPrimitive::UpdateState()
{
}

VOID CSLIDEPrimitive::PushAuxillaryGeometryAttributes()
{
  gilPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT | GL_TEXTURE_BIT);

  gilDisable(GL_LIGHTING);
  gilDisable(GL_TEXTURE_2D);
  gilShadeModel(GL_FLAT);
  gilColor3f(1.0, 1.0, 1.0);
}

VOID CSLIDEPrimitive::PopAuxillaryGeometryAttributes()
{
  gilPopAttrib();
}

VOID CSLIDEPrimitive::CheckAttributesStackDepth(UINT_32 uiDepth)
{
  if ( m_uiAttributeDepthCurr + uiDepth > m_uiAttributeDepthMax )
    {
      HandleWarning("OpenGL Attribute Stack Overflow: %ld > %ld", 
		    m_uiAttributeDepthCurr + uiDepth, m_uiAttributeDepthMax);
    }  
}

VOID CSLIDEPrimitive::PushAttributes(GLbitfield GLbAttributesMask)
{
  if ( GLbAttributesMask != 0x0 )
    {
      m_uiAttributeDepthCurr++;
      //printf("PUSH Attrib Level %ld: %s\n", m_uiAttributeDepthCurr, 
      //     (m_pcID == NULL) ? ("NULL") : (m_pcID));
      if ( m_uiAttributeDepthCurr > m_uiAttributeDepthMax )
	{
	  HandleWarning("OpenGL Attribute Stack Overflow: %ld > %ld", 
			m_uiAttributeDepthCurr, m_uiAttributeDepthMax);
	}
      gilPushAttrib(GLbAttributesMask);  
    }
}

VOID CSLIDEPrimitive::PopAttributes(GLbitfield GLbAttributesMask)
{
  if ( GLbAttributesMask != 0x0 )
    {
      ASSERT( m_uiAttributeDepthCurr > 0 );

      //printf(" POP Attrib Level %ld: %s\n", m_uiAttributeDepthCurr, 
      //     (m_pcID == NULL) ? ("NULL") : (m_pcID));

      m_uiAttributeDepthCurr--;
      gilPopAttrib();  
    }
}

GLbitfield CSLIDEPrimitive::CalculateWireFrameAttributesMask()
{
  return GL_LIGHTING_BIT | GL_TEXTURE_BIT | GL_POLYGON_BIT;
}

VOID CSLIDEPrimitive::PushWireFrameAttributes()
{
  //gilPushAttrib(GL_LIGHTING_BIT | GL_TEXTURE_BIT | GL_POLYGON_BIT);

  // Disable lighting for wire frames
  gilShadeModel(GL_FLAT);

  gilPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
}

VOID CSLIDEPrimitive::PopWireFrameAttributes()
{
  //gilPopAttrib();
}

GLbitfield CSLIDEPrimitive::CalculateFlatAttributesMask()
{
  return GL_LIGHTING_BIT | GL_TEXTURE_BIT | GL_POLYGON_BIT;
}

VOID CSLIDEPrimitive::PushFlatAttributes()
{
  //gilPushAttrib(GL_LIGHTING_BIT | GL_TEXTURE_BIT | GL_POLYGON_BIT);

  // Disable lighting for wire frames
  gilShadeModel(GL_FLAT);

  gilPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
}

VOID CSLIDEPrimitive::PopFlatAttributes()
{
  //gilPopAttrib();
}

GLbitfield CSLIDEPrimitive::CalculateSmoothAttributesMask()
{
  return GL_LIGHTING_BIT | GL_TEXTURE_BIT | GL_POLYGON_BIT;
}

VOID CSLIDEPrimitive::PushSmoothAttributes()
{
  //gilPushAttrib(GL_LIGHTING_BIT | GL_TEXTURE_BIT | GL_POLYGON_BIT);

  gilShadeModel(GL_SMOOTH);

  gilPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
}

VOID CSLIDEPrimitive::PopSmoothAttributes()
{
  //gilPopAttrib();
}

GLbitfield CSLIDEPrimitive::CalculateShadingAttributesMask(SLIDEShadingType eShading)
{
  switch ( eShading )
    {
    case slideshtSLF_INHERIT:
      break;
    case slideshtSLF_WIRE:
      return CalculateWireFrameAttributesMask();
      break;
    case slideshtSLF_FLAT:
      return CalculateFlatAttributesMask();
      break;
    case slideshtSLF_GOURAUD:
    case slideshtSLF_PHONG:
      return CalculateSmoothAttributesMask();
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  return 0x0;
}

VOID CSLIDEPrimitive::PushShadingAttributes(SLIDEShadingType eShading)
{
  switch ( eShading )
    {
    case slideshtSLF_INHERIT:
      break;
    case slideshtSLF_WIRE:
      PushWireFrameAttributes();
      break;
    case slideshtSLF_FLAT:
      PushFlatAttributes();
      break;
    case slideshtSLF_GOURAUD:
    case slideshtSLF_PHONG:
      PushSmoothAttributes();
      break;
    default:
      ASSERT( FALSE );
      break;
    }
}

VOID CSLIDEPrimitive::PopShadingAttributes(SLIDEShadingType eShading)
{
  switch ( eShading )
    {
    case slideshtSLF_INHERIT:
      break;
    case slideshtSLF_WIRE:
      PopWireFrameAttributes();
      break;
    case slideshtSLF_FLAT:
      PopFlatAttributes();
      break;
    case slideshtSLF_GOURAUD:
    case slideshtSLF_PHONG:
      PopSmoothAttributes();
      break;
    default:
      ASSERT( FALSE );
      break;
    }
}

GLbitfield CSLIDEPrimitive::CalculateSolidAttributesMask(SLIDESolidType eSolid)
{
  if ( eSolid == slidestSLF_HOLLOW )
    {
      return GL_POLYGON_BIT;
    }

  return 0x0;
}

VOID CSLIDEPrimitive::PushSolidAttributes(SLIDESolidType eSolid)
{
  if ( eSolid == slidestSLF_HOLLOW )
    {
      //gilPushAttrib(GL_POLYGON_BIT);
      
      gilDisable(GL_CULL_FACE);
    }

}

VOID CSLIDEPrimitive::PopSolidAttributes(SLIDESolidType eSolid)
{
  if ( eSolid == slidestSLF_HOLLOW )
    {
      //gilPopAttrib();
    }
}

CSLIDEInstance *CSLIDEPrimitive::FindSelection(GLuint */*puiID*/, int /*iLength*/)
{
  ASSERT(FALSE);
  return NULL;
}

VOID CSLIDEPrimitive::RenderGUISelected(GLuint */*puiID*/, int /*iLength*/,
					Matrix &/*mChildFrame*/, Matrix &/*mParentFrame*/,
					BOOL /*bDrawBBox*/, BOOL /*bDrawAxes*/, 
					BOOL /*bChildAxesSelected*/, 
					BOOL /*bXAxisSelected*/, BOOL /*bYAxisSelected*/, 
					BOOL /*bZAxisSelected*/)
{
  ASSERT(FALSE);
}

VOID CSLIDEPrimitive::WriteSLIDEFlags(FILE *fpOutputFile, UINT_32 uiIndent)
{
  WriteIndent(fpOutputFile, uiIndent);
  fprintf(fpOutputFile, "  %s %ld %ld\n",
	  gcppcSLIDEReservedWords[sliderwtGraphPosition], 
	  GetGraphPositionX(),
	  GetGraphPositionY());

  if( m_pcRIBBegin != NULL )
    {
      WriteIndent(fpOutputFile, uiIndent);
      fprintf(fpOutputFile, "  %s \"%s\"\n",
	      gcppcSLIDEReservedWords[sliderwtRIBBegin], 
	      m_pcRIBBegin);
    }

  if( m_pcRIBEnd != NULL )
    {
      WriteIndent(fpOutputFile, uiIndent);
      fprintf(fpOutputFile, "  %s \"%s\"\n",
	      gcppcSLIDEReservedWords[sliderwtRIBEnd], 
	      m_pcRIBEnd);
    }
}

VOID CSLIDEPrimitive::WriteIndent(FILE *fpOutputFile, UINT_32 uiIndent)
{
  UINT_32 i;

  ASSERT( fpOutputFile != NULL );

  for ( i = 0; i < uiIndent; i++ )
    {
      fprintf(fpOutputFile, "  ");
    }
}

VOID CSLIDEPrimitive::WriteRIBBegin(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  if ( m_pcRIBBegin != NULL )
    {
      RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE RIB BEGIN:\n%s", m_pcRIBBegin);
      RiArchiveRecord(RI_COMMENT, "   END SLIDE RIB BEGIN:");      
    }
}

VOID CSLIDEPrimitive::WriteRIBEnd(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  if ( m_pcRIBEnd != NULL )
    {
      RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE RIB   END:\n%s", m_pcRIBEnd);
      RiArchiveRecord(RI_COMMENT, "   END SLIDE RIB   END:");      
    }
}

VOID CSLIDEPrimitive::WriteRIBMultMatrix(FILE *fpOutputFile, const Matrix &m)
{
  MATRIX4D mGL;
  RtMatrix rtMatrix;
  UINT_32 uiRow, uiCol, i;

  ASSERT( fpOutputFile != NULL );

  m.Get(mGL);

  // The indices are reversed because OpenGL uses 
  // Column Major Ordered Matrices
  // while RenderMan expects Row Major Ordered Matrices

  // The for loops are based on Column Major Order because that
  // is that is the linear ordering of the OpenGL matrix
  i = 0;
  for ( uiCol = 0; uiCol < 4; uiCol++ )
    {
      for ( uiRow = 0; uiRow < 4; uiRow++ )
	{
	  //rtMatrix[uiRow][uiCol] = mGL[i];
	  rtMatrix[uiCol][uiRow] = mGL[i];
	  i++;
	}
    }

  RiConcatTransform(rtMatrix);
}

VOID CSLIDEPrimitive::WriteRIBPushWireFrameAttributes(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiAttributeBegin();

  // Disable lighting for wire frames
  //gilShadeModel(GL_FLAT);

  //gilPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
}

VOID CSLIDEPrimitive::WriteRIBPopWireFrameAttributes(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiAttributeEnd();
}

VOID CSLIDEPrimitive::WriteRIBPushFlatAttributes(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiAttributeBegin();

  RiShadingInterpolation(RI_CONSTANT);

  //gilPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
}

VOID CSLIDEPrimitive::WriteRIBPopFlatAttributes(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiAttributeEnd();
}

VOID CSLIDEPrimitive::WriteRIBPushSmoothAttributes(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiAttributeBegin();

  RiShadingInterpolation(RI_SMOOTH);

  //gilPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
}

VOID CSLIDEPrimitive::WriteRIBPopSmoothAttributes(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiAttributeEnd();
}

VOID CSLIDEPrimitive::WriteRIBPushShadingAttributes(FILE *fpOutputFile,
						    SLIDEShadingType eShading)
{
  ASSERT( fpOutputFile != NULL );

  switch ( eShading )
    {
    case slideshtSLF_INHERIT:
      break;
    case slideshtSLF_WIRE:
      WriteRIBPushWireFrameAttributes(fpOutputFile);
      break;
    case slideshtSLF_FLAT:
      WriteRIBPushFlatAttributes(fpOutputFile);
      break;
    case slideshtSLF_GOURAUD:
    case slideshtSLF_PHONG:
      WriteRIBPushSmoothAttributes(fpOutputFile);
      break;
    default:
      ASSERT( FALSE );
      break;
    }
}

VOID CSLIDEPrimitive::WriteRIBPopShadingAttributes(FILE *fpOutputFile,
						   SLIDEShadingType eShading)
{
  ASSERT( fpOutputFile != NULL );

  switch ( eShading )
    {
    case slideshtSLF_INHERIT:
      break;
    case slideshtSLF_WIRE:
      WriteRIBPopWireFrameAttributes(fpOutputFile);
      break;
    case slideshtSLF_FLAT:
      WriteRIBPopFlatAttributes(fpOutputFile);
      break;
    case slideshtSLF_GOURAUD:
    case slideshtSLF_PHONG:
      WriteRIBPopSmoothAttributes(fpOutputFile);
      break;
    default:
      ASSERT( FALSE );
      break;
    }
}

VOID CSLIDEPrimitive::WriteRIBPushSolidAttributes(FILE *fpOutputFile,
						  SLIDESolidType eSolid)
{
  ASSERT( fpOutputFile != NULL );

  if ( eSolid == slidestSLF_HOLLOW )
    {
      RiAttributeBegin();
      
      // Don't do Back Face Culling
      RiSides(2);
    }
}

VOID CSLIDEPrimitive::WriteRIBPopSolidAttributes(FILE *fpOutputFile,
						 SLIDESolidType eSolid)
{
  ASSERT( fpOutputFile != NULL );

  if ( eSolid == slidestSLF_HOLLOW )
    {
      RiAttributeEnd();
    }
}

VOID CSLIDEPrimitive::HandleError(CHAR *pcFormat, ...)
{
  va_list vaArgs;

  ASSERT( pcFormat != NULL );

  m_uiErrorCount++;
  fprintf(m_fpErrorFile, "%s: %ld: %ld: Semantic Error: ", 
	  (m_pcFile == NULL) ? ("") : (m_pcFile), m_uiLine, m_uiColumn);

  // Now print the message sent by the program
  va_start(vaArgs, pcFormat);
  vfprintf(m_fpErrorFile, pcFormat, vaArgs);
  va_end(vaArgs);

  fprintf(m_fpErrorFile, "\n"); 
}

VOID CSLIDEPrimitive::HandleWarning(CHAR *pcFormat, ...)
{
  va_list vaArgs;

  ASSERT( pcFormat != NULL );

  m_uiWarningCount++;
  fprintf(m_fpErrorFile, "%s: %ld: %ld: Semantic Warning: ", 
	  (m_pcFile == NULL) ? ("") : (m_pcFile), m_uiLine, m_uiColumn);

  // Now print the message sent by the program
  va_start(vaArgs, pcFormat);
  vfprintf(m_fpErrorFile, pcFormat, vaArgs);
  va_end(vaArgs);

  fprintf(m_fpErrorFile, "\n"); 
}

VOID CSLIDEPrimitive::WriteRIB(FILE * /*fpOutputFile*/)
{
}

VOID CSLIDEPrimitive::WritePOV(FILE * /*fpOutputFile*/, 
			       CListInt * /*plistTransformIDs*/,
			       CListInt * /*plistColors*/) 
{
}

VOID CSLIDEPrimitive::WriteWRLZ(FILE * /*fpOutputFile*/, const Matrix &/*mTransform*/,
				CSLIDESurface */*pCSLIDESurface*/)
{
}

VOID CSLIDEPrimitive::WriteSTL(FILE * /*fpOutputFile*/, const Matrix &/*mTransform*/)
{
}

VOID CSLIDEPrimitive::WriteDotID(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "ERROR");
}

VOID CSLIDEPrimitive::WriteDotNode(FILE * /*fpOutputFile*/)
{
}

VOID CSLIDEPrimitive::WriteDotEdges(FILE * /*fpOutputFile*/)
{
}
