#ifndef __SLIDEPRIMITIVE_H__
#define __SLIDEPRIMITIVE_H__

#include <tcl.h>
//#define DT_TRIANGULATE_TEST

#include "genlib.h"
#include "datalib.h"
#include "mathlib.h"
#include "bound.h"

#include "slide.h"
#include "gillib.h"
//#include <ri.h>
#include <ri.h>

#include "dynamics.h"

#ifdef __USE_CDA__
#include "cda.h"
#include "cdavertex.h"
#endif // __USE_CDA__

//////////////////////////////////////////////////////////////////////
// CSLIDEPrimitive Class
//

class CSLIDEWorld;
class CSLIDECamera;
class CSLIDEInstance;
class CSLIDECylinder;
class CSLIDEWormwarp;

class CSLIDEPrimitive
{
//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEPrimitive();
  virtual ~CSLIDEPrimitive();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  virtual VOID InsertParentInstance(CSLIDEInstance *pCSLIDEInstance);

  inline VOID SetRIBBegin(CHAR *pcRIBBegin);
  inline VOID SetRIBEnd(CHAR *pcRIBEnd);
  inline VOID SetGraphPosition(DynInt *pdyniX, DynInt *pdyniY);
  inline VOID SetGraphPosition(int x, int y);

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

  virtual VOID Remove(CSLIDEWorld *pCSLIDEWorld);
  virtual VOID RemoveParentInstance(CSLIDEInstance *pCSLIDEInstance);
  virtual VOID RemoveChildInstance(CSLIDEInstance *pCSLIDEInstance);
  virtual VOID RecalculateParent();

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual VOID Preprocess();
  virtual VOID SetUnvisited();
  virtual VOID SetBoundUnvisited();
  virtual VOID SetBoundUnchanged();

  virtual VOID Update();
  virtual VOID SetPrerenderUnvisited();

protected:
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();

//////////////////////////////////////////////////
// Rendering
//
public:
  VOID PushAuxillaryGeometryAttributes();
  VOID PopAuxillaryGeometryAttributes();

  VOID CheckAttributesStackDepth(UINT_32 uiDepth);
  VOID PushAttributes(GLbitfield GLbAttributesMask);
  VOID PopAttributes(GLbitfield GLbAttributesMask);
  GLbitfield CalculateWireFrameAttributesMask();
  VOID PushWireFrameAttributes();
  VOID PopWireFrameAttributes();
  GLbitfield CalculateFlatAttributesMask();
  VOID PushFlatAttributes();
  VOID PopFlatAttributes();
  GLbitfield CalculateSmoothAttributesMask();
  VOID PushSmoothAttributes();
  VOID PopSmoothAttributes();
  GLbitfield CalculateShadingAttributesMask(SLIDEShadingType eShading);
  VOID PushShadingAttributes(SLIDEShadingType eShading);
  VOID PopShadingAttributes(SLIDEShadingType eShading);

  GLbitfield CalculateSolidAttributesMask(SLIDESolidType eSolid);
  VOID PushSolidAttributes(SLIDESolidType eSolid);
  VOID PopSolidAttributes(SLIDESolidType eSolid);

  virtual CSLIDEInstance *FindSelection(GLuint *puiID, int iLength);  
  virtual VOID RenderGUISelected(GLuint *puiID, int iLength, 
				 Matrix &mChildFrame, Matrix &mParentFrame, 
				 BOOL bDrawBBox, BOOL bDrawAxes,
				 BOOL bChildAxesSelected, BOOL bXAxisSelected,
				 BOOL bYAxisSelected, BOOL bZAxisSelected);  

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile) = 0;
  virtual VOID WriteSLIDEFlags(FILE *fpOutputFile, UINT_32 uiIndent);
  VOID WriteIndent(FILE *fpOutputFile, UINT_32 uiIndent);

  virtual VOID WriteRIB(FILE *fpOutputFile);
  virtual VOID WritePOV(FILE *fpOutputFile, 
			CListInt *plistTransformIDs,
			CListInt *plistColors); 
  virtual VOID WriteWRLZ(FILE *fpOutputFile, const Matrix &mTransform,
			 CSLIDESurface *pCSLIDESurface);

  virtual VOID WriteSTL(FILE *fpOutputFile, const Matrix &mTransform);

  VOID WriteRIBBegin(FILE *fpOutputFile);
  VOID WriteRIBEnd(FILE *fpOutputFile);

  VOID WriteRIBMultMatrix(FILE *fpOutputFile, const Matrix &m);

  VOID WriteRIBPushWireFrameAttributes(FILE *fpOutputFile);
  VOID WriteRIBPopWireFrameAttributes(FILE *fpOutputFile);
  VOID WriteRIBPushFlatAttributes(FILE *fpOutputFile);
  VOID WriteRIBPopFlatAttributes(FILE *fpOutputFile);
  VOID WriteRIBPushSmoothAttributes(FILE *fpOutputFile);
  VOID WriteRIBPopSmoothAttributes(FILE *fpOutputFile);
  VOID WriteRIBPushShadingAttributes(FILE *fpOutputFile, SLIDEShadingType eShading);
  VOID WriteRIBPopShadingAttributes(FILE *fpOutputFile, SLIDEShadingType eShading);

  VOID WriteRIBPushSolidAttributes(FILE *fpOutputFile, SLIDESolidType eSolid);
  VOID WriteRIBPopSolidAttributes(FILE *fpOutputFile, SLIDESolidType eSolid);

  virtual VOID WriteDotID(FILE *fpOutputFile);
  virtual VOID WriteDotNode(FILE *fpOutputFile);
  virtual VOID WriteDotEdges(FILE *fpOutputFile);

//////////////////////////////////////////////////
// Error Handling
//
public:
  VOID HandleError(CHAR *pcFormat, ...);
  VOID HandleWarning(CHAR *pcFormat, ...);

//////////////////////////////////////////////////
// Class Accessor Functions
//
public:
  static inline FILE *GetErrorFile();
  static inline VOID SetErrorFile(FILE *fpErrorFile);
  static inline UINT_32 GetErrorCount();
  static inline VOID SetErrorCount(UINT_32 uiErrors);
  static inline UINT_32 GetWarningCount();
  static inline VOID SetWarningCount(UINT_32 uiWarnings);

  static inline BOOL GetRenderClipBoundEnabled();
  static inline VOID SetRenderClipBoundEnabled(BOOL bRenderClipBoundEnabled);
  static inline BOOL GetFogEnabled();
  static inline VOID SetFogEnabled(BOOL bFogEnabled);
  static inline BOOL GetLightingEnabled();
  static inline VOID SetLightingEnabled(BOOL bLightingEnabled);
  static inline BOOL GetTextureMappingEnabled();
  static inline VOID SetTextureMappingEnabled(BOOL bTextureMappingEnabled);
  static inline BOOL GetCrystalBallEnabled();
  static inline VOID SetCrystalBallEnabled(BOOL bCrystalBallEnabled);
  static inline int GetWindowCUIType();
  static inline VOID SetWindowCUIType(DynInt *pDyniType);
  static inline FLOAT GetWindowCUISpeed();
  static inline VOID SetWindowCUISpeed(DynFloat *pDynfSpeed);
  static inline BOOL GetSelectionEnabled();
  static inline VOID SetSelectionEnabled(BOOL bSelectionEnabled);

  static inline BOOL GetRenderBoundsEnabled();
  static inline VOID SetRenderBoundsEnabled(BOOL bRenderBoundsEnabled);
  static inline BOOL GetColorDynamicBounds();
  static inline VOID SetColorDynamicBounds(BOOL bColorDynamicBounds);
  static inline BOOL GetColorChangedBounds();
  static inline VOID SetColorChangedBounds(BOOL bColorChangedBounds);
  static inline BOOL GetCheckBoundChanged();
  static inline VOID SetCheckBoundChanged(BOOL bCheckBoundChanged);
  static inline BOOL GetBoundRejectEnabled();
  static inline VOID SetBoundRejectEnabled(BOOL bBoundRejectEnabled);

  static inline BOOL GetBackFaceCullingEnabled();
  static inline VOID SetBackFaceCullingEnabled(BOOL bBackFaceCullingEnabled);
  static inline BOOL GetClipEnabled();
  static inline VOID SetClipEnabled(BOOL bClipEnabled);
  static inline BOOL GetRenderNormalsEnabled();
  static inline VOID SetRenderNormalsEnabled(BOOL bRenderNormalsEnabled);
  static inline FLOAT GetNormalsRenderLength();
  static inline VOID SetNormalsRenderLength(FLOAT fNormalsRenderLength);

  static inline BOOL GetRenderCDAOperandsEnabled();
  static inline VOID SetRenderCDAOperandsEnabled(BOOL bRenderNormalsEnabled);

//////////////////////////////////////////////////
// Member Accessor Functions
//
public:
  inline CHAR *GetFileName();
  inline UINT_32 GetFileLine();
  inline UINT_32 GetFileColumn();
  virtual inline CHAR *GetTypeString();
  virtual inline SLIDEReservedWordType GetType();
  inline CHAR *GetID();
  inline CHAR *GetGUIID();
  inline INT_32 GetGraphPositionX();
  inline INT_32 GetGraphPositionY();
  virtual VOID GetParameters(Tcl_Interp *pInterp);
  virtual UINT_32 SetParameter(CSLIDEWorld *pCSLIDEWorld, 
			       SLIDEReservedWordType uiParameter, CHAR **argv);

  inline BOOL GetDynamic();
  inline VOID SetDynamic(BOOL bDynamic);

//////////////////////////////////////////////////
// Class Variables
//
protected:
  static FILE *m_fpErrorFile;
  static UINT_32 m_uiErrorCount;
  static UINT_32 m_uiWarningCount;

  static BOOL m_bRenderClipBoundEnabled;
  static BOOL m_bFogEnabled;
  static BOOL m_bLightingEnabled;
  static BOOL m_bTextureMappingEnabled;
  static BOOL m_bCrystalBallEnabled;
  static int  m_iWindowCUIType;
  static FLOAT m_fWindowCUISpeed;
  static BOOL m_bSelectionEnabled;

  static BOOL m_bRenderBoundsEnabled;
  static BOOL m_bColorDynamicBounds;
  static BOOL m_bColorChangedBounds;
  static BOOL m_bCheckBoundChanged;
  static BOOL m_bBoundRejectEnabled;

  static BOOL m_bBackFaceCullingEnabled;
  static BOOL m_bClipEnabled;
  static BOOL m_bRenderNormalsEnabled;
  static FLOAT m_fNormalsRenderLength;

  static BOOL m_bRenderCDAOperandsEnabled;

  static CSLIDECamera *m_pCSLIDECameraCurr;

  static UINT_32 m_uiAttributeDepthMax;
  static UINT_32 m_uiAttributeDepthCurr;

  static CSLIDECylinder *m_pCSLIDECylinderAxis;

//////////////////////////////////////////////////
// Member Variables
//
protected:
  // SLIDE .slf file position of the primitive for debugging the files
  CHAR *m_pcFile;
  UINT_32 m_uiLine;
  UINT_32 m_uiColumn;

  CHAR *m_pcGUIID; // This is the name that a gui user sees, but SLIDE still uses
                   // m_pcID in the string table. m_pcGUIID is for writing SLIDE
  CHAR *m_pcID; // Stored in a string table so equals checks can be done on 
                // the pointer itself
  BOOL m_bDynamic;

  CHAR *m_pcRIBBegin;
  CHAR *m_pcRIBEnd;

  CListInt m_listCSLIDEParentInstance;

  // For drawing to the hierarchy-editing graph
  DynInt m_dyniGraphPositionX;
  DynInt m_dyniGraphPositionY;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDEPrimitive::SetRIBBegin(CHAR *pcRIBBegin)
{
  m_pcRIBBegin = pcRIBBegin;
}

inline VOID CSLIDEPrimitive::SetRIBEnd(CHAR *pcRIBEnd)
{
  m_pcRIBEnd = pcRIBEnd;
}

inline VOID CSLIDEPrimitive::SetGraphPosition(DynInt *pdyniX, DynInt *pdyniY)
{
  m_dyniGraphPositionX.Set(pdyniX);
  m_dyniGraphPositionY.Set(pdyniY);
}

inline VOID CSLIDEPrimitive::SetGraphPosition(int x, int y)
{
  m_dyniGraphPositionX.SetValue(x);
  m_dyniGraphPositionY.SetValue(y);
}

inline INT_32 CSLIDEPrimitive::GetGraphPositionX()
{
  return m_dyniGraphPositionX.GetValue();
}

inline INT_32 CSLIDEPrimitive::GetGraphPositionY()
{
  return m_dyniGraphPositionY.GetValue();
}

inline CHAR *CSLIDEPrimitive::GetTypeString()
{
  CHAR *pcFoo;

  ASSERT(FALSE);

  return pcFoo;
}

inline SLIDEReservedWordType CSLIDEPrimitive::GetType()
{
  ASSERT(FALSE);

  return sliderwtSLF_INHERIT;
}

inline FILE *CSLIDEPrimitive::GetErrorFile()
{
  return m_fpErrorFile;
}

inline VOID CSLIDEPrimitive::SetErrorFile(FILE *fpErrorFile)
{
  ASSERT( fpErrorFile != NULL );

  m_fpErrorFile = fpErrorFile;
}

inline UINT_32 CSLIDEPrimitive::GetErrorCount()
{
  return m_uiErrorCount;
}
 
inline VOID CSLIDEPrimitive::SetErrorCount(UINT_32 uiErrors)
{
  m_uiErrorCount = uiErrors;
}

inline UINT_32 CSLIDEPrimitive::GetWarningCount()
{
  return m_uiWarningCount;
}

inline VOID CSLIDEPrimitive::SetWarningCount(UINT_32 uiWarnings)
{
  m_uiWarningCount = uiWarnings;
}

inline BOOL CSLIDEPrimitive::GetRenderClipBoundEnabled()
{
  return m_bRenderClipBoundEnabled;
}

inline VOID CSLIDEPrimitive::SetRenderClipBoundEnabled(BOOL bRenderClipBoundEnabled)
{
  m_bRenderClipBoundEnabled = bRenderClipBoundEnabled;
}

inline BOOL CSLIDEPrimitive::GetFogEnabled()
{
  return m_bFogEnabled;
}

inline VOID CSLIDEPrimitive::SetFogEnabled(BOOL bFogEnabled)
{
  m_bFogEnabled = bFogEnabled;
}

inline BOOL CSLIDEPrimitive::GetLightingEnabled()
{
  return m_bLightingEnabled;
}

inline VOID CSLIDEPrimitive::SetLightingEnabled(BOOL bLightingEnabled)
{
  m_bLightingEnabled = bLightingEnabled;
}

inline BOOL CSLIDEPrimitive::GetTextureMappingEnabled()
{
  return m_bTextureMappingEnabled;
}

inline VOID CSLIDEPrimitive::SetTextureMappingEnabled(BOOL bTextureMappingEnabled)
{
  m_bTextureMappingEnabled = bTextureMappingEnabled;
}

inline BOOL CSLIDEPrimitive::GetCrystalBallEnabled()
{
  return m_bCrystalBallEnabled;
}

inline VOID CSLIDEPrimitive::SetCrystalBallEnabled(BOOL bCrystalBallEnabled)
{
  m_bCrystalBallEnabled = bCrystalBallEnabled;
}

inline int CSLIDEPrimitive::GetWindowCUIType()
{
  return m_iWindowCUIType;
}

inline VOID CSLIDEPrimitive::SetWindowCUIType(DynInt *pDyniType)
{
  m_iWindowCUIType = pDyniType->GetValue();
}

inline FLOAT CSLIDEPrimitive::GetWindowCUISpeed()
{
  return m_fWindowCUISpeed;
}

inline VOID CSLIDEPrimitive::SetWindowCUISpeed(DynFloat *pDynfSpeed)
{
  m_fWindowCUISpeed = pDynfSpeed->GetValue();
}

inline BOOL CSLIDEPrimitive::GetSelectionEnabled()
{
  return m_bSelectionEnabled;
}

inline VOID CSLIDEPrimitive::SetSelectionEnabled(BOOL bSelectionEnabled)
{
  m_bSelectionEnabled = bSelectionEnabled;
}

inline BOOL CSLIDEPrimitive::GetRenderBoundsEnabled()
{
  return m_bRenderBoundsEnabled;
}

inline VOID CSLIDEPrimitive::SetRenderBoundsEnabled(BOOL bRenderBoundsEnabled)
{
  m_bRenderBoundsEnabled = bRenderBoundsEnabled;
}

inline BOOL CSLIDEPrimitive::GetColorDynamicBounds()
{
  return m_bColorDynamicBounds;
}

inline VOID CSLIDEPrimitive::SetColorDynamicBounds(BOOL bColorDynamicBounds)
{
  m_bColorDynamicBounds = bColorDynamicBounds;
}

inline BOOL CSLIDEPrimitive::GetColorChangedBounds()
{
  return m_bColorChangedBounds;
}

inline VOID CSLIDEPrimitive::SetColorChangedBounds(BOOL bColorChangedBounds)
{
  m_bColorChangedBounds = bColorChangedBounds;
}

inline BOOL CSLIDEPrimitive::GetCheckBoundChanged()
{
  return m_bCheckBoundChanged;
}

inline VOID CSLIDEPrimitive::SetCheckBoundChanged(BOOL bCheckBoundChanged)
{
  m_bCheckBoundChanged = bCheckBoundChanged;
}

inline BOOL CSLIDEPrimitive::GetBoundRejectEnabled()
{
  return m_bBoundRejectEnabled;
}

inline VOID CSLIDEPrimitive::SetBoundRejectEnabled(BOOL bBoundRejectEnabled)
{
  m_bBoundRejectEnabled = bBoundRejectEnabled;
}

inline BOOL CSLIDEPrimitive::GetBackFaceCullingEnabled()
{
  return m_bBackFaceCullingEnabled;
}

inline VOID CSLIDEPrimitive::SetBackFaceCullingEnabled(BOOL bBackFaceCullingEnabled)
{
  m_bBackFaceCullingEnabled = bBackFaceCullingEnabled;
}

inline BOOL CSLIDEPrimitive::GetClipEnabled()
{
  return m_bClipEnabled;
}

inline VOID CSLIDEPrimitive::SetClipEnabled(BOOL bClipEnabled)
{
  m_bClipEnabled = bClipEnabled;
}

inline BOOL CSLIDEPrimitive::GetRenderNormalsEnabled()
{
  return m_bRenderNormalsEnabled;
}

inline VOID CSLIDEPrimitive::SetRenderNormalsEnabled(BOOL bRenderNormalsEnabled)
{
  m_bRenderNormalsEnabled = bRenderNormalsEnabled;
}

inline FLOAT CSLIDEPrimitive::GetNormalsRenderLength()
{
  return m_fNormalsRenderLength;
}

inline VOID CSLIDEPrimitive::SetNormalsRenderLength(FLOAT fNormalsRenderLength)
{
  m_fNormalsRenderLength = fNormalsRenderLength;
}

inline BOOL CSLIDEPrimitive::GetRenderCDAOperandsEnabled()
{
  return m_bRenderCDAOperandsEnabled;
}

inline VOID CSLIDEPrimitive::SetRenderCDAOperandsEnabled(BOOL bRenderCDAOperandsEnabled)
{
  m_bRenderCDAOperandsEnabled = bRenderCDAOperandsEnabled;
}

inline CHAR *CSLIDEPrimitive::GetFileName()
{
  return m_pcFile;
}

inline UINT_32 CSLIDEPrimitive::GetFileLine()
{
  return m_uiLine;
}

inline UINT_32 CSLIDEPrimitive::GetFileColumn()
{
  return m_uiColumn;
}

inline CHAR *CSLIDEPrimitive::GetID()
{
  return m_pcID;
}

inline CHAR *CSLIDEPrimitive::GetGUIID()
{
  return m_pcGUIID;
}

inline BOOL CSLIDEPrimitive::GetDynamic()
{
  return m_bDynamic;
}

inline VOID CSLIDEPrimitive::SetDynamic(BOOL bDynamic)
{
  m_bDynamic = bDynamic;
}

#endif __SLIDEPRIMITIVE_H__
