#include "sliderender.h"

#include "slideworld.h"
#include "slideviewport.h"
#include "slidecamera.h"
#include "slidelight.h"
#include "slidearealight.h"
#include "slidenode.h"
#include "slideboundednode.h"
#include "slideface.h"
#include "slidesurface.h"
#include "slidefog.h"
#include "slideinstance.h"
#include "slidetranslate.h"
#include "slidescale.h"
#include "sliderotate.h"

#include "qemesh.h"
#include "qewrite.h"
#include "sif.h"

#define SELECTION_AXES 1234567890

static CHAR gpcBufferVariable[200];
static CHAR gpcBufferValue[200];
static CHAR gpcRIBOutputFile[500];

UINT_32 CSLIDERender::m_uiCount = 0;
Vector CSLIDERender::m_vProjMin;
Vector CSLIDERender::m_vProjMax;
CBound CSLIDERender::m_boundClip;

GLuint CSLIDERender::m_GLuiStencil = 0;
BOOL CSLIDERender::m_bUseStencil = FALSE;

CListInt *CSLIDERender::m_plistCSLIDETclVariables = NULL;

CHAR CSLIDERender::m_pcRIBOutputFileBase[500];
UINT_32 CSLIDERender::m_uiRIBFrame = 0;
DynInt CSLIDERender::m_dyniBlendOnOff=DynInt();

//////////////////////////////////////////////////////////////////////
// CSLIDERender Class
//

CSLIDERender::CSLIDERender()
{
  if ( m_uiCount == 0 )
    {
      // These are the limits of the canonical view volume
      m_vProjMin.Set(-1, -1, -1);
      m_vProjMax.Set(1, 1, 0);

      // This a bound around the canonical view volume which
      // we will use to do clipping with homogeneous coordinates
      Point pt;
      
      pt = m_vProjMin;
      m_boundClip.Insert(pt);
      pt = m_vProjMax;
      m_boundClip.Insert(pt);

      m_GLuiStencil = CSLIDEBoundedNode::GenLists();
      // GL is not yet initialized
      //ASSERT( gilIsList(m_GLuiStencil) );
      m_bUseStencil = FALSE;

      m_plistCSLIDETclVariables = new CListInt;
      ASSERT( m_plistCSLIDETclVariables != NULL );
      m_plistCSLIDETclVariables->Init();
    }
  m_uiCount++;

  m_pcTclTag = NULL;
  m_pCSLIDETclVariables = NULL;

  m_pcViewportID = NULL;
  m_pCSLIDENodeGUISelected = NULL;
  m_iSelected = -1;
  m_iSelectionHits = 0;
  m_bManipulationMode = FALSE;
  m_bChildAxesSelected = FALSE;
  m_bXAxisSelected = FALSE;
  m_bYAxisSelected = FALSE;
  m_bZAxisSelected = FALSE;
  m_bNewTransform = FALSE;
  m_bGraphSelected = FALSE;

  m_pCSLIDECameraPath = NULL;

  m_pcWorldRootID = NULL;
  m_pWorldRoot = NULL;

  m_dyniLOD.SetValue(slidelodtSLF_FULL);
  m_dyniStencil.SetValue(slidesttSLF_BOTH);

  m_pcFogID = NULL;
  m_pCSLIDEFog = NULL;
  m_dyniFogOnOff.SetValue(slideoftSLF_OFF);

  m_pmeBoundaryEdges = NULL;

  m_uiDotID = 0;
}

CSLIDERender::~CSLIDERender()
{
  m_uiCount--;
  if ( m_uiCount == 0 )
    {
      if ( gilIsList(m_GLuiStencil) )
	{
	  gilDeleteLists(m_GLuiStencil, 1);
	}
      m_GLuiStencil = 0;
      m_bUseStencil = FALSE;

      if ( m_plistCSLIDETclVariables != NULL )
	{
	  CSLIDETclVariables *pCSLIDETclVariables;

	  for ( pCSLIDETclVariables = (CSLIDETclVariables*)m_plistCSLIDETclVariables->RemoveFirst();
		pCSLIDETclVariables != NULL;
		pCSLIDETclVariables = (CSLIDETclVariables*)m_plistCSLIDETclVariables->RemoveFirst() )
	    {
	      delete pCSLIDETclVariables;
	    }
	  
	  delete m_plistCSLIDETclVariables;
	  m_plistCSLIDETclVariables = NULL;
	}
    }
}

VOID CSLIDERender::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			CHAR *pcViewportID, CSLIDECameraPath *pCSLIDECameraPath,
			CHAR *pcWorldRootID)
{
  CSLIDEPrimitive::Init(pcFile, uiLine, uiColumn, NULL);

  ASSERT( pcViewportID != NULL );
  ASSERT( pCSLIDECameraPath != NULL );
  ASSERT( pcWorldRootID != NULL );

  m_pcTclTag = NULL;
  m_pCSLIDETclVariables = NULL;

  m_pcViewportID = pcViewportID;

  m_pCSLIDECameraPath = pCSLIDECameraPath;

  m_pcWorldRootID = pcWorldRootID;
  m_pWorldRoot = NULL;

  m_dyniLOD.SetValue(slidelodtSLF_FULL);
  m_dyniStencil.SetValue(slidesttSLF_BOTH);

  m_pcFogID = NULL;
  m_pCSLIDEFog = NULL;
  m_dyniFogOnOff.SetValue(slideoftSLF_OFF);

  m_listCSLIDELightPath.Init();
  m_listCSLIDEAreaLightPath.Init();

  m_pmeBoundaryEdges = NULL;

  // Crystal Ball Variables
  m_dyniCUIType.SetValue( slidecuitSLF_CAD_CUI );

  m_iButtonCurr = GLSI_INVALID_BUTTON;
  m_fScaleLast = 1.0;

  m_iButtonsCurr = 0x00;
  m_iKeysCurr = 0x00;

  m_dyniCamZoom.SetValue(0);

  m_tvLastRender.RecordTime();
  m_dynfCUISpeed.SetValue( 1.0f );

  m_uiDotID = 0;
}

VOID CSLIDERender::Uninit()
{
  CListIterInt iter;
  CSLIDELightPath *pCSLIDELightPath;
  CSLIDEAreaLightPath *pCSLIDEAreaLightPath;

  CSLIDEPrimitive::Uninit();

  m_pcViewportID = NULL;
  m_pCSLIDENodeGUISelected = NULL;

  if ( m_pCSLIDECameraPath != NULL )
    {
      delete m_pCSLIDECameraPath;
      m_pCSLIDECameraPath = NULL;
    }

  m_pcWorldRootID = NULL;
  m_pWorldRoot = NULL;

  iter.Init(&m_listCSLIDELightPath);

  for ( pCSLIDELightPath = (CSLIDELightPath*)iter.PeekFirst();
	pCSLIDELightPath != NULL;
	pCSLIDELightPath = (CSLIDELightPath*)iter.PeekNext() )
    {
      delete pCSLIDELightPath;
    }

  iter.Init(&m_listCSLIDEAreaLightPath);

  for ( pCSLIDEAreaLightPath = (CSLIDEAreaLightPath*)iter.PeekFirst();
	pCSLIDEAreaLightPath != NULL;
	pCSLIDEAreaLightPath = (CSLIDEAreaLightPath*)iter.PeekNext() )
    {
      delete pCSLIDEAreaLightPath;
    }

  if ( m_pmeBoundaryEdges != NULL )
    {
      DeallocateQEMesh(m_pmeBoundaryEdges);
      m_pmeBoundaryEdges = NULL;
    }
}

BOOL CSLIDERender::Initialized()
{
  if ( !CSLIDEPrimitive::Initialized() )
    {
      return FALSE;
    }

  return ( m_pcViewportID != NULL &&
	   m_pCSLIDECameraPath != NULL &&
           m_pcWorldRootID != NULL );
}

BOOL CSLIDERender::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect, b;
  CSLIDEViewport *pCSLIDEViewport;

  if ( !CSLIDEPrimitive::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  // Check Tcl Tag to see if we are sharing UI Variables

  m_pCSLIDETclVariables = (CSLIDETclVariables*)m_plistCSLIDETclVariables->Peek((INT_32)m_pcTclTag);
  if ( m_pCSLIDETclVariables == NULL )
    {
      m_pCSLIDETclVariables = new CSLIDETclVariables;
      ASSERT( m_pCSLIDETclVariables != NULL);
      m_pCSLIDETclVariables->Init(m_pcTclTag);

      m_plistCSLIDETclVariables->Insert(m_pCSLIDETclVariables, 
					(INT_32)m_pcTclTag);
    }

  // Check Viewport Reference

  pCSLIDEViewport = pCSLIDEWorld->PeekViewport(m_pcViewportID);
  if ( pCSLIDEViewport == NULL )
    {
      HandleError("Undefined %s \"%s\"", 
		  gcppcSLIDEReservedWords[sliderwtViewport],
		  m_pcViewportID);
      bCorrect = FALSE;
    }
  else
    {
      pCSLIDEViewport->InsertRender(this);
    }

  // Check Camera Reference

  b = m_pCSLIDECameraPath->CheckSemantics(pCSLIDEWorld);
  bCorrect = bCorrect && b;

  // Check Root Node Reference

  m_pWorldRoot = pCSLIDEWorld->PeekNode(m_pcWorldRootID);
  if ( m_pWorldRoot == NULL )
    {
      HandleError("Undefined %s \"%s\"", 
		  gcppcSLIDEReservedWords[sliderwtObject],
		  m_pcWorldRootID);
      bCorrect = FALSE;
    }

  // Check Surface Reference

  if ( m_pcFogID != NULL )
    {
      m_pCSLIDEFog = pCSLIDEWorld->PeekFog(m_pcFogID);
      if ( m_pCSLIDEFog == NULL )
	{
	  HandleError("Undefined %s \"%s\"", 
		      gcppcSLIDEReservedWords[sliderwtFog],
		      m_pcFogID);
	  bCorrect = FALSE;
	}
    }

  // Check Light References

  CListIterInt iter;
  CSLIDELightPath *pCSLIDELightPath;

  iter.Init(&m_listCSLIDELightPath);

  for ( pCSLIDELightPath = (CSLIDELightPath*)iter.PeekFirst();
	pCSLIDELightPath != NULL;
	pCSLIDELightPath = (CSLIDELightPath*)iter.PeekNext() )
    {
      b = pCSLIDELightPath->CheckSemantics(pCSLIDEWorld);
      bCorrect = bCorrect && b;
    }

  // Check Area Light References

  CSLIDEAreaLightPath *pCSLIDEAreaLightPath;

  iter.Init(&m_listCSLIDEAreaLightPath);

  for ( pCSLIDEAreaLightPath = (CSLIDEAreaLightPath*)iter.PeekFirst();
	pCSLIDEAreaLightPath != NULL;
	pCSLIDEAreaLightPath = (CSLIDEAreaLightPath*)iter.PeekNext() )
    {
      b = pCSLIDEAreaLightPath->CheckSemantics(pCSLIDEWorld);
      bCorrect = bCorrect && b;
    }

  return bCorrect;
}

BOOL CSLIDERender::Valid()
{
  ASSERT( CSRSpecialKeysSize <= 32 );

  if ( !CSLIDEPrimitive::Valid() )
    {
      return FALSE;
    }

  return (m_pWorldRoot != NULL &&
	  ( 0 <= m_dyniLOD.GetValue() &&
	    m_dyniLOD.GetValue() < slidelodtSize ) &&
	  ( 0 <= m_dyniStencil.GetValue() &&
	    m_dyniStencil.GetValue() < slidesttSize ) );
}

VOID CSLIDERender::PreprocessGraph()
{
  m_pWorldRoot->PreprocessGraph();
}

VOID CSLIDERender::UpdateGraph()
{
  if ( m_dyniLOD.GetValue() == slideoftSLF_OFF )
    {
      return;
    }
  
  m_pWorldRoot->UpdateGraph();
}

VOID CSLIDERender::CheckDynamic()
{
  BOOL bDynamic;
  CListIterInt iter;
  CSLIDELightPath *pCSLIDELightPath;
  CSLIDEAreaLightPath *pCSLIDEAreaLightPath;

  bDynamic = ( m_dynfMergeEpsilon.GetDynamic() ||
	       m_dyniLOD.GetDynamic() ||
	       m_dyniStencil.GetDynamic() ||
	       m_dyniCUIType.GetDynamic() ||
	       m_dynfCUISpeed.GetDynamic() );

  m_pCSLIDECameraPath->Preprocess();
  bDynamic = bDynamic || m_pCSLIDECameraPath->GetDynamic();

  iter.Init(&m_listCSLIDELightPath);
  for ( pCSLIDELightPath = (CSLIDELightPath*)iter.PeekFirst();
	pCSLIDELightPath != NULL;
	pCSLIDELightPath = (CSLIDELightPath*)iter.PeekNext() )
    {
      pCSLIDELightPath->Preprocess();
      bDynamic = bDynamic || pCSLIDELightPath->GetDynamic();
    }

  iter.Init(&m_listCSLIDEAreaLightPath);
  for ( pCSLIDEAreaLightPath = (CSLIDEAreaLightPath*)iter.PeekFirst();
	pCSLIDEAreaLightPath != NULL;
	pCSLIDEAreaLightPath = (CSLIDEAreaLightPath*)iter.PeekNext() )
    {
      pCSLIDEAreaLightPath->Preprocess();
      bDynamic = bDynamic || pCSLIDEAreaLightPath->GetDynamic();
    }

  SetDynamic(bDynamic);
}

VOID CSLIDERender::UpdateState()
{
  // The type specified in the SLIDE file overrides the window UI
  if( m_dyniCUIType.GetDynamic() )
    {
      m_dyniCUIType.Update();
    }

  // The speed specified in the SLIDE file overrides the window UI
  if( m_dynfCUISpeed.GetDynamic() )
    {
      m_dynfCUISpeed.Update();
    }

  m_dynfMergeEpsilon.Update();
  if( m_dynfMergeEpsilon.GetDynamic() )
    {
      if( !( (0 <= m_dynfMergeEpsilon.GetValue()) &&
	      (m_dynfMergeEpsilon.GetValue() <= 0.5)) )
	{
	  HandleError("Illegal Merge Epsilon value: %f: 0 <= merge epsison <= 0.5", 
		      m_dynfMergeEpsilon.GetValue());
	  m_dynfMergeEpsilon.SetValue(0.4);
	}
      QEMesh::SetMergeEpsilon(m_dynfMergeEpsilon.GetValue());
    }

  m_dyniLOD.Update();
  if ( !( 0 <= m_dyniLOD.GetValue() &&
	  m_dyniLOD.GetValue() < slidelodtSize ) )
    {
      HandleError("Illegal LOD flag value:  %ld", 
		  m_dyniLOD.GetValue());
      m_dyniLOD.SetValue(slidelodtSLF_FULL);
    }

  if ( m_dyniLOD.GetValue() != slidelodtSLF_OFF )
    {
      m_dyniStencil.Update();
      if ( !( 0 <= m_dyniStencil.GetValue() &&
	      m_dyniStencil.GetValue() < slidesttSize ) )
	{
	  HandleError("Illegal Stencil flag value:  %ld", 
		      m_dyniStencil.GetValue());
	  m_dyniStencil.SetValue(slidesttSLF_BOTH);
	}
      
      m_dyniFogOnOff.Update();
      if ( !( 0 <= m_dyniFogOnOff.GetValue() &&
	      m_dyniFogOnOff.GetValue() < slideoftSize ) )
	{
	  HandleError("Illegal Fog On/Off flag value:  %ld", 
		      m_dyniFogOnOff.GetValue());
	  m_dyniFogOnOff.SetValue(slideoftSLF_OFF);
	}
      /*
      m_dyniBlendOnOff.Update();
      if ( !( 0 <= m_dyniBlendOnOff.GetValue() &&
	      m_dyniBlendOnOff.GetValue() < slideoftSize ) )
	{
	  HandleError("Illegal Blend On/Off flag value:  %ld", 
		      m_dyniBlendOnOff.GetValue());
	  m_dyniBlendOnOff.SetValue(slideoftSLF_OFF);
	}
	*/
      m_pCSLIDECameraPath->Update();

      CListIterInt iter;
      CSLIDELightPath *pCSLIDELightPath;
      
      iter.Init(&m_listCSLIDELightPath);
      for ( pCSLIDELightPath = (CSLIDELightPath*)iter.PeekFirst();
	    pCSLIDELightPath != NULL;
	    pCSLIDELightPath = (CSLIDELightPath*)iter.PeekNext() )
	{
	  pCSLIDELightPath->Update();
	}

      CSLIDEAreaLightPath *pCSLIDEAreaLightPath;
      
      iter.Init(&m_listCSLIDEAreaLightPath);
      for ( pCSLIDEAreaLightPath = (CSLIDEAreaLightPath*)iter.PeekFirst();
	    pCSLIDEAreaLightPath != NULL;
	    pCSLIDEAreaLightPath = (CSLIDEAreaLightPath*)iter.PeekNext() )
	{
	  pCSLIDEAreaLightPath->Update();
	}
    }
}

VOID CSLIDERender::HandleReshape(const Vector &vSize)
{
  if ( m_dyniLOD.GetValue() != slideoftSLF_OFF )
    {
      m_pCSLIDETclVariables->SetWidth(vSize(X));
      m_pCSLIDETclVariables->SetHeight(vSize(Y));
    }
}

int CSLIDERender::RenderSelection(GLint *viewport, GLuint *selectBuf, const Vector &vClick,
				  BOOL bRenderAxes)
{
  Matrix mGvrc_obj, mGobj_vrc;

  //printf("Selection: (%.3f, %.3f)\n", vClick(X), vClick(Y));
  
  // DRAW TO SELECTION BUFFER
  gilGetIntegerv(GL_VIEWPORT, viewport);
  gilSelectBuffer(500, selectBuf);
  (void) gilRenderMode(GL_SELECT);

  gilInitNames();
  gilPushName(55);

  // The mGproj_vrc is the Camera's projection matrix
  //   this also does gluPickMatrix(x, y, 1.0, 1.0, viewport);
  m_pCSLIDECameraPath->GetCamera()->ApplyProjectionMatrix(TRUE, vClick(X), vClick(Y));

  gilMatrixMode(GL_MODELVIEW);
  gilLoadIdentity();

  mGvrc_obj = m_pCSLIDECameraPath->GetRootToObject();
  mGobj_vrc = m_pCSLIDECameraPath->GetObjectToRoot();
  {
    if ( slidecuitSLF_RESET_CUI == m_dyniCUIType.GetValue() )
      {
	m_mCameraUI.SetIdentity();
	m_mCameraUIInverse.SetIdentity();
	m_mBall.SetIdentity();
	m_mBallInverse.SetIdentity();
      }
    
    // Apply the Camera Motion Matrix
    gilMultMatrixf(m_mCameraUI.Get());
    
    // Apply the Gvrc_world matrix
    gilMultMatrixf(mGvrc_obj.Get());
    
    mGvrc_obj = m_mCameraUI * mGvrc_obj;
    mGobj_vrc = mGobj_vrc * m_mCameraUIInverse;
    
    // Apply the Crystal Ball Matrix
    gilMultMatrixf(m_mBall.Get());
    
    mGvrc_obj = mGvrc_obj * m_mBall;
    mGobj_vrc = m_mBallInverse * mGobj_vrc;
  }

  // Save for direct Manipulation
  gilGetFloatv(GL_PROJECTION_MATRIX, m_mProjection);
  gilGetFloatv(GL_MODELVIEW_MATRIX, m_mModelview);

  // Draw World
  m_pWorldRoot->RenderSelect(m_pCSLIDECameraPath->GetCamera()->GetVRCToProjection() * mGvrc_obj,
		       &m_boundClip, slideshtSLF_INHERIT, 
		       (SLIDELODType)m_dyniLOD.GetValue());

  // Draw GUI Axes
  if( (m_iSelected != -1) && bRenderAxes )
    {
      gilPushName(SELECTION_AXES);
      {
	int i;
	i = m_puiSelectionHitList[m_iSelected];
	m_pWorldRoot->RenderGUISelected(&(m_puiSelectionNameList[i+4]),
					(int)(m_puiSelectionNameList[i] - 1),
					m_mChildFrame, m_mParentFrame, FALSE, TRUE, 
					0, 0, 0, 0);
      }
      gilPopName();
    }

  gilMatrixMode(GL_PROJECTION);
  gilLoadIdentity();
  gilMatrixMode(GL_MODELVIEW);
  gilLoadIdentity();
  gilFlush();

  // return hits
  return gilRenderMode(GL_RENDER);
}

VOID CSLIDERender::Selection(const Vector &vClick)
{
  GLint viewport[4];
  GLuint selectBuf[500];
  int i, iSelectionHits;
  GLuint j, c, names, uiSelected;
  FLOAT fZMin;

  m_bGraphSelected = FALSE;

  iSelectionHits = RenderSelection(viewport, selectBuf, vClick, TRUE);

  printf( "hits = %d\n", iSelectionHits );
  
  // Find which object was Selected (object or axes)
  fZMin = INFINITY;
  for( i = 0, c = 0; i < iSelectionHits; i++ )
    {
      names = selectBuf[c];	    

      if( names > 10000 ) // this is some error, I'm not sure what
	{
	  printf("ERROR: selection names = %d: ", names);
	  return;
	}

      // Find nearest object
      if( ((float) selectBuf[c+1]/0x7fffffff) < fZMin )
	{
	  fZMin = (float) selectBuf[c+1]/0x7fffffff;
	  uiSelected = c;  // index into start of hit in selectBuf
	} 
      c += 3 + names; // next hit
    }

  /////////////////////////////////////////////////////////////////////
  /// Selected Axis
  if( (iSelectionHits > 0) && (selectBuf[uiSelected+4] == SELECTION_AXES) )
    {
      // Make sure the name corresponds to (55, SELECTION_AXES, 0 or 1, X or Y or Z) 
      ASSERT( selectBuf[uiSelected] == 4 );

      SelectionAxis( selectBuf[uiSelected+5], selectBuf[uiSelected+6]);
      return;
    }

  /////////////////////////////////////////////////////////////////////
  ///  Selected Object - store hit list
  m_iSelected = -1;
  m_iSelectionHits = iSelectionHits;

  for( i = 0, c = 0; i < m_iSelectionHits; i++ )
    {
      names = selectBuf[c];	    
      printf(" names = %d: ", names);

      if( c == uiSelected )
	{
	  ASSERT( i < 100 );
	  m_iSelected = i;
	}
      
      m_puiSelectionHitList[i] = (UINT_32)c;
      m_puiSelectionNameList[c] = selectBuf[c];
      c++;
      m_puiSelectionNameList[c] = selectBuf[c];
      c++;
      m_puiSelectionNameList[c] = selectBuf[c];
      c++;

      for( j = 0; j < names; j++, c++ )
	{
	  m_puiSelectionNameList[c] = selectBuf[c];
	  printf("%d ", selectBuf[c]);
	}
      printf( "\n" );
    }

  INT_32 iCode;
  CHAR pcBuffer[200];
  Tcl_Interp *pInterp;
      
  if( CSLIDEWorld::GetTclGUIExists() )
    {
      // Clear the GUI graph
      pInterp = CSLIDEWorld::GetInterp();
      sprintf(pcBuffer, "Graph:SelectedPathClear 1");
      iCode = Tcl_Eval(pInterp, pcBuffer);
      //ASSERT( iCode == TCL_OK );
      if ( iCode != TCL_OK )
	{
	  printf( "%s\n", pcBuffer );
	  printf("Error %s \n", pInterp->result);
	}
    }

  m_bManipulationMode = FALSE;

  if( m_iSelectionHits > 0 )
    {
      // Find instance and mark GUI graph
      CSLIDEInstance *pCSLIDEInstance;

      i = m_puiSelectionHitList[m_iSelected];
      pCSLIDEInstance = m_pWorldRoot->FindSelection(&(m_puiSelectionNameList[i+4]),
						    (int)(m_puiSelectionNameList[i] - 1));

      if( m_pCSLIDENodeGUISelected != (CSLIDENode*)pCSLIDEInstance )
	{
	  m_pCSLIDENodeGUISelected = (CSLIDENode*)pCSLIDEInstance;
	  m_bXAxisSelected = FALSE;
	  m_bYAxisSelected = FALSE;
	  m_bZAxisSelected = FALSE;
	}
      else
	{
	  m_bManipulationMode = TRUE;
	  m_bNewTransform = TRUE;
	  m_vManipulationStart = vClick;
	}

      if( CSLIDEWorld::GetTclGUIExists() )
	{
	  int iSelected;
	  iSelected = m_iSelected;
	  sprintf(pcBuffer, "Graph:SLIDEObjectSelected %s", pCSLIDEInstance->GetID() );
	  iCode = Tcl_Eval(pInterp, pcBuffer);
	  //ASSERT( iCode == TCL_OK );
	  if ( iCode != TCL_OK )
	    {
	      printf( "%s\n", pcBuffer );
	      printf("Error %s \n", pInterp->result);
	    }
	  // Tcl function will reset m_iSelected so need to store
	  m_iSelected = iSelected;
	  m_bGraphSelected = FALSE;
	}
    }
  else
    {
      m_pCSLIDENodeGUISelected = NULL;
    }
}

VOID CSLIDERender::SelectionManipulation(const Vector &vClick)
{
  GLint viewport[4];
  GLuint selectBuf[500];
  int i, iSelectionHits;
  GLuint j, c, names, uiSelected;
  FLOAT fZMin;

  iSelectionHits = RenderSelection(viewport, selectBuf, vClick, TRUE);

  printf( "hits = %d\n", iSelectionHits );

  // Find which object was Selected (object or axes)
  fZMin = INFINITY;
  for( i = 0, c = 0; i < iSelectionHits; i++ )
    {
      names = selectBuf[c];	    

      if( names > 10000 ) // this is some error, I don't know what
	{
	  printf("ERROR: selection names = %d: ", names);
	  return;
	}

      if( ((float) selectBuf[c+1]/0x7fffffff) < fZMin )
	{
	  fZMin = (float) selectBuf[c+1]/0x7fffffff;
	  uiSelected = c;  // index into start of hit in selectBuf
	} 
      c += 3 + names; // next hit
    }

  // Didn't Select An Object
  if( (iSelectionHits == 0) ||
      ((iSelectionHits > 0) && (selectBuf[uiSelected+4] == SELECTION_AXES)) )
    {
      return;
    }

  //  Selected Object
  m_bManipulationMode = FALSE;

  if( iSelectionHits > 0 )
    {
      i = m_puiSelectionHitList[m_iSelected];

      // This comparision won't work if graph is modified
      if( selectBuf[uiSelected] == m_puiSelectionNameList[i] )
	{
	  for( j = 0; j < selectBuf[uiSelected]; j++ )
	    {
	      if( selectBuf[uiSelected+3+j] != m_puiSelectionNameList[i+3+j] )
		{
		  return;
		}
	    }
	  // The same object was clicked on
	  m_bManipulationMode = TRUE;
	  m_bNewTransform = TRUE;
	  m_vManipulationStart = vClick;
	}
    }
}

VOID CSLIDERender::SelectionCycle()
{
  if( (!GetSelectionEnabled()) || (m_iSelected == -1) || (m_iSelectionHits <= 1) )
    {
      return;
    }

  int iSelected, i;

  // Cycle through the hit list
  iSelected = m_iSelected + 1;
  if( iSelected >= m_iSelectionHits )
    {
      iSelected = 0;
    }
  i = m_puiSelectionHitList[iSelected];

  while( m_puiSelectionNameList[i+4] == SELECTION_AXES ) // Selected Axis
    {
      if( iSelected == m_iSelected )
	{
	  return;
	}

      iSelected++;
      if( iSelected >= m_iSelectionHits )
	{
	  iSelected = 0;
	}
      i = m_puiSelectionHitList[iSelected];
    } 
  
  // Selected Object
  CSLIDEInstance *pCSLIDEInstance;
  INT_32 iCode;
  CHAR pcBuffer[200];
  Tcl_Interp *pInterp;
  
  m_iSelected = iSelected;

  if( CSLIDEWorld::GetTclGUIExists() )
    {
      // Clear the GUI graph
      pInterp = CSLIDEWorld::GetInterp();
      sprintf(pcBuffer, "Graph:SelectedPathClear 1");
      iCode = Tcl_Eval(pInterp, pcBuffer);
      //ASSERT( iCode == TCL_OK );
      if ( iCode != TCL_OK )
	{
	  printf( "%s\n", pcBuffer );
	  printf("Error %s \n", pInterp->result);
	}
    }
  
  // Find instance and mark GUI graph
  pCSLIDEInstance = m_pWorldRoot->FindSelection(&(m_puiSelectionNameList[i+4]),
						(int)(m_puiSelectionNameList[i] - 1));
  m_pCSLIDENodeGUISelected = (CSLIDENode*)pCSLIDEInstance;
  
  if( CSLIDEWorld::GetTclGUIExists() )
    {
      pInterp = CSLIDEWorld::GetInterp();
      sprintf(pcBuffer, "Graph:SLIDEObjectSelected %s", pCSLIDEInstance->GetID() );
      iCode = Tcl_Eval(pInterp, pcBuffer);
      //ASSERT( iCode == TCL_OK );
      if ( iCode != TCL_OK )
	{
	  printf( "%s\n", pcBuffer );
	  printf("Error %s \n", pInterp->result);
	}
    }
}

VOID CSLIDERender::SelectionAxis(GLuint uiChild, GLuint uiAxis)
{
  if( m_bChildAxesSelected != uiChild )
    {
      m_bXAxisSelected = FALSE;
      m_bYAxisSelected = FALSE;
      m_bZAxisSelected = FALSE;
    }
  
  m_bChildAxesSelected = uiChild;
  
  if( uiAxis == X )
    {
      m_bXAxisSelected = !m_bXAxisSelected;
      m_bYAxisSelected = FALSE;
      m_bZAxisSelected = FALSE;
      printf( "X Axis Selected\n");
    }
  else if( uiAxis == Y )
    {
      m_bYAxisSelected = !m_bYAxisSelected;
      m_bXAxisSelected = FALSE;
      m_bZAxisSelected = FALSE;
      printf( "Y Axis Selected\n");
    }
  else if( uiAxis == Z )
    {
      m_bZAxisSelected = !m_bZAxisSelected;
      m_bXAxisSelected = FALSE;
      m_bYAxisSelected = FALSE;
      printf( "Z Axis Selected\n");
    }
  else
    {
      ASSERT(FALSE);
    }
}

VOID CSLIDERender::ResetCamera()
{
  m_mCameraUI.SetIdentity();
  m_mCameraUIInverse.SetIdentity();
  m_mBall.SetIdentity();
  m_mBallInverse.SetIdentity();
}

VOID CSLIDERender::EditSelectedObject()
{
  if( m_pCSLIDENodeGUISelected == NULL )
      return;

  printf( "        EDIT OBJECT: %s\n", m_pCSLIDENodeGUISelected->GetID());

  CWindow *pCWindow;
  CSLIDEViewport *pCSLIDEViewport;
  CSLIDERender *pCSLIDERender;
  CSLIDEWorld *pCSLIDEWorld;

  pCSLIDEWorld = CSLIDEWorld::GetWorld();
  pCWindow = CSLIDEWorld::GetWindowEditObject();
  pCSLIDEViewport = CSLIDEWorld::GetViewportEditObject();
  pCSLIDERender = CSLIDEWorld::GetRenderEditObject();

  if( pCWindow == NULL )
    {

      m_pcWorldRootID = m_pCSLIDENodeGUISelected->GetID();
      m_pWorldRoot = m_pCSLIDENodeGUISelected;
      //Prerender();
      

      /*
      CSLIDELightPath *pCSLIDELightPath;
      CListIterInt iter;

      pCWindow = new CWindow();
      pCSLIDEViewport = new CSLIDEViewport();
      pCSLIDERender = new CSLIDERender();

      pCWindow->Init("[Edit Object]", 0, 0, "WindowEditObject");
      pCSLIDEWorld->InsertWindow(pCWindow);
      pCSLIDEViewport->Init("[Edit Object]", 0, 0, "ViewportEditObject", 
			    "WindowEditObject");
      pCSLIDEWorld->InsertViewport(pCSLIDEViewport);
      pCSLIDERender->Init("[Edit Object]", 0, 0, "ViewportEditObject", 
			  m_pCSLIDECameraPath, m_pcWorldRootID);
				      //m_pCSLIDENodeGUISelected->GetID());
      pCSLIDEWorld->InsertRender(pCSLIDERender);
      
      iter.Init(&m_listCSLIDELightPath);
      for ( pCSLIDELightPath = (CSLIDELightPath*)iter.PeekFirst();
	    pCSLIDELightPath != NULL;
	    pCSLIDELightPath = (CSLIDELightPath*)iter.PeekNext() )
	{
	  pCSLIDERender->AddLightPath(pCSLIDELightPath);
	}

      pCWindow->CreateTclToplevel();

      pCSLIDERender->CheckSemantics(pCSLIDEWorld);
      pCSLIDEViewport->CheckSemantics(pCSLIDEWorld);
      pCWindow->CheckSemantics(pCSLIDEWorld);

      pCSLIDEViewport->Preprocess();
      pCSLIDERender->Preprocess();
      pCWindow->Preprocess();
      */
    } 
}

// Both Mouse Functions get the size of the viewport rectangle
// and the location of the mouse cursor relative to the lower 
// left hand corner
BOOL CSLIDERender::HandleMouseButton(const Vector &vSize,
				     INT_32 iButton, INT_32 iState, 
				     const Vector &vClick)
{
  BOOL bPostRedisplay;

  bPostRedisplay = FALSE;

  if ( iState == GLSI_DOWN )
    {
      if ( m_dyniLOD.GetValue() == slideoftSLF_OFF )
	{
	  return FALSE;
	}

      m_pCSLIDETclVariables->SetWidth(vSize(X));
      m_pCSLIDETclVariables->SetHeight(vSize(Y));
      m_pCSLIDETclVariables->SetMouseX(vClick(X)/vSize(X));
      m_pCSLIDETclVariables->SetMouseY(vClick(Y)/vSize(Y));

      m_vStart = vClick - (vSize / 2.0);
      m_vStart *= ((FLOAT)2.0 / MIN(vSize(X), vSize(Y)));
      m_vLatest = m_vStart;
      m_iButtonCurr = iButton;
      m_fScaleLast = 1.0;
      
      m_iButtonsCurr = m_iButtonsCurr | (0x01 << iButton);
      
      /* Detects all three buttons down */
      if ( ( (0x01 << GLSI_LEFT_BUTTON) |
	     (0x01 << GLSI_MIDDLE_BUTTON) | 
	     (0x01 << GLSI_RIGHT_BUTTON) )== m_iButtonsCurr )
	{
	  m_dyniCamZoom.SetValue(1);
	}
      
      switch ( iButton ) 
	{
	case GLSI_LEFT_BUTTON:
	  {
	    if( GetSelectionEnabled() )
	      {
		Selection(vClick);
	      }
	  }
	break;
	case GLSI_MIDDLE_BUTTON:
	  {
	    if( GetSelectionEnabled() )
	      {
		SelectionManipulation(vClick);
	      }
	  }
	break;
	case GLSI_RIGHT_BUTTON:
	  {
	  }
	break;
	default:
	  ASSERT(FALSE);
	  break;
	}
    }
  else if ( iState == GLSI_UP && iButton == m_iButtonCurr )
    {
      m_iButtonCurr = GLSI_INVALID_BUTTON;
    }

  if ( iState == GLSI_UP )
    {
      m_iButtonsCurr = m_iButtonsCurr & ~(0x01 << iButton);
    }
  
  return bPostRedisplay;
}

// We change the vClick into a vector from the center of the window
// where the radius of the best fit sphere is 1.0
BOOL CSLIDERender::HandleMouseMotion(const Vector &vSize, const Vector &vClick)
{
  BOOL bPostRedisplay;

  bPostRedisplay = FALSE;

  if ( m_dyniLOD.GetValue() == slideoftSLF_OFF )
    {
      m_iButtonCurr = GLSI_INVALID_BUTTON;
      return FALSE;
    }

  if( GetSelectionEnabled() )
    {
      // Selection UI, ignore camera ui
      return HandleMouseMotionDirectManipulation(vSize, vClick); 
    }

  // Camera UI
  if ( m_iButtonCurr != GLSI_INVALID_BUTTON )
    {
      Vector vCurr;
      FLOAT fMultiplier;
      m_pCSLIDETclVariables->SetWidth(vSize(X));
      m_pCSLIDETclVariables->SetHeight(vSize(Y));
      m_pCSLIDETclVariables->SetMouseX(vClick(X)/vSize(X));
      m_pCSLIDETclVariables->SetMouseY(vClick(Y)/vSize(Y));
      
      vCurr = vClick - (vSize / 2.0);
      vCurr *= ((FLOAT)2.0 / MIN(vSize(X), vSize(Y)));
      
      fMultiplier = ((m_iKeysCurr&(CSR_SHIFT_L|CSR_SHIFT_R))?5.0f:1.0f);
      bPostRedisplay = ApplyCUI( GetMouseUIType( m_dyniCUIType.GetValue(), m_iButtonCurr ), 
				 vCurr, 1.0f, fMultiplier, fMultiplier, fMultiplier );
      
      m_vLatest = vCurr;
    }

  return bPostRedisplay;
}

const INT_32 CSLIDERender::m_ppiKeyConversion[CSRSpecialKeysSize][2] =
{
	{ GLSI_KEY_CONTROL_L, CSR_CTRL_L },
	{ GLSI_KEY_CONTROL_R, CSR_CTRL_R },
	{ GLSI_KEY_SHIFT_L, CSR_SHIFT_L },
	{ GLSI_KEY_SHIFT_R, CSR_SHIFT_R }
};

BOOL CSLIDERender::HandleKeyboard(INT_32 iKey, INT_32 iState)
{
	INT_32 iMask = -1;

	for ( int i = 0; i < CSRSpecialKeysSize; i++ )
	{
		if ( iKey == m_ppiKeyConversion[i][0] )
		{
			iMask = m_ppiKeyConversion[i][1];
			break;
		}
	}

	if ( iMask < 0 )
	{ return FALSE; }

	BOOL bPosSet = m_iKeysCurr & ( iMask );


	if ( ! bPosSet && GLSI_DOWN == iState )
	{
		m_iKeysCurr = m_iKeysCurr | ( iMask );
	} else {
		m_iKeysCurr = m_iKeysCurr & ~( iMask );
	}

	return FALSE;
}

// Apply the camera user interface matrices
BOOL CSLIDERender::ApplyCUI(const INT_32 iMUIType, const Vector &vCurr, 
			    const FLOAT &fDeltaTime /* = 1.0f */,
			    const FLOAT &fTMultiplier /* = 1.0f */,
			    const FLOAT &fRMultiplier /* = 1.0f */,
			    const FLOAT &fSMultiplier /* = 1.0f */)
{
  BOOL bPostRedisplay = FALSE;
  Matrix mBall, mBallInverse;
  Matrix mCam, mCamInverse;
  Matrix m1, m1Inverse;
  Vector v1, v2, vNorm;
  Quater q;
  FLOAT theta;

  switch ( iMUIType )
	{
	case slidemuitSLF_NO_MUI:
		{
			bPostRedisplay = FALSE;
		}
		break;
	case slidemuitSLF_CB_GENERAL_ROTATION:
	  {
	    // General Rotations
	    
	    v1.Set(m_vLatest[X], m_vLatest[Y], 
		   (FLOAT)1.0 - (m_vLatest[X]*m_vLatest[X] + m_vLatest[Y]*m_vLatest[Y]));
	    v2.Set(vCurr(X), vCurr(Y), 
		   (FLOAT)1.0 - (vCurr(X)*vCurr(X) + vCurr(Y)*vCurr(Y)));
	    
	    if ( v1[Z] >= 0.0f && v2[Z] >= 0.0f )
	      {
		v1[Z] = (FLOAT)sqrt(v1[Z]);
		v2[Z] = (FLOAT)sqrt(v2[Z]);
	      }
	    else
	      {
		v1[Z] = 0.0;
		v2[Z] = 0.0;
	      }
	    
	    VERIFY(v1.Normalize());
	    VERIFY(v2.Normalize());
	    
	    theta = (FLOAT)MAX(MIN(v1.Dot(v2), 1.0), -1.0);
	    theta = fRMultiplier * (FLOAT)acos(theta);
	    ASSERT( theta >= 0.0f );

	    vNorm = v1 * v2;
	    
	    if ( vNorm.Normalize() )
	      {
		mBall.SetRotate(vNorm, theta);
		mBallInverse.SetRotate(vNorm, -theta);

		bPostRedisplay = TRUE;
	      }

	    q.SetRotation(v1, v2);	    
	    m_pCSLIDETclVariables->SetRotate(q*m_pCSLIDETclVariables->GetRotate()); 
	  }
	  break;
	case slidemuitSLF_CB_SCALE:
	  {
	    FLOAT fScale;

	    // Scaling
	    ASSERT( m_fScaleLast > 0.0 );
	    fScale = (FLOAT)pow(2, fSMultiplier*5*(vCurr(Y) - m_vStart[Y])/2.0);
	    mBall.SetScale(fScale / m_fScaleLast);
	    mBallInverse.SetScale(m_fScaleLast / fScale);

	    m_pCSLIDETclVariables->SetScale(m_pCSLIDETclVariables->GetScale() * 
					    fScale / m_fScaleLast);

	    bPostRedisplay = TRUE;

	    m_fScaleLast = fScale;
	  }
	  break;
	case slidemuitSLF_CB_Z_ROTATION:
	  {
	    // Rotations around the Z-axis

	    v1 = m_vLatest;
	    v2 = vCurr;
	    
	    v1.Normalize();
	    v2.Normalize();

	    theta = (FLOAT)MAX(MIN(v1.Dot(v2), 1.0), -1.0);
	    theta = fRMultiplier * (FLOAT)acos(theta);
	    ASSERT( theta >= 0.0f );
	    
	    vNorm = v1 * v2;
	    if ( vNorm.Normalize() )
	      {
		mBall.SetRotate(vNorm, theta);
		mBallInverse.SetRotate(vNorm, -theta);
	      }
	    else
	      {		    
		mBall.SetRotateZ(theta);
		    mBallInverse.SetRotateZ(-theta);
	      }
	    
	    q.SetRotation(v1, v2);	    
	    m_pCSLIDETclVariables->SetRotate(q*m_pCSLIDETclVariables->GetRotate()); 

	    if ( theta != 0.0 )
	      {
		bPostRedisplay = TRUE;
	      }
	  }
	  break;
	case slidemuitSLF_CAM_Z_TRANSLATE:
	  {
	    Vector vTrans;
	    
	    // Translation
	    vTrans[X] = 0;
	    vTrans[Y] = 0;
	    vTrans[Z] = fTMultiplier * m_dynfCUISpeed.GetValue() * 
	      (FLOAT)(vCurr(Y) - m_vLatest[Y])/2;
	    
	    mCam.SetTranslate(vTrans);
	    mCamInverse.SetTranslate(- vTrans);
	    
	    bPostRedisplay = TRUE;
	  }
	  break;
	case slidemuitSLF_CAM_PAN:
	  {
	    Vector vTrans;
	    
	    // Translation
	    vTrans[X] = fTMultiplier * m_dynfCUISpeed.GetValue() * 
	      (FLOAT)(vCurr(X) - m_vLatest[X])/2.0f;
	    vTrans[Y] = fTMultiplier * m_dynfCUISpeed.GetValue() * 
	      (FLOAT)(vCurr(Y) - m_vLatest[Y])/2.0f;
	    vTrans[Z] = 0;
	    
	    mCam.SetTranslate(vTrans);
	    mCamInverse.SetTranslate(- vTrans);
	    
	    bPostRedisplay = TRUE;
	  }
	  break;
	case slidemuitSLF_CAM_GENERAL_ROTATION:
	  {
	    // General Rotations
	    
	    v1.Set(m_vLatest[X], m_vLatest[Y], 
		   (FLOAT)1.0 - (m_vLatest[X]*m_vLatest[X] + m_vLatest[Y]*m_vLatest[Y]));
	    v2.Set(vCurr(X), vCurr(Y), 
		   (FLOAT)1.0 - (vCurr(X)*vCurr(X) + vCurr(Y)*vCurr(Y)));
	    
	    if ( v1[Z] >= 0.0f && v2[Z] >= 0.0f )
	      {
		v1[Z] = (FLOAT)sqrt(v1[Z]);
		v2[Z] = (FLOAT)sqrt(v2[Z]);
	      }
	    else
	      {
		v1[Z] = 0.0;
		v2[Z] = 0.0;
	      }
	    
	    VERIFY(v1.Normalize());
	    VERIFY(v2.Normalize());
	    
	    theta = (FLOAT)MAX(MIN(v1.Dot(v2), 1.0), -1.0);
	    theta = fRMultiplier * (FLOAT)acos(theta);
	    ASSERT( theta >= 0.0f );

	    vNorm = v1 * v2;
	    
	    if ( vNorm.Normalize() )
	      {
		mCam.SetRotate(vNorm, theta);
		mCamInverse.SetRotate(vNorm, -theta);

		bPostRedisplay = TRUE;
	      }
	  }
	  break;
    case slidemuitSLF_CAM_INTEGRAL:
	  {
		FLOAT fIter = 0.0f;
		FLOAT fTimeSlice = 0.03f;
		mCam.SetIdentity();
		mCamInverse.SetIdentity();

		  // General Rotations
		for (fIter = 0.0f; fIter < fDeltaTime; fIter += fTimeSlice )
		{
			fTimeSlice = MIN(fTimeSlice, fDeltaTime - fIter);

			v1 = vCurr;
			v1[X] = (FLOAT)MAX(MIN(v1(X), 1.0), -1.0);
			v1[Y] = (FLOAT)MAX(MIN(v1(Y), 1.0), -1.0);
			v1[Z] = (FLOAT)MAX(MIN(v1(Z), 1.0), -1.0);
			
			theta = fTimeSlice * fRMultiplier *
			  m_dynfCUISpeed.GetValue( ) * (FLOAT) asin( v1(Y) );
			m1.SetRotateX( -theta );
			mCam = mCam * m1;
			m1Inverse.SetRotateX( theta );
			mCamInverse = m1Inverse * mCamInverse;
			
			theta = fTimeSlice * fRMultiplier *
			  m_dynfCUISpeed.GetValue( ) * (FLOAT) asin( v1(X) );
			m1.SetRotateY( theta );
			mCam = mCam * m1;
			m1Inverse.SetRotateY( -theta );
			mCamInverse = m1Inverse * mCamInverse;
			
			v1.Null();
			v1[Z] = fTimeSlice * fTMultiplier * m_dynfCUISpeed.GetValue( );
			m1.SetTranslate(v1);
			mCam = mCam * m1;
			m1Inverse.SetTranslate(-v1);
			mCamInverse = m1Inverse * mCamInverse;
		}
		bPostRedisplay = TRUE;
	  }
	  break;
	} // End switch

      if ( bPostRedisplay && GetCrystalBallEnabled() )
	{
	  // Update the Camera UI Matrix if it has changed
	  m_mCameraUI = mCam * m_mCameraUI;
	  m_mCameraUIInverse = m_mCameraUIInverse * mCamInverse;
	  
	  
	  // Update the Crystal Ball Matrix if it has changed
	  Matrix mGvrp_vrc, mGvrc_vrp;
	  
	  mGvrp_vrc = m_pCSLIDECameraPath->GetCamera()->GetVRCToVRP();
	  mGvrc_vrp = m_pCSLIDECameraPath->GetCamera()->GetVRPToVRC();
	  
	  // Bring the Center of the VRC to the origin before applying scales and rotates
	  m_mBall = m_pCSLIDECameraPath->GetObjectToRoot() * m_mCameraUIInverse * 
	    mGvrc_vrp * mBall * mGvrp_vrc * m_mCameraUI * 
	    m_pCSLIDECameraPath->GetRootToObject() * m_mBall;
	  m_mBallInverse = m_mBallInverse * 
	    m_pCSLIDECameraPath->GetObjectToRoot() * 
	    m_mCameraUIInverse * mGvrc_vrp * 
	    mBallInverse * mGvrp_vrc * m_mCameraUI * 
	    m_pCSLIDECameraPath->GetRootToObject();
	  
	  // Even if the world is paused we want the UI to take effect
	  CSLIDEWorld::GetWorld()->ScheduleDynamics(0);
	}
  return bPostRedisplay;
} // End CSLIDERender::ApplyCUI

BOOL CSLIDERender::HandleMouseMotionDirectManipulation(Vector vSize, Vector vClick)
{
  if( (m_iButtonCurr == GLSI_INVALID_BUTTON) ||
      (!m_bManipulationMode) )
    {
      return FALSE;
    }
  
  FLOAT f, fProjectionMagnitude;
  DynVector *pdynv;
  Vector vCurr, v1, v2;
  Matrix mj, mGproj_vrc, mGproj_frame;
  CSLIDETransform *pCSLIDETransform;
  
  m_pCSLIDETclVariables->SetWidth(vSize(X));
  m_pCSLIDETclVariables->SetHeight(vSize(Y));
  m_pCSLIDETclVariables->SetMouseX(vClick(X)/vSize(X));
  m_pCSLIDETclVariables->SetMouseY(vClick(Y)/vSize(Y));
  
  vCurr = vClick - (vSize / 2.0);
  vCurr *= ((FLOAT)2.0 / MIN(vSize(X), vSize(Y)));
  
  // Test
  mj = m_mChildFrame;
  mj = m_mParentFrame;
  mj.Set(m_mProjection);
  mj.Set(m_mModelview);
  
  // Set Matrices
  mGproj_vrc.Set(m_mProjection);
  
  if( m_bChildAxesSelected ) // TURTLE WALK
    mGproj_frame = mGproj_vrc * m_mChildFrame;
  else                       // COMMAND POST
    mGproj_frame = mGproj_vrc * m_mParentFrame;
  
  
  switch( m_iButtonCurr )
    {
    case GLSI_LEFT_BUTTON:
      {
	// TRANSLATE
	if( !(m_bXAxisSelected || m_bYAxisSelected || m_bZAxisSelected ) )
	  {
	    return FALSE;
	  }
	
	CSLIDETranslate *pCSLIDETranslate;
	
	pCSLIDETranslate = new CSLIDETranslate();
	pCSLIDETransform = pCSLIDETranslate;
	pdynv = new DynVector();
	
	v1.Set((FLOAT)(vClick[X] - m_vManipulationStart[X]), 
	       (FLOAT)(vClick[Y] - m_vManipulationStart[Y]), 
	       0);
	
	if( m_bXAxisSelected )
	  v2.Set(1, 0, 0, 0);
	else if( m_bYAxisSelected )
	  v2.Set(0, 1, 0, 0);
	else if( m_bZAxisSelected )
	  v2.Set(0, 0, 1, 0);

	v2 = mGproj_frame * v2;
	fProjectionMagnitude = v2.Magnitude();

	v2.Normalize();
	f = v1.Dot(v2);
	// projection coords (do I need to do vCurr stuff too?)
	f = 2*f / fProjectionMagnitude;
	
	if( m_bXAxisSelected )
	  pdynv->Set(f, 0, 0, 0);
	else if( m_bYAxisSelected )
	  pdynv->Set(0, f, 0, 0);
	else if( m_bZAxisSelected )
	  pdynv->Set(0, 0, f, 0);

	pCSLIDETranslate->Init("[Manipulation]", 0, 0, pdynv);
	delete pdynv;

	printf( "************************************ TRANSLATING BY: %f\n", f);
      }
    break;
    case GLSI_MIDDLE_BUTTON:
      {
	// SCALE

	return FALSE;

	/*
	if( !(m_bXAxisSelected || m_bYAxisSelected || m_bZAxisSelected ) )
	  {
	    return FALSE;
	  }

	CSLIDEScale *pCSLIDEScale;
	
	pCSLIDEScale = new CSLIDEScale();
	pCSLIDETransform = pCSLIDEScale;
	pdynv = new DynVector();
	
	v1.Set((FLOAT)(vClick[X] - m_vManipulationStart[X]), 
	       (FLOAT)(vClick[Y] - m_vManipulationStart[Y]), 
	       0);

	if( m_bXAxisSelected )
	  v2.Set(1, 0, 0, 0);
	else if( m_bYAxisSelected )
	  v2.Set(0, 1, 0, 0);
	else if( m_bZAxisSelected )
	  v2.Set(0, 0, 1, 0);

	v2 = mGproj_frame * v2;
	fProjectionMagnitude = v2.Magnitude();

	v2.Normalize();
	f = v1.Dot(v2);
	// projection coords (do I need to do vCurr stuff too?)
	f = 2*f / fProjectionMagnitude;

	if( f < 1.0 )
	  f = 1.0;

	if( m_bXAxisSelected )
	  pdynv->Set(f, 1, 1, 0);
	else if( m_bYAxisSelected )
	  pdynv->Set(1, f, 1, 0);
	else if( m_bZAxisSelected )
	  pdynv->Set(1, 1, f, 0);

	pCSLIDEScale->Init("[Manipulation]", 0, 0, pdynv);
	delete pdynv;

	printf( "************************************ SCALING BY: %f \n", f);	
	*/
      }
    break;
    case GLSI_RIGHT_BUTTON:
      {
	// ROTATE

	return FALSE;

	/*
	if( !(m_bXAxisSelected || m_bYAxisSelected || m_bZAxisSelected ) )
	  {
	    return FALSE;
	  }
	DynFloat *pdynf;
	CSLIDERotate *pCSLIDERotate;
	
	pCSLIDERotate = new CSLIDERotate();
	pCSLIDETransform = pCSLIDERotate;
	pdynv = new DynVector();
	pdynf = new DynFloat();
	
	v1.Set((FLOAT)(vClick[X] - m_vManipulationStart[X])/vSize(X), 
	       (FLOAT)(vClick[Y] - m_vManipulationStart[Y])/vSize(Y), 
	       0);	
	
	if( m_bXAxisSelected )
	  v2.Set(1, 0, 0, 0);
	else if( m_bYAxisSelected )
	  v2.Set(0, 1, 0, 0);
	else if( m_bZAxisSelected )
	  v2.Set(0, 1, 0, 0);

	v2 = mGproj_frame * v2;
	fProjectionMagnitude = v2.Magnitude();

	v2.Normalize();
	f = v1.Dot(v2);
	// projection coords (do I need to do vCurr stuff too?)
	f = 2*f / fProjectionMagnitude;

	f = DEGTORAD(360) * f / v1.Magnitude();
	pdynf->SetValue(f);

	if( m_bXAxisSelected )
	  pdynv->Set(1, 0, 0, 0);
	else if( m_bYAxisSelected )
	  pdynv->Set(0, 1, 0, 0);
	else if( m_bZAxisSelected )
	  pdynv->Set(0, 0, 1, 0);

	pCSLIDERotate->Init("[Manipulation]", 0, 0, pdynv, pdynf);
	delete pdynv;
	delete pdynf;

	printf( "ROTATING BY: %f ************************************\n", f);	
	*/
      }
    break;
    default:
      ASSERT(FALSE);
      break;
    }

  // Update the Transforms
  if( m_bChildAxesSelected ) 
    {
      // TURTLE WALK
      ((CSLIDEInstance *)m_pCSLIDENodeGUISelected)->InsertTurtleWalkTransform(pCSLIDETransform, m_bNewTransform);
    }
  else // COMMAND POST                        
    {
      ((CSLIDEInstance *)m_pCSLIDENodeGUISelected)->InsertCommandPostTransform(pCSLIDETransform, m_bNewTransform);
    }
  
  if( CSLIDEWorld::GetTclGUIExists() )
    {
      ((CSLIDEInstance *)m_pCSLIDENodeGUISelected)->GetParameters(CSLIDEWorld::GetInterp());
    }
  
  // Redisplay (m_pCSLIDENodeGUISelected should be an instance)
  CSLIDEWorld::GetWorld()->Reprerender(m_pCSLIDENodeGUISelected);
  CSLIDEWorld::GetWorld()->ScheduleDynamics(0);
  
  m_vLatest = vCurr;
  m_bNewTransform = FALSE;
  
  return TRUE; // PostRedisplay
}

INT_32 CSLIDERender::m_ppiMUITTable[slidecuitSize][3] =
{
  // SLF_RESET_CUI
  { slidemuitSLF_NO_MUI,
    slidemuitSLF_NO_MUI,
    slidemuitSLF_NO_MUI },
  // SLF_PAUSE_CUI
  { slidemuitSLF_NO_MUI,
    slidemuitSLF_NO_MUI,
    slidemuitSLF_NO_MUI },
  // SLF_CAD_CUI
  { slidemuitSLF_CB_GENERAL_ROTATION,
    slidemuitSLF_CB_SCALE,
	slidemuitSLF_CAM_PAN },
  // SLF_FLY_CUI
  { slidemuitSLF_NO_MUI,
    slidemuitSLF_NO_MUI,
    slidemuitSLF_NO_MUI },
  // SLF_EXPLORE_CUI
  { slidemuitSLF_CAM_GENERAL_ROTATION,
    slidemuitSLF_CAM_Z_TRANSLATE,
	slidemuitSLF_CAM_PAN },
};


INT_32 CSLIDERender::GetMouseUIType( INT_32 iUIType, INT_32 iButton )
{
  INT_32 l_iButton = ( GLSI_LEFT_BUTTON == iButton ) ?
    0 : ( (GLSI_MIDDLE_BUTTON == iButton ) ? 1 : 2 );
  
  if (( slidecuitSLF_CAD_CUI == iUIType ) &&
      ( m_iKeysCurr & (CSR_CTRL_L|CSR_CTRL_R) ) )
    {
      return slidemuitSLF_NO_MUI;
    }
  
  return m_ppiMUITTable[ iUIType ][ l_iButton ];
}

INT_32 CSLIDERender::GetUpdateMouseUIType( INT_32 iUIType, INT_32 iButton )
{
  /*
  INT_32 l_iButton = ( GLSI_LEFT_BUTTON == iButton ) ?
    0 : ( (GLSI_MIDDLE_BUTTON == iButton ) ? 1 : 2 );
  */

  // May do tabular layout, but only a few cases here
  switch ( iUIType )
    {
    case slidecuitSLF_FLY_CUI:
      {
	return slidemuitSLF_CAM_INTEGRAL;
      }
    break;
    case slidecuitSLF_CAD_CUI:
      {
	if ( m_iKeysCurr & (CSR_CTRL_L|CSR_CTRL_R) )
	  {
	    return slidemuitSLF_CAM_INTEGRAL;
	  }
      }
    break;
    default:
      {
	return slidemuitSLF_NO_MUI;
      }
    break;
    } // end switch (iUIType)
  
  return slidemuitSLF_NO_MUI;
}

VOID CSLIDERender::UpdateCameraUI( )
{
  Vector v1;
  Matrix m1, m1Inverse;
  BOOL bPostRedisplay = FALSE;
  TimeValue tvCurrTime;
  FLOAT fDeltaTime;
  FLOAT fTimeMult = 1.0f;
  FLOAT fTMult = 1.0f;
  FLOAT fRMult = 1.0f;
  FLOAT fSMult = 1.0f;
  
  // If Crystal Ball not enabled, don't alter matrices
  if ( ! GetCrystalBallEnabled() )
    {
      return;
    }
  
  tvCurrTime.RecordTime();
  fDeltaTime = (FLOAT)(tvCurrTime - m_tvLastRender).GetSeconds();

  switch ( m_iButtonsCurr )
    {
      // Left
    case ( 0x01 << GLSI_LEFT_BUTTON ):
      { }
    break;
    // Middle
    case ( 0x01 << GLSI_MIDDLE_BUTTON ):
      { fTMult = 0.0f; }
    break;
    // Right
    case ( 0x01 << GLSI_RIGHT_BUTTON ):
      { 
	fTMult = -1.0f;
	fSMult = -1.0f;
      }
    break;
    default:
      return;
    } // End switch ( m_iButtonsCurr )
  
#ifdef WIN32
  // Weirdness with time. seems like NT clock moves faster
  fTimeMult *= ((m_iKeysCurr&(CSR_SHIFT_L|CSR_SHIFT_R))?5.0f:1.0f);
#else
  // Weirdness with time. seems like NT clock moves faster
  fTimeMult *= ((m_iKeysCurr&(CSR_SHIFT_L|CSR_SHIFT_R))?25.0f:5.0f);
#endif // WIN32
  switch( m_dyniCUIType.GetValue( ) )
    {
    default:
      {
	if ( slidemuitSLF_CAM_INTEGRAL ==
	     GetUpdateMouseUIType( m_dyniCUIType.GetValue(), m_iButtonCurr ) )
	  {
	    // Translation and lift
	    bPostRedisplay = ApplyCUI( slidemuitSLF_CAM_INTEGRAL,
				       m_vLatest, fTimeMult*fDeltaTime,
				       fTMult, fRMult, fSMult);
	  }
      }
    break;
    } // End switch( m_dyniCUIType.GetValue( ) )
  
  if ( bPostRedisplay )
    {
      m_mCameraUI = m1 * m_mCameraUI;
      m_mCameraUIInverse = m_mCameraUIInverse * m1Inverse;
      
      // Even if the world is paused we want the UI to take effect
      CSLIDEWorld::GetWorld()->ScheduleDynamics(0);
    } 
}

VOID CSLIDERender::Prerender()
{
  if ( !(!m_dyniLOD.GetDynamic() && 
	 m_dyniLOD.GetValue() == slideoftSLF_OFF) )
    {
      m_pWorldRoot->Prerender();
    }
}

VOID CSLIDERender::Render(const Vector &vVPOrigin, const Vector &vVPSize)
{
  Matrix mGvp_proj, mGvrc_obj, mGobj_vrc;
  Vector vOrigin, vSize, vFrustum, vRatios;

  if ( m_dyniLOD.GetValue() == slideoftSLF_OFF )
    {
      return;
    }

  ResetGL();


  UpdateCameraUI();

  //////////////////////////////////////////////////////////////////////
  // Create the Stencil and place it in a Display List
  CreateStencil(vVPOrigin, vVPSize);

  if ( m_dyniStencil.GetValue() == slidesttSLF_BOTH )
    {
      gilDisable(GL_STENCIL_TEST); 
    }
  else
    {
      // Clear the Stencil buffer and put in the correct stencil
      gilEnable(GL_STENCIL_TEST); 
      
      // clear the stencil buffer to 0x0
      gilClearStencil(0x0);
      gilClear(GL_STENCIL_BUFFER_BIT);

      // Apply the Stencil 
      // (This might need to happen multiple times if we are using CDA's)
      ApplyStencil();
    }
  // Create the Stencil and place it in a Display List
  //////////////////////////////////////////////////////////////////////

  //////////////////////////////////////////////////////////////////////
  // Fog
  if ( GetFogEnabled() && 
       (m_dyniFogOnOff.GetValue() == slideoftSLF_ON) )
    {
      ASSERT( m_pCSLIDEFog != NULL );
      m_pCSLIDEFog->Render();
    }
  // Fog
  //////////////////////////////////////////////////////////////////////

  //printf("START: Render: Max Attrib Depth = %ld\n", m_uiAttributeDepthMax);

  m_pCSLIDECameraCurr = m_pCSLIDECameraPath->GetCamera();

  // Set the NormalRenderLength based on the camera's Frustum
  vFrustum = m_pCSLIDECameraPath->GetCamera()->GetFrustumMax() -
             m_pCSLIDECameraPath->GetCamera()->GetFrustumMin();
  CSLIDEFace::SetNormalsRenderLength((FLOAT)0.05 * MIN(vFrustum[X], vFrustum[Y]));

  // Set up the Viewport mapping matrix
  ASSERT( vVPSize(X) != 0 && vVPSize(Y) != 0 );

  vRatios.Set(vFrustum(X)/vVPSize(X), vFrustum(Y)/vVPSize(Y), 0.0);
  vRatios /= MAX(vRatios[X], vRatios[Y]);
  vSize.Set(vVPSize(X) * vRatios(X), vVPSize(Y) * vRatios(Y), 0.0);
  vOrigin = vVPOrigin + ((vVPSize - vSize) / 2.0);

  gilViewport((GLint)vOrigin(X), (GLint)vOrigin(Y), 
	      (GLsizei)vSize(X), (GLsizei)vSize(Y));

  // The mGproj_vrc is the Camera's projection matrix
  m_pCSLIDECameraPath->GetCamera()->ApplyProjectionMatrix(FALSE, 0, 0);

  gilMatrixMode(GL_MODELVIEW);
  gilLoadIdentity();

  // These lines look backward but they are not
  // we need the inverse of the camera path
  mGvrc_obj = m_pCSLIDECameraPath->GetRootToObject();
  mGobj_vrc = m_pCSLIDECameraPath->GetObjectToRoot();

  //if ( ! ( GetCrystalBallEnabled() && !GetRotateClippedEnabled() ) )
  if ( ! GetCrystalBallEnabled() )
    {
    } 
  else
    {
      if ( slidecuitSLF_RESET_CUI == m_dyniCUIType.GetValue() )
	{
	  m_mCameraUI.SetIdentity();
	  m_mCameraUIInverse.SetIdentity();
	  m_mBall.SetIdentity();
	  m_mBallInverse.SetIdentity();
	}

      if (m_dyniCamZoom.GetValue() &&
	  m_pWorldRoot->GetBoundValid() &&
	  !m_pWorldRoot->GetBoundEmpty())
	{
	  Matrix mGvrc_zoom, mGzoom_vrc, m;
	  CBound bound;
	  
	  m_dyniCamZoom.SetValue(0);
	  
	  m.SetIdentity();
	  m_pWorldRoot->InsertIntoBound(m, bound);
	  m_pCSLIDECameraPath->GetCamera()->CalculateZoom(mGzoom_vrc, 
							  mGvrc_zoom, 
							  m_mCameraUI * mGvrc_obj * m_mBall,
							  bound);
	  
	  m_mCameraUI = mGzoom_vrc * m_mCameraUI;
	  m_mCameraUIInverse = m_mCameraUIInverse * mGvrc_zoom;
	}

    }  // End if

  if ( GetCrystalBallEnabled() && 
       // If gets more complex, may want function to check if lighting moves
       ( slidecuitSLF_FLY_CUI == m_dyniCUIType.GetValue() ||
	 slidecuitSLF_EXPLORE_CUI == m_dyniCUIType.GetValue() ) )
    {
      // Lights stay with object space (for FLY and EXPLORE)
      mGvrc_obj = m_mCameraUI * mGvrc_obj * m_mBall;
      mGobj_vrc = m_mBallInverse * mGobj_vrc * m_mCameraUIInverse;
    }
  

  // Apply the Gvrc_world matrix for lights
  gilPushMatrix();
  {
    gilMultMatrixf(mGvrc_obj.Get());
  
  // Deal with the lights
  CListInt *plistCSLIDELightPath;

  plistCSLIDELightPath = NULL;

  if ( GetLightingEnabled() && 
       (m_listCSLIDELightPath.GetLength() + m_listCSLIDEAreaLightPath.GetLength()) != 0 )
    {
      CListIterInt iter;
      CSLIDELightPath *pCSLIDELightPath;
      CSLIDELight *pCSLIDELight;

      gilEnable(GL_LIGHTING);
      CSLIDELight::ResetGLLights();

      iter.Init(&m_listCSLIDELightPath);
      for ( pCSLIDELightPath = (CSLIDELightPath*)iter.PeekFirst();
	    pCSLIDELightPath != NULL;
	    pCSLIDELightPath = (CSLIDELightPath*)iter.PeekNext() )
	{
	  plistCSLIDELightPath = &m_listCSLIDELightPath;
	  
	  if ( pCSLIDELightPath->GetOnOff() == slideoftSLF_ON )
	    {
	      // We need to make the Gvrc_light matrix to be part
	      // of the light path not the light, because there can be
	      // multiple instances of a single light

	      pCSLIDELight = pCSLIDELightPath->GetLight();
	      
	      pCSLIDELightPath->SetVRCToLight() = 
		pCSLIDELightPath->GetRootToObject() * mGobj_vrc;
	      pCSLIDELightPath->SetLightToVRC() = 
		mGvrc_obj * pCSLIDELightPath->GetObjectToRoot();
	      
	      gilPushMatrix();
	      {
		gilMultMatrixf(pCSLIDELightPath->GetObjectToRoot().Get());

		pCSLIDELight->Place();
	      }
	      gilPopMatrix();
	    }
	}

      CSLIDEAreaLightPath *pCSLIDEAreaLightPath;
      CSLIDEAreaLight *pCSLIDEAreaLight;

      iter.Init(&m_listCSLIDEAreaLightPath);
      for ( pCSLIDEAreaLightPath = (CSLIDEAreaLightPath*)iter.PeekFirst();
	    pCSLIDEAreaLightPath != NULL;
	    pCSLIDEAreaLightPath = (CSLIDEAreaLightPath*)iter.PeekNext() )
	{
	  if ( pCSLIDEAreaLightPath->GetOnOff() == slideoftSLF_ON )
	    {
	      // We need to make the Gvrc_light matrix to be part
	      // of the light path not the light, because there can be
	      // multiple instances of a single light

	      pCSLIDEAreaLight = pCSLIDEAreaLightPath->GetAreaLight();
	      
	      gilPushMatrix();
	      {
		gilMultMatrixf(pCSLIDEAreaLightPath->GetObjectToRoot().Get());

		pCSLIDEAreaLight->Place();
	      }
	      gilPopMatrix();
	    }
	}
    }
  else
    {
      gilDisable(GL_LIGHTING);
    }
  }
  gilPopMatrix();
  // Matrix for lights

  if ( GetCrystalBallEnabled() && 
       // If gets more complex, may want function to check if lighting moves
       !( slidecuitSLF_FLY_CUI == m_dyniCUIType.GetValue() ||
	  slidecuitSLF_EXPLORE_CUI == m_dyniCUIType.GetValue() ) )
    {
      // Lights stay with object space (for FLY and EXPLORE)
      mGvrc_obj = m_mCameraUI * mGvrc_obj * m_mBall;
      mGobj_vrc = m_mBallInverse * mGobj_vrc * m_mCameraUIInverse;
    }
  gilMultMatrixf(mGvrc_obj.Get());

  CSLIDESurface::GetSLF_INHERIT()->Render();
  m_pWorldRoot->Render(m_pCSLIDECameraPath->GetCamera()->GetVRCToProjection() * mGvrc_obj,
		       &m_boundClip, 
		       slideshtSLF_INHERIT, 
		       (SLIDELODType)m_dyniLOD.GetValue());
/*
  m_pWorldRoot->Render(mGvp_proj, 
		       m_pCSLIDECameraPath->GetCamera()->GetVRCToProjection(), 
		       m_pCSLIDECameraPath->GetCamera()->GetProjectionToVRC(), 
		       mGvrc_obj, mGobj_vrc,
		       &m_boundClip, 
		       plistCSLIDELightPath,
		       CSLIDESurface::GetSLF_INHERIT(),
		       slideshtSLF_INHERIT, 
		       (SLIDELODType)m_dyniLOD.GetValue());
*/

  // Draw GUI Selected bboxes
  if ( m_bGraphSelected )
    {
      // This is wrong because doesn't get correct path from root, but do it for now
      m_pCSLIDENodeGUISelected->RenderGraphSelectedBound();
    }
  else if( m_iSelected != -1 )
    {
      int i;
      i = m_puiSelectionHitList[m_iSelected];
      m_pWorldRoot->RenderGUISelected(&(m_puiSelectionNameList[i+4]),
				      (int)(m_puiSelectionNameList[i] - 1),
				      m_mChildFrame, m_mParentFrame, 
				      TRUE, GetSelectionEnabled(),
				      m_bChildAxesSelected, m_bXAxisSelected,
				      m_bYAxisSelected, m_bZAxisSelected);
    }

  if ( m_pmeBoundaryEdges != NULL )
    {
      QEMeshIter iterQEMesh;
      QEEdge *pe;

      PushAuxillaryGeometryAttributes();
      {
	gilColor3f(1.0, 0.0, 0.0);
	gilBegin(GL_LINES);
	{
	  iterQEMesh.Init(m_pmeBoundaryEdges);
	  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdgeBoundary() )
	    {
	      gilVertex3fv(pe->GetVertex()->GetPoint().Get());
	      gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
	    }
	}
	gilEnd();
      }
      PopAuxillaryGeometryAttributes();
    }

  gilMatrixMode(GL_PROJECTION);
  gilLoadIdentity();
  
  gilMatrixMode(GL_MODELVIEW);
  gilLoadIdentity();

  // Render Clip Bound
  if ( GetRenderClipBoundEnabled() )
    {
      PushAuxillaryGeometryAttributes();
      {
	gilColor3f(1.0, 1.0, 1.0);
	m_boundClip.Render();
      }
      PopAuxillaryGeometryAttributes();
    }
/*
  if ( TRUE )
    {
	    
      gilPushMatrix();
	  {
      gilMultMatrixf(m_pCSLIDECameraPath->GetCamera()->GetVRPToVRC().Get());
      PushAuxillaryGeometryAttributes();
      {
      gilColor3f(1.0, 1.0, 1.0);
  gilBegin(GL_LINES);
  {    
    gilVertex3f( -0.1f, -0.1f, -1.0f);
    gilVertex3f(0.1f, 0.1f, 1.0f);
  }
  gilEnd();
  gilBegin(GL_LINES);
  {    
    gilVertex3f( 0.1f, -0.1f, -1.0f);
    gilVertex3f(-0.1f, 0.1f, 1.0f);
  }
  gilEnd();
  gilBegin(GL_LINES);
  {    
    gilVertex3f( 0.0f, 0.1f, -1.0f);
    gilVertex3f(0.0f, -0.1f, 1.0f);
  }
  gilEnd();
      }
      PopAuxillaryGeometryAttributes();

	  }
      gilPopMatrix();

    }

*/
  //printf("  END: Render: Max Attrib Depth = %ld\n", m_uiAttributeDepthMax);

  // Update last render time
  m_tvLastRender.RecordTime();

}

VOID CSLIDERender::ApplyStencil()
{
  if ( m_bUseStencil )
    {
      ASSERT( gilIsList(m_GLuiStencil) );
      gilCallList(m_GLuiStencil);
    }
}

VOID CSLIDERender::CreateStencil(const Vector &/*vOrigin*/, const Vector &vSize)
{
  UINT_32 i;

  if ( m_dyniStencil.GetValue() == slidesttSLF_BOTH )
    {
      m_bUseStencil = FALSE;
    }
  else
    {
      //ASSERT( gilIsList(m_GLuiStencil) );
      gilNewList(m_GLuiStencil, GL_COMPILE);
      {
	gilPushAttrib(GL_LINE_BIT | GL_CURRENT_BIT);
	{
	  // make sure the line is thickness 1
	  gilLineWidth(1.0);
	  
	  // create a horizontally interlaced stencil area
	  gilMatrixMode(GL_PROJECTION);
	  gilPushMatrix();
	  {
	    gilLoadIdentity();
	    // set up window coordinate system from upper-left to lower-right
	    gilOrtho(0.0, vSize(X), 0.0, vSize(Y), -1000.0, 1000.0);
	    
	    gilMatrixMode(GL_MODELVIEW);
	    gilPushMatrix();
	    {
	      gilLoadIdentity();
	      
	      gilColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
	      gilDepthMask(GL_FALSE);

	      gilStencilFunc (GL_ALWAYS, 0x1, 0x1);
	      gilStencilOp (GL_REPLACE, GL_REPLACE, GL_REPLACE);
	      
	      if ( m_dyniStencil.GetValue() == slidesttSLF_ODD )
		{
		  i = 0;
		}
	      else
		{
		  i = 1;
		}

	      gilBegin(GL_LINES);
	      {
		for ( ; i < vSize(Y); i += 2 )
		  {
		    gilVertex3f(0.0, (GLfloat)i, 0.0);
		    gilVertex3f(vSize(X), (GLfloat)i, 0.0);
		  }
	      }
	      gilEnd();
	      
	      // protect ourselves from having some other code we don't know about from
	      // drawing on the stencil buffer
	      gilStencilFunc(GL_NOTEQUAL, 0x1, 0x1);
	      gilStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

	      // Restore OpenGL to drawing to the Frame Buffer
	      gilColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	      gilDepthMask(GL_TRUE);
	    }
	    gilPopMatrix();
	  }
	  gilMatrixMode(GL_PROJECTION);
	  gilPopMatrix();
	}
	gilPopAttrib();
      }
      gilEndList();

      m_bUseStencil = TRUE;
    }
}

VOID CSLIDERender::ResetGL()
{
  // Verify that we are not overloading the Attribute Stack
  int iAttributeDepthMax;
  
  gilGetIntegerv(GL_MAX_ATTRIB_STACK_DEPTH, &iAttributeDepthMax);
  ASSERT( iAttributeDepthMax >= 16 );
  m_uiAttributeDepthMax = (UINT_32)iAttributeDepthMax;

  gilShadeModel(GL_FLAT);

  // Set up Texture Mapping
  if ( GetTextureMappingEnabled() )
    {
      gilEnable(GL_TEXTURE_2D);
    }
  else
    {
      gilDisable(GL_TEXTURE_2D);
    }
  // Use the default texture
  gilBindTexture(GL_TEXTURE_2D, 0);

  GLfloat vColor[4] = {1.0, 1.0, 1.0, 1.0};

  gilTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, vColor);
  gilTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

  gilDisable(GL_FOG);
}


VOID CSLIDERender::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDELightPath *pCSLIDELightPath;
  CSLIDEAreaLightPath *pCSLIDEAreaLightPath;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s ", 
	  gcppcSLIDEReservedWords[sliderwtRender], 
	  m_pcViewportID);
  m_pCSLIDECameraPath->WriteSLIDE(fpOutputFile);
  fprintf(fpOutputFile, " %s\n", m_pcWorldRootID);

  if ( m_dyniLOD.GetValue() != slidelodtSLF_FULL )
    {
      fprintf(fpOutputFile, "  %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtLOD], 
	      gcppcSLIDELODReservedWords[m_dyniLOD.GetValue()]);
    }

  iter.Init(&m_listCSLIDELightPath);
  for ( pCSLIDELightPath = (CSLIDELightPath*)iter.PeekFirst();
	pCSLIDELightPath != NULL;
	pCSLIDELightPath = (CSLIDELightPath*)iter.PeekNext() )
    {
      fprintf(fpOutputFile, "  %s ",
	      gcppcSLIDEReservedWords[sliderwtLight]); 
      pCSLIDELightPath->WriteSLIDE(fpOutputFile);
      fprintf(fpOutputFile, "\n");
    }
  
  iter.Init(&m_listCSLIDEAreaLightPath);
  for ( pCSLIDEAreaLightPath = (CSLIDEAreaLightPath*)iter.PeekFirst();
	pCSLIDEAreaLightPath != NULL;
	pCSLIDEAreaLightPath = (CSLIDEAreaLightPath*)iter.PeekNext() )
    {
      fprintf(fpOutputFile, "  %s ",
	      gcppcSLIDEReservedWords[sliderwtAreaLight]); 
      pCSLIDEAreaLightPath->WriteSLIDE(fpOutputFile);
      fprintf(fpOutputFile, "\n");
    }
  
  CSLIDEPrimitive::WriteSLIDEFlags(fpOutputFile, 0);

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndRender]);
}

VOID CSLIDERender::WriteRIB(FILE *fpOutputFile, 
			    const Vector &vVPOrigin, const Vector &vVPSize)
{
  Matrix mGvp_proj, mGvrc_obj, mGobj_vrc, mGworld_light;
  Vector vOrigin, vSize, vFrustum, vRatios;

  if ( m_dyniLOD.GetValue() == slideoftSLF_OFF )
    {
      return;
    }

    if ( m_uiRIBFrame == 0 )
      {
	sprintf(gpcRIBOutputFile, "%s.tif", m_pcRIBOutputFileBase);
      }
    else
      {
	sprintf(gpcRIBOutputFile, "%s_%ld.tif", m_pcRIBOutputFileBase, m_uiRIBFrame);
      }
    RiDisplay(gpcRIBOutputFile, RI_FRAMEBUFFER, RI_RGB, RI_NULL);

    RiFrameBegin(m_uiRIBFrame);
    m_uiRIBFrame++;
    {

      // Set up the Viewport mapping matrix
      vFrustum = m_pCSLIDECameraPath->GetCamera()->GetFrustumMax() -
	m_pCSLIDECameraPath->GetCamera()->GetFrustumMin();

      ASSERT( vVPSize(X) != 0 && vVPSize(Y) != 0 );
      
      vRatios.Set(vFrustum(X)/vVPSize(X), vFrustum(Y)/vVPSize(Y), 0.0);
      vRatios /= MAX(vRatios[X], vRatios[Y]);
      vSize.Set(vVPSize(X) * vRatios(X), vVPSize(Y) * vRatios(Y), 0.0);
      vOrigin = vVPOrigin + ((vVPSize - vSize) / 2.0);
      
      int iWidth, iHeight;

      iWidth = (int) vSize(X);
      iHeight = (int) vSize(Y);

      RiFormat(iWidth, iHeight, 1.0);
      /*
	RiCropWindow(vOrigin[X], vOrigin[X] + vSize[X],
	vOrigin[Y], vOrigin[Y] + vSize[Y]);
      */
      
      // The mGproj_vrc is the Camera's projection matrix
      m_pCSLIDECameraPath->GetCamera()->WriteRIB(fpOutputFile);
      
      // These lines look backward but they are not
      // we need the inverse of the camera path
      mGvrc_obj = m_pCSLIDECameraPath->GetRootToObject();
      mGobj_vrc = m_pCSLIDECameraPath->GetObjectToRoot();
      
      // Apply the Gvrc_world matrix
      RiIdentity();
      RiScale(1.0, 1.0, -1.0);
      WriteRIBMultMatrix(fpOutputFile, mGvrc_obj);
      
      RiWorldBegin();
      {
	// Deal with the fog
	if ( m_dyniFogOnOff.GetValue() == slideoftSLF_ON )
	  {
	    ASSERT( m_pCSLIDEFog != NULL );
	    m_pCSLIDEFog->WriteRIB(fpOutputFile);
	  }
	
	// Deal with the lights
	if ( m_listCSLIDELightPath.GetLength() +  m_listCSLIDEAreaLightPath.GetLength() != 0 )
	  {
	    CListIterInt iter;
	    CSLIDELightPath *pCSLIDELightPath;
	    CSLIDELight *pCSLIDELight;
	    
	    iter.Init(&m_listCSLIDELightPath);
	    for ( pCSLIDELightPath = (CSLIDELightPath*)iter.PeekFirst();
		  pCSLIDELightPath != NULL;
		  pCSLIDELightPath = (CSLIDELightPath*)iter.PeekNext() )
	      {
		if ( pCSLIDELightPath->GetOnOff() == slideoftSLF_ON )
		  {
		    // We need to make the Gvrc_light matrix to be part
		    // of the light path not the light, because there can be
		    // multiple instances of a single light
		    
		    pCSLIDELight = pCSLIDELightPath->GetLight();
		    
		    mGworld_light = pCSLIDELightPath->GetObjectToRoot();
		    pCSLIDELight->WriteRIB(fpOutputFile, mGworld_light);
		  }		
	      }
	    
	    CSLIDEAreaLightPath *pCSLIDEAreaLightPath;
	    CSLIDEAreaLight *pCSLIDEAreaLight;
	    
	    iter.Init(&m_listCSLIDEAreaLightPath);
	    for ( pCSLIDEAreaLightPath = (CSLIDEAreaLightPath*)iter.PeekFirst();
		  pCSLIDEAreaLightPath != NULL;
		  pCSLIDEAreaLightPath = (CSLIDEAreaLightPath*)iter.PeekNext() )
	      {
		if ( pCSLIDEAreaLightPath->GetOnOff() == slideoftSLF_ON )
		  {
		    // We need to make the Gvrc_light matrix to be part
		    // of the light path not the light, because there can be
		    // multiple instances of a single light
		    
		    pCSLIDEAreaLight = pCSLIDEAreaLightPath->GetAreaLight();
		    
		    mGworld_light = pCSLIDEAreaLightPath->GetObjectToRoot();
		    pCSLIDEAreaLight->WriteRIBAreaLight(fpOutputFile,
							mGworld_light,
							&m_boundClip, 
							slideshtSLF_INHERIT, 
							(SLIDELODType)m_dyniLOD.GetValue());
		  }		
	      }
	  }
	else
	  {
	    // Possibly Disable Lighting somehow
	  }
	
/*	
	RiTransformBegin();
	{
	  // Do we want the crystal ball matrix going out to RenderMan?
	  if ( GetCrystalBallEnabled() )
	    {
	      WriteRIBMultMatrix(fpOutputFile, m_mBall);
	    }
*/
      
	Matrix mIdentity;

	CSLIDESurface::GetSLF_INHERIT()->WriteRIB(fpOutputFile);
	m_pWorldRoot->WriteRIB(fpOutputFile,
			       //m_pCSLIDECameraPath->GetCamera()->GetVRCToProjection() * mGvrc_obj,
			       mIdentity,
			       &m_boundClip, 
			       slideshtSLF_INHERIT, 
			       (SLIDELODType)m_dyniLOD.GetValue());
/*
        }
        RiTransformEnd();
*/
      }
      RiWorldEnd();
    }
    RiFrameEnd();
}

VOID CSLIDERender::WritePOV(FILE *fpOutputFile, 
			    CListInt *plistTransformIDs, 
			    const Vector &vVPSize, 
			    CListInt *plistColors)  
{
  ASSERT( fpOutputFile != NULL );

  Matrix mGvp_proj, mGvrc_obj, mGobj_vrc, mGworld_light;
  Vector vOrigin, vSize, vFrustum, vRatios;

  // These lines look backward but they are not
  // we need the inverse of the camera path
  mGvrc_obj = m_pCSLIDECameraPath->GetRootToObject();
  mGobj_vrc = m_pCSLIDECameraPath->GetObjectToRoot();

  m_pCSLIDECameraPath->GetCamera()->WritePOV(fpOutputFile, mGobj_vrc, vVPSize);

  // Deal with the fog
  if ( m_dyniFogOnOff.GetValue() == slideoftSLF_ON )
    {
      ASSERT( m_pCSLIDEFog != NULL );
      m_pCSLIDEFog->WritePOV(fpOutputFile);
    }
  
  // Deal with the lights
  if ( m_listCSLIDELightPath.GetLength() +  m_listCSLIDEAreaLightPath.GetLength() != 0 )
    {
      CListIterInt iter;
      CSLIDELightPath *pCSLIDELightPath;
      CSLIDELight *pCSLIDELight;
      
      iter.Init(&m_listCSLIDELightPath);
      for ( pCSLIDELightPath = (CSLIDELightPath*)iter.PeekFirst();
	    pCSLIDELightPath != NULL;
	    pCSLIDELightPath = (CSLIDELightPath*)iter.PeekNext() )
	{
	  if ( pCSLIDELightPath->GetOnOff() == slideoftSLF_ON )
	    {
	      // We need to make the Gvrc_light matrix to be part
	      // of the light path not the light, because there can be
	      // multiple instances of a single light
	      
	      pCSLIDELight = pCSLIDELightPath->GetLight();
	      
	      if (pCSLIDELightPath->GetInstanceLength() == 0) {
		mGworld_light = pCSLIDELightPath->GetObjectToRoot();
		pCSLIDELight->WritePOVLight(fpOutputFile);
	      }
	    }		
	}
      
      CSLIDEAreaLightPath *pCSLIDEAreaLightPath;
      CSLIDEAreaLight *pCSLIDEAreaLight;
	    
      iter.Init(&m_listCSLIDEAreaLightPath);
      for ( pCSLIDEAreaLightPath = (CSLIDEAreaLightPath*)iter.PeekFirst();
	    pCSLIDEAreaLightPath != NULL;
	    pCSLIDEAreaLightPath = (CSLIDEAreaLightPath*)iter.PeekNext() )
	{
	  if ( pCSLIDEAreaLightPath->GetOnOff() == slideoftSLF_ON )
	    {
	      // We need to make the Gvrc_light matrix to be part
	      // of the light path not the light, because there can be
	      // multiple instances of a single light
	      
	      pCSLIDEAreaLight = pCSLIDEAreaLightPath->GetAreaLight();
	      
	      mGworld_light = pCSLIDEAreaLightPath->GetObjectToRoot();
	      pCSLIDEAreaLight->WritePOVAreaLight(fpOutputFile, 
						  plistTransformIDs,
						  plistColors);
	    }
	}
    }
  else
    {
      // Possibly Disable Lighting somehow
    }

  Matrix mIdentity;
  
  fprintf(fpOutputFile, "#declare color_SLF_INHERIT = ");
  CSLIDESurface::GetSLF_INHERIT()->WritePOV(fpOutputFile);
  fprintf(fpOutputFile, "; \n");
  m_pWorldRoot->WritePOV(fpOutputFile, plistTransformIDs, plistColors);
}

VOID CSLIDERender::WriteWRLZ(FILE *fpOutputFile) 
{
  ASSERT( fpOutputFile != NULL );

  /*
  Matrix mGvp_proj, mGvrc_obj, mGobj_vrc, mGworld_light;
  Vector vOrigin, vSize, vFrustum, vRatios;

  // These lines look backward but they are not
  // we need the inverse of the camera path
  mGvrc_obj = m_pCSLIDECameraPath->GetRootToObject();
  mGobj_vrc = m_pCSLIDECameraPath->GetObjectToRoot();

  m_pCSLIDECameraPath->GetCamera()->WriteWRLZ(fpOutputFile, mGobj_vrc, vVPSize);

  // Deal with the fog
  if ( m_dyniFogOnOff.GetValue() == slideoftSLF_ON )
    {
      ASSERT( m_pCSLIDEFog != NULL );
      m_pCSLIDEFog->WriteWRLZ(fpOutputFile);
    }
  
  // Deal with the lights
  if ( m_listCSLIDELightPath.GetLength() +  m_listCSLIDEAreaLightPath.GetLength() != 0 )
    {
      CListIterInt iter;
      CSLIDELightPath *pCSLIDELightPath;
      CSLIDELight *pCSLIDELight;
      
      iter.Init(&m_listCSLIDELightPath);
      for ( pCSLIDELightPath = (CSLIDELightPath*)iter.PeekFirst();
	    pCSLIDELightPath != NULL;
	    pCSLIDELightPath = (CSLIDELightPath*)iter.PeekNext() )
	{
	  if ( pCSLIDELightPath->GetOnOff() == slideoftSLF_ON )
	    {
	      // We need to make the Gvrc_light matrix to be part
	      // of the light path not the light, because there can be
	      // multiple instances of a single light
	      
	      pCSLIDELight = pCSLIDELightPath->GetLight();
	      
	      if (pCSLIDELightPath->GetInstanceLength() == 0) {
		mGworld_light = pCSLIDELightPath->GetObjectToRoot();
		pCSLIDELight->WriteWRLZLight(fpOutputFile);
	      }
	    }		
	}
      
      CSLIDEAreaLightPath *pCSLIDEAreaLightPath;
      CSLIDEAreaLight *pCSLIDEAreaLight;
	    
      iter.Init(&m_listCSLIDEAreaLightPath);
      for ( pCSLIDEAreaLightPath = (CSLIDEAreaLightPath*)iter.PeekFirst();
	    pCSLIDEAreaLightPath != NULL;
	    pCSLIDEAreaLightPath = (CSLIDEAreaLightPath*)iter.PeekNext() )
	{
	  if ( pCSLIDEAreaLightPath->GetOnOff() == slideoftSLF_ON )
	    {
	      // We need to make the Gvrc_light matrix to be part
	      // of the light path not the light, because there can be
	      // multiple instances of a single light
	      
	      pCSLIDEAreaLight = pCSLIDEAreaLightPath->GetAreaLight();
	      
	      mGworld_light = pCSLIDEAreaLightPath->GetObjectToRoot();
	      pCSLIDEAreaLight->WriteWRLZAreaLight(fpOutputFile, plistTransformIDs); 
	    }
	}
    }
  else
    {
      // Possibly Disable Lighting somehow
    }
    */
  
  Matrix m;
  m.SetIdentity();

  m_pWorldRoot->CalculateLeafQEMesh();
  m_pWorldRoot->WriteWRLZ(fpOutputFile, m, CSLIDESurface::GetSLF_INHERIT());
  m_pWorldRoot->DeallocateLeafQEMesh();
}

VOID CSLIDERender::WriteSTL(FILE *fpOutputFile) 
{
  ASSERT( fpOutputFile != NULL );
  Matrix m;

  m.SetIdentity();

  m_pWorldRoot->WriteSTL(fpOutputFile, m);
}

VOID CSLIDERender::WriteSIF(FILE *fpOutputFile) 
{
  ASSERT( fpOutputFile != NULL );

  CObjectList<QEMesh> listQEMesh;
  QEMesh *pme;
  Matrix m;
  UINT_32 uiIndentLevel;


  listQEMesh.Init();

  // Remove the old SIF Shell Components
  if ( m_pmeBoundaryEdges != NULL )
    {
      DeallocateQEMesh(m_pmeBoundaryEdges);
      m_pmeBoundaryEdges = NULL;
    }

  m_pmeBoundaryEdges = AllocateQEMesh();
  ASSERT( m_pmeBoundaryEdges != NULL );
  m_pmeBoundaryEdges->Init();

  uiIndentLevel = 0;
  m.SetIdentity();

  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "(%s 2 0\n", gcppcSIFReservedWords[sifrwtSIF_SFF]);
  uiIndentLevel++;
  
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "(%s %s)\n", gcppcSIFReservedWords[sifrwtUnits], 
	  gcppcSIFReservedWords[sifrwtInches]);
  
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "(%s\n", gcppcSIFReservedWords[sifrwtSolid]);
  uiIndentLevel++;
  
  pme = m_pWorldRoot->WriteSIF(fpOutputFile, uiIndentLevel, m, m, m_pmeBoundaryEdges);

  if ( pme != NULL )
    {
      m_pmeBoundaryEdges->CopyBoundaryEdges(pme);
      pme->ExtractComponents(&listQEMesh);
      DeallocateQEMesh(pme);

      for ( pme = listQEMesh.RemoveFirst();
	    pme != NULL;
	    pme = listQEMesh.RemoveFirst() )
	{
	  ::WriteSIFShell(pme, fpOutputFile, uiIndentLevel);
	  DeallocateQEMesh(pme);
	}
    }

  uiIndentLevel--;
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, ")\n");
  
  uiIndentLevel--;
  ::WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, ")\n");

  if ( m_pmeBoundaryEdges->GetBoundaryEdgeCount() )
    {
      QEMeshIter iterQEMesh;
      QEEdge *pe;

      fprintf(stderr, "ERROR: WriteSIF: Not all shells are closed: Total unmatched edges = %ld\n", m_pmeBoundaryEdges->GetBoundaryEdgeCount()/2);

      iterQEMesh.Init(m_pmeBoundaryEdges);
      printf("START: Unmatched edges:\n");
      for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
	    pe != NULL;
	    pe = iterQEMesh.PeekNextEdgeBoundary() )
	{
	  printf("  E %p: V%ld < %f, %f, %f > ==> V%ld < %f, %f, %f >\n", 
		 pe, pe->GetVertex()->GetID(), pe->GetVertex()->GetPoint()(X), pe->GetVertex()->GetPoint()(Y), pe->GetVertex()->GetPoint()(Z), 
		 pe->GetEdge()->GetVertex()->GetID(), pe->GetEdge()->GetVertex()->GetPoint()(X), pe->GetEdge()->GetVertex()->GetPoint()(Y), pe->GetEdge()->GetVertex()->GetPoint()(Z));
	}
      printf("END  : Unmatched edges:\n");

      printf ("\nWriteSIF used Merge Epsilon: %f\n", QEMesh::GetMergeEpsilon());
    }
  else
    {
      printf ("\nWriteSIF used Merge Epsilon: %f\n", QEMesh::GetMergeEpsilon());
      printf ("WriteSIF was successful: all shells are closed.\n");
    }
}

VOID CSLIDERender::WriteLSIF(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  m_pWorldRoot->WriteLSIF(fpOutputFile, 0);
}

VOID CSLIDERender::WriteDotID(FILE *fpOutputFile)
{
  static UINT_32 suiDotID = 0;

  ASSERT( fpOutputFile != NULL );

  if ( m_uiDotID == 0 )
    {
      suiDotID++;
      m_uiDotID = suiDotID;
    }

  fprintf(fpOutputFile, "render_%ld", m_uiDotID);
}

VOID CSLIDERender::WriteDotNode(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "  ");
  WriteDotID(fpOutputFile);
  if ( GetTclTag() != NULL )
    {
      fprintf(fpOutputFile, " [ label = \"%s\", shape = box ];\n", 
	      GetTclTag());
    }
  else
    {
      fprintf(fpOutputFile, " [ label = \"Render %ld\", shape = box ];\n", 
	      m_uiDotID);
    }
}

VOID CSLIDERender::WriteDotEdges(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDELightPath *pCSLIDELightPath;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "  ");
  WriteDotID(fpOutputFile);
  fprintf(fpOutputFile, " -> ");
  m_pWorldRoot->WriteDotID(fpOutputFile);
  fprintf(fpOutputFile, "\n");

  // Camera Path
  fprintf(fpOutputFile, "  ");
  WriteDotID(fpOutputFile);
  m_pCSLIDECameraPath->WriteDotEdges(fpOutputFile);

  // Light Paths
  iter.Init(&m_listCSLIDELightPath);
  for ( pCSLIDELightPath = (CSLIDELightPath*)iter.PeekFirst();
	pCSLIDELightPath != NULL;
	pCSLIDELightPath = (CSLIDELightPath*)iter.PeekNext() )
    {
      fprintf(fpOutputFile, "  ");
      WriteDotID(fpOutputFile);
      pCSLIDELightPath->WriteDotEdges(fpOutputFile);
    }
  
}

//////////////////////////////////////////////////////////////////////
// CSLIDETclVariables Class
//

CSLIDETclVariables::CSLIDETclVariables()
{
  m_pcID = NULL;

  m_fWidth = 1.0;
  m_fHeight = 1.0;
  m_fMouseX = 0.0;
  m_fMouseY = 0.0;
  m_fScale = 1.0;
}

CSLIDETclVariables::~CSLIDETclVariables()
{
  Uninit();
}

VOID CSLIDETclVariables::Init(CHAR *pcID)
{
  Vector v;
  Quater q;

  m_pcID = pcID;

  SetWidth(1.0);
  SetHeight(1.0);
  SetMouseX(0.0);
  SetMouseY(0.0);

  v.Set(0.0, 0.0, 1.0);
  q.SetRotation(v, 0.0);

  SetRotate(q);
  SetScale(1.0);
}

VOID CSLIDETclVariables::Uninit()
{
}

BOOL CSLIDETclVariables::Initialized()
{
  return TRUE;
}

BOOL CSLIDETclVariables::Valid()
{
  return TRUE;
}

VOID CSLIDETclVariables::SetWidth(FLOAT fWidth)
{
  m_fWidth = fWidth;

  if ( m_pcID == NULL )
    {
      sprintf(gpcBufferVariable, "SLF_WIDTH");
    }
  else
    {
      sprintf(gpcBufferVariable, "SLF_%s_WIDTH", m_pcID);
    }
  sprintf(gpcBufferValue, "%f", m_fWidth);
  Tcl_SetVar(CSLIDEWorld::GetInterp(), gpcBufferVariable, 
	     gpcBufferValue, TCL_GLOBAL_ONLY);
}

VOID CSLIDETclVariables::SetHeight(FLOAT fHeight)
{
  m_fHeight = fHeight;

  if ( m_pcID == NULL )
    {
      sprintf(gpcBufferVariable, "SLF_HEIGHT");
    }
  else
    {
      sprintf(gpcBufferVariable, "SLF_%s_HEIGHT", m_pcID);
    }
  sprintf(gpcBufferValue, "%f", m_fHeight);
  Tcl_SetVar(CSLIDEWorld::GetInterp(), gpcBufferVariable, 
	     gpcBufferValue, TCL_GLOBAL_ONLY);
}

VOID CSLIDETclVariables::SetMouseX(FLOAT fMouseX)
{
  m_fMouseX = fMouseX;

  if ( m_pcID == NULL )
    {
      sprintf(gpcBufferVariable, "SLF_MOUSE_X");
    }
  else
    {
      sprintf(gpcBufferVariable, "SLF_%s_MOUSE_X", m_pcID);
    }
  sprintf(gpcBufferValue, "%f", m_fMouseX);
  Tcl_SetVar(CSLIDEWorld::GetInterp(), gpcBufferVariable, 
	     gpcBufferValue, TCL_GLOBAL_ONLY);
}

VOID CSLIDETclVariables::SetMouseY(FLOAT fMouseY)
{
  m_fMouseY = fMouseY;

  if ( m_pcID == NULL )
    {
      sprintf(gpcBufferVariable, "SLF_MOUSE_Y");
    }
  else
    {
      sprintf(gpcBufferVariable, "SLF_%s_MOUSE_Y", m_pcID);
    }
  sprintf(gpcBufferValue, "%f", m_fMouseY);
  Tcl_SetVar(CSLIDEWorld::GetInterp(), gpcBufferVariable, 
	     gpcBufferValue, TCL_GLOBAL_ONLY);
}

VOID CSLIDETclVariables::SetRotate(const Quater &qRotate)
{
  Vector v;
  FLOAT fTheta;

  m_qRotate = qRotate;

  m_qRotate.GetRotation(v, fTheta);

  if ( m_pcID == NULL )
    {
      sprintf(gpcBufferVariable, "SLF_AXIS_X");
    }
  else
    {
      sprintf(gpcBufferVariable, "SLF_%s_AXIS_X", m_pcID);
    }
  sprintf(gpcBufferValue, "%f", v[X]);
  Tcl_SetVar(CSLIDEWorld::GetInterp(), gpcBufferVariable, 
	     gpcBufferValue, TCL_GLOBAL_ONLY);

  if ( m_pcID == NULL )
    {
      sprintf(gpcBufferVariable, "SLF_AXIS_Y");
    }
  else
    {
      sprintf(gpcBufferVariable, "SLF_%s_AXIS_Y", m_pcID);
    }
  sprintf(gpcBufferValue, "%f", v[Y]);
  Tcl_SetVar(CSLIDEWorld::GetInterp(), gpcBufferVariable, 
	     gpcBufferValue, TCL_GLOBAL_ONLY);

  if ( m_pcID == NULL )
    {
      sprintf(gpcBufferVariable, "SLF_AXIS_Z");
    }
  else
    {
      sprintf(gpcBufferVariable, "SLF_%s_AXIS_Z", m_pcID);
    }
  sprintf(gpcBufferValue, "%f", v[Z]);
  Tcl_SetVar(CSLIDEWorld::GetInterp(), gpcBufferVariable, 
	     gpcBufferValue, TCL_GLOBAL_ONLY);

  if ( m_pcID == NULL )
    {
      sprintf(gpcBufferVariable, "SLF_ANGLE");
    }
  else
    {
      sprintf(gpcBufferVariable, "SLF_%s_ANGLE", m_pcID);
    }
  sprintf(gpcBufferValue, "%f", RADTODEG(fTheta));
  Tcl_SetVar(CSLIDEWorld::GetInterp(), gpcBufferVariable, 
	     gpcBufferValue, TCL_GLOBAL_ONLY);
}

VOID CSLIDETclVariables::SetScale(FLOAT fScale)
{
  m_fScale = fScale;

  if ( m_pcID == NULL )
    {
      sprintf(gpcBufferVariable, "SLF_SCALE");
    }
  else
    {
      sprintf(gpcBufferVariable, "SLF_%s_SCALE", m_pcID);
    }
  sprintf(gpcBufferValue, "%f", m_fScale);
  Tcl_SetVar(CSLIDEWorld::GetInterp(), gpcBufferVariable, 
	     gpcBufferValue, TCL_GLOBAL_ONLY);
}

VOID CSLIDERender::SetBlend(DynInt *pdyniOnOff)
{
  if ( pdyniOnOff != NULL )
    {
      CSLIDERender::m_dyniBlendOnOff.Set(pdyniOnOff);
    }
  else
    {
      CSLIDERender::m_dyniBlendOnOff.SetValue(slideoftSLF_ON);
    }
}

DynInt CSLIDERender::GetBlend()
{
  return m_dyniBlendOnOff;
}

