#ifndef __SLIDERENDER_H__
#define __SLIDERENDER_H__

#include "slideprimitive.h"
#include "timer.h"
#include "slidecamera.h"

class CSLIDEWorld;
class CSLIDECameraPath;
class CSLIDELightPath;
class CSLIDEAreaLightPath;
class CSLIDENode;
class CSLIDEFog;

class CSLIDETclVariables;

class QEMesh;

//////////////////////////////////////////////////////////////////////
// CSLIDERender Class
//

class CSLIDERender : public CSLIDEPrimitive
{
//////////////////////////////////////////////////
// Friends
//
public: friend class CSLIDEWorld;

//////////////////////////////////////////////////
// Enumerated Types
//
/*
  static enum m_eSpecialKeys
  {
	CSR_CTRL_L = 0x01,
	CSR_CTRL_R = 0x02,
	CSRSpecialKeysSize = 2
  };
*/
#define CSR_CTRL_L 0x01
#define CSR_CTRL_R 0x02
#define CSR_SHIFT_L 0x04
#define CSR_SHIFT_R 0x08
#define CSRSpecialKeysSize 4

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDERender();
  virtual ~CSLIDERender();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcViewportID, CSLIDECameraPath *pCSLIDECameraPath,
		    CHAR *pcWorldRootID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  virtual inline VOID SetMergeEpsilon(DynFloat *pdynfMergeEpsilon);
  virtual inline VOID SetTclTag(CHAR *pcTclTag);
  virtual inline const CHAR *GetTclTag() const;
  virtual inline VOID SetLOD(DynInt *pdyniLOD);
  virtual inline VOID SetCUIType(DynInt *pdyniCUIType);
  virtual inline VOID SetValueCUIType(UINT_32 iCUIType);
  virtual inline VOID SetCUISpeed(DynFloat *pdynfCUISpeed);
  virtual inline VOID SetStencil(DynInt *pdyniStencil);
  virtual inline VOID SetFogID(CHAR *pcFogID, DynInt *pdyniOnOff);
  static VOID SetBlend(DynInt *pdyniOnOff);
  virtual inline VOID AddLightPath(CSLIDELightPath *pCSLIDELightPath);
  virtual inline VOID AddAreaLightPath(CSLIDEAreaLightPath *pCSLIDEAreaLightPath);

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

public:  

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual VOID PreprocessGraph();
  virtual VOID UpdateGraph();

protected:
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();

//////////////////////////////////////////////////
// Input Functions
//
public:
  VOID HandleReshape(const Vector &vSize);
  BOOL HandleMouseButton(const Vector &vSize,
			 INT_32 iButton, INT_32 iState, 
			 const Vector &vClick);
  BOOL HandleMouseMotion(const Vector &vSize,
			 const Vector &vClick);
  BOOL HandleKeyboard(INT_32 iKey, INT_32 iState);
  BOOL HandleMouseMotionDirectManipulation(Vector vSize, Vector vClick);

  int RenderSelection(GLint *viewport, GLuint *selectBuf, const Vector &vClick,
		      BOOL bRenderAxes);
  VOID Selection(const Vector &vClick);
  VOID SelectionManipulation(const Vector &vClick);
  VOID SelectionCycle();
  VOID SelectionAxis(GLuint uiChild, GLuint uiAxis);

  VOID ResetCamera();
  VOID EditSelectedObject();

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID Prerender();
  virtual VOID Render(const Vector &vOrigin, const Vector &vSize);
  static VOID ApplyStencil();

protected:
  VOID ResetGL();
  VOID CreateStencil(const Vector &vOrigin, const Vector &vSize);
// DSC Begin
  INT_32 GetMouseUIType( INT_32 iUIType, INT_32 iButton );
  INT_32 GetUpdateMouseUIType( INT_32 iUIType, INT_32 iButton );
  VOID UpdateCameraUI( );
  BOOL ApplyCUI( INT_32 iMUIType, const Vector &vCurr, 
	  const FLOAT &fDeltaTime = 1.0f,
	  const FLOAT &fTMultiplier = 1.0f,
	  const FLOAT &fRMultiplier = 1.0f,
	  const FLOAT &fSMultiplier = 1.0f);
// DSC End


//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);
  virtual VOID WriteRIB(FILE *fpOutputFile, 
			const Vector &vOrigin, const Vector &vSize);
  virtual VOID WritePOV(FILE *fpOutputFile, 
			CListInt *plistTransformIDs, 
			const Vector &vSize,
			CListInt *plistColors); 
  virtual VOID WriteWRLZ(FILE *fpOutputFile);
  virtual VOID WriteSTL(FILE *fpOutputFile);
  virtual VOID WriteSIF(FILE *fpOutputFile);
  virtual VOID WriteLSIF(FILE *fpOutputFile);
  virtual VOID WriteDotID(FILE *fpOutputFile);
  virtual VOID WriteDotNode(FILE *fpOutputFile);
  virtual VOID WriteDotEdges(FILE *fpOutputFile);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  inline CHAR *GetWorldRootID();
  inline CHAR *GetCameraPathRootID();
  inline CHAR *GetFogID();

  inline CSLIDENode *GetNodeGUISelected();
  inline VOID SetNodeGUISelected(CSLIDENode *pCSLIDENodeGUISelected);

  static DynInt GetBlend();

//////////////////////////////////////////////////
// Class Variables
//
protected:
  static UINT_32 m_uiCount;
  static Vector m_vProjMin;
  static Vector m_vProjMax;
  static CBound m_boundClip;

  static GLuint m_GLuiStencil;
  static BOOL m_bUseStencil;

  static CListInt *m_plistCSLIDETclVariables;

  static INT_32 m_ppiMUITTable[slidecuitSize][3];
  const static INT_32 m_ppiKeyConversion[CSRSpecialKeysSize][2];

public:
  // RIB Output Stuff
  static CHAR m_pcRIBOutputFileBase[500];
  static UINT_32 m_uiRIBFrame;

  // STL Output
  static CHAR m_pcSTLOutputFileBase[500];

  static DynInt m_dyniBlendOnOff;

//////////////////////////////////////////////////
// Member Variables
//
protected:
  DynFloat m_dynfMergeEpsilon;
  CHAR *m_pcTclTag;
  CSLIDETclVariables *m_pCSLIDETclVariables;

  CHAR *m_pcViewportID;

  CSLIDECameraPath *m_pCSLIDECameraPath;

  CHAR *m_pcWorldRootID;
  CSLIDENode *m_pWorldRoot;

  DynInt m_dyniLOD;
  DynInt m_dyniStencil;

  CHAR *m_pcFogID;
  CSLIDEFog *m_pCSLIDEFog;
  DynInt m_dyniFogOnOff;

  CListInt m_listCSLIDELightPath;
  CListInt m_listCSLIDEAreaLightPath;

  // SIF Variables
  QEMesh *m_pmeBoundaryEdges;

  // Dot Output Stuff
  UINT_32 m_uiDotID;

  // Crystal Ball Variables
  DynInt m_dyniCUIType;

  Matrix m_mBall;
  Matrix m_mBallInverse;
  Matrix m_mCameraUI;
  Matrix m_mCameraUIInverse;

  INT_32 m_iButtonCurr;
  INT_32 m_iButtonsCurr;
  INT_32 m_iKeysCurr;

  Vector m_vStart;
  Vector m_vLatest;
  FLOAT m_fScaleLast;

  DynInt m_dyniCamZoom;

  // Camera speed
  DynFloat m_dynfCUISpeed;
  TimeValue m_tvLastRender;

  // Direct Manipulation
  CSLIDENode *m_pCSLIDENodeGUISelected;
  GLuint m_puiSelectionNameList[500];
  UINT_32 m_puiSelectionHitList[100];
  int m_iSelected;
  int m_iSelectionHits;
  BOOL m_bManipulationMode;
  Matrix m_mChildFrame;
  Matrix m_mParentFrame;
  MATRIX4D m_mProjection;   
  MATRIX4D m_mModelview;
  BOOL m_bChildAxesSelected;
  BOOL m_bXAxisSelected;
  BOOL m_bYAxisSelected;
  BOOL m_bZAxisSelected;
  BOOL m_bNewTransform;
  Vector m_vManipulationStart;

  BOOL m_bGraphSelected;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDERender::SetMergeEpsilon(DynFloat *pdynfMergeEpsilon)
{
  m_dynfMergeEpsilon.Set(pdynfMergeEpsilon);
}

inline VOID CSLIDERender::SetTclTag(CHAR *pcTclTag)
{
  m_pcTclTag = pcTclTag;
}

inline const CHAR *CSLIDERender::GetTclTag() const
{
  return m_pcTclTag;
}

inline VOID CSLIDERender::SetNodeGUISelected(CSLIDENode *pCSLIDENodeGUISelected)
{
  m_pCSLIDENodeGUISelected = pCSLIDENodeGUISelected;
  m_bGraphSelected = (pCSLIDENodeGUISelected != NULL);
  m_iSelected = -1;
}

inline CSLIDENode *CSLIDERender::GetNodeGUISelected()
{
  return m_pCSLIDENodeGUISelected;
}

inline VOID CSLIDERender::SetLOD(DynInt *pdyniLOD)
{
  m_dyniLOD.Set(pdyniLOD);
}

inline VOID CSLIDERender::SetCUIType(DynInt *pdyniCUIType)
{
  m_dyniCUIType.Set(pdyniCUIType);
}
inline VOID CSLIDERender::SetValueCUIType(UINT_32 iCUIType)
{
  m_dyniCUIType.SetValue(iCUIType);
}

inline VOID CSLIDERender::SetCUISpeed(DynFloat *pdynfCUISpeed)
{
  m_dynfCUISpeed.Set(pdynfCUISpeed);
}

inline VOID CSLIDERender::SetStencil(DynInt *pdyniStencil)
{
  m_dyniStencil.Set(pdyniStencil);
}

inline VOID CSLIDERender::SetFogID(CHAR *pcFogID, DynInt *pdyniOnOff)
{
  m_pcFogID = pcFogID;
  if ( pdyniOnOff != NULL )
    {
      m_dyniFogOnOff.Set(pdyniOnOff);
    }
  else
    {
      m_dyniFogOnOff.SetValue(slideoftSLF_ON);
    }
}

inline VOID CSLIDERender::AddLightPath(CSLIDELightPath *pCSLIDELightPath)
{
  ASSERT( pCSLIDELightPath != NULL );
  m_listCSLIDELightPath.InsertLast(pCSLIDELightPath, 0);
}

inline VOID CSLIDERender::AddAreaLightPath(CSLIDEAreaLightPath *pCSLIDEAreaLightPath)
{
  ASSERT( pCSLIDEAreaLightPath != NULL );
  m_listCSLIDEAreaLightPath.InsertLast(pCSLIDEAreaLightPath, 0);
}

inline CHAR *CSLIDERender::GetWorldRootID()
{
  return m_pcWorldRootID;
}

inline CHAR *CSLIDERender::GetCameraPathRootID()
{
  return m_pCSLIDECameraPath->GetRootID();
}

inline CHAR *CSLIDERender::GetFogID()
{
  return m_pcFogID;
}

//////////////////////////////////////////////////////////////////////
// CSLIDETclVariables Class
//

class CSLIDETclVariables
{
//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDETclVariables();
  virtual ~CSLIDETclVariables();
  virtual VOID Init(CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();
  virtual BOOL Valid();

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  inline CHAR *GetID() const;

  inline FLOAT GetWidth() const;
  VOID SetWidth(FLOAT fWidth);
  inline FLOAT GetHeight() const;
  VOID SetHeight(FLOAT fHeight);
  inline FLOAT GetMouseX() const;
  VOID SetMouseX(FLOAT fMouseX);
  inline FLOAT GetMouseY() const;
  VOID SetMouseY(FLOAT fMouseY);
  inline const Quater &GetRotate() const;
  VOID SetRotate(const Quater &q);
  inline FLOAT GetScale() const;
  VOID SetScale(FLOAT fScale);

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CHAR *m_pcID;

  // Window Variables
  FLOAT m_fWidth;
  FLOAT m_fHeight;
  FLOAT m_fMouseX;
  FLOAT m_fMouseY;

  // Crystal Ball Variables
  Quater m_qRotate;
  FLOAT m_fScale;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline CHAR *CSLIDETclVariables::GetID() const
{
  return m_pcID;
}

inline FLOAT CSLIDETclVariables::GetWidth() const
{
  return m_fWidth;
}

inline FLOAT CSLIDETclVariables::GetHeight() const
{
  return m_fHeight;
}

inline FLOAT CSLIDETclVariables::GetMouseX() const
{
  return m_fMouseX;
}

inline FLOAT CSLIDETclVariables::GetMouseY() const
{
  return m_fMouseY;
}

inline const Quater &CSLIDETclVariables::GetRotate() const
{
  return m_qRotate;
}

inline FLOAT CSLIDETclVariables::GetScale() const
{
  return m_fScale;
}

#endif // __SLIDERENDER_H__
