#include "sliderotate.h"

//////////////////////////////////////////////////////////////////////
// CSLIDERotate Class
//

CSLIDERotate::CSLIDERotate()
{
  m_vAxis.Set(0.0, 0.0, 1.0);
}

CSLIDERotate::~CSLIDERotate()
{
}

VOID CSLIDERotate::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			DynVector *pdynvAxis, DynFloat *pdynfTheta)
{
  CSLIDETransform::Init(pcFile, uiLine, uiColumn);

  ASSERT( pdynvAxis != NULL );
  ASSERT( pdynfTheta != NULL );

  m_vAxis.SetAll(pdynvAxis);
  m_dynfTheta.Set(pdynfTheta);

  m_m.SetRotate(m_vAxis, m_dynfTheta.GetValue());
  m_mInverse.SetRotate(m_vAxis, -m_dynfTheta.GetValue());
}

VOID CSLIDERotate::Uninit()
{
  CSLIDETransform::Uninit();
}

BOOL CSLIDERotate::Initialized()
{
  return CSLIDETransform::Initialized();
}

BOOL CSLIDERotate::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  if ( !CSLIDETransform::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  return TRUE;
}

BOOL CSLIDERotate::Valid()
{
  if ( !CSLIDETransform::Valid() )
    {
      return FALSE;
    }

  return TRUE;
}

VOID CSLIDERotate::CheckDynamic()
{
  SetDynamic( m_vAxis.GetDynamic() || m_dynfTheta.GetDynamic() );
}

VOID CSLIDERotate::UpdateState()
{
  m_vAxis.Update();
  if ( m_dynfTheta.GetDynamic() )
    {
      m_dynfTheta.Update();
      m_dynfTheta.SetValue(DEGTORAD(m_dynfTheta.GetValue()));
    }
  m_m.SetRotate(m_vAxis, m_dynfTheta.GetValue());
  m_mInverse.SetRotate(m_vAxis, -m_dynfTheta.GetValue());
}

VOID CSLIDERotate::WriteSLIDE(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "    %s ( %.3f %.3f %.3f ) ( %.3f )\n", 
	  gcppcSLIDEReservedWords[sliderwtRotate],
	  m_vAxis[X], m_vAxis[Y], m_vAxis[Z], 
	  RADTODEG(m_dynfTheta.GetValue()));
}

VOID CSLIDERotate::WriteRIB(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiRotate(RADTODEG(m_dynfTheta.GetValue()),
	   m_vAxis[X], m_vAxis[Y], m_vAxis[Z]);	  
}

VOID CSLIDERotate::WritePOV(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "  %s <%.3f %.3f %.3f>\n", 
	  gcppcSLIDEReservedWords[sliderwtRotate],
	  m_vAxis[X]* RADTODEG(m_dynfTheta.GetValue()),
	  m_vAxis[Y]* RADTODEG(m_dynfTheta.GetValue()), 
	  m_vAxis[Z]* RADTODEG(m_dynfTheta.GetValue()));
}

VOID CSLIDERotate::WriteDotNode(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "R");
}
