#ifndef __SLIDEROTATE_H__
#define __SLIDEROTATE_H__

#include "slidetransform.h"

//////////////////////////////////////////////////////////////////////
// CSLIDERotate Class
//

class CSLIDERotate : public CSLIDETransform
{
//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDERotate();
  virtual ~CSLIDERotate();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    DynVector *pdynvAxis, DynFloat *pdynfTheta);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual inline VOID ComposeWith(CSLIDETransform *pCSLIDETransform);
  virtual inline VOID Set(CSLIDETransform *pCSLIDETransform);
  virtual inline VOID Modify(CSLIDETransform *pCSLIDETransformLast, 
			     CSLIDETransform *pCSLIDETransformNew);

protected:
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();

//////////////////////////////////////////////////
// Rendering
//
public:

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);
  virtual VOID WriteRIB(FILE *fpOutputFile);
  virtual VOID WritePOV(FILE *fpOutputFile);
  virtual VOID WriteDotNode(FILE *fpOutputFile);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline CHAR *GetTypeString();
  virtual inline SLIDEReservedWordType GetType();

  virtual inline DynVector GetAxis();
  virtual inline DynFloat GetTheta();

//////////////////////////////////////////////////
// Class Variables
//
public:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  DynVector m_vAxis;
  DynFloat m_dynfTheta;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline CHAR *CSLIDERotate::GetTypeString()
{
  return gcppcSLIDEReservedWords[sliderwtRotate];
}

inline SLIDEReservedWordType CSLIDERotate::GetType()
{
  return sliderwtRotate;
}

inline DynVector CSLIDERotate::GetAxis()
{
  return m_vAxis;
}

inline DynFloat CSLIDERotate::GetTheta()
{
  return m_dynfTheta;
}

inline VOID CSLIDERotate::ComposeWith(CSLIDETransform *pCSLIDETransform)
{
  // Assumes rotation axes are the same
  m_dynfTheta.SetValue((m_dynfTheta.GetValue() + 
			((CSLIDERotate*)pCSLIDETransform)->GetTheta().GetValue()));

  m_m.SetRotate(m_vAxis, m_dynfTheta.GetValue());
  m_mInverse.SetRotate(m_vAxis, -m_dynfTheta.GetValue());
}

inline VOID CSLIDERotate::Set(CSLIDETransform *pCSLIDETransform)
{
  m_vAxis[X] = ((CSLIDERotate*)pCSLIDETransform)->GetAxis()[X];
  m_vAxis[Y] = ((CSLIDERotate*)pCSLIDETransform)->GetAxis()[Y];
  m_vAxis[Z] = ((CSLIDERotate*)pCSLIDETransform)->GetAxis()[Z];
  m_dynfTheta.SetValue(((CSLIDERotate*)pCSLIDETransform)->GetTheta().GetValue());

  m_m.SetRotate(m_vAxis, m_dynfTheta.GetValue());
  m_mInverse.SetRotate(m_vAxis, -m_dynfTheta.GetValue());
}

inline VOID CSLIDERotate::Modify(CSLIDETransform *pCSLIDETransformLast,
				 CSLIDETransform *pCSLIDETransformNew)
{
  printf( "CSLIDERotate::Modify not implementd -------------------------------------\n");
  //  if( m_vAxis == ((CSLIDERotate*)pCSLIDETransformNew)->GetAxis() )
}

#endif // __SLIDEROTATE_H__


