#ifndef __SLIDESCALE_H__
#define __SLIDESCALE_H__

#include "slidetransform.h"

//////////////////////////////////////////////////////////////////////
// CSLIDEScale Class
//

class CSLIDEScale : public CSLIDETransform
{
//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEScale();
  virtual ~CSLIDEScale();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    DynVector *pdynv);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual inline VOID ComposeWith(CSLIDETransform *pCSLIDETransform);
  virtual inline VOID Set(CSLIDETransform *pCSLIDETransform);
  virtual inline VOID Modify(CSLIDETransform *pCSLIDETransformLast, 
			     CSLIDETransform *pCSLIDETransformNew);

protected:
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();

//////////////////////////////////////////////////
// Rendering
//
public:

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);
  virtual VOID WriteRIB(FILE *fpOutputFile);
  virtual VOID WritePOV(FILE *fpOutputFile);
  virtual VOID WriteDotNode(FILE *fpOutputFile);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline CHAR *GetTypeString();
  virtual inline SLIDEReservedWordType GetType();

  virtual inline DynVector GetScale();

//////////////////////////////////////////////////
// Class Variables
//
public:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  DynVector m_vScale;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline CHAR *CSLIDEScale::GetTypeString()
{
  return gcppcSLIDEReservedWords[sliderwtScale];
}

inline SLIDEReservedWordType CSLIDEScale::GetType()
{
  return sliderwtScale;
}

inline DynVector CSLIDEScale::GetScale()
{
  return m_vScale;
}

inline VOID CSLIDEScale::ComposeWith(CSLIDETransform *pCSLIDETransform)
{
  m_vScale[X] *= ((CSLIDEScale*)pCSLIDETransform)->GetScale()[X];
  m_vScale[Y] *= ((CSLIDEScale*)pCSLIDETransform)->GetScale()[Y];
  m_vScale[Z] *= ((CSLIDEScale*)pCSLIDETransform)->GetScale()[Z];
  m_m.SetScale(m_vScale);
  m_mInverse.SetScale(-m_vScale);
}

inline VOID CSLIDEScale::Set(CSLIDETransform *pCSLIDETransform)
{
  m_vScale[X] = ((CSLIDEScale*)pCSLIDETransform)->GetScale()[X];
  m_vScale[Y] = ((CSLIDEScale*)pCSLIDETransform)->GetScale()[Y];
  m_vScale[Z] = ((CSLIDEScale*)pCSLIDETransform)->GetScale()[Z];

  m_m.SetScale(m_vScale);
  m_mInverse.SetScale(-m_vScale);
}

inline VOID CSLIDEScale::Modify(CSLIDETransform *pCSLIDETransformLast,
				CSLIDETransform *pCSLIDETransformNew)
{
  m_vScale[X] *= ( (((CSLIDEScale*)pCSLIDETransformNew)->GetScale()[X]) /
		  (((CSLIDEScale*)pCSLIDETransformLast)->GetScale()[X]) );
  m_vScale[Y] *= ( (((CSLIDEScale*)pCSLIDETransformNew)->GetScale()[Y]) /
		  (((CSLIDEScale*)pCSLIDETransformLast)->GetScale()[Y]) );
  m_vScale[Z] *= ( (((CSLIDEScale*)pCSLIDETransformNew)->GetScale()[Z]) /
		  (((CSLIDEScale*)pCSLIDETransformLast)->GetScale()[Z]) );

  m_m.SetScale(m_vScale);
  m_mInverse.SetScale(-m_vScale);
}

#endif // __SLIDESCALE_H__


