#include "slideseam.h"

#include "slideworld.h"
#include "slidesurface.h"
#include "slidecontour.h"
#include "stl.h"

//static const FLOAT cfNormalRatio = 1.0 / 10.0;
static const FLOAT cfNormalRatio = 1.0;

//////////////////////////////////////////////////////////////////////
// CSLIDESeam Class
//

CSLIDESeam::CSLIDESeam()
{
  m_dyniSolid.SetValue(slidestSLF_SOLID);
  m_pcContour1ID = NULL;
  m_pcContour2ID = NULL;
}

CSLIDESeam::~CSLIDESeam()
{
}

VOID CSLIDESeam::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID,
		      CHAR *pcContour1ID, CHAR *pcContour2ID)
{
  CSLIDEBoundedNode::Init(pcFile, uiLine, uiColumn, pcID);

  m_dyniSolid.SetValue(slidestSLF_SOLID);
  m_pcContour1ID = pcContour1ID;
  m_pcContour2ID = pcContour2ID;
}

VOID CSLIDESeam::Uninit()
{
  CSLIDEBoundedNode::Uninit();

  if ( m_pcContour1ID != NULL )
    {
      delete m_pcContour1ID;
      m_pcContour1ID = NULL;
    }

  if ( m_pcContour2ID != NULL )
    {
      delete m_pcContour2ID;
      m_pcContour2ID = NULL;
    }
}

BOOL CSLIDESeam::Initialized()
{
  if ( !CSLIDEBoundedNode::Initialized() )
    {
      return FALSE;
    }

  return ((m_pcContour1ID != NULL) && (m_pcContour2ID != NULL));
}

BOOL CSLIDESeam::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;
  CHAR *pcID;

  if ( !CSLIDEBoundedNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  // Check Solid Type
  ASSERT( 0 <= m_dyniSolid.GetValue() &&
	  m_dyniSolid.GetValue() < slidestSize );

  // Check Contour References
  pcID = m_pcContour1ID;
  m_pContour1 = (CSLIDEContour*)pCSLIDEWorld->PeekNode(pcID);
  if ( m_pContour1 == NULL )
    {
      HandleError("Undefined %s \"%s\"", gcppcSLIDEReservedWords[sliderwtContour], pcID);
      bCorrect = FALSE;
    }

  pcID = m_pcContour2ID;
  m_pContour2 = (CSLIDEContour*)pCSLIDEWorld->PeekNode(pcID);
  if ( m_pContour2 == NULL )
    {
      HandleError("Undefined %s \"%s\"", gcppcSLIDEReservedWords[sliderwtContour], pcID);
      bCorrect = FALSE;
    }

  return bCorrect;
}

BOOL CSLIDESeam::Valid()
{
  if ( !CSLIDEBoundedNode::Valid() )
    {
      return FALSE;
    }

  return ( (0 <= m_dyniSolid.GetValue()) && (m_dyniSolid.GetValue() < slidestSize) );
}


#ifdef __USE_CDA__
VOID CSLIDESeam::CalculateCDAVertices(CListInt *plistCDAVertex, 
					const Matrix &mGroot_obj)
{
}
#endif // __USE_CDA__

VOID CSLIDESeam::CheckDynamic()
{
  // Are there variables which I must call TCL to update?
  SetDynamic(m_dyniShading.GetDynamic() ||
	     m_dyniLOD.GetDynamic() ||
	     m_dyniSolid.GetDynamic() );
}

VOID CSLIDESeam::PreprocessGraph()
{
  if ( !GetVisited() )
    {
      CSLIDEContour *pCSLIDEContour;
      
      BOOL bGeometryEmpty;
      BOOL bGeometryDynamic;
      BOOL bNormalsEmpty;
      BOOL bNormalsDynamic;
      BOOL bBoundsEmpty;
      BOOL bBoundsDynamic;
      BOOL bBoundEmpty;
      BOOL bBoundDynamic;

      bGeometryEmpty = TRUE;
      bGeometryDynamic = FALSE;
      bNormalsEmpty = TRUE;
      bNormalsDynamic = FALSE;
      bBoundsEmpty = TRUE;
      bBoundsDynamic = FALSE;
      bBoundEmpty = TRUE;
      bBoundDynamic = FALSE;

      //////////////////////////////////////////////////
      // Visit Children Nodes and get starting values from them
      m_pContour1->PreprocessGraph();
      m_pContour2->PreprocessGraph();

      pCSLIDEContour = m_pContour1;

      bGeometryEmpty = ( bGeometryEmpty && 
			 pCSLIDEContour->GetGeometryEmpty() );
      bGeometryDynamic = ( bGeometryDynamic ||
			   pCSLIDEContour->GetGeometryDynamic() );
      bNormalsEmpty = ( bNormalsEmpty && 
			pCSLIDEContour->GetNormalsEmpty() );
      bNormalsDynamic = ( bNormalsDynamic ||
			  pCSLIDEContour->GetNormalsDynamic() );
      bBoundsEmpty = ( bBoundsEmpty && 
		       pCSLIDEContour->GetBoundsEmpty() );
      bBoundsDynamic = ( bBoundsDynamic ||
			 pCSLIDEContour->GetBoundsDynamic() );
      bBoundEmpty = ( bBoundEmpty && 
		      pCSLIDEContour->GetBoundEmpty() );
      bBoundDynamic = ( bBoundDynamic ||
			pCSLIDEContour->GetBoundDynamic() ||
			pCSLIDEContour->GetGeometryDynamic() ); // must add this
      
      pCSLIDEContour = m_pContour2;
      
      bGeometryEmpty = ( bGeometryEmpty && 
			 pCSLIDEContour->GetGeometryEmpty() );
      bGeometryDynamic = ( bGeometryDynamic ||
			   pCSLIDEContour->GetGeometryDynamic() );
      bNormalsEmpty = ( bNormalsEmpty && 
			pCSLIDEContour->GetNormalsEmpty() );
      bNormalsDynamic = ( bNormalsDynamic ||
			  pCSLIDEContour->GetNormalsDynamic() );
      bBoundsEmpty = ( bBoundsEmpty && 
		       pCSLIDEContour->GetBoundsEmpty() );
      bBoundsDynamic = ( bBoundsDynamic ||
			 pCSLIDEContour->GetBoundsDynamic() );
      bBoundEmpty = ( bBoundEmpty && 
		      pCSLIDEContour->GetBoundEmpty() );
      bBoundDynamic = ( bBoundDynamic ||
			pCSLIDEContour->GetBoundDynamic() ||
			pCSLIDEContour->GetGeometryDynamic() ); // must add this

      //////////////////////////////////////////////////
      // Look at Local flags

      // Update whether we are Empty based on our LOD flag
      if ( !m_dyniLOD.GetDynamic() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      // This node is always off
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	      bBoundsEmpty = TRUE;
	      bBoundEmpty = TRUE;
	    }
	  else if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	    {
	      // This node is always just a bounding box
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	    }
	}

      // Check what is really dynamic
      bGeometryDynamic = ( bGeometryDynamic ||
			   m_pCSLIDESurface->GetDynamic() ||
			   m_dyniShading.GetDynamic() ||
			   m_dyniLOD.GetDynamic() ||
			   m_dyniSolid.GetDynamic() );
      bNormalsDynamic = ( bNormalsDynamic ||
			  m_dyniShading.GetDynamic() ||
			  m_dyniLOD.GetDynamic() );
      bBoundsDynamic = ( bBoundsDynamic || 
			 m_dyniLOD.GetDynamic() );
      bBoundDynamic = ( bBoundDynamic || 
			m_dyniLOD.GetDynamic() );

      // If anything is now Empty then it will always be empty so
      // it is not dynamic
      bGeometryDynamic = ( bGeometryDynamic && !bGeometryEmpty );
      bNormalsDynamic = ( bNormalsDynamic && !bNormalsEmpty );
      bBoundsDynamic = ( bBoundsDynamic && !bBoundsEmpty );
      bBoundDynamic = ( bBoundDynamic && !bBoundEmpty );
      
      //////////////////////////////////////////////////
      // Now save these correct flag values

      SetGeometryEmpty(bGeometryEmpty);
      SetGeometryDynamic(bGeometryDynamic);
      SetNormalsEmpty(bNormalsEmpty);
      SetNormalsDynamic(bNormalsDynamic);
      SetBoundsEmpty(bBoundsEmpty);
      SetBoundsDynamic(bBoundsDynamic);
      SetBoundEmpty(bBoundEmpty);
      SetBoundDynamic(bBoundDynamic);
      
      //////////////////////////////////////////////////
      // Calculate the Bounding Box (if Empty this will do nothing)
      if ( GetBoundEmpty() )
	{
	  SetBoundValid(TRUE);
	}
      else
	{
	  CalculateBound();
	}
      SetBoundChanged(FALSE);

      //////////////////////////////////////////////////
      // Done Preproces(FLOAT)sing this Node in the Graph
      SetVisited(TRUE);
    }
}

VOID CSLIDESeam::UpdateState()
{
  CSLIDEBoundedNode::UpdateState();

  m_dyniSolid.Update();
  if ( !( 0 <= m_dyniSolid.GetValue() &&  
	  m_dyniSolid.GetValue() < slidestSize ) )
    {
      HandleError("Illegal solid flag value:  %ld", m_dyniSolid.GetValue());
      m_dyniSolid.SetValue(slidestSLF_HOLLOW);
    }
}

VOID CSLIDESeam::CalculateBound()
{
  CBound bound;
  Matrix m;

  m_pContour1->InsertIntoBound(m, bound);
  m_pContour2->InsertIntoBound(m, bound);

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);
    }
}


VOID CSLIDESeam::RenderGeometrySmooth(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{
  UINT_32 uiContour1Point, uiContour2Point, uiContour1Points, uiContour2Points;
  Point *pContour1Point, *pContour2Point;
  Point pt, p1, p2, p3, p1Curr, p2Curr;
  Vector v1, v2, vNormal1, vNormal2, vNormal3;
  Vector *vContour1Normal, *vContour2Normal;
  FLOAT fv1Mag, fv2Mag;

  uiContour1Points = m_pContour1->GetPointCount();
  pContour1Point = new Point[uiContour1Points];
  vContour1Normal = new Vector[uiContour1Points];
  m_pContour1->GetPoints(pContour1Point);
  m_pContour1->GetNormals(vContour1Normal);

  uiContour2Points = m_pContour2->GetPointCount();
  pContour2Point = new Point[uiContour2Points];
  vContour2Normal = new Vector[uiContour2Points];
  m_pContour2->GetPoints(pContour2Point);
  m_pContour2->GetNormals(vContour2Normal);

  if( !((uiContour1Points > 0) && (uiContour1Points > 0)) ) 
    {
      return;
    }

  p1 = pContour1Point[0];
  vNormal1 = vContour1Normal[0];
  p2 = pContour2Point[0];
  vNormal2 = vContour2Normal[0];

  uiContour1Point = 0;
  uiContour2Point = 0;

  if( uiContour1Points >= 2)
    {
      p1Curr = pContour1Point[1];
      uiContour1Point = 1;
    }
  
  if( uiContour2Points >= 2)
    {
      p2Curr = pContour2Point[1];
      uiContour2Point = 1;
    }
    
  while( (uiContour1Point < uiContour1Points) || (uiContour2Point < uiContour2Points))
    {
      gilBegin(GL_TRIANGLES);
      {
	gilNormal3fv(vNormal1.Get());
	gilVertex3fv(p1.Get());
	gilNormal3fv(vNormal2.Get());
	gilVertex3fv(p2.Get());

	fv2Mag = INFINITY;
	if(uiContour1Point < uiContour1Points)
	  {
	    fv2Mag = (p1Curr - p2).Magnitude();
	  }
	
	fv1Mag = INFINITY;
	if(uiContour2Point < uiContour2Points)
	  {
	    fv1Mag = (p2Curr - p1).Magnitude();
	  }
	
	if( fv1Mag < fv2Mag ) // Triangulate with shorter segment
	  {
	    p3 = p2Curr;
	    if( uiContour2Point < uiContour2Points )
	      {
		vNormal3 = vContour2Normal[uiContour2Point];
	      }
	    else
	      {
		printf( "Bar2 Jane\n");
	      }

	    uiContour2Point++;
	    if( uiContour2Point < uiContour2Points )
	      {
		p2Curr = pContour2Point[uiContour2Point];
	      }
	    
	    p2 = p3;
	    vNormal2 = vNormal3;
	  }
	else
	  {
	    p3 = p1Curr;
	    if( uiContour1Point < uiContour1Points )
	      {
		vNormal3 = vContour2Normal[uiContour1Point];
	      }
	    else
	      {
		printf( "Foo2 Jane\n");
	      }

	    uiContour1Point++;
	    if( uiContour1Point < uiContour1Points )
	      {
		p1Curr = pContour1Point[uiContour1Point];
	      }
	    
	    p1 = p3;
	    vNormal1 = vNormal3;
	  }

	gilNormal3fv(vNormal3.Get());	
	gilVertex3fv(p3.Get());
      }
      gilEnd();
    }

  delete[] pContour1Point;
  delete[] vContour1Normal;
  delete[] pContour2Point;
  delete[] vContour2Normal;
}

VOID CSLIDESeam::RenderGeometryFlat(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{
  UINT_32 uiContour1Point, uiContour2Point, uiContour1Points, uiContour2Points;
  Point *pContour1Point, *pContour2Point;
  Point p1, p2, p3, p1Curr, p2Curr;
  Vector v1, v2, vNormal;
  FLOAT fv1Mag, fv2Mag;

  uiContour1Points = m_pContour1->GetPointCount();
  pContour1Point = new Point[uiContour1Points];
  m_pContour1->GetPoints(pContour1Point);

  uiContour2Points = m_pContour2->GetPointCount();
  pContour2Point = new Point[uiContour2Points];
  m_pContour2->GetPoints(pContour2Point);

  if( !((uiContour1Points > 0) && (uiContour1Points > 0)) ) 
    {
      return;
    }

  p1 = pContour1Point[0];
  p2 = pContour2Point[0];
  
  uiContour1Point = 0;
  uiContour2Point = 0;

  if( uiContour1Points >= 2)
    {
      p1Curr = pContour1Point[1];
      uiContour1Point = 1;
    }
  
  if( uiContour2Points >= 2)
    {
      p2Curr = pContour2Point[1];
      uiContour2Point = 1;
    }
    
  while( (uiContour1Point < uiContour1Points) || (uiContour2Point < uiContour2Points))
    {
      gilBegin(GL_TRIANGLES);
      {
	gilVertex3fv(p1.Get());
	gilVertex3fv(p2.Get());
	
	fv2Mag = INFINITY;
	if(uiContour1Point < uiContour1Points)
	  {
	    fv2Mag = (p1Curr - p2).Magnitude();
	  }
	
	fv1Mag = INFINITY;
	if(uiContour2Point < uiContour2Points)
	  {
	    fv1Mag = (p2Curr - p1).Magnitude();
	  }
	
	if( fv1Mag < fv2Mag ) // Triangulate with shorter segment
	  {
	    p3 = p2Curr;
	    uiContour2Point++;
	    if( uiContour2Point < uiContour2Points )
		{
		  p2Curr = pContour2Point[uiContour2Point];
		}
	    vNormal = (p2-p1)*(p3-p1);	
	    vNormal.Normalize();

	    p2 = p3;
	  }
	else
	  {
	    p3 = p1Curr;
	    uiContour1Point++;
	    if( uiContour1Point < uiContour1Points )
	      {
		  p1Curr = pContour1Point[uiContour1Point];
	      }
	    vNormal = (p2-p1)*(p3-p1);	
	    vNormal.Normalize();

	    p1 = p3;
	  }
	
	gilNormal3fv(vNormal.Get());
	gilVertex3fv(p3.Get());
      }
      gilEnd();
    }

  delete[] pContour1Point;
  delete[] pContour2Point;
}

VOID CSLIDESeam::RenderNormalsSmooth()
{
  UINT_32 uiContour1Point, uiContour2Point, uiContour1Points, uiContour2Points;
  Point *pContour1Point, *pContour2Point;
  Point pt, p1, p2, p3, p1Curr, p2Curr;
  Vector v1, v2, vNormal1, vNormal2, vNormal3;
  Vector *vContour1Normal, *vContour2Normal;
  FLOAT fv1Mag, fv2Mag;

  uiContour1Points = m_pContour1->GetPointCount();
  pContour1Point = new Point[uiContour1Points];
  vContour1Normal = new Vector[uiContour1Points];
  m_pContour1->GetPoints(pContour1Point);
  m_pContour1->GetNormals(vContour1Normal);

  uiContour2Points = m_pContour2->GetPointCount();
  pContour2Point = new Point[uiContour2Points];
  vContour2Normal = new Vector[uiContour2Points];
  m_pContour2->GetPoints(pContour2Point);
  m_pContour2->GetNormals(vContour2Normal);

  if( !((uiContour1Points > 0) && (uiContour1Points > 0)) ) 
    {
      return;
    }

  p1 = pContour1Point[0];
  vNormal1 = vContour1Normal[0];
  p2 = pContour2Point[0];
  vNormal2 = vContour2Normal[0];

  gilBegin(GL_LINES);
  {
    gilVertex3fv(p1.Get());
    pt = p1 + vNormal1 * cfNormalRatio;
    gilVertex3fv(pt.Get());
    
    gilVertex3fv(p2.Get());
    pt = p2 + vNormal2 * cfNormalRatio;
    gilVertex3fv(pt.Get());
  }
  gilEnd();
  
  uiContour1Point = 0;
  uiContour2Point = 0;

  if( uiContour1Points >= 2)
    {
      p1Curr = pContour1Point[1];
      uiContour1Point = 1;
    }
  
  if( uiContour2Points >= 2)
    {
      p2Curr = pContour2Point[1];
      uiContour2Point = 1;
    }
    
  while( (uiContour1Point < uiContour1Points) || (uiContour2Point < uiContour2Points))
    {
      gilBegin(GL_LINES);
      {
	fv2Mag = INFINITY;
	if(uiContour1Point < uiContour1Points)
	  {
	    fv2Mag = (p1Curr - p2).Magnitude();
	  }
	
	fv1Mag = INFINITY;
	if(uiContour2Point < uiContour2Points)
	  {
	    fv1Mag = (p2Curr - p1).Magnitude();
	  }
	
	if( fv1Mag < fv2Mag ) // Triangulate with shorter segment
	  {
	    p3 = p2Curr;
	    if( uiContour2Point < uiContour2Points )
	      {
		vNormal3 = vContour2Normal[uiContour2Point];
	      }
	    else
	      {
		printf( "Bar Jane\n");
	      }

	    uiContour2Point++;
	    if( uiContour2Point < uiContour2Points )
	      {
		p2Curr = pContour2Point[uiContour2Point];
	      }
	    
	    p2 = p3;
	  }
	else
	  {
	    p3 = p1Curr;
	    if( uiContour1Point < uiContour1Points )
	      {
		vNormal3 = vContour2Normal[uiContour1Point];
	      }
	    else
	      {
		printf( "Foo Jane\n");
	      }

	    uiContour1Point++;
	    if( uiContour1Point < uiContour1Points )
	      {
		p1Curr = pContour1Point[uiContour1Point];
	      }
	    
	    p1 = p3;
	  }
	
	gilVertex3fv(p3.Get());
	pt = p3 + vNormal3 * cfNormalRatio;
	gilVertex3fv(pt.Get());
      }
      gilEnd();
    }

  delete[] pContour1Point;
  delete[] pContour2Point;
  delete[] vContour1Normal;
  delete[] vContour2Normal;
}

VOID CSLIDESeam::RenderNormalsFlat()
{
  UINT_32 uiContour1Point, uiContour2Point, uiContour1Points, uiContour2Points;
  Point *pContour1Point, *pContour2Point;
  Point pt, p1, p2, p3, p1Curr, p2Curr;
  Vector v1, v2, vNormal;
  FLOAT fv1Mag, fv2Mag;

  uiContour1Points = m_pContour1->GetPointCount();
  pContour1Point = new Point[uiContour1Points];
  m_pContour1->GetPoints(pContour1Point);

  uiContour2Points = m_pContour2->GetPointCount();
  pContour2Point = new Point[uiContour2Points];
  m_pContour2->GetPoints(pContour2Point);

  if( !((uiContour1Points > 0) && (uiContour1Points > 0)) ) 
    {
      return;
    }

  p1 = pContour1Point[0];
  p2 = pContour2Point[0];
  
  uiContour1Point = 0;
  uiContour2Point = 0;

  if( uiContour1Points >= 2)
    {
      p1Curr = pContour1Point[1];
      uiContour1Point = 1;
    }
  
  if( uiContour2Points >= 2)
    {
      p2Curr = pContour2Point[1];
      uiContour2Point = 1;
    }
    
  while( (uiContour1Point < uiContour1Points) || (uiContour2Point < uiContour2Points))
    {
      gilBegin(GL_LINES);
      {
	fv2Mag = INFINITY;
	if(uiContour1Point < uiContour1Points)
	  {
	    fv2Mag = (p1Curr - p2).Magnitude();
	  }
	
	fv1Mag = INFINITY;
	if(uiContour2Point < uiContour2Points)
	  {
	    fv1Mag = (p2Curr - p1).Magnitude();
	  }
	
	if( fv1Mag < fv2Mag ) // Triangulate with shorter segment
	  {
	    p3 = p2Curr;
	    pt = (p1 + p2 + p3) / 3;

	    uiContour2Point++;
	    if( uiContour2Point < uiContour2Points )
	      {
		p2Curr = pContour2Point[uiContour2Point];
	      }
	    vNormal = (p2-p1)*(p3-p1);	
	    vNormal.Normalize();
	    
	    p2 = p3;
	  }
	else
	  {
	    p3 = p1Curr;
	    pt = (p1 + p2 + p3) / 3;

	    uiContour1Point++;
	    if( uiContour1Point < uiContour1Points )
	      {
		p1Curr = pContour1Point[uiContour1Point];
	      }
	    vNormal = (p2-p1)*(p3-p1);	
	    vNormal.Normalize();
	    
	    p1 = p3;
	  }

	gilVertex3fv(pt.Get());

	pt += vNormal * cfNormalRatio;
	gilVertex3fv(pt.Get());
      }
      gilEnd();
    }

  delete[] pContour1Point;
  delete[] pContour2Point;
}

VOID CSLIDESeam::WriteSLIDE(FILE *fpOutputFile)
{
  /*
  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s ( ", 
	  gcppcSLIDEReservedWords[sliderwtSeam], m_pcGUIID);

  iter.Init(&m_listCSLIDEFace);

  for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst(); 
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
    {
      fprintf(fpOutputFile, "%s ", pCSLIDEFace->GetGUIID());      
    }

  fprintf(fpOutputFile, ")\n");

  WriteSLIDEFlags(fpOutputFile, 0);

  if ( m_dyniSolid.GetValue() != slidestSLF_SOLID )
    {
      fprintf(fpOutputFile, "  %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtSolid], 
	      gcppcSLIDESolidReservedWords[m_dyniSolid.GetValue()]);
    }

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndSeam]);
	  */
}

VOID CSLIDESeam::WriteRIBGeometry(FILE *fpOutputFile,
				    const Matrix &/*mGproj_obj*/, 
				    CBound * /*pCBoundClip*/, 
				    SLIDEShadingType eShading,
				    SLIDELODType eLOD)
{
  /*
  ASSERT( fpOutputFile != NULL );

  WriteRIBPushSolidAttributes(fpOutputFile,
			      (SLIDESolidType)m_dyniSolid.GetValue());
  {
    WriteRIBGeometryPolygonal(fpOutputFile, eShading, eLOD);
  }
  WriteRIBPopSolidAttributes(fpOutputFile, 
			     (SLIDESolidType)m_dyniSolid.GetValue());
*/
}

BOOL CSLIDESeam::CheckRIBPrimitive()
{
  BOOL bResult;

  bResult = FALSE;

  return bResult;
}

VOID CSLIDESeam::WriteRIBGeometryPolygonal(FILE *fpOutputFile,
					     SLIDEShadingType eShading,
					     SLIDELODType /* eLOD */)
{
  /*
  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;

  ASSERT( fpOutputFile != NULL );

  iter.Init(&m_listCSLIDEFace);
  
  for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst(); 
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
    {
      pCSLIDEFace->WriteRIB(fpOutputFile, eShading);
    }
    */
}

VOID CSLIDESeam::WriteRIBGeometryPrimitive(FILE *fpOutputFile,
					     SLIDEShadingType /* eShading */,
					     SLIDELODType /* eLOD */)
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDESeam::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE SEAM: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);
}

VOID CSLIDESeam::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE SEAM: %s", m_pcGUIID);
}

VOID CSLIDESeam::WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				   CSLIDESurface *pCSLIDESurface)
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDESeam::WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform)
{
  FLOAT f;
  f = 1000;

  ASSERT( fpOutputFile != NULL );

  UINT_32 i, uiContour1Point, uiContour2Point, uiContour1Points, uiContour2Points;
  Point *pContour1Point, *pContour2Point;
  Point p1, p2, p3, p1Curr, p2Curr;
  Vector v1, v2;
  FLOAT fv1Mag, fv2Mag;

  if( mTransform.Determinant() < 0 ) // Need to Mirror triangles
    {
      uiContour2Points = m_pContour1->GetPointCount();
      pContour2Point = new Point[uiContour2Points];
      m_pContour1->GetPoints(pContour2Point);
      
      uiContour1Points = m_pContour2->GetPointCount();
      pContour1Point = new Point[uiContour1Points];
      m_pContour2->GetPoints(pContour1Point);
    }
  else // Don't mirror triangles
    {
      uiContour1Points = m_pContour1->GetPointCount();
      pContour1Point = new Point[uiContour1Points];
      m_pContour1->GetPoints(pContour1Point);
      
      uiContour2Points = m_pContour2->GetPointCount();
      pContour2Point = new Point[uiContour2Points];
      m_pContour2->GetPoints(pContour2Point);
    }
  
  if( !((uiContour1Points > 0) && (uiContour1Points > 0)) ) 
    {
      return;
    }

  for( i = 0; i < uiContour1Points; i++ )
    {
      pContour1Point[i] = mTransform * pContour1Point[i];
    }
      
  for( i = 0; i < uiContour2Points; i++ )
    {
      pContour2Point[i] = mTransform * pContour2Point[i];
    }

  p1 = pContour1Point[0];
  p2 = pContour2Point[0];
  
  uiContour1Point = 0;
  uiContour2Point = 0;

  if( uiContour1Points >= 2)
    {
      p1Curr = pContour1Point[1];
      uiContour1Point = 1;
    }
  
  if( uiContour2Points >= 2)
    {
      p2Curr = pContour2Point[1];
      uiContour2Point = 1;
    }

  BOOL ff = FALSE;  
  while( (uiContour1Point < uiContour1Points) || (uiContour2Point < uiContour2Points))
    {
      fv2Mag = INFINITY;
      if(uiContour1Point < uiContour1Points)
	{
	  fv2Mag = (p1Curr - p2).Magnitude();
	}
      
      fv1Mag = INFINITY;
      if(uiContour2Point < uiContour2Points)
	{
	  fv1Mag = (p2Curr - p1).Magnitude();
	}
      
      if( fv1Mag < fv2Mag ) // Triangulate with shorter segment
	{
	  p3 = p2Curr;
	  uiContour2Point++;
	  if( uiContour2Point < uiContour2Points )
	    {
	      p2Curr = pContour2Point[uiContour2Point];
	    }

	  if( !((fabs(p1[X] - p2[X]) <= EPSILON) &&
		(fabs(p1[Y] - p2[Y]) <= EPSILON) &&
		(fabs(p1[Z] - p2[Z]) <= EPSILON)) &&
	      !((fabs(p1[X] - p3[X]) <= EPSILON) &&
		(fabs(p1[Y] - p3[Y]) <= EPSILON) &&
		(fabs(p1[Z] - p3[Z]) <= EPSILON)) &&
	      !((fabs(p2[X] - p3[X]) <= EPSILON) &&
		(fabs(p2[Y] - p3[Y]) <= EPSILON) &&
		(fabs(p2[Z] - p3[Z]) <= EPSILON)) )
	    {
	      if( (p1 - p2).Magnitude() < f )
		{
		  f = (p1 - p2).Magnitude();
		  ff = TRUE;
		}
	      if( (p2 - p3).Magnitude() < f )
		{
		  f = (p2 - p3).Magnitude();
		  ff = TRUE;
		}
	      if( (p3 - p1).Magnitude() < f )
		{
		  f = (p3 - p1).Magnitude();
		  ff = TRUE;
		}

	      if( (f < 0.000113) && ff)
		{
		  printf( "Here A\n");		  
		}
	      ff = FALSE;

	      WriteSTLTriangle(fpOutputFile, p1, p2, p3);
	    }
	  p2 = p3;
	}
      else
	{
	  p3 = p1Curr;
	  uiContour1Point++;
	  if( uiContour1Point < uiContour1Points )
	    {
	      p1Curr = pContour1Point[uiContour1Point];
	    }

	  if( !((fabs(p1[X] - p2[X]) <= EPSILON) &&
		(fabs(p1[Y] - p2[Y]) <= EPSILON) &&
		(fabs(p1[Z] - p2[Z]) <= EPSILON)) &&
	      !((fabs(p1[X] - p3[X]) <= EPSILON) &&
		(fabs(p1[Y] - p3[Y]) <= EPSILON) &&
		(fabs(p1[Z] - p3[Z]) <= EPSILON)) &&
	      !((fabs(p2[X] - p3[X]) <= EPSILON) &&
		(fabs(p2[Y] - p3[Y]) <= EPSILON) &&
		(fabs(p2[Z] - p3[Z]) <= EPSILON)) )
	    {
	      if( (p1 - p2).Magnitude() < f )
		{
		  f = (p1 - p2).Magnitude();
		  ff = TRUE;
		}
	      if( (p2 - p3).Magnitude() < f )
		{
		  f = (p2 - p3).Magnitude();
		  ff = TRUE;
		}
	      if( (p3 - p1).Magnitude() < f )
		{
		      f = (p3 - p1).Magnitude();
		  ff = TRUE;
		}

	      if( (f < 0.000113) && ff)
		{
		  printf( "Here A\n");		  
		}
	      ff = FALSE;

	      WriteSTLTriangle(fpOutputFile, p1, p2, p3);
	    }
	  p1 = p3;
	}
    }	
  delete[] pContour1Point;
  delete[] pContour2Point;
}

VOID CSLIDESeam::WriteDotID(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "seam_%s", GetGUIID());
}

VOID CSLIDESeam::WriteDotNode(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "  ");
  WriteDotID(fpOutputFile);
  fprintf(fpOutputFile, " [ label = \"%s\", shape = ellipse, peripheries = 2", GetGUIID());
  if ( GetBoundDynamic() )
    {
      fprintf(fpOutputFile, ", color = red, fontcolor = red");
    }
  else if ( GetGeometryDynamic() )
    {
      fprintf(fpOutputFile, ", color = blue, fontcolor = blue");
    }
  fprintf(fpOutputFile, " ];\n");
}

VOID CSLIDESeam::WriteDotEdges(FILE * /*fpOutputFile*/)
{
}


