#include <stdlib.h>
#include "slidesgroot.h"

#include "slideworld.h"
#include "slidesurface.h"
#include "qemesh.h"
#include "qevertex.h"

//////////////////////////////////////////////////////////////////////
// CSLIDESphere Class
//

CSLIDESGRoot::CSLIDESGRoot()
{
  m_pcSGRoot = NULL;
}

CSLIDESGRoot::~CSLIDESGRoot()
{
}

VOID CSLIDESGRoot::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			CHAR *pcID, CHAR *pcSGRoot)
{
  CSLIDEAtomNode::Init(pcFile, uiLine, uiColumn, pcID);

  m_pcSGRoot = pcSGRoot;
}

VOID CSLIDESGRoot::Uninit()
{
  CSLIDEAtomNode::Uninit();
}

BOOL CSLIDESGRoot::Initialized()
{
  if ( !CSLIDEAtomNode::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

BOOL CSLIDESGRoot::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;
  CListIterInt iter;

  if ( !CSLIDEAtomNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  if ( m_pcSGRoot == NULL )
    {
      HandleError("Illegal SGRoot");
      bCorrect = FALSE;
    }

  return bCorrect;
}

BOOL CSLIDESGRoot::Valid()
{
  if ( !CSLIDEAtomNode::Valid() )
    {
      return FALSE;
    }

  return ( m_pcSGRoot != NULL );
}

VOID CSLIDESGRoot::GetSide(CListInt *plistPoint, CListInt *plistNormal, UINT_32 uiSide, 
			   const Matrix &mTransform, const Matrix &mInverse, BOOL bReverse)
{
}

VOID CSLIDESGRoot::CheckDynamic()
{
  CSLIDEAtomNode::CheckDynamic();

  // Are there variables which I must call TCL to update?
  SetDynamic(TRUE);
}

VOID CSLIDESGRoot::PreprocessGraphLocal()
{
  SetGeometryEmpty(FALSE);
  SetGeometryDynamic(TRUE);
  SetNormalsEmpty(FALSE);
  SetNormalsDynamic(TRUE);
  SetBoundsEmpty(FALSE);
  SetBoundsDynamic(FALSE);
  SetBoundEmpty(FALSE);
  SetBoundDynamic(FALSE);
}

VOID CSLIDESGRoot::UpdateState()
{
  CSLIDEAtomNode::UpdateState();
}

VOID CSLIDESGRoot::CalculateBound()
{
  Point pt;

  pt.Set(INFINITY, INFINITY, INFINITY);
  m_bound.Insert(pt);
  pt.Set(-INFINITY, -INFINITY, -INFINITY);
  m_bound.Insert(pt);
}

#ifdef __USE_CDA__
VOID CSLIDESGRoot::CalculateCDAVertices(CListInt *plistCDAVertex, 
					const Matrix &mGroot_obj)
{
}
#endif // __USE_CDA__

VOID CSLIDESGRoot::CalculateQEMesh()
{
}

VOID CSLIDESGRoot::RenderGeometrySmooth()
{
  static CHAR spcBuffer[200];

  sprintf(spcBuffer, "%s Update", m_pcSGRoot);
  Tcl_GlobalEval(CSLIDEWorld::GetInterp(), spcBuffer);
  sprintf(spcBuffer, "%s Render", m_pcSGRoot);
  Tcl_GlobalEval(CSLIDEWorld::GetInterp(), spcBuffer);
}

VOID CSLIDESGRoot::RenderGeometryFlat()
{
  RenderGeometrySmooth();
}

VOID CSLIDESGRoot::RenderNormalsSmooth()
{
}

VOID CSLIDESGRoot::RenderNormalsFlat()
{
}

VOID CSLIDESGRoot::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s \"%s\"\n", 
	  gcppcSLIDEReservedWords[sliderwtSGRoot], m_pcGUIID, m_pcSGRoot);

  WriteSLIDEFlags(fpOutputFile, 0);

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndSGRoot]);
}

VOID CSLIDESGRoot::WriteRIBGeometry(FILE *fpOutputFile,
				    const Matrix &/*mGproj_obj*/, 
				    CBound * /*pCBoundClip*/, 
				    SLIDEShadingType /* eShading */,
				    SLIDELODType /* eLOD*/ )
{
}

VOID CSLIDESGRoot::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE SGROOT: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);
}

VOID CSLIDESGRoot::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE SGROOT: %s", m_pcGUIID);
}

VOID CSLIDESGRoot::WritePOVGeometry(FILE *fpOutputFile, 
				    CListInt *plistTransformIDs,
				    CListInt *plistColors)  
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDESGRoot::WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				     CSLIDESurface *pCSLIDESurface)
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDESGRoot::WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform)
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDESGRoot::WriteDotID(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "sgroot_%s", GetGUIID());
}

VOID CSLIDESGRoot::WriteDotNode(FILE * fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "  ");
  WriteDotID(fpOutputFile);
  fprintf(fpOutputFile, " [ label = \"%s\", shape = diamond ];\n", GetGUIID());
}

VOID CSLIDESGRoot::WriteDotEdges(FILE * /*fpOutputFile*/)
{
}

VOID CSLIDESGRoot::GetParameters(Tcl_Interp *pInterp)
{
}

UINT_32 CSLIDESGRoot::SetParameter(CSLIDEWorld *pCSLIDEWorld, 
				 SLIDEReservedWordType uiParameter, CHAR **argv)
{
  switch(uiParameter) 
    {
    default:
      return CSLIDEAtomNode::SetParameter(pCSLIDEWorld, uiParameter, argv);
      break;
    }
}
