#include "slidesubdivision.h"

//#include "slideworld.h"
#include "slidesurface.h"
#include "slideinstance.h"

#include "qemesh.h"
#include "qerender.h"
#include "qewrite.h"

//#include "dtmesh.h"
//#include "dtcontour.h"
//#include "dtcontourvertex.h"
//#include "dtcontouredge.h"
//#include "dtvertex.h"
//#include "dtedge.h"
//#include "dttriangle.h"

extern BOOL gbPreprocessDone;

static const UINT_32 cuiUseDisplayLists = FALSE;
static const UINT_32 cuiDrawControlMesh = FALSE;
static const UINT_32 cuiDrawCurrentMesh = FALSE;
static const UINT_32 cuiDrawVertices = FALSE;
static const SLIDESubdivisionType ceType = slidesubtSLF_SUBDIVISION_SHARP_CATMULL_CLARK;
static const SLIDEErrorMetricType ceErrorMetric = slideemtSLF_ERROR_METRIC_BEST_FIT_PLANE;
static const UINT_32 cuiFacetsMax = 0;
static const UINT_32 cuiSubdivisionLevel = 0;
static const UINT_32 cuiSubdivisionLevelSize = 6;

static const UINT_32 cuiWeightedEdges = 0;
static const FLOAT cfWeight = 1.0;
static const FLOAT cfTolerance = 1.0;

static const UINT_32 cuiA = 0;
static const UINT_32 cuiB = 0;

static FLOAT s_ppfVertexColors[8][3] = 
{
  { 1.0, 1.0, 1.0 },
  { 1.0, 1.0, 0.0 },
  { 1.0, 0.0, 0.0 },
  { 1.0, 0.0, 1.0 },
  { 0.0, 0.0, 1.0 },
  { 0.0, 1.0, 1.0 },
  { 0.0, 1.0, 0.0 },
  { 0.0, 0.0, 0.0 }
};

//////////////////////////////////////////////////////////////////////
// CSLIDESubdivision Class
//

CSLIDESubdivision::CSLIDESubdivision()
{
  m_dyniUseDisplayLists.SetValue(cuiUseDisplayLists);
  m_dyniDrawControlMesh.SetValue(cuiDrawControlMesh);
  m_dyniDrawCurrentMesh.SetValue(cuiDrawCurrentMesh);
  m_dyniDrawVertices.SetValue(cuiDrawVertices);
  m_dyniType.SetValue(ceType);
  m_dyniErrorMetric.SetValue(ceErrorMetric);
  m_dyniFacetsMax.SetValue(cuiFacetsMax);
  m_dyniSubdivisionLevel.SetValue(cuiSubdivisionLevel);
  m_uiSubdivisionLevelSize = cuiSubdivisionLevelSize;
  m_ppQEMeshSubdivision = NULL;

  m_dyniWeightedEdges.SetValue(cuiWeightedEdges);
  m_dynfWeight.SetValue(cfWeight);
  m_dynfTolerance.SetValue(cfTolerance);

  m_dyniA.SetValue(cuiA);
  m_dyniB.SetValue(cuiB);
}

CSLIDESubdivision::~CSLIDESubdivision()
{
}

VOID CSLIDESubdivision::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			     CHAR *pcID)
{
  UINT_32 i;

  CSLIDEGroup::Init(pcFile, uiLine, uiColumn, pcID);

  m_dyniUseDisplayLists.SetValue(cuiUseDisplayLists);
  m_dyniDrawControlMesh.SetValue(cuiDrawControlMesh);
  m_dyniDrawCurrentMesh.SetValue(cuiDrawCurrentMesh);
  m_dyniDrawVertices.SetValue(cuiDrawVertices);
  m_dyniType.SetValue(ceType);
  m_dyniErrorMetric.SetValue(ceErrorMetric);
  m_dyniFacetsMax.SetValue(cuiFacetsMax);
  m_dyniSubdivisionLevel.SetValue(cuiSubdivisionLevel);
  m_uiSubdivisionLevelSize = cuiSubdivisionLevelSize;
  m_ppQEMeshSubdivision = new QEMesh*[m_uiSubdivisionLevelSize];
  ASSERT( m_ppQEMeshSubdivision != NULL );
  for ( i = 0; i < m_uiSubdivisionLevelSize; i++ )
    {
      m_ppQEMeshSubdivision[i] = NULL;
    }

  m_dyniWeightedEdges.SetValue(cuiWeightedEdges);
  m_dynfWeight.SetValue(cfWeight);
  m_dynfTolerance.SetValue(cfTolerance);

  m_dyniA.SetValue(cuiA);
  m_dyniB.SetValue(cuiB);
}

VOID CSLIDESubdivision::Uninit()
{
  UINT_32 i;

  m_pQEMesh = NULL;

  CSLIDEGroup::Uninit();

  if ( m_ppQEMeshSubdivision != NULL )
    {
      for ( i = 0; i < m_uiSubdivisionLevelSize; i++ )
	{
	  if ( m_ppQEMeshSubdivision[i] != NULL )
	    {
	      DeallocateQEMesh(m_ppQEMeshSubdivision[i]);
	      m_ppQEMeshSubdivision[i] = NULL;
	    }
	}
      delete []m_ppQEMeshSubdivision;
      m_ppQEMeshSubdivision = NULL;
    }
}

BOOL CSLIDESubdivision::Initialized()
{
  if ( !CSLIDEGroup::Initialized() )
    {
      return FALSE;
    }

  return m_ppQEMeshSubdivision != NULL;
}

BOOL CSLIDESubdivision::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;

  if ( !CSLIDEGroup::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  if ( !m_dyniUseDisplayLists.GetDynamic() &&
       !( m_dyniUseDisplayLists.GetValue() == TRUE || 
	  m_dyniUseDisplayLists.GetValue() == FALSE ) )
    {
      //HandleError("Illegal usedisplaylists value:  %ld", 
      //	  m_dyniUseDisplayLists.GetValue());
      //m_dyniUseDisplayLists.SetValue(FALSE);
      //bCorrect = FALSE;
    }

  if ( !m_dyniDrawControlMesh.GetDynamic() &&
       !( m_dyniDrawControlMesh.GetValue() == TRUE || 
	  m_dyniDrawControlMesh.GetValue() == FALSE ) )
    {
      //HandleError("Illegal drawcontrolmesh value:  %ld", 
      //	  m_dyniDrawControlMesh.GetValue());
      //m_dyniDrawControlMesh.SetValue(FALSE);
      //bCorrect = FALSE;
    }

  if ( !m_dyniDrawCurrentMesh.GetDynamic() &&
       !( m_dyniDrawCurrentMesh.GetValue() == TRUE || 
	  m_dyniDrawCurrentMesh.GetValue() == FALSE ) )
    {
      //HandleError("Illegal drawCurrentmesh value:  %ld", 
      //	  m_dyniDrawCurrentMesh.GetValue());
      //m_dyniDrawCurrentMesh.SetValue(FALSE);
      //bCorrect = FALSE;
    }

  if ( !m_dyniDrawVertices.GetDynamic() &&
       !( m_dyniDrawVertices.GetValue() == TRUE || 
	  m_dyniDrawVertices.GetValue() == FALSE ) )
    {
      //HandleError("Illegal drawVertices value:  %ld", 
      //	  m_dyniDrawVertices.GetValue());
      //m_dyniDrawVertices.SetValue(FALSE);
      //bCorrect = FALSE;
    }

  m_dyniType.Update();
  if ( !( 0 <= m_dyniType.GetValue() &&
	  m_dyniType.GetValue() < slidesubtSize ) )
    {
      HandleError("Illegal Subdivision Type flag value:  %ld", 
		  m_dyniType.GetValue());
      m_dyniType.SetValue(ceType);
    }

  m_dyniErrorMetric.Update();
  if ( !( 0 <= m_dyniErrorMetric.GetValue() &&
	  m_dyniErrorMetric.GetValue() < slideemtSize ) )
    {
      HandleError("Illegal Error Metric flag value:  %ld", 
		  m_dyniErrorMetric.GetValue());
      m_dyniErrorMetric.SetValue(ceErrorMetric);
    }

  if ( !m_dyniSubdivisionLevel.GetDynamic() &&
       !( 0 <= m_dyniSubdivisionLevel.GetValue() ) )
    {
      HandleError("Illegal SubdivisionLevel value:  %ld", 
      	  m_dyniSubdivisionLevel.GetValue());
      m_dyniSubdivisionLevel.SetValue(cuiSubdivisionLevel);
      bCorrect = FALSE;
    }

  if ( !m_dyniFacetsMax.GetDynamic() &&
       !( 0 <= m_dyniFacetsMax.GetValue() ) )
    {
      HandleError("Illegal FacetsMax value:  %ld", 
      	  m_dyniFacetsMax.GetValue());
      m_dyniFacetsMax.SetValue(cuiFacetsMax);
      bCorrect = FALSE;
    }

  return bCorrect;
}

BOOL CSLIDESubdivision::Valid()
{
  if ( !CSLIDEGroup::Valid() )
    {
      return FALSE;
    }

  return TRUE;
}

CSLIDECamera *CSLIDESubdivision::FindCamera(CListIterInt *piterIDs, 
					     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtCamera],
	      gcppcSLIDEReservedWords[sliderwtSubdivision], GetID());

  return NULL;
}

CSLIDELight *CSLIDESubdivision::FindLight(CListIterInt *piterIDs, 
				     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtLight],
	      gcppcSLIDEReservedWords[sliderwtSubdivision], GetID());

  return NULL;
}

CSLIDEAreaLight *CSLIDESubdivision::FindAreaLight(CListIterInt *piterIDs, 
					    CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtAreaLight],
	      gcppcSLIDEReservedWords[sliderwtSubdivision], GetID());

  return NULL;
}

VOID CSLIDESubdivision::Preprocess()
{
  CSLIDEGroup::Preprocess();
}

VOID CSLIDESubdivision::CheckDynamic()
{
  SetDynamic(m_dyniShading.GetDynamic() || 
	     m_dyniLOD.GetDynamic() ||
             m_dyniUseDisplayLists.GetDynamic() ||
             m_dyniDrawControlMesh.GetDynamic() ||
             m_dyniDrawCurrentMesh.GetDynamic() ||
             m_dyniDrawVertices.GetDynamic() ||
             m_dyniType.GetDynamic() ||
             m_dyniErrorMetric.GetDynamic() ||
             m_dyniSubdivisionLevel.GetDynamic() ||
             m_dyniFacetsMax.GetDynamic() ||
	     m_dyniWeightedEdges.GetDynamic() ||
	     m_dynfWeight.GetDynamic() ||
	     m_dyniA.GetDynamic() ||
	     m_dyniB.GetDynamic() ||
	     m_dynfTolerance.GetDynamic());
}

VOID CSLIDESubdivision::PreprocessGraph()
{
  if ( !GetVisited() )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;

      BOOL bGeometryEmpty;
      BOOL bGeometryDynamic;
      BOOL bNormalsEmpty;
      BOOL bNormalsDynamic;
      BOOL bBoundsEmpty;
      BOOL bBoundsDynamic;
      BOOL bBoundEmpty;
      BOOL bBoundDynamic;

      BOOL bChildrenBoundsDynamic;

      bGeometryEmpty = TRUE;
      bGeometryDynamic = FALSE;
      bNormalsEmpty = TRUE;
      bNormalsDynamic = FALSE;
      bBoundsEmpty = TRUE;
      bBoundsDynamic = FALSE;

      bBoundEmpty = TRUE;
      bBoundDynamic = FALSE;

      bGeometryDynamic = ( m_dyniType.GetDynamic() ||
			   m_dyniErrorMetric.GetDynamic() ||
			   m_dyniSubdivisionLevel.GetDynamic() ||
			   m_dyniFacetsMax.GetDynamic() ||
			   m_dyniWeightedEdges.GetDynamic() ||
			   m_dynfWeight.GetDynamic() ||
			   m_dyniA.GetDynamic() ||
			   m_dyniB.GetDynamic() ||
			   m_dynfTolerance.GetDynamic() );
      bNormalsDynamic = bGeometryDynamic;

      //////////////////////////////////////////////////
      // Visit Children Nodes and get starting values from them
      iter.Init(&m_listCSLIDEInstance);
      
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pCSLIDEInstance->PreprocessGraph();

	  bGeometryEmpty = ( bGeometryEmpty && 
			     pCSLIDEInstance->GetGeometryEmpty() );
	  bGeometryDynamic = ( bGeometryDynamic ||
			       pCSLIDEInstance->GetGeometryDynamic() );
	  bNormalsEmpty = ( bNormalsEmpty && 
			    pCSLIDEInstance->GetNormalsEmpty() );
	  bNormalsDynamic = ( bNormalsDynamic ||
			      pCSLIDEInstance->GetNormalsDynamic() );
	  bBoundsEmpty = ( bBoundsEmpty && 
			   pCSLIDEInstance->GetBoundsEmpty() );
	  bBoundsDynamic = ( bBoundsDynamic ||
			     pCSLIDEInstance->GetBoundsDynamic() );
	  bBoundEmpty = ( bBoundEmpty && 
			  pCSLIDEInstance->GetBoundEmpty() );
	  bBoundDynamic = ( bBoundDynamic ||
			    pCSLIDEInstance->GetBoundDynamic() );
	}

      //////////////////////////////////////////////////
      // Look at Local flags

      // Update whether we are Empty based on our LOD flag
      if ( !m_dyniLOD.GetDynamic() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      // This node is always off
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	      bBoundsEmpty = TRUE;
	      bBoundEmpty = TRUE;
	    }
	  else if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	    {
	      // This node is always just a bounding box
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	    }
	}

      // Check what is really dynamic

      bChildrenBoundsDynamic = bBoundsDynamic;

      bGeometryDynamic = ( bGeometryDynamic ||
			   m_pCSLIDESurface->GetDynamic() ||
			   m_dyniShading.GetDynamic() ||
			   m_dyniLOD.GetDynamic() );
      bNormalsDynamic = ( bNormalsDynamic ||
			  m_dyniShading.GetDynamic() ||
			  m_dyniLOD.GetDynamic() );
      bBoundsDynamic = ( bBoundsDynamic || 
			 m_dyniLOD.GetDynamic() );
      bBoundDynamic = ( bBoundDynamic || 
			m_dyniLOD.GetDynamic() );

      // If anything is now Empty then it will always be empty so
      // it is not dynamic

      bChildrenBoundsDynamic = ( bChildrenBoundsDynamic && !bBoundsEmpty );

      bGeometryDynamic = ( bGeometryDynamic && !bGeometryEmpty );
      bNormalsDynamic = ( bNormalsDynamic && !bNormalsEmpty );
      bBoundsDynamic = ( bBoundsDynamic && !bBoundsEmpty );
      bBoundDynamic = ( bBoundDynamic && !bBoundEmpty );

      //////////////////////////////////////////////////
      // Now save these correct flag values

      SetChildrenBoundsDynamic(bChildrenBoundsDynamic);

      SetGeometryEmpty(bGeometryEmpty);
      SetGeometryDynamic(bGeometryDynamic);
      SetNormalsEmpty(bNormalsEmpty);
      SetNormalsDynamic(bNormalsDynamic);
      SetBoundsEmpty(bBoundsEmpty);
      SetBoundsDynamic(bBoundsDynamic);
      SetBoundEmpty(bBoundEmpty);
      SetBoundDynamic(bBoundDynamic);
      
      //////////////////////////////
      // START: Jordan Fix This!!!
      //

      UpdateQEMesh();
      UpdateSubdivision();
      UpdateDisplayLists();

      //
      // END: Jordan Fix This!!!
      //////////////////////////////

      //////////////////////////////////////////////////
      // Calculate the Bounding Box (if Empty this will do nothing)
      if ( GetBoundEmpty() )
	{
	  SetBoundValid(TRUE);
	}
      else
	{
	  CalculateBound();
	}
      SetBoundChanged(FALSE);

      //////////////////////////////////////////////////
      // Done Preprocessing this Node in the Graph
      SetVisited(TRUE);
    }
}

VOID CSLIDESubdivision::Update()
{
  SetBoundChanged(FALSE);
  if ( GetDynamic() ||
       GetChildrenBoundsDynamic() )
    {
      if ( GetChildrenBoundsDynamic() )
	{
	  UINT_32 i;
	  for ( i = 0; i < m_uiSubdivisionLevelSize; i++ )
	    {
	      if ( m_ppQEMeshSubdivision[i] != NULL )
		{
		  DeallocateQEMesh(m_ppQEMeshSubdivision[i]);
		  m_ppQEMeshSubdivision[i] = NULL;
		}
	    }
	  m_pQEMesh = NULL;
	  //CalculateQEMesh();
	  //m_ppQEMeshSubdivision[0] = m_pQEMesh;
	  //m_ppQEMeshSubdivision[0] = NULL;
	}
      UpdateState();
    }

  if ( GetBoundDynamic() )
    {
      SetBoundValid(FALSE);
    }

  SetVisited(FALSE);
}

VOID CSLIDESubdivision::UpdateState()
{
  UINT_32 i, uiType, uiErrorMetric;
  UINT_32 uiWeightedEdges, uiFacetsMax;
  FLOAT fWeight;
  FLOAT fTolerance;
  UINT_32 uiA, uiB;
  BOOL bSetKnotValues;
  BOOL b, bNewMesh;

  m_bChanged = FALSE;
  bNewMesh = FALSE;

  CSLIDEGroup::UpdateState();

  m_uiUseDisplayLists = m_dyniUseDisplayLists.GetValue();
  b = m_dyniUseDisplayLists.Update();
  m_bChanged = ( m_bChanged || b );
  b = m_dyniDrawControlMesh.Update();
  m_bChanged = ( m_bChanged || b );
  b = m_dyniDrawCurrentMesh.Update();
  m_bChanged = ( m_bChanged || b );
  b = m_dyniDrawVertices.Update();
  m_bChanged = ( m_bChanged || b );

  uiType = (UINT_32)m_dyniType.GetValue();
  b = m_dyniType.Update();
  m_bChanged = ( m_bChanged || b );
  if ( !( 0 <= m_dyniType.GetValue() &&
	  m_dyniType.GetValue() < slidesubtSize ) )
    {
      HandleError("Illegal Subdivision Type flag value:  %ld", 
		  m_dyniType.GetValue());
      m_dyniType.SetValue(ceType);
    }

  uiErrorMetric = (UINT_32)m_dyniErrorMetric.GetValue();
  b = m_dyniErrorMetric.Update();
  m_bChanged = ( m_bChanged || b );
  if ( !( 0 <= m_dyniErrorMetric.GetValue() &&
	  m_dyniErrorMetric.GetValue() < slideemtSize ) )
    {
      HandleError("Illegal Error Metric flag value:  %ld", 
		  m_dyniErrorMetric.GetValue());
      m_dyniErrorMetric.SetValue(ceErrorMetric);
    }

  uiWeightedEdges = (UINT_32)m_dyniWeightedEdges.GetValue();
  b = m_dyniWeightedEdges.Update();
  m_bChanged = ( m_bChanged || b );
  fWeight = m_dynfWeight.GetValue();
  b = m_dynfWeight.Update();
  m_bChanged = ( m_bChanged || b );

  fTolerance = m_dynfTolerance.GetValue();
  b = m_dynfTolerance.Update();
  m_bChanged = ( m_bChanged || b );
  if ( !( 0 < m_dynfTolerance.GetValue() ) )
    {
      HandleError("Illegal Subdivision Type flag value:  %f", 
		  m_dynfTolerance.GetValue());
      m_dynfTolerance.SetValue(cfTolerance);
    }

  uiFacetsMax = m_dyniFacetsMax.GetValue();
  b = m_dyniFacetsMax.Update();
  m_bChanged = ( m_bChanged || b );
  if ( !( 0 <= m_dyniFacetsMax.GetValue() ) )
    {
      HandleError("Illegal FacetsMax value:  %ld", 
		  m_dyniFacetsMax.GetValue());
      m_dyniFacetsMax.SetValue(cuiFacetsMax);
    }

  uiA = (UINT_32)m_dyniA.GetValue();
  b = m_dyniA.Update();
  m_bChanged = ( m_bChanged || b );
  uiB = (UINT_32)m_dyniB.GetValue();
  b = m_dyniB.Update();
  m_bChanged = ( m_bChanged || b );


  // Number of subdivisions
  b = m_dyniSubdivisionLevel.Update();
  m_bChanged = ( m_bChanged || b );
  switch ( (UINT_32)m_dyniType.GetValue() )
    {
    case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE:
    case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE_LIMIT:
    case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT:
    case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT:
    case slidesubtSLF_SUBDIVISION_CATMULL_CLARK_SELECTIVE:
    case slidesubtSLF_SUBDIVISION_DOO_SABIN_SELECTIVE:
    case slidesubtSLF_SUBDIVISION_DOO_SABIN_SELECTIVE_LIMIT:
    case slidesubtSLF_SUBDIVISION_SQRT_4_SELECTIVE:
    case slidesubtSLF_SUBDIVISION_SQRT_3_SELECTIVE:
    case slidesubtSLF_SUBDIVISION_SQRT_3_SELECTIVE_FACET:
    case slidesubtSLF_SUBDIVISION_PROGRESSIVE:
      {
	if ( 1 < m_dyniSubdivisionLevel.GetValue() )
	  {
	    HandleError("Illegal subdivisionlevel value:  %ld", 
			m_dyniSubdivisionLevel.GetValue());
	    m_dyniSubdivisionLevel.SetValue(1);
	  }
      }
    //break;
    // Intentionally allow it to drop through
    default:
      {
	if ( !( 0 <= m_dyniSubdivisionLevel.GetValue() ) )
	  {
	    HandleError("Illegal subdivisionlevel value:  %ld", 
			m_dyniSubdivisionLevel.GetValue());
	    m_dyniSubdivisionLevel.SetValue(cuiSubdivisionLevel);
	  }
      }
      break;
    }

  // See if we need to grow the mesh array
  if ( (UINT_32)m_dyniSubdivisionLevel.GetValue() + 1 > m_uiSubdivisionLevelSize )
    {
      QEMesh **ppQEMesh;

      ppQEMesh = m_ppQEMeshSubdivision;
      m_ppQEMeshSubdivision = new QEMesh*[m_dyniSubdivisionLevel.GetValue() + 1];
      ASSERT( m_ppQEMeshSubdivision != NULL );
      for ( i = 0; i < m_uiSubdivisionLevelSize; i++ )
	{
	  m_ppQEMeshSubdivision[i] = ppQEMesh[i];
	}
      m_uiSubdivisionLevelSize = m_dyniSubdivisionLevel.GetValue() + 1;
      for ( ; i < m_uiSubdivisionLevelSize; i++ )
	{
	  m_ppQEMeshSubdivision[i] = NULL;	  
	}

      delete []ppQEMesh;
    }

  //////////////////////////////////////////////////
  // Wipe out previous subdivision meshes?
  //
  bSetKnotValues = FALSE;
  /*
  bSetKnotValues = ( uiType != (UINT_32)m_dyniType.GetValue() ||
		     ( ( (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_LOOP ||
			 (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT ||
			 (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT || 
			 (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_DOO_SABIN ||
			 (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_CATMULL_CLARK ||
			 (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_SHARP_CATMULL_CLARK ) &&
		       ( uiWeightedEdges != (UINT_32)m_dyniWeightedEdges.GetValue() ||
			 fWeight != (FLOAT)m_dynfWeight.GetValue() ||
			 uiA != (UINT_32)m_dyniA.GetValue() ||
			 uiB != (UINT_32)m_dyniB.GetValue() ) ) );
			 */

  m_bUnsubdivide = ( ( (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE ||
		     (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE_LIMIT ||
		     //(UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT ||
		     //(UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT ||
		     (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_DOO_SABIN_SELECTIVE ) &&
		   ( fTolerance < (FLOAT)m_dynfTolerance.GetValue() ||
		     uiFacetsMax > (UINT_32)m_dyniFacetsMax.GetValue() ) );

  if ( uiType != (UINT_32)m_dyniType.GetValue() ||

       ( ( (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_LOOP ||
	   (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT ||
	   (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT ||
	   (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_DOO_SABIN ||
	   (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_SHARP_CATMULL_CLARK ||
	   //(UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_CATMULL_CLARK_C2 ||
	   (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_CATMULL_CLARK ) &&
	 ( uiWeightedEdges != (UINT_32)m_dyniWeightedEdges.GetValue() ||
	   fWeight != (FLOAT)m_dynfWeight.GetValue() ||
	   uiA != (UINT_32)m_dyniA.GetValue() ||
	   uiB != (UINT_32)m_dyniB.GetValue() ) ) ||

       ( ( uiType == slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE ||
	   uiType == slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE_LIMIT ||
	   uiType == slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT ||
	   uiType == slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT ||
	   uiType == slidesubtSLF_SUBDIVISION_CATMULL_CLARK_SELECTIVE ||
	   uiType == slidesubtSLF_SUBDIVISION_DOO_SABIN_SELECTIVE ||
	   uiType == slidesubtSLF_SUBDIVISION_DOO_SABIN_SELECTIVE_LIMIT ||
	   uiType == slidesubtSLF_SUBDIVISION_SQRT_4_SELECTIVE ||
	   uiType == slidesubtSLF_SUBDIVISION_SQRT_3_SELECTIVE ||
	   uiType == slidesubtSLF_SUBDIVISION_SQRT_3_SELECTIVE_FACET) && 
	 ( uiErrorMetric != (UINT_32)m_dyniErrorMetric.GetValue() ||
	   ( !( uiType == slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE ||
		uiType == slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE_LIMIT ||
		//uiType == slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT ||
		//uiType == slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT ||
		uiType == slidesubtSLF_SUBDIVISION_DOO_SABIN_SELECTIVE ) &&
	     ( fTolerance < (FLOAT)m_dynfTolerance.GetValue() ||
	       uiFacetsMax > (UINT_32)m_dyniFacetsMax.GetValue() ) ) ) ) )
    {
      m_bUnsubdivide = FALSE;
      for ( i = 1; i < m_uiSubdivisionLevelSize && m_ppQEMeshSubdivision[i] != NULL; i++ )
	{
	  DeallocateQEMesh(m_ppQEMeshSubdivision[i]);
	  m_ppQEMeshSubdivision[i] = NULL;
	}
      m_pQEMesh = m_ppQEMeshSubdivision[0];
      m_ppQEMeshSubdivision[0] = NULL;
    }

  /*
  if ( !gbPreprocessDone )
    {
      return;
    }
  */
}

VOID CSLIDESubdivision::UpdateQEMesh()
{
  if ( m_ppQEMeshSubdivision[0] == NULL )
    {
      if ( m_pQEMesh == NULL )
	{
	  CalculateQEMesh();
	}
      ASSERT( m_pQEMesh != NULL );
      m_ppQEMeshSubdivision[0] = m_pQEMesh;

      // Limit Positions and Normals of the Zeroeth Level Subdivision Mesh
      switch ( (UINT_32)m_dyniType.GetValue() )
	{
	case slidesubtSLF_SUBDIVISION_CATMULL_CLARK:
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionLimitPositionsAndNormalsCatmullClark();
	  break;
	case slidesubtSLF_SUBDIVISION_CATMULL_CLARK_C2:
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionLimitPositionsAndNormalsCatmullClarkC2();
	  break;
	case slidesubtSLF_SUBDIVISION_SHARP_CATMULL_CLARK:
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionBoundaryEdgesSharpnessesSharpCatmullClark();
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionSharpVertexTypesSharpCatmullClark();
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionLimitPositionsAndNormalsSharpCatmullClark();
	  break;
	case slidesubtSLF_SUBDIVISION_DOO_SABIN:
	case slidesubtSLF_SUBDIVISION_DOO_SABIN_SELECTIVE_LIMIT:
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionLimitPositionsAndNormalsDooSabin();
	  break;
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE:
	case slidesubtSLF_SUBDIVISION_CATMULL_CLARK_SELECTIVE:
	case slidesubtSLF_SUBDIVISION_DOO_SABIN_SELECTIVE:
	case slidesubtSLF_SUBDIVISION_SQRT_4_SELECTIVE:
	case slidesubtSLF_SUBDIVISION_SQRT_3_SELECTIVE:
	case slidesubtSLF_SUBDIVISION_SQRT_3_SELECTIVE_FACET:
	case slidesubtSLF_SUBDIVISION_PROGRESSIVE:
	  break;
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_LOOP:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE_LIMIT:
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionLimitPositionsAndNormalsLoop();
	  break;
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_LOOP:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT:
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionBoundaryEdgesSharpnessesSharpCatmullClark();
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionSharpVertexTypesSharpLoop();
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionLimitPositionsAndNormalsSharpLoop();
	  break;
	case slidesubtSLF_SUBDIVISION_GEOMETRIC:
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionLimitPositionsAndNormalsLoop();
	  break;
	case slidesubtSLF_SUBDIVISION_HETEROGENEOUS:
	case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_STAM_LOOP:
	case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_OLD:
        case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_TESSELATE:
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionBoundaryEdgesSharpnessesSharpCatmullClark();
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionSharpVertexTypesSharpCatmullClark();
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionLimitPositionsAndNormalsDefault();
	  break;
	default:
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionLimitPositionsAndNormalsDefault();
	  break;
	}
    }
}

VOID CSLIDESubdivision::UpdateSubdivision()
{
  QEMesh::ErrorMetricType eErrorMetric;
  BOOL bCalculateLimitPositions;
  UINT_32 i;

  bCalculateLimitPositions = FALSE;
  if ( m_ppQEMeshSubdivision[0] == NULL )
    {
      QEMeshIter iterQEMesh;

      CalculateQEMesh();
      m_ppQEMeshSubdivision[0] = m_pQEMesh;
      bCalculateLimitPositions = TRUE;

      BOOL bSetKnotValues;
      bSetKnotValues = FALSE;
      if ( (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_CATMULL_CLARK ||
	   (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_DOO_SABIN ||
	   (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_HETEROGENEOUS ||
	   (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_HETEROGENEOUS_STAM_LOOP ||
	   (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_HETEROGENEOUS_OLD ||
	   (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_HETEROGENEOUS_TESSELATE )
	{
	  bSetKnotValues = TRUE;
	}    
      if ( bSetKnotValues )
	{
	  QEMeshIter iterQEMesh;
	  QEVertexIter iterQEVertex;
	  QEVertex *pvx;
	  QEEdge *pe;
	  UINT_32 i;
	  FLOAT fDefaultKnotInterval;

	  fDefaultKnotInterval = 1.0;
	  if ( (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_LOOP ||
	       (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT ||
	       (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT ||
	       (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_HETEROGENEOUS ||
	       (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_HETEROGENEOUS_STAM_LOOP ||
	       (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_HETEROGENEOUS_OLD ||
	       (UINT_32)m_dyniType.GetValue() == slidesubtSLF_SUBDIVISION_HETEROGENEOUS_TESSELATE )
	    {
	      fDefaultKnotInterval = 0.0;
	    }
	  
	  iterQEMesh.Init(m_ppQEMeshSubdivision[0]);
	  for ( pe = iterQEMesh.PeekFirstEdge();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdge() )
	    {
	      pe->SetKnotInterval(fDefaultKnotInterval);
	    }
	  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdgeBoundary() )
	    {
	      pe->SetKnotInterval(fDefaultKnotInterval);
	    }

	  pvx = iterQEMesh.PeekFirstVertex();
	  if ( pvx != NULL )
	    {
	      iterQEVertex.Init(pvx);
	      pe = iterQEVertex.PeekFirstEdge();
	      ASSERT( pe != NULL );
	    }
	  for ( i = 0; pvx != NULL && i < (UINT_32)m_dyniWeightedEdges.GetValue(); i++ )
	    {
	      ASSERT( pe != NULL );
	      pe->SetKnotInterval((FLOAT)m_dynfWeight.GetValue());
	      switch ( (SLIDESubdivisionType)m_dyniType.GetValue() )
		{
		case slidesubtSLF_SUBDIVISION_CATMULL_CLARK:
		case slidesubtSLF_SUBDIVISION_CATMULL_CLARK_TESSELATE:
		case slidesubtSLF_SUBDIVISION_CATMULL_CLARK_OLD:
		  {
		    pe->GetEdge()->SetKnotInterval((FLOAT)m_dynfWeight.GetValue());
		  }
		  break;
		case slidesubtSLF_SUBDIVISION_SHARP_CATMULL_CLARK:
		case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_LOOP:
		case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT:
		case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT:
		case slidesubtSLF_SUBDIVISION_HETEROGENEOUS:
		case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_STAM_LOOP:
		case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_OLD:
		case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_TESSELATE:
		  {
		    pe->SetKnotInterval(INFINITY);
		    pe->GetEdge()->SetKnotInterval(INFINITY);
		  }
		  break;
		default:
		  {
		  }
		  break;
		}

	      // Advance to next edge
	      pe = iterQEVertex.PeekNextEdge();
	      if ( pe == NULL )
		{
		  pvx = iterQEMesh.PeekNextVertex();
		  if ( pvx != NULL )
		    {
		      iterQEVertex.Init(pvx);
		      pe = iterQEVertex.PeekFirstEdge();
		      ASSERT( pe != NULL );
		    }
		}
	    }
	}

      switch ( (UINT_32)m_dyniType.GetValue() )
	{
	case slidesubtSLF_SUBDIVISION_CATMULL_CLARK:
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionLimitPositionsAndNormalsCatmullClark();
	  break;
	case slidesubtSLF_SUBDIVISION_CATMULL_CLARK_C2:
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionLimitPositionsAndNormalsCatmullClarkC2();
	  break;
	case slidesubtSLF_SUBDIVISION_SHARP_CATMULL_CLARK:
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionBoundaryEdgesSharpnessesSharpCatmullClark();
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionSharpVertexTypesSharpCatmullClark();
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionLimitPositionsAndNormalsSharpCatmullClark();
	  break;
	case slidesubtSLF_SUBDIVISION_DOO_SABIN:
	case slidesubtSLF_SUBDIVISION_DOO_SABIN_SELECTIVE_LIMIT:
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionLimitPositionsAndNormalsDooSabin();
	  break;
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE:
	case slidesubtSLF_SUBDIVISION_CATMULL_CLARK_SELECTIVE:
	case slidesubtSLF_SUBDIVISION_DOO_SABIN_SELECTIVE:
	case slidesubtSLF_SUBDIVISION_SQRT_4_SELECTIVE:
	case slidesubtSLF_SUBDIVISION_SQRT_3_SELECTIVE:
	case slidesubtSLF_SUBDIVISION_SQRT_3_SELECTIVE_FACET:
	case slidesubtSLF_SUBDIVISION_PROGRESSIVE:
	  break;
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_LOOP:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE_LIMIT:
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionLimitPositionsAndNormalsLoop();
	  break;
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_LOOP:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT:
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionBoundaryEdgesSharpnessesSharpCatmullClark();
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionSharpVertexTypesSharpLoop();
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionLimitPositionsAndNormalsSharpLoop();
	  break;
	case slidesubtSLF_SUBDIVISION_GEOMETRIC:
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionLimitPositionsAndNormalsLoop();
	  break;
	case slidesubtSLF_SUBDIVISION_HETEROGENEOUS:
	case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_STAM_LOOP:
	case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_OLD:
        case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_TESSELATE:
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionBoundaryEdgesSharpnessesSharpCatmullClark();
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionSharpVertexTypesSharpCatmullClark();
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionLimitPositionsAndNormalsDefault();
	  break;
	default:
	  m_ppQEMeshSubdivision[0]->CalculateSubdivisionLimitPositionsAndNormalsDefault();
	  break;
	}
    }
  ASSERT( m_ppQEMeshSubdivision[0] != NULL );

  if ( m_ppQEMeshSubdivision[0]->GetBoundaryEdgeCount() > 0 &&
       //(SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_DOO_SABIN &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_LOOP &&
       //(SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT &&
       //(SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_SHARP_CATMULL_CLARK &&
       //(SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_CATMULL_CLARK_C2 &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_CATMULL_CLARK &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_HETEROGENEOUS &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_HETEROGENEOUS_STAM_LOOP &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_HETEROGENEOUS_OLD &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_HETEROGENEOUS_TESSELATE &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_GEOMETRIC &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_GEOMETRIC_TESSELATE )
    {
      m_dyniSubdivisionLevel.SetValue(0);
    }

  eErrorMetric = QEMesh::emtBestFitPlane;
  switch ( m_dyniErrorMetric.GetValue() )
    {
    case slideemtSLF_ERROR_METRIC_BEST_FIT_PLANE:
      {
	eErrorMetric = QEMesh::emtBestFitPlane;
      }
      break;
    case slideemtSLF_ERROR_METRIC_LIMIT_DISTANCE:
      {
	eErrorMetric = QEMesh::emtLimitDistance;
      }
      break;
    default:
      {
	ASSERT( FALSE );
      }
      break;
    }

  switch ( m_dyniType.GetValue() )
    {
    case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE:
      if ( (UINT_32)m_dyniSubdivisionLevel.GetValue() > 0 )
	{
	  if ( m_ppQEMeshSubdivision[1] == NULL )
	    {
	      m_ppQEMeshSubdivision[1] = m_ppQEMeshSubdivision[0]->SubdivideCornerRoundingSelective(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), (UINT_32)m_dyniFacetsMax.GetValue());
	    }
	  else if ( m_bUnsubdivide )
	    {
	      UINT_32 uiFacetsMax;
	      uiFacetsMax = (UINT_32)m_dyniFacetsMax.GetValue();
	      uiFacetsMax += 7;
	      m_ppQEMeshSubdivision[1]->UnSubdivideCornerRoundingSelective(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), uiFacetsMax); 
	    }
	  else
	    {
	      m_ppQEMeshSubdivision[1]->ReSubdivideCornerRoundingSelective(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), (UINT_32)m_dyniFacetsMax.GetValue());
	    }
	}
      break;
    case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE_LIMIT:
      if ( (UINT_32)m_dyniSubdivisionLevel.GetValue() > 0 )
	{
	  if ( m_ppQEMeshSubdivision[1] == NULL )
	    {
	      m_ppQEMeshSubdivision[1] = m_ppQEMeshSubdivision[0]->SubdivideCornerRoundingSelectiveLimit(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), (UINT_32)m_dyniFacetsMax.GetValue());
	    }
	  else if ( m_bUnsubdivide )
	    {
	      UINT_32 uiFacetsMax;
	      uiFacetsMax = (UINT_32)m_dyniFacetsMax.GetValue();
	      uiFacetsMax += 7;
	      m_ppQEMeshSubdivision[1]->UnSubdivideCornerRoundingSelectiveLimit(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), uiFacetsMax); 
	    }
	  else
	    {
	      m_ppQEMeshSubdivision[1]->ReSubdivideCornerRoundingSelectiveLimit(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), (UINT_32)m_dyniFacetsMax.GetValue());
	    }
	}
      break;
    case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT:
      if ( (UINT_32)m_dyniSubdivisionLevel.GetValue() > 0 )
	{
	  if ( m_ppQEMeshSubdivision[1] == NULL )
	    {
	      m_ppQEMeshSubdivision[1] = m_ppQEMeshSubdivision[0]->SubdivideCornerRoundingSharpSelectiveLimit(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), (UINT_32)m_dyniFacetsMax.GetValue());
	    }
	  /*
	  else if ( m_bUnsubdivide )
	    {
	      UINT_32 uiFacetsMax;
	      uiFacetsMax = (UINT_32)m_dyniFacetsMax.GetValue();
	      uiFacetsMax += 7;
	      m_ppQEMeshSubdivision[1]->UnSubdivideCornerRoundingSharpSelectiveLimit(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), uiFacetsMax); 
	    }
	    */
	  else
	    {
	      m_ppQEMeshSubdivision[1]->ReSubdivideCornerRoundingSharpSelectiveLimit(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), (UINT_32)m_dyniFacetsMax.GetValue());
	    }
	}
      break;
    case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT:
      if ( (UINT_32)m_dyniSubdivisionLevel.GetValue() > 0 )
	{
	  if ( m_ppQEMeshSubdivision[1] == NULL )
	    {
	      m_ppQEMeshSubdivision[1] = m_ppQEMeshSubdivision[0]->SubdivideCornerRoundingFacetSharpSelectiveLimit(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), (UINT_32)m_dyniFacetsMax.GetValue());
	    }
	  /*
	  else if ( m_bUnsubdivide )
	    {
	      UINT_32 uiFacetsMax;
	      uiFacetsMax = (UINT_32)m_dyniFacetsMax.GetValue();
	      uiFacetsMax += 7;
	      m_ppQEMeshSubdivision[1]->UnSubdivideCornerRoundingFacetSharpSelectiveLimit(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), uiFacetsMax); 
	    }
	    */
	  else
	    {
	      m_ppQEMeshSubdivision[1]->ReSubdivideCornerRoundingFacetSharpSelectiveLimit(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), (UINT_32)m_dyniFacetsMax.GetValue());
	    }
	}
      break;
    case slidesubtSLF_SUBDIVISION_CATMULL_CLARK_SELECTIVE:
      if ( (UINT_32)m_dyniSubdivisionLevel.GetValue() > 0 )
	{
	  if ( m_ppQEMeshSubdivision[1] == NULL )
	    {
	      m_ppQEMeshSubdivision[1] = m_ppQEMeshSubdivision[0]->SubdivideCatmullClarkSelective(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), (UINT_32)m_dyniFacetsMax.GetValue());
	    }
	  else
	    {
	      m_ppQEMeshSubdivision[1]->ReSubdivideCatmullClarkSelective(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), (UINT_32)m_dyniFacetsMax.GetValue());
	    }
	}
      break;
    case slidesubtSLF_SUBDIVISION_DOO_SABIN_SELECTIVE:
      if ( (UINT_32)m_dyniSubdivisionLevel.GetValue() > 0 )
	{
	  if ( m_ppQEMeshSubdivision[1] == NULL )
	    {
	      m_ppQEMeshSubdivision[1] = m_ppQEMeshSubdivision[0]->SubdivideDooSabinSelective(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), (UINT_32)m_dyniFacetsMax.GetValue());
	    }
	  else if ( m_bUnsubdivide )
	    {
	      UINT_32 uiFacetsMax;
	      uiFacetsMax = (UINT_32)m_dyniFacetsMax.GetValue();
	      uiFacetsMax += 6;
	      m_ppQEMeshSubdivision[1]->UnSubdivideDooSabinSelective(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), uiFacetsMax); 
	    }
	  else
	    {
	      m_ppQEMeshSubdivision[1]->ReSubdivideDooSabinSelective(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), (UINT_32)m_dyniFacetsMax.GetValue());
	    }
	}
      break;
    case slidesubtSLF_SUBDIVISION_DOO_SABIN_SELECTIVE_LIMIT:
      if ( (UINT_32)m_dyniSubdivisionLevel.GetValue() > 0 )
	{
	  if ( m_ppQEMeshSubdivision[1] == NULL )
	    {
	      m_ppQEMeshSubdivision[1] = m_ppQEMeshSubdivision[0]->SubdivideDooSabinSelectiveLimit(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), (UINT_32)m_dyniFacetsMax.GetValue());
	    }
	  else
	    {
	      m_ppQEMeshSubdivision[1]->ReSubdivideDooSabinSelectiveLimit(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), (UINT_32)m_dyniFacetsMax.GetValue());
	    }
	}
      break;
    case slidesubtSLF_SUBDIVISION_SQRT_4_SELECTIVE:
      if ( (UINT_32)m_dyniSubdivisionLevel.GetValue() > 0 )
	{
	  if ( m_ppQEMeshSubdivision[1] == NULL )
	    {
	      m_ppQEMeshSubdivision[1] = m_ppQEMeshSubdivision[0]->SubdivideSqrt4Selective(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), (UINT_32)m_dyniFacetsMax.GetValue());
	    }
	  else
	    {
	      m_ppQEMeshSubdivision[1]->ReSubdivideSqrt4Selective(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), (UINT_32)m_dyniFacetsMax.GetValue());
	    }
	}
      break;
    case slidesubtSLF_SUBDIVISION_SQRT_3_SELECTIVE:
      if ( (UINT_32)m_dyniSubdivisionLevel.GetValue() > 0 )
	{
	  if ( m_ppQEMeshSubdivision[1] == NULL )
	    {
	      m_ppQEMeshSubdivision[1] = m_ppQEMeshSubdivision[0]->SubdivideSqrt3Selective(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), (UINT_32)m_dyniFacetsMax.GetValue());
	    }
	  else
	    {
	      m_ppQEMeshSubdivision[1]->ReSubdivideSqrt3Selective(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), (UINT_32)m_dyniFacetsMax.GetValue());
	    }
	}
      break;
    case slidesubtSLF_SUBDIVISION_SQRT_3_SELECTIVE_FACET:
      if ( (UINT_32)m_dyniSubdivisionLevel.GetValue() > 0 )
	{
	  if ( m_ppQEMeshSubdivision[1] == NULL )
	    {
	      m_ppQEMeshSubdivision[1] = m_ppQEMeshSubdivision[0]->SubdivideSqrt3SelectiveFacet(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), (UINT_32)m_dyniFacetsMax.GetValue());
	    }
	  else
	    {
	      m_ppQEMeshSubdivision[1]->ReSubdivideSqrt3SelectiveFacet(eErrorMetric, (FLOAT)m_dynfTolerance.GetValue(), (UINT_32)m_dyniFacetsMax.GetValue());
	    }
	}
      break;
    case slidesubtSLF_SUBDIVISION_PROGRESSIVE:
      if ( (UINT_32)m_dyniSubdivisionLevel.GetValue() > 0 )
	{
	  if ( m_ppQEMeshSubdivision[1] == NULL )
	    {
	      m_ppQEMeshSubdivision[1] = m_ppQEMeshSubdivision[0]->CreateProgressiveMesh(eErrorMetric);
	      m_ppQEMeshSubdivision[1]->AdaptProgressiveMesh((UINT_32)m_dyniFacetsMax.GetValue());
	    }
	  else
	    {
	      m_ppQEMeshSubdivision[1]->AdaptProgressiveMesh((UINT_32)m_dyniFacetsMax.GetValue());
	    }
	}
      break;
    default:
      for ( i = 1; i <= (UINT_32)m_dyniSubdivisionLevel.GetValue(); i++ )
	{
	  if ( m_ppQEMeshSubdivision[i] == NULL )
	    {
	      // Subdivide the mesh
	      switch ( m_dyniType.GetValue() )
		{
		case slidesubtSLF_SUBDIVISION_DOO_SABIN:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideDooSabin();
		  break;

		case slidesubtSLF_SUBDIVISION_CATMULL_CLARK:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideCatmullClark();
		  break;
		case slidesubtSLF_SUBDIVISION_CATMULL_CLARK_C2:
		  if ( i < 3 ) // Stationary implementation
		  //if ( i < 2 ) // Non-stationary implementation
		    {
		      m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideSharpCatmullClark();
		    }
		  else
		    {
		      // Stationary implementation
		      m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideCatmullClarkC2(-1);
		      // Non-stationary implementation
		      //m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideCatmullClarkC2(i-2);
		    }
		  break;
		case slidesubtSLF_SUBDIVISION_SHARP_CATMULL_CLARK:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideSharpCatmullClark();
		  break;
		case slidesubtSLF_SUBDIVISION_CATMULL_CLARK_TESSELATE:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideCatmullClarkTesselate();
		  break;
		case slidesubtSLF_SUBDIVISION_CATMULL_CLARK_INTERPOLATE:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideCatmullClarkInterpolate();
		  break;
		case slidesubtSLF_SUBDIVISION_CATMULL_CLARK_OLD:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideCatmullClarkFacetCentric();
		  break;

		case slidesubtSLF_SUBDIVISION_CORNER_CUTTING:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideCornerCutting();
		  break;

		case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_TESSELATE:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideCornerRoundingTesselate();
		  break;
		case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideCornerRounding();
		  break;
		case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_LOOP:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideCornerRoundingLoop();
		  break;
		case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_LOOP:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideCornerRoundingSharpLoop();
		  break;
		case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_BUTTERFLY:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideCornerRoundingButterfly();
		  break;
		case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_ZORIN:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideCornerRoundingZorin();
		  break;

		case slidesubtSLF_SUBDIVISION_TRIANGLE_TESSELATE:
		  if ( i == 1 )
		    {
		      m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideTriangulate();
		    }
		  else
		    {
		      m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideTriangleTesselate();
		    }
		  break;
		case slidesubtSLF_SUBDIVISION_TRIANGLE_LOOP:
		  if ( i == 1 )
		    {
		      m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideTriangulate();
		    }
		  else
		    {
		      m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideTriangleLoop();
		    }
		  break;
		case slidesubtSLF_SUBDIVISION_TRIANGLE_LOOP_EDGE_RULE:
		  if ( i == 1 )
		    {
		      m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideTriangulate();
		    }
		  else
		    {
		      m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideTriangleLoopEdgeRule();
		    }
		  break;
		case slidesubtSLF_SUBDIVISION_TRIANGLE_BUTTERFLY:
		  if ( i == 1 )
		    {
		      m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideTriangulate();
		    }
		  else
		    {
		      m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideTriangleButterfly();
		    }
		  break;
		case slidesubtSLF_SUBDIVISION_TRIANGLE_ZORIN:
		  if ( i == 1 )
		    {
		      m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideTriangulate();
		    }
		  else
		    {
		      m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideTriangleZorin();
		    }
		  break;
		case slidesubtSLF_SUBDIVISION_TRIANGLE_SPHERE:
		  if ( i == 1 )
		    {
		      m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideTriangulate();
		    }
		  else
		    {
		      m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideTriangleSphere();
		    }
		  break;

		case slidesubtSLF_SUBDIVISION_SQRT_4:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideSqrt4();
		  break;

		case slidesubtSLF_SUBDIVISION_SQRT_3:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideSqrt3();
		  break;

		case slidesubtSLF_SUBDIVISION_SQRT_3_SPHERE_BARYCENTER:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideSqrt3SphereBarycenter();
		  break;
		case slidesubtSLF_SUBDIVISION_SQRT_3_SPHERE_CENTER:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideSqrt3SphereCenter();
		  break;
		case slidesubtSLF_SUBDIVISION_SQRT_3_SPHERE_RADIUS:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideSqrt3SphereRadius();
		  break;
		case slidesubtSLF_SUBDIVISION_SQRT_3_SPHERE_CAP_AREA:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideSqrt3SphereCapArea();
		  break;
		case slidesubtSLF_SUBDIVISION_SQRT_3_SPHERE_TRIANGLE_AREA:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideSqrt3SphereTriangleArea();
		  break;
		case slidesubtSLF_SUBDIVISION_SQRT_3_SPHERE_TANGENT:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideSqrt3SphereTangent();
		  break;

		case slidesubtSLF_SUBDIVISION_HETEROGENEOUS:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideHeterogeneous();
		  break;
		case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_STAM_LOOP:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideHeterogeneousStamLoop();
		  break;
		case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_OLD:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideHeterogeneousOld();
		  break;
		case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_TESSELATE:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideHeterogeneousTesselate();
		  break;

		case slidesubtSLF_SUBDIVISION_PENTAGON:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdividePentagon();
		  break;
		case slidesubtSLF_SUBDIVISION_PENTAGON_TESSELATE:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdividePentagonTesselate();
		  break;

		case slidesubtSLF_SUBDIVISION_GEOMETRIC:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideGeometricBiarc();
		  break;
		case slidesubtSLF_SUBDIVISION_GEOMETRIC_TESSELATE:
		  m_ppQEMeshSubdivision[i] = m_ppQEMeshSubdivision[i-1]->SubdivideGeometricTesselate();
		  break;

		default:
		  ASSERT( FALSE );
		  break;
		}
	      
	      ASSERT( m_ppQEMeshSubdivision[i] != NULL );	  
	    }
	}
      break;
    }

  m_pQEMesh = m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()];
}

VOID CSLIDESubdivision::UpdateDisplayLists()
{
  if ( m_uiUseDisplayLists != m_dyniUseDisplayLists.GetValue() &&
       !m_dyniUseDisplayLists.GetValue() )
    {
      BOOL bDynamic;
      bDynamic = GetDynamic();

      SetDynamic(FALSE);

      SetPrerenderUnvisited();

      SetDynamic(bDynamic);
    }

  if ( m_bChanged && m_dyniUseDisplayLists.GetValue() )
    {
      BOOL bDynamic;
      bDynamic = GetDynamic();
      BOOL bGeometryDynamic;
      bGeometryDynamic = GetGeometryDynamic();
      BOOL bNormalsDynamic;
      bNormalsDynamic = GetNormalsDynamic();
      BOOL bBoundsDynamic;
      bBoundsDynamic = GetBoundsDynamic();

      SetDynamic(FALSE);
      SetGeometryDynamic(FALSE);
      SetNormalsDynamic(FALSE);
      SetBoundsDynamic(FALSE);

      SetPrerenderUnvisited();
      Prerender();
      
      SetDynamic(bDynamic);
      SetGeometryDynamic(bGeometryDynamic);
      SetNormalsDynamic(bNormalsDynamic);
      SetBoundsDynamic(bBoundsDynamic);
    }
}

VOID CSLIDESubdivision::UpdateGraph()
{
  if ( !GetVisited() )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;
      BOOL bBoundChanged;

      if ( !GetBoundValid() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      if ( !m_bound.GetEmpty() )
		{
		  SetBoundChanged(TRUE);
		  m_bound.Null();
		}
	      return;
	    }
	  
	  bBoundChanged = FALSE;
	  
	  iter.Init(&m_listCSLIDEInstance);
	  
	  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
		pCSLIDEInstance != NULL;
		pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	    {
	      pCSLIDEInstance->UpdateGraph();
	      bBoundChanged = bBoundChanged || 
		pCSLIDEInstance->GetBoundChanged();
	    }
	  
	  UpdateQEMesh();

	  if ( bBoundChanged )
	    {
	      CalculateBound();
	    }
	  
	  SetBoundValid(TRUE);
	}

      UpdateSubdivision();
      UpdateDisplayLists();

      SetVisited(TRUE);
    }
}

VOID CSLIDESubdivision::CalculateBound()
{
  CBound bound;

/*
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  
  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->InsertIntoBound(bound);
    }
*/
  ///////////////////////////////////////
  // START: Jordan Fix This!!!
  //
  {
    QEMeshIter iterQEMesh;
    QEVertex *pvx;
    Point pt;

    iterQEMesh.Init(m_pQEMesh);
    for ( pvx = iterQEMesh.PeekFirstVertex();
	  pvx != NULL;
	  pvx = iterQEMesh.PeekNextVertex() )
      {
	pt = pvx->GetPoint();
	bound.Insert(pt);
      }
  }
  //
  // END: Jordan Fix This!!!
  ///////////////////////////////////////

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);
    }
}

VOID CSLIDESubdivision::PrerenderGeometry(SLIDEShadingType eShading)
{
  CListIterInt iter;
  //CSLIDEInstance *pCSLIDEInstance;
  BOOL bFaceted, bLimitSurface, bCalculateNormals, bSelective;

  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  PushShadingAttributes(eShading);
  {
    m_pCSLIDESurface->Push();
    {
      bLimitSurface = FALSE;
      bCalculateNormals = FALSE;
      bSelective = FALSE;
      switch ( (UINT_32)m_dyniType.GetValue() )
	{
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT:
	  bSelective = TRUE;
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_LOOP:
	case slidesubtSLF_SUBDIVISION_SHARP_CATMULL_CLARK:
	  bCalculateNormals = TRUE;
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_LOOP:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE_LIMIT:
	case slidesubtSLF_SUBDIVISION_CATMULL_CLARK:
	case slidesubtSLF_SUBDIVISION_CATMULL_CLARK_C2:
	case slidesubtSLF_SUBDIVISION_DOO_SABIN:
	case slidesubtSLF_SUBDIVISION_DOO_SABIN_SELECTIVE_LIMIT:
	case slidesubtSLF_SUBDIVISION_SQRT_3_SELECTIVE_FACET:
	  bLimitSurface = TRUE;
	  break;
	default:
	  break;
	}
      bFaceted = ( !(eShading == slideshtSLF_GOURAUD ||
		     eShading == slideshtSLF_PHONG) );
      ::RenderGeometry(m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()], 
		       bFaceted, bLimitSurface, bCalculateNormals, bSelective);
    }
    m_pCSLIDESurface->Pop();
  }
  PopShadingAttributes(eShading);

  if ( m_dyniDrawControlMesh.GetValue() )
    {
      // Initial Control Mesh
      QEMeshIter iterQEMesh;
      QEEdge *pe;

      switch ( (UINT_32)m_dyniType.GetValue() )
	{
	case slidesubtSLF_SUBDIVISION_SHARP_CATMULL_CLARK:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_LOOP:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT:
	case slidesubtSLF_SUBDIVISION_HETEROGENEOUS:
	case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_STAM_LOOP:
	case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_OLD:
	case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_TESSELATE:
	  PushAuxillaryGeometryAttributes();
	  {	
	    iterQEMesh.Init(m_ppQEMeshSubdivision[0]);
	    gilBegin(GL_LINES);
	    {
	      FLOAT fSharpness;
	      BOOL bUseWeightValue;

	      bUseWeightValue = ((UINT_32)m_dyniA.GetValue() != 0);
	      for ( pe = iterQEMesh.PeekFirstEdge();
		    pe != NULL;
		    pe = iterQEMesh.PeekNextEdge() )
		{
		  fSharpness = pe->GetKnotInterval();
		  if ( fSharpness == -1.0 )
		    {
		      if ( bUseWeightValue )
			{
			  fSharpness = (FLOAT)m_dynfWeight.GetValue();
			}
		      else
			{
			  fSharpness = INFINITY;
			}
		    }

		  if ( fSharpness == 0.0 )
		    {
		      gilColor3f(1.0, 1.0, 0.0);
		    }
		  else if ( fSharpness == INFINITY )
		    {
		      gilColor3f(1.0, 0.0, 0.0);
		    }
		  else
		    {
		      gilColor3f(0.0, 1.0, 0.0);
		    }
		  gilVertex3fv(pe->GetVertex()->GetPoint().Get());
		  gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
		}
	      for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		    pe != NULL;
		    pe = iterQEMesh.PeekNextEdgeBoundary() )
		{
		  fSharpness = pe->GetKnotInterval();
		  if ( fSharpness == -1.0 )
		    {
		      if ( bUseWeightValue )
			{
			  fSharpness = (FLOAT)m_dynfWeight.GetValue();
			}
		      else
			{
			  fSharpness = INFINITY;
			}
		    }

		  if ( fSharpness == 0.0 )
		    {
		      gilColor3f(1.0, 1.0, 0.0);
		    }
		  else if ( fSharpness == INFINITY )
		    {
		      gilColor3f(1.0, 0.0, 0.0);
		    }
		  else
		    {
		      gilColor3f(0.0, 1.0, 0.0);
		    }
		  gilVertex3fv(pe->GetVertex()->GetPoint().Get());
		  gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
		}
	    }
	    gilEnd();
	  }
	  PopAuxillaryGeometryAttributes();
	  break;
	default:
	  PushAuxillaryGeometryAttributes();
	  {	
	    iterQEMesh.Init(m_ppQEMeshSubdivision[0]);
	    gilColor3f(1.0, 1.0, 0.0);
	    gilBegin(GL_LINES);
	    {
	      for ( pe = iterQEMesh.PeekFirstEdge();
		    pe != NULL;
		    pe = iterQEMesh.PeekNextEdge() )
		{
		  gilVertex3fv(pe->GetVertex()->GetPoint().Get());
		  gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
		}
	      for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		    pe != NULL;
		    pe = iterQEMesh.PeekNextEdgeBoundary() )
		{
		  gilVertex3fv(pe->GetVertex()->GetPoint().Get());
		  gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
		}
	    }
	    gilEnd();
	  }
	  PopAuxillaryGeometryAttributes();
	  break;
	}
    }

  if ( m_dyniDrawCurrentMesh.GetValue() )
    {
      // Current Control Mesh
      QEMeshIter iterQEMesh;
      QEEdge *pe;

      switch ( (UINT_32)m_dyniType.GetValue() )
	{
	case slidesubtSLF_SUBDIVISION_SHARP_CATMULL_CLARK:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_LOOP:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT:
	case slidesubtSLF_SUBDIVISION_HETEROGENEOUS:
	case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_STAM_LOOP:
	case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_OLD:
	case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_TESSELATE:
	  PushAuxillaryGeometryAttributes();
	  {	
	    iterQEMesh.Init(m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()]);
	    gilBegin(GL_LINES);
	    {
	      FLOAT fSharpness;
	      BOOL bUseWeightValue;

	      bUseWeightValue = ((UINT_32)m_dyniA.GetValue() != 0);
	      for ( pe = iterQEMesh.PeekFirstEdge();
		    pe != NULL;
		    pe = iterQEMesh.PeekNextEdge() )
		{
		  fSharpness = pe->GetKnotInterval();
		  if ( fSharpness == -1.0 )
		    {
		      if ( bUseWeightValue )
			{
			  fSharpness = (FLOAT)m_dynfWeight.GetValue();
			}
		      else
			{
			  fSharpness = INFINITY;
			}
		    }

		  if ( fSharpness == 0.0 )
		    {
		      gilColor3f(0.0, 1.0, 1.0);
		    }
		  else if ( fSharpness == INFINITY )
		    {
		      gilColor3f(1.0, 0.0, 0.0);
		    }
		  else
		    {
		      gilColor3f(0.0, 1.0, 0.0);
		    }
		  gilVertex3fv(pe->GetVertex()->GetPoint().Get());
		  gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
		}
	      for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		    pe != NULL;
		    pe = iterQEMesh.PeekNextEdgeBoundary() )
		{
		  fSharpness = pe->GetKnotInterval();
		  if ( fSharpness == -1.0 )
		    {
		      if ( bUseWeightValue )
			{
			  fSharpness = (FLOAT)m_dynfWeight.GetValue();
			}
		      else
			{
			  fSharpness = INFINITY;
			}
		    }

		  if ( fSharpness == 0.0 )
		    {
		      gilColor3f(1.0, 1.0, 0.0);
		    }
		  else if ( fSharpness == INFINITY )
		    {
		      gilColor3f(1.0, 0.0, 0.0);
		    }
		  else
		    {
		      gilColor3f(0.0, 1.0, 0.0);
		    }
		  gilVertex3fv(pe->GetVertex()->GetPoint().Get());
		  gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
		}
	    }
	    gilEnd();
	  }
	  PopAuxillaryGeometryAttributes();
	  break;
	default:
	  PushAuxillaryGeometryAttributes();
	  {	
	    iterQEMesh.Init(m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()]);
	    gilColor3f(0.0, 1.0, 1.0);
	    gilBegin(GL_LINES);
	    {
	      for ( pe = iterQEMesh.PeekFirstEdge();
		    pe != NULL;
		    pe = iterQEMesh.PeekNextEdge() )
		{
		  gilVertex3fv(pe->GetVertex()->GetPoint().Get());
		  gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
		}
	      for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		    pe != NULL;
		    pe = iterQEMesh.PeekNextEdgeBoundary() )
		{
		  gilVertex3fv(pe->GetVertex()->GetPoint().Get());
		  gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
		}
	    }
	    gilEnd();
	  }
	  PopAuxillaryGeometryAttributes();
	  break;
	}
    }

  if ( m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()]->GetBoundaryEdgeCount() > 0 &&
       //(SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_DOO_SABIN &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_LOOP &&
       //(SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT &&
       //(SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_SHARP_CATMULL_CLARK &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_CATMULL_CLARK &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_HETEROGENEOUS &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_HETEROGENEOUS_STAM_LOOP &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_HETEROGENEOUS_OLD &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_HETEROGENEOUS_TESSELATE &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_GEOMETRIC &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_GEOMETRIC_TESSELATE )
    {
      QEMeshIter iterQEMesh;
      QEEdge *pe;

      iterQEMesh.Init(m_ppQEMeshSubdivision[0]);

      PushAuxillaryGeometryAttributes();
      {
	gilColor3f(1.0, 0.0, 0.0);
	gilBegin(GL_LINES);
	{
	  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdgeBoundary() )
	    {
	      gilVertex3fv(pe->GetVertex()->GetPoint().Get());
	      gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
	    }
	}
	gilEnd();
      }
      PopAuxillaryGeometryAttributes();
    }

  if ( m_dyniDrawVertices.GetValue() &&
       m_dyniSubdivisionLevel.GetValue() > 0 )
    {
      switch ( (SLIDESubdivisionType)(m_dyniType.GetValue()) )
	{
	case slidesubtSLF_SUBDIVISION_SQRT_3_SELECTIVE_FACET:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT:
	  {
	    QEMeshIter iterQEMesh;
	    QEFacetIter iterQEFacet;
	    QEFacet *pfa;
	    QEVertex *pvx;

	    iterQEMesh.Init(m_ppQEMeshSubdivision[1]);

	    PushAuxillaryGeometryAttributes();
	    //PushShadingAttributes(slideshtSLF_FLAT);
	    {
	      for ( pfa = iterQEMesh.PeekFirstFacet();
		    pfa != NULL;
		    pfa = iterQEMesh.PeekNextFacet() )
		{
		  glColor3fv(s_ppfVertexColors[pfa->GetGeneration() & 0x7]);
		  glBegin(GL_TRIANGLE_FAN);
		  {
		    iterQEFacet.Init(pfa);
		    for ( pvx = iterQEFacet.PeekFirstVertex();
			  pvx != NULL;
			  pvx = iterQEFacet.PeekNextVertex() )
		      {
			glVertex3fv(pvx->GetPoint().Get());
		      }
		  }
		  glEnd();
		}
	    }
	    //PopShadingAttributes(slideshtSLF_FLAT);
	    PopAuxillaryGeometryAttributes();
	  }
	// Intentionally allow it to fall through
	//break;
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE_LIMIT:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT:
	case slidesubtSLF_SUBDIVISION_CATMULL_CLARK_SELECTIVE:
	case slidesubtSLF_SUBDIVISION_DOO_SABIN_SELECTIVE:
	case slidesubtSLF_SUBDIVISION_DOO_SABIN_SELECTIVE_LIMIT:
	case slidesubtSLF_SUBDIVISION_SQRT_4_SELECTIVE:
	case slidesubtSLF_SUBDIVISION_SQRT_3_SELECTIVE:
	  {
	    QEMeshIter iterQEMesh;
	    QEVertex *pvx;

	    iterQEMesh.Init(m_ppQEMeshSubdivision[1]);

	    PushAuxillaryGeometryAttributes();
	    {
	      glPointSize(3.0);
	      glBegin(GL_POINTS);
	      {
		for ( pvx = iterQEMesh.PeekFirstVertex();
		      pvx != NULL;
		      pvx = iterQEMesh.PeekNextVertex() )
		  {
		    glColor3fv(s_ppfVertexColors[pvx->GetGeneration() & 0x7]);
		    glVertex3fv(pvx->GetPoint().Get());
		  }
	      }
	      glEnd();
	    }
	    PopAuxillaryGeometryAttributes();
	  }
	break;
	default:
	break;
	}
    }
}

VOID CSLIDESubdivision::PrerenderNormals(SLIDEShadingType eShading)
{
  CListIterInt iter;
  //CSLIDEInstance *pCSLIDEInstance;
  BOOL bFaceted, bLimitSurface, bCalculateNormals, bSelective;

  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  bLimitSurface = FALSE;
  bCalculateNormals = FALSE;
  bSelective = FALSE;
  switch ( (UINT_32)m_dyniType.GetValue() )
    {
    case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT:
    case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT:
      bSelective = TRUE;
    case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_LOOP:
    case slidesubtSLF_SUBDIVISION_SHARP_CATMULL_CLARK:
      bCalculateNormals = TRUE;
    case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_LOOP:
    case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE_LIMIT:
    case slidesubtSLF_SUBDIVISION_CATMULL_CLARK:
    case slidesubtSLF_SUBDIVISION_CATMULL_CLARK_C2:
    case slidesubtSLF_SUBDIVISION_DOO_SABIN:
    case slidesubtSLF_SUBDIVISION_DOO_SABIN_SELECTIVE_LIMIT:
    case slidesubtSLF_SUBDIVISION_SQRT_3_SELECTIVE_FACET:
      bLimitSurface = TRUE;
      break;
    case slidesubtSLF_SUBDIVISION_GEOMETRIC:
      bCalculateNormals = TRUE;
      break;
    default:
      break;
    }
  bFaceted = ( !(eShading == slideshtSLF_GOURAUD ||
		 eShading == slideshtSLF_PHONG) );
  ::RenderNormals(m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()], 
		  bFaceted, bLimitSurface, bCalculateNormals, bSelective);
}

VOID CSLIDESubdivision::PrerenderBounds()
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
    {
    }
  else 
    {
      if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	{
	}
      else
	{
	  iter.Init(&m_listCSLIDEInstance);
	  
	  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
		pCSLIDEInstance != NULL;
		pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	    {
	      pCSLIDEInstance->PrerenderBounds();
	    }
	}

      m_bound.Render();
    }
}

VOID CSLIDESubdivision::Prerender()
{
  if ( !(!m_dyniLOD.GetDynamic() && 
	 ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
	   m_dyniLOD.GetValue() == slidelodtSLF_BOUND )) )
    {
      CListIterInt iter;
      CSLIDEInstance *pCSLIDEInstance;
      
      iter.Init(&m_listCSLIDEInstance);
      
      for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	{
	  pCSLIDEInstance->Prerender();
	}
    }
      
  CSLIDEBoundedNode::Prerender();
}

VOID CSLIDESubdivision::RenderGeometry(const Matrix &mGproj_obj, 
					CBound *pCBoundClip, 
					SLIDEShadingType eShading,
					SLIDELODType eLOD,
				        CSLIDEWormwarp *pCSLIDEWarp /* = NULL */)
{

  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  BOOL bLimitSurface, bFaceted, bCalculateNormals, bSelective;

  // For Debugging
  if ( GetRenderCDAOperandsEnabled() )
    {
      gilPushAttrib(GL_ALL_ATTRIB_BITS);
      {
	PushAuxillaryGeometryAttributes();
	{
	  PushWireFrameAttributes();
	  {
	    iter.Init(&m_listCSLIDEInstance);
	    
	    for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
		  pCSLIDEInstance != NULL;
		  pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
	      {
		if (NULL == pCSLIDEWarp)
		  {
		    pCSLIDEInstance->Render(mGproj_obj, pCBoundClip, slideshtSLF_WIRE, eLOD);
		  }
		else
		  {
		    pCSLIDEInstance->RenderGeometry(mGproj_obj, pCBoundClip, slideshtSLF_WIRE, eLOD, pCSLIDEWarp);
		  }
	      }
	  }
	  PopWireFrameAttributes();
	}
	PopAuxillaryGeometryAttributes();
      }
      gilPopAttrib();
    }
      
  bLimitSurface = FALSE;
  bCalculateNormals = FALSE;
  bSelective = FALSE;
  switch ( (UINT_32)m_dyniType.GetValue() )
    {
    case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT:
    case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT:
      bSelective = TRUE;
    case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_LOOP:
    case slidesubtSLF_SUBDIVISION_SHARP_CATMULL_CLARK:
      bCalculateNormals = TRUE;
    case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_LOOP:
    case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE_LIMIT:
    case slidesubtSLF_SUBDIVISION_CATMULL_CLARK:
    case slidesubtSLF_SUBDIVISION_CATMULL_CLARK_C2:
    case slidesubtSLF_SUBDIVISION_DOO_SABIN:
    case slidesubtSLF_SUBDIVISION_DOO_SABIN_SELECTIVE_LIMIT:
    case slidesubtSLF_SUBDIVISION_SQRT_3_SELECTIVE_FACET:
      bLimitSurface = TRUE;
      break;
    default:
      break;
    }
  bFaceted = ( !(eShading == slideshtSLF_GOURAUD ||
		 eShading == slideshtSLF_PHONG) );
  ::RenderGeometry(m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()], 
		   bFaceted, bLimitSurface, bCalculateNormals, bSelective);

  if ( m_dyniDrawControlMesh.GetValue() )
    {
      // Initial Control Mesh
      QEMeshIter iterQEMesh;
      QEEdge *pe;

      switch ( (UINT_32)m_dyniType.GetValue() )
	{
	case slidesubtSLF_SUBDIVISION_SHARP_CATMULL_CLARK:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_LOOP:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT:
	case slidesubtSLF_SUBDIVISION_HETEROGENEOUS:
	case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_STAM_LOOP:
	case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_OLD:
	case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_TESSELATE:
	  PushAuxillaryGeometryAttributes();
	  {	
	    iterQEMesh.Init(m_ppQEMeshSubdivision[0]);
	    gilBegin(GL_LINES);
	    {
	      FLOAT fSharpness;
	      BOOL bUseWeightValue;

	      bUseWeightValue = ((UINT_32)m_dyniA.GetValue() != 0);
	      for ( pe = iterQEMesh.PeekFirstEdge();
		    pe != NULL;
		    pe = iterQEMesh.PeekNextEdge() )
		{
		  fSharpness = pe->GetKnotInterval();
		  if ( fSharpness == -1.0 )
		    {
		      if ( bUseWeightValue )
			{
			  fSharpness = (FLOAT)m_dynfWeight.GetValue();
			}
		      else
			{
			  fSharpness = INFINITY;
			}
		    }

		  if ( fSharpness == 0.0 )
		    {
		      gilColor3f(1.0, 1.0, 0.0);
		    }
		  else if ( fSharpness == INFINITY )
		    {
		      gilColor3f(1.0, 0.0, 0.0);
		    }
		  else
		    {
		      gilColor3f(0.0, 1.0, 0.0);
		    }
		  gilVertex3fv(pe->GetVertex()->GetPoint().Get());
		  gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
		}
	      for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		    pe != NULL;
		    pe = iterQEMesh.PeekNextEdgeBoundary() )
		{
		  fSharpness = pe->GetKnotInterval();
		  if ( fSharpness == -1.0 )
		    {
		      if ( bUseWeightValue )
			{
			  fSharpness = (FLOAT)m_dynfWeight.GetValue();
			}
		      else
			{
			  fSharpness = INFINITY;
			}
		    }

		  if ( fSharpness == 0.0 )
		    {
		      gilColor3f(1.0, 1.0, 0.0);
		    }
		  else if ( fSharpness == INFINITY )
		    {
		      gilColor3f(1.0, 0.0, 0.0);
		    }
		  else
		    {
		      gilColor3f(0.0, 1.0, 0.0);
		    }
		  gilVertex3fv(pe->GetVertex()->GetPoint().Get());
		  gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
		}
	    }
	    gilEnd();
	  }
	  PopAuxillaryGeometryAttributes();
	  break;
	default:
	  PushAuxillaryGeometryAttributes();
	  {	
	    iterQEMesh.Init(m_ppQEMeshSubdivision[0]);
	    gilColor3f(1.0, 1.0, 0.0);
	    gilBegin(GL_LINES);
	    {
	      for ( pe = iterQEMesh.PeekFirstEdge();
		    pe != NULL;
		    pe = iterQEMesh.PeekNextEdge() )
		{
		  gilVertex3fv(pe->GetVertex()->GetPoint().Get());
		  gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
		}
	      for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		    pe != NULL;
		    pe = iterQEMesh.PeekNextEdgeBoundary() )
		{
		  gilVertex3fv(pe->GetVertex()->GetPoint().Get());
		  gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
		}
	    }
	    gilEnd();
	  }
	  PopAuxillaryGeometryAttributes();
	  break;
	}
    }

  if ( m_dyniDrawCurrentMesh.GetValue() )
    {
      // Current Control Mesh
      QEMeshIter iterQEMesh;
      QEEdge *pe;

      switch ( (UINT_32)m_dyniType.GetValue() )
	{
	case slidesubtSLF_SUBDIVISION_SHARP_CATMULL_CLARK:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_LOOP:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT:
	case slidesubtSLF_SUBDIVISION_HETEROGENEOUS:
	case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_STAM_LOOP:
	case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_OLD:
	case slidesubtSLF_SUBDIVISION_HETEROGENEOUS_TESSELATE:
	  PushAuxillaryGeometryAttributes();
	  {	
	    iterQEMesh.Init(m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()]);
	    gilBegin(GL_LINES);
	    {
	      FLOAT fSharpness;
	      BOOL bUseWeightValue;

	      bUseWeightValue = ((UINT_32)m_dyniA.GetValue() != 0);
	      for ( pe = iterQEMesh.PeekFirstEdge();
		    pe != NULL;
		    pe = iterQEMesh.PeekNextEdge() )
		{
		  fSharpness = pe->GetKnotInterval();
		  if ( fSharpness == -1.0 )
		    {
		      if ( bUseWeightValue )
			{
			  fSharpness = (FLOAT)m_dynfWeight.GetValue();
			}
		      else
			{
			  fSharpness = INFINITY;
			}
		    }

		  if ( fSharpness == 0.0 )
		    {
		      gilColor3f(0.0, 1.0, 1.0);
		    }
		  else if ( fSharpness == INFINITY )
		    {
		      gilColor3f(1.0, 0.0, 0.0);
		    }
		  else
		    {
		      gilColor3f(0.0, 1.0, 0.0);
		    }
		  gilVertex3fv(pe->GetVertex()->GetPoint().Get());
		  gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
		}
	      for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		    pe != NULL;
		    pe = iterQEMesh.PeekNextEdgeBoundary() )
		{
		  fSharpness = pe->GetKnotInterval();
		  if ( fSharpness == -1.0 )
		    {
		      if ( bUseWeightValue )
			{
			  fSharpness = (FLOAT)m_dynfWeight.GetValue();
			}
		      else
			{
			  fSharpness = INFINITY;
			}
		    }

		  if ( fSharpness == 0.0 )
		    {
		      gilColor3f(1.0, 1.0, 0.0);
		    }
		  else if ( fSharpness == INFINITY )
		    {
		      gilColor3f(1.0, 0.0, 0.0);
		    }
		  else
		    {
		      gilColor3f(0.0, 1.0, 0.0);
		    }
		  gilVertex3fv(pe->GetVertex()->GetPoint().Get());
		  gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
		}
	    }
	    gilEnd();
	  }
	  PopAuxillaryGeometryAttributes();
	  break;
	default:
	  PushAuxillaryGeometryAttributes();
	  {	
	    iterQEMesh.Init(m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()]);
	    gilColor3f(0.0, 1.0, 1.0);
	    gilBegin(GL_LINES);
	    {
	      for ( pe = iterQEMesh.PeekFirstEdge();
		    pe != NULL;
		    pe = iterQEMesh.PeekNextEdge() )
		{
		  gilVertex3fv(pe->GetVertex()->GetPoint().Get());
		  gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
		}
	      for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		    pe != NULL;
		    pe = iterQEMesh.PeekNextEdgeBoundary() )
		{
		  gilVertex3fv(pe->GetVertex()->GetPoint().Get());
		  gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
		}
	    }
	    gilEnd();
	  }
	  PopAuxillaryGeometryAttributes();
	  break;
	}
    }

  if ( m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()]->GetBoundaryEdgeCount() > 0 &&
       //(SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_DOO_SABIN &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_LOOP &&
       //(SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT &&
       //(SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_SHARP_CATMULL_CLARK &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_CATMULL_CLARK &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_HETEROGENEOUS &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_HETEROGENEOUS_STAM_LOOP &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_HETEROGENEOUS_OLD &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_HETEROGENEOUS_TESSELATE &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_GEOMETRIC &&
       (SLIDESubdivisionType)(m_dyniType.GetValue()) != slidesubtSLF_SUBDIVISION_GEOMETRIC_TESSELATE )
    {
      QEMeshIter iterQEMesh;
      QEEdge *pe;

      iterQEMesh.Init(m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()]);

      PushAuxillaryGeometryAttributes();
      {
	gilColor3f(1.0, 0.0, 0.0);
	gilBegin(GL_LINES);
	{
	  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdgeBoundary() )
	    {
	      gilVertex3fv(pe->GetVertex()->GetPoint().Get());
	      gilVertex3fv(pe->GetEdge()->GetVertex()->GetPoint().Get());
	    }
	}
	gilEnd();
      }
      PopAuxillaryGeometryAttributes();
    }

  if ( m_dyniDrawVertices.GetValue() &&
       m_dyniSubdivisionLevel.GetValue() > 0 )
    {
      switch ( (SLIDESubdivisionType)(m_dyniType.GetValue()) )
	{
	case slidesubtSLF_SUBDIVISION_SQRT_3_SELECTIVE_FACET:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT:
	  {
	    QEMeshIter iterQEMesh;
	    QEFacetIter iterQEFacet;
	    QEFacet *pfa;
	    QEVertex *pvx;

	    iterQEMesh.Init(m_ppQEMeshSubdivision[1]);

	    PushAuxillaryGeometryAttributes();
	    //PushShadingAttributes(slideshtSLF_FLAT);
	    {
	      for ( pfa = iterQEMesh.PeekFirstFacet();
		    pfa != NULL;
		    pfa = iterQEMesh.PeekNextFacet() )
		{
		  glColor3fv(s_ppfVertexColors[pfa->GetGeneration() & 0x7]);
		  glBegin(GL_TRIANGLE_FAN);
		  {
		    iterQEFacet.Init(pfa);
		    for ( pvx = iterQEFacet.PeekFirstVertex();
			  pvx != NULL;
			  pvx = iterQEFacet.PeekNextVertex() )
		      {
			glVertex3fv(pvx->GetPoint().Get());
		      }
		  }
		  glEnd();
		}
	    }
	    PopAuxillaryGeometryAttributes();
	    //PopShadingAttributes(slideshtSLF_FLAT);
	  }
	// Intentionally allow it to fall through
	//break;
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE_LIMIT:
	case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT:
	case slidesubtSLF_SUBDIVISION_CATMULL_CLARK_SELECTIVE:
	case slidesubtSLF_SUBDIVISION_DOO_SABIN_SELECTIVE:
	case slidesubtSLF_SUBDIVISION_DOO_SABIN_SELECTIVE_LIMIT:
	case slidesubtSLF_SUBDIVISION_SQRT_4_SELECTIVE:
	case slidesubtSLF_SUBDIVISION_SQRT_3_SELECTIVE:
	  {
	    QEMeshIter iterQEMesh;
	    QEVertex *pvx;

	    iterQEMesh.Init(m_ppQEMeshSubdivision[1]);

	    PushAuxillaryGeometryAttributes();
	    {
	      glPointSize(3.0);
	      glBegin(GL_POINTS);
	      {
		for ( pvx = iterQEMesh.PeekFirstVertex();
		      pvx != NULL;
		      pvx = iterQEMesh.PeekNextVertex() )
		  {
		    glColor3fv(s_ppfVertexColors[pvx->GetGeneration() & 0x7]);
		    glVertex3fv(pvx->GetPoint().Get());
		  }
	      }
	      glEnd();
	    }
	    PopAuxillaryGeometryAttributes();
	  }
	break;
	default:
	break;
	}
    }
}

VOID CSLIDESubdivision::RenderNormals(SLIDEShadingType eShading)
{
  BOOL bLimitSurface, bFaceted, bCalculateNormals, bSelective;

  PushAuxillaryGeometryAttributes();
  {
    gilColor3f(1.0, 0.0, 1.0);
    
    bLimitSurface = FALSE;
    bCalculateNormals = FALSE;
    bSelective = FALSE;
    switch ( (UINT_32)m_dyniType.GetValue() )
      {
      case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_SELECTIVE_LIMIT:
      case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_FACET_SHARP_SELECTIVE_LIMIT:
	bSelective = TRUE;
      case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SHARP_LOOP:
      case slidesubtSLF_SUBDIVISION_SHARP_CATMULL_CLARK:
	bCalculateNormals = TRUE;
      case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_LOOP:
      case slidesubtSLF_SUBDIVISION_CORNER_ROUNDING_SELECTIVE_LIMIT:
      case slidesubtSLF_SUBDIVISION_CATMULL_CLARK:
      case slidesubtSLF_SUBDIVISION_CATMULL_CLARK_C2:
      case slidesubtSLF_SUBDIVISION_DOO_SABIN:
      case slidesubtSLF_SUBDIVISION_DOO_SABIN_SELECTIVE_LIMIT:
      case slidesubtSLF_SUBDIVISION_SQRT_3_SELECTIVE_FACET:
	bLimitSurface = TRUE;
	break;
      case slidesubtSLF_SUBDIVISION_GEOMETRIC:
	bCalculateNormals = TRUE;
	break;
      default:
	break;
      }
    bFaceted = ( !(eShading == slideshtSLF_GOURAUD ||
		   eShading == slideshtSLF_PHONG) );
    ::RenderNormals(m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()], 
		    bFaceted, bLimitSurface, bCalculateNormals, bSelective);
  }
  PopAuxillaryGeometryAttributes();
}

VOID CSLIDESubdivision::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtSubdivision], m_pcGUIID);

  WriteSLIDEFlags(fpOutputFile, 0);

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->WriteSLIDE(fpOutputFile);    
    }

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndSubdivision]);
}

VOID CSLIDESubdivision::WriteRIBGeometry(FILE *fpOutputFile,
				   const Matrix &mGproj_obj, 
				   CBound *pCBoundClip, 
				   SLIDEShadingType eShading,
				   SLIDELODType eLOD)
{
/*
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( fpOutputFile != NULL );

  iter.Init(&m_listCSLIDEInstance);
  
  for ( pCSLIDEInstance = (CSLIDEInstance*)iter.PeekFirst();
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)iter.PeekNext() )
    {
      pCSLIDEInstance->WriteRIB(fpOutputFile,
				mGproj_obj, pCBoundClip, eShading, eLOD);    
    }
    */
}

VOID CSLIDESubdivision::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );
  /*
  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE SUBDIVISION: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);
  */
}

VOID CSLIDESubdivision::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );
  /*
  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE SUBDIVISION: %s", m_pcGUIID);
  */
}

VOID CSLIDESubdivision::WriteWRLZ(FILE *fpOutputFile, const Matrix &m,
				  CSLIDESurface *pCSLIDESurface)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "Shape { # triangle mesh: SUBDIVISION\n\n");

  // Color Inheritance
  if( m_pCSLIDESurface == CSLIDESurface::GetSLF_INHERIT() )
    {
      pCSLIDESurface->WriteWRLZ(fpOutputFile);
    }
  else
    {
      ASSERT( m_pCSLIDESurface != NULL );
      m_pCSLIDESurface->WriteWRLZ(fpOutputFile);
    }

  // Mesh
  ::WriteWRLZ(m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()], fpOutputFile, m);

  fprintf(fpOutputFile, "} # end triangle mesh\n\n");
}

VOID CSLIDESubdivision::WriteSTL(FILE *fpOutputFile, const Matrix &m)
{
  ::WriteSTL(m_ppQEMeshSubdivision[m_dyniSubdivisionLevel.GetValue()], fpOutputFile, m);
}

QEMesh *CSLIDESubdivision::WriteSIF(FILE * /*fpOutputFile*/, UINT_32 /*uiIndentLevel*/, 
				    const Matrix &/*mGroot_obj*/, const Matrix &/*mGobj_root*/,
				    QEMesh * /*pmeBoundaryEdges*/)
{
  return GetQEMesh();
}

QEMesh *CSLIDESubdivision::GetQEMesh()
{
  UINT_32 i;

  UpdateQEMesh();

  i = m_dyniSubdivisionLevel.GetValue();

  return m_ppQEMeshSubdivision[i]->Copy();
}






