#ifndef __SLIDESUBDIVISION_H__
#define __SLIDESUBDIVISION_H__

#include "slidegroup.h"

class CSLIDEWorld;

class QEMesh;

//////////////////////////////////////////////////////////////////////
// CSLIDESubdivision Class
//

class CSLIDESubdivision : public CSLIDEGroup
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDESubdivision();
  virtual ~CSLIDESubdivision();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  inline VOID SetUseDisplayLists(DynInt *pdyniUseDisplayLists);
  inline VOID SetDrawControlMesh(DynInt *pdyniDrawControlMesh);
  inline VOID SetDrawCurrentMesh(DynInt *pdyniDrawCurrentMesh);
  inline VOID SetDrawVertices(DynInt *pdyniDrawVertices);
  inline VOID SetType(DynInt *pdyniType);
  inline VOID SetErrorMetric(DynInt *pdyniErrorMetric);
  inline VOID SetSubdivisionLevel(DynInt *pdyniSubdivisionLevel);
  inline VOID SetFacetsMax(DynInt *pdyniFacetsMax);
  inline VOID SetWeightedEdges(DynInt *pdyniWeightedEdges);
  inline VOID SetWeight(DynFloat *pdynfWeight);
  inline VOID SetA(DynInt *pdyniA);
  inline VOID SetB(DynInt *pdyniB);
  inline VOID SetTolerance(DynFloat *pdynfTolerance);

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual CSLIDECamera *FindCamera(CListIterInt *piterIDs, 
				   CListInt *plistCSLIDEInstance);
  virtual CSLIDELight *FindLight(CListIterInt *piterIDs, 
				 CListInt *plistCSLIDEInstance);
  virtual CSLIDEAreaLight *FindAreaLight(CListIterInt *piterIDs, 
					 CListInt *plistCSLIDEInstance);

  virtual VOID Preprocess();
  virtual VOID PreprocessGraph();

  virtual VOID Update();
  virtual VOID UpdateGraph();

  virtual QEMesh *GetQEMesh();

protected:
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();
  virtual VOID UpdateQEMesh();
  virtual VOID UpdateSubdivision();
  virtual VOID UpdateDisplayLists();
  virtual VOID CalculateBound();

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID PrerenderGeometry(SLIDEShadingType eShading);
  virtual VOID PrerenderNormals(SLIDEShadingType eShading);
  virtual VOID PrerenderBounds();

  virtual VOID Prerender();

  virtual VOID RenderGeometry(const Matrix &mGproj_obj,
			      CBound *pCBoundClip, 
			      SLIDEShadingType eShading,
			      SLIDELODType eLOD,
			      CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderNormals(SLIDEShadingType eShading);

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

  virtual VOID WriteWRLZ(FILE *fpOutputFile, const Matrix &mTransform,
			 CSLIDESurface *pCSLIDESurface);
  virtual VOID WriteSTL(FILE *fpOutputFile, const Matrix &mTransform);

  virtual QEMesh *WriteSIF(FILE *fpOutputFile, UINT_32 uiIndentLevel, 
			   const Matrix &mGroot_obj, const Matrix &mGobj_root,
			   QEMesh *pmeBoundaryEdges);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline BOOL GetChildrenBoundsDynamic();
  inline VOID SetChildrenBoundsDynamic(BOOL bChildrenBoundsDynamic);

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  DynInt m_dyniUseDisplayLists;
  DynInt m_dyniDrawControlMesh;
  DynInt m_dyniDrawCurrentMesh;
  DynInt m_dyniDrawVertices;
  DynInt m_dyniType;
  DynInt m_dyniErrorMetric;
  DynInt m_dyniSubdivisionLevel;
  DynInt m_dyniFacetsMax;

  DynInt m_dyniWeightedEdges;
  DynFloat m_dynfWeight;

  // Extra integers for stop gap measures
  DynInt m_dyniA;
  DynInt m_dyniB;

  DynFloat m_dynfTolerance;

  UINT_32 m_uiSubdivisionLevelSize;
  QEMesh **m_ppQEMeshSubdivision;

  BOOL m_bUnsubdivide;
  BOOL m_uiUseDisplayLists;

  BOOL m_bChildrenBoundsDynamic;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDESubdivision::SetUseDisplayLists(DynInt *pdyniUseDisplayLists)
{
  m_dyniUseDisplayLists.Set(pdyniUseDisplayLists);
}

inline VOID CSLIDESubdivision::SetDrawControlMesh(DynInt *pdyniDrawControlMesh)
{
  m_dyniDrawControlMesh.Set(pdyniDrawControlMesh);
}

inline VOID CSLIDESubdivision::SetDrawCurrentMesh(DynInt *pdyniDrawCurrentMesh)
{
  m_dyniDrawCurrentMesh.Set(pdyniDrawCurrentMesh);
}

inline VOID CSLIDESubdivision::SetDrawVertices(DynInt *pdyniDrawVertices)
{
  m_dyniDrawVertices.Set(pdyniDrawVertices);
}

inline VOID CSLIDESubdivision::SetType(DynInt *pdyniType)
{
  m_dyniType.Set(pdyniType);
}

inline VOID CSLIDESubdivision::SetErrorMetric(DynInt *pdyniErrorMetric)
{
  m_dyniErrorMetric.Set(pdyniErrorMetric);
}

inline VOID CSLIDESubdivision::SetSubdivisionLevel(DynInt *pdyniSubdivisionLevel)
{
  m_dyniSubdivisionLevel.Set(pdyniSubdivisionLevel);
}

inline VOID CSLIDESubdivision::SetFacetsMax(DynInt *pdyniFacetsMax)
{
  m_dyniFacetsMax.Set(pdyniFacetsMax);
}

inline VOID CSLIDESubdivision::SetWeightedEdges(DynInt *pdyniWeightedEdges)
{
  m_dyniWeightedEdges.Set(pdyniWeightedEdges);
}

inline VOID CSLIDESubdivision::SetWeight(DynFloat *pdynfWeight)
{
  m_dynfWeight.Set(pdynfWeight);
}

inline VOID CSLIDESubdivision::SetA(DynInt *pdyniA)
{
  m_dyniA.Set(pdyniA);
}

inline VOID CSLIDESubdivision::SetB(DynInt *pdyniB)
{
  m_dyniB.Set(pdyniB);
}

inline VOID CSLIDESubdivision::SetTolerance(DynFloat *pdynfTolerance)
{
  m_dynfTolerance.Set(pdynfTolerance);
}

inline BOOL CSLIDESubdivision::GetChildrenBoundsDynamic()
{
  return m_bChildrenBoundsDynamic;
}

inline VOID CSLIDESubdivision::SetChildrenBoundsDynamic(BOOL bChildrenBoundsDynamic)
{
  m_bChildrenBoundsDynamic = bChildrenBoundsDynamic;
}

#endif // __SLIDESUBDIVISION_H__



