#include "slidesurface.h"
#include "bitmap.h"
#include "slideworld.h"
#include "sliderender.h"
#include "sif.h"

CSLIDESurface *CSLIDESurface::m_pCSLIDESurfaceDefault = NULL;

UINT_32 CSLIDESurface::m_uiCount = 0;
CListInt *CSLIDESurface::m_plistCSLIDESurfaceStack = NULL;
CListInt *CSLIDESurface::m_plistTextureIDs = NULL;
CListInt *CSLIDESurface::m_plistCSLIDESurfaceTextureOwners = NULL;
GLuint CSLIDESurface::m_textureCount=0;


// Default Values
static const VECTOR4D cvColor = { 0.5, 0.5, 0.5, 0.0 };
static const VECTOR4D cvReflectivity = { 1.0, 1.0, 0.0, 0.0 };
static const FLOAT cfExponent = 1.0;
static const FLOAT cfMetallic = 0.0;

// The exponent in the Phong term is different in SLIDE than it is 
// in OpenGL.
// SLIDE uses (R.V)^n while OpenGL uses the half way vector (H.N)^k
//             alpha               =                   2 *   beta
// set (cos(alpha))^n = (cos(beta))^k
// k = n*(log(cos(alpha)) / log(cos(beta)))
static const FLOAT cfBetaAvg = DEGTORAD(22.5);
//static const FLOAT cfAlphaBetaScale = log(cos(2*cfBetaAvg)) / log(cos(cfBetaAvg));
static const FLOAT cfAlphaBetaScale = (FLOAT)4.377388747;


// Texture Stuff
static const UINT_32 cuiTexTarget = GL_TEXTURE_2D;
static const UINT_32 cuiTexLevels = 0;
static const UINT_32 cuiTexComponents = 3;
static const UINT_32 cuiTexWidth = 64;
static const UINT_32 cuiTexHeight = 64;
static const UINT_32 cuiTexBorder = 0;
static const UINT_32 cuiTexFormat = GL_RGB;
static const UINT_32 cuiTexType = GL_UNSIGNED_BYTE;

//static const UINT_32 cuiTexWrapS = GL_CLAMP;
//static const UINT_32 cuiTexWrapT = GL_CLAMP;
static const UINT_32 cuiTexWrapS = GL_REPEAT;
static const UINT_32 cuiTexWrapT = GL_REPEAT;
static const UINT_32 cuiTexMagFilter = GL_NEAREST;
static const UINT_32 cuiTexMinFilter = GL_NEAREST;
//static const VECTOR4D cvTexBorderColor = {0.0, 0.0, 0.0, 0.0};
static const GLfloat cvTexBorderColor[4] = {0.0, 0.0, 0.0, 0.0};


//////////////////////////////////////////////////////////////////////
// CSLIDESurface Class
//

CSLIDESurface::CSLIDESurface()
{
  // Initialize the Texture Stack
  if ( m_uiCount == 0 )
    {
      m_plistCSLIDESurfaceStack = new CListInt;
      ASSERT( m_plistCSLIDESurfaceStack != NULL );
      m_plistCSLIDESurfaceStack->Init();

      m_plistTextureIDs = new CListInt;
      ASSERT( m_plistTextureIDs != NULL );
      m_plistTextureIDs->Init();

      m_plistCSLIDESurfaceTextureOwners = new CListInt;
      ASSERT( m_plistCSLIDESurfaceTextureOwners != NULL );
      m_plistCSLIDESurfaceTextureOwners->Init();
    }
  m_uiCount++;

  ((Vector*)(&m_vColor))->Set(cvColor);
  m_vReflectivity.Set(cvReflectivity);
  m_dynfExponent.SetValue(cfExponent);
  m_dynfMetallic.SetValue(cfMetallic);

  m_pcTextureFile = NULL;

  m_GLuiTextureID = 0;

  m_uiTarget = 0;
  m_uiLevels = 0;
  m_uiComponents = 0;
  m_uiWidth = 0;
  m_uiHeight = 0;
  m_uiBorder = 0;
  m_uiFormat = 0;
  m_uiType = 0;
  m_pvImage = NULL;

  m_uiWrapS = 0;
  m_uiWrapT = 0;
  m_uiMagFilter = 0;
  m_uiMinFilter = 0;
  //memset(m_vBorderColor, 0, sizeof(VECTOR4D));
  memset(m_vBorderColor, 0, 4*sizeof(GLfloat));
}

CSLIDESurface::~CSLIDESurface()
{
  m_uiCount--;
  if ( m_uiCount == 0 )
    {
      if ( m_plistCSLIDESurfaceStack != NULL )
	{
	  delete m_plistCSLIDESurfaceStack;
	  m_plistCSLIDESurfaceStack = NULL;
	}

      if ( m_plistTextureIDs != NULL )
	{
	  GLuint GLuiTextureID;

	  for ( GLuiTextureID = (UINT_32)(m_plistTextureIDs->RemoveFirst());
		gilIsTexture(GLuiTextureID);
		GLuiTextureID = (UINT_32)(m_plistTextureIDs->RemoveFirst()) )
	    {
	      gilDeleteTextures(1, &GLuiTextureID);
	    }
	  delete m_plistTextureIDs;
	  m_plistTextureIDs = NULL;
	}
      if ( m_plistCSLIDESurfaceTextureOwners != NULL )
	{
	  delete m_plistCSLIDESurfaceTextureOwners;
	  m_plistCSLIDESurfaceTextureOwners = NULL;
	}
    }
}

VOID CSLIDESurface::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			 CHAR *pcID)
{
  CSLIDEPrimitive::Init(pcFile, uiLine, uiColumn, pcID);

  m_vColor.Set(cvColor);
  m_vReflectivity.Set(cvReflectivity);
  m_dynfExponent.SetValue(cfExponent);
  m_dynfMetallic.SetValue(cfMetallic);

  m_pcTextureFile = NULL;

  m_GLuiTextureID = 0;

  m_uiTarget = cuiTexTarget;
  m_uiLevels = cuiTexLevels;
  m_uiComponents = cuiTexComponents;
  m_uiWidth = cuiTexWidth;
  m_uiHeight = cuiTexHeight;
  m_uiBorder = cuiTexBorder;
  m_uiFormat = cuiTexFormat;
  m_uiType = cuiTexType;

  m_uiWrapS = cuiTexWrapS;
  m_uiWrapT = cuiTexWrapT;
  m_uiMagFilter = cuiTexMagFilter;
  m_uiMinFilter = cuiTexMinFilter;
  //memcpy(m_vBorderColor, cvTexBorderColor, sizeof(VECTOR4D));
  memcpy(m_vBorderColor, cvTexBorderColor, 4*sizeof(GLfloat));

  m_vAveTexColor.Null();
}

VOID CSLIDESurface::Uninit()
{
  CSLIDEPrimitive::Uninit();

  if ( m_pvImage != NULL )
    {
      delete []m_pvImage;
      m_pvImage = NULL;
    }
}

BOOL CSLIDESurface::Initialized()
{
  return CSLIDEPrimitive::Initialized();
}

BOOL CSLIDESurface::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;

  bCorrect = CSLIDEPrimitive::CheckSemantics(pCSLIDEWorld);

  return bCorrect;
}

BOOL CSLIDESurface::LoadTexture()
{
  BitMap bitMap;
  UINT_32 uiLength;
  BOOL bCorrect;

  if ( m_pcTextureFile == NULL )
    {
      // This surface does not have an associated texture
      return TRUE;
    }

  // Check if we have already loaded this texture file
  m_GLuiTextureID = (UINT_32)(m_plistTextureIDs->Peek((UINT_32)m_pcTextureFile));
  if ( gilIsTexture(m_GLuiTextureID) )
    {
      //printf("Load Surface: Reusing Texture %ld on \"%s\"\n", m_GLuiTextureID, m_pcTextureFile); 
      // We have already generated this texture so we are done
      
      // Grab the average texture color from the owner
      CSLIDESurface *pCSLIDESurface;
      pCSLIDESurface = (CSLIDESurface*)(m_plistCSLIDESurfaceTextureOwners->Peek((UINT_32)m_pcTextureFile));
      ASSERT( pCSLIDESurface != NULL );
      m_vAveTexColor = pCSLIDESurface->GetAveTexColor();

      return TRUE;
    }

  // MOD by RICH, for multiple windows.
  // NOTE: if m_GLuiTextureID == 0, then the texture ID must have been created
  //       previously in another context
  //ASSERT( m_GLuiTextureID == 0 );

  bCorrect = FALSE;

  // I only want file names that end ".gif" or ".GIF" because it will then
  // be easy for me to decide what RenderMan will be calling the ".tx" file
  // The GIF reader does not actually go on the file name in deciding that the
  // file is a GIF file, so if we were just doing the OpenGL stuff only then
  // this check would be unnecessary
  uiLength = strlen(m_pcTextureFile);
  if ( !(uiLength > 4) )
    {
      HandleError("Bad Texture File name: \"%s\" (must end \".gif\" or \".GIF\")", m_pcTextureFile); 
      m_pcTextureFile = NULL;
      return FALSE;
    }
  if ( !(strcmp(m_pcTextureFile + uiLength - 4, ".gif") == 0 ||
	 strcmp(m_pcTextureFile + uiLength - 4, ".GIF") == 0
#ifdef TIFFANDJPEG
      || strcmp(m_pcTextureFile + uiLength - 4, ".tif") == 0 ||
	 strcmp(m_pcTextureFile + uiLength - 4, ".TIF") == 0 ||
	 strcmp(m_pcTextureFile + uiLength - 4, ".jpg") == 0 ||
	 strcmp(m_pcTextureFile + uiLength - 4, ".JPG") == 0
#endif
	 ))
    {
#ifdef TIFFANDJPEG
      HandleError("Bad Texture File name: \"%s\" (must end \".gif\", \".GIF\", \".tif\", \".TIF\", \".jpg\", or \".JPG\")", m_pcTextureFile); 
#else
      HandleError("Bad Texture File name: \"%s\" (must end \".gif\" or \".GIF\"", m_pcTextureFile); 
#endif
      m_pcTextureFile = NULL;
      return FALSE;
    }

  if ( LoadBitMap(m_pcFile, m_pcTextureFile, &bitMap) )
    {
      ASSERT( bitMap.pvBuffer != NULL );
      ASSERT( bitMap.iBitsPerComp == 8 );
      m_uiComponents = (UINT_32)(bitMap.iComps);
      m_uiWidth = (UINT_32)(bitMap.iWidth);
      m_uiHeight = (UINT_32)(bitMap.iHeight);
      m_pvImage = bitMap.pvBuffer;
      
      ClipTexture();
      ComputeAverageTextureColor();
      
      // Allocate an OpenGL Texture Object
      BindTexture();
      
      //printf("Load Surface: Allocating Texture %ld for \"%s\"\n", m_GLuiTextureID, m_pcTextureFile); 
      
      ASSERT( gilIsTexture(m_GLuiTextureID) ); 
      m_plistTextureIDs->Insert((VOID*)m_GLuiTextureID, 
				(UINT_32)m_pcTextureFile);
      m_plistCSLIDESurfaceTextureOwners->Insert((VOID*)this, 
						(UINT_32)m_pcTextureFile);
      
      bCorrect = TRUE;
    }
  else
    {
#ifdef TIFFANDJPEG
      HandleError("File format error for Texture File: \"%s\" (GIF, TIF, and JPEG  are the only acceptable format)", m_pcTextureFile); 
#else
      HandleError("File format error for Texture File: \"%s\" (GIF is the only acceptable format)", m_pcTextureFile); 
#endif
      m_pcTextureFile = NULL;
    }

  return bCorrect;
}


VOID CSLIDESurface::GenTextures(GLuint *ptexID)
{
  m_textureCount ++;
  *ptexID = m_textureCount;
}

VOID CSLIDESurface::BindTexture()
{
  ASSERT( m_pcTextureFile != NULL );

  //Can have exisiting texture ID for multiple contexts
  //ASSERT( m_GLuiTextureID == 0 );.......................................................................................................................

  gilPixelStorei(GL_UNPACK_ALIGNMENT, 1);


  // MOD by Rich 9.24.98 for multiple contexts
  // if TextureID != 0, want to keep it the same ID between contexts
  if (m_GLuiTextureID == 0) {
    // Allocate an OpenGL texture object
    //gilGenTextures(1, &m_GLuiTextureID);
    GenTextures(&m_GLuiTextureID);
    // make SURE that i got a non-zero id, and that it's not already taken in this
    // context.
    ASSERT( m_GLuiTextureID != 0 );
    ASSERT( !gilIsTexture(m_GLuiTextureID) ); 
  }

  gilBindTexture(m_uiTarget, m_GLuiTextureID);

  /*
  gilTexParameterf(m_uiTarget, GL_TEXTURE_WRAP_S, m_uiWrapS);
  gilTexParameterf(m_uiTarget, GL_TEXTURE_WRAP_T, m_uiWrapT);
  gilTexParameterf(m_uiTarget, GL_TEXTURE_MAG_FILTER, m_uiMagFilter);
  gilTexParameterf(m_uiTarget, GL_TEXTURE_MIN_FILTER, m_uiMinFilter);
  */
  gilTexParameteri(m_uiTarget, GL_TEXTURE_WRAP_S, m_uiWrapS);
  gilTexParameteri(m_uiTarget, GL_TEXTURE_WRAP_T, m_uiWrapT);
  gilTexParameteri(m_uiTarget, GL_TEXTURE_MAG_FILTER, m_uiMagFilter);
  gilTexParameteri(m_uiTarget, GL_TEXTURE_MIN_FILTER, m_uiMinFilter);
  gilTexParameterfv(m_uiTarget, GL_TEXTURE_BORDER_COLOR, m_vBorderColor);
  
  gilTexImage2D(m_uiTarget, m_uiLevels, m_uiComponents, m_uiWidth, m_uiHeight, 
		m_uiBorder, m_uiFormat, m_uiType, m_pvImage);
}

// Find the largest power of 2 which is <= uiNum
UINT_32 LargestPower(UINT_32 uiNum)
{
  UINT_32 uiPower;

  if ( uiNum == 0 )
    {
      return 0;
    }

  uiPower = uiNum;
  while ( (uiPower & (uiPower - 1)) != 0 )
    {
      uiPower = (uiPower & (uiPower - 1));
    }

  return uiPower;
}

VOID CSLIDESurface::ClipTexture()
{
  UINT_32 uiWidth, uiHeight;

  uiWidth = LargestPower(m_uiWidth);
  if ( uiWidth != m_uiWidth )
    {
      uiWidth = uiWidth * 2;
    }
  uiHeight = LargestPower(m_uiHeight);
  if ( uiHeight != m_uiHeight )
    {
      uiHeight = uiHeight * 2;
    }

  if ( uiWidth == m_uiWidth && uiHeight == m_uiHeight )
    {
      return;
    }

  HandleWarning("Clipping Texture Map from file \"%s\": ( %ld, %ld ) => ( %ld, %ld )", 
	 m_pcTextureFile, m_uiWidth, m_uiHeight, uiWidth, uiHeight);

  // Clip the texture map
  UINT_32 i;
  ColorEntry *pceBuffer, *pceDst, *pceSrc;

  pceBuffer = new ColorEntry[uiWidth*uiHeight];
  ASSERT( pceBuffer != NULL );
  pceDst = pceBuffer;
  pceSrc = (ColorEntry*)m_pvImage;
  
  /*
  // This was when I was clipping out the largest 2^m X 2^n texture
  for ( i = 0; i < uiHeight; i++ )
    {
      memcpy(pceDst, pceSrc, uiWidth*sizeof(ColorEntry));
      pceDst += uiWidth;
      pceSrc += m_uiWidth;
    }
    */

  // This allocates the next largest 2^m X 2^n texture rectangle and 
  // black fills the edges
  memset(pceDst, 0, uiWidth*uiHeight*sizeof(ColorEntry)); 
  for ( i = 0; i < m_uiHeight; i++ )
    {
      memcpy(pceDst, pceSrc, m_uiWidth*sizeof(ColorEntry));
      pceDst += uiWidth;
      pceSrc += m_uiWidth;
    }

  delete m_pvImage;
  m_pvImage = pceBuffer;
  m_uiHeight = uiHeight;
  m_uiWidth = uiWidth;
}

VOID CSLIDESurface::ComputeAverageTextureColor()
{
  UINT_32 i;
  ColorEntry *pce;
  Vector vColor;
  
  m_vAveTexColor.Null();
  pce = (ColorEntry*)m_pvImage;

  for ( i = 0; i < m_uiWidth*m_uiHeight; i++ )
    {
      vColor.Set(pce->pbComp[0], pce->pbComp[1], pce->pbComp[2]);
      m_vAveTexColor += vColor;
      pce++;
    }

  m_vAveTexColor /= (i*256);
}

BOOL CSLIDESurface::Valid()
{
  if ( !CSLIDEPrimitive::Valid() )
    {
      return FALSE;
    }

  return ( 0.0 <= m_vColor[R] && m_vColor[R] <= 1.0 &&
	   0.0 <= m_vColor[G] && m_vColor[G] <= 1.0 &&
	   0.0 <= m_vColor[B] && m_vColor[B] <= 1.0 &&
	   0.0 <= m_vReflectivity[0] && m_vReflectivity[0] <= 1.0 &&
	   0.0 <= m_vReflectivity[1] && m_vReflectivity[1] <= 1.0 &&
	   0.0 <= m_vReflectivity[2] && m_vReflectivity[2] <= 1.0 );
}

VOID CSLIDESurface::Remove(CSLIDEWorld *pCSLIDEWorld)
{
  CSLIDEPrimitive::Remove(pCSLIDEWorld);
  
  pCSLIDEWorld->RemoveSurface(this);
  Uninit();
}

VOID CSLIDESurface::CheckDynamic()
{
  SetDynamic(m_vColor.GetDynamic() ||
	     m_vReflectivity.GetDynamic() ||
	     m_dynfExponent.GetDynamic() ||
	     m_dynfMetallic.GetDynamic());
}

VOID CSLIDESurface::UpdateState()
{
  m_vColor.Update();
  m_vReflectivity.Update();
  m_dynfExponent.Update();
  m_dynfMetallic.Update();
  m_dyniTexture.Update();
}

VOID CSLIDESurface::Prerender()
{
  LoadTexture();
}

GLbitfield CSLIDESurface::CalculateAttributesMask()
{
  if ( this != m_pCSLIDESurfaceDefault )
    {
      return GL_CURRENT_BIT | GL_LIGHTING_BIT | GL_TEXTURE_BIT;
    }

  return 0x0;
}

VOID CSLIDESurface::Push()
{
  if ( this != m_pCSLIDESurfaceDefault )
    {
      //gilPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT | GL_TEXTURE_BIT);

      //////////////////////////////////////////////////
      // Texture
      /*
      if ( m_pcTextureFile != NULL && GetTextureMappingEnabled() )
	{
	  CSLIDESurface *pCSLIDESurface;

	  // Look at the top of the stack
	  pCSLIDESurface = (CSLIDESurface*)(m_plistCSLIDESurfaceStack->PeekFirst());
	  // Push this surface on to the top of the stack
	  m_plistCSLIDESurfaceStack->InsertFirst(this, 0);

	  if ( this != pCSLIDESurface )
	    {
	      RenderTexture();
	    }
	}
	*/
      Render();
    }
}

VOID CSLIDESurface::Pop()
{
  if ( this != m_pCSLIDESurfaceDefault )
    {
      //////////////////////////////////////////////////
      // Texture
      /*
      if ( m_pcTextureFile != NULL && GetTextureMappingEnabled() )
	{
	  CSLIDESurface *pCSLIDESurface;

	  // Remove this surface from the top of the stack
	  pCSLIDESurface = (CSLIDESurface*)(m_plistCSLIDESurfaceStack->RemoveFirst());
	  ASSERT( this == pCSLIDESurface );

	  // Restore the surface properties of the surface before us
	  pCSLIDESurface = (CSLIDESurface*)(m_plistCSLIDESurfaceStack->PeekFirst());

	  if ( this != pCSLIDESurface && pCSLIDESurface != NULL )
	    {
	      pCSLIDESurface->RenderTexture();
	    }
	}
	*/
      //gilPopAttrib();
    }
}

VOID CSLIDESurface::RenderTexture()
{
  ASSERT( m_pcTextureFile != NULL && GetTextureMappingEnabled() );

  gilEnable(m_uiTarget);
  
  gilPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  
  gilTexImage2D(m_uiTarget, m_uiLevels, m_uiComponents, m_uiWidth, m_uiHeight, 
		m_uiBorder, m_uiFormat, m_uiType, m_pvImage);
  
  /*
  gilTexParameterf(m_uiTarget, GL_TEXTURE_WRAP_S, m_uiWrapS);
  gilTexParameterf(m_uiTarget, GL_TEXTURE_WRAP_T, m_uiWrapT);
  gilTexParameterf(m_uiTarget, GL_TEXTURE_MAG_FILTER, m_uiMagFilter);
  gilTexParameterf(m_uiTarget, GL_TEXTURE_MIN_FILTER, m_uiMinFilter);
  */
  gilTexParameteri(m_uiTarget, GL_TEXTURE_WRAP_S, m_uiWrapS);
  gilTexParameteri(m_uiTarget, GL_TEXTURE_WRAP_T, m_uiWrapT);
  gilTexParameteri(m_uiTarget, GL_TEXTURE_MAG_FILTER, m_uiMagFilter);
  gilTexParameteri(m_uiTarget, GL_TEXTURE_MIN_FILTER, m_uiMinFilter);
  gilTexParameterfv(m_uiTarget, GL_TEXTURE_BORDER_COLOR, m_vBorderColor);
  
  //VECTOR4D vColor = {1.0, 1.0, 1.0, 1.0};
  GLfloat vColor[4] = {1.0, 1.0, 1.0, 1.0};
  
  gilTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, vColor);
  gilTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  gilEnable(m_uiTarget);
}

VOID CSLIDESurface::Render()
{
  Vector vColor;
  //VECTOR4D v;
  GLfloat v[4];
  FLOAT f;
  UINT_32 i;

  if(CSLIDERender::GetBlend().GetValue()==slideoftSLF_ON && m_vColor[3]!=0){
    gilEnable(GL_BLEND);
    gilBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    glDepthMask(GL_FALSE); //make z-buffer read only
    v[3]=m_vColor[3];
  }
  else{
    gilDisable(GL_BLEND);
    glDepthMask(GL_TRUE);
    v[3]=1;
  }
  v[0]=m_vColor[0];
  v[1]=m_vColor[1];
  v[2]=m_vColor[2];

  gilColor4fv(v);

  // Setting Material Properties

  // Ambient Color
  vColor = m_vReflectivity[0]*m_vColor;
  //vColor.Get(v);
  v[0] = vColor[0];
  v[1] = vColor[1];
  v[2] = vColor[2];
  gilMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, v);

  // Diffuse Color
  vColor = m_vReflectivity[1]*m_vColor;
//  vColor.Get(v);
  v[0] = vColor[0];
  v[1] = vColor[1];
  v[2] = vColor[2];
  gilMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, v);

  // Specular Color
  f = m_dynfMetallic.GetValue();
  for ( i = 0; i < 3; i++ )
    {
      vColor[i] = f*m_vColor[i] + ((FLOAT)1.0 - f);
    }
  vColor = m_vReflectivity[2]*vColor;
//  vColor.Get(v);
  v[0] = vColor[0];
  v[1] = vColor[1];
  v[2] = vColor[2];
  gilMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, v);

  // Shininess
  gilMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 
	       cfAlphaBetaScale*m_dynfExponent.GetValue());

  // Emission
  // Use default of Zero
  v[0] = 0.0;
  v[1] = 0.0;
  v[2] = 0.0;
  v[3] = 1.0;
  gilMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, v);  

  // Color Indexes
  // Use default

  // Texture
  if ( gilIsTexture(m_GLuiTextureID) )
    {
      gilBindTexture(GL_TEXTURE_2D, m_GLuiTextureID);
      if(GetTextureMode() == slidetexturetSLF_ENVMAP){
	gilTexGeni(GL_S,GL_TEXTURE_GEN_MODE,GL_SPHERE_MAP);
	gilTexGeni(GL_T,GL_TEXTURE_GEN_MODE,GL_SPHERE_MAP);
	gilEnable(GL_TEXTURE_GEN_S);
	gilEnable(GL_TEXTURE_GEN_T);
      }
      else{
	gilDisable(GL_TEXTURE_GEN_S);
	gilDisable(GL_TEXTURE_GEN_T);
      }
    }
}

VOID CSLIDESurface::WriteSLIDE(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n",
	  gcppcSLIDEReservedWords[sliderwtSurface], m_pcGUIID); 
  fprintf(fpOutputFile, "  %s ( %.3f %.3f %.3f )\n",
	  gcppcSLIDEReservedWords[sliderwtColor], 
	  m_vColor[R], m_vColor[G], m_vColor[B]); 
  fprintf(fpOutputFile, "  %s ( %.3f %.3f %.3f )\n",
	  gcppcSLIDEReservedWords[sliderwtReflectivity], 
	  m_vReflectivity[0], m_vReflectivity[1], m_vReflectivity[2]); 
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtExponent], 
	  m_dynfExponent.GetValue());
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtMetallic], 
	  m_dynfMetallic.GetValue());
  if ( m_pcTextureFile != NULL )
    {
      fprintf(fpOutputFile, "  %s \"%s\"\n",
	      gcppcSLIDEReservedWords[sliderwtTexture], 
	      m_pcTextureFile);
    }
  CSLIDEPrimitive::WriteSLIDEFlags(fpOutputFile, 0);

  fprintf(fpOutputFile, "%s\n",
	  gcppcSLIDEReservedWords[sliderwtEndSurface]);
}

VOID CSLIDESurface::WriteRIBPush(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  if ( this != m_pCSLIDESurfaceDefault )
    {
      RiAttributeBegin();

      WriteRIB(fpOutputFile);
    }
}

VOID CSLIDESurface::WriteRIBPop(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  if ( this != m_pCSLIDESurfaceDefault )
    {
      RiAttributeEnd();
    }
}

// I am not currently using the Shininess parameter
// I need a Shiny Painted Plastic shader, but this is not standard
VOID CSLIDESurface::WriteRIB(FILE *fpOutputFile)
{
  RtColor rtColor;
  FLOAT f, fKa, fKd, fKs, fRoughness;
  UINT_32 i;

  ASSERT( fpOutputFile != NULL );

  WriteRIBBegin(fpOutputFile);

  rtColor[0] = m_vColor[0];
  rtColor[1] = m_vColor[1];
  rtColor[2] = m_vColor[2];

  RiColor(rtColor);

  fKa = m_vReflectivity[0];
  fKd = m_vReflectivity[1];
  fKs = m_vReflectivity[2];
  fRoughness = cfAlphaBetaScale*m_dynfExponent.GetValue();
  if ( fRoughness == 0 )
    {
      fRoughness = RI_INFINITY;
    }
  else
    {
      fRoughness = (FLOAT)1.0 / fRoughness;
    }

  ASSERT( m_dynfExponent.GetValue() != 0 );

  // Specular Color
  f = m_dynfMetallic.GetValue();
  for ( i = 0; i < 3; i++ )
    {
      rtColor[i] = f*m_vColor[i] + (FLOAT)(1.0 - f);
    }

  //////////////////////////////////////////////////
  // Texture
  if ( m_pcTextureFile != NULL && GetTextureMappingEnabled() )
    {
      UINT_32 uiLength;
      CHAR *pcBuffer;
      Vector vColor;

      uiLength = strlen(m_pcTextureFile);
      ASSERT( uiLength > 4 );
      ASSERT( strcmp(m_pcTextureFile + uiLength - 4, ".gif") == 0 ||
	      strcmp(m_pcTextureFile + uiLength - 4, ".GIF") == 0 );

      pcBuffer = new CHAR[uiLength+1];
      ASSERT( pcBuffer != NULL );
      strcpy(pcBuffer, m_pcTextureFile);
      strcpy(pcBuffer + uiLength - 4, ".tx");

      RiSurface(RI_PAINTEDPLASTIC,
		RI_KA, &fKa,
		RI_KD, &fKd,
		RI_KS, &fKs,
		RI_ROUGHNESS, &fRoughness,
		RI_SPECULARCOLOR, &rtColor,
		RI_TEXTURENAME, &pcBuffer, 
		RI_NULL);

      // I am not sure if I should multiply this or what???
      vColor = ComponentMultiply(m_vAveTexColor, m_vColor);
      //vColor = m_vAveTexColor;

      RiArchiveRecord(RI_COMMENT, "\nAttribute \"radiosity\" \"averagecolor\" [%.7f %.7f %.7f]", vColor[0], vColor[1], vColor[2]);

      delete pcBuffer;
    }
  else
    {
      RiSurface(RI_PLASTIC,
		RI_KA, &fKa,
		RI_KD, &fKd,
		RI_KS, &fKs,
		RI_ROUGHNESS, &fRoughness,
		RI_SPECULARCOLOR, rtColor,
		RI_NULL);
    }

  WriteRIBEnd(fpOutputFile);
}

VOID CSLIDESurface::WriteRIBCreateTexture(FILE  * /*fpOutputFile*/ )
{
  //////////////////////////////////////////////////
  // Texture
  if ( m_pcTextureFile != NULL )
    {
      UINT_32 uiLength;
      CHAR *pcBuffer;

      uiLength = strlen(m_pcTextureFile);
      ASSERT( uiLength > 4 );
      ASSERT( strcmp(m_pcTextureFile + uiLength - 4, ".gif") == 0 ||
	      strcmp(m_pcTextureFile + uiLength - 4, ".GIF") == 0 );

      pcBuffer = new CHAR[uiLength+1];
      ASSERT( pcBuffer != NULL );
      strcpy(pcBuffer, m_pcTextureFile);
      strcpy(pcBuffer + uiLength - 4, ".tx");

      RiArchiveRecord(RI_COMMENT, " %s %s", m_pcTextureFile, pcBuffer);

      delete pcBuffer;
    }
}

VOID CSLIDESurface::WritePOV(FILE  * fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf (fpOutputFile, " rgb <%.7f, %.7f, %.7f>",
	   m_vColor[0], m_vColor[1], m_vColor[2]);
}

VOID CSLIDESurface::WriteWRLZ(FILE  * fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  UINT_32 uiIndentLevel;

  uiIndentLevel = 1;

  WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "appearance Appearance {\n");
  WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "material Material {\n");

  WriteSIFIndent(fpOutputFile, uiIndentLevel+2);
  fprintf(fpOutputFile, "diffuseColor %.6f %.6f %.6f\n", 
	  m_vColor[X], m_vColor[Y], m_vColor[Z]);
  WriteSIFIndent(fpOutputFile, uiIndentLevel+2);
  fprintf(fpOutputFile, "ambientIntensity 1\n");
  WriteSIFIndent(fpOutputFile, uiIndentLevel+2);
  fprintf(fpOutputFile, "specularColor 0 0 0\n");
  WriteSIFIndent(fpOutputFile, uiIndentLevel+2);
  fprintf(fpOutputFile, "emissiveColor 0 0 0\n");
  WriteSIFIndent(fpOutputFile, uiIndentLevel+2);
  fprintf(fpOutputFile, "shininess 0\n");
  WriteSIFIndent(fpOutputFile, uiIndentLevel+2);
  fprintf(fpOutputFile, "transparency 0\n");

  WriteSIFIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "} # end material\n");
  WriteSIFIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "} # end appearance\n\n");
}

VOID CSLIDESurface::WriteWRLZColor(FILE * fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%.6f %.6f %.6f, ",  m_vColor[X], m_vColor[Y], m_vColor[Z]);
}

