#ifndef __SLIDESURFACE_H__
#define __SLIDESURFACE_H__

#include "slideprimitive.h"

//////////////////////////////////////////////////////////////////////////////
// CSLIDESurface Class
//

class CSLIDESurface : public CSLIDEPrimitive
{
//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDESurface();
  virtual ~CSLIDESurface();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  inline VOID SetColor(DynVector *pdynvColor);
  inline VOID SetReflectivity(DynVector *pdynvReflectivity);
  inline VOID SetExponent(DynFloat *pdynfExponent);
  inline VOID SetMetallic(DynFloat *pdynfMetallic);
  inline VOID SetTexture(CHAR *pcTextureFile, DynInt* pdyniTexture);

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);
  BOOL LoadTexture();
  VOID ClipTexture();
  VOID ComputeAverageTextureColor();
  VOID BindTexture();

  virtual BOOL Valid();

  virtual VOID Remove(CSLIDEWorld *pCSLIDEWorld);

//////////////////////////////////////////////////
// Geometric Computations
//
public:

protected:
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID Prerender();

  GLbitfield CalculateAttributesMask();
  virtual VOID Push();   // Use these
  virtual VOID Pop();
  virtual VOID Render(); // Do not use this directly

protected:
  VOID RenderTexture(); // Do not use this directly

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIBPush(FILE *fpOutputFile);   // Use these
  virtual VOID WriteRIBPop(FILE *fpOutputFile);
  virtual VOID WriteRIB(FILE *fpOutputFile); // Do not use this directly
  virtual VOID WriteRIBCreateTexture(FILE *fpOutputFile);

  virtual VOID WritePOV(FILE *fpOutputFile);
  virtual VOID WriteWRLZ(FILE *fpOutputFile);
  virtual VOID WriteWRLZColor(FILE *fpOutputFile);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  inline VOID SetColor(const Vector &vColor);
  inline const Vector &GetColor() const;
  inline VOID SetReflectivity(const Vector &vReflectivity);
  inline const Vector &GetReflectivity() const;
  inline VOID SetExponent(FLOAT fExponent);
  inline FLOAT GetExponent() const;
  inline VOID SetMetallic(FLOAT fMetallic);
  inline FLOAT GetMetallic() const;
  inline SLIDETextureType GetTextureMode() const;

  inline const Vector &GetAveTexColor() const;

  static inline CSLIDESurface *GetSLF_INHERIT();

  VOID GenTextures(GLuint *ptexID);


//////////////////////////////////////////////////
// Class Variables
//
public:
  static CSLIDESurface *m_pCSLIDESurfaceDefault;

  // Texture Stack
  static UINT_32 m_uiCount;
  static CListInt *m_plistCSLIDESurfaceStack;
  static CListInt *m_plistTextureIDs;
  static CListInt *m_plistCSLIDESurfaceTextureOwners;

protected:
  // total texture ID counter
  //static GLuint m_textureCount=0;
  static GLuint m_textureCount;



//////////////////////////////////////////////////
// Member Variables
//
protected:
  DynVector m_vColor;
  DynVector m_vReflectivity;
  DynFloat m_dynfExponent;
  DynFloat m_dynfMetallic;
  DynInt m_dyniTexture;

  CHAR *m_pcTextureFile;

  GLuint m_GLuiTextureID;

  UINT_32 m_uiTarget;            // GL_TEXTURE_2D
  UINT_32 m_uiLevels;            // How many LOD, for single detail = 0
  UINT_32 m_uiComponents;        // How many RGBA components are in the image buffer
  UINT_32 m_uiWidth;             // Width of image = 2^(Mw) + 2*b
  UINT_32 m_uiHeight;            // Heigth of image = 2^(Mh) + 2*b
  UINT_32 m_uiBorder;            // Border Width = 2*b
  UINT_32 m_uiFormat;            // GL_COLOR_INDEX, GL_RGB, GL_RGBA, 
                                 // GL_RED, GL_GREEN, GL_BLUE, GL_ALPHA
  UINT_32 m_uiType;              // GL_UNSIGNED_BYTE
  VOID *m_pvImage;               // The image buffer

  // Defining how it is Applied
  UINT_32 m_uiWrapS;             // GL_CLAMP, GL_REPEAT
  UINT_32 m_uiWrapT;             // GL_CLAMP, GL_REPEAT
  UINT_32 m_uiMagFilter;         // GL_NEAREST, GL_LINEAR
  UINT_32 m_uiMinFilter;         // GL_NEAREST, GL_LINEAR, 
                                 // GL_NEAREST_MIPMAP_NEAREST, GL_NEAREST_MIPMAP_LINEAR
                                 // GL_LINEAR_MIPMAP_NEAREST, GL_LINEAR_MIPMAP_LINEAR
  //VECTOR4D m_vBorderColor;         // RGBA color if no border texture is defined  
  GLfloat m_vBorderColor[4];         // RGBA color if no border texture is defined  

  Vector m_vAveTexColor;         // For radiosity with BMRT
};

//////////////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDESurface::SetColor(const Vector &vColor)
{
  m_vColor = vColor;
}

inline VOID CSLIDESurface::SetColor(DynVector *pdynvColor)
{
  m_vColor.SetAll(pdynvColor);
}

inline const Vector &CSLIDESurface::GetColor() const
{
  return m_vColor;
}

inline VOID CSLIDESurface::SetReflectivity(const Vector &vReflectivity)
{
  m_vReflectivity = vReflectivity;
}

inline VOID CSLIDESurface::SetReflectivity(DynVector *pdynvReflectivity)
{
  m_vReflectivity.SetAll(pdynvReflectivity);
}

inline const Vector &CSLIDESurface::GetReflectivity() const
{
  return m_vReflectivity;
}

inline VOID CSLIDESurface::SetExponent(FLOAT fExponent)
{
  m_dynfExponent.SetValue(fExponent);
}

inline VOID CSLIDESurface::SetExponent(DynFloat *pdynfExponent)
{
  m_dynfExponent.Set(pdynfExponent);
}

inline FLOAT CSLIDESurface::GetExponent() const
{
  return m_dynfExponent.GetValue();
}

inline VOID CSLIDESurface::SetMetallic(FLOAT fMetallic)
{
  m_dynfMetallic.SetValue(fMetallic);
}

inline VOID CSLIDESurface::SetMetallic(DynFloat *pdynfMetallic)
{
  m_dynfMetallic.Set(pdynfMetallic);
}

inline FLOAT CSLIDESurface::GetMetallic() const
{
  return m_dynfMetallic.GetValue();
}

inline const Vector &CSLIDESurface::GetAveTexColor() const
{
  return m_vAveTexColor;
}

inline CSLIDESurface *CSLIDESurface::GetSLF_INHERIT()
{
  return m_pCSLIDESurfaceDefault;
}

inline VOID CSLIDESurface::SetTexture(CHAR *pcTextureFile,DynInt* pdyniTexture)
{
  m_pcTextureFile = pcTextureFile;
  m_dyniTexture.Set(pdyniTexture);
}

inline SLIDETextureType CSLIDESurface::GetTextureMode() const
{
  return (SLIDETextureType)m_dyniTexture.GetValue();
}

#endif // __SLIDESURFACE_H__


