#ifndef __SLIDESWEEP_H__
#define __SLIDESWEEP_H__

#include "slideatomnode.h"
#include "slidesweepcrosssection.h"
#include "slidesweeppath.h"

//////////////////////////////////////////////////////////////////////////////
// CSLIDESweep Class
//

class CSLIDESweep : public CSLIDEAtomNode
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDESweep();
  virtual ~CSLIDESweep();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID Uninit();
  virtual BOOL Initialized();
  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);
  virtual BOOL Valid();

  inline VOID SetPathID(CHAR *pcID);
  inline VOID SetPathMinimizeTorsion(DynInt *pdyniMinimizeTorsion);
  inline VOID SetPathClosed(DynInt *pdyniClosed);
  inline VOID SetPathSymmetry(DynInt *pdyniCrossSectionSymmetry);
  inline VOID SetPathAzimuth(DynFloat *pdynfAzimuth);
  inline VOID SetPathTwist(DynFloat *pdynfTwist);

  VOID AddCrossSectionID(CHAR *pcID, CSLIDESweepCrossSection *pCSLIDESweepCrossSection);

  inline VOID SetMeshType(DynInt *pdyniMeshType);
  inline VOID SetDrawSweep(DynInt *pdyniDrawSweep);
  inline VOID SetDrawPath(DynInt *pdyniDrawPath);

protected:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  VOID UpdatePathPoints();

protected:
  virtual VOID CheckDynamic();
  virtual VOID PreprocessGraphLocal();
  virtual VOID Update();
  virtual VOID UpdateState();
  virtual VOID CalculateBound();
  virtual VOID CalculateQEMesh();
  virtual VOID CalculateLeafQEMesh();  // for WriteWRLZ
  virtual VOID DeallocateLeafQEMesh();  // for WriteWRLZ

//////////////////////////////////////////////////
// Rendering
//
public:

protected:
  virtual VOID RenderGeometrySmooth(CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderGeometryFlat(CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderNormalsSmooth();
  virtual VOID RenderNormalsFlat();

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

  virtual VOID WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				 CSLIDESurface *pCSLIDESurface);
  virtual VOID WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline CHAR *GetTypeString();
  virtual inline SLIDEReservedWordType GetType();

  virtual inline BOOL GetChildrenBoundsDynamic();
  inline VOID SetChildrenBoundsDynamic(BOOL bChildrenBoundsDynamic);

  inline UINT_32 GetPathSlices();
  inline Point* GetPathPoint(); 
  inline Matrix* GetPathFrenetMatrix(); 

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  FLOAT m_fNormalLength;

  // Cross Section
  CListInt m_listCrossSectionIDs;
  CListInt m_listCSLIDESweepCrossSection;

  // Path
  CHAR *m_pcPathID;
  CSLIDESweepPath *m_pPath;

  DynInt m_dyniMinimizeTorsion;
  DynInt m_dyniClosed;
  DynInt m_dyniSymmetry;
  DynFloat m_dynfAzimuth;
  DynFloat m_dynfTwist;

  // General
  DynInt m_dyniMeshType;
  DynInt m_dyniDrawPath;
  DynInt m_dyniDrawSweep;

  BOOL m_bChildrenBoundsDynamic;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDESweep::SetPathID(CHAR *pcID)
{
  m_pcPathID = pcID;
}

inline VOID CSLIDESweep::SetPathMinimizeTorsion(DynInt *pdyniMinimizeTorsion)
{
  m_dyniMinimizeTorsion.Set(pdyniMinimizeTorsion);
}

inline VOID CSLIDESweep::SetPathClosed(DynInt *pdyniClosed)
{
  m_dyniClosed.Set(pdyniClosed);
}

inline VOID CSLIDESweep::SetPathAzimuth(DynFloat *pdynfAzimuth)
{
  m_dynfAzimuth.Set(pdynfAzimuth);
}

inline VOID CSLIDESweep::SetPathTwist(DynFloat *pdynfTwist)
{
  m_dynfTwist.Set(pdynfTwist);
}

inline VOID CSLIDESweep::SetPathSymmetry(DynInt *pdyniPathSymmetry)
{
  m_dyniSymmetry.Set(pdyniPathSymmetry);
}

inline VOID CSLIDESweep::SetMeshType(DynInt *pdyniMeshType)
{
  m_dyniMeshType.Set(pdyniMeshType);
}

inline VOID CSLIDESweep::SetDrawSweep(DynInt *pdyniDrawSweep)
{
  m_dyniDrawSweep.Set(pdyniDrawSweep);
}

inline VOID CSLIDESweep::SetDrawPath(DynInt *pdyniDrawPath)
{
  m_dyniDrawPath.Set(pdyniDrawPath);
}

inline CHAR *CSLIDESweep::GetTypeString()
{
  return gcppcSLIDEReservedWords[sliderwtSweep];
}

inline SLIDEReservedWordType CSLIDESweep::GetType()
{
  return sliderwtSweep;
}

inline BOOL CSLIDESweep::GetChildrenBoundsDynamic()
{
  return m_bChildrenBoundsDynamic;
}

inline VOID CSLIDESweep::SetChildrenBoundsDynamic(BOOL bChildrenBoundsDynamic)
{
  m_bChildrenBoundsDynamic = bChildrenBoundsDynamic;
}

inline UINT_32 CSLIDESweep::GetPathSlices()
{
  return m_pPath->GetSlices();
}

inline Point* CSLIDESweep::GetPathPoint()
{
  return m_pPath->GetPoint();
}

inline Matrix* CSLIDESweep::GetPathFrenetMatrix()
{
  return m_pPath->GetFrenetMatrix();
}

#endif // __SLIDESWEEP_H__
