#include "slidesweepcrosssection.h"
#include "slidesweep.h"
#include "slideworld.h"
#include "slidesurface.h"
#include "slidetransform.h"

#include "qemesh.h"
#include "qevertex.h"
#include "qerender.h"
#include "qewrite.h"

static const UINT_32 cuiOrientation = 0;
static const UINT_32 cuiBeginCap = 0;
static const UINT_32 cuiEndCap = 0;

//////////////////////////////////////////////////////////////////////
// CSLIDESweepCrossSection Class
//

CSLIDESweepCrossSection::CSLIDESweepCrossSection()
{
  m_pCSLIDESweepParent = NULL;

  m_pCSLIDECrossSection = NULL;

  m_pcSurfaceID = NULL;
  m_pCSLIDESurface = NULL;

  m_pvTexture[0].Set(0.0, 0.0, 0.0);
  m_pvTexture[1].Set(1.0, 0.0, 0.0);
  m_pvTexture[2].Set(1.0, 1.0, 0.0);
  m_pvTexture[3].Set(0.0, 1.0, 0.0);

  m_dyniBeginCap.SetValue(cuiBeginCap);
  m_dyniEndCap.SetValue(cuiEndCap);
  m_dyniOrientation.SetValue(cuiOrientation);

  m_pQEMeshBeginCap = NULL;
  m_pQEMeshEndCap = NULL;

  m_mTransform.SetIdentity(); 
  m_bTransformDynamic = FALSE;

  m_bInitialize = TRUE;
}

CSLIDESweepCrossSection::~CSLIDESweepCrossSection()
{
}

VOID CSLIDESweepCrossSection::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
				   CHAR *pcID)
{
  CSLIDEAtomNode::Init(pcFile, uiLine, uiColumn, pcID);

  m_dyniBeginCap.SetValue(cuiBeginCap);
  m_dyniEndCap.SetValue(cuiEndCap);
  m_dyniOrientation.SetValue(cuiOrientation);

  // Initialize
  m_mTransform.SetIdentity(); 
  m_listCSLIDETransform.Init();

  m_pvTexture[0].Set(0.0, 0.0, 0.0);
  m_pvTexture[1].Set(1.0, 0.0, 0.0);
  m_pvTexture[2].Set(1.0, 1.0, 0.0);
  m_pvTexture[3].Set(0.0, 1.0, 0.0);

  m_bInitialize = TRUE;
}

VOID CSLIDESweepCrossSection::Uninit()
{
  CSLIDEAtomNode::Uninit();

 if ( m_pQEMeshBeginCap != NULL )
    {
      DeallocateQEMesh(m_pQEMeshBeginCap);
      m_pQEMeshBeginCap = NULL;
    }

 if ( m_pQEMeshEndCap != NULL )
    {
      DeallocateQEMesh(m_pQEMeshEndCap);
      m_pQEMeshEndCap = NULL;
    }
}

BOOL CSLIDESweepCrossSection::Initialized()
{
  if ( !CSLIDEAtomNode::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

VOID CSLIDESweepCrossSection::InsertTransform(CSLIDETransform *pCSLIDETransform)
{
  ASSERT( pCSLIDETransform != NULL );

  m_listCSLIDETransform.InsertLast(pCSLIDETransform, 0);
}

VOID CSLIDESweepCrossSection::SetSweepParent(CSLIDESweep *pCSLIDESweep)
{
  ASSERT( pCSLIDESweep != NULL );

  m_pCSLIDESweepParent = pCSLIDESweep;
}

BOOL CSLIDESweepCrossSection::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  if ( !CSLIDEAtomNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  BOOL bCorrect;
  CListIterInt iter;
  CSLIDETransform *pCSLIDETransform;


  bCorrect = TRUE;

  // Check Surface Reference
  if( m_pcSurfaceID != NULL )
    {
      m_pCSLIDESurface = pCSLIDEWorld->PeekSurface(m_pcSurfaceID);
      if ( m_pCSLIDESurface == NULL )
	{
	  HandleError("Undefined %s \"%s\"", 
		      gcppcSLIDEReservedWords[sliderwtSurface],
		      m_pcSurfaceID);
	  bCorrect = FALSE;
	}
    }

  // Initialize Transform
  iter.Init(&m_listCSLIDETransform);
  for ( pCSLIDETransform = (CSLIDETransform*)iter.PeekFirst(); 
	pCSLIDETransform != NULL;
	pCSLIDETransform = (CSLIDETransform*)iter.PeekNext() )
    {
      m_mTransform = pCSLIDETransform->GetMatrix() * m_mTransform;      
    }

  return bCorrect;
}

BOOL CSLIDESweepCrossSection::Valid()
{
  if ( !CSLIDEAtomNode::Valid() )
    {
      return FALSE;
    }

  return ( (m_dyniBeginCap.GetValue() >= 0) &&
	   (m_dyniEndCap.GetValue() >= 0) &&
	   ( m_dyniOrientation.GetValue() >= 0 ) );
}

VOID CSLIDESweepCrossSection::CheckDynamic()
{
  CSLIDEAtomNode::CheckDynamic();

  if( !GetDynamic() )
    {
      SetDynamic( m_dyniBeginCap.GetDynamic() ||
		  m_dyniEndCap.GetDynamic() ||
		  m_dyniOrientation.GetDynamic() ||
		  m_pvTexture[0].GetDynamic() ||
		  m_pvTexture[1].GetDynamic() ||
		  m_pvTexture[2].GetDynamic() ||
		  m_pvTexture[3].GetDynamic());
    }

  CListIterInt iter;
  CSLIDETransform *pCSLIDETransform;

  iter.Init(&m_listCSLIDETransform);
  for ( pCSLIDETransform = (CSLIDETransform*)iter.PeekFirst(); 
	pCSLIDETransform != NULL;
	pCSLIDETransform = (CSLIDETransform*)iter.PeekNext() )
    {
      if ( pCSLIDETransform->GetDynamic() )
	{ 
	  m_bTransformDynamic = TRUE;
	  SetDynamic(TRUE);
	  break;
	}
    }

  if( !GetDynamic() )
    {
      ASSERT( m_pCSLIDECrossSection != NULL );
      if( m_pCSLIDECrossSection->GetDynamic() )
	{
	  SetDynamic(TRUE);
	}
    }
}

VOID CSLIDESweepCrossSection::PreprocessGraphLocal()
{
  BOOL bRealSpaceDynamic;
  BOOL bParameterSpaceDynamic;
  BOOL bTextureSpaceDynamic;
  BOOL bBoundDynamic;
  BOOL bGeometryDynamic;
  BOOL bNormalsDynamic;

  // Crosssection may not be instanced so call preprocessgraph
  m_pCSLIDECrossSection->PreprocessGraphLocal();

  // crosssection set these appropriately
  bBoundDynamic = GetBoundDynamic();
  bGeometryDynamic = GetGeometryDynamic();
  bNormalsDynamic = GetNormalsDynamic();


  bRealSpaceDynamic = FALSE;
  bTextureSpaceDynamic = ( m_pvTexture[0].GetDynamic() ||
			   m_pvTexture[1].GetDynamic() ||
			   m_pvTexture[2].GetDynamic() ||
			   m_pvTexture[3].GetDynamic() );
  bParameterSpaceDynamic = ( m_bTransformDynamic ||
			     m_dyniBeginCap.GetDynamic() ||
			     m_dyniEndCap.GetDynamic() ||
			     m_dyniOrientation.GetDynamic() );

  SetGeometryEmpty(FALSE);
  SetGeometryDynamic(bRealSpaceDynamic || 
		     bParameterSpaceDynamic ||
		     bTextureSpaceDynamic ||
		     bGeometryDynamic ||
		     ((m_pCSLIDESurface != NULL) && m_pCSLIDESurface->GetDynamic()) );
  SetNormalsEmpty(FALSE);
  SetNormalsDynamic(bRealSpaceDynamic || 
		    bParameterSpaceDynamic ||
		    bNormalsDynamic);
  SetBoundsEmpty(FALSE);
  SetBoundsDynamic(bRealSpaceDynamic ||
		   bParameterSpaceDynamic ||
		   bBoundDynamic);
  SetBoundEmpty(FALSE);
  SetBoundDynamic(bRealSpaceDynamic ||
		  bParameterSpaceDynamic ||
		  bBoundDynamic);
  SetChildrenBoundsDynamic(bRealSpaceDynamic ||
			   bParameterSpaceDynamic ||
			   bBoundDynamic);

  /////////////////////////////////////////////////////////////////////////////////////
  if( GetBoundDynamic() )
    m_pCSLIDESweepParent->SetBoundDynamic(TRUE);
  
  if( GetGeometryDynamic() )
    m_pCSLIDESweepParent->SetGeometryDynamic(TRUE);
  
  if( GetNormalsDynamic() )
    m_pCSLIDESweepParent->SetNormalsDynamic(TRUE);
}

VOID CSLIDESweepCrossSection::Update()
{
  // For End Caps
  SetBoundChanged(FALSE);
  if ( GetDynamic() ||
       GetChildrenBoundsDynamic() )
    {
      if ( GetChildrenBoundsDynamic() )
	{
	  CSLIDEBoundedNode::InvalidateQEMesh();
	}
      UpdateState();
    }

  if ( GetBoundDynamic() )
    {
      SetBoundValid(FALSE);
    }

  SetVisited(FALSE);
}

VOID CSLIDESweepCrossSection::UpdateState()
{
  CSLIDEAtomNode::UpdateState();

  CListIterInt iter;
  CSLIDETransform *pCSLIDETransform;
  UINT_32 i;
  BOOL bBeginCapChanged, bEndCapChanged, bCrossSectionChanged;
  Matrix m;

  bCrossSectionChanged = FALSE;
  bBeginCapChanged = FALSE;
  bEndCapChanged = FALSE;

  i = m_dyniOrientation.GetValue();
  m_dyniOrientation.Update();
  bCrossSectionChanged = ( bCrossSectionChanged || 
			   ( i != (UINT_32)m_dyniOrientation.GetValue()) );

  i = m_dyniBeginCap.GetValue();
  m_dyniBeginCap.Update();
  bBeginCapChanged = ( i != (UINT_32)m_dyniBeginCap.GetValue() );

  i = m_dyniEndCap.GetValue();
  m_dyniEndCap.Update();
  bEndCapChanged = ( i != (UINT_32)m_dyniEndCap.GetValue() );

  for ( i = 0; i < 4; i++ )
    {
      m_pvTexture[i].Update();
    }

  if( m_bTransformDynamic )
    {
      m = m_mTransform;
      m_mTransform.SetIdentity(); 

      iter.Init(&m_listCSLIDETransform);
      for ( pCSLIDETransform = (CSLIDETransform*)iter.PeekFirst(); 
	    pCSLIDETransform != NULL;
	    pCSLIDETransform = (CSLIDETransform*)iter.PeekNext() )
	{
	  m_mTransform = pCSLIDETransform->GetMatrix() * m_mTransform;      
	}
    }

  if( m_bInitialize )
    {
      m_pCSLIDECrossSection->CalculateCrossSection(m_dyniOrientation.GetValue());
      CalculateEndCaps();   
      m_bInitialize = FALSE;
    }
  else
    {
      // Recalculate Cross Section
      if( bCrossSectionChanged )
	{
	  m_pCSLIDECrossSection->CalculateCrossSection(m_dyniOrientation.GetValue());
	}
      
      // Recalculate End Caps
      if( (m_dyniBeginCap.GetValue() &&
	   (bCrossSectionChanged || (m != m_mTransform) || 
	    bBeginCapChanged || (m_pQEMeshBeginCap == NULL))) ||
	  (m_dyniEndCap.GetValue() &&
	   (bCrossSectionChanged || (m != m_mTransform) || 
	    bEndCapChanged || (m_pQEMeshEndCap == NULL))) )
	{
	  CalculateEndCaps();      
	}
    }
}

VOID CSLIDESweepCrossSection::CalculateBound()
{
  Point pt;
  CBound bound;

  bound = m_pCSLIDECrossSection->GetBound();
  bound.Insert(pt);

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);
    }
}

///////////////////////////////////////////////////////////////////////////////

VOID CSLIDESweepCrossSection::CalculateEndCaps()
{
  UINT_32 uiSlices;
  Point pt1, pt2;
  Point *pPathPoint;
  Matrix *pmPathFrenet;
  Matrix m1, m2;
  BOOL bAreaOK;

  bAreaOK = TRUE;
  uiSlices = m_pCSLIDESweepParent->GetPathSlices();
  pPathPoint = m_pCSLIDESweepParent->GetPathPoint();
  pmPathFrenet = m_pCSLIDESweepParent->GetPathFrenetMatrix();

  pt1 = pPathPoint[0];
  pt2 = pPathPoint[uiSlices];
  m1 = pmPathFrenet[0];
  m2 = pmPathFrenet[uiSlices];

  if( m_pCSLIDECrossSection->GetPoints() < 3 )
    {
      return;
    }

  if( m_dyniBeginCap.GetValue() )
    {
      // First End Cap 
      bAreaOK = CalculateGeneralFace(m_pQEMeshBeginCap, pt1, m1, FALSE);

      if( !bAreaOK )
	{
	  // Bad cross section -> zero area
	  m_pQEMeshBeginCap = NULL;
	HandleWarning("Could not calculate begincap for sweep \"%s\" crosssection \"%s\"\n",
			m_pCSLIDESweepParent->GetID(),
			GetID());
	}
    }

  if( m_dyniEndCap.GetValue() )
    {
      // Last End Cap 
      bAreaOK = CalculateGeneralFace(m_pQEMeshEndCap, pt2, m2, TRUE);

      if( !bAreaOK )
	{
	  // Bad cross section -> zero area
	  m_pQEMeshEndCap = NULL;	  
	  HandleWarning("Could not calculate endcap for sweep \"%s\" crosssection \"%s\"\n",
			m_pCSLIDESweepParent->GetID(),
			GetID());
	}
    }
}

BOOL CSLIDESweepCrossSection::CalculateGeneralFace(QEMesh *(&pQEMesh), Point ptSweep, 
						   Matrix mSweep, BOOL bFlipOrientation)
{
  UINT_32 uiCSSlice, uiCrossSectionPoints, uiCrossSectionEndCapPoints;
  Point ptCentroid, ptStart, ptPrev, ptCurr;
  Point *pEndCapPoints, *pCrossSectionPoint;
  Vector vNormal;

  uiCrossSectionPoints = m_pCSLIDECrossSection->GetPoints();
  uiCrossSectionEndCapPoints = m_pCSLIDECrossSection->GetEndCapPoints();
  pCrossSectionPoint = m_pCSLIDECrossSection->GetPoint();

  ////////////////////////////////////////////////////////////////////////
  // Transform cross section points along path

  pEndCapPoints = new Point[uiCrossSectionPoints];

  if( bFlipOrientation )
    {
      UINT_32 index;
      for( uiCSSlice = 0; uiCSSlice < uiCrossSectionPoints; uiCSSlice++)
	{
	  index = uiCrossSectionPoints-uiCSSlice-1;
	  pEndCapPoints[index] = ptSweep + mSweep * m_mTransform * pCrossSectionPoint[uiCSSlice];
	  pEndCapPoints[index][W] = 1.0;
	}
    }
  else
    {
      for( uiCSSlice = 0; uiCSSlice < uiCrossSectionPoints; uiCSSlice++)
	{
	  pEndCapPoints[uiCSSlice] = ptSweep + mSweep * m_mTransform * pCrossSectionPoint[uiCSSlice];
	  pEndCapPoints[uiCSSlice][W] = 1.0;
	}
    }

  ////////////////////////////////////////////////////////////////////////
  // CalculateCentroid; assuming closed so Points-1

  ptCentroid.Set(0.0, 0.0, 0.0, 0.0);
  for( uiCSSlice = 0; uiCSSlice < uiCrossSectionEndCapPoints; uiCSSlice++)
    {
      ptCentroid += pEndCapPoints[uiCSSlice];
    }
  ptCentroid /= uiCrossSectionEndCapPoints;
  ptCentroid[W] = 1.0;

  ////////////////////////////////////////////////////////////////////////
  // CalculateNormal
  
  vNormal.Null();
  ptStart = pEndCapPoints[0];
  ptPrev = ptStart;

  for( uiCSSlice = 0; uiCSSlice < uiCrossSectionEndCapPoints; uiCSSlice++)
    {
      ptCurr = pEndCapPoints[uiCSSlice];
      ptCurr[W] = 1.0;
      
      vNormal[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
      vNormal[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
      vNormal[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);

      ptPrev = ptCurr;
    }
  ptCurr = ptStart;
  vNormal[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
  vNormal[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
  vNormal[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);

  if ( !vNormal.Normalize() )
    {
      HandleWarning("Null normal because of zero area cross section");
      printf(       "      End Caps cannot be calculated.\n");

      return FALSE;
    }

  ////////////////////////////////////////////////////////////////////////

  if ( pQEMesh != NULL )
    {
      DeallocateQEMesh(pQEMesh);
      pQEMesh = NULL;
    }

  pQEMesh = AllocateQEMesh();
  ASSERT( pQEMesh != NULL );
  pQEMesh->Init();

  CListInt listContours;
  CObjectList<QEVertex> *plistQEVertex;
  QEVertex *pvx;
  CListIterInt iter;
  Point pt;
  Matrix mGobj_xy, mGxy_obj;
  
  //////////////////////////////////////////////////
  // Calculate the Gobj_xy transform (transform pts to xy plane)
  {
    Vector vZ, vAxis, vTrans;
    Point ptOrigin;
    FLOAT fSin, fCos, fTheta;
    Matrix mTrans, mTransInv, mRot, mRotInv;
    
    ptOrigin.Set(0.0, 0.0, 0.0);
    vZ.Set(0.0, 0.0, 1.0);
    
    vTrans = ptOrigin - ptCentroid;
    mTrans.SetTranslate(vTrans);
    vTrans = -vTrans;
    mTransInv.SetTranslate(vTrans);
    
    vAxis = vNormal * vZ;
    fCos = vNormal.Dot(vZ);
    fSin = vAxis.Magnitude();
    if ( !vAxis.Normalize() )
      {
	vAxis.Set(1.0, 0.0, 0.0);
	if ( fCos < 0.0 )
	  {
	    fTheta = PI;
	  }
	else
	  {
	    fTheta = 0.0;
	  }
      }
    else
      {
	fTheta = atan2(fSin, fCos);
      }
    mRot.SetRotate(vAxis, fTheta);
    mRotInv.SetRotate(vAxis, -fTheta);	
    
    mGxy_obj = mRot * mTrans;
    mGobj_xy = mTransInv * mRotInv;
  }
  // Calculate the Gobj_xy transform
  //////////////////////////////////////////////////

  listContours.Init();

  plistQEVertex = new CObjectList<QEVertex>;
  ASSERT( plistQEVertex != NULL );
  listContours.InsertLast(plistQEVertex, 0);

  for( uiCSSlice = 0; uiCSSlice < uiCrossSectionEndCapPoints; uiCSSlice++)
    {
      pvx = AllocateQEVertex();
      ASSERT( pvx != NULL );
      pt = mGxy_obj * pEndCapPoints[uiCSSlice];
      pt[2] = 0.0;
      pvx->Init(pt);

      //printf( "%f %f %f\n", pt[X], pt[Y], pt[Z]);
	  
      plistQEVertex->InsertLast(pvx);
    }

  QEMesh::SetRandomize(FALSE);
  //QEMesh::SetAccuracyType(QEMesh::atFloat);
  QEMesh::SetAccuracyType(QEMesh::atExact);

  pQEMesh->CreateTriangulatedPolygon(&listContours);
  pQEMesh->Transform(mGobj_xy, mGxy_obj);

  delete plistQEVertex;
  delete[] pEndCapPoints;

  return TRUE;
}

////////////////////////////////////////////////////////////////////////////////////////////

VOID CSLIDESweepCrossSection::RenderEndCapGeometry()
{
  if( m_dyniBeginCap.GetValue() && (m_pQEMeshBeginCap != NULL) )
    {
      ::RenderGeometry(m_pQEMeshBeginCap);
    }

  if( m_dyniEndCap.GetValue() && (m_pQEMeshEndCap != NULL) )
    {
      ::RenderGeometry(m_pQEMeshEndCap);
    }
}

VOID CSLIDESweepCrossSection::WriteRIBEndCaps(FILE *fpOutputFile)
{
  Matrix m;

  if( m_dyniBeginCap.GetValue() && (m_pQEMeshBeginCap != NULL) )
    {
      ::WriteRIB(m_pQEMeshBeginCap, fpOutputFile, m);    
    }

  if( m_dyniEndCap.GetValue() && (m_pQEMeshEndCap != NULL) )
    {
      ::WriteRIB(m_pQEMeshEndCap, fpOutputFile, m);
    }
}

VOID CSLIDESweepCrossSection::WriteWRLZEndCaps(FILE *fpOutputFile, const Matrix &mTransform)
{
  if( m_dyniBeginCap.GetValue() && (m_pQEMeshBeginCap != NULL ) )
    {
      ::WriteWRLZ(m_pQEMeshBeginCap, fpOutputFile, mTransform);
    }

  if( m_dyniEndCap.GetValue() && (m_pQEMeshEndCap != NULL) )
    {
      ::WriteWRLZ(m_pQEMeshEndCap, fpOutputFile, mTransform);
    }
}

VOID CSLIDESweepCrossSection::WriteSTLEndCaps(FILE *fpOutputFile, const Matrix &mTransform)
{
  if( m_dyniBeginCap.GetValue() && (m_pQEMeshBeginCap != NULL ) )
    {
      ::WriteSTL(m_pQEMeshBeginCap, fpOutputFile, mTransform);
    }

  if( m_dyniEndCap.GetValue() && (m_pQEMeshEndCap != NULL) )
    {
      ::WriteSTL(m_pQEMeshEndCap, fpOutputFile, mTransform);
    }
}

////////////////////////////////////////////////////////////////////////////////////////////

VOID CSLIDESweepCrossSection::WriteSLIDE(FILE */*fpOutputFile*/)
{
}

VOID CSLIDESweepCrossSection::WriteSLIDEParent(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  CListIterInt iter;
  CSLIDETransform *pCSLIDETransform;

  fprintf(fpOutputFile, "  %s %s\n",
	  gcppcSLIDEReservedWords[sliderwtCrossSection], GetID());

  fprintf(fpOutputFile, "    %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtBeginCap],
	  m_dyniBeginCap.GetValue());

  fprintf(fpOutputFile, "    %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtEndCap],
	  m_dyniEndCap.GetValue());

  fprintf(fpOutputFile, "    %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtOrientation],
	  m_dyniOrientation.GetValue());

  fprintf(fpOutputFile, "    %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtSurface], 
	      m_pcSurfaceID);

  fprintf(fpOutputFile, "    %s ( %.3f %.3f %.3f ) ( %.3f %.3f %.3f ) ( %.3f %.3f %.3f ) ( %.3f %.3f %.3f )\n",
	  gcppcSLIDEReservedWords[sliderwtTexture],
	  m_pvTexture[0](0), m_pvTexture[0](1), m_pvTexture[0](2), 
	  m_pvTexture[1](0), m_pvTexture[1](1), m_pvTexture[1](2), 
	  m_pvTexture[2](0), m_pvTexture[2](1), m_pvTexture[2](2), 
	  m_pvTexture[3](0), m_pvTexture[3](1), m_pvTexture[3](2));

  fprintf(fpOutputFile, "    %s\n",
	      gcppcSLIDEReservedWords[sliderwtTransform] );

  iter.Init(&m_listCSLIDETransform);
  for ( pCSLIDETransform = (CSLIDETransform*)iter.PeekFirst(); 
	pCSLIDETransform != NULL;
	pCSLIDETransform = (CSLIDETransform*)iter.PeekNext() )
    {
      fprintf(fpOutputFile, "      " );
      pCSLIDETransform->WriteSLIDE(fpOutputFile);
    }

  fprintf(fpOutputFile, "    %s\n",
	      gcppcSLIDEReservedWords[sliderwtEndTransform] );

  fprintf(fpOutputFile, "  %s\n",
	  gcppcSLIDEReservedWords[sliderwtEndCrossSection]);
}

VOID CSLIDESweepCrossSection::WriteRIBGeometry(FILE */*fpOutputFile*/,
					  const Matrix &mGproj_obj,
					  CBound *pCBoundClip, 
					  SLIDEShadingType eShading,
					  SLIDELODType eLOD)
{
}

VOID CSLIDESweepCrossSection::WriteRIBPushSLIDEName(FILE */*fpOutputFile*/)
{
}

VOID CSLIDESweepCrossSection::WriteRIBPopSLIDEName(FILE */*fpOutputFile*/)
{
}
