#ifndef __SLIDESWEEPCROSSSECTION_H__
#define __SLIDESWEEPCROSSSECTION_H__

#include "slideatomnode.h"
#include "slidesurface.h"
#include "slidecrosssection.h"
#include "slidepolyline.h"
#include "slidebsplinecurve.h"

class CSLIDETransform;

//////////////////////////////////////////////////////////////////////////////
// CSLIDESweepCrossSection Class
//

class CSLIDESweepCrossSection : public CSLIDEAtomNode
{
//////////////////////////////////////////////////
// Friends
//
public: friend class CSLIDESweep;
public: friend class CSLIDECrossSection;

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDESweepCrossSection();
  virtual ~CSLIDESweepCrossSection();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID Uninit();
  virtual BOOL Initialized();
  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);
  virtual BOOL Valid();

  inline VOID SetCrossSection(CSLIDECrossSection *pCSLIDECrossSection);
  inline VOID SetSurfaceID(CHAR *pcSurfaceID);
  inline VOID SetTexture(DynVector *pdynv0, DynVector *pdynv1, 
			 DynVector *pdynv2, DynVector *pdynv3);
  inline VOID SetEndCap(DynInt *pdyniEndCap);
  inline VOID SetBeginCap(DynInt *pdyniBeginCap);
  inline VOID SetOrientation(DynInt *pdyniCrossSectionOrientation);
  VOID InsertTransform(CSLIDETransform *pCSLIDETransform);

  VOID SetSweepParent(CSLIDESweep *pCSLIDESweep);

protected:


//////////////////////////////////////////////////
// Geometric Computations
//
public:

protected:
  virtual VOID CheckDynamic();
  virtual VOID PreprocessGraphLocal();
  virtual VOID Update();
  virtual VOID UpdateState();
  virtual VOID CalculateBound();

protected:
  virtual VOID CalculateEndCaps();
  virtual BOOL CalculateGeneralFace(QEMesh *(&pQEMesh), Point pt, Matrix m, 
				    BOOL bFlipOrientation);

//////////////////////////////////////////////////
// Rendering
//
public:
  VOID RenderEndCapGeometry();

  inline VOID PushSurface();
  inline VOID PopSurface();

protected:

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);
  virtual VOID WriteSLIDEParent(FILE *fpOutputFile);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

  virtual VOID WriteRIBEndCaps(FILE *fpOutputFile);
  virtual VOID WriteWRLZEndCaps(FILE *fpOutputFile, const Matrix &mTransform);
  virtual VOID WriteSTLEndCaps(FILE *fpOutputFile, const Matrix &mTransform);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline CSLIDESurface *GetSurface();

  inline UINT_32 GetPoints();
  inline Point* GetPoint();
  inline Vector* GetNormal();
  inline DynVector* GetTexture();
  inline Matrix GetTransform();

  inline UINT_32 GetOrientationValue();
  inline QEMesh* GetMeshBeginCap();
  inline QEMesh* GetMeshEndCap();
  inline UINT_32 GetBeginCapValue();
  inline UINT_32 GetEndCapValue();

  virtual inline BOOL GetChildrenBoundsDynamic();
  inline VOID SetChildrenBoundsDynamic(BOOL bChildrenBoundsDynamic);

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:

  // Sweep
  CSLIDESweep *m_pCSLIDESweepParent;

  // Cross Section
  CSLIDECrossSection *m_pCSLIDECrossSection;

  CHAR *m_pcSurfaceID;
  CSLIDESurface *m_pCSLIDESurface;
  DynVector m_pvTexture[4];

  DynInt m_dyniOrientation;    // 0 = outward, 1 = inward

  DynInt m_dyniBeginCap;
  DynInt m_dyniEndCap;
  QEMesh *m_pQEMeshBeginCap;
  QEMesh *m_pQEMeshEndCap;

  BOOL m_bTransformDynamic;
  CListInt m_listCSLIDETransform;
  Matrix m_mTransform;

  BOOL m_bInitialize;

  BOOL m_bChildrenBoundsDynamic;

  GLbitfield m_GLbAttributesMask;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDESweepCrossSection:: SetCrossSection(CSLIDECrossSection *pCSLIDECS)
{
  m_pCSLIDECrossSection = pCSLIDECS;
}

inline VOID CSLIDESweepCrossSection::SetSurfaceID(CHAR *pcSurfaceID)
{
  m_pcSurfaceID = pcSurfaceID;
}

inline VOID CSLIDESweepCrossSection::SetTexture(DynVector *pdynv0, DynVector *pdynv1, 
						DynVector *pdynv2, DynVector *pdynv3)
{
  m_pvTexture[0].SetAll(pdynv0);
  m_pvTexture[1].SetAll(pdynv1);
  m_pvTexture[2].SetAll(pdynv2);
  m_pvTexture[3].SetAll(pdynv3);
}

inline VOID CSLIDESweepCrossSection::SetOrientation(DynInt *pdyniOrientation)
{
  m_dyniOrientation.Set(pdyniOrientation);
}

inline VOID CSLIDESweepCrossSection::SetBeginCap(DynInt *pdyniBeginCap)
{
  m_dyniBeginCap.Set(pdyniBeginCap);
}

inline VOID CSLIDESweepCrossSection::SetEndCap(DynInt *pdyniEndCap)
{
  m_dyniEndCap.Set(pdyniEndCap);
}

inline CSLIDESurface *CSLIDESweepCrossSection::GetSurface()
{
  if( m_pCSLIDESurface != NULL )
    {
      return m_pCSLIDESurface;
    }
  else
    {
      return m_pCSLIDECrossSection->GetSurface();
    }
}

inline UINT_32 CSLIDESweepCrossSection::GetPoints()
{
  return m_pCSLIDECrossSection->GetPoints();
}

inline Point* CSLIDESweepCrossSection::GetPoint()
{
  return m_pCSLIDECrossSection->GetPoint();
}

inline Vector* CSLIDESweepCrossSection::GetNormal()
{
  return m_pCSLIDECrossSection->GetNormal();
}

inline DynVector* CSLIDESweepCrossSection::GetTexture()
{
  return m_pvTexture;
}

inline UINT_32 CSLIDESweepCrossSection::GetOrientationValue()
{
  return m_dyniOrientation.GetValue();
}

inline QEMesh* CSLIDESweepCrossSection::GetMeshBeginCap()
{
  return m_pQEMeshBeginCap;
}

inline QEMesh* CSLIDESweepCrossSection::GetMeshEndCap()
{
  return m_pQEMeshEndCap;
}

inline UINT_32 CSLIDESweepCrossSection::GetBeginCapValue()
{
  return m_dyniBeginCap.GetValue();
}

inline UINT_32 CSLIDESweepCrossSection::GetEndCapValue()
{
  return m_dyniEndCap.GetValue();
}

inline Matrix CSLIDESweepCrossSection::GetTransform()
{
  return m_mTransform;
}

inline BOOL CSLIDESweepCrossSection::GetChildrenBoundsDynamic()
{
  return m_bChildrenBoundsDynamic;
}

inline VOID CSLIDESweepCrossSection::SetChildrenBoundsDynamic(BOOL bChildrenBoundsDynamic)
{
  m_bChildrenBoundsDynamic = bChildrenBoundsDynamic;
}

inline VOID CSLIDESweepCrossSection::PushSurface()
{
  CSLIDESurface *pCSLIDESurface;

  pCSLIDESurface = m_pCSLIDESurface;

  if( pCSLIDESurface == NULL )
    {
      pCSLIDESurface = m_pCSLIDECrossSection->GetSurface();
    }
  
  if( pCSLIDESurface != NULL )
    {
      pCSLIDESurface->Push();
    }
}

inline VOID CSLIDESweepCrossSection::PopSurface()
{
  CSLIDESurface *pCSLIDESurface;

  pCSLIDESurface = m_pCSLIDESurface;

  if( pCSLIDESurface == NULL )
    {
      pCSLIDESurface = m_pCSLIDECrossSection->GetSurface();
    }

  if( pCSLIDESurface != NULL )
    {
      pCSLIDESurface->Pop();
    }
}

#endif // __SLIDESWEEPCROSSSECTION_H__






