#ifndef __SLIDESWEEPMORPH_H__
#define __SLIDESWEEPMORPH_H__

#include "slideatomnode.h"
#include "slidesweeppath.h"

//////////////////////////////////////////////////////////////////////////////
// CSLIDESweepMorph Class
//

class CSLIDESweepMorph : public CSLIDEAtomNode
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDESweepMorph();
  virtual ~CSLIDESweepMorph();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID Uninit();
  virtual BOOL Initialized();
  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);
  virtual BOOL Valid();

  inline VOID SetPathID(CHAR *pcID);
  inline VOID SetPathMinimizeTorsion(DynInt *pdyniMinimizeTorsion);
  inline VOID SetPathClosed(DynInt *pdyniClosed);
  inline VOID SetPathSymmetry(DynInt *pdyniCrossSectionSymmetry);
  inline VOID SetPathAzimuth(DynFloat *pdynfAzimuth);
  inline VOID SetPathTwist(DynFloat *pdynfTwist);
  inline VOID SetEndCap(DynInt *pdyniEndCap);
  inline VOID SetBeginCap(DynInt *pdyniBeginCap);

  inline VOID SetMeshType(DynInt *pdyniMeshType);
  inline VOID SetDrawSweep(DynInt *pdyniDrawSweep);
  inline VOID SetDrawPath(DynInt *pdyniDrawPath);

protected:
  VOID ReallocateCrossSectionPoints();
  VOID DeallocateCrossSectionPoints();

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  VOID UpdatePathPoints();
  VOID UpdateCrossSectionPoints();

protected:
  virtual VOID CheckDynamic();
  virtual VOID PreprocessGraphLocal();
  virtual VOID Update();
  virtual VOID UpdateState();
  virtual VOID CalculateBound();
  virtual VOID CalculateQEMesh();
  virtual VOID CalculateLeafQEMesh();  // for WriteWRLZ
  virtual VOID DeallocateLeafQEMesh();  // for WriteWRLZ

  VOID PrintWarningCrossSectionPointsNotEqual();
  UINT_32 CheckCrossSectionPointsEqual();

  VOID CalculateCrossSection();
  VOID CalculatePolylineCrossSection(UINT_32 uiCrossSectionPoints);
  VOID CalculateBSplineCurveCrossSection(UINT_32 uiCrossSectionPoints);

  virtual VOID CalculateEndCaps();
  virtual BOOL CalculateGeneralFace(QEMesh *(&pQEMesh), 
				    CSLIDECrossSection *pCSLIDECrossSection,
				    Point pt, Matrix m, BOOL bFlipOrientation);

//////////////////////////////////////////////////
// Rendering
//
public:

protected:
  virtual VOID RenderGeometrySmooth(CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderGeometryFlat(CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderNormalsSmooth();
  virtual VOID RenderNormalsFlat();

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

  virtual VOID WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				 CSLIDESurface *pCSLIDESurface);
  virtual VOID WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline CHAR *GetTypeString();
  virtual inline SLIDEReservedWordType GetType();

  virtual inline BOOL GetChildrenBoundsDynamic();
  inline VOID SetChildrenBoundsDynamic(BOOL bChildrenBoundsDynamic);

  inline UINT_32 GetPathSlices();
  inline Point* GetPathPoint(); 
  inline Matrix* GetPathFrenetMatrix(); 

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  FLOAT m_fNormalLength;

  BOOL m_bOKToUpdate;

  // Cross Sections
  Point **m_ppCrossSectionPoint;
  UINT_32 m_uiCrossSectionPoints;
  UINT_32 m_uiCrossSectionPointsPrev;
  UINT_32 m_uiPathPoints;
  UINT_32 m_uiPathPointsPrev;

  // Path
  CHAR *m_pcPathID;
  CSLIDESweepPath *m_pPath;

  DynInt m_dyniMinimizeTorsion;
  DynInt m_dyniClosed;
  DynInt m_dyniSymmetry;
  DynFloat m_dynfAzimuth;
  DynFloat m_dynfTwist;

  DynInt m_dyniEndCapFirst;
  DynInt m_dyniEndCapLast;
  QEMesh *m_pQEMeshEndCapFirst;
  QEMesh *m_pQEMeshEndCapLast;

  // General
  DynInt m_dyniMeshType;
  DynInt m_dyniDrawPath;
  DynInt m_dyniDrawSweep;
  DynVector m_pvTexture[4];

  BOOL m_bChildrenBoundsDynamic;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDESweepMorph::SetPathID(CHAR *pcID)
{
  m_pcPathID = pcID;
}

inline VOID CSLIDESweepMorph::SetPathMinimizeTorsion(DynInt *pdyniMinimizeTorsion)
{
  m_dyniMinimizeTorsion.Set(pdyniMinimizeTorsion);
}

inline VOID CSLIDESweepMorph::SetPathClosed(DynInt *pdyniClosed)
{
  m_dyniClosed.Set(pdyniClosed);
}

inline VOID CSLIDESweepMorph::SetPathAzimuth(DynFloat *pdynfAzimuth)
{
  m_dynfAzimuth.Set(pdynfAzimuth);
}

inline VOID CSLIDESweepMorph::SetPathTwist(DynFloat *pdynfTwist)
{
  m_dynfTwist.Set(pdynfTwist);
}

inline VOID CSLIDESweepMorph::SetPathSymmetry(DynInt *pdyniPathSymmetry)
{
  m_dyniSymmetry.Set(pdyniPathSymmetry);
}

inline VOID CSLIDESweepMorph::SetMeshType(DynInt *pdyniMeshType)
{
  m_dyniMeshType.Set(pdyniMeshType);
}

inline VOID CSLIDESweepMorph::SetDrawSweep(DynInt *pdyniDrawSweep)
{
  m_dyniDrawSweep.Set(pdyniDrawSweep);
}

inline VOID CSLIDESweepMorph::SetDrawPath(DynInt *pdyniDrawPath)
{
  m_dyniDrawPath.Set(pdyniDrawPath);
}

inline VOID CSLIDESweepMorph::SetEndCap(DynInt *pdyniEndCap)
{
  m_dyniEndCapFirst.Set(pdyniEndCap);
}

inline VOID CSLIDESweepMorph::SetBeginCap(DynInt *pdyniBeginCap)
{
  m_dyniEndCapLast.Set(pdyniBeginCap);
}

inline CHAR *CSLIDESweepMorph::GetTypeString()
{
  return gcppcSLIDEReservedWords[sliderwtSweepMorph];
}

inline SLIDEReservedWordType CSLIDESweepMorph::GetType()
{
  return sliderwtSweepMorph;
}

inline BOOL CSLIDESweepMorph::GetChildrenBoundsDynamic()
{
  return m_bChildrenBoundsDynamic;
}

inline VOID CSLIDESweepMorph::SetChildrenBoundsDynamic(BOOL bChildrenBoundsDynamic)
{
  m_bChildrenBoundsDynamic = bChildrenBoundsDynamic;
}

inline UINT_32 CSLIDESweepMorph::GetPathSlices()
{
  return m_pPath->GetSlices();
}

inline Point* CSLIDESweepMorph::GetPathPoint()
{
  return m_pPath->GetPoint();
}

inline Matrix* CSLIDESweepMorph::GetPathFrenetMatrix()
{
  return m_pPath->GetFrenetMatrix();
}

#endif // __SLIDESWEEPMORPH_H__
