#include "slidesweeppath.h"
#include "slidesweep.h"
#include "slidesweepmorph.h"
#include "slidecrosssection.h"
#include "slidesurface.h"

static const UINT_32 cuiSlices = 5;

static const UINT_32 cuiMinimizeTorsion = 0;
static const UINT_32 cuiClosed = 0;
static const UINT_32 cuiSymmetry = 1;
static const FLOAT cfAzimuth = 0.0;
static const FLOAT cfTwist = 0.0;

//////////////////////////////////////////////////////////////////////
// CSLIDESweepPath Class
//

CSLIDESweepPath::CSLIDESweepPath()
{
  m_dyniSlices.SetValue(cuiSlices);
  m_uiSlices = cuiSlices;

  m_dyniMinimizeTorsion.SetValue(cuiMinimizeTorsion);
  m_dyniClosed.SetValue(cuiClosed);
  m_dyniSymmetry.SetValue(cuiSymmetry);
  m_dynfAzimuth.SetValue(cfAzimuth);
  m_dynfTwist.SetValue(cfTwist);

  m_pPoint = NULL;

  m_bSweepMorphCrossSectionDynamic = FALSE;
}

CSLIDESweepPath::~CSLIDESweepPath()
{
}

VOID CSLIDESweepPath::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID)
{
  CSLIDEAtomNode::Init(pcFile, uiLine, uiColumn, pcID);

  m_dyniSlices.SetValue(cuiSlices);
  m_uiSlices = cuiSlices;

  m_listCSLIDECrossSectionParent.Init();
  m_listCSLIDESweepParent.Init();
  m_listCSLIDESweepMorphParent.Init();

  m_bSweepMorphCrossSectionDynamic = FALSE;
}

VOID CSLIDESweepPath::Uninit()
{
  CSLIDEAtomNode::Uninit();
}

BOOL CSLIDESweepPath::Initialized()
{
  if ( !CSLIDEAtomNode::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

VOID CSLIDESweepPath::InsertCrossSectionParent(CSLIDECrossSection *pCSLIDECrossSection)
{
  ASSERT( pCSLIDECrossSection != NULL );

  m_listCSLIDECrossSectionParent.InsertLast(pCSLIDECrossSection, (UINT_32)pCSLIDECrossSection);
}

VOID CSLIDESweepPath::RemoveCrossSectionParent(CSLIDECrossSection *pCSLIDECrossSectionToRemove)
{
  ASSERT( pCSLIDECrossSectionToRemove != NULL );

  CListIterInt iter;
  CSLIDECrossSection *pCSLIDECrossSection;
  CHAR *pcID;

  pcID = pCSLIDECrossSectionToRemove->GetID();

  iter.Init(&m_listCSLIDECrossSectionParent);
  for ( pCSLIDECrossSection = (CSLIDECrossSection*)iter.PeekFirst();
	pCSLIDECrossSection != NULL; )
    {
      if( (pCSLIDECrossSection->GetID()) == pcID )
	{
	  pCSLIDECrossSection = (CSLIDECrossSection*)iter.PeekNextRemoveCurr();
	}
      else
	{
	  pCSLIDECrossSection = (CSLIDECrossSection*)iter.PeekNext();
	}
    }
}

VOID CSLIDESweepPath::InsertSweepParent(CSLIDESweep *pCSLIDESweep)
{
  ASSERT( pCSLIDESweep != NULL );

  m_listCSLIDESweepParent.InsertLast(pCSLIDESweep, (UINT_32)pCSLIDESweep);
}

VOID CSLIDESweepPath::RemoveSweepParent(CSLIDESweep *pCSLIDESweepToRemove)
{
  ASSERT( pCSLIDESweepToRemove != NULL );

  CListIterInt iter;
  CSLIDESweep *pCSLIDESweep;
  CHAR *pcID;

  pcID = pCSLIDESweepToRemove->GetID();

  iter.Init(&m_listCSLIDESweepParent);
  for ( pCSLIDESweep = (CSLIDESweep*)iter.PeekFirst();
	pCSLIDESweep != NULL; )
    {
      if( (pCSLIDESweep->GetID()) == pcID )
	{
	  pCSLIDESweep = (CSLIDESweep*)iter.PeekNextRemoveCurr();
	}
      else
	{
	  pCSLIDESweep = (CSLIDESweep*)iter.PeekNext();
	}
    }
}

VOID CSLIDESweepPath::InsertSweepMorphParent(CSLIDESweepMorph *pCSLIDESweepMorph)
{
  ASSERT( pCSLIDESweepMorph != NULL );

  m_listCSLIDESweepMorphParent.InsertLast(pCSLIDESweepMorph, (UINT_32)pCSLIDESweepMorph);
}

VOID CSLIDESweepPath::RemoveSweepMorphParent(CSLIDESweepMorph *pCSLIDESweepMorphToRemove)
{
  ASSERT( pCSLIDESweepMorphToRemove != NULL );

  CListIterInt iter;
  CSLIDESweepMorph *pCSLIDESweepMorph;
  CHAR *pcID;

  pcID = pCSLIDESweepMorphToRemove->GetID();

  iter.Init(&m_listCSLIDESweepMorphParent);
  for ( pCSLIDESweepMorph = (CSLIDESweepMorph*)iter.PeekFirst();
	pCSLIDESweepMorph != NULL; )
    {
      if( (pCSLIDESweepMorph->GetID()) == pcID )
	{
	  pCSLIDESweepMorph = (CSLIDESweepMorph*)iter.PeekNextRemoveCurr();
	}
      else
	{
	  pCSLIDESweepMorph = (CSLIDESweepMorph*)iter.PeekNext();
	}
    }
}

BOOL CSLIDESweepPath::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  if ( !CSLIDEAtomNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  return TRUE;
}

BOOL CSLIDESweepPath::Valid()
{
  if ( !CSLIDEAtomNode::Valid() )
    {
      return FALSE;
    }

  return TRUE;
}

VOID CSLIDESweepPath::CheckDynamic()
{
  CSLIDEAtomNode::CheckDynamic();
}

VOID CSLIDESweepPath::PreprocessGraphLocal()
{
  CSLIDEAtomNode::PreprocessGraphLocal();
}

VOID CSLIDESweepPath::UpdateState()
{
  CSLIDEAtomNode::UpdateState();
}

VOID CSLIDESweepPath::CalculateBound()
{
}

VOID CSLIDESweepPath::RenderGeometrySmooth(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{
}

VOID CSLIDESweepPath::RenderGeometryFlat(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{
}

VOID CSLIDESweepPath::RenderNormalsSmooth()
{
}

VOID CSLIDESweepPath::RenderNormalsFlat()
{
}

VOID CSLIDESweepPath::WriteSLIDE(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "  %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtPath], GetID());

  fprintf(fpOutputFile, "    %s %ld\n",
	      gcppcSLIDEReservedWords[sliderwtMinimizeTorsion],
	      m_dyniMinimizeTorsion.GetValue());
  fprintf(fpOutputFile, "    %s %ld\n",
	      gcppcSLIDEReservedWords[sliderwtClosed],
	      m_dyniClosed.GetValue());
  fprintf(fpOutputFile, "    %s %ld\n",
	      gcppcSLIDEReservedWords[sliderwtSymmetry],
	      m_dyniSymmetry.GetValue());
  fprintf(fpOutputFile, "    %s %.3f\n",
	      gcppcSLIDEReservedWords[sliderwtAzimuth],
	      m_dynfAzimuth.GetValue());
  fprintf(fpOutputFile, "    %s %.3f\n",
	      gcppcSLIDEReservedWords[sliderwtTwist],
	      m_dynfTwist.GetValue());


  fprintf(fpOutputFile, "  %s\n",
	      gcppcSLIDEReservedWords[sliderwtEndPath]);
}

VOID CSLIDESweepPath::WriteRIBGeometry(FILE * /*fpOutputFile */ ,
				      const Matrix & /*mGproj_obj*/, 
				      CBound * /*pCBoundClip*/, 
				      SLIDEShadingType /* eShading */,
				      SLIDELODType /* eLOD */)
{
}

VOID CSLIDESweepPath::WriteRIBPushSLIDEName(FILE */*fpOutputFile*/)
{
}

VOID CSLIDESweepPath::WriteRIBPopSLIDEName(FILE */*fpOutputFile*/)
{
}

VOID CSLIDESweepPath::WriteWRLZGeometry(FILE * /*fpOutputFile */, 
					const Matrix &/*mTransform*/,
					CSLIDESurface */*pCSLIDESurface*/)
{
}

VOID CSLIDESweepPath::WriteSTLGeometry(FILE * /*fpOutputFile */, const Matrix &/*mTransform*/)
{
}
