#ifndef __SLIDESWEEPPATH_H__
#define __SLIDESWEEPPATH_H__

#include "slideatomnode.h"

class CSLIDECrossSection;

//////////////////////////////////////////////////////////////////////////////
// CSLIDESweepPath Class
//

class CSLIDESweepPath : public CSLIDEAtomNode
{
//////////////////////////////////////////////////
// Friends
//
public: 
  friend class CSLIDESweep;
  friend class CSLIDESweepMorph;

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDESweepPath();
  virtual ~CSLIDESweepPath();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID Uninit();
  virtual BOOL Initialized();
  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);
  virtual BOOL Valid();

  virtual inline VOID SetMinimizeTorsion(DynInt *pdyniMinimizeTorsion);
  virtual inline VOID SetClosed(DynInt *pdyniClosed);
  virtual inline VOID SetAzimuth(DynFloat *pdynfAzimuth);
  virtual inline VOID SetTwist(DynFloat *pdynfTwist);
  virtual inline VOID SetSymmetry(DynInt *pdyniSymmetry);


  VOID InsertCrossSectionParent(CSLIDECrossSection *pCSLIDECrossSection);
  VOID RemoveCrossSectionParent(CSLIDECrossSection *pCSLIDECrossSectionToRemove);

  VOID InsertSweepParent(CSLIDESweep *pCSLIDESweep);
  VOID RemoveSweepParent(CSLIDESweep *pCSLIDESweepToRemove);

  VOID InsertSweepMorphParent(CSLIDESweepMorph *pCSLIDESweepMorph);
  VOID RemoveSweepMorphParent(CSLIDESweepMorph *pCSLIDESweepMorphToRemove);

//////////////////////////////////////////////////
// Geometric Computations
//
protected:
  virtual inline BOOL GetSweepMorphCrossSectionDynamic();
  virtual VOID CheckSweepMorphCrossSectionDynamic() = 0;

  virtual VOID CheckDynamic();
  virtual VOID PreprocessGraphLocal();
  virtual VOID UpdateState();
  virtual VOID CalculateBound();

//////////////////////////////////////////////////
// Rendering
//
protected:
  virtual VOID RenderGeometrySmooth(CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderGeometryFlat(CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderNormalsSmooth();
  virtual VOID RenderNormalsFlat();

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

  virtual VOID WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				 CSLIDESurface *pCSLIDESurface);
  virtual VOID WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform);

//////////////////////////////////////////////////
// Accessor Functions
//
protected:
  virtual inline UINT_32 GetSlices();
  virtual inline UINT_32 GetControlPoints();
  virtual inline Point *GetPoint(); 
  virtual inline Matrix *GetFrenetMatrix(); 

  virtual inline CListInt *GetControlPointList() = 0;
  virtual inline CListInt *GetControlPointIDList() = 0;

  virtual inline FLOAT GetTu();
  virtual inline FLOAT GetTuDelta();

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:

  DynInt m_dyniSlices;
  UINT_32 m_uiSlices;
  Point *m_pPoint;

  // Variables
  DynInt m_dyniMinimizeTorsion;
  DynInt m_dyniClosed;
  DynInt m_dyniSymmetry;
  DynFloat m_dynfAzimuth;
  DynFloat m_dynfTwist;

  // CrossSections that reference me
  CListInt m_listCSLIDECrossSectionParent;

  // Sweeps that reference me as a path
  CListInt m_listCSLIDESweepParent;
  CListInt m_listCSLIDESweepMorphParent;

  // Sweep Morph
  BOOL m_bSweepMorphCrossSectionDynamic;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline BOOL CSLIDESweepPath::GetSweepMorphCrossSectionDynamic()
{
  return m_bSweepMorphCrossSectionDynamic;
}

inline UINT_32 CSLIDESweepPath::GetSlices() 
{
  return m_uiSlices;
}

inline Point *CSLIDESweepPath::GetPoint() 
{
  return m_pPoint;
}

inline Matrix *CSLIDESweepPath::GetFrenetMatrix()
{
  ASSERT(FALSE);

  return (Matrix *)NULL;
}

inline UINT_32 CSLIDESweepPath::GetControlPoints() 
{
  ASSERT(FALSE);

  return 0;
}

inline FLOAT CSLIDESweepPath::GetTu()
{
  ASSERT(FALSE);

  return 0;
}

inline FLOAT CSLIDESweepPath::GetTuDelta()
{
  ASSERT(FALSE);

  return 0;
}

inline VOID CSLIDESweepPath::SetMinimizeTorsion(DynInt *pdyniMinimizeTorsion)
{
  m_dyniMinimizeTorsion.Set(pdyniMinimizeTorsion);
}

inline VOID CSLIDESweepPath::SetClosed(DynInt *pdyniClosed)
{
  m_dyniClosed.Set(pdyniClosed);
}

inline VOID CSLIDESweepPath::SetAzimuth(DynFloat *pdynfAzimuth)
{
  m_dynfAzimuth.Set(pdynfAzimuth);
}

inline VOID CSLIDESweepPath::SetTwist(DynFloat *pdynfTwist)
{
  m_dynfTwist.Set(pdynfTwist);
}

inline VOID CSLIDESweepPath::SetSymmetry(DynInt *pdyniSymmetry)
{
  m_dyniSymmetry.Set(pdyniSymmetry);
}

#endif // __SLIDESWEEPPATH_H__
