#ifndef __SLIDETERRAIN_H__
#define __SLIDETERRAIN_H__

#include "slideatomnode.h"

class CSLIDEPoint;

//////////////////////////////////////////////////////////////////////////////
// CSLIDETerrain Class
//

class CSLIDETerrain : public CSLIDEAtomNode
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDETerrain();
  virtual ~CSLIDETerrain();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID, 
		    CListInt *plistPointIDs);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  //inline VOID SetSolid(DynInt *pdyniSolid);
  inline VOID SetType(DynInt *pdyniType);
  inline VOID SetVoronoi(DynInt *pdyniType);
  inline VOID SetAccuracy(DynInt *pdyniAccuracy);
  inline VOID SetRandomize(DynInt *pdyniRandomize);

  inline VOID SetDrawDelaunay(DynInt *pdyniDrawDelaunay);
  inline VOID SetDrawVertices(DynInt *pdyniDrawVertices);

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual CSLIDECamera *FindCamera(CListIterInt *piterIDs, 
				   CListInt *plistCSLIDEInstance);
  virtual CSLIDELight *FindLight(CListIterInt *piterIDs, 
				 CListInt *plistCSLIDEInstance);
  virtual CSLIDEAreaLight *FindAreaLight(CListIterInt *piterIDs, 
					 CListInt *plistCSLIDEInstance);

  virtual VOID Preprocess();
  virtual VOID PreprocessGraph();

  virtual VOID Update();
  virtual VOID UpdateGraph();

#ifdef __USE_CDA__
  virtual VOID CalculateCDAVertices(CListInt *plistCDAVertex, 
				    const Matrix &mGroot_obj);
#endif // __USE_CDA__

protected:
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();
  virtual VOID CalculateBound();

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID PrerenderGeometry(SLIDEShadingType eShading);
  virtual VOID PrerenderNormals(SLIDEShadingType eShading);

  virtual VOID RenderGeometry(const Matrix &mGproj_obj,
			      CBound *pCBoundClip, 
			      SLIDEShadingType eShading,
			      SLIDELODType eLOD,
			      CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderNormals(SLIDEShadingType eShading);

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);

  BOOL CheckRIBPrimitive();
  VOID WriteRIBGeometryPolygonal(FILE *fpOutputFile,
				 SLIDEShadingType eShading,
				 SLIDELODType eLOD);
  VOID WriteRIBGeometryPrimitive(FILE *fpOutputFile,
				 SLIDEShadingType eShading,
				 SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

  virtual VOID WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				 CSLIDESurface *pCSLIDESurface);
  virtual VOID WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform);

  virtual VOID WriteDotID(FILE *fpOutputFile);
  virtual VOID WriteDotNode(FILE *fpOutputFile);
  virtual VOID WriteDotEdges(FILE *fpOutputFile);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline BOOL GetChildrenBoundsDynamic();
  inline VOID SetChildrenBoundsDynamic(BOOL bChildrenBoundsDynamic);

  UINT_32 GetPointCount();

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CListInt *m_plistPointIDs;
  CListInt m_listCSLIDEPoint;

  DynInt m_dyniType;
  DynInt m_dyniVoronoi;
  DynInt m_dyniAccuracy;
  DynInt m_dyniRandomize;

  DynInt m_dyniDrawDelaunay;
  DynInt m_dyniDrawVertices;

  QEMesh *m_pmeVoronoi;

  BOOL m_bChildrenBoundsDynamic;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDETerrain::SetType(DynInt *pdyniType)
{
  m_dyniType.Set(pdyniType);
}

inline VOID CSLIDETerrain::SetVoronoi(DynInt *pdyniVoronoi)
{
  m_dyniVoronoi.Set(pdyniVoronoi);
}

inline VOID CSLIDETerrain::SetAccuracy(DynInt *pdyniAccuracy)
{
  m_dyniAccuracy.Set(pdyniAccuracy);
}

inline VOID CSLIDETerrain::SetRandomize(DynInt *pdyniRandomize)
{
  m_dyniRandomize.Set(pdyniRandomize);
}

inline VOID CSLIDETerrain::SetDrawDelaunay(DynInt *pdyniDrawDelaunay)
{
  m_dyniDrawDelaunay.Set(pdyniDrawDelaunay);
}

inline VOID CSLIDETerrain::SetDrawVertices(DynInt *pdyniDrawVertices)
{
  m_dyniDrawVertices.Set(pdyniDrawVertices);
}

inline BOOL CSLIDETerrain::GetChildrenBoundsDynamic()
{
  return m_bChildrenBoundsDynamic;
}

inline VOID CSLIDETerrain::SetChildrenBoundsDynamic(BOOL bChildrenBoundsDynamic)
{
  m_bChildrenBoundsDynamic = bChildrenBoundsDynamic;
}

#endif // __SLIDETERRAIN_H__
