#ifndef __SLIDETORUS_H__
#define __SLIDETORUS_H__

#include "slideatomnode.h"

//////////////////////////////////////////////////////////////////////////////
// CSLIDETorus Class
//

class CSLIDETorus : public CSLIDEAtomNode
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDETorus();
  virtual ~CSLIDETorus();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  inline VOID SetMajorRadius(DynFloat *pdynfMajorRadius);
  inline VOID SetMinorRadius(DynFloat *pdynfMinorRadius);
  inline VOID SetPhiMin(DynFloat *pdynfPhiMin);
  inline VOID SetPhiMax(DynFloat *pdynfPhiMax);
  inline VOID SetThetaMax(DynFloat *pdynfThetaMax);
  inline VOID SetPhiSlices(DynInt *pdyniPhiSlices);
  inline VOID SetThetaSlices(DynInt *pdyniThetaSlices);

  inline VOID SetTexture(DynVector *pdynv0, DynVector *pdynv1, DynVector *pdynv2, DynVector *pdynv3);

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Geometric Computations
//
public:

#ifdef __USE_CDA__
  virtual VOID CalculateCDAVertices(CListInt *plistCDAVertex, 
				    const Matrix &mGroot_obj);
#endif // __USE_CDA__



protected:
  virtual VOID CheckDynamic();
  virtual VOID PreprocessGraphLocal();
  virtual VOID UpdateState();
  virtual VOID CalculateBound();
  virtual VOID CalculateQEMesh();
  virtual VOID CalculateLeafQEMesh();  // for WriteWRLZ
  virtual VOID DeallocateLeafQEMesh();  // for WriteWRLZ

//////////////////////////////////////////////////
// Rendering
//
public:

protected:
  virtual VOID RenderGeometrySmooth(CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderGeometryFlat(CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderNormalsSmooth();
  virtual VOID RenderNormalsFlat();

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

  virtual VOID WritePOVGeometry(FILE *fpOutputFile, 
				CListInt *plistTransformIDs,
				CListInt *plistColors); 

  virtual VOID WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				 CSLIDESurface *pCSLIDESurface);
  virtual VOID WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline CHAR *GetTypeString();
  virtual inline SLIDEReservedWordType GetType();

  virtual VOID GetSide(CListInt *plistPoint, CListInt *plistNormal, UINT_32 uiSide, 
		       const Matrix &mTransform, const Matrix &mInverse, BOOL bReverse);

  virtual VOID GetParameters(Tcl_Interp *pInterp);
  virtual UINT_32 SetParameter(CSLIDEWorld *pCSLIDEWorld,
			       SLIDEReservedWordType uiParameter, CHAR **argv);

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  DynFloat m_dynfMajorRadius;
  DynFloat m_dynfMinorRadius;
  DynFloat m_dynfPhiMin;
  DynFloat m_dynfPhiMax;
  DynFloat m_dynfThetaMax;

  DynInt m_dyniPhiSlices;
  DynInt m_dyniThetaSlices;  

  DynVector m_pvTexture[4];
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDETorus::SetMajorRadius(DynFloat *pdynfMajorRadius)
{
  m_dynfMajorRadius.Set(pdynfMajorRadius);
}

inline VOID CSLIDETorus::SetMinorRadius(DynFloat *pdynfMinorRadius)
{
  m_dynfMinorRadius.Set(pdynfMinorRadius);
}

inline VOID CSLIDETorus::SetPhiMin(DynFloat *pdynfPhiMin)
{
  m_dynfPhiMin.Set(pdynfPhiMin);
}

inline VOID CSLIDETorus::SetPhiMax(DynFloat *pdynfPhiMax)
{
  m_dynfPhiMax.Set(pdynfPhiMax);
}

inline VOID CSLIDETorus::SetThetaMax(DynFloat *pdynfThetaMax)
{
  m_dynfThetaMax.Set(pdynfThetaMax);
}

inline VOID CSLIDETorus::SetPhiSlices(DynInt *pdyniPhiSlices)
{
  m_dyniPhiSlices.Set(pdyniPhiSlices);
}

inline VOID CSLIDETorus::SetThetaSlices(DynInt *pdyniThetaSlices)
{
  m_dyniThetaSlices.Set(pdyniThetaSlices);
}

inline VOID CSLIDETorus::SetTexture(DynVector *pdynv0, DynVector *pdynv1, 
				       DynVector *pdynv2, DynVector *pdynv3)
{
  m_pvTexture[0].SetAll(pdynv0);
  m_pvTexture[1].SetAll(pdynv1);
  m_pvTexture[2].SetAll(pdynv2);
  m_pvTexture[3].SetAll(pdynv3);
}

inline CHAR *CSLIDETorus::GetTypeString()
{
  return gcppcSLIDEReservedWords[sliderwtTorus];
}

inline SLIDEReservedWordType CSLIDETorus::GetType()
{
  return sliderwtTorus;
}

#endif // __SLIDETORUS_H__
