#ifndef __SLIDETRANSLATE_H__
#define __SLIDETRANSLATE_H__

#include "slidetransform.h"

//////////////////////////////////////////////////////////////////////
// CSLIDETranslate Class
//

class CSLIDETranslate : public CSLIDETransform
{
//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDETranslate();
  virtual ~CSLIDETranslate();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    DynVector *pdynv);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual inline VOID ComposeWith(CSLIDETransform *pCSLIDETransform);
  virtual inline VOID Set(CSLIDETransform *pCSLIDETransform);
  virtual inline VOID Modify(CSLIDETransform *pCSLIDETransformLast, 
			     CSLIDETransform *pCSLIDETransformNew);

protected:
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();

//////////////////////////////////////////////////
// Rendering
//
public:

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);
  virtual VOID WriteRIB(FILE *fpOutputFile);
  virtual VOID WritePOV(FILE *fpOutputFile);
  virtual VOID WriteDotNode(FILE *fpOutputFile);
  
//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline CHAR *GetTypeString();
  virtual inline SLIDEReservedWordType GetType();

  virtual inline DynVector GetTranslate();

//////////////////////////////////////////////////
// Class Variables
//
public:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  DynVector m_vTranslate;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline CHAR *CSLIDETranslate::GetTypeString()
{
  return gcppcSLIDEReservedWords[sliderwtTranslate];
}

inline SLIDEReservedWordType CSLIDETranslate::GetType()
{
  return sliderwtTranslate;
}

inline DynVector CSLIDETranslate::GetTranslate()
{
  return m_vTranslate;
}

inline VOID CSLIDETranslate::ComposeWith(CSLIDETransform *pCSLIDETransform)
{
  m_vTranslate += ((CSLIDETranslate*)pCSLIDETransform)->GetTranslate();
  m_m.SetTranslate(m_vTranslate);
  m_mInverse.SetTranslate(-m_vTranslate);
}

inline VOID CSLIDETranslate::Set(CSLIDETransform *pCSLIDETransform)
{
  m_vTranslate[X] = ((CSLIDETranslate*)pCSLIDETransform)->GetTranslate()[X];
  m_vTranslate[Y] = ((CSLIDETranslate*)pCSLIDETransform)->GetTranslate()[Y];
  m_vTranslate[Z] = ((CSLIDETranslate*)pCSLIDETransform)->GetTranslate()[Z];

  m_m.SetTranslate(m_vTranslate);
  m_mInverse.SetTranslate(-m_vTranslate);
}

inline VOID CSLIDETranslate::Modify(CSLIDETransform *pCSLIDETransformLast,
				    CSLIDETransform *pCSLIDETransformNew)
{
  m_vTranslate[X] += ( (((CSLIDETranslate*)pCSLIDETransformNew)->GetTranslate()[X]) -
		      (((CSLIDETranslate*)pCSLIDETransformLast)->GetTranslate()[X]) );
  m_vTranslate[Y] += ( (((CSLIDETranslate*)pCSLIDETransformNew)->GetTranslate()[Y]) -
		      (((CSLIDETranslate*)pCSLIDETransformLast)->GetTranslate()[Y]) );
  m_vTranslate[Z] += ( (((CSLIDETranslate*)pCSLIDETransformNew)->GetTranslate()[Z]) -
		      (((CSLIDETranslate*)pCSLIDETransformLast)->GetTranslate()[Z]) );

  m_m.SetTranslate(m_vTranslate);
  m_mInverse.SetTranslate(-m_vTranslate);
}

#endif // __SLIDETRANSLATE_H__


