#include <math.h>
#include "slidewarp.h"
#include "slidepoint.h"
#include "slideworld.h"
#include "slidesurface.h"
#include "slidegroup.h"
#include "slidebsplinecurve.h"

#include "qemesh.h"
#include "qevertex.h"

static const FLOAT cfInstances = 2.0f;
static const FLOAT cfPathTu = 0.0f;
static const FLOAT cfPathTuDelta = 0.0f;
static const UINT_32 cuiCurrInst = 1;
static const UINT_32 cuiType = slidewtSLF_AXIS;
static const UINT_32 cuiPathOnOff = 1;
static const UINT_32 cuiLControl = 1;
static const UINT_32 cuiMControl = 1;
static const UINT_32 cuiNControl = 1;


//////////////////////////////////////////////////////////////////////
// CSLIDEWormwarp Class
//

CSLIDEWormwarp::CSLIDEWormwarp()
{
  m_dynfInstances.SetValue(cfInstances);
  m_dyniType.SetValue(cuiType);
  m_pWarpPrev = NULL;
  m_mCurrMatrix.SetIdentity();

  // Render init
  m_fPathTu = cfPathTu;
  m_fPathTuDelta = cfPathTuDelta;
  m_uiCurrInst = cuiCurrInst;

  m_uiLControl = cuiLControl;
  m_uiMControl = cuiMControl;
  m_uiNControl = cuiNControl;
  m_pppPointLattice = NULL;
  m_pfFactorials = NULL;

  // Group init
  m_pcGroupID = NULL;
  m_pGroup = NULL;

  // Path init
  m_pcPathID = NULL;
  m_pPath = NULL;
  m_dyniPathOnOff.SetValue(cuiPathOnOff);
}

CSLIDEWormwarp::~CSLIDEWormwarp()
{
}

VOID CSLIDEWormwarp::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID)
{
  CSLIDEBoundedNode::Init(pcFile, uiLine, uiColumn, pcID);

  m_listMatrixStack.Init();
}

VOID CSLIDEWormwarp::Uninit()
{
  CSLIDEBoundedNode::Uninit();
}

BOOL CSLIDEWormwarp::Initialized()
{
  if ( !CSLIDEBoundedNode::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

BOOL CSLIDEWormwarp::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  if ( !CSLIDEBoundedNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  BOOL bCorrect;

  bCorrect = TRUE;

  // Check Group Reference
  if( NULL != m_pcGroupID )
    {
      m_pGroup = (CSLIDEGroup *) pCSLIDEWorld->PeekNode(m_pcGroupID);
      if ( NULL == m_pGroup )
	{
	  HandleError( "Undefined group \"%s\"", m_pcGroupID );
	}
    }
  else
    {
      HandleError("wormwarp %s has no specified group", m_pcGUIID);
      bCorrect = FALSE;
    }


  // Check Path Reference
  if( NULL != m_pcPathID )
    {
      m_pPath = (CSLIDEBSplineCurve *) pCSLIDEWorld->PeekNode(m_pcPathID);
      if ( NULL == m_pPath )
	{
	  HandleError( "Undefined path \"%s\"", m_pcPathID );
	}
    }
  else
    {
      HandleError("wormwarp %s has no specified path", m_pcGUIID);
      bCorrect = FALSE;
    }

  return bCorrect;
}

BOOL CSLIDEWormwarp::Valid()
{
  if ( !CSLIDEBoundedNode::Valid() )
    {
      return FALSE;
    }

  return ( ( m_dynfInstances.GetValue() >= 0 ) );
}

VOID CSLIDEWormwarp::CheckDynamic()
{
  CSLIDEBoundedNode::CheckDynamic();

  if( !GetDynamic() )
    {
      // Is this dynamic?
      SetDynamic(m_dynfInstances.GetDynamic()
		 || m_dyniType.GetDynamic()
		 || m_dyniPathOnOff.GetDynamic());
    }

  if( !GetDynamic() )
    {
      // Is group or path dynamic?
      SetDynamic(m_pGroup->GetDynamic()
		 || m_pPath->GetDynamic() );
    }

}

CSLIDECamera *CSLIDEWormwarp::FindCamera(CListIterInt *piterIDs, 
				   CListInt *plistCSLIDEInstance)
{return NULL;}

CSLIDELight *CSLIDEWormwarp::FindLight(CListIterInt *piterIDs, 
				 CListInt *plistCSLIDEInstance)
{return NULL;}

CSLIDEAreaLight *CSLIDEWormwarp::FindAreaLight(CListIterInt *piterIDs, 
					 CListInt *plistCSLIDEInstance)
{return NULL;}

VOID CSLIDEWormwarp::Preprocess()
{
  CSLIDEBoundedNode::Preprocess();

  CheckDynamic();
  UpdateState();
}

VOID CSLIDEWormwarp::PreprocessGraph()
{
  if ( !GetVisited() )
    {
      m_pGroup->PreprocessGraph();
      m_pPath->PreprocessGraph();
      SetGeometryEmpty(FALSE);
      SetGeometryDynamic( m_pCSLIDESurface->GetDynamic()
			  || this->GetDynamic()
			  || m_pGroup->GetDynamic()
			  || m_pPath->GetDynamic() );
      SetNormalsEmpty(FALSE);
      SetNormalsDynamic( this->GetDynamic()
			 || m_pGroup->GetDynamic()
			 || m_pPath->GetDynamic());

      CalculateBound();
      SetBoundsEmpty(FALSE);
      SetBoundsDynamic(FALSE);
      SetBoundEmpty(FALSE);
      SetBoundDynamic( this->GetDynamic()
		       || m_pGroup->GetDynamic()
		       || m_pPath->GetDynamic() );
      SetVisited(TRUE);
    }
}

VOID CSLIDEWormwarp::Update()
{
  SetBoundChanged(FALSE);
  if ( GetDynamic() )
    {
      UpdateState();
    }

  if ( GetBoundsDynamic() )
    {
      //      CSLIDEBoundedNode::InvalidateQEMesh();

      //      if ( m_pmeSIF != NULL )
      //	{
      //	  DeallocateQEMesh(m_pmeSIF);
      //	  m_pmeSIF = NULL;
      //	}
    }

  if ( GetBoundDynamic() )
    {
      SetBoundValid(FALSE);
    }

  SetVisited(FALSE);
}

VOID CSLIDEWormwarp::UpdateGraph()
{
  if ( !GetVisited() )
    {
      BOOL bBoundChanged;

      if ( !GetBoundValid() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      if ( !m_bound.GetEmpty() )
		{
		  SetBoundChanged(TRUE);
		  m_bound.Null();
		}
	      return;
	    }
	  
	  bBoundChanged = FALSE;
	  
	  m_pPath->UpdateGraph();
	  m_pGroup->UpdateGraph();
	  bBoundChanged = bBoundChanged || 
	    m_pPath->GetBoundChanged() ||
	    m_pGroup->GetBoundChanged() ||
	    (m_bound.GetEmpty() && 
	     (!m_pPath->GetBoundEmpty() || !m_pGroup->GetBoundEmpty() ));

	  if ( bBoundChanged )
	    {
	      CalculateBound();
	    }

	  SetBoundValid(TRUE);
	}

      SetVisited(TRUE);
    }
}

VOID CSLIDEWormwarp::UpdateState()
{
  BOOL bCalcLattice;
  CSLIDEBoundedNode::UpdateState();

  if (m_dynfInstances.GetDynamic())
    {
      m_dynfInstances.Update();
    }
  if (m_dyniType.GetDynamic())
    {
      m_dyniType.Update();
    }
  if (m_dyniPathOnOff.GetDynamic())
    {
      m_dyniPathOnOff.Update();
    }

  bCalcLattice = FALSE;
  if ( (NULL == m_pppPointLattice)
       || ( m_pPath->GetSlices() != m_uiNControl ))
    {
      UINT_32 l, m;

      bCalcLattice = TRUE;
      if (NULL != m_pppPointLattice)
	{
	  for (l = 0; l <= m_uiLControl; l++)
	    {
	      for (m = 0; m <= m_uiMControl; m++ )
		{
		  delete m_pppPointLattice[l][m];
		}
	      delete m_pppPointLattice[l];
	    }
	  delete m_pppPointLattice;
	  m_pppPointLattice = NULL;
	}

      if (NULL != m_pfFactorials)
	{
	  delete m_pfFactorials;
	}
      m_pfFactorials = NULL;
      m_uiNControl = m_pPath->GetSlices();
      
      m_pppPointLattice = new Point **[m_uiLControl+1];
      for (l = 0; l <= m_uiLControl; l++)
	{
	  m_pppPointLattice[l] = new Point *[m_uiMControl+1];
	  for (m = 0; m <= m_uiMControl; m++ )
	    {
	      m_pppPointLattice[l][m] = new Point[m_uiNControl+1];
	    }
	}

      m_pfFactorials = new FLOAT[m_uiNControl+1];
      m_pfFactorials[0] = 1;
      for (l = 1; l <= m_uiNControl; l++)
	{
	  m_pfFactorials[l] = l * m_pfFactorials[l-1];
	}
    } // End if

  if ( bCalcLattice || m_pPath->GetDynamic() )
    {
      Point *pPathPoint;
      Matrix *pmPathFrenet;
      Point ptControl;
      UINT_32 l, m, n;

      pPathPoint = m_pPath->GetPoint();
      pmPathFrenet = m_pPath->GetFrenetMatrix();

      for (l = 0; l <= m_uiLControl; l++)
	{
	  ptControl.Set((2.0f*(FLOAT)l/(FLOAT)m_uiLControl)-1.0f, 0, 0);
	  for (m = 0; m <= m_uiMControl; m++)
	    {
	      ptControl[Y] = (2.0f*(FLOAT)m/(FLOAT)m_uiMControl)-1.0f;
	      for (n = 0; n <= m_uiNControl; n++)
		{
		  ptControl[Z] = (2.0f*(FLOAT)n/(FLOAT)m_uiNControl)-1.0f;
		  m_pppPointLattice[l][m][n] = 
		    pPathPoint[n] + pmPathFrenet[n] * ptControl;
		}
	    }
	}
      
    }

}

VOID CSLIDEWormwarp::CalculateBound()
{
  CBound bound;
  UINT_32 uiSlice, uiSlices;
  Point *pPathPoint;
  Matrix *pmPathFrenet;
  Matrix m1, m2;
  Point p1;
  Vector v1;

  uiSlices = m_pPath->GetSlices();
  pPathPoint = m_pPath->GetPoint();
  pmPathFrenet = m_pPath->GetFrenetMatrix();
  
  for( uiSlice = 0; uiSlice < uiSlices+1; uiSlice++)
    {
      p1 = pPathPoint[uiSlice];

      m1.SetScale( 0.5f / m_dynfInstances.GetValue() );
      m1 = pmPathFrenet[uiSlice] * m1;        // Get Change of Basis Matrix and Warp
      v1.Set(p1(X),p1(Y),p1(Z));
      m2.SetTranslate(v1);
      m1 = m2 * m1;

      m_pGroup->InsertIntoBound(m1,bound);
    }  

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);

      // Calculate normal length
      //      m_fNormalLength = bound.GetRadius() * cfNormalRatio;
    }

}

////////////////////////////////////////////////////////////////////////////////////////////

VOID CSLIDEWormwarp::CalculateQEMesh()
{  
}

////////////////////////////////////////////////////////////////////////////////////////////

VOID CSLIDEWormwarp::PrerenderGeometry(SLIDEShadingType eShading)
{
  Matrix m;

  m.SetIdentity();

  RenderGeometry( m, &m_bound, eShading, (SLIDELODType)m_dyniLOD.GetValue(), NULL );

}

VOID CSLIDEWormwarp::PrerenderNormals(SLIDEShadingType eShading)
{
m_pGroup->PrerenderNormals(eShading);
m_pPath->PrerenderNormals(eShading);
}

VOID CSLIDEWormwarp::RenderGeometry(const Matrix &mGproj_obj,
				    CBound *pCBoundClip, 
				    SLIDEShadingType eShading,
				    SLIDELODType eLOD,
				    CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{

INT_32 iter;
FLOAT fTuCurr;
Point pAtTu;
Matrix mAtTu, mTemp;
Vector vAtTu;
Vector vNAtTu, vBAtTu, vTAtTu;

// Save the previous warp
m_pWarpPrev = pCSLIDEWarp;
m_bound.Null();

if (sliderwtBSplineCurve == m_pPath->GetType())
  {
    m_fPathTu = m_pPath->GetTu();
    m_fPathTuDelta = m_pPath->GetTuDelta();
  }

if ( 1 == m_dyniPathOnOff.GetValue() )
  {
    m_pPath->RenderGeometry(mGproj_obj,
			    pCBoundClip,
			    eShading,
			    eLOD,
			    this);
  }

switch (m_dyniType.GetValue())
  {
  case slidewtSLF_BERNSTEIN:
    {
      m_pGroup->RenderGeometry(mGproj_obj,
			       pCBoundClip,
			       eShading,
			       eLOD,
			       this);
    }
    break;
  case slidewtSLF_AXIS:
    {
      for (iter = 0; iter < m_dynfInstances.GetValue(); iter++)
	{
	  m_uiCurrInst = iter;
	  m_pGroup->RenderGeometry(mGproj_obj,
				   pCBoundClip,
				   eShading,
				   eLOD,
				   this);
	}
    }
    break;
  case slidewtSLF_BEAD:
  default:
    {
      for (iter = 0; iter < m_dynfInstances.GetValue(); iter++)
	{
          m_uiCurrInst = iter;
	  mAtTu.SetIdentity();
	  fTuCurr = m_fPathTu + (FLOAT)(2*m_uiCurrInst+1)*
	    (0.5f * m_fPathTuDelta / (FLOAT)m_dynfInstances.GetValue());
	  
	  mTemp.SetScale( 0.5f/m_dynfInstances.GetValue() );
	  mAtTu = mTemp * mAtTu;
	  m_pPath->FrenetAt(vNAtTu,vBAtTu,vTAtTu,fTuCurr);
	  mTemp.SetRotate(vNAtTu,vBAtTu,vTAtTu);
	  m_pPath->PointAt(pAtTu, fTuCurr);
	  mAtTu = mTemp * mAtTu;
	  vAtTu.Set(pAtTu(X),pAtTu(Y),pAtTu(Z));
	  mTemp.SetTranslate(vAtTu);
	  mAtTu = mTemp * mAtTu;
	  
	  if (NULL == pCSLIDEWarp)
	    {
	      gilPushMatrix();
	      {
		gilMultMatrixf(mAtTu.Get());
		
		m_pGroup->RenderGeometry(mGproj_obj * mAtTu,
					 pCBoundClip,
					 eShading,
					 eLOD,
					 pCSLIDEWarp);
	      }
	      gilPopMatrix();
	    }
	  else
	    {
	      pCSLIDEWarp->PushMatrix();
	      pCSLIDEWarp->MultMatrix(mAtTu);
	      m_pGroup->RenderGeometry(mGproj_obj * mAtTu,
					 pCBoundClip,
					 eShading,
					 eLOD,
					 pCSLIDEWarp);
	      pCSLIDEWarp->PopMatrix();
	    }
	}
    } // End default
    break;
  } // End switch

// Forget the previous warp
m_pWarpPrev = NULL;
}

VOID CSLIDEWormwarp::RenderGeometrySelect(const Matrix &mGproj_obj,
					  CBound *pCBoundClip, 
					  SLIDEShadingType eShading,
					  SLIDELODType eLOD)
{}

VOID CSLIDEWormwarp::RenderNormals(SLIDEShadingType eShading)
{}

VOID CSLIDEWormwarp::PushMatrix()
{
  Matrix *pMatrix = new Matrix();
  *pMatrix = m_mCurrMatrix;
  ASSERT (NULL != pMatrix);
  m_listMatrixStack.InsertFirst(pMatrix,0);
}

VOID CSLIDEWormwarp::PopMatrix()
{
  Matrix *pMatrix;
  pMatrix = (Matrix *) m_listMatrixStack.RemoveFirst();
  ASSERT(NULL != pMatrix);
  m_mCurrMatrix = *pMatrix;
  delete(pMatrix);
}

VOID CSLIDEWormwarp::MultMatrix(const Matrix &m)
{
  m_mCurrMatrix = m_mCurrMatrix * m;
}

/* BOOL CalcMatrix(const Point &ptOld, Point &ptNew)             */
/* Calculate the correct transform for a point                   */
/* in the object space at Z = fPtZ using the known m_mCurrMatrix */
/* going from object->warp space, then warping that point        */
/* appropriately. also possibly querying previous warps          */
BOOL CSLIDEWormwarp::CalcWarpPoint(const Point &ptOld, Point &ptNew)
{
  Matrix mWarp, m1;
  Point p1, ptWarped;
  Vector v1;

  // Calculate what transform is needed
  ptWarped = m_mCurrMatrix * ptOld;
  mWarp.SetIdentity();
  
  switch (m_dyniType.GetValue())
    {
    case slidewtSLF_BERNSTEIN:
      {
	return ApplyBernstein(ptOld, ptNew);
      }
      break;
    case slidewtSLF_AXIS:
      {
	Matrix mAtTu;
	Vector vAtTu, vNAtTu, vBAtTu, vTAtTu;
	Point pAtTu;
	FLOAT fTuCurr;

	// Take Z plane to origin
	v1.Set( 0, 0, -ptWarped(Z));
	m1.SetTranslate(v1);
	mWarp = m1 * mWarp;
	
	// Calculate location on curve
	mAtTu.SetIdentity();
	fTuCurr = m_fPathTu + 
	  ((FLOAT)(2*m_uiCurrInst +1) + (ptWarped(Z))) *
	    (0.5f * m_fPathTuDelta /(FLOAT)m_dynfInstances.GetValue() );
	  
	m1.SetScale( 0.5f/m_dynfInstances.GetValue() );
	mAtTu = m1 * mAtTu;
	m_pPath->FrenetAt(vNAtTu,vBAtTu,vTAtTu,fTuCurr);
	m1.SetRotate(vNAtTu,vBAtTu,vTAtTu);
	m_pPath->PointAt(pAtTu, fTuCurr);
        /*pAtTu = pAtTu * 0.5;*/
	mAtTu = m1 * mAtTu;
	vAtTu.Set(pAtTu(X),pAtTu(Y),pAtTu(Z));
	m1.SetTranslate(vAtTu);
	mAtTu = m1 * mAtTu;
	
	// Calculate point coordinate after warp
	mWarp = mAtTu * mWarp;

        ptWarped = mWarp * ptWarped;
      }
      break;
    case slidewtSLF_BEAD:
    default:
      {
	ptWarped = ptOld;
      } // End default
      break;
    } // End switch

  // Apply current warp/transform
  ptNew = ptWarped;

  m_bound.Insert(ptNew);
  
  // Apply any previous warps/transforms
  if (NULL != m_pWarpPrev)
    {
      m_pWarpPrev->CalcWarpPoint( ptWarped, ptNew);
    }
  
  return TRUE;
}

BOOL CSLIDEWormwarp::ApplyBernstein(const Point &ptOld, Point &ptNew)
{
  INT_32 l, m, n;
  INT_32 i, j, k;
  FLOAT s, t, u;
  FLOAT sFactor, tFactor, uFactor;
  Point ptWarped;
  
  // Take point to current space
  ptWarped = m_mCurrMatrix * ptOld;
  ptNew.Set(0, 0, 0);

  // Apply bernstein polynomial
  l = m_uiLControl;
  m = m_uiMControl;
  n = m_uiNControl;
  s = (ptWarped(X)+1.0f)/2.0f;
  t = (ptWarped(Y)+1.0f)/2.0f;
  u = (ptWarped(Z)+1.0f)/2.0f;
  for (k = 0; k <= n; k++)
    {
      uFactor = Choose(n, k) * pow(1.0f-u, n-k) * pow(u, k);
      for (j = 0; j <= m; j++)
	{
	  tFactor = Choose(m, j) * pow(1.0f-t, m-j) * pow(t, j);
	  for (i = 0; i <= l; i++)
	    {
	      sFactor = Choose(l, i) * pow(1.0f-s, l-i) * pow(s, i);
	      ptNew = ptNew + 
		sFactor * tFactor * uFactor *
		m_pppPointLattice[i][j][k];
	    }
	}
    }
  return TRUE;
}

VOID CSLIDEWormwarp::WriteSLIDE(FILE *fpOutputFile)
{
  //  UINT_32 i;
  CListIterInt iter;
  //  CHAR *pcID;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtSweep], m_pcGUIID);
  WriteSLIDEFlags(fpOutputFile, 0);

}

VOID CSLIDEWormwarp::WriteRIBGeometry(FILE *fpOutputFile,
				    const Matrix &/*mGproj_obj*/, 
				    CBound * /*pCBoundClip*/, 
				    SLIDEShadingType eShading,
				    SLIDELODType eLOD)
{
}

VOID CSLIDEWormwarp::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE Wormwarp: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);
}

VOID CSLIDEWormwarp::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE Wormwarp: %s", m_pcGUIID);
}

VOID CSLIDEWormwarp::WritePOVGeometry(FILE *fpOutputFile, 
				      CListInt *plistTransformIDs,
				      CListInt *plistColors)  
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDEWormwarp::WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				       CSLIDESurface *pCSLIDESurface)
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDEWormwarp::WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform)
{
  ASSERT( fpOutputFile != NULL );
}
