#ifndef __SLIDEWARP_H__
#define __SLIDEWARP_H__

#include "slideboundednode.h"
//#include "slidesweeppath.h"

class CSLIDEGroup;
class CSLIDEBSplineCurve;

//////////////////////////////////////////////////////////////////////////////
// CSLIDEWormwarp Class
//

class CSLIDEWormwarp : public CSLIDEBoundedNode
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEWormwarp();
  virtual ~CSLIDEWormwarp();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID Uninit();
  virtual BOOL Initialized();
  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);
  virtual BOOL Valid();

  virtual inline VOID SetGroupID(CHAR *pcID);
  virtual inline VOID SetPathID(CHAR *pcID);
  virtual inline VOID SetInstances(DynFloat *pdynfInstances);
  virtual inline VOID SetType(DynInt *pdyniType);
  virtual inline VOID SetPathOnOff(DynInt *pdyniPathOnOff);

protected:


//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual CSLIDECamera *FindCamera(CListIterInt *piterIDs, 
				   CListInt *plistCSLIDEInstance);
  virtual CSLIDELight *FindLight(CListIterInt *piterIDs, 
				 CListInt *plistCSLIDEInstance);
  virtual CSLIDEAreaLight *FindAreaLight(CListIterInt *piterIDs, 
					 CListInt *plistCSLIDEInstance);
  virtual VOID Preprocess();
  virtual VOID PreprocessGraph();

  virtual VOID Update();
  virtual VOID UpdateGraph();

  virtual VOID CalculateQEMesh();

protected:
  virtual VOID CheckDynamic();
  virtual VOID UpdateState();
  virtual VOID CalculateBound();

protected:

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID PrerenderGeometry(SLIDEShadingType eShading);
  virtual VOID PrerenderNormals(SLIDEShadingType eShading);
  virtual VOID RenderGeometry(const Matrix &mGproj_obj,
			      CBound *pCBoundClip, 
			      SLIDEShadingType eShading,
			      SLIDELODType eLOD,
			      CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderGeometrySelect(const Matrix &mGproj_obj,
				    CBound *pCBoundClip, 
				    SLIDEShadingType eShading,
				    SLIDELODType eLOD);

  virtual VOID RenderNormals(SLIDEShadingType eShading);

  virtual VOID PushMatrix();
  virtual VOID PopMatrix();

  virtual VOID MultMatrix(const Matrix &m);

  virtual BOOL CalcWarpPoint(const Point &ptOld, Point &ptNew);

protected:
  virtual BOOL ApplyBernstein(const Point &ptOld, Point &ptNew);
  virtual inline FLOAT Choose(UINT_32 n, UINT_32 k);

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

  virtual VOID WritePOVGeometry(FILE *fpOutputFile, 
				CListInt *plistTransformIDs,
				CListInt *plistColors); 

  virtual VOID WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				 CSLIDESurface *pCSLIDESurface);
  virtual VOID WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline CHAR *GetTypeString();
  virtual inline SLIDEReservedWordType GetType();

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:

// Warp
DynFloat m_dynfInstances;
DynInt m_dyniType;
CSLIDEWormwarp *m_pWarpPrev;
Matrix m_mCurrMatrix;
CListInt m_listMatrixStack;

// Rendering
FLOAT m_fPathTu, m_fPathTuDelta;
UINT_32 m_uiCurrInst;

UINT_32 m_uiLControl, m_uiMControl, m_uiNControl;
Point *** m_pppPointLattice;
FLOAT * m_pfFactorials;

// Group
CHAR *m_pcGroupID;
CSLIDEGroup *m_pGroup;

// Path
CHAR *m_pcPathID;
CSLIDEBSplineCurve *m_pPath;
DynInt m_dyniPathOnOff;

//  FLOAT m_fNormalLength;

  // Cross Section
//  UINT_32 m_uiCrossSectionType;  // [bead, polygon, circle, cresent, star]

//  CHAR *m_pcBeadID;

//  DynInt m_dyniCrossSectionSlices;
//  DynInt m_dyniCrossSectionSymmetry;
//  UINT_32 m_uiCrossSectionSlices;  // User parameter
//  UINT_32 m_uiCrossSectionPoints;  // Actual number of points allocated

//  DynFloat m_dynfRadius;           // = outerradius
//  DynFloat m_dynfInnerRadius;
//  DynFloat m_dynfThickness;
//  DynFloat m_dynfTheta;
//  DynInt m_dyniCrossSectionOrientation;  // 0 = outward, 1 = inward

//  Point *m_pCrossSectionPoint;
//  Vector*m_pvCrossSectionNormal;
//  CListInt *m_plistCrossSectionPointIDs;
//  CListInt m_listCSLIDECrossSectionPoint;
//  BOOL m_bCrossSectionPointsDynamic;
//  BOOL m_bCrossSectionChanged;

  // Path
//  CHAR *m_pcPathID;
//  CSLIDESweepPath *m_pPath;

//  DynInt m_dyniMinimizeTorsion;
//  DynInt m_dyniClosed;
//  DynFloat m_dynfAzimuth;
//  DynFloat m_dynfTwist;

//  DynInt m_dyniDrawPath;
//  DynInt m_dyniDrawSweep;

//  CListInt *m_plistWarp; // Warps to pass on to Path
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDEWormwarp::SetGroupID(CHAR *pcID)
{
  m_pcGroupID = pcID;
}

inline VOID CSLIDEWormwarp::SetPathID(CHAR *pcID)
{
  m_pcPathID = pcID;
}

inline VOID CSLIDEWormwarp::SetInstances(DynFloat *pdynfInstances)
{
  m_dynfInstances.Set(pdynfInstances);
}

inline VOID CSLIDEWormwarp::SetType(DynInt *pdyniType)
{
  m_dyniType.Set(pdyniType);
}

inline VOID CSLIDEWormwarp::SetPathOnOff(DynInt *pdyniPathOnOff)
{
  m_dyniPathOnOff.Set(pdyniPathOnOff);
}

inline CHAR *CSLIDEWormwarp::GetTypeString()
{
  return gcppcSLIDEReservedWords[sliderwtWormwarp];
}

inline SLIDEReservedWordType CSLIDEWormwarp::GetType()
{
  return sliderwtWormwarp;
}

inline FLOAT CSLIDEWormwarp::Choose(UINT_32 n, UINT_32 k)
{
  return m_pfFactorials[n] / (m_pfFactorials[n-k] * m_pfFactorials[k]);
}

#endif // __SLIDEWARP_H__
