#include "slideworld.h"
#include "sliderender.h"
#include "slideviewport.h"
#include "slidenode.h"
#include "slideinstance.h"
#include "slidetransform.h"
#include "slideobject.h"
#include "slideface.h"
#include "slidepoint.h"
#include "slidecontrolpoint.h"
#include "slidemeshvertexproperty.h"
#include "slidemeshedgeproperty.h"
#include "slidesurface.h"
#include "slidefog.h"
#include "slidelight.h"
#include "slidearealight.h"
#include "scanner.h"

BOOL gbPreprocessDone = FALSE;
BOOL gbCheckSemanticsDone = FALSE;

static CHAR gpcBuffer[200];
const FLOAT gcfFrameRate = 30.0;
//const FLOAT gcfFrameRate = 2.0;
const FLOAT gcfFrameTime = (FLOAT)1.0 / gcfFrameRate;

Tcl_Interp *CSLIDEWorld::m_pInterp = NULL;
CSLIDEWorld *CSLIDEWorld::m_pCSLIDEWorld = NULL;

DOUBLE CSLIDEWorld::m_dSecondsPerFrameRequested = gcfFrameTime;
DOUBLE CSLIDEWorld::m_dUpdateTime = 0;

BOOL CSLIDEWorld::m_bTclGUIExists = FALSE;
CWindow *CSLIDEWorld::m_pCSLIDEWindowEditObject = NULL;
CSLIDEViewport *CSLIDEWorld::m_pCSLIDEViewportEditObject = NULL;
CSLIDERender *CSLIDEWorld::m_pCSLIDERenderEditObject = NULL;

//////////////////////////////////////////////////////////////////////
// CSLIDEWorld Class
//

CSLIDEWorld::CSLIDEWorld()
{
  CSLIDESurface::m_pCSLIDESurfaceDefault = NULL;

  CSLIDEPrimitive::SetErrorCount(0);
  CSLIDEPrimitive::SetWarningCount(0);
}

CSLIDEWorld::~CSLIDEWorld()
{
  Uninit();

  CSLIDEPrimitive::SetErrorCount(0);
  CSLIDEPrimitive::SetWarningCount(0);
}

VOID CSLIDEWorld::Init()
{
  m_listCSLIDERender.Init();
  m_listCWindow.Init();
  m_hashCWindow.Init(1);
  m_listCSLIDEViewport.Init();
  m_hashCSLIDEViewport.Init(3);
  m_listCSLIDENode.Init();
  m_hashCSLIDENode.Init(101);
  m_listCSLIDEInstance.Init();
  m_hashCSLIDEInstance.Init(101);
  m_listCSLIDETransform.Init();
  m_listCSLIDEFace.Init();
  m_hashCSLIDEFace.Init(101);
  m_listCSLIDEPoint.Init();
  m_hashCSLIDEPoint.Init(101);
  m_listCSLIDEControlPoint.Init();
  m_hashCSLIDEControlPoint.Init(101);
  m_listCSLIDESurface.Init();
  m_hashCSLIDESurface.Init(101);
  m_listCSLIDEFog.Init();
  m_hashCSLIDEFog.Init(1);
  m_listCSLIDEGlobalInitTclBlock.Init();
  m_listCSLIDEGlobalUpdateTclBlock.Init();  
  m_listCSLIDEGlobalUpdateTclObj.Init();

  m_listDynamics.Init();

  InitDefaults();

  // Dynamics
  m_bDynamicsPaused = FALSE;
  m_dSecondsPerFrameRequested = gcfFrameTime;
  m_ttkWakeUpIDDynamics = 0;
  m_dTimeBase = 0.0;
  m_uiFrame = 0;
}

VOID CSLIDEWorld::InitDefaults()
{
  // Create Default Objects

  // Default Surface
  CSLIDESurface::m_pCSLIDESurfaceDefault = new CSLIDESurface;
  ASSERT( CSLIDESurface::m_pCSLIDESurfaceDefault != NULL );
  CSLIDESurface::m_pCSLIDESurfaceDefault->Init("", 0, 0, 
	(CHAR*)(gcppcSLIDEReservedWords[sliderwtSLF_INHERIT]));
  VERIFY(InsertSurface(CSLIDESurface::m_pCSLIDESurfaceDefault));  

  // Default Vertex Property
  CSLIDEMeshVertexProperty::m_pCSLIDEMeshVertexPropertyDefault = new CSLIDEMeshVertexProperty;
  ASSERT( CSLIDEMeshVertexProperty::m_pCSLIDEMeshVertexPropertyDefault != NULL );
  CSLIDEMeshVertexProperty::m_pCSLIDEMeshVertexPropertyDefault->Init("", 0, 0, 
	(CHAR*)(gcppcSLIDEReservedWords[sliderwtSLF_INHERIT]));

  // Default Edge Property
  CSLIDEMeshEdgeProperty::m_pCSLIDEMeshEdgePropertyDefault = new CSLIDEMeshEdgeProperty;
  ASSERT( CSLIDEMeshEdgeProperty::m_pCSLIDEMeshEdgePropertyDefault != NULL );
  CSLIDEMeshEdgeProperty::m_pCSLIDEMeshEdgePropertyDefault->Init("", 0, 0, 
	(CHAR*)(gcppcSLIDEReservedWords[sliderwtSLF_INHERIT]));
}

VOID CSLIDEWorld::Uninit()
{
  CListIterInt iter;
  CSLIDERender *pCSLIDERender;
  CWindow *pCWindow;
  CSLIDEViewport *pCSLIDEViewport;
  CSLIDENode *pCSLIDENode;
  CSLIDEInstance *pCSLIDEInstance;
  CSLIDETransform *pCSLIDETransform;
  CSLIDEFace *pCSLIDEFace;
  CSLIDEPoint *pCSLIDEPoint;
  CSLIDEControlPoint *pCSLIDEControlPoint;
  CSLIDESurface *pCSLIDESurface;
  CSLIDEFog *pCSLIDEFog;
  Tcl_Obj *pto;

  UnscheduleDynamics();

  // Uninit Defaults
  if ( CSLIDESurface::m_pCSLIDESurfaceDefault != NULL )
    {
      // This is in the list of surfaces, so it will be deleted 
      // at the end of this function
      //delete CSLIDESurface::m_pCSLIDESurfaceDefault;
      CSLIDESurface::m_pCSLIDESurfaceDefault = NULL;
    }

  if ( CSLIDEMeshEdgeProperty::m_pCSLIDEMeshEdgePropertyDefault != NULL )
    {
      delete CSLIDEMeshEdgeProperty::m_pCSLIDEMeshEdgePropertyDefault;
      CSLIDEMeshEdgeProperty::m_pCSLIDEMeshEdgePropertyDefault = NULL;
    }

  
  iter.Init(&m_listCSLIDEGlobalUpdateTclObj);

  for ( pto = (Tcl_Obj*)(iter.PeekFirst());
	pto != NULL;
	pto = (Tcl_Obj*)(iter.PeekNext()) )
    {
      Tcl_DecrRefCount(pto);
    }

  iter.Init(&m_listCSLIDERender);

  for ( pCSLIDERender = (CSLIDERender*)(iter.PeekFirst());
	pCSLIDERender != NULL;
	pCSLIDERender = (CSLIDERender*)(iter.PeekNext()) )
    {
      delete pCSLIDERender;
    }

  iter.Init(&m_listCWindow);

  for ( pCWindow = (CWindow*)(iter.PeekFirst());
	pCWindow != NULL;
	pCWindow = (CWindow*)(iter.PeekNext()) )
    {
      delete pCWindow;
    }

  iter.Init(&m_listCSLIDEViewport);

  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      delete pCSLIDEViewport;
    }

  iter.Init(&m_listCSLIDENode);

  for ( pCSLIDENode = (CSLIDENode*)(iter.PeekFirst());
	pCSLIDENode != NULL;
	pCSLIDENode = (CSLIDENode*)(iter.PeekNext()) )
    {
      delete pCSLIDENode;
    }

  iter.Init(&m_listCSLIDEInstance);

  for ( pCSLIDEInstance = (CSLIDEInstance*)(iter.PeekFirst());
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)(iter.PeekNext()) )
    {
      delete pCSLIDEInstance;
    }

  iter.Init(&m_listCSLIDETransform);

  for ( pCSLIDETransform = (CSLIDETransform*)(iter.PeekFirst());
	pCSLIDETransform != NULL;
	pCSLIDETransform = (CSLIDETransform*)(iter.PeekNext()) )
    {
      delete pCSLIDETransform;
    }

  iter.Init(&m_listCSLIDEFace);

  for ( pCSLIDEFace = (CSLIDEFace*)(iter.PeekFirst());
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)(iter.PeekNext()) )
    {
      delete pCSLIDEFace;
    }

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekFirst());
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekNext()) )
    {
      delete pCSLIDEPoint;
    }

  iter.Init(&m_listCSLIDEControlPoint);

  for ( pCSLIDEControlPoint = (CSLIDEControlPoint*)(iter.PeekFirst());
	pCSLIDEControlPoint != NULL;
	pCSLIDEControlPoint = (CSLIDEControlPoint*)(iter.PeekNext()) )
    {
      delete pCSLIDEControlPoint;
    }

  iter.Init(&m_listCSLIDESurface);

  for ( pCSLIDESurface = (CSLIDESurface*)(iter.PeekFirst());
	pCSLIDESurface != NULL;
	pCSLIDESurface = (CSLIDESurface*)(iter.PeekNext()) )
    {
      delete pCSLIDESurface;
    }

  iter.Init(&m_listCSLIDEFog);

  for ( pCSLIDEFog = (CSLIDEFog*)(iter.PeekFirst());
	pCSLIDEFog != NULL;
	pCSLIDEFog = (CSLIDEFog*)(iter.PeekNext()) )
    {
      delete pCSLIDEFog;
    }
}

VOID CSLIDEWorld::UninitTcl()
{
  CListIterInt iter;
  CWindow *pCWindow;

  UnscheduleDynamics();

  iter.Init(&m_listCWindow);

  for ( pCWindow = (CWindow*)(iter.PeekFirst());
	pCWindow != NULL;
	pCWindow = (CWindow*)(iter.PeekNext()) )
    {
      Tk_DoWhenIdle(CWindow::CBHandleDestroy, (ClientData)pCWindow);
    }
}

BOOL CSLIDEWorld::Initialized()
{
  return TRUE;
}

BOOL CSLIDEWorld::InsertRender(CSLIDERender *pCSLIDERender)
{
  ASSERT( pCSLIDERender != NULL );

  m_listCSLIDERender.InsertLast(pCSLIDERender, (UINT_32)pCSLIDERender);

  return TRUE;
}

BOOL CSLIDEWorld::InsertWindow(CWindow *pCWindow)
{
  CHAR *pcID;

  ASSERT( pCWindow != NULL );

  pcID = pCWindow->GetSLIDEID();

  m_hashCWindow.Insert(pCWindow, (UINT_32)pcID);
  m_listCWindow.InsertLast(pCWindow, (UINT_32)pCWindow);

  return TRUE;
}

CWindow *CSLIDEWorld::PeekWindow(CHAR *pcID)
{
  ASSERT( pcID != NULL );

  return (CWindow*)(m_hashCWindow.Peek((UINT_32)pcID));
}

BOOL CSLIDEWorld::InsertViewport(CSLIDEViewport *pCSLIDEViewport)
{
  CHAR *pcID;

  ASSERT( pCSLIDEViewport != NULL );

  pcID = pCSLIDEViewport->GetID();

  m_hashCSLIDEViewport.Insert(pCSLIDEViewport, (UINT_32)pcID);
  m_listCSLIDEViewport.InsertLast(pCSLIDEViewport, (UINT_32)pCSLIDEViewport);

  return TRUE;
}

CSLIDEViewport *CSLIDEWorld::PeekViewport(CHAR *pcID)
{
  ASSERT( pcID != NULL );

  return (CSLIDEViewport*)(m_hashCSLIDEViewport.Peek((UINT_32)pcID));
}
/*
BOOL CSLIDEWorld::InsertCamera(CSLIDECamera *pCSLIDECamera)
{
  CHAR *pcID;

  ASSERT( pCSLIDECamera != NULL );

  pcID = pCSLIDECamera->GetID();

  m_hashCSLIDECamera.Insert(pCSLIDECamera, (UINT_32)pcID);
  m_listCSLIDECamera.InsertLast(pCSLIDECamera, 0);

  return TRUE;
}
*/

BOOL CSLIDEWorld::InsertNode(CSLIDENode *pCSLIDENode)
{
  ASSERT( pCSLIDENode != NULL );

  CHAR *pcID;
  pcID = pCSLIDENode->GetID();

  m_hashCSLIDENode.Insert(pCSLIDENode, (UINT_32)pcID);
  m_listCSLIDENode.InsertLast(pCSLIDENode, (UINT_32)pCSLIDENode);

  return TRUE;
}

BOOL CSLIDEWorld::RemoveNode(CSLIDENode *pCSLIDENode)
{
  CHAR *pcID;

  ASSERT( pCSLIDENode != NULL );

  pcID = pCSLIDENode->GetID();

  m_hashCSLIDENode.Remove((UINT_32)pcID);
  m_listCSLIDENode.RemoveUnordered((UINT_32)pCSLIDENode);
  m_listDynamics.RemoveUnordered((UINT_32)pCSLIDENode);

  return TRUE;
}

CSLIDENode *CSLIDEWorld::PeekNode(CHAR *pcID)
{
  ASSERT( pcID != NULL );

  return (CSLIDENode*)(m_hashCSLIDENode.Peek((UINT_32)pcID));
}

BOOL CSLIDEWorld::InsertInstance(CSLIDEInstance *pCSLIDEInstance)
{
  ASSERT( pCSLIDEInstance != NULL );

  m_listCSLIDEInstance.InsertLast(pCSLIDEInstance, (UINT_32)pCSLIDEInstance);

  return TRUE;
}

BOOL CSLIDEWorld::RemoveInstance(CSLIDEInstance *pCSLIDEInstance)
{
  ASSERT( pCSLIDEInstance != NULL );

  m_listCSLIDEInstance.RemoveUnordered((UINT_32)pCSLIDEInstance);
  m_listDynamics.RemoveUnordered((UINT_32)pCSLIDEInstance);

  return TRUE;
}

BOOL CSLIDEWorld::InsertInstanceID(CSLIDEInstance *pCSLIDEInstance, CHAR *pcID)
{
  m_hashCSLIDEInstance.Insert(pCSLIDEInstance, (UINT_32)pcID);

  return TRUE;
}

CSLIDEInstance *CSLIDEWorld::PeekInstance(CHAR *pcID)
{
  ASSERT( pcID != NULL );

  return (CSLIDEInstance*)(m_hashCSLIDEInstance.Peek((UINT_32)pcID));
}

BOOL CSLIDEWorld::InsertTransform(CSLIDETransform *pCSLIDETransform)
{
  ASSERT( pCSLIDETransform != NULL );

  m_listCSLIDETransform.InsertLast(pCSLIDETransform, (UINT_32)pCSLIDETransform);

  return TRUE;
}

BOOL CSLIDEWorld::RemoveTransform(CSLIDETransform *pCSLIDETransform)
{
  ASSERT( pCSLIDETransform != NULL );

  m_listCSLIDETransform.RemoveUnordered((UINT_32)pCSLIDETransform);
  m_listDynamics.RemoveUnordered((UINT_32)pCSLIDETransform);

  return TRUE;
}

BOOL CSLIDEWorld::InsertFace(CSLIDEFace *pCSLIDEFace)
{
  CHAR *pcID;

  ASSERT( pCSLIDEFace != NULL );

  pcID = pCSLIDEFace->GetID();

  m_hashCSLIDEFace.Insert(pCSLIDEFace, (UINT_32)pcID);
  m_listCSLIDEFace.InsertLast(pCSLIDEFace, (UINT_32)pCSLIDEFace);

  return TRUE;
}


BOOL CSLIDEWorld::RemoveFace(CSLIDEFace *pCSLIDEFace)
{
  CHAR *pcID;

  ASSERT( pCSLIDEFace != NULL );

  pcID = pCSLIDEFace->GetID();

  m_hashCSLIDEFace.Remove((UINT_32)pcID);
  m_listCSLIDEFace.RemoveUnordered((UINT_32)pCSLIDEFace);
  m_listDynamics.RemoveUnordered((UINT_32)pCSLIDEFace);

  return TRUE;
}

CSLIDEFace *CSLIDEWorld::PeekFace(CHAR *pcID)
{
  ASSERT( pcID != NULL );

  return (CSLIDEFace*)(m_hashCSLIDEFace.Peek((UINT_32)pcID));
}

BOOL CSLIDEWorld::InsertPoint(CSLIDEPoint *pCSLIDEPoint)
{
  CHAR *pcID;

  ASSERT( pCSLIDEPoint != NULL );

  pcID = pCSLIDEPoint->GetID();

  m_hashCSLIDEPoint.Insert(pCSLIDEPoint, (UINT_32)pcID);
  m_listCSLIDEPoint.InsertLast(pCSLIDEPoint, (UINT_32)pCSLIDEPoint);

  return TRUE;
}

BOOL CSLIDEWorld::RemovePoint(CSLIDEPoint *pCSLIDEPoint)
{
  CHAR *pcID;

  ASSERT( pCSLIDEPoint != NULL );

  pcID = pCSLIDEPoint->GetID();

  m_hashCSLIDEPoint.Remove((UINT_32)pcID);
  m_listCSLIDEPoint.RemoveUnordered((UINT_32)pCSLIDEPoint);
  m_listDynamics.RemoveUnordered((UINT_32)pCSLIDEPoint);

  return TRUE;
}

CSLIDEPoint *CSLIDEWorld::PeekPoint(CHAR *pcID)
{
  ASSERT( pcID != NULL );

  return (CSLIDEPoint*)(m_hashCSLIDEPoint.Peek((UINT_32)pcID));
}

BOOL CSLIDEWorld::InsertControlPoint(CSLIDEControlPoint *pCSLIDEControlPoint)
{
  CHAR *pcID;

  ASSERT( pCSLIDEControlPoint != NULL );

  pcID = pCSLIDEControlPoint->GetID();

  m_hashCSLIDEControlPoint.Insert(pCSLIDEControlPoint, (UINT_32)pcID);
  m_listCSLIDEControlPoint.InsertLast(pCSLIDEControlPoint, (UINT_32)pCSLIDEControlPoint);

  return TRUE;
}

BOOL CSLIDEWorld::RemoveControlPoint(CSLIDEControlPoint *pCSLIDEControlPoint)
{
  CHAR *pcID;

  ASSERT( pCSLIDEControlPoint != NULL );

  pcID = pCSLIDEControlPoint->GetID();

  m_hashCSLIDEControlPoint.Remove((UINT_32)pcID);
  m_listCSLIDEControlPoint.RemoveUnordered((UINT_32)pCSLIDEControlPoint);
  m_listDynamics.RemoveUnordered((UINT_32)pCSLIDEControlPoint);

  return TRUE;
}

CSLIDEControlPoint *CSLIDEWorld::PeekControlPoint(CHAR *pcID)
{
  ASSERT( pcID != NULL );

  return (CSLIDEControlPoint*)(m_hashCSLIDEControlPoint.Peek((UINT_32)pcID));
}

BOOL CSLIDEWorld::InsertSurface(CSLIDESurface *pCSLIDESurface)
{
  CHAR *pcID;

  ASSERT( pCSLIDESurface != NULL );

  pcID = pCSLIDESurface->GetID();

  m_hashCSLIDESurface.Insert(pCSLIDESurface, (UINT_32)pcID);
  m_listCSLIDESurface.InsertLast(pCSLIDESurface, (UINT_32)pCSLIDESurface);

  return TRUE;
}

BOOL CSLIDEWorld::RemoveSurface(CSLIDESurface *pCSLIDESurface)
{
  CHAR *pcID;

  ASSERT( pCSLIDESurface != NULL );

  pcID = pCSLIDESurface->GetID();

  m_hashCSLIDESurface.Remove((UINT_32)pcID);
  m_listCSLIDESurface.RemoveUnordered((UINT_32)pCSLIDESurface);
  m_listDynamics.RemoveUnordered((UINT_32)pCSLIDESurface);

  return TRUE;
}

CSLIDESurface *CSLIDEWorld::PeekSurface(CHAR *pcID)
{
  ASSERT( pcID != NULL );

  return (CSLIDESurface*)(m_hashCSLIDESurface.Peek((UINT_32)pcID));
}

BOOL CSLIDEWorld::InsertFog(CSLIDEFog *pCSLIDEFog)
{
  CHAR *pcID;

  ASSERT( pCSLIDEFog != NULL );

  pcID = pCSLIDEFog->GetID();

  m_hashCSLIDEFog.Insert(pCSLIDEFog, (UINT_32)pcID);
  m_listCSLIDEFog.InsertLast(pCSLIDEFog, (UINT_32)pCSLIDEFog);

  return TRUE;
}

BOOL CSLIDEWorld::RemoveFog(CSLIDEFog *pCSLIDEFog)
{
  CHAR *pcID;
  
  ASSERT( pCSLIDEFog != NULL );

  pcID = pCSLIDEFog->GetID();

  m_hashCSLIDEFog.Remove((UINT_32)pcID);
  m_listCSLIDEFog.RemoveUnordered((UINT_32)pCSLIDEFog);
  m_listDynamics.RemoveUnordered((UINT_32)pCSLIDEFog);

  return TRUE;
}

CSLIDEFog *CSLIDEWorld::PeekFog(CHAR *pcID)
{
  ASSERT( pcID != NULL );

  return (CSLIDEFog*)(m_hashCSLIDEFog.Peek((UINT_32)pcID));
}

BOOL CSLIDEWorld::InsertGlobalInitTclBlock(CHAR *pcTclBlock)
{
  ASSERT( pcTclBlock != NULL );

  m_listCSLIDEGlobalInitTclBlock.InsertLast(pcTclBlock, 0);

  return TRUE;
}

BOOL CSLIDEWorld::InsertGlobalUpdateTclBlock(CHAR *pcTclBlock)
{
  Tcl_Obj *pto;

  ASSERT( pcTclBlock != NULL );

  pto = Tcl_NewStringObj(pcTclBlock, -1);
  ASSERT( pto != NULL );

  m_listCSLIDEGlobalUpdateTclBlock.InsertLast(pcTclBlock, 0);
  m_listCSLIDEGlobalUpdateTclObj.InsertLast(pto, 0);

  return TRUE;
}

BOOL CSLIDEWorld::CheckSemantics()
{
  BOOL b, bCorrect;
  CListIterInt iter;
  CSLIDERender *pCSLIDERender;
  CWindow *pCWindow;
  CSLIDEViewport *pCSLIDEViewport;
  CSLIDENode *pCSLIDENode;
  CSLIDEInstance *pCSLIDEInstance;
  CSLIDETransform *pCSLIDETransform;
  CSLIDEFace *pCSLIDEFace;
  CSLIDEPoint *pCSLIDEPoint;
  CSLIDEControlPoint *pCSLIDEControlPoint;
  CSLIDESurface *pCSLIDESurface;
  CSLIDEFog *pCSLIDEFog;
  CHAR *pcTclBlock;
  INT_32 iCode;

  bCorrect = TRUE;

  // Create the toplevel tk widget before calling the Tcl Blocks
  iter.Init(&m_listCWindow);

  for ( pCWindow = (CWindow*)(iter.PeekFirst());
	pCWindow != NULL;
	pCWindow = (CWindow*)(iter.PeekNext()) )
    {
      pCWindow->CreateTclToplevel();
    }

  fprintf(stderr, "BEGIN: Executing global tclinit blocks...\n");

  iter.Init(&m_listCSLIDEGlobalInitTclBlock);

  for ( pcTclBlock = (CHAR*)(iter.PeekFirst());
	pcTclBlock != NULL;
	pcTclBlock = (CHAR*)(iter.PeekNext()) )
    {
      iCode = Tcl_GlobalEval(m_pInterp, pcTclBlock);
      if ( iCode != TCL_OK )
	{
	  fprintf(CSLIDEPrimitive::GetErrorFile(), 
		  "Error: Tcl returned error code \"%s\" for the block:\n%s",
		  GetInterp()->result, pcTclBlock);
	  CSLIDEPrimitive::SetErrorCount(CSLIDEPrimitive::GetErrorCount()+1);
	  bCorrect = FALSE;
	}
    }

  fprintf(stderr, "\nEND: Executing global tclinit blocks\n");

  // Create the toplevel tk widget before calling the Tcl Blocks
  iter.Init(&m_listCWindow);

  for ( pCWindow = (CWindow*)(iter.PeekFirst());
	pCWindow != NULL;
	pCWindow = (CWindow*)(iter.PeekNext()) )
    {
      pCWindow->CreateTclToplevel();
    }


  iter.Init(&m_listCSLIDESurface);

  for ( pCSLIDESurface = (CSLIDESurface*)(iter.PeekFirst());
	pCSLIDESurface != NULL;
	pCSLIDESurface = (CSLIDESurface*)(iter.PeekNext()) )
    {
      b = pCSLIDESurface->CheckSemantics(this);
      bCorrect = bCorrect && b;
    }

  iter.Init(&m_listCSLIDEFog);

  for ( pCSLIDEFog = (CSLIDEFog*)(iter.PeekFirst());
	pCSLIDEFog != NULL;
	pCSLIDEFog = (CSLIDEFog*)(iter.PeekNext()) )
    {
      b = pCSLIDEFog->CheckSemantics(this);
      bCorrect = bCorrect && b;
    }

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekFirst());
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekNext()) )
    {
      b = pCSLIDEPoint->CheckSemantics(this);
      bCorrect = bCorrect && b;
    }

  iter.Init(&m_listCSLIDEControlPoint);

  for ( pCSLIDEControlPoint = (CSLIDEControlPoint*)(iter.PeekFirst());
	pCSLIDEControlPoint != NULL;
	pCSLIDEControlPoint = (CSLIDEControlPoint*)(iter.PeekNext()) )
    {
      b = pCSLIDEControlPoint->CheckSemantics(this);
      bCorrect = bCorrect && b;
    }

  iter.Init(&m_listCSLIDEFace);

  for ( pCSLIDEFace = (CSLIDEFace*)(iter.PeekFirst());
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)(iter.PeekNext()) )
    {
      b = pCSLIDEFace->CheckSemantics(this);
      bCorrect = bCorrect && b;
    }

  iter.Init(&m_listCSLIDETransform);

  for ( pCSLIDETransform = (CSLIDETransform*)(iter.PeekFirst());
	pCSLIDETransform != NULL;
	pCSLIDETransform = (CSLIDETransform*)(iter.PeekNext()) )
    {
      b = pCSLIDETransform->CheckSemantics(this);
      bCorrect = bCorrect && b;
    }

  iter.Init(&m_listCSLIDEInstance);

  for ( pCSLIDEInstance = (CSLIDEInstance*)(iter.PeekFirst());
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)(iter.PeekNext()) )
    {
      b = pCSLIDEInstance->CheckSemantics(this);
      bCorrect = bCorrect && b;
    }

  iter.Init(&m_listCSLIDENode);

  for ( pCSLIDENode = (CSLIDENode*)(iter.PeekFirst());
	pCSLIDENode != NULL;
	pCSLIDENode = (CSLIDENode*)(iter.PeekNext()) )
    {
      b = pCSLIDENode->CheckSemantics(this);
      bCorrect = bCorrect && b;
    }

  // Resolve Render Statements 
  if ( m_listCSLIDERender.GetLength() == 0 )
    {
      fprintf(CSLIDEPrimitive::GetErrorFile(), 
	      "Warning: No %s statement was defined so nothing will be rendered.\n", 
	      gcppcSLIDEReservedWords[sliderwtRender]);
      CSLIDEPrimitive::SetWarningCount(CSLIDEPrimitive::GetWarningCount()+1);
    }

  iter.Init(&m_listCSLIDERender);

  for ( pCSLIDERender = (CSLIDERender*)(iter.PeekFirst());
	pCSLIDERender != NULL;
	pCSLIDERender = (CSLIDERender*)(iter.PeekNext()) )
    {
      b = pCSLIDERender->CheckSemantics(this);
      bCorrect = bCorrect && b;
    }

  iter.Init(&m_listCSLIDEViewport);

  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      b = pCSLIDEViewport->CheckSemantics(this);
      bCorrect = bCorrect && b;
    }

  iter.Init(&m_listCWindow);

  for ( pCWindow = (CWindow*)(iter.PeekFirst());
	pCWindow != NULL;
	pCWindow = (CWindow*)(iter.PeekNext()) )
    {
      b = pCWindow->CheckSemantics(this);
      bCorrect = bCorrect && b;
    }

  gbCheckSemanticsDone = TRUE;

  return bCorrect;
}

BOOL CSLIDEWorld::Valid()
{
  return TRUE;
}

VOID CSLIDEWorld::DrawGUIGraph()
{
  CListIterInt iter, iter2;
  CSLIDENode *pCSLIDENode;
  CSLIDEInstance *pCSLIDEInstance;
  CSLIDEFace *pCSLIDEFace;
  CSLIDEPoint *pCSLIDEPoint;
  //CSLIDEControlPoint *pCSLIDEControlPoint;
  CSLIDESurface *pCSLIDESurface;
  CSLIDEFog *pCSLIDEFog;

  INT_32 iCode;
  CHAR pcBuffer[200];

  iter.Init(&m_listCSLIDESurface);
  for ( pCSLIDESurface = (CSLIDESurface*)(iter.PeekFirst());
	pCSLIDESurface != NULL;
	pCSLIDESurface = (CSLIDESurface*)(iter.PeekNext()) )
    {
      if( strcmp(pCSLIDESurface->GetID(), "SLF_INHERIT") ) 
	  {
	    sprintf(pcBuffer, "Graph:CreateNode surface %s %ld %ld", 
		    pCSLIDESurface->GetID(), 
		    pCSLIDESurface->GetGraphPositionX(), 
		    pCSLIDESurface->GetGraphPositionY());
	    iCode = Tcl_Eval(m_pInterp, pcBuffer);
	    if ( iCode != TCL_OK )
	      {
		fprintf(stderr, "%s\n", pcBuffer );
		fprintf(stderr, "Tcl Error: \"%s\"\n", m_pInterp->result);
		ASSERT( FALSE );
	      }
	    sprintf(pcBuffer, "Editor:AddSurface %s", pCSLIDESurface->GetID());
	    iCode = Tcl_Eval(m_pInterp, pcBuffer);
	    //ASSERT( iCode == TCL_OK );
	    if ( iCode != TCL_OK )
	      {
		fprintf(stderr, "%s\n", pcBuffer );
		fprintf(stderr, "Error %s \n", m_pInterp->result);
	      }
	  }
    }

  iter.Init(&m_listCSLIDEFog);
  for ( pCSLIDEFog = (CSLIDEFog*)(iter.PeekFirst());
	pCSLIDEFog != NULL;
	pCSLIDEFog = (CSLIDEFog*)(iter.PeekNext()) )
    {
      sprintf(pcBuffer, "Graph:CreateNode fog %s %ld %ld", 
	      pCSLIDEFog->GetID(), 
	      pCSLIDEFog->GetGraphPositionX(), 
	      pCSLIDEFog->GetGraphPositionY());
      iCode = Tcl_Eval(m_pInterp, pcBuffer);
      ASSERT( iCode == TCL_OK );
    }

  iter.Init(&m_listCSLIDEPoint);
  for ( pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekFirst());
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekNext()) )
    {
      sprintf(pcBuffer, "Graph:CreateNode point %s %ld %ld", 
	      pCSLIDEPoint->GetID(), 
	      pCSLIDEPoint->GetGraphPositionX(), 
	      pCSLIDEPoint->GetGraphPositionY());
      iCode = Tcl_Eval(m_pInterp, pcBuffer);
      ASSERT( iCode == TCL_OK );
    }

  /*
  iter.Init(&m_listCSLIDEControlPoint);
  for ( pCSLIDEControlPoint = (CSLIDEControlPoint*)(iter.PeekFirst());
	pCSLIDEControlPoint != NULL;
	pCSLIDEControlPoint = (CSLIDEControlPoint*)(iter.PeekNext()) )
    {
      sprintf(pcBuffer, "Graph:CreateNode ControlPoint %s %ld %ld", 
	      pCSLIDEControlPoint->GetID(), 
	      pCSLIDEControlPoint->GetGraphPositionX(), 
	      pCSLIDEControlPoint->GetGraphPositionY());
      iCode = Tcl_Eval(m_pInterp, pcBuffer);
      ASSERT( iCode == TCL_OK );
    }
    */

  iter.Init(&m_listCSLIDEFace);
  for ( pCSLIDEFace = (CSLIDEFace*)(iter.PeekFirst());
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)(iter.PeekNext()) )
    {
      sprintf(pcBuffer, "Graph:CreateNode face %s %ld %ld", 
	      pCSLIDEFace->GetID(), 
	      pCSLIDEFace->GetGraphPositionX(), 
	      pCSLIDEFace->GetGraphPositionY());
      iCode = Tcl_Eval(m_pInterp, pcBuffer);
      ASSERT( iCode == TCL_OK );

      pCSLIDEFace->DrawGUIGraphPointReferences(m_pInterp);
    }

  iter.Init(&m_listCSLIDENode);
  for ( pCSLIDENode = (CSLIDENode*)(iter.PeekFirst());
	pCSLIDENode != NULL;
	pCSLIDENode = (CSLIDENode*)(iter.PeekNext()) )
    {
      sprintf(pcBuffer, "Graph:CreateNode %s %s %ld %ld", 
	      pCSLIDENode->GetTypeString(),
	      pCSLIDENode->GetID(), 
	      pCSLIDENode->GetGraphPositionX(), 
	      pCSLIDENode->GetGraphPositionY());
      iCode = Tcl_Eval(m_pInterp, pcBuffer);
      //ASSERT( iCode == TCL_OK );
      if ( iCode != TCL_OK )
	{
	  printf( "%s\n", pcBuffer );
	  printf("Error %s \n", m_pInterp->result);
	}

      if( pCSLIDENode->GetType() == sliderwtObject )
	{
	  ((CSLIDEObject*)pCSLIDENode)->DrawGUIGraphFaceReferences(m_pInterp);
	}
    }

  iter.Init(&m_listCSLIDEInstance);
  for ( pCSLIDEInstance = (CSLIDEInstance*)(iter.PeekFirst());
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)(iter.PeekNext()) )
    {
      sprintf(pcBuffer, "Graph:CreateInstance %s %s %s", 
	      pCSLIDEInstance->GetID(), 
	      pCSLIDEInstance->GetParentID(), 
	      pCSLIDEInstance->GetChildID() );
      iCode = Tcl_Eval(m_pInterp, pcBuffer);
      //ASSERT( iCode == TCL_OK );
      if ( iCode != TCL_OK )
	{
	  printf( "%s\n", pcBuffer );
	  printf("Error %s \n", m_pInterp->result);
	}
    }

  /*
  UINT_32 i;
  CSLIDERender *pCSLIDERender;
  CWindow *pCWindow;
  CSLIDEViewport *pCSLIDEViewport;
  CSLIDETransform *pCSLIDETransform;
  CSLIDELightPath *pCSLIDELightPath;
  CSLIDEAreaLightPath *pCSLIDEAreaLightPath;

  iter.Init(&m_listCSLIDERender);
  for ( pCSLIDERender = (CSLIDERender*)(iter.PeekFirst()), i = 1;
	pCSLIDERender != NULL;
	pCSLIDERender = (CSLIDERender*)(iter.PeekNext()), i++ )
    {
      sprintf(pcBuffer, "Graph:CreateNode render render%ld %ld %ld", i,
	      pCSLIDERender->GetGraphPositionX(), 
	      pCSLIDERender->GetGraphPositionY());
      iCode = Tcl_Eval(m_pInterp, pcBuffer);
      //      ASSERT( iCode == TCL_OK );
      if ( iCode != TCL_OK )
	{
	  printf( "%s\n", pcBuffer );
	  printf("Error %s \n", m_pInterp->result);
	}

      sprintf(pcBuffer, "Graph:CreateReference render%ld %s", i,
	      pCSLIDERender->GetWorldRootID());
      iCode = Tcl_Eval(m_pInterp, pcBuffer);
      ASSERT( iCode == TCL_OK );

      sprintf(pcBuffer, "Graph:CreateReference render%ld %s", i,
	      pCSLIDERender->GetCameraPathRootID());
      iCode = Tcl_Eval(m_pInterp, pcBuffer);
      ASSERT( iCode == TCL_OK );

      if( pCSLIDERender->GetFogID() != NULL )
	{
	  sprintf(pcBuffer, "Graph:CreateReference render%ld %s", i, 
		  pCSLIDERender->GetFogID());
	  iCode = Tcl_Eval(m_pInterp, pcBuffer);
	  ASSERT( iCode == TCL_OK );
	}

      iter2.Init(&(pCSLIDERender->m_listCSLIDELightPath));
      for ( pCSLIDELightPath = (CSLIDELightPath*)(iter2.PeekFirst());
	    pCSLIDELightPath != NULL;
	    pCSLIDELightPath = (CSLIDELightPath*)(iter2.PeekNext()) )
	{
	  sprintf(pcBuffer, "Graph:CreateReference render%ld %s", i,
		  pCSLIDELightPath->GetRootID());
	  iCode = Tcl_Eval(m_pInterp, pcBuffer);
	  ASSERT( iCode == TCL_OK );
	}

      iter2.Init(&(pCSLIDERender->m_listCSLIDEAreaLightPath));
      for ( pCSLIDEAreaLightPath = (CSLIDEAreaLightPath*)(iter2.PeekFirst());
	    pCSLIDEAreaLightPath != NULL;
	    pCSLIDEAreaLightPath = (CSLIDEAreaLightPath*)(iter2.PeekNext()) )
	{
	  sprintf(pcBuffer, "Graph:CreateReference render%ld %s", i,
		  pCSLIDEAreaLightPath->GetRootID());
	  iCode = Tcl_Eval(m_pInterp, pcBuffer);
	  ASSERT( iCode == TCL_OK );
	}
    }

  iter.Init(&m_listCSLIDEViewport);
  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      sprintf(pcBuffer, "Graph:CreateNode viewport %s %ld %ld", 
	      pCSLIDEViewport->GetID(), 
	      pCSLIDEViewport->GetGraphPositionX(), 
	      pCSLIDEViewport->GetGraphPositionY());
      iCode = Tcl_Eval(m_pInterp, pcBuffer);
      ASSERT( iCode == TCL_OK );
    }

  iter.Init(&m_listCWindow);
  for ( pCWindow = (CWindow*)(iter.PeekFirst());
	pCWindow != NULL;
	pCWindow = (CWindow*)(iter.PeekNext()) )
    {
      sprintf(pcBuffer, "Graph:CreateNode window foo %ld %ld", 
	      pCWindow->GetGraphPositionX(), 
	      pCWindow->GetGraphPositionY());
      iCode = Tcl_Eval(m_pInterp, pcBuffer);
      ASSERT( iCode == TCL_OK );
    }
    */
}

VOID CSLIDEWorld::PreprocessGUI()
{
  Preprocess();
  DrawGUIGraph();
}

VOID CSLIDEWorld::Preprocess()
{
  CListIterInt iter;
  CSLIDERender *pCSLIDERender;
  CWindow *pCWindow;
  CSLIDEViewport *pCSLIDEViewport;
  CSLIDENode *pCSLIDENode;
  CSLIDEInstance *pCSLIDEInstance;
  CSLIDETransform *pCSLIDETransform;
  CSLIDEFace *pCSLIDEFace;
  CSLIDEPoint *pCSLIDEPoint;
  CSLIDEControlPoint *pCSLIDEControlPoint;
  CSLIDESurface *pCSLIDESurface;
  CSLIDEFog *pCSLIDEFog;

  UpdateTclBlocks();

  iter.Init(&m_listCSLIDESurface);

  for ( pCSLIDESurface = (CSLIDESurface*)(iter.PeekFirst());
	pCSLIDESurface != NULL;
	pCSLIDESurface = (CSLIDESurface*)(iter.PeekNext()) )
    {
      pCSLIDESurface->Preprocess();
    }

  iter.Init(&m_listCSLIDEFog);

  for ( pCSLIDEFog = (CSLIDEFog*)(iter.PeekFirst());
	pCSLIDEFog != NULL;
	pCSLIDEFog = (CSLIDEFog*)(iter.PeekNext()) )
    {
	   pCSLIDEFog->Preprocess();
    }

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekFirst());
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekNext()) )
    {
      pCSLIDEPoint->Preprocess();
    }

  iter.Init(&m_listCSLIDEControlPoint);

  for ( pCSLIDEControlPoint = (CSLIDEControlPoint*)(iter.PeekFirst());
	pCSLIDEControlPoint != NULL;
	pCSLIDEControlPoint = (CSLIDEControlPoint*)(iter.PeekNext()) )
    {
      pCSLIDEControlPoint->Preprocess();
    }

  iter.Init(&m_listCSLIDEFace);

  for ( pCSLIDEFace = (CSLIDEFace*)(iter.PeekFirst());
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)(iter.PeekNext()) )
    {
      pCSLIDEFace->Preprocess();
    }

  iter.Init(&m_listCSLIDETransform);

  for ( pCSLIDETransform = (CSLIDETransform*)(iter.PeekFirst());
	pCSLIDETransform != NULL;
	pCSLIDETransform = (CSLIDETransform*)(iter.PeekNext()) )
    {
      pCSLIDETransform->Preprocess();
    }

  iter.Init(&m_listCSLIDEInstance);

  for ( pCSLIDEInstance = (CSLIDEInstance*)(iter.PeekFirst());
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)(iter.PeekNext()) )
    {
      pCSLIDEInstance->Preprocess();
    }

  iter.Init(&m_listCSLIDENode);

  for ( pCSLIDENode = (CSLIDENode*)(iter.PeekFirst());
	pCSLIDENode != NULL;
	pCSLIDENode = (CSLIDENode*)(iter.PeekNext()) )
    {
      pCSLIDENode->Preprocess();
    }

  iter.Init(&m_listCSLIDEFace);

  for ( pCSLIDEFace = (CSLIDEFace*)(iter.PeekFirst());
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)(iter.PeekNext()) )
    {
      pCSLIDEFace->PreprocessAutoNormals();
    }

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekFirst());
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekNext()) )
    {
      pCSLIDEPoint->PreprocessNormal();
    }

  iter.Init(&m_listCSLIDEFace);

  for ( pCSLIDEFace = (CSLIDEFace*)(iter.PeekFirst());
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)(iter.PeekNext()) )
    {
      pCSLIDEFace->CheckAutoNormalsDynamic();
    }

  iter.Init(&m_listCSLIDEViewport);

  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      pCSLIDEViewport->Preprocess();
    }

  iter.Init(&m_listCSLIDERender);

  for ( pCSLIDERender = (CSLIDERender*)(iter.PeekFirst());
	pCSLIDERender != NULL;
	pCSLIDERender = (CSLIDERender*)(iter.PeekNext()) )
    {
      pCSLIDERender->Preprocess();
    }

  gbPreprocessDone = TRUE;

  PreprocessGraphs();

  CreateDynamicsList();

  // Preprocessing the Windows will actually create the windows and
  // bring them up on the screen.  This is why we do this last.
  CListIterInt iterNode;

  iter.Init(&m_listCWindow);

  for ( pCWindow = (CWindow*)(iter.PeekFirst());
	pCWindow != NULL;
	pCWindow = (CWindow*)(iter.PeekNext()) )
    {
      iterNode.Init(&m_listCSLIDEInstance);
      
      for ( pCSLIDEInstance = (CSLIDEInstance*)(iterNode.PeekFirst());
	    pCSLIDEInstance != NULL;
	    pCSLIDEInstance = (CSLIDEInstance*)(iterNode.PeekNext()) )
	{
	  pCSLIDEInstance->SetVisited(FALSE);
	  pCSLIDEInstance->SetPrerenderVisited(FALSE);
	}
      
      iterNode.Init(&m_listCSLIDENode);
      
      for ( pCSLIDENode = (CSLIDENode*)(iterNode.PeekFirst());
	    pCSLIDENode != NULL;
	    pCSLIDENode = (CSLIDENode*)(iterNode.PeekNext()) )
	{
	  pCSLIDENode->SetVisited(FALSE);
	  pCSLIDENode->SetPrerenderVisited(FALSE);
	}
      
      pCWindow->Preprocess();
    }

  // Get the dynamics going
  RestartDynamics();
}

VOID CSLIDEWorld::PreprocessGraphs()
{
  CListIterInt iter;
  CSLIDERender *pCSLIDERender;

  // Run through the scene graphs of each of the render statements
  // to find which subtrees are static and to calculate all of the 
  // bounding boxes

  iter.Init(&m_listCSLIDERender);

  for ( pCSLIDERender = (CSLIDERender*)(iter.PeekFirst());
	pCSLIDERender != NULL;
	pCSLIDERender = (CSLIDERender*)(iter.PeekNext()) )
    {
      pCSLIDERender->PreprocessGraph();
    }
}

VOID CSLIDEWorld::CreateDynamicsList()
{
  CListIterInt iter;
  CSLIDERender *pCSLIDERender;
  CSLIDEViewport *pCSLIDEViewport;
  CSLIDENode *pCSLIDENode;
  CSLIDEInstance *pCSLIDEInstance;
  CSLIDETransform *pCSLIDETransform;
  CSLIDEFace *pCSLIDEFace;
  CSLIDEPoint *pCSLIDEPoint;
  CSLIDEControlPoint *pCSLIDEControlPoint;
  CSLIDESurface *pCSLIDESurface;
  CSLIDEFog *pCSLIDEFog;

  iter.Init(&m_listCSLIDESurface);

  for ( pCSLIDESurface = (CSLIDESurface*)(iter.PeekFirst());
	pCSLIDESurface != NULL;
	pCSLIDESurface = (CSLIDESurface*)(iter.PeekNext()) )
    {
      if ( pCSLIDESurface->GetDynamic() )
	{
	  m_listDynamics.InsertLast(pCSLIDESurface, (UINT_32)pCSLIDESurface);
	}
    }

  iter.Init(&m_listCSLIDEFog);

  for ( pCSLIDEFog = (CSLIDEFog*)(iter.PeekFirst());
	pCSLIDEFog != NULL;
	pCSLIDEFog = (CSLIDEFog*)(iter.PeekNext()) )
    {
      if ( pCSLIDEFog->GetDynamic() )
	{
	  m_listDynamics.InsertLast(pCSLIDEFog, (UINT_32)pCSLIDEFog);
	}
    }

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekFirst());
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekNext()) )
    {
      if ( pCSLIDEPoint->GetDynamic() )
	{
	  m_listDynamics.InsertLast(pCSLIDEPoint, (UINT_32)pCSLIDEPoint);
	}
    }

  iter.Init(&m_listCSLIDEControlPoint);

  for ( pCSLIDEControlPoint = (CSLIDEControlPoint*)(iter.PeekFirst());
	pCSLIDEControlPoint != NULL;
	pCSLIDEControlPoint = (CSLIDEControlPoint*)(iter.PeekNext()) )
    {
      if ( (pCSLIDEControlPoint->GetPoint())->GetDynamic() )
	{
	  m_listDynamics.InsertLast(pCSLIDEControlPoint, (UINT_32)pCSLIDEControlPoint);
	}
    }

  iter.Init(&m_listCSLIDEFace);

  for ( pCSLIDEFace = (CSLIDEFace*)(iter.PeekFirst());
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)(iter.PeekNext()) )
    {
      if ( pCSLIDEFace->GetDynamic() )
	{
	  m_listDynamics.InsertLast(pCSLIDEFace, (UINT_32)pCSLIDEFace);
	}
    }

  iter.Init(&m_listCSLIDETransform);

  for ( pCSLIDETransform = (CSLIDETransform*)(iter.PeekFirst());
	pCSLIDETransform != NULL;
	pCSLIDETransform = (CSLIDETransform*)(iter.PeekNext()) )
    {
      if ( pCSLIDETransform->GetDynamic() )
	{
	  m_listDynamics.InsertLast(pCSLIDETransform, (UINT_32)pCSLIDETransform);
	}
    }

  iter.Init(&m_listCSLIDEInstance);

  for ( pCSLIDEInstance = (CSLIDEInstance*)(iter.PeekFirst());
	pCSLIDEInstance != NULL;
	pCSLIDEInstance = (CSLIDEInstance*)(iter.PeekNext()) )
    {
      if ( pCSLIDEInstance->GetDynamic() || 
	   pCSLIDEInstance->GetTransformDynamic() || 
	   pCSLIDEInstance->GetBoundDynamic() )
	{
	  m_listDynamics.InsertLast(pCSLIDEInstance, (UINT_32)pCSLIDEInstance);
	}
    }

  iter.Init(&m_listCSLIDENode);

  for ( pCSLIDENode = (CSLIDENode*)(iter.PeekFirst());
	pCSLIDENode != NULL;
	pCSLIDENode = (CSLIDENode*)(iter.PeekNext()) )
    {
      if ( pCSLIDENode->GetDynamic() ||
	   pCSLIDENode->GetBoundDynamic() )
	{
	  m_listDynamics.InsertLast(pCSLIDENode, (UINT_32)pCSLIDENode);
	}
    }

  iter.Init(&m_listCSLIDEViewport);

  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      if ( pCSLIDEViewport->GetDynamic() )
	{
	  m_listDynamics.InsertLast(pCSLIDEViewport, (UINT_32)pCSLIDEViewport);
	}
    }

  iter.Init(&m_listCSLIDERender);

  for ( pCSLIDERender = (CSLIDERender*)(iter.PeekFirst());
	pCSLIDERender != NULL;
	pCSLIDERender = (CSLIDERender*)(iter.PeekNext()) )
    {
      if ( pCSLIDERender->GetDynamic() )
	{
	  m_listDynamics.InsertLast(pCSLIDERender, (UINT_32)pCSLIDERender);
	}
    }

/*

  // Print out which objects were dynamic
  printf("\n********* DYNAMIC OBJECTS **********\n");
  iter.Init(&m_listDynamics);
  CSLIDEPrimitive *pCSLIDEPrimitive;

  for ( pCSLIDEPrimitive = (CSLIDEPrimitive*)(iter.PeekFirst());
	pCSLIDEPrimitive != NULL;
	pCSLIDEPrimitive = (CSLIDEPrimitive*)(iter.PeekNext()) )
    {
      printf("%s: %ld: %ld: \"%s\"\n", pCSLIDEPrimitive->GetFileName(),
	     pCSLIDEPrimitive->GetFileLine(), 
	     pCSLIDEPrimitive->GetFileColumn(),
	     ( pCSLIDEPrimitive->GetID() == NULL ) ? ("No ID") :
	     (pCSLIDEPrimitive->GetID()));
    }
  printf("********* END: DYNAMIC OBJECTS **********\n");
*/

}

VOID CSLIDEWorld::Update()
{
  CListIterInt iter;
  CSLIDEPrimitive *pCSLIDEPrimitive;
  CWindow *pCWindow;

  UpdateTclBlocks();

  // Only call Update on the objects with some Dynamic aspect
  iter.Init(&m_listDynamics);

  for ( pCSLIDEPrimitive = (CSLIDEPrimitive*)(iter.PeekFirst());
	pCSLIDEPrimitive != NULL;
	pCSLIDEPrimitive = (CSLIDEPrimitive*)(iter.PeekNext()) )
    {
      pCSLIDEPrimitive->Update();
    }

  UpdateGraphs();

  // Update the windows after all other updates have happened
  // Updating a window == scheduling a Redisplay Event
  iter.Init(&m_listCWindow);

  for ( pCWindow = (CWindow*)(iter.PeekFirst());
	pCWindow != NULL;
	pCWindow = (CWindow*)(iter.PeekNext()) )
    {
      pCWindow->Update();
    }
}

VOID CSLIDEWorld::UpdateTclBlocks()
{
  CListIterInt iter, iterObj;
  CHAR *pcTclBlock;
  Tcl_Obj *pto;
  INT_32 iCode;

  iter.Init(&m_listCSLIDEGlobalUpdateTclBlock);
  iterObj.Init(&m_listCSLIDEGlobalUpdateTclObj);

  for ( pcTclBlock = (CHAR*)(iter.PeekFirst()), pto = (Tcl_Obj*)(iterObj.PeekFirst());
	pcTclBlock != NULL;
	pcTclBlock = (CHAR*)(iter.PeekNext()), pto = (Tcl_Obj*)(iterObj.PeekNext()) )
    {
      ASSERT( pto != NULL );

      iCode = Tcl_GlobalEvalObj(GetInterp(), pto);
      if ( iCode != TCL_OK )
	{
	  fprintf(CSLIDEPrimitive::GetErrorFile(), 
		  "Error: Tcl returned error code \"%s\" for the block:\n%s",
		  Tcl_GetStringResult(GetInterp()), pcTclBlock);
	  exit(1);
	}
    }
}

VOID CSLIDEWorld::UpdateGraphs()
{
  CListIterInt iter;
  CSLIDERender *pCSLIDERender;

  // Run through the scene graphs of the render statements
  // and update bounding boxes
  iter.Init(&m_listCSLIDERender);

  for ( pCSLIDERender = (CSLIDERender*)(iter.PeekFirst());
	pCSLIDERender != NULL;
	pCSLIDERender = (CSLIDERender*)(iter.PeekNext()) )
    {
      pCSLIDERender->UpdateGraph();
    }
}

// This needs to be called for each OpenGL Drawing Context
//
VOID CSLIDEWorld::Prerender()
{
  CListIterInt iter;
  CSLIDESurface *pCSLIDESurface;
  static UINT_32 uiContexts = 0;

  //ASSERT( uiContexts == 0 );

  uiContexts++;

  iter.Init(&m_listCSLIDESurface);

  for ( pCSLIDESurface = (CSLIDESurface*)(iter.PeekFirst());
	pCSLIDESurface != NULL;
	pCSLIDESurface = (CSLIDESurface*)(iter.PeekNext()) )
    {
      pCSLIDESurface->Prerender();
    }
}

VOID CSLIDEWorld::RestartDynamics()
{
  // Remove the Timer Event from the Event Queue in case 
  // dynamics are currently running
  UnscheduleDynamics();

  // Initialize the system time to zero
  m_tvStart.RecordTime();
  m_dTimeBase = 0.0;
  m_uiFrame = 0;

  // Update the data structures at time zero, this will also
  // restart the timer events if we are not paused
  HandleDynamics();
}  

VOID CSLIDEWorld::PauseDynamics()
{
  m_bDynamicsPaused = !m_bDynamicsPaused;
  if ( m_bDynamicsPaused )
    {
      TimeValue tvEnd, tvDelta;

      // Pause the Dynamics
      UnscheduleDynamics();

      // Update the Base amount of time
      tvEnd.RecordTime();
      tvDelta = tvEnd - m_tvStart;
      m_dTimeBase += tvDelta.GetSeconds();
    }
  else
    {
      // Unpause the Dynamics
      
      // Record the new start time
      m_tvStart.RecordTime();

      // Schedule an Update for as soon as possible
      ScheduleDynamics(0);
    }
}

VOID CSLIDEWorld::ScheduleDynamics(INT_32 iSleepTime)
{
  //ASSERT( m_ttkWakeUpIDDynamics == 0 );
  if ( m_ttkWakeUpIDDynamics == 0 )
    {
      /*
      m_ttkWakeUpIDDynamics = Tk_CreateTimerHandler((int)iSleepTime, 
			        CSLIDEWorld::CBHandleDynamics, 
			        (ClientData)this);
				*/
      m_ttkWakeUpIDDynamics = Tk_CreateTimerHandler((int)iSleepTime, 
			        CSLIDEWorld::CBHandleIdle, 
			        (ClientData)this);
    }
}

VOID CSLIDEWorld::UnscheduleDynamics()
{
  if ( m_ttkWakeUpIDDynamics != 0 )
    {
      Tk_DeleteTimerHandler(m_ttkWakeUpIDDynamics);
    }
  m_ttkWakeUpIDDynamics = 0;
}

VOID CSLIDEWorld::HandleDynamics()
{
  TimeValue tvStart, tvEnd, tvDelta;
  INT_32 iSleepTime;

  // The Timer Event has happened so the token is no longer valid
  m_ttkWakeUpIDDynamics = 0;

  tvStart.RecordTime();
  tvDelta = tvStart - m_tvStart;
  m_dUpdateTime = m_dTimeBase + tvDelta.GetSeconds();

  if ( !m_bDynamicsPaused )
    {
      // Update the value of the TCL SLF_TIME variable
      sprintf(gpcBuffer, "%f", m_dUpdateTime);
      Tcl_SetVar(m_pInterp, "SLF_TIME", gpcBuffer, TCL_GLOBAL_ONLY);
      m_uiFrame++;
      sprintf(gpcBuffer, "%ld", m_uiFrame);
      Tcl_SetVar(m_pInterp, "SLF_FRAME", gpcBuffer, TCL_GLOBAL_ONLY);
    }

  //printf("SLIDEWorld: Update\n");

  Update();

  tvEnd.RecordTime();
  tvDelta = tvEnd - tvStart;
  m_dUpdateTime = tvDelta.GetSeconds();

  iSleepTime = (INT_32)((m_dSecondsPerFrameRequested - m_dUpdateTime) * 1000.0);
  if ( iSleepTime < 0 )
    {
      iSleepTime = 0;
    }

  if ( !m_bDynamicsPaused )
    {
      //printf("Dynamics going to sleep for %ld ms.\n", iSleepTime);
      ScheduleDynamics(iSleepTime);
    }
}

VOID CSLIDEWorld::CBHandleDynamics(ClientData pvClientData)
{
  CSLIDEWorld *pCSLIDEWorld;

  ASSERT( pvClientData != NULL );

  // We only want the Idle function to get run once
  //Tk_CancelIdleCall(CSLIDEWorld::CBHandleDynamics, (ClientData)pvClientData);

  pCSLIDEWorld = (CSLIDEWorld*)pvClientData;

  pCSLIDEWorld->HandleDynamics();
}

// I am trying this indirection.  The timer call back calls this function
// which then calls the HandleDynamics in an Idle callback.  This way when
// the system slows down, the Updates and Redisplay events will serialize.
VOID CSLIDEWorld::CBHandleIdle(ClientData pvClientData)
{
  ASSERT( pvClientData != NULL );

  Tk_DoWhenIdle(CSLIDEWorld::CBHandleDynamics, (ClientData)pvClientData);
}

VOID CSLIDEWorld::HandleWriteSLIDE(CHAR *pcOutputFile, FILE *fpOutputFile)
{
  ASSERT( pcOutputFile != NULL );
  ASSERT( fpOutputFile != NULL );

  WriteSLIDE(fpOutputFile);
}

VOID CSLIDEWorld::HandleWriteDOT(CHAR *pcOutputFile, FILE *fpOutputFile)
{
  ASSERT( pcOutputFile != NULL );
  ASSERT( fpOutputFile != NULL );

  WriteDOT(fpOutputFile);
}

VOID CSLIDEWorld::HandleWriteRIB(CHAR *pcOutputFile, FILE *fpOutputFile)
{
  ASSERT( pcOutputFile != NULL );
  ASSERT( fpOutputFile != NULL );

  UINT_32 uiWindows;
  CListIterInt iter;
  CWindow *pCWindow;

  uiWindows = m_listCWindow.GetLength();

  if( uiWindows == 1 )
    {
      iter.Init(&m_listCWindow);
      pCWindow = (CWindow*)(iter.PeekFirst());
    }
  else
    {
      printf( "CSLIDEWorld::HandleWriteRIB - need to write tcl to ask user which window");
      ASSERT(FALSE);
      
      /*
      iter.Init(&m_listCWindow);
      for ( pCWindow = (CWindow*)(iter.PeekFirst() );
	    pCWindow != NULL;
	    pCWindow = (CWindow*)(iter.PeekNext()) )
	{
	  ASSERT( pCWindow != NULL );

	  if ( )
	    {
	      break;
	    }
	}
	*/
    }
  pCWindow->WriteRIB(pcOutputFile, fpOutputFile);
}

VOID CSLIDEWorld::HandleWritePOV(CHAR *pcOutputFile, FILE *fpOutputFile)
{
  ASSERT( pcOutputFile != NULL );
  ASSERT( fpOutputFile != NULL );

  UINT_32 uiWindows;
  CListIterInt iter;
  CWindow *pCWindow;
  CListInt *pListTransformIDs;
  CListInt *pListColors;

  uiWindows = m_listCWindow.GetLength();
  pListTransformIDs = new CListInt(); 
  pListTransformIDs->Init();
  pListColors = new CListInt();
  pListColors->Init();

  if( uiWindows == 1 )
    {
      iter.Init(&m_listCWindow);
      pCWindow = (CWindow*)(iter.PeekFirst());
    }
  else
    {
      printf( "CSLIDEWorld::HandleWritePOV - need to write tcl to ask user which window");
      ASSERT(FALSE);
      
      /*
      iter.Init(&m_listCWindow);
      for ( pCWindow = (CWindow*)(iter.PeekFirst() );
	    pCWindow != NULL;
	    pCWindow = (CWindow*)(iter.PeekNext()) )
	{
	  ASSERT( pCWindow != NULL );

	  if ( )
	    {
	      break;
	    }
	}
	*/
    }

  fprintf(fpOutputFile, "global_settings { assumed_gamma 2.2 } \n\n");
  fprintf(fpOutputFile, "#include \"colors.inc\"\n");
  fprintf(fpOutputFile, "#include \"textures.inc\"\n\n");
  fprintf(fpOutputFile, "background { color White } \n\n");

  pCWindow->WritePOV(pcOutputFile, fpOutputFile, pListTransformIDs, pListColors);
}

VOID CSLIDEWorld::HandleWriteWRLZ(CHAR *pcOutputFile, FILE *fpOutputFile)
{
  ASSERT( pcOutputFile != NULL );
  ASSERT( fpOutputFile != NULL );

  UINT_32 uiWindows;
  CListIterInt iter;
  CWindow *pCWindow;

  uiWindows = m_listCWindow.GetLength();

  if( uiWindows == 1 )
    {
      iter.Init(&m_listCWindow);
      pCWindow = (CWindow*)(iter.PeekFirst());
    }
  else
    {
      printf( "CSLIDEWorld::HandleWriteWRLZ - need to write tcl to ask user which window");
      ASSERT(FALSE);
      
      /*
      iter.Init(&m_listCWindow);
      for ( pCWindow = (CWindow*)(iter.PeekFirst() );
	    pCWindow != NULL;
	    pCWindow = (CWindow*)(iter.PeekNext()) )
	{
	  ASSERT( pCWindow != NULL );

	  if ( )
	    {
	      break;
	    }
	}
	*/
    }

  pCWindow->WriteWRLZ(pcOutputFile, fpOutputFile);
}

VOID CSLIDEWorld::HandleWriteSTL(CHAR *pcOutputFile, FILE *fpOutputFile)
{
  ASSERT( pcOutputFile != NULL );
  ASSERT( fpOutputFile != NULL );
  
  UINT_32 uiWindows;
  CListIterInt iter;
  CWindow *pCWindow;

  uiWindows = m_listCWindow.GetLength();

  if( uiWindows == 1 )
    {
      iter.Init(&m_listCWindow);
      pCWindow = (CWindow*)(iter.PeekFirst());
    }
  else
    {
      printf( "CSLIDEWorld::HandleWriteSTL - need to write tcl to ask user which window");
      ASSERT(FALSE);

      /*
      iter.Init(&m_listCWindow);
      for ( pCWindow = (CWindow*)(iter.PeekFirst() );
	    pCWindow != NULL;
	    pCWindow = (CWindow*)(iter.PeekNext()) )
	{
	  ASSERT( pCWindow != NULL );

	  if ( )
	    {
	      break;
	    }
	}
	*/
    }
  fprintf(fpOutputFile, "solid\n" );
  pCWindow->WriteSTL(pcOutputFile, fpOutputFile);
  fprintf(fpOutputFile, "endsolid\n" );
}

VOID CSLIDEWorld::HandleWriteSIF(CHAR *pcOutputFile, FILE *fpOutputFile)
{
  ASSERT( pcOutputFile != NULL );
  ASSERT( fpOutputFile != NULL );
  
  UINT_32 uiWindows;
  CListIterInt iter;
  CWindow *pCWindow;

  uiWindows = m_listCWindow.GetLength();

  if( uiWindows == 1 )
    {
      iter.Init(&m_listCWindow);
      pCWindow = (CWindow*)(iter.PeekFirst());
    }
  else
    {
      printf( "CSLIDEWorld::HandleWriteSIF - need to write tcl to ask user which window");
      ASSERT(FALSE);

      /*
      iter.Init(&m_listCWindow);
      for ( pCWindow = (CWindow*)(iter.PeekFirst() );
	    pCWindow != NULL;
	    pCWindow = (CWindow*)(iter.PeekNext()) )
	{
	  ASSERT( pCWindow != NULL );

	  if ( )
	    {
	      break;
	    }
	}
	*/
    }

  pCWindow->WriteSIF(pcOutputFile, fpOutputFile);
}

VOID CSLIDEWorld::HandleWriteLSIF(CHAR *pcOutputFile, FILE *fpOutputFile)
{
  ASSERT( pcOutputFile != NULL );
  ASSERT( fpOutputFile != NULL );
  
  UINT_32 uiWindows;
  CListIterInt iter;
  CWindow *pCWindow;

  uiWindows = m_listCWindow.GetLength();

  if( uiWindows == 1 )
    {
      iter.Init(&m_listCWindow);
      pCWindow = (CWindow*)(iter.PeekFirst());
    }
  else
    {
      printf( "CSLIDEWorld::HandleWriteSIF - need to write tcl to ask user which window");
      ASSERT(FALSE);

      /*
      iter.Init(&m_listCWindow);
      for ( pCWindow = (CWindow*)(iter.PeekFirst() );
	    pCWindow != NULL;
	    pCWindow = (CWindow*)(iter.PeekNext()) )
	{
	  ASSERT( pCWindow != NULL );

	  if ( )
	    {
	      break;
	    }
	}
	*/
    }

  pCWindow->WriteLSIF(pcOutputFile, fpOutputFile);
}

//////////////////////////////////////////////////////////////////////////////////////////

VOID CSLIDEWorld::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDESurface *pCSLIDESurface;
  CSLIDEFog *pCSLIDEFog;
  CSLIDEPoint *pCSLIDEPoint;
  CSLIDEControlPoint *pCSLIDEControlPoint;
  CSLIDEFace *pCSLIDEFace;
  CSLIDENode *pCSLIDENode;
  CWindow *pCWindow;
  CSLIDEViewport *pCSLIDEViewport;
  CSLIDERender *pCSLIDERender;

  ASSERT( fpOutputFile != NULL );

  iter.Init(&m_listCSLIDESurface);

  // Skip the Default Surface
  pCSLIDESurface = (CSLIDESurface*)(iter.PeekFirst());
  ASSERT( pCSLIDESurface == CSLIDESurface::m_pCSLIDESurfaceDefault );
  for ( pCSLIDESurface = (CSLIDESurface*)(iter.PeekNext());
	pCSLIDESurface != NULL;
	pCSLIDESurface = (CSLIDESurface*)(iter.PeekNext()) )
    {
      pCSLIDESurface->WriteSLIDE(fpOutputFile);
    }

  iter.Init(&m_listCSLIDEFog);

  for ( pCSLIDEFog = (CSLIDEFog*)(iter.PeekNext());
	pCSLIDEFog != NULL;
	pCSLIDEFog = (CSLIDEFog*)(iter.PeekNext()) )
    {
      pCSLIDEFog->WriteSLIDE(fpOutputFile);
    }

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekFirst());
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekNext()) )
    {
      pCSLIDEPoint->WriteSLIDE(fpOutputFile);
    }

  iter.Init(&m_listCSLIDEControlPoint);

  for ( pCSLIDEControlPoint = (CSLIDEControlPoint*)(iter.PeekFirst());
	pCSLIDEControlPoint != NULL;
	pCSLIDEControlPoint = (CSLIDEControlPoint*)(iter.PeekNext()) )
    {
      pCSLIDEControlPoint->WriteSLIDE(fpOutputFile);
    }

  iter.Init(&m_listCSLIDEFace);

  for ( pCSLIDEFace = (CSLIDEFace*)(iter.PeekFirst());
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)(iter.PeekNext()) )
    {
      pCSLIDEFace->WriteSLIDE(fpOutputFile);
    }

  iter.Init(&m_listCSLIDENode);

  for ( pCSLIDENode = (CSLIDENode*)(iter.PeekFirst());
	pCSLIDENode != NULL;
	pCSLIDENode = (CSLIDENode*)(iter.PeekNext()) )
    {
      pCSLIDENode->WriteSLIDE(fpOutputFile);
    }

  iter.Init(&m_listCWindow);

  for ( pCWindow = (CWindow*)(iter.PeekFirst());
	pCWindow != NULL;
	pCWindow = (CWindow*)(iter.PeekNext()) )
    {
      pCWindow->WriteSLIDE(fpOutputFile);
    }

  iter.Init(&m_listCSLIDEViewport);

  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      pCSLIDEViewport->WriteSLIDE(fpOutputFile);
    }

  iter.Init(&m_listCSLIDERender);

  for ( pCSLIDERender = (CSLIDERender*)(iter.PeekFirst());
	pCSLIDERender != NULL;
	pCSLIDERender = (CSLIDERender*)(iter.PeekNext()) )
    {
      pCSLIDERender->WriteSLIDE(fpOutputFile);
    }
}

VOID CSLIDEWorld::WriteDOT(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  CListIterInt iter;
  CSLIDESurface *pCSLIDESurface;
  CSLIDEFog *pCSLIDEFog;
  CSLIDEPoint *pCSLIDEPoint;
  CSLIDEControlPoint *pCSLIDEControlPoint;
  CSLIDEFace *pCSLIDEFace;
  CSLIDENode *pCSLIDENode;
  CWindow *pCWindow;
  CSLIDEViewport *pCSLIDEViewport;
  CSLIDERender *pCSLIDERender;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "digraph slide\n");
  fprintf(fpOutputFile, "{\n");
  fprintf(fpOutputFile, "  // Default Node and Edge Attributes\n");
  fprintf(fpOutputFile, "  node\n");
  fprintf(fpOutputFile, "  [\n");
  fprintf(fpOutputFile, "    fontsize = 20\n");
  fprintf(fpOutputFile, "    fontname = \"Times-Roman\"\n");
  fprintf(fpOutputFile, "    fontcolor = black\n");
  fprintf(fpOutputFile, "    shape = ellipse\n");
  fprintf(fpOutputFile, "    color = black\n");
  fprintf(fpOutputFile, "    width = 0.5\n");
  fprintf(fpOutputFile, "  ];\n");
  fprintf(fpOutputFile, "\n");
  fprintf(fpOutputFile, "  edge\n"); 
  fprintf(fpOutputFile, "  [\n");
  fprintf(fpOutputFile, "    fontsize = 20\n");
  fprintf(fpOutputFile, "    fontname = \"Times-Roman\"\n");
  fprintf(fpOutputFile, "    fontcolor = black\n");
  //  fprintf(fpOutputFile, "    dir = back\n");
  fprintf(fpOutputFile, "    color = black\n");
  fprintf(fpOutputFile, "    style = bold\n");
  fprintf(fpOutputFile, "  ];\n");
  fprintf(fpOutputFile, "\n");
  fprintf(fpOutputFile, "ranksep = 0.75\n");
  fprintf(fpOutputFile, "\n");

  //////////////////////////////////////////////////////////////////////
  // Write the Nodes of the Graph
  // 

  fprintf(fpOutputFile, "{ rank = same;\n");

  iter.Init(&m_listCWindow);

  for ( pCWindow = (CWindow*)(iter.PeekFirst());
	pCWindow != NULL;
	pCWindow = (CWindow*)(iter.PeekNext()) )
    {
      pCWindow->WriteDotNode(fpOutputFile);
    }

  fprintf(fpOutputFile, "}\n");
  fprintf(fpOutputFile, "\n");
  fprintf(fpOutputFile, "{ rank = same;\n");

  iter.Init(&m_listCSLIDEViewport);

  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      pCSLIDEViewport->WriteDotNode(fpOutputFile);
    }

  fprintf(fpOutputFile, "}\n");
  fprintf(fpOutputFile, "\n");
  fprintf(fpOutputFile, "{ rank = same;\n");

  iter.Init(&m_listCSLIDERender);

  for ( pCSLIDERender = (CSLIDERender*)(iter.PeekFirst());
	pCSLIDERender != NULL;
	pCSLIDERender = (CSLIDERender*)(iter.PeekNext()) )
    {
      pCSLIDERender->WriteDotNode(fpOutputFile);
    }

  fprintf(fpOutputFile, "}\n");
  fprintf(fpOutputFile, "\n");

  fprintf(fpOutputFile, "subgraph nodes {\n");
//  fprintf(fpOutputFile, "  style = invis;\n");
  fprintf(fpOutputFile, "\n");

  iter.Init(&m_listCSLIDENode);

  for ( pCSLIDENode = (CSLIDENode*)(iter.PeekFirst());
	pCSLIDENode != NULL;
	pCSLIDENode = (CSLIDENode*)(iter.PeekNext()) )
    {
      pCSLIDENode->WriteDotNode(fpOutputFile);
    }

  fprintf(fpOutputFile, "\n");

  iter.Init(&m_listCSLIDEFace);

  for ( pCSLIDEFace = (CSLIDEFace*)(iter.PeekFirst());
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)(iter.PeekNext()) )
    {
      pCSLIDEFace->WriteDotNode(fpOutputFile);
    }

  fprintf(fpOutputFile, "\n");

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekFirst());
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekNext()) )
    {
      pCSLIDEPoint->WriteDotNode(fpOutputFile);
    }

  fprintf(fpOutputFile, "\n");

  iter.Init(&m_listCSLIDEControlPoint);

  for ( pCSLIDEControlPoint = (CSLIDEControlPoint*)(iter.PeekFirst());
	pCSLIDEControlPoint != NULL;
	pCSLIDEControlPoint = (CSLIDEControlPoint*)(iter.PeekNext()) )
    {
      pCSLIDEControlPoint->WriteDotNode(fpOutputFile);
    }

  fprintf(fpOutputFile, "\n");

  iter.Init(&m_listCSLIDESurface);

  // Skip the Default Surface
  pCSLIDESurface = (CSLIDESurface*)(iter.PeekFirst());
  ASSERT( pCSLIDESurface == CSLIDESurface::m_pCSLIDESurfaceDefault );
  for ( pCSLIDESurface = (CSLIDESurface*)(iter.PeekNext());
	pCSLIDESurface != NULL;
	pCSLIDESurface = (CSLIDESurface*)(iter.PeekNext()) )
    {
      pCSLIDESurface->WriteDotNode(fpOutputFile);
    }

  fprintf(fpOutputFile, "\n");

  iter.Init(&m_listCSLIDEFog);

  for ( pCSLIDEFog = (CSLIDEFog*)(iter.PeekNext());
	pCSLIDEFog != NULL;
	pCSLIDEFog = (CSLIDEFog*)(iter.PeekNext()) )
    {
      pCSLIDEFog->WriteDotNode(fpOutputFile);
    }

  fprintf(fpOutputFile, "}\n");
  fprintf(fpOutputFile, "\n");

  //////////////////////////////////////////////////////////////////////
  // Write the Edges of the Graph
  // 

  iter.Init(&m_listCWindow);

  for ( pCWindow = (CWindow*)(iter.PeekFirst());
	pCWindow != NULL;
	pCWindow = (CWindow*)(iter.PeekNext()) )
    {
      pCWindow->WriteDotEdges(fpOutputFile);
    }

  fprintf(fpOutputFile, "\n");

  iter.Init(&m_listCSLIDEViewport);

  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      pCSLIDEViewport->WriteDotEdges(fpOutputFile);
    }

  fprintf(fpOutputFile, "\n");

  iter.Init(&m_listCSLIDERender);

  for ( pCSLIDERender = (CSLIDERender*)(iter.PeekFirst());
	pCSLIDERender != NULL;
	pCSLIDERender = (CSLIDERender*)(iter.PeekNext()) )
    {
      pCSLIDERender->WriteDotEdges(fpOutputFile);
    }

  fprintf(fpOutputFile, "\n");

  iter.Init(&m_listCSLIDENode);

  for ( pCSLIDENode = (CSLIDENode*)(iter.PeekFirst());
	pCSLIDENode != NULL;
	pCSLIDENode = (CSLIDENode*)(iter.PeekNext()) )
    {
      pCSLIDENode->WriteDotEdges(fpOutputFile);
    }

  fprintf(fpOutputFile, "\n");

  iter.Init(&m_listCSLIDEFace);

  for ( pCSLIDEFace = (CSLIDEFace*)(iter.PeekFirst());
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)(iter.PeekNext()) )
    {
      pCSLIDEFace->WriteDotEdges(fpOutputFile);
    }

  fprintf(fpOutputFile, "\n");

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekFirst());
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekNext()) )
    {
      pCSLIDEPoint->WriteDotEdges(fpOutputFile);
    }

  fprintf(fpOutputFile, "\n");

  iter.Init(&m_listCSLIDEControlPoint);

  for ( pCSLIDEControlPoint = (CSLIDEControlPoint*)(iter.PeekFirst());
	pCSLIDEControlPoint != NULL;
	pCSLIDEControlPoint = (CSLIDEControlPoint*)(iter.PeekNext()) )
    {
      pCSLIDEControlPoint->WriteDotEdges(fpOutputFile);
    }

  fprintf(fpOutputFile, "\n");

  iter.Init(&m_listCSLIDESurface);

  // Skip the Default Surface
  pCSLIDESurface = (CSLIDESurface*)(iter.PeekFirst());
  ASSERT( pCSLIDESurface == CSLIDESurface::m_pCSLIDESurfaceDefault );
  for ( pCSLIDESurface = (CSLIDESurface*)(iter.PeekNext());
	pCSLIDESurface != NULL;
	pCSLIDESurface = (CSLIDESurface*)(iter.PeekNext()) )
    {
      pCSLIDESurface->WriteDotEdges(fpOutputFile);
    }

  fprintf(fpOutputFile, "\n");

  iter.Init(&m_listCSLIDEFog);

  for ( pCSLIDEFog = (CSLIDEFog*)(iter.PeekNext());
	pCSLIDEFog != NULL;
	pCSLIDEFog = (CSLIDEFog*)(iter.PeekNext()) )
    {
      pCSLIDEFog->WriteDotEdges(fpOutputFile);
    }

  fprintf(fpOutputFile, "}\n");
}

VOID CSLIDEWorld::WriteRIB(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDESurface *pCSLIDESurface;

  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, "");
  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE CREATE TEXTURE MAPS");

  iter.Init(&m_listCSLIDESurface);

  // Skip the Default Surface
  pCSLIDESurface = (CSLIDESurface*)(iter.PeekFirst());
  ASSERT( pCSLIDESurface == CSLIDESurface::m_pCSLIDESurfaceDefault );
  for ( pCSLIDESurface = (CSLIDESurface*)(iter.PeekNext());
	pCSLIDESurface != NULL;
	pCSLIDESurface = (CSLIDESurface*)(iter.PeekNext()) )
    {
      pCSLIDESurface->WriteRIBCreateTexture(fpOutputFile);
    }

  RiArchiveRecord(RI_COMMENT, "   END SLIDE CREATE TEXTURE MAPS");
}

VOID CSLIDEWorld::WritePOV(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDEWorld::WriteWRLZ(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDEWorld::WriteSTL(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDEWorld::WriteSIF(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );
}

UINT_32 CSLIDEWorld::GetErrorCount()
{
  return CSLIDEPrimitive::GetErrorCount();
}

UINT_32 CSLIDEWorld::GetWarningCount()
{
  return CSLIDEPrimitive::GetWarningCount();
}

VOID CSLIDEWorld::Unselect()
{
  CListIterInt iter;
  CSLIDERender *pCSLIDERender;

  iter.Init(&m_listCSLIDERender);
  pCSLIDERender = (CSLIDERender*)(iter.PeekFirst());
  pCSLIDERender->SetNodeGUISelected(NULL);
}

VOID CSLIDEWorld::GetParameters(CHAR *pcID, SLIDEReservedWordType uiType)
{
  CListIterInt iter;
  CSLIDERender *pCSLIDERender;
  CSLIDENode *pCSLIDENode;
  CSLIDEInstance *pCSLIDEInstance;

  switch(uiType) 
    {
    case sliderwtFace:
    case sliderwtFog:
    case sliderwtPoint:
    case sliderwtControlPoint:
    case sliderwtSurface:
    case sliderwtViewport:
    case sliderwtWindow:
      printf( "CSLIDEWorld::GetParameters not implemented\n" );
      ASSERT(FALSE);
      break;
    case sliderwtInstance:
      pCSLIDEInstance = PeekInstance(pcID);
      ASSERT(pCSLIDEInstance != NULL);
      pCSLIDEInstance->GetParameters(m_pInterp);
      pCSLIDENode = pCSLIDEInstance;
      break;
    default:
      pCSLIDENode = PeekNode(pcID);
      ASSERT(pCSLIDENode != NULL);
      pCSLIDENode->GetParameters(m_pInterp);
      break;
    }

  // render selected object with bbox
  iter.Init(&m_listCSLIDERender);
  pCSLIDERender = (CSLIDERender*)(iter.PeekFirst());
  pCSLIDERender->SetNodeGUISelected(pCSLIDENode);
}

int CSLIDEWorld::SetParameters(CScanner *pCScanner, CHAR *pcID, 
			       SLIDEReservedWordType uiType, BOOL bPrerender,
			       int argc, CHAR **argv)
{
  UINT_32 i, uiParameter;
  CSLIDEPrimitive *pCSLIDEPrimitive;
  CSLIDENode *pCSLIDENode;
  CSLIDEInstance *pCSLIDEInstance;
  CHAR *pcParameter;

  switch(uiType) 
    {
    case sliderwtFace:
    case sliderwtPoint:
    case sliderwtControlPoint:
    case sliderwtFog:
    case sliderwtSurface:
    case sliderwtViewport:
    case sliderwtWindow:
      printf( "CSLIDEWorld::SetParameter not implemented\n" );
      ASSERT(FALSE);
      break;
    case sliderwtInstance:
      pCSLIDEInstance = PeekInstance(pcID);
      ASSERT(pCSLIDEInstance != NULL);
      pCSLIDEPrimitive = pCSLIDEInstance;
      break;
    default:
      pCSLIDENode = PeekNode(pcID);
      ASSERT(pCSLIDENode != NULL);
      pCSLIDEPrimitive = pCSLIDENode;
      break;
    }

  // Set parameters
  for( i = 0; i < (UINT_32)argc; ) 
    {
      ASSERT( argv[0] != NULL );

      pcParameter = NULL;
      pcParameter = pCScanner->FindIdentifier(argv[i], FALSE);
      i++;
      ASSERT( pcParameter );

      // Get the enum SLIDE type
      for( uiParameter=0; uiParameter < sliderwtSize; uiParameter++ )
	{
	  if( !strcmp(pcParameter, gcppcSLIDEReservedWords[uiParameter]) )
	    {
	      break;
	    }
	}

      // Make sure these enum's are valid
      if( uiParameter == sliderwtSize )
	{
	  return TCL_ERROR;
	}

      // If this is a surface parameter then get the proper ID
      if( uiParameter == sliderwtSurface )
	{
	  CHAR *pcSurfaceID;
	  pcSurfaceID = NULL;

	  if( !strcmp(argv[i], "inherit") )
	    {
	      pcSurfaceID = (CHAR*)(gcppcSLIDEReservedWords[sliderwtSLF_INHERIT]);
	    }
	  else
	    {
	      pcSurfaceID = pCScanner->FindIdentifier(argv[i], FALSE);
	    }
	  ASSERT(pcSurfaceID != NULL);
	  argv[i] = pcSurfaceID;
	}

      i += pCSLIDEPrimitive->SetParameter(this, (SLIDEReservedWordType)uiParameter,
					  &(argv[i]));
    }

  if( bPrerender == 1 )
    {
      Reprerender(pCSLIDEPrimitive);
    }

  return TCL_OK;
}

VOID CSLIDEWorld::Reprerender(CSLIDEPrimitive *pCSLIDEPrimitive)
{
  // FOR NOW pCSLIDEPrimitive should be an instance, group, or atomnode
  //  but in the future want to select points and faces

  // Mark the path up to root or dynamic node as unvisited
  pCSLIDEPrimitive->SetBoundUnvisited();
  UpdateGraphs();                        // Recalculate bboxes; sets BoundChanged=TRUE
  pCSLIDEPrimitive->SetBoundUnchanged(); // unset guiChangedTransforms & BoundChange

  pCSLIDEPrimitive->SetPrerenderUnvisited();   // Mark path as prerender unvisited 
  PrerenderWindows();                          // Re Prerender
}

VOID CSLIDEWorld::PrerenderWindows()
{
  CListIterInt iter;
  CWindow *pCWindow;

  iter.Init(&m_listCWindow);  
  for ( pCWindow = (CWindow*)(iter.PeekFirst());
	pCWindow != NULL;
	pCWindow = (CWindow*)(iter.PeekNext()) )
    {
      pCWindow->Prerender();
    }
}



