#ifndef __SLIDEWORLD_H__
#define __SLIDEWORLD_H__

#include "slideprimitive.h"
#include "window.h"
#include "timer.h"

class CScanner;
class CSLIDERender;
class CSLIDEViewport;
class CSLIDENode;
class CSLIDEInstance;
class CSLIDETransform;
class CSLIDEFace;
class CSLIDEPoint;
class CSLIDEControlPoint;
class CSLIDESurface;
class CSLIDEFog;

//////////////////////////////////////////////////////////////////////
// CSLIDEWorld Class
//

class CSLIDEWorld
{
//////////////////////////////////////////////////
// Friends
//
public: friend class CSLIDEParser;

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEWorld();
  virtual ~CSLIDEWorld();
  virtual VOID Init();
  virtual VOID Uninit();

  virtual VOID UninitTcl();

  virtual BOOL Initialized();

protected:
  virtual VOID InitDefaults();

public:  
  virtual BOOL InsertRender(CSLIDERender *pCSLIDERender);
  virtual BOOL InsertWindow(CWindow *pCWindow);
  virtual CWindow *PeekWindow(CHAR *pcID);
  virtual BOOL InsertViewport(CSLIDEViewport *pCSLIDEViewport);
  virtual CSLIDEViewport *PeekViewport(CHAR *pcID);
  virtual BOOL InsertNode(CSLIDENode *pCSLIDENode);
  virtual BOOL RemoveNode(CSLIDENode *pCSLIDENode);
  virtual CSLIDENode *PeekNode(CHAR *pcID);
  virtual BOOL InsertInstanceID(CSLIDEInstance *pCSLIDEInstance, CHAR *pcID);
  virtual BOOL InsertInstance(CSLIDEInstance *pCSLIDEInstance);
  virtual BOOL RemoveInstance(CSLIDEInstance *pCSLIDEInstance);
  virtual CSLIDEInstance *PeekInstance(CHAR *pcID);
  virtual BOOL InsertTransform(CSLIDETransform *pCSLIDETransform);
  virtual BOOL RemoveTransform(CSLIDETransform *pCSLIDETransform);
  virtual BOOL InsertFace(CSLIDEFace *pCSLIDEFace);
  virtual BOOL RemoveFace(CSLIDEFace *pCSLIDEFace);
  virtual CSLIDEFace *PeekFace(CHAR *pcID);
  virtual BOOL InsertPoint(CSLIDEPoint *pCSLIDEPoint);
  virtual BOOL RemovePoint(CSLIDEPoint *pCSLIDEPoint);
  virtual CSLIDEPoint *PeekPoint(CHAR *pcID);
  virtual BOOL InsertControlPoint(CSLIDEControlPoint *pCSLIDEControlPoint);
  virtual BOOL RemoveControlPoint(CSLIDEControlPoint *pCSLIDEControlPoint);
  virtual CSLIDEControlPoint *PeekControlPoint(CHAR *pcID);
  virtual BOOL InsertSurface(CSLIDESurface *pCSLIDESurface);
  virtual BOOL RemoveSurface(CSLIDESurface *pCSLIDESurface);
  virtual CSLIDESurface *PeekSurface(CHAR *pcID);
  virtual BOOL InsertFog(CSLIDEFog *pCSLIDEFog);
  virtual BOOL RemoveFog(CSLIDEFog *pCSLIDEFog);
  virtual CSLIDEFog *PeekFog(CHAR *pcID);
  virtual BOOL InsertGlobalInitTclBlock(CHAR *pcTclBlock);
  virtual BOOL InsertGlobalUpdateTclBlock(CHAR *pcTclBlock);

  virtual BOOL CheckSemantics();

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  virtual VOID Preprocess();
  virtual VOID PreprocessGUI();
  virtual VOID DrawGUIGraph();

protected:
  virtual VOID PreprocessGraphs();
  virtual VOID CreateDynamicsList();

public:
  virtual VOID Update();

protected:
  virtual VOID UpdateTclBlocks();
  virtual VOID UpdateGraphs();

//////////////////////////////////////////////////
// Rendering
//
protected:

public:
  virtual VOID Prerender();
  VOID Reprerender(CSLIDEPrimitive *pCSLIDEPrimitive);
  VOID PrerenderWindows();

  virtual VOID RestartDynamics();
  virtual VOID PauseDynamics();
  virtual VOID ScheduleDynamics(INT_32 iSleepTime);
  virtual VOID UnscheduleDynamics();
  virtual VOID HandleDynamics();
  static VOID CBHandleDynamics(ClientData pvClientData);
  static VOID CBHandleIdle(ClientData pvClientData);

//////////////////////////////////////////////////
// Output Functions
//
public:
  VOID HandleWriteSLIDE(CHAR *pcOutputFile, FILE *fpOutputFile);
  VOID HandleWriteDOT(CHAR *pcOutputFile, FILE *fpOutputFile);
  VOID HandleWriteRIB(CHAR *pcOutputFile, FILE *fpOutputFile);
  VOID HandleWritePOV(CHAR *pcOutputFile, FILE *fpOutputFile);
  VOID HandleWriteWRLZ(CHAR *pcOutputFile, FILE *fpOutputFile);
  VOID HandleWriteSTL(CHAR *pcOutputFile, FILE *fpOutputFile);
  VOID HandleWriteSIF(CHAR *pcOutputFile, FILE *fpOutputFile);
  VOID HandleWriteLSIF(CHAR *pcOutputFile, FILE *fpOutputFile);

  virtual VOID WriteSLIDE(FILE *fpOutputFile);
  virtual VOID WriteDOT(FILE *fpOutputFile);
  virtual VOID WriteRIB(FILE *fpOutputFile);
  virtual VOID WritePOV(FILE *fpOutputFile);
  virtual VOID WriteWRLZ(FILE *fpOutputFile);
  virtual VOID WriteSTL(FILE *fpOutputFile);
  virtual VOID WriteSIF(FILE *fpOutputFile);

//////////////////////////////////////////////////
// Functions that communicate with TCL
//
public:
  VOID Unselect();
  VOID GetParameters(CHAR *pcID, SLIDEReservedWordType uiType);
  int SetParameters(CScanner *pCScanner, CHAR *pcID, SLIDEReservedWordType uiType, 
		    BOOL bPrerender, int argc, CHAR **argv);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  UINT_32 GetErrorCount();
  UINT_32 GetWarningCount();

//////////////////////////////////////////////////
// Class Accessor Functions
//
public:
  static inline Tcl_Interp *GetInterp();
  static inline VOID SetInterp(Tcl_Interp *pInterp);
  static inline CSLIDEWorld *GetWorld();
  static inline VOID SetWorld(CSLIDEWorld *pCSLIDEWorld);
  static inline DOUBLE GetSecondsPerFrameRequested();
  static inline DOUBLE GetUpdateTime();
  static inline BOOL GetTclGUIExists();
  static inline VOID SetTclGUIExists(BOOL bExists);

  static inline CWindow *GetWindowEditObject();
  static inline CSLIDEViewport *GetViewportEditObject();
  static inline CSLIDERender *GetRenderEditObject();

//////////////////////////////////////////////////
// Class Variables
//
protected:
  static Tcl_Interp *m_pInterp;
  static CSLIDEWorld *m_pCSLIDEWorld;
  static BOOL m_bTclGUIExists;

  // Object Editing
  static CWindow *m_pCSLIDEWindowEditObject;
  static CSLIDEViewport *m_pCSLIDEViewportEditObject;
  static CSLIDERender *m_pCSLIDERenderEditObject;

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CListInt m_listCSLIDERender;

  CListInt m_listCWindow;
  CHashInt m_hashCWindow;
  CListInt m_listCSLIDEViewport;
  CHashInt m_hashCSLIDEViewport;
  CListInt m_listCSLIDENode;
  CHashInt m_hashCSLIDENode;
  CListInt m_listCSLIDEInstance;
  CHashInt m_hashCSLIDEInstance;
  CListInt m_listCSLIDETransform;
  CListInt m_listCSLIDEFace;
  CHashInt m_hashCSLIDEFace;
  CListInt m_listCSLIDEPoint;
  CHashInt m_hashCSLIDEPoint;
  CListInt m_listCSLIDEControlPoint;
  CHashInt m_hashCSLIDEControlPoint;
  CListInt m_listCSLIDESurface;
  CHashInt m_hashCSLIDESurface;
  CListInt m_listCSLIDEFog;
  CHashInt m_hashCSLIDEFog;

  // These will only be evaluated once during Preprocess
  CListInt m_listCSLIDEGlobalInitTclBlock; 
  CListInt m_listCSLIDEGlobalUpdateTclBlock; 
  CListInt m_listCSLIDEGlobalUpdateTclObj; 

  // We will only call Update on these objects
  CListInt m_listDynamics;

  // Defaults

  // Dynamics
  BOOL m_bDynamicsPaused;
  static DOUBLE m_dSecondsPerFrameRequested;
  static DOUBLE m_dUpdateTime;
  Tk_TimerToken m_ttkWakeUpIDDynamics;
  TimeValue m_tvStart;
  DOUBLE m_dTimeBase;
  UINT_32 m_uiFrame;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline Tcl_Interp *CSLIDEWorld::GetInterp()
{
  return m_pInterp;
}

inline VOID CSLIDEWorld::SetInterp(Tcl_Interp *pInterp)
{
  m_pInterp = pInterp;
}

inline CSLIDEWorld *CSLIDEWorld::GetWorld()
{
  return m_pCSLIDEWorld;
}

inline VOID CSLIDEWorld::SetWorld(CSLIDEWorld *pCSLIDEWorld)
{
  m_pCSLIDEWorld = pCSLIDEWorld;
}

inline DOUBLE CSLIDEWorld::GetSecondsPerFrameRequested()
{
  return m_dSecondsPerFrameRequested;
}

inline DOUBLE CSLIDEWorld::GetUpdateTime()
{
  return m_dUpdateTime;
}

inline BOOL CSLIDEWorld::GetTclGUIExists()
{
  return m_bTclGUIExists;
}

inline VOID CSLIDEWorld::SetTclGUIExists(BOOL bExists)
{
  m_bTclGUIExists = bExists;
}

inline CWindow* CSLIDEWorld::GetWindowEditObject()
{
  return m_pCSLIDEWindowEditObject;
}

inline CSLIDEViewport* CSLIDEWorld::GetViewportEditObject()
{
  return m_pCSLIDEViewportEditObject;
}

inline CSLIDERender* CSLIDEWorld::GetRenderEditObject()
{
  return m_pCSLIDERenderEditObject;
}

#endif // __SLIDEWORLD_H__


