#ifndef __SPLAY_H__
#define __SPLAY_H__

#include "list.h"

//////////////////////////////////////////////////////////////////////////////
// CSplayNode Class
//

template <class TPData, class TKey>
class CSplayNode
{
public:
  inline CSplayNode();
  inline ~CSplayNode();
  inline BOOL Init(TKey key, TPData pData);

protected:
  inline VOID Uninit();

public:
  // Insertion Functions
  inline VOID Insert(CSplayNode *pNode);

  // Deletion Functions
  //inline TPData Remove(TKey key);

  // Look up Functions
  //inline TPData Peek(TKey key) const;

public:
  inline VOID Display();

  // Inlined Functions
public:
  inline CSplayNode<TPData, TKey> *GetParent();
  inline VOID SetParent(CSplayNode *pParent);
  inline CSplayNode<TPData, TKey> *GetLeftChild();
  inline VOID SetLeftChild(CSplayNode *pLeftChild);
  inline CSplayNode<TPData, TKey> *GetRightChild();
  inline VOID SetRightChild(CSplayNode *pRightChild);
  inline TKey GetKey();
  inline TPData GetData(TKey *pTKey = NULL);

protected:
  CSplayNode *m_pParent;
  CSplayNode *m_pLeftChild;
  CSplayNode *m_pRightChild;

  TKey m_key;
  TPData m_pData;
};

//////////////////////
// Inlined Functions
//

template <class TPData, class TKey>
inline CSplayNode<TPData, TKey> *CSplayNode<TPData, TKey>::GetParent()
{
  return m_pParent;
}

template <class TPData, class TKey>
inline VOID CSplayNode<TPData, TKey>::SetParent(CSplayNode *pParent)
{
  m_pParent = pParent;
}

template <class TPData, class TKey>
inline CSplayNode<TPData, TKey> *CSplayNode<TPData, TKey>::GetLeftChild()
{
  return m_pLeftChild;
}

template <class TPData, class TKey>
inline VOID CSplayNode<TPData, TKey>::SetLeftChild(CSplayNode *pLeftChild)
{
  m_pLeftChild = pLeftChild;
}

template <class TPData, class TKey>
inline CSplayNode<TPData, TKey> *CSplayNode<TPData, TKey>::GetRightChild()
{
  return m_pRightChild;
}

template <class TPData, class TKey>
inline VOID CSplayNode<TPData, TKey>::SetRightChild(CSplayNode *pRightChild)
{
  m_pRightChild = pRightChild;
}

template <class TPData, class TKey>
inline TKey CSplayNode<TPData, TKey>::GetKey()
{
  return m_key;
}

template <class TPData, class TKey>
inline TPData CSplayNode<TPData, TKey>::GetData(TKey *pTKey/* = NULL*/)
{
  if ( pTKey != NULL )
    {
      *pTKey = m_key;
    }
  
  return m_pData;
}

#define CSPLAYNODEBUFFER
#ifdef CSPLAYNODEBUFFER

//////////////////////////////////////////////////////////////////////////////
// CSplayNodeBuffer Class
//

template <class TPData, class TKey>
class CSplayNodeBuffer
{
public:
  inline CSplayNodeBuffer();
  inline ~CSplayNodeBuffer();
  inline BOOL Init(UINT_32 uiSize = 2);

protected:
  inline VOID Uninit();

  // Inlined Functions
public:
  inline CSplayNodeBuffer<TPData, TKey> *GetNext();
  inline VOID SetNext(CSplayNodeBuffer *pNext);
  inline CSplayNode<TPData, TKey> *GetBuffer();

protected:
  CSplayNodeBuffer *m_pNext;

  UINT_32 m_uiCSplayNodes;
  CSplayNode<TPData, TKey> *m_pCSplayNodes;
};

//////////////////////////////////////////////////
// Inlined Functions
//

template <class TPData, class TKey>
inline CSplayNodeBuffer<TPData, TKey> *CSplayNodeBuffer<TPData, TKey>::GetNext()
{
  return m_pNext;
}

template <class TPData, class TKey>
inline VOID CSplayNodeBuffer<TPData, TKey>::SetNext(CSplayNodeBuffer *pNext)
{
  m_pNext = pNext;
}

template <class TPData, class TKey>
inline CSplayNode<TPData, TKey> *CSplayNodeBuffer<TPData, TKey>::GetBuffer()
{
  return m_pCSplayNodes;
}

#endif // CSPLAYNODEBUFFER

//////////////////////////////////////////////////////////////////////////////
// CSplay Class
//

template <class TPData, class TKey>
class CSplay
{
public:
  inline CSplay();
  inline ~CSplay();
  inline BOOL Init();

protected:
  inline VOID Uninit();
  inline BOOL Initialized() const;
  inline BOOL Valid() const;

public:
  // Insertion Functions
  inline TPData Insert(TPData pData, TKey key);

  // Deletion Functions
  inline TPData Remove(TKey key);
  inline VOID RemoveAll();

  // Look up Functions
  inline TPData Peek(TKey key);

protected:
  inline VOID RemoveAllHelper(CSplayNode<TPData, TKey> *pNode);
  inline VOID Splay(CSplayNode<TPData, TKey> *pNode);
  inline VOID RotateL(CSplayNode<TPData, TKey> *pNode,
		      CSplayNode<TPData, TKey> *pParent);
  inline VOID RotateR(CSplayNode<TPData, TKey> *pNode,
		      CSplayNode<TPData, TKey> *pParent);
  inline VOID RotateLL(CSplayNode<TPData, TKey> *pNode,
		       CSplayNode<TPData, TKey> *pParent,
		       CSplayNode<TPData, TKey> *pGrandParent);
  inline VOID RotateLR(CSplayNode<TPData, TKey> *pNode,
		       CSplayNode<TPData, TKey> *pParent,
		       CSplayNode<TPData, TKey> *pGrandParent);
  inline VOID RotateRL(CSplayNode<TPData, TKey> *pNode,
		       CSplayNode<TPData, TKey> *pParent,
		       CSplayNode<TPData, TKey> *pGrandParent);
  inline VOID RotateRR(CSplayNode<TPData, TKey> *pNode,
		       CSplayNode<TPData, TKey> *pParent,
		       CSplayNode<TPData, TKey> *pGrandParent);

public:
  // Display Functions
  inline VOID Display();

  //Inlined Functions
public:
  inline UINT_32 GetSize();

protected:
  // CSplayNode Allocation
  inline CSplayNode<TPData, TKey> *AllocateSplayNode();
  inline VOID FreeSplayNode(CSplayNode<TPData, TKey> *pNode);

  // Insertion and Deletion

protected:
  CSplayNode<TPData, TKey> *m_pRoot;

  UINT_32 m_uiAllocations;
  UINT_32 m_uiSize;

#ifdef CSPLAYNODEBUFFER
  CSplayNode<TPData, TKey> *m_pFree;
  CSplayNodeBuffer<TPData, TKey> *m_pCSplayNodeBuffer;
#endif // CSPLAYNODEBUFFER
};

//////////////////////////////////////////////////
// Inlined Functions
//

template <class TPData, class TKey>
inline BOOL CSplay<TPData, TKey>::Initialized() const
{
  return TRUE;
}

template <class TPData, class TKey>
inline BOOL CSplay<TPData, TKey>::Valid() const
{
  return Initialized() && 
         TRUE;
}

template <class TPData, class TKey>
inline UINT_32 CSplay<TPData, TKey>::GetSize()
{
  return m_uiSize;
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// Class Implementations
//

//////////////////////////////////////////////////////////////////////////////
// CSplayNode Class
//

template <class TPData, class TKey>
inline CSplayNode<TPData, TKey>::CSplayNode()
{
  m_pParent = NULL;
  m_pLeftChild = NULL;
  m_pRightChild = NULL;

  ::SetNull(m_key);
  m_pData = NULL;
}

template <class TPData, class TKey>
inline CSplayNode<TPData, TKey>::~CSplayNode()
{
  Uninit();
}

template <class TPData, class TKey>
inline BOOL CSplayNode<TPData, TKey>::Init(TKey key, TPData pData)
{
  m_pParent = NULL;
  m_pLeftChild = NULL;
  m_pRightChild = NULL;

  m_key = key;
  m_pData = pData;

  return TRUE;
}

template <class TPData, class TKey>
inline VOID CSplayNode<TPData, TKey>::Uninit()
{
  m_pParent = NULL;
  m_pLeftChild = NULL;
  m_pRightChild = NULL;

  ::SetNull(m_key);
  m_pData = NULL;
}

template <class TPData, class TKey>
inline VOID CSplayNode<TPData, TKey>::Insert(CSplayNode *pNode)
{
  ASSERT( pNode != NULL );

  // Some operations would not be consistent if multiple data items had the same key
  ASSERT( m_key != pNode->m_key );

  if ( pNode->m_key < m_key )
    {
      if ( m_pLeftChild == NULL )
	{
	  pNode->m_pParent = this;
	  m_pLeftChild = pNode;
	  return;
	}
      m_pLeftChild->Insert(pNode);
    }
  else
    {
      if ( m_pRightChild == NULL )
	{
	  pNode->m_pParent = this;
	  m_pRightChild = pNode;
	  return;
	}
      m_pRightChild->Insert(pNode);
    }
}

template <class TPData, class TKey>
inline VOID CSplayNode<TPData, TKey>::Display()
{
  printf("( k = ");
  ::DisplayValue(m_key);
  printf(", d = %p)", m_pData);
}

//////////////////////////////////////////////////////////////////////////////
// CSplayNodeBuffer Class
//

template <class TPData, class TKey>
inline CSplayNodeBuffer<TPData, TKey>::CSplayNodeBuffer()
{
  m_pNext = NULL;

  m_uiCSplayNodes = 0;
  m_pCSplayNodes = NULL;
}

template <class TPData, class TKey>
inline CSplayNodeBuffer<TPData, TKey>::~CSplayNodeBuffer()
{
  Uninit();
}

template <class TPData, class TKey>
inline BOOL CSplayNodeBuffer<TPData, TKey>::Init(UINT_32 uiSize/* = 20*/)
{
  UINT_32 i;

  ASSERT( uiSize != 0 );

  m_pNext = NULL;

  m_uiCSplayNodes = uiSize;

  m_pCSplayNodes = new CSplayNode<TPData, TKey>[m_uiCSplayNodes];
  ASSERT( m_pCSplayNodes != NULL );

  for ( i = 1; i < m_uiCSplayNodes; i++ )
    {
      m_pCSplayNodes[i-1].SetParent(&(m_pCSplayNodes[i]));
    }

  return TRUE;
}

template <class TPData, class TKey>
inline VOID CSplayNodeBuffer<TPData, TKey>::Uninit()
{
  m_pNext = NULL;

  m_uiCSplayNodes = 0;
  if ( m_pCSplayNodes != NULL )
    {
      delete []m_pCSplayNodes;
      m_pCSplayNodes = NULL;
    }
}

//////////////////////////////////////////////////////////////////////////////
// CSplay Class
//

template <class TPData, class TKey>
inline CSplay<TPData, TKey>::CSplay()
{
  m_pRoot = NULL;

  m_uiAllocations = 0;
  m_uiSize = 0;

#ifdef CSPLAYNODEBUFFER
  m_pFree = NULL;
  m_pCSplayNodeBuffer = NULL;
#endif // CSPLAYNODEBUFFER
}

template <class TPData, class TKey>
inline CSplay<TPData, TKey>::~CSplay()
{
  Uninit();
}

template <class TPData, class TKey>
inline BOOL CSplay<TPData, TKey>::Init()
{
  ASSERT( m_pRoot == NULL &&
	  m_uiSize == 0 );

  m_pRoot = NULL;

  m_uiAllocations = 0;
  m_uiSize = 0;
  ASSERT( Valid() );

  return TRUE;
}

template <class TPData, class TKey>
inline VOID CSplay<TPData, TKey>::Uninit()
{
  RemoveAll();
  ASSERT( m_pRoot == NULL &&
	  m_uiSize == 0 &&
	  m_uiAllocations == 0 );

#ifdef CSPLAYNODEBUFFER
  while ( m_pCSplayNodeBuffer != NULL )
    {
      CSplayNodeBuffer<TPData, TKey> *pNodeBuffer;

      pNodeBuffer = m_pCSplayNodeBuffer;
      m_pCSplayNodeBuffer = m_pCSplayNodeBuffer->GetNext();
      delete pNodeBuffer;
    }

  m_pFree = NULL;
#endif // CSPLAYNODEBUFFER
}

template <class TPData, class TKey>
inline TPData CSplay<TPData, TKey>::Insert(TPData pData, TKey key)
{
  ASSERT( Valid() );

  CSplayNode<TPData, TKey> *pNode;

  // Allocate and Initialize a CSplayNode
  pNode = AllocateSplayNode();
  m_uiSize++;
  ASSERT( pNode != NULL );
  pNode->Init(key, pData);

  if ( m_pRoot == NULL )
    {
      m_pRoot = pNode;
    }
  else
    {
      CSplayNode<TPData, TKey> *pCurr;

      for ( pCurr = m_pRoot; pCurr != NULL; )
	{
	  // Some operations would not be consistent if multiple data items had the same key
	  ASSERT( pCurr->GetKey() != pNode->GetKey() );
	  
	  if ( pNode->GetKey() < pCurr->GetKey() )
	    {
	      if ( pCurr->GetLeftChild() == NULL )
		{
		  pNode->SetParent(pCurr);
		  pCurr->SetLeftChild(pNode);
		  break;
		}
	      pCurr = pCurr->GetLeftChild();
	    }
	  else
	    {
	      if ( pCurr->GetRightChild() == NULL )
		{
		  pNode->SetParent(pCurr);
		  pCurr->SetRightChild(pNode);
		  break;
		}
	      pCurr = pCurr->GetRightChild();
	    }
	}
      ASSERT( pCurr != NULL );

      Splay(pNode);
    }
  
  return pData;
}

template <class TPData, class TKey>
inline TPData CSplay<TPData, TKey>::Remove(TKey key)
{
  CSplayNode<TPData, TKey> *pCurr;
  CSplayNode<TPData, TKey> *pPrev;
  CSplayNode<TPData, TKey> *pNext;
  TPData pData;

  pData = NULL;
  pPrev = NULL;
  for ( pCurr = m_pRoot; pCurr != NULL; )
    {
      if ( pCurr->GetKey() == key )
	{
	  // Found it
	  pData = pCurr->GetData();
	  break;
	}
      else if ( key < pCurr->GetKey() )
	{
	  pPrev = pCurr;
	  pCurr = pCurr->GetLeftChild();
	}
      else 
	{
	  pPrev = pCurr;
	  pCurr = pCurr->GetRightChild();	  
	}
    }

  // Check if we actually found it
  if ( pCurr != NULL )
    {
      if ( pCurr->GetLeftChild() == NULL && pCurr->GetRightChild() == NULL )
	{
	  if ( pCurr->GetParent() == NULL )
	    {
	      // This is the root node
	      m_pRoot = NULL;
	    }
	  else
	    {
	      if ( pCurr->GetParent()->GetLeftChild() == pCurr )
		{
		  pCurr->GetParent()->SetLeftChild(NULL);
		}
	      else
		{
		  ASSERT( pCurr->GetParent()->GetRightChild() == pCurr );
		  pCurr->GetParent()->SetRightChild(NULL);
		}
	    }
	}
      else if ( pCurr->GetLeftChild() == NULL )
	{
	  pPrev = pCurr->GetRightChild();
	  if ( pCurr->GetParent() == NULL )
	    {
	      // This is the root node
	      m_pRoot = pCurr->GetRightChild();
	      pCurr->GetRightChild()->SetParent(NULL);
	    }
	  else
	    {
	      if ( pCurr->GetParent()->GetLeftChild() == pCurr )
		{
		  pCurr->GetParent()->SetLeftChild(pCurr->GetRightChild());
		  pCurr->GetRightChild()->SetParent(pCurr->GetParent());
		}
	      else
		{
		  ASSERT( pCurr->GetParent()->GetRightChild() == pCurr );
		  pCurr->GetParent()->SetRightChild(pCurr->GetRightChild());
		  pCurr->GetRightChild()->SetParent(pCurr->GetParent());
		}
	    }
	}
      else if ( pCurr->GetRightChild() == NULL )
	{
	  pPrev = pCurr->GetLeftChild();
	  if ( pCurr->GetParent() == NULL )
	    {
	      // This is the root node
	      m_pRoot = pCurr->GetLeftChild();
	      pCurr->GetLeftChild()->SetParent(NULL);
	    }
	  else
	    {
	      if ( pCurr->GetParent()->GetLeftChild() == pCurr )
		{
		  pCurr->GetParent()->SetLeftChild(pCurr->GetLeftChild());
		  pCurr->GetLeftChild()->SetParent(pCurr->GetParent());
		}
	      else
		{
		  ASSERT( pCurr->GetParent()->GetRightChild() == pCurr );
		  pCurr->GetParent()->SetRightChild(pCurr->GetLeftChild());
		  pCurr->GetLeftChild()->SetParent(pCurr->GetParent());
		}
	    }
	}
      else
	{
	  // Find the next element
	  for ( pNext = pCurr->GetRightChild(); 
		pNext->GetLeftChild() != NULL;
		pNext = pNext->GetLeftChild() )
	    {
	    }
	  pPrev = pNext;

	  // Remove pNext from the tree
	  if ( pNext->GetParent()->GetLeftChild() == pNext )
	    {
	      pNext->GetParent()->SetLeftChild(pNext->GetRightChild());
	      if ( pNext->GetRightChild() != NULL )
		{
		  pNext->GetRightChild()->SetParent(pNext->GetParent());
		}
	    }
	  else
	    {
	      ASSERT( pNext->GetParent()->GetRightChild() == pNext );
	      pNext->GetParent()->SetRightChild(pNext->GetRightChild());
	      if ( pNext->GetRightChild() != NULL )
		{
		  pNext->GetRightChild()->SetParent(pNext->GetParent());
		}
	    }

	  // Replace pCurr by pNext
	  pNext->SetLeftChild(pCurr->GetLeftChild());
	  if ( pNext->GetLeftChild() != NULL )
	    {
	      pNext->GetLeftChild()->SetParent(pNext);
	    }
	  pNext->SetRightChild(pCurr->GetRightChild());
	  if ( pNext->GetRightChild() != NULL )
	    {
	      pNext->GetRightChild()->SetParent(pNext);
	    }
	  pNext->SetParent(pCurr->GetParent());
	  if ( pCurr->GetParent() == NULL )
	    {
	      // This is the root node
	      m_pRoot = pNext;
	    }
	  else
	    {
	      if ( pCurr->GetParent()->GetLeftChild() == pCurr )
		{
		  pCurr->GetParent()->SetLeftChild(pNext);
		}
	      else
		{
		  ASSERT( pCurr->GetParent()->GetRightChild() == pCurr );
		  pCurr->GetParent()->SetRightChild(pNext);
		}
	    }
	}

      // Finally deallocate the tree node
      m_uiSize--;
      FreeSplayNode(pCurr);
    }

  if ( pPrev != NULL )
    {
      Splay(pPrev);
    }

  return pData;
}

template <class TPData, class TKey>
inline VOID CSplay<TPData, TKey>::RemoveAll()
{
  if ( m_pRoot != NULL )
    {
      RemoveAllHelper(m_pRoot);
      m_pRoot = NULL;
    }
}

template <class TPData, class TKey>
inline VOID CSplay<TPData, TKey>::RemoveAllHelper(CSplayNode<TPData, TKey> *pNode)
{
  ASSERT( pNode != NULL );

  if ( pNode->GetLeftChild() != NULL )
    {      
      RemoveAllHelper(pNode->GetLeftChild());
    }
  if ( pNode->GetRightChild() != NULL )
    {      
      RemoveAllHelper(pNode->GetRightChild());
    }

  m_uiSize--;
  FreeSplayNode(pNode);
}

template <class TPData, class TKey>
inline TPData CSplay<TPData, TKey>::Peek(TKey key)
{
  CSplayNode<TPData, TKey> *pCurr;
  CSplayNode<TPData, TKey> *pPrev;
  TPData pData;

  pData = NULL;
  pPrev = NULL;
  for ( pCurr = m_pRoot; pCurr != NULL; )
    {
      pPrev = pCurr;
      if ( pCurr->GetKey() == key )
	{
	  // Found it
	  pData = pCurr->GetData();
	  break;
	}
      else if ( key < pCurr->GetKey() )
	{
	  pCurr = pCurr->GetLeftChild();
	}
      else 
	{
	  pCurr = pCurr->GetRightChild();	  
	}
    }

  if ( pPrev != NULL )
    {
      Splay(pPrev);
    }

  return pData;
}

template <class TPData, class TKey>
inline VOID CSplay<TPData, TKey>::Splay(CSplayNode<TPData, TKey> *pNode)
{
  ASSERT( pNode != NULL );

  CSplayNode<TPData, TKey> *pParent;
  CSplayNode<TPData, TKey> *pGrandParent;

  for ( pParent = pNode->GetParent();
	pParent != NULL;
	pParent = pNode->GetParent() )
    {
      pGrandParent = pParent->GetParent();
      
      if ( pGrandParent == NULL )
	{
	  if ( pNode == pParent->GetLeftChild() )
	    {
	      // L
	      RotateL(pNode, pParent);
	    }
	  else
	    {
	      ASSERT( pNode == pParent->GetRightChild() );
	      // R
	      RotateR(pNode, pParent);
	    }
	}
      else
	{
	  if ( pParent == pGrandParent->GetLeftChild() )
	    {
	      // L?
	      if ( pNode == pParent->GetLeftChild() )
		{
		  // LL
		  RotateLL(pNode, pParent, pGrandParent);
		}
	      else
		{
		  ASSERT( pNode == pParent->GetRightChild() );
		  // LR
		  RotateLR(pNode, pParent, pGrandParent);
		}
	    }
	  else
	    {
	      ASSERT( pParent == pGrandParent->GetRightChild() );
	      // R?
	      if ( pNode == pParent->GetLeftChild() )
		{
		  // RL
		  RotateRL(pNode, pParent, pGrandParent);
		}
	      else
		{
		  ASSERT( pNode == pParent->GetRightChild() );
		  // RR
		  RotateRR(pNode, pParent, pGrandParent);
		}
	    }
	}
    }

  m_pRoot = pNode;
}

template <class TPData, class TKey>
inline VOID CSplay<TPData, TKey>::RotateL(CSplayNode<TPData, TKey> *pNode,
					  CSplayNode<TPData, TKey> *pParent)
{
  ASSERT( pNode != NULL );
  ASSERT( pParent != NULL );
  ASSERT( pNode->GetParent() == pParent );
  ASSERT( pNode == pParent->GetLeftChild() );

  pParent->SetLeftChild(pNode->GetRightChild());
  if ( pParent->GetLeftChild() != NULL )
    {
      pParent->GetLeftChild()->SetParent(pParent);
    }
  pNode->SetRightChild(pParent);
  pNode->SetParent(pParent->GetParent());
  if ( pNode->GetParent() != NULL )
    {
      if ( pNode->GetParent()->GetLeftChild() == pParent )
	{
	  pNode->GetParent()->SetLeftChild(pNode);
	}
      else
	{
	  ASSERT( pNode->GetParent()->GetRightChild() == pParent );
	  pNode->GetParent()->SetRightChild(pNode);	  
	}
    }
  pParent->SetParent(pNode);
}

template <class TPData, class TKey>
inline VOID CSplay<TPData, TKey>::RotateR(CSplayNode<TPData, TKey> *pNode,
					  CSplayNode<TPData, TKey> *pParent)
{
  ASSERT( pNode != NULL );
  ASSERT( pParent != NULL );
  ASSERT( pNode->GetParent() == pParent );
  ASSERT( pNode == pParent->GetRightChild() );

  pParent->SetRightChild(pNode->GetLeftChild());
  if ( pParent->GetRightChild() != NULL )
    {
      pParent->GetRightChild()->SetParent(pParent);
    }
  pNode->SetLeftChild(pParent);
  pNode->SetParent(pParent->GetParent());
  if ( pNode->GetParent() != NULL )
    {
      if ( pNode->GetParent()->GetLeftChild() == pParent )
	{
	  pNode->GetParent()->SetLeftChild(pNode);
	}
      else
	{
	  ASSERT( pNode->GetParent()->GetRightChild() == pParent );
	  pNode->GetParent()->SetRightChild(pNode);	  
	}
    }
  pParent->SetParent(pNode);
}

template <class TPData, class TKey>
inline VOID CSplay<TPData, TKey>::RotateLL(CSplayNode<TPData, TKey> *pNode,
					   CSplayNode<TPData, TKey> *pParent,
					   CSplayNode<TPData, TKey> *pGrandParent)
{
  ASSERT( pNode != NULL );
  ASSERT( pParent != NULL );
  ASSERT( pGrandParent != NULL );
  ASSERT( pNode->GetParent() == pParent );
  ASSERT( pNode == pParent->GetLeftChild() );
  ASSERT( pParent->GetParent() == pGrandParent );
  ASSERT( pParent == pGrandParent->GetLeftChild() );

  pGrandParent->SetLeftChild(pParent->GetRightChild());
  if ( pGrandParent->GetLeftChild() != NULL )
    {
      pGrandParent->GetLeftChild()->SetParent(pGrandParent);
    }
  pParent->SetLeftChild(pNode->GetRightChild());
  if ( pParent->GetLeftChild() != NULL )
    {
      pParent->GetLeftChild()->SetParent(pParent);
    }

  pNode->SetParent(pGrandParent->GetParent());
  if ( pNode->GetParent() != NULL )
    {
      if ( pNode->GetParent()->GetLeftChild() == pGrandParent )
	{
	  pNode->GetParent()->SetLeftChild(pNode);
	}
      else
	{
	  ASSERT( pNode->GetParent()->GetRightChild() == pGrandParent );
	  pNode->GetParent()->SetRightChild(pNode);	  
	}
    }
  pNode->SetRightChild(pParent);
  pParent->SetParent(pNode);
  pParent->SetRightChild(pGrandParent);
  pGrandParent->SetParent(pParent);
}

template <class TPData, class TKey>
inline VOID CSplay<TPData, TKey>::RotateRR(CSplayNode<TPData, TKey> *pNode,
					   CSplayNode<TPData, TKey> *pParent,
					   CSplayNode<TPData, TKey> *pGrandParent)
{
  ASSERT( pNode != NULL );
  ASSERT( pParent != NULL );
  ASSERT( pGrandParent != NULL );
  ASSERT( pNode->GetParent() == pParent );
  ASSERT( pNode == pParent->GetRightChild() );
  ASSERT( pParent->GetParent() == pGrandParent );
  ASSERT( pParent == pGrandParent->GetRightChild() );

  pGrandParent->SetRightChild(pParent->GetLeftChild());
  if ( pGrandParent->GetRightChild() != NULL )
    {
      pGrandParent->GetRightChild()->SetParent(pGrandParent);
    }
  pParent->SetRightChild(pNode->GetLeftChild());
  if ( pParent->GetRightChild() != NULL )
    {
      pParent->GetRightChild()->SetParent(pParent);
    }

  pNode->SetParent(pGrandParent->GetParent());
  if ( pNode->GetParent() != NULL )
    {
      if ( pNode->GetParent()->GetLeftChild() == pGrandParent )
	{
	  pNode->GetParent()->SetLeftChild(pNode);
	}
      else
	{
	  ASSERT( pNode->GetParent()->GetRightChild() == pGrandParent );
	  pNode->GetParent()->SetRightChild(pNode);	  
	}
    }
  pNode->SetLeftChild(pParent);
  pParent->SetParent(pNode);
  pParent->SetLeftChild(pGrandParent);
  pGrandParent->SetParent(pParent);
}

template <class TPData, class TKey>
inline VOID CSplay<TPData, TKey>::RotateLR(CSplayNode<TPData, TKey> *pNode,
					   CSplayNode<TPData, TKey> *pParent,
					   CSplayNode<TPData, TKey> *pGrandParent)
{
  ASSERT( pNode != NULL );
  ASSERT( pParent != NULL );
  ASSERT( pGrandParent != NULL );
  ASSERT( pNode->GetParent() == pParent );
  ASSERT( pNode == pParent->GetRightChild() );
  ASSERT( pParent->GetParent() == pGrandParent );
  ASSERT( pParent == pGrandParent->GetLeftChild() );

  pGrandParent->SetLeftChild(pNode->GetRightChild());
  if ( pGrandParent->GetLeftChild() != NULL )
    {
      pGrandParent->GetLeftChild()->SetParent(pGrandParent);
    }
  pParent->SetRightChild(pNode->GetLeftChild());
  if ( pParent->GetRightChild() != NULL )
    {
      pParent->GetRightChild()->SetParent(pParent);
    }

  pNode->SetParent(pGrandParent->GetParent());
  if ( pNode->GetParent() != NULL )
    {
      if ( pNode->GetParent()->GetLeftChild() == pGrandParent )
	{
	  pNode->GetParent()->SetLeftChild(pNode);
	}
      else
	{
	  ASSERT( pNode->GetParent()->GetRightChild() == pGrandParent );
	  pNode->GetParent()->SetRightChild(pNode);	  
	}
    }
  pNode->SetLeftChild(pParent);
  pParent->SetParent(pNode);
  pNode->SetRightChild(pGrandParent);
  pGrandParent->SetParent(pNode);
}

template <class TPData, class TKey>
inline VOID CSplay<TPData, TKey>::RotateRL(CSplayNode<TPData, TKey> *pNode,
					   CSplayNode<TPData, TKey> *pParent,
					   CSplayNode<TPData, TKey> *pGrandParent)
{
  ASSERT( pNode != NULL );
  ASSERT( pParent != NULL );
  ASSERT( pGrandParent != NULL );
  ASSERT( pNode->GetParent() == pParent );
  ASSERT( pNode == pParent->GetLeftChild() );
  ASSERT( pParent->GetParent() == pGrandParent );
  ASSERT( pParent == pGrandParent->GetRightChild() );

  pGrandParent->SetRightChild(pNode->GetLeftChild());
  if ( pGrandParent->GetRightChild() != NULL )
    {
      pGrandParent->GetRightChild()->SetParent(pGrandParent);
    }
  pParent->SetLeftChild(pNode->GetRightChild());
  if ( pParent->GetLeftChild() != NULL )
    {
      pParent->GetLeftChild()->SetParent(pParent);
    }

  pNode->SetParent(pGrandParent->GetParent());
  if ( pNode->GetParent() != NULL )
    {
      if ( pNode->GetParent()->GetLeftChild() == pGrandParent )
	{
	  pNode->GetParent()->SetLeftChild(pNode);
	}
      else
	{
	  ASSERT( pNode->GetParent()->GetRightChild() == pGrandParent );
	  pNode->GetParent()->SetRightChild(pNode);	  
	}
    }
  pNode->SetRightChild(pParent);
  pParent->SetParent(pNode);
  pNode->SetLeftChild(pGrandParent);
  pGrandParent->SetParent(pNode);
}

/*
template <class TPData, class TKey>
inline VOID CSplay<TPData, TKey>::Splay(CSplayNode<TPData, TKey> *pNode)
{
  ASSERT( pNode != NULL );

  CSplayNode<TPData, TKey> *pParent;
  CSplayNode<TPData, TKey> *pGrandParent;

  for ( pParent = pNode->GetParent();
	pParent != NULL;
	pParent = pNode->GetParent() )
    {
      pGrandParent = pParent->GetParent();
      if ( pGrandParent != NULL )
	{
	  if ( pGrandParent->GetLeftChild() == pParent )
	    {
	      pGrandParent->SetLeftChild(pNode);
	    }
	  else
	    {
	      ASSERT( pGrandParent->GetRightChild() == pParent );
	      pGrandParent->SetRightChild(pNode);
	    }
	}
      pNode->SetParent(pGrandParent);

      if ( pParent->GetLeftChild() == pNode )
	{
	  pParent->SetLeftChild(pNode->GetRightChild());
	  if ( pParent->GetLeftChild() != NULL )
	    {
	      pParent->GetLeftChild()->SetParent(pParent);
	    }
	  pNode->SetRightChild(pParent);
	}
      else
	{
	  ASSERT( pParent->GetRightChild() == pNode );
	  pParent->SetRightChild(pNode->GetLeftChild());
	  if ( pParent->GetRightChild() != NULL )
	    {
	      pParent->GetRightChild()->SetParent(pParent);
	    }
	  pNode->SetLeftChild(pParent);
	}
      pParent->SetParent(pNode);      
    }
  m_pRoot = pNode;
}
*/

template <class TPData, class TKey>
inline VOID CSplay<TPData, TKey>::Display()
{
  ASSERT( Valid() );
  
  printf("Splay:\n");

  if ( m_pRoot == NULL )
    {
      return;
    }

  CList<CSplayNode<TPData, TKey>*, UINT_32> queue;
  CSplayNode<TPData, TKey> *pCurr;
  UINT_32 uiCurr, uiPrev;

  queue.Init();
  queue.InsertLast(m_pRoot, 0);
  uiPrev = 0;

  for ( pCurr = queue.RemoveFirst(&uiCurr); 
	pCurr != NULL; 
	pCurr = queue.RemoveFirst(&uiCurr) )
    {
      if ( uiPrev != uiCurr )
	{
	  printf("\n");
	  uiPrev = uiCurr;
	}
      if ( pCurr->GetParent() != NULL )
	{
	  ::DisplayValue(pCurr->GetParent()->GetKey());
	  if ( pCurr->GetParent()->GetLeftChild() == pCurr )
	    {
	      printf("L:");
	    }
	  else
	    {
	      ASSERT( pCurr->GetParent()->GetRightChild() == pCurr );
	      printf("R:");
	    }
	}

      pCurr->Display();
      printf(" ");

      if ( pCurr->GetLeftChild() != NULL )
	{
	  queue.InsertLast(pCurr->GetLeftChild(), uiCurr+1);
	}
      if ( pCurr->GetRightChild() != NULL )
	{
	  queue.InsertLast(pCurr->GetRightChild(), uiCurr+1);
	}
    }
  printf("\n");
}

#ifdef CSPLAYNODEBUFFER

template <class TPData, class TKey>
inline CSplayNode<TPData, TKey> *CSplay<TPData, TKey>::AllocateSplayNode()
{
  CSplayNode<TPData, TKey> *pNode;  

  ASSERT( m_uiAllocations == m_uiSize );

  if ( m_pFree == NULL )
    {
//printf("*** Allocating a new SplayNode Buffer ***\n");
      // The free splay is empty so
      // dynamically allocate a new buffer
      CSplayNodeBuffer<TPData, TKey> *pNodeBuffer;  

      pNodeBuffer = new CSplayNodeBuffer<TPData, TKey>;
      ASSERT( pNodeBuffer != NULL );

      pNodeBuffer->Init();

      // Insert the new buffer at the beginning of the buffer splay
      // so that we can deallocate it later
      pNodeBuffer->SetNext(m_pCSplayNodeBuffer);
      m_pCSplayNodeBuffer = pNodeBuffer;

      // Set the free splay to be the splay contained in the buffer
      m_pFree = m_pCSplayNodeBuffer->GetBuffer();
      ASSERT( m_pFree != NULL );
    }

  ASSERT( m_pFree != NULL );
  // Take the first free SplayNode off the free splay
  pNode = m_pFree;
  m_pFree = m_pFree->GetParent();
  pNode->SetParent(NULL);
  m_uiAllocations++;

  return pNode;
}

template <class TPData, class TKey>
inline VOID CSplay<TPData, TKey>::FreeSplayNode(CSplayNode<TPData, TKey> *pNode)
{
  ASSERT( pNode != NULL );

  // Insert the newly freed SplayNode at the beginning of the free splay to be reused
  pNode->SetParent(m_pFree);
  m_pFree = pNode;
  m_uiAllocations--;

  ASSERT( m_uiAllocations == m_uiSize );
}

#else // CSPLAYNODEBUFFER

template <class TPData, class TKey>
inline CSplayNode<TPData, TKey> *CSplay<TPData, TKey>::AllocateSplayNode()
{
  CSplayNode<TPData, TKey> *pNode;  

  ASSERT( Valid() );

  pNode = new CSplayNode<TPData, TKey>;
  ASSERT( pNode != NULL );
  m_uiAllocations++;

  return pNode;
}

template <class TPData, class TKey>
inline VOID CSplay<TPData, TKey>::FreeSplayNode(CSplayNode<TPData, TKey> *pNode)
{
  ASSERT( Initialized() );
  ASSERT( pNode != NULL );

  delete pNode;
  m_uiAllocations--;

  ASSERT( Valid() );
}

#endif // CSPLAYNODEBUFFER

#endif // __SPLAY_H__








