/**************************************
 * SLIDE - Tcl func argument parse module
 * Parses args from exported TCL functions
 * of the format { list... } -option value
 * 3.11.99
 * Richard Sun
 **************************************/

#include "tclparseargs.h"


// Given a string w/ space-delim'd entries, make a argList
//
struct argList *TCLParseArgs_List (char *liststr, int argType)
{
  struct argList *newlist = NULL;
  struct argList *templist = NULL;
  char *tempstr;
  char *expr;
  char *liststr_copy;

  ASSERT (liststr!= NULL);

  liststr_copy = new char[strlen (liststr) +2];
  strcpy (liststr_copy, liststr);
  liststr_copy[strlen(liststr)+1] = '\0';  // endbuffer

  //printf ("Parsing: %s\n",liststr_copy);


  tempstr = strtok(liststr_copy, " ");

  while ( tempstr != NULL )
    {
      //printf ("tok - %s\n", tempstr);

      if (tempstr[0]==' ' || tempstr[0]=='\t' || tempstr[0]=='\n')
	  {
		tempstr = strtok(NULL, " ");
		continue;
	  }

      // build new node
      //
      if (newlist == NULL)
	{
	  newlist = new argList;
	  templist = newlist;
	}
      else
	{
	  templist->next = new argList;
	  templist = templist->next;
	}

      templist->next = NULL;

      // do the right thing.
      //
      if (tempstr[0] == '{')
	{
	  //Expression stuff
	  expr = tempstr+1;
	  
	  int brackcnt=1;
	  for (int x=0; x< (int)strlen(expr); x++)
	    if (expr[x] == '{')
	      brackcnt++;
	  
	  templist->value = NULL;
	  while (brackcnt > 0)
	  {
		  tempstr = strtok(NULL, "}");
		  
		  if (tempstr != NULL)
		  {
			brackcnt--;
			//printf ("   subtok: %s\n",tempstr);
			for (int x=0; x< (int)strlen(tempstr); x++)
				if (tempstr[x] == '{')
					brackcnt++;
		  }
		  else
			{
			  fprintf (stderr, "ParseError: Unclosed expression in list\n");
			  return NULL;
			}
		  
		  if (templist->value == NULL)
		  {
			templist->value = new CHAR[strlen(expr) + strlen(tempstr) + 2];
			strcpy ((CHAR*)(templist->value), expr);
			strcat ((CHAR*)(templist->value), " ");
			strcat ((CHAR*)(templist->value), tempstr);
		  }
		  else
		  {
			  char *temp = new CHAR[strlen ((char*)(templist->value)) + strlen(tempstr) + 3 ];
			  strcpy (temp, (CHAR*)(templist->value));
			  strcat (temp, "}");
			  strcat (temp, tempstr);

			  delete templist->value;
			  templist->value = temp;
		  }

		// cheap hack for strings w/ }}
		if (tempstr[strlen(tempstr)+1] == '}')
		{
		  brackcnt--;
		  tempstr = strtok(NULL, "}");
		}
		  
	  }

	  templist->argType = ARG_EXPR;
	  //printf ("   subexpr final: %s\n", (char*)(templist->value));

	}
      else
	{
	  switch (argType)
	    {
	    case ARG_INT:
	      //(int*)(templist->value) = new int;
		  templist->value = new int;
	      *((int*)templist->value) = atoi (tempstr);
	      templist->argType = argType;
	      break;

	    case ARG_FLOAT:
	      //(FLOAT*)(templist->value) = new FLOAT;
		  templist->value = new FLOAT;
	      *((FLOAT*)templist->value) = (FLOAT)atof (tempstr);
	      templist->argType = argType;
	      break;

	    case ARG_STRING:
	      //(CHAR*)(templist->value)) = new CHAR[strlen(tempstr)];
		  templist->value = new CHAR[strlen(tempstr) + 1];
	      strcpy ((CHAR*)(templist->value), tempstr);
	      templist->argType = argType;
	      break;

	    default:
	      return NULL;
	    }
	}

      tempstr = strtok(NULL, " ");
    }

  delete liststr_copy;

  return newlist;
}



// Given a string w/ space-delim'd entries, make an argv array of char*
//
int TCLParseArgs_argv (char *liststr, char ***argv)
{
  char *tempstr;
  char *expr;
  char *liststr_copy;
  int argnum=0;

  ASSERT (liststr!= NULL);

  liststr_copy = new char[strlen (liststr) +1];
  strcpy (liststr_copy, liststr);

  //printf ("Parsing: %s\n",liststr_copy);


  *argv = NULL;

  tempstr = strtok(liststr_copy, " ");

  while ( tempstr != NULL )
    {
      //printf ("tok%d - %s\n",i++, tempstr);

      if (tempstr[0]==' ' || tempstr[0]=='\t' || tempstr[0]=='\n')
	  {
		tempstr = strtok(NULL, " ");
		continue;
	  }

      // make new array entry
      //
      if (*argv == NULL)
	{
	  *argv = (char **)malloc(sizeof (char **));
	}
      else
	{
	  argnum++;
	  *argv = (char **)realloc(*argv, (argnum+1) * sizeof (char **));
	}
      
      (*argv)[argnum] = NULL;

      // do the right thing.
      //
      if (tempstr[0] == '{')
	{
	  //Expression stuff
	  expr = tempstr+1;
	  
	  int brackcnt=1;

	  while (brackcnt > 0)
	  {
		  tempstr = strtok(NULL, "}");
		  
		  if (tempstr != NULL)
		  {
			brackcnt--;
			//printf ("   subtok: %s\n",tempstr);
			for (int x=0; x< (int)strlen(tempstr); x++)
				if (tempstr[x] == '{')
					brackcnt++;
		  }
		  else
			{
			  fprintf (stderr, "ParseError: Unclosed expression in list\n");
			  return NULL;
			}
		  
		  if ((*argv)[argnum] == NULL)
		  {
			(*argv)[argnum] = new CHAR[strlen(expr) + strlen(tempstr) + 2];
			strcpy ((*argv)[argnum], expr);
			strcat ((*argv)[argnum], " ");
			strcat ((*argv)[argnum], tempstr);

		  }
		  else
		  {
			  char *temp = new CHAR[strlen ((*argv)[argnum]) + strlen(tempstr) + 3 ];
			  strcpy (temp, (*argv)[argnum]);
			  strcat (temp, "}");
			  strcat (temp, tempstr);

			  delete (*argv)[argnum];
			  (*argv)[argnum] = temp;
			  
		  }
		  
	  }

	}
      else
	{
	  (*argv)[argnum] = new CHAR[strlen(tempstr) + 2];
	  strcpy ((*argv)[argnum], tempstr);	  
	}

      tempstr = strtok(NULL, " ");
    }

  delete liststr_copy;

  return argnum+1;
}


void TCLParseArgs_Usage(const CHAR *progheader, 
		     struct argFormat* aformat, int iArgtypes)
{
  int i;
  printf ("%s ", progheader);
  
  for (i=0; i<iArgtypes; i++)
    {
      if (! (aformat[i].argReq) )
	printf ("[");

      if (! (aformat[i].argType & ARG_SIMPLE) )
	printf ("-%s ", aformat[i].argName);

      if ( aformat[i].argType & ARG_LIST )
	printf ("{ ");

      if (aformat[i].argType != ARG_BOOL)
	printf ("%s", aformat[i].descShort);
      
      if ( aformat[i].argType & ARG_LIST )
	printf ( "... }");

      if (! (aformat[i].argReq) )
	printf ("]");
      printf (" ");
    }
  
  printf ("\n\n");


  for (i=0; i<iArgtypes; i++)
    {
      if (! (aformat[i].argType & ARG_SIMPLE) )
	printf ("  -%s ", aformat[i].argName);
      else
	printf ("   ");

      if (aformat[i].argType != ARG_BOOL)
	printf ("%s   ", aformat[i].descShort);

      if (aformat[i].argType & ARG_LIST)
	printf ("list of ");

      switch (aformat[i].argType & !ARG_LIST)
	{
	case ARG_BOOL:
	  printf ("(set Boolean switch)\n");
	  break;

	case ARG_STRING:
	  printf ("(String argument)\n");
	  break;

	case ARG_INT:
	  printf ("(Integer argument)\n");
	  break;

	case ARG_FLOAT:
	  printf ("(FLOAT argument)\n");
	  break;

	default:
	  if (aformat[i].argType & ARG_SIMPLE)
	    printf ("\n");
	  else
	    {
	      fprintf (stderr,  "ERROR: TCLParseArgs configured wrong, check code!\n");
	      exit(1);
	    }
	}

      
      
      if (aformat[i].argReq)
	printf ("      [REQUIRED] ");
      else printf ("      ");

      printf ("%s", aformat[i].descLong);
      
      printf ("\n");
    }

  printf ("\n");

}





// NOTE: THIS DOES NOT DO ARG TYPE CHECKING!!!!
//       ie if you try to assign a string to an int, it'll do something
//       weird.
//
int TCLParseArgs (struct argFormat* aformat, int iArgtypes,
	       int argc, CHAR **argv)
{
  int i=1;
  int k=0;
  CHAR *tempstr;  


  // array stores whether param has been used.
  //
  BOOL *fulfilled = new BOOL [iArgtypes];
  for (int x=0; x<iArgtypes; x++)
	  fulfilled[x] = FALSE;

  //printf ("argc: %d\n",argc);

  while (i < argc)
    {
      if ((aformat[k].argType) & ARG_SIMPLE)
	{
	  fulfilled[k] = TRUE;

	  /*
	  i++;
	  if (i >= argc)
	    {
	      fprintf (stderr,"Missing argument parameter for arg - %s\n",
		aformat[k].argName);

	      free (fulfilled);
	      return i;
	    }
	    */

	  int dattype = aformat[k].argType & (~ARG_SIMPLE);

	  switch (dattype)
	    {
	    case ARG_INT:
	      *((int *)aformat[k].argStorage) = atoi (argv[i]);
		  break;

		case ARG_DYNINT:
		  *(DynInt **)(aformat[k].argStorage) = new DynInt;
		  if (argv[i][0] >= '0' && argv[i][0] <= '9' || argv[i][0] == '-')
            (*(DynInt **)(aformat[k].argStorage))->SetValue (atoi (argv[i]));
		  else
		  {
			tempstr = new char[strlen(argv[i])+1];
			strcpy (tempstr, argv[i]);
			(*(DynInt **)(aformat[k].argStorage))->SetDynExpr (tempstr);
		  }
	      break;
	      
	    case ARG_STRING:
	      (*(CHAR **)(aformat[k].argStorage)) = 
		     new char [strlen(argv[i]) +1];
	      strcpy ((*(CHAR **)(aformat[k].argStorage)), argv[i]);
	      break;

	    case ARG_FLOAT:
	      *((FLOAT *)aformat[k].argStorage) = (FLOAT)atof (argv[i]);
		  break;

		case ARG_DYNFLOAT:
		  *(DynFloat **)(aformat[k].argStorage) = new DynFloat;
 		  if (argv[i][0] >= '0' && argv[i][0] <= '9' || argv[i][0] == '-')
            (*(DynFloat **)(aformat[k].argStorage))->SetValue ((FLOAT)atof (argv[i]));
		  else
		  {
			tempstr = new char[strlen(argv[i])+1];
			strcpy (tempstr, argv[i]);
			(*(DynFloat **)(aformat[k].argStorage))->SetDynExpr (tempstr);
		  }
	      break;

	    case ARG_BOOL:
	      //invalid
	      delete[] fulfilled;
	      return i;
	      
	    default:
	      if (dattype & ARG_LIST)
		{
		  (*(struct argList **)(aformat[k].argStorage)) = 
		    TCLParseArgs_List (argv[i], 
				       dattype & ~ARG_LIST);

		  if (*(struct argList **)(aformat[k].argStorage) == NULL)
		    {
		      fprintf (stderr,"Bad arg struct entry %d\n",k);
		      return i;
		    }
		    
		  break;
		}
	      else
		{
		  fprintf (stderr, "Received unknown arg type?! 1\n");
		  delete[] fulfilled;
		  return i;
		}
	      
	    }
	}  // endif simple
      else
	{
	  if (argv[i][0] != '-')
	    {
	      delete[] fulfilled;
	      return i;
	    }

	  int j;

	  for (j=0; j<iArgtypes; j++)
	    {
	      // switch match found.
	      //
	      if ((aformat[j].argName != NULL) &&
		  (!strcmp (argv[i]+1, aformat[j].argName)) )
		{
		  if (fulfilled[j])
		    {
		      fprintf (stderr, "Warning: Arg '-%s' has already been used.  Taking new value.\n", aformat[j].argName);
		    }
		  fulfilled[j] = TRUE;
		
		  i++;
		  if (i >= argc)
		    {
		      fprintf (stderr, "Missing argument parameter for arg -%s\n",
			       aformat[j].argName);
		      delete[] fulfilled;
		      return i;
		    }

		  //tempstr = argv[i];

		  switch (aformat[j].argType)
		    {
		    case ARG_INT:
		      *((int *)aformat[j].argStorage) = atoi (argv[i]);
		      break;

			case ARG_DYNINT:
			  *(DynInt **)(aformat[j].argStorage) = new DynInt;
			  if ((argv[i][0] >= '0' && argv[i][0] <= '9') || argv[i][0] == '-')
				(*(DynInt **)(aformat[j].argStorage))->SetValue (atoi (argv[i]));
			  else
			  {
				tempstr = new char[strlen(argv[i])+1];
				strcpy (tempstr, argv[i]);
				(*(DynInt **)(aformat[j].argStorage))->SetDynExpr (tempstr);
			  }
			  break;

		      
		    case ARG_STRING:
		      (*(CHAR **)(aformat[j].argStorage)) = 
			    new char [strlen(argv[i]) +1];
		      strcpy ((*(CHAR **)(aformat[j].argStorage)), argv[i]);
		      break;
		      
		    case ARG_FLOAT:
		      *((FLOAT *)aformat[j].argStorage) = (FLOAT)atof (argv[i]);
		      break;

			case ARG_DYNFLOAT:
			  *(DynFloat **)(aformat[j].argStorage) = new DynFloat;
 			  if (argv[i][0] >= '0' && argv[i][0] <= '9' || argv[i][0] == '-')
				(*(DynFloat **)(aformat[j].argStorage))->SetValue ((FLOAT)atof (argv[i]));
			  else
			  {
				tempstr = new char[strlen(argv[i])+1];
				strcpy (tempstr, argv[i]);
				(*(DynFloat **)(aformat[j].argStorage))->SetDynExpr (tempstr);
			  }
			  break;

		      
		    case ARG_BOOL:
		      *((BOOL *)aformat[j].argStorage) = TRUE;
		      break;

		    default:
		      if (aformat[j].argType & ARG_LIST)
			{
			  (*(struct argList **)(aformat[j].argStorage)) = 
			    TCLParseArgs_List (argv[i], 
					       aformat[j].argType & ~ARG_LIST);
			  if (*(struct argList **)(aformat[j].argStorage) == NULL)
			    {
			      fprintf (stderr,"Bad arg struct entry %d\n",j);
			      return i;
			    }

			  break;
			}
		      else
			{
			  fprintf (stderr, "Received unknown arg type?! 2  %x\n", aformat[j].argType);
			  delete[] fulfilled;
			  return i;
			}

		    } //endswitch

		  break;
		} //endif arg == argName
	      
	    } // endfor argtypes

	  if (j >= iArgtypes)
	    {
	      fprintf (stderr, "Ignoring unrecognized switch: %s\n", argv[i]);
	    }
	} // endelse notsimple

      i++;  // incr argv pos
      k++;  // incr argsturct pos
    } //endwhile argc

  int screwed = FALSE;

  for (int j=0; j<iArgtypes; j++)
    {
      if (aformat[j].argReq && !fulfilled[j])
	{
	  fprintf (stderr, "Err: Required arg -%s was not used!\n", aformat[j].argName);

	  screwed = TRUE;
	}
    }

  delete[] fulfilled;

  if (screwed)
    return -1;

  return 0;
}




void TCLAjaxArgList (struct argList* list)
{
  struct argList *tempptr;
  struct argList *head;
  head = list;
  tempptr = list;

  while (tempptr != NULL)
    {
	  tempptr = tempptr->next;
      if (head->argType != ARG_STRING && head->argType != ARG_EXPR)
		  delete head->value;
	  delete head;
      head = tempptr;
    }
}






