/**************************************
 * SLIDE - Tcl func argument parse module
 * Parses args from exported TCL functions
 * of the format { list... } -option value
 * 3.11.99
 * Richard Sun
 **************************************/


#ifndef TCLPARSEARGS_H
#define TCLPARSEARGS_H

#include <stdio.h>
#include <stdlib.h>
//#include <iostream.h>
#ifndef WIN32
#include <strings.h>
#endif

#include "types.h"
#include "debug.h"

#include "dynamics.h"

// Argument types (arg_simple can be &'d with others except bool)
//
#define ARG_SIMPLE   0x100   // simple non-switch arg (position sensitive)
#define ARG_STRING   0x001   // pull string
#define ARG_INT      0x002   // pull int
#define ARG_DYNINT   0x003   // pull dynint (ie int val, or dyn expr)
#define ARG_FLOAT    0x004   // pull float
#define ARG_DYNFLOAT 0x005   // pull dynfloat (ie float val, or dyn expr)
#define ARG_BOOL     0x006   // set boolean

#define ARG_LIST     0x010   // list of...

#define ARG_EXPR     0xFFF   // TCL Expression

struct argFormat
{
  char *argName;       // switch name (make this NULL for SIMPLE non-switch args)
  char *descShort;     // short desc (appears after the switch)
  char *descLong;      // long desc (appears in usage below)
  int   argType;       // arg type, (see above)
  VOID *argStorage;    // var * to store value to.
  BOOL  argReq;        // is argument required?
};

// An argformat example would look something like this:
// struct argFormat myargform[] =
//   { { [switchname], [argumentname],
//       [switch description],
//       [switch type], [ptr to associated storage], [required bool] }, ...
//


struct argList
{
  void *value;
  int   argType;

  struct argList *next;
};


// Given a string w/ space-delim'd entries, make a argList
//
struct argList *TCLParseArgs_List (char *liststr, int argType);

// Given a string w/ space-delim'd entries, make an array of char *
//
int TCLParseArgs_argv (char *string, char ***argv);


// print program usage, starting with progheader
//
void TCLParseArgs_Usage(const char *progheader, 
		     struct argFormat* aformat, int iArgtypes);


// parse args from func params up and store in the specified places
//
int TCLParseArgs (struct argFormat* aformat, int iArgtypes,
	       int argc, char **argv);


//delete ArgList (not stored data in nodes)
//
void TCLAjaxArgList (struct argList* list);


#endif // TCLParseArgs_H





