#include "timer.h"

// WindowsNT
#ifdef WIN32 
#include <windows.h>

#else
// X Windows
#include <sys/time.h>

#endif //windows

//////////////////////////////////////////////////////////////////////
// TimeValue
//

TimeValue::TimeValue()
{
  m_sec = 0;
  m_usec = 0;
}

TimeValue::~TimeValue()
{
}

#ifdef WIN32
// Win32
void TimeValue::RecordTime()
{
  SYSTEMTIME systemTime;
  
  GetSystemTime(&systemTime);
  
  m_sec = (long)(systemTime.wMinute*60 + systemTime.wSecond);
  m_usec = (long)(systemTime.wMilliseconds * 1000);
}

#else
// XWindows
void TimeValue::RecordTime()
{
  struct timeval tv;
  
  gettimeofday(&tv, NULL);

  m_sec = tv.tv_sec;
  m_usec = tv.tv_usec;
}
#endif



TimeValue TimeValue::operator+(const TimeValue& tv)
{
  TimeValue tvRes;

  tvRes.m_usec = m_usec + tv.m_usec;
  tvRes.m_sec = m_sec + tv.m_sec;

  if ( tvRes.m_usec >= ciTimer_sec_frac ) 
    {  
      tvRes.m_usec -= ciTimer_sec_frac;
      tvRes.m_sec++; 
    }

  return tvRes;
}

TimeValue TimeValue::operator-(const TimeValue& tv)
{
  TimeValue tvRes;

  if (tv.m_usec > m_usec) 
    {  
      m_usec += ciTimer_sec_frac;
      m_sec--; 
    }
  tvRes.m_usec = m_usec - tv.m_usec;
  tvRes.m_sec = m_sec - tv.m_sec;

  return tvRes;
}

//////////////////////////////////////////////////////////////////////
// Timer
//

Timer::Timer()
{
  m_bRunning = FALSE;
  m_dSecTotal = 0.0;
}

Timer::~Timer()
{
}

void Timer::Reset()
{
  ASSERT( !m_bRunning );
  m_dSecTotal = 0.0;
}

void Timer::Start()
{
  ASSERT( !m_bRunning );
  m_bRunning = TRUE;
  m_tvStart.RecordTime();
}

void Timer::Stop()
{
  TimeValue tv;

  ASSERT( m_bRunning );
  m_tvStop.RecordTime();
  tv = m_tvStop - m_tvStart;
  m_dSecTotal += tv.GetSeconds();
  m_bRunning = FALSE;
}

