#include "window.h"

#include "slideparser.h"
#include "slideworld.h"
#include "slideface.h"
#include "slidenode.h"
#include "slideviewport.h"
#include "sliderender.h"


CWindow *gpCWindow = NULL;
CWindow *gpCWindowCurr = NULL;

BOOL CWindow::m_bTclWindowExists = FALSE;

static CHAR gpcBuffer[200];

static const INT_32 ciWindowSize = 500;

//////////////////////////////////////////////////////////////////////
// CWindow Class
//

Tcl_Interp *CWindow::m_pInterp = NULL;

CHashString *CWindow::m_pHashKeyboard = NULL;

const int CWindow::m_cuiOotMethods = cuiCWindowOotMethods;

// OOT Registration Variables
OotMethodEntry CWindow::m_pOotMethods[cuiCWindowOotMethods] =
{
  CWINDOW_METHODS
};
const CHAR *CWindow::m_cppcOotFields[CWindow::ftSize] =
{
  CWINDOW_FIELDS_STRINGS
};
const CHAR *CWindow::m_cpcOotClassName = "window";
OotClass *CWindow::m_pOotClass = NULL;

UINT_32 CWindow::m_uiOotCounter = 0;

// Local Defaults
//static const BOOL cbDisplayEvents = TRUE;
static const BOOL cbDisplayEvents = FALSE;
static const CWindow::EventLevelType ceEventLevel = CWindow::eltAction;
//static const CWindow::EventLevelType ceEventLevel = CWindow::eltCallBack;

CWindow::CWindow()
{
  m_uiOotID = 0;
  m_pcOotName = NULL;
  m_pOotObject = NULL;

  m_pTogl = FALSE;
  m_bRedisplayPending = FALSE;

  m_pcFile = NULL;
  m_uiLine = 0;
  m_uiColumn = 0;
  m_pcSLIDEID = NULL;
  m_pCSLIDEWorld = NULL;

  m_pcDisplayID = NULL;
  m_eResize = slideoftSLF_ON;
  m_eAspectRatio = slideartSLF_ANY;
}

CWindow::~CWindow()
{
  Uninit();
}

VOID CWindow::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		   CHAR *pcSLIDEID)
{
  UINT_32 i;

  ASSERT( pcFile != NULL );
  ASSERT( pcSLIDEID != NULL );

  // Set Up Keyboard Interpreter
  // This is only done once for the whole CWindow Class
  if ( m_pHashKeyboard == NULL )
    {
      m_pHashKeyboard = new CHashString;
      ASSERT( m_pHashKeyboard != NULL );
      m_pHashKeyboard->Init((3*gcuiglsiKeyBindings)/2);
      
      for ( i = 0; i < gcuiglsiKeyBindings; i++ )
	{
	  m_pHashKeyboard->Insert(&(gcpglsiKeyBindings[i]), gcpglsiKeyBindings[i].pcKey);
	}
    }

  m_uiOotID = 0;
  m_pcOotName = NULL;
  m_pOotObject = NULL;

  m_pTogl = FALSE;
  m_bRedisplayPending = FALSE;

  m_iWidth = ciWindowSize;
  m_iHeight = ciWindowSize;

  m_bDisplayEvents = cbDisplayEvents;
  m_eEventLevel = ceEventLevel;

  // SLIDE Stuff
  m_pcFile = pcFile;
  m_uiLine = uiLine;
  m_uiColumn = uiColumn;
  m_pcSLIDEID = pcSLIDEID;
  m_pCSLIDEWorld = NULL;
  m_listCSLIDEViewport.Init();

  // move default settings to TCL.
  m_vOrigin.Set(-1,-1,0);
  m_vSize.Set(-1,-1,0);
  m_vBackground.Null();
  m_eResize = slideoftSLF_ON;
  m_eAspectRatio = slideartSLF_ANY;

  // Set up the OOT communication link with Tcl
  InitTcl();

  // Create the actual Togl drawing area and initialize OpenGL
  //InitTogl();

  SetDrawBounds(FALSE);
  SetColorDynamicBounds(FALSE);
  SetColorChangedBounds(FALSE);
  SetCheckBoundChanged(TRUE);
  SetDrawClipBound(FALSE);
  SetDrawViewportBound(FALSE);
  SetDrawNormals(FALSE);
  SetDrawCDAOperands(FALSE);
  SetBackFaceCullingEnabled(TRUE);
  SetClipEnabled(TRUE);
  SetBoundCullingEnabled(TRUE);
  SetLightingEnabled(TRUE);
  SetFogEnabled(TRUE);
  SetTextureMappingEnabled(TRUE);
  SetCrystalBallEnabled(TRUE);
  SetSelectionEnabled(FALSE);

  SetSecondsPerFrameRequested(CSLIDEWorld::GetSecondsPerFrameRequested());
  SetUpdateTime(0.0);
  SetRenderTime(0.0);  
}

VOID CWindow::InitTcl()
{
  //////////////////////////////////////////////////
  // Tcl Stuff
  m_uiOotID = (++m_uiOotCounter);
//  sprintf(gpcBuffer, "%s%ld", m_cpcOotClassName, m_uiOotID);
  sprintf(gpcBuffer, "slf%s", m_pcSLIDEID);
  m_pcOotName = new CHAR[strlen(gpcBuffer) + 1];
  ASSERT( m_pcOotName != NULL );
  strcpy(m_pcOotName, gpcBuffer);

  m_pOotObject = Oot_CreateObject(m_pInterp, m_pOotClass, m_pcOotName, this);
  ASSERT( m_pOotObject != NULL );
}

VOID CWindow::InitTogl()
{
  INT_32 iCode;

  ASSERT( gpCWindowCurr == NULL );
  gpCWindowCurr = this;
  
  sprintf(gpcBuffer, "CreateWindow . %s %f %f %f %f %ld %ld", m_pcOotName, 
	  m_vOrigin(0), m_vOrigin(1), m_vSize(0), m_vSize(1), (UINT_32)m_eResize, 
	  (UINT_32)m_eAspectRatio);
  iCode = Tcl_GlobalEval(m_pInterp, gpcBuffer);

  if ( iCode != TCL_OK )
    {
      printf("ERROR CreateWindow: %s\n", Tcl_GetStringResult(m_pInterp));
    }
  ASSERT( iCode == TCL_OK );

  m_bTclWindowExists = TRUE;

  Prerender();
}

VOID CWindow::Prerender()
{
  // Do the prerendering of static primitives for multiple viewports/windows
  CListIterInt iter;
  CSLIDEViewport *pCSLIDEViewport;

  ASSERT( m_pTogl != NULL );
  Togl_MakeCurrent(m_pTogl);
    
  // Set up Textures
  m_pCSLIDEWorld->Prerender();

  iter.Init(&m_listCSLIDEViewport);
    
  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      pCSLIDEViewport->Prerender();
    }
}

VOID CWindow::HandleDestroy()
{
  INT_32 iCode;

  if ( m_pInterp != NULL && m_pOotClass != NULL && m_pcOotName != NULL )
    {
      sprintf(gpcBuffer, "DestroyWindow . %s", m_pcOotName);
      iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
      ASSERT( iCode == TCL_OK );

      m_bTclWindowExists = FALSE;

      Oot_DestroyObject(m_pInterp, m_pOotClass, m_pcOotName);
    }
}

VOID CWindow::Uninit()
{
  if ( m_pcOotName != NULL )
    {
      delete []m_pcOotName;
      m_pcOotName = NULL;
    }

  // We do not own this memory
  m_pOotObject = NULL;
  m_pTogl = NULL;
  m_pCSLIDEWorld = NULL;
}

VOID CWindow::InsertViewport(CSLIDEViewport *pCSLIDEViewport)
{
  ASSERT( pCSLIDEViewport != NULL );

  m_listCSLIDEViewport.InsertLast(pCSLIDEViewport, 0);
}


BOOL CWindow::CreateTclToplevel()
{
  INT_32 iCode;

  // Create Togl Drawing Area
  if (m_pcDisplayID != NULL)
    {
#ifdef WIN32
      printf ("Warning: Window Display specified for window %s -- This doesn't work in Windows\n", m_pcOotName);
      // Do not attempt to open a window on another display
      sprintf(gpcBuffer, "toplevel .%s", m_pcOotName);

#else // WIN32

      // code for bringing up windows on different machine displays
      sprintf(gpcBuffer, "toplevel .%s -screen \"%s\"", m_pcOotName,m_pcDisplayID);

#endif // WIN32
    }
  else
    {
      sprintf(gpcBuffer, "toplevel .%s", m_pcOotName);
    }


  iCode = Tcl_GlobalEval(m_pInterp, gpcBuffer);
  return ( iCode == TCL_OK );
}

BOOL CWindow::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  m_pCSLIDEWorld = pCSLIDEWorld;

  return TRUE;
}

VOID CWindow::Preprocess()
{
  // Open the Window
  // Create the actual Togl drawing area and initialize OpenGL
  InitTogl();
}

VOID CWindow::Update()
{
  m_vBackground.Update();

  PostRedisplay();
}

VOID CWindow::WriteRIB(CHAR *pcOutputFile, FILE *&fpOutputFile)
{
  CListIterInt iter;
  CSLIDEViewport *pCSLIDEViewport;

  ASSERT( pcOutputFile != NULL );
  ASSERT( fpOutputFile != NULL );
  
  if ( fpOutputFile != stdout )
    {
      ASSERT( fpOutputFile != NULL );
      fclose(fpOutputFile);
      fpOutputFile = NULL;
    }
  
  fpOutputFile = stdout;


  if ( pcOutputFile != NULL )
    {
      RiBegin(pcOutputFile);
    }
  else
    {
      RiBegin(RI_NULL);
    }
  {
    // RenderMan has a Left-Handed Coordinate system by default
    // this is BAD, because everything we do is with Right-Handed systems
    // so we must specify this
    RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE SET DEFAULTS:");
    RiOrientation(RI_LH);
    RiShadingRate(1.0);
    RiPixelSamples(2.0, 2.0);
    RiSides(1);
    
    // Turn on shadows for all lights as a default
    // ???? RiAttribute("light", "special", "shadows", "special", "on", RI_NULL);
    RiArchiveRecord(RI_COMMENT, "\nAttribute \"light\" \"shadows\" \"on\"");

    RiArchiveRecord(RI_COMMENT, "   END SLIDE SET DEFAULTS:");

    // This should be a .tif file so I am going to modify the .rib name
    ASSERT( pcOutputFile != NULL );
    ASSERT( strlen(pcOutputFile) >= 4 );
    strcpy(CSLIDERender::m_pcRIBOutputFileBase, pcOutputFile);
    CSLIDERender::m_pcRIBOutputFileBase[strlen(CSLIDERender::m_pcRIBOutputFileBase) - 4] = '\0';
/*
    CHAR *pcRIBOutputFile; 
    pcRIBOutputFile = new CHAR[strlen(pcOutputFile)+1];
    ASSERT( pcRIBOutputFile != NULL );
    strcpy(pcRIBOutputFile, pcOutputFile);
    strcpy(pcRIBOutputFile + strlen(pcRIBOutputFile) - 4, ".tif");
    RiDisplay(pcRIBOutputFile, RI_FRAMEBUFFER, RI_RGB, RI_NULL);
    delete []pcRIBOutputFile;

    RiFormat(m_iWidth, m_iHeight, 1.0);
*/
    CSLIDERender::m_uiRIBFrame = 0;
 
    UINT_32 i;
    for ( i = 1; i <= 1; i++ )
      {
	//RiFrameBegin(i);
	{
	  iter.Init(&m_listCSLIDEViewport);
	  
	  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
		pCSLIDEViewport != NULL;
		pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
	    {
	      pCSLIDEViewport->WriteRIB(fpOutputFile, 0, 0, 
				      (UINT_32)m_iWidth, (UINT_32)m_iHeight);
	    }  
	}
	//RiFrameEnd();
      }
  }

  // Make notes about or create any offline files which are necessary for 
  // the scene like texture maps, environment maps, or shadow maps ...
  m_pCSLIDEWorld->WriteRIB(fpOutputFile);

  RiEnd();

  fpOutputFile = NULL;
}

VOID CWindow::WritePOV(CHAR *pcOutputFile, 
		       FILE *&fpOutputFile, 
		       CListInt *plistTransformIDs,
		       CListInt *plistColors)  
{
  ASSERT( pcOutputFile != NULL );
  ASSERT( fpOutputFile != NULL );
  
  CListIterInt iter;
  CSLIDEViewport *pCSLIDEViewport;

  if ( pcOutputFile != NULL )
    {
      ASSERT( pcOutputFile != NULL );
      ASSERT( strlen(pcOutputFile) >= 4 );
      
      UINT_32 i;
      for ( i = 0; i < 1; i++ )
	{
	  iter.Init(&m_listCSLIDEViewport);
	  
	  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
		pCSLIDEViewport != NULL;
		pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
	    {
	      pCSLIDEViewport->WritePOV(fpOutputFile, 
					plistTransformIDs,
					plistColors);
	    }  
	}
    }
}

VOID CWindow::WriteWRLZ(CHAR *pcOutputFile, FILE *&fpOutputFile)
{
  ASSERT( pcOutputFile != NULL );
  ASSERT( fpOutputFile != NULL );

  CListIterInt iter;
  CSLIDEViewport *pCSLIDEViewport;

  fprintf(fpOutputFile, "#VRML V2.0 utf8\n\n");
  fprintf(fpOutputFile, "Background {\n");
  fprintf(fpOutputFile, "  skyColor [%.6f %.6f %.6f]\n",   
	  m_vBackground[X], m_vBackground[Y], m_vBackground[Z]);
  fprintf(fpOutputFile, "}\n\n");


  if ( pcOutputFile != NULL )
    {
      ASSERT( pcOutputFile != NULL );
      ASSERT( strlen(pcOutputFile) >= 4 );
      
      UINT_32 i;
      for ( i = 0; i < 1; i++ )
	{
	  iter.Init(&m_listCSLIDEViewport);
	  
	  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
		pCSLIDEViewport != NULL;
		pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
	    {
	      pCSLIDEViewport->WriteWRLZ(fpOutputFile);
	    }  
	}
    }

  fprintf(fpOutputFile, "NavigationInfo {\n");
  fprintf(fpOutputFile, "  type \"EXAMINE\"\n");
  fprintf(fpOutputFile, "}\n\n");

}

VOID CWindow::WriteSTL(CHAR *pcOutputFile, FILE *&fpOutputFile)
{
  ASSERT( pcOutputFile != NULL );
  ASSERT( fpOutputFile != NULL );
  
  CListIterInt iter;
  CSLIDEViewport *pCSLIDEViewport;

  if ( pcOutputFile != NULL )
    {
      ASSERT( pcOutputFile != NULL );
      ASSERT( strlen(pcOutputFile) >= 4 );
      
      UINT_32 i;
      for ( i = 0; i < 1; i++ )
	{
	  iter.Init(&m_listCSLIDEViewport);
	  
	  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
		pCSLIDEViewport != NULL;
		pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
	    {
	      pCSLIDEViewport->WriteSTL(fpOutputFile);
	    }  
	}
    }
}

VOID CWindow::WriteSIF(CHAR *pcOutputFile, FILE *&fpOutputFile)
{
  ASSERT( pcOutputFile != NULL );
  ASSERT( fpOutputFile != NULL );
  
  CListIterInt iter;
  CSLIDEViewport *pCSLIDEViewport;

  if ( pcOutputFile != NULL )
    {
      ASSERT( pcOutputFile != NULL );
      ASSERT( strlen(pcOutputFile) >= 4 );
      
      UINT_32 i;
      for ( i = 0; i < 1; i++ )
	{
	  iter.Init(&m_listCSLIDEViewport);
	  
	  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
		pCSLIDEViewport != NULL;
		pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
	    {
	      pCSLIDEViewport->WriteSIF(fpOutputFile);
	    }  
	}
    }
}

VOID CWindow::WriteLSIF(CHAR *pcOutputFile, FILE *&fpOutputFile)
{
  ASSERT( pcOutputFile != NULL );
  ASSERT( fpOutputFile != NULL );
  
  CListIterInt iter;
  CSLIDEViewport *pCSLIDEViewport;

  if ( pcOutputFile != NULL )
    {
      ASSERT( pcOutputFile != NULL );
      ASSERT( strlen(pcOutputFile) >= 4 );
      
      UINT_32 i;
      for ( i = 0; i < 1; i++ )
	{
	  iter.Init(&m_listCSLIDEViewport);
	  
	  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
		pCSLIDEViewport != NULL;
		pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
	    {
	      pCSLIDEViewport->WriteLSIF(fpOutputFile);
	    }  
	}
    }
}

VOID CWindow::WriteSLIDE(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtWindow], m_pcSLIDEID);
  fprintf(fpOutputFile, "  %s ( %.3f %.3f )\n", 
	  gcppcSLIDEReservedWords[sliderwtOrigin], 
	  m_vOrigin[X], m_vOrigin[Y]);
  fprintf(fpOutputFile, "  %s ( %.3f %.3f )\n", 
	  gcppcSLIDEReservedWords[sliderwtWinSize], 
	  m_vSize[X], m_vSize[Y]);
  fprintf(fpOutputFile, "  %s ( %.3f %.3f %.3f )\n", 
	  gcppcSLIDEReservedWords[sliderwtBackground], 
	  m_vBackground[X], m_vBackground[Y], m_vBackground[Z]);
  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndWindow]);
}

VOID CWindow::WriteDotID(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "window_%s", GetSLIDEID());
}

VOID CWindow::WriteDotNode(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "  ");
  WriteDotID(fpOutputFile);
  fprintf(fpOutputFile, " [ label = \"%s\", shape = box ];\n", GetSLIDEID());
}

VOID CWindow::WriteDotEdges(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDEViewport *pCSLIDEViewport;

  ASSERT( fpOutputFile != NULL );

  iter.Init(&m_listCSLIDEViewport);

  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      fprintf(fpOutputFile, "  ");
      WriteDotID(fpOutputFile);
      fprintf(fpOutputFile, " -> ");
      pCSLIDEViewport->WriteDotID(fpOutputFile);
      fprintf(fpOutputFile, "\n");
    }  
}

VOID CWindow::HandleReshape(INT_32 iWidth, INT_32 iHeight)
{
  DisplayEvent(eltCallBack, "Entering HandleReshape(%3ld, %3ld)\n", 
	       iWidth, iHeight);

  CListIterInt iter;
  CSLIDEViewport *pCSLIDEViewport;

  m_iWidth = iWidth;
  m_iHeight = iHeight;

  // Send messages to all the viewports
  iter.Init(&m_listCSLIDEViewport);

  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      pCSLIDEViewport->HandleReshape((UINT_32)m_iWidth, (UINT_32)m_iHeight);
    }  
/*
  gilViewport(0, 0, m_iWidth, m_iHeight);

  gilMatrixMode(GL_PROJECTION);
  gilLoadIdentity();
  //giluOrtho2D(0, 1, 0, 1);
  //giluOrtho2D(0, m_iWidth, 0, m_iHeight);
  //gilOrtho(0, m_iWidth, 0, m_iHeight, -1000, 1000);
  gilOrtho(-1, 1, -1, 1, 0, 1);

  gilMatrixMode(GL_MODELVIEW);
  gilLoadIdentity();
*/
}

VOID CWindow::HandleDisplay()
{
  DisplayEvent(eltCallBack, "Entering HandleDisplay()\n");

  CListIterInt iter;
  CSLIDEViewport *pCSLIDEViewport;
  Timer t;

  t.Reset();
  t.Start();

//printf("Window: HandleDisplay\n");

  m_bRedisplayPending = FALSE;

  // Make our Togl Window the current OpenGL Context
  ASSERT( m_pTogl != NULL );
  Togl_MakeCurrent(m_pTogl);

  gilClearColor(m_vBackground[0], m_vBackground[1], m_vBackground[2], 
		m_vBackground[3]);
  gilClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

  // Polygon Modes -- OpenGL defaults are commented out
  gilFrontFace(GL_CCW);
  //gilPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  gilCullFace(GL_BACK);
  if ( CSLIDEFace::GetBackFaceCullingEnabled() )
    {
      gilEnable(GL_CULL_FACE);
    }
  else
    {
      gilDisable(GL_CULL_FACE);
    }
  gilDisable(GL_TEXTURE_2D);

  gilEnable(GL_DEPTH_TEST);
  gilDepthFunc(GL_LEQUAL);

  if ( CSLIDERender::GetLightingEnabled() )
    {
      gilEnable(GL_LIGHTING);
    }
  else
    {
      gilDisable(GL_LIGHTING);
    }

  gilEnable(GL_NORMALIZE);
  gilShadeModel(GL_SMOOTH);

  //gilTexEnvf(GL_TEXTURE_2D, GL_TEXTURE_ENV_MODE, GL_DECAL);

  gilColor3f(0.0, 1.0, 1.0);

  iter.Init(&m_listCSLIDEViewport);

  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      pCSLIDEViewport->Render(0, 0,
			      (UINT_32)m_iWidth, (UINT_32)m_iHeight);
    }

  Togl_SwapBuffers(m_pTogl);

  t.Stop();
  SetUpdateTime(CSLIDEWorld::GetUpdateTime());
  SetRenderTime(t.GetSeconds());
}

VOID CWindow::HandleMouseButton(INT_32 iButton, INT_32 iState, INT_32 iX, INT_32 iY)
{
  DisplayEvent(eltCallBack, "Entering HandleMouseButton(%ld, %ld, %3ld, %3ld)\n", 
	       iButton, iState, iX, iY);

  BOOL b, bPostRedisplay;
  FLOAT fX, fY;
  CListIterInt iter;
  CSLIDEViewport *pCSLIDEViewport;

  bPostRedisplay = FALSE;

  // Change Pixel coordinates into Normalize coords (0,0) -> (1,1)
  VPToNorm(fX, fY, iX, iY);

  if ( iState == GLSI_DOWN )
    {
      // Fill in the appropriate actions and remember to set
      // bPostRedisplay = TRUE if you want to redraw any changes
      // to your data structures.

      switch( iButton ) 
	{
	case GLSI_LEFT_BUTTON:
	  DisplayEvent(eltAction, "L down  [%3ld %3ld]\n", iX, iY);
	  break;
	  
	case GLSI_MIDDLE_BUTTON:
	  DisplayEvent(eltAction, "M down  [%3ld %3ld]\n", iX, iY);
	  break;
	
	case GLSI_RIGHT_BUTTON:
	  DisplayEvent(eltAction, "R down  [%3ld %3ld]\n", iX, iY);
	  break;
	}
    }
  else if ( iState == GLSI_UP )
    {
      switch( iButton ) 
	{
	case GLSI_LEFT_BUTTON:
	  DisplayEvent(eltAction, "L up    [%3ld %3ld]\n", iX, iY);
	  break;
	  
	case GLSI_MIDDLE_BUTTON:
	  DisplayEvent(eltAction, "M up    [%3ld %3ld]\n", iX, iY);
	  break;
	
	case GLSI_RIGHT_BUTTON:
	  DisplayEvent(eltAction, "R up    [%3ld %3ld]\n", iX, iY);
	  break;
	}
    }

  // Send messages to all the viewports
  iter.Init(&m_listCSLIDEViewport);

  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      b = pCSLIDEViewport->HandleMouseButton((UINT_32)m_iWidth, 
					     (UINT_32)m_iHeight,
					     iButton, iState, 
					     iX - 0, iY - 0);
      bPostRedisplay = bPostRedisplay || b;
    }

  if ( bPostRedisplay )
    {
      PostRedisplay();
    }
}

VOID CWindow::HandleMouseMotion(INT_32 iX, INT_32 iY)
{
  DisplayEvent(eltCallBack, "Entering HandleMouseMotion(%3ld, %3ld)\n", 
	       iX, iY);

  BOOL b, bPostRedisplay;
  FLOAT fX, fY;
  CListIterInt iter;
  CSLIDEViewport *pCSLIDEViewport;

  bPostRedisplay = FALSE;

  // Change Pixel coordinates into Normalize coords (0,0) -> (1,1)
  VPToNorm(fX, fY, iX, iY);

  //sprintf(gpcBuffer, "%f", MIN(1.0, MAX(fX, 0.0)));
  //Tcl_SetVar(m_pInterp, "SLF_MOUSE_X", gpcBuffer, TCL_GLOBAL_ONLY);
  //sprintf(gpcBuffer, "%f", MIN(1.0, MAX(fY, 0.0)));
  //Tcl_SetVar(m_pInterp, "SLF_MOUSE_Y", gpcBuffer, TCL_GLOBAL_ONLY);

  // Send messages to all the viewports
  iter.Init(&m_listCSLIDEViewport);

  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      b = pCSLIDEViewport->HandleMouseMotion((UINT_32)m_iWidth, 
					     (UINT_32)m_iHeight,
					     iX - 0, iY - 0);
      bPostRedisplay = bPostRedisplay || b;
    }

  if ( bPostRedisplay )
    {
      PostRedisplay();
    }
}

VOID CWindow::HandleKeyboard(INT_32 iKey, INT_32 iState, INT_32 iX, INT_32 iY)
{
  DisplayEvent(eltCallBack, "Entering HandleKeyboard('%c', %3ld, %3ld)\n",
	       (CHAR)iKey, iX, iY);

  BOOL bPostRedisplay, b;
  FLOAT fX, fY;
  CListIterInt iter;
  CSLIDEViewport *pCSLIDEViewport;

  bPostRedisplay = FALSE;

  // Change Pixel coordinates into Normalize coords (0,0) -> (1,1)
  VPToNorm(fX, fY, iX, iY);

  if ( iState != GLSI_DOWN )
    {
  } else {

  switch ( iKey )
    {
      /*
    case GLSI_KEY_ESCAPE:  // Escape Key
      exit(0);
      break;
      */

    case 'b':
      SetDrawBounds(!CSLIDEPrimitive::GetRenderBoundsEnabled());
      bPostRedisplay = TRUE;
      break;

    case 'd':
      SetColorDynamicBounds(!CSLIDENode::GetColorDynamicBounds());
      bPostRedisplay = TRUE;
      break;

    case 'a':
      SetColorChangedBounds(!CSLIDENode::GetColorChangedBounds());
      bPostRedisplay = TRUE;
      break;

    case 'h':
      SetCheckBoundChanged(!CSLIDENode::GetCheckBoundChanged());
      bPostRedisplay = TRUE;
      break;

    case 'v':
      SetDrawViewportBound(!CSLIDEViewport::GetRenderViewportBound());
      bPostRedisplay = TRUE;
      break;

    case 'c':
      SetDrawClipBound(!CSLIDEPrimitive::GetRenderClipBoundEnabled());
      bPostRedisplay = TRUE;
      break;
/*
    case 'c':
      SetClipEnabled(!CSLIDEFace::GetClipEnabled());
      bPostRedisplay = TRUE;
      break;
*/
    case 'n':
      SetDrawNormals(!CSLIDEFace::GetRenderNormalsEnabled());
      bPostRedisplay = TRUE;
      break;

    case 'o':
      SetDrawCDAOperands(!CSLIDEFace::GetRenderCDAOperandsEnabled());
      bPostRedisplay = TRUE;
      break;

    case 'x':
      SetBackFaceCullingEnabled(!CSLIDEFace::GetBackFaceCullingEnabled());
      bPostRedisplay = TRUE;
      break;

    case 'r':
      SetBoundCullingEnabled(!CSLIDEPrimitive::GetBoundRejectEnabled());
      bPostRedisplay = TRUE;
      break;

    case 'l':
      SetLightingEnabled(!CSLIDERender::GetLightingEnabled());
      bPostRedisplay = TRUE;
      break;

    case 'f':
      SetFogEnabled(!CSLIDEPrimitive::GetFogEnabled());
      bPostRedisplay = TRUE;
      break;

    case 't':
      SetTextureMappingEnabled(!CSLIDEPrimitive::GetTextureMappingEnabled());
      bPostRedisplay = TRUE;
      break;

    case 'u':
      SetCrystalBallEnabled(!CSLIDERender::GetCrystalBallEnabled());
      bPostRedisplay = TRUE;
      break;

    case 'z':
      SetSelectionEnabled(!CSLIDERender::GetSelectionEnabled());
      bPostRedisplay = TRUE;
      break;

    case 's':
      HandleStartDynamics();
      break;

    case 'p':
      HandlePauseDynamics();
      break;

    case ' ':
      {
	CListIterInt iter;
	CSLIDEViewport *pCSLIDEViewport;
	
	iter.Init(&m_listCSLIDEViewport);

	for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	      pCSLIDEViewport != NULL;
	      pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
	  {
	    pCSLIDEViewport->SelectionCycle();
	  }
      }
      break;

    default:
      DisplayEvent(eltAction, "Key '%c' [%3ld %3ld]\n", (CHAR)iKey, iX, iY);
      break;
    }
	} 

  // Send messages to all the viewports
  iter.Init(&m_listCSLIDEViewport);

  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      b = pCSLIDEViewport->HandleKeyboard((UINT_32)m_iWidth, 
					     (UINT_32)m_iHeight,
						 iKey, iState,
					     iX - 0, iY - 0);
      bPostRedisplay = bPostRedisplay || b;
    }


  if ( bPostRedisplay )
    {
      PostRedisplay();
    }
}

VOID CWindow::HandleStartDynamics()
{
  m_pCSLIDEWorld->RestartDynamics();  
}

VOID CWindow::HandlePauseDynamics()
{
  m_pCSLIDEWorld->PauseDynamics();  
}

VOID CWindow::HandleResetCamera()
{
  CListIterInt iter;
  CSLIDEViewport *pCSLIDEViewport;

  iter.Init(&m_listCSLIDEViewport);    
  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      pCSLIDEViewport->ResetCamera();
    }
}

VOID CWindow::HandleEditSelectedObject()
{
  CListIterInt iter;
  CSLIDEViewport *pCSLIDEViewport;

  iter.Init(&m_listCSLIDEViewport);
  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      pCSLIDEViewport->EditSelectedObject();
    }
}

VOID CWindow::SetDrawBounds(BOOL b)
{
  CSLIDEPrimitive::SetRenderBoundsEnabled(b);
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftDrawBounds], b);
}

VOID CWindow::SetColorDynamicBounds(BOOL b)
{
  CSLIDEPrimitive::SetColorDynamicBounds(b);
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftColorDynamicBounds], b);
}

VOID CWindow::SetColorChangedBounds(BOOL b)
{
  CSLIDEPrimitive::SetColorChangedBounds(b);
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftColorChangedBounds], b);
}

VOID CWindow::SetCheckBoundChanged(BOOL b)
{
  CSLIDEPrimitive::SetCheckBoundChanged(b);
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftCheckBoundChanged], b);
}

VOID CWindow::SetDrawClipBound(BOOL b)
{
  CSLIDEPrimitive::SetRenderClipBoundEnabled(b);
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftDrawClipBound], b);
}

VOID CWindow::SetDrawViewportBound(BOOL b)
{
  CSLIDEViewport::SetRenderViewportBound(b);
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftDrawViewportBound], b);
}

VOID CWindow::SetDrawNormals(BOOL b)
{
  CSLIDEFace::SetRenderNormalsEnabled(b);
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftDrawNormals], b);
}

VOID CWindow::SetDrawCDAOperands(BOOL b)
{
  CSLIDEFace::SetRenderCDAOperandsEnabled(b);
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftDrawCDAOperands], b);
}

VOID CWindow::SetBackFaceCullingEnabled(BOOL b)
{
  CSLIDEFace::SetBackFaceCullingEnabled(b);
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftBackFaceCullingEnabled], b);
}

VOID CWindow::SetClipEnabled(BOOL b)
{
  CSLIDEFace::SetClipEnabled(b);
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftClipEnabled], b);
}

VOID CWindow::SetBoundCullingEnabled(BOOL b)
{
  CSLIDEPrimitive::SetBoundRejectEnabled(b);
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftBoundCullingEnabled], b);
}

VOID CWindow::SetLightingEnabled(BOOL b)
{
  CSLIDERender::SetLightingEnabled(b);
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftLightingEnabled], b);
}

VOID CWindow::SetFogEnabled(BOOL b)
{
  CSLIDEPrimitive::SetFogEnabled(b);
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftFogEnabled], b);
}

VOID CWindow::SetTextureMappingEnabled(BOOL b)
{
  CSLIDERender::SetTextureMappingEnabled(b);
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftTextureMappingEnabled], b);
}

VOID CWindow::SetCrystalBallEnabled(BOOL b)
{
  CSLIDERender::SetCrystalBallEnabled(b);
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftCrystalBallEnabled], b);
}

VOID CWindow::SetCUIType(DynInt *pDyniType)
{
  CListIterInt iter;
  CSLIDEViewport *pCSLIDEViewport;

  CSLIDERender::SetWindowCUIType(pDyniType);

  iter.Init(&m_listCSLIDEViewport);    
  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      pCSLIDEViewport->SetCUIType(pDyniType);
    }
}

VOID CWindow::SetCUISpeed(DynFloat *pDynfSpeed)
{
  CListIterInt iter;
  CSLIDEViewport *pCSLIDEViewport;

  CSLIDERender::SetWindowCUISpeed(pDynfSpeed);

  iter.Init(&m_listCSLIDEViewport);    
  for ( pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekFirst());
	pCSLIDEViewport != NULL;
	pCSLIDEViewport = (CSLIDEViewport*)(iter.PeekNext()) )
    {
      pCSLIDEViewport->SetCUISpeed(pDynfSpeed);
    }
}

VOID CWindow::SetSelectionEnabled(BOOL b)
{
  CSLIDERender::SetSelectionEnabled(b);
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftSelectionEnabled], b);

  if( m_bTclWindowExists )
    {
      INT_32 iCode;
      CHAR pcBuffer[200];
      Tcl_Interp *pInterp;
      
      pInterp = CSLIDEWorld::GetInterp();
      sprintf(pcBuffer, "Window:SelectionMode .%s %i", m_pcOotName, b);
      iCode = Tcl_Eval(pInterp, pcBuffer);
      //ASSERT( iCode == TCL_OK );
      if ( iCode != TCL_OK )
	{
	  printf( "%s\n", pcBuffer );
	  printf("Error %s \n", pInterp->result);
	}
    }
}

VOID CWindow::SetSecondsPerFrameRequested(DOUBLE dSecondsPerFrame)
{
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftFramesPerSecondRequested],
		     (dSecondsPerFrame == 0.0) ? (0.0) : (1.0/dSecondsPerFrame));
}

VOID CWindow::SetUpdateTime(DOUBLE dUpdateTime)
{
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftUpdateTime], dUpdateTime);
}

VOID CWindow::SetRenderTime(DOUBLE dRenderTime)
{
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftRenderTime], dRenderTime);
}

VOID CWindow::PostRedisplay()
{
  if ( !m_bRedisplayPending )
    { 
      ASSERT( m_pTogl != NULL );
      Togl_PostRedisplay(m_pTogl);
      m_bRedisplayPending = TRUE;
    }
}

VOID CWindow::DisplayControls()
{
  fprintf(stderr, "\n2D Viewer Controls\n");
  fprintf(stderr, "------------------\n");
  fprintf(stderr, "M Button - Bounding Box Depth Slider Control\n");
  fprintf(stderr, "'b' Key  - Toggles the drawing of the Bounding Boxes\n");
  fprintf(stderr, "'w' Key  - Writes a SLIDE description to stdout\n");
  fprintf(stderr, "Esc Key  - Exits the program\n");
}

VOID CWindow::DisplayEvent(EventLevelType eEventLevel, CHAR *pcFormat, ...)
{
  if ( m_bDisplayEvents && eEventLevel <= m_eEventLevel )
    {
      va_list vaArgs;

      va_start(vaArgs, pcFormat);
      vfprintf(stderr, pcFormat, vaArgs);
      va_end(vaArgs);
    }
}

//////////////////////////////////////////////////////////////////////
// Callback Functions
//

int CWindow::CreateOotClass(Tcl_Interp *pInterp)
{
  ASSERT( pInterp != NULL );

  m_pInterp = pInterp;

  m_pOotClass = Oot_CreateClass(m_pInterp, m_cpcOotClassName, 
				CWindow::CBHandleCreate, 
				CWindow::m_pOotMethods, 
				cuiCWindowOotMethods,

				CWindow::m_cppcOotFields, 
				(UINT_32)(CWindow::ftSize));
  if ( m_pOotClass == NULL )
    {
      return TCL_ERROR;
    }

  return TCL_OK;
}
    
int CWindow::CBHandleCreate(Tcl_Interp *pInterp, OotClass *pClass, 
			     int argc, char ** /*argv*/)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pClass != NULL );

  if ( argc != 2 )
    {
      Tcl_SetResult(pInterp, "wrong args: window <name> <world>", NULL);
      return TCL_ERROR;
    }

  pCWindow = new CWindow;
  ASSERT( pCWindow != NULL );
  pCWindow->Init(NULL, 0, 0, NULL);
  /*
  if ( !pCWindow->Init() )
    {
      delete pCWindow;
      Tcl_SetResult(pInterp, "Error initializing window", NULL);
      return TCL_ERROR;      
    }
    */
  return TCL_OK;
}

VOID CWindow::CBHandleDestroy(ClientData pvClientData)
{
  CWindow *pCWindow;

  ASSERT( pvClientData != NULL );

  pCWindow = (CWindow*)pvClientData;

  pCWindow->HandleDestroy();
}

VOID CWindow::CBHandleToglCreate(struct Togl *pTogl)
{
  ASSERT( pTogl != NULL );

  ASSERT( gpCWindowCurr != NULL );
  //printf("Togl client data = %s\n", gpCWindowCurr->GetSLIDEID());
  Togl_SetClientData(pTogl, (ClientData)gpCWindowCurr);
  gpCWindowCurr->SetTogl(pTogl);
  gpCWindowCurr = NULL;
}

VOID CWindow::CBHandleToglDisplay(struct Togl *pTogl)
{
  CWindow *pCWindow;

  ASSERT( pTogl != NULL );

  pCWindow = (CWindow*)Togl_GetClientData(pTogl);
  ASSERT( pCWindow != NULL );

  pCWindow->HandleDisplay();
}

VOID CWindow::CBHandleToglReshape(struct Togl *pTogl)
{
  CWindow *pCWindow;

  ASSERT( pTogl != NULL );

  pCWindow = (CWindow*)Togl_GetClientData(pTogl);
  ASSERT( pCWindow != NULL );

  pCWindow->HandleReshape(Togl_Width(pTogl), Togl_Height(pTogl));
}

VOID CWindow::CBHandleToglDestroy(struct Togl *pTogl)
{
  CWindow *pCWindow;

  ASSERT( pTogl != NULL );

  pCWindow = (CWindow*)Togl_GetClientData(pTogl);
  ASSERT( pCWindow != NULL );

  //delete pCWindow;

  Togl_SetClientData(pTogl, NULL);
}

int CWindow::CBHandleMouseButton(Tcl_Interp *pInterp, OotObject *pObject,
				 INT_32 iButton, INT_32 iState,
				 INT_32 iX, INT_32 iY)
{
  CWindow *pCWindow;
  INT_32 iNewButton, iNewState;
  INT_32 iXVP, iYVP;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  switch ( iButton )
    {
    case 1:
      iNewButton = GLSI_LEFT_BUTTON;
      break;
    case 2:
      iNewButton = GLSI_MIDDLE_BUTTON;
      break;
    case 3:
      iNewButton = GLSI_RIGHT_BUTTON;
      break;
    default:
      ASSERT( FALSE );
      break;
    }
  
  iNewState = ( iState == 0 ) ? ( GLSI_DOWN ) : ( GLSI_UP );

  pCWindow->WinToVP(iXVP, iYVP, (INT_32)iX, (INT_32)iY);

  pCWindow->HandleMouseButton(iNewButton, iNewState, iXVP, iYVP);

  return TCL_OK;
}

int CWindow::CBHandleMouseMotion(Tcl_Interp *pInterp, OotObject *pObject,
				 INT_32 /*iButton*/, INT_32 iX, INT_32 iY)
{
  CWindow *pCWindow;
  INT_32 iXVP, iYVP;
  
  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->WinToVP(iXVP, iYVP, (INT_32)iX, (INT_32)iY);

  pCWindow->HandleMouseMotion(iXVP, iYVP);

  return TCL_OK;
}

int CWindow::CBHandleKeyboard(Tcl_Interp *pInterp, OotObject *pObject,
			      CHAR *pcKeyCode, INT_32 iState, 
			      INT_32 iX, INT_32 iY)
{
  CWindow *pCWindow;
  INT_32 iKey, iNewState;
  glsiKeyBinding *pKeyBinding;
  INT_32 iXVP, iYVP;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pKeyBinding = (glsiKeyBinding*)(m_pHashKeyboard->Peek(pcKeyCode));
  if ( pKeyBinding != NULL )
    {
      iKey = pKeyBinding->iValue;

      iNewState = ( iState == 0 ) ? ( GLSI_DOWN ) : ( GLSI_UP );
      pCWindow->WinToVP(iXVP, iYVP, (INT_32)iX, (INT_32)iY);
      
      pCWindow->HandleKeyboard(iKey, iNewState, iXVP, iYVP);
    }

  return TCL_OK;
}

int CWindow::CBHandleStartDynamics(Tcl_Interp *pInterp, OotObject *pObject)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->HandleStartDynamics();

  return TCL_OK;
}

int CWindow::CBHandlePauseDynamics(Tcl_Interp *pInterp, OotObject *pObject)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->HandlePauseDynamics();

  return TCL_OK;
}

int CWindow::CBSetDrawBounds(Tcl_Interp *pInterp, OotObject *pObject, BOOL b)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->SetDrawBounds(b);
  sprintf(gpcBuffer, "%ld", (UINT_32)(b));
  Tcl_AppendResult(pInterp, gpcBuffer, NULL);

  return TCL_OK;
}

int CWindow::CBSetColorDynamicBounds(Tcl_Interp *pInterp, OotObject *pObject, BOOL b)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->SetColorDynamicBounds(b);
  sprintf(gpcBuffer, "%ld", (UINT_32)(b));
  Tcl_AppendResult(pInterp, gpcBuffer, NULL);

  return TCL_OK;
}

int CWindow::CBSetColorChangedBounds(Tcl_Interp *pInterp, OotObject *pObject, BOOL b)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->SetColorChangedBounds(b);
  sprintf(gpcBuffer, "%ld", (UINT_32)(b));
  Tcl_AppendResult(pInterp, gpcBuffer, NULL);

  return TCL_OK;
}

int CWindow::CBSetCheckBoundChanged(Tcl_Interp *pInterp, OotObject *pObject, BOOL b)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->SetCheckBoundChanged(b);
  sprintf(gpcBuffer, "%ld", (UINT_32)(b));
  Tcl_AppendResult(pInterp, gpcBuffer, NULL);

  return TCL_OK;
}

int CWindow::CBSetDrawClipBound(Tcl_Interp *pInterp, OotObject *pObject, BOOL b)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->SetDrawClipBound(b);
  sprintf(gpcBuffer, "%ld", (UINT_32)(b));
  Tcl_AppendResult(pInterp, gpcBuffer, NULL);

  return TCL_OK;
}

int CWindow::CBSetDrawViewportBound(Tcl_Interp *pInterp, OotObject *pObject, BOOL b)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->SetDrawViewportBound(b);
  sprintf(gpcBuffer, "%ld", (UINT_32)(b));
  Tcl_AppendResult(pInterp, gpcBuffer, NULL);

  return TCL_OK;
}

int CWindow::CBSetDrawNormals(Tcl_Interp *pInterp, OotObject *pObject, BOOL b)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->SetDrawNormals(b);
  sprintf(gpcBuffer, "%ld", (UINT_32)(b));
  Tcl_AppendResult(pInterp, gpcBuffer, NULL);

  return TCL_OK;
}

int CWindow::CBSetDrawCDAOperands(Tcl_Interp *pInterp, OotObject *pObject, BOOL b)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->SetDrawCDAOperands(b);
  sprintf(gpcBuffer, "%ld", (UINT_32)(b));
  Tcl_AppendResult(pInterp, gpcBuffer, NULL);

  return TCL_OK;
}

int CWindow::CBSetBackFaceCullingEnabled(Tcl_Interp *pInterp, OotObject *pObject, BOOL b)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->SetBackFaceCullingEnabled(b);
  sprintf(gpcBuffer, "%ld", (UINT_32)(b));
  Tcl_AppendResult(pInterp, gpcBuffer, NULL);

  return TCL_OK;
}

int CWindow::CBSetClipEnabled(Tcl_Interp *pInterp, OotObject *pObject, BOOL b)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->SetClipEnabled(b);
  sprintf(gpcBuffer, "%ld", (UINT_32)(b));
  Tcl_AppendResult(pInterp, gpcBuffer, NULL);

  return TCL_OK;
}

int CWindow::CBSetBoundCullingEnabled(Tcl_Interp *pInterp, OotObject *pObject, BOOL b)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->SetBoundCullingEnabled(b);
  sprintf(gpcBuffer, "%ld", (UINT_32)(b));
  Tcl_AppendResult(pInterp, gpcBuffer, NULL);

  return TCL_OK;
}

int CWindow::CBSetLightingEnabled(Tcl_Interp *pInterp, OotObject *pObject, BOOL b)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->SetLightingEnabled(b);
  sprintf(gpcBuffer, "%ld", (UINT_32)(b));
  Tcl_AppendResult(pInterp, gpcBuffer, NULL);

  return TCL_OK;
}

int CWindow::CBSetFogEnabled(Tcl_Interp *pInterp, OotObject *pObject, BOOL b)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->SetFogEnabled(b);
  sprintf(gpcBuffer, "%ld", (UINT_32)(b));
  Tcl_AppendResult(pInterp, gpcBuffer, NULL);

  return TCL_OK;
}

int CWindow::CBSetTextureMappingEnabled(Tcl_Interp *pInterp, OotObject *pObject, BOOL b)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->SetTextureMappingEnabled(b);
  sprintf(gpcBuffer, "%ld", (UINT_32)(b));
  Tcl_AppendResult(pInterp, gpcBuffer, NULL);

  return TCL_OK;
}

int CWindow::CBSetCrystalBallEnabled(Tcl_Interp *pInterp, OotObject *pObject, BOOL b)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->SetCrystalBallEnabled(b);
  sprintf(gpcBuffer, "%ld", (UINT_32)(b));
  Tcl_AppendResult(pInterp, gpcBuffer, NULL);

  return TCL_OK;
}

int CWindow::CBHandleResetCamera(Tcl_Interp *pInterp, OotObject *pObject)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->HandleResetCamera();

  return TCL_OK;
}

int CWindow::CBSetCUIType(Tcl_Interp *pInterp, OotObject *pObject, int iType)
{
  CWindow *pCWindow;
  DynInt *pDyniType;

  pDyniType = new DynInt;
  pDyniType->SetValue(iType);

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->SetCUIType(pDyniType);

  return TCL_OK;
}

int CWindow::CBSetCUISpeed(Tcl_Interp *pInterp, OotObject *pObject, FLOAT fSpeed)
{
  CWindow *pCWindow;
  DynFloat *pDynfSpeed;

  pDynfSpeed = new DynFloat;
  pDynfSpeed->SetValue(fSpeed);

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->SetCUISpeed(pDynfSpeed);

  return TCL_OK;
}

int CWindow::CBSetSelectionEnabled(Tcl_Interp *pInterp, OotObject *pObject, BOOL b)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->SetSelectionEnabled(b);
  sprintf(gpcBuffer, "%ld", (UINT_32)(b));
  Tcl_AppendResult(pInterp, gpcBuffer, NULL);

  return TCL_OK;
}

int CWindow::CBHandleEditSelectedObject(Tcl_Interp *pInterp, OotObject *pObject)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->HandleEditSelectedObject();

  return TCL_OK;
}






