#ifndef __WINDOW_H__
#define __WINDOW_H__

#include "genlib.h"
#include "mathlib.h"
#include "gillib.h"
#include "togl.h"
#include "ootlib.h"

#include "slide.h"

class CSLIDEParser;
class CSLIDEWorld;
class CSLIDEViewport;

//////////////////////////////////////////////////////////////////////
// Class CWindow
//

const int cuiCWindowOotMethods = 25;

#define CWINDOW_METHODS \
  {"handleKeyboard", OOT_PARAM_TYPE_KEYBOARD, CWindow::CBHandleKeyboard}, \
  {"handleMouseButton", OOT_PARAM_TYPE_INT_4, CWindow::CBHandleMouseButton}, \
  {"handleMouseMotion", OOT_PARAM_TYPE_INT_3, CWindow::CBHandleMouseMotion}, \
  {"handleStartDynamics", OOT_PARAM_TYPE_VOID,CWindow::CBHandleStartDynamics},\
  {"handlePauseDynamics", OOT_PARAM_TYPE_VOID,CWindow::CBHandlePauseDynamics},\
  {"setDrawBounds", OOT_PARAM_TYPE_BOOL, CWindow::CBSetDrawBounds}, \
  {"setColorDynamicBounds", OOT_PARAM_TYPE_BOOL, CWindow::CBSetColorDynamicBounds}, \
  {"setColorChangedBounds", OOT_PARAM_TYPE_BOOL, CWindow::CBSetColorChangedBounds}, \
  {"setCheckBoundChanged", OOT_PARAM_TYPE_BOOL, CWindow::CBSetCheckBoundChanged}, \
  {"setDrawClipBound", OOT_PARAM_TYPE_BOOL, CWindow::CBSetDrawClipBound}, \
  {"setDrawViewportBound", OOT_PARAM_TYPE_BOOL, CWindow::CBSetDrawViewportBound}, \
  {"setDrawNormals", OOT_PARAM_TYPE_BOOL, CWindow::CBSetDrawNormals}, \
  {"setDrawCDAOperands", OOT_PARAM_TYPE_BOOL, CWindow::CBSetDrawCDAOperands}, \
  {"setBackFaceCullingEnabled", OOT_PARAM_TYPE_BOOL, CWindow::CBSetBackFaceCullingEnabled},\
  {"setClipEnabled", OOT_PARAM_TYPE_BOOL, CWindow::CBSetClipEnabled}, \
  {"setBoundCullingEnabled", OOT_PARAM_TYPE_BOOL, CWindow::CBSetBoundCullingEnabled}, \
  {"setLightingEnabled", OOT_PARAM_TYPE_BOOL, CWindow::CBSetLightingEnabled}, \
  {"setFogEnabled", OOT_PARAM_TYPE_BOOL, CWindow::CBSetFogEnabled}, \
  {"setTextureMappingEnabled", OOT_PARAM_TYPE_BOOL, CWindow::CBSetTextureMappingEnabled}, \
  {"setCrystalBallEnabled", OOT_PARAM_TYPE_BOOL, CWindow::CBSetCrystalBallEnabled}, \
  {"resetCamera", OOT_PARAM_TYPE_VOID, CWindow::CBHandleResetCamera}, \
  {"setCUIType", OOT_PARAM_TYPE_INT, CWindow::CBSetCUIType}, \
  {"setCUISpeed", OOT_PARAM_TYPE_FLOAT, CWindow::CBSetCUISpeed}, \
  {"setSelectionEnabled", OOT_PARAM_TYPE_BOOL, CWindow::CBSetSelectionEnabled},\
  {"editSelectedObject", OOT_PARAM_TYPE_VOID, CWindow::CBHandleEditSelectedObject}


#define CWINDOW_FIELDS_ENUMS \
  ftFramesPerSecondRequested, \
  ftUpdateTime, \
  ftRenderTime, \
  ftDrawBounds, \
  ftColorDynamicBounds, \
  ftColorChangedBounds, \
  ftCheckBoundChanged, \
  ftDrawClipBound, \
  ftDrawViewportBound, \
  ftDrawNormals, \
  ftDrawCDAOperands, \
  ftBackFaceCullingEnabled, \
  ftClipEnabled, \
  ftBoundCullingEnabled, \
  ftLightingEnabled, \
  ftFogEnabled, \
  ftTextureMappingEnabled, \
  ftCrystalBallEnabled, \
  ftSelectionEnabled

#define CWINDOW_FIELDS_STRINGS \
  "framesPerSecondRequested", \
  "updateTime", \
  "renderTime", \
  "drawBounds", \
  "colorDynamicBounds", \
  "colorChangedBounds", \
  "checkBoundChanged", \
  "drawClipBound", \
  "drawViewportBound", \
  "drawNormals", \
  "drawCDAOperands", \
  "backFaceCullingEnabled", \
  "clipEnabled", \
  "boundCullingEnabled", \
  "lightingEnabled", \
  "fogEnabled", \
  "textureMappingEnabled", \
  "crystalBallEnabled", \
  "selectionEnabled"


class CWindow
{

//////////////////////////////////////////////////
// Enumerated Types
//
public:
  enum EventLevelType 
    {
      eltNone,
      eltAction,
      eltCallBack,
      eltSize
    };

//////////////////////////////////////////////////
// Initialization
//
public:
  CWindow();
  virtual ~CWindow();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcSLIDEID);
  virtual VOID Uninit();

  virtual VOID InitTcl();
  virtual VOID InitTogl();
  virtual VOID Prerender();

  virtual VOID HandleDestroy();

//////////////////////////////////////////////////
// SLIDE Stuff
//
public:
  virtual VOID InsertViewport(CSLIDEViewport *pCSLIDEViewport);
  virtual BOOL CreateTclToplevel();
  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);
  virtual VOID Preprocess();
  virtual VOID Update();

  virtual VOID WriteRIB(CHAR *pcOutputFile, FILE *&fpOutputFile);
  virtual VOID WritePOV(CHAR *pcOutputFile, FILE *&fpOutputFile, 
			CListInt *plistTransformIDs,
			CListInt *plistColors);
  virtual VOID WriteWRLZ(CHAR *pcOutputFile, FILE *&fpOutputFile);

  virtual VOID WriteSTL(CHAR *pcOutputFile, FILE *&fpOutputFile);
  virtual VOID WriteSIF(CHAR *pcOutputFile, FILE *&fpOutputFile);
  virtual VOID WriteLSIF(CHAR *pcOutputFile, FILE *&fpOutputFile);

  virtual VOID WriteSLIDE(FILE *fpOutputFile);
  virtual VOID WriteDotID(FILE *fpOutputFile);
  virtual VOID WriteDotNode(FILE *fpOutputFile);
  virtual VOID WriteDotEdges(FILE *fpOutputFile);

//////////////////////////////////////////////////
// Handlers
//
public:
  virtual VOID HandleReshape(INT_32 iWidth, INT_32 iHeight);
  virtual VOID HandleDisplay();
  virtual VOID HandleMouseButton(INT_32 iButton, INT_32 iState, 
				 INT_32 iX, INT_32 iY);
  virtual VOID HandleMouseMotion(INT_32 iX, INT_32 iY);
  virtual VOID HandleKeyboard(INT_32 iKey, INT_32 iState, 
			      INT_32 iX, INT_32 iY);

  virtual VOID HandleStartDynamics();
  virtual VOID HandlePauseDynamics();

  virtual VOID HandleResetCamera();
  virtual VOID HandleEditSelectedObject();

  virtual VOID PostRedisplay();

//////////////////////////////////////////////////
// Utilities
//
public:
  virtual VOID DisplayControls();

protected:
  VOID DisplayEvent(EventLevelType eEventLevel, CHAR *pcFormat, ...);

public:
  inline VOID WinToVP(INT_32 &iXVP, INT_32 &iYVP, INT_32 iXWin, INT_32 iYWin); 
  inline VOID VPToNorm(FLOAT &fXNorm, FLOAT &fYNorm, INT_32 iXVP, INT_32 iYVP);

//////////////////////////////////////////////////
// CallBack Functions / Handler Dispatchers
//
public:
  static int CreateOotClass(Tcl_Interp *pInterp);
  static int CBHandleCreate(Tcl_Interp *pInterp, OotClass *pClass, 
			    int argc, char **argv);
  static VOID CBHandleDestroy(ClientData pvClientData);

  static VOID CBHandleToglCreate(struct Togl *pTogl);
  static VOID CBHandleToglDisplay(struct Togl *pTogl);
  static VOID CBHandleToglReshape(struct Togl *pTogl);
  static VOID CBHandleToglDestroy(struct Togl *pTogl);

  static int CBHandleMouseButton(Tcl_Interp *pInterp, OotObject *pObject,
				 INT_32 iButton, INT_32 iState,
				 INT_32 iX, INT_32 iY);
  static int CBHandleMouseMotion(Tcl_Interp *pInterp, OotObject *pObject,
				 INT_32 iButton, INT_32 iX, INT_32 iY);
  static int CBHandleKeyboard(Tcl_Interp *pInterp, OotObject *pObject,
			      CHAR *pcKeyCode, INT_32 iState, 
			      INT_32 iX, INT_32 iY);

  static int CBHandleStartDynamics(Tcl_Interp *pInterp, OotObject *pObject);
  static int CBHandlePauseDynamics(Tcl_Interp *pInterp, OotObject *pObject);

  static int CBSetDrawBounds(Tcl_Interp *pInterp, OotObject *pObject, 
			    BOOL b);
  static int CBSetColorDynamicBounds(Tcl_Interp *pInterp, OotObject *pObject, 
				     BOOL b);
  static int CBSetColorChangedBounds(Tcl_Interp *pInterp, OotObject *pObject, 
				     BOOL b);
  static int CBSetCheckBoundChanged(Tcl_Interp *pInterp, OotObject *pObject, 
				    BOOL b);
  static int CBSetDrawClipBound(Tcl_Interp *pInterp, OotObject *pObject, 
				BOOL b);
  static int CBSetDrawViewportBound(Tcl_Interp *pInterp, OotObject *pObject, 
				BOOL b);
  static int CBSetDrawNormals(Tcl_Interp *pInterp, OotObject *pObject, 
				BOOL b);
  static int CBSetDrawCDAOperands(Tcl_Interp *pInterp, OotObject *pObject, 
				  BOOL b);
  static int CBSetBackFaceCullingEnabled(Tcl_Interp *pInterp, OotObject *pObject, 
					 BOOL b);
  static int CBSetClipEnabled(Tcl_Interp *pInterp, OotObject *pObject, 
			      BOOL b);
  static int CBSetBoundCullingEnabled(Tcl_Interp *pInterp, OotObject *pObject, 
				      BOOL b);
  static int CBSetLightingEnabled(Tcl_Interp *pInterp, OotObject *pObject, 
				  BOOL b);
  static int CBSetFogEnabled(Tcl_Interp *pInterp, OotObject *pObject, 
				  BOOL b);
  static int CBSetTextureMappingEnabled(Tcl_Interp *pInterp, OotObject *pObject, 
					BOOL b);
  static int CBSetCrystalBallEnabled(Tcl_Interp *pInterp, OotObject *pObject, 
				     BOOL b);
  static int CBHandleResetCamera(Tcl_Interp *pInterp, OotObject *pObject);
  static int CBSetCUIType(Tcl_Interp *pInterp, OotObject *pObject, 
				     int iType);
  static int CBSetCUISpeed(Tcl_Interp *pInterp, OotObject *pObject, 
			   FLOAT fSpeed);
  static int CBSetSelectionEnabled(Tcl_Interp *pInterp, OotObject *pObject, 
				   BOOL b);
  static int CBHandleEditSelectedObject(Tcl_Interp *pInterp, OotObject *pObject);


//////////////////////////////////////////////////
// Accessor Functions
//
public:
  inline VOID SetTogl(struct Togl *pTogl);

  inline CHAR *GetFileName();
  inline UINT_32 GetFileLine();
  inline UINT_32 GetFileColumn();
  inline CHAR *GetSLIDEID();
  inline VOID SetOrigin(FLOAT fX, FLOAT fY);
  inline VOID SetSize(FLOAT fX, FLOAT fY);
  inline VOID SetBackground(DynVector *pdynvColor);
  inline VOID SetDisplay(CHAR *pcDisplayID);
  inline VOID SetResize(SLIDEOnOffType eResize);
  inline VOID SetAspectRatio(SLIDEAspectRatioType eAspectRatio);

  // UI Stuff
  VOID SetDrawBounds(BOOL b);
  VOID SetColorDynamicBounds(BOOL b);
  VOID SetColorChangedBounds(BOOL b);
  VOID SetCheckBoundChanged(BOOL b);
  VOID SetDrawClipBound(BOOL b);
  VOID SetDrawViewportBound(BOOL b);
  VOID SetDrawNormals(BOOL b);
  VOID SetDrawCDAOperands(BOOL b);
  VOID SetBackFaceCullingEnabled(BOOL b);
  VOID SetClipEnabled(BOOL b);
  VOID SetBoundCullingEnabled(BOOL b);
  VOID SetLightingEnabled(BOOL b);
  VOID SetFogEnabled(BOOL b);
  VOID SetTextureMappingEnabled(BOOL b);
  VOID SetCrystalBallEnabled(BOOL b);
  VOID SetCUIType(DynInt *pDyniType);
  VOID SetCUISpeed(DynFloat *pDynfSpeed);
  VOID SetSelectionEnabled(BOOL b);

  VOID SetSecondsPerFrameRequested(DOUBLE dSecondsPerFrame);
  VOID SetUpdateTime(DOUBLE dUpdateTime);
  VOID SetRenderTime(DOUBLE dRenderTime);

//////////////////////////////////////////////////
// Class Accessor Functions
//
public:

//////////////////////////////////////////////////
// Class Variables
//
protected:
  // Table for decoding Tcl key names in Keyboard Callback
  static CHashString *m_pHashKeyboard;

  // OOT registration variables
  enum FieldType 
    {
      CWINDOW_FIELDS_ENUMS,

      ftSize
    };

  static Tcl_Interp *m_pInterp;

  static const int m_cuiOotMethods;
  //static OotMethodEntry m_pOotMethods[CWindow::m_cuiOotMethods];
  static OotMethodEntry m_pOotMethods[cuiCWindowOotMethods];

  static const CHAR *m_cppcOotFields[ftSize];
  static const CHAR *m_cpcOotClassName;
  static OotClass *m_pOotClass;
  
  // Object Counter Variable
  static UINT_32 m_uiOotCounter;

  static BOOL m_bTclWindowExists;

//////////////////////////////////////////////////
// Member Variables
//
protected:
  // Tcl and Togl variables for Windowing
  UINT_32 m_uiOotID;
  CHAR *m_pcOotName;
  OotObject *m_pOotObject;
  struct Togl *m_pTogl;
  BOOL m_bRedisplayPending;

  INT_32 m_iWidth;
  INT_32 m_iHeight;

  // For Debugging Print Statements
  BOOL m_bDisplayEvents;        
  EventLevelType m_eEventLevel;

  // The SLIDE World which we are rendering
  CHAR *m_pcFile;
  UINT_32 m_uiLine;
  UINT_32 m_uiColumn;
  CHAR *m_pcSLIDEID;
  Vector m_vOrigin;
  Vector m_vSize;
  DynVector m_vBackground;
  CHAR *m_pcDisplayID;

  SLIDEOnOffType m_eResize;
  SLIDEAspectRatioType m_eAspectRatio;
  
  CSLIDEWorld *m_pCSLIDEWorld;
  CListInt m_listCSLIDEViewport;  
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//


inline VOID CWindow::WinToVP(INT_32 &iXVP, INT_32 &iYVP, INT_32 iXWin, INT_32 iYWin)
{
  iXVP = iXWin;
  iYVP = m_iHeight - iYWin;
}

inline VOID CWindow::VPToNorm(FLOAT &fXNorm, FLOAT &fYNorm, 
                              INT_32 iXVP, INT_32 iYVP)
{
  fXNorm = (FLOAT)iXVP / (FLOAT)(m_iWidth);
  fYNorm = (FLOAT)iYVP / (FLOAT)(m_iHeight);
}

inline VOID CWindow::SetTogl(struct Togl *pTogl)
{
  ASSERT( pTogl != NULL );
  m_pTogl = pTogl;
}

inline CHAR *CWindow::GetFileName()
{
  return m_pcFile;
}

inline UINT_32 CWindow::GetFileLine()
{
  return m_uiLine;
}

inline UINT_32 CWindow::GetFileColumn()
{
  return m_uiColumn;
}

inline CHAR *CWindow::GetSLIDEID()
{
  return m_pcSLIDEID;
}

inline VOID CWindow::SetOrigin(FLOAT fX, FLOAT fY)
{
  m_vOrigin.Set(fX, fY, 0.0);
}

inline VOID CWindow::SetSize(FLOAT fX, FLOAT fY)
{
  m_vSize.Set(fX, fY, 0.0);
}

inline VOID CWindow::SetBackground(DynVector *pdynvColor)
{
  m_vBackground.SetAll(pdynvColor);
}

inline VOID CWindow::SetDisplay(CHAR *pcDisplayID)
{
  m_pcDisplayID = pcDisplayID;
}

inline VOID CWindow::SetResize(SLIDEOnOffType eResize)
{
  m_eResize = eResize;
}

inline VOID CWindow::SetAspectRatio(SLIDEAspectRatioType eAspectRatio)
{
  m_eAspectRatio = eAspectRatio;
}

//////////////////////////////////////////////////////////////////////
// Globals
//

#endif __WINDOW_H__




