#include "apm2d.h"
#include <stdio.h>

#include "predicates.h"

#include "apm2d_adaptive.h"
#include "apm2d_mpfun.h"
#include "apm2d_mpfunfilter.h"
#include "apm2d_mpfunnorm.h"

//#define ENABLE_DEBUGGING_OUTPUT
//#define DEBUG_EPSILON_VALUES
#ifdef DEBUG_EPSILON_VALUES
static const DOUBLE s_cdOrderOfMagnitude = 1e1;
#endif // DEBUG_EPSILON_VALUES

//////////////////////////////////////////////////////////////////////
// APM2D Private Methods 
//   implemented as static functions to keep them out of the apm2d.h file

static INT_32 DistanceCompareV_VVLocal(DOUBLE dV0_x, DOUBLE dV0_y,
				       DOUBLE dV1_x, DOUBLE dV1_y, 
				       DOUBLE dV2_x, DOUBLE dV2_y);

static INT_32 DistanceCompareV_VLLocal(DOUBLE dV0_x, DOUBLE dV0_y,
				       DOUBLE dV1_x, DOUBLE dV1_y, 
				       DOUBLE dE20_x, DOUBLE dE20_y, 
				       DOUBLE dE21_x, DOUBLE dE21_y);

static INT_32 DistanceCompareV_LLLocal(DOUBLE dV0_x, DOUBLE dV0_y,
				       DOUBLE dE10_x, DOUBLE dE10_y, 
				       DOUBLE dE11_x, DOUBLE dE11_y,
				       DOUBLE dE20_x, DOUBLE dE20_y, 
				       DOUBLE dE21_x, DOUBLE dE21_y);
static INT_32 DistanceCompareV_LLAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
					  DOUBLE dE10_x, DOUBLE dE10_y, 
					  DOUBLE dE11_x, DOUBLE dE11_y,
					  DOUBLE dE20_x, DOUBLE dE20_y, 
					  DOUBLE dE21_x, DOUBLE dE21_y);

static INT_32 DistanceCompareEvv_VVLocal(DOUBLE dE0A_x, DOUBLE dE0A_y,
					 DOUBLE dE0B_x, DOUBLE dE0B_y,
					 DOUBLE dV1_x, DOUBLE dV1_y, 
					 DOUBLE dV2_x, DOUBLE dV2_y);
static INT_32 DistanceCompareEvv_VVAdaptive(DOUBLE dE0A_x, DOUBLE dE0A_y,
					    DOUBLE dE0B_x, DOUBLE dE0B_y,
					    DOUBLE dV1_x, DOUBLE dV1_y, 
					    DOUBLE dV2_x, DOUBLE dV2_y);

static INT_32 DistanceCompareEvv_VLLocal(DOUBLE dE0A_x, DOUBLE dE0A_y,
					 DOUBLE dE0B_x, DOUBLE dE0B_y,
					 DOUBLE dV1_x, DOUBLE dV1_y, 
					 DOUBLE dE20_x, DOUBLE dE20_y, 
					 DOUBLE dE21_x, DOUBLE dE21_y);
static INT_32 DistanceCompareEvv_VLAdaptive(DOUBLE dE0A_x, DOUBLE dE0A_y,
					    DOUBLE dE0B_x, DOUBLE dE0B_y,
					    DOUBLE dV1_x, DOUBLE dV1_y, 
					    DOUBLE dE20_x, DOUBLE dE20_y, 
					    DOUBLE dE21_x, DOUBLE dE21_y);

static INT_32 DistanceCompareEvv_LLLocal(DOUBLE dE0A_x, DOUBLE dE0A_y,
					 DOUBLE dE0B_x, DOUBLE dE0B_y,
					 DOUBLE dE10_x, DOUBLE dE10_y, 
					 DOUBLE dE11_x, DOUBLE dE11_y,
					 DOUBLE dE20_x, DOUBLE dE20_y, 
					 DOUBLE dE21_x, DOUBLE dE21_y);
static INT_32 DistanceCompareEvv_LLAdaptive(DOUBLE dE0A_x, DOUBLE dE0A_y,
					    DOUBLE dE0B_x, DOUBLE dE0B_y,
					    DOUBLE dE10_x, DOUBLE dE10_y, 
					    DOUBLE dE11_x, DOUBLE dE11_y,
					    DOUBLE dE20_x, DOUBLE dE20_y, 
					    DOUBLE dE21_x, DOUBLE dE21_y);

static INT_32 DistanceCompareElv_VVLocal(DOUBLE dE00_x, DOUBLE dE00_y,
					 DOUBLE dE01_x, DOUBLE dE01_y,
					 DOUBLE dE0B_x, DOUBLE dE0B_y,
					 DOUBLE dV1_x, DOUBLE dV1_y, 
					 DOUBLE dV2_x, DOUBLE dV2_y);
static INT_32 DistanceCompareElv_VVAdaptive(DOUBLE dE00_x, DOUBLE dE00_y,
					    DOUBLE dE01_x, DOUBLE dE01_y,
					    DOUBLE dE0B_x, DOUBLE dE0B_y,
					    DOUBLE dV1_x, DOUBLE dV1_y, 
					    DOUBLE dV2_x, DOUBLE dV2_y);

static INT_32 DistanceCompareElv_VLLocal(DOUBLE dE00_x, DOUBLE dE00_y,
					 DOUBLE dE01_x, DOUBLE dE01_y,
					 DOUBLE dE0B_x, DOUBLE dE0B_y,
					 DOUBLE dV1_x, DOUBLE dV1_y, 
					 DOUBLE dE20_x, DOUBLE dE20_y, 
					 DOUBLE dE21_x, DOUBLE dE21_y);
static INT_32 DistanceCompareElv_VLAdaptive(DOUBLE dE00_x, DOUBLE dE00_y,
					    DOUBLE dE01_x, DOUBLE dE01_y,
					    DOUBLE dE0B_x, DOUBLE dE0B_y,
					    DOUBLE dV1_x, DOUBLE dV1_y, 
					    DOUBLE dE20_x, DOUBLE dE20_y, 
					    DOUBLE dE21_x, DOUBLE dE21_y);

static INT_32 DistanceCompareL_VVLocal(DOUBLE dE00_x, DOUBLE dE00_y,
				       DOUBLE dE01_x, DOUBLE dE01_y,
				       DOUBLE dV1_x, DOUBLE dV1_y, 
				       DOUBLE dV2_x, DOUBLE dV2_y);
static INT_32 DistanceCompareL_VVAdaptive(DOUBLE dE00_x, DOUBLE dE00_y,
					  DOUBLE dE01_x, DOUBLE dE01_y,
					  DOUBLE dV1_x, DOUBLE dV1_y, 
					  DOUBLE dV2_x, DOUBLE dV2_y);

//////////////////////////////////////////////////////////////////////
// EE Orientaion Functions
//
static INT_32 OrientationExELocal(DOUBLE dE00_x, DOUBLE dE00_y,
				  DOUBLE dE01_x, DOUBLE dE01_y,
				  DOUBLE dE10_x, DOUBLE dE10_y,
				  DOUBLE dE11_x, DOUBLE dE11_y);
static INT_32 OrientationEdotELocal(DOUBLE dE00_x, DOUBLE dE00_y,
				    DOUBLE dE01_x, DOUBLE dE01_y,
				    DOUBLE dE10_x, DOUBLE dE10_y,
				    DOUBLE dE11_x, DOUBLE dE11_y);

//////////////////////////////////////////////////
// Inscribed Circle Center Functions
//
static APM2D::InscribedCircleCenterType 
InscribedCircleCenterVVVLocal(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			      DOUBLE dV00_x, DOUBLE dV00_y,
			      DOUBLE dV0_x, DOUBLE dV0_y,
			      DOUBLE dV01_x, DOUBLE dV01_y,
			      DOUBLE dV10_x, DOUBLE dV10_y,
			      DOUBLE dV1_x, DOUBLE dV1_y,
			      DOUBLE dV11_x, DOUBLE dV11_y,
			      DOUBLE dV20_x, DOUBLE dV20_y,
			      DOUBLE dV2_x, DOUBLE dV2_y,
			      DOUBLE dV21_x, DOUBLE dV21_y);
static APM2D::InscribedCircleCenterType 
InscribedCircleCenterVVVAdaptive(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				 DOUBLE dV00_x, DOUBLE dV00_y,
				 DOUBLE dV0_x, DOUBLE dV0_y,
				 DOUBLE dV01_x, DOUBLE dV01_y,
				 DOUBLE dV10_x, DOUBLE dV10_y,
				 DOUBLE dV1_x, DOUBLE dV1_y,
				 DOUBLE dV11_x, DOUBLE dV11_y,
				 DOUBLE dV20_x, DOUBLE dV20_y,
				 DOUBLE dV2_x, DOUBLE dV2_y,
				 DOUBLE dV21_x, DOUBLE dV21_y);
static APM2D::EdgeZoneType EdgeZoneVVVLocal(DOUBLE dV0_x, DOUBLE dV0_y,
					    DOUBLE dV1_x, DOUBLE dV1_y,
					    DOUBLE dV2_x, DOUBLE dV2_y,
					    DOUBLE dE30_x, DOUBLE dE30_y,
					    DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::EdgeZoneType EdgeZoneVVVAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
					       DOUBLE dV1_x, DOUBLE dV1_y,
					       DOUBLE dV2_x, DOUBLE dV2_y,
					       DOUBLE dE30_x, DOUBLE dE30_y,
					       DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::CircleTestType InCircleVVV_VLocal(DOUBLE dV0_x, DOUBLE dV0_y,
						DOUBLE dV1_x, DOUBLE dV1_y,
						DOUBLE dV2_x, DOUBLE dV2_y,
						DOUBLE dV3_x, DOUBLE dV3_y);
static APM2D::CircleTestType InCircleVVV_VAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dV1_x, DOUBLE dV1_y,
						   DOUBLE dV2_x, DOUBLE dV2_y,
						   DOUBLE dV3_x, DOUBLE dV3_y);
static APM2D::CircleTestType InCircleVVV_ELocal(DOUBLE dV0_x, DOUBLE dV0_y,
						DOUBLE dV1_x, DOUBLE dV1_y,
						DOUBLE dV2_x, DOUBLE dV2_y,
						DOUBLE dE30_x, DOUBLE dE30_y,
						DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::CircleTestType InCircleVVV_EAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dV1_x, DOUBLE dV1_y,
						   DOUBLE dV2_x, DOUBLE dV2_y,
						   DOUBLE dE30_x, DOUBLE dE30_y,
						   DOUBLE dE31_x, DOUBLE dE31_y);

static APM2D::InscribedCircleCenterType 
InscribedCircleCenterVVELocal(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			      DOUBLE dV00_x, DOUBLE dV00_y,
			      DOUBLE dV0_x, DOUBLE dV0_y,
			      DOUBLE dV01_x, DOUBLE dV01_y,
			      DOUBLE dV10_x, DOUBLE dV10_y,
			      DOUBLE dV1_x, DOUBLE dV1_y,
			      DOUBLE dV11_x, DOUBLE dV11_y,
			      DOUBLE dE20_x, DOUBLE dE20_y, 
			      DOUBLE dE21_x, DOUBLE dE21_y);
static APM2D::InscribedCircleCenterType 
InscribedCircleCenterVVEAdaptive(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				 DOUBLE dV00_x, DOUBLE dV00_y,
				 DOUBLE dV0_x, DOUBLE dV0_y,
				 DOUBLE dV01_x, DOUBLE dV01_y,
				 DOUBLE dV10_x, DOUBLE dV10_y,
				 DOUBLE dV1_x, DOUBLE dV1_y,
				 DOUBLE dV11_x, DOUBLE dV11_y,
				 DOUBLE dE20_x, DOUBLE dE20_y, 
				 DOUBLE dE21_x, DOUBLE dE21_y);
static APM2D::EdgeZoneType EdgeZoneVVELocal(DOUBLE dV0_x, DOUBLE dV0_y,
					    DOUBLE dV1_x, DOUBLE dV1_y,
					    DOUBLE dE20_x, DOUBLE dE20_y, 
					    DOUBLE dE21_x, DOUBLE dE21_y,
					    DOUBLE dE30_x, DOUBLE dE30_y,
					    DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::EdgeZoneType EdgeZoneVVEAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
					       DOUBLE dV1_x, DOUBLE dV1_y,
					       DOUBLE dE20_x, DOUBLE dE20_y, 
					       DOUBLE dE21_x, DOUBLE dE21_y,
					       DOUBLE dE30_x, DOUBLE dE30_y,
					       DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::CircleTestType InCircleVVE_VLocal(DOUBLE dV0_x, DOUBLE dV0_y,
						DOUBLE dV1_x, DOUBLE dV1_y,
						DOUBLE dE20_x, DOUBLE dE20_y, 
						DOUBLE dE21_x, DOUBLE dE21_y,
						DOUBLE dV3_x, DOUBLE dV3_y);
static APM2D::CircleTestType InCircleVVE_VAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dV1_x, DOUBLE dV1_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   DOUBLE dV3_x, DOUBLE dV3_y);
static APM2D::CircleTestType InCircleVVE_ELocal(DOUBLE dV0_x, DOUBLE dV0_y,
						DOUBLE dV1_x, DOUBLE dV1_y,
						DOUBLE dE20_x, DOUBLE dE20_y, 
						DOUBLE dE21_x, DOUBLE dE21_y,
						DOUBLE dE30_x, DOUBLE dE30_y,
						DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::CircleTestType InCircleVVE_EAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dV1_x, DOUBLE dV1_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   DOUBLE dE30_x, DOUBLE dE30_y,
						   DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::InscribedCircleCenterType 
InscribedCircleCenterVVEAntiparallelLocal(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					  DOUBLE dV00_x, DOUBLE dV00_y,
					  DOUBLE dV0_x, DOUBLE dV0_y,
					  DOUBLE dV01_x, DOUBLE dV01_y,
					  DOUBLE dV10_x, DOUBLE dV10_y,
					  DOUBLE dV1_x, DOUBLE dV1_y,
					  DOUBLE dV11_x, DOUBLE dV11_y,
					  DOUBLE dE20_x, DOUBLE dE20_y, 
					  DOUBLE dE21_x, DOUBLE dE21_y);
static APM2D::InscribedCircleCenterType 
InscribedCircleCenterVVEAntiparallelAdaptive(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					     DOUBLE dV00_x, DOUBLE dV00_y,
					     DOUBLE dV0_x, DOUBLE dV0_y,
					     DOUBLE dV01_x, DOUBLE dV01_y,
					     DOUBLE dV10_x, DOUBLE dV10_y,
					     DOUBLE dV1_x, DOUBLE dV1_y,
					     DOUBLE dV11_x, DOUBLE dV11_y,
					     DOUBLE dE20_x, DOUBLE dE20_y, 
					     DOUBLE dE21_x, DOUBLE dE21_y);
static APM2D::EdgeZoneType EdgeZoneVVEAntiparallelLocal(DOUBLE dV0_x, DOUBLE dV0_y,
							DOUBLE dV1_x, DOUBLE dV1_y,
							DOUBLE dE20_x, DOUBLE dE20_y, 
							DOUBLE dE21_x, DOUBLE dE21_y,
							DOUBLE dE30_x, DOUBLE dE30_y,
							DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::EdgeZoneType EdgeZoneVVEAntiparallelAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
							   DOUBLE dV1_x, DOUBLE dV1_y,
							   DOUBLE dE20_x, DOUBLE dE20_y, 
							   DOUBLE dE21_x, DOUBLE dE21_y,
							   DOUBLE dE30_x, DOUBLE dE30_y,
							   DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::CircleTestType InCircleVVE_VAntiparallelLocal(DOUBLE dV0_x, DOUBLE dV0_y,
							    DOUBLE dV1_x, DOUBLE dV1_y,
							    DOUBLE dE20_x, DOUBLE dE20_y, 
							    DOUBLE dE21_x, DOUBLE dE21_y,
							    DOUBLE dV3_x, DOUBLE dV3_y);
static APM2D::CircleTestType InCircleVVE_VAntiparallelAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
							       DOUBLE dV1_x, DOUBLE dV1_y,
							       DOUBLE dE20_x, DOUBLE dE20_y, 
							       DOUBLE dE21_x, DOUBLE dE21_y,
							       DOUBLE dV3_x, DOUBLE dV3_y);
static APM2D::CircleTestType InCircleVVE_EAntiparallelLocal(DOUBLE dV0_x, DOUBLE dV0_y,
							    DOUBLE dV1_x, DOUBLE dV1_y,
							    DOUBLE dE20_x, DOUBLE dE20_y, 
							    DOUBLE dE21_x, DOUBLE dE21_y,
							    DOUBLE dE30_x, DOUBLE dE30_y,
							    DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::CircleTestType InCircleVVE_EAntiparallelAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
							       DOUBLE dV1_x, DOUBLE dV1_y,
							       DOUBLE dE20_x, DOUBLE dE20_y, 
							       DOUBLE dE21_x, DOUBLE dE21_y,
							       DOUBLE dE30_x, DOUBLE dE30_y,
							       DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::InscribedCircleCenterType 
InscribedCircleCenterVvELocal(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			      DOUBLE dV00_x, DOUBLE dV00_y,
			      DOUBLE dV0_x, DOUBLE dV0_y,
			      DOUBLE dV01_x, DOUBLE dV01_y,
			      DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
			      DOUBLE dE21_x, DOUBLE dE21_y);
static APM2D::InscribedCircleCenterType 
InscribedCircleCenterVvEAdaptive(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				 DOUBLE dV00_x, DOUBLE dV00_y,
				 DOUBLE dV0_x, DOUBLE dV0_y,
				 DOUBLE dV01_x, DOUBLE dV01_y,
				 DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
				 DOUBLE dE21_x, DOUBLE dE21_y);
static APM2D::EdgeZoneType EdgeZoneVvELocal(DOUBLE dV0_x, DOUBLE dV0_y,
					    DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
					    DOUBLE dE21_x, DOUBLE dE21_y,
					    DOUBLE dE30_x, DOUBLE dE30_y,
					    DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::EdgeZoneType EdgeZoneVvEAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
					       DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
					       DOUBLE dE21_x, DOUBLE dE21_y,
					       DOUBLE dE30_x, DOUBLE dE30_y,
					       DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::CircleTestType InCircleVvE_VLocal(DOUBLE dV0_x, DOUBLE dV0_y,
						DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
						DOUBLE dE21_x, DOUBLE dE21_y,
						DOUBLE dV3_x, DOUBLE dV3_y);
static APM2D::CircleTestType InCircleVvE_VAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   DOUBLE dV3_x, DOUBLE dV3_y);
static APM2D::CircleTestType InCircleVvE_ELocal(DOUBLE dV0_x, DOUBLE dV0_y,
						DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
						DOUBLE dE21_x, DOUBLE dE21_y,
						DOUBLE dE30_x, DOUBLE dE30_y,
						DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::CircleTestType InCircleVvE_EAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   DOUBLE dE30_x, DOUBLE dE30_y,
						   DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::InscribedCircleCenterType 
InscribedCircleCenterEvVLocal(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			      DOUBLE dE00_x, DOUBLE dE00_y,
			      DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
			      DOUBLE dV20_x, DOUBLE dV20_y,
			      DOUBLE dV2_x, DOUBLE dV2_y,
			      DOUBLE dV21_x, DOUBLE dV21_y);
static APM2D::InscribedCircleCenterType 
InscribedCircleCenterEvVAdaptive(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				 DOUBLE dE00_x, DOUBLE dE00_y,
				 DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
				 DOUBLE dV20_x, DOUBLE dV20_y,
				 DOUBLE dV2_x, DOUBLE dV2_y,
				 DOUBLE dV21_x, DOUBLE dV21_y);
static APM2D::EdgeZoneType EdgeZoneEvVLocal(DOUBLE dE00_x, DOUBLE dE00_y,
					    DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
					    DOUBLE dV2_x, DOUBLE dV2_y,
					    DOUBLE dE30_x, DOUBLE dE30_y,
					    DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::EdgeZoneType EdgeZoneEvVAdaptive(DOUBLE dE00_x, DOUBLE dE00_y,
					       DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
					       DOUBLE dV2_x, DOUBLE dV2_y,
					       DOUBLE dE30_x, DOUBLE dE30_y,
					       DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::CircleTestType InCircleEvV_VLocal(DOUBLE dE00_x, DOUBLE dE00_y,
						DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
						DOUBLE dV2_x, DOUBLE dV2_y,
						DOUBLE dV3_x, DOUBLE dV3_y);
static APM2D::CircleTestType InCircleEvV_VAdaptive(DOUBLE dE00_x, DOUBLE dE00_y,
						   DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
						   DOUBLE dV2_x, DOUBLE dV2_y,
						   DOUBLE dV3_x, DOUBLE dV3_y);
static APM2D::CircleTestType InCircleEvV_ELocal(DOUBLE dE00_x, DOUBLE dE00_y,
						DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
						DOUBLE dV2_x, DOUBLE dV2_y,
						DOUBLE dE30_x, DOUBLE dE30_y,
						DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::CircleTestType InCircleEvV_EAdaptive(DOUBLE dE00_x, DOUBLE dE00_y,
						   DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
						   DOUBLE dV2_x, DOUBLE dV2_y,
						   DOUBLE dE30_x, DOUBLE dE30_y,
						   DOUBLE dE31_x, DOUBLE dE31_y);

static APM2D::InscribedCircleCenterType 
InscribedCircleCenterVEELocal(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			      DOUBLE dV00_x, DOUBLE dV00_y,
			      DOUBLE dV0_x, DOUBLE dV0_y,
			      DOUBLE dV01_x, DOUBLE dV01_y,
			      DOUBLE dE10_x, DOUBLE dE10_y, 
			      DOUBLE dE11_x, DOUBLE dE11_y,
			      DOUBLE dE20_x, DOUBLE dE20_y, 
			      DOUBLE dE21_x, DOUBLE dE21_y);
static APM2D::InscribedCircleCenterType 
InscribedCircleCenterVEEAdaptive(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				 DOUBLE dV00_x, DOUBLE dV00_y,
				 DOUBLE dV0_x, DOUBLE dV0_y,
				 DOUBLE dV01_x, DOUBLE dV01_y,
				 DOUBLE dE10_x, DOUBLE dE10_y, 
				 DOUBLE dE11_x, DOUBLE dE11_y,
				 DOUBLE dE20_x, DOUBLE dE20_y, 
				 DOUBLE dE21_x, DOUBLE dE21_y);
static APM2D::EdgeZoneType EdgeZoneVEELocal(DOUBLE dV0_x, DOUBLE dV0_y,
					    DOUBLE dE10_x, DOUBLE dE10_y, 
					    DOUBLE dE11_x, DOUBLE dE11_y,
					    DOUBLE dE20_x, DOUBLE dE20_y, 
					    DOUBLE dE21_x, DOUBLE dE21_y,
					    DOUBLE dE30_x, DOUBLE dE30_y,
					    DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::EdgeZoneType EdgeZoneVEEAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
					       DOUBLE dE10_x, DOUBLE dE10_y, 
					       DOUBLE dE11_x, DOUBLE dE11_y,
					       DOUBLE dE20_x, DOUBLE dE20_y, 
					       DOUBLE dE21_x, DOUBLE dE21_y,
					       DOUBLE dE30_x, DOUBLE dE30_y,
					       DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::CircleTestType InCircleVEE_VLocal(DOUBLE dV0_x, DOUBLE dV0_y,
						DOUBLE dE10_x, DOUBLE dE10_y, 
						DOUBLE dE11_x, DOUBLE dE11_y,
						DOUBLE dE20_x, DOUBLE dE20_y, 
						DOUBLE dE21_x, DOUBLE dE21_y,
						DOUBLE dV3_x, DOUBLE dV3_y);
static APM2D::CircleTestType InCircleVEE_VAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dE10_x, DOUBLE dE10_y, 
						   DOUBLE dE11_x, DOUBLE dE11_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   DOUBLE dV3_x, DOUBLE dV3_y);
static APM2D::CircleTestType InCircleVEE_ELocal(DOUBLE dV0_x, DOUBLE dV0_y,
						DOUBLE dE10_x, DOUBLE dE10_y, 
						DOUBLE dE11_x, DOUBLE dE11_y,
						DOUBLE dE20_x, DOUBLE dE20_y, 
						DOUBLE dE21_x, DOUBLE dE21_y,
						DOUBLE dE30_x, DOUBLE dE30_y,
						DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::CircleTestType InCircleVEE_EAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dE10_x, DOUBLE dE10_y, 
						   DOUBLE dE11_x, DOUBLE dE11_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   DOUBLE dE30_x, DOUBLE dE30_y,
						   DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::InscribedCircleCenterType 
InscribedCircleCenterVEEAntiparallelLocal(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					  DOUBLE dV00_x, DOUBLE dV00_y,
					  DOUBLE dV0_x, DOUBLE dV0_y,
					  DOUBLE dV01_x, DOUBLE dV01_y,
					  DOUBLE dE10_x, DOUBLE dE10_y, 
					  DOUBLE dE11_x, DOUBLE dE11_y,
					  DOUBLE dE20_x, DOUBLE dE20_y, 
					  DOUBLE dE21_x, DOUBLE dE21_y);
static APM2D::InscribedCircleCenterType 
InscribedCircleCenterVEEAntiparallelAdaptive(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					     DOUBLE dV00_x, DOUBLE dV00_y,
					     DOUBLE dV0_x, DOUBLE dV0_y,
					     DOUBLE dV01_x, DOUBLE dV01_y,
					     DOUBLE dE10_x, DOUBLE dE10_y, 
					     DOUBLE dE11_x, DOUBLE dE11_y,
					     DOUBLE dE20_x, DOUBLE dE20_y, 
					     DOUBLE dE21_x, DOUBLE dE21_y);
static APM2D::EdgeZoneType EdgeZoneVEEAntiparallelLocal(DOUBLE dV0_x, DOUBLE dV0_y,
							DOUBLE dE10_x, DOUBLE dE10_y, 
							DOUBLE dE11_x, DOUBLE dE11_y,
							DOUBLE dE20_x, DOUBLE dE20_y, 
							DOUBLE dE21_x, DOUBLE dE21_y,
							DOUBLE dE30_x, DOUBLE dE30_y,
							DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::EdgeZoneType EdgeZoneVEEAntiparallelAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
							   DOUBLE dE10_x, DOUBLE dE10_y, 
							   DOUBLE dE11_x, DOUBLE dE11_y,
							   DOUBLE dE20_x, DOUBLE dE20_y, 
							   DOUBLE dE21_x, DOUBLE dE21_y,
							   DOUBLE dE30_x, DOUBLE dE30_y,
							   DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::CircleTestType InCircleVEE_VAntiparallelLocal(DOUBLE dV0_x, DOUBLE dV0_y,
							    DOUBLE dE10_x, DOUBLE dE10_y, 
							    DOUBLE dE11_x, DOUBLE dE11_y,
							    DOUBLE dE20_x, DOUBLE dE20_y, 
							    DOUBLE dE21_x, DOUBLE dE21_y,
							    DOUBLE dV3_x, DOUBLE dV3_y);
static APM2D::CircleTestType InCircleVEE_VAntiparallelAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
							       DOUBLE dE10_x, DOUBLE dE10_y,
							       DOUBLE dE11_x, DOUBLE dE11_y,
							       DOUBLE dE20_x, DOUBLE dE20_y,
							       DOUBLE dE21_x, DOUBLE dE21_y,
							       DOUBLE dV3_x, DOUBLE dV3_y);
static APM2D::CircleTestType InCircleVEE_EAntiparallelLocal(DOUBLE dV0_x, DOUBLE dV0_y,
							    DOUBLE dE10_x, DOUBLE dE10_y, 
							    DOUBLE dE11_x, DOUBLE dE11_y,
							    DOUBLE dE20_x, DOUBLE dE20_y, 
							    DOUBLE dE21_x, DOUBLE dE21_y,
							    DOUBLE dE30_x, DOUBLE dE30_y,
							    DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::CircleTestType InCircleVEE_EAntiparallelAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
							       DOUBLE dE10_x, DOUBLE dE10_y,
							       DOUBLE dE11_x, DOUBLE dE11_y,
							       DOUBLE dE20_x, DOUBLE dE20_y,
							       DOUBLE dE21_x, DOUBLE dE21_y,
							       DOUBLE dE30_x, DOUBLE dE30_y,
							       DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::InscribedCircleCenterType 
InscribedCircleCentervEELocal(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			      DOUBLE dv0_E10_x, DOUBLE dV0_E10_y, 
			      DOUBLE dE11_x, DOUBLE dE11_y,
			      DOUBLE dE20_x, DOUBLE dE20_y, 
			      DOUBLE dE21_x, DOUBLE dE21_y);
static APM2D::InscribedCircleCenterType 
InscribedCircleCentervEEAdaptive(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				 DOUBLE dv0_E10_x, DOUBLE dV0_E10_y, 
				 DOUBLE dE11_x, DOUBLE dE11_y,
				 DOUBLE dE20_x, DOUBLE dE20_y, 
				 DOUBLE dE21_x, DOUBLE dE21_y);
static APM2D::EdgeZoneType EdgeZonevEELocal(DOUBLE dv0_E10_x, DOUBLE dV0_E10_y, 
					    DOUBLE dE11_x, DOUBLE dE11_y,
					    DOUBLE dE20_x, DOUBLE dE20_y, 
					    DOUBLE dE21_x, DOUBLE dE21_y,
					    DOUBLE dE30_x, DOUBLE dE30_y,
					    DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::EdgeZoneType EdgeZonevEEAdaptive(DOUBLE dv0_E10_x, DOUBLE dV0_E10_y, 
					       DOUBLE dE11_x, DOUBLE dE11_y,
					       DOUBLE dE20_x, DOUBLE dE20_y, 
					       DOUBLE dE21_x, DOUBLE dE21_y,
					       DOUBLE dE30_x, DOUBLE dE30_y,
					       DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::CircleTestType InCirclevEE_VLocal(DOUBLE dv0_E10_x, DOUBLE dV0_E10_y, 
						DOUBLE dE11_x, DOUBLE dE11_y,
						DOUBLE dE20_x, DOUBLE dE20_y, 
						DOUBLE dE21_x, DOUBLE dE21_y,
						DOUBLE dV3_x, DOUBLE dV3_y);
static APM2D::CircleTestType InCirclevEE_VAdaptive(DOUBLE dv0_E10_x, DOUBLE dV0_E10_y, 
						   DOUBLE dE11_x, DOUBLE dE11_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   DOUBLE dV3_x, DOUBLE dV3_y);
static APM2D::CircleTestType InCirclevEE_ELocal(DOUBLE dv0_E10_x, DOUBLE dV0_E10_y, 
						DOUBLE dE11_x, DOUBLE dE11_y,
						DOUBLE dE20_x, DOUBLE dE20_y, 
						DOUBLE dE21_x, DOUBLE dE21_y,
						DOUBLE dE30_x, DOUBLE dE30_y,
						DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::CircleTestType InCirclevEE_EAdaptive(DOUBLE dv0_E10_x, DOUBLE dV0_E10_y, 
						   DOUBLE dE11_x, DOUBLE dE11_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   DOUBLE dE30_x, DOUBLE dE30_y,
						   DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::InscribedCircleCenterType 
InscribedCircleCenterEEvLocal(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			      DOUBLE dE00_x, DOUBLE dE00_y, 
			      DOUBLE dE01_x, DOUBLE dE01_y,
			      DOUBLE dE10_x, DOUBLE dE10_y, 
			      DOUBLE dE11_V2_x, DOUBLE dE11_V2_y);
static APM2D::InscribedCircleCenterType 
InscribedCircleCenterEEvAdaptive(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				 DOUBLE dE00_x, DOUBLE dE00_y, 
				 DOUBLE dE01_x, DOUBLE dE01_y,
				 DOUBLE dE10_x, DOUBLE dE10_y, 
				 DOUBLE dE11_V2_x, DOUBLE dE11_V2_y);
static APM2D::EdgeZoneType EdgeZoneEEvLocal(DOUBLE dE00_x, DOUBLE dE00_y, 
					    DOUBLE dE01_x, DOUBLE dE01_y,
					    DOUBLE dE10_x, DOUBLE dE10_y, 
					    DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
					    DOUBLE dE30_x, DOUBLE dE30_y,
					    DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::EdgeZoneType EdgeZoneEEvAdaptive(DOUBLE dE00_x, DOUBLE dE00_y, 
					       DOUBLE dE01_x, DOUBLE dE01_y,
					       DOUBLE dE10_x, DOUBLE dE10_y, 
					       DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
					       DOUBLE dE30_x, DOUBLE dE30_y,
					       DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::CircleTestType InCircleEEv_VLocal(DOUBLE dE00_x, DOUBLE dE00_y, 
						DOUBLE dE01_x, DOUBLE dE01_y,
						DOUBLE dE10_x, DOUBLE dE10_y, 
						DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
						DOUBLE dV3_x, DOUBLE dV3_y);
static APM2D::CircleTestType InCircleEEv_VAdaptive(DOUBLE dE00_x, DOUBLE dE00_y, 
						   DOUBLE dE01_x, DOUBLE dE01_y,
						   DOUBLE dE10_x, DOUBLE dE10_y, 
						   DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
						   DOUBLE dV3_x, DOUBLE dV3_y);
static APM2D::CircleTestType InCircleEEv_ELocal(DOUBLE dE00_x, DOUBLE dE00_y, 
						DOUBLE dE01_x, DOUBLE dE01_y,
						DOUBLE dE10_x, DOUBLE dE10_y, 
						DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
						DOUBLE dE30_x, DOUBLE dE30_y,
						DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::CircleTestType InCircleEEv_EAdaptive(DOUBLE dE00_x, DOUBLE dE00_y, 
						   DOUBLE dE01_x, DOUBLE dE01_y,
						   DOUBLE dE10_x, DOUBLE dE10_y, 
						   DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
						   DOUBLE dE30_x, DOUBLE dE30_y,
						   DOUBLE dE31_x, DOUBLE dE31_y);

static APM2D::InscribedCircleCenterType 
InscribedCircleCenterEEELocal(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			      DOUBLE dE00_x, DOUBLE dE00_y, 
			      DOUBLE dE01_x, DOUBLE dE01_y,
			      DOUBLE dE10_x, DOUBLE dE10_y, 
			      DOUBLE dE11_x, DOUBLE dE11_y,
			      DOUBLE dE20_x, DOUBLE dE20_y, 
			      DOUBLE dE21_x, DOUBLE dE21_y);
static APM2D::InscribedCircleCenterType 
InscribedCircleCenterEEEAdaptive(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				 DOUBLE dE00_x, DOUBLE dE00_y, 
				 DOUBLE dE01_x, DOUBLE dE01_y,
				 DOUBLE dE10_x, DOUBLE dE10_y, 
				 DOUBLE dE11_x, DOUBLE dE11_y,
				 DOUBLE dE20_x, DOUBLE dE20_y, 
				 DOUBLE dE21_x, DOUBLE dE21_y);
static APM2D::EdgeZoneType EdgeZoneEEELocal(DOUBLE dE00_x, DOUBLE dE00_y, 
					    DOUBLE dE01_x, DOUBLE dE01_y,
					    DOUBLE dE10_x, DOUBLE dE10_y, 
					    DOUBLE dE11_x, DOUBLE dE11_y,
					    DOUBLE dE20_x, DOUBLE dE20_y, 
					    DOUBLE dE21_x, DOUBLE dE21_y,
					    DOUBLE dE30_x, DOUBLE dE30_y,
					    DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::EdgeZoneType EdgeZoneEEEAdaptive(DOUBLE dE00_x, DOUBLE dE00_y, 
					       DOUBLE dE01_x, DOUBLE dE01_y,
					       DOUBLE dE10_x, DOUBLE dE10_y, 
					       DOUBLE dE11_x, DOUBLE dE11_y,
					       DOUBLE dE20_x, DOUBLE dE20_y, 
					       DOUBLE dE21_x, DOUBLE dE21_y,
					       DOUBLE dE30_x, DOUBLE dE30_y,
					       DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::CircleTestType InCircleEEE_VLocal(DOUBLE dE00_x, DOUBLE dE00_y, 
						DOUBLE dE01_x, DOUBLE dE01_y,
						DOUBLE dE10_x, DOUBLE dE10_y, 
						DOUBLE dE11_x, DOUBLE dE11_y,
						DOUBLE dE20_x, DOUBLE dE20_y, 
						DOUBLE dE21_x, DOUBLE dE21_y,
						DOUBLE dV3_x, DOUBLE dV3_y);
static APM2D::CircleTestType InCircleEEE_VAdaptive(DOUBLE dE00_x, DOUBLE dE00_y, 
						   DOUBLE dE01_x, DOUBLE dE01_y,
						   DOUBLE dE10_x, DOUBLE dE10_y, 
						   DOUBLE dE11_x, DOUBLE dE11_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   DOUBLE dV3_x, DOUBLE dV3_y);
static APM2D::CircleTestType InCircleEEE_ELocal(DOUBLE dE00_x, DOUBLE dE00_y, 
						DOUBLE dE01_x, DOUBLE dE01_y,
						DOUBLE dE10_x, DOUBLE dE10_y, 
						DOUBLE dE11_x, DOUBLE dE11_y,
						DOUBLE dE20_x, DOUBLE dE20_y, 
						DOUBLE dE21_x, DOUBLE dE21_y,
						DOUBLE dE30_x, DOUBLE dE30_y,
						DOUBLE dE31_x, DOUBLE dE31_y);
static APM2D::CircleTestType InCircleEEE_EAdaptive(DOUBLE dE00_x, DOUBLE dE00_y, 
						   DOUBLE dE01_x, DOUBLE dE01_y,
						   DOUBLE dE10_x, DOUBLE dE10_y, 
						   DOUBLE dE11_x, DOUBLE dE11_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   DOUBLE dE30_x, DOUBLE dE30_y,
						   DOUBLE dE31_x, DOUBLE dE31_y);

//////////////////////////////////////////////////////////////////////
// APM2D Constants

static const DOUBLE s_cdEpsilonOnCircle = 1e-10;

static const DOUBLE s_cdEpsilonNullEdge = 1e-10;
static const DOUBLE s_cdEpsilonDisc = 5e-6;
static const DOUBLE s_cdEpsilonRadius = 1e-6;
//static const DOUBLE s_cdEpsilonEdgeZone = 1e-6;
//static const DOUBLE s_cdEpsilonEdgeZone = 1e-8;
//static const DOUBLE s_cdEpsilonEdgeZone = sin(1/20 degree);
static const DOUBLE s_cdEpsilonEdgeZone = 1e-3;
//static const DOUBLE s_cdEpsilonEdgeZone = 1e-2;
//static const DOUBLE s_cdEpsilonVertexZone = 1e-8;
//static const DOUBLE s_cdEpsilonVertexZone = 1e-7;
//static const DOUBLE s_cdEpsilonVertexZone = sin(1/20 degree);
static const DOUBLE s_cdEpsilonVertexZone = 1e-3;
//static const DOUBLE s_cdEpsilonVertexOnEdge = 1e-10;
//static const DOUBLE s_cdEpsilonVertexOnEdge = sin(1/20 degree);
static const DOUBLE s_cdEpsilonVertexOnEdge = 1e-3;
//static const DOUBLE s_cdEpsilonParallel = 1e-8;
//static const DOUBLE s_cdEpsilonParallel = 1e-7;
//static const DOUBLE s_cdEpsilonParallel = 1e-6;
//static const DOUBLE s_cdEpsilonParallel = 1e-5;
//static const DOUBLE s_cdEpsilonParallel = sin(1/20 degree);
static const DOUBLE s_cdEpsilonParallel = 1e-3;

static const DOUBLE s_cdEpsilonIntersectionEEParallel = s_cdEpsilonParallel;
//static const DOUBLE s_cdEpsilonIntersectionEEEdgeZone = s_cdEpsilonEdgeZone;
static const DOUBLE s_cdEpsilonIntersectionEEEdgeZone = 1e-5;

//static const DOUBLE s_cdEpsilonInscribedCircleCenterVVVParallel = s_cdEpsilonParallel;
static const DOUBLE s_cdEpsilonInscribedCircleCenterVVVRadius = s_cdEpsilonRadius;
static const DOUBLE s_cdEpsilonInscribedCircleCenterVVVVertexZone = s_cdEpsilonVertexZone;
//static const DOUBLE s_cdEpsilonInscribedCircleCenterVVVVertexZone = 1e-6;
// -2.533199e-07 should succeed
static const DOUBLE s_cdEpsilonEdgeZoneVVVEdgeZone = s_cdEpsilonEdgeZone;
//static const DOUBLE s_cdEpsilonInCircleVVV_VOnCircle = s_cdEpsilonOnCircle;
static const DOUBLE s_cdEpsilonInCircleVVV_VOnCircle = 1e-7;
// -1.5561636269500662e-08 should be on circle
// -9.6118174530701594e-10 should be on circle
//static const DOUBLE s_cdEpsilonInCircleVVV_EOnCircle = s_cdEpsilonOnCircle;
static const DOUBLE s_cdEpsilonInCircleVVV_EOnCircle = 1e-7;
// -3.1526423682851241e-08 should be on circle

static const DOUBLE s_cdEpsilonInscribedCircleCenterVVEVertexOnEdge = s_cdEpsilonVertexOnEdge;
static const DOUBLE s_cdEpsilonInscribedCircleCenterVVEAntiparallel = s_cdEpsilonParallel;
//static const DOUBLE s_cdEpsilonInscribedCircleCenterVVEAntiparallel = 1e-5;
// -1.0596381727623339e-07 should be antiparallel
// 1.0425484220366599e-07 should be antiparallel
// -1.1090316200884002e-08 should be antiparallel
static const DOUBLE s_cdEpsilonInscribedCircleCenterVVEDisc = s_cdEpsilonDisc;
static const DOUBLE s_cdEpsilonInscribedCircleCenterVVERadius = s_cdEpsilonRadius;
static const DOUBLE s_cdEpsilonInscribedCircleCenterVVEEdgeZone = s_cdEpsilonEdgeZone;
//static const DOUBLE s_cdEpsilonInscribedCircleCenterVVEEdgeZone = 1e-7;
// -3.183230344898022e-12 should succeed
//
// -1.2434490570143618e-14 should succeed
// -1.5088788836104836e-15 should fail
// -1.0999550736361812e-12 should fail
static const DOUBLE s_cdEpsilonInscribedCircleCenterVVEVertexZone = s_cdEpsilonVertexZone;
//static const DOUBLE s_cdEpsilonInscribedCircleCenterVVEVertexZone = 1e-7;
static const DOUBLE s_cdEpsilonEdgeZoneVVEEdgeZone = s_cdEpsilonEdgeZone;
//static const DOUBLE s_cdEpsilonInCircleVVE_VOnCircle = s_cdEpsilonOnCircle;
static const DOUBLE s_cdEpsilonInCircleVVE_VOnCircle = 1e-8;
// -1.081266455780679e-09 should be on circle
// -4.3523587106412815e-10 should be on circle
static const DOUBLE s_cdEpsilonInCircleVVE_EOnCircle = s_cdEpsilonOnCircle;

static const DOUBLE s_cdEpsilonInscribedCircleCenterVVEAntiparallelRadius = s_cdEpsilonRadius;
static const DOUBLE s_cdEpsilonInscribedCircleCenterVVEAntiparallelEdgeZone = s_cdEpsilonEdgeZone;
//static const DOUBLE s_cdEpsilonInscribedCircleCenterVVEAntiparallelEdgeZone = 1e-7;
// -4.765879424439845e-08 should succeed
static const DOUBLE s_cdEpsilonInscribedCircleCenterVVEAntiparallelVertexZone = s_cdEpsilonVertexZone;
//static const DOUBLE s_cdEpsilonInscribedCircleCenterVVEAntiparallelVertexZone = 1e-7;
static const DOUBLE s_cdEpsilonEdgeZoneVVEAntiparallelEdgeZone = s_cdEpsilonEdgeZone;
//static const DOUBLE s_cdEpsilonInCircleVVE_VAntiparallelOnCircle = s_cdEpsilonOnCircle;
static const DOUBLE s_cdEpsilonInCircleVVE_VAntiparallelOnCircle = 1e-8;
// -7.8203417028583073e-09 should be on circle
static const DOUBLE s_cdEpsilonInCircleVVE_EAntiparallelOnCircle = s_cdEpsilonOnCircle;

static const DOUBLE s_cdEpsilonInscribedCircleCenterVvEVertexOnEdge = s_cdEpsilonVertexOnEdge;
//static const DOUBLE s_cdEpsilonInscribedCircleCenterVvEVertexOnEdge = 1e-9;
// 1.138272e-10 should be on edge
static const DOUBLE s_cdEpsilonInscribedCircleCenterVvERadius = s_cdEpsilonRadius;
static const DOUBLE s_cdEpsilonInscribedCircleCenterVvEVertexZone = s_cdEpsilonVertexZone;
//static const DOUBLE s_cdEpsilonInscribedCircleCenterVvEVertexZone = 5e-6;
// -4.580378e-06 should succeed
// -2.4348946262774324e-07 should succeed
// -2.2814238498128242e-08 should succeed
static const DOUBLE s_cdEpsilonEdgeZoneVvEEdgeZone = s_cdEpsilonEdgeZone;
//static const DOUBLE s_cdEpsilonInCircleVvE_VOnCircle = s_cdEpsilonOnCircle;
static const DOUBLE s_cdEpsilonInCircleVvE_VOnCircle = 1e-7;
// -1.1175880665355559e-08 should be on circle
// -1.3969837370231275e-09 should be on circle
// -9.3132251910432728e-10 should be on circle
static const DOUBLE s_cdEpsilonInCircleVvE_EOnCircle = s_cdEpsilonOnCircle;

static const DOUBLE s_cdEpsilonInscribedCircleCenterEvVVertexOnEdge = s_cdEpsilonVertexOnEdge;
//static const DOUBLE s_cdEpsilonInscribedCircleCenterEvVVertexOnEdge = 1e-9;
// 1.430864e-10 should be on edge
static const DOUBLE s_cdEpsilonInscribedCircleCenterEvVRadius = s_cdEpsilonRadius;
static const DOUBLE s_cdEpsilonInscribedCircleCenterEvVVertexZone = s_cdEpsilonVertexZone;
//static const DOUBLE s_cdEpsilonInscribedCircleCenterEvVVertexZone = 5e-6;
static const DOUBLE s_cdEpsilonEdgeZoneEvVEdgeZone = s_cdEpsilonEdgeZone;
//static const DOUBLE s_cdEpsilonEdgeZoneEvVEdgeZone = 1e-7;
//static const DOUBLE s_cdEpsilonInCircleEvV_VOnCircle = s_cdEpsilonOnCircle;
static const DOUBLE s_cdEpsilonInCircleEvV_VOnCircle = 1e-7;
// -2.1373612411035154e-08 should be on circle
// -1.3969827517007225e-09 should be on circle
//static const DOUBLE s_cdEpsilonInCircleEvV_EOnCircle = s_cdEpsilonOnCircle;
static const DOUBLE s_cdEpsilonInCircleEvV_EOnCircle = 1e-8;
// -7.4210992724681024e-09 should be on circle

static const DOUBLE s_cdEpsilonInscribedCircleCenterVEEAntiparallel = s_cdEpsilonParallel;
//static const DOUBLE s_cdEpsilonInscribedCircleCenterVEEAntiparallel = 5e-7;
//static const DOUBLE s_cdEpsilonInscribedCircleCenterVEEAntiparallel = 1e-5;
// -2.086500562073305e-07 should be antiparallel
// 1.4227449440795681e-08 should be antiparallel
//static const DOUBLE s_cdEpsilonInscribedCircleCenterVEEParallel = s_cdEpsilonParallel;
static const DOUBLE s_cdEpsilonInscribedCircleCenterVEEVertexOnEdge = s_cdEpsilonVertexOnEdge;
static const DOUBLE s_cdEpsilonInscribedCircleCenterVEEDisc = s_cdEpsilonDisc;
static const DOUBLE s_cdEpsilonInscribedCircleCenterVEEEdgeZone = s_cdEpsilonEdgeZone;
//static const DOUBLE s_cdEpsilonInscribedCircleCenterVEEEdgeZone = 1e-7;
// -2.1898066159900166e-08 should succeed
static const DOUBLE s_cdEpsilonInscribedCircleCenterVEEVertexZone = s_cdEpsilonVertexZone;
//static const DOUBLE s_cdEpsilonInscribedCircleCenterVEEVertexZone = 1e-7;
static const DOUBLE s_cdEpsilonInscribedCircleCenterVEERadius = s_cdEpsilonRadius;
static const DOUBLE s_cdEpsilonEdgeZoneVEEEdgeZone = s_cdEpsilonEdgeZone;
static const DOUBLE s_cdEpsilonInCircleVEE_VOnCircle = s_cdEpsilonOnCircle;
static const DOUBLE s_cdEpsilonInCircleVEE_EOnCircle = s_cdEpsilonOnCircle;

//static const DOUBLE s_cdEpsilonInscribedCircleCenterVEvEParallel = s_cdEpsilonParallel;
static const DOUBLE s_cdEpsilonInscribedCircleCenterVEvEParallel = 1e-2;
//static const DOUBLE s_cdEpsilonInscribedCircleCenterVEvEParallel = 0.0;

static const DOUBLE s_cdEpsilonInscribedCircleCenterVEEAntiparallelNullEdge = s_cdEpsilonNullEdge;
//static const DOUBLE s_cdEpsilonInscribedCircleCenterVEEAntiparallelNullEdge = 1e-10;
static const DOUBLE s_cdEpsilonInscribedCircleCenterVEEAntiparallelParallel = s_cdEpsilonParallel;
static const DOUBLE s_cdEpsilonInscribedCircleCenterVEEAntiparallelVertexOnEdge = s_cdEpsilonVertexOnEdge;
static const DOUBLE s_cdEpsilonInscribedCircleCenterVEEAntiparallelDisc = s_cdEpsilonDisc;
static const DOUBLE s_cdEpsilonInscribedCircleCenterVEEAntiparallelEdgeZone = s_cdEpsilonEdgeZone;
static const DOUBLE s_cdEpsilonInscribedCircleCenterVEEAntiparallelVertexZone = s_cdEpsilonVertexZone;
//static const DOUBLE s_cdEpsilonInscribedCircleCenterVEEAntiparallelVertexZone = 1e-7;
// -2.0703160423249534e-08 should succeed
//static const DOUBLE s_cdEpsilonInscribedCircleCenterVEEAntiparallelRadius = s_cdEpsilonRadius;
static const DOUBLE s_cdEpsilonEdgeZoneVEEAntiparallelEdgeZone = s_cdEpsilonEdgeZone;
//static const DOUBLE s_cdEpsilonInCircleVEE_VAntiparallelOnCircle = s_cdEpsilonOnCircle;
static const DOUBLE s_cdEpsilonInCircleVEE_VAntiparallelOnCircle = 1e-8;
// -1.0996640000000000e-09 should be on circle
// -8.8824280555545727e-10 should be on circle
static const DOUBLE s_cdEpsilonInCircleVEE_EAntiparallelOnCircle = s_cdEpsilonOnCircle;

static const DOUBLE s_cdEpsilonInscribedCircleCentervEEParallel = s_cdEpsilonParallel;
static const DOUBLE s_cdEpsilonInscribedCircleCentervEEVertexOnEdge = s_cdEpsilonVertexOnEdge;
static const DOUBLE s_cdEpsilonInscribedCircleCentervEEEdgeZone = s_cdEpsilonEdgeZone;
//static const DOUBLE s_cdEpsilonInscribedCircleCentervEEEdgeZone = 1e-6;
// -1.223868e-08 should succeed
static const DOUBLE s_cdEpsilonInscribedCircleCentervEERadius = s_cdEpsilonRadius;
static const DOUBLE s_cdEpsilonEdgeZonevEEEdgeZone = s_cdEpsilonEdgeZone;
//static const DOUBLE s_cdEpsilonInCirclevEE_VOnCircle = s_cdEpsilonOnCircle;
static const DOUBLE s_cdEpsilonInCirclevEE_VOnCircle = 1e-7;
// -2.3532461120587657e-08 should be on circle
// -1.3969837370231275e-09 should be on circle
// -9.3132251910432728e-10 should be on circle
//static const DOUBLE s_cdEpsilonInCirclevEE_EOnCircle = s_cdEpsilonOnCircle;
static const DOUBLE s_cdEpsilonInCirclevEE_EOnCircle = 5e-8;
// -9.0331004765198769e-08 should be in circle
// -3.0092210801803674e-08 should be on circle

static const DOUBLE s_cdEpsilonInscribedCircleCenterEEvParallel = s_cdEpsilonParallel;
static const DOUBLE s_cdEpsilonInscribedCircleCenterEEvVertexOnEdge = s_cdEpsilonVertexOnEdge;
static const DOUBLE s_cdEpsilonInscribedCircleCenterEEvEdgeZone = s_cdEpsilonEdgeZone;
//static const DOUBLE s_cdEpsilonInscribedCircleCenterEEvEdgeZone = 1e-6;
// -5.3846724767923699e-07 should succeed
// -1.0163050616396175e-08 should succeed
static const DOUBLE s_cdEpsilonInscribedCircleCenterEEvRadius = s_cdEpsilonRadius;
static const DOUBLE s_cdEpsilonEdgeZoneEEvEdgeZone = s_cdEpsilonEdgeZone;
//static const DOUBLE s_cdEpsilonInCircleEEv_VOnCircle = s_cdEpsilonOnCircle;
static const DOUBLE s_cdEpsilonInCircleEEv_VOnCircle = 1e-7;
// -1.0649538964677963e-08 should be on circle
// -1.8626448716751016e-09 should be on circle
//static const DOUBLE s_cdEpsilonInCircleEEv_EOnCircle = s_cdEpsilonOnCircle;
static const DOUBLE s_cdEpsilonInCircleEEv_EOnCircle = 5e-9;
// -5.5786397937638614e-09 should be in circle
// -2.199940916415915e-09 should be on circle

//static const DOUBLE s_cdEpsilonInscribedCircleCenterEEEDen = 1e-7;
//static const DOUBLE s_cdEpsilonInscribedCircleCenterEEEDen = sin(1/20 degree);
static const DOUBLE s_cdEpsilonInscribedCircleCenterEEEDen = 1e-6;
// 3.0457587671243395e-06 should succeed
// 2.0650267752727922e-05 should succeed
static const DOUBLE s_cdEpsilonInscribedCircleCenterEEEEdgeZone = s_cdEpsilonEdgeZone;
static const DOUBLE s_cdEpsilonInscribedCircleCenterEEECCW = 1e-8;
// -9.6039563391320801e-15 should fail
static const DOUBLE s_cdEpsilonEdgeZoneEEEEdgeZone = s_cdEpsilonEdgeZone;
static const DOUBLE s_cdEpsilonInCircleEEE_VOnCircle = s_cdEpsilonOnCircle;
static const DOUBLE s_cdEpsilonInCircleEEE_EOnCircle = s_cdEpsilonOnCircle;

//static const DOUBLE s_cdEpsilonInscribedCircleCenterEvEEParallel = s_cdEpsilonParallel;
static const DOUBLE s_cdEpsilonInscribedCircleCenterEvEEParallel = 1e-2;
//static const DOUBLE s_cdEpsilonInscribedCircleCenterEvEEParallel = 0.0;

//static const DOUBLE s_cdEpsilonInscribedCircleCenterEvEvEParallel = s_cdEpsilonParallel;
static const DOUBLE s_cdEpsilonInscribedCircleCenterEvEvEParallel = 1e-2;
//static const DOUBLE s_cdEpsilonInscribedCircleCenterEvEvEParallel = 0.0;

//////////////////////////////////////////////////////////////////////
// APM2D Class Implementation

APM2D::APM2D()
{
}

APM2D::~APM2D()
{
}

VOID APM2D::Init()
{
}

//////////////////////////////////////////////////////////////////////
// Intersection point of two edges

static
BOOL IntersectionEELocal(DOUBLE *pdV_x, DOUBLE *pdV_y,
			 DOUBLE dE00_x, DOUBLE dE00_y, 
			 DOUBLE dE01_x, DOUBLE dE01_y,
			 DOUBLE dE10_x, DOUBLE dE10_y, 
			 DOUBLE dE11_x, DOUBLE dE11_y)
{
  ASSERT( pdV_x != NULL && pdV_y != NULL );
  // Set to zero for the case of no intersection
  *pdV_x = 0.0;
  *pdV_y = 0.0;
  
  //////////////////////////////
  // Parallel or colinear check
  DOUBLE dE0xE1;
  DOUBLE dE0dotE0, dE1dotE1;
  DOUBLE dSinTheta;
  dE0dotE0 = ( (dE01_x - dE00_x)*(dE01_x - dE00_x) + 
	       (dE01_y - dE00_y)*(dE01_y - dE00_y) ); 
  dE1dotE1 = ( (dE11_x - dE10_x)*(dE11_x - dE10_x) + 
	       (dE11_y - dE10_y)*(dE11_y - dE10_y) ); 
  dE0xE1 = ( (dE01_x - dE00_x)*(dE11_y - dE10_y) - 
	     (dE11_x - dE10_x)*(dE01_y - dE00_y) ); 
  dSinTheta = dE0xE1/sqrt(dE0dotE0*dE1dotE1);
  if ( fabs(dSinTheta) <= s_cdEpsilonIntersectionEEParallel )
    {
      // Parallel or colinear edges
      return FALSE;
    }

  //////////////////////////////
  // Check I vs E00 half plane
  DOUBLE dE10xE11_E00;
  DOUBLE dEdgeZone_E00;
  dE10xE11_E00 = ( (dE10_x - dE00_x)*(dE11_y - dE00_y) - 
		   (dE11_x - dE00_x)*(dE10_y - dE00_y) );
  dEdgeZone_E00 = dE10xE11_E00 / dE0xE1;
  if ( dEdgeZone_E00 < -s_cdEpsilonIntersectionEEEdgeZone )
    {
      return FALSE;
    }

  //////////////////////////////
  // Check I vs E01 half plane
  DOUBLE dE10xE11_E01;
  DOUBLE dEdgeZone_E01;
  dE10xE11_E01 = ( (dE10_x - dE01_x)*(dE11_y - dE01_y) - 
		   (dE11_x - dE01_x)*(dE10_y - dE01_y) );
  dEdgeZone_E01 = dE10xE11_E01 / dE0xE1;
  if ( dEdgeZone_E01 > s_cdEpsilonIntersectionEEEdgeZone )
    {
      return FALSE;
    }

  //////////////////////////////
  // Check I vs E10 half plane
  DOUBLE dE00xE01_E10;
  DOUBLE dEdgeZone_E10;
  dE00xE01_E10 = ( (dE00_x - dE10_x)*(dE01_y - dE10_y) - 
		   (dE01_x - dE10_x)*(dE00_y - dE10_y) );
  dEdgeZone_E10 = dE00xE01_E10 / dE0xE1;
  if ( dEdgeZone_E10 > s_cdEpsilonIntersectionEEEdgeZone )
    {
      return FALSE;
    }

  //////////////////////////////
  // Check I vs E11 half plane
  DOUBLE dE00xE01_E11;
  DOUBLE dEdgeZone_E11;
  dE00xE01_E11 = ( (dE00_x - dE11_x)*(dE01_y - dE11_y) - 
		   (dE01_x - dE11_x)*(dE00_y - dE11_y) );
  dEdgeZone_E11 = dE00xE01_E11 / dE0xE1;
  if (  dEdgeZone_E11 < -s_cdEpsilonIntersectionEEEdgeZone )
    {
      return FALSE;
    }

  *pdV_x = dE00_x + dE10xE11_E00*(dE01_x - dE00_x)/dE0xE1;
  *pdV_y = dE00_y + dE10xE11_E00*(dE01_y - dE00_y)/dE0xE1;
  return TRUE;
}

BOOL APM2D::IntersectionEE(DOUBLE *pdV_x, DOUBLE *pdV_y,
			   DOUBLE dE00_x, DOUBLE dE00_y, 
			   DOUBLE dE01_x, DOUBLE dE01_y,
			   DOUBLE dE10_x, DOUBLE dE10_y, 
			   DOUBLE dE11_x, DOUBLE dE11_y)
{
  ASSERT( pdV_x != NULL && pdV_y != NULL );

  BOOL bResult;

  *pdV_x = 0.0;
  *pdV_y = 0.0;
  bResult = FALSE;

#define PERFORM_BOUNDING_BOX_CHECK_FOR_EE_INTERSECTION
#ifdef PERFORM_BOUNDING_BOX_CHECK_FOR_EE_INTERSECTION
  ////////////////////////////////////////
  // Bounding box check
  DOUBLE pdE0Min[2], pdE0Max[2];
  DOUBLE pdE1Min[2], pdE1Max[2];

  if ( dE00_x < dE01_x )
    {
      pdE0Min[0] = dE00_x;
      pdE0Max[0] = dE01_x;
    }
  else
    {
      pdE0Min[0] = dE01_x;
      pdE0Max[0] = dE00_x;
    }
  if ( dE00_y < dE01_y )
    {
      pdE0Min[1] = dE00_y;
      pdE0Max[1] = dE01_y;
    }
  else
    {
      pdE0Min[1] = dE01_y;
      pdE0Max[1] = dE00_y;
    }
  if ( dE10_x < dE11_x )
    {
      pdE1Min[0] = dE10_x;
      pdE1Max[0] = dE11_x;
    }
  else
    {
      pdE1Min[0] = dE11_x;
      pdE1Max[0] = dE10_x;
    }
  if ( dE10_y < dE11_y )
    {
      pdE1Min[1] = dE10_y;
      pdE1Max[1] = dE11_y;
    }
  else
    {
      pdE1Min[1] = dE11_y;
      pdE1Max[1] = dE10_y;
    }

  if ( pdE0Min[0] > pdE1Max[0] ||
       pdE0Min[1] > pdE1Max[1] ||
       pdE1Min[0] > pdE0Max[0] ||
       pdE1Min[1] > pdE0Max[1] )
    {
      // Bounding box trivial reject
      return FALSE;
    }
#endif // PERFORM_BOUNDING_BOX_CHECK_FOR_EE_INTERSECTION

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	bResult = IntersectionEELocal(pdV_x, pdV_y,
				      dE00_x, dE00_y, 
				      dE01_x, dE01_y,
				      dE10_x, dE10_y, 
				      dE11_x, dE11_y);
      }
      break;

    case APM::atExact:
      {
	bResult = IntersectionEEAdaptive(pdV_x, pdV_y,
					 dE00_x, dE00_y, 
					 dE01_x, dE01_y,
					 dE10_x, dE10_y, 
					 dE11_x, dE11_y);
      }
      break;

    case APM::atMPFUN:
      {
	bResult = IntersectionEEMPFUN(pdV_x, pdV_y,
				      dE00_x, dE00_y, 
				      dE01_x, dE01_y,
				      dE10_x, dE10_y, 
				      dE11_x, dE11_y);
      }
      break;

    case APM::atMPFUNFilter:
      {
	bResult = IntersectionEEMPFUNFilter(pdV_x, pdV_y,
				      dE00_x, dE00_y, 
				      dE01_x, dE01_y,
				      dE10_x, dE10_y, 
				      dE11_x, dE11_y);
      }
      break;

    case APM::atMPFUNNormalized:
      {
	bResult = IntersectionEEMPFUNNormalized(pdV_x, pdV_y,
						dE00_x, dE00_y, 
						dE01_x, dE01_y,
						dE10_x, dE10_y, 
						dE11_x, dE11_y);
      }
      break;

    default:
      ASSERT( FALSE );
      break;
    }  

  return bResult;
}

//////////////////////////////////////////////////////////////////////
// Distance Comparison Functions

APM2D::EdgeZoneType APM2D::EdgeZone(DOUBLE dV0_x, DOUBLE dV0_y,
				    DOUBLE dE10_x, DOUBLE dE10_y, 
				    DOUBLE dE11_x, DOUBLE dE11_y)
{
  APM2D::EdgeZoneType eResult;

  eResult = APM2D::eztMiddle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	DOUBLE pppdV[2][2][2]; // Endpoints
	DOUBLE ppdE[2][2];     // Edge vectors
	DOUBLE pdEdgeZone[2];
	DOUBLE dE1dotE1, dE10dotE10, dE11dotE11, dE1dotE10, dE1dotE11;
	
	pppdV[0][0][0] = 0.0;
	pppdV[0][0][1] = 0.0;
	pppdV[0][1][0] = 0.0;
	pppdV[0][1][1] = 0.0;
	pppdV[1][0][0] = dE10_x - dV0_x;
	pppdV[1][0][1] = dE10_y - dV0_y;
	pppdV[1][1][0] = dE11_x - dV0_x;
	pppdV[1][1][1] = dE11_y - dV0_y;
	
	ppdE[0][0] = 0.0;
	ppdE[0][1] = 0.0;
	ppdE[1][0] = dE11_x - dE10_x;
	ppdE[1][1] = dE11_y - dE10_y;
	
	dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
	dE10dotE10 = (pppdV[1][0][0]*pppdV[1][0][0]) + (pppdV[1][0][1]*pppdV[1][0][1]);
	dE11dotE11 = (pppdV[1][1][0]*pppdV[1][1][0]) + (pppdV[1][1][1]*pppdV[1][1][1]);
	dE1dotE10 = (ppdE[1][0]*pppdV[1][0][0]) + (ppdE[1][1]*pppdV[1][0][1]);
	dE1dotE11 = (ppdE[1][0]*pppdV[1][1][0]) + (ppdE[1][1]*pppdV[1][1][1]);

	pdEdgeZone[0] = dE1dotE10/sqrt(dE1dotE1*dE10dotE10);
	pdEdgeZone[1] = -dE1dotE11/sqrt(dE1dotE1*dE11dotE11);

	if ( !( pdEdgeZone[0] < -s_cdEpsilonEdgeZone ) )
	  {
	    eResult = APM2D::eztStart;
	  }
	else if ( !( pdEdgeZone[1] < -s_cdEpsilonEdgeZone ) )
	  {
	    eResult = APM2D::eztEnd;
	  }
	else
	  {
	    eResult = APM2D::eztMiddle;
	  }
      }
      break;

    case APM::atExact:
      {
	eResult = EdgeZoneAdaptive(dV0_x, dV0_y,
				   dE10_x, dE10_y, 
				   dE11_x, dE11_y);
      }
      break;

    case APM::atMPFUN:
      {
	eResult = EdgeZoneMPFUN(dV0_x, dV0_y,
				dE10_x, dE10_y, 
				dE11_x, dE11_y);
      }
      break;

    case APM::atMPFUNFilter:
      {
	eResult = EdgeZoneMPFUNFilter(dV0_x, dV0_y,
				dE10_x, dE10_y, 
				dE11_x, dE11_y);
      }
      break;

    case APM::atMPFUNNormalized:
      {
	eResult = EdgeZoneMPFUNNormalized(dV0_x, dV0_y,
					  dE10_x, dE10_y, 
					  dE11_x, dE11_y);
      }
      break;

    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

INT_32 APM2D::DistanceCompareV_VV(DOUBLE dV0_x, DOUBLE dV0_y,
				  DOUBLE dV1_x, DOUBLE dV1_y, 
				  DOUBLE dV2_x, DOUBLE dV2_y)
{
  INT_32 iResult;

  iResult = 0;
  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	iResult = DistanceCompareV_VVLocal(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y);
      }
      break;
    case APM::atExact:
      {
	iResult = DistanceCompareV_VVAdaptive(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y);
      }
      break;
    case APM::atMPFUN:
      {
	iResult = DistanceCompareV_VVMPFUN(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y);
      }
      break;

    case APM::atMPFUNFilter:
      {
	iResult = DistanceCompareV_VVMPFUNFilter(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y);
      }
      break;
    case APM::atMPFUNNormalized:
      {
	iResult = DistanceCompareV_VVMPFUNNormalized(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y);
      }
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  return iResult;
}

static INT_32 DistanceCompareV_VVLocal(DOUBLE dV0_x, DOUBLE dV0_y,
				       DOUBLE dV1_x, DOUBLE dV1_y, 
				       DOUBLE dV2_x, DOUBLE dV2_y)
{
  DOUBLE pppdV[3][2][2]; // Endpoints
  DOUBLE d, dV1dotV1, dV2dotV2;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[1][0][0] = dV1_x - dV0_x;
  pppdV[1][0][1] = dV1_y - dV0_y;
  pppdV[2][0][0] = dV2_x - dV0_x;
  pppdV[2][0][1] = dV2_y - dV0_y;
  
  dV1dotV1 = (pppdV[1][0][0]*pppdV[1][0][0]) + (pppdV[1][0][1]*pppdV[1][0][1]);
  dV2dotV2 = (pppdV[2][0][0]*pppdV[2][0][0]) + (pppdV[2][0][1]*pppdV[2][0][1]);

  d = dV1dotV1 - dV2dotV2;
  
  if ( d == 0.0 )
    {
      return 0;
    }
  else if ( d > 0.0 )
    {
      return 1;
    }
  else
    {
      ASSERT( d < 0.0 );
      return -1;
    }
}

INT_32 APM2D::DistanceCompareV_VL(DOUBLE dV0_x, DOUBLE dV0_y,
				  DOUBLE dV1_x, DOUBLE dV1_y, 
				  DOUBLE dE20_x, DOUBLE dE20_y, 
				  DOUBLE dE21_x, DOUBLE dE21_y)
{
  INT_32 iResult;

  iResult = 0;
  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	iResult = DistanceCompareV_VLLocal(dV0_x, dV0_y, dV1_x, dV1_y, 
					   dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;
    case APM::atExact:
      {
	iResult = DistanceCompareV_VLAdaptive(dV0_x, dV0_y, dV1_x, dV1_y, 
					      dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;
    case APM::atMPFUN:
      {
	iResult = DistanceCompareV_VLMPFUN(dV0_x, dV0_y, dV1_x, dV1_y, 
					   dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;

    case APM::atMPFUNFilter:
      {
	iResult = DistanceCompareV_VLMPFUNFilter(dV0_x, dV0_y, dV1_x, dV1_y, 
					   dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;
    case APM::atMPFUNNormalized:
      {
	iResult = DistanceCompareV_VLMPFUNNormalized(dV0_x, dV0_y, dV1_x, dV1_y, 
						     dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  return iResult;
}

static INT_32 DistanceCompareV_VLLocal(DOUBLE dV0_x, DOUBLE dV0_y,
				       DOUBLE dV1_x, DOUBLE dV1_y, 
				       DOUBLE dE20_x, DOUBLE dE20_y, 
				       DOUBLE dE21_x, DOUBLE dE21_y)
{
  DOUBLE pppdV[3][2][2]; // Endpoints
  DOUBLE ppdE[3][2];     // Edge vectors
  DOUBLE d, dV1dotV1, dE2dotE2, dE20xE21;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = dV1_x - dV0_x;
  pppdV[1][0][1] = dV1_y - dV0_y;
  pppdV[2][0][0] = dE20_x - dV0_x;
  pppdV[2][0][1] = dE20_y - dV0_y;
  pppdV[2][1][0] = dE21_x - dV0_x;
  pppdV[2][1][1] = dE21_y - dV0_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;
  
  dV1dotV1 = (pppdV[1][0][0]*pppdV[1][0][0]) + (pppdV[1][0][1]*pppdV[1][0][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);

  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);

  d = dV1dotV1*dE2dotE2 - dE20xE21*dE20xE21;
  
  if ( d == 0.0 )
    {
      return 0;
    }
  else if ( d > 0.0 )
    {
      return 1;
    }
  else
    {
      ASSERT( d < 0.0 );
      return -1;
    }
}

INT_32 APM2D::DistanceCompareV_LL(DOUBLE dV0_x, DOUBLE dV0_y,
				  DOUBLE dE10_x, DOUBLE dE10_y, 
				  DOUBLE dE11_x, DOUBLE dE11_y,
				  DOUBLE dE20_x, DOUBLE dE20_y, 
				  DOUBLE dE21_x, DOUBLE dE21_y)
{
  INT_32 iResult;

  iResult = 0;
  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	iResult = DistanceCompareV_LLLocal(dV0_x, dV0_y, dE10_x, dE10_y, dE11_x, dE11_y, 
					   dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;
    case APM::atExact:
      {
	iResult = DistanceCompareV_LLAdaptive(dV0_x, dV0_y, dE10_x, dE10_y, dE11_x, dE11_y, 
					      dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;
    case APM::atMPFUN:
      {
	iResult = DistanceCompareV_LLMPFUN(dV0_x, dV0_y, dE10_x, dE10_y, dE11_x, dE11_y, 
					   dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;

    case APM::atMPFUNFilter:
      {
	iResult = DistanceCompareV_LLMPFUNFilter(dV0_x, dV0_y, dE10_x, dE10_y, dE11_x, dE11_y, 
					   dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;
    case APM::atMPFUNNormalized:
      {
	iResult = DistanceCompareV_LLMPFUNNormalized(dV0_x, dV0_y, dE10_x, dE10_y, dE11_x, dE11_y, 
						     dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;
    default:
      ASSERT( FALSE );
      break;
    }
  
 return iResult;
}

static INT_32 DistanceCompareV_LLLocal(DOUBLE dV0_x, DOUBLE dV0_y,
				       DOUBLE dE10_x, DOUBLE dE10_y, 
				       DOUBLE dE11_x, DOUBLE dE11_y,
				       DOUBLE dE20_x, DOUBLE dE20_y, 
				       DOUBLE dE21_x, DOUBLE dE21_y)
{
  DOUBLE pppdV[3][2][2]; // Endpoints
  DOUBLE ppdE[3][2];     // Edge vectors
  DOUBLE d, dE1dotE1, dE2dotE2, dE10xE11, dE20xE21;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = dE10_x - dV0_x;
  pppdV[1][0][1] = dE10_y - dV0_y;
  pppdV[1][1][0] = dE11_x - dV0_x;
  pppdV[1][1][1] = dE11_y - dV0_y;
  pppdV[2][0][0] = dE20_x - dV0_x;
  pppdV[2][0][1] = dE20_y - dV0_y;
  pppdV[2][1][0] = dE21_x - dV0_x;
  pppdV[2][1][1] = dE21_y - dV0_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[1][0] = dE11_x - dE10_x;
  ppdE[1][1] = dE11_y - dE10_y;
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;
  
  dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);

  dE10xE11 = (pppdV[1][0][0]*pppdV[1][1][1]) - (pppdV[1][0][1]*pppdV[1][1][0]);
  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);

  d = dE2dotE2*dE10xE11*dE10xE11 - dE1dotE1*dE20xE21*dE20xE21;
  
  if ( d == 0.0 )
    {
      return 0;
    }
  else if ( d > 0.0 )
    {
      return 1;
    }
  else
    {
      ASSERT( d < 0.0 );
      return -1;
    }
}

static INT_32 DistanceCompareV_LLAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
					  DOUBLE dE10_x, DOUBLE dE10_y, 
					  DOUBLE dE11_x, DOUBLE dE11_y,
					  DOUBLE dE20_x, DOUBLE dE20_y, 
					  DOUBLE dE21_x, DOUBLE dE21_y)
{
  return DistanceCompareV_LLLocal(dV0_x, dV0_y, dE10_x, dE10_y, dE11_x, dE11_y, 
				  dE20_x, dE20_y, dE21_x, dE21_y);
}

INT_32 APM2D::DistanceCompareEvv_VV(DOUBLE dE0A_x, DOUBLE dE0A_y,
				    DOUBLE dE0B_x, DOUBLE dE0B_y,
				    DOUBLE dV1_x, DOUBLE dV1_y, 
				    DOUBLE dV2_x, DOUBLE dV2_y)
{
  INT_32 iResult;

  iResult = 0;
  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	iResult = DistanceCompareEvv_VVLocal(dE0A_x, dE0A_y, dE0B_x, dE0B_y, 
					     dV1_x, dV1_y, dV2_x, dV2_y);
      }
      break;
    case APM::atExact:
      {
	iResult = DistanceCompareEvv_VVAdaptive(dE0A_x, dE0A_y, dE0B_x, dE0B_y,
						dV1_x, dV1_y, dV2_x, dV2_y);
      }
      break;
    case APM::atMPFUN:
      {
	iResult = DistanceCompareEvv_VVMPFUN(dE0A_x, dE0A_y, dE0B_x, dE0B_y,
					     dV1_x, dV1_y, dV2_x, dV2_y);
      }
      break;

    case APM::atMPFUNFilter:
      {
	iResult = DistanceCompareEvv_VVMPFUNFilter(dE0A_x, dE0A_y, dE0B_x, dE0B_y,
					     dV1_x, dV1_y, dV2_x, dV2_y);
      }
      break;
    case APM::atMPFUNNormalized:
      {
	iResult = DistanceCompareEvv_VVMPFUNNormalized(dE0A_x, dE0A_y, dE0B_x, dE0B_y,
						       dV1_x, dV1_y, dV2_x, dV2_y);
      }
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  return iResult;
}

static INT_32 DistanceCompareEvv_VVLocal(DOUBLE dE0A_x, DOUBLE dE0A_y,
					 DOUBLE dE0B_x, DOUBLE dE0B_y,
					 DOUBLE dV1_x, DOUBLE dV1_y, 
					 DOUBLE dV2_x, DOUBLE dV2_y)
{
  DOUBLE pppdV[3][2][2]; // Endpoints
  DOUBLE d, dV1dotV1, dV2dotV2;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[1][0][0] = dV1_x - dE0A_x;
  pppdV[1][0][1] = dV1_y - dE0A_y;
  pppdV[2][0][0] = dV2_x - dE0B_x;
  pppdV[2][0][1] = dV2_y - dE0B_y;
  
  dV1dotV1 = (pppdV[1][0][0]*pppdV[1][0][0]) + (pppdV[1][0][1]*pppdV[1][0][1]);
  dV2dotV2 = (pppdV[2][0][0]*pppdV[2][0][0]) + (pppdV[2][0][1]*pppdV[2][0][1]);

  d = dV1dotV1 - dV2dotV2;
  
  if ( d == 0.0 )
    {
      return 0;
    }
  else if ( d > 0.0 )
    {
      return 1;
    }
  else
    {
      ASSERT( d < 0.0 );
      return -1;
    }
}

static INT_32 DistanceCompareEvv_VVAdaptive(DOUBLE dE0A_x, DOUBLE dE0A_y,
					    DOUBLE dE0B_x, DOUBLE dE0B_y,
					    DOUBLE dV1_x, DOUBLE dV1_y, 
					    DOUBLE dV2_x, DOUBLE dV2_y)
{
  return DistanceCompareEvv_VVLocal(dE0A_x, dE0A_y, dE0B_x, dE0B_y,
				    dV1_x, dV1_y, dV2_x, dV2_y);
}

INT_32 APM2D::DistanceCompareEvv_VL(DOUBLE dE0A_x, DOUBLE dE0A_y,
				    DOUBLE dE0B_x, DOUBLE dE0B_y,
				    DOUBLE dV1_x, DOUBLE dV1_y, 
				    DOUBLE dE20_x, DOUBLE dE20_y, 
				    DOUBLE dE21_x, DOUBLE dE21_y)
{
  INT_32 iResult;

  iResult = 0;
  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	iResult = DistanceCompareEvv_VLLocal(dE0A_x, dE0A_y, dE0B_x, dE0B_y,
					     dV1_x, dV1_y, 
					     dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;
    case APM::atExact:
      {
	iResult = DistanceCompareEvv_VLAdaptive(dE0A_x, dE0A_y, dE0B_x, dE0B_y,
						dV1_x, dV1_y, 
						dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;
    case APM::atMPFUN:
      {
	iResult = DistanceCompareEvv_VLMPFUN(dE0A_x, dE0A_y, dE0B_x, dE0B_y,
					     dV1_x, dV1_y, 
					     dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;

    case APM::atMPFUNFilter:
      {
	iResult = DistanceCompareEvv_VLMPFUNFilter(dE0A_x, dE0A_y, dE0B_x, dE0B_y,
					     dV1_x, dV1_y, 
					     dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;
    case APM::atMPFUNNormalized:
      {
	iResult = DistanceCompareEvv_VLMPFUNNormalized(dE0A_x, dE0A_y, dE0B_x, dE0B_y,
						       dV1_x, dV1_y, 
						       dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  return iResult;
}

static INT_32 DistanceCompareEvv_VLLocal(DOUBLE dE0A_x, DOUBLE dE0A_y,
					 DOUBLE dE0B_x, DOUBLE dE0B_y,
					 DOUBLE dV1_x, DOUBLE dV1_y, 
					 DOUBLE dE20_x, DOUBLE dE20_y, 
					 DOUBLE dE21_x, DOUBLE dE21_y)
{
  DOUBLE pppdV[3][2][2]; // Endpoints
  DOUBLE ppdE[3][2];     // Edge vectors
  DOUBLE d, dV1dotV1, dE2dotE2, dE20xE21;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = dV1_x - dE0A_x;
  pppdV[1][0][1] = dV1_y - dE0A_y;
  pppdV[2][0][0] = dE20_x - dE0B_x;
  pppdV[2][0][1] = dE20_y - dE0B_y;
  pppdV[2][1][0] = dE21_x - dE0B_x;
  pppdV[2][1][1] = dE21_y - dE0B_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;
  
  dV1dotV1 = (pppdV[1][0][0]*pppdV[1][0][0]) + (pppdV[1][0][1]*pppdV[1][0][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);

  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);

  d = dV1dotV1*dE2dotE2 - dE20xE21*dE20xE21;
  
  if ( d == 0.0 )
    {
      return 0;
    }
  else if ( d > 0.0 )
    {
      return 1;
    }
  else
    {
      ASSERT( d < 0.0 );
      return -1;
    }
}

static INT_32 DistanceCompareEvv_VLAdaptive(DOUBLE dE0A_x, DOUBLE dE0A_y,
					    DOUBLE dE0B_x, DOUBLE dE0B_y,
					    DOUBLE dV1_x, DOUBLE dV1_y, 
					    DOUBLE dE20_x, DOUBLE dE20_y, 
					    DOUBLE dE21_x, DOUBLE dE21_y)
{
  return DistanceCompareEvv_VLLocal(dE0A_x, dE0A_y, dE0B_x, dE0B_y, 
				    dV1_x, dV1_y, 
				    dE20_x, dE20_y, dE21_x, dE21_y);
}

INT_32 APM2D::DistanceCompareEvv_LL(DOUBLE dE0A_x, DOUBLE dE0A_y,
				    DOUBLE dE0B_x, DOUBLE dE0B_y,
				    DOUBLE dE10_x, DOUBLE dE10_y, 
				    DOUBLE dE11_x, DOUBLE dE11_y,
				    DOUBLE dE20_x, DOUBLE dE20_y, 
				    DOUBLE dE21_x, DOUBLE dE21_y)
{
  INT_32 iResult;

  iResult = 0;
  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	iResult = DistanceCompareEvv_LLLocal(dE0A_x, dE0A_y, dE0B_x, dE0B_y, 
					     dE10_x, dE10_y, dE11_x, dE11_y, 
					     dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;
    case APM::atExact:
      {
	iResult = DistanceCompareEvv_LLAdaptive(dE0A_x, dE0A_y, dE0B_x, dE0B_y,
						dE10_x, dE10_y, dE11_x, dE11_y, 
						dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;
    case APM::atMPFUN:
      {
	iResult = DistanceCompareEvv_LLMPFUN(dE0A_x, dE0A_y, dE0B_x, dE0B_y,
					     dE10_x, dE10_y, dE11_x, dE11_y, 
					     dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;

    case APM::atMPFUNFilter:
      {
	iResult = DistanceCompareEvv_LLMPFUNFilter(dE0A_x, dE0A_y, dE0B_x, dE0B_y,
					     dE10_x, dE10_y, dE11_x, dE11_y, 
					     dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;
    case APM::atMPFUNNormalized:
      {
	iResult = DistanceCompareEvv_LLMPFUNNormalized(dE0A_x, dE0A_y, dE0B_x, dE0B_y,
						       dE10_x, dE10_y, dE11_x, dE11_y, 
						       dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;
    default:
      ASSERT( FALSE );
      break;
    }
  
  return iResult;
}

static INT_32 DistanceCompareEvv_LLLocal(DOUBLE dE0A_x, DOUBLE dE0A_y,
					 DOUBLE dE0B_x, DOUBLE dE0B_y,
					 DOUBLE dE10_x, DOUBLE dE10_y, 
					 DOUBLE dE11_x, DOUBLE dE11_y,
					 DOUBLE dE20_x, DOUBLE dE20_y, 
					 DOUBLE dE21_x, DOUBLE dE21_y)
{
  DOUBLE pppdV[3][2][2]; // Endpoints
  DOUBLE ppdE[3][2];     // Edge vectors
  DOUBLE d, dE1dotE1, dE2dotE2, dE10xE11, dE20xE21;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = dE10_x - dE0A_x;
  pppdV[1][0][1] = dE10_y - dE0A_y;
  pppdV[1][1][0] = dE11_x - dE0A_x;
  pppdV[1][1][1] = dE11_y - dE0A_y;
  pppdV[2][0][0] = dE20_x - dE0B_x;
  pppdV[2][0][1] = dE20_y - dE0B_y;
  pppdV[2][1][0] = dE21_x - dE0B_x;
  pppdV[2][1][1] = dE21_y - dE0B_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[1][0] = dE11_x - dE10_x;
  ppdE[1][1] = dE11_y - dE10_y;
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;
  
  dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);

  dE10xE11 = (pppdV[1][0][0]*pppdV[1][1][1]) - (pppdV[1][0][1]*pppdV[1][1][0]);
  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);

  d = dE2dotE2*dE10xE11*dE10xE11 - dE1dotE1*dE20xE21*dE20xE21;
  
  if ( d == 0.0 )
    {
      return 0;
    }
  else if ( d > 0.0 )
    {
      return 1;
    }
  else
    {
      ASSERT( d < 0.0 );
      return -1;
    }
}

static INT_32 DistanceCompareEvv_LLAdaptive(DOUBLE dE0A_x, DOUBLE dE0A_y,
					    DOUBLE dE0B_x, DOUBLE dE0B_y,
					    DOUBLE dE10_x, DOUBLE dE10_y, 
					    DOUBLE dE11_x, DOUBLE dE11_y,
					    DOUBLE dE20_x, DOUBLE dE20_y, 
					    DOUBLE dE21_x, DOUBLE dE21_y)
{
  return DistanceCompareEvv_LLLocal(dE0A_x, dE0A_y, dE0B_x, dE0B_y, 
				    dE10_x, dE10_y, dE11_x, dE11_y, 
				    dE20_x, dE20_y, dE21_x, dE21_y);
}

INT_32 APM2D::DistanceCompareElv_VV(DOUBLE dE00_x, DOUBLE dE00_y,
				    DOUBLE dE01_x, DOUBLE dE01_y,
				    DOUBLE dE0B_x, DOUBLE dE0B_y,
				    DOUBLE dV1_x, DOUBLE dV1_y, 
				    DOUBLE dV2_x, DOUBLE dV2_y)
{
  INT_32 iResult;

  iResult = 0;
  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	iResult = DistanceCompareElv_VVLocal(dE00_x, dE00_y, dE01_x, dE01_y, 
					     dE0B_x, dE0B_y, 
					     dV1_x, dV1_y, dV2_x, dV2_y);
      }
      break;
    case APM::atExact:
      {
	iResult = DistanceCompareElv_VVAdaptive(dE00_x, dE00_y, dE01_x, dE01_y, 
						dE0B_x, dE0B_y, 
						dV1_x, dV1_y, dV2_x, dV2_y);
      }
      break;
    case APM::atMPFUN:
      {
	iResult = DistanceCompareElv_VVMPFUN(dE00_x, dE00_y, dE01_x, dE01_y, 
					     dE0B_x, dE0B_y, 
					     dV1_x, dV1_y, dV2_x, dV2_y);
      }
      break;

    case APM::atMPFUNFilter:
      {
	iResult = DistanceCompareElv_VVMPFUNFilter(dE00_x, dE00_y, dE01_x, dE01_y, 
					     dE0B_x, dE0B_y, 
					     dV1_x, dV1_y, dV2_x, dV2_y);
      }
      break;
    case APM::atMPFUNNormalized:
      {
	iResult = DistanceCompareElv_VVMPFUNNormalized(dE00_x, dE00_y, dE01_x, dE01_y, 
						       dE0B_x, dE0B_y, 
						       dV1_x, dV1_y, dV2_x, dV2_y);
      }
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  return iResult;
}

static INT_32 DistanceCompareElv_VVLocal(DOUBLE dE00_x, DOUBLE dE00_y,
					 DOUBLE dE01_x, DOUBLE dE01_y,
					 DOUBLE dE0B_x, DOUBLE dE0B_y,
					 DOUBLE dV1_x, DOUBLE dV1_y, 
					 DOUBLE dV2_x, DOUBLE dV2_y)
{
  DOUBLE pppdV[3][2][2]; // Endpoints
  DOUBLE ppdE[3][2];     // Edge vectors
  DOUBLE d, dE0dotE0, dV2dotV2, dE00xE01;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = dE00_x - dV1_x;
  pppdV[0][0][1] = dE00_y - dV1_y;
  pppdV[0][1][0] = dE01_x - dV1_x;
  pppdV[0][1][1] = dE01_y - dV1_y;
  pppdV[1][0][0] = 0.0;
  pppdV[1][0][1] = 0.0;
  pppdV[2][0][0] = dV2_x - dE0B_x;
  pppdV[2][0][1] = dV2_y - dE0B_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[0][0] = dE01_x - dE00_x;
  ppdE[0][1] = dE01_y - dE00_y;

  dE0dotE0 = (ppdE[0][0]*ppdE[0][0]) + (ppdE[0][1]*ppdE[0][1]);
  dV2dotV2 = (pppdV[2][0][0]*pppdV[2][0][0]) + (pppdV[2][0][1]*pppdV[2][0][1]);

  dE00xE01 = (pppdV[0][0][0]*pppdV[0][1][1]) - (pppdV[0][0][1]*pppdV[0][1][0]);

  d = dE00xE01*dE00xE01 - dV2dotV2*dE0dotE0;
  
  if ( d == 0.0 )
    {
      return 0;
    }
  else if ( d > 0.0 )
    {
      return 1;
    }
  else
    {
      ASSERT( d < 0.0 );
      return -1;
    }
}

static INT_32 DistanceCompareElv_VVAdaptive(DOUBLE dE00_x, DOUBLE dE00_y,
					    DOUBLE dE01_x, DOUBLE dE01_y,
					    DOUBLE dE0B_x, DOUBLE dE0B_y,
					    DOUBLE dV1_x, DOUBLE dV1_y, 
					    DOUBLE dV2_x, DOUBLE dV2_y)
{
  return DistanceCompareElv_VVLocal(dE00_x, dE00_y, dE01_x, dE01_y, 
				    dE0B_x, dE0B_y, 
				    dV1_x, dV1_y, dV2_x, dV2_y);
}

INT_32 APM2D::DistanceCompareElv_VL(DOUBLE dE00_x, DOUBLE dE00_y,
				    DOUBLE dE01_x, DOUBLE dE01_y,
				    DOUBLE dE0B_x, DOUBLE dE0B_y,
				    DOUBLE dV1_x, DOUBLE dV1_y, 
				    DOUBLE dE20_x, DOUBLE dE20_y, 
				    DOUBLE dE21_x, DOUBLE dE21_y)
{
  INT_32 iResult;

  iResult = 0;
  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	iResult = DistanceCompareElv_VLLocal(dE00_x, dE00_y, dE01_x, dE01_y, 
					     dE0B_x, dE0B_y, 
					     dV1_x, dV1_y, 
					     dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;
    case APM::atExact:
      {
	iResult = DistanceCompareElv_VLAdaptive(dE00_x, dE00_y, dE01_x, dE01_y, 
						dE0B_x, dE0B_y, 
						dV1_x, dV1_y, 
						dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;
    case APM::atMPFUN:
      {
	iResult = DistanceCompareElv_VLMPFUN(dE00_x, dE00_y, dE01_x, dE01_y, 
					     dE0B_x, dE0B_y, 
					     dV1_x, dV1_y, 
					     dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;

    case APM::atMPFUNFilter:
      {
	iResult = DistanceCompareElv_VLMPFUNFilter(dE00_x, dE00_y, dE01_x, dE01_y, 
					     dE0B_x, dE0B_y, 
					     dV1_x, dV1_y, 
					     dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;
    case APM::atMPFUNNormalized:
      {
	iResult = DistanceCompareElv_VLMPFUNNormalized(dE00_x, dE00_y, dE01_x, dE01_y, 
						       dE0B_x, dE0B_y, 
						       dV1_x, dV1_y, 
						       dE20_x, dE20_y, dE21_x, dE21_y);
      }
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  return iResult;
}

static INT_32 DistanceCompareElv_VLLocal(DOUBLE dE00_x, DOUBLE dE00_y,
					 DOUBLE dE01_x, DOUBLE dE01_y,
					 DOUBLE dE0B_x, DOUBLE dE0B_y,
					 DOUBLE dV1_x, DOUBLE dV1_y, 
					 DOUBLE dE20_x, DOUBLE dE20_y, 
					 DOUBLE dE21_x, DOUBLE dE21_y)
{
  DOUBLE pppdV[3][2][2]; // Endpoints
  DOUBLE ppdE[3][2];     // Edge vectors
  DOUBLE d, dE0dotE0, dE2dotE2, dE00xE01, dE20xE21;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = dE00_x - dV1_x;
  pppdV[0][0][1] = dE00_y - dV1_y;
  pppdV[0][1][0] = dE01_x - dV1_x;
  pppdV[0][1][1] = dE01_y - dV1_y;
  pppdV[1][0][0] = 0.0;
  pppdV[1][0][1] = 0.0;
  pppdV[2][0][0] = dE20_x - dE0B_x;
  pppdV[2][0][1] = dE20_y - dE0B_y;
  pppdV[2][1][0] = dE21_x - dE0B_x;
  pppdV[2][1][1] = dE21_y - dE0B_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[0][0] = dE01_x - dE00_x;
  ppdE[0][1] = dE01_y - dE00_y;
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;
  
  dE0dotE0 = (ppdE[0][0]*ppdE[0][0]) + (ppdE[0][1]*ppdE[0][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);

  dE00xE01 = (pppdV[0][0][0]*pppdV[0][1][1]) - (pppdV[0][0][1]*pppdV[0][1][0]);
  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);

  d = dE2dotE2*dE00xE01*dE00xE01 - dE0dotE0*dE20xE21*dE20xE21;
  
  if ( d == 0.0 )
    {
      return 0;
    }
  else if ( d > 0.0 )
    {
      return 1;
    }
  else
    {
      ASSERT( d < 0.0 );
      return -1;
    }
}

static INT_32 DistanceCompareElv_VLAdaptive(DOUBLE dE00_x, DOUBLE dE00_y,
					    DOUBLE dE01_x, DOUBLE dE01_y,
					    DOUBLE dE0B_x, DOUBLE dE0B_y,
					    DOUBLE dV1_x, DOUBLE dV1_y, 
					    DOUBLE dE20_x, DOUBLE dE20_y, 
					    DOUBLE dE21_x, DOUBLE dE21_y)
{
  return DistanceCompareElv_VLLocal(dE00_x, dE00_y, dE01_x, dE01_y, 
				    dE0B_x, dE0B_y, 
				    dV1_x, dV1_y, 
				    dE20_x, dE20_y, dE21_x, dE21_y);
}

INT_32 APM2D::DistanceCompareL_VV(DOUBLE dE00_x, DOUBLE dE00_y,
				  DOUBLE dE01_x, DOUBLE dE01_y,
				  DOUBLE dV1_x, DOUBLE dV1_y, 
				  DOUBLE dV2_x, DOUBLE dV2_y)
{
  INT_32 iResult;

  iResult = 0;
  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	iResult = DistanceCompareL_VVLocal(dE00_x, dE00_y, dE01_x, dE01_y, 
					   dV1_x, dV1_y, dV2_x, dV2_y);
      }
      break;
    case APM::atExact:
      {
	iResult = DistanceCompareL_VVAdaptive(dE00_x, dE00_y, dE01_x, dE01_y, 
					      dV1_x, dV1_y, dV2_x, dV2_y);
      }
      break;
    case APM::atMPFUN:
      {
	iResult = DistanceCompareL_VVMPFUN(dE00_x, dE00_y, dE01_x, dE01_y, 
					   dV1_x, dV1_y, dV2_x, dV2_y);
      }
      break;

    case APM::atMPFUNFilter:
      {
	iResult = DistanceCompareL_VVMPFUNFilter(dE00_x, dE00_y, dE01_x, dE01_y, 
					   dV1_x, dV1_y, dV2_x, dV2_y);
      }
      break;
    case APM::atMPFUNNormalized:
      {
	iResult = DistanceCompareL_VVMPFUNNormalized(dE00_x, dE00_y, dE01_x, dE01_y, 
						     dV1_x, dV1_y, dV2_x, dV2_y);
      }
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  return iResult;
}

static INT_32 DistanceCompareL_VVLocal(DOUBLE dE00_x, DOUBLE dE00_y,
				       DOUBLE dE01_x, DOUBLE dE01_y,
				       DOUBLE dV1_x, DOUBLE dV1_y, 
				       DOUBLE dV2_x, DOUBLE dV2_y)
{
  DOUBLE pppdV[3][2][2]; // Endpoints
  DOUBLE ppdE[3][2];     // Edge vectors
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[1][0][0] = dE00_x - dV1_x;
  pppdV[1][0][1] = dE00_y - dV1_y;
  pppdV[1][1][0] = dE01_x - dV1_x;
  pppdV[1][1][1] = dE01_y - dV1_y;
  pppdV[2][0][0] = dE00_x - dV2_x;
  pppdV[2][0][1] = dE00_y - dV2_y;
  pppdV[2][1][0] = dE01_x - dV2_x;
  pppdV[2][1][1] = dE01_y - dV2_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[0][0] = dE01_x - dE00_x;
  ppdE[0][1] = dE01_y - dE00_y;

  //DOUBLE dE0dotE0 = (ppdE[0][0]*ppdE[0][0]) + (ppdE[0][1]*ppdE[0][1]);

  DOUBLE dE10xE11 = (pppdV[1][0][0]*pppdV[1][1][1]) - (pppdV[1][0][1]*pppdV[1][1][0]);
  DOUBLE dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);

  DOUBLE d = dE10xE11*dE10xE11 - dE20xE21*dE20xE21;
  
  if ( d == 0.0 )
    {
      return 0;
    }
  else if ( d > 0.0 )
    {
      return 1;
    }
  else
    {
      ASSERT( d < 0.0 );
      return -1;
    }
}

static INT_32 DistanceCompareL_VVAdaptive(DOUBLE dE00_x, DOUBLE dE00_y,
					  DOUBLE dE01_x, DOUBLE dE01_y,
					  DOUBLE dV1_x, DOUBLE dV1_y, 
					  DOUBLE dV2_x, DOUBLE dV2_y)
{
  return DistanceCompareL_VVLocal(dE00_x, dE00_y, dE01_x, dE01_y, 
				  dV1_x, dV1_y, dV2_x, dV2_y);
}

//////////////////////////////////////////////////////////////////////
// VVV Orientation Functions

DOUBLE APM2D::SinTheta(DOUBLE dV2_x, DOUBLE dV2_y,
		       DOUBLE dV0_x, DOUBLE dV0_y,
		       DOUBLE dV1_x, DOUBLE dV1_y)
{
  DOUBLE pdV1[2], pdV2[2], dV1dotV1, dV2dotV2, dDen;
  DOUBLE d;

  d = 0.0;

  pdV1[0] = dV1_x - dV0_x;
  pdV1[1] = dV1_y - dV0_y;
  pdV2[0] = dV2_x - dV0_x;
  pdV2[1] = dV2_y - dV0_y;
  
  dV1dotV1 = (pdV1[0]*pdV1[0]) + (pdV1[1]*pdV1[1]);
  dV2dotV2 = (pdV2[0]*pdV2[0]) + (pdV2[1]*pdV2[1]);
  
  d = ( (pdV1[0] * pdV2[1]) - (pdV2[0] * pdV1[1]) );
  
  dDen = sqrt(dV1dotV1*dV2dotV2);
  if ( dDen != 0.0 )
    {
      d /= dDen;
    }

  return d;
}

DOUBLE APM2D::OrientationVVVValue(DOUBLE dV0_x, DOUBLE dV0_y,
				  DOUBLE dV1_x, DOUBLE dV1_y,
				  DOUBLE dV2_x, DOUBLE dV2_y)
{
  DOUBLE d;

  d = 0.0;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
      {
	d = ( (dV0_x * dV1_y + dV1_x * dV2_y + dV2_x * dV0_y) -
	      (dV0_x * dV2_y + dV1_x * dV0_y + dV2_x * dV1_y) );
      }
      break;

    case APM::atLocal:
      {
	DOUBLE pdV1[2], pdV2[2];
	
	pdV1[0] = dV1_x - dV0_x;
	pdV1[1] = dV1_y - dV0_y;
	pdV2[0] = dV2_x - dV0_x;
	pdV2[1] = dV2_y - dV0_y;
	
	d = ( (pdV1[0] * pdV2[1]) - (pdV2[0] * pdV1[1]) );
      }
      break;

    case APM::atExact:
      {
	//DOUBLE ppdV0[2], ppdV1[2], ppdV2[2];
	
	//ppdV0[0] = dV0_x;
	//ppdV0[1] = dV0_y;
	//ppdV1[0] = dV1_x;
	//ppdV1[1] = dV1_y;
	//ppdV2[0] = dV2_x;
	//ppdV2[1] = dV2_y;
	
	//DOUBLE dShewchuk;
	//dShewchuk = orient2d(ppdV0, ppdV1, ppdV2);
	d = OrientationVVVValueAdaptive(dV0_x, dV0_y,
					dV1_x, dV1_y,
					dV2_x, dV2_y);
	//ASSERT( d == dShewchuk );
      }
      break;

    case APM::atMPFUN:
    case APM::atMPFUNFilter:
    case APM::atMPFUNNormalized:
      {
	DOUBLE ppdV0[2], ppdV1[2], ppdV2[2];
	
	ppdV0[0] = dV0_x;
	ppdV0[1] = dV0_y;
	ppdV1[0] = dV1_x;
	ppdV1[1] = dV1_y;
	ppdV2[0] = dV2_x;
	ppdV2[1] = dV2_y;
	
	d = orient2dexact(ppdV0, ppdV1, ppdV2);
      }
      break;

    default:
      ASSERT( FALSE );
      break;
    }

  return d;
}

INT_32 APM2D::OrientationVVV(DOUBLE dV0_x, DOUBLE dV0_y,
			     DOUBLE dV1_x, DOUBLE dV1_y,
			     DOUBLE dV2_x, DOUBLE dV2_y)
{
  DOUBLE d;

  d = OrientationVVVValue(dV0_x, dV0_y,
			  dV1_x, dV1_y,
			  dV2_x, dV2_y);

  if ( d == 0.0 )
    {
      return 0;
    }
  else if ( d > 0.0 )
    {
      return 1;
    }
  else
    {
      ASSERT( d < 0.0 );
      return -1;
    }
}

//////////////////////////////////////////////////////////////////////
// EE Orientaion Functions

INT_32 APM2D::OrientationExE(DOUBLE dE00_x, DOUBLE dE00_y,
			     DOUBLE dE01_x, DOUBLE dE01_y,
			     DOUBLE dE10_x, DOUBLE dE10_y,
			     DOUBLE dE11_x, DOUBLE dE11_y)
{
  INT_32 iResult;

  iResult = 0;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	iResult = OrientationExELocal(dE00_x, dE00_y,
				      dE01_x, dE01_y,
				      dE10_x, dE10_y,
				      dE11_x, dE11_y);
      }
      break;

    case APM::atExact:
      {
	DOUBLE d;
	d = OrientationExEAdaptive(dE00_x, dE00_y,
				   dE01_x, dE01_y,
				   dE10_x, dE10_y,
				   dE11_x, dE11_y);
	if ( d == 0.0 )
	  {
	    iResult = 0;
	  }
	else if ( d > 0.0 )
	  {
	    iResult = 1;
	  }
	else
	  {
	    ASSERT( d < 0.0 );
	    iResult = -1;
	  }
      }
      break;
 
    case APM::atMPFUN:
      {
	iResult = OrientationExEMPFUN(dE00_x, dE00_y,
				      dE01_x, dE01_y,
				      dE10_x, dE10_y,
				      dE11_x, dE11_y);
      }
      break;

    case APM::atMPFUNFilter:
      {
	iResult = OrientationExEMPFUNFilter(dE00_x, dE00_y,
					    dE01_x, dE01_y,
					    dE10_x, dE10_y,
					    dE11_x, dE11_y);
      }
      break;

    case APM::atMPFUNNormalized:
      {
	iResult = OrientationExEMPFUNNormalized(dE00_x, dE00_y,
						dE01_x, dE01_y,
						dE10_x, dE10_y,
						dE11_x, dE11_y);
      }
      break;

    default:
      ASSERT( FALSE );
      break;
    }

  return iResult;
}
  
static INT_32 OrientationExELocal(DOUBLE dE00_x, DOUBLE dE00_y,
				  DOUBLE dE01_x, DOUBLE dE01_y,
				  DOUBLE dE10_x, DOUBLE dE10_y,
				  DOUBLE dE11_x, DOUBLE dE11_y)
{
  DOUBLE pdE0[2], pdE1[2];
  DOUBLE d;

  d = 0.0;
  
  pdE0[0] = dE01_x - dE00_x;
  pdE0[1] = dE01_y - dE00_y;
  pdE1[0] = dE11_x - dE10_x;
  pdE1[1] = dE11_y - dE10_y;

  d = ( (pdE0[0] * pdE1[1]) - (pdE1[0] * pdE0[1]) );

  if ( d == 0.0 )
    {
      return 0;
    }
  else if ( d > 0.0 )
    {
      return 1;
    }
  else
    {
      ASSERT( d < 0.0 );
      return -1;
    }
}
  
INT_32 APM2D::OrientationEdotE(DOUBLE dE00_x, DOUBLE dE00_y,
			       DOUBLE dE01_x, DOUBLE dE01_y,
			       DOUBLE dE10_x, DOUBLE dE10_y,
			       DOUBLE dE11_x, DOUBLE dE11_y)
{
  INT_32 iResult;

  iResult = 0;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	iResult = OrientationEdotELocal(dE00_x, dE00_y,
					dE01_x, dE01_y,
					dE10_x, dE10_y,
					dE11_x, dE11_y);
      }
      break;

    case APM::atExact:
      {
	DOUBLE d;
	d = OrientationEdotEAdaptive(dE00_x, dE00_y,
					   dE01_x, dE01_y,
					   dE10_x, dE10_y,
					   dE11_x, dE11_y);
	if ( d == 0.0 )
	  {
	    iResult = 0;
	  }
	else if ( d > 0.0 )
	  {
	    iResult = 1;
	  }
	else
	  {
	    ASSERT( d < 0.0 );
	    iResult = -1;
	  }
      }
      break;

    case APM::atMPFUN:
      {
	iResult = OrientationEdotEMPFUN(dE00_x, dE00_y,
					dE01_x, dE01_y,
					dE10_x, dE10_y,
					dE11_x, dE11_y);
      }
      break;

    case APM::atMPFUNFilter:
      {
	iResult = OrientationEdotEMPFUNFilter(dE00_x, dE00_y,
					dE01_x, dE01_y,
					dE10_x, dE10_y,
					dE11_x, dE11_y);
      }
      break;
    case APM::atMPFUNNormalized:
      {
	iResult = OrientationEdotEMPFUNNormalized(dE00_x, dE00_y,
						  dE01_x, dE01_y,
						  dE10_x, dE10_y,
						  dE11_x, dE11_y);
      }
      break;

    default:
      ASSERT( FALSE );
      break;
    }
  
  return iResult;
}

static INT_32 OrientationEdotELocal(DOUBLE dE00_x, DOUBLE dE00_y,
				    DOUBLE dE01_x, DOUBLE dE01_y,
				    DOUBLE dE10_x, DOUBLE dE10_y,
				    DOUBLE dE11_x, DOUBLE dE11_y)
{
  DOUBLE d;
  DOUBLE pdE0[2], pdE1[2];

  d = 0.0;

  pdE0[0] = dE01_x - dE00_x;
  pdE0[1] = dE01_y - dE00_y;
  pdE1[0] = dE11_x - dE10_x;
  pdE1[1] = dE11_y - dE10_y;
  
  d = ( (pdE0[0] * pdE1[0]) + (pdE0[1] * pdE1[1]) );

  if ( d == 0.0 )
    {
      return 0;
    }
  else if ( d > 0.0 )
    {
      return 1;
    }
  else
    {
      ASSERT( d < 0.0 );
      return -1;
    }
}

//////////////////////////////////////////////////////////////////////
// Inscribed Circle Center Functions

APM2D::InscribedCircleCenterType 
APM2D::InscribedCircleCenterVVV(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dV00_x, DOUBLE dV00_y,
				DOUBLE dV0_x, DOUBLE dV0_y,
				DOUBLE dV01_x, DOUBLE dV01_y,
				DOUBLE dV10_x, DOUBLE dV10_y,
				DOUBLE dV1_x, DOUBLE dV1_y,
				DOUBLE dV11_x, DOUBLE dV11_y,
				DOUBLE dV20_x, DOUBLE dV20_y,
				DOUBLE dV2_x, DOUBLE dV2_y,
				DOUBLE dV21_x, DOUBLE dV21_y)
{
  ASSERT( pdC_x != NULL && pdC_y != NULL && pdR != NULL );

  APM2D::InscribedCircleCenterType eResult;

  eResult = APM2D::icctNull;

  //////////////////////////////////////////////////
  // Degenerate Touching Check
  if ( ( dV0_x == dV1_x && dV0_y == dV1_y ) &&
       ( dV0_x == dV2_x && dV0_y == dV2_y ) )
    {
      *pdC_x = dV0_x;
      *pdC_y = dV0_y;
      *pdR = 0.0;
      return APM2D::icctVVV;
    }

  //////////////////////////////////////////////////
  // CCW Checks

  if ( !Counterclockwise(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InscribedCircleCenterVVVLocal(pdC_x, pdC_y, pdR, 
						dV00_x, dV00_y, 
						dV0_x, dV0_y, 
						dV01_x, dV01_y, 
						dV10_x, dV10_y, 
						dV1_x, dV1_y, 
						dV11_x, dV11_y, 
						dV20_x, dV20_y,
						dV2_x, dV2_y,
						dV21_x, dV21_y);
      }
    break;

    case APM::atExact:
      {
	eResult =  InscribedCircleCenterVVVAdaptive(pdC_x, pdC_y, pdR, 
						    dV00_x, dV00_y, 
						    dV0_x, dV0_y, 
						    dV01_x, dV01_y, 
						    dV10_x, dV10_y, 
						    dV1_x, dV1_y, 
						    dV11_x, dV11_y, 
						    dV20_x, dV20_y,
						    dV2_x, dV2_y,
						    dV21_x, dV21_y);
      }
    break;

    case APM::atMPFUN:
      {
	eResult =  InscribedCircleCenterVVVMPFUN(pdC_x, pdC_y, pdR, 
						 dV00_x, dV00_y, 
						 dV0_x, dV0_y, 
						 dV01_x, dV01_y, 
						 dV10_x, dV10_y, 
						 dV1_x, dV1_y, 
						 dV11_x, dV11_y, 
						 dV20_x, dV20_y,
						 dV2_x, dV2_y,
						 dV21_x, dV21_y);
      }
    break;

    case APM::atMPFUNFilter:
      {
	eResult =  InscribedCircleCenterVVVMPFUNFilter(pdC_x, pdC_y, pdR, 
						 dV00_x, dV00_y, 
						 dV0_x, dV0_y, 
						 dV01_x, dV01_y, 
						 dV10_x, dV10_y, 
						 dV1_x, dV1_y, 
						 dV11_x, dV11_y, 
						 dV20_x, dV20_y,
						 dV2_x, dV2_y,
						 dV21_x, dV21_y);
      }
    break;

    case APM::atMPFUNNormalized:
      {
	eResult =  InscribedCircleCenterVVVMPFUNNormalized(pdC_x, pdC_y, pdR, 
							   dV00_x, dV00_y, 
							   dV0_x, dV0_y, 
							   dV01_x, dV01_y, 
							   dV10_x, dV10_y, 
							   dV1_x, dV1_y, 
							   dV11_x, dV11_y, 
							   dV20_x, dV20_y,
							   dV2_x, dV2_y,
							   dV21_x, dV21_y);
      }
    break;

    default:
      ASSERT( FALSE );
      break;
    }

  return eResult;
}

static APM2D::InscribedCircleCenterType 
InscribedCircleCenterVVVLocal(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
                              DOUBLE dV00_x, DOUBLE dV00_y,
                              DOUBLE dV0_x, DOUBLE dV0_y,
			      DOUBLE dV01_x, DOUBLE dV01_y,
			      DOUBLE dV10_x, DOUBLE dV10_y,
			      DOUBLE dV1_x, DOUBLE dV1_y,
			      DOUBLE dV11_x, DOUBLE dV11_y,
			      DOUBLE dV20_x, DOUBLE dV20_y,
			      DOUBLE dV2_x, DOUBLE dV2_y,
			      DOUBLE dV21_x, DOUBLE dV21_y)
{
  DOUBLE pdC[3], dR, ppdV[3][2], pppdVN[3][2][2], pdV1V2[2];
  DOUBLE ppdVertexZone[3][2];
  DOUBLE dV1dotV1, dV2dotV2, dV1xV2;
  DOUBLE dV1xV00, dV1xV01, dV1xV10, dV1xV11, dV1xV20, dV1xV21;
  DOUBLE dV2xV00, dV2xV01, dV2xV10, dV2xV11, dV2xV20, dV2xV21;
  DOUBLE dV1dotV10, dV1dotV11, dV2dotV20, dV2dotV21;
  DOUBLE dV00dotV00, dV01dotV01, dV10dotV10, dV11dotV11, dV20dotV20, dV21dotV21;
  DOUBLE dV1V2dotV1V2;
  //DOUBLE dSinTheta;
  UINT_32 i, j;

  // Translate V0 to the origin
  memset(ppdV, 0, sizeof(ppdV));
  ppdV[1][0] = dV1_x - dV0_x;
  ppdV[1][1] = dV1_y - dV0_y;
  ppdV[2][0] = dV2_x - dV0_x;
  ppdV[2][1] = dV2_y - dV0_y;

  // Vertex neighbor vectors
  memset(pppdVN, 0, sizeof(pppdVN));
  pppdVN[0][0][0] = dV0_x - dV00_x;
  pppdVN[0][0][1] = dV0_y - dV00_y;
  pppdVN[0][1][0] = dV0_x - dV01_x;
  pppdVN[0][1][1] = dV0_y - dV01_y;
  pppdVN[1][0][0] = dV1_x - dV10_x;
  pppdVN[1][0][1] = dV1_y - dV10_y;
  pppdVN[1][1][0] = dV1_x - dV11_x;
  pppdVN[1][1][1] = dV1_y - dV11_y;
  pppdVN[2][0][0] = dV2_x - dV20_x;
  pppdVN[2][0][1] = dV2_y - dV20_y;
  pppdVN[2][1][0] = dV2_x - dV21_x;
  pppdVN[2][1][1] = dV2_y - dV21_y;

  pdV1V2[0] = dV2_x - dV1_x;
  pdV1V2[1] = dV2_y - dV1_y;
  
  dV1xV2 = ( (ppdV[1][0] * ppdV[2][1]) - (ppdV[2][0] * ppdV[1][1]) );
  dV1dotV1 = (ppdV[1][0] * ppdV[1][0]) + (ppdV[1][1] * ppdV[1][1]);
  dV2dotV2 = (ppdV[2][0] * ppdV[2][0]) + (ppdV[2][1] * ppdV[2][1]);

  dV1xV00 = ( (ppdV[1][0] * pppdVN[0][0][1]) - (ppdV[1][1] * pppdVN[0][0][0]) );
  dV1xV01 = ( (ppdV[1][0] * pppdVN[0][1][1]) - (ppdV[1][1] * pppdVN[0][1][0]) );
  dV1xV10 = ( (ppdV[1][0] * pppdVN[1][0][1]) - (ppdV[1][1] * pppdVN[1][0][0]) );
  dV1xV11 = ( (ppdV[1][0] * pppdVN[1][1][1]) - (ppdV[1][1] * pppdVN[1][1][0]) );
  dV1xV20 = ( (ppdV[1][0] * pppdVN[2][0][1]) - (ppdV[1][1] * pppdVN[2][0][0]) );
  dV1xV21 = ( (ppdV[1][0] * pppdVN[2][1][1]) - (ppdV[1][1] * pppdVN[2][1][0]) );

  dV2xV00 = ( (ppdV[2][0] * pppdVN[0][0][1]) - (ppdV[2][1] * pppdVN[0][0][0]) );
  dV2xV01 = ( (ppdV[2][0] * pppdVN[0][1][1]) - (ppdV[2][1] * pppdVN[0][1][0]) );
  dV2xV10 = ( (ppdV[2][0] * pppdVN[1][0][1]) - (ppdV[2][1] * pppdVN[1][0][0]) );
  dV2xV11 = ( (ppdV[2][0] * pppdVN[1][1][1]) - (ppdV[2][1] * pppdVN[1][1][0]) );
  dV2xV20 = ( (ppdV[2][0] * pppdVN[2][0][1]) - (ppdV[2][1] * pppdVN[2][0][0]) );
  dV2xV21 = ( (ppdV[2][0] * pppdVN[2][1][1]) - (ppdV[2][1] * pppdVN[2][1][0]) );

  dV1dotV10 = ( (ppdV[1][0] * pppdVN[1][0][0]) + (ppdV[1][1] * pppdVN[1][0][1]) );
  dV1dotV11 = ( (ppdV[1][0] * pppdVN[1][1][0]) + (ppdV[1][1] * pppdVN[1][1][1]) );
  dV2dotV20 = ( (ppdV[2][0] * pppdVN[2][0][0]) + (ppdV[2][1] * pppdVN[2][0][1]) );
  dV2dotV21 = ( (ppdV[2][0] * pppdVN[2][1][0]) + (ppdV[2][1] * pppdVN[2][1][1]) );

  dV00dotV00 = ( (pppdVN[0][0][0] * pppdVN[0][0][0]) + (pppdVN[0][0][1] * pppdVN[0][0][1]) );
  dV01dotV01 = ( (pppdVN[0][1][0] * pppdVN[0][1][0]) + (pppdVN[0][1][1] * pppdVN[0][1][1]) );
  dV10dotV10 = ( (pppdVN[1][0][0] * pppdVN[1][0][0]) + (pppdVN[1][0][1] * pppdVN[1][0][1]) );
  dV11dotV11 = ( (pppdVN[1][1][0] * pppdVN[1][1][0]) + (pppdVN[1][1][1] * pppdVN[1][1][1]) );
  dV20dotV20 = ( (pppdVN[2][0][0] * pppdVN[2][0][0]) + (pppdVN[2][0][1] * pppdVN[2][0][1]) );
  dV21dotV21 = ( (pppdVN[2][1][0] * pppdVN[2][1][0]) + (pppdVN[2][1][1] * pppdVN[2][1][1]) );

  /*
  dSinTheta = dV1xV2/sqrt(dV1dotV1*dV2dotV2);
  if ( dSinTheta <= s_cdEpsilonInscribedCircleCenterVVVParallel )
  */
  if ( !APM2D::Counterclockwise(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  //////////////////////////////////////////////////
  // Compute the center and radius

  pdC[0] =  ( (dV2dotV2 * ppdV[1][1]) - (dV1dotV1 * ppdV[2][1]) ) / ( -2.0 * dV1xV2 );
  pdC[1] =  ( (dV2dotV2 * ppdV[1][0]) - (dV1dotV1 * ppdV[2][0]) ) / ( 2.0 * dV1xV2 );
  pdC[2] = sqrt( (pdC[0] * pdC[0]) + (pdC[1] * pdC[1]) );

  dV1V2dotV1V2 = (pdV1V2[0] * pdV1V2[0]) + (pdV1V2[1] * pdV1V2[1]);

  dR = sqrt(dV1dotV1 * dV2dotV2 * dV1V2dotV1V2) / ( 2.0 * dV1xV2 );

  //////////////////////////////////////////////////
  // Check the Vertex Zones

  ppdVertexZone[0][0] = ( ( dV2dotV2*dV1xV00 - dV1dotV1*dV2xV00 ) /
			  ( 2.0*dV1xV2*dR*sqrt(dV00dotV00) ) );
  ppdVertexZone[0][1] = ( ( dV2dotV2*dV1xV01 - dV1dotV1*dV2xV01 ) /
			  ( 2.0*dV1xV2*dR*sqrt(dV01dotV01) ) );
  ppdVertexZone[1][0] = ( ( dV2dotV2*dV1xV10 - dV1dotV1*dV2xV10 - 2.0*dV1xV2*dV1dotV10 ) /
			  ( 2.0*dV1xV2*dR*sqrt(dV10dotV10) ) );
  ppdVertexZone[1][1] = ( ( dV2dotV2*dV1xV11 - dV1dotV1*dV2xV11 - 2.0*dV1xV2*dV1dotV11 ) /
			  ( 2.0*dV1xV2*dR*sqrt(dV11dotV11) ) );
  ppdVertexZone[2][0] = ( ( dV2dotV2*dV1xV20 - dV1dotV1*dV2xV20 - 2.0*dV1xV2*dV2dotV20 ) /
			  ( 2.0*dV1xV2*dR*sqrt(dV20dotV20) ) );
  ppdVertexZone[2][1] = ( ( dV2dotV2*dV1xV21 - dV1dotV1*dV2xV21 - 2.0*dV1xV2*dV2dotV21 ) /
			  ( 2.0*dV1xV2*dR*sqrt(dV21dotV21) ) );

  for ( i = 0; i < 3; i++ )
    {
      for ( j = 0; j < 2; j++ )
	{
	  if ( ppdVertexZone[i][j] < -s_cdEpsilonInscribedCircleCenterVVVVertexZone )
	    {
#ifdef DEBUG_EPSILON_VALUES
	      if ( !( ppdVertexZone[i][j] < -(s_cdEpsilonInscribedCircleCenterVVVVertexZone*s_cdOrderOfMagnitude) ) )
		{
		  fprintf(stderr, "  VVV: Vertex Zone: %e < %e < %e\n",
			  s_cdEpsilonInscribedCircleCenterVVVVertexZone*s_cdOrderOfMagnitude,
			  ppdVertexZone[i][j],
			  s_cdEpsilonInscribedCircleCenterVVVVertexZone);
		}
#endif // DEBUG_EPSILON_VALUES	      
	      *pdC_x = 0.0;
	      *pdC_y = 0.0;
	      *pdR = 0.0;
	      return APM2D::icctNull;
	    }
	}
    }

  if ( !( fabs(pdC[2] - dR) < s_cdEpsilonInscribedCircleCenterVVVRadius ) )
    {
      fprintf(stderr, "VVV: C = ( %f, %f, %f ) != %f = R\n", pdC[0], pdC[1], pdC[2], dR);
      //ASSERT( FALSE );
    }

  // Translate the answer back out away from the origin
  *pdC_x = dV0_x + pdC[0];
  *pdC_y = dV0_y + pdC[1];
  *pdR = dR;
  return APM2D::icctVVV;
}

APM2D::InscribedCircleCenterType 
InscribedCircleCenterVVVAdaptive(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
	                         DOUBLE dV00_x, DOUBLE dV00_y,
	                         DOUBLE dV0_x, DOUBLE dV0_y,
				 DOUBLE dV01_x, DOUBLE dV01_y,
				 DOUBLE dV10_x, DOUBLE dV10_y,
				 DOUBLE dV1_x, DOUBLE dV1_y,
				 DOUBLE dV11_x, DOUBLE dV11_y,
				 DOUBLE dV20_x, DOUBLE dV20_y,
				 DOUBLE dV2_x, DOUBLE dV2_y,
				 DOUBLE dV21_x, DOUBLE dV21_y)
{
  return InscribedCircleCenterVVVLocal(pdC_x, pdC_y, pdR, 
				       dV00_x, dV00_y, 
				       dV0_x, dV0_y, 
				       dV01_x, dV01_y, 
				       dV10_x, dV10_y, 
				       dV1_x, dV1_y, 
				       dV11_x, dV11_y, 
				       dV20_x, dV20_y,
				       dV2_x, dV2_y,
				       dV21_x, dV21_y);
}

APM2D::EdgeZoneType APM2D::EdgeZoneVVV(DOUBLE dV0_x, DOUBLE dV0_y,
				       DOUBLE dV1_x, DOUBLE dV1_y,
				       DOUBLE dV2_x, DOUBLE dV2_y,
				       DOUBLE dE30_x, DOUBLE dE30_y,
				       DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::EdgeZoneType eResult;

  //////////////////////////////////////////////////
  // Degenerate Touching Check
  if ( ( dV0_x == dV1_x && dV0_y == dV1_y ) &&
       ( dV0_x == dV2_x && dV0_y == dV2_y ) )
    {
      return EdgeZone(dV0_x, dV0_y,
		      dE30_x, dE30_y, 
		      dE31_x, dE31_y);
    }

  eResult = APM2D::eztMiddle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = EdgeZoneVVVLocal(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y,
				   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atExact:
      {
	eResult = EdgeZoneVVVAdaptive(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y,
				      dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = EdgeZoneVVVMPFUN(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y,
				   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;

    case APM::atMPFUNFilter:
      {
	eResult = EdgeZoneVVVMPFUNFilter(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y,
				   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = EdgeZoneVVVMPFUNNormalized(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y,
					     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::EdgeZoneType EdgeZoneVVVLocal(DOUBLE dV0_x, DOUBLE dV0_y,
					    DOUBLE dV1_x, DOUBLE dV1_y,
					    DOUBLE dV2_x, DOUBLE dV2_y,
					    DOUBLE dE30_x, DOUBLE dE30_y,
					    DOUBLE dE31_x, DOUBLE dE31_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE pdC[2];
  DOUBLE pdEdgeZone[2];
  DOUBLE dV1dotV1, dV2dotV2;
  DOUBLE dE3dotE3, dE3dotE30, dE3dotE31;
  DOUBLE dMagCE30, dMagCE31;
  DOUBLE dV1xV2, dE3xV2, dE3xV1;
  DOUBLE dDen;
  APM2D::EdgeZoneType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = dV1_x - dV0_x;
  pppdV[1][0][1] = dV1_y - dV0_y;
  pppdV[2][0][0] = dV2_x - dV0_x;
  pppdV[2][0][1] = dV2_y - dV0_y;
  pppdV[3][0][0] = dE30_x - dV0_x;
  pppdV[3][0][1] = dE30_y - dV0_y;
  pppdV[3][1][0] = dE31_x - dV0_x;
  pppdV[3][1][1] = dE31_y - dV0_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[3][0] = dE31_x - dE30_x;
  ppdE[3][1] = dE31_y - dE30_y;
  
  dV1dotV1 = (pppdV[1][0][0]*pppdV[1][0][0]) + (pppdV[1][0][1]*pppdV[1][0][1]);
  dV2dotV2 = (pppdV[2][0][0]*pppdV[2][0][0]) + (pppdV[2][0][1]*pppdV[2][0][1]);

  dE3dotE3 = (ppdE[3][0]*ppdE[3][0]) + (ppdE[3][1]*ppdE[3][1]);
  dE3dotE30 = (ppdE[3][0]*pppdV[3][0][0]) + (ppdE[3][1]*pppdV[3][0][1]);
  dE3dotE31 = (ppdE[3][0]*pppdV[3][1][0]) + (ppdE[3][1]*pppdV[3][1][1]);
  
  dV1xV2 = (pppdV[1][0][0]*pppdV[2][0][1]) - (pppdV[1][0][1]*pppdV[2][0][0]);
  dE3xV1 = (ppdE[3][0]*pppdV[1][0][1]) - (ppdE[3][1]*pppdV[1][0][0]);
  dE3xV2 = (ppdE[3][0]*pppdV[2][0][1]) - (ppdE[3][1]*pppdV[2][0][0]);

  dDen = 2.0*dV1xV2; 
  ASSERT( dDen != 0.0 );

  pdC[0] =  ( (dV2dotV2 * pppdV[1][0][1]) - (dV1dotV1 * pppdV[2][0][1]) ) / ( -2.0 * dV1xV2 );
  pdC[1] =  ( (dV2dotV2 * pppdV[1][0][0]) - (dV1dotV1 * pppdV[2][0][0]) ) / ( 2.0 * dV1xV2 );

  dMagCE30 = ( (pdC[0] - pppdV[3][0][0])*(pdC[0] - pppdV[3][0][0]) + 
	       (pdC[1] - pppdV[3][0][1])*(pdC[1] - pppdV[3][0][1]) );
  dMagCE31 = ( (pdC[0] - pppdV[3][1][0])*(pdC[0] - pppdV[3][1][0]) + 
	       (pdC[1] - pppdV[3][1][1])*(pdC[1] - pppdV[3][1][1]) );

  if ( dMagCE30 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[0] = 0.0;
    }
  else
    {
      pdEdgeZone[0] = ( ( 2.0*dV1xV2*dE3dotE30 - (dV1dotV1*dE3xV2 - dV2dotV2*dE3xV1) ) / 
			( dDen*sqrt(dE3dotE3*dMagCE30) ) );
    }
  if ( dMagCE31 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[1] = 0.0;
    }
  else
    {
      pdEdgeZone[1] = -( ( 2.0*dV1xV2*dE3dotE31 - (dV1dotV1*dE3xV2 - dV2dotV2*dE3xV1) ) / 
			 ( dDen*sqrt(dE3dotE3*dMagCE31) ) );
    }
  
  if ( !( pdEdgeZone[0] < -s_cdEpsilonEdgeZoneVVVEdgeZone ) )
    {
      eResult = APM2D::eztStart;
    }
  else if ( !( pdEdgeZone[1] < -s_cdEpsilonEdgeZoneVVVEdgeZone ) )
    {
      eResult = APM2D::eztEnd;
    }
  else
    {
      eResult = APM2D::eztMiddle;
    }

  return eResult;
}

static APM2D::EdgeZoneType EdgeZoneVVVAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
					       DOUBLE dV1_x, DOUBLE dV1_y,
					       DOUBLE dV2_x, DOUBLE dV2_y,
					       DOUBLE dE30_x, DOUBLE dE30_y,
					       DOUBLE dE31_x, DOUBLE dE31_y)
{
  return EdgeZoneVVVLocal(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y,
			  dE30_x, dE30_y, dE31_x, dE31_y);
}

APM2D::CircleTestType APM2D::InCircleVVV_V(DOUBLE dV0_x, DOUBLE dV0_y,
					   DOUBLE dV1_x, DOUBLE dV1_y,
					   DOUBLE dV2_x, DOUBLE dV2_y,
					   DOUBLE dV3_x, DOUBLE dV3_y)
{
  APM2D::CircleTestType eResult;

  //////////////////////////////////////////////////
  // Degenerate Touching Check
  if ( ( dV0_x == dV1_x && dV0_y == dV1_y ) &&
       ( dV0_x == dV2_x && dV0_y == dV2_y ) )
    {
      if ( dV0_x == dV3_x && dV0_y == dV3_y )
	{
	  return APM2D::cttOnCircle;
	}
      else
	{
	  return APM2D::cttOutCircle;
	}
    }

  eResult = APM2D::cttOnCircle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InCircleVVV_VLocal(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y,
				     dV3_x, dV3_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InCircleVVV_VLocal(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y,
					dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InCircleVVV_VMPFUN(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y,
				     dV3_x, dV3_y);
      }
    break;

    case APM::atMPFUNFilter:
      {
	eResult = InCircleVVV_VMPFUNFilter(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y,
				     dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InCircleVVV_VMPFUNNormalized(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y,
					       dV3_x, dV3_y);
      }
    break;
    default:
      ASSERT( FALSE );
	eResult = InCircleVVV_VAdaptive(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y,
					dV3_x, dV3_y);
      break;
    }  

  return eResult;
}

static APM2D::CircleTestType InCircleVVV_VLocal(DOUBLE dV0_x, DOUBLE dV0_y,
						DOUBLE dV1_x, DOUBLE dV1_y,
						DOUBLE dV2_x, DOUBLE dV2_y,
						DOUBLE dV3_x, DOUBLE dV3_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE dV1dotV1, dV2dotV2, dV3dotV3;
  DOUBLE dV1xV2, dV2xV3, dV3xV1;
  DOUBLE dDen;
  DOUBLE dResult;
  APM2D::CircleTestType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = dV1_x - dV0_x;
  pppdV[1][0][1] = dV1_y - dV0_y;
  pppdV[2][0][0] = dV2_x - dV0_x;
  pppdV[2][0][1] = dV2_y - dV0_y;
  pppdV[3][0][0] = dV3_x - dV0_x;
  pppdV[3][0][1] = dV3_y - dV0_y;
  
  dV1dotV1 = (pppdV[1][0][0]*pppdV[1][0][0]) + (pppdV[1][0][1]*pppdV[1][0][1]);
  dV2dotV2 = (pppdV[2][0][0]*pppdV[2][0][0]) + (pppdV[2][0][1]*pppdV[2][0][1]);
  dV3dotV3 = (pppdV[3][0][0]*pppdV[3][0][0]) + (pppdV[3][0][1]*pppdV[3][0][1]);

  dV1xV2 = (pppdV[1][0][0]*pppdV[2][0][1]) - (pppdV[1][0][1]*pppdV[2][0][0]);
  dV2xV3 = (pppdV[2][0][0]*pppdV[3][0][1]) - (pppdV[2][0][1]*pppdV[3][0][0]);
  dV3xV1 = (pppdV[3][0][0]*pppdV[1][0][1]) - (pppdV[3][0][1]*pppdV[1][0][0]);

  dDen = dV1xV2; 
  ASSERT( dDen != 0.0 );

  dResult = ( dV1dotV1*dV2xV3 + dV2dotV2*dV3xV1 + dV3dotV3*dV1xV2 ) / dDen;
  
  if ( dResult < -s_cdEpsilonInCircleVVV_VOnCircle )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( dResult > s_cdEpsilonInCircleVVV_VOnCircle )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

static APM2D::CircleTestType InCircleVVV_VAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dV1_x, DOUBLE dV1_y,
						   DOUBLE dV2_x, DOUBLE dV2_y,
						   DOUBLE dV3_x, DOUBLE dV3_y)
{
  DOUBLE d;
  double ppdV0[2], ppdV1[2], ppdV2[2], ppdV3[2];
  
  ppdV0[0] = dV0_x;
  ppdV0[1] = dV0_y;
  ppdV1[0] = dV1_x;
  ppdV1[1] = dV1_y;
  ppdV2[0] = dV2_x;
  ppdV2[1] = dV2_y;
  ppdV3[0] = dV3_x;
  ppdV3[1] = dV3_y;
  
  d = incircleexact(ppdV0, ppdV1, ppdV2, ppdV3);

  if ( d == 0.0 )
    {
      return APM2D::cttOnCircle;
    }
  else if ( d > 0.0 )
    {
      return APM2D::cttInCircle;
    }
  else
    {
      ASSERT( d < 0.0 );
      return APM2D::cttOutCircle;
    }
}

APM2D::CircleTestType APM2D::InCircleVVV_E(DOUBLE dV0_x, DOUBLE dV0_y,
					   DOUBLE dV1_x, DOUBLE dV1_y,
					   DOUBLE dV2_x, DOUBLE dV2_y,
					   DOUBLE dE30_x, DOUBLE dE30_y,
					   DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::CircleTestType eResult;

  //////////////////////////////////////////////////
  // Degenerate Touching Check
  if ( ( dV0_x == dV1_x && dV0_y == dV1_y ) &&
       ( dV0_x == dV2_x && dV0_y == dV2_y ) )
    {
      if ( ( dV0_x == dE30_x && dV0_y == dE30_y ) ||
	   ( dV0_x == dE31_x && dV0_y == dE31_y ) )
	{
	  return APM2D::cttOnCircle;
	}
      else
	{
	  return APM2D::cttOutCircle;
	}
    }

  eResult = APM2D::cttOnCircle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InCircleVVV_ELocal(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InCircleVVV_EAdaptive(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y,
					dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InCircleVVV_EMPFUN(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;

    case APM::atMPFUNFilter:
      {
	eResult = InCircleVVV_EMPFUNFilter(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InCircleVVV_EMPFUNNormalized(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y,
					       dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::CircleTestType InCircleVVV_ELocal(DOUBLE dV0_x, DOUBLE dV0_y,
						DOUBLE dV1_x, DOUBLE dV1_y,
						DOUBLE dV2_x, DOUBLE dV2_y,
						DOUBLE dE30_x, DOUBLE dE30_y,
						DOUBLE dE31_x, DOUBLE dE31_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE pdV1V2[2];
  DOUBLE dV1dotV1, dV2dotV2, dE3dotE3, dE3dotV1, dE3dotV2, dV1V2dotV1V2;
  DOUBLE dV1xV2, dE30xE31;
  DOUBLE dDen;
  DOUBLE dResult;
  APM2D::CircleTestType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = dV1_x - dV0_x;
  pppdV[1][0][1] = dV1_y - dV0_y;
  pppdV[2][0][0] = dV2_x - dV0_x;
  pppdV[2][0][1] = dV2_y - dV0_y;
  pppdV[3][0][0] = dE30_x - dV0_x;
  pppdV[3][0][1] = dE30_y - dV0_y;
  pppdV[3][1][0] = dE31_x - dV0_x;
  pppdV[3][1][1] = dE31_y - dV0_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[3][0] = dE31_x - dE30_x;
  ppdE[3][1] = dE31_y - dE30_y;

  pdV1V2[0] = dV2_x - dV1_x;
  pdV1V2[1] = dV2_y - dV1_y;
  
  dV1dotV1 = (pppdV[1][0][0]*pppdV[1][0][0]) + (pppdV[1][0][1]*pppdV[1][0][1]);
  dV2dotV2 = (pppdV[2][0][0]*pppdV[2][0][0]) + (pppdV[2][0][1]*pppdV[2][0][1]);
  dE3dotE3 = (ppdE[3][0]*ppdE[3][0]) + (ppdE[3][1]*ppdE[3][1]);
  dE3dotV1 = (ppdE[3][0]*pppdV[1][0][0]) + (ppdE[3][1]*pppdV[1][0][1]);
  dE3dotV2 = (ppdE[3][0]*pppdV[2][0][0]) + (ppdE[3][1]*pppdV[2][0][1]);

  dV1V2dotV1V2 = (pdV1V2[0]*pdV1V2[0]) + (pdV1V2[1]*pdV1V2[1]);

  dV1xV2 = (pppdV[1][0][0]*pppdV[2][0][1]) - (pppdV[1][0][1]*pppdV[2][0][0]);
  dE30xE31 = (pppdV[3][0][0]*pppdV[3][1][1]) - (pppdV[3][0][1]*pppdV[3][1][0]);

  dDen = 4.0*dE3dotE3*dV1xV2*dV1xV2; 
  ASSERT( dDen != 0.0 );

  dResult = ( dV1dotV1*dV1dotV1*dE3dotV2*dE3dotV2 + dV2dotV2*dV2dotV2*dE3dotV1*dE3dotV1 
	      - dV1dotV1*dV2dotV2*(dV1V2dotV1V2*dE3dotE3 + 2.0*dE3dotV1*dE3dotV2)
	      + 4.0*dV1xV2*dE30xE31*(dV1xV2*dE30xE31 + 
				     dV2dotV2*dE3dotV1 - dV1dotV1*dE3dotV2) ) / dDen;
  
  if ( dResult < -s_cdEpsilonInCircleVVV_EOnCircle )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( dResult > s_cdEpsilonInCircleVVV_EOnCircle )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

static APM2D::CircleTestType InCircleVVV_EAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dV1_x, DOUBLE dV1_y,
						   DOUBLE dV2_x, DOUBLE dV2_y,
						   DOUBLE dE30_x, DOUBLE dE30_y,
						   DOUBLE dE31_x, DOUBLE dE31_y)
{
  return InCircleVVV_ELocal(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y,
			    dE30_x, dE30_y, dE31_x, dE31_y);
}

APM2D::InscribedCircleCenterType 
APM2D::InscribedCircleCenterVVE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dV00_x, DOUBLE dV00_y,
				DOUBLE dV0_x, DOUBLE dV0_y,
				DOUBLE dV01_x, DOUBLE dV01_y,
				DOUBLE dV10_x, DOUBLE dV10_y,
				DOUBLE dV1_x, DOUBLE dV1_y,
				DOUBLE dV11_x, DOUBLE dV11_y,
				DOUBLE dE20_x, DOUBLE dE20_y, 
				DOUBLE dE21_x, DOUBLE dE21_y)
{
  ASSERT( pdC_x != NULL && pdC_y != NULL && pdR != NULL );

  APM2D::InscribedCircleCenterType eResult;

  eResult = APM2D::icctNull;
  
  //////////////////////////////////////////////////
  // CCW Checks

  if ( !Counterclockwise(dV0_x, dV0_y, dE20_x, dE20_y, dE21_x, dE21_y) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  if ( !Counterclockwise(dV1_x, dV1_y, dE20_x, dE20_y, dE21_x, dE21_y) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  if ( !( Counterclockwise(dV0_x, dV0_y, dV1_x, dV1_y, dE20_x, dE20_y) ||
	  Counterclockwise(dV0_x, dV0_y, dV1_x, dV1_y, dE21_x, dE21_y) ) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  //////////////////////////////////////////////////
  // Anitparallel Check

  if ( Parallel(dV0_x, dV0_y, dV1_x, dV1_y, dE20_x, dE20_y, dE21_x, dE21_y) )
    {
      ASSERT( OrientationEdotE(dV0_x, dV0_y, dV1_x, dV1_y, 
			       dE20_x, dE20_y, dE21_x, dE21_y) < 0 );
      switch ( APM::GetAccuracyType() )
	{
	case APM::atGlobal:
	case APM::atLocal:
	  {
	    eResult = InscribedCircleCenterVVEAntiparallelLocal(pdC_x, pdC_y, pdR, 
								dV00_x, dV00_y, 
								dV0_x, dV0_y, 
								dV01_x, dV01_y, 
								dV10_x, dV10_y, 
								dV1_x, dV1_y, 
								dV11_x, dV11_y, 
								dE20_x, dE20_y, 
								dE21_x, dE21_y);
	  }
	break;
	case APM::atExact:
	  {
	    eResult = InscribedCircleCenterVVEAntiparallelAdaptive(pdC_x, pdC_y, pdR, 
								   dV00_x, dV00_y, 
								   dV0_x, dV0_y, 
								   dV01_x, dV01_y, 
								   dV10_x, dV10_y, 
								   dV1_x, dV1_y, 
								   dV11_x, dV11_y, 
								   dE20_x, dE20_y, 
								   dE21_x, dE21_y);
	  }
	break;	
	case APM::atMPFUN:
	  {
	    eResult = InscribedCircleCenterVVEAntiparallelMPFUN(pdC_x, pdC_y, pdR, 
								dV00_x, dV00_y, 
								dV0_x, dV0_y, 
								dV01_x, dV01_y, 
								dV10_x, dV10_y, 
								dV1_x, dV1_y, 
								dV11_x, dV11_y, 
								dE20_x, dE20_y, 
								dE21_x, dE21_y);
	  }
	break;	

	case APM::atMPFUNFilter:
	  {
	    eResult = InscribedCircleCenterVVEAntiparallelMPFUNFilter(pdC_x, pdC_y, pdR, 
								dV00_x, dV00_y, 
								dV0_x, dV0_y, 
								dV01_x, dV01_y, 
								dV10_x, dV10_y, 
								dV1_x, dV1_y, 
								dV11_x, dV11_y, 
								dE20_x, dE20_y, 
								dE21_x, dE21_y);
	  }
	break;	
	case APM::atMPFUNNormalized:
	  {
	    eResult = InscribedCircleCenterVVEAntiparallelMPFUNNormalized(pdC_x, pdC_y, pdR, 
									  dV00_x, dV00_y, 
									  dV0_x, dV0_y, 
									  dV01_x, dV01_y, 
									  dV10_x, dV10_y, 
									  dV1_x, dV1_y, 
									  dV11_x, dV11_y, 
									  dE20_x, dE20_y, 
									  dE21_x, dE21_y);
	  }
	break;	
	default:
	  ASSERT( FALSE );
	  break;
	}
    }
  else
    {
      switch ( APM::GetAccuracyType() )
	{
	case APM::atGlobal:
	case APM::atLocal:
	  {
	    eResult = InscribedCircleCenterVVELocal(pdC_x, pdC_y, pdR, 
						    dV00_x, dV00_y, 
						    dV0_x, dV0_y, 
						    dV01_x, dV01_y, 
						    dV10_x, dV10_y, 
						    dV1_x, dV1_y, 
						    dV11_x, dV11_y, 
						    dE20_x, dE20_y, dE21_x, dE21_y);
	  }
	break;
	case APM::atExact:
	  {
	    eResult = InscribedCircleCenterVVEAdaptive(pdC_x, pdC_y, pdR, 
						       dV00_x, dV00_y, 
						       dV0_x, dV0_y, 
						       dV01_x, dV01_y, 
						       dV10_x, dV10_y, 
						       dV1_x, dV1_y, 
						       dV11_x, dV11_y, 
						       dE20_x, dE20_y, dE21_x, dE21_y);
	  }
	break;
	case APM::atMPFUN:
	  {
	    eResult = InscribedCircleCenterVVEMPFUN(pdC_x, pdC_y, pdR, 
						    dV00_x, dV00_y, 
						    dV0_x, dV0_y, 
						    dV01_x, dV01_y, 
						    dV10_x, dV10_y, 
						    dV1_x, dV1_y, 
						    dV11_x, dV11_y, 
						    dE20_x, dE20_y, dE21_x, dE21_y);
	  }
	break;

	case APM::atMPFUNFilter:
	  {
	    eResult = InscribedCircleCenterVVEMPFUNFilter(pdC_x, pdC_y, pdR, 
						    dV00_x, dV00_y, 
						    dV0_x, dV0_y, 
						    dV01_x, dV01_y, 
						    dV10_x, dV10_y, 
						    dV1_x, dV1_y, 
						    dV11_x, dV11_y, 
						    dE20_x, dE20_y, dE21_x, dE21_y);
	  }
	break;
	case APM::atMPFUNNormalized:
	  {
	    eResult = InscribedCircleCenterVVEMPFUNNormalized(pdC_x, pdC_y, pdR, 
							      dV00_x, dV00_y, 
							      dV0_x, dV0_y, 
							      dV01_x, dV01_y, 
							      dV10_x, dV10_y, 
							      dV1_x, dV1_y, 
							      dV11_x, dV11_y, 
							      dE20_x, dE20_y, dE21_x, dE21_y);
	  }
	break;
	default:
	  ASSERT( FALSE );
	  break;
	}
    }

  return eResult;
}

static APM2D::InscribedCircleCenterType 
InscribedCircleCenterVVELocal(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			      DOUBLE dV00_x, DOUBLE dV00_y,
			      DOUBLE dV0_x, DOUBLE dV0_y,
			      DOUBLE dV01_x, DOUBLE dV01_y,
			      DOUBLE dV10_x, DOUBLE dV10_y,
			      DOUBLE dV1_x, DOUBLE dV1_y,
			      DOUBLE dV11_x, DOUBLE dV11_y,
			      DOUBLE dE20_x, DOUBLE dE20_y, 
			      DOUBLE dE21_x, DOUBLE dE21_y)
{
  DOUBLE pdC[3], dR;
  DOUBLE pppdV[3][2][2], pppdVN[2][2][2], pdE[2];
  DOUBLE dDen, dDisc, dDiscR;
  DOUBLE dV1dotV1, dE2dotE2, dE2dotV1, dE2dotE20, dE2dotE21;
  DOUBLE dE2xV1, dE20xE21;
  DOUBLE dV1xV00, dV1xV01, dV1xV10, dV1xV11;
  DOUBLE dE2xV00, dE2xV01, dE2xV10, dE2xV11;
  DOUBLE dV1dotV10, dV1dotV11;
  DOUBLE dV00dotV00, dV01dotV01, dV10dotV10, dV11dotV11;
  DOUBLE dMagCE20, dMagCE21;
  DOUBLE dSinTheta, dSinThetaTemp;
  DOUBLE ppdVertexZone[2][2];
  DOUBLE pdEdgeZone[2];
  UINT_32 i, j;
  
  // Translate V0 to the origin
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[1][0][0] = dV1_x - dV0_x;
  pppdV[1][0][1] = dV1_y - dV0_y;
  pppdV[2][0][0] = dE20_x - dV0_x;
  pppdV[2][0][1] = dE20_y - dV0_y;
  pppdV[2][1][0] = dE21_x - dV0_x;
  pppdV[2][1][1] = dE21_y - dV0_y;
  
  // Vertex neighbor vectors
  memset(pppdVN, 0, sizeof(pppdV));
  pppdVN[0][0][0] = dV0_x - dV00_x;
  pppdVN[0][0][1] = dV0_y - dV00_y;
  pppdVN[0][1][0] = dV0_x - dV01_x;
  pppdVN[0][1][1] = dV0_y - dV01_y;
  pppdVN[1][0][0] = dV1_x - dV10_x;
  pppdVN[1][0][1] = dV1_y - dV10_y;
  pppdVN[1][1][0] = dV1_x - dV11_x;
  pppdVN[1][1][1] = dV1_y - dV11_y;

  pdE[0] = dE21_x - dE20_x;
  pdE[1] = dE21_y - dE20_y;
  
  dV1dotV1 = (pppdV[1][0][0]*pppdV[1][0][0]) + (pppdV[1][0][1]*pppdV[1][0][1]);
  dE2dotE2 = (pdE[0]*pdE[0]) + (pdE[1]*pdE[1]);
  dE2dotV1 = (pdE[0]*pppdV[1][0][0]) + (pdE[1]*pppdV[1][0][1]);

  dE2dotE20 = (pdE[0]*pppdV[2][0][0]) + (pdE[1]*pppdV[2][0][1]);
  dE2dotE21 = (pdE[0]*pppdV[2][1][0]) + (pdE[1]*pppdV[2][1][1]);

  dE2xV1 = (pdE[0]*pppdV[1][0][1]) - (pdE[1]*pppdV[1][0][0]);
  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);
  
  dV00dotV00 = ( (pppdVN[0][0][0] * pppdVN[0][0][0]) + (pppdVN[0][0][1] * pppdVN[0][0][1]) );
  dV01dotV01 = ( (pppdVN[0][1][0] * pppdVN[0][1][0]) + (pppdVN[0][1][1] * pppdVN[0][1][1]) );
  dV10dotV10 = ( (pppdVN[1][0][0] * pppdVN[1][0][0]) + (pppdVN[1][0][1] * pppdVN[1][0][1]) );
  dV11dotV11 = ( (pppdVN[1][1][0] * pppdVN[1][1][0]) + (pppdVN[1][1][1] * pppdVN[1][1][1]) );

  dV1dotV10 = ( (pppdV[1][0][0] * pppdVN[1][0][0]) + (pppdV[1][0][1] * pppdVN[1][0][1]) );
  dV1dotV11 = ( (pppdV[1][0][0] * pppdVN[1][1][0]) + (pppdV[1][0][1] * pppdVN[1][1][1]) );

  dV1xV00 = ( (pppdV[1][0][0] * pppdVN[0][0][1]) - (pppdV[1][0][1] * pppdVN[0][0][0]) );
  dV1xV01 = ( (pppdV[1][0][0] * pppdVN[0][1][1]) - (pppdV[1][0][1] * pppdVN[0][1][0]) );
  dV1xV10 = ( (pppdV[1][0][0] * pppdVN[1][0][1]) - (pppdV[1][0][1] * pppdVN[1][0][0]) );
  dV1xV11 = ( (pppdV[1][0][0] * pppdVN[1][1][1]) - (pppdV[1][0][1] * pppdVN[1][1][0]) );

  dE2xV00 = ( (pdE[0] * pppdVN[0][0][1]) - (pdE[1] * pppdVN[0][0][0]) );
  dE2xV01 = ( (pdE[0] * pppdVN[0][1][1]) - (pdE[1] * pppdVN[0][1][0]) );
  dE2xV10 = ( (pdE[0] * pppdVN[1][0][1]) - (pdE[1] * pppdVN[1][0][0]) );
  dE2xV11 = ( (pdE[0] * pppdVN[1][1][1]) - (pdE[1] * pppdVN[1][1][0]) );

  dSinTheta = APM2D::SinTheta(dV1_x, dV1_y, dE20_x, dE20_y, dE21_x, dE21_y);
  dSinThetaTemp = APM2D::SinTheta(dE20_x, dE20_y, dE21_x, dE21_y, dV1_x, dV1_y);
  if ( dSinThetaTemp > dSinTheta )
    {
      dSinTheta = dSinThetaTemp;
    }
  if ( dSinTheta <= s_cdEpsilonInscribedCircleCenterVVEVertexOnEdge )
    {
      // V1 is on E2 and it is not one of the endpoints
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  dSinTheta = APM2D::SinTheta(dV0_x, dV0_y, dE20_x, dE20_y, dE21_x, dE21_y);
  dSinThetaTemp = APM2D::SinTheta(dE20_x, dE20_y, dE21_x, dE21_y, dV0_x, dV0_y);
  if ( dSinThetaTemp > dSinTheta )
    {
      dSinTheta = dSinThetaTemp;
    }
  if ( dSinTheta <= s_cdEpsilonInscribedCircleCenterVVEVertexOnEdge )
    {
      // V0 is on E1 and it is not one of the endpoints
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  dSinTheta = dE2xV1/sqrt(dE2dotE2*dV1dotV1);
  if ( fabs(dSinTheta) < s_cdEpsilonInscribedCircleCenterVVEAntiparallel )
    {
      // The two vertices lie on a line parallel to the edge
#ifdef ENABLE_DEBUGGING_OUTPUT
      fprintf(stderr, "    VVE: Misclassification: running VVE Antiparallel: %f < %f\n",
	      dE2xV1, s_cdEpsilonInscribedCircleCenterVVEAntiparallel);
#endif // ENABLE_DEBUGGING_OUTPUT

      return InscribedCircleCenterVVEAntiparallelLocal(pdC_x, pdC_y, pdR, 
						       dV00_x, dV00_y, 
						       dV0_x, dV0_y, 
						       dV01_x, dV01_y, 
						       dV10_x, dV10_y, 
						       dV1_x, dV1_y, 
						       dV11_x, dV11_y, 
						       dE20_x, dE20_y, 
						       dE21_x, dE21_y);
    }

  dDen = 2.0*dE2xV1*dE2xV1;
  ASSERT( dDen != 0.0 );

  dDisc = dV1dotV1*dE2dotE2*dE20xE21*( dE20xE21 + dE2xV1 );
  if ( -s_cdEpsilonInscribedCircleCenterVVEDisc <= dDisc && dDisc <= 0.0 )
    {
      dDisc = 0.0;
    }
  dDiscR = dV1dotV1*dE20xE21*( dE20xE21 + dE2xV1 );
  if ( dE2xV1 == 0.0 || 
       (-s_cdEpsilonInscribedCircleCenterVVEDisc <= dDiscR && dDiscR <= 0.0) )
    {
      dDiscR = 0.0;
    }
  if ( dDisc < 0.0 || dDiscR < 0.0 ) 
    {
      // Imaginary Solutions
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  //////////////////////////////////////////////////
  // Compute the center and radius

  pdC[0] = ( -( dV1dotV1*dE2xV1*pdE[1] + 2.0*dE20xE21*dE2dotV1*pppdV[1][0][1] )
	     - 2.0 * pppdV[1][0][1] * sqrt(dDisc) ) / dDen;
  pdC[1] = ( ( dV1dotV1*dE2xV1*pdE[0] + 2.0*dE20xE21*dE2dotV1*pppdV[1][0][0] )
	     + 2.0*pppdV[1][0][0]*sqrt(dDisc) ) / dDen;
  pdC[2] = sqrt((pdC[0]*pdC[0]) + (pdC[1]*pdC[1]));

  dR = ( dV1dotV1*sqrt(dE2dotE2) * (2.0*dE20xE21 + dE2xV1) 
	 + 2.0*dE2dotV1*sqrt(dDiscR) ) / dDen;
  
  //////////////////////////////////////////////////
  // Check the Edge Zone

  dMagCE20 = ( (pdC[0] - pppdV[2][0][0])*(pdC[0] - pppdV[2][0][0]) + 
	       (pdC[1] - pppdV[2][0][1])*(pdC[1] - pppdV[2][0][1]) );
  dMagCE21 = ( (pdC[0] - pppdV[2][1][0])*(pdC[0] - pppdV[2][1][0]) + 
	       (pdC[1] - pppdV[2][1][1])*(pdC[1] - pppdV[2][1][1]) );

  if ( dMagCE20 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[0] = 0.0;
    }
  else
    {
      pdEdgeZone[0] = -( ( dE2xV1*dE2dotE20 + dE20xE21*dE2dotV1 + sqrt(dDisc) ) / 
			 ( dE2xV1*sqrt(dE2dotE2*dMagCE20) ) );
    }
  if ( dMagCE21 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[1] = 0.0;
    }
  else
    {
      pdEdgeZone[1] = ( ( dE2xV1*dE2dotE21 + dE20xE21*dE2dotV1 + sqrt(dDisc) ) / 
			( dE2xV1*sqrt(dE2dotE2*dMagCE21) ) );    
    }

  for ( i = 0; i < 2; i++ )
    {
      if ( pdEdgeZone[i] < -s_cdEpsilonInscribedCircleCenterVVEEdgeZone )
	{
#ifdef DEBUG_EPSILON_VALUES
	  if ( !( pdEdgeZone[i] < -(s_cdEpsilonInscribedCircleCenterVVEEdgeZone*s_cdOrderOfMagnitude) ) )
	    {
	      fprintf(stderr, "  VVE: Edge Zone: %e < %e < %e\n",
		      s_cdEpsilonInscribedCircleCenterVVEEdgeZone*s_cdOrderOfMagnitude,
		      pdEdgeZone[i],
		      s_cdEpsilonInscribedCircleCenterVVEEdgeZone);
	    }
#endif // DEBUG_EPSILON_VALUES	      
	  *pdC_x = 0.0;
	  *pdC_y = 0.0;
	  *pdR = 0.0;
	  return APM2D::icctNull;
	}
    }

  //////////////////////////////////////////////////
  // Check the Vertex Zones

  ppdVertexZone[0][0] = ( ( dV1dotV1*dE2xV1*dE2xV00 
			    + 2.0*dE20xE21*dE2dotV1*dV1xV00 
			    + 2.0*dV1xV00*sqrt(dDisc) ) /
			  ( 2.0*dE2xV1*dE2xV1*dR*sqrt(dV00dotV00) ) );
  ppdVertexZone[0][1] = ( ( dV1dotV1*dE2xV1*dE2xV01 
			    + 2.0*dE20xE21*dE2dotV1*dV1xV01 
			    + 2.0*dV1xV01*sqrt(dDisc) ) /
			  ( 2.0*dE2xV1*dE2xV1*dR*sqrt(dV01dotV01) ) );
  ppdVertexZone[1][0] = ( ( dV1dotV1*dE2xV1*dE2xV10 
			    + 2.0*dE20xE21*dE2dotV1*dV1xV10 
			    + 2.0*dV1xV10*sqrt(dDisc)
			    - 2.0*dE2xV1*dE2xV1*dV1dotV10 ) /
			  ( 2.0*dE2xV1*dE2xV1*dR*sqrt(dV10dotV10) ) );
  ppdVertexZone[1][1] = ( ( dV1dotV1*dE2xV1*dE2xV11 
			    + 2.0*dE20xE21*dE2dotV1*dV1xV11 
			    + 2.0*dV1xV11*sqrt(dDisc)
			    - 2.0*dE2xV1*dE2xV1*dV1dotV11 ) /
			  ( 2.0*dE2xV1*dE2xV1*dR*sqrt(dV11dotV11) ) );

  for ( i = 0; i < 2; i++ )
    {
      for ( j = 0; j < 2; j++ )
	{
	  if ( ppdVertexZone[i][j] < -s_cdEpsilonInscribedCircleCenterVVEVertexZone )
	    {
#ifdef DEBUG_EPSILON_VALUES
	      if ( !( ppdVertexZone[i][j] < -(s_cdEpsilonInscribedCircleCenterVVEVertexZone*s_cdOrderOfMagnitude) ) )
		{
		  fprintf(stderr, "  VVE: Vertex Zone: %e < %e < %e\n",
			  s_cdEpsilonInscribedCircleCenterVVEVertexZone*s_cdOrderOfMagnitude,
			  ppdVertexZone[i][j],
			  s_cdEpsilonInscribedCircleCenterVVEVertexZone);
		}
#endif // DEBUG_EPSILON_VALUES	      
	      *pdC_x = 0.0;
	      *pdC_y = 0.0;
	      *pdR = 0.0;
	      return APM2D::icctNull;
	    }
	}
    }

  if ( !( fabs(pdC[2] - dR) < s_cdEpsilonInscribedCircleCenterVVERadius ) )
    {
      fprintf(stderr, "VVE: C = ( %f, %f, %f ) != %f = R\n", pdC[0], pdC[1], pdC[2], dR);
      //ASSERT( FALSE );
    }

  // Translate the answer back out away from the origin
  *pdC_x = dV0_x + pdC[0];
  *pdC_y = dV0_y + pdC[1];
  *pdR = dR;
  return APM2D::icctVVE;
}

static APM2D::InscribedCircleCenterType 
InscribedCircleCenterVVEAdaptive(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				 DOUBLE dV00_x, DOUBLE dV00_y,
				 DOUBLE dV0_x, DOUBLE dV0_y,
				 DOUBLE dV01_x, DOUBLE dV01_y,
				 DOUBLE dV10_x, DOUBLE dV10_y,
				 DOUBLE dV1_x, DOUBLE dV1_y,
				 DOUBLE dV11_x, DOUBLE dV11_y,
				 DOUBLE dE20_x, DOUBLE dE20_y, 
				 DOUBLE dE21_x, DOUBLE dE21_y)
{
  return InscribedCircleCenterVVELocal(pdC_x, pdC_y, pdR, 
				       dV00_x, dV00_y, 
				       dV0_x, dV0_y, 
				       dV01_x, dV01_y, 
				       dV10_x, dV10_y, 
				       dV1_x, dV1_y, 
				       dV11_x, dV11_y, 
				       dE20_x, dE20_y, dE21_x, dE21_y);
}

APM2D::EdgeZoneType APM2D::EdgeZoneVVE(DOUBLE dV0_x, DOUBLE dV0_y,
				       DOUBLE dV1_x, DOUBLE dV1_y,
				       DOUBLE dE20_x, DOUBLE dE20_y, 
				       DOUBLE dE21_x, DOUBLE dE21_y,
				       DOUBLE dE30_x, DOUBLE dE30_y,
				       DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::EdgeZoneType eResult;

  eResult = APM2D::eztMiddle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = EdgeZoneVVELocal(dV0_x, dV0_y, dV1_x, dV1_y,
				   dE20_x, dE20_y, dE21_x, dE21_y,
				   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atExact:
      {
	eResult = EdgeZoneVVEAdaptive(dV0_x, dV0_y, dV1_x, dV1_y,
				      dE20_x, dE20_y, dE21_x, dE21_y,
				      dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = EdgeZoneVVEMPFUN(dV0_x, dV0_y, dV1_x, dV1_y,
				   dE20_x, dE20_y, dE21_x, dE21_y,
				   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;

    case APM::atMPFUNFilter:
      {
	eResult = EdgeZoneVVEMPFUNFilter(dV0_x, dV0_y, dV1_x, dV1_y,
				   dE20_x, dE20_y, dE21_x, dE21_y,
				   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = EdgeZoneVVEMPFUNNormalized(dV0_x, dV0_y, dV1_x, dV1_y,
					     dE20_x, dE20_y, dE21_x, dE21_y,
					     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::EdgeZoneType EdgeZoneVVELocal(DOUBLE dV0_x, DOUBLE dV0_y,
					    DOUBLE dV1_x, DOUBLE dV1_y,
					    DOUBLE dE20_x, DOUBLE dE20_y, 
					    DOUBLE dE21_x, DOUBLE dE21_y,
					    DOUBLE dE30_x, DOUBLE dE30_y,
					    DOUBLE dE31_x, DOUBLE dE31_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE pdC[2];
  DOUBLE pdEdgeZone[2];
  DOUBLE dV1dotV1, dE2dotE2, dV1dotE2;
  DOUBLE dE3dotE30, dE3dotE3, dE3dotE31;
  DOUBLE dE2xV1, dE2xE3, dE3xV1, dE20xE21;
  DOUBLE dE2dotV1, dMagCE30, dMagCE31;
  DOUBLE dDisc;
  DOUBLE dDen, dSinTheta;
  APM2D::EdgeZoneType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = dV1_x - dV0_x;
  pppdV[1][0][1] = dV1_y - dV0_y;
  pppdV[2][0][0] = dE20_x - dV0_x;
  pppdV[2][0][1] = dE20_y - dV0_y;
  pppdV[2][1][0] = dE21_x - dV0_x;
  pppdV[2][1][1] = dE21_y - dV0_y;
  pppdV[3][0][0] = dE30_x - dV0_x;
  pppdV[3][0][1] = dE30_y - dV0_y;
  pppdV[3][1][0] = dE31_x - dV0_x;
  pppdV[3][1][1] = dE31_y - dV0_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;
  ppdE[3][0] = dE31_x - dE30_x;
  ppdE[3][1] = dE31_y - dE30_y;
  
  dE2xV1 = (ppdE[2][0]*pppdV[1][0][1]) - (ppdE[2][1]*pppdV[1][0][0]);

  dV1dotV1 = (pppdV[1][0][0]*pppdV[1][0][0]) + (pppdV[1][0][1]*pppdV[1][0][1]);
  dV1dotE2 = (pppdV[1][0][0]*ppdE[2][0]) + (pppdV[1][0][1]*ppdE[2][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);
  dE3dotE3 = (ppdE[3][0]*ppdE[3][0]) + (ppdE[3][1]*ppdE[3][1]);

  dSinTheta = dE2xV1/sqrt(dE2dotE2*dV1dotV1);
  if ( fabs(dSinTheta) < s_cdEpsilonInscribedCircleCenterVVEAntiparallel )
    {
      // The two vertices lie on a line parallel to the edge
      return EdgeZoneVVEAntiparallelLocal(dV0_x, dV0_y, 
					  dV1_x, dV1_y, 
					  dE20_x, dE20_y, dE21_x, dE21_y,
					  dE30_x, dE30_y, dE31_x, dE31_y);
    }

  dE3dotE30 = (ppdE[3][0]*pppdV[3][0][0]) + (ppdE[3][1]*pppdV[3][0][1]);
  dE3dotE31 = (ppdE[3][0]*pppdV[3][1][0]) + (ppdE[3][1]*pppdV[3][1][1]);
  
  dE2xE3 = (ppdE[2][0]*ppdE[3][1]) - (ppdE[2][1]*ppdE[3][0]);
  dE3xV1 = (ppdE[3][0]*pppdV[1][0][1]) - (ppdE[3][1]*pppdV[1][0][0]);
  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);

  dDisc = dV1dotV1*dE2dotE2*dE20xE21*( dE20xE21 + dE2xV1 );
  if ( -s_cdEpsilonInscribedCircleCenterVVEDisc <= dDisc && dDisc <= 0.0 )
    {
      dDisc = 0.0;
    }
  ASSERT( dDisc >= 0.0 ); 

  dDen = 2.0*dE2xV1*dE2xV1; 
  ASSERT( dDen != 0.0 );

  dE2dotV1 = (ppdE[2][0]*pppdV[1][0][0]) + (ppdE[2][1]*pppdV[1][0][1]);

  pdC[0] = ( -( dV1dotV1*dE2xV1*ppdE[2][1] + 2.0*dE20xE21*dE2dotV1*pppdV[1][0][1] )
	     - 2.0 * pppdV[1][0][1] * sqrt(dDisc) ) / dDen;
  pdC[1] = ( ( dV1dotV1*dE2xV1*ppdE[2][0] + 2.0*dE20xE21*dE2dotV1*pppdV[1][0][0] )
	     + 2.0*pppdV[1][0][0]*sqrt(dDisc) ) / dDen;

  dMagCE30 = ( (pdC[0] - pppdV[3][0][0])*(pdC[0] - pppdV[3][0][0]) + 
	       (pdC[1] - pppdV[3][0][1])*(pdC[1] - pppdV[3][0][1]) );
  dMagCE31 = ( (pdC[0] - pppdV[3][1][0])*(pdC[0] - pppdV[3][1][0]) + 
	       (pdC[1] - pppdV[3][1][1])*(pdC[1] - pppdV[3][1][1]) );

  if ( dMagCE30 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[0] = 0.0;
    }
  else
    {
      pdEdgeZone[0] = ( ( dE2xV1*( 2.0*dE2xV1*dE3dotE30 - dE2xE3*dV1dotV1 )
			  + 2.0*dE3xV1*( dE20xE21*dV1dotE2 + sqrt(dDisc) ) ) / 
			( dDen*sqrt(dE3dotE3*dMagCE30) ) );
    }
  if ( dMagCE31 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[1] = 0.0;
    }
  else
    {
      pdEdgeZone[1] = -( ( dE2xV1*( 2.0*dE2xV1*dE3dotE31 - dE2xE3*dV1dotV1 )
			   + 2.0*dE3xV1*( dE20xE21*dV1dotE2 + sqrt(dDisc) ) ) / 
			 ( dDen*sqrt(dE3dotE3*dMagCE31) ) );
    }

  if ( !( pdEdgeZone[0] < -s_cdEpsilonEdgeZoneVVEEdgeZone ) )
    {
      eResult = APM2D::eztStart;
    }
  else if ( !( pdEdgeZone[1] < -s_cdEpsilonEdgeZoneVVEEdgeZone ) )
    {
      eResult = APM2D::eztEnd;
    }
  else
    {
      eResult = APM2D::eztMiddle;
    }

  return eResult;
}

static APM2D::EdgeZoneType EdgeZoneVVEAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
					       DOUBLE dV1_x, DOUBLE dV1_y,
					       DOUBLE dE20_x, DOUBLE dE20_y, 
					       DOUBLE dE21_x, DOUBLE dE21_y,
					       DOUBLE dE30_x, DOUBLE dE30_y,
					       DOUBLE dE31_x, DOUBLE dE31_y)
{
  return EdgeZoneVVELocal(dV0_x, dV0_y, dV1_x, dV1_y,
			  dE20_x, dE20_y, dE21_x, dE21_y,
			  dE30_x, dE30_y, dE31_x, dE31_y);
}

APM2D::CircleTestType APM2D::InCircleVVE_V(DOUBLE dV0_x, DOUBLE dV0_y,
					   DOUBLE dV1_x, DOUBLE dV1_y,
					   DOUBLE dE20_x, DOUBLE dE20_y, 
					   DOUBLE dE21_x, DOUBLE dE21_y,
					   DOUBLE dV3_x, DOUBLE dV3_y)
{
  APM2D::CircleTestType eResult;

  eResult = APM2D::cttOnCircle;

  // Check to see if the vertex is in the edge envelope
  if ( !Counterclockwise(dE20_x, dE20_y, 
			 dE21_x, dE21_y,
			 dV3_x, dV3_y) )
    {
      return APM2D::cttOutCircle;
    }

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InCircleVVE_VLocal(dV0_x, dV0_y, dV1_x, dV1_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dV3_x, dV3_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InCircleVVE_VAdaptive(dV0_x, dV0_y, dV1_x, dV1_y,
					dE20_x, dE20_y, dE21_x, dE21_y,
					dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InCircleVVE_VMPFUN(dV0_x, dV0_y, dV1_x, dV1_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dV3_x, dV3_y);
      }
    break;

    case APM::atMPFUNFilter:
      {
	eResult = InCircleVVE_VMPFUNFilter(dV0_x, dV0_y, dV1_x, dV1_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InCircleVVE_VMPFUNNormalized(dV0_x, dV0_y, dV1_x, dV1_y,
					       dE20_x, dE20_y, dE21_x, dE21_y,
					       dV3_x, dV3_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::CircleTestType InCircleVVE_VLocal(DOUBLE dV0_x, DOUBLE dV0_y,
						DOUBLE dV1_x, DOUBLE dV1_y,
						DOUBLE dE20_x, DOUBLE dE20_y, 
						DOUBLE dE21_x, DOUBLE dE21_y,
						DOUBLE dV3_x, DOUBLE dV3_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE dV1dotV1, dE2dotE2, dV3dotV3, dV1dotE2;
  DOUBLE dE2xV1, dE2xV3, dV1xV3, dE20xE21;
  DOUBLE dDisc;
  DOUBLE dDen, dSinTheta;
  DOUBLE dResult;
  APM2D::CircleTestType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = dV1_x - dV0_x;
  pppdV[1][0][1] = dV1_y - dV0_y;
  pppdV[2][0][0] = dE20_x - dV0_x;
  pppdV[2][0][1] = dE20_y - dV0_y;
  pppdV[2][1][0] = dE21_x - dV0_x;
  pppdV[2][1][1] = dE21_y - dV0_y;
  pppdV[3][0][0] = dV3_x - dV0_x;
  pppdV[3][0][1] = dV3_y - dV0_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;

  dV1dotV1 = (pppdV[1][0][0]*pppdV[1][0][0]) + (pppdV[1][0][1]*pppdV[1][0][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);
  dV3dotV3 = (pppdV[3][0][0]*pppdV[3][0][0]) + (pppdV[3][0][1]*pppdV[3][0][1]);
  dV1dotE2 = (pppdV[1][0][0]*ppdE[2][0]) + (pppdV[1][0][1]*ppdE[2][1]);

  dE2xV1 = (ppdE[2][0]*pppdV[1][0][1]) - (ppdE[2][1]*pppdV[1][0][0]);
  dSinTheta = dE2xV1/sqrt(dE2dotE2*dV1dotV1);
  if ( fabs(dSinTheta) < s_cdEpsilonInscribedCircleCenterVVEAntiparallel )
    {
      // The two vertices lie on a line parallel to the edge
      return InCircleVVE_VAntiparallelLocal(dV0_x, dV0_y, 
					    dV1_x, dV1_y, 
					    dE20_x, dE20_y, dE21_x, dE21_y,
					    dV3_x, dV3_y);
    }

  dE2xV3 = (ppdE[2][0]*pppdV[3][0][1]) - (ppdE[2][1]*pppdV[3][0][0]);
  dV1xV3 = (pppdV[1][0][0]*pppdV[3][0][1]) - (pppdV[1][0][1]*pppdV[3][0][0]);
  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);

  dDisc = dV1dotV1*dE2dotE2*dE20xE21*( dE20xE21 + dE2xV1 );
  if ( -s_cdEpsilonInscribedCircleCenterVVEDisc <= dDisc && dDisc <= 0.0 )
    {
      dDisc = 0.0;
    }
  ASSERT( dDisc >= 0.0 ); 

  dDen = dE2xV1*dE2xV1; 
  ASSERT( dDen != 0.0 );

  dResult = ( dE2xV1*( dV3dotV3*dE2xV1 - dV1dotV1*dE2xV3 ) 
	      - 2.0*dV1xV3*( dE20xE21*dV1dotE2 + sqrt(dDisc) ) ) / dDen;
  
  if ( dResult < -s_cdEpsilonInCircleVVE_VOnCircle )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( dResult > s_cdEpsilonInCircleVVE_VOnCircle )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

static APM2D::CircleTestType InCircleVVE_VAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dV1_x, DOUBLE dV1_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   DOUBLE dV3_x, DOUBLE dV3_y)
{
  return InCircleVVE_VLocal(dV0_x, dV0_y, dV1_x, dV1_y,
			    dE20_x, dE20_y, dE21_x, dE21_y,
			    dV3_x, dV3_y);
}

APM2D::CircleTestType APM2D::InCircleVVE_E(DOUBLE dV0_x, DOUBLE dV0_y,
					   DOUBLE dV1_x, DOUBLE dV1_y,
					   DOUBLE dE20_x, DOUBLE dE20_y, 
					   DOUBLE dE21_x, DOUBLE dE21_y,
					   DOUBLE dE30_x, DOUBLE dE30_y,
					   DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::CircleTestType eResult;

  eResult = APM2D::cttOnCircle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InCircleVVE_ELocal(dV0_x, dV0_y, dV1_x, dV1_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InCircleVVE_EAdaptive(dV0_x, dV0_y, dV1_x, dV1_y,
					dE20_x, dE20_y, dE21_x, dE21_y,
					dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InCircleVVE_EMPFUN(dV0_x, dV0_y, dV1_x, dV1_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;

    case APM::atMPFUNFilter:
      {
	eResult = InCircleVVE_EMPFUNFilter(dV0_x, dV0_y, dV1_x, dV1_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InCircleVVE_EMPFUNNormalized(dV0_x, dV0_y, dV1_x, dV1_y,
					       dE20_x, dE20_y, dE21_x, dE21_y,
					       dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::CircleTestType InCircleVVE_ELocal(DOUBLE dV0_x, DOUBLE dV0_y,
						DOUBLE dV1_x, DOUBLE dV1_y,
						DOUBLE dE20_x, DOUBLE dE20_y, 
						DOUBLE dE21_x, DOUBLE dE21_y,
						DOUBLE dE30_x, DOUBLE dE30_y,
						DOUBLE dE31_x, DOUBLE dE31_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE dV1dotV1, dE2dotE2, dE3dotE3;
  DOUBLE dV1dotE2, dV1dotE3, dE2dotE3;
  DOUBLE dE2xV1, dE20xE21, dE30xE31;
  DOUBLE dDisc, dDiscR, dA, dB;
  DOUBLE dDen;
  DOUBLE dResult, dSinTheta;
  APM2D::CircleTestType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = dV1_x - dV0_x;
  pppdV[1][0][1] = dV1_y - dV0_y;
  pppdV[2][0][0] = dE20_x - dV0_x;
  pppdV[2][0][1] = dE20_y - dV0_y;
  pppdV[2][1][0] = dE21_x - dV0_x;
  pppdV[2][1][1] = dE21_y - dV0_y;
  pppdV[3][0][0] = dE30_x - dV0_x;
  pppdV[3][0][1] = dE30_y - dV0_y;
  pppdV[3][1][0] = dE31_x - dV0_x;
  pppdV[3][1][1] = dE31_y - dV0_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;
  ppdE[3][0] = dE31_x - dE30_x;
  ppdE[3][1] = dE31_y - dE30_y;

  dV1dotV1 = (pppdV[1][0][0]*pppdV[1][0][0]) + (pppdV[1][0][1]*pppdV[1][0][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);
  dE3dotE3 = (ppdE[3][0]*ppdE[3][0]) + (ppdE[3][1]*ppdE[3][1]);

  dE2xV1 = (ppdE[2][0]*pppdV[1][0][1]) - (ppdE[2][1]*pppdV[1][0][0]);
  dSinTheta = dE2xV1/sqrt(dE2dotE2*dV1dotV1);
  if ( fabs(dSinTheta) < s_cdEpsilonInscribedCircleCenterVVEAntiparallel )
    {
      // The two vertices lie on a line parallel to the edge
      return InCircleVVE_EAntiparallelLocal(dV0_x, dV0_y, 
					    dV1_x, dV1_y, 
					    dE20_x, dE20_y, dE21_x, dE21_y,
					    dE30_x, dE30_y, dE31_x, dE31_y);
    }

  dV1dotE2 = (pppdV[1][0][0]*ppdE[2][0]) + (pppdV[1][0][1]*ppdE[2][1]);
  dV1dotE3 = (pppdV[1][0][0]*ppdE[3][0]) + (pppdV[1][0][1]*ppdE[3][1]);
  dE2dotE3 = (ppdE[2][0]*ppdE[3][0]) + (ppdE[2][1]*ppdE[3][1]);

  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);
  dE30xE31 = (pppdV[3][0][0]*pppdV[3][1][1]) - (pppdV[3][0][1]*pppdV[3][1][0]);

  dDisc = dV1dotV1*dE2dotE2*dE20xE21*( dE20xE21 + dE2xV1 );
  if ( -s_cdEpsilonInscribedCircleCenterVVEDisc <= dDisc && dDisc <= 0.0 )
    {
      dDisc = 0.0;
    }
  ASSERT( dDisc >= 0.0 ); 

  dDiscR = dV1dotV1*dE20xE21*( dE20xE21 + dE2xV1 );
  if ( -s_cdEpsilonInscribedCircleCenterVVEDisc <= dDiscR && dDiscR <= 0.0 )
    {
      dDiscR = 0.0;
    }
  ASSERT( dDiscR >= 0.0 ); 

  dA = ( dV1dotV1*dE2xV1*dE2dotE3 + 2.0*dV1dotE3*( dE20xE21*dV1dotE2 + sqrt(dDisc) ) );
  dB = ( dV1dotV1*sqrt(dE2dotE2)*( 2.0*dE20xE21 + dE2xV1 ) + 2.0*dV1dotE2*sqrt(dDiscR) );

  dDen = 4.0*dE3dotE3*dE2xV1*dE2xV1*dE2xV1*dE2xV1;
  ASSERT( dDen != 0.0 );

  dResult = ( dA*dA
	      + 4.0*dE30xE31*dE2xV1*dE2xV1*dA
	      + 4.0*dE30xE31*dE30xE31*dE2xV1*dE2xV1*dE2xV1*dE2xV1
	      - dE3dotE3*dB*dB ) / dDen;
  
  if ( dResult < -s_cdEpsilonInCircleVVE_EOnCircle )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( dResult > s_cdEpsilonInCircleVVE_EOnCircle )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

static APM2D::CircleTestType InCircleVVE_EAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dV1_x, DOUBLE dV1_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   DOUBLE dE30_x, DOUBLE dE30_y,
						   DOUBLE dE31_x, DOUBLE dE31_y)
{
  return InCircleVVE_ELocal(dV0_x, dV0_y, dV1_x, dV1_y,
			    dE20_x, dE20_y, dE21_x, dE21_y,
			    dE30_x, dE30_y, dE31_x, dE31_y);
}

static APM2D::InscribedCircleCenterType 
InscribedCircleCenterVVEAntiparallelLocal(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					  DOUBLE dV00_x, DOUBLE dV00_y,
					  DOUBLE dV0_x, DOUBLE dV0_y,
					  DOUBLE dV01_x, DOUBLE dV01_y,
					  DOUBLE dV10_x, DOUBLE dV10_y,
					  DOUBLE dV1_x, DOUBLE dV1_y,
					  DOUBLE dV11_x, DOUBLE dV11_y,
					  DOUBLE dE20_x, DOUBLE dE20_y, 
					  DOUBLE dE21_x, DOUBLE dE21_y)
{
  DOUBLE pdC[3], dR;
  DOUBLE pppdV[3][2][2], pppdVN[2][2][2], pdE[2];
  DOUBLE dDen, dDenR;
  DOUBLE dE2dotE2, dE2dotV1, dE2dotE20, dE2dotE21;
  DOUBLE dE20xE21;
  DOUBLE dV1xV00, dV1xV01, dV1xV10, dV1xV11;
  DOUBLE dV1dotV00, dV1dotV01, dV1dotV10, dV1dotV11;
  DOUBLE dV00dotV00, dV01dotV01, dV10dotV10, dV11dotV11;
  DOUBLE dMagCE20, dMagCE21;
  DOUBLE dSinTheta, dSinThetaTemp;
  DOUBLE ppdVertexZone[2][2];
  DOUBLE pdEdgeZone[2];
  UINT_32 i, j;
  
  // Translate V0 to the origin
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[1][0][0] = dV1_x - dV0_x;
  pppdV[1][0][1] = dV1_y - dV0_y;
  pppdV[2][0][0] = dE20_x - dV0_x;
  pppdV[2][0][1] = dE20_y - dV0_y;
  pppdV[2][1][0] = dE21_x - dV0_x;
  pppdV[2][1][1] = dE21_y - dV0_y;

  // Vertex neighbor vectors
  memset(pppdVN, 0, sizeof(pppdV));
  pppdVN[0][0][0] = dV0_x - dV00_x;
  pppdVN[0][0][1] = dV0_y - dV00_y;
  pppdVN[0][1][0] = dV0_x - dV01_x;
  pppdVN[0][1][1] = dV0_y - dV01_y;
  pppdVN[1][0][0] = dV1_x - dV10_x;
  pppdVN[1][0][1] = dV1_y - dV10_y;
  pppdVN[1][1][0] = dV1_x - dV11_x;
  pppdVN[1][1][1] = dV1_y - dV11_y;
  
  pdE[0] = dE21_x - dE20_x;
  pdE[1] = dE21_y - dE20_y;
  
  dE2dotE2 = (pdE[0]*pdE[0]) + (pdE[1]*pdE[1]);
  dE2dotV1 = (pdE[0]*pppdV[1][0][0]) + (pdE[1]*pppdV[1][0][1]);

  dE2dotE20 = (pdE[0]*pppdV[2][0][0]) + (pdE[1]*pppdV[2][0][1]);
  dE2dotE21 = (pdE[0]*pppdV[2][1][0]) + (pdE[1]*pppdV[2][1][1]);

  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);
  
  dV00dotV00 = ( (pppdVN[0][0][0] * pppdVN[0][0][0]) + (pppdVN[0][0][1] * pppdVN[0][0][1]) );
  dV01dotV01 = ( (pppdVN[0][1][0] * pppdVN[0][1][0]) + (pppdVN[0][1][1] * pppdVN[0][1][1]) );
  dV10dotV10 = ( (pppdVN[1][0][0] * pppdVN[1][0][0]) + (pppdVN[1][0][1] * pppdVN[1][0][1]) );
  dV11dotV11 = ( (pppdVN[1][1][0] * pppdVN[1][1][0]) + (pppdVN[1][1][1] * pppdVN[1][1][1]) );

  dV1dotV00 = ( (pppdV[1][0][0] * pppdVN[0][0][0]) + (pppdV[1][0][1] * pppdVN[0][0][1]) );
  dV1dotV01 = ( (pppdV[1][0][0] * pppdVN[0][1][0]) + (pppdV[1][0][1] * pppdVN[0][1][1]) );
  dV1dotV10 = ( (pppdV[1][0][0] * pppdVN[1][0][0]) + (pppdV[1][0][1] * pppdVN[1][0][1]) );
  dV1dotV11 = ( (pppdV[1][0][0] * pppdVN[1][1][0]) + (pppdV[1][0][1] * pppdVN[1][1][1]) );

  dV1xV00 = ( (pppdV[1][0][0] * pppdVN[0][0][1]) - (pppdV[1][0][1] * pppdVN[0][0][0]) );
  dV1xV01 = ( (pppdV[1][0][0] * pppdVN[0][1][1]) - (pppdV[1][0][1] * pppdVN[0][1][0]) );
  dV1xV10 = ( (pppdV[1][0][0] * pppdVN[1][0][1]) - (pppdV[1][0][1] * pppdVN[1][0][0]) );
  dV1xV11 = ( (pppdV[1][0][0] * pppdVN[1][1][1]) - (pppdV[1][0][1] * pppdVN[1][1][0]) );

  dSinTheta = APM2D::SinTheta(dV1_x, dV1_y, dE20_x, dE20_y, dE21_x, dE21_y);
  dSinThetaTemp = APM2D::SinTheta(dE20_x, dE20_y, dE21_x, dE21_y, dV1_x, dV1_y);
  if ( dSinThetaTemp > dSinTheta )
    {
      dSinTheta = dSinThetaTemp;
    }
  if ( dSinTheta <= s_cdEpsilonInscribedCircleCenterVVEVertexOnEdge )
    {
      // V1 is on E2 and it is not one of the endpoints
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  //dSinTheta = dE20xE21/sqrt(dE20dotE20*dE21dotE21);
  dSinTheta = APM2D::SinTheta(dV0_x, dV0_y, dE20_x, dE20_y, dE21_x, dE21_y);
  dSinThetaTemp = APM2D::SinTheta(dE20_x, dE20_y, dE21_x, dE21_y, dV0_x, dV0_y);
  if ( dSinThetaTemp > dSinTheta )
    {
      dSinTheta = dSinThetaTemp;
    }
  if ( dSinTheta <= s_cdEpsilonInscribedCircleCenterVVEVertexOnEdge )
    {
      // V0 is on E2 and it is not one of the endpoints
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  dDen = -8.0*dE20xE21*dE2dotV1;
  dDenR = 8.0*dE20xE21*sqrt(dE2dotE2);
  if ( dDen == 0.0 || dDenR == 0.0 )
    {
      //ASSERT( FALSE );
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  //////////////////////////////////////////////////
  // Compute the center and radius

  pdC[0] = ( -(4.0*dE20xE21*dE20xE21 - dE2dotV1*dE2dotV1)*pppdV[1][0][1]
	     - (4.0*dE20xE21*dE2dotV1)*pppdV[1][0][0] ) / dDen;
  pdC[1] = ( (4.0*dE20xE21*dE20xE21 - dE2dotV1*dE2dotV1)*pppdV[1][0][0]
	     - (4.0*dE20xE21*dE2dotV1)*pppdV[1][0][1] ) / dDen;
  pdC[2] = sqrt((pdC[0]*pdC[0]) + (pdC[1]*pdC[1]));

  dR = (4.0*dE20xE21*dE20xE21 + dE2dotV1*dE2dotV1) / dDenR;
  
  //////////////////////////////////////////////////
  // Check the Edge Zone

  dMagCE20 = ( (pdC[0] - pppdV[2][0][0])*(pdC[0] - pppdV[2][0][0]) + 
	       (pdC[1] - pppdV[2][0][1])*(pdC[1] - pppdV[2][0][1]) );
  dMagCE21 = ( (pdC[0] - pppdV[2][1][0])*(pdC[0] - pppdV[2][1][0]) + 
	       (pdC[1] - pppdV[2][1][1])*(pdC[1] - pppdV[2][1][1]) );

  if ( dMagCE20 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[0] = 0.0;
    }
  else
    {
      pdEdgeZone[0] = -( ( 2.0*dE2dotE20 - dE2dotV1 ) / 
			 ( sqrt(dE2dotE2*dMagCE20) ) );
    }
  if ( dMagCE21 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[1] = 0.0;
    }
  else
    {
      pdEdgeZone[1] = ( ( 2.0*dE2dotE21 - dE2dotV1 ) /
			( sqrt(dE2dotE2*dMagCE21) ) );
    }

  for ( i = 0; i < 2; i++ )
    {
      if ( pdEdgeZone[i] < -s_cdEpsilonInscribedCircleCenterVVEAntiparallelEdgeZone )
	{
#ifdef DEBUG_EPSILON_VALUES
	  if ( !( pdEdgeZone[i] < -(s_cdEpsilonInscribedCircleCenterVVEAntiparallelEdgeZone*s_cdOrderOfMagnitude) ) )
	    {
	      fprintf(stderr, "  VVE Antiparallel: Edge Zone: %e < %e < %e\n",
		      s_cdEpsilonInscribedCircleCenterVVEAntiparallelEdgeZone*s_cdOrderOfMagnitude,
		      pdEdgeZone[i],
		      s_cdEpsilonInscribedCircleCenterVVEAntiparallelEdgeZone);
	    }
#endif // DEBUG_EPSILON_VALUES	      
	  *pdC_x = 0.0;
	  *pdC_y = 0.0;
	  *pdR = 0.0;
	  return APM2D::icctNull;
	}
    }

  //////////////////////////////////////////////////
  // Check the Vertex Zones

  ppdVertexZone[0][0] = ( ( (4.0*dE20xE21*dE20xE21 - dE2dotV1*dE2dotV1)*dV1xV00
			    - 4.0*dE20xE21*dE2dotV1*dV1dotV00 ) / 
			  dDen*dR*sqrt(dV00dotV00) );
  ppdVertexZone[0][1] = ( ( (4.0*dE20xE21*dE20xE21 - dE2dotV1*dE2dotV1)*dV1xV01
			    - 4.0*dE20xE21*dE2dotV1*dV1dotV01 ) / 
			  dDen*dR*sqrt(dV01dotV01) );
  ppdVertexZone[1][0] = ( ( (4.0*dE20xE21*dE20xE21 - dE2dotV1*dE2dotV1)*dV1xV10
			    + 4.0*dE20xE21*dE2dotV1*dV1dotV10 ) / 
			  dDen*dR*sqrt(dV10dotV10) );
  ppdVertexZone[1][1] = ( ( (4.0*dE20xE21*dE20xE21 - dE2dotV1*dE2dotV1)*dV1xV11
			    + 4.0*dE20xE21*dE2dotV1*dV1dotV11 ) / 
			  dDen*dR*sqrt(dV11dotV11) );

  for ( i = 0; i < 2; i++ )
    {
      for ( j = 0; j < 2; j++ )
	{
	  if ( ppdVertexZone[i][j] < -s_cdEpsilonInscribedCircleCenterVVEAntiparallelVertexZone )
	    {
#ifdef DEBUG_EPSILON_VALUES
	      if ( !( ppdVertexZone[i][j] < -(s_cdEpsilonInscribedCircleCenterVVEAntiparallelVertexZone*s_cdOrderOfMagnitude) ) )
		{
		  fprintf(stderr, "  VVV: Vertex Zone: %e < %e < %e\n",
			  s_cdEpsilonInscribedCircleCenterVVEAntiparallelVertexZone*s_cdOrderOfMagnitude,
			  ppdVertexZone[i][j],
			  s_cdEpsilonInscribedCircleCenterVVEAntiparallelVertexZone);
		}
#endif // DEBUG_EPSILON_VALUES	      
	      *pdC_x = 0.0;
	      *pdC_y = 0.0;
	      *pdR = 0.0;
	      return APM2D::icctNull;
	    }
	}
    }

  if ( !( fabs(pdC[2] - dR) < s_cdEpsilonInscribedCircleCenterVVEAntiparallelRadius ) )
    {
      fprintf(stderr, "VVE Antiparallel: C = ( %f, %f, %f ) != %f = R\n", pdC[0], pdC[1], pdC[2], dR);
      //ASSERT( FALSE );
    }

  // Translate the answer back out away from the origin
  *pdC_x = dV0_x + pdC[0];
  *pdC_y = dV0_y + pdC[1];
  *pdR = dR;
  return APM2D::icctVVEAntiparallel;
}

static APM2D::InscribedCircleCenterType 
InscribedCircleCenterVVEAntiparallelAdaptive(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					     DOUBLE dV00_x, DOUBLE dV00_y,
					     DOUBLE dV0_x, DOUBLE dV0_y,
					     DOUBLE dV01_x, DOUBLE dV01_y,
					     DOUBLE dV10_x, DOUBLE dV10_y,
					     DOUBLE dV1_x, DOUBLE dV1_y,
					     DOUBLE dV11_x, DOUBLE dV11_y,
					     DOUBLE dE20_x, DOUBLE dE20_y, 
					     DOUBLE dE21_x, DOUBLE dE21_y)
{
  return InscribedCircleCenterVVEAntiparallelLocal(pdC_x, pdC_y, pdR, 
						   dV00_x, dV00_y, 
						   dV0_x, dV0_y, 
						   dV01_x, dV01_y, 
						   dV10_x, dV10_y, 
						   dV1_x, dV1_y, 
						   dV11_x, dV11_y, 
						   dE20_x, dE20_y, dE21_x, dE21_y);
}

APM2D::EdgeZoneType APM2D::EdgeZoneVVEAntiparallel(DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dV1_x, DOUBLE dV1_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   DOUBLE dE30_x, DOUBLE dE30_y,
						   DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::EdgeZoneType eResult;

  eResult = APM2D::eztMiddle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = EdgeZoneVVEAntiparallelLocal(dV0_x, dV0_y, dV1_x, dV1_y,
					       dE20_x, dE20_y, dE21_x, dE21_y,
					       dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atExact:
      {
	eResult = EdgeZoneVVEAntiparallelAdaptive(dV0_x, dV0_y, dV1_x, dV1_y,
						  dE20_x, dE20_y, dE21_x, dE21_y,
						  dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = EdgeZoneVVEAntiparallelMPFUN(dV0_x, dV0_y, dV1_x, dV1_y,
					       dE20_x, dE20_y, dE21_x, dE21_y,
					       dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;

    case APM::atMPFUNFilter:
      {
	eResult = EdgeZoneVVEAntiparallelMPFUNFilter(dV0_x, dV0_y, dV1_x, dV1_y,
					       dE20_x, dE20_y, dE21_x, dE21_y,
					       dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = EdgeZoneVVEAntiparallelMPFUNNormalized(dV0_x, dV0_y, dV1_x, dV1_y,
							 dE20_x, dE20_y, dE21_x, dE21_y,
							 dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::EdgeZoneType EdgeZoneVVEAntiparallelLocal(DOUBLE dV0_x, DOUBLE dV0_y,
							DOUBLE dV1_x, DOUBLE dV1_y,
							DOUBLE dE20_x, DOUBLE dE20_y, 
							DOUBLE dE21_x, DOUBLE dE21_y,
							DOUBLE dE30_x, DOUBLE dE30_y,
							DOUBLE dE31_x, DOUBLE dE31_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE pdC[3];
  DOUBLE pdEdgeZone[2];
  DOUBLE dV1dotV1, dE2dotE2, dV1dotE3;
  DOUBLE dE3dotE30, dE3dotE31;
  DOUBLE dE3xV1, dE20xE21;
  DOUBLE dE3dotE3, dMagCE30, dMagCE31;
  DOUBLE dE2dotV1;
  DOUBLE dDen;
  APM2D::EdgeZoneType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = dV1_x - dV0_x;
  pppdV[1][0][1] = dV1_y - dV0_y;
  pppdV[2][0][0] = dE20_x - dV0_x;
  pppdV[2][0][1] = dE20_y - dV0_y;
  pppdV[2][1][0] = dE21_x - dV0_x;
  pppdV[2][1][1] = dE21_y - dV0_y;
  pppdV[3][0][0] = dE30_x - dV0_x;
  pppdV[3][0][1] = dE30_y - dV0_y;
  pppdV[3][1][0] = dE31_x - dV0_x;
  pppdV[3][1][1] = dE31_y - dV0_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;
  ppdE[3][0] = dE31_x - dE30_x;
  ppdE[3][1] = dE31_y - dE30_y;
  
  dV1dotV1 = (pppdV[1][0][0]*pppdV[1][0][0]) + (pppdV[1][0][1]*pppdV[1][0][1]);
  dV1dotE3 = (pppdV[1][0][0]*ppdE[3][0]) + (pppdV[1][0][1]*ppdE[3][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);
  dE3dotE3 = (ppdE[3][0]*ppdE[3][0]) + (ppdE[3][1]*ppdE[3][1]);

  dE3dotE30 = (ppdE[3][0]*pppdV[3][0][0]) + (ppdE[3][1]*pppdV[3][0][1]);
  dE3dotE31 = (ppdE[3][0]*pppdV[3][1][0]) + (ppdE[3][1]*pppdV[3][1][1]);
  
  dE3xV1 = (ppdE[3][0]*pppdV[1][0][1]) - (ppdE[3][1]*pppdV[1][0][0]);
  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);

  dDen = 8.0*dE20xE21*sqrt(dV1dotV1*dE2dotE2);
  ASSERT( dDen != 0.0 );

  dE2dotV1 = (ppdE[2][0]*pppdV[1][0][0]) + (ppdE[2][1]*pppdV[1][0][1]);

  pdC[0] = ( -(4.0*dE20xE21*dE20xE21 - dE2dotV1*dE2dotV1)*pppdV[1][0][1]
	     - (4.0*dE20xE21*dE2dotV1)*pppdV[1][0][0] ) / dDen;
  pdC[1] = ( (4.0*dE20xE21*dE20xE21 - dE2dotV1*dE2dotV1)*pppdV[1][0][0]
	     - (4.0*dE20xE21*dE2dotV1)*pppdV[1][0][1] ) / dDen;
  
  dMagCE30 = ( (pdC[0] - pppdV[3][0][0])*(pdC[0] - pppdV[3][0][0]) + 
	       (pdC[1] - pppdV[3][0][1])*(pdC[1] - pppdV[3][0][1]) );
  dMagCE31 = ( (pdC[0] - pppdV[3][1][0])*(pdC[0] - pppdV[3][1][0]) + 
	       (pdC[1] - pppdV[3][1][1])*(pdC[1] - pppdV[3][1][1]) );
  if ( dMagCE30 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[0] = 0.0;
    }
  else
    {
      pdEdgeZone[0] = ( ( 4.0*dE20xE21*sqrt(dV1dotV1*dE2dotE2)*( 2.0*dE3dotE30 - dV1dotE3 )
			  + ( 4.0*dE20xE21*dE20xE21 - dV1dotV1*dE2dotE2 )*dE3xV1 ) / 
			( dDen*sqrt(dE3dotE3*dMagCE30) ) );
    }
  if ( dMagCE31 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[1] = 0.0;
    }
  else
    {
      pdEdgeZone[1] = -( ( 4.0*dE20xE21*sqrt(dV1dotV1*dE2dotE2)*( 2.0*dE3dotE31 - dV1dotE3 )
			   + ( 4.0*dE20xE21*dE20xE21 - dV1dotV1*dE2dotE2 )*dE3xV1 ) / 
			 ( dDen*sqrt(dE3dotE3*dMagCE31) ) );
    }

  if ( !( pdEdgeZone[0] < -s_cdEpsilonEdgeZoneVVEAntiparallelEdgeZone ) )
    {
      eResult = APM2D::eztStart;
    }
  else if ( !( pdEdgeZone[1] < -s_cdEpsilonEdgeZoneVVEAntiparallelEdgeZone ) )
    {
      eResult = APM2D::eztEnd;
    }
  else
    {
      eResult = APM2D::eztMiddle;
    }

  return eResult;
}

static APM2D::EdgeZoneType EdgeZoneVVEAntiparallelAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
							   DOUBLE dV1_x, DOUBLE dV1_y,
							   DOUBLE dE20_x, DOUBLE dE20_y, 
							   DOUBLE dE21_x, DOUBLE dE21_y,
							   DOUBLE dE30_x, DOUBLE dE30_y,
							   DOUBLE dE31_x, DOUBLE dE31_y)
{
  return EdgeZoneVVEAntiparallelLocal(dV0_x, dV0_y, dV1_x, dV1_y,
				      dE20_x, dE20_y, dE21_x, dE21_y,
				      dE30_x, dE30_y, dE31_x, dE31_y);
}

APM2D::CircleTestType APM2D::InCircleVVE_VAntiparallel(DOUBLE dV0_x, DOUBLE dV0_y,
						       DOUBLE dV1_x, DOUBLE dV1_y,
						       DOUBLE dE20_x, DOUBLE dE20_y, 
						       DOUBLE dE21_x, DOUBLE dE21_y,
						       DOUBLE dV3_x, DOUBLE dV3_y)
{
  APM2D::CircleTestType eResult;

  eResult = APM2D::cttOnCircle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InCircleVVE_VAntiparallelLocal(dV0_x, dV0_y, dV1_x, dV1_y,
						 dE20_x, dE20_y, dE21_x, dE21_y,
						 dV3_x, dV3_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InCircleVVE_VAntiparallelAdaptive(dV0_x, dV0_y, dV1_x, dV1_y,
						    dE20_x, dE20_y, dE21_x, dE21_y,
						    dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InCircleVVE_VAntiparallelMPFUN(dV0_x, dV0_y, dV1_x, dV1_y,
						 dE20_x, dE20_y, dE21_x, dE21_y,
						 dV3_x, dV3_y);
      }
    break;

    case APM::atMPFUNFilter:
      {
	eResult = InCircleVVE_VAntiparallelMPFUNFilter(dV0_x, dV0_y, dV1_x, dV1_y,
						 dE20_x, dE20_y, dE21_x, dE21_y,
						 dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InCircleVVE_VAntiparallelMPFUNNormalized(dV0_x, dV0_y, dV1_x, dV1_y,
							   dE20_x, dE20_y, dE21_x, dE21_y,
							   dV3_x, dV3_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::CircleTestType InCircleVVE_VAntiparallelLocal(DOUBLE dV0_x, DOUBLE dV0_y,
							    DOUBLE dV1_x, DOUBLE dV1_y,
							    DOUBLE dE20_x, DOUBLE dE20_y, 
							    DOUBLE dE21_x, DOUBLE dE21_y,
							    DOUBLE dV3_x, DOUBLE dV3_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE dV1dotV1, dE2dotE2, dV3dotV3, dV1dotV3;
  DOUBLE dV1xV3, dE20xE21;
  DOUBLE dDen;
  DOUBLE dResult;
  APM2D::CircleTestType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = dV1_x - dV0_x;
  pppdV[1][0][1] = dV1_y - dV0_y;
  pppdV[2][0][0] = dE20_x - dV0_x;
  pppdV[2][0][1] = dE20_y - dV0_y;
  pppdV[2][1][0] = dE21_x - dV0_x;
  pppdV[2][1][1] = dE21_y - dV0_y;
  pppdV[3][0][0] = dV3_x - dV0_x;
  pppdV[3][0][1] = dV3_y - dV0_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;

  dV1dotV1 = (pppdV[1][0][0]*pppdV[1][0][0]) + (pppdV[1][0][1]*pppdV[1][0][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);
  dV3dotV3 = (pppdV[3][0][0]*pppdV[3][0][0]) + (pppdV[3][0][1]*pppdV[3][0][1]);
  dV1dotV3 = (pppdV[1][0][0]*pppdV[3][0][0]) + (pppdV[1][0][1]*pppdV[3][0][1]);

  dV1xV3 = (pppdV[1][0][0]*pppdV[3][0][1]) - (pppdV[1][0][1]*pppdV[3][0][0]);
  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);

  dDen = 4.0*dE20xE21*sqrt(dV1dotV1*dE2dotE2);
  ASSERT( dDen != 0.0 );

  dResult = ( 4.0*dE20xE21*sqrt(dV1dotV1*dE2dotE2)*( dV3dotV3 - dV1dotV3 )
	      - ( 4.0*dE20xE21*dE20xE21 - dV1dotV1*dE2dotE2 )*dV1xV3 ) / dDen;
  
  if ( dResult < -s_cdEpsilonInCircleVVE_VAntiparallelOnCircle )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( dResult > s_cdEpsilonInCircleVVE_VAntiparallelOnCircle )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

static APM2D::CircleTestType InCircleVVE_VAntiparallelAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
							       DOUBLE dV1_x, DOUBLE dV1_y,
							       DOUBLE dE20_x, DOUBLE dE20_y, 
							       DOUBLE dE21_x, DOUBLE dE21_y,
							       DOUBLE dV3_x, DOUBLE dV3_y)
{
  return InCircleVVE_VAntiparallelLocal(dV0_x, dV0_y, dV1_x, dV1_y,
					dE20_x, dE20_y, dE21_x, dE21_y,
					dV3_x, dV3_y);
}

APM2D::CircleTestType APM2D::InCircleVVE_EAntiparallel(DOUBLE dV0_x, DOUBLE dV0_y,
						       DOUBLE dV1_x, DOUBLE dV1_y,
						       DOUBLE dE20_x, DOUBLE dE20_y, 
						       DOUBLE dE21_x, DOUBLE dE21_y,
						       DOUBLE dE30_x, DOUBLE dE30_y,
						       DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::CircleTestType eResult;

  eResult = APM2D::cttOnCircle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InCircleVVE_EAntiparallelLocal(dV0_x, dV0_y, dV1_x, dV1_y,
						 dE20_x, dE20_y, dE21_x, dE21_y,
						 dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InCircleVVE_EAntiparallelAdaptive(dV0_x, dV0_y, dV1_x, dV1_y,
						    dE20_x, dE20_y, dE21_x, dE21_y,
						    dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InCircleVVE_EAntiparallelMPFUN(dV0_x, dV0_y, dV1_x, dV1_y,
						 dE20_x, dE20_y, dE21_x, dE21_y,
						 dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = InCircleVVE_EAntiparallelMPFUNFilter(dV0_x, dV0_y, dV1_x, dV1_y,
						       dE20_x, dE20_y, dE21_x, dE21_y,
						       dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InCircleVVE_EAntiparallelMPFUNNormalized(dV0_x, dV0_y, dV1_x, dV1_y,
							   dE20_x, dE20_y, dE21_x, dE21_y,
							   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::CircleTestType InCircleVVE_EAntiparallelLocal(DOUBLE dV0_x, DOUBLE dV0_y,
							    DOUBLE dV1_x, DOUBLE dV1_y,
							    DOUBLE dE20_x, DOUBLE dE20_y, 
							    DOUBLE dE21_x, DOUBLE dE21_y,
							    DOUBLE dE30_x, DOUBLE dE30_y,
							    DOUBLE dE31_x, DOUBLE dE31_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE dE3dotE3;
  DOUBLE dV1dotE2, dV1dotE3;
  DOUBLE dE3xV1, dE20xE21, dE30xE31;
  DOUBLE dDen;
  DOUBLE dResult;
  APM2D::CircleTestType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = dV1_x - dV0_x;
  pppdV[1][0][1] = dV1_y - dV0_y;
  pppdV[2][0][0] = dE20_x - dV0_x;
  pppdV[2][0][1] = dE20_y - dV0_y;
  pppdV[2][1][0] = dE21_x - dV0_x;
  pppdV[2][1][1] = dE21_y - dV0_y;
  pppdV[3][0][0] = dE30_x - dV0_x;
  pppdV[3][0][1] = dE30_y - dV0_y;
  pppdV[3][1][0] = dE31_x - dV0_x;
  pppdV[3][1][1] = dE31_y - dV0_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;
  ppdE[3][0] = dE31_x - dE30_x;
  ppdE[3][1] = dE31_y - dE30_y;

  dE3dotE3 = (ppdE[3][0]*ppdE[3][0]) + (ppdE[3][1]*ppdE[3][1]);

  dV1dotE2 = (pppdV[1][0][0]*ppdE[2][0]) + (pppdV[1][0][1]*ppdE[2][1]);
  dV1dotE3 = (pppdV[1][0][0]*ppdE[3][0]) + (pppdV[1][0][1]*ppdE[3][1]);

  dE3xV1 = (ppdE[3][0]*pppdV[1][0][1]) - (ppdE[3][1]*pppdV[1][0][0]);
  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);
  dE30xE31 = (pppdV[3][0][0]*pppdV[3][1][1]) - (pppdV[3][0][1]*pppdV[3][1][0]);

  dDen = 64.0*dE20xE21*dE20xE21*dV1dotE2*dV1dotE2*dE3dotE3;
  ASSERT( dDen != 0.0 );

  dResult = ( 64.0*dE20xE21*dE20xE21*dV1dotE2*dV1dotE2*dE30xE31*( dE30xE31 + dE3xV1 )
	      - dE3xV1*dE3xV1*( 4.0*dE20xE21*dE20xE21 - dV1dotE2*dV1dotE2 )*( 4.0*dE20xE21*dE20xE21 - dV1dotE2*dV1dotE2 )
	      - 16.0*dE20xE21*dE20xE21*dV1dotE2*dV1dotE2*dV1dotE3*dV1dotE3
	      - 8.0*dE20xE21*dV1dotE2*dV1dotE3*( 4.0*dE20xE21*dE20xE21 - dV1dotE2*dV1dotE2 )*( 2.0*dE30xE31 + dE3xV1 ) ) / dDen;

  if ( dResult < -s_cdEpsilonInCircleVVE_EAntiparallelOnCircle )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( dResult > s_cdEpsilonInCircleVVE_EAntiparallelOnCircle )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

static APM2D::CircleTestType InCircleVVE_EAntiparallelAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
							       DOUBLE dV1_x, DOUBLE dV1_y,
							       DOUBLE dE20_x, DOUBLE dE20_y, 
							       DOUBLE dE21_x, DOUBLE dE21_y,
							       DOUBLE dE30_x, DOUBLE dE30_y,
							       DOUBLE dE31_x, DOUBLE dE31_y)
{
  return InCircleVVE_EAntiparallelLocal(dV0_x, dV0_y, dV1_x, dV1_y,
					dE20_x, dE20_y, dE21_x, dE21_y,
					dE30_x, dE30_y, dE31_x, dE31_y);
}

APM2D::InscribedCircleCenterType 
APM2D::InscribedCircleCenterVvE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dV00_x, DOUBLE dV00_y,
				DOUBLE dV0_x, DOUBLE dV0_y,
				DOUBLE dV01_x, DOUBLE dV01_y,
				DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
				DOUBLE dE21_x, DOUBLE dE21_y)
{
  ASSERT( pdC_x != NULL && pdC_y != NULL && pdR != NULL );

  APM2D::InscribedCircleCenterType eResult;

  eResult = APM2D::icctNull;
  
  //////////////////////////////////////////////////
  // Degenerate Touching Check
  if ( dV1_E20_x  == dV0_x && dV1_E20_y == dV0_y )
    {
      *pdC_x = dV1_E20_x;
      *pdC_y = dV1_E20_y;
      *pdR = 0.0;
      return APM2D::icctVvE;
    }

  //////////////////////////////////////////////////
  // CCW Checks

  if ( !Counterclockwise(dV0_x, dV0_y, dV1_E20_x, dV1_E20_y, dE21_x, dE21_y) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InscribedCircleCenterVvELocal(pdC_x, pdC_y, pdR, 
						dV00_x, dV00_y, 
						dV0_x, dV0_y, 
						dV01_x, dV01_y, 
						dV1_E20_x, dV1_E20_y, 
						dE21_x, dE21_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InscribedCircleCenterVvEAdaptive(pdC_x, pdC_y, pdR, 
						   dV00_x, dV00_y, 
						   dV0_x, dV0_y, 
						   dV01_x, dV01_y, 
						   dV1_E20_x, dV1_E20_y, 
						   dE21_x, dE21_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InscribedCircleCenterVvEMPFUN(pdC_x, pdC_y, pdR, 
						dV00_x, dV00_y, 
						dV0_x, dV0_y, 
						dV01_x, dV01_y, 
						dV1_E20_x, dV1_E20_y, 
						dE21_x, dE21_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = InscribedCircleCenterVvEMPFUNFilter(pdC_x, pdC_y, pdR, 
						      dV00_x, dV00_y, 
						      dV0_x, dV0_y, 
						      dV01_x, dV01_y, 
						      dV1_E20_x, dV1_E20_y, 
						      dE21_x, dE21_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InscribedCircleCenterVvEMPFUNNormalized(pdC_x, pdC_y, pdR, 
							  dV00_x, dV00_y, 
							  dV0_x, dV0_y, 
							  dV01_x, dV01_y, 
							  dV1_E20_x, dV1_E20_y, 
							  dE21_x, dE21_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }

  return eResult;
}

static APM2D::InscribedCircleCenterType 
InscribedCircleCenterVvELocal(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			      DOUBLE dV00_x, DOUBLE dV00_y,
			      DOUBLE dV0_x, DOUBLE dV0_y,
			      DOUBLE dV01_x, DOUBLE dV01_y,
			      DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
			      DOUBLE dE21_x, DOUBLE dE21_y)
{
  DOUBLE pdC[3], dR;
  DOUBLE pppdV[3][2][2], pppdVN[1][2][2], ppdE[3][2];
  DOUBLE dDen;
  DOUBLE dV0dotV0, dE2dotE2;
  DOUBLE dE2xV0;
  DOUBLE dV00dotV00, dV01dotV01, dV0dotV00, dV0dotV01;
  DOUBLE dE2xV00, dE2xV01;
  DOUBLE pdVertexZone[2], dSinTheta, dSinThetaTemp;
  UINT_32 i;
  
  // Translate V1 to the origin
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = dV0_x - dV1_E20_x;
  pppdV[0][0][1] = dV0_y - dV1_E20_y;
  pppdV[2][0][0] = 0.0;
  pppdV[2][0][1] = 0.0;
  pppdV[2][1][0] = dE21_x - dV1_E20_x;
  pppdV[2][1][1] = dE21_y - dV1_E20_y;
  
  // Vertex neighbor vectors
  memset(pppdVN, 0, sizeof(pppdV));
  pppdVN[0][0][0] = dV0_x - dV00_x;
  pppdVN[0][0][1] = dV0_y - dV00_y;
  pppdVN[0][1][0] = dV0_x - dV01_x;
  pppdVN[0][1][1] = dV0_y - dV01_y;

  memset(ppdE, 0, sizeof(ppdE));
  ppdE[2][0] = dE21_x - dV1_E20_x;
  ppdE[2][1] = dE21_y - dV1_E20_y;
  
  dV0dotV0 = (pppdV[0][0][0]*pppdV[0][0][0]) + (pppdV[0][0][1]*pppdV[0][0][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0] + ppdE[2][1]*ppdE[2][1]);
  dE2xV0 = (ppdE[2][0]*pppdV[0][0][1]) - (ppdE[2][1]*pppdV[0][0][0]);
  
  dV00dotV00 = ( (pppdVN[0][0][0] * pppdVN[0][0][0]) + (pppdVN[0][0][1] * pppdVN[0][0][1]) );
  dV01dotV01 = ( (pppdVN[0][1][0] * pppdVN[0][1][0]) + (pppdVN[0][1][1] * pppdVN[0][1][1]) );

  dV0dotV00 = ( (pppdV[0][0][0] * pppdVN[0][0][0]) + (pppdV[0][0][1] * pppdVN[0][0][1]) );
  dV0dotV01 = ( (pppdV[0][0][0] * pppdVN[0][1][0]) + (pppdV[0][0][1] * pppdVN[0][1][1]) );

  dE2xV00 = ( (ppdE[2][0] * pppdVN[0][0][1]) - (ppdE[2][1] * pppdVN[0][0][0]) );
  dE2xV01 = ( (ppdE[2][0] * pppdVN[0][1][1]) - (ppdE[2][1] * pppdVN[0][1][0]) );

  //dSinTheta = dE2xV0/sqrt(dE2dotE2*dV0dotV0);
  dSinTheta = APM2D::SinTheta(dV0_x, dV0_y, dV1_E20_x, dV1_E20_y, dE21_x, dE21_y);
  dSinThetaTemp = APM2D::SinTheta(dV1_E20_x, dV1_E20_y, dE21_x, dE21_y, dV0_x, dV0_y);
  if ( dSinThetaTemp > dSinTheta )
    {
      dSinTheta = dSinThetaTemp;
    }
  if ( dSinTheta <= s_cdEpsilonInscribedCircleCenterVvEVertexOnEdge )
    {
      // V0 is on E2 and it is not one of the endpoints
      
#ifdef ENABLE_DEBUGGING_OUTPUT
      fprintf(stderr, "    VvE: V0 ( %f ) is on E2 <= %f\n", 
	      dE2xV0, s_cdEpsilonInscribedCircleCenterVvEVertexOnEdge);
      fprintf(stderr, "      V0: ( %.5f, %.5f )\n", dV0_x, dV0_y);
      fprintf(stderr, "      V1: ( %.5f, %.5f )\n", dV1_E20_x, dV1_E20_y);
      fprintf(stderr, "      E2: ( %.5f, %.5f ) => ( %.5f, %.5f )\n", 
	      dV1_E20_x, dV1_E20_y, dE21_x, dE21_y);
#endif // ENABLE_DEBUGGING_OUTPUT

      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  dDen = 2.0*dE2xV0;
  if ( dDen == 0.0 )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  //////////////////////////////////////////////////
  // Compute the center and radius

  pdC[0] = -dV0dotV0*ppdE[2][1]/dDen;
  pdC[1] = dV0dotV0*ppdE[2][0]/dDen;
  pdC[2] = sqrt((pdC[0]*pdC[0]) + (pdC[1]*pdC[1]));

  dR = (dV0dotV0*sqrt(dE2dotE2)) / dDen;
  
  //////////////////////////////////////////////////
  // Check the Vertex Zones

  pdVertexZone[0] = ( ( dV0dotV0*dE2xV00 - 2.0*dE2xV0*dV0dotV00 ) /
		      ( dDen*dR*sqrt(dV00dotV00) ) );
  pdVertexZone[1] = ( ( dV0dotV0*dE2xV01 - 2.0*dE2xV0*dV0dotV01 ) /
		      ( dDen*dR*sqrt(dV01dotV01) ) );
  
  for ( i = 0; i < 2; i++ )
    {
      if ( pdVertexZone[i] < -s_cdEpsilonInscribedCircleCenterVvEVertexZone )
	{
#ifdef DEBUG_EPSILON_VALUES
	  if ( !( pdVertexZone[i] < -(s_cdEpsilonInscribedCircleCenterVvEVertexZone*s_cdOrderOfMagnitude) ) )
		{
		  fprintf(stderr, "  VvE: Vertex Zone: %e < %e < %e\n",
			  s_cdEpsilonInscribedCircleCenterVvEVertexZone*s_cdOrderOfMagnitude,
			  pdVertexZone[i],
			  s_cdEpsilonInscribedCircleCenterVvEVertexZone);
		}
#endif // DEBUG_EPSILON_VALUES	      
	  *pdC_x = 0.0;
	  *pdC_y = 0.0;
	  *pdR = 0.0;
	  return APM2D::icctNull;
	}
    }

  if ( !( fabs(pdC[2] - dR) < s_cdEpsilonInscribedCircleCenterVvERadius ) )
    {
      fprintf(stderr, "VvE: C = ( %f, %f, %f ) != %f = R\n", pdC[0], pdC[1], pdC[2], dR);
      //ASSERT( FALSE );
    }

  // Translate the answer back out away from the origin
  *pdC_x = dV1_E20_x + pdC[0];
  *pdC_y = dV1_E20_y + pdC[1];
  *pdR = dR;
  return APM2D::icctVvE;
}

static APM2D::InscribedCircleCenterType 
InscribedCircleCenterVvEAdaptive(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				 DOUBLE dV00_x, DOUBLE dV00_y,
				 DOUBLE dV0_x, DOUBLE dV0_y,
				 DOUBLE dV01_x, DOUBLE dV01_y,
				 DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
				 DOUBLE dE21_x, DOUBLE dE21_y)
{
  return InscribedCircleCenterVvELocal(pdC_x, pdC_y, pdR, 
				       dV00_x, dV00_y, 
				       dV0_x, dV0_y, 
				       dV01_x, dV01_y, 
				       dV1_E20_x, dV1_E20_y, dE21_x, dE21_y);
}

APM2D::EdgeZoneType APM2D::EdgeZoneVvE(DOUBLE dV0_x, DOUBLE dV0_y,
				       DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
				       DOUBLE dE21_x, DOUBLE dE21_y,
				       DOUBLE dE30_x, DOUBLE dE30_y,
				       DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::EdgeZoneType eResult;

  //////////////////////////////////////////////////
  // Degenerate Touching Check
  if ( dV1_E20_x  == dV0_x && dV1_E20_y == dV0_y )
    {
      return EdgeZone(dV1_E20_x, dV1_E20_y, 
		      dE30_x, dE30_y,
		      dE31_x, dE31_y);
    }

  eResult = APM2D::eztMiddle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = EdgeZoneVvELocal(dV0_x, dV0_y,
				   dV1_E20_x, dV1_E20_y, dE21_x, dE21_y,
				   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atExact:
      {
	eResult = EdgeZoneVvEAdaptive(dV0_x, dV0_y,
				      dV1_E20_x, dV1_E20_y, dE21_x, dE21_y,
				      dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = EdgeZoneVvEMPFUN(dV0_x, dV0_y,
				   dV1_E20_x, dV1_E20_y, dE21_x, dE21_y,
				   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;

    case APM::atMPFUNFilter:
      {
	eResult = EdgeZoneVvEMPFUNFilter(dV0_x, dV0_y,
				   dV1_E20_x, dV1_E20_y, dE21_x, dE21_y,
				   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = EdgeZoneVvEMPFUNNormalized(dV0_x, dV0_y,
					     dV1_E20_x, dV1_E20_y, dE21_x, dE21_y,
					     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::EdgeZoneType EdgeZoneVvELocal(DOUBLE dV0_x, DOUBLE dV0_y,
					    DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
					    DOUBLE dE21_x, DOUBLE dE21_y,
					    DOUBLE dE30_x, DOUBLE dE30_y,
					    DOUBLE dE31_x, DOUBLE dE31_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE pdC[2];
  DOUBLE pdEdgeZone[2];
  DOUBLE dV0dotV0;
  DOUBLE dE3dotE30, dE3dotE31;
  DOUBLE dE2xV0, dE2xE3;
  DOUBLE dE3dotE3, dMagCE30, dMagCE31;
  DOUBLE dDen;
  APM2D::EdgeZoneType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = dV0_x - dV1_E20_x;
  pppdV[0][0][1] = dV0_y - dV1_E20_y;
  pppdV[1][0][0] = 0.0;
  pppdV[1][0][1] = 0.0;
  pppdV[2][0][0] = 0.0;
  pppdV[2][0][1] = 0.0;
  pppdV[2][1][0] = dE21_x - dV1_E20_x;
  pppdV[2][1][1] = dE21_y - dV1_E20_y;
  pppdV[3][0][0] = dE30_x - dV1_E20_x;
  pppdV[3][0][1] = dE30_y - dV1_E20_y;
  pppdV[3][1][0] = dE31_x - dV1_E20_x;
  pppdV[3][1][1] = dE31_y - dV1_E20_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[2][0] = dE21_x - dV1_E20_x;
  ppdE[2][1] = dE21_y - dV1_E20_y;
  ppdE[3][0] = dE31_x - dE30_x;
  ppdE[3][1] = dE31_y - dE30_y;
  
  dV0dotV0 = (pppdV[0][0][0]*pppdV[0][0][0]) + (pppdV[0][0][1]*pppdV[0][0][1]);
  dE3dotE3 = (ppdE[3][0]*ppdE[3][0]) + (ppdE[3][1]*ppdE[3][1]);

  dE3dotE30 = (ppdE[3][0]*pppdV[3][0][0]) + (ppdE[3][1]*pppdV[3][0][1]);
  dE3dotE31 = (ppdE[3][0]*pppdV[3][1][0]) + (ppdE[3][1]*pppdV[3][1][1]);
  
  dE2xV0 = (ppdE[2][0]*pppdV[0][0][1]) - (ppdE[2][1]*pppdV[0][0][0]);
  dE2xE3 = (ppdE[2][0]*ppdE[3][1]) - (ppdE[2][1]*ppdE[3][0]);

  dDen = 2.0*dE2xV0; 
  ASSERT( dDen != 0.0 );

  pdC[0] = -dV0dotV0*ppdE[2][1]/dDen;
  pdC[1] = dV0dotV0*ppdE[2][0]/dDen;

  dMagCE30 = ( (pdC[0] - pppdV[3][0][0])*(pdC[0] - pppdV[3][0][0]) + 
	       (pdC[1] - pppdV[3][0][1])*(pdC[1] - pppdV[3][0][1]) );
  dMagCE31 = ( (pdC[0] - pppdV[3][1][0])*(pdC[0] - pppdV[3][1][0]) + 
	       (pdC[1] - pppdV[3][1][1])*(pdC[1] - pppdV[3][1][1]) );

  if ( dMagCE30 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[0] = 0.0;
    }
  else
    {
      pdEdgeZone[0] = ( ( 2.0*dE2xV0*dE3dotE30 - dV0dotV0*dE2xE3 ) / 
			( dDen*sqrt(dE3dotE3*dMagCE30) ) );
    }
  if ( dMagCE31 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[1] = 0.0;
    }
  else
    {
      pdEdgeZone[1] = -( ( 2.0*dE2xV0*dE3dotE31 - dV0dotV0*dE2xE3 ) /
			 ( dDen*sqrt(dE3dotE3*dMagCE31) ) );
    }
  
  if ( !( pdEdgeZone[0] < -s_cdEpsilonEdgeZoneVvEEdgeZone ) )
    {
      eResult = APM2D::eztStart;
    }
  else if ( !( pdEdgeZone[1] < -s_cdEpsilonEdgeZoneVvEEdgeZone ) )
    {
      eResult = APM2D::eztEnd;
    }
  else
    {
      eResult = APM2D::eztMiddle;
    }

  return eResult;
}

static APM2D::EdgeZoneType EdgeZoneVvEAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
					       DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
					       DOUBLE dE21_x, DOUBLE dE21_y,
					       DOUBLE dE30_x, DOUBLE dE30_y,
					       DOUBLE dE31_x, DOUBLE dE31_y)
{
  return EdgeZoneVvELocal(dV0_x, dV0_y,
			  dV1_E20_x, dV1_E20_y, dE21_x, dE21_y,
			  dE30_x, dE30_y, dE31_x, dE31_y);
}

APM2D::CircleTestType APM2D::InCircleVvE_V(DOUBLE dV0_x, DOUBLE dV0_y,
					   DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
					   DOUBLE dE21_x, DOUBLE dE21_y,
					   DOUBLE dV3_x, DOUBLE dV3_y)
{
  APM2D::CircleTestType eResult;

  //////////////////////////////////////////////////
  // Degenerate Touching Check
  if ( dV1_E20_x == dV0_x && dV1_E20_y == dV0_y )
    {
      if ( dV1_E20_x == dV3_x && dV1_E20_y == dV3_y )
	{
	  return APM2D::cttOnCircle;
	}
      else
	{
	  return APM2D::cttOutCircle;
	}
    }

  eResult = APM2D::cttOnCircle;

  // Check to see if the vertex is in the edge envelope
  if ( !Counterclockwise(dV1_E20_x, dV1_E20_y, 
			 dE21_x, dE21_y,
			 dV3_x, dV3_y) )
    {
      return APM2D::cttOutCircle;
    }

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InCircleVvE_VLocal(dV0_x, dV0_y,
				     dV1_E20_x, dV1_E20_y, dE21_x, dE21_y,
				     dV3_x, dV3_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InCircleVvE_VAdaptive(dV0_x, dV0_y,
					dV1_E20_x, dV1_E20_y, dE21_x, dE21_y,
					dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InCircleVvE_VMPFUN(dV0_x, dV0_y,
				     dV1_E20_x, dV1_E20_y, dE21_x, dE21_y,
				     dV3_x, dV3_y);
      }
    break;

    case APM::atMPFUNFilter:
      {
	eResult = InCircleVvE_VMPFUNFilter(dV0_x, dV0_y,
				     dV1_E20_x, dV1_E20_y, dE21_x, dE21_y,
				     dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InCircleVvE_VMPFUNNormalized(dV0_x, dV0_y,
					       dV1_E20_x, dV1_E20_y, dE21_x, dE21_y,
					       dV3_x, dV3_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::CircleTestType InCircleVvE_VLocal(DOUBLE dV0_x, DOUBLE dV0_y,
						DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
						DOUBLE dE21_x, DOUBLE dE21_y,
						DOUBLE dV3_x, DOUBLE dV3_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE dV0dotV0, dV3dotV3;
  DOUBLE dE2xV0, dE2xV3;
  DOUBLE dDen;
  DOUBLE dResult;
  APM2D::CircleTestType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = dV0_x - dV1_E20_x;
  pppdV[0][0][1] = dV0_y - dV1_E20_y;
  pppdV[1][0][0] = 0.0;
  pppdV[1][0][1] = 0.0;
  pppdV[2][0][0] = 0.0;
  pppdV[2][0][1] = 0.0;
  pppdV[2][1][0] = dE21_x - dV1_E20_x;
  pppdV[2][1][1] = dE21_y - dV1_E20_y;
  pppdV[3][0][0] = dV3_x - dV1_E20_x;
  pppdV[3][0][1] = dV3_y - dV1_E20_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[2][0] = dE21_x - dV1_E20_x;
  ppdE[2][1] = dE21_y - dV1_E20_y;
  
  dV0dotV0 = (pppdV[0][0][0]*pppdV[0][0][0]) + (pppdV[0][0][1]*pppdV[0][0][1]);
  dV3dotV3 = (pppdV[3][0][0]*pppdV[3][0][0]) + (pppdV[3][0][1]*pppdV[3][0][1]);

  dE2xV0 = (ppdE[2][0]*pppdV[0][0][1]) - (ppdE[2][1]*pppdV[0][0][0]);
  dE2xV3 = (ppdE[2][0]*pppdV[3][0][1]) - (ppdE[2][1]*pppdV[3][0][0]);

  dDen = dE2xV0; 
  ASSERT( dDen != 0.0 );

  dResult = ( dV3dotV3*dE2xV0 - dV0dotV0*dE2xV3 ) / dDen;
  
  if ( dResult < -s_cdEpsilonInCircleVvE_VOnCircle )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( dResult > s_cdEpsilonInCircleVvE_VOnCircle )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

static APM2D::CircleTestType InCircleVvE_VAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   DOUBLE dV3_x, DOUBLE dV3_y)
{
  return InCircleVvE_VLocal(dV0_x, dV0_y,
			    dV1_E20_x, dV1_E20_y, dE21_x, dE21_y,
			    dV3_x, dV3_y);
}

APM2D::CircleTestType APM2D::InCircleVvE_E(DOUBLE dV0_x, DOUBLE dV0_y,
					   DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
					   DOUBLE dE21_x, DOUBLE dE21_y,
					   DOUBLE dE30_x, DOUBLE dE30_y,
					   DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::CircleTestType eResult;

  //////////////////////////////////////////////////
  // Degenerate Touching Check
  if ( dV1_E20_x == dV0_x && dV1_E20_y == dV0_y )
    {
      if ( ( dV1_E20_x == dE30_x && dV1_E20_y == dE30_y ) ||
	   ( dV1_E20_x == dE31_x && dV1_E20_y == dE31_y ) )
	{
	  return APM2D::cttOnCircle;
	}
      else
	{
	  return APM2D::cttOutCircle;
	}
    }

  eResult = APM2D::cttOnCircle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InCircleVvE_ELocal(dV0_x, dV0_y,
				     dV1_E20_x, dV1_E20_y, dE21_x, dE21_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InCircleVvE_EAdaptive(dV0_x, dV0_y,
					dV1_E20_x, dV1_E20_y, dE21_x, dE21_y,
					dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InCircleVvE_EMPFUN(dV0_x, dV0_y,
				     dV1_E20_x, dV1_E20_y, dE21_x, dE21_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;

    case APM::atMPFUNFilter:
      {
	eResult = InCircleVvE_EMPFUNFilter(dV0_x, dV0_y,
				     dV1_E20_x, dV1_E20_y, dE21_x, dE21_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InCircleVvE_EMPFUNNormalized(dV0_x, dV0_y,
					       dV1_E20_x, dV1_E20_y, dE21_x, dE21_y,
					       dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::CircleTestType InCircleVvE_ELocal(DOUBLE dV0_x, DOUBLE dV0_y,
						DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
						DOUBLE dE21_x, DOUBLE dE21_y,
						DOUBLE dE30_x, DOUBLE dE30_y,
						DOUBLE dE31_x, DOUBLE dE31_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE dV0dotV0, dE2dotE2, dE3dotE3;
  DOUBLE dE2dotE3;
  DOUBLE dE2xV0, dE30xE31;
  DOUBLE dDen;
  DOUBLE dResult;
  APM2D::CircleTestType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = dV0_x - dV1_E20_x;
  pppdV[0][0][1] = dV0_y - dV1_E20_y;
  pppdV[1][0][0] = 0.0;
  pppdV[1][0][1] = 0.0;
  pppdV[2][0][0] = 0.0;
  pppdV[2][0][1] = 0.0;
  pppdV[2][1][0] = dE21_x - dV1_E20_x;
  pppdV[2][1][1] = dE21_y - dV1_E20_y;
  pppdV[3][0][0] = dE30_x - dV1_E20_x;
  pppdV[3][0][1] = dE30_y - dV1_E20_y;
  pppdV[3][1][0] = dE31_x - dV1_E20_x;
  pppdV[3][1][1] = dE31_y - dV1_E20_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[2][0] = dE21_x - dV1_E20_x;
  ppdE[2][1] = dE21_y - dV1_E20_y;
  ppdE[3][0] = dE31_x - dE30_x;
  ppdE[3][1] = dE31_y - dE30_y;
  
  dV0dotV0 = (pppdV[0][0][0]*pppdV[0][0][0]) + (pppdV[0][0][1]*pppdV[0][0][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);
  dE3dotE3 = (ppdE[3][0]*ppdE[3][0]) + (ppdE[3][1]*ppdE[3][1]);

  dE2dotE3 = (ppdE[2][0]*ppdE[3][0]) + (ppdE[2][1]*ppdE[3][1]);

  dE2xV0 = (ppdE[2][0]*pppdV[0][0][1]) - (ppdE[2][1]*pppdV[0][0][0]);
  dE30xE31 = (pppdV[3][0][0]*pppdV[3][1][1]) - (pppdV[3][0][1]*pppdV[3][1][0]);

  dDen = 4.0*dE3dotE3*dE2xV0*dE2xV0; 
  ASSERT( dDen != 0.0 );

  dResult = ( 4.0*dE2xV0*dE2xV0*dE30xE31*dE30xE31 
	      + 4.0*dE2xV0*dE30xE31*dV0dotV0*dE2dotE3
	      - dV0dotV0*dV0dotV0*( dE2dotE2*dE3dotE3 - dE2dotE3*dE2dotE3 ) ) / dDen;

  if ( dResult < -s_cdEpsilonInCircleVvE_EOnCircle )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( dResult > s_cdEpsilonInCircleVvE_EOnCircle )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

static APM2D::CircleTestType InCircleVvE_EAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   DOUBLE dE30_x, DOUBLE dE30_y,
						   DOUBLE dE31_x, DOUBLE dE31_y)
{
  return InCircleVvE_ELocal(dV0_x, dV0_y,
			    dV1_E20_x, dV1_E20_y, dE21_x, dE21_y,
			    dE30_x, dE30_y, dE31_x, dE31_y);
}

APM2D::InscribedCircleCenterType 
APM2D::InscribedCircleCenterEvV(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dE00_x, DOUBLE dE00_y,
				DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
				DOUBLE dV20_x, DOUBLE dV20_y,
				DOUBLE dV2_x, DOUBLE dV2_y,
				DOUBLE dV21_x, DOUBLE dV21_y)
{
  ASSERT( pdC_x != NULL && pdC_y != NULL && pdR != NULL );

  APM2D::InscribedCircleCenterType eResult;

  eResult = APM2D::icctNull;
  
  //////////////////////////////////////////////////
  // Degenerate Touching Check
  if ( dE01_V1_x  == dV2_x && dE01_V1_y == dV2_y )
    {
      *pdC_x = dE01_V1_x;
      *pdC_y = dE01_V1_y;
      *pdR = 0.0;
      return APM2D::icctEvV;
    }

  //////////////////////////////////////////////////
  // CCW Checks

  if ( !Counterclockwise(dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, dV2_x, dV2_y) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InscribedCircleCenterEvVLocal(pdC_x, pdC_y, pdR, 
						dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
						dV20_x, dV20_y,
						dV2_x, dV2_y,
						dV21_x, dV21_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InscribedCircleCenterEvVAdaptive(pdC_x, pdC_y, pdR, 
						   dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
						   dV20_x, dV20_y,
						   dV2_x, dV2_y,
						   dV21_x, dV21_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InscribedCircleCenterEvVMPFUN(pdC_x, pdC_y, pdR, 
						dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
						dV20_x, dV20_y,
						dV2_x, dV2_y,
						dV21_x, dV21_y);
      }
    break;

    case APM::atMPFUNFilter:
      {
	eResult = InscribedCircleCenterEvVMPFUNFilter(pdC_x, pdC_y, pdR, 
						      dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
						      dV20_x, dV20_y,
						      dV2_x, dV2_y,
						      dV21_x, dV21_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InscribedCircleCenterEvVMPFUNNormalized(pdC_x, pdC_y, pdR, 
							  dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
							  dV20_x, dV20_y,
							  dV2_x, dV2_y,
							  dV21_x, dV21_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }

  return eResult;
}

static APM2D::InscribedCircleCenterType 
InscribedCircleCenterEvVLocal(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			      DOUBLE dE00_x, DOUBLE dE00_y,
			      DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
			      DOUBLE dV20_x, DOUBLE dV20_y,
			      DOUBLE dV2_x, DOUBLE dV2_y,
			      DOUBLE dV21_x, DOUBLE dV21_y)
{
  DOUBLE pdC[3], dR;
  DOUBLE pppdV[3][2][2], pppdVN[3][2][2]; // Endpoints
  DOUBLE ppdE[3][2];     // Edge vectors
  DOUBLE dDen;
  DOUBLE dV2dotV2, dE0dotE0;
  DOUBLE dE0xV2;
  DOUBLE dV20dotV20, dV21dotV21, dV2dotV20, dV2dotV21;
  DOUBLE dE0xV20, dE0xV21;
  DOUBLE pdVertexZone[2], dSinTheta, dSinThetaTemp;
  UINT_32 i;
  
  // Translate V1 to the origin
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = dE00_x - dE01_V1_x;
  pppdV[0][0][1] = dE00_y - dE01_V1_y;
  pppdV[0][1][0] = 0.0;
  pppdV[0][1][1] = 0.0;
  pppdV[1][0][0] = 0.0;
  pppdV[1][0][1] = 0.0;
  pppdV[2][0][0] = dV2_x - dE01_V1_x;
  pppdV[2][0][1] = dV2_y - dE01_V1_y;
  
  // Vertex neighbor vectors
  memset(pppdVN, 0, sizeof(pppdV));
  pppdVN[2][0][0] = dV2_x - dV20_x;
  pppdVN[2][0][1] = dV2_y - dV20_y;
  pppdVN[2][1][0] = dV2_x - dV21_x;
  pppdVN[2][1][1] = dV2_y - dV21_y;

  memset(ppdE, 0, sizeof(ppdE));
  ppdE[0][0] = dE01_V1_x - dE00_x;
  ppdE[0][1] = dE01_V1_y - dE00_y;
  
  dV2dotV2 = (pppdV[2][0][0]*pppdV[2][0][0]) + (pppdV[2][0][1]*pppdV[2][0][1]);
  dE0dotE0 = (ppdE[0][0]*ppdE[0][0]) + (ppdE[0][1]*ppdE[0][1]);
  dE0xV2 = (ppdE[0][0]*pppdV[2][0][1]) - (ppdE[0][1]*pppdV[2][0][0]);

  dV20dotV20 = ( (pppdVN[2][0][0] * pppdVN[2][0][0]) + (pppdVN[2][0][1] * pppdVN[2][0][1]) );
  dV21dotV21 = ( (pppdVN[2][1][0] * pppdVN[2][1][0]) + (pppdVN[2][1][1] * pppdVN[2][1][1]) );

  dV2dotV20 = ( (pppdV[2][0][0] * pppdVN[2][0][0]) + (pppdV[2][0][1] * pppdVN[2][0][1]) );
  dV2dotV21 = ( (pppdV[2][0][0] * pppdVN[2][1][0]) + (pppdV[2][0][1] * pppdVN[2][1][1]) );

  dE0xV20 = ( (ppdE[0][0] * pppdVN[2][0][1]) - (ppdE[0][1] * pppdVN[2][0][0]) );
  dE0xV21 = ( (ppdE[0][0] * pppdVN[2][1][1]) - (ppdE[0][1] * pppdVN[2][1][0]) );

  //dSinTheta = dE0xV2/sqrt(dE0dotE0*dV2dotV2); 
  dSinTheta = APM2D::SinTheta(dV2_x, dV2_y, dE00_x, dE00_y, dE01_V1_x, dE01_V1_y);
  dSinThetaTemp = APM2D::SinTheta(dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, dV2_x, dV2_y);
  if ( dSinThetaTemp > dSinTheta )
    {
      dSinTheta = dSinThetaTemp;
    }
  if ( dSinTheta <= s_cdEpsilonInscribedCircleCenterEvVVertexOnEdge )
    {
      // V0 is on E2 and it is not one of the endpoints
      
#ifdef ENABLE_DEBUGGING_OUTPUT
      fprintf(stderr, "    EvV: V2 ( %f ) is on E0 <= %f\n", 
	      dE0xV2, s_cdEpsilonInscribedCircleCenterEvVVertexOnEdge);
      fprintf(stderr, "      E0: ( %.5f, %.5f ) => ( %.5f, %.5f )\n", 
	      dE00_x, dE00_y, dE01_V1_x, dE01_V1_y);
      fprintf(stderr, "      V1: ( %.5f, %.5f )\n", dE01_V1_x, dE01_V1_y);
      fprintf(stderr, "      V2: ( %.5f, %.5f )\n", dV2_x, dV2_y);
#endif // ENABLE_DEBUGGING_OUTPUT

      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  dDen = 2.0*dE0xV2;
  if ( dDen == 0.0 )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;      
    }

  //////////////////////////////////////////////////
  // Compute the center and radius

  pdC[0] = -dV2dotV2*ppdE[0][1]/dDen;
  pdC[1] = dV2dotV2*ppdE[0][0]/dDen;
  pdC[2] = sqrt((pdC[0]*pdC[0]) + (pdC[1]*pdC[1]));

  dR = (dV2dotV2*sqrt(dE0dotE0)) / dDen;
  
  //////////////////////////////////////////////////
  // Check the Vertex Zones

  pdVertexZone[0] = ( ( dV2dotV2*dE0xV20 - 2.0*dE0xV2*dV2dotV20 ) /
		      ( dDen*dR*sqrt(dV20dotV20) ) );
  pdVertexZone[1] = ( ( dV2dotV2*dE0xV21 - 2.0*dE0xV2*dV2dotV21 ) /
		      ( dDen*dR*sqrt(dV21dotV21) ) );
  
  for ( i = 0; i < 2; i++ )
    {
      if ( pdVertexZone[i] < -s_cdEpsilonInscribedCircleCenterEvVVertexZone )
	{
#ifdef DEBUG_EPSILON_VALUES
	  if ( !( pdVertexZone[i] < -(s_cdEpsilonInscribedCircleCenterEvVVertexZone*s_cdOrderOfMagnitude) ) )
	    {
	      fprintf(stderr, "  EvV: Vertex Zone: %e < %e < %e\n",
		      s_cdEpsilonInscribedCircleCenterEvVVertexZone*s_cdOrderOfMagnitude,
		      pdVertexZone[i],
		      s_cdEpsilonInscribedCircleCenterEvVVertexZone);
	    }
#endif // DEBUG_EPSILON_VALUES	      
	  *pdC_x = 0.0;
	  *pdC_y = 0.0;
	  *pdR = 0.0;
	  return APM2D::icctNull;
	}
    }

  if ( !( fabs(pdC[2] - dR) < s_cdEpsilonInscribedCircleCenterEvVRadius ) )
    {
      fprintf(stderr, "EvV: C = ( %f, %f, %f ) != %f = R\n", pdC[0], pdC[1], pdC[2], dR);
      //ASSERT( FALSE );
    }

  // Translate the answer back out away from the origin
  *pdC_x = dE01_V1_x + pdC[0];
  *pdC_y = dE01_V1_y + pdC[1];
  *pdR = dR;
  return APM2D::icctEvV;
}

static APM2D::InscribedCircleCenterType 
InscribedCircleCenterEvVAdaptive(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				 DOUBLE dE00_x, DOUBLE dE00_y,
				 DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
				 DOUBLE dV20_x, DOUBLE dV20_y,
				 DOUBLE dV2_x, DOUBLE dV2_y,
				 DOUBLE dV21_x, DOUBLE dV21_y)
{
  return InscribedCircleCenterEvVLocal(pdC_x, pdC_y, pdR, 
				       dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
				       dV20_x, dV20_y,
				       dV2_x, dV2_y,
				       dV21_x, dV21_y);
}

APM2D::EdgeZoneType APM2D::EdgeZoneEvV(DOUBLE dE00_x, DOUBLE dE00_y,
				       DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
				       DOUBLE dV2_x, DOUBLE dV2_y,
				       DOUBLE dE30_x, DOUBLE dE30_y,
				       DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::EdgeZoneType eResult;

  //////////////////////////////////////////////////
  // Degenerate Touching Check
  if ( dE01_V1_x  == dV2_x && dE01_V1_y == dV2_y )
    {
      return EdgeZone(dE01_V1_x, dE01_V1_y, 
		      dE30_x, dE30_y,
		      dE31_x, dE31_y);
    }

  eResult = APM2D::eztMiddle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = EdgeZoneEvVLocal(dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
				   dV2_x, dV2_y,
				   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atExact:
      {
	eResult = EdgeZoneEvVAdaptive(dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
				      dV2_x, dV2_y,
				      dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = EdgeZoneEvVMPFUN(dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
				   dV2_x, dV2_y,
				   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;

    case APM::atMPFUNFilter:
      {
	eResult = EdgeZoneEvVMPFUNFilter(dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
					 dV2_x, dV2_y,
					 dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = EdgeZoneEvVMPFUNNormalized(dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
					     dV2_x, dV2_y,
					     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::EdgeZoneType EdgeZoneEvVLocal(DOUBLE dE00_x, DOUBLE dE00_y,
					    DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
					    DOUBLE dV2_x, DOUBLE dV2_y,
					    DOUBLE dE30_x, DOUBLE dE30_y,
					    DOUBLE dE31_x, DOUBLE dE31_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE pdC[2];
  DOUBLE pdEdgeZone[2];
  DOUBLE dV2dotV2;
  DOUBLE dE3dotE30, dE3dotE31;
  DOUBLE dE0xV2, dE0xE3;
  DOUBLE dE3dotE3, dMagCE30, dMagCE31;
  DOUBLE dDen;
  APM2D::EdgeZoneType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = dE00_x - dE01_V1_x;
  pppdV[0][0][1] = dE00_y - dE01_V1_y;
  pppdV[0][1][0] = 0.0;
  pppdV[0][1][1] = 0.0;
  pppdV[1][0][0] = 0.0;
  pppdV[1][0][1] = 0.0;
  pppdV[2][0][0] = dV2_x - dE01_V1_x;
  pppdV[2][0][1] = dV2_y - dE01_V1_y;
  pppdV[3][0][0] = dE30_x - dE01_V1_x;
  pppdV[3][0][1] = dE30_y - dE01_V1_y;
  pppdV[3][1][0] = dE31_x - dE01_V1_x;
  pppdV[3][1][1] = dE31_y - dE01_V1_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[0][0] = dE01_V1_x - dE00_x;
  ppdE[0][1] = dE01_V1_y - dE00_y;
  ppdE[3][0] = dE31_x - dE30_x;
  ppdE[3][1] = dE31_y - dE30_y;
  
  dV2dotV2 = (pppdV[2][0][0]*pppdV[2][0][0]) + (pppdV[2][0][1]*pppdV[2][0][1]);
  dE3dotE3 = (ppdE[3][0]*ppdE[3][0]) + (ppdE[3][1]*ppdE[3][1]);

  dE3dotE30 = (ppdE[3][0]*pppdV[3][0][0]) + (ppdE[3][1]*pppdV[3][0][1]);
  dE3dotE31 = (ppdE[3][0]*pppdV[3][1][0]) + (ppdE[3][1]*pppdV[3][1][1]);
  
  dE0xV2 = (ppdE[0][0]*pppdV[2][0][1]) - (ppdE[0][1]*pppdV[2][0][0]);
  dE0xE3 = (ppdE[0][0]*ppdE[3][1]) - (ppdE[0][1]*ppdE[3][0]);

  dDen = 2.0*dE0xV2; 
  ASSERT( dDen != 0.0 );

  pdC[0] = -dV2dotV2*ppdE[0][1]/dDen;
  pdC[1] = dV2dotV2*ppdE[0][0]/dDen;

  dMagCE30 = ( (pdC[0] - pppdV[3][0][0])*(pdC[0] - pppdV[3][0][0]) + 
	       (pdC[1] - pppdV[3][0][1])*(pdC[1] - pppdV[3][0][1]) );
  dMagCE31 = ( (pdC[0] - pppdV[3][1][0])*(pdC[0] - pppdV[3][1][0]) + 
	       (pdC[1] - pppdV[3][1][1])*(pdC[1] - pppdV[3][1][1]) );

  if ( dMagCE30 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[0] = 0.0;
    }
  else
    {
      pdEdgeZone[0] = ( ( 2.0*dE0xV2*dE3dotE30 - dV2dotV2*dE0xE3 ) / 
			( dDen*sqrt(dE3dotE3*dMagCE30) ) );
    }
  if ( dMagCE31 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[1] = 0.0;
    }
  else
    {
      pdEdgeZone[1] = -( ( 2.0*dE0xV2*dE3dotE31 - dV2dotV2*dE0xE3 ) / 
			 ( dDen*sqrt(dE3dotE3*dMagCE31) ) );
    }
  
#ifdef ENABLE_DEBUGGING_OUTPUT
  //fprintf(stderr, "START: EdgeZoneEvV\n");
  //fprintf(stderr, "  dV2dotV2 = %e\n", dV2dotV2);
  //fprintf(stderr, "  dE3dotE30 = %e\n", dE3dotE30);
  //fprintf(stderr, "  dE3dotE31 = %e\n", dE3dotE31);
  //fprintf(stderr, "  dE0xV2 = %e\n", dE0xV2);
  //fprintf(stderr, "  dE0xE3 = %e\n", dE0xE3);
  //fprintf(stderr, "  dDen = %e\n", dDen);
  //fprintf(stderr, "  pdEdgeZone[0] = %e\n", pdEdgeZone[0]);
  //fprintf(stderr, "  pdEdgeZone[1] = %e\n", pdEdgeZone[1]);
  //fprintf(stderr, "END  : EdgeZoneEvV\n");
#endif // ENABLE_DEBUGGING_OUTPUT

  if ( !( pdEdgeZone[0] < -s_cdEpsilonEdgeZoneEvVEdgeZone ) )
    {
      eResult = APM2D::eztStart;
    }
  else if ( !( pdEdgeZone[1] < -s_cdEpsilonEdgeZoneEvVEdgeZone ) )
    {
      eResult = APM2D::eztEnd;
    }
  else
    {
      eResult = APM2D::eztMiddle;
    }

  return eResult;
}

static APM2D::EdgeZoneType EdgeZoneEvVAdaptive(DOUBLE dE00_x, DOUBLE dE00_y,
					       DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
					       DOUBLE dV2_x, DOUBLE dV2_y,
					       DOUBLE dE30_x, DOUBLE dE30_y,
					       DOUBLE dE31_x, DOUBLE dE31_y)
{
  return EdgeZoneEvVLocal(dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
			  dV2_x, dV2_y,
			  dE30_x, dE30_y, dE31_x, dE31_y);
}

APM2D::CircleTestType APM2D::InCircleEvV_V(DOUBLE dE00_x, DOUBLE dE00_y,
					   DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
					   DOUBLE dV2_x, DOUBLE dV2_y,
					   DOUBLE dV3_x, DOUBLE dV3_y)
{
  APM2D::CircleTestType eResult;

  //////////////////////////////////////////////////
  // Degenerate Touching Check
  if ( dE01_V1_x  == dV2_x && dE01_V1_y == dV2_y )
    {
      if ( dE01_V1_x == dV3_x && dE01_V1_y == dV3_y )
	{
	  return APM2D::cttOnCircle;
	}
      else
	{
	  return APM2D::cttOutCircle;
	}
    }

  eResult = APM2D::cttOnCircle;

  // Check to see if the vertex is in the edge envelope
  if ( !Counterclockwise(dE00_x, dE00_y, 
			 dE01_V1_x, dE01_V1_y,
			 dV3_x, dV3_y) )
    {
      return APM2D::cttOutCircle;
    }

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InCircleEvV_VLocal(dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
				     dV2_x, dV2_y,
				     dV3_x, dV3_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InCircleEvV_VAdaptive(dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
					dV2_x, dV2_y,
					dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InCircleEvV_VMPFUN(dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
				     dV2_x, dV2_y,
				     dV3_x, dV3_y);
      }
    break;

    case APM::atMPFUNFilter:
      {
	eResult = InCircleEvV_VMPFUNFilter(dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
				     dV2_x, dV2_y,
				     dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InCircleEvV_VMPFUNNormalized(dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
					       dV2_x, dV2_y,
					       dV3_x, dV3_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::CircleTestType InCircleEvV_VLocal(DOUBLE dE00_x, DOUBLE dE00_y,
						DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
						DOUBLE dV2_x, DOUBLE dV2_y,
						DOUBLE dV3_x, DOUBLE dV3_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE dV2dotV2, dV3dotV3;
  DOUBLE dE0xV2, dE0xV3;
  DOUBLE dDen;
  DOUBLE dResult;
  APM2D::CircleTestType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = dE00_x - dE01_V1_x;
  pppdV[0][0][1] = dE00_y - dE01_V1_y;
  pppdV[0][1][0] = 0.0;
  pppdV[0][1][1] = 0.0;
  pppdV[1][0][0] = 0.0;
  pppdV[1][0][1] = 0.0;
  pppdV[2][0][0] = dV2_x - dE01_V1_x;
  pppdV[2][0][1] = dV2_y - dE01_V1_y;
  pppdV[3][0][0] = dV3_x - dE01_V1_x;
  pppdV[3][0][1] = dV3_y - dE01_V1_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[0][0] = dE01_V1_x - dE00_x;
  ppdE[0][1] = dE01_V1_y - dE00_y;
  
  dV2dotV2 = (pppdV[2][0][0]*pppdV[2][0][0]) + (pppdV[2][0][1]*pppdV[2][0][1]);
  dV3dotV3 = (pppdV[3][0][0]*pppdV[3][0][0]) + (pppdV[3][0][1]*pppdV[3][0][1]);

  dE0xV2 = (ppdE[0][0]*pppdV[2][0][1]) - (ppdE[0][1]*pppdV[2][0][0]);
  dE0xV3 = (ppdE[0][0]*pppdV[3][0][1]) - (ppdE[0][1]*pppdV[3][0][0]);

  dDen = dE0xV2;
  ASSERT( dDen != 0.0 );

  dResult = ( dV3dotV3*dE0xV2 - dV2dotV2*dE0xV3 ) / dDen;
  
  if ( dResult < -s_cdEpsilonInCircleEvV_VOnCircle )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( dResult > s_cdEpsilonInCircleEvV_VOnCircle )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

static APM2D::CircleTestType InCircleEvV_VAdaptive(DOUBLE dE00_x, DOUBLE dE00_y,
						   DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
						   DOUBLE dV2_x, DOUBLE dV2_y,
						   DOUBLE dV3_x, DOUBLE dV3_y)
{
  return InCircleEvV_VLocal(dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
			    dV2_x, dV2_y,
			    dV3_x, dV3_y);
}

APM2D::CircleTestType APM2D::InCircleEvV_E(DOUBLE dE00_x, DOUBLE dE00_y,
					   DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
					   DOUBLE dV2_x, DOUBLE dV2_y,
					   DOUBLE dE30_x, DOUBLE dE30_y,
					   DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::CircleTestType eResult;

  //////////////////////////////////////////////////
  // Degenerate Touching Check
  if ( dE01_V1_x  == dV2_x && dE01_V1_y == dV2_y )
    {
      if ( ( dE01_V1_x == dE30_x && dE01_V1_y == dE30_y ) ||
	   ( dE01_V1_x == dE31_x && dE01_V1_y == dE31_y ) )
	{
	  return APM2D::cttOnCircle;
	}
      else
	{
	  return APM2D::cttOutCircle;
	}
    }

  eResult = APM2D::cttOnCircle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InCircleEvV_ELocal(dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
				     dV2_x, dV2_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InCircleEvV_EAdaptive(dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
					dV2_x, dV2_y,
					dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InCircleEvV_EMPFUN(dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
				     dV2_x, dV2_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;

    case APM::atMPFUNFilter:
      {
	eResult = InCircleEvV_EMPFUNFilter(dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
				     dV2_x, dV2_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InCircleEvV_EMPFUNNormalized(dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
					       dV2_x, dV2_y,
					       dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::CircleTestType InCircleEvV_ELocal(DOUBLE dE00_x, DOUBLE dE00_y,
						DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
						DOUBLE dV2_x, DOUBLE dV2_y,
						DOUBLE dE30_x, DOUBLE dE30_y,
						DOUBLE dE31_x, DOUBLE dE31_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE dV2dotV2, dE0dotE0, dE3dotE3;
  DOUBLE dE0dotE3;
  DOUBLE dE0xV2, dE30xE31;
  DOUBLE dDen;
  DOUBLE dResult;
  APM2D::CircleTestType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = dE00_x - dE01_V1_x;
  pppdV[0][0][1] = dE00_y - dE01_V1_y;
  pppdV[0][1][0] = 0.0;
  pppdV[0][1][1] = 0.0;
  pppdV[1][0][0] = 0.0;
  pppdV[1][0][1] = 0.0;
  pppdV[2][0][0] = dV2_x - dE01_V1_x;
  pppdV[2][0][1] = dV2_y - dE01_V1_y;
  pppdV[3][0][0] = dE30_x - dE01_V1_x;
  pppdV[3][0][1] = dE30_y - dE01_V1_y;
  pppdV[3][1][0] = dE31_x - dE01_V1_x;
  pppdV[3][1][1] = dE31_y - dE01_V1_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[0][0] = dE01_V1_x - dE00_x;
  ppdE[0][1] = dE01_V1_y - dE00_y;
  ppdE[3][0] = dE31_x - dE30_x;
  ppdE[3][1] = dE31_y - dE30_y;
  
  dV2dotV2 = (pppdV[2][0][0]*pppdV[2][0][0]) + (pppdV[2][0][1]*pppdV[2][0][1]);
  dE0dotE0 = (ppdE[0][0]*ppdE[0][0]) + (ppdE[0][1]*ppdE[0][1]);
  dE3dotE3 = (ppdE[3][0]*ppdE[3][0]) + (ppdE[3][1]*ppdE[3][1]);

  dE0dotE3 = (ppdE[0][0]*ppdE[3][0]) + (ppdE[0][1]*ppdE[3][1]);

  dE0xV2 = (ppdE[0][0]*pppdV[2][0][1]) - (ppdE[0][1]*pppdV[2][0][0]);
  dE30xE31 = (pppdV[3][0][0]*pppdV[3][1][1]) - (pppdV[3][0][1]*pppdV[3][1][0]);

  dDen = 4.0*dE3dotE3*dE0xV2*dE0xV2;
  ASSERT( dDen != 0.0 );

  dResult = ( 4.0*dE0xV2*dE0xV2*dE30xE31*dE30xE31 
	      + 4.0*dE0xV2*dE30xE31*dV2dotV2*dE0dotE3
	      - dV2dotV2*dV2dotV2*( dE0dotE0*dE3dotE3 - dE0dotE3*dE0dotE3 ) ) / dDen;

  if ( dResult < -s_cdEpsilonInCircleEvV_EOnCircle )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( dResult > s_cdEpsilonInCircleEvV_EOnCircle )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

static APM2D::CircleTestType InCircleEvV_EAdaptive(DOUBLE dE00_x, DOUBLE dE00_y,
						   DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
						   DOUBLE dV2_x, DOUBLE dV2_y,
						   DOUBLE dE30_x, DOUBLE dE30_y,
						   DOUBLE dE31_x, DOUBLE dE31_y)
{
  return InCircleEvV_ELocal(dE00_x, dE00_y, dE01_V1_x, dE01_V1_y, 
			    dV2_x, dV2_y,
			    dE30_x, dE30_y, dE31_x, dE31_y);
}

APM2D::InscribedCircleCenterType 
APM2D::InscribedCircleCenterVEE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dV00_x, DOUBLE dV00_y,
				DOUBLE dV0_x, DOUBLE dV0_y,
				DOUBLE dV01_x, DOUBLE dV01_y,
				DOUBLE dE10_x, DOUBLE dE10_y, 
				DOUBLE dE11_x, DOUBLE dE11_y,
				DOUBLE dE20_x, DOUBLE dE20_y, 
				DOUBLE dE21_x, DOUBLE dE21_y)
{
  ASSERT( pdC_x != NULL && pdC_y != NULL && pdR != NULL );

  APM2D::InscribedCircleCenterType eResult;

  eResult = APM2D::icctNull;
  
  //////////////////////////////////////////////////
  // CCW Checks

  if ( !Counterclockwise(dE10_x, dE10_y, dE11_x, dE11_y, dV0_x, dV0_y) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  if ( !Counterclockwise(dE20_x, dE20_y, dE21_x, dE21_y, dV0_x, dV0_y) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  if ( !( Counterclockwise(dE10_x, dE10_y, dE11_x, dE11_y, dE20_x, dE20_y) ||
	  Counterclockwise(dE10_x, dE10_y, dE11_x, dE11_y, dE21_x, dE21_y) ) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  if ( !( Counterclockwise(dE20_x, dE20_y, dE21_x, dE21_y, dE10_x, dE10_y) ||
	  Counterclockwise(dE20_x, dE20_y, dE21_x, dE21_y, dE11_x, dE11_y) ) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  //////////////////////////////////////////////////
  // Anitparallel Check

  if ( Parallel(dE10_x, dE10_y, dE11_x, dE11_y, dE20_x, dE20_y, dE21_x, dE21_y) )
    {
      ASSERT( OrientationEdotE(dE10_x, dE10_y, dE11_x, dE11_y, 
			       dE20_x, dE20_y, dE21_x, dE21_y) < 0 );

      switch ( APM::GetAccuracyType() )
	{
	case APM::atGlobal:
	case APM::atLocal:
	  {
	    eResult = InscribedCircleCenterVEEAntiparallelLocal(pdC_x, pdC_y, pdR, 
								dV00_x, dV00_y, 
								dV0_x, dV0_y, 
								dV01_x, dV01_y, 
								dE10_x, dE10_y, dE11_x, dE11_y,
								dE20_x, dE20_y, dE21_x, dE21_y);
	  }
	break;
	case APM::atExact:
	  {
	    eResult = InscribedCircleCenterVEEAntiparallelAdaptive(pdC_x, pdC_y, pdR, 
								   dV00_x, dV00_y, 
								   dV0_x, dV0_y, 
								   dV01_x, dV01_y, 
								   dE10_x, dE10_y, dE11_x, dE11_y,
								   dE20_x, dE20_y, dE21_x, dE21_y);
	  }
	break;
	case APM::atMPFUN:
	  {
	    eResult = InscribedCircleCenterVEEAntiparallelMPFUN(pdC_x, pdC_y, pdR, 
								dV00_x, dV00_y, 
								dV0_x, dV0_y, 
								dV01_x, dV01_y, 
								dE10_x, dE10_y, dE11_x, dE11_y,
								dE20_x, dE20_y, dE21_x, dE21_y);
	  }
	break;

	case APM::atMPFUNFilter:
	  {
	    eResult = InscribedCircleCenterVEEAntiparallelMPFUNFilter(pdC_x, pdC_y, pdR, 
								dV00_x, dV00_y, 
								dV0_x, dV0_y, 
								dV01_x, dV01_y, 
								dE10_x, dE10_y, dE11_x, dE11_y,
								dE20_x, dE20_y, dE21_x, dE21_y);
	  }
	break;
	case APM::atMPFUNNormalized:
	  {
	    eResult = InscribedCircleCenterVEEAntiparallelMPFUNNormalized(pdC_x, pdC_y, pdR, 
									  dV00_x, dV00_y, 
									  dV0_x, dV0_y, 
									  dV01_x, dV01_y, 
									  dE10_x, dE10_y, dE11_x, dE11_y,
									  dE20_x, dE20_y, dE21_x, dE21_y);
	  }
	break;
	default:
	  ASSERT( FALSE );
	  break;
	}
    }
  else
    {
      switch ( APM::GetAccuracyType() )
	{
	case APM::atGlobal:
	case APM::atLocal:
	  {
	    eResult = InscribedCircleCenterVEELocal(pdC_x, pdC_y, pdR, 
						    dV00_x, dV00_y, 
						    dV0_x, dV0_y, 
						    dV01_x, dV01_y, 
						    dE10_x, dE10_y, dE11_x, dE11_y,
						    dE20_x, dE20_y, dE21_x, dE21_y);
	  }
	break;
	case APM::atExact:
	  {
	    eResult = InscribedCircleCenterVEEAdaptive(pdC_x, pdC_y, pdR, 
						       dV00_x, dV00_y, 
						       dV0_x, dV0_y, 
						       dV01_x, dV01_y, 
						       dE10_x, dE10_y, dE11_x, dE11_y,
						       dE20_x, dE20_y, dE21_x, dE21_y);
	  }
	break;
	case APM::atMPFUN:
	  {
	    eResult = InscribedCircleCenterVEEMPFUN(pdC_x, pdC_y, pdR, 
						    dV00_x, dV00_y, 
						    dV0_x, dV0_y, 
						    dV01_x, dV01_y, 
						    dE10_x, dE10_y, dE11_x, dE11_y,
						    dE20_x, dE20_y, dE21_x, dE21_y);
	  }
	break;
	case APM::atMPFUNFilter:
	  {
	    eResult = InscribedCircleCenterVEEMPFUNFilter(pdC_x, pdC_y, pdR, 
						    dV00_x, dV00_y, 
						    dV0_x, dV0_y, 
						    dV01_x, dV01_y, 
						    dE10_x, dE10_y, dE11_x, dE11_y,
						    dE20_x, dE20_y, dE21_x, dE21_y);
	  }
	break;
	case APM::atMPFUNNormalized:
	  {
	    eResult = InscribedCircleCenterVEEMPFUNNormalized(pdC_x, pdC_y, pdR, 
							      dV00_x, dV00_y, 
							      dV0_x, dV0_y, 
							      dV01_x, dV01_y, 
							      dE10_x, dE10_y, dE11_x, dE11_y,
							      dE20_x, dE20_y, dE21_x, dE21_y);
	  }
	break;
	default:
	  ASSERT( FALSE );
	  break;
	}
    }

  return eResult;
}

APM2D::InscribedCircleCenterType 
APM2D::InscribedCircleCenterVEvE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				 DOUBLE dV00_x, DOUBLE dV00_y,
				 DOUBLE dV0_x, DOUBLE dV0_y,
				 DOUBLE dV01_x, DOUBLE dV01_y,
				 DOUBLE dE10_x, DOUBLE dE10_y, 
				 DOUBLE dE11_E20_x, DOUBLE dE11_E20_y,
				 DOUBLE dE21_x, DOUBLE dE21_y)
{
  ASSERT( pdC_x != NULL && pdC_y != NULL && pdR != NULL );

  APM2D::InscribedCircleCenterType eResult;

  eResult = APM2D::icctNull;
  
  //////////////////////////////////////////////////
  // Colinear Check

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
    case APM::atExact:
      {
	DOUBLE dSinTheta;
	dSinTheta = APM2D::SinTheta(dE10_x, dE10_y, dE11_E20_x, dE11_E20_y, dE21_x, dE21_y);
	if ( fabs(dSinTheta) <= s_cdEpsilonInscribedCircleCenterVEvEParallel )
	  {
	    eResult = InscribedCircleCenterVvE(pdC_x, pdC_y, pdR, 
					       dV00_x, dV00_y,
					       dV0_x, dV0_y,
					       dV01_x, dV01_y,
					       dE11_E20_x, dE11_E20_y, dE21_x, dE21_y);
	    if ( eResult != APM2D::icctNull )
	      {
		eResult = icctVEcE;
	      }
	    return eResult;      
	  }
      }
      break;
    case APM::atMPFUN:
    case APM::atMPFUNFilter:
    case APM::atMPFUNNormalized:
      {
	INT_32 iResult;
	iResult = OrientationVVV(dE10_x, dE10_y, dE11_E20_x, dE11_E20_y, dE21_x, dE21_y);
	if ( iResult == 0 )
	  {
	    eResult = InscribedCircleCenterVvE(pdC_x, pdC_y, pdR, 
					       dV00_x, dV00_y,
					       dV0_x, dV0_y,
					       dV01_x, dV01_y,
					       dE11_E20_x, dE11_E20_y, dE21_x, dE21_y);
	    if ( eResult != APM2D::icctNull )
	      {
		eResult = icctVEcE;
	      }
	    return eResult;      
	  }  
      }
      break;
    default:
      {
	ASSERT( FALSE );
      }
      break;
    }

  //////////////////////////////////////////////////
  // CCW Checks

  if ( !Counterclockwise(dE10_x, dE10_y, dE11_E20_x, dE11_E20_y, dV0_x, dV0_y) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  if ( !Counterclockwise(dE11_E20_x, dE11_E20_y, dE21_x, dE21_y, dV0_x, dV0_y) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  if ( !Counterclockwise(dE10_x, dE10_y, dE11_E20_x, dE11_E20_y, dE21_x, dE21_y) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  if ( !Counterclockwise(dE11_E20_x, dE11_E20_y, dE21_x, dE21_y, dE10_x, dE10_y) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  //////////////////////////////////////////////////
  // Anitparallel Check

  // Not necessary because of the topology

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InscribedCircleCenterVEELocal(pdC_x, pdC_y, pdR, 
						dV00_x, dV00_y, 
						dV0_x, dV0_y, 
						dV01_x, dV01_y, 
						dE10_x, dE10_y, dE11_E20_x, dE11_E20_y,
						dE11_E20_x, dE11_E20_y, dE21_x, dE21_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InscribedCircleCenterVEEAdaptive(pdC_x, pdC_y, pdR, 
						   dV00_x, dV00_y, 
						   dV0_x, dV0_y, 
						   dV01_x, dV01_y, 
						   dE10_x, dE10_y, dE11_E20_x, dE11_E20_y,
						   dE11_E20_x, dE11_E20_y, dE21_x, dE21_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InscribedCircleCenterVEEMPFUN(pdC_x, pdC_y, pdR, 
						dV00_x, dV00_y, 
						dV0_x, dV0_y, 
						dV01_x, dV01_y, 
						dE10_x, dE10_y, dE11_E20_x, dE11_E20_y,
						dE11_E20_x, dE11_E20_y, dE21_x, dE21_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = InscribedCircleCenterVEEMPFUNFilter(pdC_x, pdC_y, pdR, 
						dV00_x, dV00_y, 
						dV0_x, dV0_y, 
						dV01_x, dV01_y, 
						dE10_x, dE10_y, dE11_E20_x, dE11_E20_y,
						dE11_E20_x, dE11_E20_y, dE21_x, dE21_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InscribedCircleCenterVEEMPFUNNormalized(pdC_x, pdC_y, pdR, 
							  dV00_x, dV00_y, 
							  dV0_x, dV0_y, 
							  dV01_x, dV01_y, 
							  dE10_x, dE10_y, dE11_E20_x, dE11_E20_y,
							  dE11_E20_x, dE11_E20_y, dE21_x, dE21_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }

  return eResult;
}

static APM2D::InscribedCircleCenterType 
InscribedCircleCenterVEELocal(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			      DOUBLE dV00_x, DOUBLE dV00_y,
			      DOUBLE dV0_x, DOUBLE dV0_y,
			      DOUBLE dV01_x, DOUBLE dV01_y,
			      DOUBLE dE10_x, DOUBLE dE10_y, 
			      DOUBLE dE11_x, DOUBLE dE11_y,
			      DOUBLE dE20_x, DOUBLE dE20_y, 
			      DOUBLE dE21_x, DOUBLE dE21_y)
{
  DOUBLE pdC[3], dR;
  DOUBLE pppdV[3][2][2], pppdVN[1][2][2]; // Endpoints
  DOUBLE ppdE[3][2];     // Edge vectors
  DOUBLE pdVertexZone[2];
  DOUBLE ppdEdgeZone[2][2];
  DOUBLE dDen, dDenR, dDisc;
  DOUBLE dE1dotE1, dE2dotE2, dE1dotE2;
  DOUBLE dE1dotE10, dE1dotE11, dE2dotE20, dE2dotE21;
  DOUBLE dE1xE2, dE10xE11, dE20xE21;
  DOUBLE dE1E2plusE1dotE2, dE1E2minusE1dotE2;
  DOUBLE dPlusOrMinus, dSinTheta, dSinThetaTemp;
  DOUBLE dV00dotV00, dV01dotV01, dE1dotV00, dE1dotV01, dE2dotV00, dE2dotV01;
  DOUBLE dMagCE10, dMagCE11, dMagCE20, dMagCE21;
  UINT_32 i, j;
  
  // Translate V1 to the origin
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = dE10_x - dV0_x;
  pppdV[1][0][1] = dE10_y - dV0_y;
  pppdV[1][1][0] = dE11_x - dV0_x;
  pppdV[1][1][1] = dE11_y - dV0_y;
  pppdV[2][0][0] = dE20_x - dV0_x;
  pppdV[2][0][1] = dE20_y - dV0_y;
  pppdV[2][1][0] = dE21_x - dV0_x;
  pppdV[2][1][1] = dE21_y - dV0_y;
  
  // Vertex neighbor vectors
  memset(pppdVN, 0, sizeof(pppdV));
  pppdVN[0][0][0] = dV0_x - dV00_x;
  pppdVN[0][0][1] = dV0_y - dV00_y;
  pppdVN[0][1][0] = dV0_x - dV01_x;
  pppdVN[0][1][1] = dV0_y - dV01_y;

  memset(ppdE, 0, sizeof(ppdE));
  ppdE[1][0] = dE11_x - dE10_x;
  ppdE[1][1] = dE11_y - dE10_y;
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;
  
  dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);
  dE1dotE2 = (ppdE[1][0]*ppdE[2][0]) + (ppdE[1][1]*ppdE[2][1]);
  dE1dotE10 = (ppdE[1][0]*pppdV[1][0][0]) + (ppdE[1][1]*pppdV[1][0][1]);
  dE1dotE11 = (ppdE[1][0]*pppdV[1][1][0]) + (ppdE[1][1]*pppdV[1][1][1]);
  dE2dotE20 = (ppdE[2][0]*pppdV[2][0][0]) + (ppdE[2][1]*pppdV[2][0][1]);
  dE2dotE21 = (ppdE[2][0]*pppdV[2][1][0]) + (ppdE[2][1]*pppdV[2][1][1]);

  dE1xE2 = (ppdE[1][0]*ppdE[2][1]) - (ppdE[1][1]*ppdE[2][0]);
  dE10xE11 = (pppdV[1][0][0]*pppdV[1][1][1]) - (pppdV[1][0][1]*pppdV[1][1][0]);
  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);

  dE1E2plusE1dotE2 = sqrt(dE1dotE1*dE2dotE2) + dE1dotE2;
  dE1E2minusE1dotE2 = sqrt(dE1dotE1*dE2dotE2) - dE1dotE2;

  dV00dotV00 = ( (pppdVN[0][0][0] * pppdVN[0][0][0]) + (pppdVN[0][0][1] * pppdVN[0][0][1]) );
  dV01dotV01 = ( (pppdVN[0][1][0] * pppdVN[0][1][0]) + (pppdVN[0][1][1] * pppdVN[0][1][1]) );

  dE1dotV00 = ( (ppdE[1][0] * pppdVN[0][0][0]) + (ppdE[1][1] * pppdVN[0][0][1]) );
  dE1dotV01 = ( (ppdE[1][0] * pppdVN[0][1][0]) + (ppdE[1][1] * pppdVN[0][1][1]) );
  dE2dotV00 = ( (ppdE[2][0] * pppdVN[0][0][0]) + (ppdE[2][1] * pppdVN[0][0][1]) );
  dE2dotV01 = ( (ppdE[2][0] * pppdVN[0][1][0]) + (ppdE[2][1] * pppdVN[0][1][1]) );

  dSinTheta = dE1xE2/sqrt(dE1dotE1*dE2dotE2);
  if ( fabs(dSinTheta) < s_cdEpsilonInscribedCircleCenterVEEAntiparallel )
    {
      if ( dE1dotE2 > 0.0 )
	{
	  // Colinear
	  *pdC_x = 0.0;
	  *pdC_y = 0.0;
	  *pdR = 0.0;
	  return APM2D::icctNull;
	}
      else
	{
	  return InscribedCircleCenterVEEAntiparallelLocal(pdC_x, pdC_y, pdR, 
							   dV00_x, dV00_y, 
							   dV0_x, dV0_y, 
							   dV01_x, dV01_y, 
							   dE10_x, dE10_y, 
							   dE11_x, dE11_y, 
							   dE20_x, dE20_y, 
							   dE21_x, dE21_y);
	}
    }

  //dSinTheta = dE10xE11/sqrt(dE10dotE10*dE11dotE11);
  dSinTheta = APM2D::SinTheta(dV0_x, dV0_y, dE10_x, dE10_y, dE11_x, dE11_y);
  dSinThetaTemp = APM2D::SinTheta(dE10_x, dE10_y, dE11_x, dE11_y, dV0_x, dV0_y);
  if ( dSinThetaTemp > dSinTheta )
    {
      dSinTheta = dSinThetaTemp;
    }
  if ( dSinTheta <= s_cdEpsilonInscribedCircleCenterVEEVertexOnEdge )
    {
      // V0 is on E1 and it is not one of the endpoints
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  //dSinTheta = dE20xE21/sqrt(dE20dotE20*dE21dotE21);
  dSinTheta = APM2D::SinTheta(dV0_x, dV0_y, dE20_x, dE20_y, dE21_x, dE21_y);
  dSinThetaTemp = APM2D::SinTheta(dE20_x, dE20_y, dE21_x, dE21_y, dV0_x, dV0_y);
  if ( dSinThetaTemp > dSinTheta )
    {
      dSinTheta = dSinThetaTemp;
    }
  if ( dSinTheta <= s_cdEpsilonInscribedCircleCenterVEEVertexOnEdge )
    {
      // V0 is on E2 and it is not one of the endpoints
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  dDen = dE1xE2*dE1E2minusE1dotE2;
  ASSERT( dDen != 0.0 );
  dDenR = dE1E2minusE1dotE2;
  ASSERT( dDenR != 0.0 );

  dDisc = 2.0*dE10xE11*dE20xE21*dE1E2plusE1dotE2;
  if ( -s_cdEpsilonInscribedCircleCenterVEEDisc <= dDisc && dDisc <= 0.0 )
    {
      dDisc = 0.0;
    }
  if ( dDisc < 0.0 ) 
    {
      // Imaginary Solutions
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  dPlusOrMinus = 1.0;
  if ( dE1xE2 > 0.0 )
    {
      dPlusOrMinus = -1.0;
    }

  //////////////////////////////////////////////////
  // Compute the center and radius

  pdC[0] = ( (dE1dotE1*ppdE[2][0] - dE1dotE2*ppdE[1][0])*dE20xE21
	     + (dE1dotE2*ppdE[2][0] - dE2dotE2*ppdE[1][0])*dE10xE11
	     + dPlusOrMinus*(sqrt(dE1dotE1)*ppdE[2][0] - sqrt(dE2dotE2)*ppdE[1][0])*sqrt(dDisc) ) / dDen;
  pdC[1] = ( (dE1dotE1*ppdE[2][1] - dE1dotE2*ppdE[1][1])*dE20xE21
	     + (dE1dotE2*ppdE[2][1] - dE2dotE2*ppdE[1][1])*dE10xE11
	     + dPlusOrMinus*(sqrt(dE1dotE1)*ppdE[2][1] - sqrt(dE2dotE2)*ppdE[1][1])*sqrt(dDisc) ) / dDen;
  pdC[2] = sqrt((pdC[0]*pdC[0]) + (pdC[1]*pdC[1]));

  dR = ( sqrt(dE1dotE1)*dE20xE21 + sqrt(dE2dotE2)*dE10xE11 + dPlusOrMinus*sqrt(dDisc) ) / dDenR;

  //////////////////////////////////////////////////
  // Check the Edge Zone

  dMagCE10 = ( (pdC[0] - pppdV[1][0][0])*(pdC[0] - pppdV[1][0][0]) + 
	       (pdC[1] - pppdV[1][0][1])*(pdC[1] - pppdV[1][0][1]) );
  dMagCE11 = ( (pdC[0] - pppdV[1][1][0])*(pdC[0] - pppdV[1][1][0]) + 
	       (pdC[1] - pppdV[1][1][1])*(pdC[1] - pppdV[1][1][1]) );
  dMagCE20 = ( (pdC[0] - pppdV[2][0][0])*(pdC[0] - pppdV[2][0][0]) + 
	       (pdC[1] - pppdV[2][0][1])*(pdC[1] - pppdV[2][0][1]) );
  dMagCE21 = ( (pdC[0] - pppdV[2][1][0])*(pdC[0] - pppdV[2][1][0]) + 
	       (pdC[1] - pppdV[2][1][1])*(pdC[1] - pppdV[2][1][1]) );

  if ( dMagCE10 <= s_cdEpsilonNullEdge )
    {
      ppdEdgeZone[0][0] = 0.0;
    }
  else
    {
      ppdEdgeZone[0][0] = -( ( dE1xE2*dE1dotE10 + dE10xE11*dE1E2plusE1dotE2 
			       + dPlusOrMinus*sqrt(dE1dotE1*dDisc) ) / 
			     ( dE1xE2*sqrt(dE1dotE1*dMagCE10) ) );
    }
  if ( dMagCE11 <= s_cdEpsilonNullEdge )
    {
      ppdEdgeZone[0][1] = 0.0;
    }
  else
    {
      ppdEdgeZone[0][1] = ( ( dE1xE2*dE1dotE11 + dE10xE11*dE1E2plusE1dotE2 
			      + dPlusOrMinus*sqrt(dE1dotE1*dDisc) ) / 
			    ( dE1xE2*sqrt(dE1dotE1*dMagCE11) ) );
    }
  if ( dMagCE20 <= s_cdEpsilonNullEdge )
    {
      ppdEdgeZone[1][0] = 0.0;
    }
  else
    {
      ppdEdgeZone[1][0] = -( ( dE1xE2*dE2dotE20 - dE20xE21*dE1E2plusE1dotE2 
			       - dPlusOrMinus*sqrt(dE2dotE2*dDisc) ) / 
			     ( dE1xE2*sqrt(dE2dotE2*dMagCE20) ) );
    }
  if ( dMagCE21 <= s_cdEpsilonNullEdge )
    {
      ppdEdgeZone[1][1] = 0.0;
    }
  else
    {
      ppdEdgeZone[1][1] = ( ( dE1xE2*dE2dotE21 - dE20xE21*dE1E2plusE1dotE2 
			      - dPlusOrMinus*sqrt(dE2dotE2*dDisc) ) / 
			    ( dE1xE2*sqrt(dE2dotE2*dMagCE21) ) );
    }
  
  for ( i = 0; i < 2; i++ )
    {
      for ( j = 0; j < 2; j++ )
	{
	  if ( ppdEdgeZone[i][j] < -s_cdEpsilonInscribedCircleCenterVEEEdgeZone )
	    {
#ifdef DEBUG_EPSILON_VALUES
	      if ( !( ppdEdgeZone[i][j] < -(s_cdEpsilonInscribedCircleCenterVVEEdgeZone*s_cdOrderOfMagnitude) ) )
		{
		  fprintf(stderr, "  VEE: Edge Zone: %e < %e < %e\n",
			  s_cdEpsilonInscribedCircleCenterVEEEdgeZone*s_cdOrderOfMagnitude,
			  ppdEdgeZone[i][j],
			  s_cdEpsilonInscribedCircleCenterVEEEdgeZone);
		}
#endif // DEBUG_EPSILON_VALUES	      
	      *pdC_x = 0.0;
	      *pdC_y = 0.0;
	      *pdR = 0.0;
	      return APM2D::icctNull;
	    }
	}
    }

  //////////////////////////////////////////////////
  // Check the Vertex Zone

  pdVertexZone[0] = ( ( ( dE1dotE1*dE2dotV00 - dE1dotE2*dE1dotV00 )*dE20xE21 
			+ ( dE1dotE2*dE2dotV00 - dE2dotE2*dE1dotV00 )*dE10xE11 
			+ dPlusOrMinus*( sqrt(dE1dotE1)*dE2dotV00 - sqrt(dE2dotE2)*dE1dotV00 )*sqrt(dDisc) ) /
		      ( dDen*dR*sqrt(dV00dotV00) ) );
  pdVertexZone[1] = ( ( ( dE1dotE1*dE2dotV01 - dE1dotE2*dE1dotV01 )*dE20xE21 
			+ ( dE1dotE2*dE2dotV01 - dE2dotE2*dE1dotV01 )*dE10xE11 
			+ dPlusOrMinus*( sqrt(dE1dotE1)*dE2dotV01 - sqrt(dE2dotE2)*dE1dotV01 )*sqrt(dDisc) ) /
		      ( dDen*dR*sqrt(dV01dotV01) ) );

  for ( i = 0; i < 2; i++ )
    {
      if ( pdVertexZone[i] < -s_cdEpsilonInscribedCircleCenterVEEVertexZone )
	{
#ifdef DEBUG_EPSILON_VALUES
	      if ( !( pdVertexZone[i] < -(s_cdEpsilonInscribedCircleCenterVEEVertexZone*s_cdOrderOfMagnitude) ) )
		{
		  fprintf(stderr, "  VVE: Vertex Zone: %e < %e < %e\n",
			  s_cdEpsilonInscribedCircleCenterVEEVertexZone*s_cdOrderOfMagnitude,
			  pdVertexZone[i],
			  s_cdEpsilonInscribedCircleCenterVEEVertexZone);
		}
#endif // DEBUG_EPSILON_VALUES	      
	  *pdC_x = 0.0;
	  *pdC_y = 0.0;
	  *pdR = 0.0;
	  return APM2D::icctNull;
	}
    }
  
  if ( !( fabs(pdC[2] - dR) < s_cdEpsilonInscribedCircleCenterVEERadius ) )
    {
      fprintf(stderr, "VEE: C = ( %f, %f, %f ) != %f = R\n", pdC[0], pdC[1], pdC[2], dR);
      //ASSERT( FALSE );
    }

  // Translate the answer back out away from the origin
  *pdC_x = dV0_x + pdC[0];
  *pdC_y = dV0_y + pdC[1];
  *pdR = dR;
  return APM2D::icctVEE;
}

static APM2D::InscribedCircleCenterType 
InscribedCircleCenterVEEAdaptive(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				 DOUBLE dV00_x, DOUBLE dV00_y,
				 DOUBLE dV0_x, DOUBLE dV0_y,
				 DOUBLE dV01_x, DOUBLE dV01_y,
				 DOUBLE dE10_x, DOUBLE dE10_y, 
				 DOUBLE dE11_x, DOUBLE dE11_y,
				 DOUBLE dE20_x, DOUBLE dE20_y, 
				 DOUBLE dE21_x, DOUBLE dE21_y)
{
  return InscribedCircleCenterVEELocal(pdC_x, pdC_y, pdR, 
				       dV00_x, dV00_y, 
				       dV0_x, dV0_y, 
				       dV01_x, dV01_y, 
				       dE10_x, dE10_y, dE11_x, dE11_y,
				       dE20_x, dE20_y, dE21_x, dE21_y);
}

APM2D::EdgeZoneType APM2D::EdgeZoneVEE(DOUBLE dV0_x, DOUBLE dV0_y,
				       DOUBLE dE10_x, DOUBLE dE10_y, 
				       DOUBLE dE11_x, DOUBLE dE11_y,
				       DOUBLE dE20_x, DOUBLE dE20_y, 
				       DOUBLE dE21_x, DOUBLE dE21_y,
				       DOUBLE dE30_x, DOUBLE dE30_y,
				       DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::EdgeZoneType eResult;

  eResult = APM2D::eztMiddle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = EdgeZoneVEELocal(dV0_x, dV0_y,
				   dE10_x, dE10_y, dE11_x, dE11_y,
				   dE20_x, dE20_y, dE21_x, dE21_y,
				   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atExact:
      {
	eResult = EdgeZoneVEEAdaptive(dV0_x, dV0_y,
				      dE10_x, dE10_y, dE11_x, dE11_y,
				      dE20_x, dE20_y, dE21_x, dE21_y,
				      dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = EdgeZoneVEEMPFUN(dV0_x, dV0_y,
				   dE10_x, dE10_y, dE11_x, dE11_y,
				   dE20_x, dE20_y, dE21_x, dE21_y,
				   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = EdgeZoneVEEMPFUNFilter(dV0_x, dV0_y,
				   dE10_x, dE10_y, dE11_x, dE11_y,
				   dE20_x, dE20_y, dE21_x, dE21_y,
				   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = EdgeZoneVEEMPFUNNormalized(dV0_x, dV0_y,
					     dE10_x, dE10_y, dE11_x, dE11_y,
					     dE20_x, dE20_y, dE21_x, dE21_y,
					     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::EdgeZoneType EdgeZoneVEELocal(DOUBLE dV0_x, DOUBLE dV0_y,
					    DOUBLE dE10_x, DOUBLE dE10_y, 
					    DOUBLE dE11_x, DOUBLE dE11_y,
					    DOUBLE dE20_x, DOUBLE dE20_y, 
					    DOUBLE dE21_x, DOUBLE dE21_y,
					    DOUBLE dE30_x, DOUBLE dE30_y,
					    DOUBLE dE31_x, DOUBLE dE31_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE pdC[2];
  DOUBLE pdEdgeZone[2];
  DOUBLE dE1dotE1, dE2dotE2, dE1dotE2, dE2dotE3, dE1dotE3;
  DOUBLE dE3dotE30, dE3dotE31;
  DOUBLE dE1xE2, dE10xE11, dE20xE21;
  DOUBLE dE1E2minusE1dotE2;
  DOUBLE dDisc, dPlusOrMinus;
  DOUBLE dDen;
  DOUBLE dE3dotE3, dMagCE30, dMagCE31;
  APM2D::EdgeZoneType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = dE10_x - dV0_x;
  pppdV[1][0][1] = dE10_y - dV0_y;
  pppdV[1][1][0] = dE11_x - dV0_x;
  pppdV[1][1][1] = dE11_y - dV0_y;
  pppdV[2][0][0] = dE20_x - dV0_x;
  pppdV[2][0][1] = dE20_y - dV0_y;
  pppdV[2][1][0] = dE21_x - dV0_x;
  pppdV[2][1][1] = dE21_y - dV0_y;
  pppdV[3][0][0] = dE30_x - dV0_x;
  pppdV[3][0][1] = dE30_y - dV0_y;
  pppdV[3][1][0] = dE31_x - dV0_x;
  pppdV[3][1][1] = dE31_y - dV0_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[1][0] = dE11_x - dE10_x;
  ppdE[1][1] = dE11_y - dE10_y;
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;
  ppdE[3][0] = dE31_x - dE30_x;
  ppdE[3][1] = dE31_y - dE30_y;
  
  dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);
  dE1dotE2 = (ppdE[1][0]*ppdE[2][0]) + (ppdE[1][1]*ppdE[2][1]);
  dE2dotE3 = (ppdE[2][0]*ppdE[3][0]) + (ppdE[2][1]*ppdE[3][1]);
  dE1dotE3 = (ppdE[1][0]*ppdE[3][0]) + (ppdE[1][1]*ppdE[3][1]);
  dE3dotE3 = (ppdE[3][0]*ppdE[3][0]) + (ppdE[3][1]*ppdE[3][1]);

  dE3dotE30 = (ppdE[3][0]*pppdV[3][0][0]) + (ppdE[3][1]*pppdV[3][0][1]);
  dE3dotE31 = (ppdE[3][0]*pppdV[3][1][0]) + (ppdE[3][1]*pppdV[3][1][1]);
  
  dE1xE2 = (ppdE[1][0]*ppdE[2][1]) - (ppdE[1][1]*ppdE[2][0]);
  dE10xE11 = (pppdV[1][0][0]*pppdV[1][1][1]) - (pppdV[1][0][1]*pppdV[1][1][0]);
  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);

  dE1E2minusE1dotE2 = ( sqrt(dE1dotE1*dE2dotE2) - dE1dotE2 );

  dDisc = 2.0*dE10xE11*dE20xE21*( sqrt(dE1dotE1*dE2dotE2) + dE1dotE2 );
  if ( -s_cdEpsilonInscribedCircleCenterVEEDisc <= dDisc && dDisc <= 0.0 )
    {
      dDisc = 0.0;
    }
  ASSERT( dDisc >= 0.0 ); 

  ASSERT( dE1xE2 != 0.0 );
  dPlusOrMinus = 1.0;
  if ( dE1xE2 > 0.0 )
    {
      dPlusOrMinus = -1.0;
    }
  
  dDen = dE1xE2*dE1E2minusE1dotE2;
  ASSERT( dDen != 0.0 );

  pdC[0] = ( (dE1dotE1*ppdE[2][0] - dE1dotE2*ppdE[1][0])*dE20xE21
	     + (dE1dotE2*ppdE[2][0] - dE2dotE2*ppdE[1][0])*dE10xE11
	     + dPlusOrMinus*(sqrt(dE1dotE1)*ppdE[2][0] - sqrt(dE2dotE2)*ppdE[1][0])*sqrt(dDisc) ) / dDen;
  pdC[1] = ( (dE1dotE1*ppdE[2][1] - dE1dotE2*ppdE[1][1])*dE20xE21
	     + (dE1dotE2*ppdE[2][1] - dE2dotE2*ppdE[1][1])*dE10xE11
	     + dPlusOrMinus*(sqrt(dE1dotE1)*ppdE[2][1] - sqrt(dE2dotE2)*ppdE[1][1])*sqrt(dDisc) ) / dDen;

  dMagCE30 = ( (pdC[0] - pppdV[3][0][0])*(pdC[0] - pppdV[3][0][0]) + 
	       (pdC[1] - pppdV[3][0][1])*(pdC[1] - pppdV[3][0][1]) );
  dMagCE31 = ( (pdC[0] - pppdV[3][1][0])*(pdC[0] - pppdV[3][1][0]) + 
	       (pdC[1] - pppdV[3][1][1])*(pdC[1] - pppdV[3][1][1]) );

  if ( dMagCE30 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[0] = 0.0;
    }
  else
    {
      pdEdgeZone[0] = ( ( dE1xE2*dE1E2minusE1dotE2*dE3dotE30
			  + ( dE2dotE2*dE1dotE3 - dE1dotE2*dE2dotE3 )*dE10xE11
			  - ( dE1dotE1*dE2dotE3 - dE1dotE2*dE1dotE3 )*dE20xE21
			  + dPlusOrMinus*( sqrt(dE2dotE2)*dE1dotE3 - sqrt(dE1dotE1)*dE2dotE3 )*sqrt(dDisc) ) / 
			( dDen*sqrt(dE3dotE3*dMagCE30) ) );
    }
  if ( dMagCE31 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[1] = 0.0;
    }
  else
    {
      pdEdgeZone[1] = -( ( dE1xE2*dE1E2minusE1dotE2*dE3dotE31
			   + ( dE2dotE2*dE1dotE3 - dE1dotE2*dE2dotE3 )*dE10xE11
			   - ( dE1dotE1*dE2dotE3 - dE1dotE2*dE1dotE3 )*dE20xE21
			   + dPlusOrMinus*( sqrt(dE2dotE2)*dE1dotE3 - sqrt(dE1dotE1)*dE2dotE3 )*sqrt(dDisc) ) / 
			 ( dDen*sqrt(dE3dotE3*dMagCE31) ) );
    }
  
  if ( !( pdEdgeZone[0] < -s_cdEpsilonEdgeZoneVEEEdgeZone ) )
    {
      eResult = APM2D::eztStart;
    }
  else if ( !( pdEdgeZone[1] < -s_cdEpsilonEdgeZoneVEEEdgeZone ) )
    {
      eResult = APM2D::eztEnd;
    }
  else
    {
      eResult = APM2D::eztMiddle;
    }

  return eResult;
}

static APM2D::EdgeZoneType EdgeZoneVEEAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
					       DOUBLE dE10_x, DOUBLE dE10_y, 
					       DOUBLE dE11_x, DOUBLE dE11_y,
					       DOUBLE dE20_x, DOUBLE dE20_y, 
					       DOUBLE dE21_x, DOUBLE dE21_y,
					       DOUBLE dE30_x, DOUBLE dE30_y,
					       DOUBLE dE31_x, DOUBLE dE31_y)
{
  return EdgeZoneVEELocal(dV0_x, dV0_y,
			  dE10_x, dE10_y, dE11_x, dE11_y,
			  dE20_x, dE20_y, dE21_x, dE21_y,
			  dE30_x, dE30_y, dE31_x, dE31_y);
}

APM2D::CircleTestType APM2D::InCircleVEE_V(DOUBLE dV0_x, DOUBLE dV0_y,
					   DOUBLE dE10_x, DOUBLE dE10_y, 
					   DOUBLE dE11_x, DOUBLE dE11_y,
					   DOUBLE dE20_x, DOUBLE dE20_y, 
					   DOUBLE dE21_x, DOUBLE dE21_y,
					   DOUBLE dV3_x, DOUBLE dV3_y)
{
  APM2D::CircleTestType eResult;

  eResult = APM2D::cttOnCircle;

  // Check to see if the vertex is in the edge envelope
  if ( !Counterclockwise(dE10_x, dE10_y, 
			 dE11_x, dE11_y,
			 dV3_x, dV3_y) )
    {
      return APM2D::cttOutCircle;
    }

  if ( !Counterclockwise(dE20_x, dE20_y, 
			 dE21_x, dE21_y,
			 dV3_x, dV3_y) )
    {
      return APM2D::cttOutCircle;
    }

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InCircleVEE_VLocal(dV0_x, dV0_y,
				     dE10_x, dE10_y, dE11_x, dE11_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dV3_x, dV3_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InCircleVEE_VAdaptive(dV0_x, dV0_y,
					dE10_x, dE10_y, dE11_x, dE11_y,
					dE20_x, dE20_y, dE21_x, dE21_y,
					dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InCircleVEE_VMPFUN(dV0_x, dV0_y,
				     dE10_x, dE10_y, dE11_x, dE11_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = InCircleVEE_VMPFUNFilter(dV0_x, dV0_y,
				     dE10_x, dE10_y, dE11_x, dE11_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InCircleVEE_VMPFUNNormalized(dV0_x, dV0_y,
					       dE10_x, dE10_y, dE11_x, dE11_y,
					       dE20_x, dE20_y, dE21_x, dE21_y,
					       dV3_x, dV3_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::CircleTestType InCircleVEE_VLocal(DOUBLE dV0_x, DOUBLE dV0_y,
						DOUBLE dE10_x, DOUBLE dE10_y, 
						DOUBLE dE11_x, DOUBLE dE11_y,
						DOUBLE dE20_x, DOUBLE dE20_y, 
						DOUBLE dE21_x, DOUBLE dE21_y,
						DOUBLE dV3_x, DOUBLE dV3_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE dE1dotE1, dE2dotE2, dV3dotV3;
  DOUBLE dE1dotE2, dE2dotV3, dE1dotV3;
  DOUBLE dE1xE2, dE10xE11, dE20xE21;
  DOUBLE dE1E2minusE1dotE2;
  DOUBLE dDisc, dPlusOrMinus;
  DOUBLE dDen;
  DOUBLE dResult;
  APM2D::CircleTestType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = dE10_x - dV0_x;
  pppdV[1][0][1] = dE10_y - dV0_y;
  pppdV[1][1][0] = dE11_x - dV0_x;
  pppdV[1][1][1] = dE11_y - dV0_y;
  pppdV[2][0][0] = dE20_x - dV0_x;
  pppdV[2][0][1] = dE20_y - dV0_y;
  pppdV[2][1][0] = dE21_x - dV0_x;
  pppdV[2][1][1] = dE21_y - dV0_y;
  pppdV[3][0][0] = dV3_x - dV0_x;
  pppdV[3][0][1] = dV3_y - dV0_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[1][0] = dE11_x - dE10_x;
  ppdE[1][1] = dE11_y - dE10_y;
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;

  dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);
  dV3dotV3 = (pppdV[3][0][0]*pppdV[3][0][0]) + (pppdV[3][0][1]*pppdV[3][0][1]);

  dE1dotE2 = (ppdE[1][0]*ppdE[2][0]) + (ppdE[1][1]*ppdE[2][1]);
  dE2dotV3 = (ppdE[2][0]*pppdV[3][0][0]) + (ppdE[2][1]*pppdV[3][0][1]);
  dE1dotV3 = (ppdE[1][0]*pppdV[3][0][0]) + (ppdE[1][1]*pppdV[3][0][1]);

  dE1xE2 = (ppdE[1][0]*ppdE[2][1]) - (ppdE[1][1]*ppdE[2][0]);
  dE10xE11 = (pppdV[1][0][0]*pppdV[1][1][1]) - (pppdV[1][0][1]*pppdV[1][1][0]);
  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);

  dE1E2minusE1dotE2 = ( sqrt(dE1dotE1*dE2dotE2) - dE1dotE2 );

  dDisc = 2.0*dE10xE11*dE20xE21*( sqrt(dE1dotE1*dE2dotE2) + dE1dotE2 );
  if ( -s_cdEpsilonInscribedCircleCenterVEEDisc <= dDisc && dDisc <= 0.0 )
    {
      dDisc = 0.0;
    }
  ASSERT( dDisc >= 0.0 ); 

  ASSERT( dE1xE2 != 0.0 );
  dPlusOrMinus = 1.0;
  if ( dE1xE2 > 0.0 )
    {
      dPlusOrMinus = -1.0;
    }

  dDen = dE1xE2*dE1E2minusE1dotE2;
  ASSERT( dDen != 0.0 );

  dResult = ( dE1xE2*dE1E2minusE1dotE2*dV3dotV3
	      + 2.0*( dE2dotE2*dE1dotV3 - dE1dotE2*dE2dotV3 )*dE10xE11
	      - 2.0*( dE1dotE1*dE2dotV3 - dE1dotE2*dE1dotV3 )*dE20xE21
	      + dPlusOrMinus*2.0*( sqrt(dE2dotE2)*dE1dotV3 - sqrt(dE1dotE1)*dE2dotV3 )*sqrt(dDisc) ) / dDen; 

#ifdef ENABLE_DEBUGGING_OUTPUT
  fprintf(stderr, "VEE_V Anti: dPlusOrMinus = %e\n", dPlusOrMinus);
  fprintf(stderr, "VEE_V Anti: dResult = %e\n", dResult);
#endif // ENABLE_DEBUGGING_OUTPUT

  if ( dResult < -s_cdEpsilonInCircleVEE_VOnCircle )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( dResult > s_cdEpsilonInCircleVEE_VOnCircle )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

static APM2D::CircleTestType InCircleVEE_VAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dE10_x, DOUBLE dE10_y, 
						   DOUBLE dE11_x, DOUBLE dE11_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   DOUBLE dV3_x, DOUBLE dV3_y)
{
  return InCircleVEE_VLocal(dV0_x, dV0_y,
			    dE10_x, dE10_y, dE11_x, dE11_y,
			    dE20_x, dE20_y, dE21_x, dE21_y,
			    dV3_x, dV3_y);
}

APM2D::CircleTestType APM2D::InCircleVEE_E(DOUBLE dV0_x, DOUBLE dV0_y,
					   DOUBLE dE10_x, DOUBLE dE10_y, 
					   DOUBLE dE11_x, DOUBLE dE11_y,
					   DOUBLE dE20_x, DOUBLE dE20_y, 
					   DOUBLE dE21_x, DOUBLE dE21_y,
					   DOUBLE dE30_x, DOUBLE dE30_y,
					   DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::CircleTestType eResult;

  eResult = APM2D::cttOnCircle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InCircleVEE_ELocal(dV0_x, dV0_y,
				     dE10_x, dE10_y, dE11_x, dE11_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InCircleVEE_EAdaptive(dV0_x, dV0_y,
					dE10_x, dE10_y, dE11_x, dE11_y,
					dE20_x, dE20_y, dE21_x, dE21_y,
					dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InCircleVEE_EMPFUN(dV0_x, dV0_y,
				     dE10_x, dE10_y, dE11_x, dE11_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = InCircleVEE_EMPFUNFilter(dV0_x, dV0_y,
				     dE10_x, dE10_y, dE11_x, dE11_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InCircleVEE_EMPFUNNormalized(dV0_x, dV0_y,
					       dE10_x, dE10_y, dE11_x, dE11_y,
					       dE20_x, dE20_y, dE21_x, dE21_y,
					       dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::CircleTestType InCircleVEE_ELocal(DOUBLE dV0_x, DOUBLE dV0_y,
						DOUBLE dE10_x, DOUBLE dE10_y, 
						DOUBLE dE11_x, DOUBLE dE11_y,
						DOUBLE dE20_x, DOUBLE dE20_y, 
						DOUBLE dE21_x, DOUBLE dE21_y,
						DOUBLE dE30_x, DOUBLE dE30_y,
						DOUBLE dE31_x, DOUBLE dE31_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE dE1dotE1, dE2dotE2, dE3dotE3;
  DOUBLE dE1dotE2;
  DOUBLE dE1xE2, dE2xE3, dE1xE3;
  DOUBLE dE10xE11, dE20xE21, dE30xE31;
  DOUBLE dE1E2plusE1dotE2, dE1E2minusE1dotE2;
  DOUBLE dDisc, dPlusOrMinus, dA, dB;
  DOUBLE dDen;
  DOUBLE dResult;
  APM2D::CircleTestType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = dE10_x - dV0_x;
  pppdV[1][0][1] = dE10_y - dV0_y;
  pppdV[1][1][0] = dE11_x - dV0_x;
  pppdV[1][1][1] = dE11_y - dV0_y;
  pppdV[2][0][0] = dE20_x - dV0_x;
  pppdV[2][0][1] = dE20_y - dV0_y;
  pppdV[2][1][0] = dE21_x - dV0_x;
  pppdV[2][1][1] = dE21_y - dV0_y;
  pppdV[3][0][0] = dE30_x - dV0_x;
  pppdV[3][0][1] = dE30_y - dV0_y;
  pppdV[3][1][0] = dE31_x - dV0_x;
  pppdV[3][1][1] = dE31_y - dV0_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[1][0] = dE11_x - dE10_x;
  ppdE[1][1] = dE11_y - dE10_y;
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;
  ppdE[3][0] = dE31_x - dE30_x;
  ppdE[3][1] = dE31_y - dE30_y;
  
  dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);
  dE3dotE3 = (ppdE[3][0]*ppdE[3][0]) + (ppdE[3][1]*ppdE[3][1]);
  
  dE1dotE2 = (ppdE[1][0]*ppdE[2][0]) + (ppdE[1][1]*ppdE[2][1]);
  
  dE1xE2 = (ppdE[1][0]*ppdE[2][1]) - (ppdE[1][1]*ppdE[2][0]);
  dE2xE3 = (ppdE[2][0]*ppdE[3][1]) - (ppdE[2][1]*ppdE[3][0]);
  dE1xE3 = (ppdE[1][0]*ppdE[3][1]) - (ppdE[1][1]*ppdE[3][0]);
  
  dE10xE11 = (pppdV[1][0][0]*pppdV[1][1][1]) - (pppdV[1][0][1]*pppdV[1][1][0]);
  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);
  dE30xE31 = (pppdV[3][0][0]*pppdV[3][1][1]) - (pppdV[3][0][1]*pppdV[3][1][0]);
  
  dE1E2plusE1dotE2 = ( sqrt(dE1dotE1*dE2dotE2) + dE1dotE2 );
  dE1E2minusE1dotE2 = ( sqrt(dE1dotE1*dE2dotE2) - dE1dotE2 );
  
  dDisc = 2.0*dE10xE11*dE20xE21*dE1E2plusE1dotE2;
  if ( -s_cdEpsilonInscribedCircleCenterVEEDisc <= dDisc && dDisc <= 0.0 )
    {
      dDisc = 0.0;
    }
  ASSERT( dDisc >= 0.0 ); 
  
  ASSERT( dE1xE2 != 0.0 );
  dPlusOrMinus = 1.0;
  if ( dE1xE2 > 0.0 )
    {
      dPlusOrMinus = -1.0;
    }
  
  dA = ( ( dE1dotE2*dE1xE3 - dE1dotE1*dE2xE3 )*dE20xE21
	 + ( dE2dotE2*dE1xE3 - dE1dotE2*dE2xE3 )*dE10xE11
	 + dPlusOrMinus*( sqrt(dE2dotE2)*dE1xE3 - sqrt(dE1dotE1)*dE2xE3 )*sqrt(dDisc) );
  
  dB = ( ( sqrt(dE1dotE1)*dE20xE21 + sqrt(dE2dotE2)*dE10xE11 )*( sqrt(dE1dotE1)*dE20xE21 + sqrt(dE2dotE2)*dE10xE11 )
	 + 2.0*dE10xE11*dE20xE21*dE1E2plusE1dotE2
	 + dPlusOrMinus*2.0*( sqrt(dE1dotE1)*dE20xE21 + sqrt(dE2dotE2)*dE10xE11 )*sqrt(dDisc) );
  
  dDen = dE3dotE3*dE1xE2*dE1xE2*dE1E2minusE1dotE2*dE1E2minusE1dotE2;
  ASSERT( dDen != 0.0 );
  
  dResult = ( dA*dA
	      + 2.0*dE30xE31*dE1xE2*dE1E2minusE1dotE2*dA
	      + dE30xE31*dE30xE31*dE1xE2*dE1xE2*dE1E2minusE1dotE2*dE1E2minusE1dotE2
	      - dE3dotE3*dE1xE2*dE1xE2*dB ) / dDen;
  
  if ( dResult < -s_cdEpsilonInCircleVEE_EOnCircle )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( dResult > s_cdEpsilonInCircleVEE_EOnCircle )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }
  
  return eResult;
}

static APM2D::CircleTestType InCircleVEE_EAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dE10_x, DOUBLE dE10_y, 
						   DOUBLE dE11_x, DOUBLE dE11_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   DOUBLE dE30_x, DOUBLE dE30_y,
						   DOUBLE dE31_x, DOUBLE dE31_y)
{
  return InCircleVEE_ELocal(dV0_x, dV0_y,
			    dE10_x, dE10_y, dE11_x, dE11_y,
			    dE20_x, dE20_y, dE21_x, dE21_y,
			    dE30_x, dE30_y, dE31_x, dE31_y);
}

static APM2D::InscribedCircleCenterType 
InscribedCircleCenterVEEAntiparallelLocal(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					  DOUBLE dV00_x, DOUBLE dV00_y,
					  DOUBLE dV0_x, DOUBLE dV0_y,
					  DOUBLE dV01_x, DOUBLE dV01_y,
					  DOUBLE dE10_x, DOUBLE dE10_y, 
					  DOUBLE dE11_x, DOUBLE dE11_y,
					  DOUBLE dE20_x, DOUBLE dE20_y, 
					  DOUBLE dE21_x, DOUBLE dE21_y)
{
  DOUBLE pdC[3], dR;
  DOUBLE pppdV[3][2][2], pppdVN[1][2][2]; // Endpoints
  DOUBLE ppdE[3][2];     // Edge vectors
  DOUBLE pdVertexZone[2];
  DOUBLE ppdEdgeZone[2][2];
  DOUBLE dDen, dDenR, dDisc, dDiscE2;
  DOUBLE dE1dotE1, dE2dotE2, dE1dotE2;
  DOUBLE dE1dotE10, dE1dotE11, dE2dotE20, dE2dotE21;
  DOUBLE dE10xE11, dE20xE21, dE1xE20, dE2xE10;
  DOUBLE dV00dotV00, dV01dotV01, dE1dotV00, dE1dotV01, dE1xV00, dE1xV01;
  DOUBLE dE1E2minusE1dotE2, dSinTheta, dSinThetaTemp;
  DOUBLE dMagCE10, dMagCE11, dMagCE20, dMagCE21;
  UINT_32 i, j;
  
  // Translate V1 to the origin
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = dE10_x - dV0_x;
  pppdV[1][0][1] = dE10_y - dV0_y;
  pppdV[1][1][0] = dE11_x - dV0_x;
  pppdV[1][1][1] = dE11_y - dV0_y;
  pppdV[2][0][0] = dE20_x - dV0_x;
  pppdV[2][0][1] = dE20_y - dV0_y;
  pppdV[2][1][0] = dE21_x - dV0_x;
  pppdV[2][1][1] = dE21_y - dV0_y;
  
  // Vertex neighbor vectors
  memset(pppdVN, 0, sizeof(pppdV));
  pppdVN[0][0][0] = dV0_x - dV00_x;
  pppdVN[0][0][1] = dV0_y - dV00_y;
  pppdVN[0][1][0] = dV0_x - dV01_x;
  pppdVN[0][1][1] = dV0_y - dV01_y;

  memset(ppdE, 0, sizeof(ppdE));
  ppdE[1][0] = dE11_x - dE10_x;
  ppdE[1][1] = dE11_y - dE10_y;
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;
  
  dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);
  dE1dotE2 = (ppdE[1][0]*ppdE[2][0]) + (ppdE[1][1]*ppdE[2][1]);
  dE1dotE10 = (ppdE[1][0]*pppdV[1][0][0]) + (ppdE[1][1]*pppdV[1][0][1]);
  dE1dotE11 = (ppdE[1][0]*pppdV[1][1][0]) + (ppdE[1][1]*pppdV[1][1][1]);
  dE2dotE20 = (ppdE[2][0]*pppdV[2][0][0]) + (ppdE[2][1]*pppdV[2][0][1]);
  dE2dotE21 = (ppdE[2][0]*pppdV[2][1][0]) + (ppdE[2][1]*pppdV[2][1][1]);

  dE10xE11 = (pppdV[1][0][0]*pppdV[1][1][1]) - (pppdV[1][0][1]*pppdV[1][1][0]);
  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);
  dE1xE20 = (ppdE[1][0]*pppdV[2][0][1]) - (ppdE[1][1]*pppdV[2][0][0]);
  dE2xE10 = (ppdE[2][0]*pppdV[1][0][1]) - (ppdE[2][1]*pppdV[1][0][0]);

  dV00dotV00 = ( (pppdVN[0][0][0] * pppdVN[0][0][0]) + (pppdVN[0][0][1] * pppdVN[0][0][1]) );
  dV01dotV01 = ( (pppdVN[0][1][0] * pppdVN[0][1][0]) + (pppdVN[0][1][1] * pppdVN[0][1][1]) );

  dE1dotV00 = ( (ppdE[1][0] * pppdVN[0][0][0]) + (ppdE[1][1] * pppdVN[0][0][1]) );
  dE1dotV01 = ( (ppdE[1][0] * pppdVN[0][1][0]) + (ppdE[1][1] * pppdVN[0][1][1]) );

  dE1xV00 = ( (ppdE[1][0] * pppdVN[0][0][1]) - (ppdE[1][1] * pppdVN[0][0][0]) );
  dE1xV01 = ( (ppdE[1][0] * pppdVN[0][1][1]) - (ppdE[1][1] * pppdVN[0][1][0]) );

  if ( fabs(dE1dotE1) <= s_cdEpsilonInscribedCircleCenterVEEAntiparallelNullEdge )
    {
      // Colinear
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;      
    }

  dE1E2minusE1dotE2 = sqrt(dE1dotE1*dE2dotE2) - dE1dotE2;
  dSinTheta = dE1E2minusE1dotE2/sqrt(dE1dotE1*dE2dotE2);
  if ( fabs(dSinTheta) < s_cdEpsilonInscribedCircleCenterVEEAntiparallelParallel )
    {
      // Colinear
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  //dSinTheta = dE10xE11/sqrt(dE10dotE10*dE11dotE11);
  dSinTheta = APM2D::SinTheta(dV0_x, dV0_y, dE10_x, dE10_y, dE11_x, dE11_y);
  dSinThetaTemp = APM2D::SinTheta(dE10_x, dE10_y, dE11_x, dE11_y, dV0_x, dV0_y);
  if ( dSinThetaTemp > dSinTheta )
    {
      dSinTheta = dSinThetaTemp;
    }
  if ( dSinTheta <= s_cdEpsilonInscribedCircleCenterVEEAntiparallelVertexOnEdge )
    {
      // V0 is on E1 and it is not one of the endpoints
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  //dSinTheta = dE20xE21/sqrt(dE20dotE20*dE21dotE21);
  dSinTheta = APM2D::SinTheta(dV0_x, dV0_y, dE20_x, dE20_y, dE21_x, dE21_y);
  dSinThetaTemp = APM2D::SinTheta(dE20_x, dE20_y, dE21_x, dE21_y, dV0_x, dV0_y);
  if ( dSinThetaTemp > dSinTheta )
    {
      dSinTheta = dSinThetaTemp;
    }
  if ( dSinTheta <= s_cdEpsilonInscribedCircleCenterVEEAntiparallelVertexOnEdge )
    {
      // V0 is on E2 and it is not one of the endpoints
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  dDen = 2*dE1dotE1;
  ASSERT( dDen != 0.0 );
  dDenR = 2*sqrt(dE1dotE1);
  ASSERT( dDenR != 0.0 );

  dDisc = dE1xE20*dE10xE11;
  if ( -s_cdEpsilonInscribedCircleCenterVEEAntiparallelDisc <= dDisc && dDisc <= 0.0 )
    {
      dDisc = 0.0;
    }
  if ( dDisc < 0.0 ) 
    {
      // Imaginary Solutions
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  dDiscE2 = dE2xE10*dE20xE21;
  if ( -s_cdEpsilonInscribedCircleCenterVEEAntiparallelDisc <= dDiscE2 && dDiscE2 <= 0.0 )
    {
      dDiscE2 = 0.0;
    }
  if ( dDiscE2 < 0.0 ) 
    {
      // Imaginary Solutions
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  

  //////////////////////////////////////////////////
  // Compute the center and radius

  pdC[0] = ( -(dE1xE20 - dE10xE11)*ppdE[1][1] + 2.0*ppdE[1][0]*sqrt(dDisc) ) / dDen;
  pdC[1] = ( (dE1xE20 - dE10xE11)*ppdE[1][0] + 2.0*ppdE[1][1]*sqrt(dDisc) ) / dDen;
  pdC[2] = sqrt((pdC[0]*pdC[0]) + (pdC[1]*pdC[1]));

  dR = ( dE1xE20 + dE10xE11 ) / dDenR;
  
  //////////////////////////////////////////////////
  // Check the Edge Zone

  dMagCE10 = ( (pdC[0] - pppdV[1][0][0])*(pdC[0] - pppdV[1][0][0]) + 
	       (pdC[1] - pppdV[1][0][1])*(pdC[1] - pppdV[1][0][1]) );
  dMagCE11 = ( (pdC[0] - pppdV[1][1][0])*(pdC[0] - pppdV[1][1][0]) + 
	       (pdC[1] - pppdV[1][1][1])*(pdC[1] - pppdV[1][1][1]) );
  dMagCE20 = ( (pdC[0] - pppdV[2][0][0])*(pdC[0] - pppdV[2][0][0]) + 
	       (pdC[1] - pppdV[2][0][1])*(pdC[1] - pppdV[2][0][1]) );
  dMagCE21 = ( (pdC[0] - pppdV[2][1][0])*(pdC[0] - pppdV[2][1][0]) + 
	       (pdC[1] - pppdV[2][1][1])*(pdC[1] - pppdV[2][1][1]) );

  if ( dMagCE10 <= s_cdEpsilonNullEdge )
    {
      ppdEdgeZone[0][0] = 0.0;
    }
  else
    {
      ppdEdgeZone[0][0] = -( ( dE1dotE10 - sqrt(dDisc) ) /
			     ( sqrt(dE1dotE1*dMagCE10) ) );
    }
  if ( dMagCE11 <= s_cdEpsilonNullEdge )
    {
      ppdEdgeZone[0][1] = 0.0;
    }
  else
    {
      ppdEdgeZone[0][1] = ( ( dE1dotE11 - sqrt(dDisc) ) /
			    ( sqrt(dE1dotE1*dMagCE11) ) );
    }
  if ( dMagCE20 <= s_cdEpsilonNullEdge )
    {
      ppdEdgeZone[1][0] = 0.0;
    }
  else
    {
      ppdEdgeZone[1][0] = -( ( dE2dotE20 + sqrt(dDiscE2) ) /
			     ( sqrt(dE2dotE2*dMagCE20) ) );
    }
  if ( dMagCE21 <= s_cdEpsilonNullEdge )
    {
      ppdEdgeZone[1][1] = 0.0;
    }
  else
    {
      ppdEdgeZone[1][1] = ( ( dE2dotE21 + sqrt(dDiscE2) ) /
			    ( sqrt(dE2dotE2*dMagCE21) ) );
    }
  
  for ( i = 0; i < 2; i++ )
    {
      for ( j = 0; j < 2; j++ )
	{
	  if ( ppdEdgeZone[i][j] < -s_cdEpsilonInscribedCircleCenterVEEAntiparallelEdgeZone )
	    {
#ifdef DEBUG_EPSILON_VALUES
	      if ( !( ppdEdgeZone[i][j] < -(s_cdEpsilonInscribedCircleCenterVEEAntiparallelEdgeZone*s_cdOrderOfMagnitude) ) )
		{
		  fprintf(stderr, "  VEE Antiparallel: Edge Zone: %e < %e < %e\n",
			  s_cdEpsilonInscribedCircleCenterVEEAntiparallelEdgeZone*s_cdOrderOfMagnitude,
			  ppdEdgeZone[i][j],
			  s_cdEpsilonInscribedCircleCenterVEEAntiparallelEdgeZone);
		}
#endif // DEBUG_EPSILON_VALUES	      
	      *pdC_x = 0.0;
	      *pdC_y = 0.0;
	      *pdR = 0.0;
	      return APM2D::icctNull;
	    }
	}
    }

  //////////////////////////////////////////////////
  // Check the Vertex Zone

  pdVertexZone[0] = ( ( ( dE1xE20 - dE10xE11 )*dE1xV00
			+ 2.0*dE1dotV00*sqrt(dDisc) ) / 
		      ( dDen*dR*sqrt(dV00dotV00) ) );
  pdVertexZone[1] = ( ( ( dE1xE20 - dE10xE11 )*dE1xV01
			+ 2.0*dE1dotV01*sqrt(dDisc) ) / 
		      ( dDen*dR*sqrt(dV01dotV01) ) );

  for ( i = 0; i < 2; i++ )
    {
      if ( pdVertexZone[i] < -s_cdEpsilonInscribedCircleCenterVEEAntiparallelVertexZone )
	{
#ifdef DEBUG_EPSILON_VALUES
	  if ( !( pdVertexZone[i] < -(s_cdEpsilonInscribedCircleCenterVEEAntiparallelVertexZone*s_cdOrderOfMagnitude) ) )
	    {
	      fprintf(stderr, "  VEE Antiparallel: Vertex Zone: %e < %e < %e\n",
		      s_cdEpsilonInscribedCircleCenterVEEAntiparallelVertexZone*s_cdOrderOfMagnitude,
		      pdVertexZone[i],
		      s_cdEpsilonInscribedCircleCenterVEEAntiparallelVertexZone);
	    }
#endif // DEBUG_EPSILON_VALUES	      
	  *pdC_x = 0.0;
	  *pdC_y = 0.0;
	  *pdR = 0.0;
	  return APM2D::icctNull;
	}
    }

  if ( !( fabs(pdC[2] - dR) < s_cdEpsilonInscribedCircleCenterVEERadius ) )
    {
      fprintf(stderr, "VEE Antiparallel: C = ( %f, %f, %f ) != %f = R\n", pdC[0], pdC[1], pdC[2], dR);
      //      ASSERT( FALSE );
    }

  // Translate the answer back out away from the origin
  *pdC_x = dV0_x + pdC[0];
  *pdC_y = dV0_y + pdC[1];
  *pdR = dR;
  return APM2D::icctVEEAntiparallel;
}

static APM2D::InscribedCircleCenterType 
InscribedCircleCenterVEEAntiparallelAdaptive(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					     DOUBLE dV00_x, DOUBLE dV00_y,
					     DOUBLE dV0_x, DOUBLE dV0_y,
					     DOUBLE dV01_x, DOUBLE dV01_y,
					     DOUBLE dE10_x, DOUBLE dE10_y, 
					     DOUBLE dE11_x, DOUBLE dE11_y,
					     DOUBLE dE20_x, DOUBLE dE20_y, 
					     DOUBLE dE21_x, DOUBLE dE21_y)
{
  return InscribedCircleCenterVEEAntiparallelLocal(pdC_x, pdC_y, pdR, 
						   dV00_x, dV00_y, 
						   dV0_x, dV0_y, 
						   dV01_x, dV01_y, 
						   dE10_x, dE10_y, dE11_x, dE11_y,
						   dE20_x, dE20_y, dE21_x, dE21_y);
}

APM2D::EdgeZoneType APM2D::EdgeZoneVEEAntiparallel(DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dE10_x, DOUBLE dE10_y, 
						   DOUBLE dE11_x, DOUBLE dE11_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   DOUBLE dE30_x, DOUBLE dE30_y,
						   DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::EdgeZoneType eResult;

  eResult = APM2D::eztMiddle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = EdgeZoneVEEAntiparallelLocal(dV0_x, dV0_y,
					       dE10_x, dE10_y, dE11_x, dE11_y,
					       dE20_x, dE20_y, dE21_x, dE21_y,
					       dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atExact:
      {
	eResult = EdgeZoneVEEAntiparallelAdaptive(dV0_x, dV0_y,
						  dE10_x, dE10_y, dE11_x, dE11_y,
						  dE20_x, dE20_y, dE21_x, dE21_y,
						  dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = EdgeZoneVEEAntiparallelMPFUN(dV0_x, dV0_y,
					       dE10_x, dE10_y, dE11_x, dE11_y,
					       dE20_x, dE20_y, dE21_x, dE21_y,
					       dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = EdgeZoneVEEAntiparallelMPFUNFilter(dV0_x, dV0_y,
					       dE10_x, dE10_y, dE11_x, dE11_y,
					       dE20_x, dE20_y, dE21_x, dE21_y,
					       dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = EdgeZoneVEEAntiparallelMPFUNNormalized(dV0_x, dV0_y,
							 dE10_x, dE10_y, dE11_x, dE11_y,
							 dE20_x, dE20_y, dE21_x, dE21_y,
							 dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::EdgeZoneType EdgeZoneVEEAntiparallelLocal(DOUBLE dV0_x, DOUBLE dV0_y,
							DOUBLE dE10_x, DOUBLE dE10_y, 
							DOUBLE dE11_x, DOUBLE dE11_y,
							DOUBLE dE20_x, DOUBLE dE20_y, 
							DOUBLE dE21_x, DOUBLE dE21_y,
							DOUBLE dE30_x, DOUBLE dE30_y,
							DOUBLE dE31_x, DOUBLE dE31_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE pdC[2];
  DOUBLE pdEdgeZone[2];
  DOUBLE dE1dotE1, dE1dotE3;
  DOUBLE dE3dotE30, dE3dotE31;
  DOUBLE dE1xE3, dE1xE20, dE10xE11;
  DOUBLE dDisc;
  DOUBLE dDen;
  DOUBLE dE3dotE3, dMagCE30, dMagCE31;
  APM2D::EdgeZoneType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = dE10_x - dV0_x;
  pppdV[1][0][1] = dE10_y - dV0_y;
  pppdV[1][1][0] = dE11_x - dV0_x;
  pppdV[1][1][1] = dE11_y - dV0_y;
  pppdV[2][0][0] = dE20_x - dV0_x;
  pppdV[2][0][1] = dE20_y - dV0_y;
  pppdV[2][1][0] = dE21_x - dV0_x;
  pppdV[2][1][1] = dE21_y - dV0_y;
  pppdV[3][0][0] = dE30_x - dV0_x;
  pppdV[3][0][1] = dE30_y - dV0_y;
  pppdV[3][1][0] = dE31_x - dV0_x;
  pppdV[3][1][1] = dE31_y - dV0_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[1][0] = dE11_x - dE10_x;
  ppdE[1][1] = dE11_y - dE10_y;
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;
  ppdE[3][0] = dE31_x - dE30_x;
  ppdE[3][1] = dE31_y - dE30_y;
  
  dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
  dE1dotE3 = (ppdE[1][0]*ppdE[3][0]) + (ppdE[1][1]*ppdE[3][1]);
  dE3dotE3 = (ppdE[3][0]*ppdE[3][0]) + (ppdE[3][1]*ppdE[3][1]);

  dE3dotE30 = (ppdE[3][0]*pppdV[3][0][0]) + (ppdE[3][1]*pppdV[3][0][1]);
  dE3dotE31 = (ppdE[3][0]*pppdV[3][1][0]) + (ppdE[3][1]*pppdV[3][1][1]);
  
  dE1xE3 = (ppdE[1][0]*ppdE[3][1]) - (ppdE[1][1]*ppdE[3][0]);
  dE1xE20 = (ppdE[1][0]*pppdV[2][0][1]) - (ppdE[1][1]*pppdV[2][0][0]);
  dE10xE11 = (pppdV[1][0][0]*pppdV[1][1][1]) - (pppdV[1][0][1]*pppdV[1][1][0]);

  dDisc = dE1xE20*dE10xE11;
  if ( -s_cdEpsilonInscribedCircleCenterVEEAntiparallelDisc <= dDisc && dDisc <= 0.0 )
    {
      dDisc = 0.0;
    }
  ASSERT( dDisc >= 0.0 ); 

  dDen = 2.0*dE1dotE1;
  ASSERT( dDen != 0.0 );

  pdC[0] = ( -(dE1xE20 - dE10xE11)*ppdE[1][1] + 2.0*ppdE[1][0]*sqrt(dDisc) ) / dDen;
  pdC[1] = ( (dE1xE20 - dE10xE11)*ppdE[1][0] + 2.0*ppdE[1][1]*sqrt(dDisc) ) / dDen;

  dMagCE30 = ( (pdC[0] - pppdV[3][0][0])*(pdC[0] - pppdV[3][0][0]) + 
	       (pdC[1] - pppdV[3][0][1])*(pdC[1] - pppdV[3][0][1]) );
  dMagCE31 = ( (pdC[0] - pppdV[3][1][0])*(pdC[0] - pppdV[3][1][0]) + 
	       (pdC[1] - pppdV[3][1][1])*(pdC[1] - pppdV[3][1][1]) );

  if ( dMagCE30 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[0] = 0.0;
    }
  else
    {
      pdEdgeZone[0] = ( ( 2.0*dE1dotE1*dE3dotE30 
			  - ( dE1xE20 - dE10xE11 )*dE1xE3
			  - 2.0*dE1dotE3*sqrt(dDisc) ) /
			( dDen*sqrt(dE3dotE3*dMagCE30) ) );
    }
  if ( dMagCE31 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[1] = 0.0;
    }
  else
    {
      pdEdgeZone[1] = -( ( 2.0*dE1dotE1*dE3dotE31 
			   - ( dE1xE20 - dE10xE11 )*dE1xE3
			   - 2.0*dE1dotE3*sqrt(dDisc) ) /
			 ( dDen*sqrt(dE3dotE3*dMagCE31) ) );
    }
		     
  if ( !( pdEdgeZone[0] < -s_cdEpsilonEdgeZoneVEEAntiparallelEdgeZone ) )
    {
      eResult = APM2D::eztStart;
    }
  else if ( !( pdEdgeZone[1] < -s_cdEpsilonEdgeZoneVEEAntiparallelEdgeZone ) )
    {
      eResult = APM2D::eztEnd;
    }
  else
    {
      eResult = APM2D::eztMiddle;
    }

  return eResult;
}

static APM2D::EdgeZoneType EdgeZoneVEEAntiparallelAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
							   DOUBLE dE10_x, DOUBLE dE10_y, 
							   DOUBLE dE11_x, DOUBLE dE11_y,
							   DOUBLE dE20_x, DOUBLE dE20_y, 
							   DOUBLE dE21_x, DOUBLE dE21_y,
							   DOUBLE dE30_x, DOUBLE dE30_y,
							   DOUBLE dE31_x, DOUBLE dE31_y)
{
  return EdgeZoneVEEAntiparallelLocal(dV0_x, dV0_y,
				      dE10_x, dE10_y, dE11_x, dE11_y,
				      dE20_x, dE20_y, dE21_x, dE21_y,
				      dE30_x, dE30_y, dE31_x, dE31_y);
}

APM2D::CircleTestType APM2D::InCircleVEE_VAntiparallel(DOUBLE dV0_x, DOUBLE dV0_y,
						       DOUBLE dE10_x, DOUBLE dE10_y, 
						       DOUBLE dE11_x, DOUBLE dE11_y,
						       DOUBLE dE20_x, DOUBLE dE20_y, 
						       DOUBLE dE21_x, DOUBLE dE21_y,
						       DOUBLE dV3_x, DOUBLE dV3_y)
{
  APM2D::CircleTestType eResult;

  eResult = APM2D::cttOnCircle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InCircleVEE_VAntiparallelLocal(dV0_x, dV0_y,
						 dE10_x, dE10_y, dE11_x, dE11_y,
						 dE20_x, dE20_y, dE21_x, dE21_y,
						 dV3_x, dV3_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InCircleVEE_VAntiparallelAdaptive(dV0_x, dV0_y,
						    dE10_x, dE10_y, dE11_x, dE11_y,
						    dE20_x, dE20_y, dE21_x, dE21_y,
						    dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InCircleVEE_VAntiparallelMPFUN(dV0_x, dV0_y,
						 dE10_x, dE10_y, dE11_x, dE11_y,
						 dE20_x, dE20_y, dE21_x, dE21_y,
						 dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = InCircleVEE_VAntiparallelMPFUNFilter(dV0_x, dV0_y,
						 dE10_x, dE10_y, dE11_x, dE11_y,
						 dE20_x, dE20_y, dE21_x, dE21_y,
						 dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InCircleVEE_VAntiparallelMPFUNNormalized(dV0_x, dV0_y,
							   dE10_x, dE10_y, dE11_x, dE11_y,
							   dE20_x, dE20_y, dE21_x, dE21_y,
							   dV3_x, dV3_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::CircleTestType InCircleVEE_VAntiparallelLocal(DOUBLE dV0_x, DOUBLE dV0_y,
							    DOUBLE dE10_x, DOUBLE dE10_y, 
							    DOUBLE dE11_x, DOUBLE dE11_y,
							    DOUBLE dE20_x, DOUBLE dE20_y, 
							    DOUBLE dE21_x, DOUBLE dE21_y,
							    DOUBLE dV3_x, DOUBLE dV3_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE dE1dotE1, dV3dotV3, dE1dotV3;
  DOUBLE dE1xE20, dE1xV3, dE10xE11;
  DOUBLE dDisc;
  DOUBLE dDen;
  DOUBLE dResult;
  APM2D::CircleTestType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = dE10_x - dV0_x;
  pppdV[1][0][1] = dE10_y - dV0_y;
  pppdV[1][1][0] = dE11_x - dV0_x;
  pppdV[1][1][1] = dE11_y - dV0_y;
  pppdV[2][0][0] = dE20_x - dV0_x;
  pppdV[2][0][1] = dE20_y - dV0_y;
  pppdV[2][1][0] = dE21_x - dV0_x;
  pppdV[2][1][1] = dE21_y - dV0_y;
  pppdV[3][0][0] = dV3_x - dV0_x;
  pppdV[3][0][1] = dV3_y - dV0_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[1][0] = dE11_x - dE10_x;
  ppdE[1][1] = dE11_y - dE10_y;
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;

  dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
  dV3dotV3 = (pppdV[3][0][0]*pppdV[3][0][0]) + (pppdV[3][0][1]*pppdV[3][0][1]);
  dE1dotV3 = (ppdE[1][0]*pppdV[3][0][0]) + (ppdE[1][1]*pppdV[3][0][1]);

  dE1xV3 = (ppdE[1][0]*pppdV[3][0][1]) - (ppdE[1][1]*pppdV[3][0][0]);
  dE1xE20 = (ppdE[1][0]*pppdV[2][0][1]) - (ppdE[1][1]*pppdV[2][0][0]);
  dE10xE11 = (pppdV[1][0][0]*pppdV[1][1][1]) - (pppdV[1][0][1]*pppdV[1][1][0]);

  dDisc = dE1xE20*dE10xE11;
  if ( -s_cdEpsilonInscribedCircleCenterVEEAntiparallelDisc <= dDisc && dDisc <= 0.0 )
    {
      dDisc = 0.0;
    }
  ASSERT( dDisc >= 0.0 ); 

  dDen = dE1dotE1;
  ASSERT( dDen != 0.0 );

  dResult = ( dE1dotE1*dV3dotV3 
	      - ( dE1xE20 - dE10xE11 )*dE1xV3
	      - 2.0*dE1dotV3*sqrt(dDisc) ) / dDen;
  
#ifdef ENABLE_DEBUGGING_OUTPUT
  fprintf(stderr, "VEE_V Anti: dResult = %e\n", dResult);
#endif // ENABLE_DEBUGGING_OUTPUT

  if ( dResult < -s_cdEpsilonInCircleVEE_VAntiparallelOnCircle )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( dResult > s_cdEpsilonInCircleVEE_VAntiparallelOnCircle )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

static APM2D::CircleTestType InCircleVEE_VAntiparallelAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
							       DOUBLE dE10_x, DOUBLE dE10_y,
							       DOUBLE dE11_x, DOUBLE dE11_y,
							       DOUBLE dE20_x, DOUBLE dE20_y,
							       DOUBLE dE21_x, DOUBLE dE21_y,
							       DOUBLE dV3_x, DOUBLE dV3_y)
{
  return InCircleVEE_VAntiparallelLocal(dV0_x, dV0_y,
					dE10_x, dE10_y, dE11_x, dE11_y,
					dE20_x, dE20_y, dE21_x, dE21_y,
					dV3_x, dV3_y);
}

APM2D::CircleTestType APM2D::InCircleVEE_EAntiparallel(DOUBLE dV0_x, DOUBLE dV0_y,
						       DOUBLE dE10_x, DOUBLE dE10_y, 
						       DOUBLE dE11_x, DOUBLE dE11_y,
						       DOUBLE dE20_x, DOUBLE dE20_y, 
						       DOUBLE dE21_x, DOUBLE dE21_y,
						       DOUBLE dE30_x, DOUBLE dE30_y,
						       DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::CircleTestType eResult;

  eResult = APM2D::cttOnCircle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InCircleVEE_EAntiparallelLocal(dV0_x, dV0_y,
						 dE10_x, dE10_y, dE11_x, dE11_y,
						 dE20_x, dE20_y, dE21_x, dE21_y,
						 dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InCircleVEE_EAntiparallelAdaptive(dV0_x, dV0_y,
						    dE10_x, dE10_y, dE11_x, dE11_y,
						    dE20_x, dE20_y, dE21_x, dE21_y,
						    dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InCircleVEE_EAntiparallelMPFUN(dV0_x, dV0_y,
						 dE10_x, dE10_y, dE11_x, dE11_y,
						 dE20_x, dE20_y, dE21_x, dE21_y,
						 dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = InCircleVEE_EAntiparallelMPFUNFilter(dV0_x, dV0_y,
						 dE10_x, dE10_y, dE11_x, dE11_y,
						 dE20_x, dE20_y, dE21_x, dE21_y,
						 dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InCircleVEE_EAntiparallelMPFUNNormalized(dV0_x, dV0_y,
							   dE10_x, dE10_y, dE11_x, dE11_y,
							   dE20_x, dE20_y, dE21_x, dE21_y,
							   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::CircleTestType InCircleVEE_EAntiparallelLocal(DOUBLE dV0_x, DOUBLE dV0_y,
							    DOUBLE dE10_x, DOUBLE dE10_y, 
							    DOUBLE dE11_x, DOUBLE dE11_y,
							    DOUBLE dE20_x, DOUBLE dE20_y, 
							    DOUBLE dE21_x, DOUBLE dE21_y,
							    DOUBLE dE30_x, DOUBLE dE30_y,
							    DOUBLE dE31_x, DOUBLE dE31_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE dE1dotE1, dE2dotE2, dE1dotE2, dE1dotE3;
  DOUBLE dE1xE20, dE1xE3, dE10xE11, dE20xE21, dE30xE31;
  DOUBLE dE1E2minusE1dotE2, dE1xE20minusE10xE11; 
  DOUBLE dDisc;
  DOUBLE dDen;
  DOUBLE dResult;
  APM2D::CircleTestType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = dE10_x - dV0_x;
  pppdV[1][0][1] = dE10_y - dV0_y;
  pppdV[1][1][0] = dE11_x - dV0_x;
  pppdV[1][1][1] = dE11_y - dV0_y;
  pppdV[2][0][0] = dE20_x - dV0_x;
  pppdV[2][0][1] = dE20_y - dV0_y;
  pppdV[2][1][0] = dE21_x - dV0_x;
  pppdV[2][1][1] = dE21_y - dV0_y;
  pppdV[3][0][0] = dE30_x - dV0_x;
  pppdV[3][0][1] = dE30_y - dV0_y;
  pppdV[3][1][0] = dE31_x - dV0_x;
  pppdV[3][1][1] = dE31_y - dV0_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[1][0] = dE11_x - dE10_x;
  ppdE[1][1] = dE11_y - dE10_y;
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;
  ppdE[3][0] = dE31_x - dE30_x;
  ppdE[3][1] = dE31_y - dE30_y;

  dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);
  dE1dotE2 = (ppdE[1][0]*ppdE[2][0]) + (ppdE[1][1]*ppdE[2][1]);
  dE1dotE3 = (ppdE[1][0]*ppdE[3][0]) + (ppdE[1][1]*ppdE[3][1]);

  dE1xE20 = (ppdE[1][0]*pppdV[2][0][1]) - (ppdE[1][1]*pppdV[2][0][0]);
  dE1xE3 = (ppdE[1][0]*ppdE[3][1]) - (ppdE[1][1]*ppdE[3][0]);
  dE10xE11 = (pppdV[1][0][0]*pppdV[1][1][1]) - (pppdV[1][0][1]*pppdV[1][1][0]);
  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);
  dE30xE31 = (pppdV[3][0][0]*pppdV[3][1][1]) - (pppdV[3][0][1]*pppdV[3][1][0]);

  dE1E2minusE1dotE2 = ( sqrt(dE1dotE1*dE2dotE2) - dE1dotE2 );
  dE1xE20minusE10xE11 = dE1xE20 - dE10xE11;

  dDisc = dE1xE20*dE10xE11;
  if ( -s_cdEpsilonInscribedCircleCenterVEEAntiparallelDisc <= dDisc && dDisc <= 0.0 )
    {
      dDisc = 0.0;
    }
  ASSERT( dDisc >= 0.0 ); 

  dDen = 4.0*dE1dotE1*dE1dotE1*dE1E2minusE1dotE2*dE1E2minusE1dotE2;
  ASSERT( dDen != 0.0 );

  dResult = ( 4.0*dE1dotE1*dE30xE31*( dE1dotE1*dE30xE31 + dE1dotE3*dE1xE20minusE10xE11 )
	      - 4.0*dE1dotE3*dE1dotE3*dE1xE20*dE20xE21
	      - dE1xE3*dE1xE3*dE1xE20minusE10xE11*dE1xE20minusE10xE11
	      + 4.0*dE1xE3*( 2.0*dE1dotE1*dE30xE31 + dE1dotE3*dE1xE20minusE10xE11 )*sqrt(dDisc) ) / dDen;
  
  if ( dResult < -s_cdEpsilonInCircleVEE_EAntiparallelOnCircle )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( dResult > s_cdEpsilonInCircleVEE_EAntiparallelOnCircle )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

static APM2D::CircleTestType InCircleVEE_EAntiparallelAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
							       DOUBLE dE10_x, DOUBLE dE10_y,
							       DOUBLE dE11_x, DOUBLE dE11_y,
							       DOUBLE dE20_x, DOUBLE dE20_y,
							       DOUBLE dE21_x, DOUBLE dE21_y,
							       DOUBLE dE30_x, DOUBLE dE30_y,
							       DOUBLE dE31_x, DOUBLE dE31_y)
{
  return InCircleVEE_EAntiparallelLocal(dV0_x, dV0_y,
					dE10_x, dE10_y, dE11_x, dE11_y,
					dE20_x, dE20_y, dE21_x, dE21_y,
					dE30_x, dE30_y, dE31_x, dE31_y);
}

APM2D::InscribedCircleCenterType 
APM2D::InscribedCircleCentervEE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
				DOUBLE dE11_x, DOUBLE dE11_y,
				DOUBLE dE20_x, DOUBLE dE20_y, 
				DOUBLE dE21_x, DOUBLE dE21_y)
{
  ASSERT( pdC_x != NULL && pdC_y != NULL && pdR != NULL );

  APM2D::InscribedCircleCenterType eResult;

  eResult = APM2D::icctNull;
  
  //////////////////////////////////////////////////
  // Degenerate Touching Check
  if ( ( dV0_E10_x == dE20_x && dV0_E10_y == dE20_y ) ||
       ( dV0_E10_x == dE21_x && dV0_E10_y == dE21_y ) )
    {
      *pdC_x = dV0_E10_x;
      *pdC_y = dV0_E10_y;
      *pdR = 0.0;
      return APM2D::icctvEE;
    }

  //////////////////////////////////////////////////
  // CCW Checks

  if ( !( Counterclockwise(dV0_E10_x, dV0_E10_y, dE11_x, dE11_y, dE20_x, dE20_y) ||
	  Counterclockwise(dV0_E10_x, dV0_E10_y, dE11_x, dE11_y, dE21_x, dE21_y) ) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  if ( !( Counterclockwise(dE20_x, dE20_y, dE21_x, dE21_y, dV0_E10_x, dV0_E10_y) ||
	  Counterclockwise(dE20_x, dE20_y, dE21_x, dE21_y, dE11_x, dE11_y) ) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InscribedCircleCentervEELocal(pdC_x, pdC_y, pdR, 
						dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
						dE20_x, dE20_y, dE21_x, dE21_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InscribedCircleCentervEEAdaptive(pdC_x, pdC_y, pdR, 
						   dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
						   dE20_x, dE20_y, dE21_x, dE21_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InscribedCircleCentervEEMPFUN(pdC_x, pdC_y, pdR, 
						dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
						dE20_x, dE20_y, dE21_x, dE21_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = InscribedCircleCentervEEMPFUNFilter(pdC_x, pdC_y, pdR, 
						dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
						dE20_x, dE20_y, dE21_x, dE21_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InscribedCircleCentervEEMPFUNNormalized(pdC_x, pdC_y, pdR, 
							  dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
							  dE20_x, dE20_y, dE21_x, dE21_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }

  return eResult;
}

APM2D::InscribedCircleCenterType 
APM2D::InscribedCircleCentervEvE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				 DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
				 DOUBLE dE11_E20_x, DOUBLE dE11_E20_y,
				 DOUBLE dE21_x, DOUBLE dE21_y)
{
  ASSERT( pdC_x != NULL && pdC_y != NULL && pdR != NULL );

  APM2D::InscribedCircleCenterType eResult;

  eResult = APM2D::icctNull;
  
  //////////////////////////////////////////////////
  // Degenerate Touching Check
  if ( ( dV0_E10_x == dE11_E20_x && dV0_E10_y == dE11_E20_y ) ||
       ( dV0_E10_x == dE21_x && dV0_E10_y == dE21_y ) )
    {
      ASSERT( FALSE );
      *pdC_x = dV0_E10_x;
      *pdC_y = dV0_E10_y;
      *pdR = 0.0;
      return APM2D::icctvEE;
    }

  //////////////////////////////////////////////////
  // CCW Checks

  if ( !Counterclockwise(dV0_E10_x, dV0_E10_y, dE11_E20_x, dE11_E20_y, dE21_x, dE21_y) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InscribedCircleCentervEELocal(pdC_x, pdC_y, pdR, 
						dV0_E10_x, dV0_E10_y, 
						dE11_E20_x, dE11_E20_y,
						dE11_E20_x, dE11_E20_y,
						dE21_x, dE21_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InscribedCircleCentervEEAdaptive(pdC_x, pdC_y, pdR, 
						   dV0_E10_x, dV0_E10_y,
						   dE11_E20_x, dE11_E20_y,
						   dE11_E20_x, dE11_E20_y,
						   dE21_x, dE21_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InscribedCircleCentervEEMPFUN(pdC_x, pdC_y, pdR, 
						dV0_E10_x, dV0_E10_y,
						dE11_E20_x, dE11_E20_y,
						dE11_E20_x, dE11_E20_y,
						dE21_x, dE21_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = InscribedCircleCentervEEMPFUNFilter(pdC_x, pdC_y, pdR, 
						dV0_E10_x, dV0_E10_y,
						dE11_E20_x, dE11_E20_y,
						dE11_E20_x, dE11_E20_y,
						dE21_x, dE21_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InscribedCircleCentervEEMPFUNNormalized(pdC_x, pdC_y, pdR, 
							  dV0_E10_x, dV0_E10_y,
							  dE11_E20_x, dE11_E20_y,
							  dE11_E20_x, dE11_E20_y,
							  dE21_x, dE21_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }

  return eResult;
}

static APM2D::InscribedCircleCenterType 
InscribedCircleCentervEELocal(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			      DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
			      DOUBLE dE11_x, DOUBLE dE11_y,
			      DOUBLE dE20_x, DOUBLE dE20_y, 
			      DOUBLE dE21_x, DOUBLE dE21_y)
{
  DOUBLE pdC[3], dR;
  DOUBLE pppdV[3][2][2]; // Endpoints
  DOUBLE ppdE[3][2];     // Edge vectors
  DOUBLE pdEdgeZone[2];
  DOUBLE dDen;
  DOUBLE dE1dotE1, dE2dotE2, dE1dotE2;
  DOUBLE dE2dotE20, dE2dotE21;
  DOUBLE dE1xE2, dE20xE21;
  DOUBLE dE1E2minusE1dotE2, dSinTheta, dSinThetaTemp;
  DOUBLE dMagCE20, dMagCE21;
  UINT_32 i;
  
  // Translate V1 to the origin
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = 0.0;
  pppdV[1][0][1] = 0.0;
  pppdV[1][1][0] = dE11_x - dV0_E10_x;
  pppdV[1][1][1] = dE11_y - dV0_E10_y;
  pppdV[2][0][0] = dE20_x - dV0_E10_x;
  pppdV[2][0][1] = dE20_y - dV0_E10_y;
  pppdV[2][1][0] = dE21_x - dV0_E10_x;
  pppdV[2][1][1] = dE21_y - dV0_E10_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[1][0] = dE11_x - dV0_E10_x;
  ppdE[1][1] = dE11_y - dV0_E10_y;
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;
  
  dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);
  dE1dotE2 = (ppdE[1][0]*ppdE[2][0]) + (ppdE[1][1]*ppdE[2][1]);
  dE2dotE20 = (ppdE[2][0]*pppdV[2][0][0]) + (ppdE[2][1]*pppdV[2][0][1]);
  dE2dotE21 = (ppdE[2][0]*pppdV[2][1][0]) + (ppdE[2][1]*pppdV[2][1][1]);

  dE1xE2 = (ppdE[1][0]*ppdE[2][1]) - (ppdE[1][1]*ppdE[2][0]);
  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);

  dE1E2minusE1dotE2 = sqrt(dE1dotE1*dE2dotE2) - dE1dotE2;

  dSinTheta = dE1xE2/sqrt(dE1dotE1*dE2dotE2);
  if ( fabs(dSinTheta) < s_cdEpsilonInscribedCircleCentervEEParallel &&
       dE1dotE2 > 0.0 )
    {
#ifdef ENABLE_DEBUGGING_OUTPUT
      fprintf(stderr, "dE1E2minusE1dotE2 (%e) < vEEParallel (%e)\n", dE1E2minusE1dotE2, s_cdEpsilonInscribedCircleCentervEEParallel); 
#endif // ENABLE_DEBUGGING_OUTPUT

      // Colinear
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  dSinTheta = APM2D::SinTheta(dV0_E10_x, dV0_E10_y, dE20_x, dE20_y, dE21_x, dE21_y);
  dSinThetaTemp = APM2D::SinTheta(dE20_x, dE20_y, dE21_x, dE21_y, dV0_E10_x, dV0_E10_y);
  if ( dSinThetaTemp > dSinTheta )
    {
      dSinTheta = dSinThetaTemp;
    }
  if ( dSinTheta <= s_cdEpsilonInscribedCircleCentervEEVertexOnEdge )
    {
      // V0 is on E2 and it is not one of the endpoints
      
#ifdef ENABLE_DEBUGGING_OUTPUT
      fprintf(stderr, "    vEE: V0 is on E2 ( %f ) <= %f\n", 
	      dE20xE21, s_cdEpsilonInscribedCircleCentervEEVertexOnEdge);
      fprintf(stderr, "      V0: ( %.5f, %.5f )\n", dV0_E10_x, dV0_E10_y);
      fprintf(stderr, "      E1: ( %.5f, %.5f ) => ( %.5f, %.5f )\n", 
	      dV0_E10_x, dV0_E10_y, dE11_x, dE11_y);
      fprintf(stderr, "      E2: ( %.5f, %.5f ) => ( %.5f, %.5f )\n", 
	      dE20_x, dE20_y, dE21_x, dE21_y);
#endif // ENABLE_DEBUGGING_OUTPUT

      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  dDen = dE1E2minusE1dotE2;
  ASSERT( dDen != 0.0 );

  //////////////////////////////////////////////////
  // Compute the center and radius

  pdC[0] = -( dE20xE21*ppdE[1][1] ) / dDen;
  pdC[1] = ( dE20xE21*ppdE[1][0] ) / dDen;
  pdC[2] = sqrt((pdC[0]*pdC[0]) + (pdC[1]*pdC[1]));

  dR = ( sqrt(dE1dotE1)*dE20xE21 ) / dDen;
  
  //////////////////////////////////////////////////
  // Check the Edge Zone

  dMagCE20 = ( (pdC[0] - pppdV[2][0][0])*(pdC[0] - pppdV[2][0][0]) + 
	       (pdC[1] - pppdV[2][0][1])*(pdC[1] - pppdV[2][0][1]) );
  dMagCE21 = ( (pdC[0] - pppdV[2][1][0])*(pdC[0] - pppdV[2][1][0]) + 
	       (pdC[1] - pppdV[2][1][1])*(pdC[1] - pppdV[2][1][1]) );

  if ( dMagCE20 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[0] = 0.0;
    }
  else
    {
      pdEdgeZone[0] = -( ( dE1E2minusE1dotE2*dE2dotE20 - dE20xE21*dE1xE2 ) / 
			 ( dE1E2minusE1dotE2*sqrt(dE2dotE2*dMagCE20) ) );
    }
  if ( dMagCE21 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[1] = 0.0;
    }
  else
    {
      pdEdgeZone[1] = ( ( dE1E2minusE1dotE2*dE2dotE21 - dE20xE21*dE1xE2 ) / 
			( dE1E2minusE1dotE2*sqrt(dE2dotE2*dMagCE21) ) );
    }
  
  for ( i = 0; i < 2; i++ )
    {
      if ( pdEdgeZone[i] < -s_cdEpsilonInscribedCircleCentervEEEdgeZone )
	{
#ifdef DEBUG_EPSILON_VALUES
	  if ( !( pdEdgeZone[i] < -(s_cdEpsilonInscribedCircleCentervEEEdgeZone*s_cdOrderOfMagnitude) ) )
	    {
	      fprintf(stderr, "  vEE: Edge Zone: %e < %e < %e\n",
		      s_cdEpsilonInscribedCircleCentervEEEdgeZone*s_cdOrderOfMagnitude,
		      pdEdgeZone[i],
		      s_cdEpsilonInscribedCircleCentervEEEdgeZone);
	    }
#endif // DEBUG_EPSILON_VALUES	      
	  *pdC_x = 0.0;
	  *pdC_y = 0.0;
	  *pdR = 0.0;
	  return APM2D::icctNull;
	}
    }

  if ( !( fabs(pdC[2] - dR) < s_cdEpsilonInscribedCircleCentervEERadius ) )
    {
      fprintf(stderr, "vEE: C = ( %f, %f, %f ) != %f = R\n", pdC[0], pdC[1], pdC[2], dR);
      //ASSERT( FALSE );
    }

  // Translate the answer back out away from the origin
  *pdC_x = dV0_E10_x + pdC[0];
  *pdC_y = dV0_E10_y + pdC[1];
  *pdR = dR;
  return APM2D::icctvEE;
}


static APM2D::InscribedCircleCenterType 
InscribedCircleCentervEEAdaptive(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				 DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
				 DOUBLE dE11_x, DOUBLE dE11_y,
				 DOUBLE dE20_x, DOUBLE dE20_y, 
				 DOUBLE dE21_x, DOUBLE dE21_y)
{
  return InscribedCircleCentervEELocal(pdC_x, pdC_y, pdR, 
				       dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
				       dE20_x, dE20_y, dE21_x, dE21_y);
}

APM2D::EdgeZoneType APM2D::EdgeZonevEE(DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
				       DOUBLE dE11_x, DOUBLE dE11_y,
				       DOUBLE dE20_x, DOUBLE dE20_y, 
				       DOUBLE dE21_x, DOUBLE dE21_y,
				       DOUBLE dE30_x, DOUBLE dE30_y,
				       DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::EdgeZoneType eResult;

  //////////////////////////////////////////////////
  // Degenerate Touching Check
  if ( ( dV0_E10_x == dE20_x && dV0_E10_y == dE20_y ) ||
       ( dV0_E10_x == dE21_x && dV0_E10_y == dE21_y ) )
    {
      return EdgeZone(dV0_E10_x, dV0_E10_y, 
		      dE30_x, dE30_y,
		      dE31_x, dE31_y);
    }

  eResult = APM2D::eztMiddle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = EdgeZonevEELocal(dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
				   dE20_x, dE20_y, dE21_x, dE21_y,
				   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atExact:
      {
	eResult = EdgeZonevEEAdaptive(dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
				      dE20_x, dE20_y, dE21_x, dE21_y,
				      dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = EdgeZonevEEMPFUN(dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
				   dE20_x, dE20_y, dE21_x, dE21_y,
				   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = EdgeZonevEEMPFUNFilter(dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
				   dE20_x, dE20_y, dE21_x, dE21_y,
				   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = EdgeZonevEEMPFUNNormalized(dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
					     dE20_x, dE20_y, dE21_x, dE21_y,
					     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::EdgeZoneType EdgeZonevEELocal(DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
					    DOUBLE dE11_x, DOUBLE dE11_y,
					    DOUBLE dE20_x, DOUBLE dE20_y, 
					    DOUBLE dE21_x, DOUBLE dE21_y,
					    DOUBLE dE30_x, DOUBLE dE30_y,
					    DOUBLE dE31_x, DOUBLE dE31_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE pdC[2];
  DOUBLE pdEdgeZone[2];
  DOUBLE dE1dotE1, dE2dotE2, dE1dotE2;
  DOUBLE dE3dotE3, dE3dotE30, dE3dotE31;
  DOUBLE dE1xE3, dE20xE21;
  DOUBLE dE1E2minusE1dotE2;
  DOUBLE dDen;
  DOUBLE dMagCE30, dMagCE31;
  APM2D::EdgeZoneType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = 0.0;
  pppdV[1][0][1] = 0.0;
  pppdV[1][1][0] = dE11_x - dV0_E10_x;
  pppdV[1][1][1] = dE11_y - dV0_E10_y;
  pppdV[2][0][0] = dE20_x - dV0_E10_x;
  pppdV[2][0][1] = dE20_y - dV0_E10_y;
  pppdV[2][1][0] = dE21_x - dV0_E10_x;
  pppdV[2][1][1] = dE21_y - dV0_E10_y;
  pppdV[3][0][0] = dE30_x - dV0_E10_x;
  pppdV[3][0][1] = dE30_y - dV0_E10_y;
  pppdV[3][1][0] = dE31_x - dV0_E10_x;
  pppdV[3][1][1] = dE31_y - dV0_E10_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[1][0] = dE11_x - dV0_E10_x;
  ppdE[1][1] = dE11_y - dV0_E10_y;
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;
  ppdE[3][0] = dE31_x - dE30_x;
  ppdE[3][1] = dE31_y - dE30_y;
  
  dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);
  dE1dotE2 = (ppdE[1][0]*ppdE[2][0]) + (ppdE[1][1]*ppdE[2][1]);
  dE3dotE3 = (ppdE[3][0]*ppdE[3][0]) + (ppdE[3][1]*ppdE[3][1]);
  dE3dotE30 = (ppdE[3][0]*pppdV[3][0][0]) + (ppdE[3][1]*pppdV[3][0][1]);
  dE3dotE31 = (ppdE[3][0]*pppdV[3][1][0]) + (ppdE[3][1]*pppdV[3][1][1]);

  dE1xE3 = (ppdE[1][0]*ppdE[3][1]) - (ppdE[1][1]*ppdE[3][0]);
  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);

  dE1E2minusE1dotE2 = sqrt(dE1dotE1*dE2dotE2) - dE1dotE2;

  dDen = dE1E2minusE1dotE2;
  ASSERT( dDen != 0.0 );

  pdC[0] = -( dE20xE21*ppdE[1][1] ) / dDen;
  pdC[1] = ( dE20xE21*ppdE[1][0] ) / dDen;

  dMagCE30 = ( (pdC[0] - pppdV[3][0][0])*(pdC[0] - pppdV[3][0][0]) + 
	       (pdC[1] - pppdV[3][0][1])*(pdC[1] - pppdV[3][0][1]) );
  dMagCE31 = ( (pdC[0] - pppdV[3][1][0])*(pdC[0] - pppdV[3][1][0]) + 
	       (pdC[1] - pppdV[3][1][1])*(pdC[1] - pppdV[3][1][1]) );

  if ( dMagCE30 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[0] = 0.0;
    }
  else
    {
      pdEdgeZone[0] = ( ( dE1E2minusE1dotE2*dE3dotE30 - dE20xE21*dE1xE3 ) / 
			( dDen*sqrt(dE3dotE3*dMagCE30) ) );
    }
  if ( dMagCE31 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[1] = 0.0;
    }
  else
    {
      pdEdgeZone[1] = -( ( dE1E2minusE1dotE2*dE3dotE31 - dE20xE21*dE1xE3 ) / 
			 ( dDen*sqrt(dE3dotE3*dMagCE31) ) );
    }
  
  if ( !( pdEdgeZone[0] < -s_cdEpsilonEdgeZonevEEEdgeZone ) )
    {
      eResult = APM2D::eztStart;
    }
  else if ( !( pdEdgeZone[1] < -s_cdEpsilonEdgeZonevEEEdgeZone ) )
    {
      eResult = APM2D::eztEnd;
    }
  else
    {
      eResult = APM2D::eztMiddle;
    }

  return eResult;
}

static APM2D::EdgeZoneType EdgeZonevEEAdaptive(DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
					       DOUBLE dE11_x, DOUBLE dE11_y,
					       DOUBLE dE20_x, DOUBLE dE20_y, 
					       DOUBLE dE21_x, DOUBLE dE21_y,
					       DOUBLE dE30_x, DOUBLE dE30_y,
					       DOUBLE dE31_x, DOUBLE dE31_y)
{
  return EdgeZonevEELocal(dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
			  dE20_x, dE20_y, dE21_x, dE21_y,
			  dE30_x, dE30_y, dE31_x, dE31_y);
}

APM2D::CircleTestType APM2D::InCirclevEE_V(DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
					   DOUBLE dE11_x, DOUBLE dE11_y,
					   DOUBLE dE20_x, DOUBLE dE20_y, 
					   DOUBLE dE21_x, DOUBLE dE21_y,
					   DOUBLE dV3_x, DOUBLE dV3_y)
{
  APM2D::CircleTestType eResult;

  //////////////////////////////////////////////////
  // Degenerate Touching Check
  if ( ( dV0_E10_x == dE20_x && dV0_E10_y == dE20_y ) ||
       ( dV0_E10_x == dE21_x && dV0_E10_y == dE21_y ) )
    {
      if ( dV0_E10_x == dV3_x && dV0_E10_y == dV3_y )
	{
	  return APM2D::cttOnCircle;
	}
      else
	{
	  return APM2D::cttOutCircle;
	}
    }

  eResult = APM2D::cttOnCircle;

  // Check to see if the vertex is in the edge envelope
  if ( !Counterclockwise(dV0_E10_x, dV0_E10_y, 
			 dE11_x, dE11_y,
			 dV3_x, dV3_y) )
    {
      return APM2D::cttOutCircle;
    }

  if ( !Counterclockwise(dE20_x, dE20_y, 
			 dE21_x, dE21_y,
			 dV3_x, dV3_y) )
    {
      return APM2D::cttOutCircle;
    }

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InCirclevEE_VLocal(dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dV3_x, dV3_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InCirclevEE_VAdaptive(dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
					dE20_x, dE20_y, dE21_x, dE21_y,
					dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InCirclevEE_VMPFUN(dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = InCirclevEE_VMPFUNFilter(dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InCirclevEE_VMPFUNNormalized(dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
					       dE20_x, dE20_y, dE21_x, dE21_y,
					       dV3_x, dV3_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::CircleTestType InCirclevEE_VLocal(DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
						DOUBLE dE11_x, DOUBLE dE11_y,
						DOUBLE dE20_x, DOUBLE dE20_y, 
						DOUBLE dE21_x, DOUBLE dE21_y,
						DOUBLE dV3_x, DOUBLE dV3_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE dE1dotE1, dE2dotE2, dE1dotE2, dV3dotV3;
  DOUBLE dE1xV3, dE20xE21;
  DOUBLE dE1E2minusE1dotE2;
  DOUBLE dDen;
  DOUBLE dResult;
  APM2D::CircleTestType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = 0.0;
  pppdV[1][0][1] = 0.0;
  pppdV[1][1][0] = dE11_x - dV0_E10_x;
  pppdV[1][1][1] = dE11_y - dV0_E10_y;
  pppdV[2][0][0] = dE20_x - dV0_E10_x;
  pppdV[2][0][1] = dE20_y - dV0_E10_y;
  pppdV[2][1][0] = dE21_x - dV0_E10_x;
  pppdV[2][1][1] = dE21_y - dV0_E10_y;
  pppdV[3][0][0] = dV3_x - dV0_E10_x;
  pppdV[3][0][1] = dV3_y - dV0_E10_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[1][0] = dE11_x - dV0_E10_x;
  ppdE[1][1] = dE11_y - dV0_E10_y;
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;

  dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);
  dE1dotE2 = (ppdE[1][0]*ppdE[2][0]) + (ppdE[1][1]*ppdE[2][1]);
  dV3dotV3 = (pppdV[3][0][0]*pppdV[3][0][0]) + (pppdV[3][0][1]*pppdV[3][0][1]);

  dE1xV3 = (ppdE[1][0]*pppdV[3][0][1]) - (ppdE[1][1]*pppdV[3][0][0]);
  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);

  dE1E2minusE1dotE2 = sqrt(dE1dotE1*dE2dotE2) - dE1dotE2;

  dDen = dE1E2minusE1dotE2;
  ASSERT( dDen != 0.0 );

  dResult = ( dE1E2minusE1dotE2*dV3dotV3 - 2.0*dE20xE21*dE1xV3 ) / dDen;
  
  if ( dResult < -s_cdEpsilonInCirclevEE_VOnCircle )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( dResult > s_cdEpsilonInCirclevEE_VOnCircle )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

static APM2D::CircleTestType InCirclevEE_VAdaptive(DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
						   DOUBLE dE11_x, DOUBLE dE11_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   DOUBLE dV3_x, DOUBLE dV3_y)
{
  return InCirclevEE_VLocal(dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
			    dE20_x, dE20_y, dE21_x, dE21_y,
			    dV3_x, dV3_y);
}

APM2D::CircleTestType APM2D::InCirclevEE_E(DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
					   DOUBLE dE11_x, DOUBLE dE11_y,
					   DOUBLE dE20_x, DOUBLE dE20_y, 
					   DOUBLE dE21_x, DOUBLE dE21_y,
					   DOUBLE dE30_x, DOUBLE dE30_y,
					   DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::CircleTestType eResult;

  //////////////////////////////////////////////////
  // Degenerate Touching Check
  if ( ( dV0_E10_x == dE20_x && dV0_E10_y == dE20_y ) ||
       ( dV0_E10_x == dE21_x && dV0_E10_y == dE21_y ) )
    {
      if ( ( dV0_E10_x == dE30_x && dV0_E10_y == dE30_y ) ||
	   ( dV0_E10_x == dE31_x && dV0_E10_y == dE31_y ) )
	{
	  return APM2D::cttOnCircle;
	}
      else
	{
	  return APM2D::cttOutCircle;
	}
    }

  eResult = APM2D::cttOnCircle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InCirclevEE_ELocal(dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InCirclevEE_EAdaptive(dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
					dE20_x, dE20_y, dE21_x, dE21_y,
					dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InCirclevEE_EMPFUN(dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = InCirclevEE_EMPFUNFilter(dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InCirclevEE_EMPFUNNormalized(dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
					       dE20_x, dE20_y, dE21_x, dE21_y,
					       dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::CircleTestType InCirclevEE_ELocal(DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
						DOUBLE dE11_x, DOUBLE dE11_y,
						DOUBLE dE20_x, DOUBLE dE20_y, 
						DOUBLE dE21_x, DOUBLE dE21_y,
						DOUBLE dE30_x, DOUBLE dE30_y,
						DOUBLE dE31_x, DOUBLE dE31_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE dE1dotE1, dE2dotE2, dE1dotE2, dE1dotE3, dE3dotE3;
  DOUBLE dE1xE3, dE20xE21, dE30xE31;
  DOUBLE dE1E2minusE1dotE2;
  DOUBLE dDen;
  DOUBLE dResult;
  APM2D::CircleTestType eResult;
  
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = 0.0;
  pppdV[0][0][1] = 0.0;
  pppdV[1][0][0] = 0.0;
  pppdV[1][0][1] = 0.0;
  pppdV[1][1][0] = dE11_x - dV0_E10_x;
  pppdV[1][1][1] = dE11_y - dV0_E10_y;
  pppdV[2][0][0] = dE20_x - dV0_E10_x;
  pppdV[2][0][1] = dE20_y - dV0_E10_y;
  pppdV[2][1][0] = dE21_x - dV0_E10_x;
  pppdV[2][1][1] = dE21_y - dV0_E10_y;
  pppdV[3][0][0] = dE30_x - dV0_E10_x;
  pppdV[3][0][1] = dE30_y - dV0_E10_y;
  pppdV[3][1][0] = dE31_x - dV0_E10_x;
  pppdV[3][1][1] = dE31_y - dV0_E10_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[1][0] = dE11_x - dV0_E10_x;
  ppdE[1][1] = dE11_y - dV0_E10_y;
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;
  ppdE[3][0] = dE31_x - dE30_x;
  ppdE[3][1] = dE31_y - dE30_y;
  
  dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);
  dE1dotE2 = (ppdE[1][0]*ppdE[2][0]) + (ppdE[1][1]*ppdE[2][1]);
  dE1dotE3 = (ppdE[1][0]*ppdE[3][0]) + (ppdE[1][1]*ppdE[3][1]);
  dE3dotE3 = (ppdE[3][0]*ppdE[3][0]) + (ppdE[3][1]*ppdE[3][1]);

  dE1xE3 = (ppdE[1][0]*ppdE[3][1]) - (ppdE[1][1]*ppdE[3][0]);
  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);
  dE30xE31 = (pppdV[3][0][0]*pppdV[3][1][1]) - (pppdV[3][0][1]*pppdV[3][1][0]);

  dE1E2minusE1dotE2 = sqrt(dE1dotE1*dE2dotE2) - dE1dotE2;

  dDen = dE3dotE3*dE1E2minusE1dotE2*dE1E2minusE1dotE2;
  ASSERT( dDen != 0.0 );

  dResult = ( dE1E2minusE1dotE2*dE30xE31*( dE1E2minusE1dotE2*dE30xE31 
					   + 2.0*dE20xE21*dE1dotE3 )
	      - dE20xE21*dE20xE21*dE1xE3*dE1xE3 ) / dDen;
  
  if ( dResult < -s_cdEpsilonInCirclevEE_EOnCircle )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( dResult > s_cdEpsilonInCirclevEE_EOnCircle )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

static APM2D::CircleTestType InCirclevEE_EAdaptive(DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
						   DOUBLE dE11_x, DOUBLE dE11_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   DOUBLE dE30_x, DOUBLE dE30_y,
						   DOUBLE dE31_x, DOUBLE dE31_y)
{
  return InCirclevEE_ELocal(dV0_E10_x, dV0_E10_y, dE11_x, dE11_y,
			    dE20_x, dE20_y, dE21_x, dE21_y,
			    dE30_x, dE30_y, dE31_x, dE31_y);
}

APM2D::InscribedCircleCenterType 
APM2D::InscribedCircleCenterEEv(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dE00_x, DOUBLE dE00_y, 
				DOUBLE dE01_x, DOUBLE dE01_y,
				DOUBLE dE10_x, DOUBLE dE10_y, 
				DOUBLE dE11_V2_x, DOUBLE dE11_V2_y)
{
  ASSERT( pdC_x != NULL && pdC_y != NULL && pdR != NULL );

  APM2D::InscribedCircleCenterType eResult;

  eResult = APM2D::icctNull;
  
  //////////////////////////////////////////////////
  // Degenerate Touching Check
  if ( ( dE11_V2_x == dE00_x && dE11_V2_y == dE00_y ) ||
       ( dE11_V2_x == dE01_x && dE11_V2_y == dE01_y ) )
    {
      *pdC_x = dE11_V2_x;
      *pdC_y = dE11_V2_y;
      *pdR = 0.0;
      return APM2D::icctEEv;
    }

  //////////////////////////////////////////////////
  // CCW Checks

  if ( !( Counterclockwise(dE10_x, dE10_y, dE11_V2_x, dE11_V2_y, dE00_x, dE00_y) ||
	  Counterclockwise(dE10_x, dE10_y, dE11_V2_x, dE11_V2_y, dE01_x, dE01_y) ) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  if ( !( Counterclockwise(dE00_x, dE00_y, dE01_x, dE01_y, dE10_x, dE10_y) ||
	  Counterclockwise(dE00_x, dE00_y, dE01_x, dE01_y, dE11_V2_x, dE11_V2_y) ) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InscribedCircleCenterEEvLocal(pdC_x, pdC_y, pdR, 
						dE00_x, dE00_y, dE01_x, dE01_y,
						dE10_x, dE10_y, dE11_V2_x, dE11_V2_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InscribedCircleCenterEEvAdaptive(pdC_x, pdC_y, pdR, 
						   dE00_x, dE00_y, dE01_x, dE01_y,
						   dE10_x, dE10_y, dE11_V2_x, dE11_V2_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InscribedCircleCenterEEvMPFUN(pdC_x, pdC_y, pdR, 
						dE00_x, dE00_y, dE01_x, dE01_y,
						dE10_x, dE10_y, dE11_V2_x, dE11_V2_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = InscribedCircleCenterEEvMPFUNFilter(pdC_x, pdC_y, pdR, 
						dE00_x, dE00_y, dE01_x, dE01_y,
						dE10_x, dE10_y, dE11_V2_x, dE11_V2_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InscribedCircleCenterEEvMPFUNNormalized(pdC_x, pdC_y, pdR, 
							  dE00_x, dE00_y, dE01_x, dE01_y,
							  dE10_x, dE10_y, dE11_V2_x, dE11_V2_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }
  
  return eResult;
}

APM2D::InscribedCircleCenterType 
APM2D::InscribedCircleCenterEvEv(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				 DOUBLE dE00_x, DOUBLE dE00_y, 
				 DOUBLE dE01_E10_x, DOUBLE dE01_E10_y,
				 DOUBLE dE11_V2_x, DOUBLE dE11_V2_y)
{
  ASSERT( pdC_x != NULL && pdC_y != NULL && pdR != NULL );

  APM2D::InscribedCircleCenterType eResult;

  eResult = APM2D::icctNull;
  
  //////////////////////////////////////////////////
  // Degenerate Touching Check
  if ( ( dE11_V2_x == dE00_x && dE11_V2_y == dE00_y ) ||
       ( dE11_V2_x == dE01_E10_x && dE11_V2_y == dE01_E10_y ) )
    {
      ASSERT( FALSE );
      *pdC_x = dE11_V2_x;
      *pdC_y = dE11_V2_y;
      *pdR = 0.0;
      return APM2D::icctEEv;
    }

  //////////////////////////////////////////////////
  // CCW Checks

  if ( !Counterclockwise(dE01_E10_x, dE01_E10_y, dE11_V2_x, dE11_V2_y, dE00_x, dE00_y) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InscribedCircleCenterEEvLocal(pdC_x, pdC_y, pdR, 
						dE00_x, dE00_y, 
						dE01_E10_x, dE01_E10_y,
						dE01_E10_x, dE01_E10_y, 
						dE11_V2_x, dE11_V2_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InscribedCircleCenterEEvAdaptive(pdC_x, pdC_y, pdR, 
						   dE00_x, dE00_y, 
						   dE01_E10_x, dE01_E10_y,
						   dE01_E10_x, dE01_E10_y, 
						   dE11_V2_x, dE11_V2_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InscribedCircleCenterEEvMPFUN(pdC_x, pdC_y, pdR, 
						dE00_x, dE00_y, 
						dE01_E10_x, dE01_E10_y,
						dE01_E10_x, dE01_E10_y, 
						dE11_V2_x, dE11_V2_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = InscribedCircleCenterEEvMPFUNFilter(pdC_x, pdC_y, pdR, 
						dE00_x, dE00_y, 
						dE01_E10_x, dE01_E10_y,
						dE01_E10_x, dE01_E10_y, 
						dE11_V2_x, dE11_V2_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InscribedCircleCenterEEvMPFUNNormalized(pdC_x, pdC_y, pdR, 
							  dE00_x, dE00_y, 
							  dE01_E10_x, dE01_E10_y,
							  dE01_E10_x, dE01_E10_y, 
							  dE11_V2_x, dE11_V2_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }
  
  return eResult;
}

static APM2D::InscribedCircleCenterType 
InscribedCircleCenterEEvLocal(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			      DOUBLE dE00_x, DOUBLE dE00_y, 
			      DOUBLE dE01_x, DOUBLE dE01_y,
			      DOUBLE dE10_x, DOUBLE dE10_y, 
			      DOUBLE dE11_V2_x, DOUBLE dE11_V2_y)
{
  DOUBLE pdC[3], dR;
  DOUBLE pppdV[3][2][2]; // Endpoints
  DOUBLE ppdE[3][2];     // Edge vectors
  DOUBLE pdEdgeZone[2];
  DOUBLE dDen;
  DOUBLE dE0dotE0, dE1dotE1, dE0dotE1;
  DOUBLE dE0dotE00, dE0dotE01;
  DOUBLE dE0xE1, dE00xE01;
  DOUBLE dE0E1minusE0dotE1, dSinTheta, dSinThetaTemp;
  DOUBLE dMagCE00, dMagCE01;
  UINT_32 i;
  
  // Translate V1 to the origin
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = dE00_x - dE11_V2_x;
  pppdV[0][0][1] = dE00_y - dE11_V2_y;
  pppdV[0][1][0] = dE01_x - dE11_V2_x;
  pppdV[0][1][1] = dE01_y - dE11_V2_y;
  pppdV[1][0][0] = dE10_x - dE11_V2_x;
  pppdV[1][0][1] = dE10_y - dE11_V2_y;
  pppdV[1][1][0] = 0.0;
  pppdV[1][1][1] = 0.0;
  pppdV[2][0][0] = 0.0;
  pppdV[2][0][1] = 0.0;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[0][0] = dE01_x - dE00_x;
  ppdE[0][1] = dE01_y - dE00_y;
  ppdE[1][0] = dE11_V2_x - dE10_x;
  ppdE[1][1] = dE11_V2_y - dE10_y;
  
  dE0dotE0 = (ppdE[0][0]*ppdE[0][0]) + (ppdE[0][1]*ppdE[0][1]);
  dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
  dE0dotE1 = (ppdE[0][0]*ppdE[1][0]) + (ppdE[0][1]*ppdE[1][1]);
  dE0dotE00 = (ppdE[0][0]*pppdV[0][0][0]) + (ppdE[0][1]*pppdV[0][0][1]);
  dE0dotE01 = (ppdE[0][0]*pppdV[0][1][0]) + (ppdE[0][1]*pppdV[0][1][1]);

  dE0xE1 = (ppdE[0][0]*ppdE[1][1]) - (ppdE[0][1]*ppdE[1][0]);
  dE00xE01 = (pppdV[0][0][0]*pppdV[0][1][1]) - (pppdV[0][0][1]*pppdV[0][1][0]);

  dE0E1minusE0dotE1 = sqrt(dE0dotE0*dE1dotE1) - dE0dotE1;

  //dSinTheta = dE0E1minusE0dotE1/sqrt(dE0dotE0*dE1dotE1);
  dSinTheta = dE0xE1/sqrt(dE0dotE0*dE1dotE1);
  if ( fabs(dSinTheta) < s_cdEpsilonInscribedCircleCenterEEvParallel &&
       dE0dotE1 > 0.0 )
    {
      // Colinear
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  //dSinTheta = dE00xE01/sqrt(dE00dotE00*dE01dotE01);
  dSinTheta = APM2D::SinTheta(dE11_V2_x, dE11_V2_y, dE00_x, dE00_y, dE01_x, dE01_y);
  dSinThetaTemp = APM2D::SinTheta(dE00_x, dE00_y, dE01_x, dE01_y, dE11_V2_x, dE11_V2_y);
  if ( dSinThetaTemp > dSinTheta )
    {
      dSinTheta = dSinThetaTemp;
    }
  if ( dSinTheta <= s_cdEpsilonInscribedCircleCenterEEvVertexOnEdge )
    {
      // V2 is on E0 and it is not one of the endpoints
      
#ifdef ENABLE_DEBUGGING_OUTPUT
      fprintf(stderr, "    VEE: V2 is on E0 ( %f ) <= %f\n", 
	      dE00xE01, s_cdEpsilonInscribedCircleCenterEEvVertexOnEdge);
      fprintf(stderr, "      E0: ( %.5f, %.5f ) => ( %.5f, %.5f )\n", 
	      dE00_x, dE00_y, dE01_x, dE01_y);
      fprintf(stderr, "      E1: ( %.5f, %.5f ) => ( %.5f, %.5f )\n", 
	      dE10_x, dE10_y, dE11_V2_x, dE11_V2_y);
      fprintf(stderr, "      V2: ( %.5f, %.5f )\n", dE11_V2_x, dE11_V2_y);
#endif // ENABLE_DEBUGGING_OUTPUT

      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  dDen = dE0E1minusE0dotE1;
  ASSERT( dDen != 0.0 );

  //////////////////////////////////////////////////
  // Compute the center and radius

  pdC[0] = -( dE00xE01*ppdE[1][1] ) / dDen;
  pdC[1] = ( dE00xE01*ppdE[1][0] ) / dDen;
  pdC[2] = sqrt((pdC[0]*pdC[0]) + (pdC[1]*pdC[1]));

  dR = ( sqrt(dE1dotE1)*dE00xE01 ) / dDen;
  
  //////////////////////////////////////////////////
  // Check the Edge Zone

  dMagCE00 = ( (pdC[0] - pppdV[0][0][0])*(pdC[0] - pppdV[0][0][0]) + 
	       (pdC[1] - pppdV[0][0][1])*(pdC[1] - pppdV[0][0][1]) );
  dMagCE01 = ( (pdC[0] - pppdV[0][1][0])*(pdC[0] - pppdV[0][1][0]) + 
	       (pdC[1] - pppdV[0][1][1])*(pdC[1] - pppdV[0][1][1]) );

  if ( dMagCE00 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[0] = 0.0;
    }
  else
    {
      pdEdgeZone[0] = -( ( dE0E1minusE0dotE1*dE0dotE00 + dE00xE01*dE0xE1 ) / 
			 ( dE0E1minusE0dotE1*sqrt(dE0dotE0*dMagCE00) ) ); 
    }
  if ( dMagCE01 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[1] = 0.0;
    }
  else
    {
      pdEdgeZone[1] = ( ( dE0E1minusE0dotE1*dE0dotE01 + dE00xE01*dE0xE1 ) / 
			( dE0E1minusE0dotE1*sqrt(dE0dotE0*dMagCE01) ) ); 
    }
  
  for ( i = 0; i < 2; i++ )
    {
      if ( pdEdgeZone[i] < -s_cdEpsilonInscribedCircleCenterEEvEdgeZone )
	{
#ifdef DEBUG_EPSILON_VALUES
	  if ( !( pdEdgeZone[i] < -(s_cdEpsilonInscribedCircleCenterEEvEdgeZone*s_cdOrderOfMagnitude) ) )
	    {
	      fprintf(stderr, "  EEv: Edge Zone: %e < %e < %e\n",
		      s_cdEpsilonInscribedCircleCenterEEvEdgeZone*s_cdOrderOfMagnitude,
		      pdEdgeZone[i],
		      s_cdEpsilonInscribedCircleCenterEEvEdgeZone);
	    }
#endif // DEBUG_EPSILON_VALUES	      
	  *pdC_x = 0.0;
	  *pdC_y = 0.0;
	  *pdR = 0.0;
	  return APM2D::icctNull;
	}
    }

  if ( !( fabs(pdC[2] - dR) < s_cdEpsilonInscribedCircleCenterEEvRadius ) )
    {
      fprintf(stderr, "EEv: C = ( %f, %f, %f ) != %f = R\n", pdC[0], pdC[1], pdC[2], dR);
      //ASSERT( FALSE );
    }

  // Translate the answer back out away from the origin
  *pdC_x = dE11_V2_x + pdC[0];
  *pdC_y = dE11_V2_y + pdC[1];
  *pdR = dR;
  return APM2D::icctEEv;
}

static APM2D::InscribedCircleCenterType 
InscribedCircleCenterEEvAdaptive(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				 DOUBLE dE00_x, DOUBLE dE00_y, 
				 DOUBLE dE01_x, DOUBLE dE01_y,
				 DOUBLE dE10_x, DOUBLE dE10_y, 
				 DOUBLE dE11_V2_x, DOUBLE dE11_V2_y)
{
  return InscribedCircleCenterEEvLocal(pdC_x, pdC_y, pdR, 
				       dE00_x, dE00_y, dE01_x, dE01_y,
				       dE10_x, dE10_y, dE11_V2_x, dE11_V2_y);
}

APM2D::EdgeZoneType APM2D::EdgeZoneEEv(DOUBLE dE00_x, DOUBLE dE00_y, 
				       DOUBLE dE01_x, DOUBLE dE01_y,
				       DOUBLE dE10_x, DOUBLE dE10_y, 
				       DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
				       DOUBLE dE30_x, DOUBLE dE30_y,
				       DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::EdgeZoneType eResult;

  //////////////////////////////////////////////////
  // Degenerate Touching Check
  if ( ( dE11_V2_x == dE00_x && dE11_V2_y == dE00_y ) ||
       ( dE11_V2_x == dE01_x && dE11_V2_y == dE01_y ) )
    {
      return EdgeZone(dE11_V2_x, dE11_V2_y,
		      dE30_x, dE30_y,
		      dE31_x, dE31_y);
    }

  eResult = APM2D::eztMiddle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = EdgeZoneEEvLocal(dE00_x, dE00_y, dE01_x, dE01_y,
				   dE10_x, dE10_y, dE11_V2_x, dE11_V2_y,
				   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atExact:
      {
	eResult = EdgeZoneEEvAdaptive(dE00_x, dE00_y, dE01_x, dE01_y,
				      dE10_x, dE10_y, dE11_V2_x, dE11_V2_y,
				      dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = EdgeZoneEEvMPFUN(dE00_x, dE00_y, dE01_x, dE01_y,
				   dE10_x, dE10_y, dE11_V2_x, dE11_V2_y,
				   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = EdgeZoneEEvMPFUNFilter(dE00_x, dE00_y, dE01_x, dE01_y,
				   dE10_x, dE10_y, dE11_V2_x, dE11_V2_y,
				   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = EdgeZoneEEvMPFUNNormalized(dE00_x, dE00_y, dE01_x, dE01_y,
					     dE10_x, dE10_y, dE11_V2_x, dE11_V2_y,
					     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::EdgeZoneType EdgeZoneEEvLocal(DOUBLE dE00_x, DOUBLE dE00_y, 
					    DOUBLE dE01_x, DOUBLE dE01_y,
					    DOUBLE dE10_x, DOUBLE dE10_y, 
					    DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
					    DOUBLE dE30_x, DOUBLE dE30_y,
					    DOUBLE dE31_x, DOUBLE dE31_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE pdC[2];
  DOUBLE pdEdgeZone[2];
  DOUBLE dE0dotE0, dE1dotE1, dE0dotE1;
  DOUBLE dE3dotE30, dE3dotE31;
  DOUBLE dE1xE3, dE00xE01;
  DOUBLE dE0E1minusE0dotE1;
  DOUBLE dE3dotE3, dMagCE30, dMagCE31;
  DOUBLE dDen;
  APM2D::EdgeZoneType eResult;
  
  // Translate V1 to the origin
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = dE00_x - dE11_V2_x;
  pppdV[0][0][1] = dE00_y - dE11_V2_y;
  pppdV[0][1][0] = dE01_x - dE11_V2_x;
  pppdV[0][1][1] = dE01_y - dE11_V2_y;
  pppdV[1][0][0] = dE10_x - dE11_V2_x;
  pppdV[1][0][1] = dE10_y - dE11_V2_y;
  pppdV[1][1][0] = 0.0;
  pppdV[1][1][1] = 0.0;
  pppdV[2][0][0] = 0.0;
  pppdV[2][0][1] = 0.0;
  pppdV[3][0][0] = dE30_x - dE11_V2_x;
  pppdV[3][0][1] = dE30_y - dE11_V2_y;
  pppdV[3][1][0] = dE31_x - dE11_V2_x;
  pppdV[3][1][1] = dE31_y - dE11_V2_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[0][0] = dE01_x - dE00_x;
  ppdE[0][1] = dE01_y - dE00_y;
  ppdE[1][0] = dE11_V2_x - dE10_x;
  ppdE[1][1] = dE11_V2_y - dE10_y;
  ppdE[3][0] = dE31_x - dE30_x;
  ppdE[3][1] = dE31_y - dE30_y;
  
  dE0dotE0 = (ppdE[0][0]*ppdE[0][0]) + (ppdE[0][1]*ppdE[0][1]);
  dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
  dE0dotE1 = (ppdE[0][0]*ppdE[1][0]) + (ppdE[0][1]*ppdE[1][1]);
  dE3dotE3 = (ppdE[3][0]*ppdE[3][0]) + (ppdE[3][1]*ppdE[3][1]);

  dE3dotE30 = (ppdE[3][0]*pppdV[3][0][0]) + (ppdE[3][1]*pppdV[3][0][1]);
  dE3dotE31 = (ppdE[3][0]*pppdV[3][1][0]) + (ppdE[3][1]*pppdV[3][1][1]);

  dE1xE3 = (ppdE[1][0]*ppdE[3][1]) - (ppdE[1][1]*ppdE[3][0]);
  dE00xE01 = (pppdV[0][0][0]*pppdV[0][1][1]) - (pppdV[0][0][1]*pppdV[0][1][0]);

  dE0E1minusE0dotE1 = sqrt(dE0dotE0*dE1dotE1) - dE0dotE1;

  dDen = dE0E1minusE0dotE1;
  ASSERT( dDen != 0.0 );

  pdC[0] = -( dE00xE01*ppdE[1][1] ) / dDen;
  pdC[1] = ( dE00xE01*ppdE[1][0] ) / dDen;

  dMagCE30 = ( (pdC[0] - pppdV[3][0][0])*(pdC[0] - pppdV[3][0][0]) + 
	       (pdC[1] - pppdV[3][0][1])*(pdC[1] - pppdV[3][0][1]) );
  dMagCE31 = ( (pdC[0] - pppdV[3][1][0])*(pdC[0] - pppdV[3][1][0]) + 
	       (pdC[1] - pppdV[3][1][1])*(pdC[1] - pppdV[3][1][1]) );

  if ( dMagCE30 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[0] = 0.0;
    }
  else
    {
      pdEdgeZone[0] = ( ( dE0E1minusE0dotE1*dE3dotE30 - dE00xE01*dE1xE3 ) / 
			( dDen*sqrt(dE3dotE3*dMagCE30) ) );
    }
  if ( dMagCE31 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[1] = 0.0;
    }
  else
    {
      pdEdgeZone[1] = -( ( dE0E1minusE0dotE1*dE3dotE31 - dE00xE01*dE1xE3 ) / 
			 ( dDen*sqrt(dE3dotE3*dMagCE31) ) );
    }

  if ( !( pdEdgeZone[0] < -s_cdEpsilonEdgeZoneEEvEdgeZone ) )
    {
      eResult = APM2D::eztStart;
    }
  else if ( !( pdEdgeZone[1] < -s_cdEpsilonEdgeZoneEEvEdgeZone ) )
    {
      eResult = APM2D::eztEnd;
    }
  else
    {
      eResult = APM2D::eztMiddle;
    }

  return eResult;
}

static APM2D::EdgeZoneType EdgeZoneEEvAdaptive(DOUBLE dE00_x, DOUBLE dE00_y, 
					       DOUBLE dE01_x, DOUBLE dE01_y,
					       DOUBLE dE10_x, DOUBLE dE10_y, 
					       DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
					       DOUBLE dE30_x, DOUBLE dE30_y,
					       DOUBLE dE31_x, DOUBLE dE31_y)
{
  return EdgeZoneEEvLocal(dE00_x, dE00_y, dE01_x, dE01_y,
			  dE10_x, dE10_y, dE11_V2_x, dE11_V2_y,
			  dE30_x, dE30_y, dE31_x, dE31_y);
}

APM2D::CircleTestType APM2D::InCircleEEv_V(DOUBLE dE00_x, DOUBLE dE00_y, 
					   DOUBLE dE01_x, DOUBLE dE01_y,
					   DOUBLE dE10_x, DOUBLE dE10_y, 
					   DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
					   DOUBLE dV3_x, DOUBLE dV3_y)
{
  APM2D::CircleTestType eResult;

  //////////////////////////////////////////////////
  // Degenerate Touching Check
  if ( ( dE11_V2_x == dE00_x && dE11_V2_y == dE00_y ) ||
       ( dE11_V2_x == dE01_x && dE11_V2_y == dE01_y ) )
    {
      if ( dE11_V2_x == dV3_x && dE11_V2_y == dV3_y )
	{
	  return APM2D::cttOnCircle;
	}
      else
	{
	  return APM2D::cttOutCircle;
	}
    }

  eResult = APM2D::cttOnCircle;

  // Check to see if the vertex is in the edge envelope
  if ( !Counterclockwise(dE00_x, dE00_y, 
			 dE01_x, dE01_y,
			 dV3_x, dV3_y) )
    {
      return APM2D::cttOutCircle;
    }

  if ( !Counterclockwise(dE10_x, dE10_y, 
			 dE11_V2_x, dE11_V2_y,
			 dV3_x, dV3_y) )
    {
      return APM2D::cttOutCircle;
    }

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InCircleEEv_VLocal(dE00_x, dE00_y, dE01_x, dE01_y,
				     dE10_x, dE10_y, dE11_V2_x, dE11_V2_y,
				     dV3_x, dV3_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InCircleEEv_VAdaptive(dE00_x, dE00_y, dE01_x, dE01_y,
					dE10_x, dE10_y, dE11_V2_x, dE11_V2_y,
					dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InCircleEEv_VMPFUN(dE00_x, dE00_y, dE01_x, dE01_y,
				     dE10_x, dE10_y, dE11_V2_x, dE11_V2_y,
				     dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = InCircleEEv_VMPFUNFilter(dE00_x, dE00_y, dE01_x, dE01_y,
				     dE10_x, dE10_y, dE11_V2_x, dE11_V2_y,
				     dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InCircleEEv_VMPFUNNormalized(dE00_x, dE00_y, dE01_x, dE01_y,
					       dE10_x, dE10_y, dE11_V2_x, dE11_V2_y,
					       dV3_x, dV3_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::CircleTestType InCircleEEv_VLocal(DOUBLE dE00_x, DOUBLE dE00_y, 
						DOUBLE dE01_x, DOUBLE dE01_y,
						DOUBLE dE10_x, DOUBLE dE10_y, 
						DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
						DOUBLE dV3_x, DOUBLE dV3_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE dE0dotE0, dE1dotE1, dE0dotE1, dV3dotV3;
  DOUBLE dE1xV3, dE00xE01;
  DOUBLE dE0E1minusE0dotE1;
  DOUBLE dDen;
  DOUBLE dResult;
  APM2D::CircleTestType eResult;
  
  // Translate V1 to the origin
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = dE00_x - dE11_V2_x;
  pppdV[0][0][1] = dE00_y - dE11_V2_y;
  pppdV[0][1][0] = dE01_x - dE11_V2_x;
  pppdV[0][1][1] = dE01_y - dE11_V2_y;
  pppdV[1][0][0] = dE10_x - dE11_V2_x;
  pppdV[1][0][1] = dE10_y - dE11_V2_y;
  pppdV[1][1][0] = 0.0;
  pppdV[1][1][1] = 0.0;
  pppdV[2][0][0] = 0.0;
  pppdV[2][0][1] = 0.0;
  pppdV[3][0][0] = dV3_x - dE11_V2_x;
  pppdV[3][0][1] = dV3_y - dE11_V2_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[0][0] = dE01_x - dE00_x;
  ppdE[0][1] = dE01_y - dE00_y;
  ppdE[1][0] = dE11_V2_x - dE10_x;
  ppdE[1][1] = dE11_V2_y - dE10_y;
  
  dE0dotE0 = (ppdE[0][0]*ppdE[0][0]) + (ppdE[0][1]*ppdE[0][1]);
  dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
  dE0dotE1 = (ppdE[0][0]*ppdE[1][0]) + (ppdE[0][1]*ppdE[1][1]);
  dV3dotV3 = (pppdV[3][0][0]*pppdV[3][0][0]) + (pppdV[3][0][1]*pppdV[3][0][1]);

  dE1xV3 = (ppdE[1][0]*pppdV[3][0][1]) - (ppdE[1][1]*pppdV[3][0][0]);
  dE00xE01 = (pppdV[0][0][0]*pppdV[0][1][1]) - (pppdV[0][0][1]*pppdV[0][1][0]);

  dE0E1minusE0dotE1 = sqrt(dE0dotE0*dE1dotE1) - dE0dotE1;

  dDen = dE0E1minusE0dotE1;
  ASSERT( dDen != 0.0 );

  dResult = ( dE0E1minusE0dotE1*dV3dotV3 - 2.0*dE00xE01*dE1xV3 ) / dDen;
  
  if ( dResult < -s_cdEpsilonInCircleEEv_VOnCircle )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( dResult > s_cdEpsilonInCircleEEv_VOnCircle )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

static APM2D::CircleTestType InCircleEEv_VAdaptive(DOUBLE dE00_x, DOUBLE dE00_y, 
						   DOUBLE dE01_x, DOUBLE dE01_y,
						   DOUBLE dE10_x, DOUBLE dE10_y, 
						   DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
						   DOUBLE dV3_x, DOUBLE dV3_y)
{
  return InCircleEEv_VLocal(dE00_x, dE00_y, dE01_x, dE01_y,
			    dE10_x, dE10_y, dE11_V2_x, dE11_V2_y,
			    dV3_x, dV3_y);
}

APM2D::CircleTestType APM2D::InCircleEEv_E(DOUBLE dE00_x, DOUBLE dE00_y, 
					   DOUBLE dE01_x, DOUBLE dE01_y,
					   DOUBLE dE10_x, DOUBLE dE10_y, 
					   DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
					   DOUBLE dE30_x, DOUBLE dE30_y,
					   DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::CircleTestType eResult;

  //////////////////////////////////////////////////
  // Degenerate Touching Check
  if ( ( dE11_V2_x == dE00_x && dE11_V2_y == dE00_y ) ||
       ( dE11_V2_x == dE01_x && dE11_V2_y == dE01_y ) )
    {
      if ( ( dE11_V2_x == dE30_x && dE11_V2_y == dE30_y ) ||
	   ( dE11_V2_x == dE31_x && dE11_V2_y == dE31_y ) )
	{
	  return APM2D::cttOnCircle;
	}
      else
	{
	  return APM2D::cttOutCircle;
	}
    }

  eResult = APM2D::cttOnCircle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InCircleEEv_ELocal(dE00_x, dE00_y, dE01_x, dE01_y,
				     dE10_x, dE10_y, dE11_V2_x, dE11_V2_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InCircleEEv_EAdaptive(dE00_x, dE00_y, dE01_x, dE01_y,
					dE10_x, dE10_y, dE11_V2_x, dE11_V2_y,
					dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InCircleEEv_EMPFUN(dE00_x, dE00_y, dE01_x, dE01_y,
				     dE10_x, dE10_y, dE11_V2_x, dE11_V2_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = InCircleEEv_EMPFUNFilter(dE00_x, dE00_y, dE01_x, dE01_y,
				     dE10_x, dE10_y, dE11_V2_x, dE11_V2_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InCircleEEv_EMPFUNNormalized(dE00_x, dE00_y, dE01_x, dE01_y,
					       dE10_x, dE10_y, dE11_V2_x, dE11_V2_y,
					       dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::CircleTestType InCircleEEv_ELocal(DOUBLE dE00_x, DOUBLE dE00_y, 
						DOUBLE dE01_x, DOUBLE dE01_y,
						DOUBLE dE10_x, DOUBLE dE10_y, 
						DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
						DOUBLE dE30_x, DOUBLE dE30_y,
						DOUBLE dE31_x, DOUBLE dE31_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE dE0dotE0, dE1dotE1, dE0dotE1, dE1dotE3, dE3dotE3;
  DOUBLE dE1xE3, dE00xE01, dE30xE31;
  DOUBLE dE0E1minusE0dotE1;
  DOUBLE dDen;
  DOUBLE dResult;
  APM2D::CircleTestType eResult;
  
  // Translate V1 to the origin
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = dE00_x - dE11_V2_x;
  pppdV[0][0][1] = dE00_y - dE11_V2_y;
  pppdV[0][1][0] = dE01_x - dE11_V2_x;
  pppdV[0][1][1] = dE01_y - dE11_V2_y;
  pppdV[1][0][0] = dE10_x - dE11_V2_x;
  pppdV[1][0][1] = dE10_y - dE11_V2_y;
  pppdV[1][1][0] = 0.0;
  pppdV[1][1][1] = 0.0;
  pppdV[2][0][0] = 0.0;
  pppdV[2][0][1] = 0.0;
  pppdV[3][0][0] = dE30_x - dE11_V2_x;
  pppdV[3][0][1] = dE30_y - dE11_V2_y;
  pppdV[3][1][0] = dE31_x - dE11_V2_x;
  pppdV[3][1][1] = dE31_y - dE11_V2_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[0][0] = dE01_x - dE00_x;
  ppdE[0][1] = dE01_y - dE00_y;
  ppdE[1][0] = dE11_V2_x - dE10_x;
  ppdE[1][1] = dE11_V2_y - dE10_y;
  ppdE[3][0] = dE31_x - dE30_x;
  ppdE[3][1] = dE31_y - dE30_y;
  
  dE0dotE0 = (ppdE[0][0]*ppdE[0][0]) + (ppdE[0][1]*ppdE[0][1]);
  dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
  dE0dotE1 = (ppdE[0][0]*ppdE[1][0]) + (ppdE[0][1]*ppdE[1][1]);
  dE1dotE3 = (ppdE[1][0]*ppdE[3][0]) + (ppdE[1][1]*ppdE[3][1]);
  dE3dotE3 = (ppdE[3][0]*ppdE[3][0]) + (ppdE[3][1]*ppdE[3][1]);

  dE1xE3 = (ppdE[1][0]*ppdE[3][1]) - (ppdE[1][1]*ppdE[3][0]);
  dE00xE01 = (pppdV[0][0][0]*pppdV[0][1][1]) - (pppdV[0][0][1]*pppdV[0][1][0]);
  dE30xE31 = (pppdV[3][0][0]*pppdV[3][1][1]) - (pppdV[3][0][1]*pppdV[3][1][0]);

  dE0E1minusE0dotE1 = sqrt(dE0dotE0*dE1dotE1) - dE0dotE1;

  dDen = dE3dotE3*dE0E1minusE0dotE1*dE0E1minusE0dotE1;
  ASSERT( dDen != 0.0 );

  dResult = ( dE0E1minusE0dotE1*dE30xE31*( dE0E1minusE0dotE1*dE30xE31
					   + 2.0*dE00xE01*dE1dotE3 )
	      - dE00xE01*dE00xE01*dE1xE3*dE1xE3 ) / dDen;
  
  if ( dResult < -s_cdEpsilonInCircleEEv_EOnCircle )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( dResult > s_cdEpsilonInCircleEEv_EOnCircle )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

static APM2D::CircleTestType InCircleEEv_EAdaptive(DOUBLE dE00_x, DOUBLE dE00_y, 
						   DOUBLE dE01_x, DOUBLE dE01_y,
						   DOUBLE dE10_x, DOUBLE dE10_y, 
						   DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
						   DOUBLE dE30_x, DOUBLE dE30_y,
						   DOUBLE dE31_x, DOUBLE dE31_y)
{
  return InCircleEEv_ELocal(dE00_x, dE00_y, dE01_x, dE01_y,
			    dE10_x, dE10_y, dE11_V2_x, dE11_V2_y,
			    dE30_x, dE30_y, dE31_x, dE31_y);
}

APM2D::InscribedCircleCenterType 
APM2D::InscribedCircleCenterEEE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dE00_x, DOUBLE dE00_y, 
				DOUBLE dE01_x, DOUBLE dE01_y,
				DOUBLE dE10_x, DOUBLE dE10_y, 
				DOUBLE dE11_x, DOUBLE dE11_y,
				DOUBLE dE20_x, DOUBLE dE20_y, 
				DOUBLE dE21_x, DOUBLE dE21_y)
{
  ASSERT( pdC_x != NULL && pdC_y != NULL && pdR != NULL );

  APM2D::InscribedCircleCenterType eResult;

  eResult = APM2D::icctNull;
  
  //////////////////////////////////////////////////
  // CCW Checks

  if ( !( Counterclockwise(dE00_x, dE00_y, dE01_x, dE01_y, dE10_x, dE10_y) ||
	  Counterclockwise(dE00_x, dE00_y, dE01_x, dE01_y, dE11_x, dE11_y) ) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  if ( !( Counterclockwise(dE00_x, dE00_y, dE01_x, dE01_y, dE20_x, dE20_y) ||
	  Counterclockwise(dE00_x, dE00_y, dE01_x, dE01_y, dE21_x, dE21_y) ) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  if ( !( Counterclockwise(dE10_x, dE10_y, dE11_x, dE11_y, dE00_x, dE00_y) ||
	  Counterclockwise(dE10_x, dE10_y, dE11_x, dE11_y, dE01_x, dE01_y) ) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  if ( !( Counterclockwise(dE10_x, dE10_y, dE11_x, dE11_y, dE20_x, dE20_y) ||
	  Counterclockwise(dE10_x, dE10_y, dE11_x, dE11_y, dE21_x, dE21_y) ) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  if ( !( Counterclockwise(dE20_x, dE20_y, dE21_x, dE21_y, dE00_x, dE00_y) ||
	  Counterclockwise(dE20_x, dE20_y, dE21_x, dE21_y, dE01_x, dE01_y) ) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  if ( !( Counterclockwise(dE20_x, dE20_y, dE21_x, dE21_y, dE10_x, dE10_y) ||
	  Counterclockwise(dE20_x, dE20_y, dE21_x, dE21_y, dE11_x, dE11_y) ) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InscribedCircleCenterEEELocal(pdC_x, pdC_y, pdR, 
						dE00_x, dE00_y, dE01_x, dE01_y,
						dE10_x, dE10_y, dE11_x, dE11_y,
						dE20_x, dE20_y, dE21_x, dE21_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InscribedCircleCenterEEEAdaptive(pdC_x, pdC_y, pdR, 
						   dE00_x, dE00_y, dE01_x, dE01_y,
						   dE10_x, dE10_y, dE11_x, dE11_y,
						   dE20_x, dE20_y, dE21_x, dE21_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InscribedCircleCenterEEEMPFUN(pdC_x, pdC_y, pdR, 
						dE00_x, dE00_y, dE01_x, dE01_y,
						dE10_x, dE10_y, dE11_x, dE11_y,
						dE20_x, dE20_y, dE21_x, dE21_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = InscribedCircleCenterEEEMPFUNFilter(pdC_x, pdC_y, pdR, 
						dE00_x, dE00_y, dE01_x, dE01_y,
						dE10_x, dE10_y, dE11_x, dE11_y,
						dE20_x, dE20_y, dE21_x, dE21_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InscribedCircleCenterEEEMPFUNNormalized(pdC_x, pdC_y, pdR, 
							  dE00_x, dE00_y, dE01_x, dE01_y,
							  dE10_x, dE10_y, dE11_x, dE11_y,
							  dE20_x, dE20_y, dE21_x, dE21_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }
  
  return eResult;
}

APM2D::InscribedCircleCenterType 
APM2D::InscribedCircleCenterEvEE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				 DOUBLE dE00_x, DOUBLE dE00_y, 
				 DOUBLE dE01_E10_x, DOUBLE dE01_E10_y,
				 DOUBLE dE11_x, DOUBLE dE11_y,
				 DOUBLE dE20_x, DOUBLE dE20_y, 
				 DOUBLE dE21_x, DOUBLE dE21_y)
{
  ASSERT( pdC_x != NULL && pdC_y != NULL && pdR != NULL );

  APM2D::InscribedCircleCenterType eResult;

  eResult = APM2D::icctNull;
  
  //////////////////////////////////////////////////
  // Degenerate Touching Check
  if ( ( dE01_E10_x == dE20_x && dE01_E10_y == dE20_y ) ||
       ( dE01_E10_x == dE21_x && dE01_E10_y == dE21_y ) )
    {
      *pdC_x = dE01_E10_x;
      *pdC_y = dE01_E10_y;
      *pdR = 0.0;
      return APM2D::icctvEE;
    }

  //////////////////////////////////////////////////
  // Colinear Check

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
    case APM::atExact:
      {
	DOUBLE dSinTheta;
	dSinTheta = APM2D::SinTheta(dE00_x, dE00_y, dE01_E10_x, dE01_E10_y, dE11_x, dE11_y);
	if ( fabs(dSinTheta) <= s_cdEpsilonInscribedCircleCenterEvEEParallel )
	  {
	    eResult = APM2D::InscribedCircleCentervEE(pdC_x, pdC_y, pdR, 
						      dE01_E10_x, dE01_E10_y, dE11_x, dE11_y,
						      dE20_x, dE20_y, dE21_x, dE21_y);
	    if ( eResult != APM2D::icctNull )
	      {
		eResult = icctEcEE;
	      }
	    return eResult;      
	  }
	else if ( dSinTheta < 0 )
	  {
	    *pdC_x = 0.0;
	    *pdC_y = 0.0;
	    *pdR = 0.0;
	    return APM2D::icctNull;
	  }
      }
      break;
    case APM::atMPFUN:
    case APM::atMPFUNFilter:
    case APM::atMPFUNNormalized:
      {
	INT_32 iResult;
	iResult = OrientationVVV(dE00_x, dE00_y, dE01_E10_x, dE01_E10_y, dE11_x, dE11_y);
	if ( iResult == 0 )
	  {
	    eResult = APM2D::InscribedCircleCentervEE(pdC_x, pdC_y, pdR, 
						      dE01_E10_x, dE01_E10_y, dE11_x, dE11_y,
						      dE20_x, dE20_y, dE21_x, dE21_y);
	    if ( eResult != APM2D::icctNull )
	      {
		eResult = icctEcEE;
	      }
	    return eResult;      
	  }
	else if ( iResult < 0 )
	  {
	    *pdC_x = 0.0;
	    *pdC_y = 0.0;
	    *pdR = 0.0;
	    return APM2D::icctNull;
	  }
      }
      break;
    default:
      {
	ASSERT( FALSE );
      }
      break;
    }

  //////////////////////////////////////////////////
  // CCW Checks

  if ( !( Counterclockwise(dE00_x, dE00_y, dE01_E10_x, dE01_E10_y, dE20_x, dE20_y) ||
	  Counterclockwise(dE00_x, dE00_y, dE01_E10_x, dE01_E10_y, dE21_x, dE21_y) ) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  if ( !( Counterclockwise(dE01_E10_x, dE01_E10_y, dE11_x, dE11_y, dE20_x, dE20_y) ||
	  Counterclockwise(dE01_E10_x, dE01_E10_y, dE11_x, dE11_y, dE21_x, dE21_y) ) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  if ( !Counterclockwise(dE20_x, dE20_y, dE21_x, dE21_y, dE01_E10_x, dE01_E10_y) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InscribedCircleCenterEEELocal(pdC_x, pdC_y, pdR, 
						dE00_x, dE00_y, dE01_E10_x, dE01_E10_y,
						dE01_E10_x, dE01_E10_y, dE11_x, dE11_y,
						dE20_x, dE20_y, dE21_x, dE21_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InscribedCircleCenterEEEAdaptive(pdC_x, pdC_y, pdR, 
						   dE00_x, dE00_y, dE01_E10_x, dE01_E10_y,
						   dE01_E10_x, dE01_E10_y, dE11_x, dE11_y,
						   dE20_x, dE20_y, dE21_x, dE21_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InscribedCircleCenterEEEMPFUN(pdC_x, pdC_y, pdR, 
						dE00_x, dE00_y, dE01_E10_x, dE01_E10_y,
						dE01_E10_x, dE01_E10_y, dE11_x, dE11_y,
						dE20_x, dE20_y, dE21_x, dE21_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = InscribedCircleCenterEEEMPFUNFilter(pdC_x, pdC_y, pdR, 
						dE00_x, dE00_y, dE01_E10_x, dE01_E10_y,
						dE01_E10_x, dE01_E10_y, dE11_x, dE11_y,
						dE20_x, dE20_y, dE21_x, dE21_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InscribedCircleCenterEEEMPFUNNormalized(pdC_x, pdC_y, pdR, 
							  dE00_x, dE00_y, dE01_E10_x, dE01_E10_y,
							  dE01_E10_x, dE01_E10_y, dE11_x, dE11_y,
							  dE20_x, dE20_y, dE21_x, dE21_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }
  
  return eResult;
}

APM2D::InscribedCircleCenterType 
APM2D::InscribedCircleCenterEvEvE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				  DOUBLE dE00_x, DOUBLE dE00_y, 
				  DOUBLE dE01_E10_x, DOUBLE dE01_E10_y,
				  DOUBLE dE11_E20_x, DOUBLE dE11_E20_y,
				  DOUBLE dE21_x, DOUBLE dE21_y)
{
  ASSERT( pdC_x != NULL && pdC_y != NULL && pdR != NULL );

  APM2D::InscribedCircleCenterType eResult;

  eResult = APM2D::icctNull;
  
  //////////////////////////////////////////////////
  // Colinear Check

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
    case APM::atExact:
      {
	DOUBLE dSinTheta;
	dSinTheta = APM2D::SinTheta(dE00_x, dE00_y, dE01_E10_x, dE01_E10_y, dE11_E20_x, dE11_E20_y);
	if ( fabs(dSinTheta) <= s_cdEpsilonInscribedCircleCenterEvEvEParallel )
	  {
	    eResult = APM2D::InscribedCircleCentervEvE(pdC_x, pdC_y, pdR, 
						       dE01_E10_x, dE01_E10_y, 
						       dE11_E20_x, dE11_E20_y,
						       dE21_x, dE21_y);
	    if ( eResult != APM2D::icctNull )
	      {
		eResult = icctEcEvE;
	      }
	    return eResult;      
	  }
	else if ( dSinTheta < 0 )
	  {
	    *pdC_x = 0.0;
	    *pdC_y = 0.0;
	    *pdR = 0.0;
	    return APM2D::icctNull;
	  }

	dSinTheta = APM2D::SinTheta(dE01_E10_x, dE01_E10_y, dE11_E20_x, dE11_E20_y, dE21_x, dE21_y);
	if ( fabs(dSinTheta) <= s_cdEpsilonInscribedCircleCenterEvEvEParallel )
	  {
	    eResult = APM2D::InscribedCircleCenterEvEv(pdC_x, pdC_y, pdR, 
						       dE00_x, dE00_y,
						       dE01_E10_x, dE01_E10_y, 
						       dE11_E20_x, dE11_E20_y);
	    if ( eResult != APM2D::icctNull )
	      {
		eResult = icctEvEcE;
	      }
	    return eResult;      
	  }
	else if ( dSinTheta < 0 )
	  {
	    *pdC_x = 0.0;
	    *pdC_y = 0.0;
	    *pdR = 0.0;
	    return APM2D::icctNull;
	  }
      }
      break;
    case APM::atMPFUN:
    case APM::atMPFUNFilter:
    case APM::atMPFUNNormalized:
      {
	INT_32 iResult;
	iResult = OrientationVVV(dE00_x, dE00_y, dE01_E10_x, dE01_E10_y, dE11_E20_x, dE11_E20_y);
	if ( iResult == 0 )
	  {
	    eResult = APM2D::InscribedCircleCentervEvE(pdC_x, pdC_y, pdR, 
						       dE01_E10_x, dE01_E10_y, 
						       dE11_E20_x, dE11_E20_y,
						       dE21_x, dE21_y);
	    if ( eResult != APM2D::icctNull )
	      {
		eResult = icctEcEvE;
	      }
	    return eResult;      
	  }
	else if ( iResult < 0 )
	  {
	    *pdC_x = 0.0;
	    *pdC_y = 0.0;
	    *pdR = 0.0;
	    return APM2D::icctNull;
	  }
	
	iResult = OrientationVVV(dE01_E10_x, dE01_E10_y, dE11_E20_x, dE11_E20_y, dE21_x, dE21_y);
	if ( iResult == 0 )
	  {
	    eResult = APM2D::InscribedCircleCenterEvEv(pdC_x, pdC_y, pdR, 
						       dE00_x, dE00_y,
						       dE01_E10_x, dE01_E10_y, 
						       dE11_E20_x, dE11_E20_y);
	    if ( eResult != APM2D::icctNull )
	      {
		eResult = icctEvEcE;
	      }
	    return eResult;      
	  }
	else if ( iResult < 0 )
	  {
	    *pdC_x = 0.0;
	    *pdC_y = 0.0;
	    *pdR = 0.0;
	    return APM2D::icctNull;
	  }
      }
      break;
    default:
      {
	ASSERT( FALSE );
      }
      break;
    }

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InscribedCircleCenterEEELocal(pdC_x, pdC_y, pdR, 
						dE00_x, dE00_y, dE01_E10_x, dE01_E10_y,
						dE01_E10_x, dE01_E10_y, 
						dE11_E20_x, dE11_E20_y,
						dE11_E20_x, dE11_E20_y, dE21_x, dE21_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InscribedCircleCenterEEEAdaptive(pdC_x, pdC_y, pdR, 
						   dE00_x, dE00_y, dE01_E10_x, dE01_E10_y,
						   dE01_E10_x, dE01_E10_y, 
						   dE11_E20_x, dE11_E20_y,
						   dE11_E20_x, dE11_E20_y, dE21_x, dE21_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InscribedCircleCenterEEEMPFUN(pdC_x, pdC_y, pdR, 
						dE00_x, dE00_y, dE01_E10_x, dE01_E10_y,
						dE01_E10_x, dE01_E10_y, 
						dE11_E20_x, dE11_E20_y,
						dE11_E20_x, dE11_E20_y, dE21_x, dE21_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = InscribedCircleCenterEEEMPFUNFilter(pdC_x, pdC_y, pdR, 
						dE00_x, dE00_y, dE01_E10_x, dE01_E10_y,
						dE01_E10_x, dE01_E10_y, 
						dE11_E20_x, dE11_E20_y,
						dE11_E20_x, dE11_E20_y, dE21_x, dE21_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InscribedCircleCenterEEEMPFUNNormalized(pdC_x, pdC_y, pdR, 
							  dE00_x, dE00_y, dE01_E10_x, dE01_E10_y,
							  dE01_E10_x, dE01_E10_y, 
							  dE11_E20_x, dE11_E20_y,
							  dE11_E20_x, dE11_E20_y, dE21_x, dE21_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }
  
  return eResult;
}

APM2D::InscribedCircleCenterType 
APM2D::InscribedCircleCenterEvEvEv(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				   DOUBLE dE21_E00_x, DOUBLE dE21_E00_y, 
				   DOUBLE dE01_E10_x, DOUBLE dE01_E10_y,
				   DOUBLE dE11_E20_x, DOUBLE dE11_E20_y)
{
  ASSERT( pdC_x != NULL && pdC_y != NULL && pdR != NULL );

  APM2D::InscribedCircleCenterType eResult;

  eResult = APM2D::icctNull;
  
  //////////////////////////////////////////////////
  // CCW Check

  if ( !Counterclockwise(dE21_E00_x, dE21_E00_y, 
			 dE01_E10_x, dE01_E10_y, 
			 dE11_E20_x, dE11_E20_y) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InscribedCircleCenterEEELocal(pdC_x, pdC_y, pdR, 
						dE21_E00_x, dE21_E00_y, 
						dE01_E10_x, dE01_E10_y,
						dE01_E10_x, dE01_E10_y, 
						dE11_E20_x, dE11_E20_y,
						dE11_E20_x, dE11_E20_y, 
						dE21_E00_x, dE21_E00_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InscribedCircleCenterEEEAdaptive(pdC_x, pdC_y, pdR, 
						   dE21_E00_x, dE21_E00_y, 
						   dE01_E10_x, dE01_E10_y,
						   dE01_E10_x, dE01_E10_y, 
						   dE11_E20_x, dE11_E20_y,
						   dE11_E20_x, dE11_E20_y, 
						   dE21_E00_x, dE21_E00_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InscribedCircleCenterEEEMPFUN(pdC_x, pdC_y, pdR, 
						dE21_E00_x, dE21_E00_y, 
						dE01_E10_x, dE01_E10_y,
						dE01_E10_x, dE01_E10_y, 
						dE11_E20_x, dE11_E20_y,
						dE11_E20_x, dE11_E20_y, 
						dE21_E00_x, dE21_E00_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = InscribedCircleCenterEEEMPFUNFilter(pdC_x, pdC_y, pdR, 
						dE21_E00_x, dE21_E00_y, 
						dE01_E10_x, dE01_E10_y,
						dE01_E10_x, dE01_E10_y, 
						dE11_E20_x, dE11_E20_y,
						dE11_E20_x, dE11_E20_y, 
						dE21_E00_x, dE21_E00_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InscribedCircleCenterEEEMPFUNNormalized(pdC_x, pdC_y, pdR, 
							  dE21_E00_x, dE21_E00_y, 
							  dE01_E10_x, dE01_E10_y,
							  dE01_E10_x, dE01_E10_y, 
							  dE11_E20_x, dE11_E20_y,
							  dE11_E20_x, dE11_E20_y, 
							  dE21_E00_x, dE21_E00_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }
  
  return eResult;
}

static APM2D::InscribedCircleCenterType 
InscribedCircleCenterEEELocal(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			      DOUBLE dE00_x, DOUBLE dE00_y, 
			      DOUBLE dE01_x, DOUBLE dE01_y,
			      DOUBLE dE10_x, DOUBLE dE10_y, 
			      DOUBLE dE11_x, DOUBLE dE11_y,
			      DOUBLE dE20_x, DOUBLE dE20_y, 
			      DOUBLE dE21_x, DOUBLE dE21_y)
{
  DOUBLE pdC[3], dR;
  DOUBLE pppdV[3][2][2]; // Endpoints
  DOUBLE ppdE[3][2];     // Edge vectors
  DOUBLE ppdEdgeZone[3][2];
  DOUBLE dDen;
  DOUBLE dE0dotE0, dE1dotE1, dE2dotE2, dE0dotE1, dE0dotE2, dE1dotE2;
  DOUBLE dE0mag, dE1mag, dE2mag;
  DOUBLE dE0dotE00, dE0dotE01, dE1dotE10, dE1dotE11, dE2dotE20, dE2dotE21;
  DOUBLE dE0xE1, dE1xE2, dE2xE0, dE00xE01, dE20xE21;
  DOUBLE dE0E1minusE0dotE1, dE1E2minusE1dotE2;
  DOUBLE dE0dotC, dE1dotC, dE2dotC;
  DOUBLE dMagCE00, dMagCE01, dMagCE10, dMagCE11, dMagCE20, dMagCE21;
  DOUBLE ppdProjection[3][2];
  DOUBLE dCCWCheck;
  DOUBLE dSinTheta;
  //DOUBLE dCCW;
  UINT_32 i, j;
  
  // Translate V1 to the origin
  pppdV[0][0][0] = dE00_x - dE10_x;
  pppdV[0][0][1] = dE00_y - dE10_y;
  pppdV[0][1][0] = dE01_x - dE10_x;
  pppdV[0][1][1] = dE01_y - dE10_y;
  pppdV[1][0][0] = 0.0;
  pppdV[1][0][1] = 0.0;
  pppdV[1][1][0] = dE11_x - dE10_x;
  pppdV[1][1][1] = dE11_y - dE10_y;
  pppdV[2][0][0] = dE20_x - dE10_x;
  pppdV[2][0][1] = dE20_y - dE10_y;
  pppdV[2][1][0] = dE21_x - dE10_x;
  pppdV[2][1][1] = dE21_y - dE10_y;
  
  ppdE[0][0] = dE01_x - dE00_x;
  ppdE[0][1] = dE01_y - dE00_y;
  ppdE[1][0] = dE11_x - dE10_x;
  ppdE[1][1] = dE11_y - dE10_y;
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;
  

  dE0dotE0 = (ppdE[0][0]*ppdE[0][0]) + (ppdE[0][1]*ppdE[0][1]);
  dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);

  dE0mag = sqrt(dE0dotE0);
  dE1mag = sqrt(dE1dotE1);
  dE2mag = sqrt(dE2dotE2);

  dE0dotE1 = (ppdE[0][0]*ppdE[1][0]) + (ppdE[0][1]*ppdE[1][1]);
  dE1dotE2 = (ppdE[1][0]*ppdE[2][0]) + (ppdE[1][1]*ppdE[2][1]);
  dE0dotE2 = (ppdE[0][0]*ppdE[2][0]) + (ppdE[0][1]*ppdE[2][1]);

  dE0dotE00 = (ppdE[0][0]*pppdV[0][0][0]) + (ppdE[0][1]*pppdV[0][0][1]);
  dE0dotE01 = (ppdE[0][0]*pppdV[0][1][0]) + (ppdE[0][1]*pppdV[0][1][1]);
  dE1dotE10 = (ppdE[1][0]*pppdV[1][0][0]) + (ppdE[1][1]*pppdV[1][0][1]);
  dE1dotE11 = (ppdE[1][0]*pppdV[1][1][0]) + (ppdE[1][1]*pppdV[1][1][1]);
  dE2dotE20 = (ppdE[2][0]*pppdV[2][0][0]) + (ppdE[2][1]*pppdV[2][0][1]);
  dE2dotE21 = (ppdE[2][0]*pppdV[2][1][0]) + (ppdE[2][1]*pppdV[2][1][1]);

  dE0xE1 = (ppdE[0][0]*ppdE[1][1]) - (ppdE[0][1]*ppdE[1][0]);
  dE1xE2 = (ppdE[1][0]*ppdE[2][1]) - (ppdE[1][1]*ppdE[2][0]);
  dE2xE0 = (ppdE[2][0]*ppdE[0][1]) - (ppdE[2][1]*ppdE[0][0]);

  dE00xE01 = (pppdV[0][0][0]*pppdV[0][1][1]) - (pppdV[0][0][1]*pppdV[0][1][0]);
  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);

  dE0E1minusE0dotE1 = sqrt(dE0dotE0*dE1dotE1) - dE0dotE1;
  dE1E2minusE1dotE2 = sqrt(dE1dotE1*dE2dotE2) - dE1dotE2;
  //dE0E2minusE0dotE2 = sqrt(dE0dotE0*dE2dotE2) - dE0dotE2;

  dDen = dE0mag*dE1xE2 + dE1mag*dE2xE0 + dE2mag*dE0xE1;
  dSinTheta = dDen/sqrt(dE0dotE0*dE1dotE1*dE2dotE2);
  if ( fabs(dSinTheta) < s_cdEpsilonInscribedCircleCenterEEEDen )
    {
      // Colinear
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  //////////////////////////////////////////////////
  // Compute the center and radius

  pdC[0] = ( (dE1mag*ppdE[2][0] - dE2mag*ppdE[1][0])*dE00xE01
	     + (dE0mag*ppdE[1][0] - dE1mag*ppdE[0][0])*dE20xE21 ) / dDen;
  pdC[1] = ( (dE1mag*ppdE[2][1] - dE2mag*ppdE[1][1])*dE00xE01
	     + (dE0mag*ppdE[1][1] - dE1mag*ppdE[0][1])*dE20xE21 ) / dDen;

  dR = ( dE0xE1*dE20xE21 + dE1xE2*dE00xE01 ) / dDen;
  
  //////////////////////////////////////////////////
  // Check the Edge Zones

  dMagCE00 = ( (pdC[0] - pppdV[0][0][0])*(pdC[0] - pppdV[0][0][0]) + 
	       (pdC[1] - pppdV[0][0][1])*(pdC[1] - pppdV[0][0][1]) );
  dMagCE01 = ( (pdC[0] - pppdV[0][1][0])*(pdC[0] - pppdV[0][1][0]) + 
	       (pdC[1] - pppdV[0][1][1])*(pdC[1] - pppdV[0][1][1]) );
  dMagCE10 = ( (pdC[0] - pppdV[1][0][0])*(pdC[0] - pppdV[1][0][0]) + 
	       (pdC[1] - pppdV[1][0][1])*(pdC[1] - pppdV[1][0][1]) );
  dMagCE11 = ( (pdC[0] - pppdV[1][1][0])*(pdC[0] - pppdV[1][1][0]) + 
	       (pdC[1] - pppdV[1][1][1])*(pdC[1] - pppdV[1][1][1]) );
  dMagCE20 = ( (pdC[0] - pppdV[2][0][0])*(pdC[0] - pppdV[2][0][0]) + 
	       (pdC[1] - pppdV[2][0][1])*(pdC[1] - pppdV[2][0][1]) );
  dMagCE21 = ( (pdC[0] - pppdV[2][1][0])*(pdC[0] - pppdV[2][1][0]) + 
	       (pdC[1] - pppdV[2][1][1])*(pdC[1] - pppdV[2][1][1]) );

  if ( dMagCE00 <= s_cdEpsilonNullEdge )
    {
      ppdEdgeZone[0][0] = 0.0;
    }
  else
    {
      ppdEdgeZone[0][0] = -( ( dDen*dE0dotE00 
			       - (dE1mag*dE0dotE2 - dE2mag*dE0dotE1)*dE00xE01
			       + dE0mag*dE0E1minusE0dotE1*dE20xE21 ) /
			     ( dDen*sqrt(dE0dotE0*dMagCE00) ) );
    }
  if ( dMagCE01 <= s_cdEpsilonNullEdge )
    {
      ppdEdgeZone[0][1] = 0.0;
    }
  else
    {
      ppdEdgeZone[0][1] = ( ( dDen*dE0dotE01 
			      - (dE1mag*dE0dotE2 - dE2mag*dE0dotE1)*dE00xE01
			      + dE0mag*dE0E1minusE0dotE1*dE20xE21 ) /
			    ( dDen*sqrt(dE0dotE0*dMagCE01) ) );
    }
  if ( dMagCE10 <= s_cdEpsilonNullEdge )
    {
      ppdEdgeZone[1][0] = 0.0;
    }
  else
    {
      ppdEdgeZone[1][0] = -( ( dDen*dE1dotE10 
			       + dE1mag*dE1E2minusE1dotE2*dE00xE01
			       - dE1mag*dE0E1minusE0dotE1*dE20xE21 ) /
			     ( dDen*sqrt(dE1dotE1*dMagCE10) ) );
    }
  if ( dMagCE11 <= s_cdEpsilonNullEdge )
    {
      ppdEdgeZone[1][1] = 0.0;
    }
  else
    {
      ppdEdgeZone[1][1] = ( ( dDen*dE1dotE11 
			      + dE1mag*dE1E2minusE1dotE2*dE00xE01
			      - dE1mag*dE0E1minusE0dotE1*dE20xE21 ) /
			    ( dDen*sqrt(dE1dotE1*dMagCE11) ) );
    }
  if ( dMagCE20 <= s_cdEpsilonNullEdge )
    {
      ppdEdgeZone[2][0] = 0.0;
    }
  else
    {
      ppdEdgeZone[2][0] = -( ( dDen*dE2dotE20 
			       - dE2mag*dE1E2minusE1dotE2*dE00xE01 
			       - (dE0mag*dE1dotE2 - dE1mag*dE0dotE2)*dE20xE21 ) /
			     ( dDen*sqrt(dE2dotE2*dMagCE20) ) );
    }
  if ( dMagCE21 <= s_cdEpsilonNullEdge )
    {
      ppdEdgeZone[2][1] = 0.0;
    }
  else
    {
      ppdEdgeZone[2][1] = ( ( dDen*dE2dotE21 
			      - dE2mag*dE1E2minusE1dotE2*dE00xE01 
			      - (dE0mag*dE1dotE2 - dE1mag*dE0dotE2)*dE20xE21 ) /
			    ( dDen*sqrt(dE2dotE2*dMagCE21) ) );
    }

  for ( i = 0; i < 3; i++ )
    {
      for ( j = 0; j < 2; j++ )
	{
	  if ( ppdEdgeZone[i][j] < -s_cdEpsilonInscribedCircleCenterEEEEdgeZone )
	    {
#ifdef DEBUG_EPSILON_VALUES
	  if ( !( ppdEdgeZone[i][j] < -(s_cdEpsilonInscribedCircleCenterEEEEdgeZone*s_cdOrderOfMagnitude) ) )
	    {
	      fprintf(stderr, "  EEE: Edge Zone: %e < %e < %e\n",
		      s_cdEpsilonInscribedCircleCenterEEEEdgeZone*s_cdOrderOfMagnitude,
		      ppdEdgeZone[i][j],
		      s_cdEpsilonInscribedCircleCenterEEEEdgeZone);
	    }
#endif // DEBUG_EPSILON_VALUES	      
	      *pdC_x = 0.0;
	      *pdC_y = 0.0;
	      *pdR = 0.0;
	      return APM2D::icctNull;
	    }
	}
    }

  //////////////////////////////////////////////////
  // Check if the projection of the center on to the sites is in CCW order
  /*
  dCCW = ( - ( dE00xE01*dE00xE01*dE1E2minusE1dotE2*
	       ( 2.0*dE1dotE1*dE2dotE2*( dE0xE1*dE0dotE2 + dE2xE0*dE0dotE1 ) +
		 dE1xE2*( dE0dotE0*dE1dotE1*dE2dotE2 
			  - sqrt(dE1dotE1*dE2dotE2)*dE0dotE0*dE1dotE2
			  + sqrt(dE0dotE0*dE2dotE2)*dE1dotE1*dE0dotE2
			  + sqrt(dE0dotE0*dE1dotE1)*dE2dotE2*dE0dotE1 ) ) )
	   - ( dE20xE21*dE20xE21*dE0E1minusE0dotE1*
	       ( 2.0*dE0dotE0*dE1dotE1*( dE1xE2*dE0dotE2 + dE2xE0*dE1dotE2 ) +
		 dE0xE1*( dE0dotE0*dE1dotE1*dE2dotE2 
			  + sqrt(dE1dotE1*dE2dotE2)*dE0dotE0*dE1dotE2
			  + sqrt(dE0dotE0*dE2dotE2)*dE1dotE1*dE0dotE2
			  - sqrt(dE0dotE0*dE1dotE1)*dE2dotE2*dE0dotE1 ) ) )
	   + ( dE00xE01*dE20xE21*
	       ( dE1dotE1*dE2xE0*dDen*dDen
		 + ( dE1xE2*sqrt(dE0dotE0*dE1dotE1)*
		     ( dE0dotE0*dE1dotE1*dE2dotE2
		       - dE0dotE0*sqrt(dE1dotE1*dE2dotE2)*dE1dotE2
		       + dE1dotE1*sqrt(dE0dotE0*dE2dotE2)*dE0dotE2
		       - 2.0*sqrt(dE0dotE0*dE1dotE1)*dE0dotE2*dE1dotE2
		       - sqrt(dE1dotE1*dE2dotE2)*dE0dotE1*dE0dotE2
		       + sqrt(dE0dotE0*dE2dotE2)*dE0dotE1*dE1dotE2
		       + 2.0*dE1dotE1*dE0dotE2*dE0dotE2
		       - dE2dotE2*dE0dotE1*dE0dotE1 ) )
		 + ( dE2xE0*dE1dotE1*
		     ( - dE0dotE0*dE1dotE1*dE2dotE2
		       + 2.0*dE0dotE0*sqrt(dE1dotE1*dE2dotE2)*dE1dotE2
		       + 2.0*dE2dotE2*sqrt(dE0dotE0*dE1dotE1)*dE0dotE1
		       - dE0dotE0*dE1dotE2*dE1dotE2
		       + dE1dotE1*dE0dotE2*dE0dotE2
		       - dE2dotE2*dE0dotE1*dE0dotE1
		       - 2.0*sqrt(dE0dotE0*dE2dotE2)*dE0dotE1*dE1dotE2 ) )
		 + ( dE0xE1*sqrt(dE1dotE1*dE2dotE2)*
		     ( dE0dotE0*dE1dotE1*dE2dotE2
		       - dE2dotE2*sqrt(dE0dotE0*dE1dotE1)*dE0dotE1
		       + dE1dotE1*sqrt(dE0dotE0*dE2dotE2)*dE0dotE2
		       + sqrt(dE0dotE0*dE2dotE2)*dE0dotE1*dE1dotE2
		       - 2.0*sqrt(dE1dotE1*dE2dotE2)*dE0dotE1*dE0dotE2
		       - sqrt(dE0dotE0*dE1dotE1)*dE0dotE2*dE1dotE2
		       - dE0dotE0*dE1dotE2*dE1dotE2
		       + 2.0*dE1dotE1*dE0dotE2*dE0dotE2 ) ) ) )
	   ) / dE0dotE0*dE1dotE1*dE2dotE2*dDen*dDen;
  */

  dE0dotC = (ppdE[0][0]*pdC[0]) + (ppdE[0][1]*pdC[1]);
  dE1dotC = (ppdE[1][0]*pdC[0]) + (ppdE[1][1]*pdC[1]);
  dE2dotC = (ppdE[2][0]*pdC[0]) + (ppdE[2][1]*pdC[1]);

  ppdProjection[0][0] = (-dE00xE01*(-ppdE[0][1]) + dE0dotC*ppdE[0][0]) / dE0dotE0;
  ppdProjection[0][1] = (-dE00xE01*(ppdE[0][0]) + dE0dotC*ppdE[0][1]) / dE0dotE0;

  ppdProjection[1][0] = (dE1dotC*ppdE[1][0]) / dE1dotE1;
  ppdProjection[1][1] = (dE1dotC*ppdE[1][1]) / dE1dotE1;

  ppdProjection[2][0] = (-dE20xE21*(-ppdE[2][1]) + dE2dotC*ppdE[2][0]) / dE2dotE2;
  ppdProjection[2][1] = (-dE20xE21*(ppdE[2][0]) + dE2dotC*ppdE[2][1]) / dE2dotE2;

  dCCWCheck = ( ( (ppdProjection[2][0] - ppdProjection[1][0]) *
		  (ppdProjection[0][1] - ppdProjection[1][1]) ) -
		( (ppdProjection[2][1] - ppdProjection[1][1]) *
		  (ppdProjection[0][0] - ppdProjection[1][0]) ) );

  //fprintf(stderr, "  EEE: CCW:  Expanded %e ?= %e Projection\n", dCCW, dCCWCheck);

  if ( dCCWCheck < -s_cdEpsilonInscribedCircleCenterEEECCW )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  // Translate the answer back out away from the origin
  *pdC_x = dE10_x + pdC[0];
  *pdC_y = dE10_y + pdC[1];
  *pdR = dR;
  return APM2D::icctEEE;
}

static APM2D::InscribedCircleCenterType 
InscribedCircleCenterEEEAdaptive(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				 DOUBLE dE00_x, DOUBLE dE00_y, 
				 DOUBLE dE01_x, DOUBLE dE01_y,
				 DOUBLE dE10_x, DOUBLE dE10_y, 
				 DOUBLE dE11_x, DOUBLE dE11_y,
				 DOUBLE dE20_x, DOUBLE dE20_y, 
				 DOUBLE dE21_x, DOUBLE dE21_y)
{
  return InscribedCircleCenterEEELocal(pdC_x, pdC_y, pdR, 
				       dE00_x, dE00_y, dE01_x, dE01_y,
				       dE10_x, dE10_y, dE11_x, dE11_y,
				       dE20_x, dE20_y, dE21_x, dE21_y);
}

APM2D::EdgeZoneType APM2D::EdgeZoneEEE(DOUBLE dE00_x, DOUBLE dE00_y, 
				       DOUBLE dE01_x, DOUBLE dE01_y,
				       DOUBLE dE10_x, DOUBLE dE10_y, 
				       DOUBLE dE11_x, DOUBLE dE11_y,
				       DOUBLE dE20_x, DOUBLE dE20_y, 
				       DOUBLE dE21_x, DOUBLE dE21_y,
				       DOUBLE dE30_x, DOUBLE dE30_y,
				       DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::EdgeZoneType eResult;

  eResult = APM2D::eztMiddle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = EdgeZoneEEELocal(dE00_x, dE00_y, dE01_x, dE01_y,
				   dE10_x, dE10_y, dE11_x, dE11_y,
				   dE20_x, dE20_y, dE21_x, dE21_y,
				   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atExact:
      {
	eResult = EdgeZoneEEEAdaptive(dE00_x, dE00_y, dE01_x, dE01_y,
				      dE10_x, dE10_y, dE11_x, dE11_y,
				      dE20_x, dE20_y, dE21_x, dE21_y,
				      dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = EdgeZoneEEEMPFUN(dE00_x, dE00_y, dE01_x, dE01_y,
				   dE10_x, dE10_y, dE11_x, dE11_y,
				   dE20_x, dE20_y, dE21_x, dE21_y,
				   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = EdgeZoneEEEMPFUNFilter(dE00_x, dE00_y, dE01_x, dE01_y,
				   dE10_x, dE10_y, dE11_x, dE11_y,
				   dE20_x, dE20_y, dE21_x, dE21_y,
				   dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = EdgeZoneEEEMPFUNNormalized(dE00_x, dE00_y, dE01_x, dE01_y,
					     dE10_x, dE10_y, dE11_x, dE11_y,
					     dE20_x, dE20_y, dE21_x, dE21_y,
					     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::EdgeZoneType EdgeZoneEEELocal(DOUBLE dE00_x, DOUBLE dE00_y, 
					    DOUBLE dE01_x, DOUBLE dE01_y,
					    DOUBLE dE10_x, DOUBLE dE10_y, 
					    DOUBLE dE11_x, DOUBLE dE11_y,
					    DOUBLE dE20_x, DOUBLE dE20_y, 
					    DOUBLE dE21_x, DOUBLE dE21_y,
					    DOUBLE dE30_x, DOUBLE dE30_y,
					    DOUBLE dE31_x, DOUBLE dE31_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE pdC[2];
  DOUBLE pdEdgeZone[2];
  DOUBLE dE0dotE0, dE1dotE1, dE2dotE2, dE0dotE3, dE1dotE3, dE2dotE3;
  DOUBLE dE0mag, dE1mag, dE2mag;
  DOUBLE dE3dotE30, dE3dotE31;
  DOUBLE dE0xE1, dE1xE2, dE2xE0, dE00xE01, dE20xE21;
  DOUBLE dE3dotE3;
  DOUBLE dMagCE30, dMagCE31;
  DOUBLE dDen;
  APM2D::EdgeZoneType eResult;
  
  // Translate V1 to the origin
  pppdV[0][0][0] = dE00_x - dE10_x;
  pppdV[0][0][1] = dE00_y - dE10_y;
  pppdV[0][1][0] = dE01_x - dE10_x;
  pppdV[0][1][1] = dE01_y - dE10_y;
  pppdV[1][0][0] = 0.0;
  pppdV[1][0][1] = 0.0;
  pppdV[1][1][0] = dE11_x - dE10_x;
  pppdV[1][1][1] = dE11_y - dE10_y;
  pppdV[2][0][0] = dE20_x - dE10_x;
  pppdV[2][0][1] = dE20_y - dE10_y;
  pppdV[2][1][0] = dE21_x - dE10_x;
  pppdV[2][1][1] = dE21_y - dE10_y;
  pppdV[3][0][0] = dE30_x - dE10_x;
  pppdV[3][0][1] = dE30_y - dE10_y;
  pppdV[3][1][0] = dE31_x - dE10_x;
  pppdV[3][1][1] = dE31_y - dE10_y;
  
  ppdE[0][0] = dE01_x - dE00_x;
  ppdE[0][1] = dE01_y - dE00_y;
  ppdE[1][0] = dE11_x - dE10_x;
  ppdE[1][1] = dE11_y - dE10_y;
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;
  ppdE[3][0] = dE31_x - dE30_x;
  ppdE[3][1] = dE31_y - dE30_y;
  
  dE0dotE0 = (ppdE[0][0]*ppdE[0][0]) + (ppdE[0][1]*ppdE[0][1]);
  dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);
  dE3dotE3 = (ppdE[3][0]*ppdE[3][0]) + (ppdE[3][1]*ppdE[3][1]);

  dE0mag = sqrt(dE0dotE0);
  dE1mag = sqrt(dE1dotE1);
  dE2mag = sqrt(dE2dotE2);

  dE0dotE3 = (ppdE[0][0]*ppdE[3][0]) + (ppdE[0][1]*ppdE[3][1]);
  dE1dotE3 = (ppdE[1][0]*ppdE[3][0]) + (ppdE[1][1]*ppdE[3][1]);
  dE2dotE3 = (ppdE[2][0]*ppdE[3][0]) + (ppdE[2][1]*ppdE[3][1]);

  dE3dotE30 = (ppdE[3][0]*pppdV[3][0][0]) + (ppdE[3][1]*pppdV[3][0][1]);
  dE3dotE31 = (ppdE[3][0]*pppdV[3][1][0]) + (ppdE[3][1]*pppdV[3][1][1]);

  dE0xE1 = (ppdE[0][0]*ppdE[1][1]) - (ppdE[0][1]*ppdE[1][0]);
  dE1xE2 = (ppdE[1][0]*ppdE[2][1]) - (ppdE[1][1]*ppdE[2][0]);
  dE2xE0 = (ppdE[2][0]*ppdE[0][1]) - (ppdE[2][1]*ppdE[0][0]);

  dE00xE01 = (pppdV[0][0][0]*pppdV[0][1][1]) - (pppdV[0][0][1]*pppdV[0][1][0]);
  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);

  dDen = dE0mag*dE1xE2 + dE1mag*dE2xE0 + dE2mag*dE0xE1;
  DOUBLE dSinTheta = dDen/sqrt(dE0dotE0*dE1dotE1*dE2dotE2);
  ASSERT( !( fabs(dSinTheta) < s_cdEpsilonInscribedCircleCenterEEEDen ) );

  pdC[0] = ( (dE1mag*ppdE[2][0] - dE2mag*ppdE[1][0])*dE00xE01
	     + (dE0mag*ppdE[1][0] - dE1mag*ppdE[0][0])*dE20xE21 ) / dDen;
  pdC[1] = ( (dE1mag*ppdE[2][1] - dE2mag*ppdE[1][1])*dE00xE01
	     + (dE0mag*ppdE[1][1] - dE1mag*ppdE[0][1])*dE20xE21 ) / dDen;

  dMagCE30 = ( (pdC[0] - pppdV[3][0][0])*(pdC[0] - pppdV[3][0][0]) + 
	       (pdC[1] - pppdV[3][0][1])*(pdC[1] - pppdV[3][0][1]) );
  dMagCE31 = ( (pdC[0] - pppdV[3][1][0])*(pdC[0] - pppdV[3][1][0]) + 
	       (pdC[1] - pppdV[3][1][1])*(pdC[1] - pppdV[3][1][1]) );

  if ( dMagCE30 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[0] = 0.0;
    }
  else
    {
      pdEdgeZone[0] = ( ( dDen*dE3dotE30 
			  - (dE1mag*dE2dotE3 - dE2mag*dE1dotE3)*dE00xE01
			  - (dE0mag*dE1dotE3 - dE1mag*dE0dotE3)*dE20xE21 ) /
			( dDen*sqrt(dE3dotE3*dMagCE30) ) );
    }
  if ( dMagCE31 <= s_cdEpsilonNullEdge )
    {
      pdEdgeZone[1] = 0.0;
    }
  else
    {
      pdEdgeZone[1] = -( ( dDen*dE3dotE31 
			   - (dE1mag*dE2dotE3 - dE2mag*dE1dotE3)*dE00xE01
			   - (dE0mag*dE1dotE3 - dE1mag*dE0dotE3)*dE20xE21 ) /
			 ( dDen*sqrt(dE3dotE3*dMagCE31) ) );
    }
  
  if ( !( pdEdgeZone[0] < -s_cdEpsilonEdgeZoneEEEEdgeZone ) )
    {
      eResult = APM2D::eztStart;
    }
  else if ( !( pdEdgeZone[1] < -s_cdEpsilonEdgeZoneEEEEdgeZone ) )
    {
      eResult = APM2D::eztEnd;
    }
  else
    {
      eResult = APM2D::eztMiddle;
    }

  return eResult;
}

static APM2D::EdgeZoneType EdgeZoneEEEAdaptive(DOUBLE dE00_x, DOUBLE dE00_y, 
					       DOUBLE dE01_x, DOUBLE dE01_y,
					       DOUBLE dE10_x, DOUBLE dE10_y, 
					       DOUBLE dE11_x, DOUBLE dE11_y,
					       DOUBLE dE20_x, DOUBLE dE20_y, 
					       DOUBLE dE21_x, DOUBLE dE21_y,
					       DOUBLE dE30_x, DOUBLE dE30_y,
					       DOUBLE dE31_x, DOUBLE dE31_y)
{
  return EdgeZoneEEELocal(dE00_x, dE00_y, dE01_x, dE01_y,
			  dE10_x, dE10_y, dE11_x, dE11_y,
			  dE20_x, dE20_y, dE21_x, dE21_y,
			  dE30_x, dE30_y, dE31_x, dE31_y);
}

APM2D::CircleTestType APM2D::InCircleEEE_V(DOUBLE dE00_x, DOUBLE dE00_y, 
					   DOUBLE dE01_x, DOUBLE dE01_y,
					   DOUBLE dE10_x, DOUBLE dE10_y, 
					   DOUBLE dE11_x, DOUBLE dE11_y,
					   DOUBLE dE20_x, DOUBLE dE20_y, 
					   DOUBLE dE21_x, DOUBLE dE21_y,
					   DOUBLE dV3_x, DOUBLE dV3_y)
{
  APM2D::CircleTestType eResult;

  eResult = APM2D::cttOnCircle;

  // Check to see if the vertex is in the edge envelope
  if ( !Counterclockwise(dE00_x, dE00_y, 
			 dE01_x, dE01_y,
			 dV3_x, dV3_y) )
    {
      return APM2D::cttOutCircle;
    }

  if ( !Counterclockwise(dE10_x, dE10_y, 
			 dE11_x, dE11_y,
			 dV3_x, dV3_y) )
    {
      return APM2D::cttOutCircle;
    }

  if ( !Counterclockwise(dE20_x, dE20_y, 
			 dE21_x, dE21_y,
			 dV3_x, dV3_y) )
    {
      return APM2D::cttOutCircle;
    }


  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InCircleEEE_VLocal(dE00_x, dE00_y, dE01_x, dE01_y,
				     dE10_x, dE10_y, dE11_x, dE11_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dV3_x, dV3_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InCircleEEE_VAdaptive(dE00_x, dE00_y, dE01_x, dE01_y,
					dE10_x, dE10_y, dE11_x, dE11_y,
					dE20_x, dE20_y, dE21_x, dE21_y,
					dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InCircleEEE_VMPFUN(dE00_x, dE00_y, dE01_x, dE01_y,
				     dE10_x, dE10_y, dE11_x, dE11_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = InCircleEEE_VMPFUNFilter(dE00_x, dE00_y, dE01_x, dE01_y,
				     dE10_x, dE10_y, dE11_x, dE11_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dV3_x, dV3_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InCircleEEE_VMPFUNNormalized(dE00_x, dE00_y, dE01_x, dE01_y,
					       dE10_x, dE10_y, dE11_x, dE11_y,
					       dE20_x, dE20_y, dE21_x, dE21_y,
					       dV3_x, dV3_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::CircleTestType InCircleEEE_VLocal(DOUBLE dE00_x, DOUBLE dE00_y, 
						DOUBLE dE01_x, DOUBLE dE01_y,
						DOUBLE dE10_x, DOUBLE dE10_y, 
						DOUBLE dE11_x, DOUBLE dE11_y,
						DOUBLE dE20_x, DOUBLE dE20_y, 
						DOUBLE dE21_x, DOUBLE dE21_y,
						DOUBLE dV3_x, DOUBLE dV3_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE dE0dotE0, dE1dotE1, dE2dotE2, dV3dotV3, dE0dotV3, dE1dotV3, dE2dotV3;
  DOUBLE dE0dotE1, dE1dotE2, dE0dotE2;
  DOUBLE dE0mag, dE1mag, dE2mag;
  DOUBLE dE0xE1, dE1xE2, dE2xE0, dE00xE01, dE20xE21;
  DOUBLE dE1E2minusE1dotE2, dE0E1minusE0dotE1, dE0E2plusE0dotE2;
  DOUBLE dDen;
  DOUBLE dResult;
  APM2D::CircleTestType eResult;
  
  // Translate V1 to the origin
  memset(pppdV, 0, sizeof(pppdV));
  pppdV[0][0][0] = dE00_x - dE10_x;
  pppdV[0][0][1] = dE00_y - dE10_y;
  pppdV[0][1][0] = dE01_x - dE10_x;
  pppdV[0][1][1] = dE01_y - dE10_y;
  pppdV[1][0][0] = 0.0;
  pppdV[1][0][1] = 0.0;
  pppdV[1][1][0] = dE11_x - dE10_x;
  pppdV[1][1][1] = dE11_y - dE10_y;
  pppdV[2][0][0] = dE20_x - dE10_x;
  pppdV[2][0][1] = dE20_y - dE10_y;
  pppdV[2][1][0] = dE21_x - dE10_x;
  pppdV[2][1][1] = dE21_y - dE10_y;
  pppdV[3][0][0] = dV3_x - dE10_x;
  pppdV[3][0][1] = dV3_y - dE10_y;
  
  memset(ppdE, 0, sizeof(ppdE));
  ppdE[0][0] = dE01_x - dE00_x;
  ppdE[0][1] = dE01_y - dE00_y;
  ppdE[1][0] = dE11_x - dE10_x;
  ppdE[1][1] = dE11_y - dE10_y;
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;
  
  dE0dotE0 = (ppdE[0][0]*ppdE[0][0]) + (ppdE[0][1]*ppdE[0][1]);
  dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);
  dV3dotV3 = (pppdV[3][0][0]*pppdV[3][0][0]) + (pppdV[3][0][1]*pppdV[3][0][1]);

  dE0mag = sqrt(dE0dotE0);
  dE1mag = sqrt(dE1dotE1);
  dE2mag = sqrt(dE2dotE2);

  dE0dotV3 = (ppdE[0][0]*pppdV[3][0][0]) + (ppdE[0][1]*pppdV[3][0][1]);
  dE1dotV3 = (ppdE[1][0]*pppdV[3][0][0]) + (ppdE[1][1]*pppdV[3][0][1]);
  dE2dotV3 = (ppdE[2][0]*pppdV[3][0][0]) + (ppdE[2][1]*pppdV[3][0][1]);

  dE0dotE1 = (ppdE[0][0]*ppdE[1][0]) + (ppdE[0][1]*ppdE[1][1]);
  dE1dotE2 = (ppdE[1][0]*ppdE[2][0]) + (ppdE[1][1]*ppdE[2][1]);
  dE0dotE2 = (ppdE[0][0]*ppdE[2][0]) + (ppdE[0][1]*ppdE[2][1]);

  dE0xE1 = (ppdE[0][0]*ppdE[1][1]) - (ppdE[0][1]*ppdE[1][0]);
  dE1xE2 = (ppdE[1][0]*ppdE[2][1]) - (ppdE[1][1]*ppdE[2][0]);
  dE2xE0 = (ppdE[2][0]*ppdE[0][1]) - (ppdE[2][1]*ppdE[0][0]);

  dE00xE01 = (pppdV[0][0][0]*pppdV[0][1][1]) - (pppdV[0][0][1]*pppdV[0][1][0]);
  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);

  dE0E1minusE0dotE1 = sqrt(dE0dotE0*dE1dotE1) - dE0dotE1;
  dE1E2minusE1dotE2 = sqrt(dE1dotE1*dE2dotE2) - dE1dotE2;
  dE0E2plusE0dotE2 = sqrt(dE0dotE0*dE2dotE2) + dE0dotE2;

  dDen = dE0mag*dE1xE2 + dE1mag*dE2xE0 + dE2mag*dE0xE1;
  DOUBLE dSinTheta = dDen/sqrt(dE0dotE0*dE1dotE1*dE2dotE2);
  ASSERT( !( fabs(dSinTheta) < s_cdEpsilonInscribedCircleCenterEEEDen ) );

  dResult = ( dDen*( dDen*dV3dotV3 
		     + 2.0*( dE0dotV3*dE1mag*dE20xE21 + dE1dotV3*( dE2mag*dE00xE01 - dE0mag*dE20xE21 ) - dE2dotV3*dE1mag*dE00xE01 ) )
	      + dE00xE01*dE00xE01*dE1E2minusE1dotE2*dE1E2minusE1dotE2
	      + 2.0*dE00xE01*dE20xE21*( dE1mag*( dE0mag*dE1dotE2 + dE2mag*dE0dotE1 ) - dE1dotE1*dE0E2plusE0dotE2 - dE0xE1*dE1xE2 )
	      + dE20xE21*dE20xE21*dE0E1minusE0dotE1*dE0E1minusE0dotE1 ) / (dDen*dDen);
  
  if ( dResult < -s_cdEpsilonInCircleEEE_VOnCircle )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( dResult > s_cdEpsilonInCircleEEE_VOnCircle )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

static APM2D::CircleTestType InCircleEEE_VAdaptive(DOUBLE dE00_x, DOUBLE dE00_y, 
						   DOUBLE dE01_x, DOUBLE dE01_y,
						   DOUBLE dE10_x, DOUBLE dE10_y, 
						   DOUBLE dE11_x, DOUBLE dE11_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   DOUBLE dV3_x, DOUBLE dV3_y)
{
  return InCircleEEE_VLocal(dE00_x, dE00_y, dE01_x, dE01_y,
			    dE10_x, dE10_y, dE11_x, dE11_y,
			    dE20_x, dE20_y, dE21_x, dE21_y,
			    dV3_x, dV3_y);
}

APM2D::CircleTestType APM2D::InCircleEEE_E(DOUBLE dE00_x, DOUBLE dE00_y, 
					   DOUBLE dE01_x, DOUBLE dE01_y,
					   DOUBLE dE10_x, DOUBLE dE10_y, 
					   DOUBLE dE11_x, DOUBLE dE11_y,
					   DOUBLE dE20_x, DOUBLE dE20_y, 
					   DOUBLE dE21_x, DOUBLE dE21_y,
					   DOUBLE dE30_x, DOUBLE dE30_y,
					   DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::CircleTestType eResult;

  eResult = APM2D::cttOnCircle;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
    case APM::atLocal:
      {
	eResult = InCircleEEE_ELocal(dE00_x, dE00_y, dE01_x, dE01_y,
				     dE10_x, dE10_y, dE11_x, dE11_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atExact:
      {
	eResult = InCircleEEE_EAdaptive(dE00_x, dE00_y, dE01_x, dE01_y,
					dE10_x, dE10_y, dE11_x, dE11_y,
					dE20_x, dE20_y, dE21_x, dE21_y,
					dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUN:
      {
	eResult = InCircleEEE_EMPFUN(dE00_x, dE00_y, dE01_x, dE01_y,
				     dE10_x, dE10_y, dE11_x, dE11_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNFilter:
      {
	eResult = InCircleEEE_EMPFUNFilter(dE00_x, dE00_y, dE01_x, dE01_y,
				     dE10_x, dE10_y, dE11_x, dE11_y,
				     dE20_x, dE20_y, dE21_x, dE21_y,
				     dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    case APM::atMPFUNNormalized:
      {
	eResult = InCircleEEE_EMPFUNNormalized(dE00_x, dE00_y, dE01_x, dE01_y,
					       dE10_x, dE10_y, dE11_x, dE11_y,
					       dE20_x, dE20_y, dE21_x, dE21_y,
					       dE30_x, dE30_y, dE31_x, dE31_y);
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }  

  return eResult;
}

static APM2D::CircleTestType InCircleEEE_ELocal(DOUBLE dE00_x, DOUBLE dE00_y, 
						DOUBLE dE01_x, DOUBLE dE01_y,
						DOUBLE dE10_x, DOUBLE dE10_y, 
						DOUBLE dE11_x, DOUBLE dE11_y,
						DOUBLE dE20_x, DOUBLE dE20_y, 
						DOUBLE dE21_x, DOUBLE dE21_y,
						DOUBLE dE30_x, DOUBLE dE30_y,
						DOUBLE dE31_x, DOUBLE dE31_y)
{
  DOUBLE pppdV[4][2][2]; // Endpoints
  DOUBLE ppdE[4][2];     // Edge vectors
  DOUBLE dE0dotE0, dE1dotE1, dE2dotE2, dE3dotE3;
  DOUBLE dE0mag, dE1mag, dE2mag;
  DOUBLE dE0xE1, dE1xE2, dE2xE0;
  DOUBLE dE0xE3, dE1xE3, dE2xE3;
  DOUBLE dE00xE01, dE20xE21, dE30xE31;
  DOUBLE dA, dB;
  DOUBLE dDen;
  DOUBLE dResult;
  APM2D::CircleTestType eResult;
  
  // Translate V1 to the origin
  pppdV[0][0][0] = dE00_x - dE10_x;
  pppdV[0][0][1] = dE00_y - dE10_y;
  pppdV[0][1][0] = dE01_x - dE10_x;
  pppdV[0][1][1] = dE01_y - dE10_y;
  pppdV[1][0][0] = 0.0;
  pppdV[1][0][1] = 0.0;
  pppdV[1][1][0] = dE11_x - dE10_x;
  pppdV[1][1][1] = dE11_y - dE10_y;
  pppdV[2][0][0] = dE20_x - dE10_x;
  pppdV[2][0][1] = dE20_y - dE10_y;
  pppdV[2][1][0] = dE21_x - dE10_x;
  pppdV[2][1][1] = dE21_y - dE10_y;
  pppdV[3][0][0] = dE30_x - dE10_x;
  pppdV[3][0][1] = dE30_y - dE10_y;
  pppdV[3][1][0] = dE31_x - dE10_x;
  pppdV[3][1][1] = dE31_y - dE10_y;
  
  ppdE[0][0] = dE01_x - dE00_x;
  ppdE[0][1] = dE01_y - dE00_y;
  ppdE[1][0] = dE11_x - dE10_x;
  ppdE[1][1] = dE11_y - dE10_y;
  ppdE[2][0] = dE21_x - dE20_x;
  ppdE[2][1] = dE21_y - dE20_y;
  ppdE[3][0] = dE31_x - dE30_x;
  ppdE[3][1] = dE31_y - dE30_y;
  
  dE0dotE0 = (ppdE[0][0]*ppdE[0][0]) + (ppdE[0][1]*ppdE[0][1]);
  dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
  dE2dotE2 = (ppdE[2][0]*ppdE[2][0]) + (ppdE[2][1]*ppdE[2][1]);
  dE3dotE3 = (ppdE[3][0]*ppdE[3][0]) + (ppdE[3][1]*ppdE[3][1]);
  
  dE0mag = sqrt(dE0dotE0);
  dE1mag = sqrt(dE1dotE1);
  dE2mag = sqrt(dE2dotE2);
  
  dE0xE1 = (ppdE[0][0]*ppdE[1][1]) - (ppdE[0][1]*ppdE[1][0]);
  dE1xE2 = (ppdE[1][0]*ppdE[2][1]) - (ppdE[1][1]*ppdE[2][0]);
  dE2xE0 = (ppdE[2][0]*ppdE[0][1]) - (ppdE[2][1]*ppdE[0][0]);
    
  dE0xE3 = (ppdE[0][0]*ppdE[3][1]) - (ppdE[0][1]*ppdE[3][0]);
  dE1xE3 = (ppdE[1][0]*ppdE[3][1]) - (ppdE[1][1]*ppdE[3][0]);
  dE2xE3 = (ppdE[2][0]*ppdE[3][1]) - (ppdE[2][1]*ppdE[3][0]);
    
  dE00xE01 = (pppdV[0][0][0]*pppdV[0][1][1]) - (pppdV[0][0][1]*pppdV[0][1][0]);
  dE20xE21 = (pppdV[2][0][0]*pppdV[2][1][1]) - (pppdV[2][0][1]*pppdV[2][1][0]);
  dE30xE31 = (pppdV[3][0][0]*pppdV[3][1][1]) - (pppdV[3][0][1]*pppdV[3][1][0]);
  
  dA = ( ( sqrt(dE2dotE2)*dE1xE3 - sqrt(dE1dotE1)*dE2xE3 )*dE00xE01
	 + ( sqrt(dE1dotE1)*dE0xE3 - sqrt(dE0dotE0)*dE1xE3 )*dE20xE21 );
  dB = ( dE0xE1*dE20xE21 + dE1xE2*dE00xE01 );
  dB = dB*dB;

  dDen = dE0mag*dE1xE2 + dE1mag*dE2xE0 + dE2mag*dE0xE1;
  DOUBLE dSinTheta = dDen/sqrt(dE0dotE0*dE1dotE1*dE2dotE2);
  ASSERT( !( fabs(dSinTheta) < s_cdEpsilonInscribedCircleCenterEEEDen ) );
  
  dResult = ( dA*dA
	      + 2.0*dE30xE31*dDen*dA 
	      + dE30xE31*dE30xE31*dDen*dDen
	      - dE3dotE3*dB ) / (dE3dotE3*dDen*dDen);

  if ( dResult < -s_cdEpsilonInCircleEEE_EOnCircle )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( dResult > s_cdEpsilonInCircleEEE_EOnCircle )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

static APM2D::CircleTestType InCircleEEE_EAdaptive(DOUBLE dE00_x, DOUBLE dE00_y, 
						   DOUBLE dE01_x, DOUBLE dE01_y,
						   DOUBLE dE10_x, DOUBLE dE10_y, 
						   DOUBLE dE11_x, DOUBLE dE11_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   DOUBLE dE30_x, DOUBLE dE30_y,
						   DOUBLE dE31_x, DOUBLE dE31_y)
{
  return InCircleEEE_ELocal(dE00_x, dE00_y, dE01_x, dE01_y,
			    dE10_x, dE10_y, dE11_x, dE11_y,
			    dE20_x, dE20_y, dE21_x, dE21_y,
			    dE30_x, dE30_y, dE31_x, dE31_y);
}

//////////////////////////////////////////////////////////////////////
// In Circle Functions

DOUBLE APM2D::OrientationCircleValue(DOUBLE dV0_x, DOUBLE dV0_y,
				     DOUBLE dV1_x, DOUBLE dV1_y,
				     DOUBLE dV2_x, DOUBLE dV2_y,
				     DOUBLE dV3_x, DOUBLE dV3_y)
{
  DOUBLE d;

  d = 0.0;

  switch ( APM::GetAccuracyType() )
    {
    case APM::atGlobal:
      {
	DOUBLE ppdV[4][3];

	ppdV[0][0] = dV0_x;
	ppdV[0][1] = dV0_y;
	ppdV[0][2] = (ppdV[0][0]*ppdV[0][0]) + (ppdV[0][1]*ppdV[0][1]);
	ppdV[1][0] = dV1_x;
	ppdV[1][1] = dV1_y;
	ppdV[1][2] = (ppdV[1][0]*ppdV[1][0]) + (ppdV[1][1]*ppdV[1][1]);
	ppdV[2][0] = dV2_x;
	ppdV[2][1] = dV2_y;
	ppdV[2][2] = (ppdV[2][0]*ppdV[2][0]) + (ppdV[2][1]*ppdV[2][1]);
	ppdV[3][0] = dV3_x;
	ppdV[3][1] = dV3_y;
	ppdV[3][2] = (ppdV[3][0]*ppdV[3][0]) + (ppdV[3][1]*ppdV[3][1]);

	d = ( - ( ( (ppdV[1][0]*ppdV[2][1]*ppdV[3][2]) + 
		    (ppdV[2][0]*ppdV[3][1]*ppdV[1][2]) + 
		    (ppdV[3][0]*ppdV[1][1]*ppdV[2][2]) ) -
		  ( (ppdV[1][0]*ppdV[3][1]*ppdV[2][2]) + 
		    (ppdV[2][0]*ppdV[1][1]*ppdV[3][2]) + 
		    (ppdV[3][0]*ppdV[2][1]*ppdV[1][2]) ) ) 
	      + ( ( (ppdV[0][0]*ppdV[2][1]*ppdV[3][2]) + 
		    (ppdV[2][0]*ppdV[3][1]*ppdV[0][2]) + 
		    (ppdV[3][0]*ppdV[0][1]*ppdV[2][2]) ) -
		  ( (ppdV[0][0]*ppdV[3][1]*ppdV[2][2]) + 
		    (ppdV[2][0]*ppdV[0][1]*ppdV[3][2]) + 
		    (ppdV[3][0]*ppdV[2][1]*ppdV[0][2]) ) ) 
	      - ( ( (ppdV[0][0]*ppdV[1][1]*ppdV[3][2]) + 
		    (ppdV[1][0]*ppdV[3][1]*ppdV[0][2]) + 
		    (ppdV[3][0]*ppdV[0][1]*ppdV[1][2]) ) -
		  ( (ppdV[0][0]*ppdV[3][1]*ppdV[1][2]) + 
		    (ppdV[1][0]*ppdV[0][1]*ppdV[3][2]) + 
		    (ppdV[3][0]*ppdV[1][1]*ppdV[0][2]) ) )
	      + ( ( (ppdV[0][0]*ppdV[1][1]*ppdV[2][2]) + 
		    (ppdV[1][0]*ppdV[2][1]*ppdV[0][2]) + 
		    (ppdV[2][0]*ppdV[0][1]*ppdV[1][2]) ) -
		  ( (ppdV[0][0]*ppdV[2][1]*ppdV[1][2]) + 
		    (ppdV[1][0]*ppdV[0][1]*ppdV[2][2]) + 
		    (ppdV[2][0]*ppdV[1][1]*ppdV[0][2]) ) ) );
      }
    break;

    case APM::atLocal:
      {
	DOUBLE ppdV[3][3];

	ppdV[0][0] = dV0_x - dV3_x;
	ppdV[0][1] = dV0_y - dV3_y;
	ppdV[0][2] = (ppdV[0][0]*ppdV[0][0]) + (ppdV[0][1]*ppdV[0][1]);
	ppdV[1][0] = dV1_x - dV3_x;
	ppdV[1][1] = dV1_y - dV3_y;
	ppdV[1][2] = (ppdV[1][0]*ppdV[1][0]) + (ppdV[1][1]*ppdV[1][1]);
	ppdV[2][0] = dV2_x - dV3_x;
	ppdV[2][1] = dV2_y - dV3_y;
	ppdV[2][2] = (ppdV[2][0]*ppdV[2][0]) + (ppdV[2][1]*ppdV[2][1]);

	d = ( ( (ppdV[0][0]*ppdV[1][1]*ppdV[2][2]) + 
		(ppdV[1][0]*ppdV[2][1]*ppdV[0][2]) + 
		(ppdV[2][0]*ppdV[0][1]*ppdV[1][2]) ) -
	      ( (ppdV[0][0]*ppdV[2][1]*ppdV[1][2]) + 
		(ppdV[1][0]*ppdV[0][1]*ppdV[2][2]) + 
		(ppdV[2][0]*ppdV[1][1]*ppdV[0][2]) ) );
      }
    break;
	
    case APM::atExact:
    case APM::atMPFUN:
    case APM::atMPFUNFilter:
    case APM::atMPFUNNormalized:
      {
	double ppdV0[2], ppdV1[2], ppdV2[2], ppdV3[2];
	  
	ppdV0[0] = dV0_x;
	ppdV0[1] = dV0_y;
	ppdV1[0] = dV1_x;
	ppdV1[1] = dV1_y;
	ppdV2[0] = dV2_x;
	ppdV2[1] = dV2_y;
	ppdV3[0] = dV3_x;
	ppdV3[1] = dV3_y;
	  
	d = incircleexact(ppdV0, ppdV1, ppdV2, ppdV3);
      }
    break;

    default:
      ASSERT( FALSE );
      break;
    }

  return d;
}

APM2D::CircleTestType APM2D::OrientationCircle(DOUBLE dV0_x, DOUBLE dV0_y,
					       DOUBLE dV1_x, DOUBLE dV1_y,
					       DOUBLE dV2_x, DOUBLE dV2_y,
					       DOUBLE dV3_x, DOUBLE dV3_y)
{
  DOUBLE d;

  d = OrientationCircle(dV0_x, dV0_y,
			dV1_x, dV1_y,
			dV2_x, dV2_y,
			dV3_x, dV3_y);

  if ( d == 0.0 )
    {
      return APM2D::cttOnCircle;
    }
  else if ( d > 0.0 )
    {
      return APM2D::cttOutCircle;
    }
  else
    {
      ASSERT( d < 0.0 );
      return APM2D::cttInCircle;
    }
}

//////////////////////////////////////////////////////////////////////
// MPFUN Filter Counters

UINT_32 APM2D::m_uiEdgeZoneMPFUNFilter = 0;
UINT_32 APM2D::m_uiEdgeZoneMPFUN = 0;

UINT_32 APM2D::m_uiDistanceCompareV_VVMPFUNFilter = 0;
UINT_32 APM2D::m_uiDistanceCompareV_VVMPFUN = 0;

UINT_32 APM2D::m_uiDistanceCompareV_VLMPFUNFilter = 0;
UINT_32 APM2D::m_uiDistanceCompareV_VLMPFUN = 0;

UINT_32 APM2D::m_uiDistanceCompareV_LLMPFUNFilter = 0;
UINT_32 APM2D::m_uiDistanceCompareV_LLMPFUN = 0;

UINT_32 APM2D::m_uiDistanceCompareEvv_VVMPFUNFilter = 0;
UINT_32 APM2D::m_uiDistanceCompareEvv_VVMPFUN = 0;

UINT_32 APM2D::m_uiDistanceCompareEvv_VLMPFUNFilter = 0;
UINT_32 APM2D::m_uiDistanceCompareEvv_VLMPFUN = 0;

UINT_32 APM2D::m_uiDistanceCompareEvv_LLMPFUNFilter = 0;
UINT_32 APM2D::m_uiDistanceCompareEvv_LLMPFUN = 0;

UINT_32 APM2D::m_uiDistanceCompareElv_VVMPFUNFilter = 0;
UINT_32 APM2D::m_uiDistanceCompareElv_VVMPFUN = 0;

UINT_32 APM2D::m_uiDistanceCompareElv_VLMPFUNFilter = 0;
UINT_32 APM2D::m_uiDistanceCompareElv_VLMPFUN = 0;

UINT_32 APM2D::m_uiDistanceCompareL_VVMPFUNFilter = 0;
UINT_32 APM2D::m_uiDistanceCompareL_VVMPFUN = 0;

UINT_32 APM2D::m_uiOrientationExEMPFUNFilter = 0;
UINT_32 APM2D::m_uiOrientationExEMPFUN = 0;

UINT_32 APM2D::m_uiOrientationEdotEMPFUNFilter = 0;
UINT_32 APM2D::m_uiOrientationEdotEMPFUN = 0;

UINT_32 APM2D::m_uiInscribedCircleCenterVVVMPFUNFilter = 0;
UINT_32 APM2D::m_uiInscribedCircleCenterVVVMPFUN = 0;
UINT_32 APM2D::m_uiInscribedCircleCenterVVVMPFUNFinal = 0;

UINT_32 APM2D::m_uiEdgeZoneVVVMPFUNFilter = 0;
UINT_32 APM2D::m_uiEdgeZoneVVVMPFUN = 0;

UINT_32 APM2D::m_uiInCircleVVV_VMPFUNFilter = 0;
UINT_32 APM2D::m_uiInCircleVVV_VMPFUN = 0;

UINT_32 APM2D::m_uiInCircleVVV_EMPFUNFilter = 0;
UINT_32 APM2D::m_uiInCircleVVV_EMPFUN = 0;

UINT_32 APM2D::m_uiInscribedCircleCenterVVEMPFUNFilter = 0;
UINT_32 APM2D::m_uiInscribedCircleCenterVVEMPFUN = 0;
UINT_32 APM2D::m_uiInscribedCircleCenterVVEMPFUNFinal = 0;

UINT_32 APM2D::m_uiEdgeZoneVVEMPFUNFilter = 0;
UINT_32 APM2D::m_uiEdgeZoneVVEMPFUN = 0;

UINT_32 APM2D::m_uiInCircleVVE_VMPFUNFilter = 0;
UINT_32 APM2D::m_uiInCircleVVE_VMPFUN = 0;

UINT_32 APM2D::m_uiInCircleVVE_EMPFUNFilter = 0;
UINT_32 APM2D::m_uiInCircleVVE_EMPFUN = 0;

UINT_32 APM2D::m_uiInscribedCircleCenterVVEAntiparallelMPFUNFilter = 0;
UINT_32 APM2D::m_uiInscribedCircleCenterVVEAntiparallelMPFUN = 0;
UINT_32 APM2D::m_uiInscribedCircleCenterVVEAntiparallelMPFUNFinal = 0;

UINT_32 APM2D::m_uiEdgeZoneVVEAntiparallelMPFUNFilter = 0;
UINT_32 APM2D::m_uiEdgeZoneVVEAntiparallelMPFUN = 0;

UINT_32 APM2D::m_uiInCircleVVE_VAntiparallelMPFUNFilter = 0;
UINT_32 APM2D::m_uiInCircleVVE_VAntiparallelMPFUN = 0;

UINT_32 APM2D::m_uiInCircleVVE_EAntiparallelMPFUNFilter = 0;
UINT_32 APM2D::m_uiInCircleVVE_EAntiparallelMPFUN = 0;

UINT_32 APM2D::m_uiInscribedCircleCenterVvEMPFUNFilter = 0;
UINT_32 APM2D::m_uiInscribedCircleCenterVvEMPFUN = 0;
UINT_32 APM2D::m_uiInscribedCircleCenterVvEMPFUNFinal = 0;

UINT_32 APM2D::m_uiEdgeZoneVvEMPFUNFilter = 0;
UINT_32 APM2D::m_uiEdgeZoneVvEMPFUN = 0;

UINT_32 APM2D::m_uiInCircleVvE_VMPFUNFilter = 0;
UINT_32 APM2D::m_uiInCircleVvE_VMPFUN = 0;

UINT_32 APM2D::m_uiInCircleVvE_EMPFUNFilter = 0;
UINT_32 APM2D::m_uiInCircleVvE_EMPFUN = 0;

UINT_32 APM2D::m_uiInscribedCircleCenterEvVMPFUNFilter = 0;
UINT_32 APM2D::m_uiInscribedCircleCenterEvVMPFUN = 0;
UINT_32 APM2D::m_uiInscribedCircleCenterEvVMPFUNFinal = 0;

UINT_32 APM2D::m_uiEdgeZoneEvVMPFUNFilter = 0;
UINT_32 APM2D::m_uiEdgeZoneEvVMPFUN = 0;

UINT_32 APM2D::m_uiInCircleEvV_VMPFUNFilter = 0;
UINT_32 APM2D::m_uiInCircleEvV_VMPFUN = 0;

UINT_32 APM2D::m_uiInCircleEvV_EMPFUNFilter = 0;
UINT_32 APM2D::m_uiInCircleEvV_EMPFUN = 0;

UINT_32 APM2D::m_uiInscribedCircleCenterVEEMPFUNFilter = 0;
UINT_32 APM2D::m_uiInscribedCircleCenterVEEMPFUN = 0;
UINT_32 APM2D::m_uiInscribedCircleCenterVEEMPFUNFinal = 0;

UINT_32 APM2D::m_uiEdgeZoneVEEMPFUNFilter = 0;
UINT_32 APM2D::m_uiEdgeZoneVEEMPFUN = 0;

UINT_32 APM2D::m_uiInCircleVEE_VMPFUNFilter = 0;
UINT_32 APM2D::m_uiInCircleVEE_VMPFUN = 0;

UINT_32 APM2D::m_uiInCircleVEE_EMPFUNFilter = 0;
UINT_32 APM2D::m_uiInCircleVEE_EMPFUN = 0;

UINT_32 APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUNFilter = 0;
UINT_32 APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUN = 0;
UINT_32 APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUNFinal = 0;

UINT_32 APM2D::m_uiEdgeZoneVEEAntiparallelMPFUNFilter = 0;
UINT_32 APM2D::m_uiEdgeZoneVEEAntiparallelMPFUN = 0;

UINT_32 APM2D::m_uiInCircleVEE_VAntiparallelMPFUNFilter = 0;
UINT_32 APM2D::m_uiInCircleVEE_VAntiparallelMPFUN = 0;

UINT_32 APM2D::m_uiInCircleVEE_EAntiparallelMPFUNFilter = 0;
UINT_32 APM2D::m_uiInCircleVEE_EAntiparallelMPFUN = 0;

UINT_32 APM2D::m_uiInscribedCircleCentervEEMPFUNFilter = 0;
UINT_32 APM2D::m_uiInscribedCircleCentervEEMPFUN = 0;
UINT_32 APM2D::m_uiInscribedCircleCentervEEMPFUNFinal = 0;

UINT_32 APM2D::m_uiEdgeZonevEEMPFUNFilter = 0;
UINT_32 APM2D::m_uiEdgeZonevEEMPFUN = 0;

UINT_32 APM2D::m_uiInCirclevEE_VMPFUNFilter = 0;
UINT_32 APM2D::m_uiInCirclevEE_VMPFUN = 0;

UINT_32 APM2D::m_uiInCirclevEE_EMPFUNFilter = 0;
UINT_32 APM2D::m_uiInCirclevEE_EMPFUN = 0;

UINT_32 APM2D::m_uiInscribedCircleCenterEEvMPFUNFilter = 0;
UINT_32 APM2D::m_uiInscribedCircleCenterEEvMPFUN = 0;
UINT_32 APM2D::m_uiInscribedCircleCenterEEvMPFUNFinal = 0;

UINT_32 APM2D::m_uiEdgeZoneEEvMPFUNFilter = 0;
UINT_32 APM2D::m_uiEdgeZoneEEvMPFUN = 0;

UINT_32 APM2D::m_uiInCircleEEv_VMPFUNFilter = 0;
UINT_32 APM2D::m_uiInCircleEEv_VMPFUN = 0;

UINT_32 APM2D::m_uiInCircleEEv_EMPFUNFilter = 0;
UINT_32 APM2D::m_uiInCircleEEv_EMPFUN = 0;

UINT_32 APM2D::m_uiInscribedCircleCenterEEEMPFUNFilter = 0;
UINT_32 APM2D::m_uiInscribedCircleCenterEEEMPFUN = 0;
UINT_32 APM2D::m_uiInscribedCircleCenterEEEMPFUNFinal = 0;

UINT_32 APM2D::m_uiEdgeZoneEEEMPFUNFilter = 0;
UINT_32 APM2D::m_uiEdgeZoneEEEMPFUN = 0;

UINT_32 APM2D::m_uiInCircleEEE_VMPFUNFilter = 0;
UINT_32 APM2D::m_uiInCircleEEE_VMPFUN = 0;

UINT_32 APM2D::m_uiInCircleEEE_EMPFUNFilter = 0;
UINT_32 APM2D::m_uiInCircleEEE_EMPFUN = 0;

VOID APM2D::DisplayMPFUNFilterCounters(FILE *fpOutput)
{
  fprintf(fpOutput, "START: MPFUN Filter Counters\n");
  fprintf(fpOutput, "  %ld / %ld : EdgeZone\n", m_uiEdgeZoneMPFUN, m_uiEdgeZoneMPFUNFilter);

  fprintf(fpOutput, "  %ld / %ld : DistanceCompareV_VV\n", m_uiDistanceCompareV_VVMPFUN, m_uiDistanceCompareV_VVMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : DistanceCompareV_VL\n", m_uiDistanceCompareV_VLMPFUN, m_uiDistanceCompareV_VLMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : DistanceCompareV_LL\n", m_uiDistanceCompareV_LLMPFUN, m_uiDistanceCompareV_LLMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : DistanceCompareEvv_VV\n", m_uiDistanceCompareEvv_VVMPFUN, m_uiDistanceCompareEvv_VVMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : DistanceCompareEvv_VL\n", m_uiDistanceCompareEvv_VLMPFUN, m_uiDistanceCompareEvv_VLMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : DistanceCompareEvv_LL\n", m_uiDistanceCompareEvv_LLMPFUN, m_uiDistanceCompareEvv_LLMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : DistanceCompareElv_VV\n", m_uiDistanceCompareElv_VVMPFUN, m_uiDistanceCompareElv_VVMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : DistanceCompareElv_VL\n", m_uiDistanceCompareElv_VLMPFUN, m_uiDistanceCompareElv_VLMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : DistanceCompareL_VV\n", m_uiDistanceCompareL_VVMPFUN, m_uiDistanceCompareL_VVMPFUNFilter);

  fprintf(fpOutput, "  %ld / %ld : OrientationExE\n", m_uiOrientationExEMPFUN, m_uiOrientationExEMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : OrientationEdotE\n", m_uiOrientationEdotEMPFUN, m_uiOrientationEdotEMPFUNFilter);

  fprintf(fpOutput, "  ( %ld, %ld ) / %ld : InscribedCircleCenterVVV\n", m_uiInscribedCircleCenterVVVMPFUN, m_uiInscribedCircleCenterVVVMPFUNFinal, m_uiInscribedCircleCenterVVVMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : EdgeZoneVVV\n", m_uiEdgeZoneVVVMPFUN, m_uiEdgeZoneVVVMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : InCircleVVV_V\n", m_uiInCircleVVV_VMPFUN, m_uiInCircleVVV_VMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : InCircleVVV_E\n", m_uiInCircleVVV_EMPFUN, m_uiInCircleVVV_EMPFUNFilter);

  fprintf(fpOutput, "  ( %ld, %ld ) / %ld : InscribedCircleCenterVVE\n", m_uiInscribedCircleCenterVVEMPFUN, m_uiInscribedCircleCenterVVEMPFUNFinal, m_uiInscribedCircleCenterVVEMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : EdgeZoneVVE\n", m_uiEdgeZoneVVEMPFUN, m_uiEdgeZoneVVEMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : InCircleVVE_V\n", m_uiInCircleVVE_VMPFUN, m_uiInCircleVVE_VMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : InCircleVVE_E\n", m_uiInCircleVVE_EMPFUN, m_uiInCircleVVE_EMPFUNFilter);


  fprintf(fpOutput, "  ( %ld, %ld ) / %ld : InscribedCircleCenterVVEAntiparallel\n", m_uiInscribedCircleCenterVVEAntiparallelMPFUN, m_uiInscribedCircleCenterVVEAntiparallelMPFUNFinal, m_uiInscribedCircleCenterVVEAntiparallelMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : EdgeZoneVVEAntiparallel\n", m_uiEdgeZoneVVEAntiparallelMPFUN, m_uiEdgeZoneVVEAntiparallelMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : InCircleVVE_VAntiparallel\n", m_uiInCircleVVE_VAntiparallelMPFUN, m_uiInCircleVVE_VAntiparallelMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : InCircleVVE_EAntiparallel\n", m_uiInCircleVVE_EAntiparallelMPFUN, m_uiInCircleVVE_EAntiparallelMPFUNFilter);

  fprintf(fpOutput, "  ( %ld, %ld ) / %ld : InscribedCircleCenterVvE\n", m_uiInscribedCircleCenterVvEMPFUN, m_uiInscribedCircleCenterVvEMPFUNFinal, m_uiInscribedCircleCenterVvEMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : EdgeZoneVvE\n", m_uiEdgeZoneVvEMPFUN, m_uiEdgeZoneVvEMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : InCircleVvE_V\n", m_uiInCircleVvE_VMPFUN, m_uiInCircleVvE_VMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : InCircleVvE_E\n", m_uiInCircleVvE_EMPFUN, m_uiInCircleVvE_EMPFUNFilter);

  fprintf(fpOutput, "  ( %ld, %ld ) / %ld : InscribedCircleCenterEvV\n", m_uiInscribedCircleCenterEvVMPFUN, m_uiInscribedCircleCenterEvVMPFUNFinal, m_uiInscribedCircleCenterEvVMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : EdgeZoneEvV\n", m_uiEdgeZoneEvVMPFUN, m_uiEdgeZoneEvVMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : InCircleEvV_V\n", m_uiInCircleEvV_VMPFUN, m_uiInCircleEvV_VMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : InCircleEvV_E\n", m_uiInCircleEvV_EMPFUN, m_uiInCircleEvV_EMPFUNFilter);

  fprintf(fpOutput, "  ( %ld, %ld ) / %ld : InscribedCircleCenterVEE\n", m_uiInscribedCircleCenterVEEMPFUN, m_uiInscribedCircleCenterVEEMPFUNFinal, m_uiInscribedCircleCenterVEEMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : EdgeZoneVEE\n", m_uiEdgeZoneVEEMPFUN, m_uiEdgeZoneVEEMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : InCircleVEE_V\n", m_uiInCircleVEE_VMPFUN, m_uiInCircleVEE_VMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : InCircleVEE_E\n", m_uiInCircleVEE_EMPFUN, m_uiInCircleVEE_EMPFUNFilter);

  fprintf(fpOutput, "  ( %ld, %ld ) / %ld : InscribedCircleCenterVEEAntiparallel\n", m_uiInscribedCircleCenterVEEAntiparallelMPFUN, m_uiInscribedCircleCenterVEEAntiparallelMPFUNFinal, m_uiInscribedCircleCenterVEEAntiparallelMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : EdgeZoneVEEAntiparallel\n", m_uiEdgeZoneVEEAntiparallelMPFUN, m_uiEdgeZoneVEEAntiparallelMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : InCircleVEE_VAntiparallel\n", m_uiInCircleVEE_VAntiparallelMPFUN, m_uiInCircleVEE_VAntiparallelMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : InCircleVEE_EAntiparallel\n", m_uiInCircleVEE_EAntiparallelMPFUN, m_uiInCircleVEE_EAntiparallelMPFUNFilter);

  fprintf(fpOutput, "  ( %ld, %ld ) / %ld : InscribedCircleCentervEE\n", m_uiInscribedCircleCentervEEMPFUN, m_uiInscribedCircleCentervEEMPFUNFinal, m_uiInscribedCircleCentervEEMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : EdgeZonevEE\n", m_uiEdgeZonevEEMPFUN, m_uiEdgeZonevEEMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : InCirclevEE_V\n", m_uiInCirclevEE_VMPFUN, m_uiInCirclevEE_VMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : InCirclevEE_E\n", m_uiInCirclevEE_EMPFUN, m_uiInCirclevEE_EMPFUNFilter);

  fprintf(fpOutput, "  ( %ld, %ld ) / %ld : InscribedCircleCenterEEv\n", m_uiInscribedCircleCenterEEvMPFUN, m_uiInscribedCircleCenterEEvMPFUNFinal, m_uiInscribedCircleCenterEEvMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : EdgeZoneEEv\n", m_uiEdgeZoneEEvMPFUN, m_uiEdgeZoneEEvMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : InCircleEEv_V\n", m_uiInCircleEEv_VMPFUN, m_uiInCircleEEv_VMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : InCircleEEv_E\n", m_uiInCircleEEv_EMPFUN, m_uiInCircleEEv_EMPFUNFilter);


  fprintf(fpOutput, "  ( %ld, %ld ) / %ld : InscribedCircleCenterEEE\n", m_uiInscribedCircleCenterEEEMPFUN, m_uiInscribedCircleCenterEEEMPFUNFinal, m_uiInscribedCircleCenterEEEMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : EdgeZoneEEE\n", m_uiEdgeZoneEEEMPFUN, m_uiEdgeZoneEEEMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : InCircleEEE_V\n", m_uiInCircleEEE_VMPFUN, m_uiInCircleEEE_VMPFUNFilter);
  fprintf(fpOutput, "  %ld / %ld : InCircleEEE_E\n", m_uiInCircleEEE_EMPFUN, m_uiInCircleEEE_EMPFUNFilter);

  fprintf(fpOutput, "END  : MPFUN Filter Counters\n");
}
