#include "apm2d_mpfun.h"
#include <stdio.h>

#undef MAX
#undef MIN
#include "mpreal.h"

//////////////////////////////////////////////////////////////////////
// APM2D MPFUN Implementation

//////////////////////////////////////////////////////////////////////
// Intersection point of two edges

//static
BOOL IntersectionEEMPFUN(DOUBLE *pdV_x, DOUBLE *pdV_y,
			 DOUBLE dE00_x, DOUBLE dE00_y, 
			 DOUBLE dE01_x, DOUBLE dE01_y,
			 DOUBLE dE10_x, DOUBLE dE10_y, 
			 DOUBLE dE11_x, DOUBLE dE11_y)
{
  ASSERT( pdV_x != NULL && pdV_y != NULL );
  // Set to zero for the case of no intersection
  *pdV_x = 0.0;
  *pdV_y = 0.0;
  
  mp_real rE00_x = dE00_x;
  mp_real rE00_y = dE00_y; 
  mp_real rE01_x = dE01_x;
  mp_real rE01_y = dE01_y;
  mp_real rE10_x = dE10_x;
  mp_real rE10_y = dE10_y;
  mp_real rE11_x = dE11_x;
  mp_real rE11_y = dE11_y;

  //////////////////////////////
  // Parallel or colinear check
  mp_real rE0xE1;
  rE0xE1 = ( (rE01_x - rE00_x)*(rE11_y - rE10_y) - 
	     (rE11_x - rE10_x)*(rE01_y - rE00_y) ); 
  if ( rE0xE1 == 0.0 )
    {
      // Parallel or colinear edges
      return FALSE;
    }
  BOOL bPositive;
  bPositive = (rE0xE1 > 0.0);

  //////////////////////////////
  // Check I vs E00 half plane
  mp_real rE10xE11_E00;
  rE10xE11_E00 = ( (rE10_x - rE00_x)*(rE11_y - rE00_y) - 
		   (rE11_x - rE00_x)*(rE10_y - rE00_y) );
  if ( !bPositive )
    {
      if ( 0.0 < rE10xE11_E00 )
	{
	  return FALSE;
	}
    }
  else
    {
      if ( 0.0 > rE10xE11_E00 )
	{
	  return FALSE;
	}
    }

  //////////////////////////////
  // Check I vs E01 half plane
  mp_real rE10xE11_E01;
  rE10xE11_E01 = ( (rE10_x - rE01_x)*(rE11_y - rE01_y) - 
		   (rE11_x - rE01_x)*(rE10_y - rE01_y) );
  if ( !bPositive )
    {
      if ( 0.0 > rE10xE11_E01 )
	{
	  return FALSE;
	}
    }
  else
    {
      if ( 0.0 < rE10xE11_E01 )
	{
	  return FALSE;
	}
    }

  //////////////////////////////
  // Check I vs E10 half plane
  mp_real rE00xE01_E10;
  rE00xE01_E10 = ( (rE00_x - rE10_x)*(rE01_y - rE10_y) - 
		   (rE01_x - rE10_x)*(rE00_y - rE10_y) );
  if ( bPositive )
    {
      if ( 0.0 < rE00xE01_E10 )
	{
	  return FALSE;
	}
    }
  else
    {
      if ( 0.0 > rE00xE01_E10 )
	{
	  return FALSE;
	}
    }

  //////////////////////////////
  // Check I vs E11 half plane
  mp_real rE00xE01_E11;
  rE00xE01_E11 = ( (rE00_x - rE11_x)*(rE01_y - rE11_y) - 
		   (rE01_x - rE11_x)*(rE00_y - rE11_y) );
  if ( bPositive )
    {
      if ( 0.0 > rE00xE01_E11 )
	{
	  return FALSE;
	}
    }
  else
    {
      if ( 0.0 < rE00xE01_E11 )
	{
	  return FALSE;
	}
    }

  mp_real rV_x, rV_y;
  rV_x = rE00_x + rE10xE11_E00*(rE01_x - rE00_x)/rE0xE1;
  rV_y = rE00_y + rE10xE11_E00*(rE01_y - rE00_y)/rE0xE1;
  *pdV_x = dble(rV_x);
  *pdV_y = dble(rV_y);
  return TRUE;
}

//////////////////////////////////////////////////////////////////////
// Distance Comparison Functions

//static
APM2D::EdgeZoneType EdgeZoneMPFUN(DOUBLE dV0_x, DOUBLE dV0_y,
				  DOUBLE dE10_x, DOUBLE dE10_y, 
				  DOUBLE dE11_x, DOUBLE dE11_y)
{
  APM2D::EdgeZoneType eResult;
  eResult = APM2D::eztMiddle;

  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rE10_x = dE10_x;
  mp_real rE10_y = dE10_y;
  mp_real rE11_x = dE11_x;
  mp_real rE11_y = dE11_y;

  mp_real pprV[2][2]; // Endpoints
  mp_real prE[2];     // Edge vectors
  mp_real rE1dotE10, rE1dotE11;
  
  prE[0] = rE11_x - rE10_x;
  prE[1] = rE11_y - rE10_y;
  
  pprV[0][0] = rE10_x - rV0_x;
  pprV[0][1] = rE10_y - rV0_y;
  rE1dotE10 = (prE[0]*pprV[0][0]) + (prE[1]*pprV[0][1]);
  if ( rE1dotE10 > 0.0 )
    {
      eResult = APM2D::eztStart;
    }
  else 
    {
      pprV[1][0] = rE11_x - rV0_x;
      pprV[1][1] = rE11_y - rV0_y;
      rE1dotE11 = (prE[0]*pprV[1][0]) + (prE[1]*pprV[1][1]);
      if ( rE1dotE11 < 0.0 )
	{
	  eResult = APM2D::eztEnd;
	}
      else
	{
	  eResult = APM2D::eztMiddle;
	}
    }

  return eResult;
}

//static 
INT_32 DistanceCompareV_VVMPFUN(DOUBLE dV0_x, DOUBLE dV0_y,
				DOUBLE dV1_x, DOUBLE dV1_y, 
				DOUBLE dV2_x, DOUBLE dV2_y)
{
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rV1_x = dV1_x;
  mp_real rV1_y = dV1_y;
  mp_real rV2_x = dV2_x;
  mp_real rV2_y = dV2_y;

  INT_32 iResult;
  mp_real ppprV[3][2][2]; // Endpoints
  mp_real r, rV1dotV1, rV2dotV2;
  
  iResult = 0;

  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[1][0][0] = rV1_x - rV0_x;
  ppprV[1][0][1] = rV1_y - rV0_y;
  ppprV[2][0][0] = rV2_x - rV0_x;
  ppprV[2][0][1] = rV2_y - rV0_y;
  
  rV1dotV1 = (ppprV[1][0][0]*ppprV[1][0][0]) + (ppprV[1][0][1]*ppprV[1][0][1]);
  rV2dotV2 = (ppprV[2][0][0]*ppprV[2][0][0]) + (ppprV[2][0][1]*ppprV[2][0][1]);

  r = rV1dotV1 - rV2dotV2;
  
  if ( r == 0.0 )
    {
      iResult = 0;
    }
  else if ( r > 0.0 )
    {
      iResult = 1;
    }
  else
    {
      ASSERT( r < 0.0 );
      iResult = -1;
    }

  return iResult;
}

//static 
INT_32 DistanceCompareV_VLMPFUN(DOUBLE dV0_x, DOUBLE dV0_y,
				DOUBLE dV1_x, DOUBLE dV1_y, 
				DOUBLE dE20_x, DOUBLE dE20_y, 
				DOUBLE dE21_x, DOUBLE dE21_y)
{
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rV1_x = dV1_x;
  mp_real rV1_y = dV1_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y;
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;

  INT_32 iResult;
  mp_real ppprV[3][2][2]; // Endpoints
  mp_real pprE[3][2];     // Edge vectors
  mp_real r, rV1dotV1, rE2dotE2, rE20xE21;
  
  iResult = 0;

  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = rV1_x - rV0_x;
  ppprV[1][0][1] = rV1_y - rV0_y;
  ppprV[2][0][0] = rE20_x - rV0_x;
  ppprV[2][0][1] = rE20_y - rV0_y;
  ppprV[2][1][0] = rE21_x - rV0_x;
  ppprV[2][1][1] = rE21_y - rV0_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;
  
  rV1dotV1 = (ppprV[1][0][0]*ppprV[1][0][0]) + (ppprV[1][0][1]*ppprV[1][0][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);

  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);

  r = rV1dotV1*rE2dotE2 - rE20xE21*rE20xE21;
  
  if ( r == 0.0 )
    {
      iResult = 0;
    }
  else if ( r > 0.0 )
    {
      iResult = 1;
    }
  else
    {
      ASSERT( r < 0.0 );
      iResult = -1;
    }

  return iResult;
}

//static 
INT_32 DistanceCompareV_LLMPFUN(DOUBLE dV0_x, DOUBLE dV0_y,
				DOUBLE dE10_x, DOUBLE dE10_y, 
				DOUBLE dE11_x, DOUBLE dE11_y,
				DOUBLE dE20_x, DOUBLE dE20_y, 
				DOUBLE dE21_x, DOUBLE dE21_y)
{
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rE10_x = dE10_x;
  mp_real rE10_y = dE10_y;
  mp_real rE11_x = dE11_x;
  mp_real rE11_y = dE11_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y;
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;

  INT_32 iResult;
  mp_real ppprV[3][2][2]; // Endpoints
  mp_real pprE[3][2];     // Edge vectors
  mp_real r, rE1dotE1, rE2dotE2, rE10xE11, rE20xE21;
  
  iResult = 0;

  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = rE10_x - rV0_x;
  ppprV[1][0][1] = rE10_y - rV0_y;
  ppprV[1][1][0] = rE11_x - rV0_x;
  ppprV[1][1][1] = rE11_y - rV0_y;
  ppprV[2][0][0] = rE20_x - rV0_x;
  ppprV[2][0][1] = rE20_y - rV0_y;
  ppprV[2][1][0] = rE21_x - rV0_x;
  ppprV[2][1][1] = rE21_y - rV0_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[1][0] = rE11_x - rE10_x;
  pprE[1][1] = rE11_y - rE10_y;
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;
  
  rE1dotE1 = (pprE[1][0]*pprE[1][0]) + (pprE[1][1]*pprE[1][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);

  rE10xE11 = (ppprV[1][0][0]*ppprV[1][1][1]) - (ppprV[1][0][1]*ppprV[1][1][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);

  r = rE2dotE2*rE10xE11*rE10xE11 - rE1dotE1*rE20xE21*rE20xE21;
  
  if ( r == 0.0 )
    {
      iResult = 0;
    }
  else if ( r > 0.0 )
    {
      iResult = 1;
    }
  else
    {
      ASSERT( r < 0.0 );
      iResult = -1;
    }

  return iResult;
}

//static 
INT_32 DistanceCompareEvv_VVMPFUN(DOUBLE dE0A_x, DOUBLE dE0A_y,
				  DOUBLE dE0B_x, DOUBLE dE0B_y,
				  DOUBLE dV1_x, DOUBLE dV1_y, 
				  DOUBLE dV2_x, DOUBLE dV2_y)
{
  mp_real rE0A_x = dE0A_x;
  mp_real rE0A_y = dE0A_y;
  mp_real rE0B_x = dE0B_x;
  mp_real rE0B_y = dE0B_y;
  mp_real rV1_x = dV1_x;
  mp_real rV1_y = dV1_y;
  mp_real rV2_x = dV2_x;
  mp_real rV2_y = dV2_y;
  
  INT_32 iResult;
  mp_real ppprV[3][2][2]; // Endpoints
  mp_real r, rV1dotV1, rV2dotV2;
  
  iResult = 0;

  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[1][0][0] = rV1_x - rE0A_x;
  ppprV[1][0][1] = rV1_y - rE0A_y;
  ppprV[2][0][0] = rV2_x - rE0B_x;
  ppprV[2][0][1] = rV2_y - rE0B_y;
  
  rV1dotV1 = (ppprV[1][0][0]*ppprV[1][0][0]) + (ppprV[1][0][1]*ppprV[1][0][1]);
  rV2dotV2 = (ppprV[2][0][0]*ppprV[2][0][0]) + (ppprV[2][0][1]*ppprV[2][0][1]);

  r = rV1dotV1 - rV2dotV2;
  
  if ( r == 0.0 )
    {
      iResult = 0;
    }
  else if ( r > 0.0 )
    {
      iResult = 1;
    }
  else
    {
      ASSERT( r < 0.0 );
      iResult = -1;
    }

  return iResult;
}

//static 
INT_32 DistanceCompareEvv_VLMPFUN(DOUBLE dE0A_x, DOUBLE dE0A_y,
				  DOUBLE dE0B_x, DOUBLE dE0B_y,
				  DOUBLE dV1_x, DOUBLE dV1_y, 
				  DOUBLE dE20_x, DOUBLE dE20_y, 
				  DOUBLE dE21_x, DOUBLE dE21_y)
{
  mp_real rE0A_x = dE0A_x;
  mp_real rE0A_y = dE0A_y;
  mp_real rE0B_x = dE0B_x;
  mp_real rE0B_y = dE0B_y;
  mp_real rV1_x = dV1_x;
  mp_real rV1_y = dV1_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y;
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;

  INT_32 iResult;
  mp_real ppprV[3][2][2]; // Endpoints
  mp_real pprE[3][2];     // Edge vectors
  mp_real r, rV1dotV1, rE2dotE2, rE20xE21;
  
  iResult = 0;

  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = rV1_x - rE0A_x;
  ppprV[1][0][1] = rV1_y - rE0A_y;
  ppprV[2][0][0] = rE20_x - rE0B_x;
  ppprV[2][0][1] = rE20_y - rE0B_y;
  ppprV[2][1][0] = rE21_x - rE0B_x;
  ppprV[2][1][1] = rE21_y - rE0B_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;
  
  rV1dotV1 = (ppprV[1][0][0]*ppprV[1][0][0]) + (ppprV[1][0][1]*ppprV[1][0][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);

  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);

  r = rV1dotV1*rE2dotE2 - rE20xE21*rE20xE21;
  
  if ( r == 0.0 )
    {
      iResult = 0;
    }
  else if ( r > 0.0 )
    {
      iResult = 1;
    }
  else
    {
      ASSERT( r < 0.0 );
      iResult = -1;
    }

  return iResult;
}

//static 
INT_32 DistanceCompareEvv_LLMPFUN(DOUBLE dE0A_x, DOUBLE dE0A_y,
				  DOUBLE dE0B_x, DOUBLE dE0B_y,
				  DOUBLE dE10_x, DOUBLE dE10_y, 
				  DOUBLE dE11_x, DOUBLE dE11_y,
				  DOUBLE dE20_x, DOUBLE dE20_y, 
				  DOUBLE dE21_x, DOUBLE dE21_y)
{
  mp_real rE0A_x = dE0A_x;
  mp_real rE0A_y = dE0A_y;
  mp_real rE0B_x = dE0B_x;
  mp_real rE0B_y = dE0B_y;
  mp_real rE10_x = dE10_x;
  mp_real rE10_y = dE10_y;
  mp_real rE11_x = dE11_x;
  mp_real rE11_y = dE11_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y;
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;

  INT_32 iResult;
  mp_real ppprV[3][2][2]; // Endpoints
  mp_real pprE[3][2];     // Edge vectors
  mp_real r, rE1dotE1, rE2dotE2, rE10xE11, rE20xE21;
  
  iResult = 0;

  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = rE10_x - rE0A_x;
  ppprV[1][0][1] = rE10_y - rE0A_y;
  ppprV[1][1][0] = rE11_x - rE0A_x;
  ppprV[1][1][1] = rE11_y - rE0A_y;
  ppprV[2][0][0] = rE20_x - rE0B_x;
  ppprV[2][0][1] = rE20_y - rE0B_y;
  ppprV[2][1][0] = rE21_x - rE0B_x;
  ppprV[2][1][1] = rE21_y - rE0B_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[1][0] = rE11_x - rE10_x;
  pprE[1][1] = rE11_y - rE10_y;
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;
  
  rE1dotE1 = (pprE[1][0]*pprE[1][0]) + (pprE[1][1]*pprE[1][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);

  rE10xE11 = (ppprV[1][0][0]*ppprV[1][1][1]) - (ppprV[1][0][1]*ppprV[1][1][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);

  r = rE2dotE2*rE10xE11*rE10xE11 - rE1dotE1*rE20xE21*rE20xE21;
  
  if ( r == 0.0 )
    {
      iResult = 0;
    }
  else if ( r > 0.0 )
    {
      iResult = 1;
    }
  else
    {
      ASSERT( r < 0.0 );
      iResult = -1;
    }

  return iResult;
}

//static 
INT_32 DistanceCompareElv_VVMPFUN(DOUBLE dE00_x, DOUBLE dE00_y,
				  DOUBLE dE01_x, DOUBLE dE01_y,
				  DOUBLE dE0B_x, DOUBLE dE0B_y,
				  DOUBLE dV1_x, DOUBLE dV1_y, 
				  DOUBLE dV2_x, DOUBLE dV2_y)
{
  mp_real rE00_x = dE00_x;
  mp_real rE00_y = dE00_y;
  mp_real rE01_x = dE01_x;
  mp_real rE01_y = dE01_y;
  mp_real rE0B_x = dE0B_x;
  mp_real rE0B_y = dE0B_y;
  mp_real rV1_x = dV1_x;
  mp_real rV1_y = dV1_y;
  mp_real rV2_x = dV2_x;
  mp_real rV2_y = dV2_y;
  
  INT_32 iResult;
  mp_real ppprV[3][2][2]; // Endpoints
  mp_real pprE[3][2];     // Edge vectors
  mp_real r, rE0dotE0, rV2dotV2, rE00xE01;
  
  iResult = 0;

  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = rE00_x - rV1_x;
  ppprV[0][0][1] = rE00_y - rV1_y;
  ppprV[0][1][0] = rE01_x - rV1_x;
  ppprV[0][1][1] = rE01_y - rV1_y;
  ppprV[1][0][0] = 0.0;
  ppprV[1][0][1] = 0.0;
  ppprV[2][0][0] = rV2_x - rE0B_x;
  ppprV[2][0][1] = rV2_y - rE0B_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[0][0] = rE01_x - rE00_x;
  pprE[0][1] = rE01_y - rE00_y;

  rE0dotE0 = (pprE[0][0]*pprE[0][0]) + (pprE[0][1]*pprE[0][1]);
  rV2dotV2 = (ppprV[2][0][0]*ppprV[2][0][0]) + (ppprV[2][0][1]*ppprV[2][0][1]);

  rE00xE01 = (ppprV[0][0][0]*ppprV[0][1][1]) - (ppprV[0][0][1]*ppprV[0][1][0]);

  r = rE00xE01*rE00xE01 - rV2dotV2*rE0dotE0;
  
  if ( r == 0.0 )
    {
      iResult = 0;
    }
  else if ( r > 0.0 )
    {
      iResult = 1;
    }
  else
    {
      ASSERT( r < 0.0 );
      iResult = -1;
    }

  return iResult;
}

//static 
INT_32 DistanceCompareElv_VLMPFUN(DOUBLE dE00_x, DOUBLE dE00_y,
				  DOUBLE dE01_x, DOUBLE dE01_y,
				  DOUBLE dE0B_x, DOUBLE dE0B_y,
				  DOUBLE dV1_x, DOUBLE dV1_y, 
				  DOUBLE dE20_x, DOUBLE dE20_y, 
				  DOUBLE dE21_x, DOUBLE dE21_y)
{
  mp_real rE00_x = dE00_x;
  mp_real rE00_y = dE00_y;
  mp_real rE01_x = dE01_x;
  mp_real rE01_y = dE01_y;
  mp_real rE0B_x = dE0B_x;
  mp_real rE0B_y = dE0B_y;
  mp_real rV1_x = dV1_x;
  mp_real rV1_y = dV1_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y;
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;

  INT_32 iResult;
  mp_real ppprV[3][2][2]; // Endpoints
  mp_real pprE[3][2];     // Edge vectors
  mp_real r, rE0dotE0, rE2dotE2, rE00xE01, rE20xE21;
  
  iResult = 0;

  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = rE00_x - rV1_x;
  ppprV[0][0][1] = rE00_y - rV1_y;
  ppprV[0][1][0] = rE01_x - rV1_x;
  ppprV[0][1][1] = rE01_y - rV1_y;
  ppprV[1][0][0] = 0.0;
  ppprV[1][0][1] = 0.0;
  ppprV[2][0][0] = rE20_x - rE0B_x;
  ppprV[2][0][1] = rE20_y - rE0B_y;
  ppprV[2][1][0] = rE21_x - rE0B_x;
  ppprV[2][1][1] = rE21_y - rE0B_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[0][0] = rE01_x - rE00_x;
  pprE[0][1] = rE01_y - rE00_y;
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;
  
  rE0dotE0 = (pprE[0][0]*pprE[0][0]) + (pprE[0][1]*pprE[0][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);

  rE00xE01 = (ppprV[0][0][0]*ppprV[0][1][1]) - (ppprV[0][0][1]*ppprV[0][1][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);

  r = rE2dotE2*rE00xE01*rE00xE01 - rE0dotE0*rE20xE21*rE20xE21;
  
  if ( r == 0.0 )
    {
      iResult = 0;
    }
  else if ( r > 0.0 )
    {
      iResult = 1;
    }
  else
    {
      ASSERT( r < 0.0 );
      iResult = -1;
    }

  return iResult;
}

//static 
INT_32 DistanceCompareL_VVMPFUN(DOUBLE dE00_x, DOUBLE dE00_y,
				DOUBLE dE01_x, DOUBLE dE01_y,
				DOUBLE dV1_x, DOUBLE dV1_y, 
				DOUBLE dV2_x, DOUBLE dV2_y)
{
  mp_real rE00_x = dE00_x;
  mp_real rE00_y = dE00_y;
  mp_real rE01_x = dE01_x;
  mp_real rE01_y = dE01_y;
  mp_real rV1_x = dV1_x;
  mp_real rV1_y = dV1_y;
  mp_real rV2_x = dV2_x;
  mp_real rV2_y = dV2_y;

  INT_32 iResult;
  mp_real ppprV[3][2][2]; // Endpoints
  mp_real pprE[3][2];     // Edge vectors
  mp_real r, rE0dotE0, rE10xE11, rE20xE21;
  
  iResult = 0;

  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[1][0][0] = rE00_x - rV1_x;
  ppprV[1][0][1] = rE00_y - rV1_y;
  ppprV[1][1][0] = rE01_x - rV1_x;
  ppprV[1][1][1] = rE01_y - rV1_y;
  ppprV[2][0][0] = rE00_x - rV2_x;
  ppprV[2][0][1] = rE00_y - rV2_y;
  ppprV[2][1][0] = rE01_x - rV2_x;
  ppprV[2][1][1] = rE01_y - rV2_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[0][0] = rE01_x - rE00_x;
  pprE[0][1] = rE01_y - rE00_y;

  rE0dotE0 = (pprE[0][0]*pprE[0][0]) + (pprE[0][1]*pprE[0][1]);

  rE10xE11 = (ppprV[1][0][0]*ppprV[1][1][1]) - (ppprV[1][0][1]*ppprV[1][1][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);

  r = rE10xE11*rE10xE11 - rE20xE21*rE20xE21;
  
  if ( r == 0.0 )
    {
      iResult = 0;
    }
  else if ( r > 0.0 )
    {
      iResult = 1;
    }
  else
    {
      ASSERT( r < 0.0 );
      iResult = -1;
    }

  return iResult;
}

//////////////////////////////////////////////////////////////////////
// VVV Orientation Functions

static
mp_real SinThetaMPFUN(mp_real rV2_x, mp_real rV2_y,
		      mp_real rV0_x, mp_real rV0_y,
		      mp_real rV1_x, mp_real rV1_y)
{
  mp_real prV1[2], prV2[2], rV1dotV1, rV2dotV2;
  //mp_real rDen;
  mp_real r;

  r = 0.0;

  prV1[0] = rV1_x - rV0_x;
  prV1[1] = rV1_y - rV0_y;
  prV2[0] = rV2_x - rV0_x;
  prV2[1] = rV2_y - rV0_y;
  
  rV1dotV1 = (prV1[0]*prV1[0]) + (prV1[1]*prV1[1]);
  rV2dotV2 = (prV2[0]*prV2[0]) + (prV2[1]*prV2[1]);
  
  r = ( (prV1[0] * prV2[1]) - (prV2[0] * prV1[1]) );
  /*
  rDen = sqrt(rV1dotV1*rV2dotV2);
  if ( rDen == 0.0 )
    {
      r = 0.0;
    }
  else
    {
      r /= rDen;
    }
  */
  return r;
}

//////////////////////////////////////////////////////////////////////
// EE Orientaion Functions

//static
INT_32 OrientationExEMPFUN(DOUBLE dE00_x, DOUBLE dE00_y,
			   DOUBLE dE01_x, DOUBLE dE01_y,
			   DOUBLE dE10_x, DOUBLE dE10_y,
			   DOUBLE dE11_x, DOUBLE dE11_y)
{
  mp_real rE00_x = dE00_x;
  mp_real rE00_y = dE00_y;
  mp_real rE01_x = dE01_x;
  mp_real rE01_y = dE01_y;
  mp_real rE10_x = dE10_x;
  mp_real rE10_y = dE10_y;
  mp_real rE11_x = dE11_x;
  mp_real rE11_y = dE11_y;

  INT_32 iResult;
  mp_real r;

  iResult = 0;
  r = 0.0;

  mp_real prE0[2], prE1[2];
  
  prE0[0] = rE01_x - rE00_x;
  prE0[1] = rE01_y - rE00_y;
  prE1[0] = rE11_x - rE10_x;
  prE1[1] = rE11_y - rE10_y;
  
  r = ( (prE0[0] * prE1[1]) - (prE1[0] * prE0[1]) );

  if ( r == 0.0 )
    {
      iResult = 0;
    }
  else if ( r > 0.0 )
    {
      iResult = 1;
    }
  else
    {
      ASSERT( r < 0.0 );
      iResult = -1;
    }

  return iResult;
}

//static  
INT_32 OrientationEdotEMPFUN(DOUBLE dE00_x, DOUBLE dE00_y,
			     DOUBLE dE01_x, DOUBLE dE01_y,
			     DOUBLE dE10_x, DOUBLE dE10_y,
			     DOUBLE dE11_x, DOUBLE dE11_y)
{
  mp_real rE00_x = dE00_x;
  mp_real rE00_y = dE00_y;
  mp_real rE01_x = dE01_x;
  mp_real rE01_y = dE01_y;
  mp_real rE10_x = dE10_x;
  mp_real rE10_y = dE10_y;
  mp_real rE11_x = dE11_x;
  mp_real rE11_y = dE11_y;

  INT_32 iResult;
  mp_real r;

  iResult = 0;
  r = 0.0;

  mp_real prE0[2], prE1[2];
  
  prE0[0] = rE01_x - rE00_x;
  prE0[1] = rE01_y - rE00_y;
  prE1[0] = rE11_x - rE10_x;
  prE1[1] = rE11_y - rE10_y;
  
  r = ( (prE0[0] * prE1[0]) + (prE0[1] * prE1[1]) );

  if ( r == 0.0 )
    {
      iResult = 0;
    }
  else if ( r > 0.0 )
    {
      iResult = 1;
    }
  else
    {
      ASSERT( r < 0.0 );
      iResult = -1;
    }

  return iResult;
}

//////////////////////////////////////////////////////////////////////
// Inscribed Circle Center Functions

//static 
APM2D::InscribedCircleCenterType 
InscribedCircleCenterVVVMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
                              DOUBLE dV00_x, DOUBLE dV00_y,
                              DOUBLE dV0_x, DOUBLE dV0_y,
			      DOUBLE dV01_x, DOUBLE dV01_y,
			      DOUBLE dV10_x, DOUBLE dV10_y,
			      DOUBLE dV1_x, DOUBLE dV1_y,
			      DOUBLE dV11_x, DOUBLE dV11_y,
			      DOUBLE dV20_x, DOUBLE dV20_y,
			      DOUBLE dV2_x, DOUBLE dV2_y,
			      DOUBLE dV21_x, DOUBLE dV21_y)
{
  mp_real prC[3], rR;
  mp_real rV00_x = dV00_x; 
  mp_real rV00_y = dV00_y;
  mp_real rV0_x = dV0_x; 
  mp_real rV0_y = dV0_y;
  mp_real rV01_x = dV01_x; 
  mp_real rV01_y = dV01_y;
  mp_real rV10_x = dV10_x; 
  mp_real rV10_y = dV10_y;
  mp_real rV1_x = dV1_x; 
  mp_real rV1_y = dV1_y;
  mp_real rV11_x = dV11_x; 
  mp_real rV11_y = dV11_y;
  mp_real rV20_x = dV20_x; 
  mp_real rV20_y = dV20_y;
  mp_real rV2_x = dV2_x; 
  mp_real rV2_y = dV2_y;
  mp_real rV21_x = dV21_x; 
  mp_real rV21_y = dV21_y;
  
  mp_real pprV[3][2], ppprVN[3][2][2], prV1V2[2];
  mp_real pprVertexZone[3][2];
  mp_real rV1dotV1, rV2dotV2, rV1xV2;
  mp_real rV1xV00, rV1xV01, rV1xV10, rV1xV11, rV1xV20, rV1xV21;
  mp_real rV2xV00, rV2xV01, rV2xV10, rV2xV11, rV2xV20, rV2xV21;
  mp_real rV1dotV10, rV1dotV11, rV2dotV20, rV2dotV21;
  mp_real rV00dotV00, rV01dotV01, rV10dotV10, rV11dotV11, rV20dotV20, rV21dotV21;
  mp_real rV1V2dotV1V2;
  //DOUBLE dSinTheta;
  UINT_32 i, j;

  // Translate V0 to the origin
  //memset(pprV, 0, sizeof(pprV));
  pprV[1][0] = rV1_x - rV0_x;
  pprV[1][1] = rV1_y - rV0_y;
  pprV[2][0] = rV2_x - rV0_x;
  pprV[2][1] = rV2_y - rV0_y;

  // Vertex neighbor vectors
  //memset(ppprVN, 0, sizeof(ppprVN));
  ppprVN[0][0][0] = rV0_x - rV00_x;
  ppprVN[0][0][1] = rV0_y - rV00_y;
  ppprVN[0][1][0] = rV0_x - rV01_x;
  ppprVN[0][1][1] = rV0_y - rV01_y;
  ppprVN[1][0][0] = rV1_x - rV10_x;
  ppprVN[1][0][1] = rV1_y - rV10_y;
  ppprVN[1][1][0] = rV1_x - rV11_x;
  ppprVN[1][1][1] = rV1_y - rV11_y;
  ppprVN[2][0][0] = rV2_x - rV20_x;
  ppprVN[2][0][1] = rV2_y - rV20_y;
  ppprVN[2][1][0] = rV2_x - rV21_x;
  ppprVN[2][1][1] = rV2_y - rV21_y;

  prV1V2[0] = rV2_x - rV1_x;
  prV1V2[1] = rV2_y - rV1_y;
  
  rV1xV2 = ( (pprV[1][0] * pprV[2][1]) - (pprV[2][0] * pprV[1][1]) );
  rV1dotV1 = (pprV[1][0] * pprV[1][0]) + (pprV[1][1] * pprV[1][1]);
  rV2dotV2 = (pprV[2][0] * pprV[2][0]) + (pprV[2][1] * pprV[2][1]);

  rV1xV00 = ( (pprV[1][0] * ppprVN[0][0][1]) - (pprV[1][1] * ppprVN[0][0][0]) );
  rV1xV01 = ( (pprV[1][0] * ppprVN[0][1][1]) - (pprV[1][1] * ppprVN[0][1][0]) );
  rV1xV10 = ( (pprV[1][0] * ppprVN[1][0][1]) - (pprV[1][1] * ppprVN[1][0][0]) );
  rV1xV11 = ( (pprV[1][0] * ppprVN[1][1][1]) - (pprV[1][1] * ppprVN[1][1][0]) );
  rV1xV20 = ( (pprV[1][0] * ppprVN[2][0][1]) - (pprV[1][1] * ppprVN[2][0][0]) );
  rV1xV21 = ( (pprV[1][0] * ppprVN[2][1][1]) - (pprV[1][1] * ppprVN[2][1][0]) );

  rV2xV00 = ( (pprV[2][0] * ppprVN[0][0][1]) - (pprV[2][1] * ppprVN[0][0][0]) );
  rV2xV01 = ( (pprV[2][0] * ppprVN[0][1][1]) - (pprV[2][1] * ppprVN[0][1][0]) );
  rV2xV10 = ( (pprV[2][0] * ppprVN[1][0][1]) - (pprV[2][1] * ppprVN[1][0][0]) );
  rV2xV11 = ( (pprV[2][0] * ppprVN[1][1][1]) - (pprV[2][1] * ppprVN[1][1][0]) );
  rV2xV20 = ( (pprV[2][0] * ppprVN[2][0][1]) - (pprV[2][1] * ppprVN[2][0][0]) );
  rV2xV21 = ( (pprV[2][0] * ppprVN[2][1][1]) - (pprV[2][1] * ppprVN[2][1][0]) );

  rV1dotV10 = ( (pprV[1][0] * ppprVN[1][0][0]) + (pprV[1][1] * ppprVN[1][0][1]) );
  rV1dotV11 = ( (pprV[1][0] * ppprVN[1][1][0]) + (pprV[1][1] * ppprVN[1][1][1]) );
  rV2dotV20 = ( (pprV[2][0] * ppprVN[2][0][0]) + (pprV[2][1] * ppprVN[2][0][1]) );
  rV2dotV21 = ( (pprV[2][0] * ppprVN[2][1][0]) + (pprV[2][1] * ppprVN[2][1][1]) );

  rV00dotV00 = ( (ppprVN[0][0][0] * ppprVN[0][0][0]) + (ppprVN[0][0][1] * ppprVN[0][0][1]) );
  rV01dotV01 = ( (ppprVN[0][1][0] * ppprVN[0][1][0]) + (ppprVN[0][1][1] * ppprVN[0][1][1]) );
  rV10dotV10 = ( (ppprVN[1][0][0] * ppprVN[1][0][0]) + (ppprVN[1][0][1] * ppprVN[1][0][1]) );
  rV11dotV11 = ( (ppprVN[1][1][0] * ppprVN[1][1][0]) + (ppprVN[1][1][1] * ppprVN[1][1][1]) );
  rV20dotV20 = ( (ppprVN[2][0][0] * ppprVN[2][0][0]) + (ppprVN[2][0][1] * ppprVN[2][0][1]) );
  rV21dotV21 = ( (ppprVN[2][1][0] * ppprVN[2][1][0]) + (ppprVN[2][1][1] * ppprVN[2][1][1]) );

  if ( !APM2D::Counterclockwise(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y) )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  //////////////////////////////////////////////////
  // Check the Vertex Zones

  pprVertexZone[0][0] = ( rV2dotV2*rV1xV00 - rV1dotV1*rV2xV00 );
  pprVertexZone[0][1] = ( rV2dotV2*rV1xV01 - rV1dotV1*rV2xV01 );
  pprVertexZone[1][0] = ( rV2dotV2*rV1xV10 - rV1dotV1*rV2xV10 - 2.0*rV1xV2*rV1dotV10 );
  pprVertexZone[1][1] = ( rV2dotV2*rV1xV11 - rV1dotV1*rV2xV11 - 2.0*rV1xV2*rV1dotV11 );
  pprVertexZone[2][0] = ( rV2dotV2*rV1xV20 - rV1dotV1*rV2xV20 - 2.0*rV1xV2*rV2dotV20 );
  pprVertexZone[2][1] = ( rV2dotV2*rV1xV21 - rV1dotV1*rV2xV21 - 2.0*rV1xV2*rV2dotV21 );

  for ( i = 0; i < 3; i++ )
    {
      for ( j = 0; j < 2; j++ )
	{
	  if ( pprVertexZone[i][j] < 0.0 )
	    {
	      *pdC_x = 0.0;
	      *pdC_y = 0.0;
	      *pdR = 0.0;
	      return APM2D::icctNull;
	    }
	}
    }

  //////////////////////////////////////////////////
  // Compute the center and radius

  prC[0] =  ( (rV2dotV2 * pprV[1][1]) - (rV1dotV1 * pprV[2][1]) ) / ( -2.0 * rV1xV2 );
  prC[1] =  ( (rV2dotV2 * pprV[1][0]) - (rV1dotV1 * pprV[2][0]) ) / ( 2.0 * rV1xV2 );
  //prC[2] = sqrt( (prC[0] * prC[0]) + (prC[1] * prC[1]) );

  rV1V2dotV1V2 = (prV1V2[0] * prV1V2[0]) + (prV1V2[1] * prV1V2[1]);

  rR = sqrt(rV1dotV1 * rV2dotV2 * rV1V2dotV1V2) / ( 2.0 * rV1xV2 );

  /*
  if ( !(prC[2] - rR) < s_crEpsilonInscribedCircleCenterVVVRadius ) )
    {
      fprintf(strerr, "VVV: C = ( %f, %f, %f ) != %f = R\n", prC[0], prC[1], prC[2], rR);
      //ASSERT( FALSE );
    }
  */

  // Translate the answer back out away from the origin 
  prC[0] = rV0_x + prC[0];
  prC[1] = rV0_y + prC[1];

  *pdC_x = dble(prC[0]);
  *pdC_y = dble(prC[1]);
  *pdR = dble(rR);
  return APM2D::icctVVV;
}

//static 
APM2D::EdgeZoneType EdgeZoneVVVMPFUN(DOUBLE dV0_x, DOUBLE dV0_y,
				     DOUBLE dV1_x, DOUBLE dV1_y,
				     DOUBLE dV2_x, DOUBLE dV2_y,
				     DOUBLE dE30_x, DOUBLE dE30_y,
				     DOUBLE dE31_x, DOUBLE dE31_y)
{
  mp_real rV0_x = dV0_x; 
  mp_real rV0_y = dV0_y;
  mp_real rV1_x = dV1_x;
  mp_real rV1_y = dV1_y;
  mp_real rV2_x = dV2_x;
  mp_real rV2_y = dV2_y;
  mp_real rE30_x = dE30_x;
  mp_real rE30_y = dE30_y;
  mp_real rE31_x = dE31_x;
  mp_real rE31_y = dE31_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real prEdgeZone[2];
  mp_real rV1dotV1, rV2dotV2;
  mp_real rE3dotE3, rE3dotE30, rE3dotE31;
  mp_real rV1xV2, rE3xV2, rE3xV1;
  //mp_real rDen;
  APM2D::EdgeZoneType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = rV1_x - rV0_x;
  ppprV[1][0][1] = rV1_y - rV0_y;
  ppprV[2][0][0] = rV2_x - rV0_x;
  ppprV[2][0][1] = rV2_y - rV0_y;
  ppprV[3][0][0] = rE30_x - rV0_x;
  ppprV[3][0][1] = rE30_y - rV0_y;
  ppprV[3][1][0] = rE31_x - rV0_x;
  ppprV[3][1][1] = rE31_y - rV0_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[3][0] = rE31_x - rE30_x;
  pprE[3][1] = rE31_y - rE30_y;
  
  rV1dotV1 = (ppprV[1][0][0]*ppprV[1][0][0]) + (ppprV[1][0][1]*ppprV[1][0][1]);
  rV2dotV2 = (ppprV[2][0][0]*ppprV[2][0][0]) + (ppprV[2][0][1]*ppprV[2][0][1]);

  rE3dotE3 = (pprE[3][0]*pprE[3][0]) + (pprE[3][1]*pprE[3][1]);
  rE3dotE30 = (pprE[3][0]*ppprV[3][0][0]) + (pprE[3][1]*ppprV[3][0][1]);
  rE3dotE31 = (pprE[3][0]*ppprV[3][1][0]) + (pprE[3][1]*ppprV[3][1][1]);
  
  rV1xV2 = (ppprV[1][0][0]*ppprV[2][0][1]) - (ppprV[1][0][1]*ppprV[2][0][0]);
  rE3xV1 = (pprE[3][0]*ppprV[1][0][1]) - (pprE[3][1]*ppprV[1][0][0]);
  rE3xV2 = (pprE[3][0]*ppprV[2][0][1]) - (pprE[3][1]*ppprV[2][0][0]);

  //rDen = 2.0*rV1xV2; 
  //ASSERT( rDen != 0.0 );

  /*
  prC[0] =  ( (rV2dotV2 * ppprV[1][0][1]) - (rV1dotV1 * ppprV[2][0][1]) ) / ( -2.0 * rV1xV2 );
  prC[1] =  ( (rV2dotV2 * ppprV[1][0][0]) - (rV1dotV1 * ppprV[2][0][0]) ) / ( 2.0 * rV1xV2 );

  rMagCE30 = ( (prC[0] - ppprV[3][0][0])*(prC[0] - ppprV[3][0][0]) + 
	       (prC[1] - ppprV[3][0][1])*(prC[1] - ppprV[3][0][1]) );
  rMagCE31 = ( (prC[0] - ppprV[3][1][0])*(prC[0] - ppprV[3][1][0]) + 
	       (prC[1] - ppprV[3][1][1])*(prC[1] - ppprV[3][1][1]) );

  if ( rMagCE30 <= 0.0 )
    {
      prEdgeZone[0] = 0.0;
    }
  else
    {
      prEdgeZone[0] = ( 2.0*rV1xV2*rE3dotE30 - (rV1dotV1*rE3xV2 - rV2dotV2*rE3xV1);
    }
  if ( rMagCE31 <= 0.0 )
    {
      prEdgeZone[1] = 0.0;
    }
  else
    {
      prEdgeZone[1] = -( 2.0*rV1xV2*rE3dotE31 - (rV1dotV1*rE3xV2 - rV2dotV2*rE3xV1);
    }
  */

  prEdgeZone[0] = ( 2.0*rV1xV2*rE3dotE30 - (rV1dotV1*rE3xV2 - rV2dotV2*rE3xV1) );
  prEdgeZone[1] = -( 2.0*rV1xV2*rE3dotE31 - (rV1dotV1*rE3xV2 - rV2dotV2*rE3xV1) );
  
  if ( !( prEdgeZone[0] < 0.0 ) )
    {
      eResult = APM2D::eztStart;
    }
  else if ( !( prEdgeZone[1] < 0.0 ) )
    {
      eResult = APM2D::eztEnd;
    }
  else
    {
      eResult = APM2D::eztMiddle;
    }

  return eResult;
}

//static 
APM2D::CircleTestType InCircleVVV_VMPFUN(DOUBLE dV0_x, DOUBLE dV0_y,
					 DOUBLE dV1_x, DOUBLE dV1_y,
					 DOUBLE dV2_x, DOUBLE dV2_y,
					 DOUBLE dV3_x, DOUBLE dV3_y)
{
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rV1_x = dV1_x;
  mp_real rV1_y = dV1_y;
  mp_real rV2_x = dV2_x;
  mp_real rV2_y = dV2_y;
  mp_real rV3_x = dV3_x;
  mp_real rV3_y = dV3_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real rV1dotV1, rV2dotV2, rV3dotV3;
  mp_real rV1xV2, rV2xV3, rV3xV1;
  //mp_real rDen;
  mp_real rResult;
  APM2D::CircleTestType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = rV1_x - rV0_x;
  ppprV[1][0][1] = rV1_y - rV0_y;
  ppprV[2][0][0] = rV2_x - rV0_x;
  ppprV[2][0][1] = rV2_y - rV0_y;
  ppprV[3][0][0] = rV3_x - rV0_x;
  ppprV[3][0][1] = rV3_y - rV0_y;
  
  rV1dotV1 = (ppprV[1][0][0]*ppprV[1][0][0]) + (ppprV[1][0][1]*ppprV[1][0][1]);
  rV2dotV2 = (ppprV[2][0][0]*ppprV[2][0][0]) + (ppprV[2][0][1]*ppprV[2][0][1]);
  rV3dotV3 = (ppprV[3][0][0]*ppprV[3][0][0]) + (ppprV[3][0][1]*ppprV[3][0][1]);

  rV1xV2 = (ppprV[1][0][0]*ppprV[2][0][1]) - (ppprV[1][0][1]*ppprV[2][0][0]);
  rV2xV3 = (ppprV[2][0][0]*ppprV[3][0][1]) - (ppprV[2][0][1]*ppprV[3][0][0]);
  rV3xV1 = (ppprV[3][0][0]*ppprV[1][0][1]) - (ppprV[3][0][1]*ppprV[1][0][0]);

  //rDen = rV1xV2; 
  //ASSERT( rDen != 0.0 );

  rResult = ( rV1dotV1*rV2xV3 + rV2dotV2*rV3xV1 + rV3dotV3*rV1xV2 );
  
  if ( rResult < 0.0 )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( rResult > 0.0 )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

//static 
APM2D::CircleTestType InCircleVVV_EMPFUN(DOUBLE dV0_x, DOUBLE dV0_y,
					 DOUBLE dV1_x, DOUBLE dV1_y,
					 DOUBLE dV2_x, DOUBLE dV2_y,
					 DOUBLE dE30_x, DOUBLE dE30_y,
					 DOUBLE dE31_x, DOUBLE dE31_y)
{
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rV1_x = dV1_x;
  mp_real rV1_y = dV1_y;
  mp_real rV2_x = dV2_x;
  mp_real rV2_y = dV2_y;
  mp_real rE30_x = dE30_x;
  mp_real rE30_y = dE30_y;
  mp_real rE31_x = dE31_x;
  mp_real rE31_y = dE31_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real prV1V2[2];
  mp_real rV1dotV1, rV2dotV2, rE3dotE3, rE3dotV1, rE3dotV2, rV1V2dotV1V2;
  mp_real rE3dotE30, rE3dotE31;
  mp_real rV1xV2, rE30xE31;
  //mp_real rDen;
  mp_real rResult;
  APM2D::CircleTestType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = rV1_x - rV0_x;
  ppprV[1][0][1] = rV1_y - rV0_y;
  ppprV[2][0][0] = rV2_x - rV0_x;
  ppprV[2][0][1] = rV2_y - rV0_y;
  ppprV[3][0][0] = rE30_x - rV0_x;
  ppprV[3][0][1] = rE30_y - rV0_y;
  ppprV[3][1][0] = rE31_x - rV0_x;
  ppprV[3][1][1] = rE31_y - rV0_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[3][0] = rE31_x - rE30_x;
  pprE[3][1] = rE31_y - rE30_y;

  prV1V2[0] = rV2_x - rV1_x;
  prV1V2[1] = rV2_y - rV1_y;
  
  rV1dotV1 = (ppprV[1][0][0]*ppprV[1][0][0]) + (ppprV[1][0][1]*ppprV[1][0][1]);
  rV2dotV2 = (ppprV[2][0][0]*ppprV[2][0][0]) + (ppprV[2][0][1]*ppprV[2][0][1]);
  rE3dotE3 = (pprE[3][0]*pprE[3][0]) + (pprE[3][1]*pprE[3][1]);
  rE3dotV1 = (pprE[3][0]*ppprV[1][0][0]) + (pprE[3][1]*ppprV[1][0][1]);
  rE3dotV2 = (pprE[3][0]*ppprV[2][0][0]) + (pprE[3][1]*ppprV[2][0][1]);

  rV1V2dotV1V2 = (prV1V2[0]*prV1V2[0]) + (prV1V2[1]*prV1V2[1]);

  rE3dotE30 = (pprE[3][0]*ppprV[3][0][0]) + (pprE[3][1]*ppprV[3][0][1]);
  rE3dotE31 = (pprE[3][0]*ppprV[3][1][0]) + (pprE[3][1]*ppprV[3][1][1]);
  
  rV1xV2 = (ppprV[1][0][0]*ppprV[2][0][1]) - (ppprV[1][0][1]*ppprV[2][0][0]);
  rE30xE31 = (ppprV[3][0][0]*ppprV[3][1][1]) - (ppprV[3][0][1]*ppprV[3][1][0]);

  //rDen = 4.0*rE3dotE3*rV1xV2*rV1xV2; 
  //ASSERT( rDen != 0.0 );

  rResult = ( rV1dotV1*rV1dotV1*rE3dotV2*rE3dotV2 + rV2dotV2*rV2dotV2*rE3dotV1*rE3dotV1 
	      - rV1dotV1*rV2dotV2*(rV1V2dotV1V2*rE3dotE3 + 2.0*rE3dotV1*rE3dotV2)
	      + 4.0*rV1xV2*rE30xE31*(rV1xV2*rE30xE31 + 
				     rV2dotV2*rE3dotV1 - rV1dotV1*rE3dotV2) );
  
  if ( rResult < 0.0 )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( rResult > 0.0 )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

//static 
APM2D::InscribedCircleCenterType 
InscribedCircleCenterVVEMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			      DOUBLE dV00_x, DOUBLE dV00_y,
			      DOUBLE dV0_x, DOUBLE dV0_y,
			      DOUBLE dV01_x, DOUBLE dV01_y,
			      DOUBLE dV10_x, DOUBLE dV10_y,
			      DOUBLE dV1_x, DOUBLE dV1_y,
			      DOUBLE dV11_x, DOUBLE dV11_y,
			      DOUBLE dE20_x, DOUBLE dE20_y, 
			      DOUBLE dE21_x, DOUBLE dE21_y)
{
  mp_real prC[3], rR;
  mp_real rV00_x = dV00_x;
  mp_real rV00_y = dV00_y;
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rV01_x = dV01_x;
  mp_real rV01_y = dV01_y;
  mp_real rV10_x = dV10_x;
  mp_real rV10_y = dV10_y;
  mp_real rV1_x = dV1_x;
  mp_real rV1_y = dV1_y;
  mp_real rV11_x = dV11_x;
  mp_real rV11_y = dV11_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;

  mp_real ppprV[3][2][2], ppprVN[2][2][2], prE[2];
  mp_real rDen; 
  mp_real rDisc, rDiscR;
  mp_real rV1dotV1, rE2dotE2, rE2dotV1, rE2dotE20, rE2dotE21;
  mp_real rE2xV1, rE20xE21;
  mp_real rV1xV00, rV1xV01, rV1xV10, rV1xV11;
  mp_real rE2xV00, rE2xV01, rE2xV10, rE2xV11;
  mp_real rV1dotV10, rV1dotV11;
  mp_real rV00dotV00, rV01dotV01, rV10dotV10, rV11dotV11;
  mp_real rE20dotE20, rE21dotE21;
  mp_real rV1onE2;
  mp_real rSinTheta;
  mp_real pprVertexZone[2][2];
  mp_real prEdgeZone[2];
  UINT_32 i, j;
  
  // Translate V0 to the origin
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[1][0][0] = rV1_x - rV0_x;
  ppprV[1][0][1] = rV1_y - rV0_y;
  ppprV[2][0][0] = rE20_x - rV0_x;
  ppprV[2][0][1] = rE20_y - rV0_y;
  ppprV[2][1][0] = rE21_x - rV0_x;
  ppprV[2][1][1] = rE21_y - rV0_y;
  
  // Vertex neighbor vectors
  //memset(ppprVN, 0, sizeof(ppprV));
  ppprVN[0][0][0] = rV0_x - rV00_x;
  ppprVN[0][0][1] = rV0_y - rV00_y;
  ppprVN[0][1][0] = rV0_x - rV01_x;
  ppprVN[0][1][1] = rV0_y - rV01_y;
  ppprVN[1][0][0] = rV1_x - rV10_x;
  ppprVN[1][0][1] = rV1_y - rV10_y;
  ppprVN[1][1][0] = rV1_x - rV11_x;
  ppprVN[1][1][1] = rV1_y - rV11_y;

  prE[0] = rE21_x - rE20_x;
  prE[1] = rE21_y - rE20_y;
  
  rV1dotV1 = (ppprV[1][0][0]*ppprV[1][0][0]) + (ppprV[1][0][1]*ppprV[1][0][1]);
  rE2dotE2 = (prE[0]*prE[0]) + (prE[1]*prE[1]);
  rE2dotV1 = (prE[0]*ppprV[1][0][0]) + (prE[1]*ppprV[1][0][1]);

  rE2dotE20 = (prE[0]*ppprV[2][0][0]) + (prE[1]*ppprV[2][0][1]);
  rE2dotE21 = (prE[0]*ppprV[2][1][0]) + (prE[1]*ppprV[2][1][1]);

  rE2xV1 = (prE[0]*ppprV[1][0][1]) - (prE[1]*ppprV[1][0][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);
  
  rE20dotE20 = (ppprV[2][0][0]*ppprV[2][0][0]) + (ppprV[2][0][1]*ppprV[2][0][1]);
  rE21dotE21 = (ppprV[2][1][0]*ppprV[2][1][0]) + (ppprV[2][1][1]*ppprV[2][1][1]);

  rV00dotV00 = ( (ppprVN[0][0][0] * ppprVN[0][0][0]) + (ppprVN[0][0][1] * ppprVN[0][0][1]) );
  rV01dotV01 = ( (ppprVN[0][1][0] * ppprVN[0][1][0]) + (ppprVN[0][1][1] * ppprVN[0][1][1]) );
  rV10dotV10 = ( (ppprVN[1][0][0] * ppprVN[1][0][0]) + (ppprVN[1][0][1] * ppprVN[1][0][1]) );
  rV11dotV11 = ( (ppprVN[1][1][0] * ppprVN[1][1][0]) + (ppprVN[1][1][1] * ppprVN[1][1][1]) );

  rV1dotV10 = ( (ppprV[1][0][0] * ppprVN[1][0][0]) + (ppprV[1][0][1] * ppprVN[1][0][1]) );
  rV1dotV11 = ( (ppprV[1][0][0] * ppprVN[1][1][0]) + (ppprV[1][0][1] * ppprVN[1][1][1]) );

  rV1xV00 = ( (ppprV[1][0][0] * ppprVN[0][0][1]) - (ppprV[1][0][1] * ppprVN[0][0][0]) );
  rV1xV01 = ( (ppprV[1][0][0] * ppprVN[0][1][1]) - (ppprV[1][0][1] * ppprVN[0][1][0]) );
  rV1xV10 = ( (ppprV[1][0][0] * ppprVN[1][0][1]) - (ppprV[1][0][1] * ppprVN[1][0][0]) );
  rV1xV11 = ( (ppprV[1][0][0] * ppprVN[1][1][1]) - (ppprV[1][0][1] * ppprVN[1][1][0]) );

  rE2xV00 = ( (prE[0] * ppprVN[0][0][1]) - (prE[1] * ppprVN[0][0][0]) );
  rE2xV01 = ( (prE[0] * ppprVN[0][1][1]) - (prE[1] * ppprVN[0][1][0]) );
  rE2xV10 = ( (prE[0] * ppprVN[1][0][1]) - (prE[1] * ppprVN[1][0][0]) );
  rE2xV11 = ( (prE[0] * ppprVN[1][1][1]) - (prE[1] * ppprVN[1][1][0]) );

  rV1onE2 = ((rE20_x - rV1_x)*(rE21_y - rV1_y) - (rE20_y - rV1_y)*(rE21_x - rV1_x));  
  //rSinTheta = rV1onE2/sqrt(((rE20_x - rV1_x)*(rE20_x - rV1_x) + (rE20_y - rV1_y)*(rE20_y - rV1_y))*((rE21_x - rV1_x)*(rE21_x - rV1_x) + (rE21_y - rV1_y)*(rE21_y - rV1_y)));
  rSinTheta = SinThetaMPFUN(rV1_x, rV1_y, rE20_x, rE20_y, rE21_x, rE21_y);
  /*
  rSinThetaTemp = SinThetaMPFUN(rE20_x, rE20_y, rE21_x, rE21_y, rV1_x, rV1_y);
  if ( rSinThetaTemp > rSinTheta )
    {
      rSinTheta = rSinThetaTemp;
    }
  */
  if ( rSinTheta <= 0.0 )
    {
      // V1 is on E2 and it is not one of the endpoints
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  //rSinTheta = rE20xE21/sqrt(rE20dotE20*rE21dotE21); 
  rSinTheta = SinThetaMPFUN(rV0_x, rV0_y, rE20_x, rE20_y, rE21_x, rE21_y);
  /*
  rSinThetaTemp = SinThetaMPFUN(rE20_x, rE20_y, rE21_x, rE21_y, rV0_x, rV0_y);
  if ( rSinThetaTemp > rSinTheta )
    {
      rSinTheta = rSinThetaTemp;
    }
  */
  if ( rSinTheta <= 0.0 )
    {
      // V0 is on E1 and it is not one of the endpoints
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  rSinTheta = rE2xV1;
  if ( rSinTheta == 0.0 )
    {
      // The two vertices lie on a line parallel to the edge
      return InscribedCircleCenterVVEAntiparallelMPFUN(pdC_x, pdC_y, pdR, 
						       dV00_x, dV00_y, 
						       dV0_x, dV0_y, 
						       dV01_x, dV01_y, 
						       dV10_x, dV10_y, 
						       dV1_x, dV1_y, 
						       dV11_x, dV11_y, 
						       dE20_x, dE20_y, 
						       dE21_x, dE21_y);
    }

  rDen = 2.0*rE2xV1*rE2xV1;
  ASSERT( rDen != 0.0 );

  rDiscR = rV1dotV1*rE20xE21*( rE20xE21 + rE2xV1 );
  rDisc = rE2dotE2*rDiscR;
  if ( rDisc < 0.0 || rDiscR < 0.0 ) 
    {
      // Imaginary Solutions
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  mp_real rSqrtDisc = sqrt(rDisc);
  mp_real r2E20xE21timesE2dotV1 = 2.0*rE20xE21*rE2dotV1;


  //////////////////////////////////////////////////
  // Check the Edge Zone

  /*
  rMagCE20 = ( (prC[0] - ppprV[2][0][0])*(prC[0] - ppprV[2][0][0]) + 
	       (prC[1] - ppprV[2][0][1])*(prC[1] - ppprV[2][0][1]) );
  rMagCE21 = ( (prC[0] - ppprV[2][1][0])*(prC[0] - ppprV[2][1][0]) + 
	       (prC[1] - ppprV[2][1][1])*(prC[1] - ppprV[2][1][1]) );

  if ( rMagCE20 <= 0.0 )
    {
      prEdgeZone[0] = 0.0;
    }
  else
    {
      prEdgeZone[0] = -( ( rE2xV1*rE2dotE20 + rE20xE21*rE2dotV1 + rSqrtDisc ) / 
			 ( rE2xV1*sqrt(rE2dotE2*rMagCE20) ) );
    }
  if ( rMagCE21 <= 0.0 )
    {
      prEdgeZone[1] = 0.0;
    }
  else
    {
      prEdgeZone[1] = ( ( rE2xV1*rE2dotE21 + rE20xE21*rE2dotV1 + rSqrtDisc ) / 
			( rE2xV1*sqrt(rE2dotE2*rMagCE21) ) );    
    }
  */

  if ( rE2xV1 >= 0.0 )
    {
      prEdgeZone[0] = -( rE2xV1*rE2dotE20 + rE20xE21*rE2dotV1 + rSqrtDisc );
      prEdgeZone[1] = ( rE2xV1*rE2dotE21 + rE20xE21*rE2dotV1 + rSqrtDisc );
    }
  else
    {
      prEdgeZone[0] = ( rE2xV1*rE2dotE20 + rE20xE21*rE2dotV1 + rSqrtDisc );
      prEdgeZone[1] = -( rE2xV1*rE2dotE21 + rE20xE21*rE2dotV1 + rSqrtDisc );
    }

  for ( i = 0; i < 2; i++ )
    {
      if ( prEdgeZone[i] < 0.0 )
	{
	  *pdC_x = 0.0;
	  *pdC_y = 0.0;
	  *pdR = 0.0;
	  return APM2D::icctNull;
	}
    }

  //////////////////////////////////////////////////
  // Check the Vertex Zones

  pprVertexZone[0][0] = ( rV1dotV1*rE2xV1*rE2xV00 
			  + r2E20xE21timesE2dotV1*rV1xV00 
			  + 2.0*rV1xV00*rSqrtDisc );
  pprVertexZone[0][1] = ( rV1dotV1*rE2xV1*rE2xV01 
			  + r2E20xE21timesE2dotV1*rV1xV01 
			  + 2.0*rV1xV01*rSqrtDisc );
  pprVertexZone[1][0] = ( rV1dotV1*rE2xV1*rE2xV10 
			  + r2E20xE21timesE2dotV1*rV1xV10 
			  + 2.0*rV1xV10*rSqrtDisc
			  - 2.0*rE2xV1*rE2xV1*rV1dotV10 );
  pprVertexZone[1][1] = ( rV1dotV1*rE2xV1*rE2xV11 
			  + r2E20xE21timesE2dotV1*rV1xV11 
			  + 2.0*rV1xV11*rSqrtDisc
			  - 2.0*rE2xV1*rE2xV1*rV1dotV11 );

  for ( i = 0; i < 2; i++ )
    {
      for ( j = 0; j < 2; j++ )
	{
	  if ( pprVertexZone[i][j] < 0.0 )
	    {
	      *pdC_x = 0.0;
	      *pdC_y = 0.0;
	      *pdR = 0.0;
	      return APM2D::icctNull;
	    }
	}
    }

  //////////////////////////////////////////////////
  // Compute the center and radius

  prC[0] = ( -( rV1dotV1*rE2xV1*prE[1] + r2E20xE21timesE2dotV1*ppprV[1][0][1] )
	     - 2.0 * ppprV[1][0][1] * rSqrtDisc ) / rDen;
  prC[1] = ( ( rV1dotV1*rE2xV1*prE[0] + r2E20xE21timesE2dotV1*ppprV[1][0][0] )
	     + 2.0*ppprV[1][0][0]*rSqrtDisc ) / rDen;
  //prC[2] = sqrt((prC[0]*prC[0]) + (prC[1]*prC[1]));

  rR = ( rV1dotV1*sqrt(rE2dotE2) * (2.0*rE20xE21 + rE2xV1) 
	 + 2.0*rE2dotV1*sqrt(rDiscR) ) / rDen;
  
  /*
  if ( !( fabs(prC[2] - rR) < s_crEpsilonInscribedCircleCenterVVERadius ) )
    {
      fprintf(strerr, "VVE: C = ( %f, %f, %f ) != %f = R\n", prC[0], prC[1], prC[2], rR);
      //ASSERT( FALSE );
    }
  */

  // Translate the answer back out away from the origin
  prC[0] += rV0_x;
  prC[1] += rV0_y;

  *pdC_x = dble(prC[0]);
  *pdC_y = dble(prC[1]);
  *pdR = dble(rR);
  return APM2D::icctVVE;
}

//static 
APM2D::EdgeZoneType EdgeZoneVVEMPFUN(DOUBLE dV0_x, DOUBLE dV0_y,
				     DOUBLE dV1_x, DOUBLE dV1_y,
				     DOUBLE dE20_x, DOUBLE dE20_y, 
				     DOUBLE dE21_x, DOUBLE dE21_y,
				     DOUBLE dE30_x, DOUBLE dE30_y,
				     DOUBLE dE31_x, DOUBLE dE31_y)
{
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rV1_x = dV1_x;
  mp_real rV1_y = dV1_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;
  mp_real rE30_x = dE30_x;
  mp_real rE30_y = dE30_y;
  mp_real rE31_x = dE31_x;
  mp_real rE31_y = dE31_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real prEdgeZone[2];
  mp_real rV1dotV1, rE2dotE2, rV1dotE2;
  mp_real rE3dotE30, rE3dotE3, rE3dotE31;
  mp_real rE2xV1, rE2xE3, rE3xV1, rE20xE21;
  mp_real rE2dotV1;
  mp_real rDisc;
  mp_real rDen, rSinTheta;
  APM2D::EdgeZoneType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = rV1_x - rV0_x;
  ppprV[1][0][1] = rV1_y - rV0_y;
  ppprV[2][0][0] = rE20_x - rV0_x;
  ppprV[2][0][1] = rE20_y - rV0_y;
  ppprV[2][1][0] = rE21_x - rV0_x;
  ppprV[2][1][1] = rE21_y - rV0_y;
  ppprV[3][0][0] = rE30_x - rV0_x;
  ppprV[3][0][1] = rE30_y - rV0_y;
  ppprV[3][1][0] = rE31_x - rV0_x;
  ppprV[3][1][1] = rE31_y - rV0_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;
  pprE[3][0] = rE31_x - rE30_x;
  pprE[3][1] = rE31_y - rE30_y;
  
  rE2xV1 = (pprE[2][0]*ppprV[1][0][1]) - (pprE[2][1]*ppprV[1][0][0]);

  rV1dotV1 = (ppprV[1][0][0]*ppprV[1][0][0]) + (ppprV[1][0][1]*ppprV[1][0][1]);
  rV1dotE2 = (ppprV[1][0][0]*pprE[2][0]) + (ppprV[1][0][1]*pprE[2][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);
  rE3dotE3 = (pprE[3][0]*pprE[3][0]) + (pprE[3][1]*pprE[3][1]);

  rSinTheta = rE2xV1;
  if ( rSinTheta == 0.0 )
    {
      // The two vertices lie on a line parallel to the edge
      return EdgeZoneVVEAntiparallelMPFUN(dV0_x, dV0_y, 
					  dV1_x, dV1_y, 
					  dE20_x, dE20_y, dE21_x, dE21_y,
					  dE30_x, dE30_y, dE31_x, dE31_y);
    }

  rE3dotE30 = (pprE[3][0]*ppprV[3][0][0]) + (pprE[3][1]*ppprV[3][0][1]);
  rE3dotE31 = (pprE[3][0]*ppprV[3][1][0]) + (pprE[3][1]*ppprV[3][1][1]);
  
  rE2xE3 = (pprE[2][0]*pprE[3][1]) - (pprE[2][1]*pprE[3][0]);
  rE3xV1 = (pprE[3][0]*ppprV[1][0][1]) - (pprE[3][1]*ppprV[1][0][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);

  rDisc = rV1dotV1*rE2dotE2*rE20xE21*( rE20xE21 + rE2xV1 );
  ASSERT( rDisc >= 0.0 ); 
  mp_real rSqrtDisc = sqrt(rDisc);

  rDen = 2.0*rE2xV1*rE2xV1; 
  ASSERT( rDen != 0.0 );

  rE2dotV1 = (pprE[2][0]*ppprV[1][0][0]) + (pprE[2][1]*ppprV[1][0][1]);
  /*
  mp_real r2E20xE21timesE2dotV1 = 2.0*rE20xE21*rE2dotV1;
  mp_real rV1dotV1timesE2xV1 = rV1dotV1*rE2xV1;

  prC[0] = ( -( rV1dotV1timesE2xV1*pprE[2][1] + r2E20xE21timesE2dotV1*ppprV[1][0][1] )
	     - 2.0 * ppprV[1][0][1] * rSqrtDisc ) / rDen;
  prC[1] = ( ( rV1dotV1timesE2xV1*pprE[2][0] + r2E20xE21timesE2dotV1*ppprV[1][0][0] )
	     + 2.0*ppprV[1][0][0]*rSqrtDisc ) / rDen;

  rMagCE30 = ( (prC[0] - ppprV[3][0][0])*(prC[0] - ppprV[3][0][0]) + 
	       (prC[1] - ppprV[3][0][1])*(prC[1] - ppprV[3][0][1]) );
  rMagCE31 = ( (prC[0] - ppprV[3][1][0])*(prC[0] - ppprV[3][1][0]) + 
	       (prC[1] - ppprV[3][1][1])*(prC[1] - ppprV[3][1][1]) );

  if ( rMagCE30 <= 0.0 )
    {
      prEdgeZone[0] = 0.0;
    }
  else
    {
      prEdgeZone[0] = ( ( rE2xV1*( 2.0*rE2xV1*rE3dotE30 - rE2xE3*rV1dotV1 )
			  + 2.0*rE3xV1*( rE20xE21*rV1dotE2 + rSqrtDisc ) ) / 
			( rDen*sqrt(rE3dotE3*rMagCE30) ) );
    }
  if ( rMagCE31 <= 0.0 )
    {
      prEdgeZone[1] = 0.0;
    }
  else
    {
      prEdgeZone[1] = -( ( rE2xV1*( 2.0*rE2xV1*rE3dotE31 - rE2xE3*rV1dotV1 )
			   + 2.0*rE3xV1*( rE20xE21*rV1dotE2 + rSqrtDisc ) ) / 
			 ( rDen*sqrt(rE3dotE3*rMagCE31) ) );
    }
*/
  mp_real rE2xE3timesV1dotV1 = rE2xE3*rV1dotV1;
  mp_real rCoef = 2.0*rE3xV1*( rE20xE21*rV1dotE2 + rSqrtDisc );

  prEdgeZone[0] = ( rE2xV1*( 2.0*rE2xV1*rE3dotE30 - rE2xE3timesV1dotV1 )
		    + rCoef );
  prEdgeZone[1] = -( rE2xV1*( 2.0*rE2xV1*rE3dotE31 - rE2xE3timesV1dotV1 )
		     + rCoef );
  
  if ( !( prEdgeZone[0] < 0.0 ) )
    {
      eResult = APM2D::eztStart;
    }
  else if ( !( prEdgeZone[1] < 0.0 ) )
    {
      eResult = APM2D::eztEnd;
    }
  else
    {
      eResult = APM2D::eztMiddle;
    }

  return eResult;
}

//static 
APM2D::CircleTestType InCircleVVE_VMPFUN(DOUBLE dV0_x, DOUBLE dV0_y,
					 DOUBLE dV1_x, DOUBLE dV1_y,
					 DOUBLE dE20_x, DOUBLE dE20_y, 
					 DOUBLE dE21_x, DOUBLE dE21_y,
					 DOUBLE dV3_x, DOUBLE dV3_y)
{
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rV1_x = dV1_x;
  mp_real rV1_y = dV1_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;
  mp_real rV3_x = dV3_x;
  mp_real rV3_y = dV3_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real rV1dotV1, rE2dotE2, rV3dotV3, rV1dotE2;
  mp_real rE2xV1, rE2xV3, rV1xV3, rE20xE21;
  mp_real rDisc;
  //mp_real rDen;
  mp_real rSinTheta;
  mp_real rResult;
  APM2D::CircleTestType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = rV1_x - rV0_x;
  ppprV[1][0][1] = rV1_y - rV0_y;
  ppprV[2][0][0] = rE20_x - rV0_x;
  ppprV[2][0][1] = rE20_y - rV0_y;
  ppprV[2][1][0] = rE21_x - rV0_x;
  ppprV[2][1][1] = rE21_y - rV0_y;
  ppprV[3][0][0] = rV3_x - rV0_x;
  ppprV[3][0][1] = rV3_y - rV0_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;

  rV1dotV1 = (ppprV[1][0][0]*ppprV[1][0][0]) + (ppprV[1][0][1]*ppprV[1][0][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);
  rV3dotV3 = (ppprV[3][0][0]*ppprV[3][0][0]) + (ppprV[3][0][1]*ppprV[3][0][1]);
  rV1dotE2 = (ppprV[1][0][0]*pprE[2][0]) + (ppprV[1][0][1]*pprE[2][1]);

  rE2xV1 = (pprE[2][0]*ppprV[1][0][1]) - (pprE[2][1]*ppprV[1][0][0]);
  rSinTheta = rE2xV1;
  if ( rSinTheta == 0.0 )
    {
      // The two vertices lie on a line parallel to the edge
      return InCircleVVE_VAntiparallelMPFUN(dV0_x, dV0_y, 
					    dV1_x, dV1_y, 
					    dE20_x, dE20_y, dE21_x, dE21_y,
					    dV3_x, dV3_y);
    }

  rE2xV3 = (pprE[2][0]*ppprV[3][0][1]) - (pprE[2][1]*ppprV[3][0][0]);
  rV1xV3 = (ppprV[1][0][0]*ppprV[3][0][1]) - (ppprV[1][0][1]*ppprV[3][0][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);

  rDisc = rV1dotV1*rE2dotE2*rE20xE21*( rE20xE21 + rE2xV1 );
  ASSERT( rDisc >= 0.0 ); 
  mp_real rSqrtDisc = sqrt(rDisc);

  //rDen = rE2xV1*rE2xV1; 
  //ASSERT( rDen > 0.0 );

  rResult = ( rE2xV1*( rV3dotV3*rE2xV1 - rV1dotV1*rE2xV3 ) 
	      - 2.0*rV1xV3*( rE20xE21*rV1dotE2 + rSqrtDisc ) );
  
  if ( rResult < 0.0 )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( rResult > 0.0 )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

//static 
APM2D::CircleTestType InCircleVVE_EMPFUN(DOUBLE dV0_x, DOUBLE dV0_y,
					 DOUBLE dV1_x, DOUBLE dV1_y,
					 DOUBLE dE20_x, DOUBLE dE20_y, 
					 DOUBLE dE21_x, DOUBLE dE21_y,
					 DOUBLE dE30_x, DOUBLE dE30_y,
					 DOUBLE dE31_x, DOUBLE dE31_y)
{
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rV1_x = dV1_x;
  mp_real rV1_y = dV1_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;
  mp_real rE30_x = dE30_x;
  mp_real rE30_y = dE30_y;
  mp_real rE31_x = dE31_x;
  mp_real rE31_y = dE31_y;
  
  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real rV1dotV1, rE2dotE2, rE3dotE3;
  mp_real rV1dotE2, rV1dotE3, rE2dotE3;
  mp_real rE2xV1, rE2xE3, rE20xE21, rE30xE31;
  mp_real rDisc, rDiscR, rA, rB;
  mp_real rResult, rSinTheta;
  APM2D::CircleTestType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = rV1_x - rV0_x;
  ppprV[1][0][1] = rV1_y - rV0_y;
  ppprV[2][0][0] = rE20_x - rV0_x;
  ppprV[2][0][1] = rE20_y - rV0_y;
  ppprV[2][1][0] = rE21_x - rV0_x;
  ppprV[2][1][1] = rE21_y - rV0_y;
  ppprV[3][0][0] = rE30_x - rV0_x;
  ppprV[3][0][1] = rE30_y - rV0_y;
  ppprV[3][1][0] = rE31_x - rV0_x;
  ppprV[3][1][1] = rE31_y - rV0_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;
  pprE[3][0] = rE31_x - rE30_x;
  pprE[3][1] = rE31_y - rE30_y;

  rV1dotV1 = (ppprV[1][0][0]*ppprV[1][0][0]) + (ppprV[1][0][1]*ppprV[1][0][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);
  rE3dotE3 = (pprE[3][0]*pprE[3][0]) + (pprE[3][1]*pprE[3][1]);

  rE2xV1 = (pprE[2][0]*ppprV[1][0][1]) - (pprE[2][1]*ppprV[1][0][0]);
  rSinTheta = rE2xV1;
  if ( rSinTheta == 0.0 )
    {
      // The two vertices lie on a line parallel to the edge
      return InCircleVVE_EAntiparallelMPFUN(dV0_x, dV0_y, 
					    dV1_x, dV1_y, 
					    dE20_x, dE20_y, dE21_x, dE21_y,
					    dE30_x, dE30_y, dE31_x, dE31_y);
    }

  rV1dotE2 = (ppprV[1][0][0]*pprE[2][0]) + (ppprV[1][0][1]*pprE[2][1]);
  rV1dotE3 = (ppprV[1][0][0]*pprE[3][0]) + (ppprV[1][0][1]*pprE[3][1]);
  rE2dotE3 = (pprE[2][0]*pprE[3][0]) + (pprE[2][1]*pprE[3][1]);

  rE2xE3 = (pprE[2][0]*pprE[3][1]) - (pprE[2][1]*pprE[3][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);
  rE30xE31 = (ppprV[3][0][0]*ppprV[3][1][1]) - (ppprV[3][0][1]*ppprV[3][1][0]);

  rDisc = rV1dotV1*rE2dotE2*rE20xE21*( rE20xE21 + rE2xV1 );
  ASSERT( rDisc >= 0.0 ); 
  mp_real rSqrtDisc = sqrt(rDisc);

  rDiscR = rV1dotV1*rE20xE21*( rE20xE21 + rE2xV1 );
  ASSERT( rDiscR >= 0.0 ); 

  rA = ( rV1dotV1*rE2xV1*rE2dotE3 + 2.0*rV1dotE3*( rE20xE21*rV1dotE2 + rSqrtDisc ) );
  rB = ( rV1dotV1*sqrt(rE2dotE2)*( 2.0*rE20xE21 + rE2xV1 ) + 2.0*rV1dotE2*sqrt(rDiscR) );

  //rDen = 4.0*rE3dotE3*rE2xV1*rE2xV1*rE2xV1*rE2xV1;
  //ASSERT( rDen != 0.0 );

  mp_real rE30xE31timesE2xV1timesE2xV1 = rE30xE31*rE2xV1*rE2xV1;

  rResult = ( rA*rA
	      + 4.0*rE30xE31timesE2xV1timesE2xV1*rA
	      + 4.0*rE30xE31timesE2xV1timesE2xV1*rE30xE31timesE2xV1timesE2xV1
	      - rE3dotE3*rB*rB );
  
  if ( rResult < 0.0 )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( rResult > 0.0 )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

//static 
APM2D::InscribedCircleCenterType 
InscribedCircleCenterVVEAntiparallelMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					  DOUBLE dV00_x, DOUBLE dV00_y,
					  DOUBLE dV0_x, DOUBLE dV0_y,
					  DOUBLE dV01_x, DOUBLE dV01_y,
					  DOUBLE dV10_x, DOUBLE dV10_y,
					  DOUBLE dV1_x, DOUBLE dV1_y,
					  DOUBLE dV11_x, DOUBLE dV11_y,
					  DOUBLE dE20_x, DOUBLE dE20_y, 
					  DOUBLE dE21_x, DOUBLE dE21_y)
{
  mp_real prC[3], rR;
  mp_real rV00_x = dV00_x;
  mp_real rV00_y = dV00_y;
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rV01_x = dV01_x;
  mp_real rV01_y = dV01_y;
  mp_real rV10_x = dV10_x;
  mp_real rV10_y = dV10_y;
  mp_real rV1_x = dV1_x;
  mp_real rV1_y = dV1_y;
  mp_real rV11_x = dV11_x;
  mp_real rV11_y = dV11_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;

  mp_real ppprV[3][2][2], ppprVN[2][2][2], prE[2];
  mp_real rDen, rDenR;
  mp_real rV1dotV1, rE2dotE2, rE2dotV1, rE2dotE20, rE2dotE21;
  mp_real rE20xE21;
  mp_real rV1xV00, rV1xV01, rV1xV10, rV1xV11;
  mp_real rV1dotV00, rV1dotV01, rV1dotV10, rV1dotV11;
  mp_real rV00dotV00, rV01dotV01, rV10dotV10, rV11dotV11;
  mp_real rE20dotE20, rE21dotE21;
  mp_real rV1onE2, rSinTheta;
  mp_real pprVertexZone[2][2];
  mp_real prEdgeZone[2];
  UINT_32 i, j;
  
  // Translate V0 to the origin
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[1][0][0] = rV1_x - rV0_x;
  ppprV[1][0][1] = rV1_y - rV0_y;
  ppprV[2][0][0] = rE20_x - rV0_x;
  ppprV[2][0][1] = rE20_y - rV0_y;
  ppprV[2][1][0] = rE21_x - rV0_x;
  ppprV[2][1][1] = rE21_y - rV0_y;

  // Vertex neighbor vectors
  //memset(ppprVN, 0, sizeof(ppprV));
  ppprVN[0][0][0] = rV0_x - rV00_x;
  ppprVN[0][0][1] = rV0_y - rV00_y;
  ppprVN[0][1][0] = rV0_x - rV01_x;
  ppprVN[0][1][1] = rV0_y - rV01_y;
  ppprVN[1][0][0] = rV1_x - rV10_x;
  ppprVN[1][0][1] = rV1_y - rV10_y;
  ppprVN[1][1][0] = rV1_x - rV11_x;
  ppprVN[1][1][1] = rV1_y - rV11_y;
  
  prE[0] = rE21_x - rE20_x;
  prE[1] = rE21_y - rE20_y;
  
  rV1dotV1 = (ppprV[1][0][0]*ppprV[1][0][0]) + (ppprV[1][0][1]*ppprV[1][0][1]);
  rE2dotE2 = (prE[0]*prE[0]) + (prE[1]*prE[1]);
  rE2dotV1 = (prE[0]*ppprV[1][0][0]) + (prE[1]*ppprV[1][0][1]);

  rE2dotE20 = (prE[0]*ppprV[2][0][0]) + (prE[1]*ppprV[2][0][1]);
  rE2dotE21 = (prE[0]*ppprV[2][1][0]) + (prE[1]*ppprV[2][1][1]);

  rE20dotE20 = (ppprV[2][0][0]*ppprV[2][0][0]) + (ppprV[2][0][1]*ppprV[2][0][1]);
  rE21dotE21 = (ppprV[2][1][0]*ppprV[2][1][0]) + (ppprV[2][1][1]*ppprV[2][1][1]);

  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);
  
  rV00dotV00 = ( (ppprVN[0][0][0] * ppprVN[0][0][0]) + (ppprVN[0][0][1] * ppprVN[0][0][1]) );
  rV01dotV01 = ( (ppprVN[0][1][0] * ppprVN[0][1][0]) + (ppprVN[0][1][1] * ppprVN[0][1][1]) );
  rV10dotV10 = ( (ppprVN[1][0][0] * ppprVN[1][0][0]) + (ppprVN[1][0][1] * ppprVN[1][0][1]) );
  rV11dotV11 = ( (ppprVN[1][1][0] * ppprVN[1][1][0]) + (ppprVN[1][1][1] * ppprVN[1][1][1]) );

  rV1dotV00 = ( (ppprV[1][0][0] * ppprVN[0][0][0]) + (ppprV[1][0][1] * ppprVN[0][0][1]) );
  rV1dotV01 = ( (ppprV[1][0][0] * ppprVN[0][1][0]) + (ppprV[1][0][1] * ppprVN[0][1][1]) );
  rV1dotV10 = ( (ppprV[1][0][0] * ppprVN[1][0][0]) + (ppprV[1][0][1] * ppprVN[1][0][1]) );
  rV1dotV11 = ( (ppprV[1][0][0] * ppprVN[1][1][0]) + (ppprV[1][0][1] * ppprVN[1][1][1]) );

  rV1xV00 = ( (ppprV[1][0][0] * ppprVN[0][0][1]) - (ppprV[1][0][1] * ppprVN[0][0][0]) );
  rV1xV01 = ( (ppprV[1][0][0] * ppprVN[0][1][1]) - (ppprV[1][0][1] * ppprVN[0][1][0]) );
  rV1xV10 = ( (ppprV[1][0][0] * ppprVN[1][0][1]) - (ppprV[1][0][1] * ppprVN[1][0][0]) );
  rV1xV11 = ( (ppprV[1][0][0] * ppprVN[1][1][1]) - (ppprV[1][0][1] * ppprVN[1][1][0]) );

  rV1onE2 = ((rE20_x - rV1_x)*(rE21_y - rV1_y) - (rE20_y - rV1_y)*(rE21_x - rV1_x));
  //rSinTheta = rV1onE2/sqrt(((rE20_x - rV1_x)*(rE20_x - rV1_x) + (rE20_y - rV1_y)*(rE20_y - rV1_y))*((rE21_x - rV1_x)*(rE21_x - rV1_x) + (rE21_y - rV1_y)*(rE21_y - rV1_y)));
  rSinTheta = SinThetaMPFUN(rV1_x, rV1_y, rE20_x, rE20_y, rE21_x, rE21_y);
  /*
  rSinThetaTemp = SinThetaMPFUN(rE20_x, rE20_y, rE21_x, rE21_y, rV1_x, rV1_y);
  if ( rSinThetaTemp > rSinTheta )
    {
      rSinTheta = rSinThetaTemp;
    }
  */
  if ( rSinTheta <= 0.0 )
    {
      // V1 is on E2 and it is not one of the endpoints
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  //rSinTheta = rE20xE21/sqrt(rE20dotE20*rE21dotE21);
  rSinTheta = SinThetaMPFUN(rV0_x, rV0_y, rE20_x, rE20_y, rE21_x, rE21_y);
  /*
  rSinThetaTemp = SinThetaMPFUN(rE20_x, rE20_y, rE21_x, rE21_y, rV0_x, rV0_y);
  if ( rSinThetaTemp > rSinTheta )
    {
      rSinTheta = rSinThetaTemp;
    }
  */
  if ( rSinTheta <= 0.0 )
    {
      // V0 is on E2 and it is not one of the endpoints
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  rDen = -8.0*rE20xE21*rE2dotV1;
  rDenR = 8.0*rE20xE21*sqrt(rE2dotE2);
  if ( rDen == 0.0 || rDenR == 0.0 )
    {
      ASSERT( FALSE );
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  mp_real r4E20xE21timesE20xE21minusE2dotV1timesE2dotV1 = 4.0*rE20xE21*rE20xE21 - rE2dotV1*rE2dotV1;

  //////////////////////////////////////////////////
  // Check the Edge Zone
  /*
  rMagCE20 = ( (prC[0] - ppprV[2][0][0])*(prC[0] - ppprV[2][0][0]) + 
	       (prC[1] - ppprV[2][0][1])*(prC[1] - ppprV[2][0][1]) );
  rMagCE21 = ( (prC[0] - ppprV[2][1][0])*(prC[0] - ppprV[2][1][0]) + 
	       (prC[1] - ppprV[2][1][1])*(prC[1] - ppprV[2][1][1]) );

  if ( rMagCE20 <= 0.0 )
    {
      prEdgeZone[0] = 0.0;
    }
  else
    {
      prEdgeZone[0] = -( ( 2.0*rE2dotE20 - rE2dotV1 ) / 
			 ( sqrt(rE2dotE2*rMagCE20) ) );
    }
  if ( rMagCE21 <= 0.0 )
    {
      prEdgeZone[1] = 0.0;
    }
  else
    {
      prEdgeZone[1] = ( ( 2.0*rE2dotE21 - rE2dotV1 ) /
			( sqrt(rE2dotE2*rMagCE21) ) );
    }
  */
  prEdgeZone[0] = -( 2.0*rE2dotE20 - rE2dotV1 );
  prEdgeZone[1] = ( 2.0*rE2dotE21 - rE2dotV1 );

  for ( i = 0; i < 2; i++ )
    {
      if ( prEdgeZone[i] < 0.0 )
	{
	  *pdC_x = 0.0;
	  *pdC_y = 0.0;
	  *pdR = 0.0;
	  return APM2D::icctNull;
	}
    }

  //////////////////////////////////////////////////
  // Check the Vertex Zones

  mp_real r4E20xE21timesE2dotV1 = 4.0*rE20xE21*rE2dotV1;

  pprVertexZone[0][0] = ( (r4E20xE21timesE20xE21minusE2dotV1timesE2dotV1)*rV1xV00
			  - r4E20xE21timesE2dotV1*rV1dotV00 );
  pprVertexZone[0][1] = ( (r4E20xE21timesE20xE21minusE2dotV1timesE2dotV1)*rV1xV01
			  - r4E20xE21timesE2dotV1*rV1dotV01 );
  pprVertexZone[1][0] = ( (r4E20xE21timesE20xE21minusE2dotV1timesE2dotV1)*rV1xV10
			  + r4E20xE21timesE2dotV1*rV1dotV10 );
  pprVertexZone[1][1] = ( (r4E20xE21timesE20xE21minusE2dotV1timesE2dotV1)*rV1xV11
			  + r4E20xE21timesE2dotV1*rV1dotV11 );

  for ( i = 0; i < 2; i++ )
    {
      for ( j = 0; j < 2; j++ )
	{
	  if ( pprVertexZone[i][j] < 0.0 )
	    {
	      *pdC_x = 0.0;
	      *pdC_y = 0.0;
	      *pdR = 0.0;
	      return APM2D::icctNull;
	    }
	}
    }

  //////////////////////////////////////////////////
  // Compute the center and radius

  prC[0] = ( -(r4E20xE21timesE20xE21minusE2dotV1timesE2dotV1)*ppprV[1][0][1]
	     - (4.0*rE20xE21*rE2dotV1)*ppprV[1][0][0] ) / rDen;
  prC[1] = ( (r4E20xE21timesE20xE21minusE2dotV1timesE2dotV1)*ppprV[1][0][0]
	     - (4.0*rE20xE21*rE2dotV1)*ppprV[1][0][1] ) / rDen;
  //prC[2] = sqrt((prC[0]*prC[0]) + (prC[1]*prC[1]));

  rR = (4.0*rE20xE21*rE20xE21 + rE2dotV1*rE2dotV1) / rDenR;
  
  /*
  if ( !( fabs(prC[2] - rR) < s_crEpsilonInscribedCircleCenterVVEAntiparallelRadius ) )
    {
      fprintf(strerr, "VVE Antiparallel: C = ( %f, %f, %f ) != %f = R\n", prC[0], prC[1], prC[2], rR);
      //ASSERT( FALSE );
    }
  */

  // Translate the answer back out away from the origin
  prC[0] += rV0_x;
  prC[1] += rV0_y;

  *pdC_x = dble(prC[0]);
  *pdC_y = dble(prC[1]);
  *pdR = dble(rR);
  return APM2D::icctVVEAntiparallel;
}

//static 
APM2D::EdgeZoneType EdgeZoneVVEAntiparallelMPFUN(DOUBLE dV0_x, DOUBLE dV0_y,
						 DOUBLE dV1_x, DOUBLE dV1_y,
						 DOUBLE dE20_x, DOUBLE dE20_y, 
						 DOUBLE dE21_x, DOUBLE dE21_y,
						 DOUBLE dE30_x, DOUBLE dE30_y,
						 DOUBLE dE31_x, DOUBLE dE31_y)
{
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rV1_x = dV1_x;
  mp_real rV1_y = dV1_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;
  mp_real rE30_x = dE30_x;
  mp_real rE30_y = dE30_y;
  mp_real rE31_x = dE31_x;
  mp_real rE31_y = dE31_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real prEdgeZone[2];
  mp_real rV1dotV1, rE2dotE2, rV1dotE3;
  mp_real rE3dotE30, rE3dotE31;
  mp_real rE3xV1, rE20xE21;
  mp_real rE3dotE3;
  mp_real rE2dotV1;
  mp_real rDen;
  APM2D::EdgeZoneType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = rV1_x - rV0_x;
  ppprV[1][0][1] = rV1_y - rV0_y;
  ppprV[2][0][0] = rE20_x - rV0_x;
  ppprV[2][0][1] = rE20_y - rV0_y;
  ppprV[2][1][0] = rE21_x - rV0_x;
  ppprV[2][1][1] = rE21_y - rV0_y;
  ppprV[3][0][0] = rE30_x - rV0_x;
  ppprV[3][0][1] = rE30_y - rV0_y;
  ppprV[3][1][0] = rE31_x - rV0_x;
  ppprV[3][1][1] = rE31_y - rV0_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;
  pprE[3][0] = rE31_x - rE30_x;
  pprE[3][1] = rE31_y - rE30_y;
  
  rV1dotV1 = (ppprV[1][0][0]*ppprV[1][0][0]) + (ppprV[1][0][1]*ppprV[1][0][1]);
  rV1dotE3 = (ppprV[1][0][0]*pprE[3][0]) + (ppprV[1][0][1]*pprE[3][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);
  rE3dotE3 = (pprE[3][0]*pprE[3][0]) + (pprE[3][1]*pprE[3][1]);

  rE3dotE30 = (pprE[3][0]*ppprV[3][0][0]) + (pprE[3][1]*ppprV[3][0][1]);
  rE3dotE31 = (pprE[3][0]*ppprV[3][1][0]) + (pprE[3][1]*ppprV[3][1][1]);
  
  rE3xV1 = (pprE[3][0]*ppprV[1][0][1]) - (pprE[3][1]*ppprV[1][0][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);

  mp_real r4E20xE21timesSqrtrV1dotV1timesE2dotE2 = 4.0*rE20xE21*sqrt(rV1dotV1*rE2dotE2);
  rDen = 2.0*r4E20xE21timesSqrtrV1dotV1timesE2dotE2;
  ASSERT( rDen != 0.0 );

  rE2dotV1 = (pprE[2][0]*ppprV[1][0][0]) + (pprE[2][1]*ppprV[1][0][1]);

  mp_real r4E20xE21timesE20xE21minusV1dotV1timesE2dotE2 = 4.0*rE20xE21*rE20xE21 - rV1dotV1*rE2dotE2;

  /*
  mp_real r4E20xE21timesE20xE21minusE2dotV1timesE2dotV1 = 4.0*rE20xE21*rE20xE21 - rE2dotV1*rE2dotV1;
  mp_real r4E20xE21timesE2dotV1 = 4.0*rE20xE21*rE2dotV1;

  prC[0] = ( -(r4E20xE21timesE20xE21minusE2dotV1timesE2dotV1)*ppprV[1][0][1]
	     - (r4E20xE21timesE2dotV1)*ppprV[1][0][0] ) / rDen;
  prC[1] = ( (r4E20xE21timesE20xE21minusE2dotV1timesE2dotV1)*ppprV[1][0][0]
	     - (r4E20xE21timesE2dotV1)*ppprV[1][0][1] ) / rDen;
  
  rMagCE30 = ( (prC[0] - ppprV[3][0][0])*(prC[0] - ppprV[3][0][0]) + 
	       (prC[1] - ppprV[3][0][1])*(prC[1] - ppprV[3][0][1]) );
  rMagCE31 = ( (prC[0] - ppprV[3][1][0])*(prC[0] - ppprV[3][1][0]) + 
	       (prC[1] - ppprV[3][1][1])*(prC[1] - ppprV[3][1][1]) );

  if ( rMagCE30 <= 0.0 )
    {
      prEdgeZone[0] = 0.0;
    }
  else
    {
      prEdgeZone[0] = ( ( r4E20xE21timesSqrtrV1dotV1timesE2dotE2*( 2.0*rE3dotE30 - rV1dotE3 )
			  + ( r4E20xE21timesE20xE21minusV1dotV1timesE2dotE2 )*rE3xV1 ) / 
			( rDen*sqrt(rE3dotE3*rMagCE30) ) );
    }
  if ( rMagCE31 <= 0.0 )
    {
      prEdgeZone[1] = 0.0;
    }
  else
    {
      prEdgeZone[1] = -( ( r4E20xE21timesSqrtrV1dotV1timesE2dotE2*( 2.0*rE3dotE31 - rV1dotE3 )
			   + ( r4E20xE21timesE20xE21minusV1dotV1timesE2dotE2 )*rE3xV1 ) / 
			 ( rDen*sqrt(rE3dotE3*rMagCE31) ) );
    }
  */

  prEdgeZone[0] = ( r4E20xE21timesSqrtrV1dotV1timesE2dotE2*( 2.0*rE3dotE30 - rV1dotE3 )
		    + ( r4E20xE21timesE20xE21minusV1dotV1timesE2dotE2 )*rE3xV1 );
  prEdgeZone[1] = -( r4E20xE21timesSqrtrV1dotV1timesE2dotE2*( 2.0*rE3dotE31 - rV1dotE3 )
		     + ( r4E20xE21timesE20xE21minusV1dotV1timesE2dotE2 )*rE3xV1 );

  if ( !( prEdgeZone[0] < 0.0 ) )
    {
      eResult = APM2D::eztStart;
    }
  else if ( !( prEdgeZone[1] < 0.0 ) )
    {
      eResult = APM2D::eztEnd;
    }
  else
    {
      eResult = APM2D::eztMiddle;
    }

  return eResult;
}

//static 
APM2D::CircleTestType InCircleVVE_VAntiparallelMPFUN(DOUBLE dV0_x, DOUBLE dV0_y,
						     DOUBLE dV1_x, DOUBLE dV1_y,
						     DOUBLE dE20_x, DOUBLE dE20_y, 
						     DOUBLE dE21_x, DOUBLE dE21_y,
						     DOUBLE dV3_x, DOUBLE dV3_y)
{
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rV1_x = dV1_x;
  mp_real rV1_y = dV1_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;
  mp_real rV3_x = dV3_x;
  mp_real rV3_y = dV3_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real rV1dotV1, rE2dotE2, rV3dotV3, rV1dotV3;
  mp_real rV1xV3, rE20xE21;
  mp_real rDen;
  mp_real rResult;
  APM2D::CircleTestType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = rV1_x - rV0_x;
  ppprV[1][0][1] = rV1_y - rV0_y;
  ppprV[2][0][0] = rE20_x - rV0_x;
  ppprV[2][0][1] = rE20_y - rV0_y;
  ppprV[2][1][0] = rE21_x - rV0_x;
  ppprV[2][1][1] = rE21_y - rV0_y;
  ppprV[3][0][0] = rV3_x - rV0_x;
  ppprV[3][0][1] = rV3_y - rV0_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;

  rV1dotV1 = (ppprV[1][0][0]*ppprV[1][0][0]) + (ppprV[1][0][1]*ppprV[1][0][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);
  rV3dotV3 = (ppprV[3][0][0]*ppprV[3][0][0]) + (ppprV[3][0][1]*ppprV[3][0][1]);
  rV1dotV3 = (ppprV[1][0][0]*ppprV[3][0][0]) + (ppprV[1][0][1]*ppprV[3][0][1]);

  rV1xV3 = (ppprV[1][0][0]*ppprV[3][0][1]) - (ppprV[1][0][1]*ppprV[3][0][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);

  rDen = 4.0*rE20xE21*sqrt(rV1dotV1*rE2dotE2);
  ASSERT( rDen != 0.0 );

  rResult = ( rDen*( rV3dotV3 - rV1dotV3 )
	      - ( 4.0*rE20xE21*rE20xE21 - rV1dotV1*rE2dotE2 )*rV1xV3 );
  
  if ( rResult < 0.0 )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( rResult > 0.0 )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

//static 
APM2D::CircleTestType InCircleVVE_EAntiparallelMPFUN(DOUBLE dV0_x, DOUBLE dV0_y,
						     DOUBLE dV1_x, DOUBLE dV1_y,
						     DOUBLE dE20_x, DOUBLE dE20_y, 
						     DOUBLE dE21_x, DOUBLE dE21_y,
						     DOUBLE dE30_x, DOUBLE dE30_y,
						     DOUBLE dE31_x, DOUBLE dE31_y)
{
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rV1_x = dV1_x;
  mp_real rV1_y = dV1_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;
  mp_real rE30_x = dE30_x;
  mp_real rE30_y = dE30_y;
  mp_real rE31_x = dE31_x;
  mp_real rE31_y = dE31_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real rV1dotV1, rE2dotE2, rE3dotE3;
  mp_real rV1dotE2, rV1dotE3;
  mp_real rE3xV1, rE20xE21, rE30xE31;
  //mp_real rDen;
  mp_real rResult;
  APM2D::CircleTestType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = rV1_x - rV0_x;
  ppprV[1][0][1] = rV1_y - rV0_y;
  ppprV[2][0][0] = rE20_x - rV0_x;
  ppprV[2][0][1] = rE20_y - rV0_y;
  ppprV[2][1][0] = rE21_x - rV0_x;
  ppprV[2][1][1] = rE21_y - rV0_y;
  ppprV[3][0][0] = rE30_x - rV0_x;
  ppprV[3][0][1] = rE30_y - rV0_y;
  ppprV[3][1][0] = rE31_x - rV0_x;
  ppprV[3][1][1] = rE31_y - rV0_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;
  pprE[3][0] = rE31_x - rE30_x;
  pprE[3][1] = rE31_y - rE30_y;

  rV1dotV1 = (ppprV[1][0][0]*ppprV[1][0][0]) + (ppprV[1][0][1]*ppprV[1][0][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);
  rE3dotE3 = (pprE[3][0]*pprE[3][0]) + (pprE[3][1]*pprE[3][1]);

  rV1dotE2 = (ppprV[1][0][0]*pprE[2][0]) + (ppprV[1][0][1]*pprE[2][1]);
  rV1dotE3 = (ppprV[1][0][0]*pprE[3][0]) + (ppprV[1][0][1]*pprE[3][1]);

  rE3xV1 = (pprE[3][0]*ppprV[1][0][1]) - (pprE[3][1]*ppprV[1][0][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);
  rE30xE31 = (ppprV[3][0][0]*ppprV[3][1][1]) - (ppprV[3][0][1]*ppprV[3][1][0]);

  //rDen = 64.0*rE20xE21*rE20xE21*rV1dotE2*rV1dotE2*rE3dotE3;
  //ASSERT( rDen != 0.0 );

  rResult = ( 64.0*rE20xE21*rE20xE21*rV1dotE2*rV1dotE2*rE30xE31*( rE30xE31 + rE3xV1 )
	      - rE3xV1*rE3xV1*( 4.0*rE20xE21*rE20xE21 - rV1dotE2*rV1dotE2 )*( 4.0*rE20xE21*rE20xE21 - rV1dotE2*rV1dotE2 )
	      - 16.0*rE20xE21*rE20xE21*rV1dotE2*rV1dotE2*rV1dotE3*rV1dotE3
	      - 8.0*rE20xE21*rV1dotE2*rV1dotE3*( 4.0*rE20xE21*rE20xE21 - rV1dotE2*rV1dotE2 )*( 2.0*rE30xE31 + rE3xV1 ) );

  if ( rResult < 0.0 )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( rResult > 0.0 )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

//static 
APM2D::InscribedCircleCenterType 
InscribedCircleCenterVvEMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			      DOUBLE dV00_x, DOUBLE dV00_y,
			      DOUBLE dV0_x, DOUBLE dV0_y,
			      DOUBLE dV01_x, DOUBLE dV01_y,
			      DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
			      DOUBLE dE21_x, DOUBLE dE21_y)
{
  mp_real prC[3], rR;
  mp_real rV00_x = dV00_x;
  mp_real rV00_y = dV00_y;
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rV01_x = dV01_x;
  mp_real rV01_y = dV01_y;
  mp_real rV1_E20_x = dV1_E20_x;
  mp_real rV1_E20_y = dV1_E20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;

  mp_real ppprV[3][2][2], ppprVN[1][2][2], pprE[3][2];
  mp_real rDen;
  mp_real rV0dotV0, rE2dotE2;
  mp_real rE2xV0;
  mp_real rV00dotV00, rV01dotV01, rV0dotV00, rV0dotV01;
  mp_real rE2xV00, rE2xV01;
  mp_real prVertexZone[2], rSinTheta;
  UINT_32 i;
  
  // Translate V1 to the origin
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = rV0_x - rV1_E20_x;
  ppprV[0][0][1] = rV0_y - rV1_E20_y;
  ppprV[2][0][0] = 0.0;
  ppprV[2][0][1] = 0.0;
  ppprV[2][1][0] = rE21_x - rV1_E20_x;
  ppprV[2][1][1] = rE21_y - rV1_E20_y;
  
  // Vertex neighbor vectors
  //memset(ppprVN, 0, sizeof(ppprV));
  ppprVN[0][0][0] = rV0_x - rV00_x;
  ppprVN[0][0][1] = rV0_y - rV00_y;
  ppprVN[0][1][0] = rV0_x - rV01_x;
  ppprVN[0][1][1] = rV0_y - rV01_y;

  //memset(pprE, 0, sizeof(pprE));
  pprE[2][0] = rE21_x - rV1_E20_x;
  pprE[2][1] = rE21_y - rV1_E20_y;
  
  rV0dotV0 = (ppprV[0][0][0]*ppprV[0][0][0]) + (ppprV[0][0][1]*ppprV[0][0][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0] + pprE[2][1]*pprE[2][1]);
  rE2xV0 = (pprE[2][0]*ppprV[0][0][1]) - (pprE[2][1]*ppprV[0][0][0]);
  
  rV00dotV00 = ( (ppprVN[0][0][0] * ppprVN[0][0][0]) + (ppprVN[0][0][1] * ppprVN[0][0][1]) );
  rV01dotV01 = ( (ppprVN[0][1][0] * ppprVN[0][1][0]) + (ppprVN[0][1][1] * ppprVN[0][1][1]) );

  rV0dotV00 = ( (ppprV[0][0][0] * ppprVN[0][0][0]) + (ppprV[0][0][1] * ppprVN[0][0][1]) );
  rV0dotV01 = ( (ppprV[0][0][0] * ppprVN[0][1][0]) + (ppprV[0][0][1] * ppprVN[0][1][1]) );

  rE2xV00 = ( (pprE[2][0] * ppprVN[0][0][1]) - (pprE[2][1] * ppprVN[0][0][0]) );
  rE2xV01 = ( (pprE[2][0] * ppprVN[0][1][1]) - (pprE[2][1] * ppprVN[0][1][0]) );

  //rSinTheta = rE2xV0/sqrt(rE2dotE2*rV0dotV0);
  rSinTheta = SinThetaMPFUN(rV0_x, rV0_y, rV1_E20_x, rV1_E20_y, rE21_x, rE21_y);
  /*
  rSinThetaTemp = SinThetaMPFUN(rV1_E20_x, rV1_E20_y, rE21_x, rE21_y, rV0_x, rV0_y);
  if ( rSinThetaTemp > rSinTheta )
    {
      rSinTheta = rSinThetaTemp;
    }
  */
  if ( rSinTheta <= 0.0 )
    {
      // V0 is on E2 and it is not one of the endpoints
      
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  rDen = 2.0*rE2xV0;
  ASSERT( rDen != 0.0 );

  //////////////////////////////////////////////////
  // Check the Vertex Zones

  prVertexZone[0] = ( rV0dotV0*rE2xV00 - 2.0*rE2xV0*rV0dotV00 );
  prVertexZone[1] = ( rV0dotV0*rE2xV01 - 2.0*rE2xV0*rV0dotV01 );
  
  for ( i = 0; i < 2; i++ )
    {
      if ( prVertexZone[i] < 0.0 )
	{
	  *pdC_x = 0.0;
	  *pdC_y = 0.0;
	  *pdR = 0.0;
	  return APM2D::icctNull;
	}
    }

  //////////////////////////////////////////////////
  // Compute the center and radius

  prC[0] = -rV0dotV0*pprE[2][1]/rDen;
  prC[1] = rV0dotV0*pprE[2][0]/rDen;
  //prC[2] = sqrt((prC[0]*prC[0]) + (prC[1]*prC[1]));

  rR = (rV0dotV0*sqrt(rE2dotE2)) / rDen;
  
  /*
  if ( !( fabs(prC[2] - rR) < s_crEpsilonInscribedCircleCenterVvERadius ) )
    {
      fprintf(strerr, "VvE: C = ( %f, %f, %f ) != %f = R\n", prC[0], prC[1], prC[2], rR);
      //ASSERT( FALSE );
    }
  */

  // Translate the answer back out away from the origin
  prC[0] += rV1_E20_x;
  prC[1] += rV1_E20_y;

  *pdC_x = dble(prC[0]);
  *pdC_y = dble(prC[1]);
  *pdR = dble(rR);
  return APM2D::icctVvE;
}

//static 
APM2D::EdgeZoneType EdgeZoneVvEMPFUN(DOUBLE dV0_x, DOUBLE dV0_y,
				     DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
				     DOUBLE dE21_x, DOUBLE dE21_y,
				     DOUBLE dE30_x, DOUBLE dE30_y,
				     DOUBLE dE31_x, DOUBLE dE31_y)
{
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rV1_E20_x = dV1_E20_x;
  mp_real rV1_E20_y = dV1_E20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;
  mp_real rE30_x = dE30_x;
  mp_real rE30_y = dE30_y;
  mp_real rE31_x = dE31_x;
  mp_real rE31_y = dE31_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real prEdgeZone[2];
  mp_real rV0dotV0;
  mp_real rE3dotE30, rE3dotE31;
  mp_real rE2xV0, rE2xE3;
  mp_real rE3dotE3;
  mp_real rDen;
  APM2D::EdgeZoneType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = rV0_x - rV1_E20_x;
  ppprV[0][0][1] = rV0_y - rV1_E20_y;
  ppprV[1][0][0] = 0.0;
  ppprV[1][0][1] = 0.0;
  ppprV[2][0][0] = 0.0;
  ppprV[2][0][1] = 0.0;
  ppprV[2][1][0] = rE21_x - rV1_E20_x;
  ppprV[2][1][1] = rE21_y - rV1_E20_y;
  ppprV[3][0][0] = rE30_x - rV1_E20_x;
  ppprV[3][0][1] = rE30_y - rV1_E20_y;
  ppprV[3][1][0] = rE31_x - rV1_E20_x;
  ppprV[3][1][1] = rE31_y - rV1_E20_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[2][0] = rE21_x - rV1_E20_x;
  pprE[2][1] = rE21_y - rV1_E20_y;
  pprE[3][0] = rE31_x - rE30_x;
  pprE[3][1] = rE31_y - rE30_y;
  
  rV0dotV0 = (ppprV[0][0][0]*ppprV[0][0][0]) + (ppprV[0][0][1]*ppprV[0][0][1]);
  rE3dotE3 = (pprE[3][0]*pprE[3][0]) + (pprE[3][1]*pprE[3][1]);

  rE3dotE30 = (pprE[3][0]*ppprV[3][0][0]) + (pprE[3][1]*ppprV[3][0][1]);
  rE3dotE31 = (pprE[3][0]*ppprV[3][1][0]) + (pprE[3][1]*ppprV[3][1][1]);
  
  rE2xV0 = (pprE[2][0]*ppprV[0][0][1]) - (pprE[2][1]*ppprV[0][0][0]);
  rE2xE3 = (pprE[2][0]*pprE[3][1]) - (pprE[2][1]*pprE[3][0]);

  rDen = 2.0*rE2xV0; 
  ASSERT( rDen != 0.0 );
  /*
  prC[0] = -rV0dotV0*pprE[2][1]/rDen;
  prC[1] = rV0dotV0*pprE[2][0]/rDen;
  */
  mp_real rV0dotV0timesE2xE3 = rV0dotV0*rE2xE3;

  /*
  rMagCE30 = ( (prC[0] - ppprV[3][0][0])*(prC[0] - ppprV[3][0][0]) + 
	       (prC[1] - ppprV[3][0][1])*(prC[1] - ppprV[3][0][1]) );
  rMagCE31 = ( (prC[0] - ppprV[3][1][0])*(prC[0] - ppprV[3][1][0]) + 
	       (prC[1] - ppprV[3][1][1])*(prC[1] - ppprV[3][1][1]) );

  if ( rMagCE30 <= 0.0 )
    {
      prEdgeZone[0] = 0.0;
    }
  else
    {
      prEdgeZone[0] = ( ( 2.0*rE2xV0*rE3dotE30 - rV0dotV0timesE2xE3 ) / 
			( rDen*sqrt(rE3dotE3*rMagCE30) ) );
    }
  if ( rMagCE31 <= 0.0 )
    {
      prEdgeZone[1] = 0.0;
    }
  else
    {
      prEdgeZone[1] = -( ( 2.0*rE2xV0*rE3dotE31 - rV0dotV0timesE2xE3 ) /
			 ( rDen*sqrt(rE3dotE3*rMagCE31) ) );
    }
  */
  prEdgeZone[0] = ( 2.0*rE2xV0*rE3dotE30 - rV0dotV0timesE2xE3 );
  prEdgeZone[1] = -( 2.0*rE2xV0*rE3dotE31 - rV0dotV0timesE2xE3 );
  
  if ( !( prEdgeZone[0] < 0.0 ) )
    {
      eResult = APM2D::eztStart;
    }
  else if ( !( prEdgeZone[1] < 0.0 ) )
    {
      eResult = APM2D::eztEnd;
    }
  else
    {
      eResult = APM2D::eztMiddle;
    }

  return eResult;
}

//static 
APM2D::CircleTestType InCircleVvE_VMPFUN(DOUBLE dV0_x, DOUBLE dV0_y,
					 DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
					 DOUBLE dE21_x, DOUBLE dE21_y,
					 DOUBLE dV3_x, DOUBLE dV3_y)
{
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rV1_E20_x = dV1_E20_x;
  mp_real rV1_E20_y = dV1_E20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;
  mp_real rV3_x = dV3_x;
  mp_real rV3_y = dV3_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real rV0dotV0, rV3dotV3;
  mp_real rE2xV0, rE2xV3;
  //mp_real rDen;
  mp_real rResult;
  APM2D::CircleTestType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = rV0_x - rV1_E20_x;
  ppprV[0][0][1] = rV0_y - rV1_E20_y;
  ppprV[1][0][0] = 0.0;
  ppprV[1][0][1] = 0.0;
  ppprV[2][0][0] = 0.0;
  ppprV[2][0][1] = 0.0;
  ppprV[2][1][0] = rE21_x - rV1_E20_x;
  ppprV[2][1][1] = rE21_y - rV1_E20_y;
  ppprV[3][0][0] = rV3_x - rV1_E20_x;
  ppprV[3][0][1] = rV3_y - rV1_E20_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[2][0] = rE21_x - rV1_E20_x;
  pprE[2][1] = rE21_y - rV1_E20_y;
  
  rV0dotV0 = (ppprV[0][0][0]*ppprV[0][0][0]) + (ppprV[0][0][1]*ppprV[0][0][1]);
  rV3dotV3 = (ppprV[3][0][0]*ppprV[3][0][0]) + (ppprV[3][0][1]*ppprV[3][0][1]);

  rE2xV0 = (pprE[2][0]*ppprV[0][0][1]) - (pprE[2][1]*ppprV[0][0][0]);
  rE2xV3 = (pprE[2][0]*ppprV[3][0][1]) - (pprE[2][1]*ppprV[3][0][0]);

  //rDen = rE2xV0; 
  //ASSERT( rDen != 0.0 );

  rResult = ( rV3dotV3*rE2xV0 - rV0dotV0*rE2xV3 );
  
  if ( rResult < 0.0 )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( rResult > 0.0 )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

//static 
APM2D::CircleTestType InCircleVvE_EMPFUN(DOUBLE dV0_x, DOUBLE dV0_y,
					 DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
					 DOUBLE dE21_x, DOUBLE dE21_y,
					 DOUBLE dE30_x, DOUBLE dE30_y,
					 DOUBLE dE31_x, DOUBLE dE31_y)
{
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rV1_E20_x = dV1_E20_x;
  mp_real rV1_E20_y = dV1_E20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;
  mp_real rE30_x = dE30_x;
  mp_real rE30_y = dE30_y;
  mp_real rE31_x = dE31_x;
  mp_real rE31_y = dE31_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real rV0dotV0, rE2dotE2, rE3dotE3;
  mp_real rE2dotE3;
  mp_real rE2xV0, rE30xE31;
  //mp_real rDen;
  mp_real rResult;
  APM2D::CircleTestType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = rV0_x - rV1_E20_x;
  ppprV[0][0][1] = rV0_y - rV1_E20_y;
  ppprV[1][0][0] = 0.0;
  ppprV[1][0][1] = 0.0;
  ppprV[2][0][0] = 0.0;
  ppprV[2][0][1] = 0.0;
  ppprV[2][1][0] = rE21_x - rV1_E20_x;
  ppprV[2][1][1] = rE21_y - rV1_E20_y;
  ppprV[3][0][0] = rE30_x - rV1_E20_x;
  ppprV[3][0][1] = rE30_y - rV1_E20_y;
  ppprV[3][1][0] = rE31_x - rV1_E20_x;
  ppprV[3][1][1] = rE31_y - rV1_E20_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[2][0] = rE21_x - rV1_E20_x;
  pprE[2][1] = rE21_y - rV1_E20_y;
  pprE[3][0] = rE31_x - rE30_x;
  pprE[3][1] = rE31_y - rE30_y;
  
  rV0dotV0 = (ppprV[0][0][0]*ppprV[0][0][0]) + (ppprV[0][0][1]*ppprV[0][0][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);
  rE3dotE3 = (pprE[3][0]*pprE[3][0]) + (pprE[3][1]*pprE[3][1]);

  rE2dotE3 = (pprE[2][0]*pprE[3][0]) + (pprE[2][1]*pprE[3][1]);

  rE2xV0 = (pprE[2][0]*ppprV[0][0][1]) - (pprE[2][1]*ppprV[0][0][0]);
  rE30xE31 = (ppprV[3][0][0]*ppprV[3][1][1]) - (ppprV[3][0][1]*ppprV[3][1][0]);

  //rDen = 4.0*rE3dotE3*rE2xV0*rE2xV0; 
  //ASSERT( rDen != 0.0 );

  rResult = ( 4.0*rE2xV0*rE30xE31*( rE2xV0*rE30xE31 + rV0dotV0*rE2dotE3 )
	      - rV0dotV0*rV0dotV0*( rE2dotE2*rE3dotE3 - rE2dotE3*rE2dotE3 ) );

  if ( rResult < 0.0 )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( rResult > 0.0 )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

//static 
APM2D::InscribedCircleCenterType 
InscribedCircleCenterEvVMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			      DOUBLE dE00_x, DOUBLE dE00_y,
			      DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
			      DOUBLE dV20_x, DOUBLE dV20_y,
			      DOUBLE dV2_x, DOUBLE dV2_y,
			      DOUBLE dV21_x, DOUBLE dV21_y)
{
  mp_real prC[3], rR;
  mp_real rE00_x = dE00_x;
  mp_real rE00_y = dE00_y;
  mp_real rE01_V1_x = dE01_V1_x;
  mp_real rE01_V1_y = dE01_V1_y; 
  mp_real rV20_x = dV20_x;
  mp_real rV20_y = dV20_y;
  mp_real rV2_x = dV2_x;
  mp_real rV2_y = dV2_y;
  mp_real rV21_x = dV21_x;
  mp_real rV21_y = dV21_y;

  mp_real ppprV[3][2][2], ppprVN[3][2][2]; // Endpoints
  mp_real pprE[3][2];     // Edge vectors
  mp_real rDen;
  mp_real rV2dotV2, rE0dotE0;
  mp_real rE0xV2;
  mp_real rV20dotV20, rV21dotV21, rV2dotV20, rV2dotV21;
  mp_real rE0xV20, rE0xV21;
  mp_real prVertexZone[2], rSinTheta;
  UINT_32 i;
  
  // Translate V1 to the origin
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = rE00_x - rE01_V1_x;
  ppprV[0][0][1] = rE00_y - rE01_V1_y;
  ppprV[0][1][0] = 0.0;
  ppprV[0][1][1] = 0.0;
  ppprV[1][0][0] = 0.0;
  ppprV[1][0][1] = 0.0;
  ppprV[2][0][0] = rV2_x - rE01_V1_x;
  ppprV[2][0][1] = rV2_y - rE01_V1_y;
  
  // Vertex neighbor vectors
  //memset(ppprVN, 0, sizeof(ppprV));
  ppprVN[2][0][0] = rV2_x - rV20_x;
  ppprVN[2][0][1] = rV2_y - rV20_y;
  ppprVN[2][1][0] = rV2_x - rV21_x;
  ppprVN[2][1][1] = rV2_y - rV21_y;

  //memset(pprE, 0, sizeof(pprE));
  pprE[0][0] = rE01_V1_x - rE00_x;
  pprE[0][1] = rE01_V1_y - rE00_y;
  
  rV2dotV2 = (ppprV[2][0][0]*ppprV[2][0][0]) + (ppprV[2][0][1]*ppprV[2][0][1]);
  rE0dotE0 = (pprE[0][0]*pprE[0][0]) + (pprE[0][1]*pprE[0][1]);
  rE0xV2 = (pprE[0][0]*ppprV[2][0][1]) - (pprE[0][1]*ppprV[2][0][0]);

  rV20dotV20 = ( (ppprVN[2][0][0] * ppprVN[2][0][0]) + (ppprVN[2][0][1] * ppprVN[2][0][1]) );
  rV21dotV21 = ( (ppprVN[2][1][0] * ppprVN[2][1][0]) + (ppprVN[2][1][1] * ppprVN[2][1][1]) );

  rV2dotV20 = ( (ppprV[2][0][0] * ppprVN[2][0][0]) + (ppprV[2][0][1] * ppprVN[2][0][1]) );
  rV2dotV21 = ( (ppprV[2][0][0] * ppprVN[2][1][0]) + (ppprV[2][0][1] * ppprVN[2][1][1]) );

  rE0xV20 = ( (pprE[0][0] * ppprVN[2][0][1]) - (pprE[0][1] * ppprVN[2][0][0]) );
  rE0xV21 = ( (pprE[0][0] * ppprVN[2][1][1]) - (pprE[0][1] * ppprVN[2][1][0]) );

  //rSinTheta = rE0xV2/sqrt(rE0dotE0*rV2dotV2); 
  rSinTheta = SinThetaMPFUN(rV2_x, rV2_y, rE00_x, rE00_y, rE01_V1_x, rE01_V1_y);
  /*
  rSinThetaTemp = SinThetaMPFUN(rE00_x, rE00_y, rE01_V1_x, rE01_V1_y, rV2_x, rV2_y);
  if ( rSinThetaTemp > rSinTheta )
    {
      rSinTheta = rSinThetaTemp;
    }
  */
  if ( rSinTheta <= 0.0 )
    {
      // V0 is on E2 and it is not one of the endpoints
      
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  rDen = 2.0*rE0xV2;
  ASSERT( rDen != 0.0 );

  //////////////////////////////////////////////////
  // Check the Vertex Zones

  prVertexZone[0] = ( rV2dotV2*rE0xV20 - 2.0*rE0xV2*rV2dotV20 );
  prVertexZone[1] = ( rV2dotV2*rE0xV21 - 2.0*rE0xV2*rV2dotV21 );
  
  for ( i = 0; i < 2; i++ )
    {
      if ( prVertexZone[i] < 0.0 )
	{
	  *pdC_x = 0.0;
	  *pdC_y = 0.0;
	  *pdR = 0.0;
	  return APM2D::icctNull;
	}
    }

  //////////////////////////////////////////////////
  // Compute the center and radius

  prC[0] = -rV2dotV2*pprE[0][1]/rDen;
  prC[1] = rV2dotV2*pprE[0][0]/rDen;
  //prC[2] = sqrt((prC[0]*prC[0]) + (prC[1]*prC[1]));

  rR = (rV2dotV2*sqrt(rE0dotE0)) / rDen;
  
  /*
  if ( !( fabs(prC[2] - rR) < s_crEpsilonInscribedCircleCenterEvVRadius ) )
    {
      fprintf(strerr, "EvV: C = ( %f, %f, %f ) != %f = R\n", prC[0], prC[1], prC[2], rR);
      //ASSERT( FALSE );
    }
  */

  // Translate the answer back out away from the origin
  prC[0] += rE01_V1_x;
  prC[1] += rE01_V1_y;

  *pdC_x = dble(prC[0]);
  *pdC_y = dble(prC[1]);
  *pdR = dble(rR);
  return APM2D::icctEvV;
}

//static 
APM2D::EdgeZoneType EdgeZoneEvVMPFUN(DOUBLE dE00_x, DOUBLE dE00_y,
				     DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
				     DOUBLE dV2_x, DOUBLE dV2_y,
				     DOUBLE dE30_x, DOUBLE dE30_y,
				     DOUBLE dE31_x, DOUBLE dE31_y)
{
  mp_real rE00_x = dE00_x;
  mp_real rE00_y = dE00_y;
  mp_real rE01_V1_x = dE01_V1_x;
  mp_real rE01_V1_y = dE01_V1_y; 
  mp_real rV2_x = dV2_x;
  mp_real rV2_y = dV2_y;
  mp_real rE30_x = dE30_x;
  mp_real rE30_y = dE30_y;
  mp_real rE31_x = dE31_x;
  mp_real rE31_y = dE31_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real prEdgeZone[2];
  mp_real rV2dotV2;
  mp_real rE3dotE30, rE3dotE31;
  mp_real rE0xV2, rE0xE3;
  mp_real rE3dotE3;
  mp_real rDen;
  APM2D::EdgeZoneType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = rE00_x - rE01_V1_x;
  ppprV[0][0][1] = rE00_y - rE01_V1_y;
  ppprV[0][1][0] = 0.0;
  ppprV[0][1][1] = 0.0;
  ppprV[1][0][0] = 0.0;
  ppprV[1][0][1] = 0.0;
  ppprV[2][0][0] = rV2_x - rE01_V1_x;
  ppprV[2][0][1] = rV2_y - rE01_V1_y;
  ppprV[3][0][0] = rE30_x - rE01_V1_x;
  ppprV[3][0][1] = rE30_y - rE01_V1_y;
  ppprV[3][1][0] = rE31_x - rE01_V1_x;
  ppprV[3][1][1] = rE31_y - rE01_V1_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[0][0] = rE01_V1_x - rE00_x;
  pprE[0][1] = rE01_V1_y - rE00_y;
  pprE[3][0] = rE31_x - rE30_x;
  pprE[3][1] = rE31_y - rE30_y;
  
  rV2dotV2 = (ppprV[2][0][0]*ppprV[2][0][0]) + (ppprV[2][0][1]*ppprV[2][0][1]);
  rE3dotE3 = (pprE[3][0]*pprE[3][0]) + (pprE[3][1]*pprE[3][1]);

  rE3dotE30 = (pprE[3][0]*ppprV[3][0][0]) + (pprE[3][1]*ppprV[3][0][1]);
  rE3dotE31 = (pprE[3][0]*ppprV[3][1][0]) + (pprE[3][1]*ppprV[3][1][1]);
  
  rE0xV2 = (pprE[0][0]*ppprV[2][0][1]) - (pprE[0][1]*ppprV[2][0][0]);
  rE0xE3 = (pprE[0][0]*pprE[3][1]) - (pprE[0][1]*pprE[3][0]);

  rDen = 2.0*rE0xV2; 
  ASSERT( rDen != 0.0 );
  /*
  prC[0] = -rV2dotV2*pprE[0][1]/rDen;
  prC[1] = rV2dotV2*pprE[0][0]/rDen;
  */
  mp_real rV2dotV2timesE0xE3 = rV2dotV2*rE0xE3;
  /*
  rMagCE30 = ( (prC[0] - ppprV[3][0][0])*(prC[0] - ppprV[3][0][0]) + 
	       (prC[1] - ppprV[3][0][1])*(prC[1] - ppprV[3][0][1]) );
  rMagCE31 = ( (prC[0] - ppprV[3][1][0])*(prC[0] - ppprV[3][1][0]) + 
	       (prC[1] - ppprV[3][1][1])*(prC[1] - ppprV[3][1][1]) );

  if ( rMagCE30 <= 0.0 )
    {
      prEdgeZone[0] = 0.0;
    }
  else
    {
      prEdgeZone[0] = ( ( 2.0*rE0xV2*rE3dotE30 - rV2dotV2timesE0xE3 ) / 
			( rDen*sqrt(rE3dotE3*rMagCE30) ) );
    }
  if ( rMagCE31 <= 0.0 )
    {
      prEdgeZone[1] = 0.0;
    }
  else
    {
      prEdgeZone[1] = -( ( 2.0*rE0xV2*rE3dotE31 - rV2dotV2timesE0xE3 ) / 
			 ( rDen*sqrt(rE3dotE3*rMagCE31) ) );
    }
  */
  prEdgeZone[0] = ( 2.0*rE0xV2*rE3dotE30 - rV2dotV2timesE0xE3 );
  prEdgeZone[1] = -( 2.0*rE0xV2*rE3dotE31 - rV2dotV2timesE0xE3 );

  if ( !( prEdgeZone[0] < 0.0 ) )
    {
      eResult = APM2D::eztStart;
    }
  else if ( !( prEdgeZone[1] < 0.0 ) )
    {
      eResult = APM2D::eztEnd;
    }
  else
    {
      eResult = APM2D::eztMiddle;
    }

  return eResult;
}

//static 
APM2D::CircleTestType InCircleEvV_VMPFUN(DOUBLE dE00_x, DOUBLE dE00_y,
					 DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
					 DOUBLE dV2_x, DOUBLE dV2_y,
					 DOUBLE dV3_x, DOUBLE dV3_y)
{
  mp_real rE00_x = dE00_x;
  mp_real rE00_y = dE00_y;
  mp_real rE01_V1_x = dE01_V1_x;
  mp_real rE01_V1_y = dE01_V1_y; 
  mp_real rV2_x = dV2_x;
  mp_real rV2_y = dV2_y;
  mp_real rV3_x = dV3_x;
  mp_real rV3_y = dV3_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real rV2dotV2, rV3dotV3;
  mp_real rE0xV2, rE0xV3;
  //mp_real rDen;
  mp_real rResult;
  APM2D::CircleTestType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = rE00_x - rE01_V1_x;
  ppprV[0][0][1] = rE00_y - rE01_V1_y;
  ppprV[0][1][0] = 0.0;
  ppprV[0][1][1] = 0.0;
  ppprV[1][0][0] = 0.0;
  ppprV[1][0][1] = 0.0;
  ppprV[2][0][0] = rV2_x - rE01_V1_x;
  ppprV[2][0][1] = rV2_y - rE01_V1_y;
  ppprV[3][0][0] = rV3_x - rE01_V1_x;
  ppprV[3][0][1] = rV3_y - rE01_V1_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[0][0] = rE01_V1_x - rE00_x;
  pprE[0][1] = rE01_V1_y - rE00_y;
  
  rV2dotV2 = (ppprV[2][0][0]*ppprV[2][0][0]) + (ppprV[2][0][1]*ppprV[2][0][1]);
  rV3dotV3 = (ppprV[3][0][0]*ppprV[3][0][0]) + (ppprV[3][0][1]*ppprV[3][0][1]);

  rE0xV2 = (pprE[0][0]*ppprV[2][0][1]) - (pprE[0][1]*ppprV[2][0][0]);
  rE0xV3 = (pprE[0][0]*ppprV[3][0][1]) - (pprE[0][1]*ppprV[3][0][0]);

  //rDen = rE0xV2;
  //ASSERT( rDen != 0.0 );

  rResult = ( rV3dotV3*rE0xV2 - rV2dotV2*rE0xV3 );
  
  if ( rResult < 0.0 )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( rResult > 0.0 )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

//static 
APM2D::CircleTestType InCircleEvV_EMPFUN(DOUBLE dE00_x, DOUBLE dE00_y,
					 DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
					 DOUBLE dV2_x, DOUBLE dV2_y,
					 DOUBLE dE30_x, DOUBLE dE30_y,
					 DOUBLE dE31_x, DOUBLE dE31_y)
{
  mp_real rE00_x = dE00_x;
  mp_real rE00_y = dE00_y;
  mp_real rE01_V1_x = dE01_V1_x;
  mp_real rE01_V1_y = dE01_V1_y; 
  mp_real rV2_x = dV2_x;
  mp_real rV2_y = dV2_y;
  mp_real rE30_x = dE30_x;
  mp_real rE30_y = dE30_y;
  mp_real rE31_x = dE31_x;
  mp_real rE31_y = dE31_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real rV2dotV2, rE0dotE0, rE3dotE3;
  mp_real rE0dotE3;
  mp_real rE0xV2, rE30xE31;
  //mp_real rDen;
  mp_real rResult;
  APM2D::CircleTestType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = rE00_x - rE01_V1_x;
  ppprV[0][0][1] = rE00_y - rE01_V1_y;
  ppprV[0][1][0] = 0.0;
  ppprV[0][1][1] = 0.0;
  ppprV[1][0][0] = 0.0;
  ppprV[1][0][1] = 0.0;
  ppprV[2][0][0] = rV2_x - rE01_V1_x;
  ppprV[2][0][1] = rV2_y - rE01_V1_y;
  ppprV[3][0][0] = rE30_x - rE01_V1_x;
  ppprV[3][0][1] = rE30_y - rE01_V1_y;
  ppprV[3][1][0] = rE31_x - rE01_V1_x;
  ppprV[3][1][1] = rE31_y - rE01_V1_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[0][0] = rE01_V1_x - rE00_x;
  pprE[0][1] = rE01_V1_y - rE00_y;
  pprE[3][0] = rE31_x - rE30_x;
  pprE[3][1] = rE31_y - rE30_y;
  
  rV2dotV2 = (ppprV[2][0][0]*ppprV[2][0][0]) + (ppprV[2][0][1]*ppprV[2][0][1]);
  rE0dotE0 = (pprE[0][0]*pprE[0][0]) + (pprE[0][1]*pprE[0][1]);
  rE3dotE3 = (pprE[3][0]*pprE[3][0]) + (pprE[3][1]*pprE[3][1]);

  rE0dotE3 = (pprE[0][0]*pprE[3][0]) + (pprE[0][1]*pprE[3][1]);

  rE0xV2 = (pprE[0][0]*ppprV[2][0][1]) - (pprE[0][1]*ppprV[2][0][0]);
  rE30xE31 = (ppprV[3][0][0]*ppprV[3][1][1]) - (ppprV[3][0][1]*ppprV[3][1][0]);

  //rDen = 4.0*rE3dotE3*rE0xV2*rE0xV2;
  //ASSERT( rDen != 0.0 );

  rResult = ( 4.0*rE0xV2*rE30xE31*( rE0xV2*rE30xE31 + rV2dotV2*rE0dotE3 )
	      - rV2dotV2*rV2dotV2*( rE0dotE0*rE3dotE3 - rE0dotE3*rE0dotE3 ) );

  if ( rResult < 0.0 )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( rResult > 0.0 )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

//static 
APM2D::InscribedCircleCenterType 
InscribedCircleCenterVEEMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			      DOUBLE dV00_x, DOUBLE dV00_y,
			      DOUBLE dV0_x, DOUBLE dV0_y,
			      DOUBLE dV01_x, DOUBLE dV01_y,
			      DOUBLE dE10_x, DOUBLE dE10_y, 
			      DOUBLE dE11_x, DOUBLE dE11_y,
			      DOUBLE dE20_x, DOUBLE dE20_y, 
			      DOUBLE dE21_x, DOUBLE dE21_y)
{
  mp_real prC[3], rR;
  mp_real rV00_x = dV00_x;
  mp_real rV00_y = dV00_y;
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rV01_x = dV01_x;
  mp_real rV01_y = dV01_y;
  mp_real rE10_x = dE10_x;
  mp_real rE10_y = dE10_y; 
  mp_real rE11_x = dE11_x;
  mp_real rE11_y = dE11_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;

  mp_real ppprV[3][2][2], ppprVN[1][2][2]; // Endpoints
  mp_real pprE[3][2];     // Edge vectors
  mp_real prVertexZone[2];
  mp_real pprEdgeZone[2][2];
  mp_real rDen, rDenR, rDisc;
  mp_real rE1dotE1, rE2dotE2, rE1dotE2;
  mp_real rE1dotE10, rE1dotE11, rE2dotE20, rE2dotE21;
  mp_real rE1xE2, rE10xE11, rE20xE21;
  mp_real rE1E2plusE1dotE2, rE1E2minusE1dotE2;
  mp_real rPlusOrMinus, rSinTheta;
  mp_real rV00dotV00, rV01dotV01, rE1dotV00, rE1dotV01, rE2dotV00, rE2dotV01;
  mp_real rE10dotE10, rE11dotE11, rE20dotE20, rE21dotE21;
  UINT_32 i, j;
  
  // Translate V1 to the origin
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = rE10_x - rV0_x;
  ppprV[1][0][1] = rE10_y - rV0_y;
  ppprV[1][1][0] = rE11_x - rV0_x;
  ppprV[1][1][1] = rE11_y - rV0_y;
  ppprV[2][0][0] = rE20_x - rV0_x;
  ppprV[2][0][1] = rE20_y - rV0_y;
  ppprV[2][1][0] = rE21_x - rV0_x;
  ppprV[2][1][1] = rE21_y - rV0_y;
  
  // Vertex neighbor vectors
  //memset(ppprVN, 0, sizeof(ppprV));
  ppprVN[0][0][0] = rV0_x - rV00_x;
  ppprVN[0][0][1] = rV0_y - rV00_y;
  ppprVN[0][1][0] = rV0_x - rV01_x;
  ppprVN[0][1][1] = rV0_y - rV01_y;

  //memset(pprE, 0, sizeof(pprE));
  pprE[1][0] = rE11_x - rE10_x;
  pprE[1][1] = rE11_y - rE10_y;
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;
  
  rE1dotE1 = (pprE[1][0]*pprE[1][0]) + (pprE[1][1]*pprE[1][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);
  rE1dotE2 = (pprE[1][0]*pprE[2][0]) + (pprE[1][1]*pprE[2][1]);
  rE1dotE10 = (pprE[1][0]*ppprV[1][0][0]) + (pprE[1][1]*ppprV[1][0][1]);
  rE1dotE11 = (pprE[1][0]*ppprV[1][1][0]) + (pprE[1][1]*ppprV[1][1][1]);
  rE2dotE20 = (pprE[2][0]*ppprV[2][0][0]) + (pprE[2][1]*ppprV[2][0][1]);
  rE2dotE21 = (pprE[2][0]*ppprV[2][1][0]) + (pprE[2][1]*ppprV[2][1][1]);

  rE1xE2 = (pprE[1][0]*pprE[2][1]) - (pprE[1][1]*pprE[2][0]);
  rE10xE11 = (ppprV[1][0][0]*ppprV[1][1][1]) - (ppprV[1][0][1]*ppprV[1][1][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);

  rE1E2plusE1dotE2 = sqrt(rE1dotE1*rE2dotE2) + rE1dotE2;
  rE1E2minusE1dotE2 = sqrt(rE1dotE1*rE2dotE2) - rE1dotE2;

  rV00dotV00 = ( (ppprVN[0][0][0] * ppprVN[0][0][0]) + (ppprVN[0][0][1] * ppprVN[0][0][1]) );
  rV01dotV01 = ( (ppprVN[0][1][0] * ppprVN[0][1][0]) + (ppprVN[0][1][1] * ppprVN[0][1][1]) );

  rE10dotE10 = (ppprV[1][0][0]*ppprV[1][0][0]) + (ppprV[1][0][1]*ppprV[1][0][1]);
  rE11dotE11 = (ppprV[1][1][0]*ppprV[1][1][0]) + (ppprV[1][1][1]*ppprV[1][1][1]);
  rE20dotE20 = (ppprV[2][0][0]*ppprV[2][0][0]) + (ppprV[2][0][1]*ppprV[2][0][1]);
  rE21dotE21 = (ppprV[2][1][0]*ppprV[2][1][0]) + (ppprV[2][1][1]*ppprV[2][1][1]);

  rE1dotV00 = ( (pprE[1][0] * ppprVN[0][0][0]) + (pprE[1][1] * ppprVN[0][0][1]) );
  rE1dotV01 = ( (pprE[1][0] * ppprVN[0][1][0]) + (pprE[1][1] * ppprVN[0][1][1]) );
  rE2dotV00 = ( (pprE[2][0] * ppprVN[0][0][0]) + (pprE[2][1] * ppprVN[0][0][1]) );
  rE2dotV01 = ( (pprE[2][0] * ppprVN[0][1][0]) + (pprE[2][1] * ppprVN[0][1][1]) );

  rSinTheta = rE1xE2;
  if ( rSinTheta == 0.0 )
    {
      if ( rE1dotE2 >= 0.0 )
	{
	  // Colinear
	  *pdC_x = 0.0;
	  *pdC_y = 0.0;
	  *pdR = 0.0;
	  return APM2D::icctNull;
	}
      else
	{
	  return InscribedCircleCenterVEEAntiparallelMPFUN(pdC_x, pdC_y, pdR, 
							   dV00_x, dV00_y, 
							   dV0_x, dV0_y, 
							   dV01_x, dV01_y, 
							   dE10_x, dE10_y, 
							   dE11_x, dE11_y, 
							   dE20_x, dE20_y, 
							   dE21_x, dE21_y);
	}
    }

  //rSinTheta = rE10xE11/sqrt(rE10dotE10*rE11dotE11);
  rSinTheta = SinThetaMPFUN(rV0_x, rV0_y, rE10_x, rE10_y, rE11_x, rE11_y);
  /*
  rSinThetaTemp = SinThetaMPFUN(rE10_x, rE10_y, rE11_x, rE11_y, rV0_x, rV0_y);
  if ( rSinThetaTemp > rSinTheta )
    {
      rSinTheta = rSinThetaTemp;
    }
  */
  if ( rSinTheta <= 0.0 )
    {
      // V0 is on E1 and it is not one of the endpoints
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  //rSinTheta = rE20xE21/sqrt(rE20dotE20*rE21dotE21);
  rSinTheta = SinThetaMPFUN(rV0_x, rV0_y, rE20_x, rE20_y, rE21_x, rE21_y);
  /*
  rSinThetaTemp = SinThetaMPFUN(rE20_x, rE20_y, rE21_x, rE21_y, rV0_x, rV0_y);
  if ( rSinThetaTemp > rSinTheta )
    {
      rSinTheta = rSinThetaTemp;
    }
  */
  if ( rSinTheta <= 0.0 )
    {
      // V0 is on E2 and it is not one of the endpoints
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  rDen = rE1xE2*rE1E2minusE1dotE2;
  ASSERT( rDen != 0.0 );
  rDenR = rE1E2minusE1dotE2;
  ASSERT( rDenR != 0.0 );

  rDisc = 2.0*rE10xE11*rE20xE21*rE1E2plusE1dotE2;
  if ( rDisc < 0.0 ) 
    {
      // Imaginary Solutions
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  mp_real rSqrtDisc = sqrt(rDisc);
  
  rPlusOrMinus = 1.0;
  if ( rE1xE2 > 0.0 )
    {
      rPlusOrMinus = -1.0;
    }

  //////////////////////////////////////////////////
  // Check the Edge Zone

  mp_real rCoef1 = ( rE10xE11*rE1E2plusE1dotE2 
		     + rPlusOrMinus*sqrt(rE1dotE1*rDisc) );
  mp_real rCoef2 = ( - rE20xE21*rE1E2plusE1dotE2 
		     - rPlusOrMinus*sqrt(rE2dotE2*rDisc) );
  /*
  rMagCE10 = ( (prC[0] - ppprV[1][0][0])*(prC[0] - ppprV[1][0][0]) + 
	       (prC[1] - ppprV[1][0][1])*(prC[1] - ppprV[1][0][1]) );
  rMagCE11 = ( (prC[0] - ppprV[1][1][0])*(prC[0] - ppprV[1][1][0]) + 
	       (prC[1] - ppprV[1][1][1])*(prC[1] - ppprV[1][1][1]) );
  rMagCE20 = ( (prC[0] - ppprV[2][0][0])*(prC[0] - ppprV[2][0][0]) + 
	       (prC[1] - ppprV[2][0][1])*(prC[1] - ppprV[2][0][1]) );
  rMagCE21 = ( (prC[0] - ppprV[2][1][0])*(prC[0] - ppprV[2][1][0]) + 
	       (prC[1] - ppprV[2][1][1])*(prC[1] - ppprV[2][1][1]) );

  if ( rMagCE10 <= 0.0 )
    {
      pprEdgeZone[0][0] = 0.0;
    }
  else
    {
      pprEdgeZone[0][0] = -( ( rE1xE2*rE1dotE10 
			       + rCoef1 ) / 
			     ( rE1xE2*sqrt(rE1dotE1*rMagCE10) ) );
    }
  if ( rMagCE11 <= 0.0 )
    {
      pprEdgeZone[0][1] = 0.0;
    }
  else
    {
      pprEdgeZone[0][1] = ( ( rE1xE2*rE1dotE11 
			      + rCoef1 ) / 
			    ( rE1xE2*sqrt(rE1dotE1*rMagCE11) ) );
    }
  if ( rMagCE20 <= 0.0 )
    {
      pprEdgeZone[1][0] = 0.0;
    }
  else
    {
      pprEdgeZone[1][0] = -( ( rE1xE2*rE2dotE20 
			       + rCoef2 ) / 
			     ( rE1xE2*sqrt(rE2dotE2*rMagCE20) ) );
    }
  if ( rMagCE21 <= 0.0 )
    {
      pprEdgeZone[1][1] = 0.0;
    }
  else
    {
      pprEdgeZone[1][1] = ( ( rE1xE2*rE2dotE21 
			      + rCoef2 ) / 
			    ( rE1xE2*sqrt(rE2dotE2*rMagCE21) ) );
    }
  */

  if ( rE1xE2 >= 0.0 )
    {
      pprEdgeZone[0][0] = -( rE1xE2*rE1dotE10 
			     + rCoef1 );
      pprEdgeZone[0][1] = ( rE1xE2*rE1dotE11 
			    + rCoef1 );
      pprEdgeZone[1][0] = -( rE1xE2*rE2dotE20 
			     + rCoef2 );
      pprEdgeZone[1][1] = ( rE1xE2*rE2dotE21 
			    + rCoef2 );
    }
  else
    {
      pprEdgeZone[0][0] = ( rE1xE2*rE1dotE10 
			    + rCoef1 );
      pprEdgeZone[0][1] = -( rE1xE2*rE1dotE11 
			     + rCoef1 );
      pprEdgeZone[1][0] = ( rE1xE2*rE2dotE20 
			    + rCoef2 );
      pprEdgeZone[1][1] = -( rE1xE2*rE2dotE21 
			     + rCoef2 );
    }

  for ( i = 0; i < 2; i++ )
    {
      for ( j = 0; j < 2; j++ )
	{
	  if ( pprEdgeZone[i][j] < 0.0 )
	    {
	      *pdC_x = 0.0;
	      *pdC_y = 0.0;
	      *pdR = 0.0;
	      return APM2D::icctNull;
	    }
	}
    }

  //////////////////////////////////////////////////
  // Check the Vertex Zone

  if ( rE1xE2 >= 0.0 )
    {
      prVertexZone[0] = ( ( rE1dotE1*rE2dotV00 - rE1dotE2*rE1dotV00 )*rE20xE21 
			  + ( rE1dotE2*rE2dotV00 - rE2dotE2*rE1dotV00 )*rE10xE11 
			  + rPlusOrMinus*( sqrt(rE1dotE1)*rE2dotV00 - sqrt(rE2dotE2)*rE1dotV00 )*rSqrtDisc );
      prVertexZone[1] = ( ( rE1dotE1*rE2dotV01 - rE1dotE2*rE1dotV01 )*rE20xE21 
			  + ( rE1dotE2*rE2dotV01 - rE2dotE2*rE1dotV01 )*rE10xE11 
			  + rPlusOrMinus*( sqrt(rE1dotE1)*rE2dotV01 - sqrt(rE2dotE2)*rE1dotV01 )*rSqrtDisc );
    }
  else
    {
      prVertexZone[0] = -( ( rE1dotE1*rE2dotV00 - rE1dotE2*rE1dotV00 )*rE20xE21 
			   + ( rE1dotE2*rE2dotV00 - rE2dotE2*rE1dotV00 )*rE10xE11 
			   + rPlusOrMinus*( sqrt(rE1dotE1)*rE2dotV00 - sqrt(rE2dotE2)*rE1dotV00 )*rSqrtDisc );
      prVertexZone[1] = -( ( rE1dotE1*rE2dotV01 - rE1dotE2*rE1dotV01 )*rE20xE21 
			   + ( rE1dotE2*rE2dotV01 - rE2dotE2*rE1dotV01 )*rE10xE11 
			   + rPlusOrMinus*( sqrt(rE1dotE1)*rE2dotV01 - sqrt(rE2dotE2)*rE1dotV01 )*rSqrtDisc );
    }

  for ( i = 0; i < 2; i++ )
    {
      if ( prVertexZone[i] < 0.0 )
	{
	  *pdC_x = 0.0;
	  *pdC_y = 0.0;
	  *pdR = 0.0;
	  return APM2D::icctNull;
	}
    }

  //////////////////////////////////////////////////
  // Compute the center and radius

  prC[0] = ( (rE1dotE1*pprE[2][0] - rE1dotE2*pprE[1][0])*rE20xE21
	     + (rE1dotE2*pprE[2][0] - rE2dotE2*pprE[1][0])*rE10xE11
	     + rPlusOrMinus*(sqrt(rE1dotE1)*pprE[2][0] - sqrt(rE2dotE2)*pprE[1][0])*rSqrtDisc ) / rDen;
  prC[1] = ( (rE1dotE1*pprE[2][1] - rE1dotE2*pprE[1][1])*rE20xE21
	     + (rE1dotE2*pprE[2][1] - rE2dotE2*pprE[1][1])*rE10xE11
	     + rPlusOrMinus*(sqrt(rE1dotE1)*pprE[2][1] - sqrt(rE2dotE2)*pprE[1][1])*rSqrtDisc ) / rDen;
  //prC[2] = sqrt((prC[0]*prC[0]) + (prC[1]*prC[1]));

  rR = ( sqrt(rE1dotE1)*rE20xE21 + sqrt(rE2dotE2)*rE10xE11 + rPlusOrMinus*rSqrtDisc ) / rDenR;

  /*  
  if ( !( fabs(prC[2] - rR) < s_crEpsilonInscribedCircleCenterVEERadius ) )
    {
      fprintf(strerr, "VEE: C = ( %f, %f, %f ) != %f = R\n", prC[0], prC[1], prC[2], rR);
      //ASSERT( FALSE );
    }
  */

  // Translate the answer back out away from the origin
  prC[0] += rV0_x;
  prC[1] += rV0_y;

  *pdC_x = dble(prC[0]);
  *pdC_y = dble(prC[1]);
  *pdR = dble(rR);
  return APM2D::icctVEE;
}

//static 
APM2D::EdgeZoneType EdgeZoneVEEMPFUN(DOUBLE dV0_x, DOUBLE dV0_y,
				     DOUBLE dE10_x, DOUBLE dE10_y, 
				     DOUBLE dE11_x, DOUBLE dE11_y,
				     DOUBLE dE20_x, DOUBLE dE20_y, 
				     DOUBLE dE21_x, DOUBLE dE21_y,
				     DOUBLE dE30_x, DOUBLE dE30_y,
				     DOUBLE dE31_x, DOUBLE dE31_y)
{
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rE10_x = dE10_x;
  mp_real rE10_y = dE10_y; 
  mp_real rE11_x = dE11_x;
  mp_real rE11_y = dE11_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;
  mp_real rE30_x = dE30_x;
  mp_real rE30_y = dE30_y;
  mp_real rE31_x = dE31_x;
  mp_real rE31_y = dE31_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real prEdgeZone[2];
  mp_real rE1dotE1, rE2dotE2, rE1dotE2, rE2dotE3, rE1dotE3;
  mp_real rE3dotE30, rE3dotE31;
  mp_real rE1xE2, rE10xE11, rE20xE21;
  mp_real rE1E2minusE1dotE2;
  mp_real rDisc, rPlusOrMinus;
  mp_real rDen;
  mp_real rE3dotE3;
  APM2D::EdgeZoneType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = rE10_x - rV0_x;
  ppprV[1][0][1] = rE10_y - rV0_y;
  ppprV[1][1][0] = rE11_x - rV0_x;
  ppprV[1][1][1] = rE11_y - rV0_y;
  ppprV[2][0][0] = rE20_x - rV0_x;
  ppprV[2][0][1] = rE20_y - rV0_y;
  ppprV[2][1][0] = rE21_x - rV0_x;
  ppprV[2][1][1] = rE21_y - rV0_y;
  ppprV[3][0][0] = rE30_x - rV0_x;
  ppprV[3][0][1] = rE30_y - rV0_y;
  ppprV[3][1][0] = rE31_x - rV0_x;
  ppprV[3][1][1] = rE31_y - rV0_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[1][0] = rE11_x - rE10_x;
  pprE[1][1] = rE11_y - rE10_y;
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;
  pprE[3][0] = rE31_x - rE30_x;
  pprE[3][1] = rE31_y - rE30_y;
  
  rE1dotE1 = (pprE[1][0]*pprE[1][0]) + (pprE[1][1]*pprE[1][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);
  rE1dotE2 = (pprE[1][0]*pprE[2][0]) + (pprE[1][1]*pprE[2][1]);
  rE2dotE3 = (pprE[2][0]*pprE[3][0]) + (pprE[2][1]*pprE[3][1]);
  rE1dotE3 = (pprE[1][0]*pprE[3][0]) + (pprE[1][1]*pprE[3][1]);
  rE3dotE3 = (pprE[3][0]*pprE[3][0]) + (pprE[3][1]*pprE[3][1]);

  rE3dotE30 = (pprE[3][0]*ppprV[3][0][0]) + (pprE[3][1]*ppprV[3][0][1]);
  rE3dotE31 = (pprE[3][0]*ppprV[3][1][0]) + (pprE[3][1]*ppprV[3][1][1]);
  
  rE1xE2 = (pprE[1][0]*pprE[2][1]) - (pprE[1][1]*pprE[2][0]);
  rE10xE11 = (ppprV[1][0][0]*ppprV[1][1][1]) - (ppprV[1][0][1]*ppprV[1][1][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);

  rE1E2minusE1dotE2 = ( sqrt(rE1dotE1*rE2dotE2) - rE1dotE2 );

  rDisc = 2.0*rE10xE11*rE20xE21*( sqrt(rE1dotE1*rE2dotE2) + rE1dotE2 );
  ASSERT( rDisc >= 0.0 ); 
  mp_real rSqrtDisc = sqrt(rDisc);

  ASSERT( rE1xE2 != 0.0 );
  rPlusOrMinus = 1.0;
  if ( rE1xE2 > 0.0 )
    {
      rPlusOrMinus = -1.0;
    }
  
  rDen = rE1xE2*rE1E2minusE1dotE2;
  ASSERT( rDen != 0.0 );

  mp_real rCoef1 = rE1xE2*rE1E2minusE1dotE2;
  mp_real rCoef2 = ( ( rE2dotE2*rE1dotE3 - rE1dotE2*rE2dotE3 )*rE10xE11
		     - ( rE1dotE1*rE2dotE3 - rE1dotE2*rE1dotE3 )*rE20xE21
		     + rPlusOrMinus*( sqrt(rE2dotE2)*rE1dotE3 - sqrt(rE1dotE1)*rE2dotE3 )*rSqrtDisc );
  /*
  prC[0] = ( (rE1dotE1*pprE[2][0] - rE1dotE2*pprE[1][0])*rE20xE21
	     + (rE1dotE2*pprE[2][0] - rE2dotE2*pprE[1][0])*rE10xE11
	     + rPlusOrMinus*(sqrt(rE1dotE1)*pprE[2][0] - sqrt(rE2dotE2)*pprE[1][0])*rSqrtDisc ) / rDen;
  prC[1] = ( (rE1dotE1*pprE[2][1] - rE1dotE2*pprE[1][1])*rE20xE21
	     + (rE1dotE2*pprE[2][1] - rE2dotE2*pprE[1][1])*rE10xE11
	     + rPlusOrMinus*(sqrt(rE1dotE1)*pprE[2][1] - sqrt(rE2dotE2)*pprE[1][1])*rSqrtDisc ) / rDen;

  rMagCE30 = ( (prC[0] - ppprV[3][0][0])*(prC[0] - ppprV[3][0][0]) + 
	       (prC[1] - ppprV[3][0][1])*(prC[1] - ppprV[3][0][1]) );
  rMagCE31 = ( (prC[0] - ppprV[3][1][0])*(prC[0] - ppprV[3][1][0]) + 
	       (prC[1] - ppprV[3][1][1])*(prC[1] - ppprV[3][1][1]) );

  if ( rMagCE30 <= 0.0 )
    {
      prEdgeZone[0] = 0.0;
    }
  else
    {
      prEdgeZone[0] = ( ( rCoef1*rE3dotE30
			  + rCoef2 ) /
			( rDen*sqrt(rE3dotE3*rMagCE30) ) );
    }
  if ( rMagCE31 <= 0.0 )
    {
      prEdgeZone[1] = 0.0;
    }
  else
    {
      prEdgeZone[1] = -( ( rCoef1*rE3dotE31
			   + rCoef2 ) /
			 ( rDen*sqrt(rE3dotE3*rMagCE31) ) );
    }
  */

  if ( rE1xE2 >= 0.0 )
    {
      prEdgeZone[0] = ( rCoef1*rE3dotE30
			+ rCoef2 );
      prEdgeZone[1] = -( rCoef1*rE3dotE31
			 + rCoef2 );
    }
  else
    {
      prEdgeZone[0] = -( rCoef1*rE3dotE30
			 + rCoef2 );
      prEdgeZone[1] = ( rCoef1*rE3dotE31
			+ rCoef2 );
    }

  if ( !( prEdgeZone[0] < 0.0 ) )
    {
      eResult = APM2D::eztStart;
    }
  else if ( !( prEdgeZone[1] < 0.0 ) )
    {
      eResult = APM2D::eztEnd;
    }
  else
    {
      eResult = APM2D::eztMiddle;
    }

  return eResult;
}

//static 
APM2D::CircleTestType InCircleVEE_VMPFUN(DOUBLE dV0_x, DOUBLE dV0_y,
					 DOUBLE dE10_x, DOUBLE dE10_y, 
					 DOUBLE dE11_x, DOUBLE dE11_y,
					 DOUBLE dE20_x, DOUBLE dE20_y, 
					 DOUBLE dE21_x, DOUBLE dE21_y,
					 DOUBLE dV3_x, DOUBLE dV3_y)
{
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rE10_x = dE10_x;
  mp_real rE10_y = dE10_y; 
  mp_real rE11_x = dE11_x;
  mp_real rE11_y = dE11_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;
  mp_real rV3_x = dV3_x;
  mp_real rV3_y = dV3_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real rE1dotE1, rE2dotE2, rV3dotV3;
  mp_real rE1dotE2, rE2dotV3, rE1dotV3;
  mp_real rE1xE2, rE10xE11, rE20xE21;
  mp_real rE1E2minusE1dotE2;
  mp_real rDisc, rPlusOrMinus;
  mp_real rDen;
  mp_real rResult;
  APM2D::CircleTestType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = rE10_x - rV0_x;
  ppprV[1][0][1] = rE10_y - rV0_y;
  ppprV[1][1][0] = rE11_x - rV0_x;
  ppprV[1][1][1] = rE11_y - rV0_y;
  ppprV[2][0][0] = rE20_x - rV0_x;
  ppprV[2][0][1] = rE20_y - rV0_y;
  ppprV[2][1][0] = rE21_x - rV0_x;
  ppprV[2][1][1] = rE21_y - rV0_y;
  ppprV[3][0][0] = rV3_x - rV0_x;
  ppprV[3][0][1] = rV3_y - rV0_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[1][0] = rE11_x - rE10_x;
  pprE[1][1] = rE11_y - rE10_y;
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;

  rE1dotE1 = (pprE[1][0]*pprE[1][0]) + (pprE[1][1]*pprE[1][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);
  rV3dotV3 = (ppprV[3][0][0]*ppprV[3][0][0]) + (ppprV[3][0][1]*ppprV[3][0][1]);

  rE1dotE2 = (pprE[1][0]*pprE[2][0]) + (pprE[1][1]*pprE[2][1]);
  rE2dotV3 = (pprE[2][0]*ppprV[3][0][0]) + (pprE[2][1]*ppprV[3][0][1]);
  rE1dotV3 = (pprE[1][0]*ppprV[3][0][0]) + (pprE[1][1]*ppprV[3][0][1]);

  rE1xE2 = (pprE[1][0]*pprE[2][1]) - (pprE[1][1]*pprE[2][0]);
  rE10xE11 = (ppprV[1][0][0]*ppprV[1][1][1]) - (ppprV[1][0][1]*ppprV[1][1][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);

  rE1E2minusE1dotE2 = ( sqrt(rE1dotE1*rE2dotE2) - rE1dotE2 );

  rDisc = 2.0*rE10xE11*rE20xE21*( sqrt(rE1dotE1*rE2dotE2) + rE1dotE2 );
  ASSERT( rDisc >= 0.0 ); 
  mp_real rSqrtDisc = sqrt(rDisc);

  ASSERT( rE1xE2 != 0.0 );
  rPlusOrMinus = 1.0;
  if ( rE1xE2 > 0.0 )
    {
      rPlusOrMinus = -1.0;
    }

  rDen = rE1xE2*rE1E2minusE1dotE2;
  ASSERT( rDen != 0.0 );

  rResult = ( rDen*rV3dotV3
	      + 2.0*( rE2dotE2*rE1dotV3 - rE1dotE2*rE2dotV3 )*rE10xE11
	      - 2.0*( rE1dotE1*rE2dotV3 - rE1dotE2*rE1dotV3 )*rE20xE21
	      + rPlusOrMinus*2.0*( sqrt(rE2dotE2)*rE1dotV3 - sqrt(rE1dotE1)*rE2dotV3 )*rSqrtDisc ); 
  if ( rE1xE2 < 0.0 )
    {
      rResult = -rResult;
    }

  if ( rResult < 0.0 )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( rResult > 0.0 )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

//static 
APM2D::CircleTestType InCircleVEE_EMPFUN(DOUBLE dV0_x, DOUBLE dV0_y,
					 DOUBLE dE10_x, DOUBLE dE10_y, 
					 DOUBLE dE11_x, DOUBLE dE11_y,
					 DOUBLE dE20_x, DOUBLE dE20_y, 
					 DOUBLE dE21_x, DOUBLE dE21_y,
					 DOUBLE dE30_x, DOUBLE dE30_y,
					 DOUBLE dE31_x, DOUBLE dE31_y)
{
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rE10_x = dE10_x;
  mp_real rE10_y = dE10_y; 
  mp_real rE11_x = dE11_x;
  mp_real rE11_y = dE11_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;
  mp_real rE30_x = dE30_x;
  mp_real rE30_y = dE30_y;
  mp_real rE31_x = dE31_x;
  mp_real rE31_y = dE31_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real rE1dotE1, rE2dotE2, rE3dotE3;
  mp_real rE1dotE2;
  mp_real rE1xE2, rE2xE3, rE1xE3;
  mp_real rE10xE11, rE20xE21, rE30xE31;
  mp_real rE1E2plusE1dotE2, rE1E2minusE1dotE2;
  mp_real rDisc, rPlusOrMinus, rA, rB;
  //mp_real rDen;
  mp_real rResult;
  APM2D::CircleTestType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = rE10_x - rV0_x;
  ppprV[1][0][1] = rE10_y - rV0_y;
  ppprV[1][1][0] = rE11_x - rV0_x;
  ppprV[1][1][1] = rE11_y - rV0_y;
  ppprV[2][0][0] = rE20_x - rV0_x;
  ppprV[2][0][1] = rE20_y - rV0_y;
  ppprV[2][1][0] = rE21_x - rV0_x;
  ppprV[2][1][1] = rE21_y - rV0_y;
  ppprV[3][0][0] = rE30_x - rV0_x;
  ppprV[3][0][1] = rE30_y - rV0_y;
  ppprV[3][1][0] = rE31_x - rV0_x;
  ppprV[3][1][1] = rE31_y - rV0_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[1][0] = rE11_x - rE10_x;
  pprE[1][1] = rE11_y - rE10_y;
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;
  pprE[3][0] = rE31_x - rE30_x;
  pprE[3][1] = rE31_y - rE30_y;
  
  rE1dotE1 = (pprE[1][0]*pprE[1][0]) + (pprE[1][1]*pprE[1][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);
  rE3dotE3 = (pprE[3][0]*pprE[3][0]) + (pprE[3][1]*pprE[3][1]);
  
  rE1dotE2 = (pprE[1][0]*pprE[2][0]) + (pprE[1][1]*pprE[2][1]);
  
  rE1xE2 = (pprE[1][0]*pprE[2][1]) - (pprE[1][1]*pprE[2][0]);
  rE2xE3 = (pprE[2][0]*pprE[3][1]) - (pprE[2][1]*pprE[3][0]);
  rE1xE3 = (pprE[1][0]*pprE[3][1]) - (pprE[1][1]*pprE[3][0]);
  
  rE10xE11 = (ppprV[1][0][0]*ppprV[1][1][1]) - (ppprV[1][0][1]*ppprV[1][1][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);
  rE30xE31 = (ppprV[3][0][0]*ppprV[3][1][1]) - (ppprV[3][0][1]*ppprV[3][1][0]);
  
  rE1E2plusE1dotE2 = ( sqrt(rE1dotE1*rE2dotE2) + rE1dotE2 );
  rE1E2minusE1dotE2 = ( sqrt(rE1dotE1*rE2dotE2) - rE1dotE2 );
  
  rDisc = 2.0*rE10xE11*rE20xE21*rE1E2plusE1dotE2;
  ASSERT( rDisc >= 0.0 ); 
  mp_real rSqrtDisc = sqrt(rDisc);
  
  ASSERT( rE1xE2 != 0.0 );
  rPlusOrMinus = 1.0;
  if ( rE1xE2 > 0.0 )
    {
      rPlusOrMinus = -1.0;
    }
  
  rA = ( ( rE1dotE2*rE1xE3 - rE1dotE1*rE2xE3 )*rE20xE21
	 + ( rE2dotE2*rE1xE3 - rE1dotE2*rE2xE3 )*rE10xE11
	 + rPlusOrMinus*( sqrt(rE2dotE2)*rE1xE3 - sqrt(rE1dotE1)*rE2xE3 )*rSqrtDisc );
  
  rB = ( ( sqrt(rE1dotE1)*rE20xE21 + sqrt(rE2dotE2)*rE10xE11 )*( sqrt(rE1dotE1)*rE20xE21 + sqrt(rE2dotE2)*rE10xE11 )
	 + 2.0*rE10xE11*rE20xE21*rE1E2plusE1dotE2
	 + rPlusOrMinus*2.0*( sqrt(rE1dotE1)*rE20xE21 + sqrt(rE2dotE2)*rE10xE11 )*rSqrtDisc );
  
  //rDen = rE3dotE3*rE1xE2*rE1xE2*rE1E2minusE1dotE2*rE1E2minusE1dotE2;
  //ASSERT( rDen != 0.0 );
  
  rResult = ( rA*rA
	      + 2.0*rE30xE31*rE1xE2*rE1E2minusE1dotE2*rA
	      + rE30xE31*rE30xE31*rE1xE2*rE1xE2*rE1E2minusE1dotE2*rE1E2minusE1dotE2
	      - rE3dotE3*rE1xE2*rE1xE2*rB );
  
  if ( rResult < 0.0 )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( rResult > 0.0 )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }
  
  return eResult;
}

//static 
APM2D::InscribedCircleCenterType 
InscribedCircleCenterVEEAntiparallelMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					  DOUBLE dV00_x, DOUBLE dV00_y,
					  DOUBLE dV0_x, DOUBLE dV0_y,
					  DOUBLE dV01_x, DOUBLE dV01_y,
					  DOUBLE dE10_x, DOUBLE dE10_y, 
					  DOUBLE dE11_x, DOUBLE dE11_y,
					  DOUBLE dE20_x, DOUBLE dE20_y, 
					  DOUBLE dE21_x, DOUBLE dE21_y)
{
  mp_real prC[3], rR;
  mp_real rV00_x = dV00_x;
  mp_real rV00_y = dV00_y;
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rV01_x = dV01_x;
  mp_real rV01_y = dV01_y;
  mp_real rE10_x = dE10_x;
  mp_real rE10_y = dE10_y; 
  mp_real rE11_x = dE11_x;
  mp_real rE11_y = dE11_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;
  
  mp_real ppprV[3][2][2], ppprVN[1][2][2]; // Endpoints
  mp_real pprE[3][2];     // Edge vectors
  mp_real prVertexZone[2];
  mp_real pprEdgeZone[2][2];
  mp_real rDen, rDenR, rDisc, rDiscE2;
  mp_real rE1dotE1, rE2dotE2, rE1dotE2;
  mp_real rE1dotE10, rE1dotE11, rE2dotE20, rE2dotE21;
  mp_real rE10xE11, rE20xE21, rE1xE20, rE2xE10;
  mp_real rV00dotV00, rV01dotV01, rE1dotV00, rE1dotV01, rE1xV00, rE1xV01;
  mp_real rE10dotE10, rE11dotE11, rE20dotE20, rE21dotE21;
  mp_real rE1E2minusE1dotE2, rSinTheta;
  UINT_32 i, j;
  
  // Translate V1 to the origin
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = rE10_x - rV0_x;
  ppprV[1][0][1] = rE10_y - rV0_y;
  ppprV[1][1][0] = rE11_x - rV0_x;
  ppprV[1][1][1] = rE11_y - rV0_y;
  ppprV[2][0][0] = rE20_x - rV0_x;
  ppprV[2][0][1] = rE20_y - rV0_y;
  ppprV[2][1][0] = rE21_x - rV0_x;
  ppprV[2][1][1] = rE21_y - rV0_y;
  
  // Vertex neighbor vectors
  //memset(ppprVN, 0, sizeof(ppprV));
  ppprVN[0][0][0] = rV0_x - rV00_x;
  ppprVN[0][0][1] = rV0_y - rV00_y;
  ppprVN[0][1][0] = rV0_x - rV01_x;
  ppprVN[0][1][1] = rV0_y - rV01_y;

  //memset(pprE, 0, sizeof(pprE));
  pprE[1][0] = rE11_x - rE10_x;
  pprE[1][1] = rE11_y - rE10_y;
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;
  
  rE1dotE1 = (pprE[1][0]*pprE[1][0]) + (pprE[1][1]*pprE[1][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);
  rE1dotE2 = (pprE[1][0]*pprE[2][0]) + (pprE[1][1]*pprE[2][1]);
  rE1dotE10 = (pprE[1][0]*ppprV[1][0][0]) + (pprE[1][1]*ppprV[1][0][1]);
  rE1dotE11 = (pprE[1][0]*ppprV[1][1][0]) + (pprE[1][1]*ppprV[1][1][1]);
  rE2dotE20 = (pprE[2][0]*ppprV[2][0][0]) + (pprE[2][1]*ppprV[2][0][1]);
  rE2dotE21 = (pprE[2][0]*ppprV[2][1][0]) + (pprE[2][1]*ppprV[2][1][1]);

  rE10xE11 = (ppprV[1][0][0]*ppprV[1][1][1]) - (ppprV[1][0][1]*ppprV[1][1][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);
  rE1xE20 = (pprE[1][0]*ppprV[2][0][1]) - (pprE[1][1]*ppprV[2][0][0]);
  rE2xE10 = (pprE[2][0]*ppprV[1][0][1]) - (pprE[2][1]*ppprV[1][0][0]);

  rV00dotV00 = ( (ppprVN[0][0][0] * ppprVN[0][0][0]) + (ppprVN[0][0][1] * ppprVN[0][0][1]) );
  rV01dotV01 = ( (ppprVN[0][1][0] * ppprVN[0][1][0]) + (ppprVN[0][1][1] * ppprVN[0][1][1]) );
  rE10dotE10 = (ppprV[1][0][0]*ppprV[1][0][0]) + (ppprV[1][0][1]*ppprV[1][0][1]);
  rE11dotE11 = (ppprV[1][1][0]*ppprV[1][1][0]) + (ppprV[1][1][1]*ppprV[1][1][1]);
  rE20dotE20 = (ppprV[2][0][0]*ppprV[2][0][0]) + (ppprV[2][0][1]*ppprV[2][0][1]);
  rE21dotE21 = (ppprV[2][1][0]*ppprV[2][1][0]) + (ppprV[2][1][1]*ppprV[2][1][1]);

  rE1dotV00 = ( (pprE[1][0] * ppprVN[0][0][0]) + (pprE[1][1] * ppprVN[0][0][1]) );
  rE1dotV01 = ( (pprE[1][0] * ppprVN[0][1][0]) + (pprE[1][1] * ppprVN[0][1][1]) );

  rE1xV00 = ( (pprE[1][0] * ppprVN[0][0][1]) - (pprE[1][1] * ppprVN[0][0][0]) );
  rE1xV01 = ( (pprE[1][0] * ppprVN[0][1][1]) - (pprE[1][1] * ppprVN[0][1][0]) );

  /*
  // JORDAN
  ASSERT( !( fabs(rE1dotE1) <= s_crEpsilonInscribedCircleCenterVEEAntiparallelNullEdge ) );
  */

  rE1E2minusE1dotE2 = sqrt(rE1dotE1*rE2dotE2) - rE1dotE2;
  rSinTheta = rE1E2minusE1dotE2;
  if ( rSinTheta == 0.0 )
    {
      // Colinear
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  //rSinTheta = rE10xE11/sqrt(rE10dotE10*rE11dotE11);
  rSinTheta = SinThetaMPFUN(rV0_x, rV0_y, rE10_x, rE10_y, rE11_x, rE11_y);
  /*
  rSinThetaTemp = SinThetaMPFUN(rE10_x, rE10_y, rE11_x, rE11_y, rV0_x, rV0_y);
  if ( rSinThetaTemp > rSinTheta )
    {
      rSinTheta = rSinThetaTemp;
    }
  */
  if ( rSinTheta <= 0.0 )
    {
      // V0 is on E1 and it is not one of the endpoints
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  //rSinTheta = rE20xE21/sqrt(rE20dotE20*rE21dotE21);
  rSinTheta = SinThetaMPFUN(rV0_x, rV0_y, rE20_x, rE20_y, rE21_x, rE21_y);
  /*
  rSinThetaTemp = SinThetaMPFUN(rE20_x, rE20_y, rE21_x, rE21_y, rV0_x, rV0_y);
  if ( rSinThetaTemp > rSinTheta )
    {
      rSinTheta = rSinThetaTemp;
    }
  */
  if ( rSinTheta <= 0.0 )
    {
      // V0 is on E2 and it is not one of the endpoints
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  rDen = 2*rE1dotE1;
  ASSERT( rDen != 0.0 );
  rDenR = 2*sqrt(rE1dotE1);
  ASSERT( rDenR != 0.0 );

  rDisc = rE1xE20*rE10xE11;
  if ( rDisc < 0.0 ) 
    {
      // Imaginary Solutions
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  mp_real rSqrtDisc = sqrt(rDisc);
  
  rDiscE2 = rE2xE10*rE20xE21;
  if ( rDiscE2 < 0.0 ) 
    {
      // Imaginary Solutions
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  mp_real rSqrtDiscE2 = sqrt(rDiscE2);
  

  //////////////////////////////////////////////////
  // Check the Edge Zone
  /*
  rMagCE10 = ( (prC[0] - ppprV[1][0][0])*(prC[0] - ppprV[1][0][0]) + 
	       (prC[1] - ppprV[1][0][1])*(prC[1] - ppprV[1][0][1]) );
  rMagCE11 = ( (prC[0] - ppprV[1][1][0])*(prC[0] - ppprV[1][1][0]) + 
	       (prC[1] - ppprV[1][1][1])*(prC[1] - ppprV[1][1][1]) );
  rMagCE20 = ( (prC[0] - ppprV[2][0][0])*(prC[0] - ppprV[2][0][0]) + 
	       (prC[1] - ppprV[2][0][1])*(prC[1] - ppprV[2][0][1]) );
  rMagCE21 = ( (prC[0] - ppprV[2][1][0])*(prC[0] - ppprV[2][1][0]) + 
	       (prC[1] - ppprV[2][1][1])*(prC[1] - ppprV[2][1][1]) );

  if ( rMagCE10 <= 0.0 )
    {
      pprEdgeZone[0][0] = 0.0;
    }
  else
    {
      pprEdgeZone[0][0] = -( ( rE1dotE10 - rSqrtDisc ) /
			     ( sqrt(rE1dotE1*rMagCE10) ) );
    }
  if ( rMagCE11 <= 0.0 )
    {
      pprEdgeZone[0][1] = 0.0;
    }
  else
    {
      pprEdgeZone[0][1] = ( ( rE1dotE11 - rSqrtDisc ) /
			    ( sqrt(rE1dotE1*rMagCE11) ) );
    }
  if ( rMagCE20 <= 0.0 )
    {
      pprEdgeZone[1][0] = 0.0;
    }
  else
    {
      pprEdgeZone[1][0] = -( ( rE2dotE20 + rSqrtDiscE2 ) /
			     ( sqrt(rE2dotE2*rMagCE20) ) );
    }
  if ( rMagCE21 <= 0.0 )
    {
      pprEdgeZone[1][1] = 0.0;
    }
  else
    {
      pprEdgeZone[1][1] = ( ( rE2dotE21 + rSqrtDiscE2 ) /
			    ( sqrt(rE2dotE2*rMagCE21) ) );
    }
  */
  pprEdgeZone[0][0] = -( rE1dotE10 - rSqrtDisc );
  pprEdgeZone[0][1] = ( rE1dotE11 - rSqrtDisc );
  pprEdgeZone[1][0] = -( rE2dotE20 + rSqrtDiscE2 );
  pprEdgeZone[1][1] = ( rE2dotE21 + rSqrtDiscE2 );

  for ( i = 0; i < 2; i++ )
    {
      for ( j = 0; j < 2; j++ )
	{
	  if ( pprEdgeZone[i][j] < 0.0 )
	    {
	      *pdC_x = 0.0;
	      *pdC_y = 0.0;
	      *pdR = 0.0;
	      return APM2D::icctNull;
	    }
	}
    }

  //////////////////////////////////////////////////
  // Check the Vertex Zone

  //mp_real rDentimesR = rDen*rR;
  mp_real rE1xE20timesE10xE11 = rE1xE20 - rE10xE11;

  //  prVertexZone[0] = ( ( ( rE1xE20timesE10xE11 )*rE1xV00
  //			+ 2.0*rE1dotV00*rSqrtDisc ) / 
  //		      ( rDentimesR*sqrt(rV00dotV00) ) );
  //  prVertexZone[1] = ( ( ( rE1xE20timesE10xE11 )*rE1xV01
  //			+ 2.0*rE1dotV01*rSqrtDisc ) / 
  //		      ( rDentimesR*sqrt(rV01dotV01) ) );

  prVertexZone[0] = ( ( rE1xE20timesE10xE11 )*rE1xV00
		      + 2.0*rE1dotV00*rSqrtDisc );
  prVertexZone[1] = ( ( rE1xE20timesE10xE11 )*rE1xV01
			+ 2.0*rE1dotV01*rSqrtDisc );

  for ( i = 0; i < 2; i++ )
    {
      if ( prVertexZone[i] < 0.0 )
	{
	  *pdC_x = 0.0;
	  *pdC_y = 0.0;
	  *pdR = 0.0;
	  return APM2D::icctNull;
	}
    }

  //////////////////////////////////////////////////
  // Compute the center and radius

  prC[0] = ( -(rE1xE20 - rE10xE11)*pprE[1][1] + 2.0*pprE[1][0]*rSqrtDisc ) / rDen;
  prC[1] = ( (rE1xE20 - rE10xE11)*pprE[1][0] + 2.0*pprE[1][1]*rSqrtDisc ) / rDen;
  //prC[2] = sqrt((prC[0]*prC[0]) + (prC[1]*prC[1]));

  rR = ( rE1xE20 + rE10xE11 ) / rDenR;
  
  /*
  if ( !( fabs(prC[2] - rR) < s_crEpsilonInscribedCircleCenterVEERadius ) )
    {
      fprintf(strerr, "VEE Antiparallel: C = ( %f, %f, %f ) != %f = R\n", prC[0], prC[1], prC[2], rR);
      //      ASSERT( FALSE );
    }
  */

  // Translate the answer back out away from the origin
  prC[0] += rV0_x;
  prC[1] += rV0_y;

  *pdC_x = dble(prC[0]);
  *pdC_y = dble(prC[1]);
  *pdR = dble(rR);
  return APM2D::icctVEEAntiparallel;
}

//static 
APM2D::EdgeZoneType EdgeZoneVEEAntiparallelMPFUN(DOUBLE dV0_x, DOUBLE dV0_y,
						 DOUBLE dE10_x, DOUBLE dE10_y, 
						 DOUBLE dE11_x, DOUBLE dE11_y,
						 DOUBLE dE20_x, DOUBLE dE20_y, 
						 DOUBLE dE21_x, DOUBLE dE21_y,
						 DOUBLE dE30_x, DOUBLE dE30_y,
						 DOUBLE dE31_x, DOUBLE dE31_y)
{
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rE10_x = dE10_x;
  mp_real rE10_y = dE10_y; 
  mp_real rE11_x = dE11_x;
  mp_real rE11_y = dE11_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;
  mp_real rE30_x = dE30_x;
  mp_real rE30_y = dE30_y;
  mp_real rE31_x = dE31_x;
  mp_real rE31_y = dE31_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real prEdgeZone[2];
  mp_real rE1dotE1, rE1dotE3;
  mp_real rE3dotE30, rE3dotE31;
  mp_real rE1xE3, rE1xE20, rE10xE11;
  mp_real rDisc;
  //mp_real rDen;
  mp_real rE3dotE3;
  APM2D::EdgeZoneType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = rE10_x - rV0_x;
  ppprV[1][0][1] = rE10_y - rV0_y;
  ppprV[1][1][0] = rE11_x - rV0_x;
  ppprV[1][1][1] = rE11_y - rV0_y;
  ppprV[2][0][0] = rE20_x - rV0_x;
  ppprV[2][0][1] = rE20_y - rV0_y;
  ppprV[2][1][0] = rE21_x - rV0_x;
  ppprV[2][1][1] = rE21_y - rV0_y;
  ppprV[3][0][0] = rE30_x - rV0_x;
  ppprV[3][0][1] = rE30_y - rV0_y;
  ppprV[3][1][0] = rE31_x - rV0_x;
  ppprV[3][1][1] = rE31_y - rV0_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[1][0] = rE11_x - rE10_x;
  pprE[1][1] = rE11_y - rE10_y;
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;
  pprE[3][0] = rE31_x - rE30_x;
  pprE[3][1] = rE31_y - rE30_y;
  
  rE1dotE1 = (pprE[1][0]*pprE[1][0]) + (pprE[1][1]*pprE[1][1]);
  rE1dotE3 = (pprE[1][0]*pprE[3][0]) + (pprE[1][1]*pprE[3][1]);
  rE3dotE3 = (pprE[3][0]*pprE[3][0]) + (pprE[3][1]*pprE[3][1]);

  rE3dotE30 = (pprE[3][0]*ppprV[3][0][0]) + (pprE[3][1]*ppprV[3][0][1]);
  rE3dotE31 = (pprE[3][0]*ppprV[3][1][0]) + (pprE[3][1]*ppprV[3][1][1]);
  
  rE1xE3 = (pprE[1][0]*pprE[3][1]) - (pprE[1][1]*pprE[3][0]);
  rE1xE20 = (pprE[1][0]*ppprV[2][0][1]) - (pprE[1][1]*ppprV[2][0][0]);
  rE10xE11 = (ppprV[1][0][0]*ppprV[1][1][1]) - (ppprV[1][0][1]*ppprV[1][1][0]);

  rDisc = rE1xE20*rE10xE11;
  ASSERT( rDisc >= 0.0 ); 
  mp_real rSqrtDisc = sqrt(rDisc);

  //rDen = 2.0*rE1dotE1;
  //ASSERT( rDen != 0.0 );

  mp_real rE1xE20minusE10xE11 = rE1xE20 - rE10xE11;
  mp_real rCoef1 = ( - ( rE1xE20minusE10xE11 )*rE1xE3
		     - 2.0*rE1dotE3*rSqrtDisc );
  /*
  prC[0] = ( -(rE1xE20minusE10xE11)*pprE[1][1] + 2.0*pprE[1][0]*rSqrtDisc ) / rDen;
  prC[1] = ( (rE1xE20minusE10xE11)*pprE[1][0] + 2.0*pprE[1][1]*rSqrtDisc ) / rDen;

  rMagCE30 = ( (prC[0] - ppprV[3][0][0])*(prC[0] - ppprV[3][0][0]) + 
	       (prC[1] - ppprV[3][0][1])*(prC[1] - ppprV[3][0][1]) );
  rMagCE31 = ( (prC[0] - ppprV[3][1][0])*(prC[0] - ppprV[3][1][0]) + 
	       (prC[1] - ppprV[3][1][1])*(prC[1] - ppprV[3][1][1]) );

  if ( rMagCE30 <= 0.0 )
    {
      prEdgeZone[0] = 0.0;
    }
  else
    {
      prEdgeZone[0] = ( ( 2.0*rE1dotE1*rE3dotE30 
			  + rCoef1 ) /
			( rDen*sqrt(rE3dotE3*rMagCE30) ) );
    }
  if ( rMagCE31 <= 0.0 )
    {
      prEdgeZone[1] = 0.0;
    }
  else
    {
      prEdgeZone[1] = -( ( 2.0*rE1dotE1*rE3dotE31 
			   + rCoef1 ) /
			 ( rDen*sqrt(rE3dotE3*rMagCE31) ) );
    }
  */
  prEdgeZone[0] = ( 2.0*rE1dotE1*rE3dotE30 
		    + rCoef1 );
  prEdgeZone[1] = -( 2.0*rE1dotE1*rE3dotE31 
		     + rCoef1 );
		     
  if ( !( prEdgeZone[0] < 0.0 ) )
    {
      eResult = APM2D::eztStart;
    }
  else if ( !( prEdgeZone[1] < 0.0 ) )
    {
      eResult = APM2D::eztEnd;
    }
  else
    {
      eResult = APM2D::eztMiddle;
    }

  return eResult;
}

//static 
APM2D::CircleTestType InCircleVEE_VAntiparallelMPFUN(DOUBLE dV0_x, DOUBLE dV0_y,
						     DOUBLE dE10_x, DOUBLE dE10_y, 
						     DOUBLE dE11_x, DOUBLE dE11_y,
						     DOUBLE dE20_x, DOUBLE dE20_y, 
						     DOUBLE dE21_x, DOUBLE dE21_y,
						     DOUBLE dV3_x, DOUBLE dV3_y)
{
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rE10_x = dE10_x;
  mp_real rE10_y = dE10_y; 
  mp_real rE11_x = dE11_x;
  mp_real rE11_y = dE11_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;
  mp_real rV3_x = dV3_x;
  mp_real rV3_y = dV3_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real rE1dotE1, rV3dotV3, rE1dotV3;
  mp_real rE1xE20, rE1xV3, rE10xE11;
  mp_real rDisc;
  //mp_real rDen;
  mp_real rResult;
  APM2D::CircleTestType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = rE10_x - rV0_x;
  ppprV[1][0][1] = rE10_y - rV0_y;
  ppprV[1][1][0] = rE11_x - rV0_x;
  ppprV[1][1][1] = rE11_y - rV0_y;
  ppprV[2][0][0] = rE20_x - rV0_x;
  ppprV[2][0][1] = rE20_y - rV0_y;
  ppprV[2][1][0] = rE21_x - rV0_x;
  ppprV[2][1][1] = rE21_y - rV0_y;
  ppprV[3][0][0] = rV3_x - rV0_x;
  ppprV[3][0][1] = rV3_y - rV0_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[1][0] = rE11_x - rE10_x;
  pprE[1][1] = rE11_y - rE10_y;
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;

  rE1dotE1 = (pprE[1][0]*pprE[1][0]) + (pprE[1][1]*pprE[1][1]);
  rV3dotV3 = (ppprV[3][0][0]*ppprV[3][0][0]) + (ppprV[3][0][1]*ppprV[3][0][1]);
  rE1dotV3 = (pprE[1][0]*ppprV[3][0][0]) + (pprE[1][1]*ppprV[3][0][1]);

  rE1xV3 = (pprE[1][0]*ppprV[3][0][1]) - (pprE[1][1]*ppprV[3][0][0]);
  rE1xE20 = (pprE[1][0]*ppprV[2][0][1]) - (pprE[1][1]*ppprV[2][0][0]);
  rE10xE11 = (ppprV[1][0][0]*ppprV[1][1][1]) - (ppprV[1][0][1]*ppprV[1][1][0]);

  rDisc = rE1xE20*rE10xE11;
  ASSERT( rDisc >= 0.0 ); 
  mp_real rSqrtDisc = sqrt(rDisc);

  //rDen = rE1dotE1;
  //ASSERT( rDen != 0.0 );

  rResult = ( rE1dotE1*rV3dotV3 
	      - ( rE1xE20 - rE10xE11 )*rE1xV3
	      - 2.0*rE1dotV3*rSqrtDisc );

  if ( rResult < 0.0 )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( rResult > 0.0 )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

//static 
APM2D::CircleTestType InCircleVEE_EAntiparallelMPFUN(DOUBLE dV0_x, DOUBLE dV0_y,
						     DOUBLE dE10_x, DOUBLE dE10_y, 
						     DOUBLE dE11_x, DOUBLE dE11_y,
						     DOUBLE dE20_x, DOUBLE dE20_y, 
						     DOUBLE dE21_x, DOUBLE dE21_y,
						     DOUBLE dE30_x, DOUBLE dE30_y,
						     DOUBLE dE31_x, DOUBLE dE31_y)
{
  mp_real rV0_x = dV0_x;
  mp_real rV0_y = dV0_y;
  mp_real rE10_x = dE10_x;
  mp_real rE10_y = dE10_y; 
  mp_real rE11_x = dE11_x;
  mp_real rE11_y = dE11_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;
  mp_real rE30_x = dE30_x;
  mp_real rE30_y = dE30_y;
  mp_real rE31_x = dE31_x;
  mp_real rE31_y = dE31_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real rE1dotE1, rE2dotE2, rE3dotE3, rE1dotE2, rE1dotE3;
  mp_real rE1xE20, rE1xE3, rE10xE11, rE20xE21, rE30xE31;
  mp_real rE1E2minusE1dotE2, rE1xE20minusE10xE11; 
  mp_real rDisc;
  //mp_real rDen;
  mp_real rResult;
  APM2D::CircleTestType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = rE10_x - rV0_x;
  ppprV[1][0][1] = rE10_y - rV0_y;
  ppprV[1][1][0] = rE11_x - rV0_x;
  ppprV[1][1][1] = rE11_y - rV0_y;
  ppprV[2][0][0] = rE20_x - rV0_x;
  ppprV[2][0][1] = rE20_y - rV0_y;
  ppprV[2][1][0] = rE21_x - rV0_x;
  ppprV[2][1][1] = rE21_y - rV0_y;
  ppprV[3][0][0] = rE30_x - rV0_x;
  ppprV[3][0][1] = rE30_y - rV0_y;
  ppprV[3][1][0] = rE31_x - rV0_x;
  ppprV[3][1][1] = rE31_y - rV0_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[1][0] = rE11_x - rE10_x;
  pprE[1][1] = rE11_y - rE10_y;
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;
  pprE[3][0] = rE31_x - rE30_x;
  pprE[3][1] = rE31_y - rE30_y;

  rE1dotE1 = (pprE[1][0]*pprE[1][0]) + (pprE[1][1]*pprE[1][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);
  rE3dotE3 = (pprE[3][0]*pprE[3][0]) + (pprE[3][1]*pprE[3][1]);
  rE1dotE2 = (pprE[1][0]*pprE[2][0]) + (pprE[1][1]*pprE[2][1]);
  rE1dotE3 = (pprE[1][0]*pprE[3][0]) + (pprE[1][1]*pprE[3][1]);

  rE1xE20 = (pprE[1][0]*ppprV[2][0][1]) - (pprE[1][1]*ppprV[2][0][0]);
  rE1xE3 = (pprE[1][0]*pprE[3][1]) - (pprE[1][1]*pprE[3][0]);
  rE10xE11 = (ppprV[1][0][0]*ppprV[1][1][1]) - (ppprV[1][0][1]*ppprV[1][1][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);
  rE30xE31 = (ppprV[3][0][0]*ppprV[3][1][1]) - (ppprV[3][0][1]*ppprV[3][1][0]);

  rE1E2minusE1dotE2 = ( sqrt(rE1dotE1*rE2dotE2) - rE1dotE2 );
  rE1xE20minusE10xE11 = rE1xE20 - rE10xE11;

  rDisc = rE1xE20*rE10xE11;
  ASSERT( rDisc >= 0.0 ); 
  mp_real rSqrtDisc = sqrt(rDisc);

  //rDen = 4.0*rE1dotE1*rE1dotE1*rE1E2minusE1dotE2*rE1E2minusE1dotE2;
  //ASSERT( rDen != 0.0 );

  rResult = ( 4.0*rE1dotE1*rE30xE31*( rE1dotE1*rE30xE31 + rE1dotE3*rE1xE20minusE10xE11 )
	      - 4.0*rE1dotE3*rE1dotE3*rE1xE20*rE20xE21
	      - rE1xE3*rE1xE3*rE1xE20minusE10xE11*rE1xE20minusE10xE11
	      + 4.0*rE1xE3*( 2.0*rE1dotE1*rE30xE31 + rE1dotE3*rE1xE20minusE10xE11 )*rSqrtDisc );
  
  if ( rResult < 0.0 )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( rResult > 0.0 )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

//static 
APM2D::InscribedCircleCenterType 
InscribedCircleCentervEEMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			      DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
			      DOUBLE dE11_x, DOUBLE dE11_y,
			      DOUBLE dE20_x, DOUBLE dE20_y, 
			      DOUBLE dE21_x, DOUBLE dE21_y)
{
  mp_real prC[3], rR;
  mp_real rV0_E10_x = dV0_E10_x;
  mp_real rV0_E10_y = dV0_E10_y; 
  mp_real rE11_x = dE11_x;
  mp_real rE11_y = dE11_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;

  mp_real ppprV[3][2][2]; // Endpoints
  mp_real pprE[3][2];     // Edge vectors
  mp_real prEdgeZone[2];
  mp_real rDen;
  mp_real rE1dotE1, rE2dotE2, rE1dotE2;
  mp_real rE2dotE20, rE2dotE21;
  mp_real rE1xE2, rE20xE21;
  mp_real rE20dotE20, rE21dotE21;
  mp_real rE1E2minusE1dotE2, rSinTheta;
  UINT_32 i;
  
  // Translate V1 to the origin
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = 0.0;
  ppprV[1][0][1] = 0.0;
  ppprV[1][1][0] = rE11_x - rV0_E10_x;
  ppprV[1][1][1] = rE11_y - rV0_E10_y;
  ppprV[2][0][0] = rE20_x - rV0_E10_x;
  ppprV[2][0][1] = rE20_y - rV0_E10_y;
  ppprV[2][1][0] = rE21_x - rV0_E10_x;
  ppprV[2][1][1] = rE21_y - rV0_E10_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[1][0] = rE11_x - rV0_E10_x;
  pprE[1][1] = rE11_y - rV0_E10_y;
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;
  
  rE1dotE1 = (pprE[1][0]*pprE[1][0]) + (pprE[1][1]*pprE[1][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);
  rE1dotE2 = (pprE[1][0]*pprE[2][0]) + (pprE[1][1]*pprE[2][1]);
  rE2dotE20 = (pprE[2][0]*ppprV[2][0][0]) + (pprE[2][1]*ppprV[2][0][1]);
  rE2dotE21 = (pprE[2][0]*ppprV[2][1][0]) + (pprE[2][1]*ppprV[2][1][1]);

  rE1xE2 = (pprE[1][0]*pprE[2][1]) - (pprE[1][1]*pprE[2][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);

  rE20dotE20 = (ppprV[2][0][0]*ppprV[2][0][0]) + (ppprV[2][0][1]*ppprV[2][0][1]);
  rE21dotE21 = (ppprV[2][1][0]*ppprV[2][1][0]) + (ppprV[2][1][1]*ppprV[2][1][1]);

  rE1E2minusE1dotE2 = sqrt(rE1dotE1*rE2dotE2) - rE1dotE2;

  //rOneMinusCosTheta = rE1E2minusE1dotE2/sqrt(rE1dotE1*rE2dotE2);
  rSinTheta = rE1xE2;
  if ( rSinTheta == 0.0 && rE1dotE2 > 0.0 )
    {
      // Colinear
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  //rSinTheta = rE20xE21/sqrt(rE20dotE20*rE21dotE21);
  rSinTheta = SinThetaMPFUN(rV0_E10_x, rV0_E10_y, rE20_x, rE20_y, rE21_x, rE21_y);
  /*
  rSinThetaTemp = SinThetaMPFUN(rE20_x, rE20_y, rE21_x, rE21_y, rV0_E10_x, rV0_E10_y);
  if ( rSinThetaTemp > rSinTheta )
    {
      rSinTheta = rSinThetaTemp;
    }
  */
  if ( rSinTheta <= 0.0 )
    {
      // V0 is on E2 and it is not one of the endpoints
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  rDen = rE1E2minusE1dotE2;
  ASSERT( rDen != 0.0 );

  //////////////////////////////////////////////////
  // Check the Edge Zone

  mp_real rE20xE21timesE1xE2 = rE20xE21*rE1xE2;
  /*
  rMagCE20 = ( (prC[0] - ppprV[2][0][0])*(prC[0] - ppprV[2][0][0]) + 
	       (prC[1] - ppprV[2][0][1])*(prC[1] - ppprV[2][0][1]) );
  rMagCE21 = ( (prC[0] - ppprV[2][1][0])*(prC[0] - ppprV[2][1][0]) + 
	       (prC[1] - ppprV[2][1][1])*(prC[1] - ppprV[2][1][1]) );

  if ( rMagCE20 <= 0.0 )
    {
      prEdgeZone[0] = 0.0;
    }
  else
    {
      prEdgeZone[0] = -( ( rE1E2minusE1dotE2*rE2dotE20 - rE20xE21timesE1xE2 ) / 
			 ( rE1E2minusE1dotE2*sqrt(rE2dotE2*rMagCE20) ) );
    }
  if ( rMagCE21 <= 0.0 )
    {
      prEdgeZone[1] = 0.0;
    }
  else
    {
      prEdgeZone[1] = ( ( rE1E2minusE1dotE2*rE2dotE21 - rE20xE21timesE1xE2 ) / 
			( rE1E2minusE1dotE2*sqrt(rE2dotE2*rMagCE21) ) );
    }
  */

  prEdgeZone[0] = -( rE1E2minusE1dotE2*rE2dotE20 - rE20xE21timesE1xE2 );
  prEdgeZone[1] = ( rE1E2minusE1dotE2*rE2dotE21 - rE20xE21timesE1xE2 );
  
  for ( i = 0; i < 2; i++ )
    {
      if ( prEdgeZone[i] < 0.0 )
	{
	  *pdC_x = 0.0;
	  *pdC_y = 0.0;
	  *pdR = 0.0;
	  return APM2D::icctNull;
	}
    }

  //////////////////////////////////////////////////
  // Compute the center and radius

  prC[0] = -( rE20xE21*pprE[1][1] ) / rDen;
  prC[1] = ( rE20xE21*pprE[1][0] ) / rDen;
  //prC[2] = sqrt((prC[0]*prC[0]) + (prC[1]*prC[1]));

  rR = ( sqrt(rE1dotE1)*rE20xE21 ) / rDen;
  
  /*
  if ( !( fabs(prC[2] - rR) < s_crEpsilonInscribedCircleCentervEERadius ) )
    {
      fprintf(strerr, "vEE: C = ( %f, %f, %f ) != %f = R\n", prC[0], prC[1], prC[2], rR);
      //ASSERT( FALSE );
    }
  */

  // Translate the answer back out away from the origin
  prC[0] += rV0_E10_x;
  prC[1] += rV0_E10_y;

  *pdC_x = dble(prC[0]);
  *pdC_y = dble(prC[1]);
  *pdR = dble(rR);
  return APM2D::icctvEE;
}

//static 
APM2D::EdgeZoneType EdgeZonevEEMPFUN(DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
				     DOUBLE dE11_x, DOUBLE dE11_y,
				     DOUBLE dE20_x, DOUBLE dE20_y, 
				     DOUBLE dE21_x, DOUBLE dE21_y,
				     DOUBLE dE30_x, DOUBLE dE30_y,
				     DOUBLE dE31_x, DOUBLE dE31_y)
{
  mp_real rV0_E10_x = dV0_E10_x;
  mp_real rV0_E10_y = dV0_E10_y; 
  mp_real rE11_x = dE11_x;
  mp_real rE11_y = dE11_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;
  mp_real rE30_x = dE30_x;
  mp_real rE30_y = dE30_y;
  mp_real rE31_x = dE31_x;
  mp_real rE31_y = dE31_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real prEdgeZone[2];
  mp_real rE1dotE1, rE2dotE2, rE1dotE2;
  mp_real rE3dotE3, rE3dotE30, rE3dotE31;
  mp_real rE1xE3, rE20xE21;
  mp_real rE1E2minusE1dotE2;
  //mp_real rDen;
  //mp_real rMagCE30, rMagCE31;
  APM2D::EdgeZoneType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = 0.0;
  ppprV[1][0][1] = 0.0;
  ppprV[1][1][0] = rE11_x - rV0_E10_x;
  ppprV[1][1][1] = rE11_y - rV0_E10_y;
  ppprV[2][0][0] = rE20_x - rV0_E10_x;
  ppprV[2][0][1] = rE20_y - rV0_E10_y;
  ppprV[2][1][0] = rE21_x - rV0_E10_x;
  ppprV[2][1][1] = rE21_y - rV0_E10_y;
  ppprV[3][0][0] = rE30_x - rV0_E10_x;
  ppprV[3][0][1] = rE30_y - rV0_E10_y;
  ppprV[3][1][0] = rE31_x - rV0_E10_x;
  ppprV[3][1][1] = rE31_y - rV0_E10_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[1][0] = rE11_x - rV0_E10_x;
  pprE[1][1] = rE11_y - rV0_E10_y;
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;
  pprE[3][0] = rE31_x - rE30_x;
  pprE[3][1] = rE31_y - rE30_y;
  
  rE1dotE1 = (pprE[1][0]*pprE[1][0]) + (pprE[1][1]*pprE[1][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);
  rE1dotE2 = (pprE[1][0]*pprE[2][0]) + (pprE[1][1]*pprE[2][1]);
  rE3dotE3 = (pprE[3][0]*pprE[3][0]) + (pprE[3][1]*pprE[3][1]);
  rE3dotE30 = (pprE[3][0]*ppprV[3][0][0]) + (pprE[3][1]*ppprV[3][0][1]);
  rE3dotE31 = (pprE[3][0]*ppprV[3][1][0]) + (pprE[3][1]*ppprV[3][1][1]);

  rE1xE3 = (pprE[1][0]*pprE[3][1]) - (pprE[1][1]*pprE[3][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);

  rE1E2minusE1dotE2 = sqrt(rE1dotE1*rE2dotE2) - rE1dotE2;

  //rDen = rE1E2minusE1dotE2;
  //ASSERT( rDen != 0.0 );

  mp_real rE20xE21timesE1xE3 = rE20xE21*rE1xE3;
  /*
  prC[0] = -( rE20xE21*pprE[1][1] ) / rDen;
  prC[1] = ( rE20xE21*pprE[1][0] ) / rDen;

  rMagCE30 = ( (prC[0] - ppprV[3][0][0])*(prC[0] - ppprV[3][0][0]) + 
	       (prC[1] - ppprV[3][0][1])*(prC[1] - ppprV[3][0][1]) );
  rMagCE31 = ( (prC[0] - ppprV[3][1][0])*(prC[0] - ppprV[3][1][0]) + 
	       (prC[1] - ppprV[3][1][1])*(prC[1] - ppprV[3][1][1]) );

  if ( rMagCE30 <= 0.0 )
    {
      prEdgeZone[0] = 0.0;
    }
  else
    {
      prEdgeZone[0] = ( ( rE1E2minusE1dotE2*rE3dotE30 - rE20xE21timesE1xE3 ) / 
			( rDen*sqrt(rE3dotE3*rMagCE30) ) );
    }
  if ( rMagCE31 <= 0.0 )
    {
      prEdgeZone[1] = 0.0;
    }
  else
    {
      prEdgeZone[1] = -( ( rE1E2minusE1dotE2*rE3dotE31 - rE20xE21timesE1xE3 ) / 
			 ( rDen*sqrt(rE3dotE3*rMagCE31) ) );
    }
  */

  prEdgeZone[0] = ( rE1E2minusE1dotE2*rE3dotE30 - rE20xE21timesE1xE3 );
  prEdgeZone[1] = -( rE1E2minusE1dotE2*rE3dotE31 - rE20xE21timesE1xE3 );

  if ( !( prEdgeZone[0] < 0.0 ) )
    {
      eResult = APM2D::eztStart;
    }
  else if ( !( prEdgeZone[1] < 0.0 ) )
    {
      eResult = APM2D::eztEnd;
    }
  else
    {
      eResult = APM2D::eztMiddle;
    }

  return eResult;
}

//static 
APM2D::CircleTestType InCirclevEE_VMPFUN(DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
					 DOUBLE dE11_x, DOUBLE dE11_y,
					 DOUBLE dE20_x, DOUBLE dE20_y, 
					 DOUBLE dE21_x, DOUBLE dE21_y,
					 DOUBLE dV3_x, DOUBLE dV3_y)
{
  mp_real rV0_E10_x = dV0_E10_x;
  mp_real rV0_E10_y = dV0_E10_y; 
  mp_real rE11_x = dE11_x;
  mp_real rE11_y = dE11_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;
  mp_real rV3_x = dV3_x;
  mp_real rV3_y = dV3_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real rE1dotE1, rE2dotE2, rE1dotE2, rV3dotV3;
  mp_real rE1xV3, rE20xE21;
  mp_real rE1E2minusE1dotE2;
  //mp_real rDen;
  mp_real rResult;
  APM2D::CircleTestType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = 0.0;
  ppprV[1][0][1] = 0.0;
  ppprV[1][1][0] = rE11_x - rV0_E10_x;
  ppprV[1][1][1] = rE11_y - rV0_E10_y;
  ppprV[2][0][0] = rE20_x - rV0_E10_x;
  ppprV[2][0][1] = rE20_y - rV0_E10_y;
  ppprV[2][1][0] = rE21_x - rV0_E10_x;
  ppprV[2][1][1] = rE21_y - rV0_E10_y;
  ppprV[3][0][0] = rV3_x - rV0_E10_x;
  ppprV[3][0][1] = rV3_y - rV0_E10_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[1][0] = rE11_x - rV0_E10_x;
  pprE[1][1] = rE11_y - rV0_E10_y;
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;

  rE1dotE1 = (pprE[1][0]*pprE[1][0]) + (pprE[1][1]*pprE[1][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);
  rE1dotE2 = (pprE[1][0]*pprE[2][0]) + (pprE[1][1]*pprE[2][1]);
  rV3dotV3 = (ppprV[3][0][0]*ppprV[3][0][0]) + (ppprV[3][0][1]*ppprV[3][0][1]);

  rE1xV3 = (pprE[1][0]*ppprV[3][0][1]) - (pprE[1][1]*ppprV[3][0][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);

  rE1E2minusE1dotE2 = sqrt(rE1dotE1*rE2dotE2) - rE1dotE2;

  //rDen = rE1E2minusE1dotE2;
  //ASSERT( rDen != 0.0 );

  rResult = ( rE1E2minusE1dotE2*rV3dotV3 - 2.0*rE20xE21*rE1xV3 );
  
  if ( rResult < 0.0 )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( rResult > 0.0 )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

//static 
APM2D::CircleTestType InCirclevEE_EMPFUN(DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
					 DOUBLE dE11_x, DOUBLE dE11_y,
					 DOUBLE dE20_x, DOUBLE dE20_y, 
					 DOUBLE dE21_x, DOUBLE dE21_y,
					 DOUBLE dE30_x, DOUBLE dE30_y,
					 DOUBLE dE31_x, DOUBLE dE31_y)
{
  mp_real rV0_E10_x = dV0_E10_x;
  mp_real rV0_E10_y = dV0_E10_y; 
  mp_real rE11_x = dE11_x;
  mp_real rE11_y = dE11_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;
  mp_real rE30_x = dE30_x;
  mp_real rE30_y = dE30_y;
  mp_real rE31_x = dE31_x;
  mp_real rE31_y = dE31_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real rE1dotE1, rE2dotE2, rE1dotE2, rE1dotE3, rE3dotE3;
  mp_real rE1xE3, rE20xE21, rE30xE31;
  mp_real rE1E2minusE1dotE2;
  //mp_real rDen;
  mp_real rResult;
  APM2D::CircleTestType eResult;
  
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = 0.0;
  ppprV[1][0][1] = 0.0;
  ppprV[1][1][0] = rE11_x - rV0_E10_x;
  ppprV[1][1][1] = rE11_y - rV0_E10_y;
  ppprV[2][0][0] = rE20_x - rV0_E10_x;
  ppprV[2][0][1] = rE20_y - rV0_E10_y;
  ppprV[2][1][0] = rE21_x - rV0_E10_x;
  ppprV[2][1][1] = rE21_y - rV0_E10_y;
  ppprV[3][0][0] = rE30_x - rV0_E10_x;
  ppprV[3][0][1] = rE30_y - rV0_E10_y;
  ppprV[3][1][0] = rE31_x - rV0_E10_x;
  ppprV[3][1][1] = rE31_y - rV0_E10_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[1][0] = rE11_x - rV0_E10_x;
  pprE[1][1] = rE11_y - rV0_E10_y;
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;
  pprE[3][0] = rE31_x - rE30_x;
  pprE[3][1] = rE31_y - rE30_y;
  
  rE1dotE1 = (pprE[1][0]*pprE[1][0]) + (pprE[1][1]*pprE[1][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);
  rE1dotE2 = (pprE[1][0]*pprE[2][0]) + (pprE[1][1]*pprE[2][1]);
  rE1dotE3 = (pprE[1][0]*pprE[3][0]) + (pprE[1][1]*pprE[3][1]);
  rE3dotE3 = (pprE[3][0]*pprE[3][0]) + (pprE[3][1]*pprE[3][1]);

  rE1xE3 = (pprE[1][0]*pprE[3][1]) - (pprE[1][1]*pprE[3][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);
  rE30xE31 = (ppprV[3][0][0]*ppprV[3][1][1]) - (ppprV[3][0][1]*ppprV[3][1][0]);

  rE1E2minusE1dotE2 = sqrt(rE1dotE1*rE2dotE2) - rE1dotE2;

  //rDen = rE3dotE3*rE1E2minusE1dotE2*rE1E2minusE1dotE2;
  //ASSERT( rDen != 0.0 );

  rResult = ( rE1E2minusE1dotE2*rE30xE31*( rE1E2minusE1dotE2*rE30xE31 
					   + 2.0*rE20xE21*rE1dotE3 )
	      - rE20xE21*rE20xE21*rE1xE3*rE1xE3 );
  
  if ( rResult < 0.0 )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( rResult > 0.0 )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

//static 
APM2D::InscribedCircleCenterType 
InscribedCircleCenterEEvMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			      DOUBLE dE00_x, DOUBLE dE00_y, 
			      DOUBLE dE01_x, DOUBLE dE01_y,
			      DOUBLE dE10_x, DOUBLE dE10_y, 
			      DOUBLE dE11_V2_x, DOUBLE dE11_V2_y)
{
  mp_real prC[3], rR;
  mp_real rE00_x = dE00_x;
  mp_real rE00_y = dE00_y; 
  mp_real rE01_x = dE01_x;
  mp_real rE01_y = dE01_y;
  mp_real rE10_x = dE10_x;
  mp_real rE10_y = dE10_y; 
  mp_real rE11_V2_x = dE11_V2_x;
  mp_real rE11_V2_y = dE11_V2_y;

  mp_real ppprV[3][2][2]; // Endpoints
  mp_real pprE[3][2];     // Edge vectors
  mp_real prEdgeZone[2];
  mp_real rDen;
  mp_real rE0dotE0, rE1dotE1, rE0dotE1;
  mp_real rE0dotE00, rE0dotE01;
  mp_real rE0xE1, rE00xE01;
  mp_real rE00dotE00, rE01dotE01;
  mp_real rE0E1minusE0dotE1, rSinTheta;
  UINT_32 i;
  
  // Translate V1 to the origin
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = rE00_x - rE11_V2_x;
  ppprV[0][0][1] = rE00_y - rE11_V2_y;
  ppprV[0][1][0] = rE01_x - rE11_V2_x;
  ppprV[0][1][1] = rE01_y - rE11_V2_y;
  ppprV[1][0][0] = rE10_x - rE11_V2_x;
  ppprV[1][0][1] = rE10_y - rE11_V2_y;
  ppprV[1][1][0] = 0.0;
  ppprV[1][1][1] = 0.0;
  ppprV[2][0][0] = 0.0;
  ppprV[2][0][1] = 0.0;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[0][0] = rE01_x - rE00_x;
  pprE[0][1] = rE01_y - rE00_y;
  pprE[1][0] = rE11_V2_x - rE10_x;
  pprE[1][1] = rE11_V2_y - rE10_y;
  
  rE0dotE0 = (pprE[0][0]*pprE[0][0]) + (pprE[0][1]*pprE[0][1]);
  rE1dotE1 = (pprE[1][0]*pprE[1][0]) + (pprE[1][1]*pprE[1][1]);
  rE0dotE1 = (pprE[0][0]*pprE[1][0]) + (pprE[0][1]*pprE[1][1]);
  rE0dotE00 = (pprE[0][0]*ppprV[0][0][0]) + (pprE[0][1]*ppprV[0][0][1]);
  rE0dotE01 = (pprE[0][0]*ppprV[0][1][0]) + (pprE[0][1]*ppprV[0][1][1]);

  rE0xE1 = (pprE[0][0]*pprE[1][1]) - (pprE[0][1]*pprE[1][0]);
  rE00xE01 = (ppprV[0][0][0]*ppprV[0][1][1]) - (ppprV[0][0][1]*ppprV[0][1][0]);

  rE00dotE00 = (ppprV[0][0][0]*ppprV[0][0][0]) + (ppprV[0][0][1]*ppprV[0][0][1]);
  rE01dotE01 = (ppprV[0][1][0]*ppprV[0][1][0]) + (ppprV[0][1][1]*ppprV[0][1][1]);

  rE0E1minusE0dotE1 = sqrt(rE0dotE0*rE1dotE1) - rE0dotE1;

  //rSinTheta = rE0E1minusE0dotE1/sqrt(rE0dotE0*rE1dotE1);
  rSinTheta = rE0xE1;
  if ( rSinTheta == 0.0 &&
       rE0dotE1 > 0.0 )
    {
      // Colinear
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  //rSinTheta = rE00xE01/sqrt(rE00dotE00*rE01dotE01);
  rSinTheta = SinThetaMPFUN(rE11_V2_x, rE11_V2_y, rE00_x, rE00_y, rE01_x, rE01_y);
  /*
  rSinThetaTemp = SinThetaMPFUN(rE00_x, rE00_y, rE01_x, rE01_y, rE11_V2_x, rE11_V2_y);
  if ( rSinThetaTemp > rSinTheta )
    {
      rSinTheta = rSinThetaTemp;
    }
  */
  if ( rSinTheta <= 0.0 )
    {
      // V2 is on E0 and it is not one of the endpoints
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }
  
  rDen = rE0E1minusE0dotE1;
  ASSERT( rDen != 0.0 );

  //////////////////////////////////////////////////
  // Check the Edge Zone

  mp_real rE00xE01timesE0xE1 = rE00xE01*rE0xE1;
  /*
  rMagCE00 = ( (prC[0] - ppprV[0][0][0])*(prC[0] - ppprV[0][0][0]) + 
	       (prC[1] - ppprV[0][0][1])*(prC[1] - ppprV[0][0][1]) );
  rMagCE01 = ( (prC[0] - ppprV[0][1][0])*(prC[0] - ppprV[0][1][0]) + 
	       (prC[1] - ppprV[0][1][1])*(prC[1] - ppprV[0][1][1]) );

  if ( rMagCE00 <= 0.0 )
    {
      prEdgeZone[0] = 0.0;
    }
  else
    {
      prEdgeZone[0] = -( ( rE0E1minusE0dotE1*rE0dotE00 + rE00xE01timesE0xE1 ) / 
			 ( rE0E1minusE0dotE1*sqrt(rE0dotE0*rMagCE00) ) ); 
    }
  if ( rMagCE01 <= 0.0 )
    {
      prEdgeZone[1] = 0.0;
    }
  else
    {
      prEdgeZone[1] = ( ( rE0E1minusE0dotE1*rE0dotE01 + rE00xE01timesE0xE1 ) / 
			( rE0E1minusE0dotE1*sqrt(rE0dotE0*rMagCE01) ) ); 
    }
  */

  prEdgeZone[0] = -( rE0E1minusE0dotE1*rE0dotE00 + rE00xE01timesE0xE1 );
  prEdgeZone[1] = ( rE0E1minusE0dotE1*rE0dotE01 + rE00xE01timesE0xE1 );

  for ( i = 0; i < 2; i++ )
    {
      if ( prEdgeZone[i] < 0.0 )
	{
	  *pdC_x = 0.0;
	  *pdC_y = 0.0;
	  *pdR = 0.0;
	  return APM2D::icctNull;
	}
    }

  //////////////////////////////////////////////////
  // Compute the center and radius

  prC[0] = -( rE00xE01*pprE[1][1] ) / rDen;
  prC[1] = ( rE00xE01*pprE[1][0] ) / rDen;
  //prC[2] = sqrt((prC[0]*prC[0]) + (prC[1]*prC[1]));

  rR = ( sqrt(rE1dotE1)*rE00xE01 ) / rDen;
  
  /*
  if ( !( fabs(prC[2] - rR) < s_crEpsilonInscribedCircleCenterEEvRadius ) )
    {
      fprintf(strerr, "EEv: C = ( %f, %f, %f ) != %f = R\n", prC[0], prC[1], prC[2], rR);
      //ASSERT( FALSE );
    }
  */

  // Translate the answer back out away from the origin
  prC[0] += rE11_V2_x;
  prC[1] += rE11_V2_y;

  *pdC_x = dble(prC[0]);
  *pdC_y = dble(prC[1]);
  *pdR = dble(rR);
  return APM2D::icctEEv;
}

//static 
APM2D::EdgeZoneType EdgeZoneEEvMPFUN(DOUBLE dE00_x, DOUBLE dE00_y, 
				     DOUBLE dE01_x, DOUBLE dE01_y,
				     DOUBLE dE10_x, DOUBLE dE10_y, 
				     DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
				     DOUBLE dE30_x, DOUBLE dE30_y,
				     DOUBLE dE31_x, DOUBLE dE31_y)
{
  mp_real rE00_x = dE00_x;
  mp_real rE00_y = dE00_y; 
  mp_real rE01_x = dE01_x;
  mp_real rE01_y = dE01_y;
  mp_real rE10_x = dE10_x;
  mp_real rE10_y = dE10_y; 
  mp_real rE11_V2_x = dE11_V2_x;
  mp_real rE11_V2_y = dE11_V2_y;
  mp_real rE30_x = dE30_x;
  mp_real rE30_y = dE30_y;
  mp_real rE31_x = dE31_x;
  mp_real rE31_y = dE31_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real prEdgeZone[2];
  mp_real rE0dotE0, rE1dotE1, rE0dotE1;
  mp_real rE3dotE30, rE3dotE31;
  mp_real rE1xE3, rE00xE01;
  mp_real rE0E1minusE0dotE1;
  //mp_real rE3dotE3, rMagCE30, rMagCE31;
  //mp_real rDen;
  APM2D::EdgeZoneType eResult;
  
  // Translate V1 to the origin
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = rE00_x - rE11_V2_x;
  ppprV[0][0][1] = rE00_y - rE11_V2_y;
  ppprV[0][1][0] = rE01_x - rE11_V2_x;
  ppprV[0][1][1] = rE01_y - rE11_V2_y;
  ppprV[1][0][0] = rE10_x - rE11_V2_x;
  ppprV[1][0][1] = rE10_y - rE11_V2_y;
  ppprV[1][1][0] = 0.0;
  ppprV[1][1][1] = 0.0;
  ppprV[2][0][0] = 0.0;
  ppprV[2][0][1] = 0.0;
  ppprV[3][0][0] = rE30_x - rE11_V2_x;
  ppprV[3][0][1] = rE30_y - rE11_V2_y;
  ppprV[3][1][0] = rE31_x - rE11_V2_x;
  ppprV[3][1][1] = rE31_y - rE11_V2_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[0][0] = rE01_x - rE00_x;
  pprE[0][1] = rE01_y - rE00_y;
  pprE[1][0] = rE11_V2_x - rE10_x;
  pprE[1][1] = rE11_V2_y - rE10_y;
  pprE[3][0] = rE31_x - rE30_x;
  pprE[3][1] = rE31_y - rE30_y;
  
  rE0dotE0 = (pprE[0][0]*pprE[0][0]) + (pprE[0][1]*pprE[0][1]);
  rE1dotE1 = (pprE[1][0]*pprE[1][0]) + (pprE[1][1]*pprE[1][1]);
  rE0dotE1 = (pprE[0][0]*pprE[1][0]) + (pprE[0][1]*pprE[1][1]);
  //rE3dotE3 = (pprE[3][0]*pprE[3][0]) + (pprE[3][1]*pprE[3][1]);

  rE3dotE30 = (pprE[3][0]*ppprV[3][0][0]) + (pprE[3][1]*ppprV[3][0][1]);
  rE3dotE31 = (pprE[3][0]*ppprV[3][1][0]) + (pprE[3][1]*ppprV[3][1][1]);

  rE1xE3 = (pprE[1][0]*pprE[3][1]) - (pprE[1][1]*pprE[3][0]);
  rE00xE01 = (ppprV[0][0][0]*ppprV[0][1][1]) - (ppprV[0][0][1]*ppprV[0][1][0]);

  rE0E1minusE0dotE1 = sqrt(rE0dotE0*rE1dotE1) - rE0dotE1;

  //rDen = rE0E1minusE0dotE1;
  //ASSERT( rDen != 0.0 );

  mp_real rE00xE01timesE1xE3 = rE00xE01*rE1xE3;
  /*
  prC[0] = -( rE00xE01*pprE[1][1] ) / rDen;
  prC[1] = ( rE00xE01*pprE[1][0] ) / rDen;

  rMagCE30 = ( (prC[0] - ppprV[3][0][0])*(prC[0] - ppprV[3][0][0]) + 
	       (prC[1] - ppprV[3][0][1])*(prC[1] - ppprV[3][0][1]) );
  rMagCE31 = ( (prC[0] - ppprV[3][1][0])*(prC[0] - ppprV[3][1][0]) + 
	       (prC[1] - ppprV[3][1][1])*(prC[1] - ppprV[3][1][1]) );

  if ( rMagCE30 <= 0.0 )
    {
      prEdgeZone[0] = 0.0;
    }
  else
    {
      prEdgeZone[0] = ( ( rE0E1minusE0dotE1*rE3dotE30 - rE00xE01timesE1xE3 ) / 
			( rDen*sqrt(rE3dotE3*rMagCE30) ) );
    }
  if ( rMagCE31 <= 0.0 )
    {
      prEdgeZone[1] = 0.0;
    }
  else
    {
      prEdgeZone[1] = -( ( rE0E1minusE0dotE1*rE3dotE31 - rE00xE01timesE1xE3 ) / 
			 ( rDen*sqrt(rE3dotE3*rMagCE31) ) );
    }
  */

  prEdgeZone[0] = ( rE0E1minusE0dotE1*rE3dotE30 - rE00xE01timesE1xE3 );
  prEdgeZone[1] = -( rE0E1minusE0dotE1*rE3dotE31 - rE00xE01timesE1xE3 );

  if ( !( prEdgeZone[0] < 0.0 ) )
    {
      eResult = APM2D::eztStart;
    }
  else if ( !( prEdgeZone[1] < 0.0 ) )
    {
      eResult = APM2D::eztEnd;
    }
  else
    {
      eResult = APM2D::eztMiddle;
    }

  return eResult;
}

//static 
APM2D::CircleTestType InCircleEEv_VMPFUN(DOUBLE dE00_x, DOUBLE dE00_y, 
					 DOUBLE dE01_x, DOUBLE dE01_y,
					 DOUBLE dE10_x, DOUBLE dE10_y, 
					 DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
					 DOUBLE dV3_x, DOUBLE dV3_y)
{
  mp_real rE00_x = dE00_x;
  mp_real rE00_y = dE00_y; 
  mp_real rE01_x = dE01_x;
  mp_real rE01_y = dE01_y;
  mp_real rE10_x = dE10_x;
  mp_real rE10_y = dE10_y; 
  mp_real rE11_V2_x = dE11_V2_x;
  mp_real rE11_V2_y = dE11_V2_y;
  mp_real rV3_x = dV3_x;
  mp_real rV3_y = dV3_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real rE0dotE0, rE1dotE1, rE0dotE1, rV3dotV3;
  mp_real rE1xV3, rE00xE01;
  mp_real rE0E1minusE0dotE1;
  //mp_real rDen;
  mp_real rResult;
  APM2D::CircleTestType eResult;
  
  // Translate V1 to the origin
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = rE00_x - rE11_V2_x;
  ppprV[0][0][1] = rE00_y - rE11_V2_y;
  ppprV[0][1][0] = rE01_x - rE11_V2_x;
  ppprV[0][1][1] = rE01_y - rE11_V2_y;
  ppprV[1][0][0] = rE10_x - rE11_V2_x;
  ppprV[1][0][1] = rE10_y - rE11_V2_y;
  ppprV[1][1][0] = 0.0;
  ppprV[1][1][1] = 0.0;
  ppprV[2][0][0] = 0.0;
  ppprV[2][0][1] = 0.0;
  ppprV[3][0][0] = rV3_x - rE11_V2_x;
  ppprV[3][0][1] = rV3_y - rE11_V2_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[0][0] = rE01_x - rE00_x;
  pprE[0][1] = rE01_y - rE00_y;
  pprE[1][0] = rE11_V2_x - rE10_x;
  pprE[1][1] = rE11_V2_y - rE10_y;
  
  rE0dotE0 = (pprE[0][0]*pprE[0][0]) + (pprE[0][1]*pprE[0][1]);
  rE1dotE1 = (pprE[1][0]*pprE[1][0]) + (pprE[1][1]*pprE[1][1]);
  rE0dotE1 = (pprE[0][0]*pprE[1][0]) + (pprE[0][1]*pprE[1][1]);
  rV3dotV3 = (ppprV[3][0][0]*ppprV[3][0][0]) + (ppprV[3][0][1]*ppprV[3][0][1]);

  rE1xV3 = (pprE[1][0]*ppprV[3][0][1]) - (pprE[1][1]*ppprV[3][0][0]);
  rE00xE01 = (ppprV[0][0][0]*ppprV[0][1][1]) - (ppprV[0][0][1]*ppprV[0][1][0]);

  rE0E1minusE0dotE1 = sqrt(rE0dotE0*rE1dotE1) - rE0dotE1;

  //rDen = rE0E1minusE0dotE1;
  //ASSERT( rDen != 0.0 );

  rResult = ( rE0E1minusE0dotE1*rV3dotV3 - 2.0*rE00xE01*rE1xV3 );
  
  if ( rResult < 0.0 )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( rResult > 0.0 )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

//static 
APM2D::CircleTestType InCircleEEv_EMPFUN(DOUBLE dE00_x, DOUBLE dE00_y, 
					 DOUBLE dE01_x, DOUBLE dE01_y,
					 DOUBLE dE10_x, DOUBLE dE10_y, 
					 DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
					 DOUBLE dE30_x, DOUBLE dE30_y,
					 DOUBLE dE31_x, DOUBLE dE31_y)
{
  mp_real rE00_x = dE00_x;
  mp_real rE00_y = dE00_y; 
  mp_real rE01_x = dE01_x;
  mp_real rE01_y = dE01_y;
  mp_real rE10_x = dE10_x;
  mp_real rE10_y = dE10_y; 
  mp_real rE11_V2_x = dE11_V2_x;
  mp_real rE11_V2_y = dE11_V2_y;
  mp_real rE30_x = dE30_x;
  mp_real rE30_y = dE30_y;
  mp_real rE31_x = dE31_x;
  mp_real rE31_y = dE31_y;
  
  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real rE0dotE0, rE1dotE1, rE0dotE1, rE1dotE3;
  //mp_real rE3dotE3;
  mp_real rE1xE3, rE00xE01, rE30xE31;
  mp_real rE0E1minusE0dotE1;
  //mp_real rDen;
  mp_real rResult;
  APM2D::CircleTestType eResult;
  
  // Translate V1 to the origin
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = rE00_x - rE11_V2_x;
  ppprV[0][0][1] = rE00_y - rE11_V2_y;
  ppprV[0][1][0] = rE01_x - rE11_V2_x;
  ppprV[0][1][1] = rE01_y - rE11_V2_y;
  ppprV[1][0][0] = rE10_x - rE11_V2_x;
  ppprV[1][0][1] = rE10_y - rE11_V2_y;
  ppprV[1][1][0] = 0.0;
  ppprV[1][1][1] = 0.0;
  ppprV[2][0][0] = 0.0;
  ppprV[2][0][1] = 0.0;
  ppprV[3][0][0] = rE30_x - rE11_V2_x;
  ppprV[3][0][1] = rE30_y - rE11_V2_y;
  ppprV[3][1][0] = rE31_x - rE11_V2_x;
  ppprV[3][1][1] = rE31_y - rE11_V2_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[0][0] = rE01_x - rE00_x;
  pprE[0][1] = rE01_y - rE00_y;
  pprE[1][0] = rE11_V2_x - rE10_x;
  pprE[1][1] = rE11_V2_y - rE10_y;
  pprE[3][0] = rE31_x - rE30_x;
  pprE[3][1] = rE31_y - rE30_y;
  
  rE0dotE0 = (pprE[0][0]*pprE[0][0]) + (pprE[0][1]*pprE[0][1]);
  rE1dotE1 = (pprE[1][0]*pprE[1][0]) + (pprE[1][1]*pprE[1][1]);
  rE0dotE1 = (pprE[0][0]*pprE[1][0]) + (pprE[0][1]*pprE[1][1]);
  rE1dotE3 = (pprE[1][0]*pprE[3][0]) + (pprE[1][1]*pprE[3][1]);
  //rE3dotE3 = (pprE[3][0]*pprE[3][0]) + (pprE[3][1]*pprE[3][1]);

  rE1xE3 = (pprE[1][0]*pprE[3][1]) - (pprE[1][1]*pprE[3][0]);
  rE00xE01 = (ppprV[0][0][0]*ppprV[0][1][1]) - (ppprV[0][0][1]*ppprV[0][1][0]);
  rE30xE31 = (ppprV[3][0][0]*ppprV[3][1][1]) - (ppprV[3][0][1]*ppprV[3][1][0]);

  rE0E1minusE0dotE1 = sqrt(rE0dotE0*rE1dotE1) - rE0dotE1;

  //rDen = rE3dotE3*rE0E1minusE0dotE1*rE0E1minusE0dotE1;
  //ASSERT( rDen != 0.0 );

  rResult = ( rE0E1minusE0dotE1*rE30xE31*( rE0E1minusE0dotE1*rE30xE31
					   + 2.0*rE00xE01*rE1dotE3 )
	      - rE00xE01*rE00xE01*rE1xE3*rE1xE3 );
  
  if ( rResult < 0.0 )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( rResult > 0.0 )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

//static 
APM2D::InscribedCircleCenterType 
InscribedCircleCenterEEEMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			      DOUBLE dE00_x, DOUBLE dE00_y, 
			      DOUBLE dE01_x, DOUBLE dE01_y,
			      DOUBLE dE10_x, DOUBLE dE10_y, 
			      DOUBLE dE11_x, DOUBLE dE11_y,
			      DOUBLE dE20_x, DOUBLE dE20_y, 
			      DOUBLE dE21_x, DOUBLE dE21_y)
{
  mp_real rE00_x = dE00_x;
  mp_real rE00_y = dE00_y; 
  mp_real rE01_x = dE01_x;
  mp_real rE01_y = dE01_y;
  mp_real rE10_x = dE10_x;
  mp_real rE10_y = dE10_y; 
  mp_real rE11_x = dE11_x;
  mp_real rE11_y = dE11_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;
  
  mp_real prC[3], rR;
  mp_real ppprV[3][2][2]; // Endpoints
  mp_real pprE[3][2];     // Edge vectors
  mp_real pprEdgeZone[3][2];
  mp_real rDen;
  mp_real rE0dotE0, rE1dotE1, rE2dotE2, rE0dotE1, rE0dotE2, rE1dotE2;
  mp_real rE0mag, rE1mag, rE2mag;
  mp_real rE0dotE00, rE0dotE01, rE1dotE10, rE1dotE11, rE2dotE20, rE2dotE21;
  mp_real rE0xE1, rE1xE2, rE2xE0, rE00xE01, rE20xE21;
  mp_real rE0E1minusE0dotE1, rE1E2minusE1dotE2, rE0E2minusE0dotE2;
  mp_real rE0dotC, rE1dotC, rE2dotC;
  mp_real rMagCE00, rMagCE01, rMagCE10, rMagCE11, rMagCE20, rMagCE21;
  mp_real pprProjection[3][2];
  mp_real rCCWCheck;
  mp_real rSinTheta;
  //DOUBLE dCCW;
  UINT_32 i, j;
  
  // Translate V1 to the origin
  ppprV[0][0][0] = rE00_x - rE10_x;
  ppprV[0][0][1] = rE00_y - rE10_y;
  ppprV[0][1][0] = rE01_x - rE10_x;
  ppprV[0][1][1] = rE01_y - rE10_y;
  ppprV[1][0][0] = 0.0;
  ppprV[1][0][1] = 0.0;
  ppprV[1][1][0] = rE11_x - rE10_x;
  ppprV[1][1][1] = rE11_y - rE10_y;
  ppprV[2][0][0] = rE20_x - rE10_x;
  ppprV[2][0][1] = rE20_y - rE10_y;
  ppprV[2][1][0] = rE21_x - rE10_x;
  ppprV[2][1][1] = rE21_y - rE10_y;
  
  pprE[0][0] = rE01_x - rE00_x;
  pprE[0][1] = rE01_y - rE00_y;
  pprE[1][0] = rE11_x - rE10_x;
  pprE[1][1] = rE11_y - rE10_y;
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;
  

  rE0dotE0 = (pprE[0][0]*pprE[0][0]) + (pprE[0][1]*pprE[0][1]);
  rE1dotE1 = (pprE[1][0]*pprE[1][0]) + (pprE[1][1]*pprE[1][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);

  rE0mag = sqrt(rE0dotE0);
  rE1mag = sqrt(rE1dotE1);
  rE2mag = sqrt(rE2dotE2);

  rE0dotE1 = (pprE[0][0]*pprE[1][0]) + (pprE[0][1]*pprE[1][1]);
  rE1dotE2 = (pprE[1][0]*pprE[2][0]) + (pprE[1][1]*pprE[2][1]);
  rE0dotE2 = (pprE[0][0]*pprE[2][0]) + (pprE[0][1]*pprE[2][1]);

  rE0dotE00 = (pprE[0][0]*ppprV[0][0][0]) + (pprE[0][1]*ppprV[0][0][1]);
  rE0dotE01 = (pprE[0][0]*ppprV[0][1][0]) + (pprE[0][1]*ppprV[0][1][1]);
  rE1dotE10 = (pprE[1][0]*ppprV[1][0][0]) + (pprE[1][1]*ppprV[1][0][1]);
  rE1dotE11 = (pprE[1][0]*ppprV[1][1][0]) + (pprE[1][1]*ppprV[1][1][1]);
  rE2dotE20 = (pprE[2][0]*ppprV[2][0][0]) + (pprE[2][1]*ppprV[2][0][1]);
  rE2dotE21 = (pprE[2][0]*ppprV[2][1][0]) + (pprE[2][1]*ppprV[2][1][1]);

  rE0xE1 = (pprE[0][0]*pprE[1][1]) - (pprE[0][1]*pprE[1][0]);
  rE1xE2 = (pprE[1][0]*pprE[2][1]) - (pprE[1][1]*pprE[2][0]);
  rE2xE0 = (pprE[2][0]*pprE[0][1]) - (pprE[2][1]*pprE[0][0]);

  rE00xE01 = (ppprV[0][0][0]*ppprV[0][1][1]) - (ppprV[0][0][1]*ppprV[0][1][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);

  rE0E1minusE0dotE1 = sqrt(rE0dotE0*rE1dotE1) - rE0dotE1;
  rE1E2minusE1dotE2 = sqrt(rE1dotE1*rE2dotE2) - rE1dotE2;
  rE0E2minusE0dotE2 = sqrt(rE0dotE0*rE2dotE2) - rE0dotE2;

  rDen = rE0mag*rE1xE2 + rE1mag*rE2xE0 + rE2mag*rE0xE1;
  rSinTheta = rDen;
  if ( rSinTheta == 0.0 )
    {
      // Colinear
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  //////////////////////////////////////////////////
  // Check the Edge Zones

  rMagCE00 = ( (prC[0] - ppprV[0][0][0])*(prC[0] - ppprV[0][0][0]) + 
	       (prC[1] - ppprV[0][0][1])*(prC[1] - ppprV[0][0][1]) );
  rMagCE01 = ( (prC[0] - ppprV[0][1][0])*(prC[0] - ppprV[0][1][0]) + 
	       (prC[1] - ppprV[0][1][1])*(prC[1] - ppprV[0][1][1]) );
  rMagCE10 = ( (prC[0] - ppprV[1][0][0])*(prC[0] - ppprV[1][0][0]) + 
	       (prC[1] - ppprV[1][0][1])*(prC[1] - ppprV[1][0][1]) );
  rMagCE11 = ( (prC[0] - ppprV[1][1][0])*(prC[0] - ppprV[1][1][0]) + 
	       (prC[1] - ppprV[1][1][1])*(prC[1] - ppprV[1][1][1]) );
  rMagCE20 = ( (prC[0] - ppprV[2][0][0])*(prC[0] - ppprV[2][0][0]) + 
	       (prC[1] - ppprV[2][0][1])*(prC[1] - ppprV[2][0][1]) );
  rMagCE21 = ( (prC[0] - ppprV[2][1][0])*(prC[0] - ppprV[2][1][0]) + 
	       (prC[1] - ppprV[2][1][1])*(prC[1] - ppprV[2][1][1]) );
  /*
  if ( rMagCE00 <= 0.0 )
    {
      pprEdgeZone[0][0] = 0.0;
    }
  else
    {
      pprEdgeZone[0][0] = -( ( rDen*rE0dotE00 
			       - (rE1mag*rE0dotE2 - rE2mag*rE0dotE1)*rE00xE01
			       + rE0mag*rE0E1minusE0dotE1*rE20xE21 ) /
			     ( rDen*sqrt(rE0dotE0*rMagCE00) ) );
    }
  if ( rMagCE01 <= 0.0 )
    {
      pprEdgeZone[0][1] = 0.0;
    }
  else
    {
      pprEdgeZone[0][1] = ( ( rDen*rE0dotE01 
			      - (rE1mag*rE0dotE2 - rE2mag*rE0dotE1)*rE00xE01
			      + rE0mag*rE0E1minusE0dotE1*rE20xE21 ) /
			    ( rDen*sqrt(rE0dotE0*rMagCE01) ) );
    }
  if ( rMagCE10 <= 0.0 )
    {
      pprEdgeZone[1][0] = 0.0;
    }
  else
    {
      pprEdgeZone[1][0] = -( ( rDen*rE1dotE10 
			       + rE1mag*rE1E2minusE1dotE2*rE00xE01
			       - rE1mag*rE0E1minusE0dotE1*rE20xE21 ) /
			     ( rDen*sqrt(rE1dotE1*rMagCE10) ) );
    }
  if ( rMagCE11 <= 0.0 )
    {
      pprEdgeZone[1][1] = 0.0;
    }
  else
    {
      pprEdgeZone[1][1] = ( ( rDen*rE1dotE11 
			      + rE1mag*rE1E2minusE1dotE2*rE00xE01
			      - rE1mag*rE0E1minusE0dotE1*rE20xE21 ) /
			    ( rDen*sqrt(rE1dotE1*rMagCE11) ) );
    }
  if ( rMagCE20 <= 0.0 )
    {
      pprEdgeZone[2][0] = 0.0;
    }
  else
    {
      pprEdgeZone[2][0] = -( ( rDen*rE2dotE20 
			       - rE2mag*rE1E2minusE1dotE2*rE00xE01 
			       - (rE0mag*rE1dotE2 - rE1mag*rE0dotE2)*rE20xE21 ) /
			     ( rDen*sqrt(rE2dotE2*rMagCE20) ) );
    }
  if ( rMagCE21 <= 0.0 )
    {
      pprEdgeZone[2][1] = 0.0;
    }
  else
    {
      pprEdgeZone[2][1] = ( ( rDen*rE2dotE21 
			      - rE2mag*rE1E2minusE1dotE2*rE00xE01 
			      - (rE0mag*rE1dotE2 - rE1mag*rE0dotE2)*rE20xE21 ) /
			    ( rDen*sqrt(rE2dotE2*rMagCE21) ) );
    }
  */

  pprEdgeZone[0][0] = -( rDen*rE0dotE00 
			 - (rE1mag*rE0dotE2 - rE2mag*rE0dotE1)*rE00xE01
			 + rE0mag*rE0E1minusE0dotE1*rE20xE21 );
  pprEdgeZone[0][1] = ( rDen*rE0dotE01 
			- (rE1mag*rE0dotE2 - rE2mag*rE0dotE1)*rE00xE01
			+ rE0mag*rE0E1minusE0dotE1*rE20xE21 );
  pprEdgeZone[1][0] = -( rDen*rE1dotE10 
			 + rE1mag*rE1E2minusE1dotE2*rE00xE01
			 - rE1mag*rE0E1minusE0dotE1*rE20xE21 );
  pprEdgeZone[1][1] = ( rDen*rE1dotE11 
			+ rE1mag*rE1E2minusE1dotE2*rE00xE01
			- rE1mag*rE0E1minusE0dotE1*rE20xE21 );
  pprEdgeZone[2][0] = -( rDen*rE2dotE20 
			 - rE2mag*rE1E2minusE1dotE2*rE00xE01 
			 - (rE0mag*rE1dotE2 - rE1mag*rE0dotE2)*rE20xE21 ); 
  pprEdgeZone[2][1] = ( rDen*rE2dotE21 
			- rE2mag*rE1E2minusE1dotE2*rE00xE01 
			- (rE0mag*rE1dotE2 - rE1mag*rE0dotE2)*rE20xE21 );
  if ( rDen < 0.0 )
    {
      for ( i = 0; i < 3; i++ )
	{
	  for ( j = 0; j < 2; j++ )
	    {
	      pprEdgeZone[i][j] = -pprEdgeZone[i][j];
	    }
	}
    }

  for ( i = 0; i < 3; i++ )
    {
      for ( j = 0; j < 2; j++ )
	{
	  if ( pprEdgeZone[i][j] < 0.0 )
	    {
	      *pdC_x = 0.0;
	      *pdC_y = 0.0;
	      *pdR = 0.0;
	      return APM2D::icctNull;
	    }
	}
    }

  //////////////////////////////////////////////////
  // Compute the center and radius

  prC[0] = ( (rE1mag*pprE[2][0] - rE2mag*pprE[1][0])*rE00xE01
	     + (rE0mag*pprE[1][0] - rE1mag*pprE[0][0])*rE20xE21 ) / rDen;
  prC[1] = ( (rE1mag*pprE[2][1] - rE2mag*pprE[1][1])*rE00xE01
	     + (rE0mag*pprE[1][1] - rE1mag*pprE[0][1])*rE20xE21 ) / rDen;

  rR = ( rE0xE1*rE20xE21 + rE1xE2*rE00xE01 ) / rDen;
  
  //////////////////////////////////////////////////
  // Check if the projection of the center on to the sites is in CCW orrer

  rE0dotC = (pprE[0][0]*prC[0]) + (pprE[0][1]*prC[1]);
  rE1dotC = (pprE[1][0]*prC[0]) + (pprE[1][1]*prC[1]);
  rE2dotC = (pprE[2][0]*prC[0]) + (pprE[2][1]*prC[1]);

  pprProjection[0][0] = (-rE00xE01*(-pprE[0][1]) + rE0dotC*pprE[0][0]) / rE0dotE0;
  pprProjection[0][1] = (-rE00xE01*(pprE[0][0]) + rE0dotC*pprE[0][1]) / rE0dotE0;

  pprProjection[1][0] = (rE1dotC*pprE[1][0]) / rE1dotE1;
  pprProjection[1][1] = (rE1dotC*pprE[1][1]) / rE1dotE1;

  pprProjection[2][0] = (-rE20xE21*(-pprE[2][1]) + rE2dotC*pprE[2][0]) / rE2dotE2;
  pprProjection[2][1] = (-rE20xE21*(pprE[2][0]) + rE2dotC*pprE[2][1]) / rE2dotE2;

  rCCWCheck = ( ( (pprProjection[2][0] - pprProjection[1][0]) *
		  (pprProjection[0][1] - pprProjection[1][1]) ) -
		( (pprProjection[2][1] - pprProjection[1][1]) *
		  (pprProjection[0][0] - pprProjection[1][0]) ) );

  //fprintf(strerr, "  EEE: CCW:  Expander %e ?= %e Projection\n", rCCW, rCCWCheck);

  if ( rCCWCheck < 0.0 )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdR = 0.0;
      return APM2D::icctNull;
    }

  // Translate the answer back out away from the origin
  prC[0] += rE10_x;
  prC[1] += rE10_y;

  *pdC_x = dble(prC[0]);
  *pdC_y = dble(prC[1]);
  *pdR = dble(rR);
  return APM2D::icctEEE;
}

//static 
APM2D::EdgeZoneType EdgeZoneEEEMPFUN(DOUBLE dE00_x, DOUBLE dE00_y, 
				     DOUBLE dE01_x, DOUBLE dE01_y,
				     DOUBLE dE10_x, DOUBLE dE10_y, 
				     DOUBLE dE11_x, DOUBLE dE11_y,
				     DOUBLE dE20_x, DOUBLE dE20_y, 
				     DOUBLE dE21_x, DOUBLE dE21_y,
				     DOUBLE dE30_x, DOUBLE dE30_y,
				     DOUBLE dE31_x, DOUBLE dE31_y)
{
  mp_real rE00_x = dE00_x;
  mp_real rE00_y = dE00_y; 
  mp_real rE01_x = dE01_x;
  mp_real rE01_y = dE01_y;
  mp_real rE10_x = dE10_x;
  mp_real rE10_y = dE10_y; 
  mp_real rE11_x = dE11_x;
  mp_real rE11_y = dE11_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;
  mp_real rE30_x = dE30_x;
  mp_real rE30_y = dE30_y;
  mp_real rE31_x = dE31_x;
  mp_real rE31_y = dE31_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real prEdgeZone[2];
  mp_real rE0dotE0, rE1dotE1, rE2dotE2, rE0dotE3, rE1dotE3, rE2dotE3;
  mp_real rE0mag, rE1mag, rE2mag;
  mp_real rE3dotE30, rE3dotE31;
  mp_real rE0xE1, rE1xE2, rE2xE0, rE00xE01, rE20xE21;
  mp_real rE3dotE3;
  mp_real rDen;
  mp_real rSinTheta;
  APM2D::EdgeZoneType eResult;
  
  // Translate V1 to the origin
  ppprV[0][0][0] = rE00_x - rE10_x;
  ppprV[0][0][1] = rE00_y - rE10_y;
  ppprV[0][1][0] = rE01_x - rE10_x;
  ppprV[0][1][1] = rE01_y - rE10_y;
  ppprV[1][0][0] = 0.0;
  ppprV[1][0][1] = 0.0;
  ppprV[1][1][0] = rE11_x - rE10_x;
  ppprV[1][1][1] = rE11_y - rE10_y;
  ppprV[2][0][0] = rE20_x - rE10_x;
  ppprV[2][0][1] = rE20_y - rE10_y;
  ppprV[2][1][0] = rE21_x - rE10_x;
  ppprV[2][1][1] = rE21_y - rE10_y;
  ppprV[3][0][0] = rE30_x - rE10_x;
  ppprV[3][0][1] = rE30_y - rE10_y;
  ppprV[3][1][0] = rE31_x - rE10_x;
  ppprV[3][1][1] = rE31_y - rE10_y;
  
  pprE[0][0] = rE01_x - rE00_x;
  pprE[0][1] = rE01_y - rE00_y;
  pprE[1][0] = rE11_x - rE10_x;
  pprE[1][1] = rE11_y - rE10_y;
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;
  pprE[3][0] = rE31_x - rE30_x;
  pprE[3][1] = rE31_y - rE30_y;
  
  rE0dotE0 = (pprE[0][0]*pprE[0][0]) + (pprE[0][1]*pprE[0][1]);
  rE1dotE1 = (pprE[1][0]*pprE[1][0]) + (pprE[1][1]*pprE[1][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);
  rE3dotE3 = (pprE[3][0]*pprE[3][0]) + (pprE[3][1]*pprE[3][1]);

  rE0mag = sqrt(rE0dotE0);
  rE1mag = sqrt(rE1dotE1);
  rE2mag = sqrt(rE2dotE2);

  rE0dotE3 = (pprE[0][0]*pprE[3][0]) + (pprE[0][1]*pprE[3][1]);
  rE1dotE3 = (pprE[1][0]*pprE[3][0]) + (pprE[1][1]*pprE[3][1]);
  rE2dotE3 = (pprE[2][0]*pprE[3][0]) + (pprE[2][1]*pprE[3][1]);

  rE3dotE30 = (pprE[3][0]*ppprV[3][0][0]) + (pprE[3][1]*ppprV[3][0][1]);
  rE3dotE31 = (pprE[3][0]*ppprV[3][1][0]) + (pprE[3][1]*ppprV[3][1][1]);

  rE0xE1 = (pprE[0][0]*pprE[1][1]) - (pprE[0][1]*pprE[1][0]);
  rE1xE2 = (pprE[1][0]*pprE[2][1]) - (pprE[1][1]*pprE[2][0]);
  rE2xE0 = (pprE[2][0]*pprE[0][1]) - (pprE[2][1]*pprE[0][0]);

  rE00xE01 = (ppprV[0][0][0]*ppprV[0][1][1]) - (ppprV[0][0][1]*ppprV[0][1][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);

  rDen = rE0mag*rE1xE2 + rE1mag*rE2xE0 + rE2mag*rE0xE1;
  rSinTheta = rDen;
  ASSERT( !( rSinTheta == 0.0 ) );

  mp_real rCoef1 = ( - (rE1mag*rE2dotE3 - rE2mag*rE1dotE3)*rE00xE01
		     - (rE0mag*rE1dotE3 - rE1mag*rE0dotE3)*rE20xE21 );
  /*
  prC[0] = ( (rE1mag*pprE[2][0] - rE2mag*pprE[1][0])*rE00xE01
	     + (rE0mag*pprE[1][0] - rE1mag*pprE[0][0])*rE20xE21 ) / rDen;
  prC[1] = ( (rE1mag*pprE[2][1] - rE2mag*pprE[1][1])*rE00xE01
	     + (rE0mag*pprE[1][1] - rE1mag*pprE[0][1])*rE20xE21 ) / rDen;

  rMagCE30 = ( (prC[0] - ppprV[3][0][0])*(prC[0] - ppprV[3][0][0]) + 
	       (prC[1] - ppprV[3][0][1])*(prC[1] - ppprV[3][0][1]) );
  rMagCE31 = ( (prC[0] - ppprV[3][1][0])*(prC[0] - ppprV[3][1][0]) + 
	       (prC[1] - ppprV[3][1][1])*(prC[1] - ppprV[3][1][1]) );

  if ( rMagCE30 <= 0.0 )
    {
      prEdgeZone[0] = 0.0;
    }
  else
    {
      prEdgeZone[0] = ( ( rDen*rE3dotE30 
			  + rCoef1 ) /
			( rDen*sqrt(rE3dotE3*rMagCE30) ) );
    }
  if ( rMagCE31 <= 0.0 )
    {
      prEdgeZone[1] = 0.0;
    }
  else
    {
      prEdgeZone[1] = -( ( rDen*rE3dotE31 
			  + rCoef1 ) /
			 ( rDen*sqrt(rE3dotE3*rMagCE31) ) );
    }
*/
  
  prEdgeZone[0] = ( rDen*rE3dotE30 
		    + rCoef1 );
  prEdgeZone[1] = -( rDen*rE3dotE31 
		     + rCoef1 );
  if ( rDen < 0.0 )
    {
      prEdgeZone[0] = -prEdgeZone[0];
      prEdgeZone[1] = -prEdgeZone[1];
    }


  if ( !( prEdgeZone[0] < 0.0 ) )
    {
      eResult = APM2D::eztStart;
    }
  else if ( !( prEdgeZone[1] < 0.0 ) )
    {
      eResult = APM2D::eztEnd;
    }
  else
    {
      eResult = APM2D::eztMiddle;
    }

  return eResult;
}

//static 
APM2D::CircleTestType InCircleEEE_VMPFUN(DOUBLE dE00_x, DOUBLE dE00_y, 
					 DOUBLE dE01_x, DOUBLE dE01_y,
					 DOUBLE dE10_x, DOUBLE dE10_y, 
					 DOUBLE dE11_x, DOUBLE dE11_y,
					 DOUBLE dE20_x, DOUBLE dE20_y, 
					 DOUBLE dE21_x, DOUBLE dE21_y,
					 DOUBLE dV3_x, DOUBLE dV3_y)
{
  mp_real rE00_x = dE00_x;
  mp_real rE00_y = dE00_y; 
  mp_real rE01_x = dE01_x;
  mp_real rE01_y = dE01_y;
  mp_real rE10_x = dE10_x;
  mp_real rE10_y = dE10_y; 
  mp_real rE11_x = dE11_x;
  mp_real rE11_y = dE11_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;
  mp_real rV3_x = dV3_x;
  mp_real rV3_y = dV3_y;

  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real rE0dotE0, rE1dotE1, rE2dotE2, rV3dotV3, rE0dotV3, rE1dotV3, rE2dotV3;
  mp_real rE0dotE1, rE1dotE2, rE0dotE2;
  mp_real rE0mag, rE1mag, rE2mag;
  mp_real rE0xE1, rE1xE2, rE2xE0, rE00xE01, rE20xE21;
  mp_real rE1E2minusE1dotE2, rE0E1minusE0dotE1, rE0E2plusE0dotE2;
  mp_real rDen, rSinTheta;
  mp_real rResult;
  APM2D::CircleTestType eResult;
  
  // Translate V1 to the origin
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = rE00_x - rE10_x;
  ppprV[0][0][1] = rE00_y - rE10_y;
  ppprV[0][1][0] = rE01_x - rE10_x;
  ppprV[0][1][1] = rE01_y - rE10_y;
  ppprV[1][0][0] = 0.0;
  ppprV[1][0][1] = 0.0;
  ppprV[1][1][0] = rE11_x - rE10_x;
  ppprV[1][1][1] = rE11_y - rE10_y;
  ppprV[2][0][0] = rE20_x - rE10_x;
  ppprV[2][0][1] = rE20_y - rE10_y;
  ppprV[2][1][0] = rE21_x - rE10_x;
  ppprV[2][1][1] = rE21_y - rE10_y;
  ppprV[3][0][0] = rV3_x - rE10_x;
  ppprV[3][0][1] = rV3_y - rE10_y;
  
  //memset(pprE, 0, sizeof(pprE));
  pprE[0][0] = rE01_x - rE00_x;
  pprE[0][1] = rE01_y - rE00_y;
  pprE[1][0] = rE11_x - rE10_x;
  pprE[1][1] = rE11_y - rE10_y;
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;
  
  rE0dotE0 = (pprE[0][0]*pprE[0][0]) + (pprE[0][1]*pprE[0][1]);
  rE1dotE1 = (pprE[1][0]*pprE[1][0]) + (pprE[1][1]*pprE[1][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);
  rV3dotV3 = (ppprV[3][0][0]*ppprV[3][0][0]) + (ppprV[3][0][1]*ppprV[3][0][1]);

  rE0mag = sqrt(rE0dotE0);
  rE1mag = sqrt(rE1dotE1);
  rE2mag = sqrt(rE2dotE2);

  rE0dotV3 = (pprE[0][0]*ppprV[3][0][0]) + (pprE[0][1]*ppprV[3][0][1]);
  rE1dotV3 = (pprE[1][0]*ppprV[3][0][0]) + (pprE[1][1]*ppprV[3][0][1]);
  rE2dotV3 = (pprE[2][0]*ppprV[3][0][0]) + (pprE[2][1]*ppprV[3][0][1]);

  rE0dotE1 = (pprE[0][0]*pprE[1][0]) + (pprE[0][1]*pprE[1][1]);
  rE1dotE2 = (pprE[1][0]*pprE[2][0]) + (pprE[1][1]*pprE[2][1]);
  rE0dotE2 = (pprE[0][0]*pprE[2][0]) + (pprE[0][1]*pprE[2][1]);

  rE0xE1 = (pprE[0][0]*pprE[1][1]) - (pprE[0][1]*pprE[1][0]);
  rE1xE2 = (pprE[1][0]*pprE[2][1]) - (pprE[1][1]*pprE[2][0]);
  rE2xE0 = (pprE[2][0]*pprE[0][1]) - (pprE[2][1]*pprE[0][0]);

  rE00xE01 = (ppprV[0][0][0]*ppprV[0][1][1]) - (ppprV[0][0][1]*ppprV[0][1][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);

  rE0E1minusE0dotE1 = sqrt(rE0dotE0*rE1dotE1) - rE0dotE1;
  rE1E2minusE1dotE2 = sqrt(rE1dotE1*rE2dotE2) - rE1dotE2;
  rE0E2plusE0dotE2 = sqrt(rE0dotE0*rE2dotE2) + rE0dotE2;

  rDen = rE0mag*rE1xE2 + rE1mag*rE2xE0 + rE2mag*rE0xE1;
  rSinTheta = rDen;
  ASSERT( !( rSinTheta == 0.0 ) );

  rResult = ( rDen*( rDen*rV3dotV3 
		     + 2.0*( rE0dotV3*rE1mag*rE20xE21 + rE1dotV3*( rE2mag*rE00xE01 - rE0mag*rE20xE21 ) - rE2dotV3*rE1mag*rE00xE01 ) )
	      + rE00xE01*rE00xE01*rE1E2minusE1dotE2*rE1E2minusE1dotE2
	      + 2.0*rE00xE01*rE20xE21*( rE1mag*( rE0mag*rE1dotE2 + rE2mag*rE0dotE1 ) - rE1dotE1*rE0E2plusE0dotE2 - rE0xE1*rE1xE2 )
	      + rE20xE21*rE20xE21*rE0E1minusE0dotE1*rE0E1minusE0dotE1 );
  
  if ( rResult < 0.0 )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( rResult > 0.0 )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

//static 
APM2D::CircleTestType InCircleEEE_EMPFUN(DOUBLE dE00_x, DOUBLE dE00_y, 
					 DOUBLE dE01_x, DOUBLE dE01_y,
					 DOUBLE dE10_x, DOUBLE dE10_y, 
					 DOUBLE dE11_x, DOUBLE dE11_y,
					 DOUBLE dE20_x, DOUBLE dE20_y, 
					 DOUBLE dE21_x, DOUBLE dE21_y,
					 DOUBLE dE30_x, DOUBLE dE30_y,
					 DOUBLE dE31_x, DOUBLE dE31_y)
{
  mp_real rE00_x = dE00_x;
  mp_real rE00_y = dE00_y; 
  mp_real rE01_x = dE01_x;
  mp_real rE01_y = dE01_y;
  mp_real rE10_x = dE10_x;
  mp_real rE10_y = dE10_y; 
  mp_real rE11_x = dE11_x;
  mp_real rE11_y = dE11_y;
  mp_real rE20_x = dE20_x;
  mp_real rE20_y = dE20_y; 
  mp_real rE21_x = dE21_x;
  mp_real rE21_y = dE21_y;
  mp_real rE30_x = dE30_x;
  mp_real rE30_y = dE30_y;
  mp_real rE31_x = dE31_x;
  mp_real rE31_y = dE31_y;
  
  mp_real ppprV[4][2][2]; // Endpoints
  mp_real pprE[4][2];     // Edge vectors
  mp_real rE0dotE0, rE1dotE1, rE2dotE2, rE3dotE3;
  mp_real rE0mag, rE1mag, rE2mag;
  mp_real rE0xE1, rE1xE2, rE2xE0;
  mp_real rE0xE3, rE1xE3, rE2xE3;
  mp_real rE00xE01, rE20xE21, rE30xE31;
  mp_real rA, rB;
  mp_real rDen, rSinTheta;
  mp_real rResult;
  APM2D::CircleTestType eResult;
  
  // Translate V1 to the origin
  ppprV[0][0][0] = rE00_x - rE10_x;
  ppprV[0][0][1] = rE00_y - rE10_y;
  ppprV[0][1][0] = rE01_x - rE10_x;
  ppprV[0][1][1] = rE01_y - rE10_y;
  ppprV[1][0][0] = 0.0;
  ppprV[1][0][1] = 0.0;
  ppprV[1][1][0] = rE11_x - rE10_x;
  ppprV[1][1][1] = rE11_y - rE10_y;
  ppprV[2][0][0] = rE20_x - rE10_x;
  ppprV[2][0][1] = rE20_y - rE10_y;
  ppprV[2][1][0] = rE21_x - rE10_x;
  ppprV[2][1][1] = rE21_y - rE10_y;
  ppprV[3][0][0] = rE30_x - rE10_x;
  ppprV[3][0][1] = rE30_y - rE10_y;
  ppprV[3][1][0] = rE31_x - rE10_x;
  ppprV[3][1][1] = rE31_y - rE10_y;
  
  pprE[0][0] = rE01_x - rE00_x;
  pprE[0][1] = rE01_y - rE00_y;
  pprE[1][0] = rE11_x - rE10_x;
  pprE[1][1] = rE11_y - rE10_y;
  pprE[2][0] = rE21_x - rE20_x;
  pprE[2][1] = rE21_y - rE20_y;
  pprE[3][0] = rE31_x - rE30_x;
  pprE[3][1] = rE31_y - rE30_y;
  
  rE0dotE0 = (pprE[0][0]*pprE[0][0]) + (pprE[0][1]*pprE[0][1]);
  rE1dotE1 = (pprE[1][0]*pprE[1][0]) + (pprE[1][1]*pprE[1][1]);
  rE2dotE2 = (pprE[2][0]*pprE[2][0]) + (pprE[2][1]*pprE[2][1]);
  rE3dotE3 = (pprE[3][0]*pprE[3][0]) + (pprE[3][1]*pprE[3][1]);
  
  rE0mag = sqrt(rE0dotE0);
  rE1mag = sqrt(rE1dotE1);
  rE2mag = sqrt(rE2dotE2);
  
  rE0xE1 = (pprE[0][0]*pprE[1][1]) - (pprE[0][1]*pprE[1][0]);
  rE1xE2 = (pprE[1][0]*pprE[2][1]) - (pprE[1][1]*pprE[2][0]);
  rE2xE0 = (pprE[2][0]*pprE[0][1]) - (pprE[2][1]*pprE[0][0]);
    
  rE0xE3 = (pprE[0][0]*pprE[3][1]) - (pprE[0][1]*pprE[3][0]);
  rE1xE3 = (pprE[1][0]*pprE[3][1]) - (pprE[1][1]*pprE[3][0]);
  rE2xE3 = (pprE[2][0]*pprE[3][1]) - (pprE[2][1]*pprE[3][0]);
    
  rE00xE01 = (ppprV[0][0][0]*ppprV[0][1][1]) - (ppprV[0][0][1]*ppprV[0][1][0]);
  rE20xE21 = (ppprV[2][0][0]*ppprV[2][1][1]) - (ppprV[2][0][1]*ppprV[2][1][0]);
  rE30xE31 = (ppprV[3][0][0]*ppprV[3][1][1]) - (ppprV[3][0][1]*ppprV[3][1][0]);
  
  rA = ( ( sqrt(rE2dotE2)*rE1xE3 - sqrt(rE1dotE1)*rE2xE3 )*rE00xE01
	 + ( sqrt(rE1dotE1)*rE0xE3 - sqrt(rE0dotE0)*rE1xE3 )*rE20xE21 );
  rB = ( rE0xE1*rE20xE21 + rE1xE2*rE00xE01 );
  rB = rB*rB;

  rDen = rE0mag*rE1xE2 + rE1mag*rE2xE0 + rE2mag*rE0xE1;
  rSinTheta = rDen;
  ASSERT( !( rSinTheta == 0.0 ) );
  
  rResult = ( rA*rA
	      + 2.0*rE30xE31*rDen*rA 
	      + rE30xE31*rE30xE31*rDen*rDen
	      - rE3dotE3*rB );

  if ( rResult < 0.0 )
    {
      eResult = APM2D::cttInCircle;
    }
  else if ( rResult > 0.0 )
    {
      eResult = APM2D::cttOutCircle;
    }
  else
    {
      eResult = APM2D::cttOnCircle;
    }

  return eResult;
}

