#include "apm2d_mpfunfilter.h"
#include "apm2d_mpfun.h"
#include <stdio.h>

#include "interval.h"

#undef MAX
#undef MIN
#include "mpreal.h"

#define COMPILER_PROBLEM

//static 
void Dot(Interval &valResult, 
	 const Interval &valE0_x, const Interval &valE0_y, 
	 const Interval &valE1_x, const Interval &valE1_y)
{
  valResult = ( valE0_x * valE1_x ) + ( valE0_y * valE1_y ); 
}
//static 
void Dot(Interval &valResult, 
	 const Interval &valE00_x, const Interval &valE00_y, 
	 const Interval &valE01_x, const Interval &valE01_y, 
	 const Interval &valE10_x, const Interval &valE10_y,
	 const Interval &valE11_x, const Interval &valE11_y)
{
  valResult = ( ( ( valE01_x - valE00_x ) * ( valE11_x - valE10_x ) )
	      + ( ( valE01_y - valE00_y ) * ( valE11_y - valE10_y ) ) ); 
}
//static 
void Cross(Interval &valResult, 
	   const Interval &valE0_x, const Interval &valE0_y, 
	   const Interval &valE1_x, const Interval &valE1_y)
{
  valResult = ( valE0_x * valE1_y ) - ( valE0_y * valE1_x ); 
}
//static 
void Cross(Interval &valResult, 
	   const Interval &valE00_x, const Interval &valE00_y, 
	   const Interval &valE01_x, const Interval &valE01_y, 
	   const Interval &valE10_x, const Interval &valE10_y,
	   const Interval &valE11_x, const Interval &valE11_y)
{
  valResult = ( ( ( valE01_x - valE00_x ) * ( valE11_y - valE10_y ) )
	      - ( ( valE01_y - valE00_y ) * ( valE11_x - valE10_x ) ) ); 
}

static 
void Dot(mp_real &rResult, 
	 const mp_real &rE0_x, const mp_real &rE0_y, 
	 const mp_real &rE1_x, const mp_real &rE1_y)
{
  rResult = ( rE0_x * rE1_x ) + ( rE0_y * rE1_y ); 
}
//static 
void Dot(mp_real &rResult, 
	 const mp_real &rE00_x, const mp_real &rE00_y, 
	 const mp_real &rE01_x, const mp_real &rE01_y, 
	 const mp_real &rE10_x, const mp_real &rE10_y,
	 const mp_real &rE11_x, const mp_real &rE11_y)
{
  rResult = ( ( ( rE01_x - rE00_x ) * ( rE11_x - rE10_x ) )
            + ( ( rE01_y - rE00_y ) * ( rE11_y - rE10_y ) ) ); 
}
static 
void Cross(mp_real &rResult, 
	   const mp_real &rE0_x, const mp_real &rE0_y, 
	   const mp_real &rE1_x, const mp_real &rE1_y)
{
  rResult = ( rE0_x * rE1_y ) - ( rE0_y * rE1_x ); 
}
static 
void Cross(mp_real &rResult, 
	   const mp_real &rE00_x, const mp_real &rE00_y, 
	   const mp_real &rE01_x, const mp_real &rE01_y, 
	   const mp_real &rE10_x, const mp_real &rE10_y,
	   const mp_real &rE11_x, const mp_real &rE11_y)
{
  rResult = ( ( ( rE01_x - rE00_x ) * ( rE11_y - rE10_y ) )
            - ( ( rE01_y - rE00_y ) * ( rE11_x - rE10_x ) ) ); 
}

//static
inline void SinTheta(Interval &valResult,
		     const Interval &valV2_x, const Interval &valV2_y,
		     const Interval &valV0_x, const Interval &valV0_y,
		     const Interval &valV1_x, const Interval &valV1_y)
{
  Cross(valResult, 
	valV0_x, valV0_y, valV1_x, valV1_y, 
	valV0_x, valV0_y, valV2_x, valV2_y);
}

static
inline void SinTheta(mp_real &rResult,
		     const mp_real &rV2_x, const mp_real &rV2_y,
		     const mp_real &rV0_x, const mp_real &rV0_y,
		     const mp_real &rV1_x, const mp_real &rV1_y)
{
  Cross(rResult, 
	rV0_x, rV0_y, rV1_x, rV1_y, 
	rV0_x, rV0_y, rV2_x, rV2_y);
}

//////////////////////////////////////////////////////////////////////
// APM2D MPFUNFilter Implementation

//////////////////////////////////////////////////////////////////////
// Intersection point of two edges

//static
BOOL IntersectionEEMPFUNFilter(DOUBLE *pdV_x, DOUBLE *pdV_y,
			       DOUBLE dE00_x, DOUBLE dE00_y, 
			       DOUBLE dE01_x, DOUBLE dE01_y,
			       DOUBLE dE10_x, DOUBLE dE10_y, 
			       DOUBLE dE11_x, DOUBLE dE11_y)
{
  ASSERT( pdV_x != NULL && pdV_y != NULL );
  // Set to zero for the case of no intersection
  *pdV_x = 0.0;
  *pdV_y = 0.0;
  
  Interval valE00_x = dE00_x;
  Interval valE00_y = dE00_y; 
  Interval valE01_x = dE01_x;
  Interval valE01_y = dE01_y;
  Interval valE10_x = dE10_x;
  Interval valE10_y = dE10_y;
  Interval valE11_x = dE11_x;
  Interval valE11_y = dE11_y;

  mp_real rE00_x = dE00_x;
  mp_real rE00_y = dE00_y; 
  mp_real rE01_x = dE01_x;
  mp_real rE01_y = dE01_y;
  mp_real rE10_x = dE10_x;
  mp_real rE10_y = dE10_y;
  mp_real rE11_x = dE11_x;
  mp_real rE11_y = dE11_y;

  //////////////////////////////
  // Parallel or colinear check
  Interval valE0xE1;
  mp_real rE0xE1;
  int iE0xE1, bPositive;
  valE0xE1 = ( (valE01_x - valE00_x)*(valE11_y - valE10_y) - 
	     (valE11_x - valE10_x)*(valE01_y - valE00_y) ); 
  iE0xE1 = ( valE0xE1 == 0.0 );
  if ( iE0xE1 != -1 )
    {
      if ( iE0xE1 )
	{
	  // Parallel or colinear edges
	  return FALSE;
	}
      bPositive = (valE0xE1 > 0.0);
    }
  else
    {
      rE0xE1 = ( (rE01_x - rE00_x)*(rE11_y - rE10_y) - 
		 (rE11_x - rE10_x)*(rE01_y - rE00_y) ); 
      if ( rE0xE1 == 0.0 )
	{
	  // Parallel or colinear edges
	  return FALSE;
	}
      bPositive = (rE0xE1 > 0.0);
    }

  //////////////////////////////
  // Check I vs E00 half plane
  int iE10xE11_E00;
  Interval valE10xE11_E00;
  valE10xE11_E00 = ( (valE10_x - valE00_x)*(valE11_y - valE00_y) - 
		   (valE11_x - valE00_x)*(valE10_y - valE00_y) );
  if ( !bPositive )
    {
      iE10xE11_E00 = ( 0.0 < valE10xE11_E00 );
    }
  else
    {
      iE10xE11_E00 = ( 0.0 > valE10xE11_E00 );
    }
  if ( iE10xE11_E00 == 1 )
    {
      return FALSE;
    }

  //////////////////////////////
  // Check I vs E01 half plane
  int iE10xE11_E01;
  Interval valE10xE11_E01;
  valE10xE11_E01 = ( (valE10_x - valE01_x)*(valE11_y - valE01_y) - 
		     (valE11_x - valE01_x)*(valE10_y - valE01_y) );
  if ( !bPositive )
    {
      iE10xE11_E01 = ( 0.0 > valE10xE11_E01 );
    }
  else
    {
      iE10xE11_E01 = ( 0.0 < valE10xE11_E01 );
    }
  if ( iE10xE11_E01 == 1 )
    {
      return FALSE;
    }

  //////////////////////////////
  // Check I vs E10 half plane
  int iE00xE01_E10;
  Interval valE00xE01_E10;
  valE00xE01_E10 = ( (valE00_x - valE10_x)*(valE01_y - valE10_y) - 
		   (valE01_x - valE10_x)*(valE00_y - valE10_y) );
  if ( bPositive )
    {
      iE00xE01_E10 = ( 0.0 < valE00xE01_E10 );
    }
  else
    {
      iE00xE01_E10 = ( 0.0 > valE00xE01_E10 );
    }
  if ( iE00xE01_E10 == 1 )
    {
      return FALSE;
    }

  //////////////////////////////
  // Check I vs E11 half plane
  int iE00xE01_E11;
  Interval valE00xE01_E11;
  valE00xE01_E11 = ( (valE00_x - valE11_x)*(valE01_y - valE11_y) - 
		     (valE01_x - valE11_x)*(valE00_y - valE11_y) );
  if ( bPositive )
    {
      iE00xE01_E11 = ( 0.0 > valE00xE01_E11 );
    }
  else
    {
      iE00xE01_E11 = ( 0.0 < valE00xE01_E11 );
    }
  if ( iE00xE01_E11 == 1 )
    {
      return FALSE;
    }

  //////////////////////////////////////////////////
  // MPFUN Tests

  //////////////////////////////
  // Check I vs E01 half plane
  if ( iE10xE11_E01 == -1 )
    {
      mp_real rE10xE11_E01;
      rE10xE11_E01 = ( (rE10_x - rE01_x)*(rE11_y - rE01_y) - 
		       (rE11_x - rE01_x)*(rE10_y - rE01_y) );
      if ( !bPositive )
	{
	  if ( 0.0 > rE10xE11_E01 )
	    {
	      return FALSE;
	    }
	}
      else
	{
	  if ( 0.0 < rE10xE11_E01 )
	    {
	      return FALSE;
	    }
	}
    }

  //////////////////////////////
  // Check I vs E10 half plane
  if ( iE00xE01_E10 == -1 )
    {
      mp_real rE00xE01_E10;
      rE00xE01_E10 = ( (rE00_x - rE10_x)*(rE01_y - rE10_y) - 
		       (rE01_x - rE10_x)*(rE00_y - rE10_y) );
      if ( bPositive )
	{
	  if ( 0.0 < rE00xE01_E10 )
	    {
	      return FALSE;
	    }
	}
      else
	{
	  if ( 0.0 > rE00xE01_E10 )
	    {
	      return FALSE;
	    }
	}
    }

  //////////////////////////////
  // Check I vs E11 half plane
  if ( iE00xE01_E11 == -1 )
    {
      mp_real rE00xE01_E11;
      rE00xE01_E11 = ( (rE00_x - rE11_x)*(rE01_y - rE11_y) - 
		       (rE01_x - rE11_x)*(rE00_y - rE11_y) );
      if ( bPositive )
	{
	  if ( 0.0 > rE00xE01_E11 )
	    {
	      return FALSE;
	    }
	}
      else
	{
	  if ( 0.0 < rE00xE01_E11 )
	    {
	      return FALSE;
	    }
	}
    }

  //////////////////////////////
  // Check I vs E00 half plane
  mp_real rE10xE11_E00;
  rE10xE11_E00 = ( (rE10_x - rE00_x)*(rE11_y - rE00_y) - 
		   (rE11_x - rE00_x)*(rE10_y - rE00_y) );
  if ( iE10xE11_E00 == -1 )
    {
      if ( !bPositive )
	{
	  if ( 0.0 < rE10xE11_E00 )
	    {
	      return FALSE;
	    }
	}
      else
	{
	  if ( 0.0 > rE10xE11_E00 )
	    {
	      return FALSE;
	    }
	}
    }

  //////////////////////////////
  // Force MPFUN computation of the denominator
  if ( iE0xE1 != -1 )
    {
      rE0xE1 = ( (rE01_x - rE00_x)*(rE11_y - rE10_y) - 
		 (rE11_x - rE10_x)*(rE01_y - rE00_y) ); 
    }

  mp_real rV_x, rV_y;
  rV_x = rE00_x + rE10xE11_E00*(rE01_x - rE00_x)/rE0xE1;
  rV_y = rE00_y + rE10xE11_E00*(rE01_y - rE00_y)/rE0xE1;
  *pdV_x = dble(rV_x);
  *pdV_y = dble(rV_y);
  return TRUE;
}

//////////////////////////////////////////////////////////////////////
// Distance Comparison Functions

//static
APM2D::EdgeZoneType EdgeZoneMPFUNFilter(DOUBLE dV0_x, DOUBLE dV0_y,
					DOUBLE dE10_x, DOUBLE dE10_y, 
					DOUBLE dE11_x, DOUBLE dE11_y)
{
  int piEdgeZone[2];

  APM2D::m_uiEdgeZoneMPFUNFilter++;

  Interval valV0_x = dV0_x;
  Interval valV0_y = dV0_y;
  Interval valE10_x = dE10_x;
  Interval valE10_y = dE10_y;
  Interval valE11_x = dE11_x;
  Interval valE11_y = dE11_y;

  Interval ppvalV[2][2]; // Endpoints
  Interval pvalE[2];     // Edge vectors
  Interval valE1dotE10, valE1dotE11;
  
  pvalE[0] = valE11_x - valE10_x;
  pvalE[1] = valE11_y - valE10_y;
  
  ppvalV[0][0] = valE10_x - valV0_x;
  ppvalV[0][1] = valE10_y - valV0_y;
  valE1dotE10 = (pvalE[0]*ppvalV[0][0]) + (pvalE[1]*ppvalV[0][1]);
  piEdgeZone[0] = ( valE1dotE10 >= 0.0 );
  if ( piEdgeZone[0] == 1 )
    {
      return APM2D::eztStart;
    }

  ppvalV[1][0] = valE11_x - valV0_x;
  ppvalV[1][1] = valE11_y - valV0_y;
  valE1dotE11 = (pvalE[0]*ppvalV[1][0]) + (pvalE[1]*ppvalV[1][1]);
  piEdgeZone[1] = ( valE1dotE11 <= 0.0 );
  if ( piEdgeZone[1] == 1 )
    {
      return APM2D::eztEnd;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( piEdgeZone[0] == -1 || piEdgeZone[0] == -1 )
    {
      APM2D::m_uiEdgeZoneMPFUN++;

      mp_real rV0_x = dV0_x;
      mp_real rV0_y = dV0_y;
      mp_real rE10_x = dE10_x;
      mp_real rE10_y = dE10_y;
      mp_real rE11_x = dE11_x;
      mp_real rE11_y = dE11_y;

      mp_real pprV[2][2]; // Endpoints
      mp_real prE[2];     // Edge vectors
      mp_real rE1dotE10, rE1dotE11;
      
      prE[0] = rE11_x - rE10_x;
      prE[1] = rE11_y - rE10_y;
  
      if ( piEdgeZone[0] == -1 )
	{      
	  pprV[0][0] = rE10_x - rV0_x;
	  pprV[0][1] = rE10_y - rV0_y;
	  rE1dotE10 = (prE[0]*pprV[0][0]) + (prE[1]*pprV[0][1]);
	  if ( rE1dotE10 >= 0.0 )
	    {
	      return APM2D::eztStart;
	    }
	}

      if ( piEdgeZone[1] == -1 )
	{      
	  pprV[1][0] = rE11_x - rV0_x;
	  pprV[1][1] = rE11_y - rV0_y;
	  rE1dotE11 = (prE[0]*pprV[1][0]) + (prE[1]*pprV[1][1]);
	  if ( rE1dotE11 <= 0.0 )
	    {
	      return APM2D::eztEnd;
	    }
	}
    }

  return APM2D::eztMiddle;
}

//static 
INT_32 DistanceCompareV_VVMPFUNFilter(DOUBLE dV0_x, DOUBLE dV0_y,
				      DOUBLE dV1_x, DOUBLE dV1_y, 
				      DOUBLE dV2_x, DOUBLE dV2_y)
{
  INT_32 iResult;
  iResult = 0;

  APM2D::m_uiDistanceCompareV_VVMPFUNFilter++;

  //////////////////////////////////////////////////
  // Double precision interval arithmetic filter

  Interval valV0_x = dV0_x;
  Interval valV0_y = dV0_y;
  Interval valV1_x = dV1_x;
  Interval valV1_y = dV1_y;
  Interval valV2_x = dV2_x;
  Interval valV2_y = dV2_y;

  Interval pppvalV[3][2][2]; // Endpoints
  Interval val, valV1dotV1, valV2dotV2;
  
  //memset(pppvalV, 0, sizeof(pppvalV));
  pppvalV[1][0][0] = valV1_x - valV0_x;
  pppvalV[1][0][1] = valV1_y - valV0_y;
  pppvalV[2][0][0] = valV2_x - valV0_x;
  pppvalV[2][0][1] = valV2_y - valV0_y;
  
  valV1dotV1 = (pppvalV[1][0][0]*pppvalV[1][0][0]) + (pppvalV[1][0][1]*pppvalV[1][0][1]);
  valV2dotV2 = (pppvalV[2][0][0]*pppvalV[2][0][0]) + (pppvalV[2][0][1]*pppvalV[2][0][1]);

  val = valV1dotV1 - valV2dotV2;
  
  int i = ( val == 0.0 );
  if ( i != -1 )
    {
      if ( i )
	{
	  iResult = 0;
	}
      else
	{
	  i = (val > 0.0);
	  ASSERT( i >= 0 );
	  if ( i )
	    {
	      iResult = 1;
	    }
	  else
	    {
	      i = (val < 0.0);
	      ASSERT( i == 1 );
	      iResult = -1;
	    }
	}

      return iResult;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  APM2D::m_uiDistanceCompareV_VVMPFUN++;
  return DistanceCompareV_VVMPFUN(dV0_x, dV0_y,
				  dV1_x, dV1_y, 
				  dV2_x, dV2_y);
}

//static 
INT_32 DistanceCompareV_VLMPFUNFilter(DOUBLE dV0_x, DOUBLE dV0_y,
				      DOUBLE dV1_x, DOUBLE dV1_y, 
				      DOUBLE dE20_x, DOUBLE dE20_y, 
				      DOUBLE dE21_x, DOUBLE dE21_y)
{
  INT_32 iResult;
  iResult = 0;

  APM2D::m_uiDistanceCompareV_VLMPFUNFilter++;

  //////////////////////////////////////////////////
  // Double precision interval arithmetic filter

  Interval valV0_x = dV0_x;
  Interval valV0_y = dV0_y;
  Interval valV1_x = dV1_x;
  Interval valV1_y = dV1_y;
  Interval valE20_x = dE20_x;
  Interval valE20_y = dE20_y;
  Interval valE21_x = dE21_x;
  Interval valE21_y = dE21_y;

  Interval pppvalV[3][2][2]; // Endpoints
  Interval ppvalE[3][2];     // Edge vectors
  Interval val, valV1dotV1, valE2dotE2, valE20xE21;
  
  //memset(pppvalV, 0, sizeof(pppvalV));
  //pppvalV[0][0][0] = 0.0;
  //pppvalV[0][0][1] = 0.0;
  pppvalV[1][0][0] = valV1_x - valV0_x;
  pppvalV[1][0][1] = valV1_y - valV0_y;
  pppvalV[2][0][0] = valE20_x - valV0_x;
  pppvalV[2][0][1] = valE20_y - valV0_y;
  pppvalV[2][1][0] = valE21_x - valV0_x;
  pppvalV[2][1][1] = valE21_y - valV0_y;
  
  //memset(ppvalE, 0, sizeof(ppvalE));
  ppvalE[2][0] = valE21_x - valE20_x;
  ppvalE[2][1] = valE21_y - valE20_y;
  
  valV1dotV1 = (pppvalV[1][0][0]*pppvalV[1][0][0]) + (pppvalV[1][0][1]*pppvalV[1][0][1]);
  valE2dotE2 = (ppvalE[2][0]*ppvalE[2][0]) + (ppvalE[2][1]*ppvalE[2][1]);

  valE20xE21 = (pppvalV[2][0][0]*pppvalV[2][1][1]) - (pppvalV[2][0][1]*pppvalV[2][1][0]);

  val = valV1dotV1*valE2dotE2 - valE20xE21*valE20xE21;

  int i = ( val == 0.0 );
  if ( i != -1 )
    {
      if ( i )
	{
	  iResult = 0;
	}
      else
	{
	  i = (val > 0.0);
	  ASSERT( i >= 0 );
	  if ( i )
	    {
	      iResult = 1;
	    }
	  else
	    {
	      i = (val < 0.0);
	      ASSERT( i == 1 );
	      iResult = -1;
	    }
	}

      return iResult;
    }
  
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  APM2D::m_uiDistanceCompareV_VLMPFUN++;
  return DistanceCompareV_VLMPFUN(dV0_x, dV0_y,
				  dV1_x, dV1_y, 
				  dE20_x, dE20_y, 
				  dE21_x, dE21_y);
}

//static 
INT_32 DistanceCompareV_LLMPFUNFilter(DOUBLE dV0_x, DOUBLE dV0_y,
				      DOUBLE dE10_x, DOUBLE dE10_y, 
				      DOUBLE dE11_x, DOUBLE dE11_y,
				      DOUBLE dE20_x, DOUBLE dE20_y, 
				      DOUBLE dE21_x, DOUBLE dE21_y)
{
  INT_32 iResult;
  iResult = 0;

  APM2D::m_uiDistanceCompareV_LLMPFUNFilter++;

  Interval valV0_x = dV0_x;
  Interval valV0_y = dV0_y;
  Interval valE10_x = dE10_x;
  Interval valE10_y = dE10_y;
  Interval valE11_x = dE11_x;
  Interval valE11_y = dE11_y;
  Interval valE20_x = dE20_x;
  Interval valE20_y = dE20_y;
  Interval valE21_x = dE21_x;
  Interval valE21_y = dE21_y;

  Interval pppvalV[3][2][2]; // Endpoints
  Interval ppvalE[3][2];     // Edge vectors
  Interval val, valE1dotE1, valE2dotE2, valE10xE11, valE20xE21;
  
  //memset(pppvalV, 0, sizeof(pppvalV));
  pppvalV[0][0][0] = 0.0;
  pppvalV[0][0][1] = 0.0;
  pppvalV[1][0][0] = valE10_x - valV0_x;
  pppvalV[1][0][1] = valE10_y - valV0_y;
  pppvalV[1][1][0] = valE11_x - valV0_x;
  pppvalV[1][1][1] = valE11_y - valV0_y;
  pppvalV[2][0][0] = valE20_x - valV0_x;
  pppvalV[2][0][1] = valE20_y - valV0_y;
  pppvalV[2][1][0] = valE21_x - valV0_x;
  pppvalV[2][1][1] = valE21_y - valV0_y;
  
  //memset(ppvalE, 0, sizeof(ppvalE));
  ppvalE[1][0] = valE11_x - valE10_x;
  ppvalE[1][1] = valE11_y - valE10_y;
  ppvalE[2][0] = valE21_x - valE20_x;
  ppvalE[2][1] = valE21_y - valE20_y;
  
  valE1dotE1 = (ppvalE[1][0]*ppvalE[1][0]) + (ppvalE[1][1]*ppvalE[1][1]);
  valE2dotE2 = (ppvalE[2][0]*ppvalE[2][0]) + (ppvalE[2][1]*ppvalE[2][1]);

  valE10xE11 = (pppvalV[1][0][0]*pppvalV[1][1][1]) - (pppvalV[1][0][1]*pppvalV[1][1][0]);
  valE20xE21 = (pppvalV[2][0][0]*pppvalV[2][1][1]) - (pppvalV[2][0][1]*pppvalV[2][1][0]);

  val = valE2dotE2*valE10xE11*valE10xE11 - valE1dotE1*valE20xE21*valE20xE21;
  
  int i = ( val == 0.0 ); 
  if ( i != -1 )
    {
      if ( i )
	{
	  iResult = 0;
	}
      else if ( val > 0.0 )
	{
	  iResult = 1;
	}
      else
	{
	  ASSERT( val < 0.0 );
	  iResult = -1;
	}

      return iResult;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  APM2D::m_uiDistanceCompareV_LLMPFUN++;
  return DistanceCompareV_LLMPFUN(dV0_x, dV0_y,
				  dE10_x, dE10_y, 
				  dE11_x, dE11_y,
				  dE20_x, dE20_y, 
				  dE21_x, dE21_y);
}

//static 
INT_32 DistanceCompareEvv_VVMPFUNFilter(DOUBLE dE0A_x, DOUBLE dE0A_y,
					DOUBLE dE0B_x, DOUBLE dE0B_y,
					DOUBLE dV1_x, DOUBLE dV1_y, 
					DOUBLE dV2_x, DOUBLE dV2_y)
{
  INT_32 iResult;
  iResult = 0;

  APM2D::m_uiDistanceCompareEvv_VVMPFUNFilter++;

  Interval valE0A_x = dE0A_x;
  Interval valE0A_y = dE0A_y;
  Interval valE0B_x = dE0B_x;
  Interval valE0B_y = dE0B_y;
  Interval valV1_x = dV1_x;
  Interval valV1_y = dV1_y;
  Interval valV2_x = dV2_x;
  Interval valV2_y = dV2_y;
  
  Interval pppvalV[3][2][2]; // Endpoints
  Interval val, valV1dotV1, valV2dotV2;
  
  //memset(pppvalV, 0, sizeof(pppvalV));
  pppvalV[1][0][0] = valV1_x - valE0A_x;
  pppvalV[1][0][1] = valV1_y - valE0A_y;
  pppvalV[2][0][0] = valV2_x - valE0B_x;
  pppvalV[2][0][1] = valV2_y - valE0B_y;
  
  valV1dotV1 = (pppvalV[1][0][0]*pppvalV[1][0][0]) + (pppvalV[1][0][1]*pppvalV[1][0][1]);
  valV2dotV2 = (pppvalV[2][0][0]*pppvalV[2][0][0]) + (pppvalV[2][0][1]*pppvalV[2][0][1]);

  val = valV1dotV1 - valV2dotV2;
  
  int i = ( val == 0.0 );
  if ( i != -1 )
    {
      if ( i )
	{
	  iResult = 0;
	}
      else if ( val > 0.0 )
	{
	  iResult = 1;
	}
      else
	{
	  ASSERT( val < 0.0 );
	  iResult = -1;
	}

      return iResult;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  APM2D::m_uiDistanceCompareEvv_VVMPFUN++;
  return DistanceCompareEvv_VVMPFUN(dE0A_x, dE0A_y,
				    dE0B_x, dE0B_y,
				    dV1_x, dV1_y, 
				    dV2_x, dV2_y);
}

//static 
INT_32 DistanceCompareEvv_VLMPFUNFilter(DOUBLE dE0A_x, DOUBLE dE0A_y,
					DOUBLE dE0B_x, DOUBLE dE0B_y,
					DOUBLE dV1_x, DOUBLE dV1_y, 
					DOUBLE dE20_x, DOUBLE dE20_y, 
					DOUBLE dE21_x, DOUBLE dE21_y)
{
  INT_32 iResult;
  iResult = 0;

  APM2D::m_uiDistanceCompareEvv_VLMPFUNFilter++;

  Interval valE0A_x = dE0A_x;
  Interval valE0A_y = dE0A_y;
  Interval valE0B_x = dE0B_x;
  Interval valE0B_y = dE0B_y;
  Interval valV1_x = dV1_x;
  Interval valV1_y = dV1_y;
  Interval valE20_x = dE20_x;
  Interval valE20_y = dE20_y;
  Interval valE21_x = dE21_x;
  Interval valE21_y = dE21_y;

  Interval pppvalV[3][2][2]; // Endpoints
  Interval ppvalE[3][2];     // Edge vectors
  Interval val, valV1dotV1, valE2dotE2, valE20xE21;
  
  //memset(pppvalV, 0, sizeof(pppvalV));
  pppvalV[0][0][0] = 0.0;
  pppvalV[0][0][1] = 0.0;
  pppvalV[1][0][0] = valV1_x - valE0A_x;
  pppvalV[1][0][1] = valV1_y - valE0A_y;
  pppvalV[2][0][0] = valE20_x - valE0B_x;
  pppvalV[2][0][1] = valE20_y - valE0B_y;
  pppvalV[2][1][0] = valE21_x - valE0B_x;
  pppvalV[2][1][1] = valE21_y - valE0B_y;
  
  //memset(ppvalE, 0, sizeof(ppvalE));
  ppvalE[2][0] = valE21_x - valE20_x;
  ppvalE[2][1] = valE21_y - valE20_y;
  
  valV1dotV1 = (pppvalV[1][0][0]*pppvalV[1][0][0]) + (pppvalV[1][0][1]*pppvalV[1][0][1]);
  valE2dotE2 = (ppvalE[2][0]*ppvalE[2][0]) + (ppvalE[2][1]*ppvalE[2][1]);

  valE20xE21 = (pppvalV[2][0][0]*pppvalV[2][1][1]) - (pppvalV[2][0][1]*pppvalV[2][1][0]);

  val = valV1dotV1*valE2dotE2 - valE20xE21*valE20xE21;
  
  int i = ( val == 0.0 );
  if ( i != -1 )
    {
      if ( i )
	{
	  iResult = 0;
	}
      else if ( val > 0.0 )
	{
	  iResult = 1;
	}
      else
	{
	  ASSERT( val < 0.0 );
	  iResult = -1;
	}
      
      return iResult;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  APM2D::m_uiDistanceCompareEvv_VLMPFUN++;
  return DistanceCompareEvv_VLMPFUN(dE0A_x, dE0A_y,
				    dE0B_x, dE0B_y,
				    dV1_x, dV1_y, 
				    dE20_x, dE20_y, 
				    dE21_x, dE21_y);
}

//static 
INT_32 DistanceCompareEvv_LLMPFUNFilter(DOUBLE dE0A_x, DOUBLE dE0A_y,
					DOUBLE dE0B_x, DOUBLE dE0B_y,
					DOUBLE dE10_x, DOUBLE dE10_y, 
					DOUBLE dE11_x, DOUBLE dE11_y,
					DOUBLE dE20_x, DOUBLE dE20_y, 
					DOUBLE dE21_x, DOUBLE dE21_y)
{
  INT_32 iResult;
  iResult = 0;

  APM2D::m_uiDistanceCompareEvv_LLMPFUNFilter++;

  Interval valE0A_x = dE0A_x;
  Interval valE0A_y = dE0A_y;
  Interval valE0B_x = dE0B_x;
  Interval valE0B_y = dE0B_y;
  Interval valE10_x = dE10_x;
  Interval valE10_y = dE10_y;
  Interval valE11_x = dE11_x;
  Interval valE11_y = dE11_y;
  Interval valE20_x = dE20_x;
  Interval valE20_y = dE20_y;
  Interval valE21_x = dE21_x;
  Interval valE21_y = dE21_y;

  Interval pppvalV[3][2][2]; // Endpoints
  Interval ppvalE[3][2];     // Edge vectors
  Interval val, valE1dotE1, valE2dotE2, valE10xE11, valE20xE21;
  
  //memset(pppvalV, 0, sizeof(pppvalV));
  pppvalV[0][0][0] = 0.0;
  pppvalV[0][0][1] = 0.0;
  pppvalV[1][0][0] = valE10_x - valE0A_x;
  pppvalV[1][0][1] = valE10_y - valE0A_y;
  pppvalV[1][1][0] = valE11_x - valE0A_x;
  pppvalV[1][1][1] = valE11_y - valE0A_y;
  pppvalV[2][0][0] = valE20_x - valE0B_x;
  pppvalV[2][0][1] = valE20_y - valE0B_y;
  pppvalV[2][1][0] = valE21_x - valE0B_x;
  pppvalV[2][1][1] = valE21_y - valE0B_y;
  
  //memset(ppvalE, 0, sizeof(ppvalE));
  ppvalE[1][0] = valE11_x - valE10_x;
  ppvalE[1][1] = valE11_y - valE10_y;
  ppvalE[2][0] = valE21_x - valE20_x;
  ppvalE[2][1] = valE21_y - valE20_y;
  
  valE1dotE1 = (ppvalE[1][0]*ppvalE[1][0]) + (ppvalE[1][1]*ppvalE[1][1]);
  valE2dotE2 = (ppvalE[2][0]*ppvalE[2][0]) + (ppvalE[2][1]*ppvalE[2][1]);

  valE10xE11 = (pppvalV[1][0][0]*pppvalV[1][1][1]) - (pppvalV[1][0][1]*pppvalV[1][1][0]);
  valE20xE21 = (pppvalV[2][0][0]*pppvalV[2][1][1]) - (pppvalV[2][0][1]*pppvalV[2][1][0]);

  val = valE2dotE2*valE10xE11*valE10xE11 - valE1dotE1*valE20xE21*valE20xE21;
  
  int i = ( val == 0.0 );
  if ( i != -1 )
    {
      if ( i )
	{
	  iResult = 0;
	}
      else if ( val > 0.0 )
	{
	  iResult = 1;
	}
      else
	{
	  ASSERT( val < 0.0 );
	  iResult = -1;
	}

      return iResult;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  APM2D::m_uiDistanceCompareEvv_LLMPFUNFilter++;
  return DistanceCompareEvv_LLMPFUNFilter(dE0A_x, dE0A_y,
					  dE0B_x, dE0B_y,
					  dE10_x, dE10_y, 
					  dE11_x, dE11_y,
					  dE20_x, dE20_y, 
					  dE21_x, dE21_y);
}

//static 
INT_32 DistanceCompareElv_VVMPFUNFilter(DOUBLE dE00_x, DOUBLE dE00_y,
					DOUBLE dE01_x, DOUBLE dE01_y,
					DOUBLE dE0B_x, DOUBLE dE0B_y,
					DOUBLE dV1_x, DOUBLE dV1_y, 
					DOUBLE dV2_x, DOUBLE dV2_y)
{
  INT_32 iResult;
  iResult = 0;

  APM2D::m_uiDistanceCompareElv_VVMPFUNFilter++;

  Interval valE00_x = dE00_x;
  Interval valE00_y = dE00_y;
  Interval valE01_x = dE01_x;
  Interval valE01_y = dE01_y;
  Interval valE0B_x = dE0B_x;
  Interval valE0B_y = dE0B_y;
  Interval valV1_x = dV1_x;
  Interval valV1_y = dV1_y;
  Interval valV2_x = dV2_x;
  Interval valV2_y = dV2_y;
  
  Interval pppvalV[3][2][2]; // Endpoints
  Interval ppvalE[3][2];     // Edge vectors
  Interval val, valE0dotE0, valV2dotV2, valE00xE01;
  

  //memset(pppvalV, 0, sizeof(pppvalV));
  pppvalV[0][0][0] = valE00_x - valV1_x;
  pppvalV[0][0][1] = valE00_y - valV1_y;
  pppvalV[0][1][0] = valE01_x - valV1_x;
  pppvalV[0][1][1] = valE01_y - valV1_y;
  pppvalV[1][0][0] = 0.0;
  pppvalV[1][0][1] = 0.0;
  pppvalV[2][0][0] = valV2_x - valE0B_x;
  pppvalV[2][0][1] = valV2_y - valE0B_y;
  
  //memset(ppvalE, 0, sizeof(ppvalE));
  ppvalE[0][0] = valE01_x - valE00_x;
  ppvalE[0][1] = valE01_y - valE00_y;

  valE0dotE0 = (ppvalE[0][0]*ppvalE[0][0]) + (ppvalE[0][1]*ppvalE[0][1]);
  valV2dotV2 = (pppvalV[2][0][0]*pppvalV[2][0][0]) + (pppvalV[2][0][1]*pppvalV[2][0][1]);

  valE00xE01 = (pppvalV[0][0][0]*pppvalV[0][1][1]) - (pppvalV[0][0][1]*pppvalV[0][1][0]);

  val = valE00xE01*valE00xE01 - valV2dotV2*valE0dotE0;
  
  int i = ( val == 0.0 );
  if ( i != -1 )
    {
      if ( i )
	{
	  iResult = 0;
	}
      else if ( val > 0.0 )
	{
	  iResult = 1;
	}
      else
	{
	  ASSERT( val < 0.0 );
	  iResult = -1;
	}

      return iResult;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  APM2D::m_uiDistanceCompareElv_VVMPFUN++;
  return DistanceCompareElv_VVMPFUN(dE00_x, dE00_y,
				    dE01_x, dE01_y,
				    dE0B_x, dE0B_y,
				    dV1_x, dV1_y, 
				    dV2_x, dV2_y);
}

//static 
INT_32 DistanceCompareElv_VLMPFUNFilter(DOUBLE dE00_x, DOUBLE dE00_y,
					DOUBLE dE01_x, DOUBLE dE01_y,
					DOUBLE dE0B_x, DOUBLE dE0B_y,
					DOUBLE dV1_x, DOUBLE dV1_y, 
					DOUBLE dE20_x, DOUBLE dE20_y, 
					DOUBLE dE21_x, DOUBLE dE21_y)
{
  INT_32 iResult;
  iResult = 0;

  APM2D::m_uiDistanceCompareElv_VLMPFUNFilter++;

  Interval valE00_x = dE00_x;
  Interval valE00_y = dE00_y;
  Interval valE01_x = dE01_x;
  Interval valE01_y = dE01_y;
  Interval valE0B_x = dE0B_x;
  Interval valE0B_y = dE0B_y;
  Interval valV1_x = dV1_x;
  Interval valV1_y = dV1_y;
  Interval valE20_x = dE20_x;
  Interval valE20_y = dE20_y;
  Interval valE21_x = dE21_x;
  Interval valE21_y = dE21_y;

  Interval pppvalV[3][2][2]; // Endpoints
  Interval ppvalE[3][2];     // Edge vectors
  Interval val, valE0dotE0, valE2dotE2, valE00xE01, valE20xE21;
  
  //memset(pppvalV, 0, sizeof(pppvalV));
  pppvalV[0][0][0] = valE00_x - valV1_x;
  pppvalV[0][0][1] = valE00_y - valV1_y;
  pppvalV[0][1][0] = valE01_x - valV1_x;
  pppvalV[0][1][1] = valE01_y - valV1_y;
  pppvalV[1][0][0] = 0.0;
  pppvalV[1][0][1] = 0.0;
  pppvalV[2][0][0] = valE20_x - valE0B_x;
  pppvalV[2][0][1] = valE20_y - valE0B_y;
  pppvalV[2][1][0] = valE21_x - valE0B_x;
  pppvalV[2][1][1] = valE21_y - valE0B_y;
  
  //memset(ppvalE, 0, sizeof(ppvalE));
  ppvalE[0][0] = valE01_x - valE00_x;
  ppvalE[0][1] = valE01_y - valE00_y;
  ppvalE[2][0] = valE21_x - valE20_x;
  ppvalE[2][1] = valE21_y - valE20_y;
  
  valE0dotE0 = (ppvalE[0][0]*ppvalE[0][0]) + (ppvalE[0][1]*ppvalE[0][1]);
  valE2dotE2 = (ppvalE[2][0]*ppvalE[2][0]) + (ppvalE[2][1]*ppvalE[2][1]);

  valE00xE01 = (pppvalV[0][0][0]*pppvalV[0][1][1]) - (pppvalV[0][0][1]*pppvalV[0][1][0]);
  valE20xE21 = (pppvalV[2][0][0]*pppvalV[2][1][1]) - (pppvalV[2][0][1]*pppvalV[2][1][0]);

  val = valE2dotE2*valE00xE01*valE00xE01 - valE0dotE0*valE20xE21*valE20xE21;
  
  int i = ( val == 0.0 );
  if ( i != -1 )
    {
      if ( i )
	{
	  iResult = 0;
	}
      else if ( val > 0.0 )
	{
	  iResult = 1;
	}
      else
	{
	  ASSERT( val < 0.0 );
	  iResult = -1;
	}
      
      return iResult;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  APM2D::m_uiDistanceCompareElv_VLMPFUN++;
  return DistanceCompareElv_VLMPFUN(dE00_x, dE00_y,
				    dE01_x, dE01_y,
				    dE0B_x, dE0B_y,
				    dV1_x, dV1_y, 
				    dE20_x, dE20_y, 
				    dE21_x, dE21_y);
}

//static 
INT_32 DistanceCompareL_VVMPFUNFilter(DOUBLE dE00_x, DOUBLE dE00_y,
				      DOUBLE dE01_x, DOUBLE dE01_y,
				      DOUBLE dV1_x, DOUBLE dV1_y, 
				      DOUBLE dV2_x, DOUBLE dV2_y)
{
  INT_32 iResult;
  iResult = 0;

  APM2D::m_uiDistanceCompareL_VVMPFUNFilter++;

  Interval valE00_x = dE00_x;
  Interval valE00_y = dE00_y;
  Interval valE01_x = dE01_x;
  Interval valE01_y = dE01_y;
  Interval valV1_x = dV1_x;
  Interval valV1_y = dV1_y;
  Interval valV2_x = dV2_x;
  Interval valV2_y = dV2_y;

  Interval pppvalV[3][2][2]; // Endpoints
  Interval ppvalE[3][2];     // Edge vectors
  Interval val, valE0dotE0, valE10xE11, valE20xE21;
  
  //memset(pppvalV, 0, sizeof(pppvalV));
  pppvalV[1][0][0] = valE00_x - valV1_x;
  pppvalV[1][0][1] = valE00_y - valV1_y;
  pppvalV[1][1][0] = valE01_x - valV1_x;
  pppvalV[1][1][1] = valE01_y - valV1_y;
  pppvalV[2][0][0] = valE00_x - valV2_x;
  pppvalV[2][0][1] = valE00_y - valV2_y;
  pppvalV[2][1][0] = valE01_x - valV2_x;
  pppvalV[2][1][1] = valE01_y - valV2_y;
  
  //memset(ppvalE, 0, sizeof(ppvalE));
  ppvalE[0][0] = valE01_x - valE00_x;
  ppvalE[0][1] = valE01_y - valE00_y;

  valE0dotE0 = (ppvalE[0][0]*ppvalE[0][0]) + (ppvalE[0][1]*ppvalE[0][1]);

  valE10xE11 = (pppvalV[1][0][0]*pppvalV[1][1][1]) - (pppvalV[1][0][1]*pppvalV[1][1][0]);
  valE20xE21 = (pppvalV[2][0][0]*pppvalV[2][1][1]) - (pppvalV[2][0][1]*pppvalV[2][1][0]);

  val = valE10xE11*valE10xE11 - valE20xE21*valE20xE21;
  
  int i = ( val == 0.0 );
  if ( i != -1 )
    {
      if ( i )
	{
	  iResult = 0;
	}
      else if ( val > 0.0 )
	{
	  iResult = 1;
	}
      else
	{
	  ASSERT( val < 0.0 );
	  iResult = -1;
	}
      
      return iResult;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  APM2D::m_uiDistanceCompareL_VVMPFUN++;
  return DistanceCompareL_VVMPFUN(dE00_x, dE00_y,
				  dE01_x, dE01_y,
				  dV1_x, dV1_y, 
				  dV2_x, dV2_y);
}

//////////////////////////////////////////////////////////////////////
// VVV Orientation Functions

//////////////////////////////////////////////////////////////////////
// EE Orientaion Functions

//static
INT_32 OrientationExEMPFUNFilter(DOUBLE dE00_x, DOUBLE dE00_y,
				 DOUBLE dE01_x, DOUBLE dE01_y,
				 DOUBLE dE10_x, DOUBLE dE10_y,
				 DOUBLE dE11_x, DOUBLE dE11_y)
{
  INT_32 iResult;
  iResult = 0;

  APM2D::m_uiOrientationExEMPFUNFilter++;

  Interval valE00_x = dE00_x;
  Interval valE00_y = dE00_y;
  Interval valE01_x = dE01_x;
  Interval valE01_y = dE01_y;
  Interval valE10_x = dE10_x;
  Interval valE10_y = dE10_y;
  Interval valE11_x = dE11_x;
  Interval valE11_y = dE11_y;

  Interval val;
  Interval pvalE0[2], pvalE1[2];
  
  pvalE0[0] = valE01_x - valE00_x;
  pvalE0[1] = valE01_y - valE00_y;
  pvalE1[0] = valE11_x - valE10_x;
  pvalE1[1] = valE11_y - valE10_y;
  
  val = ( (pvalE0[0] * pvalE1[1]) - (pvalE1[0] * pvalE0[1]) );

  int i = ( val == 0.0 );
  if ( i != -1 )
    {
      if ( i )
	{
	  iResult = 0;
	}
      else if ( val > 0.0 )
	{
	  iResult = 1;
	}
      else
	{
	  ASSERT( val < 0.0 );
	  iResult = -1;
	}

      return iResult;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  APM2D::m_uiOrientationExEMPFUN++;
  return OrientationExEMPFUN(dE00_x, dE00_y,
			     dE01_x, dE01_y,
			     dE10_x, dE10_y,
			     dE11_x, dE11_y);
}

//static  
INT_32 OrientationEdotEMPFUNFilter(DOUBLE dE00_x, DOUBLE dE00_y,
				   DOUBLE dE01_x, DOUBLE dE01_y,
				   DOUBLE dE10_x, DOUBLE dE10_y,
				   DOUBLE dE11_x, DOUBLE dE11_y)
{
  INT_32 iResult;
  iResult = 0;

  APM2D::m_uiOrientationEdotEMPFUNFilter++;

  Interval valE00_x = dE00_x;
  Interval valE00_y = dE00_y;
  Interval valE01_x = dE01_x;
  Interval valE01_y = dE01_y;
  Interval valE10_x = dE10_x;
  Interval valE10_y = dE10_y;
  Interval valE11_x = dE11_x;
  Interval valE11_y = dE11_y;

  Interval val;
  Interval pvalE0[2], pvalE1[2];
  
  pvalE0[0] = valE01_x - valE00_x;
  pvalE0[1] = valE01_y - valE00_y;
  pvalE1[0] = valE11_x - valE10_x;
  pvalE1[1] = valE11_y - valE10_y;
  
  val = ( (pvalE0[0] * pvalE1[0]) + (pvalE0[1] * pvalE1[1]) );

  int i = ( val == 0.0 );
  if ( i != -1 )
    {
      if ( i )
	{
	  iResult = 0;
	}
      else if ( val > 0.0 )
	{
	  iResult = 1;
	}
      else
	{
	  ASSERT( val < 0.0 );
	  iResult = -1;
	}

      return iResult;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  APM2D::m_uiOrientationEdotEMPFUN++;
  return OrientationEdotEMPFUN(dE00_x, dE00_y,
			       dE01_x, dE01_y,
			       dE10_x, dE10_y,
			       dE11_x, dE11_y);
}

//////////////////////////////////////////////////////////////////////
// Inscribed Circle Center Functions

//////////////////////////////////////////////////
// VVV

class ICC_VVV_MPFUNFilter
{
public:
  ICC_VVV_MPFUNFilter();
  ~ICC_VVV_MPFUNFilter();

public:
  APM2D::InscribedCircleCenterType 
  InscribedCircleCenterVVV(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			   DOUBLE dV00_x, DOUBLE dV00_y,
			   DOUBLE dV0_x, DOUBLE dV0_y,
			   DOUBLE dV01_x, DOUBLE dV01_y,
			   DOUBLE dV10_x, DOUBLE dV10_y,
			   DOUBLE dV1_x, DOUBLE dV1_y,
			   DOUBLE dV11_x, DOUBLE dV11_y,
			   DOUBLE dV20_x, DOUBLE dV20_y,
			   DOUBLE dV2_x, DOUBLE dV2_y,
			   DOUBLE dV21_x, DOUBLE dV21_y);
protected:
  APM2D::InscribedCircleCenterType 
  InscribedCircleCenterVVVMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dV00_x, DOUBLE dV00_y,
				DOUBLE dV0_x, DOUBLE dV0_y,
				DOUBLE dV01_x, DOUBLE dV01_y,
				DOUBLE dV10_x, DOUBLE dV10_y,
				DOUBLE dV1_x, DOUBLE dV1_y,
				DOUBLE dV11_x, DOUBLE dV11_y,
				DOUBLE dV20_x, DOUBLE dV20_y,
				DOUBLE dV2_x, DOUBLE dV2_y,
				DOUBLE dV21_x, DOUBLE dV21_y);
protected:
  INT_32 m_ppiVertexZone[3][2];
};

ICC_VVV_MPFUNFilter::ICC_VVV_MPFUNFilter()
{
}

ICC_VVV_MPFUNFilter::~ICC_VVV_MPFUNFilter()
{
}

APM2D::InscribedCircleCenterType 
ICC_VVV_MPFUNFilter::InscribedCircleCenterVVV(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					      DOUBLE dV00_x, DOUBLE dV00_y,
					      DOUBLE dV0_x, DOUBLE dV0_y,
					      DOUBLE dV01_x, DOUBLE dV01_y,
					      DOUBLE dV10_x, DOUBLE dV10_y,
					      DOUBLE dV1_x, DOUBLE dV1_y,
					      DOUBLE dV11_x, DOUBLE dV11_y,
					      DOUBLE dV20_x, DOUBLE dV20_y,
					      DOUBLE dV2_x, DOUBLE dV2_y,
					      DOUBLE dV21_x, DOUBLE dV21_y)
{
  APM2D::m_uiInscribedCircleCenterVVVMPFUNFilter++;

  Interval valV00_x = dV00_x; 
  Interval valV00_y = dV00_y;
  Interval valV0_x = dV0_x; 
  Interval valV0_y = dV0_y;
  Interval valV01_x = dV01_x; 
  Interval valV01_y = dV01_y;
  Interval valV10_x = dV10_x; 
  Interval valV10_y = dV10_y;
  Interval valV1_x = dV1_x; 
  Interval valV1_y = dV1_y;
  Interval valV11_x = dV11_x; 
  Interval valV11_y = dV11_y;
  Interval valV20_x = dV20_x; 
  Interval valV20_y = dV20_y;
  Interval valV2_x = dV2_x; 
  Interval valV2_y = dV2_y;
  Interval valV21_x = dV21_x; 
  Interval valV21_y = dV21_y;
  
  // Translate V0 to the origin
  Interval ppvalV[3][2];
  //memset(pprV, 0, sizeof(ppvalV));
  ppvalV[1][0] = valV1_x - valV0_x;
  ppvalV[1][1] = valV1_y - valV0_y;
  ppvalV[2][0] = valV2_x - valV0_x;
  ppvalV[2][1] = valV2_y - valV0_y;

  // Vevaltex neighboval vectovals
  Interval pppvalVN[3][2][2];
  //memset(pppvalVN, 0, sizeof(pppvalVN));
  pppvalVN[0][0][0] = valV0_x - valV00_x;
  pppvalVN[0][0][1] = valV0_y - valV00_y;
  pppvalVN[0][1][0] = valV0_x - valV01_x;
  pppvalVN[0][1][1] = valV0_y - valV01_y;
  pppvalVN[1][0][0] = valV1_x - valV10_x;
  pppvalVN[1][0][1] = valV1_y - valV10_y;
  pppvalVN[1][1][0] = valV1_x - valV11_x;
  pppvalVN[1][1][1] = valV1_y - valV11_y;
  pppvalVN[2][0][0] = valV2_x - valV20_x;
  pppvalVN[2][0][1] = valV2_y - valV20_y;
  pppvalVN[2][1][0] = valV2_x - valV21_x;
  pppvalVN[2][1][1] = valV2_y - valV21_y;

  Interval valV1xV2;
  Cross(valV1xV2, ppvalV[1][0], ppvalV[1][1], ppvalV[2][0], ppvalV[2][1]);
  Interval valV1dotV1;
  Dot(valV1dotV1, ppvalV[1][0], ppvalV[1][1], ppvalV[1][0], ppvalV[1][1]);
  Interval valV2dotV2;
  Dot(valV2dotV2, ppvalV[2][0], ppvalV[2][1], ppvalV[2][0], ppvalV[2][1]);

  //////////////////////////////////////////////////
  // Check the Vertex Zones

  Interval ppvalVertexZone[3][2];

  Interval valV1xV00;
  Cross(valV1xV00, ppvalV[1][0], ppvalV[1][1], pppvalVN[0][0][0], pppvalVN[0][0][1]);
  Interval valV2xV00;
  Cross(valV2xV00, ppvalV[2][0], ppvalV[2][1], pppvalVN[0][0][0], pppvalVN[0][0][1]);
  ppvalVertexZone[0][0] = ( valV2dotV2*valV1xV00 - valV1dotV1*valV2xV00 );
  m_ppiVertexZone[0][0] = ( ppvalVertexZone[0][0] < 0.0 );
  if ( m_ppiVertexZone[0][0] == 1 )
    {
      return APM2D::icctNull;
    }
  Interval valV1xV01;
  Cross(valV1xV01, ppvalV[1][0], ppvalV[1][1], pppvalVN[0][1][0], pppvalVN[0][1][1]);
  Interval valV2xV01;
  Cross(valV2xV01, ppvalV[2][0], ppvalV[2][1], pppvalVN[0][1][0], pppvalVN[0][1][1]);
  ppvalVertexZone[0][1] = ( valV2dotV2*valV1xV01 - valV1dotV1*valV2xV01 );
  m_ppiVertexZone[0][1] = ( ppvalVertexZone[0][1] < 0.0 );
  if ( m_ppiVertexZone[0][1] == 1 )
    {
      return APM2D::icctNull;
    }
  Interval valV1xV10;
  Cross(valV1xV10, ppvalV[1][0], ppvalV[1][1], pppvalVN[1][0][0], pppvalVN[1][0][1]);
  Interval valV2xV10;
  Cross(valV2xV10, ppvalV[2][0], ppvalV[2][1], pppvalVN[1][0][0], pppvalVN[1][0][1]);
  Interval valV1dotV10;
  Dot(valV1dotV10, ppvalV[1][0], ppvalV[1][1], pppvalVN[1][0][0], pppvalVN[1][0][1]);
  ppvalVertexZone[1][0] = ( valV2dotV2*valV1xV10 - valV1dotV1*valV2xV10 - 2.0*valV1xV2*valV1dotV10 );
  m_ppiVertexZone[1][0] = ( ppvalVertexZone[1][0] < 0.0 );
  if ( m_ppiVertexZone[1][0] == 1 )
    {
      return APM2D::icctNull;
    }
  Interval valV1xV11;
  Cross(valV1xV11, ppvalV[1][0], ppvalV[1][1], pppvalVN[1][1][0], pppvalVN[1][1][1]);
  Interval valV2xV11;
  Cross(valV2xV11, ppvalV[2][0], ppvalV[2][1], pppvalVN[1][1][0], pppvalVN[1][1][1]);
  Interval valV1dotV11;
  Dot(valV1dotV11, ppvalV[1][0], ppvalV[1][1], pppvalVN[1][1][0], pppvalVN[1][1][1]);
  ppvalVertexZone[1][1] = ( valV2dotV2*valV1xV11 - valV1dotV1*valV2xV11 - 2.0*valV1xV2*valV1dotV11 );
  m_ppiVertexZone[1][1] = ( ppvalVertexZone[1][1] < 0.0 );
  if ( m_ppiVertexZone[1][1] == 1 )
    {
      return APM2D::icctNull;
    }
  Interval valV1xV20;
  Cross(valV1xV20, ppvalV[1][0], ppvalV[1][1], pppvalVN[2][0][0], pppvalVN[2][0][1]);
  Interval valV2xV20;
  Cross(valV2xV20, ppvalV[2][0], ppvalV[2][1], pppvalVN[2][0][0], pppvalVN[2][0][1]);
  Interval valV2dotV20;
  Dot(valV2dotV20, ppvalV[2][0], ppvalV[2][1], pppvalVN[2][0][0], pppvalVN[2][0][1]);
  ppvalVertexZone[2][0] = ( valV2dotV2*valV1xV20 - valV1dotV1*valV2xV20 - 2.0*valV1xV2*valV2dotV20 );
  m_ppiVertexZone[2][0] = ( ppvalVertexZone[2][0] < 0.0 );
  if ( m_ppiVertexZone[2][0] == 1 )
    {
      return APM2D::icctNull;
    }
  Interval valV1xV21;
  Cross(valV1xV21, ppvalV[1][0], ppvalV[1][1], pppvalVN[2][1][0], pppvalVN[2][1][1]);
  Interval valV2xV21;
  Cross(valV2xV21, ppvalV[2][0], ppvalV[2][1], pppvalVN[2][1][0], pppvalVN[2][1][1]);
  Interval valV2dotV21;
  Dot(valV2dotV21, ppvalV[2][0], ppvalV[2][1], pppvalVN[2][1][0], pppvalVN[2][1][1]);
  ppvalVertexZone[2][1] = ( valV2dotV2*valV1xV21 - valV1dotV1*valV2xV21 - 2.0*valV1xV2*valV2dotV21 );
  m_ppiVertexZone[2][1] = ( ppvalVertexZone[2][1] < 0.0 );
  if ( m_ppiVertexZone[2][1] == 1 )
    {
      return APM2D::icctNull;
    }

  if ( m_ppiVertexZone[0][0] == -1 ||
       m_ppiVertexZone[0][1] == -1 ||
       m_ppiVertexZone[1][0] == -1 ||
       m_ppiVertexZone[1][1] == -1 ||
       m_ppiVertexZone[2][0] == -1 ||
       m_ppiVertexZone[2][1] == -1 )
    {
      return InscribedCircleCenterVVVMPFUN(pdC_x, pdC_y, pdR, 
					   dV00_x, dV00_y,
					   dV0_x, dV0_y,
					   dV01_x, dV01_y,
					   dV10_x, dV10_y,
					   dV1_x, dV1_y,
					   dV11_x, dV11_y,
					   dV20_x, dV20_y,
					   dV2_x, dV2_y,
					   dV21_x, dV21_y);
    }

  //////////////////////////////////////////////////
  // Compute the center and radius

  Interval valDen = 2.0 * valV1xV2;
  if ( ( valDen != 0.0 ) != 1 )
    {
      return InscribedCircleCenterVVVMPFUN(pdC_x, pdC_y, pdR, 
					   dV00_x, dV00_y,
					   dV0_x, dV0_y,
					   dV01_x, dV01_y,
					   dV10_x, dV10_y,
					   dV1_x, dV1_y,
					   dV11_x, dV11_y,
					   dV20_x, dV20_y,
					   dV2_x, dV2_y,
					   dV21_x, dV21_y);
    }
      
  Interval pvalC[2];
  pvalC[0] = ( (valV2dotV2 * ppvalV[1][1]) - (valV1dotV1 * ppvalV[2][1]) ) / -valDen;
  pvalC[1] = ( (valV2dotV2 * ppvalV[1][0]) - (valV1dotV1 * ppvalV[2][0]) ) / valDen;

  Interval pvalV1V2[2];
  pvalV1V2[0] = valV2_x - valV1_x;
  pvalV1V2[1] = valV2_y - valV1_y;
  Interval valV1V2dotV1V2 = (pvalV1V2[0] * pvalV1V2[0]) + (pvalV1V2[1] * pvalV1V2[1]);

  Interval valR = sqrt(valV1dotV1 * valV2dotV2 * valV1V2dotV1V2) / valDen;

  // Translate the answer back out away from the origin 
  pvalC[0] = valV0_x + pvalC[0];
  pvalC[1] = valV0_y + pvalC[1];

  *pdC_x = dble(pvalC[0]);
  *pdC_y = dble(pvalC[1]);
  *pdR = dble(valR);
  return APM2D::icctVVV;
}

APM2D::InscribedCircleCenterType 
ICC_VVV_MPFUNFilter::InscribedCircleCenterVVVMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dV00_x, DOUBLE dV00_y,
				DOUBLE dV0_x, DOUBLE dV0_y,
				DOUBLE dV01_x, DOUBLE dV01_y,
				DOUBLE dV10_x, DOUBLE dV10_y,
				DOUBLE dV1_x, DOUBLE dV1_y,
				DOUBLE dV11_x, DOUBLE dV11_y,
				DOUBLE dV20_x, DOUBLE dV20_y,
				DOUBLE dV2_x, DOUBLE dV2_y,
				DOUBLE dV21_x, DOUBLE dV21_y)
{
  APM2D::m_uiInscribedCircleCenterVVVMPFUN++;
  APM2D::m_uiInscribedCircleCenterVVVMPFUNFinal++;

  mp_real rV00_x = dV00_x; 
  mp_real rV00_y = dV00_y;
  mp_real rV0_x = dV0_x; 
  mp_real rV0_y = dV0_y;
  mp_real rV01_x = dV01_x; 
  mp_real rV01_y = dV01_y;
  mp_real rV10_x = dV10_x; 
  mp_real rV10_y = dV10_y;
  mp_real rV1_x = dV1_x; 
  mp_real rV1_y = dV1_y;
  mp_real rV11_x = dV11_x; 
  mp_real rV11_y = dV11_y;
  mp_real rV20_x = dV20_x; 
  mp_real rV20_y = dV20_y;
  mp_real rV2_x = dV2_x; 
  mp_real rV2_y = dV2_y;
  mp_real rV21_x = dV21_x; 
  mp_real rV21_y = dV21_y;
  
  // Translate V0 to the origin
  mp_real pprV[3][2];
  //memset(pprV, 0, sizeof(pprV));
  pprV[1][0] = rV1_x - rV0_x;
  pprV[1][1] = rV1_y - rV0_y;
  pprV[2][0] = rV2_x - rV0_x;
  pprV[2][1] = rV2_y - rV0_y;

  // Vertex neighbor vectors
  mp_real ppprVN[3][2][2];
  //memset(ppprVN, 0, sizeof(ppprVN));
  ppprVN[0][0][0] = rV0_x - rV00_x;
  ppprVN[0][0][1] = rV0_y - rV00_y;
  ppprVN[0][1][0] = rV0_x - rV01_x;
  ppprVN[0][1][1] = rV0_y - rV01_y;
  ppprVN[1][0][0] = rV1_x - rV10_x;
  ppprVN[1][0][1] = rV1_y - rV10_y;
  ppprVN[1][1][0] = rV1_x - rV11_x;
  ppprVN[1][1][1] = rV1_y - rV11_y;
  ppprVN[2][0][0] = rV2_x - rV20_x;
  ppprVN[2][0][1] = rV2_y - rV20_y;
  ppprVN[2][1][0] = rV2_x - rV21_x;
  ppprVN[2][1][1] = rV2_y - rV21_y;

  //////////////////////////////////////////////////
  // Check the Vertex Zones

  mp_real pprVertexZone[3][2];

  mp_real rV1xV2 = ( (pprV[1][0] * pprV[2][1]) - (pprV[2][0] * pprV[1][1]) );
  mp_real rV1dotV1 = (pprV[1][0] * pprV[1][0]) + (pprV[1][1] * pprV[1][1]);
  mp_real rV2dotV2 = (pprV[2][0] * pprV[2][0]) + (pprV[2][1] * pprV[2][1]);

  if ( m_ppiVertexZone[0][0] == -1 )
    {
      mp_real rV1xV00 = ( (pprV[1][0] * ppprVN[0][0][1]) - (pprV[1][1] * ppprVN[0][0][0]) );
      mp_real rV2xV00 = ( (pprV[2][0] * ppprVN[0][0][1]) - (pprV[2][1] * ppprVN[0][0][0]) );
      pprVertexZone[0][0] = ( rV2dotV2*rV1xV00 - rV1dotV1*rV2xV00 );
      if ( pprVertexZone[0][0] < 0.0 )
	{
	  return APM2D::icctNull;
	}
    }
  if ( m_ppiVertexZone[0][1] == -1 )
    {
      mp_real rV1xV01 = ( (pprV[1][0] * ppprVN[0][1][1]) - (pprV[1][1] * ppprVN[0][1][0]) );
      mp_real rV2xV01 = ( (pprV[2][0] * ppprVN[0][1][1]) - (pprV[2][1] * ppprVN[0][1][0]) );
      pprVertexZone[0][1] = ( rV2dotV2*rV1xV01 - rV1dotV1*rV2xV01 );
      if ( pprVertexZone[0][1] < 0.0 )
	{
	  return APM2D::icctNull;
	}
    }
  if ( m_ppiVertexZone[1][0] == -1 )
    {
      mp_real rV1xV10 = ( (pprV[1][0] * ppprVN[1][0][1]) - (pprV[1][1] * ppprVN[1][0][0]) );
      mp_real rV2xV10 = ( (pprV[2][0] * ppprVN[1][0][1]) - (pprV[2][1] * ppprVN[1][0][0]) );
      mp_real rV1dotV10 = ( (pprV[1][0] * ppprVN[1][0][0]) + (pprV[1][1] * ppprVN[1][0][1]) );
      pprVertexZone[1][0] = ( rV2dotV2*rV1xV10 - rV1dotV1*rV2xV10 - 2.0*rV1xV2*rV1dotV10 );
      if ( pprVertexZone[1][0] < 0.0 )
	{
	  return APM2D::icctNull;
	}
    }
  if ( m_ppiVertexZone[1][1] == -1 )
    {
      mp_real rV1xV11 = ( (pprV[1][0] * ppprVN[1][1][1]) - (pprV[1][1] * ppprVN[1][1][0]) );
      mp_real rV2xV11 = ( (pprV[2][0] * ppprVN[1][1][1]) - (pprV[2][1] * ppprVN[1][1][0]) );
      mp_real rV1dotV11 = ( (pprV[1][0] * ppprVN[1][1][0]) + (pprV[1][1] * ppprVN[1][1][1]) );
      pprVertexZone[1][1] = ( rV2dotV2*rV1xV11 - rV1dotV1*rV2xV11 - 2.0*rV1xV2*rV1dotV11 );
      if ( pprVertexZone[1][1] < 0.0 )
	{
	  return APM2D::icctNull;
	}
    }
  if ( m_ppiVertexZone[2][0] == -1 )
    {
      mp_real rV1xV20 = ( (pprV[1][0] * ppprVN[2][0][1]) - (pprV[1][1] * ppprVN[2][0][0]) );
      mp_real rV2xV20 = ( (pprV[2][0] * ppprVN[2][0][1]) - (pprV[2][1] * ppprVN[2][0][0]) );
      mp_real rV2dotV20 = ( (pprV[2][0] * ppprVN[2][0][0]) + (pprV[2][1] * ppprVN[2][0][1]) );
      pprVertexZone[2][0] = ( rV2dotV2*rV1xV20 - rV1dotV1*rV2xV20 - 2.0*rV1xV2*rV2dotV20 );
      if ( pprVertexZone[2][0] < 0.0 )
	{
	  return APM2D::icctNull;
	}
    }
  if ( m_ppiVertexZone[2][1] == -1 )
    {
      mp_real rV1xV21 = ( (pprV[1][0] * ppprVN[2][1][1]) - (pprV[1][1] * ppprVN[2][1][0]) );
      mp_real rV2xV21 = ( (pprV[2][0] * ppprVN[2][1][1]) - (pprV[2][1] * ppprVN[2][1][0]) );
      mp_real rV2dotV21 = ( (pprV[2][0] * ppprVN[2][1][0]) + (pprV[2][1] * ppprVN[2][1][1]) );
      pprVertexZone[2][1] = ( rV2dotV2*rV1xV21 - rV1dotV1*rV2xV21 - 2.0*rV1xV2*rV2dotV21 );
      if ( pprVertexZone[2][1] < 0.0 )
	{
	  return APM2D::icctNull;
	}
    }

  //////////////////////////////////////////////////
  // Compute the center and radius

  mp_real prC[2];
  prC[0] =  ( (rV2dotV2 * pprV[1][1]) - (rV1dotV1 * pprV[2][1]) ) / ( -2.0 * rV1xV2 );
  prC[1] =  ( (rV2dotV2 * pprV[1][0]) - (rV1dotV1 * pprV[2][0]) ) / ( 2.0 * rV1xV2 );

  mp_real prV1V2[2];
  prV1V2[0] = rV2_x - rV1_x;
  prV1V2[1] = rV2_y - rV1_y;
  mp_real rV1V2dotV1V2 = (prV1V2[0] * prV1V2[0]) + (prV1V2[1] * prV1V2[1]);

  mp_real rR = sqrt(rV1dotV1 * rV2dotV2 * rV1V2dotV1V2) / ( 2.0 * rV1xV2 );

  // Translate the answer back out away from the origin 
  prC[0] = rV0_x + prC[0];
  prC[1] = rV0_y + prC[1];

  *pdC_x = dble(prC[0]);
  *pdC_y = dble(prC[1]);
  *pdR = dble(rR);
  return APM2D::icctVVV;
}


//static 
APM2D::InscribedCircleCenterType 
InscribedCircleCenterVVVMPFUNFilter(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				    DOUBLE dV00_x, DOUBLE dV00_y,
				    DOUBLE dV0_x, DOUBLE dV0_y,
				    DOUBLE dV01_x, DOUBLE dV01_y,
				    DOUBLE dV10_x, DOUBLE dV10_y,
				    DOUBLE dV1_x, DOUBLE dV1_y,
				    DOUBLE dV11_x, DOUBLE dV11_y,
				    DOUBLE dV20_x, DOUBLE dV20_y,
				    DOUBLE dV2_x, DOUBLE dV2_y,
				    DOUBLE dV21_x, DOUBLE dV21_y)
{
  *pdC_x = 0.0;
  *pdC_y = 0.0;
  *pdR = 0.0;

  if ( !APM2D::Counterclockwise(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y) )
    {
      return APM2D::icctNull;
    }

  ICC_VVV_MPFUNFilter iccVVV;

  return iccVVV.InscribedCircleCenterVVV(pdC_x, pdC_y, pdR, 
					 dV00_x, dV00_y,
					 dV0_x, dV0_y,
					 dV01_x, dV01_y,
					 dV10_x, dV10_y,
					 dV1_x, dV1_y,
					 dV11_x, dV11_y,
					 dV20_x, dV20_y,
					 dV2_x, dV2_y,
					 dV21_x, dV21_y);
}

//static 
APM2D::EdgeZoneType EdgeZoneVVVMPFUNFilter(DOUBLE dV0_x, DOUBLE dV0_y,
					   DOUBLE dV1_x, DOUBLE dV1_y,
					   DOUBLE dV2_x, DOUBLE dV2_y,
					   DOUBLE dE30_x, DOUBLE dE30_y,
					   DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::m_uiEdgeZoneVVVMPFUNFilter++;

  Interval valV0_x = dV0_x; 
  Interval valV0_y = dV0_y;
  Interval valV1_x = dV1_x;
  Interval valV1_y = dV1_y;
  Interval valV2_x = dV2_x;
  Interval valV2_y = dV2_y;
  Interval valE30_x = dE30_x;
  Interval valE30_y = dE30_y;
  Interval valE31_x = dE31_x;
  Interval valE31_y = dE31_y;

  Interval pppvalV[4][2][2]; // Endpoints
  //memset(pppvalV, 0, sizeof(pppvalV));
  //pppvalV[0][0][0] = 0.0;
  //pppvalV[0][0][1] = 0.0;
  pppvalV[1][0][0] = valV1_x - valV0_x;
  pppvalV[1][0][1] = valV1_y - valV0_y;
  pppvalV[2][0][0] = valV2_x - valV0_x;
  pppvalV[2][0][1] = valV2_y - valV0_y;
  pppvalV[3][0][0] = valE30_x - valV0_x;
  pppvalV[3][0][1] = valE30_y - valV0_y;
  pppvalV[3][1][0] = valE31_x - valV0_x;
  pppvalV[3][1][1] = valE31_y - valV0_y;
  
  Interval ppvalE[4][2];     // Edge vectors
  //memset(ppvalE, 0, sizeof(ppvalE));
  ppvalE[3][0] = valE31_x - valE30_x;
  ppvalE[3][1] = valE31_y - valE30_y;
  
  Interval valV1dotV1 = (pppvalV[1][0][0]*pppvalV[1][0][0]) + (pppvalV[1][0][1]*pppvalV[1][0][1]);
  Interval valV2dotV2 = (pppvalV[2][0][0]*pppvalV[2][0][0]) + (pppvalV[2][0][1]*pppvalV[2][0][1]);

  Interval valV1xV2 = (pppvalV[1][0][0]*pppvalV[2][0][1]) - (pppvalV[1][0][1]*pppvalV[2][0][0]);
  Interval valE3xV1 = (ppvalE[3][0]*pppvalV[1][0][1]) - (ppvalE[3][1]*pppvalV[1][0][0]);
  Interval valE3xV2 = (ppvalE[3][0]*pppvalV[2][0][1]) - (ppvalE[3][1]*pppvalV[2][0][0]);

  Interval pvalEdgeZone[2];
  int piEdgeZone[2];

  Interval val2V1xV2 = 2.0*valV1xV2;
  Interval valV1dotV1timesE3xV2minusV2dotV2timesE3xV1 = ( valV1dotV1*valE3xV2 - valV2dotV2*valE3xV1 );
  Interval valE3dotE30 = (ppvalE[3][0]*pppvalV[3][0][0]) + (ppvalE[3][1]*pppvalV[3][0][1]);
  pvalEdgeZone[0] = ( val2V1xV2*valE3dotE30 - valV1dotV1timesE3xV2minusV2dotV2timesE3xV1 );
  piEdgeZone[0] = ( pvalEdgeZone[0] >= 0.0 );
  if ( piEdgeZone[0] == 1 )
    {
      return APM2D::eztStart;
    }
  Interval valE3dotE31 = (ppvalE[3][0]*pppvalV[3][1][0]) + (ppvalE[3][1]*pppvalV[3][1][1]);
  pvalEdgeZone[1] = -( val2V1xV2*valE3dotE31 - valV1dotV1timesE3xV2minusV2dotV2timesE3xV1 );
  piEdgeZone[1] = ( pvalEdgeZone[1] >= 0.0 );
  if ( piEdgeZone[1] == 1 )
    {
      return APM2D::eztEnd;
    }
  
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( piEdgeZone[0] == -1 || piEdgeZone[0] == -1 )
    {
      APM2D::m_uiEdgeZoneVVVMPFUNFilter++;

      mp_real rV0_x = dV0_x; 
      mp_real rV0_y = dV0_y;
      mp_real rV1_x = dV1_x;
      mp_real rV1_y = dV1_y;
      mp_real rV2_x = dV2_x;
      mp_real rV2_y = dV2_y;
      mp_real rE30_x = dE30_x;
      mp_real rE30_y = dE30_y;
      mp_real rE31_x = dE31_x;
      mp_real rE31_y = dE31_y;

      mp_real ppprV[4][2][2]; // Endpoints
      //memset(ppprV, 0, sizeof(ppprV));
      //ppprV[0][0][0] = 0.0;
      //ppprV[0][0][1] = 0.0;
      ppprV[1][0][0] = rV1_x - rV0_x;
      ppprV[1][0][1] = rV1_y - rV0_y;
      ppprV[2][0][0] = rV2_x - rV0_x;
      ppprV[2][0][1] = rV2_y - rV0_y;
      ppprV[3][0][0] = rE30_x - rV0_x;
      ppprV[3][0][1] = rE30_y - rV0_y;
      ppprV[3][1][0] = rE31_x - rV0_x;
      ppprV[3][1][1] = rE31_y - rV0_y;
      
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[3][0] = rE31_x - rE30_x;
      pprE[3][1] = rE31_y - rE30_y;
  
      mp_real rV1dotV1 = (ppprV[1][0][0]*ppprV[1][0][0]) + (ppprV[1][0][1]*ppprV[1][0][1]);
      mp_real rV2dotV2 = (ppprV[2][0][0]*ppprV[2][0][0]) + (ppprV[2][0][1]*ppprV[2][0][1]);
      
      mp_real rV1xV2 = (ppprV[1][0][0]*ppprV[2][0][1]) - (ppprV[1][0][1]*ppprV[2][0][0]);
      mp_real rE3xV1 = (pprE[3][0]*ppprV[1][0][1]) - (pprE[3][1]*ppprV[1][0][0]);
      mp_real rE3xV2 = (pprE[3][0]*ppprV[2][0][1]) - (pprE[3][1]*ppprV[2][0][0]);

      mp_real prEdgeZone[2];

      mp_real r2V1xV2 = 2.0*rV1xV2;
      mp_real rV1dotV1timesE3xV2minusV2dotV2timesE3xV1 = ( rV1dotV1*rE3xV2 - rV2dotV2*rE3xV1 );

      if ( piEdgeZone[0] == -1 )
	{      
	  mp_real rE3dotE30 = (pprE[3][0]*ppprV[3][0][0]) + (pprE[3][1]*ppprV[3][0][1]);
	  prEdgeZone[0] = ( r2V1xV2*rE3dotE30 - rV1dotV1timesE3xV2minusV2dotV2timesE3xV1 );
	  if ( prEdgeZone[0] >= 0.0 )
	    {
	      return APM2D::eztStart;
	    }
	}

      if ( piEdgeZone[1] == -1 )
	{      
	  mp_real rE3dotE31 = (pprE[3][0]*ppprV[3][1][0]) + (pprE[3][1]*ppprV[3][1][1]);
	  prEdgeZone[1] = -( r2V1xV2*rE3dotE31 - rV1dotV1timesE3xV2minusV2dotV2timesE3xV1 );
	  if ( prEdgeZone[1] >= 0.0 )
	    {
	      return APM2D::eztEnd;
	    }
	}
    }

  return APM2D::eztMiddle;
}

//static 
APM2D::CircleTestType InCircleVVV_VMPFUNFilter(DOUBLE dV0_x, DOUBLE dV0_y,
					       DOUBLE dV1_x, DOUBLE dV1_y,
					       DOUBLE dV2_x, DOUBLE dV2_y,
					       DOUBLE dV3_x, DOUBLE dV3_y)
{
  APM2D::m_uiInCircleVVV_VMPFUNFilter++;

  Interval valV0_x = dV0_x;
  Interval valV0_y = dV0_y;
  Interval valV1_x = dV1_x;
  Interval valV1_y = dV1_y;
  Interval valV2_x = dV2_x;
  Interval valV2_y = dV2_y;
  Interval valV3_x = dV3_x;
  Interval valV3_y = dV3_y;

  Interval pppvalV[4][2][2]; // Endpoints
  //memset(pppvalV, 0, sizeof(pppvalV));
  //pppvalV[0][0][0] = 0.0;
  //pppvalV[0][0][1] = 0.0;
  pppvalV[1][0][0] = valV1_x - valV0_x;
  pppvalV[1][0][1] = valV1_y - valV0_y;
  pppvalV[2][0][0] = valV2_x - valV0_x;
  pppvalV[2][0][1] = valV2_y - valV0_y;
  pppvalV[3][0][0] = valV3_x - valV0_x;
  pppvalV[3][0][1] = valV3_y - valV0_y;
  
  Interval valV1dotV1 = (pppvalV[1][0][0]*pppvalV[1][0][0]) + (pppvalV[1][0][1]*pppvalV[1][0][1]);
  Interval valV2dotV2 = (pppvalV[2][0][0]*pppvalV[2][0][0]) + (pppvalV[2][0][1]*pppvalV[2][0][1]);
  Interval valV3dotV3 = (pppvalV[3][0][0]*pppvalV[3][0][0]) + (pppvalV[3][0][1]*pppvalV[3][0][1]);

  Interval valV1xV2 = (pppvalV[1][0][0]*pppvalV[2][0][1]) - (pppvalV[1][0][1]*pppvalV[2][0][0]);
  Interval valV2xV3 = (pppvalV[2][0][0]*pppvalV[3][0][1]) - (pppvalV[2][0][1]*pppvalV[3][0][0]);
  Interval valV3xV1 = (pppvalV[3][0][0]*pppvalV[1][0][1]) - (pppvalV[3][0][1]*pppvalV[1][0][0]);

  Interval valResult = ( valV1dotV1*valV2xV3 + valV2dotV2*valV3xV1 + valV3dotV3*valV1xV2 );
  
  int i = ( valResult < 0.0 );
  if ( i != -1 )
    {
      if ( i )
	{
	  return APM2D::cttInCircle;
	}
      else if ( valResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}

      return APM2D::cttOnCircle;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  APM2D::m_uiInCircleVVV_VMPFUN++;
  return InCircleVVV_VMPFUN(dV0_x, dV0_y,
			    dV1_x, dV1_y,
			    dV2_x, dV2_y,
			    dV3_x, dV3_y);
}

//static 
APM2D::CircleTestType InCircleVVV_EMPFUNFilter(DOUBLE dV0_x, DOUBLE dV0_y,
					       DOUBLE dV1_x, DOUBLE dV1_y,
					       DOUBLE dV2_x, DOUBLE dV2_y,
					       DOUBLE dE30_x, DOUBLE dE30_y,
					       DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::m_uiInCircleVVV_EMPFUNFilter++;

  Interval valV0_x = dV0_x;
  Interval valV0_y = dV0_y;
  Interval valV1_x = dV1_x;
  Interval valV1_y = dV1_y;
  Interval valV2_x = dV2_x;
  Interval valV2_y = dV2_y;
  Interval valE30_x = dE30_x;
  Interval valE30_y = dE30_y;
  Interval valE31_x = dE31_x;
  Interval valE31_y = dE31_y;

  Interval pppvalV[4][2][2]; // Endpoints
  //memset(pppvalV, 0, sizeof(pppvalV));
  pppvalV[0][0][0] = 0.0;
  pppvalV[0][0][1] = 0.0;
  pppvalV[1][0][0] = valV1_x - valV0_x;
  pppvalV[1][0][1] = valV1_y - valV0_y;
  pppvalV[2][0][0] = valV2_x - valV0_x;
  pppvalV[2][0][1] = valV2_y - valV0_y;
  pppvalV[3][0][0] = valE30_x - valV0_x;
  pppvalV[3][0][1] = valE30_y - valV0_y;
  pppvalV[3][1][0] = valE31_x - valV0_x;
  pppvalV[3][1][1] = valE31_y - valV0_y;
  
  Interval ppvalE[4][2];     // Edge vectors
  //memset(ppvalE, 0, sizeof(ppvalE));
  ppvalE[3][0] = valE31_x - valE30_x;
  ppvalE[3][1] = valE31_y - valE30_y;

  Interval pvalV1V2[2];
  pvalV1V2[0] = valV2_x - valV1_x;
  pvalV1V2[1] = valV2_y - valV1_y;
  
  Interval valV1dotV1 = (pppvalV[1][0][0]*pppvalV[1][0][0]) + (pppvalV[1][0][1]*pppvalV[1][0][1]);
  Interval valV2dotV2 = (pppvalV[2][0][0]*pppvalV[2][0][0]) + (pppvalV[2][0][1]*pppvalV[2][0][1]);
  Interval valE3dotE3 = (ppvalE[3][0]*ppvalE[3][0]) + (ppvalE[3][1]*ppvalE[3][1]);
  Interval valE3dotV1 = (ppvalE[3][0]*pppvalV[1][0][0]) + (ppvalE[3][1]*pppvalV[1][0][1]);
  Interval valE3dotV2 = (ppvalE[3][0]*pppvalV[2][0][0]) + (ppvalE[3][1]*pppvalV[2][0][1]);

  Interval valV1V2dotV1V2 = (pvalV1V2[0]*pvalV1V2[0]) + (pvalV1V2[1]*pvalV1V2[1]);

  Interval valV1xV2 = (pppvalV[1][0][0]*pppvalV[2][0][1]) - (pppvalV[1][0][1]*pppvalV[2][0][0]);
  Interval valE30xE31 = (pppvalV[3][0][0]*pppvalV[3][1][1]) - (pppvalV[3][0][1]*pppvalV[3][1][0]);

  Interval valResult = ( valV1dotV1*valV1dotV1*valE3dotV2*valE3dotV2 + valV2dotV2*valV2dotV2*valE3dotV1*valE3dotV1 
		      - valV1dotV1*valV2dotV2*(valV1V2dotV1V2*valE3dotE3 + 2.0*valE3dotV1*valE3dotV2)
		      + 4.0*valV1xV2*valE30xE31*(valV1xV2*valE30xE31 + 
					     valV2dotV2*valE3dotV1 - valV1dotV1*valE3dotV2) );
  
  int i = ( valResult < 0.0 );
  if ( i != -1 )
    {
      if ( i )
	{
	  return APM2D::cttInCircle;
	}
      else if ( valResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}

      return APM2D::cttOnCircle;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  APM2D::m_uiInCircleVVV_EMPFUN++;
  return InCircleVVV_EMPFUN(dV0_x, dV0_y,
			    dV1_x, dV1_y,
			    dV2_x, dV2_y,
			    dE30_x, dE30_y,
			    dE31_x, dE31_y);
}

//////////////////////////////////////////////////
// VVE

class ICC_VVE_MPFUNFilter
{
protected:
  enum InstructionPointer {ipStart, ipV0orV1onE2, ipE2xV1, ipE2xV1Sign, 
			   ipDiscR, ipDisc, 
			   ipEdgeZones, ipVertexZones, ipEnd};

public:
  ICC_VVE_MPFUNFilter();
  ~ICC_VVE_MPFUNFilter();

public:
  APM2D::InscribedCircleCenterType 
  InscribedCircleCenterVVE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			   DOUBLE dV00_x, DOUBLE dV00_y,
			   DOUBLE dV0_x, DOUBLE dV0_y,
			   DOUBLE dV01_x, DOUBLE dV01_y,
			   DOUBLE dV10_x, DOUBLE dV10_y,
			   DOUBLE dV1_x, DOUBLE dV1_y,
			   DOUBLE dV11_x, DOUBLE dV11_y,
			   DOUBLE dE20_x, DOUBLE dE20_y, 
			   DOUBLE dE21_x, DOUBLE dE21_y);
  APM2D::InscribedCircleCenterType 
  InscribedCircleCenterVVEForce(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dV00_x, DOUBLE dV00_y,
				DOUBLE dV0_x, DOUBLE dV0_y,
				DOUBLE dV01_x, DOUBLE dV01_y,
				DOUBLE dV10_x, DOUBLE dV10_y,
				DOUBLE dV1_x, DOUBLE dV1_y,
				DOUBLE dV11_x, DOUBLE dV11_y,
				DOUBLE dE20_x, DOUBLE dE20_y, 
				DOUBLE dE21_x, DOUBLE dE21_y);

protected:
  BOOL
  InscribedCircleCenterVVEMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dV00_x, DOUBLE dV00_y,
				DOUBLE dV0_x, DOUBLE dV0_y,
				DOUBLE dV01_x, DOUBLE dV01_y,
				DOUBLE dV10_x, DOUBLE dV10_y,
				DOUBLE dV1_x, DOUBLE dV1_y,
				DOUBLE dV11_x, DOUBLE dV11_y,
				DOUBLE dE20_x, DOUBLE dE20_y, 
				DOUBLE dE21_x, DOUBLE dE21_y,
				InstructionPointer eIPMax = ipEnd);

protected:
  InstructionPointer m_eIP;
  APM2D::InscribedCircleCenterType m_eResult;
  INT_32 m_iV0onE2, m_iV1onE2;
  INT_32 m_iE2xV1, m_iE2xV1Sign;
  INT_32 m_iDiscR;
  INT_32 m_iDisc;
  INT_32 m_iDen;

  INT_32 m_piEdgeZone[2];
  INT_32 m_ppiVertexZone[2][2];

  Interval m_valE2xV1;
  Interval m_valDiscR;
  Interval m_valDisc;

  mp_real m_rV00_x;
  mp_real m_rV00_y;
  mp_real m_rV0_x;
  mp_real m_rV0_y;
  mp_real m_rV01_x;
  mp_real m_rV01_y;
  mp_real m_rV10_x;
  mp_real m_rV10_y;
  mp_real m_rV1_x;
  mp_real m_rV1_y;
  mp_real m_rV11_x;
  mp_real m_rV11_y;
  mp_real m_rE20_x;
  mp_real m_rE20_y; 
  mp_real m_rE21_x;
  mp_real m_rE21_y;
  mp_real m_ppprV[3][2][2];
  mp_real m_ppprVN[2][2][2];
  mp_real m_prE[2];
  mp_real m_rE2xV1;
  mp_real m_rV1dotV1;
  mp_real m_rE2dotE2;
  mp_real m_rE20xE21;
  mp_real m_rDiscR;
  mp_real m_rDisc;
  mp_real m_rE2dotV1;
  mp_real m_rSqrtDisc;
  mp_real m_r2E20xE21timesE2dotV1;
};

ICC_VVE_MPFUNFilter::ICC_VVE_MPFUNFilter()
{
}

ICC_VVE_MPFUNFilter::~ICC_VVE_MPFUNFilter()
{
}

APM2D::InscribedCircleCenterType 
ICC_VVE_MPFUNFilter::InscribedCircleCenterVVE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					      DOUBLE dV00_x, DOUBLE dV00_y,
					      DOUBLE dV0_x, DOUBLE dV0_y,
					      DOUBLE dV01_x, DOUBLE dV01_y,
					      DOUBLE dV10_x, DOUBLE dV10_y,
					      DOUBLE dV1_x, DOUBLE dV1_y,
					      DOUBLE dV11_x, DOUBLE dV11_y,
					      DOUBLE dE20_x, DOUBLE dE20_y, 
					      DOUBLE dE21_x, DOUBLE dE21_y)
{
  APM2D::m_uiInscribedCircleCenterVVEMPFUNFilter++;

  *pdC_x = 0.0;
  *pdC_y = 0.0;
  *pdR = 0.0;

  m_eIP = ipStart;
  m_eResult = APM2D::icctNull;
  m_iV0onE2 = -1; 
  m_iV1onE2 = -1;
  m_iE2xV1 = -1; 
  m_iE2xV1Sign = -1;
  m_iDiscR = -1;
  m_iDisc = -1;
  m_iDen = -1;

  m_piEdgeZone[0] = -1;
  m_piEdgeZone[1] = -1;
  m_ppiVertexZone[0][0] = -1;  
  m_ppiVertexZone[0][1] = -1;  
  m_ppiVertexZone[1][0] = -1;  
  m_ppiVertexZone[1][1] = -1;  

  BOOL bDone;

  Interval valV00_x = dV00_x;
  Interval valV00_y = dV00_y;
  Interval valV0_x = dV0_x;
  Interval valV0_y = dV0_y;
  Interval valV01_x = dV01_x;
  Interval valV01_y = dV01_y;
  Interval valV10_x = dV10_x;
  Interval valV10_y = dV10_y;
  Interval valV1_x = dV1_x;
  Interval valV1_y = dV1_y;
  Interval valV11_x = dV11_x;
  Interval valV11_y = dV11_y;
  Interval valE20_x = dE20_x;
  Interval valE20_y = dE20_y; 
  Interval valE21_x = dE21_x;
  Interval valE21_y = dE21_y;
      
  //////////////////////////////////////////////////
  //  Check for V0 or V1 on E2
  
  Interval valV0onE2;
  SinTheta(valV0onE2, valV0_x, valV0_y, valE20_x, valE20_y, valE21_x, valE21_y);
  m_iV0onE2 = ( valV0onE2 <= 0.0 );
  if ( m_iV0onE2 == 1 )
    {
      // V0 is on E1 and it is not one of the endpoints
      return APM2D::icctNull;
    }
  
  Interval valV1onE2;
  SinTheta(valV1onE2, valV1_x, valV1_y, valE20_x, valE20_y, valE21_x, valE21_y);
  m_iV1onE2 = ( valV1onE2 <= 0.0 );
  if ( m_iV1onE2 == 1 )
    {
      // V1 is on E2 and it is not one of the endpoints
      return APM2D::icctNull;
    }

  if ( m_iV0onE2 == -1 || m_iV1onE2 == -1 )
    {
      bDone = InscribedCircleCenterVVEMPFUN(pdC_x, pdC_y, pdR, 
					    dV00_x, dV00_y,
					    dV0_x, dV0_y,
					    dV01_x, dV01_y,
					    dV10_x, dV10_y,
					    dV1_x, dV1_y,
					    dV11_x, dV11_y,
					    dE20_x, dE20_y, 
					    dE21_x, dE21_y,
					    ipV0orV1onE2);  
      if ( bDone )
	{
	  return m_eResult;
	}
    }
  
  //////////////////////////////////////////////////
  //  Check E2xV1

  Interval pppvalV[3][2][2];
  // Translate V0 to the origin
  pppvalV[1][0][0] = valV1_x - valV0_x;
  pppvalV[1][0][1] = valV1_y - valV0_y;
  pppvalV[2][0][0] = valE20_x - valV0_x;
  pppvalV[2][0][1] = valE20_y - valV0_y;
  pppvalV[2][1][0] = valE21_x - valV0_x;
  pppvalV[2][1][1] = valE21_y - valV0_y;
      
  Interval pppvalVN[2][2][2];
  // Vertex neighbor vectors
  pppvalVN[0][0][0] = valV0_x - valV00_x;
  pppvalVN[0][0][1] = valV0_y - valV00_y;
  pppvalVN[0][1][0] = valV0_x - valV01_x;
  pppvalVN[0][1][1] = valV0_y - valV01_y;
  pppvalVN[1][0][0] = valV1_x - valV10_x;
  pppvalVN[1][0][1] = valV1_y - valV10_y;
  pppvalVN[1][1][0] = valV1_x - valV11_x;
  pppvalVN[1][1][1] = valV1_y - valV11_y;
  
  Interval pvalE[2];
  pvalE[0] = valE21_x - valE20_x;
  pvalE[1] = valE21_y - valE20_y;
  
  Cross(m_valE2xV1, pvalE[0], pvalE[1], pppvalV[1][0][0], pppvalV[1][0][1]);

  m_iE2xV1 = ( m_valE2xV1 == 0.0 );
  if ( m_iE2xV1 == -1 )
    {
      bDone = InscribedCircleCenterVVEMPFUN(pdC_x, pdC_y, pdR, 
					    dV00_x, dV00_y,
					    dV0_x, dV0_y,
					    dV01_x, dV01_y,
					    dV10_x, dV10_y,
					    dV1_x, dV1_y,
					    dV11_x, dV11_y,
					    dE20_x, dE20_y, 
					    dE21_x, dE21_y,
					    ipE2xV1);  
      if ( bDone )
	{
	  return m_eResult;
	}
      if ( m_iE2xV1 == -1 )
	{
	  bDone = InscribedCircleCenterVVEMPFUN(pdC_x, pdC_y, pdR, 
						dV00_x, dV00_y,
						dV0_x, dV0_y,
						dV01_x, dV01_y,
						dV10_x, dV10_y,
						dV1_x, dV1_y,
						dV11_x, dV11_y,
						dE20_x, dE20_y, 
						dE21_x, dE21_y,
						ipEnd);  
	  ASSERT( bDone );
	  return m_eResult;
	}
    }

  if ( m_iE2xV1 == 1 )
    {
      m_iE2xV1Sign = 0;
    }
  else if ( m_valE2xV1 < 0.0 )
    {
      m_iE2xV1Sign = -1;
    }
  else
    {
      m_iE2xV1Sign = 1;
    }
  if ( m_iE2xV1Sign == 0 )
    {
      // The two vertices lie on a line parallel to the edge
      return InscribedCircleCenterVVEAntiparallelMPFUNFilter(pdC_x, pdC_y, pdR,
							     dV00_x, dV00_y, 
							     dV0_x, dV0_y, 
							     dV01_x, dV01_y, 
							     dV10_x, dV10_y, 
							     dV1_x, dV1_y, 
							     dV11_x, dV11_y, 
							     dE20_x, dE20_y, 
							     dE21_x, dE21_y);
    }

  //////////////////////////////////////////////////
  // Check DiscR

  Interval valV1dotV1;
  Dot(valV1dotV1, pppvalV[1][0][0], pppvalV[1][0][1], pppvalV[1][0][0], pppvalV[1][0][1]);
  Interval valE20xE21;
  Cross(valE20xE21, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][1][0], pppvalV[2][1][1]);
  
  m_valDiscR = valV1dotV1*valE20xE21*( valE20xE21 + m_valE2xV1 );
  m_iDiscR = ( m_valDiscR < 0.0 );
  if ( m_iDiscR == -1 )
    {
      bDone = InscribedCircleCenterVVEMPFUN(pdC_x, pdC_y, pdR, 
					    dV00_x, dV00_y,
					    dV0_x, dV0_y,
					    dV01_x, dV01_y,
					    dV10_x, dV10_y,
					    dV1_x, dV1_y,
					    dV11_x, dV11_y,
					    dE20_x, dE20_y, 
					    dE21_x, dE21_y,
					    ipDiscR);  
      if ( bDone )
	{
	  return m_eResult;
	}
      if ( m_iDiscR == -1 )
	{
	  bDone = InscribedCircleCenterVVEMPFUN(pdC_x, pdC_y, pdR, 
						dV00_x, dV00_y,
						dV0_x, dV0_y,
						dV01_x, dV01_y,
						dV10_x, dV10_y,
						dV1_x, dV1_y,
						dV11_x, dV11_y,
						dE20_x, dE20_y, 
						dE21_x, dE21_y,
						ipEnd);  
	  ASSERT( bDone );
	  return m_eResult;
	}
    }
  if ( m_iDiscR == 1 )
    {
      // Imaginary Solutions
      return APM2D::icctNull;
    }

  //////////////////////////////////////////////////
  // Check Disc

  Interval valE2dotE2;
  Dot(valE2dotE2, pvalE[0], pvalE[1], pvalE[0], pvalE[1]);

  m_valDisc = valE2dotE2*m_valDiscR;
  m_iDisc = ( m_valDisc < 0.0 );
  if ( m_iDisc == -1 )
    {
      ASSERT( FALSE );
      bDone = InscribedCircleCenterVVEMPFUN(pdC_x, pdC_y, pdR, 
					    dV00_x, dV00_y,
					    dV0_x, dV0_y,
					    dV01_x, dV01_y,
					    dV10_x, dV10_y,
					    dV1_x, dV1_y,
					    dV11_x, dV11_y,
					    dE20_x, dE20_y, 
					    dE21_x, dE21_y,
					    ipEnd);  
      if ( bDone )
	{
	  return m_eResult;
	}
      ASSERT( m_iDisc != -1 );
    }
  if ( m_iDisc == 1 ) 
    {
      // Imaginary Solutions
      return APM2D::icctNull;
    }
  
  //////////////////////////////////////////////////
  // Check the Edge Zone

  Interval valE2dotV1;
  Dot(valE2dotV1, pvalE[0], pvalE[1], pppvalV[1][0][0], pppvalV[1][0][1]);
  Interval valSqrtDisc;
  valSqrtDisc = sqrt(m_valDisc);

  Interval pvalEdgeZone[2];
  
  Interval valE2dotE20;
  Dot(valE2dotE20, pvalE[0], pvalE[1], pppvalV[2][0][0], pppvalV[2][0][1]);
  
  if ( m_iE2xV1Sign > 0 )
    {
      pvalEdgeZone[0] = -( m_valE2xV1*valE2dotE20 + valE20xE21*valE2dotV1 + valSqrtDisc );
    }
  else
    {
      pvalEdgeZone[0] = ( m_valE2xV1*valE2dotE20 + valE20xE21*valE2dotV1 + valSqrtDisc );
    }
  m_piEdgeZone[0] = ( pvalEdgeZone[0] < 0.0 );
  if ( m_piEdgeZone[0] == 1 )
    {
      return APM2D::icctNull;
    }
  
  Interval valE2dotE21;
  Dot(valE2dotE21, pvalE[0], pvalE[1], pppvalV[2][1][0], pppvalV[2][1][1]);
  
  if ( m_iE2xV1Sign > 0 )
    {
      pvalEdgeZone[1] = ( m_valE2xV1*valE2dotE21 + valE20xE21*valE2dotV1 + valSqrtDisc );
    }
  else
    {
      pvalEdgeZone[1] = -( m_valE2xV1*valE2dotE21 + valE20xE21*valE2dotV1 + valSqrtDisc );
    }
  m_piEdgeZone[1] = ( pvalEdgeZone[1] < 0.0 );
  if ( m_piEdgeZone[1] == 1 )
    {
      return APM2D::icctNull;
    }

  if ( m_piEdgeZone[0] == -1 || m_piEdgeZone[1] == -1 )
    {
      bDone = InscribedCircleCenterVVEMPFUN(pdC_x, pdC_y, pdR, 
					    dV00_x, dV00_y,
					    dV0_x, dV0_y,
					    dV01_x, dV01_y,
					    dV10_x, dV10_y,
					    dV1_x, dV1_y,
					    dV11_x, dV11_y,
					    dE20_x, dE20_y, 
					    dE21_x, dE21_y,
					    ipEdgeZones);  
      if ( bDone )
	{
	  return m_eResult;
	}
    }
  
  //////////////////////////////////////////////////
  // Check the Vertex Zones

  Interval val2E20xE21timesE2dotV1;
  val2E20xE21timesE2dotV1 = 2.0*valE20xE21*valE2dotV1;
  
  Interval ppvalVertexZone[2][2];
  
  Interval valE2xV00;
  Cross(valE2xV00, pvalE[0], pvalE[1], pppvalVN[0][0][0], pppvalVN[0][0][1]);
  Interval valV1xV00;
  Cross(valV1xV00, pppvalV[1][0][0], pppvalV[1][0][1], pppvalVN[0][0][0], pppvalVN[0][0][1]);
  ppvalVertexZone[0][0] = ( valV1dotV1*m_valE2xV1*valE2xV00 
			    + val2E20xE21timesE2dotV1*valV1xV00 
			    + 2.0*valV1xV00*valSqrtDisc );
  m_ppiVertexZone[0][0] = ( ppvalVertexZone[0][0] < 0.0 );
  if ( m_ppiVertexZone[0][0] == 1 ) 
    {
      return APM2D::icctNull;
    }

  Interval valE2xV01;
  Cross(valE2xV01, pvalE[0], pvalE[1], pppvalVN[0][1][0], pppvalVN[0][1][1]);
  Interval valV1xV01;
  Cross(valV1xV01, pppvalV[1][0][0], pppvalV[1][0][1], pppvalVN[0][1][0], pppvalVN[0][1][1]);
  ppvalVertexZone[0][1] = ( valV1dotV1*m_valE2xV1*valE2xV01 
			    + val2E20xE21timesE2dotV1*valV1xV01 
			    + 2.0*valV1xV01*valSqrtDisc );
  m_ppiVertexZone[0][1] = ( ppvalVertexZone[0][1] < 0.0 );
  if ( m_ppiVertexZone[0][1] == 1 ) 
    {
      return APM2D::icctNull;
    }

  Interval valE2xV10;
  Cross(valE2xV10, pvalE[0], pvalE[1], pppvalVN[1][0][0], pppvalVN[1][0][1]);
  Interval valV1xV10;
  Cross(valV1xV10, pppvalV[1][0][0], pppvalV[1][0][1], pppvalVN[1][0][0], pppvalVN[1][0][1]);      
  Interval valV1dotV10;
  Dot(valV1dotV10, pppvalV[1][0][0], pppvalV[1][0][1], pppvalVN[1][0][0], pppvalVN[1][0][1]);
  ppvalVertexZone[1][0] = ( valV1dotV1*m_valE2xV1*valE2xV10 
			    + val2E20xE21timesE2dotV1*valV1xV10 
			    + 2.0*valV1xV10*valSqrtDisc
			    - 2.0*m_valE2xV1*m_valE2xV1*valV1dotV10 );
  m_ppiVertexZone[1][0] = ( ppvalVertexZone[1][0] < 0.0 );
  if ( m_ppiVertexZone[1][0] == 1 ) 
    {
      return APM2D::icctNull;
    }

  Interval valE2xV11;
  Cross(valE2xV11, pvalE[0], pvalE[1], pppvalVN[1][1][0], pppvalVN[1][1][1]);
  Interval valV1xV11;
  Cross(valV1xV11, pppvalV[1][0][0], pppvalV[1][0][1], pppvalVN[1][1][0], pppvalVN[1][1][1]);
  Interval valV1dotV11;
  Dot(valV1dotV11, pppvalV[1][0][0], pppvalV[1][0][1], pppvalVN[1][1][0], pppvalVN[1][1][1]);
  ppvalVertexZone[1][1] = ( valV1dotV1*m_valE2xV1*valE2xV11 
			    + val2E20xE21timesE2dotV1*valV1xV11 
			    + 2.0*valV1xV11*valSqrtDisc
			    - 2.0*m_valE2xV1*m_valE2xV1*valV1dotV11 );
  m_ppiVertexZone[1][1] = ( ppvalVertexZone[1][1] < 0.0 );
  if ( m_ppiVertexZone[1][1] == 1 ) 
    {
      return APM2D::icctNull;
    }

  if ( m_ppiVertexZone[0][0] == -1 || 
       m_ppiVertexZone[0][1] == -1 || 
       m_ppiVertexZone[1][0] == -1 || 
       m_ppiVertexZone[1][1] == -1 ) 
    {      
      bDone = InscribedCircleCenterVVEMPFUN(pdC_x, pdC_y, pdR, 
					    dV00_x, dV00_y,
					    dV0_x, dV0_y,
					    dV01_x, dV01_y,
					    dV10_x, dV10_y,
					    dV1_x, dV1_y,
					    dV11_x, dV11_y,
					    dE20_x, dE20_y, 
					    dE21_x, dE21_y,
					    ipVertexZones);  
      if ( bDone )
	{
	  return m_eResult;
	}
    }
  
  //////////////////////////////////////////////////
  // Compute the center and radius

  Interval valDen = 2.0*m_valE2xV1*m_valE2xV1;
  
  m_iDen = ( valDen > 0.0 );
  ASSERT( m_iDen == 1 );
  if ( m_iDen == -1 )
    {
      bDone = InscribedCircleCenterVVEMPFUN(pdC_x, pdC_y, pdR, 
					    dV00_x, dV00_y,
					    dV0_x, dV0_y,
					    dV01_x, dV01_y,
					    dV10_x, dV10_y,
					    dV1_x, dV1_y,
					    dV11_x, dV11_y,
					    dE20_x, dE20_y, 
					    dE21_x, dE21_y,
					    ipEnd);  
      ASSERT( bDone );
      return m_eResult;
    }

  Interval pvalC[2], valR;
  pvalC[0] = ( -( valV1dotV1*m_valE2xV1*pvalE[1] + val2E20xE21timesE2dotV1*pppvalV[1][0][1] )
	     - 2.0 * pppvalV[1][0][1] * valSqrtDisc ) / valDen;
  pvalC[1] = ( ( valV1dotV1*m_valE2xV1*pvalE[0] + val2E20xE21timesE2dotV1*pppvalV[1][0][0] )
	     + 2.0*pppvalV[1][0][0]*valSqrtDisc ) / valDen;
  valR = ( valV1dotV1*sqrt(valE2dotE2) * (2.0*valE20xE21 + m_valE2xV1) 
	 + 2.0*valE2dotV1*sqrt(m_valDiscR) ) / valDen;
  
  // Translate the answer back out away from the origin
  pvalC[0] += valV0_x;
  pvalC[1] += valV0_y;
  
  *pdC_x = dble(pvalC[0]);
  *pdC_y = dble(pvalC[1]);
  *pdR = dble(valR);
  return APM2D::icctVVE;
}

APM2D::InscribedCircleCenterType 
ICC_VVE_MPFUNFilter::InscribedCircleCenterVVEForce(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
						   DOUBLE dV00_x, DOUBLE dV00_y,
						   DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dV01_x, DOUBLE dV01_y,
						   DOUBLE dV10_x, DOUBLE dV10_y,
						   DOUBLE dV1_x, DOUBLE dV1_y,
						   DOUBLE dV11_x, DOUBLE dV11_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y)
{
  *pdC_x = 0.0;
  *pdC_y = 0.0;
  *pdR = 0.0;
 
  m_eIP = ipStart;
  m_eResult = APM2D::icctNull;
  m_iV0onE2 = -1; 
  m_iV1onE2 = -1;
  m_iE2xV1 = -1; 
  m_iE2xV1Sign = -1;
  m_iDiscR = -1;
  m_iDisc = -1;
  m_iDen = -1;

  m_piEdgeZone[0] = -1;
  m_piEdgeZone[1] = -1;
  m_ppiVertexZone[0][0] = -1;  
  m_ppiVertexZone[0][1] = -1;  
  m_ppiVertexZone[1][0] = -1;  
  m_ppiVertexZone[1][1] = -1;  

  BOOL bDone;
  bDone = InscribedCircleCenterVVEMPFUN(pdC_x, pdC_y, pdR, 
					dV00_x, dV00_y,
					dV0_x, dV0_y,
					dV01_x, dV01_y,
					dV10_x, dV10_y,
					dV1_x, dV1_y,
					dV11_x, dV11_y,
					dE20_x, dE20_y, 
					dE21_x, dE21_y);
  if ( !bDone ) 
    {
      fprintf(stderr, "InscribedCircleCenterVVEMPFUN does not work!!!\n");
      exit(1);
    }
  return m_eResult;
}

BOOL
ICC_VVE_MPFUNFilter::InscribedCircleCenterVVEMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
						   DOUBLE dV00_x, DOUBLE dV00_y,
						   DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dV01_x, DOUBLE dV01_y,
						   DOUBLE dV10_x, DOUBLE dV10_y,
						   DOUBLE dV1_x, DOUBLE dV1_y,
						   DOUBLE dV11_x, DOUBLE dV11_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   ICC_VVE_MPFUNFilter::InstructionPointer eIPMax/* = ICC_VVE_MPFUNFilter::ipEnd*/)
{
  APM2D::m_uiInscribedCircleCenterVVEMPFUN++;

  switch ( m_eIP )
    {
    case ipStart:
      {
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	*pdC_x = 0.0;
	*pdC_y = 0.0;
	*pdR = 0.0;
	
	m_rV00_x = dV00_x;
	m_rV00_y = dV00_y;
	m_rV0_x = dV0_x;
	m_rV0_y = dV0_y;
	m_rV01_x = dV01_x;
	m_rV01_y = dV01_y;
	m_rV10_x = dV10_x;
	m_rV10_y = dV10_y;
	m_rV1_x = dV1_x;
	m_rV1_y = dV1_y;
	m_rV11_x = dV11_x;
	m_rV11_y = dV11_y;
	m_rE20_x = dE20_x;
	m_rE20_y = dE20_y; 
	m_rE21_x = dE21_x;
	m_rE21_y = dE21_y;
      
	//////////////////////////////////////////////////
	// Check for V0 or V1 on E2
      
	if ( m_iV0onE2 == -1 )
	  {
	    mp_real rV0onE2;
	    SinTheta(rV0onE2, m_rV0_x, m_rV0_y, m_rE20_x, m_rE20_y, m_rE21_x, m_rE21_y);
	    if ( rV0onE2 <= 0.0 )
	      {
		// V0 is on E1 and it is not one of the endpoints
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterVVEMPFUNFinal++;
		return TRUE;
	      }
	  }
  
	if ( m_iV1onE2 == -1 )
	  {
	    mp_real rV1onE2;
	    SinTheta(rV1onE2, m_rV1_x, m_rV1_y, m_rE20_x, m_rE20_y, m_rE21_x, m_rE21_y);
	    if ( rV1onE2 <= 0.0 )
	      {
		// V1 is on E2 and it is not one of the endpoints
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterVVEMPFUNFinal++;
		return TRUE;
	      }
	  }

      }
    case ipV0orV1onE2:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Check E2xV1
      
	// Translate V0 to the origin
	m_ppprV[1][0][0] = m_rV1_x - m_rV0_x;
	m_ppprV[1][0][1] = m_rV1_y - m_rV0_y;
	m_ppprV[2][0][0] = m_rE20_x - m_rV0_x;
	m_ppprV[2][0][1] = m_rE20_y - m_rV0_y;
	m_ppprV[2][1][0] = m_rE21_x - m_rV0_x;
	m_ppprV[2][1][1] = m_rE21_y - m_rV0_y;
      
	// Vertex neighbor vectors
	m_ppprVN[0][0][0] = m_rV0_x - m_rV00_x;
	m_ppprVN[0][0][1] = m_rV0_y - m_rV00_y;
	m_ppprVN[0][1][0] = m_rV0_x - m_rV01_x;
	m_ppprVN[0][1][1] = m_rV0_y - m_rV01_y;
	m_ppprVN[1][0][0] = m_rV1_x - m_rV10_x;
	m_ppprVN[1][0][1] = m_rV1_y - m_rV10_y;
	m_ppprVN[1][1][0] = m_rV1_x - m_rV11_x;
	m_ppprVN[1][1][1] = m_rV1_y - m_rV11_y;

	m_prE[0] = m_rE21_x - m_rE20_x;
	m_prE[1] = m_rE21_y - m_rE20_y;
  
	Cross(m_rE2xV1, m_prE[0], m_prE[1], m_ppprV[1][0][0], m_ppprV[1][0][1]);

	if ( eIPMax != ipEnd &&
	     m_iE2xV1 == -1 )
	  {
	    m_valE2xV1.ConservativeBound(dble(m_rE2xV1));
	    m_iE2xV1 = ( m_valE2xV1 == 0.0 );
	  }

      }
    case ipE2xV1:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	if ( m_iE2xV1 == -1 )
	  {
	    if ( m_rE2xV1 == 0.0 )
	      {
		m_iE2xV1Sign = 0;
	      }
	    else if ( m_rE2xV1 < 0.0 )
	      {
		m_iE2xV1Sign = -1;
	      }
	    else
	      {
		m_iE2xV1Sign = 1;
	      }
	    if ( m_iE2xV1Sign == 0 )
	      {
		// The two vertices lie on a line parallel to the edge
		m_eResult = InscribedCircleCenterVVEAntiparallelMPFUNFilter(pdC_x, pdC_y, pdR,
									    dV00_x, dV00_y, 
									    dV0_x, dV0_y, 
									    dV01_x, dV01_y, 
									    dV10_x, dV10_y, 
									    dV1_x, dV1_y, 
									    dV11_x, dV11_y, 
									    dE20_x, dE20_y, 
									    dE21_x, dE21_y);
		APM2D::m_uiInscribedCircleCenterVVEMPFUNFinal++;
		return TRUE;
	      }
	  }

      }
    case ipE2xV1Sign:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Check DiscR
      
	Dot(m_rV1dotV1, m_ppprV[1][0][0], m_ppprV[1][0][1], m_ppprV[1][0][0], m_ppprV[1][0][1]);
	Cross(m_rE20xE21, m_ppprV[2][0][0], m_ppprV[2][0][1], m_ppprV[2][1][0], m_ppprV[2][1][1]);
  
	m_rDiscR = m_rV1dotV1*m_rE20xE21*( m_rE20xE21 + m_rE2xV1 );
	if ( eIPMax != ipEnd &&
	     m_iDiscR == -1 )
	  {
	    m_valDiscR.ConservativeBound(dble(m_rDiscR));
	    m_iDiscR = ( m_valDiscR < 0.0 );
	  }

      }
    case ipDiscR:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	if ( m_iDiscR == -1 )
	  {
	    if ( m_rDiscR < 0.0 ) 
	      {
		// Imaginary Solutions
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterVVEMPFUNFinal++;
		return TRUE;
	      }
	  }

	//////////////////////////////////////////////////
	// Check Disc

	Dot(m_rE2dotE2, m_prE[0], m_prE[1], m_prE[0], m_prE[1]);      

	m_rDisc = m_rE2dotE2*m_rDiscR;
	if ( m_iDisc == -1 )
	  {
	    if ( m_rDisc < 0.0 ) 
	      {
		// Imaginary Solutions
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterVVEMPFUNFinal++;
		return TRUE;
	      }
	  }

      }
    case ipDisc:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Check the Edge Zone

	Dot(m_rE2dotV1, m_prE[0], m_prE[1], m_ppprV[1][0][0], m_ppprV[1][0][1]);
	m_rSqrtDisc = sqrt(m_rDisc);

	if ( m_piEdgeZone[0] == -1 || m_piEdgeZone[1] == -1 )
	  {
	    mp_real prEdgeZone[2];

	    if ( m_piEdgeZone[0] == -1 )
	      {
		mp_real rE2dotE20;
		Dot(rE2dotE20, m_prE[0], m_prE[1], m_ppprV[2][0][0], m_ppprV[2][0][1]);

		if ( m_iE2xV1Sign > 0 )
		  {
		    prEdgeZone[0] = -( m_rE2xV1*rE2dotE20 + m_rE20xE21*m_rE2dotV1 + m_rSqrtDisc );
		  }
		else
		  {
		    prEdgeZone[0] = ( m_rE2xV1*rE2dotE20 + m_rE20xE21*m_rE2dotV1 + m_rSqrtDisc );
		  }
		if ( prEdgeZone[0] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVVEMPFUNFinal++;
		    return TRUE;
		  }
	      }

	    if ( m_piEdgeZone[1] == -1 )
	      {
		mp_real rE2dotE21;
		Dot(rE2dotE21, m_prE[0], m_prE[1], m_ppprV[2][1][0], m_ppprV[2][1][1]);

		if ( m_iE2xV1Sign > 0 )
		  {
		    prEdgeZone[1] = ( m_rE2xV1*rE2dotE21 + m_rE20xE21*m_rE2dotV1 + m_rSqrtDisc );
		  }
		else
		  {
		    prEdgeZone[1] = -( m_rE2xV1*rE2dotE21 + m_rE20xE21*m_rE2dotV1 + m_rSqrtDisc );
		  }
		if ( prEdgeZone[1] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVVEMPFUNFinal++;
		    return TRUE;
		  }
	      }
	  }

      }
    case ipEdgeZones:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Check the Vertex Zones

	m_r2E20xE21timesE2dotV1 = 2.0*m_rE20xE21*m_rE2dotV1;
      
	if ( m_ppiVertexZone[0][0] == -1 || 
	     m_ppiVertexZone[0][1] == -1 || 
	     m_ppiVertexZone[1][0] == -1 || 
	     m_ppiVertexZone[1][1] == -1 ) 
	  {      
	    mp_real pprVertexZone[2][2];
      
	    if ( m_ppiVertexZone[0][0] == -1 ) 
	      {
		mp_real rE2xV00;
		Cross(rE2xV00, m_prE[0], m_prE[1], m_ppprVN[0][0][0], m_ppprVN[0][0][1]);
		mp_real rV1xV00;
		Cross(rV1xV00, m_ppprV[1][0][0], m_ppprV[1][0][1], m_ppprVN[0][0][0], m_ppprVN[0][0][1]);
		pprVertexZone[0][0] = ( m_rV1dotV1*m_rE2xV1*rE2xV00 
					+ m_r2E20xE21timesE2dotV1*rV1xV00 
					+ 2.0*rV1xV00*m_rSqrtDisc );
		if ( pprVertexZone[0][0] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVVEMPFUNFinal++;
		    return TRUE;
		  }
	      }
	    if ( m_ppiVertexZone[0][1] == -1 ) 
	      {
		mp_real rE2xV01;
		Cross(rE2xV01, m_prE[0], m_prE[1], m_ppprVN[0][1][0], m_ppprVN[0][1][1]);
		mp_real rV1xV01;
		Cross(rV1xV01, m_ppprV[1][0][0], m_ppprV[1][0][1], m_ppprVN[0][1][0], m_ppprVN[0][1][1]);
		pprVertexZone[0][1] = ( m_rV1dotV1*m_rE2xV1*rE2xV01 
					+ m_r2E20xE21timesE2dotV1*rV1xV01 
					+ 2.0*rV1xV01*m_rSqrtDisc );
		if ( pprVertexZone[0][1] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVVEMPFUNFinal++;
		    return TRUE;
		  }
	      }
	    if ( m_ppiVertexZone[1][0] == -1 ) 
	      {
		mp_real rE2xV10;
		Cross(rE2xV10, m_prE[0], m_prE[1], m_ppprVN[1][0][0], m_ppprVN[1][0][1]);
		mp_real rV1xV10;
		Cross(rV1xV10, m_ppprV[1][0][0], m_ppprV[1][0][1], m_ppprVN[1][0][0], m_ppprVN[1][0][1]);      
		mp_real rV1dotV10;
		Dot(rV1dotV10, m_ppprV[1][0][0], m_ppprV[1][0][1], m_ppprVN[1][0][0], m_ppprVN[1][0][1]);
		pprVertexZone[1][0] = ( m_rV1dotV1*m_rE2xV1*rE2xV10 
					+ m_r2E20xE21timesE2dotV1*rV1xV10 
					+ 2.0*rV1xV10*m_rSqrtDisc
					- 2.0*m_rE2xV1*m_rE2xV1*rV1dotV10 );
		if ( pprVertexZone[1][0] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVVEMPFUNFinal++;
		    return TRUE;
		  }
	      }
	    if ( m_ppiVertexZone[1][1] == -1 ) 
	      {
		mp_real rE2xV11;
		Cross(rE2xV11, m_prE[0], m_prE[1], m_ppprVN[1][1][0], m_ppprVN[1][1][1]);
		mp_real rV1xV11;
		Cross(rV1xV11, m_ppprV[1][0][0], m_ppprV[1][0][1], m_ppprVN[1][1][0], m_ppprVN[1][1][1]);
		mp_real rV1dotV11;
		Dot(rV1dotV11, m_ppprV[1][0][0], m_ppprV[1][0][1], m_ppprVN[1][1][0], m_ppprVN[1][1][1]);
		pprVertexZone[1][1] = ( m_rV1dotV1*m_rE2xV1*rE2xV11 
					+ m_r2E20xE21timesE2dotV1*rV1xV11 
					+ 2.0*rV1xV11*m_rSqrtDisc
					- 2.0*m_rE2xV1*m_rE2xV1*rV1dotV11 );
		if ( pprVertexZone[1][1] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVVEMPFUNFinal++;
		    return TRUE;
		  }
	      }
	  }

      }
    case ipVertexZones:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Compute the center and radius
      
	mp_real rDen = 2.0*m_rE2xV1*m_rE2xV1;

	mp_real prC[2], rR;
	prC[0] = ( -( m_rV1dotV1*m_rE2xV1*m_prE[1] + m_r2E20xE21timesE2dotV1*m_ppprV[1][0][1] )
		   - 2.0 * m_ppprV[1][0][1] * m_rSqrtDisc ) / rDen;
	prC[1] = ( ( m_rV1dotV1*m_rE2xV1*m_prE[0] + m_r2E20xE21timesE2dotV1*m_ppprV[1][0][0] )
		   + 2.0*m_ppprV[1][0][0]*m_rSqrtDisc ) / rDen;
	rR = ( m_rV1dotV1*sqrt(m_rE2dotE2) * (2.0*m_rE20xE21 + m_rE2xV1) 
	       + 2.0*m_rE2dotV1*sqrt(m_rDiscR) ) / rDen;

	// Translate the answer back out away from the origin
	prC[0] += m_rV0_x;
	prC[1] += m_rV0_y;

	*pdC_x = dble(prC[0]);
	*pdC_y = dble(prC[1]);
	*pdR = dble(rR);
	m_eResult = APM2D::icctVVE;
      }
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  APM2D::m_uiInscribedCircleCenterVVEMPFUNFinal++;
  return TRUE;
}

//static 
APM2D::InscribedCircleCenterType 
InscribedCircleCenterVVEMPFUNFilter(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				    DOUBLE dV00_x, DOUBLE dV00_y,
				    DOUBLE dV0_x, DOUBLE dV0_y,
				    DOUBLE dV01_x, DOUBLE dV01_y,
				    DOUBLE dV10_x, DOUBLE dV10_y,
				    DOUBLE dV1_x, DOUBLE dV1_y,
				    DOUBLE dV11_x, DOUBLE dV11_y,
				    DOUBLE dE20_x, DOUBLE dE20_y, 
				    DOUBLE dE21_x, DOUBLE dE21_y)
{
  ICC_VVE_MPFUNFilter iccVVE;

  return iccVVE.InscribedCircleCenterVVE(pdC_x, pdC_y, pdR, 
					 dV00_x, dV00_y,
					 dV0_x, dV0_y,
					 dV01_x, dV01_y,
					 dV10_x, dV10_y,
					 dV1_x, dV1_y,
					 dV11_x, dV11_y,
					 dE20_x, dE20_y, 
					 dE21_x, dE21_y);
}

//static 
APM2D::EdgeZoneType EdgeZoneVVEMPFUNFilter(DOUBLE dV0_x, DOUBLE dV0_y,
					   DOUBLE dV1_x, DOUBLE dV1_y,
					   DOUBLE dE20_x, DOUBLE dE20_y, 
					   DOUBLE dE21_x, DOUBLE dE21_y,
					   DOUBLE dE30_x, DOUBLE dE30_y,
					   DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::m_uiEdgeZoneVVEMPFUNFilter++;

  INT_32 iE2xV1;
  INT_32 piEdgeZone[2];
  piEdgeZone[0] = -1;
  piEdgeZone[1] = -1;

  switch ( 0 )
    {
    default:
      Interval valV0_x = dV0_x;
      Interval valV0_y = dV0_y;
      Interval valV1_x = dV1_x;
      Interval valV1_y = dV1_y;
      Interval valE20_x = dE20_x;
      Interval valE20_y = dE20_y; 
      Interval valE21_x = dE21_x;
      Interval valE21_y = dE21_y;
      Interval valE30_x = dE30_x;
      Interval valE30_y = dE30_y;
      Interval valE31_x = dE31_x;
      Interval valE31_y = dE31_y;

      Interval pppvalV[4][2][2]; // Endpoints
      //memset(pppvalV, 0, sizeof(pppvalV));
      pppvalV[0][0][0] = 0.0;
      pppvalV[0][0][1] = 0.0;
      pppvalV[1][0][0] = valV1_x - valV0_x;
      pppvalV[1][0][1] = valV1_y - valV0_y;
      pppvalV[2][0][0] = valE20_x - valV0_x;
      pppvalV[2][0][1] = valE20_y - valV0_y;
      pppvalV[2][1][0] = valE21_x - valV0_x;
      pppvalV[2][1][1] = valE21_y - valV0_y;
      pppvalV[3][0][0] = valE30_x - valV0_x;
      pppvalV[3][0][1] = valE30_y - valV0_y;
      pppvalV[3][1][0] = valE31_x - valV0_x;
      pppvalV[3][1][1] = valE31_y - valV0_y;
  
      Interval ppvalE[4][2];     // Edge vectors
      //memset(ppvalE, 0, sizeof(ppvalE));
      ppvalE[2][0] = valE21_x - valE20_x;
      ppvalE[2][1] = valE21_y - valE20_y;
      ppvalE[3][0] = valE31_x - valE30_x;
      ppvalE[3][1] = valE31_y - valE30_y;
  
      Interval valE2xV1;
      Cross(valE2xV1, ppvalE[2][0], ppvalE[2][1], pppvalV[1][0][0], pppvalV[1][0][1]);
      iE2xV1 = ( valE2xV1 == 0.0 );
      if ( iE2xV1 == -1 )
	{
	  break;
	}
      if ( iE2xV1 == 1 )
	{
	  // The two vertices lie on a line parallel to the edge
	  return EdgeZoneVVEAntiparallelMPFUNFilter(dV0_x, dV0_y, 
						    dV1_x, dV1_y, 
						    dE20_x, dE20_y, dE21_x, dE21_y,
						    dE30_x, dE30_y, dE31_x, dE31_y);
	}
      
      Interval valV1dotV1;
      Dot(valV1dotV1, pppvalV[1][0][0], pppvalV[1][0][1], pppvalV[1][0][0], pppvalV[1][0][1]);
      Interval valV1dotE2;
      Dot(valV1dotE2, pppvalV[1][0][0], pppvalV[1][0][1], ppvalE[2][0], ppvalE[2][1]);
      Interval valE2dotE2;
      Dot(valE2dotE2, ppvalE[2][0], ppvalE[2][1], ppvalE[2][0], ppvalE[2][1]);
      Interval valE3dotE3;
      Dot(valE3dotE3, ppvalE[3][0], ppvalE[3][1], ppvalE[3][0], ppvalE[3][1]);
      
      Interval valE20xE21;
      Cross(valE20xE21, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][1][0], pppvalV[2][1][1]);

      Interval valDisc;
      valDisc = valV1dotV1*valE2dotE2*valE20xE21*( valE20xE21 + valE2xV1 );
      INT_32 iDisc;
      iDisc = ( valDisc >= 0.0 );
      if ( iDisc == -1 )
	{
	  break;
	}
      ASSERT( iDisc == 1 );
      Interval valSqrtDisc = sqrt(valDisc);
  
      Interval valDen;  
      valDen = 2.0*valE2xV1*valE2xV1; 
      INT_32 iDen;
      iDen = ( valDen != 0.0 );
      if ( iDen == -1 )
	{
	  break;
	}
      ASSERT( iDen );

      Interval valE2xE3;
      Cross(valE2xE3, ppvalE[2][0], ppvalE[2][1], ppvalE[3][0], ppvalE[3][1]);
  
      Interval valE2xE3timesV1dotV1 = valE2xE3*valV1dotV1;
	  
      Interval valE3xV1;
      Cross(valE3xV1, ppvalE[3][0], ppvalE[3][1], pppvalV[1][0][0], pppvalV[1][0][1]);
  
      Interval valCoef = 2.0*valE3xV1*( valE20xE21*valV1dotE2 + valSqrtDisc );
  
      Interval pvalEdgeZone[2];
  
      Interval valE3dotE30;
      Dot(valE3dotE30, ppvalE[3][0], ppvalE[3][1], pppvalV[3][0][0], pppvalV[3][0][1]);
      pvalEdgeZone[0] = ( valE2xV1*( 2.0*valE2xV1*valE3dotE30 - valE2xE3timesV1dotV1 )
			  + valCoef );
      piEdgeZone[0] = ( pvalEdgeZone[0] >= 0.0 );
      if ( piEdgeZone[0] == 1 )  
	{
	  return APM2D::eztStart;
	}

      Interval valE3dotE31;
      Dot(valE3dotE31, ppvalE[3][0], ppvalE[3][1], pppvalV[3][1][0], pppvalV[3][1][1]);
      pvalEdgeZone[1] = -( valE2xV1*( 2.0*valE2xV1*valE3dotE31 - valE2xE3timesV1dotV1 )
			   + valCoef );
      piEdgeZone[1] = ( pvalEdgeZone[1] >= 0.0 );
      if ( piEdgeZone[1] == 1 )
	{
	  return APM2D::eztEnd;
	}
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( piEdgeZone[0] == -1 || piEdgeZone[1] == -1 )
    {
      APM2D::m_uiEdgeZoneVVEMPFUN++;

      mp_real rV0_x = dV0_x;
      mp_real rV0_y = dV0_y;
      mp_real rV1_x = dV1_x;
      mp_real rV1_y = dV1_y;
      mp_real rE20_x = dE20_x;
      mp_real rE20_y = dE20_y; 
      mp_real rE21_x = dE21_x;
      mp_real rE21_y = dE21_y;
      mp_real rE30_x = dE30_x;
      mp_real rE30_y = dE30_y;
      mp_real rE31_x = dE31_x;
      mp_real rE31_y = dE31_y;

      mp_real ppprV[4][2][2]; // Endpoints
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = 0.0;
      ppprV[0][0][1] = 0.0;
      ppprV[1][0][0] = rV1_x - rV0_x;
      ppprV[1][0][1] = rV1_y - rV0_y;
      ppprV[2][0][0] = rE20_x - rV0_x;
      ppprV[2][0][1] = rE20_y - rV0_y;
      ppprV[2][1][0] = rE21_x - rV0_x;
      ppprV[2][1][1] = rE21_y - rV0_y;
      ppprV[3][0][0] = rE30_x - rV0_x;
      ppprV[3][0][1] = rE30_y - rV0_y;
      ppprV[3][1][0] = rE31_x - rV0_x;
      ppprV[3][1][1] = rE31_y - rV0_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[2][0] = rE21_x - rE20_x;
      pprE[2][1] = rE21_y - rE20_y;
      pprE[3][0] = rE31_x - rE30_x;
      pprE[3][1] = rE31_y - rE30_y;
  
      mp_real rE2xV1;
      Cross(rE2xV1, pprE[2][0], pprE[2][1], ppprV[1][0][0], ppprV[1][0][1]);
      if ( iE2xV1 == -1 )
	{
	  if ( rE2xV1 == 0.0 )
	    {
	      // The two vertices lie on a line parallel to the edge
	      return EdgeZoneVVEAntiparallelMPFUNFilter(dV0_x, dV0_y, 
							dV1_x, dV1_y, 
							dE20_x, dE20_y, dE21_x, dE21_y,
							dE30_x, dE30_y, dE31_x, dE31_y);
	    }
	}

      mp_real rV1dotV1;
      Dot(rV1dotV1, ppprV[1][0][0], ppprV[1][0][1], ppprV[1][0][0], ppprV[1][0][1]);
      mp_real rV1dotE2;
      Dot(rV1dotE2, ppprV[1][0][0], ppprV[1][0][1], pprE[2][0], pprE[2][1]);
      mp_real rE2dotE2;
      Dot(rE2dotE2, pprE[2][0], pprE[2][1], pprE[2][0], pprE[2][1]);
      mp_real rE3dotE3;
      Dot(rE3dotE3, pprE[3][0], pprE[3][1], pprE[3][0], pprE[3][1]);
      
      mp_real rE20xE21;
      Cross(rE20xE21, ppprV[2][0][0], ppprV[2][0][1], ppprV[2][1][0], ppprV[2][1][1]);

      mp_real rDisc;
      rDisc = rV1dotV1*rE2dotE2*rE20xE21*( rE20xE21 + rE2xV1 );
      ASSERT( rDisc >= 0.0 ); 
      mp_real rSqrtDisc = sqrt(rDisc);

      mp_real rDen;  
      rDen = 2.0*rE2xV1*rE2xV1; 
      ASSERT( rDen != 0.0 );

      mp_real rE2xE3;
      Cross(rE2xE3, pprE[2][0], pprE[2][1], pprE[3][0], pprE[3][1]);

      mp_real rE2xE3timesV1dotV1 = rE2xE3*rV1dotV1;

      mp_real rE3xV1;
      Cross(rE3xV1, pprE[3][0], pprE[3][1], ppprV[1][0][0], ppprV[1][0][1]);

      mp_real rCoef = 2.0*rE3xV1*( rE20xE21*rV1dotE2 + rSqrtDisc );
      
      mp_real prEdgeZone[2];

      if ( piEdgeZone[0] == -1 )
	{
	  mp_real rE3dotE30;
	  Dot(rE3dotE30, pprE[3][0], pprE[3][1], ppprV[3][0][0], ppprV[3][0][1]);
	  prEdgeZone[0] = ( rE2xV1*( 2.0*rE2xV1*rE3dotE30 - rE2xE3timesV1dotV1 )
			    + rCoef );
	  if ( prEdgeZone[0] >= 0.0 )
	    {
	      return APM2D::eztStart;
	    }
	}

      if ( piEdgeZone[1] == -1 )
	{
	  mp_real rE3dotE31;
	  Dot(rE3dotE31, pprE[3][0], pprE[3][1], ppprV[3][1][0], ppprV[3][1][1]);
	  prEdgeZone[1] = -( rE2xV1*( 2.0*rE2xV1*rE3dotE31 - rE2xE3timesV1dotV1 )
			     + rCoef );
	  if ( prEdgeZone[1] >= 0.0 )
	    {
	      return APM2D::eztEnd;
	    }
	}
    }

  return  APM2D::eztMiddle;
}

//static 
APM2D::CircleTestType InCircleVVE_VMPFUNFilter(DOUBLE dV0_x, DOUBLE dV0_y,
					       DOUBLE dV1_x, DOUBLE dV1_y,
					       DOUBLE dE20_x, DOUBLE dE20_y, 
					       DOUBLE dE21_x, DOUBLE dE21_y,
					       DOUBLE dV3_x, DOUBLE dV3_y)
{
  APM2D::m_uiInCircleVVE_VMPFUNFilter++;

  INT_32 iResult;
  INT_32 iE2xV1;

  iResult = -1;

  switch ( 0 )
    {
    default:
      Interval valV0_x = dV0_x;
      Interval valV0_y = dV0_y;
      Interval valV1_x = dV1_x;
      Interval valV1_y = dV1_y;
      Interval valE20_x = dE20_x;
      Interval valE20_y = dE20_y; 
      Interval valE21_x = dE21_x;
      Interval valE21_y = dE21_y;
      Interval valV3_x = dV3_x;
      Interval valV3_y = dV3_y;

      Interval pppvalV[4][2][2]; // Endpoints
      //memset(pppvalV, 0, sizeof(pppvalV));
      pppvalV[0][0][0] = 0.0;
      pppvalV[0][0][1] = 0.0;
      pppvalV[1][0][0] = valV1_x - valV0_x;
      pppvalV[1][0][1] = valV1_y - valV0_y;
      pppvalV[2][0][0] = valE20_x - valV0_x;
      pppvalV[2][0][1] = valE20_y - valV0_y;
      pppvalV[2][1][0] = valE21_x - valV0_x;
      pppvalV[2][1][1] = valE21_y - valV0_y;
      pppvalV[3][0][0] = valV3_x - valV0_x;
      pppvalV[3][0][1] = valV3_y - valV0_y;
  
      Interval ppvalE[4][2];     // Edge vectors
      //memset(ppvalE, 0, sizeof(ppvalE));
      ppvalE[2][0] = valE21_x - valE20_x;
      ppvalE[2][1] = valE21_y - valE20_y;
      
      Interval valV1dotV1;
      Dot(valV1dotV1, pppvalV[1][0][0], pppvalV[1][0][1], pppvalV[1][0][0], pppvalV[1][0][1]);
      Interval valE2dotE2;
      Dot(valE2dotE2, ppvalE[2][0], ppvalE[2][1], ppvalE[2][0], ppvalE[2][1]);
      Interval valV3dotV3;
      Dot(valV3dotV3, pppvalV[3][0][0], pppvalV[3][0][1], pppvalV[3][0][0], pppvalV[3][0][1]);
      Interval valV1dotE2;
      Dot(valV1dotE2, pppvalV[1][0][0], pppvalV[1][0][1], ppvalE[2][0], ppvalE[2][1]);

      Interval valE2xV1;
      Cross(valE2xV1, ppvalE[2][0], ppvalE[2][1], pppvalV[1][0][0], pppvalV[1][0][1]);
      iE2xV1 = ( valE2xV1 == 0.0 );
      if ( iE2xV1 == -1 )
	{
	  break;
	}
      if ( iE2xV1 == 1 )
	{
	  // The two vertices lie on a line parallel to the edge
	  return InCircleVVE_VAntiparallelMPFUNFilter(dV0_x, dV0_y, 
						      dV1_x, dV1_y, 
						      dE20_x, dE20_y, dE21_x, dE21_y,
						      dV3_x, dV3_y);
	}

      Interval valE2xV3;
      Cross(valE2xV3, ppvalE[2][0], ppvalE[2][1], pppvalV[3][0][0], pppvalV[3][0][1]);
      Interval valV1xV3;
      Cross(valV1xV3, pppvalV[1][0][0], pppvalV[1][0][1], pppvalV[3][0][0], pppvalV[3][0][1]);
      Interval valE20xE21;
      Cross(valE20xE21, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][1][0], pppvalV[2][1][1]);
      
      Interval valDisc;
      valDisc = valV1dotV1*valE2dotE2*valE20xE21*( valE20xE21 + valE2xV1 );
      INT_32 iDisc;
      iDisc = ( valDisc >= 0.0 ); 
      if ( iDisc == -1 )
	{
	  break;
	}
      ASSERT( iDisc == 1 ); 
      Interval valSqrtDisc = sqrt(valDisc);

      Interval valResult;
      valResult = ( valE2xV1*( valV3dotV3*valE2xV1 - valV1dotV1*valE2xV3 ) 
		    - 2.0*valV1xV3*( valE20xE21*valV1dotE2 + valSqrtDisc ) );
  
      iResult = ( valResult < 0.0 );
      if ( iResult == -1 )
	{
	  break;
	}
      if ( iResult == 1 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( valResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
      return APM2D::cttOnCircle;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( iResult == -1 )
    {
      APM2D::m_uiInCircleVVE_VMPFUN++;

      mp_real rV0_x = dV0_x;
      mp_real rV0_y = dV0_y;
      mp_real rV1_x = dV1_x;
      mp_real rV1_y = dV1_y;
      mp_real rE20_x = dE20_x;
      mp_real rE20_y = dE20_y; 
      mp_real rE21_x = dE21_x;
      mp_real rE21_y = dE21_y;
      mp_real rV3_x = dV3_x;
      mp_real rV3_y = dV3_y;
      
      mp_real ppprV[4][2][2]; // Endpoints
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = 0.0;
      ppprV[0][0][1] = 0.0;
      ppprV[1][0][0] = rV1_x - rV0_x;
      ppprV[1][0][1] = rV1_y - rV0_y;
      ppprV[2][0][0] = rE20_x - rV0_x;
      ppprV[2][0][1] = rE20_y - rV0_y;
      ppprV[2][1][0] = rE21_x - rV0_x;
      ppprV[2][1][1] = rE21_y - rV0_y;
      ppprV[3][0][0] = rV3_x - rV0_x;
      ppprV[3][0][1] = rV3_y - rV0_y;
      
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[2][0] = rE21_x - rE20_x;
      pprE[2][1] = rE21_y - rE20_y;

      mp_real rV1dotV1;
      Dot(rV1dotV1, ppprV[1][0][0], ppprV[1][0][1], ppprV[1][0][0], ppprV[1][0][1]);
      mp_real rE2dotE2;
      Dot(rE2dotE2, pprE[2][0], pprE[2][1], pprE[2][0], pprE[2][1]);
      mp_real rV3dotV3;
      Dot(rV3dotV3, ppprV[3][0][0], ppprV[3][0][1], ppprV[3][0][0], ppprV[3][0][1]);
      mp_real rV1dotE2;
      Dot(rV1dotE2, ppprV[1][0][0], ppprV[1][0][1], pprE[2][0], pprE[2][1]);
      
      mp_real rE2xV1;
      Cross(rE2xV1, pprE[2][0], pprE[2][1], ppprV[1][0][0], ppprV[1][0][1]);
      if ( iE2xV1 == -1 )
	{
	  if ( rE2xV1 == 0.0 )
	    {
	      // The two vertices lie on a line parallel to the edge
	      return InCircleVVE_VAntiparallelMPFUNFilter(dV0_x, dV0_y, 
							  dV1_x, dV1_y, 
							  dE20_x, dE20_y, dE21_x, dE21_y,
							  dV3_x, dV3_y);
	    }
	}

      mp_real rE2xV3;
      Cross(rE2xV3, pprE[2][0], pprE[2][1], ppprV[3][0][0], ppprV[3][0][1]);
      mp_real rV1xV3;
      Cross(rV1xV3, ppprV[1][0][0], ppprV[1][0][1], ppprV[3][0][0], ppprV[3][0][1]);
      mp_real rE20xE21;
      Cross(rE20xE21, ppprV[2][0][0], ppprV[2][0][1], ppprV[2][1][0], ppprV[2][1][1]);
      
      mp_real rDisc;
      rDisc = rV1dotV1*rE2dotE2*rE20xE21*( rE20xE21 + rE2xV1 );
      ASSERT( rDisc >= 0.0 ); 
      mp_real rSqrtDisc = sqrt(rDisc);

      mp_real rResult;
      rResult = ( rE2xV1*( rV3dotV3*rE2xV1 - rV1dotV1*rE2xV3 ) 
		  - 2.0*rV1xV3*( rE20xE21*rV1dotE2 + rSqrtDisc ) );
      
      if ( rResult < 0.0 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( rResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
    }

  return APM2D::cttOnCircle;
}

//static 
APM2D::CircleTestType InCircleVVE_EMPFUNFilter(DOUBLE dV0_x, DOUBLE dV0_y,
					       DOUBLE dV1_x, DOUBLE dV1_y,
					       DOUBLE dE20_x, DOUBLE dE20_y, 
					       DOUBLE dE21_x, DOUBLE dE21_y,
					       DOUBLE dE30_x, DOUBLE dE30_y,
					       DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::m_uiInCircleVVE_EMPFUNFilter++;

  INT_32 iResult;
  INT_32 iE2xV1;

  iResult = -1;

  switch ( 0 )
    {
    default:
      Interval valV0_x = dV0_x;
      Interval valV0_y = dV0_y;
      Interval valV1_x = dV1_x;
      Interval valV1_y = dV1_y;
      Interval valE20_x = dE20_x;
      Interval valE20_y = dE20_y; 
      Interval valE21_x = dE21_x;
      Interval valE21_y = dE21_y;
      Interval valE30_x = dE30_x;
      Interval valE30_y = dE30_y;
      Interval valE31_x = dE31_x;
      Interval valE31_y = dE31_y;
  
      Interval pppvalV[4][2][2]; // Endpoints
      //memset(pppvalV, 0, sizeof(pppvalV));
      pppvalV[0][0][0] = 0.0;
      pppvalV[0][0][1] = 0.0;
      pppvalV[1][0][0] = valV1_x - valV0_x;
      pppvalV[1][0][1] = valV1_y - valV0_y;
      pppvalV[2][0][0] = valE20_x - valV0_x;
      pppvalV[2][0][1] = valE20_y - valV0_y;
      pppvalV[2][1][0] = valE21_x - valV0_x;
      pppvalV[2][1][1] = valE21_y - valV0_y;
      pppvalV[3][0][0] = valE30_x - valV0_x;
      pppvalV[3][0][1] = valE30_y - valV0_y;
      pppvalV[3][1][0] = valE31_x - valV0_x;
      pppvalV[3][1][1] = valE31_y - valV0_y;
  
      Interval ppvalE[4][2];     // Edge vectors
      //memset(ppvalE, 0, sizeof(ppvalE));
      ppvalE[2][0] = valE21_x - valE20_x;
      ppvalE[2][1] = valE21_y - valE20_y;
      ppvalE[3][0] = valE31_x - valE30_x;
      ppvalE[3][1] = valE31_y - valE30_y;

      Interval valV1dotV1;
      Dot(valV1dotV1, pppvalV[1][0][0], pppvalV[1][0][1], pppvalV[1][0][0], pppvalV[1][0][1]);
      Interval valE2dotE2;
      Dot(valE2dotE2, ppvalE[2][0], ppvalE[2][1], ppvalE[2][0], ppvalE[2][1]);
      Interval valE3dotE3;
      Dot(valE3dotE3, ppvalE[3][0], ppvalE[3][1], ppvalE[3][0], ppvalE[3][1]);

      Interval valE2xV1;
      Cross(valE2xV1, ppvalE[2][0], ppvalE[2][1], pppvalV[1][0][0], pppvalV[1][0][1]);
      iE2xV1 = ( valE2xV1 == 0.0 );
      if ( iE2xV1 == -1 )
	{
	  break;
	}
      if ( iE2xV1 == 1 )
	{
	  // The two vertices lie on a line parallel to the edge
	  return InCircleVVE_EAntiparallelMPFUNFilter(dV0_x, dV0_y, 
						      dV1_x, dV1_y, 
						      dE20_x, dE20_y, dE21_x, dE21_y,
						      dE30_x, dE30_y, dE31_x, dE31_y);
	}

      Interval valV1dotE2;
      Dot(valV1dotE2, pppvalV[1][0][0], pppvalV[1][0][1], ppvalE[2][0], ppvalE[2][1]);
      Interval valV1dotE3;
      Dot(valV1dotE3, pppvalV[1][0][0], pppvalV[1][0][1], ppvalE[3][0], ppvalE[3][1]);
      Interval valE2dotE3;
      Dot(valE2dotE3, ppvalE[2][0], ppvalE[2][1], ppvalE[3][0], ppvalE[3][1]);

      Interval valE2xE3;
      Cross(valE2xE3, ppvalE[2][0], ppvalE[2][1], ppvalE[3][0], ppvalE[3][1]);
      Interval valE20xE21;
      Cross(valE20xE21, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][1][0], pppvalV[2][1][1]);
      Interval valE30xE31;
      Cross(valE30xE31, pppvalV[3][0][0], pppvalV[3][0][1], pppvalV[3][1][0], pppvalV[3][1][1]);

      Interval valDiscR;
      valDiscR = valV1dotV1*valE20xE21*( valE20xE21 + valE2xV1 );
      INT_32 iDiscR;
      iDiscR = ( valDiscR >= 0.0 ); 
      if ( iDiscR == -1 )
	{
	  break;
	}
      ASSERT( iDiscR == 1 );

      Interval valDisc;
      valDisc = valE2dotE2*valDiscR;
      INT_32 iDisc;
      iDisc = ( valDisc >= 0.0 ); 
      if ( iDisc == -1 )
	{
	  break;
	}
      ASSERT( iDisc == 1 );
      Interval valSqrtDisc = sqrt(valDisc);

      Interval valA;
      valA = ( valV1dotV1*valE2xV1*valE2dotE3 + 2.0*valV1dotE3*( valE20xE21*valV1dotE2 + valSqrtDisc ) );
      Interval valB;
      valB = ( valV1dotV1*sqrt(valE2dotE2)*( 2.0*valE20xE21 + valE2xV1 ) + 2.0*valV1dotE2*sqrt(valDiscR) );

      Interval valE30xE31timesE2xV1timesE2xV1 = valE30xE31*valE2xV1*valE2xV1;

      Interval valResult;
      valResult = ( valA*valA
		  + 4.0*valE30xE31timesE2xV1timesE2xV1*valA
		  + 4.0*valE30xE31timesE2xV1timesE2xV1*valE30xE31timesE2xV1timesE2xV1
		  - valE3dotE3*valB*valB );
  
      iResult = ( valResult < 0.0 );
      if ( iResult == -1 )
	{
	  break;
	}
      if ( iResult == 1 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( valResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
      else
	{
	  return APM2D::cttOnCircle;
	}
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( iResult == -1 )
    {
      APM2D::m_uiInCircleVVE_EMPFUN++;

      mp_real rV0_x = dV0_x;
      mp_real rV0_y = dV0_y;
      mp_real rV1_x = dV1_x;
      mp_real rV1_y = dV1_y;
      mp_real rE20_x = dE20_x;
      mp_real rE20_y = dE20_y; 
      mp_real rE21_x = dE21_x;
      mp_real rE21_y = dE21_y;
      mp_real rE30_x = dE30_x;
      mp_real rE30_y = dE30_y;
      mp_real rE31_x = dE31_x;
      mp_real rE31_y = dE31_y;
      
      mp_real ppprV[4][2][2]; // Endpoints
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = 0.0;
      ppprV[0][0][1] = 0.0;
      ppprV[1][0][0] = rV1_x - rV0_x;
      ppprV[1][0][1] = rV1_y - rV0_y;
      ppprV[2][0][0] = rE20_x - rV0_x;
      ppprV[2][0][1] = rE20_y - rV0_y;
      ppprV[2][1][0] = rE21_x - rV0_x;
      ppprV[2][1][1] = rE21_y - rV0_y;
      ppprV[3][0][0] = rE30_x - rV0_x;
      ppprV[3][0][1] = rE30_y - rV0_y;
      ppprV[3][1][0] = rE31_x - rV0_x;
      ppprV[3][1][1] = rE31_y - rV0_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[2][0] = rE21_x - rE20_x;
      pprE[2][1] = rE21_y - rE20_y;
      pprE[3][0] = rE31_x - rE30_x;
      pprE[3][1] = rE31_y - rE30_y;

      mp_real rV1dotV1;
      Dot(rV1dotV1, ppprV[1][0][0], ppprV[1][0][1], ppprV[1][0][0], ppprV[1][0][1]);
      mp_real rE2dotE2;
      Dot(rE2dotE2, pprE[2][0], pprE[2][1], pprE[2][0], pprE[2][1]);
      mp_real rE3dotE3;
      Dot(rE3dotE3, pprE[3][0], pprE[3][1], pprE[3][0], pprE[3][1]);
      
      mp_real rE2xV1;
      Cross(rE2xV1, pprE[2][0], pprE[2][1], ppprV[1][0][0], ppprV[1][0][1]);
      if ( rE2xV1 == 0.0 )
	{
	  // The two vertices lie on a line parallel to the edge
	  return InCircleVVE_EAntiparallelMPFUNFilter(dV0_x, dV0_y, 
						      dV1_x, dV1_y, 
						      dE20_x, dE20_y, dE21_x, dE21_y,
						      dE30_x, dE30_y, dE31_x, dE31_y);
	}

      mp_real rV1dotE2;
      Dot(rV1dotE2, ppprV[1][0][0], ppprV[1][0][1], pprE[2][0], pprE[2][1]);
      mp_real rV1dotE3;
      Dot(rV1dotE3, ppprV[1][0][0], ppprV[1][0][1], pprE[3][0], pprE[3][1]);
      mp_real rE2dotE3;
      Dot(rE2dotE3, pprE[2][0], pprE[2][1], pprE[3][0], pprE[3][1]);

      mp_real rE2xE3;
      Cross(rE2xE3, pprE[2][0], pprE[2][1], pprE[3][0], pprE[3][1]);
      mp_real rE20xE21;
      Cross(rE20xE21, ppprV[2][0][0], ppprV[2][0][1], ppprV[2][1][0], ppprV[2][1][1]);
      mp_real rE30xE31;
      Cross(rE30xE31, ppprV[3][0][0], ppprV[3][0][1], ppprV[3][1][0], ppprV[3][1][1]);
      
      mp_real rDiscR;
      rDiscR = rV1dotV1*rE20xE21*( rE20xE21 + rE2xV1 );
      ASSERT( rDiscR >= 0.0 ); 
      
      mp_real rDisc;
      rDisc = rE2dotE2*rDiscR;
      ASSERT( rDisc >= 0.0 ); 
      mp_real rSqrtDisc = sqrt(rDisc);
      
      mp_real rA;
      rA = ( rV1dotV1*rE2xV1*rE2dotE3 + 2.0*rV1dotE3*( rE20xE21*rV1dotE2 + rSqrtDisc ) );
      mp_real rB;
      rB = ( rV1dotV1*sqrt(rE2dotE2)*( 2.0*rE20xE21 + rE2xV1 ) + 2.0*rV1dotE2*sqrt(rDiscR) );

      mp_real rE30xE31timesE2xV1timesE2xV1 = rE30xE31*rE2xV1*rE2xV1;

      mp_real rResult;
      rResult = ( rA*rA
		  + 4.0*rE30xE31timesE2xV1timesE2xV1*rA
		  + 4.0*rE30xE31timesE2xV1timesE2xV1*rE30xE31timesE2xV1timesE2xV1
		  - rE3dotE3*rB*rB );
  
      if ( rResult < 0.0 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( rResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
    }

  return APM2D::cttOnCircle;
}

//////////////////////////////////////////////////
// VVE Antiparallel

class ICC_VVEAntiparallel_MPFUNFilter
{
protected:
  enum InstructionPointer {ipStart, ipV0orV1onE2,
			   ipEdgeZones, ipVertexZones, 
			   ipDen, ipDenR, ipEnd};

public:
  ICC_VVEAntiparallel_MPFUNFilter();
  ~ICC_VVEAntiparallel_MPFUNFilter();

public:
  APM2D::InscribedCircleCenterType 
  InscribedCircleCenterVVEAntiparallel(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				       DOUBLE dV00_x, DOUBLE dV00_y,
				       DOUBLE dV0_x, DOUBLE dV0_y,
				       DOUBLE dV01_x, DOUBLE dV01_y,
				       DOUBLE dV10_x, DOUBLE dV10_y,
				       DOUBLE dV1_x, DOUBLE dV1_y,
				       DOUBLE dV11_x, DOUBLE dV11_y,
				       DOUBLE dE20_x, DOUBLE dE20_y, 
				       DOUBLE dE21_x, DOUBLE dE21_y);
  APM2D::InscribedCircleCenterType 
  InscribedCircleCenterVVEAntiparallelForce(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					    DOUBLE dV00_x, DOUBLE dV00_y,
					    DOUBLE dV0_x, DOUBLE dV0_y,
					    DOUBLE dV01_x, DOUBLE dV01_y,
					    DOUBLE dV10_x, DOUBLE dV10_y,
					    DOUBLE dV1_x, DOUBLE dV1_y,
					    DOUBLE dV11_x, DOUBLE dV11_y,
					    DOUBLE dE20_x, DOUBLE dE20_y, 
					    DOUBLE dE21_x, DOUBLE dE21_y);

protected:
  BOOL
  InscribedCircleCenterVVEAntiparallelMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					    DOUBLE dV00_x, DOUBLE dV00_y,
					    DOUBLE dV0_x, DOUBLE dV0_y,
					    DOUBLE dV01_x, DOUBLE dV01_y,
					    DOUBLE dV10_x, DOUBLE dV10_y,
					    DOUBLE dV1_x, DOUBLE dV1_y,
					    DOUBLE dV11_x, DOUBLE dV11_y,
					    DOUBLE dE20_x, DOUBLE dE20_y, 
					    DOUBLE dE21_x, DOUBLE dE21_y,
					    InstructionPointer eIPMax = ipEnd);

protected:
  InstructionPointer m_eIP;
  APM2D::InscribedCircleCenterType m_eResult;
  INT_32 m_iV0onE2, m_iV1onE2;
  INT_32 m_iDen;
  INT_32 m_iDenR;

  INT_32 m_piEdgeZone[2];
  INT_32 m_ppiVertexZone[2][2];

  Interval m_valE2xV1;
  Interval m_valDiscR;
  Interval m_valDisc;

  mp_real m_rV00_x;
  mp_real m_rV00_y;
  mp_real m_rV0_x;
  mp_real m_rV0_y;
  mp_real m_rV01_x;
  mp_real m_rV01_y;
  mp_real m_rV10_x;
  mp_real m_rV10_y;
  mp_real m_rV1_x;
  mp_real m_rV1_y;
  mp_real m_rV11_x;
  mp_real m_rV11_y;
  mp_real m_rE20_x;
  mp_real m_rE20_y; 
  mp_real m_rE21_x;
  mp_real m_rE21_y;
  mp_real m_ppprV[3][2][2];
  mp_real m_prE[2];
  mp_real m_rE2dotV1;
  mp_real m_rE20xE21;
  mp_real m_rDen;
  mp_real m_rDenR;
  mp_real m_r4E20xE21timesE20xE21minusE2dotV1timesE2dotV1;
};

ICC_VVEAntiparallel_MPFUNFilter::ICC_VVEAntiparallel_MPFUNFilter()
{
}

ICC_VVEAntiparallel_MPFUNFilter::~ICC_VVEAntiparallel_MPFUNFilter()
{
}

APM2D::InscribedCircleCenterType
ICC_VVEAntiparallel_MPFUNFilter::InscribedCircleCenterVVEAntiparallel(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
								      DOUBLE dV00_x, DOUBLE dV00_y,
								      DOUBLE dV0_x, DOUBLE dV0_y,
								      DOUBLE dV01_x, DOUBLE dV01_y,
								      DOUBLE dV10_x, DOUBLE dV10_y,
								      DOUBLE dV1_x, DOUBLE dV1_y,
								      DOUBLE dV11_x, DOUBLE dV11_y,
								      DOUBLE dE20_x, DOUBLE dE20_y, 
								      DOUBLE dE21_x, DOUBLE dE21_y)
{
  APM2D::m_uiInscribedCircleCenterVVEAntiparallelMPFUNFilter++;

  *pdC_x = 0.0;
  *pdC_y = 0.0;
  *pdR = 0.0;
  
  m_eIP = ipStart;
  m_eResult = APM2D::icctNull;
  m_iV0onE2 = -1; 
  m_iV1onE2 = -1;
  m_iDen = -1;
  m_iDenR = -1;

  m_piEdgeZone[0] = -1;
  m_piEdgeZone[1] = -1;
  m_ppiVertexZone[0][0] = -1;  
  m_ppiVertexZone[0][1] = -1;  
  m_ppiVertexZone[1][0] = -1;  
  m_ppiVertexZone[1][1] = -1;  

  BOOL bDone;

  Interval valV00_x = dV00_x;
  Interval valV00_y = dV00_y;
  Interval valV0_x = dV0_x;
  Interval valV0_y = dV0_y;
  Interval valV01_x = dV01_x;
  Interval valV01_y = dV01_y;
  Interval valV10_x = dV10_x;
  Interval valV10_y = dV10_y;
  Interval valV1_x = dV1_x;
  Interval valV1_y = dV1_y;
  Interval valV11_x = dV11_x;
  Interval valV11_y = dV11_y;
  Interval valE20_x = dE20_x;
  Interval valE20_y = dE20_y; 
  Interval valE21_x = dE21_x;
  Interval valE21_y = dE21_y;
  
  //////////////////////////////////////////////////
  // Check for V0 or V1 on E2
  
  Interval valV0onE2;
  SinTheta(valV0onE2, valV0_x, valV0_y, valE20_x, valE20_y, valE21_x, valE21_y);
  m_iV0onE2 = ( valV0onE2 <= 0.0 );
  if ( m_iV0onE2 == 1 )
    {
      // V0 is on E2 and it is not one of the endpoints
      return APM2D::icctNull;
    }
  
  Interval valV1onE2;
  SinTheta(valV1onE2, valV1_x, valV1_y, valE20_x, valE20_y, valE21_x, valE21_y);
  m_iV1onE2 = ( valV1onE2 <= 0.0 );
  if ( m_iV1onE2 == 1 )
    {
      // V1 is on E2 and it is not one of the endpoints
      return APM2D::icctNull;
    }

  if ( m_iV0onE2 == -1 || m_iV1onE2 == -1 )
    {
      bDone = InscribedCircleCenterVVEAntiparallelMPFUN(pdC_x, pdC_y, pdR, 
							dV00_x, dV00_y,
							dV0_x, dV0_y,
							dV01_x, dV01_y,
							dV10_x, dV10_y,
							dV1_x, dV1_y,
							dV11_x, dV11_y,
							dE20_x, dE20_y, 
							dE21_x, dE21_y,
							ipV0orV1onE2);  
      if ( bDone )
	{
	  return m_eResult;
	}
    }

  //////////////////////////////////////////////////
  // Check the Edge Zone
  
  Interval pppvalV[3][2][2];
  // Translate V0 to the origin
  //memset(pppvalV, 0, sizeof(pppvalV));
  pppvalV[1][0][0] = valV1_x - valV0_x;
  pppvalV[1][0][1] = valV1_y - valV0_y;
  pppvalV[2][0][0] = valE20_x - valV0_x;
  pppvalV[2][0][1] = valE20_y - valV0_y;
  pppvalV[2][1][0] = valE21_x - valV0_x;
  pppvalV[2][1][1] = valE21_y - valV0_y;
	
  Interval pvalE[2];
  pvalE[0] = valE21_x - valE20_x;
  pvalE[1] = valE21_y - valE20_y;
  
  Interval valE2dotV1;
  Dot(valE2dotV1, pvalE[0], pvalE[1], pppvalV[1][0][0], pppvalV[1][0][1]);
  
  Interval pvalEdgeZone[2];
  
  Interval valE2dotE20;
  Dot(valE2dotE20, pvalE[0], pvalE[1], pppvalV[2][0][0], pppvalV[2][0][1]);
  pvalEdgeZone[0] = -( 2.0*valE2dotE20 - valE2dotV1 );
  m_piEdgeZone[0] = ( pvalEdgeZone[0] < 0.0 );
  if ( m_piEdgeZone[0] == 1 )
    {
      return APM2D::icctNull;
    }
  
  Interval valE2dotE21;
  Dot(valE2dotE21, pvalE[0], pvalE[1], pppvalV[2][1][0], pppvalV[2][1][1]);
  pvalEdgeZone[1] = ( 2.0*valE2dotE21 - valE2dotV1 );
  m_piEdgeZone[1] = ( pvalEdgeZone[1] < 0.0 );
  if ( m_piEdgeZone[1] == 1 )
    {
      return APM2D::icctNull;
    }

  if ( m_piEdgeZone[0] == -1 || m_piEdgeZone[1] == -1 )
    {
      bDone = InscribedCircleCenterVVEAntiparallelMPFUN(pdC_x, pdC_y, pdR, 
							dV00_x, dV00_y,
							dV0_x, dV0_y,
							dV01_x, dV01_y,
							dV10_x, dV10_y,
							dV1_x, dV1_y,
							dV11_x, dV11_y,
							dE20_x, dE20_y, 
							dE21_x, dE21_y,
							ipEdgeZones);  
      if ( bDone )
	{
	  return m_eResult;
	}
    }

  //////////////////////////////////////////////////
  // Check the Vertex Zones
  
  Interval valE20xE21;
  Cross(valE20xE21, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][1][0], pppvalV[2][1][1]);
  Interval val4E20xE21timesE20xE21minusE2dotV1timesE2dotV1 = 4.0*valE20xE21*valE20xE21 - valE2dotV1*valE2dotV1;
  Interval val4E20xE21timesE2dotV1 = 4.0*valE20xE21*valE2dotV1;
  
  Interval pppvalVN[2][2][2];
  // Vertex neighbor vectors
  //memset(pppvalVN, 0, sizeof(pppvalV));
  pppvalVN[0][0][0] = valV0_x - valV00_x;
  pppvalVN[0][0][1] = valV0_y - valV00_y;
  pppvalVN[0][1][0] = valV0_x - valV01_x;
  pppvalVN[0][1][1] = valV0_y - valV01_y;
  pppvalVN[1][0][0] = valV1_x - valV10_x;
  pppvalVN[1][0][1] = valV1_y - valV10_y;
  pppvalVN[1][1][0] = valV1_x - valV11_x;
  pppvalVN[1][1][1] = valV1_y - valV11_y;
      
  Interval ppvalVertexZone[2][2];
  
  Interval valV1dotV00;
  Dot(valV1dotV00, pppvalV[1][0][0], pppvalV[1][0][1], pppvalVN[0][0][0], pppvalVN[0][0][1]);
  Interval valV1xV00;
  Cross(valV1xV00, pppvalV[1][0][0], pppvalV[1][0][1], pppvalVN[0][0][0], pppvalVN[0][0][1]);
  ppvalVertexZone[0][0] = ( (val4E20xE21timesE20xE21minusE2dotV1timesE2dotV1)*valV1xV00
				    - val4E20xE21timesE2dotV1*valV1dotV00 );
  m_ppiVertexZone[0][0] = ( ppvalVertexZone[0][0] < 0.0 );
  if ( m_ppiVertexZone[0][0] == 1 ) 
    {
      return APM2D::icctNull;
    }

  Interval valV1dotV01;
  Dot(valV1dotV01, pppvalV[1][0][0], pppvalV[1][0][1], pppvalVN[0][1][0], pppvalVN[0][1][1]);
  Interval valV1xV01;
  Cross(valV1xV01, pppvalV[1][0][0], pppvalV[1][0][1], pppvalVN[0][1][0], pppvalVN[0][1][1]);
  ppvalVertexZone[0][1] = ( (val4E20xE21timesE20xE21minusE2dotV1timesE2dotV1)*valV1xV01
			    - val4E20xE21timesE2dotV1*valV1dotV01 );
  m_ppiVertexZone[0][1] = ( ppvalVertexZone[0][1] < 0.0 );
  if ( m_ppiVertexZone[0][1] == 1 ) 
    {
      return APM2D::icctNull;
    }

  Interval valV1dotV10;
  Dot(valV1dotV10, pppvalV[1][0][0], pppvalV[1][0][1], pppvalVN[1][0][0], pppvalVN[1][0][1]);
  Interval valV1xV10;
  Cross(valV1xV10, pppvalV[1][0][0], pppvalV[1][0][1], pppvalVN[1][0][0], pppvalVN[1][0][1]);
  ppvalVertexZone[1][0] = ( (val4E20xE21timesE20xE21minusE2dotV1timesE2dotV1)*valV1xV10
			    + val4E20xE21timesE2dotV1*valV1dotV10 );
  m_ppiVertexZone[1][0] = ( ppvalVertexZone[1][0] < 0.0 );
  if ( m_ppiVertexZone[1][0] == 1 ) 
    {
      return APM2D::icctNull;
    }

  Interval valV1dotV11;
  Dot(valV1dotV11, pppvalV[1][0][0], pppvalV[1][0][1], pppvalVN[1][1][0], pppvalVN[1][1][1]);
  Interval valV1xV11;
  Cross(valV1xV11, pppvalV[1][0][0], pppvalV[1][0][1], pppvalVN[1][1][0], pppvalVN[1][1][1]);
  ppvalVertexZone[1][1] = ( (val4E20xE21timesE20xE21minusE2dotV1timesE2dotV1)*valV1xV11
			    + val4E20xE21timesE2dotV1*valV1dotV11 );
  m_ppiVertexZone[1][1] = ( ppvalVertexZone[1][1] < 0.0 );
  if ( m_ppiVertexZone[1][1] == 1 ) 
    {
      return APM2D::icctNull;
    }

  if ( m_ppiVertexZone[0][0] == -1 || 
       m_ppiVertexZone[0][1] == -1 || 
       m_ppiVertexZone[1][0] == -1 || 
       m_ppiVertexZone[1][1] == -1 ) 
    {      
      bDone = InscribedCircleCenterVVEAntiparallelMPFUN(pdC_x, pdC_y, pdR, 
							dV00_x, dV00_y,
							dV0_x, dV0_y,
							dV01_x, dV01_y,
							dV10_x, dV10_y,
							dV1_x, dV1_y,
							dV11_x, dV11_y,
							dE20_x, dE20_y, 
							dE21_x, dE21_y,
							ipVertexZones);  
      if ( bDone )
	{
	  return m_eResult;
	}
    }

  //////////////////////////////////////////////////
  // Check Den
  
  Interval valDen;
  valDen = -8.0*valE20xE21*valE2dotV1;
  m_iDen = ( valDen != 0.0 );
  ASSERT( m_iDen == 1 );
  if ( m_iDen == -1 )
    {
      bDone = InscribedCircleCenterVVEAntiparallelMPFUN(pdC_x, pdC_y, pdR, 
							dV00_x, dV00_y,
							dV0_x, dV0_y,
							dV01_x, dV01_y,
							dV10_x, dV10_y,
							dV1_x, dV1_y,
							dV11_x, dV11_y,
							dE20_x, dE20_y, 
							dE21_x, dE21_y,
							ipEnd);  
      ASSERT( bDone );
      return m_eResult;
    }

  //////////////////////////////////////////////////
  // Check DenR
  
  Interval valE2dotE2;
  Dot(valE2dotE2, pvalE[0], pvalE[1], pvalE[0], pvalE[1]);
  
  Interval valDenR;
  valDenR = 8.0*valE20xE21*sqrt(valE2dotE2);
  m_iDenR = ( valDenR != 0.0 );
  ASSERT( m_iDenR == 1 );
  if ( m_iDenR == -1 )
    {
      bDone = InscribedCircleCenterVVEAntiparallelMPFUN(pdC_x, pdC_y, pdR, 
							dV00_x, dV00_y,
							dV0_x, dV0_y,
							dV01_x, dV01_y,
							dV10_x, dV10_y,
							dV1_x, dV1_y,
							dV11_x, dV11_y,
							dE20_x, dE20_y, 
							dE21_x, dE21_y,
							ipEnd);  
      ASSERT( bDone );
      return m_eResult;
    }

  //////////////////////////////////////////////////
  // Compute the center and radius
  
  Interval pvalC[2];
  pvalC[0] = ( -(val4E20xE21timesE20xE21minusE2dotV1timesE2dotV1)*pppvalV[1][0][1]
		   - (4.0*valE20xE21*valE2dotV1)*pppvalV[1][0][0] ) / valDen;
  pvalC[1] = ( (val4E20xE21timesE20xE21minusE2dotV1timesE2dotV1)*pppvalV[1][0][0]
	       - (4.0*valE20xE21*valE2dotV1)*pppvalV[1][0][1] ) / valDen;
  
  Interval valR;
  valR = (4.0*valE20xE21*valE20xE21 + valE2dotV1*valE2dotV1) / valDenR;
  
  // Translate the answer back out away from the origin
  pvalC[0] += valV0_x;
  pvalC[1] += valV0_y;
  
  *pdC_x = dble(pvalC[0]);
  *pdC_y = dble(pvalC[1]);
  *pdR = dble(valR);
  return APM2D::icctVVEAntiparallel;
}

APM2D::InscribedCircleCenterType
ICC_VVEAntiparallel_MPFUNFilter::InscribedCircleCenterVVEAntiparallelForce(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
									   DOUBLE dV00_x, DOUBLE dV00_y,
									   DOUBLE dV0_x, DOUBLE dV0_y,
									   DOUBLE dV01_x, DOUBLE dV01_y,
									   DOUBLE dV10_x, DOUBLE dV10_y,
									   DOUBLE dV1_x, DOUBLE dV1_y,
									   DOUBLE dV11_x, DOUBLE dV11_y,
									   DOUBLE dE20_x, DOUBLE dE20_y, 
									   DOUBLE dE21_x, DOUBLE dE21_y)
{
  *pdC_x = 0.0;
  *pdC_y = 0.0;
  *pdR = 0.0;
  
  m_eIP = ipStart;
  m_eResult = APM2D::icctNull;
  m_iV0onE2 = -1; 
  m_iV1onE2 = -1;
  m_iDen = -1;
  m_iDenR = -1;

  m_piEdgeZone[0] = -1;
  m_piEdgeZone[1] = -1;
  m_ppiVertexZone[0][0] = -1;  
  m_ppiVertexZone[0][1] = -1;  
  m_ppiVertexZone[1][0] = -1;  
  m_ppiVertexZone[1][1] = -1;  

  BOOL bDone;
  bDone = InscribedCircleCenterVVEAntiparallelMPFUN(pdC_x, pdC_y, pdR, 
						    dV00_x, dV00_y,
						    dV0_x, dV0_y,
						    dV01_x, dV01_y,
						    dV10_x, dV10_y,
						    dV1_x, dV1_y,
						    dV11_x, dV11_y,
						    dE20_x, dE20_y, 
						    dE21_x, dE21_y);  
  if ( bDone )
    {
      fprintf(stderr, "InscribedCircleCenterVVEAntiparallelMPFUN does not work!!!\n");
      exit(1);
    }
  return m_eResult;
}

BOOL
ICC_VVEAntiparallel_MPFUNFilter::InscribedCircleCenterVVEAntiparallelMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
									   DOUBLE dV00_x, DOUBLE dV00_y,
									   DOUBLE dV0_x, DOUBLE dV0_y,
									   DOUBLE dV01_x, DOUBLE dV01_y,
									   DOUBLE dV10_x, DOUBLE dV10_y,
									   DOUBLE dV1_x, DOUBLE dV1_y,
									   DOUBLE dV11_x, DOUBLE dV11_y,
									   DOUBLE dE20_x, DOUBLE dE20_y, 
									   DOUBLE dE21_x, DOUBLE dE21_y,
									   ICC_VVEAntiparallel_MPFUNFilter::InstructionPointer eIPMax/* = ipEnd*/)
{
  APM2D::m_uiInscribedCircleCenterVVEAntiparallelMPFUN++;

  switch ( m_eIP )
    {
    case ipStart:
      {
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	*pdC_x = 0.0;
	*pdC_y = 0.0;
	*pdR = 0.0;

	m_rV00_x = dV00_x;
	m_rV00_y = dV00_y;
	m_rV0_x = dV0_x;
	m_rV0_y = dV0_y;
	m_rV01_x = dV01_x;
	m_rV01_y = dV01_y;
	m_rV10_x = dV10_x;
	m_rV10_y = dV10_y;
	m_rV1_x = dV1_x;
	m_rV1_y = dV1_y;
	m_rV11_x = dV11_x;
	m_rV11_y = dV11_y;
	m_rE20_x = dE20_x;
	m_rE20_y = dE20_y; 
	m_rE21_x = dE21_x;
	m_rE21_y = dE21_y;

	//////////////////////////////////////////////////
	// Check for V0 or V1 on E2
      
	if ( m_iV0onE2 == -1 )
	  {
	    mp_real rV0onE2;
	    SinTheta(rV0onE2, m_rV0_x, m_rV0_y, m_rE20_x, m_rE20_y, m_rE21_x, m_rE21_y);
	    if ( rV0onE2 <= 0.0 )
	      {
		// V0 is on E2 and it is not one of the endpoints
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterVVEAntiparallelMPFUNFinal++;
		return TRUE;
	      }
	  }
  
	if ( m_iV1onE2 == -1 )
	  {
	    mp_real rV1onE2;
	    SinTheta(rV1onE2, m_rV1_x, m_rV1_y, m_rE20_x, m_rE20_y, m_rE21_x, m_rE21_y);
	    if ( rV1onE2 <= 0.0 )
	      {
		// V1 is on E2 and it is not one of the endpoints
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterVVEAntiparallelMPFUNFinal++;
		return TRUE;
	      }
	  }
  
      }
    case ipV0orV1onE2:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	// Translate V0 to the origin
	//memset(m_ppprV, 0, sizeof(m_ppprV));
	m_ppprV[1][0][0] = m_rV1_x - m_rV0_x;
	m_ppprV[1][0][1] = m_rV1_y - m_rV0_y;
	m_ppprV[2][0][0] = m_rE20_x - m_rV0_x;
	m_ppprV[2][0][1] = m_rE20_y - m_rV0_y;
	m_ppprV[2][1][0] = m_rE21_x - m_rV0_x;
	m_ppprV[2][1][1] = m_rE21_y - m_rV0_y;
	
	m_prE[0] = m_rE21_x - m_rE20_x;
	m_prE[1] = m_rE21_y - m_rE20_y;
  
	Dot(m_rE2dotV1, m_prE[0], m_prE[1], m_ppprV[1][0][0], m_ppprV[1][0][1]);
	
	//////////////////////////////////////////////////
	// Check the Edge Zone

	if ( m_piEdgeZone[0] == -1 || m_piEdgeZone[1] == -1 )
	  {
	    mp_real prEdgeZone[2];

	    if ( m_piEdgeZone[0] == -1 )
	      {
		mp_real rE2dotE20;
		Dot(rE2dotE20, m_prE[0], m_prE[1], m_ppprV[2][0][0], m_ppprV[2][0][1]);
		prEdgeZone[0] = -( 2.0*rE2dotE20 - m_rE2dotV1 );
		if ( prEdgeZone[0] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVVEAntiparallelMPFUNFinal++;
		    return TRUE;
		  }
	      }

	    if ( m_piEdgeZone[1] == -1 )
	      {
		mp_real rE2dotE21;
		Dot(rE2dotE21, m_prE[0], m_prE[1], m_ppprV[2][1][0], m_ppprV[2][1][1]);
		prEdgeZone[1] = ( 2.0*rE2dotE21 - m_rE2dotV1 );
		if ( prEdgeZone[1] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVVEAntiparallelMPFUNFinal++;
		    return TRUE;
		  }
	      }
	  }

      }
    case ipEdgeZones:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Check the Vertex Zones

	Cross(m_rE20xE21, m_ppprV[2][0][0], m_ppprV[2][0][1], m_ppprV[2][1][0], m_ppprV[2][1][1]);
	m_r4E20xE21timesE20xE21minusE2dotV1timesE2dotV1 = 4.0*m_rE20xE21*m_rE20xE21 - m_rE2dotV1*m_rE2dotV1;
	mp_real r4E20xE21timesE2dotV1 = 4.0*m_rE20xE21*m_rE2dotV1;

	if ( m_ppiVertexZone[0][0] == -1 || 
	     m_ppiVertexZone[0][1] == -1 || 
	     m_ppiVertexZone[1][0] == -1 || 
	     m_ppiVertexZone[1][1] == -1 ) 
	  {
	    // Vertex neighbor vectors
	    mp_real ppprVN[2][2][2];
	    //memset(ppprVN, 0, sizeof(ppprV));
	    ppprVN[0][0][0] = m_rV0_x - m_rV00_x;
	    ppprVN[0][0][1] = m_rV0_y - m_rV00_y;
	    ppprVN[0][1][0] = m_rV0_x - m_rV01_x;
	    ppprVN[0][1][1] = m_rV0_y - m_rV01_y;
	    ppprVN[1][0][0] = m_rV1_x - m_rV10_x;
	    ppprVN[1][0][1] = m_rV1_y - m_rV10_y;
	    ppprVN[1][1][0] = m_rV1_x - m_rV11_x;
	    ppprVN[1][1][1] = m_rV1_y - m_rV11_y;
  
	    mp_real pprVertexZone[2][2];

	    if ( m_ppiVertexZone[0][0] == -1 ) 
	      {
		mp_real rV1dotV00;
		Dot(rV1dotV00, m_ppprV[1][0][0], m_ppprV[1][0][1], ppprVN[0][0][0], ppprVN[0][0][1]);
		mp_real rV1xV00;
		Cross(rV1xV00, m_ppprV[1][0][0], m_ppprV[1][0][1], ppprVN[0][0][0], ppprVN[0][0][1]);
		pprVertexZone[0][0] = ( (m_r4E20xE21timesE20xE21minusE2dotV1timesE2dotV1)*rV1xV00
					- r4E20xE21timesE2dotV1*rV1dotV00 );
		if ( pprVertexZone[0][0] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVVEAntiparallelMPFUNFinal++;
		    return TRUE;
		  }
	      }
	    if ( m_ppiVertexZone[0][1] == -1 ) 
	      {
		mp_real rV1dotV01;
		Dot(rV1dotV01, m_ppprV[1][0][0], m_ppprV[1][0][1], ppprVN[0][1][0], ppprVN[0][1][1]);
		mp_real rV1xV01;
		Cross(rV1xV01, m_ppprV[1][0][0], m_ppprV[1][0][1], ppprVN[0][1][0], ppprVN[0][1][1]);
		pprVertexZone[0][1] = ( (m_r4E20xE21timesE20xE21minusE2dotV1timesE2dotV1)*rV1xV01
					- r4E20xE21timesE2dotV1*rV1dotV01 );
		if ( pprVertexZone[0][1] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVVEAntiparallelMPFUNFinal++;
		    return TRUE;
		  }
	      }
	    if ( m_ppiVertexZone[1][0] == -1 ) 
	      {
		mp_real rV1dotV10;
		Dot(rV1dotV10, m_ppprV[1][0][0], m_ppprV[1][0][1], ppprVN[1][0][0], ppprVN[1][0][1]);
		mp_real rV1xV10;
		Cross(rV1xV10, m_ppprV[1][0][0], m_ppprV[1][0][1], ppprVN[1][0][0], ppprVN[1][0][1]);
		pprVertexZone[1][0] = ( (m_r4E20xE21timesE20xE21minusE2dotV1timesE2dotV1)*rV1xV10
					+ r4E20xE21timesE2dotV1*rV1dotV10 );
		if ( pprVertexZone[1][0] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVVEAntiparallelMPFUNFinal++;
		    return TRUE;
		  }
	      }
	    if ( m_ppiVertexZone[1][1] == -1 ) 
	      {
		mp_real rV1dotV11;
		Dot(rV1dotV11, m_ppprV[1][0][0], m_ppprV[1][0][1], ppprVN[1][1][0], ppprVN[1][1][1]);
		mp_real rV1xV11;
		Cross(rV1xV11, m_ppprV[1][0][0], m_ppprV[1][0][1], ppprVN[1][1][0], ppprVN[1][1][1]);
		pprVertexZone[1][1] = ( (m_r4E20xE21timesE20xE21minusE2dotV1timesE2dotV1)*rV1xV11
					+ r4E20xE21timesE2dotV1*rV1dotV11 );
		if ( pprVertexZone[1][1] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVVEAntiparallelMPFUNFinal++;
		    return TRUE;
		  }
	      }
	  }

      }
    case ipVertexZones:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Check Den
      
	m_rDen = -8.0*m_rE20xE21*m_rE2dotV1;
	if ( m_rDen == 0.0 )
	  {
	    ASSERT( FALSE );
	    m_eResult = APM2D::icctNull;
	    APM2D::m_uiInscribedCircleCenterVVEAntiparallelMPFUNFinal++;
	    return TRUE;
	  }

      }
    case ipDen:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Check DenR

	mp_real rE2dotE2;
	Dot(rE2dotE2, m_prE[0], m_prE[1], m_prE[0], m_prE[1]);

	m_rDenR = 8.0*m_rE20xE21*sqrt(rE2dotE2);
	if ( m_rDenR == 0.0 )
	  {
	    ASSERT( FALSE );
	    m_eResult = APM2D::icctNull;
	    APM2D::m_uiInscribedCircleCenterVVEAntiparallelMPFUNFinal++;
	    return TRUE;
	  }

      }
    case ipDenR:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Compute the center and radius

	mp_real prC[2];
	prC[0] = ( -(m_r4E20xE21timesE20xE21minusE2dotV1timesE2dotV1)*m_ppprV[1][0][1]
		   - (4.0*m_rE20xE21*m_rE2dotV1)*m_ppprV[1][0][0] ) / m_rDen;
	prC[1] = ( (m_r4E20xE21timesE20xE21minusE2dotV1timesE2dotV1)*m_ppprV[1][0][0]
		   - (4.0*m_rE20xE21*m_rE2dotV1)*m_ppprV[1][0][1] ) / m_rDen;

	mp_real rR;
	rR = (4.0*m_rE20xE21*m_rE20xE21 + m_rE2dotV1*m_rE2dotV1) / m_rDenR;
  
	// Translate the answer back out away from the origin
	prC[0] += m_rV0_x;
	prC[1] += m_rV0_y;

	*pdC_x = dble(prC[0]);
	*pdC_y = dble(prC[1]);
	*pdR = dble(rR);
	m_eResult =  APM2D::icctVVEAntiparallel;
      }
      break;
      /*
    default:
      ASSERT( FALSE );
      break;
      */
    }

  APM2D::m_uiInscribedCircleCenterVVEAntiparallelMPFUNFinal++;
  return TRUE;
}

//static 
APM2D::InscribedCircleCenterType 
InscribedCircleCenterVVEAntiparallelMPFUNFilter(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
						DOUBLE dV00_x, DOUBLE dV00_y,
						DOUBLE dV0_x, DOUBLE dV0_y,
						DOUBLE dV01_x, DOUBLE dV01_y,
						DOUBLE dV10_x, DOUBLE dV10_y,
						DOUBLE dV1_x, DOUBLE dV1_y,
						DOUBLE dV11_x, DOUBLE dV11_y,
						DOUBLE dE20_x, DOUBLE dE20_y, 
						DOUBLE dE21_x, DOUBLE dE21_y)
{
  ICC_VVEAntiparallel_MPFUNFilter iccVVEAntiparallel;

  return iccVVEAntiparallel.InscribedCircleCenterVVEAntiparallel(pdC_x, pdC_y, pdR, 
								 dV00_x, dV00_y,
								 dV0_x, dV0_y,
								 dV01_x, dV01_y,
								 dV10_x, dV10_y,
								 dV1_x, dV1_y,
								 dV11_x, dV11_y,
								 dE20_x, dE20_y, 
								 dE21_x, dE21_y);
}

//static 
APM2D::EdgeZoneType EdgeZoneVVEAntiparallelMPFUNFilter(DOUBLE dV0_x, DOUBLE dV0_y,
						       DOUBLE dV1_x, DOUBLE dV1_y,
						       DOUBLE dE20_x, DOUBLE dE20_y, 
						       DOUBLE dE21_x, DOUBLE dE21_y,
						       DOUBLE dE30_x, DOUBLE dE30_y,
						       DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::m_uiEdgeZoneVVEAntiparallelMPFUNFilter++;

  INT_32 piEdgeZone[2];
  piEdgeZone[0] = -1;
  piEdgeZone[1] = -1;

  switch ( 0 )
  {
  default:
      Interval valV0_x = dV0_x;
      Interval valV0_y = dV0_y;
      Interval valV1_x = dV1_x;
      Interval valV1_y = dV1_y;
      Interval valE20_x = dE20_x;
      Interval valE20_y = dE20_y; 
      Interval valE21_x = dE21_x;
      Interval valE21_y = dE21_y;
      Interval valE30_x = dE30_x;
      Interval valE30_y = dE30_y;
      Interval valE31_x = dE31_x;
      Interval valE31_y = dE31_y;

      Interval pppvalV[4][2][2]; // Endpoints
      //memset(pppvalV, 0, sizeof(pppvalV));
      pppvalV[0][0][0] = 0.0;
      pppvalV[0][0][1] = 0.0;
      pppvalV[1][0][0] = valV1_x - valV0_x;
      pppvalV[1][0][1] = valV1_y - valV0_y;
      pppvalV[2][0][0] = valE20_x - valV0_x;
      pppvalV[2][0][1] = valE20_y - valV0_y;
      pppvalV[2][1][0] = valE21_x - valV0_x;
      pppvalV[2][1][1] = valE21_y - valV0_y;
      pppvalV[3][0][0] = valE30_x - valV0_x;
      pppvalV[3][0][1] = valE30_y - valV0_y;
      pppvalV[3][1][0] = valE31_x - valV0_x;
      pppvalV[3][1][1] = valE31_y - valV0_y;
  
      Interval ppvalE[4][2];     // Edge vectors
      //memset(ppvalE, 0, sizeof(ppvalE));
      ppvalE[2][0] = valE21_x - valE20_x;
      ppvalE[2][1] = valE21_y - valE20_y;
      ppvalE[3][0] = valE31_x - valE30_x;
      ppvalE[3][1] = valE31_y - valE30_y;
  
      Interval valV1dotV1;
      Dot(valV1dotV1, pppvalV[1][0][0], pppvalV[1][0][1], pppvalV[1][0][0], pppvalV[1][0][1]);
      Interval valV1dotE3;
      Dot(valV1dotE3, pppvalV[1][0][0], pppvalV[1][0][1], ppvalE[3][0], ppvalE[3][1]);
      Interval valE2dotE2;
      Dot(valE2dotE2, ppvalE[2][0], ppvalE[2][1], ppvalE[2][0], ppvalE[2][1]);
      Interval valE3dotE3;
      Dot(valE3dotE3, ppvalE[3][0], ppvalE[3][1], ppvalE[3][0], ppvalE[3][1]);

      Interval valE3xV1;
      Cross(valE3xV1, ppvalE[3][0], ppvalE[3][1], pppvalV[1][0][0], pppvalV[1][0][1]);
      Interval valE20xE21;
      Cross(valE20xE21, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][1][0], pppvalV[2][1][1]);

      Interval val4E20xE21timesSqrtV1dotV1timesE2dotE2 = 4.0*valE20xE21*sqrt(valV1dotV1*valE2dotE2);

      Interval valE2dotV1;
      Dot(valE2dotV1, ppvalE[2][0], ppvalE[2][1], pppvalV[1][0][0], pppvalV[1][0][1]);

      Interval pvalEdgeZone[2];
      Interval val4E20xE21timesE20xE21minusV1dotV1timesE2dotE2 = 4.0*valE20xE21*valE20xE21 - valV1dotV1*valE2dotE2;

      Interval valE3dotE30;
      Dot(valE3dotE30, ppvalE[3][0], ppvalE[3][1], pppvalV[3][0][0], pppvalV[3][0][1]);
      pvalEdgeZone[0] = ( val4E20xE21timesSqrtV1dotV1timesE2dotE2*( 2.0*valE3dotE30 - valV1dotE3 )
			  + ( val4E20xE21timesE20xE21minusV1dotV1timesE2dotE2 )*valE3xV1 );
      piEdgeZone[0] = ( pvalEdgeZone[0] >= 0.0 );
      if ( piEdgeZone[0] == 1 )
	{
	  return APM2D::eztStart;
	}

      Interval valE3dotE31;
      Dot(valE3dotE31, ppvalE[3][0], ppvalE[3][1], pppvalV[3][1][0], pppvalV[3][1][1]);
      pvalEdgeZone[1] = -( val4E20xE21timesSqrtV1dotV1timesE2dotE2*( 2.0*valE3dotE31 - valV1dotE3 )
			   + ( val4E20xE21timesE20xE21minusV1dotV1timesE2dotE2 )*valE3xV1 );
      piEdgeZone[1] = ( pvalEdgeZone[1] >= 0.0 );
      if ( piEdgeZone[1] == 1 )
	{
	  return APM2D::eztEnd;
	}
  }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( piEdgeZone[0] == -1 || piEdgeZone[1] == -1 )
    {
      APM2D::m_uiEdgeZoneVVEAntiparallelMPFUN++;

      mp_real rV0_x = dV0_x;
      mp_real rV0_y = dV0_y;
      mp_real rV1_x = dV1_x;
      mp_real rV1_y = dV1_y;
      mp_real rE20_x = dE20_x;
      mp_real rE20_y = dE20_y; 
      mp_real rE21_x = dE21_x;
      mp_real rE21_y = dE21_y;
      mp_real rE30_x = dE30_x;
      mp_real rE30_y = dE30_y;
      mp_real rE31_x = dE31_x;
      mp_real rE31_y = dE31_y;

      mp_real ppprV[4][2][2]; // Endpoints
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = 0.0;
      ppprV[0][0][1] = 0.0;
      ppprV[1][0][0] = rV1_x - rV0_x;
      ppprV[1][0][1] = rV1_y - rV0_y;
      ppprV[2][0][0] = rE20_x - rV0_x;
      ppprV[2][0][1] = rE20_y - rV0_y;
      ppprV[2][1][0] = rE21_x - rV0_x;
      ppprV[2][1][1] = rE21_y - rV0_y;
      ppprV[3][0][0] = rE30_x - rV0_x;
      ppprV[3][0][1] = rE30_y - rV0_y;
      ppprV[3][1][0] = rE31_x - rV0_x;
      ppprV[3][1][1] = rE31_y - rV0_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[2][0] = rE21_x - rE20_x;
      pprE[2][1] = rE21_y - rE20_y;
      pprE[3][0] = rE31_x - rE30_x;
      pprE[3][1] = rE31_y - rE30_y;
  
      mp_real rV1dotV1;
      Dot(rV1dotV1, ppprV[1][0][0], ppprV[1][0][1], ppprV[1][0][0], ppprV[1][0][1]);
      mp_real rV1dotE3;
      Dot(rV1dotE3, ppprV[1][0][0], ppprV[1][0][1], pprE[3][0], pprE[3][1]);
      mp_real rE2dotE2;
      Dot(rE2dotE2, pprE[2][0], pprE[2][1], pprE[2][0], pprE[2][1]);
      mp_real rE3dotE3;
      Dot(rE3dotE3, pprE[3][0], pprE[3][1], pprE[3][0], pprE[3][1]);

      mp_real rE3xV1;
      Cross(rE3xV1, pprE[3][0], pprE[3][1], ppprV[1][0][0], ppprV[1][0][1]);
      mp_real rE20xE21;
      Cross(rE20xE21, ppprV[2][0][0], ppprV[2][0][1], ppprV[2][1][0], ppprV[2][1][1]);

      mp_real r4E20xE21timesSqrtV1dotV1timesE2dotE2 = 4.0*rE20xE21*sqrt(rV1dotV1*rE2dotE2);

      mp_real rE2dotV1;
      Dot(rE2dotV1, pprE[2][0], pprE[2][1], ppprV[1][0][0], ppprV[1][0][1]);

      mp_real prEdgeZone[2];
      mp_real r4E20xE21timesE20xE21minusV1dotV1timesE2dotE2 = 4.0*rE20xE21*rE20xE21 - rV1dotV1*rE2dotE2;

      if ( piEdgeZone[0] == -1 )
	{
	  mp_real rE3dotE30;
	  Dot(rE3dotE30, pprE[3][0], pprE[3][1], ppprV[3][0][0], ppprV[3][0][1]);  
	  prEdgeZone[0] = ( r4E20xE21timesSqrtV1dotV1timesE2dotE2*( 2.0*rE3dotE30 - rV1dotE3 )
			    + ( r4E20xE21timesE20xE21minusV1dotV1timesE2dotE2 )*rE3xV1 );
	  if ( prEdgeZone[0] >= 0.0 )
	    {
	      return APM2D::eztStart;
	    }
	}

      if ( piEdgeZone[1] == -1 )
	{
	  mp_real rE3dotE31;
	  Dot(rE3dotE31, pprE[3][0], pprE[3][1], ppprV[3][1][0], ppprV[3][1][1]);
	  prEdgeZone[1] = -( r4E20xE21timesSqrtV1dotV1timesE2dotE2*( 2.0*rE3dotE31 - rV1dotE3 )
			     + ( r4E20xE21timesE20xE21minusV1dotV1timesE2dotE2 )*rE3xV1 );
	  if ( prEdgeZone[1] >= 0.0 )
	    {
	      return APM2D::eztEnd;
	    }
	}
    }

  return  APM2D::eztMiddle;
}

//static 
APM2D::CircleTestType InCircleVVE_VAntiparallelMPFUNFilter(DOUBLE dV0_x, DOUBLE dV0_y,
							   DOUBLE dV1_x, DOUBLE dV1_y,
							   DOUBLE dE20_x, DOUBLE dE20_y, 
							   DOUBLE dE21_x, DOUBLE dE21_y,
							   DOUBLE dV3_x, DOUBLE dV3_y)
{
  APM2D::m_uiInCircleVVE_VAntiparallelMPFUNFilter++;

  INT_32 iResult;

  iResult = -1;

  switch ( 0 )
    {
    default:
      Interval valV0_x = dV0_x;
      Interval valV0_y = dV0_y;
      Interval valV1_x = dV1_x;
      Interval valV1_y = dV1_y;
      Interval valE20_x = dE20_x;
      Interval valE20_y = dE20_y; 
      Interval valE21_x = dE21_x;
      Interval valE21_y = dE21_y;
      Interval valV3_x = dV3_x;
      Interval valV3_y = dV3_y;

      Interval pppvalV[4][2][2]; // Endpoints
      //memset(pppvalV, 0, sizeof(pppvalV));
      pppvalV[0][0][0] = 0.0;
      pppvalV[0][0][1] = 0.0;
      pppvalV[1][0][0] = valV1_x - valV0_x;
      pppvalV[1][0][1] = valV1_y - valV0_y;
      pppvalV[2][0][0] = valE20_x - valV0_x;
      pppvalV[2][0][1] = valE20_y - valV0_y;
      pppvalV[2][1][0] = valE21_x - valV0_x;
      pppvalV[2][1][1] = valE21_y - valV0_y;
      pppvalV[3][0][0] = valV3_x - valV0_x;
      pppvalV[3][0][1] = valV3_y - valV0_y;
  
      Interval ppvalE[4][2];     // Edge vectors
      //memset(ppvalE, 0, sizeof(ppvalE));
      ppvalE[2][0] = valE21_x - valE20_x;
      ppvalE[2][1] = valE21_y - valE20_y;

      Interval valV1dotV1;
      Dot(valV1dotV1, pppvalV[1][0][0], pppvalV[1][0][1], pppvalV[1][0][0], pppvalV[1][0][1]);
      Interval valE2dotE2;
      Dot(valE2dotE2, ppvalE[2][0], ppvalE[2][1], ppvalE[2][0], ppvalE[2][1]);
      Interval valV3dotV3;
      Dot(valV3dotV3, pppvalV[3][0][0], pppvalV[3][0][1], pppvalV[3][0][0], pppvalV[3][0][1]);
      Interval valV1dotV3;
      Dot(valV1dotV3, pppvalV[1][0][0], pppvalV[1][0][1], pppvalV[3][0][0], pppvalV[3][0][1]);

      Interval valV1xV3;
      Cross(valV1xV3, pppvalV[1][0][0], pppvalV[1][0][1], pppvalV[3][0][0], pppvalV[3][0][1]);
      Interval valE20xE21;
      Cross(valE20xE21, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][1][0], pppvalV[2][1][1]);

      Interval valDen;
      valDen = 4.0*valE20xE21*sqrt(valV1dotV1*valE2dotE2);
      INT_32 iDen;
      iDen = ( valDen != 0.0 );
      if ( iDen == -1 )
	{
	  break;
	}
      ASSERT( iDen == 1 );

      Interval valResult;
      valResult = ( valDen*( valV3dotV3 - valV1dotV3 )
		  - ( 4.0*valE20xE21*valE20xE21 - valV1dotV1*valE2dotE2 )*valV1xV3 );
  
      iResult = ( valResult < 0.0 );
      if ( iResult == -1 )
	{
	  break;
	}
      if ( iResult == 1 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( valResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
      return APM2D::cttOnCircle;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( iResult == -1 )
    {
      APM2D::m_uiInCircleVVE_VAntiparallelMPFUN++;

      mp_real rV0_x = dV0_x;
      mp_real rV0_y = dV0_y;
      mp_real rV1_x = dV1_x;
      mp_real rV1_y = dV1_y;
      mp_real rE20_x = dE20_x;
      mp_real rE20_y = dE20_y; 
      mp_real rE21_x = dE21_x;
      mp_real rE21_y = dE21_y;
      mp_real rV3_x = dV3_x;
      mp_real rV3_y = dV3_y;

      mp_real ppprV[4][2][2]; // Endpoints
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = 0.0;
      ppprV[0][0][1] = 0.0;
      ppprV[1][0][0] = rV1_x - rV0_x;
      ppprV[1][0][1] = rV1_y - rV0_y;
      ppprV[2][0][0] = rE20_x - rV0_x;
      ppprV[2][0][1] = rE20_y - rV0_y;
      ppprV[2][1][0] = rE21_x - rV0_x;
      ppprV[2][1][1] = rE21_y - rV0_y;
      ppprV[3][0][0] = rV3_x - rV0_x;
      ppprV[3][0][1] = rV3_y - rV0_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[2][0] = rE21_x - rE20_x;
      pprE[2][1] = rE21_y - rE20_y;

      mp_real rV1dotV1;
      Dot(rV1dotV1, ppprV[1][0][0], ppprV[1][0][1], ppprV[1][0][0], ppprV[1][0][1]);
      mp_real rE2dotE2;
      Dot(rE2dotE2, pprE[2][0], pprE[2][1], pprE[2][0], pprE[2][1]);
      mp_real rV3dotV3;
      Dot(rV3dotV3, ppprV[3][0][0], ppprV[3][0][1], ppprV[3][0][0], ppprV[3][0][1]);
      mp_real rV1dotV3;
      Dot(rV1dotV3, ppprV[1][0][0], ppprV[1][0][1], ppprV[3][0][0], ppprV[3][0][1]);

      mp_real rV1xV3;
      Cross(rV1xV3, ppprV[1][0][0], ppprV[1][0][1], ppprV[3][0][0], ppprV[3][0][1]);
      mp_real rE20xE21;
      Cross(rE20xE21, ppprV[2][0][0], ppprV[2][0][1], ppprV[2][1][0], ppprV[2][1][1]);

      mp_real rDen;
      rDen = 4.0*rE20xE21*sqrt(rV1dotV1*rE2dotE2);
      ASSERT( rDen != 0.0 );

      mp_real rResult;
      rResult = ( rDen*( rV3dotV3 - rV1dotV3 )
		  - ( 4.0*rE20xE21*rE20xE21 - rV1dotV1*rE2dotE2 )*rV1xV3 );
  
      if ( rResult < 0.0 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( rResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
    }

  return APM2D::cttOnCircle;
}

//static 
APM2D::CircleTestType InCircleVVE_EAntiparallelMPFUNFilter(DOUBLE dV0_x, DOUBLE dV0_y,
						     DOUBLE dV1_x, DOUBLE dV1_y,
						     DOUBLE dE20_x, DOUBLE dE20_y, 
						     DOUBLE dE21_x, DOUBLE dE21_y,
						     DOUBLE dE30_x, DOUBLE dE30_y,
						     DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::m_uiInCircleVVE_EAntiparallelMPFUNFilter++;

  INT_32 iResult;

  iResult = -1;

  switch ( 0 )
    {
    default:
      Interval valV0_x = dV0_x;
      Interval valV0_y = dV0_y;
      Interval valV1_x = dV1_x;
      Interval valV1_y = dV1_y;
      Interval valE20_x = dE20_x;
      Interval valE20_y = dE20_y; 
      Interval valE21_x = dE21_x;
      Interval valE21_y = dE21_y;
      Interval valE30_x = dE30_x;
      Interval valE30_y = dE30_y;
      Interval valE31_x = dE31_x;
      Interval valE31_y = dE31_y;

      Interval pppvalV[4][2][2]; // Endpoints
      //memset(pppvalV, 0, sizeof(pppvalV));
      pppvalV[0][0][0] = 0.0;
      pppvalV[0][0][1] = 0.0;
      pppvalV[1][0][0] = valV1_x - valV0_x;
      pppvalV[1][0][1] = valV1_y - valV0_y;
      pppvalV[2][0][0] = valE20_x - valV0_x;
      pppvalV[2][0][1] = valE20_y - valV0_y;
      pppvalV[2][1][0] = valE21_x - valV0_x;
      pppvalV[2][1][1] = valE21_y - valV0_y;
      pppvalV[3][0][0] = valE30_x - valV0_x;
      pppvalV[3][0][1] = valE30_y - valV0_y;
      pppvalV[3][1][0] = valE31_x - valV0_x;
      pppvalV[3][1][1] = valE31_y - valV0_y;
  
      Interval ppvalE[4][2];     // Edge vectors
      //memset(ppvalE, 0, sizeof(ppvalE));
      ppvalE[2][0] = valE21_x - valE20_x;
      ppvalE[2][1] = valE21_y - valE20_y;
      ppvalE[3][0] = valE31_x - valE30_x;
      ppvalE[3][1] = valE31_y - valE30_y;

      Interval valV1dotV1;
      Dot(valV1dotV1, pppvalV[1][0][0], pppvalV[1][0][1], pppvalV[1][0][0], pppvalV[1][0][1]);
      Interval valE2dotE2;
      Dot(valE2dotE2, ppvalE[2][0], ppvalE[2][1], ppvalE[2][0], ppvalE[2][1]);
      Interval valE3dotE3;
      Dot(valE3dotE3, ppvalE[3][0], ppvalE[3][1], ppvalE[3][0], ppvalE[3][1]);

      Interval valV1dotE2;
      Dot(valV1dotE2, pppvalV[1][0][0], pppvalV[1][0][1], ppvalE[2][0], ppvalE[2][1]);
      Interval valV1dotE3;
      Dot(valV1dotE3, pppvalV[1][0][0], pppvalV[1][0][1], ppvalE[3][0], ppvalE[3][1]);

      Interval valE3xV1;
      Cross(valE3xV1, ppvalE[3][0], ppvalE[3][1], pppvalV[1][0][0], pppvalV[1][0][1]);
      Interval valE20xE21;
      Cross(valE20xE21, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][1][0], pppvalV[2][1][1]);
      Interval valE30xE31;
      Cross(valE30xE31, pppvalV[3][0][0], pppvalV[3][0][1], pppvalV[3][1][0], pppvalV[3][1][1]);

      Interval valResult;
      valResult = ( 64.0*valE20xE21*valE20xE21*valV1dotE2*valV1dotE2*valE30xE31*( valE30xE31 + valE3xV1 )
		  - valE3xV1*valE3xV1*( 4.0*valE20xE21*valE20xE21 - valV1dotE2*valV1dotE2 )*( 4.0*valE20xE21*valE20xE21 - valV1dotE2*valV1dotE2 )
		  - 16.0*valE20xE21*valE20xE21*valV1dotE2*valV1dotE2*valV1dotE3*valV1dotE3
		  - 8.0*valE20xE21*valV1dotE2*valV1dotE3*( 4.0*valE20xE21*valE20xE21 - valV1dotE2*valV1dotE2 )*( 2.0*valE30xE31 + valE3xV1 ) );

      iResult = ( valResult < 0.0 );
      if ( iResult == -1 )
	{
	  break;
	}
      if ( iResult == 1 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( valResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
      return APM2D::cttOnCircle;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( iResult == -1 )
    {
      APM2D::m_uiInCircleVVE_EAntiparallelMPFUN++;

      mp_real rV0_x = dV0_x;
      mp_real rV0_y = dV0_y;
      mp_real rV1_x = dV1_x;
      mp_real rV1_y = dV1_y;
      mp_real rE20_x = dE20_x;
      mp_real rE20_y = dE20_y; 
      mp_real rE21_x = dE21_x;
      mp_real rE21_y = dE21_y;
      mp_real rE30_x = dE30_x;
      mp_real rE30_y = dE30_y;
      mp_real rE31_x = dE31_x;
      mp_real rE31_y = dE31_y;

      mp_real ppprV[4][2][2]; // Endpoints
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = 0.0;
      ppprV[0][0][1] = 0.0;
      ppprV[1][0][0] = rV1_x - rV0_x;
      ppprV[1][0][1] = rV1_y - rV0_y;
      ppprV[2][0][0] = rE20_x - rV0_x;
      ppprV[2][0][1] = rE20_y - rV0_y;
      ppprV[2][1][0] = rE21_x - rV0_x;
      ppprV[2][1][1] = rE21_y - rV0_y;
      ppprV[3][0][0] = rE30_x - rV0_x;
      ppprV[3][0][1] = rE30_y - rV0_y;
      ppprV[3][1][0] = rE31_x - rV0_x;
      ppprV[3][1][1] = rE31_y - rV0_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[2][0] = rE21_x - rE20_x;
      pprE[2][1] = rE21_y - rE20_y;
      pprE[3][0] = rE31_x - rE30_x;
      pprE[3][1] = rE31_y - rE30_y;

      mp_real rV1dotV1;
      Dot(rV1dotV1, ppprV[1][0][0], ppprV[1][0][1], ppprV[1][0][0], ppprV[1][0][1]);
      mp_real rE2dotE2;
      Dot(rE2dotE2, pprE[2][0], pprE[2][1], pprE[2][0], pprE[2][1]);
      mp_real rE3dotE3;
      Dot(rE3dotE3, pprE[3][0], pprE[3][1], pprE[3][0], pprE[3][1]);

      mp_real rV1dotE2;
      Dot(rV1dotE2, ppprV[1][0][0], ppprV[1][0][1], pprE[2][0], pprE[2][1]);
      mp_real rV1dotE3;
      Dot(rV1dotE3, ppprV[1][0][0], ppprV[1][0][1], pprE[3][0], pprE[3][1]);

      mp_real rE3xV1;
      Cross(rE3xV1, pprE[3][0], pprE[3][1], ppprV[1][0][0], ppprV[1][0][1]);
      mp_real rE20xE21;
      Cross(rE20xE21, ppprV[2][0][0], ppprV[2][0][1], ppprV[2][1][0], ppprV[2][1][1]);
      mp_real rE30xE31;
      Cross(rE30xE31, ppprV[3][0][0], ppprV[3][0][1], ppprV[3][1][0], ppprV[3][1][1]);

      mp_real rResult;
      rResult = ( 64.0*rE20xE21*rE20xE21*rV1dotE2*rV1dotE2*rE30xE31*( rE30xE31 + rE3xV1 )
		  - rE3xV1*rE3xV1*( 4.0*rE20xE21*rE20xE21 - rV1dotE2*rV1dotE2 )*( 4.0*rE20xE21*rE20xE21 - rV1dotE2*rV1dotE2 )
		  - 16.0*rE20xE21*rE20xE21*rV1dotE2*rV1dotE2*rV1dotE3*rV1dotE3
		  - 8.0*rE20xE21*rV1dotE2*rV1dotE3*( 4.0*rE20xE21*rE20xE21 - rV1dotE2*rV1dotE2 )*( 2.0*rE30xE31 + rE3xV1 ) );

      if ( rResult < 0.0 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( rResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
    }

  return APM2D::cttOnCircle;
}

//////////////////////////////////////////////////
// VvE

class ICC_VvE_MPFUNFilter
{
protected:
  enum InstructionPointer {ipStart, ipV0onE2, 
			   ipDen, 
			   ipVertexZones, ipEnd};

public:
  ICC_VvE_MPFUNFilter();
  ~ICC_VvE_MPFUNFilter();

public:
  APM2D::InscribedCircleCenterType 
  InscribedCircleCenterVvE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			   DOUBLE dV00_x, DOUBLE dV00_y,
			   DOUBLE dV0_x, DOUBLE dV0_y,
			   DOUBLE dV01_x, DOUBLE dV01_y,
			   DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
			   DOUBLE dE21_x, DOUBLE dE21_y);
  APM2D::InscribedCircleCenterType 
  InscribedCircleCenterVvEForce(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dV00_x, DOUBLE dV00_y,
				DOUBLE dV0_x, DOUBLE dV0_y,
				DOUBLE dV01_x, DOUBLE dV01_y,
				DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
				DOUBLE dE21_x, DOUBLE dE21_y);

protected:
  BOOL
  InscribedCircleCenterVvEMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dV00_x, DOUBLE dV00_y,
				DOUBLE dV0_x, DOUBLE dV0_y,
				DOUBLE dV01_x, DOUBLE dV01_y,
				DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
				DOUBLE dE21_x, DOUBLE dE21_y,
				InstructionPointer eIPMax = ipEnd);

protected:
  InstructionPointer m_eIP;
  APM2D::InscribedCircleCenterType m_eResult;
  INT_32 m_iV0onE2;
  INT_32 m_iDen;

  INT_32 m_piVertexZone[2];

  Interval m_valE2xV0;
  Interval m_valDen;

  mp_real m_rV00_x;
  mp_real m_rV00_y;
  mp_real m_rV0_x;
  mp_real m_rV0_y;
  mp_real m_rV01_x;
  mp_real m_rV01_y;
  mp_real m_rV1_E20_x;
  mp_real m_rV1_E20_y; 
  mp_real m_rE21_x;
  mp_real m_rE21_y;
  mp_real m_ppprV[3][2][2];
  mp_real m_pprE[3][2];
  mp_real m_rE2xV0;
  mp_real m_rDen;
  mp_real m_rV0dotV0;
};

ICC_VvE_MPFUNFilter::ICC_VvE_MPFUNFilter()
{
}

ICC_VvE_MPFUNFilter::~ICC_VvE_MPFUNFilter()
{
}

APM2D::InscribedCircleCenterType 
ICC_VvE_MPFUNFilter::InscribedCircleCenterVvE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					      DOUBLE dV00_x, DOUBLE dV00_y,
					      DOUBLE dV0_x, DOUBLE dV0_y,
					      DOUBLE dV01_x, DOUBLE dV01_y,
					      DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
					      DOUBLE dE21_x, DOUBLE dE21_y)
{
  APM2D::m_uiInscribedCircleCenterVvEMPFUNFilter++;

  *pdC_x = 0.0;
  *pdC_y = 0.0;
  *pdR = 0.0;
 
  m_eIP = ipStart;
  m_eResult = APM2D::icctNull;
  m_iV0onE2 = -1; 
  m_iDen = -1;

  m_piVertexZone[0] = -1;  
  m_piVertexZone[1] = -1;  

  BOOL bDone;

  Interval valV00_x = dV00_x;
  Interval valV00_y = dV00_y;
  Interval valV0_x = dV0_x;
  Interval valV0_y = dV0_y;
  Interval valV01_x = dV01_x;
  Interval valV01_y = dV01_y;
  Interval valV1_E20_x = dV1_E20_x;
  Interval valV1_E20_y = dV1_E20_y; 
  Interval valE21_x = dE21_x;
  Interval valE21_y = dE21_y;

  //////////////////////////////////////////////////
  // Check for V0 on E2

  Interval valV0onE2;
  SinTheta(valV0onE2, valV0_x, valV0_y, valV1_E20_x, valV1_E20_y, valE21_x, valE21_y);
  m_iV0onE2 = ( valV0onE2 <= 0.0 );
  if ( m_iV0onE2 == 1 )
    {
      // V0 is on E2 and it is not one of the endpoints
      return APM2D::icctNull;
    }
  if ( m_iV0onE2 == -1 )
    {
      bDone = InscribedCircleCenterVvEMPFUN(pdC_x, pdC_y, pdR, 
					    dV00_x, dV00_y,
					    dV0_x, dV0_y,
					    dV01_x, dV01_y,
					    dV1_E20_x, dV1_E20_y, 
					    dE21_x, dE21_y,
					    ipV0onE2);  
      if ( bDone )
	{
	  return m_eResult;
	}
    }

  //////////////////////////////////////////////////
  // Check E2xV0 and Den

  Interval pppvalV[3][2][2];
  // Translate V1 to the origin
  //memset(pppvalV, 0, sizeof(pppvalV));
  pppvalV[0][0][0] = valV0_x - valV1_E20_x;
  pppvalV[0][0][1] = valV0_y - valV1_E20_y;
  pppvalV[2][0][0] = 0.0;
  pppvalV[2][0][1] = 0.0;
  pppvalV[2][1][0] = valE21_x - valV1_E20_x;
  pppvalV[2][1][1] = valE21_y - valV1_E20_y;
  
  Interval ppvalE[3][2];
  //memset(ppvalE, 0, sizeof(ppvalE));
  ppvalE[2][0] = valE21_x - valV1_E20_x;
  ppvalE[2][1] = valE21_y - valV1_E20_y;
  
  Cross(m_valE2xV0, ppvalE[2][0], ppvalE[2][1], pppvalV[0][0][0], pppvalV[0][0][1]);
  m_valDen = 2.0*m_valE2xV0;
  m_iDen = ( m_valDen != 0.0 );
  if ( m_iDen == -1 )
    {
      bDone = InscribedCircleCenterVvEMPFUN(pdC_x, pdC_y, pdR, 
					    dV00_x, dV00_y,
					    dV0_x, dV0_y,
					    dV01_x, dV01_y,
					    dV1_E20_x, dV1_E20_y, 
					    dE21_x, dE21_y,
					    ipDen);  
      if ( bDone )
	{
	  return m_eResult;
	}
      if ( m_iDen == -1 )
	{
	  bDone = InscribedCircleCenterVvEMPFUN(pdC_x, pdC_y, pdR, 
						dV00_x, dV00_y,
						dV0_x, dV0_y,
						dV01_x, dV01_y,
						dV1_E20_x, dV1_E20_y, 
						dE21_x, dE21_y,
						ipEnd);  
	  ASSERT( bDone );
	  return m_eResult;
	}
    }
  ASSERT( m_iDen == 1 );

  Interval valV0dotV0;
  Dot(valV0dotV0, pppvalV[0][0][0], pppvalV[0][0][1], pppvalV[0][0][0], pppvalV[0][0][1]);
  
  //////////////////////////////////////////////////
  // Check the Vertex Zones

  Interval pppvalVN[1][2][2];
  // Vertex neighbor vectors
  //memset(pppvalVN, 0, sizeof(pppvalV));
  pppvalVN[0][0][0] = valV0_x - valV00_x;
  pppvalVN[0][0][1] = valV0_y - valV00_y;
  pppvalVN[0][1][0] = valV0_x - valV01_x;
  pppvalVN[0][1][1] = valV0_y - valV01_y;

  Interval pvalVertexZone[2];

  Interval valV0dotV00;
  Dot(valV0dotV00, pppvalV[0][0][0], pppvalV[0][0][1], pppvalVN[0][0][0], pppvalVN[0][0][1]);
  Interval valE2xV00;
  Cross(valE2xV00, ppvalE[2][0], ppvalE[2][1], pppvalVN[0][0][0], pppvalVN[0][0][1]);
  pvalVertexZone[0] = ( valV0dotV0*valE2xV00 - 2.0*m_valE2xV0*valV0dotV00 );
  m_piVertexZone[0] = ( pvalVertexZone[0] < 0.0 );
  if ( m_piVertexZone[0] == 1 ) 
    {
      return APM2D::icctNull;
    }
  Interval valV0dotV01;
  Dot(valV0dotV01, pppvalV[0][0][0], pppvalV[0][0][1], pppvalVN[0][1][0], pppvalVN[0][1][1]);
  Interval valE2xV01;
  Cross(valE2xV01, ppvalE[2][0], ppvalE[2][1], pppvalVN[0][1][0], pppvalVN[0][1][1]);
  pvalVertexZone[1] = ( valV0dotV0*valE2xV01 - 2.0*m_valE2xV0*valV0dotV01 );
  m_piVertexZone[1] = ( pvalVertexZone[1] < 0.0 );
  if ( m_piVertexZone[1] == 1 ) 
    {
      return APM2D::icctNull;
    }

  if ( m_piVertexZone[0] == -1 ||
       m_piVertexZone[1] == -1 )
    {
      bDone = InscribedCircleCenterVvEMPFUN(pdC_x, pdC_y, pdR, 
					    dV00_x, dV00_y,
					    dV0_x, dV0_y,
					    dV01_x, dV01_y,
					    dV1_E20_x, dV1_E20_y, 
					    dE21_x, dE21_y,
					    ipVertexZones);  
      if ( bDone )
	{
	  return m_eResult;
	}
    }

  //////////////////////////////////////////////////
  // Compute the center and radius

  Interval pvalC[2];
  pvalC[0] = -valV0dotV0*ppvalE[2][1]/m_valDen;
  pvalC[1] = valV0dotV0*ppvalE[2][0]/m_valDen;

  Interval valE2dotE2;
  Dot(valE2dotE2, ppvalE[2][0], ppvalE[2][1], ppvalE[2][0], ppvalE[2][1]);

  Interval valR;
  valR = (valV0dotV0*sqrt(valE2dotE2)) / m_valDen;

  // Translate the answer back out away from the origin
  pvalC[0] += valV1_E20_x;
  pvalC[1] += valV1_E20_y;

  *pdC_x = dble(pvalC[0]);
  *pdC_y = dble(pvalC[1]);
  *pdR = dble(valR);
  return APM2D::icctVvE;
}

APM2D::InscribedCircleCenterType 
ICC_VvE_MPFUNFilter::InscribedCircleCenterVvEForce(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
						   DOUBLE dV00_x, DOUBLE dV00_y,
						   DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dV01_x, DOUBLE dV01_y,
						   DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y)
{
  *pdC_x = 0.0;
  *pdC_y = 0.0;
  *pdR = 0.0;
 
  m_eIP = ipStart;
  m_eResult = APM2D::icctNull;
  m_iV0onE2 = -1; 
  m_iDen = -1;

  m_piVertexZone[0] = -1;  
  m_piVertexZone[1] = -1;  

  BOOL bDone;
  bDone = InscribedCircleCenterVvEMPFUN(pdC_x, pdC_y, pdR, 
					dV00_x, dV00_y,
				        dV0_x, dV0_y,
					dV01_x, dV01_y,
					dV1_E20_x, dV1_E20_y, 
					dE21_x, dE21_y);
  if ( !bDone ) 
    {
      fprintf(stderr, "InscribedCircleCenterVvEMPFUN does not work!!!\n");
      exit(1);
    }
  return m_eResult;
}

BOOL
ICC_VvE_MPFUNFilter::InscribedCircleCenterVvEMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
						   DOUBLE dV00_x, DOUBLE dV00_y,
						   DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dV01_x, DOUBLE dV01_y,
						   DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   InstructionPointer eIPMax/* = ipEnd*/)
{
  APM2D::m_uiInscribedCircleCenterVvEMPFUN++;

  switch ( m_eIP )
    {
    case ipStart:
      {
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	*pdC_x = 0.0;
	*pdC_y = 0.0;
	*pdR = 0.0;

	m_rV00_x = dV00_x;
	m_rV00_y = dV00_y;
	m_rV0_x = dV0_x;
	m_rV0_y = dV0_y;
	m_rV01_x = dV01_x;
	m_rV01_y = dV01_y;
	m_rV1_E20_x = dV1_E20_x;
	m_rV1_E20_y = dV1_E20_y; 
	m_rE21_x = dE21_x;
	m_rE21_y = dE21_y;

	//////////////////////////////////////////////////
	// Check for V0 on E2
      
	if ( m_iV0onE2 == -1 )
	  {
	    mp_real rV0onE2;
	    SinTheta(rV0onE2, m_rV0_x, m_rV0_y, m_rV1_E20_x, m_rV1_E20_y, m_rE21_x, m_rE21_y);
	    if ( rV0onE2 <= 0.0 )
	      {
		// V0 is on E2 and it is not one of the endpoints
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterVvEMPFUNFinal++;
		return TRUE;
	      }
	  }
  
      }
    case ipV0onE2:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Check E2xV0 and Den
      
	// Translate V1 to the origin
	//memset(m_ppprV, 0, sizeof(m_ppprV));
	m_ppprV[0][0][0] = m_rV0_x - m_rV1_E20_x;
	m_ppprV[0][0][1] = m_rV0_y - m_rV1_E20_y;
	m_ppprV[2][0][0] = 0.0;
	m_ppprV[2][0][1] = 0.0;
	m_ppprV[2][1][0] = m_rE21_x - m_rV1_E20_x;
	m_ppprV[2][1][1] = m_rE21_y - m_rV1_E20_y;
  
	//memset(m_pprE, 0, sizeof(m_pprE));
	m_pprE[2][0] = m_rE21_x - m_rV1_E20_x;
	m_pprE[2][1] = m_rE21_y - m_rV1_E20_y;
  
	Cross(m_rE2xV0, m_pprE[2][0], m_pprE[2][1], m_ppprV[0][0][0], m_ppprV[0][0][1]);
  
	m_rDen = 2.0*m_rE2xV0;

	if ( m_iDen == -1 )
	  {
	    ASSERT( m_rDen != 0.0 );
	    if ( eIPMax != ipEnd )
	      { 
		m_valE2xV0.ConservativeBound(dble(m_rE2xV0));
		m_valDen.ConservativeBound(dble(m_rDen));	    
		m_iDen = ( m_valDen != 0.0 );      
	      }
	  }
      }
    case ipDen:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Check the Vertex Zones
	Dot(m_rV0dotV0, m_ppprV[0][0][0], m_ppprV[0][0][1], m_ppprV[0][0][0], m_ppprV[0][0][1]);

	if ( m_piVertexZone[0] == -1 || 
	     m_piVertexZone[1] == -1 )
	  {      
	    mp_real ppprVN[1][2][2];
	    // Vertex neighbor vectors
	    //memset(ppprVN, 0, sizeof(ppprVN));
	    ppprVN[0][0][0] = m_rV0_x - m_rV00_x;
	    ppprVN[0][0][1] = m_rV0_y - m_rV00_y;
	    ppprVN[0][1][0] = m_rV0_x - m_rV01_x;
	    ppprVN[0][1][1] = m_rV0_y - m_rV01_y;

	    mp_real prVertexZone[2];
	    if ( m_piVertexZone[0] == -1 )
	      {
		mp_real rV0dotV00;
		Dot(rV0dotV00, m_ppprV[0][0][0], m_ppprV[0][0][1], ppprVN[0][0][0], ppprVN[0][0][1]);
		mp_real rE2xV00;
		Cross(rE2xV00, m_pprE[2][0], m_pprE[2][1], ppprVN[0][0][0], ppprVN[0][0][1]);
		prVertexZone[0] = ( m_rV0dotV0*rE2xV00 - 2.0*m_rE2xV0*rV0dotV00 );
		if ( prVertexZone[0] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVvEMPFUNFinal++;
		    return TRUE;
		  }
	      }

	    if ( m_piVertexZone[0] == -1 )
	      {
		mp_real rV0dotV01;
		Dot(rV0dotV01, m_ppprV[0][0][0], m_ppprV[0][0][1], ppprVN[0][1][0], ppprVN[0][1][1]);
		mp_real rE2xV01;
		Cross(rE2xV01, m_pprE[2][0], m_pprE[2][1], ppprVN[0][1][0], ppprVN[0][1][1]);
		prVertexZone[1] = ( m_rV0dotV0*rE2xV01 - 2.0*m_rE2xV0*rV0dotV01 );
		if ( prVertexZone[0] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVvEMPFUNFinal++;
		    return TRUE;
		  }
	      }
	  }

      }
    case ipVertexZones:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Compute the center and radius
 
	mp_real prC[2];
	prC[0] = -m_rV0dotV0*m_pprE[2][1]/m_rDen;
	prC[1] = m_rV0dotV0*m_pprE[2][0]/m_rDen;

	mp_real rE2dotE2;
	Dot(rE2dotE2, m_pprE[2][0], m_pprE[2][1], m_pprE[2][0], m_pprE[2][1]);

	mp_real rR;
	rR = (m_rV0dotV0*sqrt(rE2dotE2)) / m_rDen;

	// Translate the answer back out away from the origin
	prC[0] += m_rV1_E20_x;
	prC[1] += m_rV1_E20_y;

	*pdC_x = dble(prC[0]);
	*pdC_y = dble(prC[1]);
	*pdR = dble(rR);
	m_eResult = APM2D::icctVvE;
      }
      break;
      /*
    default:
      ASSERT( FALSE );
      break;
      */
    }

  APM2D::m_uiInscribedCircleCenterVvEMPFUNFinal++;
  return TRUE;
}

//static 
APM2D::InscribedCircleCenterType 
InscribedCircleCenterVvEMPFUNFilter(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				    DOUBLE dV00_x, DOUBLE dV00_y,
				    DOUBLE dV0_x, DOUBLE dV0_y,
				    DOUBLE dV01_x, DOUBLE dV01_y,
				    DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
				    DOUBLE dE21_x, DOUBLE dE21_y)
{
  ICC_VvE_MPFUNFilter iccVvE;

  return iccVvE.InscribedCircleCenterVvE(pdC_x, pdC_y, pdR, 
					 dV00_x, dV00_y,
					 dV0_x, dV0_y,
					 dV01_x, dV01_y,
					 dV1_E20_x, dV1_E20_y, 
					 dE21_x, dE21_y);
}

//static 
APM2D::EdgeZoneType EdgeZoneVvEMPFUNFilter(DOUBLE dV0_x, DOUBLE dV0_y,
					   DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
					   DOUBLE dE21_x, DOUBLE dE21_y,
					   DOUBLE dE30_x, DOUBLE dE30_y,
					   DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::m_uiEdgeZoneVvEMPFUNFilter++;

  INT_32 piEdgeZone[2];
  piEdgeZone[0] = -1;
  piEdgeZone[1] = -1;

  switch ( 0 )
  {
  default:
    Interval valV0_x = dV0_x;
    Interval valV0_y = dV0_y;
    Interval valV1_E20_x = dV1_E20_x;
    Interval valV1_E20_y = dV1_E20_y; 
    Interval valE21_x = dE21_x;
    Interval valE21_y = dE21_y;
    Interval valE30_x = dE30_x;
    Interval valE30_y = dE30_y;
    Interval valE31_x = dE31_x;
    Interval valE31_y = dE31_y;

    Interval pppvalV[4][2][2]; // Endpoints
    //memset(pppvalV, 0, sizeof(pppvalV));
    pppvalV[0][0][0] = valV0_x - valV1_E20_x;
    pppvalV[0][0][1] = valV0_y - valV1_E20_y;
    pppvalV[1][0][0] = 0.0;
    pppvalV[1][0][1] = 0.0;
    pppvalV[2][0][0] = 0.0;
    pppvalV[2][0][1] = 0.0;
    pppvalV[2][1][0] = valE21_x - valV1_E20_x;
    pppvalV[2][1][1] = valE21_y - valV1_E20_y;
    pppvalV[3][0][0] = valE30_x - valV1_E20_x;
    pppvalV[3][0][1] = valE30_y - valV1_E20_y;
    pppvalV[3][1][0] = valE31_x - valV1_E20_x;
    pppvalV[3][1][1] = valE31_y - valV1_E20_y;
  
    Interval ppvalE[4][2];     // Edge vectors
    //memset(ppvalE, 0, sizeof(ppvalE));
    ppvalE[2][0] = valE21_x - valV1_E20_x;
    ppvalE[2][1] = valE21_y - valV1_E20_y;
    ppvalE[3][0] = valE31_x - valE30_x;
    ppvalE[3][1] = valE31_y - valE30_y;
  
    Interval valV0dotV0;
    Dot(valV0dotV0, pppvalV[0][0][0], pppvalV[0][0][1], pppvalV[0][0][0], pppvalV[0][0][1]);

    Interval valE2xV0;
    Cross(valE2xV0, ppvalE[2][0], ppvalE[2][1], pppvalV[0][0][0], pppvalV[0][0][1]);
    Interval valE2xE3;
    Cross(valE2xE3, ppvalE[2][0], ppvalE[2][1], ppvalE[3][0], ppvalE[3][1]);

    Interval valV0dotV0timesE2xE3 = valV0dotV0*valE2xE3;

    Interval pvalEdgeZone[2];
  
    Interval valE3dotE30;
    Dot(valE3dotE30, ppvalE[3][0], ppvalE[3][1], pppvalV[3][0][0], pppvalV[3][0][1]);
    pvalEdgeZone[0] = ( 2.0*valE2xV0*valE3dotE30 - valV0dotV0timesE2xE3 );
    piEdgeZone[0] = ( pvalEdgeZone[0] >= 0.0 );
    if ( piEdgeZone[0] == 1 )
      {
	return APM2D::eztStart;
      }

    Interval valE3dotE31;
    Dot(valE3dotE31, ppvalE[3][0], ppvalE[3][1], pppvalV[3][1][0], pppvalV[3][1][1]);
    pvalEdgeZone[1] = -( 2.0*valE2xV0*valE3dotE31 - valV0dotV0timesE2xE3 );
    piEdgeZone[1] = ( pvalEdgeZone[1] >= 0.0 );
    if ( piEdgeZone[1] == 1 )
      {
	return APM2D::eztEnd;
      }
  }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( piEdgeZone[0] == -1 || piEdgeZone[1] == -1 )
    {
      APM2D::m_uiEdgeZoneVvEMPFUN++;

      mp_real rV0_x = dV0_x;
      mp_real rV0_y = dV0_y;
      mp_real rV1_E20_x = dV1_E20_x;
      mp_real rV1_E20_y = dV1_E20_y; 
      mp_real rE21_x = dE21_x;
      mp_real rE21_y = dE21_y;
      mp_real rE30_x = dE30_x;
      mp_real rE30_y = dE30_y;
      mp_real rE31_x = dE31_x;
      mp_real rE31_y = dE31_y;

      mp_real ppprV[4][2][2]; // Endpoints
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = rV0_x - rV1_E20_x;
      ppprV[0][0][1] = rV0_y - rV1_E20_y;
      ppprV[1][0][0] = 0.0;
      ppprV[1][0][1] = 0.0;
      ppprV[2][0][0] = 0.0;
      ppprV[2][0][1] = 0.0;
      ppprV[2][1][0] = rE21_x - rV1_E20_x;
      ppprV[2][1][1] = rE21_y - rV1_E20_y;
      ppprV[3][0][0] = rE30_x - rV1_E20_x;
      ppprV[3][0][1] = rE30_y - rV1_E20_y;
      ppprV[3][1][0] = rE31_x - rV1_E20_x;
      ppprV[3][1][1] = rE31_y - rV1_E20_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[2][0] = rE21_x - rV1_E20_x;
      pprE[2][1] = rE21_y - rV1_E20_y;
      pprE[3][0] = rE31_x - rE30_x;
      pprE[3][1] = rE31_y - rE30_y;
  
      mp_real rV0dotV0;
      Dot(rV0dotV0, ppprV[0][0][0], ppprV[0][0][1], ppprV[0][0][0], ppprV[0][0][1]);

      mp_real rE2xV0;
      Cross(rE2xV0, pprE[2][0], pprE[2][1], ppprV[0][0][0], ppprV[0][0][1]);
      mp_real rE2xE3;
      Cross(rE2xE3, pprE[2][0], pprE[2][1], pprE[3][0], pprE[3][1]);

      mp_real rV0dotV0timesE2xE3 = rV0dotV0*rE2xE3;

      mp_real prEdgeZone[2];
  
      if ( piEdgeZone[0] == -1 )
	{
	  mp_real rE3dotE30;
	  Dot(rE3dotE30, pprE[3][0], pprE[3][1], ppprV[3][0][0], ppprV[3][0][1]);
	  prEdgeZone[0] = ( 2.0*rE2xV0*rE3dotE30 - rV0dotV0timesE2xE3 );
	  if ( prEdgeZone[0] >= 0.0 )
	    {
	      return APM2D::eztStart;
	    }
	}

      if ( piEdgeZone[1] == -1 )
	{
	  mp_real rE3dotE31;
	  Dot(rE3dotE31, pprE[3][0], pprE[3][1], ppprV[3][1][0], ppprV[3][1][1]);
	  prEdgeZone[1] = -( 2.0*rE2xV0*rE3dotE31 - rV0dotV0timesE2xE3 );
	  if ( prEdgeZone[1] >= 0.0 )
	    {
	      return APM2D::eztEnd;
	    }
	}
    }

  return APM2D::eztMiddle;
}

//static 
APM2D::CircleTestType InCircleVvE_VMPFUNFilter(DOUBLE dV0_x, DOUBLE dV0_y,
					       DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
					       DOUBLE dE21_x, DOUBLE dE21_y,
					       DOUBLE dV3_x, DOUBLE dV3_y)
{
  APM2D::m_uiInCircleVvE_VMPFUNFilter++;

  INT_32 iResult;

  iResult = -1;

  switch ( 0 )
    {
    default:
      Interval valV0_x = dV0_x;
      Interval valV0_y = dV0_y;
      Interval valV1_E20_x = dV1_E20_x;
      Interval valV1_E20_y = dV1_E20_y; 
      Interval valE21_x = dE21_x;
      Interval valE21_y = dE21_y;
      Interval valV3_x = dV3_x;
      Interval valV3_y = dV3_y;

      Interval pppvalV[4][2][2]; // Endpoints
      //memset(pppvalV, 0, sizeof(pppvalV));
      pppvalV[0][0][0] = valV0_x - valV1_E20_x;
      pppvalV[0][0][1] = valV0_y - valV1_E20_y;
      pppvalV[1][0][0] = 0.0;
      pppvalV[1][0][1] = 0.0;
      pppvalV[2][0][0] = 0.0;
      pppvalV[2][0][1] = 0.0;
      pppvalV[2][1][0] = valE21_x - valV1_E20_x;
      pppvalV[2][1][1] = valE21_y - valV1_E20_y;
      pppvalV[3][0][0] = valV3_x - valV1_E20_x;
      pppvalV[3][0][1] = valV3_y - valV1_E20_y;
  
      Interval ppvalE[4][2];     // Edge vectors
      //memset(ppvalE, 0, sizeof(ppvalE));
      ppvalE[2][0] = valE21_x - valV1_E20_x;
      ppvalE[2][1] = valE21_y - valV1_E20_y;
  
      Interval valV0dotV0;
      Dot(valV0dotV0, pppvalV[0][0][0], pppvalV[0][0][1], pppvalV[0][0][0], pppvalV[0][0][1]);
      Interval valV3dotV3;
      Dot(valV3dotV3, pppvalV[3][0][0], pppvalV[3][0][1], pppvalV[3][0][0], pppvalV[3][0][1]);

      Interval valE2xV0;
      Cross(valE2xV0, ppvalE[2][0], ppvalE[2][1], pppvalV[0][0][0], pppvalV[0][0][1]);
      Interval valE2xV3;
      Cross(valE2xV3, ppvalE[2][0], ppvalE[2][1], pppvalV[3][0][0], pppvalV[3][0][1]);

      Interval valResult;
      valResult = ( valV3dotV3*valE2xV0 - valV0dotV0*valE2xV3 );

      iResult = ( valResult < 0.0 );
      if ( iResult == -1 )
	{
	  break;
	}
      if ( iResult == 1 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( valResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
      return APM2D::cttOnCircle;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( iResult == -1 )
    {
      APM2D::m_uiInCircleVvE_VMPFUN++;

      mp_real rV0_x = dV0_x;
      mp_real rV0_y = dV0_y;
      mp_real rV1_E20_x = dV1_E20_x;
      mp_real rV1_E20_y = dV1_E20_y; 
      mp_real rE21_x = dE21_x;
      mp_real rE21_y = dE21_y;
      mp_real rV3_x = dV3_x;
      mp_real rV3_y = dV3_y;

      mp_real ppprV[4][2][2]; // Endpoints
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = rV0_x - rV1_E20_x;
      ppprV[0][0][1] = rV0_y - rV1_E20_y;
      ppprV[1][0][0] = 0.0;
      ppprV[1][0][1] = 0.0;
      ppprV[2][0][0] = 0.0;
      ppprV[2][0][1] = 0.0;
      ppprV[2][1][0] = rE21_x - rV1_E20_x;
      ppprV[2][1][1] = rE21_y - rV1_E20_y;
      ppprV[3][0][0] = rV3_x - rV1_E20_x;
      ppprV[3][0][1] = rV3_y - rV1_E20_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[2][0] = rE21_x - rV1_E20_x;
      pprE[2][1] = rE21_y - rV1_E20_y;
  
      mp_real rV0dotV0;
      Dot(rV0dotV0, ppprV[0][0][0], ppprV[0][0][1], ppprV[0][0][0], ppprV[0][0][1]);
      mp_real rV3dotV3;
      Dot(rV3dotV3, ppprV[3][0][0], ppprV[3][0][1], ppprV[3][0][0], ppprV[3][0][1]);

      mp_real rE2xV0;
      Cross(rE2xV0, pprE[2][0], pprE[2][1], ppprV[0][0][0], ppprV[0][0][1]);
      mp_real rE2xV3;
      Cross(rE2xV3, pprE[2][0], pprE[2][1], ppprV[3][0][0], ppprV[3][0][1]);

      mp_real rResult;
      rResult = ( rV3dotV3*rE2xV0 - rV0dotV0*rE2xV3 );
  
      if ( rResult < 0.0 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( rResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
    }

  return APM2D::cttOnCircle;
}

//static 
APM2D::CircleTestType InCircleVvE_EMPFUNFilter(DOUBLE dV0_x, DOUBLE dV0_y,
					       DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
					       DOUBLE dE21_x, DOUBLE dE21_y,
					       DOUBLE dE30_x, DOUBLE dE30_y,
					       DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::m_uiInCircleVvE_EMPFUNFilter++;

  INT_32 iResult;

  iResult = -1;

  switch ( 0 )
    {
    default:
      Interval valV0_x = dV0_x;
      Interval valV0_y = dV0_y;
      Interval valV1_E20_x = dV1_E20_x;
      Interval valV1_E20_y = dV1_E20_y; 
      Interval valE21_x = dE21_x;
      Interval valE21_y = dE21_y;
      Interval valE30_x = dE30_x;
      Interval valE30_y = dE30_y;
      Interval valE31_x = dE31_x;
      Interval valE31_y = dE31_y;

      Interval pppvalV[4][2][2]; // Endpoints
      //memset(pppvalV, 0, sizeof(pppvalV));
      pppvalV[0][0][0] = valV0_x - valV1_E20_x;
      pppvalV[0][0][1] = valV0_y - valV1_E20_y;
      pppvalV[1][0][0] = 0.0;
      pppvalV[1][0][1] = 0.0;
      pppvalV[2][0][0] = 0.0;
      pppvalV[2][0][1] = 0.0;
      pppvalV[2][1][0] = valE21_x - valV1_E20_x;
      pppvalV[2][1][1] = valE21_y - valV1_E20_y;
      pppvalV[3][0][0] = valE30_x - valV1_E20_x;
      pppvalV[3][0][1] = valE30_y - valV1_E20_y;
      pppvalV[3][1][0] = valE31_x - valV1_E20_x;
      pppvalV[3][1][1] = valE31_y - valV1_E20_y;
  
      Interval ppvalE[4][2];     // Edge vectors
      //memset(ppvalE, 0, sizeof(ppvalE));
      ppvalE[2][0] = valE21_x - valV1_E20_x;
      ppvalE[2][1] = valE21_y - valV1_E20_y;
      ppvalE[3][0] = valE31_x - valE30_x;
      ppvalE[3][1] = valE31_y - valE30_y;
  
      Interval valV0dotV0;
      Dot(valV0dotV0, pppvalV[0][0][0], pppvalV[0][0][1], pppvalV[0][0][0], pppvalV[0][0][1]);
      Interval valE2dotE2;
      Dot(valE2dotE2, ppvalE[2][0], ppvalE[2][1], ppvalE[2][0], ppvalE[2][1]);
      Interval valE3dotE3;
      Dot(valE3dotE3, ppvalE[3][0], ppvalE[3][1], ppvalE[3][0], ppvalE[3][1]);

      Interval valE2dotE3;
      Dot(valE2dotE3, ppvalE[2][0], ppvalE[2][1], ppvalE[3][0], ppvalE[3][1]);

      Interval valE2xV0;
      Cross(valE2xV0, ppvalE[2][0], ppvalE[2][1], pppvalV[0][0][0], pppvalV[0][0][1]);
      Interval valE30xE31;
      Cross(valE30xE31, pppvalV[3][0][0], pppvalV[3][0][1], pppvalV[3][1][0], pppvalV[3][1][1]);

      Interval valResult;
      valResult = ( 4.0*valE2xV0*valE30xE31*( valE2xV0*valE30xE31 + valV0dotV0*valE2dotE3 )
		  - valV0dotV0*valV0dotV0*( valE2dotE2*valE3dotE3 - valE2dotE3*valE2dotE3 ) );

      iResult = ( valResult < 0.0 );
      if ( iResult == -1 )
	{
	  break;
	}
      if ( iResult == 1 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( valResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
      return APM2D::cttOnCircle;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( iResult == -1 )
    {
      APM2D::m_uiInCircleVvE_EMPFUN++;

      mp_real rV0_x = dV0_x;
      mp_real rV0_y = dV0_y;
      mp_real rV1_E20_x = dV1_E20_x;
      mp_real rV1_E20_y = dV1_E20_y; 
      mp_real rE21_x = dE21_x;
      mp_real rE21_y = dE21_y;
      mp_real rE30_x = dE30_x;
      mp_real rE30_y = dE30_y;
      mp_real rE31_x = dE31_x;
      mp_real rE31_y = dE31_y;

      mp_real ppprV[4][2][2]; // Endpoints
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = rV0_x - rV1_E20_x;
      ppprV[0][0][1] = rV0_y - rV1_E20_y;
      ppprV[1][0][0] = 0.0;
      ppprV[1][0][1] = 0.0;
      ppprV[2][0][0] = 0.0;
      ppprV[2][0][1] = 0.0;
      ppprV[2][1][0] = rE21_x - rV1_E20_x;
      ppprV[2][1][1] = rE21_y - rV1_E20_y;
      ppprV[3][0][0] = rE30_x - rV1_E20_x;
      ppprV[3][0][1] = rE30_y - rV1_E20_y;
      ppprV[3][1][0] = rE31_x - rV1_E20_x;
      ppprV[3][1][1] = rE31_y - rV1_E20_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[2][0] = rE21_x - rV1_E20_x;
      pprE[2][1] = rE21_y - rV1_E20_y;
      pprE[3][0] = rE31_x - rE30_x;
      pprE[3][1] = rE31_y - rE30_y;
  
      mp_real rV0dotV0;
      Dot(rV0dotV0, ppprV[0][0][0], ppprV[0][0][1], ppprV[0][0][0], ppprV[0][0][1]);
      mp_real rE2dotE2;
      Dot(rE2dotE2, pprE[2][0], pprE[2][1], pprE[2][0], pprE[2][1]);
      mp_real rE3dotE3;
      Dot(rE3dotE3, pprE[3][0], pprE[3][1], pprE[3][0], pprE[3][1]);

      mp_real rE2dotE3;
      Dot(rE2dotE3, pprE[2][0], pprE[2][1], pprE[3][0], pprE[3][1]);

      mp_real rE2xV0;
      Cross(rE2xV0, pprE[2][0], pprE[2][1], ppprV[0][0][0], ppprV[0][0][1]);
      mp_real rE30xE31;
      Cross(rE30xE31, ppprV[3][0][0], ppprV[3][0][1], ppprV[3][1][0], ppprV[3][1][1]);

      mp_real rResult;
      rResult = ( 4.0*rE2xV0*rE30xE31*( rE2xV0*rE30xE31 + rV0dotV0*rE2dotE3 )
		  - rV0dotV0*rV0dotV0*( rE2dotE2*rE3dotE3 - rE2dotE3*rE2dotE3 ) );

      if ( rResult < 0.0 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( rResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
    }
  
  return APM2D::cttOnCircle;
}

//////////////////////////////////////////////////
// EvV

class ICC_EvV_MPFUNFilter
{
protected:
  enum InstructionPointer {ipStart, ipV2onE0, 
			   ipDen, 
			   ipVertexZones, ipEnd};

public:
  ICC_EvV_MPFUNFilter();
  ~ICC_EvV_MPFUNFilter();

public:
  APM2D::InscribedCircleCenterType 
  InscribedCircleCenterEvV(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			   DOUBLE dE00_x, DOUBLE dE00_y,
			   DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
			   DOUBLE dV20_x, DOUBLE dV20_y,
			   DOUBLE dV2_x, DOUBLE dV2_y,
			   DOUBLE dV21_x, DOUBLE dV21_y);
  APM2D::InscribedCircleCenterType 
  InscribedCircleCenterEvVForce(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dE00_x, DOUBLE dE00_y,
				DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
				DOUBLE dV20_x, DOUBLE dV20_y,
				DOUBLE dV2_x, DOUBLE dV2_y,
				DOUBLE dV21_x, DOUBLE dV21_y);

protected:
  BOOL
  InscribedCircleCenterEvVMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
                                DOUBLE dE00_x, DOUBLE dE00_y,
                                DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
				DOUBLE dV20_x, DOUBLE dV20_y,
				DOUBLE dV2_x, DOUBLE dV2_y,
				DOUBLE dV21_x, DOUBLE dV21_y,
                                InstructionPointer eIPMax = ipEnd);

protected:
  InstructionPointer m_eIP;
  APM2D::InscribedCircleCenterType m_eResult;
  INT_32 m_iV2onE0;
  INT_32 m_iDen;

  INT_32 m_piVertexZone[2];

  Interval m_valE0xV2;
  Interval m_valDen;

  mp_real m_rE00_x;
  mp_real m_rE00_y;
  mp_real m_rE01_V1_x;
  mp_real m_rE01_V1_y; 
  mp_real m_rV20_x;
  mp_real m_rV20_y;
  mp_real m_rV2_x;
  mp_real m_rV2_y;
  mp_real m_rV21_x;
  mp_real m_rV21_y;
  mp_real m_ppprV[3][2][2];
  mp_real m_pprE[3][2];
  mp_real m_rE0xV2;
  mp_real m_rDen;
  mp_real m_rV2dotV2;
};

ICC_EvV_MPFUNFilter::ICC_EvV_MPFUNFilter()
{
}

ICC_EvV_MPFUNFilter::~ICC_EvV_MPFUNFilter()
{
}

APM2D::InscribedCircleCenterType 
ICC_EvV_MPFUNFilter::InscribedCircleCenterEvV(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					      DOUBLE dE00_x, DOUBLE dE00_y,
					      DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
					      DOUBLE dV20_x, DOUBLE dV20_y,
					      DOUBLE dV2_x, DOUBLE dV2_y,
					      DOUBLE dV21_x, DOUBLE dV21_y)
{
  APM2D::m_uiInscribedCircleCenterEvVMPFUNFilter++;

  *pdC_x = 0.0;
  *pdC_y = 0.0;
  *pdR = 0.0;
 
  m_eIP = ipStart;
  m_eResult = APM2D::icctNull;
  m_iV2onE0 = -1; 
  m_iDen = -1;

  m_piVertexZone[0] = -1;  
  m_piVertexZone[1] = -1;  

  BOOL bDone;

  Interval valE00_x = dE00_x;
  Interval valE00_y = dE00_y;
  Interval valE01_V1_x = dE01_V1_x;
  Interval valE01_V1_y = dE01_V1_y; 
  Interval valV20_x = dV20_x;
  Interval valV20_y = dV20_y;
  Interval valV2_x = dV2_x;
  Interval valV2_y = dV2_y;
  Interval valV21_x = dV21_x;
  Interval valV21_y = dV21_y;
  
  //////////////////////////////////////////////////
  // Check for V2 on E0
  
  Interval valV2onE0;
  SinTheta(valV2onE0, valV2_x, valV2_y, valE00_x, valE00_y, valE01_V1_x, valE01_V1_y);
  m_iV2onE0 = ( valV2onE0 <= 0.0 );
  if ( m_iV2onE0 == 1 )
    {
      // V2 is on E0 and it is not one of the endpoints
      return APM2D::icctNull;
    }
  if ( m_iV2onE0 == -1 )
    {
      bDone = InscribedCircleCenterEvVMPFUN(pdC_x, pdC_y, pdR, 
					    dE00_x, dE00_y,
					    dE01_V1_x, dE01_V1_y, 
					    dV20_x, dV20_y,
					    dV2_x, dV2_y,
					    dV21_x, dV21_y,
					    ipV2onE0);
      if ( bDone )
	{
	  return m_eResult;
	}
    }

  //////////////////////////////////////////////////
  // Check E0xV2 and Den
  
  Interval pppvalV[3][2][2]; // Endpoints
  // Translate V1 to the origin
  //memset(pppvalV, 0, sizeof(pppvalV));
  pppvalV[0][0][0] = valE00_x - valE01_V1_x;
  pppvalV[0][0][1] = valE00_y - valE01_V1_y;
  pppvalV[0][1][0] = 0.0;
  pppvalV[0][1][1] = 0.0;
  pppvalV[1][0][0] = 0.0;
  pppvalV[1][0][1] = 0.0;
  pppvalV[2][0][0] = valV2_x - valE01_V1_x;
  pppvalV[2][0][1] = valV2_y - valE01_V1_y;
  
  Interval ppvalE[3][2];     // Edge vectors
  //memset(ppvalE, 0, sizeof(ppvalE));
  ppvalE[0][0] = valE01_V1_x - valE00_x;
  ppvalE[0][1] = valE01_V1_y - valE00_y;
  
  Cross(m_valE0xV2, ppvalE[0][0], ppvalE[0][1], pppvalV[2][0][0], pppvalV[2][0][1]);
  
  m_valDen = 2.0*m_valE0xV2;
  m_iDen = ( m_valDen != 0.0 );
  if ( m_iDen == -1 )
    {
      bDone = InscribedCircleCenterEvVMPFUN(pdC_x, pdC_y, pdR, 
				       dE00_x, dE00_y,
				       dE01_V1_x, dE01_V1_y, 
				       dV20_x, dV20_y,
				       dV2_x, dV2_y,
				       dV21_x, dV21_y,
				       ipDen);
      if ( bDone )
	{
	  return m_eResult;
	}
      if ( m_iDen == -1 )
	{
	  bDone = InscribedCircleCenterEvVMPFUN(pdC_x, pdC_y, pdR, 
					   dE00_x, dE00_y,
					   dE01_V1_x, dE01_V1_y, 
					   dV20_x, dV20_y,
					   dV2_x, dV2_y,
					   dV21_x, dV21_y,
					   ipEnd);
	  ASSERT( bDone );
	  return m_eResult;
	}
    }
  ASSERT( m_iDen == 1 );

  //////////////////////////////////////////////////
  // Check the Vertex Zones

  Interval valV2dotV2;
  Dot(valV2dotV2, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][0][0], pppvalV[2][0][1]);
  
  Interval pppvalVN[3][2][2]; // Endpoints
  // Vertex neighbor vectors
  //memset(pppvalVN, 0, sizeof(pppvalV));
  pppvalVN[2][0][0] = valV2_x - valV20_x;
  pppvalVN[2][0][1] = valV2_y - valV20_y;
  pppvalVN[2][1][0] = valV2_x - valV21_x;
  pppvalVN[2][1][1] = valV2_y - valV21_y;
  
  Interval pvalVertexZone[2];
  
  Interval valV2dotV20;
  Dot(valV2dotV20, pppvalV[2][0][0], pppvalV[2][0][1], pppvalVN[2][0][0], pppvalVN[2][0][1]);
  Interval valE0xV20;
  Cross(valE0xV20, ppvalE[0][0], ppvalE[0][1], pppvalVN[2][0][0], pppvalVN[2][0][1]);
  pvalVertexZone[0] = ( valV2dotV2*valE0xV20 - 2.0*m_valE0xV2*valV2dotV20 );
  m_piVertexZone[0] = ( pvalVertexZone[0] < 0.0 );
  if ( m_piVertexZone[0] == 1 ) 
    {
      return APM2D::icctNull;
    }
  
  Interval valV2dotV21;
  Dot(valV2dotV21, pppvalV[2][0][0], pppvalV[2][0][1], pppvalVN[2][1][0], pppvalVN[2][1][1]);
  Interval valE0xV21;
  Cross(valE0xV21, ppvalE[0][0], ppvalE[0][1], pppvalVN[2][1][0], pppvalVN[2][1][1]);
  pvalVertexZone[1] = ( valV2dotV2*valE0xV21 - 2.0*m_valE0xV2*valV2dotV21 );
  m_piVertexZone[1] = ( pvalVertexZone[1] < 0.0 );
  if ( m_piVertexZone[1] == 1 ) 
    {
      return APM2D::icctNull;
    }

  if ( m_piVertexZone[0] == -1 ||
       m_piVertexZone[1] == -1 )
    {
      bDone = InscribedCircleCenterEvVMPFUN(pdC_x, pdC_y, pdR, 
				       dE00_x, dE00_y,
				       dE01_V1_x, dE01_V1_y, 
				       dV20_x, dV20_y,
				       dV2_x, dV2_y,
				       dV21_x, dV21_y,
				       ipVertexZones);
      if ( bDone )
	{
	  return m_eResult;
	}
    }

  //////////////////////////////////////////////////
  // Compute the center and radius
  
  Interval valE0dotE0;
  Dot(valE0dotE0, ppvalE[0][0], ppvalE[0][1], ppvalE[0][0], ppvalE[0][1]);
  
  Interval pvalC[2];
  pvalC[0] = -valV2dotV2*ppvalE[0][1]/m_valDen;
  pvalC[1] = valV2dotV2*ppvalE[0][0]/m_valDen;

  Interval valR;
  valR = (valV2dotV2*sqrt(valE0dotE0)) / m_valDen;
  
  // Translate the answer back out away from the origin
  pvalC[0] += valE01_V1_x;
  pvalC[1] += valE01_V1_y;
  
  *pdC_x = dble(pvalC[0]);
  *pdC_y = dble(pvalC[1]);
  *pdR = dble(valR);
  return APM2D::icctEvV;
}

APM2D::InscribedCircleCenterType 
ICC_EvV_MPFUNFilter::InscribedCircleCenterEvVForce(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
						   DOUBLE dE00_x, DOUBLE dE00_y,
						   DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
						   DOUBLE dV20_x, DOUBLE dV20_y,
						   DOUBLE dV2_x, DOUBLE dV2_y,
						   DOUBLE dV21_x, DOUBLE dV21_y)
{
  *pdC_x = 0.0;
  *pdC_y = 0.0;
  *pdR = 0.0;
 
  m_eIP = ipStart;
  m_eResult = APM2D::icctNull;
  m_iV2onE0 = -1; 
  m_iDen = -1;

  m_piVertexZone[0] = -1;  
  m_piVertexZone[1] = -1;  

  BOOL bDone;

  bDone = InscribedCircleCenterEvVMPFUN(pdC_x, pdC_y, pdR, 
					dE00_x, dE00_y,
					dE01_V1_x, dE01_V1_y, 
					dV20_x, dV20_y,
					dV2_x, dV2_y,
					dV21_x, dV21_y,
					ipEnd);
  ASSERT( bDone );
  return m_eResult;
}

BOOL
ICC_EvV_MPFUNFilter::InscribedCircleCenterEvVMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
						   DOUBLE dE00_x, DOUBLE dE00_y,
						   DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
						   DOUBLE dV20_x, DOUBLE dV20_y,
						   DOUBLE dV2_x, DOUBLE dV2_y,
						   DOUBLE dV21_x, DOUBLE dV21_y,
						   InstructionPointer eIPMax/* = ipEnd*/)
{
  APM2D::m_uiInscribedCircleCenterEvVMPFUN++;

  switch ( m_eIP )
    {
    case ipStart:
      {
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	*pdC_x = 0.0;
	*pdC_y = 0.0;
	*pdR = 0.0;

	m_rE00_x = dE00_x;
	m_rE00_y = dE00_y;
	m_rE01_V1_x = dE01_V1_x;
	m_rE01_V1_y = dE01_V1_y; 
	m_rV20_x = dV20_x;
	m_rV20_y = dV20_y;
	m_rV2_x = dV2_x;
	m_rV2_y = dV2_y;
	m_rV21_x = dV21_x;
	m_rV21_y = dV21_y;

	//////////////////////////////////////////////////
	// Check for V2 on E0
      
	if ( m_iV2onE0 == -1 )
	  {
	    mp_real rV2onE0;
	    SinTheta(rV2onE0, m_rV2_x, m_rV2_y, m_rE00_x, m_rE00_y, m_rE01_V1_x, m_rE01_V1_y);
	    if ( rV2onE0 <= 0.0 )
	      {
		// V2 is on E0 and it is not one of the endpoints
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterEvVMPFUNFinal++;
		return TRUE;
	      }
	  }

      }
    case ipV2onE0:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Check Den
  
	// Translate V1 to the origin
	//memset(m_ppprV, 0, sizeof(m_ppprV));
	m_ppprV[0][0][0] = m_rE00_x - m_rE01_V1_x;
	m_ppprV[0][0][1] = m_rE00_y - m_rE01_V1_y;
	m_ppprV[0][1][0] = 0.0;
	m_ppprV[0][1][1] = 0.0;
	m_ppprV[1][0][0] = 0.0;
	m_ppprV[1][0][1] = 0.0;
	m_ppprV[2][0][0] = m_rV2_x - m_rE01_V1_x;
	m_ppprV[2][0][1] = m_rV2_y - m_rE01_V1_y;
  
	//memset(m_pprE, 0, sizeof(m_pprE));
	m_pprE[0][0] = m_rE01_V1_x - m_rE00_x;
	m_pprE[0][1] = m_rE01_V1_y - m_rE00_y;
  
	Cross(m_rE0xV2, m_pprE[0][0], m_pprE[0][1], m_ppprV[2][0][0], m_ppprV[2][0][1]);

	m_rDen = 2.0*m_rE0xV2;
	ASSERT( m_rDen != 0.0 );

	if ( eIPMax != ipEnd &&
	     m_iDen == -1 )
	  { 
	    ASSERT( m_rDen != 0.0 );
	    m_valE0xV2.ConservativeBound(dble(m_rE0xV2));
	    m_valDen.ConservativeBound(dble(m_rDen));	    
	    m_iDen = ( m_valDen != 0.0 );      
	  }
      }
    case ipDen:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Check the Vertex Zones

	Dot(m_rV2dotV2, m_ppprV[2][0][0], m_ppprV[2][0][1], m_ppprV[2][0][0], m_ppprV[2][0][1]);

	if ( m_piVertexZone[0] == -1 || 
	     m_piVertexZone[1] == -1 )
	  {      
	    mp_real ppprVN[3][2][2]; // Endpoints
	    // Vertex neighbor vectors
	    //memset(ppprVN, 0, sizeof(ppprV));
	    ppprVN[2][0][0] = m_rV2_x - m_rV20_x;
	    ppprVN[2][0][1] = m_rV2_y - m_rV20_y;
	    ppprVN[2][1][0] = m_rV2_x - m_rV21_x;
	    ppprVN[2][1][1] = m_rV2_y - m_rV21_y;

	    mp_real prVertexZone[2];

	    if ( m_piVertexZone[0] == -1 )
	      {
		mp_real rV2dotV20;
		Dot(rV2dotV20, m_ppprV[2][0][0], m_ppprV[2][0][1], ppprVN[2][0][0], ppprVN[2][0][1]);
		mp_real rE0xV20;
		Cross(rE0xV20, m_pprE[0][0], m_pprE[0][1], ppprVN[2][0][0], ppprVN[2][0][1]);
		prVertexZone[0] = ( m_rV2dotV2*rE0xV20 - 2.0*m_rE0xV2*rV2dotV20 );
		if ( prVertexZone[0] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterEvVMPFUNFinal++;
		    return TRUE;
		  }
	      }

	    if ( m_piVertexZone[0] == -1 )
	      {
		mp_real rV2dotV21;
		Dot(rV2dotV21, m_ppprV[2][0][0], m_ppprV[2][0][1], ppprVN[2][1][0], ppprVN[2][1][1]);
		mp_real rE0xV21;
		Cross(rE0xV21, m_pprE[0][0], m_pprE[0][1], ppprVN[2][1][0], ppprVN[2][1][1]);
		prVertexZone[1] = ( m_rV2dotV2*rE0xV21 - 2.0*m_rE0xV2*rV2dotV21 );
		if ( prVertexZone[0] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterEvVMPFUNFinal++;
		    return TRUE;
		  }
	      }
	  }

      }
    case ipVertexZones:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Compute the center and radius

	mp_real rE0dotE0;
	Dot(rE0dotE0, m_pprE[0][0], m_pprE[0][1], m_pprE[0][0], m_pprE[0][1]);

	mp_real prC[2];
	prC[0] = -m_rV2dotV2*m_pprE[0][1]/m_rDen;
	prC[1] = m_rV2dotV2*m_pprE[0][0]/m_rDen;

	mp_real rR;
	rR = (m_rV2dotV2*sqrt(rE0dotE0)) / m_rDen;
  
	// Translate the answer back out away from the origin
	prC[0] += m_rE01_V1_x;
	prC[1] += m_rE01_V1_y;

	*pdC_x = dble(prC[0]);
	*pdC_y = dble(prC[1]);
	*pdR = dble(rR);
	m_eResult = APM2D::icctEvV;
      }
      break;
      /*
    default:
      ASSERT( FALSE );
      break;
      */
    }

  APM2D::m_uiInscribedCircleCenterEvVMPFUNFinal++;
  return TRUE;
}

//static 
APM2D::InscribedCircleCenterType 
InscribedCircleCenterEvVMPFUNFilter(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				    DOUBLE dE00_x, DOUBLE dE00_y,
				    DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
				    DOUBLE dV20_x, DOUBLE dV20_y,
				    DOUBLE dV2_x, DOUBLE dV2_y,
				    DOUBLE dV21_x, DOUBLE dV21_y)
{
  ICC_EvV_MPFUNFilter iccEvV;

  return iccEvV.InscribedCircleCenterEvV(pdC_x, pdC_y, pdR, 
					 dE00_x, dE00_y,
					 dE01_V1_x, dE01_V1_y, 
					 dV20_x, dV20_y,
					 dV2_x, dV2_y,
					 dV21_x, dV21_y);
}

//static 
APM2D::EdgeZoneType EdgeZoneEvVMPFUNFilter(DOUBLE dE00_x, DOUBLE dE00_y,
					   DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
					   DOUBLE dV2_x, DOUBLE dV2_y,
					   DOUBLE dE30_x, DOUBLE dE30_y,
					   DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::m_uiEdgeZoneEvVMPFUNFilter++;

  INT_32 piEdgeZone[2];
  piEdgeZone[0] = -1;
  piEdgeZone[1] = -1;

  switch ( 0 )
  {
  default:
    Interval valE00_x = dE00_x;
    Interval valE00_y = dE00_y;
    Interval valE01_V1_x = dE01_V1_x;
    Interval valE01_V1_y = dE01_V1_y; 
    Interval valV2_x = dV2_x;
    Interval valV2_y = dV2_y;
    Interval valE30_x = dE30_x;
    Interval valE30_y = dE30_y;
    Interval valE31_x = dE31_x;
    Interval valE31_y = dE31_y;

    Interval pppvalV[4][2][2]; // Endpoints
    //memset(pppvalV, 0, sizeof(pppvalV));
    pppvalV[0][0][0] = valE00_x - valE01_V1_x;
    pppvalV[0][0][1] = valE00_y - valE01_V1_y;
    pppvalV[0][1][0] = 0.0;
    pppvalV[0][1][1] = 0.0;
    pppvalV[1][0][0] = 0.0;
    pppvalV[1][0][1] = 0.0;
    pppvalV[2][0][0] = valV2_x - valE01_V1_x;
    pppvalV[2][0][1] = valV2_y - valE01_V1_y;
    pppvalV[3][0][0] = valE30_x - valE01_V1_x;
    pppvalV[3][0][1] = valE30_y - valE01_V1_y;
    pppvalV[3][1][0] = valE31_x - valE01_V1_x;
    pppvalV[3][1][1] = valE31_y - valE01_V1_y;
  
    Interval ppvalE[4][2];     // Edge vectors
    //memset(ppvalE, 0, sizeof(ppvalE));
    ppvalE[0][0] = valE01_V1_x - valE00_x;
    ppvalE[0][1] = valE01_V1_y - valE00_y;
    ppvalE[3][0] = valE31_x - valE30_x;
    ppvalE[3][1] = valE31_y - valE30_y;
  
    Interval valV2dotV2;
    Dot(valV2dotV2, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][0][0], pppvalV[2][0][1]);
    Interval valE3dotE3;
    Dot(valE3dotE3, ppvalE[3][0], ppvalE[3][1], ppvalE[3][0], ppvalE[3][1]);

    Interval valE0xV2;
    Cross(valE0xV2, ppvalE[0][0], ppvalE[0][1], pppvalV[2][0][0], pppvalV[2][0][1]);
    Interval valE0xE3;
    Cross(valE0xE3, ppvalE[0][0], ppvalE[0][1], ppvalE[3][0], ppvalE[3][1]);

    Interval valV2dotV2timesE0xE3 = valV2dotV2*valE0xE3;

    Interval pvalEdgeZone[2];

    Interval valE3dotE30;
    Dot(valE3dotE30, ppvalE[3][0], ppvalE[3][1], pppvalV[3][0][0], pppvalV[3][0][1]);
    pvalEdgeZone[0] = ( 2.0*valE0xV2*valE3dotE30 - valV2dotV2timesE0xE3 );
    piEdgeZone[0] = ( pvalEdgeZone[0] >= 0.0 );
    if ( piEdgeZone[0] == 1 )
      {
	return APM2D::eztStart;
      }

    Interval valE3dotE31;
    Dot(valE3dotE31, ppvalE[3][0], ppvalE[3][1], pppvalV[3][1][0], pppvalV[3][1][1]);
    pvalEdgeZone[1] = -( 2.0*valE0xV2*valE3dotE31 - valV2dotV2timesE0xE3 );
    piEdgeZone[1] = ( pvalEdgeZone[1] >= 0.0 );
    if ( piEdgeZone[1] == 1 )
      {
	return APM2D::eztEnd;
      }
  }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( piEdgeZone[0] == -1 || piEdgeZone[1] == -1 )
    {
      APM2D::m_uiEdgeZoneEvVMPFUN++;

      mp_real rE00_x = dE00_x;
      mp_real rE00_y = dE00_y;
      mp_real rE01_V1_x = dE01_V1_x;
      mp_real rE01_V1_y = dE01_V1_y; 
      mp_real rV2_x = dV2_x;
      mp_real rV2_y = dV2_y;
      mp_real rE30_x = dE30_x;
      mp_real rE30_y = dE30_y;
      mp_real rE31_x = dE31_x;
      mp_real rE31_y = dE31_y;

      mp_real ppprV[4][2][2]; // Endpoints
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = rE00_x - rE01_V1_x;
      ppprV[0][0][1] = rE00_y - rE01_V1_y;
      ppprV[0][1][0] = 0.0;
      ppprV[0][1][1] = 0.0;
      ppprV[1][0][0] = 0.0;
      ppprV[1][0][1] = 0.0;
      ppprV[2][0][0] = rV2_x - rE01_V1_x;
      ppprV[2][0][1] = rV2_y - rE01_V1_y;
      ppprV[3][0][0] = rE30_x - rE01_V1_x;
      ppprV[3][0][1] = rE30_y - rE01_V1_y;
      ppprV[3][1][0] = rE31_x - rE01_V1_x;
      ppprV[3][1][1] = rE31_y - rE01_V1_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[0][0] = rE01_V1_x - rE00_x;
      pprE[0][1] = rE01_V1_y - rE00_y;
      pprE[3][0] = rE31_x - rE30_x;
      pprE[3][1] = rE31_y - rE30_y;
  
      mp_real rV2dotV2;
      Dot(rV2dotV2, ppprV[2][0][0], ppprV[2][0][1], ppprV[2][0][0], ppprV[2][0][1]);
      mp_real rE3dotE3;
      Dot(rE3dotE3, pprE[3][0], pprE[3][1], pprE[3][0], pprE[3][1]);

      mp_real rE0xV2;
      Cross(rE0xV2, pprE[0][0], pprE[0][1], ppprV[2][0][0], ppprV[2][0][1]);
      mp_real rE0xE3;
      Cross(rE0xE3, pprE[0][0], pprE[0][1], pprE[3][0], pprE[3][1]);

      mp_real rV2dotV2timesE0xE3 = rV2dotV2*rE0xE3;

      mp_real prEdgeZone[2];

      if ( piEdgeZone[0] == -1 )
	{
	  mp_real rE3dotE30;
	  Dot(rE3dotE30, pprE[3][0], pprE[3][1], ppprV[3][0][0], ppprV[3][0][1]);
	  prEdgeZone[0] = ( 2.0*rE0xV2*rE3dotE30 - rV2dotV2timesE0xE3 );
	  if ( prEdgeZone[0] >= 0.0 )
	    {
	      return APM2D::eztStart;
	    }
	}

      if ( piEdgeZone[1] == -1 )
	{
	  mp_real rE3dotE31;
	  Dot(rE3dotE31, pprE[3][0], pprE[3][1], ppprV[3][1][0], ppprV[3][1][1]);
	  prEdgeZone[1] = -( 2.0*rE0xV2*rE3dotE31 - rV2dotV2timesE0xE3 );
	  if ( prEdgeZone[1] >= 0.0 )
	    {
	      return APM2D::eztEnd;
	    }
	}
    }

  return APM2D::eztMiddle;
}

//static 
APM2D::CircleTestType InCircleEvV_VMPFUNFilter(DOUBLE dE00_x, DOUBLE dE00_y,
					       DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
					       DOUBLE dV2_x, DOUBLE dV2_y,
					       DOUBLE dV3_x, DOUBLE dV3_y)
{
  APM2D::m_uiInCircleEvV_VMPFUNFilter++;

  INT_32 iResult;

  iResult = -1;

  switch ( 0 )
    {
    default:
      Interval valE00_x = dE00_x;
      Interval valE00_y = dE00_y;
      Interval valE01_V1_x = dE01_V1_x;
      Interval valE01_V1_y = dE01_V1_y; 
      Interval valV2_x = dV2_x;
      Interval valV2_y = dV2_y;
      Interval valV3_x = dV3_x;
      Interval valV3_y = dV3_y;

      Interval pppvalV[4][2][2]; // Endpoints
      //memset(pppvalV, 0, sizeof(pppvalV));
      pppvalV[0][0][0] = valE00_x - valE01_V1_x;
      pppvalV[0][0][1] = valE00_y - valE01_V1_y;
      pppvalV[0][1][0] = 0.0;
      pppvalV[0][1][1] = 0.0;
      pppvalV[1][0][0] = 0.0;
      pppvalV[1][0][1] = 0.0;
      pppvalV[2][0][0] = valV2_x - valE01_V1_x;
      pppvalV[2][0][1] = valV2_y - valE01_V1_y;
      pppvalV[3][0][0] = valV3_x - valE01_V1_x;
      pppvalV[3][0][1] = valV3_y - valE01_V1_y;
  
      Interval ppvalE[4][2];     // Edge vectors
      //memset(ppvalE, 0, sizeof(ppvalE));
      ppvalE[0][0] = valE01_V1_x - valE00_x;
      ppvalE[0][1] = valE01_V1_y - valE00_y;
  
      Interval valV2dotV2;
      Dot(valV2dotV2, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][0][0], pppvalV[2][0][1]);
      Interval valV3dotV3;
      Dot(valV3dotV3, pppvalV[3][0][0], pppvalV[3][0][1], pppvalV[3][0][0], pppvalV[3][0][1]);

      Interval valE0xV2;
      Cross(valE0xV2, ppvalE[0][0], ppvalE[0][1], pppvalV[2][0][0], pppvalV[2][0][1]);
      Interval valE0xV3;
      Cross(valE0xV3, ppvalE[0][0], ppvalE[0][1], pppvalV[3][0][0], pppvalV[3][0][1]);

      Interval valResult;
      valResult = ( valV3dotV3*valE0xV2 - valV2dotV2*valE0xV3 );

      iResult = ( valResult < 0.0 );
      if ( iResult == -1 )
	{
	  break;
	}
      if ( iResult == 1 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( valResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
      return APM2D::cttOnCircle;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( iResult == -1 )
    {
      APM2D::m_uiInCircleEvV_VMPFUN++;

      mp_real rE00_x = dE00_x;
      mp_real rE00_y = dE00_y;
      mp_real rE01_V1_x = dE01_V1_x;
      mp_real rE01_V1_y = dE01_V1_y; 
      mp_real rV2_x = dV2_x;
      mp_real rV2_y = dV2_y;
      mp_real rV3_x = dV3_x;
      mp_real rV3_y = dV3_y;

      mp_real ppprV[4][2][2]; // Endpoints
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = rE00_x - rE01_V1_x;
      ppprV[0][0][1] = rE00_y - rE01_V1_y;
      ppprV[0][1][0] = 0.0;
      ppprV[0][1][1] = 0.0;
      ppprV[1][0][0] = 0.0;
      ppprV[1][0][1] = 0.0;
      ppprV[2][0][0] = rV2_x - rE01_V1_x;
      ppprV[2][0][1] = rV2_y - rE01_V1_y;
      ppprV[3][0][0] = rV3_x - rE01_V1_x;
      ppprV[3][0][1] = rV3_y - rE01_V1_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[0][0] = rE01_V1_x - rE00_x;
      pprE[0][1] = rE01_V1_y - rE00_y;
  
      mp_real rV2dotV2;
      Dot(rV2dotV2, ppprV[2][0][0], ppprV[2][0][1], ppprV[2][0][0], ppprV[2][0][1]);
      mp_real rV3dotV3;
      Dot(rV3dotV3, ppprV[3][0][0], ppprV[3][0][1], ppprV[3][0][0], ppprV[3][0][1]);

      mp_real rE0xV2;
      Cross(rE0xV2, pprE[0][0], pprE[0][1], ppprV[2][0][0], ppprV[2][0][1]);
      mp_real rE0xV3;
      Cross(rE0xV3, pprE[0][0], pprE[0][1], ppprV[3][0][0], ppprV[3][0][1]);

      mp_real rResult;
      rResult = ( rV3dotV3*rE0xV2 - rV2dotV2*rE0xV3 );
  
      if ( rResult < 0.0 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( rResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
    }

  return APM2D::cttOnCircle;
}

//static 
APM2D::CircleTestType InCircleEvV_EMPFUNFilter(DOUBLE dE00_x, DOUBLE dE00_y,
					       DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
					       DOUBLE dV2_x, DOUBLE dV2_y,
					       DOUBLE dE30_x, DOUBLE dE30_y,
					       DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::m_uiInCircleEvV_EMPFUNFilter++;

  INT_32 iResult;

  iResult = -1;

  switch ( 0 )
    {
    default:
      Interval valE00_x = dE00_x;
      Interval valE00_y = dE00_y;
      Interval valE01_V1_x = dE01_V1_x;
      Interval valE01_V1_y = dE01_V1_y; 
      Interval valV2_x = dV2_x;
      Interval valV2_y = dV2_y;
      Interval valE30_x = dE30_x;
      Interval valE30_y = dE30_y;
      Interval valE31_x = dE31_x;
      Interval valE31_y = dE31_y;

      Interval pppvalV[4][2][2]; // Endpoints
      //memset(pppvalV, 0, sizeof(pppvalV));
      pppvalV[0][0][0] = valE00_x - valE01_V1_x;
      pppvalV[0][0][1] = valE00_y - valE01_V1_y;
      pppvalV[0][1][0] = 0.0;
      pppvalV[0][1][1] = 0.0;
      pppvalV[1][0][0] = 0.0;
      pppvalV[1][0][1] = 0.0;
      pppvalV[2][0][0] = valV2_x - valE01_V1_x;
      pppvalV[2][0][1] = valV2_y - valE01_V1_y;
      pppvalV[3][0][0] = valE30_x - valE01_V1_x;
      pppvalV[3][0][1] = valE30_y - valE01_V1_y;
      pppvalV[3][1][0] = valE31_x - valE01_V1_x;
      pppvalV[3][1][1] = valE31_y - valE01_V1_y;
  
      Interval ppvalE[4][2];     // Edge vectors
      //memset(ppvalE, 0, sizeof(ppvalE));
      ppvalE[0][0] = valE01_V1_x - valE00_x;
      ppvalE[0][1] = valE01_V1_y - valE00_y;
      ppvalE[3][0] = valE31_x - valE30_x;
      ppvalE[3][1] = valE31_y - valE30_y;
  
      Interval valV2dotV2;
      Dot(valV2dotV2, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][0][0], pppvalV[2][0][1]);
      Interval valE0dotE0;
      Dot(valE0dotE0, ppvalE[0][0], ppvalE[0][1], ppvalE[0][0], ppvalE[0][1]);
      Interval valE3dotE3;
      Dot(valE3dotE3, ppvalE[3][0], ppvalE[3][1], ppvalE[3][0], ppvalE[3][1]);

      Interval valE0dotE3;
      Dot(valE0dotE3, ppvalE[0][0], ppvalE[0][1], ppvalE[3][0], ppvalE[3][1]);

      Interval valE0xV2;
      Cross(valE0xV2, ppvalE[0][0], ppvalE[0][1], pppvalV[2][0][0], pppvalV[2][0][1]);
      Interval valE30xE31;
      Cross(valE30xE31, pppvalV[3][0][0], pppvalV[3][0][1], pppvalV[3][1][0], pppvalV[3][1][1]);

      Interval valResult;
      valResult = ( 4.0*valE0xV2*valE30xE31*( valE0xV2*valE30xE31 + valV2dotV2*valE0dotE3 )
		  - valV2dotV2*valV2dotV2*( valE0dotE0*valE3dotE3 - valE0dotE3*valE0dotE3 ) );

      iResult = ( valResult < 0.0 );
      if ( iResult == -1 )
	{
	  break;
	}
      if ( iResult == 1 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( valResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
      return APM2D::cttOnCircle;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( iResult == -1 )
    {
      APM2D::m_uiInCircleEvV_EMPFUN++;

      mp_real rE00_x = dE00_x;
      mp_real rE00_y = dE00_y;
      mp_real rE01_V1_x = dE01_V1_x;
      mp_real rE01_V1_y = dE01_V1_y; 
      mp_real rV2_x = dV2_x;
      mp_real rV2_y = dV2_y;
      mp_real rE30_x = dE30_x;
      mp_real rE30_y = dE30_y;
      mp_real rE31_x = dE31_x;
      mp_real rE31_y = dE31_y;

      mp_real ppprV[4][2][2]; // Endpoints
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = rE00_x - rE01_V1_x;
      ppprV[0][0][1] = rE00_y - rE01_V1_y;
      ppprV[0][1][0] = 0.0;
      ppprV[0][1][1] = 0.0;
      ppprV[1][0][0] = 0.0;
      ppprV[1][0][1] = 0.0;
      ppprV[2][0][0] = rV2_x - rE01_V1_x;
      ppprV[2][0][1] = rV2_y - rE01_V1_y;
      ppprV[3][0][0] = rE30_x - rE01_V1_x;
      ppprV[3][0][1] = rE30_y - rE01_V1_y;
      ppprV[3][1][0] = rE31_x - rE01_V1_x;
      ppprV[3][1][1] = rE31_y - rE01_V1_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[0][0] = rE01_V1_x - rE00_x;
      pprE[0][1] = rE01_V1_y - rE00_y;
      pprE[3][0] = rE31_x - rE30_x;
      pprE[3][1] = rE31_y - rE30_y;
  
      mp_real rV2dotV2;
      Dot(rV2dotV2, ppprV[2][0][0], ppprV[2][0][1], ppprV[2][0][0], ppprV[2][0][1]);
      mp_real rE0dotE0;
      Dot(rE0dotE0, pprE[0][0], pprE[0][1], pprE[0][0], pprE[0][1]);
      mp_real rE3dotE3;
      Dot(rE3dotE3, pprE[3][0], pprE[3][1], pprE[3][0], pprE[3][1]);

      mp_real rE0dotE3;
      Dot(rE0dotE3, pprE[0][0], pprE[0][1], pprE[3][0], pprE[3][1]);

      mp_real rE0xV2;
      Cross(rE0xV2, pprE[0][0], pprE[0][1], ppprV[2][0][0], ppprV[2][0][1]);
      mp_real rE30xE31;
      Cross(rE30xE31, ppprV[3][0][0], ppprV[3][0][1], ppprV[3][1][0], ppprV[3][1][1]);

      mp_real rResult;
      rResult = ( 4.0*rE0xV2*rE30xE31*( rE0xV2*rE30xE31 + rV2dotV2*rE0dotE3 )
		  - rV2dotV2*rV2dotV2*( rE0dotE0*rE3dotE3 - rE0dotE3*rE0dotE3 ) );

      if ( rResult < 0.0 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( rResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
    }

  return APM2D::cttOnCircle;
}

//////////////////////////////////////////////////
// VEE

class ICC_VEE_MPFUNFilter
{
protected:
  enum InstructionPointer {ipStart, ipE1xE2, ipV0onE1orE2, 
			   ipDenR, ipDen, ipDisc, 
			   ipEdgeZones, ipVertexZones, ipEnd};

public:
  ICC_VEE_MPFUNFilter();
  ~ICC_VEE_MPFUNFilter();

public:
  APM2D::InscribedCircleCenterType 
  InscribedCircleCenterVEE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			   DOUBLE dV00_x, DOUBLE dV00_y,
			   DOUBLE dV0_x, DOUBLE dV0_y,
			   DOUBLE dV01_x, DOUBLE dV01_y,
			   DOUBLE dE10_x, DOUBLE dE10_y, 
			   DOUBLE dE11_x, DOUBLE dE11_y,
			   DOUBLE dE20_x, DOUBLE dE20_y, 
			   DOUBLE dE21_x, DOUBLE dE21_y);
  APM2D::InscribedCircleCenterType 
  InscribedCircleCenterVEEForce(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dV00_x, DOUBLE dV00_y,
				DOUBLE dV0_x, DOUBLE dV0_y,
				DOUBLE dV01_x, DOUBLE dV01_y,
				DOUBLE dE10_x, DOUBLE dE10_y, 
				DOUBLE dE11_x, DOUBLE dE11_y,
				DOUBLE dE20_x, DOUBLE dE20_y, 
				DOUBLE dE21_x, DOUBLE dE21_y);
  
protected:
  BOOL
  InscribedCircleCenterVEEMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dV00_x, DOUBLE dV00_y,
				DOUBLE dV0_x, DOUBLE dV0_y,
				DOUBLE dV01_x, DOUBLE dV01_y,
				DOUBLE dE10_x, DOUBLE dE10_y, 
				DOUBLE dE11_x, DOUBLE dE11_y,
				DOUBLE dE20_x, DOUBLE dE20_y, 
				DOUBLE dE21_x, DOUBLE dE21_y,
				InstructionPointer eIPMax = ipEnd);

protected:
  InstructionPointer m_eIP;
  APM2D::InscribedCircleCenterType m_eResult;
  INT_32 m_iE1xE2, m_iE1xE2Sign;
  INT_32 m_iE1dotE2;
  INT_32 m_iV0onE1, m_iV0onE2;
  INT_32 m_iDenR;
  INT_32 m_iDen;
  INT_32 m_iDisc;

  INT_32 m_ppiEdgeZone[2][2];
  INT_32 m_piVertexZone[2];

  Interval m_valE1xE2;
  Interval m_valE1dotE2;
  Interval m_valDenR;
  Interval m_valDen;
  Interval m_valDisc;

  mp_real m_rV00_x;
  mp_real m_rV00_y;
  mp_real m_rV0_x;
  mp_real m_rV0_y;
  mp_real m_rV01_x;
  mp_real m_rV01_y;
  mp_real m_rE10_x;
  mp_real m_rE10_y;
  mp_real m_rE11_x;
  mp_real m_rE11_y;
  mp_real m_rE20_x;
  mp_real m_rE20_y;
  mp_real m_rE21_x;
  mp_real m_rE21_y;
  mp_real m_pprE[3][2];
  mp_real m_rE1xE2;
  mp_real m_rE1dotE2;
  mp_real m_rPlusOrMinus;
  mp_real m_rE1dotE1;
  mp_real m_rE2dotE2;
  mp_real m_rE1E2minusE1dotE2;
  mp_real m_rDenR;
  mp_real m_rDen;
  mp_real m_ppprV[3][2][2];
  mp_real m_rE10xE11;
  mp_real m_rE20xE21;
  mp_real m_rE1E2plusE1dotE2;
  mp_real m_rDisc;
  mp_real m_rSqrtDisc;
};

ICC_VEE_MPFUNFilter::ICC_VEE_MPFUNFilter()
{
}

ICC_VEE_MPFUNFilter::~ICC_VEE_MPFUNFilter()
{
}

APM2D::InscribedCircleCenterType 
ICC_VEE_MPFUNFilter::InscribedCircleCenterVEE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					      DOUBLE dV00_x, DOUBLE dV00_y,
					      DOUBLE dV0_x, DOUBLE dV0_y,
					      DOUBLE dV01_x, DOUBLE dV01_y,
					      DOUBLE dE10_x, DOUBLE dE10_y, 
					      DOUBLE dE11_x, DOUBLE dE11_y,
					      DOUBLE dE20_x, DOUBLE dE20_y, 
					      DOUBLE dE21_x, DOUBLE dE21_y)
{
  APM2D::m_uiInscribedCircleCenterVEEMPFUNFilter++;

  *pdC_x = 0.0;
  *pdC_y = 0.0;
  *pdR = 0.0;

  m_eIP = ipStart;
  m_eResult = APM2D::icctNull;
  m_iE1xE2 = -1;
  m_iE1xE2Sign = -1;
  m_iE1dotE2 = -1;
  m_iV0onE1 = -1;
  m_iV0onE2 = -1;
  m_iDenR = -1;
  m_iDen = -1;
  m_iDisc = -1;
  m_ppiEdgeZone[0][0] = -1;
  m_ppiEdgeZone[0][1] = -1;
  m_ppiEdgeZone[1][0] = -1;
  m_ppiEdgeZone[1][1] = -1;
  m_piVertexZone[0] = -1;
  m_piVertexZone[1] = -1;

  BOOL bDone;

  //////////////////////////////////////////////////
  // Check E1xE2

  Interval valV00_x = dV00_x;
  Interval valV00_y = dV00_y;
  Interval valV0_x = dV0_x;
  Interval valV0_y = dV0_y;
  Interval valV01_x = dV01_x;
  Interval valV01_y = dV01_y;
  Interval valE10_x = dE10_x;
  Interval valE10_y = dE10_y; 
  Interval valE11_x = dE11_x;
  Interval valE11_y = dE11_y;
  Interval valE20_x = dE20_x;
  Interval valE20_y = dE20_y; 
  Interval valE21_x = dE21_x;
  Interval valE21_y = dE21_y;

  Interval ppvalE[3][2];     // Edge vectors
  //memset(ppvalE, 0, sizeof(ppvalE));
  ppvalE[1][0] = valE11_x - valE10_x;
  ppvalE[1][1] = valE11_y - valE10_y;
  ppvalE[2][0] = valE21_x - valE20_x;
  ppvalE[2][1] = valE21_y - valE20_y;
  
  Cross(m_valE1xE2, ppvalE[1][0], ppvalE[1][1], ppvalE[2][0], ppvalE[2][1]);
  Dot(m_valE1dotE2, ppvalE[1][0], ppvalE[1][1], ppvalE[2][0], ppvalE[2][1]);

  m_iE1xE2 = ( m_valE1xE2 == 0.0 );
  m_iE1xE2Sign = ( m_valE1xE2 < 0.0 );
  m_iE1dotE2 = ( m_valE1dotE2 >= 0.0 );

  if ( m_iE1xE2 == -1 ||
       m_iE1dotE2 == -1 )
    {
      bDone = InscribedCircleCenterVEEMPFUN(pdC_x, pdC_y, pdR, 
					    dV00_x, dV00_y,
					    dV0_x, dV0_y,
					    dV01_x, dV01_y,
					    dE10_x, dE10_y, 
					    dE11_x, dE11_y,
					    dE20_x, dE20_y, 
					    dE21_x, dE21_y,
					    ipE1xE2);
      if ( bDone )
	{
	  return m_eResult;
	}
      if ( m_iE1xE2 == -1 ||
	   m_iE1dotE2 == -1 )
	{
	  bDone = InscribedCircleCenterVEEMPFUN(pdC_x, pdC_y, pdR, 
						dV00_x, dV00_y,
						dV0_x, dV0_y,
						dV01_x, dV01_y,
						dE10_x, dE10_y, 
						dE11_x, dE11_y,
						dE20_x, dE20_y, 
						dE21_x, dE21_y,
						ipEnd);
	  ASSERT( bDone );
	  return m_eResult;
	}
    }

  if ( m_iE1xE2 == 1 )
    {
      if ( m_iE1dotE2 == 1 )
	{
	  // Colinear
	  return APM2D::icctNull;
	}
      else
	{
	  return InscribedCircleCenterVEEAntiparallelMPFUNFilter(pdC_x, pdC_y, pdR, 
								 dV00_x, dV00_y, 
								 dV0_x, dV0_y, 
								 dV01_x, dV01_y, 
								 dE10_x, dE10_y, 
								 dE11_x, dE11_y, 
								 dE20_x, dE20_y, 
								 dE21_x, dE21_y);
	}
    }
  Interval valPlusOrMinus;
  valPlusOrMinus = 1.0;
  if ( m_iE1xE2Sign == 0 )
    {
      valPlusOrMinus = -1.0;
    }

  //////////////////////////////////////////////////
  // Check V0 on E1 or E2

  Interval valV0onE1;
  SinTheta(valV0onE1, valV0_x, valV0_y, valE10_x, valE10_y, valE11_x, valE11_y);
  m_iV0onE1 = ( valV0onE1 <= 0.0 );
  if ( m_iV0onE1 == 1 )
    {
      // V0 is on E1 and it is not one of the endpoints
      return APM2D::icctNull;
    }
  
  Interval valV0onE2;
  SinTheta(valV0onE2, valV0_x, valV0_y, valE20_x, valE20_y, valE21_x, valE21_y);
  m_iV0onE2 = ( valV0onE2 <= 0.0 );
  if ( m_iV0onE2 == 1 )
    {
      // V0 is on E2 and it is not one of the endpoints
      return APM2D::icctNull;
    }

  if ( m_iV0onE1 == -1 ||
       m_iV0onE2 == -1 )
    {
      bDone = InscribedCircleCenterVEEMPFUN(pdC_x, pdC_y, pdR, 
					    dV00_x, dV00_y,
					    dV0_x, dV0_y,
					    dV01_x, dV01_y,
					    dE10_x, dE10_y, 
					    dE11_x, dE11_y,
					    dE20_x, dE20_y, 
					    dE21_x, dE21_y,
					    ipV0onE1orE2);
      if ( bDone )
	{
	  return m_eResult;
	}
    }
  
  //////////////////////////////////////////////////
  // Check DenR

  Interval valE1dotE1;
  Dot(valE1dotE1, ppvalE[1][0], ppvalE[1][1], ppvalE[1][0], ppvalE[1][1]);
  Interval valE2dotE2;
  Dot(valE2dotE2, ppvalE[2][0], ppvalE[2][1], ppvalE[2][0], ppvalE[2][1]);

  Interval valE1E2minusE1dotE2;
  valE1E2minusE1dotE2 = sqrt(valE1dotE1*valE2dotE2) - m_valE1dotE2;

  m_valDenR = valE1E2minusE1dotE2;
  m_iDenR = ( m_valDenR != 0.0 );
  if ( m_iDenR == -1 )
    {
      bDone = InscribedCircleCenterVEEMPFUN(pdC_x, pdC_y, pdR, 
					    dV00_x, dV00_y,
					    dV0_x, dV0_y,
					    dV01_x, dV01_y,
					    dE10_x, dE10_y, 
					    dE11_x, dE11_y,
					    dE20_x, dE20_y, 
					    dE21_x, dE21_y,
					    ipDenR);
      if ( bDone )
	{
	  return m_eResult;
	}
      if ( m_iDenR == -1 )
	{
	  bDone = InscribedCircleCenterVEEMPFUN(pdC_x, pdC_y, pdR, 
						dV00_x, dV00_y,
						dV0_x, dV0_y,
						dV01_x, dV01_y,
						dE10_x, dE10_y, 
						dE11_x, dE11_y,
						dE20_x, dE20_y, 
						dE21_x, dE21_y,
						ipEnd);
	  ASSERT( bDone );
	  return m_eResult;
	}
    }
  ASSERT( m_iDenR == 1 );

  //////////////////////////////////////////////////
  // Check Den

  m_valDen = m_valE1xE2*valE1E2minusE1dotE2;
  m_iDen = ( m_valDen != 0.0 );
  if ( m_iDen == -1 )
    {
      bDone = InscribedCircleCenterVEEMPFUN(pdC_x, pdC_y, pdR, 
					    dV00_x, dV00_y,
					    dV0_x, dV0_y,
					    dV01_x, dV01_y,
					    dE10_x, dE10_y, 
					    dE11_x, dE11_y,
					    dE20_x, dE20_y, 
					    dE21_x, dE21_y,
					    ipDen);
      if ( bDone )
	{
	  return m_eResult;
	}
      if ( m_iDen == -1 )
	{
	  bDone = InscribedCircleCenterVEEMPFUN(pdC_x, pdC_y, pdR, 
						dV00_x, dV00_y,
						dV0_x, dV0_y,
						dV01_x, dV01_y,
						dE10_x, dE10_y, 
						dE11_x, dE11_y,
						dE20_x, dE20_y, 
						dE21_x, dE21_y,
						ipEnd);
	  ASSERT( bDone );
	  return m_eResult;
	}
    }
  ASSERT( m_iDen == 1 );

  //////////////////////////////////////////////////
  // Check Disc

  Interval pppvalV[3][2][2]; // Endpoints
  // Translate V1 to the origin
  //memset(pppvalV, 0, sizeof(pppvalV));
  pppvalV[0][0][0] = 0.0;
  pppvalV[0][0][1] = 0.0;
  pppvalV[1][0][0] = valE10_x - valV0_x;
  pppvalV[1][0][1] = valE10_y - valV0_y;
  pppvalV[1][1][0] = valE11_x - valV0_x;
  pppvalV[1][1][1] = valE11_y - valV0_y;
  pppvalV[2][0][0] = valE20_x - valV0_x;
  pppvalV[2][0][1] = valE20_y - valV0_y;
  pppvalV[2][1][0] = valE21_x - valV0_x;
  pppvalV[2][1][1] = valE21_y - valV0_y;
  
  Interval valE10xE11;
  Cross(valE10xE11, pppvalV[1][0][0], pppvalV[1][0][1], pppvalV[1][1][0], pppvalV[1][1][1]);
  Interval valE20xE21;
  Cross(valE20xE21, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][1][0], pppvalV[2][1][1]);

  Interval valE1E2plusE1dotE2;
  valE1E2plusE1dotE2 = sqrt(valE1dotE1*valE2dotE2) + m_valE1dotE2;

  m_valDisc = 2.0*valE10xE11*valE20xE21*valE1E2plusE1dotE2;
  m_iDisc = ( m_valDisc < 0.0 );
  if ( m_iDisc == 1 ) 
    {
      // Imaginary Solutions
      return APM2D::icctNull;
    }
  if ( m_iDisc == -1 )
    {
      bDone = InscribedCircleCenterVEEMPFUN(pdC_x, pdC_y, pdR, 
					    dV00_x, dV00_y,
					    dV0_x, dV0_y,
					    dV01_x, dV01_y,
					    dE10_x, dE10_y, 
					    dE11_x, dE11_y,
					    dE20_x, dE20_y, 
					    dE21_x, dE21_y,
					    ipDisc);
      if ( bDone )
	{
	  return m_eResult;
	}
      if ( m_iDisc == -1 )
	{
	  bDone = InscribedCircleCenterVEEMPFUN(pdC_x, pdC_y, pdR, 
						dV00_x, dV00_y,
						dV0_x, dV0_y,
						dV01_x, dV01_y,
						dE10_x, dE10_y, 
						dE11_x, dE11_y,
						dE20_x, dE20_y, 
						dE21_x, dE21_y,
						ipEnd);
	  ASSERT( bDone );
	  return m_eResult;
	}
    }
  Interval valSqrtDisc = sqrt(m_valDisc);
  
  //////////////////////////////////////////////////
  // Check the Edge Zones

  Interval valCoef1 = ( valE10xE11*valE1E2plusE1dotE2 
		     + valPlusOrMinus*sqrt(valE1dotE1*m_valDisc) );
  Interval valCoef2 = ( - valE20xE21*valE1E2plusE1dotE2 
		     - valPlusOrMinus*sqrt(valE2dotE2*m_valDisc) );

  Interval ppvalEdgeZone[2][2];

  Interval valE1dotE10;
  Dot(valE1dotE10, ppvalE[1][0], ppvalE[1][1], pppvalV[1][0][0], pppvalV[1][0][1]);
  ppvalEdgeZone[0][0] = -( m_valE1xE2*valE1dotE10 
			 + valCoef1 );
  if ( m_iE1xE2Sign == 1 )
    {
      ppvalEdgeZone[0][0] = -ppvalEdgeZone[0][0];
    }
  m_ppiEdgeZone[0][0] = ( ppvalEdgeZone[0][0] < 0.0 );
  if ( m_ppiEdgeZone[0][0] == 1 )
    {
      return APM2D::icctNull;
    }

  Interval valE1dotE11;
  Dot(valE1dotE11, ppvalE[1][0], ppvalE[1][1], pppvalV[1][1][0], pppvalV[1][1][1]);
  ppvalEdgeZone[0][1] = ( m_valE1xE2*valE1dotE11 
			+ valCoef1 );
  if ( m_iE1xE2Sign == 1 )
    {
      ppvalEdgeZone[0][1] = -ppvalEdgeZone[0][1];
    }
  m_ppiEdgeZone[0][1] = ( ppvalEdgeZone[0][1] < 0.0 );
  if ( m_ppiEdgeZone[0][1] == 1 )
    {
      return APM2D::icctNull;
    }

  Interval valE2dotE20;
  Dot(valE2dotE20, ppvalE[2][0], ppvalE[2][1], pppvalV[2][0][0], pppvalV[2][0][1]);
  ppvalEdgeZone[1][0] = -( m_valE1xE2*valE2dotE20 
			 + valCoef2 );
  if ( m_iE1xE2Sign == 1 )
    {
      ppvalEdgeZone[1][0] = -ppvalEdgeZone[1][0];
    }
  m_ppiEdgeZone[1][0] = ( ppvalEdgeZone[1][0] < 0.0 );
  if ( m_ppiEdgeZone[1][0] == 1 )
    {
      return APM2D::icctNull;
    }

  Interval valE2dotE21;
  Dot(valE2dotE21, ppvalE[2][0], ppvalE[2][1], pppvalV[2][1][0], pppvalV[2][1][1]);
  ppvalEdgeZone[1][1] = ( m_valE1xE2*valE2dotE21 
			+ valCoef2 );
  if ( m_iE1xE2Sign == 1 )
    {
      ppvalEdgeZone[1][1] = -ppvalEdgeZone[1][1];
    }
  m_ppiEdgeZone[1][1] = ( ppvalEdgeZone[1][1] < 0.0 );
  if ( m_ppiEdgeZone[1][1] == 1 )
    {
      return APM2D::icctNull;
    }

  if ( m_ppiEdgeZone[0][0] == -1 ||
       m_ppiEdgeZone[0][1] == -1 ||
       m_ppiEdgeZone[1][0] == -1 ||
       m_ppiEdgeZone[1][1] == -1 )
    {
      bDone = InscribedCircleCenterVEEMPFUN(pdC_x, pdC_y, pdR, 
					    dV00_x, dV00_y,
					    dV0_x, dV0_y,
					    dV01_x, dV01_y,
					    dE10_x, dE10_y, 
					    dE11_x, dE11_y,
					    dE20_x, dE20_y, 
					    dE21_x, dE21_y,
					    ipEdgeZones);
      if ( bDone )
	{
	  return m_eResult;
	}
    }

  //////////////////////////////////////////////////
  // Check the Vertex Zone

  Interval pppvalVN[1][2][2]; // Endpoints
  // Vertex neighbor vectors
  //memset(pppvalVN, 0, sizeof(pppvalV));
  pppvalVN[0][0][0] = valV0_x - valV00_x;
  pppvalVN[0][0][1] = valV0_y - valV00_y;
  pppvalVN[0][1][0] = valV0_x - valV01_x;
  pppvalVN[0][1][1] = valV0_y - valV01_y;

  Interval pvalVertexZone[2];

  Interval valE1dotV00;
  Dot(valE1dotV00, ppvalE[1][0], ppvalE[1][1], pppvalVN[0][0][0], pppvalVN[0][0][1]);
  Interval valE2dotV00;
  Dot(valE2dotV00, ppvalE[2][0], ppvalE[2][1], pppvalVN[0][0][0], pppvalVN[0][0][1]);
  pvalVertexZone[0] = ( ( valE1dotE1*valE2dotV00 - m_valE1dotE2*valE1dotV00 )*valE20xE21 
		      + ( m_valE1dotE2*valE2dotV00 - valE2dotE2*valE1dotV00 )*valE10xE11 
		      + valPlusOrMinus*( sqrt(valE1dotE1)*valE2dotV00 - sqrt(valE2dotE2)*valE1dotV00 )*valSqrtDisc );
  if ( m_iE1xE2Sign == 1 )
    {
      pvalVertexZone[0] = -pvalVertexZone[0];
    }
  m_piVertexZone[0] = ( pvalVertexZone[0] < 0.0 );
  if ( m_piVertexZone[0] == 1 )
    {
      return APM2D::icctNull;
    }

  Interval valE1dotV01;
  Dot(valE1dotV01, ppvalE[1][0], ppvalE[1][1], pppvalVN[0][1][0], pppvalVN[0][1][1]);
  Interval valE2dotV01;
  Dot(valE2dotV01, ppvalE[2][0], ppvalE[2][1], pppvalVN[0][1][0], pppvalVN[0][1][1]);
  pvalVertexZone[1] = ( ( valE1dotE1*valE2dotV01 - m_valE1dotE2*valE1dotV01 )*valE20xE21 
		      + ( m_valE1dotE2*valE2dotV01 - valE2dotE2*valE1dotV01 )*valE10xE11 
		      + valPlusOrMinus*( sqrt(valE1dotE1)*valE2dotV01 - sqrt(valE2dotE2)*valE1dotV01 )*valSqrtDisc );
  if ( m_iE1xE2Sign == 1 )
    {
      pvalVertexZone[1] = -pvalVertexZone[1];
    }
  m_piVertexZone[1] = ( pvalVertexZone[1] < 0.0 );
  if ( m_piVertexZone[1] == 1 )
    {
      return APM2D::icctNull;
    }

  if ( m_piVertexZone[0] == -1 ||
       m_piVertexZone[1] == -1 )
    {
      bDone = InscribedCircleCenterVEEMPFUN(pdC_x, pdC_y, pdR, 
					    dV00_x, dV00_y,
					    dV0_x, dV0_y,
					    dV01_x, dV01_y,
					    dE10_x, dE10_y, 
					    dE11_x, dE11_y,
					    dE20_x, dE20_y, 
					    dE21_x, dE21_y,
					    ipVertexZones);
      if ( bDone )
	{
	  return m_eResult;
	}
    }

  //////////////////////////////////////////////////
  // Compute the center and radius

  Interval pvalC[2];
  pvalC[0] = ( (valE1dotE1*ppvalE[2][0] - m_valE1dotE2*ppvalE[1][0])*valE20xE21
	     + (m_valE1dotE2*ppvalE[2][0] - valE2dotE2*ppvalE[1][0])*valE10xE11
	     + valPlusOrMinus*(sqrt(valE1dotE1)*ppvalE[2][0] - sqrt(valE2dotE2)*ppvalE[1][0])*valSqrtDisc ) / m_valDen;
  pvalC[1] = ( (valE1dotE1*ppvalE[2][1] - m_valE1dotE2*ppvalE[1][1])*valE20xE21
	     + (m_valE1dotE2*ppvalE[2][1] - valE2dotE2*ppvalE[1][1])*valE10xE11
	     + valPlusOrMinus*(sqrt(valE1dotE1)*ppvalE[2][1] - sqrt(valE2dotE2)*ppvalE[1][1])*valSqrtDisc ) / m_valDen;

  Interval valR;
  valR = ( sqrt(valE1dotE1)*valE20xE21 + sqrt(valE2dotE2)*valE10xE11 + valPlusOrMinus*valSqrtDisc ) / m_valDenR;

  // Translate the answer back out away from the origin
  pvalC[0] += valV0_x;
  pvalC[1] += valV0_y;

  *pdC_x = dble(pvalC[0]);
  *pdC_y = dble(pvalC[1]);
  *pdR = dble(valR);
  return APM2D::icctVEE;
}

APM2D::InscribedCircleCenterType 
ICC_VEE_MPFUNFilter::InscribedCircleCenterVEEForce(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
						   DOUBLE dV00_x, DOUBLE dV00_y,
						   DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dV01_x, DOUBLE dV01_y,
						   DOUBLE dE10_x, DOUBLE dE10_y, 
						   DOUBLE dE11_x, DOUBLE dE11_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y)
{
  *pdC_x = 0.0;
  *pdC_y = 0.0;
  *pdR = 0.0;

  m_eIP = ipStart;
  m_eResult = APM2D::icctNull;
  m_iE1xE2 = -1;
  m_iE1xE2Sign = -1;
  m_iE1dotE2 = -1;
  m_iV0onE1 = -1;
  m_iV0onE2 = -1;
  m_iDenR = -1;
  m_iDen = -1;
  m_iDisc = -1;
  m_ppiEdgeZone[0][0] = -1;
  m_ppiEdgeZone[0][1] = -1;
  m_ppiEdgeZone[1][0] = -1;
  m_ppiEdgeZone[1][1] = -1;
  m_piVertexZone[0] = -1;
  m_piVertexZone[1] = -1;

  BOOL bDone;
  bDone = InscribedCircleCenterVEEMPFUN(pdC_x, pdC_y, pdR, 
					dV00_x, dV00_y,
					dV0_x, dV0_y,
					dV01_x, dV01_y,
					dE10_x, dE10_y, 
					dE11_x, dE11_y,
					dE20_x, dE20_y, 
					dE21_x, dE21_y,
					ipEnd);
  ASSERT( bDone );
  return m_eResult;
}

BOOL 
ICC_VEE_MPFUNFilter::InscribedCircleCenterVEEMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
						   DOUBLE dV00_x, DOUBLE dV00_y,
						   DOUBLE dV0_x, DOUBLE dV0_y,
						   DOUBLE dV01_x, DOUBLE dV01_y,
						   DOUBLE dE10_x, DOUBLE dE10_y, 
						   DOUBLE dE11_x, DOUBLE dE11_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   InstructionPointer eIPMax/* = ICC_VEE_MPFUNFilter::ipEnd*/)
{
  APM2D::m_uiInscribedCircleCenterVEEMPFUN++;

  switch ( m_eIP )
    {
    case ipStart:
      {
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	*pdC_x = 0.0;
	*pdC_y = 0.0;
	*pdR = 0.0;

	//////////////////////////////////////////////////
	// Check E1xE2

	m_rV00_x = dV00_x;
	m_rV00_y = dV00_y;
	m_rV0_x = dV0_x;
	m_rV0_y = dV0_y;
	m_rV01_x = dV01_x;
	m_rV01_y = dV01_y;
	m_rE10_x = dE10_x;
	m_rE10_y = dE10_y; 
	m_rE11_x = dE11_x;
	m_rE11_y = dE11_y;
	m_rE20_x = dE20_x;
	m_rE20_y = dE20_y; 
	m_rE21_x = dE21_x;
	m_rE21_y = dE21_y;

	//memset(m_pprE, 0, sizeof(m_pprE));
	m_pprE[1][0] = m_rE11_x - m_rE10_x;
	m_pprE[1][1] = m_rE11_y - m_rE10_y;
	m_pprE[2][0] = m_rE21_x - m_rE20_x;
	m_pprE[2][1] = m_rE21_y - m_rE20_y;
  
	Cross(m_rE1xE2, m_pprE[1][0], m_pprE[1][1], m_pprE[2][0], m_pprE[2][1]);
	Dot(m_rE1dotE2, m_pprE[1][0], m_pprE[1][1], m_pprE[2][0], m_pprE[2][1]);

	if ( m_iE1xE2 == -1 )
	  {
	    if ( m_rE1xE2 == 0.0 )
	      {
		if ( m_rE1dotE2 >= 0.0 )
		  {
		    // Colinear
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVEEMPFUNFinal++;
		    return TRUE;
		  }
		else
		  {
		    m_eResult = InscribedCircleCenterVEEAntiparallelMPFUNFilter(pdC_x, pdC_y, pdR, 
										dV00_x, dV00_y, 
										dV0_x, dV0_y, 
										dV01_x, dV01_y, 
										dE10_x, dE10_y, 
										dE11_x, dE11_y, 
										dE20_x, dE20_y, 
										dE21_x, dE21_y);
		    APM2D::m_uiInscribedCircleCenterVEEMPFUNFinal++;
		    return TRUE;
		  }
	      }
	    m_valE1xE2.ConservativeBound(dble(m_rE1xE2));
	    m_iE1xE2 = ( m_valE1xE2 == 0.0 );	    
	    m_iE1xE2Sign = ( m_rE1xE2 < 0.0 );
	  }
	if ( m_iE1dotE2 == -1 )
	  {
	    m_valE1dotE2.ConservativeBound(dble(m_rE1dotE2));
	    m_iE1dotE2 = ( m_valE1dotE2 >= 0.0 );
	  }

	m_rPlusOrMinus = 1.0;
	if ( !m_iE1xE2Sign )
	  {
	    m_rPlusOrMinus = -1.0;
	  }

      }
    case ipE1xE2:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Check V0 on E1 or E2

	if ( m_iV0onE1 == -1 )
	  {
	    mp_real rV0onE1;
	    SinTheta(rV0onE1, m_rV0_x, m_rV0_y, m_rE10_x, m_rE10_y, m_rE11_x, m_rE11_y);
	    if ( rV0onE1 <= 0.0 )
	      {
		// V0 is on E1 and it is not one of the endpoints
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterVEEMPFUNFinal++;
		return TRUE;
	      }
	  }
  
	if ( m_iV0onE2 == -1 )
	  {
	    mp_real rV0onE2;
	    SinTheta(rV0onE2, m_rV0_x, m_rV0_y, m_rE20_x, m_rE20_y, m_rE21_x, m_rE21_y);
	    if ( rV0onE2 <= 0.0 )
	      {
		// V0 is on E2 and it is not one of the endpoints
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterVEEMPFUNFinal++;
		return TRUE;
	      }
	  }
  
      }
    case ipV0onE1orE2:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Check DenR

	Dot(m_rE1dotE1, m_pprE[1][0], m_pprE[1][1], m_pprE[1][0], m_pprE[1][1]);
	Dot(m_rE2dotE2, m_pprE[2][0], m_pprE[2][1], m_pprE[2][0], m_pprE[2][1]);

	m_rE1E2minusE1dotE2 = sqrt(m_rE1dotE1*m_rE2dotE2) - m_rE1dotE2;

	m_rDenR = m_rE1E2minusE1dotE2;
	if ( m_iDenR == -1 )
	  {
	    ASSERT( m_rDenR != 0.0 );
	    m_valDenR.ConservativeBound(dble(m_rDenR));
	    m_iDenR = ( m_valDenR != 0.0 );
	    
	  }

      }
    case ipDenR:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Check Den

	m_rDen = m_rE1xE2*m_rE1E2minusE1dotE2;
	if ( m_iDen == -1 )
	  {
	    ASSERT( m_rDen != 0.0 );
	    m_valDen.ConservativeBound(dble(m_rDen));
	    m_iDen = ( m_valDen != 0.0 );
	  }

      }
    case ipDen:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Check Disc

	// Translate V1 to the origin
	//memset(m_ppprV, 0, sizeof(m_ppprV));
	m_ppprV[0][0][0] = 0.0;
	m_ppprV[0][0][1] = 0.0;
	m_ppprV[1][0][0] = m_rE10_x - m_rV0_x;
	m_ppprV[1][0][1] = m_rE10_y - m_rV0_y;
	m_ppprV[1][1][0] = m_rE11_x - m_rV0_x;
	m_ppprV[1][1][1] = m_rE11_y - m_rV0_y;
	m_ppprV[2][0][0] = m_rE20_x - m_rV0_x;
	m_ppprV[2][0][1] = m_rE20_y - m_rV0_y;
	m_ppprV[2][1][0] = m_rE21_x - m_rV0_x;
	m_ppprV[2][1][1] = m_rE21_y - m_rV0_y;
  
	Cross(m_rE10xE11, m_ppprV[1][0][0], m_ppprV[1][0][1], m_ppprV[1][1][0], m_ppprV[1][1][1]);
	Cross(m_rE20xE21, m_ppprV[2][0][0], m_ppprV[2][0][1], m_ppprV[2][1][0], m_ppprV[2][1][1]);

	m_rE1E2plusE1dotE2 = sqrt(m_rE1dotE1*m_rE2dotE2) + m_rE1dotE2;

	m_rDisc = 2.0*m_rE10xE11*m_rE20xE21*m_rE1E2plusE1dotE2;
	if ( m_iDisc == -1 )
	  {
	    if ( m_rDisc < 0.0 ) 
	      {
		// Imaginary Solutions
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterVEEMPFUNFinal++;
		return TRUE;
	      }
	    m_valDisc.ConservativeBound(dble(m_rDisc));
	    m_iDisc = ( m_valDisc < 0.0 );	    
	  }
	m_rSqrtDisc = sqrt(m_rDisc);
  
      }
    case ipDisc:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Check the Edge Zones

	if ( m_ppiEdgeZone[0][0] == -1 ||
	     m_ppiEdgeZone[0][1] == -1 ||
	     m_ppiEdgeZone[1][0] == -1 ||
	     m_ppiEdgeZone[1][1] == -1 )
	  {
	    mp_real rCoef1 = ( m_rE10xE11*m_rE1E2plusE1dotE2 
			       + m_rPlusOrMinus*sqrt(m_rE1dotE1*m_rDisc) );
	    mp_real rCoef2 = ( - m_rE20xE21*m_rE1E2plusE1dotE2 
			       - m_rPlusOrMinus*sqrt(m_rE2dotE2*m_rDisc) );

	    mp_real pprEdgeZone[2][2];

	    if ( m_ppiEdgeZone[0][0] == -1 )
	      {
		mp_real rE1dotE10;
		Dot(rE1dotE10, m_pprE[1][0], m_pprE[1][1], m_ppprV[1][0][0], m_ppprV[1][0][1]);
		pprEdgeZone[0][0] = -( m_rE1xE2*rE1dotE10 
				       + rCoef1 );
		if ( m_iE1xE2Sign )
		  {
		    pprEdgeZone[0][0] = -pprEdgeZone[0][0];
		  }
		if ( pprEdgeZone[0][0] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVEEMPFUNFinal++;
		    return TRUE;
		  }
	      }

	    if ( m_ppiEdgeZone[0][1] == -1 )
	      {
		mp_real rE1dotE11;
		Dot(rE1dotE11, m_pprE[1][0], m_pprE[1][1], m_ppprV[1][1][0], m_ppprV[1][1][1]);
		pprEdgeZone[0][1] = ( m_rE1xE2*rE1dotE11 
				      + rCoef1 );
		if ( m_iE1xE2Sign )
		  {
		    pprEdgeZone[0][1] = -pprEdgeZone[0][1];
		  }
		if ( pprEdgeZone[0][1] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVEEMPFUNFinal++;
		    return TRUE;
		  }
	      }

	    if ( m_ppiEdgeZone[1][0] == -1 )
	      {
		mp_real rE2dotE20;
		Dot(rE2dotE20, m_pprE[2][0], m_pprE[2][1], m_ppprV[2][0][0], m_ppprV[2][0][1]);
		pprEdgeZone[1][0] = -( m_rE1xE2*rE2dotE20 
				       + rCoef2 );
		if ( m_iE1xE2Sign )
		  {
		    pprEdgeZone[1][0] = -pprEdgeZone[1][0];
		  }
		if ( pprEdgeZone[1][0] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVEEMPFUNFinal++;
		    return TRUE;
		  }
	      }

	    if ( m_ppiEdgeZone[1][1] == -1 )
	      {
		mp_real rE2dotE21;
		Dot(rE2dotE21, m_pprE[2][0], m_pprE[2][1], m_ppprV[2][1][0], m_ppprV[2][1][1]);
		pprEdgeZone[1][1] = ( m_rE1xE2*rE2dotE21 
				      + rCoef2 );
		if ( m_iE1xE2Sign )
		  {
		    pprEdgeZone[1][1] = -pprEdgeZone[1][1];
		  }
		if ( pprEdgeZone[1][1] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVEEMPFUNFinal++;
		    return TRUE;
		  }
	      }
	  }
      }
    case ipEdgeZones:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Check the Vertex Zone

	if ( m_piVertexZone[0] == -1 ||
	     m_piVertexZone[0] == -1 )
	  {
	    mp_real ppprVN[1][2][2]; // Endpoints
	    // Vertex neighbor vectors
	    //memset(ppprVN, 0, sizeof(ppprV));
	    ppprVN[0][0][0] = m_rV0_x - m_rV00_x;
	    ppprVN[0][0][1] = m_rV0_y - m_rV00_y;
	    ppprVN[0][1][0] = m_rV0_x - m_rV01_x;
	    ppprVN[0][1][1] = m_rV0_y - m_rV01_y;

	    mp_real prVertexZone[2];

	    if ( m_piVertexZone[0] == -1 )
	      {
		mp_real rE1dotV00;
		Dot(rE1dotV00, m_pprE[1][0], m_pprE[1][1], ppprVN[0][0][0], ppprVN[0][0][1]);
		mp_real rE2dotV00;
		Dot(rE2dotV00, m_pprE[2][0], m_pprE[2][1], ppprVN[0][0][0], ppprVN[0][0][1]);
		prVertexZone[0] = ( ( m_rE1dotE1*rE2dotV00 - m_rE1dotE2*rE1dotV00 )*m_rE20xE21 
				    + ( m_rE1dotE2*rE2dotV00 - m_rE2dotE2*rE1dotV00 )*m_rE10xE11 
				    + m_rPlusOrMinus*( sqrt(m_rE1dotE1)*rE2dotV00 - sqrt(m_rE2dotE2)*rE1dotV00 )*m_rSqrtDisc );
		if ( m_iE1xE2Sign )
		  {
		    prVertexZone[0] = -prVertexZone[0];
		  }
		if ( prVertexZone[0] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVEEMPFUNFinal++;
		    return TRUE;
		  }
	      }

	    if ( m_piVertexZone[1] == -1 )
	      {
		mp_real rE1dotV01;
		Dot(rE1dotV01, m_pprE[1][0], m_pprE[1][1], ppprVN[0][1][0], ppprVN[0][1][1]);
		mp_real rE2dotV01;
		Dot(rE2dotV01, m_pprE[2][0], m_pprE[2][1], ppprVN[0][1][0], ppprVN[0][1][1]);
		prVertexZone[1] = ( ( m_rE1dotE1*rE2dotV01 - m_rE1dotE2*rE1dotV01 )*m_rE20xE21 
				    + ( m_rE1dotE2*rE2dotV01 - m_rE2dotE2*rE1dotV01 )*m_rE10xE11 
				    + m_rPlusOrMinus*( sqrt(m_rE1dotE1)*rE2dotV01 - sqrt(m_rE2dotE2)*rE1dotV01 )*m_rSqrtDisc );
		if ( m_iE1xE2Sign )
		  {
		    prVertexZone[1] = -prVertexZone[1];
		  }
		if ( prVertexZone[1] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVEEMPFUNFinal++;
		    return TRUE;
		  }
	      }
	  }

      }
    case ipVertexZones:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Compute the center and radius

	mp_real prC[2];
	prC[0] = ( (m_rE1dotE1*m_pprE[2][0] - m_rE1dotE2*m_pprE[1][0])*m_rE20xE21
		   + (m_rE1dotE2*m_pprE[2][0] - m_rE2dotE2*m_pprE[1][0])*m_rE10xE11
		   + m_rPlusOrMinus*(sqrt(m_rE1dotE1)*m_pprE[2][0] - sqrt(m_rE2dotE2)*m_pprE[1][0])*m_rSqrtDisc ) / m_rDen;
	prC[1] = ( (m_rE1dotE1*m_pprE[2][1] - m_rE1dotE2*m_pprE[1][1])*m_rE20xE21
		   + (m_rE1dotE2*m_pprE[2][1] - m_rE2dotE2*m_pprE[1][1])*m_rE10xE11
		   + m_rPlusOrMinus*(sqrt(m_rE1dotE1)*m_pprE[2][1] - sqrt(m_rE2dotE2)*m_pprE[1][1])*m_rSqrtDisc ) / m_rDen;

	mp_real rR;
	rR = ( sqrt(m_rE1dotE1)*m_rE20xE21 + sqrt(m_rE2dotE2)*m_rE10xE11 + m_rPlusOrMinus*m_rSqrtDisc ) / m_rDenR;

	// Translate the answer back out away from the origin
	prC[0] += m_rV0_x;
	prC[1] += m_rV0_y;

	*pdC_x = dble(prC[0]);
	*pdC_y = dble(prC[1]);
	*pdR = dble(rR);
	m_eResult = APM2D::icctVEE;
      }
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  APM2D::m_uiInscribedCircleCenterVEEMPFUNFinal++;
  return TRUE;
}

//static 
APM2D::InscribedCircleCenterType 
InscribedCircleCenterVEEMPFUNFilter(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				    DOUBLE dV00_x, DOUBLE dV00_y,
				    DOUBLE dV0_x, DOUBLE dV0_y,
				    DOUBLE dV01_x, DOUBLE dV01_y,
				    DOUBLE dE10_x, DOUBLE dE10_y, 
				    DOUBLE dE11_x, DOUBLE dE11_y,
				    DOUBLE dE20_x, DOUBLE dE20_y, 
				    DOUBLE dE21_x, DOUBLE dE21_y)
{
  ICC_VEE_MPFUNFilter iccVEE;

  return iccVEE.InscribedCircleCenterVEE(pdC_x, pdC_y, pdR, 
					 dV00_x, dV00_y,
					 dV0_x, dV0_y,
					 dV01_x, dV01_y,
					 dE10_x, dE10_y, 
					 dE11_x, dE11_y,
					 dE20_x, dE20_y, 
					 dE21_x, dE21_y);
}

//static 
APM2D::EdgeZoneType EdgeZoneVEEMPFUNFilter(DOUBLE dV0_x, DOUBLE dV0_y,
					   DOUBLE dE10_x, DOUBLE dE10_y, 
					   DOUBLE dE11_x, DOUBLE dE11_y,
					   DOUBLE dE20_x, DOUBLE dE20_y, 
					   DOUBLE dE21_x, DOUBLE dE21_y,
					   DOUBLE dE30_x, DOUBLE dE30_y,
					   DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::m_uiEdgeZoneVEEMPFUNFilter++;

  INT_32 iE1xE2 = -1;
  INT_32 iDisc = -1;
  INT_32 piEdgeZone[2];
  piEdgeZone[0] = -1;
  piEdgeZone[1] = -1;

  switch ( 0 )
  {
  default:
    Interval valV0_x = dV0_x;
    Interval valV0_y = dV0_y;
    Interval valE10_x = dE10_x;
    Interval valE10_y = dE10_y; 
    Interval valE11_x = dE11_x;
    Interval valE11_y = dE11_y;
    Interval valE20_x = dE20_x;
    Interval valE20_y = dE20_y; 
    Interval valE21_x = dE21_x;
    Interval valE21_y = dE21_y;
    Interval valE30_x = dE30_x;
    Interval valE30_y = dE30_y;
    Interval valE31_x = dE31_x;
    Interval valE31_y = dE31_y;

    Interval pppvalV[4][2][2]; // Endpoints
    //memset(pppvalV, 0, sizeof(pppvalV));
    pppvalV[0][0][0] = 0.0;
    pppvalV[0][0][1] = 0.0;
    pppvalV[1][0][0] = valE10_x - valV0_x;
    pppvalV[1][0][1] = valE10_y - valV0_y;
    pppvalV[1][1][0] = valE11_x - valV0_x;
    pppvalV[1][1][1] = valE11_y - valV0_y;
    pppvalV[2][0][0] = valE20_x - valV0_x;
    pppvalV[2][0][1] = valE20_y - valV0_y;
    pppvalV[2][1][0] = valE21_x - valV0_x;
    pppvalV[2][1][1] = valE21_y - valV0_y;
    pppvalV[3][0][0] = valE30_x - valV0_x;
    pppvalV[3][0][1] = valE30_y - valV0_y;
    pppvalV[3][1][0] = valE31_x - valV0_x;
    pppvalV[3][1][1] = valE31_y - valV0_y;
  
    Interval ppvalE[4][2];     // Edge vectors
    //memset(ppvalE, 0, sizeof(ppvalE));
    ppvalE[1][0] = valE11_x - valE10_x;
    ppvalE[1][1] = valE11_y - valE10_y;
    ppvalE[2][0] = valE21_x - valE20_x;
    ppvalE[2][1] = valE21_y - valE20_y;
    ppvalE[3][0] = valE31_x - valE30_x;
    ppvalE[3][1] = valE31_y - valE30_y;
  
    Interval valE1xE2;
    Cross(valE1xE2, ppvalE[1][0], ppvalE[1][1], ppvalE[2][0], ppvalE[2][1]);

    Interval valPlusOrMinus;
    valPlusOrMinus = 1.0;
    iE1xE2 = ( valE1xE2 > 0.0 );
    if ( iE1xE2 == -1 )
      {
	break;
      }
    ASSERT( valE1xE2 != 0.0 );
    if ( iE1xE2 == 1 )
      {
	valPlusOrMinus = -1.0;
      }
  
    Interval valE1dotE1;
    Dot(valE1dotE1, ppvalE[1][0], ppvalE[1][1], ppvalE[1][0], ppvalE[1][1]);
    Interval valE2dotE2;
    Dot(valE2dotE2, ppvalE[2][0], ppvalE[2][1], ppvalE[2][0], ppvalE[2][1]);
    Interval valE1dotE2;
    Dot(valE1dotE2, ppvalE[1][0], ppvalE[1][1], ppvalE[2][0], ppvalE[2][1]);

    Interval valE10xE11;
    Cross(valE10xE11, pppvalV[1][0][0], pppvalV[1][0][1], pppvalV[1][1][0], pppvalV[1][1][1]);
    Interval valE20xE21;
    Cross(valE20xE21, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][1][0], pppvalV[2][1][1]);

    Interval valDisc;
    valDisc = 2.0*valE10xE11*valE20xE21*( sqrt(valE1dotE1*valE2dotE2) + valE1dotE2 );
    iDisc = ( valDisc >= 0.0 ); 
    if ( iDisc == -1 )
      {
	break;
      }
    ASSERT( iDisc == 1 ); 
    Interval valSqrtDisc = sqrt(valDisc);

    Interval valE2dotE3;
    Dot(valE2dotE3, ppvalE[2][0], ppvalE[2][1], ppvalE[3][0], ppvalE[3][1]);
    Interval valE1dotE3;
    Dot(valE1dotE3, ppvalE[1][0], ppvalE[1][1], ppvalE[3][0], ppvalE[3][1]);

    Interval valE1E2minusE1dotE2;
    valE1E2minusE1dotE2 = ( sqrt(valE1dotE1*valE2dotE2) - valE1dotE2 );

    Interval valCoef1 = valE1xE2*valE1E2minusE1dotE2;
    Interval valCoef2 = ( ( valE2dotE2*valE1dotE3 - valE1dotE2*valE2dotE3 )*valE10xE11
		       - ( valE1dotE1*valE2dotE3 - valE1dotE2*valE1dotE3 )*valE20xE21
		       + valPlusOrMinus*( sqrt(valE2dotE2)*valE1dotE3 - sqrt(valE1dotE1)*valE2dotE3 )*valSqrtDisc );

    Interval pvalEdgeZone[2];

    Interval valE3dotE30;
    Dot(valE3dotE30, ppvalE[3][0], ppvalE[3][1], pppvalV[3][0][0], pppvalV[3][0][1]);
    if ( iE1xE2 == 1 )
      {
	pvalEdgeZone[0] = ( valCoef1*valE3dotE30
			  + valCoef2 );
      }
    else
      {
	pvalEdgeZone[0] = -( valCoef1*valE3dotE30
			   + valCoef2 );
      }
    piEdgeZone[0] = ( pvalEdgeZone[0] >= 0.0 );
    if ( piEdgeZone[0] == 1 )  
      {
	return APM2D::eztStart;
      }
    
    Interval valE3dotE31;
    Dot(valE3dotE31, ppvalE[3][0], ppvalE[3][1], pppvalV[3][1][0], pppvalV[3][1][1]);
    if ( iE1xE2 == 1 )
      {
	pvalEdgeZone[1] = -( valCoef1*valE3dotE31
			   + valCoef2 );
      }
    else
      {
	pvalEdgeZone[1] = ( valCoef1*valE3dotE31
			  + valCoef2 );
      }
    piEdgeZone[1] = ( pvalEdgeZone[1] >= 0.0 );
    if ( piEdgeZone[1] == 1 )
      {
	return APM2D::eztEnd;
      }
  }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( piEdgeZone[0] == -1 || piEdgeZone[1] == -1 )
    {
      APM2D::m_uiEdgeZoneVEEMPFUNFilter++;

      mp_real rV0_x = dV0_x;
      mp_real rV0_y = dV0_y;
      mp_real rE10_x = dE10_x;
      mp_real rE10_y = dE10_y; 
      mp_real rE11_x = dE11_x;
      mp_real rE11_y = dE11_y;
      mp_real rE20_x = dE20_x;
      mp_real rE20_y = dE20_y; 
      mp_real rE21_x = dE21_x;
      mp_real rE21_y = dE21_y;
      mp_real rE30_x = dE30_x;
      mp_real rE30_y = dE30_y;
      mp_real rE31_x = dE31_x;
      mp_real rE31_y = dE31_y;

      mp_real ppprV[4][2][2]; // Endpoints
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = 0.0;
      ppprV[0][0][1] = 0.0;
      ppprV[1][0][0] = rE10_x - rV0_x;
      ppprV[1][0][1] = rE10_y - rV0_y;
      ppprV[1][1][0] = rE11_x - rV0_x;
      ppprV[1][1][1] = rE11_y - rV0_y;
      ppprV[2][0][0] = rE20_x - rV0_x;
      ppprV[2][0][1] = rE20_y - rV0_y;
      ppprV[2][1][0] = rE21_x - rV0_x;
      ppprV[2][1][1] = rE21_y - rV0_y;
      ppprV[3][0][0] = rE30_x - rV0_x;
      ppprV[3][0][1] = rE30_y - rV0_y;
      ppprV[3][1][0] = rE31_x - rV0_x;
      ppprV[3][1][1] = rE31_y - rV0_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[1][0] = rE11_x - rE10_x;
      pprE[1][1] = rE11_y - rE10_y;
      pprE[2][0] = rE21_x - rE20_x;
      pprE[2][1] = rE21_y - rE20_y;
      pprE[3][0] = rE31_x - rE30_x;
      pprE[3][1] = rE31_y - rE30_y;
  
      mp_real rE1xE2;
      Cross(rE1xE2, pprE[1][0], pprE[1][1], pprE[2][0], pprE[2][1]);

      mp_real rPlusOrMinus;
      rPlusOrMinus = 1.0;
      iE1xE2 = 0;
      ASSERT( rE1xE2 != 0.0 );
      if ( rE1xE2 > 0.0 )
	{
	  iE1xE2 = 1;
	  rPlusOrMinus = -1.0;
	}
  
      mp_real rE1dotE1;
      Dot(rE1dotE1, pprE[1][0], pprE[1][1], pprE[1][0], pprE[1][1]);
      mp_real rE2dotE2;
      Dot(rE2dotE2, pprE[2][0], pprE[2][1], pprE[2][0], pprE[2][1]);
      mp_real rE1dotE2;
      Dot(rE1dotE2, pprE[1][0], pprE[1][1], pprE[2][0], pprE[2][1]);

      mp_real rE10xE11;
      Cross(rE10xE11, ppprV[1][0][0], ppprV[1][0][1], ppprV[1][1][0], ppprV[1][1][1]);
      mp_real rE20xE21;
      Cross(rE20xE21, ppprV[2][0][0], ppprV[2][0][1], ppprV[2][1][0], ppprV[2][1][1]);

      mp_real rDisc;
      rDisc = 2.0*rE10xE11*rE20xE21*( sqrt(rE1dotE1*rE2dotE2) + rE1dotE2 );
      ASSERT( rDisc >= 0.0 ); 
      mp_real rSqrtDisc = sqrt(rDisc);

      mp_real rE2dotE3;
      Dot(rE2dotE3, pprE[2][0], pprE[2][1], pprE[3][0], pprE[3][1]);
      mp_real rE1dotE3;
      Dot(rE1dotE3, pprE[1][0], pprE[1][1], pprE[3][0], pprE[3][1]);

      mp_real rE1E2minusE1dotE2;
      rE1E2minusE1dotE2 = ( sqrt(rE1dotE1*rE2dotE2) - rE1dotE2 );

      mp_real rCoef1 = rE1xE2*rE1E2minusE1dotE2;
      mp_real rCoef2 = ( ( rE2dotE2*rE1dotE3 - rE1dotE2*rE2dotE3 )*rE10xE11
			 - ( rE1dotE1*rE2dotE3 - rE1dotE2*rE1dotE3 )*rE20xE21
			 + rPlusOrMinus*( sqrt(rE2dotE2)*rE1dotE3 - sqrt(rE1dotE1)*rE2dotE3 )*rSqrtDisc );

      mp_real prEdgeZone[2];

      if ( piEdgeZone[0] == -1 )
	{
	  mp_real rE3dotE30;
	  Dot(rE3dotE30, pprE[3][0], pprE[3][1], ppprV[3][0][0], ppprV[3][0][1]);
	  if ( iE1xE2 == 1 )
	    {
	      prEdgeZone[0] = ( rCoef1*rE3dotE30
				+ rCoef2 );
	    }
	  else
	    {
	      prEdgeZone[0] = -( rCoef1*rE3dotE30
				 + rCoef2 );
	    }
	  if ( prEdgeZone[0] >= 0.0 )
	    {
	      return APM2D::eztStart;
	    }
	}

      if ( piEdgeZone[1] == -1 )
	{
	  mp_real rE3dotE31;
	  Dot(rE3dotE31, pprE[3][0], pprE[3][1], ppprV[3][1][0], ppprV[3][1][1]);
	  if ( iE1xE2 == 1 )
	    {
	      prEdgeZone[1] = -( rCoef1*rE3dotE31
				 + rCoef2 );
	    }
	  else
	    {
	      prEdgeZone[1] = ( rCoef1*rE3dotE31
				+ rCoef2 );
	    }
	  if ( prEdgeZone[1] >= 0.0 )
	    {
	      return APM2D::eztEnd;
	    }
	}
    }

  return APM2D::eztMiddle;
}

//static 
APM2D::CircleTestType InCircleVEE_VMPFUNFilter(DOUBLE dV0_x, DOUBLE dV0_y,
					       DOUBLE dE10_x, DOUBLE dE10_y, 
					       DOUBLE dE11_x, DOUBLE dE11_y,
					       DOUBLE dE20_x, DOUBLE dE20_y, 
					       DOUBLE dE21_x, DOUBLE dE21_y,
					       DOUBLE dV3_x, DOUBLE dV3_y)
{
  APM2D::m_uiInCircleVEE_VMPFUNFilter++;

  INT_32 iE1xE2 = -1;
  INT_32 iDen = -1;
  INT_32 iDisc = -1;
  INT_32 iResult = -1;

  switch ( 0 )
    {
    default:
      Interval valV0_x = dV0_x;
      Interval valV0_y = dV0_y;
      Interval valE10_x = dE10_x;
      Interval valE10_y = dE10_y; 
      Interval valE11_x = dE11_x;
      Interval valE11_y = dE11_y;
      Interval valE20_x = dE20_x;
      Interval valE20_y = dE20_y; 
      Interval valE21_x = dE21_x;
      Interval valE21_y = dE21_y;
      Interval valV3_x = dV3_x;
      Interval valV3_y = dV3_y;

      Interval pppvalV[4][2][2]; // Endpoints
      //memset(pppvalV, 0, sizeof(pppvalV));
      pppvalV[0][0][0] = 0.0;
      pppvalV[0][0][1] = 0.0;
      pppvalV[1][0][0] = valE10_x - valV0_x;
      pppvalV[1][0][1] = valE10_y - valV0_y;
      pppvalV[1][1][0] = valE11_x - valV0_x;
      pppvalV[1][1][1] = valE11_y - valV0_y;
      pppvalV[2][0][0] = valE20_x - valV0_x;
      pppvalV[2][0][1] = valE20_y - valV0_y;
      pppvalV[2][1][0] = valE21_x - valV0_x;
      pppvalV[2][1][1] = valE21_y - valV0_y;
      pppvalV[3][0][0] = valV3_x - valV0_x;
      pppvalV[3][0][1] = valV3_y - valV0_y;
  
      Interval ppvalE[4][2];     // Edge vectors
      //memset(ppvalE, 0, sizeof(ppvalE));
      ppvalE[1][0] = valE11_x - valE10_x;
      ppvalE[1][1] = valE11_y - valE10_y;
      ppvalE[2][0] = valE21_x - valE20_x;
      ppvalE[2][1] = valE21_y - valE20_y;

      Interval valE1xE2;
      Cross(valE1xE2, ppvalE[1][0], ppvalE[1][1], ppvalE[2][0], ppvalE[2][1]);

      Interval valPlusOrMinus;
      valPlusOrMinus = 1.0;
      iE1xE2 = ( valE1xE2 > 0.0 );
      if ( iE1xE2 == -1 )
	{
	  break;
	}
      ASSERT( valE1xE2 != 0.0 );
      if ( iE1xE2 == 1 )
	{
	  valPlusOrMinus = -1.0;
	}

      Interval valE1dotE1;
      Dot(valE1dotE1, ppvalE[1][0], ppvalE[1][1], ppvalE[1][0], ppvalE[1][1]);
      Interval valE2dotE2;
      Dot(valE2dotE2, ppvalE[2][0], ppvalE[2][1], ppvalE[2][0], ppvalE[2][1]);
      Interval valE1dotE2;
      Dot(valE1dotE2, ppvalE[1][0], ppvalE[1][1], ppvalE[2][0], ppvalE[2][1]);

      Interval valE1E2minusE1dotE2;
      valE1E2minusE1dotE2 = ( sqrt(valE1dotE1*valE2dotE2) - valE1dotE2 );

      Interval valDen;
      valDen = valE1xE2*valE1E2minusE1dotE2;
      iDen = ( valDen != 0.0 );
      if ( iDen == -1 )
	{
	  break;
	}
      ASSERT( iDen == 1 );

      Interval valE10xE11;
      Cross(valE10xE11, pppvalV[1][0][0], pppvalV[1][0][1], pppvalV[1][1][0], pppvalV[1][1][1]);
      Interval valE20xE21;
      Cross(valE20xE21, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][1][0], pppvalV[2][1][1]);

      Interval valDisc;
      valDisc = 2.0*valE10xE11*valE20xE21*( sqrt(valE1dotE1*valE2dotE2) + valE1dotE2 );
      iDisc = ( valDisc >= 0.0 ); 
      if ( iDisc == -1 )
	{
	  break;
	}
      ASSERT( iDisc == 1 ); 
      Interval valSqrtDisc = sqrt(valDisc);

      Interval valV3dotV3;
      Dot(valV3dotV3, pppvalV[3][0][0], pppvalV[3][0][1], pppvalV[3][0][0], pppvalV[3][0][1]);

      Interval valE2dotV3;
      Dot(valE2dotV3, ppvalE[2][0], ppvalE[2][1], pppvalV[3][0][0], pppvalV[3][0][1]);
      Interval valE1dotV3;
      Dot(valE1dotV3, ppvalE[1][0], ppvalE[1][1], pppvalV[3][0][0], pppvalV[3][0][1]);

      Interval valResult;
      valResult = ( valDen*valV3dotV3
		  + 2.0*( valE2dotE2*valE1dotV3 - valE1dotE2*valE2dotV3 )*valE10xE11
		  - 2.0*( valE1dotE1*valE2dotV3 - valE1dotE2*valE1dotV3 )*valE20xE21
		  + valPlusOrMinus*2.0*( sqrt(valE2dotE2)*valE1dotV3 - sqrt(valE1dotE1)*valE2dotV3 )*valSqrtDisc ); 
      if ( iE1xE2 == 0 )
	{
	  valResult = -valResult;
	}
      iResult = ( valResult < 0.0 );
      if ( iResult == -1 )
	{
	  break;
	}
      if ( iResult == 1 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( valResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
      return APM2D::cttOnCircle;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( iResult == -1 )
    {
      APM2D::m_uiInCircleVEE_VMPFUN++;

      mp_real rV0_x = dV0_x;
      mp_real rV0_y = dV0_y;
      mp_real rE10_x = dE10_x;
      mp_real rE10_y = dE10_y; 
      mp_real rE11_x = dE11_x;
      mp_real rE11_y = dE11_y;
      mp_real rE20_x = dE20_x;
      mp_real rE20_y = dE20_y; 
      mp_real rE21_x = dE21_x;
      mp_real rE21_y = dE21_y;
      mp_real rV3_x = dV3_x;
      mp_real rV3_y = dV3_y;

      mp_real ppprV[4][2][2]; // Endpoints
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = 0.0;
      ppprV[0][0][1] = 0.0;
      ppprV[1][0][0] = rE10_x - rV0_x;
      ppprV[1][0][1] = rE10_y - rV0_y;
      ppprV[1][1][0] = rE11_x - rV0_x;
      ppprV[1][1][1] = rE11_y - rV0_y;
      ppprV[2][0][0] = rE20_x - rV0_x;
      ppprV[2][0][1] = rE20_y - rV0_y;
      ppprV[2][1][0] = rE21_x - rV0_x;
      ppprV[2][1][1] = rE21_y - rV0_y;
      ppprV[3][0][0] = rV3_x - rV0_x;
      ppprV[3][0][1] = rV3_y - rV0_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[1][0] = rE11_x - rE10_x;
      pprE[1][1] = rE11_y - rE10_y;
      pprE[2][0] = rE21_x - rE20_x;
      pprE[2][1] = rE21_y - rE20_y;

      mp_real rE1xE2;
      Cross(rE1xE2, pprE[1][0], pprE[1][1], pprE[2][0], pprE[2][1]);

      mp_real rPlusOrMinus;
      rPlusOrMinus = 1.0;
      iE1xE2 = 0;
      ASSERT( rE1xE2 != 0.0 );
      if ( rE1xE2 > 0.0 )
	{
	  iE1xE2 = 1;
	  rPlusOrMinus = -1.0;
	}
 
      mp_real rE1dotE1;
      Dot(rE1dotE1, pprE[1][0], pprE[1][1], pprE[1][0], pprE[1][1]);
      mp_real rE2dotE2;
      Dot(rE2dotE2, pprE[2][0], pprE[2][1], pprE[2][0], pprE[2][1]);
      mp_real rE1dotE2;
      Dot(rE1dotE2, pprE[1][0], pprE[1][1], pprE[2][0], pprE[2][1]);

      mp_real rE1E2minusE1dotE2;
      rE1E2minusE1dotE2 = ( sqrt(rE1dotE1*rE2dotE2) - rE1dotE2 );

      mp_real rDen;
      rDen = rE1xE2*rE1E2minusE1dotE2;
      ASSERT( rDen != 0.0 );

      mp_real rE10xE11;
      Cross(rE10xE11, ppprV[1][0][0], ppprV[1][0][1], ppprV[1][1][0], ppprV[1][1][1]);
      mp_real rE20xE21;
      Cross(rE20xE21, ppprV[2][0][0], ppprV[2][0][1], ppprV[2][1][0], ppprV[2][1][1]);

      mp_real rDisc;
      rDisc = 2.0*rE10xE11*rE20xE21*( sqrt(rE1dotE1*rE2dotE2) + rE1dotE2 );
      ASSERT( rDisc >= 0.0 );
      mp_real rSqrtDisc = sqrt(rDisc);

      mp_real rV3dotV3;
      Dot(rV3dotV3, ppprV[3][0][0], ppprV[3][0][1], ppprV[3][0][0], ppprV[3][0][1]);

      mp_real rE2dotV3;
      Dot(rE2dotV3, pprE[2][0], pprE[2][1], ppprV[3][0][0], ppprV[3][0][1]);
      mp_real rE1dotV3;
      Dot(rE1dotV3, pprE[1][0], pprE[1][1], ppprV[3][0][0], ppprV[3][0][1]);

      mp_real rResult;
      rResult = ( rDen*rV3dotV3
		  + 2.0*( rE2dotE2*rE1dotV3 - rE1dotE2*rE2dotV3 )*rE10xE11
		  - 2.0*( rE1dotE1*rE2dotV3 - rE1dotE2*rE1dotV3 )*rE20xE21
		  + rPlusOrMinus*2.0*( sqrt(rE2dotE2)*rE1dotV3 - sqrt(rE1dotE1)*rE2dotV3 )*rSqrtDisc ); 
      if ( iE1xE2 == 0 )
	{
	  rResult = -rResult;
	}
      if ( rResult < 0.0 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( rResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
    }

  return APM2D::cttOnCircle;
}

//static 
APM2D::CircleTestType InCircleVEE_EMPFUNFilter(DOUBLE dV0_x, DOUBLE dV0_y,
					       DOUBLE dE10_x, DOUBLE dE10_y, 
					       DOUBLE dE11_x, DOUBLE dE11_y,
					       DOUBLE dE20_x, DOUBLE dE20_y, 
					       DOUBLE dE21_x, DOUBLE dE21_y,
					       DOUBLE dE30_x, DOUBLE dE30_y,
					       DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::m_uiInCircleVEE_EMPFUNFilter++;

  INT_32 iE1xE2 = -1;
  INT_32 iDisc = -1;
  INT_32 iResult = -1;

  switch ( 0 )
    {
    default:
      Interval valV0_x = dV0_x;
      Interval valV0_y = dV0_y;
      Interval valE10_x = dE10_x;
      Interval valE10_y = dE10_y; 
      Interval valE11_x = dE11_x;
      Interval valE11_y = dE11_y;
      Interval valE20_x = dE20_x;
      Interval valE20_y = dE20_y; 
      Interval valE21_x = dE21_x;
      Interval valE21_y = dE21_y;
      Interval valE30_x = dE30_x;
      Interval valE30_y = dE30_y;
      Interval valE31_x = dE31_x;
      Interval valE31_y = dE31_y;

      Interval pppvalV[4][2][2]; // Endpoints
      //memset(pppvalV, 0, sizeof(pppvalV));
      pppvalV[0][0][0] = 0.0;
      pppvalV[0][0][1] = 0.0;
      pppvalV[1][0][0] = valE10_x - valV0_x;
      pppvalV[1][0][1] = valE10_y - valV0_y;
      pppvalV[1][1][0] = valE11_x - valV0_x;
      pppvalV[1][1][1] = valE11_y - valV0_y;
      pppvalV[2][0][0] = valE20_x - valV0_x;
      pppvalV[2][0][1] = valE20_y - valV0_y;
      pppvalV[2][1][0] = valE21_x - valV0_x;
      pppvalV[2][1][1] = valE21_y - valV0_y;
      pppvalV[3][0][0] = valE30_x - valV0_x;
      pppvalV[3][0][1] = valE30_y - valV0_y;
      pppvalV[3][1][0] = valE31_x - valV0_x;
      pppvalV[3][1][1] = valE31_y - valV0_y;
  
      Interval ppvalE[4][2];     // Edge vectors
      //memset(ppvalE, 0, sizeof(ppvalE));
      ppvalE[1][0] = valE11_x - valE10_x;
      ppvalE[1][1] = valE11_y - valE10_y;
      ppvalE[2][0] = valE21_x - valE20_x;
      ppvalE[2][1] = valE21_y - valE20_y;
      ppvalE[3][0] = valE31_x - valE30_x;
      ppvalE[3][1] = valE31_y - valE30_y;
  
      Interval valE1xE2;
      Cross(valE1xE2, ppvalE[1][0], ppvalE[1][1], ppvalE[2][0], ppvalE[2][1]);

      Interval valPlusOrMinus;
      valPlusOrMinus = 1.0;
      iE1xE2 = ( valE1xE2 > 0.0 );
      if ( iE1xE2 == -1 )
	{
	  break;
	}
      ASSERT( valE1xE2 != 0.0 );
      if ( iE1xE2 == 1 )
	{
	  valPlusOrMinus = -1.0;
	}
  
      Interval valE10xE11;
      Cross(valE10xE11, pppvalV[1][0][0], pppvalV[1][0][1], pppvalV[1][1][0], pppvalV[1][1][1]);
      Interval valE20xE21;
      Cross(valE20xE21, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][1][0], pppvalV[2][1][1]);

      Interval valE1dotE1;
      Dot(valE1dotE1, ppvalE[1][0], ppvalE[1][1], ppvalE[1][0], ppvalE[1][1]);
      Interval valE2dotE2;
      Dot(valE2dotE2, ppvalE[2][0], ppvalE[2][1], ppvalE[2][0], ppvalE[2][1]);
      Interval valE1dotE2;
      Dot(valE1dotE2, ppvalE[1][0], ppvalE[1][1], ppvalE[2][0], ppvalE[2][1]);

      Interval valE1E2plusE1dotE2;
      valE1E2plusE1dotE2 = ( sqrt(valE1dotE1*valE2dotE2) + valE1dotE2 );

      Interval valDisc;
      valDisc = 2.0*valE10xE11*valE20xE21*valE1E2plusE1dotE2;
      iDisc = ( valDisc >= 0.0 ); 
      if ( iDisc == -1 )
	{
	  break;
	}
      ASSERT( iDisc == 1 ); 
      Interval valSqrtDisc = sqrt(valDisc);
  
      Interval valE3dotE3;
      Dot(valE3dotE3, ppvalE[3][0], ppvalE[3][1], ppvalE[3][0], ppvalE[3][1]);
  
      Interval valE2xE3;
      Cross(valE2xE3, ppvalE[2][0], ppvalE[2][1], ppvalE[3][0], ppvalE[3][1]);
      Interval valE1xE3;
      Cross(valE1xE3, ppvalE[1][0], ppvalE[1][1], ppvalE[3][0], ppvalE[3][1]);
  
      Interval valE30xE31;
      Cross(valE30xE31, pppvalV[3][0][0], pppvalV[3][0][1], pppvalV[3][1][0], pppvalV[3][1][1]);
  
      Interval valE1E2minusE1dotE2;
      valE1E2minusE1dotE2 = ( sqrt(valE1dotE1*valE2dotE2) - valE1dotE2 );
  
      Interval valA;
      valA = ( ( valE1dotE2*valE1xE3 - valE1dotE1*valE2xE3 )*valE20xE21
	     + ( valE2dotE2*valE1xE3 - valE1dotE2*valE2xE3 )*valE10xE11
	     + valPlusOrMinus*( sqrt(valE2dotE2)*valE1xE3 - sqrt(valE1dotE1)*valE2xE3 )*valSqrtDisc );
  
      Interval valB;
      valB = ( ( sqrt(valE1dotE1)*valE20xE21 + sqrt(valE2dotE2)*valE10xE11 )*( sqrt(valE1dotE1)*valE20xE21 + sqrt(valE2dotE2)*valE10xE11 )
	     + 2.0*valE10xE11*valE20xE21*valE1E2plusE1dotE2
	     + valPlusOrMinus*2.0*( sqrt(valE1dotE1)*valE20xE21 + sqrt(valE2dotE2)*valE10xE11 )*valSqrtDisc );
  
      Interval valResult;  
      valResult = ( valA*valA
		  + 2.0*valE30xE31*valE1xE2*valE1E2minusE1dotE2*valA
		  + valE30xE31*valE30xE31*valE1xE2*valE1xE2*valE1E2minusE1dotE2*valE1E2minusE1dotE2
		  - valE3dotE3*valE1xE2*valE1xE2*valB );

      iResult = ( valResult < 0.0 );
      if ( iResult == -1 )
	{
	  break;
	}
      if ( iResult == 1 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( valResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
      return APM2D::cttOnCircle;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( iResult == -1 )
    {
      APM2D::m_uiInCircleVEE_EMPFUN++;

      mp_real rV0_x = dV0_x;
      mp_real rV0_y = dV0_y;
      mp_real rE10_x = dE10_x;
      mp_real rE10_y = dE10_y; 
      mp_real rE11_x = dE11_x;
      mp_real rE11_y = dE11_y;
      mp_real rE20_x = dE20_x;
      mp_real rE20_y = dE20_y; 
      mp_real rE21_x = dE21_x;
      mp_real rE21_y = dE21_y;
      mp_real rE30_x = dE30_x;
      mp_real rE30_y = dE30_y;
      mp_real rE31_x = dE31_x;
      mp_real rE31_y = dE31_y;

      mp_real ppprV[4][2][2]; // Endpoints
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = 0.0;
      ppprV[0][0][1] = 0.0;
      ppprV[1][0][0] = rE10_x - rV0_x;
      ppprV[1][0][1] = rE10_y - rV0_y;
      ppprV[1][1][0] = rE11_x - rV0_x;
      ppprV[1][1][1] = rE11_y - rV0_y;
      ppprV[2][0][0] = rE20_x - rV0_x;
      ppprV[2][0][1] = rE20_y - rV0_y;
      ppprV[2][1][0] = rE21_x - rV0_x;
      ppprV[2][1][1] = rE21_y - rV0_y;
      ppprV[3][0][0] = rE30_x - rV0_x;
      ppprV[3][0][1] = rE30_y - rV0_y;
      ppprV[3][1][0] = rE31_x - rV0_x;
      ppprV[3][1][1] = rE31_y - rV0_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[1][0] = rE11_x - rE10_x;
      pprE[1][1] = rE11_y - rE10_y;
      pprE[2][0] = rE21_x - rE20_x;
      pprE[2][1] = rE21_y - rE20_y;
      pprE[3][0] = rE31_x - rE30_x;
      pprE[3][1] = rE31_y - rE30_y;
  
      mp_real rE1xE2;
      Cross(rE1xE2, pprE[1][0], pprE[1][1], pprE[2][0], pprE[2][1]);

      mp_real rPlusOrMinus;
      rPlusOrMinus = 1.0;
      ASSERT( rE1xE2 != 0.0 );
      if ( rE1xE2 > 0.0 )
	{
	  rPlusOrMinus = -1.0;
	}
  
      mp_real rE10xE11;
      Cross(rE10xE11, ppprV[1][0][0], ppprV[1][0][1], ppprV[1][1][0], ppprV[1][1][1]);
      mp_real rE20xE21;
      Cross(rE20xE21, ppprV[2][0][0], ppprV[2][0][1], ppprV[2][1][0], ppprV[2][1][1]);

      mp_real rE1dotE1;
      Dot(rE1dotE1, pprE[1][0], pprE[1][1], pprE[1][0], pprE[1][1]);
      mp_real rE2dotE2;
      Dot(rE2dotE2, pprE[2][0], pprE[2][1], pprE[2][0], pprE[2][1]);
      mp_real rE1dotE2;
      Dot(rE1dotE2, pprE[1][0], pprE[1][1], pprE[2][0], pprE[2][1]);

      mp_real rE1E2plusE1dotE2;
      rE1E2plusE1dotE2 = ( sqrt(rE1dotE1*rE2dotE2) + rE1dotE2 );

      mp_real rDisc;
      rDisc = 2.0*rE10xE11*rE20xE21*rE1E2plusE1dotE2;
      ASSERT( rDisc >= 0.0 ); 
      mp_real rSqrtDisc = sqrt(rDisc);
  
      mp_real rE3dotE3;
      Dot(rE3dotE3, pprE[3][0], pprE[3][1], pprE[3][0], pprE[3][1]);
  
      mp_real rE2xE3;
      Cross(rE2xE3, pprE[2][0], pprE[2][1], pprE[3][0], pprE[3][1]);
      mp_real rE1xE3;
      Cross(rE1xE3, pprE[1][0], pprE[1][1], pprE[3][0], pprE[3][1]);
  
      mp_real rE30xE31;
      Cross(rE30xE31, ppprV[3][0][0], ppprV[3][0][1], ppprV[3][1][0], ppprV[3][1][1]);
  
      mp_real rE1E2minusE1dotE2;
      rE1E2minusE1dotE2 = ( sqrt(rE1dotE1*rE2dotE2) - rE1dotE2 );
  
      mp_real rA;
      rA = ( ( rE1dotE2*rE1xE3 - rE1dotE1*rE2xE3 )*rE20xE21
	     + ( rE2dotE2*rE1xE3 - rE1dotE2*rE2xE3 )*rE10xE11
	     + rPlusOrMinus*( sqrt(rE2dotE2)*rE1xE3 - sqrt(rE1dotE1)*rE2xE3 )*rSqrtDisc );
  
      mp_real rB;
      rB = ( ( sqrt(rE1dotE1)*rE20xE21 + sqrt(rE2dotE2)*rE10xE11 )*( sqrt(rE1dotE1)*rE20xE21 + sqrt(rE2dotE2)*rE10xE11 )
	     + 2.0*rE10xE11*rE20xE21*rE1E2plusE1dotE2
	     + rPlusOrMinus*2.0*( sqrt(rE1dotE1)*rE20xE21 + sqrt(rE2dotE2)*rE10xE11 )*rSqrtDisc );
  
      mp_real rResult;  
      rResult = ( rA*rA
		  + 2.0*rE30xE31*rE1xE2*rE1E2minusE1dotE2*rA
		  + rE30xE31*rE30xE31*rE1xE2*rE1xE2*rE1E2minusE1dotE2*rE1E2minusE1dotE2
		  - rE3dotE3*rE1xE2*rE1xE2*rB );

      if ( rResult < 0.0 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( rResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
    }

  return APM2D::cttOnCircle;
}

//////////////////////////////////////////////////
// VEE Antiparallel

class ICC_VEEAntiparallel_MPFUNFilter
{
protected:
  enum InstructionPointer {ipStart, ipV0onE1orE2, 
			   ipColinear, ipDen, ipDisc, 
			   ipEdgeZones, ipVertexZones, ipEnd};

public:
  ICC_VEEAntiparallel_MPFUNFilter();
  ~ICC_VEEAntiparallel_MPFUNFilter();

public:
  APM2D::InscribedCircleCenterType 
  InscribedCircleCenterVEEAntiparallel(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			   DOUBLE dV00_x, DOUBLE dV00_y,
			   DOUBLE dV0_x, DOUBLE dV0_y,
			   DOUBLE dV01_x, DOUBLE dV01_y,
			   DOUBLE dE10_x, DOUBLE dE10_y, 
			   DOUBLE dE11_x, DOUBLE dE11_y,
			   DOUBLE dE20_x, DOUBLE dE20_y, 
			   DOUBLE dE21_x, DOUBLE dE21_y);
  APM2D::InscribedCircleCenterType 
  InscribedCircleCenterVEEAntiparallelForce(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dV00_x, DOUBLE dV00_y,
				DOUBLE dV0_x, DOUBLE dV0_y,
				DOUBLE dV01_x, DOUBLE dV01_y,
				DOUBLE dE10_x, DOUBLE dE10_y, 
				DOUBLE dE11_x, DOUBLE dE11_y,
				DOUBLE dE20_x, DOUBLE dE20_y, 
				DOUBLE dE21_x, DOUBLE dE21_y);
  
protected:
#ifndef COMPILER_PROBLEM
  BOOL
  InscribedCircleCenterVEEAntiparallelMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					    DOUBLE dV00_x, DOUBLE dV00_y,
					    DOUBLE dV0_x, DOUBLE dV0_y,
					    DOUBLE dV01_x, DOUBLE dV01_y,
					    DOUBLE dE10_x, DOUBLE dE10_y, 
					    DOUBLE dE11_x, DOUBLE dE11_y,
					    DOUBLE dE20_x, DOUBLE dE20_y, 
					    DOUBLE dE21_x, DOUBLE dE21_y,
					    InstructionPointer eIPMax = ipEnd);
#else // COMPILER_PROBLEM
  BOOL
  InscribedCircleCenterVEEAntiparallelMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					    DOUBLE dV00_x, DOUBLE dV00_y,
					    DOUBLE dV0_x, DOUBLE dV0_y,
					    DOUBLE dV01_x, DOUBLE dV01_y,
					    DOUBLE dE10_x, DOUBLE dE10_y, 
					    DOUBLE dE11_x, DOUBLE dE11_y,
					    DOUBLE dE20_x, DOUBLE dE20_y, 
					    DOUBLE dE21_x, DOUBLE dE21_y,
					    InstructionPointer eIPMax = ipEnd);
  BOOL Colinear(InstructionPointer eIPMax);
  BOOL Den(InstructionPointer eIPMax);
  BOOL Disc(InstructionPointer eIPMax);
  BOOL EdgeZones(InstructionPointer eIPMax);
  BOOL VertexZones(InstructionPointer eIPMax);
  /*
  VOID CenterAndRadius(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
		       InstructionPointer eIPMax);
  */
#endif // COMPILER_PROBLEM

protected:
  InstructionPointer m_eIP;
  APM2D::InscribedCircleCenterType m_eResult;
  INT_32 m_iV0onE1, m_iV0onE2;
  INT_32 m_iE1E2minusE1dotE2;
  INT_32 m_iDen, m_iDenR;
  INT_32 m_iDisc, m_iDiscE2;
  INT_32 m_ppiEdgeZone[2][2];
  INT_32 m_piVertexZone[2];

  Interval m_valE1dotE1;
  Interval m_valDen;
  Interval m_valDenR;
  Interval m_valDisc;
  Interval m_valDiscE2;

  mp_real m_rV00_x;
  mp_real m_rV00_y;
  mp_real m_rV0_x;
  mp_real m_rV0_y;
  mp_real m_rV01_x;
  mp_real m_rV01_y;
  mp_real m_rE10_x;
  mp_real m_rE10_y;
  mp_real m_rE11_x;
  mp_real m_rE11_y;
  mp_real m_rE20_x;
  mp_real m_rE20_y;
  mp_real m_rE21_x;
  mp_real m_rE21_y;
  mp_real m_pprE[3][2];
  mp_real m_rE1dotE1;
  mp_real m_rE2dotE2;
  mp_real m_rE1dotE2;
  mp_real m_rDen;
  mp_real m_rDenR;
  mp_real m_ppprV[3][2][2];
  mp_real m_rE10xE11;
  mp_real m_rE1xE20;
  mp_real m_rSqrtDisc;
  mp_real m_rSqrtDiscE2;

};

ICC_VEEAntiparallel_MPFUNFilter::ICC_VEEAntiparallel_MPFUNFilter()
{
}

ICC_VEEAntiparallel_MPFUNFilter::~ICC_VEEAntiparallel_MPFUNFilter()
{
}

APM2D::InscribedCircleCenterType 
ICC_VEEAntiparallel_MPFUNFilter::InscribedCircleCenterVEEAntiparallel(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					      DOUBLE dV00_x, DOUBLE dV00_y,
					      DOUBLE dV0_x, DOUBLE dV0_y,
					      DOUBLE dV01_x, DOUBLE dV01_y,
					      DOUBLE dE10_x, DOUBLE dE10_y, 
					      DOUBLE dE11_x, DOUBLE dE11_y,
					      DOUBLE dE20_x, DOUBLE dE20_y, 
					      DOUBLE dE21_x, DOUBLE dE21_y)
{
  APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUNFilter++;

  *pdC_x = 0.0;
  *pdC_y = 0.0;
  *pdR = 0.0;

  m_eIP = ipStart;
  m_eResult = APM2D::icctNull;

  m_iV0onE1 = -1;
  m_iV0onE2 = -1;
  m_iE1E2minusE1dotE2 = -1;
  m_iDen = -1;
  m_iDenR = -1;
  m_iDisc = -1;
  m_iDiscE2 = -1;
  m_ppiEdgeZone[0][0] = -1;
  m_ppiEdgeZone[0][1] = -1;
  m_ppiEdgeZone[1][0] = -1;
  m_ppiEdgeZone[1][1] = -1;
  m_piVertexZone[0] = -1;
  m_piVertexZone[1] = -1;

  BOOL bDone;

  //////////////////////////////////////////////////
  // Check V0 on E1 or E2

  Interval valV00_x = dV00_x;
  Interval valV00_y = dV00_y;
  Interval valV0_x = dV0_x;
  Interval valV0_y = dV0_y;
  Interval valV01_x = dV01_x;
  Interval valV01_y = dV01_y;
  Interval valE10_x = dE10_x;
  Interval valE10_y = dE10_y; 
  Interval valE11_x = dE11_x;
  Interval valE11_y = dE11_y;
  Interval valE20_x = dE20_x;
  Interval valE20_y = dE20_y; 
  Interval valE21_x = dE21_x;
  Interval valE21_y = dE21_y;
  
  Interval valV0onE1;
  SinTheta(valV0onE1, valV0_x, valV0_y, valE10_x, valE10_y, valE11_x, valE11_y);
  m_iV0onE1 = ( valV0onE1 <= 0.0 );
  if ( m_iV0onE1 == 1 )
    {
      // V0 is on E1 and it is not one of the endpoints
      return APM2D::icctNull;
    }
  
  Interval valV0onE2;
  SinTheta(valV0onE2, valV0_x, valV0_y, valE20_x, valE20_y, valE21_x, valE21_y);
  m_iV0onE2 = ( valV0onE2 <= 0.0 );
  if ( m_iV0onE2 == 1 )
    {
      // V0 is on E2 and it is not one of the endpoints
      return APM2D::icctNull;
    }

  if ( m_iV0onE1 == -1 ||
       m_iV0onE2 == -1 )
    {
      bDone = InscribedCircleCenterVEEAntiparallelMPFUN(pdC_x, pdC_y, pdR, 
							dV00_x, dV00_y,
							dV0_x, dV0_y,
							dV01_x, dV01_y,
							dE10_x, dE10_y, 
							dE11_x, dE11_y,
							dE20_x, dE20_y, 
							dE21_x, dE21_y,
							ipV0onE1orE2);
      if ( bDone )
	{
	  return m_eResult;
	}
    }
  
  //////////////////////////////////////////////////
  // Check Colinear

  Interval ppvalE[3][2];     // Edge vectors
  //memset(ppvalE, 0, sizeof(ppvalE));
  ppvalE[1][0] = valE11_x - valE10_x;
  ppvalE[1][1] = valE11_y - valE10_y;
  ppvalE[2][0] = valE21_x - valE20_x;
  ppvalE[2][1] = valE21_y - valE20_y;
  
  Dot(m_valE1dotE1, ppvalE[1][0], ppvalE[1][1], ppvalE[1][0], ppvalE[1][1]);
  Interval valE2dotE2;
  Dot(valE2dotE2, ppvalE[2][0], ppvalE[2][1], ppvalE[2][0], ppvalE[2][1]);
  Interval valE1dotE2;
  Dot(valE1dotE2, ppvalE[1][0], ppvalE[1][1], ppvalE[2][0], ppvalE[2][1]);

  Interval valE1E2minusE1dotE2;  
  valE1E2minusE1dotE2 = sqrt(m_valE1dotE1*valE2dotE2) - valE1dotE2;
  m_iE1E2minusE1dotE2 = ( valE1E2minusE1dotE2 == 0.0 );
  if ( m_iE1E2minusE1dotE2 == 1 )
    {
      // Colinear
      return APM2D::icctNull;
    }
  if ( m_iE1E2minusE1dotE2 == -1 )
    {
      bDone = InscribedCircleCenterVEEAntiparallelMPFUN(pdC_x, pdC_y, pdR, 
							dV00_x, dV00_y,
							dV0_x, dV0_y,
							dV01_x, dV01_y,
							dE10_x, dE10_y, 
							dE11_x, dE11_y,
							dE20_x, dE20_y, 
							dE21_x, dE21_y,
							ipColinear);
      if ( bDone )
	{
	  return m_eResult;
	}
    }

  //////////////////////////////////////////////////
  // Check Den

  m_valDen = 2*m_valE1dotE1;
  m_iDen = ( m_valDen != 0.0 );

  m_valDenR = 2*sqrt(m_valE1dotE1);
  m_iDenR = ( m_valDenR != 0.0 );
  if ( m_iDen == -1 ||
       m_iDenR == -1 )
    {
      bDone = InscribedCircleCenterVEEAntiparallelMPFUN(pdC_x, pdC_y, pdR, 
							dV00_x, dV00_y,
							dV0_x, dV0_y,
							dV01_x, dV01_y,
							dE10_x, dE10_y, 
							dE11_x, dE11_y,
							dE20_x, dE20_y, 
							dE21_x, dE21_y,
							ipDen);
      if ( bDone )
	{
	  return m_eResult;
	}
      if ( m_iDen == -1 ||
	   m_iDenR == -1 )
	{
	  bDone = InscribedCircleCenterVEEAntiparallelMPFUN(pdC_x, pdC_y, pdR, 
							    dV00_x, dV00_y,
							    dV0_x, dV0_y,
							    dV01_x, dV01_y,
							    dE10_x, dE10_y, 
							    dE11_x, dE11_y,
							    dE20_x, dE20_y, 
							    dE21_x, dE21_y,
							    ipEnd);
	  ASSERT( bDone );
	  return m_eResult;
	}
    }
  ASSERT( m_iDen == 1 );
  ASSERT( m_iDenR == 1 );

  //////////////////////////////////////////////////
  // Check Disc

  Interval pppvalV[3][2][2]; // Endpoints
  // Translate V1 to the origin
  //memset(pppvalV, 0, sizeof(pppvalV));
  pppvalV[0][0][0] = 0.0;
  pppvalV[0][0][1] = 0.0;
  pppvalV[1][0][0] = valE10_x - valV0_x;
  pppvalV[1][0][1] = valE10_y - valV0_y;
  pppvalV[1][1][0] = valE11_x - valV0_x;
  pppvalV[1][1][1] = valE11_y - valV0_y;
  pppvalV[2][0][0] = valE20_x - valV0_x;
  pppvalV[2][0][1] = valE20_y - valV0_y;
  pppvalV[2][1][0] = valE21_x - valV0_x;
  pppvalV[2][1][1] = valE21_y - valV0_y;
  
  Interval valE10xE11;
  Cross(valE10xE11, pppvalV[1][0][0], pppvalV[1][0][1], pppvalV[1][1][0], pppvalV[1][1][1]);
  Interval valE20xE21;
  Cross(valE20xE21, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][1][0], pppvalV[2][1][1]);
  Interval valE1xE20;
  Cross(valE1xE20, ppvalE[1][0], ppvalE[1][1], pppvalV[2][0][0], pppvalV[2][0][1]);
  Interval valE2xE10;
  Cross(valE2xE10, ppvalE[2][0], ppvalE[2][1], pppvalV[1][0][0], pppvalV[1][0][1]);

  m_valDisc = valE1xE20*valE10xE11;
  m_iDisc = ( m_valDisc < 0.0 );
  if ( m_iDisc == 1 ) 
    {
      // Imaginary Solutions
      return APM2D::icctNull;
    }
  
  m_valDiscE2 = valE2xE10*valE20xE21;
  m_iDiscE2 = ( m_valDiscE2 < 0.0 ) ;
  if ( m_iDiscE2 == 1 ) 
    {
      // Imaginary Solutions
      return APM2D::icctNull;
    }

  if ( m_iDisc == -1 ||
       m_iDiscE2 == -1 )
    {
      bDone = InscribedCircleCenterVEEAntiparallelMPFUN(pdC_x, pdC_y, pdR, 
							dV00_x, dV00_y,
							dV0_x, dV0_y,
							dV01_x, dV01_y,
							dE10_x, dE10_y, 
							dE11_x, dE11_y,
							dE20_x, dE20_y, 
							dE21_x, dE21_y,
							ipDisc);
      if ( bDone )
	{
	  return m_eResult;
	}
      if ( m_iDisc == -1 ||
	   m_iDiscE2 == -1 )
	{
	  bDone = InscribedCircleCenterVEEAntiparallelMPFUN(pdC_x, pdC_y, pdR, 
							    dV00_x, dV00_y,
							    dV0_x, dV0_y,
							    dV01_x, dV01_y,
							    dE10_x, dE10_y, 
							    dE11_x, dE11_y,
							    dE20_x, dE20_y, 
							    dE21_x, dE21_y,
							    ipEnd);
	  ASSERT( bDone );
	  return m_eResult;
	}
    }
  Interval valSqrtDisc = sqrt(m_valDisc);
  Interval valSqrtDiscE2 = sqrt(m_valDiscE2);

  //////////////////////////////////////////////////
  // Check the Edge Zone

  Interval ppvalEdgeZone[2][2];

  Interval valE1dotE10;
  Dot(valE1dotE10, ppvalE[1][0], ppvalE[1][1], pppvalV[1][0][0], pppvalV[1][0][1]);
  ppvalEdgeZone[0][0] = -( valE1dotE10 - valSqrtDisc );
  m_ppiEdgeZone[0][0] = ( ppvalEdgeZone[0][0] < 0.0 );
  if ( m_ppiEdgeZone[0][0] == 1 )
    {
      return APM2D::icctNull;      
    }

  Interval valE1dotE11;
  Dot(valE1dotE11, ppvalE[1][0], ppvalE[1][1], pppvalV[1][1][0], pppvalV[1][1][1]);
  ppvalEdgeZone[0][1] = ( valE1dotE11 - valSqrtDisc );
  m_ppiEdgeZone[0][1] = ( ppvalEdgeZone[0][1] < 0.0 );
  if ( m_ppiEdgeZone[0][1] == 1 )
    {
      return APM2D::icctNull;      
    }

  Interval valE2dotE20;
  Dot(valE2dotE20, ppvalE[2][0], ppvalE[2][1], pppvalV[2][0][0], pppvalV[2][0][1]);
  ppvalEdgeZone[1][0] = -( valE2dotE20 + valSqrtDiscE2 );
  m_ppiEdgeZone[1][0] = ( ppvalEdgeZone[1][0] < 0.0 );
  if ( m_ppiEdgeZone[1][0] == 1 )
    {
      return APM2D::icctNull;      
    }

  Interval valE2dotE21;
  Dot(valE2dotE21, ppvalE[2][0], ppvalE[2][1], pppvalV[2][1][0], pppvalV[2][1][1]);
  ppvalEdgeZone[1][1] = ( valE2dotE21 + valSqrtDiscE2 );
  m_ppiEdgeZone[1][1] = ( ppvalEdgeZone[1][1] < 0.0 );
  if ( m_ppiEdgeZone[1][1] == 1 )
    {
      return APM2D::icctNull;      
    }

  if ( m_ppiEdgeZone[0][0] == -1 ||
       m_ppiEdgeZone[0][1] == -1 ||
       m_ppiEdgeZone[1][0] == -1 ||
       m_ppiEdgeZone[1][1] == -1 )
    {
      bDone = InscribedCircleCenterVEEAntiparallelMPFUN(pdC_x, pdC_y, pdR, 
							dV00_x, dV00_y,
							dV0_x, dV0_y,
							dV01_x, dV01_y,
							dE10_x, dE10_y, 
							dE11_x, dE11_y,
							dE20_x, dE20_y, 
							dE21_x, dE21_y,
							ipEdgeZones);
      if ( bDone )
	{
	  return m_eResult;
	}
    }

  //////////////////////////////////////////////////
  // Check the Vertex Zone

  Interval pppvalVN[1][2][2]; // Endpoints
  // Vertex neighbor vectors
  //memset(pppvalVN, 0, sizeof(pppvalV));
  pppvalVN[0][0][0] = valV0_x - valV00_x;
  pppvalVN[0][0][1] = valV0_y - valV00_y;
  pppvalVN[0][1][0] = valV0_x - valV01_x;
  pppvalVN[0][1][1] = valV0_y - valV01_y;

  Interval valE1xE20minusE10xE11 = valE1xE20 - valE10xE11;

  Interval pvalVertexZone[2];

  Interval valE1dotV00;
  Dot(valE1dotV00, ppvalE[1][0], ppvalE[1][1], pppvalVN[0][0][0], pppvalVN[0][0][1]);
  Interval valE1xV00;
  Cross(valE1xV00, ppvalE[1][0], ppvalE[1][1], pppvalVN[0][0][0], pppvalVN[0][0][1]);
  pvalVertexZone[0] = ( ( valE1xE20minusE10xE11 )*valE1xV00
		      + 2.0*valE1dotV00*valSqrtDisc );
  m_piVertexZone[0] = ( pvalVertexZone[0] < 0.0 );
  if ( m_piVertexZone[0] == 1 )
    {
      return APM2D::icctNull;
    }

  Interval valE1dotV01;
  Dot(valE1dotV01, ppvalE[1][0], ppvalE[1][1], pppvalVN[0][1][0], pppvalVN[0][1][1]);
  Interval valE1xV01;
  Cross(valE1xV01, ppvalE[1][0], ppvalE[1][1], pppvalVN[0][1][0], pppvalVN[0][1][1]);
  pvalVertexZone[1] = ( ( valE1xE20minusE10xE11 )*valE1xV01
			+ 2.0*valE1dotV01*valSqrtDisc );
  m_piVertexZone[1] = ( pvalVertexZone[1] < 0.0 );
  if ( m_piVertexZone[1] == 1 )
    {
      return APM2D::icctNull;
    }

  if ( m_piVertexZone[0] == -1 ||
       m_piVertexZone[1] == -1 )
    {
      bDone = InscribedCircleCenterVEEAntiparallelMPFUN(pdC_x, pdC_y, pdR, 
							dV00_x, dV00_y,
							dV0_x, dV0_y,
							dV01_x, dV01_y,
							dE10_x, dE10_y, 
							dE11_x, dE11_y,
							dE20_x, dE20_y, 
							dE21_x, dE21_y,
							ipVertexZones);
      if ( bDone )
	{
	  return m_eResult;
	}
    }

  //////////////////////////////////////////////////
  // Compute the center and radius

  Interval pvalC[2];
  pvalC[0] = ( -(valE1xE20 - valE10xE11)*ppvalE[1][1] + 2.0*ppvalE[1][0]*valSqrtDisc ) / m_valDen;
  pvalC[1] = ( (valE1xE20 - valE10xE11)*ppvalE[1][0] + 2.0*ppvalE[1][1]*valSqrtDisc ) / m_valDen;

  Interval valR;
  valR = ( valE1xE20 + valE10xE11 ) / m_valDenR;
  
  // Translate the answer back out away from the origin
  pvalC[0] += valV0_x;
  pvalC[1] += valV0_y;

  *pdC_x = dble(pvalC[0]);
  *pdC_y = dble(pvalC[1]);
  *pdR = dble(valR);
  return APM2D::icctVEEAntiparallel;
}

APM2D::InscribedCircleCenterType 
ICC_VEEAntiparallel_MPFUNFilter::InscribedCircleCenterVEEAntiparallelForce(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					      DOUBLE dV00_x, DOUBLE dV00_y,
					      DOUBLE dV0_x, DOUBLE dV0_y,
					      DOUBLE dV01_x, DOUBLE dV01_y,
					      DOUBLE dE10_x, DOUBLE dE10_y, 
					      DOUBLE dE11_x, DOUBLE dE11_y,
					      DOUBLE dE20_x, DOUBLE dE20_y, 
					      DOUBLE dE21_x, DOUBLE dE21_y)
{
  *pdC_x = 0.0;
  *pdC_y = 0.0;
  *pdR = 0.0;

  m_eIP = ipStart;
  m_eResult = APM2D::icctNull;

  m_iV0onE1 = -1;
  m_iV0onE2 = -1;
  m_iE1E2minusE1dotE2 = -1;
  m_iDen = -1;
  m_iDenR = -1;
  m_iDisc = -1;
  m_iDiscE2 = -1;
  m_ppiEdgeZone[0][0] = -1;
  m_ppiEdgeZone[0][1] = -1;
  m_ppiEdgeZone[1][0] = -1;
  m_ppiEdgeZone[1][1] = -1;
  m_piVertexZone[0] = -1;
  m_piVertexZone[1] = -1;

  BOOL bDone;
  bDone = InscribedCircleCenterVEEAntiparallelMPFUN(pdC_x, pdC_y, pdR, 
						    dV00_x, dV00_y,
						    dV0_x, dV0_y,
						    dV01_x, dV01_y,
						    dE10_x, dE10_y, 
						    dE11_x, dE11_y,
						    dE20_x, dE20_y, 
						    dE21_x, dE21_y,
						    ipEnd);
  ASSERT( bDone );
  return m_eResult;
}

#ifndef COMPILER_PROBLEM
BOOL 
ICC_VEEAntiparallel_MPFUNFilter::InscribedCircleCenterVEEAntiparallelMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					      DOUBLE dV00_x, DOUBLE dV00_y,
					      DOUBLE dV0_x, DOUBLE dV0_y,
					      DOUBLE dV01_x, DOUBLE dV01_y,
					      DOUBLE dE10_x, DOUBLE dE10_y, 
					      DOUBLE dE11_x, DOUBLE dE11_y,
					      DOUBLE dE20_x, DOUBLE dE20_y, 
					      DOUBLE dE21_x, DOUBLE dE21_y,
					      InstructionPointer eIPMax/* = ipEnd*/)
{
  APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUN++;

  switch ( m_eIP )
    {
    case ipStart:
      {
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	*pdC_x = 0.0;
	*pdC_y = 0.0;
	*pdR = 0.0;

	//////////////////////////////////////////////////
	// Check V0 on E1 or E2

	m_rV00_x = dV00_x;
	m_rV00_y = dV00_y;
	m_rV0_x = dV0_x;
	m_rV0_y = dV0_y;
	m_rV01_x = dV01_x;
	m_rV01_y = dV01_y;
	m_rE10_x = dE10_x;
	m_rE10_y = dE10_y; 
	m_rE11_x = dE11_x;
	m_rE11_y = dE11_y;
	m_rE20_x = dE20_x;
	m_rE20_y = dE20_y; 
	m_rE21_x = dE21_x;
	m_rE21_y = dE21_y;
  
	if ( m_iV0onE1 == -1 )
	  {
	    mp_real rV0onE1;
	    SinTheta(rV0onE1, m_rV0_x, m_rV0_y, m_rE10_x, m_rE10_y, m_rE11_x, m_rE11_y);
	    if ( rV0onE1 <= 0.0 )
	      {
		// V0 is on E1 and it is not one of the endpoints
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUNFinal++;
		return TRUE;
	      }
  	  }
  
	if ( m_iV0onE2 == -1 )
	  {
	    mp_real rV0onE2;
	    SinTheta(rV0onE2, m_rV0_x, m_rV0_y, m_rE20_x, m_rE20_y, m_rE21_x, m_rE21_y);
	    if ( rV0onE2 <= 0.0 )
	      {
		// V0 is on E2 and it is not one of the endpoints
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUNFinal++;
		return TRUE;
	      }
	  }
  
      }
    case ipV0onE1orE2:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Check Colinear

	//memset(m_pprE, 0, sizeof(m_pprE));
	m_pprE[1][0] = m_rE11_x - m_rE10_x;
	m_pprE[1][1] = m_rE11_y - m_rE10_y;
	m_pprE[2][0] = m_rE21_x - m_rE20_x;
	m_pprE[2][1] = m_rE21_y - m_rE20_y;
  
	Dot(m_rE1dotE1, m_pprE[1][0], m_pprE[1][1], m_pprE[1][0], m_pprE[1][1]);
	Dot(m_rE2dotE2, m_pprE[2][0], m_pprE[2][1], m_pprE[2][0], m_pprE[2][1]);
	Dot(m_rE1dotE2, m_pprE[1][0], m_pprE[1][1], m_pprE[2][0], m_pprE[2][1]);

	if ( m_iE1E2minusE1dotE2 == -1 )
	  {
	    mp_real rE1E2minusE1dotE2;  
	    rE1E2minusE1dotE2 = sqrt(m_rE1dotE1*m_rE2dotE2) - m_rE1dotE2;
	    if ( rE1E2minusE1dotE2 == 0.0 )
	      {
		// Colinear
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUNFinal++;
		return TRUE;
	      }
	  }

      }
    case ipColinear:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Check Den

	m_rDen = 2*m_rE1dotE1;
	if ( m_iDen == -1 )
	  {
	    ASSERT( m_rDen != 0.0 );
	    m_valE1dotE1.ConservativeBound(dble(m_rE1dotE1));
	    m_valDen.ConservativeBound(dble(m_rDen));
	    m_iDen = ( m_valDen != 0.0 );
	  }

	m_rDenR = 2*sqrt(m_rE1dotE1);
	if ( m_iDenR == -1 )
	  {
	    ASSERT( m_rDenR != 0.0 );
	    m_valDenR.ConservativeBound(dble(m_rDenR));
	    m_iDenR = ( m_valDenR != 0.0 );
	  }

      }
    case ipDen:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Check Disc

	// Translate V1 to the origin
	//memset(m_ppprV, 0, sizeof(m_ppprV));
	m_ppprV[0][0][0] = 0.0;
	m_ppprV[0][0][1] = 0.0;
	m_ppprV[1][0][0] = m_rE10_x - m_rV0_x;
	m_ppprV[1][0][1] = m_rE10_y - m_rV0_y;
	m_ppprV[1][1][0] = m_rE11_x - m_rV0_x;
	m_ppprV[1][1][1] = m_rE11_y - m_rV0_y;
	m_ppprV[2][0][0] = m_rE20_x - m_rV0_x;
	m_ppprV[2][0][1] = m_rE20_y - m_rV0_y;
	m_ppprV[2][1][0] = m_rE21_x - m_rV0_x;
	m_ppprV[2][1][1] = m_rE21_y - m_rV0_y;
  
	Cross(m_rE10xE11, m_ppprV[1][0][0], m_ppprV[1][0][1], m_ppprV[1][1][0], m_ppprV[1][1][1]);
	mp_real rE20xE21;
	Cross(rE20xE21, m_ppprV[2][0][0], m_ppprV[2][0][1], m_ppprV[2][1][0], m_ppprV[2][1][1]);
	Cross(m_rE1xE20, m_pprE[1][0], m_pprE[1][1], m_ppprV[2][0][0], m_ppprV[2][0][1]);
	mp_real rE2xE10;
	Cross(rE2xE10, m_pprE[2][0], m_pprE[2][1], m_ppprV[1][0][0], m_ppprV[1][0][1]);

	mp_real rDisc;
	rDisc = m_rE1xE20*m_rE10xE11;
	if ( m_iDisc == -1 )
	  {
	    if ( rDisc < 0.0 ) 
	      {
		// Imaginary Solutions
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUNFinal++;
		return TRUE;
	      }
	    m_valDisc.ConservativeBound(dble(rDisc));
	    m_iDisc = ( m_valDisc < 0.0 );	    
	  }
	m_rSqrtDisc = sqrt(rDisc);
  
	mp_real rDiscE2;
	rDiscE2 = rE2xE10*rE20xE21;
	if ( m_iDiscE2 == -1 )
	  {
	    if ( rDiscE2 < 0.0 ) 
	      {
		// Imaginary Solutions
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUNFinal++;
		return TRUE;
	      }
	    m_valDiscE2.ConservativeBound(dble(rDiscE2));
	    m_iDiscE2 = ( m_valDiscE2 < 0.0 );	    
	  }
	m_rSqrtDiscE2 = sqrt(rDiscE2);
  
      }
    case ipDisc:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Check the Edge Zones

	if ( m_ppiEdgeZone[0][0] == -1 ||
	     m_ppiEdgeZone[0][1] == -1 ||
	     m_ppiEdgeZone[1][0] == -1 ||
	     m_ppiEdgeZone[1][1] == -1 )
	  {
	    mp_real pprEdgeZone[2][2];

	    if ( m_ppiEdgeZone[0][0] == -1 )
	      {
		mp_real rE1dotE10;
		Dot(rE1dotE10, m_pprE[1][0], m_pprE[1][1], m_ppprV[1][0][0], m_ppprV[1][0][1]);
		pprEdgeZone[0][0] = -( rE1dotE10 - m_rSqrtDisc );
		if ( pprEdgeZone[0][0] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUNFinal++;
		    return TRUE;
		  }
	      }

	    if ( m_ppiEdgeZone[0][1] == -1 )
	      {
		mp_real rE1dotE11;
		Dot(rE1dotE11, m_pprE[1][0], m_pprE[1][1], m_ppprV[1][1][0], m_ppprV[1][1][1]);
		pprEdgeZone[0][1] = ( rE1dotE11 - m_rSqrtDisc );
		if ( pprEdgeZone[0][1] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUNFinal++;
		    return TRUE;
		  }
	      }

	    if ( m_ppiEdgeZone[1][0] == -1 )
	      {
		mp_real rE2dotE20;
		Dot(rE2dotE20, m_pprE[2][0], m_pprE[2][1], m_ppprV[2][0][0], m_ppprV[2][0][1]);
		pprEdgeZone[1][0] = -( rE2dotE20 + m_rSqrtDiscE2 );
		if ( pprEdgeZone[1][0] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUNFinal++;
		    return TRUE;
		  }
	      }

	    if ( m_ppiEdgeZone[1][1] == -1 )
	      {
		mp_real rE2dotE21;
		Dot(rE2dotE21, m_pprE[2][0], m_pprE[2][1], m_ppprV[2][1][0], m_ppprV[2][1][1]);
		pprEdgeZone[1][1] = ( rE2dotE21 + m_rSqrtDiscE2 );
		if ( pprEdgeZone[1][1] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUNFinal++;
		    return TRUE;
		  }
	      }
	  }

      }
    case ipEdgeZones:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Check the Vertex Zone

	mp_real ppprVN[1][2][2]; // Endpoints
	// Vertex neighbor vectors
	//memset(ppprVN, 0, sizeof(ppprV));
	ppprVN[0][0][0] = m_rV0_x - m_rV00_x;
	ppprVN[0][0][1] = m_rV0_y - m_rV00_y;
	ppprVN[0][1][0] = m_rV0_x - m_rV01_x;
	ppprVN[0][1][1] = m_rV0_y - m_rV01_y;

	if ( m_piVertexZone[0] == -1 ||
	     m_piVertexZone[1] == -1 )
	  {
	    mp_real rE1xE20minusE10xE11 = m_rE1xE20 - m_rE10xE11;

	    mp_real prVertexZone[2];

	    if ( m_piVertexZone[0] == -1 )
	      {
		mp_real rE1dotV00;
		Dot(rE1dotV00, m_pprE[1][0], m_pprE[1][1], ppprVN[0][0][0], ppprVN[0][0][1]);
		mp_real rE1xV00;
		Cross(rE1xV00, m_pprE[1][0], m_pprE[1][1], ppprVN[0][0][0], ppprVN[0][0][1]);
		prVertexZone[0] = ( ( rE1xE20minusE10xE11 )*rE1xV00
				    + 2.0*rE1dotV00*m_rSqrtDisc );
		if ( prVertexZone[0] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUNFinal++;
		    return TRUE;
		  }
	      }

	    if ( m_piVertexZone[1] == -1 )
	      {
		mp_real rE1dotV01;
		Dot(rE1dotV01, m_pprE[1][0], m_pprE[1][1], ppprVN[0][1][0], ppprVN[0][1][1]);
		mp_real rE1xV01;
		Cross(rE1xV01, m_pprE[1][0], m_pprE[1][1], ppprVN[0][1][0], ppprVN[0][1][1]);
		prVertexZone[1] = ( ( rE1xE20minusE10xE11 )*rE1xV01
				    + 2.0*rE1dotV01*m_rSqrtDisc );
		if ( prVertexZone[1] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUNFinal++;
		    return TRUE;
		  }
	      }
	  }

      }
    case ipVertexZones:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Compute the center and radius

	mp_real prC[2];
	prC[0] = ( -(m_rE1xE20 - m_rE10xE11)*m_pprE[1][1] + 2.0*m_pprE[1][0]*m_rSqrtDisc ) / m_rDen;
	prC[1] = ( (m_rE1xE20 - m_rE10xE11)*m_pprE[1][0] + 2.0*m_pprE[1][1]*m_rSqrtDisc ) / m_rDen;

	mp_real rR;
	rR = ( m_rE1xE20 + m_rE10xE11 ) / m_rDenR;
  
	// Translate the answer back out away from the origin
	prC[0] += m_rV0_x;
	prC[1] += m_rV0_y;

	*pdC_x = dble(prC[0]);
	*pdC_y = dble(prC[1]);
	*pdR = dble(rR);
	m_eResult = APM2D::icctVEEAntiparallel;
      }
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUNFinal++;
  return TRUE;
}

#else // COMPILER_PROBLEM

BOOL 
ICC_VEEAntiparallel_MPFUNFilter::InscribedCircleCenterVEEAntiparallelMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					      DOUBLE dV00_x, DOUBLE dV00_y,
					      DOUBLE dV0_x, DOUBLE dV0_y,
					      DOUBLE dV01_x, DOUBLE dV01_y,
					      DOUBLE dE10_x, DOUBLE dE10_y, 
					      DOUBLE dE11_x, DOUBLE dE11_y,
					      DOUBLE dE20_x, DOUBLE dE20_y, 
					      DOUBLE dE21_x, DOUBLE dE21_y,
					      InstructionPointer eIPMax/* = ipEnd*/)
{
  APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUN++;

  switch ( m_eIP )
    {
    case ipStart:
      {
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	*pdC_x = 0.0;
	*pdC_y = 0.0;
	*pdR = 0.0;

	//////////////////////////////////////////////////
	// Check V0 on E1 or E2

	m_rV00_x = dV00_x;
	m_rV00_y = dV00_y;
	m_rV0_x = dV0_x;
	m_rV0_y = dV0_y;
	m_rV01_x = dV01_x;
	m_rV01_y = dV01_y;
	m_rE10_x = dE10_x;
	m_rE10_y = dE10_y; 
	m_rE11_x = dE11_x;
	m_rE11_y = dE11_y;
	m_rE20_x = dE20_x;
	m_rE20_y = dE20_y; 
	m_rE21_x = dE21_x;
	m_rE21_y = dE21_y;
  
	if ( m_iV0onE1 == -1 )
	  {
	    mp_real rV0onE1;
	    SinTheta(rV0onE1, m_rV0_x, m_rV0_y, m_rE10_x, m_rE10_y, m_rE11_x, m_rE11_y);
	    if ( rV0onE1 <= 0.0 )
	      {
		// V0 is on E1 and it is not one of the endpoints
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUNFinal++;
		return TRUE;
	      }
  	  }
  
	if ( m_iV0onE2 == -1 )
	  {
	    mp_real rV0onE2;
	    SinTheta(rV0onE2, m_rV0_x, m_rV0_y, m_rE20_x, m_rE20_y, m_rE21_x, m_rE21_y);
	    if ( rV0onE2 <= 0.0 )
	      {
		// V0 is on E2 and it is not one of the endpoints
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUNFinal++;
		return TRUE;
	      }
	  }
  
      }
    case ipV0onE1orE2:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	if ( Colinear(eIPMax) )
	  {
	    APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUNFinal++;
	    return TRUE;
	  }
      }
    case ipColinear:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	if ( Den(eIPMax) )
	  {
	    APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUNFinal++;
	    return TRUE;
	  }
      }
    case ipDen:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	if ( Disc(eIPMax) )
	  {
	    APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUNFinal++;
	    return TRUE;
	  }
      }
    case ipDisc:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	if ( EdgeZones(eIPMax) )
	  {
	    APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUNFinal++;
	    return TRUE;
	  }
      }
    case ipEdgeZones:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	if ( VertexZones(eIPMax) )
	  {
	    APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUNFinal++;
	    return TRUE;
	  }
      }
    case ipVertexZones:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Compute the center and radius

	mp_real prC[2];
	prC[0] = ( -(m_rE1xE20 - m_rE10xE11)*m_pprE[1][1] + 2.0*m_pprE[1][0]*m_rSqrtDisc ) / m_rDen;
	prC[1] = ( (m_rE1xE20 - m_rE10xE11)*m_pprE[1][0] + 2.0*m_pprE[1][1]*m_rSqrtDisc ) / m_rDen;

	mp_real rR;
	rR = ( m_rE1xE20 + m_rE10xE11 ) / m_rDenR;
  
	// Translate the answer back out away from the origin
	prC[0] += m_rV0_x;
	prC[1] += m_rV0_y;

	*pdC_x = dble(prC[0]);
	*pdC_y = dble(prC[1]);
	*pdR = dble(rR);
	m_eResult = APM2D::icctVEEAntiparallel;
      }
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  APM2D::m_uiInscribedCircleCenterVEEAntiparallelMPFUNFinal++;
  return TRUE;
}

BOOL 
ICC_VEEAntiparallel_MPFUNFilter::Colinear(InstructionPointer /*eIPMax*/)
{
  //////////////////////////////////////////////////
  // Check Colinear

  //memset(m_pprE, 0, sizeof(m_pprE));
  m_pprE[1][0] = m_rE11_x - m_rE10_x;
  m_pprE[1][1] = m_rE11_y - m_rE10_y;
  m_pprE[2][0] = m_rE21_x - m_rE20_x;
  m_pprE[2][1] = m_rE21_y - m_rE20_y;
  
  Dot(m_rE1dotE1, m_pprE[1][0], m_pprE[1][1], m_pprE[1][0], m_pprE[1][1]);
  Dot(m_rE2dotE2, m_pprE[2][0], m_pprE[2][1], m_pprE[2][0], m_pprE[2][1]);
  Dot(m_rE1dotE2, m_pprE[1][0], m_pprE[1][1], m_pprE[2][0], m_pprE[2][1]);

  if ( m_iE1E2minusE1dotE2 == -1 )
    {
      mp_real rE1E2minusE1dotE2;  
      rE1E2minusE1dotE2 = sqrt(m_rE1dotE1*m_rE2dotE2) - m_rE1dotE2;
      if ( rE1E2minusE1dotE2 == 0.0 )
	{
	  // Colinear
	  m_eResult = APM2D::icctNull;
	  return TRUE;
	}
    }

  return FALSE;
}

BOOL 
ICC_VEEAntiparallel_MPFUNFilter::Den(InstructionPointer /*eIPMax*/)
{
  //////////////////////////////////////////////////
  // Check Den

  m_rDen = 2*m_rE1dotE1;
  if ( m_iDen == -1 )
    {
      ASSERT( m_rDen != 0.0 );
      m_valE1dotE1.ConservativeBound(dble(m_rE1dotE1));
      m_valDen.ConservativeBound(dble(m_rDen));
      m_iDen = ( m_valDen != 0.0 );
    }

  m_rDenR = 2*sqrt(m_rE1dotE1);
  if ( m_iDenR == -1 )
    {
      ASSERT( m_rDenR != 0.0 );
      m_valDenR.ConservativeBound(dble(m_rDenR));
      m_iDenR = ( m_valDenR != 0.0 );
    }

  return FALSE;
}

BOOL 
ICC_VEEAntiparallel_MPFUNFilter::Disc(InstructionPointer /*eIPMax*/)
{
  //////////////////////////////////////////////////
  // Check Disc

  // Translate V1 to the origin
  //memset(m_ppprV, 0, sizeof(m_ppprV));
  m_ppprV[0][0][0] = 0.0;
  m_ppprV[0][0][1] = 0.0;
  m_ppprV[1][0][0] = m_rE10_x - m_rV0_x;
  m_ppprV[1][0][1] = m_rE10_y - m_rV0_y;
  m_ppprV[1][1][0] = m_rE11_x - m_rV0_x;
  m_ppprV[1][1][1] = m_rE11_y - m_rV0_y;
  m_ppprV[2][0][0] = m_rE20_x - m_rV0_x;
  m_ppprV[2][0][1] = m_rE20_y - m_rV0_y;
  m_ppprV[2][1][0] = m_rE21_x - m_rV0_x;
  m_ppprV[2][1][1] = m_rE21_y - m_rV0_y;
  
  Cross(m_rE10xE11, m_ppprV[1][0][0], m_ppprV[1][0][1], m_ppprV[1][1][0], m_ppprV[1][1][1]);
  mp_real rE20xE21;
  Cross(rE20xE21, m_ppprV[2][0][0], m_ppprV[2][0][1], m_ppprV[2][1][0], m_ppprV[2][1][1]);
  Cross(m_rE1xE20, m_pprE[1][0], m_pprE[1][1], m_ppprV[2][0][0], m_ppprV[2][0][1]);
  mp_real rE2xE10;
  Cross(rE2xE10, m_pprE[2][0], m_pprE[2][1], m_ppprV[1][0][0], m_ppprV[1][0][1]);
  
  mp_real rDisc;
  rDisc = m_rE1xE20*m_rE10xE11;
  if ( m_iDisc == -1 )
    {
      if ( rDisc < 0.0 ) 
	{
	  // Imaginary Solutions
	  m_eResult = APM2D::icctNull;
	  return TRUE;
	}
      m_valDisc.ConservativeBound(dble(rDisc));
      m_iDisc = ( m_valDisc < 0.0 );	    
    }
  m_rSqrtDisc = sqrt(rDisc);
  
  mp_real rDiscE2;
  rDiscE2 = rE2xE10*rE20xE21;
  if ( m_iDiscE2 == -1 )
    {
      if ( rDiscE2 < 0.0 ) 
	{
	  // Imaginary Solutions
	  m_eResult = APM2D::icctNull;
	  return TRUE;
	}
      m_valDiscE2.ConservativeBound(dble(rDiscE2));
      m_iDiscE2 = ( m_valDiscE2 < 0.0 );	    
    }
  m_rSqrtDiscE2 = sqrt(rDiscE2);
  
  return FALSE;
}  

BOOL 
ICC_VEEAntiparallel_MPFUNFilter::EdgeZones(InstructionPointer /*eIPMax*/)
{
  //////////////////////////////////////////////////
  // Check the Edge Zones

  if ( m_ppiEdgeZone[0][0] == -1 ||
       m_ppiEdgeZone[0][1] == -1 ||
       m_ppiEdgeZone[1][0] == -1 ||
       m_ppiEdgeZone[1][1] == -1 )
    {
      mp_real pprEdgeZone[2][2];
      
      if ( m_ppiEdgeZone[0][0] == -1 )
	{
	  mp_real rE1dotE10;
	  Dot(rE1dotE10, m_pprE[1][0], m_pprE[1][1], m_ppprV[1][0][0], m_ppprV[1][0][1]);
	  pprEdgeZone[0][0] = -( rE1dotE10 - m_rSqrtDisc );
	  if ( pprEdgeZone[0][0] < 0.0 )
	    {
	      m_eResult = APM2D::icctNull;
	      return TRUE;
	    }
	}
      
      if ( m_ppiEdgeZone[0][1] == -1 )
	{
	  mp_real rE1dotE11;
	  Dot(rE1dotE11, m_pprE[1][0], m_pprE[1][1], m_ppprV[1][1][0], m_ppprV[1][1][1]);
	  pprEdgeZone[0][1] = ( rE1dotE11 - m_rSqrtDisc );
	  if ( pprEdgeZone[0][1] < 0.0 )
	    {
	      m_eResult = APM2D::icctNull;
	      return TRUE;
	    }
	}
      
      if ( m_ppiEdgeZone[1][0] == -1 )
	{
	  mp_real rE2dotE20;
	  Dot(rE2dotE20, m_pprE[2][0], m_pprE[2][1], m_ppprV[2][0][0], m_ppprV[2][0][1]);
	  pprEdgeZone[1][0] = -( rE2dotE20 + m_rSqrtDiscE2 );
	  if ( pprEdgeZone[1][0] < 0.0 )
	    {
	      m_eResult = APM2D::icctNull;
	      return TRUE;
	    }
	}
      
      if ( m_ppiEdgeZone[1][1] == -1 )
	{
	  mp_real rE2dotE21;
	  Dot(rE2dotE21, m_pprE[2][0], m_pprE[2][1], m_ppprV[2][1][0], m_ppprV[2][1][1]);
	  pprEdgeZone[1][1] = ( rE2dotE21 + m_rSqrtDiscE2 );
	  if ( pprEdgeZone[1][1] < 0.0 )
	    {
	      m_eResult = APM2D::icctNull;
	      return TRUE;
	    }
	}
    }

  return FALSE;
}

BOOL 
ICC_VEEAntiparallel_MPFUNFilter::VertexZones(InstructionPointer /*eIPMax*/)
{
  //////////////////////////////////////////////////
  // Check the Vertex Zone
  
  mp_real ppprVN[1][2][2]; // Endpoints
  // Vertex neighbor vectors
  //memset(ppprVN, 0, sizeof(ppprV));
  ppprVN[0][0][0] = m_rV0_x - m_rV00_x;
  ppprVN[0][0][1] = m_rV0_y - m_rV00_y;
  ppprVN[0][1][0] = m_rV0_x - m_rV01_x;
  ppprVN[0][1][1] = m_rV0_y - m_rV01_y;
  
  if ( m_piVertexZone[0] == -1 ||
       m_piVertexZone[1] == -1 )
    {
      mp_real rE1xE20minusE10xE11 = m_rE1xE20 - m_rE10xE11;
      
      mp_real prVertexZone[2];
      
      if ( m_piVertexZone[0] == -1 )
	{
	  mp_real rE1dotV00;
	  Dot(rE1dotV00, m_pprE[1][0], m_pprE[1][1], ppprVN[0][0][0], ppprVN[0][0][1]);
	  mp_real rE1xV00;
	  Cross(rE1xV00, m_pprE[1][0], m_pprE[1][1], ppprVN[0][0][0], ppprVN[0][0][1]);
	  prVertexZone[0] = ( ( rE1xE20minusE10xE11 )*rE1xV00
			      + 2.0*rE1dotV00*m_rSqrtDisc );
	  if ( prVertexZone[0] < 0.0 )
	    {
	      m_eResult = APM2D::icctNull;
	      return TRUE;
	    }
	}
      
      if ( m_piVertexZone[1] == -1 )
	{
	  mp_real rE1dotV01;
	  Dot(rE1dotV01, m_pprE[1][0], m_pprE[1][1], ppprVN[0][1][0], ppprVN[0][1][1]);
	  mp_real rE1xV01;
	  Cross(rE1xV01, m_pprE[1][0], m_pprE[1][1], ppprVN[0][1][0], ppprVN[0][1][1]);
	  prVertexZone[1] = ( ( rE1xE20minusE10xE11 )*rE1xV01
			      + 2.0*rE1dotV01*m_rSqrtDisc );
	  if ( prVertexZone[1] < 0.0 )
	    {
	      m_eResult = APM2D::icctNull;
	      return TRUE;
	    }
	}
    }

  return FALSE;
}

#endif // COMPILER_PROBLEM

//static 
APM2D::InscribedCircleCenterType 
InscribedCircleCenterVEEAntiparallelMPFUNFilter(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
						DOUBLE dV00_x, DOUBLE dV00_y,
						DOUBLE dV0_x, DOUBLE dV0_y,
						DOUBLE dV01_x, DOUBLE dV01_y,
						DOUBLE dE10_x, DOUBLE dE10_y, 
						DOUBLE dE11_x, DOUBLE dE11_y,
						DOUBLE dE20_x, DOUBLE dE20_y, 
						DOUBLE dE21_x, DOUBLE dE21_y)
{
  ICC_VEEAntiparallel_MPFUNFilter iccVEEAntiparallel;

  return iccVEEAntiparallel.InscribedCircleCenterVEEAntiparallel(pdC_x, pdC_y, pdR, 
								 dV00_x, dV00_y,
								 dV0_x, dV0_y,
								 dV01_x, dV01_y,
								 dE10_x, dE10_y, 
								 dE11_x, dE11_y,
								 dE20_x, dE20_y, 
								 dE21_x, dE21_y);
}

//static 
APM2D::EdgeZoneType EdgeZoneVEEAntiparallelMPFUNFilter(DOUBLE dV0_x, DOUBLE dV0_y,
						       DOUBLE dE10_x, DOUBLE dE10_y, 
						       DOUBLE dE11_x, DOUBLE dE11_y,
						       DOUBLE dE20_x, DOUBLE dE20_y, 
						       DOUBLE dE21_x, DOUBLE dE21_y,
						       DOUBLE dE30_x, DOUBLE dE30_y,
						       DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::m_uiEdgeZoneVEEAntiparallelMPFUNFilter++;

  INT_32 iDisc = -1;
  INT_32 piEdgeZone[2];
  piEdgeZone[0] = -1;
  piEdgeZone[1] = -1;

  switch ( 0 )
  {
  default:
    Interval valV0_x = dV0_x;
    Interval valV0_y = dV0_y;
    Interval valE10_x = dE10_x;
    Interval valE10_y = dE10_y; 
    Interval valE11_x = dE11_x;
    Interval valE11_y = dE11_y;
    Interval valE20_x = dE20_x;
    Interval valE20_y = dE20_y; 
    Interval valE21_x = dE21_x;
    Interval valE21_y = dE21_y;
    Interval valE30_x = dE30_x;
    Interval valE30_y = dE30_y;
    Interval valE31_x = dE31_x;
    Interval valE31_y = dE31_y;

    Interval pppvalV[4][2][2]; // Endpoints
    //memset(pppvalV, 0, sizeof(pppvalV));
    pppvalV[0][0][0] = 0.0;
    pppvalV[0][0][1] = 0.0;
    pppvalV[1][0][0] = valE10_x - valV0_x;
    pppvalV[1][0][1] = valE10_y - valV0_y;
    pppvalV[1][1][0] = valE11_x - valV0_x;
    pppvalV[1][1][1] = valE11_y - valV0_y;
    pppvalV[2][0][0] = valE20_x - valV0_x;
    pppvalV[2][0][1] = valE20_y - valV0_y;
    pppvalV[2][1][0] = valE21_x - valV0_x;
    pppvalV[2][1][1] = valE21_y - valV0_y;
    pppvalV[3][0][0] = valE30_x - valV0_x;
    pppvalV[3][0][1] = valE30_y - valV0_y;
    pppvalV[3][1][0] = valE31_x - valV0_x;
    pppvalV[3][1][1] = valE31_y - valV0_y;
  
    Interval ppvalE[4][2];     // Edge vectors
    //memset(ppvalE, 0, sizeof(ppvalE));
    ppvalE[1][0] = valE11_x - valE10_x;
    ppvalE[1][1] = valE11_y - valE10_y;
    ppvalE[2][0] = valE21_x - valE20_x;
    ppvalE[2][1] = valE21_y - valE20_y;
    ppvalE[3][0] = valE31_x - valE30_x;
    ppvalE[3][1] = valE31_y - valE30_y;
  
    Interval valE1xE20;
    Cross(valE1xE20, ppvalE[1][0], ppvalE[1][1], pppvalV[2][0][0], pppvalV[2][0][1]);
    Interval valE10xE11;
    Cross(valE10xE11, pppvalV[1][0][0], pppvalV[1][0][1], pppvalV[1][1][0], pppvalV[1][1][1]);

    Interval valDisc;
    valDisc = valE1xE20*valE10xE11;
    iDisc = ( valDisc >= 0.0 ); 
    if ( iDisc == -1 )
      {
	break;
      }
    ASSERT( iDisc == 1 ); 
    Interval valSqrtDisc = sqrt(valDisc);

    Interval valE1dotE1;
    Dot(valE1dotE1, ppvalE[1][0], ppvalE[1][1], ppvalE[1][0], ppvalE[1][1]);
    Interval valE1dotE3;
    Dot(valE1dotE3, ppvalE[1][0], ppvalE[1][1], ppvalE[3][0], ppvalE[3][1]);
    Interval valE3dotE3;
    Dot(valE3dotE3, ppvalE[3][0], ppvalE[3][1], ppvalE[3][0], ppvalE[3][1]);

    Interval valE1xE3;
    Cross(valE1xE3, ppvalE[1][0], ppvalE[1][1], ppvalE[3][0], ppvalE[3][1]);

    Interval valE1xE20minusE10xE11 = valE1xE20 - valE10xE11;
    Interval valCoef1 = ( - ( valE1xE20minusE10xE11 )*valE1xE3
		       - 2.0*valE1dotE3*valSqrtDisc );

    Interval pvalEdgeZone[2];

    Interval valE3dotE30;
    Dot(valE3dotE30, ppvalE[3][0], ppvalE[3][1], pppvalV[3][0][0], pppvalV[3][0][1]);
    pvalEdgeZone[0] = ( 2.0*valE1dotE1*valE3dotE30 
		      + valCoef1 );
    piEdgeZone[0] = ( pvalEdgeZone[0] >= 0.0 );
    if ( piEdgeZone[0] == 1 )  
      {
	return APM2D::eztStart;
      }

    Interval valE3dotE31;
    Dot(valE3dotE31, ppvalE[3][0], ppvalE[3][1], pppvalV[3][1][0], pppvalV[3][1][1]);  
    pvalEdgeZone[1] = -( 2.0*valE1dotE1*valE3dotE31 
		       + valCoef1 );
    piEdgeZone[1] = ( pvalEdgeZone[1] >= 0.0 );
    if ( piEdgeZone[1] == 1 )
      {
	return APM2D::eztEnd;
      }
  }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( piEdgeZone[0] == -1 || piEdgeZone[1] == -1 )
    {
      APM2D::m_uiEdgeZoneVEEAntiparallelMPFUN++;

      mp_real rV0_x = dV0_x;
      mp_real rV0_y = dV0_y;
      mp_real rE10_x = dE10_x;
      mp_real rE10_y = dE10_y; 
      mp_real rE11_x = dE11_x;
      mp_real rE11_y = dE11_y;
      mp_real rE20_x = dE20_x;
      mp_real rE20_y = dE20_y; 
      mp_real rE21_x = dE21_x;
      mp_real rE21_y = dE21_y;
      mp_real rE30_x = dE30_x;
      mp_real rE30_y = dE30_y;
      mp_real rE31_x = dE31_x;
      mp_real rE31_y = dE31_y;

      mp_real ppprV[4][2][2]; // Endpoints
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = 0.0;
      ppprV[0][0][1] = 0.0;
      ppprV[1][0][0] = rE10_x - rV0_x;
      ppprV[1][0][1] = rE10_y - rV0_y;
      ppprV[1][1][0] = rE11_x - rV0_x;
      ppprV[1][1][1] = rE11_y - rV0_y;
      ppprV[2][0][0] = rE20_x - rV0_x;
      ppprV[2][0][1] = rE20_y - rV0_y;
      ppprV[2][1][0] = rE21_x - rV0_x;
      ppprV[2][1][1] = rE21_y - rV0_y;
      ppprV[3][0][0] = rE30_x - rV0_x;
      ppprV[3][0][1] = rE30_y - rV0_y;
      ppprV[3][1][0] = rE31_x - rV0_x;
      ppprV[3][1][1] = rE31_y - rV0_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[1][0] = rE11_x - rE10_x;
      pprE[1][1] = rE11_y - rE10_y;
      pprE[2][0] = rE21_x - rE20_x;
      pprE[2][1] = rE21_y - rE20_y;
      pprE[3][0] = rE31_x - rE30_x;
      pprE[3][1] = rE31_y - rE30_y;
  
      mp_real rE1xE20;
      Cross(rE1xE20, pprE[1][0], pprE[1][1], ppprV[2][0][0], ppprV[2][0][1]);
      mp_real rE10xE11;
      Cross(rE10xE11, ppprV[1][0][0], ppprV[1][0][1], ppprV[1][1][0], ppprV[1][1][1]);

      mp_real rDisc;
      rDisc = rE1xE20*rE10xE11;
      ASSERT( rDisc >= 0.0 ); 
      mp_real rSqrtDisc = sqrt(rDisc);

      mp_real rE1dotE1;
      Dot(rE1dotE1, pprE[1][0], pprE[1][1], pprE[1][0], pprE[1][1]);
      mp_real rE1dotE3;
      Dot(rE1dotE3, pprE[1][0], pprE[1][1], pprE[3][0], pprE[3][1]);
      mp_real rE3dotE3;
      Dot(rE3dotE3, pprE[3][0], pprE[3][1], pprE[3][0], pprE[3][1]);

      mp_real rE1xE3;
      Cross(rE1xE3, pprE[1][0], pprE[1][1], pprE[3][0], pprE[3][1]);

      mp_real rE1xE20minusE10xE11 = rE1xE20 - rE10xE11;
      mp_real rCoef1 = ( - ( rE1xE20minusE10xE11 )*rE1xE3
			 - 2.0*rE1dotE3*rSqrtDisc );

      mp_real prEdgeZone[2];

      if ( piEdgeZone[0] == -1 )
	{
	  mp_real rE3dotE30;
	  Dot(rE3dotE30, pprE[3][0], pprE[3][1], ppprV[3][0][0], ppprV[3][0][1]);
	  prEdgeZone[0] = ( 2.0*rE1dotE1*rE3dotE30 
			    + rCoef1 );
	  if ( prEdgeZone[0] >= 0.0 )
	    {
	      return APM2D::eztStart;
	    }
	}

      if ( piEdgeZone[1] == -1 )
	{
	  mp_real rE3dotE31;
	  Dot(rE3dotE31, pprE[3][0], pprE[3][1], ppprV[3][1][0], ppprV[3][1][1]);  
	  prEdgeZone[1] = -( 2.0*rE1dotE1*rE3dotE31 
			     + rCoef1 );
	  if ( prEdgeZone[1] >= 0.0 )
	    {
	      return APM2D::eztEnd;
	    }
	}
    }

  return APM2D::eztMiddle;
}

//static 
APM2D::CircleTestType InCircleVEE_VAntiparallelMPFUNFilter(DOUBLE dV0_x, DOUBLE dV0_y,
							   DOUBLE dE10_x, DOUBLE dE10_y, 
							   DOUBLE dE11_x, DOUBLE dE11_y,
							   DOUBLE dE20_x, DOUBLE dE20_y, 
							   DOUBLE dE21_x, DOUBLE dE21_y,
							   DOUBLE dV3_x, DOUBLE dV3_y)
{
  APM2D::m_uiInCircleVEE_VAntiparallelMPFUNFilter++;

  INT_32 iResult = -1;
  INT_32 iDisc = -1;

  switch ( 0 )
    {
    default:
      Interval valV0_x = dV0_x;
      Interval valV0_y = dV0_y;
      Interval valE10_x = dE10_x;
      Interval valE10_y = dE10_y; 
      Interval valE11_x = dE11_x;
      Interval valE11_y = dE11_y;
      Interval valE20_x = dE20_x;
      Interval valE20_y = dE20_y; 
      Interval valE21_x = dE21_x;
      Interval valE21_y = dE21_y;
      Interval valV3_x = dV3_x;
      Interval valV3_y = dV3_y;

      Interval pppvalV[4][2][2]; // Endpoints
      //memset(pppvalV, 0, sizeof(pppvalV));
      pppvalV[0][0][0] = 0.0;
      pppvalV[0][0][1] = 0.0;
      pppvalV[1][0][0] = valE10_x - valV0_x;
      pppvalV[1][0][1] = valE10_y - valV0_y;
      pppvalV[1][1][0] = valE11_x - valV0_x;
      pppvalV[1][1][1] = valE11_y - valV0_y;
      pppvalV[2][0][0] = valE20_x - valV0_x;
      pppvalV[2][0][1] = valE20_y - valV0_y;
      pppvalV[2][1][0] = valE21_x - valV0_x;
      pppvalV[2][1][1] = valE21_y - valV0_y;
      pppvalV[3][0][0] = valV3_x - valV0_x;
      pppvalV[3][0][1] = valV3_y - valV0_y;
  
      Interval ppvalE[4][2];     // Edge vectors
      //memset(ppvalE, 0, sizeof(ppvalE));
      ppvalE[1][0] = valE11_x - valE10_x;
      ppvalE[1][1] = valE11_y - valE10_y;
      ppvalE[2][0] = valE21_x - valE20_x;
      ppvalE[2][1] = valE21_y - valE20_y;

      Interval valE1xE20;
      Cross(valE1xE20, ppvalE[1][0], ppvalE[1][1], pppvalV[2][0][0], pppvalV[2][0][1]);
      Interval valE10xE11;
      Cross(valE10xE11, pppvalV[1][0][0], pppvalV[1][0][1], pppvalV[1][1][0], pppvalV[1][1][1]);

      Interval valDisc;
      valDisc = valE1xE20*valE10xE11;
      iDisc = ( valDisc >= 0.0 ); 
      if ( iDisc == -1 )
	{
	  break;
	}
      ASSERT( iDisc == 1 ); 
      Interval valSqrtDisc = sqrt(valDisc);

      Interval valE1dotE1;
      Dot(valE1dotE1, ppvalE[1][0], ppvalE[1][1], ppvalE[1][0], ppvalE[1][1]);
      Interval valV3dotV3;
      Dot(valV3dotV3, pppvalV[3][0][0], pppvalV[3][0][1], pppvalV[3][0][0], pppvalV[3][0][1]);
      Interval valE1dotV3;
      Dot(valE1dotV3, ppvalE[1][0], ppvalE[1][1], pppvalV[3][0][0], pppvalV[3][0][1]);

      Interval valE1xV3;
      Cross(valE1xV3, ppvalE[1][0], ppvalE[1][1], pppvalV[3][0][0], pppvalV[3][0][1]);

      Interval valResult;
      valResult = ( valE1dotE1*valV3dotV3 
		  - ( valE1xE20 - valE10xE11 )*valE1xV3
		  - 2.0*valE1dotV3*valSqrtDisc );

      iResult = ( valResult < 0.0 );
      if ( iResult == -1 )
	{
	  break;
	}
      if ( iResult == 1 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( valResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
      return APM2D::cttOnCircle;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( iResult == -1 )
    {
      APM2D::m_uiInCircleVEE_VAntiparallelMPFUN++;

      mp_real rV0_x = dV0_x;
      mp_real rV0_y = dV0_y;
      mp_real rE10_x = dE10_x;
      mp_real rE10_y = dE10_y; 
      mp_real rE11_x = dE11_x;
      mp_real rE11_y = dE11_y;
      mp_real rE20_x = dE20_x;
      mp_real rE20_y = dE20_y; 
      mp_real rE21_x = dE21_x;
      mp_real rE21_y = dE21_y;
      mp_real rV3_x = dV3_x;
      mp_real rV3_y = dV3_y;

      mp_real ppprV[4][2][2]; // Endpoints
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = 0.0;
      ppprV[0][0][1] = 0.0;
      ppprV[1][0][0] = rE10_x - rV0_x;
      ppprV[1][0][1] = rE10_y - rV0_y;
      ppprV[1][1][0] = rE11_x - rV0_x;
      ppprV[1][1][1] = rE11_y - rV0_y;
      ppprV[2][0][0] = rE20_x - rV0_x;
      ppprV[2][0][1] = rE20_y - rV0_y;
      ppprV[2][1][0] = rE21_x - rV0_x;
      ppprV[2][1][1] = rE21_y - rV0_y;
      ppprV[3][0][0] = rV3_x - rV0_x;
      ppprV[3][0][1] = rV3_y - rV0_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[1][0] = rE11_x - rE10_x;
      pprE[1][1] = rE11_y - rE10_y;
      pprE[2][0] = rE21_x - rE20_x;
      pprE[2][1] = rE21_y - rE20_y;

      mp_real rE1xE20;
      Cross(rE1xE20, pprE[1][0], pprE[1][1], ppprV[2][0][0], ppprV[2][0][1]);
      mp_real rE10xE11;
      Cross(rE10xE11, ppprV[1][0][0], ppprV[1][0][1], ppprV[1][1][0], ppprV[1][1][1]);

      mp_real rDisc;
      rDisc = rE1xE20*rE10xE11;
      ASSERT( rDisc >= 0.0 ); 
      mp_real rSqrtDisc = sqrt(rDisc);

      mp_real rE1dotE1;
      Dot(rE1dotE1, pprE[1][0], pprE[1][1], pprE[1][0], pprE[1][1]);
      mp_real rV3dotV3;
      Dot(rV3dotV3, ppprV[3][0][0], ppprV[3][0][1], ppprV[3][0][0], ppprV[3][0][1]);
      mp_real rE1dotV3;
      Dot(rE1dotV3, pprE[1][0], pprE[1][1], ppprV[3][0][0], ppprV[3][0][1]);

      mp_real rE1xV3;
      Cross(rE1xV3, pprE[1][0], pprE[1][1], ppprV[3][0][0], ppprV[3][0][1]);

      mp_real rResult;
      rResult = ( rE1dotE1*rV3dotV3 
		  - ( rE1xE20 - rE10xE11 )*rE1xV3
		  - 2.0*rE1dotV3*rSqrtDisc );

      if ( rResult < 0.0 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( rResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
    }

  return APM2D::cttOnCircle;
}

//static 
APM2D::CircleTestType InCircleVEE_EAntiparallelMPFUNFilter(DOUBLE dV0_x, DOUBLE dV0_y,
						     DOUBLE dE10_x, DOUBLE dE10_y, 
						     DOUBLE dE11_x, DOUBLE dE11_y,
						     DOUBLE dE20_x, DOUBLE dE20_y, 
						     DOUBLE dE21_x, DOUBLE dE21_y,
						     DOUBLE dE30_x, DOUBLE dE30_y,
						     DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::m_uiInCircleVEE_EAntiparallelMPFUNFilter++;

  INT_32 iResult = -1;
  INT_32 iDisc = -1;

  switch ( 0 )
    {
    default:
      Interval valV0_x = dV0_x;
      Interval valV0_y = dV0_y;
      Interval valE10_x = dE10_x;
      Interval valE10_y = dE10_y; 
      Interval valE11_x = dE11_x;
      Interval valE11_y = dE11_y;
      Interval valE20_x = dE20_x;
      Interval valE20_y = dE20_y; 
      Interval valE21_x = dE21_x;
      Interval valE21_y = dE21_y;
      Interval valE30_x = dE30_x;
      Interval valE30_y = dE30_y;
      Interval valE31_x = dE31_x;
      Interval valE31_y = dE31_y;

      Interval pppvalV[4][2][2]; // Endpoints
      //memset(pppvalV, 0, sizeof(pppvalV));
      pppvalV[0][0][0] = 0.0;
      pppvalV[0][0][1] = 0.0;
      pppvalV[1][0][0] = valE10_x - valV0_x;
      pppvalV[1][0][1] = valE10_y - valV0_y;
      pppvalV[1][1][0] = valE11_x - valV0_x;
      pppvalV[1][1][1] = valE11_y - valV0_y;
      pppvalV[2][0][0] = valE20_x - valV0_x;
      pppvalV[2][0][1] = valE20_y - valV0_y;
      pppvalV[2][1][0] = valE21_x - valV0_x;
      pppvalV[2][1][1] = valE21_y - valV0_y;
      pppvalV[3][0][0] = valE30_x - valV0_x;
      pppvalV[3][0][1] = valE30_y - valV0_y;
      pppvalV[3][1][0] = valE31_x - valV0_x;
      pppvalV[3][1][1] = valE31_y - valV0_y;
  
      Interval ppvalE[4][2];     // Edge vectors
      //memset(ppvalE, 0, sizeof(ppvalE));
      ppvalE[1][0] = valE11_x - valE10_x;
      ppvalE[1][1] = valE11_y - valE10_y;
      ppvalE[2][0] = valE21_x - valE20_x;
      ppvalE[2][1] = valE21_y - valE20_y;
      ppvalE[3][0] = valE31_x - valE30_x;
      ppvalE[3][1] = valE31_y - valE30_y;

      Interval valE1xE20;
      Cross(valE1xE20, ppvalE[1][0], ppvalE[1][1], pppvalV[2][0][0], pppvalV[2][0][1]);
      Interval valE10xE11;
      Cross(valE10xE11, pppvalV[1][0][0], pppvalV[1][0][1], pppvalV[1][1][0], pppvalV[1][1][1]);

      Interval valDisc;
      valDisc = valE1xE20*valE10xE11;
      iDisc = ( valDisc >= 0.0 ); 
      if ( iDisc == -1 )
	{
	  break;
	}
      ASSERT( iDisc == 1 ); 
      Interval valSqrtDisc = sqrt(valDisc);

      Interval valE1dotE1;
      Dot(valE1dotE1, ppvalE[1][0], ppvalE[1][1], ppvalE[1][0], ppvalE[1][1]);
      Interval valE2dotE2;
      Dot(valE2dotE2, ppvalE[2][0], ppvalE[2][1], ppvalE[2][0], ppvalE[2][1]);
      Interval valE3dotE3;
      Dot(valE3dotE3, ppvalE[3][0], ppvalE[3][1], ppvalE[3][0], ppvalE[3][1]);
      Interval valE1dotE2;
      Dot(valE1dotE2, ppvalE[1][0], ppvalE[1][1], ppvalE[2][0], ppvalE[2][1]);
      Interval valE1dotE3;
      Dot(valE1dotE3, ppvalE[1][0], ppvalE[1][1], ppvalE[3][0], ppvalE[3][1]);

      Interval valE1xE3;
      Cross(valE1xE3, ppvalE[1][0], ppvalE[1][1], ppvalE[3][0], ppvalE[3][1]);
      Interval valE20xE21;
      Cross(valE20xE21, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][1][0], pppvalV[2][1][1]);
      Interval valE30xE31;
      Cross(valE30xE31, pppvalV[3][0][0], pppvalV[3][0][1], pppvalV[3][1][0], pppvalV[3][1][1]);

      Interval valE1E2minusE1dotE2, valE1xE20minusE10xE11; 
      valE1E2minusE1dotE2 = ( sqrt(valE1dotE1*valE2dotE2) - valE1dotE2 );
      valE1xE20minusE10xE11 = valE1xE20 - valE10xE11;

      Interval valResult;
      valResult = ( 4.0*valE1dotE1*valE30xE31*( valE1dotE1*valE30xE31 + valE1dotE3*valE1xE20minusE10xE11 )
		  - 4.0*valE1dotE3*valE1dotE3*valE1xE20*valE20xE21
		  - valE1xE3*valE1xE3*valE1xE20minusE10xE11*valE1xE20minusE10xE11
		  + 4.0*valE1xE3*( 2.0*valE1dotE1*valE30xE31 + valE1dotE3*valE1xE20minusE10xE11 )*valSqrtDisc );

      iResult = ( valResult < 0.0 );
      if ( iResult == -1 )
	{
	  break;
	}
      if ( iResult == 1 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( valResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
      return APM2D::cttOnCircle;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( iResult == -1 )
    {
      APM2D::m_uiInCircleVEE_EAntiparallelMPFUN++;

      mp_real rV0_x = dV0_x;
      mp_real rV0_y = dV0_y;
      mp_real rE10_x = dE10_x;
      mp_real rE10_y = dE10_y; 
      mp_real rE11_x = dE11_x;
      mp_real rE11_y = dE11_y;
      mp_real rE20_x = dE20_x;
      mp_real rE20_y = dE20_y; 
      mp_real rE21_x = dE21_x;
      mp_real rE21_y = dE21_y;
      mp_real rE30_x = dE30_x;
      mp_real rE30_y = dE30_y;
      mp_real rE31_x = dE31_x;
      mp_real rE31_y = dE31_y;

      mp_real ppprV[4][2][2]; // Endpoints
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = 0.0;
      ppprV[0][0][1] = 0.0;
      ppprV[1][0][0] = rE10_x - rV0_x;
      ppprV[1][0][1] = rE10_y - rV0_y;
      ppprV[1][1][0] = rE11_x - rV0_x;
      ppprV[1][1][1] = rE11_y - rV0_y;
      ppprV[2][0][0] = rE20_x - rV0_x;
      ppprV[2][0][1] = rE20_y - rV0_y;
      ppprV[2][1][0] = rE21_x - rV0_x;
      ppprV[2][1][1] = rE21_y - rV0_y;
      ppprV[3][0][0] = rE30_x - rV0_x;
      ppprV[3][0][1] = rE30_y - rV0_y;
      ppprV[3][1][0] = rE31_x - rV0_x;
      ppprV[3][1][1] = rE31_y - rV0_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[1][0] = rE11_x - rE10_x;
      pprE[1][1] = rE11_y - rE10_y;
      pprE[2][0] = rE21_x - rE20_x;
      pprE[2][1] = rE21_y - rE20_y;
      pprE[3][0] = rE31_x - rE30_x;
      pprE[3][1] = rE31_y - rE30_y;

      mp_real rE1xE20;
      Cross(rE1xE20, pprE[1][0], pprE[1][1], ppprV[2][0][0], ppprV[2][0][1]);
      mp_real rE10xE11;
      Cross(rE10xE11, ppprV[1][0][0], ppprV[1][0][1], ppprV[1][1][0], ppprV[1][1][1]);

      mp_real rDisc;
      rDisc = rE1xE20*rE10xE11;
      ASSERT( rDisc >= 0.0 ); 
      mp_real rSqrtDisc = sqrt(rDisc);

      mp_real rE1dotE1;
      Dot(rE1dotE1, pprE[1][0], pprE[1][1], pprE[1][0], pprE[1][1]);
      mp_real rE2dotE2;
      Dot(rE2dotE2, pprE[2][0], pprE[2][1], pprE[2][0], pprE[2][1]);
      mp_real rE3dotE3;
      Dot(rE3dotE3, pprE[3][0], pprE[3][1], pprE[3][0], pprE[3][1]);
      mp_real rE1dotE2;
      Dot(rE1dotE2, pprE[1][0], pprE[1][1], pprE[2][0], pprE[2][1]);
      mp_real rE1dotE3;
      Dot(rE1dotE3, pprE[1][0], pprE[1][1], pprE[3][0], pprE[3][1]);

      mp_real rE1xE3;
      Cross(rE1xE3, pprE[1][0], pprE[1][1], pprE[3][0], pprE[3][1]);
      mp_real rE20xE21;
      Cross(rE20xE21, ppprV[2][0][0], ppprV[2][0][1], ppprV[2][1][0], ppprV[2][1][1]);
      mp_real rE30xE31;
      Cross(rE30xE31, ppprV[3][0][0], ppprV[3][0][1], ppprV[3][1][0], ppprV[3][1][1]);

      mp_real rE1E2minusE1dotE2, rE1xE20minusE10xE11; 
      rE1E2minusE1dotE2 = ( sqrt(rE1dotE1*rE2dotE2) - rE1dotE2 );
      rE1xE20minusE10xE11 = rE1xE20 - rE10xE11;

      mp_real rResult;
      rResult = ( 4.0*rE1dotE1*rE30xE31*( rE1dotE1*rE30xE31 + rE1dotE3*rE1xE20minusE10xE11 )
		  - 4.0*rE1dotE3*rE1dotE3*rE1xE20*rE20xE21
		  - rE1xE3*rE1xE3*rE1xE20minusE10xE11*rE1xE20minusE10xE11
		  + 4.0*rE1xE3*( 2.0*rE1dotE1*rE30xE31 + rE1dotE3*rE1xE20minusE10xE11 )*rSqrtDisc );
  
      if ( rResult < 0.0 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( rResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
    }

  return APM2D::cttOnCircle;
}

//////////////////////////////////////////////////
// vEE

class ICC_vEE_MPFUNFilter
{
protected:
  enum InstructionPointer {ipStart, ipV0onE2, 
			   ipColinear, ipDen, 
			   ipEdgeZones, ipEnd};

public:
  ICC_vEE_MPFUNFilter();
  ~ICC_vEE_MPFUNFilter();

public:
  APM2D::InscribedCircleCenterType 
  InscribedCircleCentervEE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			   DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
			   DOUBLE dE11_x, DOUBLE dE11_y,
			   DOUBLE dE20_x, DOUBLE dE20_y, 
			   DOUBLE dE21_x, DOUBLE dE21_y);
  APM2D::InscribedCircleCenterType 
  InscribedCircleCentervEEForce(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
				DOUBLE dE11_x, DOUBLE dE11_y,
				DOUBLE dE20_x, DOUBLE dE20_y, 
				DOUBLE dE21_x, DOUBLE dE21_y);

protected:
#ifndef COMPILER_PROBLEM
  BOOL
  InscribedCircleCentervEEMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
				DOUBLE dE11_x, DOUBLE dE11_y,
				DOUBLE dE20_x, DOUBLE dE20_y, 
				DOUBLE dE21_x, DOUBLE dE21_y,
				InstructionPointer eIPMax = ipEnd);
#else // COMPILER_PROBLEM
  BOOL
  InscribedCircleCentervEEMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
				DOUBLE dE11_x, DOUBLE dE11_y,
				DOUBLE dE20_x, DOUBLE dE20_y, 
				DOUBLE dE21_x, DOUBLE dE21_y,
				InstructionPointer eIPMax = ipEnd);

  BOOL Colinear(InstructionPointer eIPMax);
  BOOL Den(InstructionPointer eIPMax);
  BOOL EdgeZones(InstructionPointer eIPMax);
  VOID CenterAndRadius(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
		       InstructionPointer eIPMax);
#endif // COMPILER_PROBLEM

protected:
  InstructionPointer m_eIP;
  APM2D::InscribedCircleCenterType m_eResult;
  INT_32 m_iV0onE2;
  INT_32 m_iE1xE2;
  INT_32 m_iE1dotE2;
  INT_32 m_iDen;
  INT_32 m_piEdgeZone[2];

  Interval m_valE1xE2;
  Interval m_valE1dotE2;
  Interval m_valDen;

  mp_real m_rV0_E10_x;
  mp_real m_rV0_E10_y;
  mp_real m_rE11_x;
  mp_real m_rE11_y;
  mp_real m_rE20_x;
  mp_real m_rE20_y;
  mp_real m_rE21_x;
  mp_real m_rE21_y;
  mp_real m_pprE[3][2];
  mp_real m_rE1dotE2;
  mp_real m_rE1xE2;
  mp_real m_rE1dotE1;
  mp_real m_rE1E2minusE1dotE2;
  mp_real m_rDen;
  mp_real m_rE20xE21;
};

ICC_vEE_MPFUNFilter::ICC_vEE_MPFUNFilter()
{
}

ICC_vEE_MPFUNFilter::~ICC_vEE_MPFUNFilter()
{
}

APM2D::InscribedCircleCenterType 
ICC_vEE_MPFUNFilter::InscribedCircleCentervEE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					      DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
					      DOUBLE dE11_x, DOUBLE dE11_y,
					      DOUBLE dE20_x, DOUBLE dE20_y, 
					      DOUBLE dE21_x, DOUBLE dE21_y)
{
  APM2D::m_uiInscribedCircleCentervEEMPFUNFilter++;

  *pdC_x = 0.0;
  *pdC_y = 0.0;
  *pdR = 0.0;
  
  m_eIP = ipStart;
  m_eResult = APM2D::icctNull;
  m_iV0onE2 = -1;
  m_iE1xE2 = -1;
  m_iE1dotE2 = -1;
  m_iDen = -1;
  m_piEdgeZone[0] = -1;
  m_piEdgeZone[1] = -1;

  BOOL bDone;

  Interval valV0_E10_x = dV0_E10_x;
  Interval valV0_E10_y = dV0_E10_y; 
  Interval valE11_x = dE11_x;
  Interval valE11_y = dE11_y;
  Interval valE20_x = dE20_x;
  Interval valE20_y = dE20_y; 
  Interval valE21_x = dE21_x;
  Interval valE21_y = dE21_y;
  
  //////////////////////////////////////////////////
  // Check for V0 on E2
  
  Interval valV0onE2;
  SinTheta(valV0onE2, valV0_E10_x, valV0_E10_y, valE20_x, valE20_y, valE21_x, valE21_y);
  m_iV0onE2 = ( valV0onE2 <= 0.0 );
  if ( m_iV0onE2 == 1 )
    {
      // V0 is on E2 and it is not one of the endpoints
      return APM2D::icctNull;
    }
  if ( m_iV0onE2 == -1 )
    {
      bDone = InscribedCircleCentervEEMPFUN(pdC_x, pdC_y, pdR, 
					    dV0_E10_x, dV0_E10_y, 
					    dE11_x, dE11_y,
					    dE20_x, dE20_y, 
					    dE21_x, dE21_y,
					    ipV0onE2);  
      if ( bDone )
	{
	  return m_eResult;
	}
    }

  //////////////////////////////////////////////////
  // Check Colinear
  
  Interval ppvalE[3][2];     // Edge vectors
  //memset(ppvalE, 0, sizeof(ppvalE));
  ppvalE[1][0] = valE11_x - valV0_E10_x;
  ppvalE[1][1] = valE11_y - valV0_E10_y;
  ppvalE[2][0] = valE21_x - valE20_x;
  ppvalE[2][1] = valE21_y - valE20_y;
  
  Dot(m_valE1dotE2, ppvalE[1][0], ppvalE[1][1], ppvalE[2][0], ppvalE[2][1]);
  Cross(m_valE1xE2, ppvalE[1][0], ppvalE[1][1], ppvalE[2][0], ppvalE[2][1]);

  m_iE1xE2 = ( m_valE1xE2 == 0.0 );
  m_iE1dotE2 = ( m_valE1dotE2 > 0.0 );
  if ( m_iE1xE2 == -1 || m_iE1dotE2 == -1 )
    {
      bDone = InscribedCircleCentervEEMPFUN(pdC_x, pdC_y, pdR, 
					    dV0_E10_x, dV0_E10_y, 
					    dE11_x, dE11_y,
					    dE20_x, dE20_y, 
					    dE21_x, dE21_y,
					    ipColinear);  
      if ( bDone )
	{
	  return m_eResult;
	}
      if ( m_iE1xE2 == -1 || m_iE1dotE2 == -1 )
	{
	  bDone = InscribedCircleCentervEEMPFUN(pdC_x, pdC_y, pdR, 
						dV0_E10_x, dV0_E10_y, 
						dE11_x, dE11_y,
						dE20_x, dE20_y, 
						dE21_x, dE21_y,
						ipEnd);  
	  ASSERT( bDone );
	  return m_eResult;
	}
    }
  if ( m_iE1xE2 == 1 && m_iE1dotE2 == 1 )
    {
      // Colinear
      return APM2D::icctNull;
    }

  //////////////////////////////////////////////////
  // Check Den
  
  Interval valE1dotE1;
  Dot(valE1dotE1, ppvalE[1][0], ppvalE[1][1], ppvalE[1][0], ppvalE[1][1]);
  Interval valE2dotE2;
  Dot(valE2dotE2, ppvalE[2][0], ppvalE[2][1], ppvalE[2][0], ppvalE[2][1]);
  
  Interval valE1E2minusE1dotE2;
  valE1E2minusE1dotE2 = sqrt(valE1dotE1*valE2dotE2) - m_valE1dotE2;
  
  m_valDen = valE1E2minusE1dotE2;
  m_iDen = ( m_valDen != 0.0 );
  if ( m_iDen == -1 )
    {
      bDone = InscribedCircleCentervEEMPFUN(pdC_x, pdC_y, pdR, 
					    dV0_E10_x, dV0_E10_y, 
					    dE11_x, dE11_y,
					    dE20_x, dE20_y, 
					    dE21_x, dE21_y,
					    ipDen);  
      if ( bDone )
	{
	  return m_eResult;
	}
      if ( m_iDen == -1 )
	{
	  bDone = InscribedCircleCentervEEMPFUN(pdC_x, pdC_y, pdR, 
						dV0_E10_x, dV0_E10_y, 
						dE11_x, dE11_y,
						dE20_x, dE20_y, 
						dE21_x, dE21_y,
						ipEnd);  
	  ASSERT( bDone );
	  return m_eResult;
	}
    }
  ASSERT( m_iDen == 1 );

  //////////////////////////////////////////////////
  // Check the Edge Zone
  
  Interval pppvalV[3][2][2]; // Endpoints
  // Translate V1 to the origin
  //memset(pppvalV, 0, sizeof(pppvalV));
  pppvalV[0][0][0] = 0.0;
  pppvalV[0][0][1] = 0.0;
  pppvalV[1][0][0] = 0.0;
  pppvalV[1][0][1] = 0.0;
  pppvalV[1][1][0] = valE11_x - valV0_E10_x;
  pppvalV[1][1][1] = valE11_y - valV0_E10_y;
  pppvalV[2][0][0] = valE20_x - valV0_E10_x;
  pppvalV[2][0][1] = valE20_y - valV0_E10_y;
  pppvalV[2][1][0] = valE21_x - valV0_E10_x;
  pppvalV[2][1][1] = valE21_y - valV0_E10_y;
  
  Interval valE20xE21;
  Cross(valE20xE21, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][1][0], pppvalV[2][1][1]);
  
  Interval valE20xE21timesE1xE2 = valE20xE21*m_valE1xE2;
  
  Interval pvalEdgeZone[2];
  
  Interval valE2dotE20;
  Dot(valE2dotE20, ppvalE[2][0], ppvalE[2][1], pppvalV[2][0][0], pppvalV[2][0][1]);
  pvalEdgeZone[0] = -( valE1E2minusE1dotE2*valE2dotE20 - valE20xE21timesE1xE2 );
  m_piEdgeZone[0] = ( pvalEdgeZone[0] < 0.0 );
  if ( m_piEdgeZone[0] == 1 )
    {
      return APM2D::icctNull;
    }
  
  Interval valE2dotE21;
  Dot(valE2dotE21, ppvalE[2][0], ppvalE[2][1], pppvalV[2][1][0], pppvalV[2][1][1]);
  pvalEdgeZone[1] = ( valE1E2minusE1dotE2*valE2dotE21 - valE20xE21timesE1xE2 );
  m_piEdgeZone[1] = ( pvalEdgeZone[1] < 0.0 );
  if ( m_piEdgeZone[1] == 1 )
    {
      return APM2D::icctNull;
    }

  if ( m_piEdgeZone[0] == -1 ||
       m_piEdgeZone[1] == -1 )
    {
      bDone = InscribedCircleCentervEEMPFUN(pdC_x, pdC_y, pdR, 
					    dV0_E10_x, dV0_E10_y, 
					    dE11_x, dE11_y,
					    dE20_x, dE20_y, 
					    dE21_x, dE21_y,
					    ipEdgeZones);  
      if ( bDone )
	{
	  return m_eResult;
	}
    }

  //////////////////////////////////////////////////
  // Compute the center and radius
  
  Interval pvalC[2];
  pvalC[0] = -( valE20xE21*ppvalE[1][1] ) / m_valDen;
  pvalC[1] = ( valE20xE21*ppvalE[1][0] ) / m_valDen;
  
  Interval valR;
  valR = ( sqrt(valE1dotE1)*valE20xE21 ) / m_valDen;
  
  // Translate the answer back out away from the origin
  pvalC[0] += valV0_E10_x;
  pvalC[1] += valV0_E10_y;
  
  *pdC_x = dble(pvalC[0]);
  *pdC_y = dble(pvalC[1]);
  *pdR = dble(valR);
  return APM2D::icctvEE;
}

APM2D::InscribedCircleCenterType 
ICC_vEE_MPFUNFilter::InscribedCircleCentervEEForce(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
						   DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
						   DOUBLE dE11_x, DOUBLE dE11_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y)
{
  *pdC_x = 0.0;
  *pdC_y = 0.0;
  *pdR = 0.0;
  
  m_eIP = ipStart;
  m_eResult = APM2D::icctNull;
  m_iV0onE2 = -1;
  m_iE1xE2 = -1;
  m_iE1dotE2 = -1;
  m_iDen = -1;
  m_piEdgeZone[0] = -1;
  m_piEdgeZone[1] = -1;

  BOOL bDone;
  
  bDone = InscribedCircleCentervEEMPFUN(pdC_x, pdC_y, pdR, 
					dV0_E10_x, dV0_E10_y, 
					dE11_x, dE11_y,
					dE20_x, dE20_y, 
					dE21_x, dE21_y,
					ipEnd);
  ASSERT( bDone );
  return m_eResult;
}

#ifndef COMPILER_PROBLEM

BOOL 
ICC_vEE_MPFUNFilter::InscribedCircleCentervEEMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
						   DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
						   DOUBLE dE11_x, DOUBLE dE11_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   InstructionPointer eIPMax/* = ipEnd*/)
{
  APM2D::m_uiInscribedCircleCentervEEMPFUN++;

  switch ( m_eIP )
    {
    case ipStart:
      goto labStart;
    case ipV0onE2:
      goto labV0onE2;
    case ipColinear:
      goto labColinear;
    case ipDen:
      goto labDen;
    case ipEdgeZones:
      goto labEdgeZones;
    default:
      ASSERT( FALSE );
      goto labEnd;
    }

 labStart:
  {
    m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);
    
    *pdC_x = 0.0;
    *pdC_y = 0.0;
    *pdR = 0.0;
    
    m_rV0_E10_x = dV0_E10_x;
    m_rV0_E10_y = dV0_E10_y; 
    m_rE11_x = dE11_x;
    m_rE11_y = dE11_y;
    m_rE20_x = dE20_x;
    m_rE20_y = dE20_y; 
    m_rE21_x = dE21_x;
    m_rE21_y = dE21_y;
    
    //////////////////////////////////////////////////
    // Check for V0 on E2
    
    if ( m_iV0onE2 == -1 )
      {
	mp_real rV0onE2;
	SinTheta(rV0onE2, m_rV0_E10_x, m_rV0_E10_y, m_rE20_x, m_rE20_y, m_rE21_x, m_rE21_y);
	if ( rV0onE2 <= 0.0 )
	  {
	    // V0 is on E2 and it is not one of the endpoints
	    m_eResult = APM2D::icctNull;
	    APM2D::m_uiInscribedCircleCentervEEMPFUNFinal++;
	    return TRUE;
	  }
      }
    
  }
 labV0onE2:
  {
    if ( m_eIP >= eIPMax )
      {
	return FALSE;
      }
    m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);
    
    //////////////////////////////////////////////////
    // Check Colinear
    
    //memset(m_pprE, 0, sizeof(m_pprE));
    m_pprE[1][0] = m_rE11_x - m_rV0_E10_x;
    m_pprE[1][1] = m_rE11_y - m_rV0_E10_y;
    m_pprE[2][0] = m_rE21_x - m_rE20_x;
    m_pprE[2][1] = m_rE21_y - m_rE20_y;
    
    Dot(m_rE1dotE2, m_pprE[1][0], m_pprE[1][1], m_pprE[2][0], m_pprE[2][1]);
    Cross(m_rE1xE2, m_pprE[1][0], m_pprE[1][1], m_pprE[2][0], m_pprE[2][1]);
	
    if ( m_iE1xE2 == -1 || m_iE1dotE2 == -1 )
      {
	if ( m_rE1xE2 == 0.0 && m_rE1dotE2 > 0.0 )
	  {
	    // Colinear
	    m_eResult = APM2D::icctNull;
	    APM2D::m_uiInscribedCircleCentervEEMPFUNFinal++;
	    return TRUE;
	  }
	if ( eIPMax != ipEnd )
	  {
	    if ( m_iE1xE2 == -1 )
	      {
		m_valE1xE2.ConservativeBound(dble(m_rE1xE2));
		m_iE1xE2 = ( m_valE1xE2 == 0.0 );
	      }
	    if ( m_iE1dotE2 == -1 )
	      {
		m_valE1dotE2.ConservativeBound(dble(m_rE1dotE2));
		m_iE1dotE2 = ( m_valE1dotE2 > 0.0 );
	      }
	  }	    
      }

  }
 labColinear:
  {
    if ( m_eIP >= eIPMax )
      {
	return FALSE;
      }
    m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

    //////////////////////////////////////////////////
    // Check Den

    Dot(m_rE1dotE1, m_pprE[1][0], m_pprE[1][1], m_pprE[1][0], m_pprE[1][1]);
    mp_real rE2dotE2;
    Dot(rE2dotE2, m_pprE[2][0], m_pprE[2][1], m_pprE[2][0], m_pprE[2][1]);

    m_rE1E2minusE1dotE2 = sqrt(m_rE1dotE1*rE2dotE2) - m_rE1dotE2;

    m_rDen = m_rE1E2minusE1dotE2;
    if ( m_iDen == -1 )
      {
	ASSERT( m_rDen != 0.0 );
	if ( eIPMax != ipEnd )
	  { 
	    m_valDen.ConservativeBound(dble(m_rDen));	    
	    m_iDen = ( m_valDen != 0.0 );      
	  }
      }
  }
 labDen:
  {
    if ( m_eIP >= eIPMax )
      {
	return FALSE;
      }
    m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

    //////////////////////////////////////////////////
    // Check the Edge Zone

    mp_real ppprV[3][2][2]; // Endpoints
    // Translate V1 to the origin
    //memset(ppprV, 0, sizeof(ppprV));
    ppprV[0][0][0] = 0.0;
    ppprV[0][0][1] = 0.0;
    ppprV[1][0][0] = 0.0;
    ppprV[1][0][1] = 0.0;
    ppprV[1][1][0] = m_rE11_x - m_rV0_E10_x;
    ppprV[1][1][1] = m_rE11_y - m_rV0_E10_y;
    ppprV[2][0][0] = m_rE20_x - m_rV0_E10_x;
    ppprV[2][0][1] = m_rE20_y - m_rV0_E10_y;
    ppprV[2][1][0] = m_rE21_x - m_rV0_E10_x;
    ppprV[2][1][1] = m_rE21_y - m_rV0_E10_y;

    Cross(m_rE20xE21, ppprV[2][0][0], ppprV[2][0][1], ppprV[2][1][0], ppprV[2][1][1]);

    if ( m_piEdgeZone[0] == -1 || 
	 m_piEdgeZone[1] == -1 )
      {      
	mp_real rE20xE21timesE1xE2 = m_rE20xE21*m_rE1xE2;
	    
	mp_real prEdgeZone[2];
	
	if ( m_piEdgeZone[0] == -1 )
	  {
	    mp_real rE2dotE20;
	    Dot(rE2dotE20, m_pprE[2][0], m_pprE[2][1], ppprV[2][0][0], ppprV[2][0][1]);
	    prEdgeZone[0] = -( m_rE1E2minusE1dotE2*rE2dotE20 - rE20xE21timesE1xE2 );
	    if ( prEdgeZone[0] < 0.0 )
	      {
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCentervEEMPFUNFinal++;
		return TRUE;
	      }
	  }

	if ( m_piEdgeZone[1] == -1 )
	  {
	    mp_real rE2dotE21;
	    Dot(rE2dotE21, m_pprE[2][0], m_pprE[2][1], ppprV[2][1][0], ppprV[2][1][1]);
	    prEdgeZone[1] = ( m_rE1E2minusE1dotE2*rE2dotE21 - rE20xE21timesE1xE2 );
	    if ( prEdgeZone[1] < 0.0 )
	      {
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCentervEEMPFUNFinal++;
		return TRUE;
	      }
	  }
      }

  }
 labEdgeZones:
  {
    if ( m_eIP >= eIPMax )
      {
	return FALSE;
      }
    m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

    //////////////////////////////////////////////////
    // Compute the center and radius

    mp_real prC[2];
    prC[0] = -( m_rE20xE21*m_pprE[1][1] ) / m_rDen;
    prC[1] = ( m_rE20xE21*m_pprE[1][0] ) / m_rDen;

    mp_real rR;
    rR = ( sqrt(m_rE1dotE1)*m_rE20xE21 ) / m_rDen;

    // Translate the answer back out away from the origin
    prC[0] += m_rV0_E10_x;
    prC[1] += m_rV0_E10_y;

    *pdC_x = dble(prC[0]);
    *pdC_y = dble(prC[1]);
    *pdR = dble(rR);
    m_eResult = APM2D::icctvEE;
  }
 labEnd:

  APM2D::m_uiInscribedCircleCentervEEMPFUNFinal++;
  return TRUE;
}

#else // COMPILER_PROBLEM

BOOL 
ICC_vEE_MPFUNFilter::InscribedCircleCentervEEMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
						   DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
						   DOUBLE dE11_x, DOUBLE dE11_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   InstructionPointer eIPMax/* = ipEnd*/)
{
  APM2D::m_uiInscribedCircleCentervEEMPFUN++;

  switch ( m_eIP )
    {
    case ipStart:
      {
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);
	
	*pdC_x = 0.0;
	*pdC_y = 0.0;
	*pdR = 0.0;
	
	m_rV0_E10_x = dV0_E10_x;
	m_rV0_E10_y = dV0_E10_y; 
	m_rE11_x = dE11_x;
	m_rE11_y = dE11_y;
	m_rE20_x = dE20_x;
	m_rE20_y = dE20_y; 
	m_rE21_x = dE21_x;
	m_rE21_y = dE21_y;
	
	//////////////////////////////////////////////////
	// Check for V0 on E2
	
	if ( m_iV0onE2 == -1 )
	  {
	    mp_real rV0onE2;
	    SinTheta(rV0onE2, m_rV0_E10_x, m_rV0_E10_y, m_rE20_x, m_rE20_y, m_rE21_x, m_rE21_y);
	    if ( rV0onE2 <= 0.0 )
	      {
		// V0 is on E2 and it is not one of the endpoints
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCentervEEMPFUNFinal++;
		return TRUE;
	      }
	  }
	
      }
    case ipV0onE2:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	if ( Colinear(eIPMax) )
	  {
	    APM2D::m_uiInscribedCircleCentervEEMPFUNFinal++;
	    return TRUE;
	  }
      }
    case ipColinear:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	if ( Den(eIPMax) )
	  {
	    APM2D::m_uiInscribedCircleCentervEEMPFUNFinal++;
	    return TRUE;
	  }
      }
    case ipDen:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	if ( EdgeZones(eIPMax) )
	  {
	    APM2D::m_uiInscribedCircleCentervEEMPFUNFinal++;
	    return TRUE;
	  }
      }
    case ipEdgeZones:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	CenterAndRadius(pdC_x, pdC_y, pdR, 
			eIPMax);
      }
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  APM2D::m_uiInscribedCircleCentervEEMPFUNFinal++;
  return TRUE;
}

BOOL 
ICC_vEE_MPFUNFilter::Colinear(InstructionPointer eIPMax)
{
  //////////////////////////////////////////////////
  // Check Colinear
  
  //memset(m_pprE, 0, sizeof(m_pprE));
  m_pprE[1][0] = m_rE11_x - m_rV0_E10_x;
  m_pprE[1][1] = m_rE11_y - m_rV0_E10_y;
  m_pprE[2][0] = m_rE21_x - m_rE20_x;
  m_pprE[2][1] = m_rE21_y - m_rE20_y;
  
  Dot(m_rE1dotE2, m_pprE[1][0], m_pprE[1][1], m_pprE[2][0], m_pprE[2][1]);
  Cross(m_rE1xE2, m_pprE[1][0], m_pprE[1][1], m_pprE[2][0], m_pprE[2][1]);
  
  if ( m_iE1xE2 == -1 || m_iE1dotE2 == -1 )
    {
      if ( m_rE1xE2 == 0.0 && m_rE1dotE2 > 0.0 )
	{
	  // Colinear
	  m_eResult = APM2D::icctNull;
	  return TRUE;
	}
      if ( eIPMax != ipEnd )
	{
	  if ( m_iE1xE2 == -1 )
	    {
	      m_valE1xE2.ConservativeBound(dble(m_rE1xE2));
	      m_iE1xE2 = ( m_valE1xE2 == 0.0 );
	    }
	  if ( m_iE1dotE2 == -1 )
	    {
	      m_valE1dotE2.ConservativeBound(dble(m_rE1dotE2));
	      m_iE1dotE2 = ( m_valE1dotE2 > 0.0 );
	    }
	}	    
    }
 
  return FALSE;
}


BOOL 
ICC_vEE_MPFUNFilter::Den(InstructionPointer eIPMax)
{
  //////////////////////////////////////////////////
  // Check Den
  
  Dot(m_rE1dotE1, m_pprE[1][0], m_pprE[1][1], m_pprE[1][0], m_pprE[1][1]);
  mp_real rE2dotE2;
  Dot(rE2dotE2, m_pprE[2][0], m_pprE[2][1], m_pprE[2][0], m_pprE[2][1]);

  m_rE1E2minusE1dotE2 = sqrt(m_rE1dotE1*rE2dotE2) - m_rE1dotE2;
  
  m_rDen = m_rE1E2minusE1dotE2;
  if ( m_iDen == -1 )
    {
      ASSERT( m_rDen != 0.0 );
      if ( eIPMax != ipEnd )
	{ 
	  m_valDen.ConservativeBound(dble(m_rDen));	    
	  m_iDen = ( m_valDen != 0.0 );      
	}
    }

  return FALSE;
}

BOOL 
ICC_vEE_MPFUNFilter::EdgeZones(InstructionPointer /*eIPMax*/)
{
  //////////////////////////////////////////////////
  // Check the Edge Zone
  
  mp_real ppprV[3][2][2]; // Endpoints
  // Translate V1 to the origin
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = 0.0;
  ppprV[0][0][1] = 0.0;
  ppprV[1][0][0] = 0.0;
  ppprV[1][0][1] = 0.0;
  ppprV[1][1][0] = m_rE11_x - m_rV0_E10_x;
  ppprV[1][1][1] = m_rE11_y - m_rV0_E10_y;
  ppprV[2][0][0] = m_rE20_x - m_rV0_E10_x;
  ppprV[2][0][1] = m_rE20_y - m_rV0_E10_y;
  ppprV[2][1][0] = m_rE21_x - m_rV0_E10_x;
  ppprV[2][1][1] = m_rE21_y - m_rV0_E10_y;
  
  Cross(m_rE20xE21, ppprV[2][0][0], ppprV[2][0][1], ppprV[2][1][0], ppprV[2][1][1]);
  
  if ( m_piEdgeZone[0] == -1 || 
       m_piEdgeZone[1] == -1 )
    {      
      mp_real rE20xE21timesE1xE2 = m_rE20xE21*m_rE1xE2;
      
      mp_real prEdgeZone[2];
      
      if ( m_piEdgeZone[0] == -1 )
	{
	  mp_real rE2dotE20;
	  Dot(rE2dotE20, m_pprE[2][0], m_pprE[2][1], ppprV[2][0][0], ppprV[2][0][1]);
	  prEdgeZone[0] = -( m_rE1E2minusE1dotE2*rE2dotE20 - rE20xE21timesE1xE2 );
	  if ( prEdgeZone[0] < 0.0 )
	    {
	      m_eResult = APM2D::icctNull;
	      return TRUE;
	    }
	}
      
      if ( m_piEdgeZone[1] == -1 )
	{
	  mp_real rE2dotE21;
	  Dot(rE2dotE21, m_pprE[2][0], m_pprE[2][1], ppprV[2][1][0], ppprV[2][1][1]);
	  prEdgeZone[1] = ( m_rE1E2minusE1dotE2*rE2dotE21 - rE20xE21timesE1xE2 );
	  if ( prEdgeZone[1] < 0.0 )
	    {
	      m_eResult = APM2D::icctNull;
	      return TRUE;
	    }
	}
    }
 
  return FALSE;
}

VOID 
ICC_vEE_MPFUNFilter::CenterAndRadius(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				     InstructionPointer /*eIPMax*/)
{
  //////////////////////////////////////////////////
  // Compute the center and radius
  
  mp_real prC[2];
  prC[0] = -( m_rE20xE21*m_pprE[1][1] ) / m_rDen;
  prC[1] = ( m_rE20xE21*m_pprE[1][0] ) / m_rDen;
  
  mp_real rR;
  rR = ( sqrt(m_rE1dotE1)*m_rE20xE21 ) / m_rDen;

  // Translate the answer back out away from the origin
  prC[0] += m_rV0_E10_x;
  prC[1] += m_rV0_E10_y;
  
  *pdC_x = dble(prC[0]);
  *pdC_y = dble(prC[1]);
  *pdR = dble(rR);
  m_eResult = APM2D::icctvEE;
}

#endif // COMPILER_PROBLEM

//static 
APM2D::InscribedCircleCenterType 
InscribedCircleCentervEEMPFUNFilter(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				    DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
				    DOUBLE dE11_x, DOUBLE dE11_y,
				    DOUBLE dE20_x, DOUBLE dE20_y, 
				    DOUBLE dE21_x, DOUBLE dE21_y)
{
  ICC_vEE_MPFUNFilter iccvEE;

  return iccvEE.InscribedCircleCentervEE(pdC_x, pdC_y, pdR, 
					 dV0_E10_x, dV0_E10_y, 
					 dE11_x, dE11_y,
					 dE20_x, dE20_y, 
					 dE21_x, dE21_y);
}

//static 
APM2D::EdgeZoneType EdgeZonevEEMPFUNFilter(DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
					   DOUBLE dE11_x, DOUBLE dE11_y,
					   DOUBLE dE20_x, DOUBLE dE20_y, 
					   DOUBLE dE21_x, DOUBLE dE21_y,
					   DOUBLE dE30_x, DOUBLE dE30_y,
					   DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::m_uiEdgeZonevEEMPFUNFilter++;

  INT_32 piEdgeZone[2];
  piEdgeZone[0] = -1;
  piEdgeZone[1] = -1;

  switch ( 0 )
  {
  default:
    Interval valV0_E10_x = dV0_E10_x;
    Interval valV0_E10_y = dV0_E10_y; 
    Interval valE11_x = dE11_x;
    Interval valE11_y = dE11_y;
    Interval valE20_x = dE20_x;
    Interval valE20_y = dE20_y; 
    Interval valE21_x = dE21_x;
    Interval valE21_y = dE21_y;
    Interval valE30_x = dE30_x;
    Interval valE30_y = dE30_y;
    Interval valE31_x = dE31_x;
    Interval valE31_y = dE31_y;

    Interval pppvalV[4][2][2]; // Endpoints
    //memset(pppvalV, 0, sizeof(pppvalV));
    pppvalV[0][0][0] = 0.0;
    pppvalV[0][0][1] = 0.0;
    pppvalV[1][0][0] = 0.0;
    pppvalV[1][0][1] = 0.0;
    pppvalV[1][1][0] = valE11_x - valV0_E10_x;
    pppvalV[1][1][1] = valE11_y - valV0_E10_y;
    pppvalV[2][0][0] = valE20_x - valV0_E10_x;
    pppvalV[2][0][1] = valE20_y - valV0_E10_y;
    pppvalV[2][1][0] = valE21_x - valV0_E10_x;
    pppvalV[2][1][1] = valE21_y - valV0_E10_y;
    pppvalV[3][0][0] = valE30_x - valV0_E10_x;
    pppvalV[3][0][1] = valE30_y - valV0_E10_y;
    pppvalV[3][1][0] = valE31_x - valV0_E10_x;
    pppvalV[3][1][1] = valE31_y - valV0_E10_y;
  
    Interval ppvalE[4][2];     // Edge vectors
    //memset(ppvalE, 0, sizeof(ppvalE));
    ppvalE[1][0] = valE11_x - valV0_E10_x;
    ppvalE[1][1] = valE11_y - valV0_E10_y;
    ppvalE[2][0] = valE21_x - valE20_x;
    ppvalE[2][1] = valE21_y - valE20_y;
    ppvalE[3][0] = valE31_x - valE30_x;
    ppvalE[3][1] = valE31_y - valE30_y;
  
    Interval valE1dotE1;
    Dot(valE1dotE1, ppvalE[1][0], ppvalE[1][1], ppvalE[1][0], ppvalE[1][1]);
    Interval valE2dotE2;
    Dot(valE2dotE2, ppvalE[2][0], ppvalE[2][1], ppvalE[2][0], ppvalE[2][1]);
    Interval valE1dotE2;
    Dot(valE1dotE2, ppvalE[1][0], ppvalE[1][1], ppvalE[2][0], ppvalE[2][1]);

    Interval valE1xE3;
    Cross(valE1xE3, ppvalE[1][0], ppvalE[1][1], ppvalE[3][0], ppvalE[3][1]);
    Interval valE20xE21;
    Cross(valE20xE21, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][1][0], pppvalV[2][1][1]);

    Interval valE1E2minusE1dotE2;
    valE1E2minusE1dotE2 = sqrt(valE1dotE1*valE2dotE2) - valE1dotE2;

    Interval valE20xE21timesE1xE3 = valE20xE21*valE1xE3;

    Interval pvalEdgeZone[2];

    Interval valE3dotE30;
    Dot(valE3dotE30, ppvalE[3][0], ppvalE[3][1], pppvalV[3][0][0], pppvalV[3][0][1]);
    pvalEdgeZone[0] = ( valE1E2minusE1dotE2*valE3dotE30 - valE20xE21timesE1xE3 );
    piEdgeZone[0] = ( pvalEdgeZone[0] >= 0.0 );
    if ( piEdgeZone[0] == 1 )  
      {
	return APM2D::eztStart;
      }

    Interval valE3dotE31;
    Dot(valE3dotE31, ppvalE[3][0], ppvalE[3][1], pppvalV[3][1][0], pppvalV[3][1][1]);
    pvalEdgeZone[1] = -( valE1E2minusE1dotE2*valE3dotE31 - valE20xE21timesE1xE3 );
    piEdgeZone[1] = ( pvalEdgeZone[1] >= 0.0 );
    if ( piEdgeZone[1] == 1 )
      {
	return APM2D::eztEnd;
      }
  }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( piEdgeZone[0] == -1 || piEdgeZone[1] == -1 )
    {
      APM2D::m_uiEdgeZonevEEMPFUN++;

      mp_real rV0_E10_x = dV0_E10_x;
      mp_real rV0_E10_y = dV0_E10_y; 
      mp_real rE11_x = dE11_x;
      mp_real rE11_y = dE11_y;
      mp_real rE20_x = dE20_x;
      mp_real rE20_y = dE20_y; 
      mp_real rE21_x = dE21_x;
      mp_real rE21_y = dE21_y;
      mp_real rE30_x = dE30_x;
      mp_real rE30_y = dE30_y;
      mp_real rE31_x = dE31_x;
      mp_real rE31_y = dE31_y;

      mp_real ppprV[4][2][2]; // Endpoints
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = 0.0;
      ppprV[0][0][1] = 0.0;
      ppprV[1][0][0] = 0.0;
      ppprV[1][0][1] = 0.0;
      ppprV[1][1][0] = rE11_x - rV0_E10_x;
      ppprV[1][1][1] = rE11_y - rV0_E10_y;
      ppprV[2][0][0] = rE20_x - rV0_E10_x;
      ppprV[2][0][1] = rE20_y - rV0_E10_y;
      ppprV[2][1][0] = rE21_x - rV0_E10_x;
      ppprV[2][1][1] = rE21_y - rV0_E10_y;
      ppprV[3][0][0] = rE30_x - rV0_E10_x;
      ppprV[3][0][1] = rE30_y - rV0_E10_y;
      ppprV[3][1][0] = rE31_x - rV0_E10_x;
      ppprV[3][1][1] = rE31_y - rV0_E10_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[1][0] = rE11_x - rV0_E10_x;
      pprE[1][1] = rE11_y - rV0_E10_y;
      pprE[2][0] = rE21_x - rE20_x;
      pprE[2][1] = rE21_y - rE20_y;
      pprE[3][0] = rE31_x - rE30_x;
      pprE[3][1] = rE31_y - rE30_y;
  
      mp_real rE1dotE1;
      Dot(rE1dotE1, pprE[1][0], pprE[1][1], pprE[1][0], pprE[1][1]);
      mp_real rE2dotE2;
      Dot(rE2dotE2, pprE[2][0], pprE[2][1], pprE[2][0], pprE[2][1]);
      mp_real rE1dotE2;
      Dot(rE1dotE2, pprE[1][0], pprE[1][1], pprE[2][0], pprE[2][1]);

      mp_real rE1xE3;
      Cross(rE1xE3, pprE[1][0], pprE[1][1], pprE[3][0], pprE[3][1]);
      mp_real rE20xE21;
      Cross(rE20xE21, ppprV[2][0][0], ppprV[2][0][1], ppprV[2][1][0], ppprV[2][1][1]);

      mp_real rE1E2minusE1dotE2;
      rE1E2minusE1dotE2 = sqrt(rE1dotE1*rE2dotE2) - rE1dotE2;

      mp_real rE20xE21timesE1xE3 = rE20xE21*rE1xE3;

      mp_real prEdgeZone[2];

      if ( piEdgeZone[0] == -1 )
	{
	  mp_real rE3dotE30;
	  Dot(rE3dotE30, pprE[3][0], pprE[3][1], ppprV[3][0][0], ppprV[3][0][1]);
	  prEdgeZone[0] = ( rE1E2minusE1dotE2*rE3dotE30 - rE20xE21timesE1xE3 );
	  if ( prEdgeZone[0] >= 0.0 )
	    {
	      return APM2D::eztStart;
	    }
	}

      if ( piEdgeZone[1] == -1 )
	{
	  mp_real rE3dotE31;
	  Dot(rE3dotE31, pprE[3][0], pprE[3][1], ppprV[3][1][0], ppprV[3][1][1]);
	  prEdgeZone[1] = -( rE1E2minusE1dotE2*rE3dotE31 - rE20xE21timesE1xE3 );
	  if ( prEdgeZone[1] >= 0.0 )
	    {
	      return APM2D::eztEnd;
	    }
	}
    }

  return APM2D::eztMiddle;
}

//static 
APM2D::CircleTestType InCirclevEE_VMPFUNFilter(DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
					 DOUBLE dE11_x, DOUBLE dE11_y,
					 DOUBLE dE20_x, DOUBLE dE20_y, 
					 DOUBLE dE21_x, DOUBLE dE21_y,
					 DOUBLE dV3_x, DOUBLE dV3_y)
{
  APM2D::m_uiInCirclevEE_VMPFUNFilter++;

  INT_32 iResult;

  iResult = -1;

  switch ( 0 )
    {
    default:
      Interval valV0_E10_x = dV0_E10_x;
      Interval valV0_E10_y = dV0_E10_y; 
      Interval valE11_x = dE11_x;
      Interval valE11_y = dE11_y;
      Interval valE20_x = dE20_x;
      Interval valE20_y = dE20_y; 
      Interval valE21_x = dE21_x;
      Interval valE21_y = dE21_y;
      Interval valV3_x = dV3_x;
      Interval valV3_y = dV3_y;

      Interval pppvalV[4][2][2]; // Endpoints
      //memset(pppvalV, 0, sizeof(pppvalV));
      pppvalV[0][0][0] = 0.0;
      pppvalV[0][0][1] = 0.0;
      pppvalV[1][0][0] = 0.0;
      pppvalV[1][0][1] = 0.0;
      pppvalV[1][1][0] = valE11_x - valV0_E10_x;
      pppvalV[1][1][1] = valE11_y - valV0_E10_y;
      pppvalV[2][0][0] = valE20_x - valV0_E10_x;
      pppvalV[2][0][1] = valE20_y - valV0_E10_y;
      pppvalV[2][1][0] = valE21_x - valV0_E10_x;
      pppvalV[2][1][1] = valE21_y - valV0_E10_y;
      pppvalV[3][0][0] = valV3_x - valV0_E10_x;
      pppvalV[3][0][1] = valV3_y - valV0_E10_y;
  
      Interval ppvalE[4][2];     // Edge vectors
      //memset(ppvalE, 0, sizeof(ppvalE));
      ppvalE[1][0] = valE11_x - valV0_E10_x;
      ppvalE[1][1] = valE11_y - valV0_E10_y;
      ppvalE[2][0] = valE21_x - valE20_x;
      ppvalE[2][1] = valE21_y - valE20_y;

      Interval valE1dotE1;
      Dot(valE1dotE1, ppvalE[1][0], ppvalE[1][1], ppvalE[1][0], ppvalE[1][1]);
      Interval valE2dotE2;
      Dot(valE2dotE2, ppvalE[2][0], ppvalE[2][1], ppvalE[2][0], ppvalE[2][1]);
      Interval valE1dotE2;
      Dot(valE1dotE2, ppvalE[1][0], ppvalE[1][1], ppvalE[2][0], ppvalE[2][1]);
      Interval valV3dotV3;
      Dot(valV3dotV3, pppvalV[3][0][0], pppvalV[3][0][1], pppvalV[3][0][0], pppvalV[3][0][1]);

      Interval valE1xV3;
      Cross(valE1xV3, ppvalE[1][0], ppvalE[1][1], pppvalV[3][0][0], pppvalV[3][0][1]);
      Interval valE20xE21;
      Cross(valE20xE21, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][1][0], pppvalV[2][1][1]);

      Interval valE1E2minusE1dotE2;
      valE1E2minusE1dotE2 = sqrt(valE1dotE1*valE2dotE2) - valE1dotE2;

      Interval valResult;
      valResult = ( valE1E2minusE1dotE2*valV3dotV3 - 2.0*valE20xE21*valE1xV3 );

      iResult = ( valResult < 0.0 );
      if ( iResult == -1 )
	{
	  break;
	}
      if ( iResult == 1 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( valResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
      return APM2D::cttOnCircle;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( iResult == -1 )
    {
      APM2D::m_uiInCirclevEE_VMPFUN++;

      mp_real rV0_E10_x = dV0_E10_x;
      mp_real rV0_E10_y = dV0_E10_y; 
      mp_real rE11_x = dE11_x;
      mp_real rE11_y = dE11_y;
      mp_real rE20_x = dE20_x;
      mp_real rE20_y = dE20_y; 
      mp_real rE21_x = dE21_x;
      mp_real rE21_y = dE21_y;
      mp_real rV3_x = dV3_x;
      mp_real rV3_y = dV3_y;

      mp_real ppprV[4][2][2]; // Endpoints
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = 0.0;
      ppprV[0][0][1] = 0.0;
      ppprV[1][0][0] = 0.0;
      ppprV[1][0][1] = 0.0;
      ppprV[1][1][0] = rE11_x - rV0_E10_x;
      ppprV[1][1][1] = rE11_y - rV0_E10_y;
      ppprV[2][0][0] = rE20_x - rV0_E10_x;
      ppprV[2][0][1] = rE20_y - rV0_E10_y;
      ppprV[2][1][0] = rE21_x - rV0_E10_x;
      ppprV[2][1][1] = rE21_y - rV0_E10_y;
      ppprV[3][0][0] = rV3_x - rV0_E10_x;
      ppprV[3][0][1] = rV3_y - rV0_E10_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[1][0] = rE11_x - rV0_E10_x;
      pprE[1][1] = rE11_y - rV0_E10_y;
      pprE[2][0] = rE21_x - rE20_x;
      pprE[2][1] = rE21_y - rE20_y;

      mp_real rE1dotE1;
      Dot(rE1dotE1, pprE[1][0], pprE[1][1], pprE[1][0], pprE[1][1]);
      mp_real rE2dotE2;
      Dot(rE2dotE2, pprE[2][0], pprE[2][1], pprE[2][0], pprE[2][1]);
      mp_real rE1dotE2;
      Dot(rE1dotE2, pprE[1][0], pprE[1][1], pprE[2][0], pprE[2][1]);
      mp_real rV3dotV3;
      Dot(rV3dotV3, ppprV[3][0][0], ppprV[3][0][1], ppprV[3][0][0], ppprV[3][0][1]);

      mp_real rE1xV3;
      Cross(rE1xV3, pprE[1][0], pprE[1][1], ppprV[3][0][0], ppprV[3][0][1]);
      mp_real rE20xE21;
      Cross(rE20xE21, ppprV[2][0][0], ppprV[2][0][1], ppprV[2][1][0], ppprV[2][1][1]);

      mp_real rE1E2minusE1dotE2;
      rE1E2minusE1dotE2 = sqrt(rE1dotE1*rE2dotE2) - rE1dotE2;

      mp_real rResult;
      rResult = ( rE1E2minusE1dotE2*rV3dotV3 - 2.0*rE20xE21*rE1xV3 );
  
      if ( rResult < 0.0 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( rResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
    }

  return APM2D::cttOnCircle;
}

//static 
APM2D::CircleTestType InCirclevEE_EMPFUNFilter(DOUBLE dV0_E10_x, DOUBLE dV0_E10_y, 
					 DOUBLE dE11_x, DOUBLE dE11_y,
					 DOUBLE dE20_x, DOUBLE dE20_y, 
					 DOUBLE dE21_x, DOUBLE dE21_y,
					 DOUBLE dE30_x, DOUBLE dE30_y,
					 DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::m_uiInCirclevEE_EMPFUNFilter++;

  INT_32 iResult;

  iResult = -1;

  switch ( 0 )
    {
    default:
      Interval valV0_E10_x = dV0_E10_x;
      Interval valV0_E10_y = dV0_E10_y; 
      Interval valE11_x = dE11_x;
      Interval valE11_y = dE11_y;
      Interval valE20_x = dE20_x;
      Interval valE20_y = dE20_y; 
      Interval valE21_x = dE21_x;
      Interval valE21_y = dE21_y;
      Interval valE30_x = dE30_x;
      Interval valE30_y = dE30_y;
      Interval valE31_x = dE31_x;
      Interval valE31_y = dE31_y;

      Interval pppvalV[4][2][2]; // Endpoints
      //memset(pppvalV, 0, sizeof(pppvalV));
      pppvalV[0][0][0] = 0.0;
      pppvalV[0][0][1] = 0.0;
      pppvalV[1][0][0] = 0.0;
      pppvalV[1][0][1] = 0.0;
      pppvalV[1][1][0] = valE11_x - valV0_E10_x;
      pppvalV[1][1][1] = valE11_y - valV0_E10_y;
      pppvalV[2][0][0] = valE20_x - valV0_E10_x;
      pppvalV[2][0][1] = valE20_y - valV0_E10_y;
      pppvalV[2][1][0] = valE21_x - valV0_E10_x;
      pppvalV[2][1][1] = valE21_y - valV0_E10_y;
      pppvalV[3][0][0] = valE30_x - valV0_E10_x;
      pppvalV[3][0][1] = valE30_y - valV0_E10_y;
      pppvalV[3][1][0] = valE31_x - valV0_E10_x;
      pppvalV[3][1][1] = valE31_y - valV0_E10_y;
  
      Interval ppvalE[4][2];     // Edge vectors
      //memset(ppvalE, 0, sizeof(ppvalE));
      ppvalE[1][0] = valE11_x - valV0_E10_x;
      ppvalE[1][1] = valE11_y - valV0_E10_y;
      ppvalE[2][0] = valE21_x - valE20_x;
      ppvalE[2][1] = valE21_y - valE20_y;
      ppvalE[3][0] = valE31_x - valE30_x;
      ppvalE[3][1] = valE31_y - valE30_y;
  
      Interval valE1dotE1;
      Dot(valE1dotE1, ppvalE[1][0], ppvalE[1][1], ppvalE[1][0], ppvalE[1][1]);
      Interval valE2dotE2;
      Dot(valE2dotE2, ppvalE[2][0], ppvalE[2][1], ppvalE[2][0], ppvalE[2][1]);
      Interval valE1dotE2;
      Dot(valE1dotE2, ppvalE[1][0], ppvalE[1][1], ppvalE[2][0], ppvalE[2][1]);
      Interval valE1dotE3;
      Dot(valE1dotE3, ppvalE[1][0], ppvalE[1][1], ppvalE[3][0], ppvalE[3][1]);
      Interval valE3dotE3;
      Dot(valE3dotE3, ppvalE[3][0], ppvalE[3][1], ppvalE[3][0], ppvalE[3][1]);

      Interval valE1xE3;
      Cross(valE1xE3, ppvalE[1][0], ppvalE[1][1], ppvalE[3][0], ppvalE[3][1]);
      Interval valE20xE21;
      Cross(valE20xE21, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][1][0], pppvalV[2][1][1]);
      Interval valE30xE31;
      Cross(valE30xE31, pppvalV[3][0][0], pppvalV[3][0][1], pppvalV[3][1][0], pppvalV[3][1][1]);

      Interval valE1E2minusE1dotE2;
      valE1E2minusE1dotE2 = sqrt(valE1dotE1*valE2dotE2) - valE1dotE2;

      Interval valResult;
      valResult = ( valE1E2minusE1dotE2*valE30xE31*( valE1E2minusE1dotE2*valE30xE31 
					       + 2.0*valE20xE21*valE1dotE3 )
		  - valE20xE21*valE20xE21*valE1xE3*valE1xE3 );

      iResult = ( valResult < 0.0 );
      if ( iResult == -1 )
	{
	  break;
	}
      if ( iResult == 1 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( valResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
      return APM2D::cttOnCircle;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( iResult == -1 )
    {
      APM2D::m_uiInCirclevEE_EMPFUN++;

      mp_real rV0_E10_x = dV0_E10_x;
      mp_real rV0_E10_y = dV0_E10_y; 
      mp_real rE11_x = dE11_x;
      mp_real rE11_y = dE11_y;
      mp_real rE20_x = dE20_x;
      mp_real rE20_y = dE20_y; 
      mp_real rE21_x = dE21_x;
      mp_real rE21_y = dE21_y;
      mp_real rE30_x = dE30_x;
      mp_real rE30_y = dE30_y;
      mp_real rE31_x = dE31_x;
      mp_real rE31_y = dE31_y;

      mp_real ppprV[4][2][2]; // Endpoints
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = 0.0;
      ppprV[0][0][1] = 0.0;
      ppprV[1][0][0] = 0.0;
      ppprV[1][0][1] = 0.0;
      ppprV[1][1][0] = rE11_x - rV0_E10_x;
      ppprV[1][1][1] = rE11_y - rV0_E10_y;
      ppprV[2][0][0] = rE20_x - rV0_E10_x;
      ppprV[2][0][1] = rE20_y - rV0_E10_y;
      ppprV[2][1][0] = rE21_x - rV0_E10_x;
      ppprV[2][1][1] = rE21_y - rV0_E10_y;
      ppprV[3][0][0] = rE30_x - rV0_E10_x;
      ppprV[3][0][1] = rE30_y - rV0_E10_y;
      ppprV[3][1][0] = rE31_x - rV0_E10_x;
      ppprV[3][1][1] = rE31_y - rV0_E10_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[1][0] = rE11_x - rV0_E10_x;
      pprE[1][1] = rE11_y - rV0_E10_y;
      pprE[2][0] = rE21_x - rE20_x;
      pprE[2][1] = rE21_y - rE20_y;
      pprE[3][0] = rE31_x - rE30_x;
      pprE[3][1] = rE31_y - rE30_y;
  
      mp_real rE1dotE1;
      Dot(rE1dotE1, pprE[1][0], pprE[1][1], pprE[1][0], pprE[1][1]);
      mp_real rE2dotE2;
      Dot(rE2dotE2, pprE[2][0], pprE[2][1], pprE[2][0], pprE[2][1]);
      mp_real rE1dotE2;
      Dot(rE1dotE2, pprE[1][0], pprE[1][1], pprE[2][0], pprE[2][1]);
      mp_real rE1dotE3;
      Dot(rE1dotE3, pprE[1][0], pprE[1][1], pprE[3][0], pprE[3][1]);
      mp_real rE3dotE3;
      Dot(rE3dotE3, pprE[3][0], pprE[3][1], pprE[3][0], pprE[3][1]);

      mp_real rE1xE3;
      Cross(rE1xE3, pprE[1][0], pprE[1][1], pprE[3][0], pprE[3][1]);
      mp_real rE20xE21;
      Cross(rE20xE21, ppprV[2][0][0], ppprV[2][0][1], ppprV[2][1][0], ppprV[2][1][1]);
      mp_real rE30xE31;
      Cross(rE30xE31, ppprV[3][0][0], ppprV[3][0][1], ppprV[3][1][0], ppprV[3][1][1]);

      mp_real rE1E2minusE1dotE2;
      rE1E2minusE1dotE2 = sqrt(rE1dotE1*rE2dotE2) - rE1dotE2;

      mp_real rResult;
      rResult = ( rE1E2minusE1dotE2*rE30xE31*( rE1E2minusE1dotE2*rE30xE31 
					       + 2.0*rE20xE21*rE1dotE3 )
		  - rE20xE21*rE20xE21*rE1xE3*rE1xE3 );
  
      if ( rResult < 0.0 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( rResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
    }

  return APM2D::cttOnCircle;
}

//////////////////////////////////////////////////
// EEv

class ICC_EEv_MPFUNFilter
{
protected:
  enum InstructionPointer {ipStart, ipV2onE0, 
			   ipColinear, ipDen, 
			   ipEdgeZones, ipEnd};

public:
  ICC_EEv_MPFUNFilter();
  ~ICC_EEv_MPFUNFilter();

public:
  APM2D::InscribedCircleCenterType 
  InscribedCircleCenterEEv(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			   DOUBLE dE00_x, DOUBLE dE00_y, 
			   DOUBLE dE01_x, DOUBLE dE01_y,
			   DOUBLE dE10_x, DOUBLE dE10_y, 
			   DOUBLE dE11_V2_x, DOUBLE dE11_V2_y);
  APM2D::InscribedCircleCenterType 
  InscribedCircleCenterEEvForce(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dE00_x, DOUBLE dE00_y, 
				DOUBLE dE01_x, DOUBLE dE01_y,
				DOUBLE dE10_x, DOUBLE dE10_y, 
				DOUBLE dE11_V2_x, DOUBLE dE11_V2_y);

protected:
#ifndef COMPILER_PROBLEM
  BOOL
  InscribedCircleCenterEEvMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dE00_x, DOUBLE dE00_y, 
				DOUBLE dE01_x, DOUBLE dE01_y,
				DOUBLE dE10_x, DOUBLE dE10_y, 
				DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
				InstructionPointer eIPMax = ipEnd);
#else // COMPILER_PROBLEM
  BOOL
  InscribedCircleCenterEEvMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dE00_x, DOUBLE dE00_y, 
				DOUBLE dE01_x, DOUBLE dE01_y,
				DOUBLE dE10_x, DOUBLE dE10_y, 
				DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
				InstructionPointer eIPMax = ipEnd);

  BOOL Colinear(InstructionPointer eIPMax);
  BOOL Den(InstructionPointer eIPMax);
  BOOL EdgeZones(InstructionPointer eIPMax);
  /*
  VOID CenterAndRadius(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
		       InstructionPointer eIPMax);
  */
#endif // COMPILER_PROBLEM

protected:
  InstructionPointer m_eIP;
  APM2D::InscribedCircleCenterType m_eResult;
  INT_32 m_iV2onE0;
  INT_32 m_iE0xE1;
  INT_32 m_iE0dotE1;
  INT_32 m_iDen;
  INT_32 m_piEdgeZone[2];

  Interval m_valE0xE1;
  Interval m_valE0dotE1;
  Interval m_valDen;

  mp_real m_rE00_x;
  mp_real m_rE00_y;
  mp_real m_rE01_x;
  mp_real m_rE01_y;
  mp_real m_rE10_x;
  mp_real m_rE10_y;
  mp_real m_rE11_V2_x;
  mp_real m_rE11_V2_y;
  mp_real m_pprE[3][2];
  mp_real m_rE0dotE1;
  mp_real m_rE0xE1;
  mp_real m_rE1dotE1;
  mp_real m_rE0E1minusE0dotE1;
  mp_real m_rDen;
  mp_real m_rE00xE01;
};

ICC_EEv_MPFUNFilter::ICC_EEv_MPFUNFilter()
{
}

ICC_EEv_MPFUNFilter::~ICC_EEv_MPFUNFilter()
{
}

APM2D::InscribedCircleCenterType 
ICC_EEv_MPFUNFilter::InscribedCircleCenterEEv(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					      DOUBLE dE00_x, DOUBLE dE00_y, 
					      DOUBLE dE01_x, DOUBLE dE01_y,
					      DOUBLE dE10_x, DOUBLE dE10_y, 
					      DOUBLE dE11_V2_x, DOUBLE dE11_V2_y)
{
  APM2D::m_uiInscribedCircleCenterEEvMPFUNFilter++;

  *pdC_x = 0.0;
  *pdC_y = 0.0;
  *pdR = 0.0;
  
  m_eIP = ipStart;
  m_eResult = APM2D::icctNull;
  m_iV2onE0 = -1;
  m_iE0xE1 = -1;
  m_iE0dotE1 = -1;
  m_iDen = -1;
  m_piEdgeZone[0] = -1;
  m_piEdgeZone[1] = -1;

  BOOL bDone;

  Interval valE00_x = dE00_x;
  Interval valE00_y = dE00_y; 
  Interval valE01_x = dE01_x;
  Interval valE01_y = dE01_y;
  Interval valE10_x = dE10_x;
  Interval valE10_y = dE10_y; 
  Interval valE11_V2_x = dE11_V2_x;
  Interval valE11_V2_y = dE11_V2_y;
  
  Interval valV2onE0;
  SinTheta(valV2onE0, valE11_V2_x, valE11_V2_y, valE00_x, valE00_y, valE01_x, valE01_y);
  m_iV2onE0 = ( valV2onE0 <= 0.0 );
  if ( m_iV2onE0 == 1 )
    {
      // V2 is on E0 and it is not one of the endpoints
      return APM2D::icctNull;
    }
  if ( m_iV2onE0 == -1 )
    {
      bDone = InscribedCircleCenterEEvMPFUN(pdC_x, pdC_y, pdR, 
					    dE00_x, dE00_y, 
					    dE01_x, dE01_y,
					    dE10_x, dE10_y, 
					    dE11_V2_x, dE11_V2_y,
					    ipV2onE0);
      if ( bDone )
	{
	  return m_eResult;
	}
    }
	
  //////////////////////////////////////////////////
  // Check Colinear
    
  Interval ppvalE[3][2];     // Edge vectors
  //memset(ppvalE, 0, sizeof(ppvalE));
  ppvalE[0][0] = valE01_x - valE00_x;
  ppvalE[0][1] = valE01_y - valE00_y;
  ppvalE[1][0] = valE11_V2_x - valE10_x;
  ppvalE[1][1] = valE11_V2_y - valE10_y;
  
  Dot(m_valE0dotE1, ppvalE[0][0], ppvalE[0][1], ppvalE[1][0], ppvalE[1][1]);
  Cross(m_valE0xE1, ppvalE[0][0], ppvalE[0][1], ppvalE[1][0], ppvalE[1][1]);
  
  m_iE0xE1 = ( m_valE0xE1 == 0.0 );
  m_iE0dotE1 = ( m_valE0dotE1 > 0.0 );
  if ( m_iE0xE1 == -1 || m_iE0dotE1 == -1 )
    {
      bDone = InscribedCircleCenterEEvMPFUN(pdC_x, pdC_y, pdR, 
					    dE00_x, dE00_y, 
					    dE01_x, dE01_y,
					    dE10_x, dE10_y, 
					    dE11_V2_x, dE11_V2_y,
					    ipColinear);
      if ( bDone )
	{
	  return m_eResult;
	}
      if ( m_iE0xE1 == -1 || m_iE0dotE1 == -1 )
	{
	  bDone = InscribedCircleCenterEEvMPFUN(pdC_x, pdC_y, pdR, 
						dE00_x, dE00_y, 
						dE01_x, dE01_y,
						dE10_x, dE10_y, 
						dE11_V2_x, dE11_V2_y,
						ipEnd);
	  ASSERT( bDone );
	  return m_eResult;
	}
    }
  if ( m_iE0xE1 == 1 && m_iE0dotE1 == 1 )
    {
      // Colinear
      return APM2D::icctNull;
    }
	
  //////////////////////////////////////////////////
  // Check Den
  
  Interval valE0dotE0;
  Dot(valE0dotE0, ppvalE[0][0], ppvalE[0][1], ppvalE[0][0], ppvalE[0][1]);
  Interval valE1dotE1;
  Dot(valE1dotE1, ppvalE[1][0], ppvalE[1][1], ppvalE[1][0], ppvalE[1][1]);
  
  Interval valE0E1minusE0dotE1;
  valE0E1minusE0dotE1 = sqrt(valE0dotE0*valE1dotE1) - m_valE0dotE1;
  
  m_valDen = valE0E1minusE0dotE1;
  m_iDen = ( m_valDen != 0.0 );
  if ( m_iDen == -1 )
    {
      bDone = InscribedCircleCenterEEvMPFUN(pdC_x, pdC_y, pdR, 
					    dE00_x, dE00_y, 
					    dE01_x, dE01_y,
					    dE10_x, dE10_y, 
					    dE11_V2_x, dE11_V2_y,
					    ipDen);
      if ( bDone )
	{
	  return m_eResult;
	}
      if ( m_iDen == -1 )
	{
	  bDone = InscribedCircleCenterEEvMPFUN(pdC_x, pdC_y, pdR, 
						dE00_x, dE00_y, 
						dE01_x, dE01_y,
						dE10_x, dE10_y, 
						dE11_V2_x, dE11_V2_y,
						ipEnd);
	  ASSERT( bDone );
	  return m_eResult;
	}
    }
  ASSERT( m_iDen == 1 );
	
  //////////////////////////////////////////////////
  // Check the Edge Zone
  
  Interval pppvalV[3][2][2]; // Endpoints
  // Translate V1 to the origin
  //memset(pppvalV, 0, sizeof(pppvalV));
  pppvalV[0][0][0] = valE00_x - valE11_V2_x;
  pppvalV[0][0][1] = valE00_y - valE11_V2_y;
  pppvalV[0][1][0] = valE01_x - valE11_V2_x;
  pppvalV[0][1][1] = valE01_y - valE11_V2_y;
  pppvalV[1][0][0] = valE10_x - valE11_V2_x;
  pppvalV[1][0][1] = valE10_y - valE11_V2_y;
  pppvalV[1][1][0] = 0.0;
  pppvalV[1][1][1] = 0.0;
  pppvalV[2][0][0] = 0.0;
  pppvalV[2][0][1] = 0.0;
  
  Interval valE00xE01;
  Cross(valE00xE01, pppvalV[0][0][0], pppvalV[0][0][1], pppvalV[0][1][0], pppvalV[0][1][1]);
  
  Interval valE00xE01timesE0xE1 = valE00xE01*m_valE0xE1;
  
  Interval pvalEdgeZone[2];
  
  Interval valE0dotE00;
  Dot(valE0dotE00, ppvalE[0][0], ppvalE[0][1], pppvalV[0][0][0], pppvalV[0][0][1]);
  pvalEdgeZone[0] = -( valE0E1minusE0dotE1*valE0dotE00 + valE00xE01timesE0xE1 );
  m_piEdgeZone[0] = ( pvalEdgeZone[0] < 0.0 );
  if ( m_piEdgeZone[0] == 1 )
    {
      return APM2D::icctNull;
    }
  
  Interval valE0dotE01;
  Dot(valE0dotE01, ppvalE[0][0], ppvalE[0][1], pppvalV[0][1][0], pppvalV[0][1][1]);
  pvalEdgeZone[1] = ( valE0E1minusE0dotE1*valE0dotE01 + valE00xE01timesE0xE1 );
  m_piEdgeZone[1] = ( pvalEdgeZone[1] < 0.0 );
  if ( m_piEdgeZone[1] == 1 )
    {
      return APM2D::icctNull;
    }

  if ( m_piEdgeZone[0] == -1 ||
       m_piEdgeZone[1] == -1 )
    {
      bDone = InscribedCircleCenterEEvMPFUN(pdC_x, pdC_y, pdR, 
					    dE00_x, dE00_y, 
					    dE01_x, dE01_y,
					    dE10_x, dE10_y, 
					    dE11_V2_x, dE11_V2_y,
					    ipEdgeZones);
      if ( bDone )
	{
	  return m_eResult;
	}
    }
  
  //////////////////////////////////////////////////
  // Compute the center and radius
  
  Interval pvalC[2];
  pvalC[0] = -( valE00xE01*ppvalE[1][1] ) / m_valDen;
  pvalC[1] = ( valE00xE01*ppvalE[1][0] ) / m_valDen;
  
  Interval valR;
  valR = ( sqrt(valE1dotE1)*valE00xE01 ) / m_valDen;
  
  // Translate the answer back out away from the origin
  pvalC[0] += valE11_V2_x;
  pvalC[1] += valE11_V2_y;
  
  *pdC_x = dble(pvalC[0]);
  *pdC_y = dble(pvalC[1]);
  *pdR = dble(valR);
  return APM2D::icctEEv;
}

APM2D::InscribedCircleCenterType 
ICC_EEv_MPFUNFilter::InscribedCircleCenterEEvForce(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
						   DOUBLE dE00_x, DOUBLE dE00_y, 
						   DOUBLE dE01_x, DOUBLE dE01_y,
						   DOUBLE dE10_x, DOUBLE dE10_y, 
						   DOUBLE dE11_V2_x, DOUBLE dE11_V2_y)
{
  *pdC_x = 0.0;
  *pdC_y = 0.0;
  *pdR = 0.0;
  
  m_eIP = ipStart;
  m_eResult = APM2D::icctNull;
  m_iV2onE0 = -1;
  m_iE0xE1 = -1;
  m_iE0dotE1 = -1;
  m_iDen = -1;
  m_piEdgeZone[0] = -1;
  m_piEdgeZone[1] = -1;

  BOOL bDone;

  bDone = InscribedCircleCenterEEvMPFUN(pdC_x, pdC_y, pdR, 
					dE00_x, dE00_y, 
					dE01_x, dE01_y,
					dE10_x, dE10_y, 
					dE11_V2_x, dE11_V2_y,
					ipEnd);
  ASSERT( bDone );
  return m_eResult;
}

#ifndef COMPILER_PROBLEM

BOOL 
ICC_EEv_MPFUNFilter::InscribedCircleCenterEEvMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
						   DOUBLE dE00_x, DOUBLE dE00_y, 
						   DOUBLE dE01_x, DOUBLE dE01_y,
						   DOUBLE dE10_x, DOUBLE dE10_y, 
						   DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
						   InstructionPointer eIPMax/* = ipEnd*/)
{
  APM2D::m_uiInscribedCircleCenterEEvMPFUN++;

  switch ( m_eIP )
    {
    case ipStart:
      {
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);
    
	*pdC_x = 0.0;
	*pdC_y = 0.0;
	*pdR = 0.0;
    
	m_rE00_x = dE00_x;
	m_rE00_y = dE00_y; 
	m_rE01_x = dE01_x;
	m_rE01_y = dE01_y;
	m_rE10_x = dE10_x;
	m_rE10_y = dE10_y; 
	m_rE11_V2_x = dE11_V2_x;
	m_rE11_V2_y = dE11_V2_y;

	if ( m_iV2onE0 == -1 )
	  {
	    mp_real rV2onE0;
	    SinTheta(rV2onE0, m_rE11_V2_x, m_rE11_V2_y, m_rE00_x, m_rE00_y, m_rE01_x, m_rE01_y);
	    if ( rV2onE0 <= 0.0 )
	      {
		// V2 is on E0 and it is not one of the endpoints
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterEEvMPFUNFinal++;
		return TRUE;
	      }
	  }
  
      }
    case ipV2onE0:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);
	
	//////////////////////////////////////////////////
	// Check Colinear
    
	//memset(m_pprE, 0, sizeof(m_pprE));
	m_pprE[0][0] = m_rE01_x - m_rE00_x;
	m_pprE[0][1] = m_rE01_y - m_rE00_y;
	m_pprE[1][0] = m_rE11_V2_x - m_rE10_x;
	m_pprE[1][1] = m_rE11_V2_y - m_rE10_y;
  
	Dot(m_rE0dotE1, m_pprE[0][0], m_pprE[0][1], m_pprE[1][0], m_pprE[1][1]);
	Cross(m_rE0xE1, m_pprE[0][0], m_pprE[0][1], m_pprE[1][0], m_pprE[1][1]);

	if ( m_iE0xE1 == -1 || m_iE0dotE1 == -1 )
	  {
	    if ( m_rE0xE1 == 0.0 &&
		 m_rE0dotE1 > 0.0 )
	      {
		// Colinear
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterEEvMPFUNFinal++;
		return TRUE;
	      }
	    if ( eIPMax != ipEnd )
	      {
		if ( m_iE0xE1 == -1 )
		  {
		    m_valE0xE1.ConservativeBound(dble(m_rE0xE1));
		    m_iE0xE1 = ( m_valE0xE1 == 0.0 );
		  }
		if ( m_iE0dotE1 == -1 )
		  {
		    m_valE0dotE1.ConservativeBound(dble(m_rE0dotE1));
		    m_iE0dotE1 = ( m_valE0dotE1 > 0.0 );
		  }
	      }	    
	  }

      }
    case ipColinear:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);
	
	//////////////////////////////////////////////////
	// Check Den

	mp_real rE0dotE0;
	Dot(rE0dotE0, m_pprE[0][0], m_pprE[0][1], m_pprE[0][0], m_pprE[0][1]);
	Dot(m_rE1dotE1, m_pprE[1][0], m_pprE[1][1], m_pprE[1][0], m_pprE[1][1]);

	m_rE0E1minusE0dotE1 = sqrt(rE0dotE0*m_rE1dotE1) - m_rE0dotE1;

	m_rDen = m_rE0E1minusE0dotE1;
	if ( m_iDen == -1 )
	  {
	    ASSERT( m_rDen != 0.0 );
	    if ( eIPMax != ipEnd )
	      { 
		m_valDen.ConservativeBound(dble(m_rDen));	    
		m_iDen = ( m_valDen != 0.0 );      
	      }
	  }

      }
    case ipDen:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);
	
	//////////////////////////////////////////////////
	// Check the Edge Zone

	mp_real ppprV[3][2][2]; // Endpoints
	// Translate V1 to the origin
	//memset(ppprV, 0, sizeof(ppprV));
	ppprV[0][0][0] = m_rE00_x - m_rE11_V2_x;
	ppprV[0][0][1] = m_rE00_y - m_rE11_V2_y;
	ppprV[0][1][0] = m_rE01_x - m_rE11_V2_x;
	ppprV[0][1][1] = m_rE01_y - m_rE11_V2_y;
	ppprV[1][0][0] = m_rE10_x - m_rE11_V2_x;
	ppprV[1][0][1] = m_rE10_y - m_rE11_V2_y;
	ppprV[1][1][0] = 0.0;
	ppprV[1][1][1] = 0.0;
	ppprV[2][0][0] = 0.0;
	ppprV[2][0][1] = 0.0;
  
	Cross(m_rE00xE01, ppprV[0][0][0], ppprV[0][0][1], ppprV[0][1][0], ppprV[0][1][1]);

	if ( m_piEdgeZone[0] == -1 || 
	     m_piEdgeZone[1] == -1 )
	  {      
	    mp_real rE00xE01timesE0xE1 = m_rE00xE01*m_rE0xE1;
	    
	    mp_real prEdgeZone[2];
	    
	    if ( m_piEdgeZone[0] == -1 )
	      {
		mp_real rE0dotE00;
		Dot(rE0dotE00, m_pprE[0][0], m_pprE[0][1], ppprV[0][0][0], ppprV[0][0][1]);
		prEdgeZone[0] = -( m_rE0E1minusE0dotE1*rE0dotE00 + rE00xE01timesE0xE1 );
		if ( prEdgeZone[0] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterEEvMPFUNFinal++;
		    return TRUE;
		  }
	      }
	    
	    if ( m_piEdgeZone[1] == -1 )
	      {
		mp_real rE0dotE01;
		Dot(rE0dotE01, m_pprE[0][0], m_pprE[0][1], ppprV[0][1][0], ppprV[0][1][1]);
		prEdgeZone[1] = ( m_rE0E1minusE0dotE1*rE0dotE01 + rE00xE01timesE0xE1 );
		if ( prEdgeZone[1] < 0.0 )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterEEvMPFUNFinal++;
		    return TRUE;
		  }
	      }
	  }

      }
    case ipEdgeZones:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Compute the center and radius

	mp_real prC[2];
	prC[0] = -( m_rE00xE01*m_pprE[1][1] ) / m_rDen;
	prC[1] = ( m_rE00xE01*m_pprE[1][0] ) / m_rDen;

	mp_real rR;
	rR = ( sqrt(m_rE1dotE1)*m_rE00xE01 ) / m_rDen;

	// Translate the answer back out away from the origin
	prC[0] += m_rE11_V2_x;
	prC[1] += m_rE11_V2_y;

	*pdC_x = dble(prC[0]);
	*pdC_y = dble(prC[1]);
	*pdR = dble(rR);
	m_eResult = APM2D::icctEEv;
      }
    }

  APM2D::m_uiInscribedCircleCenterEEvMPFUNFinal++;
  return TRUE;
}

#else // COMPILER_PROBLEM

BOOL 
ICC_EEv_MPFUNFilter::InscribedCircleCenterEEvMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
						   DOUBLE dE00_x, DOUBLE dE00_y, 
						   DOUBLE dE01_x, DOUBLE dE01_y,
						   DOUBLE dE10_x, DOUBLE dE10_y, 
						   DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
						   InstructionPointer eIPMax/* = ipEnd*/)
{
  APM2D::m_uiInscribedCircleCenterEEvMPFUN++;

  switch ( m_eIP )
    {
    case ipStart:
      {
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);
    
	*pdC_x = 0.0;
	*pdC_y = 0.0;
	*pdR = 0.0;
    
	m_rE00_x = dE00_x;
	m_rE00_y = dE00_y; 
	m_rE01_x = dE01_x;
	m_rE01_y = dE01_y;
	m_rE10_x = dE10_x;
	m_rE10_y = dE10_y; 
	m_rE11_V2_x = dE11_V2_x;
	m_rE11_V2_y = dE11_V2_y;

	if ( m_iV2onE0 == -1 )
	  {
	    mp_real rV2onE0;
	    SinTheta(rV2onE0, m_rE11_V2_x, m_rE11_V2_y, m_rE00_x, m_rE00_y, m_rE01_x, m_rE01_y);
	    if ( rV2onE0 <= 0.0 )
	      {
		// V2 is on E0 and it is not one of the endpoints
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterEEvMPFUNFinal++;
		return TRUE;
	      }
	  }
  
      }
    case ipV2onE0:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);
	
	if ( Colinear(eIPMax) )
	  {
	    APM2D::m_uiInscribedCircleCenterEEvMPFUNFinal++;
	    return TRUE;
	  }
      }
    case ipColinear:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);
	
	if ( Den(eIPMax) )
	  {
	    APM2D::m_uiInscribedCircleCenterEEvMPFUNFinal++;
	    return TRUE;
	  }
      }
    case ipDen:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);
	
	if ( EdgeZones(eIPMax) )
	  {
	    APM2D::m_uiInscribedCircleCenterEEvMPFUNFinal++;
	    return TRUE;
	  }
      }
    case ipEdgeZones:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Compute the center and radius

	mp_real prC[2];
	prC[0] = -( m_rE00xE01*m_pprE[1][1] ) / m_rDen;
	prC[1] = ( m_rE00xE01*m_pprE[1][0] ) / m_rDen;

	mp_real rR;
	rR = ( sqrt(m_rE1dotE1)*m_rE00xE01 ) / m_rDen;

	// Translate the answer back out away from the origin
	prC[0] += m_rE11_V2_x;
	prC[1] += m_rE11_V2_y;

	*pdC_x = dble(prC[0]);
	*pdC_y = dble(prC[1]);
	*pdR = dble(rR);
	m_eResult = APM2D::icctEEv;
      }
    }

  APM2D::m_uiInscribedCircleCenterEEvMPFUNFinal++;
  return TRUE;
}

BOOL 
ICC_EEv_MPFUNFilter::Colinear(InstructionPointer eIPMax)
{
  //////////////////////////////////////////////////
  // Check Colinear
  
  //memset(m_pprE, 0, sizeof(m_pprE));
  m_pprE[0][0] = m_rE01_x - m_rE00_x;
  m_pprE[0][1] = m_rE01_y - m_rE00_y;
  m_pprE[1][0] = m_rE11_V2_x - m_rE10_x;
  m_pprE[1][1] = m_rE11_V2_y - m_rE10_y;
  
  Dot(m_rE0dotE1, m_pprE[0][0], m_pprE[0][1], m_pprE[1][0], m_pprE[1][1]);
  Cross(m_rE0xE1, m_pprE[0][0], m_pprE[0][1], m_pprE[1][0], m_pprE[1][1]);
  
  if ( m_iE0xE1 == -1 || m_iE0dotE1 == -1 )
    {
      if ( m_rE0xE1 == 0.0 &&
	   m_rE0dotE1 > 0.0 )
	{
	  // Colinear
	  m_eResult = APM2D::icctNull;
	  return TRUE;
	}
      if ( eIPMax != ipEnd )
	{
	  if ( m_iE0xE1 == -1 )
	    {
	      m_valE0xE1.ConservativeBound(dble(m_rE0xE1));
	      m_iE0xE1 = ( m_valE0xE1 == 0.0 );
	    }
	  if ( m_iE0dotE1 == -1 )
	    {
	      m_valE0dotE1.ConservativeBound(dble(m_rE0dotE1));
	      m_iE0dotE1 = ( m_valE0dotE1 > 0.0 );
	    }
	}	    
    }

  return FALSE;
}

BOOL 
ICC_EEv_MPFUNFilter::Den(InstructionPointer eIPMax)
{
  //////////////////////////////////////////////////
  // Check Den
  
  mp_real rE0dotE0;
  Dot(rE0dotE0, m_pprE[0][0], m_pprE[0][1], m_pprE[0][0], m_pprE[0][1]);
  Dot(m_rE1dotE1, m_pprE[1][0], m_pprE[1][1], m_pprE[1][0], m_pprE[1][1]);
  
  m_rE0E1minusE0dotE1 = sqrt(rE0dotE0*m_rE1dotE1) - m_rE0dotE1;
  
  m_rDen = m_rE0E1minusE0dotE1;
  if ( m_iDen == -1 )
    {
      ASSERT( m_rDen != 0.0 );
      if ( eIPMax != ipEnd )
	{ 
	  m_valDen.ConservativeBound(dble(m_rDen));	    
	  m_iDen = ( m_valDen != 0.0 );      
	}
    }

  return FALSE;
}

BOOL 
ICC_EEv_MPFUNFilter::EdgeZones(InstructionPointer /*eIPMax*/)
{
  //////////////////////////////////////////////////
  // Check the Edge Zone

  mp_real ppprV[3][2][2]; // Endpoints
  // Translate V1 to the origin
  //memset(ppprV, 0, sizeof(ppprV));
  ppprV[0][0][0] = m_rE00_x - m_rE11_V2_x;
  ppprV[0][0][1] = m_rE00_y - m_rE11_V2_y;
  ppprV[0][1][0] = m_rE01_x - m_rE11_V2_x;
  ppprV[0][1][1] = m_rE01_y - m_rE11_V2_y;
  ppprV[1][0][0] = m_rE10_x - m_rE11_V2_x;
  ppprV[1][0][1] = m_rE10_y - m_rE11_V2_y;
  ppprV[1][1][0] = 0.0;
  ppprV[1][1][1] = 0.0;
  ppprV[2][0][0] = 0.0;
  ppprV[2][0][1] = 0.0;
  
  Cross(m_rE00xE01, ppprV[0][0][0], ppprV[0][0][1], ppprV[0][1][0], ppprV[0][1][1]);

  if ( m_piEdgeZone[0] == -1 || 
       m_piEdgeZone[1] == -1 )
    {      
      mp_real rE00xE01timesE0xE1 = m_rE00xE01*m_rE0xE1;
	    
      mp_real prEdgeZone[2];
	    
      if ( m_piEdgeZone[0] == -1 )
	{
	  mp_real rE0dotE00;
	  Dot(rE0dotE00, m_pprE[0][0], m_pprE[0][1], ppprV[0][0][0], ppprV[0][0][1]);
	  prEdgeZone[0] = -( m_rE0E1minusE0dotE1*rE0dotE00 + rE00xE01timesE0xE1 );
	  if ( prEdgeZone[0] < 0.0 )
	    {
	      m_eResult = APM2D::icctNull;
	      return TRUE;
	    }
	}
	    
      if ( m_piEdgeZone[1] == -1 )
	{
	  mp_real rE0dotE01;
	  Dot(rE0dotE01, m_pprE[0][0], m_pprE[0][1], ppprV[0][1][0], ppprV[0][1][1]);
	  prEdgeZone[1] = ( m_rE0E1minusE0dotE1*rE0dotE01 + rE00xE01timesE0xE1 );
	  if ( prEdgeZone[1] < 0.0 )
	    {
	      m_eResult = APM2D::icctNull;
	      return TRUE;
	    }
	}
    }

  return FALSE;
}

#endif // COMPILER_PROBLEM

//static 
APM2D::InscribedCircleCenterType 
InscribedCircleCenterEEvMPFUNFilter(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				    DOUBLE dE00_x, DOUBLE dE00_y, 
				    DOUBLE dE01_x, DOUBLE dE01_y,
				    DOUBLE dE10_x, DOUBLE dE10_y, 
				    DOUBLE dE11_V2_x, DOUBLE dE11_V2_y)
{
  ICC_EEv_MPFUNFilter iccEEv;

  return iccEEv.InscribedCircleCenterEEv(pdC_x, pdC_y, pdR, 
					 dE00_x, dE00_y, 
					 dE01_x, dE01_y,
					 dE10_x, dE10_y, 
					 dE11_V2_x, dE11_V2_y);
}

//static 
APM2D::EdgeZoneType EdgeZoneEEvMPFUNFilter(DOUBLE dE00_x, DOUBLE dE00_y, 
				     DOUBLE dE01_x, DOUBLE dE01_y,
				     DOUBLE dE10_x, DOUBLE dE10_y, 
				     DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
				     DOUBLE dE30_x, DOUBLE dE30_y,
				     DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::m_uiEdgeZoneEEvMPFUNFilter++;

  INT_32 piEdgeZone[2];
  piEdgeZone[0] = -1;
  piEdgeZone[1] = -1;

  switch ( 0 )
  {
  default:
    Interval valE00_x = dE00_x;
    Interval valE00_y = dE00_y; 
    Interval valE01_x = dE01_x;
    Interval valE01_y = dE01_y;
    Interval valE10_x = dE10_x;
    Interval valE10_y = dE10_y; 
    Interval valE11_V2_x = dE11_V2_x;
    Interval valE11_V2_y = dE11_V2_y;
    Interval valE30_x = dE30_x;
    Interval valE30_y = dE30_y;
    Interval valE31_x = dE31_x;
    Interval valE31_y = dE31_y;

    Interval pppvalV[4][2][2]; // Endpoints
    // Translate V1 to the origin
    //memset(pppvalV, 0, sizeof(pppvalV));
    pppvalV[0][0][0] = valE00_x - valE11_V2_x;
    pppvalV[0][0][1] = valE00_y - valE11_V2_y;
    pppvalV[0][1][0] = valE01_x - valE11_V2_x;
    pppvalV[0][1][1] = valE01_y - valE11_V2_y;
    pppvalV[1][0][0] = valE10_x - valE11_V2_x;
    pppvalV[1][0][1] = valE10_y - valE11_V2_y;
    pppvalV[1][1][0] = 0.0;
    pppvalV[1][1][1] = 0.0;
    pppvalV[2][0][0] = 0.0;
    pppvalV[2][0][1] = 0.0;
    pppvalV[3][0][0] = valE30_x - valE11_V2_x;
    pppvalV[3][0][1] = valE30_y - valE11_V2_y;
    pppvalV[3][1][0] = valE31_x - valE11_V2_x;
    pppvalV[3][1][1] = valE31_y - valE11_V2_y;
  
    Interval ppvalE[4][2];     // Edge vectors
    //memset(ppvalE, 0, sizeof(ppvalE));
    ppvalE[0][0] = valE01_x - valE00_x;
    ppvalE[0][1] = valE01_y - valE00_y;
    ppvalE[1][0] = valE11_V2_x - valE10_x;
    ppvalE[1][1] = valE11_V2_y - valE10_y;
    ppvalE[3][0] = valE31_x - valE30_x;
    ppvalE[3][1] = valE31_y - valE30_y;
  
    Interval valE0dotE0;
    Dot(valE0dotE0, ppvalE[0][0], ppvalE[0][1], ppvalE[0][0], ppvalE[0][1]);
    Interval valE1dotE1;
    Dot(valE1dotE1, ppvalE[1][0], ppvalE[1][1], ppvalE[1][0], ppvalE[1][1]);
    Interval valE0dotE1;
    Dot(valE0dotE1, ppvalE[0][0], ppvalE[0][1], ppvalE[1][0], ppvalE[1][1]);

    Interval valE1xE3;
    Cross(valE1xE3, ppvalE[1][0], ppvalE[1][1], ppvalE[3][0], ppvalE[3][1]);
    Interval valE00xE01;
    Cross(valE00xE01, pppvalV[0][0][0], pppvalV[0][0][1], pppvalV[0][1][0], pppvalV[0][1][1]);

    Interval valE0E1minusE0dotE1;
    valE0E1minusE0dotE1 = sqrt(valE0dotE0*valE1dotE1) - valE0dotE1;

    Interval valE00xE01timesE1xE3 = valE00xE01*valE1xE3;

    Interval pvalEdgeZone[2];

    Interval valE3dotE30;
    Dot(valE3dotE30, ppvalE[3][0], ppvalE[3][1], pppvalV[3][0][0], pppvalV[3][0][1]);
    pvalEdgeZone[0] = ( valE0E1minusE0dotE1*valE3dotE30 - valE00xE01timesE1xE3 );
    piEdgeZone[0] = ( pvalEdgeZone[0] >= 0.0 );
    if ( piEdgeZone[0] == 1 )  
      {
	return APM2D::eztStart;
      }

    Interval valE3dotE31;
    Dot(valE3dotE31, ppvalE[3][0], ppvalE[3][1], pppvalV[3][1][0], pppvalV[3][1][1]);
    pvalEdgeZone[1] = -( valE0E1minusE0dotE1*valE3dotE31 - valE00xE01timesE1xE3 );
    piEdgeZone[1] = ( pvalEdgeZone[1] >= 0.0 );
    if ( piEdgeZone[1] == 1 )
      {
	return APM2D::eztEnd;
      }
  }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( piEdgeZone[0] == -1 || piEdgeZone[1] == -1 )
    {
      APM2D::m_uiEdgeZoneEEvMPFUN++;

      mp_real rE00_x = dE00_x;
      mp_real rE00_y = dE00_y; 
      mp_real rE01_x = dE01_x;
      mp_real rE01_y = dE01_y;
      mp_real rE10_x = dE10_x;
      mp_real rE10_y = dE10_y; 
      mp_real rE11_V2_x = dE11_V2_x;
      mp_real rE11_V2_y = dE11_V2_y;
      mp_real rE30_x = dE30_x;
      mp_real rE30_y = dE30_y;
      mp_real rE31_x = dE31_x;
      mp_real rE31_y = dE31_y;

      mp_real ppprV[4][2][2]; // Endpoints
      // Translate V1 to the origin
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = rE00_x - rE11_V2_x;
      ppprV[0][0][1] = rE00_y - rE11_V2_y;
      ppprV[0][1][0] = rE01_x - rE11_V2_x;
      ppprV[0][1][1] = rE01_y - rE11_V2_y;
      ppprV[1][0][0] = rE10_x - rE11_V2_x;
      ppprV[1][0][1] = rE10_y - rE11_V2_y;
      ppprV[1][1][0] = 0.0;
      ppprV[1][1][1] = 0.0;
      ppprV[2][0][0] = 0.0;
      ppprV[2][0][1] = 0.0;
      ppprV[3][0][0] = rE30_x - rE11_V2_x;
      ppprV[3][0][1] = rE30_y - rE11_V2_y;
      ppprV[3][1][0] = rE31_x - rE11_V2_x;
      ppprV[3][1][1] = rE31_y - rE11_V2_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[0][0] = rE01_x - rE00_x;
      pprE[0][1] = rE01_y - rE00_y;
      pprE[1][0] = rE11_V2_x - rE10_x;
      pprE[1][1] = rE11_V2_y - rE10_y;
      pprE[3][0] = rE31_x - rE30_x;
      pprE[3][1] = rE31_y - rE30_y;
  
      mp_real rE0dotE0;
      Dot(rE0dotE0, pprE[0][0], pprE[0][1], pprE[0][0], pprE[0][1]);
      mp_real rE1dotE1;
      Dot(rE1dotE1, pprE[1][0], pprE[1][1], pprE[1][0], pprE[1][1]);
      mp_real rE0dotE1;
      Dot(rE0dotE1, pprE[0][0], pprE[0][1], pprE[1][0], pprE[1][1]);

      mp_real rE1xE3;
      Cross(rE1xE3, pprE[1][0], pprE[1][1], pprE[3][0], pprE[3][1]);
      mp_real rE00xE01;
      Cross(rE00xE01, ppprV[0][0][0], ppprV[0][0][1], ppprV[0][1][0], ppprV[0][1][1]);

      mp_real rE0E1minusE0dotE1;
      rE0E1minusE0dotE1 = sqrt(rE0dotE0*rE1dotE1) - rE0dotE1;

      mp_real rE00xE01timesE1xE3 = rE00xE01*rE1xE3;

      mp_real prEdgeZone[2];

      if ( piEdgeZone[0] == -1 )
	{
	  mp_real rE3dotE30;
	  Dot(rE3dotE30, pprE[3][0], pprE[3][1], ppprV[3][0][0], ppprV[3][0][1]);
	  prEdgeZone[0] = ( rE0E1minusE0dotE1*rE3dotE30 - rE00xE01timesE1xE3 );
	  if ( prEdgeZone[0] >= 0.0 )
	    {
	      return APM2D::eztStart;
	    }
	}

      if ( piEdgeZone[1] == -1 )
	{
	  mp_real rE3dotE31;
	  Dot(rE3dotE31, pprE[3][0], pprE[3][1], ppprV[3][1][0], ppprV[3][1][1]);
	  prEdgeZone[1] = -( rE0E1minusE0dotE1*rE3dotE31 - rE00xE01timesE1xE3 );
	  if ( prEdgeZone[1] >= 0.0 )
	    {
	      return APM2D::eztEnd;
	    }
	}
    }

  return APM2D::eztMiddle;
}

//static 
APM2D::CircleTestType InCircleEEv_VMPFUNFilter(DOUBLE dE00_x, DOUBLE dE00_y, 
					 DOUBLE dE01_x, DOUBLE dE01_y,
					 DOUBLE dE10_x, DOUBLE dE10_y, 
					 DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
					 DOUBLE dV3_x, DOUBLE dV3_y)
{
  APM2D::m_uiInCircleEEv_VMPFUNFilter++;

  INT_32 iResult;

  iResult = -1;

  switch ( 0 )
    {
    default:
      Interval valE00_x = dE00_x;
      Interval valE00_y = dE00_y; 
      Interval valE01_x = dE01_x;
      Interval valE01_y = dE01_y;
      Interval valE10_x = dE10_x;
      Interval valE10_y = dE10_y; 
      Interval valE11_V2_x = dE11_V2_x;
      Interval valE11_V2_y = dE11_V2_y;
      Interval valV3_x = dV3_x;
      Interval valV3_y = dV3_y;

      Interval pppvalV[4][2][2]; // Endpoints
      // Translate V1 to the origin
      //memset(pppvalV, 0, sizeof(pppvalV));
      pppvalV[0][0][0] = valE00_x - valE11_V2_x;
      pppvalV[0][0][1] = valE00_y - valE11_V2_y;
      pppvalV[0][1][0] = valE01_x - valE11_V2_x;
      pppvalV[0][1][1] = valE01_y - valE11_V2_y;
      pppvalV[1][0][0] = valE10_x - valE11_V2_x;
      pppvalV[1][0][1] = valE10_y - valE11_V2_y;
      pppvalV[1][1][0] = 0.0;
      pppvalV[1][1][1] = 0.0;
      pppvalV[2][0][0] = 0.0;
      pppvalV[2][0][1] = 0.0;
      pppvalV[3][0][0] = valV3_x - valE11_V2_x;
      pppvalV[3][0][1] = valV3_y - valE11_V2_y;
  
      Interval ppvalE[4][2];     // Edge vectors
      //memset(ppvalE, 0, sizeof(ppvalE));
      ppvalE[0][0] = valE01_x - valE00_x;
      ppvalE[0][1] = valE01_y - valE00_y;
      ppvalE[1][0] = valE11_V2_x - valE10_x;
      ppvalE[1][1] = valE11_V2_y - valE10_y;
  
      Interval valE0dotE0;
      Dot(valE0dotE0, ppvalE[0][0], ppvalE[0][1], ppvalE[0][0], ppvalE[0][1]);
      Interval valE1dotE1;
      Dot(valE1dotE1, ppvalE[1][0], ppvalE[1][1], ppvalE[1][0], ppvalE[1][1]);
      Interval valE0dotE1;
      Dot(valE0dotE1, ppvalE[0][0], ppvalE[0][1], ppvalE[1][0], ppvalE[1][1]);
      Interval valV3dotV3;
      Dot(valV3dotV3, pppvalV[3][0][0], pppvalV[3][0][1], pppvalV[3][0][0], pppvalV[3][0][1]);

      Interval valE1xV3;
      Cross(valE1xV3, ppvalE[1][0], ppvalE[1][1], pppvalV[3][0][0], pppvalV[3][0][1]);
      Interval valE00xE01;
      Cross(valE00xE01, pppvalV[0][0][0], pppvalV[0][0][1], pppvalV[0][1][0], pppvalV[0][1][1]);

      Interval valE0E1minusE0dotE1;
      valE0E1minusE0dotE1 = sqrt(valE0dotE0*valE1dotE1) - valE0dotE1;

      Interval valResult;
      valResult = ( valE0E1minusE0dotE1*valV3dotV3 - 2.0*valE00xE01*valE1xV3 );

      iResult = ( valResult < 0.0 );
      if ( iResult == -1 )
	{
	  break;
	}
      if ( iResult == 1 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( valResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
      return APM2D::cttOnCircle;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( iResult == -1 )
    {
      APM2D::m_uiInCircleEEv_VMPFUN++;

      mp_real rE00_x = dE00_x;
      mp_real rE00_y = dE00_y; 
      mp_real rE01_x = dE01_x;
      mp_real rE01_y = dE01_y;
      mp_real rE10_x = dE10_x;
      mp_real rE10_y = dE10_y; 
      mp_real rE11_V2_x = dE11_V2_x;
      mp_real rE11_V2_y = dE11_V2_y;
      mp_real rV3_x = dV3_x;
      mp_real rV3_y = dV3_y;

      mp_real ppprV[4][2][2]; // Endpoints
      // Translate V1 to the origin
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = rE00_x - rE11_V2_x;
      ppprV[0][0][1] = rE00_y - rE11_V2_y;
      ppprV[0][1][0] = rE01_x - rE11_V2_x;
      ppprV[0][1][1] = rE01_y - rE11_V2_y;
      ppprV[1][0][0] = rE10_x - rE11_V2_x;
      ppprV[1][0][1] = rE10_y - rE11_V2_y;
      ppprV[1][1][0] = 0.0;
      ppprV[1][1][1] = 0.0;
      ppprV[2][0][0] = 0.0;
      ppprV[2][0][1] = 0.0;
      ppprV[3][0][0] = rV3_x - rE11_V2_x;
      ppprV[3][0][1] = rV3_y - rE11_V2_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[0][0] = rE01_x - rE00_x;
      pprE[0][1] = rE01_y - rE00_y;
      pprE[1][0] = rE11_V2_x - rE10_x;
      pprE[1][1] = rE11_V2_y - rE10_y;
  
      mp_real rE0dotE0;
      Dot(rE0dotE0, pprE[0][0], pprE[0][1], pprE[0][0], pprE[0][1]);
      mp_real rE1dotE1;
      Dot(rE1dotE1, pprE[1][0], pprE[1][1], pprE[1][0], pprE[1][1]);
      mp_real rE0dotE1;
      Dot(rE0dotE1, pprE[0][0], pprE[0][1], pprE[1][0], pprE[1][1]);
      mp_real rV3dotV3;
      Dot(rV3dotV3, ppprV[3][0][0], ppprV[3][0][1], ppprV[3][0][0], ppprV[3][0][1]);

      mp_real rE1xV3;
      Cross(rE1xV3, pprE[1][0], pprE[1][1], ppprV[3][0][0], ppprV[3][0][1]);
      mp_real rE00xE01;
      Cross(rE00xE01, ppprV[0][0][0], ppprV[0][0][1], ppprV[0][1][0], ppprV[0][1][1]);

      mp_real rE0E1minusE0dotE1;
      rE0E1minusE0dotE1 = sqrt(rE0dotE0*rE1dotE1) - rE0dotE1;

      mp_real rResult;
      rResult = ( rE0E1minusE0dotE1*rV3dotV3 - 2.0*rE00xE01*rE1xV3 );
  
      if ( rResult < 0.0 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( rResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
    }

  return APM2D::cttOnCircle;
}

//static 
APM2D::CircleTestType InCircleEEv_EMPFUNFilter(DOUBLE dE00_x, DOUBLE dE00_y, 
					 DOUBLE dE01_x, DOUBLE dE01_y,
					 DOUBLE dE10_x, DOUBLE dE10_y, 
					 DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
					 DOUBLE dE30_x, DOUBLE dE30_y,
					 DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::m_uiInCircleEEv_EMPFUNFilter++;

  INT_32 iResult;

  iResult = -1;

  switch ( 0 )
    {
    default:
      Interval valE00_x = dE00_x;
      Interval valE00_y = dE00_y; 
      Interval valE01_x = dE01_x;
      Interval valE01_y = dE01_y;
      Interval valE10_x = dE10_x;
      Interval valE10_y = dE10_y; 
      Interval valE11_V2_x = dE11_V2_x;
      Interval valE11_V2_y = dE11_V2_y;
      Interval valE30_x = dE30_x;
      Interval valE30_y = dE30_y;
      Interval valE31_x = dE31_x;
      Interval valE31_y = dE31_y;
  
      Interval pppvalV[4][2][2]; // Endpoints
      // Translate V1 to the origin
      //memset(pppvalV, 0, sizeof(pppvalV));
      pppvalV[0][0][0] = valE00_x - valE11_V2_x;
      pppvalV[0][0][1] = valE00_y - valE11_V2_y;
      pppvalV[0][1][0] = valE01_x - valE11_V2_x;
      pppvalV[0][1][1] = valE01_y - valE11_V2_y;
      pppvalV[1][0][0] = valE10_x - valE11_V2_x;
      pppvalV[1][0][1] = valE10_y - valE11_V2_y;
      pppvalV[1][1][0] = 0.0;
      pppvalV[1][1][1] = 0.0;
      pppvalV[2][0][0] = 0.0;
      pppvalV[2][0][1] = 0.0;
      pppvalV[3][0][0] = valE30_x - valE11_V2_x;
      pppvalV[3][0][1] = valE30_y - valE11_V2_y;
      pppvalV[3][1][0] = valE31_x - valE11_V2_x;
      pppvalV[3][1][1] = valE31_y - valE11_V2_y;
  
      Interval ppvalE[4][2];     // Edge vectors
      //memset(ppvalE, 0, sizeof(ppvalE));
      ppvalE[0][0] = valE01_x - valE00_x;
      ppvalE[0][1] = valE01_y - valE00_y;
      ppvalE[1][0] = valE11_V2_x - valE10_x;
      ppvalE[1][1] = valE11_V2_y - valE10_y;
      ppvalE[3][0] = valE31_x - valE30_x;
      ppvalE[3][1] = valE31_y - valE30_y;
  
      Interval valE0dotE0;
      Dot(valE0dotE0, ppvalE[0][0], ppvalE[0][1], ppvalE[0][0], ppvalE[0][1]);
      Interval valE1dotE1;
      Dot(valE1dotE1, ppvalE[1][0], ppvalE[1][1], ppvalE[1][0], ppvalE[1][1]);
      Interval valE0dotE1;
      Dot(valE0dotE1, ppvalE[0][0], ppvalE[0][1], ppvalE[1][0], ppvalE[1][1]);
      Interval valE1dotE3;
      Dot(valE1dotE3, ppvalE[1][0], ppvalE[1][1], ppvalE[3][0], ppvalE[3][1]);

      Interval valE1xE3;
      Cross(valE1xE3, ppvalE[1][0], ppvalE[1][1], ppvalE[3][0], ppvalE[3][1]);
      Interval valE00xE01;
      Cross(valE00xE01, pppvalV[0][0][0], pppvalV[0][0][1], pppvalV[0][1][0], pppvalV[0][1][1]);
      Interval valE30xE31;
      Cross(valE30xE31, pppvalV[3][0][0], pppvalV[3][0][1], pppvalV[3][1][0], pppvalV[3][1][1]);

      Interval valE0E1minusE0dotE1;
      valE0E1minusE0dotE1 = sqrt(valE0dotE0*valE1dotE1) - valE0dotE1;

      Interval valResult;
      valResult = ( valE0E1minusE0dotE1*valE30xE31*( valE0E1minusE0dotE1*valE30xE31
					       + 2.0*valE00xE01*valE1dotE3 )
		  - valE00xE01*valE00xE01*valE1xE3*valE1xE3 );

      iResult = ( valResult < 0.0 );
      if ( iResult == -1 )
	{
	  break;
	}
      if ( iResult == 1 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( valResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
      return APM2D::cttOnCircle;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( iResult == -1 )
    {
      APM2D::m_uiInCircleEEv_EMPFUN++;

      mp_real rE00_x = dE00_x;
      mp_real rE00_y = dE00_y; 
      mp_real rE01_x = dE01_x;
      mp_real rE01_y = dE01_y;
      mp_real rE10_x = dE10_x;
      mp_real rE10_y = dE10_y; 
      mp_real rE11_V2_x = dE11_V2_x;
      mp_real rE11_V2_y = dE11_V2_y;
      mp_real rE30_x = dE30_x;
      mp_real rE30_y = dE30_y;
      mp_real rE31_x = dE31_x;
      mp_real rE31_y = dE31_y;
  
      mp_real ppprV[4][2][2]; // Endpoints
      // Translate V1 to the origin
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = rE00_x - rE11_V2_x;
      ppprV[0][0][1] = rE00_y - rE11_V2_y;
      ppprV[0][1][0] = rE01_x - rE11_V2_x;
      ppprV[0][1][1] = rE01_y - rE11_V2_y;
      ppprV[1][0][0] = rE10_x - rE11_V2_x;
      ppprV[1][0][1] = rE10_y - rE11_V2_y;
      ppprV[1][1][0] = 0.0;
      ppprV[1][1][1] = 0.0;
      ppprV[2][0][0] = 0.0;
      ppprV[2][0][1] = 0.0;
      ppprV[3][0][0] = rE30_x - rE11_V2_x;
      ppprV[3][0][1] = rE30_y - rE11_V2_y;
      ppprV[3][1][0] = rE31_x - rE11_V2_x;
      ppprV[3][1][1] = rE31_y - rE11_V2_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[0][0] = rE01_x - rE00_x;
      pprE[0][1] = rE01_y - rE00_y;
      pprE[1][0] = rE11_V2_x - rE10_x;
      pprE[1][1] = rE11_V2_y - rE10_y;
      pprE[3][0] = rE31_x - rE30_x;
      pprE[3][1] = rE31_y - rE30_y;
  
      mp_real rE0dotE0;
      Dot(rE0dotE0, pprE[0][0], pprE[0][1], pprE[0][0], pprE[0][1]);
      mp_real rE1dotE1;
      Dot(rE1dotE1, pprE[1][0], pprE[1][1], pprE[1][0], pprE[1][1]);
      mp_real rE0dotE1;
      Dot(rE0dotE1, pprE[0][0], pprE[0][1], pprE[1][0], pprE[1][1]);
      mp_real rE1dotE3;
      Dot(rE1dotE3, pprE[1][0], pprE[1][1], pprE[3][0], pprE[3][1]);

      mp_real rE1xE3;
      Cross(rE1xE3, pprE[1][0], pprE[1][1], pprE[3][0], pprE[3][1]);
      mp_real rE00xE01;
      Cross(rE00xE01, ppprV[0][0][0], ppprV[0][0][1], ppprV[0][1][0], ppprV[0][1][1]);
      mp_real rE30xE31;
      Cross(rE30xE31, ppprV[3][0][0], ppprV[3][0][1], ppprV[3][1][0], ppprV[3][1][1]);

      mp_real rE0E1minusE0dotE1;
      rE0E1minusE0dotE1 = sqrt(rE0dotE0*rE1dotE1) - rE0dotE1;

      mp_real rResult;
      rResult = ( rE0E1minusE0dotE1*rE30xE31*( rE0E1minusE0dotE1*rE30xE31
					       + 2.0*rE00xE01*rE1dotE3 )
		  - rE00xE01*rE00xE01*rE1xE3*rE1xE3 );
  
      if ( rResult < 0.0 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( rResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
    }

  return APM2D::cttOnCircle;
}

//////////////////////////////////////////////////
// EEE

class ICC_EEE_MPFUNFilter
{
protected:
  enum InstructionPointer {ipStart, ipDen, 
			   ipEdgeZones, ipCCW, ipEnd};

public:
  ICC_EEE_MPFUNFilter();
  ~ICC_EEE_MPFUNFilter();

public:
  APM2D::InscribedCircleCenterType 
  InscribedCircleCenterEEE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
			   DOUBLE dE00_x, DOUBLE dE00_y, 
			   DOUBLE dE01_x, DOUBLE dE01_y,
			   DOUBLE dE10_x, DOUBLE dE10_y, 
			   DOUBLE dE11_x, DOUBLE dE11_y,
			   DOUBLE dE20_x, DOUBLE dE20_y, 
			   DOUBLE dE21_x, DOUBLE dE21_y);
  APM2D::InscribedCircleCenterType 
  InscribedCircleCenterEEEForce(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dE00_x, DOUBLE dE00_y, 
				DOUBLE dE01_x, DOUBLE dE01_y,
				DOUBLE dE10_x, DOUBLE dE10_y, 
				DOUBLE dE11_x, DOUBLE dE11_y,
				DOUBLE dE20_x, DOUBLE dE20_y, 
				DOUBLE dE21_x, DOUBLE dE21_y);

protected:
#ifndef COMPILER_PROBLEM
  BOOL
  InscribedCircleCenterEEEMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dE00_x, DOUBLE dE00_y, 
				DOUBLE dE01_x, DOUBLE dE01_y,
				DOUBLE dE10_x, DOUBLE dE10_y, 
				DOUBLE dE11_x, DOUBLE dE11_y,
				DOUBLE dE20_x, DOUBLE dE20_y, 
				DOUBLE dE21_x, DOUBLE dE21_y,
				InstructionPointer eIPMax = ipEnd);
#else // COMPILER_PROBLEM
  BOOL
  InscribedCircleCenterEEEMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				DOUBLE dE00_x, DOUBLE dE00_y, 
				DOUBLE dE01_x, DOUBLE dE01_y,
				DOUBLE dE10_x, DOUBLE dE10_y, 
				DOUBLE dE11_x, DOUBLE dE11_y,
				DOUBLE dE20_x, DOUBLE dE20_y, 
				DOUBLE dE21_x, DOUBLE dE21_y,
				InstructionPointer eIPMax = ipEnd);
  BOOL EdgeZones(InstructionPointer eIPMax);
  VOID CenterAndRadius(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
		       InstructionPointer eIPMax);
#endif // COMPILER_PROBLEM

protected:
  InstructionPointer m_eIP;
  APM2D::InscribedCircleCenterType m_eResult;
  INT_32 m_iDen;
  INT_32 m_iDenSign;
  INT_32 m_ppiEdgeZone[3][2];
  INT_32 m_iCCW;

  Interval m_valDen;

  mp_real m_rE00_x;
  mp_real m_rE00_y;
  mp_real m_rE01_x;
  mp_real m_rE01_y;
  mp_real m_rE10_x;
  mp_real m_rE10_y;
  mp_real m_rE11_x;
  mp_real m_rE11_y;
  mp_real m_rE20_x;
  mp_real m_rE20_y;
  mp_real m_rE21_x;
  mp_real m_rE21_y;
  mp_real m_pprE[3][2];
  mp_real m_rE0dotE0;
  mp_real m_rE1dotE1;
  mp_real m_rE2dotE2;  
  mp_real m_rE0mag;
  mp_real m_rE1mag;
  mp_real m_rE2mag;
  mp_real m_rE0xE1;
  mp_real m_rE1xE2;
  mp_real m_rE2xE0;
  mp_real m_rDen;
  mp_real m_rE00xE01;
  mp_real m_rE20xE21;
};

ICC_EEE_MPFUNFilter::ICC_EEE_MPFUNFilter()
{
}

ICC_EEE_MPFUNFilter::~ICC_EEE_MPFUNFilter()
{
}

APM2D::InscribedCircleCenterType 
ICC_EEE_MPFUNFilter::InscribedCircleCenterEEE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
					      DOUBLE dE00_x, DOUBLE dE00_y, 
					      DOUBLE dE01_x, DOUBLE dE01_y,
					      DOUBLE dE10_x, DOUBLE dE10_y, 
					      DOUBLE dE11_x, DOUBLE dE11_y,
					      DOUBLE dE20_x, DOUBLE dE20_y, 
					      DOUBLE dE21_x, DOUBLE dE21_y)
{
  APM2D::m_uiInscribedCircleCenterEEEMPFUNFilter++;

  *pdC_x = 0.0;
  *pdC_y = 0.0;
  *pdR = 0.0;
  
  m_eIP = ipStart;
  m_eResult = APM2D::icctNull;
  m_iDen = -1;
  m_iDenSign = -1;
  m_ppiEdgeZone[0][0] = -1;
  m_ppiEdgeZone[0][1] = -1;
  m_ppiEdgeZone[1][0] = -1;
  m_ppiEdgeZone[1][1] = -1;
  m_ppiEdgeZone[2][0] = -1;
  m_ppiEdgeZone[2][1] = -1;
  m_iCCW = -1;

  BOOL bDone;

  Interval valE00_x = dE00_x;
  Interval valE00_y = dE00_y; 
  Interval valE01_x = dE01_x;
  Interval valE01_y = dE01_y;
  Interval valE10_x = dE10_x;
  Interval valE10_y = dE10_y; 
  Interval valE11_x = dE11_x;
  Interval valE11_y = dE11_y;
  Interval valE20_x = dE20_x;
  Interval valE20_y = dE20_y; 
  Interval valE21_x = dE21_x;
  Interval valE21_y = dE21_y;
  
  Interval ppvalE[3][2];     // Edge vectors
  ppvalE[0][0] = valE01_x - valE00_x;
  ppvalE[0][1] = valE01_y - valE00_y;
  ppvalE[1][0] = valE11_x - valE10_x;
  ppvalE[1][1] = valE11_y - valE10_y;
  ppvalE[2][0] = valE21_x - valE20_x;
  ppvalE[2][1] = valE21_y - valE20_y;
  
  Interval valE0dotE0;
  Dot(valE0dotE0, ppvalE[0][0], ppvalE[0][1], ppvalE[0][0], ppvalE[0][1]);
  Interval valE1dotE1;
  Dot(valE1dotE1, ppvalE[1][0], ppvalE[1][1], ppvalE[1][0], ppvalE[1][1]);
  Interval valE2dotE2;
  Dot(valE2dotE2, ppvalE[2][0], ppvalE[2][1], ppvalE[2][0], ppvalE[2][1]);

  Interval valE0mag;
  valE0mag = sqrt(valE0dotE0);
  Interval valE1mag;
  valE1mag = sqrt(valE1dotE1);
  Interval valE2mag;
  valE2mag = sqrt(valE2dotE2);

  Interval valE0xE1;
  Cross(valE0xE1, ppvalE[0][0], ppvalE[0][1], ppvalE[1][0], ppvalE[1][1]);
  Interval valE1xE2;
  Cross(valE1xE2, ppvalE[1][0], ppvalE[1][1], ppvalE[2][0], ppvalE[2][1]);
  Interval valE2xE0;
  Cross(valE2xE0, ppvalE[2][0], ppvalE[2][1], ppvalE[0][0], ppvalE[0][1]);

  m_valDen = valE0mag*valE1xE2 + valE1mag*valE2xE0 + valE2mag*valE0xE1;

  m_iDen = ( m_valDen == 0.0 );
  if ( m_iDen == -1 )
    {
      bDone = InscribedCircleCenterEEEMPFUN(pdC_x, pdC_y, pdR, 
					    dE00_x, dE00_y, 
					    dE01_x, dE01_y,
					    dE10_x, dE10_y, 
					    dE11_x, dE11_y,
					    dE20_x, dE20_y, 
					    dE21_x, dE21_y,
					    ipDen);
      if ( bDone )
	{
	  return m_eResult;
	}
      if ( m_iDen == -1 )
	{
	  bDone = InscribedCircleCenterEEEMPFUN(pdC_x, pdC_y, pdR, 
						dE00_x, dE00_y, 
						dE01_x, dE01_y,
						dE10_x, dE10_y, 
						dE11_x, dE11_y,
						dE20_x, dE20_y, 
						dE21_x, dE21_y,
						ipEnd);
	  ASSERT( bDone );
	  return m_eResult;
	}
    }
  if ( m_iDen == 1 )
    {
      // Colinear
      return APM2D::icctNull;
    }
  m_iDenSign = ( m_valDen < 0.0 );
  ASSERT( m_iDenSign != -1 );

  //////////////////////////////////////////////////
  // Check the Edge Zones

  Interval pppvalV[3][2][2]; // Endpoints
  // Translate V1 to the origin
  pppvalV[0][0][0] = valE00_x - valE10_x;
  pppvalV[0][0][1] = valE00_y - valE10_y;
  pppvalV[0][1][0] = valE01_x - valE10_x;
  pppvalV[0][1][1] = valE01_y - valE10_y;
  pppvalV[1][0][0] = 0.0;
  pppvalV[1][0][1] = 0.0;
  pppvalV[1][1][0] = valE11_x - valE10_x;
  pppvalV[1][1][1] = valE11_y - valE10_y;
  pppvalV[2][0][0] = valE20_x - valE10_x;
  pppvalV[2][0][1] = valE20_y - valE10_y;
  pppvalV[2][1][0] = valE21_x - valE10_x;
  pppvalV[2][1][1] = valE21_y - valE10_y;
  
  Interval valE0dotE1;
  Dot(valE0dotE1, ppvalE[0][0], ppvalE[0][1], ppvalE[1][0], ppvalE[1][1]);
  Interval valE1dotE2;
  Dot(valE1dotE2, ppvalE[1][0], ppvalE[1][1], ppvalE[2][0], ppvalE[2][1]);
  Interval valE0dotE2;
  Dot(valE0dotE2, ppvalE[0][0], ppvalE[0][1], ppvalE[2][0], ppvalE[2][1]);

  Interval valE00xE01;
  Cross(valE00xE01, pppvalV[0][0][0], pppvalV[0][0][1], pppvalV[0][1][0], pppvalV[0][1][1]);
  Interval valE20xE21;
  Cross(valE20xE21, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][1][0], pppvalV[2][1][1]);

  Interval valE0E1minusE0dotE1;
  valE0E1minusE0dotE1 = sqrt(valE0dotE0*valE1dotE1) - valE0dotE1;
  Interval valE1E2minusE1dotE2;
  valE1E2minusE1dotE2 = sqrt(valE1dotE1*valE2dotE2) - valE1dotE2;
  Interval valE0E2minusE0dotE2;
  valE0E2minusE0dotE2 = sqrt(valE0dotE0*valE2dotE2) - valE0dotE2;

  Interval ppvalEdgeZone[3][2];

  Interval valE0dotE00;
  Dot(valE0dotE00, ppvalE[0][0], ppvalE[0][1], pppvalV[0][0][0], pppvalV[0][0][1]);
  ppvalEdgeZone[0][0] = -( m_valDen*valE0dotE00 
			 - (valE1mag*valE0dotE2 - valE2mag*valE0dotE1)*valE00xE01
			 + valE0mag*valE0E1minusE0dotE1*valE20xE21 );
  if ( m_iDenSign == 1 )
    {
      ppvalEdgeZone[0][0] = -ppvalEdgeZone[0][0];
    }
  m_ppiEdgeZone[0][0] = ( ppvalEdgeZone[0][0] < 0.0 );
  if ( m_ppiEdgeZone[0][0] == 1 )
    {
      return APM2D::icctNull;
    }
  Interval valE0dotE01;
  Dot(valE0dotE01, ppvalE[0][0], ppvalE[0][1], pppvalV[0][1][0], pppvalV[0][1][1]);
  ppvalEdgeZone[0][1] = ( m_valDen*valE0dotE01 
			- (valE1mag*valE0dotE2 - valE2mag*valE0dotE1)*valE00xE01
			+ valE0mag*valE0E1minusE0dotE1*valE20xE21 );
  if ( m_iDenSign == 1 )
    {
      ppvalEdgeZone[0][1] = -ppvalEdgeZone[0][1];
    }
  m_ppiEdgeZone[0][1] = ( ppvalEdgeZone[0][1] < 0.0 );
  if ( m_ppiEdgeZone[0][1] == 1 )
    {
      return APM2D::icctNull;
    }
  Interval valE1dotE10;
  Dot(valE1dotE10, ppvalE[1][0], ppvalE[1][1], pppvalV[1][0][0], pppvalV[1][0][1]);
  ppvalEdgeZone[1][0] = -( m_valDen*valE1dotE10 
			 + valE1mag*valE1E2minusE1dotE2*valE00xE01
			 - valE1mag*valE0E1minusE0dotE1*valE20xE21 );
  if ( m_iDenSign == 1 )
    {
      ppvalEdgeZone[1][0] = -ppvalEdgeZone[1][0];
    }
  m_ppiEdgeZone[1][0] = ( ppvalEdgeZone[1][0] < 0.0 );
  if ( m_ppiEdgeZone[1][0] == 1 )
    {
      return APM2D::icctNull;
    }
  Interval valE1dotE11;
  Dot(valE1dotE11, ppvalE[1][0], ppvalE[1][1], pppvalV[1][1][0], pppvalV[1][1][1]);
  ppvalEdgeZone[1][1] = ( m_valDen*valE1dotE11 
			+ valE1mag*valE1E2minusE1dotE2*valE00xE01
			- valE1mag*valE0E1minusE0dotE1*valE20xE21 );
  if ( m_iDenSign == 1 )
    {
      ppvalEdgeZone[1][1] = -ppvalEdgeZone[1][1];
    }
  m_ppiEdgeZone[1][1] = ( ppvalEdgeZone[1][1] < 0.0 );
  if ( m_ppiEdgeZone[1][1] == 1 )
    {
      return APM2D::icctNull;
    }
  Interval valE2dotE20;
  Dot(valE2dotE20, ppvalE[2][0], ppvalE[2][1], pppvalV[2][0][0], pppvalV[2][0][1]);
  ppvalEdgeZone[2][0] = -( m_valDen*valE2dotE20 
			 - valE2mag*valE1E2minusE1dotE2*valE00xE01 
			 - (valE0mag*valE1dotE2 - valE1mag*valE0dotE2)*valE20xE21 ); 
  if ( m_iDenSign == 1 )
    {
      ppvalEdgeZone[2][0] = -ppvalEdgeZone[2][0];
    }
  m_ppiEdgeZone[2][0] = ( ppvalEdgeZone[2][0] < 0.0 );
  if ( m_ppiEdgeZone[2][0] == 1 )
    {
      return APM2D::icctNull;
    }
  Interval valE2dotE21;
  Dot(valE2dotE21, ppvalE[2][0], ppvalE[2][1], pppvalV[2][1][0], pppvalV[2][1][1]);
  ppvalEdgeZone[2][1] = ( m_valDen*valE2dotE21 
			- valE2mag*valE1E2minusE1dotE2*valE00xE01 
			- (valE0mag*valE1dotE2 - valE1mag*valE0dotE2)*valE20xE21 );
  if ( m_iDenSign == 1 )
    {
      ppvalEdgeZone[2][1] = -ppvalEdgeZone[2][1];
    }
  m_ppiEdgeZone[2][1] = ( ppvalEdgeZone[2][1] < 0.0 );
  if ( m_ppiEdgeZone[2][1] == 1 )
    {
      return APM2D::icctNull;
    }

  if ( m_ppiEdgeZone[0][0] == -1 ||
       m_ppiEdgeZone[0][1] == -1 ||
       m_ppiEdgeZone[1][0] == -1 ||
       m_ppiEdgeZone[1][1] == -1 ||
       m_ppiEdgeZone[2][0] == -1 ||
       m_ppiEdgeZone[2][1] == -1 )
    {
      bDone = InscribedCircleCenterEEEMPFUN(pdC_x, pdC_y, pdR, 
					    dE00_x, dE00_y, 
					    dE01_x, dE01_y,
					    dE10_x, dE10_y, 
					    dE11_x, dE11_y,
					    dE20_x, dE20_y, 
					    dE21_x, dE21_y,
					    ipEdgeZones);
      if ( bDone )
	{
	  return m_eResult;
	}
    }

  //////////////////////////////////////////////////
  // Compute the center and radius

  Interval pvalC[2];  
  pvalC[0] = ( (valE1mag*ppvalE[2][0] - valE2mag*ppvalE[1][0])*valE00xE01
	     + (valE0mag*ppvalE[1][0] - valE1mag*ppvalE[0][0])*valE20xE21 ) / m_valDen;
  pvalC[1] = ( (valE1mag*ppvalE[2][1] - valE2mag*ppvalE[1][1])*valE00xE01
	     + (valE0mag*ppvalE[1][1] - valE1mag*ppvalE[0][1])*valE20xE21 ) / m_valDen;

  Interval valR;  
  valR = ( valE0xE1*valE20xE21 + valE1xE2*valE00xE01 ) / m_valDen;
  
  //////////////////////////////////////////////////
  // Check if the projection of the center on to the sites is in CCW orrer

  Interval valE0dotC, valE1dotC, valE2dotC;
  Dot(valE0dotC, ppvalE[0][0], ppvalE[0][1], pvalC[0], pvalC[1]);
  Dot(valE1dotC, ppvalE[1][0], ppvalE[1][1], pvalC[0], pvalC[1]);
  Dot(valE2dotC, ppvalE[2][0], ppvalE[2][1], pvalC[0], pvalC[1]);

  Interval ppvalProjection[3][2];
  ppvalProjection[0][0] = (-valE00xE01*(-ppvalE[0][1]) + valE0dotC*ppvalE[0][0]) / valE0dotE0;
  ppvalProjection[0][1] = (-valE00xE01*(ppvalE[0][0]) + valE0dotC*ppvalE[0][1]) / valE0dotE0;

  ppvalProjection[1][0] = (valE1dotC*ppvalE[1][0]) / valE1dotE1;
  ppvalProjection[1][1] = (valE1dotC*ppvalE[1][1]) / valE1dotE1;

  ppvalProjection[2][0] = (-valE20xE21*(-ppvalE[2][1]) + valE2dotC*ppvalE[2][0]) / valE2dotE2;
  ppvalProjection[2][1] = (-valE20xE21*(ppvalE[2][0]) + valE2dotC*ppvalE[2][1]) / valE2dotE2;

  Interval valCCWCheck;
  valCCWCheck = ( ( (ppvalProjection[2][0] - ppvalProjection[1][0]) *
		  (ppvalProjection[0][1] - ppvalProjection[1][1]) ) -
		( (ppvalProjection[2][1] - ppvalProjection[1][1]) *
		  (ppvalProjection[0][0] - ppvalProjection[1][0]) ) );

  m_iCCW = ( valCCWCheck < 0.0 );
  if ( m_iCCW == -1 )
    {
      bDone = InscribedCircleCenterEEEMPFUN(pdC_x, pdC_y, pdR, 
					    dE00_x, dE00_y, 
					    dE01_x, dE01_y,
					    dE10_x, dE10_y, 
					    dE11_x, dE11_y,
					    dE20_x, dE20_y, 
					    dE21_x, dE21_y,
					    ipCCW);
      if ( bDone )
	{
	  return m_eResult;
	}
    }
  if ( m_iCCW == 1 )
    {
      return APM2D::icctNull;
    }

  // Translate the answer back out away from the origin
  pvalC[0] += valE10_x;
  pvalC[1] += valE10_y;

  *pdC_x = dble(pvalC[0]);
  *pdC_y = dble(pvalC[1]);
  *pdR = dble(valR);
  return APM2D::icctEEE;
}

APM2D::InscribedCircleCenterType 
ICC_EEE_MPFUNFilter::InscribedCircleCenterEEEForce(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
						   DOUBLE dE00_x, DOUBLE dE00_y, 
						   DOUBLE dE01_x, DOUBLE dE01_y,
						   DOUBLE dE10_x, DOUBLE dE10_y, 
						   DOUBLE dE11_x, DOUBLE dE11_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y)
{
  *pdC_x = 0.0;
  *pdC_y = 0.0;
  *pdR = 0.0;
  
  m_eIP = ipStart;
  m_eResult = APM2D::icctNull;
  m_iDen = -1;
  m_iDenSign = -1;
  m_ppiEdgeZone[0][0] = -1;
  m_ppiEdgeZone[0][1] = -1;
  m_ppiEdgeZone[1][0] = -1;
  m_ppiEdgeZone[1][1] = -1;
  m_ppiEdgeZone[2][0] = -1;
  m_ppiEdgeZone[2][1] = -1;
  m_iCCW = -1;

  BOOL bDone;

  bDone = InscribedCircleCenterEEEMPFUN(pdC_x, pdC_y, pdR, 
					dE00_x, dE00_y, 
					dE01_x, dE01_y,
					dE10_x, dE10_y, 
					dE11_x, dE11_y,
					dE20_x, dE20_y, 
					dE21_x, dE21_y,
					ipEnd);
  ASSERT( bDone );
  return m_eResult;
}

#ifndef COMPILER_PROBLEM

BOOL 
ICC_EEE_MPFUNFilter::InscribedCircleCenterEEEMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
						   DOUBLE dE00_x, DOUBLE dE00_y, 
						   DOUBLE dE01_x, DOUBLE dE01_y,
						   DOUBLE dE10_x, DOUBLE dE10_y, 
						   DOUBLE dE11_x, DOUBLE dE11_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   InstructionPointer eIPMax/* = ipEnd*/)
{
  APM2D::m_uiInscribedCircleCenterEEEMPFUN++;

  switch ( m_eIP )
    {
    case ipStart:
      {
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);
    
	*pdC_x = 0.0;
	*pdC_y = 0.0;
	*pdR = 0.0;
    
	m_rE00_x = dE00_x;
	m_rE00_y = dE00_y; 
	m_rE01_x = dE01_x;
	m_rE01_y = dE01_y;
	m_rE10_x = dE10_x;
	m_rE10_y = dE10_y; 
	m_rE11_x = dE11_x;
	m_rE11_y = dE11_y;
	m_rE20_x = dE20_x;
	m_rE20_y = dE20_y; 
	m_rE21_x = dE21_x;
	m_rE21_y = dE21_y;
  
	m_pprE[0][0] = m_rE01_x - m_rE00_x;
	m_pprE[0][1] = m_rE01_y - m_rE00_y;
	m_pprE[1][0] = m_rE11_x - m_rE10_x;
	m_pprE[1][1] = m_rE11_y - m_rE10_y;
	m_pprE[2][0] = m_rE21_x - m_rE20_x;
	m_pprE[2][1] = m_rE21_y - m_rE20_y;
  
	Dot(m_rE0dotE0, m_pprE[0][0], m_pprE[0][1], m_pprE[0][0], m_pprE[0][1]);
	Dot(m_rE1dotE1, m_pprE[1][0], m_pprE[1][1], m_pprE[1][0], m_pprE[1][1]);
	Dot(m_rE2dotE2, m_pprE[2][0], m_pprE[2][1], m_pprE[2][0], m_pprE[2][1]);

	m_rE0mag = sqrt(m_rE0dotE0);
	m_rE1mag = sqrt(m_rE1dotE1);
	m_rE2mag = sqrt(m_rE2dotE2);

	Cross(m_rE0xE1, m_pprE[0][0], m_pprE[0][1], m_pprE[1][0], m_pprE[1][1]);
	Cross(m_rE1xE2, m_pprE[1][0], m_pprE[1][1], m_pprE[2][0], m_pprE[2][1]);
	Cross(m_rE2xE0, m_pprE[2][0], m_pprE[2][1], m_pprE[0][0], m_pprE[0][1]);

	m_rDen = m_rE0mag*m_rE1xE2 + m_rE1mag*m_rE2xE0 + m_rE2mag*m_rE0xE1;
	if ( m_iDen == -1 )
	  {
	    if ( m_rDen == 0.0 )
	      {
		// Colinear
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterEEEMPFUNFinal++;
		return TRUE;
	      }
	    m_iDenSign = ( m_rDen < 0.0 );
	    if ( eIPMax != ipEnd )
	      { 
		m_valDen.ConservativeBound(dble(m_rDen));	    
		m_iDen = ( m_valDen == 0.0 );      
	      }
	  }

      }
    case ipDen:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);
	
	//////////////////////////////////////////////////
	// Check the Edge Zones

	mp_real ppprV[3][2][2]; // Endpoints
	// Translate V1 to the origin
	ppprV[0][0][0] = m_rE00_x - m_rE10_x;
	ppprV[0][0][1] = m_rE00_y - m_rE10_y;
	ppprV[0][1][0] = m_rE01_x - m_rE10_x;
	ppprV[0][1][1] = m_rE01_y - m_rE10_y;
	ppprV[1][0][0] = 0.0;
	ppprV[1][0][1] = 0.0;
	ppprV[1][1][0] = m_rE11_x - m_rE10_x;
	ppprV[1][1][1] = m_rE11_y - m_rE10_y;
	ppprV[2][0][0] = m_rE20_x - m_rE10_x;
	ppprV[2][0][1] = m_rE20_y - m_rE10_y;
	ppprV[2][1][0] = m_rE21_x - m_rE10_x;
	ppprV[2][1][1] = m_rE21_y - m_rE10_y;
  
	Cross(m_rE00xE01, ppprV[0][0][0], ppprV[0][0][1], ppprV[0][1][0], ppprV[0][1][1]);
	Cross(m_rE20xE21, ppprV[2][0][0], ppprV[2][0][1], ppprV[2][1][0], ppprV[2][1][1]);

	mp_real rE0dotE1;
	Dot(rE0dotE1, m_pprE[0][0], m_pprE[0][1], m_pprE[1][0], m_pprE[1][1]);
	mp_real rE1dotE2;
	Dot(rE1dotE2, m_pprE[1][0], m_pprE[1][1], m_pprE[2][0], m_pprE[2][1]);
	mp_real rE0dotE2;
	Dot(rE0dotE2, m_pprE[0][0], m_pprE[0][1], m_pprE[2][0], m_pprE[2][1]);

	if ( m_ppiEdgeZone[0][0] == -1 ||
	     m_ppiEdgeZone[0][1] == -1 ||
	     m_ppiEdgeZone[1][0] == -1 ||
	     m_ppiEdgeZone[1][1] == -1 ||
	     m_ppiEdgeZone[2][0] == -1 ||
	     m_ppiEdgeZone[2][1] == -1 )
	  {
	    mp_real rE0E1minusE0dotE1;
	    rE0E1minusE0dotE1 = sqrt(m_rE0dotE0*m_rE1dotE1) - rE0dotE1;
	    mp_real rE1E2minusE1dotE2;
	    rE1E2minusE1dotE2 = sqrt(m_rE1dotE1*m_rE2dotE2) - rE1dotE2;
	    mp_real rE0E2minusE0dotE2;
	    rE0E2minusE0dotE2 = sqrt(m_rE0dotE0*m_rE2dotE2) - rE0dotE2;
	    
	    mp_real pprEdgeZone[3][2];

	    if ( m_ppiEdgeZone[0][0] == -1 )
	      {
		mp_real rE0dotE00;
		Dot(rE0dotE00, m_pprE[0][0], m_pprE[0][1], ppprV[0][0][0], ppprV[0][0][1]);
		pprEdgeZone[0][0] = -( m_rDen*rE0dotE00 
				       - (m_rE1mag*rE0dotE2 - m_rE2mag*rE0dotE1)*m_rE00xE01
				       + m_rE0mag*rE0E1minusE0dotE1*m_rE20xE21 );
		if ( ( m_iDenSign == 0 && pprEdgeZone[0][0] < 0.0 ) ||
		     ( m_iDenSign == 1 && pprEdgeZone[0][0] > 0.0 ) )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterEEEMPFUNFinal++;
		    return TRUE;
		  }
	      }
	    if ( m_ppiEdgeZone[0][1] == -1 )
	      {
		mp_real rE0dotE01;
		Dot(rE0dotE01, m_pprE[0][0], m_pprE[0][1], ppprV[0][1][0], ppprV[0][1][1]);
		pprEdgeZone[0][1] = ( m_rDen*rE0dotE01 
				      - (m_rE1mag*rE0dotE2 - m_rE2mag*rE0dotE1)*m_rE00xE01
				      + m_rE0mag*rE0E1minusE0dotE1*m_rE20xE21 );
		if ( ( m_iDenSign == 0 && pprEdgeZone[0][1] < 0.0 ) ||
		     ( m_iDenSign == 1 && pprEdgeZone[0][1] > 0.0 ) )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterEEEMPFUNFinal++;
		    return TRUE;
		  }
	      }
	    if ( m_ppiEdgeZone[1][0] == -1 )
	      {
		mp_real rE1dotE10;
		Dot(rE1dotE10, m_pprE[1][0], m_pprE[1][1], ppprV[1][0][0], ppprV[1][0][1]);
		pprEdgeZone[1][0] = -( m_rDen*rE1dotE10 
				       + m_rE1mag*rE1E2minusE1dotE2*m_rE00xE01
				       - m_rE1mag*rE0E1minusE0dotE1*m_rE20xE21 );
		if ( ( m_iDenSign == 0 && pprEdgeZone[1][0] < 0.0 ) ||
		     ( m_iDenSign == 1 && pprEdgeZone[1][0] > 0.0 ) )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterEEEMPFUNFinal++;
		    return TRUE;
		  }
	      }
	    if ( m_ppiEdgeZone[1][1] == -1 )
	      {
		mp_real rE1dotE11;
		Dot(rE1dotE11, m_pprE[1][0], m_pprE[1][1], ppprV[1][1][0], ppprV[1][1][1]);
		pprEdgeZone[1][1] = ( m_rDen*rE1dotE11 
				      + m_rE1mag*rE1E2minusE1dotE2*m_rE00xE01
				      - m_rE1mag*rE0E1minusE0dotE1*m_rE20xE21 );
		if ( ( m_iDenSign == 0 && pprEdgeZone[1][1] < 0.0 ) ||
		     ( m_iDenSign == 1 && pprEdgeZone[1][1] > 0.0 ) )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterEEEMPFUNFinal++;
		    return TRUE;
		  }
	      }
	    if ( m_ppiEdgeZone[2][0] == -1 )
	      {
		mp_real rE2dotE20;
		Dot(rE2dotE20, m_pprE[2][0], m_pprE[2][1], ppprV[2][0][0], ppprV[2][0][1]);
		pprEdgeZone[2][0] = -( m_rDen*rE2dotE20 
				       - m_rE2mag*rE1E2minusE1dotE2*m_rE00xE01 
				       - (m_rE0mag*rE1dotE2 - m_rE1mag*rE0dotE2)*m_rE20xE21 ); 
		if ( ( m_iDenSign == 0 && pprEdgeZone[2][0] < 0.0 ) ||
		     ( m_iDenSign == 1 && pprEdgeZone[2][0] > 0.0 ) )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterEEEMPFUNFinal++;
		    return TRUE;
		  }
	      }
	    if ( m_ppiEdgeZone[2][1] == -1 )
	      {
		mp_real rE2dotE21;
		Dot(rE2dotE21, m_pprE[2][0], m_pprE[2][1], ppprV[2][1][0], ppprV[2][1][1]);
		pprEdgeZone[2][1] = ( m_rDen*rE2dotE21 
				      - m_rE2mag*rE1E2minusE1dotE2*m_rE00xE01 
				      - (m_rE0mag*rE1dotE2 - m_rE1mag*rE0dotE2)*m_rE20xE21 );
		if ( ( m_iDenSign == 0 && pprEdgeZone[2][1] < 0.0 ) ||
		     ( m_iDenSign == 1 && pprEdgeZone[2][1] > 0.0 ) )
		  {
		    m_eResult = APM2D::icctNull;
		    APM2D::m_uiInscribedCircleCenterEEEMPFUNFinal++;
		    return TRUE;
		  }
	      }
	  }

      }
    case ipEdgeZones:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	//////////////////////////////////////////////////
	// Compute the center and radius

	mp_real prC[2];  
	prC[0] = ( (m_rE1mag*m_pprE[2][0] - m_rE2mag*m_pprE[1][0])*m_rE00xE01
		   + (m_rE0mag*m_pprE[1][0] - m_rE1mag*m_pprE[0][0])*m_rE20xE21 ) / m_rDen;
	prC[1] = ( (m_rE1mag*m_pprE[2][1] - m_rE2mag*m_pprE[1][1])*m_rE00xE01
		   + (m_rE0mag*m_pprE[1][1] - m_rE1mag*m_pprE[0][1])*m_rE20xE21 ) / m_rDen;

	mp_real rR;  
	rR = ( m_rE0xE1*m_rE20xE21 + m_rE1xE2*m_rE00xE01 ) / m_rDen;
  
	//////////////////////////////////////////////////
	// Check if the projection of the center on to the sites is in CCW orrer

	if ( m_iCCW == -1 )
	  {
	    mp_real rE0dotC, rE1dotC, rE2dotC;
	    Dot(rE0dotC, m_pprE[0][0], m_pprE[0][1], prC[0], prC[1]);
	    Dot(rE1dotC, m_pprE[1][0], m_pprE[1][1], prC[0], prC[1]);
	    Dot(rE2dotC, m_pprE[2][0], m_pprE[2][1], prC[0], prC[1]);

	    mp_real pprProjection[3][2];
	    pprProjection[0][0] = (-m_rE00xE01*(-m_pprE[0][1]) + rE0dotC*m_pprE[0][0]) / m_rE0dotE0;
	    pprProjection[0][1] = (-m_rE00xE01*(m_pprE[0][0]) + rE0dotC*m_pprE[0][1]) / m_rE0dotE0;

	    pprProjection[1][0] = (rE1dotC*m_pprE[1][0]) / m_rE1dotE1;
	    pprProjection[1][1] = (rE1dotC*m_pprE[1][1]) / m_rE1dotE1;

	    pprProjection[2][0] = (-m_rE20xE21*(-m_pprE[2][1]) + rE2dotC*m_pprE[2][0]) / m_rE2dotE2;
	    pprProjection[2][1] = (-m_rE20xE21*(m_pprE[2][0]) + rE2dotC*m_pprE[2][1]) / m_rE2dotE2;

	    mp_real rCCWCheck;
	    rCCWCheck = ( ( (pprProjection[2][0] - pprProjection[1][0]) *
			    (pprProjection[0][1] - pprProjection[1][1]) ) -
			  ( (pprProjection[2][1] - pprProjection[1][1]) *
			    (pprProjection[0][0] - pprProjection[1][0]) ) );

	    if ( rCCWCheck < 0.0 )
	      {
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterEEEMPFUNFinal++;
		return TRUE;
	      }
	  }

	// Translate the answer back out away from the origin
	prC[0] += m_rE10_x;
	prC[1] += m_rE10_y;

	*pdC_x = dble(prC[0]);
	*pdC_y = dble(prC[1]);
	*pdR = dble(rR);
	m_eResult = APM2D::icctEEE;
      }
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  APM2D::m_uiInscribedCircleCenterEEEMPFUNFinal++;
  return TRUE;
}

#else // COMPILER_PROBLEM

BOOL 
ICC_EEE_MPFUNFilter::InscribedCircleCenterEEEMPFUN(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
						   DOUBLE dE00_x, DOUBLE dE00_y, 
						   DOUBLE dE01_x, DOUBLE dE01_y,
						   DOUBLE dE10_x, DOUBLE dE10_y, 
						   DOUBLE dE11_x, DOUBLE dE11_y,
						   DOUBLE dE20_x, DOUBLE dE20_y, 
						   DOUBLE dE21_x, DOUBLE dE21_y,
						   InstructionPointer eIPMax/* = ipEnd*/)
{
  APM2D::m_uiInscribedCircleCenterEEEMPFUN++;

  switch ( m_eIP )
    {
    case ipStart:
      {
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);
    
	*pdC_x = 0.0;
	*pdC_y = 0.0;
	*pdR = 0.0;
    
	m_rE00_x = dE00_x;
	m_rE00_y = dE00_y; 
	m_rE01_x = dE01_x;
	m_rE01_y = dE01_y;
	m_rE10_x = dE10_x;
	m_rE10_y = dE10_y; 
	m_rE11_x = dE11_x;
	m_rE11_y = dE11_y;
	m_rE20_x = dE20_x;
	m_rE20_y = dE20_y; 
	m_rE21_x = dE21_x;
	m_rE21_y = dE21_y;
  
	m_pprE[0][0] = m_rE01_x - m_rE00_x;
	m_pprE[0][1] = m_rE01_y - m_rE00_y;
	m_pprE[1][0] = m_rE11_x - m_rE10_x;
	m_pprE[1][1] = m_rE11_y - m_rE10_y;
	m_pprE[2][0] = m_rE21_x - m_rE20_x;
	m_pprE[2][1] = m_rE21_y - m_rE20_y;
  
	Dot(m_rE0dotE0, m_pprE[0][0], m_pprE[0][1], m_pprE[0][0], m_pprE[0][1]);
	Dot(m_rE1dotE1, m_pprE[1][0], m_pprE[1][1], m_pprE[1][0], m_pprE[1][1]);
	Dot(m_rE2dotE2, m_pprE[2][0], m_pprE[2][1], m_pprE[2][0], m_pprE[2][1]);

	m_rE0mag = sqrt(m_rE0dotE0);
	m_rE1mag = sqrt(m_rE1dotE1);
	m_rE2mag = sqrt(m_rE2dotE2);

	Cross(m_rE0xE1, m_pprE[0][0], m_pprE[0][1], m_pprE[1][0], m_pprE[1][1]);
	Cross(m_rE1xE2, m_pprE[1][0], m_pprE[1][1], m_pprE[2][0], m_pprE[2][1]);
	Cross(m_rE2xE0, m_pprE[2][0], m_pprE[2][1], m_pprE[0][0], m_pprE[0][1]);

	m_rDen = m_rE0mag*m_rE1xE2 + m_rE1mag*m_rE2xE0 + m_rE2mag*m_rE0xE1;
	if ( m_iDen == -1 )
	  {
	    if ( m_rDen == 0.0 )
	      {
		// Colinear
		m_eResult = APM2D::icctNull;
		APM2D::m_uiInscribedCircleCenterEEEMPFUNFinal++;
		return TRUE;
	      }
	    m_iDenSign = ( m_rDen < 0.0 );
	    if ( eIPMax != ipEnd )
	      { 
		m_valDen.ConservativeBound(dble(m_rDen));	    
		m_iDen = ( m_valDen == 0.0 );      
	      }
	  }

      }
    case ipDen:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);
	
	if ( EdgeZones(eIPMax) )
	  {
	    APM2D::m_uiInscribedCircleCenterEEEMPFUNFinal++;
	    return TRUE;
	  }
      }
    case ipEdgeZones:
      {
	if ( m_eIP >= eIPMax )
	  {
	    return FALSE;
	  }
	m_eIP = (InstructionPointer)((unsigned int)m_eIP + 1);

	CenterAndRadius(pdC_x, pdC_y, pdR, 
			eIPMax);
      }
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  APM2D::m_uiInscribedCircleCenterEEEMPFUNFinal++;
  return TRUE;
}

BOOL 
ICC_EEE_MPFUNFilter::EdgeZones(InstructionPointer /*eIPMax*/)
{
  //////////////////////////////////////////////////
  // Check the Edge Zones

  mp_real ppprV[3][2][2]; // Endpoints
  // Translate V1 to the origin
  ppprV[0][0][0] = m_rE00_x - m_rE10_x;
  ppprV[0][0][1] = m_rE00_y - m_rE10_y;
  ppprV[0][1][0] = m_rE01_x - m_rE10_x;
  ppprV[0][1][1] = m_rE01_y - m_rE10_y;
  ppprV[1][0][0] = 0.0;
  ppprV[1][0][1] = 0.0;
  ppprV[1][1][0] = m_rE11_x - m_rE10_x;
  ppprV[1][1][1] = m_rE11_y - m_rE10_y;
  ppprV[2][0][0] = m_rE20_x - m_rE10_x;
  ppprV[2][0][1] = m_rE20_y - m_rE10_y;
  ppprV[2][1][0] = m_rE21_x - m_rE10_x;
  ppprV[2][1][1] = m_rE21_y - m_rE10_y;
  
  Cross(m_rE00xE01, ppprV[0][0][0], ppprV[0][0][1], ppprV[0][1][0], ppprV[0][1][1]);
  Cross(m_rE20xE21, ppprV[2][0][0], ppprV[2][0][1], ppprV[2][1][0], ppprV[2][1][1]);

  mp_real rE0dotE1;
  Dot(rE0dotE1, m_pprE[0][0], m_pprE[0][1], m_pprE[1][0], m_pprE[1][1]);
  mp_real rE1dotE2;
  Dot(rE1dotE2, m_pprE[1][0], m_pprE[1][1], m_pprE[2][0], m_pprE[2][1]);
  mp_real rE0dotE2;
  Dot(rE0dotE2, m_pprE[0][0], m_pprE[0][1], m_pprE[2][0], m_pprE[2][1]);

  if ( m_ppiEdgeZone[0][0] == -1 ||
       m_ppiEdgeZone[0][1] == -1 ||
       m_ppiEdgeZone[1][0] == -1 ||
       m_ppiEdgeZone[1][1] == -1 ||
       m_ppiEdgeZone[2][0] == -1 ||
       m_ppiEdgeZone[2][1] == -1 )
    {
      mp_real rE0E1minusE0dotE1;
      rE0E1minusE0dotE1 = sqrt(m_rE0dotE0*m_rE1dotE1) - rE0dotE1;
      mp_real rE1E2minusE1dotE2;
      rE1E2minusE1dotE2 = sqrt(m_rE1dotE1*m_rE2dotE2) - rE1dotE2;
      mp_real rE0E2minusE0dotE2;
      rE0E2minusE0dotE2 = sqrt(m_rE0dotE0*m_rE2dotE2) - rE0dotE2;
	    
      mp_real pprEdgeZone[3][2];

      if ( m_ppiEdgeZone[0][0] == -1 )
	{
	  mp_real rE0dotE00;
	  Dot(rE0dotE00, m_pprE[0][0], m_pprE[0][1], ppprV[0][0][0], ppprV[0][0][1]);
	  pprEdgeZone[0][0] = -( m_rDen*rE0dotE00 
				 - (m_rE1mag*rE0dotE2 - m_rE2mag*rE0dotE1)*m_rE00xE01
				 + m_rE0mag*rE0E1minusE0dotE1*m_rE20xE21 );
	  if ( ( m_iDenSign == 0 && pprEdgeZone[0][0] < 0.0 ) ||
	       ( m_iDenSign == 1 && pprEdgeZone[0][0] > 0.0 ) )
	    {
	      m_eResult = APM2D::icctNull;
	      return TRUE;
	    }
	}
      if ( m_ppiEdgeZone[0][1] == -1 )
	{
	  mp_real rE0dotE01;
	  Dot(rE0dotE01, m_pprE[0][0], m_pprE[0][1], ppprV[0][1][0], ppprV[0][1][1]);
	  pprEdgeZone[0][1] = ( m_rDen*rE0dotE01 
				- (m_rE1mag*rE0dotE2 - m_rE2mag*rE0dotE1)*m_rE00xE01
				+ m_rE0mag*rE0E1minusE0dotE1*m_rE20xE21 );
	  if ( ( m_iDenSign == 0 && pprEdgeZone[0][1] < 0.0 ) ||
	       ( m_iDenSign == 1 && pprEdgeZone[0][1] > 0.0 ) )
	    {
	      m_eResult = APM2D::icctNull;
	      return TRUE;
	    }
	}
      if ( m_ppiEdgeZone[1][0] == -1 )
	{
	  mp_real rE1dotE10;
	  Dot(rE1dotE10, m_pprE[1][0], m_pprE[1][1], ppprV[1][0][0], ppprV[1][0][1]);
	  pprEdgeZone[1][0] = -( m_rDen*rE1dotE10 
				 + m_rE1mag*rE1E2minusE1dotE2*m_rE00xE01
				 - m_rE1mag*rE0E1minusE0dotE1*m_rE20xE21 );
	  if ( ( m_iDenSign == 0 && pprEdgeZone[1][0] < 0.0 ) ||
	       ( m_iDenSign == 1 && pprEdgeZone[1][0] > 0.0 ) )
	    {
	      m_eResult = APM2D::icctNull;
	      return TRUE;
	    }
	}
      if ( m_ppiEdgeZone[1][1] == -1 )
	{
	  mp_real rE1dotE11;
	  Dot(rE1dotE11, m_pprE[1][0], m_pprE[1][1], ppprV[1][1][0], ppprV[1][1][1]);
	  pprEdgeZone[1][1] = ( m_rDen*rE1dotE11 
				+ m_rE1mag*rE1E2minusE1dotE2*m_rE00xE01
				- m_rE1mag*rE0E1minusE0dotE1*m_rE20xE21 );
	  if ( ( m_iDenSign == 0 && pprEdgeZone[1][1] < 0.0 ) ||
	       ( m_iDenSign == 1 && pprEdgeZone[1][1] > 0.0 ) )
	    {
	      m_eResult = APM2D::icctNull;
	      return TRUE;
	    }
	}
      if ( m_ppiEdgeZone[2][0] == -1 )
	{
	  mp_real rE2dotE20;
	  Dot(rE2dotE20, m_pprE[2][0], m_pprE[2][1], ppprV[2][0][0], ppprV[2][0][1]);
	  pprEdgeZone[2][0] = -( m_rDen*rE2dotE20 
				 - m_rE2mag*rE1E2minusE1dotE2*m_rE00xE01 
				 - (m_rE0mag*rE1dotE2 - m_rE1mag*rE0dotE2)*m_rE20xE21 ); 
	  if ( ( m_iDenSign == 0 && pprEdgeZone[2][0] < 0.0 ) ||
	       ( m_iDenSign == 1 && pprEdgeZone[2][0] > 0.0 ) )
	    {
	      m_eResult = APM2D::icctNull;
	      return TRUE;
	    }
	}
      if ( m_ppiEdgeZone[2][1] == -1 )
	{
	  mp_real rE2dotE21;
	  Dot(rE2dotE21, m_pprE[2][0], m_pprE[2][1], ppprV[2][1][0], ppprV[2][1][1]);
	  pprEdgeZone[2][1] = ( m_rDen*rE2dotE21 
				- m_rE2mag*rE1E2minusE1dotE2*m_rE00xE01 
				- (m_rE0mag*rE1dotE2 - m_rE1mag*rE0dotE2)*m_rE20xE21 );
	  if ( ( m_iDenSign == 0 && pprEdgeZone[2][1] < 0.0 ) ||
	       ( m_iDenSign == 1 && pprEdgeZone[2][1] > 0.0 ) )
	    {
	      m_eResult = APM2D::icctNull;
	      return TRUE;
	    }
	}
    }

  return FALSE;
}

VOID 
ICC_EEE_MPFUNFilter::CenterAndRadius(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				     InstructionPointer /*eIPMax*/)
{
  //////////////////////////////////////////////////
  // Compute the center and radius

  mp_real prC[2];  
  prC[0] = ( (m_rE1mag*m_pprE[2][0] - m_rE2mag*m_pprE[1][0])*m_rE00xE01
	     + (m_rE0mag*m_pprE[1][0] - m_rE1mag*m_pprE[0][0])*m_rE20xE21 ) / m_rDen;
  prC[1] = ( (m_rE1mag*m_pprE[2][1] - m_rE2mag*m_pprE[1][1])*m_rE00xE01
	     + (m_rE0mag*m_pprE[1][1] - m_rE1mag*m_pprE[0][1])*m_rE20xE21 ) / m_rDen;

  mp_real rR;  
  rR = ( m_rE0xE1*m_rE20xE21 + m_rE1xE2*m_rE00xE01 ) / m_rDen;
  
  //////////////////////////////////////////////////
  // Check if the projection of the center on to the sites is in CCW orrer

  if ( m_iCCW == -1 )
    {
      mp_real rE0dotC, rE1dotC, rE2dotC;
      Dot(rE0dotC, m_pprE[0][0], m_pprE[0][1], prC[0], prC[1]);
      Dot(rE1dotC, m_pprE[1][0], m_pprE[1][1], prC[0], prC[1]);
      Dot(rE2dotC, m_pprE[2][0], m_pprE[2][1], prC[0], prC[1]);

      mp_real pprProjection[3][2];
      pprProjection[0][0] = (-m_rE00xE01*(-m_pprE[0][1]) + rE0dotC*m_pprE[0][0]) / m_rE0dotE0;
      pprProjection[0][1] = (-m_rE00xE01*(m_pprE[0][0]) + rE0dotC*m_pprE[0][1]) / m_rE0dotE0;

      pprProjection[1][0] = (rE1dotC*m_pprE[1][0]) / m_rE1dotE1;
      pprProjection[1][1] = (rE1dotC*m_pprE[1][1]) / m_rE1dotE1;

      pprProjection[2][0] = (-m_rE20xE21*(-m_pprE[2][1]) + rE2dotC*m_pprE[2][0]) / m_rE2dotE2;
      pprProjection[2][1] = (-m_rE20xE21*(m_pprE[2][0]) + rE2dotC*m_pprE[2][1]) / m_rE2dotE2;

      mp_real rCCWCheck;
      rCCWCheck = ( ( (pprProjection[2][0] - pprProjection[1][0]) *
		      (pprProjection[0][1] - pprProjection[1][1]) ) -
		    ( (pprProjection[2][1] - pprProjection[1][1]) *
		      (pprProjection[0][0] - pprProjection[1][0]) ) );

      if ( rCCWCheck < 0.0 )
	{
	  m_eResult = APM2D::icctNull;
	  //return TRUE;
	  return;
	}
    }

  // Translate the answer back out away from the origin
  prC[0] += m_rE10_x;
  prC[1] += m_rE10_y;

  *pdC_x = dble(prC[0]);
  *pdC_y = dble(prC[1]);
  *pdR = dble(rR);
  m_eResult = APM2D::icctEEE;
}

#endif // COMPILER_PROBLEM

//static 
APM2D::InscribedCircleCenterType 
InscribedCircleCenterEEEMPFUNFilter(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
				    DOUBLE dE00_x, DOUBLE dE00_y, 
				    DOUBLE dE01_x, DOUBLE dE01_y,
				    DOUBLE dE10_x, DOUBLE dE10_y, 
				    DOUBLE dE11_x, DOUBLE dE11_y,
				    DOUBLE dE20_x, DOUBLE dE20_y, 
				    DOUBLE dE21_x, DOUBLE dE21_y)
{
  ICC_EEE_MPFUNFilter iccEEE;

  return iccEEE.InscribedCircleCenterEEE(pdC_x, pdC_y, pdR, 
					 dE00_x, dE00_y, 
					 dE01_x, dE01_y,
					 dE10_x, dE10_y, 
					 dE11_x, dE11_y,
					 dE20_x, dE20_y, 
					 dE21_x, dE21_y);
}

//static 
APM2D::EdgeZoneType EdgeZoneEEEMPFUNFilter(DOUBLE dE00_x, DOUBLE dE00_y, 
					   DOUBLE dE01_x, DOUBLE dE01_y,
					   DOUBLE dE10_x, DOUBLE dE10_y, 
					   DOUBLE dE11_x, DOUBLE dE11_y,
					   DOUBLE dE20_x, DOUBLE dE20_y, 
					   DOUBLE dE21_x, DOUBLE dE21_y,
					   DOUBLE dE30_x, DOUBLE dE30_y,
					   DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::m_uiEdgeZoneEEEMPFUNFilter++;

  INT_32 iDen = -1;
  INT_32 piEdgeZone[2];
  piEdgeZone[0] = -1;
  piEdgeZone[1] = -1;

  switch ( 0 )
  {
  default:
    Interval valE00_x = dE00_x;
    Interval valE00_y = dE00_y; 
    Interval valE01_x = dE01_x;
    Interval valE01_y = dE01_y;
    Interval valE10_x = dE10_x;
    Interval valE10_y = dE10_y; 
    Interval valE11_x = dE11_x;
    Interval valE11_y = dE11_y;
    Interval valE20_x = dE20_x;
    Interval valE20_y = dE20_y; 
    Interval valE21_x = dE21_x;
    Interval valE21_y = dE21_y;
    Interval valE30_x = dE30_x;
    Interval valE30_y = dE30_y;
    Interval valE31_x = dE31_x;
    Interval valE31_y = dE31_y;

    Interval pppvalV[4][2][2]; // Endpoints
    // Translate V1 to the origin
    pppvalV[0][0][0] = valE00_x - valE10_x;
    pppvalV[0][0][1] = valE00_y - valE10_y;
    pppvalV[0][1][0] = valE01_x - valE10_x;
    pppvalV[0][1][1] = valE01_y - valE10_y;
    pppvalV[1][0][0] = 0.0;
    pppvalV[1][0][1] = 0.0;
    pppvalV[1][1][0] = valE11_x - valE10_x;
    pppvalV[1][1][1] = valE11_y - valE10_y;
    pppvalV[2][0][0] = valE20_x - valE10_x;
    pppvalV[2][0][1] = valE20_y - valE10_y;
    pppvalV[2][1][0] = valE21_x - valE10_x;
    pppvalV[2][1][1] = valE21_y - valE10_y;
    pppvalV[3][0][0] = valE30_x - valE10_x;
    pppvalV[3][0][1] = valE30_y - valE10_y;
    pppvalV[3][1][0] = valE31_x - valE10_x;
    pppvalV[3][1][1] = valE31_y - valE10_y;
  
    Interval ppvalE[4][2];     // Edge vectors
    ppvalE[0][0] = valE01_x - valE00_x;
    ppvalE[0][1] = valE01_y - valE00_y;
    ppvalE[1][0] = valE11_x - valE10_x;
    ppvalE[1][1] = valE11_y - valE10_y;
    ppvalE[2][0] = valE21_x - valE20_x;
    ppvalE[2][1] = valE21_y - valE20_y;
    ppvalE[3][0] = valE31_x - valE30_x;
    ppvalE[3][1] = valE31_y - valE30_y;
  
    Interval valE0dotE0;
    Dot(valE0dotE0, ppvalE[0][0], ppvalE[0][1], ppvalE[0][0], ppvalE[0][1]);
    Interval valE1dotE1;
    Dot(valE1dotE1, ppvalE[1][0], ppvalE[1][1], ppvalE[1][0], ppvalE[1][1]);
    Interval valE2dotE2;
    Dot(valE2dotE2, ppvalE[2][0], ppvalE[2][1], ppvalE[2][0], ppvalE[2][1]);
    Interval valE3dotE3;
    Dot(valE3dotE3, ppvalE[3][0], ppvalE[3][1], ppvalE[3][0], ppvalE[3][1]);

    Interval valE0mag;
    valE0mag = sqrt(valE0dotE0);
    Interval valE1mag;
    valE1mag = sqrt(valE1dotE1);
    Interval valE2mag;
    valE2mag = sqrt(valE2dotE2);

    Interval valE0xE1;
    Cross(valE0xE1, ppvalE[0][0], ppvalE[0][1], ppvalE[1][0], ppvalE[1][1]);
    Interval valE1xE2;
    Cross(valE1xE2, ppvalE[1][0], ppvalE[1][1], ppvalE[2][0], ppvalE[2][1]);
    Interval valE2xE0;
    Cross(valE2xE0, ppvalE[2][0], ppvalE[2][1], ppvalE[0][0], ppvalE[0][1]);

    Interval valDen;
    valDen = valE0mag*valE1xE2 + valE1mag*valE2xE0 + valE2mag*valE0xE1;
    iDen = ( valDen < 0.0 );
    if ( iDen == -1 )
      {
	break;
      }
    ASSERT( valDen != 0.0 );

    Interval valE0dotE3;
    Dot(valE0dotE3, ppvalE[0][0], ppvalE[0][1], ppvalE[3][0], ppvalE[3][1]);
    Interval valE1dotE3;
    Dot(valE1dotE3, ppvalE[1][0], ppvalE[1][1], ppvalE[3][0], ppvalE[3][1]);
    Interval valE2dotE3;
    Dot(valE2dotE3, ppvalE[2][0], ppvalE[2][1], ppvalE[3][0], ppvalE[3][1]);

    Interval valE00xE01;
    Cross(valE00xE01, pppvalV[0][0][0], pppvalV[0][0][1], pppvalV[0][1][0], pppvalV[0][1][1]);
    Interval valE20xE21;
    Cross(valE20xE21, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][1][0], pppvalV[2][1][1]);

    Interval valCoef1 = ( - (valE1mag*valE2dotE3 - valE2mag*valE1dotE3)*valE00xE01
		       - (valE0mag*valE1dotE3 - valE1mag*valE0dotE3)*valE20xE21 );
  
    Interval pvalEdgeZone[2];

    Interval valE3dotE30;
    Dot(valE3dotE30, ppvalE[3][0], ppvalE[3][1], pppvalV[3][0][0], pppvalV[3][0][1]);
    pvalEdgeZone[0] = ( valDen*valE3dotE30 
		      + valCoef1 );
    if ( iDen == 1 )
      {
	pvalEdgeZone[0] = -pvalEdgeZone[0];
      }
    piEdgeZone[0] = ( pvalEdgeZone[0] >= 0.0 );
    if ( piEdgeZone[0] == 1 )  
      {
	return APM2D::eztStart;
      }

    Interval valE3dotE31;
    Dot(valE3dotE31, ppvalE[3][0], ppvalE[3][1], pppvalV[3][1][0], pppvalV[3][1][1]);
    pvalEdgeZone[1] = -( valDen*valE3dotE31 
		       + valCoef1 );
    if ( iDen == 1 )
      {
	pvalEdgeZone[1] = -pvalEdgeZone[1];
      }
    piEdgeZone[1] = ( pvalEdgeZone[1] >= 0.0 );
    if ( piEdgeZone[1] == 1 )
      {
	return APM2D::eztEnd;
      }
  }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( piEdgeZone[0] == -1 || piEdgeZone[1] == -1 )
    {
      APM2D::m_uiEdgeZoneEEEMPFUN++;

      mp_real rE00_x = dE00_x;
      mp_real rE00_y = dE00_y; 
      mp_real rE01_x = dE01_x;
      mp_real rE01_y = dE01_y;
      mp_real rE10_x = dE10_x;
      mp_real rE10_y = dE10_y; 
      mp_real rE11_x = dE11_x;
      mp_real rE11_y = dE11_y;
      mp_real rE20_x = dE20_x;
      mp_real rE20_y = dE20_y; 
      mp_real rE21_x = dE21_x;
      mp_real rE21_y = dE21_y;
      mp_real rE30_x = dE30_x;
      mp_real rE30_y = dE30_y;
      mp_real rE31_x = dE31_x;
      mp_real rE31_y = dE31_y;

      mp_real ppprV[4][2][2]; // Endpoints
      // Translate V1 to the origin
      ppprV[0][0][0] = rE00_x - rE10_x;
      ppprV[0][0][1] = rE00_y - rE10_y;
      ppprV[0][1][0] = rE01_x - rE10_x;
      ppprV[0][1][1] = rE01_y - rE10_y;
      ppprV[1][0][0] = 0.0;
      ppprV[1][0][1] = 0.0;
      ppprV[1][1][0] = rE11_x - rE10_x;
      ppprV[1][1][1] = rE11_y - rE10_y;
      ppprV[2][0][0] = rE20_x - rE10_x;
      ppprV[2][0][1] = rE20_y - rE10_y;
      ppprV[2][1][0] = rE21_x - rE10_x;
      ppprV[2][1][1] = rE21_y - rE10_y;
      ppprV[3][0][0] = rE30_x - rE10_x;
      ppprV[3][0][1] = rE30_y - rE10_y;
      ppprV[3][1][0] = rE31_x - rE10_x;
      ppprV[3][1][1] = rE31_y - rE10_y;
  
      mp_real pprE[4][2];     // Edge vectors
      pprE[0][0] = rE01_x - rE00_x;
      pprE[0][1] = rE01_y - rE00_y;
      pprE[1][0] = rE11_x - rE10_x;
      pprE[1][1] = rE11_y - rE10_y;
      pprE[2][0] = rE21_x - rE20_x;
      pprE[2][1] = rE21_y - rE20_y;
      pprE[3][0] = rE31_x - rE30_x;
      pprE[3][1] = rE31_y - rE30_y;
  
      mp_real rE0dotE0;
      Dot(rE0dotE0, pprE[0][0], pprE[0][1], pprE[0][0], pprE[0][1]);
      mp_real rE1dotE1;
      Dot(rE1dotE1, pprE[1][0], pprE[1][1], pprE[1][0], pprE[1][1]);
      mp_real rE2dotE2;
      Dot(rE2dotE2, pprE[2][0], pprE[2][1], pprE[2][0], pprE[2][1]);
      mp_real rE3dotE3;
      Dot(rE3dotE3, pprE[3][0], pprE[3][1], pprE[3][0], pprE[3][1]);

      mp_real rE0mag;
      rE0mag = sqrt(rE0dotE0);
      mp_real rE1mag;
      rE1mag = sqrt(rE1dotE1);
      mp_real rE2mag;
      rE2mag = sqrt(rE2dotE2);

      mp_real rE0xE1;
      Cross(rE0xE1, pprE[0][0], pprE[0][1], pprE[1][0], pprE[1][1]);
      mp_real rE1xE2;
      Cross(rE1xE2, pprE[1][0], pprE[1][1], pprE[2][0], pprE[2][1]);
      mp_real rE2xE0;
      Cross(rE2xE0, pprE[2][0], pprE[2][1], pprE[0][0], pprE[0][1]);

      mp_real rDen;
      rDen = rE0mag*rE1xE2 + rE1mag*rE2xE0 + rE2mag*rE0xE1;
      if ( iDen == -1 )
	{
	  ASSERT( rDen != 0.0 );
	  iDen = ( rDen < 0.0 );
	}

      mp_real rE0dotE3;
      Dot(rE0dotE3, pprE[0][0], pprE[0][1], pprE[3][0], pprE[3][1]);
      mp_real rE1dotE3;
      Dot(rE1dotE3, pprE[1][0], pprE[1][1], pprE[3][0], pprE[3][1]);
      mp_real rE2dotE3;
      Dot(rE2dotE3, pprE[2][0], pprE[2][1], pprE[3][0], pprE[3][1]);

      mp_real rE00xE01;
      Cross(rE00xE01, ppprV[0][0][0], ppprV[0][0][1], ppprV[0][1][0], ppprV[0][1][1]);
      mp_real rE20xE21;
      Cross(rE20xE21, ppprV[2][0][0], ppprV[2][0][1], ppprV[2][1][0], ppprV[2][1][1]);

      mp_real rCoef1 = ( - (rE1mag*rE2dotE3 - rE2mag*rE1dotE3)*rE00xE01
			 - (rE0mag*rE1dotE3 - rE1mag*rE0dotE3)*rE20xE21 );
  
      mp_real prEdgeZone[2];

      if ( piEdgeZone[0] == -1 )
	{
	  mp_real rE3dotE30;
	  Dot(rE3dotE30, pprE[3][0], pprE[3][1], ppprV[3][0][0], ppprV[3][0][1]);
	  prEdgeZone[0] = ( rDen*rE3dotE30 
			    + rCoef1 );
	  if ( iDen )
	    {
	      prEdgeZone[0] = -prEdgeZone[0];
	    }
	  if ( prEdgeZone[0] >= 0.0 )
	    {
	      return APM2D::eztStart;
	    }
	}

      if ( piEdgeZone[1] == -1 )
	{
	  mp_real rE3dotE31;
	  Dot(rE3dotE31, pprE[3][0], pprE[3][1], ppprV[3][1][0], ppprV[3][1][1]);
	  prEdgeZone[1] = -( rDen*rE3dotE31 
			     + rCoef1 );
	  if ( iDen )
	    {
	      prEdgeZone[1] = -prEdgeZone[1];
	    }
	  if ( prEdgeZone[1] >= 0.0 )
	    {
	      return APM2D::eztEnd;
	    }
	}
    }

  return APM2D::eztMiddle;
}

//static 
APM2D::CircleTestType InCircleEEE_VMPFUNFilter(DOUBLE dE00_x, DOUBLE dE00_y, 
					 DOUBLE dE01_x, DOUBLE dE01_y,
					 DOUBLE dE10_x, DOUBLE dE10_y, 
					 DOUBLE dE11_x, DOUBLE dE11_y,
					 DOUBLE dE20_x, DOUBLE dE20_y, 
					 DOUBLE dE21_x, DOUBLE dE21_y,
					 DOUBLE dV3_x, DOUBLE dV3_y)
{
  APM2D::m_uiInCircleEEE_VMPFUNFilter++;

  INT_32 iDen = -1;
  INT_32 iResult = -1;

  switch ( 0 )
    {
    default:
      Interval valE00_x = dE00_x;
      Interval valE00_y = dE00_y; 
      Interval valE01_x = dE01_x;
      Interval valE01_y = dE01_y;
      Interval valE10_x = dE10_x;
      Interval valE10_y = dE10_y; 
      Interval valE11_x = dE11_x;
      Interval valE11_y = dE11_y;
      Interval valE20_x = dE20_x;
      Interval valE20_y = dE20_y; 
      Interval valE21_x = dE21_x;
      Interval valE21_y = dE21_y;
      Interval valV3_x = dV3_x;
      Interval valV3_y = dV3_y;

      Interval pppvalV[4][2][2]; // Endpoints
      // Translate V1 to the origin
      //memset(pppvalV, 0, sizeof(pppvalV));
      pppvalV[0][0][0] = valE00_x - valE10_x;
      pppvalV[0][0][1] = valE00_y - valE10_y;
      pppvalV[0][1][0] = valE01_x - valE10_x;
      pppvalV[0][1][1] = valE01_y - valE10_y;
      pppvalV[1][0][0] = 0.0;
      pppvalV[1][0][1] = 0.0;
      pppvalV[1][1][0] = valE11_x - valE10_x;
      pppvalV[1][1][1] = valE11_y - valE10_y;
      pppvalV[2][0][0] = valE20_x - valE10_x;
      pppvalV[2][0][1] = valE20_y - valE10_y;
      pppvalV[2][1][0] = valE21_x - valE10_x;
      pppvalV[2][1][1] = valE21_y - valE10_y;
      pppvalV[3][0][0] = valV3_x - valE10_x;
      pppvalV[3][0][1] = valV3_y - valE10_y;
  
      Interval ppvalE[4][2];     // Edge vectors
      //memset(ppvalE, 0, sizeof(ppvalE));
      ppvalE[0][0] = valE01_x - valE00_x;
      ppvalE[0][1] = valE01_y - valE00_y;
      ppvalE[1][0] = valE11_x - valE10_x;
      ppvalE[1][1] = valE11_y - valE10_y;
      ppvalE[2][0] = valE21_x - valE20_x;
      ppvalE[2][1] = valE21_y - valE20_y;
  
      Interval valE0dotE0;
      Dot(valE0dotE0, ppvalE[0][0], ppvalE[0][1], ppvalE[0][0], ppvalE[0][1]);
      Interval valE1dotE1;
      Dot(valE1dotE1, ppvalE[1][0], ppvalE[1][1], ppvalE[1][0], ppvalE[1][1]);
      Interval valE2dotE2;
      Dot(valE2dotE2, ppvalE[2][0], ppvalE[2][1], ppvalE[2][0], ppvalE[2][1]);
      Interval valV3dotV3;
      Dot(valV3dotV3, pppvalV[3][0][0], pppvalV[3][0][1], pppvalV[3][0][0], pppvalV[3][0][1]);

      Interval valE0mag;
      valE0mag = sqrt(valE0dotE0);
      Interval valE1mag;
      valE1mag = sqrt(valE1dotE1);
      Interval valE2mag;
      valE2mag = sqrt(valE2dotE2);

      Interval valE0xE1;
      Cross(valE0xE1, ppvalE[0][0], ppvalE[0][1], ppvalE[1][0], ppvalE[1][1]);
      Interval valE1xE2;
      Cross(valE1xE2, ppvalE[1][0], ppvalE[1][1], ppvalE[2][0], ppvalE[2][1]);
      Interval valE2xE0;
      Cross(valE2xE0, ppvalE[2][0], ppvalE[2][1], ppvalE[0][0], ppvalE[0][1]);

      Interval valDen;
      valDen = valE0mag*valE1xE2 + valE1mag*valE2xE0 + valE2mag*valE0xE1;
      iDen = ( valDen != 0.0 );
      if ( iDen == -1 )
	{
	  break;
	}
      ASSERT( iDen == 1 );

      Interval valE0dotV3;
      Dot(valE0dotV3, ppvalE[0][0], ppvalE[0][1], pppvalV[3][0][0], pppvalV[3][0][1]);
      Interval valE1dotV3;
      Dot(valE1dotV3, ppvalE[1][0], ppvalE[1][1], pppvalV[3][0][0], pppvalV[3][0][1]);
      Interval valE2dotV3;
      Dot(valE2dotV3, ppvalE[2][0], ppvalE[2][1], pppvalV[3][0][0], pppvalV[3][0][1]);

      Interval valE0dotE1;
      Dot(valE0dotE1, ppvalE[0][0], ppvalE[0][1], ppvalE[1][0], ppvalE[1][1]);
      Interval valE1dotE2;
      Dot(valE1dotE2, ppvalE[1][0], ppvalE[1][1], ppvalE[2][0], ppvalE[2][1]);
      Interval valE0dotE2;
      Dot(valE0dotE2, ppvalE[0][0], ppvalE[0][1], ppvalE[2][0], ppvalE[2][1]);

      Interval valE00xE01;
      Cross(valE00xE01, pppvalV[0][0][0], pppvalV[0][0][1], pppvalV[0][1][0], pppvalV[0][1][1]);
      Interval valE20xE21;
      Cross(valE20xE21, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][1][0], pppvalV[2][1][1]);

      Interval valE0E1minusE0dotE1;
      valE0E1minusE0dotE1 = sqrt(valE0dotE0*valE1dotE1) - valE0dotE1;
      Interval valE1E2minusE1dotE2;
      valE1E2minusE1dotE2 = sqrt(valE1dotE1*valE2dotE2) - valE1dotE2;
      Interval valE0E2plusE0dotE2;
      valE0E2plusE0dotE2 = sqrt(valE0dotE0*valE2dotE2) + valE0dotE2;

      Interval valResult;
      valResult = ( valDen*( valDen*valV3dotV3 
			 + 2.0*( valE0dotV3*valE1mag*valE20xE21 + valE1dotV3*( valE2mag*valE00xE01 - valE0mag*valE20xE21 ) - valE2dotV3*valE1mag*valE00xE01 ) )
		  + valE00xE01*valE00xE01*valE1E2minusE1dotE2*valE1E2minusE1dotE2
		  + 2.0*valE00xE01*valE20xE21*( valE1mag*( valE0mag*valE1dotE2 + valE2mag*valE0dotE1 ) - valE1dotE1*valE0E2plusE0dotE2 - valE0xE1*valE1xE2 )
		  + valE20xE21*valE20xE21*valE0E1minusE0dotE1*valE0E1minusE0dotE1 );

      iResult = ( valResult < 0.0 );
      if ( iResult == -1 )
	{
	  break;
	}
      if ( iResult == 1 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( valResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
      return APM2D::cttOnCircle;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( iResult == -1 )
    {
      APM2D::m_uiInCircleEEE_VMPFUN++;

      mp_real rE00_x = dE00_x;
      mp_real rE00_y = dE00_y; 
      mp_real rE01_x = dE01_x;
      mp_real rE01_y = dE01_y;
      mp_real rE10_x = dE10_x;
      mp_real rE10_y = dE10_y; 
      mp_real rE11_x = dE11_x;
      mp_real rE11_y = dE11_y;
      mp_real rE20_x = dE20_x;
      mp_real rE20_y = dE20_y; 
      mp_real rE21_x = dE21_x;
      mp_real rE21_y = dE21_y;
      mp_real rV3_x = dV3_x;
      mp_real rV3_y = dV3_y;

      mp_real ppprV[4][2][2]; // Endpoints
      // Translate V1 to the origin
      //memset(ppprV, 0, sizeof(ppprV));
      ppprV[0][0][0] = rE00_x - rE10_x;
      ppprV[0][0][1] = rE00_y - rE10_y;
      ppprV[0][1][0] = rE01_x - rE10_x;
      ppprV[0][1][1] = rE01_y - rE10_y;
      ppprV[1][0][0] = 0.0;
      ppprV[1][0][1] = 0.0;
      ppprV[1][1][0] = rE11_x - rE10_x;
      ppprV[1][1][1] = rE11_y - rE10_y;
      ppprV[2][0][0] = rE20_x - rE10_x;
      ppprV[2][0][1] = rE20_y - rE10_y;
      ppprV[2][1][0] = rE21_x - rE10_x;
      ppprV[2][1][1] = rE21_y - rE10_y;
      ppprV[3][0][0] = rV3_x - rE10_x;
      ppprV[3][0][1] = rV3_y - rE10_y;
  
      mp_real pprE[4][2];     // Edge vectors
      //memset(pprE, 0, sizeof(pprE));
      pprE[0][0] = rE01_x - rE00_x;
      pprE[0][1] = rE01_y - rE00_y;
      pprE[1][0] = rE11_x - rE10_x;
      pprE[1][1] = rE11_y - rE10_y;
      pprE[2][0] = rE21_x - rE20_x;
      pprE[2][1] = rE21_y - rE20_y;
  
      mp_real rE0dotE0;
      Dot(rE0dotE0, pprE[0][0], pprE[0][1], pprE[0][0], pprE[0][1]);
      mp_real rE1dotE1;
      Dot(rE1dotE1, pprE[1][0], pprE[1][1], pprE[1][0], pprE[1][1]);
      mp_real rE2dotE2;
      Dot(rE2dotE2, pprE[2][0], pprE[2][1], pprE[2][0], pprE[2][1]);
      mp_real rV3dotV3;
      Dot(rV3dotV3, ppprV[3][0][0], ppprV[3][0][1], ppprV[3][0][0], ppprV[3][0][1]);

      mp_real rE0mag;
      rE0mag = sqrt(rE0dotE0);
      mp_real rE1mag;
      rE1mag = sqrt(rE1dotE1);
      mp_real rE2mag;
      rE2mag = sqrt(rE2dotE2);

      mp_real rE0xE1;
      Cross(rE0xE1, pprE[0][0], pprE[0][1], pprE[1][0], pprE[1][1]);
      mp_real rE1xE2;
      Cross(rE1xE2, pprE[1][0], pprE[1][1], pprE[2][0], pprE[2][1]);
      mp_real rE2xE0;
      Cross(rE2xE0, pprE[2][0], pprE[2][1], pprE[0][0], pprE[0][1]);

      mp_real rDen;
      rDen = rE0mag*rE1xE2 + rE1mag*rE2xE0 + rE2mag*rE0xE1;
      ASSERT( !( rDen == 0.0 ) );

      mp_real rE0dotV3;
      Dot(rE0dotV3, pprE[0][0], pprE[0][1], ppprV[3][0][0], ppprV[3][0][1]);
      mp_real rE1dotV3;
      Dot(rE1dotV3, pprE[1][0], pprE[1][1], ppprV[3][0][0], ppprV[3][0][1]);
      mp_real rE2dotV3;
      Dot(rE2dotV3, pprE[2][0], pprE[2][1], ppprV[3][0][0], ppprV[3][0][1]);

      mp_real rE0dotE1;
      Dot(rE0dotE1, pprE[0][0], pprE[0][1], pprE[1][0], pprE[1][1]);
      mp_real rE1dotE2;
      Dot(rE1dotE2, pprE[1][0], pprE[1][1], pprE[2][0], pprE[2][1]);
      mp_real rE0dotE2;
      Dot(rE0dotE2, pprE[0][0], pprE[0][1], pprE[2][0], pprE[2][1]);

      mp_real rE00xE01;
      Cross(rE00xE01, ppprV[0][0][0], ppprV[0][0][1], ppprV[0][1][0], ppprV[0][1][1]);
      mp_real rE20xE21;
      Cross(rE20xE21, ppprV[2][0][0], ppprV[2][0][1], ppprV[2][1][0], ppprV[2][1][1]);

      mp_real rE0E1minusE0dotE1;
      rE0E1minusE0dotE1 = sqrt(rE0dotE0*rE1dotE1) - rE0dotE1;
      mp_real rE1E2minusE1dotE2;
      rE1E2minusE1dotE2 = sqrt(rE1dotE1*rE2dotE2) - rE1dotE2;
      mp_real rE0E2plusE0dotE2;
      rE0E2plusE0dotE2 = sqrt(rE0dotE0*rE2dotE2) + rE0dotE2;

      mp_real rResult;
      rResult = ( rDen*( rDen*rV3dotV3 
			 + 2.0*( rE0dotV3*rE1mag*rE20xE21 + rE1dotV3*( rE2mag*rE00xE01 - rE0mag*rE20xE21 ) - rE2dotV3*rE1mag*rE00xE01 ) )
		  + rE00xE01*rE00xE01*rE1E2minusE1dotE2*rE1E2minusE1dotE2
		  + 2.0*rE00xE01*rE20xE21*( rE1mag*( rE0mag*rE1dotE2 + rE2mag*rE0dotE1 ) - rE1dotE1*rE0E2plusE0dotE2 - rE0xE1*rE1xE2 )
		  + rE20xE21*rE20xE21*rE0E1minusE0dotE1*rE0E1minusE0dotE1 );
  
      if ( rResult < 0.0 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( rResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
    }

  return APM2D::cttOnCircle;
}

//static 
APM2D::CircleTestType InCircleEEE_EMPFUNFilter(DOUBLE dE00_x, DOUBLE dE00_y, 
					       DOUBLE dE01_x, DOUBLE dE01_y,
					       DOUBLE dE10_x, DOUBLE dE10_y, 
					       DOUBLE dE11_x, DOUBLE dE11_y,
					       DOUBLE dE20_x, DOUBLE dE20_y, 
					       DOUBLE dE21_x, DOUBLE dE21_y,
					       DOUBLE dE30_x, DOUBLE dE30_y,
					       DOUBLE dE31_x, DOUBLE dE31_y)
{
  APM2D::m_uiInCircleEEE_EMPFUNFilter++;

  INT_32 iDen = -1;
  INT_32 iResult = -1;

  switch ( 0 )
    {
    default:
      Interval valE00_x = dE00_x;
      Interval valE00_y = dE00_y; 
      Interval valE01_x = dE01_x;
      Interval valE01_y = dE01_y;
      Interval valE10_x = dE10_x;
      Interval valE10_y = dE10_y; 
      Interval valE11_x = dE11_x;
      Interval valE11_y = dE11_y;
      Interval valE20_x = dE20_x;
      Interval valE20_y = dE20_y; 
      Interval valE21_x = dE21_x;
      Interval valE21_y = dE21_y;
      Interval valE30_x = dE30_x;
      Interval valE30_y = dE30_y;
      Interval valE31_x = dE31_x;
      Interval valE31_y = dE31_y;
  
      Interval pppvalV[4][2][2]; // Endpoints
      // Translate V1 to the origin
      pppvalV[0][0][0] = valE00_x - valE10_x;
      pppvalV[0][0][1] = valE00_y - valE10_y;
      pppvalV[0][1][0] = valE01_x - valE10_x;
      pppvalV[0][1][1] = valE01_y - valE10_y;
      pppvalV[1][0][0] = 0.0;
      pppvalV[1][0][1] = 0.0;
      pppvalV[1][1][0] = valE11_x - valE10_x;
      pppvalV[1][1][1] = valE11_y - valE10_y;
      pppvalV[2][0][0] = valE20_x - valE10_x;
      pppvalV[2][0][1] = valE20_y - valE10_y;
      pppvalV[2][1][0] = valE21_x - valE10_x;
      pppvalV[2][1][1] = valE21_y - valE10_y;
      pppvalV[3][0][0] = valE30_x - valE10_x;
      pppvalV[3][0][1] = valE30_y - valE10_y;
      pppvalV[3][1][0] = valE31_x - valE10_x;
      pppvalV[3][1][1] = valE31_y - valE10_y;
  
      Interval ppvalE[4][2];     // Edge vectors
      ppvalE[0][0] = valE01_x - valE00_x;
      ppvalE[0][1] = valE01_y - valE00_y;
      ppvalE[1][0] = valE11_x - valE10_x;
      ppvalE[1][1] = valE11_y - valE10_y;
      ppvalE[2][0] = valE21_x - valE20_x;
      ppvalE[2][1] = valE21_y - valE20_y;
      ppvalE[3][0] = valE31_x - valE30_x;
      ppvalE[3][1] = valE31_y - valE30_y;
  
      Interval valE0dotE0;
      Dot(valE0dotE0, ppvalE[0][0], ppvalE[0][1], ppvalE[0][0], ppvalE[0][1]);
      Interval valE1dotE1;
      Dot(valE1dotE1, ppvalE[1][0], ppvalE[1][1], ppvalE[1][0], ppvalE[1][1]);
      Interval valE2dotE2;
      Dot(valE2dotE2, ppvalE[2][0], ppvalE[2][1], ppvalE[2][0], ppvalE[2][1]);
      Interval valE3dotE3;
      Dot(valE3dotE3, ppvalE[3][0], ppvalE[3][1], ppvalE[3][0], ppvalE[3][1]);
  
      Interval valE0mag;
      valE0mag = sqrt(valE0dotE0);
      Interval valE1mag;
      valE1mag = sqrt(valE1dotE1);
      Interval valE2mag;
      valE2mag = sqrt(valE2dotE2);
  
      Interval valE0xE1;
      Cross(valE0xE1, ppvalE[0][0], ppvalE[0][1], ppvalE[1][0], ppvalE[1][1]);
      Interval valE1xE2;
      Cross(valE1xE2, ppvalE[1][0], ppvalE[1][1], ppvalE[2][0], ppvalE[2][1]);
      Interval valE2xE0;
      Cross(valE2xE0, ppvalE[2][0], ppvalE[2][1], ppvalE[0][0], ppvalE[0][1]);
    
      Interval valDen;
      valDen = valE0mag*valE1xE2 + valE1mag*valE2xE0 + valE2mag*valE0xE1;
      iDen = ( valDen != 0.0 );
      if ( iDen == -1 )
	{
	  break;
	}
      ASSERT( iDen == 1 );
  
      Interval valE0xE3;
      Cross(valE0xE3, ppvalE[0][0], ppvalE[0][1], ppvalE[3][0], ppvalE[3][1]);
      Interval valE1xE3;
      Cross(valE1xE3, ppvalE[1][0], ppvalE[1][1], ppvalE[3][0], ppvalE[3][1]);
      Interval valE2xE3;
      Cross(valE2xE3, ppvalE[2][0], ppvalE[2][1], ppvalE[3][0], ppvalE[3][1]);
    
      Interval valE00xE01;
      Cross(valE00xE01, pppvalV[0][0][0], pppvalV[0][0][1], pppvalV[0][1][0], pppvalV[0][1][1]);
      Interval valE20xE21;
      Cross(valE20xE21, pppvalV[2][0][0], pppvalV[2][0][1], pppvalV[2][1][0], pppvalV[2][1][1]);
      Interval valE30xE31;
      Cross(valE30xE31, pppvalV[3][0][0], pppvalV[3][0][1], pppvalV[3][1][0], pppvalV[3][1][1]);
  
      Interval valA, valB;
      valA = ( ( sqrt(valE2dotE2)*valE1xE3 - sqrt(valE1dotE1)*valE2xE3 )*valE00xE01
	     + ( sqrt(valE1dotE1)*valE0xE3 - sqrt(valE0dotE0)*valE1xE3 )*valE20xE21 );
      valB = ( valE0xE1*valE20xE21 + valE1xE2*valE00xE01 );
      valB = valB*valB;

      Interval valResult;
      valResult = ( valA*valA
		  + 2.0*valE30xE31*valDen*valA 
		  + valE30xE31*valE30xE31*valDen*valDen
		  - valE3dotE3*valB );

      iResult = ( valResult < 0.0 );
      if ( iResult == -1 )
	{
	  break;
	}
      if ( iResult == 1 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( valResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
      return APM2D::cttOnCircle;
    }

  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  //////////////////////////////////////////////////
  // MPFUN

  if ( iResult == -1 )
    {
      APM2D::m_uiInCircleEEE_EMPFUN++;

      mp_real rE00_x = dE00_x;
      mp_real rE00_y = dE00_y; 
      mp_real rE01_x = dE01_x;
      mp_real rE01_y = dE01_y;
      mp_real rE10_x = dE10_x;
      mp_real rE10_y = dE10_y; 
      mp_real rE11_x = dE11_x;
      mp_real rE11_y = dE11_y;
      mp_real rE20_x = dE20_x;
      mp_real rE20_y = dE20_y; 
      mp_real rE21_x = dE21_x;
      mp_real rE21_y = dE21_y;
      mp_real rE30_x = dE30_x;
      mp_real rE30_y = dE30_y;
      mp_real rE31_x = dE31_x;
      mp_real rE31_y = dE31_y;
  
      mp_real ppprV[4][2][2]; // Endpoints
      // Translate V1 to the origin
      ppprV[0][0][0] = rE00_x - rE10_x;
      ppprV[0][0][1] = rE00_y - rE10_y;
      ppprV[0][1][0] = rE01_x - rE10_x;
      ppprV[0][1][1] = rE01_y - rE10_y;
      ppprV[1][0][0] = 0.0;
      ppprV[1][0][1] = 0.0;
      ppprV[1][1][0] = rE11_x - rE10_x;
      ppprV[1][1][1] = rE11_y - rE10_y;
      ppprV[2][0][0] = rE20_x - rE10_x;
      ppprV[2][0][1] = rE20_y - rE10_y;
      ppprV[2][1][0] = rE21_x - rE10_x;
      ppprV[2][1][1] = rE21_y - rE10_y;
      ppprV[3][0][0] = rE30_x - rE10_x;
      ppprV[3][0][1] = rE30_y - rE10_y;
      ppprV[3][1][0] = rE31_x - rE10_x;
      ppprV[3][1][1] = rE31_y - rE10_y;
  
      mp_real pprE[4][2];     // Edge vectors
      pprE[0][0] = rE01_x - rE00_x;
      pprE[0][1] = rE01_y - rE00_y;
      pprE[1][0] = rE11_x - rE10_x;
      pprE[1][1] = rE11_y - rE10_y;
      pprE[2][0] = rE21_x - rE20_x;
      pprE[2][1] = rE21_y - rE20_y;
      pprE[3][0] = rE31_x - rE30_x;
      pprE[3][1] = rE31_y - rE30_y;
  
      mp_real rE0dotE0;
      Dot(rE0dotE0, pprE[0][0], pprE[0][1], pprE[0][0], pprE[0][1]);
      mp_real rE1dotE1;
      Dot(rE1dotE1, pprE[1][0], pprE[1][1], pprE[1][0], pprE[1][1]);
      mp_real rE2dotE2;
      Dot(rE2dotE2, pprE[2][0], pprE[2][1], pprE[2][0], pprE[2][1]);
      mp_real rE3dotE3;
      Dot(rE3dotE3, pprE[3][0], pprE[3][1], pprE[3][0], pprE[3][1]);
  
      mp_real rE0mag;
      rE0mag = sqrt(rE0dotE0);
      mp_real rE1mag;
      rE1mag = sqrt(rE1dotE1);
      mp_real rE2mag;
      rE2mag = sqrt(rE2dotE2);
  
      mp_real rE0xE1;
      Cross(rE0xE1, pprE[0][0], pprE[0][1], pprE[1][0], pprE[1][1]);
      mp_real rE1xE2;
      Cross(rE1xE2, pprE[1][0], pprE[1][1], pprE[2][0], pprE[2][1]);
      mp_real rE2xE0;
      Cross(rE2xE0, pprE[2][0], pprE[2][1], pprE[0][0], pprE[0][1]);
    
      mp_real rDen;
      rDen = rE0mag*rE1xE2 + rE1mag*rE2xE0 + rE2mag*rE0xE1;
      ASSERT( !( rDen == 0.0 ) );
  
      mp_real rE0xE3;
      Cross(rE0xE3, pprE[0][0], pprE[0][1], pprE[3][0], pprE[3][1]);
      mp_real rE1xE3;
      Cross(rE1xE3, pprE[1][0], pprE[1][1], pprE[3][0], pprE[3][1]);
      mp_real rE2xE3;
      Cross(rE2xE3, pprE[2][0], pprE[2][1], pprE[3][0], pprE[3][1]);
    
      mp_real rE00xE01;
      Cross(rE00xE01, ppprV[0][0][0], ppprV[0][0][1], ppprV[0][1][0], ppprV[0][1][1]);
      mp_real rE20xE21;
      Cross(rE20xE21, ppprV[2][0][0], ppprV[2][0][1], ppprV[2][1][0], ppprV[2][1][1]);
      mp_real rE30xE31;
      Cross(rE30xE31, ppprV[3][0][0], ppprV[3][0][1], ppprV[3][1][0], ppprV[3][1][1]);
  
      mp_real rA, rB;
      rA = ( ( sqrt(rE2dotE2)*rE1xE3 - sqrt(rE1dotE1)*rE2xE3 )*rE00xE01
	     + ( sqrt(rE1dotE1)*rE0xE3 - sqrt(rE0dotE0)*rE1xE3 )*rE20xE21 );
      rB = ( rE0xE1*rE20xE21 + rE1xE2*rE00xE01 );
      rB = rB*rB;

      mp_real rResult;
      rResult = ( rA*rA
		  + 2.0*rE30xE31*rDen*rA 
		  + rE30xE31*rE30xE31*rDen*rDen
		  - rE3dotE3*rB );

      if ( rResult < 0.0 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( rResult > 0.0 )
	{
	  return APM2D::cttOutCircle;
	}
    }

  return APM2D::cttOnCircle;
}
