#include "hashtable.h"
#include "qemesh.h"

#include "qenode.h"
#include "qeedgecollapse.h"

#include "qevertex.h"
#include "qewedge.h"
#include "qeedge.h"
#include "qefacet.h"

#include "predicates.h"

//#define COMPILER_PROBLEM

QEFacet QEMesh::m_faNULL;
QEFacet *QEMesh::m_pfaNULL = NULL;
BOOL QEMesh::m_bRandomize = FALSE;
QEMesh::AccuracyType QEMesh::m_eAccuracy = QEMesh::atFloat;

//const FLOAT QEMesh::m_cfEpsilon = 0.0;
//const FLOAT QEMesh::m_cfEpsilon = 0.000001;
const FLOAT QEMesh::m_cfEpsilon = 0.000005;
//const FLOAT QEMesh::m_cfEpsilon = 0.00001;
FLOAT QEMesh::m_fMergeEpsilon = 0.40;

//////////////////////////////////////////////////
// Temporary Fix to infinite edge problem

//#define ENABLE_VORONOI_POLYGON_DEBUGGING
//#define ENABLE_VORONOI_POLYGON_DEBUGGING_PROJECTION
#define ENABLE_VORONOI_POLYGONS_DEBUGGING

#define ENABLE_VORONOI_POLYGON_CHECK_EDGE_ZONES
//#define ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES

static QEVertex *s_pvxOpenEndpoint = NULL;
static Point s_ptOpenEndpoint;
static BOOL s_bOpenEndpoint;
static const FLOAT s_cfDistanceEpsilon = 0.0001;
//static const FLOAT s_cfDistanceEpsilon = 0.00001;
//static const FLOAT s_cfDistanceEpsilon = 0.0000001;
//static const FLOAT s_cfVoronoiInfinity = 500.0;
//static const FLOAT s_cfVoronoiInfinity = 1000.0;
//static const FLOAT s_cfVoronoiInfinity = 3000.0;
static const FLOAT s_cfVoronoiInfinity = 640.0;
//static const FLOAT s_cfVoronoiInfinity = 1000.0;
//static const FLOAT s_cfVoronoiInfinity = 10000.0;
//static const FLOAT s_cfVoronoiInfinity = 100000.0;
//static const FLOAT s_cfVoronoiInfinity = 1000000.0;

//static const FLOAT s_cfEpsilonEdgeZone = 0.00001;
static const FLOAT s_cfEpsilonEdgeZone = 0.000001;

static const FLOAT s_cfEpsilonVoronoiVEVertexOnEdge = 0.0000000001;
static const FLOAT s_cfEpsilonVoronoiVVE = 0.0001;
static const FLOAT s_cfEpsilonVoronoiVEE = 0.0001;
//static const FLOAT s_cfEpsilonVoronoiVVEAntiparallel = 0.00000001;
static const FLOAT s_cfEpsilonVoronoiVVEAntiparallel = 0.0000001;
static const FLOAT s_cfEpsilonVoronoiVVEVertexOnEdge = 0.0000000001;
static const FLOAT s_cfEpsilonVoronoiVEEVertexOnEdge = 0.0000000001;
static const FLOAT s_cfEpsilonVoronoiVEEParallel = 0.0000000001;
//static const FLOAT s_cfEpsilonVoronoiVEEAntiparallel = 0.00000001;
static const FLOAT s_cfEpsilonVoronoiVEEAntiparallel = 0.0000001;
static const FLOAT s_cfEpsilonVoronoiVVEVertexZone = 0.0000001;
static const FLOAT s_cfEpsilonVoronoiVEEVertexZone = 0.0000001;
static const FLOAT s_cfEpsilonVoronoiZMonotone = 0.0001;

// problems with offsetting KochIsland at 1.25
static const FLOAT s_cfEpsilonVoronoiOffsetDen = 0.0;
//static const FLOAT s_cfEpsilonVoronoiOffsetDen = 1e-7;

//////////////////////////////////////////////////////////////////////
// QEMesh Class
//

QEMesh::QEMesh()
{
  if ( m_pfaNULL == NULL )
    {
      // Jonathan Schewchuck's exact arithmetic stuff
      exactinit();

      m_pfaNULL = &m_faNULL;
      m_faNULL.Init();
    }

  UINT_32 i;

  m_plistQEVertex = NULL;
  m_plistQEWedge = NULL;
  m_plistQEEdge = NULL;
  m_plistQEEdgeBoundary = NULL;
  m_plistQEFacet = NULL;

  m_plistQEEdgeCollapse = NULL;

  m_plistQEEdgeNode = NULL;
  m_plistQEFacetNode = NULL;
  m_pheapSelective = NULL;
  m_pheapSelectiveReverse = NULL;
  m_plistSelective = NULL;

  m_penRoot = NULL;

  for ( i = 0; i < 2; i++ )
    {
      m_pmeLink[i] = NULL;
    }
}

QEMesh::~QEMesh()
{
  Uninit();
}

VOID QEMesh::Init()
{
  UINT_32 i;

  QE::Init();

  m_plistQEVertex = new CObjectList<QEVertex>;
  ASSERT( m_plistQEVertex != NULL );
  m_plistQEVertex->Init();

  m_plistQEWedge = new CObjectList<QEWedge>;
  ASSERT( m_plistQEWedge != NULL );
  m_plistQEWedge->Init();

  m_plistQEEdge = new CObjectList<QEEdge>;
  ASSERT( m_plistQEEdge != NULL );
  m_plistQEEdge->Init();

  m_plistQEEdgeBoundary = new CObjectList<QEEdge>;
  ASSERT( m_plistQEEdgeBoundary != NULL );
  m_plistQEEdgeBoundary->Init();

  m_plistQEFacet = new CObjectList<QEFacet>;
  ASSERT( m_plistQEFacet != NULL );
  m_plistQEFacet->Init();

  m_plistQEEdgeCollapse = new CObjectList<QEEdgeCollapse>;
  ASSERT( m_plistQEEdgeCollapse != NULL );
  m_plistQEEdgeCollapse->Init();

  m_plistQEEdgeNode = new CObjectList<QEEdgeNode>;
  ASSERT( m_plistQEEdgeNode != NULL );
  m_plistQEEdgeNode->Init();

  m_plistQEFacetNode = new CObjectList<QEFacetNode>;
  ASSERT( m_plistQEFacetNode != NULL );
  m_plistQEFacetNode->Init();

  m_pheapSelective = new CHeap<VOID*, FLOAT>;
  ASSERT( m_pheapSelective != NULL );
  m_pheapSelective->Init();

  m_pheapSelectiveReverse = new CHeap<VOID*, FLOAT>;
  ASSERT( m_pheapSelectiveReverse != NULL );
  m_pheapSelectiveReverse->Init();
  m_pheapSelectiveReverse->SetBiggestFirst(FALSE);

  m_plistSelective = new CListInt;
  ASSERT( m_plistSelective != NULL );
  m_plistSelective->Init();

  m_penRoot = NULL;

  for ( i = 0; i < 2; i++ )
    {
      ASSERT( m_pmeLink[i] == NULL );
    }
}

VOID QEMesh::Uninit()
{
  UINT_32 i;
  QEVertex *pvx;
  QEWedge *pwe;
  QEEdge *pe;
  QEFacet *pfa;
  QEEdgeCollapse *pec;
  QEEdgeNode *pen;
  QEFacetNode *pfn;

  //fprintf(stderr, "START: QEMesh::Uninit %p\n", this);

  //DisplayEuler();
  //Display();

  if ( m_plistQEVertex != NULL )
    {
      //fprintf(stderr, "  before: length = %ld\n", m_plistQEVertex->GetLength());
      for ( pvx = m_plistQEVertex->RemoveFirst();
	    pvx != NULL;
	    pvx = m_plistQEVertex->RemoveFirst() )
	{
	  //fprintf(stderr, "  deallocate vertex = %p\n", pvx);
	  DeallocateQEVertex(pvx);
	  //fprintf(stderr, "  loop  : length = %ld\n", m_plistQEVertex->GetLength());
	}
      //fprintf(stderr, "  after : length = %ld\n", m_plistQEVertex->GetLength());

      delete m_plistQEVertex;
      m_plistQEVertex = NULL;
    }

  if ( m_plistQEWedge != NULL )
    {
      for ( pwe = m_plistQEWedge->RemoveFirst();
	    pwe != NULL;
	    pwe = m_plistQEWedge->RemoveFirst() )
	{
	  DeallocateQEWedge(pwe);
	}
      delete m_plistQEWedge;
      m_plistQEWedge = NULL;
    }

  if ( m_plistQEEdge != NULL )
    {
      for ( pe = m_plistQEEdge->RemoveFirst();
	    pe != NULL;
	    pe = m_plistQEEdge->RemoveFirst() )
	{
	  DeallocateQEEdge(pe);
	}
      delete m_plistQEEdge;
      m_plistQEEdge = NULL;
    }

  if ( m_plistQEEdgeBoundary != NULL )
    {
      for ( pe = m_plistQEEdgeBoundary->RemoveFirst();
	    pe != NULL;
	    pe = m_plistQEEdgeBoundary->RemoveFirst() )
	{
	  DeallocateQEEdge(pe);
	}
      delete m_plistQEEdgeBoundary;
      m_plistQEEdgeBoundary = NULL;
    }

  if ( m_plistQEFacet != NULL )
    {
      for ( pfa = m_plistQEFacet->RemoveFirst();
	    pfa != NULL;
	    pfa = m_plistQEFacet->RemoveFirst() )
	{
	  DeallocateQEFacet(pfa);
	}
      delete m_plistQEFacet;
      m_plistQEFacet = NULL;
    }

  if ( m_plistQEEdgeCollapse != NULL )
    {
      for ( pec = m_plistQEEdgeCollapse->RemoveFirst();
	    pec != NULL;
	    pec = m_plistQEEdgeCollapse->RemoveFirst() )
	{
	  DeallocateQEEdgeCollapse(pec);
	}
      delete m_plistQEEdgeCollapse;
      m_plistQEEdgeCollapse = NULL;
    }

  if ( m_plistQEEdgeNode != NULL )
    {
      for ( pen = m_plistQEEdgeNode->RemoveFirst();
	    pen != NULL;
	    pen = m_plistQEEdgeNode->RemoveFirst() )
	{
	  DeallocateQEEdgeNode(pen);
	}
      delete m_plistQEEdgeNode;
      m_plistQEEdgeNode = NULL;
    }

  if ( m_plistQEFacetNode != NULL )
    {
      for ( pfn = m_plistQEFacetNode->RemoveFirst();
	    pfn != NULL;
	    pfn = m_plistQEFacetNode->RemoveFirst() )
	{
	  DeallocateQEFacetNode(pfn);
	}
      delete m_plistQEFacetNode;
      m_plistQEFacetNode = NULL;
    }

  if ( m_pheapSelective != NULL )
    {
      delete m_pheapSelective;
      m_pheapSelective = NULL;
    }

  if ( m_pheapSelectiveReverse != NULL )
    {
      delete m_pheapSelectiveReverse;
      m_pheapSelectiveReverse = NULL;
    }

  if ( m_plistSelective != NULL )
    {
      delete m_plistSelective;
      m_plistSelective = NULL;
    }

  m_penRoot = NULL;

  for ( i = 0; i < 2; i++ )
    {
      m_pmeLink[i] = NULL;
    }

  //fprintf(stderr, "END  : QEMesh::Uninit %p\n", this);
}

BOOL QEMesh::Initialized()
{
  return TRUE;
}

BOOL QEMesh::Valid()
{
  return TRUE;
}

VOID QEMesh::CreateContours(CListInt *plistContours)
{
  ASSERT( plistContours != NULL );

  CListInt listQEEdge;
  CListIterInt iterContours;
  CObjectList<QEVertex> *plistQEVertex;
  CObjectListIter<QEVertex> iterQEVertex;  

  QEEdge *peCurr, *pePrev, *peStart;
  QEVertex *pvx;
  QEFacet *pfa;
  Point pt;

  listQEEdge.Init();

  //////////////////////////////////////////////////////////////////////
  // Copy all of the input vertices, form the contour edge loops, and 
  // place the edges in a heap sorted by the Y value of their originating
  // endpoint.

  iterContours.Init(plistContours);
  for ( plistQEVertex = (CObjectList<QEVertex>*)iterContours.PeekFirst();
	plistQEVertex != NULL;
	plistQEVertex = (CObjectList<QEVertex>*)iterContours.PeekNext() )
    {
      iterQEVertex.Init(plistQEVertex);

      pvx = iterQEVertex.PeekFirst();
      pt = pvx->GetPoint();
      pfa = CreateFacet();
      pvx = CreateVertex(pt);
      ASSERT( pvx != NULL );
      peCurr = CreateEdge();
      peCurr->SetFacet(pfa);
      peCurr->GetEdge()->SetFacet(m_pfaNULL);
      peCurr->SetVertex(pvx);

      m_plistQEEdge->Remove(peCurr->GetEdge());
      m_plistQEEdgeBoundary->InsertLast(peCurr->GetEdge());

      listQEEdge.InsertLast(peCurr, 0);
      
      pePrev = peCurr;
      peStart = peCurr;

      for ( pvx = iterQEVertex.PeekNext();
	    pvx != NULL;
	    pvx = iterQEVertex.PeekNext() )
	{
	  pt = pvx->GetPoint();
	  pvx = CreateVertex(pt);
	  ASSERT( pvx != NULL );

	  peCurr = CreateEdge();
	  peCurr->SetFacet(pfa);
	  peCurr->GetEdge()->SetFacet(m_pfaNULL);
	  peCurr->SetVertex(pvx);

	  m_plistQEEdge->Remove(peCurr->GetEdge());
	  m_plistQEEdgeBoundary->InsertLast(peCurr->GetEdge());

	  listQEEdge.InsertLast(peCurr, 0);

	  pePrev->GetEdge()->SetVertex(pvx);
	  Splice(1, peCurr, pePrev->GetEdge());

	  pePrev = peCurr;
	}

      pePrev->GetEdge()->SetVertex(peStart->GetVertex());
      Splice(1, peStart, pePrev->GetEdge());
    }

  Preprocess();
}
 
VOID QEMesh::CreateTriangulatedPolygon(CListInt *plistContours)
{
  ASSERT( plistContours != NULL );

  CListInt listQEEdge;
  CListIterInt iterContours;
  CObjectList<QEVertex> *plistQEVertex;
  CObjectListIter<QEVertex> iterQEVertex;  

  QEEdge *peCurr, *pePrev, *peStart;
  QEVertex *pvx;
  Point pt;

  listQEEdge.Init();

  //////////////////////////////////////////////////////////////////////
  // Copy all of the input vertices, form the contour edge loops, and 
  // place the edges in a heap sorted by the Y value of their originating
  // endpoint.

  iterContours.Init(plistContours);
  for ( plistQEVertex = (CObjectList<QEVertex>*)iterContours.PeekFirst();
	plistQEVertex != NULL;
	plistQEVertex = (CObjectList<QEVertex>*)iterContours.PeekNext() )
    {
      iterQEVertex.Init(plistQEVertex);

      pvx = iterQEVertex.PeekFirst();
      pt = pvx->GetPoint();
      pvx = CreateVertex(pt);
      ASSERT( pvx != NULL );
      peCurr = CreateEdge();
      peCurr->SetFacet(m_pfaNULL);
      peCurr->GetEdge()->SetFacet(m_pfaNULL);
      peCurr->SetVertex(pvx);

      m_plistQEEdge->Remove(peCurr->GetEdge());
      m_plistQEEdgeBoundary->InsertLast(peCurr->GetEdge());

      listQEEdge.InsertLast(peCurr, 0);
      
      pePrev = peCurr;
      peStart = peCurr;

      for ( pvx = iterQEVertex.PeekNext();
	    pvx != NULL;
	    pvx = iterQEVertex.PeekNext() )
	{
	  pt = pvx->GetPoint();
	  pvx = CreateVertex(pt);
	  ASSERT( pvx != NULL );

	  peCurr = CreateEdge();
	  peCurr->SetFacet(m_pfaNULL);
	  peCurr->GetEdge()->SetFacet(m_pfaNULL);
	  peCurr->SetVertex(pvx);

	  m_plistQEEdge->Remove(peCurr->GetEdge());
	  m_plistQEEdgeBoundary->InsertLast(peCurr->GetEdge());

	  listQEEdge.InsertLast(peCurr, 0);

	  pePrev->GetEdge()->SetVertex(pvx);
	  Splice(1, peCurr, pePrev->GetEdge());

	  pePrev = peCurr;
	}

      pePrev->GetEdge()->SetVertex(peStart->GetVertex());
      Splice(1, peStart, pePrev->GetEdge());
    }

  TriangulatePolygon(&listQEEdge);
}
 
VOID QEMesh::TriangulatePolygon(CListInt *plistQEEdge)
{
  static BOOL bFirst = TRUE;
  static CHeap<QEEdge*, QEVertex> heapQEEdge;
  static CList<QEEdge*, UINT_32> listContoursYMonotone;
  static CList<QEEdge*, UINT_32> listScan;
  static CListIter<QEEdge*, UINT_32> iterScan;

  if ( bFirst )
    {
      bFirst = FALSE;
      heapQEEdge.Init();
      listContoursYMonotone.Init();
      listScan.Init();
      iterScan.Init(&listScan);
    }

  CListInt listFlipCandidateEdges;
  CListIterInt iterQEEdge;
  QEEdge *peCurr, *pePrev, *pe, *peScan;
  Point ptCurr;
  //Point pt;
  FLOAT fX;
  QEEdge::VertexType eVertexType;

  //fprintf(stderr, "START: TriangulatePolygon\n");

  QEVertex::SetComparisonType(QEVertex::ctYXZ);

  listFlipCandidateEdges.Init();

  iterQEEdge.Init(plistQEEdge);
  for ( pe = (QEEdge*)(iterQEEdge.PeekFirst());
	pe != NULL;
	pe = (QEEdge*)(iterQEEdge.PeekNext()) )
    {
      heapQEEdge.Insert(*(pe->GetVertex()), pe);
      listFlipCandidateEdges.InsertLast(pe, 0);
    }

  //////////////////////////////////////////////////////////////////////
  // Process the Vertices

  for ( pe = heapQEEdge.RemoveTop();
	pe != NULL;
	pe = heapQEEdge.RemoveTop() )
    {
      peCurr = pe;
      pePrev = peCurr->GetEdgePrev(0);

      ptCurr = peCurr->GetVertex()->GetPoint();

      peCurr->SetVertexType(peCurr->CalculateVertexType());

      switch( peCurr->GetVertexType() )
	{
	case QEEdge::vtStart:
	  {
	    // Start Vertex
	    //fprintf(stderr, "Start Vertex:\n");
	    //peCurr->GetVertex()->Display();
	    
	    // Create a new Y Monotone contour
	    listContoursYMonotone.InsertLast(peCurr, 0);
	    
	    // Update Helper(Ei) <= Vi
	    peCurr->SetForwardEdge(peCurr);
	    
	    // Insert Ei into Scan Line
	    for ( peScan = iterScan.PeekFirst();
		  peScan != NULL;
		  peScan = iterScan.PeekNext() )
	      {
		fX = peScan->CalculateScanX(ptCurr(Y));
		if ( ptCurr(X) < fX )
		  {
		    iterScan.InsertPrev(peCurr, (UINT_32)peCurr);
		    break;
		  }
	      }
	    if ( peScan == NULL )
	      {
		listScan.InsertLast(peCurr, (UINT_32)peCurr);
	      }
	  }
	  break;
	case QEEdge::vtSplit:
	  {
	    // Split Vertex
	    //fprintf(stderr, "Split Vertex:\n");
	    //peCurr->GetVertex()->Display();
	    
	    // Find Ej in the Scan Line
	    pePrev = iterScan.PeekFirst();
	    for ( peScan = iterScan.PeekNext();
		  peScan != NULL;
		  peScan = iterScan.PeekNext() )
	      {
		fX = peScan->CalculateScanX(ptCurr(Y));
		if ( ptCurr(X) < fX )
		  {
		    break;
		  }
		pePrev = peScan;
	      }
	    ASSERT( pePrev != NULL );

	    // Create Diagonal Vi <=> Helper(Ej)
	    eVertexType = pePrev->GetForwardEdge()->GetVertexType();
	    //pe = InsertDiagonal(peCurr, pePrev->GetForwardEdge());
	    pe = InsertDiagonal(pePrev->GetForwardEdge(), peCurr);
	    listFlipCandidateEdges.InsertLast(pe, 0);

	    pe->SetVertexType(pe->CalculateVertexType());
	    pe->GetEdgePrev(1)->SetVertexType(pe->GetEdgePrev(1)->CalculateVertexType());
	    pe->GetEdge()->SetVertexType(pe->GetEdge()->CalculateVertexType());
	    pe->GetEdge()->GetEdgePrev(1)->SetVertexType(pe->GetEdge()->GetEdgePrev(1)->CalculateVertexType());
	    
	    // We should add at most one Y Monotone contour
	    ASSERT ( !( (eVertexType != QEEdge::vtStart && 
			 pePrev->GetForwardEdge()->GetVertexType() == QEEdge::vtStart) &&
			(pe->GetVertexType() == QEEdge::vtStart) ) );

	    if ( eVertexType != QEEdge::vtStart && 
		 pePrev->GetForwardEdge()->GetVertexType() == QEEdge::vtStart )
	      {
		// Create a new Y Monotone contour
		listContoursYMonotone.InsertLast(pePrev->GetForwardEdge(), 0);
	      }
	    if ( pe->GetVertexType() == QEEdge::vtStart )
	      {
		// Create a new Y Monotone contour
		listContoursYMonotone.InsertLast(pe, 0);
	      }

	    // Update Helper(Ej) <= Vi
	    pePrev->SetForwardEdge(pe->GetEdge());
	    // Update Helper(Ei) <= Vi'
	    peCurr->SetForwardEdge(peCurr);

	    // Insert Ei into Scan Line
	    for ( peScan = iterScan.PeekFirst();
		  peScan != NULL;
		  peScan = iterScan.PeekNext() )
	      {
		fX = peScan->CalculateScanX(ptCurr(Y));
		if ( ptCurr(X) < fX )
		  {
		    iterScan.InsertPrev(peCurr, (UINT_32)peCurr);
		    break;
		  }
	      }
	    if ( peScan == NULL )
	      {
		listScan.InsertLast(peCurr, (UINT_32)peCurr);
	      }
	  }
	  break;
	case QEEdge::vtMerge:
	  {
	    // Merge Vertex
	    //fprintf(stderr, "Merge Vertex:\n");
	    //peCurr->GetVertex()->Display();
	    
	    // If Helper(Ei-1) is Merge Vertex then Create Diagonal Vi <=> Helper(Ei-1)
	    if ( pePrev->GetForwardEdge()->GetVertexType() == QEEdge::vtMerge )
	      {
		pe = InsertDiagonal(pePrev->GetForwardEdge(), peCurr);
		listFlipCandidateEdges.InsertLast(pe, 0);

		pe->SetVertexType(pe->CalculateVertexType());
		pe->GetEdgePrev(1)->SetVertexType(pe->GetEdgePrev(1)->CalculateVertexType());
		pe->GetEdge()->SetVertexType(pe->GetEdge()->CalculateVertexType());
		pe->GetEdge()->GetEdgePrev(1)->SetVertexType(pe->GetEdge()->GetEdgePrev(1)->CalculateVertexType());
	      }
	    // Delete Ei-1 from the Scan Line
	    for ( peScan = iterScan.PeekFirst();
		  peScan != NULL;
		  peScan = iterScan.PeekNext() )
	      {
		if ( peScan == pePrev )
		  {
		    iterScan.PeekNextRemoveCurr();
		    break;
		  }
	      }
	    ASSERT( peScan == pePrev );

	    // Find Ej in the Scan Line
	    pePrev = iterScan.PeekFirst();
	    for ( peScan = iterScan.PeekNext();
		  peScan != NULL;
		  peScan = iterScan.PeekNext() )
	      {
		fX = peScan->CalculateScanX(ptCurr(Y));
		if ( ptCurr(X) < fX )
		  {
		    break;
		  }
		pePrev = peScan;
	      }
	    ASSERT( pePrev != NULL );

	    // If Helper(Ej) is a Merge Vertex then Create Diagonal Vi <=> Helper(Ej)
	    if ( pePrev->GetForwardEdge()->GetVertexType() == QEEdge::vtMerge )
	      {
		pe = InsertDiagonal(peCurr, pePrev->GetForwardEdge());
		listFlipCandidateEdges.InsertLast(pe, 0);

		pe->SetVertexType(pe->CalculateVertexType());
		pe->GetEdgePrev(1)->SetVertexType(pe->GetEdgePrev(1)->CalculateVertexType());
		pe->GetEdge()->SetVertexType(pe->GetEdge()->CalculateVertexType());
		pe->GetEdge()->GetEdgePrev(1)->SetVertexType(pe->GetEdge()->GetEdgePrev(1)->CalculateVertexType());
	      }

	    // Update Helper(Ej) <= Vi
	    pePrev->SetForwardEdge(pe);
	  }
	  break;
	case QEEdge::vtEnd:
	  {
	    // End Vertex
	    //fprintf(stderr, "End Vertex:\n");
	    //peCurr->GetVertex()->Display();
	    
	    // If Helper(Ei-1) is a Merge Vertex then Create Diagonal Vi <=> Helper(Ei-1)
	    if ( pePrev->GetForwardEdge()->GetVertexType() == QEEdge::vtMerge )
	      {
		pe = InsertDiagonal(peCurr, pePrev->GetForwardEdge());
		listFlipCandidateEdges.InsertLast(pe, 0);

		pe->SetVertexType(pe->CalculateVertexType());
		pe->GetEdgePrev(1)->SetVertexType(pe->GetEdgePrev(1)->CalculateVertexType());
		pe->GetEdge()->SetVertexType(pe->GetEdge()->CalculateVertexType());
		pe->GetEdge()->GetEdgePrev(1)->SetVertexType(pe->GetEdge()->GetEdgePrev(1)->CalculateVertexType());
	      }
	    // Delete Ei-1 from the Scan Line
	    for ( peScan = iterScan.PeekFirst();
		  peScan != NULL;
		  peScan = iterScan.PeekNext() )
	      {
		if ( peScan == pePrev )
		  {
		    iterScan.PeekNextRemoveCurr();
		    break;
		  }
	      }
	    ASSERT( peScan == pePrev );
	  }
	  break;
	case QEEdge::vtRight:
	  {
	    // Right Side Edge -- Material is to the left
	    //fprintf(stderr, "Regular Right Vertex:\n");
	    //peCurr->GetVertex()->Display();
	    
	    // Find Ej
	    pePrev = iterScan.PeekFirst();
	    for ( peScan = iterScan.PeekNext();
		  peScan != NULL;
		  peScan = iterScan.PeekNext() )
	      {
		fX = peScan->CalculateScanX(ptCurr(Y));
		if ( ptCurr(X) < fX )
		  {
		    break;
		  }
		pePrev = peScan;
	      }
	    ASSERT( pePrev != NULL );
	    // If Helper(Ej) is a Merge Vertex then Create Diagonal Vi <=> Helper(Ej)
	    if ( pePrev->GetForwardEdge()->GetVertexType() == QEEdge::vtMerge )
	      {
		pe = InsertDiagonal(peCurr, pePrev->GetForwardEdge());
		listFlipCandidateEdges.InsertLast(pe, 0);

		pe->SetVertexType(pe->CalculateVertexType());
		pe->GetEdgePrev(1)->SetVertexType(pe->GetEdgePrev(1)->CalculateVertexType());
		pe->GetEdge()->SetVertexType(pe->GetEdge()->CalculateVertexType());
		pe->GetEdge()->GetEdgePrev(1)->SetVertexType(pe->GetEdge()->GetEdgePrev(1)->CalculateVertexType());
	      }
	    // Update Helper(Ej) <= Vi
	    pePrev->SetForwardEdge(pe);
	  }
	  break;
	case QEEdge::vtLeft:
	  {
	    // Left Side Edge -- Material is to the right
	    //fprintf(stderr, "Regular Left Vertex:\n");
	    //peCurr->GetVertex()->Display();
	    
	    // If Helper(Ei-1) is a Merge Vertex then Create Diagonal Vi <=> Helper(Ei-1)
	    if ( pePrev->GetForwardEdge()->GetVertexType() == QEEdge::vtMerge )
	      {
		pe = InsertDiagonal(pePrev->GetForwardEdge(), peCurr);
		listFlipCandidateEdges.InsertLast(pe, 0);

		pe->SetVertexType(pe->CalculateVertexType());
		pe->GetEdgePrev(1)->SetVertexType(pe->GetEdgePrev(1)->CalculateVertexType());
		pe->GetEdge()->SetVertexType(pe->GetEdge()->CalculateVertexType());
		pe->GetEdge()->GetEdgePrev(1)->SetVertexType(pe->GetEdge()->GetEdgePrev(1)->CalculateVertexType());
	      }
	    // Delete Ei-1 from the Scan Line
	    for ( peScan = iterScan.PeekFirst();
		  peScan != NULL;
		  peScan = iterScan.PeekNext() )
	      {
		if ( peScan == pePrev )
		  {
		    iterScan.PeekNextRemoveCurr();
		    break;
		  }
	      }
	    ASSERT( peScan == pePrev );
	    
	    // Insert Ei into the Scan Line
	    for ( peScan = iterScan.PeekFirst();
		  peScan != NULL;
		  peScan = iterScan.PeekNext() )
	      {
		fX = peScan->CalculateScanX(ptCurr(Y));
		if ( ptCurr(X) < fX )
		  {
		    iterScan.InsertPrev(peCurr, (UINT_32)peCurr);
		    break;
		  }
	      }
	    if ( peScan == NULL )
	      {
		listScan.InsertLast(peCurr, (UINT_32)peCurr);
	      }
	    // Update Helper(Ei) <= Vi
	    peCurr->SetForwardEdge(peCurr);
	  }
	  break;
	default:
	  {
	    ASSERT( FALSE );
	  }
	  break;
	}
    }

  // Triangulate each of the Y Monotone contours
  for ( pe = listContoursYMonotone.RemoveFirst();
	pe != NULL;
	pe = listContoursYMonotone.RemoveFirst() )
    {      
      /*
      fprintf(stderr, "START: Y MONO\n");
      peCurr = pe;
      do
	{
	  peCurr->GetVertex()->Display();

	  peCurr = peCurr->GetEdgeNext(0);
	}
      while ( peCurr != pe );
      fprintf(stderr, "END  : Y MONO\n");
      fprintf(stderr, "START REV: Y MONO\n");
      peCurr = pe->GetEdgePrev(0);
      do
	{
	  peCurr->GetVertex()->Display();

	  peCurr = peCurr->GetEdgePrev(0);
	}
      while ( peCurr != pe->GetEdgePrev(0) );
      fprintf(stderr, "END REV   : Y MONO\n");
      */
      TriangulateYMonotone(pe, &listFlipCandidateEdges);
    }

  ASSERT( listContoursYMonotone.GetLength() == 0 );
  ASSERT( heapQEEdge.GetSize() == 0 );
  ASSERT( listScan.GetLength() == 0 );
/*
  QEMeshIter iterQEMesh;

  iterQEMesh.Init(this);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      ASSERT( pvx->GetEdge() != NULL );
      if ( pvx->GetEdge()->GetVertex() != pvx )
	{
	  fprintf(stderr, "VERTEX DOES NOT MATCH EDGE: After Tesselating\n");
	  pvx->GetPoint().Display();
	  pvx->GetEdge()->GetVertex()->GetPoint().Display();
	  pvx->GetEdge()->GetEdge()->GetVertex()->GetPoint().Display();	 
	  ASSERT( FALSE );
	}
    }
    */
  ConstrainedDelaunayEdgeFlipping(&listFlipCandidateEdges);
  /*
  iterQEMesh.Init(this);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      ASSERT( pvx->GetEdge() != NULL );
      if ( pvx->GetEdge()->GetVertex() != pvx )
	{
	  fprintf(stderr, "VERTEX DOES NOT MATCH EDGE: After edge flipping\n");
	  pvx->GetPoint().Display();
	  pvx->GetEdge()->GetVertex()->GetPoint().Display();
	  pvx->GetEdge()->GetEdge()->GetVertex()->GetPoint().Display();	 
	  ASSERT( FALSE );
	}
    }
    */

  //fprintf(stderr, "START: Triangulated Contours\n");
  //Display();
  //fprintf(stderr, "END  : Triangulated Contours\n");

  //fprintf(stderr, "END  : TriangulatePolygon\n");

  listFlipCandidateEdges.RemoveAll();
}

VOID QEMesh::TriangulateYMonotone(QEEdge *peContour, CListInt *plistFlipCandidateEdges)
{
  ASSERT( peContour != NULL );

  static BOOL bFirst = TRUE;
  static CList<QEEdge*, UINT_32> list;
  static CList<QEEdge*, UINT_32> stack;

  if ( bFirst )
    {
      bFirst = FALSE;
      list.Init();
      stack.Init();
    }

  QEEdge *peLeft, *peRight, *pe;
  QEEdge *peCurr, *pePrev, *pe0, *pe1, *pe2, *pePivot;
  QEFacet *pfa;
  UINT_32 iSide, iSidePrev;
  BOOL bFoundIt;

  // Find the top vertex
  bFoundIt = FALSE;
  peCurr = peContour;
  do
    {
      if ( (*(peCurr->GetVertex())) > (*(peCurr->GetEdgePrev(0)->GetVertex())) &&
	   (*(peCurr->GetVertex())) > (*(peCurr->GetEdge()->GetVertex())) )
	{
	  bFoundIt = TRUE;
	  /*
	  ASSERT( Counterclockwise(peCurr->GetEdgePrev(0)->GetVertex()->GetPoint(),
				   peCurr->GetVertex()->GetPoint(),
				   peCurr->GetEdge()->GetVertex()->GetPoint()) );
				   */

	  // Sort in Y order using one pass of merge sort

	  peLeft = peCurr;
	  peRight = peCurr->GetEdgePrev(0);

	  while ( peLeft != peRight )
	    {
	      if ( (*(peLeft->GetVertex())) < (*(peRight->GetVertex())) )
		{
		  list.InsertLast(peRight, 1);
		  peRight = peRight->GetEdgePrev(0);
		}
	      else
		{
		  list.InsertLast(peLeft, 0);
		  peLeft = peLeft->GetEdgeNext(0);
		}
	    }
	  // Insert the bottom vertex
	  list.InsertLast(peLeft, 0);

	  break;
	}

      peCurr = peCurr->GetEdgeNext(0);
    }
  while ( peCurr != peContour );
  ASSERT( bFoundIt );

  //////////////////////////////////////////////////////////////////////
  // Process the Vertices in order and create the triangles
  //

  peCurr = list.RemoveFirst(&iSide);
  stack.InsertFirst(peCurr, 0);
  peCurr = list.RemoveFirst(&iSide);
  stack.InsertFirst(peCurr, 0);

  iSidePrev = iSide;
  for( peCurr = list.RemoveFirst(&iSide);
       list.GetLength() > 0;
       peCurr = list.RemoveFirst(&iSide) )
    {
      if ( iSide != iSidePrev )
	{
	  // Different sides

	  if ( iSide == 0 )
	    {
	      // Left Side -- Interior to the Right

	      pePrev = stack.RemoveFirst();
	      pe0 = pePrev;
	      pePrev = pePrev->GetEdgePrev(0);
	      pePivot = peCurr;
	      for ( pe1 = stack.RemoveFirst();
		    pe1 != NULL;
		    pe1 = stack.RemoveFirst() )
		{
		  pfa = pe0->GetEdgePrev(0)->GetFacet();
		  
		  pe = InsertDiagonal(pePivot, pe0);
		  plistFlipCandidateEdges->InsertLast(pe, 0);

		  pe->GetEdge()->SetFacet(pfa);
		  pfa = CreateFacet();
		  pe->GetEdgePrev(0)->SetFacet(pfa);
		  pe->SetFacet(pfa);
		  pe->GetEdgeNext(0)->SetFacet(pfa);
		  pe->GetEdgeNext(0)->GetEdgeNext(0)->SetFacet(pfa);

		  pePivot = pe;
		  pe0 = pe1;
		}
	      pePrev = pePrev->GetEdgeNext(0);
	    }
	  else
	    {
	      // Right Side -- Interior to the Left

	      pePrev = stack.RemoveFirst();
	      pe0 = pePrev;
	      pePivot = peCurr;
	      peCurr = peCurr->GetEdgePrev(0);

	      for ( pe1 = stack.RemoveFirst();
		    pe1 != NULL;
		    pe1 = stack.RemoveFirst() )
		{
		  pfa = pe0->GetFacet();
		  pe = InsertDiagonal(pe0, pePivot);
		  plistFlipCandidateEdges->InsertLast(pe, 0);

		  pe->GetEdge()->SetFacet(pfa);
		  pfa = CreateFacet();
		  pe->GetEdgeNext(0)->SetFacet(pfa);
		  pe->SetFacet(pfa);
		  pe->GetEdgePrev(0)->SetFacet(pfa);

		  pe0 = pe1;
		}
	      peCurr = peCurr->GetEdgeNext(0);
	    }
	  stack.InsertFirst(pePrev, 0);
	  stack.InsertFirst(peCurr, 0);
	}
      else
	{
	  // Same side

	  if ( iSide == 0 )
	    {
	      // Left Side -- Interior to the Right

	      pe0 = stack.RemoveFirst();

	      for ( pe1 = stack.PeekFirst();
		    pe1 != NULL;
		    pe1 = stack.PeekFirst() )
		{
		  if ( Counterclockwise(pe1->GetVertex()->GetPoint(),
					pe0->GetVertex()->GetPoint(),
					peCurr->GetVertex()->GetPoint()) )
		    {
		      pe = InsertDiagonal(peCurr, pe1);
		      plistFlipCandidateEdges->InsertLast(pe, 0);
		      
		      pfa = CreateFacet();
		      pe->GetEdgePrev(0)->SetFacet(pfa);
		      pe->SetFacet(pfa);
		      pe->GetEdgeNext(0)->SetFacet(pfa);

		      pe0 = stack.RemoveFirst();
		      ASSERT( pe0 == pe1 );
		      pe0 = pe->GetEdge();
		    }
		  else
		    {
		      break;
		    }
		}
	    }
	  else
	    {
	      // Right Side -- Interior to the Left

	      pe0 = stack.RemoveFirst();

	      for ( pe1 = stack.PeekFirst();
		    pe1 != NULL;
		    pe1 = stack.PeekFirst() )
		{
		  if ( Counterclockwise(peCurr->GetVertex()->GetPoint(),
					pe0->GetVertex()->GetPoint(),
					pe1->GetVertex()->GetPoint()) )
		    {
		      pe = InsertDiagonal(pe1, peCurr);
		      plistFlipCandidateEdges->InsertLast(pe, 0);
		      
		      pfa = CreateFacet();
		      pe->GetEdgePrev(0)->SetFacet(pfa);
		      pe->SetFacet(pfa);
		      pe->GetEdgeNext(0)->SetFacet(pfa);

		      pe0 = stack.RemoveFirst();
		      ASSERT( pe0 == pe1 );
		      peCurr = pe->GetEdge();
		    }
		  else
		    {
		      break;
		    }
		}
	    }

	  stack.InsertFirst(pe0, 0);
	  stack.InsertFirst(peCurr, 0);
	}

      iSidePrev = iSide;
    }
  
  // Bottom vertex
  if ( iSidePrev == 0 )
    {
      // Left Side -- Interior to the Right

      pePivot = peCurr;
      pe0 = stack.RemoveFirst();
      ASSERT( pe0 != NULL );
      pe1 = stack.RemoveFirst();
      ASSERT( pe1 != NULL );
      for ( pe2 = stack.RemoveFirst();
	    pe2 != NULL;
	    pe2 = stack.RemoveFirst() )
	{
	  pe = InsertDiagonal(pePivot, pe1);
	  plistFlipCandidateEdges->InsertLast(pe, 0);
	  
	  pfa = CreateFacet();
	  pe->GetEdgePrev(0)->SetFacet(pfa);
	  pe->SetFacet(pfa);
	  pe->GetEdgeNext(0)->SetFacet(pfa);

	  pe0 = pe->GetEdge();
	  pe1 = pe2;
	}
    }
  else
    {
      // Right Side -- Interior to the Left

      pePivot = peCurr;
      pe0 = stack.RemoveFirst();
      ASSERT( pe0 != NULL );
      pe1 = stack.RemoveFirst();
      ASSERT( pe1 != NULL );
      for ( pe2 = stack.RemoveFirst();
	    pe2 != NULL;
	    pe2 = stack.RemoveFirst() )
	{
	  pe = InsertDiagonal(pe1, pePivot);
	  plistFlipCandidateEdges->InsertLast(pe, 0);
	  
	  pfa = CreateFacet();
	  pe->GetEdgePrev(0)->SetFacet(pfa);
	  pe->SetFacet(pfa);
	  pe->GetEdgeNext(0)->SetFacet(pfa);

	  pePivot = pe->GetEdge();
	  pe0 = pe1;
	  pe1 = pe2;
	}
    }

  // Make the last Triangle
  pfa = CreateFacet();
  pe0->GetEdgePrev(0)->SetFacet(pfa);
  pe0->SetFacet(pfa);
  pe0->GetEdgeNext(0)->SetFacet(pfa);

  ASSERT( list.GetLength() == 0 );
  ASSERT( stack.GetLength() == 0 );  
}

VOID QEMesh::ConstrainedDelaunayEdgeFlipping(CListInt *plistFlipCandidateEdges)
{
  static BOOL bFirst = TRUE;
  static CListInt list;

  if ( bFirst )
    {
      bFirst = FALSE;
      list.Init();
    }

  ASSERT( plistFlipCandidateEdges != NULL );

  CListIterInt iter;
  QEEdge *pe;

  iter.Init(plistFlipCandidateEdges);
  for ( pe = (QEEdge*)(iter.PeekFirst());
	pe != NULL;
	pe = (QEEdge*)(iter.PeekNext()) )
    {
      pe->SetForwardVertex(NULL);
      list.InsertLast(pe, 0);
    }

  for ( pe = (QEEdge*)list.RemoveFirst();
	pe != NULL;
	pe = (QEEdge*)list.RemoveFirst() )
    {
      if ( !ConstrainedDelaunayEdge(pe) &&
	   !(pe->GetForwardVertex() == pe->GetEdgePrev(0)->GetVertex() &&
	     pe->GetEdge()->GetForwardVertex() == pe->GetEdge()->GetEdgePrev(0)->GetVertex() ) )
	{
	  //fprintf(stderr, "Flip: %p: ", pe);
	  //pe->Display();
	  DelaunayFlip(pe, &list);
	  pe->SetForwardVertex(pe->GetEdgePrev(0)->GetVertex());
	  pe->GetEdge()->SetForwardVertex(pe->GetEdge()->GetEdgePrev(0)->GetVertex());
	}
    }
}

//////////////////////////////////////////////////////////////////////
// CreateVoronoiPolygon
//
VOID QEMesh::CreateVoronoiPolygon(VoronoiType eType, QEMesh *pmeCDT, BOOL bCleanUpContour)
{
  ASSERT( pmeCDT != NULL );

  switch ( eType )
    {
    case vtDelaunay:
      CreateVoronoiPolygonCDT(pmeCDT);
      break;
    case vtDivideAndConquer:
      CreateVoronoiPolygonDivideAndConquer(pmeCDT);
      break;
    case vtDivideAndConquerAlternating:
      CreateVoronoiPolygonHeld(pmeCDT, bCleanUpContour);
      break;
    case vtCones:
    case vtPlanes:
      CreateVoronoiPolygonChordalAxis(pmeCDT);
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  Preprocess();

  fprintf(stderr, "Finished computing Voronoi Diagram of a Polygon:\n");
  DisplayEuler();
}

//////////////////////////////////////////////////////////////////////
// CreateOffsetPolygon
//
VOID QEMesh::CreateOffsetPolygon(QEMesh *pmeCDT, QEMesh *pmeVoronoi, FLOAT fOffset)
{
  ASSERT( pmeCDT != NULL );
  ASSERT( pmeVoronoi != NULL );

  QEMeshIter iterVoronoi;
  QEFacet *pfaOffset;
  QEEdge *pe, *peNew, *peVoronoi, *peOffset, *peOffsetStart;
  QEVertex *pvxNew;
  Point ppt[2], pptResults[2], pt;
  UINT_32 uiResults;

  if ( fOffset == 0.0 )
    {
      //////////////////////////////////////////////////
      // This is necessary because the Voronoi diagram 
      //   includes the original contour and this caused
      //   numeric problems.
      pfaOffset = NULL;

      iterVoronoi.Init(pmeCDT);
      for ( peVoronoi = iterVoronoi.PeekFirstEdgeBoundary();
	    peVoronoi != NULL;
	    peVoronoi = iterVoronoi.PeekNextEdgeBoundary() )
	{
	  peVoronoi->SetForwardEdge(NULL);
	}

      for ( peVoronoi = iterVoronoi.PeekFirstEdgeBoundary();
	    peVoronoi != NULL;
	    peVoronoi = iterVoronoi.PeekNextEdgeBoundary() )
	{
	  if ( peVoronoi->GetForwardEdge() == NULL )
	    {
	      //////////////////////////////////////////////////
	      // Copy this contour and create a new face
	      peVoronoi->SetForwardEdge(peVoronoi);

	      pe = peVoronoi;

	      pfaOffset = CreateFacet();

	      pt = pe->GetEdge()->GetVertex()->GetPoint();
	      pt[2] = 0.0;
	      pvxNew = CreateVertex(pt);
	      peNew = CreateEdge();
	      peNew->SetKnotInterval(0.0);
	      peNew->GetEdge()->SetKnotInterval(0.0);
	      peNew->GetEdge()->SetFacet(pfaOffset);
	      peNew->GetEdge()->SetVertex(pvxNew);
	      peOffset = peNew;
	      peOffsetStart = peNew;

	      ASSERT( pe->GetEdge()->GetFacet()->GetForwardVertex() == NULL );

	      for ( pe = pe->GetEdgePrev(0);
		    pe != peVoronoi;
		    pe = pe->GetEdgePrev(0) )
		{
		  pe->SetForwardEdge(pe);

		  pt = pe->GetEdge()->GetVertex()->GetPoint();
		  pt[2] = 0.0;
		  pvxNew = CreateVertex(pt);
		  peNew = CreateEdge();
		  peNew->GetEdge()->SetFacet(pfaOffset);
		  peNew->GetEdge()->SetVertex(pvxNew);
	      
		  ASSERT( pe->GetEdge()->GetFacet()->GetForwardVertex() == NULL );

		  peOffset->SetVertex(peNew->GetEdge()->GetVertex());
		  Splice(1, peNew->GetEdge(), peOffset);
	      
		  peOffset = peNew;
		}
	      peOffset->SetVertex(peOffsetStart->GetEdge()->GetVertex());
	      Splice(1, peOffset, peOffsetStart->GetEdge());
	    }
	}
    }
  else
    {
      pfaOffset = NULL;

      iterVoronoi.Init(pmeVoronoi);
      for ( peVoronoi = iterVoronoi.PeekFirstEdge();
	    peVoronoi != NULL;
	    peVoronoi = iterVoronoi.PeekNextEdge() )
	{
	  peVoronoi->SetForwardEdge(NULL);
	}

      for ( peVoronoi = iterVoronoi.PeekFirstEdge();
	    peVoronoi != NULL;
	    peVoronoi = iterVoronoi.PeekNextEdge() )
	{
	  if ( peVoronoi->GetForwardEdge() == NULL )
	    {
	      peVoronoi->SetForwardEdge(peVoronoi);
	      peVoronoi->GetEdge()->SetForwardEdge(peVoronoi->GetEdge());

	      //////////////////////////////////////////////////
	      // This is necessary because the Voronoi diagram 
	      //   includes the original contour and this caused
	      //   numeric problems.
	      if ( peVoronoi->GetFacet()->GetForwardEdge() == peVoronoi )
		{
		  continue;
		}

	      ppt[0] = peVoronoi->GetVertex()->GetPoint();
	      ppt[1] = peVoronoi->GetEdge()->GetVertex()->GetPoint();
      
	      if ( ( ppt[0](3) != 0.0 && ppt[0](2) <= fOffset &&
		     ppt[1](3) != 0.0 && ppt[1](2) >= fOffset ) ||
		   ( ppt[0](3) != 0.0 && ppt[0](2) >= fOffset &&
		     ppt[1](3) != 0.0 && ppt[1](2) <= fOffset ) ||
		   ( ppt[0](3) == 0.0 &&
		     ppt[1](3) != 0.0 && ppt[1](2) > fOffset ) ||	     
		   ( ppt[0](3) != 0.0 && ppt[0](2) > fOffset &&
		     ppt[1](3) == 0.0 ) )
		{
		  if ( pfaOffset == NULL )
		    {
		      pfaOffset = CreateFacet();
		    }

		  if ( ppt[0](3) == 0.0 ||
		       ( ppt[1](3) != 0.0 && fabs(ppt[0](2)) > fabs(ppt[1](2)) ) )
		    {
		      peVoronoi = peVoronoi->GetEdge();
		    }

		  CalculateOffsetPoint(pptResults, uiResults, 
				       peVoronoi, 
				       peVoronoi->GetFacet(), peVoronoi->GetEdge()->GetFacet(), 
				       fOffset);
		  switch ( uiResults )
		    {
		    case 0:
		      {
			//ASSERT( FALSE );
			uiResults = 1;
			if ( ppt[0](3) == 0.0 )
			  {
			    pptResults[0] = ppt[1];
			  }
			else if ( ppt[1](3) == 0.0 ) 
			  {
			    pptResults[0] = ppt[0];
			  }
			else if ( ppt[0](2) - fOffset < ppt[1](2) - fOffset ) 
			  {
			    pptResults[0] = ppt[0];
			  }
			else
			  {
			    pptResults[0] = ppt[1];
			  }
		      }
		    break;
		    case 1:
		      {
		      }
		    break;
		    case 2:
		      {
			if ( ( peVoronoi->GetFacet()->GetForwardVertex() != NULL &&
			       peVoronoi->GetEdge()->GetFacet()->GetForwardEdge() != NULL ) ||
			     ( peVoronoi->GetFacet()->GetForwardEdge() != NULL &&
			       peVoronoi->GetEdge()->GetFacet()->GetForwardVertex() != NULL ) ||
			     ( peVoronoi->GetFacet()->GetForwardVertex() != NULL &&
			       peVoronoi->GetEdge()->GetFacet()->GetForwardVertex() != NULL ) )
			  {
			    ppt[0] = peVoronoi->GetVertex()->GetPoint();
			    ppt[0][2] = 0.0;
			    ppt[1] = peVoronoi->GetEdge()->GetVertex()->GetPoint();
			    ppt[1][2] = 0.0;		    

			    if ( DistanceFromEdgeZone(pptResults[1], ppt[0], ppt[1]) < 
				 DistanceFromEdgeZone(pptResults[0], ppt[0], ppt[1]) )
			      {
				pptResults[0] = pptResults[1];
			      }
			  }
		      }
		    break;
		    default:
		      {
			ASSERT( FALSE );
		      }
		    break;
		    }

		  pvxNew = CreateVertex(pptResults[0]);
		  peNew = CreateEdge();
	      
		  peNew->GetEdge()->SetFacet(pfaOffset);
		  peNew->GetEdge()->SetVertex(pvxNew);
		  peOffsetStart = peNew;
		  peOffset = peNew;
	      
		  if ( peVoronoi->GetEdge()->GetFacet()->GetForwardVertex() != NULL )
		    {
		      ASSERT( peVoronoi->GetEdge()->GetFacet()->GetForwardVertex()->GetEdge()->GetVertex() == peVoronoi->GetEdge()->GetFacet()->GetForwardVertex() );
		      peNew->SetForwardEdge(peVoronoi->GetEdge()->GetFacet()->GetForwardVertex()->GetEdge());
		      ASSERT( peVoronoi->GetEdge()->GetFacet()->GetForwardVertex()->GetEdge()->GetVertex() == peVoronoi->GetEdge()->GetFacet()->GetForwardVertex() );
		      peNew->GetEdge()->SetForwardEdge(peVoronoi->GetEdge()->GetFacet()->GetForwardVertex()->GetEdge());
		    }

		  for ( pe = peVoronoi->GetEdge()->GetEdgePrev(0);
			pe != peVoronoi;
			pe = pe->GetEdgePrev(0) )
		    {
		      if ( pe->GetEdge()->GetFacet() == GetNullFacet() )
			{
			  //ASSERT( FALSE );
			  pe = pe->GetEdgePrev(0);
			  if ( pe == peVoronoi )
			    {
			      break;
			    }
			}

		      pe->SetForwardEdge(pe);
		      pe->GetEdge()->SetForwardEdge(pe->GetEdge());

		      ppt[0] = pe->GetVertex()->GetPoint();
		      ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
	      
		      if ( ( ppt[0](3) != 0.0 && ppt[0](2) <= fOffset &&
			     ppt[1](3) != 0.0 && ppt[1](2) >= fOffset ) ||
			   ( ppt[0](3) != 0.0 && ppt[0](2) >= fOffset &&
			     ppt[1](3) != 0.0 && ppt[1](2) <= fOffset ) ||
			   ( ppt[0](3) == 0.0 &&
			     ppt[1](3) != 0.0 && ppt[1](2) > fOffset ) ||	     
			   ( ppt[0](3) != 0.0 && ppt[0](2) > fOffset &&
			     ppt[1](3) == 0.0 ) )
			{
			  CalculateOffsetPoint(pptResults, uiResults, 
					       pe, 
					       pe->GetFacet(), pe->GetEdge()->GetFacet(), 
					       fOffset);
			  switch ( uiResults )
			    {
			    case 0:
			      {
				//ASSERT( FALSE );
				uiResults = 1;
				if ( ppt[0](3) == 0.0 )
				  {
				    pptResults[0] = ppt[1];
				  }
				else if ( ppt[1](3) == 0.0 ) 
				  {
				    pptResults[0] = ppt[0];
				  }
				else if ( ppt[0](2) - fOffset < ppt[1](2) - fOffset ) 
				  {
				    pptResults[0] = ppt[0];
				  }
				else
				  {
				    pptResults[0] = ppt[1];
				  }
			      }
			    break;
			    case 1:
			      {
			      }
			    break;
			    case 2:
			      {
				if ( ( pe->GetFacet()->GetForwardVertex() != NULL &&
				       pe->GetEdge()->GetFacet()->GetForwardEdge() != NULL ) ||
				     ( pe->GetFacet()->GetForwardEdge() != NULL &&
				       pe->GetEdge()->GetFacet()->GetForwardVertex() != NULL ) ||
				     ( pe->GetFacet()->GetForwardVertex() != NULL &&
				       pe->GetEdge()->GetFacet()->GetForwardVertex() != NULL ) )
				  {
				    ppt[0] = pe->GetVertex()->GetPoint();
				    ppt[0][2] = 0.0;
				    ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
				    ppt[1][2] = 0.0;		    
				
				    if ( DistanceFromEdgeZone(pptResults[1], ppt[0], ppt[1]) < 
					 DistanceFromEdgeZone(pptResults[0], ppt[0], ppt[1]) )
				      {
					pptResults[0] = pptResults[1];
				      }
				  }
			      }
			    break;

			    default:
			      {
				ASSERT( FALSE );
			      }
			    break;
			    }

			  pvxNew = CreateVertex(pptResults[0]);
			  peNew = CreateEdge();
			
			  peNew->GetEdge()->SetFacet(pfaOffset);
			  peOffset->SetVertex(pvxNew);
			  peNew->GetEdge()->SetVertex(pvxNew);
			  Splice(1, peOffset, peNew->GetEdge());
			  peOffset = peNew;

			  if ( pe->GetEdge()->GetFacet()->GetForwardVertex() != NULL )
			    {
			      ASSERT( pe->GetEdge()->GetFacet()->GetForwardVertex()->GetEdge()->GetVertex() == pe->GetEdge()->GetFacet()->GetForwardVertex() );
			      peNew->SetForwardEdge(pe->GetEdge()->GetFacet()->GetForwardVertex()->GetEdge());
			      ASSERT( pe->GetEdge()->GetFacet()->GetForwardVertex()->GetEdge()->GetVertex() == pe->GetEdge()->GetFacet()->GetForwardVertex() );
			      peNew->GetEdge()->SetForwardEdge(pe->GetEdge()->GetFacet()->GetForwardVertex()->GetEdge());
			    }

			  pe = pe->GetEdge();
			}
		    }

		  if ( peOffset == peOffsetStart )
		    {
		      ASSERT( FALSE );
		      DestroyVertex(peOffsetStart->GetEdge()->GetVertex());
		      DestroyEdge(peOffsetStart->GetEdge());
		      DestroyEdge(peOffsetStart);
		    }
		  else
		    {
		      peOffset->SetVertex(peOffsetStart->GetEdge()->GetVertex());
		      Splice(1, peOffset, peOffsetStart->GetEdge());
		    }
		}
	    }
	}
    }

  //fprintf(stderr, "Finished computing Offset:\n");
  //DisplayEuler();
}

/*
VOID QEMesh::CreateOffsetPolygon(QEMesh *pmeCDT, QEMesh *pmeVoronoi, FLOAT fOffset)
{
  ASSERT( pmeCDT != NULL );
  ASSERT( pmeVoronoi != NULL );

  QEMeshIter iterVoronoi;
  QEEdge *pe, *peNew, *peVoronoi, *peCDT, *peCDTStart, *peOffset, *peOffsetStart;
  QEVertex *pvxCDT, *pvxNew;
  Point ppt[2], pptResults[2];
  UINT_32 uiResults;

  CList<QEEdge*, UINT_32> listTStack;
  CList<QEEdge*, UINT_32> listCStack;
  QEMesh::OffsetType eState;

  listTStack.Init();
  listCStack.Init();
  eState = otStart;

  peVoronoi = NULL;
  peCDT = NULL;
  pvxCDT = NULL;
  iterVoronoi.Init(pmeVoronoi);
  for ( pe = iterVoronoi.PeekFirstEdge();
	pe != NULL;
	pe = iterVoronoi.PeekNextEdge() )
    {
      if ( pe->GetEdge()->GetFacet() == m_pfaNULL &&
	   pe->GetFacet() != m_pfaNULL &&
	   pe->GetFacet()->GetForwardVertex() != NULL )
	{
	  pvxCDT = pe->GetFacet()->GetForwardVertex();

	  peVoronoi = pe->GetEdgeNext(0);
	  while ( !( peVoronoi->GetEdge()->GetFacet()->GetForwardEdge() != NULL &&
		     peVoronoi->GetEdge()->GetFacet()->GetForwardEdge()->GetVertex() == pvxCDT ) )
	    {
	      peVoronoi = peVoronoi->GetEdgeNext(0);
	    }
	  peVoronoi = peVoronoi->GetEdge();
	  peCDT = peVoronoi->GetFacet()->GetForwardEdge()->GetEdge();
	  pvxCDT = NULL;
	  break;
	}
    }

  peCDTStart = peCDT;
  peOffset = CreateEdge();
  peOffsetStart = peOffset;

  do
    {
      for ( ; 
	    !( ( peVoronoi->GetEdge()->GetFacet()->GetForwardVertex() != NULL &&
		 peVoronoi->GetEdge()->GetFacet()->GetForwardVertex() == peCDT->GetVertex() ) ||
	       ( peVoronoi->GetEdge()->GetFacet()->GetForwardEdge() != NULL &&
		 peVoronoi->GetEdge()->GetFacet()->GetForwardEdge() == peCDT->GetEdgePrev(0)->GetEdge() ) );
	    peVoronoi = peVoronoi->GetEdgePrev(0) )
	{
	  ppt[0] = peVoronoi->GetVertex()->GetPoint();
	  ppt[1] = peVoronoi->GetEdge()->GetVertex()->GetPoint();
	  
	  if ( ( ppt[0](3) != 0.0 && ppt[0](2) <= fOffset &&
		 ppt[1](3) != 0.0 && ppt[1](2) >= fOffset ) ||
	       ( ppt[0](3) != 0.0 && ppt[0](2) >= fOffset &&
		 ppt[1](3) != 0.0 && ppt[1](2) <= fOffset ) ||
	       ( ppt[0](3) == 0.0 &&
		 ppt[1](3) != 0.0 && ppt[1](2) > fOffset ) ||
	       ( ppt[0](3) != 0.0 && ppt[0](2) > fOffset &&
		 ppt[1](3) == 0.0 ) )
	    {
	      CalculateOffsetPoint(pptResults, uiResults, 
				   peVoronoi, 
				   peVoronoi->GetFacet(), peVoronoi->GetEdge()->GetFacet(), 
				   fOffset);
	      switch ( uiResults )
		{
		case 0:
		  {
		  }
		break;
		case 1:
		  {
		    pvxNew = CreateVertex(pptResults[0]);
		    peNew = CreateEdge();

		    peOffset->SetVertex(pvxNew);
		    peNew->GetEdge()->SetVertex(pvxNew);
		    Splice(1, peOffset, peNew->GetEdge());
		    peOffset = peNew;
		  }
		break;
		case 2:
		  {
		    pvxNew = CreateVertex(pptResults[0]);
		    peNew = CreateEdge();

		    peOffset->SetVertex(pvxNew);
		    peNew->GetEdge()->SetVertex(pvxNew);
		    Splice(1, peOffset, peNew->GetEdge());
		    peOffset = peNew;

		    pvxNew = CreateVertex(pptResults[1]);
		    peNew = CreateEdge();

		    peOffset->SetVertex(pvxNew);
		    peNew->GetEdge()->SetVertex(pvxNew);
		    Splice(1, peOffset, peNew->GetEdge());
		    peOffset = peNew;
		  }
		break;
		default:
		  {
		    ASSERT( FALSE );
		  }
		break;
		}
	    }
	}

      if ( peVoronoi->GetEdge()->GetFacet()->GetForwardVertex() != NULL &&
	   peVoronoi->GetEdge()->GetFacet()->GetForwardVertex() == peCDT->GetVertex() ) 
	{
	  pvxCDT = peCDT->GetVertex();
	}
      else if ( peVoronoi->GetEdge()->GetFacet()->GetForwardEdge() != NULL &&
		peVoronoi->GetEdge()->GetFacet()->GetForwardEdge() == peCDT->GetEdgePrev(0)->GetEdge() )
	{
	  pvxCDT = NULL;
	  peCDT = peCDT->GetEdgePrev(0);
	}
      peVoronoi = peVoronoi->GetEdge();
    }
  while ( !( peCDT == peCDTStart && pvxCDT == NULL ) );

  if ( peOffset != peOffsetStart )
    {
      peOffset->SetVertex(peOffsetStart->GetVertex());
      Splice(1, peOffsetStart->GetEdgePrev(1), peOffset);
      Splice(1, peOffsetStart->GetEdgePrev(1), peOffsetStart);
    }

  DestroyEdge(peOffsetStart->GetEdge());
  DestroyEdge(peOffsetStart);
}
*/

//////////////////////////////////////////////////////////////////////
// CreateMillingContour
//
VOID QEMesh::CreateMillingContour(QEMesh *pmeCDT, QEMesh *pmeVoronoi, FLOAT fToolRadius)
{
  ASSERT( pmeCDT != NULL );
  ASSERT( pmeVoronoi != NULL );

  QEMeshIter iterVoronoi;
  QEFacet *pfaOffset;
  QEEdge *pe, *peNew, *peVoronoi, *peOffset, *peOffsetStart;
  QEVertex *pvxNew;
  Point ppt[2], pptResults[2], pt;
  //Point ptStart;
  UINT_32 uiResults;
  BOOL bRemoveLastEdge;

  if ( fToolRadius == 0.0 )
    {
      //////////////////////////////////////////////////
      // This is necessary because the Voronoi diagram 
      //   includes the original contour and this caused
      //   numeric problems.
      pfaOffset = NULL;

      iterVoronoi.Init(pmeCDT);
      for ( peVoronoi = iterVoronoi.PeekFirstEdgeBoundary();
	    peVoronoi != NULL;
	    peVoronoi = iterVoronoi.PeekNextEdgeBoundary() )
	{
	  peVoronoi->SetForwardEdge(NULL);
	}

      for ( peVoronoi = iterVoronoi.PeekFirstEdgeBoundary();
	    peVoronoi != NULL;
	    peVoronoi = iterVoronoi.PeekNextEdgeBoundary() )
	{
	  if ( peVoronoi->GetForwardEdge() == NULL )
	    {
	      //////////////////////////////////////////////////
	      // Copy this contour and create a new face
	      peVoronoi->SetForwardEdge(peVoronoi);

	      pe = peVoronoi;

	      pfaOffset = CreateFacet();

	      pt = pe->GetEdge()->GetVertex()->GetPoint();
	      pt[2] = 0.0;
	      pvxNew = CreateVertex(pt);
	      peNew = CreateEdge();
	      peNew->GetEdge()->SetFacet(pfaOffset);
	      peNew->GetEdge()->SetVertex(pvxNew);
	      peOffset = peNew;
	      peOffsetStart = peNew;

	      ASSERT( pe->GetEdge()->GetFacet()->GetForwardVertex() == NULL );

	      for ( pe = pe->GetEdgePrev(0);
		    pe != peVoronoi;
		    pe = pe->GetEdgePrev(0) )
		{
		  pe->SetForwardEdge(pe);

		  pt = pe->GetEdge()->GetVertex()->GetPoint();
		  pt[2] = 0.0;
		  pvxNew = CreateVertex(pt);
		  peNew = CreateEdge();
		  peNew->GetEdge()->SetFacet(pfaOffset);
		  peNew->GetEdge()->SetVertex(pvxNew);
	      
		  ASSERT( pe->GetEdge()->GetFacet()->GetForwardVertex() == NULL );

		  peOffset->SetVertex(peNew->GetEdge()->GetVertex());
		  Splice(1, peNew->GetEdge(), peOffset);
	      
		  peOffset = peNew;
		}
	      peOffset->SetVertex(peOffsetStart->GetEdge()->GetVertex());
	      Splice(1, peOffset, peOffsetStart->GetEdge());
	    }
	}
    }
  else
    {
      pfaOffset = NULL;

      iterVoronoi.Init(pmeVoronoi);
      for ( peVoronoi = iterVoronoi.PeekFirstEdge();
	    peVoronoi != NULL;
	    peVoronoi = iterVoronoi.PeekNextEdge() )
	{
	  peVoronoi->SetForwardEdge(NULL);
	}

      for ( peVoronoi = iterVoronoi.PeekFirstEdge();
	    peVoronoi != NULL;
	    peVoronoi = iterVoronoi.PeekNextEdge() )
	{
	  if ( peVoronoi->GetForwardEdge() == NULL )
	    {
	      peVoronoi->SetForwardEdge(peVoronoi);
	      peVoronoi->GetEdge()->SetForwardEdge(peVoronoi->GetEdge());

	      //////////////////////////////////////////////////
	      // This is necessary because the Voronoi diagram 
	      //   includes the original contour and this caused
	      //   numeric problems.
	      if ( peVoronoi->GetFacet()->GetForwardEdge() == peVoronoi )
		{
		  continue;
		}

	      ppt[0] = peVoronoi->GetVertex()->GetPoint();
	      ppt[1] = peVoronoi->GetEdge()->GetVertex()->GetPoint();
      
	      if ( ( ppt[0](3) != 0.0 && ppt[0](2) <= fToolRadius &&
		     ppt[1](3) != 0.0 && ppt[1](2) >= fToolRadius ) ||
		   ( ppt[0](3) != 0.0 && ppt[0](2) >= fToolRadius &&
		     ppt[1](3) != 0.0 && ppt[1](2) <= fToolRadius ) ||
		   ( ppt[0](3) == 0.0 &&
		     ppt[1](3) != 0.0 && ppt[1](2) > fToolRadius ) ||	     
		   ( ppt[0](3) != 0.0 && ppt[0](2) > fToolRadius &&
		     ppt[1](3) == 0.0 ) )
		{
		  if ( pfaOffset == NULL )
		    {
		      pfaOffset = CreateFacet();
		    }

		  if ( ppt[1](3) == 0.0 ||
		       ( ppt[0](3) != 0.0 && ppt[0](2) > ppt[1](2) ) )
		    {
		      peVoronoi = peVoronoi->GetEdge();
		    }

		  bRemoveLastEdge = FALSE;
		  pe = peVoronoi;

		  CalculateOffsetPoint(pptResults, uiResults, 
				       pe, 
				       pe->GetFacet(), pe->GetEdge()->GetFacet(), 
				       fToolRadius);
		  switch ( uiResults )
		    {
		    case 0:
		      {
			//ASSERT( FALSE );
			uiResults = 1;
			if ( ppt[0](3) == 0.0 )
			  {
			    pptResults[0] = ppt[1];
			  }
			else if ( ppt[1](3) == 0.0 ) 
			  {
			    pptResults[0] = ppt[0];
			  }
			else if ( ppt[0](2) - fToolRadius < ppt[1](2) - fToolRadius ) 
			  {
			    pptResults[0] = ppt[0];
			  }
			else
			  {
			    pptResults[0] = ppt[1];
			  }
		      }
		    break;
		    case 1:
		      {
		      }
		    break;
		    case 2:
		      {
			if ( ( pe->GetFacet()->GetForwardVertex() != NULL &&
			       pe->GetEdge()->GetFacet()->GetForwardEdge() != NULL ) ||
			     ( pe->GetFacet()->GetForwardEdge() != NULL &&
			       pe->GetEdge()->GetFacet()->GetForwardVertex() != NULL ) ||
			     ( pe->GetFacet()->GetForwardVertex() != NULL &&
			       pe->GetEdge()->GetFacet()->GetForwardVertex() != NULL ) )
			  {
			    ppt[0] = pe->GetVertex()->GetPoint();
			    ppt[0][2] = 0.0;
			    ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
			    ppt[1][2] = 0.0;		    
			    
			    if ( DistanceFromEdgeZone(pptResults[1], ppt[0], ppt[1]) < 
				 DistanceFromEdgeZone(pptResults[0], ppt[0], ppt[1]) )
			      {
				pptResults[0] = pptResults[1];
			      }
			  }
		      }
		    break;
		    default:
		      {
			ASSERT( FALSE );
		      }
		    break;
		    }

		  if ( pe->GetFacet()->GetForwardVertex() != NULL &&
		       pe->GetEdge()->GetFacet()->GetForwardVertex() != NULL &&
		       (pe->GetFacet()->GetForwardVertex() ==
			pe->GetEdge()->GetFacet()->GetForwardVertex()) )
		    {
		      // Crossing from vertex zone to same vertex zone
		      pvxNew = CreateVertex(pe->GetFacet()->GetForwardVertex()->GetPoint());
		      peNew = CreateEdge();
		      
		      peNew->GetEdge()->SetFacet(pfaOffset);
		      peNew->GetEdge()->SetVertex(pvxNew);
		      peOffset = peNew;
		      peOffsetStart = peNew;
		      bRemoveLastEdge = TRUE;
		    }
		  else if ( pe->GetFacet()->GetForwardVertex() != NULL &&
		       pe->GetEdge()->GetFacet()->GetForwardEdge() != NULL &&
		       (pe->GetFacet()->GetForwardVertex() ==
			pe->GetEdge()->GetFacet()->GetForwardEdge()->GetVertex()) )
		    {
		      // Crossing from vertex zone to adjacent edge zone 
		      pvxNew = CreateVertex(pe->GetFacet()->GetForwardVertex()->GetPoint());
		      peNew = CreateEdge();
		      
		      peNew->GetEdge()->SetFacet(pfaOffset);
		      peNew->GetEdge()->SetVertex(pvxNew);
		      peOffset = peNew;
		      peOffsetStart = peNew;
		      bRemoveLastEdge = TRUE;
		    }
		  else if ( pe->GetFacet()->GetForwardEdge() != NULL &&
			    pe->GetEdge()->GetFacet()->GetForwardVertex() != NULL &&
			    (pe->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex() == 
			     pe->GetEdge()->GetFacet()->GetForwardVertex()) )
		    {
		      // Crossing from edge zone to adjacent vertex zone 
		      pvxNew = CreateVertex(pe->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint());
		      peNew = CreateEdge();
		      
		      peNew->GetEdge()->SetFacet(pfaOffset);
		      peNew->GetEdge()->SetVertex(pvxNew);
		      peOffset = peNew;
		      peOffsetStart = peNew;
		    }
		  else 
		    {
		      if ( pe->GetFacet()->GetForwardVertex() != NULL )
			{			  
			  pt = pe->GetFacet()->GetForwardVertex()->GetPoint();
			}
		      else
			{			  
			  ppt[0] = pe->GetFacet()->GetForwardEdge()->GetVertex()->GetPoint();
			  ppt[1] = pe->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint();
			  pt = CalculatePointProjectionOnLine(pptResults[0], ppt[0], ppt[1], FALSE);
			}

		      pvxNew = CreateVertex(pt);
		      peNew = CreateEdge();

		      peNew->GetEdge()->SetVertex(pvxNew);
		      peNew->GetEdge()->SetFacet(pfaOffset);
		      peOffset = peNew;
		      peOffsetStart = peNew;
		      
		      // Create circular arc center which will be unattached
		      pvxNew = CreateVertex(pptResults[0]);
		      peNew = CreateEdge();
		      peNew->SetVertex(pvxNew);
		      peOffset->SetForwardEdge(peNew);
		      peOffset->GetEdge()->SetForwardEdge(peNew);

		      if ( pe->GetEdge()->GetFacet()->GetForwardVertex() != NULL )
			{			  
			  pt = pe->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint();
			}
		      else
			{			  
			  ppt[0] = pe->GetEdge()->GetFacet()->GetForwardEdge()->GetVertex()->GetPoint();
			  ppt[1] = pe->GetEdge()->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint();
			  pt = CalculatePointProjectionOnLine(pptResults[0], ppt[0], ppt[1], FALSE);
			}
		      
		      pvxNew = CreateVertex(pt);
		      peNew = CreateEdge();
		      
		      peNew->GetEdge()->SetFacet(pfaOffset);
		      peOffset->SetVertex(pvxNew);
		      peNew->GetEdge()->SetVertex(pvxNew);
		      Splice(1, peOffset, peNew->GetEdge());
		      peOffset = peNew;
		    }

		  for ( pe = pe->GetEdge()->GetEdgePrev(0);
			pe != peVoronoi;
			pe = pe->GetEdgePrev(0) )
		    {
		      pe->SetForwardEdge(pe);
		      pe->GetEdge()->SetForwardEdge(pe->GetEdge());

		      ppt[0] = pe->GetVertex()->GetPoint();
		      ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
	      
		      if ( ( ppt[0](3) != 0.0 && ppt[0](2) <= fToolRadius &&
			     ppt[1](3) != 0.0 && ppt[1](2) >= fToolRadius ) ||
			   ( ppt[0](3) != 0.0 && ppt[0](2) >= fToolRadius &&
			     ppt[1](3) != 0.0 && ppt[1](2) <= fToolRadius ) ||
			   ( ppt[0](3) == 0.0 &&
			     ppt[1](3) != 0.0 && ppt[1](2) > fToolRadius ) ||	     
			   ( ppt[0](3) != 0.0 && ppt[0](2) > fToolRadius &&
			     ppt[1](3) == 0.0 ) )
			{
			  CalculateOffsetPoint(pptResults, uiResults, 
					       pe, 
					       pe->GetFacet(), pe->GetEdge()->GetFacet(), 
					       fToolRadius);
			  switch ( uiResults )
			    {
			    case 0:
			      {
				//ASSERT( FALSE );
				uiResults = 1;
				if ( ppt[0](3) == 0.0 )
				  {
				    pptResults[0] = ppt[1];
				  }
				else if ( ppt[1](3) == 0.0 ) 
				  {
				    pptResults[0] = ppt[0];
				  }
				else if ( ppt[0](2) - fToolRadius < ppt[1](2) - fToolRadius ) 
				  {
				    pptResults[0] = ppt[0];
				  }
				else
				  {
				    pptResults[0] = ppt[1];
				  }
			      }
			    break;
			    case 1:
			      {
			      }
			    break;
			    case 2:
			      {
				if ( ( pe->GetFacet()->GetForwardVertex() != NULL &&
				       pe->GetEdge()->GetFacet()->GetForwardEdge() != NULL ) ||
				     ( pe->GetFacet()->GetForwardEdge() != NULL &&
				       pe->GetEdge()->GetFacet()->GetForwardVertex() != NULL ) ||
				     ( pe->GetFacet()->GetForwardVertex() != NULL &&
				       pe->GetEdge()->GetFacet()->GetForwardVertex() != NULL ) )
				  {
				    ppt[0] = pe->GetVertex()->GetPoint();
				    ppt[0][2] = 0.0;
				    ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
				    ppt[1][2] = 0.0;		    
				
				    if ( DistanceFromEdgeZone(pptResults[1], ppt[0], ppt[1]) < 
					 DistanceFromEdgeZone(pptResults[0], ppt[0], ppt[1]) )
				      {
					pptResults[0] = pptResults[1];
				      }
				  }
			      }
			    break;

			    default:
			      {
				ASSERT( FALSE );
			      }
			    break;
			    }
			
			  if ( pe->GetFacet()->GetForwardVertex() != NULL &&
			       pe->GetEdge()->GetFacet()->GetForwardEdge() != NULL &&
			       (pe->GetFacet()->GetForwardVertex() ==
				pe->GetEdge()->GetFacet()->GetForwardEdge()->GetVertex()) )
			    {
			      // Crossing from vertex zone to adjacent edge zone 
			    }
			  else if ( pe->GetFacet()->GetForwardEdge() != NULL &&
			       pe->GetEdge()->GetFacet()->GetForwardVertex() != NULL &&
			       (pe->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex() == 
				pe->GetEdge()->GetFacet()->GetForwardVertex()) )
			    {
			      // Crossing from edge zone to adjacent vertex zone 
			      pvxNew = CreateVertex(pe->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint());
			      peNew = CreateEdge();

			      peNew->GetEdge()->SetFacet(pfaOffset);
			      peOffset->SetVertex(pvxNew);
			      peNew->GetEdge()->SetVertex(pvxNew);
			      Splice(1, peOffset, peNew->GetEdge());
			      peOffset = peNew;
			    }
			  else 
			    {
			      if ( pe->GetFacet()->GetForwardEdge() != NULL )
				{			  
				  ppt[0] = pe->GetFacet()->GetForwardEdge()->GetVertex()->GetPoint();
				  ppt[1] = pe->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint();
				  pt = CalculatePointProjectionOnLine(pptResults[0], ppt[0], ppt[1], FALSE);

				  pvxNew = CreateVertex(pt);
				  peNew = CreateEdge();
			
				  peNew->GetEdge()->SetFacet(pfaOffset);
				  peOffset->SetVertex(pvxNew);
				  peNew->GetEdge()->SetVertex(pvxNew);
				  Splice(1, peOffset, peNew->GetEdge());
				  peOffset = peNew;
				}

			      // Create circular arc center which will be unattached
			      pvxNew = CreateVertex(pptResults[0]);
			      peNew = CreateEdge();
			      peNew->SetVertex(pvxNew);
			      peOffset->SetForwardEdge(peNew);
			      peOffset->GetEdge()->SetForwardEdge(peNew);

			      if ( pe->GetEdge()->GetFacet()->GetForwardVertex() != NULL )
				{			  
				  pt = pe->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint();
				}
			      else
				{			  
				  ppt[0] = pe->GetEdge()->GetFacet()->GetForwardEdge()->GetVertex()->GetPoint();
				  ppt[1] = pe->GetEdge()->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint();
				  pt = CalculatePointProjectionOnLine(pptResults[0], ppt[0], ppt[1], FALSE);
				}

			      pvxNew = CreateVertex(pt);
			      peNew = CreateEdge();
			
			      peNew->GetEdge()->SetFacet(pfaOffset);
			      peOffset->SetVertex(pvxNew);
			      peNew->GetEdge()->SetVertex(pvxNew);
			      Splice(1, peOffset, peNew->GetEdge());
			      peOffset = peNew;
			    }

			  pe = pe->GetEdge();
			}
		    }

		  if ( peOffset == peOffsetStart )
		    {
		      ASSERT( FALSE );
		      DestroyVertex(peOffsetStart->GetEdge()->GetVertex());
		      DestroyEdge(peOffsetStart->GetEdge());
		      DestroyEdge(peOffsetStart);
		    }
		  else
		    {
		      if ( bRemoveLastEdge )
			{
			  pe = peOffset;
			  peOffset = peOffset->GetEdgeNext(0);
			  Splice(1, peOffset, pe->GetEdge());
			  DestroyVertex(pe->GetEdge()->GetVertex());
			  DestroyEdge(pe->GetEdge());
			  DestroyEdge(pe);			  
			}
		      
		      if ( peOffset == peOffsetStart )
			{
			  ASSERT( FALSE );
			  DestroyVertex(peOffsetStart->GetEdge()->GetVertex());
			  DestroyEdge(peOffsetStart->GetEdge());
			  DestroyEdge(peOffsetStart);
			}

		      peOffset->SetVertex(peOffsetStart->GetEdge()->GetVertex());
		      Splice(1, peOffset, peOffsetStart->GetEdge());
		    }
		}
	    }
	}
    }

  //fprintf(stderr, "Finished computing Offset:\n");
  //DisplayEuler();
}

VOID QEMesh::DisplayMinChannelVertices()
{
  CHeap<VOID*, FLOAT> heapQEVertex;
  CHeapEntry<VOID*, FLOAT> *pCHeapEntry;
  QEMeshIter iterQEMesh;
  QEVertex *pvx;
  QEEdge *pe, *peCurr;
  QEEdge *ppeRising[2];
  FLOAT fZ;
  Point pt;
  UINT_32 uiRisingEdge, uiMinChannelVertices;

  heapQEVertex.Init();
  heapQEVertex.SetBiggestFirst(FALSE);
  iterQEMesh.Init(this);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      fZ = pvx->GetPoint()(2);
      if ( fZ > 0.0 )
	{
	  pCHeapEntry = heapQEVertex.Insert(fZ, (VOID*)pvx);
	  pvx->SetHeapEntry(pCHeapEntry);
	}
    }

  printf("START: Min Channel Vertices for Voronoi Mesh %p\n", this);
  uiMinChannelVertices = 0;
  for ( pvx = (QEVertex*)(heapQEVertex.RemoveTop());
	pvx != NULL;
	pvx = (QEVertex*)(heapQEVertex.RemoveTop()) )
    {
      pvx->SetHeapEntry(NULL);
      pt = pvx->GetPoint();
      fZ = pt[2];

      ppeRising[0] = NULL;
      ppeRising[1] = NULL;
      uiRisingEdge = 0;

      pe = pvx->GetEdge();
      peCurr = pe;
      do
	{
	  pt = peCurr->GetEdge()->GetVertex()->GetPoint();
	  if ( pt(2) == fZ )
	    {
	      pCHeapEntry = peCurr->GetEdge()->GetVertex()->GetHeapEntry();
	      if ( pCHeapEntry != NULL )
		{
		  heapQEVertex.Remove(pCHeapEntry);
		  peCurr->GetEdge()->GetVertex()->SetHeapEntry(NULL);
		}
	      peCurr = peCurr->GetEdge();
	    }
	  else if ( pt(2) > fZ )
	    {
	      if ( uiRisingEdge == 0 )
		{
		  ppeRising[uiRisingEdge] = peCurr;
		  uiRisingEdge++;
		}
	      else if ( uiRisingEdge == 1 )
		{
		  /*
		    // Don't consider adjacent rising edges
		  ASSERT( peCurr != ppeRising[0] );
		  if ( ( ppeRising[0]->GetEdgeNext(1) != peCurr &&
			 ppeRising[0] != peCurr->GetEdgeNext(1) ) ||
		       ( ppeRising[0]->GetEdgeNext(1) == peCurr &&
			 ppeRising[0] == peCurr->GetEdgeNext(1) ) )
		    {
		      ppeRising[uiRisingEdge] = peCurr;
		      uiRisingEdge++;
		      break;
		    }
		    */
		  ppeRising[uiRisingEdge] = peCurr;
		  uiRisingEdge++;
		  break;
		}
	      else
		{
		  ASSERT( FALSE );
		}
	    }

	    peCurr = peCurr->GetEdgeNext(1);
	}
      while ( peCurr != pe );

      
      if ( uiRisingEdge > 1 )
	{
	  uiMinChannelVertices++;
	  pt = pvx->GetPoint();
	  printf("  Min Channel Vertex %ld: R = %f: ( %.3f, %.3f, %.3f ): %p\n",
		 uiMinChannelVertices, pt[2], pt[0], pt[1], pt[2], pvx); 
	}
    }

  printf("END  : Min Channel Vertices for Voronoi Mesh %p\n", this);
}


VOID QEMesh::CreateTerrain(DelaunayType eType, CObjectList<QEVertex> *plistQEVertex)
{
  ASSERT( plistQEVertex != NULL );
  
  //CHashTable<CList<QEVertex*, QEVertex>, QEVertex*, QEVertex> hash;
  QEMeshIter iterQEMesh;
  CObjectListIter<QEVertex> iterQEVertex;
  //CObjectList<QEVertex> listQEVertex;
  //CObjectListIter<QEVertex> iterQEVertex;
  QEVertex *pvx, *pvxNew, *pvxForward;
  Point pt;

  //hash.Init();
  iterQEMesh.Init(this);
  iterQEVertex.Init(plistQEVertex);

  for ( pvx = iterQEVertex.PeekFirst();
	pvx != NULL;
	pvx = iterQEVertex.PeekNext() )
    {
      pt = pvx->GetPoint();
      pt[Z] = 0.0;

      pvxNew = CreateVertex(pt);
      ASSERT( pvxNew != NULL );
      pvxNew->SetID(pvx->GetID());

      pvxNew->SetForwardVertex(pvx);
/*
      pvxNew = hash.Peek(*pvx);
      if ( pvxNew != NULL )
	{
	  // This is a duplicate vertex
	  DestroyVertex(pvx);
	}

      //else
	//{
	  //pt = pCSLIDEPoint->GetPoint();
	  //pt[Z] = 0.0;
	  
	  //pvxNew = AllocateQEVertex(pt);
	  //ASSERT( pvxNew != NULL );
	  //listQEVertex.InsertLast(pvxNew);
	//}
*/
    }

  switch ( eType )
    {
    case dtIncremental:
      TriangulateDelaunayIncremental();
      break;
    case dtIncrementalSlow:
      TriangulateDelaunayIncrementalSlow();
      break;
    case dtIncrementalNoBound:
      TriangulateDelaunayIncrementalNoBound();
      break;
    case dtDivideAndConquer:
      TriangulateDelaunayDivideAndConquer();
      break;
    case dtDivideAndConquerAlternating:
      TriangulateDelaunayDivideAndConquerAlternating();
      break;
    case dtConvexHull:
      TriangulateDelaunayConvexHull();
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pvxForward = pvx->GetForwardVertex();
      if ( pvxForward != NULL )
	{
	  pt = pvxForward->GetPoint();

	  // The Parabaloid Lifting Map
	  //pt = pvx->GetPoint();
	  //pt[Z] = -(pt[X]*pt[X] + pt[Y]*pt[Y]);
	  
	  pvx->SetPoint(pt);
	  pvx->SetID(pvxForward->GetID());
	}
    }
  
  Preprocess();
}

VOID QEMesh::CreateDelaunay(DelaunayType eType, CObjectList<QEVertex> *plistQEVertex)
{
  ASSERT( plistQEVertex != NULL );
  
  ASSERT( m_plistQEVertex->GetLength() == 0 );
  m_plistQEVertex->Splice(plistQEVertex);

  switch ( eType )
    {
    case dtIncremental:
      TriangulateDelaunayIncremental();
      break;
    case dtIncrementalSlow:
      TriangulateDelaunayIncrementalSlow();
      break;
    case dtIncrementalNoBound:
      TriangulateDelaunayIncrementalNoBound();
      break;
    case dtDivideAndConquer:
      TriangulateDelaunayDivideAndConquer();
      break;
    case dtDivideAndConquerAlternating:
      TriangulateDelaunayDivideAndConquerAlternating();
      break;
    case dtConvexHull:
      TriangulateDelaunayConvexHull();
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  Preprocess();
}

VOID QEMesh::CreateVoronoi(VoronoiType eType, QEMesh *pmeDelaunay)
{
  ASSERT( pmeDelaunay != NULL );

  switch ( eType )
    {
    case vtDelaunay:
      CreateVoronoiDelaunay(pmeDelaunay);
      break;
    case vtDivideAndConquer:
      CreateVoronoiDivideAndConquer(pmeDelaunay);
      break;
    case vtDivideAndConquerAlternating:
      CreateVoronoiDivideAndConquerAlternating(pmeDelaunay);
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  Preprocess();
/*
  fprintf(stderr, "Finished computing Voronoi Diagram:\n");
  DisplayEuler();
*/
}

VOID QEMesh::CreateTetrahedron(const Point &pt0, const Point &pt1, const Point &pt2, 
			       const Point &pt3)
{
  QEFacet *pfa;
  QEEdge *pe, *peCurr;
  QEVertex *pvx;
  CObjectListIter<QEFacet> iterQEFacet;

  pfa = CreateFacet();
  ASSERT( pfa != NULL );

  pvx = CreateVertex(pt0);
  ASSERT( pvx != NULL );
  pe = CreateContour(pfa, pvx);
  ASSERT( pe != NULL );
  
  pfa = CreateFacet();
  ASSERT( pfa != NULL );
  pe->GetEdge()->SetFacet(pfa);

  pvx = CreateVertex(pt1);
  ASSERT( pvx != NULL );
  peCurr = AddContourVertex(pe, pvx);
  ASSERT( peCurr != NULL );
  peCurr->GetEdge()->SetFacet(pfa);

  pvx = CreateVertex(pt2);
  ASSERT( pvx != NULL );
  peCurr = AddContourVertex(pe, pvx);
  ASSERT( peCurr != NULL );
  peCurr->GetEdge()->SetFacet(pfa);

  pvx = CreateVertex(pt3);
  ASSERT( pvx != NULL );
  InsertVertex(pfa, pvx);

  iterQEFacet.Init(m_plistQEFacet);
  for ( pfa = iterQEFacet.PeekFirst();
	pfa != NULL;
	pfa = iterQEFacet.PeekNext() )
    {
      pfa->Preprocess();
    }
}

VOID QEMesh::CreateTetrahedron(QEVertex *pvx0, QEVertex *pvx1, 
			       QEVertex *pvx2, QEVertex *pvx3)
{
  QEFacet *pfa;
  QEEdge *pe, *peCurr;
  CObjectListIter<QEFacet> iterQEFacet;

  pfa = CreateFacet();
  ASSERT( pfa != NULL );

  pe = CreateContour(pfa, pvx0);
  ASSERT( pe != NULL );
  
  pfa = CreateFacet();
  ASSERT( pfa != NULL );
  pe->GetEdge()->SetFacet(pfa);

  peCurr = AddContourVertex(pe, pvx1);
  ASSERT( peCurr != NULL );
  peCurr->GetEdge()->SetFacet(pfa);

  peCurr = AddContourVertex(pe, pvx2);
  ASSERT( peCurr != NULL );
  peCurr->GetEdge()->SetFacet(pfa);

  InsertVertex(pfa, pvx3);

  iterQEFacet.Init(m_plistQEFacet);
  for ( pfa = iterQEFacet.PeekFirst();
	pfa != NULL;
	pfa = iterQEFacet.PeekNext() )
    {
      pfa->Preprocess();
    }
}

// This uses the QuickHull algorithm for computing the convex hull of
// a point set ( URL == http://www.geom.umn.edu/software/qhull/ ).
//
// Return value is the Dimension of the Convex Hull (3, 2, 1, or 0)
UINT_32 QEMesh::CreateConvexHull(CObjectList<QEVertex> *plistQEVertex)
{
  ASSERT( plistQEVertex != NULL );

  UINT_32 i;
  QEVertex *ppvxSimplex[4], *pvx;
  //QEVertex *pvxNew;
  CObjectList<QEVertex> listQEVertexUnassigned;
  CObjectListIter<QEVertex> iterQEVertex;
  Vector vL, vV, vNormal;
  //Point pt;
  FLOAT fDistance, fDistanceMax;

  //////////////////////////////////////////////////
  // Copy the vertices
  /*
  iterQEVertex.Init(plistQEVertex);
  for ( pvx = iterQEVertex.PeekFirst();
	pvx != NULL;
	pvx = iterQEVertex.PeekNext() )
    {
      //pvx->Display();

      pvxNew = AllocateQEVertex();
      ASSERT( pvxNew != NULL );
      pvxNew->Init(pvx->GetPoint());
      pvxNew->SetForwardVertex(pvx->GetForwardVertex());
      pvxNew->SetID(pvx->GetID());

      listQEVertexUnassigned.InsertLast(pvxNew);
    }
    */
  listQEVertexUnassigned.Splice(plistQEVertex);

  //////////////////////////////////////////////////
  // Create an initial simplex (tetrahedron)

  for ( i = 0; i < 4; i++ )
    {
      ppvxSimplex[i] = NULL;
    }

  QEVertex::SetComparisonType(QEVertex::ctXYZ);

  iterQEVertex.Init(&listQEVertexUnassigned);
  pvx = iterQEVertex.PeekFirst();
  ppvxSimplex[0] = pvx;
  ppvxSimplex[1] = pvx;

  if ( ppvxSimplex[0] == NULL )
    {
      // There are no points so it is dimension -1D
      ASSERT( FALSE );
      return 0;
    }

  for ( pvx = iterQEVertex.PeekNext();
	pvx != NULL;
	pvx = iterQEVertex.PeekNext() )
    {
      if ( (*pvx) < (*(ppvxSimplex[0])) )
	{
	  ppvxSimplex[0] = pvx;
	}
      if ( (*pvx) > (*(ppvxSimplex[1])) )
	{
	  ppvxSimplex[1] = pvx;
	}
    }
  QEVertex::SetComparisonTypeDefault();

  if ( ppvxSimplex[0] == ppvxSimplex[1] || 
       ppvxSimplex[0]->GetPoint()(X) == ppvxSimplex[1]->GetPoint()(X) )
    {
      // Dimension is at most 2D
      ASSERT( FALSE );
      // We would still need to construct the convex hull
      return 2;
    }

  // Find a Third Point which is furthest from the line between the first two
  vL = ppvxSimplex[1]->GetPoint() - ppvxSimplex[0]->GetPoint();
  VERIFY( vL.Normalize() );

  fDistanceMax = 0.0;
  ppvxSimplex[2] = NULL;

  for ( pvx = iterQEVertex.PeekFirst();
	pvx != NULL;
	pvx = iterQEVertex.PeekNext() )
    {
      if ( ppvxSimplex[0] != pvx && ppvxSimplex[1] != pvx )
	{
	  vV = pvx->GetPoint() - ppvxSimplex[0]->GetPoint();
	  vNormal = vV * vL;
	  fDistance = vNormal.Magnitude();

	  if ( fDistance > fDistanceMax )
	    {
	      fDistanceMax = fDistance;
	      ppvxSimplex[2] = pvx;
	    }
	}
    }

  if ( ppvxSimplex[2] == NULL )
    {
      // Our Point set is 1D along the line from Left to Right
      ASSERT( FALSE );

      return 1;
    }

  // Find a fourth point which is furthest from the plane of the triangle
  vV = ppvxSimplex[2]->GetPoint() - ppvxSimplex[0]->GetPoint();
  vNormal = vL * vV;
  VERIFY( vNormal.Normalize() );
  vNormal.CompletePlaneEquation(ppvxSimplex[0]->GetPoint());

  ppvxSimplex[3] = NULL;
  fDistanceMax = 0.0;

  for ( pvx = iterQEVertex.PeekFirst();
	pvx != NULL;
	pvx = iterQEVertex.PeekNext() )
    {
      if ( ppvxSimplex[0] != pvx && 
	   ppvxSimplex[1] != pvx &&
	   ppvxSimplex[2] != pvx )
	{
	  vV = pvx->GetPoint();
	  vV[W] = 1.0;
	  fDistance = vNormal.Dot(vV);

	  if ( fabs(fDistance) > fabs(fDistanceMax) )
	    {
	      fDistanceMax = fDistance;
	      ppvxSimplex[3] = pvx;
	    }
	}
    }

  if ( ppvxSimplex[3] == NULL )
    {
      // Our Point set is 2D in the plane we have defined
      ASSERT( FALSE );

      return 2;
    }

  if ( fDistanceMax > 0.0 )
    {
      // Our Triangle is oriented incorrectly because the fourth
      // vertex of our tetrahedron was above it.
      pvx = ppvxSimplex[1];
      ppvxSimplex[1] = ppvxSimplex[2];
      ppvxSimplex[2] = pvx;
    }

  // Create the tetrahedron
  for ( i = 0; i < 4; i++ )
    {
      listQEVertexUnassigned.Remove(ppvxSimplex[i]);
      AddVertex(ppvxSimplex[i]);
      //fprintf(stderr, "Tetra %ld: ", i);
      //ppvxSimplex[i]->Display();
    }
  CreateTetrahedron(ppvxSimplex[0], ppvxSimplex[1], ppvxSimplex[2], ppvxSimplex[3]);

  ////////////////////////////////////////////////////////////
  // The real algorithm

  CHeap<QEFacet*, FLOAT> heapQEFacet;
  CObjectList<QEFacet> listQEFacet;
  CObjectList<QEEdge> listQEEdge;
  //CObjectListIter<QEFacet> iterQEFacet;
  QEVertexIter vxIter;
  QEFacet *pfaEye, *pfa;
  QEVertex *pvxEye;
  QEEdge *peStart, *pe, *peOpp, *peCurr, *peHorizon;

  heapQEFacet.Init();

  //////////////////////////////////////////////////
  // Initial Partitioning of Point Set to Outside Sets
  // Points should go to the face which they are furthest outside of 

  listQEFacet.Splice(m_plistQEFacet);
  PartitionVertices(&heapQEFacet, &listQEFacet, &listQEVertexUnassigned);
  m_plistQEFacet->Splice(&listQEFacet);

  ASSERT( listQEVertexUnassigned.GetLength() == 0 );

  //////////////////////////////////////////////////
  // Do the Main Loop

  for ( pfaEye = heapQEFacet.PeekTop();
	pfaEye != NULL;
	pfaEye = heapQEFacet.PeekTop() )
    {
      ASSERT( listQEFacet.GetLength() == 0 );
      ASSERT( listQEEdge.GetLength() == 0 );
      ASSERT( listQEVertexUnassigned.GetLength() == 0 );

      ASSERT( pfaEye->GetPrevLink() == NULL );
      ASSERT( pfaEye->GetNextLink() == NULL );
      ASSERT( pfaEye->GetPriority() != NULL );

      pvxEye = pfaEye->RemoveFurthestOutsideVertex();
      ASSERT( pvxEye != NULL );
      //fprintf(stderr, "Eye: ");
      //pvxEye->Display();
      
      //////////////////////////////////////////////////
      // Initialize the Visible Set of Faces
      // Find the Horizon Edges
      // Recover all the Vertices in the outside sets of the 
      // visible Faces
      
      peStart = pfaEye->GetEdge();
      CalculateHorizon(peStart, pvxEye, 
		       &heapQEFacet, &listQEVertexUnassigned, 
		       &listQEEdge, TRUE);
      /*
      if ( peStart->GetEdge()->GetFacet() == m_pfaNULL )
	{
	  // We must remove the edge and possibly the two endpoints
	  RemoveEdge(peStart);
	  listQEEdge.InsertLast(peStart);
	}
	*/
      for ( pe = listQEEdge.RemoveFirst();
	    pe != NULL;
	    pe = listQEEdge.RemoveFirst() )
	{
	  peOpp = pe->GetEdge();
	  listQEEdge.Remove(peOpp);

	  ASSERT( pe->GetFacet() == m_pfaNULL );
	  if ( pe->GetEdgeNext(1) == pe )
	    {
	      DestroyVertex(pe->GetVertex());
	    }
	  else
	    {
	      Splice(1, pe->GetEdgePrev(1), pe);
	    }

	  ASSERT( peOpp->GetFacet() == m_pfaNULL );
	  if ( peOpp->GetEdgeNext(1) == peOpp )
	    {
	      DestroyVertex(peOpp->GetVertex());
	    }
	  else
	    {
	      Splice(1, peOpp->GetEdgePrev(1), peOpp);
	    }

	  DeallocateQEEdge(pe);
	  DeallocateQEEdge(peOpp);
	}

      peHorizon = m_pfaNULL->GetEdge();
      ASSERT( peHorizon != NULL );

      ////////////////////////////////////////////////////////////
      // Create a cone from the eye point vertex to the horizon

      pfa = CreateFacet();
      ASSERT( pfa != NULL );

      pe = m_pfaNULL->GetEdge();
      ASSERT( pe != NULL );

      peCurr = pe;
      do
	{
	  peCurr->SetFacet(pfa);
	  
	  peCurr = peCurr->GetEdgeNext(0);
	}
      while ( peCurr != pe );

      AddVertex(pvxEye);
      InsertVertex(pfa, pvxEye);

      ////////////////////////////////////////////////////////////
      // Partition the outside vertices
      vxIter.Init(pvxEye);
      for ( pfa = vxIter.PeekFirstFacet();
	    pfa != NULL;
	    pfa = vxIter.PeekNextFacet() )
	{
	  pfa->Preprocess();
	  m_plistQEFacet->Remove(pfa);
	  listQEFacet.InsertLast(pfa);
	}

      PartitionVertices(&heapQEFacet, &listQEFacet, &listQEVertexUnassigned);
      m_plistQEFacet->Splice(&listQEFacet);

      ASSERT( listQEFacet.GetLength() == 0 );
      ASSERT( listQEEdge.GetLength() == 0 );
      ASSERT( listQEVertexUnassigned.GetLength() == 0 );
    }

  ASSERT( heapQEFacet.GetSize() == 0 );
  ASSERT( listQEFacet.GetLength() == 0 );
  ASSERT( listQEEdge.GetLength() == 0 );
  ASSERT( listQEVertexUnassigned.GetLength() == 0 );

  return 3;
}

VOID QEMesh::PartitionVertices(CHeap<QEFacet*, FLOAT> *pheapQEFacet,
			       CObjectList<QEFacet> *plistQEFacet,
			       CObjectList<QEVertex> *plistQEVertex)
			       
{
  ASSERT( pheapQEFacet != NULL );
  ASSERT( plistQEFacet != NULL );
  ASSERT( plistQEVertex != NULL );

  CObjectListIter<QEFacet> iterQEFacet;
  CHeapEntry<QEFacet*, FLOAT> *pCHeapEntry;
  QEVertex *pvx;
  QEFacet *pfa, *pfaMax, *pfaPrev;
  FLOAT fDistanceMax, fDistance;

  iterQEFacet.Init(plistQEFacet);
  for ( pvx = plistQEVertex->RemoveFirst();
	pvx != NULL;
	pvx = plistQEVertex->RemoveFirst() )
    {
      pfaMax = iterQEFacet.PeekFirst();
      ASSERT( pfaMax != NULL );

      fDistanceMax = pfaMax->CalculateDistance(pvx->GetPoint());

      for ( pfa = iterQEFacet.PeekNext();
	    pfa != NULL;
	    pfa = iterQEFacet.PeekNext() )
	{
	  fDistance = pfa->CalculateDistance(pvx->GetPoint());
	  if ( fDistance > fDistanceMax )
	    {
	      pfaMax = pfa;
	      fDistanceMax = fDistance;
	    }
	}

      if ( fDistanceMax > 0.0 /*&& !pfaMax->CheckCoplanar(pvx)*/ )
	{
	  pfaMax->AddOutsideVertex(fDistanceMax, pvx);
	}
      else
	{
	  DeallocateQEVertex(pvx);
	}
    }

  // Add the faces which have outside vertices to the priority queue of 
  // unprocessed faces
  for ( pfa = iterQEFacet.PeekFirst();
	pfa != NULL; )
    {
      pvx = pfa->PeekFurthestOutsideVertex(&fDistance);
      if ( pvx != NULL )
	{
	  pCHeapEntry = pfa->GetPriority();
	  if ( pCHeapEntry != NULL )
	    {
	      if ( pCHeapEntry->GetKey() < fDistance )
		{
		  pheapQEFacet->Update(pCHeapEntry, fDistance);
		}
	    }
	  else
	    {
	      pCHeapEntry = pheapQEFacet->Insert(fDistance, pfa);
	      ASSERT( pCHeapEntry != NULL );
	      pfa->SetPriority(pCHeapEntry);
	    }

	  
	  pfaPrev = pfa;
	  pfa = iterQEFacet.PeekNext();
	  plistQEFacet->Remove(pfaPrev);
	}
      else
	{
	  ASSERT( pfa->GetPriority() == NULL );
	  pfa = iterQEFacet.PeekNext();
	}
    }  
}

VOID QEMesh::CalculateHorizon(QEEdge *peContour, QEVertex *pvxEye, 
			      CHeap<QEFacet*, FLOAT> *pheapQEFacet, 
			      CObjectList<QEVertex> *plistQEVertexUnassigned, 
			      CObjectList<QEEdge> *plistQEEdgeRemoved,
			      BOOL bStart)
{
  ASSERT( peContour != NULL );
  ASSERT( pvxEye != NULL );
  ASSERT( pheapQEFacet != NULL );
  ASSERT( plistQEVertexUnassigned != NULL );
  ASSERT( plistQEEdgeRemoved != NULL );

  QEFacet *pfa;
  QEEdge *pe, *pePrev;
  QEVertex *pvx;
  CHeapEntry<QEFacet*, FLOAT> *pCHeapEntry;

  pfa = peContour->GetFacet();
  ASSERT( pfa != NULL && pfa != m_pfaNULL );

  if ( !pfa->CheckOutside(pvxEye) )
    {
      ASSERT( !bStart );
      // This face is not visible so we are at a leaf of our depth first search
      ASSERT( peContour->GetEdge()->GetFacet() == m_pfaNULL );
      // Do this so that m_pfaNULL will point to one of the horizon edges in the end
      peContour->GetEdge()->SetFacet(m_pfaNULL);

      return;
    }

  // The facet is visible by the eye point so we will need to remove it
  
  // Set its edges' facet pointers to m_pfaNULL
  pe = peContour;
  do
    {
      ASSERT( pe->GetFacet() == pfa );
      pe->SetFacet(m_pfaNULL);

      pe = pe->GetEdgeNext(0);
    }
  while ( pe != peContour );

  // Release the outside set of vertices
  for ( pvx = pfa->RemoveOutsideVertex();
	pvx != NULL;
	pvx = pfa->RemoveOutsideVertex() )
    {
      plistQEVertexUnassigned->InsertLast(pvx);
    }
  
  pCHeapEntry = pfa->GetPriority();
  if ( pCHeapEntry != NULL )
    {
      ASSERT( pfa->GetPrevLink() == NULL );
      ASSERT( pfa->GetNextLink() == NULL );
      pheapQEFacet->Remove(pCHeapEntry);
      pfa->SetPriority(NULL);
    }
  else
    {
      ASSERT( pfa->GetPrevLink() != NULL );
      ASSERT( pfa->GetNextLink() != NULL );
      RemoveFacet(pfa);
    }

  DeallocateQEFacet(pfa);

  if ( bStart )
    {
      pe = peContour;
    }
  else
    {
      ASSERT( peContour->GetEdge()->GetFacet() == m_pfaNULL );
      // We must remove the edge and possibly the two endpoints
      RemoveEdge(peContour);
      plistQEEdgeRemoved->InsertLast(peContour);

      pe = peContour->GetEdgeNext(0);
    }
  do
    {
      if ( pe->GetEdge()->GetFacet() != m_pfaNULL )
	{
	  CalculateHorizon(pe->GetEdge(), pvxEye, pheapQEFacet, 
			   plistQEVertexUnassigned, plistQEEdgeRemoved, FALSE);
	}

      pePrev = pe;
      pe = pe->GetEdgeNext(0);
      if ( pePrev->GetEdge()->GetFacet() == m_pfaNULL )
	{
	  // We must remove the edge and possibly the two endpoints
	  RemoveEdge(pePrev);
	  plistQEEdgeRemoved->InsertLast(pePrev);
	}
    }
  while ( pe != peContour );
}

QEMesh *QEMesh::CreateOffsetMeshRounded(FLOAT fOffset)
{
  ASSERT( m_plistQEEdgeBoundary->GetLength() == 0 );

  QEMeshIter iterQEMesh;
  QEMesh *pme;
  QEVertex *pvx, *pvxNew;
  QEVertexIter iterQEVertex;
  QEFacet *pfa, *pfaNew;
  QEEdge *pe, *peNew, *peCurr, *peNewCurr, *peNewPrev;
  QEVertex *ppvx[6];
  QEFacet *ppfa[2];
  Vector pv[2], vEdge, v;
  Point ppt[2], pt;
  FLOAT fWeight;

  if ( fOffset == 0.0 )
    {
      return Copy();
    }

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  ClearForwardPointers();

  iterQEMesh.Init(this);

  // Offset Facets
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      pfaNew = pme->CreateFacet();

      pfa->SetForwardFacet(pfaNew);
    }

  // Offset Edges
  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      if ( pe->GetForwardEdge() == NULL )
	{
	  peNew = pme->CreateEdge();

	  pe->SetForwardEdge(peNew);
	  pe->GetEdge()->SetForwardEdge(peNew->GetEdge());

	  // Create the vertices
	  ppfa[0] = pe->GetFacet();
	  ppfa[1] = pe->GetEdge()->GetFacet();

	  pv[0] = ppfa[0]->GetNormal();
	  pv[0][3] = 0.0;
	  pv[1] = ppfa[1]->GetNormal();
	  pv[1][3] = 0.0;

	  vEdge = pv[0] + pv[1];
	  vEdge.Normalize();

	  ppt[0] = pe->GetVertex()->GetPoint();
	  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();

	  pfaNew = pme->CreateFacet();

	  ////////////////////////////////////////
	  // v0
	  pt = ppt[0] + (fOffset * vEdge);
	  ppvx[0] = pme->CreateVertex(pt);

	  peNew->SetVertex(ppvx[0]);
	  peNew->SetFacet(pfaNew);

	  peNewPrev = peNew;

	  ////////////////////////////////////////
	  // v1
	  peNewCurr = pme->CreateEdge();
	  peNewCurr->GetEdge()->SetVertex(ppvx[0]);
	  peNewCurr->SetFacet(pfaNew);
	  pme->Splice(1, peNewPrev, peNewCurr->GetEdge());

	  peCurr = pe->GetEdgePrev(0)->GetForwardEdge();
	  if ( peCurr != NULL )
	    {
	      ppvx[1] = peCurr->GetEdgeNext(0)->GetEdge()->GetVertex();
	      pme->Splice(1, peNewCurr, peCurr->GetEdgeNext(0)->GetEdge());
	    }
	  else
	    {
	      pt = ppt[0] + (fOffset * pv[0]);
	      ppvx[1] = pme->CreateVertex(pt);
	    }
	  peNewCurr->SetVertex(ppvx[1]);

	  peNewPrev = peNewCurr;

	  ////////////////////////////////////////
	  // v2
	  peNewCurr = pme->CreateEdge();
	  peNewCurr->GetEdge()->SetVertex(ppvx[1]);
	  peNewCurr->SetFacet(pfaNew);
	  peNewCurr->GetEdge()->SetFacet(pe->GetFacet()->GetForwardFacet());
	  pme->Splice(1, peNewPrev, peNewCurr->GetEdge());

	  peCurr = pe->GetEdgeNext(0)->GetForwardEdge();
	  if ( peCurr != NULL )
	    {
	      ppvx[2] = peCurr->GetEdgePrev(0)->GetVertex();
	      pme->Splice(1, peNewCurr, peCurr->GetEdgeNext(0)->GetEdgeNext(0)->GetEdge());
	    }
	  else
	    {
	      pt = ppt[1] + (fOffset * pv[0]);
	      ppvx[2] = pme->CreateVertex(pt);
	    }
	  peNewCurr->SetVertex(ppvx[2]);

	  peNewPrev = peNewCurr;

	  ////////////////////////////////////////
	  // v3
	  pt = ppt[1] + (fOffset * vEdge);
	  ppvx[3] = pme->CreateVertex(pt);
	  peNew->GetEdge()->SetVertex(ppvx[3]);

	  peNewCurr = pme->CreateEdge();
	  peNewCurr->GetEdge()->SetVertex(ppvx[2]);
	  peNewCurr->SetVertex(ppvx[3]);
	  peNewCurr->SetFacet(pfaNew);
	  pme->Splice(1, peNewPrev, peNewCurr->GetEdge());
	  pme->Splice(1, peNewCurr, peNew->GetEdge());
	  
	  pfaNew = pme->CreateFacet();
	  peNew->GetEdge()->SetFacet(pfaNew);
	  peNewPrev = peNew->GetEdge();

	  ////////////////////////////////////////
	  // v4
	  peNewCurr = pme->CreateEdge();
	  peNewCurr->GetEdge()->SetVertex(ppvx[3]);
	  peNewCurr->SetFacet(pfaNew);
	  pme->Splice(1, peNewPrev, peNewCurr->GetEdge());

	  peCurr = pe->GetEdge()->GetEdgePrev(0)->GetForwardEdge();
	  if ( peCurr != NULL )
	    {
	      ppvx[4] = peCurr->GetEdgeNext(0)->GetEdge()->GetVertex();
	      pme->Splice(1, peNewCurr, peCurr->GetEdgeNext(0)->GetEdge());
	    }
	  else
	    {
	      pt = ppt[1] + (fOffset * pv[1]);
	      ppvx[4] = pme->CreateVertex(pt);
	    }
	  peNewCurr->SetVertex(ppvx[4]);

	  peNewPrev = peNewCurr;

	  ////////////////////////////////////////
	  // v5
	  peNewCurr = pme->CreateEdge();
	  peNewCurr->GetEdge()->SetVertex(ppvx[4]);
	  peNewCurr->SetFacet(pfaNew);
	  peNewCurr->GetEdge()->SetFacet(pe->GetEdge()->GetFacet()->GetForwardFacet());
	  pme->Splice(1, peNewPrev, peNewCurr->GetEdge());

	  peCurr = pe->GetEdge()->GetEdgeNext(0)->GetForwardEdge();
	  if ( peCurr != NULL )
	    {
	      ppvx[5] = peCurr->GetEdgePrev(0)->GetVertex();
	      pme->Splice(1, peNewCurr, peCurr->GetEdgeNext(0)->GetEdgeNext(0)->GetEdge());
	    }
	  else
	    {
	      pt = ppt[0] + (fOffset * pv[1]);
	      ppvx[5] = pme->CreateVertex(pt);
	    }
	  peNewCurr->SetVertex(ppvx[5]);

	  peNewPrev = peNewCurr;

	  ////////////////////////////////////////
	  // Final Edge
	  peNewCurr = pme->CreateEdge();
	  peNewCurr->GetEdge()->SetVertex(ppvx[5]);
	  peNewCurr->SetVertex(ppvx[0]);
	  peNewCurr->SetFacet(pfaNew);
	  pme->Splice(1, peNewPrev, peNewCurr->GetEdge());
	  pme->Splice(1, peNewCurr, peNew->GetEdgeNext(1));
	}
    }

  // Offset Vertices
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pfaNew = pme->CreateFacet();

      iterQEVertex.Init(pvx);

      v.Null();
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  peCurr = pe->GetForwardEdge();
	  ASSERT( peCurr->GetEdgePrev(1)->GetEdge()->GetFacet() == m_pfaNULL );
	  peCurr->GetEdgePrev(1)->GetEdge()->SetFacet(pfaNew);
	  ASSERT( peCurr->GetEdgeNext(1)->GetFacet() == m_pfaNULL );
	  peCurr->GetEdgeNext(1)->SetFacet(pfaNew);

	  pv[0] = pe->GetEdge()->GetVertex()->GetPoint() - pe->GetVertex()->GetPoint();
	  pv[0].Normalize();
	  pv[1] = pe->GetEdgePrev(0)->GetVertex()->GetPoint() - pe->GetVertex()->GetPoint();
	  pv[1].Normalize();
	  fWeight = pv[0].Dot(pv[1]);
	  fWeight = acos(fWeight);
	  
	  v += (fWeight * pe->GetFacet()->GetNormal());
	}
      v[3] = 0.0;
      v.Normalize();

      pt = pvx->GetPoint() + (fOffset * v);
      pvxNew = pme->CreateVertex(pt);

      pvx->SetForwardVertex(pvxNew);

      pme->InsertVertex(pfaNew, pvxNew);
    }

  pme->Preprocess();

  return pme;
}

//This is the function that I hope will be able to create the grid like 
//structures from the polygonal mesh, now upgraded for open edges.
QEMesh* QEMesh::CreateOffsetMeshAngular(FLOAT fOffset)
{
  QEMeshIter iterQEMesh;
  QEMesh *pme;
  QEVertexIter iterQEVertex;
  QEVertex *pvx, *pvxNew;
  QEFacet *pfa, *pfaNew;
  QEEdge *pe, *peNew, *peCurr;
  Vector v, pv[2];
  Point pt;
  FLOAT fWeight;

  if ( fOffset == 0.0 )
    {
      return Copy();
    }

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  iterQEMesh.Init(this);

  //////////////////////////////////////////////////////////////////////
  // Offset Positively

  // Clear Edge Forward Pointers
  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      pe->SetForwardEdge(NULL);
    }
  // Clear Boundary Edge Forward Pointers
  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdgeBoundary() )
    {
      pe->SetForwardEdge(NULL);
    }

  // Copy Vertices
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      iterQEVertex.Init(pvx);

      v.Null();
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  pv[0] = pe->GetEdge()->GetVertex()->GetPoint() - pe->GetVertex()->GetPoint();
	  pv[0].Normalize();
	  pv[1] = pe->GetEdgePrev(0)->GetVertex()->GetPoint() - pe->GetVertex()->GetPoint();
	  pv[1].Normalize();
	  fWeight = pv[0].Dot(pv[1]);
	  fWeight = acos(fWeight);
	  
	  v += (fWeight * pe->GetFacet()->GetNormal());
	}
      v[3] = 0.0;
      v.Normalize();

      pt = pvx->GetPoint() + (fOffset * v);

      pvxNew = pme->CreateVertex(pt);
      pvxNew->SetNormal(pvx->GetNormal());
      pvxNew->SetTexture(pvx->GetTexture());
      pvx->SetForwardVertex(pvxNew);
    }

  // Copy Facets
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      pfaNew = pme->CreateFacet();

      pfa->SetForwardFacet(pfaNew);
    }

  // Copy Edges
  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      if ( pe->GetForwardEdge() == NULL )
	{
	  peNew = pme->CreateEdge();

	  peNew->SetForwardEdge(NULL);
	  peNew->GetEdge()->SetForwardEdge(NULL);

	  peNew->SetKnotInterval(pe->GetKnotInterval());
	  peNew->GetEdge()->SetKnotInterval(pe->GetEdge()->GetKnotInterval());

	  pe->SetForwardEdge(peNew);
	  pe->GetEdge()->SetForwardEdge(peNew->GetEdge());

	  if ( pe->GetFacet() == m_pfaNULL )
	    {
	    }
	  else
	    {
	      peNew->SetFacet(pe->GetFacet()->GetForwardFacet());
	    }

	  if ( pe->GetEdge()->GetFacet() == m_pfaNULL )
	    {
	    }
	  else
	    {
	      peNew->GetEdge()->SetFacet(pe->GetEdge()->GetFacet()->GetForwardFacet());
	    }

	  peNew->SetVertex(pe->GetVertex()->GetForwardVertex());
	  peNew->GetEdge()->SetVertex(pe->GetEdge()->GetVertex()->GetForwardVertex());

	  for ( peCurr = pe->GetEdgePrev(1);
		peCurr != pe;
		peCurr = peCurr->GetEdgePrev(1) )
	    {
	      if ( peCurr->GetForwardEdge() != NULL )
		{
		  Splice(1, peNew, peCurr->GetForwardEdge());
		  break;
		}
	    }

	  for ( peCurr = pe->GetEdge()->GetEdgePrev(1);
		peCurr != pe->GetEdge();
		peCurr = peCurr->GetEdgePrev(1) )
	    {
	      if ( peCurr->GetForwardEdge() != NULL )
		{
		  Splice(1, peNew->GetEdge(), peCurr->GetForwardEdge());
		  break;
		}
	    }
	}
    }

  //////////////////////////////////////////////////////////////////////
  // Offset Negatively

  // Copy Vertices
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      iterQEVertex.Init(pvx);

      v.Null();
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  pv[0] = pe->GetEdge()->GetVertex()->GetPoint() - pe->GetVertex()->GetPoint();
	  pv[0].Normalize();
	  pv[1] = pe->GetEdgePrev(0)->GetVertex()->GetPoint() - pe->GetVertex()->GetPoint();
	  pv[1].Normalize();
	  fWeight = pv[0].Dot(pv[1]);
	  fWeight = acos(fWeight);
	  
	  v += (fWeight * pe->GetFacet()->GetNormal());
	}
      v[3] = 0.0;
      v.Normalize();

      pt = pvx->GetPoint() + ((-fOffset) * v);

      pvxNew = pme->CreateVertex(pt);
      v = pvx->GetNormal();
      v *= -1.0;
      pvxNew->SetNormal(v);
      pvxNew->SetTexture(pvx->GetTexture());
      pvx->SetForwardVertex(pvxNew);
    }

  // Copy Facets
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      pfaNew = pme->CreateFacet();

      pfa->SetForwardFacet(pfaNew);
    }

  // Copy Edges
  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      ASSERT( pe->GetForwardEdge() != NULL &&
	      pe->GetEdge()->GetForwardEdge() != NULL );
      if ( pe->GetForwardEdge()->GetForwardEdge() == NULL )
	{
	  peNew = pme->CreateEdge();
	  peNew->SetKnotInterval(pe->GetKnotInterval());
	  peNew->GetEdge()->SetKnotInterval(pe->GetEdge()->GetKnotInterval());

	  pe->GetForwardEdge()->SetForwardEdge(peNew);
	  pe->GetEdge()->GetForwardEdge()->SetForwardEdge(peNew->GetEdge());

	  if ( pe->GetFacet() == m_pfaNULL )
	    {
	    }
	  else
	    {
	      // Swap orientation of the mesh
	      peNew->GetEdge()->SetFacet(pe->GetFacet()->GetForwardFacet());
	    }

	  if ( pe->GetEdge()->GetFacet() == m_pfaNULL )
	    {
	    }
	  else
	    {
	      // Swap orientation of the mesh
	      peNew->SetFacet(pe->GetEdge()->GetFacet()->GetForwardFacet());
	    }

	  peNew->SetVertex(pe->GetVertex()->GetForwardVertex());
	  peNew->GetEdge()->SetVertex(pe->GetEdge()->GetVertex()->GetForwardVertex());

	  for ( peCurr = pe->GetEdgeNext(1);
		peCurr != pe;
		peCurr = peCurr->GetEdgeNext(1) )
	    {
	      if ( peCurr->GetForwardEdge()->GetForwardEdge() != NULL )
		{
		  Splice(1, peNew, peCurr->GetForwardEdge()->GetForwardEdge());
		  break;
		}
	    }

	  for ( peCurr = pe->GetEdge()->GetEdgeNext(1);
		peCurr != pe->GetEdge();
		peCurr = peCurr->GetEdgeNext(1) )
	    {
	      if ( peCurr->GetForwardEdge()->GetForwardEdge() != NULL )
		{
		  Splice(1, peNew->GetEdge(), peCurr->GetForwardEdge()->GetForwardEdge());
		  break;
		}
	    }
	}
    }

  //////////////////////////////////////////////////////////////////////
  // Create Facets for the Boundary Edges

  // Clear Boundary Edge Forward Pointers
  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdgeBoundary() )
    {
      peCurr = pe->GetForwardEdge();
      ASSERT( peCurr != NULL );
      ASSERT( peCurr->GetForwardEdge() != NULL );

      if ( peCurr->GetEdgeNext(0) == pe->GetEdgeNext(0)->GetForwardEdge() )
	{
	  peNew = pme->CreateEdge();

	  peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	  peNew->GetEdge()->SetVertex(peCurr->GetForwardEdge()->GetEdge()->GetVertex());

	  Splice(1, peNew, peCurr->GetEdge()->GetEdgePrev(1));
	  Splice(1, peNew->GetEdge(), peCurr->GetForwardEdge()->GetEdge());
	}
      if ( peCurr->GetEdgePrev(0) == pe->GetEdgePrev(0)->GetForwardEdge() )
	{
	  peNew = pme->CreateEdge();

	  peNew->SetVertex(peCurr->GetVertex());
	  peNew->GetEdge()->SetVertex(peCurr->GetForwardEdge()->GetVertex());

	  Splice(1, peNew, peCurr);
	  Splice(1, peNew->GetEdge(), peCurr->GetForwardEdge()->GetEdgePrev(1));
	}

      pfaNew = pme->CreateFacet();

      peNew = peCurr;
      do
	{
	  peNew->SetFacet(pfaNew);
	  peNew = peNew->GetEdgeNext(0);
	}
      while ( peNew != peCurr );
    }

  pme->Preprocess();

  return pme;
}  

/*This is the function that I hope will be able to create the grid like 
  structures from the polygonal mesh, now upgraded for open edges.*/
QEMesh* QEMesh::CreateOffsetMeshPerforated(FLOAT fOffsetNormal, FLOAT fOffsetTangential)
{
  FLOAT w_offset, t_offset;

  w_offset = fOffsetTangential;
  t_offset = fOffsetNormal;

  QEMesh* new_mesh = AllocateQEMesh();
  ASSERT(new_mesh != NULL);
  new_mesh->Init();
  ClearForwardPointers();

  if ( w_offset == 0.0 || t_offset == 0.0 )
  //if ( t_offset == 0.0 )
    {
      return Copy();
    }

  //declare some variables...
  QEMeshIter iterQEMesh;
  iterQEMesh.Init(this);
  QEFacetIter iterQEFacet;
  QEVertexIter iterQEVertex;
  float fWeight, theta;
  Point p_hinge, offset_hinge;
  Vector pv[2], v, offset_v;
  QEVertex* in_vertex;
  QEVertex* new_v = NULL;  /*new vertex that gets created if old_edge
				   or old_edge->GetEdge() have no forwarding
				   pointers for vertices*/
 
  //iterate thru facets...
  QEFacet* cur_facet = NULL;
  QEFacet* new_facet = NULL;

  QEEdge* old_edge = NULL;
  QEEdge* new_edge = NULL;  /*new edge that gets created, if old edge has
			      no forwarding pointer*/
  QEEdge* new_edge_offset = NULL;  /*inwardly offset new_edge to get a grid
				       like structure*/

  for(cur_facet = iterQEMesh.PeekFirstFacet(); cur_facet != NULL; cur_facet = iterQEMesh.PeekNextFacet())
  {
    iterQEFacet.Init(cur_facet);
    QEEdge* current = NULL;
    QEEdge* first_edge = NULL;

    //we have to create a new facet corresponding to the old one...
    
    for(old_edge = iterQEFacet.PeekFirstEdge(); old_edge != NULL; old_edge = iterQEFacet.PeekNextEdge())
    {
      //check if there is a forward edge already.. if there is none, create
      //a new one taking care not to dupicate vertices. so check for forward
      //vertices...
      if(old_edge->GetForwardEdge() == NULL)
      {
	new_edge = new_mesh->CreateEdge();
	old_edge->SetForwardEdge(new_edge);
	old_edge->GetEdge()->SetForwardEdge(new_edge->GetEdge());

	//prevent duplication of vertices...
	if(old_edge->GetVertex()->GetForwardVertex() == NULL)
	{
	  /*just cut and pasted this from jordan's code for angular offset*/
	  v.Null();
	  iterQEVertex.Init(old_edge->GetVertex());
	  for (current = iterQEVertex.PeekFirstEdge(); current != NULL; current = iterQEVertex.PeekNextEdge() )
	  {
	    pv[0] = current->GetEdge()->GetVertex()->GetPoint() - current->GetVertex()->GetPoint();
	    pv[0].Normalize();
	    pv[1] = current->GetEdgePrev(0)->GetVertex()->GetPoint() - current->GetVertex()->GetPoint();
	    pv[1].Normalize();
	    fWeight = pv[0].Dot(pv[1]);
	    fWeight = acos(fWeight);
    	    
	    if(current->GetFacet() != NULL)
	      v += (fWeight * current->GetFacet()->GetNormal());
	  }
	  v[3] = 0.0;
	  v.Normalize();

	  new_v =
	    new_mesh->CreateVertex(old_edge->GetVertex()->GetPoint() +
				   (0.5*t_offset * v));
	  old_edge->GetVertex()->SetForwardVertex(new_v);
	}	  

	if(old_edge->GetEdge()->GetVertex()->GetForwardVertex() == NULL)
	{
	  /*just cut and pasted this from jordan's code for angular offset*/
	  v.Null();
	  iterQEVertex.Init(old_edge->GetEdge()->GetVertex());
	  for (current = iterQEVertex.PeekFirstEdge(); current != NULL; current = iterQEVertex.PeekNextEdge() )
	  {
	    pv[0] = current->GetEdge()->GetVertex()->GetPoint() - current->GetVertex()->GetPoint();
	    pv[0].Normalize();
	    pv[1] = current->GetEdgePrev(0)->GetVertex()->GetPoint() - current->GetVertex()->GetPoint();
	    pv[1].Normalize();
	    fWeight = pv[0].Dot(pv[1]);
	    fWeight = acos(fWeight);
    	    
	    if(current->GetFacet() != NULL)
	      v += (fWeight * current->GetFacet()->GetNormal());
	  }
	  v[3] = 0.0;
	  v.Normalize();


	  new_v =
	    new_mesh->CreateVertex(old_edge->GetEdge()->GetVertex()->GetPoint()+ (0.5*t_offset * v));
	  old_edge->GetEdge()->GetVertex()->SetForwardVertex(new_v);
	}	  
	
	//specify vertices for the new edge that was created...
	new_edge->SetVertex(old_edge->GetVertex()->GetForwardVertex());
	new_edge->GetEdge()->SetVertex(old_edge->GetEdge()->GetVertex()->GetForwardVertex());

	//we have to insert the forward edge of the current old_edge and 
	//old_edge->GetEdge() into the existing scenario...
	//printf("START SPLICING \n\n");
	
	for(current = old_edge->GetEdgePrev(1); current != old_edge; current = current->GetEdgePrev(1))
	{
	  if(current->GetForwardEdge() != NULL)
	  {
	    new_mesh->Splice(1, old_edge->GetForwardEdge(), current->GetForwardEdge());
	    //printf("Splicing...");
	    //old_edge->GetForwardEdge()->Display();
	    //printf(" and ");
	    //current->GetForwardEdge()->Display();
	    break;
	  }
	}
	

	for(current = old_edge->GetEdge()->GetEdgePrev(1); current != old_edge->GetEdge(); current = current->GetEdgePrev(1))
	{
	  if(current->GetForwardEdge() != NULL)
	  {
	    new_mesh->Splice(1, old_edge->GetEdge()->GetForwardEdge(), current->GetForwardEdge());
	    //printf("Splicing...");
	    //old_edge->GetEdge()->GetForwardEdge()->Display();
	    //printf(" and ");
	    //current->GetForwardEdge()->Display();
	    break;
	  }
	}
      }  //endif (old_edge...)
      else
      {
	new_edge = old_edge->GetForwardEdge();
      }      
    }

    //now we have copied the original mesh (except for the fact that all the
    //new edges have no facets). we now have to add angular bisectors at each
    //vertex of the new "mesh" and then connect them up corectly...
    
    //go thru each of the old edges...
    QEEdge* bisector = NULL;  //bisector edge...

    for(old_edge = iterQEFacet.PeekFirstEdge(); old_edge != NULL; old_edge = iterQEFacet.PeekNextEdge())
    {
      //get the angular bisector at this vertex displaced by the correct
      //offset distance...
      bisector = new_mesh->CreateOffsetMeshBisector(new_mesh, old_edge, w_offset);

      //splice to connect up...
      new_mesh->Splice(1, bisector, old_edge->GetForwardEdge());
    }

    //now connect the bisectors to get the final configuration. remember to
    //assign new facets to the newly created edges...
    
    //iterate thru old edges...
    first_edge =  iterQEFacet.PeekFirstEdge();
    for(old_edge = first_edge; old_edge != NULL; old_edge = iterQEFacet.PeekNextEdge())
    {
      new_edge = new_mesh->CreateEdge();

      //vertex of the connecting edge will be the "destination" of the bisector
      //edge...
      new_edge->SetVertex(old_edge->GetForwardEdge()->GetEdgeNext(1)->GetEdge()->GetVertex());
      //new_edge->GetEdge()->SetVertex(old_edge->GetForwardEdge()->GetEdgeNext(0)->GetEdge()->GetVertex());
      new_edge->GetEdge()->SetVertex(old_edge->GetEdgeNext(0)->GetForwardEdge()->GetEdgeNext(1)->GetEdge()->GetVertex());
      
      //splice correctly...
      new_mesh->Splice(1, new_edge, old_edge->GetForwardEdge()->GetEdgeNext(1)->GetEdge());

      //if this is not the last connector between the bisectors...
      if(old_edge != first_edge->GetEdgePrev(0))
	new_mesh->Splice(1, new_edge->GetEdge(), old_edge->GetEdgeNext(0)->GetForwardEdge()->GetEdgeNext(1)->GetEdge());
    }
    
    //splice the last connector correctly...
    new_mesh->Splice(1, new_edge->GetEdge(), first_edge->GetForwardEdge()->GetEdgePrev(0)->GetEdgeNext(1));

    //some variables...
    //QEFacetIter cur_iter;

    //set facets to all new edges correctly- iterate thru old edges...
    for(old_edge = iterQEFacet.PeekFirstEdge(); old_edge != NULL; old_edge = iterQEFacet.PeekNextEdge())
    {
      new_edge = old_edge->GetForwardEdge();
      
      //create a new facet...
      new_facet = new_mesh->CreateFacet();
     
      //set facets correctly...
      do
      {
	new_edge->SetFacet(new_facet);
	new_edge = new_edge->GetEdgeNext(0);
      } while(new_edge != old_edge->GetForwardEdge());
    }

    //now we have cut out a smaller facet from each original facet. after this
    //we have to thicken this new facet...
    for(old_edge = iterQEFacet.PeekFirstEdge(); old_edge != NULL; old_edge = iterQEFacet.PeekNextEdge())
    {
      //check if the "new_edge" corresponding to this old edge has a forward
      //pointer. if not, create one so that it can represent the offset edge
      //corresponding to the new edge- this offset is inwards...
      if(old_edge->GetForwardEdge()->GetForwardEdge() == NULL)
      {
	new_edge = old_edge->GetForwardEdge();
	//new_edge_offset = new_mesh->CreateEdge();

	//remember to set the correct vertices for new_edge_offset later...
	//new_edge->SetForwardEdge(new_edge_offset);
	//new_edge->GetEdge()->SetForwardEdge(new_edge_offset->GetEdge());
	  
	//check for forward pointers for vertices in new_edge...
	if(new_edge->GetVertex()->GetForwardVertex() == NULL)
	{
	  //calculate the new vertex. this is got by inwardly offsetting the
	  //old vertex by the average normal at this vertex...

	  /*just cut and pasted this from jordan's code for angular offset*/
	  v.Null();
	  iterQEVertex.Init(old_edge->GetVertex());
	  for (current = iterQEVertex.PeekFirstEdge(); current != NULL; current = iterQEVertex.PeekNextEdge() )
	  {
	    pv[0] = current->GetEdge()->GetVertex()->GetPoint() - current->GetVertex()->GetPoint();
	    pv[0].Normalize();
	    pv[1] = current->GetEdgePrev(0)->GetVertex()->GetPoint() - current->GetVertex()->GetPoint();
	    pv[1].Normalize();
	    fWeight = pv[0].Dot(pv[1]);
	    fWeight = acos(fWeight);
    	    
	    //the '-' sign is because we are moving inwards...
	    if(current->GetFacet() != NULL)
	      v += (-fWeight * current->GetFacet()->GetNormal());
	  }
	  v[3] = 0.0;
	  v.Normalize();
	  
	  new_v =
	    new_mesh->CreateVertex(old_edge->GetVertex()->GetPoint() +
				   (0.5*t_offset * v));
	  new_edge->GetVertex()->SetForwardVertex(new_v);
	}
      }  //endif (old_edge->GetForward...)
    }  //endfor

    //now we have created inwardly offset vertices for the new_edges and also
    //created corresponding edges for the inside mesh. but the edges do not
    //yet have vertices assigned to them. below, additional facets will be
    //created to "extrude" the new facet...
    
    //create bisectors for the inner facet...
    for(old_edge = iterQEFacet.PeekFirstEdge(); old_edge != NULL; old_edge = iterQEFacet.PeekNextEdge())
    {
      new_edge = old_edge->GetForwardEdge();
      pv[0] = new_edge->GetEdge()->GetVertex()->GetForwardVertex()->GetPoint() - new_edge->GetVertex()->GetForwardVertex()->GetPoint();
      pv[1] = old_edge->GetEdgePrev(0)->GetForwardEdge()->GetVertex()->GetForwardVertex()->GetPoint() - new_edge->GetVertex()->GetForwardVertex()->GetPoint();
      
      pv[0].Normalize();
      pv[1].Normalize();
      theta = acos(pv[0].Dot(pv[1]));
      
      p_hinge = new_edge->GetVertex()->GetForwardVertex()->GetPoint();
      Vector v_hinge;
      
      v_hinge = p_hinge;

      offset_v =  v_hinge + (w_offset/sin(theta)) * (pv[0] + pv[1]);
      offset_v[3] = 0;
      
      offset_hinge = offset_v;
      in_vertex = new_mesh->CreateVertex(offset_hinge);
      
      new_edge_offset = new_mesh->CreateEdge();
      new_edge_offset->SetVertex(new_edge->GetEdgePrev(0)->GetVertex());
      new_edge_offset->GetEdge()->SetVertex(in_vertex);
      
      //splice new_edge_offset correctly...
      new_mesh->Splice(1, new_edge_offset, new_edge->GetEdgePrev(0)->GetEdgeNext(1));
    }  //endfor

    //now create connector edges and splice them correctly...
    first_edge = iterQEFacet.PeekFirstEdge();
    for(old_edge = first_edge; old_edge != NULL; old_edge = iterQEFacet.PeekNextEdge())
    {
      new_edge = old_edge->GetForwardEdge();
      
      //create connector edge...
      new_edge_offset = new_mesh->CreateEdge();
      new_edge_offset->SetVertex(new_edge->GetEdgePrev(0)->GetEdgeNext(1)->GetEdgePrev(0)->GetVertex());
      new_edge_offset->GetEdge()->SetVertex(new_edge->GetEdgeNext(0)->GetEdgeNext(0)->GetEdgePrev(1)->GetEdge()->GetVertex());
      
      //splice the connector correctly ...
      new_mesh->Splice(1, new_edge_offset, new_edge->GetEdgePrev(0)->GetEdgeNext(1)->GetEdgePrev(0));
      if(old_edge != first_edge->GetEdgePrev(0))
	new_mesh->Splice(1, new_edge_offset->GetEdge(), new_edge->GetEdgeNext(0)->GetEdgeNext(0)->GetEdgePrev(1)->GetEdge());
    }  //endfor...
    
    //Splice last connector correctly..
    new_mesh->Splice(1, new_edge_offset->GetEdge(), first_edge->GetForwardEdge()->GetEdgePrev(0)->GetEdgeNext(1)->GetEdgePrev(0)->GetEdgeNext(1));

    //set facets correctly...
    for(old_edge = iterQEFacet.PeekFirstEdge(); old_edge != NULL; old_edge = iterQEFacet.PeekNextEdge())
    {
      new_edge = old_edge->GetForwardEdge()->GetEdgeNext(0)->GetEdgeNext(0)->GetEdge();

      //create new facet...
      new_facet = new_mesh->CreateFacet();

      //set facets...
      do
      {
	new_edge->SetFacet(new_facet);
	new_edge = new_edge->GetEdgeNext(0);
      } while(new_edge != old_edge->GetForwardEdge()->GetEdgeNext(0)->GetEdgeNext(0)->GetEdge());
    }

    //ok... cool.. now i have created some new facets and am really relieved 
    //to find that they are perfectly alright.. below i am creating new facets
    //for the inside mesh, so that i can get the grid structure...
    
    //added later...
    for(old_edge = iterQEFacet.PeekFirstEdge(); old_edge != NULL; old_edge = iterQEFacet.PeekNextEdge())
    {
      new_edge = old_edge->GetForwardEdge();
      //create forward edges for each of the "new_edge"s and splice them 
      //correctly and assign facets to them...
    
      //if new_edge does not have forward pointers, create them...
      if(new_edge->GetForwardEdge() == NULL)
      {  
	new_edge_offset = new_mesh->CreateEdge();
	new_edge->SetForwardEdge(new_edge_offset);
	new_edge->GetEdge()->SetForwardEdge(new_edge_offset->GetEdge());
	
	//assign vertices correctly...
	new_edge_offset->SetVertex(new_edge->GetVertex()->GetForwardVertex());
	new_edge_offset->GetEdge()->SetVertex(new_edge->GetEdge()->GetVertex()->GetForwardVertex());

	//splice new_edge_offset correctly...
	for(current = old_edge->GetEdgeNext(1); current != old_edge; current = current->GetEdgeNext(1))
	{
	  if(current->GetForwardEdge() != NULL)
	  {
	    if(current->GetForwardEdge()->GetForwardEdge() != NULL)
	    {
	      new_mesh->Splice(1, new_edge_offset, current->GetForwardEdge()->GetForwardEdge());
	      break;
	    }
	  }	
	}

	for(current = old_edge->GetEdge()->GetEdgeNext(1); current != old_edge->GetEdge(); current = current->GetEdgeNext(1))
	{
	  if(current->GetForwardEdge() != NULL)
	  {
	    if(current->GetForwardEdge()->GetForwardEdge() != NULL)
	    {
	      new_mesh->Splice(1, new_edge_offset->GetEdge(), current->GetForwardEdge()->GetForwardEdge());
	      break;
	    }
	  }	
	}
      }
    }

    first_edge = iterQEFacet.PeekFirstEdge();
    for(old_edge = first_edge; old_edge != NULL; old_edge = iterQEFacet.PeekNextEdge())
    {
      new_edge = old_edge->GetForwardEdge();
      new_edge_offset = new_mesh->CreateEdge();
      new_edge_offset->SetVertex(new_edge->GetEdgePrev(0)->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge()->GetVertex());
      new_edge_offset->GetEdge()->SetVertex(new_edge->GetVertex()->GetForwardVertex());
      
      //splice correctly...
      new_mesh->Splice(1, new_edge_offset,
		       new_edge->GetEdgePrev(0)->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge()->GetEdgeNext(1));

      new_mesh->Splice(1, new_edge_offset->GetEdge(), old_edge->GetEdgeNext(1)->GetForwardEdge()->GetForwardEdge());
    }
    

    //assign facets...
    iterQEFacet.Init(cur_facet);
    for(old_edge = iterQEFacet.PeekFirstEdge(); old_edge != NULL; old_edge = iterQEFacet.PeekNextEdge())
    {
      new_edge = old_edge->GetForwardEdge();
      new_facet = new_mesh->CreateFacet();
      new_edge_offset =  old_edge->GetForwardEdge()->GetEdgeNext(0)->GetEdgeNext(0)->GetEdgePrev(1)->GetEdgeNext(0)->GetEdge();
      
      do
      {
	new_edge_offset->SetFacet(new_facet);
	new_edge_offset = new_edge_offset->GetEdgeNext(0);
      } while(new_edge_offset !=
	      old_edge->GetForwardEdge()->GetEdgeNext(0)->GetEdgeNext(0)->GetEdgePrev(1)->GetEdgeNext(0)->GetEdge());
    }
  }

  //printf("\n Trying to create end caps!!\n");

  //printf("\n # boundary edges = %ld\n", m_plistQEEdgeBoundary->GetLength());
  
  CObjectListIter<QEEdge> bdry_iter;
  bdry_iter.Init(m_plistQEEdgeBoundary);
  QEEdge* bdpe = NULL;

  for(bdpe = bdry_iter.PeekFirst(); bdpe != NULL; bdpe =
	bdry_iter.PeekNext())
  {
    //create connectors.
    new_edge_offset = new_mesh->CreateEdge();

    ASSERT(bdpe->GetForwardEdge() != NULL);
    new_edge_offset->SetVertex(bdpe->GetForwardEdge()->GetVertex());
    new_edge_offset->GetEdge()->SetVertex(bdpe->GetForwardEdge()->GetForwardEdge()->GetVertex());

    //splice this connector.
    new_mesh->Splice(1, new_edge_offset, bdpe->GetForwardEdge());
    new_mesh->Splice(1, new_edge_offset->GetEdge(),
		     bdpe->GetEdgeNext(1)->GetForwardEdge()->GetForwardEdge());
  }    

  //create end-cap facets.
  for(bdpe = bdry_iter.PeekFirst(); bdpe != NULL; bdpe =
	bdry_iter.PeekNext())
  {
    new_edge_offset = bdpe->GetForwardEdge();
    new_facet = new_mesh->CreateFacet();
    
    do
    {
      new_edge_offset->SetFacet(new_facet);
      new_edge_offset = new_edge_offset->GetEdgeNext(0);
    } while(new_edge_offset != bdpe->GetForwardEdge());
  }
    
  /*
  //now, for all edges that did not have facets, we have to create
  //facets- these are pretty much like end caps.
  for(cur_facet = iterQEMesh.PeekFirstFacet(); cur_facet != NULL;
      cur_facet = iterQEMesh.PeekNextFacet())  
  {
    iterQEFacet.Init(cur_facet);
    for(old_edge = iterQEFacet.PeekFirstEdge(); old_edge != NULL;
	old_edge = iterQEFacet.PeekNextEdge())
    {
      if(old_edge->GetForwardEdge()->GetEdge()->GetFacet() == NULL)
      {
	printf("\n Creating connectors\n");
	
	//create connector. 
	new_edge_offset = new_mesh->CreateEdge();
	new_edge_offset->SetVertex(old_edge->GetForwardEdge()->GetVertex());
	new_edge_offset->GetEdge()->SetVertex(old_edge->GetForwardEdge()->GetVertex()->GetForwardVertex());

	//splice this connector.
	new_mesh->Splice(1, new_edge_offset,
			 old_edge->GetEdgePrev(1)->GetForwardEdge());
	new_mesh->Splice(1, new_edge_offset->GetEdge(),
			 old_edge->GetForwardEdge()->GetForwardEdge());
      }
    }
  }


  //create end cap facets.
  for(cur_facet = iterQEMesh.PeekFirstFacet(); cur_facet != NULL;
      cur_facet = iterQEMesh.PeekNextFacet())
  {
    iterQEFacet.Init(cur_facet);
    for(old_edge = iterQEFacet.PeekFirstEdge(); old_edge != NULL;
	old_edge = iterQEFacet.PeekNextEdge())
    {
      if(old_edge->GetForwardEdge()->GetEdge()->GetFacet() == NULL)
      {
	printf("\n Creating end caps\n");

	new_facet = new_mesh->CreateFacet();
	new_edge_offset =
	  old_edge->GetForwardEdge()->GetEdge();
	do
	{
	  new_edge_offset->SetFacet(new_facet);
	  new_edge_offset = new_edge_offset->GetEdgeNext(0);
	} while(new_edge_offset != old_edge->GetForwardEdge()->GetEdge());
      }
    }
  }
  
  */

  new_mesh->Preprocess();
    
  return new_mesh;
}  

//function that calculates the angular bisector with a given offset...
QEEdge* QEMesh::CreateOffsetMeshBisector(QEMesh* new_mesh, QEEdge* old_edge, float offset)
{
  QEEdge* e1 = old_edge->GetForwardEdge();  //old edge...
  QEEdge* e2 = old_edge->GetEdgeNext(1)->GetForwardEdge();

  Vector vec[2];  //vectors along the edges e1 and e2...
  vec[0] = e1->GetEdge()->GetVertex()->GetPoint() - e1->GetVertex()->GetPoint();
  vec[1] = e2->GetEdge()->GetVertex()->GetPoint() - e2->GetVertex()->GetPoint();
  vec[0].Normalize();
  vec[1].Normalize();

  //for the vextex where e1 and e2 intersect...
  Point p_hinge = e1->GetVertex()->GetPoint();
  Vector v_hinge(p_hinge[0], p_hinge[1], p_hinge[2], p_hinge[3]);

  //find out the location of the inwardly offset "hinge" point...
  float theta = acos(vec[0].Dot(vec[1]));
  Vector offset_v =  v_hinge + (offset/sin(theta)) * (vec[0] + vec[1]);
  Point offset_hinge(offset_v[0], offset_v[1], offset_v[2], offset_v[3]);

  //this is the new vertex that is the inward offset of the "hinge" vertex...
  QEVertex* in_vertex = new_mesh->CreateVertex(offset_hinge);

  QEEdge* bisector = new_mesh->CreateEdge();  //the bisecting edge...

  //set the vertices for the bisector edge correctly...
  bisector->SetVertex(old_edge->GetForwardEdge()->GetVertex());
  bisector->GetEdge()->SetVertex(in_vertex);

  return bisector;
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

QEMesh *QEMesh::Copy()
{
  QEMeshIter iterQEMesh;
  QEMesh *pme;
  QEVertex *pvx, *pvxNew;
  QEFacet *pfa, *pfaNew;
  QEEdge *pe, *peNew, *peCurr;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  ClearForwardPointers();

  iterQEMesh.Init(this);

  // Copy Vertices
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pvxNew = pme->CreateVertex(pvx->GetPoint());
      pvxNew->SetNormal(pvx->GetNormal());
      pvxNew->SetTexture(pvx->GetTexture());
      pvxNew->SetSharpVertexType(pvx->GetSharpVertexType());
      pvx->SetForwardVertex(pvxNew);
    }

  // Copy Facets
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      pfaNew = pme->CreateFacet();

      pfa->SetForwardFacet(pfaNew);
    }

  // Copy Edges
  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      if ( pe->GetForwardEdge() == NULL )
	{
	  peNew = pme->CreateEdge();
	  peNew->SetKnotInterval(pe->GetKnotInterval());
	  peNew->GetEdge()->SetKnotInterval(pe->GetEdge()->GetKnotInterval());

	  pe->SetForwardEdge(peNew);
	  pe->GetEdge()->SetForwardEdge(peNew->GetEdge());

	  if ( pe->GetFacet() == m_pfaNULL )
	    {
	      peNew->SetFacet(m_pfaNULL);
	      pme->m_plistQEEdge->Remove(peNew);
	      pme->m_plistQEEdgeBoundary->InsertLast(peNew);
	    }
	  else
	    {
	      peNew->SetFacet(pe->GetFacet()->GetForwardFacet());
	    }

	  if ( pe->GetEdge()->GetFacet() == m_pfaNULL )
	    {
	      peNew->GetEdge()->SetFacet(m_pfaNULL);
	      pme->m_plistQEEdge->Remove(peNew->GetEdge());
	      pme->m_plistQEEdgeBoundary->InsertLast(peNew->GetEdge());
	    }
	  else
	    {
	      peNew->GetEdge()->SetFacet(pe->GetEdge()->GetFacet()->GetForwardFacet());
	    }

	  peNew->SetVertex(pe->GetVertex()->GetForwardVertex());
	  peNew->GetEdge()->SetVertex(pe->GetEdge()->GetVertex()->GetForwardVertex());

	  for ( peCurr = pe->GetEdgePrev(1);
		peCurr != pe;
		peCurr = peCurr->GetEdgePrev(1) )
	    {
	      if ( peCurr->GetForwardEdge() != NULL )
		{
		  Splice(1, peNew, peCurr->GetForwardEdge());
		  break;
		}
	    }

	  for ( peCurr = pe->GetEdge()->GetEdgePrev(1);
		peCurr != pe->GetEdge();
		peCurr = peCurr->GetEdgePrev(1) )
	    {
	      if ( peCurr->GetForwardEdge() != NULL )
		{
		  Splice(1, peNew->GetEdge(), peCurr->GetForwardEdge());
		  break;
		}
	    }
	}
    }

  pme->Preprocess();

  return pme;
}

QEMesh *QEMesh::CopyVoronoiPolygon()
{
  QEMeshIter iterQEMesh;
  QEMesh *pme;
  QEVertex *pvx, *pvxNew;
  QEFacet *pfa, *pfaNew;
  QEEdge *pe, *peNew, *peCurr;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  iterQEMesh.Init(this);

  //////////////////////////////////////////////////////////////////////
  // Clear Forward Pointers, except for the pointers from facets to their related site

  /*
  // Clear Vertex Forward Pointers
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pvx->SetForwardVertex(NULL);
    }

  // Clear Facet Forward Pointers to Facets ONLY (leave vertex and edge ones alone)
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      pfa->SetForwardFacet(NULL);
    }
    */

  // Clear Edge Forward Pointers
  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      pe->SetForwardEdge(NULL);
      pe->GetEdge()->SetForwardEdge(NULL);
    }

  //////////////////////////////////////////////////////////////////////
  // Copy the geometry

  // Copy Vertices
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pvxNew = pme->CreateVertex(pvx->GetPoint());
      pvxNew->SetLimitPosition(pvx->GetLimitPosition());
      pvxNew->SetNormal(pvx->GetNormal());
      pvxNew->SetTexture(pvx->GetTexture());
      pvxNew->SetInscribedCircleCenterType(pvx->GetInscribedCircleCenterType());
      pvx->SetForwardVertex(pvxNew);
    }

  // Copy Facets
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      pfaNew = pme->CreateFacet();

      pfa->SetForwardFacet(pfaNew);
    }

  // Copy Edges
  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      if ( pe->GetForwardEdge() == NULL )
	{
	  peNew = pme->CreateEdge();
	  
	  pe->SetForwardEdge(peNew);
	  pe->GetEdge()->SetForwardEdge(peNew->GetEdge());

	  if ( pe->GetFacet() == GetNullFacet() )
	    {
	      peNew->SetFacet(GetNullFacet());
	      //pme->m_plistQEEdge->Remove(peNew);
	      //pme->m_plistQEEdgeBoundary->InsertLast(peNew);
	    }
	  else
	    {
	      peNew->SetFacet(pe->GetFacet()->GetForwardFacet());
	    }

	  if ( pe->GetEdge()->GetFacet() == GetNullFacet() )
	    {
	      peNew->GetEdge()->SetFacet(GetNullFacet());
	      //pme->m_plistQEEdge->Remove(peNew->GetEdge());
	      //pme->m_plistQEEdgeBoundary->InsertLast(peNew->GetEdge());
	    }
	  else
	    {
	      peNew->GetEdge()->SetFacet(pe->GetEdge()->GetFacet()->GetForwardFacet());
	    }

	  peNew->SetVertex(pe->GetVertex()->GetForwardVertex());
	  peNew->GetEdge()->SetVertex(pe->GetEdge()->GetVertex()->GetForwardVertex());

	  for ( peCurr = pe->GetEdgePrev(1);
		peCurr != pe;
		peCurr = peCurr->GetEdgePrev(1) )
	    {
	      if ( peCurr->GetForwardEdge() != NULL )
		{
		  Splice(1, peNew, peCurr->GetForwardEdge());
		  break;
		}
	    }

	  for ( peCurr = pe->GetEdge()->GetEdgePrev(1);
		peCurr != pe->GetEdge();
		peCurr = peCurr->GetEdgePrev(1) )
	    {
	      if ( peCurr->GetForwardEdge() != NULL )
		{
		  Splice(1, peNew->GetEdge(), peCurr->GetForwardEdge());
		  break;
		}
	    }
	}

      // Set New Facet Forward Pointers for edge and vertex sites
      for ( pfa = iterQEMesh.PeekFirstFacet();
	    pfa != NULL;
	    pfa = iterQEMesh.PeekNextFacet() )
	{
	  pfaNew = pfa->GetForwardFacet();

	  pvx = pfa->GetForwardVertex();
	  if ( pvx != NULL )
	    {
	      pfaNew->SetForwardVertex(pvx->GetForwardVertex());
	    }
	  else
	    {
	      pfaNew->SetForwardVertex(NULL);
	    }

	  pe = pfa->GetForwardEdge();
	  if ( pe != NULL )
	    {
	      pfaNew->SetForwardEdge(pe->GetForwardEdge());
	    }	  
	  else
	    {
	      pfaNew->SetForwardEdge(NULL);
	    }
	}
    }

  // Make sure the vertex is pointing at the same relative edge for ICCT
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pvx->GetForwardVertex()->SetEdge(pvx->GetEdge()->GetForwardEdge());
    }

  pme->Preprocess();

  return pme;
}

QEMesh *QEMesh::CreateMirror()
{
  QEMeshIter iterQEMesh;
  QEMesh *pme;
  QEVertex *pvx, *pvxNew;
  QEFacet *pfa, *pfaNew;
  QEEdge *pe, *peNew, *peCurr;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  ClearForwardPointers();

  iterQEMesh.Init(this);

  // Copy Vertices
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pvxNew = pme->CreateVertex(pvx->GetPoint());
      pvxNew->SetNormal(pvx->GetNormal());
      pvxNew->SetTexture(pvx->GetTexture());
      pvxNew->SetSharpVertexType(pvx->GetSharpVertexType());
      pvx->SetForwardVertex(pvxNew);
    }

  // Copy Facets
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      pfaNew = pme->CreateFacet();

      pfa->SetForwardFacet(pfaNew);
    }

  // Copy Edges
  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      if ( pe->GetForwardEdge() == NULL )
	{
	  peNew = pme->CreateEdge();
	  peNew->SetKnotInterval(pe->GetKnotInterval());
	  peNew->GetEdge()->SetKnotInterval(pe->GetEdge()->GetKnotInterval());

	  pe->SetForwardEdge(peNew);
	  pe->GetEdge()->SetForwardEdge(peNew->GetEdge());

	  if ( pe->GetEdge()->GetFacet() == m_pfaNULL )
	    {
	      peNew->SetFacet(m_pfaNULL);
	      pme->m_plistQEEdge->Remove(peNew);
	      pme->m_plistQEEdgeBoundary->InsertLast(peNew);
	    }
	  else
	    {
	      peNew->SetFacet(pe->GetEdge()->GetFacet()->GetForwardFacet());
	    }

	  if ( pe->GetFacet() == m_pfaNULL )
	    {
	      peNew->GetEdge()->SetFacet(m_pfaNULL);
	      pme->m_plistQEEdge->Remove(peNew->GetEdge());
	      pme->m_plistQEEdgeBoundary->InsertLast(peNew->GetEdge());
	    }
	  else
	    {
	      peNew->GetEdge()->SetFacet(pe->GetFacet()->GetForwardFacet());
	    }

	  peNew->SetVertex(pe->GetVertex()->GetForwardVertex());
	  peNew->GetEdge()->SetVertex(pe->GetEdge()->GetVertex()->GetForwardVertex());

	  for ( peCurr = pe->GetEdgeNext(1);
		peCurr != pe;
		peCurr = peCurr->GetEdgeNext(1) )
	    {
	      if ( peCurr->GetForwardEdge() != NULL )
		{
		  Splice(1, peNew, peCurr->GetForwardEdge());
		  break;
		}
	    }

	  for ( peCurr = pe->GetEdge()->GetEdgeNext(1);
		peCurr != pe->GetEdge();
		peCurr = peCurr->GetEdgeNext(1) )
	    {
	      if ( peCurr->GetForwardEdge() != NULL )
		{
		  Splice(1, peNew->GetEdge(), peCurr->GetForwardEdge());
		  break;
		}
	    }
	}
    }

  pme->Preprocess();

  return pme;
}

QEMesh *QEMesh::CreateDual()
{
  QEMeshIter iterQEMesh;
  QEFacetIter iterQEFacet;
  QEVertexIter iterQEVertex;
  QEMesh *pme;
  QEVertex *pvx, *pvxNew;
  QEFacet *pfa, *pfaNew;
  QEEdge *pe, *peNew, *peCurr;
  Point pt;
  Vector v;
  UINT_32 uiN;
  BOOL bCreateFacet;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  ClearForwardPointers();

  iterQEMesh.Init(this);

  // Dualize Facets
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      pt.Null();
      v.Null();

      iterQEFacet.Init(pfa);
      for ( uiN = 0, pvx = iterQEFacet.PeekFirstVertex();
	    pvx != NULL;
	    uiN++, pvx = iterQEFacet.PeekNextVertex() )
	{
	  pt += pvx->GetPoint();
	  v += pvx->GetTexture();
	}
      pt /= (FLOAT)uiN;
      pt[3] = 1.0;
      v /= (FLOAT)uiN;
      v[3] = 0.0;

      pvxNew = pme->CreateVertex(pt);
      pvxNew->SetTexture(v);
      v = pfa->GetNormal();
      pvxNew->SetNormal(v);

      pfa->SetForwardVertex(pvxNew);
    }

  // Dualize Vertices
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      bCreateFacet = TRUE;
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  if ( pe->GetFacet() == pme->GetNullFacet() )
	    {
	      ASSERT( FALSE );
	      bCreateFacet = FALSE;
	    }
	}

      if ( bCreateFacet )
	{
	  pfaNew = pme->CreateFacet();
	}
      else
	{
	  pfaNew = pme->GetNullFacet();	  
	}
      pvx->SetForwardFacet(pfaNew);
    }

  // Copy Edges
  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      if ( pe->GetForwardEdge() == NULL &&
	   pe->GetFacet() != pme->GetNullFacet() &&
	   pe->GetEdge()->GetFacet() != pme->GetNullFacet() )
	{
	  peNew = pme->CreateEdge();
	  peNew->SetKnotInterval(pe->GetKnotInterval());
	  peNew->GetEdge()->SetKnotInterval(pe->GetEdge()->GetKnotInterval());

	  pe->SetForwardEdge(peNew);
	  pe->GetEdge()->SetForwardEdge(peNew->GetEdge());

	  peNew->SetVertex(pe->GetFacet()->GetForwardVertex());
	  peNew->GetEdge()->SetVertex(pe->GetEdge()->GetFacet()->GetForwardVertex());

	  peNew->SetFacet(pe->GetEdge()->GetVertex()->GetForwardFacet());
	  peNew->GetEdge()->SetFacet(pe->GetVertex()->GetForwardFacet());

	  for ( peCurr = pe->GetEdgePrev(0);
		peCurr != pe;
		peCurr = peCurr->GetEdgePrev(0) )
	    {
	      if ( peCurr->GetForwardEdge() != NULL )
		{
		  Splice(1, peNew, peCurr->GetForwardEdge());
		  break;
		}
	    }

	  for ( peCurr = pe->GetEdge()->GetEdgePrev(0);
		peCurr != pe->GetEdge();
		peCurr = peCurr->GetEdgePrev(0) )
	    {
	      if ( peCurr->GetForwardEdge() != NULL )
		{
		  Splice(1, peNew->GetEdge(), peCurr->GetForwardEdge());
		  break;
		}
	    }
	}
    }

  pme->Preprocess();

  return pme;
}

VOID QEMesh::Transform(const Matrix &m, const Matrix &/*mInv*/)
{
  QEMeshIter iter;
  QEVertex *pvx;
  Point pt;

  iter.Init(this);
  for ( pvx = iter.PeekFirstVertex();
	pvx != NULL;
	pvx = iter.PeekNextVertex() )
    {
      pt = m * pvx->GetPoint();
      pvx->SetPoint(pt);
    }

  Preprocess();
}

VOID QEMesh::CopyBoundaryEdges(QEMesh *pmeSrc)
{
  ASSERT( pmeSrc != NULL );

  QEMeshIter iterQEMesh;
  QEVertex *pvx, *pvxNew;
  QEEdge *pe, *peNew;

  pmeSrc->ClearForwardPointers();

  iterQEMesh.Init(pmeSrc);

  // Copy Boundary Edges
  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdgeBoundary() )
    {
      if ( pe->GetForwardEdge() == NULL )
	{
	  peNew = CreateEdge();
	  
	  pe->SetForwardEdge(peNew);
	  pe->GetEdge()->SetForwardEdge(peNew->GetEdge());

	  m_plistQEEdge->Remove(peNew);
	  m_plistQEEdgeBoundary->InsertLast(peNew);
	  m_plistQEEdge->Remove(peNew->GetEdge());
	  m_plistQEEdgeBoundary->InsertLast(peNew->GetEdge());

	  pvx = pe->GetVertex()->GetForwardVertex();
	  if ( pvx == NULL ) 
	    {
	      pvx = pe->GetVertex();
	      pvxNew = CreateVertex(pvx->GetPoint());	      
	      pvx->SetForwardVertex(pvxNew);
	      pvx = pvxNew;
	    }
	  peNew->SetVertex(pvx);

	  pvx = pe->GetEdge()->GetVertex()->GetForwardVertex();
	  if ( pvx == NULL ) 
	    {
	      pvx = pe->GetEdge()->GetVertex();
	      pvxNew = CreateVertex(pvx->GetPoint());	      
	      pvx->SetForwardVertex(pvxNew);
	      pvx = pvxNew;
	    }
	  peNew->GetEdge()->SetVertex(pvx);
	}
    }
}

VOID QEMesh::ExtractComponents(CObjectList<QEMesh> *plistQEMesh)
{
  ASSERT( plistQEMesh != NULL );

  QEMeshIter iterQEMesh;
  //QEFacetIter iterQEFacet;
  QEFacet *pfa;
  QEVertex *pvx;
  QEMesh *pme;
  UINT_32 uiVertexID, uiFacetID;

  if ( m_plistQEEdgeBoundary->GetLength() > 0 )
    {
      //fprintf(stderr, "ERROR: QEMesh::WriteSIF:  The mesh is not closed:  boundary edges = %ld\n", m_plistQEEdgeBoundary->GetLength());
    }

  iterQEMesh.Init(this);

  // Initialize the vertex and facet ID's to 0
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pvx->SetID(0);
    }

  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      pfa->SetID(0);
    }

  for ( pfa = m_plistQEFacet->PeekFirst();
	pfa != NULL;
	pfa = m_plistQEFacet->PeekFirst() )
    {
      uiVertexID = 0;
      uiFacetID = 0;

      pme = AllocateQEMesh();
      ASSERT( pme != NULL );
      pme->Init();
      ExtractComponent(pme, uiVertexID, uiFacetID, pfa);
      plistQEMesh->InsertLast(pme);
    }

  ASSERT( m_plistQEVertex->GetLength() == 0 );
  ASSERT( m_plistQEEdge->GetLength() == 0 );
  ASSERT( m_plistQEEdgeBoundary->GetLength() == 0 );
  ASSERT( m_plistQEFacet->GetLength() == 0 );
}

VOID QEMesh::ExtractComponent(QEMesh *pmeComponent, 
			      UINT_32 &uiVertexID, UINT_32 &uiFacetID, 
			      QEFacet *pfaCurr)
{
  QEFacetIter iterQEFacet;
  QEFacet *pfa;
  QEEdge *pe;
  QEVertex *pvx;

  ASSERT( pmeComponent != NULL );
  ASSERT( pfaCurr != NULL );

  uiFacetID++;
  pfaCurr->SetID(uiFacetID);
  m_plistQEFacet->Remove(pfaCurr);
  pmeComponent->m_plistQEFacet->InsertLast(pfaCurr);	  

  iterQEFacet.Init(pfaCurr);
  
  for ( pe = iterQEFacet.PeekFirstEdge();
	pe != NULL;
	pe = iterQEFacet.PeekNextEdge() )
    {
      pvx = pe->GetVertex();
      if ( pvx->GetID() == 0 )
	{
	  uiVertexID++;
	  pvx->SetID(uiVertexID);
	  m_plistQEVertex->Remove(pvx);
	  pmeComponent->m_plistQEVertex->InsertLast(pvx);	  
	}
      
      pfa = pe->GetEdge()->GetFacet();

      m_plistQEEdge->Remove(pe);
      pmeComponent->m_plistQEEdge->InsertLast(pe);

      if ( pfa == m_pfaNULL )
	{
	  // Transfer over the boundary edge
	  m_plistQEEdgeBoundary->Remove(pe->GetEdge());
	  pmeComponent->m_plistQEEdgeBoundary->InsertLast(pe->GetEdge());
	}
      else if ( pfa->GetID() == 0 )
	{
	  ExtractComponent(pmeComponent, uiVertexID, uiFacetID, pfa);
	}
    }
}

VOID QEMesh::Preprocess()
{
  QEMeshIter iterQEMesh;
  QEVertexIter iterQEVertex;
  QEVertex *pvx;
  QEEdge *pe;
  QEFacet *pfa;
  UINT_32 uiN;

  iterQEMesh.Init(this);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge(), uiN = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), uiN++ )
	{
	}
      pvx->SetValence(uiN);
    }
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      pfa->Preprocess();
    }
}

QEVertex *QEMesh::CreateVertex(const Point &pt)
{
  QEVertex *pvx;

  pvx = AllocateQEVertex();
  ASSERT( pvx != NULL );
  pvx->Init(pt);
  AddVertex(pvx);

  return pvx;
}

QEFacet *QEMesh::CreateFacet()
{
  QEFacet *pfa;

  pfa = AllocateQEFacet();
  ASSERT( pfa != NULL );
  pfa->Init();
  AddFacet(pfa);

  return pfa;
}

QEEdge *QEMesh::CreateEdge()
{
  QEEdge *ppe[2];
  UINT_32 i;

  for ( i = 0; i < 2; i++ )
    {
      ppe[i] = AllocateQEEdge();
      ASSERT( ppe[i] != NULL );
      ppe[i]->Init();
      AddEdge(ppe[i]);

      ppe[i]->SetEdgeNext(1, ppe[i]);
      ppe[i]->SetEdgePrev(1, ppe[i]);
    }

  ppe[0]->SetFacet(m_pfaNULL);
  ppe[1]->SetFacet(m_pfaNULL);

  ppe[0]->SetEdge(ppe[1]);
  ppe[1]->SetEdge(ppe[0]);

  ppe[0]->SetEdgeNext(0, ppe[1]);
  ppe[1]->SetEdgePrev(0, ppe[0]);
  ppe[0]->SetEdgePrev(0, ppe[1]);
  ppe[1]->SetEdgeNext(0, ppe[0]);

  return ppe[0];
}

QEEdgeCollapse *QEMesh::CreateEdgeCollapse()
{
  QEEdgeCollapse *pec;

  pec = AllocateQEEdgeCollapse();
  ASSERT( pec != NULL );
  pec->Init();
  AddEdgeCollapse(pec);

  return pec;
}

QEEdgeNode *QEMesh::CreateEdgeNode(const Point &ptFrom, const Point &ptTo)
{
  QEEdgeNode *pen;

  pen = AllocateQEEdgeNode();
  ASSERT( pen != NULL );
  pen->Init(ptFrom, ptTo);
  AddEdgeNode(pen);

  return pen;
}

QEFacetNode *QEMesh::CreateFacetNode()
{
  QEFacetNode *pfn;

  pfn = AllocateQEFacetNode();
  ASSERT( pfn != NULL );
  pfn->Init();
  AddFacetNode(pfn);

  return pfn;
}

VOID QEMesh::RemoveAllVertices(CObjectList<QEVertex> *plistQEVertex)
{
  plistQEVertex->Splice(m_plistQEVertex);
}

VOID QEMesh::AddVertex(QEVertex *pvx)
{
  ASSERT( pvx != NULL );

  m_plistQEVertex->InsertLast(pvx);
}

VOID QEMesh::AddWedge(QEWedge *pwe)
{
  ASSERT( pwe != NULL );

  m_plistQEWedge->InsertLast(pwe);
}

VOID QEMesh::AddEdge(QEEdge *pe)
{
  ASSERT( pe != NULL );

  m_plistQEEdge->InsertLast(pe);
}

VOID QEMesh::AddFacet(QEFacet *pfa)
{
  ASSERT( pfa != NULL );

  m_plistQEFacet->InsertLast(pfa);
}

VOID QEMesh::AddEdgeCollapse(QEEdgeCollapse *pec)
{
  ASSERT( pec != NULL );

  m_plistQEEdgeCollapse->InsertLast(pec);
}

VOID QEMesh::AddEdgeNode(QEEdgeNode *pen)
{
  ASSERT( pen != NULL );

  m_plistQEEdgeNode->InsertLast(pen);
}

VOID QEMesh::AddFacetNode(QEFacetNode *pfn)
{
  ASSERT( pfn != NULL );

  m_plistQEFacetNode->InsertLast(pfn);
}

VOID QEMesh::DestroyVertex(QEVertex *pvx)
{
  RemoveVertex(pvx);
  DeallocateQEVertex(pvx);
}

VOID QEMesh::DestroyWedge(QEWedge *pwe)
{
  RemoveWedge(pwe);
  DeallocateQEWedge(pwe);
}

VOID QEMesh::DestroyEdge(QEEdge *pe)
{
  RemoveEdge(pe);
  DeallocateQEEdge(pe);
}

VOID QEMesh::DestroyFacet(QEFacet *pfa)
{
  RemoveFacet(pfa);
  DeallocateQEFacet(pfa);
}

VOID QEMesh::DestroyEdgeCollapse(QEEdgeCollapse *pec)
{
  RemoveEdgeCollapse(pec);
  DeallocateQEEdgeCollapse(pec);
}

VOID QEMesh::DestroyEdgeNode(QEEdgeNode *pen)
{
  RemoveEdgeNode(pen);
  DeallocateQEEdgeNode(pen);
}

VOID QEMesh::DestroyFacetNode(QEFacetNode *pfn)
{
  RemoveFacetNode(pfn);
  DeallocateQEFacetNode(pfn);
}

VOID QEMesh::CollapseEdge(QEEdge *pe)
{
  QEEdge *peCurr;
  QEVertex *pvx;

  // Set the two facets to point to edges which will continue to exist
  if ( pe->GetFacet() != GetNullFacet() )
    {
      ASSERT( pe->GetFacet() == pe->GetEdgeNext(0)->GetFacet() );
      pe->GetEdgeNext(0)->SetFacet(pe->GetEdgeNext(0)->GetFacet());
    }
  if ( pe->GetEdge()->GetFacet() != GetNullFacet() )
    {
      ASSERT( pe->GetEdge()->GetFacet() == pe->GetEdge()->GetEdgePrev(0)->GetFacet() );
      pe->GetEdge()->GetEdgePrev(0)->SetFacet(pe->GetEdge()->GetEdgePrev(0)->GetFacet());
    }

  pvx = pe->GetEdge()->GetVertex();
  for ( peCurr = pe->GetEdgeNext(1);
	peCurr != pe;
	peCurr = peCurr->GetEdgeNext(1) )
    {
      peCurr->SetVertex(pvx);
    }

  peCurr = pe->GetEdgePrev(1);
  if ( peCurr != pe )
    {
      Splice(1, pe, peCurr);
      Splice(1, pe->GetEdge(), peCurr);
    }

  peCurr = pe->GetEdge()->GetEdgePrev(1);
  ASSERT( peCurr != pe->GetEdge() );
  Splice(1, pe->GetEdge(), peCurr); 

  DestroyVertex(pe->GetVertex());
  DestroyEdge(pe->GetEdge());
  DestroyEdge(pe);
}

VOID QEMesh::RemoveVertex(QEVertex *pvx)
{
  ASSERT( pvx != NULL );

  m_plistQEVertex->Remove(pvx);
}

VOID QEMesh::RemoveWedge(QEWedge *pwe)
{
  ASSERT( pwe != NULL );

  m_plistQEWedge->Remove(pwe);
}

VOID QEMesh::RemoveEdge(QEEdge *pe)
{
  ASSERT( pe != NULL );

  m_plistQEEdge->Remove(pe);
}

VOID QEMesh::RemoveFacet(QEFacet *pfa)
{
  ASSERT( pfa != NULL );

  m_plistQEFacet->Remove(pfa);
}

VOID QEMesh::RemoveEdgeCollapse(QEEdgeCollapse *pec)
{
  ASSERT( pec != NULL );

  m_plistQEEdgeCollapse->Remove(pec);
}

VOID QEMesh::RemoveEdgeNode(QEEdgeNode *pen)
{
  ASSERT( pen != NULL );

  m_plistQEEdgeNode->Remove(pen);
}

VOID QEMesh::RemoveFacetNode(QEFacetNode *pfn)
{
  ASSERT( pfn != NULL );

  m_plistQEFacetNode->Remove(pfn);
}

QEEdge *QEMesh::CreateContour(QEFacet *pfa, QEVertex *pvx)
{
  ASSERT( pfa != NULL );
  ASSERT( pvx != NULL );

  QEEdge *pe;

  pe = CreateEdge();
  ASSERT( pe != NULL );
  pe->SetVertex(pvx);
  pe->SetFacet(pfa);
  pe->GetEdge()->SetVertex(pvx);

  return pe;
}

QEEdge *QEMesh::AddContourVertex(QEEdge *peContour, QEVertex *pvx)
{
  ASSERT( peContour != NULL );
  ASSERT( pvx != NULL );

  QEEdge *pe, *peLast;

  pe = CreateEdge();
  ASSERT( pe != NULL );

  peLast = peContour->GetEdgePrev(0);
  if ( peLast == peContour->GetEdge() )
    {
      peLast = peContour;
    }
  else
    {
      Splice(1, peContour, peLast->GetEdge());
    }

  pe->SetVertex(pvx);
  pe->SetFacet(peContour->GetFacet());
  pe->GetEdge()->SetVertex(peContour->GetVertex());
  peLast->GetEdge()->SetVertex(pvx);

  Splice(1, peContour, pe->GetEdge());
  Splice(1, pe, peLast->GetEdge());

  return pe;
}

VOID QEMesh::RandomizeVertexList(CObjectList<QEVertex> *plistQEVertex)
{
  QEVertex **ppvx, *pvx;
  UINT_32 i, uiSize;

  uiSize = plistQEVertex->GetLength();
  if ( uiSize == 0 ) 
    {
      return;
    }

  ppvx = new QEVertex*[uiSize];

  for ( i = 0, pvx = plistQEVertex->RemoveFirst();
	pvx != NULL;
	i++, pvx = plistQEVertex->RemoveFirst() )
    {
      ppvx[i] = pvx;
    }
  ASSERT( i == uiSize );

  RandomizeArray(uiSize, (VOID**)ppvx);

  for ( i = 0; i < uiSize; i++ )
    {
      plistQEVertex->InsertLast(ppvx[i]);
    }

  delete []ppvx;
}

VOID QEMesh::RandomizeArray(UINT_32 uiSize, VOID **ppv)
{
  UINT_32 i, j, uiRand;

  for ( i = uiSize; i > 0; i-- )
    {
      uiRand = rand();

      // Bias towards the lower end
      //j = uiRand % i;

      // More evenly distributed
      j = (UINT_32)(((FLOAT)uiRand / ( ((FLOAT)RAND_MAX) + 1.0 )) * (FLOAT)i);

      Swap(ppv+i-1, ppv+j);
    }
}

VOID QEMesh::Swap(VOID **ppv0, VOID **ppv1)
{
  VOID *pv;

  pv = *ppv0;
  *ppv0 = *ppv1;
  *ppv1 = pv;
}

VOID QEMesh::Splice(UINT_32 uiDual, QEEdge *pe0, QEEdge *pe1)
{
  ASSERT( uiDual < 2 );
  ASSERT( pe0 != NULL && pe1 != NULL );

  QEEdge *pppe[2][3];
  UINT_32 i, j, uiDualInv;

  uiDualInv = uiDual ^ 0x1;

  pppe[0][0] = pe0;
  pppe[1][0] = pe1;
  for ( i = 0; i < 2; i++ )
    {
      pppe[i][1] = pppe[i][0]->GetEdgeNext(uiDual);
      pppe[i][2] = pppe[i][1]->GetEdge();
    }

  j = 1;
  for ( i = 0; i < 2; i++ )
    {
      pppe[i][0]->SetEdgeNext(uiDual, pppe[j][1]);
      pppe[i][0]->SetEdgePrev(uiDualInv, pppe[j][2]);
      pppe[i][1]->SetEdgePrev(uiDual, pppe[j][0]);
      pppe[i][2]->SetEdgeNext(uiDualInv, pppe[j][0]);

      j = i;
    }
}

////////////////////////////////////////////////////////////////////
// QEMesh::InsertDiagonal                                         //
//                                                                //
//           Old Facet                     Split Facet            //
//                                                                //
//  -v4--e0'/e0--v0--e2'/e2--v2-    -v4--e0'/e0--v0--e2'/e2--v2-  //
//                                                |               //
//              Fold             =>       Fnew  e'\e  Fold        //
//                                                |               //
//  -v3--e3/e3'--v1--e1/e1'--v5-    -v3--e3/e3'--v1--e1/e1'--v5-  //
//                                                                //
QEEdge *QEMesh::InsertDiagonal(QEEdge *pe0, QEEdge *pe1)
{
  ASSERT( pe0 != NULL );
  ASSERT( pe1 != NULL );
  ASSERT( pe0 != pe1 );
  ASSERT( pe0->GetFacet() == pe1->GetFacet() );

  QEEdge *pe;

  pe = CreateEdge();
  ASSERT( pe != NULL );

  pe->SetKnotInterval(0.0);
  pe->GetEdge()->SetKnotInterval(0.0);
  pe->SetVertex(pe0->GetVertex());
  pe->GetEdge()->SetVertex(pe1->GetVertex());
  pe->SetFacet(pe1->GetFacet());
  pe->GetEdge()->SetFacet(pe0->GetFacet());
  Splice(1, pe, pe0);
  Splice(1, pe->GetEdge(), pe1);

  return pe;
}

VOID QEMesh::InsertDiagonal(QEEdge *pe0, QEEdge *pe1, QEEdge *pe)
{
  ASSERT( pe0 != NULL );
  ASSERT( pe1 != NULL );
  ASSERT( pe0 != pe1 );
  ASSERT( pe0->GetFacet() == pe1->GetFacet() );

  ASSERT( pe != NULL );

  pe->SetVertex(pe0->GetVertex());
  pe->GetEdge()->SetVertex(pe1->GetVertex());
  pe->SetFacet(pe1->GetFacet());
  pe->GetEdge()->SetFacet(pe0->GetFacet());
  Splice(1, pe, pe0);
  Splice(1, pe->GetEdge(), pe1);
}

VOID QEMesh::RemoveDiagonal(QEEdge *pe)
{
  ASSERT( pe != NULL );

  Splice(1, pe, pe->GetEdgePrev(1));
  Splice(1, pe->GetEdge(), pe->GetEdge()->GetEdgePrev(1));  
}

VOID QEMesh::InsertVertex(QEFacet *pfa, QEVertex *pvx)
{
  ASSERT( pfa != NULL );
  ASSERT( pvx != NULL );

  QEEdge *pe, *peCurr;
  QEEdge *peNew, *peNewStart, *peNewPrev;
  QEFacet *pfaNew;

  pe = pfa->GetEdge();

  peNew = CreateEdge();
  ASSERT( peNew != NULL );

  peNew->SetVertex(pe->GetVertex());
  peNew->GetEdge()->SetVertex(pvx);
  peNew->SetFacet(pfa);

  Splice(1, pe, peNew);

  peNewStart = peNew;
  peNewPrev = peNew;
  peCurr = pe->GetEdgeNext(0);
  while ( peCurr != peNewStart )
    {
      pfaNew = CreateFacet();
      ASSERT( pfaNew != NULL );

      peNew = CreateEdge();
      ASSERT( peNew != NULL );

      peNew->SetVertex(peCurr->GetVertex());
      peNew->GetEdge()->SetVertex(pvx);
      
      Splice(1, peCurr, peNew);
      Splice(1, peNewPrev->GetEdge(), peNew->GetEdge());

      peNew->SetFacet(pfaNew);
      peNew->GetEdgeNext(0)->SetFacet(pfaNew);
      peNew->GetEdgePrev(0)->SetFacet(pfaNew);

      peNewPrev = peNew;
      peCurr = peCurr->GetEdgeNext(0);
    }
  peNew->GetEdge()->SetFacet(pfa);
}

VOID QEMesh::InsertVertex(QEFacet *pfa, QEVertex *pvx, CListInt *plistQEEdgeQueue)
{
  ASSERT( pfa != NULL );
  ASSERT( pvx != NULL );
  ASSERT( plistQEEdgeQueue != NULL );

  QEEdge *pe, *peCurr;

  pe = pfa->GetEdge();
  peCurr = pe;
  do
    {
      plistQEEdgeQueue->InsertLast(peCurr, 0);

      peCurr = peCurr->GetEdgeNext(0);
    }
  while ( peCurr != pe );

  InsertVertex(pfa, pvx);
}

VOID QEMesh::InsertVertexUpdateSearch(QEFacet *pfa, QEVertex *pvx, 
				      CListInt *plistQEEdgeQueue)
{
  ASSERT( pfa != NULL );
  ASSERT( pvx != NULL );
  ASSERT( plistQEEdgeQueue != NULL );

  QEVertex *ppvx[3];
  QEEdge *ppe[3], *pe, *peCurr;
  QEFacet *ppfa[3];
  QEEdgeNode *ppen[3];
  QEFacetNode *ppfn[3];
  UINT_32 i;

  InsertVertex(pfa, pvx, plistQEEdgeQueue);

  pe = pvx->GetEdge();
  i = 0;
  peCurr = pe;
  do
    {
      ASSERT( i < 3 );
      ppe[i] = peCurr;
      ppvx[i] = ppe[i]->GetEdge()->GetVertex();
      ppfa[i] = ppe[i]->GetFacet();

      // Allocate the new search DAG nodes
      ppen[i] = CreateEdgeNode(pvx->GetPoint(), ppvx[i]->GetPoint());
      ASSERT( ppen[i] != NULL ); 

      peCurr = peCurr->GetEdgeNext(1);
      i++;
    }
  while ( peCurr != pe );
  ASSERT( i == 3 );

  // Update the search DAG structure
  ppfn[0] = pfa->GetFacetNode();
  ASSERT( ppfn[0]->GetParent(0) != NULL );
  for ( i = 0; i < 2 && ppfn[0]->GetParent(i) != NULL; i++ )
    {
      if ( ppfn[0] == ppfn[0]->GetParent(i)->GetChild(0) )
	{
	  ppfn[0]->GetParent(i)->SetChild(0, ppen[0]);
	}
      else
	{
	  ASSERT( ppfn[0] == ppfn[0]->GetParent(i)->GetChild(1) );
	  ppfn[0]->GetParent(i)->SetChild(1, ppen[0]);
	}
    }

  ppen[0]->SetChild(0, ppen[1]);
  ppen[0]->SetChild(1, ppen[2]);

  ASSERT( ppfa[0] == pfa );
  ASSERT( ppfn[0]->GetFacet() == ppfa[0] );

  for ( i = 1; i < 3; i++ )
    {
      ppfn[i] = CreateFacetNode();
      ASSERT( ppfn[i] != NULL ); 
      ppfn[i]->SetFacet(ppfa[i]);
      ppfa[i]->SetFacetNode(ppfn[i]);
    }
      
  ppen[1]->SetChild(0, ppfn[1]);
  ppen[1]->SetChild(1, ppfn[0]);

  ppen[2]->SetChild(0, ppfn[2]);
  ppen[2]->SetChild(1, ppfn[1]);

  ppfn[0]->SetParents(ppen[1], NULL);
  ppfn[1]->SetParents(ppen[1], ppen[2]);
  ppfn[2]->SetParents(ppen[2], NULL);
}

VOID QEMesh::InsertVertex(QEEdge *pe, QEVertex *pvx)
{
  ASSERT( pe != NULL );
  ASSERT( pvx != NULL );

  QEEdge *peCurr, *peNew;
  QEFacet *pfaNew;
  UINT_32 i;

  peNew = CreateEdge();
  ASSERT( peNew != NULL );
  peNew->SetFacet(pe->GetFacet());
  peNew->GetEdge()->SetFacet(pe->GetEdge()->GetFacet());
  peNew->SetVertex(pe->GetVertex());
  peNew->GetEdge()->SetVertex(pvx);
  pe->SetVertex(pvx);
  
  peCurr = pe->GetEdgePrev(1);
  Splice(1, peCurr, peNew);
  Splice(1, peNew, pe);
  Splice(1, peNew->GetEdge(), pe);

  for ( i = 0; i < 2; i++ )
    {
      if ( pe->GetFacet() != m_pfaNULL )
	{
	  peCurr = pe->GetEdgePrev(0)->GetEdgePrev(0);
	  while ( peCurr != pe->GetEdgeNext(0) )
	    {
	      pfaNew = CreateFacet();
	      ASSERT( pfaNew != NULL );
	      
	      peNew = CreateEdge();
	      ASSERT( peNew != NULL );
	      
	      Splice(1, peCurr, peNew->GetEdge());
	      Splice(1, pe, peNew);
	  
	      peNew->GetEdge()->SetVertex(peCurr->GetVertex());
	      peNew->SetVertex(pvx);
	      peNew->SetFacet(pfaNew);
	      peNew->GetEdgeNext(0)->SetFacet(pfaNew);
	      peNew->GetEdgePrev(0)->SetFacet(pfaNew);
	  
	      peCurr = peNew->GetEdge()->GetEdgePrev(0);
	    }
	  peNew->GetEdge()->SetFacet(pe->GetFacet());
	}

      pe = pe->GetEdgePrev(1);
    }

}

VOID QEMesh::InsertVertex(QEEdge *pe, QEVertex *pvx, CListInt *plistQEEdgeQueue)
{
  ASSERT( pe != NULL );
  ASSERT( pvx != NULL );
  ASSERT( plistQEEdgeQueue != NULL );

  QEEdge *peStart, *peCurr;
  UINT_32 i;

  peStart = pe;
  for ( i = 0; i < 2; i++ )
    {
      if ( peStart->GetFacet() != m_pfaNULL )
	{
	  for ( peCurr = peStart->GetEdgeNext(0);	   
		peCurr != peStart;
		peCurr = peCurr->GetEdgeNext(0) )
	    {
	      plistQEEdgeQueue->InsertLast(peCurr, 0);
	    }
	}
      peStart = pe->GetEdge();
    }

  InsertVertex(pe, pvx);
}

VOID QEMesh::InsertVertexUpdateSearch(QEEdge *pe, QEVertex *pvx, 
				      CListInt *plistQEEdgeQueue)
{
  ASSERT( pe != NULL );
  ASSERT( pvx != NULL );
  ASSERT( plistQEEdgeQueue != NULL );
 
  //QEFacetIter iterQEFacet;
  QEVertex *ppvx[4];
  QEEdge *ppe[4], *peCurr;
  QEFacet *ppfa[4];
  QEEdgeNode *ppen[2];
  QEFacetNode *ppfn[4];
  UINT_32 i, j;

  InsertVertex(pe, pvx, plistQEEdgeQueue);

  i = 0;
  peCurr = pe;
  do
    {
      ASSERT( i < 4 );
      ppe[i] = peCurr;
      ppvx[i] = ppe[i]->GetEdge()->GetVertex();
      ppfa[i] = ppe[i]->GetFacet();

      peCurr = peCurr->GetEdgeNext(1);
      i++;
    }
  while ( peCurr != pe );
  ASSERT( i == 4 );

  for ( i = 0; i < 2; i++ )
    {
      // Allocate the new search DAG nodes
      ppen[i] = CreateEdgeNode(pvx->GetPoint(), ppvx[2*i + 1]->GetPoint());
      ASSERT( ppen[i] != NULL ); 
    }

  // Create the two new Facet Nodes
  for ( i = 0; i < 4; i++ )
    {
      ppfn[i] = ppfa[i]->GetFacetNode();
      ASSERT( ppfn[i] != NULL );
      i++;
      ppfn[i] = ppfa[i]->GetFacetNode();
      ASSERT( ppfn[i] == NULL );
      ppfn[i] = CreateFacetNode();
      ASSERT( ppfn[i] != NULL ); 
      ppfn[i]->SetFacet(ppfa[i]);
      ppfa[i]->SetFacetNode(ppfn[i]);
    }

  // Update the search DAG structure
  for ( i = 0; i < 2; i++ )
    {
      ASSERT( ppfn[2*i]->GetParent(0) != NULL );
      for ( j = 0; j < 2 && ppfn[2*i]->GetParent(j) != NULL; j++ )
	{
	  if ( ppfn[2*i] == ppfn[2*i]->GetParent(j)->GetChild(0) )
	    {
	      ppfn[2*i]->GetParent(j)->SetChild(0, ppen[i]);
	    }
	  else
	    {
	      ASSERT( ppfn[2*i] == ppfn[2*i]->GetParent(j)->GetChild(1) );
	      ppfn[2*i]->GetParent(j)->SetChild(1, ppen[i]);
	    }
	}

      ppen[i]->SetChild(0, ppfn[2*i+1]);
      ppen[i]->SetChild(1, ppfn[2*i]);

      ppfn[2*i+1]->SetParents(ppen[i], NULL);
      ppfn[2*i]->SetParents(ppen[i], NULL);
    }
}

//////////////////////////////////////////////////////////////
// QEMesh::DelaunayFlip                                     //
//                                                          //
//     Old Quadrilateral           Flipped Quadrilateral    //
//                                                          //
//    v3-----e0'/e0----v0           v3-----e0'/e0----v0     //
//     |              / |            | \              |     //
//     |   f0       /   |            |   \       f0   |     //
//     |          /     |            |     \          |     //
//  e3'\e3    e'/e    e2\e2'  =>  e3'\e3    e\e'     e2\e2'  //
//     |     /          |            |          \     |     //
//     |   /       f1   |            |   f1       \   |     //
//     | /              |            |              \ |     //
//    v1-----e1/e1'----v2           v1-----e1/e1'----v2     //
//                                                          //
VOID QEMesh::DelaunayFlip(QEEdge *pe)
{
  ASSERT( pe != NULL );

  QEVertex *ppvx[4];
  QEEdge *ppe[4];
  QEFacet *ppfa[2];

  pe->GetVertex()->SetEdge(pe->GetEdgeNext(1));
  pe->GetEdge()->GetVertex()->SetEdge(pe->GetEdge()->GetEdgeNext(1));
  pe->GetFacet()->SetEdge(pe->GetEdgeNext(0));
  pe->GetEdge()->GetFacet()->SetEdge(pe->GetEdge()->GetEdgeNext(0));

  ppfa[0] = pe->GetEdge()->GetFacet();
  ppfa[1] = pe->GetFacet();

  ppe[0] = pe->GetEdge()->GetEdgeNext(0);
  ppvx[0] = ppe[0]->GetVertex();
  ppe[1] = pe->GetEdgeNext(0);
  ppvx[1] = ppe[1]->GetVertex();
  ppe[2] = pe->GetEdgePrev(0);
  ppvx[2] = ppe[2]->GetVertex();
  ppe[3] = pe->GetEdge()->GetEdgePrev(0);
  ppvx[3] = ppe[3]->GetVertex();

  ASSERT( ppe[0]->GetEdgeNext(0) == ppe[3] );
  ASSERT( ppe[1]->GetEdgeNext(0) == ppe[2] );

  // Detach the diagonal
  Splice(1, pe, ppe[0]);
  Splice(1, pe->GetEdge(), ppe[1]);

  // Reattach the diagonal
  Splice(1, pe, ppe[2]);
  Splice(1, pe->GetEdge(), ppe[3]);

  pe->SetVertex(ppvx[2]);
  pe->GetEdge()->SetVertex(ppvx[3]);
  ppe[2]->SetFacet(ppfa[0]);
  ppe[3]->SetFacet(ppfa[1]);
}

VOID QEMesh::DelaunayFlip(QEEdge *pe, CListInt *plistQEEdgeQueue)
{
  ASSERT( pe != NULL );
  ASSERT( plistQEEdgeQueue != NULL );

  DelaunayFlip(pe);

  plistQEEdgeQueue->InsertLast(pe->GetEdgeNext(0), 0);
  plistQEEdgeQueue->InsertLast(pe->GetEdgePrev(0), 0);
  plistQEEdgeQueue->InsertLast(pe->GetEdge()->GetEdgeNext(0), 0);
  plistQEEdgeQueue->InsertLast(pe->GetEdge()->GetEdgePrev(0), 0);
}

VOID QEMesh::DelaunayFlipIncremental(QEEdge *pe, CListInt *plistQEEdgeQueue)
{
  ASSERT( pe != NULL );
  ASSERT( plistQEEdgeQueue != NULL );

  //plistQEEdgeQueue->InsertLast(pe->GetEdgeNext(0), 0);
  //plistQEEdgeQueue->InsertLast(pe->GetEdgePrev(0), 0);
  plistQEEdgeQueue->InsertLast(pe->GetEdge()->GetEdgeNext(0), 0);
  plistQEEdgeQueue->InsertLast(pe->GetEdge()->GetEdgePrev(0), 0);

  DelaunayFlip(pe);
}

VOID QEMesh::DelaunayFlipUpdateSearch(QEEdge *pe, CListInt *plistQEEdgeQueue)
{
  QEVertex *ppvx[2];
  QEFacet *ppfa[2];
  QEFacetNode *ppfn[2];
  QEEdgeNode *penNew;
  UINT_32 i, j;

  DelaunayFlipIncremental(pe, plistQEEdgeQueue);

  ppvx[0] = pe->GetVertex();
  ppfa[0] = pe->GetFacet();
  ppfn[0] = ppfa[0]->GetFacetNode();
  ppvx[1] = pe->GetEdge()->GetVertex();
  ppfa[1] = pe->GetEdge()->GetFacet();
  ppfn[1] = ppfa[1]->GetFacetNode();

  ASSERT( (ppfn[0] != NULL && ppfn[1] != NULL) ||
	  (ppfn[0] == NULL && ppfn[1] == NULL) );

  if ( ppfn[0] != NULL )
    {
      penNew = CreateEdgeNode(ppvx[0]->GetPoint(), ppvx[1]->GetPoint());
      ASSERT( penNew != NULL ); 

      for ( j = 0; j < 2; j++ )
	{
	  ASSERT( ppfn[j]->GetParent(0) != NULL );
	  for ( i = 0; i < 2 && ppfn[j]->GetParent(i) != NULL; i++ )
	    {
	      if ( ppfn[j] == ppfn[j]->GetParent(i)->GetChild(0) )
		{
		  ppfn[j]->GetParent(i)->SetChild(0, penNew);
		}
	      else
		{
		  ASSERT( ppfn[j] == ppfn[j]->GetParent(i)->GetChild(1) );
		  ppfn[j]->GetParent(i)->SetChild(1, penNew);
		}
	    }
	}

      penNew->SetChild(0, ppfn[0]);
      ppfn[0]->SetParents(penNew, NULL);
      penNew->SetChild(1, ppfn[1]);
      ppfn[1]->SetParents(penNew, NULL);
    }
}

VOID QEMesh::CalculateBound()
{
  QEMeshIter iter;
  QEVertex *pvxCurr;

  iter.Init(this);
  for ( pvxCurr = iter.PeekFirstVertex();
	pvxCurr != NULL;
	pvxCurr = iter.PeekNextVertex() )
    {
      //pCBound->InsertIntoBound(pvxCurr->GetPoint());
    }
}

BOOL QEMesh::ConstrainedDelaunayEdge(QEEdge *pe)
{
  if ( pe->GetFacet() == QEMesh::GetNullFacet() ||
       pe->GetEdge()->GetFacet() == QEMesh::GetNullFacet() ||
       pe->GetKnotInterval() == -1.0 ||
       pe->GetEdge()->GetKnotInterval() == -1.0 )
    {
      return TRUE;
    } 

  return DelaunayEdge(pe);
}

BOOL QEMesh::DelaunayEdge(QEEdge *pe)
{
  QEVertex *ppvx[4];

  ppvx[0] = pe->GetVertex();
  ppvx[1] = pe->GetEdgePrev(1)->GetEdge()->GetVertex();
  ppvx[2] = pe->GetEdge()->GetVertex();
  ppvx[3] = pe->GetEdgeNext(1)->GetEdge()->GetVertex();

  return !InCircumcircle(ppvx[0]->GetPoint(), 
			 ppvx[1]->GetPoint(), 
			 ppvx[2]->GetPoint(), 
			 ppvx[3]->GetPoint());
}

BOOL QEMesh::InCircumcircle(const Point &ptTri0, const Point &ptTri1, const Point &ptTri2, 
			    const Point &pt)
{
  FLOAT f;

  f = 0.0;

#ifndef COMPILER_PROBLEM
  switch ( m_eAccuracy )
    {
      case atFloatNaive:
	{
	  MATRIX4D m;
	  Matrix mDet;

	  m[0] = ptTri0(0);
	  m[1] = ptTri0(1);
	  m[2] = ptTri0(0) * ptTri0(0) + ptTri0(1) * ptTri0(1);
	  m[3] = 1.0;
	  m[4] = ptTri1(0);
	  m[5] = ptTri1(1);
	  m[6] = ptTri1(0) * ptTri1(0) + ptTri1(1) * ptTri1(1);
	  m[7] = 1.0;
	  m[8] = ptTri2(0);
	  m[9] = ptTri2(1);
	  m[10] = ptTri2(0) * ptTri2(0) + ptTri2(1) * ptTri2(1);
	  m[11] = 1.0;
	  m[12] = pt(0);
	  m[13] = pt(1);
	  m[14] = pt(0) * pt(0) + pt(1) * pt(1);
	  m[15] = 1.0;

	  mDet.Set(m);

	  f = mDet.Determinant();
	}
        break;

      case atFloat:
	{
	  Point ppt[3];

	  ppt[0] = ptTri0 - pt;
	  ppt[0][2] = (ppt[0](0)*ppt[0](0)) + (ppt[0](1)*ppt[0](1));
	  ppt[1] = ptTri1 - pt;
	  ppt[1][2] = (ppt[1](0)*ppt[1](0)) + (ppt[1](1)*ppt[1](1));
	  ppt[2] = ptTri2 - pt;
	  ppt[2][2] = (ppt[2](0)*ppt[2](0)) + (ppt[2](1)*ppt[2](1));

	  f = ( ( (ppt[0](0)*ppt[1](1)*ppt[2](2)) + 
		  (ppt[1](0)*ppt[2](1)*ppt[0](2)) + 
		  (ppt[2](0)*ppt[0](1)*ppt[1](2)) ) -
		( (ppt[0](0)*ppt[2](1)*ppt[1](2)) + 
		  (ppt[1](0)*ppt[0](1)*ppt[2](2)) + 
		  (ppt[2](0)*ppt[1](1)*ppt[0](2)) ) );
	}
        break;
	
      case atExact:
	{
	  double ppd0[2], ppd1[2], ppd2[2], ppd3[2];
	  
	  ppd0[0] = ptTri0(0);
	  ppd0[1] = ptTri0(1);
	  ppd1[0] = ptTri1(0);
	  ppd1[1] = ptTri1(1);
	  ppd2[0] = ptTri2(0);
	  ppd2[1] = ptTri2(1);
	  ppd3[0] = pt(0);
	  ppd3[1] = pt(1);
	  
	  f = incircleexact(ppd0, ppd1, ppd2, ppd3);
	}
        break;

    default:
      ASSERT(FALSE);
      break;
    }
#endif // COMPILER_PROBLEM

  return f > 0.0;
}

BOOL QEMesh::Counterclockwise(const Point &ptTri0, const Point &ptTri1, const Point &ptTri2)
{
  return CounterclockwiseValue(ptTri0, ptTri1, ptTri2) > 0.0;
}

DOUBLE QEMesh::CounterclockwiseValue(const Point &ptTri0, const Point &ptTri1, 
				     const Point &ptTri2)
{
  DOUBLE d;

  d = 0.0;

#ifndef COMPILER_PROBLEM
  switch ( m_eAccuracy )
    {
    case atFloatNaive:
      {
	d = ( (ptTri0(0) * ptTri1(1) + ptTri1(0) * ptTri2(1) + ptTri2(0) * ptTri0(1)) -
	      (ptTri0(0) * ptTri2(1) + ptTri1(0) * ptTri0(1) + ptTri2(0) * ptTri1(1)) );
      }
      break;

    case atFloat:
      {
	Point ppt[2];
	ppt[0] = ptTri1 - ptTri0;
	ppt[1] = ptTri2 - ptTri0;
	d = ( (ppt[0](0) * ppt[1](1)) - (ppt[1](0) * ppt[0](1)) );
      }
      break;

    case atExact:
      {
	double ppd0[2], ppd1[2], ppd2[2];

	ppd0[0] = ptTri0(0);
	ppd0[1] = ptTri0(1);
	ppd1[0] = ptTri1(0);
	ppd1[1] = ptTri1(1);
	ppd2[0] = ptTri2(0);
	ppd2[1] = ptTri2(1);

	d = orient2dexact(ppd0, ppd1, ppd2);
      }
      break;

    default:
      ASSERT( FALSE );
      break;
    }
#endif // COMPILER_PROBLEM

  return d;
}

BOOL QEMesh::RightOfEdge(QEEdge *pe, const Point &pt)
{
  return Counterclockwise(pt, pe->GetEdge()->GetVertex()->GetPoint(),
			  pe->GetVertex()->GetPoint());
			  
}

BOOL QEMesh::LeftOfEdge(QEEdge *pe, const Point &pt)
{
  return Counterclockwise(pt, pe->GetVertex()->GetPoint(), 
			  pe->GetEdge()->GetVertex()->GetPoint()); 
}

QEFacet *QEMesh::FindFacet(const Point &pt)
{
  QEEdge *pe;

  pe = FindFacetEdge(pt);

  return pe->GetFacet();
}

QEEdge *QEMesh::FindFacetEdge(const Point &pt)
{
  QEEdge *pe, *peCurr;
  BOOL bDone;

  pe = m_plistQEEdge->PeekLast();
  ASSERT( pe != NULL );

  if ( pe->GetFacet() == m_pfaNULL )
    {
      pe = pe->GetEdge();
    }
  if ( RightOfEdge(pe, pt) )
    {
      pe = pe->GetEdge();
    }

  bDone = FALSE;
  while ( !bDone )
    {
      if ( pe->GetFacet() == m_pfaNULL )
	{
	  // We are outside of the convex hull
	  break;
	}

      bDone = TRUE;
      for ( peCurr = pe->GetEdgeNext(0);
	    pe != peCurr;
	    peCurr = peCurr->GetEdgeNext(0) )
	{
	  if ( RightOfEdge(peCurr, pt) )
	    {
	      pe = peCurr->GetEdge();
	      bDone = FALSE;
	      break;
	    }
	}
    }

  return pe;
}

Point QEMesh::CalculateCircumcenter(const Point &pt0, const Point &pt1, const Point &pt2, BOOL bRadiusInZ/* = FALSE*/)
{
  Point pt;
  Vector v;

  FLOAT fDet;
  FLOAT pfMag[3];

#ifndef COMPILER_PROBLEM
  switch ( m_eAccuracy )
    {
      case atFloatNaive:
	{
	  fDet = ( (pt0(0) * pt1(1) + pt1(0) * pt2(1) + pt2(0) * pt0(1)) -
		   (pt0(0) * pt2(1) + pt1(0) * pt0(1) + pt2(0) * pt1(1)) );
	  
	  pfMag[0] = pt0(0) * pt0(0) + pt0(1) * pt0(1);
	  pfMag[1] = pt1(0) * pt1(0) + pt1(1) * pt1(1);
	  pfMag[2] = pt2(0) * pt2(0) + pt2(1) * pt2(1);
	  
	  pt[0] = ( pfMag[0] * (pt1(1) - pt2(1)) + 
		    pfMag[1] * (pt2(1) - pt0(1)) + 
		    pfMag[2] * (pt0(1) - pt1(1)) ) / (2.0 * fDet);
	  pt[1] = ( pfMag[0] * (pt1(0) - pt2(0)) + 
		    pfMag[1] * (pt2(0) - pt0(0)) + 
		    pfMag[2] * (pt0(0) - pt1(0)) ) / (-2.0 * fDet);
	}
        break;

      case atFloat:
      case atExact:
	{
	  Vector pv[2];

	  // Translate one point to the origin
	  pv[0] = pt1 - pt0;
	  pv[1] = pt2 - pt0;

	  fDet = ( (pv[0](0) * pv[1](1)) - (pv[1](0) * pv[0](1)) );

	  pfMag[0] = pv[0](0) * pv[0](0) + pv[0](1) * pv[0](1);
	  pfMag[1] = pv[1](0) * pv[1](0) + pv[1](1) * pv[1](1);
	  
	  v.Null();
	  v[0] = ( (pfMag[0] * pv[1](1)) - (pfMag[1] * pv[0](1)) ) / (2.0 * fDet);
	  v[1] = ( (pfMag[0] * pv[1](0)) - (pfMag[1] * pv[0](0)) ) / (-2.0 * fDet);

	  // Translate the answer back out away from the origin
	  pt = pt0 + v;
	}
        break;

    default:
      ASSERT( FALSE );
      break;
    }
#endif // COMPILER_PROBLEM

  if ( bRadiusInZ )
    {
      v = pt0 - pt;
      pt[Z] = v.Magnitude();
    }

  return pt;
}

Point QEMesh::CalculateInscribedCircumcenter(const Point &pt0, const Point &pt1, const Point &pt2, BOOL bRadiusInZ/* = FALSE*/)
{
  Point pt;
  Vector pvV[4], pvM[4];
  Matrix m;

  FLOAT fDen, fX, fY, fR;
  UINT_32 i, j;

  switch ( m_eAccuracy )
    {
      case atFloatNaive:
	{
	  pvV[0] = pt0; 
	  pvV[1] = pt1; 
	  pvV[2] = pt2; 
	  pvV[3].Null();
	}
        break;

      case atFloat:
      case atExact:
	{
	  // Translate one point to the origin
	  pvV[0].Null();
	  pvV[1] = pt1 - pt0;
	  pvV[2] = pt2 - pt0;
	  pvV[3] = pt0;
	  pvV[3][3] = 0.0;
	}
        break;

    default:
      ASSERT( FALSE );
      break;
    }

  // The A coeffs
  j = 2;
  for ( i = 0; i < 3; i++ )
    {
      pvM[0][j] = pvV[i][1] - pvV[j][1];
      j = i;
    }
  // The B coeffs
  j = 2;
  for ( i = 0; i < 3; i++ )
    {
      pvM[1][j] = -(pvV[i][0] - pvV[j][0]);
      j = i;
    }
  // The R coeffs
  for ( i = 0; i < 3; i++ )
    {
      pvM[2][i] = sqrt(pvM[0][i]*pvM[0][i] + pvM[1][i]*pvM[1][i]);
    }
  // The Constants
  j = 2;
  for ( i = 0; i < 3; i++ )
    {
      pvM[3][j] = pvV[j][0]*pvV[i][1] - pvV[i][0]*pvV[j][1];
      j = i;
    }

  //////////////////////////////////////////////////////////////////////
  // Use Kramer's Rule
  
  pt.Null();

  ////////////////////////////////////////
  // The Denominator Term

  m.SetIdentity();
  for ( i = 0; i < 3; i++ )
    {
      m.Set(i, 0, pvM[0][i]);
    }
  for ( i = 0; i < 3; i++ )
    {
      m.Set(i, 1, pvM[1][i]);
    }
  for ( i = 0; i < 3; i++ )
    {
      m.Set(i, 2, pvM[2][i]);
    }

  fDen = m.Determinant();

  if ( fDen == 0.0 )
    {
      ASSERT( FALSE );
      return pt;
    }

  ////////////////////////////////////////
  // The X Term

  for ( i = 0; i < 3; i++ )
    {
      m.Set(i, 0, pvM[3][i]);
    }

  fX = m.Determinant();
  pt[0] = fX / fDen;

  ////////////////////////////////////////
  // The Y Term

  for ( i = 0; i < 3; i++ )
    {
      m.Set(i, 0, pvM[0][i]);
    }
  for ( i = 0; i < 3; i++ )
    {
      m.Set(i, 1, pvM[3][i]);
    }

  fY = m.Determinant();

  pt[1] = fY / fDen;


  ////////////////////////////////////////
  // The R Term

  if ( bRadiusInZ )
    {
      for ( i = 0; i < 3; i++ )
	{
	  m.Set(i, 1, pvM[1][i]);
	}
      for ( i = 0; i < 3; i++ )
	{
	  m.Set(i, 2, pvM[3][i]);
	}
      
      fR = m.Determinant();
      
      pt[2] = fR / fDen;
    }

  pt += pvV[3];

  return pt;
}

BOOL QEMesh::PointInEdgeZone(const Point &pt, const Point &ptE0, const Point &ptE1)
{
  Vector v, u;
  FLOAT fDot;

  v = ptE1 - ptE0;
  u = pt - ptE0;

  fDot = (v(0)*u(0) + v(1)*u(1));
  if ( fDot < -s_cfEpsilonEdgeZone )
    {
      return FALSE;
    }

  v = ptE0 - ptE1;
  u = pt - ptE1;

  fDot = (v(0)*u(0) + v(1)*u(1));
  if ( fDot < -s_cfEpsilonEdgeZone )
    {
      return FALSE;
    }

  return TRUE;
}

FLOAT QEMesh::DistanceFromEdgeZone(const Point &pt, const Point &ptE0, const Point &ptE1)
{
  Vector v, vPlane;
  FLOAT fDot;

  ASSERT( ptE0(3) != 0.0 || ptE1(3) != 0.0 );

  if ( ptE0(3) == 0.0 ) 
    {
      vPlane = ptE0;
      vPlane = -vPlane;
      vPlane[2] = 0.0;
      vPlane[3] = 0.0;
      if ( !vPlane.Normalize() )
	{
	  v = pt - ptE1;
	  v[2] = 0.0;
	  v[3] = 0.0;
	  return v.Magnitude();
	}
    }
  else if ( ptE1(3) == 0.0 )
    {
      vPlane = ptE1;
      vPlane[2] = 0.0;
      vPlane[3] = 0.0;
      if ( !vPlane.Normalize() )
	{
	  v = pt - ptE0;
	  v[2] = 0.0;
	  v[3] = 0.0;
	  return v.Magnitude();
	}
    }
  else
    {
      vPlane = ptE0 - ptE1;
      vPlane[2] = 0.0;
      vPlane[3] = 0.0;
      if ( !vPlane.Normalize() )
	{
	  v = pt - ptE0;
	  v[2] = 0.0;
	  v[3] = 0.0;
	  return v.Magnitude();
	}
    }

  if ( ptE0(3) != 0.0 )
    {
      v = pt - ptE0;
      v[2] = 0.0;
      v[3] = 0.0;
      fDot = vPlane.Dot(v);
      if ( fDot < 0.0 )
	{
	  return -fDot;
	}
    }

  if ( ptE1(3) != 0.0 )
    {
      vPlane = -vPlane;
      v = pt - ptE1;
      v[2] = 0.0;
      v[3] = 0.0;
      fDot = vPlane.Dot(v);
      if ( fDot < 0.0 )
	{
	  return -fDot;
	}
    }

  return 0.0;
}


Point QEMesh::CalculatePointProjectionOnLine(const Point &pt, 
					     const Point &ptE0, const Point &ptE1, 
					     BOOL bRadiusInZ/* = FALSE*/)
{
  Point ptResult;
  Vector v, v2;
  FLOAT fDen;

  ptResult.Null();

  v = ptE1 - ptE0;
  v2[0] = v(0)*v(0);
  v2[1] = v(1)*v(1);
  v2[2] = v(0)*v(1);

  //fDen = -(v2(0) + v2(1));
  fDen = v2(0) + v2(1);

  if ( fDen == 0.0 )
    {
      ASSERT( FALSE );
      return ptResult;
    }

  ptResult.Set( (pt(0)*v2(0) + ptE0(0)*v2(1) + (pt(1) - ptE0(1))*v2(2)) / fDen,
		(pt(1)*v2(1) + ptE0(1)*v2(0) + (pt(0) - ptE0(0))*v2(2)) / fDen,
		0.0 );

  if ( bRadiusInZ )
    {
      v = pt - ptResult;
      ptResult[2] = v.Magnitude();
    }

  return ptResult;
}

VOID QEMesh::CalculatePointProjectionOnLineSegment(Point &ptResult, UINT_32 &uiResults,
						   const Point &pt, 
						   const Point &ptE0, const Point &ptE1, 
						   BOOL bRadiusInZ/* = FALSE*/)
{
  Vector v, u;
  FLOAT fDot;

  ptResult = CalculatePointProjectionOnLine(pt, ptE0, ptE1, bRadiusInZ);

  u = ptResult;
  u[2] = 0.0;
  u[3] = 1.0;

  v = ptE1 - ptE0;
  v[2] = 0.0;
  v[3] = 0.0;
  v.CompletePlaneEquation(ptE1);
  fDot = v.Dot(u);
  if ( fDot > 0.0 )
    {
      ptResult.Null();
      uiResults = 0;
      return;
    }

  v = -v;
  v[2] = 0.0;
  v[3] = 0.0;
  v.CompletePlaneEquation(ptE0);
  fDot = v.Dot(u);
  if ( fDot > 0.0 )
    {
      ptResult.Null();
      uiResults = 0;
      return;
    }

  uiResults = 1;
}

Vector QEMesh::CalculateBisectorDirection(QEFacet *pfa0, BOOL bInsideContour0, 
					  QEFacet *pfa1, BOOL bInsideContour1, BOOL bBelow)
{
  Point ppt[2];
  Vector v;

  if ( pfa0->GetForwardVertex() != NULL &&
       pfa1->GetForwardVertex() != NULL )
    {
      ppt[0] = pfa0->GetForwardVertex()->GetPoint();
      ppt[1] = pfa1->GetForwardVertex()->GetPoint();

      v = ppt[1] - ppt[0];
      if ( bBelow )
	{
	  v = -v;
	}
      v.Set(-v(1), v(0), 0.0);
      v.Normalize();
    }
  else if ( pfa0->GetForwardVertex() != NULL ||
	    pfa1->GetForwardVertex() != NULL )
    {
      if ( pfa0->GetForwardEdge() != NULL )
	{
	  ppt[0] = pfa0->GetForwardEdge()->GetVertex()->GetPoint();
	  ppt[1] = pfa0->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint();
	  
	  v = ppt[1] - ppt[0];
	  if ( !bInsideContour0 )
	    {
	      v = -v;
	    }      
	}
      else 
	{
	  ASSERT( pfa1->GetForwardEdge() != NULL );

	  ppt[0] = pfa1->GetForwardEdge()->GetVertex()->GetPoint();
	  ppt[1] = pfa1->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint();
	  
	  v = ppt[1] - ppt[0];
	  if ( !bInsideContour1 )
	    {
	      v = -v;
	    }      
	}

      v.Set(-v(1), v(0), 0.0);
      v.Normalize();
    }
  else
    {
      Vector u;
      FLOAT fCross, fVMag, fUMag;

      ppt[0] = pfa0->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint();
      ppt[1] = pfa0->GetForwardEdge()->GetVertex()->GetPoint();
      v = ppt[1] - ppt[0];
      fVMag = v.Magnitude();
      v.Normalize();
      ppt[0] = pfa1->GetForwardEdge()->GetVertex()->GetPoint();
      ppt[1] = pfa1->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint();
      u = ppt[1] - ppt[0];
      fUMag = u.Magnitude();
      u.Normalize();

      fCross = u(0)*v(1) - u(1)*v(0);
      if ( fabs(fCross) <= 1e-3 )
	{
	  if ( fUMag > fVMag )
	    {
	      v = u;
	    }
	  else
	    {
	      v = -v;
	    }
	  v.Set(-v(1), v(0), 0.0);
	}
      else
	{
	  v += u;
	  v.Normalize();
	}
    }
  
  return v;
}

BOOL QEMesh::Counterclockwise(QEFacet *pfa0, QEFacet *pfa1, QEFacet *pfa2)
{
  QEFacet *ppfa[3];
  UINT_32 puiID[3], uiID, uiIDMin;
  BOOL pbEdge[3], bEdge, bResult;
  UINT_32 i;

  ppfa[0] = pfa0;
  ppfa[1] = pfa1;
  ppfa[2] = pfa2;

  for ( i = 0; i < 3; i++ )
    {
      if ( ppfa[i]->GetForwardVertex() != NULL )
	{
	  ASSERT( ppfa[i]->GetForwardEdge() == NULL );
	  puiID[i] = ppfa[i]->GetForwardVertex()->GetID();
	  pbEdge[i] = FALSE;
	}
      else
	{
	  ASSERT( ppfa[i]->GetForwardEdge() != NULL );
	  puiID[i] = ppfa[i]->GetForwardEdge()->GetVertex()->GetID();
	  pbEdge[i] = TRUE;
	}
    }  

  i = 0;
  uiIDMin = puiID[i];
  for ( i++; i < 3; i++ )
    {
      if ( puiID[i] < uiIDMin )
	{
	  uiIDMin = puiID[i];
	}
    }

  while ( puiID[0] > uiIDMin )
    {
      uiID = puiID[0];
      puiID[0] = puiID[1];
      puiID[1] = puiID[2];
      puiID[2] = uiID;

      bEdge = pbEdge[0];
      pbEdge[0] = pbEdge[1];
      pbEdge[1] = pbEdge[2];
      pbEdge[2] = bEdge;
    }

  bResult = TRUE;
  for ( i = 0; i < 2; i++ )
    {
      bResult = (bResult && ( (puiID[i] < puiID[i+1]) || 
			      ( (!pbEdge[i] && pbEdge[i+1]) && (puiID[i] == puiID[i+1]) )));
    }

  return bResult;
}

BOOL QEMesh::Counterclockwise(QEFacet *pfa0, BOOL bInsideContour0,
			      QEFacet *pfa1, BOOL bInsideContour1,
			      QEFacet *pfa2, BOOL bInsideContour2)
{
  QEFacet *ppfa[3], *pfa;
  QEVertex *ppvx[3];
  QEVertex *pppvx[3][2];
  QEEdge *ppe[3];
  BOOL pbInsideContour[3], b;
  Point pppt[3][2];
  UINT_32 i, uiEdges;

  ppfa[0] = pfa0;
  ppfa[1] = pfa1;
  ppfa[2] = pfa2;

  pbInsideContour[0] = bInsideContour0;
  pbInsideContour[1] = bInsideContour1;
  pbInsideContour[2] = bInsideContour2;

  ppvx[0] = NULL;
  ppvx[1] = NULL;
  ppvx[2] = NULL;

  ppe[0] = NULL;
  ppe[1] = NULL;
  ppe[2] = NULL;

  uiEdges = 0;
  for ( i = 0; i < 3; i++ )
    {
      if ( ppfa[i]->GetForwardEdge() != NULL )
	{
	  ASSERT( ppfa[i]->GetForwardVertex() == NULL );
	  uiEdges++;
	}
      else
	{
	  ASSERT( ppfa[i]->GetForwardVertex() != NULL );
	}
    }  

  if ( 0 < uiEdges && uiEdges < 3 )
    {
      // Swap the Edges to the front
      while ( ppfa[0]->GetForwardEdge() == NULL ||
	      ppfa[2]->GetForwardVertex() == NULL )
	{
	  pfa = ppfa[0];
	  ppfa[0] = ppfa[1];
	  ppfa[1] = ppfa[2];
	  ppfa[2] = pfa;

	  b = pbInsideContour[0];
	  pbInsideContour[0] = pbInsideContour[1];
	  pbInsideContour[1] = pbInsideContour[2];
	  pbInsideContour[2] = b;
	}
    }

  for ( i = 0; i < uiEdges; i++ )
    {
      if ( pbInsideContour[i] )
	{
	  ppe[i] = ppfa[i]->GetForwardEdge();
	}
      else
	{
	  ppe[i] = ppfa[i]->GetForwardEdge()->GetEdge();
	}
      pppt[i][0] = ppe[i]->GetVertex()->GetPoint();
      pppt[i][1] = ppe[i]->GetEdge()->GetVertex()->GetPoint();
      pppvx[i][0] = ppe[i]->GetVertex();
      pppvx[i][1] = ppe[i]->GetEdge()->GetVertex();
    }
  for ( ; i < 3; i++ )
    {
      ppvx[i] = ppfa[i]->GetForwardVertex();
      pppt[i][0] = ppvx[i]->GetPoint();
      pppt[i][1] = ppvx[i]->GetPoint();
      pppvx[i][0] = ppvx[i];
      pppvx[i][1] = ppvx[i];
    }

  /*
#ifndef COMPILER_PROBLEM
  switch ( uiEdges )
    {
    case 2:
      {
	if ( ppe[0]->GetVertex() == ppvx[2] &&
	     ppe[1]->GetEdge()->GetVertex() != ppvx[2] )
	  {
	    if ( Counterclockwise(pppt[0][0], pppt[0][1], pppt[1][1]) &&
		 Counterclockwise(pppt[1][0], pppt[1][1], pppt[0][0]) )
	      {
		return TRUE;
	      }
	  }
      }
      break;
    default:
      {
      }
      break;
    }
#endif // COMPILER_PROBLEM
    */

  return Counterclockwise(pppvx, pppt, uiEdges);
}

BOOL QEMesh::Counterclockwise(QEVertex *pppvx[3][2],
			      Point pppt[3][2], UINT_32 uiEdges)
{
  BOOL bResult;
  UINT_32 pui[3], puiMax[3];
  UINT_32 i, uiPrev, uiNext;

  for ( i = 0; i < 3; i++ )
    {
      if ( i < uiEdges )
	{
	  puiMax[i] = 2;
	}
      else
	{
	  puiMax[i] = 1;
	}
    }

  for ( pui[0] = 0; pui[0] < puiMax[0]; pui[0]++ )
    {
      for ( pui[1] = 0; pui[1] < puiMax[1]; pui[1]++ )
	{
	  for ( pui[2] = 0; pui[2] < puiMax[2]; pui[2]++ )
	    {
	      if ( Counterclockwise(pppt[0][pui[0]], pppt[1][pui[1]], pppt[2][pui[2]]) )
		{
		  bResult = TRUE;
		  uiPrev = 2;
		  for ( i = 0; bResult && i < uiEdges; i++ )
		    {
		      uiNext = i+1;
		      if ( uiNext == 3 )
			{
			  uiNext = 0;
			}

		      if ( pppvx[uiPrev][1] == pppvx[i][0] &&
			   pppvx[i][1] == pppvx[uiNext][0] )
			{
			  // The edge shares its endpoints with the two other objects
			  if ( !( ( ( pppvx[uiPrev][0] == pppvx[i][0] ||
				      !Counterclockwise(pppt[i][1], pppt[i][0], pppt[uiPrev][0]) ) &&
				    Counterclockwise(pppt[i][0], pppt[i][1], pppt[uiNext][1]) ) ||
				  ( Counterclockwise(pppt[i][0], pppt[i][1], pppt[uiPrev][0]) &&
				    ( pppvx[i][1] == pppvx[uiNext][1] || 
				      !Counterclockwise(pppt[i][1], pppt[i][0], pppt[uiNext][1]) ) ) ) )
			    {
			      bResult = FALSE;
			    }
			}
		      else if ( pppvx[uiPrev][1] == pppvx[i][0] )
			{
			  // The edge shares its endpoint with the previous object
			  if ( !( ( pppvx[uiPrev][0] == pppvx[i][0] ||
				    !Counterclockwise(pppt[i][1], pppt[i][0], pppt[uiPrev][0]) ) &&
				  ( Counterclockwise(pppt[i][0], pppt[i][1], pppt[uiNext][0]) ||
				    Counterclockwise(pppt[i][0], pppt[i][1], pppt[uiNext][1]) ) ) )
			    {
			      bResult = FALSE;
			    }
			}
		      else if ( pppvx[i][1] == pppvx[uiNext][0] )
			{
			  // The edge shares its endpoint with the next object
			  if ( !( ( Counterclockwise(pppt[i][0], pppt[i][1], pppt[uiPrev][0]) ||
				    Counterclockwise(pppt[i][0], pppt[i][1], pppt[uiPrev][1]) ) &&
				  ( pppvx[i][1] == pppvx[uiNext][1] || 
				    !Counterclockwise(pppt[i][1], pppt[i][0], pppt[uiNext][1]) ) ) )
			    {
			      bResult = FALSE;
			    }
			}
		      else
			{
			  if ( !( ( Counterclockwise(pppt[i][0], pppt[i][1], pppt[uiPrev][0]) ||
				    Counterclockwise(pppt[i][0], pppt[i][1], pppt[uiPrev][1]) ) &&
			          ( Counterclockwise(pppt[i][0], pppt[i][1], pppt[uiNext][0]) ||
				    Counterclockwise(pppt[i][0], pppt[i][1], pppt[uiNext][1]) ) ) )
			    {
			      bResult = FALSE;
			    }
			}

		      uiPrev = i;
		    }

		  if ( bResult )
		    {
		      return TRUE;
		    }
		}
	    }
	}
    }

  return FALSE;
}
      
FLOAT QEMesh::CalculateDistance(QEFacet *pfa, const Point &pt, QEVertex **ppvxEndpoint)
{
  Vector u, v;
  FLOAT fDistance;

  ASSERT( ppvxEndpoint != NULL );

  *ppvxEndpoint = NULL;
  if ( pfa->GetForwardVertex() != NULL )
    {
      v = pfa->GetForwardVertex()->GetPoint() - pt;
      v[2] = 0.0;
      fDistance = v.Magnitude();
    }
  else
    {
      ASSERT( pfa->GetForwardEdge() != NULL );

      v = (pfa->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint() - 
	   pfa->GetForwardEdge()->GetVertex()->GetPoint());
      if ( !v.Normalize() )
	{
	  v = pfa->GetForwardEdge()->GetVertex()->GetPoint() - pt;
	  v[2] = 0.0;
	  fDistance = v.Magnitude();
	}
      else
	{
	  v.Set(-v(1), v(0), 0.0);
	  v.CompletePlaneEquation(pfa->GetForwardEdge()->GetVertex()->GetPoint());
	  u = pt;
	  u[2] = 0.0;
	  u[3] = 1.0;
	  fDistance = v.Dot(u);
	  if ( fDistance < 0.0 ) 
	    {
	      fDistance = -fDistance;
	    }

	  // Check if the answer is in the valid zone for the finite edge segment
	  {
	    QEVertex *ppvx[2];
	    Vector pvPlane[2];
	    FLOAT fDot;
	    BOOL bOnEdge;
	    UINT_32 i, j;

	    ppvx[0] = pfa->GetForwardEdge()->GetVertex();
	    ppvx[1] = pfa->GetForwardEdge()->GetEdge()->GetVertex();
	  
	    i = 1;
	    for ( j = 0; j < 2; j++ )
	      {
		if ( s_bOpenEndpoint && s_pvxOpenEndpoint == ppvx[j] )
		  {
		    // This is the open endpoint so just duplicate the other side's condition
		    pvPlane[j] = ppvx[j]->GetPoint() - ppvx[i]->GetPoint();
		    pvPlane[j].Normalize();
		    pvPlane[j].CompletePlaneEquation(ppvx[i]->GetPoint());
		    i = j;
		  }
		else
		  {
		    // Normal
		    pvPlane[j] = ppvx[i]->GetPoint() - ppvx[j]->GetPoint();
		    pvPlane[j].Normalize();
		    pvPlane[j].CompletePlaneEquation(ppvx[j]->GetPoint());
		    i = j;
		  }
	      }
	  
	    v = pt;
	    v[2] = 0.0;
	    bOnEdge = TRUE;
	    for ( j = 0; j < 2; j++ )
	      {
		fDot = pvPlane[j].Dot(v);
		if ( fDot < 0.0 )
		  {
		    bOnEdge = FALSE;
		    break;
		  }
	      }
	    
	    if ( !bOnEdge )
	      {
		// The distance is the shorter of the two distances to the endpoints
		v = pfa->GetForwardEdge()->GetVertex()->GetPoint() - pt;
		v[2] = 0.0;
		fDistance = v.Magnitude();

		v = pfa->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint() - pt;
		v[2] = 0.0;
		fDot = v.Magnitude();
		
		*ppvxEndpoint = pfa->GetForwardEdge()->GetVertex();
		if ( fDot < fDistance )
		  {
		    fDistance = fDot;
		    *ppvxEndpoint = pfa->GetForwardEdge()->GetEdge()->GetVertex();
		  }
	      }
	  }
	}
    }

  return fDistance;
}
  
VOID QEMesh::CalculateVoronoiPoint(Point *pptResults, UINT_32 &uiResults, 
				   QEFacet *pfa0, QEFacet *pfa1, QEFacet *pfa2,
				   BOOL bRadiusInZ/* = FALSE*/)
{
  ASSERT( pptResults != NULL );
  ASSERT( pfa0 != NULL && pfa1 != NULL && pfa2 != NULL );

  QEFacet *ppfa[3], *pfa;
  UINT_32 uiEdges, uiVertices, i;

  ppfa[0] = pfa0;
  ppfa[1] = pfa1;
  ppfa[2] = pfa2;

  uiEdges = 0;
  uiVertices = 0;
  for ( i = 0; i < 3; i++ )
    {
      if ( ppfa[i]->GetForwardVertex() != NULL )
	{
	  ASSERT( ppfa[i]->GetForwardEdge() == NULL );
	  uiVertices++;
	}
      else
	{
	  ASSERT( ppfa[i]->GetForwardEdge() != NULL );
	  uiEdges++;
	}
    }

  switch ( uiEdges )
    {
    case 0:
      {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	fprintf(stderr, "START: VVV\n");

	fprintf(stderr, "  VVV: V0: ( %.3f, %.3f )\n", 
	       ppfa[0]->GetForwardVertex()->GetPoint()(X),
	       ppfa[0]->GetForwardVertex()->GetPoint()(Y));
	fprintf(stderr, "  VVV: V1: ( %.3f, %.3f )\n", 
	       ppfa[1]->GetForwardVertex()->GetPoint()(X),
	       ppfa[1]->GetForwardVertex()->GetPoint()(Y));
	fprintf(stderr, "  VVV: V2: ( %.3f, %.3f )\n", 
	       ppfa[2]->GetForwardVertex()->GetPoint()(X),
	       ppfa[2]->GetForwardVertex()->GetPoint()(Y));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	CalculateVoronoiPointVVV(pptResults, uiResults, 
				 ppfa[0]->GetForwardVertex()->GetPoint(),
				 ppfa[1]->GetForwardVertex()->GetPoint(),
				 ppfa[2]->GetForwardVertex()->GetPoint(),
				 bRadiusInZ);

	if (  uiResults > 0 && pptResults[0](2) > s_cfVoronoiInfinity )
	  {
	    uiResults = 0;
	    fprintf(stderr, "  VVV: Point at Infinity: ( %.6f, %.6f, %.6f )\n", 
		   pptResults[0](X), pptResults[0](Y), pptResults[0](Z));
	  }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	if ( uiResults == 0 )
	  {
	    fprintf(stderr, "  VVV: No answers\n");
	  }
	else
	  {
	    fprintf(stderr, "  VVV: Center: ( %.6f, %.6f, %.6f )\n", 
		   pptResults[0](X), pptResults[0](Y), pptResults[0](Z));
	  }

	fprintf(stderr, "END  : VVV\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
      }
      break;

    case 1:
      {
	while ( ppfa[2]->GetForwardEdge() == NULL )
	  {
	    pfa = ppfa[0];
	    ppfa[0] = ppfa[1];
	    ppfa[1] = ppfa[2];
	    ppfa[2] = pfa;
	  }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	fprintf(stderr, "START: VVE\n");

	fprintf(stderr, "  VVE: V0: ( %.3f, %.3f )\n", 
	       ppfa[0]->GetForwardVertex()->GetPoint()(X),
	       ppfa[0]->GetForwardVertex()->GetPoint()(Y));
	fprintf(stderr, "  VVE: V1: ( %.3f, %.3f )\n", 
	       ppfa[1]->GetForwardVertex()->GetPoint()(X),
	       ppfa[1]->GetForwardVertex()->GetPoint()(Y));
	fprintf(stderr, "  VVE: E2: ( %.3f, %.3f ) -> ( %.3f, %.3f )\n", 
	       ppfa[2]->GetForwardEdge()->GetVertex()->GetPoint()(X),
	       ppfa[2]->GetForwardEdge()->GetVertex()->GetPoint()(Y),
	       ppfa[2]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint()(X),
	       ppfa[2]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint()(Y));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	ASSERT( !( ppfa[0]->GetForwardVertex() == ppfa[2]->GetForwardEdge()->GetEdge()->GetVertex() &&
		   ppfa[1]->GetForwardVertex() == ppfa[2]->GetForwardEdge()->GetVertex() ) );

	if ( ppfa[0]->GetForwardVertex() == ppfa[2]->GetForwardEdge()->GetEdge()->GetVertex() )
	  {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	    fprintf(stderr, "  VVE: V0 on E2\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	    CalculateVoronoiPointVE(pptResults, uiResults, 
				    ppfa[1]->GetForwardVertex()->GetPoint(),
				    ppfa[2]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint(),
				    ppfa[2]->GetForwardEdge()->GetVertex()->GetPoint(),
				    bRadiusInZ);
	  }
	else if ( ppfa[1]->GetForwardVertex() == ppfa[2]->GetForwardEdge()->GetVertex() )
	  {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	    fprintf(stderr, "  VVE: V1 on E2\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	    CalculateVoronoiPointVE(pptResults, uiResults, 
				    ppfa[0]->GetForwardVertex()->GetPoint(),
				    ppfa[2]->GetForwardEdge()->GetVertex()->GetPoint(),
				    ppfa[2]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint(),
				    bRadiusInZ);
	  }
	else
	  {
	    CalculateVoronoiPointVVE(pptResults, uiResults, 
				     ppfa[0]->GetForwardVertex()->GetPoint(),
				     ppfa[0]->GetForwardVertex()->GetPoint(),
				     ppfa[0]->GetForwardVertex()->GetPoint(),
				     ppfa[1]->GetForwardVertex()->GetPoint(),
				     ppfa[1]->GetForwardVertex()->GetPoint(),
				     ppfa[1]->GetForwardVertex()->GetPoint(),
				     ppfa[2]->GetForwardEdge()->GetVertex()->GetPoint(),
				     ppfa[2]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint(),
				     bRadiusInZ);
	  }

	if ( uiResults > 0 && pptResults[0](2) > s_cfVoronoiInfinity )
	  {
	    uiResults = 0;
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	    fprintf(stderr, "  VVE: Point at Infinity: ( %.6f, %.6f, %.6f )\n", 
		   pptResults[0](X), pptResults[0](Y), pptResults[0](Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	  }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	if ( uiResults == 0 )
	  {
	    fprintf(stderr, "  VVE: No answers\n");
	  }
	else
	  {
	    fprintf(stderr, "  VVE: Center: ( %.6f, %.6f, %.6f )\n", 
		   pptResults[0](X), pptResults[0](Y), pptResults[0](Z));
	  }

	fprintf(stderr, "END  : VVE\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
      }
      break;
    
    case 2:
      {
	while ( ppfa[0]->GetForwardVertex() == NULL )
	  {
	    pfa = ppfa[0];
	    ppfa[0] = ppfa[1];
	    ppfa[1] = ppfa[2];
	    ppfa[2] = pfa;
	  }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	fprintf(stderr, "START: VEE\n");

	fprintf(stderr, "  VEE: V0: ( %.3f, %.3f )\n", 
	       ppfa[0]->GetForwardVertex()->GetPoint()(X),
	       ppfa[0]->GetForwardVertex()->GetPoint()(Y));
	fprintf(stderr, "  VEE: E1: ( %.3f, %.3f ) -> ( %.3f, %.3f )\n", 
	       ppfa[1]->GetForwardEdge()->GetVertex()->GetPoint()(X),
	       ppfa[1]->GetForwardEdge()->GetVertex()->GetPoint()(Y),
	       ppfa[1]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint()(X),
	       ppfa[1]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint()(Y));
	fprintf(stderr, "  VEE: E2: ( %.3f, %.3f ) -> ( %.3f, %.3f )\n", 
	       ppfa[2]->GetForwardEdge()->GetVertex()->GetPoint()(X),
	       ppfa[2]->GetForwardEdge()->GetVertex()->GetPoint()(Y),
	       ppfa[2]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint()(X),
	       ppfa[2]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint()(Y));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	if ( ppfa[0]->GetForwardVertex() == ppfa[1]->GetForwardEdge()->GetVertex() &&
	     ppfa[0]->GetForwardVertex() == ppfa[2]->GetForwardEdge()->GetEdge()->GetVertex() )
	  {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	    fprintf(stderr, "  VEE: V0 on E1 and E2\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	    uiResults = 1;
	    pptResults[0] = ppfa[0]->GetForwardVertex()->GetPoint();
	    pptResults[0][Z] = 0.0; 
	  }
	else if ( ppfa[0]->GetForwardVertex() == ppfa[1]->GetForwardEdge()->GetVertex() )
	  {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	    fprintf(stderr, "  VEE: V0 on E1\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	    CalculateVoronoiPointEE(pptResults, uiResults, 
				    ppfa[1]->GetForwardEdge()->GetVertex()->GetPoint(),
				    ppfa[1]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint(),
				    ppfa[2]->GetForwardEdge()->GetVertex()->GetPoint(),
				    ppfa[2]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint(),
				    bRadiusInZ, TRUE);
	  }
	else if ( ppfa[0]->GetForwardVertex() == ppfa[2]->GetForwardEdge()->GetEdge()->GetVertex() )
	  {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	    fprintf(stderr, "  VEE: V0 on E2\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	    CalculateVoronoiPointEE(pptResults, uiResults, 
				    ppfa[2]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint(),
				    ppfa[2]->GetForwardEdge()->GetVertex()->GetPoint(),
				    ppfa[1]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint(),
				    ppfa[1]->GetForwardEdge()->GetVertex()->GetPoint(),
				    bRadiusInZ, FALSE);
	  }
	else
	  {
	    CalculateVoronoiPointVEE(pptResults, uiResults, 
				     ppfa[0]->GetForwardVertex()->GetPoint(),
				     ppfa[0]->GetForwardVertex()->GetPoint(),
				     ppfa[0]->GetForwardVertex()->GetPoint(),
				     ppfa[1]->GetForwardEdge()->GetVertex()->GetPoint(),
				     ppfa[1]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint(),
				     ppfa[2]->GetForwardEdge()->GetVertex()->GetPoint(),
				     ppfa[2]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint(),
				     bRadiusInZ);
	  }

	if ( uiResults > 0 && pptResults[0](2) > s_cfVoronoiInfinity )
	  {
	    uiResults = 0;
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	    fprintf(stderr, "  VEE: Point at Infinity: ( %.6f, %.6f, %.6f )\n", 
		   pptResults[0](X), pptResults[0](Y), pptResults[0](Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	  }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	if ( uiResults == 0 )
	  {
	    fprintf(stderr, "  VEE: No answers\n");
	  }
	else
	  {
	    fprintf(stderr, "  VEE: Center: ( %.6f, %.6f, %.6f )\n", 
		   pptResults[0](X), pptResults[0](Y), pptResults[0](Z));
	  }

	fprintf(stderr, "END  : VEE\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
      }
      break;
    
    case 3:
      {
	//UINT_32 j, uiMatchingVertices;

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	fprintf(stderr, "START: EEE\n");

	fprintf(stderr, "  EEE: E0: ( %.3f, %.3f ) -> ( %.3f, %.3f )\n", 
	       ppfa[0]->GetForwardEdge()->GetVertex()->GetPoint()(X),
	       ppfa[0]->GetForwardEdge()->GetVertex()->GetPoint()(Y),
	       ppfa[0]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint()(X),
	       ppfa[0]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint()(Y));
	fprintf(stderr, "  EEE: E1: ( %.3f, %.3f ) -> ( %.3f, %.3f )\n", 
	       ppfa[1]->GetForwardEdge()->GetVertex()->GetPoint()(X),
	       ppfa[1]->GetForwardEdge()->GetVertex()->GetPoint()(Y),
	       ppfa[1]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint()(X),
	       ppfa[1]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint()(Y));
	fprintf(stderr, "  EEE: E2: ( %.3f, %.3f ) -> ( %.3f, %.3f )\n", 
	       ppfa[2]->GetForwardEdge()->GetVertex()->GetPoint()(X),
	       ppfa[2]->GetForwardEdge()->GetVertex()->GetPoint()(Y),
	       ppfa[2]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint()(X),
	       ppfa[2]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint()(Y));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
/*
	uiMatchingVertices = 0;
	j = 2;
	for ( i = 0; i < 3; i++ )
	  {
	    if ( ppfa[j]->GetForwardEdge()->GetEdge()->GetVertex() == 
		 ppfa[i]->GetForwardEdge()->GetVertex() )
	      {
		uiMatchingVertices++;
	      }
	    j = i;
	  }

	switch ( uiMatchingVertices )
	  {
	    //case 1:
	  case 6:
	    {
	      while ( ppfa[0]->GetForwardEdge()->GetEdge()->GetVertex() != 
		      ppfa[1]->GetForwardEdge()->GetVertex() )
		{
		  pfa = ppfa[0];
		  ppfa[0] = ppfa[1];
		  ppfa[1] = ppfa[2];
		  ppfa[2] = pfa;
		}

	      fprintf(stderr, "  EEE: E0_1 == E1_0\n");
	      CalculateVoronoiPointEvEE(pptResults, uiResults, 
					ppfa[0]->GetForwardEdge()->GetVertex()->GetPoint(),
					ppfa[1]->GetForwardEdge()->GetVertex()->GetPoint(),
					ppfa[1]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint(),
					ppfa[2]->GetForwardEdge()->GetVertex()->GetPoint(),
					ppfa[2]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint(),
					bRadiusInZ);
	    }
	    break;
	  default:
	    {
	      CalculateVoronoiPointEEE(pptResults, uiResults,
				       ppfa[0]->GetForwardEdge()->GetVertex()->GetPoint(),
				       ppfa[0]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint(),
				       ppfa[1]->GetForwardEdge()->GetVertex()->GetPoint(),
				       ppfa[1]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint(),
				       ppfa[2]->GetForwardEdge()->GetVertex()->GetPoint(),
				       ppfa[2]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint(),
				       bRadiusInZ);
	    }
	    break;
	  }
	  */

	      CalculateVoronoiPointEEE(pptResults, uiResults, 
				       ppfa[0]->GetForwardEdge()->GetVertex()->GetPoint(),
				       ppfa[0]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint(),
				       ppfa[1]->GetForwardEdge()->GetVertex()->GetPoint(),
				       ppfa[1]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint(),
				       ppfa[2]->GetForwardEdge()->GetVertex()->GetPoint(),
				       ppfa[2]->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint(),
				       bRadiusInZ);

	if ( uiResults > 0 && pptResults[0](2) > s_cfVoronoiInfinity )
	  {
	    uiResults = 0;
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	    fprintf(stderr, "  EEE: Point at Infinity: ( %.6f, %.6f, %.6f )\n", 
		   pptResults[0](X), pptResults[0](Y), pptResults[0](Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	  }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	if ( uiResults == 0 )
	  {
	    fprintf(stderr, "  EEE: No answers\n");
	  }
	else
	  {
	    fprintf(stderr, "  EEE: Center: ( %.6f, %.6f, %.6f )\n", 
		   pptResults[0](X), pptResults[0](Y), pptResults[0](Z));
	  }

	fprintf(stderr, "END  : EEE\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
      }
      break;
    
    default:
      {
	ASSERT( FALSE );
      }
      break;
    }
}

VOID QEMesh::CalculateVoronoiPoint(Point *pptResults, UINT_32 &uiResults, 
				   QEFacet *pfa0, BOOL bInsideContour0, 
				   QEFacet *pfa1, BOOL bInsideContour1, 
				   QEFacet *pfa2, BOOL bInsideContour2, 
				   BOOL bRadiusInZ/* = FALSE*/)
{
  ASSERT( pptResults != NULL );
  ASSERT( pfa0 != NULL && pfa1 != NULL && pfa2 != NULL );

  QEFacet *ppfa[3], *pfa;
  QEVertex *ppvx[3];
  QEEdge *ppe[3];
  BOOL pbInsideContour[3], b;
  UINT_32 uiEdges, uiVertices, i;

  ppfa[0] = pfa0;
  ppfa[1] = pfa1;
  ppfa[2] = pfa2;

  pbInsideContour[0] = bInsideContour0;
  pbInsideContour[1] = bInsideContour1;
  pbInsideContour[2] = bInsideContour2;

  ppvx[0] = NULL;
  ppvx[1] = NULL;
  ppvx[2] = NULL;

  ppe[0] = NULL;
  ppe[1] = NULL;
  ppe[2] = NULL;

  uiEdges = 0;
  uiVertices = 0;
  for ( i = 0; i < 3; i++ )
    {
      if ( ppfa[i]->GetForwardVertex() != NULL )
	{
	  ASSERT( ppfa[i]->GetForwardEdge() == NULL );
	  uiVertices++;
	}
      else
	{
	  ASSERT( ppfa[i]->GetForwardEdge() != NULL );
	  uiEdges++;
	}
    }

  switch ( uiEdges )
    {
    case 0:
      {
	for ( i = 0; i < 3; i++ )
	  {
	    ppvx[i] = ppfa[i]->GetForwardVertex();
	  }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	fprintf(stderr, "START: VVV\n");

	fprintf(stderr, "  VVV: V0: ( %.3f, %.3f )\n", 
	       ppvx[0]->GetPoint()(X),
	       ppvx[0]->GetPoint()(Y));
	fprintf(stderr, "  VVV: V1: ( %.3f, %.3f )\n", 
	       ppvx[1]->GetPoint()(X),
	       ppvx[1]->GetPoint()(Y));
	fprintf(stderr, "  VVV: V2: ( %.3f, %.3f )\n", 
	       ppvx[2]->GetPoint()(X),
	       ppvx[2]->GetPoint()(Y));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	CalculateVoronoiPointVVV(pptResults, uiResults,
				 ppvx[0]->GetPoint(),
				 ppvx[1]->GetPoint(),
				 ppvx[2]->GetPoint(),
				 bRadiusInZ);

	if ( uiResults > 0 && pptResults[0](2) > s_cfVoronoiInfinity )
	  {
	    uiResults = 0;
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	    fprintf(stderr, "  VVV: Point at Infinity: ( %.6f, %.6f, %.6f )\n", 
		   pptResults[0](X), pptResults[0](Y), pptResults[0](Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	  }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	if ( uiResults == 0 )
	  {
	    fprintf(stderr, "  VVV: No answers\n");
	  }
	else
	  {
	    fprintf(stderr, "  VVV: Center: ( %.6f, %.6f, %.6f )\n", 
		   pptResults[0](X), pptResults[0](Y), pptResults[0](Z));
	  }

	fprintf(stderr, "END  : VVV\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
      }
      break;

    case 1:
      {
	while ( ppfa[2]->GetForwardEdge() == NULL )
	  {
	    pfa = ppfa[0];
	    ppfa[0] = ppfa[1];
	    ppfa[1] = ppfa[2];
	    ppfa[2] = pfa;

	    b = pbInsideContour[0];
	    pbInsideContour[0] = pbInsideContour[1];
	    pbInsideContour[1] = pbInsideContour[2];
	    pbInsideContour[2] = b;
	  }

	for ( i = 0; i < 2; i++ )
	  {
	    ppvx[i] = ppfa[i]->GetForwardVertex();
	  }
	for ( i = 2; i < 3; i++ )
	  {
	    if ( pbInsideContour[i] )
	      {
		ppe[i] = ppfa[i]->GetForwardEdge();
	      }
	    else
	      {
		ppe[i] = ppfa[i]->GetForwardEdge()->GetEdge();
	      }
	  }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	fprintf(stderr, "START: VVE\n");

	fprintf(stderr, "  VVE: V0: ( %.3f, %.3f )\n", 
	       ppvx[0]->GetPoint()(X),
	       ppvx[0]->GetPoint()(Y));
	fprintf(stderr, "  VVE: V1: ( %.3f, %.3f )\n", 
	       ppvx[1]->GetPoint()(X),
	       ppvx[1]->GetPoint()(Y));
	fprintf(stderr, "  VVE: E2: ( %.3f, %.3f ) -> ( %.3f, %.3f )\n", 
	       ppe[2]->GetVertex()->GetPoint()(X),
	       ppe[2]->GetVertex()->GetPoint()(Y),
	       ppe[2]->GetEdge()->GetVertex()->GetPoint()(X),
	       ppe[2]->GetEdge()->GetVertex()->GetPoint()(Y));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	ASSERT( !( ppvx[0] == ppe[2]->GetEdge()->GetVertex() &&
		   ppvx[1] == ppe[2]->GetVertex() ) );
	
	if ( ppvx[0] == ppe[2]->GetEdge()->GetVertex() )
	  {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	    fprintf(stderr, "  VVE: V0 on E2\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	    CalculateVoronoiPointVE(pptResults, uiResults, 
				    ppvx[1]->GetPoint(),
				    ppe[2]->GetEdge()->GetVertex()->GetPoint(),
				    ppe[2]->GetVertex()->GetPoint(),
				    bRadiusInZ);
	  }
	else if ( ppvx[1] == ppe[2]->GetVertex() )
	  {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	    fprintf(stderr, "  VVE: V1 on E2\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	    CalculateVoronoiPointVE(pptResults, uiResults, 
				    ppvx[0]->GetPoint(),
				    ppe[2]->GetVertex()->GetPoint(),
				    ppe[2]->GetEdge()->GetVertex()->GetPoint(),
				    bRadiusInZ);
	  }
	else
	  {
	    QEEdge *ppeVertexNeighbors[2];

	    for ( i = 0; i < 2; i++ )
	      {
		for ( ppeVertexNeighbors[i] = ppvx[i]->GetEdge(); 
		      !( ppeVertexNeighbors[i]->GetFacet() != GetNullFacet() &&
			 ppeVertexNeighbors[i]->GetFacet()->GetForwardVertex() == ppvx[i] );
		      ppeVertexNeighbors[i] = ppeVertexNeighbors[i]->GetEdgeNext(1) )
		  {
		  }
	      }

	    CalculateVoronoiPointVVE(pptResults, uiResults, 
				     ppeVertexNeighbors[0]->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge()->GetVertex()->GetPoint(),
				     ppvx[0]->GetPoint(),
				     ppeVertexNeighbors[0]->GetEdgePrev(1)->GetEdge()->GetVertex()->GetPoint(),
				     ppeVertexNeighbors[1]->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge()->GetVertex()->GetPoint(),
				     ppvx[1]->GetPoint(),
				     ppeVertexNeighbors[1]->GetEdgePrev(1)->GetEdge()->GetVertex()->GetPoint(),
				     ppe[2]->GetVertex()->GetPoint(),
				     ppe[2]->GetEdge()->GetVertex()->GetPoint(),
				     bRadiusInZ);
	  }


	if ( uiResults > 0 && pptResults[0](2) > s_cfVoronoiInfinity )
	  {
	    uiResults = 0;
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	    fprintf(stderr, "  VVE: Point at Infinity: ( %.6f, %.6f, %.6f )\n", 
		   pptResults[0](X), pptResults[0](Y), pptResults[0](Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	  }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	if ( uiResults == 0 )
	  {
	    fprintf(stderr, "  VVE: No answers\n");
	  }
	else
	  {
	    fprintf(stderr, "  VVE: Center: ( %.6f, %.6f, %.6f )\n", 
		   pptResults[0](X), pptResults[0](Y), pptResults[0](Z));
	  }

	fprintf(stderr, "END  : VVE\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
      }
      break;
    
    case 2:
      {
	while ( ppfa[0]->GetForwardVertex() == NULL )
	  {
	    pfa = ppfa[0];
	    ppfa[0] = ppfa[1];
	    ppfa[1] = ppfa[2];
	    ppfa[2] = pfa;

	    b = pbInsideContour[0];
	    pbInsideContour[0] = pbInsideContour[1];
	    pbInsideContour[1] = pbInsideContour[2];
	    pbInsideContour[2] = b;
	  }

	for ( i = 0; i < 1; i++ )
	  {
	    ppvx[i] = ppfa[i]->GetForwardVertex();
	  }
	for ( i = 1; i < 3; i++ )
	  {
	    if ( pbInsideContour[i] )
	      {
		ppe[i] = ppfa[i]->GetForwardEdge();
	      }
	    else
	      {
		ppe[i] = ppfa[i]->GetForwardEdge()->GetEdge();
	      }
	  }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	fprintf(stderr, "START: VEE\n");

	fprintf(stderr, "  VEE: V0: ( %.3f, %.3f )\n", 
	       ppvx[0]->GetPoint()(X),
	       ppvx[0]->GetPoint()(Y));
	fprintf(stderr, "  VEE: E1: ( %.3f, %.3f ) -> ( %.3f, %.3f )\n", 
	       ppe[1]->GetVertex()->GetPoint()(X),
	       ppe[1]->GetVertex()->GetPoint()(Y),
	       ppe[1]->GetEdge()->GetVertex()->GetPoint()(X),
	       ppe[1]->GetEdge()->GetVertex()->GetPoint()(Y));
	fprintf(stderr, "  VEE: E2: ( %.3f, %.3f ) -> ( %.3f, %.3f )\n", 
	       ppe[2]->GetVertex()->GetPoint()(X),
	       ppe[2]->GetVertex()->GetPoint()(Y),
	       ppe[2]->GetEdge()->GetVertex()->GetPoint()(X),
	       ppe[2]->GetEdge()->GetVertex()->GetPoint()(Y));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	if ( ppvx[0] == ppe[1]->GetVertex() &&
	     ppvx[0] == ppe[2]->GetEdge()->GetVertex() )
	  {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	    fprintf(stderr, "  VEE: V0 on E1 and E2\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	    uiResults = 1;
	    pptResults[0] = ppvx[0]->GetPoint();
	    pptResults[0][Z] = 0.0; 
	  }
	else if ( ppvx[0] == ppe[1]->GetVertex() )
	  {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	    fprintf(stderr, "  VEE: V0 on E1\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	    CalculateVoronoiPointEE(pptResults, uiResults, 
				    ppe[1]->GetVertex()->GetPoint(),
				    ppe[1]->GetEdge()->GetVertex()->GetPoint(),
				    ppe[2]->GetVertex()->GetPoint(),
				    ppe[2]->GetEdge()->GetVertex()->GetPoint(),
				    bRadiusInZ, TRUE);
	  }
	else if ( ppvx[0] == ppe[2]->GetEdge()->GetVertex() )
	  {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	    fprintf(stderr, "  VEE: V0 on E2\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	    CalculateVoronoiPointEE(pptResults, uiResults, 
				    ppe[2]->GetEdge()->GetVertex()->GetPoint(),
				    ppe[2]->GetVertex()->GetPoint(),
				    ppe[1]->GetEdge()->GetVertex()->GetPoint(),
				    ppe[1]->GetVertex()->GetPoint(),
				    bRadiusInZ, FALSE);
	  }
	else
	  {
	    QEEdge *ppeVertexNeighbors[1];

	    for ( i = 0; i < 1; i++ )
	      {
		for ( ppeVertexNeighbors[i] = ppvx[i]->GetEdge(); 
		      !( ppeVertexNeighbors[i]->GetFacet() != GetNullFacet() &&
			 ppeVertexNeighbors[i]->GetFacet()->GetForwardVertex() == ppvx[i] );
		      ppeVertexNeighbors[i] = ppeVertexNeighbors[i]->GetEdgeNext(1) )
		  {
		  }
	      }

	    CalculateVoronoiPointVEE(pptResults, uiResults, 
				     ppeVertexNeighbors[0]->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge()->GetVertex()->GetPoint(),
				     ppvx[0]->GetPoint(),
				     ppeVertexNeighbors[0]->GetEdgePrev(1)->GetEdge()->GetVertex()->GetPoint(),
				     ppe[1]->GetVertex()->GetPoint(),
				     ppe[1]->GetEdge()->GetVertex()->GetPoint(),
				     ppe[2]->GetVertex()->GetPoint(),
				     ppe[2]->GetEdge()->GetVertex()->GetPoint(),
				     bRadiusInZ);
	  }
	  
	if ( uiResults > 0 && pptResults[0](2) > s_cfVoronoiInfinity )
	  {
	    uiResults = 0;
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	    fprintf(stderr, "  VEE: Point at Infinity: ( %.6f, %.6f, %.6f )\n", 
		   pptResults[0](X), pptResults[0](Y), pptResults[0](Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	  }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	if ( uiResults == 0 )
	  {
	    fprintf(stderr, "  VEE: No answers\n");
	  }
	else
	  {
	    fprintf(stderr, "  VEE: Center: ( %.6f, %.6f, %.6f )\n", 
		   pptResults[0](X), pptResults[0](Y), pptResults[0](Z));
	  }

	fprintf(stderr, "END  : VEE\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
      }
      break;
    
    case 3:
      {
	//UINT_32 j, uiMatchingVertices;

	for ( i = 0; i < 3; i++ )
	  {
	    if ( pbInsideContour[i] )
	      {
		ppe[i] = ppfa[i]->GetForwardEdge();
	      }
	    else
	      {
		ppe[i] = ppfa[i]->GetForwardEdge()->GetEdge();
	      }
	  }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	fprintf(stderr, "START: EEE\n");

	fprintf(stderr, "  EEE: E0: ( %.3f, %.3f ) -> ( %.3f, %.3f )\n", 
	       ppe[0]->GetVertex()->GetPoint()(X),
	       ppe[0]->GetVertex()->GetPoint()(Y),
	       ppe[0]->GetEdge()->GetVertex()->GetPoint()(X),
	       ppe[0]->GetEdge()->GetVertex()->GetPoint()(Y));
	fprintf(stderr, "  EEE: E1: ( %.3f, %.3f ) -> ( %.3f, %.3f )\n", 
	       ppe[1]->GetVertex()->GetPoint()(X),
	       ppe[1]->GetVertex()->GetPoint()(Y),
	       ppe[1]->GetEdge()->GetVertex()->GetPoint()(X),
	       ppe[1]->GetEdge()->GetVertex()->GetPoint()(Y));
	fprintf(stderr, "  EEE: E2: ( %.3f, %.3f ) -> ( %.3f, %.3f )\n", 
	       ppe[2]->GetVertex()->GetPoint()(X),
	       ppe[2]->GetVertex()->GetPoint()(Y),
	       ppe[2]->GetEdge()->GetVertex()->GetPoint()(X),
	       ppe[2]->GetEdge()->GetVertex()->GetPoint()(Y));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	CalculateVoronoiPointEEE(pptResults, uiResults, 
				 ppe[0]->GetVertex()->GetPoint(),
				 ppe[0]->GetEdge()->GetVertex()->GetPoint(),
				 ppe[1]->GetVertex()->GetPoint(),
				 ppe[1]->GetEdge()->GetVertex()->GetPoint(),
				 ppe[2]->GetVertex()->GetPoint(),
				 ppe[2]->GetEdge()->GetVertex()->GetPoint(),
				 bRadiusInZ);
	
	if ( uiResults > 0 && pptResults[0](2) > s_cfVoronoiInfinity )
	  {
	    uiResults = 0;
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	    fprintf(stderr, "  EEE: Point at Infinity: ( %.6f, %.6f, %.6f )\n", 
		   pptResults[0](X), pptResults[0](Y), pptResults[0](Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	  }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	if ( uiResults == 0 )
	  {
	    fprintf(stderr, "  EEE: No answers\n");
	  }
	else
	  {
	    fprintf(stderr, "  EEE: Center: ( %.6f, %.6f, %.6f )\n", 
		   pptResults[0](X), pptResults[0](Y), pptResults[0](Z));
	  }

	fprintf(stderr, "END  : EEE\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
      }
      break;
    
    default:
      {
	ASSERT( FALSE );
      }
      break;
    }
}

VOID QEMesh::CalculateVoronoiPointVVV(Point *pptResults, UINT_32 &uiResults, 
				      const Point &pt0, 
				      const Point &pt1, 
				      const Point &pt2, 
				      BOOL bRadiusInZ/* = FALSE*/)
{
  Vector v;

  FLOAT fDet;
  FLOAT pfMag[3];

  uiResults = 1;

#ifndef COMPILER_PROBLEM
  switch ( m_eAccuracy )
    {
      case atFloatNaive:
	{
	  fDet = ( (pt0(0) * pt1(1) + pt1(0) * pt2(1) + pt2(0) * pt0(1)) -
		   (pt0(0) * pt2(1) + pt1(0) * pt0(1) + pt2(0) * pt1(1)) );
	  
	  if ( fDet == 0.0 )
	    {
	      uiResults = 0;
	      return;
	    }

	  pfMag[0] = pt0(0) * pt0(0) + pt0(1) * pt0(1);
	  pfMag[1] = pt1(0) * pt1(0) + pt1(1) * pt1(1);
	  pfMag[2] = pt2(0) * pt2(0) + pt2(1) * pt2(1);
	  
	  pptResults[0][0] = ( pfMag[0] * (pt1(1) - pt2(1)) + 
		    pfMag[1] * (pt2(1) - pt0(1)) + 
		    pfMag[2] * (pt0(1) - pt1(1)) ) / (2.0 * fDet);
	  pptResults[0][1] = ( pfMag[0] * (pt1(0) - pt2(0)) + 
		    pfMag[1] * (pt2(0) - pt0(0)) + 
		    pfMag[2] * (pt0(0) - pt1(0)) ) / (-2.0 * fDet);
	}
        break;

      case atFloat:
      case atExact:
	{
	  Vector pv[2];

	  // Translate one point to the origin
	  pv[0] = pt1 - pt0;
	  pv[1] = pt2 - pt0;

	  fDet = ( (pv[0](0) * pv[1](1)) - (pv[1](0) * pv[0](1)) );

	  if ( fDet == 0.0 )
	    {
	      uiResults = 0;
	      return;
	    }

	  pfMag[0] = pv[0](0) * pv[0](0) + pv[0](1) * pv[0](1);
	  pfMag[1] = pv[1](0) * pv[1](0) + pv[1](1) * pv[1](1);
	  
	  v.Null();
	  v[0] = ( (pfMag[0] * pv[1](1)) - (pfMag[1] * pv[0](1)) ) / (2.0 * fDet);
	  v[1] = ( (pfMag[0] * pv[1](0)) - (pfMag[1] * pv[0](0)) ) / (-2.0 * fDet);

	  // Translate the answer back out away from the origin
	  pptResults[0] = pt0 + v;
	}
        break;

    default:
      ASSERT( FALSE );
      break;
    }
#endif // COMPILER_PROBLEM

  if ( bRadiusInZ )
    {
      v = pt0 - pptResults[0];
      pptResults[0][Z] = v.Magnitude();
    }
}

VOID QEMesh::CalculateVoronoiPointVE(Point *pptResults, UINT_32 &uiResults, 
				     const Point &ptV0, 
				     const Point &ptE1_0, const Point &ptE1_1, 
				     BOOL bRadiusInZ/* = FALSE*/)
{
  Vector pvV[3], vTrans;
  Vector v, pvCoeffs[2];
  Point ptCenter;
  //Point pt;
  FLOAT fDen;

  switch ( m_eAccuracy )
    {
      case atFloatNaive:
	{
	  ptCenter.Null();
	  vTrans.Null();
	}
        break;

      case atFloat:
      case atExact:
	{
	  // Translate the average of the 6 vertices to the origin
	  ptCenter = ptE1_0;
	  vTrans = ptCenter;
	  vTrans[3] = 0.0;
	}
        break;

    default:
      ASSERT( FALSE );
      break;
    }

  pvV[0] = ptV0 - ptCenter; 
  pvV[1] = ptE1_0 - ptCenter; 
  pvV[2] = ptE1_1 - ptCenter; 

  uiResults = 0;

  //////////////////////////////
  // Make sure the vertex is on the positive side of the edge
  /*
  {
    Vector vPlane;
    FLOAT fDot;

    v = pvV[2] - pvV[0];
    vPlane.Set(-v(1), v(0), 0.0);
    pt = pvV[1];
    pt[2] = 0.0;
    pt[3] = 1.0;
    vPlane.CompletePlaneEquation(pt);
    v = pvV[0];
    v[2] = 0.0;
    v[3] = 1.0;
    fDot = vPlane.Dot(v);
    if ( fDot < s_cfDistanceEpsilon )
      {
	pptResults[0].Null();
	return;      
      }
  }
  */

  pvCoeffs[0][0] = pvV[0](0) - pvV[1](0);
  pvCoeffs[0][1] = pvV[0](1) - pvV[1](1);
  pvCoeffs[0][2] = ( (pvV[0](0)*pvV[0](0) + pvV[0](1)*pvV[0](1)) - 
		     (pvV[1](0)*pvV[1](0) + pvV[1](1)*pvV[1](1)) ) / 2.0;
  pvCoeffs[1][0] = pvV[2](0) - pvV[1](0);
  pvCoeffs[1][1] = pvV[2](1) - pvV[1](1);
  pvCoeffs[1][2] = ( (pvV[1](0)*pvV[2](0) + pvV[1](1)*pvV[2](1)) - 
		     (pvV[1](0)*pvV[1](0) + pvV[1](1)*pvV[1](1)) );

  fDen = (pvCoeffs[0][0]*pvCoeffs[1][1]) - (pvCoeffs[1][0]*pvCoeffs[0][1]);
  //if ( fDen == 0.0 )
  if ( fabs(fDen) <= s_cfEpsilonVoronoiVEVertexOnEdge )
    {
      pptResults[0].Null();
      return;
    }
  
  uiResults = 1;
  pptResults[0].Set( ((pvCoeffs[0][2]*pvCoeffs[1][1]) - (pvCoeffs[1][2]*pvCoeffs[0][1])) / fDen,
		     ((pvCoeffs[0][0]*pvCoeffs[1][2]) - (pvCoeffs[1][0]*pvCoeffs[0][2])) / fDen,
		     0.0 );

  if ( bRadiusInZ )
    {
      v = pptResults[0];
      v = v - pvV[0];
      v[2] = 0.0;
      v[3] = 0.0;
      pptResults[0][2] = v.Magnitude();
    }

  pptResults[0] += vTrans;
}

VOID QEMesh::CalculateVoronoiPointVVE(Point *pptResults, UINT_32 &uiResults, 
				      const Point &ptV0_0, const Point &ptV0, const Point &ptV0_1, 
				      const Point &ptV1_0, const Point &ptV1, const Point &ptV1_1, 
				      const Point &ptE2_0, const Point &ptE2_1, 
				      BOOL bRadiusInZ/* = FALSE*/)
{
  Vector vTrans;
#ifdef ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES
  //BOOL bCheckVertexZones;
  
  //bCheckVertexZones = TRUE;
#endif // ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES

#ifndef COMPILER_PROBLEM
  switch ( m_eAccuracy )
    {
      case atFloatNaive:
	{
	  Vector u, v2, ppvE[3][2], vTrans;
	  Point pt, ptCenter;
	  FLOAT a, b, c, d, e, f, g, h, k;
	  FLOAT fDen, fDisc;
	  FLOAT fXLeft, fXRight, fYLeft, fYRight;
	  UINT_32 i;
	  Vector v;
	  //Vector pvPlane[2];

	  // Translate the average of the 6 vertices to the origin
	  ptCenter = ptV0;
	  //ptCenter = ( (2.0*ptV0) + (2.0*ptV1) + ptE2_0 + ptE2_1 ) / 6.0;
	  ptCenter[3] = 1.0;
	  vTrans = ptCenter;
	  vTrans[3] = 0.0;

	  ppvE[0][0] = ptV0 - ptCenter; 
	  ppvE[1][0] = ptV1 - ptCenter; 
	  ppvE[2][0] = ptE2_0 - ptCenter; 
	  ppvE[2][1] = ptE2_1 - ptCenter; 

	  uiResults = 0;
  
	  v2 = ppvE[2][1] - ppvE[2][0];
	  v2.Set(-v2(1), v2(0), 0.0);
	  if ( !v2.Normalize() )
	    {
	      ASSERT( FALSE );
	      return;
	    }
	  pt = ppvE[2][0];
	  v2.CompletePlaneEquation(pt);
    
	  a = (1 - v2(0)*v2(0));
	  b = -2.0*v2(0)*v2(1); 
	  c = (1 - v2(1)*v2(1));
	  d = -2.0*(ppvE[0][0](0) + v2(0)*v2(3)); 
	  e = -2.0*(ppvE[0][0](1) + v2(1)*v2(3)); 
	  f = (ppvE[0][0](0)*ppvE[0][0](0) + ppvE[0][0](1)*ppvE[0][0](1) - v2(3)*v2(3));

	  g = 2.0*(ppvE[1][0](0) - ppvE[0][0](0));
	  h = 2.0*(ppvE[1][0](1) - ppvE[0][0](1));
	  k = (ppvE[0][0](0)*ppvE[0][0](0) + ppvE[0][0](1)*ppvE[0][0](1)) - (ppvE[1][0](0)*ppvE[1][0](0) + ppvE[1][0](1)*ppvE[1][0](1));

	  fDen = 2.0 * ( a*h*h - b*g*h + c*g*g );
	  if ( fabs(fDen) < s_cfDistanceEpsilon )
	    {
	      // I believe this means that the two vertices lie on a line parallel to the edge
	      FLOAT d2, h;

	      v = ppvE[0][0] - ppvE[1][0];
	      v.Set(-v(1), v(0), 0.0);
	      v[3] = -( v(0)*ppvE[2][0](0) + v(1)*ppvE[2][0](1) );

	      d2 = ( (ppvE[1][0](0) - ppvE[0][0](0))*(ppvE[1][0](0) - ppvE[0][0](0)) +
		     (ppvE[1][0](1) - ppvE[0][0](1))*(ppvE[1][0](1) - ppvE[0][0](1)) );
	      pt = (ppvE[1][0] + ppvE[0][0]) / 2.0;
	      u = pt;
	      u[3] = 1.0;
	      h = v.Dot(u);
	      if ( h <= 0 || d2 <= 0 )
		{
		  ASSERT( FALSE );
		  return;
		}
	      f = - (4.0*h*h - d2*d2) / (8.0*h*d2);

	      uiResults = 1;
	      pptResults[0] = pt + f*v;
	      pptResults[0][3] = 1.0;
	      if ( bRadiusInZ )
		{
		  d = sqrt(d2);
		  pptResults[0][2] = (h/d) + (d*f);
		}

	      // Translate back to the origin
	      for ( i = 0; i < uiResults; i++ )
		{
		  pptResults[i] += vTrans;
		}

	      goto end;
	    }
	  fDisc = (e*e*g*g + d*d*h*h + b*b*k*k) - 4.0*(c*f*g*g + a*f*h*h + a*c*k*k) + 4.0*(b*f*g*h + a*e*h*k + c*d*g*k) - 2.0*(b*d*h*k + d*e*g*h + b*e*g*k);
	  //if ( -m_cfEpsilon <= fDisc && fDisc <= m_cfEpsilon )
	  if ( -m_cfEpsilon <= fDisc && fDisc <= 0.0 )
	    {
	      fDisc = 0.0;
	    }
	  if ( fDisc < 0.0 ) 
	    {
	      // Imaginary Solutions
	      return;
	    }
	  fDisc = sqrt(fDisc);

	  // X
	  fXLeft = -(2.0*c*g*k - b*h*k - e*g*h + d*h*h);
	  fXRight = h*fDisc;

	  // Y
	  fYLeft = (2.0*a*h*k - b*g*k - d*g*h + e*g*g);
	  fYRight = g*fDisc;

	  if ( fXRight == 0.0 && fYRight == 0.0 )
	    {
#ifdef ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES
	      bCheckVertexZones = FALSE;
#endif // ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES

	      uiResults = 1;
	      pptResults[0].Set(fXLeft/fDen, -fYLeft/fDen, 0.0);

	      if ( bRadiusInZ )
		{
		  v2 = pptResults[0];
		  v2 -= ppvE[0][0];
		  v2[2] = 0.0;
		  v2[3] = 0.0;
		  pptResults[0][2] = v2.Magnitude();
		}
	    }
	  else
	    {
	      uiResults = 2;
	      pptResults[0].Set((fXLeft + fXRight)/fDen, -(fYLeft + fYRight)/fDen, 0.0);
	      pptResults[1].Set((fXLeft - fXRight)/fDen, -(fYLeft - fYRight)/fDen, 0.0);

	      for ( i = 0; i < uiResults; i++ )
		{
		  v2 = pptResults[i];
		  v2 -= ppvE[0][0];
		  v2[2] = 0.0;
		  v2[3] = 0.0;
		  pptResults[i][2] = v2.Magnitude();
		}

	      if ( pptResults[1][2] < pptResults[0][2] )
		{
		  pt = pptResults[1];
		  pptResults[1] = pptResults[0];
		  pptResults[0] = pt;
		}

	      if ( !bRadiusInZ )
		{
		  for ( i = 0; i < uiResults; i++ )
		    {
		      pptResults[i][2] = 0.0;
		    }
		}
	    }
	}
        break;

      case atFloat:
      case atExact:
	{
	  Vector ppvE[3][2], vE;
	  Point ptCenter;
	  //Point pt;
	  //FLOAT f;
	  FLOAT fDen, fDenR, fDisc, fDiscR;
	  FLOAT fVDot, fEDot, fEVDot;
	  FLOAT fEVCross, fECross;
	  FLOAT fEVCross2, fECross2;
	  //FLOAT fEVDot2;
	  FLOAT fV1onE2;
	  FLOAT fXLeft, fXRight, fYLeft, fYRight, fRLeft, fRRight, pfR[2];
	  UINT_32 i;

	  uiResults = 0;

	  // Translate the average of the 6 vertices to the origin
	  ptCenter = ptV0;
	  vTrans = ptCenter;
	  vTrans[3] = 0.0;

	  ppvE[0][0] = ptV0 - ptCenter;
	  ppvE[1][0] = ptV1 - ptCenter;
	  ppvE[2][0] = ptE2_0 - ptCenter;
	  ppvE[2][1] = ptE2_1 - ptCenter;
	  vE = ptE2_1 - ptE2_0;

	  fVDot = ppvE[1][0](0)*ppvE[1][0](0) + ppvE[1][0](1)*ppvE[1][0](1);
	  fEDot = vE(0)*vE(0) + vE(1)*vE(1);
	  fEVDot = ppvE[1][0](0)*vE(0) + ppvE[1][0](1)*vE(1);
	  fEVCross = vE(0)*ppvE[1][0](1) - vE(1)*ppvE[1][0](0);
	  fECross = ppvE[2][0](0)*ppvE[2][1](1) - ppvE[2][0](1)*ppvE[2][1](0);

	  fEVCross2 = fEVCross*fEVCross;
	  fECross2 = fECross*fECross;
	  //fEVDot2 = fEVDot*fEVDot;

	  fV1onE2 = ((ptE2_0(0) - ptV1(0))*(ptE2_1(1) - ptV1(1)) - 
		     (ptE2_0(1) - ptV1(1))*(ptE2_1(0) - ptV1(0)));

	  fDen = 2.0*fEVCross2;

	  if ( fabs(fECross) <= s_cfEpsilonVoronoiVVEVertexOnEdge ||
	       fabs(fV1onE2) <= s_cfEpsilonVoronoiVVEVertexOnEdge )
	    {
	      // V0 or V1 is on E1 and it is not one of the endpoints

	      fprintf(stderr, "    VVE: V0 ( %f ) or V1 ( %f ) is on E2\n", fECross, fV1onE2);
	      fprintf(stderr, "      V0: ( %.5f, %.5f, %.5f )\n", ptV0(0), ptV0(1), ptV0(2));
	      fprintf(stderr, "      V1: ( %.5f, %.5f, %.5f )\n", 
		      ptV1(0), ptV1(1), ptV1(2));
	      fprintf(stderr, "      E2: ( %.5f, %.5f, %.5f ) => ( %.5f, %.5f, %.5f )\n", 
		      ptE2_0(0), ptE2_0(1), ptE2_0(2),
		      ptE2_1(0), ptE2_1(1), ptE2_1(2));

	      return;
	    }

	  if ( fabs(fEVCross) < s_cfEpsilonVoronoiVVEAntiparallel )
	    {
	      // The two vertices lie on a line parallel to the edge

#ifdef ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES
	      bCheckVertexZones = FALSE;
#endif // ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES
	      uiResults = 1;

	      fDen = 8.0*fECross*sqrt(fEDot*fVDot);
	      fDenR = 8.0*fECross*sqrt(fEDot);
	      if ( fDen == 0.0 || fDenR == 0.0 )
		{
		  ASSERT( FALSE );
		  uiResults = 0;
		  return;
		}

	      fXLeft = ( -(4.0*fECross2 - fEDot*fVDot)*ppvE[1][0](1) +
			 (4.0*fECross*sqrt(fEDot*fVDot))*ppvE[1][0](0) );
	      fYLeft = ( (4.0*fECross2 - fEDot*fVDot)*ppvE[1][0](0) +
			 (4.0*fECross*sqrt(fEDot*fVDot))*ppvE[1][0](1) );
	      fRLeft = 4.0*fECross2 + fEDot*fVDot;
							
	      pfR[0] = fRLeft/fDenR;
	      pptResults[0].Set(fXLeft/fDen, fYLeft/fDen, 0.0);
	      pptResults[0][2] = sqrt(pptResults[0](0)*pptResults[0](0) + pptResults[0](1)*pptResults[0](1));
	      if ( !( fabs(pptResults[0](2) - pfR[0]) < s_cfEpsilonVoronoiVVE ) )
		{
		  ASSERT( FALSE );
		}
	      pptResults[0][2] = pfR[0];
	    }
	  else
	    {
	      fDisc = fVDot*fEDot*fECross*(fECross + fEVCross);
	      if ( -m_cfEpsilon <= fDisc && fDisc <= 0.0 )
		{
		  fDisc = 0.0;
		}

	      fDiscR = fVDot*fECross*(fECross + fEVCross);
	      if ( fEVDot == 0.0 || (-m_cfEpsilon <= fDiscR && fDiscR <= 0.0) )
		{
		  fDiscR = 0.0;
		}

	      if ( fDisc < 0.0 || fDiscR < 0.0 ) 
		{
		  // Imaginary Solutions
		  return;
		}

	      // X
	      fXLeft = -( fVDot*fEVCross*vE(1) + 2.0*fECross*fEVDot*ppvE[1][0](1) );
	      fXRight = 2.0*ppvE[1][0](1)*sqrt(fDisc);

	      // Y
	      fYLeft = fVDot*fEVCross*vE(0) + 2.0*fECross*fEVDot*ppvE[1][0](0);
	      fYRight = -2.0*ppvE[1][0](0)*sqrt(fDisc);

	      // R
	      fRLeft = sqrt(fEDot)*fVDot*(2.0*fECross + fEVCross);
	      fRRight = -2.0*fEVDot*sqrt(fDiscR);

	      // There can be only one!
	      uiResults = 1;
	      pptResults[0].Set((fXLeft - fXRight)/fDen, (fYLeft - fYRight)/fDen, 0.0);
	      pfR[0] = (fRLeft - fRRight)/fDen;
	      //pfR[0] = (fRLeft/fDen) - (fRRight/fDen);
	      
	      for ( i = 0; i < uiResults; i++ )
		{
		  pptResults[i][2] = sqrt(pptResults[i](0)*pptResults[i](0) + pptResults[i](1)*pptResults[i](1));
		}
	      
	      for ( i = 0; i < uiResults; i++ )
		{
		  if ( !( fabs(pptResults[i](2) - pfR[i]) < s_cfEpsilonVoronoiVVE ) )
		    {
		      fprintf(stderr, "      VVE: %ld: ( %.5f, %.5f, %.5f ) - %.5f !< %.5f\n", i, pptResults[i](0), pptResults[i](1), pptResults[i](2), pfR[i], s_cfEpsilonVoronoiVVE);
		      if ( pfR[i] < s_cfVoronoiInfinity )
			{
			  ASSERT( FALSE );
			}
		    }
		  pptResults[i][2] = pfR[i];
		}


	      /*
	      if ( fXRight == 0.0 && fYRight == 0.0 && fRRight == 0.0 )
		{
#ifdef ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES
		  bCheckVertexZones = FALSE;
#endif // ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES
		  uiResults = 1;
		  pptResults[0].Set(fXLeft/fDen, fYLeft/fDen, 0.0);
	      
		  pptResults[0][2] = sqrt(pptResults[0](0)*pptResults[0](0) + pptResults[0](1)*pptResults[0](1));
	      
		  pfR[0] = fRLeft/fDen;
		  ASSERT( fabs(pptResults[0](2) - pfR[0]) < s_cfEpsilonVoronoiVVE );
		  pptResults[0][2] = pfR[0];
		}
	      else
		{
		  uiResults = 2;
		  pptResults[0].Set((fXLeft + fXRight)/fDen, (fYLeft + fYRight)/fDen, 0.0);
		  pptResults[1].Set((fXLeft - fXRight)/fDen, (fYLeft - fYRight)/fDen, 0.0);
		  //pfR[0] = (fRLeft + fRRight)/fDen;
		  //pfR[1] = (fRLeft - fRRight)/fDen;
		  pfR[0] = (fRLeft/fDen) + (fRRight/fDen);
		  pfR[1] = (fRLeft/fDen) - (fRRight/fDen);
	      
		  for ( i = 0; i < uiResults; i++ )
		    {
		      pptResults[i][2] = sqrt(pptResults[i](0)*pptResults[i](0) + pptResults[i](1)*pptResults[i](1));
		    }

		  for ( i = 0; i < uiResults; i++ )
		    {
		      if ( !( fabs(pptResults[i](2) - pfR[i]) < s_cfEpsilonVoronoiVVE ) )
			{
			  fprintf(stderr, "      VVE: %ld: ( %.5f, %.5f, %.5f ) - %.5f !< %.5f\n", i, pptResults[i](0), pptResults[i](1), pptResults[i](2), pfR[i], s_cfEpsilonVoronoiVVE);
			  if ( pfR[i] < s_cfVoronoiInfinity )
			    {
			      ASSERT( FALSE );
			    }
			}
		      pptResults[i][2] = pfR[i];
		    }
		}
		*/
	    }

	  if ( !bRadiusInZ )
	    {
	      for ( i = 0; i < uiResults; i++ )
		{
		  pptResults[i][2] = 0.0;
		}
	    }
	}
        break;

    default:
      ASSERT( FALSE );
      break;
    }
#endif // COMPILER_PROBLEM

end:
  {
    // Check if the answer is in the valid zone for the finite edge segment
    Vector ppvVZone[2][2];
    Point ppptEdge[1][2];
    Point ptProjection;
    UINT_32 i, j;
#ifdef ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES
    UINT_32 k;
    FLOAT fDot;
#endif // ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES
    BOOL bValid;
    UINT_32 uiCCWRejects;

    uiCCWRejects = 0;
    //bCheckVertexZones = TRUE;

    ppvVZone[0][0] = ptV0 - ptV0_0;
    ppvVZone[0][0][3] = 0.0;
    ppvVZone[0][1] = ptV0 - ptV0_1;
    ppvVZone[0][1][3] = 0.0;
    ppvVZone[1][0] = ptV1 - ptV1_0;
    ppvVZone[1][0][3] = -(ppvVZone[1][0](0)*(ptV1(0) - ptV0(0)) + 
			  ppvVZone[1][0](1)*(ptV1(1) - ptV0(1)));
    ppvVZone[1][1] = ptV1 - ptV1_1;
    ppvVZone[1][1][3] = -(ppvVZone[1][1](0)*(ptV1(0) - ptV0(0)) + 
			  ppvVZone[1][1](1)*(ptV1(1) - ptV0(1)));

    ppptEdge[0][0] = ptE2_0; 
    ppptEdge[0][1] = ptE2_1; 

    for ( i = 0; i < uiResults; )
      {
	bValid = TRUE;
	
#ifdef ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES
	if ( bCheckVertexZones )
	  {
	    for ( j = 0; bValid && j < 2; j++ )
	      {
		for ( k = 0; bValid && k < 2; k++ )
		  {
		    fDot = ( pptResults[i](0)*ppvVZone[j][k](0) + 
			     pptResults[i](1)*ppvVZone[j][k](1) +
			     ppvVZone[j][k](3) );
		    if ( fDot < -s_cfEpsilonVoronoiVVEVertexZone )
		      {
			bValid = FALSE;
		      }
		  }
	      }
	  }
#endif // ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES

	//fprintf(stderr, "  VVE: Circumcenter before translating back:\n");

	// Translate back to the origin
	//fprintf(stderr, "    pptResults[%ld] = ( %f, %f, %f )\n", i, pptResults[i](0), pptResults[i](1), pptResults[i](2));
	pptResults[i] += vTrans;

#ifdef ENABLE_VORONOI_POLYGON_CHECK_EDGE_ZONES
	for ( j = 0; bValid && j < 1; j++ )
	  {
	    if ( !PointInEdgeZone(pptResults[i], ppptEdge[j][0], ppptEdge[j][1]) )
	      {
		bValid = FALSE;
	      }
	  }
#endif // ENABLE_VORONOI_POLYGON_CHECK_EDGE_ZONES

	// Check Counterclockwise
	ptProjection = CalculatePointProjectionOnLine(pptResults[i], ptE2_0, ptE2_1);
	if ( Counterclockwise(ptV1, ptV0, ptProjection) ) 
	  {
	    //ASSERT( FALSE );
	    uiCCWRejects++;
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING_PROJECTION
	    //fprintf(stderr, "    VVE CCW: rejecting circle center [%ld]: valid %ld: ( %.3f, %.3f )\n", i, (UINT_32)bValid, pptResults[i](X), pptResults[i](Y));
	    if ( bValid )
	      {
		fprintf(stderr, "    VVE CCW: rejecting circle center [%ld]: valid %ld: ( %.3f, %.3f )\n", 
			i, (UINT_32)bValid, pptResults[i](X), pptResults[i](Y));
		fprintf(stderr, "      VVE CCW: V0: ( %.3f, %.3f )\n", ptV0(X), ptV0(Y));
		fprintf(stderr, "      VVE CCW: V1: ( %.3f, %.3f )\n", ptV1(X), ptV1(Y));
		fprintf(stderr, "      VVE CCW: E2: ( %.3f, %.3f ) on ( %.3f, %.3f ) -> ( %.3f, %.3f )\n", 
			ptProjection(X), ptProjection(Y),
			ptE2_0(X), ptE2_0(Y),
			ptE2_1(X), ptE2_1(Y));
	      }
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING_PROJECTION

	    bValid = FALSE;
	  }

	if ( bValid )
	  {
	    i++;
	  }
	else
	  {
	    if ( uiResults > 1 )
	      {
		for ( j = i+1; j < uiResults; j++ )
		  {
		    pptResults[j-1] = pptResults[j];
		  }
	      }
	    uiResults--;
	  }
      }

    if ( uiCCWRejects > 1 )
      {
	fprintf(stderr, "VVE CCW: REJECTED BOTH SOLUTIONS\n");
	ASSERT( FALSE );
      }
  }

  ASSERT( uiResults == 0 || uiResults == 1 );
}

VOID QEMesh::CalculateVoronoiPointVEE(Point *pptResults, UINT_32 &uiResults, 
				      const Point &ptV0_0, const Point &ptV0, const Point &ptV0_1, 
				      const Point &ptE1_0, const Point &ptE1_1, 
				      const Point &ptE2_0, const Point &ptE2_1, 
				      BOOL bRadiusInZ/* = FALSE*/)
{
  Vector vTrans;
#ifdef ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES
  BOOL bCheckVertexZones;
  
  bCheckVertexZones = TRUE;
#endif // ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES

#ifndef COMPILER_PROBLEM
  switch ( m_eAccuracy )
    {
      case atFloatNaive:
	{
	  Vector v1, v2, ppvE[3][2];
	  Point pt, ptCenter;
	  FLOAT a, b, c, d, e, f, g, h, k;
	  FLOAT fDen, fDisc;
	  FLOAT fXLeft, fXRight, fYLeft, fYRight;
	  UINT_32 i;

	  // Translate the average of the 6 vertices to the origin
	  ptCenter = ptV0;
	  //ptCenter = ( (2.0*ptV0) + ptE1_0 + ptE1_1 + ptE2_0 + ptE2_1 ) / 6.0;
	  ptCenter[3] = 1.0;
	  vTrans = ptCenter;
	  vTrans[3] = 0.0;

	  ppvE[0][0] = ptV0 - ptCenter; 
	  ppvE[1][0] = ptE1_0 - ptCenter; 
	  ppvE[1][1] = ptE1_1 - ptCenter; 
	  ppvE[2][0] = ptE2_0 - ptCenter; 
	  ppvE[2][1] = ptE2_1 - ptCenter; 

	  uiResults = 0;
  
	  v1 = ppvE[1][1] - ppvE[1][0];
	  v1.Set(-v1(1), v1(0), 0.0);
	  if ( !v1.Normalize() )
	    {
	      ASSERT( FALSE );
	      return;
	    }
	  pt = ppvE[1][0];
	  v1.CompletePlaneEquation(pt);
  
	  v2 = ppvE[2][1] - ppvE[2][0];
	  v2.Set(-v2(1), v2(0), 0.0);
	  if ( !v2.Normalize() )
	    {
	      ASSERT( FALSE );
	      return;
	    }
	  pt = ppvE[2][0];
	  v2.CompletePlaneEquation(pt);
  
	  a = (1 - v1(0)*v1(0));
	  b = -2.0*v1(0)*v1(1); 
	  c = (1 - v1(1)*v1(1));
	  d = -2.0*(ppvE[0][0](0) + v1(0)*v1(3)); 
	  e = -2.0*(ppvE[0][0](1) + v1(1)*v1(3)); 
	  f = (ppvE[0][0](0)*ppvE[0][0](0) + ppvE[0][0](1)*ppvE[0][0](1) - v1(3)*v1(3));

	  g = (v2(0) - v1(0));
	  h = (v2(1) - v1(1));
	  k = (v2(3) - v1(3));

	  fDen = 2.0 * ( a*h*h - b*g*h + c*g*g );
	  if ( fDen == 0.0 )
	    {
	      // Collinear edge segments
	      //ASSERT( FALSE );
	      return;
	    }
	  fDisc = (e*e*g*g + d*d*h*h + b*b*k*k) - 4.0*(c*f*g*g + a*f*h*h + a*c*k*k) + 4.0*(b*f*g*h + a*e*h*k + c*d*g*k) - 2.0*(b*d*h*k + d*e*g*h + b*e*g*k);
	  //if ( -m_cfEpsilon <= fDisc && fDisc <= m_cfEpsilon )
	  if ( -m_cfEpsilon <= fDisc && fDisc <= 0.0 )
	    {
	      fDisc = 0.0;
	    }
	  if ( fDisc < 0.0 ) 
	    {
	      // Imaginary Solutions
	      return;
	    }
	  fDisc = sqrt(fDisc);

	  // X
	  fXLeft = -(2.0*c*g*k - b*h*k - e*g*h + d*h*h);
	  fXRight = h*fDisc;

	  // Y
	  fYLeft = (2.0*a*h*k - b*g*k - d*g*h + e*g*g);
	  fYRight = g*fDisc;

	  if ( fXRight == 0.0 && fYRight == 0.0 )
	    {
	      uiResults = 1;
	      pptResults[0].Set(fXLeft/fDen, -fYLeft/fDen, 0.0);

	      if ( bRadiusInZ )
		{
		  v2 = pptResults[0];
		  v2 -= ppvE[0][0];
		  v2[2] = 0.0;
		  v2[3] = 0.0;
		  pptResults[0][2] = v2.Magnitude();
		}
	    }
	  else
	    {
	      uiResults = 2;
	      pptResults[0].Set((fXLeft + fXRight)/fDen, -(fYLeft + fYRight)/fDen, 0.0);
	      pptResults[1].Set((fXLeft - fXRight)/fDen, -(fYLeft - fYRight)/fDen, 0.0);

	      for ( i = 0; i < uiResults; i++ )
		{
		  v2 = pptResults[i];
		  v2 -= ppvE[0][0];
		  v2[2] = 0.0;
		  v2[3] = 0.0;
		  pptResults[i][2] = v2.Magnitude();
		}

	      if ( pptResults[1][2] < pptResults[0][2] )
		{
		  pt = pptResults[1];
		  pptResults[1] = pptResults[0];
		  pptResults[0] = pt;
		}

	      if ( !bRadiusInZ )
		{
		  for ( i = 0; i < uiResults; i++ )
		    {
		      pptResults[i][2] = 0.0;
		    }
		}
	    }
	}
        break;

      case atFloat:
      case atExact:
	{
	  Vector ppvE[3][2], vE1, vE2;
	  //Vector v;
	  //Point pt;
	  Point ptCenter;
	  //FLOAT f;
	  FLOAT fDen, fDisc, fDenR;
	  FLOAT fE1dotE1, fE2dotE2, fE1dotE2, fE1E2_E1dotE2;
	  FLOAT fE10xE11, fE20xE21, fE1xE2;
	  FLOAT fE1mag, fE2mag;
	  FLOAT fXLeft, fXRight, fYLeft, fYRight, fRLeft, fRRight, pfR[2];
	  FLOAT fE1CoeffLeft, fE1CoeffRight, fE1Coeff;
	  FLOAT fE2CoeffLeft, fE2CoeffRight, fE2Coeff;
	  UINT_32 i;

	  uiResults = 0;
  
	  // Translate the average of the 6 vertices to the origin
	  ptCenter = ptV0;
	  ptCenter[3] = 1.0;
	  vTrans = ptCenter;
	  vTrans[3] = 0.0;

	  ppvE[0][0] = ptV0 - ptCenter; 
	  ppvE[1][0] = ptE1_0 - ptCenter; 
	  ppvE[1][1] = ptE1_1 - ptCenter; 
	  ppvE[2][0] = ptE2_0 - ptCenter; 
	  ppvE[2][1] = ptE2_1 - ptCenter; 
	  vE1 = ptE1_1 - ptE1_0;
	  vE2 = ptE2_1 - ptE2_0;

	  fE1dotE1 = vE1(0)*vE1(0) + vE1(1)*vE1(1);
	  fE2dotE2 = vE2(0)*vE2(0) + vE2(1)*vE2(1);
	  fE1dotE2 = vE1(0)*vE2(0) + vE1(1)*vE2(1);
	  fE1E2_E1dotE2 = sqrt(fE1dotE1*fE2dotE2) - fE1dotE2;
	  fE10xE11 = ppvE[1][0](0)*ppvE[1][1](1) - ppvE[1][0](1)*ppvE[1][1](0);
	  fE20xE21 = ppvE[2][0](0)*ppvE[2][1](1) - ppvE[2][0](1)*ppvE[2][1](0);
	  fE1xE2 = vE1(0)*vE2(1) - vE1(1)*vE2(0);

	  fE1mag = sqrt(fE1dotE1);
	  fE2mag = sqrt(fE2dotE2);

	  fDen = fE1xE2*fE1E2_E1dotE2;
	  fDenR = fE1E2_E1dotE2;

	  if ( fabs(fE10xE11) <= s_cfEpsilonVoronoiVEEVertexOnEdge ||
	       fabs(fE20xE21) <= s_cfEpsilonVoronoiVEEVertexOnEdge )
	    {
	      // V0 is on E1 or E2 and it is not one of the endpoints

	      fprintf(stderr, "    VEE: V0 is on E1 ( %f ) or on E2 ( %f )\n", fE10xE11, fE20xE21);
	      fprintf(stderr, "      V0: ( %.5f, %.5f, %.5f )\n", ptV0(0), ptV0(1), ptV0(2));
	      fprintf(stderr, "      E1: ( %.5f, %.5f, %.5f ) => ( %.5f, %.5f, %.5f )\n", 
		      ptE1_0(0), ptE1_0(1), ptE1_0(2),
		      ptE1_1(0), ptE1_1(1), ptE1_1(2));
	      fprintf(stderr, "      E2: ( %.5f, %.5f, %.5f ) => ( %.5f, %.5f, %.5f )\n", 
		      ptE2_0(0), ptE2_0(1), ptE2_0(2),
		      ptE2_1(0), ptE2_1(1), ptE2_1(2));

	      return;
	    }

	  if ( fabs(fDenR) <= s_cfEpsilonVoronoiVEEParallel )
	    {
	      // Collinear edge segments
	      return;
	    }

	  fDisc = 2.0*fE10xE11*fE20xE21*(sqrt(fE1dotE1*fE2dotE2) + fE1dotE2);
	  if ( -m_cfEpsilon <= fDisc && fDisc <= 0.0 )
	    {
	      fDisc = 0.0;
	    }

	  if ( fDisc < 0.0 ) 
	    {
	      // Imaginary Solutions
	      return;
	    }

	  // R
	  fRLeft = fE1mag*fE20xE21 + fE2mag*fE10xE11;
	  fRRight = sqrt(fDisc);

	  //if ( fabs(fRRight) < s_cfEpsilonVoronoiVEEAntiparallel || 
	  //fabs(fDen) < s_cfEpsilonVoronoiVEEAntiparallel )
	  if ( fabs(fE1xE2) < s_cfEpsilonVoronoiVEEAntiparallel )
	    {
	      FLOAT fE1xE20;

	      // The two edges are anti-parallel
	      uiResults = 2;
	      pfR[0] = fRLeft/fDenR;
	      pfR[1] = pfR[0];

	      fE1xE20 = vE1(0)*ppvE[2][0](1) - vE1(1)*ppvE[2][0](0);

	      fDen = 2.0*fE1dotE1;
	      ASSERT( fDen > 0.0 );
	      fDisc = fE1xE20*fE10xE11;
	      if ( -m_cfEpsilon <= fDisc && fDisc <= 0.0 )
		{
		  fDisc = 0.0;
		}

	      if ( fDisc < 0.0 ) 
		{
		  // Imaginary Solutions
		  ASSERT( FALSE );
		  return;
		}

	      fXLeft = -(fE1xE20 - fE10xE11)*vE1(1);
	      fXRight = 2.0*vE1(0)*sqrt(fDisc);

	      fYLeft = (fE1xE20 - fE10xE11)*vE1(0);
	      fYRight = 2.0*vE1(1)*sqrt(fDisc);

	      if ( fDisc == 0.0 )
		{
		  // One solution
#ifdef ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES
		  bCheckVertexZones = FALSE;
#endif // ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES
		  uiResults = 1;

		  pptResults[0].Set(fXLeft/fDen, fYLeft/fDen, 0.0);
		  pptResults[0][2] = sqrt(pptResults[0](0)*pptResults[0](0) + pptResults[0](1)*pptResults[0](1));
		  if ( !( fabs(pptResults[0](2) - pfR[0]) < s_cfEpsilonVoronoiVEE ) )
		    {
		      i = 0;
		      fprintf(stderr, "      VEE Anti-parallel: %ld: ( %.5f, %.5f, %.5f ) - %.5f !< %.5f\n", i, pptResults[i](0), pptResults[i](1), pptResults[i](2), pfR[i], s_cfEpsilonVoronoiVEE);
		      ASSERT( FALSE );
		    }
		  pptResults[0][2] = pfR[0];
		}
	      else
		{
		  uiResults = 2;

		  pptResults[0].Set((fXLeft - fXRight)/fDen, (fYLeft - fYRight)/fDen, 0.0);
		  pptResults[1].Set((fXLeft + fXRight)/fDen, (fYLeft + fYRight)/fDen, 0.0);
		  for ( i = 0; i < uiResults; i++ )
		    {
		      pptResults[i][2] = sqrt(pptResults[i](0)*pptResults[i](0) + pptResults[i](1)*pptResults[i](1));
		    }
		  
		  /*
		  if ( pptResults[1][2] < pptResults[0][2] )
		    {
		      pt = pptResults[1];
		      pptResults[1] = pptResults[0];
		      pptResults[0] = pt;
		    }
		    */

		  for ( i = 0; i < uiResults; i++ )
		    {
		      if ( !( fabs(pptResults[i](2) - pfR[i]) < s_cfEpsilonVoronoiVEE ) )
			{
			  fprintf(stderr, "      VEE Anti-parallel: %ld: ( %.5f, %.5f, %.5f ) - %.5f !< %.5f\n", i, pptResults[i](0), pptResults[i](1), pptResults[i](2), pfR[i], s_cfEpsilonVoronoiVEE);
			  ASSERT( FALSE );
			}
		      pptResults[i][2] = pfR[i];
		    }
		}

	      if ( !bRadiusInZ )
		{
		  for ( i = 0; i < uiResults; i++ )
		    {
		      pptResults[i][2] = 0.0;
		    }
		}
	    }
	  else
	    {
	      Point pptResultsOld[2];

	      ASSERT( fabs(fDen) > 0.0 );

	      fE1CoeffLeft = fE2dotE2*fE10xE11 + fE1dotE2*fE20xE21;
	      fE1CoeffRight = fE2mag*sqrt(fDisc);

	      fE2CoeffLeft = fE1dotE1*fE20xE21 + fE1dotE2*fE10xE11;
	      fE2CoeffRight = fE1mag*sqrt(fDisc);

	      // X
	      fXLeft = ( (fE20xE21*(vE2(0)*fE1dotE1 - vE1(0)*fE1dotE2)) -
			 (fE10xE11*(vE1(0)*fE2dotE2 - vE2(0)*fE1dotE2)) );
	      fXRight = (vE2(0)*fE1mag - vE1(0)*fE2mag)*sqrt(fDisc);
	      
	      // Y
	      fYLeft = ( (fE20xE21*(vE2(1)*fE1dotE1 - vE1(1)*fE1dotE2)) -
			 (fE10xE11*(vE1(1)*fE2dotE2 - vE2(1)*fE1dotE2)) );
	      fYRight = (vE2(1)*fE1mag - vE1(1)*fE2mag)*sqrt(fDisc);

	      if ( fDisc == 0.0 )
		{
		  // One solution
#ifdef ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES
		  bCheckVertexZones = FALSE;
#endif // ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES
		  uiResults = 1;

		  fE1Coeff = fE1CoeffLeft/fDen;
		  fE2Coeff = fE2CoeffLeft/fDen;

		  pptResults[0].Set(fE2Coeff*vE2(0) - fE1Coeff*vE1(0), 
				    fE2Coeff*vE2(1) - fE1Coeff*vE1(1), 0.0);
		  pptResults[0][2] = sqrt(pptResults[0](0)*pptResults[0](0) + pptResults[0](1)*pptResults[0](1));

		  pptResultsOld[0].Set(fXLeft/fDen, fYLeft/fDen, 0.0);
		  pptResultsOld[0][2] = sqrt(pptResults[0](0)*pptResults[0](0) + pptResults[0](1)*pptResults[0](1));
	      
		  pfR[0] = fRLeft/fDen;

		  i = 0;
		  //fprintf(stderr, "      VEE: %ld: RADIUS = %.3f\n", i, pfR[i]);
		  //fprintf(stderr, "      VEE: %ld: NEW: ( %.3f, %.3f, %.3f )\n", i, pptResults[i](0), pptResults[i](1), pptResults[i](2));
		  //fprintf(stderr, "      VEE: %ld: OLD: ( %.3f, %.3f, %.3f )\n", i, pptResultsOld[i](0), pptResultsOld[i](1), pptResultsOld[i](2));
		  
		  ASSERT( fabs(pptResults[0](2) - pfR[0]) < s_cfEpsilonVoronoiVEE );
		  pptResults[0][2] = pfR[0];

		  if ( !bRadiusInZ )
		    {
		      pptResults[0][2] = 0.0;
		    }
		}
	      else 
		{
		  uiResults = 2;

		  fE1Coeff = (fE1CoeffLeft - fE1CoeffRight)/fDen;
		  fE2Coeff = (fE2CoeffLeft - fE2CoeffRight)/fDen;
		  pptResults[0].Set(fE2Coeff*vE2(0) - fE1Coeff*vE1(0), 
				    fE2Coeff*vE2(1) - fE1Coeff*vE1(1), 0.0);

		  fE1Coeff = (fE1CoeffLeft + fE1CoeffRight)/fDen;
		  fE2Coeff = (fE2CoeffLeft + fE2CoeffRight)/fDen;
		  pptResults[1].Set(fE2Coeff*vE2(0) - fE1Coeff*vE1(0), 
				    fE2Coeff*vE2(1) - fE1Coeff*vE1(1), 0.0);

		  pptResultsOld[0].Set((fXLeft - fXRight)/fDen, (fYLeft - fYRight)/fDen, 0.0);
		  pptResultsOld[1].Set((fXLeft + fXRight)/fDen, (fYLeft + fYRight)/fDen, 0.0);

		  pfR[0] = (fRLeft - fRRight)/fDenR;
		  pfR[1] = (fRLeft + fRRight)/fDenR;
		  
		  for ( i = 0; i < uiResults; i++ )
		    {
		      pptResults[i][2] = sqrt(pptResults[i](0)*pptResults[i](0) + pptResults[i](1)*pptResults[i](1));
		      pptResultsOld[i][2] = sqrt(pptResultsOld[i](0)*pptResultsOld[i](0) + pptResultsOld[i](1)*pptResultsOld[i](1));

		      //fprintf(stderr, "      VEE: %ld: RADIUS = %.3f\n", i, pfR[i]);
		      //fprintf(stderr, "      VEE: %ld: NEW: ( %.3f, %.3f, %.3f )\n", i, pptResults[i](0), pptResults[i](1), pptResults[i](2));
		      //fprintf(stderr, "      VEE: %ld: OLD: ( %.3f, %.3f, %.3f )\n", i, pptResultsOld[i](0), pptResultsOld[i](1), pptResultsOld[i](2));
		    }
		  
		  /*
		  // Choose the answer with minimum radius -- this is WRONG
		  if ( pfR[1] < pfR[0] )
		    {
		      pt = pptResults[1];
		      pptResults[1] = pptResults[0];
		      pptResults[0] = pt;
		      
		      f = pfR[1];
		      pfR[1] = pfR[0];
		      pfR[0] = f;
		    }
		    */

		  for ( i = 0; i < uiResults; i++ )
		    {
		      if ( !( fabs(pptResults[i](2) - pfR[i]) < s_cfEpsilonVoronoiVEE ) )
			{
			  ASSERT( FALSE );
			}
		      pptResults[i][2] = pfR[i];
		    }

		  if ( !bRadiusInZ )
		    {
		      for ( i = 0; i < uiResults; i++ )
			{
			  pptResults[i][2] = 0.0;
			}
		    }
		}
	    }
	}
        break;

    default:
      ASSERT( FALSE );
      break;
    }
#endif // COMPILER_PROBLEM

  {
    // Check if the answer is in the valid zone for the finite edge segment
    Vector ppvVZone[1][2];
    Point ppptEdge[2][2];
    Point pptProjection[2];
    UINT_32 i, j;
#ifdef ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES
    UINT_32 k;
    FLOAT fDot;
#endif // ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES
    BOOL bValid;
    UINT_32 uiCCWRejects;

    uiCCWRejects = 0;    
#ifdef ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES
    //bCheckVertexZones = TRUE;
#endif // ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES

    ppvVZone[0][0] = ptV0 - ptV0_0;
    ppvVZone[0][1] = ptV0 - ptV0_1;

    ppptEdge[0][0] = ptE1_0; 
    ppptEdge[0][1] = ptE1_1; 
    ppptEdge[1][0] = ptE2_0; 
    ppptEdge[1][1] = ptE2_1; 

    for ( i = 0; i < uiResults; )
      {
	bValid = TRUE;

#ifdef ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES
	if ( bCheckVertexZones )
	  {
	    for ( j = 0; bValid && j < 1; j++ )
	      {
		for ( k = 0; bValid && k < 2; k++ )
		  {
		    fDot = pptResults[i](0)*ppvVZone[j][k](0) + pptResults[i](1)*ppvVZone[j][k](1);
		    if ( fDot < -s_cfEpsilonVoronoiVEEVertexZone )
		      {
			bValid = FALSE;
		      }
		  }
	      }
	  }
#endif // ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES

	//fprintf(stderr, "  VEE: Circumcenter before translating back:\n");
	
	// Translate the answer back
	//fprintf(stderr, "    pptResults[%ld] = ( %f, %f, %f )\n", i, pptResults[i](0), pptResults[i](1), pptResults[i](2));
	pptResults[i] += vTrans;

#ifdef ENABLE_VORONOI_POLYGON_CHECK_EDGE_ZONES
	for ( j = 0; bValid && j < 2; j++ )
	  {
	    if ( !PointInEdgeZone(pptResults[i], ppptEdge[j][0], ppptEdge[j][1]) )
	      {
		bValid = FALSE;
	      }
	  }
#endif // ENABLE_VORONOI_POLYGON_CHECK_EDGE_ZONES

	// Check Counterclockwise
	pptProjection[0] = CalculatePointProjectionOnLine(pptResults[i], ptE2_0, ptE2_1);
	pptProjection[1] = CalculatePointProjectionOnLine(pptResults[i], ptE1_0, ptE1_1);
	if ( Counterclockwise(ptV0, pptProjection[0], pptProjection[1]) )
	  {
	    uiCCWRejects++;
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING_PROJECTION
	    //fprintf(stderr, "    VEE CCW: rejecting circle center [%ld]: valid %ld: ( %.3f, %.3f )\n", i, (UINT_32)bValid, pptResults[i](X), pptResults[i](Y));
	    if ( bValid )
	      {
		fprintf(stderr, "    VEE CCW: rejecting circle center [%ld]: valid %ld: ( %.3f, %.3f )\n", 
			i, (UINT_32)bValid, pptResults[i](X), pptResults[i](Y));
		fprintf(stderr, "      VEE CCW: V0: ( %.3f, %.3f )\n", ptV0(X), ptV0(Y));
		fprintf(stderr, "      VEE CCW: E1: ( %.3f, %.3f ) on ( %.3f, %.3f ) -> ( %.3f, %.3f )\n", 
			pptProjection[1](X), pptProjection[1](Y),
			ptE1_0(X), ptE1_0(Y),
			ptE1_1(X), ptE1_1(Y));
		fprintf(stderr, "      VEE CCW: E2: ( %.3f, %.3f ) on ( %.3f, %.3f ) -> ( %.3f, %.3f )\n", 
			pptProjection[0](X), pptProjection[0](Y),
			ptE2_0(X), ptE2_0(Y),
			ptE2_1(X), ptE2_1(Y));
	      }
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING_PROJECTION

	    bValid = FALSE;
	  }

	if ( bValid )
	  {
	    i++;
	  }
	else
	  {
	    if ( uiResults > 1 )
	      {
		for ( j = i+1; j < uiResults; j++ )
		  {
		    pptResults[j-1] = pptResults[j];
		  }
	      }
	    uiResults--;
	  }
      }

    if ( uiCCWRejects > 1 )
      {
	fprintf(stderr, "VVE CCW: REJECTED BOTH SOLUTIONS\n");
	ASSERT( FALSE );
      }
  }

  ASSERT( uiResults == 0 || uiResults == 1 );
}

VOID QEMesh::CalculateVoronoiPointEE(Point *pptResults, UINT_32 &uiResults,
				     const Point &ptE0_0, const Point &ptE0_1, 
				     const Point &ptE1_0, const Point &ptE1_1, 
				     BOOL bRadiusInZ/* = FALSE*/, 
				     BOOL bCounterclockwise/* = TRUE*/)
{
  Point ptCenter;
  Vector ppvE[2][2], pvM[4], vTrans, v;
  Matrix m;

  FLOAT fDen, fX, fY;
  UINT_32 i;
  //Vector pvPlane[2];
  //Point pt;
  FLOAT fCCW;

  fCCW = 1.0;
  if ( !bCounterclockwise )
    {
      fCCW = -1.0;
    }

#ifndef COMPILER_PROBLEM
  switch ( m_eAccuracy )
    {
      case atFloatNaive:
	{
	  ptCenter.Null();
	  vTrans.Null();
	}
        break;

      case atFloat:
      case atExact:
	{
	  // Translate the average of the 6 vertices to the origin
	  ptCenter = ptE0_0;
	  ptCenter[3] = 1.0;
	  vTrans = ptCenter;
	  vTrans[3] = 0.0;
	}
        break;

    default:
      ASSERT( FALSE );
      break;
    }
#endif // COMPILER_PROBLEM

  ppvE[0][0] = ptE0_0 - ptCenter; 
  ppvE[0][1] = ptE0_1 - ptCenter; 
  ppvE[1][0] = ptE1_0 - ptCenter; 
  ppvE[1][1] = ptE1_1 - ptCenter; 

  //////////////////////////////////////////////////////////////////////

  for ( i = 0; i < 2; i++ )
    {
      // The A coeff
      pvM[0][i] = ppvE[i][1][Y] - ppvE[i][0][Y];
      pvM[0][i] *= fCCW;
      // The B coeff
      pvM[1][i] = -(ppvE[i][1][X] - ppvE[i][0][X]);
      pvM[1][i] *= fCCW;
      // The R coeff
      pvM[2][i] = sqrt(pvM[0][i]*pvM[0][i] + pvM[1][i]*pvM[1][i]);
      // The Constants
      pvM[3][i] = ppvE[i][0][X]*ppvE[i][1][Y] - ppvE[i][1][X]*ppvE[i][0][Y];
      pvM[3][i] *= fCCW;
    }

  // The A coeff
  pvM[0][2] = ppvE[0][1](X) - ppvE[0][0](X);
  // The B coeff
  pvM[1][2] = ppvE[0][1](Y) - ppvE[0][0](Y);
  // The R coeff
  pvM[2][2] = 0.0;
  // The Constants
  pvM[3][2] = ( (ppvE[0][0](X)*ppvE[0][1](X) + ppvE[0][0](Y)*ppvE[0][1](Y)) -
		(ppvE[0][0](X)*ppvE[0][0](X) + ppvE[0][0](Y)*ppvE[0][0](Y)) );
  //  pvM[3][2] = (ppvE[0][1](X) - ppvE[0][0](X))*ppvE[0][0](X) + (ppvE[0][1](Y) - ppvE[0][0](Y))*ppvE[0][0](Y);

  //////////////////////////////////////////////////////////////////////
  // Use Kramer's Rule
  
  uiResults = 0;
  pptResults[0].Null();

  ////////////////////////////////////////
  // The Denominator Term

  m.SetIdentity();
  for ( i = 0; i < 3; i++ )
    {
      m.Set(i, 0, pvM[0][i]);
    }
  for ( i = 0; i < 3; i++ )
    {
      m.Set(i, 1, pvM[1][i]);
    }
  for ( i = 0; i < 3; i++ )
    {
      m.Set(i, 2, pvM[2][i]);
    }

  fDen = m.Determinant();

  if ( fDen == 0.0 )
    {
      uiResults = 0;
      return;
    }

  ////////////////////////////////////////
  // The X Term

  for ( i = 0; i < 3; i++ )
    {
      m.Set(i, 0, pvM[3][i]);
    }

  fX = m.Determinant();

  ////////////////////////////////////////
  // The Y Term

  for ( i = 0; i < 3; i++ )
    {
      m.Set(i, 0, pvM[0][i]);
    }
  for ( i = 0; i < 3; i++ )
    {
      m.Set(i, 1, pvM[3][i]);
    }

  fY = m.Determinant();

  uiResults = 1;
  pptResults[0].Set(fX/fDen, fY/fDen, 0.0);

  ////////////////////////////////////////
  // The R Term

  if ( bRadiusInZ )
    {
      v = pptResults[0];
      v[3] = 0.0;
      v = v - ppvE[0][0];
      v[2] = 0.0;
      v[3] = 0.0;

      pptResults[0][2] = v.Magnitude();
    }

  pptResults[0] += vTrans;

  {
    // Check if the answer is in the valid zone for the finite edge segment
    Point ppptEdge[1][2];
    UINT_32 i, j;
    BOOL bValid;
    
    ppptEdge[0][0] = ptE1_0; 
    ppptEdge[0][1] = ptE1_1; 

    for ( i = 0; i < uiResults; )
      {
	bValid = TRUE;
#ifdef ENABLE_VORONOI_POLYGON_CHECK_EDGE_ZONES
	for ( j = 0; j < 1; j++ )
	  {
	    //Point ptProjection;

	    //ptProjection = CalculatePointProjectionOnLine(pptResults[i], ppptEdge[j][0], ppptEdge[j][1]);

	    //fprintf(stderr, "START EE:\n");
	    //fprintf(stderr, "  Center: ( %e, %e, %e )\n", pptResults[i](0), pptResults[i](1), pptResults[i](2));
	    //fprintf(stderr, "  E1_0  : ( %e, %e )\n", ppptEdge[j][0](0), ppptEdge[j][0](1));
	    //fprintf(stderr, "  Proj  : ( %e, %e )\n", ptProjection(0), ptProjection(1));
	    //fprintf(stderr, "  E1_1  : ( %e, %e )\n", ppptEdge[j][1](0), ppptEdge[j][1](1));
	    //fprintf(stderr, "END   EE:\n");

	    if ( !PointInEdgeZone(pptResults[i], ppptEdge[j][0], ppptEdge[j][1]) )
	      {
		bValid = FALSE;
		break;
	      }

	  }
#endif // ENABLE_VORONOI_POLYGON_CHECK_EDGE_ZONES

	if ( bValid )
	  {
	    i++;
	  }
	else
	  {
	    if ( uiResults > 1 )
	      {
		for ( j = i+1; j < uiResults; j++ )
		  {
		    pptResults[j-1] = pptResults[j];
		  }
	      }
	    uiResults--;
	  }
      }
  }
}

VOID QEMesh::CalculateVoronoiPointEvEE(Point *pptResults, UINT_32 &uiResults,
				       const Point &ptE0_0,
				       const Point &ptE1_0, const Point &ptE1_1, 
				       const Point &ptE2_0, const Point &ptE2_1, 
				       BOOL bRadiusInZ/* = FALSE*/)
{
  Point ptCenter;
  Vector ppvE[3][2], pvM[4], vTrans;
  //Vector v;
  Matrix m;

  FLOAT fDen, fX, fY, fR;
  UINT_32 i;

#ifndef COMPILER_PROBLEM
  switch ( m_eAccuracy )
    {
      case atFloatNaive:
	{
	  ptCenter.Null();
	  vTrans.Null();
	}
        break;

      case atFloat:
      case atExact:
	{
	  // Translate the average of the 6 vertices to the origin
	  ptCenter = ptE1_0;
	  ptCenter[3] = 1.0;
	  vTrans = ptCenter;
	  vTrans[3] = 0.0;
	}
        break;

    default:
      ASSERT( FALSE );
      break;
    }
#endif // COMPILER_PROBLEM

  ppvE[0][0] = ptE0_0 - ptCenter; 
  ppvE[0][1] = ptE1_0 - ptCenter; // this should not need to be used 
  ppvE[1][0] = ptE1_0 - ptCenter; 
  ppvE[1][1] = ptE1_1 - ptCenter; 
  ppvE[2][0] = ptE2_0 - ptCenter; 
  ppvE[2][1] = ptE2_1 - ptCenter; 

  //////////////////////////////////////////////////////////////////////

  // The A coeff
  pvM[0][0] = -( ppvE[0][0](Y) - 2.0*ppvE[1][0](Y) + ppvE[1][1](Y) );
  // The B coeff
  pvM[1][0] = ppvE[0][0](X) - 2.0*ppvE[1][0](X) + ppvE[1][1](X);
  // The R coeff
  pvM[2][0] = 0.0;
  // The Constants
  pvM[3][0] = ( ( ppvE[0][0](X)*ppvE[1][0](Y) + ppvE[1][1](X)*ppvE[1][0](Y) ) -
		( ppvE[1][0](X)*ppvE[0][0](Y) + ppvE[1][0](X)*ppvE[1][1](Y) ) );

  for ( i = 1; i < 3; i++ )
    {
      // The A coeff
      pvM[0][i] = ppvE[i][1][Y] - ppvE[i][0][Y];
      // The B coeff
      pvM[1][i] = -(ppvE[i][1][X] - ppvE[i][0][X]);
      // The R coeff
      pvM[2][i] = sqrt(pvM[0][i]*pvM[0][i] + pvM[1][i]*pvM[1][i]);
      // The Constants
      pvM[3][i] = ppvE[i][0][X]*ppvE[i][1][Y] - ppvE[i][1][X]*ppvE[i][0][Y];
    }

  //////////////////////////////////////////////////////////////////////
  // Use Kramer's Rule
  
  uiResults = 0;
  pptResults[0].Null();

  ////////////////////////////////////////
  // The Denominator Term

  m.SetIdentity();
  for ( i = 0; i < 3; i++ )
    {
      m.Set(i, 0, pvM[0][i]);
    }
  for ( i = 0; i < 3; i++ )
    {
      m.Set(i, 1, pvM[1][i]);
    }
  for ( i = 0; i < 3; i++ )
    {
      m.Set(i, 2, pvM[2][i]);
    }

  fDen = m.Determinant();

  if ( fDen == 0.0 )
    {
      uiResults = 0;
      return;
    }

  ////////////////////////////////////////
  // The X Term

  for ( i = 0; i < 3; i++ )
    {
      m.Set(i, 0, pvM[3][i]);
    }

  fX = m.Determinant();

  ////////////////////////////////////////
  // The Y Term

  for ( i = 0; i < 3; i++ )
    {
      m.Set(i, 0, pvM[0][i]);
    }
  for ( i = 0; i < 3; i++ )
    {
      m.Set(i, 1, pvM[3][i]);
    }

  fY = m.Determinant();

  uiResults = 1;
  pptResults[0].Set(fX/fDen, fY/fDen, 0.0);

  ////////////////////////////////////////
  // The R Term

  if ( bRadiusInZ )
    {
      for ( i = 0; i < 3; i++ )
	{
	  m.Set(i, 1, pvM[1][i]);
	}
      for ( i = 0; i < 3; i++ )
	{
	  m.Set(i, 2, pvM[3][i]);
	}
      
      fR = m.Determinant();
      
      pptResults[0][2] = fR / fDen;
    }

  pptResults[0] += vTrans;

  {
    // Check if the answer is in the valid zone for the finite edge segment
    Point ppptEdge[3][2];
    UINT_32 i, j;
    BOOL bValid;
    
    ppptEdge[0][0] = ptE0_0; 
    ppptEdge[0][1] = ptE1_0; 
    ppptEdge[1][0] = ptE1_0; 
    ppptEdge[1][1] = ptE1_1; 
    ppptEdge[2][0] = ptE2_0; 
    ppptEdge[2][1] = ptE2_1; 

    for ( i = 0; i < uiResults; )
      {
	bValid = TRUE;
#ifdef ENABLE_VORONOI_POLYGON_CHECK_EDGE_ZONES
	for ( j = 0; j < 3; j++ )
	  {
	    if ( !PointInEdgeZone(pptResults[i], ppptEdge[j][0], ppptEdge[j][1]) )
	      {
		bValid = FALSE;
		break;
	      }
	  }
#endif // ENABLE_VORONOI_POLYGON_CHECK_EDGE_ZONES

	if ( bValid )
	  {
	    i++;
	  }
	else
	  {
	    if ( uiResults > 1 )
	      {
		for ( j = i+1; j < uiResults; j++ )
		  {
		    pptResults[j-1] = pptResults[j];
		  }
	      }
	    uiResults--;
	  }
      }
  }
}

VOID QEMesh::CalculateVoronoiPointEEE(Point *pptResults, UINT_32 &uiResults,
				      const Point &ptE0_0, const Point &ptE0_1, 
				      const Point &ptE1_0, const Point &ptE1_1, 
				      const Point &ptE2_0, const Point &ptE2_1, 
				      BOOL bRadiusInZ/* = FALSE*/)
{
  Point ptCenter;
  Vector ppvE[3][2], vTrans;
  Vector ppvLine[3];

  FLOAT pfDot[3];
  FLOAT fDen, fX, fY, fR;
  //Vector v, pvPlane[2];
  //Point pt;

#ifndef COMPILER_PROBLEM
  switch ( m_eAccuracy )
    {
      case atFloatNaive:
	{
	  ptCenter.Null();
	  vTrans.Null();
	}
        break;

      case atFloat:
      case atExact:
	{
	  // Translate the average of the 6 vertices to the origin
	  ptCenter = ptE1_0;
	  ptCenter[3] = 1.0;
	  vTrans = ptCenter;
	  vTrans[3] = 0.0;
	}
        break;

    default:
      ASSERT( FALSE );
      break;
    }
#endif // COMPILER_PROBLEM

  ppvE[0][0] = ptE0_0 - ptCenter; 
  ppvE[0][1] = ptE0_1 - ptCenter; 
  ppvE[1][0] = ptE1_0 - ptCenter; 
  ppvE[1][1] = ptE1_1 - ptCenter; 
  ppvE[2][0] = ptE2_0 - ptCenter; 
  ppvE[2][1] = ptE2_1 - ptCenter; 

  //////////////////////////////////////////////////
  // Construct the line equation coefficients

  ppvLine[0][0] = (ppvE[0][0](1) - ppvE[0][1](1));
  ppvLine[0][1] = (ppvE[0][1](0) - ppvE[0][0](0));
  ppvLine[0][2] = 0.0;
  ppvLine[0][3] = (ppvE[0][0](0)*ppvE[0][1](1) - ppvE[0][1](0)*ppvE[0][0](1));
  pfDot[0] = (ppvLine[0](0)*ppvLine[0](0) + ppvLine[0](1)*ppvLine[0](1));

  // This one is oriented opposite to the other two
  ppvLine[1][0] = (ppvE[1][1](1) - ppvE[1][0](1));
  ppvLine[1][1] = (ppvE[1][0](0) - ppvE[1][1](0));
  ppvLine[1][2] = 0.0;
  ppvLine[1][3] = (ppvE[1][1](0)*ppvE[1][0](1) - ppvE[1][0](0)*ppvE[1][1](1));
  pfDot[1] = (ppvLine[1](0)*ppvLine[1](0) + ppvLine[1](1)*ppvLine[1](1));

  ppvLine[2][0] = (ppvE[2][0](1) - ppvE[2][1](1));
  ppvLine[2][1] = (ppvE[2][1](0) - ppvE[2][0](0));
  ppvLine[2][2] = 0.0;
  ppvLine[2][3] = (ppvE[2][0](0)*ppvE[2][1](1) - ppvE[2][1](0)*ppvE[2][0](1));
  pfDot[2] = (ppvLine[2](0)*ppvLine[2](0) + ppvLine[2](1)*ppvLine[2](1));

  //////////////////////////////////////////////////////////////////////
  // Compute the answer based on the expanded Kramer's rule
  //   system of 2 equations which are the bisector lines of the 2 pairs of edges

  ////////////////////////////////////////
  // The Denominator

  fDen = ( (ppvLine[0](0)*ppvLine[2](1) - ppvLine[2](0)*ppvLine[0](1))*pfDot[1] +
	   (ppvLine[1](0)*ppvLine[2](1) - ppvLine[2](0)*ppvLine[1](1))*sqrt(pfDot[0]*pfDot[1]) +
	   (ppvLine[0](0)*ppvLine[1](1) - ppvLine[1](0)*ppvLine[0](1))*sqrt(pfDot[1]*pfDot[2]) );

  if ( fDen == 0.0 )
    {
      ASSERT( FALSE );
      return;
    }

  ////////////////////////////////////////
  // The X Term

  fX = -( (ppvLine[0](3)*ppvLine[2](1) - ppvLine[2](3)*ppvLine[0](1))*pfDot[1] +
	 (ppvLine[1](3)*ppvLine[2](1) - ppvLine[2](3)*ppvLine[1](1))*sqrt(pfDot[0]*pfDot[1]) +
	 (ppvLine[0](3)*ppvLine[1](1) - ppvLine[1](3)*ppvLine[0](1))*sqrt(pfDot[1]*pfDot[2]) );

  ////////////////////////////////////////
  // The Y Term

  fY = -( (ppvLine[0](0)*ppvLine[2](3) - ppvLine[2](0)*ppvLine[0](3))*pfDot[1] +
	 (ppvLine[1](0)*ppvLine[2](3) - ppvLine[2](0)*ppvLine[1](3))*sqrt(pfDot[0]*pfDot[1]) +
	 (ppvLine[0](0)*ppvLine[1](3) - ppvLine[1](0)*ppvLine[0](3))*sqrt(pfDot[1]*pfDot[2]) );

  uiResults = 1;
  pptResults[0].Set(fX/fDen, fY/fDen, 0.0);


  ////////////////////////////////////////
  // The R Term

  if ( bRadiusInZ )
    {
      fR = ( ((ppvLine[0](3)*(ppvLine[1](0)*ppvLine[2](1) - ppvLine[2](0)*ppvLine[1](1)) +
	       ppvLine[1](3)*(ppvLine[0](0)*ppvLine[2](1) - ppvLine[2](0)*ppvLine[0](1)) +
	       ppvLine[2](3)*(ppvLine[0](0)*ppvLine[1](1) - ppvLine[1](0)*ppvLine[0](1))) *
	      sqrt(pfDot[1])
	     ) +
	     (2.0*ppvLine[1](3) * 
	      ( (ppvLine[1](0)*ppvLine[2](1) - ppvLine[2](0)*ppvLine[1](1))*sqrt(pfDot[0]) +
		(ppvLine[0](0)*ppvLine[1](1) - ppvLine[1](0)*ppvLine[0](1))*sqrt(pfDot[2]) )
	     )
	   );
      
      pptResults[0][2] = fR / fDen;
    }

  // Translate back to the origin
  pptResults[0] += vTrans;

  {
    // Check if the answer is in the valid zone for the finite edge segment
    Point ppptEdge[3][2];
    UINT_32 i, j;
    BOOL bValid;
    
    ppptEdge[0][0] = ptE0_0; 
    ppptEdge[0][1] = ptE0_1; 
    ppptEdge[1][0] = ptE1_0; 
    ppptEdge[1][1] = ptE1_1; 
    ppptEdge[2][0] = ptE2_0; 
    ppptEdge[2][1] = ptE2_1; 

    for ( i = 0; i < uiResults; )
      {
	bValid = TRUE;
#ifdef ENABLE_VORONOI_POLYGON_CHECK_EDGE_ZONES
	for ( j = 0; j < 3; j++ )
	  {
	    if ( !PointInEdgeZone(pptResults[i], ppptEdge[j][0], ppptEdge[j][1]) )
	      {
		bValid = FALSE;
		break;
	      }
	  }
#endif // ENABLE_VORONOI_POLYGON_CHECK_EDGE_ZONES

	if ( bValid )
	  {
	    i++;
	  }
	else
	  {
	    if ( uiResults > 1 )
	      {
		for ( j = i+1; j < uiResults; j++ )
		  {
		    pptResults[j-1] = pptResults[j];
		  }
	      }
	    uiResults--;
	  }
      }
  }
}

VOID QEMesh::CalculateOffsetPoint(Point *pptResults, UINT_32 &uiResults, 
				  QEEdge *peVoronoi, QEFacet *pfa0, QEFacet *pfa1, 
				  FLOAT fOffset)
{
  QEFacet *ppfa[2], *pfa;
  QEVertex *ppvx[2], *pvx;
  QEEdge *ppe[2], *pe;
  Vector pvPlane[2], v;
  Point ppt[2], pt;
  FLOAT fDisc, fDen, fX, fY, fMag;
  UINT_32 i, uiEdges;

  ppfa[0] = pfa0;
  ppfa[1] = pfa1;

  uiEdges = 0;
  for ( i = 0; i < 2; i++ )
    {
      ppvx[i] = ppfa[i]->GetForwardVertex();
      ppe[i] = ppfa[i]->GetForwardEdge();
      if ( ppfa[i]->GetForwardEdge() != NULL )
	{
	  //////////////////////////////////////////////////
	  // This is necessary because the Voronoi diagram 
	  //   includes the original contour and this caused
	  //   numeric problems.
	  if ( ppe[i]->GetEdgePrev(0)->GetVertex()->GetPoint()(Z) < 0.0 )
	    {
	      ppe[i] = ppe[i]->GetEdge();
	    }
	  uiEdges++;
	}
    }

  switch ( uiEdges )
    {
    case 0:
      {
	if ( ppvx[0] == ppvx[1] )
	  {
	    ASSERT( ppvx[0] == peVoronoi->GetVertex() ||
		    ppvx[0] == peVoronoi->GetEdge()->GetVertex() );
	    if ( ppvx[0] != peVoronoi->GetVertex() )
	      {
		peVoronoi = peVoronoi->GetEdge();
	      }
	    ASSERT( ppvx[0] == peVoronoi->GetVertex() );

	    ppt[0] = peVoronoi->GetVertex()->GetPoint();
	    ppt[1] = peVoronoi->GetEdge()->GetVertex()->GetPoint();

	    ASSERT( ppt[0](2) == 0.0 && ppt[0](3) == 1.0 );
	    ASSERT( ppt[1](3) == 1.0 );
	    v = ppt[1] - ppt[0];
	    v[2] = 0.0;
	    v[3] = 0.0;

	    fMag = v.Magnitude();
	    if ( !v.Normalize() || 
		 fabs(fMag) <= s_cfEpsilonVoronoiOffsetDen )
	      {
		uiResults = 0;
		return;
	      }

	    pptResults[0] = ppt[0] + fabs(fOffset)*v;
	    pptResults[0][2] = fOffset;
	    
	    uiResults = 1;
	  }
	else
	  {
	    ppt[0] = ppvx[0]->GetPoint();
	    ppt[1] = ppvx[1]->GetPoint();
	
	    pt = (ppt[0] + ppt[1]) / 2.0;
	    pt[3] = 1.0;
	
	    v = ppt[1] - ppt[0];
	    v.Set(-v(1), v(0), 0.0, 0.0);
	
	    fDisc = -(v(0)*v(0) + v(1)*v(1))/4.0;
	    fDisc += fOffset * fOffset;
	
	    fMag = v.Magnitude();
	    if ( fDisc < 0.0 ||
		 !v.Normalize() || 
		 fabs(fMag) <= s_cfEpsilonVoronoiOffsetDen )
	      {
		uiResults = 0;
		return;
	      }
	
	    fDisc = sqrt(fDisc);
	
	    if ( fDisc == 0.0 )
	      {
		uiResults = 1;
		pptResults[0] = pt;
		pptResults[0][2] = fOffset;
	      }
	    else
	      {
		uiResults = 2;
		if ( fOffset < 0.0 )
		  {
		    fDisc = -fDisc;
		  }
		pptResults[0] = pt + (v * fDisc);
		pptResults[0][2] = fOffset;
		pptResults[1] = pt + (v * -fDisc);
		pptResults[1][2] = fOffset;
	      }
	  }
      }
      break;
    case 1:
      {
	while ( ppvx[0] == NULL )
	  {
	    pfa = ppfa[0];
	    ppfa[0] = ppfa[1];
	    ppfa[1] = pfa;

	    pvx = ppvx[0];
	    ppvx[0] = ppvx[1];
	    ppvx[1] = pvx;

	    pe = ppe[0];
	    ppe[0] = ppe[1];
	    ppe[1] = pe;
	  }

	ppt[0] = ppe[1]->GetVertex()->GetPoint();
	ppt[1] = ppe[1]->GetEdge()->GetVertex()->GetPoint();
	pvPlane[1] = ppt[1] - ppt[0];
	pvPlane[1].Set(-pvPlane[1](1), pvPlane[1](0), 0.0, 0.0);

	fMag = pvPlane[1].Magnitude();
	if ( !pvPlane[1].Normalize() ||
	     fabs(fMag) <= s_cfEpsilonVoronoiOffsetDen )
	  {
	    uiResults = 0;
	    return;
	  }

	if ( ppe[1]->GetVertex() == ppvx[0] )
	  {
	    uiResults = 1;
	    pptResults[0] = ppt[0] + (fOffset * pvPlane[1]);
	    pptResults[0][2] = fOffset;
	    return;
	  }
	else if ( ppe[1]->GetEdge()->GetVertex() == ppvx[0] )
	  {
	    uiResults = 1;
	    pptResults[0] = ppt[1] + (fOffset * pvPlane[1]);
	    pptResults[0][2] = fOffset;
	    return;
	  }

	pvPlane[1].CompletePlaneEquation(ppt[0]);
	pvPlane[1][3] -= fOffset;
	
	ppt[0] = ppvx[0]->GetPoint();

	fDisc = ( pvPlane[1](0) * ppt[0](0) + pvPlane[1](1) * ppt[0](1) + pvPlane[1](3) );
	fDisc = fDisc * fDisc;
	fDisc = fOffset * fOffset - fDisc;

	if ( fDisc < 0.0 )
	  {
	    uiResults = 0;
	    return;
	  }

	fDisc = sqrt( fDisc );

	fX = ( pvPlane[1](1) * ( pvPlane[1](1) * ppt[0](0) - pvPlane[1](0) * ppt[0](1) ) - 
	       pvPlane[1](0) * ( pvPlane[1](3) ) );
	fY = ( pvPlane[1](0) * ( pvPlane[1](1) * ppt[0](0) - pvPlane[1](0) * ppt[0](1) ) + 
	       pvPlane[1](1) * ( pvPlane[1](3) ) );

	pptResults[0].Set(fX + pvPlane[1](1) * fDisc, 
			  -(fY + pvPlane[1](0) * fDisc), 
			  fOffset);
	pptResults[1].Set(fX - pvPlane[1](1) * fDisc, 
			  -(fY - pvPlane[1](0) * fDisc), 
			  fOffset);
	if ( fDisc == 0.0 )
	  {
	    uiResults = 1;
	  }
	else
	  {
	    QEVertex *ppvxEndpoint[2];
	    FLOAT pfDistance[2];

	    uiResults = 2;

	    pfDistance[0] = CalculateDistance(ppfa[1], pptResults[0], &(ppvxEndpoint[0])); 
	    pfDistance[1] = CalculateDistance(ppfa[1], pptResults[1], &(ppvxEndpoint[1])); 

	    if ( ppvxEndpoint[0] != NULL )
	      {
		pptResults[0] = pptResults[1];
		uiResults--;
	      }
	    if ( ppvxEndpoint[1] != NULL )
	      {
		uiResults--;
	      }
	  }
      }
      break;
    case 2:
      {
	for ( i = 0; i < 2; i++ )
	  {
	    ppt[0] = ppe[i]->GetVertex()->GetPoint();
	    ppt[1] = ppe[i]->GetEdge()->GetVertex()->GetPoint();
	    pvPlane[i] = ppt[1] - ppt[0];
	    pvPlane[i].Set(-pvPlane[i](1), pvPlane[i](0), 0.0, 0.0);

	    fMag = pvPlane[i].Magnitude();
	    if ( !pvPlane[i].Normalize() ||
		 fabs(fMag) <= s_cfEpsilonVoronoiOffsetDen )
	      {
		uiResults = 0;
		return;
	      }

	    pvPlane[i].CompletePlaneEquation(ppt[0]);
	    pvPlane[i][3] -= fOffset;
	    pvPlane[i][3] = -pvPlane[i][3];
	  }

	fDen = pvPlane[0](0)*pvPlane[1](1) - pvPlane[1](0)*pvPlane[0](1);
	fX = pvPlane[0](3)*pvPlane[1](1) - pvPlane[1](3)*pvPlane[0](1);
	fY = pvPlane[0](0)*pvPlane[1](3) - pvPlane[1](0)*pvPlane[0](3);
	
	if ( fabs(fDen) <= s_cfEpsilonVoronoiOffsetDen )
	  {
	    if ( ppe[0]->GetEdge()->GetVertex() == ppe[1]->GetVertex() )
	      {
		pptResults[0] = ppe[1]->GetVertex()->GetPoint();
		pvPlane[0][3] = 0.0;
		pptResults[0] += pvPlane[0]*fOffset;
		pptResults[0][2] = fOffset;
		uiResults = 1;
	      }
	    else if ( ppe[0]->GetVertex() == ppe[1]->GetEdge()->GetVertex() )
	      {
		pptResults[0] = ppe[0]->GetVertex()->GetPoint();
		pvPlane[0][3] = 0.0;
		pptResults[0] += pvPlane[0]*fOffset;
		pptResults[0][2] = fOffset;
		uiResults = 1;
	      }
	    else
	      {
		uiResults = 0;
		pptResults[0].Null();
	      }
	  }
	else
	  {
	    uiResults = 1;
	    pptResults[0].Set(fX/fDen, fY/fDen, fOffset);
	  }
      }
      break;
    default:
      {
	uiResults = 0;
      }
      break;
    }
}

VOID QEMesh::TriangulateDelaunayIncremental()
{
  ASSERT( m_plistQEVertex->GetLength() >= 3 );

  static BOOL bFirst = TRUE;
  static CListInt listQEEdgeQueue;

  if ( bFirst )
    {
      bFirst = FALSE;
      listQEEdgeQueue.Init();
    }

  CObjectListIter<QEVertex> iter;

  QEFacetIter iterQEFacet;
  QEFacet *ppfaInit[2];
  QEVertex *ppvxInit[4], *pvx;
  QEEdge *ppeInit[4], *pe, *peCurr;
  QEFacetNode *pfnLeaf;
  QEFacet *pfa;

  Point ptMin, ptMax, pt;
  Point ptCenter;
  Vector v, w;
  UINT_32 i, uiIncidentEdges;

  if ( GetRandomize() )
    {
      RandomizeVertexList(m_plistQEVertex);
    }

  //////////////////////////////////////////////////
  // Compute the bounding box

  iter.Init(m_plistQEVertex);
  pvx = iter.PeekFirst();
  ASSERT( pvx != NULL );
  pt = pvx->GetPoint();
  ptMin = pt;
  ptMax = pt;
  for ( pvx = iter.PeekNext();
	pvx != NULL;
	pvx = iter.PeekNext() )
    {
      pt = pvx->GetPoint();
      ptMin[0] = MIN(ptMin[0], pt[0]);
      ptMin[1] = MIN(ptMin[1], pt[1]);
      ptMin[2] = MIN(ptMin[2], pt[2]);
      ptMax[0] = MAX(ptMax[0], pt[0]);
      ptMax[1] = MAX(ptMax[1], pt[1]);
      ptMax[2] = MAX(ptMax[2], pt[2]);
    }

  ptCenter = ( ptMin + ptMax ) / 2.0;
  v = ptMax - ptMin;

  //////////////////////////////////////////////////
  // START: Construct the initial 2 triangles

  ppfaInit[0] = CreateFacet();
  ASSERT( ppfaInit[0] != NULL );

  for ( i = 0; i < 4; i++ )
    {
      w = v;
      w = 100.0 * v;
      switch ( i )
	{
	case 0:
	  break;
	case 1:
	  w[X] = -w[X];
	  break;
	case 2:
	  w[X] = -w[X];
	  w[Y] = -w[Y];
	  break;
	case 3:
	  w[Y] = -w[Y];
	  break;
	default:
	  ASSERT( FALSE );
	  break;
	}

      pt = ptCenter + w;

      ppvxInit[i] = CreateVertex(pt);
      ASSERT( ppvxInit[i] != NULL );
    }

  ppeInit[0] = CreateContour(ppfaInit[0], ppvxInit[0]);
  for ( i = 1; i < 4; i++ )
    {
      ppeInit[i] = AddContourVertex(ppeInit[0], ppvxInit[i]);
    }

  pe = InsertDiagonal(ppeInit[0], ppeInit[2]);
  
  ppfaInit[1] = CreateFacet();
  ASSERT( ppfaInit[1] != NULL );
  peCurr = pe;
  do
    {
      peCurr->SetFacet(ppfaInit[1]);
      peCurr = peCurr->GetEdgeNext(0);
    }
  while ( peCurr != pe );

  // END  : Construct the initial 2 triangles
  //////////////////////////////////////////////////

  //////////////////////////////////////////////////
  // Construct the triangles and the initial search DAG
  m_penRoot = CreateEdgeNode(ppvxInit[2]->GetPoint(), ppvxInit[0]->GetPoint());
  ASSERT( m_penRoot != NULL ); 

  for ( i = 0; i < 2; i++ )
    {
      pfnLeaf = CreateFacetNode();
      ASSERT( pfnLeaf != NULL ); 
      
      pfnLeaf->SetParents(m_penRoot, NULL);
      pfnLeaf->SetFacet(ppfaInit[i]);
      ppfaInit[i]->SetFacetNode(pfnLeaf);

      m_penRoot->SetChild(i, pfnLeaf);
    }

  //////////////////////////////////////////////////
  // Randomly insert vertices
  iter.Init(m_plistQEVertex);
  for ( pvx = iter.PeekFirst(), i = 0;
	pvx != NULL && i < m_plistQEVertex->GetLength()-4;
	pvx = iter.PeekNext(), i++ )
    {
      pfa = m_penRoot->FindFacet(pvx->GetPoint());
      ASSERT( pfa != NULL && pfa != m_pfaNULL );

      uiIncidentEdges = 0;
      pe = NULL;
      iterQEFacet.Init(pfa);
      for ( peCurr = iterQEFacet.PeekFirstEdge();
	    peCurr != NULL;
	    peCurr = iterQEFacet.PeekNextEdge() )
	{
	  if ( !LeftOfEdge(peCurr, pvx->GetPoint()) )
	    {
	      pe = peCurr;
	      uiIncidentEdges++;
	    }
	}

      switch ( uiIncidentEdges )
	{
	case 0:
	  InsertVertexUpdateSearch(pfa, pvx, &listQEEdgeQueue);
	  break;
	case 1:
	  InsertVertexUpdateSearch(pe, pvx, &listQEEdgeQueue);
	  break;
	case 2:
	  //DestroyVertex, pvx);
	  break;
	default:
	  ASSERT( FALSE );
	}

      for ( pe = (QEEdge*)listQEEdgeQueue.RemoveFirst();
	    pe != NULL;
	    pe = (QEEdge*)listQEEdgeQueue.RemoveFirst() )
	{
	  if ( !ConstrainedDelaunayEdge(pe) )
	    {
	      //fprintf(stderr, "Flip: %p: ", pe);
	      //pe->GetVertex()->Display();
	      //pe->GetEdge()->GetVertex()->Display();
	      DelaunayFlipUpdateSearch(pe, &listQEEdgeQueue);
	    }
	}
    }
  ASSERT( pvx != NULL && i == m_plistQEVertex->GetLength()-4 );

  ASSERT( listQEEdgeQueue.GetLength() == 0 );

  RemoveDelaunayBoundingBox();
}

VOID QEMesh::TriangulateDelaunayIncrementalSlow()
{
  ASSERT( m_plistQEVertex->GetLength() >= 3 );

  static BOOL bFirst = TRUE;
  static CListInt listQEEdgeQueue;

  if ( bFirst )
    {
      bFirst = FALSE;
      listQEEdgeQueue.Init();
    }

  CObjectListIter<QEVertex> iter;

  QEFacetIter iterQEFacet;
  QEFacet *ppfaInit[2];
  QEVertex *ppvxInit[4], *pvx;
  QEEdge *ppeInit[4], *pe, *peCurr;
  QEFacet *pfa;

  Point ptMin, ptMax, pt;
  Point ptCenter;
  Vector v, w;
  UINT_32 i, uiIncidentEdges;

  if ( GetRandomize() )
    {
      RandomizeVertexList(m_plistQEVertex);
    }

  //////////////////////////////////////////////////
  // Compute the bounding box

  iter.Init(m_plistQEVertex);
  pvx = iter.PeekFirst();
  ASSERT( pvx != NULL );
  pt = pvx->GetPoint();
  ptMin = pt;
  ptMax = pt;
  for ( pvx = iter.PeekNext();
	pvx != NULL;
	pvx = iter.PeekNext() )
    {
      pt = pvx->GetPoint();
      ptMin[0] = MIN(ptMin[0], pt[0]);
      ptMin[1] = MIN(ptMin[1], pt[1]);
      ptMin[2] = MIN(ptMin[2], pt[2]);
      ptMax[0] = MAX(ptMax[0], pt[0]);
      ptMax[1] = MAX(ptMax[1], pt[1]);
      ptMax[2] = MAX(ptMax[2], pt[2]);
    }

  ptCenter = ( ptMin + ptMax ) / 2.0;
  v = ptMax - ptMin;

  //////////////////////////////////////////////////
  // START: Construct the initial 2 triangles

  ppfaInit[0] = CreateFacet();
  ASSERT( ppfaInit[0] != NULL );

  for ( i = 0; i < 4; i++ )
    {
      //w = v;
      w = 100.0 * v;
      switch ( i )
	{
	case 0:
	  break;
	case 1:
	  w[X] = -w[X];
	  break;
	case 2:
	  w[X] = -w[X];
	  w[Y] = -w[Y];
	  break;
	case 3:
	  w[Y] = -w[Y];
	  break;
	default:
	  ASSERT( FALSE );
	  break;
	}

      pt = ptCenter + w;

      ppvxInit[i] = CreateVertex(pt);
      ASSERT( ppvxInit[i] != NULL );
    }

  ppeInit[0] = CreateContour(ppfaInit[0], ppvxInit[0]);
  for ( i = 1; i < 4; i++ )
    {
      ppeInit[i] = AddContourVertex(ppeInit[0], ppvxInit[i]);
    }

  pe = InsertDiagonal(ppeInit[0], ppeInit[2]);
  
  ppfaInit[1] = CreateFacet();
  ASSERT( ppfaInit[1] != NULL );
  peCurr = pe;
  do
    {
      peCurr->SetFacet(ppfaInit[1]);
      peCurr = peCurr->GetEdgeNext(0);
    }
  while ( peCurr != pe );

  // END  : Construct the initial 2 triangles
  //////////////////////////////////////////////////

  //////////////////////////////////////////////////
  // Randomly insert vertices
  iter.Init(m_plistQEVertex);
  for ( pvx = iter.PeekFirst(), i = 0;
	pvx != NULL && i < m_plistQEVertex->GetLength()-4;
	pvx = iter.PeekNext(), i++ )
    {
      pfa = FindFacet(pvx->GetPoint());
      ASSERT( pfa != NULL && pfa != m_pfaNULL );

      uiIncidentEdges = 0;
      pe = NULL;
      iterQEFacet.Init(pfa);
      for ( peCurr = iterQEFacet.PeekFirstEdge();
	    peCurr != NULL;
	    peCurr = iterQEFacet.PeekNextEdge() )
	{
	  if ( !LeftOfEdge(peCurr, pvx->GetPoint()) )
	    {
	      pe = peCurr;
	      uiIncidentEdges++;
	    }
	}

      switch ( uiIncidentEdges )
	{
	case 0:
	  InsertVertex(pfa, pvx, &listQEEdgeQueue);
	  break;
	case 1:
	  InsertVertex(pe, pvx, &listQEEdgeQueue);
	  break;
	case 2:
	  //DestroyVertex, pvx);
	  break;
	default:
	  ASSERT( FALSE );
	}

      for ( pe = (QEEdge*)listQEEdgeQueue.RemoveFirst();
	    pe != NULL;
	    pe = (QEEdge*)listQEEdgeQueue.RemoveFirst() )
	{
	  if ( !ConstrainedDelaunayEdge(pe) )
	    {
	      //fprintf(stderr, "Flip: %p: ", pe);
	      //pe->GetVertex()->Display();
	      //pe->GetEdge()->GetVertex()->Display();
	      DelaunayFlipIncremental(pe, &listQEEdgeQueue);
	    }
	}
    }
  ASSERT( pvx != NULL && i == m_plistQEVertex->GetLength()-4 );

  ASSERT( listQEEdgeQueue.GetLength() == 0 );

  RemoveDelaunayBoundingBox();
}

VOID QEMesh::RemoveDelaunayBoundingBox()
{
  QEVertex *ppvx[4];
  QEEdge *pe, *peCurr, *peNext;
  UINT_32 i;

  for ( i = 0; i < 4; i++ )
    {
      ppvx[i] = m_plistQEVertex->RemoveLast();
      ASSERT( ppvx[i] != NULL );
    }

  for ( i = 0; i < 4; i++ )
    {
      pe = ppvx[i]->GetEdge();
      ASSERT( pe != NULL );

      while ( pe->GetFacet() != m_pfaNULL )
	{
	  pe = pe->GetEdgeNext(1);
	}

      while ( TRUE )
	{
	  ASSERT( pe->GetFacet() == m_pfaNULL );
	  peNext = pe->GetEdgePrev(1);
	  if ( peNext == pe )
	    {
	      ASSERT( pe->GetEdge()->GetFacet() == m_pfaNULL );
	      
	      peCurr = pe->GetEdge()->GetEdgePrev(1);
	      if ( peCurr != pe->GetEdge() )
		{
		  ASSERT( peCurr->GetFacet() == m_pfaNULL );
		  Splice(1, peCurr, pe->GetEdge());
		}

	      DestroyEdge(pe->GetEdge());
	      DestroyEdge(pe);

	      break;
	    }

	  ASSERT( pe->GetEdge()->GetFacet() != m_pfaNULL );
	  DestroyFacet(pe->GetEdge()->GetFacet());

	  peCurr = pe->GetEdge();
	  peCurr->SetFacet(m_pfaNULL);
	  peCurr = peCurr->GetEdgeNext(0);
	  ASSERT( peCurr != pe->GetEdge() );
	  peCurr->SetFacet(m_pfaNULL);
	  peCurr = peCurr->GetEdgeNext(0);
	  ASSERT( peCurr != pe->GetEdge() );
	  peCurr->SetFacet(m_pfaNULL);
	  peCurr = peCurr->GetEdgeNext(0);
	  ASSERT( peCurr == pe->GetEdge() );

	  Splice(1, peNext, pe);

	  peCurr = pe->GetEdge()->GetEdgePrev(1);
	  if ( peCurr != pe->GetEdge() )
	    {
	      ASSERT( peCurr->GetFacet() == m_pfaNULL );
	      Splice(1, peCurr, pe->GetEdge());
	    }
	  
	  DestroyEdge(pe->GetEdge());
	  DestroyEdge(pe);
	  
	  pe = peNext;
	}

    }

  for ( i = 0; i < 4; i++ )
    {
      DeallocateQEVertex(ppvx[i]);
    }
}

VOID QEMesh::TriangulateDelaunayIncrementalNoBound()
{
  ASSERT( m_plistQEVertex->GetLength() >= 3 );

  static BOOL bFirst = TRUE;
  static CListInt listQEEdgeQueue;

  if ( bFirst )
    {
      bFirst = FALSE;
      listQEEdgeQueue.Init();
    }

  CObjectListIter<QEVertex> iter;

  QEVertex *pvx;
  QEEdge *pe, *peCurr, *pePrev, *peStart, *peNew;
  QEFacet *pfa;
  QEFacetIter iterQEFacet;
  UINT_32 uiIncidentEdges;

  if ( GetRandomize() )
    {
      RandomizeVertexList(m_plistQEVertex);
    }

  //////////////////////////////////////////////////
  // Construct an edge between any two vertices

  pe = CreateEdge();
  ASSERT( pe != NULL );

  iter.Init(m_plistQEVertex);
  pvx = iter.PeekFirst();
  pe->SetVertex(pvx);
  pvx = iter.PeekNext();
  pe->GetEdge()->SetVertex(pvx);

  //////////////////////////////////////////////////
  // Randomly insert vertices
  for ( pvx = iter.PeekNext();
	pvx != NULL;
	pvx = iter.PeekNext() )
    {
      pe = FindFacetEdge(pvx->GetPoint());
      ASSERT( pe != NULL );
      pfa = pe->GetFacet();
      if ( pfa == m_pfaNULL )
	{
	  if ( !LeftOfEdge(pe, pvx->GetPoint()) )
	    {
	      ASSERT( pe->GetEdge()->GetFacet() == m_pfaNULL );

	      if ( *(pe->GetEdge()->GetVertex()) < *(pe->GetVertex()) )
		{
		  pe = pe->GetEdge();
		}

	      while ( TRUE )
		{
		  if ( *pvx < *(pe->GetVertex()) )
		    {
		      if ( pe->GetEdge() == pe->GetEdgePrev(0) )
			{
			  peNew = CreateEdge();
			  ASSERT( peNew != NULL );
			  peNew->SetVertex(pvx);
			  peNew->GetEdge()->SetVertex(pe->GetVertex());
			  Splice(1, peNew->GetEdge(), pe);
			  break;
			}
		      pe = pe->GetEdgePrev(0);
		    }
		  if ( *pvx > *(pe->GetEdge()->GetVertex()) )
		    {
		      if ( pe->GetEdge() == pe->GetEdgeNext(0) )
			{
			  peNew = CreateEdge();
			  ASSERT( peNew != NULL );
			  peNew->GetEdge()->SetVertex(pvx);
			  peNew->SetVertex(pe->GetEdge()->GetVertex());
			  Splice(1, peNew, pe->GetEdge());
			  break;
			}
		      pe = pe->GetEdgeNext(0);
		    }
		  else
		    {
		      ASSERT( *pvx != *(pe->GetVertex()) &&
			      *pvx != *(pe->GetEdge()->GetVertex()) );
		      peNew = CreateEdge();
		      ASSERT( peNew != NULL );
		      peNew->SetVertex(pe->GetVertex());
		      peNew->GetEdge()->SetVertex(pvx);
		      pe->SetVertex(pvx);
		      Splice(1, peNew, pe->GetEdgePrev(1));
		      Splice(1, peNew, pe);
		      Splice(1, peNew->GetEdge(), pe);
		      break;
		    }
		}
	    }
	  else
	    {
	      // It is outside the convex hull
	      peCurr = pe->GetEdgePrev(0);
	      
	      peStart = CreateEdge();
	      ASSERT( peStart != NULL );
	      peStart->SetVertex(pvx);
	      peStart->GetEdge()->SetVertex(pe->GetVertex());
	      Splice(1, peStart->GetEdge(), pe);
	      
	      while ( Counterclockwise(pvx->GetPoint(), 
				       peCurr->GetVertex()->GetPoint(),
				       peCurr->GetEdge()->GetVertex()->GetPoint()) )
		{
		  listQEEdgeQueue.InsertLast(peCurr, 0);
		  
		  pfa = CreateFacet();
		  ASSERT( pfa != NULL );
		  pePrev = peCurr;
		  peCurr = peCurr->GetEdgePrev(0);
		  
		  InsertDiagonal(peStart, pePrev);
		  
		  pePrev->SetFacet(pfa);
		  pePrev->GetEdgePrev(0)->SetFacet(pfa);
		  pePrev->GetEdgeNext(0)->SetFacet(pfa);	      
		}
	      
	      peCurr = pe;
	      while ( Counterclockwise(pvx->GetPoint(), 
				       peCurr->GetVertex()->GetPoint(),
				       peCurr->GetEdge()->GetVertex()->GetPoint()) )
		{
		  listQEEdgeQueue.InsertLast(peCurr, 0);
		  
		  pfa = CreateFacet();
		  ASSERT( pfa != NULL );
		  pePrev = peCurr;
		  peCurr = peCurr->GetEdgeNext(0);
		  
		  InsertDiagonal(peCurr, pePrev->GetEdgePrev(0));

		  pePrev->SetFacet(pfa);
		  pePrev->GetEdgePrev(0)->SetFacet(pfa);
		  pePrev->GetEdgeNext(0)->SetFacet(pfa);	      
		}
	    }
	}
      else
	{
	  uiIncidentEdges = 0;
	  pe = NULL;
	  iterQEFacet.Init(pfa);
	  for ( peCurr = iterQEFacet.PeekFirstEdge();
		peCurr != NULL;
		peCurr = iterQEFacet.PeekNextEdge() )
	    {
	      if ( !LeftOfEdge(peCurr, pvx->GetPoint()) )
		{
		  pe = peCurr;
		  uiIncidentEdges++;
		}
	    }

	  switch ( uiIncidentEdges )
	    {
	    case 0:
	      InsertVertex(pfa, pvx, &listQEEdgeQueue);
	      break;
	    case 1:
	      InsertVertex(pe, pvx, &listQEEdgeQueue);
	      break;
	    case 2:
	      //DestroyVertex, pvx);
	      break;
	    default:
	      ASSERT( FALSE );
	    }
	}

      for ( pe = (QEEdge*)listQEEdgeQueue.RemoveFirst();
	    pe != NULL;
	    pe = (QEEdge*)listQEEdgeQueue.RemoveFirst() )
	{
	  if ( !ConstrainedDelaunayEdge(pe) )
	    {
	      //fprintf(stderr, "Flip: %p: ", pe);
	      //pe->GetVertex()->Display();
	      //pe->GetEdge()->GetVertex()->Display();
	      DelaunayFlipIncremental(pe, &listQEEdgeQueue);
	    }
	}
    }
  ASSERT( listQEEdgeQueue.GetLength() == 0 );
}

VOID QEMesh::TriangulateDelaunayDivideAndConquer()
{
  CObjectListIter<QEVertex> iter;
  CHeap<QEVertex*, QEVertex> heapQEVertex;
  QEVertex **ppvx, *pvx;
  QEEdge *ppe[2];
  UINT_32 i, uiSize;

  uiSize = m_plistQEVertex->GetLength();
  ASSERT( uiSize > 2 );

  //////////////////////////////////////////////////
  // Sort the Vertices
  heapQEVertex.Init(uiSize);
  ppvx = new QEVertex*[uiSize];
  ASSERT( ppvx != NULL );

  iter.Init(m_plistQEVertex);
  for ( pvx = iter.PeekFirst();
	pvx != NULL;
	pvx = iter.PeekNext() )
    {
      heapQEVertex.Insert(*pvx, pvx);
    }

  i = 0;
  for ( pvx = heapQEVertex.RemoveTop();
	pvx != NULL;
	pvx = heapQEVertex.RemoveTop() )
    {
      ASSERT( i < uiSize );
      ppvx[i] = pvx;
      i++;
    }
  ASSERT( i == uiSize );

  TriangulateDelaunayDivideAndConquer(uiSize, ppvx, ppe); 

  delete []ppvx;
}

VOID QEMesh::TriangulateDelaunayDivideAndConquer(UINT_32 uiSize, QEVertex **ppvx, 
						 QEEdge **ppeLeftAndRight)
{
  ASSERT( uiSize > 1 );
  ASSERT( ppvx != NULL );

  switch ( uiSize )
    {
    case 2:
      {
	ppeLeftAndRight[0] = CreateEdge();
	ASSERT( ppeLeftAndRight[0] != NULL );
	ppeLeftAndRight[0]->SetVertex(ppvx[0]);
	ppeLeftAndRight[0]->GetEdge()->SetVertex(ppvx[1]);
	ppeLeftAndRight[1] = ppeLeftAndRight[0]->GetEdge();
      }
      break;
    case 3:
      {
	QEFacet *pfa;
	QEEdge *pe;

	if ( Counterclockwise(ppvx[0]->GetPoint(), 
			      ppvx[1]->GetPoint(), 
			      ppvx[2]->GetPoint()) )
	  {
	    pfa = CreateFacet();
	    ASSERT( pfa != NULL );
	    pe = CreateContour(pfa, ppvx[0]);
	    ASSERT( pe != NULL );

	    AddContourVertex(pe, ppvx[1]);

	    ppeLeftAndRight[1] = AddContourVertex(pe, ppvx[2]);

	    ppeLeftAndRight[0] = pe->GetEdgeNext(1);
	    ppeLeftAndRight[1] = ppeLeftAndRight[1]->GetEdgeNext(1);
	  }
	else if ( Counterclockwise(ppvx[0]->GetPoint(), 
				   ppvx[2]->GetPoint(), 
				   ppvx[1]->GetPoint()) )
	  {
	    pfa = CreateFacet();
	    ASSERT( pfa != NULL );
	    pe = CreateContour(pfa, ppvx[0]);
	    ASSERT( pe != NULL );

	    ppeLeftAndRight[1] = AddContourVertex(pe, ppvx[2]);

	    AddContourVertex(pe, ppvx[1]);

	    ppeLeftAndRight[0] = pe->GetEdgeNext(1);
	    ppeLeftAndRight[1] = ppeLeftAndRight[1]->GetEdgeNext(1);
	  }
	else
	  {
	    // Colinear
	    ppeLeftAndRight[0] = CreateEdge();
	    ASSERT( ppeLeftAndRight[0] != NULL );
	    ppeLeftAndRight[0]->SetVertex(ppvx[0]);
	    ppeLeftAndRight[0]->GetEdge()->SetVertex(ppvx[1]);

	    ppeLeftAndRight[1] = CreateEdge();
	    ASSERT( ppeLeftAndRight[1] != NULL );
	    ppeLeftAndRight[1]->SetVertex(ppvx[2]);
	    ppeLeftAndRight[1]->GetEdge()->SetVertex(ppvx[1]);	    

	    Splice(1, ppeLeftAndRight[0]->GetEdge(), ppeLeftAndRight[1]->GetEdge());
	  }
      }
      break;
    default:
      {
	QEEdge *ppeLeft[2], *ppeRight[2], *peBase, *peLeft, *peRight, *pe, *peCurr;
	//QEEdge *peJordan;
	QEFacet *pfa;
	UINT_32 uiHalf;
	
	uiHalf = uiSize / 2;
	TriangulateDelaunayDivideAndConquer(uiHalf, ppvx, ppeLeft);
	TriangulateDelaunayDivideAndConquer(uiSize - uiHalf, ppvx + uiHalf, ppeRight);

	ASSERT( ppeLeft[0]->GetVertex() == ppvx[0] );
	ASSERT( ppeLeft[0]->GetFacet() == m_pfaNULL );
	ASSERT( ppeRight[1]->GetVertex() == ppvx[uiSize - 1] );
	ASSERT( ppeRight[1]->GetFacet() == m_pfaNULL );

	// Find the bottom two vertices
	while ( TRUE )
	  {
	    if ( Counterclockwise(ppeRight[0]->GetVertex()->GetPoint(),
				  ppeLeft[1]->GetVertex()->GetPoint(),
				  ppeLeft[1]->GetEdge()->GetVertex()->GetPoint()) )
	      
	      {
		ppeLeft[1] = ppeLeft[1]->GetEdgeNext(0);
	      }
	    else if ( Counterclockwise(ppeLeft[1]->GetVertex()->GetPoint(),
				       ppeRight[0]->GetEdgePrev(0)->GetVertex()->GetPoint(),
				       ppeRight[0]->GetVertex()->GetPoint()) )
	      
	      {
		ppeRight[0] = ppeRight[0]->GetEdgePrev(0);
	      }
	    else
	      {
		break;
	      }
	  }

	ASSERT( ppeLeft[1]->GetFacet() == m_pfaNULL );
	ASSERT( ppeRight[0]->GetFacet() == m_pfaNULL );
	peBase = InsertDiagonal(ppeLeft[1], ppeRight[0]);
	
	// The left can't be a problem, yet
	if ( ppeRight[1] == ppeRight[0] )
	  {
	    ppeRight[1] = ppeRight[1]->GetEdgeNext(1);
	  }
	//peJordan = ppeLeft[0]->GetEdgePrev(0);
	ppeLeft[0] = ppeLeft[0]->GetEdgePrev(0);
	
	//////////////////////////////////////////////////
	// Merge
	
	while ( TRUE )
	  {
	    peLeft = peBase->GetEdgePrev(0);
	    if ( peLeft->GetEdge()->GetFacet() != m_pfaNULL &&
		 LeftOfEdge(peBase, peLeft->GetVertex()->GetPoint()) )
	      {
		for ( peCurr = peLeft->GetEdge()->GetEdgeNext(1)->GetEdge();
		      ( peLeft->GetEdge()->GetFacet() != m_pfaNULL &&
			InCircumcircle(peBase->GetVertex()->GetPoint(),
				       peBase->GetEdge()->GetVertex()->GetPoint(),
				       peLeft->GetVertex()->GetPoint(),
				       peCurr->GetVertex()->GetPoint()) );
		      peCurr = peCurr->GetEdge()->GetEdgeNext(1)->GetEdge() )
		  {
		    DestroyFacet(peLeft->GetEdge()->GetFacet());

		    peLeft->GetEdge()->SetFacet(m_pfaNULL);
		    peLeft->GetEdge()->GetEdgeNext(0)->SetFacet(m_pfaNULL);
		    peLeft->GetEdge()->GetEdgePrev(0)->SetFacet(m_pfaNULL);

		    RemoveDiagonal(peLeft);	    

		    DestroyEdge(peLeft->GetEdge());
		    DestroyEdge(peLeft);

		    peLeft = peCurr;
		  }
	      }

	    peRight = peBase->GetEdgeNext(0);
	    if ( peRight->GetEdge()->GetFacet() != m_pfaNULL &&
		 LeftOfEdge(peBase, peRight->GetEdge()->GetVertex()->GetPoint()) )
	      {
		for ( peCurr = peRight->GetEdgePrev(1);
		      ( peRight->GetEdge()->GetFacet() != m_pfaNULL &&
			InCircumcircle(peBase->GetVertex()->GetPoint(),
				       peBase->GetEdge()->GetVertex()->GetPoint(),
				       peRight->GetEdge()->GetVertex()->GetPoint(),
				       peCurr->GetEdge()->GetVertex()->GetPoint()) );
		      peCurr = peCurr->GetEdgePrev(1) )
		  {
		    DestroyFacet(peRight->GetEdge()->GetFacet());

		    peRight->GetEdge()->SetFacet(m_pfaNULL);
		    peRight->GetEdge()->GetEdgeNext(0)->SetFacet(m_pfaNULL);
		    peRight->GetEdge()->GetEdgePrev(0)->SetFacet(m_pfaNULL);

		    RemoveDiagonal(peRight);	    

		    DestroyEdge(peRight->GetEdge());
		    DestroyEdge(peRight);

		    peRight = peCurr;
		  }
	      }

	    if ( !LeftOfEdge(peBase, peLeft->GetVertex()->GetPoint()) &&
		 !LeftOfEdge(peBase, peRight->GetEdge()->GetVertex()->GetPoint()) )
	      {
		break;
	      }
	    
	    // Create a diagonal
	    pfa = CreateFacet();
	    ASSERT( pfa != NULL );
	    pe = CreateEdge();
	    ASSERT( pe != NULL );
	    if ( !LeftOfEdge(peBase, peLeft->GetVertex()->GetPoint()) ||
		 (LeftOfEdge(peBase, peRight->GetEdge()->GetVertex()->GetPoint()) &&
		  InCircumcircle(peLeft->GetVertex()->GetPoint(),
				 peLeft->GetEdge()->GetVertex()->GetPoint(),
				 peRight->GetVertex()->GetPoint(),
				 peRight->GetEdge()->GetVertex()->GetPoint())) )
	      {
		InsertDiagonal(peBase, peRight->GetEdgeNext(0), pe);

		pe->GetEdge()->SetFacet(pfa);
		pe->GetEdge()->GetEdgeNext(0)->SetFacet(pfa);
		pe->GetEdge()->GetEdgePrev(0)->SetFacet(pfa);

		peBase = pe;
	      }
	    else
	      {
		InsertDiagonal(peLeft, peRight, pe);

		pe->GetEdge()->SetFacet(pfa);
		pe->GetEdge()->GetEdgeNext(0)->SetFacet(pfa);
		pe->GetEdge()->GetEdgePrev(0)->SetFacet(pfa);

		peBase = pe;
	      }
	  }

	ppeLeft[0] = ppeLeft[0]->GetEdgeNext(0);
	ppeLeftAndRight[0] = ppeLeft[0];
	ppeLeftAndRight[1] = ppeRight[1];

	ASSERT( ppeLeftAndRight[0]->GetVertex() == ppvx[0] );
	ASSERT( ppeLeftAndRight[0]->GetFacet() == m_pfaNULL );
	ASSERT( ppeLeftAndRight[1]->GetVertex() == ppvx[uiSize-1] );
	ASSERT( ppeLeftAndRight[1]->GetFacet() == m_pfaNULL );
      }
      break;
    }
}

VOID QEMesh::TriangulateDelaunayDivideAndConquerAlternating()
{
  CObjectListIter<QEVertex> iter;
  QEVertex **ppvx, *pvx;
  QEEdge *pppe[2][2];
  UINT_32 i, uiSize;

  uiSize = m_plistQEVertex->GetLength();
  ASSERT( uiSize > 2 );

  //////////////////////////////////////////////////
  // Sort the Vertices
  ppvx = new QEVertex*[uiSize];
  ASSERT( ppvx != NULL );

  iter.Init(m_plistQEVertex);
  i = 0;
  for ( pvx = iter.PeekFirst();
	pvx != NULL;
	pvx = iter.PeekNext() )
    {
      ASSERT( i < uiSize );
      ppvx[i] = pvx;
      //ppvxTemp[i] = pvx;
      i++;
    }
  ASSERT( i == uiSize );

  SortAlternating(QEVertex::ctYXZ, uiSize, ppvx);
  TriangulateDelaunayDivideAndConquerAlternating(QEVertex::ctYXZ, uiSize, ppvx, pppe); 

  delete []ppvx;
}

VOID QEMesh::TriangulateDelaunayDivideAndConquerAlternating(QEVertex::ComparisonType eDirection, 
							    UINT_32 uiSize, QEVertex **ppvx,
							    QEEdge *pppeLeftAndRight[2][2])
{
  ASSERT( eDirection == QEVertex::ctXYZ || eDirection == QEVertex::ctYXZ );
  ASSERT( uiSize > 1 );
  ASSERT( ppvx != NULL );
  ASSERT( pppeLeftAndRight != NULL );

  QEVertex::ComparisonType eOpposite;
  UINT_32 uiDir, uiOpp;

  if ( eDirection == QEVertex::ctXYZ )
    {
      uiDir = 0;
      uiOpp = 1;
      eOpposite = QEVertex::ctYXZ;
    }
  else
    {
      uiDir = 1;
      uiOpp = 0;
      eOpposite = QEVertex::ctXYZ;
    }

  switch ( uiSize )
    {
    case 2:
      {
	pppeLeftAndRight[uiDir][0] = CreateEdge();
	ASSERT( pppeLeftAndRight[uiDir][0] != NULL );
	pppeLeftAndRight[uiDir][0]->SetVertex(ppvx[0]);
	pppeLeftAndRight[uiDir][0]->GetEdge()->SetVertex(ppvx[1]);
	pppeLeftAndRight[uiDir][1] = pppeLeftAndRight[uiDir][0]->GetEdge();

	QEVertex::SetComparisonType(eOpposite);
	if ( (*ppvx[0]) > (*ppvx[1]) )
	  {
	    pppeLeftAndRight[uiOpp][0] = pppeLeftAndRight[uiDir][1];
	    pppeLeftAndRight[uiOpp][1] = pppeLeftAndRight[uiDir][0];
	  }
	else
	  {
	    pppeLeftAndRight[uiOpp][0] = pppeLeftAndRight[uiDir][0];
	    pppeLeftAndRight[uiOpp][1] = pppeLeftAndRight[uiDir][1];
	  }
	QEVertex::SetComparisonTypeDefault();

	ASSERT( pppeLeftAndRight[uiDir][0]->GetFacet() == m_pfaNULL );
	ASSERT( pppeLeftAndRight[uiDir][1]->GetFacet() == m_pfaNULL );
	ASSERT( pppeLeftAndRight[uiOpp][0]->GetFacet() == m_pfaNULL );
	ASSERT( pppeLeftAndRight[uiOpp][1]->GetFacet() == m_pfaNULL );
      }
      break;
    case 3:
      {
	QEFacet *pfa;
	QEEdge *pe;

	if ( Counterclockwise(ppvx[0]->GetPoint(), 
			      ppvx[1]->GetPoint(), 
			      ppvx[2]->GetPoint()) )
	  {
	    pfa = CreateFacet();
	    ASSERT( pfa != NULL );
	    pe = CreateContour(pfa, ppvx[0]);
	    ASSERT( pe != NULL );

	    AddContourVertex(pe, ppvx[1]);

	    pppeLeftAndRight[uiDir][1] = AddContourVertex(pe, ppvx[2]);

	    pppeLeftAndRight[uiDir][0] = pe->GetEdgeNext(1);
	    pppeLeftAndRight[uiDir][1] = pppeLeftAndRight[uiDir][1]->GetEdgeNext(1);

	    QEVertex::SetComparisonType(eOpposite);
	    pe = pppeLeftAndRight[uiDir][0];
	    do
	      {
		if ( (*(pe->GetVertex())) < (*(pe->GetEdgeNext(0)->GetVertex())) && 
		     (*(pe->GetVertex())) < (*(pe->GetEdgePrev(0)->GetVertex())) ) 
		  {
		    pppeLeftAndRight[uiOpp][0] = pe;
		  }
		if ( (*(pe->GetVertex())) > (*(pe->GetEdgeNext(0)->GetVertex())) && 
		     (*(pe->GetVertex())) > (*(pe->GetEdgePrev(0)->GetVertex())) ) 
		  {
		    pppeLeftAndRight[uiOpp][1] = pe;
		  }

		  pe = pe->GetEdgeNext(0);
	      }
	    while ( pe !=  pppeLeftAndRight[uiDir][0] );
	    QEVertex::SetComparisonTypeDefault();
	  }
	else if ( Counterclockwise(ppvx[0]->GetPoint(), 
				   ppvx[2]->GetPoint(), 
				   ppvx[1]->GetPoint()) )
	  {
	    pfa = CreateFacet();
	    ASSERT( pfa != NULL );
	    pe = CreateContour(pfa, ppvx[0]);
	    ASSERT( pe != NULL );

	    pppeLeftAndRight[uiDir][1] = AddContourVertex(pe, ppvx[2]);

	    AddContourVertex(pe, ppvx[1]);

	    pppeLeftAndRight[uiDir][0] = pe->GetEdgeNext(1);
	    pppeLeftAndRight[uiDir][1] = pppeLeftAndRight[uiDir][1]->GetEdgeNext(1);

	    QEVertex::SetComparisonType(eOpposite);
	    pe = pppeLeftAndRight[uiDir][0];
	    do
	      {
		if ( (*(pe->GetVertex())) < (*(pe->GetEdgeNext(0)->GetVertex())) && 
		     (*(pe->GetVertex())) < (*(pe->GetEdgePrev(0)->GetVertex())) ) 
		  {
		    pppeLeftAndRight[uiOpp][0] = pe;
		  }
		if ( (*(pe->GetVertex())) > (*(pe->GetEdgeNext(0)->GetVertex())) && 
		     (*(pe->GetVertex())) > (*(pe->GetEdgePrev(0)->GetVertex())) ) 
		  {
		    pppeLeftAndRight[uiOpp][1] = pe;
		  }

		  pe = pe->GetEdgeNext(0);
	      }
	    while ( pe !=  pppeLeftAndRight[uiDir][0] );
	    QEVertex::SetComparisonTypeDefault();
	  }
	else
	  {
	    // Colinear
	    pppeLeftAndRight[uiDir][0] = CreateEdge();
	    ASSERT( pppeLeftAndRight[uiDir][0] != NULL );
	    pppeLeftAndRight[uiDir][0]->SetVertex(ppvx[0]);
	    pppeLeftAndRight[uiDir][0]->GetEdge()->SetVertex(ppvx[1]);

	    pppeLeftAndRight[uiDir][1] = CreateEdge();
	    ASSERT( pppeLeftAndRight[uiDir][1] != NULL );
	    pppeLeftAndRight[uiDir][1]->SetVertex(ppvx[2]);
	    pppeLeftAndRight[uiDir][1]->GetEdge()->SetVertex(ppvx[1]);	    

	    Splice(1, pppeLeftAndRight[uiDir][0]->GetEdge(), 
		   pppeLeftAndRight[uiDir][1]->GetEdge());

	    QEVertex::SetComparisonType(eOpposite);
	    if ( (*ppvx[0]) > (*ppvx[2]) )
	      {
		pppeLeftAndRight[uiOpp][0] = pppeLeftAndRight[uiDir][1];
		pppeLeftAndRight[uiOpp][1] = pppeLeftAndRight[uiDir][0];
	      }
	    else
	      {
		pppeLeftAndRight[uiOpp][0] = pppeLeftAndRight[uiDir][0];
		pppeLeftAndRight[uiOpp][1] = pppeLeftAndRight[uiDir][1];
	      }
	    QEVertex::SetComparisonTypeDefault();
	  }

	ASSERT( pppeLeftAndRight[uiDir][0]->GetFacet() == m_pfaNULL );
	ASSERT( pppeLeftAndRight[uiDir][1]->GetFacet() == m_pfaNULL );
	ASSERT( pppeLeftAndRight[uiOpp][0]->GetFacet() == m_pfaNULL );
	ASSERT( pppeLeftAndRight[uiOpp][1]->GetFacet() == m_pfaNULL );
      }
      break;
    default:
      {
	QEEdge *pppeLeft[2][2], *pppeRight[2][2], *peBase, *peLeft, *peRight, *pe, *peCurr;
	QEFacet *pfa;
	UINT_32 uiHalf;
	
	uiHalf = uiSize / 2;
	TriangulateDelaunayDivideAndConquerAlternating(eOpposite, uiHalf, ppvx, pppeLeft);
	TriangulateDelaunayDivideAndConquerAlternating(eOpposite, uiSize - uiHalf, ppvx + uiHalf, pppeRight);

	ASSERT( pppeLeft[uiDir][0]->GetFacet() == m_pfaNULL );
	ASSERT( pppeRight[uiDir][1]->GetFacet() == m_pfaNULL );

	// Find the bottom two vertices
	while ( TRUE )
	  {
	    if ( Counterclockwise(pppeRight[uiDir][0]->GetVertex()->GetPoint(),
				  pppeLeft[uiDir][1]->GetVertex()->GetPoint(),
				  pppeLeft[uiDir][1]->GetEdge()->GetVertex()->GetPoint()) )
	      
	      {
		pppeLeft[uiDir][1] = pppeLeft[uiDir][1]->GetEdgeNext(0);
	      }
	    else if ( Counterclockwise(pppeLeft[uiDir][1]->GetVertex()->GetPoint(),
				       pppeRight[uiDir][0]->GetEdgePrev(0)->GetVertex()->GetPoint(),
				       pppeRight[uiDir][0]->GetVertex()->GetPoint()) )
	      
	      {
		pppeRight[uiDir][0] = pppeRight[uiDir][0]->GetEdgePrev(0);
	      }
	    else
	      {
		break;
	      }
	  }

	ASSERT( pppeLeft[uiDir][1]->GetFacet() == m_pfaNULL );
	ASSERT( pppeRight[uiDir][0]->GetFacet() == m_pfaNULL );
	peBase = InsertDiagonal(pppeLeft[uiDir][1], pppeRight[uiDir][0]);
	
	// Check if the right edge has been enclosed by the bottom base edge
	if ( pppeRight[uiDir][1] == pppeRight[uiDir][0] )
	  {
	    pppeRight[uiDir][1] = pppeRight[uiDir][1]->GetEdgeNext(1);
	    ASSERT( pppeRight[uiDir][1] == peBase->GetEdge() );
	  }
	// Move the left edge one position ccw around the convex hull,
	// so that we can recover the left edge again at the end
	pppeLeft[uiDir][0] = pppeLeft[uiDir][0]->GetEdgePrev(0);
	
	// Set the bottom edge
	pppeLeftAndRight[uiOpp][uiDir] = peBase->GetEdge();
	pppeLeftAndRight[uiOpp][uiOpp] = peBase;

	//////////////////////////////////////////////////
	// Merge
	
	while ( TRUE )
	  {
	    peLeft = peBase->GetEdgePrev(0);
	    if ( peLeft->GetEdge()->GetFacet() != m_pfaNULL &&
		 LeftOfEdge(peBase, peLeft->GetVertex()->GetPoint()) )
	      {
		for ( peCurr = peLeft->GetEdge()->GetEdgeNext(1)->GetEdge();
		      ( peLeft->GetEdge()->GetFacet() != m_pfaNULL &&
			InCircumcircle(peBase->GetVertex()->GetPoint(),
				       peBase->GetEdge()->GetVertex()->GetPoint(),
				       peLeft->GetVertex()->GetPoint(),
				       peCurr->GetVertex()->GetPoint()) );
		      peCurr = peCurr->GetEdge()->GetEdgeNext(1)->GetEdge() )
		  {
		    DestroyFacet(peLeft->GetEdge()->GetFacet());

		    peLeft->GetEdge()->SetFacet(m_pfaNULL);
		    peLeft->GetEdge()->GetEdgeNext(0)->SetFacet(m_pfaNULL);
		    peLeft->GetEdge()->GetEdgePrev(0)->SetFacet(m_pfaNULL);

		    RemoveDiagonal(peLeft);	    

		    DestroyEdge(peLeft->GetEdge());
		    DestroyEdge(peLeft);

		    peLeft = peCurr;
		  }
	      }

	    peRight = peBase->GetEdgeNext(0);
	    if ( peRight->GetEdge()->GetFacet() != m_pfaNULL &&
		 LeftOfEdge(peBase, peRight->GetEdge()->GetVertex()->GetPoint()) )
	      {
		for ( peCurr = peRight->GetEdgePrev(1);
		      ( peRight->GetEdge()->GetFacet() != m_pfaNULL &&
			InCircumcircle(peBase->GetVertex()->GetPoint(),
				       peBase->GetEdge()->GetVertex()->GetPoint(),
				       peRight->GetEdge()->GetVertex()->GetPoint(),
				       peCurr->GetEdge()->GetVertex()->GetPoint()) );
		      peCurr = peCurr->GetEdgePrev(1) )
		  {
		    DestroyFacet(peRight->GetEdge()->GetFacet());

		    peRight->GetEdge()->SetFacet(m_pfaNULL);
		    peRight->GetEdge()->GetEdgeNext(0)->SetFacet(m_pfaNULL);
		    peRight->GetEdge()->GetEdgePrev(0)->SetFacet(m_pfaNULL);

		    RemoveDiagonal(peRight);	    

		    DestroyEdge(peRight->GetEdge());
		    DestroyEdge(peRight);

		    peRight = peCurr;
		  }
	      }

	    if ( !LeftOfEdge(peBase, peLeft->GetVertex()->GetPoint()) &&
		 !LeftOfEdge(peBase, peRight->GetEdge()->GetVertex()->GetPoint()) )
	      {
		break;
	      }
	    
	    // Create a diagonal
	    pfa = CreateFacet();
	    ASSERT( pfa != NULL );
	    pe = CreateEdge();
	    ASSERT( pe != NULL );
	    if ( !LeftOfEdge(peBase, peLeft->GetVertex()->GetPoint()) ||
		 (LeftOfEdge(peBase, peRight->GetEdge()->GetVertex()->GetPoint()) &&
		  InCircumcircle(peLeft->GetVertex()->GetPoint(),
				 peLeft->GetEdge()->GetVertex()->GetPoint(),
				 peRight->GetVertex()->GetPoint(),
				 peRight->GetEdge()->GetVertex()->GetPoint())) )
	      {
		InsertDiagonal(peBase, peRight->GetEdgeNext(0), pe);

		pe->GetEdge()->SetFacet(pfa);
		pe->GetEdge()->GetEdgeNext(0)->SetFacet(pfa);
		pe->GetEdge()->GetEdgePrev(0)->SetFacet(pfa);

		peBase = pe;
	      }
	    else
	      {
		InsertDiagonal(peLeft, peRight, pe);

		pe->GetEdge()->SetFacet(pfa);
		pe->GetEdge()->GetEdgeNext(0)->SetFacet(pfa);
		pe->GetEdge()->GetEdgePrev(0)->SetFacet(pfa);

		peBase = pe;
	      }

	    // Set the top edge
	    pppeLeftAndRight[uiOpp][uiOpp] = peBase;
	  }

	pppeLeft[uiDir][0] = pppeLeft[uiDir][0]->GetEdgeNext(0);
	pppeLeftAndRight[uiDir][0] = pppeLeft[uiDir][0];
	pppeLeftAndRight[uiDir][1] = pppeRight[uiDir][1];

	// Set the bottom and the top
	QEVertex::SetComparisonType(eOpposite);
	// Bottom
	while ( (*(pppeLeftAndRight[uiOpp][0]->GetVertex())) > 
		(*(pppeLeftAndRight[uiOpp][0]->GetEdgeNext(0)->GetVertex())) )
	  {
	    pppeLeftAndRight[uiOpp][0] = pppeLeftAndRight[uiOpp][0]->GetEdgeNext(0);
	  }
	while ( (*(pppeLeftAndRight[uiOpp][0]->GetVertex())) > 
		(*(pppeLeftAndRight[uiOpp][0]->GetEdgePrev(0)->GetVertex())) )
	  {
	    pppeLeftAndRight[uiOpp][0] = pppeLeftAndRight[uiOpp][0]->GetEdgePrev(0);
	  }
	
	// Top
	while ( (*(pppeLeftAndRight[uiOpp][1]->GetVertex())) <
		(*(pppeLeftAndRight[uiOpp][1]->GetEdgeNext(0)->GetVertex())) )
	  {
	    pppeLeftAndRight[uiOpp][1] = pppeLeftAndRight[uiOpp][1]->GetEdgeNext(0);
	  }
	while ( (*(pppeLeftAndRight[uiOpp][1]->GetVertex())) <
		(*(pppeLeftAndRight[uiOpp][1]->GetEdgePrev(0)->GetVertex())) )
	  {
	    pppeLeftAndRight[uiOpp][1] = pppeLeftAndRight[uiOpp][1]->GetEdgePrev(0);
	  }
	QEVertex::SetComparisonTypeDefault();

	ASSERT( pppeLeftAndRight[uiDir][0]->GetFacet() == m_pfaNULL );
	ASSERT( pppeLeftAndRight[uiDir][1]->GetFacet() == m_pfaNULL );
	ASSERT( pppeLeftAndRight[uiOpp][0]->GetFacet() == m_pfaNULL );
	ASSERT( pppeLeftAndRight[uiOpp][1]->GetFacet() == m_pfaNULL );
      }
      break;
    }
}

VOID QEMesh::TriangulateDelaunayConvexHull()
{
  CObjectList<QEVertex> listQEVertex;
  CObjectList<QEEdge> listQEEdge;
  CObjectListIter<QEVertex> iterQEVertex;
  //CObjectListIter<QEEdge> iterQEEdge;
  CObjectListIter<QEFacet> iterQEFacet;
  QEVertex *pvx;
  QEEdge *pe, *peContour, *peOpp;
  QEFacet *pfa, *pfaNext;
  Point pt;

  listQEVertex.Splice(m_plistQEVertex);
  iterQEVertex.Init(&listQEVertex);
  for ( pvx = iterQEVertex.PeekFirst();
	pvx != NULL;
	pvx = iterQEVertex.PeekNext() )
    {
      pt = pvx->GetPoint();
      pt[Z] = -(pt(X) * pt(X) + pt(Y) * pt(Y));
      pvx->SetPoint(pt);
    }

  CreateConvexHull(&listQEVertex);

  for ( pvx = listQEVertex.RemoveFirst();
	pvx != NULL;
	pvx = listQEVertex.RemoveFirst() )
    {
      DeallocateQEVertex(pvx);
    }

  // Remove the negatively facing facets
  iterQEFacet.Init(m_plistQEFacet);
  for ( pfa = iterQEFacet.PeekFirst();
	pfa != NULL; )
    {
      pfaNext = iterQEFacet.PeekNext();

      if ( pfa->GetNormal()(Z) <= 0.0 )
	{
	  peContour = pfa->GetEdge();

	  // Set its edges' facet pointers to m_pfaNULL
	  pe = peContour;
	  do
	    {
	      pe->SetFacet(m_pfaNULL);
	      if ( pe->GetEdge()->GetFacet() == m_pfaNULL ||
		   pe->GetEdge()->GetFacet()->GetNormal()(Z) <= 0.0 )
		{
		  RemoveEdge(pe);
		  listQEEdge.InsertLast(pe);
		}
	      pe = pe->GetEdgeNext(0);
	    }
	  while ( pe != peContour );
	  
	  DestroyFacet(pfa);
	}
      pfa = pfaNext;
    }

  for ( pe = listQEEdge.RemoveFirst();
	pe != NULL;
	pe = listQEEdge.RemoveFirst() )
    {
      peOpp = pe->GetEdge();
      listQEEdge.Remove(peOpp);
      
      ASSERT( pe->GetFacet() == m_pfaNULL );
      if ( pe->GetEdgeNext(1) == pe )
	{
	  DestroyVertex(pe->GetVertex());
	}
      else
	{
	  Splice(1, pe->GetEdgePrev(1), pe);
	}
      
      ASSERT( peOpp->GetFacet() == m_pfaNULL );
      if ( peOpp->GetEdgeNext(1) == peOpp )
	{
	  DestroyVertex(peOpp->GetVertex());
	}
      else
	{
	  Splice(1, peOpp->GetEdgePrev(1), peOpp);
	}
      
      DeallocateQEEdge(pe);
      DeallocateQEEdge(peOpp);
    }

  // Project the paraboloid back onto the XY-plane
  iterQEVertex.Init(m_plistQEVertex);
  for ( pvx = iterQEVertex.PeekFirst();
	pvx != NULL;
	pvx = iterQEVertex.PeekNext() )
    {
      pt = pvx->GetPoint();
      pt[Z] = 0.0;
      pvx->SetPoint(pt);
    }
}

VOID QEMesh::CreateVoronoiDelaunay(QEMesh *pmeDelaunay)
{
  ASSERT( pmeDelaunay != NULL );

  QEMeshIter iterQEMesh;
  //QEFacetIter iterQEFacet;
  QEVertex *pvx, *pvxNew;
  QEFacet *pfa;
  QEEdge *pe, *peNew, *peBoundary;
  Point pt;
  Vector v;

  pmeDelaunay->ClearForwardPointers();

  iterQEMesh.Init(pmeDelaunay);

  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pfa = CreateFacet();
      pvx->SetForwardFacet(pfa);
    }

  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      pt = pfa->CalculateCircumcenter();
      pvx = CreateVertex(pt);
      pfa->SetForwardVertex(pvx);
    }

  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      if ( pe->GetForwardEdge() == NULL )
	{
	  if ( pe->GetFacet() == m_pfaNULL )
	    {
	      pe = pe->GetEdge();
	    }

	  peNew = CreateEdge();
	  pe->SetForwardEdge(peNew);
	  pe->GetEdge()->SetForwardEdge(peNew->GetEdge());

	  peNew->SetFacet(pe->GetEdge()->GetVertex()->GetForwardFacet());
	  peNew->GetEdge()->SetFacet(pe->GetVertex()->GetForwardFacet());

	  ASSERT( pe->GetFacet() != m_pfaNULL );
	  peNew->SetVertex(pe->GetFacet()->GetForwardVertex());
	  if ( pe->GetEdgeNext(0)->GetForwardEdge() != NULL )
	    {
	      if ( pe->GetEdgePrev(0)->GetForwardEdge() != NULL &&
		   pe->GetEdgeNext(0)->GetForwardEdge()->GetEdgePrev(1) != pe->GetEdgePrev(0)->GetForwardEdge() )
		{
		  Splice(1, peNew, pe->GetEdgePrev(0)->GetForwardEdge());
		}
	      Splice(1, peNew, pe->GetEdgeNext(0)->GetForwardEdge()->GetEdgePrev(1));
	    }

	  pe = pe->GetEdge();
	  peNew = peNew->GetEdge();
	  if ( pe->GetFacet() != m_pfaNULL )
	    {
	      peNew->SetVertex(pe->GetFacet()->GetForwardVertex());
	      if ( pe->GetEdgeNext(0)->GetForwardEdge() != NULL )
		{
		  if ( pe->GetEdgePrev(0)->GetForwardEdge() != NULL &&
		       pe->GetEdgeNext(0)->GetForwardEdge()->GetEdgePrev(1) != pe->GetEdgePrev(0)->GetForwardEdge() )
		    {
		      Splice(1, peNew, pe->GetEdgePrev(0)->GetForwardEdge());
		    }
		  Splice(1, peNew, pe->GetEdgeNext(0)->GetForwardEdge()->GetEdgePrev(1));
		}
	    }
	  else
	    {
	      v = pe->GetEdge()->GetVertex()->GetPoint() - pe->GetVertex()->GetPoint();
	      v.Set(-v(1), v(0), 0.0);
	      v.Normalize();
	      pt = v;
	      pt[3] = 0.0;

	      pvxNew = CreateVertex(pt);
	      peNew->SetVertex(pvxNew);

	      if ( pe->GetEdgeNext(0)->GetForwardEdge() != NULL )
		{
		  peBoundary = pe->GetEdgeNext(0)->GetForwardEdge()->GetEdgePrev(1)->GetEdge();
		}
	      else
		{
		  peBoundary = CreateEdge();
		  peBoundary->GetEdge()->SetFacet(peNew->GetFacet());
		}
	      peBoundary->SetVertex(peNew->GetVertex());
	      Splice(1, peNew, peBoundary);

	      if ( pe->GetEdgePrev(0)->GetForwardEdge() != NULL )
		{
		  peBoundary = pe->GetEdgePrev(0)->GetForwardEdge()->GetEdgePrev(0);
		}
	      else
		{
		  peBoundary = CreateEdge();
		  peBoundary->SetFacet(peNew->GetEdge()->GetFacet());
		}
	      peBoundary->SetVertex(peNew->GetVertex());
	      Splice(1, peNew->GetEdgePrev(1), peBoundary);
	    }
	}
    }
}

VOID QEMesh::CreateVoronoiDivideAndConquer(QEMesh *pmeDelaunay)
{
  CObjectListIter<QEVertex> iter;
  CHeap<QEVertex*, QEVertex> heapQEVertex;
  QEVertex **ppvx, *pvx;
  QEEdge *ppe[2];
  UINT_32 i, uiSize;

  ASSERT( pmeDelaunay != NULL );

  uiSize = pmeDelaunay->m_plistQEVertex->GetLength();
  ASSERT( uiSize > 1 );

  //////////////////////////////////////////////////
  // Sort the Vertices
  heapQEVertex.Init(uiSize);
  ppvx = new QEVertex*[uiSize];
  ASSERT( ppvx != NULL );

  iter.Init(pmeDelaunay->m_plistQEVertex);
  for ( pvx = iter.PeekFirst();
	pvx != NULL;
	pvx = iter.PeekNext() )
    {
      heapQEVertex.Insert(*pvx, pvx);
    }

  i = 0;
  for ( pvx = heapQEVertex.RemoveTop();
	pvx != NULL;
	pvx = heapQEVertex.RemoveTop() )
    {
      ASSERT( i < uiSize );
      ppvx[i] = pvx;
      i++;
    }
  ASSERT( i == uiSize );

  CreateVoronoiDivideAndConquer(uiSize, ppvx, ppe); 

  delete []ppvx;
}

VOID QEMesh::CreateVoronoiDivideAndConquer(UINT_32 uiSize, QEVertex **ppvxSite, QEEdge **ppeLeftAndRight)
{
  ASSERT( uiSize > 1 );
  ASSERT( ppvxSite != NULL );

  switch ( uiSize )
    {
    case 2:
      {
	QEFacet *ppfaNew[2];
	QEVertex *ppvxNew[2];
	QEEdge *ppeNew[3];
	Point pt;
	Vector v;
	UINT_32 i;

	for ( i = 0; i < 3; i++ )
	  {
	    ppeNew[i] = CreateEdge();
	  }

	for ( i = 0; i < 2; i++ )
	  {
	    ppfaNew[i] = CreateFacet();
	    ppfaNew[i]->SetForwardVertex(ppvxSite[i]);
	    ppvxSite[i]->SetForwardFacet(ppfaNew[i]);
	  }

	v = ppvxSite[1]->GetPoint() - ppvxSite[0]->GetPoint();
	v.Set(-v(1), v(0), 0.0);
	VERIFY( v.Normalize() );

	pt = v;
	pt[3] = 0.0;
	ppvxNew[0] = CreateVertex(pt);

	pt = (-v);
	pt[3] = 0.0;
	ppvxNew[1] = CreateVertex(pt);

	Splice(1, ppeNew[0], ppeNew[2]);
	Splice(1, ppeNew[2], ppeNew[1]->GetEdge());
	Splice(1, ppeNew[0]->GetEdge(), ppeNew[1]);
	Splice(1, ppeNew[1], ppeNew[2]->GetEdge());

	ppeNew[0]->SetFacet(ppfaNew[0]);
	ppeNew[2]->GetEdge()->SetFacet(ppfaNew[0]);
	ppeNew[2]->SetFacet(ppfaNew[1]);
	ppeNew[1]->SetFacet(ppfaNew[1]);

	ppeNew[0]->SetVertex(ppvxNew[0]);
	ppeNew[1]->GetEdge()->SetVertex(ppvxNew[0]);
	ppeNew[2]->SetVertex(ppvxNew[0]);

	ppeNew[0]->GetEdge()->SetVertex(ppvxNew[1]);
	ppeNew[1]->SetVertex(ppvxNew[1]);
	ppeNew[2]->GetEdge()->SetVertex(ppvxNew[1]);

	ppeLeftAndRight[0] = ppeNew[0]->GetEdge();
	ppeLeftAndRight[1] = ppeNew[1]->GetEdge();	
      }
      break;

    case 3:
      {
	QEFacet *ppfaNew[3];
	QEVertex *ppvxNew[4];
	QEEdge *ppeNew[6];
	Vector v;
	Point pt, ptCenter;
	UINT_32 i, j;

	for ( i = 0; i < 3; i++ )
	  {
	    ppfaNew[i] = CreateFacet();
	    ppfaNew[i]->SetForwardVertex(ppvxSite[i]);
	    ppvxSite[i]->SetForwardFacet(ppfaNew[i]);
	  }

	for ( i = 0; i < 6; i++ )
	  {
	    ppeNew[i] = CreateEdge();
	  }

	if ( Counterclockwise(ppvxSite[0]->GetPoint(), 
			      ppvxSite[1]->GetPoint(), 
			      ppvxSite[2]->GetPoint()) )
	  {
	    ptCenter = CalculateCircumcenter(ppvxSite[0]->GetPoint(), 
					     ppvxSite[1]->GetPoint(), 
					     ppvxSite[2]->GetPoint());

	    ppvxNew[3] = CreateVertex(ptCenter);

	    j = 2;
	    for ( i = 0; i < 3; i++ )
	      {
		v = ppvxSite[j]->GetPoint() - ppvxSite[i]->GetPoint();
		v.Set(-v(1), v(0), 0.0);
		VERIFY( v.Normalize() );
		pt = v;
		pt[3] = 0.0;
		ppvxNew[i] = CreateVertex(pt);

		j = i;
	      }

	    j = 2;
	    for ( i = 0; i < 3; i++ )
	      {
		ppeNew[i]->SetFacet(ppfaNew[i]);
		ppeNew[i+3]->SetFacet(ppfaNew[i]);
		ppeNew[j+3]->GetEdge()->SetFacet(ppfaNew[i]);

		ppeNew[i]->SetVertex(ppvxNew[i]);
		ppeNew[j]->GetEdge()->SetVertex(ppvxNew[i]);
		ppeNew[j+3]->SetVertex(ppvxNew[i]);

		ppeNew[i+3]->GetEdge()->SetVertex(ppvxNew[3]);

		Splice(1, ppeNew[i], ppeNew[j+3]);
		Splice(1, ppeNew[j+3], ppeNew[j]->GetEdge());

		j = i;
	      }

	    Splice(1, ppeNew[3]->GetEdge(), ppeNew[4]->GetEdge());	    
	    Splice(1, ppeNew[4]->GetEdge(), ppeNew[5]->GetEdge());	    

	    ppeLeftAndRight[0] = ppeNew[0]->GetEdge();
	    ppeLeftAndRight[1] = ppeNew[2]->GetEdge();
	  }
	else if ( Counterclockwise(ppvxSite[0]->GetPoint(), 
				   ppvxSite[2]->GetPoint(), 
				   ppvxSite[1]->GetPoint()) )
	  {
	    ptCenter = CalculateCircumcenter(ppvxSite[0]->GetPoint(), 
					     ppvxSite[2]->GetPoint(), 
					     ppvxSite[1]->GetPoint());

	    ppvxNew[3] = CreateVertex(ptCenter);

	    j = 2;
	    for ( i = 0; i < 3; i++ )
	      {
		v = ppvxSite[i]->GetPoint() - ppvxSite[j]->GetPoint();
		v.Set(-v(1), v(0), 0.0);
		VERIFY( v.Normalize() );
		pt = v;
		pt[3] = 0.0;
		ppvxNew[j] = CreateVertex(pt);

		j = i;
	      }

	    j = 2;
	    for ( i = 0; i < 3; i++ )
	      {
		ppeNew[i]->SetFacet(ppfaNew[i]);
		ppeNew[i+3]->SetFacet(ppfaNew[i]);
		ppeNew[j+3]->GetEdge()->SetFacet(ppfaNew[i]);

		ppeNew[i]->GetEdge()->SetVertex(ppvxNew[j]);
		ppeNew[j]->SetVertex(ppvxNew[j]);
		ppeNew[j+3]->GetEdge()->SetVertex(ppvxNew[j]);

		ppeNew[i+3]->SetVertex(ppvxNew[3]);

		Splice(1, ppeNew[j+3]->GetEdge(), ppeNew[j]);
		Splice(1, ppeNew[i]->GetEdge(), ppeNew[j+3]->GetEdge());

		j = i;
	      }

	    Splice(1, ppeNew[4], ppeNew[5]);	    
	    Splice(1, ppeNew[3], ppeNew[4]);	    

	    ppeLeftAndRight[0] = ppeNew[0]->GetEdge();
	    ppeLeftAndRight[1] = ppeNew[2]->GetEdge();
	  }
	else
	  {
	    // Colinear
	    for ( i = 0; i < 2; i++ )
	      {
		v = ppvxSite[i+1]->GetPoint() - ppvxSite[i]->GetPoint();
		v.Set(-v(1), v(0), 0.0);
		VERIFY( v.Normalize() );

		pt = v;
		pt[3] = 0.0;
		ppvxNew[2*i] = CreateVertex(pt);

		pt = (-v);
		pt[3] = 0.0;
		ppvxNew[2*i + 1] = CreateVertex(pt);
	      }

	    for ( i = 0; i < 2; i++ )
	      {
		ppeNew[2*i]->SetFacet(ppfaNew[2*i]);
		ppeNew[2*i+1]->SetFacet(ppfaNew[2*i]);

		ppeNew[2*i]->SetVertex(ppvxNew[2*i]);
		ppeNew[2*i+1]->GetEdge()->SetVertex(ppvxNew[2*i]);
		ppeNew[2*i]->GetEdge()->SetVertex(ppvxNew[2*i+1]);
		ppeNew[2*i+1]->SetVertex(ppvxNew[2*i+1]);
	      }

	    ppeNew[1]->GetEdge()->SetFacet(ppfaNew[1]);
	    ppeNew[2]->GetEdge()->SetFacet(ppfaNew[1]);
	    ppeNew[4]->SetFacet(ppfaNew[1]);
	    ppeNew[5]->SetFacet(ppfaNew[1]);

	    ppeNew[4]->GetEdge()->SetVertex(ppvxNew[0]);
	    ppeNew[4]->SetVertex(ppvxNew[2]);
	    ppeNew[5]->SetVertex(ppvxNew[1]);
	    ppeNew[5]->GetEdge()->SetVertex(ppvxNew[3]);

	    Splice(1, ppeNew[4]->GetEdge(), ppeNew[0]);
	    Splice(1, ppeNew[0], ppeNew[1]->GetEdge());

	    Splice(1, ppeNew[5], ppeNew[1]);
	    Splice(1, ppeNew[1], ppeNew[0]->GetEdge());

	    Splice(1, ppeNew[4], ppeNew[2]);
	    Splice(1, ppeNew[2], ppeNew[3]->GetEdge());

	    Splice(1, ppeNew[5]->GetEdge(), ppeNew[3]);
	    Splice(1, ppeNew[3], ppeNew[2]->GetEdge());

	    ppeLeftAndRight[0] = ppeNew[0]->GetEdge();
	    ppeLeftAndRight[1] = ppeNew[3]->GetEdge();
	  }
      }
      break;
    default:
      {
	QEEdge *ppeLeft[2], *ppeRight[2], *peBisector;
	QEEdge *peLeft, *peLeftNext, *peRight, *peRightNext, *peNew;
	QEVertex *pvxNew;
	Point pt;
	Vector v;
	UINT_32 uiHalf;
	
	uiHalf = uiSize / 2;
	CreateVoronoiDivideAndConquer(uiHalf, ppvxSite, ppeLeft);
	CreateVoronoiDivideAndConquer(uiSize - uiHalf, ppvxSite + uiHalf, ppeRight);

	ASSERT( ppeLeft[0]->GetEdge()->GetFacet()->GetForwardVertex() == ppvxSite[0] );
	ASSERT( ppeLeft[0]->GetFacet() == m_pfaNULL );
	ASSERT( ppeRight[1]->GetEdge()->GetFacet()->GetForwardVertex() == ppvxSite[uiSize - 1] );
	ASSERT( ppeRight[1]->GetFacet() == m_pfaNULL );

	// Find the bottom two vertices
	while ( TRUE )
	  {
	    if ( Counterclockwise(ppeRight[0]->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(),
				  ppeLeft[1]->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(),
				  ppeLeft[1]->GetEdgeNext(0)->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint()) )
	      
	      {
		ppeLeft[1] = ppeLeft[1]->GetEdgeNext(0);
	      }
	    else if ( Counterclockwise(ppeLeft[1]->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(),
				       ppeRight[0]->GetEdgePrev(0)->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(),
				       ppeRight[0]->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint()) )
	      {
		ppeRight[0] = ppeRight[0]->GetEdgePrev(0);
	      }
	    else
	      {
		break;
	      }
	  }

	ASSERT( ppeLeft[1]->GetFacet() == m_pfaNULL );
	ASSERT( ppeRight[0]->GetFacet() == m_pfaNULL );

	//////////////////////////////////////////////////
	// Merge
	
	// Break the outside edges and merge them together
	peLeft = ppeLeft[1]->GetEdge();
	peRight = ppeRight[0]->GetEdge();

	peLeftNext = peLeft->GetEdgeNext(0);
	peRightNext = peRight->GetEdgePrev(0);

	peLeft->GetEdge()->GetEdgePrev(1)->SetVertex(peLeft->GetEdge()->GetEdgePrev(1)->GetVertex());
	peRight->GetEdgePrev(1)->SetVertex(peRight->GetEdgePrev(1)->GetVertex());

	Splice(1, peLeft->GetEdge(), peLeft->GetEdge()->GetEdgePrev(1));
	Splice(1, peRight, peRight->GetEdgePrev(1));

	peBisector = CreateEdge();

	peBisector->GetEdge()->SetFacet(peLeft->GetFacet());
	peBisector->SetFacet(peRight->GetFacet());

	v = peBisector->GetFacet()->GetForwardVertex()->GetPoint() - peBisector->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint();
	v.Set(v(1), -v(0), 0.0);
	VERIFY( v.Normalize() );
	pt = v;
	pt[3] = 0.0;

	pvxNew = CreateVertex(pt);

	peLeft->GetEdge()->SetVertex(pvxNew);
	peRight->SetVertex(pvxNew);
	peBisector->GetEdge()->SetVertex(pvxNew);

	Splice(1, peRight, peLeft->GetEdge());
	Splice(1, peRight, peBisector->GetEdge());

	peLeft = peLeftNext;
	peRight = peRightNext;
	
	//////////////////////////////////////////////////
	// Store the left and right extreme sites

	ppeLeftAndRight[0] = ppeLeft[0]->GetEdge()->GetEdgeNext(0);
	ppeLeftAndRight[1] = ppeRight[1]->GetEdge()->GetEdgePrev(0);
	
	while ( TRUE )
	  {
	    ASSERT( peLeft->GetEdge()->GetFacet() != m_pfaNULL );
	    if ( Counterclockwise(peLeft->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(), 
				  peBisector->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(),
				  peBisector->GetFacet()->GetForwardVertex()->GetPoint()) )
	      {
		peLeftNext = peLeft->GetEdgeNext(0);
		if ( peLeftNext->GetEdge()->GetFacet() == m_pfaNULL )
		  {
		    peLeftNext = peLeftNext->GetEdgeNext(0);
		  }
		while ( peLeftNext != peBisector->GetEdge() &&
			InCircumcircle(peBisector->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(),
				       peBisector->GetFacet()->GetForwardVertex()->GetPoint(),
				       peLeft->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(),
				       peLeftNext->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint()) )
		  {
		    if ( peLeft->GetEdgeNext(0)->GetFacet() == m_pfaNULL )
		      {
			CollapseEdge(peLeft->GetEdgeNext(0));			
		      }
		    CollapseEdge(peLeft);

		    // Advance the pointer
		    peLeft = peLeftNext;

		    peLeftNext = peLeft->GetEdgeNext(0);
		    if ( peLeftNext->GetEdge()->GetFacet() == m_pfaNULL )
		      {
			peLeftNext = peLeftNext->GetEdgeNext(0);
		      }
		  }
	      }

	    ASSERT( peRight->GetEdge()->GetFacet() != m_pfaNULL );
	    if ( Counterclockwise(peRight->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(), 
				  peBisector->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(), 
				  peBisector->GetFacet()->GetForwardVertex()->GetPoint()) )
	      {
		peRightNext = peRight->GetEdgePrev(0);
		if ( peRightNext->GetEdge()->GetFacet() == m_pfaNULL )
		  {
		    peRightNext = peRightNext->GetEdgePrev(0);
		  }
		while ( peRightNext != peBisector &&
			InCircumcircle(peRight->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(),
				       peBisector->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(),
				       peBisector->GetFacet()->GetForwardVertex()->GetPoint(),

				       peRightNext->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint()) )
		  {
		    if ( peRight->GetEdgePrev(0)->GetFacet() == m_pfaNULL )
		      {
			CollapseEdge(peRight->GetEdgePrev(0)->GetEdge());
		      }
		    CollapseEdge(peRight->GetEdge());

		    // Advance the pointer
		    peRight = peRightNext;

		    peRightNext = peRight->GetEdgePrev(0);
		    if ( peRightNext->GetEdge()->GetFacet() == m_pfaNULL )
		      {
			peRightNext = peRightNext->GetEdgePrev(0);
		      }
		  }
	      }

	    ASSERT( peLeft->GetEdge()->GetFacet() != m_pfaNULL );
	    ASSERT( peRight->GetEdge()->GetFacet() != m_pfaNULL );
	    if ( !Counterclockwise(peLeft->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(), 
				   peBisector->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(),
				   peBisector->GetFacet()->GetForwardVertex()->GetPoint()) &&
		 !Counterclockwise(peRight->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(), 
				   peBisector->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(), 
				   peBisector->GetFacet()->GetForwardVertex()->GetPoint()) )
	      {
		break;
	      }

	    // Extend the bisector
	    if ( !Counterclockwise(peLeft->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(), 
				   peBisector->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(),
				   peBisector->GetFacet()->GetForwardVertex()->GetPoint()) ||
		 (Counterclockwise(peRight->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(), 
				   peBisector->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(), 
				   peBisector->GetFacet()->GetForwardVertex()->GetPoint()) &&
		  InCircumcircle(peLeft->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(),
				 peBisector->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(),
				 peBisector->GetFacet()->GetForwardVertex()->GetPoint(),
				 peRight->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint())) )
	      {
		ASSERT( peRight->GetEdge()->GetEdgePrev(1) != peRight->GetEdge() );
		
		peRightNext = peRight->GetEdge()->GetEdgePrev(0);
		
		pt = CalculateCircumcenter(peBisector->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(),
					   peBisector->GetFacet()->GetForwardVertex()->GetPoint(),
					   peRight->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint());
		pvxNew = CreateVertex(pt);
		peNew = CreateEdge();

		peRight->GetEdge()->GetEdgePrev(1)->SetVertex(peRight->GetEdge()->GetEdgePrev(1)->GetVertex());

		peRight->GetEdge()->SetVertex(pvxNew);
		peNew->GetEdge()->SetVertex(pvxNew);
		peBisector->SetVertex(pvxNew);

		peNew->GetEdge()->SetFacet(peBisector->GetEdge()->GetFacet());
		peNew->SetFacet(peRight->GetEdge()->GetFacet());

		Splice(1, peRight->GetEdge()->GetEdgePrev(1), peRight->GetEdge());
		Splice(1, peRight->GetEdge(), peBisector);
		Splice(1, peRight->GetEdge(), peNew->GetEdge());

		peBisector = peNew;
		peRight = peRightNext;
		if ( peRight->GetEdge()->GetFacet() == m_pfaNULL )
		  {
		    peRightNext = peRight->GetEdgePrev(0);
		    CollapseEdge(peRight->GetEdge());
		    peRight = peRightNext;		    
		  }
	      }
	    else
	      {
		ASSERT( peLeft->GetEdgePrev(1) != peLeft );

		peLeftNext = peLeft->GetEdge()->GetEdgeNext(0);
		
		pt = CalculateCircumcenter(peLeft->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(),
					   peBisector->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint(),
					   peBisector->GetFacet()->GetForwardVertex()->GetPoint());
		pvxNew = CreateVertex(pt);
		peNew = CreateEdge();

		peLeft->GetEdgePrev(1)->SetVertex(peLeft->GetEdgePrev(1)->GetVertex());

		peLeft->SetVertex(pvxNew);
		peNew->GetEdge()->SetVertex(pvxNew);
		peBisector->SetVertex(pvxNew);

		peNew->SetFacet(peBisector->GetFacet());
		peNew->GetEdge()->SetFacet(peLeft->GetEdge()->GetFacet());

		Splice(1, peLeft->GetEdgePrev(1), peLeft);
		Splice(1, peLeft, peNew->GetEdge());
		Splice(1, peLeft, peBisector);

		peBisector = peNew;
		peLeft = peLeftNext;
		if ( peLeft->GetEdge()->GetFacet() == m_pfaNULL )
		  {
		    peLeftNext = peLeft->GetEdgeNext(0);
		    CollapseEdge(peLeft);
		    peLeft = peLeftNext;		    
		  }
	      }
	  }

	v = peBisector->GetFacet()->GetForwardVertex()->GetPoint() - peBisector->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint();
	v.Set(-v(1), v(0), 0.0);
	VERIFY( v.Normalize() );
	pt = v;
	pt[3] = 0.0;
	pvxNew = CreateVertex(pt);

	peBisector->SetVertex(pvxNew);

	peNew = CreateEdge();

	peNew->GetEdge()->SetFacet(m_pfaNULL);
	peNew->SetFacet(peBisector->GetEdge()->GetFacet());
	peNew->SetVertex(pvxNew);
	peNew->GetEdge()->SetVertex(peLeft->GetVertex());

	Splice(1, peLeft, peNew->GetEdge());
	Splice(1, peBisector, peNew);

	peNew = CreateEdge();

	peNew->GetEdge()->SetFacet(m_pfaNULL);
	peNew->SetFacet(peBisector->GetFacet());
	peNew->SetVertex(peRight->GetEdge()->GetVertex());
	peNew->GetEdge()->SetVertex(pvxNew);

	Splice(1, peBisector, peNew->GetEdge());
	Splice(1, peRight->GetEdge()->GetEdgePrev(1), peNew);

	//////////////////////////////
	// Recover Left and Right extremes

	ppeLeftAndRight[0] = ppeLeftAndRight[0]->GetEdgePrev(0)->GetEdge();
	ppeLeftAndRight[1] = ppeLeftAndRight[1]->GetEdgeNext(0)->GetEdge();

	ASSERT( ppeLeftAndRight[0]->GetEdge()->GetFacet()->GetForwardVertex() == ppvxSite[0] );
	ASSERT( ppeLeftAndRight[0]->GetFacet() == m_pfaNULL );
	ASSERT( ppeLeftAndRight[1]->GetEdge()->GetFacet()->GetForwardVertex() == ppvxSite[uiSize-1] );
	ASSERT( ppeLeftAndRight[1]->GetFacet() == m_pfaNULL );
      }
      break;
    }
}

VOID QEMesh::CreateVoronoiDivideAndConquerAlternating(QEMesh *pmeDelaunay)
{
  CreateVoronoiDivideAndConquer(pmeDelaunay);
}

VOID QEMesh::SortAlternating(QEVertex::ComparisonType eDirection, 
			     UINT_32 uiSize, QEVertex **ppvx)
{
  ASSERT( eDirection == QEVertex::ctXYZ || eDirection == QEVertex::ctYXZ );
  ASSERT( uiSize > 1 );
  ASSERT( ppvx != NULL );

  QEVertex::SetComparisonType(eDirection);

  switch ( uiSize )
    {
    case 2:
      {
	if ( (*ppvx[0]) > (*ppvx[1]) )
	  {
	    Swap(&(ppvx[0]), &(ppvx[1]));
	  }
      }
      break;
    case 3:
      {
	UINT_32 i, j;

	// Bubble Sort
	for ( i = 1; i < uiSize; i++ )
	  {
	    for ( j = 1; j < uiSize + 1 - i; j++ )
	      {
		if ( (*ppvx[j-1]) > (*ppvx[j]) )
		  {
		    Swap(&(ppvx[j-1]), &(ppvx[j]));
		  }
	      }
	  }
      }
      break;
    default:
      {
	UINT_32 uiHalf;

	uiHalf = uiSize / 2;
	QuickSelect(uiSize, ppvx, uiHalf);

	if ( eDirection == QEVertex::ctXYZ )
	  {
	    eDirection = QEVertex::ctYXZ;
	  }
	else
	  {
	    eDirection = QEVertex::ctXYZ;
	  }
	SortAlternating(eDirection, uiHalf, ppvx);
	SortAlternating(eDirection, uiSize - uiHalf, ppvx + uiHalf);
      }
      break;
    }
  
  QEVertex::SetComparisonTypeDefault();
}

VOID QEMesh::QuickSelect(UINT_32 uiSize, QEVertex **ppvx, UINT_32 uiPos)
{
  ASSERT( uiSize > 0 );
  ASSERT( ppvx != NULL );
  ASSERT( uiPos < uiSize );

  UINT_32 uiSplit, uiLeftHalfSize;

  if ( uiSize < 2 )
    {
      return;
    }

  uiSplit = QuickPartition(uiSize, ppvx);
  uiLeftHalfSize = uiSplit + 1;

  if ( uiPos < uiLeftHalfSize )
    {
      QuickSelect(uiLeftHalfSize, ppvx, uiPos);
    }
  else
    {
      QuickSelect(uiSize - uiLeftHalfSize, ppvx + uiLeftHalfSize, uiPos - uiLeftHalfSize);
    }
}

UINT_32 QEMesh::QuickPartition(UINT_32 uiSize, QEVertex **ppvx)
{
  ASSERT( uiSize > 0 );
  ASSERT( ppvx != NULL );

  QEVertex *pvx;
  INT_32 i, j;

  // We should pick a random value instead of the midpoint!
  Swap(&(ppvx[0]), &(ppvx[uiSize/2]));

  pvx = ppvx[0];

  i = -1;
  j = (INT_32)uiSize;

  while ( TRUE )
    {
      for ( j--; (*(ppvx[j])) > (*pvx); j-- )
	{
	}
      for ( i++; (*(ppvx[i])) < (*pvx); i++ )
	{
	}

      if ( i < j )
	{
	  Swap(&(ppvx[i]), &(ppvx[j]));
	}
      else
	{
	  break;
	}
    }

  return (UINT_32)j;
}

VOID QEMesh::Swap(QEVertex **ppvx0, QEVertex **ppvx1)
{
  ASSERT( ppvx0 != NULL );
  ASSERT( ppvx1 != NULL );

  QEVertex *pvx;

  pvx = *ppvx0;
  *ppvx0 = *ppvx1;
  *ppvx1 = pvx;
}

VOID QEMesh::SortAlternatingBubble(QEVertex::ComparisonType eDirection, 
				   UINT_32 uiSize, QEVertex **ppvx)
{
  ASSERT( eDirection == QEVertex::ctXYZ || eDirection == QEVertex::ctYXZ );
  ASSERT( uiSize > 1 );
  ASSERT( ppvx != NULL );

  QEVertex *pvxSwap;
  UINT_32 i, j;

  QEVertex::SetComparisonType(eDirection);

  switch ( uiSize )
    {
    case 2:
      {
	if ( (*ppvx[0]) > (*ppvx[1]) )
	  {
	    pvxSwap = ppvx[1];
	    ppvx[1] = ppvx[0];
	    ppvx[0] = pvxSwap;
	  }
      }
      break;
    case 3:
      {
	// Bubble Sort
	for ( i = 1; i < uiSize; i++ )
	  {
	    for ( j = 1; j < uiSize + 1 - i; j++ )
	      {
		if ( (*ppvx[j-1]) > (*ppvx[j]) )
		  {
		    pvxSwap = ppvx[j];
		    ppvx[j] = ppvx[j-1];
		    ppvx[j-1] = pvxSwap;
		  }
	      }
	  }
      }
      break;
    default:
      {
	UINT_32 uiHalf, uiSwaps;
	// Bubble Sort
	uiSwaps = 1;
	for ( i = 1; i < uiSize && uiSwaps > 0; i++ )
	  {
	    uiSwaps = 0;
	    for ( j = 1; j < uiSize + 1 - i; j++ )
	      {
		if ( (*ppvx[j-1]) > (*ppvx[j]) )
		  {
		    pvxSwap = ppvx[j];
		    ppvx[j] = ppvx[j-1];
		    ppvx[j-1] = pvxSwap;
		    uiSwaps++;
		  }
	      }
	  }

	if ( eDirection == QEVertex::ctXYZ )
	  {
	    eDirection = QEVertex::ctYXZ;
	  }
	else
	  {
	    eDirection = QEVertex::ctXYZ;
	  }
	uiHalf = uiSize / 2;
	SortAlternatingBubble(eDirection, uiHalf, ppvx);
	SortAlternatingBubble(eDirection, uiSize - uiHalf, ppvx + uiHalf);
      }
      break;
    }
  
  QEVertex::SetComparisonTypeDefault();
}

//////////////////////////////////////////////////////////////////////
// CreateVoronoiPolygonChordalAxis
//
// Creates the Voronoi Diagram of a Polygon
//
// Currently: approximating the Voronoi Diagram by retriangulating 
//            the CDT (Constrained Delaunay Triangulation) of the polygon
//            using the chordal axis mentioned in SIGGRAPH 1999 p. 413
//            the Teddy paper
//            NOTE: The CDT of a polygon can be created using CreateTriangulatedPolygon
//
VOID QEMesh::CreateVoronoiPolygonChordalAxis(QEMesh *pmeCDT)
{
  ASSERT( pmeCDT != NULL );

  QEMeshIter iterQEMesh;
  QEFacetIter iterQEFacet;
  QEFacet *pfa, *pfaNew; 
  QEVertex *pvx, *pvxNew;
  QEEdge *pe, *peCurr; 
  QEEdge *ppe[2], *ppeNew[2];
  UINT_32 i, uiEdges, uiBoundaryEdges, uiResults;
  Point ppt[4], pt;
  Vector v;

  BOOL bLift;

  bLift = TRUE;

  pmeCDT->ClearForwardPointers();

  iterQEMesh.Init(pmeCDT);

  // Copy the Vertices
  for ( pvx = iterQEMesh.PeekFirstVertex();
        pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pvxNew = CreateVertex(pvx->GetPoint());
      pvx->SetForwardVertex(pvxNew);
    }

  // Copy the edges
  for ( pe = iterQEMesh.PeekFirstEdge();
        pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      ASSERT( (pe->GetForwardEdge() == NULL && pe->GetEdge()->GetForwardEdge() == NULL) ||
	      (pe->GetForwardEdge() != NULL && pe->GetEdge()->GetForwardEdge() != NULL) );
      if ( pe->GetForwardEdge() == NULL )
	{
	  if ( pe->GetEdge()->GetFacet() == m_pfaNULL )
	    {
	      ASSERT( pe->GetFacet() != m_pfaNULL );
	      pe = pe->GetEdge();
	    }

	  ppe[0] = pe;
	  ppe[1] = pe->GetEdge();
	  if ( ppe[0]->GetFacet() == m_pfaNULL )
	    {
	      ppeNew[0] = CreateEdge();
	      ppeNew[1] = ppeNew[0]->GetEdge();

	      // External Edge
	      m_plistQEEdge->Remove(ppeNew[0]);
	      m_plistQEEdgeBoundary->InsertLast(ppeNew[0]);
	    }
	  else
	    {
	      // Internal Edge -- Create the midpoint
	      pt = (ppe[0]->GetVertex()->GetPoint() + ppe[1]->GetVertex()->GetPoint()) / 2.0; 
	      if ( bLift )
		{
		  v = ppe[1]->GetVertex()->GetPoint() - ppe[0]->GetVertex()->GetPoint();
		  pt[Z] = v.Magnitude()/2.0;
		}

	      pvxNew = CreateVertex(pt);

	      ppeNew[0] = CreateEdge();
	      ppeNew[1] = CreateEdge();

	      Splice(1, ppeNew[0]->GetEdge(), ppeNew[1]->GetEdge());
	      ppeNew[0]->GetEdge()->SetVertex(pvxNew);
	      ppeNew[1]->GetEdge()->SetVertex(pvxNew);
	    }

	  ppe[0]->SetForwardEdge(ppeNew[0]);
	  ppe[1]->SetForwardEdge(ppeNew[1]);

	  ppeNew[0]->SetVertex(ppe[0]->GetVertex()->GetForwardVertex());
	  ppeNew[1]->SetVertex(ppe[1]->GetVertex()->GetForwardVertex());

	  for ( i = 0; i < 2; i++ )
	    {
	      for ( peCurr = ppe[i]->GetEdgePrev(1);
		    peCurr != ppe[i];
		    peCurr = peCurr->GetEdgePrev(1) )
		{
		  if ( peCurr->GetForwardEdge() != NULL )
		    {
		      Splice(1, peCurr->GetForwardEdge(), ppeNew[i]);
		      break;
		    }
		}
	    }
	}
    }

  // Triangulate the new facets
  for ( pfa = iterQEMesh.PeekFirstFacet();
        pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      ASSERT( pfa != m_pfaNULL );

      iterQEFacet.Init(pfa);
      uiEdges = 0;
      uiBoundaryEdges = 0;
      for ( pe = iterQEFacet.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEFacet.PeekNextEdge() )
	{
	  uiEdges++;
	  if ( pe->GetEdge()->GetFacet() == m_pfaNULL )
	    {
	      uiBoundaryEdges++;
	    }
	}
      ASSERT( uiEdges == 3 );

      switch ( uiBoundaryEdges )
	{
	case 0:
	  {
	    pfaNew = CreateFacet();
	    for ( i = 0, pe = iterQEFacet.PeekFirstEdge();
		  pe != NULL;
		  i++, pe = iterQEFacet.PeekNextEdge() )
	      {
		ppt[i] = pe->GetVertex()->GetPoint();

		pe->GetForwardEdge()->SetFacet(pfaNew);
		pe->GetForwardEdge()->GetEdgeNext(0)->SetFacet(pfaNew);
	      }

	    pt = CalculateCircumcenter(ppt[0], ppt[1], ppt[2], bLift);
	    pvxNew = CreateVertex(pt);

	    InsertVertex(pfaNew, pvxNew);
	  }
	  break;
	case 1:
	  {
	    for ( pe = iterQEFacet.PeekFirstEdge();
		  pe != NULL;
		  pe = iterQEFacet.PeekNextEdge() )
	      {
		if ( pe->GetEdge()->GetFacet() == m_pfaNULL )
		  {
		    // Connect the two chord bisectors creating a triangle
		    ppeNew[0] = CreateEdge();
		    ppeNew[1] = ppeNew[0]->GetEdge();

		    ppeNew[0]->SetVertex(pe->GetForwardEdge()->GetEdgePrev(0)->GetVertex());
		    ppeNew[1]->SetVertex(pe->GetForwardEdge()->GetEdgeNext(0)->GetEdgeNext(0)->GetVertex());

		    pfaNew = CreateFacet();
		    ppeNew[0]->SetFacet(pfaNew);
		    pe->GetForwardEdge()->GetEdgePrev(0)->GetEdgePrev(0)->SetFacet(pfaNew);
		    pe->GetForwardEdge()->GetEdgeNext(0)->GetEdgeNext(0)->SetFacet(pfaNew);
		    
		    Splice(1, ppeNew[1], pe->GetForwardEdge()->GetEdgeNext(0)->GetEdgeNext(0));
		    Splice(1, ppeNew[0], pe->GetForwardEdge()->GetEdgePrev(0));

		    // Triangulate the quadrilateral in Delaunay fashion

		    ppt[0] = pe->GetVertex()->GetPoint();
		    ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
		    ppt[2] = pe->GetForwardEdge()->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
		    ppt[3] = pe->GetForwardEdge()->GetEdgePrev(0)->GetVertex()->GetPoint();

		    if ( InCircumcircle(ppt[0], ppt[1], ppt[2], ppt[3]) )
		      {
			ppeNew[0] = CreateEdge();
			ppeNew[1] = ppeNew[0]->GetEdge();

			ppeNew[0]->SetVertex(pe->GetForwardEdge()->GetEdge()->GetVertex());
			ppeNew[1]->SetVertex(pe->GetForwardEdge()->GetEdgePrev(0)->GetVertex());

			pfaNew = CreateFacet();
			ppeNew[1]->SetFacet(pfaNew);
			pe->GetForwardEdge()->GetEdgeNext(0)->SetFacet(pfaNew);
			pe->GetForwardEdge()->GetEdgeNext(0)->GetEdgeNext(0)->SetFacet(pfaNew);
			pfaNew = CreateFacet();
			ppeNew[0]->SetFacet(pfaNew);
			pe->GetForwardEdge()->SetFacet(pfaNew);
			pe->GetForwardEdge()->GetEdgePrev(0)->SetFacet(pfaNew);

			Splice(1, ppeNew[0], pe->GetForwardEdge()->GetEdgeNext(0));
			Splice(1, ppeNew[1], pe->GetForwardEdge()->GetEdgePrev(0));
		      }
		    else
		      {
			ppeNew[0] = CreateEdge();
			ppeNew[1] = ppeNew[0]->GetEdge();

			ppeNew[0]->SetVertex(pe->GetForwardEdge()->GetEdgeNext(0)->GetEdge()->GetVertex());
			ppeNew[1]->SetVertex(pe->GetForwardEdge()->GetVertex());

			pfaNew = CreateFacet();
			ppeNew[1]->SetFacet(pfaNew);
			pe->GetForwardEdge()->GetEdgePrev(0)->SetFacet(pfaNew);
			pe->GetForwardEdge()->GetEdgePrev(0)->GetEdgePrev(0)->SetFacet(pfaNew);
			pfaNew = CreateFacet();
			ppeNew[0]->SetFacet(pfaNew);
			pe->GetForwardEdge()->SetFacet(pfaNew);
			pe->GetForwardEdge()->GetEdgeNext(0)->SetFacet(pfaNew);

			Splice(1, ppeNew[0], pe->GetForwardEdge()->GetEdgeNext(0)->GetEdgeNext(0));
			Splice(1, ppeNew[1], pe->GetForwardEdge());
		      }
		    break;
		  }
	      }
	  }
	  break;
	case 2:
	  {
	    for ( pe = iterQEFacet.PeekFirstEdge();
		  pe != NULL;
		  pe = iterQEFacet.PeekNextEdge() )
	      {
		if ( pe->GetEdge()->GetFacet() != m_pfaNULL )
		  {
		    ppeNew[0] = CreateEdge();
		    ppeNew[1] = ppeNew[0]->GetEdge();

		    ppeNew[0]->SetVertex(pe->GetForwardEdge()->GetEdge()->GetVertex());
		    ppeNew[1]->SetVertex(pe->GetForwardEdge()->GetEdgePrev(0)->GetVertex());

		    pfaNew = CreateFacet();
		    ppeNew[0]->SetFacet(pfaNew);
		    pe->GetForwardEdge()->SetFacet(pfaNew);
		    pe->GetForwardEdge()->GetEdgePrev(0)->SetFacet(pfaNew);
		    
		    pfaNew = CreateFacet();
		    ppeNew[1]->SetFacet(pfaNew);
		    pe->GetForwardEdge()->GetEdgeNext(0)->SetFacet(pfaNew);
		    pe->GetForwardEdge()->GetEdgeNext(0)->GetEdgeNext(0)->SetFacet(pfaNew);
		    
		    Splice(1, ppeNew[0], pe->GetForwardEdge()->GetEdgeNext(0));
		    Splice(1, ppeNew[1], pe->GetForwardEdge()->GetEdgePrev(0));

		    break;
		  }
	      }
	  }
	  break;
	case 3:
	  {
	    pfaNew = CreateFacet();
	    for ( i = 0, pe = iterQEFacet.PeekFirstEdge();
		  pe != NULL;
		  i++, pe = iterQEFacet.PeekNextEdge() )
	      {
		ppt[i] = pe->GetVertex()->GetPoint();

		pe->GetForwardEdge()->SetFacet(pfaNew);
	      }

	    CalculateVoronoiPointEEE(&pt, uiResults, 
				     ppt[0], ppt[1], 
				     ppt[1], ppt[2], 
				     ppt[2], ppt[0], bLift);
	    ASSERT( uiResults == 1 );
	    pvxNew = CreateVertex(pt);

	    InsertVertex(pfaNew, pvxNew);
	  }
	  break;
	default:
	  {
	    ASSERT( FALSE );
	  }
	  break;
	}
    }
}


VOID QEMesh::SubdivideCDTBoundary()
{
  static BOOL bFirst = TRUE;
  //static CHeap<QEFacet*, FLOAT> heapQEFacet;
  static CListInt listQEEdge;
  static CListInt listQEEdgeFlip;
  static CListInt listQEEdgeCandidates;
  static CListIterInt iterQEEdge;

  if ( GetBoundaryEdgeCount() == 3 )
    {
      return;
    }

  if ( bFirst )
    {
      bFirst = FALSE;
      //heapQEFacet.Init();
      listQEEdge.Init();
      iterQEEdge.Init(&listQEEdge);
      listQEEdgeFlip.Init();
      listQEEdgeCandidates.Init();
    }

  const FLOAT cfThreshold = 0.17;

  QEMeshIter iterQEMesh;
  QEEdge *pe, *peCurr, *peNew;
  QEFacet *pfaNew;
  QEVertex *pvxNew;
  Vector u, v, w;
  Point pt;
  FLOAT pfLength[2];

  iterQEMesh.Init(this);

  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdgeBoundary() )
    {
      listQEEdgeCandidates.InsertLast(pe, 0);
    }

  while ( TRUE )
    {
      for ( pe = (QEEdge*)(listQEEdgeCandidates.RemoveLast());
	    pe != NULL;
	    pe = (QEEdge*)(listQEEdgeCandidates.RemoveLast()) )
	{ 
	  v = pe->GetVertex()->GetPoint() - pe->GetEdge()->GetVertex()->GetPoint();
	  pfLength[0] = v.Magnitude();

	  u = pe->GetEdge()->GetVertex()->GetPoint() - pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
	  v = pe->GetVertex()->GetPoint() - pe->GetEdge()->GetVertex()->GetPoint();

	  u.Normalize();
	  v.Normalize();
	  w = u * v;

	  //if ( Counterclockwise(pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint(),
	  //			pe->GetEdge()->GetVertex()->GetPoint(),
	  //		pe->GetVertex()->GetPoint()) &&
	  //   pe->GetEdge()->GetEdgeNext(1) != pe->GetEdge()->GetEdgePrev(1) )
	  if ( w(Z) > cfThreshold &&
	       pe->GetEdge()->GetEdgeNext(1) != pe->GetEdge()->GetEdgePrev(1) )
	    {
	      v = pe->GetEdgeNext(0)->GetVertex()->GetPoint() - pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
	      pfLength[1] = v.Magnitude();

	      if ( pfLength[0] > pfLength[1] )
		{
		  for ( peCurr = (QEEdge*)(iterQEEdge.PeekFirst());
			peCurr != NULL;
			peCurr = (QEEdge*)(iterQEEdge.PeekNext()) )
		    {
		      if ( peCurr == pe )
			{
			  break;
			}
		    }
		  if ( peCurr == NULL )
		    {
		      listQEEdge.InsertFirst(pe, 0);
		    }
		}
	      else
		{
		  for ( peCurr = (QEEdge*)(iterQEEdge.PeekFirst());
			peCurr != NULL;
			peCurr = (QEEdge*)(iterQEEdge.PeekNext()) )
		    {
		      if ( peCurr == pe->GetEdgeNext(0) )
			{
			  break;
			}
		    }
		  if ( peCurr == NULL )
		    {
		      listQEEdge.InsertFirst(pe->GetEdgeNext(0), 0);
		    }
		}
	    }
	}

      if ( listQEEdge.GetLength() == 0 )
	{
	  pe = iterQEMesh.PeekFirstEdgeBoundary();
	  v = pe->GetVertex()->GetPoint() - pe->GetEdge()->GetVertex()->GetPoint();
	  pfLength[0] = v.Magnitude();
	  for ( pe = iterQEMesh.PeekNextEdgeBoundary();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdgeBoundary() )
	    {
	      v = pe->GetVertex()->GetPoint() - pe->GetEdge()->GetVertex()->GetPoint();
	      pfLength[1] = v.Magnitude();
	      
	      if ( pfLength[1] < pfLength[0] )
		{
		  pfLength[0] = pfLength[1];
		}
	    }

	  //pfLength[0] *= 2.0;
	  pfLength[0] *= sqrt(2.0);

	  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdgeBoundary() )
	    {
	      v = pe->GetVertex()->GetPoint() - pe->GetEdge()->GetVertex()->GetPoint();
	      pfLength[1] = v.Magnitude();
	      
	      if ( pfLength[1] > pfLength[0] )
		{
		  listQEEdge.InsertLast(pe, 0);
		}
	    }

	  if ( listQEEdge.GetLength() == 0 )
	    {
	      break;
	    }
	}

      for ( pe = (QEEdge*)(listQEEdge.RemoveLast());
	    pe != NULL;
	    pe = (QEEdge*)(listQEEdge.RemoveLast()) )
	{
	  // Split the boundary edge and do Delaunay flipping
	  pt = (pe->GetVertex()->GetPoint() + pe->GetEdge()->GetVertex()->GetPoint()) / 2.0;
	  pt[3] = 1.0;

	  pvxNew = CreateVertex(pt);
	  pfaNew = CreateFacet();

	  peNew = CreateEdge();
	  m_plistQEEdge->Remove(peNew);
	  m_plistQEEdgeBoundary->InsertLast(peNew);

	  peNew->SetVertex(pe->GetVertex());
	  peNew->GetEdge()->SetVertex(pvxNew);
	  pe->SetVertex(pvxNew);
      
	  peNew->SetFacet(m_pfaNULL);
	  peNew->GetEdge()->SetFacet(pfaNew);

	  pe->GetEdge()->GetEdgeNext(0)->SetFacet(pfaNew);
	  pe->GetEdge()->SetFacet(pe->GetEdge()->GetFacet());

	  Splice(1, pe->GetEdgePrev(1), peNew);
	  Splice(1, pe->GetEdgePrev(1), pe);
	  Splice(1, pe, peNew->GetEdge());

	  peNew = CreateEdge();

	  peNew->SetFacet(pfaNew);
	  peNew->GetEdge()->SetFacet(pe->GetEdge()->GetFacet());
	  peNew->SetVertex(pe->GetEdge()->GetEdgePrev(0)->GetVertex());
	  peNew->GetEdge()->SetVertex(pvxNew);

	  Splice(1, pe->GetEdge()->GetEdgePrev(0), peNew);
	  Splice(1, pe->GetEdgePrev(1), peNew->GetEdge());

	  // Put the other two edges of the triangle as edges to possibly flip
	  listQEEdgeCandidates.InsertLast(pe->GetEdgeNext(0), 0);
	  listQEEdgeCandidates.InsertLast(pe, 0);
	  listQEEdgeCandidates.InsertLast(pe->GetEdgePrev(0), 0);
	  listQEEdgeCandidates.InsertLast(pe->GetEdgePrev(0)->GetEdgePrev(0), 0);

	  listQEEdgeFlip.InsertLast(peNew->GetEdgePrev(1), 0);
	  listQEEdgeFlip.InsertLast(peNew->GetEdgeNext(1), 0);

	  for ( peCurr = (QEEdge*)(listQEEdgeFlip.RemoveFirst());
		peCurr != NULL;
		peCurr = (QEEdge*)(listQEEdgeFlip.RemoveFirst()) )
	    {
	      if ( !ConstrainedDelaunayEdge(peCurr) )
		{
		  DelaunayFlip(peCurr, &listQEEdgeFlip);

		  if ( peCurr->GetEdgeNext(0)->GetEdge()->GetFacet() == m_pfaNULL )
		    {
		      listQEEdgeCandidates.InsertLast(peCurr->GetEdgeNext(0)->GetEdge(), 0);
		    }
		  if ( peCurr->GetEdgePrev(0)->GetEdge()->GetFacet() == m_pfaNULL )
		    {
		      listQEEdgeCandidates.InsertLast(peCurr->GetEdgePrev(0)->GetEdge(), 0);
		    }
		  if ( peCurr->GetEdge()->GetEdgeNext(0)->GetEdge()->GetFacet() == m_pfaNULL )
		    {
		      listQEEdgeCandidates.InsertLast(peCurr->GetEdge()->GetEdgeNext(0)->GetEdge(), 0);
		    }
		  if ( peCurr->GetEdge()->GetEdgePrev(0)->GetEdge()->GetFacet() == m_pfaNULL )
		    {
		      listQEEdgeCandidates.InsertLast(peCurr->GetEdge()->GetEdgePrev(0)->GetEdge(), 0);
		    }
		}
	    }
	}
    }
}

//////////////////////////////////////////////////////////////////////
// CreateVoronoiPolygonCDT
//
// Creates an approximation to the Voronoi Diagram of a Polygon using the CDT
//
VOID QEMesh::CreateVoronoiPolygonCDT(QEMesh *pmeCDT)
{
  ASSERT( pmeCDT != NULL );

  QEMesh *pmeCDTCopy;
  QEMeshIter iterQEMesh;
  QEFacetIter iterQEFacet;
  QEFacet *pfa, *pfaNew; 
  QEVertex *pvx, *pvxNew;
  QEEdge *pe, *peCurr, *pePrev, *peNew; 
  QEEdge *ppe[2], *ppeNew[2];
  UINT_32 i, uiEdges, uiBoundaryEdges;
  //Point ppt[2];
  Point ppptE[3][2];
  UINT_32 uiResults;
  Point pptResults[2];
  //Vector u, v;

  BOOL bLift;

  bLift = TRUE;

  pmeCDTCopy = pmeCDT->Copy();
  pmeCDTCopy->SubdivideCDTBoundary();

  pmeCDTCopy->ClearForwardPointers();

  iterQEMesh.Init(pmeCDTCopy);

  // Copy the Vertices
  for ( pvx = iterQEMesh.PeekFirstVertex();
        pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pvxNew = CreateVertex(pvx->GetPoint());
      pvx->SetForwardVertex(pvxNew);
    }

  // Copy the edges
  for ( pe = iterQEMesh.PeekFirstEdge();
        pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      ASSERT( (pe->GetForwardEdge() == NULL && pe->GetEdge()->GetForwardEdge() == NULL) ||
	      (pe->GetForwardEdge() != NULL && pe->GetEdge()->GetForwardEdge() != NULL) );
      if ( pe->GetForwardEdge() == NULL )
	{
	  if ( pe->GetEdge()->GetFacet() == m_pfaNULL )
	    {
	      ASSERT( pe->GetFacet() != m_pfaNULL );
	      pe = pe->GetEdge();
	    }

	  ppe[0] = pe;
	  ppe[1] = pe->GetEdge();
	  if ( ppe[0]->GetFacet() == m_pfaNULL )
	    {
	      ppeNew[0] = CreateEdge();
	      ppeNew[1] = ppeNew[0]->GetEdge();

	      // External Edge
	      m_plistQEEdge->Remove(ppeNew[0]);
	      m_plistQEEdgeBoundary->InsertLast(ppeNew[0]);

	      ppeNew[0]->SetFacet(m_pfaNULL);
	      pfaNew = CreateFacet();
	      ppeNew[1]->SetFacet(pfaNew);

	      ppe[0]->SetForwardEdge(ppeNew[0]);
	      ppe[1]->SetForwardEdge(ppeNew[1]);

	      ppeNew[0]->SetVertex(ppe[0]->GetVertex()->GetForwardVertex());
	      ppeNew[1]->SetVertex(ppe[1]->GetVertex()->GetForwardVertex());
	    }
	  else
	    {
	      // Internal Edge -- Create an edge and two facets
	      ppeNew[0] = CreateEdge();
	      ppeNew[1] = ppeNew[0]->GetEdge();

	      ppe[0]->SetForwardEdge(ppeNew[0]);
	      ppe[1]->SetForwardEdge(ppeNew[1]);

	      pfaNew = CreateFacet();
	      ppeNew[0]->SetFacet(pfaNew);
	      pfaNew = CreateFacet();
	      ppeNew[1]->SetFacet(pfaNew);

	      ppeNew[0]->SetVertex(ppe[0]->GetVertex()->GetForwardVertex());
	      ppeNew[1]->SetVertex(ppe[1]->GetVertex()->GetForwardVertex());
	    }

	  for ( i = 0; i < 2; i++ )
	    {
	      for ( peCurr = ppe[i]->GetEdgePrev(1);
		    peCurr != ppe[i];
		    peCurr = peCurr->GetEdgePrev(1) )
		{
		  if ( peCurr->GetForwardEdge() != NULL )
		    {
		      Splice(1, peCurr->GetForwardEdge(), ppeNew[i]);
		      break;
		    }
		}
	    }
	}
    }

  // Triangulate the new facets
  for ( pfa = iterQEMesh.PeekFirstFacet();
        pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      ASSERT( pfa != m_pfaNULL );

      iterQEFacet.Init(pfa);
      uiEdges = 0;
      uiBoundaryEdges = 0;
      for ( pe = iterQEFacet.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEFacet.PeekNextEdge() )
	{
	  ASSERT( uiEdges < 3 );

	  ppptE[uiEdges][0] = pe->GetVertex()->GetPoint();
	  ppptE[uiEdges][1] = pe->GetEdge()->GetVertex()->GetPoint();

	  uiEdges++;

	  if ( pe->GetEdge()->GetFacet() == m_pfaNULL )
	    {
	      uiBoundaryEdges++;
	    }
	}
      ASSERT( uiEdges == 3 );

      switch ( uiBoundaryEdges )
	{
	case 0:
	  {
	    CalculateVoronoiPointVVV(pptResults, uiResults, 
				     ppptE[0][0], 
				     ppptE[1][0], 
				     ppptE[2][0], bLift);
	    ASSERT( uiResults == 1 );
	  }
	  break;
	case 1:
	  {
	    CalculateVoronoiPointVVV(pptResults, uiResults, 
				     ppptE[0][0], 
				     ppptE[1][0], 
				     ppptE[2][0], bLift);
	    ASSERT( uiResults == 1 );
	    /*
	    for ( pe = iterQEFacet.PeekFirstEdge();
		  pe != NULL;
		  pe = iterQEFacet.PeekNextEdge() )
	      {
		if ( pe->GetEdge()->GetFacet() == m_pfaNULL )
		  {
		    v = pe->GetEdge()->GetVertex()->GetPoint() - pe->GetVertex()->GetPoint();
		    v.Set(-v(1), v(0), 0.0);
		    v.CompletePlaneEquation(pe->GetVertex()->GetPoint());
		    if ( v.Normalize() )
		      {
			u = pptResults[0];
			u[2] = 0.0;
			u[3] = 1.0;
			pptResults[0][2] = v.Dot(u);
		      }
		    break;
		  }
	      }
	      */
	    /*
	    UINT_32 ppuiIndex[3][3] = { { 2, 0, 1 }, { 1, 2, 0 }, { 0, 1, 2 } };

	    for ( i = 0, pe = iterQEFacet.PeekFirstEdge();
		  pe != NULL;
		  i++, pe = iterQEFacet.PeekNextEdge() )
	      {
		if ( pe->GetEdge()->GetFacet() == m_pfaNULL )
		  {
		    break;
		  }
	      }
	    ASSERT( i < 3 );

	    CalculateVoronoiPointVVE(pptResults, uiResults, 
				     ppptE[ppuiIndex[i][0]][0], 
				     ppptE[ppuiIndex[i][1]][0], 
				     ppptE[ppuiIndex[i][2]][0], ppptE[ppuiIndex[i][2]][1], bLift);
	    ASSERT( uiResults > 0 );
	    */
	  }
	  break;
	case 2:
	  {
	    CalculateVoronoiPointEEE(pptResults, uiResults, 
				     ppptE[0][0], ppptE[0][1], 
				     ppptE[1][0], ppptE[1][1], 
				     ppptE[2][0], ppptE[2][1], bLift);
	    ASSERT( uiResults == 1 );
	    /*
	    UINT_32 ppuiIndex[3][3] = { { 0, 1, 2 }, { 2, 0, 1 }, { 1, 2, 0 } };

	    for ( i = 0, pe = iterQEFacet.PeekFirstEdge();
		  pe != NULL;
		  i++, pe = iterQEFacet.PeekNextEdge() )
	      {
		if ( pe->GetEdge()->GetFacet() != m_pfaNULL )
		  {
		    break;
		  }
	      }
	    ASSERT( i < 3 );

	    CalculateVoronoiPointVEE(pptResults, uiResults, 
				     ppptE[ppuiIndex[i][0]][0], 
				     ppptE[ppuiIndex[i][1]][0], ppptE[ppuiIndex[i][1]][1], 
				     ppptE[ppuiIndex[i][2]][0], ppptE[ppuiIndex[i][2]][1], bLift);
	    ASSERT( uiResults > 0 );
	    */
	  }
	  break;
	case 3:
	  {
	    CalculateVoronoiPointEEE(pptResults, uiResults,
				     ppptE[0][0], ppptE[0][1], 
				     ppptE[1][0], ppptE[1][1], 
				     ppptE[2][0], ppptE[2][1], bLift);
	    ASSERT( uiResults == 1 );
	  }
	  break;
	default:
	  {
	    ASSERT( FALSE );
	  }
	  break;
	}

      pvxNew = CreateVertex(pptResults[0]);

      pePrev = NULL;
      for ( pe = iterQEFacet.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEFacet.PeekNextEdge() )
	{
	  peNew = CreateEdge();

	  if ( pe->GetEdge()->GetFacet() == m_pfaNULL )
	    {
	      peNew->SetFacet(pe->GetForwardEdge()->GetFacet());
	      peNew->GetEdge()->SetFacet(pe->GetEdgePrev(0)->GetForwardEdge()->GetFacet());
	      peNew->SetVertex(pvxNew);
	      peNew->GetEdge()->SetVertex(pe->GetVertex()->GetForwardVertex());
	  
	      Splice(1, peNew->GetEdge(), pe->GetForwardEdge());
	      if ( pePrev != NULL )
		{
		  Splice(1, pePrev, peNew);
		}
	      pePrev = peNew;
	    }
	  else
	    {
	      peNew->SetFacet(pe->GetForwardEdge()->GetEdge()->GetFacet());
	      peNew->GetEdge()->SetFacet(pe->GetEdgePrev(0)->GetForwardEdge()->GetFacet());
	      peNew->SetVertex(pvxNew);
	      peNew->GetEdge()->SetVertex(pe->GetVertex()->GetForwardVertex());
	  
	      if ( pe->GetForwardEdge()->GetEdgePrev(1) == pe->GetForwardEdge() )
		{
		  // It has already been turned
		  Splice(1, peNew->GetEdge(), pe->GetForwardEdge()->GetEdge()->GetEdgePrev(0));
		  if ( pePrev != NULL )
		    {
		      Splice(1, pePrev, peNew);
		    }
		  Splice(1, peNew, pe->GetForwardEdge());
		  pe->GetForwardEdge()->SetVertex(pvxNew);
		}
	      else
		{
		  // We need to turn the edge
		  Splice(1, peNew->GetEdge(), pe->GetForwardEdge());
		  if ( pePrev != NULL )
		    {
		      Splice(1, pePrev, peNew);
		    }

		  // Remove the edge from its temporary place
		  Splice(1, pe->GetForwardEdge()->GetEdge(), pe->GetForwardEdge()->GetEdge()->GetEdgePrev(1));
		  Splice(1, pe->GetForwardEdge(), pe->GetForwardEdge()->GetEdgePrev(1));

		  // Splice the edge onto the new vertex ring
		  Splice(1, peNew, pe->GetForwardEdge());
		  pe->GetForwardEdge()->SetVertex(pvxNew);
		}

	      pePrev = pe->GetForwardEdge();
	    }
	}
    }
  DeallocateQEMesh(pmeCDTCopy);
}

VOID QEMesh::CreateVoronoiPolygonDivideAndConquer(QEMesh *pmeCDT)
{
  QEEdge *ppe[2], *pe, *peCurr;
  QEEdge **ppeContour;
  QEMeshIter iter;
  QEVertex *pvx;
  Point pt;
  UINT_32 i, uiSize;

  ASSERT( pmeCDT != NULL );

  uiSize = pmeCDT->GetBoundaryEdgeCount();
  ASSERT( uiSize > 0 );

  ppeContour = new QEEdge*[uiSize];
  ASSERT( ppeContour != NULL );

  //////////////////////////////////////////////////
  // Count the Boundary Edges
  pe = pmeCDT->m_plistQEEdgeBoundary->PeekFirst();
  ASSERT( pe != NULL );

  i = 0;
  peCurr = pe;
  peCurr->GetVertex()->SetID(i);
  ppeContour[i] = peCurr;
  for ( i++, peCurr = peCurr->GetEdgePrev(0);
	i < uiSize && peCurr != pe;
	i++, peCurr = peCurr->GetEdgePrev(0) )
    {
      peCurr->GetVertex()->SetID(i);
      ppeContour[i] = peCurr;
    }
  ASSERT( peCurr == pe );
  ASSERT( i == uiSize );

  CreateVoronoiPolygonDivideAndConquer(uiSize, ppeContour, ppe); 

  iter.Init(this);
  for ( pvx = iter.PeekFirstVertex();
	pvx != NULL;
	pvx = iter.PeekNextVertex() )
    {
      pt = pvx->GetPoint();
      if ( pt[3] == -1.0 )
	{
	  pt[2] = -pt[2];
	  pt[3] = 1.0;
	  pvx->SetPoint(pt);
	}
    }

  //////////////////////////////////////////////////
  // Z Monotone the edges

  {
    CObjectList<QEEdge> listQEEdge;
    QEEdge *pe;
    QEEdge *peSite, *peNew;
    QEVertex *pvxSite, *pvxNew, *pvxEndpoint;
    QEFacet *ppfa[4];
    Point pt;
    FLOAT fDistance, fDist;
    UINT_32 uiResults;
    BOOL bSplit;

    listQEEdge.Init();
    listQEEdge.Splice(m_plistQEEdge);

    for ( pe = listQEEdge.RemoveFirst();
	  pe != NULL;
	  pe = listQEEdge.RemoveFirst() )
      {
	m_plistQEEdge->InsertLast(pe);
	listQEEdge.Remove(pe->GetEdge());
	m_plistQEEdge->InsertLast(pe->GetEdge());

	if ( pe->GetFacet() != m_pfaNULL &&
	     pe->GetEdge()->GetFacet() != m_pfaNULL )
	  {
	    if ( ( pe->GetFacet()->GetForwardEdge() != NULL &&
		   pe->GetEdge()->GetFacet()->GetForwardVertex() != NULL ) ||
		 ( pe->GetFacet()->GetForwardVertex() != NULL &&
		   pe->GetEdge()->GetFacet()->GetForwardEdge() != NULL ) )
	      {
		if ( pe->GetFacet()->GetForwardEdge() != NULL &&
		     pe->GetEdge()->GetFacet()->GetForwardVertex() != NULL )
		  {
		    peSite = pe->GetFacet()->GetForwardEdge();
		    pvxSite = pe->GetEdge()->GetFacet()->GetForwardVertex();
		  }
		else
		  {
		    ASSERT( pe->GetFacet()->GetForwardVertex() != NULL &&
			    pe->GetEdge()->GetFacet()->GetForwardEdge() != NULL );

		    pvxSite = pe->GetFacet()->GetForwardVertex();
		    peSite = pe->GetEdge()->GetFacet()->GetForwardEdge();
		  }

		if ( peSite->GetVertex() != pvxSite &&
		     peSite->GetEdge()->GetVertex() != pvxSite )
		  {
		    CalculatePointProjectionOnLineSegment(pt, uiResults,
						   pvxSite->GetPoint(), 
						   peSite->GetVertex()->GetPoint(), 
						   peSite->GetEdge()->GetVertex()->GetPoint(),
						   TRUE);
		    if ( uiResults == 1 )
		      {
			fDistance = pt[2]/2.0;
			pt = (pvxSite->GetPoint() + pt) / 2.0;
			pt[2] = 0.0;
			pt[3] = 1.0;
		    
			ppfa[0] = pe->GetEdgePrev(1)->GetEdge()->GetFacet();
			ppfa[1] = pe->GetEdgeNext(1)->GetFacet();
			ppfa[2] = pe->GetEdge()->GetEdgePrev(1)->GetEdge()->GetFacet();
			ppfa[3] = pe->GetEdge()->GetEdgeNext(1)->GetFacet();
			
			bSplit = TRUE;
			for ( i = 0; i < 4; i++ )
			  {
			    if ( ppfa[i] != m_pfaNULL &&
				 ppfa[i] != pe->GetFacet() &&
				 ppfa[i] != pe->GetEdge()->GetFacet() )
			      {
				fDist = CalculateDistance(ppfa[i], pt, &pvxEndpoint);

				if ( ( pvxEndpoint == NULL || 
				       ( pvxEndpoint != NULL && pvxEndpoint != pvxSite ) ) &&
				     fDist < fDistance )
				  {
				    bSplit = FALSE;
				    break;
				  }
			      }
			  }

			if ( bSplit )
			  {
			    pt[2] = fDistance;
			    
			    fprintf(stderr, "Split Parabolic Edge: ( %.3f, %.3f ) && ( %.3f, %.3f ) <=> ( %.3f, %.3f ) at ( %.3f, %.3f, %.3f )\n",
				    pvxSite->GetPoint()(X),
				    pvxSite->GetPoint()(Y),
				    peSite->GetVertex()->GetPoint()(X),
				    peSite->GetVertex()->GetPoint()(Y),
				    peSite->GetEdge()->GetVertex()->GetPoint()(X),
				    peSite->GetEdge()->GetVertex()->GetPoint()(Y),
				    pt(X),
				    pt(Y),
				    pt(Z));
			    
			    pvxNew = CreateVertex(pt);
			    peNew = CreateEdge();
			    
			    peNew->SetFacet(pe->GetFacet());
			    peNew->GetEdge()->SetFacet(pe->GetEdge()->GetFacet());
			    peNew->SetVertex(pe->GetVertex());
			    peNew->GetEdge()->SetVertex(pvxNew);
			    pe->SetVertex(pvxNew);

			    Splice(1, pe->GetEdgePrev(1), peNew);
			    Splice(1, pe->GetEdgePrev(1), pe);
			    Splice(1, pe, peNew->GetEdge());
			  }
		      }
		  }
	      }
	  }
      }
  }
 
  delete []ppeContour;
}

VOID QEMesh::CreateVoronoiPolygonDivideAndConquer(UINT_32 uiSize, QEEdge **ppeContour, 
						  QEEdge **ppeLeftAndRight)
{
  ASSERT( uiSize > 0 );
  ASSERT( ppeContour != NULL );
  ASSERT( ppeLeftAndRight != NULL );

#ifndef COMPILER_PROBLEM
  switch ( uiSize )
    {
    case 1:
      {
	QEFacet *ppfaNew[2];
	QEVertex *ppvxNew[2];
	QEEdge *ppeNew[3];
	Point ppt[2], pt;
	Vector v;
	UINT_32 i;

	//////////////////////////////////////////////////
	// Create the Vertices

	ppt[0] = ppeContour[0]->GetEdge()->GetVertex()->GetPoint();
	ppt[1] = ppeContour[0]->GetVertex()->GetPoint();

	v = ppt[1] - ppt[0];
	v.Set(-v(1), v(0), 0.0);
	v.Normalize();

	pt.Set(v(0), v(1), -1.0, 0.0);

	ppvxNew[0] = CreateVertex(pt);

	v = -v;

	pt.Set(v(0), v(1), -1.0, 0.0);

	ppvxNew[1] = CreateVertex(pt);

	//////////////////////////////
	// Create the Edges

	for ( i = 0; i < 3; i++ )
	  {
	    ppeNew[i] = CreateEdge();
	  }

	//////////////////////////////
	// Create the Facets

	ppfaNew[0] = CreateFacet();
	ppfaNew[0]->SetForwardVertex(ppeContour[0]->GetEdge()->GetVertex());

	ppfaNew[1] = CreateFacet();
	ppfaNew[1]->SetForwardEdge(ppeContour[0]->GetEdge());

	//////////////////////////////
	// Link everything together

	Splice(1, ppeNew[0], ppeNew[1]);
	Splice(1, ppeNew[1], ppeNew[2]);
	Splice(1, ppeNew[0]->GetEdge(), ppeNew[1]->GetEdge());
	Splice(1, ppeNew[0]->GetEdge(), ppeNew[2]->GetEdge());

	ppeNew[0]->SetFacet(ppfaNew[0]);
	ppeNew[1]->GetEdge()->SetFacet(ppfaNew[0]);
	ppeNew[1]->SetFacet(ppfaNew[1]);
	ppeNew[2]->GetEdge()->SetFacet(ppfaNew[1]);

	ppeNew[0]->SetVertex(ppvxNew[0]);
	ppeNew[1]->SetVertex(ppvxNew[0]);
	ppeNew[2]->SetVertex(ppvxNew[0]);

	ppeNew[0]->GetEdge()->SetVertex(ppvxNew[1]);
	ppeNew[1]->GetEdge()->SetVertex(ppvxNew[1]);
	ppeNew[2]->GetEdge()->SetVertex(ppvxNew[1]);

	ppeLeftAndRight[0] = ppeNew[0]->GetEdge();
	ppeLeftAndRight[1] = ppeNew[2];
      }
      break;

    default:
      {
	QEEdge *ppeLeft[2], *ppeRight[2], *peBisector;
	QEEdge *peLeft, *peLeftNext, *peRight, *peRightNext, *peNew;
	QEEdge *peLeftStart, *peRightStart/*, *peCurr*/;
	QEVertex *pvxNew, *ppvxEndpoint[2];
	Point pt, ptLeft, ptRight, pptResults[2];
	Vector v;
	UINT_32 uiHalf, uiResults;
	BOOL bLeftCandidate, bRightCandidate;
	BOOL bLeftInsideContour, bRightInsideContour;
	BOOL bEnteredContour;
	BOOL bAttachToRight;
	FLOAT pfDistance[2];

	uiHalf = uiSize / 2;
	CreateVoronoiPolygonDivideAndConquer(uiHalf, ppeContour, ppeLeft);
	CreateVoronoiPolygonDivideAndConquer(uiSize - uiHalf, ppeContour + uiHalf, ppeRight);

	//////////////////////////////
	// Temporary Fix

	if ( ppeContour[uiSize-1]->GetVertex() == ppeContour[0]->GetEdge()->GetVertex() )
	  {
	    s_bOpenEndpoint = FALSE;
	  } 
	else
	  {
	    s_bOpenEndpoint = TRUE;
	    s_pvxOpenEndpoint = ppeContour[uiSize-1]->GetVertex();
	    s_ptOpenEndpoint = ppeContour[uiSize-1]->GetVertex()->GetPoint();
	  }

	// Temporary Fix
	//////////////////////////////


	ASSERT( ppeLeft[0]->GetEdge()->GetFacet()->GetForwardVertex() == ppeContour[0]->GetEdge()->GetVertex() );
	ASSERT( ppeLeft[0]->GetFacet() == m_pfaNULL );
	//ASSERT( ppeRight[1]->GetEdge()->GetFacet()->GetForwardEdge() == ppeContour[uiSize - 1]->GetEdge() );
	ASSERT( ppeRight[1]->GetFacet() == m_pfaNULL );

	// Find the bottom two vertices
	bLeftInsideContour = FALSE;
	bRightInsideContour = FALSE;
	while ( TRUE )
	  {
	    if ( Counterclockwise(ppeRight[0]->GetEdge()->GetFacet(), bRightInsideContour,
				  ppeLeft[1]->GetEdge()->GetFacet(), bLeftInsideContour,
				  ppeLeft[1]->GetEdgeNext(0)->GetEdge()->GetFacet(), bLeftInsideContour ) )
	      {
		ppeLeft[1] = ppeLeft[1]->GetEdgeNext(0);
	      }
	    else if ( Counterclockwise(ppeLeft[1]->GetEdge()->GetFacet(), bLeftInsideContour,
				       ppeRight[0]->GetEdgePrev(0)->GetEdge()->GetFacet(), bRightInsideContour,
				       ppeRight[0]->GetEdge()->GetFacet(), bRightInsideContour) )
	      {
		ppeRight[0] = ppeRight[0]->GetEdgePrev(0);
	      }
	    else
	      {
		break;
	      }
	  }

	ASSERT( ppeLeft[1]->GetFacet() == m_pfaNULL );
	ASSERT( ppeRight[0]->GetFacet() == m_pfaNULL );

	//////////////////////////////////////////////////
	// Merge
	
	// Break the outside edges and merge them together
	peLeft = ppeLeft[1]->GetEdge();
	peRight = ppeRight[0]->GetEdge();

	peLeftNext = peLeft->GetEdgeNext(0);
	peRightNext = peRight->GetEdgePrev(0);

	peLeft->GetEdge()->GetEdgePrev(1)->SetVertex(peLeft->GetEdge()->GetEdgePrev(1)->GetVertex());
	peRight->GetEdgePrev(1)->SetVertex(peRight->GetEdgePrev(1)->GetVertex());

	Splice(1, peLeft->GetEdge(), peLeft->GetEdge()->GetEdgePrev(1));
	Splice(1, peRight, peRight->GetEdgePrev(1));

	peBisector = CreateEdge();

	peBisector->GetEdge()->SetFacet(peLeft->GetFacet());
	peBisector->SetFacet(peRight->GetFacet());

	// Calculate bisector direction????
	v = CalculateBisectorDirection(peBisector->GetEdge()->GetFacet(), bLeftInsideContour, 
				       peBisector->GetFacet(), bRightInsideContour, TRUE);
	pt.Set(v(0), v(1), -1.0, 0.0);

	pvxNew = CreateVertex(pt);

	peLeft->GetEdge()->SetVertex(pvxNew);
	peRight->SetVertex(pvxNew);
	peBisector->GetEdge()->SetVertex(pvxNew);

	Splice(1, peRight, peLeft->GetEdge());
	Splice(1, peRight, peBisector->GetEdge());

	peLeft = peLeftNext;
	peRight = peRightNext;
	
	//////////////////////////////////////////////////
	// Store the left and right extreme sites

	ppeLeftAndRight[0] = ppeLeft[0]->GetEdge()->GetEdgeNext(0);
	ppeLeftAndRight[1] = ppeRight[1]->GetEdge()->GetEdgePrev(0);
	
	fprintf(stderr, "START MERGE PROCESS: OUTSIDE CONTOUR: Size = %ld\n", uiSize);
	bEnteredContour = FALSE;
	while ( TRUE )
	  {
	    fprintf(stderr, "START MERGE LOOP: LEFT %s: RIGHT %s\n", 
		   (bLeftInsideContour) ? ("INSIDE") : ("OUTSIDE"), 
		   (bRightInsideContour) ? ("INSIDE") : ("OUTSIDE"));

	    bLeftCandidate = FALSE;
	    bRightCandidate = FALSE;
	    ptLeft.Set(0.0, 0.0, INFINITY);
	    ptRight.Set(0.0, 0.0, INFINITY);

	    //////////////////////////////////////////////////
	    // Find the left candidate
		
	    fprintf(stderr, "  START: Choosing Left Candidate\n");
	    ASSERT( peLeft->GetEdge()->GetFacet() != m_pfaNULL );
	    peLeftStart = peLeft;
	    peLeftNext = peLeft;

	    if ( ( bLeftInsideContour && bRightInsideContour ) && 
		 peLeft->GetFacet()->GetForwardVertex() != NULL &&
		 peRight->GetFacet()->GetForwardEdge() != NULL &&
		 peLeft->GetFacet()->GetForwardVertex() ==
		 peRight->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex() )
	      {
		// Exiting the contour

		fprintf(stderr, "USING EXIT HELPER FOR CONCAVE VERTEX\n");

		peLeftNext = peLeft->GetEdgeNext(0);
		while ( peLeft != peBisector->GetEdge() &&
			( peLeft->GetEdge()->GetFacet() == m_pfaNULL ||
			  peLeft->GetEdge()->GetFacet()->GetForwardEdge() == NULL ||
			  peLeft->GetEdge()->GetFacet()->GetForwardEdge()->GetVertex() !=
			  peLeft->GetFacet()->GetForwardVertex() ) )
		  {
		    //CollapseEdge(peLeft);

		    peLeft = peLeftNext;
		    peLeftNext = peLeft->GetEdgeNext(0);
		  }
		ASSERT( peLeft != peBisector->GetEdge() );

		bLeftCandidate = TRUE;
		ptLeft = peLeft->GetFacet()->GetForwardVertex()->GetPoint();
	      }
	    else 
	      {
		while ( peLeftNext != peBisector->GetEdge() )
		  {
		    if ( !bLeftCandidate )
		      {
			// We are searching for any candidate site
			if ( Counterclockwise(peLeftNext->GetEdge()->GetFacet(), bLeftInsideContour,
					      peLeft->GetFacet(), bLeftInsideContour,
					      peRight->GetFacet(), bRightInsideContour) )
			  {
			    CalculateVoronoiPoint(pptResults, uiResults, 
						  peLeftNext->GetEdge()->GetFacet(), bLeftInsideContour,
						  peLeft->GetFacet(), bLeftInsideContour,
						  peRight->GetFacet(), bRightInsideContour, 
						  TRUE);
			    if ( uiResults > 0 )
			      {
				bLeftCandidate = TRUE;
				ptLeft = pptResults[0];
				/*
				// Remove all the previous site edges
				while ( peLeft != peLeftNext )
				  {
				    peCurr = peLeft;
				    peLeft = peLeft->GetEdgeNext(0);
				    CollapseEdge(peCurr);
				  }
				  */
				peLeft = peLeftNext;
			      }
			  }

			// Advance the next pointer
			peLeftNext = peLeftNext->GetEdgeNext(0);
			if ( peLeftNext->GetEdge()->GetFacet() == m_pfaNULL )
			  {
			    peLeftNext = peLeftNext->GetEdgeNext(0);
			  }
		      }
		    else
		      {
			// We have a candidate site and are looking for the optimal site
			if ( Counterclockwise(peLeftNext->GetEdge()->GetFacet(), bLeftInsideContour,
					      peLeft->GetFacet(), bLeftInsideContour,
					      peRight->GetFacet(), bRightInsideContour) )
			  {
			    CalculateVoronoiPoint(pptResults, uiResults, 
						  peLeftNext->GetEdge()->GetFacet(), bLeftInsideContour,
						  peLeft->GetFacet(), bLeftInsideContour,
						  peRight->GetFacet(), bRightInsideContour, 
						  TRUE);
			    if ( uiResults == 0 )
			      {
				break;
			      }
			    else
			      {
				pfDistance[0] = CalculateDistance(peLeftNext->GetEdge()->GetFacet(), ptLeft, &(ppvxEndpoint[0]));
				pfDistance[1] = CalculateDistance(peLeft->GetEdge()->GetFacet(), pptResults[0], &(ppvxEndpoint[1]));
			
				if ( ( ppvxEndpoint[0] != NULL &&
				       ( peLeft->GetEdge()->GetFacet()->GetForwardVertex() == ppvxEndpoint[0] ||
					 peLeft->GetFacet()->GetForwardVertex() == ppvxEndpoint[0] ||
					 peRight->GetFacet()->GetForwardVertex() == ppvxEndpoint[0] ) ) ||
				     pfDistance[0] == ptLeft(Z) )
				  {
				    if ( ( ppvxEndpoint[1] != NULL &&
					   ( peLeftNext->GetEdge()->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
					     peLeft->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
					     peRight->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ) ) ||
					 pfDistance[1] == pptResults[0](Z) )
				      {
					if ( peLeft->GetEdge()->GetFacet()->GetForwardVertex() != NULL )
					  {
					    break;
					  }
					else if ( peLeftNext->GetEdge()->GetFacet()->GetForwardVertex() != NULL )
					  {
					  }
					else
					  {
					    break;
					  }
				      }
				    else if ( pfDistance[1] > pptResults[0](Z) )
				      {
				      }
				    else
				      {
					break;
				      }
				  }
				else if ( pfDistance[0] > ptLeft(Z) )
				  {
				    ASSERT( pptResults[0](Z) >= pfDistance[1] ||
					    ( ppvxEndpoint[1] != NULL &&
					      ( peLeftNext->GetEdge()->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
						peLeft->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
						peRight->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ) ) );
				    break;
				  }
				else
				  {
				    ASSERT( pptResults[0](Z) <= pfDistance[1] ||
					    ( ppvxEndpoint[1] != NULL &&
					      ( peLeftNext->GetEdge()->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
						peLeft->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
						peRight->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ) ) );
				  }

				ptLeft = pptResults[0];
				/*
				if ( peLeft->GetEdgeNext(0)->GetFacet() == m_pfaNULL )
				  {
				    CollapseEdge(peLeft->GetEdgeNext(0));
				  }
				CollapseEdge(peLeft);
				*/
				peLeft = peLeftNext;
			    
				// Advance the next pointer
				peLeftNext = peLeftNext->GetEdgeNext(0);
				if ( peLeftNext->GetEdge()->GetFacet() == m_pfaNULL )
				  {
				    peLeftNext = peLeftNext->GetEdgeNext(0);
				  }
			      }
			  }
			else
			  {
			    break;
			  }
		      }
		  }
	      }
	    fprintf(stderr, "  END  : Choosing Left Candidate\n");

	    //////////////////////////////////////////////////
	    // Find the right candidate
		
	    fprintf(stderr, "  START: Choosing Right Candidate\n");
	    ASSERT( peRight->GetEdge()->GetFacet() != m_pfaNULL );
	    peRightStart = peRight;
	    peRightNext = peRight;

	    if ( !bEnteredContour && 
		 peLeft->GetFacet()->GetForwardEdge() != NULL &&
		 peRight->GetFacet()->GetForwardVertex() != NULL &&
		 peLeft->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex() ==
		 peRight->GetFacet()->GetForwardVertex() )
	      {
		// Entering the contour
		fprintf(stderr, "USING ENTER HELPER FOR CONVEX VERTEX\n");

		peRightNext = peRight->GetEdgePrev(0);
		while ( peRight != peBisector &&
			( peRight->GetEdge()->GetFacet() == m_pfaNULL ||
			  peRight->GetEdge()->GetFacet()->GetForwardEdge() == NULL ||
			  peRight->GetFacet()->GetForwardVertex() !=
			  peRight->GetEdge()->GetFacet()->GetForwardEdge()->GetVertex() ) )
		  {
		    //CollapseEdge(peRight->GetEdge());
			    
		    peRight = peRightNext;
		    peRightNext = peRight->GetEdgePrev(0);
		  }
		ASSERT( peRight != peBisector );
		
		bRightCandidate = TRUE;
		ptRight = peRight->GetFacet()->GetForwardVertex()->GetPoint();
	      }
	    else 
	      {
		while ( peRightNext != peBisector )
		  {
		    if ( !bRightCandidate )
		      {
			// We are searching for any candidate site
			if ( Counterclockwise(peLeft->GetFacet(), bLeftInsideContour,
					      peRight->GetFacet(), bRightInsideContour,
					      peRightNext->GetEdge()->GetFacet(), bRightInsideContour) )
			  {
			    CalculateVoronoiPoint(pptResults, uiResults, 
						  peLeft->GetFacet(), bLeftInsideContour,
						  peRight->GetFacet(), bRightInsideContour,
						  peRightNext->GetEdge()->GetFacet(), bRightInsideContour, 
						  TRUE);
			    if ( uiResults > 0 )
			      {
				bRightCandidate = TRUE;
				ptRight = pptResults[0];
				/*
				// Remove all the previous site edges
				while ( peRight != peRightNext )
				  {
				    peCurr = peRight;
				    peRight = peRight->GetEdgePrev(0);
				    CollapseEdge(peCurr->GetEdge());
				  }
				  */
				peRight = peRightNext;
			      }
			  }

			// Advance the next pointer
			peRightNext = peRightNext->GetEdgePrev(0);
			if ( peRightNext->GetEdge()->GetFacet() == m_pfaNULL )
			  {
			    peRightNext = peRightNext->GetEdgePrev(0);
			  }
		      }
		    else
		      {
			// We have a candidate site and are looking for the optimal site
			if ( Counterclockwise(peLeft->GetFacet(), bLeftInsideContour,
					      peRight->GetFacet(), bRightInsideContour,
					      peRightNext->GetEdge()->GetFacet(), bRightInsideContour) )
			  {
			    CalculateVoronoiPoint(pptResults, uiResults, 
						  peLeft->GetFacet(), bLeftInsideContour,
						  peRight->GetFacet(), bRightInsideContour,
						  peRightNext->GetEdge()->GetFacet(), bRightInsideContour, 
						  TRUE);
			    if ( uiResults == 0 )
			      {
				break;
			      }

			    pfDistance[0] = CalculateDistance(peRightNext->GetEdge()->GetFacet(), ptRight, &(ppvxEndpoint[0]));
			    pfDistance[1] = CalculateDistance(peRight->GetEdge()->GetFacet(), pptResults[0], &(ppvxEndpoint[1]));

			    if ( ( ppvxEndpoint[0] != NULL &&
				   ( peLeft->GetFacet()->GetForwardVertex() == ppvxEndpoint[0] ||
				     peRight->GetFacet()->GetForwardVertex() == ppvxEndpoint[0] ||
				     peRight->GetEdge()->GetFacet()->GetForwardVertex() == ppvxEndpoint[0] ) ) ||
				 pfDistance[0] == ptRight(Z) )
			      {
				if ( ( ppvxEndpoint[1] != NULL &&
				       ( peLeft->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
					 peRight->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
					 peRightNext->GetEdge()->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ) ) ||
				     pfDistance[1] == pptResults[0](Z) )
				  {
				    if ( peRight->GetEdge()->GetFacet()->GetForwardVertex() != NULL )
				      {
					break;
				      }
				    else if ( peRightNext->GetEdge()->GetFacet()->GetForwardVertex() != NULL )
				      {
				      }
				    else
				      {
					break;
				      }
				  }
				else if ( pfDistance[1] > pptResults[0](Z) )
				  {
				  }
				else
				  {
				    break;
				  }
			      }
			    else if ( pfDistance[0] > ptRight(Z) )
			      {
				ASSERT( pptResults[0](Z) >= pfDistance[1] ||
					( ppvxEndpoint[1] != NULL &&
					  ( peLeft->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
					    peRight->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
					    peRightNext->GetEdge()->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ) ) );
				break;
			      }
			    else
			      {
				ASSERT( pptResults[0](Z) <= pfDistance[1] ||
					( ppvxEndpoint[1] != NULL &&
					  ( peLeft->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
					    peRight->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
					    peRightNext->GetEdge()->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ) ) );
			      }

			    ptRight = pptResults[0];
			    /*		    
			    if ( peRight->GetEdgePrev(0)->GetFacet() == m_pfaNULL )
			      {
				CollapseEdge(peRight->GetEdgePrev(0)->GetEdge());
			      }
			    CollapseEdge(peRight->GetEdge());
			    */
			    peRight = peRightNext;

			    // Advance the next pointer
			    peRightNext = peRight->GetEdgePrev(0);
			    if ( peRightNext->GetEdge()->GetFacet() == m_pfaNULL )
			      {
				peRightNext = peRightNext->GetEdgePrev(0);
			      }
			  }
			else
			  {
			    break;
			  }
		      }
		  }
	      }

	    fprintf(stderr, "  END  : Choosing Right Candidate\n");

	    //////////////////////////////////////////////////
	    // Check if we are done

	    ASSERT( peLeft->GetEdge()->GetFacet() != m_pfaNULL );
	    ASSERT( peRight->GetEdge()->GetFacet() != m_pfaNULL );
	    if ( !bLeftCandidate && !bRightCandidate )
	      {
		if ( peLeft->GetFacet()->GetForwardEdge() != NULL &&
		     peRight->GetFacet()->GetForwardEdge() != NULL )
		  {
		    ASSERT( peRight->GetFacet()->GetForwardEdge()->GetEdge() == ppeContour[uiSize-1] );
		    peLeft = peLeftStart;
		    peRight = peRightStart;
		    bRightInsideContour = !bRightInsideContour;
		    continue;
		  }
		break;
	      }
		
	    //////////////////////////////////////////////////
	    // Decide whether to attach to the left or the right

	    if ( bLeftCandidate && bRightCandidate )
	      {
		pfDistance[0] = CalculateDistance(peLeft->GetEdge()->GetFacet(), ptRight, &(ppvxEndpoint[0]));
		pfDistance[1] = CalculateDistance(peRight->GetEdge()->GetFacet(), ptLeft, &(ppvxEndpoint[1]));

		if ( ( ppvxEndpoint[0] != NULL &&
		       ( peLeft->GetFacet()->GetForwardVertex() == ppvxEndpoint[0] ||
			 peRight->GetFacet()->GetForwardVertex() == ppvxEndpoint[0] ||
			 peRight->GetEdge()->GetFacet()->GetForwardVertex() == ppvxEndpoint[0] ) ) ||
		     pfDistance[0] >= ptRight(Z) )
		  {
		    if ( ( ppvxEndpoint[1] != NULL &&
			   ( peLeft->GetEdge()->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
			     peLeft->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
			     peRight->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ) ) ||
			 pfDistance[1] == ptLeft(Z) ) 
		      {
			if ( peRight->GetEdge()->GetFacet()->GetForwardVertex() != NULL )
			  {
			    bAttachToRight = TRUE;
			  }
			else if ( peLeft->GetEdge()->GetFacet()->GetForwardVertex() != NULL )
			  {
			    bAttachToRight = FALSE;
			  }
			else
			  {
			    bAttachToRight = TRUE;
			  }
		      }
		    else if ( pfDistance[1] > ptLeft(Z) )
		      {
			bAttachToRight = FALSE;
		      }
		    else
		      {
			bAttachToRight = TRUE;
		      }
		  }
		else if ( pfDistance[0] > ptRight(Z) )
		  {
		    ASSERT( ptLeft(Z) >= pfDistance[1] ||
			    ( ppvxEndpoint[1] != NULL &&
			      ( peLeft->GetEdge()->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
				peLeft->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
				peRight->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ) ) );
				    
		    bAttachToRight = TRUE;
		  }
		else
		  {
		    ASSERT( ptLeft(Z) <= pfDistance[1] ||
			    ( ppvxEndpoint[1] != NULL &&
			      ( peLeft->GetEdge()->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
				peLeft->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
				peRight->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ) ) );
		    bAttachToRight = FALSE;
		  }
	      }
	    else if ( bLeftCandidate )
	      {
		bAttachToRight = FALSE;
	      }
	    else
	      {
		ASSERT( bRightCandidate );
		bAttachToRight = TRUE;
	      }

	    //////////////////////////////////////////////////
	    // Extend the bisector and advance the pointer

	    if ( bAttachToRight )
	      {
		fprintf(stderr, "  Attach to Right Side: ( %.6f, %.6f, %.6f )\n", 
		       ptRight(X), ptRight(Y), ptRight(Z));

		// Remove all the previous site edges
		peLeft = peLeftStart;
		while ( peRightStart != peRight )
		  {
		    peRightNext = peRightStart->GetEdgePrev(0);
		    CollapseEdge(peRightStart->GetEdge());
		    peRightStart = peRightNext;
		  }

		if ( !bLeftInsideContour || !bRightInsideContour )
		  {
		    ptRight[3] = -1.0;
		  }

		if ( ptRight(Z) == 0.0 )
		  {
		    fprintf(stderr, "TRANSITION: OUT to IN\n");
		    bLeftInsideContour = !bLeftInsideContour;
		    bRightInsideContour = !bRightInsideContour;
		    bEnteredContour = TRUE;
		  }

		// Right Side
		ASSERT( bRightCandidate );
		ASSERT( peRight->GetEdge()->GetEdgePrev(1) != peRight->GetEdge() );

		peRightNext = peRight->GetEdge()->GetEdgePrev(0);

		pvxNew = CreateVertex(ptRight);
		peNew = CreateEdge();

		peRight->GetEdge()->GetEdgePrev(1)->SetVertex(peRight->GetEdge()->GetEdgePrev(1)->GetVertex());

		peRight->GetEdge()->SetVertex(pvxNew);
		peNew->GetEdge()->SetVertex(pvxNew);
		peBisector->SetVertex(pvxNew);

		peNew->GetEdge()->SetFacet(peBisector->GetEdge()->GetFacet());
		peNew->SetFacet(peRight->GetEdge()->GetFacet());

		Splice(1, peRight->GetEdge()->GetEdgePrev(1), peRight->GetEdge());
		Splice(1, peRight->GetEdge(), peBisector);
		Splice(1, peRight->GetEdge(), peNew->GetEdge());

		peBisector = peNew;
		peRight = peRightNext;
		if ( peRight->GetEdge()->GetFacet() == m_pfaNULL )
		  {
		    peRightNext = peRight->GetEdgePrev(0);
		    CollapseEdge(peRight->GetEdge());
		    peRight = peRightNext;		    
		  }		    
	      }
	    else
	      {
		fprintf(stderr, "  Attach to Left Side: ( %.6f, %.6f, %.6f )\n", 
		       ptLeft(X), ptLeft(Y), ptLeft(Z));
		
		// Remove all the previous site edges
		peRight = peRightStart;
		while ( peLeftStart != peLeft )
		  {
		    peLeftNext = peLeftStart->GetEdgeNext(0);
		    CollapseEdge(peLeftStart);
		    peLeftStart = peLeftNext;
		  }

		if ( !bLeftInsideContour || !bRightInsideContour )
		  {
		    ptLeft[3] = -1.0;
		  }

		if ( ptLeft(Z) == 0.0 )
		  {
		    fprintf(stderr, "TRANSITION: IN to OUT\n");
		    bLeftInsideContour = !bLeftInsideContour;
		    bRightInsideContour = !bRightInsideContour;
		  }

		// Left Side
		ASSERT( bLeftCandidate );
		ASSERT( peLeft->GetEdgePrev(1) != peLeft );

		peLeftNext = peLeft->GetEdge()->GetEdgeNext(0);
		
		pvxNew = CreateVertex(ptLeft);
		peNew = CreateEdge();

		peLeft->GetEdgePrev(1)->SetVertex(peLeft->GetEdgePrev(1)->GetVertex());

		peLeft->SetVertex(pvxNew);
		peNew->GetEdge()->SetVertex(pvxNew);
		peBisector->SetVertex(pvxNew);

		peNew->SetFacet(peBisector->GetFacet());
		peNew->GetEdge()->SetFacet(peLeft->GetEdge()->GetFacet());

		Splice(1, peLeft->GetEdgePrev(1), peLeft);
		Splice(1, peLeft, peNew->GetEdge());
		Splice(1, peLeft, peBisector);

		peBisector = peNew;
		peLeft = peLeftNext;
		if ( peLeft->GetEdge()->GetFacet() == m_pfaNULL )
		  {
		    peLeftNext = peLeft->GetEdgeNext(0);
		    CollapseEdge(peLeft);
		    peLeft = peLeftNext;		    
		  }
	      }

	    fprintf(stderr, "END   MERGE LOOP\n");
	  }

	fprintf(stderr, "END   MERGE PROCESS: LEFT %s: RIGHT %s: Size = %ld\n", 
	       (bLeftInsideContour) ? ("INSIDE") : ("OUTSIDE"), 
	       (bRightInsideContour) ? ("INSIDE") : ("OUTSIDE"),
	       uiSize);

	//////////////////////////////////////////////////
	// Close off the top

	v = CalculateBisectorDirection(peBisector->GetEdge()->GetFacet(), bLeftInsideContour, 
				       peBisector->GetFacet(), bRightInsideContour,
				       FALSE);
	pt.Set(v(0), v(1), -1.0, 0.0);
	pvxNew = CreateVertex(pt);

	peBisector->SetVertex(pvxNew);

	peNew = CreateEdge();
	peLeftStart = peNew->GetEdge();

	peNew->GetEdge()->SetFacet(m_pfaNULL);
	peNew->SetFacet(peBisector->GetEdge()->GetFacet());

	peNew->SetVertex(pvxNew);
	peNew->GetEdge()->SetVertex(peLeft->GetVertex());

	Splice(1, peLeft, peNew->GetEdge());
	Splice(1, peBisector, peNew);

	peNew = CreateEdge();
	peRightStart = peNew->GetEdge();

	peNew->GetEdge()->SetFacet(m_pfaNULL);
	peNew->SetFacet(peBisector->GetFacet());
	peNew->SetVertex(peRight->GetEdge()->GetVertex());
	peNew->GetEdge()->SetVertex(pvxNew);

	Splice(1, peBisector, peNew->GetEdge());
	Splice(1, peRight->GetEdge()->GetEdgePrev(1), peNew);

	//////////////////////////////
	// Recover Left and Right extremes

	ppeLeftAndRight[0] = ppeLeftAndRight[0]->GetEdgePrev(0)->GetEdge();
	if ( ppeLeftAndRight[0]->GetFacet() != m_pfaNULL )
	  {
	    ppeLeftAndRight[0] = peLeftStart;
	  }
	ppeLeftAndRight[1] = ppeLeftAndRight[1]->GetEdgeNext(0)->GetEdge();
	if ( ppeLeftAndRight[1]->GetFacet() != m_pfaNULL )
	  {
	    ppeLeftAndRight[1] = peRightStart;
	  }

	/*
	ASSERT( ppeLeftAndRight[0]->GetEdge()->GetFacet()->GetForwardVertex() == ppvxSite[0] );
	ASSERT( ppeLeftAndRight[0]->GetFacet() == m_pfaNULL );
	ASSERT( ppeLeftAndRight[1]->GetEdge()->GetFacet()->GetForwardVertex() == ppvxSite[uiSize-1] );
	ASSERT( ppeLeftAndRight[1]->GetFacet() == m_pfaNULL );
	*/
      }
      break;
    }
#endif // COMPILER_PROBLEM
}

VOID QEMesh::CreateVoronoiPolygonHeld(QEMesh *pmeCDT, BOOL bCleanUpContour)
{
  QEEdge *pe, *peCurr, *pePrev, *peStart, *peNext;
  QEEdge **pppeContourSegments[2];
  QEMeshIter iter;
  QEVertex *pvx, *pvxNew;
  Point pt;
  UINT_32 i, uiSize, uiSegment;
  BOOL bConvex;

  ASSERT( pmeCDT != NULL );

  //////////////////////////////////////////////////
  // Weird Static Variable Dependencies
  s_bOpenEndpoint = FALSE;
  // Weird Static Variable Dependencies
  //////////////////////////////////////////////////

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "START: CreateVoronoiPolygonHeld\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

  uiSize = pmeCDT->GetBoundaryEdgeCount();
  ASSERT( uiSize > 0 );

  pppeContourSegments[0] = new QEEdge*[uiSize];
  ASSERT( pppeContourSegments[0] != NULL );
  pppeContourSegments[1] = new QEEdge*[uiSize];
  ASSERT( pppeContourSegments[1] != NULL );

  //////////////////////////////////////////////////
  // Count the Boundary Edges
  pe = pmeCDT->m_plistQEEdgeBoundary->PeekFirst();
  ASSERT( pe != NULL );

  //////////////////////////////////////////////////
  // Copy the contour

  peCurr = CreateEdge();
  pt = pe->GetVertex()->GetPoint();
  pt[2] = 0.0;
  pvxNew = CreateVertex(pt);
  peCurr->SetVertex(pvxNew);

  pePrev = peCurr;

  peStart = peCurr;
  for ( pe = pe->GetEdgePrev(0);
	pe != pmeCDT->m_plistQEEdgeBoundary->PeekFirst();
	pe = pe->GetEdgePrev(0) )
    {
      peCurr = CreateEdge();
      pt = pe->GetVertex()->GetPoint();
      pt[2] = 0.0;
      pvxNew = CreateVertex(pt);
      peCurr->SetVertex(pvxNew);

      peCurr->GetEdge()->SetVertex(pePrev->GetVertex());
      Splice(1, peCurr->GetEdge(), pePrev);
      
      pePrev = peCurr;
    }
  peStart->GetEdge()->SetVertex(peCurr->GetVertex());
  Splice(1, peCurr, peStart->GetEdge());

  //////////////////////////////////////////////////
  // Remove epsilon length edges and 
  // collinear adjacent edge line segments
  if ( bCleanUpContour )
    {
      Point ppt[3];
      Vector v;
      FLOAT fMagnitude;

      fprintf(stderr, "START: Cleaning Up Contour: size = %ld\n", uiSize);

      for ( pe = peStart->GetEdgePrev(0);
	    pe != peStart;
	    pe = peNext )
	{
	  peNext = pe->GetEdgePrev(0);

	  ppt[0] = pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
	  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
	  ppt[2] = pe->GetVertex()->GetPoint();
	  v = ppt[2] - ppt[1];
	  v[2] = 0.0;
	  v[3] = 0.0;
	  fMagnitude = v.Magnitude();
	  if ( CounterclockwiseValue(ppt[0], ppt[1], ppt[2]) == 0.0 ||
	       fMagnitude < s_cfDistanceEpsilon )
	    {
	      // Remove unnecessary vertex and edge
	      fprintf(stderr, "  Removing edge: %f < %f\n    ( %f, %f ) ==> ( %f, %f ) ==> ( %f, %f )\n", fMagnitude, s_cfDistanceEpsilon, ppt[0](0), ppt[0](1), ppt[1](0), ppt[1](1), ppt[2](0), ppt[2](1));
	      uiSize--;
	      pePrev = pe->GetEdgeNext(0);
	      pePrev->SetVertex(pe->GetVertex());
	      Splice(1, pePrev, pe->GetEdge());
	      Splice(1, pe, peNext->GetEdge());
	      Splice(1, pePrev, peNext->GetEdge());
	      DestroyVertex(pe->GetEdge()->GetVertex());
	      DestroyEdge(pe->GetEdge());
	      DestroyEdge(pe);
	    }
	}

      if ( peStart == peStart->GetEdgePrev(0) )
	{
	  ASSERT( peStart->GetVertex() == peStart->GetEdge()->GetVertex() );
	  DestroyVertex(peStart->GetEdge()->GetVertex());
	  DestroyEdge(peStart->GetEdge());
	  DestroyEdge(peStart);
	  goto end;
	}
      else if ( peStart == peStart->GetEdgePrev(0)->GetEdgePrev(0) )
	{
	  pe = peStart->GetEdgePrev(0);
	  Splice(1, pe->GetEdge(), peStart);
	  Splice(1, pe, peStart->GetEdge());

	  DestroyVertex(pe->GetEdge()->GetVertex());
	  DestroyEdge(pe->GetEdge());
	  DestroyEdge(pe);

	  DestroyVertex(peStart->GetEdge()->GetVertex());
	  DestroyEdge(peStart->GetEdge());
	  DestroyEdge(peStart);

	  goto end;
	}
      else
	{
	  pe = peStart;
	  ppt[0] = pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
	  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
	  ppt[2] = pe->GetVertex()->GetPoint();
	  v = ppt[2] - ppt[1];
	  v[2] = 0.0;
	  v[3] = 0.0;
	  fMagnitude = v.Magnitude();
	  if ( CounterclockwiseValue(ppt[0], ppt[1], ppt[2]) == 0.0 ||
	       fMagnitude < s_cfDistanceEpsilon )
	    {
	      peStart = pe->GetEdgePrev(0);

	      // Remove unnecessary vertex and edge
	      fprintf(stderr, "  Removing edge: %f < %f\n    ( %f, %f ) ==> ( %f, %f ) ==> ( %f, %f )\n", fMagnitude, s_cfDistanceEpsilon, ppt[0](0), ppt[0](1), ppt[1](0), ppt[1](1), ppt[2](0), ppt[2](1));

	      uiSize--;
	      pePrev = pe->GetEdgeNext(0);
	      peNext = pe->GetEdgePrev(0);
	      pePrev->SetVertex(pe->GetVertex());
	      Splice(1, pePrev, pe->GetEdge());
	      Splice(1, pe, peNext->GetEdge());
	      Splice(1, pePrev, peNext->GetEdge());
	      DestroyVertex(pe->GetEdge()->GetVertex());
	      DestroyEdge(pe->GetEdge());
	      DestroyEdge(pe);

	      if ( peStart == peStart->GetEdgePrev(0)->GetEdgePrev(0) )
		{
		  pe = peStart->GetEdgePrev(0);
		  Splice(1, pe->GetEdge(), peStart);
		  Splice(1, pe, peStart->GetEdge());
		  
		  DestroyVertex(pe->GetEdge()->GetVertex());
		  DestroyEdge(pe->GetEdge());
		  DestroyEdge(pe);

		  DestroyVertex(peStart->GetEdge()->GetVertex());
		  DestroyEdge(peStart->GetEdge());
		  DestroyEdge(peStart);

		  goto end;
		}
	    }
	}

      fprintf(stderr, "END  : Cleaning Up Contour: size = %ld\n", uiSize);
    }


  //////////////////////////////////////////////////////////////////////
  // Create the Voronoi Diagram of the inside of the contour
  peStart = peStart->GetEdge();
  memset(pppeContourSegments[0], 0, uiSize*sizeof(QEEdge*));
  memset(pppeContourSegments[1], 0, uiSize*sizeof(QEEdge*));

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "  START: Inside Voronoi\n");
  fprintf(stderr, "    START: Find Convex Corner\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
  //////////////////////////////////////////////////
  // Locate a Convex corner which will start the first segment
  bConvex = TRUE;
  peCurr = peStart;
  do
    {
      if ( Counterclockwise(peCurr->GetEdgePrev(0)->GetVertex()->GetPoint(),
			    peCurr->GetVertex()->GetPoint(),
			    peCurr->GetEdge()->GetVertex()->GetPoint()) )
	{
	  // Convex corner starts a new segment
	  pppeContourSegments[0][0] = peCurr;
	  pppeContourSegments[1][0] = peCurr;
	  bConvex = FALSE;
	  break;
	}
      peCurr = peCurr->GetEdgeNext(0);
    }
  while ( peCurr != peStart );
  ASSERT( !bConvex );
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "    END  : Find Convex Corner\n");

  fprintf(stderr, "    START: Split contour into maximal reflex segments\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
  //////////////////////////////////////////////////
  // Divide the contour into maximal reflex segments
  pe = pppeContourSegments[0][0];
  i = 0;
  uiSegment = 0;
  peCurr = pe;
  peCurr->GetVertex()->SetID(i);
  for ( i++, peCurr = peCurr->GetEdgeNext(0);
	i < uiSize && peCurr != pe;
	i++, peCurr = peCurr->GetEdgeNext(0) )
    {
      peCurr->GetVertex()->SetID(i);
      if ( Counterclockwise(peCurr->GetEdgePrev(0)->GetVertex()->GetPoint(),
			    peCurr->GetVertex()->GetPoint(),
			    peCurr->GetEdge()->GetVertex()->GetPoint()) )
	{
	  // Convex corner starts a new segment
	  uiSegment++;
	  pppeContourSegments[0][uiSegment] = peCurr;
	  pppeContourSegments[1][uiSegment] = peCurr;	  
	}
      else
	{
	  // Concave corner or straight joint
	  pppeContourSegments[1][uiSegment] = peCurr;	  
	}
    }
  ASSERT( peCurr == pe );
  ASSERT( i == uiSize );
  uiSegment++;
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "    END  : Split contour into maximal reflex segments\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

  CreateVoronoiPolygonHeld(uiSegment, pppeContourSegments[0], pppeContourSegments[1], 1, bConvex, TRUE); 
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "  END  : Inside Voronoi\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

  //////////////////////////////////////////////////////////////////////
  // Create the Voronoi Diagram of the outside of the contour
  peStart = peStart->GetEdge();
  memset(pppeContourSegments[0], 0, uiSize*sizeof(QEEdge*));
  memset(pppeContourSegments[1], 0, uiSize*sizeof(QEEdge*));

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "  START: Outside Voronoi\n");
  fprintf(stderr, "    START: Find Convex Corner\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
  //////////////////////////////////////////////////
  // Locate a Convex corner which will start the first segment
  bConvex = TRUE;
  peCurr = peStart;
  do
    {
      if ( Counterclockwise(peCurr->GetEdgePrev(0)->GetVertex()->GetPoint(),
			    peCurr->GetVertex()->GetPoint(),
			    peCurr->GetEdge()->GetVertex()->GetPoint()) )
	{
	  // Convex corner starts a new segment
	  pppeContourSegments[0][0] = peCurr;
	  pppeContourSegments[1][0] = peCurr;
	  bConvex = FALSE;
	  break;
	}
      peCurr = peCurr->GetEdgeNext(0);
    }
  while ( peCurr != peStart );
  if ( bConvex )
    {
      // The contour is convex so there are no convex outer vertices
      pppeContourSegments[0][0] = peCurr;
      pppeContourSegments[0][0] = peCurr;
    }
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "    END  : Find Convex Corner\n");

  fprintf(stderr, "    START: Split contour into maximal reflex segments\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
  //////////////////////////////////////////////////
  // Divide the contour into maximal reflex segments
  pe = pppeContourSegments[0][0];
  i = 0;
  uiSegment = 0;
  peCurr = pe;
  peCurr->GetVertex()->SetID(i);
  for ( i++, peCurr = peCurr->GetEdgeNext(0);
	i < uiSize && peCurr != pe;
	i++, peCurr = peCurr->GetEdgeNext(0) )
    {
      peCurr->GetVertex()->SetID(i);
      if ( Counterclockwise(peCurr->GetEdgePrev(0)->GetVertex()->GetPoint(),
			    peCurr->GetVertex()->GetPoint(),
			    peCurr->GetEdge()->GetVertex()->GetPoint()) )
	{
	  // Convex corner starts a new segment
	  uiSegment++;
	  pppeContourSegments[0][uiSegment] = peCurr;
	  pppeContourSegments[1][uiSegment] = peCurr;	  
	}
      else
	{
	  // Concave corner or straight joint
	  pppeContourSegments[1][uiSegment] = peCurr;	  
	}
    }
  ASSERT( peCurr == pe );
  ASSERT( i == uiSize );
  uiSegment++;
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "    END  : Split contour into maximal reflex segments\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

  CreateVoronoiPolygonHeld(uiSegment, pppeContourSegments[0], pppeContourSegments[1], -1, bConvex, FALSE); 
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "  END  : Outside Voronoi\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

  //////////////////////////////////////////////////
  // Make the distance values on the exterior negative
  iter.Init(this);
  for ( pvx = iter.PeekFirstVertex();
	pvx != NULL;
	pvx = iter.PeekNextVertex() )
    {
      pt = pvx->GetPoint();
      if ( pt[3] == -1.0 )
	{
	  pt[2] = -pt[2];
	  pt[3] = 1.0;
	  pvx->SetPoint(pt);
	}
    }

end:
  delete []pppeContourSegments[0];
  delete []pppeContourSegments[1];

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "END  : CreateVoronoiPolygonHeld\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
}

VOID QEMesh::CreateVoronoiPolygonHeld(CObjectList<QEVertex> *plistQEVertex, 
				      BOOL bCleanUpContour)
{
  QEEdge *pe, *peCurr, *pePrev, *peStart, *peNext;
  QEEdge **pppeContourSegments[2];
  //QEMeshIter iter;
  CObjectListIter<QEVertex> iterQEVertex;  
  QEVertex *pvx, *pvxNew;
  Point pt;
  UINT_32 i, uiSize, uiSegment;
  BOOL bConvex;
  FLOAT fArea, fW;

  ASSERT( plistQEVertex != NULL );

  //////////////////////////////////////////////////
  // Weird Static Variable Dependencies
  s_bOpenEndpoint = FALSE;
  // Weird Static Variable Dependencies
  //////////////////////////////////////////////////

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "START: CreateVoronoiPolygonHeld\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

  uiSize = plistQEVertex->GetLength();
  ASSERT( uiSize > 0 );

  pppeContourSegments[0] = new QEEdge*[uiSize];
  ASSERT( pppeContourSegments[0] != NULL );
  pppeContourSegments[1] = new QEEdge*[uiSize];
  ASSERT( pppeContourSegments[1] != NULL );

  //////////////////////////////////////////////////
  // Determine the orientation of the contour

  fArea = CalculateContourArea(plistQEVertex);

  fW = 1.0;
  if ( fArea < 0.0 )
    {
      fW = -1.0;
    }

  //////////////////////////////////////////////////
  // Count the Boundary Edges

  iterQEVertex.Init(plistQEVertex);

  //////////////////////////////////////////////////
  // Copy the contour

  if ( fW == 1.0 )
    {
      pvx = iterQEVertex.PeekFirst();
    }
  else
    {
      pvx = iterQEVertex.PeekLast();
    }
  ASSERT( pvx != NULL );
  
  peCurr = CreateEdge();
  pt = pvx->GetPoint();
  pt[2] = 0.0;
  pvxNew = CreateVertex(pt);
  peCurr->SetVertex(pvxNew);
      
  pePrev = peCurr;
      
  peStart = peCurr;

  if ( fW == 1.0 )
    {
      pvx = iterQEVertex.PeekNext();
    }
  else
    {
      pvx = iterQEVertex.PeekPrev();
    }
  while ( pvx != NULL )
    {
      peCurr = CreateEdge();
      pt = pvx->GetPoint();
      pt[2] = 0.0;
      pvxNew = CreateVertex(pt);
      peCurr->SetVertex(pvxNew);
	  
      peCurr->GetEdge()->SetVertex(pePrev->GetVertex());
      Splice(1, peCurr->GetEdge(), pePrev);
      
      pePrev = peCurr;
      
      if ( fW == 1.0 )
	{
	  pvx = iterQEVertex.PeekNext();
	}
      else
	{
	  pvx = iterQEVertex.PeekPrev();
	}
    }
  peStart->GetEdge()->SetVertex(peCurr->GetVertex());
  Splice(1, peCurr, peStart->GetEdge());


  //////////////////////////////////////////////////
  // Remove epsilon length edges and 
  // collinear adjacent edge line segments
  if ( bCleanUpContour )
    {
      Point ppt[3];
      Vector v;
      FLOAT fMagnitude;

      fprintf(stderr, "START: Cleaning Up Contour: size = %ld\n", uiSize);

      for ( pe = peStart->GetEdgePrev(0);
	    pe != peStart;
	    pe = peNext )
	{
	  peNext = pe->GetEdgePrev(0);

	  ppt[0] = pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
	  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
	  ppt[2] = pe->GetVertex()->GetPoint();
	  v = ppt[2] - ppt[1];
	  v[2] = 0.0;
	  v[3] = 0.0;
	  fMagnitude = v.Magnitude();
	  if ( CounterclockwiseValue(ppt[0], ppt[1], ppt[2]) == 0.0 ||
	       fMagnitude < s_cfDistanceEpsilon )
	    {
	      // Remove unnecessary vertex and edge
	      fprintf(stderr, "  Removing edge: %f < %f\n    ( %f, %f ) ==> ( %f, %f ) ==> ( %f, %f )\n", fMagnitude, s_cfDistanceEpsilon, ppt[0](0), ppt[0](1), ppt[1](0), ppt[1](1), ppt[2](0), ppt[2](1));
	      uiSize--;
	      pePrev = pe->GetEdgeNext(0);
	      pePrev->SetVertex(pe->GetVertex());
	      Splice(1, pePrev, pe->GetEdge());
	      Splice(1, pe, peNext->GetEdge());
	      Splice(1, pePrev, peNext->GetEdge());
	      DestroyVertex(pe->GetEdge()->GetVertex());
	      DestroyEdge(pe->GetEdge());
	      DestroyEdge(pe);
	    }
	}

      if ( peStart == peStart->GetEdgePrev(0) )
	{
	  ASSERT( peStart->GetVertex() == peStart->GetEdge()->GetVertex() );
	  DestroyVertex(peStart->GetEdge()->GetVertex());
	  DestroyEdge(peStart->GetEdge());
	  DestroyEdge(peStart);
	  goto end;
	}
      else if ( peStart == peStart->GetEdgePrev(0)->GetEdgePrev(0) )
	{
	  pe = peStart->GetEdgePrev(0);
	  Splice(1, pe->GetEdge(), peStart);
	  Splice(1, pe, peStart->GetEdge());

	  DestroyVertex(pe->GetEdge()->GetVertex());
	  DestroyEdge(pe->GetEdge());
	  DestroyEdge(pe);

	  DestroyVertex(peStart->GetEdge()->GetVertex());
	  DestroyEdge(peStart->GetEdge());
	  DestroyEdge(peStart);

	  goto end;
	}
      else
	{
	  pe = peStart;
	  ppt[0] = pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
	  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
	  ppt[2] = pe->GetVertex()->GetPoint();
	  v = ppt[2] - ppt[1];
	  v[2] = 0.0;
	  v[3] = 0.0;
	  fMagnitude = v.Magnitude();
	  if ( CounterclockwiseValue(ppt[0], ppt[1], ppt[2]) == 0.0 ||
	       fMagnitude < s_cfDistanceEpsilon )
	    {
	      peStart = pe->GetEdgePrev(0);

	      // Remove unnecessary vertex and edge
	      fprintf(stderr, "  Removing edge: %f < %f\n    ( %f, %f ) ==> ( %f, %f ) ==> ( %f, %f )\n", fMagnitude, s_cfDistanceEpsilon, ppt[0](0), ppt[0](1), ppt[1](0), ppt[1](1), ppt[2](0), ppt[2](1));

	      uiSize--;
	      pePrev = pe->GetEdgeNext(0);
	      peNext = pe->GetEdgePrev(0);
	      pePrev->SetVertex(pe->GetVertex());
	      Splice(1, pePrev, pe->GetEdge());
	      Splice(1, pe, peNext->GetEdge());
	      Splice(1, pePrev, peNext->GetEdge());
	      DestroyVertex(pe->GetEdge()->GetVertex());
	      DestroyEdge(pe->GetEdge());
	      DestroyEdge(pe);

	      if ( peStart == peStart->GetEdgePrev(0)->GetEdgePrev(0) )
		{
		  pe = peStart->GetEdgePrev(0);
		  Splice(1, pe->GetEdge(), peStart);
		  Splice(1, pe, peStart->GetEdge());
		  
		  DestroyVertex(pe->GetEdge()->GetVertex());
		  DestroyEdge(pe->GetEdge());
		  DestroyEdge(pe);

		  DestroyVertex(peStart->GetEdge()->GetVertex());
		  DestroyEdge(peStart->GetEdge());
		  DestroyEdge(peStart);

		  goto end;
		}
	    }
	}

      fprintf(stderr, "END  : Cleaning Up Contour: size = %ld\n", uiSize);
    }

  //////////////////////////////////////////////////////////////////////
  // Create the Voronoi Diagram of the inside of the contour
  peStart = peStart->GetEdge();
  memset(pppeContourSegments[0], 0, uiSize*sizeof(QEEdge*));
  memset(pppeContourSegments[1], 0, uiSize*sizeof(QEEdge*));

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "  START: Inside Voronoi\n");
  fprintf(stderr, "    START: Find Convex Corner\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
  //////////////////////////////////////////////////
  // Locate a Convex corner which will start the first segment
  bConvex = TRUE;
  peCurr = peStart;
  do
    {
      if ( Counterclockwise(peCurr->GetEdgePrev(0)->GetVertex()->GetPoint(),
			    peCurr->GetVertex()->GetPoint(),
			    peCurr->GetEdge()->GetVertex()->GetPoint()) )
	{
	  // Convex corner starts a new segment
	  pppeContourSegments[0][0] = peCurr;
	  pppeContourSegments[1][0] = peCurr;
	  bConvex = FALSE;
	  break;
	}
      peCurr = peCurr->GetEdgeNext(0);
    }
  while ( peCurr != peStart );
  ASSERT( !bConvex );
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "    END  : Find Convex Corner\n");

  fprintf(stderr, "    START: Split contour into maximal reflex segments\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
  //////////////////////////////////////////////////
  // Divide the contour into maximal reflex segments
  pe = pppeContourSegments[0][0];
  i = 0;
  uiSegment = 0;
  peCurr = pe;
  peCurr->GetVertex()->SetID(i);
  for ( i++, peCurr = peCurr->GetEdgeNext(0);
	i < uiSize && peCurr != pe;
	i++, peCurr = peCurr->GetEdgeNext(0) )
    {
      peCurr->GetVertex()->SetID(i);
      if ( Counterclockwise(peCurr->GetEdgePrev(0)->GetVertex()->GetPoint(),
			    peCurr->GetVertex()->GetPoint(),
			    peCurr->GetEdge()->GetVertex()->GetPoint()) )
	{
	  // Convex corner starts a new segment
	  uiSegment++;
	  pppeContourSegments[0][uiSegment] = peCurr;
	  pppeContourSegments[1][uiSegment] = peCurr;	  
	}
      else
	{
	  // Concave corner or straight joint
	  pppeContourSegments[1][uiSegment] = peCurr;	  
	}
    }
  ASSERT( peCurr == pe );
  ASSERT( i == uiSize );
  uiSegment++;
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "    END  : Split contour into maximal reflex segments\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

  CreateVoronoiPolygonHeld(uiSegment, pppeContourSegments[0], pppeContourSegments[1], 
			   fW, bConvex, TRUE); 
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "  END  : Inside Voronoi\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

  //////////////////////////////////////////////////////////////////////
  // Create the Voronoi Diagram of the outside of the contour
  peStart = peStart->GetEdge();
  memset(pppeContourSegments[0], 0, uiSize*sizeof(QEEdge*));
  memset(pppeContourSegments[1], 0, uiSize*sizeof(QEEdge*));

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "  START: Outside Voronoi\n");
  fprintf(stderr, "    START: Find Convex Corner\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
  //////////////////////////////////////////////////
  // Locate a Convex corner which will start the first segment
  bConvex = TRUE;
  peCurr = peStart;
  do
    {
      if ( Counterclockwise(peCurr->GetEdgePrev(0)->GetVertex()->GetPoint(),
			    peCurr->GetVertex()->GetPoint(),
			    peCurr->GetEdge()->GetVertex()->GetPoint()) )
	{
	  // Convex corner starts a new segment
	  pppeContourSegments[0][0] = peCurr;
	  pppeContourSegments[1][0] = peCurr;
	  bConvex = FALSE;
	  break;
	}
      peCurr = peCurr->GetEdgeNext(0);
    }
  while ( peCurr != peStart );
  if ( bConvex )
    {
      // The contour is convex so there are no convex outer vertices
      pppeContourSegments[0][0] = peCurr;
      pppeContourSegments[0][0] = peCurr;
    }
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "    END  : Find Convex Corner\n");

  fprintf(stderr, "    START: Split contour into maximal reflex segments\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
  //////////////////////////////////////////////////
  // Divide the contour into maximal reflex segments
  pe = pppeContourSegments[0][0];
  i = 0;
  uiSegment = 0;
  peCurr = pe;
  peCurr->GetVertex()->SetID(i);
  for ( i++, peCurr = peCurr->GetEdgeNext(0);
	i < uiSize && peCurr != pe;
	i++, peCurr = peCurr->GetEdgeNext(0) )
    {
      peCurr->GetVertex()->SetID(i);
      if ( Counterclockwise(peCurr->GetEdgePrev(0)->GetVertex()->GetPoint(),
			    peCurr->GetVertex()->GetPoint(),
			    peCurr->GetEdge()->GetVertex()->GetPoint()) )
	{
	  // Convex corner starts a new segment
	  uiSegment++;
	  pppeContourSegments[0][uiSegment] = peCurr;
	  pppeContourSegments[1][uiSegment] = peCurr;	  
	}
      else
	{
	  // Concave corner or straight joint
	  pppeContourSegments[1][uiSegment] = peCurr;	  
	}
    }
  ASSERT( peCurr == pe );
  ASSERT( i == uiSize );
  uiSegment++;
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "    END  : Split contour into maximal reflex segments\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

  CreateVoronoiPolygonHeld(uiSegment, pppeContourSegments[0], pppeContourSegments[1], 
			   -fW, bConvex, FALSE); 
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "  END  : Outside Voronoi\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

end:
  delete []pppeContourSegments[0];
  delete []pppeContourSegments[1];

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "END  : CreateVoronoiPolygonHeld\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
}

FLOAT QEMesh::CalculateContourArea(CObjectList<QEVertex> *plistQEVertex)
{
  ASSERT( plistQEVertex != NULL );

  CObjectListIter<QEVertex> iter;
  QEVertex *pvx;
  Point pt, ptOrigin;
  Vector vCurr, vPrev;
  FLOAT fArea;

  // We can skip the first and the last edge because we are using the first vertex
  // as the origin, so those edges will have zero area triangles

  if ( plistQEVertex->GetLength() < 3 )
    {
      return 0.0;
    }

  iter.Init(plistQEVertex);

  fArea = 0.0;

  pvx = iter.PeekFirst();
  ASSERT( pvx != NULL );
  ptOrigin = pvx->GetPoint();

  pvx = iter.PeekNext();
  ASSERT( pvx != NULL );
  pt = pvx->GetPoint();

  vPrev = pt - ptOrigin;

  for ( pvx = iter.PeekNext();
	pvx != NULL;
	pvx = iter.PeekNext() )
    {
      pt = pvx->GetPoint();
      
      vCurr = pt - ptOrigin;
      
      fArea += ( vPrev(0)*vCurr(1) - vPrev(1)*vCurr(0) );

      vPrev = vCurr;
    }

  fArea /= 2.0;

  return fArea;
}

VOID QEMesh::MergeVoronoiPolygonsHeld(QEMesh *pmeVoronoi)
{
  ASSERT( pmeVoronoi != NULL );

  QEMeshIter piter[2];
  QEFacetIter iterQEFacet;
  QEEdge *pe, *ppe[2], *peCurr;
  QEFacet *pfaMin;
  Point pt;

#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
  fprintf(stderr, "START: MergeVoronoiPolygons\n");
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING

  piter[0].Init(this);
  piter[1].Init(pmeVoronoi);
  
  //////////////////////////////////////////////////
  // Find a vertex zone which extends to infinity
  //
  ppe[1] = NULL;
  for ( pe = piter[1].PeekFirstEdge();
	pe != NULL;
	pe = piter[1].PeekNextEdge() )
    {
      if ( pe->GetFacet() == GetNullFacet() )
	{
	  pe = pe->GetEdge();
	  ASSERT( pe->GetFacet() != GetNullFacet() );

	  if ( pe->GetFacet()->GetForwardVertex() != NULL )
	    {
	      ppe[1] = pe;
	      if ( pe->GetVertex() != pe->GetFacet()->GetForwardVertex() )
		{
		  for ( peCurr = pe->GetEdgeNext(0); 
			peCurr != pe;
			peCurr = peCurr->GetEdgeNext(0) )
		    {
		      if ( peCurr->GetVertex() == pe->GetFacet()->GetForwardVertex() )
			{
			  break;
			} 
		    }
		  ASSERT( peCurr != pe );
		  ppe[1] = peCurr;
		}

	      ASSERT( ppe[1]->GetVertex() == ppe[1]->GetFacet()->GetForwardVertex() &&
		      ppe[1]->GetFacet() == pe->GetFacet() );
	      break;
	    }
	}
    }
  ASSERT( ppe[1] != NULL );

  //////////////////////////////////////////////////
  // Find the facet in the base mesh which contains this vertex
  //

  pt = ppe[1]->GetVertex()->GetPoint();
  pt[2] = 0.0;
  pt[3] = 1.0;
  pfaMin = PointLocateVoronoiPolygonHeld(pt);

  ppe[0] = NULL;
  if ( pfaMin->GetForwardVertex() != NULL )
    {
      iterQEFacet.Init(pfaMin);
      for ( pe = iterQEFacet.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEFacet.PeekFirstEdge() )
	{
	  if ( pe->GetVertex() == pfaMin->GetForwardVertex() )
	    {
	      break;
	    }
	}
      ASSERT( pe != NULL );
      ppe[0] = pe;
    }
  else
    {
      ASSERT( pfaMin->GetForwardEdge() != NULL );

      pe = pfaMin->GetForwardEdge();
      ASSERT( Counterclockwise(pe->GetVertex()->GetPoint(), 
			       pe->GetEdge()->GetVertex()->GetPoint(),
			       pt) );
      ppe[0] = pe;
    }
  ASSERT( ppe[0] != NULL );

  //////////////////////////////////////////////////
  // Debugging Info
  //

#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
  if ( ppe[0]->GetFacet()->GetForwardVertex() != NULL )
    {
      fprintf(stderr, "  Vertex %p: ( %f, %f ) lies in the region of vertex %p: ( %f, %f )\n", ppe[1]->GetVertex(), ppe[1]->GetVertex()->GetPoint()(0), ppe[1]->GetVertex()->GetPoint()(1), ppe[0]->GetVertex(), ppe[0]->GetVertex()->GetPoint()(0), ppe[0]->GetVertex()->GetPoint()(1));
    }
  else
    {
      fprintf(stderr, "  Vertex %p: ( %f, %f ) lies in the region of edge %p => %p: ( %f, %f ) => ( %f, %f )\n", ppe[1]->GetVertex(), ppe[1]->GetVertex()->GetPoint()(0), ppe[1]->GetVertex()->GetPoint()(1), ppe[0]->GetVertex(), ppe[0]->GetEdge()->GetVertex(), ppe[0]->GetVertex()->GetPoint()(0), ppe[0]->GetVertex()->GetPoint()(1), ppe[0]->GetEdge()->GetVertex()->GetPoint()(0), ppe[0]->GetEdge()->GetVertex()->GetPoint()(1));
    }
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING

  //////////////////////////////////////////////////
  // Check for Inside or Outside Merge
  //
  BOOL bInside;
  BOOL bLeftInside, bRightInside;
  FLOAT fW;
  QEEdge *peBisector, *peBisectorPrev, *peBisectorEnd;
  QEEdge *peLeft, *peRight;
  QEEdge *peLeftNext, *peRightNext;
  QEEdge *peLeftStart, *peRightStart;
  QEEdge *peLeftEnd, *peRightEnd;
  QEFacet *pfaLeft, *pfaRight, *pfaTest;
  QEVertex *pvxNew, *ppvxEndpoint[2];
  FLOAT pfDistance[2];
  Point ptLeft, ptRight, pptResults[2];
  UINT_32 uiResults, uiMergeDirection;
  BOOL bReplacePoint;

  peLeftStart = ppe[1];
  peRightStart = ppe[0];
  peLeft = peLeftStart;
  peRight = peRightStart;

  pt = peLeft->GetEdgeNext(0)->GetVertex()->GetPoint();
  bLeftInside = !(pt(2) < 0.0 || pt(3) < 0.0);
  pt = peRight->GetEdgePrev(0)->GetVertex()->GetPoint();
  bRightInside = !(pt(2) < 0.0 || pt(3) < 0.0);
  
  if ( bLeftInside != bRightInside )
    {
      fprintf(stderr, "  ERROR: Distance orientation does not match: Base = %d != Merge = %d\n", bRightInside, bLeftInside);
      ASSERT( FALSE );
    }

  bInside = bLeftInside && bRightInside;
  fW = 1.0;
  if ( !bInside )
    {
      fW = -1.0;
    }

  //////////////////////////////////////////////////
  // Move all the geometry over into the base mesh
  //
  m_plistQEVertex->Splice(pmeVoronoi->m_plistQEVertex);
  m_plistQEFacet->Splice(pmeVoronoi->m_plistQEFacet);
  m_plistQEEdge->Splice(pmeVoronoi->m_plistQEEdge);

  //////////////////////////////////////////////////
  // Find the starting points
  //

  peBisector = NULL;
  peBisectorEnd = NULL;
  peLeft = peLeftStart;
  peRight = peRightStart;
  peLeftEnd = NULL;
  peRightEnd = NULL;

  while ( peLeftEnd == NULL || peRightEnd == NULL )
    {
      pfaLeft = NULL;
      pfaRight = NULL;

      //////////////////////////////////////////////////
      // Search for left starting point
      if ( peLeftEnd == NULL )
	{
#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
	  fprintf(stderr, "        Searching for minimum left starting point\n");
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING

	  ASSERT( peLeft->GetFacet() == peLeftStart->GetFacet() );
	  peLeft = peLeftStart;

	  for ( peLeftNext = peLeftStart->GetEdgePrev(0);
		peLeftNext != peLeftStart && 
		  peLeftNext->GetEdge()->GetFacet() != GetNullFacet();
		peLeftNext = peLeftNext->GetEdgePrev(0) )
	    {
	      ASSERT( peLeftNext->GetEdge()->GetFacet() != GetNullFacet() );
	      if ( Counterclockwise(peLeftNext->GetEdge()->GetFacet(), TRUE,
				    peLeft->GetFacet(), TRUE,
				    peRight->GetFacet(), TRUE) )
		{
#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
		  fprintf(stderr, "        Calculating left start candidate\n");
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING
		  CalculateVoronoiPoint(pptResults, uiResults, 
					peLeftNext->GetEdge()->GetFacet(), TRUE,
					peLeft->GetFacet(), TRUE,
					peRight->GetFacet(), TRUE, 
					TRUE);
		  if ( uiResults > 0 )
		    {
		      bReplacePoint = FALSE;
		      if ( peLeft == peLeftStart )
			{
			  //////////////////////////////
			  // No left start candidate yet
			  
#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
			  fprintf(stderr, "        Found first left start candidate\n");
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING
			  bReplacePoint = TRUE;
			}
		      else
			{
			  //////////////////////////////
			  // Do an InCircle test with the current left point
			  
			  ASSERT( peLeft->GetFacet() != GetNullFacet() &&
				  peLeft->GetFacet()->GetForwardEdge() != peLeft );
			  pfaTest = peLeft->GetEdge()->GetFacet();

			  bReplacePoint = FALSE;

			  //pfDistance[0] = CalculateDistance(peLeft->GetEdge()->GetFacet(), ptLeft, &(ppvxEndpoint[0]));
			  pfDistance[1] = CalculateDistance(pfaTest, pptResults[0], &(ppvxEndpoint[1]));

			  if ( pptResults[0](Z) <= ptLeft(Z) &&
			       ( pptResults[0](Z) <= pfDistance[1] ||
				 ( ppvxEndpoint[1] != NULL &&
				   ( peLeftNext->GetEdge()->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
				     peLeft->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ) ) ) )
			    {
			      //////////////////////////////
			      // Replacing left candidate

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			      fprintf(stderr, "        Updated left start candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			      bReplacePoint = TRUE;
			    }
			}
		      
		      if ( bReplacePoint )
			{
			  ptLeft = pptResults[0];
			  peLeft = peLeftNext;
			  pfaLeft = peLeft->GetEdge()->GetFacet();
			}
		    }
		}
	    }
	}

      //////////////////////////////////////////////////
      // Search for right starting point
      if ( peRightEnd == NULL )
	{
#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
	  fprintf(stderr, "        Searching for minimum right starting point\n");
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING

	  ASSERT( peRight->GetFacet() == peRightStart->GetFacet() );
	  peRight = peRightStart;

	  for ( peRightNext = peRightStart->GetEdgeNext(0);
		peRightNext != peRightStart && 
		  peRightNext->GetEdge()->GetFacet() != GetNullFacet();
		peRightNext = peRightNext->GetEdgeNext(0) )
	    {
	      ASSERT( peRightNext->GetEdge()->GetFacet() != GetNullFacet() );
	      if ( Counterclockwise(peLeft->GetFacet(), TRUE,
				    peRight->GetFacet(), TRUE,
				    peRightNext->GetEdge()->GetFacet(), TRUE) )
		{
#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
		  fprintf(stderr, "        Calculating right start candidate\n");
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING
		  CalculateVoronoiPoint(pptResults, uiResults, 
					peLeft->GetFacet(), TRUE,
					peRight->GetFacet(), TRUE, 
					peRightNext->GetEdge()->GetFacet(), TRUE,
					TRUE);
		  if ( uiResults > 0 )
		    {
		      bReplacePoint = FALSE;
		      if ( peRight == peRightStart )
			{
			  //////////////////////////////
			  // No right start candidate yet
			  
#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
			  fprintf(stderr, "        Found first right start candidate\n");
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING
			  bReplacePoint = TRUE;
			}
		      else
			{
			  //////////////////////////////
			  // Do an InCircle test with the current right point
			  
			  ASSERT( peRight->GetFacet() != GetNullFacet() &&
				  peRight->GetFacet()->GetForwardEdge() != peRight );
			  pfaTest = peRight->GetEdge()->GetFacet();

			  bReplacePoint = FALSE;

			  //pfDistance[0] = CalculateDistance(peRight->GetEdge()->GetFacet(), ptRight, &(ppvxEndpoint[0]));
			  pfDistance[1] = CalculateDistance(pfaTest, pptResults[0], &(ppvxEndpoint[1]));

			  if ( pptResults[0](Z) <= ptRight(Z) &&
			       ( pptResults[0](Z) <= pfDistance[1] ||
				 ( ppvxEndpoint[1] != NULL &&
				   ( peRightNext->GetEdge()->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
				     peRight->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ) ) ) )
			    {
			      //////////////////////////////
			      // Replacing right candidate

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			      fprintf(stderr, "        Updated right start candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			      bReplacePoint = TRUE;
			    }
			}
		      
		      if ( bReplacePoint )
			{
			  ptRight = pptResults[0];
			  peRight = peRightNext;
			  pfaRight = peRight->GetEdge()->GetFacet();
			}
		    }
		}
	    }
	}

      //////////////////////////////////////////////////
      // Decide which side to attach to
      uiMergeDirection = 0;
      if ( pfaLeft != NULL && pfaRight == NULL )
	{
	  uiMergeDirection = 1;
	}
      else if ( pfaLeft == NULL && pfaRight != NULL )
	{
	  uiMergeDirection = 2;
	}
      else if ( pfaLeft != NULL && pfaRight != NULL )
	{
	  if ( ptLeft(2) < ptRight(2) )
	    {
	      uiMergeDirection = 1;
	    }
	  else
	    {
	      uiMergeDirection = 2;
	    }
	}
      else
	{
	  ASSERT( FALSE );
	}

      //////////////////////////////////////////////////
      // Attach to the correct side
      switch ( uiMergeDirection )
	{
	case 1:
	  {
	    //////////////////////////////////////////////////
	    // Attach to the left
#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
		fprintf(stderr, "      Attaching to left ( %.3f, %.3f, %.3f )\n", ptLeft(X), ptLeft(Y), ptLeft(Z));
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING

	    if ( peLeftEnd == NULL )
	      {
#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
		fprintf(stderr, "      Starting on left\n");
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING

		for ( peLeftEnd = peLeft->GetEdgePrev(0);
		      peLeftEnd != peLeft &&
			peLeftEnd->GetEdge()->GetFacet() != GetNullFacet();
		      peLeftEnd = peLeftEnd->GetEdgePrev(0) )
		  {
		  }

		ASSERT( peLeftEnd->GetEdge()->GetFacet() == GetNullFacet() );
		if ( peLeftEnd->GetEdge()->GetFacet() == GetNullFacet() )
		  {
		    //////////////////////////////
		    // Detach the left end edge
		    peLeftEnd->GetEdge()->GetEdgePrev(1)->SetVertex(peLeftEnd->GetEdge()->GetEdgePrev(1)->GetVertex());
		    Splice(1, peLeftEnd->GetEdge(), peLeftEnd->GetEdge()->GetEdgePrev(1));
		  }

		//////////////////////////////////////////////////
		// Start the bisector to the left side
		ptLeft[3] = fW;
		
		peLeftNext = peLeft->GetEdge()->GetEdgeNext(0);
		peBisectorPrev = peBisector;

		pvxNew = CreateVertex(ptLeft);
		peBisector = CreateEdge();
		
		peLeft->GetEdgePrev(1)->SetVertex(peLeft->GetEdgePrev(1)->GetVertex());
		if ( peLeftEnd->GetEdge()->GetFacet() != GetNullFacet() )
		  {
		    peLeftEnd = CreateEdge();
		    peLeftEnd->SetFacet(peLeft->GetFacet());
		    peLeftEnd->SetVertex(peLeft->GetVertex());
		    Splice(1, peLeftEnd, peLeft->GetEdgePrev(1));
		  }
		Splice(1, peLeft, peLeft->GetEdgePrev(1));
		
		peLeft->SetVertex(pvxNew);
		peBisector->GetEdge()->SetVertex(pvxNew);
		if ( peBisectorPrev != NULL )
		  {
		    peBisectorPrev->SetVertex(pvxNew);
		    Splice(1, peBisectorPrev, peBisector->GetEdge());
		  }
		else
		  {
		    peBisectorEnd = peLeft;
		  }
		Splice(1, peLeft, peBisector->GetEdge());
		
		peLeft = peLeftNext;
		
		peBisector->SetFacet(peRight->GetFacet());
		peBisector->GetEdge()->SetFacet(peLeft->GetFacet());
		
		if ( peLeft->GetEdge()->GetFacet() == GetNullFacet() &&
		     peLeft->GetFacet()->GetForwardEdge() != peLeft && 
		     peLeft->GetFacet()->GetForwardVertex() != peLeft->GetEdge()->GetVertex() )
		  {
		    peLeftNext = peLeft->GetEdgeNext(0);
		    CollapseEdge(peLeft);
		    peLeft = peLeftNext;
		  }
	      }
	    else
	      {
		ASSERT( FALSE );
	      }
	  }
	break;
	case 2:
	  {
	    //////////////////////////////////////////////////
	    // Attach to the right
#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
		fprintf(stderr, "      Attaching to right ( %.3f, %.3f, %.3f )\n", ptRight(X), ptRight(Y), ptRight(Z));
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING

	    if ( peRightEnd == NULL )
	      {
#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
		fprintf(stderr, "      Starting on right\n");
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING
		for ( peRightEnd = peRight->GetEdgePrev(0);
		      peRightEnd != peRight &&
			peRightEnd->GetEdge()->GetFacet() != GetNullFacet();
		      peRightEnd = peRightEnd->GetEdgePrev(0) )
		  {
		  }

		if ( peRightEnd->GetEdge()->GetFacet() == GetNullFacet() )
		  {
		    //////////////////////////////
		    // Detach the right end edge
		    peRightEnd->GetEdgePrev(1)->SetVertex(peRightEnd->GetEdgePrev(1)->GetVertex());
		    Splice(1, peRightEnd, peRightEnd->GetEdgePrev(1));
		  }

		//////////////////////////////////////////////////
		// Start the bisector to the left side
		ptRight[3] = fW;
		
		peRightNext = peRight->GetEdge()->GetEdgePrev(0);
		peBisectorPrev = peBisector;

		pvxNew = CreateVertex(ptRight);
		peBisector = CreateEdge();
		
		peRight->GetEdge()->GetEdgePrev(1)->SetVertex(peRight->GetEdge()->GetEdgePrev(1)->GetVertex());
		if ( peRightEnd->GetEdge()->GetFacet() != GetNullFacet() )
		  {
		    peRightEnd = CreateEdge();
		    peRightEnd->SetFacet(peRight->GetFacet());
		    peRightEnd->GetEdge()->SetFacet(GetNullFacet());
		    peRightEnd->GetEdge()->SetVertex(peRight->GetEdge()->GetVertex());
		    Splice(1, peRightEnd->GetEdge(), peRight->GetEdge()->GetEdgePrev(1));
		  }
		Splice(1, peRight->GetEdge(), peRight->GetEdge()->GetEdgePrev(1));
		
		peRight->GetEdge()->SetVertex(pvxNew);
		peBisector->GetEdge()->SetVertex(pvxNew);
		Splice(1, peRight->GetEdge(), peBisector->GetEdge());
		if ( peBisectorPrev != NULL )
		  {
		    peBisectorPrev->SetVertex(pvxNew);
		    Splice(1, peBisectorPrev, peBisector->GetEdge());
		  }
		else
		  {
		    peBisectorEnd = peBisector->GetEdge();
		  }
		
		peRight = peRightNext;
		
		peBisector->SetFacet(peRight->GetFacet());
		peBisector->GetEdge()->SetFacet(peLeft->GetFacet());
		
		if ( peRight->GetEdge()->GetFacet() == GetNullFacet() &&
		     peRight->GetFacet()->GetForwardEdge() != peRight &&
		     peRight->GetFacet()->GetForwardVertex() != peRight->GetVertex() )
		  {
		    peRightNext = peRight->GetEdgePrev(0);
		    CollapseEdge(peRight->GetEdge());
		    peRight = peRightNext;
		  }
	      }
	    else
	      {
		ASSERT( FALSE );
	      }
	  }
	break;
	default:
	  {
	    ASSERT( FALSE );
	  }
	  break;
	}

    }

  //////////////////////////////////////////////////
  // Run the normal merge over most of the rest of the bisector path
  //
  MergeVoronoiPolygonsHeld(peLeftEnd, peLeft, peBisector, peRight, peRightEnd, fW);

  //////////////////////////////////////////////////
  // Connect up the bisector loop
  //
  
  ASSERT( peLeftEnd == peLeft );
  ASSERT( peLeftEnd->GetEdgeNext(1) == peLeftEnd );
  ASSERT( peLeftEnd->GetEdge()->GetEdgeNext(1) == peLeftEnd->GetEdge() );
  ASSERT( peRightEnd == peRight );
  ASSERT( peRightEnd->GetEdgeNext(1) == peRightEnd );
  ASSERT( peRightEnd->GetEdge()->GetEdgeNext(1) == peRightEnd->GetEdge() );
  ASSERT( peBisector->GetFacet() == peBisectorEnd->GetEdgePrev(0)->GetFacet() );
  ASSERT( peBisector->GetEdge()->GetFacet() == peBisectorEnd->GetFacet() );

  peBisector->SetVertex(peBisectorEnd->GetVertex());
  Splice(1, peBisector, peBisectorEnd);

  DestroyVertex(peLeftEnd->GetVertex());
  DestroyVertex(peRightEnd->GetEdge()->GetVertex());
  DestroyEdge(peLeftEnd->GetEdge());
  DestroyEdge(peLeftEnd);
  DestroyEdge(peRightEnd->GetEdge());
  DestroyEdge(peRightEnd);

  //////////////////////////////////////////////////
  // Deallocate the operand mesh
  //
  DeallocateQEMesh(pmeVoronoi);

#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
  fprintf(stderr, "END  : MergeVoronoiPolygons\n");
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING
}

QEFacet *QEMesh::PointLocateVoronoiPolygonHeld(const Point &pt)
{
  QEFacet *ppfa[2];
  QEVertex *ppvxEndpoint[2];
  FLOAT pfDistance[2];

  ppfa[0] = PointLocateVoronoiPolygonHeldGradientDescent(pt);
  ppfa[1] = PointLocateVoronoiPolygonHeldNaive(pt);
  if ( ppfa[0] != ppfa[1] )
    {
      pfDistance[0] = CalculateDistance(ppfa[0], pt, &ppvxEndpoint[0]);
      pfDistance[1] = CalculateDistance(ppfa[1], pt, &ppvxEndpoint[1]);
      ASSERT( pfDistance[0] - EPSILON <= pfDistance[1] && 
	      pfDistance[0] + EPSILON >= pfDistance[1] );
    }

  return ppfa[0];
}

QEFacet *QEMesh::PointLocateVoronoiPolygonHeldNaive(const Point &pt)
{
  QEMeshIter iterQEMesh;
  QEFacet *pfa, *pfaMin;
  QEEdge *pe;
  QEVertex *pvxEndpoint;
  FLOAT fDistance, fDistanceMin;

  iterQEMesh.Init(this);

  pfaMin = NULL;
  fDistanceMin = INFINITY;
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      fDistance = CalculateDistance(pfa, pt, &pvxEndpoint);
      if ( pvxEndpoint == NULL )
	{
	  if ( pfaMin == NULL ||
	       fDistance < fDistanceMin )
	    {
	      fDistanceMin = fDistance;
	      pfaMin = pfa;
	    }
	}
    }
  ASSERT( pfaMin != NULL );

  pe = pfaMin->GetForwardEdge();
  if ( pe != NULL )
    {
      if ( Counterclockwise(pe->GetEdge()->GetVertex()->GetPoint(),
			    pe->GetVertex()->GetPoint(),
			    pt) )
	{
	  pfaMin = pe->GetEdge()->GetFacet();
	}
    }
  ASSERT( pfaMin != NULL );
  ASSERT( pfaMin != GetNullFacet() );

  return pfaMin;
}

QEFacet *QEMesh::PointLocateVoronoiPolygonHeldGradientDescent(const Point &pt)
{
  QEEdge *pe, *peCurr;
  QEVertex *pvxEndpoint;
  FLOAT fDistance, fDistanceMin;
  BOOL bDone;

  pe = m_plistQEEdge->PeekFirst();
  if ( pe->GetFacet() == GetNullFacet() )
    {
      pe = pe->GetEdge();
    }
  ASSERT( pe->GetFacet() != GetNullFacet() );

  fDistanceMin = CalculateDistance(pe->GetFacet(), pt, &pvxEndpoint);
  if ( pvxEndpoint != NULL )
    {
      ASSERT( pe->GetFacet()->GetForwardEdge() != NULL );
      pe = pe->GetFacet()->GetForwardEdge();
      if ( pe->GetVertex() != pvxEndpoint )
	{
	  pe = pe->GetEdge();
	}
      ASSERT( pe->GetVertex() == pvxEndpoint );
      
      if ( pe->GetEdgeNext(1)->GetFacet()->GetForwardVertex() == pvxEndpoint )
	{
	  pe = pe->GetEdgeNext(1);
	}
      else
	{
	  pe = pe->GetEdgePrev(1)->GetEdgePrev(1);
	}
      ASSERT( pe->GetFacet()->GetForwardVertex() == pvxEndpoint );
    }
  else if ( pe->GetFacet()->GetForwardEdge() == pe )
    {
      if ( Counterclockwise(pe->GetEdge()->GetVertex()->GetPoint(),
			    pe->GetVertex()->GetPoint(),
			    pt) )
	{
	  pe = pe->GetEdge();
	}
    }

  bDone = FALSE;
  while ( !bDone )
    {
      bDone = TRUE;
      peCurr = pe;
      do
	{
	  if ( peCurr->GetEdge()->GetFacet() != GetNullFacet() )
	    {
	      if ( peCurr->GetFacet()->GetForwardEdge() == peCurr )
		{
		  if ( Counterclockwise(peCurr->GetEdge()->GetVertex()->GetPoint(),
					peCurr->GetVertex()->GetPoint(),
					pt) )
		    {
		      pe = peCurr->GetEdge();
		      bDone = FALSE;
		      break;
		    }
		}
	      else
		{
		  fDistance = CalculateDistance(peCurr->GetEdge()->GetFacet(), pt, &pvxEndpoint);
		  if ( fDistance < fDistanceMin &&
		       ( pvxEndpoint == NULL ||
			 peCurr->GetFacet()->GetForwardVertex() != pvxEndpoint ) )
		    {
		      fDistanceMin = fDistance;
		      pe = peCurr->GetEdge();
		      if ( pvxEndpoint != NULL )
			{
			  ASSERT( pe->GetFacet()->GetForwardEdge() != NULL );
			  pe = pe->GetFacet()->GetForwardEdge();
			  if ( pe->GetVertex() != pvxEndpoint )
			    {
			      pe = pe->GetEdge();
			    }
			  ASSERT( pe->GetVertex() == pvxEndpoint );
			  
			  if ( pe->GetEdgeNext(1)->GetFacet()->GetForwardVertex() == pvxEndpoint )
			    {
			      pe = pe->GetEdgeNext(1);
			    }
			  else
			    {
			      pe = pe->GetEdgePrev(1)->GetEdgePrev(1);
			    }
			  ASSERT( pe->GetFacet()->GetForwardVertex() == pvxEndpoint );
			}
		      else if ( pe->GetFacet()->GetForwardEdge() != NULL )
			{
			  peCurr = pe->GetFacet()->GetForwardEdge();
			  if ( Counterclockwise(peCurr->GetEdge()->GetVertex()->GetPoint(),
						peCurr->GetVertex()->GetPoint(),
						pt) )
			    {
			      pe = peCurr->GetEdge();
			    }
			}
		      
		      bDone = FALSE;
		      break;
		    }
		}
	    }

	  peCurr = peCurr->GetEdgeNext(0);
	}
      while ( peCurr != pe );
    }

  return pe->GetFacet();
}

VOID QEMesh::MergeVoronoiPolygonsHeld(QEEdge *peLeftEnd, QEEdge *&peLeft,
				      QEEdge *&peBisector,
				      QEEdge *&peRight, QEEdge *peRightEnd, FLOAT fW)
{
  //QEEdge *peLeft, *peRight, *peBisector;
  QEEdge *peLeftNext, *peRightNext;
  QEEdge *pe;
  QEVertex *pvxNew, *ppvxEndpoint[2];
  Point pt, pptResults[2];
  UINT_32 uiResults;
  //Vector v;
  FLOAT pfDistance[2];

  UINT_32 uiMergeDirection;
  Point ptMerge, ptLeft, ptRight;
  QEFacet *pfaMerge, *pfaTest, *pfaLeft, *pfaRight;
  BOOL bReplacePoint, bContinueSearch;

  //////////////////////////////////////////////////////////////////////
  // Merge

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  //fprintf(stderr, "    START: Merge Left = ( %.3f, %.3f ) Right = ( %.3f, %.3f )\n", peLeftMiddle->GetVertex()->GetPoint()(X), peLeftMiddle->GetVertex()->GetPoint()(Y), peRightMiddle->GetVertex()->GetPoint()(X), peRightMiddle->GetVertex()->GetPoint()(Y));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

  ////////////////////////////////////////
  // Merge Loop
  while ( TRUE )
    {
      uiMergeDirection = 0;
      ptMerge.Set(0.0, 0.0, INFINITY);
      pfaMerge = NULL;
      pfaLeft = NULL;
      pfaRight = NULL;

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
      fprintf(stderr, "      START: Searching for candidate point\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

      ////////////////////////////////////////////////////////////
      // START: Collapse edges from the left and right
      //   until they are in the sector bounded by the 
      //   two contour edges
      
      // Left
      ASSERT( peLeft->GetEdge()->GetFacet() != GetNullFacet() ||
	      peLeft == peLeftEnd ||
	      peLeft == peLeft->GetFacet()->GetForwardEdge() );
      while ( peLeft != peLeftEnd &&
	      peLeft != peLeft->GetFacet()->GetForwardEdge() &&
	      peLeft->GetEdgeNext(0) != peLeft->GetEdgeNext(0)->GetFacet()->GetForwardEdge() &&
	      peLeft->GetEdge()->GetVertex() != peLeft->GetFacet()->GetForwardVertex() )

	{
	  // Check if the Voronoi vertex is closer to the left or right edge
	  pt = peLeft->GetEdge()->GetVertex()->GetPoint();
	  ASSERT( pt(3) != 0.0 );
	  //pfDistance[0] = CalculateDistance(peLeft->GetFacet(), pt, &(ppvxEndpoint[0]));
	  pfDistance[0] = pt[2];
	  pt[2] = 0.0;
	  pt[3] = 1.0;
	  pfDistance[1] = CalculateDistance(peRight->GetFacet(), pt, &(ppvxEndpoint[1]));
	  if ( pfDistance[0] < pfDistance[1] )
	    {
	      break;
	    }
	  
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	  fprintf(stderr, "        Left side skipping uniformed Voronoi edge\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	  // Advance the left next pointer
	  peLeftNext = peLeft->GetEdgeNext(0);
	  if ( peLeftNext->GetEdge()->GetFacet() == GetNullFacet() &&
	       peLeftNext != peLeftEnd &&
	       peLeftNext != peLeftNext->GetFacet()->GetForwardEdge() )
	    {
	      peLeftNext = peLeftNext->GetEdgeNext(0);
	    }

	  //////////////////////////////
	  // Advance peLeft cleaning up extraneous edges
	  if ( peLeft->GetEdgeNext(0)->GetEdge()->GetFacet() == GetNullFacet() &&
	       peLeft->GetEdgeNext(0) != peLeftEnd &&
	       peLeft->GetEdgeNext(0) != peLeft->GetEdgeNext(0)->GetFacet()->GetForwardEdge() )
	    {
	      ASSERT( peLeft->GetEdgeNext(0) != peLeftEnd &&
		      peLeft->GetEdgeNext(0) != peLeft->GetEdgeNext(0)->GetFacet()->GetForwardEdge() );
	      CollapseEdge(peLeft->GetEdgeNext(0));
	    }
	  ASSERT( peLeft != peLeftEnd &&
		  peLeft != peLeft->GetFacet()->GetForwardEdge() );
	  CollapseEdge(peLeft);
	  peLeft = peLeftNext;

	  ASSERT( peLeft->GetEdge()->GetFacet() != GetNullFacet() ||
		  peLeft == peLeftEnd );
	}
      
      // Right
      ASSERT( peRight->GetEdge()->GetFacet() != GetNullFacet() ||
	      peRight == peRightEnd ||
	      peRight == peRight->GetFacet()->GetForwardEdge() );
      while ( peRight != peRightEnd &&
	      peRight != peRight->GetFacet()->GetForwardEdge() &&
	      peRight->GetEdgePrev(0) != peRight->GetEdgePrev(0)->GetFacet()->GetForwardEdge() &&
	      peRight->GetVertex() != peRight->GetFacet()->GetForwardVertex() )
	{
	  // Check if the Voronoi vertex is closer to the left or right edge
	  pt = peRight->GetVertex()->GetPoint();
	  ASSERT( pt(3) != 0.0 );
	  //pfDistance[0] = CalculateDistance(peRight->GetFacet(), pt, &(ppvxEndpoint[0]));
	  pfDistance[0] = pt[2];
	  pt[2] = 0.0;
	  pt[3] = 1.0;
	  pfDistance[1] = CalculateDistance(peLeft->GetFacet(), pt, &(ppvxEndpoint[1]));
	  if ( pfDistance[0] < pfDistance[1] )
	    {
	      break;
	    }
	  
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	  fprintf(stderr, "        Right side skipping uniformed Voronoi edge\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	  // Advance the right next pointer
	  peRightNext = peRight->GetEdgePrev(0);
	  if ( peRightNext->GetEdge()->GetFacet() == GetNullFacet() &&
	       peRightNext != peRightEnd &&
	       peRightNext != peRightNext->GetFacet()->GetForwardEdge() )
	    {
	      peRightNext = peRightNext->GetEdgePrev(0);
	    }
	  
	  //////////////////////////////
	  // Advance peRight cleaning up extraneous edges
	  if ( peRight->GetEdgePrev(0)->GetEdge()->GetFacet() == GetNullFacet() &&
	       peRight->GetEdgePrev(0) != peRightEnd &&
	       peRight->GetEdgePrev(0) != peRight->GetEdgePrev(0)->GetFacet()->GetForwardEdge() )
	    {
	      ASSERT( peRight->GetEdgePrev(0) != peRightEnd &&
		      peRight->GetEdgePrev(0) != peRight->GetEdgePrev(0)->GetFacet()->GetForwardEdge() );
	      CollapseEdge(peRight->GetEdgePrev(0)->GetEdge());
	    }
	  ASSERT( peRight != peRightEnd &&
		  peRight != peRight->GetFacet()->GetForwardEdge() );
	  CollapseEdge(peRight->GetEdge());
	  peRight = peRightNext;

	  ASSERT( peRight->GetEdge()->GetFacet() != GetNullFacet() ||
		  peRight == peRightEnd );
	}

      // END  : Collapse edges from the left and right
      //   until they are in the sector bounded by the 
      //   two contour edges
      ////////////////////////////////////////////////////////////      

      peLeftNext = peLeft;
      peRightNext = peRight;

      ////////////////////////////////////////
      // Search for the best candidate
      bContinueSearch = TRUE;
      while ( bContinueSearch )
	{
	  bContinueSearch = FALSE;
	  bReplacePoint = FALSE;

	  ////////////////////////////////////////
	  // Search for Left candidate
	      
	  uiResults = 0;

	  // Do not go past a contour edge
	  if ( peLeftNext != peLeftEnd &&
	       peLeftNext != peLeftNext->GetFacet()->GetForwardEdge() &&
	       (peLeftNext->GetFacet()->GetForwardVertex() == NULL ||
		peLeftNext->GetVertex() != peLeftNext->GetFacet()->GetForwardVertex()) &&
	       peLeftNext != peBisector->GetEdge() &&
	       peLeftNext->GetEdgeNext(0) != peBisector->GetEdge() )
	    {
	      ASSERT( peLeftNext->GetEdge()->GetFacet() != GetNullFacet() );
	      if ( Counterclockwise(peLeftNext->GetEdge()->GetFacet(), TRUE,
				    peLeft->GetFacet(), TRUE,
				    peRight->GetFacet(), TRUE) )
		{
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
		  fprintf(stderr, "        Calculating left candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
		  CalculateVoronoiPoint(pptResults, uiResults, 
					peLeftNext->GetEdge()->GetFacet(), TRUE,
					peLeft->GetFacet(), TRUE,
					peRight->GetFacet(), TRUE, 
					TRUE);
		  if ( uiResults > 0 ) 
		    {
		      //////////////////////////////
		      // Check if this candidate is closer than the current right point
		      bReplacePoint = FALSE;
		      if ( peLeft == peLeftNext )
			{
			  //////////////////////////////
			  // No left candidate yet
			  
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			  fprintf(stderr, "        Found first left candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			  bReplacePoint = TRUE;
			}
		      else
			{
			  //////////////////////////////
			  // Do an InCircle test with the current left point
			  
			  ASSERT( peLeft->GetFacet() != GetNullFacet() &&
				  peLeft->GetFacet()->GetForwardEdge() != peLeft );
			  pfaTest = peLeft->GetEdge()->GetFacet();

			  bReplacePoint = FALSE;

			  //pfDistance[0] = CalculateDistance(peLeftNext->GetEdge()->GetFacet(), ptLeft, &(ppvxEndpoint[0]));
			  pfDistance[1] = CalculateDistance(pfaTest, pptResults[0], &(ppvxEndpoint[1]));

			  if ( pptResults[0](Z) <= ptLeft(Z) &&
			       ( pptResults[0](Z) <= pfDistance[1] ||
				 ( ppvxEndpoint[1] != NULL &&
				   ( peLeftNext->GetEdge()->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
				     peLeft->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
				     peRight->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ) ) ) )
			    {
			      //////////////////////////////
			      // Replacing left candidate

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			      fprintf(stderr, "        Updated left candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			      bReplacePoint = TRUE;
			    }
			}

		      if ( bReplacePoint ) 
			{
			  pfaLeft = peLeftNext->GetEdge()->GetFacet();
			  ptLeft = pptResults[0];

			  //////////////////////////////
			  // Check if this candidate is closer than the current merge point

			  pfaTest = pfaMerge;
			  if ( pfaTest == NULL &&
			       peRight != peRightEnd &&
			       peRight->GetFacet()->GetForwardEdge() != peRight )
			    {
			      pfaTest = peRight->GetEdge()->GetFacet();
			      ASSERT( pfaTest != GetNullFacet() );
			    }

			  bReplacePoint = FALSE;
			  if ( pfaTest == NULL )
			    {
			      bReplacePoint = TRUE;
			    }
			  else
			    {
			      //pfDistance[0] = CalculateDistance(peLeftNext->GetEdge()->GetFacet(), ptMerge, &(ppvxEndpoint[0]));
			      pfDistance[1] = CalculateDistance(pfaTest, pptResults[0], &(ppvxEndpoint[1]));

			      if ( pptResults[0](Z) <= ptMerge(Z) &&
				   ( pptResults[0](Z) <= pfDistance[1] ||
				     ( ppvxEndpoint[1] != NULL &&
				       ( peLeftNext->GetEdge()->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
					 peLeft->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
					 peRight->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ) ) ) )
				{
				  bReplacePoint = TRUE;
				}
			    }

			  if ( bReplacePoint )
			    {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			      fprintf(stderr, "        Replacing merge point with left candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			      bContinueSearch = TRUE;
			      uiMergeDirection = 1;
			      ptMerge = pptResults[0];
			      pfaMerge = peLeftNext->GetEdge()->GetFacet();
			    }

			  if ( peLeft != peLeftNext )
			    {
			      //////////////////////////////
			      // Advance peLeft cleaning up extraneous edges
			      if ( peLeft->GetEdgeNext(0)->GetEdge()->GetFacet() == GetNullFacet() &&
				   peLeft->GetEdgeNext(0) != peLeftEnd )
				{
				  ASSERT( peLeft->GetEdgeNext(0) != peLeftEnd &&
					  peLeft->GetEdgeNext(0) != peLeft->GetEdgeNext(0)->GetFacet()->GetForwardEdge() );
				  CollapseEdge(peLeft->GetEdgeNext(0));
				}
			      ASSERT( peLeft != peLeftEnd &&
				      peLeft != peLeft->GetFacet()->GetForwardEdge() );
			      CollapseEdge(peLeft);
			      peLeft = peLeftNext;
			    }

			  // Advance the left next pointer
			  peLeftNext = peLeft->GetEdgeNext(0);
			  if ( peLeftNext->GetEdge()->GetFacet() == GetNullFacet() &&
			       peLeftNext != peLeftEnd &&
			       peLeftNext != peLeftNext->GetFacet()->GetForwardEdge() &&
			       peLeftNext->GetEdge()->GetVertex() != peLeftNext->GetFacet()->GetForwardVertex() )
			    {
			      peLeftNext = peLeftNext->GetEdgeNext(0);
			    }
			}
		    }
		}
	    }

	  ////////////////////////////////////////
	  // Search for Right candidate
	      
	  uiResults = 0;

	  // Do not go past a contour edge
	  if ( peRightNext != peRightEnd &&
	       peRightNext != peRightNext->GetFacet()->GetForwardEdge() &&
	       (peRightNext->GetFacet()->GetForwardVertex() == NULL ||
		peRightNext->GetEdge()->GetVertex() != peRightNext->GetFacet()->GetForwardVertex()) &&
	       peRightNext != peBisector &&
	       peRightNext->GetEdgePrev(0) != peBisector )
	    {
	      ASSERT( peRightNext->GetEdge()->GetFacet() != GetNullFacet() );
	      if ( Counterclockwise(peLeft->GetFacet(), TRUE,
				    peRight->GetFacet(), TRUE,
				    peRightNext->GetEdge()->GetFacet(), TRUE) )
		{
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
		  fprintf(stderr, "        Calculating right candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
		  CalculateVoronoiPoint(pptResults, uiResults, 
					peLeft->GetFacet(), TRUE, 
					peRight->GetFacet(), TRUE,
					peRightNext->GetEdge()->GetFacet(), TRUE,
					TRUE);
		  if ( uiResults > 0 )
		    {
		      //////////////////////////////
		      // Check if this candidate is closer than the current right point
		      bReplacePoint = FALSE;
		      if ( peRight == peRightNext )
			{
			  //////////////////////////////
			  // No right candidate yet
			  
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			  fprintf(stderr, "        Found first right candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			  bReplacePoint = TRUE;
			}
		      else
			{
			  //////////////////////////////
			  // Do an InCircle test with the curren right point
			  
			  ASSERT( peRight->GetFacet() != GetNullFacet() &&
				  peRight->GetFacet()->GetForwardEdge() != peRight );
			  pfaTest = peRight->GetEdge()->GetFacet();

			  bReplacePoint = FALSE;

			  //pfDistance[0] = CalculateDistance(peRightNext->GetEdge()->GetFacet(), ptRight, &(ppvxEndpoint[0]));
			  pfDistance[1] = CalculateDistance(pfaTest, pptResults[0], &(ppvxEndpoint[1]));

			  if ( pptResults[0](Z) <= ptRight(Z) &&
			       ( pptResults[0](Z) <= pfDistance[1] ||
				 ( ppvxEndpoint[1] != NULL &&
				   ( peLeft->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
				     peRight->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
				     peRightNext->GetEdge()->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ) ) ) )
			    {
			      //////////////////////////////
			      // Replacing right candidate

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			      fprintf(stderr, "        Updated right candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			      bReplacePoint = TRUE;
			    }
			}

		      if ( bReplacePoint ) 
			{
			  pfaRight = peRightNext->GetEdge()->GetFacet();
			  ptRight = pptResults[0];

			  //////////////////////////////
			  // Check if this candidate is closer than the current merge point

			  pfaTest = pfaMerge;
			  if ( pfaTest == NULL &&
			       peLeft != peLeftEnd &&
			       peLeft->GetFacet()->GetForwardEdge() != peLeft )
			    {
			      pfaTest = peLeft->GetEdge()->GetFacet();
			      ASSERT( pfaTest != GetNullFacet() );
			    }

			  bReplacePoint = FALSE;
			  if ( pfaTest == NULL )
			    {
			      bReplacePoint = TRUE;
			    }
			  else
			    {
			      //pfDistance[0] = CalculateDistance(peRightNext->GetEdge()->GetFacet(), ptMerge, &(ppvxEndpoint[0]));
			      pfDistance[1] = CalculateDistance(pfaTest, pptResults[0], &(ppvxEndpoint[1]));

			      if ( pptResults[0](Z) <= ptMerge(Z) &&
				   ( pptResults[0](Z) <= pfDistance[1] ||
				     ( ppvxEndpoint[1] != NULL &&
				       ( peLeft->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
					 peRight->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
					 peRightNext->GetEdge()->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ) ) ) )
			    {
				  bReplacePoint = TRUE;
				}
			    }

			  if ( bReplacePoint )
			    {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			      fprintf(stderr, "        Replacing merge point with right candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			      bContinueSearch = TRUE;
			      uiMergeDirection = 2;
			      ptMerge = pptResults[0];
			      pfaMerge = peRightNext->GetEdge()->GetFacet();
			    }

			  if ( peRight != peRightNext )
			    {
			      //////////////////////////////
			      // Advance peRight cleaning up extraneous edges
			      if ( peRight->GetEdgePrev(0)->GetEdge()->GetFacet() == GetNullFacet() &&
				   peRight->GetEdgePrev(0) != peRightEnd )
				{
				  ASSERT( peRight->GetEdgePrev(0) != peRightEnd &&
					  peRight->GetEdgePrev(0) != peRight->GetEdgePrev(0)->GetFacet()->GetForwardEdge() );
				  CollapseEdge(peRight->GetEdgePrev(0)->GetEdge());
				}
			      ASSERT( peRight != peRightEnd &&
				      peRight != peRight->GetFacet()->GetForwardEdge() );
			      CollapseEdge(peRight->GetEdge());
			      peRight = peRightNext;
			    }

			  // Advance the right next pointer
			  peRightNext = peRight->GetEdgePrev(0);
			  if ( peRightNext->GetEdge()->GetFacet() == GetNullFacet() &&
			       peRightNext != peRightEnd &&
			       peRightNext != peRightNext->GetFacet()->GetForwardEdge() &&
			       peRightNext->GetVertex() != peRightNext->GetFacet()->GetForwardVertex() )
			    {
			      peRightNext = peRightNext->GetEdgePrev(0);
			    }
			}
		    }
		}

	      //////////////////////////////////////////////////
	      // Because of numeric issues, the left and right might
	      // get equally good answers which will confuse things
	      // and cancel out.  We need to override that possibility
	      if ( pfaMerge == NULL &&
		   ( pfaLeft != NULL || pfaRight != NULL ) )
		{
		  if ( pfaLeft != NULL )
		    {
		      pfDistance[0] = CalculateDistance(peRight->GetEdge()->GetFacet(), ptLeft, &(ppvxEndpoint[0]));
		    }
		  if ( pfaRight != NULL )
		    {
		      pfDistance[1] = CalculateDistance(peLeft->GetEdge()->GetFacet(), ptRight, &(ppvxEndpoint[1]));
		    }

		  if ( ( pfaRight == NULL ||
			 ( pfaRight != NULL && pfaLeft != NULL &&
			   !( ptRight(Z) <= pfDistance[1] + s_cfDistanceEpsilon ) ) ) &&
		       ptLeft(Z) <= pfDistance[0] + s_cfDistanceEpsilon )
		    {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
		      fprintf(stderr, "        Special replace merge point with left candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
		      bContinueSearch = TRUE;
		      uiMergeDirection = 1;
		      ptMerge = ptLeft;
		      pfaMerge = pfaLeft;
		    }
		  else if ( ( pfaLeft == NULL ||
			 ( pfaRight != NULL && pfaLeft != NULL &&
			   !( ptLeft(Z) <= pfDistance[0] + s_cfDistanceEpsilon ) ) ) &&
		       ptRight(Z) <= pfDistance[1] + s_cfDistanceEpsilon )
		    {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
		      fprintf(stderr, "        Special replace merge point with right candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
		      bContinueSearch = TRUE;
		      uiMergeDirection = 2;
		      ptMerge = ptRight;
		      pfaMerge = pfaRight;
		    }
		  else if ( pfaRight != NULL && pfaLeft != NULL &&
			    ptLeft(Z) <= pfDistance[0] + s_cfDistanceEpsilon &&
			    ptRight(Z) <= pfDistance[1] + s_cfDistanceEpsilon )
		    {
		      if ( ptLeft(Z) < ptRight(Z) )
			{
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			  fprintf(stderr, "        Special replace merge point with left candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			  bContinueSearch = TRUE;
			  uiMergeDirection = 1;
			  ptMerge = ptLeft;
			  pfaMerge = pfaLeft;
			}
		      else
			{
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			  fprintf(stderr, "        Special replace merge point with right candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			  bContinueSearch = TRUE;
			  uiMergeDirection = 2;
			  ptMerge = ptRight;
			  pfaMerge = pfaRight;
			}
		    }
		}
	    }
	}

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
      fprintf(stderr, "      END  : Searching for candidate point\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

      //////////////////////////////////////////////////
      // Check if we have a valid candidate
      if ( uiMergeDirection == 0 ) 
	{
	  break;
	} 
      else if ( uiMergeDirection == 1 ) 
	{
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	  fprintf(stderr, "      Attaching to left ( %.3f, %.3f, %.3f )\n", ptMerge(X), ptMerge(Y), ptMerge(Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	  //////////////////////////////////////////////////
	  // Attach the bisector to the left side

	  ptMerge[3] = fW;

	  peLeftNext = peLeft->GetEdge()->GetEdgeNext(0);
	  pe = peBisector;

	  pvxNew = CreateVertex(ptMerge);
	  peBisector = CreateEdge();

	  peLeft->GetEdgePrev(1)->SetVertex(peLeft->GetEdgePrev(1)->GetVertex());
	  Splice(1, peLeft, peLeft->GetEdgePrev(1));

	  peLeft->SetVertex(pvxNew);
	  pe->SetVertex(pvxNew);
	  peBisector->GetEdge()->SetVertex(pvxNew);

	  Splice(1, pe, peBisector->GetEdge());
	  Splice(1, peLeft, peBisector->GetEdge());

	  peLeft = peLeftNext;

	  peBisector->SetFacet(peRight->GetFacet());
	  peBisector->GetEdge()->SetFacet(peLeft->GetFacet());

	  if ( peLeft->GetEdge()->GetFacet() == GetNullFacet() &&
	       peLeft != peLeftEnd &&
	       peLeft->GetFacet()->GetForwardEdge() != peLeft && 
	       peLeft->GetFacet()->GetForwardVertex() != peLeft->GetEdge()->GetVertex() )
	    {
	      peLeftNext = peLeft->GetEdgeNext(0);
	      CollapseEdge(peLeft);
	      peLeft = peLeftNext;
	    }
	}
      else
	{
	  //////////////////////////////////////////////////
	  // Attach the bisector to the right side

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	  fprintf(stderr, "      Attaching to right ( %.3f, %.3f, %.3f )\n", ptMerge(X), ptMerge(Y), ptMerge(Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	  ptMerge[3] = fW;

	  peRightNext = peRight->GetEdge()->GetEdgePrev(0);
	  pe = peBisector;

	  pvxNew = CreateVertex(ptMerge);
	  peBisector = CreateEdge();

	  peRight->GetEdge()->GetEdgePrev(1)->SetVertex(peRight->GetEdge()->GetEdgePrev(1)->GetVertex());
	  Splice(1, peRight->GetEdge(), peRight->GetEdge()->GetEdgePrev(1));

	  peRight->GetEdge()->SetVertex(pvxNew);
	  pe->SetVertex(pvxNew);
	  peBisector->GetEdge()->SetVertex(pvxNew);

	  Splice(1, pe, peBisector->GetEdge());
	  Splice(1, peRight->GetEdge(), pe);

	  peRight = peRightNext;

	  peBisector->SetFacet(peRight->GetFacet());
	  peBisector->GetEdge()->SetFacet(peLeft->GetFacet());

	  if ( peRight->GetEdge()->GetFacet() == GetNullFacet() &&
	       peRight != peRightEnd &&
	       peRight->GetFacet()->GetForwardEdge() != peRight &&
	       peRight->GetFacet()->GetForwardVertex() != peRight->GetVertex() )
	    {
	      peRightNext = peRight->GetEdgePrev(0);
	      CollapseEdge(peRight->GetEdge());
	      peRight = peRightNext;
	    }
	}

      peLeftNext = peLeft;
      peRightNext = peRight;
    }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  //fprintf(stderr, "    END  : Merge Left = ( %.3f, %.3f ) Right = ( %.3f, %.3f )\n", peLeftMiddle->GetVertex()->GetPoint()(X), peLeftMiddle->GetVertex()->GetPoint()(Y), peRightMiddle->GetVertex()->GetPoint()(X), peRightMiddle->GetVertex()->GetPoint()(Y));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
}

//////////////////////////////////////////////////
// Z Monotone the edges
VOID QEMesh::ZMonotoneVoronoiPolygon()
{
  CObjectListIter<QEEdge> iterQEEdge;
  CObjectList<QEEdge> listQEEdge;
  QEEdge *pe, *peCurr;
  QEEdge *peSite, *peNew, *peNext;
  QEVertex *pvxSite, *ppvxSite[2], *pvxNew, *pvxEndpoint;
  QEFacet *ppfa[4], *pfaNew;
  Point pt, ppt[2];
  Vector v;
  FLOAT fDot, fDistance, fDist, fMagnitude;
  BOOL bSplit;
  UINT_32 i, uiResults;

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "START: ZMonotoneVoronoiPolygon\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

  {
    QEMeshIter iter;
    QEVertex *pvx;
    Point pt;

    //////////////////////////////////////////////////
    // Make the distance values on the exterior negative
    iter.Init(this);
    for ( pvx = iter.PeekFirstVertex();
	  pvx != NULL;
	  pvx = iter.PeekNextVertex() )
      {
	pt = pvx->GetPoint();
	if ( pt[3] == -1.0 )
	  {
	    pt[2] = -pt[2];
	    pt[3] = 1.0;
	    pvx->SetPoint(pt);
	  }
      }
  }

  iterQEEdge.Init(m_plistQEEdge);
  
  for ( pe = iterQEEdge.PeekFirst();
	pe != NULL;
	pe = peNext )
    {
      peNext = iterQEEdge.PeekNext();
      if ( peNext == pe->GetEdge() )
	{
	  peNext = iterQEEdge.PeekNext();	    
	}
      
      if ( pe->GetFacet() == GetNullFacet() )
	{
	  pe = pe->GetEdge();
	}
      ASSERT( pe->GetFacet() != GetNullFacet() );
      
      //////////////////////////////////////////////////
      // Do not touch edges which touch the input contour
      if ( pe == pe->GetFacet()->GetForwardEdge() ||
	   pe->GetEdgePrev(0) == pe->GetFacet()->GetForwardEdge() ||
	   pe->GetEdgeNext(0) == pe->GetFacet()->GetForwardEdge() ||
	   pe->GetVertex() == pe->GetFacet()->GetForwardVertex() ||
	   pe->GetEdge()->GetVertex() == pe->GetFacet()->GetForwardVertex() ||
	   ( pe->GetEdge()->GetFacet() != GetNullFacet() &&
	     ( pe->GetEdge() == pe->GetEdge()->GetFacet()->GetForwardEdge() ||
	       pe->GetEdge()->GetEdgePrev(0) == pe->GetEdge()->GetFacet()->GetForwardEdge() ||
	       pe->GetEdge()->GetEdgeNext(0) == pe->GetEdge()->GetFacet()->GetForwardEdge() ||
	       pe->GetVertex() == pe->GetEdge()->GetFacet()->GetForwardVertex() ||
	       pe->GetEdge()->GetVertex() == pe->GetEdge()->GetFacet()->GetForwardVertex() ) ) )
	{
	  continue;
	}
      
      //////////////////////////////////////////////////
      // Remove epsilon length edges
      ppt[0] = pe->GetVertex()->GetPoint();
      ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
      if ( ppt[0](3) != 0.0 && ppt[1](3) != 0.0 )
	{
	  v = ppt[1] - ppt[0];
	  v[2] = 0.0;
	  v[3] = 0.0;
	  fMagnitude = v.Magnitude();
	  if ( fMagnitude <= s_cfEpsilonVoronoiZMonotone )
	    {
	      //fprintf(stderr, "  Collapsing Edge: %f: \n    %p ( %f, %f ) <==> %p ( %f, %f )\n", fMagnitude, pe->GetVertex(), ppt[0](0), ppt[0](1), pe->GetEdge()->GetVertex(), ppt[1](0), ppt[1](1));
	      ppfa[0] = pe->GetFacet();
	      ppfa[1] = pe->GetEdge()->GetFacet();
	      ASSERT( ppfa[0] != ppfa[1] );
	      CollapseEdge(pe);
	      for ( i = 0; i < 2; i++ )
		{
		  if ( ppfa[i] != GetNullFacet() )
		    {
		      pe = ppfa[i]->GetEdge();
		      if ( pe->GetEdgeNext(0) == pe->GetEdgePrev(0) )
			{
			  ASSERT( pe != pe->GetEdgeNext(0) );
			  pe->GetEdgeNext(0)->SetFacet(pe->GetEdge()->GetFacet());
			  pe->GetEdgePrev(1)->SetVertex(pe->GetEdgePrev(1)->GetVertex());
			  pe->GetEdge()->GetEdgePrev(1)->SetVertex(pe->GetEdge()->GetEdgePrev(1)->GetVertex());
			  Splice(1, pe->GetEdgePrev(1), pe);
			  Splice(1, pe->GetEdge()->GetEdgePrev(1), pe->GetEdge());
			  
			  DestroyFacet(ppfa[i]);
			  while ( peNext != NULL && (peNext == pe || peNext == pe->GetEdge()) )
			    {
			      peNext = iterQEEdge.PeekNext();	    
			    }
			  DestroyEdge(pe->GetEdge());
			  DestroyEdge(pe);
			}
		    }
		}
	    }
	}
    }
  
  listQEEdge.Init();
  listQEEdge.Splice(m_plistQEEdge);
  
  for ( pe = listQEEdge.RemoveFirst();
	pe != NULL;
	pe = listQEEdge.RemoveFirst() )
    {
      m_plistQEEdge->InsertLast(pe);
      listQEEdge.Remove(pe->GetEdge());
      m_plistQEEdge->InsertLast(pe->GetEdge());
      
      if ( pe->GetFacet() == GetNullFacet() )
	{
	  pe = pe->GetEdge();
	}
      ASSERT( pe->GetFacet() != GetNullFacet() );
      
      //////////////////////////////////////////////////
      // Split edges at minima
      if ( pe->GetFacet() != m_pfaNULL &&
	   pe->GetEdge()->GetFacet() != m_pfaNULL )
	{
	  bSplit = FALSE;
	  pvxSite = NULL;
	  peSite = NULL;
	  ppvxSite[0] = NULL;
	  ppvxSite[1] = NULL;
	  
	  if ( ( pe->GetFacet()->GetForwardEdge() != NULL &&
		 pe->GetEdge()->GetFacet()->GetForwardVertex() != NULL ) ||
	       ( pe->GetFacet()->GetForwardVertex() != NULL &&
		 pe->GetEdge()->GetFacet()->GetForwardEdge() != NULL ) )
	    {
	      //////////////////////////////////////////////////
	      // Split Parabolic Edge created by an edge and a vertex
	      if ( pe->GetFacet()->GetForwardEdge() != NULL &&
		   pe->GetEdge()->GetFacet()->GetForwardVertex() != NULL )
		{
		  pe = pe->GetEdge();
		}
	      else
		{
		  ASSERT( pe->GetFacet()->GetForwardVertex() != NULL &&
			  pe->GetEdge()->GetFacet()->GetForwardEdge() != NULL );
		}
	      pvxSite = pe->GetFacet()->GetForwardVertex();
	      peSite = pe->GetEdge()->GetFacet()->GetForwardEdge();

	      if ( peSite->GetVertex() != pvxSite &&
		   peSite->GetEdge()->GetVertex() != pvxSite )
		{
		  CalculatePointProjectionOnLineSegment(pt, uiResults,
							pvxSite->GetPoint(), 
							peSite->GetVertex()->GetPoint(), 
							peSite->GetEdge()->GetVertex()->GetPoint(),
							TRUE);
		  if ( uiResults > 0 )
		    {
		      
		      fDistance = pt[2]/2.0;
		      pt = (pvxSite->GetPoint() + pt) / 2.0;
		      pt[2] = 0.0;
		      pt[3] = 1.0;
		      
		      ppfa[0] = pe->GetEdgePrev(1)->GetEdge()->GetFacet();
		      ppfa[1] = pe->GetEdgeNext(1)->GetFacet();
		      ppfa[2] = pe->GetEdge()->GetEdgePrev(1)->GetEdge()->GetFacet();
		      ppfa[3] = pe->GetEdge()->GetEdgeNext(1)->GetFacet();

		      bSplit = TRUE;
		      for ( i = 0; i < 4; i++ )
			{
			  if ( ppfa[i] != m_pfaNULL &&
			       ppfa[i] != pe->GetFacet() &&
			       ppfa[i] != pe->GetEdge()->GetFacet() )
			    {
			      fDist = CalculateDistance(ppfa[i], pt, &pvxEndpoint);

			      if ( ( pvxEndpoint == NULL || 
				     ( pvxEndpoint != NULL && pvxEndpoint != pvxSite ) ) &&
				   fDist < fDistance )
				{
				  bSplit = FALSE;
				  break;
				}
			    }
			}

		      v = pt - pe->GetVertex()->GetPoint();
		      v[2] = 0.0;
		      v[3] = 0.0;
		      fDot = v.Magnitude();
		      if ( fDot < s_cfEpsilonVoronoiZMonotone )
			{
			  bSplit = FALSE;
			}

		      v = pt - pe->GetEdge()->GetVertex()->GetPoint();
		      v[2] = 0.0;
		      v[3] = 0.0;
		      fDot = v.Magnitude();
		      if ( fDot < s_cfEpsilonVoronoiZMonotone )
			{
			  bSplit = FALSE;
			}

		      pt[2] = fDistance;

		      if ( pe->GetVertex()->GetPoint()(2) < 0.0 || 
			   pe->GetEdge()->GetVertex()->GetPoint()(2) < 0.0 )
			{
			  pt[2] = -pt[2];
			} 
		    }

		  if ( bSplit )
		    {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
		      fprintf(stderr, "Split Parabolic Edge VE: ( %.3f, %.3f ) && ( %.3f, %.3f ) <=> ( %.3f, %.3f ) at ( %.3f, %.3f, %.3f )\n",
			      pvxSite->GetPoint()(X),
			      pvxSite->GetPoint()(Y),
			      peSite->GetVertex()->GetPoint()(X),
			      peSite->GetVertex()->GetPoint()(Y),
			      peSite->GetEdge()->GetVertex()->GetPoint()(X),
			      peSite->GetEdge()->GetVertex()->GetPoint()(Y),
			      pt(X),
			      pt(Y),
			      pt(Z));
		      fprintf(stderr, "      Parabolic Edge endpoints: ( %.3f, %.3f, %.3f ) <==> ( %.3f, %.3f, %.3f )\n",
			      pe->GetVertex()->GetPoint()(X),
			      pe->GetVertex()->GetPoint()(Y),
			      pe->GetVertex()->GetPoint()(Z),
			      pe->GetEdge()->GetVertex()->GetPoint()(X),
			      pe->GetEdge()->GetVertex()->GetPoint()(Y),
			      pe->GetEdge()->GetVertex()->GetPoint()(Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

		      //////////////////////////////////////////////////
		      // Split the parabolic edge in two
		      pvxNew = CreateVertex(pt);
		      peNew = CreateEdge();
		
		      peNew->SetFacet(pe->GetFacet());
		      peNew->GetEdge()->SetFacet(pe->GetEdge()->GetFacet());
		      peNew->SetVertex(pe->GetVertex());
		      peNew->GetEdge()->SetVertex(pvxNew);
		      pe->SetVertex(pvxNew);
		
		      Splice(1, pe->GetEdgePrev(1), peNew);
		      Splice(1, pe->GetEdgePrev(1), pe);
		      Splice(1, pe, peNew->GetEdge());

		      //////////////////////////////////////////////////
		      // Split the cone face in two
		      pfaNew = CreateFacet();
		      pfaNew->SetForwardVertex(pvxSite);

		      pe->SetFacet(pfaNew);
		      for ( peCurr = pe->GetEdgeNext(0);
			    peCurr->GetVertex() != pvxSite;
			    peCurr = peCurr->GetEdgeNext(0) )
			{
			  ASSERT( peCurr != pe );
			  peCurr->SetFacet(pfaNew);
			}

		      peNew = CreateEdge();
		      peNew->SetVertex(pe->GetVertex());
		      peNew->SetFacet(peCurr->GetFacet());
		      peNew->GetEdge()->SetVertex(peCurr->GetVertex());
		      peNew->GetEdge()->SetFacet(pfaNew);

		      Splice(1, peCurr, peNew->GetEdge());
		      Splice(1, pe, peNew);
		    }
		}
	    }
	  else if ( pe->GetFacet()->GetForwardVertex() != NULL &&
		    pe->GetEdge()->GetFacet()->GetForwardVertex() != NULL )
	    {
	      //////////////////////////////////////////////////
	      // Split Hyperbolic Edge created by two vertices
	      ppvxSite[0] = pe->GetFacet()->GetForwardVertex();
	      ppvxSite[1] = pe->GetEdge()->GetFacet()->GetForwardVertex();

	      pt = ( ppvxSite[1]->GetPoint() + ppvxSite[0]->GetPoint() ) / 2.0;
	      pt[2] = 0.0;
	      pt[3] = 1.0;

	      v = ppvxSite[1]->GetPoint() - ppvxSite[0]->GetPoint();
	      fDot = v.Dot(v);
	      fDistance = fDot/4.0;
		
	      ppfa[0] = pe->GetEdgePrev(1)->GetEdge()->GetFacet();
	      ppfa[1] = pe->GetEdgeNext(1)->GetFacet();
	      ppfa[2] = pe->GetEdge()->GetEdgePrev(1)->GetEdge()->GetFacet();
	      ppfa[3] = pe->GetEdge()->GetEdgeNext(1)->GetFacet();
		
	      bSplit = TRUE;
	      for ( i = 0; i < 4; i++ )
		{
		  if ( ppfa[i] != m_pfaNULL &&
		       ppfa[i] != pe->GetFacet() &&
		       ppfa[i] != pe->GetEdge()->GetFacet() )
		    {
		      fDist = CalculateDistance(ppfa[i], pt, &pvxEndpoint);
		      fDist *= fDist;

		      if ( ( pvxEndpoint == NULL || 
			     ( pvxEndpoint != NULL && 
			       pvxEndpoint != ppvxSite[0] && 
			       pvxEndpoint != ppvxSite[1] ) ) &&
			   fDist < fDistance )
			{
			  bSplit = FALSE;
			  break;
			}
		    }
		}

	      v = pt - pe->GetVertex()->GetPoint();
	      v[2] = 0.0;
	      v[3] = 0.0;
	      fDot = v.Magnitude();
	      if ( fDot < s_cfEpsilonVoronoiZMonotone )
		{
		  bSplit = FALSE;
		}
		
	      v = pt - pe->GetEdge()->GetVertex()->GetPoint();
	      v[2] = 0.0;
	      v[3] = 0.0;
	      fDot = v.Magnitude();
	      if ( fDot < s_cfEpsilonVoronoiZMonotone )
		{
		  bSplit = FALSE;
		}

	      pt[2] = sqrt(fDistance);
	      if ( pe->GetVertex()->GetPoint()(2) < 0.0 || 
		   pe->GetEdge()->GetVertex()->GetPoint()(2) < 0.0 )
		{
		  pt[2] = -pt[2];
		} 

	      if ( bSplit )
		{
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
		  fprintf(stderr, "Split Hyperbolic Edge VV: ( %.3f, %.3f ) && ( %.3f, %.3f ) at ( %.3f, %.3f, %.3f )\n",
			  ppvxSite[0]->GetPoint()(X),
			  ppvxSite[0]->GetPoint()(Y),
			  ppvxSite[1]->GetPoint()(X),
			  ppvxSite[1]->GetPoint()(Y),
			  pt(X),
			  pt(Y),
			  pt(Z));
		  fprintf(stderr, "      Hyperbolic Edge endpoints: ( %.3f, %.3f, %.3f ) <==> ( %.3f, %.3f, %.3f )\n",
			  pe->GetVertex()->GetPoint()(X),
			  pe->GetVertex()->GetPoint()(Y),
			  pe->GetVertex()->GetPoint()(Z),
			  pe->GetEdge()->GetVertex()->GetPoint()(X),
			  pe->GetEdge()->GetVertex()->GetPoint()(Y),
			  pe->GetEdge()->GetVertex()->GetPoint()(Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

		  //////////////////////////////////////////////////
		  // Split the parabolic edge in two
		  pvxNew = CreateVertex(pt);
		  peNew = CreateEdge();
		
		  peNew->SetFacet(pe->GetFacet());
		  peNew->GetEdge()->SetFacet(pe->GetEdge()->GetFacet());
		  peNew->SetVertex(pe->GetVertex());
		  peNew->GetEdge()->SetVertex(pvxNew);
		  pe->SetVertex(pvxNew);
		
		  Splice(1, pe->GetEdgePrev(1), peNew);
		  Splice(1, pe->GetEdgePrev(1), pe);
		  Splice(1, pe, peNew->GetEdge());

		  for ( i = 0; i < 2; i++ )
		    {
		      //////////////////////////////////////////////////
		      // Split the cone face in two
		      pfaNew = CreateFacet();
		      pfaNew->SetForwardVertex(ppvxSite[i]);
			
		      pe->SetFacet(pfaNew);
		      for ( peCurr = pe->GetEdgeNext(0);
			    peCurr->GetVertex() != ppvxSite[i];
			    peCurr = peCurr->GetEdgeNext(0) )
			{
			  ASSERT( peCurr != pe );
			  peCurr->SetFacet(pfaNew);
			}
			
		      peNew = CreateEdge();
		      peNew->SetVertex(pe->GetVertex());
		      peNew->SetFacet(peCurr->GetFacet());
		      peNew->GetEdge()->SetVertex(peCurr->GetVertex());
		      peNew->GetEdge()->SetFacet(pfaNew);

		      Splice(1, peCurr, peNew->GetEdge());
		      Splice(1, pe, peNew);

		      pe = pe->GetEdge()->GetEdgeNext(0);
		    }
		}
	    }
	}
    }
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "END  : ZMonotoneVoronoiPolygon\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
}
 
VOID QEMesh::CreateVoronoiPolygonHeld(UINT_32 uiSize, QEEdge **ppeContourLeft, 
				      QEEdge **ppeContourRight, FLOAT fW,
				      BOOL bConvex, BOOL bInside)
{
  QEEdge *peLeftEnd, *peLeftMiddle, *peRightEnd, *peRightMiddle;
  UINT_32 uiHalf;
  BOOL bMergeSingleSegment, bPossiblyMergeTwice;

  if ( uiSize == 1 )
    {
      QEEdge *pe, *peNext, *peCurr, *peNew, *peStart;
      QEVertex *pvxNew;
      QEFacet *pfaNew;
      Point pt;
      Vector v;

      bPossiblyMergeTwice = FALSE;

      if ( ppeContourLeft[0] == ppeContourRight[0] )
	{
	  pe = ppeContourLeft[0];

	  peNew = CreateEdge();
	  
	  pfaNew = CreateFacet();
	  pfaNew->SetForwardEdge(pe);

	  peNew->SetVertex(pe->GetEdge()->GetVertex());
	  peNew->GetEdge()->SetVertex(pe->GetVertex());
	  peNew->SetFacet(pfaNew);
	  pe->SetFacet(pfaNew);

	  Splice(1, pe, peNew->GetEdge());
	  Splice(1, pe->GetEdge()->GetEdgePrev(1), peNew);

	  return;
	}
      else if ( ppeContourLeft[0]->GetEdgePrev(0) == ppeContourRight[0] && bConvex )
	{
	  ////////////////////////////////////////////////////////////
	  // This constructs the complete outer Voronoi diagram of a convex contour 
	  ASSERT( !bInside );
	  pe = ppeContourLeft[0];

	  //////////////////////////////////////////////////
	  // Special Case the first edge
	  peNext = pe->GetEdgeNext(0);

	  v = pe->GetEdge()->GetVertex()->GetPoint() - pe->GetVertex()->GetPoint();
	  v.Set(-v(1), v(0), 0.0);
	  v.Normalize();
	  v[2] = fW;
	  pt = v;
	  pt[3] = 0.0;

	  peCurr = pe;

	  pfaNew = CreateFacet();
	  pfaNew->SetForwardEdge(pe);
	  pe->SetFacet(pfaNew);

	  pvxNew = CreateVertex(pt);
	  peNew = CreateEdge();
	  peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	  peNew->GetEdge()->SetVertex(pvxNew);
	  peNew->SetFacet(pfaNew);
	  Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);

	  peCurr = peNew;

	  pvxNew = CreateVertex(pt);
	  peNew = CreateEdge();
	  peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	  peNew->GetEdge()->SetVertex(pvxNew);
	  peNew->SetFacet(pfaNew);
	  Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);

	  peCurr = peNew;

	  peNew = CreateEdge();
	  peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	  peNew->GetEdge()->SetVertex(pe->GetVertex());
	  peNew->SetFacet(pfaNew);
	  Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);
	  Splice(1, pe, peNew->GetEdge());

	  peStart = peNew->GetEdge();

	  for ( pe = peNext; pe != peStart; pe = peNext )
	    {
	      peNext = pe->GetEdgeNext(0);

	      //////////////////////////////////////////////////
	      // The normal case
	      v = pe->GetEdge()->GetVertex()->GetPoint() - pe->GetVertex()->GetPoint();
	      v.Set(-v(1), v(0), 0.0);
	      v.Normalize();	  
	      v[2] = fW;
	      pt = v;
	      pt[3] = 0.0;

	      peCurr = pe;

	      pfaNew = CreateFacet();
	      pfaNew->SetForwardEdge(pe);
	      pe->SetFacet(pfaNew);

	      pvxNew = CreateVertex(pt);
	      peNew = CreateEdge();
	      peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	      peNew->GetEdge()->SetVertex(pvxNew);
	      peNew->SetFacet(pfaNew);
	      Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);

	      peCurr = peNew;

	      pvxNew = CreateVertex(pt);
	      peNew = CreateEdge();
	      peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	      peNew->GetEdge()->SetVertex(pvxNew);
	      peNew->SetFacet(pfaNew);
	      Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);

	      peCurr = peNew;

	      peNew = CreateEdge();
	      peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	      peNew->GetEdge()->SetVertex(pe->GetVertex());
	      peNew->SetFacet(pfaNew);
	      Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);
	      Splice(1, pe, peNew->GetEdge());

	      peCurr = peNew->GetEdge();

	      pfaNew = CreateFacet();
	      pfaNew->SetForwardVertex(peCurr->GetVertex());
	      peCurr->SetFacet(pfaNew);
	      peCurr->GetEdgePrev(0)->SetFacet(pfaNew);

	      peNew = CreateEdge();
	      peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	      peNew->GetEdge()->SetVertex(peCurr->GetEdgePrev(0)->GetVertex());
	      peNew->SetFacet(pfaNew);
	      Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);
	      Splice(1, peCurr->GetEdgePrev(0), peNew->GetEdge());
	    }

	  //////////////////////////////////////////////////
	  // Special Case the last vertex
	  peCurr = peStart;

	  pfaNew = CreateFacet();
	  pfaNew->SetForwardVertex(peCurr->GetVertex());
	  peCurr->SetFacet(pfaNew);
	  peCurr->GetEdgePrev(0)->SetFacet(pfaNew);

	  peNew = CreateEdge();
	  peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	  peNew->GetEdge()->SetVertex(peCurr->GetEdgePrev(0)->GetVertex());
	  peNew->SetFacet(pfaNew);
	  Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);
	  Splice(1, peCurr->GetEdgePrev(0), peNew->GetEdge());

	  return;
	}
      else
	{
	  ////////////////////////////////////////////////////////////
	  // Open Multi-edge segment

	  bMergeSingleSegment = (ppeContourLeft[0]->GetEdgePrev(0) == ppeContourRight[0]);

	  pe = ppeContourLeft[0];

	  //////////////////////////////////////////////////
	  // Special Case the first edge
	  peNext = pe->GetEdgeNext(0);

	  v = pe->GetEdge()->GetVertex()->GetPoint() - pe->GetVertex()->GetPoint();
	  v.Set(-v(1), v(0), 0.0);
	  v.Normalize(); 
	  v[2] = fW;
	  pt = v;
	  pt[3] = 0.0;
	  
	  peCurr = pe;

	  pfaNew = CreateFacet();
	  pfaNew->SetForwardEdge(pe);
	  pe->SetFacet(pfaNew);

	  pvxNew = CreateVertex(pt);
	  peNew = CreateEdge();
	  peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	  peNew->GetEdge()->SetVertex(pvxNew);
	  peNew->SetFacet(pfaNew);
	  Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);

	  peCurr = peNew;

	  peNew = CreateEdge();
	  peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	  peNew->GetEdge()->SetVertex(pe->GetVertex());
	  peNew->SetFacet(pfaNew);
	  Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);
	  Splice(1, pe, peNew->GetEdge());

	  for ( pe = peNext; pe != ppeContourRight[0]; pe = peNext )
	    {
	      peNext = pe->GetEdgeNext(0);

	      //////////////////////////////////////////////////
	      // The normal case
	      v = pe->GetEdge()->GetVertex()->GetPoint() - pe->GetVertex()->GetPoint();
	      v.Set(-v(1), v(0), 0.0);
	      v.Normalize();
	      v[2] = fW;
	      pt = v;
	      pt[3] = 0.0;
	      
	      peCurr = pe;

	      pfaNew = CreateFacet();
	      pfaNew->SetForwardEdge(pe);
	      pe->SetFacet(pfaNew);

	      pvxNew = CreateVertex(pt);
	      peNew = CreateEdge();
	      peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	      peNew->GetEdge()->SetVertex(pvxNew);
	      peNew->SetFacet(pfaNew);
	      Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);

	      peCurr = peNew;

	      pvxNew = CreateVertex(pt);
	      peNew = CreateEdge();
	      peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	      peNew->GetEdge()->SetVertex(pvxNew);
	      peNew->SetFacet(pfaNew);
	      Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);

	      peCurr = peNew;

	      peNew = CreateEdge();
	      peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	      peNew->GetEdge()->SetVertex(pe->GetVertex());
	      peNew->SetFacet(pfaNew);
	      Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);
	      Splice(1, pe, peNew->GetEdge());

	      peCurr = peNew->GetEdge();

	      pfaNew = CreateFacet();
	      pfaNew->SetForwardVertex(peCurr->GetVertex());
	      peCurr->SetFacet(pfaNew);
	      peCurr->GetEdgePrev(0)->SetFacet(pfaNew);

	      peNew = CreateEdge();
	      peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	      peNew->GetEdge()->SetVertex(peCurr->GetEdgePrev(0)->GetVertex());
	      peNew->SetFacet(pfaNew);
	      Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);
	      Splice(1, peCurr->GetEdgePrev(0), peNew->GetEdge());
	    }

	  //////////////////////////////////////////////////
	  // Special Case the last edge
	  v = pe->GetEdge()->GetVertex()->GetPoint() - pe->GetVertex()->GetPoint();
	  v.Set(-v(1), v(0), 0.0);
	  v.Normalize();  
	  v[2] = fW;
	  pt = v;
	  pt[3] = 0.0;
	  
	  peCurr = pe;

	  pfaNew = CreateFacet();
	  pfaNew->SetForwardEdge(pe);
	  pe->SetFacet(pfaNew);

	  pvxNew = CreateVertex(pt);
	  peNew = CreateEdge();
	  peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	  peNew->GetEdge()->SetVertex(pvxNew);
	  peNew->SetFacet(pfaNew);
	  Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);

	  peCurr = peNew;

	  peNew = CreateEdge();
	  peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	  peNew->GetEdge()->SetVertex(pe->GetVertex());
	  peNew->SetFacet(pfaNew);
	  Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);
	  Splice(1, pe, peNew->GetEdge());

	  peCurr = peNew->GetEdge();

	  pfaNew = CreateFacet();
	  pfaNew->SetForwardVertex(peCurr->GetVertex());
	  peCurr->SetFacet(pfaNew);
	  peCurr->GetEdgePrev(0)->SetFacet(pfaNew);

	  peNew = CreateEdge();
	  peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	  peNew->GetEdge()->SetVertex(peCurr->GetEdgePrev(0)->GetVertex());
	  peNew->SetFacet(pfaNew);
	  Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);
	  Splice(1, peCurr->GetEdgePrev(0), peNew->GetEdge());

	  if ( !bMergeSingleSegment )
	    {
	      // This is the normal case
	      return;
	    }
	}

      //////////////////////////////////////////////////
      // If there is exactly one concave vertex then the outer
      //   Voronoi diagram will have a single segment but will still need to be merged
      ASSERT( !bInside );

      peLeftEnd = ppeContourLeft[0];
      peLeftMiddle = ppeContourRight[0];
      peRightMiddle = ppeContourLeft[0];
      peRightEnd = ppeContourRight[0];
    }
  else
    {
      ASSERT( uiSize > 1 );

      ////////////////////////////////////////////////////////////
      // Divide and Conquer

      //////////////////////////////////////////////////
      // Split and recursively compute sub-Voronoi diagrams
      uiHalf = uiSize / 2;

      peLeftEnd = ppeContourLeft[0];
      peLeftMiddle = ppeContourRight[uiHalf-1];
      peRightMiddle = ppeContourLeft[uiHalf];
      peRightEnd = ppeContourRight[uiSize-1];

      bPossiblyMergeTwice = (peLeftEnd->GetEdgePrev(0) == peRightEnd);

      CreateVoronoiPolygonHeld(uiHalf, ppeContourLeft, ppeContourRight, fW, FALSE, bInside);
      CreateVoronoiPolygonHeld(uiSize - uiHalf, ppeContourLeft + uiHalf, 
			       ppeContourRight + uiHalf, fW, FALSE, bInside);
    }

  MergeVoronoiPolygonHeld(peLeftEnd, peLeftMiddle, peRightMiddle, peRightEnd, fW);
  if ( bPossiblyMergeTwice && 
       peLeftEnd->GetEdgeNext(1) != peRightEnd->GetEdgeNext(0) )
    {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
      fprintf(stderr, "Merge again because this is the last outside merge!\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
      ASSERT( !bInside );
      MergeVoronoiPolygonHeld(peRightMiddle, peRightEnd, peLeftEnd, peLeftMiddle, fW);
    }
}

VOID QEMesh::MergeVoronoiPolygonHeld(QEEdge *peLeftEnd, QEEdge *peLeftMiddle, 
				     QEEdge *peRightMiddle, QEEdge *peRightEnd, FLOAT fW)
{
  QEEdge *peLeft, *peRight, *peBisector;
  QEEdge *peLeftNext, *peRightNext;
  QEEdge *pe, *peNew;
  QEVertex *pvxNew, *ppvxEndpoint[2];
  Point pt, pptResults[2];
  UINT_32 uiResults;
  Vector v;
  FLOAT pfDistance[2];

  UINT_32 uiMergeDirection;
  Point ptMerge, ptLeft, ptRight;
  QEFacet *pfaMerge, *pfaTest, *pfaLeft, *pfaRight;
  BOOL bReplacePoint, bContinueSearch;

  //////////////////////////////////////////////////////////////////////
  // Merge

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "    START: Merge Left = ( %.3f, %.3f ) Right = ( %.3f, %.3f )\n", peLeftMiddle->GetVertex()->GetPoint()(X), peLeftMiddle->GetVertex()->GetPoint()(Y), peRightMiddle->GetVertex()->GetPoint()(X), peRightMiddle->GetVertex()->GetPoint()(Y));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

  ////////////////////////////////////////
  // Start the Bisector

  //////////////////////////////
  // Left side
  pe = peLeftMiddle->GetEdgeNext(0);
  peLeft = pe->GetEdgeNext(0);

  // Remove the edge at infinity
  ASSERT( pe->GetEdge()->GetFacet() == GetNullFacet() );
  ASSERT( pe != peLeftEnd && 
	  pe != peLeftMiddle &&
	  pe != peLeft );
  pe->GetEdge()->GetEdgePrev(1)->SetVertex(pe->GetEdge()->GetEdgePrev(1)->GetVertex());
  pe->GetEdgePrev(1)->SetVertex(pe->GetEdgePrev(1)->GetVertex());
  peLeftMiddle->SetFacet(peLeftMiddle->GetFacet());
  Splice(1, pe->GetEdge(), pe->GetEdge()->GetEdgePrev(1));
  Splice(1, pe, pe->GetEdgePrev(1));
  DestroyEdge(pe->GetEdge());
  DestroyEdge(pe);

  //////////////////////////////
  // Right side
  pe = peRightMiddle->GetEdgePrev(0);
  peRight = pe->GetEdgePrev(0);

  // Remove the edge at infinity, and recycle it as the first bisector
  ASSERT( pe->GetEdge()->GetFacet() == GetNullFacet() );
  ASSERT( pe != peRightEnd && 
	  pe != peRightMiddle &&
	  pe != peRight );
  pe->GetEdge()->GetEdgePrev(1)->SetVertex(pe->GetEdge()->GetEdgePrev(1)->GetVertex());
  pe->GetEdgePrev(1)->SetVertex(pe->GetEdgePrev(1)->GetVertex());
  peRightMiddle->SetFacet(peRightMiddle->GetFacet());
  Splice(1, pe->GetEdge(), pe->GetEdge()->GetEdgePrev(1));
  Splice(1, pe, pe->GetEdgePrev(1));

  peBisector = pe;

  //////////////////////////////
  // Attach the Bisector
  ASSERT( peLeftMiddle->GetEdgeNext(0) == peRightMiddle );
  peBisector->SetVertex(peRightMiddle->GetVertex()); // This is really unset
  peBisector->GetEdge()->SetVertex(peRightMiddle->GetVertex());
  peBisector->SetFacet(peRightMiddle->GetFacet());
  peBisector->GetEdge()->SetFacet(peLeftMiddle->GetFacet());

  Splice(1, peRightMiddle, peBisector->GetEdge());

  ////////////////////////////////////////
  // Merge Loop
  while ( TRUE )
    {
      uiMergeDirection = 0;
      ptMerge.Set(0.0, 0.0, INFINITY);
      pfaMerge = NULL;
      pfaLeft = NULL;
      pfaRight = NULL;

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
      fprintf(stderr, "      START: Searching for candidate point\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

      ////////////////////////////////////////////////////////////
      // START: Collapse edges from the left and right
      //   until they are in the sector bounded by the 
      //   two contour edges
      
      // Left
      ASSERT( peLeft->GetEdge()->GetFacet() != GetNullFacet() ||
	      peLeft == peLeft->GetFacet()->GetForwardEdge() );
      while ( peLeft != peLeft->GetFacet()->GetForwardEdge() &&
	      peLeft->GetEdgeNext(0) != peLeft->GetEdgeNext(0)->GetFacet()->GetForwardEdge() &&
	      peLeft->GetEdge()->GetVertex() != peLeft->GetFacet()->GetForwardVertex() )

	{
	  // Check if the Voronoi vertex is closer to the left or right edge
	  pt = peLeft->GetEdge()->GetVertex()->GetPoint();
	  ASSERT( pt(3) != 0.0 );
	  //pfDistance[0] = CalculateDistance(peLeft->GetFacet(), pt, &(ppvxEndpoint[0]));
	  pfDistance[0] = pt[2];
	  pt[2] = 0.0;
	  pt[3] = 1.0;
	  pfDistance[1] = CalculateDistance(peRight->GetFacet(), pt, &(ppvxEndpoint[1]));
	  if ( pfDistance[0] < pfDistance[1] )
	    {
	      break;
	    }
	  
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	  fprintf(stderr, "        Left side skipping uniformed Voronoi edge\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	  // Advance the left next pointer
	  peLeftNext = peLeft->GetEdgeNext(0);
	  if ( peLeftNext->GetEdge()->GetFacet() == GetNullFacet() &&
	       peLeftNext != peLeftNext->GetFacet()->GetForwardEdge() )
	    {
	      peLeftNext = peLeftNext->GetEdgeNext(0);
	    }

	  //////////////////////////////
	  // Advance peLeft cleaning up extraneous edges
	  if ( peLeft->GetEdgeNext(0)->GetEdge()->GetFacet() == GetNullFacet() )
	    {
	      ASSERT( peLeft->GetEdgeNext(0) != peLeft->GetEdgeNext(0)->GetFacet()->GetForwardEdge() );
	      CollapseEdge(peLeft->GetEdgeNext(0));
	    }
	  ASSERT( peLeft != peLeft->GetFacet()->GetForwardEdge() );
	  CollapseEdge(peLeft);
	  peLeft = peLeftNext;

	  ASSERT( peLeft->GetEdge()->GetFacet() != GetNullFacet() );
	}
      
      // Right
      ASSERT( peRight->GetEdge()->GetFacet() != GetNullFacet() ||
	      peRight == peRight->GetFacet()->GetForwardEdge() );
      while ( peRight != peRight->GetFacet()->GetForwardEdge() &&
	      peRight->GetEdgePrev(0) != peRight->GetEdgePrev(0)->GetFacet()->GetForwardEdge() &&
	      peRight->GetVertex() != peRight->GetFacet()->GetForwardVertex() )
	{
	  // Check if the Voronoi vertex is closer to the left or right edge
	  pt = peRight->GetVertex()->GetPoint();
	  ASSERT( pt(3) != 0.0 );
	  //pfDistance[0] = CalculateDistance(peRight->GetFacet(), pt, &(ppvxEndpoint[0]));
	  pfDistance[0] = pt[2];
	  pt[2] = 0.0;
	  pt[3] = 1.0;
	  pfDistance[1] = CalculateDistance(peLeft->GetFacet(), pt, &(ppvxEndpoint[1]));
	  if ( pfDistance[0] < pfDistance[1] )
	    {
	      break;
	    }
	  
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	  fprintf(stderr, "        Right side skipping uniformed Voronoi edge\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	  // Advance the right next pointer
	  peRightNext = peRight->GetEdgePrev(0);
	  if ( peRightNext->GetEdge()->GetFacet() == GetNullFacet() &&
	       peRightNext != peRightNext->GetFacet()->GetForwardEdge() )
	    {
	      peRightNext = peRightNext->GetEdgePrev(0);
	    }
	  
	  //////////////////////////////
	  // Advance peRight cleaning up extraneous edges
	  if ( peRight->GetEdgePrev(0)->GetEdge()->GetFacet() == GetNullFacet() )
	    {
	      ASSERT( peRight->GetEdgePrev(0) != peRight->GetEdgePrev(0)->GetFacet()->GetForwardEdge() );
	      CollapseEdge(peRight->GetEdgePrev(0)->GetEdge());
	    }
	  ASSERT( peRight != peRight->GetFacet()->GetForwardEdge() );
	  CollapseEdge(peRight->GetEdge());
	  peRight = peRightNext;

	  ASSERT( peRight->GetEdge()->GetFacet() != GetNullFacet() );
	}

      // END  : Collapse edges from the left and right
      //   until they are in the sector bounded by the 
      //   two contour edges
      ////////////////////////////////////////////////////////////      

      peLeftNext = peLeft;
      peRightNext = peRight;

      ////////////////////////////////////////
      // Search for the best candidate
      bContinueSearch = TRUE;
      while ( bContinueSearch )
	{
	  bContinueSearch = FALSE;
	  bReplacePoint = FALSE;

	  ////////////////////////////////////////
	  // Search for Left candidate
	      
	  uiResults = 0;

	  // Do not go past a contour edge
	  if ( peLeftNext != peLeftNext->GetFacet()->GetForwardEdge() &&
	       (peLeftNext->GetFacet()->GetForwardVertex() == NULL ||
		peLeftNext->GetVertex() != peLeftNext->GetFacet()->GetForwardVertex()) &&
	       peLeftNext != peBisector->GetEdge() &&
	       peLeftNext->GetEdgeNext(0) != peBisector->GetEdge() )
	    {
	      ASSERT( peLeftNext->GetEdge()->GetFacet() != GetNullFacet() );
	      if ( Counterclockwise(peLeftNext->GetEdge()->GetFacet(), TRUE,
				    peLeft->GetFacet(), TRUE,
				    peRight->GetFacet(), TRUE) )
		{
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
		  fprintf(stderr, "        Calculating left candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
		  CalculateVoronoiPoint(pptResults, uiResults, 
					peLeftNext->GetEdge()->GetFacet(), TRUE,
					peLeft->GetFacet(), TRUE,
					peRight->GetFacet(), TRUE, 
					TRUE);
		  if ( uiResults > 0 ) 
		    {
		      //////////////////////////////
		      // Check if this candidate is closer than the current right point
		      bReplacePoint = FALSE;
		      if ( peLeft == peLeftNext )
			{
			  //////////////////////////////
			  // No left candidate yet
			  
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			  fprintf(stderr, "        Found first left candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			  bReplacePoint = TRUE;
			}
		      else
			{
			  //////////////////////////////
			  // Do an InCircle test with the current left point
			  
			  ASSERT( peLeft->GetFacet() != GetNullFacet() &&
				  peLeft->GetFacet()->GetForwardEdge() != peLeft );
			  pfaTest = peLeft->GetEdge()->GetFacet();

			  bReplacePoint = FALSE;

			  //pfDistance[0] = CalculateDistance(peLeftNext->GetEdge()->GetFacet(), ptLeft, &(ppvxEndpoint[0]));
			  pfDistance[1] = CalculateDistance(pfaTest, pptResults[0], &(ppvxEndpoint[1]));

			  if ( pptResults[0](Z) <= ptLeft(Z) &&
			       ( pptResults[0](Z) <= pfDistance[1] ||
				 ( ppvxEndpoint[1] != NULL &&
				   ( peLeftNext->GetEdge()->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
				     peLeft->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
				     peRight->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ) ) ) )
			    {
			      //////////////////////////////
			      // Replacing left candidate

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			      fprintf(stderr, "        Updated left candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			      bReplacePoint = TRUE;
			    }
			}

		      if ( bReplacePoint ) 
			{
			  pfaLeft = peLeftNext->GetEdge()->GetFacet();
			  ptLeft = pptResults[0];

			  //////////////////////////////
			  // Check if this candidate is closer than the current merge point

			  pfaTest = pfaMerge;
			  if ( pfaTest == NULL &&
			       peRight->GetFacet()->GetForwardEdge() != peRight )
			    {
			      pfaTest = peRight->GetEdge()->GetFacet();
			      ASSERT( pfaTest != GetNullFacet() );
			    }

			  bReplacePoint = FALSE;
			  if ( pfaTest == NULL )
			    {
			      bReplacePoint = TRUE;
			    }
			  else
			    {
			      //pfDistance[0] = CalculateDistance(peLeftNext->GetEdge()->GetFacet(), ptMerge, &(ppvxEndpoint[0]));
			      pfDistance[1] = CalculateDistance(pfaTest, pptResults[0], &(ppvxEndpoint[1]));

			      if ( pptResults[0](Z) <= ptMerge(Z) &&
				   ( pptResults[0](Z) <= pfDistance[1] ||
				     ( ppvxEndpoint[1] != NULL &&
				       ( peLeftNext->GetEdge()->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
					 peLeft->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
					 peRight->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ) ) ) )
				{
				  bReplacePoint = TRUE;
				}
			    }

			  if ( bReplacePoint )
			    {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			      fprintf(stderr, "        Replacing merge point with left candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			      bContinueSearch = TRUE;
			      uiMergeDirection = 1;
			      ptMerge = pptResults[0];
			      pfaMerge = peLeftNext->GetEdge()->GetFacet();
			    }

			  if ( peLeft != peLeftNext )
			    {
			      //////////////////////////////
			      // Advance peLeft cleaning up extraneous edges
			      if ( peLeft->GetEdgeNext(0)->GetEdge()->GetFacet() == GetNullFacet() )
				{
				  ASSERT( peLeft->GetEdgeNext(0) != peLeft->GetEdgeNext(0)->GetFacet()->GetForwardEdge() );
				  CollapseEdge(peLeft->GetEdgeNext(0));
				}
			      ASSERT( peLeft != peLeft->GetFacet()->GetForwardEdge() );
			      CollapseEdge(peLeft);
			      peLeft = peLeftNext;
			    }

			  // Advance the left next pointer
			  peLeftNext = peLeft->GetEdgeNext(0);
			  if ( peLeftNext->GetEdge()->GetFacet() == GetNullFacet() &&
			       peLeftNext != peLeftNext->GetFacet()->GetForwardEdge() &&
			       peLeftNext->GetEdge()->GetVertex() != peLeftNext->GetFacet()->GetForwardVertex() )
			    {
			      peLeftNext = peLeftNext->GetEdgeNext(0);
			    }
			}
		    }
		}
	    }

	  ////////////////////////////////////////
	  // Search for Right candidate
	      
	  uiResults = 0;

	  // Do not go past a contour edge
	  if ( peRightNext != peRightNext->GetFacet()->GetForwardEdge() &&
	       (peRightNext->GetFacet()->GetForwardVertex() == NULL ||
		peRightNext->GetEdge()->GetVertex() != peRightNext->GetFacet()->GetForwardVertex()) &&
	       peRightNext != peBisector &&
	       peRightNext->GetEdgePrev(0) != peBisector )
	    {
	      ASSERT( peRightNext->GetEdge()->GetFacet() != GetNullFacet() );
	      if ( Counterclockwise(peLeft->GetFacet(), TRUE,
				    peRight->GetFacet(), TRUE,
				    peRightNext->GetEdge()->GetFacet(), TRUE) )
		{
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
		  fprintf(stderr, "        Calculating right candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
		  CalculateVoronoiPoint(pptResults, uiResults, 
					peLeft->GetFacet(), TRUE, 
					peRight->GetFacet(), TRUE,
					peRightNext->GetEdge()->GetFacet(), TRUE,
					TRUE);
		  if ( uiResults > 0 )
		    {
		      //////////////////////////////
		      // Check if this candidate is closer than the current right point
		      bReplacePoint = FALSE;
		      if ( peRight == peRightNext )
			{
			  //////////////////////////////
			  // No right candidate yet
			  
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			  fprintf(stderr, "        Found first right candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			  bReplacePoint = TRUE;
			}
		      else
			{
			  //////////////////////////////
			  // Do an InCircle test with the curren right point
			  
			  ASSERT( peRight->GetFacet() != GetNullFacet() &&
				  peRight->GetFacet()->GetForwardEdge() != peRight );
			  pfaTest = peRight->GetEdge()->GetFacet();

			  bReplacePoint = FALSE;

			  //pfDistance[0] = CalculateDistance(peRightNext->GetEdge()->GetFacet(), ptRight, &(ppvxEndpoint[0]));
			  pfDistance[1] = CalculateDistance(pfaTest, pptResults[0], &(ppvxEndpoint[1]));

			  if ( pptResults[0](Z) <= ptRight(Z) &&
			       ( pptResults[0](Z) <= pfDistance[1] ||
				 ( ppvxEndpoint[1] != NULL &&
				   ( peLeft->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
				     peRight->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
				     peRightNext->GetEdge()->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ) ) ) )
			    {
			      //////////////////////////////
			      // Replacing right candidate

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			      fprintf(stderr, "        Updated right candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			      bReplacePoint = TRUE;
			    }
			}

		      if ( bReplacePoint ) 
			{
			  pfaRight = peRightNext->GetEdge()->GetFacet();
			  ptRight = pptResults[0];

			  //////////////////////////////
			  // Check if this candidate is closer than the current merge point

			  pfaTest = pfaMerge;
			  if ( pfaTest == NULL &&
			       peLeft->GetFacet()->GetForwardEdge() != peLeft )
			    {
			      pfaTest = peLeft->GetEdge()->GetFacet();
			      ASSERT( pfaTest != GetNullFacet() );
			    }

			  bReplacePoint = FALSE;
			  if ( pfaTest == NULL )
			    {
			      bReplacePoint = TRUE;
			    }
			  else
			    {
			      //pfDistance[0] = CalculateDistance(peRightNext->GetEdge()->GetFacet(), ptMerge, &(ppvxEndpoint[0]));
			      pfDistance[1] = CalculateDistance(pfaTest, pptResults[0], &(ppvxEndpoint[1]));

			      if ( pptResults[0](Z) <= ptMerge(Z) &&
				   ( pptResults[0](Z) <= pfDistance[1] ||
				     ( ppvxEndpoint[1] != NULL &&
				       ( peLeft->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
					 peRight->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ||
					 peRightNext->GetEdge()->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] ) ) ) )
			    {
				  bReplacePoint = TRUE;
				}
			    }

			  if ( bReplacePoint )
			    {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			      fprintf(stderr, "        Replacing merge point with right candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			      bContinueSearch = TRUE;
			      uiMergeDirection = 2;
			      ptMerge = pptResults[0];
			      pfaMerge = peRightNext->GetEdge()->GetFacet();
			    }

			  if ( peRight != peRightNext )
			    {
			      //////////////////////////////
			      // Advance peRight cleaning up extraneous edges
			      if ( peRight->GetEdgePrev(0)->GetEdge()->GetFacet() == GetNullFacet() )
				{
				  ASSERT( peRight->GetEdgePrev(0) != peRight->GetEdgePrev(0)->GetFacet()->GetForwardEdge() );
				  CollapseEdge(peRight->GetEdgePrev(0)->GetEdge());
				}
			      ASSERT( peRight != peRight->GetFacet()->GetForwardEdge() );
			      CollapseEdge(peRight->GetEdge());
			      peRight = peRightNext;
			    }

			  // Advance the right next pointer
			  peRightNext = peRight->GetEdgePrev(0);
			  if ( peRightNext->GetEdge()->GetFacet() == GetNullFacet() &&
			       peRightNext != peRightNext->GetFacet()->GetForwardEdge() &&
			       peRightNext->GetVertex() != peRightNext->GetFacet()->GetForwardVertex() )
			    {
			      peRightNext = peRightNext->GetEdgePrev(0);
			    }
			}
		    }
		}

	      //////////////////////////////////////////////////
	      // Because of numeric issues, the left and right might
	      // get equally good answers which will confuse things
	      // and cancel out.  We need to override that possibility
	      if ( pfaMerge == NULL &&
		   ( pfaLeft != NULL || pfaRight != NULL ) )
		{
		  if ( pfaLeft != NULL )
		    {
		      pfDistance[0] = CalculateDistance(peRight->GetEdge()->GetFacet(), ptLeft, &(ppvxEndpoint[0]));
		    }
		  if ( pfaRight != NULL )
		    {
		      pfDistance[1] = CalculateDistance(peLeft->GetEdge()->GetFacet(), ptRight, &(ppvxEndpoint[1]));
		    }

		  if ( ( pfaRight == NULL ||
			 ( pfaRight != NULL && pfaLeft != NULL &&
			   !( ptRight(Z) <= pfDistance[1] + s_cfDistanceEpsilon ) ) ) &&
		       ptLeft(Z) <= pfDistance[0] + s_cfDistanceEpsilon )
		    {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
		      fprintf(stderr, "        Special replace merge point with left candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
		      bContinueSearch = TRUE;
		      uiMergeDirection = 1;
		      ptMerge = ptLeft;
		      pfaMerge = pfaLeft;
		    }
		  else if ( ( pfaLeft == NULL ||
			 ( pfaRight != NULL && pfaLeft != NULL &&
			   !( ptLeft(Z) <= pfDistance[0] + s_cfDistanceEpsilon ) ) ) &&
		       ptRight(Z) <= pfDistance[1] + s_cfDistanceEpsilon )
		    {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
		      fprintf(stderr, "        Special replace merge point with right candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
		      bContinueSearch = TRUE;
		      uiMergeDirection = 2;
		      ptMerge = ptRight;
		      pfaMerge = pfaRight;
		    }
		  else if ( pfaRight != NULL && pfaLeft != NULL &&
			    ptLeft(Z) <= pfDistance[0] + s_cfDistanceEpsilon &&
			    ptRight(Z) <= pfDistance[1] + s_cfDistanceEpsilon )
		    {
		      if ( ptLeft(Z) < ptRight(Z) )
			{
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			  fprintf(stderr, "        Special replace merge point with left candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			  bContinueSearch = TRUE;
			  uiMergeDirection = 1;
			  ptMerge = ptLeft;
			  pfaMerge = pfaLeft;
			}
		      else
			{
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			  fprintf(stderr, "        Special replace merge point with right candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			  bContinueSearch = TRUE;
			  uiMergeDirection = 2;
			  ptMerge = ptRight;
			  pfaMerge = pfaRight;
			}
		    }
		}
	    }
	}

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
      fprintf(stderr, "      END  : Searching for candidate point\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

      //////////////////////////////////////////////////
      // Check if we have a valid candidate
      if ( uiMergeDirection == 0 ) 
	{
	  break;
	} 
      else if ( uiMergeDirection == 1 ) 
	{
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	  fprintf(stderr, "      Attaching to left ( %.3f, %.3f, %.3f )\n", ptMerge(X), ptMerge(Y), ptMerge(Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	  //////////////////////////////////////////////////
	  // Attach the bisector to the left side

	  ptMerge[3] = fW;

	  peLeftNext = peLeft->GetEdge()->GetEdgeNext(0);
	  pe = peBisector;

	  pvxNew = CreateVertex(ptMerge);
	  peBisector = CreateEdge();

	  peLeft->GetEdgePrev(1)->SetVertex(peLeft->GetEdgePrev(1)->GetVertex());
	  Splice(1, peLeft, peLeft->GetEdgePrev(1));

	  peLeft->SetVertex(pvxNew);
	  pe->SetVertex(pvxNew);
	  peBisector->GetEdge()->SetVertex(pvxNew);

	  Splice(1, pe, peBisector->GetEdge());
	  Splice(1, peLeft, peBisector->GetEdge());

	  peLeft = peLeftNext;

	  peBisector->SetFacet(peRight->GetFacet());
	  peBisector->GetEdge()->SetFacet(peLeft->GetFacet());

	  if ( peLeft->GetEdge()->GetFacet() == GetNullFacet() &&
	       peLeft->GetFacet()->GetForwardEdge() != peLeft && 
	       peLeft->GetFacet()->GetForwardVertex() != peLeft->GetEdge()->GetVertex() )
	    {
	      peLeftNext = peLeft->GetEdgeNext(0);
	      CollapseEdge(peLeft);
	      peLeft = peLeftNext;
	    }
	}
      else
	{
	  //////////////////////////////////////////////////
	  // Attach the bisector to the right side

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	  fprintf(stderr, "      Attaching to right ( %.3f, %.3f, %.3f )\n", ptMerge(X), ptMerge(Y), ptMerge(Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	  ptMerge[3] = fW;

	  peRightNext = peRight->GetEdge()->GetEdgePrev(0);
	  pe = peBisector;

	  pvxNew = CreateVertex(ptMerge);
	  peBisector = CreateEdge();

	  peRight->GetEdge()->GetEdgePrev(1)->SetVertex(peRight->GetEdge()->GetEdgePrev(1)->GetVertex());
	  Splice(1, peRight->GetEdge(), peRight->GetEdge()->GetEdgePrev(1));

	  peRight->GetEdge()->SetVertex(pvxNew);
	  pe->SetVertex(pvxNew);
	  peBisector->GetEdge()->SetVertex(pvxNew);

	  Splice(1, pe, peBisector->GetEdge());
	  Splice(1, peRight->GetEdge(), pe);

	  peRight = peRightNext;

	  peBisector->SetFacet(peRight->GetFacet());
	  peBisector->GetEdge()->SetFacet(peLeft->GetFacet());

	  if ( peRight->GetEdge()->GetFacet() == GetNullFacet() &&
	       peRight->GetFacet()->GetForwardEdge() != peRight &&
	       peRight->GetFacet()->GetForwardVertex() != peRight->GetVertex() )
	    {
	      peRightNext = peRight->GetEdgePrev(0);
	      CollapseEdge(peRight->GetEdge());
	      peRight = peRightNext;
	    }
	}

      peLeftNext = peLeft;
      peRightNext = peRight;
    }

  //////////////////////////////////////////////////
  // Seal off the bisector
  if ( peRight->GetEdgeNext(0) == peLeft )
    {
      //////////////////////////////////////////////////
      // Finishing off the whole contour
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
      fprintf(stderr, "      Contour done attaching bisector to vertex ( %.3f, %.3f, %.3f )\n", peLeft->GetVertex()->GetPoint()(X), peLeft->GetVertex()->GetPoint()(Y), peLeft->GetVertex()->GetPoint()(Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

      ASSERT( peLeft->GetFacet()->GetForwardEdge() == peLeft &&
	      peRight->GetFacet()->GetForwardEdge() == peRight );
      Splice(1, peLeft, peBisector);
      peBisector->SetVertex(peLeft->GetVertex());
    }
  else
    {
      ASSERT( peLeft->GetVertex() != peRight->GetEdge()->GetVertex() );

      v = CalculateBisectorDirection(peLeft->GetFacet(), TRUE, 
				     peRight->GetFacet(), TRUE, FALSE);
      v[2] = fW;
      pt = v;
      pt[3] = 0.0;
      pvxNew = CreateVertex(pt);
      peBisector->SetVertex(pvxNew);

      peNew = CreateEdge();
      peNew->SetVertex(pvxNew);
      peNew->GetEdge()->SetVertex(peLeft->GetVertex());
      peNew->SetFacet(peLeft->GetFacet());
      Splice(1, peLeft, peNew->GetEdge());
      Splice(1, peBisector, peNew);

      peNew = CreateEdge();
      peNew->SetVertex(pvxNew);
      peNew->GetEdge()->SetVertex(peRight->GetEdge()->GetVertex());
      peNew->GetEdge()->SetFacet(peRight->GetFacet());
      Splice(1, peRight->GetEdge()->GetEdgePrev(1), peNew->GetEdge());
      Splice(1, peBisector, peNew);
    }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "    END  : Merge Left = ( %.3f, %.3f ) Right = ( %.3f, %.3f )\n", peLeftMiddle->GetVertex()->GetPoint()(X), peLeftMiddle->GetVertex()->GetPoint()(Y), peRightMiddle->GetVertex()->GetPoint()(X), peRightMiddle->GetVertex()->GetPoint()(Y));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
}

//////////////////////////////////////////////////////////////////////
// General Mesh Functions

VOID QEMesh::ClearForwardPointers()
{
  QEMeshIter iter;
  QEVertex *pvx;
  QEEdge *pe;
  QEFacet *pfa;

  iter.Init(this);

  for ( pvx = iter.PeekFirstVertex();
	pvx != NULL;
	pvx = iter.PeekNextVertex() )
    {
      pvx->SetForwardVertex(NULL);
      pvx->SetForwardFacet(NULL);
    }

  for ( pe = iter.PeekFirstEdge();
	pe != NULL;
	pe = iter.PeekNextEdge() )
    {
      pe->SetForwardEdge(NULL);
    }

  for ( pe = iter.PeekFirstEdgeBoundary();
	pe != NULL;
	pe = iter.PeekNextEdgeBoundary() )
    {
      pe->SetForwardEdge(NULL);
    }

  for ( pfa = iter.PeekFirstFacet();
	pfa != NULL;
	pfa = iter.PeekNextFacet() )
    {
      pfa->SetForwardFacet(NULL);
      pfa->SetForwardVertex(NULL);
      pfa->SetForwardEdge(NULL);
    }
}

VOID QEMesh::MergeLocalFacet(QEFacet *pfa)
{
  ASSERT( pfa != NULL );

  CObjectListIter<QEEdge> iterQEEdge;
  QEFacetIter iterQEFacet;
  QEEdge *pe, *peBoundary;

  iterQEEdge.Init(m_plistQEEdgeBoundary);

  iterQEFacet.Init(pfa);
  for ( pe = iterQEFacet.PeekFirstEdge();
	pe != NULL;
	pe = iterQEFacet.PeekNextEdge() )
    {
      for ( peBoundary = iterQEEdge.PeekFirst();
	    peBoundary != NULL;
	    peBoundary = iterQEEdge.PeekNext() )
	{
	  if ( pe->GetVertex() == peBoundary->GetVertex() &&
	       pe->GetEdge()->GetVertex() == peBoundary->GetEdge()->GetVertex() )
	    {
	      // Found the matching edge
	      break;
	    }
	}

      if ( peBoundary == NULL )
	{
	  // There is no matching edge yet, so put the partner edge on the boundary list
	  m_plistQEEdge->Remove(pe->GetEdge());
	  m_plistQEEdgeBoundary->InsertLast(pe->GetEdge());
	}
      else
	{
	  pe->GetEdge()->SetFacet(peBoundary->GetEdge()->GetFacet());
	  pe->SetVertex(pe->GetVertex());
	  pe->GetEdge()->SetVertex(pe->GetEdge()->GetVertex());

	  Splice(1, peBoundary, pe->GetEdgePrev(1));
	  Splice(1, peBoundary->GetEdge()->GetEdgePrev(1), pe->GetEdge());
	  Splice(1, peBoundary, peBoundary->GetEdgePrev(1));
	  Splice(1, peBoundary->GetEdge(), peBoundary->GetEdge()->GetEdgePrev(1));

	  m_plistQEEdgeBoundary->Remove(peBoundary);
	  m_plistQEEdge->InsertLast(peBoundary);
	  DestroyEdge(peBoundary->GetEdge());
	  DestroyEdge(peBoundary);
	}
    }
}

VOID QEMesh::MergeFacet(CObjectList<QEVertex> *plistQEVertex)
{
  ASSERT( plistQEVertex != NULL );
  ASSERT( plistQEVertex->GetLength() > 2 );

  QEMesh *pme;
  QEFacet *pfa;
  QEVertex *pvx;
  QEEdge *pe, *peCurr;
  CObjectListIter<QEVertex> iterQEVertex;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  pme->m_plistQEVertex->Splice(plistQEVertex);

  pfa = pme->CreateFacet();
  ASSERT( pfa != NULL );
  
  iterQEVertex.Init(pme->m_plistQEVertex);
  pvx = iterQEVertex.PeekFirst();

  pe = pme->CreateContour(pfa, pvx);
  pme->m_plistQEEdge->Remove(pe->GetEdge());
  pme->m_plistQEEdgeBoundary->InsertLast(pe->GetEdge());

  for ( pvx = iterQEVertex.PeekNext();
	pvx != NULL;
	pvx = iterQEVertex.PeekNext() )
    {
      peCurr = pme->AddContourVertex(pe, pvx);
      pme->m_plistQEEdge->Remove(peCurr->GetEdge());
      pme->m_plistQEEdgeBoundary->InsertLast(peCurr->GetEdge());
    }

  MergeMesh(pme);
}

VOID QEMesh::MergeFacet(CObjectList<QEVertex> *plistQEVertex, CSLIDESurface *pCSLIDESurface)
{
  ASSERT( plistQEVertex != NULL );
  ASSERT( plistQEVertex->GetLength() > 2 );

  QEMesh *pme;
  QEFacet *pfa;
  QEVertex *pvx;
  QEEdge *pe, *peCurr;
  CObjectListIter<QEVertex> iterQEVertex;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  pme->m_plistQEVertex->Splice(plistQEVertex);

  pfa = pme->CreateFacet();
  ASSERT( pfa != NULL );
  pfa->SetSurface(pCSLIDESurface);
  
  iterQEVertex.Init(pme->m_plistQEVertex);
  pvx = iterQEVertex.PeekFirst();

  pe = pme->CreateContour(pfa, pvx);
  pme->m_plistQEEdge->Remove(pe->GetEdge());
  pme->m_plistQEEdgeBoundary->InsertLast(pe->GetEdge());

  for ( pvx = iterQEVertex.PeekNext();
	pvx != NULL;
	pvx = iterQEVertex.PeekNext() )
    {
      peCurr = pme->AddContourVertex(pe, pvx);
      pme->m_plistQEEdge->Remove(peCurr->GetEdge());
      pme->m_plistQEEdgeBoundary->InsertLast(peCurr->GetEdge());
    }

  MergeMesh(pme);
}

BOOL QEMesh::EqualEdge(QEEdge *pe, QEEdge *peResult) {
  Point pppt[2][2];
  Vector pv[2];
  FLOAT fEPSILON;
  
  fEPSILON = EPSILON;
  fEPSILON = 0.0001;
  
  pppt[0][0] = pe->GetVertex()->GetPoint();
  pppt[0][1] = pe->GetEdge()->GetVertex()->GetPoint();
  pppt[1][1] = peResult->GetVertex()->GetPoint();
  pppt[1][0] = peResult->GetEdge()->GetVertex()->GetPoint();
  pv[0] = pppt[1][0] - pppt[0][0];
  pv[1] = pppt[1][1] - pppt[0][1];
  if ( ( pe->GetVertex() == peResult->GetEdge()->GetVertex() || 
	 ( -fEPSILON <= pv[0][0] && pv[0][0] <= fEPSILON &&
	   -fEPSILON <= pv[0][1] && pv[0][1] <= fEPSILON &&
	   -fEPSILON <= pv[0][2] && pv[0][2] <= fEPSILON ) ) &&
       ( pe->GetEdge()->GetVertex() == peResult->GetVertex() || 
	 ( -fEPSILON <= pv[1][0] && pv[1][0] <= fEPSILON &&
	   -fEPSILON <= pv[1][1] && pv[1][1] <= fEPSILON &&
	   -fEPSILON <= pv[1][2] && pv[1][2] <= fEPSILON ) ) )
    {
      return 1;
    } 
  return 0;  
}

///////////////////////////////////////////////////////////
//// this method takes in two edges and a mesh.  the first edge
//// and the second edge should both have a null facet.  
//// The first edge is removed from the mesh provided as an argument.
//////////////////////////////////////////////////////////
VOID QEMesh::MergeEdge(QEEdge *pe, QEEdge *peMerge, QEMesh *pme) {
  ASSERT( peMerge->GetFacet() == m_pfaNULL );
  ASSERT( pe->GetEdge()->GetFacet() != m_pfaNULL );

  QEEdge *peCurr;

  peMerge->SetFacet(pe->GetEdge()->GetFacet());
  m_plistQEEdgeBoundary->Remove(peMerge);
  m_plistQEEdge->InsertLast(peMerge);
  
  if ( pe->GetEdgePrev(0) != peMerge )
    {	      
      DestroyVertex(pe->GetVertex());
      
      peCurr = pe;
      do
	{
	  peCurr->SetVertex(peMerge->GetEdge()->GetVertex());
	  
	  peCurr = peCurr->GetEdgeNext(1);
	}
      while ( peCurr != pe );
      
      Splice(1, pe, peMerge->GetEdgeNext(0));
    }
  
  if ( pe->GetEdgeNext(0) != peMerge )
    {	      
      DestroyVertex(pe->GetEdge()->GetVertex());
      
      peCurr = pe->GetEdge();
      do
	{
	  peCurr->SetVertex(peMerge->GetVertex());
	  
	  peCurr = peCurr->GetEdgeNext(1);
	}
      while ( peCurr != pe->GetEdge() );
      
      Splice(1, pe->GetEdgeNext(0), peMerge);
    }
  
  pe->GetVertex()->SetEdge(pe->GetEdgePrev(1));
  pe->GetEdge()->GetVertex()->SetEdge(pe->GetEdge()->GetEdgePrev(1));
  Splice(1, pe, pe->GetEdgePrev(1));
  Splice(1, pe->GetEdge(), pe->GetEdge()->GetEdgePrev(1));
  pme->m_plistQEEdge->Remove(pe->GetEdge());
  pme->m_plistQEEdgeBoundary->Remove(pe);
  DeallocateQEEdge(pe->GetEdge());
  DeallocateQEEdge(pe);
}

VOID QEMesh::MergeTriangleStrip(CObjectList<QEVertex> *plistQEVertex)
{
  ASSERT( plistQEVertex != NULL );
  ASSERT( plistQEVertex->GetLength() > 2 );

  QEMesh *pme;
  QEFacet *pfa;
  QEVertex *pvx;
  QEEdge *pe, *peCurr, *peLeft, *peFirst;
  CObjectListIter<QEVertex> iterQEVertex;
  int iVertexCount;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  pme->m_plistQEVertex->Splice(plistQEVertex);

  iterQEVertex.Init(pme->m_plistQEVertex);
  pvx = iterQEVertex.PeekFirst();
  
  pe = pme->CreateEdge();
  ASSERT(pe != NULL);
  pe->SetVertex(pvx);
  //first edge is a boundary edge, for now... merge later if necessary
  pme->m_plistQEEdge->Remove(pe->GetEdge());
  pme->m_plistQEEdgeBoundary->InsertLast(pe->GetEdge());
  pvx = iterQEVertex.PeekNext();
  pe->GetEdge()->SetVertex(pvx);
  peFirst = pe;
  peLeft = pe;
  // now the first edge has been built

  iVertexCount=2; 
  for ( pvx = iterQEVertex.PeekNext();
	pvx != NULL;
	pvx = iterQEVertex.PeekNext() )
    {
      // first set up facet to left edge
      pfa = pme->CreateFacet();
      ASSERT(pfa != NULL);
	  peLeft->SetFacet(pfa);
      //first make top (on even) or right (on odd) edge..
      peCurr = pme->CreateEdge();
      ASSERT(peCurr != NULL);
      peCurr->SetVertex(pvx);
      peCurr->GetEdge()->SetVertex(peLeft->GetVertex());
      //splice this into the left edge's vertex loop
      Splice(1,peLeft,peCurr->GetEdge());
      peCurr->SetFacet(pfa);

      //now make diagonal (on even) or bottom edge (on odd)
      pe = pme->CreateEdge();
      ASSERT (pe != NULL);
      pe->SetVertex(peLeft->GetEdge()->GetVertex());
      pe->GetEdge()->SetVertex(pvx);

      //now splice into left's vertex loop and tops vertex loop
      //edge next? or edge prev?
      Splice(1,peLeft->GetEdge()->GetEdgePrev(1), pe);
      Splice(1,peCurr,pe->GetEdge());
      pe->SetFacet(pfa);

      if (iVertexCount%2 == 0) { //if even
	//top edge is boundary edge
	pme->m_plistQEEdge->Remove(peCurr->GetEdge());
	pme->m_plistQEEdgeBoundary->InsertLast(peCurr->GetEdge());      
	//diag is next left edge
	peLeft=pe->GetEdge();
      } else { //odd
	//bottom edge is boundary edge
	pme->m_plistQEEdge->Remove(pe->GetEdge());
	pme->m_plistQEEdgeBoundary->InsertLast(pe->GetEdge()); 
	//right is next left edge
	peLeft=peCurr->GetEdge();
      }

      iVertexCount++;

    } 
  //last edge a boundary edge
  pme->m_plistQEEdge->Remove(peLeft);
  pme->m_plistQEEdgeBoundary->InsertLast(peLeft);      
  if (EqualEdge(peLeft,peFirst->GetEdge())) {
    pme->MergeEdge(peLeft,peFirst->GetEdge(),pme);
  }
  MergeMesh(pme);
}

VOID QEMesh::MergeTriangleStrip(CObjectList<QEVertex> *plistQEVertex,
				CSLIDESurface *pCSLIDESurface)
{
  ASSERT( plistQEVertex != NULL );
  ASSERT( plistQEVertex->GetLength() > 2 );

  QEMesh *pme;
  QEFacet *pfa;
  QEVertex *pvx;
  QEEdge *pe, *peCurr, *peLeft, *peFirst;
  CObjectListIter<QEVertex> iterQEVertex;
  int iVertexCount;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  pme->m_plistQEVertex->Splice(plistQEVertex);

  iterQEVertex.Init(pme->m_plistQEVertex);
  pvx = iterQEVertex.PeekFirst();
  
  pe = pme->CreateEdge();
  ASSERT(pe != NULL);
  pe->SetVertex(pvx);
  //first edge is a boundary edge, for now... merge later if necessary
  pme->m_plistQEEdge->Remove(pe->GetEdge());
  pme->m_plistQEEdgeBoundary->InsertLast(pe->GetEdge());
  pvx = iterQEVertex.PeekNext();
  pe->GetEdge()->SetVertex(pvx);
  peFirst = pe;
  peLeft = pe;
  // now the first edge has been built

  iVertexCount=2; 
  for ( pvx = iterQEVertex.PeekNext();
	pvx != NULL;
	pvx = iterQEVertex.PeekNext() )
    {
      // first set up facet to left edge
      pfa = pme->CreateFacet();
      ASSERT(pfa != NULL);
      pfa->SetSurface(pCSLIDESurface);
      peLeft->SetFacet(pfa);
      //first make top (on even) or right (on odd) edge..
      peCurr = pme->CreateEdge();
      ASSERT(peCurr != NULL);
      peCurr->SetVertex(pvx);
      peCurr->GetEdge()->SetVertex(peLeft->GetVertex());
      //splice this into the left edge's vertex loop
      Splice(1,peLeft,peCurr->GetEdge());
      peCurr->SetFacet(pfa);

      //now make diagonal (on even) or bottom edge (on odd)
      pe = pme->CreateEdge();
      ASSERT (pe != NULL);
      pe->SetVertex(peLeft->GetEdge()->GetVertex());
      pe->GetEdge()->SetVertex(pvx);

      //now splice into left's vertex loop and tops vertex loop
      //edge next? or edge prev?
      Splice(1,peLeft->GetEdge()->GetEdgePrev(1), pe);
      Splice(1,peCurr,pe->GetEdge());
      pe->SetFacet(pfa);

      if (iVertexCount%2 == 0) { //if even
	//top edge is boundary edge
	pme->m_plistQEEdge->Remove(peCurr->GetEdge());
	pme->m_plistQEEdgeBoundary->InsertLast(peCurr->GetEdge());      
	//diag is next left edge
	peLeft=pe->GetEdge();
      } else { //odd
	//bottom edge is boundary edge
	pme->m_plistQEEdge->Remove(pe->GetEdge());
	pme->m_plistQEEdgeBoundary->InsertLast(pe->GetEdge()); 
	//right is next left edge
	peLeft=peCurr->GetEdge();
      }

      iVertexCount++;

    } 
  //last edge a boundary edge
  pme->m_plistQEEdge->Remove(peLeft);
  pme->m_plistQEEdgeBoundary->InsertLast(peLeft);      
  if (EqualEdge(peLeft,peFirst->GetEdge())) {
    pme->MergeEdge(peLeft,peFirst->GetEdge(),pme);
  }
  MergeMesh(pme);
}

VOID QEMesh::MergeTriangleFan(CObjectList<QEVertex> *plistQEVertex)
{
  ASSERT( plistQEVertex != NULL );
  ASSERT( plistQEVertex->GetLength() > 2 );

  QEMesh *pme;
  QEFacet *pfa;
  QEVertex *pvx;
  QEEdge *pe, *peCurr, *peLeft, *peFirst;
  CObjectListIter<QEVertex> iterQEVertex;
  int iVertexCount;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  pme->m_plistQEVertex->Splice(plistQEVertex);

  iterQEVertex.Init(pme->m_plistQEVertex);
  pvx = iterQEVertex.PeekFirst();
  //pme->AddVertex(pvx);
  
  pe = pme->CreateEdge();
  ASSERT(pe != NULL);
  pe->SetVertex(pvx);
  //first edge is a boundary edge, for now... merge later if necessary
  pme->m_plistQEEdge->Remove(pe->GetEdge());
  pme->m_plistQEEdgeBoundary->InsertLast(pe->GetEdge());
  pvx = iterQEVertex.PeekNext();
  pe->GetEdge()->SetVertex(pvx);
  peFirst = pe;
  peLeft = pe;
  // now the first edge has been built

  for ( pvx = iterQEVertex.PeekNext();
	pvx != NULL;
	pvx = iterQEVertex.PeekNext() )
    {
      // first set up facet to left edge
      pfa = pme->CreateFacet();
      ASSERT(pfa != NULL);
	  peLeft->SetFacet(pfa);
      //first make right edge..
      peCurr = pme->CreateEdge();
      ASSERT(peCurr != NULL);
      peCurr->SetVertex(pvx);
      peCurr->GetEdge()->SetVertex(peLeft->GetVertex());
      //splice this into the left edge's vertex loop
      Splice(1,peLeft,peCurr->GetEdge());
      peCurr->SetFacet(pfa);

      //now make bottom edge
      pe = pme->CreateEdge();
      ASSERT (pe != NULL);
      pe->SetVertex(peLeft->GetEdge()->GetVertex());
      pe->GetEdge()->SetVertex(pvx);

      //now splice into left's vertex loop and tops vertex loop
      //edge next? or edge prev?
      Splice(1,peLeft->GetEdge()->GetEdgePrev(1), pe);
      Splice(1,peCurr,pe->GetEdge());
      pe->SetFacet(pfa);

      //bottom edge is boundary edge
      pme->m_plistQEEdge->Remove(pe->GetEdge());
      pme->m_plistQEEdgeBoundary->InsertLast(pe->GetEdge()); 
      //right is next left edge
      peLeft=peCurr->GetEdge();

      iVertexCount++;

    } 
  //last edge a boundary edge
  pme->m_plistQEEdge->Remove(peLeft);
  pme->m_plistQEEdgeBoundary->InsertLast(peLeft);      
  if (EqualEdge(peLeft,peFirst->GetEdge())) {
    pme->MergeEdge(peLeft,peFirst->GetEdge(),pme);
  }
  MergeMesh(pme);
}

VOID QEMesh::MergeQuadStrip(CObjectList<QEVertex> *plistQEVertex)
{
  ASSERT( plistQEVertex != NULL );
  ASSERT( plistQEVertex->GetLength() > 2 );

  QEMesh *pme;
  QEFacet *pfa;
  QEVertex *pvx, *pvx2;
  QEEdge *pe, *peFirst, *peLeft, *peTop, *peBot, *peRight;
  CObjectListIter<QEVertex> iterQEVertex;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();
  pme->m_plistQEVertex->Splice(plistQEVertex);

  iterQEVertex.Init(pme->m_plistQEVertex);
  pvx = iterQEVertex.PeekFirst();
  pvx2 = iterQEVertex.PeekNext();
  
  pe = pme->CreateEdge();
  ASSERT(pe != NULL);
  pe->GetEdge()->SetVertex(pvx);
  pe->SetVertex(pvx2);
  //first edge is a boundary edge, for now... merge later if necessary
  pme->m_plistQEEdge->Remove(pe);
  pme->m_plistQEEdgeBoundary->InsertLast(pe);
  peFirst = pe->GetEdge();
  peLeft = peFirst;
  //now the first edge has been built

  for ( pvx = iterQEVertex.PeekNext();
	pvx != NULL;
	pvx = iterQEVertex.PeekNext() )
    {
      pvx2 = iterQEVertex.PeekNext();
      ASSERT(pvx2 != NULL);

      //first set up facet to left edge
      pfa = pme->CreateFacet();
      ASSERT(pfa != NULL);
      peLeft->SetFacet(pfa);

      //make top
      peTop = pme->CreateEdge();
      ASSERT(peTop != NULL);
      peTop->SetVertex(pvx);
      peTop->GetEdge()->SetVertex(peLeft->GetVertex());
      //splice this into the left edge's vertex loop
      Splice(1,peLeft,peTop->GetEdge());
      peTop->SetFacet(pfa);
      //top edge is boundary edge
      pme->m_plistQEEdge->Remove(peTop->GetEdge());
      pme->m_plistQEEdgeBoundary->InsertLast(peTop->GetEdge());      

      //make bottom
      peBot = pme->CreateEdge();
      ASSERT(peBot != NULL);
      peBot->SetVertex(peLeft->GetEdge()->GetVertex());
      peBot->GetEdge()->SetVertex(pvx2);
      //splice this into the left edge's vertex loop
      Splice(1,peLeft->GetEdge()->GetEdgePrev(1),peBot);
      peBot->SetFacet(pfa);
      //bottom edge is boundary edge
      pme->m_plistQEEdge->Remove(peBot->GetEdge());
      pme->m_plistQEEdgeBoundary->InsertLast(peBot->GetEdge());      

      //make right
      peRight = pme->CreateEdge();
      ASSERT(peRight != NULL);
      peRight->SetVertex(pvx2);
      peRight->GetEdge()->SetVertex(pvx);
      //splice this into the top and bot edge's vertex loop
      Splice(1,peTop,peRight->GetEdge());
      Splice(1,peBot->GetEdge(),peRight);
      peRight->SetFacet(pfa);

      peLeft=peRight->GetEdge();
    } 
  //last edge a boundary edge
  pme->m_plistQEEdge->Remove(peLeft);
  pme->m_plistQEEdgeBoundary->InsertLast(peLeft);      
  if (EqualEdge(peLeft,peFirst->GetEdge())) 
    {
      pme->MergeEdge(peLeft,peFirst->GetEdge(),pme);
    }
  MergeMesh(pme);
}

VOID QEMesh::MergeQuadStrip(CObjectList<QEVertex> *plistQEVertex,
			    CSLIDESurface *pCSLIDESurface)
{
  ASSERT( plistQEVertex != NULL );
  ASSERT( plistQEVertex->GetLength() > 2 );

  QEMesh *pme;
  QEFacet *pfa;
  QEVertex *pvx, *pvx2;
  QEEdge *pe, *peFirst, *peLeft, *peTop, *peBot, *peRight;
  CObjectListIter<QEVertex> iterQEVertex;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();
  pme->m_plistQEVertex->Splice(plistQEVertex);

  iterQEVertex.Init(pme->m_plistQEVertex);
  pvx = iterQEVertex.PeekFirst();
  pvx2 = iterQEVertex.PeekNext();
  
  pe = pme->CreateEdge();
  ASSERT(pe != NULL);
  pe->GetEdge()->SetVertex(pvx);
  pe->SetVertex(pvx2);
  //first edge is a boundary edge, for now... merge later if necessary
  pme->m_plistQEEdge->Remove(pe);
  pme->m_plistQEEdgeBoundary->InsertLast(pe);
  peFirst = pe->GetEdge();
  peLeft = peFirst;
  //now the first edge has been built

  for ( pvx = iterQEVertex.PeekNext();
	pvx != NULL;
	pvx = iterQEVertex.PeekNext() )
    {
      pvx2 = iterQEVertex.PeekNext();
      ASSERT(pvx2 != NULL);

      //first set up facet to left edge
      pfa = pme->CreateFacet();
      ASSERT(pfa != NULL);
      pfa->SetSurface(pCSLIDESurface);
      peLeft->SetFacet(pfa);

      //make top
      peTop = pme->CreateEdge();
      ASSERT(peTop != NULL);
      peTop->SetVertex(pvx);
      peTop->GetEdge()->SetVertex(peLeft->GetVertex());
      //splice this into the left edge's vertex loop
      Splice(1,peLeft,peTop->GetEdge());
      peTop->SetFacet(pfa);
      //top edge is boundary edge
      pme->m_plistQEEdge->Remove(peTop->GetEdge());
      pme->m_plistQEEdgeBoundary->InsertLast(peTop->GetEdge());      

      //make bottom
      peBot = pme->CreateEdge();
      ASSERT(peBot != NULL);
      peBot->SetVertex(peLeft->GetEdge()->GetVertex());
      peBot->GetEdge()->SetVertex(pvx2);
      //splice this into the left edge's vertex loop
      Splice(1,peLeft->GetEdge()->GetEdgePrev(1),peBot);
      peBot->SetFacet(pfa);
      //bottom edge is boundary edge
      pme->m_plistQEEdge->Remove(peBot->GetEdge());
      pme->m_plistQEEdgeBoundary->InsertLast(peBot->GetEdge());      

      //make right
      peRight = pme->CreateEdge();
      ASSERT(peRight != NULL);
      peRight->SetVertex(pvx2);
      peRight->GetEdge()->SetVertex(pvx);
      //splice this into the top and bot edge's vertex loop
      Splice(1,peTop,peRight->GetEdge());
      Splice(1,peBot->GetEdge(),peRight);
      peRight->SetFacet(pfa);

      peLeft=peRight->GetEdge();
    } 
  //last edge a boundary edge
  pme->m_plistQEEdge->Remove(peLeft);
  pme->m_plistQEEdgeBoundary->InsertLast(peLeft);      
  if (EqualEdge(peLeft, peFirst->GetEdge())) 
    {
      pme->MergeEdge(peLeft,peFirst->GetEdge(),pme);
    }
  MergeMesh(pme);
}

VOID QEMesh::SpliceMesh(QEMesh *pme)
{
  ASSERT( pme != NULL );

  //QEMeshIter iterQEMesh;

  // Transfer all of the new geometry into this mesh
  m_plistQEVertex->Splice(pme->m_plistQEVertex);
  m_plistQEWedge->Splice(pme->m_plistQEWedge);
  m_plistQEFacet->Splice(pme->m_plistQEFacet);
  m_plistQEEdge->Splice(pme->m_plistQEEdge);
  m_plistQEEdgeBoundary->Splice(pme->m_plistQEEdgeBoundary);

  // Destroy the merged in mesh
  DeallocateQEMesh(pme);
}

VOID QEMesh::MergeMesh(QEMesh *pme)
{
  ASSERT( pme != NULL );

  //QEMeshIter iterQEMesh;

  //fprintf(stderr, "START: QEMesh::MergeMesh(%p, %p)\n", this, pme);

  //fprintf(stderr, "  START: Current %p\n", this);
  //this->Display();
  //fprintf(stderr, "  END  : Current %p\n", this);

  //fprintf(stderr, "  START: Addition %p\n", pme);
  //pme->Display();
  //fprintf(stderr, "  END  : Addition %p\n", pme);

  // Transfer all of the new geometry except the edges into this mesh
  m_plistQEVertex->Splice(pme->m_plistQEVertex);
  m_plistQEWedge->Splice(pme->m_plistQEWedge);
  m_plistQEFacet->Splice(pme->m_plistQEFacet);

  MergeMesh(0, pme);
  //fprintf(stderr, "MERGED MESH: EXACT VERTICES\n");
  //DisplayEuler();
  //pme->DisplayEuler();
  MergeMesh(1, pme);

  // Transfer all of the edges into this mesh
  m_plistQEEdge->Splice(pme->m_plistQEEdge);
  m_plistQEEdgeBoundary->Splice(pme->m_plistQEEdgeBoundary);

  //fprintf(stderr, "MERGED MESH: EPSILON VERTICES\n");
  //DisplayEuler();
  //Display();

  // Destroy the merged in mesh
  DeallocateQEMesh(pme);

  //fprintf(stderr, "  START: Result %p\n", this);
  //this->Display();
  //fprintf(stderr, "  END  : Result %p\n", this);

  //fprintf(stderr, "END  : QEMesh::MergeMesh(%p, %p)\n", this, pme);
}

VOID QEMesh::MergeMesh(QEMesh *pme, CSLIDESurface *pCSLIDESurface)
{
  ASSERT( pme != NULL );

  QEMeshIter iterQEMesh;
  QEFacet *pfa;

  // Color 
  iterQEMesh.Init(pme);
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      pfa->SetSurface(pCSLIDESurface);
    }

  // Transfer all of the new geometry except the edges into this mesh
  m_plistQEVertex->Splice(pme->m_plistQEVertex);
  m_plistQEWedge->Splice(pme->m_plistQEWedge);
  m_plistQEFacet->Splice(pme->m_plistQEFacet);

  MergeMesh(0, pme);
  //fprintf(stderr, "MERGED MESH: EXACT VERTICES\n");
  //DisplayEuler();
  //pme->DisplayEuler();
  MergeMesh(1, pme);

  // Transfer all of the edges into this mesh
  m_plistQEEdge->Splice(pme->m_plistQEEdge);
  m_plistQEEdgeBoundary->Splice(pme->m_plistQEEdgeBoundary);

  //fprintf(stderr, "MERGED MESH: EPSILON VERTICES\n");
  //DisplayEuler();
  //Display();

  // Destroy the merged in mesh
  DeallocateQEMesh(pme);
}

VOID QEMesh::MergeMesh(UINT_32 uiSearchType, QEMesh *pme)
{
  if ( m_plistQEEdgeBoundary->GetLength() > 0 &&
       pme->m_plistQEEdgeBoundary->GetLength() > 0 )
    {
      CObjectListIter<QEEdge> iterQEEdge;
      QEEdge *pe, *peMerge, *peNext;
      Vector v;
      FLOAT fEpsilon, fMag;

      fEpsilon = INFINITY;

      iterQEEdge.Init(m_plistQEEdge);
      for ( pe = iterQEEdge.PeekFirst();
	    pe != NULL; 
	    pe = iterQEEdge.PeekNext() )
	{
	  v = pe->GetEdge()->GetVertex()->GetPoint() - pe->GetVertex()->GetPoint();
	  fMag = v.Magnitude();
	  fEpsilon = MIN(fMag, fEpsilon);
	}

      iterQEEdge.Init(pme->m_plistQEEdge);
      for ( pe = iterQEEdge.PeekFirst();
	    pe != NULL; 
	    pe = iterQEEdge.PeekNext() )
	{
	  v = pe->GetEdge()->GetVertex()->GetPoint() - pe->GetVertex()->GetPoint();
	  fMag = v.Magnitude();
	  fEpsilon = MIN(fMag, fEpsilon);
	}

      fEpsilon *= m_fMergeEpsilon;

      iterQEEdge.Init(pme->m_plistQEEdgeBoundary);
      for ( pe = iterQEEdge.PeekFirst();
	    pe != NULL; 
	    pe = peNext )
	{
	  peNext = iterQEEdge.PeekNext();

	  peMerge = SearchEdge(uiSearchType, pe, fEpsilon);
	  if ( peMerge != NULL )
	    {
	      MergeEdge(pe, peMerge, pme);
	    }
	}
    }
}

QEEdge *QEMesh::SearchEdge(UINT_32 uiSearchType, QEEdge *pe, FLOAT fEpsilon)
{
  CObjectListIter<QEEdge> iterQEEdge;
  QEEdge *peResult;

  iterQEEdge.Init(m_plistQEEdgeBoundary);
  peResult = NULL;
#ifndef COMPILER_PROBLEM
  switch ( uiSearchType )
    {
    case 0:
      {
	for ( peResult = iterQEEdge.PeekFirst();
	      peResult != NULL;
	      peResult = iterQEEdge.PeekNext() )
	  {
	    if ( ( pe->GetVertex() == peResult->GetEdge()->GetVertex() || 
		   *(pe->GetVertex()) == *(peResult->GetEdge()->GetVertex()) ) &&
		 ( pe->GetEdge()->GetVertex() == peResult->GetVertex() ||
		   *(pe->GetEdge()->GetVertex()) == *(peResult->GetVertex()) ) )
	      {
		break;
	      }
	  }
      }
      break;

    case 1:
      {
	Point pppt[2][2];
	Vector pv[2];

	pppt[0][0] = pe->GetVertex()->GetPoint();
	pppt[0][1] = pe->GetEdge()->GetVertex()->GetPoint();
	for ( peResult = iterQEEdge.PeekFirst();
	      peResult != NULL;
	      peResult = iterQEEdge.PeekNext() )
	  {
	    pppt[1][1] = peResult->GetVertex()->GetPoint();
	    pppt[1][0] = peResult->GetEdge()->GetVertex()->GetPoint();
	    pv[0] = pppt[1][0] - pppt[0][0];
	    pv[1] = pppt[1][1] - pppt[0][1];
	    if ( ( pe->GetVertex() == peResult->GetEdge()->GetVertex() || 
		   ( -fEpsilon <= pv[0][0] && pv[0][0] <= fEpsilon &&
		     -fEpsilon <= pv[0][1] && pv[0][1] <= fEpsilon &&
		     -fEpsilon <= pv[0][2] && pv[0][2] <= fEpsilon ) ) &&
		 ( pe->GetEdge()->GetVertex() == peResult->GetVertex() || 
		   ( -fEpsilon <= pv[1][0] && pv[1][0] <= fEpsilon &&
		     -fEpsilon <= pv[1][1] && pv[1][1] <= fEpsilon &&
		     -fEpsilon <= pv[1][2] && pv[1][2] <= fEpsilon ) ) )
	      {
		break;
	      }
	  }
      }
      break;

    default:
      ASSERT( FALSE );
      break;
    }
#endif // COMPILER_PROBLEM

  return peResult;
}

//////////////////////////////////////////////////
// CreateCleanMesh
//   This attempts to clean up meshes sent into the slide mesh construct
//
QEMesh *QEMesh::CreateCleanMesh()
{
  CHeap<QEVertex*, UINT_32> heapQEVertex;
  CHeapEntry<QEVertex*, UINT_32> *phe;
  QEMesh *pme;
  QEMeshIter iterQEMesh;
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peCurr, *peTemp;
  QEVertex *pvx, *pvxNext, *pvxNew;
  Point pt;
  UINT_32 uiValence, uiBoundaryEdges;
  BOOL b;

  pme = Copy();
  ASSERT( pme != NULL );

  fprintf(stderr, "START: QEMesh::CreateCleanMesh\n");

  heapQEVertex.Init((3*pme->GetVertexCount())/2);

  iterQEMesh.Init(pme);

  //////////////////////////////////////////////////
  // Duplicate vertices as necessary
  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      b = FALSE;
      pvx = pe->GetVertex();
      iterQEVertex.Init(pvx);
      for ( peCurr = iterQEVertex.PeekFirstEdge();
	    peCurr != NULL;
	    peCurr = iterQEVertex.PeekNextEdge() )
	{
	  if ( pe == peCurr )
	    {
	      b = TRUE;
	      break;
	    }
	}
      if ( !b )
	{
	  pt = pvx->GetPoint();
	  pvxNew = pme->CreateVertex(pt);
	  //fprintf(stderr, "  Duplicating vertex %p to %p ( %f %f %f )\n", pvx, pvxNew, pt(0), pt(1), pt(2));
	  pe->SetVertex(pvxNew);
	  for ( peCurr = pe->GetEdgeNext(1);
		peCurr != pe;
		peCurr = peCurr->GetEdgeNext(1) )
	    {
	      peCurr->SetVertex(pvxNew);
	    }
	}
    }

  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdgeBoundary() )
    {
      b = FALSE;
      pvx = pe->GetVertex();
      iterQEVertex.Init(pvx);
      for ( peCurr = iterQEVertex.PeekFirstEdge();
	    peCurr != NULL;
	    peCurr = iterQEVertex.PeekNextEdge() )
	{
	  if ( pe == peCurr )
	    {
	      b = TRUE;
	      break;
	    }
	}
      if ( !b )
	{
	  pt = pvx->GetPoint();
	  pvxNew = pme->CreateVertex(pt);
	  //fprintf(stderr, "  BE: Duplicating vertex %p to %p ( %f %f %f )\n", pvx, pvxNew, pt(0), pt(1), pt(2));
	  pe->SetVertex(pvxNew);
	  for ( peCurr = pe->GetEdgeNext(1);
		peCurr != pe;
		peCurr = peCurr->GetEdgeNext(1) )
	    {
	      peCurr->SetVertex(pvxNew);
	    }
	}
    }

  //////////////////////////////////////////////////
  // Split components that share a single edge
  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      for ( peCurr = pe->GetEdgeNext(1);
	    peCurr != pe;
	    peCurr = peCurr->GetEdgeNext(1) )
	{
	  if ( pe->GetEdge()->GetVertex() == peCurr->GetEdge()->GetVertex() )
	    {
	      // Assuming that this a single edge shared by two components
	      // THIS IS NOT ALWAYS A VALID ASSUMPTION!!!!!

	      pvx = pe->GetVertex();
	      pt = pvx->GetPoint();
	      pvxNew = pme->CreateVertex(pt);
	      //fprintf(stderr, "  Split duplicating vertex %p to %p ( %f %f %f )\n", pvx, pvxNew, pt(0), pt(1), pt(2));

	      Splice(1, pe, peCurr);
	      Splice(1, pe->GetEdge()->GetEdgePrev(1), peCurr->GetEdge()->GetEdgePrev(1));
	      pe->SetFacet(pe->GetEdgePrev(0)->GetFacet());
	      peCurr->SetFacet(peCurr->GetEdgePrev(0)->GetFacet());

	      peTemp = pe;
	      do
		{
		  peTemp->SetVertex(pvx);
		  peTemp = peTemp->GetEdgeNext(1);
		}
	      while ( peTemp != pe );
	      peTemp = peCurr;
	      do
		{
		  peTemp->SetVertex(pvxNew);
		  peTemp = peTemp->GetEdgeNext(1);
		}
	      while ( peTemp != peCurr );

	      pe = pe->GetEdge();
	      peCurr = peCurr->GetEdge();

	      pvx = pe->GetVertex();
	      pt = pvx->GetPoint();
	      pvxNew = pme->CreateVertex(pt);
	      //fprintf(stderr, "  Split duplicating vertex %p to %p ( %f %f %f )\n", pvx, pvxNew, pt(0), pt(1), pt(2));
	      peTemp = pe;
	      do
		{
		  peTemp->SetVertex(pvx);
		  peTemp = peTemp->GetEdgeNext(1);
		}
	      while ( peTemp != pe );
	      peTemp = peCurr;
	      do
		{
		  peTemp->SetVertex(pvxNew);
		  peTemp = peTemp->GetEdgeNext(1);
		}
	      while ( peTemp != peCurr );

	      break;
	    }
	}
    }

  //////////////////////////////////////////////////
  // Remove interior valence 2 vertices
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge(), uiValence = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), uiValence++ )
	{
	}
      phe = heapQEVertex.Insert(uiValence, pvx);
      pvx->SetHeapEntry((CHeapEntry<VOID*,FLOAT>*)(VOID*)phe);
    }

  for ( pvx = heapQEVertex.RemoveTop();
	pvx != NULL;
	pvx = heapQEVertex.RemoveTop() )
    {
      pvx->SetHeapEntry(NULL);

      uiBoundaryEdges = 0;
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge(), uiValence = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), uiValence++ )
	{
	  if ( pe->GetFacet() == QEMesh::GetNullFacet() || 
	       pe->GetEdge()->GetFacet() == QEMesh::GetNullFacet() )
	    {
	      uiBoundaryEdges++;
	    }
	}

      switch ( uiValence )
	{
	case 0:
	  pt = pvx->GetPoint();
	  //fprintf(stderr, "  Removing unused vertex %p ( %f %f %f )\n", pvx, pt(0), pt(1), pt(2));
	  pme->DestroyVertex(pvx);
	  break;
	case 1:
	  ASSERT( FALSE );
	  break;
	case 2:
	  if ( uiBoundaryEdges < 2 )
	    {
	      pt = pvx->GetPoint();
	      //fprintf(stderr, "  START: Removing valence 2 vertex %p ( %f %f %f )\n", pvx, pt(0), pt(1), pt(2));

	      /////////////////////////////////////////////////
	      // Remove the vertex and one of the two edges
	      pe = iterQEVertex.PeekFirstEdge();
	      peCurr = pe->GetEdgeNext(1);
	      ASSERT( peCurr->GetEdgeNext(1) == pe &&
		      peCurr != pe );
	      Splice(1, pe, peCurr);
	      Splice(1, peCurr->GetEdge(), pe);
	      pe->SetVertex(peCurr->GetEdge()->GetVertex());
	      pe->SetFacet(pe->GetFacet());
	      pe->GetEdge()->SetFacet(pe->GetEdge()->GetFacet());
	      Splice(1, peCurr->GetEdge(), peCurr->GetEdge()->GetEdgePrev(1));

	      pvxNext = peCurr->GetVertex();
	      phe = (CHeapEntry<QEVertex*,UINT_32>*)(VOID*)(pvxNext->GetHeapEntry());
	      if ( phe != NULL )
		{
		  heapQEVertex.Remove(phe);
		}
	      pme->DestroyVertex(pvxNext);
	      pme->DestroyEdge(peCurr->GetEdge());
	      pme->DestroyEdge(peCurr);

	      //////////////////////////////
	      // Possibly destroy the facet
	      peCurr = pe->GetEdgeNext(0); 
	      ASSERT( peCurr != pe );
	      if ( peCurr->GetEdgeNext(0) == pe )
		{
		  //fprintf(stderr, "    Destroy left facet, pe = %p\n", pe);
		  pe->GetEdgePrev(1)->SetVertex(pe->GetEdgePrev(1)->GetVertex());
		  pe->GetEdge()->GetEdgePrev(1)->SetVertex(pe->GetEdge()->GetEdgePrev(1)->GetVertex());
		  peCurr->SetFacet(pe->GetEdge()->GetFacet());
		  Splice(1, pe, pe->GetEdgePrev(1));
		  Splice(1, pe->GetEdge(), pe->GetEdge()->GetEdgePrev(1));
		  pme->DestroyFacet(pe->GetFacet());
		  pme->DestroyEdge(pe->GetEdge());
		  pme->DestroyEdge(pe);

		  pe = peCurr;

		  pvxNext = pe->GetVertex();
		  iterQEVertex.Init(pvxNext);
		  for ( peCurr = iterQEVertex.PeekFirstEdge(), uiValence = 0;
			peCurr != NULL;
			peCurr = iterQEVertex.PeekNextEdge(), uiValence++ )
		    {
		    }
		  phe = (CHeapEntry<QEVertex*,UINT_32>*)(VOID*)(pvxNext->GetHeapEntry());
		  if ( phe != NULL )
		    {
		      heapQEVertex.Update(phe, uiValence);
		    }
		  else
		    {
		      phe = heapQEVertex.Insert(uiValence, pvxNext);
		      pvxNext->SetHeapEntry((CHeapEntry<VOID*,FLOAT>*)(VOID*)phe);
		    }

		  pvxNext = pe->GetEdge()->GetVertex();
		  iterQEVertex.Init(pvxNext);
		  for ( peCurr = iterQEVertex.PeekFirstEdge(), uiValence = 0;
			peCurr != NULL;
			peCurr = iterQEVertex.PeekNextEdge(), uiValence++ )
		    {
		    }
		  phe = (CHeapEntry<QEVertex*,UINT_32>*)(VOID*)(pvxNext->GetHeapEntry());
		  if ( phe != NULL )
		    {
		      heapQEVertex.Update(phe, uiValence);
		    }
		  else
		    {
		      phe = heapQEVertex.Insert(uiValence, pvxNext);
		      pvxNext->SetHeapEntry((CHeapEntry<VOID*,FLOAT>*)(VOID*)phe);
		    }
		}
	      else
		{
		  pe = pe->GetEdge();
		}

	      //////////////////////////////
	      // Possibly destroy the facet
	      peCurr = pe->GetEdgeNext(0); 
	      ASSERT( peCurr != pe );
	      if ( peCurr->GetEdgeNext(0) == pe )
		{
		  if ( pe->GetEdge() == peCurr )
		    {
		      //fprintf(stderr, "    Destroy right facet whole component, pe = %p\n", pe);
		      pme->DestroyFacet(pe->GetFacet());

		      pvxNext = pe->GetEdge()->GetVertex();
		      phe = (CHeapEntry<QEVertex*,UINT_32>*)(VOID*)(pvxNext->GetHeapEntry());
		      if ( phe != NULL )
			{
			  heapQEVertex.Remove(phe);
			}
		      pme->DestroyVertex(pvxNext);

		      pvxNext = pe->GetVertex();
		      phe = (CHeapEntry<QEVertex*,UINT_32>*)(VOID*)(pvxNext->GetHeapEntry());
		      if ( phe != NULL )
			{
			  heapQEVertex.Remove(phe);
			}
		      pme->DestroyVertex(pvxNext);

		      pme->DestroyEdge(pe->GetEdge());
		      pme->DestroyEdge(pe);
		    }
		  else
		    {
		      ASSERT( !( pe->GetEdge()->GetFacet() == GetNullFacet() &&
				  peCurr->GetEdge()->GetFacet() == GetNullFacet() ) );
		      //fprintf(stderr, "    Destroy right facet, pe = %p\n", pe);
		      if ( pe->GetEdge()->GetFacet() == GetNullFacet() )
			{
			  pe = peCurr;
			  peCurr = pe->GetEdgeNext(0);
			  //fprintf(stderr, "      swap right facet, pe = %p\n", pe);
			}
		      pe->GetEdgePrev(1)->SetVertex(pe->GetEdgePrev(1)->GetVertex());
		      pe->GetEdge()->GetEdgePrev(1)->SetVertex(pe->GetEdge()->GetEdgePrev(1)->GetVertex());
		      peCurr->SetFacet(pe->GetEdge()->GetFacet());
		      Splice(1, pe, pe->GetEdgePrev(1));
		      Splice(1, pe->GetEdge(), pe->GetEdge()->GetEdgePrev(1));
		      pme->DestroyFacet(pe->GetFacet());
		      pme->DestroyEdge(pe->GetEdge());
		      pme->DestroyEdge(pe);

		      pe = peCurr;
		      
		      pvxNext = pe->GetVertex();
		      iterQEVertex.Init(pvxNext);
		      for ( peCurr = iterQEVertex.PeekFirstEdge(), uiValence = 0;
			    peCurr != NULL;
			    peCurr = iterQEVertex.PeekNextEdge(), uiValence++ )
			{
			}
		      phe = (CHeapEntry<QEVertex*,UINT_32>*)(VOID*)(pvxNext->GetHeapEntry());
		      if ( phe != NULL )
			{
			  heapQEVertex.Update(phe, uiValence);
			}
		      else
			{
			  phe = heapQEVertex.Insert(uiValence, pvxNext);
			  pvxNext->SetHeapEntry((CHeapEntry<VOID*,FLOAT>*)(VOID*)phe);
			}
		      
		      pvxNext = pe->GetEdge()->GetVertex();
		      iterQEVertex.Init(pvxNext);
		      for ( peCurr = iterQEVertex.PeekFirstEdge(), uiValence = 0;
			    peCurr != NULL;
			    peCurr = iterQEVertex.PeekNextEdge(), uiValence++ )
			{
			}
		      phe = (CHeapEntry<QEVertex*,UINT_32>*)(VOID*)(pvxNext->GetHeapEntry());
		      if ( phe != NULL )
			{
			  heapQEVertex.Update(phe, uiValence);
			}
		      else
			{
			  phe = heapQEVertex.Insert(uiValence, pvxNext);
			  pvxNext->SetHeapEntry((CHeapEntry<VOID*,FLOAT>*)(VOID*)phe);
			}
		    }
		}
	      //fprintf(stderr, "  END  : Removing valence 2 vertex %p ( %f %f %f )\n", pvx, pt(0), pt(1), pt(2));
	    }
	  break;
	default:
	  break;
	}     
    }

  pme->Preprocess();
  
  fprintf(stderr, "END  : QEMesh::CreateCleanMesh\n");

  return pme;
}

/*
VOID QEMesh::Render()
{
  RenderGeometry();
}

VOID QEMesh::RenderGeometry()
{
  QEMeshIter iter;
  QEFacet *pfa;

  iter.Init(this);
  for ( pfa = iter.PeekFirstFacet();
	pfa != NULL;
	pfa = iter.PeekNextFacet() )
    {
      pfa->RenderGeometry();
    }

#ifdef RENDER_VORONOI
  QEEdge *pe;
  Point ptCenter, pt;
  Vector v;

  glDisable(GL_LIGHTING);
  glColor3f(0.0, 0.0, 0.0);
  glBegin(GL_LINES);
  {
    for ( pe = iter.PeekFirstEdge();
	  pe != NULL;
	  pe = iter.PeekNextEdge() )
      {
	if ( pe->GetFacet() == m_pfaNULL )
	  {
	    pe = pe->GetEdge();
	    ASSERT( pe->GetFacet() != m_pfaNULL );
	  }
	ptCenter = pe->GetFacet()->GetCircumcenter();

	if ( pe->GetEdge()->GetFacet() == m_pfaNULL )
	  {
	    pt = ( pe->GetVertex()->GetPoint() + 
		   pe->GetNeighborVertex()->GetPoint() ) / 2.0; 
	    v = pt - ptCenter;
	    v.Normalize();
	    if ( RightOfEdge(pe, ptCenter) )
	      {
		v *= -100.0;
	      }
	    else
	      {
		v *= 100.0;
	      }
	    pt = ptCenter + v;
	  }
	else
	  {
	    pt = pe->GetEdge()->GetFacet()->GetCircumcenter();
	  }

	gilVertex3fv(ptCenter.Get());
	gilVertex3fv(pt.Get());
      }
  }
  glEnd();
#endif // RENDER_VORONOI
}

VOID QEMesh::RenderNormals()
{
  QEMeshIter iter;
  QEFacet *pfa;

  iter.Init(this);
  for ( pfa = iter.PeekFirstFacet();
	pfa != NULL;
	pfa = iter.PeekNextFacet() )
    {
      pfa->RenderNormals();
    }
}
*/
VOID QEMesh::Display()
{
  QEMeshIter iter;
  QEVertex *pvxCurr;
  QEEdge *peCurr;
  QEFacet *pfaCurr;

  iter.Init(this);

  fprintf(stderr, "    START: Mesh %p\n", this);

  DisplayEuler();

  fprintf(stderr, "      START: Vertices\n");
  
  for ( pvxCurr = iter.PeekFirstVertex();
	pvxCurr != NULL;
	pvxCurr = iter.PeekNextVertex() )
    {
      fprintf(stderr, "    ");
      pvxCurr->Display();
    }
  fprintf(stderr, "      END:   Vertices\n");

  fprintf(stderr, "      START: Edges\n");

  for ( peCurr = iter.PeekFirstEdge();
	peCurr != NULL;
	peCurr = iter.PeekNextEdge() )
    {
      fprintf(stderr, "    ");
      peCurr->Display();
    }
  fprintf(stderr, "      END:   Edges\n");

  fprintf(stderr, "      START: Facets\n");
  iter.Init(this);
  
  for ( pfaCurr = iter.PeekFirstFacet();
	pfaCurr != NULL;
	pfaCurr = iter.PeekNextFacet() )
    {
      fprintf(stderr, "    ");
      pfaCurr->Display();
    }
  fprintf(stderr, "      END:   Facets\n");

  fprintf(stderr, "    END  : Mesh %p\n", this);
}

VOID QEMesh::DisplayEuler()
{
  fprintf(stderr, "      START: Mesh Euler Info: %p\n", this);

  fprintf(stderr, "        Euler's Formula:  V - EU/2 + (F + 1) = 2\n");
  fprintf(stderr, "                          %ld - %ld/2 + (%ld + 1) = %ld\n", 
	 m_plistQEVertex->GetLength(),
	 m_plistQEEdge->GetLength() + m_plistQEEdgeBoundary->GetLength(),
	 m_plistQEFacet->GetLength(),
	 m_plistQEVertex->GetLength() - 
	 (m_plistQEEdge->GetLength() + m_plistQEEdgeBoundary->GetLength())/2 + 
	 m_plistQEFacet->GetLength() + 1);
  fprintf(stderr, "        Boundary Edges = %ld\n", m_plistQEEdgeBoundary->GetLength());

  fprintf(stderr, "      END  : Mesh Euler Info: %p\n", this);
}

//////////////////////////////////////////////////////////////////////
// QEMeshIter Class
//

QEMeshIter::QEMeshIter()
{
  m_pQEMesh = NULL;
}

QEMeshIter::~QEMeshIter()
{
  Uninit();
}

VOID QEMeshIter::Init(QEMesh *pQEMesh)
{
  ASSERT( pQEMesh != NULL );

  m_iterQEVertex.Init(pQEMesh->m_plistQEVertex);
  m_iterQEWedge.Init(pQEMesh->m_plistQEWedge);
  m_iterQEEdge.Init(pQEMesh->m_plistQEEdge);
  m_iterQEEdgeBoundary.Init(pQEMesh->m_plistQEEdgeBoundary);
  m_iterQEFacet.Init(pQEMesh->m_plistQEFacet);

  m_pQEMesh = pQEMesh;
}

VOID QEMeshIter::Uninit()
{
  m_pQEMesh = NULL;
}

BOOL QEMeshIter::Initialized()
{
  return TRUE;
}

BOOL QEMeshIter::Valid()
{
  return Initialized();
}

QEVertex *QEMeshIter::PeekFirstVertex()
{
  return m_iterQEVertex.PeekFirst();
}

QEVertex *QEMeshIter::PeekNextVertex()
{
  return m_iterQEVertex.PeekNext();
}

QEVertex *QEMeshIter::PeekLastVertex()
{
  return m_iterQEVertex.PeekLast();
}

QEVertex *QEMeshIter::PeekPrevVertex()
{
  return m_iterQEVertex.PeekPrev();
}

QEWedge *QEMeshIter::PeekFirstWedge()
{
  return m_iterQEWedge.PeekFirst();
}

QEWedge *QEMeshIter::PeekNextWedge()
{
  return m_iterQEWedge.PeekNext();
}

QEWedge *QEMeshIter::PeekLastWedge()
{
  return m_iterQEWedge.PeekLast();
}

QEWedge *QEMeshIter::PeekPrevWedge()
{
  return m_iterQEWedge.PeekPrev();
}

QEEdge *QEMeshIter::PeekFirstEdge()
{
  return m_iterQEEdge.PeekFirst();
}

QEEdge *QEMeshIter::PeekNextEdge()
{
  return m_iterQEEdge.PeekNext();
}

QEEdge *QEMeshIter::PeekLastEdge()
{
  return m_iterQEEdge.PeekLast();
}

QEEdge *QEMeshIter::PeekPrevEdge()
{
  return m_iterQEEdge.PeekPrev();
}

QEEdge *QEMeshIter::PeekFirstEdgeBoundary()
{
  return m_iterQEEdgeBoundary.PeekFirst();
}

QEEdge *QEMeshIter::PeekNextEdgeBoundary()
{
  return m_iterQEEdgeBoundary.PeekNext();
}

QEEdge *QEMeshIter::PeekLastEdgeBoundary()
{
  return m_iterQEEdgeBoundary.PeekLast();
}

QEEdge *QEMeshIter::PeekPrevEdgeBoundary()
{
  return m_iterQEEdgeBoundary.PeekPrev();
}

QEFacet *QEMeshIter::PeekFirstFacet()
{
  return m_iterQEFacet.PeekFirst();
}

QEFacet *QEMeshIter::PeekNextFacet()
{
  return m_iterQEFacet.PeekNext();
}

QEFacet *QEMeshIter::PeekLastFacet()
{
  return m_iterQEFacet.PeekLast();
}

QEFacet *QEMeshIter::PeekPrevFacet()
{
  return m_iterQEFacet.PeekPrev();
}


