#include "hashtable.h"
#include "qemesh.h"

#include "qenode.h"
#include "qeedgecollapse.h"

#include "qevertex.h"
#include "qewedge.h"
#include "qeedge.h"
#include "qefacet.h"

#include "apm2d.h"

//////////////////////////////////////////////////
// Temporary Fix to infinite edge problem

static QEMesh *spmeJordan = NULL;

//#define ENABLE_VORONOI_POLYGON_CLEANING_UP_CONTOUR
//#define ENABLE_VORONOI_POLYGON_DEBUGGING
//#define ENABLE_VORONOI_POLYGON_DEBUGGING_PROJECTION
//#define ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
//#define ENABLE_VORONOI_POLYGONS_DEBUGGING

//////////////////////////////////////////////////////////////////////
// START: Movie stuff
//#define ENABLE_VORONOI_POLYGON_DEBUGGING_OUTPUT_FILES
//#define ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
static UINT_32 suiDebugFile = 0;
//static const FLOAT cfInfinityMagnitude = 0.2;
//static const FLOAT cfInfinityMagnitude = 3.0;
static const FLOAT cfInfinityMagnitude = 10.0;

static VOID OutputDebugVoronoiMesh(QEMesh *pme)
{
  ASSERT( pme != NULL );

  CHAR spcDebugFile[32];
  FILE *fpDebug;
  
  sprintf(spcDebugFile, "debug%ld.slf", suiDebugFile);
  
  fpDebug = fopen(spcDebugFile, "w");
  ASSERT( fpDebug != 0 );
  pme->WriteDebugVoronoiPolygonShewchuk(fpDebug);
  fclose(fpDebug);
  suiDebugFile++;
}
// END  : Movie stuff
//////////////////////////////////////////////////////////////////////

#define ENABLE_VORONOI_POLYGON_CHECK_EDGE_ZONES
//#define ENABLE_VORONOI_POLYGON_CHECK_VERTEX_ZONES

//static QEVertex *s_pvxOpenEndpoint = NULL;
//static Point s_ptOpenEndpoint;
//static BOOL s_bOpenEndpoint;
static const FLOAT s_cfDistanceEpsilon = 0.0001;
//static const FLOAT s_cfDistanceEpsilon = 0.00001;
//static const FLOAT s_cfDistanceEpsilon = 0.0000001;
//static const FLOAT s_cfVoronoiInfinity = 500.0;
//static const FLOAT s_cfVoronoiInfinity = 1000.0;
//static const FLOAT s_cfVoronoiInfinity = 3000.0;
static const FLOAT s_cfVoronoiInfinity = 800.0;
//static const FLOAT s_cfVoronoiInfinity = 640.0; // this was working before
//static const FLOAT s_cfVoronoiInfinity = 1000.0;
//static const FLOAT s_cfVoronoiInfinity = 10000.0;
//static const FLOAT s_cfVoronoiInfinity = 100000.0;
//static const FLOAT s_cfVoronoiInfinity = 1000000.0;

//static const FLOAT s_cfEpsilonEdgeZone = 0.00001;
static const FLOAT s_cfEpsilonEdgeZone = 0.000001;

static const FLOAT s_cfEpsilonVoronoiVVE = 0.0001;
static const FLOAT s_cfEpsilonVoronoiVEE = 0.0001;
static const FLOAT s_cfEpsilonVoronoiVVEAntiparallel = 0.00000001;
static const FLOAT s_cfEpsilonVoronoiVVEVertexOnEdge = 0.0000000001;
static const FLOAT s_cfEpsilonVoronoiVEEVertexOnEdge = 0.0000000001;
static const FLOAT s_cfEpsilonVoronoiVEEParallel = 0.0000000001;
static const FLOAT s_cfEpsilonVoronoiVEEAntiparallel = 0.00000001;
static const FLOAT s_cfEpsilonVoronoiVVEVertexZone = 0.0000001;
static const FLOAT s_cfEpsilonVoronoiVEEVertexZone = 0.0000001;
static const FLOAT s_cfEpsilonVoronoiZMonotone = 0.0001;

//static const FLOAT s_cfEpsilonCompareCenter = 0.000001;
//static const FLOAT s_cfEpsilonCompareCenter = 0.00001;
static const FLOAT s_cfEpsilonCompareCenter = 0.0001;
static const FLOAT s_cfInfinityCompareCenter = 100.0;

static const FLOAT s_cfNullEdgeEpsilon = 1e-3;
static const DOUBLE s_cdStraightnessEpsilon = 1e-2;

//////////////////////////////////////////////////////////////////////
// QEMesh Voronoi Diagram of polygonal chains (Shewchuk version)
//

static FLOAT CalculateContourArea(QEEdge *peStart)
{
  QEEdge *pe;
  QEVertex *pvx;
  Point pt, ptOrigin;
  Vector vCurr, vPrev;
  FLOAT fArea;

  // We can skip the first and the last edge because we are using the first vertex
  // as the origin, so those edges will have zero area triangles

  if ( peStart == NULL ||
       peStart->GetEdgePrev(0) == peStart )
    {
      return 0.0;
    }

  pe = peStart->GetEdgePrev(0);

  fArea = 0.0;

  pvx = peStart->GetVertex();
  ASSERT( pvx != NULL );
  ptOrigin = pvx->GetPoint();

  pvx = pe->GetVertex();
  ASSERT( pvx != NULL );
  pt = pvx->GetPoint();

  vPrev = pt - ptOrigin;

  for ( pe = pe->GetEdgePrev(0);
	pe != peStart;
	pe = pe->GetEdgePrev(0) )
    {
      pvx = pe->GetVertex();
      pt = pvx->GetPoint();
      
      vCurr = pt - ptOrigin;
      
      fArea += ( vPrev(0)*vCurr(1) - vPrev(1)*vCurr(0) );

      vPrev = vCurr;
    }

  fArea /= 2.0;

  return fArea;  
}

VOID QEMesh::VerifyVoronoiPolygonsShewchuk()
{
  QEMeshIter iterQEMesh;
  QEVertexIter iterQEVertex;
  QEFacetIter iterQEFacet;
  QEVertex *pvx;
  QEFacet *pfa;
  QEEdge *pe;
  BOOL b;

  iterQEMesh.Init(this);

  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  ASSERT( pe->GetVertex() == pvx );
	}
    }

  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      iterQEFacet.Init(pfa);
      b = FALSE;
      if ( pfa->GetForwardVertex() != NULL )
	{
	  ASSERT( pfa->GetForwardEdge() == NULL );
	  BOOL b2 = TRUE;
	  for ( pe = iterQEFacet.PeekFirstEdge();
		pe != NULL;
		pe = iterQEFacet.PeekNextEdge() )
	    {
	      if ( pe->GetFacet() != pfa )
		{
		  fprintf(stderr, "pe = %p --> pfa = %p  !=  %p\n", pe, pe->GetFacet(), pfa);
		  b2 = FALSE;
		}
	      //ASSERT( pe->GetFacet() == pfa );
	      b = ( b || pe->GetVertex() == pfa->GetForwardVertex() );
	    }
	  ASSERT( b2 );
	  ASSERT( b );
	}
      else
	{
	  ASSERT( pfa->GetForwardEdge() != NULL );
	  for ( pe = iterQEFacet.PeekFirstEdge();
		pe != NULL;
		pe = iterQEFacet.PeekNextEdge() )
	    {
	      ASSERT( pe->GetFacet() == pfa );
	      b = ( b || pe == pfa->GetForwardEdge() );
	    }
	  ASSERT( b );
	}
    }
}


VOID QEMesh::WriteDebugVoronoiPolygonShewchuk(FILE *fpOutput)
{
  QEMeshIter iterQEMesh;
  QEVertexIter iterQEVertex;
  QEVertex *pvx;
  QEEdge *pe, *peCurr;
  Point pt;
  Vector u, v;
  UINT_32 i, j, uiNull;

  iterQEMesh.Init(this);

  fprintf(fpOutput, "surface s%ld_Bisector color ( 1 0 0 ) endsurface\n", suiDebugFile);
  fprintf(fpOutput, "surface s%ld_Infinity color ( 1 1 1 ) endsurface\n", suiDebugFile);
  fprintf(fpOutput, "surface s%ld_InfinityError color ( 1 1 0 ) endsurface\n", suiDebugFile);
  fprintf(fpOutput, "surface s%ld_Contour color ( 0 0 1 ) endsurface\n", suiDebugFile);
  fprintf(fpOutput, "surface s%ld_ContourError color ( 0 0 1 ) endsurface\n", suiDebugFile);
  fprintf(fpOutput, "surface s%ld_Voronoi color ( 0 0 0 ) endsurface\n", suiDebugFile);
  fprintf(fpOutput, "surface s%ld_VoronoiError color ( 0 1 0 ) endsurface\n", suiDebugFile);

  fprintf(fpOutput, "point p%ld_NULL ( 0 0 0 ) endpoint\n\n", suiDebugFile);

  for ( pvx = iterQEMesh.PeekFirstVertex(), i = 0;
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex(), i++ )
    {
      pvx->SetID(i);
      pt = pvx->GetLimitPosition();
      /*
      pt = pvx->GetPoint();
      if ( pvx->GetInscribedCircleCenterType() == APM2D::icctInfinity )
	{
	  v = pt;
	  pt.Null();
	  iterQEVertex.Init(pvx);
	  for ( pe = iterQEVertex.PeekFirstEdge();
		pe != NULL;
		pe = iterQEVertex.PeekNextEdge() )
	    {
	      //if ( pe->GetEdge()->GetVertex()->GetInscribedCircleCenterType() != APM2D::icctInfinity )
	      if ( pe->GetEdge()->GetFacet() != NULL && 
		   pe->GetEdge()->GetFacet() != GetNullFacet() && 
		   pe->GetEdge()->GetEdge()->GetFacet() != NULL && 
		   pe->GetEdge()->GetEdge()->GetFacet() != GetNullFacet() )
		{
		  pt = pe->GetEdge()->GetVertex()->GetPoint();
		  pt += (0.01*v);
		  break;
		}
	    }
	}
      */
      fprintf(fpOutput, "point p%ld_%ld ( %f %f %f ) endpoint\n", 
	      suiDebugFile, i, pt(0), pt(1), pt(2));
    }

  fprintf(fpOutput, "\n");

  uiNull = 0;
  for ( pe = iterQEMesh.PeekFirstEdge(), i = 0;
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge(), i++ )
    {
      if ( pe->GetVertex() == NULL && pe->GetEdge()->GetVertex() == NULL )
	{
	  fprintf(fpOutput, "face f%ld_%ld ( p%ld_NULL p%ld_NULL p%ld_NULL )\n",
		  suiDebugFile, i, suiDebugFile, suiDebugFile, suiDebugFile);
	}
      else if ( pe->GetEdge()->GetVertex() == NULL )
	{
	  v.Null();
	  for ( peCurr = pe->GetEdgeNext(1);
		peCurr != pe;
		peCurr = peCurr->GetEdgeNext(1) )
	    {
	      if ( peCurr->GetEdge()->GetVertex() != NULL )
		{
		  u = peCurr->GetVertex()->GetLimitPosition() - peCurr->GetEdge()->GetVertex()->GetLimitPosition();
		  u[2] = 0.0;
		  u.Normalize();
		  v += u;
		}
	    }
	  v.Normalize();
	  v[2] = 1.0;
	  pt = pe->GetVertex()->GetLimitPosition() + cfInfinityMagnitude*v;
	  fprintf(fpOutput, "point p%ld_NULL_%ld ( %f %f %f ) endpoint\n", 
		  suiDebugFile, uiNull, pt(0), pt(1), pt(2));
	  fprintf(fpOutput, "face f%ld_%ld ( p%ld_%ld p%ld_%ld p%ld_NULL_%ld )\n",
		  suiDebugFile, i,
		  suiDebugFile, pe->GetVertex()->GetID(),
		  suiDebugFile, pe->GetVertex()->GetID(),
		  suiDebugFile, uiNull);
	  uiNull++;
	}
      else if ( pe->GetVertex() == NULL )
	{
	  v.Null();
	  for ( peCurr = pe->GetEdge()->GetEdgeNext(1);
		peCurr != pe->GetEdge();
		peCurr = peCurr->GetEdgeNext(1) )
	    {
	      if ( peCurr->GetEdge()->GetVertex() != NULL )
		{
		  u = peCurr->GetVertex()->GetLimitPosition() - peCurr->GetEdge()->GetVertex()->GetLimitPosition();
		  u[2] = 0.0;
		  u.Normalize();
		  v += u;
		}
	    }
	  v.Normalize();
	  v[2] = 1.0;
	  pt = pe->GetEdge()->GetVertex()->GetLimitPosition() + cfInfinityMagnitude*v;
	  fprintf(fpOutput, "point p%ld_NULL_%ld ( %f %f %f ) endpoint\n", 
		  suiDebugFile, uiNull, pt(0), pt(1), pt(2));
	  fprintf(fpOutput, "face f%ld_%ld ( p%ld_NULL_%ld p%ld_NULL_%ld p%ld_%ld )\n",
		  suiDebugFile, i,
		  suiDebugFile, uiNull, 
		  suiDebugFile, uiNull,
		  suiDebugFile, pe->GetEdge()->GetVertex()->GetID());
	  uiNull++;
	}
      else
	{
	  fprintf(fpOutput, "face f%ld_%ld ( p%ld_%ld p%ld_%ld p%ld_%ld )\n",
		  suiDebugFile, i,
		  suiDebugFile, pe->GetVertex()->GetID(),
		  suiDebugFile, pe->GetVertex()->GetID(),
		  suiDebugFile, pe->GetEdge()->GetVertex()->GetID());
	}

      if ( pe->GetVertex() == NULL || pe->GetEdge()->GetVertex() == NULL )
	{
	  fprintf(fpOutput, "  surface s%ld_Bisector\n", suiDebugFile);
	}
      else if ( pe->GetFacet() == NULL || pe->GetEdge()->GetFacet() == NULL )
	{
	  if ( pe->GetFacet() == GetNullFacet() || pe->GetEdge()->GetFacet() == GetNullFacet() )
	    {
	      fprintf(fpOutput, "  surface s%ld_InfinityError\n", suiDebugFile);
	    }
	  else if ( pe->GetVertex()->GetPoint()(3) == 0.0 &&
		    pe->GetEdge()->GetVertex()->GetPoint()(3) == 0.0 )
	    {
	      fprintf(fpOutput, "  surface s%ld_ContourError\n", suiDebugFile);
	    }
	  else
	    {
	      fprintf(fpOutput, "  surface s%ld_VoronoiError\n", suiDebugFile);
	    }
	}
      else 
	{
	  if ( pe->GetFacet() == GetNullFacet() || pe->GetEdge()->GetFacet() == GetNullFacet() )
	    {
	      fprintf(fpOutput, "  surface s%ld_Infinity\n", suiDebugFile);
	    }
	  else if ( pe->GetVertex()->GetPoint()(3) == 0.0 &&
		    pe->GetEdge()->GetVertex()->GetPoint()(3) == 0.0 )
	    {
	      fprintf(fpOutput, "  surface s%ld_Contour\n", suiDebugFile);
	    }
	  else
	    {
	      fprintf(fpOutput, "  surface s%ld_Voronoi\n", suiDebugFile);
	    }
	}

      fprintf(fpOutput, "endface\n");
    }

  fprintf(fpOutput, "\n");

  fprintf(fpOutput, "object oDebugVoronoi%ld (", suiDebugFile); 
  for ( j = 0; j < i; j++ )
    {
      fprintf(fpOutput, " f%ld_%ld", suiDebugFile, j); 
    }  
  fprintf(fpOutput, " )\n"); 
  fprintf(fpOutput, "  shading SLF_WIRE\n"); 
  fprintf(fpOutput, "endobject\n"); 
}

VOID QEMesh::CreateVoronoiPolygonShewchuk(QEMesh *pmeCDT, BOOL bCleanUpContour)
{
  QEEdge *pe, *peEndLoop, *peCurr, *pePrev, *peStart, *peNext;
  QEEdge **pppeContourSegments[2];
  QEMeshIter iter;
  QEVertex *pvx, *pvxNew;
  Point pt;
  UINT_32 i, uiSize, uiSegment;
  BOOL bConvex;
  FLOAT fArea, fW;
  DOUBLE d;

  ASSERT( pmeCDT != NULL );

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING_OUTPUT_FILES
  //  suiDebugFile = 0;
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING_OUTPUT_FILES

  //////////////////////////////////////////////////
  // Weird Static Variable Dependencies
  //s_bOpenEndpoint = FALSE;
  // Weird Static Variable Dependencies
  //////////////////////////////////////////////////

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "START: CreateVoronoiPolygonShewchuk\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

  uiSize = pmeCDT->GetBoundaryEdgeCount();
  ASSERT( uiSize > 0 );

  pppeContourSegments[0] = new QEEdge*[uiSize];
  ASSERT( pppeContourSegments[0] != NULL );
  pppeContourSegments[1] = new QEEdge*[uiSize];
  ASSERT( pppeContourSegments[1] != NULL );

  //////////////////////////////////////////////////
  // Determine the orientation of the contour

  pe = pmeCDT->m_plistQEEdgeBoundary->PeekFirst();
  ASSERT( pe != NULL );

  fArea = ::CalculateContourArea(pe);

  fW = 1.0;
  if ( fArea < 0.0 )
    {
      fW = -1.0;
    }

  //////////////////////////////////////////////////
  // Count the Boundary Edges

  //////////////////////////////////////////////////
  // Copy the contour

  peCurr = CreateEdge();
  pt = pe->GetVertex()->GetPoint();
  pt[2] = 0.0;
  pvxNew = CreateVertex(pt);
  pvxNew->SetLimitPosition(pt);
  pvxNew->SetInscribedCircleCenterType(APM2D::icctNull);
  peCurr->SetVertex(pvxNew);

  pePrev = peCurr;

  peStart = peCurr;
  peEndLoop = pmeCDT->m_plistQEEdgeBoundary->PeekFirst();
  pe = peEndLoop;
  if ( fW == 1.0 )
    {
      pe = pe->GetEdgePrev(0);
    }
  else
    {
      pe = pe->GetEdgeNext(0);
    }
  while ( pe != peEndLoop )
    {
      peCurr = CreateEdge();
      pt = pe->GetVertex()->GetPoint();
      pt[2] = 0.0;
      pvxNew = CreateVertex(pt);
      pvxNew->SetLimitPosition(pt);
      pvxNew->SetInscribedCircleCenterType(APM2D::icctNull);
      peCurr->SetVertex(pvxNew);

      peCurr->GetEdge()->SetVertex(pePrev->GetVertex());
      Splice(1, peCurr->GetEdge(), pePrev);
      
      pePrev = peCurr;

      if ( fW == 1.0 )
	{
	  pe = pe->GetEdgePrev(0);
	}
      else
	{
	  pe = pe->GetEdgeNext(0);
	}
    }
  peStart->GetEdge()->SetVertex(peCurr->GetVertex());
  Splice(1, peCurr, peStart->GetEdge());

  //////////////////////////////////////////////////
  // Remove epsilon length edges and 
  // collinear adjacent edge line segments
  if ( bCleanUpContour )
    {
      Point ppt[3];
      Vector v;
      FLOAT fMagnitude;

#ifdef ENABLE_VORONOI_POLYGON_CLEANING_UP_CONTOUR
      fprintf(stderr, "START: Cleaning Up Contour: size = %ld\n", uiSize);
#endif // ENABLE_VORONOI_POLYGON_CLEANING_UP_CONTOUR

      for ( pe = peStart->GetEdgePrev(0);
	    pe != peStart;
	    pe = peNext )
	{
	  peNext = pe->GetEdgePrev(0);

	  ppt[0] = pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
	  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
	  ppt[2] = pe->GetVertex()->GetPoint();
	  v = ppt[2] - ppt[1];
	  v[2] = 0.0;
	  v[3] = 0.0;
	  fMagnitude = v.Magnitude();
	  d = APM2D::SinTheta(ppt[0](0), ppt[0](1), 
			      ppt[1](0), ppt[1](1), 
			      ppt[2](0), ppt[2](1));
	  if ( fabs(d) <= s_cdStraightnessEpsilon ||
	       fMagnitude < s_cfNullEdgeEpsilon )
	    {
	      // Remove unnecessary vertex and edge
	      fprintf(stderr, "  Removing edge: %f < %f\n    ( %f, %f ) ==> ( %f, %f ) ==> ( %f, %f )\n", fMagnitude, s_cfNullEdgeEpsilon, ppt[0](0), ppt[0](1), ppt[1](0), ppt[1](1), ppt[2](0), ppt[2](1));
	      uiSize--;
	      pePrev = pe->GetEdgeNext(0);
	      pePrev->SetVertex(pe->GetVertex());
	      Splice(1, pePrev, pe->GetEdge());
	      Splice(1, pe, peNext->GetEdge());
	      Splice(1, pePrev, peNext->GetEdge());
	      DestroyVertex(pe->GetEdge()->GetVertex());
	      DestroyEdge(pe->GetEdge());
	      DestroyEdge(pe);
	    }
	}

      if ( peStart == peStart->GetEdgePrev(0) )
	{
	  ASSERT( peStart->GetVertex() == peStart->GetEdge()->GetVertex() );
	  DestroyVertex(peStart->GetEdge()->GetVertex());
	  DestroyEdge(peStart->GetEdge());
	  DestroyEdge(peStart);
	  goto end;
	}
      else if ( peStart == peStart->GetEdgePrev(0)->GetEdgePrev(0) )
	{
	  pe = peStart->GetEdgePrev(0);
	  Splice(1, pe->GetEdge(), peStart);
	  Splice(1, pe, peStart->GetEdge());

	  DestroyVertex(pe->GetEdge()->GetVertex());
	  DestroyEdge(pe->GetEdge());
	  DestroyEdge(pe);

	  DestroyVertex(peStart->GetEdge()->GetVertex());
	  DestroyEdge(peStart->GetEdge());
	  DestroyEdge(peStart);

	  goto end;
	}
      else
	{
	  pe = peStart;
	  ppt[0] = pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
	  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
	  ppt[2] = pe->GetVertex()->GetPoint();
	  v = ppt[2] - ppt[1];
	  v[2] = 0.0;
	  v[3] = 0.0;
	  fMagnitude = v.Magnitude();
	  d = APM2D::SinTheta(ppt[0](0), ppt[0](1), 
			      ppt[1](0), ppt[1](1), 
			      ppt[2](0), ppt[2](1));
	  if ( fabs(d) <= s_cdStraightnessEpsilon ||
	       fMagnitude < s_cfNullEdgeEpsilon )
	    {
	      peStart = pe->GetEdgePrev(0);

	      // Remove unnecessary vertex and edge
	      fprintf(stderr, "  Removing edge: %f < %f\n    ( %f, %f ) ==> ( %f, %f ) ==> ( %f, %f )\n", fMagnitude, s_cfNullEdgeEpsilon, ppt[0](0), ppt[0](1), ppt[1](0), ppt[1](1), ppt[2](0), ppt[2](1));

	      uiSize--;
	      pePrev = pe->GetEdgeNext(0);
	      peNext = pe->GetEdgePrev(0);
	      pePrev->SetVertex(pe->GetVertex());
	      Splice(1, pePrev, pe->GetEdge());
	      Splice(1, pe, peNext->GetEdge());
	      Splice(1, pePrev, peNext->GetEdge());
	      DestroyVertex(pe->GetEdge()->GetVertex());
	      DestroyEdge(pe->GetEdge());
	      DestroyEdge(pe);

	      if ( peStart == peStart->GetEdgePrev(0)->GetEdgePrev(0) )
		{
		  pe = peStart->GetEdgePrev(0);
		  Splice(1, pe->GetEdge(), peStart);
		  Splice(1, pe, peStart->GetEdge());
		  
		  DestroyVertex(pe->GetEdge()->GetVertex());
		  DestroyEdge(pe->GetEdge());
		  DestroyEdge(pe);

		  DestroyVertex(peStart->GetEdge()->GetVertex());
		  DestroyEdge(peStart->GetEdge());
		  DestroyEdge(peStart);

		  goto end;
		}
	    }
	}

#ifdef ENABLE_VORONOI_POLYGON_CLEANING_UP_CONTOUR
      fprintf(stderr, "END  : Cleaning Up Contour: size = %ld\n", uiSize);
#endif // ENABLE_VORONOI_POLYGON_CLEANING_UP_CONTOUR
    }


  //////////////////////////////////////////////////////////////////////
  // Create the Voronoi Diagram of the inside of the contour
  peStart = peStart->GetEdge();
  memset(pppeContourSegments[0], 0, uiSize*sizeof(QEEdge*));
  memset(pppeContourSegments[1], 0, uiSize*sizeof(QEEdge*));

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "  START: Inside Voronoi\n");
  fprintf(stderr, "    START: Find Convex Corner\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
  //////////////////////////////////////////////////
  // Locate a Convex corner which will start the first segment
  bConvex = TRUE;
  peCurr = peStart;
  do
    {
      if ( Counterclockwise(peCurr->GetEdgePrev(0)->GetVertex()->GetPoint(),
			    peCurr->GetVertex()->GetPoint(),
			    peCurr->GetEdge()->GetVertex()->GetPoint()) )
	{
	  // Convex corner starts a new segment
	  pppeContourSegments[0][0] = peCurr;
	  pppeContourSegments[1][0] = peCurr;
	  bConvex = FALSE;
	  break;
	}
      peCurr = peCurr->GetEdgeNext(0);
    }
  while ( peCurr != peStart );
  ASSERT( !bConvex );
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "    END  : Find Convex Corner\n");

  fprintf(stderr, "    START: Split contour into maximal reflex segments\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
  //////////////////////////////////////////////////
  // Divide the contour into maximal reflex segments
  pe = pppeContourSegments[0][0];
  i = 0;
  uiSegment = 0;
  peCurr = pe;
  peCurr->GetVertex()->SetID(i);
  for ( i++, peCurr = peCurr->GetEdgeNext(0);
	i < uiSize && peCurr != pe;
	i++, peCurr = peCurr->GetEdgeNext(0) )
    {
      peCurr->GetVertex()->SetID(i);
      if ( Counterclockwise(peCurr->GetEdgePrev(0)->GetVertex()->GetPoint(),
			    peCurr->GetVertex()->GetPoint(),
			    peCurr->GetEdge()->GetVertex()->GetPoint()) )
      /*
      if ( !Counterclockwise(peCurr->GetVertex()->GetPoint(),
			     peCurr->GetEdgePrev(0)->GetVertex()->GetPoint(),
			     peCurr->GetEdge()->GetVertex()->GetPoint()) )
      */
	{
	  // Convex corner starts a new segment
	  uiSegment++;
	  pppeContourSegments[0][uiSegment] = peCurr;
	  pppeContourSegments[1][uiSegment] = peCurr;	  
	}
      else
	{
	  // Concave corner or straight joint
	  pppeContourSegments[1][uiSegment] = peCurr;	  
	}
    }
  ASSERT( peCurr == pe );
  ASSERT( i == uiSize );
  uiSegment++;
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "    END  : Split contour into maximal reflex segments\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

  CreateVoronoiPolygonShewchuk(uiSegment, pppeContourSegments[0], pppeContourSegments[1], fW, bConvex, TRUE); 
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "  END  : Inside Voronoi\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

  //////////////////////////////////////////////////////////////////////
  // Create the Voronoi Diagram of the outside of the contour
  peStart = peStart->GetEdge();
  memset(pppeContourSegments[0], 0, uiSize*sizeof(QEEdge*));
  memset(pppeContourSegments[1], 0, uiSize*sizeof(QEEdge*));

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "  START: Outside Voronoi\n");
  fprintf(stderr, "    START: Find Convex Corner\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
  //////////////////////////////////////////////////
  // Locate a Convex corner which will start the first segment
  bConvex = TRUE;
  peCurr = peStart;
  do
    {
      if ( Counterclockwise(peCurr->GetEdgePrev(0)->GetVertex()->GetPoint(),
			    peCurr->GetVertex()->GetPoint(),
			    peCurr->GetEdge()->GetVertex()->GetPoint()) )
	{
	  // Convex corner starts a new segment
	  pppeContourSegments[0][0] = peCurr;
	  pppeContourSegments[1][0] = peCurr;
	  bConvex = FALSE;
	  break;
	}
      peCurr = peCurr->GetEdgeNext(0);
    }
  while ( peCurr != peStart );
  if ( bConvex )
    {
      // The contour is convex so there are no convex outer vertices
      pppeContourSegments[0][0] = peCurr;
      pppeContourSegments[0][0] = peCurr;
    }
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "    END  : Find Convex Corner\n");

  fprintf(stderr, "    START: Split contour into maximal reflex segments\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
  //////////////////////////////////////////////////
  // Divide the contour into maximal reflex segments
  pe = pppeContourSegments[0][0];
  i = 0;
  uiSegment = 0;
  peCurr = pe;
  peCurr->GetVertex()->SetID(i);
  for ( i++, peCurr = peCurr->GetEdgeNext(0);
	i < uiSize && peCurr != pe;
	i++, peCurr = peCurr->GetEdgeNext(0) )
    {
      peCurr->GetVertex()->SetID(i);
      if ( Counterclockwise(peCurr->GetEdgePrev(0)->GetVertex()->GetPoint(),
			    peCurr->GetVertex()->GetPoint(),
			    peCurr->GetEdge()->GetVertex()->GetPoint()) )
      /*
      if ( !Counterclockwise(peCurr->GetVertex()->GetPoint(),
			     peCurr->GetEdgePrev(0)->GetVertex()->GetPoint(),
			     peCurr->GetEdge()->GetVertex()->GetPoint()) )
      */
	{
	  // Convex corner starts a new segment
	  uiSegment++;
	  pppeContourSegments[0][uiSegment] = peCurr;
	  pppeContourSegments[1][uiSegment] = peCurr;	  
	}
      else
	{
	  // Concave corner or straight joint
	  pppeContourSegments[1][uiSegment] = peCurr;	  
	}
    }
  ASSERT( peCurr == pe );
  ASSERT( i == uiSize );
  uiSegment++;
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "    END  : Split contour into maximal reflex segments\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

  CreateVoronoiPolygonShewchuk(uiSegment, pppeContourSegments[0], pppeContourSegments[1], -fW, bConvex, FALSE); 
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "  END  : Outside Voronoi\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

  //////////////////////////////////////////////////
  // Make the distance values on the exterior negative
  iter.Init(this);
  for ( pvx = iter.PeekFirstVertex();
	pvx != NULL;
	pvx = iter.PeekNextVertex() )
    {
      pt = pvx->GetPoint();
      if ( pt[3] == -1.0 )
	{
	  pt[2] = -pt[2];
	  pt[3] = 1.0;
	  pvx->SetPoint(pt);
	}
    }

end:
  delete []pppeContourSegments[0];
  delete []pppeContourSegments[1];

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "END  : CreateVoronoiPolygonShewchuk\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
}

VOID QEMesh::CreateVoronoiPolygonShewchuk(CObjectList<QEVertex> *plistQEVertex, 
					  BOOL bCleanUpContour)
{
  QEEdge *pe, *peCurr, *pePrev, *peStart, *peNext;
  QEEdge **pppeContourSegments[2];
  //QEMeshIter iter;
  CObjectListIter<QEVertex> iterQEVertex;  
  QEVertex *pvx, *pvxNew;
  Point pt;
  UINT_32 i, uiSize, uiSegment;
  BOOL bConvex;
  FLOAT fArea, fW;
  DOUBLE d;

  ASSERT( plistQEVertex != NULL );

  //////////////////////////////////////////////////
  // Weird Static Variable Dependencies
  //s_bOpenEndpoint = FALSE;
  // Weird Static Variable Dependencies
  //////////////////////////////////////////////////

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "START: CreateVoronoiPolygonShewchuk\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

  uiSize = plistQEVertex->GetLength();
  ASSERT( uiSize > 0 );

  pppeContourSegments[0] = new QEEdge*[uiSize];
  ASSERT( pppeContourSegments[0] != NULL );
  pppeContourSegments[1] = new QEEdge*[uiSize];
  ASSERT( pppeContourSegments[1] != NULL );

  //////////////////////////////////////////////////
  // Determine the orientation of the contour

  fArea = CalculateContourArea(plistQEVertex);

  fW = 1.0;
  if ( fArea < 0.0 )
    {
      fW = -1.0;
    }

  //////////////////////////////////////////////////
  // Count the Boundary Edges

  iterQEVertex.Init(plistQEVertex);

  //////////////////////////////////////////////////
  // Copy the contour

  if ( fW == 1.0 )
    {
      pvx = iterQEVertex.PeekFirst();
    }
  else
    {
      pvx = iterQEVertex.PeekLast();
    }
  ASSERT( pvx != NULL );
  
  peCurr = CreateEdge();
  pt = pvx->GetPoint();
  pt[2] = 0.0;
  pvxNew = CreateVertex(pt);
  pvxNew->SetLimitPosition(pt);
  pvxNew->SetInscribedCircleCenterType(APM2D::icctNull);
  peCurr->SetVertex(pvxNew);
      
  pePrev = peCurr;
      
  peStart = peCurr;

  if ( fW == 1.0 )
    {
      pvx = iterQEVertex.PeekNext();
    }
  else
    {
      pvx = iterQEVertex.PeekPrev();
    }
  while ( pvx != NULL )
    {
      peCurr = CreateEdge();
      pt = pvx->GetPoint();
      pt[2] = 0.0;
      pvxNew = CreateVertex(pt);
      pvxNew->SetLimitPosition(pt);
      pvxNew->SetInscribedCircleCenterType(APM2D::icctNull);
      peCurr->SetVertex(pvxNew);
	  
      peCurr->GetEdge()->SetVertex(pePrev->GetVertex());
      Splice(1, peCurr->GetEdge(), pePrev);
      
      pePrev = peCurr;
      
      if ( fW == 1.0 )
	{
	  pvx = iterQEVertex.PeekNext();
	}
      else
	{
	  pvx = iterQEVertex.PeekPrev();
	}
    }
  peStart->GetEdge()->SetVertex(peCurr->GetVertex());
  Splice(1, peCurr, peStart->GetEdge());


  //////////////////////////////////////////////////
  // Remove epsilon length edges and 
  // collinear adjacent edge line segments
  if ( bCleanUpContour )
    {
      Point ppt[3];
      Vector v;
      FLOAT fMagnitude;

#ifdef ENABLE_VORONOI_POLYGON_CLEANING_UP_CONTOUR
      fprintf(stderr, "START: Cleaning Up Contour: size = %ld\n", uiSize);
#endif // ENABLE_VORONOI_POLYGON_CLEANING_UP_CONTOUR

      for ( pe = peStart->GetEdgePrev(0);
	    pe != peStart;
	    pe = peNext )
	{
	  peNext = pe->GetEdgePrev(0);

	  ppt[0] = pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
	  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
	  ppt[2] = pe->GetVertex()->GetPoint();
	  v = ppt[2] - ppt[1];
	  v[2] = 0.0;
	  v[3] = 0.0;
	  fMagnitude = v.Magnitude();
	  d = APM2D::SinTheta(ppt[0](0), ppt[0](1), 
			      ppt[1](0), ppt[1](1), 
			      ppt[2](0), ppt[2](1));
	  if ( fabs(d) <= s_cdStraightnessEpsilon ||
	       fMagnitude < s_cfNullEdgeEpsilon )
	    {
	      // Remove unnecessary vertex and edge
	      fprintf(stderr, "  Removing edge: %f < %f\n    ( %f, %f ) ==> ( %f, %f ) ==> ( %f, %f )\n", fMagnitude, s_cfNullEdgeEpsilon, ppt[0](0), ppt[0](1), ppt[1](0), ppt[1](1), ppt[2](0), ppt[2](1));
	      uiSize--;
	      pePrev = pe->GetEdgeNext(0);
	      pePrev->SetVertex(pe->GetVertex());
	      Splice(1, pePrev, pe->GetEdge());
	      Splice(1, pe, peNext->GetEdge());
	      Splice(1, pePrev, peNext->GetEdge());
	      DestroyVertex(pe->GetEdge()->GetVertex());
	      DestroyEdge(pe->GetEdge());
	      DestroyEdge(pe);
	    }
	}

      if ( peStart == peStart->GetEdgePrev(0) )
	{
	  ASSERT( peStart->GetVertex() == peStart->GetEdge()->GetVertex() );
	  DestroyVertex(peStart->GetEdge()->GetVertex());
	  DestroyEdge(peStart->GetEdge());
	  DestroyEdge(peStart);
	  goto end;
	}
      else if ( peStart == peStart->GetEdgePrev(0)->GetEdgePrev(0) )
	{
	  pe = peStart->GetEdgePrev(0);
	  Splice(1, pe->GetEdge(), peStart);
	  Splice(1, pe, peStart->GetEdge());

	  DestroyVertex(pe->GetEdge()->GetVertex());
	  DestroyEdge(pe->GetEdge());
	  DestroyEdge(pe);

	  DestroyVertex(peStart->GetEdge()->GetVertex());
	  DestroyEdge(peStart->GetEdge());
	  DestroyEdge(peStart);

	  goto end;
	}
      else
	{
	  pe = peStart;
	  ppt[0] = pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
	  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
	  ppt[2] = pe->GetVertex()->GetPoint();
	  v = ppt[2] - ppt[1];
	  v[2] = 0.0;
	  v[3] = 0.0;
	  fMagnitude = v.Magnitude();
	  d = APM2D::SinTheta(ppt[0](0), ppt[0](1), 
			      ppt[1](0), ppt[1](1), 
			      ppt[2](0), ppt[2](1));
	  if ( fabs(d) <= s_cdStraightnessEpsilon ||
	       fMagnitude < s_cfNullEdgeEpsilon )
	    {
	      peStart = pe->GetEdgePrev(0);

	      // Remove unnecessary vertex and edge
	      fprintf(stderr, "  Removing edge: %f < %f\n    ( %f, %f ) ==> ( %f, %f ) ==> ( %f, %f )\n", fMagnitude, s_cfNullEdgeEpsilon, ppt[0](0), ppt[0](1), ppt[1](0), ppt[1](1), ppt[2](0), ppt[2](1));

	      uiSize--;
	      pePrev = pe->GetEdgeNext(0);
	      peNext = pe->GetEdgePrev(0);
	      pePrev->SetVertex(pe->GetVertex());
	      Splice(1, pePrev, pe->GetEdge());
	      Splice(1, pe, peNext->GetEdge());
	      Splice(1, pePrev, peNext->GetEdge());
	      DestroyVertex(pe->GetEdge()->GetVertex());
	      DestroyEdge(pe->GetEdge());
	      DestroyEdge(pe);

	      if ( peStart == peStart->GetEdgePrev(0)->GetEdgePrev(0) )
		{
		  pe = peStart->GetEdgePrev(0);
		  Splice(1, pe->GetEdge(), peStart);
		  Splice(1, pe, peStart->GetEdge());
		  
		  DestroyVertex(pe->GetEdge()->GetVertex());
		  DestroyEdge(pe->GetEdge());
		  DestroyEdge(pe);

		  DestroyVertex(peStart->GetEdge()->GetVertex());
		  DestroyEdge(peStart->GetEdge());
		  DestroyEdge(peStart);

		  goto end;
		}
	    }
	}

#ifdef ENABLE_VORONOI_POLYGON_CLEANING_UP_CONTOUR
      fprintf(stderr, "END  : Cleaning Up Contour: size = %ld\n", uiSize);
#endif // ENABLE_VORONOI_POLYGON_CLEANING_UP_CONTOUR
    }

  //////////////////////////////////////////////////////////////////////
  // Create the Voronoi Diagram of the inside of the contour
  peStart = peStart->GetEdge();
  memset(pppeContourSegments[0], 0, uiSize*sizeof(QEEdge*));
  memset(pppeContourSegments[1], 0, uiSize*sizeof(QEEdge*));

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "  START: Inside Voronoi\n");
  fprintf(stderr, "    START: Find Convex Corner\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
  //////////////////////////////////////////////////
  // Locate a Convex corner which will start the first segment
  bConvex = TRUE;
  peCurr = peStart;
  do
    {
      if ( Counterclockwise(peCurr->GetEdgePrev(0)->GetVertex()->GetPoint(),
			    peCurr->GetVertex()->GetPoint(),
			    peCurr->GetEdge()->GetVertex()->GetPoint()) )
	{
	  // Convex corner starts a new segment
	  pppeContourSegments[0][0] = peCurr;
	  pppeContourSegments[1][0] = peCurr;
	  bConvex = FALSE;
	  break;
	}
      peCurr = peCurr->GetEdgeNext(0);
    }
  while ( peCurr != peStart );
  ASSERT( !bConvex );
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "    END  : Find Convex Corner\n");

  fprintf(stderr, "    START: Split contour into maximal reflex segments\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
  //////////////////////////////////////////////////
  // Divide the contour into maximal reflex segments
  pe = pppeContourSegments[0][0];
  i = 0;
  uiSegment = 0;
  peCurr = pe;
  peCurr->GetVertex()->SetID(i);
  for ( i++, peCurr = peCurr->GetEdgeNext(0);
	i < uiSize && peCurr != pe;
	i++, peCurr = peCurr->GetEdgeNext(0) )
    {
      peCurr->GetVertex()->SetID(i);
      if ( Counterclockwise(peCurr->GetEdgePrev(0)->GetVertex()->GetPoint(),
			    peCurr->GetVertex()->GetPoint(),
			    peCurr->GetEdge()->GetVertex()->GetPoint()) )
      /*
      if ( !Counterclockwise(peCurr->GetVertex()->GetPoint(),
			     peCurr->GetEdgePrev(0)->GetVertex()->GetPoint(),
			     peCurr->GetEdge()->GetVertex()->GetPoint()) )
      */
	{
	  // Convex corner starts a new segment
	  uiSegment++;
	  pppeContourSegments[0][uiSegment] = peCurr;
	  pppeContourSegments[1][uiSegment] = peCurr;	  
	}
      else
	{
	  // Concave corner or straight joint
	  pppeContourSegments[1][uiSegment] = peCurr;	  
	}
    }
  ASSERT( peCurr == pe );
  ASSERT( i == uiSize );
  uiSegment++;
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "    END  : Split contour into maximal reflex segments\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

  CreateVoronoiPolygonShewchuk(uiSegment, pppeContourSegments[0], pppeContourSegments[1], 
			   fW, bConvex, TRUE); 
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "  END  : Inside Voronoi\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

  //////////////////////////////////////////////////////////////////////
  // Create the Voronoi Diagram of the outside of the contour
  peStart = peStart->GetEdge();
  memset(pppeContourSegments[0], 0, uiSize*sizeof(QEEdge*));
  memset(pppeContourSegments[1], 0, uiSize*sizeof(QEEdge*));

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "  START: Outside Voronoi\n");
  fprintf(stderr, "    START: Find Convex Corner\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
  //////////////////////////////////////////////////
  // Locate a Convex corner which will start the first segment
  bConvex = TRUE;
  peCurr = peStart;
  do
    {
      if ( Counterclockwise(peCurr->GetEdgePrev(0)->GetVertex()->GetPoint(),
			    peCurr->GetVertex()->GetPoint(),
			    peCurr->GetEdge()->GetVertex()->GetPoint()) )
	{
	  // Convex corner starts a new segment
	  pppeContourSegments[0][0] = peCurr;
	  pppeContourSegments[1][0] = peCurr;
	  bConvex = FALSE;
	  break;
	}
      peCurr = peCurr->GetEdgeNext(0);
    }
  while ( peCurr != peStart );
  if ( bConvex )
    {
      // The contour is convex so there are no convex outer vertices
      pppeContourSegments[0][0] = peCurr;
      pppeContourSegments[0][0] = peCurr;
    }
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "    END  : Find Convex Corner\n");

  fprintf(stderr, "    START: Split contour into maximal reflex segments\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
  //////////////////////////////////////////////////
  // Divide the contour into maximal reflex segments
  pe = pppeContourSegments[0][0];
  i = 0;
  uiSegment = 0;
  peCurr = pe;
  peCurr->GetVertex()->SetID(i);
  for ( i++, peCurr = peCurr->GetEdgeNext(0);
	i < uiSize && peCurr != pe;
	i++, peCurr = peCurr->GetEdgeNext(0) )
    {
      peCurr->GetVertex()->SetID(i);
      if ( Counterclockwise(peCurr->GetEdgePrev(0)->GetVertex()->GetPoint(),
			    peCurr->GetVertex()->GetPoint(),
			    peCurr->GetEdge()->GetVertex()->GetPoint()) )
      /*
      if ( !Counterclockwise(peCurr->GetVertex()->GetPoint(),
			     peCurr->GetEdgePrev(0)->GetVertex()->GetPoint(),
			     peCurr->GetEdge()->GetVertex()->GetPoint()) )
      */
	{
	  // Convex corner starts a new segment
	  uiSegment++;
	  pppeContourSegments[0][uiSegment] = peCurr;
	  pppeContourSegments[1][uiSegment] = peCurr;	  
	}
      else
	{
	  // Concave corner or straight joint
	  pppeContourSegments[1][uiSegment] = peCurr;	  
	}
    }
  ASSERT( peCurr == pe );
  ASSERT( i == uiSize );
  uiSegment++;
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "    END  : Split contour into maximal reflex segments\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

  CreateVoronoiPolygonShewchuk(uiSegment, pppeContourSegments[0], pppeContourSegments[1], 
			   -fW, bConvex, FALSE); 
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "  END  : Outside Voronoi\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

end:
  delete []pppeContourSegments[0];
  delete []pppeContourSegments[1];

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "END  : CreateVoronoiPolygonShewchuk\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
}

VOID QEMesh::CreateVoronoiPolygonShewchuk(UINT_32 uiSize, QEEdge **ppeContourLeft, 
					  QEEdge **ppeContourRight, FLOAT fW,
					  BOOL bConvex, BOOL bInside)
{
  QEEdge *peLeftEnd, *peLeftMiddle, *peRightEnd, *peRightMiddle;
  UINT_32 uiHalf;
  BOOL bMergeSingleSegment, bPossiblyMergeTwice;

  if ( uiSize == 1 )
    {
      QEEdge *pe, *peNext, *peCurr, *peNew, *peStart, *pePrev;
      QEVertex *pvxNew;
      QEFacet *pfaNew;
      Point pt;
      Point ptLimit;
      Vector v;

      bPossiblyMergeTwice = FALSE;

      if ( ppeContourLeft[0] == ppeContourRight[0] )
	{
	  pe = ppeContourLeft[0];

	  peNew = CreateEdge();
	  
	  pfaNew = CreateFacet();
	  pfaNew->SetForwardEdge(pe);

	  peNew->SetVertex(pe->GetEdge()->GetVertex());
	  peNew->GetEdge()->SetVertex(pe->GetVertex());
	  peNew->SetFacet(pfaNew);
	  pe->SetFacet(pfaNew);

	  Splice(1, pe, peNew->GetEdge());
	  Splice(1, pe->GetEdge()->GetEdgePrev(1), peNew);

	  return;
	}
      else if ( ppeContourLeft[0]->GetEdgePrev(0) == ppeContourRight[0] && bConvex )
	{
	  ////////////////////////////////////////////////////////////
	  // This constructs the complete outer Voronoi diagram of a convex contour 
	  ASSERT( !bInside );
	  pe = ppeContourLeft[0];

	  //////////////////////////////////////////////////
	  // Special Case the first edge
	  peNext = pe->GetEdgeNext(0);

	  v = pe->GetEdge()->GetVertex()->GetPoint() - pe->GetVertex()->GetPoint();
	  v.Set(-v(1), v(0), 0.0);
	  v.Normalize();
	  v[2] = fW;
	  pt = v;
	  pt[3] = 0.0;

	  peCurr = pe;

	  pfaNew = CreateFacet();
	  pfaNew->SetForwardEdge(pe);
	  pe->SetFacet(pfaNew);

	  pvxNew = CreateVertex(pt);
	  ptLimit = pe->GetEdge()->GetVertex()->GetPoint() + cfInfinityMagnitude*v;
	  ptLimit[3] = 1.0;
	  pvxNew->SetLimitPosition(ptLimit);
	  pvxNew->SetInscribedCircleCenterType(APM2D::icctInfinity);
	  peNew = CreateEdge();
	  peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	  peNew->GetEdge()->SetVertex(pvxNew);
	  peNew->SetFacet(pfaNew);
	  Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);

	  peCurr = peNew;

	  pvxNew = CreateVertex(pt);
	  ptLimit = pe->GetVertex()->GetPoint() + cfInfinityMagnitude*v;
	  ptLimit[3] = 1.0;
	  pvxNew->SetLimitPosition(ptLimit);
	  pvxNew->SetInscribedCircleCenterType(APM2D::icctInfinity);
	  peNew = CreateEdge();
	  peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	  peNew->GetEdge()->SetVertex(pvxNew);
	  peNew->SetFacet(pfaNew);
	  Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);

	  peCurr = peNew;

	  peNew = CreateEdge();
	  peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	  peNew->GetEdge()->SetVertex(pe->GetVertex());
	  peNew->SetFacet(pfaNew);
	  Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);
	  Splice(1, pe, peNew->GetEdge());

	  peStart = peNew->GetEdge();

	  for ( pePrev = pe, pe = peNext; 
		pe != peStart; 
		pePrev = pe, pe = peNext )
	    {
	      peNext = pe->GetEdgeNext(0);

	      //////////////////////////////////////////////////
	      // The normal case
	      v = pe->GetEdge()->GetVertex()->GetPoint() - pe->GetVertex()->GetPoint();
	      v.Set(-v(1), v(0), 0.0);
	      v.Normalize();	  
	      v[2] = fW;
	      pt = v;
	      pt[3] = 0.0;

	      peCurr = pe;

	      pfaNew = CreateFacet();
	      pfaNew->SetForwardEdge(pe);
	      pe->SetFacet(pfaNew);

	      pvxNew = CreateVertex(pt);
	      ptLimit = pe->GetEdge()->GetVertex()->GetPoint() + cfInfinityMagnitude*v;
	      ptLimit[3] = 1.0;
	      pvxNew->SetLimitPosition(ptLimit);
	      pvxNew->SetInscribedCircleCenterType(APM2D::icctInfinity);
	      peNew = CreateEdge();
	      peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	      peNew->GetEdge()->SetVertex(pvxNew);
	      peNew->SetFacet(pfaNew);
	      Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);

	      peCurr = peNew;

	      if ( Counterclockwise(pe->GetEdge()->GetVertex()->GetPoint(),
				    pe->GetVertex()->GetPoint(),
				    pePrev->GetVertex()->GetPoint()) )
		{
		  peNew = CreateEdge();
		  peNew->SetVertex(peCurr->GetEdge()->GetVertex());
		  peNew->SetFacet(pfaNew);
		  Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);

		  pvxNew = CreateVertex(pt);
		  ptLimit = pe->GetVertex()->GetPoint() + cfInfinityMagnitude*v;
		  ptLimit[3] = 1.0;
		  pvxNew->SetLimitPosition(ptLimit);
		  pvxNew->SetInscribedCircleCenterType(APM2D::icctInfinity);
		  peNew->GetEdge()->SetVertex(pvxNew);

		  peCurr = peNew;

		  peNew = CreateEdge();
		  peNew->SetVertex(peCurr->GetEdge()->GetVertex());
		  peNew->GetEdge()->SetVertex(pe->GetVertex());
		  peNew->SetFacet(pfaNew);
		  Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);
		  Splice(1, pe, peNew->GetEdge());

		  peCurr = peNew->GetEdge();

		  pfaNew = CreateFacet();
		  pfaNew->SetForwardVertex(peCurr->GetVertex());
		  peCurr->SetFacet(pfaNew);
		  peCurr->GetEdgePrev(0)->SetFacet(pfaNew);

		  peNew = CreateEdge();
		  peNew->SetVertex(peCurr->GetEdge()->GetVertex());
		  peNew->GetEdge()->SetVertex(peCurr->GetEdgePrev(0)->GetVertex());
		  peNew->SetFacet(pfaNew);
		  Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);
		  Splice(1, peCurr->GetEdgePrev(0), peNew->GetEdge());
		}
	      else
		{
		  // Colinear Segments
		  peNew->GetEdge()->SetVertex(pe->GetEdgePrev(0)->GetVertex());
		  pe->GetEdgePrev(0)->SetFacet(pfaNew);
		  Splice(1, pe->GetEdgePrev(0), peNew->GetEdge());
		}	      
	    }

	  pe = pe->GetEdgePrev(1);

	  //////////////////////////////////////////////////
	  // Special Case the last vertex
	  if ( Counterclockwise(pe->GetEdge()->GetVertex()->GetPoint(),
				pe->GetVertex()->GetPoint(),
				pePrev->GetVertex()->GetPoint()) )
	    {
	      peCurr = peStart;

	      pfaNew = CreateFacet();
	      pfaNew->SetForwardVertex(peCurr->GetVertex());
	      peCurr->SetFacet(pfaNew);
	      peCurr->GetEdgePrev(0)->SetFacet(pfaNew);

	      peNew = CreateEdge();
	      peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	      peNew->GetEdge()->SetVertex(peCurr->GetEdgePrev(0)->GetVertex());
	      peNew->SetFacet(pfaNew);
	      Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);
	      Splice(1, peCurr->GetEdgePrev(0), peNew->GetEdge());
	    }
	  else
	    {
	      // Colinear Segments
	      peCurr = peStart->GetEdgePrev(0);
	      peNew = peCurr->GetEdgePrev(1);
	      DestroyVertex(peCurr->GetVertex());
	      Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peCurr->GetEdge());
	      Splice(1, peCurr->GetEdgePrev(1), peCurr);
	      DestroyEdge(peCurr->GetEdge());
	      DestroyEdge(peCurr);
	      
	      peCurr = peNew;
	      peCurr->SetVertex(peStart->GetEdge()->GetVertex());
	      peStart->SetFacet(peCurr->GetFacet());
	      Splice(1, peCurr, peStart->GetEdge()->GetEdgePrev(1));
	    }

#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES

	  return;
	}
      else
	{
	  ////////////////////////////////////////////////////////////
	  // Open Multi-edge segment

	  bMergeSingleSegment = (ppeContourLeft[0]->GetEdgePrev(0) == ppeContourRight[0]);

	  pe = ppeContourLeft[0];

	  //////////////////////////////////////////////////
	  // Special Case the first edge
	  peNext = pe->GetEdgeNext(0);

	  v = pe->GetEdge()->GetVertex()->GetPoint() - pe->GetVertex()->GetPoint();
	  v.Set(-v(1), v(0), 0.0);
	  v.Normalize(); 
	  v[2] = fW;
	  pt = v;
	  pt[3] = 0.0;
	  
	  peCurr = pe;

	  pfaNew = CreateFacet();
	  pfaNew->SetForwardEdge(pe);
	  pe->SetFacet(pfaNew);

	  pvxNew = CreateVertex(pt);
	  ptLimit = peCurr->GetEdge()->GetVertex()->GetPoint() + cfInfinityMagnitude*v;
	  ptLimit[3] = 1.0;
	  pvxNew->SetLimitPosition(ptLimit);
	  pvxNew->SetInscribedCircleCenterType(APM2D::icctInfinity);
	  peNew = CreateEdge();
	  peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	  peNew->GetEdge()->SetVertex(pvxNew);
	  peNew->SetFacet(pfaNew);
	  Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);

	  peCurr = peNew;

	  peNew = CreateEdge();
	  peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	  peNew->GetEdge()->SetVertex(pe->GetVertex());
	  peNew->SetFacet(pfaNew);
	  Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);
	  Splice(1, pe, peNew->GetEdge());

	  for ( pePrev = pe, pe = peNext; 
		pe != ppeContourRight[0]; 
		pePrev = pe, pe = peNext )
	    {
	      peNext = pe->GetEdgeNext(0);

	      //////////////////////////////////////////////////
	      // The normal case
	      v = pe->GetEdge()->GetVertex()->GetPoint() - pe->GetVertex()->GetPoint();
	      v.Set(-v(1), v(0), 0.0);
	      v.Normalize();
	      v[2] = fW;
	      pt = v;
	      pt[3] = 0.0;
	      
	      peCurr = pe;

	      pfaNew = CreateFacet();
	      pfaNew->SetForwardEdge(pe);
	      pe->SetFacet(pfaNew);

	      pvxNew = CreateVertex(pt);
	      ptLimit = peCurr->GetEdge()->GetVertex()->GetPoint() + cfInfinityMagnitude*v;
	      ptLimit[3] = 1.0;
	      pvxNew->SetLimitPosition(ptLimit);
	      pvxNew->SetInscribedCircleCenterType(APM2D::icctInfinity);
	      peNew = CreateEdge();
	      peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	      peNew->GetEdge()->SetVertex(pvxNew);
	      peNew->SetFacet(pfaNew);
	      Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);

	      peCurr = peNew;

	      peNew = CreateEdge();
	      peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	      peNew->SetFacet(pfaNew);
	      Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);

	      if ( Counterclockwise(pe->GetEdge()->GetVertex()->GetPoint(),
				    pe->GetVertex()->GetPoint(),
				    pePrev->GetVertex()->GetPoint()) )
		{
		  pvxNew = CreateVertex(pt);
		  ptLimit = pe->GetVertex()->GetPoint() + cfInfinityMagnitude*v;
		  ptLimit[3] = 1.0;
		  pvxNew->SetLimitPosition(ptLimit);
		  pvxNew->SetInscribedCircleCenterType(APM2D::icctInfinity);
		  peNew->GetEdge()->SetVertex(pvxNew);

		  peCurr = peNew;

		  peNew = CreateEdge();
		  peNew->SetVertex(peCurr->GetEdge()->GetVertex());
		  peNew->GetEdge()->SetVertex(pe->GetVertex());
		  peNew->SetFacet(pfaNew);
		  Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);
		  Splice(1, pe, peNew->GetEdge());

		  peCurr = peNew->GetEdge();

		  pfaNew = CreateFacet();
		  pfaNew->SetForwardVertex(peCurr->GetVertex());
		  peCurr->SetFacet(pfaNew);
		  peCurr->GetEdgePrev(0)->SetFacet(pfaNew);
		  
		  peNew = CreateEdge();
		  peNew->SetVertex(peCurr->GetEdge()->GetVertex());
		  peNew->GetEdge()->SetVertex(peCurr->GetEdgePrev(0)->GetVertex());
		  peNew->SetFacet(pfaNew);
		  Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);
		  Splice(1, peCurr->GetEdgePrev(0), peNew->GetEdge());
		}
	      else
		{
		  // Colinear Segments
		  peNew->GetEdge()->SetVertex(pe->GetEdgePrev(0)->GetVertex());
		  pe->GetEdgePrev(0)->SetFacet(pfaNew);
		  Splice(1, pe->GetEdgePrev(0), peNew->GetEdge());
		}
	    }

	  //////////////////////////////////////////////////
	  // Special Case the last edge
	  v = pe->GetEdge()->GetVertex()->GetPoint() - pe->GetVertex()->GetPoint();
	  v.Set(-v(1), v(0), 0.0);
	  v.Normalize();  
	  v[2] = fW;
	  pt = v;
	  pt[3] = 0.0;
	  
	  peCurr = pe;

	  pfaNew = CreateFacet();
	  pfaNew->SetForwardEdge(pe);
	  pe->SetFacet(pfaNew);

	  peNew = CreateEdge();
	  peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	  peNew->SetFacet(pfaNew);
	  Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);

	  if ( Counterclockwise(pe->GetEdge()->GetVertex()->GetPoint(),
				pe->GetVertex()->GetPoint(),
				pePrev->GetVertex()->GetPoint()) )
	    {
	      pvxNew = CreateVertex(pt);
	      ptLimit = peCurr->GetVertex()->GetPoint() + cfInfinityMagnitude*v;
	      ptLimit[3] = 1.0;
	      pvxNew->SetLimitPosition(ptLimit);
	      pvxNew->SetInscribedCircleCenterType(APM2D::icctInfinity);
	      peNew->GetEdge()->SetVertex(pvxNew);

	      peCurr = peNew;

	      peNew = CreateEdge();
	      peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	      peNew->GetEdge()->SetVertex(pe->GetVertex());
	      peNew->SetFacet(pfaNew);
	      Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);
	      Splice(1, pe, peNew->GetEdge());

	      peCurr = peNew->GetEdge();

	      pfaNew = CreateFacet();
	      pfaNew->SetForwardVertex(peCurr->GetVertex());
	      peCurr->SetFacet(pfaNew);
	      peCurr->GetEdgePrev(0)->SetFacet(pfaNew);

	      peNew = CreateEdge();
	      peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	      peNew->GetEdge()->SetVertex(peCurr->GetEdgePrev(0)->GetVertex());
	      peNew->SetFacet(pfaNew);
	      Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peNew);
	      Splice(1, peCurr->GetEdgePrev(0), peNew->GetEdge());
	    }
	  else
	    {
	      // Colinear Segments
	      peNew->GetEdge()->SetVertex(pe->GetEdgePrev(0)->GetVertex());
	      pe->GetEdgePrev(0)->SetFacet(pfaNew);
	      Splice(1, pe->GetEdgePrev(0), peNew->GetEdge());
	    }

	  if ( !bMergeSingleSegment )
	    {
	      // This is the normal case
	      return;
	    }
	}

      //////////////////////////////////////////////////
      // If there is exactly one concave vertex then the outer
      //   Voronoi diagram will have a single segment but will still need to be merged
      ASSERT( !bInside );

      peLeftEnd = ppeContourLeft[0];
      peLeftMiddle = ppeContourRight[0];
      peRightMiddle = ppeContourLeft[0];
      peRightEnd = ppeContourRight[0];
    }
  else
    {
      ASSERT( uiSize > 1 );

      ////////////////////////////////////////////////////////////
      // Divide and Conquer

      //////////////////////////////////////////////////
      // Split and recursively compute sub-Voronoi diagrams
      uiHalf = uiSize / 2;

      peLeftEnd = ppeContourLeft[0];
      peLeftMiddle = ppeContourRight[uiHalf-1];
      peRightMiddle = ppeContourLeft[uiHalf];
      peRightEnd = ppeContourRight[uiSize-1];

      bPossiblyMergeTwice = (peLeftEnd->GetEdgePrev(0) == peRightEnd);

      CreateVoronoiPolygonShewchuk(uiHalf, ppeContourLeft, ppeContourRight, fW, FALSE, bInside);
      CreateVoronoiPolygonShewchuk(uiSize - uiHalf, ppeContourLeft + uiHalf, 
				   ppeContourRight + uiHalf, fW, FALSE, bInside);
    }

#define THERE_CAN_BE_ONLY_ONE_MERGE_PROCEDURE
#ifdef THERE_CAN_BE_ONLY_ONE_MERGE_PROCEDURE

  QEEdge *peLeft, *peRight, *pe, *peBisector, *peBisectorEnd;
  QEEdge *peLeftNext, *peRightNext;
  QEVertex *pvxNew;
  Vector v;
  Point pt;
  Point ptLimit;
  UINT_32 i, uiMergesMax;

  uiMergesMax = ( bPossiblyMergeTwice ) ? ( 2 ) : ( 1 );

  for ( i = 0; i < uiMergesMax; i++ )
    {
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
      ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES

      peBisector = NULL;
      peBisectorEnd = peRightMiddle;	  
      
      // Break open the left side
      pe = peLeftMiddle->GetEdgeNext(0);
      ASSERT( pe->GetEdge()->GetFacet() == GetNullFacet() );
      peLeft = pe->GetEdgeNext(0);
      peLeft->SetFacet(peLeft->GetFacet());
      pe->GetEdgePrev(1)->SetVertex(pe->GetEdgePrev(1)->GetVertex());
      pe->GetEdge()->GetEdgePrev(1)->SetVertex(pe->GetEdge()->GetEdgePrev(1)->GetVertex());
      Splice(1, pe, pe->GetEdgePrev(1));
      Splice(1, pe->GetEdge(), pe->GetEdge()->GetEdgePrev(1));
      DestroyEdge(pe->GetEdge());
      DestroyEdge(pe);

      // Break open the right side
      pe = peRightMiddle->GetEdgePrev(0);
      ASSERT( pe->GetEdge()->GetFacet() == GetNullFacet() );
      peRight = pe->GetEdgePrev(0);
      peRight->SetFacet(peRight->GetFacet());
      pe->GetEdgePrev(1)->SetVertex(pe->GetEdgePrev(1)->GetVertex());
      pe->GetEdge()->GetEdgePrev(1)->SetVertex(pe->GetEdge()->GetEdgePrev(1)->GetVertex());
      Splice(1, pe, pe->GetEdgePrev(1));
      Splice(1, pe->GetEdge(), pe->GetEdge()->GetEdgePrev(1));
      DestroyEdge(pe->GetEdge());
      DestroyEdge(pe);

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
      fprintf(stderr, "    START: Merge Left = ( %.3f, %.3f ) Right = ( %.3f, %.3f )\n", 
	      peLeftMiddle->GetVertex()->GetPoint()(X), peLeftMiddle->GetVertex()->GetPoint()(Y), 
	      peRightMiddle->GetVertex()->GetPoint()(X), peRightMiddle->GetVertex()->GetPoint()(Y));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
      MergeVoronoiPolygonsShewchuk(peLeftEnd, peLeft,
				   peBisector, peBisectorEnd,
				   peRight, peRightEnd, fW);
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
      fprintf(stderr, "    END  : Merge Left = ( %.3f, %.3f ) Right = ( %.3f, %.3f )\n", 
	      peLeftMiddle->GetVertex()->GetPoint()(X), peLeftMiddle->GetVertex()->GetPoint()(Y), 
	      peRightMiddle->GetVertex()->GetPoint()(X), peRightMiddle->GetVertex()->GetPoint()(Y));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
      if ( peLeft == peLeftEnd && peRight == peRightEnd )
	{
	  ASSERT( peBisector != NULL );
	  if ( peLeft == peRight->GetEdgeNext(0) )
	    {
	      // This is the final step in the merge
	      peBisector->SetVertex(peLeft->GetVertex());
	      Splice(1, peLeft, peBisector);
	    }
	  else
	    {
	      // Create a point at infinity
	      v = CalculateBisectorDirection(peLeft->GetFacet(), TRUE, 
					     peRight->GetFacet(), TRUE, FALSE);
	      v[2] = fW;
	      pt = v;
	      pt[3] = 0.0;
	      pvxNew = CreateVertex(pt);
	      ptLimit = peBisector->GetEdge()->GetVertex()->GetLimitPosition() + cfInfinityMagnitude*v;
	      ptLimit[3] = 1.0;
	      pvxNew->SetLimitPosition(ptLimit);
	      pvxNew->SetInscribedCircleCenterType(APM2D::icctInfinity);
	      
	      peLeftNext = CreateEdge();
	      peLeftNext->SetVertex(pvxNew);
	      peLeftNext->GetEdge()->SetVertex(peLeft->GetVertex());
	      peLeftNext->SetFacet(peLeft->GetFacet());
	      Splice(1, peLeft, peLeftNext->GetEdge());
	  
	      peRightNext = CreateEdge();
	      peRightNext->SetVertex(pvxNew);
	      peRightNext->GetEdge()->SetVertex(peRight->GetEdge()->GetVertex());
	      peRightNext->GetEdge()->SetFacet(peRight->GetFacet());
	      Splice(1, peRight->GetEdge()->GetEdgePrev(1), peRightNext->GetEdge());
	      
	      ASSERT( peBisectorEnd != NULL && peBisector != NULL );
	      peBisector->SetVertex(pvxNew);
	      Splice(1, peLeftNext, peBisector);
	      Splice(1, peBisector, peRightNext);
	    }
	  break;
	}

      if ( bPossiblyMergeTwice )
	{
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	  fprintf(stderr, "Merge again because this is the last outside merge!\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	  if ( bInside )
	    {
	      WriteDebugVoronoiPolygonShewchuk(stderr);
	    }
	  ASSERT( !bInside );
	  
	  peLeft = peLeftEnd;
	  peRight = peRightEnd;
	  peLeftEnd = peRightMiddle;
	  peRightEnd = peLeftMiddle;
	  peLeftMiddle = peRight;
	  peRightMiddle = peLeft;
	}
    }
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
      ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING_OUTPUT_FILES
      ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING_OUTPUT_FILES

#else // THERE_CAN_BE_ONLY_ONE_MERGE_PROCEDURE

  MergeVoronoiPolygonShewchuk(peLeftEnd, peLeftMiddle, peRightMiddle, peRightEnd, fW);
  if ( bPossiblyMergeTwice && 
       peLeftEnd->GetEdgeNext(1) != peRightEnd->GetEdgeNext(0) )
    {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
      fprintf(stderr, "Merge again because this is the last outside merge!\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
      ASSERT( !bInside );
      MergeVoronoiPolygonShewchuk(peRightMiddle, peRightEnd, peLeftEnd, peLeftMiddle, fW);
    }

#endif // THERE_CAN_BE_ONLY_ONE_MERGE_PROCEDURE
}

VOID QEMesh::MergeVoronoiPolygonShewchuk(QEEdge *peLeftEnd, QEEdge *peLeftMiddle, 
					 QEEdge *peRightMiddle, QEEdge *peRightEnd, FLOAT fW)
{
  QEEdge *peLeft, *peRight, *peBisector;
  QEEdge *peLeftNext, *peRightNext;
  QEEdge *pe, *peNew;
  QEVertex *pvxNew, *pvxLeft, *pvxRight;
  //QEVertex *ppvxEndpoint[2];
  //FLOAT pfDistance[2];
  Point pt, ptCenter;
  Vector v;

  UINT_32 uiMergeDirection;
  Point ptMerge, ptLeft, ptRight;
  Point ptLimit;
  QEFacet *pfaMerge, *pfaLeft, *pfaRight;
  APM2D::InscribedCircleCenterType eMergeICCT, eLeftICCT, eRightICCT, eICCT;
  BOOL bReplacePoint, bContinueSearch;

  //////////////////////////////////////////////////////////////////////
  // Merge

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "    START: Merge Left = ( %.3f, %.3f ) Right = ( %.3f, %.3f )\n", peLeftMiddle->GetVertex()->GetPoint()(X), peLeftMiddle->GetVertex()->GetPoint()(Y), peRightMiddle->GetVertex()->GetPoint()(X), peRightMiddle->GetVertex()->GetPoint()(Y));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

  ////////////////////////////////////////
  // Start the Bisector

  //////////////////////////////
  // Left side
  pe = peLeftMiddle->GetEdgeNext(0);
  peLeft = pe->GetEdgeNext(0);

  // Remove the edge at infinity
  ASSERT( pe->GetEdge()->GetFacet() == GetNullFacet() );
  ASSERT( pe != peLeftEnd && 
	  pe != peLeftMiddle &&
	  pe != peLeft );
  pe->GetEdge()->GetEdgePrev(1)->SetVertex(pe->GetEdge()->GetEdgePrev(1)->GetVertex());
  pe->GetEdgePrev(1)->SetVertex(pe->GetEdgePrev(1)->GetVertex());
  peLeftMiddle->SetFacet(peLeftMiddle->GetFacet());
  Splice(1, pe->GetEdge(), pe->GetEdge()->GetEdgePrev(1));
  Splice(1, pe, pe->GetEdgePrev(1));
  DestroyEdge(pe->GetEdge());
  DestroyEdge(pe);

  //////////////////////////////
  // Right side
  pe = peRightMiddle->GetEdgePrev(0);
  peRight = pe->GetEdgePrev(0);

  // Remove the edge at infinity, and recycle it as the first bisector
  ASSERT( pe->GetEdge()->GetFacet() == GetNullFacet() );
  ASSERT( pe != peRightEnd && 
	  pe != peRightMiddle &&
	  pe != peRight );
  pe->GetEdge()->GetEdgePrev(1)->SetVertex(pe->GetEdge()->GetEdgePrev(1)->GetVertex());
  pe->GetEdgePrev(1)->SetVertex(pe->GetEdgePrev(1)->GetVertex());
  peRightMiddle->SetFacet(peRightMiddle->GetFacet());
  Splice(1, pe->GetEdge(), pe->GetEdge()->GetEdgePrev(1));
  Splice(1, pe, pe->GetEdgePrev(1));

  peBisector = pe;

  //////////////////////////////
  // Attach the Bisector
  ASSERT( peLeftMiddle->GetEdgeNext(0) == peRightMiddle );
  peBisector->SetVertex(peRightMiddle->GetVertex()); // This is really unset
  peBisector->GetEdge()->SetVertex(peRightMiddle->GetVertex());
  peBisector->SetFacet(peRightMiddle->GetFacet());
  peBisector->GetEdge()->SetFacet(peLeftMiddle->GetFacet());

  Splice(1, peRightMiddle, peBisector->GetEdge());

  ////////////////////////////////////////
  // Merge Loop
  while ( TRUE )
    {
      uiMergeDirection = 0;
      ptMerge.Set(0.0, 0.0, INFINITY);
      pfaMerge = NULL;
      pfaLeft = NULL;
      pfaRight = NULL;
      eLeftICCT = APM2D::icctNull;
      eRightICCT = APM2D::icctNull;

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
      fprintf(stderr, "      START: Searching for candidate point\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

      ////////////////////////////////////////////////////////////
      // START: Collapse edges from the left and right
      //   until they are in the sector bounded by the 
      //   two contour edges
      
      // Left
      ASSERT( peLeft->GetEdge()->GetFacet() != GetNullFacet() ||
	      peLeft == peLeft->GetFacet()->GetForwardEdge() );
      while ( peLeft != peLeft->GetFacet()->GetForwardEdge() &&
	      peLeft->GetEdgeNext(0) != peLeft->GetEdgeNext(0)->GetFacet()->GetForwardEdge() &&
	      peLeft->GetEdge()->GetVertex() != peLeft->GetFacet()->GetForwardVertex() )

	{
	  // Check if the Voronoi vertex is closer to the left or right edge
	  pvxLeft = peLeft->GetEdge()->GetVertex();
	  ASSERT( pvxLeft->GetEdge()->GetEdgeNext(1)->GetEdgeNext(1)->GetEdgeNext(1) ==
		  pvxLeft->GetEdge() );
	  ptCenter = pvxLeft->GetPoint();
	  if ( APM2D::cttOutCircle == InCircleSSS_SShewchuk(pvxLeft->GetInscribedCircleCenterType(),
						      pvxLeft->GetEdge()->GetFacet(),
						      pvxLeft->GetEdge()->GetEdgeNext(1)->GetFacet(),
						      pvxLeft->GetEdge()->GetEdgePrev(1)->GetFacet(),
						      ptCenter,
							    peRight->GetFacet()) )
	    {
	      break;
	    }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	  fprintf(stderr, "        Left side skipping uniformed Voronoi edge\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	  // Advance the left next pointer
	  peLeftNext = peLeft->GetEdgeNext(0);
	  if ( peLeftNext->GetEdge()->GetFacet() == GetNullFacet() &&
	       peLeftNext != peLeftNext->GetFacet()->GetForwardEdge() )
	    {
	      peLeftNext = peLeftNext->GetEdgeNext(0);
	    }

	  //////////////////////////////
	  // Advance peLeft cleaning up extraneous edges
	  if ( peLeft->GetEdgeNext(0)->GetEdge()->GetFacet() == GetNullFacet() )
	    {
	      ASSERT( peLeft->GetEdgeNext(0) != peLeft->GetEdgeNext(0)->GetFacet()->GetForwardEdge() );
	      CollapseEdge(peLeft->GetEdgeNext(0));
	    }
	  ASSERT( peLeft != peLeft->GetFacet()->GetForwardEdge() );
	  CollapseEdge(peLeft);
	  peLeft = peLeftNext;

	  ASSERT( peLeft->GetEdge()->GetFacet() != GetNullFacet() );

#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	}
      
      // Right
      ASSERT( peRight->GetEdge()->GetFacet() != GetNullFacet() ||
	      peRight == peRight->GetFacet()->GetForwardEdge() );
      while ( peRight != peRight->GetFacet()->GetForwardEdge() &&
	      peRight->GetEdgePrev(0) != peRight->GetEdgePrev(0)->GetFacet()->GetForwardEdge() &&
	      peRight->GetVertex() != peRight->GetFacet()->GetForwardVertex() )
	{
	  // Check if the Voronoi vertex is closer to the left or right edge
	  pvxRight = peRight->GetVertex();
	  ASSERT( pvxRight->GetEdge()->GetEdgeNext(1)->GetEdgeNext(1)->GetEdgeNext(1) ==
		  pvxRight->GetEdge() );
	  ptCenter = pvxRight->GetPoint();
	  if ( APM2D::cttOutCircle == InCircleSSS_SShewchuk(pvxRight->GetInscribedCircleCenterType(),
							    pvxRight->GetEdge()->GetFacet(),
							    pvxRight->GetEdge()->GetEdgeNext(1)->GetFacet(),
							    pvxRight->GetEdge()->GetEdgePrev(1)->GetFacet(),
							    ptCenter,
							    peLeft->GetFacet()) )
	    {
	      break;
	    }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	  fprintf(stderr, "        Right side skipping uniformed Voronoi edge\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	  // Advance the right next pointer
	  peRightNext = peRight->GetEdgePrev(0);
	  if ( peRightNext->GetEdge()->GetFacet() == GetNullFacet() &&
	       peRightNext != peRightNext->GetFacet()->GetForwardEdge() )
	    {
	      peRightNext = peRightNext->GetEdgePrev(0);
	    }
	  
	  //////////////////////////////
	  // Advance peRight cleaning up extraneous edges
	  if ( peRight->GetEdgePrev(0)->GetEdge()->GetFacet() == GetNullFacet() )
	    {
	      ASSERT( peRight->GetEdgePrev(0) != peRight->GetEdgePrev(0)->GetFacet()->GetForwardEdge() );
	      CollapseEdge(peRight->GetEdgePrev(0)->GetEdge());
	    }
	  ASSERT( peRight != peRight->GetFacet()->GetForwardEdge() );
	  CollapseEdge(peRight->GetEdge());
	  peRight = peRightNext;

	  ASSERT( peRight->GetEdge()->GetFacet() != GetNullFacet() );

#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	}

      // END  : Collapse edges from the left and right
      //   until they are in the sector bounded by the 
      //   two contour edges
      ////////////////////////////////////////////////////////////      

      peLeftNext = peLeft;
      peRightNext = peRight;

      ////////////////////////////////////////
      // Search for the best candidate
      bContinueSearch = TRUE;
      while ( bContinueSearch )
	{
	  bContinueSearch = FALSE;
	  bReplacePoint = FALSE;

	  ////////////////////////////////////////
	  // Search for Left candidate
	      
	  eICCT = APM2D::icctNull;

	  // Do not go past a contour edge
	  if ( peLeftNext != peLeftNext->GetFacet()->GetForwardEdge() &&
	       (peLeftNext->GetFacet()->GetForwardVertex() == NULL ||
		peLeftNext->GetVertex() != peLeftNext->GetFacet()->GetForwardVertex()) &&
	       peLeftNext != peBisector->GetEdge() &&
	       peLeftNext->GetEdgeNext(0) != peBisector->GetEdge() )
	    {
	      ASSERT( peLeftNext->GetEdge()->GetFacet() != GetNullFacet() );

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	      fprintf(stderr, "        Calculating left candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	      eICCT = InscribedCircleCenterSSSShewchuk(ptCenter, 
						       peLeftNext->GetEdge()->GetFacet(),
						       peLeft->GetFacet(),
						       peRight->GetFacet(), 
						       TRUE);
	      if ( eICCT != APM2D::icctNull )
		{
		  //////////////////////////////
		  // Check if this candidate is closer than the current left point
		  bReplacePoint = FALSE;
		  if ( peLeft == peLeftNext )
		    {
		      //////////////////////////////
		      // No left candidate yet
		      
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
		      fprintf(stderr, "        Found first left candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
		      bReplacePoint = TRUE;
		    }
		  else
		    {
		      //////////////////////////////
		      // Do an InCircle test with the current left point
		      
		      ASSERT( peLeft->GetFacet() != GetNullFacet() &&
			      peLeft->GetFacet()->GetForwardEdge() != peLeft );
		      
		      bReplacePoint = ReplaceCircleCenterShewchuk(eICCT, 
								  peLeftNext->GetEdge()->GetFacet(),
								  peLeft->GetFacet(),
								  peRight->GetFacet(), 
								  ptCenter,
								  peLeft->GetEdge()->GetFacet());

		      if ( bReplacePoint )
			{
			  //////////////////////////////
			  // Replacing left candidate
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			  fprintf(stderr, "        Updated left candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			}
		    }
		  
		  if ( bReplacePoint ) 
		    {
		      pfaLeft = peLeftNext->GetEdge()->GetFacet();
		      ptLeft = ptCenter;
		      eLeftICCT = eICCT;
		      
		      //////////////////////////////
		      // Check if this candidate is closer than the current merge point
		      
		      ASSERT( pfaMerge == NULL ||
			      pfaMerge == peLeft->GetEdge()->GetFacet() ||
			      pfaMerge == peRight->GetEdge()->GetFacet() );

		      bReplacePoint = FALSE;
		      if ( pfaMerge == peLeft->GetEdge()->GetFacet() ||
			   ( pfaMerge == NULL && 
			     peRight->GetFacet()->GetForwardEdge() == peRight ) )
			{
			  bReplacePoint = TRUE;
			}
		      else
			{
			  ASSERT( pfaMerge == NULL ||
				  pfaMerge == peRight->GetEdge()->GetFacet() );
			  ASSERT( peRight->GetEdge()->GetFacet() != GetNullFacet() );

			  bReplacePoint = ReplaceCircleCenterShewchuk(eLeftICCT, 
								      pfaLeft,
								      peLeft->GetFacet(),
								      peRight->GetFacet(), 
								      ptLeft,
								      peRight->GetEdge()->GetFacet());
			}
		      
		      if ( bReplacePoint )
			{
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			  fprintf(stderr, "        Replacing merge point with left candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			  bContinueSearch = TRUE;
			  uiMergeDirection = 1;
			  ptMerge = ptLeft;
			  pfaMerge = peLeftNext->GetEdge()->GetFacet();
			  eMergeICCT = eLeftICCT;
			}
		      
		      if ( peLeft != peLeftNext )
			{
			  //////////////////////////////
			  // Advance peLeft cleaning up extraneous edges
			  if ( peLeft->GetEdgeNext(0)->GetEdge()->GetFacet() == GetNullFacet() )
			    {
			      ASSERT( peLeft->GetEdgeNext(0) != peLeft->GetEdgeNext(0)->GetFacet()->GetForwardEdge() );
			      CollapseEdge(peLeft->GetEdgeNext(0));
			    }
			  ASSERT( peLeft != peLeft->GetFacet()->GetForwardEdge() );
			  CollapseEdge(peLeft);
			  peLeft = peLeftNext;

#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
			  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
			}
		      
		      // Advance the left next pointer
		      peLeftNext = peLeft->GetEdgeNext(0);
		      if ( peLeftNext->GetEdge()->GetFacet() == GetNullFacet() &&
			   peLeftNext != peLeftNext->GetFacet()->GetForwardEdge() &&
			   peLeftNext->GetEdge()->GetVertex() != peLeftNext->GetFacet()->GetForwardVertex() )
			{
			  peLeftNext = peLeftNext->GetEdgeNext(0);
			}
		    }
		}
	    }

	  ////////////////////////////////////////
	  // Search for Right candidate
	  
	  eICCT = APM2D::icctNull;

	  // Do not go past a contour edge
	  if ( peRightNext != peRightNext->GetFacet()->GetForwardEdge() &&
	       (peRightNext->GetFacet()->GetForwardVertex() == NULL ||
		peRightNext->GetEdge()->GetVertex() != peRightNext->GetFacet()->GetForwardVertex()) &&
	       peRightNext != peBisector &&
	       peRightNext->GetEdgePrev(0) != peBisector )
	    {
	      ASSERT( peRightNext->GetEdge()->GetFacet() != GetNullFacet() );

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	      fprintf(stderr, "        Calculating right candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	      eICCT = InscribedCircleCenterSSSShewchuk(ptCenter, 
						       peLeft->GetFacet(), 
						       peRight->GetFacet(),
						       peRightNext->GetEdge()->GetFacet(),
						       TRUE);
	      if ( eICCT != APM2D::icctNull )
		{
		  //////////////////////////////
		  // Check if this candidate is closer than the current right point
		  bReplacePoint = FALSE;
		  if ( peRight == peRightNext )
		    {
		      //////////////////////////////
		      // No right candidate yet
		      
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
		      fprintf(stderr, "        Found first right candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
		      bReplacePoint = TRUE;
		    }
		  else
		    {
		      //////////////////////////////
		      // Do an InCircle test with the current right point
		      
		      ASSERT( peRight->GetFacet() != GetNullFacet() &&
			      peRight->GetFacet()->GetForwardEdge() != peRight );
		      
		      bReplacePoint = ReplaceCircleCenterShewchuk(eICCT,
								  peLeft->GetFacet(), 
								  peRight->GetFacet(),
								  peRightNext->GetEdge()->GetFacet(),
								  ptCenter,
								  peRight->GetEdge()->GetFacet());

		      if ( bReplacePoint )
			{
			  //////////////////////////////
			  // Replacing right candidate
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			  fprintf(stderr, "        Updated right candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			}
		    }
		  
		  if ( bReplacePoint ) 
		    {
		      pfaRight = peRightNext->GetEdge()->GetFacet();
		      ptRight = ptCenter;
		      eRightICCT = eICCT;
		      
		      //////////////////////////////
		      // Check if this candidate is closer than the current merge point
		      
		      ASSERT( pfaMerge == NULL ||
			      pfaMerge == peLeft->GetEdge()->GetFacet() ||
			      pfaMerge == peRight->GetEdge()->GetFacet() );

		      bReplacePoint = FALSE;
		      if ( pfaMerge == peRight->GetEdge()->GetFacet() ||
			   ( pfaMerge == NULL && 
			     peLeft->GetFacet()->GetForwardEdge() == peLeft ) )
			{
			  bReplacePoint = TRUE;
			}
		      else
			{
			  ASSERT( pfaMerge == NULL ||
				  pfaMerge == peLeft->GetEdge()->GetFacet() );
			  ASSERT( peLeft->GetEdge()->GetFacet() != GetNullFacet() );

			  bReplacePoint = ReplaceCircleCenterShewchuk(eRightICCT, 
								      peLeft->GetFacet(),
								      peRight->GetFacet(), 
								      pfaRight,
								      ptRight,
								      peLeft->GetEdge()->GetFacet());
			}
		      
		      if ( bReplacePoint )
			{
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			  fprintf(stderr, "        Replacing merge point with right candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			  bContinueSearch = TRUE;
			  uiMergeDirection = 2;
			  ptMerge = ptCenter;
			  pfaMerge = peRightNext->GetEdge()->GetFacet();
			  eMergeICCT = eRightICCT;
			}
		      
		      if ( peRight != peRightNext )
			{
			  //////////////////////////////
			  // Advance peRight cleaning up extraneous edges
			  if ( peRight->GetEdgePrev(0)->GetEdge()->GetFacet() == GetNullFacet() )
			    {
			      ASSERT( peRight->GetEdgePrev(0) != peRight->GetEdgePrev(0)->GetFacet()->GetForwardEdge() );
			      CollapseEdge(peRight->GetEdgePrev(0)->GetEdge());
			    }
			  ASSERT( peRight != peRight->GetFacet()->GetForwardEdge() );
			  CollapseEdge(peRight->GetEdge());
			  peRight = peRightNext;

#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
			  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
			}
		      
		      // Advance the right next pointer
		      peRightNext = peRight->GetEdgePrev(0);
		      if ( peRightNext->GetEdge()->GetFacet() == GetNullFacet() &&
			   peRightNext != peRightNext->GetFacet()->GetForwardEdge() &&
			   peRightNext->GetVertex() != peRightNext->GetFacet()->GetForwardVertex() )
			{
			  peRightNext = peRightNext->GetEdgePrev(0);
			}
		    }
		}

//#define USE_SPECIAL_FIX_UP_CODE
#ifdef USE_SPECIAL_FIX_UP_CODE
	      //////////////////////////////////////////////////
	      // Because of numeric issues, the left and right might
	      // get equally good answers which will confuse things
	      // and cancel out.  We need to override that possibility
	      if ( pfaMerge == NULL &&
		   ( pfaLeft != NULL || pfaRight != NULL ) )
		{
		  if ( pfaLeft != NULL )
		    {
		      pfDistance[0] = CalculateDistance(peRight->GetEdge()->GetFacet(), ptLeft, &(ppvxEndpoint[0]));
		    }
		  if ( pfaRight != NULL )
		    {
		      pfDistance[1] = CalculateDistance(peLeft->GetEdge()->GetFacet(), ptRight, &(ppvxEndpoint[1]));
		    }

		  if ( ( pfaRight == NULL ||
			 ( pfaRight != NULL && pfaLeft != NULL &&
			   !( ptRight(Z) <= pfDistance[1] + s_cfDistanceEpsilon ) ) ) &&
		       ptLeft(Z) <= pfDistance[0] + s_cfDistanceEpsilon )
		    {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
		      fprintf(stderr, "        Special replace merge point with left candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
		      bContinueSearch = TRUE;
		      uiMergeDirection = 1;
		      ptMerge = ptLeft;
		      pfaMerge = pfaLeft;
		      eMergeICCT = eLeftICCT;
		    }
		  else if ( ( pfaLeft == NULL ||
			 ( pfaRight != NULL && pfaLeft != NULL &&
			   !( ptLeft(Z) <= pfDistance[0] + s_cfDistanceEpsilon ) ) ) &&
		       ptRight(Z) <= pfDistance[1] + s_cfDistanceEpsilon )
		    {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
		      fprintf(stderr, "        Special replace merge point with right candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
		      bContinueSearch = TRUE;
		      uiMergeDirection = 2;
		      ptMerge = ptRight;
		      pfaMerge = pfaRight;
		      eMergeICCT = eRightICCT;
		    }
		  else if ( pfaRight != NULL && pfaLeft != NULL &&
			    ptLeft(Z) <= pfDistance[0] + s_cfDistanceEpsilon &&
			    ptRight(Z) <= pfDistance[1] + s_cfDistanceEpsilon )
		    {
		      if ( ptLeft(Z) < ptRight(Z) )
			{
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			  fprintf(stderr, "        Special replace merge point with left candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			  bContinueSearch = TRUE;
			  uiMergeDirection = 1;
			  ptMerge = ptLeft;
			  pfaMerge = pfaLeft;
			  eMergeICCT = eLeftICCT;
			}
		      else
			{
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			  fprintf(stderr, "        Special replace merge point with right candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			  bContinueSearch = TRUE;
			  uiMergeDirection = 2;
			  ptMerge = ptRight;
			  pfaMerge = pfaRight;
			  eMergeICCT = eRightICCT;
			}
		    }
		}
#endif // USE_SPECIAL_FIX_UP_CODE
	    }
	}

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
      fprintf(stderr, "      END  : Searching for candidate point\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

      //////////////////////////////////////////////////
      // Check if we have a valid candidate
      if ( uiMergeDirection == 0 ) 
	{
	  break;
	} 
      else if ( uiMergeDirection == 1 ) 
	{
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	  fprintf(stderr, "      Attaching to left ( %.3f, %.3f, %.3f )\n", ptMerge(X), ptMerge(Y), ptMerge(Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	  //////////////////////////////////////////////////
	  // Attach the bisector to the left side

	  ptMerge[3] = fW;

	  peLeftNext = peLeft->GetEdge()->GetEdgeNext(0);
	  pe = peBisector;

	  pvxNew = CreateVertex(ptMerge);
	  ptLimit = ptMerge;
	  ptLimit[2] *= ptLimit[3];
	  ptLimit[3] = 1.0;
	  pvxNew->SetLimitPosition(ptLimit);
	  pvxNew->SetInscribedCircleCenterType(eMergeICCT);
	  peBisector = CreateEdge();

	  peLeft->GetEdgePrev(1)->SetVertex(peLeft->GetEdgePrev(1)->GetVertex());
	  Splice(1, peLeft, peLeft->GetEdgePrev(1));

	  peLeft->SetVertex(pvxNew);
	  pe->SetVertex(pvxNew);
	  peBisector->GetEdge()->SetVertex(pvxNew);
	  // The vertex needs to point at the bisector edge for the ICCT
	  pvxNew->SetEdge(peBisector->GetEdge());

	  Splice(1, pe, peBisector->GetEdge());
	  Splice(1, peLeft, peBisector->GetEdge());

	  peLeft = peLeftNext;

	  peBisector->SetFacet(peRight->GetFacet());
	  peBisector->GetEdge()->SetFacet(peLeft->GetFacet());

	  if ( peLeft->GetEdge()->GetFacet() == GetNullFacet() &&
	       peLeft->GetFacet()->GetForwardEdge() != peLeft && 
	       peLeft->GetFacet()->GetForwardVertex() != peLeft->GetEdge()->GetVertex() )
	    {
	      peLeftNext = peLeft->GetEdgeNext(0);
	      CollapseEdge(peLeft);
	      peLeft = peLeftNext;
	    }
	}
      else
	{
	  //////////////////////////////////////////////////
	  // Attach the bisector to the right side

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	  fprintf(stderr, "      Attaching to right ( %.3f, %.3f, %.3f )\n", ptMerge(X), ptMerge(Y), ptMerge(Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	  ptMerge[3] = fW;

	  peRightNext = peRight->GetEdge()->GetEdgePrev(0);
	  pe = peBisector;

	  pvxNew = CreateVertex(ptMerge);
	  ptLimit = ptMerge;
	  ptLimit[2] *= ptLimit[3];
	  ptLimit[3] = 1.0;
	  pvxNew->SetLimitPosition(ptLimit);
	  pvxNew->SetInscribedCircleCenterType(eMergeICCT);
	  peBisector = CreateEdge();

	  peRight->GetEdge()->GetEdgePrev(1)->SetVertex(peRight->GetEdge()->GetEdgePrev(1)->GetVertex());
	  Splice(1, peRight->GetEdge(), peRight->GetEdge()->GetEdgePrev(1));

	  peRight->GetEdge()->SetVertex(pvxNew);
	  pe->SetVertex(pvxNew);
	  peBisector->GetEdge()->SetVertex(pvxNew);
	  // The vertex needs to point at the bisector edge for the ICCT
	  pvxNew->SetEdge(peBisector->GetEdge());

	  Splice(1, pe, peBisector->GetEdge());
	  Splice(1, peRight->GetEdge(), pe);

	  peRight = peRightNext;

	  peBisector->SetFacet(peRight->GetFacet());
	  peBisector->GetEdge()->SetFacet(peLeft->GetFacet());

	  if ( peRight->GetEdge()->GetFacet() == GetNullFacet() &&
	       peRight->GetFacet()->GetForwardEdge() != peRight &&
	       peRight->GetFacet()->GetForwardVertex() != peRight->GetVertex() )
	    {
	      peRightNext = peRight->GetEdgePrev(0);
	      CollapseEdge(peRight->GetEdge());
	      peRight = peRightNext;
	    }
	}

      peLeftNext = peLeft;
      peRightNext = peRight;
    }

  //////////////////////////////////////////////////
  // Seal off the bisector
  if ( peRight->GetEdgeNext(0) == peLeft )
    {
      //////////////////////////////////////////////////
      // Finishing off the whole contour
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
      fprintf(stderr, "      Contour done attaching bisector to vertex ( %.3f, %.3f, %.3f )\n", peLeft->GetVertex()->GetPoint()(X), peLeft->GetVertex()->GetPoint()(Y), peLeft->GetVertex()->GetPoint()(Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

      ASSERT( peLeft->GetFacet()->GetForwardEdge() == peLeft &&
	      peRight->GetFacet()->GetForwardEdge() == peRight );
      Splice(1, peLeft, peBisector);
      peBisector->SetVertex(peLeft->GetVertex());
    }
  else
    {
      ASSERT( peLeft->GetVertex() != peRight->GetEdge()->GetVertex() );

      v = CalculateBisectorDirection(peLeft->GetFacet(), TRUE, 
				     peRight->GetFacet(), TRUE, FALSE);
      v[2] = fW;
      pt = v;
      pt[3] = 0.0;
      pvxNew = CreateVertex(pt);
      ptLimit = peBisector->GetEdge()->GetVertex()->GetLimitPosition() + cfInfinityMagnitude*v;
      ptLimit[3] = 1.0;
      pvxNew->SetLimitPosition(ptLimit);
      pvxNew->SetInscribedCircleCenterType(APM2D::icctInfinity);
      peBisector->SetVertex(pvxNew);

      peNew = CreateEdge();
      peNew->SetVertex(pvxNew);
      peNew->GetEdge()->SetVertex(peLeft->GetVertex());
      peNew->SetFacet(peLeft->GetFacet());
      Splice(1, peLeft, peNew->GetEdge());
      Splice(1, peBisector, peNew);

      peNew = CreateEdge();
      peNew->SetVertex(pvxNew);
      peNew->GetEdge()->SetVertex(peRight->GetEdge()->GetVertex());
      peNew->GetEdge()->SetFacet(peRight->GetFacet());
      Splice(1, peRight->GetEdge()->GetEdgePrev(1), peNew->GetEdge());
      Splice(1, peBisector, peNew);
    }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  fprintf(stderr, "    END  : Merge Left = ( %.3f, %.3f ) Right = ( %.3f, %.3f )\n", peLeftMiddle->GetVertex()->GetPoint()(X), peLeftMiddle->GetVertex()->GetPoint()(Y), peRightMiddle->GetVertex()->GetPoint()(X), peRightMiddle->GetVertex()->GetPoint()(Y));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING_OUTPUT_FILES
  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING_OUTPUT_FILES
}

VOID QEMesh::CalculateInscribedCircleCenterSSSShewchuk(Point *pptResults, UINT_32 &uiResults
, 
						       QEFacet *pfa0, 
						       QEFacet *pfa1, 
						       QEFacet *pfa2, 
						       BOOL bRadiusInZ/* = FALSE*/)
{
  ASSERT( pptResults != NULL );
  ASSERT( pfa0 != NULL && pfa1 != NULL && pfa2 != NULL );

  ASSERT( FALSE );
}

BOOL QEMesh::CounterclockwiseSSSShewchuk(QEFacet * /*pfa0*/, QEFacet * /*pfa1*/, 
					 QEFacet * /*pfa2*/)
{
  return TRUE;
}

APM2D::InscribedCircleCenterType QEMesh::InscribedCircleCenterSSSShewchuk(Point &ptCenter, 
									  QEFacet *pfa0, QEFacet *pfa1, QEFacet *pfa2, 
									  BOOL bRadiusInZ)
{
  ASSERT( pfa0 != NULL && pfa1 != NULL && pfa2 != NULL );

  APM2D::InscribedCircleCenterType eICCT;
  DOUBLE pdC[3];

  QEFacet *ppfa[3], *pfaSwap;
  QEVertex *ppvx[3];
  QEEdge *ppe[3], *pppeNeighbors[3][2], *pe;
  UINT_32 uiEdges, uiVertices, i;
  Point ppt[3];

  ppfa[0] = pfa0;
  ppfa[1] = pfa1;
  ppfa[2] = pfa2;

  ppvx[0] = NULL;
  ppvx[1] = NULL;
  ppvx[2] = NULL;

  ppe[0] = NULL;
  ppe[1] = NULL;
  ppe[2] = NULL;

  pppeNeighbors[0][0] = NULL;
  pppeNeighbors[0][1] = NULL;
  pppeNeighbors[1][0] = NULL;
  pppeNeighbors[1][1] = NULL;
  pppeNeighbors[2][0] = NULL;
  pppeNeighbors[2][1] = NULL;

  uiEdges = 0;
  uiVertices = 0;
  for ( i = 0; i < 3; i++ )
    {
      if ( ppfa[i]->GetForwardVertex() != NULL )
	{
	  ASSERT( ppfa[i]->GetForwardEdge() == NULL );
	  uiVertices++;
	}
      else
	{
	  ASSERT( ppfa[i]->GetForwardEdge() != NULL );
	  uiEdges++;
	}
    }

  switch ( uiEdges )
    {
    case 0:
      {
	for ( i = 0; i < 3; i++ )
	  {
	    ppvx[i] = ppfa[i]->GetForwardVertex();
	    for ( pe = ppvx[i]->GetEdge();
		  pe->GetFacet() == GetNullFacet() ||
		    pe->GetFacet()->GetForwardVertex() != ppvx[i];
		  pe = pe->GetEdgeNext(1) )
	      {
	      }
	    pppeNeighbors[i][0] = pe->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge();
	    ASSERT( pppeNeighbors[i][0]->GetFacet()->GetForwardEdge() == pppeNeighbors[i][0] );
	    pppeNeighbors[i][1] = pe->GetEdgePrev(1)->GetEdge();
	    ASSERT( pppeNeighbors[i][1]->GetEdge()->GetFacet()->GetForwardEdge() == pppeNeighbors[i][1]->GetEdge() );
	  }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	fprintf(stderr, "START: VVV\n");

	fprintf(stderr, "  VVV: V0: ( %.3f, %.3f ) => ( %.3f, %.3f ) => ( %.3f, %.3f )\n", 
		pppeNeighbors[0][0]->GetVertex()->GetPoint()(X),
		pppeNeighbors[0][0]->GetVertex()->GetPoint()(Y),
		ppvx[0]->GetPoint()(X),
		ppvx[0]->GetPoint()(Y),
		pppeNeighbors[0][1]->GetVertex()->GetPoint()(X),
		pppeNeighbors[0][1]->GetVertex()->GetPoint()(Y));
	fprintf(stderr, "  VVV: V1: ( %.3f, %.3f ) => ( %.3f, %.3f ) => ( %.3f, %.3f )\n", 
		pppeNeighbors[1][0]->GetVertex()->GetPoint()(X),
		pppeNeighbors[1][0]->GetVertex()->GetPoint()(Y),
		ppvx[1]->GetPoint()(X),
		ppvx[1]->GetPoint()(Y),
		pppeNeighbors[1][1]->GetVertex()->GetPoint()(X),
		pppeNeighbors[1][1]->GetVertex()->GetPoint()(Y));
	fprintf(stderr, "  VVV: V2: ( %.3f, %.3f ) => ( %.3f, %.3f ) => ( %.3f, %.3f )\n", 
		pppeNeighbors[2][0]->GetVertex()->GetPoint()(X),
		pppeNeighbors[2][0]->GetVertex()->GetPoint()(Y),
		ppvx[2]->GetPoint()(X),
		ppvx[2]->GetPoint()(Y),
		pppeNeighbors[2][1]->GetVertex()->GetPoint()(X),
		pppeNeighbors[2][1]->GetVertex()->GetPoint()(Y));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

        eICCT = APM2D::InscribedCircleCenterVVV(&pdC[0], &pdC[1], &pdC[2], 
						pppeNeighbors[0][0]->GetVertex()->GetPoint()(X),
						pppeNeighbors[0][0]->GetVertex()->GetPoint()(Y),
						ppvx[0]->GetPoint()(X),
						ppvx[0]->GetPoint()(Y),
						pppeNeighbors[0][1]->GetVertex()->GetPoint()(X),
						pppeNeighbors[0][1]->GetVertex()->GetPoint()(Y),
						pppeNeighbors[1][0]->GetVertex()->GetPoint()(X),
						pppeNeighbors[1][0]->GetVertex()->GetPoint()(Y),
						ppvx[1]->GetPoint()(X),
						ppvx[1]->GetPoint()(Y),
						pppeNeighbors[1][1]->GetVertex()->GetPoint()(X),
						pppeNeighbors[1][1]->GetVertex()->GetPoint()(Y),
						pppeNeighbors[2][0]->GetVertex()->GetPoint()(X),
						pppeNeighbors[2][0]->GetVertex()->GetPoint()(Y),
						ppvx[2]->GetPoint()(X),
						ppvx[2]->GetPoint()(Y),
						pppeNeighbors[2][1]->GetVertex()->GetPoint()(X),
						pppeNeighbors[2][1]->GetVertex()->GetPoint()(Y));
	ptCenter.Set(pdC[0], pdC[1], pdC[2]);
		
	switch ( APM::GetAccuracyType() )
	  {
	  case APM::atGlobal:
	  case APM::atLocal:
	  case APM::atExact:
	    if ( eICCT != APM2D::icctNull && ptCenter(2) > s_cfVoronoiInfinity )
	      {
		eICCT = APM2D::icctNull;
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
		fprintf(stderr, "  VVV: Point at Infinity: ( %.6f, %.6f, %.6f )\n", 
			ptCenter(X), ptCenter(Y), ptCenter(Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	      }
	    break;
	  case APM::atMPFUN:
	  case APM::atMPFUNFilter:
	  case APM::atMPFUNNormalized:
	    break;
	  default:
	    ASSERT( FALSE );
	    break;
	  }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	if ( eICCT == APM2D::icctNull )
	  {
	    fprintf(stderr, "  VVV: No answers\n");
	  }
	else
	  {
	    fprintf(stderr, "  VVV: Center: ( %.6f, %.6f, %.6f )\n", 
		   ptCenter(X), ptCenter(Y), ptCenter(Z));
	  }

	fprintf(stderr, "END  : VVV\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
      }
      break;

    case 1:
      {
	while ( ppfa[2]->GetForwardEdge() == NULL )
	  {
	    pfaSwap = ppfa[0];
	    ppfa[0] = ppfa[1];
	    ppfa[1] = ppfa[2];
	    ppfa[2] = pfaSwap;
	  }

	for ( i = 0; i < 2; i++ )
	  {
	    ppvx[i] = ppfa[i]->GetForwardVertex();
	    for ( pe = ppvx[i]->GetEdge();
		  pe->GetFacet() == GetNullFacet() ||
		    pe->GetFacet()->GetForwardVertex() != ppvx[i];
		  pe = pe->GetEdgeNext(1) )
	      {
	      }
	    pppeNeighbors[i][0] = pe->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge();
	    ASSERT( pppeNeighbors[i][0]->GetFacet()->GetForwardEdge() == pppeNeighbors[i][0] );
	    pppeNeighbors[i][1] = pe->GetEdgePrev(1)->GetEdge();
	    ASSERT( pppeNeighbors[i][1]->GetEdge()->GetFacet()->GetForwardEdge() == pppeNeighbors[i][1]->GetEdge() );
	  }
	for ( i = 2; i < 3; i++ )
	  {
	    ppe[i] = ppfa[i]->GetForwardEdge();
	  }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	fprintf(stderr, "START: VVE\n");

	fprintf(stderr, "  VVE: V0: ( %.3f, %.3f ) => ( %.3f, %.3f ) => ( %.3f, %.3f )\n", 
		pppeNeighbors[0][0]->GetVertex()->GetPoint()(X),
		pppeNeighbors[0][0]->GetVertex()->GetPoint()(Y),
		ppvx[0]->GetPoint()(X),
		ppvx[0]->GetPoint()(Y),
		pppeNeighbors[0][1]->GetVertex()->GetPoint()(X),
		pppeNeighbors[0][1]->GetVertex()->GetPoint()(Y));
	fprintf(stderr, "  VVE: V1: ( %.3f, %.3f ) => ( %.3f, %.3f ) => ( %.3f, %.3f )\n", 
		pppeNeighbors[1][0]->GetVertex()->GetPoint()(X),
		pppeNeighbors[1][0]->GetVertex()->GetPoint()(Y),
		ppvx[1]->GetPoint()(X),
		ppvx[1]->GetPoint()(Y),
		pppeNeighbors[1][1]->GetVertex()->GetPoint()(X),
		pppeNeighbors[1][1]->GetVertex()->GetPoint()(Y));
	fprintf(stderr, "  VVE: E2: ( %.3f, %.3f ) -> ( %.3f, %.3f )\n", 
	       ppe[2]->GetVertex()->GetPoint()(X),
	       ppe[2]->GetVertex()->GetPoint()(Y),
	       ppe[2]->GetEdge()->GetVertex()->GetPoint()(X),
	       ppe[2]->GetEdge()->GetVertex()->GetPoint()(Y));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	ASSERT( !( ppvx[0] == ppe[2]->GetEdge()->GetVertex() &&
		   ppvx[1] == ppe[2]->GetVertex() ) );
	
	if ( ppvx[0] == ppe[2]->GetEdge()->GetVertex() )
	  {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	    fprintf(stderr, "  VVE: V0 on E2\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	    eICCT = APM2D::InscribedCircleCenterEvV(&pdC[0], &pdC[1], &pdC[2], 
						    ppe[2]->GetVertex()->GetPoint()(0),
						    ppe[2]->GetVertex()->GetPoint()(1),
						    ppe[2]->GetEdge()->GetVertex()->GetPoint()(0),
						    ppe[2]->GetEdge()->GetVertex()->GetPoint()(1),
						    pppeNeighbors[1][0]->GetVertex()->GetPoint()(X),
						    pppeNeighbors[1][0]->GetVertex()->GetPoint()(Y),
						    ppvx[1]->GetPoint()(X),
						    ppvx[1]->GetPoint()(Y),
						    pppeNeighbors[1][1]->GetVertex()->GetPoint()(X),
						    pppeNeighbors[1][1]->GetVertex()->GetPoint()(Y));
	    ptCenter.Set(pdC[0], pdC[1], pdC[2]);
	  }
	else if ( ppvx[1] == ppe[2]->GetVertex() )
	  {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	    fprintf(stderr, "  VVE: V1 on E2\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	    eICCT = APM2D::InscribedCircleCenterVvE(&pdC[0], &pdC[1], &pdC[2], 
						    pppeNeighbors[0][0]->GetVertex()->GetPoint()(X),
						    pppeNeighbors[0][0]->GetVertex()->GetPoint()(Y),
						    ppvx[0]->GetPoint()(X),
						    ppvx[0]->GetPoint()(Y),
						    pppeNeighbors[0][1]->GetVertex()->GetPoint()(X),
						    pppeNeighbors[0][1]->GetVertex()->GetPoint()(Y),
						    ppe[2]->GetVertex()->GetPoint()(0),
						    ppe[2]->GetVertex()->GetPoint()(1),
						    ppe[2]->GetEdge()->GetVertex()->GetPoint()(0),
						    ppe[2]->GetEdge()->GetVertex()->GetPoint()(1));
	    ptCenter.Set(pdC[0], pdC[1], pdC[2]);
	  }
	else
	  {
	    eICCT = APM2D::InscribedCircleCenterVVE(&pdC[0], &pdC[1], &pdC[2], 
						    pppeNeighbors[0][0]->GetVertex()->GetPoint()(X),
						    pppeNeighbors[0][0]->GetVertex()->GetPoint()(Y),
						    ppvx[0]->GetPoint()(X),
						    ppvx[0]->GetPoint()(Y),
						    pppeNeighbors[0][1]->GetVertex()->GetPoint()(X),
						    pppeNeighbors[0][1]->GetVertex()->GetPoint()(Y),
						    pppeNeighbors[1][0]->GetVertex()->GetPoint()(X),
						    pppeNeighbors[1][0]->GetVertex()->GetPoint()(Y),
						    ppvx[1]->GetPoint()(X),
						    ppvx[1]->GetPoint()(Y),
						    pppeNeighbors[1][1]->GetVertex()->GetPoint()(X),
						    pppeNeighbors[1][1]->GetVertex()->GetPoint()(Y),
						    ppe[2]->GetVertex()->GetPoint()(0),
						    ppe[2]->GetVertex()->GetPoint()(1),
						    ppe[2]->GetEdge()->GetVertex()->GetPoint()(0),
						    ppe[2]->GetEdge()->GetVertex()->GetPoint()(1));
	    ptCenter.Set(pdC[0], pdC[1], pdC[2]);
	  }

	switch ( APM::GetAccuracyType() )
	  {
	  case APM::atGlobal:
	  case APM::atLocal:
	  case APM::atExact:
	    if ( eICCT != APM2D::icctNull && ptCenter(2) > s_cfVoronoiInfinity )
	      {
		eICCT = APM2D::icctNull;
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
		fprintf(stderr, "  VVE: Point at Infinity: ( %.6f, %.6f, %.6f )\n", 
			ptCenter(X), ptCenter(Y), ptCenter(Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	      }
	    break;
	  case APM::atMPFUN:
	  case APM::atMPFUNFilter:
	  case APM::atMPFUNNormalized:
	    break;
	  default:
	    ASSERT( FALSE );
	    break;
	  }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	if ( eICCT == APM2D::icctNull )
	  {
	    fprintf(stderr, "  VVE: No answers\n");
	  }
	else
	  {
	    fprintf(stderr, "  VVE: Center: ( %.6f, %.6f, %.6f )\n", 
		   ptCenter(X), ptCenter(Y), ptCenter(Z));
	  }

	fprintf(stderr, "END  : VVE\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
      }
      break;
    
    case 2:
      {
	while ( ppfa[0]->GetForwardVertex() == NULL )
	  {
	    pfaSwap = ppfa[0];
	    ppfa[0] = ppfa[1];
	    ppfa[1] = ppfa[2];
	    ppfa[2] = pfaSwap;
	  }

	for ( i = 0; i < 1; i++ )
	  {
	    ppvx[i] = ppfa[i]->GetForwardVertex();
	    for ( pe = ppvx[i]->GetEdge();
		  pe->GetFacet() == GetNullFacet() ||
		    pe->GetFacet()->GetForwardVertex() != ppvx[i];
		  pe = pe->GetEdgeNext(1) )
	      {
	      }
	    pppeNeighbors[i][0] = pe->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge();
	    ASSERT( pppeNeighbors[i][0]->GetFacet()->GetForwardEdge() == pppeNeighbors[i][0] );
	    pppeNeighbors[i][1] = pe->GetEdgePrev(1)->GetEdge();
	    ASSERT( pppeNeighbors[i][1]->GetEdge()->GetFacet()->GetForwardEdge() == pppeNeighbors[i][1]->GetEdge() );
	  }
	for ( i = 1; i < 3; i++ )
	  {
	    ppe[i] = ppfa[i]->GetForwardEdge();
	  }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	fprintf(stderr, "START: VEE\n");

	fprintf(stderr, "  VEE: V0: ( %.3f, %.3f ) => ( %.3f, %.3f ) => ( %.3f, %.3f )\n", 
		pppeNeighbors[0][0]->GetVertex()->GetPoint()(X),
		pppeNeighbors[0][0]->GetVertex()->GetPoint()(Y),
		ppvx[0]->GetPoint()(X),
		ppvx[0]->GetPoint()(Y),
		pppeNeighbors[0][1]->GetVertex()->GetPoint()(X),
		pppeNeighbors[0][1]->GetVertex()->GetPoint()(Y));
	fprintf(stderr, "  VEE: E1: ( %.3f, %.3f ) -> ( %.3f, %.3f )\n", 
	       ppe[1]->GetVertex()->GetPoint()(X),
	       ppe[1]->GetVertex()->GetPoint()(Y),
	       ppe[1]->GetEdge()->GetVertex()->GetPoint()(X),
	       ppe[1]->GetEdge()->GetVertex()->GetPoint()(Y));
	fprintf(stderr, "  VEE: E2: ( %.3f, %.3f ) -> ( %.3f, %.3f )\n", 
	       ppe[2]->GetVertex()->GetPoint()(X),
	       ppe[2]->GetVertex()->GetPoint()(Y),
	       ppe[2]->GetEdge()->GetVertex()->GetPoint()(X),
	       ppe[2]->GetEdge()->GetVertex()->GetPoint()(Y));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	if ( ppvx[0] == ppe[1]->GetVertex() &&
	     ppvx[0] == ppe[2]->GetEdge()->GetVertex() )
	  {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	    fprintf(stderr, "  VEE: V0 on E1 and E2\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	    eICCT = APM2D::icctEvE;
	    ptCenter = ppvx[0]->GetPoint();
	    ptCenter[Z] = 0.0; 
	  }
	else if ( ppvx[0] == ppe[1]->GetVertex() )
	  {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	    fprintf(stderr, "  VEE: V0 on E1\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	    if ( ppe[1]->GetEdge()->GetVertex() == ppe[2]->GetVertex() )
	      {
		eICCT = APM2D::InscribedCircleCentervEvE(&pdC[0], &pdC[1], &pdC[2], 
							 ppe[1]->GetVertex()->GetPoint()(0),
							 ppe[1]->GetVertex()->GetPoint()(1),
							 ppe[1]->GetEdge()->GetVertex()->GetPoint()(0),
							 ppe[1]->GetEdge()->GetVertex()->GetPoint()(1),
							 ppe[2]->GetEdge()->GetVertex()->GetPoint()(0),
							 ppe[2]->GetEdge()->GetVertex()->GetPoint()(1));
		ptCenter.Set(pdC[0], pdC[1], pdC[2]);
	      }
	    else
	      {
		eICCT = APM2D::InscribedCircleCentervEE(&pdC[0], &pdC[1], &pdC[2], 
							ppe[1]->GetVertex()->GetPoint()(0),
							ppe[1]->GetVertex()->GetPoint()(1),
							ppe[1]->GetEdge()->GetVertex()->GetPoint()(0),
							ppe[1]->GetEdge()->GetVertex()->GetPoint()(1),
							ppe[2]->GetVertex()->GetPoint()(0),
							ppe[2]->GetVertex()->GetPoint()(1),
							ppe[2]->GetEdge()->GetVertex()->GetPoint()(0),
							ppe[2]->GetEdge()->GetVertex()->GetPoint()(1));
		ptCenter.Set(pdC[0], pdC[1], pdC[2]);
	      }
	  }
	else if ( ppvx[0] == ppe[2]->GetEdge()->GetVertex() )
	  {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	    fprintf(stderr, "  VEE: V0 on E2\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	    if ( ppe[1]->GetEdge()->GetVertex() == ppe[2]->GetVertex() )
	      {
		eICCT = APM2D::InscribedCircleCenterEvEv(&pdC[0], &pdC[1], &pdC[2], 
							 ppe[1]->GetVertex()->GetPoint()(0),
							 ppe[1]->GetVertex()->GetPoint()(1),
							 ppe[1]->GetEdge()->GetVertex()->GetPoint()(0),
							 ppe[1]->GetEdge()->GetVertex()->GetPoint()(1),
							 ppe[2]->GetEdge()->GetVertex()->GetPoint()(0),
							 ppe[2]->GetEdge()->GetVertex()->GetPoint()(1));
		ptCenter.Set(pdC[0], pdC[1], pdC[2]);
	      }
	    else
	      {
		eICCT = APM2D::InscribedCircleCenterEEv(&pdC[0], &pdC[1], &pdC[2], 
							ppe[1]->GetVertex()->GetPoint()(0),
							ppe[1]->GetVertex()->GetPoint()(1),
							ppe[1]->GetEdge()->GetVertex()->GetPoint()(0),
							ppe[1]->GetEdge()->GetVertex()->GetPoint()(1),
							ppe[2]->GetVertex()->GetPoint()(0),
							ppe[2]->GetVertex()->GetPoint()(1),
							ppe[2]->GetEdge()->GetVertex()->GetPoint()(0),
							ppe[2]->GetEdge()->GetVertex()->GetPoint()(1));
		ptCenter.Set(pdC[0], pdC[1], pdC[2]);
	      }
	  }
	else
	  {
	    if ( ppe[1]->GetEdge()->GetVertex() == ppe[2]->GetVertex() )
	      {
		eICCT = APM2D::InscribedCircleCenterVEvE(&pdC[0], &pdC[1], &pdC[2], 
							 pppeNeighbors[0][0]->GetVertex()->GetPoint()(X),
							 pppeNeighbors[0][0]->GetVertex()->GetPoint()(Y),
							 ppvx[0]->GetPoint()(X),
							 ppvx[0]->GetPoint()(Y),
							 pppeNeighbors[0][1]->GetVertex()->GetPoint()(X),
							 pppeNeighbors[0][1]->GetVertex()->GetPoint()(Y),
							 ppe[1]->GetVertex()->GetPoint()(0),
							 ppe[1]->GetVertex()->GetPoint()(1),
							 ppe[1]->GetEdge()->GetVertex()->GetPoint()(0),
							 ppe[1]->GetEdge()->GetVertex()->GetPoint()(1),
							 ppe[2]->GetEdge()->GetVertex()->GetPoint()(0),
							 ppe[2]->GetEdge()->GetVertex()->GetPoint()(1));
		ptCenter.Set(pdC[0], pdC[1], pdC[2]);
	      }
	    else
	      {
		eICCT = APM2D::InscribedCircleCenterVEE(&pdC[0], &pdC[1], &pdC[2], 
							pppeNeighbors[0][0]->GetVertex()->GetPoint()(X),
							pppeNeighbors[0][0]->GetVertex()->GetPoint()(Y),
							ppvx[0]->GetPoint()(X),
							ppvx[0]->GetPoint()(Y),
							pppeNeighbors[0][1]->GetVertex()->GetPoint()(X),
							pppeNeighbors[0][1]->GetVertex()->GetPoint()(Y),
							ppe[1]->GetVertex()->GetPoint()(0),
							ppe[1]->GetVertex()->GetPoint()(1),
							ppe[1]->GetEdge()->GetVertex()->GetPoint()(0),
							ppe[1]->GetEdge()->GetVertex()->GetPoint()(1),
							ppe[2]->GetVertex()->GetPoint()(0),
							ppe[2]->GetVertex()->GetPoint()(1),
							ppe[2]->GetEdge()->GetVertex()->GetPoint()(0),
							ppe[2]->GetEdge()->GetVertex()->GetPoint()(1));
		ptCenter.Set(pdC[0], pdC[1], pdC[2]);
	      }
	  }
	  
	switch ( APM::GetAccuracyType() )
	  {
	  case APM::atGlobal:
	  case APM::atLocal:
	  case APM::atExact:
	    if ( eICCT != APM2D::icctNull && ptCenter(2) > s_cfVoronoiInfinity )
	      {
		eICCT = APM2D::icctNull;
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
		fprintf(stderr, "  VEE: Point at Infinity: ( %.6f, %.6f, %.6f )\n", 
			ptCenter(X), ptCenter(Y), ptCenter(Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	      }
	    break;
	  case APM::atMPFUN:
	  case APM::atMPFUNFilter:
	  case APM::atMPFUNNormalized:
	    break;
	  default:
	    ASSERT( FALSE );
	    break;
	  }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	if ( eICCT == APM2D::icctNull )
	  {
	    fprintf(stderr, "  VEE: No answers\n");
	  }
	else
	  {
	    fprintf(stderr, "  VEE: Center: ( %.6f, %.6f, %.6f )\n", 
		   ptCenter(X), ptCenter(Y), ptCenter(Z));
	  }

	fprintf(stderr, "END  : VEE\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
      }
      break;
    
    case 3:
      {
	UINT_32 j, uiSharedVertices;
	
	j = 2;
	uiSharedVertices = 0;
	for ( i = 0; i < 3; i++ )
	  {
	    if ( ppfa[j]->GetForwardEdge()->GetEdge()->GetVertex() == 
		 ppfa[i]->GetForwardEdge()->GetVertex() )
	      {
		uiSharedVertices++;
	      }
	    j = i;
	  }

	if ( uiSharedVertices > 0 && uiSharedVertices < 3 )
	  {
	    while ( ( ppfa[0]->GetForwardEdge()->GetEdge()->GetVertex() != 
		      ppfa[1]->GetForwardEdge()->GetVertex() ) ||
		    ( ppfa[2]->GetForwardEdge()->GetEdge()->GetVertex() == 
		      ppfa[0]->GetForwardEdge()->GetVertex() ) )
	      {
		pfaSwap = ppfa[0];
		ppfa[0] = ppfa[1];
		ppfa[1] = ppfa[2];
		ppfa[2] = pfaSwap;
	      }	    
	  }

	for ( i = 0; i < 3; i++ )
	  {
	    ppe[i] = ppfa[i]->GetForwardEdge();
	  }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	fprintf(stderr, "START: EEE\n");

	fprintf(stderr, "  EEE: E0: ( %.3f, %.3f ) -> ( %.3f, %.3f )\n", 
	       ppe[0]->GetVertex()->GetPoint()(X),
	       ppe[0]->GetVertex()->GetPoint()(Y),
	       ppe[0]->GetEdge()->GetVertex()->GetPoint()(X),
	       ppe[0]->GetEdge()->GetVertex()->GetPoint()(Y));
	fprintf(stderr, "  EEE: E1: ( %.3f, %.3f ) -> ( %.3f, %.3f )\n", 
	       ppe[1]->GetVertex()->GetPoint()(X),
	       ppe[1]->GetVertex()->GetPoint()(Y),
	       ppe[1]->GetEdge()->GetVertex()->GetPoint()(X),
	       ppe[1]->GetEdge()->GetVertex()->GetPoint()(Y));
	fprintf(stderr, "  EEE: E2: ( %.3f, %.3f ) -> ( %.3f, %.3f )\n", 
	       ppe[2]->GetVertex()->GetPoint()(X),
	       ppe[2]->GetVertex()->GetPoint()(Y),
	       ppe[2]->GetEdge()->GetVertex()->GetPoint()(X),
	       ppe[2]->GetEdge()->GetVertex()->GetPoint()(Y));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	switch ( uiSharedVertices )
	  {
	  case 0:
	    eICCT = APM2D::InscribedCircleCenterEEE(&pdC[0], &pdC[1], &pdC[2], 
						      ppe[0]->GetVertex()->GetPoint()(0),
						      ppe[0]->GetVertex()->GetPoint()(1),
						      ppe[0]->GetEdge()->GetVertex()->GetPoint()(0),
						      ppe[0]->GetEdge()->GetVertex()->GetPoint()(1),
						      ppe[1]->GetVertex()->GetPoint()(0),
						      ppe[1]->GetVertex()->GetPoint()(1),
						      ppe[1]->GetEdge()->GetVertex()->GetPoint()(0),
						      ppe[1]->GetEdge()->GetVertex()->GetPoint()(1),
						      ppe[2]->GetVertex()->GetPoint()(0),
						      ppe[2]->GetVertex()->GetPoint()(1),
						      ppe[2]->GetEdge()->GetVertex()->GetPoint()(0),
						      ppe[2]->GetEdge()->GetVertex()->GetPoint()(1));
	    ptCenter.Set(pdC[0], pdC[1], pdC[2]);
	    break;
	  case 1:
	    eICCT = APM2D::InscribedCircleCenterEvEE(&pdC[0], &pdC[1], &pdC[2], 
						     ppe[0]->GetVertex()->GetPoint()(0),
						     ppe[0]->GetVertex()->GetPoint()(1),
						     ppe[1]->GetVertex()->GetPoint()(0),
						     ppe[1]->GetVertex()->GetPoint()(1),
						     ppe[1]->GetEdge()->GetVertex()->GetPoint()(0),
						     ppe[1]->GetEdge()->GetVertex()->GetPoint()(1),
						     ppe[2]->GetVertex()->GetPoint()(0),
						     ppe[2]->GetVertex()->GetPoint()(1),
						     ppe[2]->GetEdge()->GetVertex()->GetPoint()(0),
						     ppe[2]->GetEdge()->GetVertex()->GetPoint()(1));
	    ptCenter.Set(pdC[0], pdC[1], pdC[2]);
	    break;
	  case 2:
	    eICCT = APM2D::InscribedCircleCenterEvEvE(&pdC[0], &pdC[1], &pdC[2], 
							ppe[0]->GetVertex()->GetPoint()(0),
							ppe[0]->GetVertex()->GetPoint()(1),
							ppe[1]->GetVertex()->GetPoint()(0),
							ppe[1]->GetVertex()->GetPoint()(1),
							ppe[2]->GetVertex()->GetPoint()(0),
							ppe[2]->GetVertex()->GetPoint()(1),
							ppe[2]->GetEdge()->GetVertex()->GetPoint()(0),
							ppe[2]->GetEdge()->GetVertex()->GetPoint()(1));
	    ptCenter.Set(pdC[0], pdC[1], pdC[2]);
	    break;
	  case 3:
	    eICCT = APM2D::InscribedCircleCenterEvEvEv(&pdC[0], &pdC[1], &pdC[2], 
							 ppe[0]->GetVertex()->GetPoint()(0),
							 ppe[0]->GetVertex()->GetPoint()(1),
							 ppe[1]->GetVertex()->GetPoint()(0),
							 ppe[1]->GetVertex()->GetPoint()(1),
							 ppe[2]->GetVertex()->GetPoint()(0),
							 ppe[2]->GetVertex()->GetPoint()(1));
	    ptCenter.Set(pdC[0], pdC[1], pdC[2]);
	    break;
	  default:
	    ASSERT( FALSE );
	    break;
	  }

	switch ( APM::GetAccuracyType() )
	  {
	  case APM::atGlobal:
	  case APM::atLocal:
	  case APM::atExact:
	    if ( eICCT != APM2D::icctNull && ptCenter(2) > s_cfVoronoiInfinity )
	      {
		eICCT = APM2D::icctNull;
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
		fprintf(stderr, "  EEE: Point at Infinity: ( %.6f, %.6f, %.6f )\n", 
			ptCenter(X), ptCenter(Y), ptCenter(Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	      }
	    break;
	  case APM::atMPFUN:
	  case APM::atMPFUNFilter:
	  case APM::atMPFUNNormalized:
	    break;
	  default:
	    ASSERT( FALSE );
	    break;
	  }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	if ( eICCT == APM2D::icctNull )
	  {
	    fprintf(stderr, "  EEE: No answers\n");
	  }
	else
	  {
	    fprintf(stderr, "  EEE: Center: ( %.6f, %.6f, %.6f )\n", 
		   ptCenter(X), ptCenter(Y), ptCenter(Z));
	  }

	fprintf(stderr, "END  : EEE\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
      }
      break;
    
    default:
      {
	ASSERT( FALSE );
      }
      break;
    }

  if ( !bRadiusInZ )
    {
      ptCenter[2] = 0.0;
    }

  return eICCT;
}

BOOL QEMesh::ReplaceCircleCenterShewchuk(APM2D::InscribedCircleCenterType eICCT,
					 QEFacet *pfa0, QEFacet *pfa1, QEFacet *pfa2, 
					 Point &ptCenter,
					 QEFacet *pfa3)
{
  APM2D::CircleTestType eCircleTest; 

  eCircleTest = InCircleSSS_SShewchuk(eICCT,
				      pfa0, pfa1, pfa2, 
				      ptCenter,
				      pfa3);
  //return ( eCircleTest == APM2D::cttOutCircle );

  // I don't know what to do about the point on circle case!!!!!
  BOOL bInCircle;
  bInCircle = ( eCircleTest == APM2D::cttInCircle );
  return !bInCircle;
}

APM2D::CircleTestType QEMesh::InCircleSSS_SShewchuk(APM2D::InscribedCircleCenterType eICCT,
						    QEFacet *pfa0, 
						    QEFacet *pfa1, 
						    QEFacet *pfa2, 
						    Point &ptCenter,
						    QEFacet *pfa3)
{
  ASSERT( pfa0 != NULL && pfa1 != NULL && pfa2 != NULL && pfa3 != NULL );

  if ( pfa3->GetForwardVertex() != NULL )
    {
      return InCircleSSS_VShewchuk(eICCT,
				   pfa0, pfa1, pfa2, 
				   ptCenter,
				   pfa3, pfa3->GetForwardVertex());
    }
  else
    {
      ASSERT( pfa3->GetForwardEdge() != NULL );
      return InCircleSSS_EShewchuk(eICCT,
				   pfa0, pfa1, pfa2, 
				   ptCenter,
				   pfa3, pfa3->GetForwardEdge());
    }
}

APM2D::CircleTestType QEMesh::InCircleSSS_VShewchuk(APM2D::InscribedCircleCenterType eICCT,
						    QEFacet *pfa0, 
						    QEFacet *pfa1, 
						    QEFacet *pfa2, 
						    Point &/*ptCenter*/,
						    QEFacet */*pfa3*/, QEVertex *pvx3)
{
  ASSERT( pfa0 != NULL && pfa1 != NULL && pfa2 != NULL && pvx3 != NULL );

  QEFacet *ppfa[3], *pfaSwap;
  QEVertex *ppvx[3];
  QEEdge *ppe[3];
  Point pppt[4][2];
  UINT_32 uiEdges, uiVertices, uiSharedVertices;
  UINT_32 i, j;
  APM2D::CircleTestType eCircleTest;

  ppfa[0] = pfa0;
  ppfa[1] = pfa1;
  ppfa[2] = pfa2;

  ppvx[0] = NULL;
  ppvx[1] = NULL;
  ppvx[2] = NULL;

  ppe[0] = NULL;
  ppe[1] = NULL;
  ppe[2] = NULL;

  //////////////////////////////////////////////////
  // The fourth site is treated differently
  pppt[3][0] = pvx3->GetPoint();

  //////////////////////////////////////////////////
  // Deal with points at infinity
  if ( eICCT == APM2D::icctInfinity )
    {
      INT_32 iResult;

      ASSERT( ppfa[0] == GetNullFacet() ||
	      ppfa[1] == GetNullFacet() ||
	      ppfa[2] == GetNullFacet() );
      while ( ppfa[2] != GetNullFacet() )
	{
	  pfaSwap = ppfa[0];
	  ppfa[0] = ppfa[1];
	  ppfa[1] = ppfa[2];
	  ppfa[2] = pfaSwap;
	}
      ASSERT( ppfa[0] != GetNullFacet() &&
	      ppfa[1] != GetNullFacet() );

      ppvx[0] = ppfa[0]->GetForwardVertex();
      ppvx[1] = ppfa[1]->GetForwardVertex();
      ASSERT( ppvx[0] != NULL || ppvx[1] != NULL );
      if ( ppvx[0] == NULL )
	{
	switch ( APM::GetAccuracyType() )
	  {
	  case APM::atGlobal:
	  case APM::atLocal:
	  case APM::atExact:
	    ppe[0] = ppfa[0]->GetForwardEdge();
	    // This assertion is less correct because finite line segments
	    // should not be able to create points at infinity with non-adjacent vertices
	    // but numerical issues make us do a work around.  Use the more distant endpoint
	    ASSERT( ppe[0] != NULL &&
		    ppvx[1] != NULL );
	    pppt[0][0] = ppe[0]->GetVertex()->GetPoint();
	    //pppt[1][0] = ppe[0]->GetEdge()->GetVertex()->GetPoint();
	    pppt[1][0] = ppvx[1]->GetPoint();
	    break;
	  case APM::atMPFUN:
	  case APM::atMPFUNFilter:
	  case APM::atMPFUNNormalized:
	    ppe[0] = ppfa[0]->GetForwardEdge();
	    ASSERT( ppe[0] != NULL &&
		    ppvx[1] != NULL &&
		    ( ppe[0]->GetVertex() == ppvx[1] ||
		      ppe[0]->GetEdge()->GetVertex() == ppvx[1] ) );
	    pppt[0][0] = ppe[0]->GetVertex()->GetPoint();
	    pppt[1][0] = ppe[0]->GetEdge()->GetVertex()->GetPoint();
	    break;
	  default:
	    ASSERT( FALSE );
	    break;
	  }
	}
      else if ( ppvx[1] == NULL )
	{
	switch ( APM::GetAccuracyType() )
	  {
	  case APM::atGlobal:
	  case APM::atLocal:
	  case APM::atExact:
	    ppe[1] = ppfa[1]->GetForwardEdge();
	    // This assertion is less correct because finite line segments
	    // should not be able to create points at infinity with non-adjacent vertices
	    // but numerical issues make us do a work around.  Use the more distant endpoint
	    ASSERT( ppe[1] != NULL &&
		    ppvx[0] != NULL );
	    pppt[0][0] = ppvx[0]->GetPoint();
	    //pppt[0][0] = ppe[1]->GetVertex()->GetPoint();
	    pppt[1][0] = ppe[1]->GetEdge()->GetVertex()->GetPoint();
	    break;
	  case APM::atMPFUN:
	  case APM::atMPFUNFilter:
	  case APM::atMPFUNNormalized:
	    ppe[1] = ppfa[1]->GetForwardEdge();
	    ASSERT( ppe[1] != NULL &&
		    ppvx[0] != NULL &&
		    ( ppe[1]->GetVertex() == ppvx[0] ||
		      ppe[1]->GetEdge()->GetVertex() == ppvx[0] ) );
	    pppt[0][0] = ppe[1]->GetVertex()->GetPoint();
	    pppt[1][0] = ppe[1]->GetEdge()->GetVertex()->GetPoint();
	    break;
	  default:
	    ASSERT( FALSE );
	    break;
	  }
	}
      else
	{
	  pppt[0][0] = ppvx[0]->GetPoint();
	  pppt[1][0] = ppvx[1]->GetPoint();
	}

      iResult = APM2D::OrientationVVV(pppt[0][0](0), pppt[0][0](1), 
				      pppt[1][0](0), pppt[1][0](1), 
				      pppt[3][0](0), pppt[3][0](1));
      if ( iResult > 0 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( iResult == 0 )
	{
	  return APM2D::cttOnCircle;
	}
      else
	{
	  return APM2D::cttOutCircle;
	}
    }

  uiEdges = 0;
  uiVertices = 0;
  uiSharedVertices = 0;
  for ( i = 0; i < 3; i++ )
    {
      if ( ppfa[i]->GetForwardVertex() != NULL )
	{
	  ASSERT( ppfa[i]->GetForwardEdge() == NULL );
	  uiVertices++;
	  if ( pvx3 == ppfa[i]->GetForwardVertex() )
	    {
	      return APM2D::cttOnCircle;
	    }
	}
      else
	{
	  ASSERT( ppfa[i]->GetForwardEdge() != NULL );
	  uiEdges++;
	}
    }

  switch ( uiEdges )
    {
    case 0:
      {
	for ( i = 0; i < 3; i++ )
	  {
	    ppvx[i] = ppfa[i]->GetForwardVertex();
	    pppt[i][0] = ppvx[i]->GetPoint();
	  }
      }
    break;
    case 1:
      {
	while ( ppfa[2]->GetForwardEdge() == NULL )
	  {
	    pfaSwap = ppfa[0];
	    ppfa[0] = ppfa[1];
	    ppfa[1] = ppfa[2];
	    ppfa[2] = pfaSwap;
	  }

	for ( i = 0; i < 2; i++ )
	  {
	    ppvx[i] = ppfa[i]->GetForwardVertex();
	    pppt[i][0] = ppvx[i]->GetPoint();
	  }
	for ( i = 2; i < 3; i++ )
	  {
	    ppe[i] = ppfa[i]->GetForwardEdge();
	    pppt[i][0] = ppe[i]->GetVertex()->GetPoint();
	    pppt[i][1] = ppe[i]->GetEdge()->GetVertex()->GetPoint();
	  }
      }
    break;
    case 2:
      {
	while ( ppfa[0]->GetForwardVertex() == NULL )
	  {
	    pfaSwap = ppfa[0];
	    ppfa[0] = ppfa[1];
	    ppfa[1] = ppfa[2];
	    ppfa[2] = pfaSwap;
	  }

	for ( i = 0; i < 1; i++ )
	  {
	    ppvx[i] = ppfa[i]->GetForwardVertex();
	    pppt[i][0] = ppvx[i]->GetPoint();
	  }
	for ( i = 1; i < 3; i++ )
	  {
	    ppe[i] = ppfa[i]->GetForwardEdge();
	    pppt[i][0] = ppe[i]->GetVertex()->GetPoint();
	    pppt[i][1] = ppe[i]->GetEdge()->GetVertex()->GetPoint();
	  }
      }
    break;
    case 3:
      {
	j = 2;
	uiSharedVertices = 0;
	for ( i = 0; i < 3; i++ )
	  {
	    if ( ppfa[j]->GetForwardEdge()->GetEdge()->GetVertex() == 
		 ppfa[i]->GetForwardEdge()->GetVertex() )
	      {
		uiSharedVertices++;
	      }
	    j = i;
	  }

	if ( uiSharedVertices > 0 && uiSharedVertices < 3 )
	  {
	    while ( ( ppfa[0]->GetForwardEdge()->GetEdge()->GetVertex() != 
		      ppfa[1]->GetForwardEdge()->GetVertex() ) ||
		    ( ppfa[2]->GetForwardEdge()->GetEdge()->GetVertex() == 
		      ppfa[0]->GetForwardEdge()->GetVertex() ) )
	      {
		pfaSwap = ppfa[0];
		ppfa[0] = ppfa[1];
		ppfa[1] = ppfa[2];
		ppfa[2] = pfaSwap;
	      }	    
	  }

	for ( i = 0; i < 3; i++ )
	  {
	    ppe[i] = ppfa[i]->GetForwardEdge();
	    pppt[i][0] = ppe[i]->GetVertex()->GetPoint();
	    pppt[i][1] = ppe[i]->GetEdge()->GetVertex()->GetPoint();
	  }
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
  switch ( uiEdges )
    {
    case 0:
      {
	fprintf(stderr, "START: VVV_V\n");	    
	
	fprintf(stderr, "  VVV_V: V0: ( %f, %f )\n", 
		pppt[0][0](0), pppt[0][0](1));
	fprintf(stderr, "  VVV_V: V1: ( %f, %f )\n", 
		pppt[1][0](0), pppt[1][0](1));
	fprintf(stderr, "  VVV_V: V2: ( %f, %f )\n", 
		pppt[2][0](0), pppt[2][0](1));
	fprintf(stderr, "  VVV_V: V3: ( %f, %f )\n", 
		pppt[3][0](0), pppt[3][0](1));
      }
    break;
    case 1:
      {
	fprintf(stderr, "START: VVE_V\n");	    
	
	fprintf(stderr, "  VVE_V: V0: ( %f, %f )\n", 
		pppt[0][0](0), pppt[0][0](1));
	fprintf(stderr, "  VVE_V: V1: ( %f, %f )\n", 
		pppt[1][0](0), pppt[1][0](1));
	fprintf(stderr, "  VVE_V: E2: ( %f, %f ) -> ( %f, %f )\n", 
		pppt[2][0](0), pppt[2][0](1),
		pppt[2][1](0), pppt[2][1](1));
	fprintf(stderr, "  VVE_V: V3: ( %f, %f )\n", 
		pppt[3][0](0), pppt[3][0](1));
      }
    break;
    case 2:
      {
	fprintf(stderr, "START: VEE_V\n");	    
	
	fprintf(stderr, "  VEE_V: V0: ( %f, %f )\n", 
		pppt[0][0](0), pppt[0][0](1));
	fprintf(stderr, "  VEE_V: E1: ( %f, %f ) -> ( %f, %f )\n", 
		pppt[1][0](0), pppt[1][0](1),
		pppt[1][1](0), pppt[1][1](1));
	fprintf(stderr, "  VEE_V: E2: ( %f, %f ) -> ( %f, %f )\n", 
		pppt[2][0](0), pppt[2][0](1),
		pppt[2][1](0), pppt[2][1](1));
	fprintf(stderr, "  VEE_V: V3: ( %f, %f )\n", 
		pppt[3][0](0), pppt[3][0](1));
      }
    break;
    case 3:
      {
	fprintf(stderr, "START: EEE_V\n");	    
	
	fprintf(stderr, "  EEE_V: E0: ( %f, %f ) -> ( %f, %f )\n", 
		pppt[0][0](0), pppt[0][0](1),
		pppt[0][1](0), pppt[0][1](1));
	fprintf(stderr, "  EEE_V: E1: ( %f, %f ) -> ( %f, %f )\n", 
		pppt[1][0](0), pppt[1][0](1),
		pppt[1][1](0), pppt[1][1](1));
	fprintf(stderr, "  EEE_V: E2: ( %f, %f ) -> ( %f, %f )\n", 
		pppt[2][0](0), pppt[2][0](1),
		pppt[2][1](0), pppt[2][1](1));
	fprintf(stderr, "  EEE_V: V3: ( %f, %f )\n", 
		pppt[3][0](0), pppt[3][0](1));
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

  //////////////////////////////////////////////////
  // Call the correct InCircle test
  switch ( (APM2D::InscribedCircleCenterType)eICCT )
    {
    case APM2D::icctVVV:
      {
	ASSERT( uiVertices == 3 );

	eCircleTest = APM2D::InCircleVVV_V(pppt[0][0](0), pppt[0][0](1),
					   pppt[1][0](0), pppt[1][0](1),
					   pppt[2][0](0), pppt[2][0](1),
					   pppt[3][0](0), pppt[3][0](1));
      }
    break;
    case APM2D::icctVVE:
      {
	ASSERT( uiVertices == 2 && uiEdges == 1 &&
		ppvx[0] != ppe[2]->GetEdge()->GetVertex() &&
		ppvx[1] != ppe[2]->GetVertex() );

	eCircleTest = APM2D::InCircleVVE_V(pppt[0][0](0), pppt[0][0](1),
					   pppt[1][0](0), pppt[1][0](1),
					   pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					   pppt[3][0](0), pppt[3][0](1));
      }
    break;
    case APM2D::icctVVEAntiparallel:
      {
	ASSERT( uiVertices == 2 && uiEdges == 1 &&
		ppvx[0] != ppe[2]->GetEdge()->GetVertex() &&
		ppvx[1] != ppe[2]->GetVertex() );

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	fprintf(stderr, "  VVE_V Anti:\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	eCircleTest = APM2D::InCircleVVE_VAntiparallel(pppt[0][0](0), pppt[0][0](1),
						       pppt[1][0](0), pppt[1][0](1),
						       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
						       pppt[3][0](0), pppt[3][0](1));
      }
    break;
    case APM2D::icctVvE:
      {
	ASSERT( uiVertices == 2 && uiEdges == 1 &&
		ppvx[0] != ppe[2]->GetEdge()->GetVertex() &&
		ppvx[1] == ppe[2]->GetVertex() );

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	fprintf(stderr, "  VvE_V:\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	eCircleTest = APM2D::InCircleVvE_V(pppt[0][0](0), pppt[0][0](1),
					   pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					   pppt[3][0](0), pppt[3][0](1));
      }
    break;
    case APM2D::icctEvV:
      {
	ASSERT( uiVertices == 2 && uiEdges == 1 &&
		ppvx[0] == ppe[2]->GetEdge()->GetVertex() &&
		ppvx[1] != ppe[2]->GetVertex() );

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	fprintf(stderr, "  EvV_V:\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	eCircleTest = APM2D::InCircleEvV_V(pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					   pppt[1][0](0), pppt[1][0](1),
					   pppt[3][0](0), pppt[3][0](1));
      }
    break;
    case APM2D::icctVEE:
    case APM2D::icctVEvE:
      {
	ASSERT( uiVertices == 1 && uiEdges == 2 &&
		ppvx[0] != ppe[2]->GetEdge()->GetVertex() &&
		ppvx[0] != ppe[1]->GetVertex() );

	eCircleTest = APM2D::InCircleVEE_V(pppt[0][0](0), pppt[0][0](1),
					   pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					   pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					   pppt[3][0](0), pppt[3][0](1));
      }
    break;
    case APM2D::icctVEEAntiparallel:
      {
	ASSERT( uiVertices == 1 && uiEdges == 2 &&
		ppvx[0] != ppe[2]->GetEdge()->GetVertex() &&
		ppvx[0] != ppe[1]->GetVertex() &&
		ppe[1]->GetEdge()->GetVertex() != ppe[2]->GetVertex() );

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	fprintf(stderr, "  VEE_V Anti:\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	eCircleTest = APM2D::InCircleVEE_VAntiparallel(pppt[0][0](0), pppt[0][0](1),
						       pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
						       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
						       pppt[3][0](0), pppt[3][0](1));
      }
    break;
    case APM2D::icctVEcE:
      {
	ASSERT( uiVertices == 1 && uiEdges == 2 &&
		ppvx[0] != ppe[2]->GetEdge()->GetVertex() &&
		ppvx[0] != ppe[1]->GetVertex() &&
		ppe[1]->GetEdge()->GetVertex() == ppe[2]->GetVertex() );

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	fprintf(stderr, "  VEcE_V:\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	eCircleTest = APM2D::InCircleVvE_V(pppt[0][0](0), pppt[0][0](1),
					   pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					   pppt[3][0](0), pppt[3][0](1));
      }
    break;
    case APM2D::icctvEE:
    case APM2D::icctvEvE:
      {
	ASSERT( uiVertices == 1 && uiEdges == 2 &&
		ppvx[0] != ppe[2]->GetEdge()->GetVertex() &&
		ppvx[0] == ppe[1]->GetVertex() );

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	fprintf(stderr, "  vEE_V:\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	eCircleTest = APM2D::InCirclevEE_V(pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					   pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					   pppt[3][0](0), pppt[3][0](1));
      }
    break;
    case APM2D::icctEEv: 
    case APM2D::icctEvEv:
      {
	ASSERT( uiVertices == 1 && uiEdges == 2 &&
		ppvx[0] == ppe[2]->GetEdge()->GetVertex() &&
		ppvx[0] != ppe[1]->GetVertex() );

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	fprintf(stderr, "  EEv_V:\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	eCircleTest = APM2D::InCircleEEv_V(pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					   pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					   pppt[3][0](0), pppt[3][0](1));
      }
    break;
    case APM2D::icctEEE:
    case APM2D::icctEvEE:
    case APM2D::icctEvEvE:
    case APM2D::icctEvEvEv:
      {
	ASSERT( uiVertices == 0 && uiEdges == 3 );

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	fprintf(stderr, "  EEE_V:\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	eCircleTest = APM2D::InCircleEEE_V(pppt[0][0](0), pppt[0][0](1), pppt[0][1](0), pppt[0][1](1),
					   pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					   pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					   pppt[3][0](0), pppt[3][0](1));
      }
    break;
    case APM2D::icctEcEE:
    case APM2D::icctEcEvE:
      {
	ASSERT( uiVertices == 0 && uiEdges == 3 &&
		ppe[0]->GetEdge()->GetVertex() == ppe[1]->GetVertex() );

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	fprintf(stderr, "  EcEE_V:\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	eCircleTest = APM2D::InCirclevEE_V(pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					   pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					   pppt[3][0](0), pppt[3][0](1));
      }
    break;
    case APM2D::icctEvEcE:
      {
	ASSERT( uiVertices == 0 && uiEdges == 3 &&
		ppe[1]->GetEdge()->GetVertex() == ppe[2]->GetVertex() );

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	fprintf(stderr, "  EvEcE_V:\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	eCircleTest = APM2D::InCircleEEv_V(pppt[0][0](0), pppt[0][0](1), pppt[0][1](0), pppt[0][1](1),
					   pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					   pppt[3][0](0), pppt[3][0](1));
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
  switch ( uiEdges )
    {
    case 0:
      {
	fprintf(stderr, "END  : VVV_V\n");	    
      }
    break;
    case 1:
      {
	fprintf(stderr, "END  : VVE_V\n");	    
      }
    break;
    case 2:
      {
	fprintf(stderr, "END  : VEE_V\n");	    
      }
    break;
    case 3:
      {
	fprintf(stderr, "END  : EEE_V\n");	    
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

  return eCircleTest;
}

APM2D::CircleTestType QEMesh::InCircleSSS_EShewchuk(APM2D::InscribedCircleCenterType eICCT,
						    QEFacet *pfa0, 
						    QEFacet *pfa1, 
						    QEFacet *pfa2, 
						    Point &ptCenter,
						    QEFacet *pfa3, QEEdge *pe3)
{
  ASSERT( pfa0 != NULL && pfa1 != NULL && pfa2 != NULL && pe3 != NULL );

  QEFacet *ppfa[3], *pfaSwap;
  QEVertex *ppvx[3];
  QEEdge *ppe[3];
  Point pppt[4][2];
  UINT_32 uiEdges, uiVertices, uiSharedVertices;
  UINT_32 i, j;
  APM2D::EdgeZoneType eEdgeZone;
  APM2D::CircleTestType eCircleTest;

  ppfa[0] = pfa0;
  ppfa[1] = pfa1;
  ppfa[2] = pfa2;

  ppvx[0] = NULL;
  ppvx[1] = NULL;
  ppvx[2] = NULL;

  ppe[0] = NULL;
  ppe[1] = NULL;
  ppe[2] = NULL;

  //////////////////////////////////////////////////
  // The fourth site is treated differently
  pppt[3][0] = pe3->GetVertex()->GetPoint();
  pppt[3][1] = pe3->GetEdge()->GetVertex()->GetPoint();

  //////////////////////////////////////////////////
  // Deal with points at infinity
  if ( eICCT == APM2D::icctInfinity )
    {
      INT_32 piResult[2];

      ASSERT( ppfa[0] == GetNullFacet() ||
	      ppfa[1] == GetNullFacet() ||
	      ppfa[2] == GetNullFacet() );
      while ( ppfa[2] != GetNullFacet() )
	{
	  pfaSwap = ppfa[0];
	  ppfa[0] = ppfa[1];
	  ppfa[1] = ppfa[2];
	  ppfa[2] = pfaSwap;
	}
      ASSERT( ppfa[0] != GetNullFacet() &&
	      ppfa[1] != GetNullFacet() );

      ppvx[0] = ppfa[0]->GetForwardVertex();
      ppvx[1] = ppfa[1]->GetForwardVertex();
      ASSERT( ppvx[0] != NULL || ppvx[1] != NULL );
      if ( ppvx[0] == NULL )
	{
	switch ( APM::GetAccuracyType() )
	  {
	  case APM::atGlobal:
	  case APM::atLocal:
	  case APM::atExact:
	    ppe[0] = ppfa[0]->GetForwardEdge();
	    // This assertion is less correct because finite line segments
	    // should not be able to create points at infinity with non-adjacent vertices
	    // but numerical issues make us do a work around.  Use the more distant endpoint
	    ASSERT( ppe[0] != NULL &&
		    ppvx[1] != NULL );
	    pppt[0][0] = ppe[0]->GetVertex()->GetPoint();
	    //pppt[1][0] = ppe[0]->GetEdge()->GetVertex()->GetPoint();
	    pppt[1][0] = ppvx[1]->GetPoint();
	    break;
	  case APM::atMPFUN:
	  case APM::atMPFUNFilter:
	  case APM::atMPFUNNormalized:
	    ppe[0] = ppfa[0]->GetForwardEdge();
	    ASSERT( ppe[0] != NULL &&
		    ppvx[1] != NULL &&
		    ( ppe[0]->GetVertex() == ppvx[1] ||
		      ppe[0]->GetEdge()->GetVertex() == ppvx[1] ) );
	    pppt[0][0] = ppe[0]->GetVertex()->GetPoint();
	    pppt[1][0] = ppe[0]->GetEdge()->GetVertex()->GetPoint();
	    break;
	  default:
	    ASSERT( FALSE );
	    break;
	  }
	}
      else if ( ppvx[1] == NULL )
	{
	switch ( APM::GetAccuracyType() )
	  {
	  case APM::atGlobal:
	  case APM::atLocal:
	  case APM::atExact:
	    ppe[1] = ppfa[1]->GetForwardEdge();
	    // This assertion is less correct because finite line segments
	    // should not be able to create points at infinity with non-adjacent vertices
	    // but numerical issues make us do a work around.  Use the more distant endpoint
	    ASSERT( ppe[1] != NULL &&
		    ppvx[0] != NULL );
	    pppt[0][0] = ppvx[0]->GetPoint();
	    //pppt[0][0] = ppe[1]->GetVertex()->GetPoint();
	    pppt[1][0] = ppe[1]->GetEdge()->GetVertex()->GetPoint();
	    break;
	  case APM::atMPFUN:
	  case APM::atMPFUNFilter:
	  case APM::atMPFUNNormalized:
	    ppe[1] = ppfa[1]->GetForwardEdge();
	    ASSERT( ppe[1] != NULL &&
		    ppvx[0] != NULL &&
		    ( ppe[1]->GetVertex() == ppvx[0] ||
		      ppe[1]->GetEdge()->GetVertex() == ppvx[0] ) );
	    pppt[0][0] = ppe[1]->GetVertex()->GetPoint();
	    pppt[1][0] = ppe[1]->GetEdge()->GetVertex()->GetPoint();
	    break;
	  default:
	    ASSERT( FALSE );
	    break;
	  }
	}
      else
	{
	  pppt[0][0] = ppvx[0]->GetPoint();
	  pppt[1][0] = ppvx[1]->GetPoint();
	}

      piResult[0] = APM2D::OrientationVVV(pppt[0][0](0), pppt[0][0](1), 
					  pppt[1][0](0), pppt[1][0](1), 
					  pppt[3][0](0), pppt[3][0](1));
      piResult[1] = APM2D::OrientationVVV(pppt[0][0](0), pppt[0][0](1), 
					  pppt[1][0](0), pppt[1][0](1), 
					  pppt[3][1](0), pppt[3][1](1));
      if ( piResult[0] > 0 || piResult[1] > 0 )
	{
	  return APM2D::cttInCircle;
	}
      else if ( piResult[0] == 0 || piResult[1] == 0 )
	{
	  return APM2D::cttOnCircle;
	}
      else
	{
	  return APM2D::cttOutCircle;
	}
    }

  uiEdges = 0;
  uiVertices = 0;
  uiSharedVertices = 0;
  for ( i = 0; i < 3; i++ )
    {
      if ( ppfa[i]->GetForwardVertex() != NULL )
	{
	  ASSERT( ppfa[i]->GetForwardEdge() == NULL );
	  uiVertices++;
	}
      else
	{
	  ASSERT( ppfa[i]->GetForwardEdge() != NULL );
	  uiEdges++;
	}
    }

  switch ( uiEdges )
    {
    case 0:
      {
	for ( i = 0; i < 3; i++ )
	  {
	    ppvx[i] = ppfa[i]->GetForwardVertex();
	    pppt[i][0] = ppvx[i]->GetPoint();
	  }
      }
    break;
    case 1:
      {
	while ( ppfa[2]->GetForwardEdge() == NULL )
	  {
	    pfaSwap = ppfa[0];
	    ppfa[0] = ppfa[1];
	    ppfa[1] = ppfa[2];
	    ppfa[2] = pfaSwap;
	  }

	for ( i = 0; i < 2; i++ )
	  {
	    ppvx[i] = ppfa[i]->GetForwardVertex();
	    pppt[i][0] = ppvx[i]->GetPoint();
	  }
	for ( i = 2; i < 3; i++ )
	  {
	    ppe[i] = ppfa[i]->GetForwardEdge();
	    pppt[i][0] = ppe[i]->GetVertex()->GetPoint();
	    pppt[i][1] = ppe[i]->GetEdge()->GetVertex()->GetPoint();
	  }
      }
    break;
    case 2:
      {
	while ( ppfa[0]->GetForwardVertex() == NULL )
	  {
	    pfaSwap = ppfa[0];
	    ppfa[0] = ppfa[1];
	    ppfa[1] = ppfa[2];
	    ppfa[2] = pfaSwap;
	  }

	for ( i = 0; i < 1; i++ )
	  {
	    ppvx[i] = ppfa[i]->GetForwardVertex();
	    pppt[i][0] = ppvx[i]->GetPoint();
	  }
	for ( i = 1; i < 3; i++ )
	  {
	    ppe[i] = ppfa[i]->GetForwardEdge();
	    pppt[i][0] = ppe[i]->GetVertex()->GetPoint();
	    pppt[i][1] = ppe[i]->GetEdge()->GetVertex()->GetPoint();
	  }
      }
    break;
    case 3:
      {
	j = 2;
	uiSharedVertices = 0;
	for ( i = 0; i < 3; i++ )
	  {
	    if ( ppfa[j]->GetForwardEdge()->GetEdge()->GetVertex() == 
		 ppfa[i]->GetForwardEdge()->GetVertex() )
	      {
		uiSharedVertices++;
	      }
	    j = i;
	  }

	if ( uiSharedVertices > 0 && uiSharedVertices < 3 )
	  {
	    while ( ( ppfa[0]->GetForwardEdge()->GetEdge()->GetVertex() != 
		      ppfa[1]->GetForwardEdge()->GetVertex() ) ||
		    ( ppfa[2]->GetForwardEdge()->GetEdge()->GetVertex() == 
		      ppfa[0]->GetForwardEdge()->GetVertex() ) )
	      {
		pfaSwap = ppfa[0];
		ppfa[0] = ppfa[1];
		ppfa[1] = ppfa[2];
		ppfa[2] = pfaSwap;
	      }	    
	  }

	for ( i = 0; i < 3; i++ )
	  {
	    ppe[i] = ppfa[i]->GetForwardEdge();
	    pppt[i][0] = ppe[i]->GetVertex()->GetPoint();
	    pppt[i][1] = ppe[i]->GetEdge()->GetVertex()->GetPoint();
	  }
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
  switch ( uiEdges )
    {
    case 0:
      {
	fprintf(stderr, "START: VVV_E\n");	    

	fprintf(stderr, "  VVV_E: V0: ( %f, %f )\n", 
		pppt[0][0](0), pppt[0][0](1));
	fprintf(stderr, "  VVV_E: V1: ( %f, %f )\n", 
		pppt[1][0](0), pppt[1][0](1));
	fprintf(stderr, "  VVV_E: V2: ( %f, %f )\n", 
		pppt[2][0](0), pppt[2][0](1));
	fprintf(stderr, "  VVV_E: E3: ( %f, %f ) -> ( %f, %f )\n", 
		pppt[3][0](0), pppt[3][0](1),
		pppt[3][1](0), pppt[3][1](1));
      }
    break;
    case 1:
      {
	fprintf(stderr, "START: VVE_E\n");	    
	
	fprintf(stderr, "  VVE_E: V0: ( %f, %f )\n", 
		pppt[0][0](0), pppt[0][0](1));
	fprintf(stderr, "  VVE_E: V1: ( %f, %f )\n", 
		pppt[1][0](0), pppt[1][0](1));
	fprintf(stderr, "  VVE_E: E2: ( %f, %f ) -> ( %f, %f )\n", 
		pppt[2][0](0), pppt[2][0](1),
		pppt[2][1](0), pppt[2][1](1));
	fprintf(stderr, "  VVE_E: E3: ( %f, %f ) -> ( %f, %f )\n", 
		pppt[3][0](0), pppt[3][0](1),
		pppt[3][1](0), pppt[3][1](1));
      }
    break;
    case 2:
      {
	fprintf(stderr, "START: VEE_E\n");	    
	
	fprintf(stderr, "  VEE_E: V0: ( %f, %f )\n", 
		pppt[0][0](0), pppt[0][0](1));
	fprintf(stderr, "  VEE_E: E1: ( %f, %f ) -> ( %f, %f )\n", 
		pppt[1][0](0), pppt[1][0](1),
		pppt[1][1](0), pppt[1][1](1));
	fprintf(stderr, "  VEE_E: E2: ( %f, %f ) -> ( %f, %f )\n", 
		pppt[2][0](0), pppt[2][0](1),
		pppt[2][1](0), pppt[2][1](1));
	fprintf(stderr, "  VEE_E: E3: ( %f, %f ) -> ( %f, %f )\n", 
		pppt[3][0](0), pppt[3][0](1),
		pppt[3][1](0), pppt[3][1](1));
      }
    break;
    case 3:
      {
	fprintf(stderr, "START: EEE_E\n");	    
	
	fprintf(stderr, "  EEE_E: E0: ( %f, %f ) -> ( %f, %f )\n", 
		pppt[0][0](0), pppt[0][0](1),
		pppt[0][1](0), pppt[0][1](1));
	fprintf(stderr, "  EEE_E: E1: ( %f, %f ) -> ( %f, %f )\n", 
		pppt[1][0](0), pppt[1][0](1),
		pppt[1][1](0), pppt[1][1](1));
	fprintf(stderr, "  EEE_E: E2: ( %f, %f ) -> ( %f, %f )\n", 
		pppt[2][0](0), pppt[2][0](1),
		pppt[2][1](0), pppt[2][1](1));
	fprintf(stderr, "  EEE_E: E3: ( %f, %f ) -> ( %f, %f )\n", 
		pppt[3][0](0), pppt[3][0](1),
		pppt[3][1](0), pppt[3][1](1));
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

  //////////////////////////////////////////////////
  // Call the correct InCircle test
  switch ( (APM2D::InscribedCircleCenterType)eICCT )
    {
    case APM2D::icctVVV:
      {
	ASSERT( uiVertices == 3 );

	eEdgeZone = APM2D::EdgeZoneVVV(pppt[0][0](0), pppt[0][0](1),
				       pppt[1][0](0), pppt[1][0](1),
				       pppt[2][0](0), pppt[2][0](1),
				       pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	switch ( eEdgeZone )
	  {
	  case APM2D::eztStart:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  VVV_E: E30 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	    /*
	    eCircleTest = APM2D::InCircleVVV_V(pppt[0][0](0), pppt[0][0](1),
					       pppt[1][0](0), pppt[1][0](1),
					       pppt[2][0](0), pppt[2][0](1),
					       pppt[3][0](0), pppt[3][0](1));
					       */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetVertex());
	    break;
	  case APM2D::eztEnd:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  VVV_E: E31 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    
	    /*
	    eCircleTest = APM2D::InCircleVVV_V(pppt[0][0](0), pppt[0][0](1),
					       pppt[1][0](0), pppt[1][0](1),
					       pppt[2][0](0), pppt[2][0](1),
					       pppt[3][1](0), pppt[3][1](1));
					       */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetEdge()->GetVertex());
	    break;
	  case APM2D::eztMiddle:
	    eCircleTest = APM2D::InCircleVVV_E(pppt[0][0](0), pppt[0][0](1),
					       pppt[1][0](0), pppt[1][0](1),
					       pppt[2][0](0), pppt[2][0](1),
					       pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	    break;
	  default:
	    ASSERT( FALSE );
	    break;
	  }
      }
    break;
    case APM2D::icctVVE:
      {
	ASSERT( uiVertices == 2 && uiEdges == 1 &&
		ppvx[0] != ppe[2]->GetEdge()->GetVertex() &&
		ppvx[1] != ppe[2]->GetVertex() );

	eEdgeZone = APM2D::EdgeZoneVVE(pppt[0][0](0), pppt[0][0](1),
				       pppt[1][0](0), pppt[1][0](1),
				       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
				       pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	switch ( eEdgeZone )
	  {
	  case APM2D::eztStart:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  VVE_E: E30 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    
	    /*
	    eCircleTest = APM2D::InCircleVVE_V(pppt[0][0](0), pppt[0][0](1),
					       pppt[1][0](0), pppt[1][0](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][0](0), pppt[3][0](1));
					       */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetVertex());
	    break;
	  case APM2D::eztEnd:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  VVE_E: E31 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    /*
	    eCircleTest = APM2D::InCircleVVE_V(pppt[0][0](0), pppt[0][0](1),
					       pppt[1][0](0), pppt[1][0](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][1](0), pppt[3][1](1));
					       */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetEdge()->GetVertex());
	    break;
	  case APM2D::eztMiddle:
	    eCircleTest = APM2D::InCircleVVE_E(pppt[0][0](0), pppt[0][0](1),
					       pppt[1][0](0), pppt[1][0](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	    break;
	  default:
	    ASSERT( FALSE );
	    break;
	  }
      }
    break;
    case APM2D::icctVVEAntiparallel:
      {
	ASSERT( uiVertices == 2 && uiEdges == 1 &&
		ppvx[0] != ppe[2]->GetEdge()->GetVertex() &&
		ppvx[1] != ppe[2]->GetVertex() );

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	fprintf(stderr, "  VVE_E Anti\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	
	eEdgeZone = APM2D::EdgeZoneVVEAntiparallel(pppt[0][0](0), pppt[0][0](1),
						   pppt[1][0](0), pppt[1][0](1),
						   pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
						   pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	switch ( eEdgeZone )
	  {
	  case APM2D::eztStart:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  VVE_E Anti: E30 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    
	    /*
	    eCircleTest = APM2D::InCircleVVE_VAntiparallel(pppt[0][0](0), pppt[0][0](1),
							   pppt[1][0](0), pppt[1][0](1),
							   pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
							   pppt[3][0](0), pppt[3][0](1));
							   */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetVertex());
	    break;
	  case APM2D::eztEnd:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  VVE_E Anti: E31 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    
	    /*
	    eCircleTest = APM2D::InCircleVVE_VAntiparallel(pppt[0][0](0), pppt[0][0](1),
							   pppt[1][0](0), pppt[1][0](1),
							   pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
							   pppt[3][1](0), pppt[3][1](1));
							   */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetEdge()->GetVertex());
	    break;
	  case APM2D::eztMiddle:
	    eCircleTest = APM2D::InCircleVVE_EAntiparallel(pppt[0][0](0), pppt[0][0](1),
							   pppt[1][0](0), pppt[1][0](1),
							   pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
							   pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	    break;
	  default:
	    ASSERT( FALSE );
	    break;
	  }
      }
    break;
    case APM2D::icctVvE:
      {
	ASSERT( uiVertices == 2 && uiEdges == 1 &&
		ppvx[0] != ppe[2]->GetEdge()->GetVertex() &&
		ppvx[1] == ppe[2]->GetVertex() );

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	fprintf(stderr, "  VvE_E:\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	eEdgeZone = APM2D::EdgeZoneVvE(pppt[0][0](0), pppt[0][0](1),
				       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
				       pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	switch ( eEdgeZone )
	  {
	  case APM2D::eztStart:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  VvE_E: E30 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    
	    /*
	    eCircleTest = APM2D::InCircleVvE_V(pppt[0][0](0), pppt[0][0](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][0](0), pppt[3][0](1));
					       */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetVertex());
	    break;
	  case APM2D::eztEnd:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  VvE_E: E31 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    
	    /*
	    eCircleTest = APM2D::InCircleVvE_V(pppt[0][0](0), pppt[0][0](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][1](0), pppt[3][1](1));
					       */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetEdge()->GetVertex());
	    break;
	  case APM2D::eztMiddle:
	    eCircleTest = APM2D::InCircleVvE_E(pppt[0][0](0), pppt[0][0](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	    break;
	  default:
	    ASSERT( FALSE );
	    break;
	  }
      }
    break;
    case APM2D::icctEvV:
      {
	ASSERT( uiVertices == 2 && uiEdges == 1 &&
		ppvx[0] == ppe[2]->GetEdge()->GetVertex() &&
		ppvx[1] != ppe[2]->GetVertex() );

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	fprintf(stderr, "  EvV_E:\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	eEdgeZone = APM2D::EdgeZoneEvV(pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
				       pppt[1][0](0), pppt[1][0](1),
				       pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	switch ( eEdgeZone )
	  {
	  case APM2D::eztStart:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  EvV_E: E30 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    
	    /*
	    eCircleTest = APM2D::InCircleEvV_V(pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[1][0](0), pppt[1][0](1),
					       pppt[3][0](0), pppt[3][0](1));
					       */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetVertex());
	    break;
	  case APM2D::eztEnd:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  EvV_E: E31 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    
	    /*
	    eCircleTest = APM2D::InCircleEvV_V(pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[1][0](0), pppt[1][0](1),
					       pppt[3][1](0), pppt[3][1](1));
					       */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetEdge()->GetVertex());
	    break;
	  case APM2D::eztMiddle:
	    eCircleTest = APM2D::InCircleEvV_E(pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[1][0](0), pppt[1][0](1),
					       pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	    break;
	  default:
	    ASSERT( FALSE );
	    break;
	  }
      }
    break;
    case APM2D::icctVEE:
    case APM2D::icctVEvE:
      {
	ASSERT( uiVertices == 1 && uiEdges == 2 &&
		ppvx[0] != ppe[2]->GetEdge()->GetVertex() &&
		ppvx[0] != ppe[1]->GetVertex() );

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	fprintf(stderr, "  VEE_E:\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	eEdgeZone = APM2D::EdgeZoneVEE(pppt[0][0](0), pppt[0][0](1),
				       pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
				       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
				       pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	switch ( eEdgeZone )
	  {
	  case APM2D::eztStart:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  VEE_E: E30 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    
	    /*
	    eCircleTest = APM2D::InCircleVEE_V(pppt[0][0](0), pppt[0][0](1),
					       pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][0](0), pppt[3][0](1));
					       */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetVertex());
	    break;
	  case APM2D::eztEnd:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  VEE_E: E31 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    
	    /*
	    eCircleTest = APM2D::InCircleVEE_V(pppt[0][0](0), pppt[0][0](1),
					       pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][1](0), pppt[3][1](1));
					       */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetEdge()->GetVertex());
	    break;
	  case APM2D::eztMiddle:
	    eCircleTest = APM2D::InCircleVEE_E(pppt[0][0](0), pppt[0][0](1),
					       pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	    break;
	  default:
	    ASSERT( FALSE );
	    break;
	  }
      }
    break;
    case APM2D::icctVEEAntiparallel:
      {
	ASSERT( uiVertices == 1 && uiEdges == 2 &&
		ppvx[0] != ppe[2]->GetEdge()->GetVertex() &&
		ppvx[0] != ppe[1]->GetVertex() &&
		ppe[1]->GetEdge()->GetVertex() != ppe[2]->GetVertex() );

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	fprintf(stderr, "  VEE_E Anti:\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	
	eEdgeZone = APM2D::EdgeZoneVEEAntiparallel(pppt[0][0](0), pppt[0][0](1),
						   pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
						   pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
						   pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	switch ( eEdgeZone )
	  {
	  case APM2D::eztStart:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  VEE_E Anti: E30 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    
	    /*
	    eCircleTest = APM2D::InCircleVEE_VAntiparallel(pppt[0][0](0), pppt[0][0](1),
							   pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
							   pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
							   pppt[3][0](0), pppt[3][0](1));
							   */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetVertex());
	    break;
	  case APM2D::eztEnd:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  VEE_E Anti: E31 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	    /*
	    eCircleTest = APM2D::InCircleVEE_VAntiparallel(pppt[0][0](0), pppt[0][0](1),
							   pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
							   pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
							   pppt[3][1](0), pppt[3][1](1));
							   */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetEdge()->GetVertex());
	    break;
	  case APM2D::eztMiddle:
	    eCircleTest = APM2D::InCircleVEE_EAntiparallel(pppt[0][0](0), pppt[0][0](1),
							   pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
							   pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
							   pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	    break;
	  default:
	    ASSERT( FALSE );
	    break;
	  }
      }
    break;
    case APM2D::icctVEcE:
      {
	ASSERT( uiVertices == 1 && uiEdges == 2 &&
		ppvx[0] != ppe[2]->GetEdge()->GetVertex() &&
		ppvx[0] != ppe[1]->GetVertex() &&
		ppe[1]->GetEdge()->GetVertex() == ppe[2]->GetVertex() );

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	fprintf(stderr, "  VEcE_E:\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	eEdgeZone = APM2D::EdgeZoneVvE(pppt[0][0](0), pppt[0][0](1),
				       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
				       pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	switch ( eEdgeZone )
	  {
	  case APM2D::eztStart:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  VEcE_E: E30 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    
	    /*
	    eCircleTest = APM2D::InCircleVvE_V(pppt[0][0](0), pppt[0][0](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][0](0), pppt[3][0](1));
					       */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetVertex());
	    break;
	  case APM2D::eztEnd:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  VEcE_E: E31 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	    /*
	    eCircleTest = APM2D::InCircleVvE_V(pppt[0][0](0), pppt[0][0](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][1](0), pppt[3][1](1));
					       */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetEdge()->GetVertex());
	    break;
	  case APM2D::eztMiddle:
	    eCircleTest = APM2D::InCircleVvE_E(pppt[0][0](0), pppt[0][0](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	    break;
	  default:
	    ASSERT( FALSE );
	    break;
	  }
      }
    break;
    case APM2D::icctvEE:
    case APM2D::icctvEvE:
      {
	ASSERT( uiVertices == 1 && uiEdges == 2 &&
		ppvx[0] != ppe[2]->GetEdge()->GetVertex() &&
		ppvx[0] == ppe[1]->GetVertex() );

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	fprintf(stderr, "  vEE_E:\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	eEdgeZone = APM2D::EdgeZonevEE(pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
				       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
				       pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	switch ( eEdgeZone )
	  {
	  case APM2D::eztStart:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  vEE_E: E30 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    
	    /*
	    eCircleTest = APM2D::InCirclevEE_V(pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][0](0), pppt[3][0](1));
					       */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetVertex());
	    break;
	  case APM2D::eztEnd:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  vEE_E: E31 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    
	    /*
	    eCircleTest = APM2D::InCirclevEE_V(pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][1](0), pppt[3][1](1));
					       */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetEdge()->GetVertex());
	    break;
	  case APM2D::eztMiddle:
	    eCircleTest = APM2D::InCirclevEE_E(pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	    break;
	  default:
	    ASSERT( FALSE );
	    break;
	  }
      }
    break;
    case APM2D::icctEEv: 
    case APM2D::icctEvEv:
      {
	ASSERT( uiVertices == 1 && uiEdges == 2 &&
		ppvx[0] == ppe[2]->GetEdge()->GetVertex() &&
		ppvx[0] != ppe[1]->GetVertex() );

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	fprintf(stderr, "  EEv_E:\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	eEdgeZone = APM2D::EdgeZoneEEv(pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
				       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
				       pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	switch ( eEdgeZone )
	  {
	  case APM2D::eztStart:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  EEv_E: E30 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    
	    /*
	    eCircleTest = APM2D::InCircleEEv_V(pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][0](0), pppt[3][0](1));
					       */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetVertex());
	    break;
	  case APM2D::eztEnd:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  EEv_E: E31 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    
	    /*
	    eCircleTest = APM2D::InCircleEEv_V(pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][1](0), pppt[3][1](1));
					       */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetEdge()->GetVertex());
	    break;
	  case APM2D::eztMiddle:
	    eCircleTest = APM2D::InCircleEEv_E(pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	    break;
	  default:
	    ASSERT( FALSE );
	    break;
	  }
      }
    break;
    case APM2D::icctEEE:
    case APM2D::icctEvEE:
    case APM2D::icctEvEvE:
    case APM2D::icctEvEvEv:
      {
	ASSERT( uiVertices == 0 && uiEdges == 3 );

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	fprintf(stderr, "  EEE_E:\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	eEdgeZone = APM2D::EdgeZoneEEE(pppt[0][0](0), pppt[0][0](1), pppt[0][1](0), pppt[0][1](1),
				       pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
				       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
				       pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	switch ( eEdgeZone )
	  {
	  case APM2D::eztStart:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  EEE_E: E30 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    
	    /*
	    eCircleTest = APM2D::InCircleEEE_V(pppt[0][0](0), pppt[0][0](1), pppt[0][1](0), pppt[0][1](1),
					       pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][0](0), pppt[3][0](1));
					       */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetVertex());
	    break;
	  case APM2D::eztEnd:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  EEE_E: E31 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    
	    /*
	    eCircleTest = APM2D::InCircleEEE_V(pppt[0][0](0), pppt[0][0](1), pppt[0][1](0), pppt[0][1](1),
					       pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][1](0), pppt[3][1](1));
					       */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetEdge()->GetVertex());
	    break;
	  case APM2D::eztMiddle:
	    eCircleTest = APM2D::InCircleEEE_E(pppt[0][0](0), pppt[0][0](1), pppt[0][1](0), pppt[0][1](1),
					       pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	    break;
	  default:
	    ASSERT( FALSE );
	    break;
	  }
      }
    break;
    case APM2D::icctEcEE:
    case APM2D::icctEcEvE:
      {
	ASSERT( uiVertices == 0 && uiEdges == 3 &&
		ppe[0]->GetEdge()->GetVertex() == ppe[1]->GetVertex() );

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	fprintf(stderr, "  EcEE_E:\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	eEdgeZone = APM2D::EdgeZonevEE(pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
				       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
				       pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	switch ( eEdgeZone )
	  {
	  case APM2D::eztStart:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  EcEE_E: E30 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	    /*
	    eCircleTest = APM2D::InCirclevEE_V(pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][0](0), pppt[3][0](1));
					       */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetVertex());
	    break;
	  case APM2D::eztEnd:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  EcEE_E: E31 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    
	    /*
	    eCircleTest = APM2D::InCirclevEE_V(pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][1](0), pppt[3][1](1));
					       */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetEdge()->GetVertex());
	    break;
	  case APM2D::eztMiddle:
	    eCircleTest = APM2D::InCirclevEE_E(pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					       pppt[2][0](0), pppt[2][0](1), pppt[2][1](0), pppt[2][1](1),
					       pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	    break;
	  default:
	    ASSERT( FALSE );
	    break;
	  }
      }
    break;
    case APM2D::icctEvEcE:
      {
	ASSERT( uiVertices == 0 && uiEdges == 3 &&
		ppe[1]->GetEdge()->GetVertex() == ppe[2]->GetVertex() );

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	fprintf(stderr, "  EvEcE_E:\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	eEdgeZone = APM2D::EdgeZoneEEv(pppt[0][0](0), pppt[0][0](1), pppt[0][1](0), pppt[0][1](1),
				       pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
				       pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	switch ( eEdgeZone )
	  {
	  case APM2D::eztStart:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  EvEcE_E: E30 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	    /*
	    eCircleTest = APM2D::InCircleEEv_V(pppt[0][0](0), pppt[0][0](1), pppt[0][1](0), pppt[0][1](1),
					       pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					       pppt[3][0](0), pppt[3][0](1));
					       */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetVertex());
	    break;
	  case APM2D::eztEnd:
#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
	    fprintf(stderr, "  EvEcE_E: E31 is closest\n");
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

	    /*
	    eCircleTest = APM2D::InCircleEEv_V(pppt[0][0](0), pppt[0][0](1), pppt[0][1](0), pppt[0][1](1),
					       pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					       pppt[3][1](0), pppt[3][1](1));
					       */
	    eCircleTest = InCircleSSS_VShewchuk(eICCT,
						pfa0, pfa1, pfa2, 
						ptCenter,
						pfa3, pe3->GetEdge()->GetVertex());
	    break;
	  case APM2D::eztMiddle:
	    eCircleTest = APM2D::InCircleEEv_E(pppt[0][0](0), pppt[0][0](1), pppt[0][1](0), pppt[0][1](1),
					       pppt[1][0](0), pppt[1][0](1), pppt[1][1](0), pppt[1][1](1),
					       pppt[3][0](0), pppt[3][0](1), pppt[3][1](0), pppt[3][1](1));
	    break;
	  default:
	    ASSERT( FALSE );
	    break;
	  }
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }

#ifdef ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING
  switch ( uiEdges )
    {
    case 0:
      {
	fprintf(stderr, "END  : VVV_E\n");	    
      }
    break;
    case 1:
      {
	fprintf(stderr, "END  : VVE_E\n");	    
      }
    break;
    case 2:
      {
	fprintf(stderr, "END  : VEE_E\n");	    
      }
    break;
    case 3:
      {
	fprintf(stderr, "END  : EEE_E\n");	    
      }
    break;
    default:
      ASSERT( FALSE );
      break;
    }
#endif // ENABLE_VORONOI_POLYGON_IN_CIRCLE_DEBUGGING

  return eCircleTest;
}

VOID QEMesh::MergeVoronoiPolygonsShewchuk(QEMesh *pmeVoronoi)
{
  ASSERT( pmeVoronoi != NULL );

  QEMeshIter piter[2];
  QEFacetIter iterQEFacet;
  QEEdge *pe, *ppe[2], *peCurr;
  QEFacet *pfaMin, *pfaPrev;
  Point pt;
  UINT_32 i, j;

#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
  fprintf(stderr, "START: MergeVoronoiPolygonsShewchuk\n");
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING_OUTPUT_FILES
  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING_OUTPUT_FILES

  //VerifyVoronoiPolygonsShewchuk();
  //pmeVoronoi->VerifyVoronoiPolygonsShewchuk();

  piter[0].Init(this);
  piter[1].Init(pmeVoronoi);
  
  //////////////////////////////////////////////////
  // Find a vertex zone which extends to infinity
  //
  ppe[1] = NULL;
  for ( pe = piter[1].PeekFirstEdge();
	pe != NULL;
	pe = piter[1].PeekNextEdge() )
    {
      if ( pe->GetFacet() == GetNullFacet() )
	{
	  pe = pe->GetEdge();
	  ASSERT( pe->GetFacet() != GetNullFacet() );

	  if ( pe->GetFacet()->GetForwardVertex() != NULL )
	    {
	      ppe[1] = pe;
	      if ( pe->GetVertex() != pe->GetFacet()->GetForwardVertex() )
		{
		  for ( peCurr = pe->GetEdgeNext(0); 
			peCurr != pe;
			peCurr = peCurr->GetEdgeNext(0) )
		    {
		      if ( peCurr->GetVertex() == pe->GetFacet()->GetForwardVertex() )
			{
			  break;
			} 
		    }
		  ASSERT( peCurr != pe );
		  ppe[1] = peCurr;
		}

	      ASSERT( ppe[1]->GetVertex() == ppe[1]->GetFacet()->GetForwardVertex() &&
		      ppe[1]->GetFacet() == pe->GetFacet() );
	      break;
	    }
	}
    }
  ASSERT( ppe[1] != NULL );
  ppe[0] = NULL;

  //////////////////////////////////////////////////
  // Find the facet in the base mesh which contains this vertex
  //

#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
  fprintf(stderr, "  START: Searching for initial sites\n");
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING

  pt = ppe[1]->GetVertex()->GetPoint();
  pt[2] = 0.0;
  pt[3] = 1.0;

  pfaPrev = NULL;
  pfaMin = PointLocateVoronoiPolygonShewchuk(ppe[1]->GetFacet(), ppe[1]->GetFacet()->GetForwardVertex());
  ASSERT( pfaMin != NULL );

  for ( i = 0; pfaMin != pfaPrev; i = j )
    {
      j = ((i + 1) & 0x1);
      if ( pfaMin->GetForwardVertex() != NULL )
	{
	  iterQEFacet.Init(pfaMin);
	  for ( pe = iterQEFacet.PeekFirstEdge();
		pe != NULL;
		pe = iterQEFacet.PeekNextEdge() )
	    {
	      if ( pe->GetVertex() == pfaMin->GetForwardVertex() )
		{
		  break;
		}
	    }
	  ASSERT( pe != NULL );
	  ppe[i] = pe;
	}
      else
	{
	  ASSERT( pfaMin->GetForwardEdge() != NULL );
	  
	  pe = pfaMin->GetForwardEdge();

	  /*
	  if ( ppe[j]->GetFacet()->GetForwardVertex() != NULL )
	    {
	      pt = ppe[j]->GetFacet()->GetForwardVertex()->GetPoint();
	    }
	  else
	    {
	      // either endpoint should work
	      pt = ppe[j]->GetFacet()->GetForwardEdge()->GetVertex()->GetPoint();
	    }
	  ASSERT( Counterclockwise(pe->GetVertex()->GetPoint(), 
				   pe->GetEdge()->GetVertex()->GetPoint(),
				   pt) );
	  */

	  ASSERT( SideOfEdgeShewchuk(pe, ppe[j]->GetFacet()) >= 0 );

	  ppe[i] = pe;
	}
      ASSERT( ppe[i] != NULL );

#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
      if ( i == 0 )
	{
	  fprintf(stderr, "    Searching Right: ");    
	}
      else
	{
	  fprintf(stderr, "    Searching Left : ");    
	}

      if ( ppe[1]->GetFacet()->GetForwardVertex() != NULL )
	{
	  fprintf(stderr, "LV %p: ( %f, %f )  ::  ", 
		  ppe[1]->GetVertex(), 
		  ppe[1]->GetVertex()->GetPoint()(0), 
		  ppe[1]->GetVertex()->GetPoint()(1));
	}
      else
	{
	  fprintf(stderr, "LE %p => %p: ( %f, %f ) => ( %f, %f )  ::  ", 
		  ppe[1]->GetVertex(), 
		  ppe[1]->GetEdge()->GetVertex(), 
		  ppe[1]->GetVertex()->GetPoint()(0), 
		  ppe[1]->GetVertex()->GetPoint()(1),
		  ppe[1]->GetEdge()->GetVertex()->GetPoint()(0), 
		  ppe[1]->GetEdge()->GetVertex()->GetPoint()(1));
	}

      if ( ppe[0]->GetFacet()->GetForwardVertex() != NULL )
	{
	  fprintf(stderr, "RV %p: ( %f, %f )\n", 
		  ppe[0]->GetVertex(), 
		  ppe[0]->GetVertex()->GetPoint()(0), 
		  ppe[0]->GetVertex()->GetPoint()(1));
	}
      else
	{
	  fprintf(stderr, "RE %p => %p: ( %f, %f ) => ( %f, %f )\n", 
		  ppe[0]->GetVertex(), 
		  ppe[0]->GetEdge()->GetVertex(), 
		  ppe[0]->GetVertex()->GetPoint()(0), 
		  ppe[0]->GetVertex()->GetPoint()(1),
		  ppe[0]->GetEdge()->GetVertex()->GetPoint()(0), 
		  ppe[0]->GetEdge()->GetVertex()->GetPoint()(1));
	}
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING
      
      pfaPrev = ppe[j]->GetFacet();
      pfaMin = PointLocateVoronoiPolygonShewchuk(ppe[i]->GetFacet(), ppe[j]);
    }
  
  if ( ppe[0]->GetFacet()->GetForwardVertex() != NULL )
    {
      ppe[0] = ppe[0]->GetEdgePrev(0);
    }

#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
  fprintf(stderr, "  END  : Searching for initial sites\n");
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING

  //////////////////////////////////////////////////
  // Check for Degenerate Touching
  {
    Point ppt[2];
      
    if ( ppe[0]->GetFacet()->GetForwardVertex() != NULL &&
	 ppe[1]->GetFacet()->GetForwardVertex() != NULL )
      {
	ppt[0] = ppe[0]->GetFacet()->GetForwardVertex()->GetPoint();
	ppt[1] = ppe[1]->GetFacet()->GetForwardVertex()->GetPoint();
	if ( ppt[0](0) == ppt[1](0) && ppt[0](1) == ppt[1](1) )
	  {
	    MergeVoronoiPolygonsDegenerateShewchuk(ppe[0]->GetFacet()->GetForwardVertex(),
							  pmeVoronoi,
							  ppe[1]->GetFacet()->GetForwardVertex());
	    return;
	  }
      }
    else if ( ppe[0]->GetFacet()->GetForwardVertex() != NULL &&
	      ppe[1]->GetFacet()->GetForwardEdge() != NULL )
      {
	ppt[0] = ppe[0]->GetFacet()->GetForwardVertex()->GetPoint();
	ppt[1] = ppe[1]->GetFacet()->GetForwardEdge()->GetVertex()->GetPoint();
	if ( ppt[0](0) == ppt[1](0) && ppt[0](1) == ppt[1](1) )
	  {
	    MergeVoronoiPolygonsDegenerateShewchuk(ppe[0]->GetFacet()->GetForwardVertex(),
							  pmeVoronoi,
							  ppe[1]->GetFacet()->GetForwardEdge()->GetVertex());
	    return;
	  }
	ppt[1] = ppe[1]->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint();
	if ( ppt[0](0) == ppt[1](0) && ppt[0](1) == ppt[1](1) )
	  {
	    MergeVoronoiPolygonsDegenerateShewchuk(ppe[0]->GetFacet()->GetForwardVertex(),
							  pmeVoronoi,
							  ppe[1]->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex());
	    return;
	  }
      }
    else if ( ppe[0]->GetFacet()->GetForwardEdge() != NULL &&
	      ppe[1]->GetFacet()->GetForwardVertex() != NULL )
      {
	ppt[1] = ppe[1]->GetFacet()->GetForwardVertex()->GetPoint();
	ppt[0] = ppe[0]->GetFacet()->GetForwardEdge()->GetVertex()->GetPoint();
	if ( ppt[0](0) == ppt[1](0) && ppt[0](1) == ppt[1](1) )
	  {
	    MergeVoronoiPolygonsDegenerateShewchuk(ppe[0]->GetFacet()->GetForwardEdge()->GetVertex(),
							  pmeVoronoi,
							  ppe[1]->GetFacet()->GetForwardVertex());
	    return;
	  }
	ppt[0] = ppe[0]->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint();
	if ( ppt[0](0) == ppt[1](0) && ppt[0](1) == ppt[1](1) )
	  {
	    MergeVoronoiPolygonsDegenerateShewchuk(ppe[0]->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex(),
							  pmeVoronoi,
							  ppe[1]->GetFacet()->GetForwardVertex());
	    return;
	  }
      }
    else if ( ppe[0]->GetFacet()->GetForwardEdge() != NULL &&
	      ppe[1]->GetFacet()->GetForwardEdge() != NULL )
      {
	ppt[0] = ppe[0]->GetFacet()->GetForwardEdge()->GetVertex()->GetPoint();
	ppt[1] = ppe[1]->GetFacet()->GetForwardEdge()->GetVertex()->GetPoint();
	if ( ppt[0](0) == ppt[1](0) && ppt[0](1) == ppt[1](1) )
	  {
	    MergeVoronoiPolygonsDegenerateShewchuk(ppe[0]->GetFacet()->GetForwardEdge()->GetVertex(),
							  pmeVoronoi,
							  ppe[1]->GetFacet()->GetForwardEdge()->GetVertex());
	    return;
	  }
	ppt[1] = ppe[1]->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint();
	if ( ppt[0](0) == ppt[1](0) && ppt[0](1) == ppt[1](1) )
	  {
	    MergeVoronoiPolygonsDegenerateShewchuk(ppe[0]->GetFacet()->GetForwardEdge()->GetVertex(),
							  pmeVoronoi,
							  ppe[1]->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex());
	    return;
	  }

	ppt[0] = ppe[0]->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint();
	ppt[1] = ppe[1]->GetFacet()->GetForwardEdge()->GetVertex()->GetPoint();
	if ( ppt[0](0) == ppt[1](0) && ppt[0](1) == ppt[1](1) )
	  {
	    MergeVoronoiPolygonsDegenerateShewchuk(ppe[0]->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex(),
							  pmeVoronoi,
							  ppe[1]->GetFacet()->GetForwardEdge()->GetVertex());
	    return;
	  }
	ppt[1] = ppe[1]->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint();
	if ( ppt[0](0) == ppt[1](0) && ppt[0](1) == ppt[1](1) )
	  {
	    MergeVoronoiPolygonsDegenerateShewchuk(ppe[0]->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex(),
							  pmeVoronoi,
							  ppe[1]->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex());
	    return;
	  }
      }
  }

  //////////////////////////////////////////////////
  // Check for Inside or Outside Merge
  //
  BOOL bInside;
  BOOL bLeftInside, bRightInside;
  FLOAT fW;
  QEEdge *peBisector, *peBisectorEnd;
  QEEdge *peLeft, *peRight;
  QEEdge *peLeftStart, *peRightStart;
  QEEdge *peLeftEnd, *peRightEnd;

  peLeftStart = ppe[1];
  peRightStart = ppe[0];
  peLeft = peLeftStart;
  peRight = peRightStart;

  pt = peLeft->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
  bLeftInside = !(pt(2) < 0.0 || pt(3) < 0.0);
  pt = peRight->GetEdgePrev(0)->GetVertex()->GetPoint();
  bRightInside = !(pt(2) < 0.0 || pt(3) < 0.0);
  
  if ( bLeftInside != bRightInside )
    {
      fprintf(stderr, "  ERROR: Distance orientation does not match: Base = %d != Merge = %d\n", bRightInside, bLeftInside);
      ASSERT( FALSE );
    }

  bInside = bLeftInside && bRightInside;
  fW = 1.0;
  if ( !bInside )
    {
      fW = -1.0;
    }

  //////////////////////////////////////////////////
  // Move all the geometry over into the base mesh
  //
  m_plistQEVertex->Splice(pmeVoronoi->m_plistQEVertex);
  m_plistQEFacet->Splice(pmeVoronoi->m_plistQEFacet);
  m_plistQEEdge->Splice(pmeVoronoi->m_plistQEEdge);

  //////////////////////////////////////////////////
  // Find the starting points
  //

  peBisector = NULL;
  peBisectorEnd = NULL;
  peLeft = peLeftStart;
  peRight = peRightStart;
  peLeftEnd = NULL;
  peRightEnd = NULL;

  MergeVoronoiPolygonsShewchuk(peLeftEnd, peLeft,
			       peBisector, peBisectorEnd,
			       peRight, peRightEnd, fW);

  if ( peLeft == peLeftEnd && peRight == peRightEnd )
    {
      //////////////////////////////////////////////////
      // Connect up the bisector loop
      //
      peBisector->SetVertex(peBisectorEnd->GetVertex());
      Splice(1, peBisector, peBisectorEnd);
    }
  else
    {
      QEEdge *peLeftNext, *peRightNext;
      QEVertex *pvxDummy;
      Point ptDummy;

      ASSERT( peLeft != peLeftEnd && peRight != peRightEnd );

      ASSERT( peRightEnd != NULL );
      peLeft = peRightEnd->GetEdgeNext(0);
      if ( peLeft->GetEdge()->GetFacet() == GetNullFacet() &&
	   peLeft->GetFacet()->GetForwardEdge() != peLeft && 
	   peLeft->GetFacet()->GetForwardVertex() != peLeft->GetEdge()->GetVertex() )
	{
	  peLeftNext = peLeft->GetEdgeNext(0);
	  CollapseEdge(peLeft);
	  peLeft = peLeftNext;
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	}
      ASSERT( peLeft->GetFacet() == peRightEnd->GetFacet() );
      ASSERT( peLeft->GetFacet() != GetNullFacet() );

      ASSERT( peLeftEnd != NULL );
      peRight = peLeftEnd->GetEdgePrev(0);
      if ( peRight->GetEdge()->GetFacet() == GetNullFacet() &&
	   peRight->GetFacet()->GetForwardEdge() != peRight &&
	   peRight->GetFacet()->GetForwardVertex() != peRight->GetVertex() )
	{
	  peRightNext = peRight->GetEdgePrev(0);
	  CollapseEdge(peRight->GetEdge());
	  peRight = peRightNext;
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	}
      ASSERT( peRight->GetFacet() == peLeftEnd->GetFacet() );
      ASSERT( peRight->GetFacet() != GetNullFacet() );

      // Disconnect the LeftEnd and RightEnd edges
      ASSERT( peLeftEnd->GetEdgePrev(1) != peLeftEnd );
      peLeftEnd->GetEdgePrev(1)->SetVertex(peLeftEnd->GetEdgePrev(1)->GetVertex());
      //ptDummy = ( peLeftEnd->GetVertex()->GetLimitPosition() + peLeftEnd->GetEdge()->GetVertex()->GetLimitPosition() ) / 2.0;
      ptDummy.Set(-10.0, 0.0, 0.0);
      //ptDummy.Null();
      pvxDummy = CreateVertex(ptDummy);
      pvxDummy->SetLimitPosition(ptDummy);
      Splice(1, peLeftEnd, peLeftEnd->GetEdgePrev(1));
      peLeftEnd->SetVertex(pvxDummy);
      peLeftEnd->SetFacet(GetNullFacet());
      peLeftEnd->GetEdge()->SetFacet(GetNullFacet());
      peRight->SetFacet(peRight->GetFacet());

      ASSERT( peRightEnd->GetEdge()->GetEdgePrev(1) != peRightEnd->GetEdge() );
      peRightEnd->GetEdge()->GetEdgePrev(1)->SetVertex(peRightEnd->GetEdge()->GetEdgePrev(1)->GetVertex());
      //ptDummy = ( peRightEnd->GetVertex()->GetLimitPosition() + peRightEnd->GetEdge()->GetVertex()->GetLimitPosition() ) / 2.0;
      ptDummy.Set(10.0, 0.0, 0.0);
      //ptDummy.Null();
      pvxDummy = CreateVertex(ptDummy);
      pvxDummy->SetLimitPosition(ptDummy);
      Splice(1, peRightEnd->GetEdge(), peRightEnd->GetEdge()->GetEdgePrev(1));
      peRightEnd->GetEdge()->SetVertex(pvxDummy);
      peRightEnd->SetFacet(GetNullFacet());
      peRightEnd->GetEdge()->SetFacet(GetNullFacet());
      peLeft->SetFacet(peLeft->GetFacet());

      MergeVoronoiPolygonsShewchuk(peLeftEnd, peLeft,
				   peBisector, peBisectorEnd,
				   peRight, peRightEnd, fW);

    }
  
  ASSERT( peLeftEnd->GetEdgeNext(1) == peLeftEnd );
  ASSERT( peLeftEnd->GetEdge()->GetEdgeNext(1) == peLeftEnd->GetEdge() );
  ASSERT( peRightEnd->GetEdgeNext(1) == peRightEnd );
  ASSERT( peRightEnd->GetEdge()->GetEdgeNext(1) == peRightEnd->GetEdge() );
  //ASSERT( peBisector->GetFacet() == peBisectorEnd->GetEdgePrev(0)->GetFacet() );
  //ASSERT( peBisector->GetEdge()->GetFacet() == peBisectorEnd->GetFacet() );
      
  DestroyVertex(peLeftEnd->GetVertex());
  DestroyVertex(peRightEnd->GetEdge()->GetVertex());
  DestroyEdge(peLeftEnd->GetEdge());
  DestroyEdge(peLeftEnd);
  DestroyEdge(peRightEnd->GetEdge());
  DestroyEdge(peRightEnd);
      
  //////////////////////////////////////////////////
  // Deallocate the operand mesh
  //
  DeallocateQEMesh(pmeVoronoi);

#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
  fprintf(stderr, "END  : MergeVoronoiPolygons\n");
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING_OUTPUT_FILES
  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING_OUTPUT_FILES
}

VOID QEMesh::MergeVoronoiPolygonsDegenerateShewchuk(QEVertex *pvxBase,
						    QEMesh *pmeVoronoi, QEVertex *pvxVoronoi)
{
  ASSERT( pvxBase != NULL );
  ASSERT( pmeVoronoi != NULL );
  ASSERT( pvxVoronoi != NULL );

  Point ppt[2];

  ppt[0] = pvxBase->GetPoint();
  ASSERT( ppt[0](2) == 0.0 );
  ppt[1] = pvxVoronoi->GetPoint();
  ASSERT( ppt[1](2) == 0.0 );

  ASSERT( ppt[0](0) == ppt[1](0) && ppt[0](1) == ppt[1](1) );

  //////////////////////////////////////////////////
  // Only handling a very simple example so far, 2 triangles touching
  //
  QEVertexIter piterQEVertex[2];
  QEVertex *ppvx[2];
  QEEdge *ppe[2];
  UINT_32 i;
  
  ppvx[0] = pvxBase;
  ppvx[1] = pvxVoronoi;

  piterQEVertex[0].Init(pvxBase);
  piterQEVertex[1].Init(pvxVoronoi);
  for ( i = 0; i < 2; i++ )
    {
      for ( ppe[i] = piterQEVertex[i].PeekFirstEdge();
	    ppe[i] != NULL;
	    ppe[i] = piterQEVertex[i].PeekNextEdge() )
	{
	  ASSERT( ppe[i]->GetFacet() != GetNullFacet() );
	  if ( ppe[i]->GetFacet()->GetForwardVertex() == ppvx[i] )
	    {
	      break;
	    }
	}
      ASSERT( ppe[i] != NULL );
    }

  //////////////////////////////////////////////////
  // Check for Inside or Outside Merge
  //
  BOOL bInside;
  BOOL bLeftInside, bRightInside;
  FLOAT fW;
  QEEdge *peBisector, *peBisectorEnd;
  QEEdge *peLeft, *peRight;
  QEEdge *peLeftStart, *peRightStart;
  QEEdge *peLeftEnd, *peRightEnd;
  QEEdge *peLeftNext, *peRightNext;
  QEEdge *pe;
  Point pt;

  peLeftStart = ppe[1];
  peRightStart = ppe[0];
  peLeft = peLeftStart;
  peRight = peRightStart;

  pt = peLeft->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
  bLeftInside = !(pt(2) < 0.0 || pt(3) < 0.0);
  pt = peRight->GetEdgePrev(0)->GetVertex()->GetPoint();
  bRightInside = !(pt(2) < 0.0 || pt(3) < 0.0);
  
  if ( bLeftInside != bRightInside )
    {
      fprintf(stderr, "  ERROR: Distance orientation does not match: Base = %d != Merge = %d\n", bRightInside, bLeftInside);
      ASSERT( FALSE );
    }

  bInside = bLeftInside && bRightInside;
  fW = 1.0;
  if ( !bInside )
    {
      fW = -1.0;
    }

  //////////////////////////////////////////////////
  // Move all the geometry over into the base mesh
  //
  m_plistQEVertex->Splice(pmeVoronoi->m_plistQEVertex);
  m_plistQEFacet->Splice(pmeVoronoi->m_plistQEFacet);
  m_plistQEEdge->Splice(pmeVoronoi->m_plistQEEdge);

  //////////////////////////////////////////////////
  // Find the starting points
  //

  DestroyFacet(ppe[0]->GetFacet());
  pe = ppe[0];
  do
    {
      pe->SetFacet(GetNullFacet());
      pe = pe->GetEdgeNext(0);
    }
  while ( pe != ppe[0] );
  DestroyFacet(ppe[1]->GetFacet());
  pe = ppe[1];
  do
    {
      pe->SetFacet(GetNullFacet());
      pe = pe->GetEdgeNext(0);
    }
  while ( pe != ppe[1] );
  for ( pe = piterQEVertex[0].PeekFirstEdge();
	pe != NULL;
	pe = piterQEVertex[0].PeekNextEdge() )
    {
      pe->SetVertex(ppvx[1]);
    }
  DestroyVertex(ppvx[1]);

  peBisector = NULL;
  peBisectorEnd = NULL;
  peLeft = peLeftStart->GetEdgePrev(0);
  peRight = peRightStart;
  peLeftEnd = peLeftStart;
  peRightEnd = peRightStart->GetEdgePrev(0);

  // Remove any double null edges
  for ( pe = peLeftEnd->GetEdgeNext(0);
	pe != peLeftEnd;
	pe = pe->GetEdgeNext(0) )
    {
      if ( pe->GetEdge()->GetFacet() == GetNullFacet() )
	{
	  pe->GetEdgePrev(1)->SetVertex(pe->GetEdgePrev(1)->GetVertex());
	  pe->GetEdge()->GetEdgePrev(1)->SetVertex(pe->GetEdge()->GetEdgePrev(1)->GetVertex());
	  Splice(1, pe->GetEdge(), pe->GetEdge()->GetEdgePrev(1));
	  Splice(1, pe, pe->GetEdgePrev(1));
	  DestroyEdge(pe->GetEdge());
	  DestroyEdge(pe);
	  break;
	}
    }
  for ( pe = peRightEnd->GetEdgeNext(0);
	pe != peRightEnd;
	pe = pe->GetEdgeNext(0) )
    {
      if ( pe->GetEdge()->GetFacet() == GetNullFacet() )
	{
	  pe->GetEdgePrev(1)->SetVertex(pe->GetEdgePrev(1)->GetVertex());
	  pe->GetEdge()->GetEdgePrev(1)->SetVertex(pe->GetEdge()->GetEdgePrev(1)->GetVertex());
	  Splice(1, pe->GetEdge(), pe->GetEdge()->GetEdgePrev(1));
	  Splice(1, pe, pe->GetEdgePrev(1));
	  DestroyEdge(pe->GetEdge());
	  DestroyEdge(pe);
	  break;
	}
    }

  // Merge the vertices
  Splice(1, peLeftEnd, peRight);

  peLeftStart = peLeft->GetEdge()->GetEdgePrev(0);
  ASSERT( peLeftStart->GetFacet()->GetForwardEdge() == peLeftStart );
  peRightStart = peRight->GetEdgePrev(1);
  ASSERT( peRightStart->GetFacet()->GetForwardEdge() == peRightStart );

  peLeftEnd = peLeftEnd->GetEdgePrev(1);
  ASSERT( peLeftEnd->GetFacet()->GetForwardEdge() == peLeftEnd );
  peRightEnd = peRightEnd->GetEdge()->GetEdgePrev(0);
  ASSERT( peRightEnd->GetFacet()->GetForwardEdge() == peRightEnd );

  for ( peLeft = peLeftEnd->GetEdgePrev(0)->GetEdgePrev(0);
	peLeft->GetEdge()->GetFacet() == GetNullFacet();
	peLeft = peLeftNext )
    {
      peLeftNext = peLeft->GetEdgePrev(0);
      CollapseEdge(peLeft->GetEdge());
    }
  for ( peRight = peRightEnd->GetEdgeNext(0)->GetEdgeNext(0);
	peRight->GetEdge()->GetFacet() == GetNullFacet();
	peRight = peRightNext )
    {
      peRightNext = peRight->GetEdgeNext(0);
      CollapseEdge(peRight);
    }

  for ( peLeft = peLeftStart->GetEdgeNext(0)->GetEdgeNext(0);
	peLeft->GetEdge()->GetFacet() == GetNullFacet();
	peLeft = peLeftNext )
    {
      peLeftNext = peLeft->GetEdgeNext(0);
      CollapseEdge(peLeft);
    }
  for ( peRight = peRightStart->GetEdgePrev(0)->GetEdgePrev(0);
	peRight->GetEdge()->GetFacet() == GetNullFacet();
	peRight = peRightNext )
    {
      peRightNext = peRight->GetEdgePrev(0);
      CollapseEdge(peRight->GetEdge());
    }


  //////////////////////////////////////////////////
  // Start the bisector

  peLeft = peLeftStart->GetEdgeNext(0);
  peLeftNext = peLeft->GetEdgeNext(0);
  peRight = peRightStart->GetEdgePrev(0);
  peRightNext = peRight->GetEdgePrev(0);
  
  peLeft->GetEdge()->GetEdgePrev(1)->SetVertex(peLeft->GetEdge()->GetEdgePrev(1)->GetVertex());
  peLeft->GetEdgePrev(1)->SetVertex(peLeft->GetEdgePrev(1)->GetVertex());
  Splice(1, peLeft->GetEdge(), peLeft->GetEdge()->GetEdgePrev(1));
  Splice(1, peLeft, peLeft->GetEdgePrev(1));
  DestroyEdge(peLeft->GetEdge());
  DestroyEdge(peLeft);
  peLeft = peLeftNext;
  peLeft->SetFacet(peLeft->GetFacet());

  peRight->GetEdge()->GetEdgePrev(1)->SetVertex(peRight->GetEdge()->GetEdgePrev(1)->GetVertex());
  peRight->GetEdgePrev(1)->SetVertex(peRight->GetEdgePrev(1)->GetVertex());
  Splice(1, peRight->GetEdge(), peRight->GetEdge()->GetEdgePrev(1));
  Splice(1, peRight, peRight->GetEdgePrev(1));
  DestroyEdge(peRight->GetEdge());
  DestroyEdge(peRight);
  peRight = peRightNext;
  peRight->SetFacet(peRight->GetFacet());

  peBisector = NULL;
  peBisectorEnd = peRightStart;

  MergeVoronoiPolygonsShewchuk(peLeftEnd, peLeft,
			       peBisector, peBisectorEnd,
			       peRight, peRightEnd, fW);

  if ( peLeft == peLeftEnd && peRight == peRightEnd )
    {
      //////////////////////////////////////////////////
      // Connect up the bisector loop
      //
      peBisector->SetVertex(peBisectorEnd->GetVertex());
      Splice(1, peBisector, peBisectorEnd);
    }
  else
    {
      peLeft = peLeftEnd->GetEdgePrev(0);
      peLeftNext = peLeft->GetEdgePrev(0);
      peRight = peRightEnd->GetEdgeNext(0);
      peRightNext = peRight->GetEdgeNext(0);
  
      // Break open left side
      peLeft->GetEdge()->GetEdgePrev(1)->SetVertex(peLeft->GetEdge()->GetEdgePrev(1)->GetVertex());
      peLeft->GetEdgePrev(1)->SetVertex(peLeft->GetEdgePrev(1)->GetVertex());
      Splice(1, peLeft->GetEdge(), peLeft->GetEdge()->GetEdgePrev(1));
      Splice(1, peLeft, peLeft->GetEdgePrev(1));
      DestroyEdge(peLeft->GetEdge());
      DestroyEdge(peLeft);
      peLeft = peLeftNext;
      peLeft->SetFacet(peLeft->GetFacet());

      // Break open right side
      peRight->GetEdge()->GetEdgePrev(1)->SetVertex(peRight->GetEdge()->GetEdgePrev(1)->GetVertex());
      peRight->GetEdgePrev(1)->SetVertex(peRight->GetEdgePrev(1)->GetVertex());
      Splice(1, peRight->GetEdge(), peRight->GetEdge()->GetEdgePrev(1));
      Splice(1, peRight, peRight->GetEdgePrev(1));
      DestroyEdge(peRight->GetEdge());
      DestroyEdge(peRight);
      peRight = peRightNext;
      peRight->SetFacet(peRight->GetFacet());

      peBisector = NULL;
      peBisectorEnd = peLeftEnd;

      MergeVoronoiPolygonsShewchuk(peRightEnd, peRight,
				   peBisector, peBisectorEnd,
				   peLeft, peLeftEnd, fW);
    }

  //////////////////////////////////////////////////
  // Deallocate the operand mesh
  //
  DeallocateQEMesh(pmeVoronoi);

#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
  fprintf(stderr, "END  : MergeVoronoiPolygons\n");
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
}

VOID QEMesh::MergeVoronoiPolygonsVertexZoneShewchuk(QEEdge *peLeftUp, QEEdge *peLeftDown, 
						    QEEdge *peRightUp, QEEdge *peRightDown)
{
  QEVertex *pvxLeft, *pvxRight;
  Point ptLeft, ptRight;

  ASSERT( peLeftUp != NULL );
  ASSERT( peLeftDown != NULL );
  ASSERT( peRightUp != NULL );
  ASSERT( peRightDown != NULL );

  ASSERT( peLeftUp->GetFacet()->GetForwardEdge() == peLeftUp );
  ASSERT( peLeftDown->GetFacet()->GetForwardEdge() == peLeftDown );
  ASSERT( peRightUp->GetFacet()->GetForwardEdge() == peRightUp );
  ASSERT( peRightDown->GetFacet()->GetForwardEdge() == peRightDown );

  ASSERT( peLeftUp->GetEdge()->GetVertex() == peLeftDown->GetVertex() );
  ASSERT( peRightUp->GetVertex() == peRightDown->GetEdge()->GetVertex() );

  pvxLeft = peLeftUp->GetEdge()->GetVertex();
  pvxRight = peRightUp->GetVertex();

  ASSERT( pvxLeft != pvxRight );

  ptLeft = pvxLeft->GetPoint();
  ptRight = pvxRight->GetPoint();
  
  ASSERT( ptLeft(0) == ptRight(0) && ptLeft(1) == ptRight(1) );

}

QEFacet *QEMesh::PointLocateVoronoiPolygonShewchuk(QEFacet *pfaTest, QEVertex *pvxTest)
{
  QEFacet *ppfa[2];

  ASSERT( pfaTest->GetForwardVertex() == pvxTest );

  ppfa[0] = PointLocateVoronoiPolygonShewchukGradientDescent(pfaTest, pvxTest);
  /*
  ppfa[1] = PointLocateVoronoiPolygonShewchukNaive(pfaTest);
  if ( ppfa[0] != ppfa[1] )
    {
      QEVertex *ppvxEndpoint[2];
      FLOAT pfDistance[2];
      Point pt;

      pt = pvxTest->GetPoint();
      
      pfDistance[0] = CalculateDistance(ppfa[0], pt, &ppvxEndpoint[0]);
      pfDistance[1] = CalculateDistance(ppfa[1], pt, &ppvxEndpoint[1]);
      ASSERT( pfDistance[0] - EPSILON <= pfDistance[1] && 
	      pfDistance[0] + EPSILON >= pfDistance[1] );
    }
  */

  return ppfa[0];
}

QEFacet *QEMesh::PointLocateVoronoiPolygonShewchuk(QEFacet *pfaTest, QEEdge *peStart)
{
  QEFacet *ppfa[2];

  ppfa[0] = PointLocateVoronoiPolygonShewchukGradientDescent(pfaTest, peStart);
  /*
  ppfa[1] = PointLocateVoronoiPolygonShewchukNaive(pfaTest);
  if ( ppfa[0] != ppfa[1] )
    {
      QEVertex *ppvxEndpoint[2];
      FLOAT pfDistance[2];
      
      pfDistance[0] = CalculateDistance(ppfa[0], pt, &ppvxEndpoint[0]);
      pfDistance[1] = CalculateDistance(ppfa[1], pt, &ppvxEndpoint[1]);
      ASSERT( pfDistance[0] - EPSILON <= pfDistance[1] && 
	      pfDistance[0] + EPSILON >= pfDistance[1] );
    }
  */

  return ppfa[0];
}

APM2D::EdgeZoneType QEMesh::EdgeZone(const Point &pt, QEEdge *pe, 
				     QEVertex **ppvxEndpoint)
{
  ASSERT( pe != NULL && 
	  ppvxEndpoint != NULL );

  QEVertex *ppvx[2];
  APM2D::EdgeZoneType eEZT;
  Point ppt[2];

  *ppvxEndpoint = NULL;

  ppvx[0] = pe->GetVertex();
  ppvx[1] = pe->GetEdge()->GetVertex();
  ppt[0] = ppvx[0]->GetPoint();
  ppt[1] = ppvx[1]->GetPoint();

  eEZT = APM2D::EdgeZone(pt(0), pt(1), 
			 ppt[0](0), ppt[0](1),
			 ppt[1](0), ppt[1](1));
  switch ( eEZT )
    {
    case APM2D::eztStart:
      {
	*ppvxEndpoint = ppvx[0];
      }
    break;
    case APM2D::eztEnd:
      {
	*ppvxEndpoint = ppvx[1];
      }
    break;
    case APM2D::eztMiddle:
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  return eEZT;
}

INT_32 QEMesh::DistanceCompareS_SS(QEFacet *pfaTest, 
				   QEVertex **ppvxEndpointTest0,
				   QEVertex **ppvxEndpointTest1,
				   QEFacet *pfa0, QEVertex **ppvxEndpoint0,
				   QEFacet *pfa1, QEVertex **ppvxEndpoint1)
{
  ASSERT( pfaTest != NULL && 
	  ppvxEndpointTest0 != NULL &&
	  ppvxEndpointTest1 != NULL &&
	  pfa0 != NULL &&
	  ppvxEndpoint0 != NULL &&
	  pfa1 != NULL && 
	  ppvxEndpoint1 != NULL );

  *ppvxEndpointTest0 = NULL;
  *ppvxEndpointTest1 = NULL;
  *ppvxEndpoint0 = NULL;
  *ppvxEndpoint1 = NULL;

#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
  fprintf(stderr, "START: DistanceCompareS_SS\n");
  if ( pfaTest->GetForwardVertex() != NULL )
    {
      fprintf(stderr, "  Test Vertex: ( %.3f, %.3f, %.3f )\n", 
	      pfaTest->GetForwardVertex()->GetPoint()(0),
	      pfaTest->GetForwardVertex()->GetPoint()(1),
	      pfaTest->GetForwardVertex()->GetPoint()(2));
    }
  else
    {
      fprintf(stderr, "  Test Edge: ( %.3f, %.3f, %.3f ) -> ( %.3f, %.3f, %.3f )\n", 
	      pfaTest->GetForwardEdge()->GetVertex()->GetPoint()(0),
	      pfaTest->GetForwardEdge()->GetVertex()->GetPoint()(1),
	      pfaTest->GetForwardEdge()->GetVertex()->GetPoint()(2),
	      pfaTest->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint()(0),
	      pfaTest->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint()(1),
	      pfaTest->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint()(2));
    }
  if ( pfa0->GetForwardVertex() != NULL )
    {
      fprintf(stderr, "  Vertex 0: ( %.3f, %.3f, %.3f )\n", 
	      pfa0->GetForwardVertex()->GetPoint()(0),
	      pfa0->GetForwardVertex()->GetPoint()(1),
	      pfa0->GetForwardVertex()->GetPoint()(2));
    }
  else
    {
      fprintf(stderr, "  Edge 0: ( %.3f, %.3f, %.3f ) -> ( %.3f, %.3f, %.3f )\n", 
	      pfa0->GetForwardEdge()->GetVertex()->GetPoint()(0),
	      pfa0->GetForwardEdge()->GetVertex()->GetPoint()(1),
	      pfa0->GetForwardEdge()->GetVertex()->GetPoint()(2),
	      pfa0->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint()(0),
	      pfa0->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint()(1),
	      pfa0->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint()(2));
    }
  if ( pfa1->GetForwardVertex() != NULL )
    {
      fprintf(stderr, "  Vertex 1: ( %.3f, %.3f, %.3f )\n", 
	      pfa1->GetForwardVertex()->GetPoint()(0),
	      pfa1->GetForwardVertex()->GetPoint()(1),
	      pfa1->GetForwardVertex()->GetPoint()(2));
    }
  else
    {
      fprintf(stderr, "  Edge 1: ( %.3f, %.3f, %.3f ) -> ( %.3f, %.3f, %.3f )\n", 
	      pfa1->GetForwardEdge()->GetVertex()->GetPoint()(0),
	      pfa1->GetForwardEdge()->GetVertex()->GetPoint()(1),
	      pfa1->GetForwardEdge()->GetVertex()->GetPoint()(2),
	      pfa1->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint()(0),
	      pfa1->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint()(1),
	      pfa1->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint()(2));
    }
  fprintf(stderr, "END  : DistanceCompareS_SS\n");
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING

  if ( pfaTest->GetForwardVertex() != NULL )
    {
      return DistanceCompareV_SS(pfaTest->GetForwardVertex(), 
				 pfa0, ppvxEndpoint0,
				 pfa1, ppvxEndpoint1);
    }
  else
    {
      ASSERT( pfaTest->GetForwardEdge() != NULL );
      return DistanceCompareE_SS(pfaTest->GetForwardEdge(), 
				 ppvxEndpointTest0,
				 ppvxEndpointTest1,
				 pfa0, ppvxEndpoint0,
				 pfa1, ppvxEndpoint1);
    }
}

INT_32 QEMesh::DistanceCompareV_SS(QEVertex *pvxTest,
				   QEFacet *pfa0, QEVertex **ppvxEndpoint0,
				   QEFacet *pfa1, QEVertex **ppvxEndpoint1)
{
  ASSERT( pvxTest != NULL && 
	  pfa0 != NULL &&
	  ppvxEndpoint0 != NULL &&
	  pfa1 != NULL && 
	  ppvxEndpoint1 != NULL );

  QEFacet *ppfa[2];
  QEVertex **pppvxEndpoint[2];
  QEVertex *ppvx[2];
  QEEdge *ppe[2];
  APM2D::EdgeZoneType peEZT[2];
  Point pt, pppt[2][2];
  UINT_32 i;
  INT_32 iResult;

  pt = pvxTest->GetPoint();

  ppfa[0] = pfa0;
  ppfa[1] = pfa1;
  *ppvxEndpoint0 = NULL;
  pppvxEndpoint[0] = ppvxEndpoint0;
  *ppvxEndpoint1 = NULL;
  pppvxEndpoint[1] = ppvxEndpoint1;
  ppvx[0] = NULL;
  ppvx[1] = NULL;
  ppe[0] = NULL;
  ppe[1] = NULL;
  peEZT[0] = APM2D::eztMiddle;
  peEZT[1] = APM2D::eztMiddle;

  for ( i = 0; i < 2; i++ )
    {
      if ( ppfa[i]->GetForwardVertex() != NULL )
	{
	  ppvx[i] = ppfa[i]->GetForwardVertex();
	  pppt[i][0] = ppvx[i]->GetPoint();
	}
      else
	{
	  ASSERT( ppfa[i]->GetForwardEdge() != NULL );
	  ppe[i] = ppfa[i]->GetForwardEdge();

	  pppt[i][0] = ppe[i]->GetVertex()->GetPoint();
	  pppt[i][1] = ppe[i]->GetEdge()->GetVertex()->GetPoint();

	  peEZT[i] = APM2D::EdgeZone(pt(0), pt(1), 
				     pppt[i][0](0), pppt[i][0](1),
				     pppt[i][1](0), pppt[i][1](1));
	  switch ( peEZT[i] )
	    {
	    case APM2D::eztStart:
	      {
		ppvx[i] = ppe[i]->GetVertex();
		*pppvxEndpoint[i] = ppvx[i];
	      }
	      break;
	    case APM2D::eztEnd:
	      {
		ppvx[i] = ppe[i]->GetEdge()->GetVertex();
		*pppvxEndpoint[i] = ppvx[i];
		pppt[i][0] = pppt[i][1];
	      }
	      break;
	    case APM2D::eztMiddle:
	      break;
	    default:
	      ASSERT( FALSE );
	      break;
	    }
	}
    }

  if ( ppvx[0] != NULL && ppvx[1] != NULL )
    {
      if ( ppvx[0] == ppvx[1] )
	{
	  iResult = 0;
	}
      else
	{
	  iResult = APM2D::DistanceCompareV_VV(pt(0), pt(1),
					       pppt[0][0](0), pppt[0][0](1),
					       pppt[1][0](0), pppt[1][0](1));
	}
    }
  else if ( ppvx[0] != NULL )
    {
      iResult = APM2D::DistanceCompareV_VL(pt(0), pt(1),
					   pppt[0][0](0), pppt[0][0](1),
					   pppt[1][0](0), pppt[1][0](1),
					   pppt[1][1](0), pppt[1][1](1));
    }
  else if ( ppvx[1] != NULL )
    {
      iResult = APM2D::DistanceCompareV_VL(pt(0), pt(1),
					   pppt[1][0](0), pppt[1][0](1),
					   pppt[0][0](0), pppt[0][0](1),
					   pppt[0][1](0), pppt[0][1](1));
      iResult = -iResult;
    }
  else
    {
      iResult = APM2D::DistanceCompareV_LL(pt(0), pt(1),
					   pppt[0][0](0), pppt[0][0](1),
					   pppt[0][1](0), pppt[0][1](1),
					   pppt[1][0](0), pppt[1][0](1),
					   pppt[1][1](0), pppt[1][1](1));
    }

  return iResult;
}

INT_32 QEMesh::DistanceCompareE_SS(QEEdge *peTest,
				    QEVertex **ppvxEndpointTest0,
				    QEVertex **ppvxEndpointTest1,
				    QEFacet *pfa0, QEVertex **ppvxEndpoint0,
				    QEFacet *pfa1, QEVertex **ppvxEndpoint1)
{
  ASSERT( peTest != NULL && 
	  ppvxEndpointTest0 != NULL &&
	  ppvxEndpointTest1 != NULL &&
	  pfa0 != NULL &&
	  ppvxEndpoint0 != NULL &&
	  pfa1 != NULL && 
	  ppvxEndpoint1 != NULL );

  QEVertex *ppvxTest[2];
  Point ppptTest[2][2];
  QEVertex **pppvxEndpointTest[2];
  APM2D::EdgeZoneType ppeEZTTest[2][2];

  QEFacet *ppfa[2];
  QEVertex **pppvxEndpoint[2];
  QEVertex *ppvx[2];
  QEEdge *ppe[2];
  APM2D::EdgeZoneType ppeEZT[2][2];
  //Point pt;
  Point pppt[2][2];
  UINT_32 i, j, uiEndpoint;
  INT_32 iResult;

  ppvxTest[0] = NULL;
  ppvxTest[1] = NULL;
  ppptTest[0][0] = peTest->GetVertex()->GetPoint();
  ppptTest[0][1] = peTest->GetEdge()->GetVertex()->GetPoint();
  ppptTest[1][0] = ppptTest[0][0];
  ppptTest[1][1] = ppptTest[0][1];
  *ppvxEndpointTest0 = NULL;
  pppvxEndpointTest[0] = ppvxEndpointTest0;
  *ppvxEndpointTest1 = NULL;
  pppvxEndpointTest[1] = ppvxEndpointTest1;
  ppeEZTTest[0][0] = APM2D::eztMiddle;
  ppeEZTTest[0][1] = APM2D::eztMiddle;
  ppeEZTTest[1][0] = APM2D::eztMiddle;
  ppeEZTTest[1][1] = APM2D::eztMiddle;

  ppfa[0] = pfa0;
  ppfa[1] = pfa1;
  *ppvxEndpoint0 = NULL;
  pppvxEndpoint[0] = ppvxEndpoint0;
  *ppvxEndpoint1 = NULL;
  pppvxEndpoint[1] = ppvxEndpoint1;
  ppvx[0] = NULL;
  ppvx[1] = NULL;
  ppe[0] = NULL;
  ppe[1] = NULL;
  ppeEZT[0][0] = APM2D::eztMiddle;
  ppeEZT[0][1] = APM2D::eztMiddle;
  ppeEZT[1][0] = APM2D::eztMiddle;
  ppeEZT[1][1] = APM2D::eztMiddle;

  for ( i = 0; i < 2; i++ )
    {
      if ( ppfa[i]->GetForwardVertex() != NULL )
	{
	  ppvx[i] = ppfa[i]->GetForwardVertex();
	  pppt[i][0] = ppvx[i]->GetPoint();

	  ppeEZTTest[i][0] = APM2D::EdgeZone(pppt[i][0](0), pppt[i][0](1), 
					     ppptTest[i][0](0), ppptTest[i][0](1),
					     ppptTest[i][1](0), ppptTest[i][1](1));
	  switch ( ppeEZTTest[i][0] )
	    {
	    case APM2D::eztStart:
	      {
		ppvxTest[i] = peTest->GetVertex();
		*pppvxEndpointTest[i] = ppvxTest[i];
	      }
	      break;
	    case APM2D::eztEnd:
	      {
		ppvxTest[i] = peTest->GetEdge()->GetVertex();
		*pppvxEndpointTest[i] = ppvxTest[i];
	      }
	      break;
	    case APM2D::eztMiddle:
	      break;
	    default:
	      ASSERT( FALSE );
	      break;
	    }
	}
      else
	{
	  ASSERT( ppfa[i]->GetForwardEdge() != NULL );
	  ppe[i] = ppfa[i]->GetForwardEdge();

	  pppt[i][0] = ppe[i]->GetVertex()->GetPoint();
	  pppt[i][1] = ppe[i]->GetEdge()->GetVertex()->GetPoint();

	  for ( j = 0; j < 2; j++ )
	    {
	      ppeEZTTest[i][j] = APM2D::EdgeZone(pppt[i][j](0), pppt[i][j](1), 
						 ppptTest[i][0](0), ppptTest[i][0](1),
						 ppptTest[i][1](0), ppptTest[i][1](1));
	      ppeEZT[i][j] = APM2D::EdgeZone(ppptTest[i][j](0), ppptTest[i][j](1), 
					     pppt[i][0](0), pppt[i][0](1),
					     pppt[i][1](0), pppt[i][1](1));
	    }

	  // Check to see if we can treat ppe[i]->GetForwardEdge() like one of its endpoints
	  if ( ppeEZT[i][0] == ppeEZT[i][1] ||
	       ppeEZTTest[i][0] == ppeEZTTest[i][1] )
	    {
	      if ( ppeEZTTest[i][0] == APM2D::eztStart &&
		   ppeEZTTest[i][1] == APM2D::eztStart )
		{
		  ppvxTest[i] = peTest->GetVertex();
		  *pppvxEndpointTest[i] = ppvxTest[i];
		}
	      else if ( ppeEZTTest[i][0] == APM2D::eztEnd &&
			ppeEZTTest[i][1] == APM2D::eztEnd )
		{
		  ppvxTest[i] = peTest->GetEdge()->GetVertex();
		  *pppvxEndpointTest[i] = ppvxTest[i];
		}
	      else if ( ppeEZT[i][0] == APM2D::eztMiddle &&
			ppeEZT[i][1] == APM2D::eztMiddle )
		{
		  iResult = APM2D::DistanceCompareL_VV(pppt[i][0](0), pppt[i][0](1),
						       pppt[i][1](0), pppt[i][1](1),
						       ppptTest[i][0](0), ppptTest[i][0](1), 
						       ppptTest[i][1](0), ppptTest[i][1](1));
		  if ( iResult < 0 )
		    {
		      ppvxTest[i] = peTest->GetVertex();
		      *pppvxEndpointTest[i] = ppvxTest[i];
		    }
		  else
		    {
		      ppvxTest[i] = peTest->GetEdge()->GetVertex();
		      *pppvxEndpointTest[i] = ppvxTest[i];
		    }
		}

	      if ( ppeEZT[i][0] == APM2D::eztStart &&
		   ppeEZT[i][1] == APM2D::eztStart )
		{
		  ppvx[i] = ppe[i]->GetVertex();
		  *pppvxEndpoint[i] = ppvx[i];
		}
	      else if ( ppeEZT[i][0] == APM2D::eztEnd &&
			ppeEZT[i][1] == APM2D::eztEnd )
		{
		  ppvx[i] = ppe[i]->GetEdge()->GetVertex();
		  *pppvxEndpoint[i] = ppvx[i];
		}
	      else if ( ppeEZTTest[i][0] == APM2D::eztMiddle &&
			ppeEZTTest[i][1] == APM2D::eztMiddle )
		{
		  iResult = APM2D::DistanceCompareL_VV(ppptTest[i][0](0), ppptTest[i][0](1), 
						       ppptTest[i][1](0), ppptTest[i][1](1),
						       pppt[i][0](0), pppt[i][0](1),
						       pppt[i][1](0), pppt[i][1](1));
		  if ( iResult < 0 )
		    {
		      ppvx[i] = ppe[i]->GetVertex();
		      *pppvxEndpoint[i] = ppvx[i];
		    }
		  else
		    {
		      ppvx[i] = ppe[i]->GetEdge()->GetVertex();
		      *pppvxEndpoint[i] = ppvx[i];
		    }
		}

	      ASSERT( ppvx[i] != NULL || ppvxTest[i] != NULL );
	      if ( ppvx[i] != NULL && ppvxTest[i] == NULL )
		{
		  if ( ppe[i]->GetVertex() == ppvx[i] )
		    {
		      uiEndpoint = 0;
		    }
		  else
		    {
		      uiEndpoint = 1;
		    }

		  switch ( ppeEZTTest[i][uiEndpoint] )
		    { 
		    case APM2D::eztStart:
		      {
			ppvxTest[i] = peTest->GetVertex();
			*pppvxEndpointTest[i] = ppvxTest[i];
		      }
		      break;
		    case APM2D::eztEnd:
		      {
			ppvxTest[i] = peTest->GetEdge()->GetVertex();
			*pppvxEndpointTest[i] = ppvxTest[i];
		      }
		      break;
		    case APM2D::eztMiddle:
		      break;
		    default:
		      ASSERT( FALSE );
		      break;
		    }
		}
	      else if ( ppvx[i] == NULL && ppvxTest[i] != NULL )
		{
		  if ( peTest->GetVertex() == ppvxTest[i] )
		    {
		      uiEndpoint = 0;
		    }
		  else
		    {
		      uiEndpoint = 1;
		    }

		  switch ( ppeEZT[i][uiEndpoint] )
		    { 
		    case APM2D::eztStart:
		      {
			ppvx[i] = ppe[i]->GetVertex();
			*pppvxEndpoint[i] = ppvx[i];
		      }
		      break;
		    case APM2D::eztEnd:
		      {
			ppvx[i] = ppe[i]->GetEdge()->GetVertex();
			*pppvxEndpoint[i] = ppvx[i];
		      }
		      break;
		    case APM2D::eztMiddle:
		      break;
		    default:
		      ASSERT( FALSE );
		      break;
		    }
		}
	    }
	  else
	    {
	      // Check if each edge has one vertex in the middle zone of the other edge
	      if ( ( ( ppeEZT[i][0] == APM2D::eztMiddle &&
		       ( ppeEZTTest[i][0] == APM2D::eztStart ||
			 ppeEZTTest[i][1] == APM2D::eztStart ) ) ||
		     ( ppeEZT[i][1] == APM2D::eztMiddle &&
		       ( ppeEZTTest[i][0] == APM2D::eztEnd ||
			 ppeEZTTest[i][1] == APM2D::eztEnd ) ) ) &&
		   ( ( ppeEZTTest[i][0] == APM2D::eztMiddle &&
		       ( ppeEZT[i][0] == APM2D::eztStart ||
			 ppeEZT[i][1] == APM2D::eztStart ) ) ||
		     ( ppeEZTTest[i][1] == APM2D::eztMiddle &&
		       ( ppeEZT[i][0] == APM2D::eztEnd ||
			 ppeEZT[i][1] == APM2D::eztEnd ) ) ) )
		{
		  QEVertex *ppvxTemp[2];

		  if ( ppeEZT[i][0] == APM2D::eztMiddle )
		    {
		      ppvxTemp[0] = peTest->GetVertex();
		    }
		  else 
		    {
		      ASSERT( ppeEZT[i][1] == APM2D::eztMiddle );
		      ppvxTemp[0] = peTest->GetEdge()->GetVertex();
		    }
		  
		  if ( ppeEZTTest[i][0] == APM2D::eztMiddle )
		    {
		      ppvxTemp[1] = ppe[i]->GetVertex();
		    }
		  else
		    {
		      ASSERT( ppeEZTTest[i][1] == APM2D::eztMiddle );
		      ppvxTemp[1] = ppe[i]->GetEdge()->GetVertex();
		    }

		  iResult = APM2D::DistanceCompareEvv_LL(ppvxTemp[0]->GetPoint()(0), ppvxTemp[0]->GetPoint()(1), 
							 ppvxTemp[1]->GetPoint()(0), ppvxTemp[1]->GetPoint()(1), 
							 pppt[i][0](0), pppt[i][0](1),
							 pppt[i][1](0), pppt[i][1](1),
							 ppptTest[i][0](0), ppptTest[i][0](1), 
							 ppptTest[i][1](0), ppptTest[i][1](1));

		  if ( iResult < 0 )
		    {
		      ppvxTest[i] = ppvxTemp[0];
		      *pppvxEndpointTest[i] = ppvxTest[i];
		    }
		  else
		    {
		      ppvx[i] = ppvxTemp[1];
		      *pppvxEndpoint[i] = ppvx[i];		  
		    }
		}
	      else if ( ppeEZT[i][0] == APM2D::eztStart &&
			ppeEZTTest[i][0] == APM2D::eztStart )
		{
		  ASSERT( !( ( ppeEZT[i][1] == APM2D::eztStart &&
			       ppeEZTTest[i][0] == APM2D::eztEnd ) ||
			     ( ppeEZT[i][0] == APM2D::eztEnd &&
			       ppeEZTTest[i][1] == APM2D::eztStart ) ||
			     ( ppeEZT[i][1] == APM2D::eztEnd &&
			       ppeEZTTest[i][1] == APM2D::eztEnd ) ) );
		  
		  ppvx[i] = ppe[i]->GetVertex();
		  *pppvxEndpoint[i] = ppvx[i];		  
		  ppvxTest[i] = peTest->GetVertex();
		  *pppvxEndpointTest[i] = ppvxTest[i];
		}
	      else if ( ppeEZT[i][1] == APM2D::eztStart &&
			ppeEZTTest[i][0] == APM2D::eztEnd )
		{
		  ASSERT( !( ( ppeEZT[i][0] == APM2D::eztStart &&
			       ppeEZTTest[i][0] == APM2D::eztStart ) ||
			     ( ppeEZT[i][0] == APM2D::eztEnd &&
			       ppeEZTTest[i][1] == APM2D::eztStart ) ||
			     ( ppeEZT[i][1] == APM2D::eztEnd &&
			       ppeEZTTest[i][1] == APM2D::eztEnd ) ) );

		  ppvx[i] = ppe[i]->GetVertex();
		  *pppvxEndpoint[i] = ppvx[i];		  
		  ppvxTest[i] = peTest->GetEdge()->GetVertex();
		  *pppvxEndpointTest[i] = ppvxTest[i];
		}
	      else if ( ppeEZT[i][0] == APM2D::eztEnd &&
			ppeEZTTest[i][1] == APM2D::eztStart )
		{
		  ASSERT( !( ( ppeEZT[i][0] == APM2D::eztStart &&
			       ppeEZTTest[i][0] == APM2D::eztStart ) ||
			     ( ppeEZT[i][1] == APM2D::eztStart &&
			       ppeEZTTest[i][0] == APM2D::eztEnd ) ||
			     ( ppeEZT[i][1] == APM2D::eztEnd &&
			       ppeEZTTest[i][1] == APM2D::eztEnd ) ) );

		  ppvx[i] = ppe[i]->GetEdge()->GetVertex();
		  *pppvxEndpoint[i] = ppvx[i];		  
		  ppvxTest[i] = peTest->GetVertex();
		  *pppvxEndpointTest[i] = ppvxTest[i];
		}
	      else 
		{
		  ASSERT( ppeEZT[i][1] == APM2D::eztEnd &&
			  ppeEZTTest[i][1] == APM2D::eztEnd );

		  ASSERT( !( ( ppeEZT[i][0] == APM2D::eztStart &&
			       ppeEZTTest[i][0] == APM2D::eztStart ) ||
			     ( ppeEZT[i][1] == APM2D::eztStart &&
			       ppeEZTTest[i][0] == APM2D::eztEnd ) ||
			     ( ppeEZT[i][0] == APM2D::eztEnd &&
			       ppeEZTTest[i][1] == APM2D::eztStart ) ) );

		  ppvx[i] = ppe[i]->GetEdge()->GetVertex();
		  *pppvxEndpoint[i] = ppvx[i];		  
		  ppvxTest[i] = peTest->GetEdge()->GetVertex();
		  *pppvxEndpointTest[i] = ppvxTest[i];
		}
	    }
	}
    }

  if ( ppvxTest[0] != NULL )
    {
      ppptTest[0][0] = ppvxTest[0]->GetPoint();
    }
  if ( ppvxTest[1] != NULL )
    {
      ppptTest[1][0] = ppvxTest[1]->GetPoint();
    }
  if ( ppvx[0] != NULL )
    {
      pppt[0][0] = ppvx[0]->GetPoint();
    }
  if ( ppvx[1] != NULL )
    {
      pppt[1][0] = ppvx[1]->GetPoint();
    }

  if ( ppvxTest[0] != NULL && ppvxTest[1] != NULL )
    {
      if ( ppvx[0] != NULL && ppvx[1] != NULL )
	{
	  iResult = APM2D::DistanceCompareEvv_VV(ppptTest[0][0](0), ppptTest[0][0](1),
						 ppptTest[1][0](0), ppptTest[1][0](1),
						 pppt[0][0](0), pppt[0][0](1),
						 pppt[1][0](0), pppt[1][0](1));
	}
      else if ( ppvx[0] != NULL )
	{
	  iResult = APM2D::DistanceCompareEvv_VL(ppptTest[0][0](0), ppptTest[0][0](1),
						 ppptTest[1][0](0), ppptTest[1][0](1),
						 pppt[0][0](0), pppt[0][0](1),
						 pppt[1][0](0), pppt[1][0](1),
						 pppt[1][1](0), pppt[1][1](1));
	}
      else if ( ppvx[1] != NULL )
	{
	  iResult = APM2D::DistanceCompareEvv_VL(ppptTest[1][0](0), ppptTest[1][0](1),
						 ppptTest[0][0](0), ppptTest[0][0](1),
						 pppt[1][0](0), pppt[1][0](1),
						 pppt[0][0](0), pppt[0][0](1),
						 pppt[0][1](0), pppt[0][1](1));
	  iResult = -iResult;
	}
      else
	{
	  iResult = APM2D::DistanceCompareEvv_LL(ppptTest[0][0](0), ppptTest[0][0](1),
						 ppptTest[1][0](0), ppptTest[1][0](1),
						 pppt[0][0](0), pppt[0][0](1),
						 pppt[0][1](0), pppt[0][1](1),
						 pppt[1][0](0), pppt[1][0](1),
						 pppt[1][1](0), pppt[1][1](1));
	}
    }
  else if ( ppvxTest[0] == NULL && ppvxTest[1] != NULL )
    {
      ASSERT( ppvx[0] != NULL );
      if ( ppvx[1] != NULL )
	{
	  iResult = APM2D::DistanceCompareElv_VV(ppptTest[0][0](0), ppptTest[0][0](1),
						 ppptTest[0][1](0), ppptTest[0][1](1),
						 ppptTest[1][0](0), ppptTest[1][0](1),
						 pppt[0][0](0), pppt[0][0](1),
						 pppt[1][0](0), pppt[1][0](1));
	}
      else
	{
	  iResult = APM2D::DistanceCompareElv_VL(ppptTest[0][0](0), ppptTest[0][0](1),
						 ppptTest[0][1](0), ppptTest[0][1](1),
						 ppptTest[1][0](0), ppptTest[1][0](1),
						 pppt[0][0](0), pppt[0][0](1),
						 pppt[1][0](0), pppt[1][0](1),
						 pppt[1][1](0), pppt[1][1](1));
	}
    }
  else if ( ppvxTest[0] != NULL && ppvxTest[1] == NULL )
    {
      ASSERT( ppvx[1] != NULL );
      if ( ppvx[0] != NULL )
	{
	  iResult = APM2D::DistanceCompareElv_VV(ppptTest[1][0](0), ppptTest[1][0](1),
						 ppptTest[1][1](0), ppptTest[1][1](1),
						 ppptTest[0][0](0), ppptTest[0][0](1),
						 pppt[1][0](0), pppt[1][0](1),
						 pppt[0][0](0), pppt[0][0](1));
	}
      else
	{
	  iResult = APM2D::DistanceCompareElv_VL(ppptTest[1][0](0), ppptTest[1][0](1),
						 ppptTest[1][1](0), ppptTest[1][1](1),
						 ppptTest[0][0](0), ppptTest[0][0](1),
						 pppt[1][0](0), pppt[1][0](1),
						 pppt[0][0](0), pppt[0][0](1),
						 pppt[0][1](0), pppt[0][1](1));
	}
      iResult = -iResult;
    }
  else
    {
      ASSERT( ppvxTest[0] == NULL && ppvxTest[1] == NULL );
      iResult = APM2D::DistanceCompareL_VV(ppptTest[0][0](0), ppptTest[0][0](1),
					   ppptTest[0][1](0), ppptTest[0][1](1),
					   pppt[0][0](0), pppt[0][0](1),
					   pppt[1][0](0), pppt[1][0](1));
    }

  return iResult;
}

QEFacet *QEMesh::PointLocateVoronoiPolygonShewchukNaive(QEFacet *pfaTest)
{
  QEMeshIter iterQEMesh;
  QEFacet *pfa, *pfaMin;
  QEEdge *pe;
  QEVertex *ppvxEndpointTest[2];
  QEVertex *ppvxEndpoint[2];
  //Point pt;
  INT_32 iResult;

  iterQEMesh.Init(this);

  pfaMin = iterQEMesh.PeekFirstFacet();
  for ( pfa = iterQEMesh.PeekNextFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      iResult = DistanceCompareS_SS(pfaTest,
				    &ppvxEndpointTest[0],
				    &ppvxEndpointTest[1],
				    pfaMin, &ppvxEndpoint[0],
				    pfa, &ppvxEndpoint[1]);

      if ( ppvxEndpoint[0] != NULL && 
	   pfa->GetForwardVertex() == ppvxEndpoint[0] )
	{
	  ASSERT( iResult == 0 );
	  pfaMin = pfa;
	}
      else if ( iResult > 0 )
	{
	  ASSERT( !( ppvxEndpoint[1] != NULL &&
		     pfaMin->GetForwardVertex() == ppvxEndpoint[1] ) );
	  pfaMin = pfa;
	}
    }

  pe = pfaMin->GetForwardEdge();
  if ( pe != NULL )
    {
      /*
      if ( pfaTest->GetForwardVertex() != NULL )
	{
	  pt = pfaTest->GetForwardVertex()->GetPoint();
	}
      else
	{
	  ASSERT( pfaTest->GetForwardEdge() != NULL );

	  // We can use either endpoint
	  pt = pfaTest->GetForwardEdge()->GetVertex()->GetPoint();
	}

      if ( Counterclockwise(pe->GetEdge()->GetVertex()->GetPoint(),
			    pe->GetVertex()->GetPoint(),
			    pt) )
	{
	  pfaMin = pe->GetEdge()->GetFacet();
	}
      */

      if ( SideOfEdgeShewchuk(pe, pfaTest) < 0 )
	{
	  pfaMin = pe->GetEdge()->GetFacet();
	}
    }
  ASSERT( pfaMin != NULL );
  ASSERT( pfaMin != GetNullFacet() );

  return pfaMin;
}

QEFacet *QEMesh::PointLocateVoronoiPolygonShewchukGradientDescent(QEFacet *pfaTest, QEVertex *pvxTest)
{
  QEEdge *pe, *peColinear;
  QEVertex *ppvxEndpoint[2];
  //Point ppt[2];
  Point pt;
  APM2D::EdgeZoneType eEZT;

  ASSERT( pfaTest != NULL && pfaTest != GetNullFacet() );
  ASSERT( pvxTest != NULL );
  ASSERT( pfaTest->GetForwardVertex() == pvxTest );
  pt = pvxTest->GetPoint();

  pe = m_plistQEEdge->PeekFirst();
  if ( pe->GetFacet() == GetNullFacet() )
    {
      pe = pe->GetEdge();
    }
  ASSERT( pe->GetFacet() != GetNullFacet() );

  if ( pe->GetFacet()->GetForwardEdge() != NULL )
    {
      pe = pe->GetFacet()->GetForwardEdge();
      peColinear = NULL;
      while ( TRUE )
	{
	  eEZT = EdgeZone(pt, pe, &ppvxEndpoint[0]);
	  
	  if ( eEZT != APM2D::eztMiddle )
	    {
	      ASSERT( ppvxEndpoint[0] != NULL );
	      if ( eEZT == APM2D::eztEnd )
		{
		  pe = pe->GetEdge();
		}
	      ASSERT( pe->GetVertex() == ppvxEndpoint[0] );
	      
	      if ( pe->GetEdgeNext(1)->GetFacet()->GetForwardVertex() == ppvxEndpoint[0] )
		{
		  pe = pe->GetEdgeNext(1);
		  break;
		}
	      else if ( pe->GetEdgePrev(1)->GetEdgePrev(1)->GetFacet()->GetForwardVertex() == ppvxEndpoint[0] )
		{
		  pe = pe->GetEdgePrev(1)->GetEdgePrev(1);
		  break;
		}
	      else
		{
		  // These are two colinear edge segments
		  if ( peColinear == pe->GetEdgePrev(1)->GetEdgePrev(1) )
		    {
		      if ( Counterclockwise(pe->GetEdge()->GetVertex()->GetPoint(),
					    pe->GetVertex()->GetPoint(),
					    pt) )
			{
			  pe = pe->GetEdge();
			}
		      break;
		    }
		  peColinear = pe;
		  pe = pe->GetEdgePrev(1)->GetEdgePrev(1);
		  ASSERT( pe->GetFacet()->GetForwardEdge() == pe );
		}
	    }
	  else 
	    {
	      if ( Counterclockwise(pe->GetEdge()->GetVertex()->GetPoint(),
				    pe->GetVertex()->GetPoint(),
				    pt) )
		{
		  pe = pe->GetEdge();
		}
	      break;
	    }
	}
    }

  return PointLocateVoronoiPolygonShewchukGradientDescent(pfaTest, pe);
}

QEFacet *QEMesh::PointLocateVoronoiPolygonShewchukGradientDescent(QEFacet *pfaTest, 
								  QEEdge *peStart)
{
  QEEdge *pe, *peCurr, *peTemp, *peColinear;
  QEVertex *ppvxEndpointTest[2];
  QEVertex *ppvxEndpoint[2];
  //Point ppt[2];
  //Point pt;
  INT_32 iResult;
  BOOL bDone;

  /*
  if ( pfaTest->GetForwardVertex() != NULL )
    {
      pt = pfaTest->GetForwardVertex()->GetPoint();
    }
  else
    {
      ASSERT( pfaTest->GetForwardEdge() != NULL );
      // Either endpoint should be ok
      pt = pfaTest->GetForwardEdge()->GetVertex()->GetPoint();
    }
  */

  pe = peStart;

  peColinear = NULL;
  bDone = FALSE;
  while ( !bDone )
    {
      bDone = TRUE;
      peCurr = pe;
      do
	{
	  if ( peCurr->GetEdge()->GetFacet() != GetNullFacet() )
	    {
	      if ( peCurr->GetFacet()->GetForwardEdge() == peCurr )
		{
		  /*
		  if ( Counterclockwise(peCurr->GetEdge()->GetVertex()->GetPoint(),
					peCurr->GetVertex()->GetPoint(),
					pt) )
		  */
		  if ( SideOfEdgeShewchuk(peCurr, pfaTest) < 0 )
		    {
		      pe = peCurr->GetEdge();
		      bDone = FALSE;
		      break;
		    }
		}
	      else
		{
		  iResult = DistanceCompareS_SS(pfaTest, 
						&ppvxEndpointTest[0], &ppvxEndpointTest[1],
						peCurr->GetFacet(), &ppvxEndpoint[0],
						peCurr->GetEdge()->GetFacet(), &ppvxEndpoint[1]);
		  //ASSERT( ppvxEndpoint[0] == NULL );

		  if ( iResult > 0 ||
		       ( ppvxEndpoint[0] != NULL &&
			 ppvxEndpoint[0] == peCurr->GetEdge()->GetFacet()->GetForwardVertex() ) )
		    {
		      ASSERT( ppvxEndpoint[1] == NULL ||
			      peCurr->GetFacet()->GetForwardVertex() != ppvxEndpoint[1] );
		      peTemp = pe;
		      pe = peCurr->GetEdge();
		      if ( ppvxEndpoint[1] != NULL )
			{
			  ASSERT( pe->GetFacet()->GetForwardEdge() != NULL );
			  pe = pe->GetFacet()->GetForwardEdge();
			  if ( pe->GetVertex() != ppvxEndpoint[1] )
			    {
			      pe = pe->GetEdge();
			    }
			  ASSERT( pe->GetVertex() == ppvxEndpoint[1] );
			  
			  if ( pe->GetEdgeNext(1)->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] )
			    {
			      pe = pe->GetEdgeNext(1);
			    }
			  else if ( pe->GetEdgePrev(1)->GetEdgePrev(1)->GetFacet()->GetForwardVertex() == ppvxEndpoint[1] )
			    {
			      pe = pe->GetEdgePrev(1)->GetEdgePrev(1);
			    }
			  else
			    {
			      // These are two colinear edge segments
			      if ( peColinear == pe->GetEdgePrev(1)->GetEdgePrev(1) )
				{
				  pe = peTemp;
				  peCurr = peCurr->GetEdgeNext(0);
				  continue;
				}
			      peColinear = pe;
			      pe = pe->GetEdgePrev(1)->GetEdgePrev(1);
			      ASSERT( pe->GetFacet()->GetForwardEdge() == pe );
			    }
			}
		      else if ( pe->GetFacet()->GetForwardEdge() != NULL )
			{
			  peCurr = pe->GetFacet()->GetForwardEdge();
			  /*
			  if ( Counterclockwise(peCurr->GetEdge()->GetVertex()->GetPoint(),
						peCurr->GetVertex()->GetPoint(),
						pt) )
			  */
			  if ( SideOfEdgeShewchuk(peCurr, pfaTest) < 0 )
			    {
			      pe = peCurr->GetEdge();
			    }
			}
		      
		      bDone = FALSE;
		      break;
		    }
		}
	    }

	  peCurr = peCurr->GetEdgeNext(0);
	}
      while ( peCurr != pe );
    }

  return pe->GetFacet();
}

// Decides which side of edge pe is the site closer to
// result > 0  => Current side is closer
// result = 0  => Both sides are equivalent
// result < 0  => Opposite side is closer
INT_32 QEMesh::SideOfEdgeShewchuk(QEEdge *pe, QEFacet *pfaTest)
{
  INT_32 iResult;
  Point pptTest[2], ppt[2];
  APM2D::EdgeZoneType peEZT[2];
  INT_32 piTest[2], pi[2], iDistanceResult;
  UINT_32 i;

  iResult = 0;

  ppt[0] = pe->GetVertex()->GetPoint();
  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();

  if ( pfaTest->GetForwardVertex() != NULL )
    {
      pptTest[0] = pfaTest->GetForwardVertex()->GetPoint();
      iResult = APM2D::OrientationVVV(ppt[0](0), ppt[0](1),
				      ppt[1](0), ppt[1](1),
				      pptTest[0](0), pptTest[0](1));
    }
  else
    {
      ASSERT( pfaTest->GetForwardEdge() != NULL );

      pptTest[0] = pfaTest->GetForwardEdge()->GetVertex()->GetPoint();
      pptTest[1] = pfaTest->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint();

      for ( i = 0; i < 2; i++ )
	{
	  piTest[i] = APM2D::OrientationVVV(ppt[0](0), ppt[0](1),
					    ppt[1](0), ppt[1](1),
					    pptTest[i](0), pptTest[i](1));
	}

      if ( piTest[0] == piTest[1] )
	{
	  iResult = piTest[0];
	}
      else if ( piTest[0] >= 0 && piTest[1] >= 0 )
	{
	  iResult = 1;
	}
      else if ( piTest[0] <= 0 && piTest[1] <= 0 )
	{
	  iResult = -1;
	}
      else
	{
	  // The test edge straddles the edge
	  for ( i = 0; i < 2; i++ )
	    {
	      peEZT[i] = APM2D::EdgeZone(pptTest[i](0), pptTest[i](1),
					 ppt[0](0), ppt[0](1),
					 ppt[1](0), ppt[1](1));
	    }

	  if ( peEZT[0] == peEZT[1] )
	    {
	      switch ( peEZT[0] )
		{
		case APM2D::eztStart:
		  {
		    iDistanceResult = APM2D::DistanceCompareV_VV(ppt[0](0), ppt[0](1),
								 pptTest[0](0), pptTest[0](1),
								 pptTest[1](0), pptTest[1](1));
		    if ( iDistanceResult == 0 )
		      {
			// Perpendicular
			iResult = 0;
		      }
		    else
		      {
			iResult = ( iDistanceResult < 0 ) ? ( piTest[0] ) : ( piTest[1] );
		      }
		  }
		  break;
		case APM2D::eztEnd:
		  {
		    iDistanceResult = APM2D::DistanceCompareV_VV(ppt[1](0), ppt[1](1),
								 pptTest[0](0), pptTest[0](1),
								 pptTest[1](0), pptTest[1](1));
		    if ( iDistanceResult == 0 )
		      {
			// Perpendicular
			iResult = 0;
		      }
		    else
		      {
			iResult = ( iDistanceResult < 0 ) ? ( piTest[0] ) : ( piTest[1] );
		      }
		  }
		  break;
		case APM2D::eztMiddle:
		  // This can only happen if the edges intersect
		  // which violates the input assumption
		  iResult = 0;
		  ASSERT( FALSE );
		  break;
		default:
		  ASSERT( FALSE );
		  break;
		}
	    }
	  else if ( peEZT[0] == APM2D::eztMiddle )
	    {
	      iResult = piTest[0];
	    }
	  else if ( peEZT[1] == APM2D::eztMiddle )
	    {
	      iResult = piTest[1];
	    }
	  else
	    {
	      if ( peEZT[1] == APM2D::eztStart )
		{
		  ASSERT( peEZT[0] == APM2D::eztEnd );
		  for ( i = 0; i < 2; i++ )
		    {
		      pi[i] = APM2D::OrientationVVV(pptTest[0](0), pptTest[0](1),
						    pptTest[1](0), pptTest[1](1),
						    ppt[i](0), ppt[i](1));
		    }
		}
	      else
		{
		  ASSERT( peEZT[0] == APM2D::eztStart );
		  ASSERT( peEZT[1] == APM2D::eztEnd );
		  for ( i = 0; i < 2; i++ )
		    {
		      pi[i] = APM2D::OrientationVVV(pptTest[1](0), pptTest[1](1),
						    pptTest[0](0), pptTest[0](1),
						    ppt[i](0), ppt[i](1));
		    }
		}

	      if ( pi[0] == pi[1] )
		{
		  iResult = pi[0];
		}
	      else if ( pi[0] == 0 )
		{
		  iResult = pi[1];
		}
	      else if ( pi[1] == 0 )
		{
		  iResult = pi[0];
		}
	      else
		{
		  // This can only happen if the edges intersect
		  // which violates the input assumption
		  iResult = 0;
		  ASSERT( FALSE );
		}
	    }
	}
    }

  return iResult;
}

VOID QEMesh::MergeVoronoiPolygonsShewchuk(QEEdge *&peLeftEnd, QEEdge *&peLeft,
					  QEEdge *&peBisector, QEEdge *&peBisectorEnd,
					  QEEdge *&peRight, QEEdge *&peRightEnd, FLOAT fW)
{
  QEEdge *peBisectorPrev;
  QEEdge *peLeftStart, *peRightStart;
  QEEdge *peLeftNext, *peRightNext;
  QEEdge *pe;
  QEVertex *pvxNew, *pvxDummy;
  QEVertex *pvxLeft, *pvxRight;
  Point pt, ptCenter, ptDummy;
  Point ptLimit;
  Vector v;

  UINT_32 uiMergeDirection;
  Point ptLeft, ptRight;
  QEFacet *pfaLeft, *pfaRight;
  APM2D::InscribedCircleCenterType eLeftICCT, eRightICCT, eICCT;
  BOOL bReplacePoint;
  //BOOL bContinueSearch;

#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES

  //////////////////////////////////////////////////////////////////////
  // Merge

  if ( peBisectorEnd != NULL )
    {
      /// This is the second merge which must proceed out to infinity

      // Start the bisector
      peBisector = CreateEdge();
      peBisector->GetEdge()->SetVertex(peBisectorEnd->GetVertex());
      Splice(1, peBisector->GetEdge(), peBisectorEnd);
      peBisector->SetFacet(peBisector->GetEdgeNext(0)->GetFacet());
      peBisector->GetEdge()->SetFacet(peBisector->GetEdge()->GetEdgePrev(0)->GetFacet());

      ASSERT( peBisector->GetEdge()->GetFacet() == peLeft->GetFacet() );
      ASSERT( peBisector->GetFacet() == peRight->GetFacet() );

#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
    }

  peLeftStart = peLeft;
  peRightStart = peRight;

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  //fprintf(stderr, "    START: Merge Left = ( %.3f, %.3f ) Right = ( %.3f, %.3f )\n", peLeftMiddle->GetVertex()->GetPoint()(X), peLeftMiddle->GetVertex()->GetPoint()(Y), peRightMiddle->GetVertex()->GetPoint()(X), peRightMiddle->GetVertex()->GetPoint()(Y));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

  ////////////////////////////////////////
  // Merge Loop
  while ( TRUE )
    {
      uiMergeDirection = 0;
      pfaLeft = NULL;
      pfaRight = NULL;
      eLeftICCT = APM2D::icctNull;
      eRightICCT = APM2D::icctNull;

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
      fprintf(stderr, "      START: Searching for candidate point\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
      fprintf(stderr, "        START: LEFT SIDE\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

      // Left
      if ( peLeftEnd == NULL )
	{
	  //////////////////////////////////////////////////
	  // Search for left starting point
	  
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	  fprintf(stderr, "        Searching for minimum left starting point\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	  
	  ASSERT( peLeft->GetFacet() == peLeftStart->GetFacet() );
	  peLeft = peLeftStart;
	  
	  for ( peLeftNext = peLeftStart->GetEdgePrev(0);
		peLeftNext != peLeftStart && 
		  peLeftNext->GetEdge()->GetFacet() != GetNullFacet();
		peLeftNext = peLeftNext->GetEdgePrev(0) )
	    {
	      ASSERT( peLeftNext->GetEdge()->GetFacet() != GetNullFacet() );
		  
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	      fprintf(stderr, "        Calculating left start candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	      eICCT = InscribedCircleCenterSSSShewchuk(ptCenter, 
						       peLeftNext->GetEdge()->GetFacet(),
						       peLeft->GetFacet(),
						       peRight->GetFacet(), 
						       TRUE);
	      if ( eICCT != APM2D::icctNull )
		{
		  bReplacePoint = FALSE;
		  if ( peLeft == peLeftStart )
		    {
		      //////////////////////////////
		      // No left start candidate yet
		      
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
		      fprintf(stderr, "        Found first left start candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
		      bReplacePoint = TRUE;
		    }
		  else
		    {
		      //////////////////////////////
		      // Do an InCircle test with the current left point
		      
		      ASSERT( peLeft->GetFacet() != GetNullFacet() &&
			      peLeft->GetFacet()->GetForwardEdge() != peLeft );
		      
		      bReplacePoint = ReplaceCircleCenterShewchuk(eICCT, 
								  peLeftNext->GetEdge()->GetFacet(),
								  peLeft->GetFacet(),
								  peRight->GetFacet(), 
								  ptCenter,
								  peLeft->GetEdge()->GetFacet());
		      
		      if ( bReplacePoint )
			{
			  //////////////////////////////
			  // Replacing left candidate
			  
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			  fprintf(stderr, "        Updated left start candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			  bReplacePoint = TRUE;
			}
		    }
		      
		  if ( bReplacePoint )
		    {
		      ptLeft = ptCenter;
		      peLeft = peLeftNext;
		      pfaLeft = peLeft->GetEdge()->GetFacet();
		      eLeftICCT = eICCT;
		    }
		}
	    }
	}
      else
	{
	  ////////////////////////////////////////////////////////////
	  // START: Collapse edges from the left and right
	  //   until they are in the sector bounded by the 
	  //   two contour edges
	  
	  // Left
	  ASSERT( peLeft->GetEdge()->GetFacet() != GetNullFacet() ||
		  peLeft == peLeftEnd ||
		  peLeft == peLeft->GetFacet()->GetForwardEdge() );
	  while ( peLeft != peLeftEnd &&
		  peLeft != peLeft->GetFacet()->GetForwardEdge() &&
		  peLeft->GetEdgeNext(0) != peLeft->GetEdgeNext(0)->GetFacet()->GetForwardEdge() &&
		  peLeft->GetEdge()->GetVertex() != peLeft->GetFacet()->GetForwardVertex() )
	    
	    {
	      // Check if the Voronoi vertex is closer to the left or right edge
	      pvxLeft = peLeft->GetEdge()->GetVertex();
	      ASSERT( pvxLeft->GetEdge()->GetEdgeNext(1)->GetEdgeNext(1)->GetEdgeNext(1) ==
		      pvxLeft->GetEdge() );
	      ptCenter = pvxLeft->GetPoint();
	      spmeJordan = this;
	      if ( APM2D::cttOutCircle == InCircleSSS_SShewchuk(pvxLeft->GetInscribedCircleCenterType(),
								pvxLeft->GetEdge()->GetFacet(),
								pvxLeft->GetEdge()->GetEdgeNext(1)->GetFacet(),
								pvxLeft->GetEdge()->GetEdgePrev(1)->GetFacet(),
								ptCenter,
								peRight->GetFacet()) )
		{
		  break;
		}
	  
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	      fprintf(stderr, "        Left side skipping uniformed Voronoi edge\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	      // Advance the left next pointer
	      peLeftNext = peLeft->GetEdgeNext(0);
	      ASSERT( peLeft != peLeftEnd &&
		      peLeft != peLeft->GetFacet()->GetForwardEdge() );
	      CollapseEdge(peLeft);
	      peLeft = peLeftNext;

	      if ( peLeft->GetEdge()->GetFacet() == GetNullFacet() &&
		   peLeft != peLeftEnd &&
		   peLeft != peLeftNext->GetFacet()->GetForwardEdge() )
		{
		  peLeftNext = peLeft->GetEdgeNext(0);

		  pe = peLeft;
		  pe->GetEdgePrev(1)->SetVertex(pe->GetEdgePrev(1)->GetVertex());
		  pe->GetEdge()->GetEdgePrev(1)->SetVertex(pe->GetEdge()->GetEdgePrev(1)->GetVertex());
		  Splice(1, pe, pe->GetEdgePrev(1));
		  Splice(1, pe->GetEdge(), pe->GetEdge()->GetEdgePrev(1));
		  DestroyEdge(pe->GetEdge());
		  DestroyEdge(pe);

		  peLeft = peLeftNext;
		  peLeft->SetFacet(peLeft->GetFacet());
		}
	      
	      ASSERT( peLeft->GetEdge()->GetFacet() != GetNullFacet() ||
		      peLeft == peLeftEnd );

#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	      ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	    }
      
	  // END  : Collapse edges from the left and right
	  //   until they are in the sector bounded by the 
	  //   two contour edges
	  ////////////////////////////////////////////////////////////      

	  peLeftNext = peLeft;
	  
	  ////////////////////////////////////////
	  // Search for the best candidate
	  bReplacePoint = TRUE;
	  while ( bReplacePoint )
	    {
	      bReplacePoint = FALSE;
	      
	      ////////////////////////////////////////
	      // Search for Left candidate
	      
	      eICCT = APM2D::icctNull;
	      
	      // Do not go past a contour edge
	      if ( peLeftNext != peLeftEnd &&
		   peLeftNext != peLeftNext->GetFacet()->GetForwardEdge() &&
		   (peLeftNext->GetFacet()->GetForwardVertex() == NULL ||
		    peLeftNext->GetVertex() != peLeftNext->GetFacet()->GetForwardVertex()) &&
		   peLeftNext != peBisector->GetEdge() &&
		   peLeftNext->GetEdgeNext(0) != peBisector->GetEdge() )
		{
		  ASSERT( peLeftNext->GetEdge()->GetFacet() != GetNullFacet() );
		  
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
		  fprintf(stderr, "        Calculating left candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
		  
		  eICCT = InscribedCircleCenterSSSShewchuk(ptCenter, 
							   peLeftNext->GetEdge()->GetFacet(),
							   peLeft->GetFacet(),
							   peRight->GetFacet(), 
							   TRUE);
		  if ( eICCT != APM2D::icctNull )
		    {
		      //////////////////////////////
		      // Check if this candidate is closer than the current right point
		      bReplacePoint = FALSE;
		      if ( peLeft == peLeftNext )
			{
			  //////////////////////////////
			  // No left candidate yet
			  
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			  fprintf(stderr, "        Found first left candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			  bReplacePoint = TRUE;
			}
		      else
			{
			  //////////////////////////////
			  // Do an InCircle test with the current left point
			  
			  ASSERT( peLeft->GetFacet() != GetNullFacet() &&
				  peLeft->GetFacet()->GetForwardEdge() != peLeft );
			  
			  bReplacePoint = ReplaceCircleCenterShewchuk(eICCT, 
								      peLeftNext->GetEdge()->GetFacet(),
								      peLeft->GetFacet(),
								      peRight->GetFacet(), 
								      ptCenter,
								      peLeft->GetEdge()->GetFacet());
			  
			  if ( bReplacePoint )
			    {
			      //////////////////////////////
			      // Replacing left candidate
			      
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			      fprintf(stderr, "        Updated left candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			      bReplacePoint = TRUE;
			    }
			}
		      
		      if ( bReplacePoint ) 
			{
			  pfaLeft = peLeftNext->GetEdge()->GetFacet();
			  ptLeft = ptCenter;
			  eLeftICCT = eICCT;

			  if ( peLeft != peLeftNext )
			    {
			      //////////////////////////////
			      // Advance peLeft cleaning up extraneous edges
			      if ( peLeft->GetEdgeNext(0)->GetEdge()->GetFacet() == GetNullFacet() &&
				   peLeft->GetEdgeNext(0) != peLeftEnd )
				{
				  ASSERT( peLeft->GetEdgeNext(0) != peLeftEnd &&
					  peLeft->GetEdgeNext(0) != peLeft->GetEdgeNext(0)->GetFacet()->GetForwardEdge() );
				  CollapseEdge(peLeft->GetEdgeNext(0));
				}
			      ASSERT( peLeft != peLeftEnd &&
				      peLeft != peLeft->GetFacet()->GetForwardEdge() );
			      CollapseEdge(peLeft);
			      peLeft = peLeftNext;

#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
			      ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
			    }
			  
			  // Advance the left next pointer
			  peLeftNext = peLeft->GetEdgeNext(0);
			  if ( peLeftNext->GetEdge()->GetFacet() == GetNullFacet() &&
			       peLeftNext != peLeftEnd &&
			       peLeftNext != peLeftNext->GetFacet()->GetForwardEdge() &&
			       peLeftNext->GetEdge()->GetVertex() != peLeftNext->GetFacet()->GetForwardVertex() )
			    {
			      peLeftNext = peLeftNext->GetEdgeNext(0);
			    }
			}
		    }
		}
	    }
	}

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
      fprintf(stderr, "        END  : LEFT SIDE\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
      fprintf(stderr, "        START: RIGHT SIDE\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

      if ( peRightEnd == NULL )
	{
#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
	  fprintf(stderr, "        Searching for minimum right starting point\n");
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING

	  ASSERT( peRight->GetFacet() == peRightStart->GetFacet() );
	  peRight = peRightStart;

	  for ( peRightNext = peRightStart->GetEdgeNext(0);
		peRightNext != peRightStart && 
		  peRightNext->GetEdge()->GetFacet() != GetNullFacet();
		peRightNext = peRightNext->GetEdgeNext(0) )
	    {
	      ASSERT( peRightNext->GetEdge()->GetFacet() != GetNullFacet() );

#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
		  fprintf(stderr, "        Calculating right start candidate\n");
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING

	      eICCT = InscribedCircleCenterSSSShewchuk(ptCenter, 
						       peLeft->GetFacet(), 
						       peRight->GetFacet(),
						       peRightNext->GetEdge()->GetFacet(),
						       TRUE);
	      if ( eICCT != APM2D::icctNull )
		{
		  bReplacePoint = FALSE;
		  if ( peRight == peRightStart )
		    {
		      //////////////////////////////
		      // No right start candidate yet
			  
#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
		      fprintf(stderr, "        Found first right start candidate\n");
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING
		      bReplacePoint = TRUE;
		    }
		  else
		    {
		      //////////////////////////////
		      // Do an InCircle test with the current right point
		      
		      ASSERT( peRight->GetFacet() != GetNullFacet() &&
			      peRight->GetFacet()->GetForwardEdge() != peRight );

		      bReplacePoint = ReplaceCircleCenterShewchuk(eICCT,
								  peLeft->GetFacet(), 
								  peRight->GetFacet(),
								  peRightNext->GetEdge()->GetFacet(),
								  ptCenter,
								  peRight->GetEdge()->GetFacet());

		      if ( bReplacePoint )
			{
			  //////////////////////////////
			  // Replacing right candidate
			  
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			  fprintf(stderr, "        Updated right start candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			  bReplacePoint = TRUE;
			}
		    }
		  
		  if ( bReplacePoint )
		    {
		      ptRight = ptCenter;
		      peRight = peRightNext;
		      pfaRight = peRight->GetEdge()->GetFacet();
		      eRightICCT = eICCT;
		    }
		}
	    }
	}
      else
	{
	  ////////////////////////////////////////////////////////////
	  // START: Collapse edges from the left and right
	  //   until they are in the sector bounded by the 
	  //   two contour edges
	  
	  // Right
	  ASSERT( peRight->GetEdge()->GetFacet() != GetNullFacet() ||
		  peRight == peRightEnd ||
		  peRight == peRight->GetFacet()->GetForwardEdge() );
	  while ( peRight != peRightEnd &&
		  peRight != peRight->GetFacet()->GetForwardEdge() &&
		  peRight->GetEdgePrev(0) != peRight->GetEdgePrev(0)->GetFacet()->GetForwardEdge() &&
		  peRight->GetVertex() != peRight->GetFacet()->GetForwardVertex() )
	    {
	      // Check if the Voronoi vertex is closer to the left or right edge
	      pvxRight = peRight->GetVertex();
	      ASSERT( pvxRight->GetEdge()->GetEdgeNext(1)->GetEdgeNext(1)->GetEdgeNext(1) ==
		      pvxRight->GetEdge() );
	      ptCenter = pvxRight->GetPoint();
	      if ( APM2D::cttOutCircle == InCircleSSS_SShewchuk(pvxRight->GetInscribedCircleCenterType(),
								pvxRight->GetEdge()->GetFacet(),
								pvxRight->GetEdge()->GetEdgeNext(1)->GetFacet(),
								pvxRight->GetEdge()->GetEdgePrev(1)->GetFacet(),
								ptCenter,
								peLeft->GetFacet()) )
		{
		  break;
		}
	      
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	      fprintf(stderr, "        Right side skipping uniformed Voronoi edge\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	      
	      // Advance the right next pointer
	      peRightNext = peRight->GetEdgePrev(0);
	      ASSERT( peRight != peRightEnd &&
		      peRight != peRight->GetFacet()->GetForwardEdge() );
	      CollapseEdge(peRight->GetEdge());
	      peRight = peRightNext;

	      if ( peRight->GetEdge()->GetFacet() == GetNullFacet() &&
		   peRight != peRightEnd &&
		   peRight != peRightNext->GetFacet()->GetForwardEdge() )
		{
		  peRightNext = peRightNext->GetEdgePrev(0);

		  pe = peRight;
		  pe->GetEdgePrev(1)->SetVertex(pe->GetEdgePrev(1)->GetVertex());
		  pe->GetEdge()->GetEdgePrev(1)->SetVertex(pe->GetEdge()->GetEdgePrev(1)->GetVertex());
		  Splice(1, pe, pe->GetEdgePrev(1));
		  Splice(1, pe->GetEdge(), pe->GetEdge()->GetEdgePrev(1));
		  DestroyEdge(pe->GetEdge());
		  DestroyEdge(pe);

		  peRight = peRightNext;
		  peRight->SetFacet(peRight->GetFacet());
		}
	      
	      ASSERT( peRight->GetEdge()->GetFacet() != GetNullFacet() ||
		      peRight == peRightEnd );

#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	      ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	    }
	  
	  // END  : Collapse edges from the left and right
	  //   until they are in the sector bounded by the 
	  //   two contour edges
	  ////////////////////////////////////////////////////////////      

	  peRightNext = peRight;

	  ////////////////////////////////////////
	  // Search for Right candidate
	      
	  eICCT = APM2D::icctNull;

	  bReplacePoint = TRUE;
	  while ( bReplacePoint )
	    {
	      bReplacePoint = FALSE;

	      // Do not go past a contour edge
	      if ( peRightNext != peRightEnd &&
		   peRightNext != peRightNext->GetFacet()->GetForwardEdge() &&
		   (peRightNext->GetFacet()->GetForwardVertex() == NULL ||
		    peRightNext->GetEdge()->GetVertex() != peRightNext->GetFacet()->GetForwardVertex()) &&
		   peRightNext != peBisector &&
		   peRightNext->GetEdgePrev(0) != peBisector )
		{
		  ASSERT( peRightNext->GetEdge()->GetFacet() != GetNullFacet() );
		  
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
		  fprintf(stderr, "        Calculating right candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
		  eICCT = InscribedCircleCenterSSSShewchuk(ptCenter, 
							   peLeft->GetFacet(), 
							   peRight->GetFacet(),
							   peRightNext->GetEdge()->GetFacet(),
							   TRUE);
		  if ( eICCT != APM2D::icctNull )
		    {
		      //////////////////////////////
		      // Check if this candidate is closer than the current right point
		      bReplacePoint = FALSE;
		      if ( peRight == peRightNext )
			{
			  //////////////////////////////
			  // No right candidate yet
		      
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			  fprintf(stderr, "        Found first right candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			  bReplacePoint = TRUE;
			}
		      else
			{
			  //////////////////////////////
			  // Do an InCircle test with the curren right point
			  
			  ASSERT( peRight->GetFacet() != GetNullFacet() &&
				  peRight->GetFacet()->GetForwardEdge() != peRight );
			  
			  bReplacePoint = ReplaceCircleCenterShewchuk(eICCT,
								      peLeft->GetFacet(), 
								      peRight->GetFacet(),
								      peRightNext->GetEdge()->GetFacet(),
								      ptCenter,
								      peRight->GetEdge()->GetFacet());

			  if ( bReplacePoint )
			    {
			      //////////////////////////////
			      // Replacing right candidate
			      
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
			      fprintf(stderr, "        Updated right candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
			      bReplacePoint = TRUE;
			    }
			}

		      if ( bReplacePoint ) 
			{
			  pfaRight = peRightNext->GetEdge()->GetFacet();
			  ptRight = ptCenter;
			  eRightICCT = eICCT;
		      
			  if ( peRight != peRightNext )
			    {
			      //////////////////////////////
			      // Advance peRight cleaning up extraneous edges
			      if ( peRight->GetEdgePrev(0)->GetEdge()->GetFacet() == GetNullFacet() &&
				   peRight->GetEdgePrev(0) != peRightEnd )
				{
				  ASSERT( peRight->GetEdgePrev(0) != peRightEnd &&
					  peRight->GetEdgePrev(0) != peRight->GetEdgePrev(0)->GetFacet()->GetForwardEdge() );
				  CollapseEdge(peRight->GetEdgePrev(0)->GetEdge());
				}
			      ASSERT( peRight != peRightEnd &&
				      peRight != peRight->GetFacet()->GetForwardEdge() );
			      CollapseEdge(peRight->GetEdge());
			      peRight = peRightNext;
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	      ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
			    }
			  
			  // Advance the right next pointer
			  peRightNext = peRight->GetEdgePrev(0);
			  if ( peRightNext->GetEdge()->GetFacet() == GetNullFacet() &&
			       peRightNext != peRightEnd &&
			       peRightNext != peRightNext->GetFacet()->GetForwardEdge() &&
			       peRightNext->GetVertex() != peRightNext->GetFacet()->GetForwardVertex() )
			    {
			      peRightNext = peRightNext->GetEdgePrev(0);
			    }
			}
		    }
		}
	    }
	}

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
      fprintf(stderr, "        END  : RIGHT SIDE\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
      fprintf(stderr, "      END  : Searching for candidate point\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

      //////////////////////////////////////////////////
      // Decide which side to attach to
      uiMergeDirection = 0;
      if ( pfaLeft != NULL && pfaRight == NULL )
	{
	  uiMergeDirection = 1;
	}
      else if ( pfaLeft == NULL && pfaRight != NULL )
	{
	  uiMergeDirection = 2;
	}
      else if ( pfaLeft != NULL && pfaRight != NULL )
	{
	  BOOL bLeftInCircle, bRightInCircle;
	  INT_32 iResult;

	  //////////////////////////////
	  // Check Degenerate Touching
	  QEVertex *ppvx[2];
	  QEEdge *ppe[2];
	  Point ppt[2];
	  ppvx[0] = NULL;
	  ppvx[1] = NULL;
	  ppe[0] = NULL;
	  ppe[1] = NULL;
	  if ( peLeft->GetEdge()->GetFacet()->GetForwardVertex() != NULL )
	    {
	      ppvx[0] = peLeft->GetEdge()->GetFacet()->GetForwardVertex();
	    }
	  else
	    {
	      ppe[0] = peLeft->GetEdge()->GetFacet()->GetForwardEdge();
	    }
	  if ( peRight->GetEdge()->GetFacet()->GetForwardVertex() != NULL )
	    {
	      ppvx[1] = peRight->GetEdge()->GetFacet()->GetForwardVertex();
	    }
	  else
	    {
	      ppe[1] = peRight->GetEdge()->GetFacet()->GetForwardEdge();
	    }

	  if ( ppvx[0] != NULL && ppvx[1] != NULL )
	    {
	      ppt[0] = ppvx[0]->GetPoint();
	      ppt[1] = ppvx[1]->GetPoint();
	      if ( ppt[0](0) == ppt[1](0) && ppt[0](1) == ppt[1](1) )
		{
		  uiMergeDirection = 3;
		}
	    }
	  else if ( ppvx[0] != NULL && ppe[1] != NULL )
	    {
	      ppt[0] = ppvx[0]->GetPoint();
	      ppt[1] = ppe[1]->GetVertex()->GetPoint();
	      if ( ( ppt[0](0) == ppt[1](0) && ppt[0](1) == ppt[1](1) ) ||
		   ( ppt[0](0) == ppt[2](0) && ppt[0](1) == ppt[2](1) ) )
		{
		  uiMergeDirection = 4;
		}
	    }
	  else if ( ppe[0] != NULL && ppvx[1] != NULL )
	    {
	      ppt[0] = ppe[0]->GetEdge()->GetVertex()->GetPoint();
	      ppt[1] = ppvx[1]->GetPoint();
	      if ( ( ppt[0](0) == ppt[1](0) && ppt[0](1) == ppt[1](1) ) ||
		   ( ppt[0](0) == ppt[2](0) && ppt[0](1) == ppt[2](1) ) )
		{
		  uiMergeDirection = 5;
		}
	    }

	  if ( uiMergeDirection == 0 )
	    {
	      bLeftInCircle = !ReplaceCircleCenterShewchuk(eLeftICCT, 
							   pfaLeft,
							   peLeft->GetFacet(),
							   peRight->GetFacet(), 
							   ptLeft,
							   pfaRight);
	      bRightInCircle = !ReplaceCircleCenterShewchuk(eRightICCT, 
							    peLeft->GetFacet(),
							    peRight->GetFacet(), 
							    pfaRight,
							    ptRight,
							    pfaLeft);
	      if ( bRightInCircle == bLeftInCircle )
		{
		  if ( peBisector != NULL )
		    {
		      ASSERT( peBisector->GetEdge()->GetVertex() != NULL );
		      iResult = APM2D::DistanceCompareV_VV(peBisector->GetEdge()->GetVertex()->GetPoint()(0),
							   peBisector->GetEdge()->GetVertex()->GetPoint()(1),
							   ptLeft(0), ptLeft(1),
							   ptRight(0), ptRight(1));
		      if ( iResult < 0 ||
			   (iResult == 0 && peLeftEnd == NULL) )
			{
			  uiMergeDirection = 1;
			}
		      else
			{
			  uiMergeDirection = 2;
			}
		    }
		  else
		    {
		      if ( ptLeft(2) < ptRight(2) ||
			   (ptLeft(2) == ptRight(2) && peLeftEnd == NULL) )
			{
			  uiMergeDirection = 1;
			}
		      else
			{
			  uiMergeDirection = 2;
			}
		    }
		}
	      else if ( bRightInCircle )
		{
		  ASSERT( !bLeftInCircle );
		  uiMergeDirection = 1;
		} 
	      else if ( bLeftInCircle )
		{
		  ASSERT( !bRightInCircle );
		  uiMergeDirection = 2;
		}
	      else
		{
		  ASSERT( FALSE );
		}
	    }
	}
      else
	{
	  // We are done!!
	  break;
	}

      //////////////////////////////////////////////////
      // Check if we have a valid candidate
      if ( uiMergeDirection == 0 ) 
	{
	  break;
	} 
      else if ( uiMergeDirection == 1 ) 
	{
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	  fprintf(stderr, "      Attaching to left ( %.3f, %.3f, %.3f )\n", ptLeft(X), ptLeft(Y), ptLeft(Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	  if ( peLeftEnd == NULL )
	    {
#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
	      fprintf(stderr, "      Starting on left\n");
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING
	      
//#define LEFT_AND_RIGHT_STARTING_CODE_DIFFERENT
#ifdef LEFT_AND_RIGHT_STARTING_CODE_DIFFERENT
	      for ( peLeftEnd = peLeft->GetEdgePrev(0);
		    peLeftEnd != peLeft &&
		      peLeftEnd->GetEdge()->GetFacet() != GetNullFacet();
		    peLeftEnd = peLeftEnd->GetEdgePrev(0) )
		{
		}
	      
	      ASSERT( peLeftEnd->GetEdge()->GetFacet() == GetNullFacet() );
	      if ( peLeftEnd->GetEdge()->GetFacet() == GetNullFacet() )
		{
		  //////////////////////////////
		  // Detach the left end edge and contract the in between edges
		  for ( pe = peLeftEnd->GetEdgeNext(0);
			pe != peLeft;
			pe = peLeftEnd->GetEdgeNext(0) )
		    {
		      CollapseEdge(pe);
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		    }
		  peLeftEnd->GetEdge()->GetEdgePrev(1)->SetVertex(peLeftEnd->GetEdge()->GetEdgePrev(1)->GetVertex());
		  Splice(1, peLeftEnd->GetEdge(), peLeftEnd->GetEdge()->GetEdgePrev(1));
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		}
#endif // LEFT_AND_RIGHT_STARTING_CODE_DIFFERENT
	      
	      //////////////////////////////////////////////////
	      // Start the bisector to the left side
	      ptLeft[3] = fW;
	      
	      peLeftNext = peLeft->GetEdge()->GetEdgeNext(0);
	      peBisectorPrev = peBisector;
	      
	      pvxNew = CreateVertex(ptLeft);
	      ptLimit = ptLeft;
	      ptLimit[2] *= ptLimit[3];
	      ptLimit[3] = 1.0;
	      pvxNew->SetLimitPosition(ptLimit);
	      pvxNew->SetInscribedCircleCenterType(eLeftICCT);
	      peBisector = CreateEdge();
	      
	      peLeft->GetEdgePrev(1)->SetVertex(peLeft->GetEdgePrev(1)->GetVertex());
#ifdef LEFT_AND_RIGHT_STARTING_CODE_DIFFERENT
	      if ( peLeftEnd->GetEdge()->GetFacet() != GetNullFacet() )
		{
#endif // LEFT_AND_RIGHT_STARTING_CODE_DIFFERENT
		  ptDummy = (peLeft->GetVertex()->GetLimitPosition() + peLeft->GetEdge()->GetVertex()->GetLimitPosition())/ 2.0;
		  peLeftNext = CreateEdge();
		  peLeftNext->GetEdge()->SetFacet(peLeft->GetFacet());
		  peLeftNext->SetFacet(peLeft->GetEdge()->GetFacet());
		  peLeftNext->GetEdge()->SetVertex(peLeft->GetVertex());
		  Splice(1, peLeftNext->GetEdge(), peLeft);
		    
		  // Dummy vertex
		  //ptDummy.Null();
		  pvxDummy = CreateVertex(ptDummy);
		  pvxDummy->SetLimitPosition(ptDummy);
		  pvxDummy->SetInscribedCircleCenterType(APM2D::icctNull);
		  
		  peLeftNext->SetVertex(pvxDummy);
		  
		  peLeftEnd = CreateEdge();
		  peLeftEnd->SetFacet(peLeft->GetFacet());
		  peLeftEnd->GetEdge()->SetFacet(GetNullFacet());
		  peLeftEnd->SetVertex(pvxDummy);
		  Splice(1, peLeftNext, peLeftEnd);
		  
#ifdef LEFT_AND_RIGHT_STARTING_CODE_DIFFERENT
		}
#endif // LEFT_AND_RIGHT_STARTING_CODE_DIFFERENT
	      Splice(1, peLeft, peLeft->GetEdgePrev(1));
	      
	      peLeft->SetVertex(pvxNew);
	      peBisector->GetEdge()->SetVertex(pvxNew);
	      if ( peBisectorPrev != NULL )
		{
		  peBisectorPrev->SetVertex(pvxNew);
		  Splice(1, peBisectorPrev, peBisector->GetEdge());
		}
	      else
		{
		  peBisectorEnd = peLeft;
		}
	      Splice(1, peLeft, peBisector->GetEdge());
	      
	      // The vertex needs to point at the bisector edge for the ICCT
	      pvxNew->SetEdge(peBisector->GetEdge());
	      
	      peLeft = peLeftNext;
	      
	      peBisector->SetFacet(peRight->GetFacet());
	      peBisector->GetEdge()->SetFacet(peLeft->GetFacet());
	      
	      if ( peLeft->GetEdge()->GetFacet() == GetNullFacet() &&
		   peLeft->GetFacet()->GetForwardEdge() != peLeft && 
		   peLeft->GetFacet()->GetForwardVertex() != peLeft->GetEdge()->GetVertex() )
		{
		  peLeftNext = peLeft->GetEdgeNext(0);
		  CollapseEdge(peLeft);
		  peLeft = peLeftNext;
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		}
	    }
	  else
	    {
	      //////////////////////////////////////////////////
	      // Attach the bisector to the left side
	      ptLeft[3] = fW;

	      peLeftNext = peLeft->GetEdge()->GetEdgeNext(0);
	      pe = peBisector;
	      
	      pvxNew = CreateVertex(ptLeft);
	      ptLimit = ptLeft;
	      ptLimit[2] *= ptLimit[3];
	      ptLimit[3] = 1.0;
	      pvxNew->SetLimitPosition(ptLimit);
	      pvxNew->SetInscribedCircleCenterType(eLeftICCT);
	      peBisector = CreateEdge();

	      peLeft->GetEdgePrev(1)->SetVertex(peLeft->GetEdgePrev(1)->GetVertex());
	      Splice(1, peLeft, peLeft->GetEdgePrev(1));
	      
	      peLeft->SetVertex(pvxNew);
	      pe->SetVertex(pvxNew);
	      peBisector->GetEdge()->SetVertex(pvxNew);
	      // The vertex needs to point at the bisector edge for the ICCT
	      pvxNew->SetEdge(peBisector->GetEdge());

	      Splice(1, pe, peBisector->GetEdge());
	      Splice(1, peLeft, peBisector->GetEdge());
	      
	      peLeft = peLeftNext;
	      
	      peBisector->SetFacet(peRight->GetFacet());
	      peBisector->GetEdge()->SetFacet(peLeft->GetFacet());
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	      ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES

	      if ( peLeft->GetEdge()->GetFacet() == GetNullFacet() &&
		   peLeft != peLeftEnd &&
		   peLeft->GetFacet()->GetForwardEdge() != peLeft && 
		   peLeft->GetFacet()->GetForwardVertex() != peLeft->GetEdge()->GetVertex() )
		{
		  peLeftNext = peLeft->GetEdgeNext(0);
		  CollapseEdge(peLeft);
		  peLeft = peLeftNext;
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		  peBisector->GetEdge()->SetFacet(peBisector->GetEdge()->GetFacet());
		}
	    }
	}
      else if ( uiMergeDirection == 2 )
	{
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	  fprintf(stderr, "      Attaching to right ( %.3f, %.3f, %.3f )\n", ptRight(X), ptRight(Y), ptRight(Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	  if ( peRightEnd == NULL )
	    {
#ifdef ENABLE_VORONOI_POLYGONS_DEBUGGING
	      fprintf(stderr, "      Starting on right\n");
#endif // ENABLE_VORONOI_POLYGONS_DEBUGGING
	      
#ifdef LEFT_AND_RIGHT_STARTING_CODE_DIFFERENT
	      /*
		// I think this is wrong
		for ( peRightEnd = peRight->GetEdgePrev(0);
		      peRightEnd != peRight &&
		      peRightEnd->GetEdge()->GetFacet() != GetNullFacet();
		      peRightEnd = peRightEnd->GetEdgePrev(0) )
		      {
		      }
	      */
	      for ( peRightEnd = peRight->GetEdgeNext(0);
		      peRightEnd != peRight &&
		      peRightEnd->GetEdge()->GetFacet() != GetNullFacet();
		    peRightEnd = peRightEnd->GetEdgeNext(0) )
		{
		}
	      
	      if ( peRightEnd->GetEdge()->GetFacet() == GetNullFacet() )
		{
		  //////////////////////////////
		  // Detach the right end edge
		  for ( pe = peRightEnd->GetEdgePrev(0);
			pe != peRight;
			pe = peRightEnd->GetEdgePrev(0) )
		    {
		      CollapseEdge(pe->GetEdge());
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		    }
		  peRightEnd->GetEdgePrev(1)->SetVertex(peRightEnd->GetEdgePrev(1)->GetVertex());
		  Splice(1, peRightEnd, peRightEnd->GetEdgePrev(1));
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		}
#endif // LEFT_AND_RIGHT_STARTING_CODE_DIFFERENT

	      //////////////////////////////////////////////////
	      // Start the bisector to the left side
	      ptRight[3] = fW;
	      
	      peRightNext = peRight->GetEdge()->GetEdgePrev(0);
	      peBisectorPrev = peBisector;
	      
	      pvxNew = CreateVertex(ptRight);
	      ptLimit = ptRight;
	      ptLimit[2] *= ptLimit[3];
	      ptLimit[3] = 1.0;
	      pvxNew->SetLimitPosition(ptLimit);
	      pvxNew->SetInscribedCircleCenterType(eRightICCT);
	      peBisector = CreateEdge();
	      
	      peRight->GetEdge()->GetEdgePrev(1)->SetVertex(peRight->GetEdge()->GetEdgePrev(1)->GetVertex());
#ifdef LEFT_AND_RIGHT_STARTING_CODE_DIFFERENT
	      if ( peRightEnd->GetEdge()->GetFacet() != GetNullFacet() )
		{
#endif // LEFT_AND_RIGHT_STARTING_CODE_DIFFERENT
		  ptDummy = (peRight->GetVertex()->GetLimitPosition() + peRight->GetEdge()->GetVertex()->GetLimitPosition())/ 2.0;
		  peRightNext = CreateEdge();
		  peRightNext->GetEdge()->SetFacet(peRight->GetFacet());
		  peRightNext->SetFacet(peRight->GetEdge()->GetFacet());
		  peRightNext->SetVertex(peRight->GetEdge()->GetVertex());
		  Splice(1, peRightNext, peRight->GetEdge()->GetEdgePrev(1));
		  
		  // Dummy vertex
		  //ptDummy.Null();
		  pvxDummy = CreateVertex(ptDummy);
		  pvxDummy->SetLimitPosition(ptDummy);
		  pvxDummy->SetInscribedCircleCenterType(APM2D::icctNull);
		    
		  peRightNext->GetEdge()->SetVertex(pvxDummy);

		  peRightEnd = CreateEdge();
		  peRightEnd->SetFacet(peRight->GetFacet());
		  peRightEnd->GetEdge()->SetFacet(GetNullFacet());
		  peRightEnd->GetEdge()->SetVertex(pvxDummy);
		  Splice(1, peRightNext->GetEdge(), peRightEnd->GetEdge());
		  
#ifdef LEFT_AND_RIGHT_STARTING_CODE_DIFFERENT
		}
#endif // LEFT_AND_RIGHT_STARTING_CODE_DIFFERENT
	      Splice(1, peRight->GetEdge(), peRight->GetEdge()->GetEdgePrev(1));
	      
	      peRight->GetEdge()->SetVertex(pvxNew);
	      peBisector->GetEdge()->SetVertex(pvxNew);
	      Splice(1, peRight->GetEdge(), peBisector->GetEdge());
	      if ( peBisectorPrev != NULL )
		{
		  peBisectorPrev->SetVertex(pvxNew);
		  Splice(1, peBisectorPrev, peBisector->GetEdge());
		}
	      else
		{
		  peBisectorEnd = peBisector->GetEdge();
		}
		
	      // The vertex needs to point at the bisector edge for the ICCT
	      pvxNew->SetEdge(peBisector->GetEdge());
	      
	      peRight = peRightNext;
	      
	      peBisector->SetFacet(peRight->GetFacet());
	      peBisector->GetEdge()->SetFacet(peLeft->GetFacet());
	      
	      if ( peRight->GetEdge()->GetFacet() == GetNullFacet() &&
		   peRight->GetFacet()->GetForwardEdge() != peRight &&
		   peRight->GetFacet()->GetForwardVertex() != peRight->GetVertex() )
		{
		  peRightNext = peRight->GetEdgePrev(0);
		  CollapseEdge(peRight->GetEdge());
		  peRight = peRightNext;
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		}
	    }
	  else
	    {
	      //////////////////////////////////////////////////
	      // Attach the bisector to the right side
	      
	      ptRight[3] = fW;
	      
	      peRightNext = peRight->GetEdge()->GetEdgePrev(0);
	      pe = peBisector;
		
	      pvxNew = CreateVertex(ptRight);
	      ptLimit = ptRight;
	      ptLimit[2] *= ptLimit[3];
	      ptLimit[3] = 1.0;
	      pvxNew->SetLimitPosition(ptLimit);
	      pvxNew->SetInscribedCircleCenterType(eRightICCT);
	      peBisector = CreateEdge();
	      
	      peRight->GetEdge()->GetEdgePrev(1)->SetVertex(peRight->GetEdge()->GetEdgePrev(1)->GetVertex());
	      Splice(1, peRight->GetEdge(), peRight->GetEdge()->GetEdgePrev(1));
		
	      peRight->GetEdge()->SetVertex(pvxNew);
	      pe->SetVertex(pvxNew);
	      peBisector->GetEdge()->SetVertex(pvxNew);
	      // The vertex needs to point at the bisector edge for the ICCT
	      pvxNew->SetEdge(peBisector->GetEdge());
		
	      Splice(1, pe, peBisector->GetEdge());
	      Splice(1, peRight->GetEdge(), pe);
		
	      peRight = peRightNext;
		
	      peBisector->SetFacet(peRight->GetFacet());
	      peBisector->GetEdge()->SetFacet(peLeft->GetFacet());
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
      ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		
	      if ( peRight->GetEdge()->GetFacet() == GetNullFacet() &&
		   peRight != peRightEnd &&
		   peRight->GetFacet()->GetForwardEdge() != peRight &&
		   peRight->GetFacet()->GetForwardVertex() != peRight->GetVertex() )
		{
		  peRightNext = peRight->GetEdgePrev(0);
		  CollapseEdge(peRight->GetEdge());
		  peRight = peRightNext;
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		  peBisector->SetFacet(peBisector->GetFacet());
		}
	    }
	}
      else if ( uiMergeDirection == 3 ) 
	{
	  //////////////////////////////
	  // Two coincident vertex zones
	  INT_32 iResult;
	  ASSERT( peBisector != NULL );

	  pt = peLeft->GetEdge()->GetVertex()->GetPoint();
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	  fprintf(stderr, "    START: Fusing identical vertex zones ( %.3f, %.3f, %.3f )\n", pt(X), pt(Y), pt(Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	  peLeftNext = peLeft->GetEdge()->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge();
	  ASSERT( peLeftNext->GetFacet()->GetForwardEdge() == peLeftNext );
	  ptLeft = peLeftNext->GetVertex()->GetPoint();
	  peRightNext = peRight->GetEdgePrev(1)->GetEdgePrev(1);
	  ASSERT( peRightNext->GetFacet()->GetForwardEdge() == peRightNext );
	  ptRight = peRightNext->GetEdge()->GetVertex()->GetPoint();

	  iResult = APM2D::OrientationVVV(ptLeft(0), ptLeft(1),
					  pt(0), pt(1),
					  ptRight(0), ptRight(1));

	  DestroyVertex(peRight->GetVertex());
	  for ( peRightNext = peRight->GetEdgeNext(1);
		peRightNext != peRight;
		peRightNext = peRightNext->GetEdgeNext(1) )
	    {
	      peRightNext->SetVertex(peLeft->GetEdge()->GetVertex());
	    }
	  DestroyFacet(peRight->GetEdge()->GetFacet());
	  for ( peRightNext = peRight->GetEdge()->GetEdgeNext(0);
		peRightNext != peRight->GetEdge();
		peRightNext = peRightNext->GetEdgeNext(0) )
	    {
	      peRightNext->SetFacet(peLeft->GetEdge()->GetFacet());
	    }

	  peBisector->SetVertex(peLeft->GetEdge()->GetVertex());
	  Splice(1, peBisector, peLeft->GetEdge()->GetEdgePrev(1));
	  peBisector = peLeft->GetEdge()->GetEdgePrev(0);
	  Splice(1, peLeft->GetEdge(), peRight->GetEdgePrev(1));

	  peLeftNext = peLeft->GetEdgePrev(1);
	  Splice(1, peLeft->GetEdge(), peLeft->GetEdge()->GetEdgePrev(1));
	  Splice(1, peLeft, peLeft->GetEdgePrev(1));
	  DestroyEdge(peLeft->GetEdge());
	  DestroyEdge(peLeft);

	  peRightNext = peRight->GetEdge()->GetEdgePrev(0);
	  Splice(1, peRight->GetEdge(), peRight->GetEdge()->GetEdgePrev(1));
	  Splice(1, peRight, peRight->GetEdgePrev(1));
	  DestroyEdge(peRight->GetEdge());
	  DestroyEdge(peRight);

	  if ( iResult >= 0 )
	    {
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	  fprintf(stderr, "      Vertex zone annihilated\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	      peLeft = peLeftNext;
	      for ( peLeftNext = peLeft->GetEdgeNext(0);
		    peLeft != peBisector;
		    peLeftNext = peLeft->GetEdgeNext(0) )
		{
		  CollapseEdge(peLeft);
		  peLeft = peLeftNext;
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		}
	      peRight = peRightNext;
	      for ( peRightNext = peRight->GetEdgePrev(0);
		    peRightNext != peBisector;
		    peRightNext = peRight->GetEdgePrev(0) )
		{
		  CollapseEdge(peRight->GetEdge());
		  peRight = peRightNext;
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		}

	      DestroyFacet(peBisector->GetFacet());
	      peBisector->SetFacet(peRight->GetEdge()->GetFacet());

	      peLeft = peBisector->GetEdgePrev(1);
	      peLeft->SetVertex(peLeft->GetVertex());
	      Splice(1, peBisector, peBisector->GetEdgePrev(1));
	      
	      peRightNext = peRight->GetEdge()->GetEdgePrev(0);
	      peRight->GetEdgePrev(1)->SetVertex(peRight->GetEdgePrev(1)->GetVertex());
	      peRight->GetEdge()->GetEdgePrev(1)->SetVertex(peRight->GetEdge()->GetEdgePrev(1)->GetVertex());
	      Splice(1, peRight, peRight->GetEdgePrev(1));
	      Splice(1, peRight->GetEdge(), peRight->GetEdge()->GetEdgePrev(1));
	      DestroyEdge(peRight->GetEdge());
	      DestroyEdge(peRight);
	      peRight = peRightNext;
	    }
	  else
	    {
	      // Search for the next circumcenter
	      peBisector = NULL;
	      peLeft = peLeftNext;
	      peRight = peRightNext;

	      if ( peLeft->GetEdge()->GetFacet() == GetNullFacet() &&
		   peLeft != peLeftEnd &&
		   peLeft->GetFacet()->GetForwardEdge() != peLeft && 
		   peLeft->GetFacet()->GetForwardVertex() != peLeft->GetEdge()->GetVertex() )
		{
		  peLeftNext = peLeft->GetEdgeNext(0);
		  CollapseEdge(peLeft);
		  peLeft = peLeftNext;
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		}

	      if ( peRight->GetEdge()->GetFacet() == GetNullFacet() &&
		   peRight != peRightEnd &&
		   peRight->GetFacet()->GetForwardEdge() != peRight &&
		   peRight->GetFacet()->GetForwardVertex() != peRight->GetVertex() )
		{
		  peRightNext = peRight->GetEdgePrev(0);
		  CollapseEdge(peRight->GetEdge());
		  peRight = peRightNext;
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		}

	      ////////////////////////////////////////////////////////////
	      // START: Collapse edges from the left and right
	      //   until they are in the sector bounded by the 
	      //   two contour edges
		  
	      ASSERT( peLeft->GetEdge()->GetFacet() != GetNullFacet() ||
		      peLeft == peLeftEnd ||
		      peLeft == peLeft->GetFacet()->GetForwardEdge() );
	      ASSERT( peRight->GetEdge()->GetFacet() != GetNullFacet() ||
		      peRight == peRightEnd ||
		      peRight == peRight->GetFacet()->GetForwardEdge() );
	      BOOL bCollapsing;
	      bCollapsing = TRUE;
	      while ( bCollapsing )
		{
		  bCollapsing = FALSE;
		  while ( peLeft != peLeftEnd &&
			  peLeft != peLeft->GetFacet()->GetForwardEdge() &&
			  peLeft->GetEdgeNext(0) != peLeft->GetEdgeNext(0)->GetFacet()->GetForwardEdge() &&
			  peLeft->GetEdge()->GetVertex() != peLeft->GetFacet()->GetForwardVertex() )
			
		    {
		      // Check if the Voronoi vertex is closer to the left or right edge
		      pvxLeft = peLeft->GetEdge()->GetVertex();
		      ASSERT( pvxLeft->GetEdge()->GetEdgeNext(1)->GetEdgeNext(1)->GetEdgeNext(1) ==
			      pvxLeft->GetEdge() );
		      ptCenter = pvxLeft->GetPoint();
		      if ( APM2D::cttInCircle != InCircleSSS_SShewchuk(pvxLeft->GetInscribedCircleCenterType(),
								       pvxLeft->GetEdge()->GetFacet(),
								       pvxLeft->GetEdge()->GetEdgeNext(1)->GetFacet(),
								       pvxLeft->GetEdge()->GetEdgePrev(1)->GetFacet(),
								       ptCenter,
								       peRight->GetEdge()->GetFacet()) )
			{
			  break;
			}

		      bCollapsing = TRUE;
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
		      fprintf(stderr, "        Left side skipping uniformed Voronoi edge\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

		      // Advance the left next pointer
		      peLeftNext = peLeft->GetEdgeNext(0);
		      ASSERT( peLeft != peLeftEnd &&
			      peLeft != peLeft->GetFacet()->GetForwardEdge() );
		      CollapseEdge(peLeft);
		      peLeft = peLeftNext;
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		      ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
			  
		      if ( peLeft->GetEdge()->GetFacet() == GetNullFacet() &&
			   peLeft != peLeftEnd &&
			   peLeft != peLeftNext->GetFacet()->GetForwardEdge() )
			{
			  peLeftNext = peLeft->GetEdgeNext(0);
			      
			  pe = peLeft;
			  pe->GetEdgePrev(1)->SetVertex(pe->GetEdgePrev(1)->GetVertex());
			  pe->GetEdge()->GetEdgePrev(1)->SetVertex(pe->GetEdge()->GetEdgePrev(1)->GetVertex());
			  Splice(1, pe, pe->GetEdgePrev(1));
			  Splice(1, pe->GetEdge(), pe->GetEdge()->GetEdgePrev(1));
			  DestroyEdge(pe->GetEdge());
			  DestroyEdge(pe);

			  peLeft = peLeftNext;
			  peLeft->SetFacet(peLeft->GetFacet());
			}
			  
		      ASSERT( peLeft->GetEdge()->GetFacet() != GetNullFacet() ||
			      peLeft == peLeftEnd );
		    }
		}
      
	      while ( peRight != peRightEnd &&
		      peRight != peRight->GetFacet()->GetForwardEdge() &&
		      peRight->GetEdgePrev(0) != peRight->GetEdgePrev(0)->GetFacet()->GetForwardEdge() &&
		      peRight->GetVertex() != peRight->GetFacet()->GetForwardVertex() )
		{
		  // Check if the Voronoi vertex is closer to the left or right edge
		  pvxRight = peRight->GetVertex();
		  ASSERT( pvxRight->GetEdge()->GetEdgeNext(1)->GetEdgeNext(1)->GetEdgeNext(1) ==
			  pvxRight->GetEdge() );
		  ptCenter = pvxRight->GetPoint();
		  if ( APM2D::cttInCircle != InCircleSSS_SShewchuk(pvxRight->GetInscribedCircleCenterType(),
								   pvxRight->GetEdge()->GetFacet(),
								   pvxRight->GetEdge()->GetEdgeNext(1)->GetFacet(),
								   pvxRight->GetEdge()->GetEdgePrev(1)->GetFacet(),
								   ptCenter,
								   peLeft->GetEdge()->GetFacet()) )
		    {
		      break;
		    }

		  bCollapsing = TRUE;
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
		  fprintf(stderr, "        Right side skipping uniformed Voronoi edge\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	      
		  // Advance the right next pointer
		  peRightNext = peRight->GetEdgePrev(0);
		  ASSERT( peRight != peRightEnd &&
			  peRight != peRight->GetFacet()->GetForwardEdge() );
		  CollapseEdge(peRight->GetEdge());
		  peRight = peRightNext;
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES

		  if ( peRight->GetEdge()->GetFacet() == GetNullFacet() &&
		       peRight != peRightEnd &&
		       peRight != peRightNext->GetFacet()->GetForwardEdge() )
		    {
		      peRightNext = peRightNext->GetEdgePrev(0);

		      pe = peRight;
		      pe->GetEdgePrev(1)->SetVertex(pe->GetEdgePrev(1)->GetVertex());
		      pe->GetEdge()->GetEdgePrev(1)->SetVertex(pe->GetEdge()->GetEdgePrev(1)->GetVertex());
		      Splice(1, pe, pe->GetEdgePrev(1));
		      Splice(1, pe->GetEdge(), pe->GetEdge()->GetEdgePrev(1));
		      DestroyEdge(pe->GetEdge());
		      DestroyEdge(pe);

		      peRight = peRightNext;
		      peRight->SetFacet(peRight->GetFacet());
		    }
	      
		  ASSERT( peRight->GetEdge()->GetFacet() != GetNullFacet() ||
			  peRight == peRightEnd );
		}

	      // END  : Collapse edges from the left and right
	      //   until they are in the sector bounded by the 
	      //   two contour edges
	      ////////////////////////////////////////////////////////////      

	      ////////////////////////////////////////////////////////////
	      // START: Check if a circle center can be formed

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	      fprintf(stderr, "        Calculating vertex zone candidate\n");
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	      if ( peLeft->GetEdgeNext(0) == peRight )
		{
		  // Merge is done
		  ASSERT( peRight->GetEdgeNext(0)->GetFacet() == GetNullFacet() );
		  ASSERT( peLeft->GetEdgePrev(0)->GetFacet() == GetNullFacet() );
		  peLeftNext = CreateEdge();
		  peLeftNext->SetFacet(peLeft->GetFacet());
		  peLeftNext->SetVertex(peRight->GetEdge()->GetVertex());
		  peLeftNext->GetEdge()->SetVertex(peLeft->GetVertex());
		  Splice(1, peLeftNext, peRight->GetEdge()->GetEdgePrev(1));
		  Splice(1, peLeftNext->GetEdge(), peLeft);
		  peBisector = NULL;
		  return;
		}

	      eICCT = InscribedCircleCenterSSSShewchuk(ptCenter, 
						       peLeft->GetEdge()->GetFacet(), 
						       peRight->GetFacet(),
						       peRight->GetEdge()->GetFacet(),
						       TRUE);
	      if ( eICCT == APM2D::icctNull )
		{
		  // Merge is done
		  ASSERT( peRight->GetEdgeNext(0)->GetFacet() == GetNullFacet() );
		  ASSERT( peLeft->GetEdgePrev(0)->GetFacet() == GetNullFacet() );
		  peLeftNext = CreateEdge();
		  peLeftNext->SetFacet(peLeft->GetFacet());
		  peLeftNext->SetVertex(peRight->GetEdge()->GetVertex());
		  peLeftNext->GetEdge()->SetVertex(peLeft->GetVertex());
		  Splice(1, peLeftNext, peRight->GetEdge()->GetEdgePrev(1));
		  Splice(1, peLeftNext->GetEdge(), peLeft);
		  peBisector = NULL;
		  return;
		}
	      
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	      fprintf(stderr, "    Vertex zone circle center ( %.3f, %.3f, %.3f )\n", ptCenter(X), ptCenter(Y), ptCenter(Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	      //////////////////////////////
	      // Start a new bisector
	      ptCenter[3] = fW;
	      
	      peLeftNext = peLeft->GetEdge()->GetEdgeNext(0);
	      peRightNext = peRight->GetEdge()->GetEdgePrev(0);

	      pvxNew = CreateVertex(ptCenter);
	      ptLimit = ptCenter;
	      ptLimit[2] *= ptLimit[3];
	      ptLimit[3] = 1.0;
	      pvxNew->SetLimitPosition(ptLimit);
	      pvxNew->SetInscribedCircleCenterType(eICCT);
	      peBisector = CreateEdge();

	      Splice(1, peLeft, peLeft->GetEdgePrev(1));
	      Splice(1, peRight->GetEdge(), peRight->GetEdge()->GetEdgePrev(1));
	      
	      peLeft->SetVertex(pvxNew);
	      peRight->GetEdge()->SetVertex(pvxNew);
	      peBisector->GetEdge()->SetVertex(pvxNew);

	      Splice(1, peLeft, peBisector->GetEdge());
	      Splice(1, peLeft, peRight->GetEdge());
	      
	      // The vertex needs to point at the bisector edge for the ICCT
	      pvxNew->SetEdge(peBisector->GetEdge());

	      peLeft = peLeftNext;
	      peRight = peRightNext;
	      
	      peBisector->SetFacet(peRight->GetFacet());
	      peBisector->GetEdge()->SetFacet(peLeft->GetFacet());

	      // END  : Check if a circle center can be formed
	      ////////////////////////////////////////////////////////////
	    }

	  if ( peLeft->GetEdge()->GetFacet() == GetNullFacet() &&
	       peLeft != peLeftEnd &&
	       peLeft->GetFacet()->GetForwardEdge() != peLeft && 
	       peLeft->GetFacet()->GetForwardVertex() != peLeft->GetEdge()->GetVertex() )
	    {
	      peLeftNext = peLeft->GetEdgeNext(0);
	      CollapseEdge(peLeft);
	      peLeft = peLeftNext;
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	      ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	    }

	  if ( peRight->GetEdge()->GetFacet() == GetNullFacet() &&
	       peRight != peRightEnd &&
	       peRight->GetFacet()->GetForwardEdge() != peRight &&
	       peRight->GetFacet()->GetForwardVertex() != peRight->GetVertex() )
	    {
	      peRightNext = peRight->GetEdgePrev(0);
	      CollapseEdge(peRight->GetEdge());
	      peRight = peRightNext;
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
		  ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	    }
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	  fprintf(stderr, "    END  : Fusing identical vertex zones ( %.3f, %.3f, %.3f )\n", pt(X), pt(Y), pt(Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	}
      else if ( uiMergeDirection == 4 ) 
	{
	  //////////////////////////////
	  // Coincident left vertex zone with right convex edges
	  ASSERT( peBisector != NULL );

	  pt = peLeft->GetEdge()->GetVertex()->GetPoint();

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	  fprintf(stderr, "    START: Fusing left side vertex zone into right concave edges ( %.3f, %.3f, %.3f )\n", pt(X), pt(Y), pt(Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	  DestroyVertex(peLeft->GetEdge()->GetVertex());
	  for ( peLeftNext = peLeft->GetEdge()->GetEdgeNext(1);
		peLeftNext != peLeft->GetEdge();
		peLeftNext = peLeftNext->GetEdgeNext(1) )
	    {
	      peLeftNext->SetVertex(peRight->GetVertex());
	    }
	  DestroyFacet(peLeft->GetEdge()->GetFacet());

	  peBisector->SetVertex(peRight->GetVertex());
	  Splice(1, peBisector, peRight);
	  Splice(1, peRight, peLeft->GetEdge()->GetEdgePrev(1));
	  peBisector = peRight->GetEdge();

	  peLeftNext = peLeft->GetEdgePrev(1);
	  Splice(1, peLeft->GetEdge(), peLeft->GetEdge()->GetEdgePrev(1));
	  Splice(1, peLeft, peLeft->GetEdgePrev(1));
	  DestroyEdge(peLeft->GetEdge());
	  DestroyEdge(peLeft);

	  peLeft = peLeftNext;
	  for ( peLeftNext = peLeft->GetEdgeNext(0);
		peLeftNext != peBisector->GetEdge();
		peLeftNext = peLeft->GetEdgeNext(0) )
	    {
	      CollapseEdge(peLeft);
	      peLeft = peLeftNext;
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	      ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	    }

	  peBisector->GetEdge()->SetFacet(peLeft->GetEdge()->GetFacet());

	  peRight = peBisector->GetEdgePrev(0);
	  peRight->GetEdge()->SetVertex(peRight->GetEdge()->GetVertex());
	  Splice(1, peBisector, peBisector->GetEdgePrev(1));
	  
	  peLeftNext = peLeft->GetEdgePrev(1);
	  peLeft->GetEdgePrev(1)->SetVertex(peLeft->GetEdgePrev(1)->GetVertex());
	  peLeft->GetEdge()->GetEdgePrev(1)->SetVertex(peLeft->GetEdge()->GetEdgePrev(1)->GetVertex());
	  Splice(1, peLeft, peLeft->GetEdgePrev(1));
	  Splice(1, peLeft->GetEdge(), peLeft->GetEdge()->GetEdgePrev(1));
	  DestroyEdge(peLeft->GetEdge());
	  DestroyEdge(peLeft);
	  peLeft = peLeftNext;

	  if ( peLeft->GetEdge()->GetFacet() == GetNullFacet() &&
	       peLeft != peLeftEnd &&
	       peLeft->GetFacet()->GetForwardEdge() != peLeft && 
	       peLeft->GetFacet()->GetForwardVertex() != peLeft->GetEdge()->GetVertex() )
	    {
	      peLeftNext = peLeft->GetEdgeNext(0);
	      CollapseEdge(peLeft);
	      peLeft = peLeftNext;
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	      ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	    }

	  if ( peRight->GetEdge()->GetFacet() == GetNullFacet() &&
	       peRight != peRightEnd &&
	       peRight->GetFacet()->GetForwardEdge() != peRight &&
	       peRight->GetFacet()->GetForwardVertex() != peRight->GetVertex() )
	    {
	      peRightNext = peRight->GetEdgePrev(0);
	      CollapseEdge(peRight->GetEdge());
	      peRight = peRightNext;
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	      ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	    }
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	  fprintf(stderr, "    END  : Fusing left side vertex zone into right concave edges ( %.3f, %.3f, %.3f )\n", pt(X), pt(Y), pt(Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	}
      else if ( uiMergeDirection == 5 ) 
	{
	  //////////////////////////////
	  // Coincident right vertex zone with left convex edges
	  ASSERT( peBisector != NULL );

	  pt = peRight->GetVertex()->GetPoint();

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	  fprintf(stderr, "    START: Fusing right side vertex zone into left concave edges ( %.3f, %.3f, %.3f )\n", pt(X), pt(Y), pt(Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING

	  DestroyVertex(peRight->GetVertex());
	  for ( peRightNext = peRight->GetEdgeNext(1);
		peRightNext != peRight;
		peRightNext = peRightNext->GetEdgeNext(1) )
	    {
	      peRightNext->SetVertex(peLeft->GetEdge()->GetVertex());
	    }
	  DestroyFacet(peRight->GetEdge()->GetFacet());

	  peBisector->SetVertex(peLeft->GetEdge()->GetVertex());
	  Splice(1, peBisector, peLeft->GetEdge()->GetEdgePrev(1));
	  Splice(1, peBisector, peRight->GetEdgePrev(1));
	  peBisector = peLeft;

	  peRightNext = peRight->GetEdge()->GetEdgePrev(0);
	  Splice(1, peRight->GetEdge(), peRight->GetEdge()->GetEdgePrev(1));
	  Splice(1, peRight, peRight->GetEdgePrev(1));
	  DestroyEdge(peRight->GetEdge());
	  DestroyEdge(peRight);

	  peRight = peRightNext;
	  for ( peRightNext = peRight->GetEdgePrev(0);
		peRightNext != peBisector;
		peRightNext = peRight->GetEdgePrev(0) )
	    {
	      CollapseEdge(peRight->GetEdge());
	      peRight = peRightNext;
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	      ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	    }

	  peBisector->SetFacet(peRight->GetEdge()->GetFacet());

	  peLeft = peBisector->GetEdgePrev(1);
	  peLeft->SetVertex(peLeft->GetVertex());
	  Splice(1, peBisector, peBisector->GetEdgePrev(1));
	  
	  peRightNext = peRight->GetEdge()->GetEdgePrev(0);
	  peRight->GetEdgePrev(1)->SetVertex(peRight->GetEdgePrev(1)->GetVertex());
	  peRight->GetEdge()->GetEdgePrev(1)->SetVertex(peRight->GetEdge()->GetEdgePrev(1)->GetVertex());
	  Splice(1, peRight, peRight->GetEdgePrev(1));
	  Splice(1, peRight->GetEdge(), peRight->GetEdge()->GetEdgePrev(1));
	  DestroyEdge(peRight->GetEdge());
	  DestroyEdge(peRight);
	  peRight = peRightNext;

	  if ( peLeft->GetEdge()->GetFacet() == GetNullFacet() &&
	       peLeft != peLeftEnd &&
	       peLeft->GetFacet()->GetForwardEdge() != peLeft && 
	       peLeft->GetFacet()->GetForwardVertex() != peLeft->GetEdge()->GetVertex() )
	    {
	      peLeftNext = peLeft->GetEdgeNext(0);
	      CollapseEdge(peLeft);
	      peLeft = peLeftNext;
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	      ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	    }

	  if ( peRight->GetEdge()->GetFacet() == GetNullFacet() &&
	       peRight != peRightEnd &&
	       peRight->GetFacet()->GetForwardEdge() != peRight &&
	       peRight->GetFacet()->GetForwardVertex() != peRight->GetVertex() )
	    {
	      peRightNext = peRight->GetEdgePrev(0);
	      CollapseEdge(peRight->GetEdge());
	      peRight = peRightNext;
#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	      ::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
	    }
#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
	  fprintf(stderr, "    END  : Fusing right side vertex zone into left concave edges ( %.3f, %.3f, %.3f )\n", pt(X), pt(Y), pt(Z));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
	}
      else
	{
	  ASSERT( FALSE );
	}

#ifdef ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES
      //::OutputDebugVoronoiMesh(this);
#endif // ENABLE_VORONOI_POLYGON_MOVIE_OUTPUT_FILES

      peLeftNext = peLeft;
      peRightNext = peRight;
    }
      
  if ( peLeft != peLeftEnd || peRight != peRightEnd )
    {
      //ASSERT( peLeft != peLeftEnd && peRight != peRightEnd );
      if ( peLeftEnd == NULL )
	{
	  for ( peLeftEnd = peLeft->GetEdgePrev(0);
		peLeftEnd != peLeft;
		peLeftEnd = peLeftEnd->GetEdgePrev(0) )
	    {
	      if ( peLeftEnd->GetEdge()->GetFacet() == GetNullFacet() )
		{
		  break;
		}
	    }
	  ASSERT( peLeftEnd->GetEdge()->GetFacet() == GetNullFacet() );
	  peLeft = peLeftEnd->GetEdgeNext(0);
	  peLeft->SetVertex(peLeft->GetVertex());
	  peLeft->SetFacet(peLeft->GetFacet());
	  Splice(1, peLeftEnd->GetEdge(), peLeftEnd->GetEdge()->GetEdgePrev(1));
	}

      if ( peRightEnd == NULL )
	{
	  for ( peRightEnd = peRight->GetEdgeNext(0);
		peRightEnd != peRight;
		peRightEnd = peRightEnd->GetEdgeNext(0) )
	    {
	      if ( peRightEnd->GetEdge()->GetFacet() == GetNullFacet() )
		{
		  break;
		}
	    }
	  ASSERT( peRightEnd->GetEdge()->GetFacet() == GetNullFacet() );
	  peRight = peRightEnd->GetEdgePrev(0);
	  peRight->GetEdge()->SetVertex(peRight->GetEdge()->GetVertex());
	  peRight->SetFacet(peRight->GetFacet());
	  Splice(1, peRightEnd, peRightEnd->GetEdgePrev(1));
	}

      v = CalculateBisectorDirection(peLeft->GetFacet(), TRUE, 
				     peRight->GetFacet(), TRUE, FALSE);
      v[2] = fW;
      pt = v;
      pt[3] = 0.0;
      pvxNew = CreateVertex(pt);
      if ( peBisector != NULL )
	{
	  ptLimit = peBisector->GetEdge()->GetVertex()->GetLimitPosition() + cfInfinityMagnitude*v;
	  ptLimit[3] = 1.0;
	}
      else
	{
	  ptLimit = peLeft->GetVertex()->GetLimitPosition() + cfInfinityMagnitude*v;
	  ptLimit[3] = 1.0;
	}
      pvxNew->SetLimitPosition(ptLimit);
      pvxNew->SetInscribedCircleCenterType(APM2D::icctInfinity);

      peLeftNext = CreateEdge();
      peLeftNext->SetVertex(pvxNew);
      peLeftNext->GetEdge()->SetVertex(peLeft->GetVertex());
      peLeftNext->SetFacet(peLeft->GetFacet());
      Splice(1, peLeft, peLeftNext->GetEdge());

      peRightNext = CreateEdge();
      peRightNext->SetVertex(pvxNew);
      peRightNext->GetEdge()->SetVertex(peRight->GetEdge()->GetVertex());
      peRightNext->GetEdge()->SetFacet(peRight->GetFacet());
      Splice(1, peRight->GetEdge()->GetEdgePrev(1), peRightNext->GetEdge());

      if ( peBisectorEnd == NULL || peBisector == NULL )
	{
	  ASSERT( peBisectorEnd == NULL && peBisector == NULL );
	  peBisectorEnd = peLeftNext;
	  Splice(1, peLeftNext, peRightNext);
	}
      else
	{
	  peBisector->SetVertex(pvxNew);
	  Splice(1, peLeftNext, peBisector);
	  Splice(1, peBisector, peRightNext);
	}
    }

#ifdef ENABLE_VORONOI_POLYGON_DEBUGGING
  //fprintf(stderr, "    END  : Merge Left = ( %.3f, %.3f ) Right = ( %.3f, %.3f )\n", peLeftMiddle->GetVertex()->GetPoint()(X), peLeftMiddle->GetVertex()->GetPoint()(Y), peRightMiddle->GetVertex()->GetPoint()(X), peRightMiddle->GetVertex()->GetPoint()(Y));
#endif // ENABLE_VORONOI_POLYGON_DEBUGGING
}

QEMesh *QEMesh::CopyVoronoiMountainMesh()
{
  CListInt listContour;
  CObjectList<QEEdge> listQEEdge;
  CObjectList<QEFacet> listQEFacet;
  QEMesh *pme;
  QEMeshIter iterQEMesh;
  QEFacetIter iterQEFacet;
  QEVertex *pvx, *pvxNew;
  QEEdge *pe, *peStart, *peCurr, *peNew, *peCenter;
  QEFacet *pfa, *pfaNew;
  Vector v;
  Point pt, ptStart, ptMin, ptMax, ptOrigin;
  FLOAT fMag;
  UINT_32 i;

  listContour.Init();
  listQEEdge.Init();
  listQEFacet.Init();

  pme = CopyVoronoiPolygon();
  iterQEMesh.Init(pme);

  //////////////////////////////////////////////////
  // Calculate the bounds of the mesh
  ptMin.Set(0.0, 0.0, 0.0);
  ptMax.Set(0.0, 0.0, 0.0);
  for ( pvx = iterQEMesh.PeekFirstVertex(); 
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pt = pvx->GetPoint();
      if ( pt(3) != 0.0 )
	{
	  for ( i = 0; i < 3; i++ )
	    {
	      ptMin[i] = MIN(ptMin(i), pt(i));
	      ptMax[i] = MAX(ptMax(i), pt(i));
	    }
	}
    }

  v = ptMax - ptMin;
  v[2] = 0.0;

  fMag = v.Magnitude();
  ptMin[2] -= (fMag/10.0);

  ptOrigin = (ptMin + ptMax) / 2.0;
  ptOrigin[2] = ptMin[2];
  ptOrigin[3] = 1.0;

  //////////////////////////////////////////////////
  // Set all edges sharp
  peStart = NULL;
  for ( pe = iterQEMesh.PeekFirstEdge(); 
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    { 
      pe->SetKnotInterval(-1.0);
      if ( pe->GetFacet() == QEMesh::GetNullFacet() )
	{
	  peStart = pe;
	}
    }
  ASSERT( peStart != NULL );

  //////////////////////////////////////////////////
  // Turn points at infinity into finite points
  peCurr = peStart;
  do
    {
      ASSERT( peCurr != NULL );
      ASSERT( peCurr->GetFacet() == QEMesh::GetNullFacet() );
      ASSERT( peCurr->GetEdge()->GetFacet() != QEMesh::GetNullFacet() );

      pe = peCurr->GetEdgePrev(1);
      ASSERT( pe != NULL );
      ASSERT( pe->GetEdgePrev(1)->GetEdge()->GetFacet() == QEMesh::GetNullFacet() &&
	      pe->GetEdgeNext(1)->GetFacet() == QEMesh::GetNullFacet() );
      pvx = pe->GetVertex();
      pt = pvx->GetPoint();
      ASSERT( pt(3) == 0.0 );

      if ( pe->GetFacet()->GetForwardVertex() != NULL &&
	   pe->GetEdge()->GetFacet()->GetForwardVertex() != NULL )
	{
	  // Hyperbola
	  v = (pe->GetFacet()->GetForwardVertex()->GetPoint() -
	       pe->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint());
	  v.Set(-v(1), v(0), 0.0);
	  ptStart = (pe->GetFacet()->GetForwardVertex()->GetPoint() +
		     pe->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint()) / 2.0;
	  fMag = v.Dot(v);
	  fMag = sqrt((ptMin(2)*ptMin(2) - (fMag/4.0))/fMag);
	  
	  pt = ptStart + (fMag*v);
	  pt[2] = ptMin(2);
	}
      else 
	{
	  // Straight line rays
	  // Parabola
	  v = pt;
	  ptStart = pe->GetEdge()->GetVertex()->GetPoint();
	  fMag = (ptStart(2) - ptMin(2))/(-v(2));
	  
	  pt = ptStart + (fMag*v);
	}

      pt[3] = 1.0;
      pvx->SetPoint(pt);

      peCurr = peCurr->GetEdgeNext(0);
    }
  while ( peCurr != peStart );

  //////////////////////////////////////////////////
  // Subdivide conical facets that extend to infinity
  // into a control structure that approximates a cone
  listQEEdge.Splice(pme->m_plistQEEdge);
  for ( pe = listQEEdge.RemoveFirst();
	pe != NULL;
	pe = listQEEdge.RemoveFirst() )
    {
      pme->m_plistQEEdge->InsertLast(pe);
      listQEEdge.Remove(pe->GetEdge());
      pme->m_plistQEEdge->InsertLast(pe->GetEdge());

      if ( pe->GetFacet() == QEMesh::GetNullFacet() )
	{
	  pe = pe->GetEdge();
	}
      else if ( pe->GetEdge()->GetFacet() != QEMesh::GetNullFacet() )
	{
	  if ( pe->GetFacet()->GetForwardEdge() != NULL &&
	       pe->GetEdge()->GetFacet()->GetForwardVertex() != NULL )
	    {
	      pe = pe->GetEdge();
	    }
	}
      ASSERT( pe->GetFacet() != QEMesh::GetNullFacet() );

      if ( pe->GetFacet()->GetForwardVertex() != NULL &&
	   pe->GetEdge()->GetFacet() == QEMesh::GetNullFacet() )
	{
	  //////////////////////////////////////////////////
	  // Approximating a circle at the base of a cone

	  Point ptCenter, ppt[2], pptControl[2];
	  Vector pv[2];
	  FLOAT pfMag[2], fCos, fSin, fTheta, fA; 

	  for ( peCenter = pe->GetEdgeNext(0);
		peCenter != pe;
		peCenter = peCenter->GetEdgeNext(0) )
	    {
	      if ( peCenter->GetVertex() == pe->GetFacet()->GetForwardVertex() )
		{
		  break;
		}
	    }
	  ASSERT( peCenter->GetVertex() == pe->GetFacet()->GetForwardVertex() );

	  ptCenter = peCenter->GetVertex()->GetPoint();
	  ppt[0] = pe->GetVertex()->GetPoint();
	  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();

	  pv[0] = ppt[0] - ptCenter;
	  pv[0][2] = 0.0;
	  pv[1] = ppt[1] - ptCenter;
	  pv[1][2] = 0.0;

	  pfMag[0] = pv[0].Magnitude();
	  pfMag[1] = pv[1].Magnitude();

	  fCos = pv[0].Dot(pv[1])/(pfMag[0]*pfMag[1]);
	  fSin = (pv[0](0)*pv[1](1) - pv[0](1)*pv[1](0))/(pfMag[0]*pfMag[1]);
	  ASSERT( fSin != 0.0 );
	  fTheta = atan2(fSin, fCos);
	  fTheta /= 2.0;
	  fCos = cos(fTheta);
	  fSin = sin(fTheta);

	  fA = 4.0*(1.0 - fCos) / (3.0 * fSin);
	  pv[0].Set(-pv[0](1), pv[0](0), 0.0); 
	  pv[1].Set(pv[1](1), -pv[1](0), 0.0); 

	  pptControl[0] = ppt[0] + pv[0]*fA;
	  pptControl[1] = ppt[1] + pv[1]*fA;

	  for ( i = 2; i > 0; i-- )
	    {
	      pvxNew = pme->CreateVertex(pptControl[i-1]);

	      peNew = pme->CreateEdge();
	      peNew->SetKnotInterval(-1.0);
	      peNew->GetEdge()->SetKnotInterval(-1.0);
	      peNew->GetEdge()->SetVertex(pe->GetEdge()->GetVertex());
	      peNew->SetVertex(pvxNew);
	      peNew->SetFacet(pe->GetFacet());
	      peNew->GetEdge()->SetFacet(pe->GetEdge()->GetFacet());
	      pe->GetEdge()->SetVertex(pvxNew);
	      Splice(1, pe->GetEdge()->GetEdgePrev(1), peNew->GetEdge());
	      Splice(1, pe->GetEdge()->GetEdgePrev(1), pe->GetEdge());
	      Splice(1, pe->GetEdge(), peNew);
	    }
	}
      else if ( pe->GetFacet()->GetForwardVertex() != NULL &&
		pe->GetEdge()->GetFacet()->GetForwardEdge() != NULL &&
		( pe->GetFacet()->GetForwardVertex() != pe->GetEdge()->GetFacet()->GetForwardEdge()->GetVertex() &&
		  pe->GetFacet()->GetForwardVertex() != pe->GetEdge()->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex() ) )
	{
	  //////////////////////////////////////////////////
	  // Parabolic edge

	  Point ptV0, ptE10, ptE11, ptP0, ptP1, pptControl[2], ptIntersection;
	  DOUBLE pppdV[2][2][2]; // Endpoints
	  DOUBLE ppdP[2][2]; // Parabola Points
	  DOUBLE ppdE[2][2];     // Edge vectors
	  DOUBLE dE10xE11;
	  DOUBLE dE1dotE1, dE1dotP0, dE1xP0, dE1dotP1, dE1xP1;
	  DOUBLE dR0, dR1;
	  DOUBLE dR0term, dR1term, dSqrtR0term, dSqrtR1term;
	  DOUBLE dDen, dD, dX, dY, dZ, dPlusOrMinus, dMax;

	  for ( peCenter = pe->GetEdgeNext(0);
		peCenter != pe;
		peCenter = peCenter->GetEdgeNext(0) )
	    {
	      if ( peCenter->GetVertex() == pe->GetFacet()->GetForwardVertex() )
		{
		  break;
		}
	    }
	  ASSERT( peCenter->GetVertex() == pe->GetFacet()->GetForwardVertex() );

	  ptV0 = peCenter->GetVertex()->GetPoint();
	  ptE10 = pe->GetEdge()->GetFacet()->GetForwardEdge()->GetVertex()->GetPoint();
	  ptE11 = pe->GetEdge()->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint();
	  ptP0 = pe->GetVertex()->GetPoint();
	  ptP1 = pe->GetEdge()->GetVertex()->GetPoint();

	  dR0 = fabs(ptP0(2));
	  dR1 = fabs(ptP1(2));

	  pppdV[0][0][0] = 0.0;
	  pppdV[0][0][1] = 0.0;
	  pppdV[0][1][0] = 0.0;
	  pppdV[0][1][1] = 0.0;
	  pppdV[1][0][0] = ptE10(0) - ptV0(0);
	  pppdV[1][0][1] = ptE10(1) - ptV0(1);
	  pppdV[1][1][0] = ptE11(0) - ptV0(0);
	  pppdV[1][1][1] = ptE11(1) - ptV0(1);
	  
	  ppdE[0][0] = 0.0;
	  ppdE[0][1] = 0.0;
	  ppdE[1][0] = ptE11(0) - ptE10(0);
	  ppdE[1][1] = ptE11(1) - ptE10(1);

	  ppdP[0][0] = ptP0(0) - ptV0(0);
	  ppdP[0][1] = ptP0(1) - ptV0(1);
	  ppdP[1][0] = ptP1(0) - ptV0(0);
	  ppdP[1][1] = ptP1(1) - ptV0(1);
	  
	  dE10xE11 = (pppdV[1][0][0]*pppdV[1][1][1]) - (pppdV[1][0][1]*pppdV[1][1][0]);
	  dE1dotE1 = (ppdE[1][0]*ppdE[1][0]) + (ppdE[1][1]*ppdE[1][1]);
	  dE1xP0 = (ppdE[1][0]*ppdP[0][1]) - (ppdE[1][1]*ppdP[0][0]);
	  dE1dotP0 = (ppdE[1][0]*ppdP[0][0]) + (ppdE[1][1]*ppdP[0][1]);
	  dE1xP1 = (ppdE[1][0]*ppdP[1][1]) - (ppdE[1][1]*ppdP[1][0]);
	  dE1dotP1 = (ppdE[1][0]*ppdP[1][0]) + (ppdE[1][1]*ppdP[1][1]);

	  dD = dE10xE11 / sqrt(dE1dotE1);
	  ASSERT( dD > 0.0 );

	  dPlusOrMinus = dE1dotP0 / sqrt(dE1dotE1);
	  dMax = dE1dotP1 / sqrt(dE1dotE1);
	  if ( fabs(dMax) > fabs(dPlusOrMinus) )
	    {
	      dPlusOrMinus = dMax;
	    }
	  if ( dPlusOrMinus < 0.0 )
	    {
	      dPlusOrMinus = -1.0;
	    }
	  else
	    {
	      dPlusOrMinus = 1.0;
	    }
 
	  dR0term = 2.0*dD*dR0 - dD*dD;
	  dR0term = MAX(0.0, dR0term);
	  dR1term = 2.0*dD*dR1 - dD*dD;
	  dR1term = MAX(0.0, dR1term);
	  dSqrtR0term = sqrt(dR0term);
	  dSqrtR1term = sqrt(dR1term);

	  dDen = dSqrtR1term - dSqrtR0term;
	  ASSERT( dDen != 0.0 );

	  dX = ( dPlusOrMinus*dSqrtR1term*dE1dotP1 - dPlusOrMinus*dSqrtR0term*dE1dotP0
		 - dD*(dE1xP1 - dE1xP0) ) / ( dDen*sqrt(dE1dotE1) );
	  dY = ( dPlusOrMinus*dSqrtR0term*dSqrtR1term*(dE1dotP1 - dE1dotP0)
		 - dD*dSqrtR0term*dE1xP1
		 + dD*dSqrtR1term*dE1xP0 ) / ( dD*dDen*sqrt(dE1dotE1) );
	  dZ = dD + dY;
	  ASSERT( dZ >= 0.0 );

	  ptIntersection[0] = ( ptV0(0) 
				+ (dPlusOrMinus*dX*ppdE[1][0] - dY*ppdE[1][1])/sqrt(dE1dotE1) );
	  ptIntersection[1] = ( ptV0(1) 
				+ (dPlusOrMinus*dX*ppdE[1][1] + dY*ppdE[1][0])/sqrt(dE1dotE1) );
	  if ( ( dR0 > dR1 ) ? ( ptP0(2) < 0.0 ) : ( ptP1(2) < 0.0 ) )
	    {
	      ptIntersection[2] = -dZ;
	    }
	  else
	    {
	      ptIntersection[2] = dZ;
	    }
	  ptIntersection[3] = 1.0;

	  pptControl[0] = (ptP0 + 2.0*ptIntersection) / 3.0;
	  pptControl[0][3] = 1.0;
	  pptControl[1] = (ptP1 + 2.0*ptIntersection) / 3.0;
	  pptControl[1][3] = 1.0;

	  for ( i = 2; i > 0; i-- )
	    {
	      pvxNew = pme->CreateVertex(pptControl[i-1]);

	      peNew = pme->CreateEdge();
	      peNew->SetKnotInterval(-1.0);
	      peNew->GetEdge()->SetKnotInterval(-1.0);
	      peNew->GetEdge()->SetVertex(pe->GetEdge()->GetVertex());
	      peNew->SetVertex(pvxNew);
	      peNew->SetFacet(pe->GetFacet());
	      peNew->GetEdge()->SetFacet(pe->GetEdge()->GetFacet());
	      pe->GetEdge()->SetVertex(pvxNew);
	      Splice(1, pe->GetEdge()->GetEdgePrev(1), peNew->GetEdge());
	      Splice(1, pe->GetEdge()->GetEdgePrev(1), pe->GetEdge());
	      Splice(1, pe->GetEdge(), peNew);
	    }
	}
      else if ( pe->GetFacet()->GetForwardVertex() != NULL &&
		pe->GetEdge()->GetFacet()->GetForwardVertex() != NULL &&
		pe->GetFacet()->GetForwardVertex() != pe->GetEdge()->GetFacet()->GetForwardVertex() )
	{
	  //////////////////////////////////////////////////
	  // Approximate hyperbola edge

	  Point ptV0, ptV1, ptP0, ptP1, pptControl[2], ptIntersection;
	  //Point ptE10, ptE11;
	  DOUBLE ppdV[2][2]; // Endpoints
	  DOUBLE ppdP[2][2];     // Hyperbola Points
	  DOUBLE dV1dotV1, dV1xP0, dV1xP1;
	  DOUBLE dR0, dR1;
	  DOUBLE dR0term, dR1term, dSqrtR0term, dSqrtR1term;
	  DOUBLE dDen, dA, dD, dX, dY, dZ, dMZ, dPlusOrMinus, dMax;

	  QEEdge *ppeCenter[2], *ppe[2];

	  ppe[0] = pe;
	  ppe[1] = pe->GetEdge();
	  for ( i = 0; i < 2; i++ )
	    {
	      for ( ppeCenter[i] = ppe[i]->GetEdgeNext(0);
		    ppeCenter[i] != ppe[i];
		    ppeCenter[i] = ppeCenter[i]->GetEdgeNext(0) )
		{
		  if ( ppeCenter[i]->GetVertex() == ppe[i]->GetFacet()->GetForwardVertex() )
		    {
		      break;
		    }
		}
	      ASSERT( ppeCenter[i]->GetVertex() == ppe[i]->GetFacet()->GetForwardVertex() );
	    }
	  ppeCenter[1] = ppeCenter[1]->GetEdgePrev(0);

	  ptV0 = pe->GetFacet()->GetForwardVertex()->GetPoint();
	  ptV1 = pe->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint();

	  ptP0 = pe->GetVertex()->GetPoint();
	  ptP1 = pe->GetEdge()->GetVertex()->GetPoint();

	  dR0 = fabs(ptP0(2));
	  dR1 = fabs(ptP1(2));

	  ppdV[0][0] = 0.0;
	  ppdV[0][1] = 0.0;
	  ppdV[1][0] = ptV1(0) - ptV0(0);
	  ppdV[1][1] = ptV1(1) - ptV0(1);

	  ppdP[0][0] = ptP0(0) - ptV0(0);
	  ppdP[0][1] = ptP0(1) - ptV0(1);
	  ppdP[1][0] = ptP1(0) - ptV0(0);
	  ppdP[1][1] = ptP1(1) - ptV0(1);
	  
	  dV1dotV1 = (ppdV[1][0]*ppdV[1][0]) + (ppdV[1][1]*ppdV[1][1]);
	  dV1xP0 = (ppdV[1][0]*ppdP[0][1]) - (ppdV[1][1]*ppdP[0][0]);
	  dV1xP1 = (ppdV[1][0]*ppdP[1][1]) - (ppdV[1][1]*ppdP[1][0]);

	  dD = sqrt(dV1dotV1);
	  ASSERT( dD > 0.0 );

	  dPlusOrMinus = dV1xP0 / sqrt(dV1dotV1);
	  dMax = dV1xP1 / sqrt(dV1dotV1);
	  if ( fabs(dMax) > fabs(dPlusOrMinus) )
	    {
	      dPlusOrMinus = dMax;
	    }
	  if ( dPlusOrMinus < 0.0 )
	    {
	      dPlusOrMinus = -1.0;
	    }
	  else
	    {
	      dPlusOrMinus = 1.0;
	    }
 
	  dR0term = 4.0*dR0*dR0 - dD*dD;
	  dR0term = MAX(0.0, dR0term);
	  dR1term = 4.0*dR1*dR1 - dD*dD;
	  dR1term = MAX(0.0, dR1term);
	  dSqrtR0term = sqrt(dR0term);
	  dSqrtR1term = sqrt(dR1term);

	  dDen = dR0*dSqrtR1term - dR1*dSqrtR0term;
	  ASSERT( dDen != 0.0 );

	  dX = ( dD / 2.0 );
	  dY = ( dD*dD*(dR1 - dR0) ) / ( 2.0*dDen );
	  dZ = ( dD*dD*(dSqrtR1term - dSqrtR0term) ) / ( 4.0*dDen );
	  ASSERT( dZ >= 0.0 );
	  dA = dSqrtR1term - dSqrtR0term;
	  dMZ = sqrt( ( dD*dD*dA*dA ) / ( 4.0*( dA*dA - 4.0*(dR1 - dR0)*(dR1 - dR0) ) ) );

	  dA = 4.0*(2.0*dMZ - ( dR0 + dR1 )) / (3.0*(2.0*dZ - ( dR0 + dR1 ) ));
	  ASSERT( 0.0 <= dA && dA <= 1.0 );

	  ptIntersection[0] = ( ptV0(0) 
				+ (dX*ppdV[1][0] - dPlusOrMinus*dY*ppdV[1][1])/dD );
	  ptIntersection[1] = ( ptV0(1) 
				+ (dX*ppdV[1][1] + dPlusOrMinus*dY*ppdV[1][0])/dD );
	  if ( ( dR0 > dR1 ) ? ( ptP0(2) < 0.0 ) : ( ptP1(2) < 0.0 ) )
	    {
	      ptIntersection[2] = -dZ;
	    }
	  else
	    {
	      ptIntersection[2] = dZ;
	    }
	  ptIntersection[3] = 1.0;

	  //dA = 2.0/3.0;
	  pptControl[0] = ((1.0 - dA)*ptP0 + dA*ptIntersection);
	  pptControl[0][3] = 1.0;
	  pptControl[1] = ((1.0 - dA)*ptP1 + dA*ptIntersection);
	  pptControl[1][3] = 1.0;

	  for ( i = 2; i > 0; i-- )
	    {
	      pvxNew = pme->CreateVertex(pptControl[i-1]);
	      
	      peNew = pme->CreateEdge();
	      peNew->SetKnotInterval(-1.0);
	      peNew->GetEdge()->SetKnotInterval(-1.0);
	      peNew->GetEdge()->SetVertex(pe->GetEdge()->GetVertex());
	      peNew->SetVertex(pvxNew);
	      peNew->SetFacet(pe->GetFacet());
	      peNew->GetEdge()->SetFacet(pe->GetEdge()->GetFacet());
	      pe->GetEdge()->SetVertex(pvxNew);
	      Splice(1, pe->GetEdge()->GetEdgePrev(1), peNew->GetEdge());
	      Splice(1, pe->GetEdge()->GetEdgePrev(1), pe->GetEdge());
	      Splice(1, pe->GetEdge(), peNew);
	    }
	}
    }

  //////////////////////////////////////////////////
  // Triangulate all faces
  listQEFacet.Splice(pme->m_plistQEFacet);
  for ( pfa = listQEFacet.RemoveFirst();
	pfa != NULL;
	pfa = listQEFacet.RemoveFirst() )
    {
      pme->m_plistQEFacet->InsertLast(pfa);
      iterQEFacet.Init(pfa);

      if ( pfa->GetForwardVertex() != NULL )
	{
	  pvx = pfa->GetForwardVertex();	  
	}
      else
	{
	  pvx = pfa->GetForwardEdge()->GetVertex();	  
	}

      for ( i = 0, pe = iterQEFacet.PeekFirstEdge();
	    pe != NULL;
	    i++, pe = iterQEFacet.PeekNextEdge() )
	{
	  if ( pe->GetVertex() == pvx )
	    {
	      peCenter = pe;
	    }
	}
      ASSERT( peCenter->GetVertex() == pvx );

      if ( i > 3 )
	{
	  if ( pfa->GetForwardVertex() != NULL )
	    {
	      //////////////////////////////
	      // Make a triangle fan from the vertex
	      for ( pe = peCenter->GetEdgePrev(0)->GetEdgePrev(0)->GetEdgePrev(0);
		    pe != peCenter;
		    pe = pe->GetEdgePrev(0) )
		{
		  pfaNew = pme->CreateFacet();
		  pfaNew->SetForwardVertex(peCenter->GetFacet()->GetForwardVertex());
		  pfaNew->SetForwardEdge(peCenter->GetFacet()->GetForwardEdge());
		  
		  peNew = pme->CreateEdge();
		  peNew->SetKnotInterval(0.0);
		  peNew->GetEdge()->SetKnotInterval(0.0);
		  peNew->GetEdge()->SetFacet(peCenter->GetFacet());
		  peNew->GetEdge()->SetVertex(pe->GetEdge()->GetVertex());
		  peNew->SetVertex(peCenter->GetVertex());
		  Splice(1, pe->GetEdge()->GetEdgePrev(1), peNew->GetEdge());
		  Splice(1, peCenter, peNew);
		  
		  peCurr = peNew;
		  do
		    {
		      peCurr->SetFacet(pfaNew);

		      peCurr = peCurr->GetEdgeNext(0);
		    }
		  while ( peCurr != peNew );
		}
	    }
	  else
	    {
	      ASSERT( pfa->GetForwardEdge() != NULL );

	      //////////////////////////////
	      // Triangulate the possibly concave polygon
	      for ( pe = iterQEFacet.PeekFirstEdge();
		    pe != NULL;
		    pe = iterQEFacet.PeekNextEdge() )
		{
		  listContour.InsertLast(pe, 0);
		}
	      
	      pme->TriangulatePolygon(&listContour);
	      // TriangulatePolygon does not expect a facet to already exist
	      pme->DestroyFacet(pfa);

	      listContour.RemoveAll();
	    }
	}
    }


  //////////////////////////////////////////////////
  // Turn the bottom into a triangle fan
  peCurr = peStart;
  pvxNew = pme->CreateVertex(ptOrigin);

  pfaNew = pme->CreateFacet();
  peNew = pme->CreateEdge();
  peNew->SetKnotInterval(0.0);
  peNew->GetEdge()->SetKnotInterval(0.0);
  peNew->SetVertex(peCurr->GetVertex());
  peNew->GetEdge()->SetVertex(pvxNew);
  peNew->GetEdge()->SetFacet(pfaNew);
  peCurr->SetFacet(pfaNew);

  Splice(1, peNew, peCurr);
  peStart = peNew;

  for ( peCurr = peCurr->GetEdgeNext(0);
	peCurr != peStart;
	peCurr = peCurr->GetEdgeNext(0) )
    {
      pfaNew = pme->CreateFacet();
      peNew = pme->CreateEdge();
      peNew->SetKnotInterval(0.0);
      peNew->GetEdge()->SetKnotInterval(0.0);
      peNew->SetVertex(peCurr->GetVertex());
      peNew->GetEdge()->SetVertex(pvxNew);
      peNew->SetFacet(peCurr->GetEdgePrev(0)->GetFacet());
      peNew->GetEdge()->SetFacet(pfaNew);
      peCurr->SetFacet(pfaNew);
      
      Splice(1, peNew->GetEdge(), peCurr->GetEdgePrev(0)->GetEdgePrev(0));
      Splice(1, peNew, peCurr);
    }
  peStart->SetFacet(pfaNew);

  //pme->Display();

  return pme;
}



