#include "slideparser.h"

#include "slideworld.h"
#include "sliderender.h"
#include "window.h"
#include "slideviewport.h"
#include "slidecamera.h"
#include "slidelight.h"
#include "slidearealight.h"
#include "slidegroup.h"
#include "slideinstance.h"
#include "slideseam.h"
#include "slidecontour.h"
#include "slidesubdivision.h"
#include "slideoffset.h"
#include "slidedual.h"
#include "slideevolver.h"
#include "slidemirror.h"
#include "slidegeodesic.h"
#include "slideunion.h"
#include "slideintersection.h"
#include "slidedifference.h"
#include "slidesgroot.h"
#include "slidetransform.h"
#include "slidetranslate.h"
#include "slidescale.h"
#include "sliderotate.h"
#include "slidelookat.h"
#include "slideobject.h"
#include "slidemesh.h"
#include "slidemeshvertexproperty.h"
#include "slidemeshedgeproperty.h"
#include "slidemeshvertex.h"
#include "slidemeshfacet.h"
#include "slidemeshedge.h"
#include "slideconvexhull.h"
#include "slidegeneralface.h"
#include "slideterrain.h"
#include "slidesphere.h"
#include "slidecylinder.h"
#include "slidecone.h"
#include "slidetorus.h"
#include "slidecyclide.h"
#include "slidebsplinepatch.h"
#include "slideheightfield.h"
#include "slidenurbpatch.h"
#include "slidebezierpatch.h"
#include "slidebsplinecurve.h"
#include "slidepolyline.h"
#include "slidecrosssection.h"
#include "slidesweep.h"
#include "slidesweepmorph.h"
#include "slidesweepcrosssection.h"
#include "slideface.h"
#include "slidepoint.h"
#include "slidecontrolpoint.h"
#include "slidesurface.h"
#include "slidewarp.h"
#include "slidefog.h"

#include "tclparseargs.h"

extern BOOL gbPreprocessDone;
extern BOOL gbCheckSemanticsDone;

CSLIDEParser *CSLIDEParser::c_pCCurrParser = NULL;
Tcl_Interp *CSLIDEParser::m_pInterp = NULL;
const CHAR *CSLIDEParser::m_cpcOotClassName = "parser";
OotStaticClass *CSLIDEParser::m_pOotClass = NULL;
const int CSLIDEParser::m_cuiOotMethods = cuiCSLIDEParserOotMethods;

// "slide" - OOT Registration Variables
OotMethodEntry CSLIDEParser::m_pOotMethods[cuiCSLIDEParserOotMethods] =
{
  CSLIDEPARSER_METHODS
};


////////////////////////////////////////////////////////////////////////////////
// API Command Methods

// slide api
int ( CSLIDEParser::*m_pAPICmdMethods[] ) ( int, char** ) =
{
  CSLIDEAPI_CMDMETHODS
};

// slide create
int ( CSLIDEParser::*m_pCreateCmdMethods[] ) ( VOID* &, int, char** ) =
{
  CSLIDECREATE_CMDMETHODS
};

// slide modify
int ( CSLIDEParser::*m_pModifyCmdMethods[] ) ( VOID* &, int, char** ) =
{
  CSLIDEMODIFY_CMDMETHODS
};

// API Command Methods
////////////////////////////////////////////////////////////////////////////////


// Cheap WIN32 compile hack
#ifdef CreateWindow
#undef CreateWindow
#endif

CSLIDEParser::CSLIDEParser()
{
  m_pCSLIDEWorld = NULL;
  m_pCSLIDEPrimitiveCurr = NULL;
  m_pCSLIDERenderCurr = NULL;
  m_pCWindowCurr = NULL;
  m_pCSLIDEViewportCurr = NULL;
  m_pCSLIDECameraCurr = NULL;
  m_pCSLIDELightCurr = NULL;
  m_pCSLIDENodeCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;
  m_pCSLIDEAreaLightCurr = NULL;
  m_pCSLIDEGroupCurr = NULL;
  m_pCSLIDESubdivisionCurr = NULL;
  m_pCSLIDEOffsetCurr = NULL;
  m_pCSLIDEDualCurr = NULL;
  m_pCSLIDEEvolverCurr = NULL;
  m_pCSLIDEMirrorCurr = NULL;
  m_pCSLIDEGeodesicCurr = NULL;
  m_pCSLIDEInstanceCurr = NULL;
  m_pCSLIDESeamCurr = NULL;
  m_pCSLIDEContourCurr = NULL;
  m_pCSLIDEObjectCurr = NULL;
  m_pCSLIDEMeshCurr = NULL;
  m_pCSLIDEMeshVertexPropertyCurr = NULL;
  m_pCSLIDEMeshEdgePropertyCurr = NULL;
  m_pCSLIDEMeshVertexCurr = NULL;
  m_pCSLIDEMeshFacetCurr = NULL;
  m_pCSLIDEMeshEdgeCurr = NULL;
  m_pCSLIDEConvexHullCurr = NULL;
  m_pCSLIDEGeneralFaceCurr = NULL;
  m_pCSLIDETerrainCurr = NULL;
  m_pCSLIDESGRootCurr = NULL;
  m_pCSLIDESphereCurr = NULL;
  m_pCSLIDECylinderCurr = NULL;
  m_pCSLIDEConeCurr = NULL;
  m_pCSLIDETorusCurr = NULL;
  m_pCSLIDECyclideCurr = NULL;
  m_pCSLIDEBSplinePatchCurr = NULL;
  m_pCSLIDEHeightfieldCurr = NULL;
  m_pCSLIDENurbPatchCurr = NULL;
  m_pCSLIDEBezierPatchCurr = NULL;
  m_pCSLIDEBSplineCurveCurr = NULL;
  m_pCSLIDEPolylineCurr = NULL;
  m_pCSLIDECrossSectionCurr = NULL;
  m_pCSLIDESweepCurr = NULL;
  m_pCSLIDESweepMorphCurr = NULL;
  m_pCSLIDESweepCrossSectionCurr = NULL;
  m_pCSLIDEFaceCurr = NULL;
  m_pCSLIDEPointCurr = NULL;
  m_pCSLIDEControlPointCurr = NULL;
  m_pCSLIDESurfaceCurr = NULL;
  m_pCSLIDEWormwarpCurr = NULL;
  m_pCSLIDEFogCurr = NULL;
  m_pCSLIDELookAtCurr = NULL;

  // slide api
  static const char* pcAPICommandNames[cuiCSLIDEAPICommands] = 
  {
    CSLIDEAPI_COMMANDS
  };

  m_hashOotAPIFuncMap.Init(101);
  
  int i; 

  for (i=0; i<cuiCSLIDEAPICommands; i++)
    {
      int *pos = new int;
      *pos = i;
      m_hashOotAPIFuncMap.Insert(pos, pcAPICommandNames[i]);
    }

  // slide create
  static const char* pcCreateCommandNames[cuiCSLIDECreateCommands] = 
  {
    CSLIDECREATE_COMMANDS
  };

  m_hashOotCreateFuncMap.Init(101);
  
  for (i=0; i<cuiCSLIDECreateCommands; i++)
    {
      int *pos = new int;
      *pos = i;
      m_hashOotCreateFuncMap.Insert(pos, pcCreateCommandNames[i]);
    }

  // slide modify
  static const char* pcModifyCommandNames[cuiCSLIDEModifyCommands] = 
  {
    CSLIDEMODIFY_COMMANDS
  };

  m_hashOotModifyFuncMap.Init(101);
  
  for (i=0; i<cuiCSLIDEModifyCommands; i++)
    {
      int *pos = new int;
      *pos = i;
      m_hashOotModifyFuncMap.Insert(pos, pcModifyCommandNames[i]);
    }

  m_uiInstances = 1;
}

CSLIDEParser::~CSLIDEParser()
{
  // unset curr parser pointer
  //
  CSLIDEParser::c_pCCurrParser = NULL;
}

VOID CSLIDEParser::Init(CHAR *pcFile, CHAR **ppcDirectories, UINT_32 uiDirectories, 
			CSLIDEParser **ppyy_CSLIDEParser, 
			YYLTYPE *pyylloc, YY_PARSE_TYPE *pfyy_Parse,
			CScanner **ppyy_CScanner,
			int *pyyleng,
			YY_BEGIN_INITIAL_TYPE *pfyy_BeginInitial,
			YY_GET_CURRENT_BUFFER_TYPE *pfyy_GetCurrentBuffer,
			YY_CREATE_BUFFER_TYPE *pfyy_CreateBuffer,
			YY_DELETE_BUFFER_TYPE *pfyy_DeleteBuffer,
			YY_SWITCH_TO_BUFFER_TYPE *pfyy_SwitchToBuffer,
			YY_SET_VALUE_INT_TYPE *pfyy_SetValueInt,
			YY_SET_VALUE_FLOAT_TYPE *pfyy_SetValueFloat,
			YY_SET_VALUE_STRING_TYPE *pfyy_SetValueString)
{
  CParser::Init(pcFile, ppcDirectories, uiDirectories, 
		(CParser**)ppyy_CSLIDEParser, 
		pyylloc, pfyy_Parse,
		ppyy_CScanner,
		pyyleng,
		pfyy_BeginInitial,
		pfyy_GetCurrentBuffer,
		pfyy_CreateBuffer,
		pfyy_DeleteBuffer,
		pfyy_SwitchToBuffer,
		pfyy_SetValueInt,
		pfyy_SetValueFloat,
		pfyy_SetValueString);


  m_pCSLIDEWorld = CreateWorld();
  ASSERT( m_pCSLIDEWorld != NULL );
  m_pCSLIDEPrimitiveCurr = NULL;
  m_pCSLIDERenderCurr = NULL;
  m_pCWindowCurr = NULL;
  m_pCSLIDEViewportCurr = NULL;
  m_pCSLIDECameraCurr = NULL;
  m_pCSLIDELightCurr = NULL;
  m_pCSLIDENodeCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;
  m_pCSLIDEAreaLightCurr = NULL;
  m_pCSLIDEGroupCurr = NULL;
  m_pCSLIDESubdivisionCurr = NULL;
  m_pCSLIDEOffsetCurr = NULL;
  m_pCSLIDEDualCurr = NULL;
  m_pCSLIDEEvolverCurr = NULL;
  m_pCSLIDEMirrorCurr = NULL;
  m_pCSLIDEGeodesicCurr = NULL;
  m_pCSLIDEInstanceCurr = NULL;
  m_pCSLIDESeamCurr = NULL;
  m_pCSLIDEContourCurr = NULL;
  m_pCSLIDEObjectCurr = NULL;
  m_pCSLIDEMeshCurr = NULL;
  m_pCSLIDEMeshVertexPropertyCurr = NULL;
  m_pCSLIDEMeshEdgePropertyCurr = NULL;
  m_pCSLIDEMeshVertexCurr = NULL;
  m_pCSLIDEMeshFacetCurr = NULL;
  m_pCSLIDEMeshEdgeCurr = NULL;
  m_pCSLIDEConvexHullCurr = NULL;
  m_pCSLIDEGeneralFaceCurr = NULL;
  m_pCSLIDETerrainCurr = NULL;
  m_pCSLIDESphereCurr = NULL;
  m_pCSLIDECylinderCurr = NULL;
  m_pCSLIDEConeCurr = NULL;
  m_pCSLIDECyclideCurr = NULL;
  m_pCSLIDEBSplinePatchCurr = NULL;
  m_pCSLIDEHeightfieldCurr = NULL;
  m_pCSLIDENurbPatchCurr = NULL;
  m_pCSLIDEBezierPatchCurr = NULL;
  m_pCSLIDEBSplineCurveCurr = NULL;
  m_pCSLIDEPolylineCurr = NULL;
  m_pCSLIDECrossSectionCurr = NULL;
  m_pCSLIDESweepCurr = NULL;
  m_pCSLIDESweepMorphCurr = NULL;
  m_pCSLIDESweepCrossSectionCurr = NULL;
  m_pCSLIDETorusCurr = NULL;
  m_pCSLIDEFaceCurr = NULL;
  m_pCSLIDEPointCurr = NULL;
  m_pCSLIDEControlPointCurr = NULL;
  m_pCSLIDESurfaceCurr = NULL;
  m_pCSLIDEFogCurr = NULL;
  m_pCSLIDELookAtCurr = NULL;

  SetAST(m_pCSLIDEWorld);

  // for TCL OOT
  // Only will be 1 active slide parser existent at a time.
  //
  ASSERT (c_pCCurrParser == NULL);
  c_pCCurrParser=this;

  m_uiInstances = 1;
}

VOID CSLIDEParser::Uninit()
{
  CParser::Uninit();
}

VOID CSLIDEParser::SetPrimitiveRIBBegin(CHAR *pcRIBBegin)
{
  ASSERT( m_pCSLIDEPrimitiveCurr != NULL );

  m_pCSLIDEPrimitiveCurr->SetRIBBegin(pcRIBBegin);
}

VOID CSLIDEParser::SetPrimitiveRIBEnd(CHAR *pcRIBEnd)
{
  ASSERT( m_pCSLIDEPrimitiveCurr != NULL );

  m_pCSLIDEPrimitiveCurr->SetRIBEnd(pcRIBEnd);
}

VOID CSLIDEParser::SetPrimitiveGraphPosition(DynInt *pdyniX, DynInt *pdyniY)
{
  ASSERT( m_pCSLIDEPrimitiveCurr != NULL );

  m_pCSLIDEPrimitiveCurr->SetGraphPosition(pdyniX, pdyniY);
  delete pdyniX;
  delete pdyniY;
}

VOID CSLIDEParser::AddRender(CHAR *pcFile, UINT_32 uiLine, 
			     UINT_32 uiColumn,
			     CHAR *pcViewportID, CListInt *plistCameraPath, 
			     CHAR *pcRootID)
{
  CSLIDERender *pCSLIDERender;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDERenderCurr == NULL );

  pCSLIDERender = CreateRender(pcFile, uiLine, uiColumn, 
			       pcViewportID, plistCameraPath, 
			       pcRootID);
  ASSERT( pCSLIDERender != NULL );

  VERIFY(m_pCSLIDEWorld->InsertRender(pCSLIDERender));
  m_pCSLIDEPrimitiveCurr = pCSLIDERender;
  m_pCSLIDERenderCurr = pCSLIDERender;
}

VOID CSLIDEParser::EndRender()
{
  ASSERT( m_pCSLIDERenderCurr != NULL );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDERenderCurr );
  m_pCSLIDERenderCurr = NULL;
  m_pCSLIDEPrimitiveCurr = NULL;
}

VOID CSLIDEParser::SetRenderMergeEpsilon(DynFloat *pdynfMergeEpsilon)
{
  ASSERT( m_pCSLIDERenderCurr != NULL );

  m_pCSLIDERenderCurr->SetMergeEpsilon(pdynfMergeEpsilon);
}

VOID CSLIDEParser::SetRenderTclTag(CHAR *pcTclTag)
{
  ASSERT( m_pCSLIDERenderCurr != NULL );

  m_pCSLIDERenderCurr->SetTclTag(pcTclTag);
}

VOID CSLIDEParser::SetRenderLOD(DynInt *pdyniLOD)
{
  ASSERT( m_pCSLIDERenderCurr != NULL );

  m_pCSLIDERenderCurr->SetLOD(pdyniLOD);
  delete pdyniLOD;
}

VOID CSLIDEParser::SetBlendFlag(DynInt *pdyniOnOff)
{
  ASSERT( m_pCSLIDERenderCurr != NULL );

  m_pCSLIDERenderCurr->SetBlend(pdyniOnOff);
  delete pdyniOnOff;
}

VOID CSLIDEParser::SetRenderCUIType(DynInt *pdyniCUIType)
{
  ASSERT( m_pCSLIDERenderCurr != NULL );

  m_pCSLIDERenderCurr->SetCUIType(pdyniCUIType);
  delete pdyniCUIType;
}
VOID CSLIDEParser::SetRenderCUISpeed(DynFloat *pdynfCUISpeed)
{
  ASSERT( m_pCSLIDERenderCurr != NULL );

  m_pCSLIDERenderCurr->SetCUISpeed(pdynfCUISpeed);
  delete pdynfCUISpeed;
}

VOID CSLIDEParser::SetRenderStencil(DynInt *pdyniStencil)
{
  ASSERT( m_pCSLIDERenderCurr != NULL );

  m_pCSLIDERenderCurr->SetStencil(pdyniStencil);
  delete pdyniStencil;
}

VOID CSLIDEParser::SetRenderFogID(CHAR *pcFogID, DynInt *pdyniOnOff)
{
  ASSERT( m_pCSLIDERenderCurr != NULL );

  m_pCSLIDERenderCurr->SetFogID(pcFogID, pdyniOnOff);
}

VOID CSLIDEParser::AddRenderLightPath(CHAR *pcFile, UINT_32 uiLine, 
				      UINT_32 uiColumn, CListInt *plistLightPath,
				      DynInt *pdyniOnOff)
{
  CSLIDELightPath *pCSLIDELightPath;

  ASSERT( m_pCSLIDERenderCurr != NULL );

  pCSLIDELightPath = CreateLightPath(pcFile, uiLine, uiColumn, plistLightPath,
				     pdyniOnOff);

  m_pCSLIDERenderCurr->AddLightPath(pCSLIDELightPath);
}

VOID CSLIDEParser::AddRenderAreaLightPath(CHAR *pcFile, UINT_32 uiLine, 
				      UINT_32 uiColumn, CListInt *plistAreaLightPath,
				      DynInt *pdyniOnOff)
{
  CSLIDEAreaLightPath *pCSLIDEAreaLightPath;

  ASSERT( m_pCSLIDERenderCurr != NULL );

  pCSLIDEAreaLightPath = CreateAreaLightPath(pcFile, uiLine, uiColumn, plistAreaLightPath,
				     pdyniOnOff);

  m_pCSLIDERenderCurr->AddAreaLightPath(pCSLIDEAreaLightPath);
}

VOID CSLIDEParser::AddWindow(CHAR *pcFile, UINT_32 uiLine, 
			      UINT_32 uiColumn,
			      CHAR *pcID)
{
  CWindow *pCWindow;

  ASSERT( m_pCWindowCurr == NULL );

  pCWindow = m_pCSLIDEWorld->PeekWindow(pcID);
  if ( pCWindow != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtWindow],
		  pcID, pCWindow->GetFileName(), 
		  pCWindow->GetFileLine(), 
		  pCWindow->GetFileColumn());
    }

  pCWindow = CreateWindow(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCWindow != NULL );

  VERIFY(m_pCSLIDEWorld->InsertWindow(pCWindow));
  m_pCWindowCurr = pCWindow;
}

VOID CSLIDEParser::EndWindow()
{
  ASSERT( m_pCWindowCurr != NULL );
  m_pCWindowCurr = NULL;
}

VOID CSLIDEParser::SetWindowOrigin(FLOAT fX, FLOAT fY)
{
  ASSERT( m_pCWindowCurr != NULL );

  m_pCWindowCurr->SetOrigin(fX, fY);
}

VOID CSLIDEParser::SetWindowSize(FLOAT fX, FLOAT fY)
{
  ASSERT( m_pCWindowCurr != NULL );

  m_pCWindowCurr->SetSize(fX, fY);
}

VOID CSLIDEParser::SetWindowResize(INT_32 iResize)
{
  ASSERT( m_pCWindowCurr != NULL );

  m_pCWindowCurr->SetResize((SLIDEOnOffType)iResize);
}

VOID CSLIDEParser::SetWindowAspectRatio(INT_32 iAspectRatio)
{
  ASSERT( m_pCWindowCurr != NULL );

  m_pCWindowCurr->SetAspectRatio((SLIDEAspectRatioType)iAspectRatio);
}

VOID CSLIDEParser::SetWindowBackground(DynVector *pdynv)
{
  ASSERT( m_pCWindowCurr != NULL );

   m_pCWindowCurr->SetBackground(pdynv);
}

// added for remote display support
// by Rich 9.28.98
VOID CSLIDEParser::SetWindowDisplay(CHAR *pcDisplayID)
{
  ASSERT( m_pCWindowCurr != NULL );

  m_pCWindowCurr->SetDisplay(pcDisplayID);
}

VOID CSLIDEParser::AddViewport(CHAR *pcFile, UINT_32 uiLine, 
			      UINT_32 uiColumn,
			      CHAR *pcID, CHAR *pcWindowID)
{
  CSLIDEViewport *pCSLIDEViewport;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDEViewportCurr == NULL );

  pCSLIDEViewport = m_pCSLIDEWorld->PeekViewport(pcID);
  if ( pCSLIDEViewport != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtViewport],
		  pcID, pCSLIDEViewport->GetFileName(), 
		  pCSLIDEViewport->GetFileLine(), 
		  pCSLIDEViewport->GetFileColumn());
    }

  pCSLIDEViewport = CreateViewport(pcFile, uiLine, uiColumn, pcID,
				   pcWindowID);
  ASSERT( pCSLIDEViewport != NULL );

  VERIFY(m_pCSLIDEWorld->InsertViewport(pCSLIDEViewport));
  m_pCSLIDEPrimitiveCurr = pCSLIDEViewport;
  m_pCSLIDEViewportCurr = pCSLIDEViewport;
}

VOID CSLIDEParser::EndViewport()
{
  ASSERT( m_pCSLIDEViewportCurr != NULL );
  ASSERT( m_pCSLIDEViewportCurr == m_pCSLIDEPrimitiveCurr );
  m_pCSLIDEViewportCurr = NULL;
  m_pCSLIDEPrimitiveCurr = NULL;
}

VOID CSLIDEParser::SetViewportOrigin(DynVector *pdynv)
{
  ASSERT( m_pCSLIDEViewportCurr != NULL );

  m_pCSLIDEViewportCurr->SetOrigin(pdynv);
  delete pdynv;
}

VOID CSLIDEParser::SetViewportSize(DynVector *pdynv)
{
  ASSERT( m_pCSLIDEViewportCurr != NULL );

  m_pCSLIDEViewportCurr->SetSize(pdynv);
  delete pdynv;
}

VOID CSLIDEParser::AddCamera(CHAR *pcFile, UINT_32 uiLine, 
			     UINT_32 uiColumn,
			     CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDECamera *pCSLIDECamera;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDECameraCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtCamera],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDECamera = CreateCamera(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDECamera != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDECamera));
  m_pCSLIDEPrimitiveCurr = pCSLIDECamera;
  m_pCSLIDECameraCurr = pCSLIDECamera;
}

VOID CSLIDEParser::EndCamera()
{
  ASSERT( m_pCSLIDECameraCurr != NULL );
  ASSERT( m_pCSLIDECameraCurr == m_pCSLIDEPrimitiveCurr );
  m_pCSLIDECameraCurr = NULL;
  m_pCSLIDEPrimitiveCurr = NULL;
}

VOID CSLIDEParser::SetCameraProjection(DynInt *pdyniProjection)
{
  ASSERT( m_pCSLIDECameraCurr != NULL );

  m_pCSLIDECameraCurr->SetProjection(pdyniProjection);
  delete pdyniProjection;
}

VOID CSLIDEParser::SetCameraFrustum(DynVector *pdynvMin, 
				    DynVector *pdynvMax)
{
  ASSERT( m_pCSLIDECameraCurr != NULL );

  m_pCSLIDECameraCurr->SetFrustumMin(pdynvMin);
  m_pCSLIDECameraCurr->SetFrustumMax(pdynvMax);
  delete pdynvMin;
  delete pdynvMax;
}

VOID CSLIDEParser::AddLight(CHAR *pcFile, UINT_32 uiLine, 
			     UINT_32 uiColumn,
			     CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDELight *pCSLIDELight;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDELightCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtLight],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDELight = CreateLight(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDELight != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDELight));
  m_pCSLIDEPrimitiveCurr = pCSLIDELight;
  m_pCSLIDELightCurr = pCSLIDELight;
}

VOID CSLIDEParser::EndLight()
{
  ASSERT( m_pCSLIDELightCurr != NULL );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDELightCurr );
  m_pCSLIDELightCurr = NULL;
  m_pCSLIDEPrimitiveCurr = NULL;
}

VOID CSLIDEParser::SetLightType(DynInt *pdyniType)
{
  ASSERT( m_pCSLIDELightCurr != NULL );

  m_pCSLIDELightCurr->SetType(pdyniType);
  delete pdyniType;
}

VOID CSLIDEParser::SetLightColor(DynVector *pdynvColor)
{
  ASSERT( m_pCSLIDELightCurr != NULL );

  m_pCSLIDELightCurr->SetColor(pdynvColor);

  delete pdynvColor;
}

VOID CSLIDEParser::SetLightDeadDistance(DynFloat *pdynfDeadDistance)
{
  ASSERT( m_pCSLIDELightCurr != NULL );

  m_pCSLIDELightCurr->SetDeadDistance(pdynfDeadDistance);

  delete pdynfDeadDistance;
}

VOID CSLIDEParser::SetLightFallOff(DynFloat *pdynfFallOff)
{
  ASSERT( m_pCSLIDELightCurr != NULL );

  m_pCSLIDELightCurr->SetFallOff(pdynfFallOff);

  delete pdynfFallOff;
}

VOID CSLIDEParser::SetLightAngularFallOff(DynFloat *pdynfAngularFallOff)
{
  ASSERT( m_pCSLIDELightCurr != NULL );

  m_pCSLIDELightCurr->SetAngularFallOff(pdynfAngularFallOff);

  delete pdynfAngularFallOff;
}

VOID CSLIDEParser::SetNodeSurfaceID(CHAR *pcSurfaceID)
{
  ASSERT( m_pCSLIDENodeCurr != NULL );

  m_pCSLIDENodeCurr->SetSurfaceID(pcSurfaceID);
}

VOID CSLIDEParser::SetNodeShading(DynInt *pdyniShading)
{
  ASSERT( m_pCSLIDENodeCurr != NULL );

  m_pCSLIDENodeCurr->SetShading(pdyniShading);
  delete pdyniShading;
}

VOID CSLIDEParser::SetNodeLOD(DynInt *pdyniLOD)
{
  ASSERT( m_pCSLIDENodeCurr != NULL );

  m_pCSLIDENodeCurr->SetLOD(pdyniLOD);
  delete pdyniLOD;
}

VOID CSLIDEParser::SetAtomNodeSolid(DynInt *pdyniSolid)
{
  ASSERT( m_pCSLIDEAtomNodeCurr != NULL );

  m_pCSLIDEAtomNodeCurr->SetSolid(pdyniSolid);
  delete pdyniSolid;
}

VOID CSLIDEParser::AddAreaLight(CHAR *pcFile, UINT_32 uiLine, 
			    UINT_32 uiColumn,
			    CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDEAreaLight *pCSLIDEAreaLight;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEGroupCurr == NULL );
  ASSERT( m_pCSLIDEAreaLightCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtObject],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDEAreaLight = CreateAreaLight(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDEAreaLight != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDEAreaLight));
  m_pCSLIDEPrimitiveCurr = pCSLIDEAreaLight;
  m_pCSLIDENodeCurr = pCSLIDEAreaLight;
  m_pCSLIDEGroupCurr = pCSLIDEAreaLight;
  m_pCSLIDEAreaLightCurr = pCSLIDEAreaLight;
}

VOID CSLIDEParser::EndAreaLight()
{
  ASSERT( m_pCSLIDEAreaLightCurr != NULL );
  ASSERT( m_pCSLIDEGroupCurr == m_pCSLIDEAreaLightCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDEAreaLightCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEAreaLightCurr );
  m_pCSLIDEAreaLightCurr = NULL;
  m_pCSLIDEGroupCurr = NULL;
  m_pCSLIDENodeCurr = NULL;
  m_pCSLIDEPrimitiveCurr = NULL;
}

VOID CSLIDEParser::SetAreaLightColor(DynVector *pdynvColor)
{
  ASSERT( m_pCSLIDEAreaLightCurr != NULL );

  m_pCSLIDEAreaLightCurr->SetColor(pdynvColor);

  delete pdynvColor;
}

VOID CSLIDEParser::SetAreaLightSamples(DynInt *pdyniSamples)
{
  ASSERT( m_pCSLIDEAreaLightCurr != NULL );

  m_pCSLIDEAreaLightCurr->SetSamples(pdyniSamples);
  delete pdyniSamples;
}

VOID CSLIDEParser::AddGroup(CHAR *pcFile, UINT_32 uiLine, 
			    UINT_32 uiColumn,
			    CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDEGroup *pCSLIDEGroup;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEGroupCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtObject],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDEGroup = CreateGroup(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDEGroup != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDEGroup));
  m_pCSLIDEPrimitiveCurr = pCSLIDEGroup;
  m_pCSLIDENodeCurr = pCSLIDEGroup;
  m_pCSLIDEGroupCurr = pCSLIDEGroup;
}

VOID CSLIDEParser::EndGroup()
{
  ASSERT( m_pCSLIDEGroupCurr != NULL );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDEGroupCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEGroupCurr );
  m_pCSLIDEGroupCurr = NULL;
  m_pCSLIDENodeCurr = NULL;
  m_pCSLIDEPrimitiveCurr = NULL;
}

VOID CSLIDEParser::SelectGroup(CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  
  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEGroupCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode == NULL )
    {
      HandleError(0, 0, "%s \"%s\" Tried to reference undefined group", 
		  gcppcSLIDEReservedWords[sliderwtObject],
		  pcID);
    }

  m_pCSLIDEPrimitiveCurr = (CSLIDEGroup *)pCSLIDENode;
  m_pCSLIDENodeCurr = (CSLIDEGroup *)pCSLIDENode;
  m_pCSLIDEGroupCurr = (CSLIDEGroup *)pCSLIDENode;
}

VOID CSLIDEParser::AddSubdivision(CHAR *pcFile, UINT_32 uiLine, 
				  UINT_32 uiColumn,
				  CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDESubdivision *pCSLIDESubdivision;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEGroupCurr == NULL );
  ASSERT( m_pCSLIDESubdivisionCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtObject],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDESubdivision = CreateSubdivision(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDESubdivision != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDESubdivision));
  m_pCSLIDEPrimitiveCurr = pCSLIDESubdivision;
  m_pCSLIDENodeCurr = pCSLIDESubdivision;
  m_pCSLIDEGroupCurr = pCSLIDESubdivision;
  m_pCSLIDESubdivisionCurr = pCSLIDESubdivision;
}

VOID CSLIDEParser::EndSubdivision()
{
  ASSERT( m_pCSLIDESubdivisionCurr != NULL );
  ASSERT( m_pCSLIDEGroupCurr == m_pCSLIDESubdivisionCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDESubdivisionCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDESubdivisionCurr );
  m_pCSLIDESubdivisionCurr = NULL;
  m_pCSLIDEGroupCurr = NULL;
  m_pCSLIDENodeCurr = NULL;
  m_pCSLIDEPrimitiveCurr = NULL;
}

VOID CSLIDEParser::SetSubdivisionUseDisplayLists(DynInt *pdyniUseDisplayLists)
{
  ASSERT( m_pCSLIDESubdivisionCurr != NULL );

  m_pCSLIDESubdivisionCurr->SetUseDisplayLists(pdyniUseDisplayLists);
  delete pdyniUseDisplayLists;
}

VOID CSLIDEParser::SetSubdivisionDrawControlMesh(DynInt *pdyniDrawControlMesh)
{
  ASSERT( m_pCSLIDESubdivisionCurr != NULL );

  m_pCSLIDESubdivisionCurr->SetDrawControlMesh(pdyniDrawControlMesh);
  delete pdyniDrawControlMesh;
}

VOID CSLIDEParser::SetSubdivisionDrawCurrentMesh(DynInt *pdyniDrawCurrentMesh)
{
  ASSERT( m_pCSLIDESubdivisionCurr != NULL );

  m_pCSLIDESubdivisionCurr->SetDrawCurrentMesh(pdyniDrawCurrentMesh);
  delete pdyniDrawCurrentMesh;
}

VOID CSLIDEParser::SetSubdivisionDrawVertices(DynInt *pdyniDrawVertices)
{
  ASSERT( m_pCSLIDESubdivisionCurr != NULL );

  m_pCSLIDESubdivisionCurr->SetDrawVertices(pdyniDrawVertices);
  delete pdyniDrawVertices;
}

VOID CSLIDEParser::SetSubdivisionType(DynInt *pdyniType)
{
  ASSERT( m_pCSLIDESubdivisionCurr != NULL );

  m_pCSLIDESubdivisionCurr->SetType(pdyniType);
  delete pdyniType;
}

VOID CSLIDEParser::SetSubdivisionErrorMetric(DynInt *pdyniErrorMetric)
{
  ASSERT( m_pCSLIDESubdivisionCurr != NULL );

  m_pCSLIDESubdivisionCurr->SetErrorMetric(pdyniErrorMetric);
  delete pdyniErrorMetric;
}

VOID CSLIDEParser::SetSubdivisionSubdivisionLevel(DynInt *pdyniSubdivisionLevel)
{
  ASSERT( m_pCSLIDESubdivisionCurr != NULL );

  m_pCSLIDESubdivisionCurr->SetSubdivisionLevel(pdyniSubdivisionLevel);
  delete pdyniSubdivisionLevel;
}

VOID CSLIDEParser::SetSubdivisionFacetsMax(DynInt *pdyniFacetsMax)
{
  ASSERT( m_pCSLIDESubdivisionCurr != NULL );

  m_pCSLIDESubdivisionCurr->SetFacetsMax(pdyniFacetsMax);
  delete pdyniFacetsMax;
}

VOID CSLIDEParser::SetSubdivisionWeightedEdges(DynInt *pdyniWeightedEdges)
{
  ASSERT( m_pCSLIDESubdivisionCurr != NULL );

  m_pCSLIDESubdivisionCurr->SetWeightedEdges(pdyniWeightedEdges);
  delete pdyniWeightedEdges;
}

VOID CSLIDEParser::SetSubdivisionWeight(DynFloat *pdynfWeight)
{
  ASSERT( m_pCSLIDESubdivisionCurr != NULL );

  m_pCSLIDESubdivisionCurr->SetWeight(pdynfWeight);
  delete pdynfWeight;
}

VOID CSLIDEParser::SetSubdivisionA(DynInt *pdyniA)
{
  ASSERT( m_pCSLIDESubdivisionCurr != NULL );

  m_pCSLIDESubdivisionCurr->SetA(pdyniA);
  delete pdyniA;
}

VOID CSLIDEParser::SetSubdivisionB(DynInt *pdyniB)
{
  ASSERT( m_pCSLIDESubdivisionCurr != NULL );

  m_pCSLIDESubdivisionCurr->SetB(pdyniB);
  delete pdyniB;
}

VOID CSLIDEParser::SetSubdivisionTolerance(DynFloat *pdynfTolerance)
{
  ASSERT( m_pCSLIDESubdivisionCurr != NULL );

  m_pCSLIDESubdivisionCurr->SetTolerance(pdynfTolerance);
  delete pdynfTolerance;
}

VOID CSLIDEParser::AddOffset(CHAR *pcFile, UINT_32 uiLine, 
			    UINT_32 uiColumn,
			    CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDEOffset *pCSLIDEOffset;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEGroupCurr == NULL );
  ASSERT( m_pCSLIDEOffsetCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtObject],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDEOffset = CreateOffset(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDEOffset != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDEOffset));
  m_pCSLIDEPrimitiveCurr = pCSLIDEOffset;
  m_pCSLIDENodeCurr = pCSLIDEOffset;
  m_pCSLIDEGroupCurr = pCSLIDEOffset;
  m_pCSLIDEOffsetCurr = pCSLIDEOffset;
}

VOID CSLIDEParser::EndOffset()
{
  ASSERT( m_pCSLIDEOffsetCurr != NULL );
  ASSERT( m_pCSLIDEGroupCurr == m_pCSLIDEOffsetCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDEOffsetCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEOffsetCurr );
  m_pCSLIDEOffsetCurr = NULL;
  m_pCSLIDEGroupCurr = NULL;
  m_pCSLIDENodeCurr = NULL;
  m_pCSLIDEPrimitiveCurr = NULL;
}

VOID CSLIDEParser::SetOffsetDrawControlMesh(DynInt *pdyniDrawControlMesh)
{
  ASSERT( m_pCSLIDEOffsetCurr != NULL );

  m_pCSLIDEOffsetCurr->SetDrawControlMesh(pdyniDrawControlMesh);
  delete pdyniDrawControlMesh;
}

VOID CSLIDEParser::SetOffsetType(DynInt *pdyniType)
{
  ASSERT( m_pCSLIDEOffsetCurr != NULL );

  m_pCSLIDEOffsetCurr->SetType(pdyniType);
  delete pdyniType;
}

VOID CSLIDEParser::SetOffsetRadius(DynFloat *pdynfRadius)
{
  ASSERT( m_pCSLIDEOffsetCurr != NULL );

  m_pCSLIDEOffsetCurr->SetOffset(pdynfRadius);
  delete pdynfRadius;
}

VOID CSLIDEParser::SetOffsetWidth(DynFloat *pdynfWidth)
{
  ASSERT( m_pCSLIDEOffsetCurr != NULL );

  m_pCSLIDEOffsetCurr->SetWidth(pdynfWidth);
  delete pdynfWidth;
}

VOID CSLIDEParser::AddDual(CHAR *pcFile, UINT_32 uiLine, 
			    UINT_32 uiColumn,
			    CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDEDual *pCSLIDEDual;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEGroupCurr == NULL );
  ASSERT( m_pCSLIDEDualCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtObject],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDEDual = CreateDual(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDEDual != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDEDual));
  m_pCSLIDEPrimitiveCurr = pCSLIDEDual;
  m_pCSLIDENodeCurr = pCSLIDEDual;
  m_pCSLIDEGroupCurr = pCSLIDEDual;
  m_pCSLIDEDualCurr = pCSLIDEDual;
}

VOID CSLIDEParser::EndDual()
{
  ASSERT( m_pCSLIDEDualCurr != NULL );
  ASSERT( m_pCSLIDEGroupCurr == m_pCSLIDEDualCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDEDualCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEDualCurr );
  m_pCSLIDEDualCurr = NULL;
  m_pCSLIDEGroupCurr = NULL;
  m_pCSLIDENodeCurr = NULL;
  m_pCSLIDEPrimitiveCurr = NULL;
}

VOID CSLIDEParser::SetDualDrawControlMesh(DynInt *pdyniDrawControlMesh)
{
  ASSERT( m_pCSLIDEDualCurr != NULL );

  m_pCSLIDEDualCurr->SetDrawControlMesh(pdyniDrawControlMesh);
  delete pdyniDrawControlMesh;
}

VOID CSLIDEParser::AddEvolver(CHAR *pcFile, UINT_32 uiLine, 
			    UINT_32 uiColumn,
			    CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDEEvolver *pCSLIDEEvolver;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEGroupCurr == NULL );
  ASSERT( m_pCSLIDEEvolverCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtObject],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDEEvolver = CreateEvolver(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDEEvolver != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDEEvolver));
  m_pCSLIDEPrimitiveCurr = pCSLIDEEvolver;
  m_pCSLIDENodeCurr = pCSLIDEEvolver;
  m_pCSLIDEGroupCurr = pCSLIDEEvolver;
  m_pCSLIDEEvolverCurr = pCSLIDEEvolver;
}

VOID CSLIDEParser::EndEvolver()
{
  ASSERT( m_pCSLIDEEvolverCurr != NULL );
  ASSERT( m_pCSLIDEGroupCurr == m_pCSLIDEEvolverCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDEEvolverCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEEvolverCurr );
  m_pCSLIDEEvolverCurr = NULL;
  m_pCSLIDEGroupCurr = NULL;
  m_pCSLIDENodeCurr = NULL;
  m_pCSLIDEPrimitiveCurr = NULL;
}

VOID CSLIDEParser::SetEvolverDrawControlMesh(DynInt *pdyniDrawControlMesh)
{
  ASSERT( m_pCSLIDEEvolverCurr != NULL );

  m_pCSLIDEEvolverCurr->SetDrawControlMesh(pdyniDrawControlMesh);
  delete pdyniDrawControlMesh;
}

VOID CSLIDEParser::SetEvolverDrawCurrentMesh(DynInt *pdyniDrawCurrentMesh)
{
  ASSERT( m_pCSLIDEEvolverCurr != NULL );

  m_pCSLIDEEvolverCurr->SetDrawCurrentMesh(pdyniDrawCurrentMesh);
  delete pdyniDrawCurrentMesh;
}

VOID CSLIDEParser::SetEvolverSubdivisionLevel(DynInt *pdyniSubdivisionLevel)
{
  ASSERT( m_pCSLIDEEvolverCurr != NULL );

  m_pCSLIDEEvolverCurr->SetSubdivisionLevel(pdyniSubdivisionLevel);
  delete pdyniSubdivisionLevel;
}

VOID CSLIDEParser::AddMirror(CHAR *pcFile, UINT_32 uiLine, 
			    UINT_32 uiColumn,
			    CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDEMirror *pCSLIDEMirror;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEGroupCurr == NULL );
  ASSERT( m_pCSLIDEMirrorCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtObject],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDEMirror = CreateMirror(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDEMirror != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDEMirror));
  m_pCSLIDEPrimitiveCurr = pCSLIDEMirror;
  m_pCSLIDENodeCurr = pCSLIDEMirror;
  m_pCSLIDEGroupCurr = pCSLIDEMirror;
  m_pCSLIDEMirrorCurr = pCSLIDEMirror;
}

VOID CSLIDEParser::EndMirror()
{
  ASSERT( m_pCSLIDEMirrorCurr != NULL );
  ASSERT( m_pCSLIDEGroupCurr == m_pCSLIDEMirrorCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDEMirrorCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEMirrorCurr );
  m_pCSLIDEMirrorCurr = NULL;
  m_pCSLIDEGroupCurr = NULL;
  m_pCSLIDENodeCurr = NULL;
  m_pCSLIDEPrimitiveCurr = NULL;
}

VOID CSLIDEParser::SetMirrorDrawControlMesh(DynInt *pdyniDrawControlMesh)
{
  ASSERT( m_pCSLIDEMirrorCurr != NULL );

  m_pCSLIDEMirrorCurr->SetDrawControlMesh(pdyniDrawControlMesh);
  delete pdyniDrawControlMesh;
}

VOID CSLIDEParser::AddGeodesic(CHAR *pcFile, UINT_32 uiLine, 
			    UINT_32 uiColumn,
			    CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDEGeodesic *pCSLIDEGeodesic;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEGroupCurr == NULL );
  ASSERT( m_pCSLIDEGeodesicCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtObject],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDEGeodesic = CreateGeodesic(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDEGeodesic != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDEGeodesic));
  m_pCSLIDEPrimitiveCurr = pCSLIDEGeodesic;
  m_pCSLIDENodeCurr = pCSLIDEGeodesic;
  m_pCSLIDEGroupCurr = pCSLIDEGeodesic;
  m_pCSLIDEGeodesicCurr = pCSLIDEGeodesic;
}

VOID CSLIDEParser::EndGeodesic()
{
  ASSERT( m_pCSLIDEGeodesicCurr != NULL );
  ASSERT( m_pCSLIDEGroupCurr == m_pCSLIDEGeodesicCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDEGeodesicCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEGeodesicCurr );
  m_pCSLIDEGeodesicCurr = NULL;
  m_pCSLIDEGroupCurr = NULL;
  m_pCSLIDENodeCurr = NULL;
  m_pCSLIDEPrimitiveCurr = NULL;
}

VOID CSLIDEParser::SetGeodesicDrawControlMesh(DynInt *pdyniDrawControlMesh)
{
  ASSERT( m_pCSLIDEGeodesicCurr != NULL );

  m_pCSLIDEGeodesicCurr->SetDrawControlMesh(pdyniDrawControlMesh);
  delete pdyniDrawControlMesh;
}

VOID CSLIDEParser::SetGeodesicTheta(DynFloat *pdynfTheta)
{
  ASSERT( m_pCSLIDEGeodesicCurr != NULL );

  m_pCSLIDEGeodesicCurr->SetTheta(pdynfTheta);
  delete pdynfTheta;
}

VOID CSLIDEParser::SetGeodesicIntersections(DynInt *pdyniIntersections)
{
  ASSERT( m_pCSLIDEGeodesicCurr != NULL );

  m_pCSLIDEGeodesicCurr->SetIntersections(pdyniIntersections);
  delete pdyniIntersections;
}

VOID CSLIDEParser::AddUnion(CHAR *pcFile, UINT_32 uiLine, 
			    UINT_32 uiColumn,
			    CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDEUnion *pCSLIDEUnion;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEGroupCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtObject],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDEUnion = CreateUnion(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDEUnion != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDEUnion));
  m_pCSLIDEPrimitiveCurr = pCSLIDEUnion;
  m_pCSLIDENodeCurr = pCSLIDEUnion;
  m_pCSLIDEGroupCurr = pCSLIDEUnion;
}

VOID CSLIDEParser::EndUnion()
{
  ASSERT( m_pCSLIDEGroupCurr != NULL );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDEGroupCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEGroupCurr );
  m_pCSLIDEGroupCurr = NULL;
  m_pCSLIDENodeCurr = NULL;
  m_pCSLIDEPrimitiveCurr = NULL;
}

VOID CSLIDEParser::AddIntersection(CHAR *pcFile, UINT_32 uiLine, 
			    UINT_32 uiColumn,
			    CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDEIntersection *pCSLIDEIntersection;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEGroupCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtObject],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDEIntersection = CreateIntersection(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDEIntersection != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDEIntersection));
  m_pCSLIDEPrimitiveCurr = pCSLIDEIntersection;
  m_pCSLIDENodeCurr = pCSLIDEIntersection;
  m_pCSLIDEGroupCurr = pCSLIDEIntersection;
}

VOID CSLIDEParser::EndIntersection()
{
  ASSERT( m_pCSLIDEGroupCurr != NULL );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDEGroupCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEGroupCurr );
  m_pCSLIDEGroupCurr = NULL;
  m_pCSLIDENodeCurr = NULL;
  m_pCSLIDEPrimitiveCurr = NULL;
}

VOID CSLIDEParser::AddDifference(CHAR *pcFile, UINT_32 uiLine, 
			    UINT_32 uiColumn,
			    CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDEDifference *pCSLIDEDifference;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEGroupCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtObject],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDEDifference = CreateDifference(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDEDifference != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDEDifference));
  m_pCSLIDEPrimitiveCurr = pCSLIDEDifference;
  m_pCSLIDENodeCurr = pCSLIDEDifference;
  m_pCSLIDEGroupCurr = pCSLIDEDifference;
}

VOID CSLIDEParser::EndDifference()
{
  ASSERT( m_pCSLIDEGroupCurr != NULL );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDEGroupCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEGroupCurr );
  m_pCSLIDEGroupCurr = NULL;
  m_pCSLIDENodeCurr = NULL;
  m_pCSLIDEPrimitiveCurr = NULL;
}

VOID CSLIDEParser::AddSGRoot(CHAR *pcFile, UINT_32 uiLine, 
			     UINT_32 uiColumn,
			     CHAR *pcID, CHAR *pcSGRoot)
{
  CSLIDENode *pCSLIDENode;
  CSLIDESGRoot *pCSLIDESGRoot;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == NULL );
  ASSERT( m_pCSLIDESGRootCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtSGRoot],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDESGRoot = CreateSGRoot(pcFile, uiLine, uiColumn,
			       pcID, pcSGRoot);
  ASSERT( pCSLIDESGRoot != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDESGRoot));
  m_pCSLIDEPrimitiveCurr = pCSLIDESGRoot;
  m_pCSLIDENodeCurr = pCSLIDESGRoot;
  m_pCSLIDEAtomNodeCurr = pCSLIDESGRoot;
  m_pCSLIDESGRootCurr = pCSLIDESGRoot;
}

VOID CSLIDEParser::EndSGRoot()
{
  ASSERT( m_pCSLIDESGRootCurr != NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == m_pCSLIDESGRootCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDESGRootCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDESGRootCurr );
  m_pCSLIDESGRootCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;  
  m_pCSLIDENodeCurr = NULL;  
  m_pCSLIDEPrimitiveCurr = NULL;  
}

VOID CSLIDEParser::AddInstance(CHAR *pcFile, UINT_32 uiLine, 
			       UINT_32 uiColumn, CHAR *pcChildID)
{
  CSLIDEInstance *pCSLIDEInstance;

  ASSERT( m_pCSLIDEGroupCurr != NULL );
  ASSERT( m_pCSLIDEGroupCurr == m_pCSLIDENodeCurr );
  ASSERT( m_pCSLIDEGroupCurr == m_pCSLIDEPrimitiveCurr );
  ASSERT( m_pCSLIDEInstanceCurr == NULL );

  pCSLIDEInstance = CreateInstance(pcFile, uiLine, uiColumn, pcChildID,
				   m_pCSLIDEGroupCurr->GetID());

  ASSERT( pCSLIDEInstance != NULL );

  VERIFY(m_pCSLIDEWorld->InsertInstance(pCSLIDEInstance));
  m_pCSLIDEGroupCurr->InsertInstance(pCSLIDEInstance);
  m_pCSLIDEInstanceCurr = pCSLIDEInstance;
  m_pCSLIDEPrimitiveCurr = m_pCSLIDEInstanceCurr;
  m_pCSLIDENodeCurr = m_pCSLIDEInstanceCurr;

  // Default name
  CHAR *pcInstanceID;
  pcInstanceID = new CHAR[13+4];
  sprintf( pcInstanceID, "slideinstance%ld", m_uiInstances);
  SetInstanceID(pcInstanceID);
  m_uiInstances++;
}

VOID CSLIDEParser::EndInstance()
{
  ASSERT( m_pCSLIDEInstanceCurr != NULL );
  ASSERT( m_pCSLIDEInstanceCurr == m_pCSLIDENodeCurr );
  ASSERT( m_pCSLIDEInstanceCurr == m_pCSLIDEPrimitiveCurr );
  ASSERT( m_pCSLIDEGroupCurr != NULL );
  m_pCSLIDEInstanceCurr = NULL;
  m_pCSLIDENodeCurr = m_pCSLIDEGroupCurr;
  m_pCSLIDEPrimitiveCurr = m_pCSLIDEGroupCurr;
}

VOID CSLIDEParser::SetInstanceID(CHAR *pcID)
{
  ASSERT( m_pCSLIDEInstanceCurr != NULL );

  // gui
  VERIFY(m_pCSLIDEWorld->InsertInstanceID(m_pCSLIDEInstanceCurr, pcID));

  m_pCSLIDEInstanceCurr->SetID(pcID);
}

VOID CSLIDEParser::AddTranslate(CHAR *pcFile, UINT_32 uiLine, 
				UINT_32 uiColumn,
				DynVector *pdynv)
{
  CSLIDETransform *pCSLIDETransform;

  ASSERT( m_pCSLIDEInstanceCurr != NULL );

  pCSLIDETransform = CreateTranslate(pcFile, uiLine, uiColumn, 
				     pdynv);
  ASSERT( pCSLIDETransform != NULL );

  VERIFY(m_pCSLIDEWorld->InsertTransform(pCSLIDETransform));
  m_pCSLIDEInstanceCurr->InsertTransform(pCSLIDETransform);
}

VOID CSLIDEParser::AddScale(CHAR *pcFile, UINT_32 uiLine, 
			    UINT_32 uiColumn,
			    DynVector *pdynv)
{
  CSLIDETransform *pCSLIDETransform;

  ASSERT( m_pCSLIDEInstanceCurr != NULL );

  pCSLIDETransform = CreateScale(pcFile, uiLine, uiColumn, 
				 pdynv);
  ASSERT( pCSLIDETransform != NULL );

  VERIFY(m_pCSLIDEWorld->InsertTransform(pCSLIDETransform));
  m_pCSLIDEInstanceCurr->InsertTransform(pCSLIDETransform);
}

VOID CSLIDEParser::AddRotate(CHAR *pcFile, UINT_32 uiLine, 
			     UINT_32 uiColumn,
			     DynVector *pdynv, DynFloat *pdynfAngle)
{
  CSLIDETransform *pCSLIDETransform;

  ASSERT( m_pCSLIDEInstanceCurr != NULL );

  pCSLIDETransform = CreateRotate(pcFile, uiLine, uiColumn, 
				  pdynv, pdynfAngle);
  ASSERT( pCSLIDETransform != NULL );

  VERIFY(m_pCSLIDEWorld->InsertTransform(pCSLIDETransform));
  m_pCSLIDEInstanceCurr->InsertTransform(pCSLIDETransform);
}

VOID CSLIDEParser::AddLookAt(CHAR *pcFile, UINT_32 uiLine, 
			     UINT_32 uiColumn)
{
  CSLIDELookAt *pCSLIDELookAt;

  ASSERT( m_pCSLIDEInstanceCurr != NULL );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEInstanceCurr );
  ASSERT( m_pCSLIDELookAtCurr == NULL );

  pCSLIDELookAt = CreateLookAt(pcFile, uiLine, uiColumn);
  ASSERT( pCSLIDELookAt != NULL );

  VERIFY(m_pCSLIDEWorld->InsertTransform(pCSLIDELookAt));
  m_pCSLIDEInstanceCurr->InsertTransform(pCSLIDELookAt);
  m_pCSLIDELookAtCurr = pCSLIDELookAt;
  m_pCSLIDEPrimitiveCurr = pCSLIDELookAt;
}

VOID CSLIDEParser::EndLookAt()
{
  ASSERT( m_pCSLIDELookAtCurr != NULL );
  ASSERT( m_pCSLIDELookAtCurr == m_pCSLIDEPrimitiveCurr );
  m_pCSLIDELookAtCurr = NULL;
  m_pCSLIDEPrimitiveCurr = m_pCSLIDEInstanceCurr;  
}

VOID CSLIDEParser::SetLookAtEye(DynVector *pdynv)
{
  ASSERT( m_pCSLIDELookAtCurr != NULL );

  m_pCSLIDELookAtCurr->SetEye(pdynv);
  delete pdynv;
}

VOID CSLIDEParser::SetLookAtTarget(DynVector *pdynv)
{
  ASSERT( m_pCSLIDELookAtCurr != NULL );

  m_pCSLIDELookAtCurr->SetTarget(pdynv);
  delete pdynv;
}

VOID CSLIDEParser::SetLookAtUp(DynVector *pdynv)
{
  ASSERT( m_pCSLIDELookAtCurr != NULL );

  m_pCSLIDELookAtCurr->SetUp(pdynv);
  delete pdynv;
}

#ifdef SLIDE_HAS_GENERAL_MATRICES
VOID CSLIDEParser::AddGeneral(CHAR *pcFile, UINT_32 uiLine, 
			      UINT_32 uiColumn,
			      FLOAT f0, FLOAT f1, FLOAT f2, FLOAT f3, 
			      FLOAT f4, FLOAT f5, FLOAT f6, FLOAT f7, 
			      FLOAT f8, FLOAT f9, FLOAT f10, FLOAT f11, 
			      FLOAT f12, FLOAT f13, FLOAT f14, FLOAT f15) 
{
  CSLIDETransform *pCSLIDETransform;
  Matrix m;

  ASSERT( m_pCSLIDEInstanceCurr != NULL );

  // Row Major Order -- Column Matrices
  /*
  m.Set(0, 0, f0);  m.Set(0, 1, f1);  m.Set(0, 2, f2);  m.Set(0, 3, f3); 
  m.Set(1, 0, f4);  m.Set(1, 1, f5);  m.Set(1, 2, f6);  m.Set(1, 3, f7); 
  m.Set(2, 0, f8);  m.Set(2, 1, f9);  m.Set(2, 2, f10); m.Set(2, 3, f11); 
  m.Set(3, 0, f12); m.Set(3, 1, f13); m.Set(3, 2, f14); m.Set(3, 3, f15); 
  */

  // Column Major Order -- Column Matrices
  m.Set(0, 0, f0);  m.Set(0, 1, f4);  m.Set(0, 2, f8);  m.Set(0, 3, f12); 
  m.Set(1, 0, f1);  m.Set(1, 1, f5);  m.Set(1, 2, f9);  m.Set(1, 3, f13); 
  m.Set(2, 0, f2);  m.Set(2, 1, f6);  m.Set(2, 2, f10); m.Set(2, 3, f14); 
  m.Set(3, 0, f3);  m.Set(3, 1, f7);  m.Set(3, 2, f11); m.Set(3, 3, f15); 

  pCSLIDETransform = CreateGeneral(pcFile, uiLine, uiColumn, m);
  ASSERT( pCSLIDETransform != NULL );

  VERIFY(m_pCSLIDEWorld->InsertTransform(pCSLIDETransform));
  m_pCSLIDEInstanceCurr->InsertTransform(pCSLIDETransform);
}
#endif // SLIDE_HAS_GENERAL_MATRICES

VOID CSLIDEParser::AddObject(CHAR *pcFile, UINT_32 uiLine, 
			     UINT_32 uiColumn,
			     CHAR *pcID, 
			     CListInt *plistFaceIDs)
{
  CSLIDENode *pCSLIDENode;
  CSLIDEObject *pCSLIDEObject;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == NULL );
  ASSERT( m_pCSLIDEObjectCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtObject],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDEObject = CreateObject(pcFile, uiLine, uiColumn,
			     pcID, plistFaceIDs);
  ASSERT( pCSLIDEObject != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDEObject));
  m_pCSLIDEPrimitiveCurr = pCSLIDEObject;
  m_pCSLIDENodeCurr = pCSLIDEObject;
  m_pCSLIDEAtomNodeCurr = pCSLIDEObject;
  m_pCSLIDEObjectCurr = pCSLIDEObject;
}

VOID CSLIDEParser::EndObject()
{
  ASSERT( m_pCSLIDEObjectCurr != NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == m_pCSLIDEObjectCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDEObjectCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEObjectCurr );
  m_pCSLIDEObjectCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;  
  m_pCSLIDENodeCurr = NULL;  
  m_pCSLIDEPrimitiveCurr = NULL;  
}

VOID CSLIDEParser::AddMesh(CHAR *pcFile, UINT_32 uiLine, 
			   UINT_32 uiColumn,
			   CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDEMesh *pCSLIDEMesh;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == NULL );
  ASSERT( m_pCSLIDEConvexHullCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtMesh],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDEMesh = CreateMesh(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDEMesh != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDEMesh));
  m_pCSLIDEPrimitiveCurr = pCSLIDEMesh;
  m_pCSLIDENodeCurr = pCSLIDEMesh;
  m_pCSLIDEAtomNodeCurr = pCSLIDEMesh;
  m_pCSLIDEMeshCurr = pCSLIDEMesh;
}

VOID CSLIDEParser::EndMesh()
{
  ASSERT( m_pCSLIDEMeshCurr != NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == m_pCSLIDEMeshCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDEMeshCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEMeshCurr );
  m_pCSLIDEMeshCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;  
  m_pCSLIDENodeCurr = NULL;  
  m_pCSLIDEPrimitiveCurr = NULL;  
}

VOID CSLIDEParser::AddMeshVertexProperty(CHAR *pcFile, UINT_32 uiLine, 
				       UINT_32 uiColumn,
				       CHAR *pcID)
{
  CSLIDEMeshVertexProperty *pCSLIDEMeshVertexProperty;

  ASSERT( m_pCSLIDEMeshCurr != NULL );

  pCSLIDEMeshVertexProperty = m_pCSLIDEMeshCurr->PeekMeshVertexProperty(pcID);
  if ( pCSLIDEMeshVertexProperty != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtMeshVertexProperty],
		  pcID, pCSLIDEMeshVertexProperty->GetFileName(), 
		  pCSLIDEMeshVertexProperty->GetFileLine(), 
		  pCSLIDEMeshVertexProperty->GetFileColumn());
    }

  pCSLIDEMeshVertexProperty = CreateMeshVertexProperty(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDEMeshVertexProperty != NULL );

  VERIFY(m_pCSLIDEMeshCurr->InsertMeshVertexProperty(pCSLIDEMeshVertexProperty));
  m_pCSLIDEMeshVertexPropertyCurr = pCSLIDEMeshVertexProperty;
}

VOID CSLIDEParser::EndMeshVertexProperty()
{
  ASSERT( m_pCSLIDEMeshVertexPropertyCurr != NULL );
  m_pCSLIDEMeshVertexPropertyCurr = NULL;
}

VOID CSLIDEParser::SetMeshVertexPropertyType(DynInt *pdyniType)
{
  ASSERT( m_pCSLIDEMeshVertexPropertyCurr != NULL );

  m_pCSLIDEMeshVertexPropertyCurr->SetType(pdyniType);
  delete pdyniType;
}

VOID CSLIDEParser::AddMeshEdgeProperty(CHAR *pcFile, UINT_32 uiLine, 
				       UINT_32 uiColumn,
				       CHAR *pcID)
{
  CSLIDEMeshEdgeProperty *pCSLIDEMeshEdgeProperty;

  ASSERT( m_pCSLIDEMeshCurr != NULL );

  pCSLIDEMeshEdgeProperty = m_pCSLIDEMeshCurr->PeekMeshEdgeProperty(pcID);
  if ( pCSLIDEMeshEdgeProperty != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtMeshEdgeProperty],
		  pcID, pCSLIDEMeshEdgeProperty->GetFileName(), 
		  pCSLIDEMeshEdgeProperty->GetFileLine(), 
		  pCSLIDEMeshEdgeProperty->GetFileColumn());
    }

  pCSLIDEMeshEdgeProperty = CreateMeshEdgeProperty(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDEMeshEdgeProperty != NULL );

  VERIFY(m_pCSLIDEMeshCurr->InsertMeshEdgeProperty(pCSLIDEMeshEdgeProperty));
  m_pCSLIDEMeshEdgePropertyCurr = pCSLIDEMeshEdgeProperty;
}

VOID CSLIDEParser::EndMeshEdgeProperty()
{
  ASSERT( m_pCSLIDEMeshEdgePropertyCurr != NULL );
  m_pCSLIDEMeshEdgePropertyCurr = NULL;
}

VOID CSLIDEParser::SetMeshEdgePropertySharpness(DynFloat *pdynfSharpness)
{
  ASSERT( m_pCSLIDEMeshEdgePropertyCurr != NULL );

  m_pCSLIDEMeshEdgePropertyCurr->SetSharpness(pdynfSharpness);
  delete pdynfSharpness;
}

VOID CSLIDEParser::AddMeshVertex(CHAR *pcFile, UINT_32 uiLine, 
				 UINT_32 uiColumn,
				 CHAR *pcID, 
				 DynPoint *pdynpt)
{
  CSLIDEMeshVertex *pCSLIDEMeshVertex;

  ASSERT( m_pCSLIDEMeshCurr != NULL );

  pCSLIDEMeshVertex = m_pCSLIDEMeshCurr->PeekMeshVertex(pcID);
  if ( pCSLIDEMeshVertex != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtMeshVertex],
		  pcID, pCSLIDEMeshVertex->GetFileName(), 
		  pCSLIDEMeshVertex->GetFileLine(), 
		  pCSLIDEMeshVertex->GetFileColumn());
    }

  pCSLIDEMeshVertex = CreateMeshVertex(pcFile, uiLine, uiColumn, pcID, pdynpt);
  ASSERT( pCSLIDEMeshVertex != NULL );

  VERIFY(m_pCSLIDEMeshCurr->InsertMeshVertex(pCSLIDEMeshVertex));
  m_pCSLIDEMeshVertexCurr = pCSLIDEMeshVertex;
}

VOID CSLIDEParser::EndMeshVertex()
{
  ASSERT( m_pCSLIDEMeshVertexCurr != NULL );
  m_pCSLIDEMeshVertexCurr = NULL;
}

VOID CSLIDEParser::SetMeshVertexVertexPropertyID(CHAR *pcVertexPropertyID)
{
  ASSERT( m_pCSLIDEMeshVertexCurr != NULL );

  m_pCSLIDEMeshVertexCurr->SetVertexPropertyID(pcVertexPropertyID);
}

VOID CSLIDEParser::SetMeshVertexNormal(DynVector *pdynv)
{
  ASSERT( m_pCSLIDEMeshVertexCurr != NULL );

  m_pCSLIDEMeshVertexCurr->SetNormal(pdynv);
  delete pdynv;
}

VOID CSLIDEParser::SetMeshVertexTexture(DynPoint *pdynpt)
{
  ASSERT( m_pCSLIDEMeshVertexCurr != NULL );

  m_pCSLIDEMeshVertexCurr->SetTexture(pdynpt);
  delete pdynpt;
}

VOID CSLIDEParser::AddMeshFacet(CHAR *pcFile, UINT_32 uiLine, 
				UINT_32 uiColumn,
				CHAR *pcID, 
				CListInt *plistPointIDs)
{
  CSLIDEMeshFacet *pCSLIDEMeshFacet;

  ASSERT( m_pCSLIDEMeshCurr != NULL );

  pCSLIDEMeshFacet = m_pCSLIDEMeshCurr->PeekMeshFacet(pcID);
  if ( pCSLIDEMeshFacet != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtMeshFacet],
		  pcID, pCSLIDEMeshFacet->GetFileName(), 
		  pCSLIDEMeshFacet->GetFileLine(), 
		  pCSLIDEMeshFacet->GetFileColumn());
    }

  pCSLIDEMeshFacet = CreateMeshFacet(pcFile, uiLine, uiColumn, pcID, plistPointIDs);
  ASSERT( pCSLIDEMeshFacet != NULL );

  VERIFY(m_pCSLIDEMeshCurr->InsertMeshFacet(pCSLIDEMeshFacet));
  m_pCSLIDEMeshFacetCurr = pCSLIDEMeshFacet;
}

VOID CSLIDEParser::EndMeshFacet()
{
  ASSERT( m_pCSLIDEMeshFacetCurr != NULL );
  m_pCSLIDEMeshFacetCurr = NULL;
}

VOID CSLIDEParser::AddMeshEdge(CHAR *pcFile, UINT_32 uiLine, 
			       UINT_32 uiColumn,
			       CHAR *pcID, 
			       CHAR *pcMeshVertexID0, CHAR *pcMeshVertexID1)
{
  CSLIDEMeshEdge *pCSLIDEMeshEdge;

  ASSERT( m_pCSLIDEMeshCurr != NULL );

  pCSLIDEMeshEdge = m_pCSLIDEMeshCurr->PeekMeshEdge(pcID);
  if ( pCSLIDEMeshEdge != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtMeshEdge],
		  pcID, pCSLIDEMeshEdge->GetFileName(), 
		  pCSLIDEMeshEdge->GetFileLine(), 
		  pCSLIDEMeshEdge->GetFileColumn());
    }

  pCSLIDEMeshEdge = CreateMeshEdge(pcFile, uiLine, uiColumn, pcID, 
				   pcMeshVertexID0, pcMeshVertexID1);
  ASSERT( pCSLIDEMeshEdge != NULL );

  VERIFY(m_pCSLIDEMeshCurr->InsertMeshEdge(pCSLIDEMeshEdge));
  m_pCSLIDEMeshEdgeCurr = pCSLIDEMeshEdge;
}

VOID CSLIDEParser::EndMeshEdge()
{
  ASSERT( m_pCSLIDEMeshEdgeCurr != NULL );
  m_pCSLIDEMeshEdgeCurr = NULL;
}

VOID CSLIDEParser::SetMeshEdgeEdgePropertyID(CHAR *pcEdgePropertyID)
{
  ASSERT( m_pCSLIDEMeshEdgeCurr != NULL );

  m_pCSLIDEMeshEdgeCurr->SetEdgePropertyID(pcEdgePropertyID);
}

VOID CSLIDEParser::AddConvexHull(CHAR *pcFile, UINT_32 uiLine, 
				 UINT_32 uiColumn,
				 CHAR *pcID, 
				 CListInt *plistPointIDs)
{
  CSLIDENode *pCSLIDENode;
  CSLIDEConvexHull *pCSLIDEConvexHull;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == NULL );
  ASSERT( m_pCSLIDEConvexHullCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtConvexHull],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDEConvexHull = CreateConvexHull(pcFile, uiLine, uiColumn,
				       pcID, plistPointIDs);
  ASSERT( pCSLIDEConvexHull != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDEConvexHull));
  m_pCSLIDEPrimitiveCurr = pCSLIDEConvexHull;
  m_pCSLIDENodeCurr = pCSLIDEConvexHull;
  m_pCSLIDEAtomNodeCurr = pCSLIDEConvexHull;
  m_pCSLIDEConvexHullCurr = pCSLIDEConvexHull;
}

VOID CSLIDEParser::EndConvexHull()
{
  ASSERT( m_pCSLIDEConvexHullCurr != NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == m_pCSLIDEConvexHullCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDEConvexHullCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEConvexHullCurr );
  m_pCSLIDEConvexHullCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;  
  m_pCSLIDENodeCurr = NULL;  
  m_pCSLIDEPrimitiveCurr = NULL;  
}

VOID CSLIDEParser::SelectGeneralFace(CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDEGeneralFace *pCSLIDEGeneralFace;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == NULL );
  ASSERT( m_pCSLIDEGeneralFaceCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode == NULL )
    {
      HandleError(0, 0, "%s \"%s\" Tried to reference undefined generalface", 
		  gcppcSLIDEReservedWords[sliderwtObject],
		  pcID);
    }
  pCSLIDEGeneralFace = (CSLIDEGeneralFace*)pCSLIDENode;

  m_pCSLIDEPrimitiveCurr = pCSLIDEGeneralFace;
  m_pCSLIDENodeCurr = pCSLIDEGeneralFace;
  m_pCSLIDEAtomNodeCurr = pCSLIDEGeneralFace;
  m_pCSLIDEGeneralFaceCurr = pCSLIDEGeneralFace;
}

VOID CSLIDEParser::AddGeneralFace(CHAR *pcFile, UINT_32 uiLine, 
			     UINT_32 uiColumn,
			     CHAR *pcID, 
			     CListInt *plistPointIDs)
{
  CSLIDENode *pCSLIDENode;
  CSLIDEGeneralFace *pCSLIDEGeneralFace;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == NULL );
  ASSERT( m_pCSLIDEGeneralFaceCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtGeneralFace],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDEGeneralFace = CreateGeneralFace(pcFile, uiLine, uiColumn,
			     pcID, plistPointIDs);
  ASSERT( pCSLIDEGeneralFace != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDEGeneralFace));
  m_pCSLIDEPrimitiveCurr = pCSLIDEGeneralFace;
  m_pCSLIDENodeCurr = pCSLIDEGeneralFace;
  m_pCSLIDEAtomNodeCurr = pCSLIDEGeneralFace;
  m_pCSLIDEGeneralFaceCurr = pCSLIDEGeneralFace;
}

VOID CSLIDEParser::EndGeneralFace()
{
  ASSERT( m_pCSLIDEGeneralFaceCurr != NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == m_pCSLIDEGeneralFaceCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDEGeneralFaceCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEGeneralFaceCurr );
  m_pCSLIDEGeneralFaceCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;  
  m_pCSLIDENodeCurr = NULL;  
  m_pCSLIDEPrimitiveCurr = NULL;  
}

VOID CSLIDEParser::InsertGeneralFaceContour(CListInt *plistPointIDs)
{
  ASSERT( m_pCSLIDEGeneralFaceCurr != NULL );

  m_pCSLIDEGeneralFaceCurr->InsertContour(plistPointIDs);
}

VOID CSLIDEParser::SetGeneralFaceType(DynInt *pdyniType)
{
  ASSERT( m_pCSLIDEGeneralFaceCurr != NULL );

  m_pCSLIDEGeneralFaceCurr->SetType(pdyniType);
  delete pdyniType;
}

VOID CSLIDEParser::SetGeneralFaceVoronoi(DynInt *pdyniVoronoi)
{
  ASSERT( m_pCSLIDEGeneralFaceCurr != NULL );

  m_pCSLIDEGeneralFaceCurr->SetVoronoi(pdyniVoronoi);
  delete pdyniVoronoi;
}

VOID CSLIDEParser::SetGeneralFaceAccuracy(DynInt *pdyniAccuracy)
{
  ASSERT( m_pCSLIDEGeneralFaceCurr != NULL );

  m_pCSLIDEGeneralFaceCurr->SetAccuracy(pdyniAccuracy);
  delete pdyniAccuracy;
}

VOID CSLIDEParser::SetGeneralFaceRandom(DynInt *pdyniRandom)
{
  ASSERT( m_pCSLIDEGeneralFaceCurr != NULL );

  m_pCSLIDEGeneralFaceCurr->SetRandomize(pdyniRandom);
  delete pdyniRandom;
}

VOID CSLIDEParser::SetGeneralFaceDrawContours(DynInt *pdyniDrawContours)
{
  ASSERT( m_pCSLIDEGeneralFaceCurr != NULL );

  m_pCSLIDEGeneralFaceCurr->SetDrawContours(pdyniDrawContours);
  delete pdyniDrawContours;
}

VOID CSLIDEParser::SetGeneralFaceDrawVertices(DynInt *pdyniDrawVertices)
{
  ASSERT( m_pCSLIDEGeneralFaceCurr != NULL );

  m_pCSLIDEGeneralFaceCurr->SetDrawVertices(pdyniDrawVertices);
  delete pdyniDrawVertices;
}

VOID CSLIDEParser::SetGeneralFaceDrawCDT(DynInt *pdyniDrawCDT)
{
  ASSERT( m_pCSLIDEGeneralFaceCurr != NULL );

  m_pCSLIDEGeneralFaceCurr->SetDrawCDT(pdyniDrawCDT);
  delete pdyniDrawCDT;
}

VOID CSLIDEParser::SetGeneralFaceDrawIndividualVoronoi(DynInt *pdyniDrawIndividualVoronoi)
{
  ASSERT( m_pCSLIDEGeneralFaceCurr != NULL );

  m_pCSLIDEGeneralFaceCurr->SetDrawIndividualVoronoi(pdyniDrawIndividualVoronoi);
  delete pdyniDrawIndividualVoronoi;
}

VOID CSLIDEParser::SetGeneralFaceDrawVoronoi(DynInt *pdyniDrawVoronoi)
{
  ASSERT( m_pCSLIDEGeneralFaceCurr != NULL );

  m_pCSLIDEGeneralFaceCurr->SetDrawVoronoi(pdyniDrawVoronoi);
  delete pdyniDrawVoronoi;
}

VOID CSLIDEParser::SetGeneralFaceDrawHeight(DynInt *pdyniDrawHeight)
{
  ASSERT( m_pCSLIDEGeneralFaceCurr != NULL );

  m_pCSLIDEGeneralFaceCurr->SetDrawHeight(pdyniDrawHeight);
  delete pdyniDrawHeight;
}

VOID CSLIDEParser::SetGeneralFaceDrawOffset(DynInt *pdyniDrawOffset)
{
  ASSERT( m_pCSLIDEGeneralFaceCurr != NULL );

  m_pCSLIDEGeneralFaceCurr->SetDrawOffset(pdyniDrawOffset);
  delete pdyniDrawOffset;
}

VOID CSLIDEParser::SetGeneralFaceDrawMillingContour(DynInt *pdyniDrawMillingContour)
{
  ASSERT( m_pCSLIDEGeneralFaceCurr != NULL );

  m_pCSLIDEGeneralFaceCurr->SetDrawMillingContour(pdyniDrawMillingContour);
  delete pdyniDrawMillingContour;
}

VOID CSLIDEParser::SetGeneralFaceMillingContour(DynInt *pdyniMillingContour)
{
  ASSERT( m_pCSLIDEGeneralFaceCurr != NULL );

  m_pCSLIDEGeneralFaceCurr->SetMillingContour(pdyniMillingContour);
  delete pdyniMillingContour;
}

VOID CSLIDEParser::SetGeneralFaceCleanUpContour(DynInt *pdyniCleanUpContour)
{
  ASSERT( m_pCSLIDEGeneralFaceCurr != NULL );

  m_pCSLIDEGeneralFaceCurr->SetCleanUpContour(pdyniCleanUpContour);
  delete pdyniCleanUpContour;
}

VOID CSLIDEParser::SetGeneralFaceOffsets(DynInt *pdyniOffsets)
{
  ASSERT( m_pCSLIDEGeneralFaceCurr != NULL );

  m_pCSLIDEGeneralFaceCurr->SetOffsets(pdyniOffsets);
  delete pdyniOffsets;
}

VOID CSLIDEParser::SetGeneralFaceOffsetMin(DynFloat *pdynfOffsetMin)
{
  ASSERT( m_pCSLIDEGeneralFaceCurr != NULL );

  m_pCSLIDEGeneralFaceCurr->SetOffsetMin(pdynfOffsetMin);
  delete pdynfOffsetMin;
}

VOID CSLIDEParser::SetGeneralFaceOffsetMax(DynFloat *pdynfOffsetMax)
{
  ASSERT( m_pCSLIDEGeneralFaceCurr != NULL );

  m_pCSLIDEGeneralFaceCurr->SetOffsetMax(pdynfOffsetMax);
  delete pdynfOffsetMax;
}

VOID CSLIDEParser::SetGeneralFaceIndividualVoronoiMin(DynInt *pdyniIndividualVoronoiMin)
{
  ASSERT( m_pCSLIDEGeneralFaceCurr != NULL );

  m_pCSLIDEGeneralFaceCurr->SetIndividualVoronoiMin(pdyniIndividualVoronoiMin);
  delete pdyniIndividualVoronoiMin;
}

VOID CSLIDEParser::SetGeneralFaceIndividualVoronoiMax(DynInt *pdyniIndividualVoronoiMax)
{
  ASSERT( m_pCSLIDEGeneralFaceCurr != NULL );

  m_pCSLIDEGeneralFaceCurr->SetIndividualVoronoiMax(pdyniIndividualVoronoiMax);
  delete pdyniIndividualVoronoiMax;
}

VOID CSLIDEParser::SetGeneralFaceSubdivisionLevel(DynInt *pdyniSubdivisionLevel)
{
  ASSERT( m_pCSLIDEGeneralFaceCurr != NULL );

  m_pCSLIDEGeneralFaceCurr->SetSubdivisionLevel(pdyniSubdivisionLevel);
  delete pdyniSubdivisionLevel;
}

VOID CSLIDEParser::SetGeneralFaceTolerance(DynFloat *pdynfTolerance)
{
  ASSERT( m_pCSLIDEGeneralFaceCurr != NULL );

  m_pCSLIDEGeneralFaceCurr->SetTolerance(pdynfTolerance);
  delete pdynfTolerance;
}

VOID CSLIDEParser::AddTerrain(CHAR *pcFile, UINT_32 uiLine, 
			     UINT_32 uiColumn,
			     CHAR *pcID, 
			     CListInt *plistPointIDs)
{
  CSLIDENode *pCSLIDENode;
  CSLIDETerrain *pCSLIDETerrain;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == NULL );
  ASSERT( m_pCSLIDETerrainCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtTerrain],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDETerrain = CreateTerrain(pcFile, uiLine, uiColumn,
			     pcID, plistPointIDs);
  ASSERT( pCSLIDETerrain != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDETerrain));
  m_pCSLIDEPrimitiveCurr = pCSLIDETerrain;
  m_pCSLIDENodeCurr = pCSLIDETerrain;
  m_pCSLIDEAtomNodeCurr = pCSLIDETerrain;
  m_pCSLIDETerrainCurr = pCSLIDETerrain;
}

VOID CSLIDEParser::EndTerrain()
{
  ASSERT( m_pCSLIDETerrainCurr != NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == m_pCSLIDETerrainCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDETerrainCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDETerrainCurr );
  m_pCSLIDETerrainCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;  
  m_pCSLIDENodeCurr = NULL;  
  m_pCSLIDEPrimitiveCurr = NULL;  
}

VOID CSLIDEParser::SetTerrainType(DynInt *pdyniType)
{
  ASSERT( m_pCSLIDETerrainCurr != NULL );

  m_pCSLIDETerrainCurr->SetType(pdyniType);
  delete pdyniType;
}

VOID CSLIDEParser::SetTerrainVoronoi(DynInt *pdyniVoronoi)
{
  ASSERT( m_pCSLIDETerrainCurr != NULL );

  m_pCSLIDETerrainCurr->SetVoronoi(pdyniVoronoi);
  delete pdyniVoronoi;
}

VOID CSLIDEParser::SetTerrainAccuracy(DynInt *pdyniAccuracy)
{
  ASSERT( m_pCSLIDETerrainCurr != NULL );

  m_pCSLIDETerrainCurr->SetAccuracy(pdyniAccuracy);
  delete pdyniAccuracy;
}

VOID CSLIDEParser::SetTerrainRandom(DynInt *pdyniRandom)
{
  ASSERT( m_pCSLIDETerrainCurr != NULL );

  m_pCSLIDETerrainCurr->SetRandomize(pdyniRandom);
  delete pdyniRandom;
}

VOID CSLIDEParser::SetTerrainDrawDelaunay(DynInt *pdyniDrawDelaunay)
{
  ASSERT( m_pCSLIDETerrainCurr != NULL );

  m_pCSLIDETerrainCurr->SetDrawDelaunay(pdyniDrawDelaunay);
  delete pdyniDrawDelaunay;
}

VOID CSLIDEParser::SetTerrainDrawVertices(DynInt *pdyniDrawVertices)
{
  ASSERT( m_pCSLIDETerrainCurr != NULL );

  m_pCSLIDETerrainCurr->SetDrawVertices(pdyniDrawVertices);
  delete pdyniDrawVertices;
}

VOID CSLIDEParser::AddSeam(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID, 
			   CHAR *pcContour1ID, CHAR *pcContour2ID) 
{
  CSLIDENode *pCSLIDENode;
  CSLIDESeam *pCSLIDESeam;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == NULL );
  ASSERT( m_pCSLIDESeamCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtObject],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDESeam = CreateSeam(pcFile, uiLine, uiColumn, pcID, pcContour1ID, pcContour2ID);
  ASSERT( pCSLIDESeam != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDESeam));
  m_pCSLIDEPrimitiveCurr = pCSLIDESeam;
  m_pCSLIDENodeCurr = pCSLIDESeam;
  m_pCSLIDEAtomNodeCurr = pCSLIDESeam;
  m_pCSLIDESeamCurr = pCSLIDESeam;
}

VOID CSLIDEParser::EndSeam()
{
  ASSERT( m_pCSLIDESeamCurr != NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == m_pCSLIDESeamCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDESeamCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDESeamCurr );
  m_pCSLIDESeamCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;  
  m_pCSLIDENodeCurr = NULL;
  m_pCSLIDEPrimitiveCurr = NULL;  
}

VOID CSLIDEParser::AddContour(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDEContour *pCSLIDEContour;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == NULL );
  ASSERT( m_pCSLIDEContourCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtObject],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDEContour = CreateContour(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDEContour != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDEContour));
  m_pCSLIDEPrimitiveCurr = pCSLIDEContour;
  m_pCSLIDENodeCurr = pCSLIDEContour;
  m_pCSLIDEAtomNodeCurr = pCSLIDEContour;
  m_pCSLIDEContourCurr = pCSLIDEContour;
}

VOID CSLIDEParser::EndContour()
{
  ASSERT( m_pCSLIDEContourCurr != NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == m_pCSLIDEContourCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDEContourCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEContourCurr );
  m_pCSLIDEContourCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;  
  m_pCSLIDENodeCurr = NULL;  
  m_pCSLIDEPrimitiveCurr = NULL;  
}

VOID CSLIDEParser::AddContourSegment(CHAR *pcID, DynInt *pDyniSide, BOOL bReverse)
{
  ASSERT( m_pCSLIDEContourCurr != NULL );

  m_pCSLIDEContourCurr->AddSegment(pcID, pDyniSide, bReverse);
}

VOID CSLIDEParser::AddSphere(CHAR *pcFile, UINT_32 uiLine, 
			     UINT_32 uiColumn,
			     CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDESphere *pCSLIDESphere;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == NULL );
  ASSERT( m_pCSLIDESphereCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtSphere],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDESphere = CreateSphere(pcFile, uiLine, uiColumn,
			       pcID);
  ASSERT( pCSLIDESphere != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDESphere));
  m_pCSLIDEPrimitiveCurr = pCSLIDESphere;
  m_pCSLIDENodeCurr = pCSLIDESphere;
  m_pCSLIDEAtomNodeCurr = pCSLIDESphere;
  m_pCSLIDESphereCurr = pCSLIDESphere;
}

VOID CSLIDEParser::EndSphere()
{
  ASSERT( m_pCSLIDESphereCurr != NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == m_pCSLIDESphereCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDESphereCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDESphereCurr );
  m_pCSLIDESphereCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;  
  m_pCSLIDENodeCurr = NULL;  
  m_pCSLIDEPrimitiveCurr = NULL;  
}

VOID CSLIDEParser::SetSphereRadius(DynFloat *pdynfRadius)
{
  ASSERT( m_pCSLIDESphereCurr != NULL );

  m_pCSLIDESphereCurr->SetRadius(pdynfRadius);
  delete pdynfRadius;
}

VOID CSLIDEParser::SetSphereZMin(DynFloat *pdynfZMin)
{
  ASSERT( m_pCSLIDESphereCurr != NULL );

  m_pCSLIDESphereCurr->SetZMin(pdynfZMin);
  delete pdynfZMin;
}

VOID CSLIDEParser::SetSphereZMax(DynFloat *pdynfZMax)
{
  ASSERT( m_pCSLIDESphereCurr != NULL );

  m_pCSLIDESphereCurr->SetZMax(pdynfZMax);
  delete pdynfZMax;
}

VOID CSLIDEParser::SetSphereThetaMax(DynFloat *pdynfThetaMax)
{
  ASSERT( m_pCSLIDESphereCurr != NULL );

  pdynfThetaMax->SetValue(DEGTORAD(pdynfThetaMax->GetValue()));
  m_pCSLIDESphereCurr->SetThetaMax(pdynfThetaMax);
  delete pdynfThetaMax;
}

VOID CSLIDEParser::SetSphereZSlices(DynInt *pdyniZSlices)
{
  ASSERT( m_pCSLIDESphereCurr != NULL );

  m_pCSLIDESphereCurr->SetZSlices(pdyniZSlices);
  delete pdyniZSlices;
}

VOID CSLIDEParser::SetSphereThetaSlices(DynInt *pdyniThetaSlices)
{
  ASSERT( m_pCSLIDESphereCurr != NULL );

  m_pCSLIDESphereCurr->SetThetaSlices(pdyniThetaSlices);
  delete pdyniThetaSlices;
}

VOID CSLIDEParser::SetSphereTexture(DynVector *pdynv0, DynVector *pdynv1, DynVector *pdynv2, DynVector *pdynv3)
{
  ASSERT( m_pCSLIDESphereCurr != NULL );

  m_pCSLIDESphereCurr->SetTexture(pdynv0, pdynv1, pdynv2, pdynv3);
  delete pdynv0;
  delete pdynv1;
  delete pdynv2;
  delete pdynv3;
}

VOID CSLIDEParser::AddCylinder(CHAR *pcFile, UINT_32 uiLine, 
			       UINT_32 uiColumn,
			       CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDECylinder *pCSLIDECylinder;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == NULL );
  ASSERT( m_pCSLIDECylinderCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtCylinder],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDECylinder = CreateCylinder(pcFile, uiLine, uiColumn,
			       pcID);
  ASSERT( pCSLIDECylinder != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDECylinder));
  m_pCSLIDEPrimitiveCurr = pCSLIDECylinder;
  m_pCSLIDENodeCurr = pCSLIDECylinder;
  m_pCSLIDEAtomNodeCurr = pCSLIDECylinder;
  m_pCSLIDECylinderCurr = pCSLIDECylinder;
}

VOID CSLIDEParser::EndCylinder()
{
  ASSERT( m_pCSLIDECylinderCurr != NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == m_pCSLIDECylinderCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDECylinderCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDECylinderCurr );
  m_pCSLIDECylinderCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;  
  m_pCSLIDENodeCurr = NULL;  
  m_pCSLIDEPrimitiveCurr = NULL;  
}

VOID CSLIDEParser::SetCylinderRadius(DynFloat *pdynfRadius)
{
  ASSERT( m_pCSLIDECylinderCurr != NULL );

  m_pCSLIDECylinderCurr->SetRadius(pdynfRadius);
  delete pdynfRadius;
}

VOID CSLIDEParser::SetCylinderZMin(DynFloat *pdynfZMin)
{
  ASSERT( m_pCSLIDECylinderCurr != NULL );

  m_pCSLIDECylinderCurr->SetZMin(pdynfZMin);
  delete pdynfZMin;
}

VOID CSLIDEParser::SetCylinderZMax(DynFloat *pdynfZMax)
{
  ASSERT( m_pCSLIDECylinderCurr != NULL );

  m_pCSLIDECylinderCurr->SetZMax(pdynfZMax);
  delete pdynfZMax;
}

VOID CSLIDEParser::SetCylinderThetaMax(DynFloat *pdynfThetaMax)
{
  ASSERT( m_pCSLIDECylinderCurr != NULL );

  pdynfThetaMax->SetValue(DEGTORAD(pdynfThetaMax->GetValue()));
  m_pCSLIDECylinderCurr->SetThetaMax(pdynfThetaMax);
  delete pdynfThetaMax;
}

VOID CSLIDEParser::SetCylinderZSlices(DynInt *pdyniZSlices)
{
  ASSERT( m_pCSLIDECylinderCurr != NULL );

  m_pCSLIDECylinderCurr->SetZSlices(pdyniZSlices);
  delete pdyniZSlices;
}

VOID CSLIDEParser::SetCylinderThetaSlices(DynInt *pdyniThetaSlices)
{
  ASSERT( m_pCSLIDECylinderCurr != NULL );

  m_pCSLIDECylinderCurr->SetThetaSlices(pdyniThetaSlices);
  delete pdyniThetaSlices;
}

VOID CSLIDEParser::SetCylinderEndCap(DynInt *pdyniEndCap)
{
  ASSERT( m_pCSLIDECylinderCurr != NULL );

  m_pCSLIDECylinderCurr->SetEndCap(pdyniEndCap);
  delete pdyniEndCap;
}

VOID CSLIDEParser::SetCylinderBeginCap(DynInt *pdyniBeginCap)
{
  ASSERT( m_pCSLIDECylinderCurr != NULL );

  m_pCSLIDECylinderCurr->SetBeginCap(pdyniBeginCap);
  delete pdyniBeginCap;
}

VOID CSLIDEParser::SetCylinderTexture(DynVector *pdynv0, DynVector *pdynv1, DynVector *pdynv2, DynVector *pdynv3)
{
  ASSERT( m_pCSLIDECylinderCurr != NULL );

  m_pCSLIDECylinderCurr->SetTexture(pdynv0, pdynv1, pdynv2, pdynv3);
  delete pdynv0;
  delete pdynv1;
  delete pdynv2;
  delete pdynv3;
}

VOID CSLIDEParser::AddCone(CHAR *pcFile, UINT_32 uiLine, 
			   UINT_32 uiColumn,
			   CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDECone *pCSLIDECone;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == NULL );
  ASSERT( m_pCSLIDEConeCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtCone],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDECone = CreateCone(pcFile, uiLine, uiColumn,
			       pcID);
  ASSERT( pCSLIDECone != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDECone));
  m_pCSLIDEPrimitiveCurr = pCSLIDECone;
  m_pCSLIDENodeCurr = pCSLIDECone;
  m_pCSLIDEAtomNodeCurr = pCSLIDECone;
  m_pCSLIDEConeCurr = pCSLIDECone;
}

VOID CSLIDEParser::EndCone()
{
  ASSERT( m_pCSLIDEConeCurr != NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == m_pCSLIDEConeCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDEConeCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEConeCurr );
  m_pCSLIDEConeCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;  
  m_pCSLIDENodeCurr = NULL;  
  m_pCSLIDEPrimitiveCurr = NULL;  
}

VOID CSLIDEParser::SetConeHeight(DynFloat *pdynfHeight)
{
  ASSERT( m_pCSLIDEConeCurr != NULL );

  m_pCSLIDEConeCurr->SetHeight(pdynfHeight);
  delete pdynfHeight;
}

VOID CSLIDEParser::SetConeZMin(DynFloat *pdynfZMin)
{
  ASSERT( m_pCSLIDEConeCurr != NULL );

  m_pCSLIDEConeCurr->SetZMin(pdynfZMin);
  delete pdynfZMin;
}

VOID CSLIDEParser::SetConeZMax(DynFloat *pdynfZMax)
{
  ASSERT( m_pCSLIDEConeCurr != NULL );

  m_pCSLIDEConeCurr->SetZMax(pdynfZMax);
  delete pdynfZMax;
}

VOID CSLIDEParser::SetConeRadius(DynFloat *pdynfRadius)
{
  ASSERT( m_pCSLIDEConeCurr != NULL );

  m_pCSLIDEConeCurr->SetRadius(pdynfRadius);
  delete pdynfRadius;
}

VOID CSLIDEParser::SetConeThetaMax(DynFloat *pdynfThetaMax)
{
  ASSERT( m_pCSLIDEConeCurr != NULL );

  pdynfThetaMax->SetValue(DEGTORAD(pdynfThetaMax->GetValue()));
  m_pCSLIDEConeCurr->SetThetaMax(pdynfThetaMax);
  delete pdynfThetaMax;
}

VOID CSLIDEParser::SetConeZSlices(DynInt *pdyniZSlices)
{
  ASSERT( m_pCSLIDEConeCurr != NULL );

  m_pCSLIDEConeCurr->SetZSlices(pdyniZSlices);
  delete pdyniZSlices;
}

VOID CSLIDEParser::SetConeThetaSlices(DynInt *pdyniThetaSlices)
{
  ASSERT( m_pCSLIDEConeCurr != NULL );

  m_pCSLIDEConeCurr->SetThetaSlices(pdyniThetaSlices);
  delete pdyniThetaSlices;
}

VOID CSLIDEParser::SetConeBeginCap(DynInt *pdyniBeginCap)
{
  ASSERT( m_pCSLIDEConeCurr != NULL );

  m_pCSLIDEConeCurr->SetBeginCap(pdyniBeginCap);
  delete pdyniBeginCap;
}

VOID CSLIDEParser::SetConeTexture(DynVector *pdynv0, DynVector *pdynv1, DynVector *pdynv2, DynVector *pdynv3)
{
  ASSERT( m_pCSLIDEConeCurr != NULL );

  m_pCSLIDEConeCurr->SetTexture(pdynv0, pdynv1, pdynv2, pdynv3);
  delete pdynv0;
  delete pdynv1;
  delete pdynv2;
  delete pdynv3;
}

VOID CSLIDEParser::AddTorus(CHAR *pcFile, UINT_32 uiLine, 
			     UINT_32 uiColumn,
			     CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDETorus *pCSLIDETorus;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == NULL );
  ASSERT( m_pCSLIDETorusCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtTorus],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDETorus = CreateTorus(pcFile, uiLine, uiColumn,
			       pcID);
  ASSERT( pCSLIDETorus != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDETorus));
  m_pCSLIDEPrimitiveCurr = pCSLIDETorus;
  m_pCSLIDENodeCurr = pCSLIDETorus;
  m_pCSLIDEAtomNodeCurr = pCSLIDETorus;
  m_pCSLIDETorusCurr = pCSLIDETorus;
}

VOID CSLIDEParser::EndTorus()
{
  ASSERT( m_pCSLIDETorusCurr != NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == m_pCSLIDETorusCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDETorusCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDETorusCurr );
  m_pCSLIDETorusCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;  
  m_pCSLIDENodeCurr = NULL;  
  m_pCSLIDEPrimitiveCurr = NULL;  
}

VOID CSLIDEParser::SetTorusMajorRadius(DynFloat *pdynfMajorRadius)
{
  ASSERT( m_pCSLIDETorusCurr != NULL );

  m_pCSLIDETorusCurr->SetMajorRadius(pdynfMajorRadius);
  delete pdynfMajorRadius;
}

VOID CSLIDEParser::SetTorusMinorRadius(DynFloat *pdynfMinorRadius)
{
  ASSERT( m_pCSLIDETorusCurr != NULL );

  m_pCSLIDETorusCurr->SetMinorRadius(pdynfMinorRadius);
  delete pdynfMinorRadius;
}

VOID CSLIDEParser::SetTorusPhiMin(DynFloat *pdynfPhiMin)
{
  ASSERT( m_pCSLIDETorusCurr != NULL );

  pdynfPhiMin->SetValue(DEGTORAD(pdynfPhiMin->GetValue()));
  m_pCSLIDETorusCurr->SetPhiMin(pdynfPhiMin);
  delete pdynfPhiMin;
}

VOID CSLIDEParser::SetTorusPhiMax(DynFloat *pdynfPhiMax)
{
  ASSERT( m_pCSLIDETorusCurr != NULL );

  pdynfPhiMax->SetValue(DEGTORAD(pdynfPhiMax->GetValue()));
  m_pCSLIDETorusCurr->SetPhiMax(pdynfPhiMax);
  delete pdynfPhiMax;
}

VOID CSLIDEParser::SetTorusThetaMax(DynFloat *pdynfThetaMax)
{
  ASSERT( m_pCSLIDETorusCurr != NULL );

  pdynfThetaMax->SetValue(DEGTORAD(pdynfThetaMax->GetValue()));
  m_pCSLIDETorusCurr->SetThetaMax(pdynfThetaMax);
  delete pdynfThetaMax;
}

VOID CSLIDEParser::SetTorusPhiSlices(DynInt *pdyniPhiSlices)
{
  ASSERT( m_pCSLIDETorusCurr != NULL );

  m_pCSLIDETorusCurr->SetPhiSlices(pdyniPhiSlices);
  delete pdyniPhiSlices;
}

VOID CSLIDEParser::SetTorusThetaSlices(DynInt *pdyniThetaSlices)
{
  ASSERT( m_pCSLIDETorusCurr != NULL );

  m_pCSLIDETorusCurr->SetThetaSlices(pdyniThetaSlices);
  delete pdyniThetaSlices;
}

VOID CSLIDEParser::SetTorusTexture(DynVector *pdynv0, DynVector *pdynv1, DynVector *pdynv2, DynVector *pdynv3)
{
  ASSERT( m_pCSLIDETorusCurr != NULL );

  m_pCSLIDETorusCurr->SetTexture(pdynv0, pdynv1, pdynv2, pdynv3);
  delete pdynv0;
  delete pdynv1;
  delete pdynv2;
  delete pdynv3;
}

VOID CSLIDEParser::AddCyclide(CHAR *pcFile, UINT_32 uiLine, 
			     UINT_32 uiColumn,
			     CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDECyclide *pCSLIDECyclide;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == NULL );
  ASSERT( m_pCSLIDECyclideCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtCyclide],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDECyclide = CreateCyclide(pcFile, uiLine, uiColumn,
			       pcID);
  ASSERT( pCSLIDECyclide != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDECyclide));
  m_pCSLIDEPrimitiveCurr = pCSLIDECyclide;
  m_pCSLIDENodeCurr = pCSLIDECyclide;
  m_pCSLIDEAtomNodeCurr = pCSLIDECyclide;
  m_pCSLIDECyclideCurr = pCSLIDECyclide;
}

VOID CSLIDEParser::EndCyclide()
{
  ASSERT( m_pCSLIDECyclideCurr != NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == m_pCSLIDECyclideCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDECyclideCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDECyclideCurr );
  m_pCSLIDECyclideCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;  
  m_pCSLIDENodeCurr = NULL;  
  m_pCSLIDEPrimitiveCurr = NULL;  
}

VOID CSLIDEParser::SetCyclideXMajorA(DynFloat *pdynfXMajorA)
{
  ASSERT( m_pCSLIDECyclideCurr != NULL );

  pdynfXMajorA->SetValue(pdynfXMajorA->GetValue());
  m_pCSLIDECyclideCurr->SetXMajorA(pdynfXMajorA);
  delete pdynfXMajorA;
}

VOID CSLIDEParser::SetCyclideYMinorB(DynFloat *pdynfYMinorB)
{
  ASSERT( m_pCSLIDECyclideCurr != NULL );

  pdynfYMinorB->SetValue(pdynfYMinorB->GetValue());
  m_pCSLIDECyclideCurr->SetYMinorB(pdynfYMinorB);
  delete pdynfYMinorB;
}

VOID CSLIDEParser::SetCyclideStringLength(DynFloat *pdynfStringLength)
{
  ASSERT( m_pCSLIDECyclideCurr != NULL );

  pdynfStringLength->SetValue(pdynfStringLength->GetValue());
  m_pCSLIDECyclideCurr->SetStringLength(pdynfStringLength);
  delete pdynfStringLength;
}

VOID CSLIDEParser::SetCyclidePhiMin(DynFloat *pdynfPhiMin)
{
  ASSERT( m_pCSLIDECyclideCurr != NULL );

  pdynfPhiMin->SetValue(DEGTORAD(pdynfPhiMin->GetValue()));
  m_pCSLIDECyclideCurr->SetPhiMin(pdynfPhiMin);
  delete pdynfPhiMin;
}

VOID CSLIDEParser::SetCyclidePhiMax(DynFloat *pdynfPhiMax)
{
  ASSERT( m_pCSLIDECyclideCurr != NULL );

  pdynfPhiMax->SetValue(DEGTORAD(pdynfPhiMax->GetValue()));
  m_pCSLIDECyclideCurr->SetPhiMax(pdynfPhiMax);
  delete pdynfPhiMax;
}

VOID CSLIDEParser::SetCyclideThetaMin(DynFloat *pdynfThetaMin)
{
  ASSERT( m_pCSLIDECyclideCurr != NULL );

  pdynfThetaMin->SetValue(DEGTORAD(pdynfThetaMin->GetValue()));
  m_pCSLIDECyclideCurr->SetThetaMin(pdynfThetaMin);
  delete pdynfThetaMin;
}

VOID CSLIDEParser::SetCyclideThetaMax(DynFloat *pdynfThetaMax)
{
  ASSERT( m_pCSLIDECyclideCurr != NULL );

  pdynfThetaMax->SetValue(DEGTORAD(pdynfThetaMax->GetValue()));
  m_pCSLIDECyclideCurr->SetThetaMax(pdynfThetaMax);
  delete pdynfThetaMax;
}

VOID CSLIDEParser::SetCyclidePhiSlices(DynInt *pdyniPhiSlices)
{
  ASSERT( m_pCSLIDECyclideCurr != NULL );

  m_pCSLIDECyclideCurr->SetPhiSlices(pdyniPhiSlices);
  delete pdyniPhiSlices;
}

VOID CSLIDEParser::SetCyclideThetaSlices(DynInt *pdyniThetaSlices)
{
  ASSERT( m_pCSLIDECyclideCurr != NULL );

  m_pCSLIDECyclideCurr->SetThetaSlices(pdyniThetaSlices);
  delete pdyniThetaSlices;
}

VOID CSLIDEParser::SetCyclideDrawEllipse(DynInt *pdyniDrawEllipse)
{
  ASSERT( m_pCSLIDECyclideCurr != NULL );

  m_pCSLIDECyclideCurr->SetDrawEllipse(pdyniDrawEllipse);
  delete pdyniDrawEllipse;
}

VOID CSLIDEParser::SetCyclideDrawHyperbola(DynInt *pdyniDrawHyperbola)
{
  ASSERT( m_pCSLIDECyclideCurr != NULL );

  m_pCSLIDECyclideCurr->SetDrawHyperbola(pdyniDrawHyperbola);
  delete pdyniDrawHyperbola;
}

VOID CSLIDEParser::SetCyclideTexture(DynVector *pdynv0, DynVector *pdynv1, DynVector *pdynv2, DynVector *pdynv3)
{
  ASSERT( m_pCSLIDECyclideCurr != NULL );

  m_pCSLIDECyclideCurr->SetTexture(pdynv0, pdynv1, pdynv2, pdynv3);
  delete pdynv0;
  delete pdynv1;
  delete pdynv2;
  delete pdynv3;
}

VOID CSLIDEParser::AddBSplinePatch(CHAR *pcFile, UINT_32 uiLine, 
				   UINT_32 uiColumn,
				   CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDEBSplinePatch *pCSLIDEBSplinePatch;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == NULL );
  ASSERT( m_pCSLIDEBSplinePatchCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtBSplinePatch],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDEBSplinePatch = CreateBSplinePatch(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDEBSplinePatch != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDEBSplinePatch));
  m_pCSLIDEPrimitiveCurr = pCSLIDEBSplinePatch;
  m_pCSLIDENodeCurr = pCSLIDEBSplinePatch;
  m_pCSLIDEAtomNodeCurr = pCSLIDEBSplinePatch;
  m_pCSLIDEBSplinePatchCurr = pCSLIDEBSplinePatch;
}

VOID CSLIDEParser::EndBSplinePatch()
{
  ASSERT( m_pCSLIDEBSplinePatchCurr != NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == m_pCSLIDEBSplinePatchCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDEBSplinePatchCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEBSplinePatchCurr );
  m_pCSLIDEBSplinePatchCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;  
  m_pCSLIDENodeCurr = NULL;  
  m_pCSLIDEPrimitiveCurr = NULL;  
}

VOID CSLIDEParser::SetBSplinePatchUOrder(DynInt *pdyniUOrder)
{
  ASSERT( m_pCSLIDEBSplinePatchCurr != NULL );

  pdyniUOrder->SetValue(pdyniUOrder->GetValue());
  m_pCSLIDEBSplinePatchCurr->SetUOrder(pdyniUOrder);
  delete pdyniUOrder;
}

VOID CSLIDEParser::SetBSplinePatchUKnots(DynInt *pdyniUKnots)
{
  ASSERT( m_pCSLIDEBSplinePatchCurr != NULL );

  pdyniUKnots->SetValue(pdyniUKnots->GetValue());
  m_pCSLIDEBSplinePatchCurr->SetUKnots(pdyniUKnots);
  delete pdyniUKnots;
}

VOID CSLIDEParser::SetBSplinePatchUKnotList(CListInt *plistUKnot)
{
  ASSERT( m_pCSLIDEBSplinePatchCurr != NULL );
  m_pCSLIDEBSplinePatchCurr->SetUKnotList(plistUKnot);
}

VOID CSLIDEParser::SetBSplinePatchVOrder(DynInt *pdyniVOrder)
{
  ASSERT( m_pCSLIDEBSplinePatchCurr != NULL );

  pdyniVOrder->SetValue(pdyniVOrder->GetValue());
  m_pCSLIDEBSplinePatchCurr->SetVOrder(pdyniVOrder);
  delete pdyniVOrder;
}

VOID CSLIDEParser::SetBSplinePatchVKnots(DynInt *pdyniVKnots)
{
  ASSERT( m_pCSLIDEBSplinePatchCurr != NULL );

  pdyniVKnots->SetValue(pdyniVKnots->GetValue());
  m_pCSLIDEBSplinePatchCurr->SetVKnots(pdyniVKnots);
  delete pdyniVKnots;
}

VOID CSLIDEParser::SetBSplinePatchVKnotList(CListInt *plistVKnot)
{
  ASSERT( m_pCSLIDEBSplinePatchCurr != NULL );
  m_pCSLIDEBSplinePatchCurr->SetVKnotList(plistVKnot);
}

VOID CSLIDEParser::SetBSplinePatchControlPointList(CListInt *plistControlPoint)
{
  ASSERT( m_pCSLIDEBSplinePatchCurr != NULL );
  m_pCSLIDEBSplinePatchCurr->SetControlPointList(plistControlPoint);
}

VOID CSLIDEParser::SetBSplinePatchUSlices(DynInt *pdyniUSlices)
{
  ASSERT( m_pCSLIDEBSplinePatchCurr != NULL );

  pdyniUSlices->SetValue(pdyniUSlices->GetValue());
  m_pCSLIDEBSplinePatchCurr->SetUSlices(pdyniUSlices);
  delete pdyniUSlices;
}

VOID CSLIDEParser::SetBSplinePatchVSlices(DynInt *pdyniVSlices)
{
  ASSERT( m_pCSLIDEBSplinePatchCurr != NULL );

  pdyniVSlices->SetValue(pdyniVSlices->GetValue());
  m_pCSLIDEBSplinePatchCurr->SetVSlices(pdyniVSlices);
  delete pdyniVSlices;
}

VOID CSLIDEParser::SetBSplinePatchDrawControls(DynInt *pdyniDrawControls)
{
  ASSERT( m_pCSLIDEBSplinePatchCurr != NULL );

  m_pCSLIDEBSplinePatchCurr->SetDrawControls(pdyniDrawControls);
  delete pdyniDrawControls;
}

VOID CSLIDEParser::SetBSplinePatchTexture(DynVector *pdynv0, DynVector *pdynv1, DynVector *pdynv2, DynVector *pdynv3)
{
  ASSERT( m_pCSLIDEBSplinePatchCurr != NULL );

  m_pCSLIDEBSplinePatchCurr->SetTexture(pdynv0, pdynv1, pdynv2, pdynv3);
  delete pdynv0;
  delete pdynv1;
  delete pdynv2;
  delete pdynv3;
}

VOID CSLIDEParser::SetBSplinePatchOffset(DynFloat *pdynfOffset, DynInt *pdyniOffsetConnected)
{
  ASSERT( m_pCSLIDEBSplinePatchCurr != NULL );

  pdynfOffset->SetValue(pdynfOffset->GetValue());
  pdyniOffsetConnected->SetValue(pdyniOffsetConnected->GetValue());
  m_pCSLIDEBSplinePatchCurr->SetOffset(pdynfOffset, pdyniOffsetConnected);
  delete pdynfOffset;
  delete pdyniOffsetConnected;
}

VOID CSLIDEParser::AddHeightfield(CHAR *pcFile, UINT_32 uiLine, 
				   UINT_32 uiColumn,
				   CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDEHeightfield *pCSLIDEHeightfield;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == NULL );
  ASSERT( m_pCSLIDEHeightfieldCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtHeightfield],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDEHeightfield = CreateHeightfield(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDEHeightfield != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDEHeightfield));
  m_pCSLIDEPrimitiveCurr = pCSLIDEHeightfield;
  m_pCSLIDENodeCurr = pCSLIDEHeightfield;
  m_pCSLIDEAtomNodeCurr = pCSLIDEHeightfield;
  m_pCSLIDEHeightfieldCurr = pCSLIDEHeightfield;
}

VOID CSLIDEParser::EndHeightfield()
{
  ASSERT( m_pCSLIDEHeightfieldCurr != NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == m_pCSLIDEHeightfieldCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDEHeightfieldCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEHeightfieldCurr );
  m_pCSLIDEHeightfieldCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;  
  m_pCSLIDENodeCurr = NULL;  
  m_pCSLIDEPrimitiveCurr = NULL;  
}

VOID CSLIDEParser::SetHeightfieldUMin(DynFloat *pdynfUMin)
{
  ASSERT( m_pCSLIDEHeightfieldCurr != NULL );

  pdynfUMin->SetValue(pdynfUMin->GetValue());
  m_pCSLIDEHeightfieldCurr->SetUMin(pdynfUMin);
  delete pdynfUMin;
}

VOID CSLIDEParser::SetHeightfieldUMax(DynFloat *pdynfUMax)
{
  ASSERT( m_pCSLIDEHeightfieldCurr != NULL );

  pdynfUMax->SetValue(pdynfUMax->GetValue());
  m_pCSLIDEHeightfieldCurr->SetUMax(pdynfUMax);
  delete pdynfUMax;
}

VOID CSLIDEParser::SetHeightfieldVMin(DynFloat *pdynfVMin)
{
  ASSERT( m_pCSLIDEHeightfieldCurr != NULL );

  pdynfVMin->SetValue(pdynfVMin->GetValue());
  m_pCSLIDEHeightfieldCurr->SetVMin(pdynfVMin);
  delete pdynfVMin;
}

VOID CSLIDEParser::SetHeightfieldVMax(DynFloat *pdynfVMax)
{
  ASSERT( m_pCSLIDEHeightfieldCurr != NULL );

  pdynfVMax->SetValue(pdynfVMax->GetValue());
  m_pCSLIDEHeightfieldCurr->SetVMax(pdynfVMax);
  delete pdynfVMax;
}

VOID CSLIDEParser::SetHeightfieldUSlices(DynInt *pdyniUSlices)
{
  ASSERT( m_pCSLIDEHeightfieldCurr != NULL );

  pdyniUSlices->SetValue(pdyniUSlices->GetValue());
  m_pCSLIDEHeightfieldCurr->SetUSlices(pdyniUSlices);
  delete pdyniUSlices;
}

VOID CSLIDEParser::SetHeightfieldVSlices(DynInt *pdyniVSlices)
{
  ASSERT( m_pCSLIDEHeightfieldCurr != NULL );

  pdyniVSlices->SetValue(pdyniVSlices->GetValue());
  m_pCSLIDEHeightfieldCurr->SetVSlices(pdyniVSlices);
  delete pdyniVSlices;
}

VOID CSLIDEParser::SetHeightfieldFunc(DynVector *pdynvFunc)
{
  ASSERT( m_pCSLIDEHeightfieldCurr != NULL );

  m_pCSLIDEHeightfieldCurr->SetFunc(pdynvFunc);
  delete pdynvFunc;
}

VOID CSLIDEParser::AddNurbPatch(CHAR *pcFile, UINT_32 uiLine, 
				   UINT_32 uiColumn,
				   CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDENurbPatch *pCSLIDENurbPatch;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == NULL );
  ASSERT( m_pCSLIDENurbPatchCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtNurbPatch],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDENurbPatch = CreateNurbPatch(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDENurbPatch != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDENurbPatch));
  m_pCSLIDEPrimitiveCurr = pCSLIDENurbPatch;
  m_pCSLIDENodeCurr = pCSLIDENurbPatch;
  m_pCSLIDEAtomNodeCurr = pCSLIDENurbPatch;
  m_pCSLIDENurbPatchCurr = pCSLIDENurbPatch;
}

VOID CSLIDEParser::EndNurbPatch()
{
  ASSERT( m_pCSLIDENurbPatchCurr != NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == m_pCSLIDENurbPatchCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDENurbPatchCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDENurbPatchCurr );
  m_pCSLIDENurbPatchCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;  
  m_pCSLIDENodeCurr = NULL;  
  m_pCSLIDEPrimitiveCurr = NULL;  
}

VOID CSLIDEParser::SetNurbPatchUOrder(DynInt *pdyniUOrder)
{
  ASSERT( m_pCSLIDENurbPatchCurr != NULL );

  pdyniUOrder->SetValue(pdyniUOrder->GetValue());
  m_pCSLIDENurbPatchCurr->SetUOrder(pdyniUOrder);
  delete pdyniUOrder;
}

VOID CSLIDEParser::SetNurbPatchUKnots(DynInt *pdyniUKnots)
{
  ASSERT( m_pCSLIDENurbPatchCurr != NULL );

  pdyniUKnots->SetValue(pdyniUKnots->GetValue());
  m_pCSLIDENurbPatchCurr->SetUKnots(pdyniUKnots);
  delete pdyniUKnots;
}

VOID CSLIDEParser::SetNurbPatchUKnotList(CListInt *plistUKnot)
{
  ASSERT( m_pCSLIDENurbPatchCurr != NULL );
  m_pCSLIDENurbPatchCurr->SetUKnotList(plistUKnot);
}

VOID CSLIDEParser::SetNurbPatchVOrder(DynInt *pdyniVOrder)
{
  ASSERT( m_pCSLIDENurbPatchCurr != NULL );

  pdyniVOrder->SetValue(pdyniVOrder->GetValue());
  m_pCSLIDENurbPatchCurr->SetVOrder(pdyniVOrder);
  delete pdyniVOrder;
}

VOID CSLIDEParser::SetNurbPatchVKnots(DynInt *pdyniVKnots)
{
  ASSERT( m_pCSLIDENurbPatchCurr != NULL );

  pdyniVKnots->SetValue(pdyniVKnots->GetValue());
  m_pCSLIDENurbPatchCurr->SetVKnots(pdyniVKnots);
  delete pdyniVKnots;
}

VOID CSLIDEParser::SetNurbPatchVKnotList(CListInt *plistVKnot)
{
  ASSERT( m_pCSLIDENurbPatchCurr != NULL );
  m_pCSLIDENurbPatchCurr->SetVKnotList(plistVKnot);
}

VOID CSLIDEParser::SetNurbPatchControlPointList(CListInt *plistControlPoint)
{
  ASSERT( m_pCSLIDENurbPatchCurr != NULL );
  m_pCSLIDENurbPatchCurr->SetControlPointList(plistControlPoint);
}

VOID CSLIDEParser::SetNurbPatchUSlices(DynInt *pdyniUSlices)
{
  ASSERT( m_pCSLIDENurbPatchCurr != NULL );

  pdyniUSlices->SetValue(pdyniUSlices->GetValue());
  m_pCSLIDENurbPatchCurr->SetUSlices(pdyniUSlices);
  delete pdyniUSlices;
}

VOID CSLIDEParser::SetNurbPatchVSlices(DynInt *pdyniVSlices)
{
  ASSERT( m_pCSLIDENurbPatchCurr != NULL );

  pdyniVSlices->SetValue(pdyniVSlices->GetValue());
  m_pCSLIDENurbPatchCurr->SetVSlices(pdyniVSlices);
  delete pdyniVSlices;
}

VOID CSLIDEParser::SetNurbPatchDrawControls(DynInt *pdyniDrawControls)
{
  ASSERT( m_pCSLIDENurbPatchCurr != NULL );

  m_pCSLIDENurbPatchCurr->SetDrawControls(pdyniDrawControls);
  delete pdyniDrawControls;
}

VOID CSLIDEParser::SetNurbPatchTexture(DynVector *pdynv0, DynVector *pdynv1, DynVector *pdynv2, DynVector *pdynv3)
{
  ASSERT( m_pCSLIDENurbPatchCurr != NULL );

  m_pCSLIDENurbPatchCurr->SetTexture(pdynv0, pdynv1, pdynv2, pdynv3);
  delete pdynv0;
  delete pdynv1;
  delete pdynv2;
  delete pdynv3;
}

VOID CSLIDEParser::SetNurbPatchOffset(DynFloat *pdynfOffset, DynInt *pdyniOffsetConnected)
{
  ASSERT( m_pCSLIDENurbPatchCurr != NULL );

  pdynfOffset->SetValue(pdynfOffset->GetValue());
  pdyniOffsetConnected->SetValue(pdyniOffsetConnected->GetValue());
  m_pCSLIDENurbPatchCurr->SetOffset(pdynfOffset, pdyniOffsetConnected);
  delete pdynfOffset;
  delete pdyniOffsetConnected;
}

VOID CSLIDEParser::AddBezierPatch(CHAR *pcFile, UINT_32 uiLine, 
				   UINT_32 uiColumn,
				   CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDEBezierPatch *pCSLIDEBezierPatch;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == NULL );
  ASSERT( m_pCSLIDEBezierPatchCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtBezierPatch],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDEBezierPatch = CreateBezierPatch(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDEBezierPatch != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDEBezierPatch));
  m_pCSLIDEPrimitiveCurr = pCSLIDEBezierPatch;
  m_pCSLIDENodeCurr = pCSLIDEBezierPatch;
  m_pCSLIDEAtomNodeCurr = pCSLIDEBezierPatch;
  m_pCSLIDEBezierPatchCurr = pCSLIDEBezierPatch;
}

VOID CSLIDEParser::EndBezierPatch()
{
  ASSERT( m_pCSLIDEBezierPatchCurr != NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == m_pCSLIDEBezierPatchCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDEBezierPatchCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEBezierPatchCurr );
  m_pCSLIDEBezierPatchCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;  
  m_pCSLIDENodeCurr = NULL;  
  m_pCSLIDEPrimitiveCurr = NULL;  
}

VOID CSLIDEParser::SetBezierPatchUOrder(DynInt *pdyniUOrder)
{
  ASSERT( m_pCSLIDEBezierPatchCurr != NULL );

  pdyniUOrder->SetValue(pdyniUOrder->GetValue());
  m_pCSLIDEBezierPatchCurr->SetUOrder(pdyniUOrder);
  delete pdyniUOrder;
}

VOID CSLIDEParser::SetBezierPatchVOrder(DynInt *pdyniVOrder)
{
  ASSERT( m_pCSLIDEBezierPatchCurr != NULL );

  pdyniVOrder->SetValue(pdyniVOrder->GetValue());
  m_pCSLIDEBezierPatchCurr->SetVOrder(pdyniVOrder);
  delete pdyniVOrder;
}

VOID CSLIDEParser::SetBezierPatchControlPointList(CListInt *plistControlPoint)
{
  ASSERT( m_pCSLIDEBezierPatchCurr != NULL );
  m_pCSLIDEBezierPatchCurr->SetControlPointList(plistControlPoint);
}

VOID CSLIDEParser::SetBezierPatchUSlices(DynInt *pdyniUSlices)
{
  ASSERT( m_pCSLIDEBezierPatchCurr != NULL );

  pdyniUSlices->SetValue(pdyniUSlices->GetValue());
  m_pCSLIDEBezierPatchCurr->SetUSlices(pdyniUSlices);
  delete pdyniUSlices;
}

VOID CSLIDEParser::SetBezierPatchVSlices(DynInt *pdyniVSlices)
{
  ASSERT( m_pCSLIDEBezierPatchCurr != NULL );

  pdyniVSlices->SetValue(pdyniVSlices->GetValue());
  m_pCSLIDEBezierPatchCurr->SetVSlices(pdyniVSlices);
  delete pdyniVSlices;
}

VOID CSLIDEParser::SetBezierPatchDrawControls(DynInt *pdyniDrawControls)
{
  ASSERT( m_pCSLIDEBezierPatchCurr != NULL );

  m_pCSLIDEBezierPatchCurr->SetDrawControls(pdyniDrawControls);
  delete pdyniDrawControls;
}

VOID CSLIDEParser::SetBezierPatchTexture(DynVector *pdynv0, DynVector *pdynv1, DynVector *pdynv2, DynVector *pdynv3)
{
  ASSERT( m_pCSLIDEBezierPatchCurr != NULL );

  m_pCSLIDEBezierPatchCurr->SetTexture(pdynv0, pdynv1, pdynv2, pdynv3);
  delete pdynv0;
  delete pdynv1;
  delete pdynv2;
  delete pdynv3;
}

VOID CSLIDEParser::SetBezierPatchOffset(DynFloat *pdynfOffset, DynInt *pdyniOffsetConnected)
{
  ASSERT( m_pCSLIDEBezierPatchCurr != NULL );

  pdynfOffset->SetValue(pdynfOffset->GetValue());
  pdyniOffsetConnected->SetValue(pdyniOffsetConnected->GetValue());
  m_pCSLIDEBezierPatchCurr->SetOffset(pdynfOffset, pdyniOffsetConnected);
  delete pdynfOffset;
  delete pdyniOffsetConnected;
}

VOID CSLIDEParser::AddBSplineCurve(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
				   CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDEBSplineCurve *pCSLIDEBSplineCurve;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == NULL );
  ASSERT( m_pCSLIDEBSplineCurveCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtBSplineCurve],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDEBSplineCurve = CreateBSplineCurve(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDEBSplineCurve != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDEBSplineCurve));
  m_pCSLIDEPrimitiveCurr = pCSLIDEBSplineCurve;
  m_pCSLIDENodeCurr = pCSLIDEBSplineCurve;
  m_pCSLIDEAtomNodeCurr = pCSLIDEBSplineCurve;
  m_pCSLIDEBSplineCurveCurr = pCSLIDEBSplineCurve;
}

VOID CSLIDEParser::EndBSplineCurve()
{
  ASSERT( m_pCSLIDEBSplineCurveCurr != NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == m_pCSLIDEBSplineCurveCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDEBSplineCurveCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEBSplineCurveCurr );
  m_pCSLIDEBSplineCurveCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;  
  m_pCSLIDENodeCurr = NULL;  
  m_pCSLIDEPrimitiveCurr = NULL;  
}

VOID CSLIDEParser::SetBSplineCurveType(UINT_32 uiType)
{
  m_pCSLIDEBSplineCurveCurr->SetType(uiType);
}


VOID CSLIDEParser::SetBSplineCurveOrder(DynInt *pdyniOrder)
{
  ASSERT( m_pCSLIDEBSplineCurveCurr != NULL );

  pdyniOrder->SetValue(pdyniOrder->GetValue());
  m_pCSLIDEBSplineCurveCurr->SetOrder(pdyniOrder);
  delete pdyniOrder;
}

VOID CSLIDEParser::SetBSplineCurveKnotList(CListInt *plistKnot)
{
  ASSERT( m_pCSLIDEBSplineCurveCurr != NULL );
  m_pCSLIDEBSplineCurveCurr->SetKnotList(plistKnot);
}

VOID CSLIDEParser::SetBSplineCurveControlPointList(CListInt *plistControlPoint)
{
  ASSERT( m_pCSLIDEBSplineCurveCurr != NULL );
  m_pCSLIDEBSplineCurveCurr->SetControlPointList(plistControlPoint);
}

VOID CSLIDEParser::SetBSplineCurveBezierControlPointList(CListInt *plistControlPoint)
{
  ASSERT( m_pCSLIDEBSplineCurveCurr != NULL );
  m_pCSLIDEBSplineCurveCurr->SetBezierControlPointList(plistControlPoint);
}

VOID CSLIDEParser::SetBSplineCurveClosed(DynInt *pdyniClosed)
{
  ASSERT( m_pCSLIDEBSplineCurveCurr != NULL );

  pdyniClosed->SetValue(pdyniClosed->GetValue());
  m_pCSLIDEBSplineCurveCurr->SetCurveClosed(pdyniClosed);
  delete pdyniClosed;
}

VOID CSLIDEParser::SetBSplineCurveSlices(DynInt *pdyniSlices)
{
  ASSERT( m_pCSLIDEBSplineCurveCurr != NULL );

  pdyniSlices->SetValue(pdyniSlices->GetValue());
  m_pCSLIDEBSplineCurveCurr->SetSlices(pdyniSlices);
  delete pdyniSlices;
}

VOID CSLIDEParser::SetBSplineCurveDrawControls(DynInt *pdyniDrawControls)
{
  ASSERT( m_pCSLIDEBSplineCurveCurr != NULL );

  m_pCSLIDEBSplineCurveCurr->SetDrawControls(pdyniDrawControls);
  delete pdyniDrawControls;
}

VOID CSLIDEParser::AddPolyline(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
				   CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDEPolyline *pCSLIDEPolyline;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == NULL );
  ASSERT( m_pCSLIDEPolylineCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtPolyline],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDEPolyline = CreatePolyline(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDEPolyline != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDEPolyline));
  m_pCSLIDEPrimitiveCurr = pCSLIDEPolyline;
  m_pCSLIDENodeCurr = pCSLIDEPolyline;
  m_pCSLIDEAtomNodeCurr = pCSLIDEPolyline;
  m_pCSLIDEPolylineCurr = pCSLIDEPolyline;
}

VOID CSLIDEParser::EndPolyline()
{
  ASSERT( m_pCSLIDEPolylineCurr != NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == m_pCSLIDEPolylineCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDEPolylineCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEPolylineCurr );
  m_pCSLIDEPolylineCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;  
  m_pCSLIDENodeCurr = NULL;  
  m_pCSLIDEPrimitiveCurr = NULL;  
}

VOID CSLIDEParser::SetPolylinePointList(CListInt *plistPoint)
{
  ASSERT( m_pCSLIDEPolylineCurr != NULL );
  m_pCSLIDEPolylineCurr->SetPointList(plistPoint);
}

VOID CSLIDEParser::SetPolylineDrawPoints(DynInt *pdyniDrawPoints)
{
  ASSERT( m_pCSLIDEPolylineCurr != NULL );

  pdyniDrawPoints->SetValue(pdyniDrawPoints->GetValue());
  m_pCSLIDEPolylineCurr->SetDrawPoints(pdyniDrawPoints);
  delete pdyniDrawPoints;
}

VOID CSLIDEParser::AddCrossSection(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
				   CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDECrossSection *pCSLIDECrossSection;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == NULL );
  ASSERT( m_pCSLIDECrossSectionCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtCrossSection],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDECrossSection = CreateCrossSection(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDECrossSection != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDECrossSection));
  m_pCSLIDEPrimitiveCurr = pCSLIDECrossSection;
  m_pCSLIDENodeCurr = pCSLIDECrossSection;
  m_pCSLIDEAtomNodeCurr = pCSLIDECrossSection;
  m_pCSLIDECrossSectionCurr = pCSLIDECrossSection;
}

VOID CSLIDEParser::EndCrossSection()
{
  ASSERT( m_pCSLIDECrossSectionCurr != NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == m_pCSLIDECrossSectionCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDECrossSectionCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDECrossSectionCurr );
  m_pCSLIDECrossSectionCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;  
  m_pCSLIDENodeCurr = NULL;  
  m_pCSLIDEPrimitiveCurr = NULL;  
}

VOID CSLIDEParser::SetCrossSectionSurfaceID(CHAR *pcSurfaceID)
{
  ASSERT( m_pCSLIDECrossSectionCurr != NULL );

  m_pCSLIDECrossSectionCurr->SetSurfaceID(pcSurfaceID);
}

VOID CSLIDEParser::SetCrossSectionType(UINT_32 uiType)
{
  ASSERT( m_pCSLIDECrossSectionCurr != NULL );

  m_pCSLIDECrossSectionCurr->SetType(uiType);
}

VOID CSLIDEParser::SetCrossSectionID(CHAR *pcID)
{
  ASSERT( m_pCSLIDECrossSectionCurr != NULL );

  m_pCSLIDECrossSectionCurr->SetID(pcID);
}

VOID CSLIDEParser::SetCrossSectionRadius(DynFloat *pdynfRadius)
{
  ASSERT( m_pCSLIDECrossSectionCurr != NULL );

  m_pCSLIDECrossSectionCurr->SetRadius(pdynfRadius);
  delete pdynfRadius;
}

VOID CSLIDEParser::SetCrossSectionInnerRadius(DynFloat *pdynfInnerRadius)
{
  ASSERT( m_pCSLIDECrossSectionCurr != NULL );

  m_pCSLIDECrossSectionCurr->SetInnerRadius(pdynfInnerRadius);
  delete pdynfInnerRadius;
}

VOID CSLIDEParser::SetCrossSectionThickness(DynFloat *pdynfThickness)
{
  ASSERT( m_pCSLIDECrossSectionCurr != NULL );

  m_pCSLIDECrossSectionCurr->SetThickness(pdynfThickness);
  delete pdynfThickness;
}

VOID CSLIDEParser::SetCrossSectionTheta(DynFloat *pdynfTheta)
{
  ASSERT( m_pCSLIDECrossSectionCurr != NULL );

  m_pCSLIDECrossSectionCurr->SetTheta(pdynfTheta);
  delete pdynfTheta;
}

VOID CSLIDEParser::SetCrossSectionSlices(DynInt *pdyniCrossSectionSlices)
{
  ASSERT( m_pCSLIDECrossSectionCurr != NULL );

  pdyniCrossSectionSlices->SetValue(pdyniCrossSectionSlices->GetValue());
  m_pCSLIDECrossSectionCurr->SetSlices(pdyniCrossSectionSlices);
  delete pdyniCrossSectionSlices;
}

VOID CSLIDEParser::AddSweep(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDESweep *pCSLIDESweep;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == NULL );
  ASSERT( m_pCSLIDESweepCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtSweep],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDESweep = CreateSweep(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDESweep != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDESweep));
  m_pCSLIDEPrimitiveCurr = pCSLIDESweep;
  m_pCSLIDENodeCurr = pCSLIDESweep;
  m_pCSLIDEAtomNodeCurr = pCSLIDESweep;
  m_pCSLIDESweepCurr = pCSLIDESweep;
}

VOID CSLIDEParser::EndSweep()
{
  ASSERT( m_pCSLIDESweepCurr != NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == m_pCSLIDESweepCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDESweepCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDESweepCurr );
  m_pCSLIDESweepCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;  
  m_pCSLIDENodeCurr = NULL;  
  m_pCSLIDEPrimitiveCurr = NULL;  
}

VOID CSLIDEParser::EndSweepCrossSection()
{
  ASSERT( m_pCSLIDESweepCrossSectionCurr != NULL );

  m_pCSLIDESweepCrossSectionCurr = NULL;  
}

VOID CSLIDEParser::SetSweepPathID(CHAR *pcID)
{
  ASSERT( m_pCSLIDESweepCurr != NULL );

  m_pCSLIDESweepCurr->SetPathID(pcID);
}

VOID CSLIDEParser::SetSweepPathMinimizeTorsion(DynInt *pdyniMinimizeTorsion)
{
  ASSERT( m_pCSLIDESweepCurr != NULL );

  pdyniMinimizeTorsion->SetValue(pdyniMinimizeTorsion->GetValue());
  m_pCSLIDESweepCurr->SetPathMinimizeTorsion(pdyniMinimizeTorsion);
  delete pdyniMinimizeTorsion;
}

VOID CSLIDEParser::SetSweepPathClosed(DynInt *pdyniClosed)
{
  ASSERT( m_pCSLIDESweepCurr != NULL );

  pdyniClosed->SetValue(pdyniClosed->GetValue());
  m_pCSLIDESweepCurr->SetPathClosed(pdyniClosed);
  delete pdyniClosed;
}

VOID CSLIDEParser::SetSweepPathSymmetry(DynInt *pdyniSymmetry)
{
  ASSERT( m_pCSLIDESweepCurr != NULL );

  m_pCSLIDESweepCurr->SetPathSymmetry(pdyniSymmetry);
  delete pdyniSymmetry;
}

VOID CSLIDEParser::SetSweepPathAzimuth(DynFloat *pdynfAzimuth)
{
  ASSERT( m_pCSLIDESweepCurr != NULL );

  pdynfAzimuth->SetValue(pdynfAzimuth->GetValue());
  m_pCSLIDESweepCurr->SetPathAzimuth(pdynfAzimuth);
  delete pdynfAzimuth;
}

VOID CSLIDEParser::SetSweepPathTwist(DynFloat *pdynfTwist)
{
  ASSERT( m_pCSLIDESweepCurr != NULL );

  pdynfTwist->SetValue(pdynfTwist->GetValue());
  m_pCSLIDESweepCurr->SetPathTwist(pdynfTwist);
  delete pdynfTwist;
}

VOID CSLIDEParser::AddSweepCrossSectionID(CHAR *pcID)
{
  ASSERT( m_pCSLIDESweepCurr != NULL );

  // Create SweepCrossSection
  CSLIDESweepCrossSection *pCSLIDESweepCS;

  pCSLIDESweepCS = CreateSweepCrossSection("[SLIDE generated]", 0, 0, pcID);
  ASSERT( pCSLIDESweepCS != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDESweepCS));

  m_pCSLIDESweepCrossSectionCurr = pCSLIDESweepCS;



  m_pCSLIDESweepCurr->AddCrossSectionID(pcID, pCSLIDESweepCS);
}

VOID CSLIDEParser::SetSweepCrossSectionSurfaceID(CHAR *pcID)
{
  ASSERT( m_pCSLIDESweepCrossSectionCurr != NULL );

  m_pCSLIDESweepCrossSectionCurr->SetSurfaceID(pcID);
}

VOID CSLIDEParser::SetSweepCrossSectionTexture(DynVector *pdynv0, DynVector *pdynv1, 
					       DynVector *pdynv2, DynVector *pdynv3)
{
  ASSERT( m_pCSLIDESweepCrossSectionCurr != NULL );

  m_pCSLIDESweepCrossSectionCurr->SetTexture(pdynv0, pdynv1, pdynv2, pdynv3);
  delete pdynv0;
  delete pdynv1;
  delete pdynv2;
  delete pdynv3;
}

VOID CSLIDEParser::SetSweepCrossSectionEndCap(DynInt *pdyniEndCap)
{
  ASSERT( m_pCSLIDESweepCrossSectionCurr != NULL );

  m_pCSLIDESweepCrossSectionCurr->SetEndCap(pdyniEndCap);
  delete pdyniEndCap;
}

VOID CSLIDEParser::SetSweepCrossSectionBeginCap(DynInt *pdyniBeginCap)
{
  ASSERT( m_pCSLIDESweepCrossSectionCurr != NULL );

  m_pCSLIDESweepCrossSectionCurr->SetBeginCap(pdyniBeginCap);
  delete pdyniBeginCap;
}

VOID CSLIDEParser::SetSweepCrossSectionOrientation(DynInt *pdyniOrientation)
{
  ASSERT( m_pCSLIDESweepCrossSectionCurr != NULL );

  m_pCSLIDESweepCrossSectionCurr->SetOrientation(pdyniOrientation);
  delete pdyniOrientation;
}

VOID CSLIDEParser::AddSweepCrossSectionTranslate(CHAR *pcFile, UINT_32 uiLine, 
						 UINT_32 uiColumn,
						 DynVector *pdynv)
{
  ASSERT( m_pCSLIDESweepCrossSectionCurr != NULL );

  CSLIDETransform *pCSLIDETransform;

  if( pcFile == NULL )
    pCSLIDETransform = CreateTranslate("[SLIDE sweep]", uiLine, uiColumn, pdynv);
  else
    pCSLIDETransform = CreateTranslate(pcFile, uiLine, uiColumn, pdynv);

  ASSERT( pCSLIDETransform != NULL );

  VERIFY(m_pCSLIDEWorld->InsertTransform(pCSLIDETransform));
  m_pCSLIDESweepCrossSectionCurr->InsertTransform(pCSLIDETransform);
}

VOID CSLIDEParser::AddSweepCrossSectionScale(CHAR *pcFile, UINT_32 uiLine, 
					     UINT_32 uiColumn,
					     DynVector *pdynv)
{
  ASSERT( m_pCSLIDESweepCrossSectionCurr != NULL );

  CSLIDETransform *pCSLIDETransform;

  if( pcFile == NULL )
    pCSLIDETransform = CreateScale("[SLIDE sweep]", uiLine, uiColumn, pdynv);
  else
    pCSLIDETransform = CreateScale(pcFile, uiLine, uiColumn, pdynv);

  ASSERT( pCSLIDETransform != NULL );

  VERIFY(m_pCSLIDEWorld->InsertTransform(pCSLIDETransform));
  m_pCSLIDESweepCrossSectionCurr->InsertTransform(pCSLIDETransform);
}

VOID CSLIDEParser::AddSweepCrossSectionRotate(CHAR *pcFile, UINT_32 uiLine, 
					      UINT_32 uiColumn,
					      DynVector *pdynv, DynFloat *pdynfAngle)
{
  ASSERT( m_pCSLIDESweepCrossSectionCurr != NULL );

  CSLIDETransform *pCSLIDETransform;

  if( pcFile == NULL )
    pCSLIDETransform = CreateRotate("[SLIDE sweep]", uiLine, uiColumn, pdynv, pdynfAngle);
  else
    pCSLIDETransform = CreateRotate(pcFile, uiLine, uiColumn, pdynv, pdynfAngle);

  ASSERT( pCSLIDETransform != NULL );

  VERIFY(m_pCSLIDEWorld->InsertTransform(pCSLIDETransform));
  m_pCSLIDESweepCrossSectionCurr->InsertTransform(pCSLIDETransform);
}

VOID CSLIDEParser::SetSweepDrawSweep(DynInt *pdyniDrawSweep)
{
  ASSERT( m_pCSLIDESweepCurr != NULL );

  m_pCSLIDESweepCurr->SetDrawSweep(pdyniDrawSweep);
  delete pdyniDrawSweep;
}

VOID CSLIDEParser::SetSweepDrawPath(DynInt *pdyniDrawPath)
{
  ASSERT( m_pCSLIDESweepCurr != NULL );

  m_pCSLIDESweepCurr->SetDrawPath(pdyniDrawPath);
  delete pdyniDrawPath;
}

VOID CSLIDEParser::SetSweepMeshType(DynInt *pdyniMeshType)
{
  ASSERT( m_pCSLIDESweepCurr != NULL );

  m_pCSLIDESweepCurr->SetMeshType(pdyniMeshType);
  delete pdyniMeshType;
}

VOID CSLIDEParser::AddSweepMorph(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDESweepMorph *pCSLIDESweepMorph;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == NULL );
  ASSERT( m_pCSLIDESweepMorphCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtSweepMorph],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDESweepMorph = CreateSweepMorph(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDESweepMorph != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDESweepMorph));
  m_pCSLIDEPrimitiveCurr = pCSLIDESweepMorph;
  m_pCSLIDENodeCurr = pCSLIDESweepMorph;
  m_pCSLIDEAtomNodeCurr = pCSLIDESweepMorph;
  m_pCSLIDESweepMorphCurr = pCSLIDESweepMorph;
}

VOID CSLIDEParser::EndSweepMorph()
{
  ASSERT( m_pCSLIDESweepMorphCurr != NULL );
  ASSERT( m_pCSLIDEAtomNodeCurr == m_pCSLIDESweepMorphCurr );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDESweepMorphCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDESweepMorphCurr );
  m_pCSLIDESweepMorphCurr = NULL;
  m_pCSLIDEAtomNodeCurr = NULL;  
  m_pCSLIDENodeCurr = NULL;  
  m_pCSLIDEPrimitiveCurr = NULL;  
}

VOID CSLIDEParser::SetSweepMorphPathID(CHAR *pcID)
{
  ASSERT( m_pCSLIDESweepMorphCurr != NULL );

  m_pCSLIDESweepMorphCurr->SetPathID(pcID);
}

VOID CSLIDEParser::SetSweepMorphPathMinimizeTorsion(DynInt *pdyniMinimizeTorsion)
{
  ASSERT( m_pCSLIDESweepMorphCurr != NULL );

  pdyniMinimizeTorsion->SetValue(pdyniMinimizeTorsion->GetValue());
  m_pCSLIDESweepMorphCurr->SetPathMinimizeTorsion(pdyniMinimizeTorsion);
  delete pdyniMinimizeTorsion;
}

VOID CSLIDEParser::SetSweepMorphPathClosed(DynInt *pdyniClosed)
{
  ASSERT( m_pCSLIDESweepMorphCurr != NULL );

  pdyniClosed->SetValue(pdyniClosed->GetValue());
  m_pCSLIDESweepMorphCurr->SetPathClosed(pdyniClosed);
  delete pdyniClosed;
}

VOID CSLIDEParser::SetSweepMorphPathSymmetry(DynInt *pdyniSymmetry)
{
  ASSERT( m_pCSLIDESweepMorphCurr != NULL );

  m_pCSLIDESweepMorphCurr->SetPathSymmetry(pdyniSymmetry);
  delete pdyniSymmetry;
}

VOID CSLIDEParser::SetSweepMorphPathAzimuth(DynFloat *pdynfAzimuth)
{
  ASSERT( m_pCSLIDESweepMorphCurr != NULL );

  pdynfAzimuth->SetValue(pdynfAzimuth->GetValue());
  m_pCSLIDESweepMorphCurr->SetPathAzimuth(pdynfAzimuth);
  delete pdynfAzimuth;
}

VOID CSLIDEParser::SetSweepMorphPathTwist(DynFloat *pdynfTwist)
{
  ASSERT( m_pCSLIDESweepMorphCurr != NULL );

  pdynfTwist->SetValue(pdynfTwist->GetValue());
  m_pCSLIDESweepMorphCurr->SetPathTwist(pdynfTwist);
  delete pdynfTwist;
}

VOID CSLIDEParser::SetSweepMorphMeshType(DynInt *pdyniMeshType)
{
  ASSERT( m_pCSLIDESweepMorphCurr != NULL );

  m_pCSLIDESweepMorphCurr->SetMeshType(pdyniMeshType);
  delete pdyniMeshType;
}

VOID CSLIDEParser::SetSweepMorphDrawSweep(DynInt *pdyniDrawSweep)
{
  ASSERT( m_pCSLIDESweepMorphCurr != NULL );

  m_pCSLIDESweepMorphCurr->SetDrawSweep(pdyniDrawSweep);
  delete pdyniDrawSweep;
}

VOID CSLIDEParser::SetSweepMorphDrawPath(DynInt *pdyniDrawPath)
{
  ASSERT( m_pCSLIDESweepMorphCurr != NULL );

  m_pCSLIDESweepMorphCurr->SetDrawPath(pdyniDrawPath);
  delete pdyniDrawPath;
}

VOID CSLIDEParser::SetSweepMorphEndCap(DynInt *pdyniEndCap)
{
  ASSERT( m_pCSLIDESweepMorphCurr != NULL );

  m_pCSLIDESweepMorphCurr->SetEndCap(pdyniEndCap);
  delete pdyniEndCap;
}

VOID CSLIDEParser::SetSweepMorphBeginCap(DynInt *pdyniBeginCap)
{
  ASSERT( m_pCSLIDESweepMorphCurr != NULL );

  m_pCSLIDESweepMorphCurr->SetBeginCap(pdyniBeginCap);
  delete pdyniBeginCap;
}

VOID CSLIDEParser::AddFace(CHAR *pcFile, UINT_32 uiLine, 
			   UINT_32 uiColumn,
			   CHAR *pcID, 
			   CListInt *plistPointIDs)
{
  CSLIDEFace *pCSLIDEFace;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDEFaceCurr == NULL );

  pCSLIDEFace = m_pCSLIDEWorld->PeekFace(pcID);
  if ( pCSLIDEFace != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld",  
		  gcppcSLIDEReservedWords[sliderwtFace],
		  pcID, pCSLIDEFace->GetFileName(), 
		  pCSLIDEFace->GetFileLine(), 
		  pCSLIDEFace->GetFileColumn());
    }

  pCSLIDEFace = CreateFace(pcFile, uiLine, uiColumn, 
			   pcID, plistPointIDs);
  ASSERT( pCSLIDEFace != NULL );

  VERIFY(m_pCSLIDEWorld->InsertFace(pCSLIDEFace));
  m_pCSLIDEPrimitiveCurr = pCSLIDEFace;
  m_pCSLIDEFaceCurr = pCSLIDEFace;
}

VOID CSLIDEParser::EndFace()
{
  ASSERT( m_pCSLIDEFaceCurr != NULL );
  ASSERT( m_pCSLIDEFaceCurr == m_pCSLIDEPrimitiveCurr );
  m_pCSLIDEFaceCurr = NULL;
  m_pCSLIDEPrimitiveCurr = NULL;
}

VOID CSLIDEParser::SetFaceSurfaceID(CHAR *pcSurfaceID)
{
  ASSERT( m_pCSLIDEFaceCurr != NULL );

  m_pCSLIDEFaceCurr->SetSurfaceID(pcSurfaceID);
}

VOID CSLIDEParser::AddPoint(CHAR *pcFile, UINT_32 uiLine, 
			    UINT_32 uiColumn, CHAR *pcID,
			    DynPoint *pdynpt)
{
  CSLIDEPoint *pCSLIDEPoint;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDEPointCurr == NULL );

  pCSLIDEPoint = m_pCSLIDEWorld->PeekPoint(pcID);
  if ( pCSLIDEPoint != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtPoint],
		  pcID, pCSLIDEPoint->GetFileName(), 
		  pCSLIDEPoint->GetFileLine(), 
		  pCSLIDEPoint->GetFileColumn());
    }

  pCSLIDEPoint = CreatePoint(pcFile, uiLine, uiColumn,
			     pcID, pdynpt);
  ASSERT( pCSLIDEPoint != NULL );

  VERIFY(m_pCSLIDEWorld->InsertPoint(pCSLIDEPoint));
  m_pCSLIDEPrimitiveCurr = pCSLIDEPoint;
  m_pCSLIDEPointCurr = pCSLIDEPoint;
}

VOID CSLIDEParser::EndPoint()
{
  ASSERT( m_pCSLIDEPointCurr != NULL );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEPointCurr );
  m_pCSLIDEPointCurr = NULL;
  m_pCSLIDEPrimitiveCurr = NULL;
}

VOID CSLIDEParser::SetPointSurfaceID(CHAR *pcSurfaceID)
{
  ASSERT( m_pCSLIDEPointCurr != NULL );

  m_pCSLIDEPointCurr->SetSurfaceID(pcSurfaceID);
}

VOID CSLIDEParser::SetPointNormal(DynVector *pdynv)
{
  ASSERT( m_pCSLIDEPointCurr != NULL );

  m_pCSLIDEPointCurr->SetNormal(pdynv);
  delete pdynv;
}

VOID CSLIDEParser::SetPointTexture(DynPoint *pdynpt)
{
  ASSERT( m_pCSLIDEPointCurr != NULL );

  m_pCSLIDEPointCurr->SetTexture(pdynpt);
  delete pdynpt;
}

VOID CSLIDEParser::AddControlPoint(CHAR *pcFile, UINT_32 uiLine, 
			    UINT_32 uiColumn, CHAR *pcID)
{
  CSLIDEControlPoint *pCSLIDEControlPoint;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDEControlPointCurr == NULL );

  pCSLIDEControlPoint = m_pCSLIDEWorld->PeekControlPoint(pcID);
  if ( pCSLIDEControlPoint != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtControlPoint],
		  pcID, pCSLIDEControlPoint->GetFileName(), 
		  pCSLIDEControlPoint->GetFileLine(), 
		  pCSLIDEControlPoint->GetFileColumn());
    }

  pCSLIDEControlPoint = CreateControlPoint(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDEControlPoint != NULL );

  VERIFY(m_pCSLIDEWorld->InsertControlPoint(pCSLIDEControlPoint));
  m_pCSLIDEPrimitiveCurr = pCSLIDEControlPoint;
  m_pCSLIDEControlPointCurr = pCSLIDEControlPoint;
}

VOID CSLIDEParser::EndControlPoint()
{
  ASSERT( m_pCSLIDEControlPointCurr != NULL );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEControlPointCurr );
  m_pCSLIDEControlPointCurr = NULL;
  m_pCSLIDEPrimitiveCurr = NULL;
}

VOID CSLIDEParser::SetControlPointPointID(CHAR *pcPointID)
{
  ASSERT( m_pCSLIDEControlPointCurr != NULL );

  m_pCSLIDEControlPointCurr->SetPointID(pcPointID);
}

VOID CSLIDEParser::SetControlPointCrossSectionID(CHAR *pcCrossSectionID)
{
  ASSERT( m_pCSLIDEControlPointCurr != NULL );

  m_pCSLIDEControlPointCurr->SetCrossSectionID(pcCrossSectionID);
}

VOID CSLIDEParser::SetControlPointScale(DynVector *pdynv)
{
  ASSERT( m_pCSLIDEControlPointCurr != NULL );

  m_pCSLIDEControlPointCurr->SetScale(pdynv);
  delete pdynv;
}

VOID CSLIDEParser::SetControlPointRotate(DynVector *pdynv)
{
  ASSERT( m_pCSLIDEControlPointCurr != NULL );

  m_pCSLIDEControlPointCurr->SetRotate(pdynv);
  delete pdynv;
}

VOID CSLIDEParser::AddSurface(CHAR *pcFile, UINT_32 uiLine, 
			      UINT_32 uiColumn,
			      CHAR *pcID)
{
  CSLIDESurface *pCSLIDESurface;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDESurfaceCurr == NULL );

  pCSLIDESurface = m_pCSLIDEWorld->PeekSurface(pcID);
  if ( pCSLIDESurface != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtSurface],
		  pcID, pCSLIDESurface->GetFileName(), 
		  pCSLIDESurface->GetFileLine(), 
		  pCSLIDESurface->GetFileColumn());
    }

  pCSLIDESurface = CreateSurface(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDESurface != NULL );

  VERIFY(m_pCSLIDEWorld->InsertSurface(pCSLIDESurface));
  m_pCSLIDEPrimitiveCurr = pCSLIDESurface;
  m_pCSLIDESurfaceCurr = pCSLIDESurface;
}

VOID CSLIDEParser::EndSurface()
{
  ASSERT( m_pCSLIDESurfaceCurr != NULL );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDESurfaceCurr );
  m_pCSLIDESurfaceCurr = NULL;
  m_pCSLIDEPrimitiveCurr = NULL;
}

VOID CSLIDEParser::SetSurfaceColor(DynVector *pdynv)
{
  ASSERT( m_pCSLIDESurfaceCurr != NULL );

  m_pCSLIDESurfaceCurr->SetColor(pdynv);
  delete pdynv;
}

VOID CSLIDEParser::SetSurfaceReflectivity(DynVector *pdynv)
{
  ASSERT( m_pCSLIDESurfaceCurr != NULL );

  m_pCSLIDESurfaceCurr->SetReflectivity(pdynv);
  delete pdynv;
}

VOID CSLIDEParser::SetSurfaceExponent(DynFloat *pdynfExponent)
{
  ASSERT( m_pCSLIDESurfaceCurr != NULL );

  m_pCSLIDESurfaceCurr->SetExponent(pdynfExponent);
  delete pdynfExponent;
}

VOID CSLIDEParser::SetSurfaceMetallic(DynFloat *pdynfMetallic)
{
  ASSERT( m_pCSLIDESurfaceCurr != NULL );

  m_pCSLIDESurfaceCurr->SetMetallic(pdynfMetallic);
  delete pdynfMetallic;
}

VOID CSLIDEParser::SetSurfaceTexture(CHAR *pcTextureFile, DynInt *pdyniTexture)
{
  ASSERT( m_pCSLIDESurfaceCurr != NULL );

  m_pCSLIDESurfaceCurr->SetTexture(pcTextureFile, pdyniTexture);
  delete pdyniTexture;
}

VOID CSLIDEParser::AddWormwarp(CHAR *pcFile, UINT_32 uiLine, 
			      UINT_32 uiColumn,
			      CHAR *pcID)
{
  CSLIDENode *pCSLIDENode;
  CSLIDEWormwarp *pCSLIDEWormwarp;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDENodeCurr == NULL );
  ASSERT( m_pCSLIDEWormwarpCurr == NULL );

  pCSLIDENode = m_pCSLIDEWorld->PeekNode(pcID);
  if ( pCSLIDENode != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtWormwarp],
		  pcID, pCSLIDENode->GetFileName(), 
		  pCSLIDENode->GetFileLine(), 
		  pCSLIDENode->GetFileColumn());
    }

  pCSLIDEWormwarp = CreateWormwarp(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDEWormwarp != NULL );

  VERIFY(m_pCSLIDEWorld->InsertNode(pCSLIDEWormwarp));
  m_pCSLIDEPrimitiveCurr = pCSLIDEWormwarp;
  m_pCSLIDENodeCurr = pCSLIDEWormwarp;
  m_pCSLIDEWormwarpCurr = pCSLIDEWormwarp;
}

VOID CSLIDEParser::EndWormwarp()
{
  ASSERT( m_pCSLIDEWormwarpCurr != NULL );
  ASSERT( m_pCSLIDENodeCurr == m_pCSLIDEWormwarpCurr );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEWormwarpCurr );
  m_pCSLIDEWormwarpCurr = NULL;
  m_pCSLIDENodeCurr = NULL;  
  m_pCSLIDEPrimitiveCurr = NULL;  
}

VOID CSLIDEParser::SetWormwarpGroupID(CHAR *pcGroupID)
{
  ASSERT( m_pCSLIDEWormwarpCurr != NULL );

  m_pCSLIDEWormwarpCurr->SetGroupID(pcGroupID);
}

VOID CSLIDEParser::SetWormwarpPathID(CHAR *pcPathID)
{
  ASSERT( m_pCSLIDEWormwarpCurr != NULL );

  m_pCSLIDEWormwarpCurr->SetPathID(pcPathID);
}

VOID CSLIDEParser::SetWormwarpInstances(DynFloat *pdynfInstances)
{
  ASSERT( m_pCSLIDEWormwarpCurr != NULL );

  m_pCSLIDEWormwarpCurr->SetInstances(pdynfInstances);
}

VOID CSLIDEParser::SetWormwarpType(DynInt *pdyniType)
{
  ASSERT( m_pCSLIDEWormwarpCurr != NULL );
  m_pCSLIDEWormwarpCurr->SetType(pdyniType);
}

VOID CSLIDEParser::SetWormwarpPathOnOff(DynInt *pdyniPathOnOff)
{
  ASSERT( m_pCSLIDEWormwarpCurr != NULL );
  m_pCSLIDEWormwarpCurr->SetPathOnOff(pdyniPathOnOff);
}

VOID CSLIDEParser::AddFog(CHAR *pcFile, UINT_32 uiLine, 
			      UINT_32 uiColumn,
			      CHAR *pcID)
{
  CSLIDEFog *pCSLIDEFog;

  ASSERT( m_pCSLIDEPrimitiveCurr == NULL );
  ASSERT( m_pCSLIDEFogCurr == NULL );

  pCSLIDEFog = m_pCSLIDEWorld->PeekFog(pcID);
  if ( pCSLIDEFog != NULL )
    {
      HandleError(uiLine, uiColumn, "%s \"%s\" previously defined at %s: %ld: %ld", 
		  gcppcSLIDEReservedWords[sliderwtFog],
		  pcID, pCSLIDEFog->GetFileName(), 
		  pCSLIDEFog->GetFileLine(), 
		  pCSLIDEFog->GetFileColumn());
    }

  pCSLIDEFog = CreateFog(pcFile, uiLine, uiColumn, pcID);
  ASSERT( pCSLIDEFog != NULL );

  VERIFY(m_pCSLIDEWorld->InsertFog(pCSLIDEFog));
  m_pCSLIDEPrimitiveCurr = pCSLIDEFog;
  m_pCSLIDEFogCurr = pCSLIDEFog;
}

VOID CSLIDEParser::EndFog()
{
  ASSERT( m_pCSLIDEFogCurr != NULL );
  ASSERT( m_pCSLIDEPrimitiveCurr == m_pCSLIDEFogCurr );
  m_pCSLIDEFogCurr = NULL;
  m_pCSLIDEPrimitiveCurr = NULL;
}

VOID CSLIDEParser::SetFogType(DynInt *pdyniType)
{
  ASSERT( m_pCSLIDEFogCurr != NULL );

  m_pCSLIDEFogCurr->SetType(pdyniType);
  delete pdyniType;
}

VOID CSLIDEParser::SetFogDensity(DynFloat *pdynfDensity)
{
  ASSERT( m_pCSLIDEFogCurr != NULL );

  m_pCSLIDEFogCurr->SetDensity(pdynfDensity);
  delete pdynfDensity;
}

VOID CSLIDEParser::SetFogStart(DynFloat *pdynfStart)
{
  ASSERT( m_pCSLIDEFogCurr != NULL );

  m_pCSLIDEFogCurr->SetStart(pdynfStart);
  delete pdynfStart;
}

VOID CSLIDEParser::SetFogEnd(DynFloat *pdynfEnd)
{
  ASSERT( m_pCSLIDEFogCurr != NULL );

  m_pCSLIDEFogCurr->SetEnd(pdynfEnd);
  delete pdynfEnd;
}

VOID CSLIDEParser::SetFogColor(DynVector *pdynv)
{
  ASSERT( m_pCSLIDEFogCurr != NULL );

  m_pCSLIDEFogCurr->SetColor(pdynv);
  delete pdynv;
}

VOID CSLIDEParser::AddGlobalInitTclBlock(CHAR *pcTclBlock)
{
  VERIFY(m_pCSLIDEWorld->InsertGlobalInitTclBlock(pcTclBlock));
}

VOID CSLIDEParser::AddGlobalUpdateTclBlock(CHAR *pcTclBlock)
{
  VERIFY(m_pCSLIDEWorld->InsertGlobalUpdateTclBlock(pcTclBlock));
}

CListInt *CSLIDEParser::CreateIDList()
{
  return CreateList();
}

CListInt *CSLIDEParser::InsertIDList(CListInt *pCListInt, CHAR *pcID)
{
  return InsertList(pCListInt, pcID);
}

CListInt *CSLIDEParser::CreateDynvalueList()
{
  return CreateList();
}

CListInt *CSLIDEParser::InsertDynvalueList(CListInt *pCListInt, DynFloat *pdfValue)
{
  return InsertList(pCListInt, pdfValue);
}

CSLIDEWorld *CSLIDEParser::CreateWorld()
{
  CSLIDEWorld *pCSLIDEWorld;

  pCSLIDEWorld = new CSLIDEWorld();
  ASSERT( pCSLIDEWorld != NULL );
  pCSLIDEWorld->Init();

  return pCSLIDEWorld;
}

CSLIDERender *CSLIDEParser::CreateRender(CHAR *pcFile, UINT_32 uiLine, 
					 UINT_32 uiColumn,
					 CHAR *pcViewportID, 
					 CListInt *plistCameraPath, 
					 CHAR *pcRootID)
{
  CSLIDECameraPath *pCSLIDECameraPath;
  CSLIDERender *pCSLIDERender;

  pCSLIDECameraPath = new CSLIDECameraPath();
  ASSERT( pCSLIDECameraPath != NULL );
  pCSLIDECameraPath->Init(pcFile, uiLine, uiColumn, plistCameraPath);

  pCSLIDERender = new CSLIDERender();
  ASSERT( pCSLIDERender != NULL );
  pCSLIDERender->Init(pcFile, uiLine, uiColumn, pcViewportID, 
		      pCSLIDECameraPath, pcRootID);

  return pCSLIDERender;
}

CSLIDELightPath *CSLIDEParser::CreateLightPath(CHAR *pcFile, UINT_32 uiLine, 
					       UINT_32 uiColumn,
					       CListInt *plistLightPath,
					       DynInt *pdyniOnOff)
{
  CSLIDELightPath *pCSLIDELightPath;

  pCSLIDELightPath = new CSLIDELightPath();
  ASSERT( pCSLIDELightPath != NULL );
  pCSLIDELightPath->Init(pcFile, uiLine, uiColumn, plistLightPath, pdyniOnOff);

  if ( pdyniOnOff != NULL )
    {
      delete pdyniOnOff;
    }

  return pCSLIDELightPath;
}

CSLIDEAreaLightPath *CSLIDEParser::CreateAreaLightPath(CHAR *pcFile, UINT_32 uiLine, 
					       UINT_32 uiColumn,
					       CListInt *plistAreaLightPath,
					       DynInt *pdyniOnOff)
{
  CSLIDEAreaLightPath *pCSLIDEAreaLightPath;

  pCSLIDEAreaLightPath = new CSLIDEAreaLightPath();
  ASSERT( pCSLIDEAreaLightPath != NULL );
  pCSLIDEAreaLightPath->Init(pcFile, uiLine, uiColumn, plistAreaLightPath, pdyniOnOff);

  if ( pdyniOnOff != NULL )
    {
      delete pdyniOnOff;
    }

  return pCSLIDEAreaLightPath;
}

CWindow *CSLIDEParser::CreateWindow(CHAR *pcFile, UINT_32 uiLine, 
				    UINT_32 uiColumn,
				    CHAR *pcID)
{
  CWindow *pCWindow;

  pCWindow = new CWindow();
  ASSERT( pCWindow != NULL );
  pCWindow->Init(pcFile, uiLine, uiColumn, pcID);

  return pCWindow;
}

CSLIDEViewport *CSLIDEParser::CreateViewport(CHAR *pcFile, 
					     UINT_32 uiLine, 
					     UINT_32 uiColumn,
					     CHAR *pcID, 
					     CHAR *pcWindowID)
{
  CSLIDEViewport *pCSLIDEViewport;

  pCSLIDEViewport = new CSLIDEViewport();
  ASSERT( pCSLIDEViewport != NULL );
  pCSLIDEViewport->Init(pcFile, uiLine, uiColumn, pcID, pcWindowID);

  return pCSLIDEViewport;
}

CSLIDECamera *CSLIDEParser::CreateCamera(CHAR *pcFile, UINT_32 uiLine, 
					 UINT_32 uiColumn,
					 CHAR *pcID)
{
  CSLIDECamera *pCSLIDECamera;

  pCSLIDECamera = new CSLIDECamera();
  ASSERT( pCSLIDECamera != NULL );
  pCSLIDECamera->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDECamera;
}

CSLIDELight *CSLIDEParser::CreateLight(CHAR *pcFile, UINT_32 uiLine, 
					 UINT_32 uiColumn,
					 CHAR *pcID)
{
  CSLIDELight *pCSLIDELight;

  pCSLIDELight = new CSLIDELight();
  ASSERT( pCSLIDELight != NULL );
  pCSLIDELight->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDELight;
}

CSLIDEAreaLight *CSLIDEParser::CreateAreaLight(CHAR *pcFile, UINT_32 uiLine, 
				       UINT_32 uiColumn,
				       CHAR *pcID)
{
  CSLIDEAreaLight *pCSLIDEAreaLight;

  pCSLIDEAreaLight = new CSLIDEAreaLight();
  ASSERT( pCSLIDEAreaLight != NULL );
  pCSLIDEAreaLight->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDEAreaLight;
}

CSLIDEGroup *CSLIDEParser::CreateGroup(CHAR *pcFile, UINT_32 uiLine, 
				       UINT_32 uiColumn,
				       CHAR *pcID)
{
  CSLIDEGroup *pCSLIDEGroup;

  pCSLIDEGroup = new CSLIDEGroup();
  ASSERT( pCSLIDEGroup != NULL );
  pCSLIDEGroup->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDEGroup;
}

CSLIDESeam *CSLIDEParser::CreateSeam(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn,
				     CHAR *pcID, CHAR *pcContour1ID, CHAR *pcContour2ID)
{
  CSLIDESeam *pCSLIDESeam;

  pCSLIDESeam = new CSLIDESeam();
  ASSERT( pCSLIDESeam != NULL );
  pCSLIDESeam->Init(pcFile, uiLine, uiColumn, pcID, pcContour1ID, pcContour2ID);

  return pCSLIDESeam;
}

CSLIDEContour *CSLIDEParser::CreateContour(CHAR *pcFile, 
					     UINT_32 uiLine, 
					     UINT_32 uiColumn,
					     CHAR *pcChildID)
{
  CSLIDEContour *pCSLIDEContour;

  pCSLIDEContour = new CSLIDEContour();
  ASSERT( pCSLIDEContour != NULL );
  pCSLIDEContour->Init(pcFile, uiLine, uiColumn, pcChildID);

  return pCSLIDEContour;
}

CSLIDESubdivision *CSLIDEParser::CreateSubdivision(CHAR *pcFile, UINT_32 uiLine, 
				       UINT_32 uiColumn,
				       CHAR *pcID)
{
  CSLIDESubdivision *pCSLIDESubdivision;

  pCSLIDESubdivision = new CSLIDESubdivision();
  ASSERT( pCSLIDESubdivision != NULL );
  pCSLIDESubdivision->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDESubdivision;
}

CSLIDEOffset *CSLIDEParser::CreateOffset(CHAR *pcFile, UINT_32 uiLine, 
				       UINT_32 uiColumn,
				       CHAR *pcID)
{
  CSLIDEOffset *pCSLIDEOffset;

  pCSLIDEOffset = new CSLIDEOffset();
  ASSERT( pCSLIDEOffset != NULL );
  pCSLIDEOffset->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDEOffset;
}

CSLIDEDual *CSLIDEParser::CreateDual(CHAR *pcFile, UINT_32 uiLine, 
				     UINT_32 uiColumn,
				     CHAR *pcID)
{
  CSLIDEDual *pCSLIDEDual;

  pCSLIDEDual = new CSLIDEDual();
  ASSERT( pCSLIDEDual != NULL );
  pCSLIDEDual->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDEDual;
}

CSLIDEEvolver *CSLIDEParser::CreateEvolver(CHAR *pcFile, UINT_32 uiLine, 
				     UINT_32 uiColumn,
				     CHAR *pcID)
{
  CSLIDEEvolver *pCSLIDEEvolver;

  pCSLIDEEvolver = new CSLIDEEvolver();
  ASSERT( pCSLIDEEvolver != NULL );
  pCSLIDEEvolver->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDEEvolver;
}

CSLIDEMirror *CSLIDEParser::CreateMirror(CHAR *pcFile, UINT_32 uiLine, 
				     UINT_32 uiColumn,
				     CHAR *pcID)
{
  CSLIDEMirror *pCSLIDEMirror;

  pCSLIDEMirror = new CSLIDEMirror();
  ASSERT( pCSLIDEMirror != NULL );
  pCSLIDEMirror->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDEMirror;
}

CSLIDEGeodesic *CSLIDEParser::CreateGeodesic(CHAR *pcFile, UINT_32 uiLine, 
				     UINT_32 uiColumn,
				     CHAR *pcID)
{
  CSLIDEGeodesic *pCSLIDEGeodesic;

  pCSLIDEGeodesic = new CSLIDEGeodesic();
  ASSERT( pCSLIDEGeodesic != NULL );
  pCSLIDEGeodesic->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDEGeodesic;
}

CSLIDEUnion *CSLIDEParser::CreateUnion(CHAR *pcFile, UINT_32 uiLine, 
				       UINT_32 uiColumn,
				       CHAR *pcID)
{
  CSLIDEUnion *pCSLIDEUnion;

  pCSLIDEUnion = new CSLIDEUnion();
  ASSERT( pCSLIDEUnion != NULL );
  pCSLIDEUnion->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDEUnion;
}

CSLIDEIntersection *CSLIDEParser::CreateIntersection(CHAR *pcFile, UINT_32 uiLine, 
				       UINT_32 uiColumn,
				       CHAR *pcID)
{
  CSLIDEIntersection *pCSLIDEIntersection;

  pCSLIDEIntersection = new CSLIDEIntersection();
  ASSERT( pCSLIDEIntersection != NULL );
  pCSLIDEIntersection->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDEIntersection;
}

CSLIDEDifference *CSLIDEParser::CreateDifference(CHAR *pcFile, UINT_32 uiLine, 
						 UINT_32 uiColumn,
						 CHAR *pcID)
{
  CSLIDEDifference *pCSLIDEDifference;

  pCSLIDEDifference = new CSLIDEDifference();
  ASSERT( pCSLIDEDifference != NULL );
  pCSLIDEDifference->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDEDifference;
}

CSLIDESGRoot *CSLIDEParser::CreateSGRoot(CHAR *pcFile, UINT_32 uiLine, 
					 UINT_32 uiColumn,
					 CHAR *pcID, CHAR *pcSGRoot)
{
  CSLIDESGRoot *pCSLIDESGRoot;

  pCSLIDESGRoot = new CSLIDESGRoot();
  ASSERT( pCSLIDESGRoot != NULL );
  pCSLIDESGRoot->Init(pcFile, uiLine, uiColumn, pcID, pcSGRoot);

  return pCSLIDESGRoot;
}

CSLIDEInstance *CSLIDEParser::CreateInstance(CHAR *pcFile, 
					     UINT_32 uiLine, 
					     UINT_32 uiColumn,
					     CHAR *pcChildID,
					     CHAR *pcParentID)
{
  CSLIDEInstance *pCSLIDEInstance;

  pCSLIDEInstance = new CSLIDEInstance();
  ASSERT( pCSLIDEInstance != NULL );
  pCSLIDEInstance->Init(pcFile, uiLine, uiColumn, pcChildID, pcParentID);

  return pCSLIDEInstance;
}

CSLIDETranslate *CSLIDEParser::CreateTranslate(CHAR *pcFile, 
					       UINT_32 uiLine,
					       UINT_32 uiColumn,
					       DynVector *pdynv)
{
  CSLIDETranslate *pCSLIDETranslate;

  pCSLIDETranslate = new CSLIDETranslate();
  ASSERT( pCSLIDETranslate != NULL );
  pCSLIDETranslate->Init(pcFile, uiLine, uiColumn, pdynv);

  delete pdynv;

  return pCSLIDETranslate;
}

CSLIDEScale *CSLIDEParser::CreateScale(CHAR *pcFile, UINT_32 uiLine, 
				       UINT_32 uiColumn,
				       DynVector *pdynv)
{
  CSLIDEScale *pCSLIDEScale;

  pCSLIDEScale = new CSLIDEScale();
  ASSERT( pCSLIDEScale != NULL );
  pCSLIDEScale->Init(pcFile, uiLine, uiColumn, pdynv);

  delete pdynv;

  return pCSLIDEScale;
}

CSLIDERotate *CSLIDEParser::CreateRotate(CHAR *pcFile, UINT_32 uiLine, 
					 UINT_32 uiColumn,
					 DynVector *pdynvAxis,
					 DynFloat *pdynfAngle)
{
  CSLIDERotate *pCSLIDERotate;

  pCSLIDERotate = new CSLIDERotate();
  ASSERT( pCSLIDERotate != NULL );

  pdynfAngle->SetValue(DEGTORAD(pdynfAngle->GetValue()));
  pCSLIDERotate->Init(pcFile, uiLine, uiColumn, pdynvAxis,
		      pdynfAngle);

  delete pdynvAxis;
  delete pdynfAngle;

  return pCSLIDERotate;
}

CSLIDELookAt *CSLIDEParser::CreateLookAt(CHAR *pcFile, UINT_32 uiLine, 
					 UINT_32 uiColumn)
{
  CSLIDELookAt *pCSLIDELookAt;

  pCSLIDELookAt = new CSLIDELookAt();
  ASSERT( pCSLIDELookAt != NULL );
  pCSLIDELookAt->Init(pcFile, uiLine, uiColumn);

  return pCSLIDELookAt;
}

/*
CSLIDETransform *CSLIDEParser::CreateGeneral(CHAR *pcFile, 
					     UINT_32 uiLine, 
					     UINT_32 uiColumn,
					     const Matrix &m)
{
  CSLIDETransform *pCSLIDETransform;

  pCSLIDETransform = new CSLIDETransform();
  ASSERT( pCSLIDETransform != NULL );
  pCSLIDETransform->Init(pcFile, uiLine, uiColumn);
  pCSLIDETransform->SetMatrix(m);

  return pCSLIDETransform;
}
*/
CSLIDEObject *CSLIDEParser::CreateObject(CHAR *pcFile, UINT_32 uiLine, 
					 UINT_32 uiColumn,
					 CHAR *pcID, 
					 CListInt *plistFaceIDs)
{
  CSLIDEObject *pCSLIDEObject;

  pCSLIDEObject = new CSLIDEObject();
  ASSERT( pCSLIDEObject != NULL );
  pCSLIDEObject->Init(pcFile, uiLine, uiColumn, pcID,
		      plistFaceIDs);

  return pCSLIDEObject;
}

CSLIDEMesh *CSLIDEParser::CreateMesh(CHAR *pcFile, UINT_32 uiLine, 
				     UINT_32 uiColumn,
				     CHAR *pcID)
{
  CSLIDEMesh *pCSLIDEMesh;

  pCSLIDEMesh = new CSLIDEMesh();
  ASSERT( pCSLIDEMesh != NULL );
  pCSLIDEMesh->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDEMesh;
}

CSLIDEMeshVertexProperty *CSLIDEParser::CreateMeshVertexProperty(CHAR *pcFile, UINT_32 uiLine, 
							     UINT_32 uiColumn,
							     CHAR *pcID)
{
  CSLIDEMeshVertexProperty *pCSLIDEMeshVertexProperty;

  pCSLIDEMeshVertexProperty = new CSLIDEMeshVertexProperty();
  ASSERT( pCSLIDEMeshVertexProperty != NULL );
  pCSLIDEMeshVertexProperty->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDEMeshVertexProperty;
}

CSLIDEMeshEdgeProperty *CSLIDEParser::CreateMeshEdgeProperty(CHAR *pcFile, UINT_32 uiLine, 
							     UINT_32 uiColumn,
							     CHAR *pcID)
{
  CSLIDEMeshEdgeProperty *pCSLIDEMeshEdgeProperty;

  pCSLIDEMeshEdgeProperty = new CSLIDEMeshEdgeProperty();
  ASSERT( pCSLIDEMeshEdgeProperty != NULL );
  pCSLIDEMeshEdgeProperty->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDEMeshEdgeProperty;
}

CSLIDEMeshVertex *CSLIDEParser::CreateMeshVertex(CHAR *pcFile, UINT_32 uiLine, 
						 UINT_32 uiColumn,
						 CHAR *pcID, 
						 DynPoint *pdynpt)
{
  CSLIDEMeshVertex *pCSLIDEMeshVertex;

  pCSLIDEMeshVertex = new CSLIDEMeshVertex();
  ASSERT( pCSLIDEMeshVertex != NULL );
  pCSLIDEMeshVertex->Init(pcFile, uiLine, uiColumn, pcID,
			  pdynpt);

  return pCSLIDEMeshVertex;
}

CSLIDEMeshFacet *CSLIDEParser::CreateMeshFacet(CHAR *pcFile, UINT_32 uiLine, 
					       UINT_32 uiColumn,
					       CHAR *pcID, 
					       CListInt *plistPointIDs)
{
  CSLIDEMeshFacet *pCSLIDEMeshFacet;

  pCSLIDEMeshFacet = new CSLIDEMeshFacet();
  ASSERT( pCSLIDEMeshFacet != NULL );
  pCSLIDEMeshFacet->Init(pcFile, uiLine, uiColumn, pcID,
		    plistPointIDs);

  return pCSLIDEMeshFacet;
}

CSLIDEMeshEdge *CSLIDEParser::CreateMeshEdge(CHAR *pcFile, UINT_32 uiLine, 
					     UINT_32 uiColumn,
					     CHAR *pcID, 
					     CHAR *pcMeshVertexID0, CHAR *pcMeshVertexID1)
{
  CSLIDEMeshEdge *pCSLIDEMeshEdge;

  pCSLIDEMeshEdge = new CSLIDEMeshEdge();
  ASSERT( pCSLIDEMeshEdge != NULL );
  pCSLIDEMeshEdge->Init(pcFile, uiLine, uiColumn, pcID,
			pcMeshVertexID0, pcMeshVertexID1);

  return pCSLIDEMeshEdge;
}

CSLIDEConvexHull *CSLIDEParser::CreateConvexHull(CHAR *pcFile, UINT_32 uiLine, 
						 UINT_32 uiColumn,
						 CHAR *pcID, 
						 CListInt *plistPointIDs)
{
  CSLIDEConvexHull *pCSLIDEConvexHull;

  pCSLIDEConvexHull = new CSLIDEConvexHull();
  ASSERT( pCSLIDEConvexHull != NULL );
  pCSLIDEConvexHull->Init(pcFile, uiLine, uiColumn, pcID,
		       plistPointIDs);

  return pCSLIDEConvexHull;
}

CSLIDEGeneralFace *CSLIDEParser::CreateGeneralFace(CHAR *pcFile, UINT_32 uiLine, 
					   UINT_32 uiColumn,
					   CHAR *pcID, 
					   CListInt *plistPointIDs)
{
  CSLIDEGeneralFace *pCSLIDEGeneralFace;

  pCSLIDEGeneralFace = new CSLIDEGeneralFace();
  ASSERT( pCSLIDEGeneralFace != NULL );
  pCSLIDEGeneralFace->Init(pcFile, uiLine, uiColumn, pcID,
		       plistPointIDs);

  return pCSLIDEGeneralFace;
}

CSLIDETerrain *CSLIDEParser::CreateTerrain(CHAR *pcFile, UINT_32 uiLine, 
					   UINT_32 uiColumn,
					   CHAR *pcID, 
					   CListInt *plistPointIDs)
{
  CSLIDETerrain *pCSLIDETerrain;

  pCSLIDETerrain = new CSLIDETerrain();
  ASSERT( pCSLIDETerrain != NULL );
  pCSLIDETerrain->Init(pcFile, uiLine, uiColumn, pcID,
		       plistPointIDs);

  return pCSLIDETerrain;
}

CSLIDESphere *CSLIDEParser::CreateSphere(CHAR *pcFile, UINT_32 uiLine, 
					 UINT_32 uiColumn,
					 CHAR *pcID)
{
  CSLIDESphere *pCSLIDESphere;

  pCSLIDESphere = new CSLIDESphere();
  ASSERT( pCSLIDESphere != NULL );
  pCSLIDESphere->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDESphere;
}

CSLIDECylinder *CSLIDEParser::CreateCylinder(CHAR *pcFile, UINT_32 uiLine, 
					 UINT_32 uiColumn,
					 CHAR *pcID)
{
  CSLIDECylinder *pCSLIDECylinder;

  pCSLIDECylinder = new CSLIDECylinder();
  ASSERT( pCSLIDECylinder != NULL );
  pCSLIDECylinder->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDECylinder;
}

CSLIDECone *CSLIDEParser::CreateCone(CHAR *pcFile, UINT_32 uiLine, 
					 UINT_32 uiColumn,
					 CHAR *pcID)
{
  CSLIDECone *pCSLIDECone;

  pCSLIDECone = new CSLIDECone();
  ASSERT( pCSLIDECone != NULL );
  pCSLIDECone->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDECone;
}

CSLIDETorus *CSLIDEParser::CreateTorus(CHAR *pcFile, UINT_32 uiLine, 
					 UINT_32 uiColumn,
					 CHAR *pcID)
{
  CSLIDETorus *pCSLIDETorus;

  pCSLIDETorus = new CSLIDETorus();
  ASSERT( pCSLIDETorus != NULL );
  pCSLIDETorus->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDETorus;
}

CSLIDECyclide *CSLIDEParser::CreateCyclide(CHAR *pcFile, UINT_32 uiLine, 
					 UINT_32 uiColumn,
					 CHAR *pcID)
{
  CSLIDECyclide *pCSLIDECyclide;

  pCSLIDECyclide = new CSLIDECyclide();
  ASSERT( pCSLIDECyclide != NULL );
  pCSLIDECyclide->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDECyclide;
}

CSLIDEBSplinePatch *CSLIDEParser::CreateBSplinePatch(CHAR *pcFile, UINT_32 uiLine, 
					 UINT_32 uiColumn,
					 CHAR *pcID)
{
  CSLIDEBSplinePatch *pCSLIDEBSplinePatch;

  pCSLIDEBSplinePatch = new CSLIDEBSplinePatch();
  ASSERT( pCSLIDEBSplinePatch != NULL );
  pCSLIDEBSplinePatch->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDEBSplinePatch;
}

CSLIDEHeightfield *CSLIDEParser::CreateHeightfield(CHAR *pcFile, UINT_32 uiLine, 
					 UINT_32 uiColumn,
					 CHAR *pcID)
{
  CSLIDEHeightfield *pCSLIDEHeightfield;

  pCSLIDEHeightfield = new CSLIDEHeightfield();
  ASSERT( pCSLIDEHeightfield != NULL );
  pCSLIDEHeightfield->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDEHeightfield;
}

CSLIDENurbPatch *CSLIDEParser::CreateNurbPatch(CHAR *pcFile, UINT_32 uiLine, 
					 UINT_32 uiColumn,
					 CHAR *pcID)
{
  CSLIDENurbPatch *pCSLIDENurbPatch;

  pCSLIDENurbPatch = new CSLIDENurbPatch();
  ASSERT( pCSLIDENurbPatch != NULL );
  pCSLIDENurbPatch->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDENurbPatch;
}

CSLIDEBezierPatch *CSLIDEParser::CreateBezierPatch(CHAR *pcFile, UINT_32 uiLine, 
					 UINT_32 uiColumn,
					 CHAR *pcID)
{
  CSLIDEBezierPatch *pCSLIDEBezierPatch;

  pCSLIDEBezierPatch = new CSLIDEBezierPatch();
  ASSERT( pCSLIDEBezierPatch != NULL );
  pCSLIDEBezierPatch->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDEBezierPatch;
}

CSLIDEPolyline *CSLIDEParser::CreatePolyline(CHAR *pcFile, UINT_32 uiLine, 
				       UINT_32 uiColumn,
				       CHAR *pcID)
{
  CSLIDEPolyline *pCSLIDEPolyline;

  pCSLIDEPolyline = new CSLIDEPolyline();
  ASSERT( pCSLIDEPolyline != NULL );
  pCSLIDEPolyline->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDEPolyline;
}

CSLIDEBSplineCurve *CSLIDEParser::CreateBSplineCurve(CHAR *pcFile, UINT_32 uiLine, 
				       UINT_32 uiColumn,
				       CHAR *pcID)
{
  CSLIDEBSplineCurve *pCSLIDEBSplineCurve;

  pCSLIDEBSplineCurve = new CSLIDEBSplineCurve();
  ASSERT( pCSLIDEBSplineCurve != NULL );
  pCSLIDEBSplineCurve->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDEBSplineCurve;
}

CSLIDECrossSection *CSLIDEParser::CreateCrossSection(CHAR *pcFile, UINT_32 uiLine, 
						     UINT_32 uiColumn,
						     CHAR *pcID)
{
  CSLIDECrossSection *pCSLIDECrossSection;

  pCSLIDECrossSection = new CSLIDECrossSection();
  ASSERT( pCSLIDECrossSection != NULL );
  pCSLIDECrossSection->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDECrossSection;
}

CSLIDESweep *CSLIDEParser::CreateSweep(CHAR *pcFile, UINT_32 uiLine, 
				       UINT_32 uiColumn,
				       CHAR *pcID)
{
  CSLIDESweep *pCSLIDESweep;

  pCSLIDESweep = new CSLIDESweep();
  ASSERT( pCSLIDESweep != NULL );
  pCSLIDESweep->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDESweep;
}

CSLIDESweepMorph *CSLIDEParser::CreateSweepMorph(CHAR *pcFile, UINT_32 uiLine, 
						 UINT_32 uiColumn,
						 CHAR *pcID)
{
  CSLIDESweepMorph *pCSLIDESweepMorph;

  pCSLIDESweepMorph = new CSLIDESweepMorph();
  ASSERT( pCSLIDESweepMorph != NULL );
  pCSLIDESweepMorph->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDESweepMorph;
}

CSLIDESweepCrossSection *CSLIDEParser::CreateSweepCrossSection(CHAR *pcFile, UINT_32 uiLine,
						   UINT_32 uiColumn,
						   CHAR *pcID)
{
  CSLIDESweepCrossSection *pCSLIDESweepCS;

  pCSLIDESweepCS = new CSLIDESweepCrossSection();
  ASSERT( pCSLIDESweepCS != NULL );
  pCSLIDESweepCS->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDESweepCS;
}

CSLIDEFace *CSLIDEParser::CreateFace(CHAR *pcFile, UINT_32 uiLine, 
				     UINT_32 uiColumn,
				     CHAR *pcID, 
				     CListInt *plistPointIDs)
{
  CSLIDEFace *pCSLIDEFace;

  pCSLIDEFace = new CSLIDEFace();
  ASSERT( pCSLIDEFace != NULL );
  pCSLIDEFace->Init(pcFile, uiLine, uiColumn, pcID, plistPointIDs);

  return pCSLIDEFace;
}

CSLIDEPoint *CSLIDEParser::CreatePoint(CHAR *pcFile, UINT_32 uiLine, 
				       UINT_32 uiColumn,
				       CHAR *pcID, 
				       DynPoint *pdynpt)
{
  CSLIDEPoint *pCSLIDEPoint;

  pCSLIDEPoint = new CSLIDEPoint();
  ASSERT( pCSLIDEPoint != NULL );
  pCSLIDEPoint->Init(pcFile, uiLine, uiColumn, pcID, pdynpt);

  delete pdynpt;

  return pCSLIDEPoint;
}

CSLIDEControlPoint *CSLIDEParser::CreateControlPoint(CHAR *pcFile, UINT_32 uiLine, 
						     UINT_32 uiColumn,
						     CHAR *pcID)
{
  CSLIDEControlPoint *pCSLIDEControlPoint;

  pCSLIDEControlPoint = new CSLIDEControlPoint();
  ASSERT( pCSLIDEControlPoint != NULL );
  pCSLIDEControlPoint->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDEControlPoint;
}

CSLIDESurface *CSLIDEParser::CreateSurface(CHAR *pcFile, UINT_32 uiLine, 
					   UINT_32 uiColumn,
					   CHAR *pcID)
{
  CSLIDESurface *pCSLIDESurface;

  pCSLIDESurface = new CSLIDESurface();
  ASSERT( pCSLIDESurface != NULL );
  pCSLIDESurface->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDESurface;
}

CSLIDEWormwarp *CSLIDEParser::CreateWormwarp(CHAR *pcFile, UINT_32 uiLine, 
					   UINT_32 uiColumn,
					   CHAR *pcID)
{
  CSLIDEWormwarp *pCSLIDEWormwarp;

  pCSLIDEWormwarp = new CSLIDEWormwarp();
  ASSERT( pCSLIDEWormwarp != NULL);
  pCSLIDEWormwarp->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDEWormwarp;
}

CSLIDEFog *CSLIDEParser::CreateFog(CHAR *pcFile, UINT_32 uiLine, 
					   UINT_32 uiColumn,
					   CHAR *pcID)
{
  CSLIDEFog *pCSLIDEFog;

  pCSLIDEFog = new CSLIDEFog();
  ASSERT( pCSLIDEFog != NULL );
  pCSLIDEFog->Init(pcFile, uiLine, uiColumn, pcID);

  return pCSLIDEFog;
}

CListInt *CSLIDEParser::CreateList()
{
  CListInt *pCListInt;

  pCListInt = new CListInt();
  ASSERT( pCListInt != NULL );

  pCListInt->Init();

  return pCListInt;
}

CListInt *CSLIDEParser::InsertList(CListInt *pCListInt, VOID *pvData)
{
  ASSERT( pCListInt != NULL );

  pCListInt->InsertLast(pvData, 0);

  return pCListInt;
}


CListString *CSLIDEParser::CreateStringList()
{
  CListString *pCListString;

  pCListString = new CListString();
  ASSERT( pCListString != NULL );

  pCListString->Init();

  return pCListString;
}

CListString *CSLIDEParser::InsertStringList(CListString *pCListString, VOID *pvData)
{
  ASSERT( pCListString != NULL );

  pCListString->InsertLast(pvData, 0);

  return pCListString;
}




//////////////////////////////////////////////////////
// Oot stuff

DynInt *CSLIDEParser::TranslateOnOffFlag(CHAR *pcFlag)
{
  DynInt *pdyniFlag = new DynInt;
  BOOL found = FALSE;

  for (int i=0; i<slideoftSize && !found; i++)
    {
      if (!strcmp(pcFlag, gcppcSLIDEOnOffReservedWords[i]))
	{
	  pdyniFlag->SetValue(i);
	  found = TRUE;
	}
    }
  
  if (!found)
    {
      pdyniFlag->SetDynExpr(pcFlag);
    }

  return pdyniFlag;
}

DynInt *CSLIDEParser::TranslateSolidFlag(CHAR *pcFlag)
{
  DynInt *pdyniFlag = new DynInt;
  BOOL found = FALSE;

  for (int i=0; i<slidestSize && !found; i++)
    {
      if (!strcmp(pcFlag, gcppcSLIDESolidReservedWords[i]))
	{
	  pdyniFlag->SetValue(i);
	  found = TRUE;
	}
    }
  
  if (!found)
    {
      pdyniFlag->SetDynExpr(pcFlag);
    }

  return pdyniFlag;
}

DynInt *CSLIDEParser::TranslateShadingFlag(CHAR *pcFlag)
{
  DynInt *pdyniFlag = new DynInt;
  BOOL found = FALSE;

  for (int i=0; i<slideshtSize && !found; i++)
    {
      if (!strcmp(pcFlag, gcppcSLIDEShadingReservedWords[i]))
	{
	  pdyniFlag->SetValue(i);
	  found = TRUE;
	}
    }
  
  if (!found)
    {
      pdyniFlag->SetDynExpr(pcFlag);
    }

  return pdyniFlag;
}

DynInt *CSLIDEParser::TranslateLODFlag(CHAR *pcFlag)
{
  DynInt *pdyniFlag = new DynInt;
  BOOL found = FALSE;

  for (int i=0; i<slidelodtSize && !found; i++)
    {
      if (!strcmp(pcFlag, gcppcSLIDELODReservedWords[i]))
	{
	  pdyniFlag->SetValue(i);
	  found = TRUE;
	}
    }
  
  if (!found)
    {
      pdyniFlag->SetDynExpr(pcFlag);
    }

  return pdyniFlag;
}

DynInt *CSLIDEParser::TranslateTextureFlag(CHAR *pcFlag)
{
  DynInt *pdyniFlag = new DynInt;
  BOOL found = FALSE;

  for (int i=0; i<slidetexturetSize && !found; i++)
    {
      if (!strcmp(pcFlag, gcppcSLIDETextureReservedWords[i]))
	{
	  pdyniFlag->SetValue(i);
	  found = TRUE;
	}
    }

  if (!found)
    {
      pdyniFlag->SetDynExpr(pcFlag);
    }

  return pdyniFlag;
}

DynInt *CSLIDEParser::TranslateProjectionFlag(CHAR *pcFlag)
{
  DynInt *pdyniFlag = new DynInt;
  BOOL found = FALSE;

  for (int i=0; i<slidepjtSize && !found; i++)
    {
      if (!strcmp(pcFlag, gcppcSLIDEProjectionReservedWords[i]))
	{
	  pdyniFlag->SetValue(i);
	  found = TRUE;
	}
    }
  
  if (!found)
    {
      pdyniFlag->SetDynExpr(pcFlag);
    }

  return pdyniFlag;
}

DynInt *CSLIDEParser::TranslateLightFlag(CHAR *pcFlag)
{
  DynInt *pdyniFlag = new DynInt;
  BOOL found = FALSE;

  for (int i=0; i<slideltSize && !found; i++)
    {
      if (!strcmp(pcFlag, gcppcSLIDELightReservedWords[i]))
	{
	  pdyniFlag->SetValue(i);
	  found = TRUE;
	}
    }
  
  if (!found)
    {
      pdyniFlag->SetDynExpr(pcFlag);
    }

  return pdyniFlag;
}

DynInt *CSLIDEParser::TranslateFogFlag(CHAR *pcFlag)
{
  DynInt *pdyniFlag = new DynInt;
  BOOL found = FALSE;

  for (int i=0; i<slidefogtSize && !found; i++)
    {
      if (!strcmp(pcFlag, gcppcSLIDEFogReservedWords[i]))
	{
	  pdyniFlag->SetValue(i);
	  found = TRUE;
	}
    }
  
  if (!found)
    {
      pdyniFlag->SetDynExpr(pcFlag);
    }

  return pdyniFlag;
}

DynInt *CSLIDEParser::TranslateCUIType(CHAR *pcCUIT)
{
  DynInt *pdyniCUIT = new DynInt;
  BOOL found = FALSE;

  for (int i=0; i<slidecuitSize && !found; i++)
    {
      if (!strcmp(pcCUIT, gcppcSLIDECameraUserInterfaceReservedWords[i]))
	{
	  pdyniCUIT->SetValue(i);
	  found = TRUE;
	}
    }
  
  if (!found)
    {
      pdyniCUIT->SetDynExpr(pcCUIT);
    }

  return pdyniCUIT;
}

DynInt *CSLIDEParser::TranslateTerrainTypeFlag(CHAR *pcFlag)
{
  DynInt *pdyniFlag = new DynInt;
  BOOL found = FALSE;

  for (int i=0; i<slidetertSize && !found; i++)
    {
      if (strcmp(pcFlag, gcppcSLIDETerrainReservedWords[i])==0)
	{
	  pdyniFlag->SetValue(i);
	  found = TRUE;
	}
    }

  if (!found)
    {
      pdyniFlag->SetDynExpr(pcFlag);
    }

  return pdyniFlag;
}

DynInt *CSLIDEParser::TranslateVoronoiFlag(CHAR *pcFlag)
{
  DynInt *pdyniFlag = new DynInt;
  BOOL found = FALSE;

  for (int i=0; i<slidevtSize && !found; i++)
    {
      if (strcmp(pcFlag, gcppcSLIDEVoronoiReservedWords[i])==0)
	{
	  pdyniFlag->SetValue(i);
	  found = TRUE;
	}
    }

  if (!found)
    {
      pdyniFlag->SetDynExpr(pcFlag);
    }

  return pdyniFlag;
}

DynInt *CSLIDEParser::TranslateAccuracyFlag(CHAR *pcFlag)
{
  DynInt *pdyniFlag = new DynInt;
  BOOL found = FALSE;

  for (int i=0; i<slideatSize && !found; i++)
    {
      if (strcmp(pcFlag, gcppcSLIDEAccuracyReservedWords[i])==0)
	{
	  pdyniFlag->SetValue(i);
	  found = TRUE;
	}
    }

  if (!found)
    {
      pdyniFlag->SetDynExpr(pcFlag);
    }

  return pdyniFlag;
}

/////////////////////////////////////////////////////////
//
// TCL procedural geometry generation stuff

int CSLIDEParser::HandleAddPoint(VOID* &pCreatedObject, int argc, char **argv)
{
  char *ptname = NULL;
  struct argList *ptcoords = NULL;
  char *surfid = NULL;
  struct argList *ptnormal = NULL;
  struct argList *pttexcoords = NULL;
  char *ribbegin = NULL;
  char *ribend = NULL;

  DynFloat* pdynVectf[4];
  DynPoint *pdynpt=NULL;
  DynVector *pdynvect=NULL;
  int axis=0;

  struct argFormat tclargform[] =
    { { NULL, "name",
	"Name for the new point",
	ARG_SIMPLE | ARG_STRING, &ptname, TRUE },
      { NULL, "X Y Z", 
	"X, Y, and Z values for new point",
	ARG_SIMPLE | ARG_LIST | ARG_FLOAT, &ptcoords, TRUE },
      { "normal", "X Y Z",
	"Point normal X Y Z",
	ARG_LIST | ARG_FLOAT, &ptnormal, FALSE },
      { "surface", "surf_id",
	"Point Surface ID",
	ARG_STRING, &surfid, FALSE },
      { "texture", "X Y Z",
	"texture mapping coordinates X Y Z",
	ARG_LIST | ARG_FLOAT, &pttexcoords, FALSE },

	  { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
	  { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }
    };

  const int argCount = 7;
  
  //printf ("HandleAddPoint: %d\n",argc);
  
  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create point: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }
  
  struct argList *tempptr;
  //printf ("ptname: %s\n",ptname);
  //printf ("ptcoords: \n");
  for (tempptr = ptcoords; tempptr != NULL; tempptr = tempptr->next)
    {
      pdynVectf[axis] = new DynFloat;
      
      if (tempptr->argType == ARG_EXPR)
	{
	  //printf ("  %s\n", (CHAR *)(tempptr->value));
	  pdynVectf[axis]->SetDynExpr((CHAR *)(tempptr->value));
	}
      else
	{
	  //printf ("  %f\n", *(FLOAT *)(tempptr->value));
	  pdynVectf[axis]->SetValue(*(FLOAT*)(tempptr->value));
	}
      
      axis++;
      
      // if too many things, just jump out of loop and throw err
      if (axis > 3 && tempptr->next != NULL)
	{
	  axis = -1;
	  break;
	}
    }
  
  if (axis != 3)
    {
      // Ajax
      //
      TCLAjaxArgList (pttexcoords);
      TCLAjaxArgList (ptcoords);
      TCLAjaxArgList (ptnormal);

      HandleError(0, 0, "%s \"%s\" wrong # of arguments for pt coordinates", 
		  gcppcSLIDEReservedWords[sliderwtPoint],
		  ptname);
      
      return TCL_ERROR;
    }
  
  pdynpt = new DynPoint;
  ASSERT( pdynpt != NULL );
  pdynpt->SetAll(pdynVectf[0], pdynVectf[1], pdynVectf[2]);
  
  ptname = m_pCScanner->FindIdentifier (ptname, TRUE);
  AddPoint ("[TCL generated Point]", 0, 0, ptname, pdynpt);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;
  
  if (surfid != NULL)
    {
      //printf ("surfid: %s\n",surfid);
      surfid = m_pCScanner->FindIdentifier (surfid, TRUE);
      SetPointSurfaceID (surfid);
    }
  
  if (ptnormal != NULL)
    {
      //printf ("ptnormal: \n");
      
      axis = 0;
      for (tempptr = ptnormal; tempptr != NULL; tempptr = tempptr->next)
	{
	  pdynVectf[axis] = new DynFloat;
	  
	  if (tempptr->argType == ARG_EXPR)
	    {
	      //printf ("	 %s\n", (CHAR *)(tempptr->value));
	      pdynVectf[axis]->SetDynExpr((CHAR *)(tempptr->value));
	    }
	  else
	    {
	      //printf ("	 %f\n", *(FLOAT *)(tempptr->value));
	      pdynVectf[axis]->SetValue(*(FLOAT*)(tempptr->value));
	    }
	  
	  axis++;
	  
	  // if too many things, just jump out of loop and throw err
	  if (axis > 3 && tempptr->next != NULL)
	    {
	      axis = -1;
	      break;
	    }
	}
      
      if (axis != 3)
	{
	  // Ajax
	  //
	  TCLAjaxArgList (pttexcoords);
	  TCLAjaxArgList (ptcoords);
	  TCLAjaxArgList (ptnormal);

	  HandleError(0, 0, "%s \"%s\" wrong # of arguments for pt normal", 
		      gcppcSLIDEReservedWords[sliderwtPoint],
		      ptname);
	  
	  return TCL_ERROR;
	}
      
      pdynvect = new DynVector;
      ASSERT( pdynvect != NULL );
      pdynvect->SetAll(pdynVectf[0], pdynVectf[1], pdynVectf[2]);

      
      SetPointNormal (pdynvect);
    }
  
  if (pttexcoords != NULL)
    {
      //printf ("pttexcoords: \n");
      
      axis = 0;
      for (tempptr = pttexcoords; tempptr != NULL; tempptr = tempptr->next)
	{
	  pdynVectf[axis] = new DynFloat;
	  
	  if (tempptr->argType == ARG_EXPR)
	    {
	      //printf ("	 %s\n", (CHAR *)(tempptr->value));
	      pdynVectf[axis]->SetDynExpr((CHAR *)(tempptr->value));
	    }
	  else
	    {
		//printf ("  %f\n", *(FLOAT *)(tempptr->value));
		pdynVectf[axis]->SetValue(*(FLOAT*)(tempptr->value));
	    }
	  
	  axis++;

	  // if too many things, just jump out of loop and throw err
	  if (axis > 3 && tempptr->next != NULL)
	    {
	      axis = -1;
	      break;
	    }
	}

      if (axis != 3)
	{
	  // Ajax
	  //
	  TCLAjaxArgList (pttexcoords);
	  TCLAjaxArgList (ptcoords);
	  TCLAjaxArgList (ptnormal);

	  HandleError(0, 0, "%s \"%s\" wrong # of arguments for pt texture coords", 
		      gcppcSLIDEReservedWords[sliderwtPoint],
		      ptname);
	 
	  return TCL_ERROR;
	}
      
      pdynpt = new DynPoint;
      ASSERT( pdynpt != NULL );
      pdynpt->SetAll(pdynVectf[0], pdynVectf[1], pdynVectf[2]);
      
      SetPointTexture (pdynpt);
      
    }
  
  if (ribbegin != NULL)
    SetPrimitiveRIBBegin (ribbegin);

  if (ribend != NULL)
    SetPrimitiveRIBBegin (ribend);

  //printf ("\n");
  EndPoint();
  
  // Ajax
  //
  TCLAjaxArgList (pttexcoords);
  TCLAjaxArgList (ptcoords);
  TCLAjaxArgList (ptnormal);

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleAddFace(VOID* &pCreatedObject, int argc, char **argv)
{
  char *fname = NULL;
  struct argList *fpoints = NULL;
  char *surfid = NULL;
  CListInt *fpointList=NULL;
  char *ribbegin = NULL;
  char *ribend = NULL;
  
  struct argFormat tclargform[] =
    { { NULL, "name",
	"Name for the new point",
	ARG_SIMPLE | ARG_STRING, &fname, TRUE },
      { NULL, "pt1 pt2 pt3 ...", 
	"List of points for face",
	ARG_SIMPLE | ARG_LIST | ARG_STRING, &fpoints, TRUE },
      { "surface", "surf_id",
	"Face Surface ID",
	ARG_STRING, &surfid, FALSE },
      
      { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
      { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }
    };

  const int argCount = 5;

  //printf ("HandleAddFace: %d\n",argc);

  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create face: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  struct argList *tempptr;
  //printf ("fname: %s\n",fname);

  fpointList = CreateList();
  //printf ("fpoints: \n");
  for (tempptr = fpoints; tempptr != NULL; tempptr = tempptr->next)
    {
      //printf ("	 %s\n", (CHAR *)(tempptr->value));
      tempptr->value = (VOID*)(m_pCScanner->FindIdentifier ((CHAR*)(tempptr->value), 
								    TRUE) );
      InsertList (fpointList, (CHAR*)(tempptr->value));
    }

  fname = m_pCScanner->FindIdentifier (fname, TRUE);
  AddFace ("[TCL Generated Face]", 0, 0, fname, fpointList);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  if (surfid != NULL)
  {
    //printf ("surfid: %s\n",surfid);
    surfid = m_pCScanner->FindIdentifier (surfid, TRUE);
    SetFaceSurfaceID (surfid);
  }

  //printf ("\n");

  if (ribbegin != NULL)
    SetPrimitiveRIBBegin (ribbegin);

  if (ribend != NULL)
    SetPrimitiveRIBBegin (ribend);


  EndFace();

  // Ajax
  //
  TCLAjaxArgList (fpoints);

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleAddObject(VOID* &pCreatedObject, int argc, char **argv)
{
  char *oname = NULL;
  struct argList *ofaces = NULL;
  char *surfid = NULL;
  char *shading = NULL;
  char *lod = NULL;
  char *solid = NULL;
  char *ribbegin = NULL;
  char *ribend = NULL;
  CListInt *ofaceList=NULL;

  struct argFormat tclargform[] =
    { { NULL, "name",
	"Name for the new object",
	ARG_SIMPLE | ARG_STRING, &oname, TRUE },
      { NULL, "face1 face2 face3 ...", 
	"List of faces for object",
	ARG_SIMPLE | ARG_LIST | ARG_STRING, &ofaces, TRUE },
      { "solid", "solid_flag",
	"Object Solidity flag",
	ARG_STRING, &solid, FALSE },
      { "surface", "surf_id",
	"Object Surface ID",
	ARG_STRING, &surfid, FALSE },
      { "shading", "shading_flag",
	"Shading type",
	ARG_STRING, &shading, FALSE },
      { "lod", "lod_flag",
	"Level of Detail flag",
	ARG_STRING, &lod, FALSE },

	  { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
	  { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }

    };

  const int argCount = 8;

  //printf ("HandleAddObject: %d\n",argc);

  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create object: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  struct argList *tempptr;
  //printf ("oname: %s\n",oname);

  ofaceList = CreateList();
  //printf ("ofaces: \n");
  for (tempptr = ofaces; tempptr != NULL; tempptr = tempptr->next)
    {
      //printf ("	 %s\n", (CHAR *)(tempptr->value));
      tempptr->value = (VOID*)(m_pCScanner->FindIdentifier ((CHAR*)(tempptr->value), 
								    TRUE) );
      InsertList (ofaceList, (CHAR*)(tempptr->value));
    }	

  oname = m_pCScanner->FindIdentifier (oname, TRUE);
  AddObject ("[TCL Generated Object]", 0, 0, oname, ofaceList);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  if (surfid != NULL)
  {
    //printf ("surfid: %s\n",surfid);
    surfid = m_pCScanner->FindIdentifier (surfid, TRUE);
    SetNodeSurfaceID(surfid);
  }

  if (shading != NULL)
  {
	//printf ("shading: %s\n",shading);
	DynInt *pdyniShading = TranslateShadingFlag(shading);
	SetNodeShading (pdyniShading);
  }

  if (lod != NULL)
  {
    //printf ("LOD: %s\n",lod);
	DynInt *pdyniLOD = TranslateLODFlag(lod);
	SetNodeLOD (pdyniLOD);
  }

  if (solid != NULL)
  {
    //printf ("solid: %s\n",solid);
	DynInt *pdyniSolid = TranslateSolidFlag(solid);
	SetAtomNodeSolid (pdyniSolid);
  }


  //printf ("\n");

  if (ribbegin != NULL)
    SetPrimitiveRIBBegin (ribbegin);

  if (ribend != NULL)
    SetPrimitiveRIBBegin (ribend);


  
  EndObject();

  // Ajax
  //
  TCLAjaxArgList (ofaces);

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleAddConvexHull(VOID* &pCreatedObject, int argc, char **argv)
{
  char *oname = NULL;
  struct argList *opoints = NULL;
  char *surfid = NULL;
  char *shading = NULL;
  char *lod = NULL;
  char *solid = NULL;
  char *ribbegin = NULL;
  char *ribend = NULL;
  CListInt *opointList=NULL;

  struct argFormat tclargform[] =
    { { NULL, "name",
	"Name for the new convexhull",
	ARG_SIMPLE | ARG_STRING, &oname, TRUE },
      { NULL, "point1 point2 point3 ...", 
	"List of points for convexhull",
	ARG_SIMPLE | ARG_LIST | ARG_STRING, &opoints, TRUE },
      { "solid", "solid_flag",
	"convexhull Solidity flag",
	ARG_STRING, &solid, FALSE },
      { "surface", "surf_id",
	"convexhull Surface ID",
	ARG_STRING, &surfid, FALSE },
      { "shading", "shading_flag",
	"Shading type",
	ARG_STRING, &shading, FALSE },
      { "lod", "lod_flag",
	"Level of Detail flag",
	ARG_STRING, &lod, FALSE },

      { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
      { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }

    };

  const int argCount = 8;

  //printf ("HandleAddConvexHull: %d\n",argc);

  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create convexhull: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  struct argList *tempptr;
  //printf ("oname: %s\n",oname);

  opointList = CreateList();
  //printf ("opoints: \n");
  for (tempptr = opoints; tempptr != NULL; tempptr = tempptr->next)
    {
      //printf ("	 %s\n", (CHAR *)(tempptr->value));
      tempptr->value = (VOID*)(m_pCScanner->FindIdentifier ((CHAR*)(tempptr->value), 
								    TRUE) );
      InsertList (opointList, (CHAR*)(tempptr->value));
    }	

  oname = m_pCScanner->FindIdentifier (oname, TRUE);
  AddConvexHull("[TCL Generated convexhull]", 0, 0, oname, opointList);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  if (surfid != NULL)
  {
    //printf ("surfid: %s\n",surfid);
    surfid = m_pCScanner->FindIdentifier (surfid, TRUE);
    SetNodeSurfaceID(surfid);
  }

  if (shading != NULL)
  {
    //printf ("shading: %s\n",shading);
    DynInt *pdyniShading = TranslateShadingFlag(shading);
    SetNodeShading (pdyniShading);
  }

  if (lod != NULL)
  {
    //printf ("LOD: %s\n",lod);
    DynInt *pdyniLOD = TranslateLODFlag(lod);
    SetNodeLOD (pdyniLOD);
  }

  if (solid != NULL)
  {
    //printf ("solid: %s\n",solid);
    DynInt *pdyniSolid = TranslateSolidFlag(solid);
    SetAtomNodeSolid (pdyniSolid);
  }


  //printf ("\n");

  if (ribbegin != NULL)
    {
      SetPrimitiveRIBBegin (ribbegin);
    }

  if (ribend != NULL)
    {
      SetPrimitiveRIBBegin (ribend);
    }

  EndConvexHull();

  // Ajax
  //
  TCLAjaxArgList (opoints);

  if (m_uiErrorCount > 0)
    {
      return TCL_ERROR;
    }

  return TCL_OK;
}

int CSLIDEParser::HandleAddGeneralFace(VOID* &pCreatedObject, int argc, char **argv)
{
  char *oname = NULL;
  struct argList *opoints = NULL;
  char *surfid = NULL;
  char *shading = NULL;
  char *lod = NULL;
  char *solid = NULL;
  char *terraintype = NULL;
  char *voronoi = NULL;
  char *accuracy = NULL;
  char *random = NULL;
  char *drawcontours = NULL;
  char *drawvertices = NULL;
  char *drawCDT = NULL;
  char *drawindividualvoronoi = NULL;
  char *drawvoronoi = NULL;
  char *drawheight = NULL;
  char *drawoffset = NULL;
  char *drawmillingcontour = NULL;
  char *millingcontour = NULL;
  char *cleanupcontour = NULL;
  DynInt *offsets = NULL;
  DynInt *individualvoronoimin = NULL;
  DynInt *individualvoronoimax = NULL;
  DynFloat *offsetmin = NULL;
  DynFloat *offsetmax = NULL;
  DynInt *subdivisionlevel = NULL;
  DynFloat *tolerance = NULL;
  char *ribbegin = NULL;
  char *ribend = NULL;

  CListInt *opointList=NULL;

  struct argFormat tclargform[] =
    { { NULL, "name",
	"Name for the new GeneralFace",
	ARG_SIMPLE | ARG_STRING, &oname, TRUE },
      { NULL, "point1 point2 point3 ...", 
	"List of points for GeneralFace",
	ARG_SIMPLE | ARG_LIST | ARG_STRING, &opoints, TRUE },
      { "solid", "solid_flag",
	"GeneralFace Solidity flag",
	ARG_STRING, &solid, FALSE },
      { "surface", "surf_id",
	"GeneralFace Surface ID",
	ARG_STRING, &surfid, FALSE },
      { "shading", "shading_flag",
	"Shading type",
	ARG_STRING, &shading, FALSE },
      { "lod", "lod_flag",
	"Level of Detail flag",
	ARG_STRING, &lod, FALSE },
      { "type", "type_flag",
	"Type of Delaunay Triangulation",
	ARG_STRING, &terraintype, FALSE },
      { "voronoi", "voronoi_flag",
	"Type of Voronoi Diagram",
	ARG_STRING, &voronoi, FALSE },
      { "accuracy", "accuracy_flag",
	"Type of Accuracy for numeric calculations",
	ARG_STRING, &accuracy, FALSE },
      { "random", "random_flag",
	"Randomize the algorithm",
	ARG_STRING, &random, FALSE },
      { "drawcontours", "drawcontours_flag",
	"Draw the contours",
	ARG_STRING, &drawcontours, FALSE },
      { "drawvertices", "drawvertices_flag",
	"Draw the contour vertices",
	ARG_STRING, &drawvertices, FALSE },
      { "drawCDT", "drawCDT_flag",
	"Draw the CDT",
	ARG_STRING, &drawCDT, FALSE },
      { "drawindividualvoronoi", "drawindividualvoronoi_flag",
	"Draw individual contour voronoi",
	ARG_STRING, &drawindividualvoronoi, FALSE },
      { "drawvoronoi", "drawvoronoi_flag",
	"Draw the contour voronoi",
	ARG_STRING, &drawvoronoi, FALSE },
      { "drawheight", "drawheight_flag",
	"Draw the contour height",
	ARG_STRING, &drawheight, FALSE },
      { "drawoffset", "drawoffset_flag",
	"Draw the contour offset",
	ARG_STRING, &drawoffset, FALSE },
      { "drawmillingcontour", "drawmillingcontour_flag",
	"Draw the contour millingcontour",
	ARG_STRING, &drawmillingcontour, FALSE },
      { "millingcontour", "millingcontour_flag",
	"Calculate the contour millingcontour",
	ARG_STRING, &millingcontour, FALSE },
      { "cleanupcontour", "cleanupcontour_flag",
	"Calculate the contour cleanupcontour",
	ARG_STRING, &cleanupcontour, FALSE },
      { "offsets", "int/intexpr",
	"# of offset contours",
	ARG_DYNINT, &offsets, FALSE },
      { "offsetmin", "float/floatexpr",
	"Offset distance min",
	ARG_DYNFLOAT, &offsetmin, FALSE },
      { "offsetmax", "float/floatexpr",
	"Offset distance max",
	ARG_DYNFLOAT, &offsetmax, FALSE },
      { "individualvoronoimin", "int/intexpr",
	"Min contour voronoi diagram to draw",
	ARG_DYNINT, &individualvoronoimin, FALSE },
      { "individualvoronoimax", "int/intexpr",
	"Max contour voronoi diagram to draw",
	ARG_DYNINT, &individualvoronoimax, FALSE },
      { "subdivisions", "int/intexpr",
	"Subdivisions of curves to draw",
	ARG_DYNINT, &subdivisionlevel, FALSE },
      { "tolerance", "float/floatexpr",
	"Tolerance of curve rendering",
	ARG_DYNFLOAT, &tolerance, FALSE },
      { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
      { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }

    };

  const int argCount = 29;

  //printf ("HandleAddGeneralFace: %d\n",argc);

  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create generalface: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  struct argList *tempptr;
  //printf ("oname: %s\n",oname);

  opointList = CreateList();
  //printf ("opoints: \n");
  for (tempptr = opoints; tempptr != NULL; tempptr = tempptr->next)
    {
      //printf ("	 %s\n", (CHAR *)(tempptr->value));
      tempptr->value = (VOID*)(m_pCScanner->FindIdentifier ((CHAR*)(tempptr->value), 
								    TRUE) );
      InsertList (opointList, (CHAR*)(tempptr->value));
    }	

  oname = m_pCScanner->FindIdentifier (oname, TRUE);
  AddGeneralFace ("[TCL Generated GeneralFace]", 0, 0, oname, opointList);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  if (surfid != NULL)
  {
    //printf ("surfid: %s\n",surfid);
    surfid = m_pCScanner->FindIdentifier (surfid, TRUE);
    SetNodeSurfaceID(surfid);
  }

  if (shading != NULL)
  {
    //printf ("shading: %s\n",shading);
    DynInt *pdyniShading = TranslateShadingFlag(shading);
    SetNodeShading (pdyniShading);
  }

  if (lod != NULL)
  {
    //printf ("LOD: %s\n",lod);
    DynInt *pdyniLOD = TranslateLODFlag(lod);
    SetNodeLOD (pdyniLOD);
  }

  if (solid != NULL)
  {
    //printf ("solid: %s\n",solid);
    DynInt *pdyniSolid = TranslateSolidFlag(solid);
    SetAtomNodeSolid (pdyniSolid);
  }

  if (terraintype != NULL)
    {
      DynInt *pdyniTerrainType = TranslateTerrainTypeFlag(terraintype);
      SetGeneralFaceType(pdyniTerrainType);
    }

  if (voronoi != NULL)
    {
      DynInt *pdyniVoronoi = TranslateVoronoiFlag(voronoi);
      SetGeneralFaceVoronoi(pdyniVoronoi);
    }

  if (accuracy != NULL)
    {
      DynInt *pdyniAccuracy = TranslateAccuracyFlag(accuracy);
      SetGeneralFaceAccuracy(pdyniAccuracy);
    }

  if (random != NULL)
    {
      DynInt *pdyniRandom = TranslateOnOffFlag(random);
      SetGeneralFaceRandom(pdyniRandom);
    }

  if (drawcontours != NULL)
    {
      DynInt *pdyniDrawContours = TranslateOnOffFlag(drawcontours);
      SetGeneralFaceDrawContours(pdyniDrawContours);
    }

  if (drawvertices != NULL)
    {
      DynInt *pdyniDrawVertices = TranslateOnOffFlag(drawvertices);
      SetGeneralFaceDrawVertices(pdyniDrawVertices);
    }

  if (drawCDT != NULL)
    {
      DynInt *pdyniDrawCDT = TranslateOnOffFlag(drawCDT);
      SetGeneralFaceDrawCDT(pdyniDrawCDT);
    }

  if (drawindividualvoronoi != NULL)
    {
      DynInt *pdyniDrawIndividualVoronoi = TranslateOnOffFlag(drawindividualvoronoi);
      SetGeneralFaceDrawIndividualVoronoi(pdyniDrawIndividualVoronoi);
    }

  if (drawvoronoi != NULL)
    {
      DynInt *pdyniDrawVoronoi = TranslateOnOffFlag(drawvoronoi);
      SetGeneralFaceDrawVoronoi(pdyniDrawVoronoi);
    }

  if (drawheight != NULL)
    {
      DynInt *pdyniDrawHeight = TranslateOnOffFlag(drawheight);
      SetGeneralFaceDrawHeight(pdyniDrawHeight);
    }

  if (drawoffset != NULL)
    {
      DynInt *pdyniDrawOffset = TranslateOnOffFlag(drawoffset);
      SetGeneralFaceDrawOffset(pdyniDrawOffset);
    }

  if (drawmillingcontour != NULL)
    {
      DynInt *pdyniDrawMillingContour = TranslateOnOffFlag(drawmillingcontour);
      SetGeneralFaceDrawMillingContour(pdyniDrawMillingContour);
    }

  if (millingcontour != NULL)
    {
      DynInt *pdyniMillingContour = TranslateOnOffFlag(millingcontour);
      SetGeneralFaceMillingContour(pdyniMillingContour);
    }

  if (cleanupcontour != NULL)
    {
      DynInt *pdyniCleanUpContour = TranslateOnOffFlag(cleanupcontour);
      SetGeneralFaceCleanUpContour(pdyniCleanUpContour);
    }

  if (offsets != NULL)
    {
      SetGeneralFaceOffsets(offsets);
    }

  if (offsetmin != NULL)
    {
      SetGeneralFaceOffsetMin(offsetmin);
    }

  if (offsetmax != NULL)
    {
      SetGeneralFaceOffsetMax(offsetmax);
    }

  if (individualvoronoimin != NULL)
    {
      SetGeneralFaceIndividualVoronoiMin(individualvoronoimin);
    }

  if (individualvoronoimax != NULL)
    {
      SetGeneralFaceIndividualVoronoiMax(individualvoronoimax);
    }

  if (subdivisionlevel != NULL)
    {
      SetGeneralFaceSubdivisionLevel(subdivisionlevel);
    }

  if (tolerance != NULL)
    {
      SetGeneralFaceTolerance(tolerance);
    }

  //printf ("\n");

  if (ribbegin != NULL)
    {
      SetPrimitiveRIBBegin (ribbegin);
    }

  if (ribend != NULL)
    {
      SetPrimitiveRIBBegin (ribend);
    }

  EndGeneralFace();

  // Ajax
  //
  TCLAjaxArgList (opoints);

  if (m_uiErrorCount > 0)
    {
      return TCL_ERROR;
    }

  return TCL_OK;
}

int CSLIDEParser::HandleAddContour(VOID* &pCreatedObject, int argc, char **argv)
{
  char *gfname = NULL;
  struct argList *opoints = NULL;
  CListInt *opointList=NULL;

  struct argFormat tclargform[] =
  { 
    { NULL, "generalfacename",
      "Name of a generalface to insert into",
      ARG_SIMPLE | ARG_STRING, &gfname, TRUE },
    { NULL, "point1 point2 point3 ...", 
      "List of points for GeneralFace",
      ARG_SIMPLE | ARG_LIST | ARG_STRING, &opoints, TRUE }
  };

  const int argCount = 2;
  
  //printf ("HandleAddInstance: %d\n",argc);
  
  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create instance: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }
  
  //printf ("gname: %s\n",gname);
  //printf ("childname: %s\n",childname);
  
  gfname = m_pCScanner->FindIdentifier (gfname, TRUE);
  SelectGeneralFace (gfname);
  if (m_pCSLIDEGeneralFaceCurr == NULL)
    {
      return TCL_ERROR;
    }
  
  struct argList *tempptr;

  opointList = CreateList();
  //printf ("opoints: \n");
  for (tempptr = opoints; tempptr != NULL; tempptr = tempptr->next)
    {
      //printf ("	 %s\n", (CHAR *)(tempptr->value));
      tempptr->value = (VOID*)(m_pCScanner->FindIdentifier ((CHAR*)(tempptr->value), 
								    TRUE) );
      InsertList (opointList, (CHAR*)(tempptr->value));
    }	

  InsertGeneralFaceContour(opointList);

  EndGeneralFace();
  
  if (m_uiErrorCount > 0)
    {
      return TCL_ERROR;
    }

  return TCL_OK;
}

int CSLIDEParser::HandleAddTerrain(VOID* &pCreatedObject, int argc, char **argv)
{
  char *oname = NULL;
  struct argList *opoints = NULL;
  char *surfid = NULL;
  char *shading = NULL;
  char *lod = NULL;
  char *solid = NULL;
  char *terraintype = NULL;
  char *voronoi = NULL;
  char *accuracy = NULL;
  char *random = NULL;
  char *drawdelaunay = NULL;
  char *drawvertices = NULL;
  char *ribbegin = NULL;
  char *ribend = NULL;
  CListInt *opointList=NULL;

  struct argFormat tclargform[] =
    { { NULL, "name",
	"Name for the new terrain",
	ARG_SIMPLE | ARG_STRING, &oname, TRUE },
      { NULL, "point1 point2 point3 ...", 
	"List of points for terrain",
	ARG_SIMPLE | ARG_LIST | ARG_STRING, &opoints, TRUE },
      { "solid", "solid_flag",
	"Terrain Solidity flag",
	ARG_STRING, &solid, FALSE },
      { "surface", "surf_id",
	"Terrain Surface ID",
	ARG_STRING, &surfid, FALSE },
      { "shading", "shading_flag",
	"Shading type",
	ARG_STRING, &shading, FALSE },
      { "lod", "lod_flag",
	"Level of Detail flag",
	ARG_STRING, &lod, FALSE },
      { "type", "type_flag",
	"Type of Delaunay Triangulation",
	ARG_STRING, &terraintype, FALSE },
      { "voronoi", "voronoi_flag",
	"Type of Voronoi Diagram",
	ARG_STRING, &voronoi, FALSE },
      { "accuracy", "accuracy_flag",
	"Type of Accuracy for numeric calculations",
	ARG_STRING, &accuracy, FALSE },
      { "random", "random_flag",
	"Randomize the algorithm",
	ARG_STRING, &random, FALSE },
      { "drawdelaunay", "drawdelaunay_flag",
	"Draw the Delaunay triangulation",
	ARG_STRING, &drawdelaunay, FALSE },
      { "drawvertices", "drawvertices_flag",
	"Draw the Delaunay vertices",
	ARG_STRING, &drawvertices, FALSE },

	  { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
	  { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }

    };

  const int argCount = 14;

  //printf ("HandleAddTerrain: %d\n",argc);

  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create terrain: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  struct argList *tempptr;
  //printf ("oname: %s\n",oname);

  opointList = CreateList();
  //printf ("opoints: \n");
  for (tempptr = opoints; tempptr != NULL; tempptr = tempptr->next)
    {
      //printf ("	 %s\n", (CHAR *)(tempptr->value));
      tempptr->value = (VOID*)(m_pCScanner->FindIdentifier ((CHAR*)(tempptr->value), 
								    TRUE) );
      InsertList (opointList, (CHAR*)(tempptr->value));
    }	

  oname = m_pCScanner->FindIdentifier (oname, TRUE);
  AddTerrain ("[TCL Generated Terrain]", 0, 0, oname, opointList);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  if (surfid != NULL)
  {
    //printf ("surfid: %s\n",surfid);
    surfid = m_pCScanner->FindIdentifier (surfid, TRUE);
    SetNodeSurfaceID(surfid);
  }

  if (shading != NULL)
  {
    //printf ("shading: %s\n",shading);
    DynInt *pdyniShading = TranslateShadingFlag(shading);
    SetNodeShading (pdyniShading);
  }

  if (lod != NULL)
  {
    //printf ("LOD: %s\n",lod);
    DynInt *pdyniLOD = TranslateLODFlag(lod);
    SetNodeLOD (pdyniLOD);
  }

  if (solid != NULL)
  {
    //printf ("solid: %s\n",solid);
    DynInt *pdyniSolid = TranslateSolidFlag(solid);
    SetAtomNodeSolid (pdyniSolid);
  }

  if (terraintype != NULL)
    {
      DynInt *pdyniTerrainType = TranslateTerrainTypeFlag(terraintype);
      SetTerrainType(pdyniTerrainType);
    }

  if (voronoi != NULL)
    {
      DynInt *pdyniVoronoi = TranslateVoronoiFlag(voronoi);
      SetTerrainVoronoi(pdyniVoronoi);
    }

  if (accuracy != NULL)
    {
      DynInt *pdyniAccuracy = TranslateAccuracyFlag(accuracy);
      SetTerrainAccuracy(pdyniAccuracy);
    }

  if (random != NULL)
    {
      DynInt *pdyniRandom = TranslateOnOffFlag(random);
      SetTerrainRandom(pdyniRandom);
    }

  if (drawdelaunay != NULL)
    {
      DynInt *pdyniDrawDelaunay = TranslateOnOffFlag(drawdelaunay);
      SetTerrainDrawDelaunay(pdyniDrawDelaunay);
    }

  if (drawvertices != NULL)
    {
      DynInt *pdyniDrawVertices = TranslateOnOffFlag(drawvertices);
      SetTerrainDrawVertices(pdyniDrawVertices);
    }


  //printf ("\n");

  if (ribbegin != NULL)
    {
      SetPrimitiveRIBBegin (ribbegin);
    }

  if (ribend != NULL)
    {
      SetPrimitiveRIBBegin (ribend);
    }

  EndTerrain();

  // Ajax
  //
  TCLAjaxArgList (opoints);

  if (m_uiErrorCount > 0)
    {
      return TCL_ERROR;
    }

  return TCL_OK;
}

int CSLIDEParser::HandleAddGroup(VOID* &pCreatedObject, int argc, char **argv)
{
  char *gname = NULL;
  char *surfid = NULL;
  char *shading = NULL;
  char *lod = NULL;
  char *ribbegin=NULL;
  char *ribend=NULL;

  struct argFormat tclargform[] =
    { { NULL, "groupname",
	"Name of new group",
	ARG_SIMPLE | ARG_STRING, &gname, TRUE },
	{ "surface", "surf_id",
	"Group Surface ID",
	ARG_STRING, &surfid, FALSE },
      { "shading", "shading_flag",
	"Shading type",
	ARG_STRING, &shading, FALSE },
      { "lod", "lod_flag",
	"Level of Detail flag",
	ARG_STRING, &lod, FALSE },

	  { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
	  { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }
    };

  const int argCount = 6;

  //printf ("HandleAddGroup: %d\n",argc);

  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create group: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  //printf ("gname: %s\n",gname);

  gname = m_pCScanner->FindIdentifier (gname, TRUE);
  AddGroup("[TCL Generated Group]", 0, 0, gname);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  if (surfid != NULL)
  {
    //printf ("surfid: %s\n",surfid);
    surfid = m_pCScanner->FindIdentifier (surfid, TRUE);
    SetNodeSurfaceID(surfid);
  }

  if (shading != NULL)
  {
    //printf ("shading: %s\n",shading);
    DynInt *pdyniShading = TranslateShadingFlag(shading);
    SetNodeShading (pdyniShading);
  }

  if (lod != NULL)
  {
    //printf ("LOD: %s\n",lod);
    DynInt *pdyniLOD = TranslateLODFlag(lod);
    SetNodeLOD (pdyniLOD);
  }

   
  //printf ("\n");

  if (ribbegin != NULL)
    {
      SetPrimitiveRIBBegin(ribbegin);
    }

  if (ribend != NULL)
    {
      SetPrimitiveRIBEnd(ribend);
    }

  EndGroup();

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleAddInstance(VOID* &pCreatedObject, int argc, char **argv)
{
  char *id = NULL;
  char *gname = NULL;
  //char *iname = NULL;
  char *childname = NULL;
  char *surfid = NULL;
  char *shading = NULL;
  char *lod = NULL;
  char *ribbegin=NULL;
  char *ribend=NULL;
  struct argList *xforms = NULL;

  struct argFormat tclargform[] =
  { { NULL, "id",
      "Name for the Instance",
      ARG_SIMPLE | ARG_STRING, &id, TRUE },
    { NULL, "groupname",
      "Name of group to insert into",
      ARG_SIMPLE | ARG_STRING, &gname, TRUE },
    { NULL, "childname",
      "Name of child entity",
      ARG_SIMPLE | ARG_STRING, &childname, TRUE },
    /*
    { "id", "instance_id",
      "Instance ID",
      ARG_STRING, &iname, FALSE },
      */
    { "surface", "surf_id",
      "Instance Surface ID",
      ARG_STRING, &surfid, FALSE },
    { "shading", "shading_flag",
      "Shading type",
      ARG_STRING, &shading, FALSE },
    { "lod", "lod_flag",
      "Level of Detail flag",
      ARG_STRING, &lod, FALSE },    
    { "ribbegin", "rib_string",
      "Begin RIB string",
      ARG_STRING, &ribbegin, FALSE },
	  { "ribend", "rib_string",
	    "End RIB string",
	ARG_STRING, &ribend, FALSE },
    
    /****************
      xform format:
      Translate:  {translate x y z}
      Scale:	     {scale x y z}
      Rotate:     {rotate x y z angle}
      Lookat:	 {lookat -eye {x y z} -target {x y z} -up {x y z} } 
      ****************/
    { "xforms", "xform1 xform2 xform3 ...", 
      "List of transforms (see docs for xform format)",
      ARG_LIST | ARG_STRING, &xforms, FALSE }
  };

  const int argCount = 9;
  
  //printf ("HandleAddInstance: %d\n",argc);
  
  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create instance: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }
  
  //printf ("gname: %s\n",gname);
  //printf ("childname: %s\n",childname);
  
  gname = m_pCScanner->FindIdentifier (gname, TRUE);
  SelectGroup (gname);
  if (m_pCSLIDEGroupCurr == NULL)
    {
      return TCL_ERROR;
    }
  
  id = m_pCScanner->FindIdentifier(id, TRUE);  
  childname = m_pCScanner->FindIdentifier(childname, TRUE);
  AddInstance ("[TCL Generated Instance]", 0, 0, childname);
  SetInstanceID(id);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;
  
  /*
  if (iname != NULL)
    {
      //printf ("iname: %s\n",iname);
      iname = m_pCScanner->FindIdentifier (iname, TRUE);
      SetInstanceID(iname);
    }
    */  

  if (surfid != NULL)
    {
      //printf ("surfid: %s\n",surfid);
      surfid = m_pCScanner->FindIdentifier (surfid, TRUE);
      SetNodeSurfaceID(surfid);
    }
  
  if (shading != NULL)
    {
      //printf ("shading: %s\n",shading);
      DynInt *pdyniShading = TranslateShadingFlag(shading);
      SetNodeShading (pdyniShading);
    }
  
  if (lod != NULL)
    {
      //printf ("LOD: %s\n",lod);
      DynInt *pdyniLOD = TranslateLODFlag(lod);
      SetNodeLOD (pdyniLOD);
    }
  
  struct argList *tempptr;

  //printf ("xforms: \n");
  for (tempptr = xforms; tempptr != NULL; tempptr = tempptr->next)
    {
      //printf ("	 %s\n", (CHAR *)(tempptr->value));
      
      struct argList *subptr;
      struct argList *subtempptr;
      
      subptr = TCLParseArgs_List((CHAR *)(tempptr->value), ARG_FLOAT);
      subtempptr = subptr->next;
      
      if (!strncmp ((CHAR*)(tempptr->value), "lookat", 6))
	{
	  struct argList *lookat_eye;
	  struct argList *lookat_target;
	  struct argList *lookat_up;
	  struct argList *lookat_templist;
	  char **lookat_argv;
	  int    lookat_argc;
	  
	  struct argFormat lookatform[] =
	  { { "eye", "{x y z}",
	      "eye x y z",
	      ARG_LIST | ARG_FLOAT, &lookat_eye, FALSE },
	    { "target", "{x y z}",
	      "target x y z",
	      ARG_LIST | ARG_FLOAT, &lookat_target, FALSE },
	    { "up", "{x y z}",
	      "up x y z",
	      ARG_LIST | ARG_FLOAT, &lookat_up, FALSE }
	  };
	  int lookat_argCount = 3;
	  
	  lookat_argc = TCLParseArgs_argv((CHAR *)(tempptr->value), 
					  &lookat_argv);
	  
	  int err = TCLParseArgs(lookatform, lookat_argCount, 
				 lookat_argc, lookat_argv);
	  if (err)
	    {
	      printf ("Syntax Error: slide create instance lookat xform: arg #%d: \"%s\"\n",
		      err, argv[err]);
	      return TCL_ERROR;
	    }
	  
	  AddLookAt ("[TCL Generated Instance:Lookat]", 0, 0);
	  
	  if (lookat_eye != NULL)
	    {
	      DynVector *pdynVectf = new DynVector;
	      DynFloat *pdynFloat[4];
	      
	      int axis = 0;
	      
	      lookat_templist = lookat_eye;
	      while (lookat_templist != NULL)
		{
		  pdynFloat[axis] = new DynFloat;
		  
		  if (lookat_templist->argType == ARG_EXPR)
		    {
		      //printf ("    %s\n", (CHAR *)(lookat_templist->value));
		      pdynFloat[axis]->SetDynExpr((CHAR *)(lookat_templist->value));
		    }
		  else
		    {
		      //printf ("    %f\n", *(FLOAT *)(lookat_templist->value));
		      pdynFloat[axis]->SetValue(*(FLOAT*)(lookat_templist->value));
		    }
		  
		  lookat_templist = lookat_templist->next;
		  axis++;
		  
		  // if too many things, just jump out of loop and throw err
		  if (axis > 3 && lookat_templist->next != NULL)
		    {
		      lookat_templist = NULL;
		      axis = -1;
		    }
		}

	      // delete sublist.
	      TCLAjaxArgList (lookat_eye);
	      
	      if (axis != 3)
		{
		  TCLAjaxArgList (xforms);
		  
		  HandleError(0, 0, "%s \"%s\" wrong # of arguments for lookat - eye", 
			      gcppcSLIDEReservedWords[sliderwtInstance],
			      gname);
		  
		  return TCL_ERROR;
		}
	      
	      pdynVectf->SetAll(pdynFloat[0], pdynFloat[1], pdynFloat[2]);	      
	      SetLookAtEye(pdynVectf);
	    }

	  if (lookat_target != NULL)
	    {
	      DynVector *pdynVectf = new DynVector;
	      DynFloat *pdynFloat[4];
	      
	      int axis = 0;
	      
	      lookat_templist = lookat_target;
	      while (lookat_templist != NULL)
		{
		  pdynFloat[axis] = new DynFloat;
		  
		  if (lookat_templist->argType == ARG_EXPR)
		    {
		      //printf ("    %s\n", (CHAR *)(lookat_templist->value));
		      pdynFloat[axis]->SetDynExpr((CHAR *)(lookat_templist->value));
		    }
		  else
		    {
		      //printf ("    %f\n", *(FLOAT *)(lookat_templist->value));
		      pdynFloat[axis]->SetValue(*(FLOAT*)(lookat_templist->value));
		    }
		  
		  lookat_templist = lookat_templist->next;
		  axis++;
		  
		  // if too many things, just jump out of loop and throw err
		  if (axis > 3)
		    {
		      lookat_templist = NULL;
		      axis = -1;
		    }
		}
	      
	      // delete sublist.
	      TCLAjaxArgList (lookat_target);
	      
	      if (axis != 3)
		{
		  TCLAjaxArgList (xforms);
		  HandleError(0, 0, "%s \"%s\" wrong # arguments for lookat - eye", 
			      gcppcSLIDEReservedWords[sliderwtInstance],
			      gname);
		  
		  return TCL_ERROR;
		}
	      
	      pdynVectf->SetAll(pdynFloat[0], pdynFloat[1], pdynFloat[2]);
	      
	      SetLookAtTarget(pdynVectf);
	    }
	  
	  if (lookat_up != NULL)
	    {
	      DynVector *pdynVectf = new DynVector;
	      DynFloat *pdynFloat[4];
	      
	      int axis = 0;
	      
	      lookat_templist = lookat_up;
	      while (lookat_templist != NULL)
		{
		  pdynFloat[axis] = new DynFloat;
		  
		  if (lookat_templist->argType == ARG_EXPR)
		    {
		      //printf ("    %s\n", (CHAR *)(lookat_templist->value));
		      pdynFloat[axis]->SetDynExpr((CHAR *)(lookat_templist->value));
		    }
		  else
		    {
		      //printf ("    %f\n", *(FLOAT *)(lookat_templist->value));
		      pdynFloat[axis]->SetValue(*(FLOAT*)(lookat_templist->value));
		    }
		  
		  lookat_templist = lookat_templist->next;
		  axis++;
		  
		  // if too many things, just jump out of loop and throw err
		  if (axis > 3)
		    {
		      lookat_templist = NULL;
		      axis = -1;
		    }
		}
		  
	      // delete sublist.
	      TCLAjaxArgList (lookat_up);
	      
	      if (axis != 3)
		{
		  TCLAjaxArgList (xforms);
		  HandleError(0, 0, "%s \"%s\" wrong # of arguments for lookat - eye", 
			      gcppcSLIDEReservedWords[sliderwtInstance],
			      gname);
		  
		  return TCL_ERROR;
		} 

	      pdynVectf->SetAll(pdynFloat[0], pdynFloat[1], pdynFloat[2]);	      
	      SetLookAtUp(pdynVectf);
	    }	  
	  free (lookat_argv);
	  
	  EndLookAt();	  
	} // if LOOKAT
      else
	{
	  DynVector *pdynVectf = new DynVector;
	  DynFloat *pdynFloat[4];
	  
	  int axis = 0;
	  
	  while (subtempptr != NULL)
	    {
	      pdynFloat[axis] = new DynFloat;
	      
	      if (subtempptr->argType == ARG_EXPR)
		{
		  //printf ("    %s\n", (CHAR *)(subtempptr->value));
		  pdynFloat[axis]->SetDynExpr((CHAR *)(subtempptr->value));
		}
	      else
		{
		  //printf ("    %f\n", *(FLOAT *)(subtempptr->value));
		  pdynFloat[axis]->SetValue(*(FLOAT*)(subtempptr->value));
		}
	      
	      subtempptr = subtempptr->next;
	      axis++;
	      // if too many things, just jump out of loop and throw err
	      if (axis > 3 && subtempptr != NULL)
		{
		  subtempptr = NULL;
		  axis = -1;
		}
	    }
	  
	  if (axis < 3)
	    {
	      TCLAjaxArgList (xforms);
	      HandleError(0, 0, "%s \"%s\" wrong # of arguments for transform", 
			  gcppcSLIDEReservedWords[sliderwtInstance],
			  gname);
	      
	      return TCL_ERROR;	      
	    }
	  
	  pdynVectf->SetAll(pdynFloat[0], pdynFloat[1], pdynFloat[2]);
	  
	  if (!strncmp ((CHAR*)(tempptr->value), "translate ", 10))
	    {
	      if (axis != 3)
		{
		  TCLAjaxArgList (xforms);
		  HandleError(0, 0, "%s \"%s\" wrong # of arguments for translate", 
			      gcppcSLIDEReservedWords[sliderwtInstance],
			      gname);
		  
		  return TCL_ERROR;
		}
	      
	      AddTranslate("[TCL Generated Instance:Translate]",0,0, pdynVectf);
	    }	  
	  else if (!strncmp ((CHAR*)(tempptr->value), "scale ", 6))
	    {
	      if (axis != 3)
		{
		  TCLAjaxArgList (xforms);
		  HandleError(0, 0, "%s \"%s\" wrong # of arguments for scale", 
			      gcppcSLIDEReservedWords[sliderwtInstance],
			      gname);
		  
		  return TCL_ERROR;
		}
	      
	      AddScale("[TCL Generated Instance:Scale]",0,0, pdynVectf);
	    }
	  else if (!strncmp ((CHAR*)(tempptr->value), "rotate ", 7))
	    {
	      if (axis != 4)
		{
		  TCLAjaxArgList (xforms);
		  HandleError(0, 0, "%s \"%s\" wrong # of arguments for rotate", 
					  gcppcSLIDEReservedWords[sliderwtInstance],
			      gname);
		  
		  return TCL_ERROR;
		}
	      AddRotate("[TCL Generated Instance:Rotate]",0,0, pdynVectf, pdynFloat[3]);
	    }
	  else
	    {
	      TCLAjaxArgList (xforms);
	      printf ("Parse Error in addinstance: Unknown transform type in \"%s\"\n",
		      (CHAR*)(tempptr->value));
	      return TCL_ERROR;
	    }
	}
      // delete sublist.
      TCLAjaxArgList (subptr);
    }
  
  //printf ("\n");
  if (ribbegin != NULL)
    SetPrimitiveRIBBegin (ribbegin);
  
  if (ribend != NULL)
    SetPrimitiveRIBBegin (ribend);
  
  EndInstance();
  EndGroup();
  
  // Ajax
  //
  TCLAjaxArgList (xforms);

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleAddSGRoot(VOID* &pCreatedObject, int argc, char **argv)
{
  char *id = NULL;
  char *sgroot = NULL;
  char *surfid = NULL;
  char *shading = NULL;
  char *lod = NULL;
  char *solid = NULL;
  char *ribbegin=NULL;
  char *ribend=NULL;

  struct argFormat tclargform[] =
    { { NULL, "id",
	"Name for the Sphere",
	ARG_SIMPLE | ARG_STRING, &id, TRUE },
      { NULL, "SGRoot",
	"Name for the Scene Graph Root Node",
	ARG_SIMPLE | ARG_STRING, &sgroot, TRUE },
      { "surface", "surfid",
	"Shurface ID",
	ARG_STRING, &surfid, FALSE },
      { "shading", "shading_flag",
	"Shading type",
	ARG_STRING, &shading, FALSE },
      { "lod", "lod_flag",
	"Level of Detail flag",
	ARG_STRING, &lod, FALSE },
      { "solid", "solid_flag",
	"Object Solidity flag",
	ARG_STRING, &solid, FALSE },
      { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
      { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }
  };


  const int argCount = 8;

  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create sgroot: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  id = m_pCScanner->FindIdentifier (id, TRUE);  
  sgroot = m_pCScanner->FindIdentifier (sgroot, TRUE);  
  AddSGRoot ("[TCL Generated SGRoot]", 0, 0, id, sgroot);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  if (surfid != NULL)
    {
      surfid = m_pCScanner->FindIdentifier (surfid, TRUE);
      SetNodeSurfaceID(surfid);
    }

  if (shading != NULL)
  {
	DynInt *pdyniShading = TranslateShadingFlag(shading);
	SetNodeShading (pdyniShading);
  }

  if (lod != NULL)
  {
	DynInt *pdyniLOD = TranslateLODFlag(lod);
	SetNodeLOD (pdyniLOD);
  }

  if (solid != NULL)
  {
	DynInt *pdyniSolid = TranslateSolidFlag(solid);
	SetAtomNodeSolid (pdyniSolid);
  }

  if (ribbegin != NULL)
    SetPrimitiveRIBBegin (ribbegin);

  if (ribend != NULL)
    SetPrimitiveRIBBegin (ribend);

  EndSGRoot();

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleAddSphere(VOID* &pCreatedObject, int argc, char **argv)
{
  char *id = NULL;
  DynFloat *radius = NULL;
  DynFloat *zmin = NULL;
  DynFloat *zmax = NULL;
  DynFloat *thetamax = NULL;
  DynInt *zslices = NULL;
  DynInt *thetaslices = NULL;
  char *surfid = NULL;
  char *shading = NULL;
  char *lod = NULL;
  char *solid = NULL;
  argList *tex_coords = NULL;
  char *ribbegin=NULL;
  char *ribend=NULL;

  struct argFormat tclargform[] =
    { { NULL, "id",
	"Name for the Sphere",
	ARG_SIMPLE | ARG_STRING, &id, TRUE },
      { "radius", "float/floatexpr",
	"Radius of the sphere",
	ARG_DYNFLOAT, &radius, FALSE },
      { "zmin", "float/floatexpr",
	"Z-Min",
	ARG_DYNFLOAT, &zmin, FALSE },
      { "zmax", "float/floatexpr",
	"Z-max",
	ARG_DYNFLOAT, &zmax, FALSE },
      { "thetamax", "float/floatexpr",
	"Theta-max",
	ARG_DYNFLOAT, &thetamax, FALSE },
      { "zslices", "int/intexpr",
	"Z-slices",
	ARG_DYNINT, &zslices, FALSE },
      { "thetaslices", "int/intexpr",
	"Theta-slices",
	ARG_DYNINT, &thetaslices, FALSE },
      { "surface", "surfid",
	"Shurface ID",
	ARG_STRING, &surfid, FALSE },
      { "shading", "shading_flag",
	"Shading type",
	ARG_STRING, &shading, FALSE },
      { "texture", "{ texture_triple } [{ texture_triple }] { texture_triple } [{ texture_triple }]",
	"Texture coordinates",
	ARG_STRING | ARG_LIST, &tex_coords, FALSE},
      { "lod", "lod_flag",
	"Level of Detail flag",
	ARG_STRING, &lod, FALSE },
      { "solid", "solid_flag",
	"Object Solidity flag",
	ARG_STRING, &solid, FALSE },
      { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
      { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }
  };


  const int argCount = 14;

  //printf ("HandleAddSphere: %d\n",argc);

  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create sphere: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  id = m_pCScanner->FindIdentifier (id, TRUE);  
  AddSphere ("[TCL Generated Sphere]", 0, 0, id);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  if (radius != NULL)
    SetSphereRadius(radius);

  if (zmin != NULL)
    SetSphereZMin(zmin);

  if (zmax != NULL)
    SetSphereZMax(zmax);

  if (thetamax != NULL)
    SetSphereThetaMax(thetamax);

  if (zslices != NULL)
    SetSphereZSlices(zslices);

  if (thetaslices != NULL)
    SetSphereThetaSlices(thetaslices);

  if (surfid != NULL)
    {
      surfid = m_pCScanner->FindIdentifier (surfid, TRUE);
      SetNodeSurfaceID(surfid);
    }

  if (shading != NULL)
  {
	DynInt *pdyniShading = TranslateShadingFlag(shading);
	SetNodeShading (pdyniShading);
  }

  if (lod != NULL)
  {
	DynInt *pdyniLOD = TranslateLODFlag(lod);
	SetNodeLOD (pdyniLOD);
  }

  if (solid != NULL)
  {
	DynInt *pdyniSolid = TranslateSolidFlag(solid);
	SetAtomNodeSolid (pdyniSolid);
  }


  if (tex_coords != NULL)
  {
    struct argList *subptr = tex_coords;
    struct argList *tc_values = NULL;
    DynVector *tex_vects[4];
    int tex_vectnum=0;

    while (subptr != NULL)
      {
	struct argList *tc_temp;
	int axis=0;
	DynFloat *tempdfloats[3];

	tc_values = TCLParseArgs_List ((CHAR *)(subptr->value), ARG_FLOAT);
	
	for (tc_temp=tc_values; tc_temp != NULL; tc_temp=tc_temp->next)
	  {
	    tempdfloats[axis] = new DynFloat;
	    
	    if (tc_temp->argType == ARG_EXPR)	      
	      {	
		tempdfloats[axis]->SetDynExpr((CHAR *)(tc_temp->value));
	      }	
	    else	
	      {
		tempdfloats[axis]->SetValue(*(FLOAT*)(tc_temp->value));
	      }
      
	    axis++;
	    // if too many things, just jump out of loop and throw err
	    if (axis > 3 && tc_temp->next!=NULL)
	      {
		axis = -1;
		break;
	      }
	  }

	if (axis == 2)
	{
	   tempdfloats[2] = new DynFloat;
	   tempdfloats[2]->SetValue((FLOAT)0.0);
	}
	else if (axis != 3)
	  {
	    TCLAjaxArgList (tex_coords);
	    HandleError(0, 0, "%s \"%s\" wrong # of arguments for Sphere texture coordinates", 
			gcppcSLIDEReservedWords[sliderwtSphere],
			id);
      
	    return TCL_ERROR;
	  }

	tex_vects[tex_vectnum] = new DynVector;
	tex_vects[tex_vectnum]->SetAll (tempdfloats[0], tempdfloats[1],
				       tempdfloats[2]);	

	for (;axis >0; axis--)
	  delete tempdfloats[axis-1];
	
	tex_vectnum ++;
	subptr = subptr->next;

	// if too many things, just jump out of loop and throw err
	if (tex_vectnum > 4 && subptr->next != NULL)
	  {
	    subptr = NULL;
	    tex_vectnum = -1;
	  }

      }

    // if only 2 given, generate the other two.
    if (tex_vectnum == 2)
      {
	tex_vects[2] = tex_vects[1];
	tex_vects[1] = new DynVector;
	tex_vects[1]->Set ((*tex_vects[2])(0), (*tex_vects[0])(1), 0.0);
	tex_vects[1]->SetDynExprs (tex_vects[2]->GetDynExpr(0),
				   tex_vects[0]->GetDynExpr(1), NULL);
	
	tex_vects[3] = new DynVector;
	tex_vects[3]->Set ((*tex_vects[0])(0), (*tex_vects[2])(1), 0.0);
	tex_vects[3]->SetDynExprs (tex_vects[0]->GetDynExpr(0),
				   tex_vects[2]->GetDynExpr(1), NULL);
      }

    else if (tex_vectnum != 4)
      {
	TCLAjaxArgList (tex_coords);
	HandleError(0, 0, "%s \"%s\" wrong # of arguments for Sphere texture points", 
		    gcppcSLIDEReservedWords[sliderwtSphere],
		    id);
      
	return TCL_ERROR;
      }

    SetSphereTexture(tex_vects[0], tex_vects[1], tex_vects[2], tex_vects[3]);
  }


  //printf ("\n");

  if (ribbegin != NULL)
    SetPrimitiveRIBBegin (ribbegin);

  if (ribend != NULL)
    SetPrimitiveRIBBegin (ribend);


  
  EndSphere();

  // Ajax
  //
  TCLAjaxArgList (tex_coords);    

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleAddCylinder(VOID* &pCreatedObject, int argc, char **argv)
{
  char *id = NULL;
  DynFloat *radius = NULL;
  DynFloat *zmin = NULL;
  DynFloat *zmax = NULL;
  DynFloat *thetamax = NULL;
  DynInt *zslices = NULL;
  DynInt *thetaslices = NULL;
  char *surfid = NULL;
  char *shading = NULL;
  char *lod = NULL;
  char *solid = NULL;
  char *ribbegin=NULL;
  char *ribend=NULL;
  struct argList *tex_coords=NULL;

   struct argFormat tclargform[] =
    { { NULL, "id",
	"Name for the Cylinder",
	ARG_SIMPLE | ARG_STRING, &id, TRUE },
      { "radius", "float/floatexpr",
	"Radius of the Cylinder",
	ARG_DYNFLOAT, &radius, FALSE },
      { "zmin", "float/floatexpr",
	"Z-Min",
	ARG_DYNFLOAT, &zmin, FALSE },
      { "zmax", "float/floatexpr",
	"Z-max",
	ARG_DYNFLOAT, &zmax, FALSE },
      { "thetamax", "float/floatexpr",
	"Theta-max",
	ARG_DYNFLOAT, &thetamax, FALSE },
      { "zslices", "int/intexpr",
	"Z-slices",
	ARG_DYNINT, &zslices, FALSE },
      { "thetaslices", "int/intexpr",
	"Theta-slices",
	ARG_DYNINT, &thetaslices, FALSE },
      { "texture", "{ texture_triple } [{ texture_triple }] { texture_triple } [{ texture_triple }]",
	"Texture coordinates",
	ARG_STRING | ARG_LIST, &tex_coords, FALSE},
      { "surface", "surfid",
	"Shurface ID",
	ARG_STRING, &surfid, FALSE },
      { "shading", "shading_flag",
	"Shading type",
	ARG_STRING, &shading, FALSE },
      { "lod", "lod_flag",
	"Level of Detail flag",
	ARG_STRING, &lod, FALSE },
      { "solid", "solid_flag",
	"Object Solidity flag",
	ARG_STRING, &solid, FALSE },

	  { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
	  { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }

  };

  const int argCount = 14;

  
  //printf ("HandleAddCylinder: %d\n",argc);

  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create cylinder: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  id = m_pCScanner->FindIdentifier (id, TRUE);    
  AddCylinder( "[TCL Generated Cylinder]", 0, 0, id);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  if (radius != NULL)
    SetCylinderRadius(radius);

  if (zmin != NULL)
    SetCylinderZMin(zmin);

  if (zmax != NULL)
    SetCylinderZMax(zmax);

  if (thetamax != NULL)
    SetCylinderThetaMax(thetamax);

  if (zslices != NULL)
    SetCylinderZSlices(zslices);

  if (thetaslices != NULL)
    SetCylinderThetaSlices(thetaslices);

  if (surfid != NULL)
    {
      surfid = m_pCScanner->FindIdentifier (surfid, TRUE);
      SetNodeSurfaceID(surfid);
    }

  if (shading != NULL)
  {
	DynInt *pdyniShading = TranslateShadingFlag(shading);
	SetNodeShading (pdyniShading);
  }

  if (lod != NULL)
  {
	DynInt *pdyniLOD = TranslateLODFlag(lod);
	SetNodeLOD (pdyniLOD);
  }

  if (solid != NULL)
  {
	DynInt *pdyniSolid = TranslateSolidFlag(solid);
	SetAtomNodeSolid (pdyniSolid);
  }


  if (tex_coords != NULL)
  {
    struct argList *subptr = tex_coords;
    struct argList *tc_values = NULL;
    DynVector *tex_vects[4];
    int tex_vectnum=0;

    while (subptr != NULL)
      {
	struct argList *tc_temp;
	int axis=0;
	DynFloat *tempdfloats[3];

	tc_values = TCLParseArgs_List ((CHAR *)(subptr->value), ARG_FLOAT);
	
	for (tc_temp=tc_values; tc_temp != NULL; tc_temp=tc_temp->next)
	  {
	    tempdfloats[axis] = new DynFloat;
	    
	    if (tc_temp->argType == ARG_EXPR)	      
	      {	
		//printf ("  %s\n", (CHAR *)(tc_temp->value));
		tempdfloats[axis]->SetDynExpr((CHAR *)(tc_temp->value));
	      }	
	    else	
	      {
		//printf ("  %f\n", *(FLOAT *)(tc_temp->value));
		tempdfloats[axis]->SetValue(*(FLOAT*)(tc_temp->value));
	      }
      
	    axis++;

	    // if too many things, just jump out of loop and throw err
	    if (axis > 2 && tc_temp->next != NULL)
	      {
		axis = -1;
		break;
	      }

	  }

	if (axis == 2)
	{
	   tempdfloats[2] = new DynFloat;
	   tempdfloats[2]->SetValue((FLOAT)0.0);
	}
	else if (axis != 3)
	  {
	    TCLAjaxArgList (tex_coords);
	    HandleError(0, 0, "%s \"%s\" wrong # of arguments for Cylinder texture coordinates", 
			gcppcSLIDEReservedWords[sliderwtCylinder],
			id);
      
	    return TCL_ERROR;
	  }

	tex_vects[tex_vectnum] = new DynVector;
	tex_vects[tex_vectnum]->SetAll (tempdfloats[0], tempdfloats[1],
				       tempdfloats[2]);	

	for (;axis >0; axis--)
	  delete tempdfloats[axis-1];
	
	tex_vectnum ++;
	subptr = subptr->next;

	// if too many things, just jump out of loop and throw err
	if (tex_vectnum > 3 && subptr != NULL)
	  {
	    tex_vectnum = -1;
	    subptr = NULL;
	  }
      }

    // if only 2 given, generate the other two.
    if (tex_vectnum == 2)
      {
	tex_vects[2] = tex_vects[1];
	tex_vects[1] = new DynVector;
	tex_vects[1]->Set ((*tex_vects[2])(0), (*tex_vects[0])(1), 0.0);
	tex_vects[1]->SetDynExprs (tex_vects[2]->GetDynExpr(0),
				   tex_vects[0]->GetDynExpr(1), NULL);
	
	tex_vects[3] = new DynVector;
	tex_vects[3]->Set ((*tex_vects[0])(0), (*tex_vects[2])(1), 0.0);
	tex_vects[3]->SetDynExprs (tex_vects[0]->GetDynExpr(0),
				   tex_vects[2]->GetDynExpr(1), NULL);
      }

    else if (tex_vectnum != 4)
      {
	TCLAjaxArgList (tex_coords);
	HandleError(0, 0, "%s \"%s\" wrong # of arguments for Cylinder texture points", 
		    gcppcSLIDEReservedWords[sliderwtCylinder],
		    id);
      
	return TCL_ERROR;
      }

    SetCylinderTexture(tex_vects[0], tex_vects[1], tex_vects[2], tex_vects[3]);
				       
  }


  //printf ("\n");

  if (ribbegin != NULL)
    SetPrimitiveRIBBegin (ribbegin);

  if (ribend != NULL)
    SetPrimitiveRIBBegin (ribend);


  
  EndCylinder();

  // Ajax
  //
  TCLAjaxArgList (tex_coords);    

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleAddCone(VOID* &pCreatedObject, int argc, char **argv)
{
  char *id = NULL;
  DynFloat *radius = NULL;
  DynFloat *height = NULL;
  DynFloat *thetamax = NULL;
  DynInt *zslices = NULL;
  DynInt *thetaslices = NULL;
  char *surfid = NULL;
  char *shading = NULL;
  char *lod = NULL;
  char *solid = NULL;
  char *ribbegin=NULL;
  char *ribend=NULL;
  struct argList *tex_coords=NULL;

  struct argFormat tclargform[] =
    { { NULL, "id",
	"Name for the Cone",
	ARG_SIMPLE | ARG_STRING, &id, TRUE },
      { "radius", "float/floatexpr",
	"Radius of the Cone",
	ARG_DYNFLOAT, &radius, FALSE },
      { "height", "float/floatexpr",
	"Height of the Cone",
	ARG_DYNFLOAT, &height, FALSE },
      { "thetamax", "float/floatexpr",
	"Theta-max",
	ARG_DYNFLOAT, &thetamax, FALSE },
      { "zslices", "int/intexpr",
	"Z-slices",
	ARG_DYNINT, &zslices, FALSE },
      { "thetaslices", "int/intexpr",
	"Theta-slices",
	ARG_DYNINT, &thetaslices, FALSE },
      { "surface", "surfid",
	"Shurface ID",
	ARG_STRING, &surfid, FALSE },
      { "texture", "{ texture_triple } [{ texture_triple }] { texture_triple } [{ texture_triple }]",
	"Texture coordinates",
	ARG_STRING | ARG_LIST, &tex_coords, FALSE},
      { "shading", "shading_flag",
	"Shading type",
	ARG_STRING, &shading, FALSE },
      { "lod", "lod_flag",
	"Level of Detail flag",
	ARG_STRING, &lod, FALSE },
      { "solid", "solid_flag",
	"Object Solidity flag",
	ARG_STRING, &solid, FALSE },

	  { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
	  { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }

  };

  const int argCount = 12;

  
  //printf ("HandleAddCone: %d\n",argc);

  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create cone: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  id = m_pCScanner->FindIdentifier (id, TRUE);    
  AddCone( "[TCL Generated Cone]", 0, 0, id);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;  

  if (radius != NULL)
    SetConeRadius(radius);

  if (height != NULL)
    SetConeHeight(height);

  if (thetamax != NULL)
    SetConeThetaMax(thetamax);

  if (zslices != NULL)
    SetConeZSlices(zslices);

  if (thetaslices != NULL)
    SetConeThetaSlices(thetaslices);

  if (surfid != NULL)
    {
      surfid = m_pCScanner->FindIdentifier (surfid, TRUE);  
      SetNodeSurfaceID(surfid);
    }

  if (shading != NULL)
  {
	DynInt *pdyniShading = TranslateShadingFlag(shading);
	SetNodeShading (pdyniShading);
  }

  if (lod != NULL)
  {
	DynInt *pdyniLOD = TranslateLODFlag(lod);
	SetNodeLOD (pdyniLOD);
  }

  if (solid != NULL)
  {
	DynInt *pdyniSolid = TranslateSolidFlag(solid);
	SetAtomNodeSolid (pdyniSolid);
  }


  if (tex_coords != NULL)
  {
    struct argList *subptr = tex_coords;
    struct argList *tc_values = NULL;
    DynVector *tex_vects[4];
    int tex_vectnum=0;

    while (subptr != NULL)
      {
	struct argList *tc_temp;
	int axis=0;
	DynFloat *tempdfloats[3];

	tc_values = TCLParseArgs_List ((CHAR *)(subptr->value), ARG_FLOAT);
	
	for (tc_temp=tc_values; tc_temp != NULL; tc_temp=tc_temp->next)
	  {
	    tempdfloats[axis] = new DynFloat;
	    
	    if (tc_temp->argType == ARG_EXPR)	      
	      {	
		tempdfloats[axis]->SetDynExpr((CHAR *)(tc_temp->value));
	      }	
	    else	
	      {
		tempdfloats[axis]->SetValue(*(FLOAT*)(tc_temp->value));
	      }
      
	    axis++;
	    // if too many things, just jump out of loop and throw err
	    if (axis > 2 && tc_temp->next != NULL)
	      {
		axis = -1;
		break;
	      }
	  }

	if (axis == 2)
	{
	   tempdfloats[2] = new DynFloat;
	   tempdfloats[2]->SetValue((FLOAT)0.0);
	}
	else if (axis != 3)
	  {
	    TCLAjaxArgList (tex_coords);
	    HandleError(0, 0, "%s \"%s\" wrong # of arguments for Cone texture coordinates", 
			gcppcSLIDEReservedWords[sliderwtCone],
			id);
      
	    return TCL_ERROR;
	  }

	tex_vects[tex_vectnum] = new DynVector;
	tex_vects[tex_vectnum]->SetAll (tempdfloats[0], tempdfloats[1],
				       tempdfloats[2]);	

	for (;axis >0; axis--)
	  delete tempdfloats[axis-1];
	
	tex_vectnum ++;
	subptr = subptr->next;
	// if too many things, just jump out of loop and throw err
	if (tex_vectnum > 3 && subptr != NULL)
	  {
	    tex_vectnum = -1;
	    subptr = NULL;
	  }
      }

    // if only 2 given, generate the other two.
    if (tex_vectnum == 2)
      {
	tex_vects[2] = tex_vects[1];
	tex_vects[1] = new DynVector;
	tex_vects[1]->Set ((*tex_vects[2])(0), (*tex_vects[0])(1), 0.0);
	tex_vects[1]->SetDynExprs (tex_vects[2]->GetDynExpr(0),
				   tex_vects[0]->GetDynExpr(1), NULL);
	
	tex_vects[3] = new DynVector;
	tex_vects[3]->Set ((*tex_vects[0])(0), (*tex_vects[2])(1), 0.0);
	tex_vects[3]->SetDynExprs (tex_vects[0]->GetDynExpr(0),
				   tex_vects[2]->GetDynExpr(1), NULL);
      }

    else if (tex_vectnum != 4)
      {
	TCLAjaxArgList (tex_coords);
	HandleError(0, 0, "%s \"%s\" wrong # of arguments for Cone texture points", 
		    gcppcSLIDEReservedWords[sliderwtCone],
		    id);
      
	return TCL_ERROR;
      }

    SetConeTexture(tex_vects[0], tex_vects[1], tex_vects[2], tex_vects[3]);
  }



  if (ribbegin != NULL)
    SetPrimitiveRIBBegin (ribbegin);

  if (ribend != NULL)
    SetPrimitiveRIBBegin (ribend);


  EndCone();

  // Ajax
  //
  TCLAjaxArgList (tex_coords);    

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleAddTorus(VOID* &pCreatedObject, int argc, char **argv)
{
  char *id = NULL;
  DynFloat *majorradius = NULL;
  DynFloat *minorradius = NULL;
  DynFloat *thetamax = NULL;
  DynFloat *phimin = NULL;
  DynFloat *phimax = NULL; 
  DynInt *thetaslices = NULL;
  DynInt *phislices = NULL;
  char *surfid = NULL;
  char *shading = NULL;
  char *lod = NULL;
  char *solid = NULL;
  char *ribbegin = NULL;
  char *ribend = NULL;
  struct argList *tex_coords = NULL;


  struct argFormat tclargform[] =
    { { NULL, "id",
	"Name for the Torus",
	ARG_SIMPLE | ARG_STRING, &id, TRUE },
      { "majorradius", "float/floatexpr",
	"Major Radius of the Torus",
	ARG_DYNFLOAT, &majorradius, FALSE },
      { "minorradius", "float/floatexpr",
	"Minor Radius of the Torus",
	ARG_DYNFLOAT, &minorradius, FALSE },
      { "thetamax", "float/floatexpr",
	"Theta-max",
	ARG_DYNFLOAT, &thetamax, FALSE },
      { "phimin", "float/floatexpr",
	"phi-min",
	ARG_DYNFLOAT, &phimin, FALSE },
      { "phimax", "float/floatexpr",
	"phi-max",
	ARG_DYNFLOAT, &phimax, FALSE },
      { "thetaslices", "int/intexpr",
	"Theta-slices",
	ARG_DYNINT, &thetaslices, FALSE },
      { "phislices", "int/intexpr",
	"Phi-slices",
	ARG_DYNINT, &phislices, FALSE },
      { "texture", "{ texture_triple } [{ texture_triple }] { texture_triple } [{ texture_triple }]",
	"Texture coordinates",
	ARG_STRING | ARG_LIST, &tex_coords, FALSE},
      { "surface", "surfid",
	"Surface ID",
	ARG_STRING, &surfid, FALSE },
      { "shading", "shading_flag",
	"Shading type",
	ARG_STRING, &shading, FALSE },
      { "lod", "lod_flag",
	"Level of Detail flag",
	ARG_STRING, &lod, FALSE },
      { "solid", "solid_flag",
	"Object Solidity flag",
	ARG_STRING, &solid, FALSE },

      { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
      { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }
  };

  const int argCount = 15;

  
  //printf ("HandleAddTorus: %d\n",argc);

  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create torus: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  id = m_pCScanner->FindIdentifier (id, TRUE);    
  AddTorus( "[TCL Generated Torus]", 0, 0, id);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;
  
  if (majorradius != NULL)
    SetTorusMajorRadius(majorradius);

  if (minorradius != NULL)
    SetTorusMinorRadius(minorradius);

  if (thetamax != NULL)
    SetTorusThetaMax(thetamax);

  if (phimin != NULL)
    SetTorusPhiMin(phimin);

  if (phimax != NULL)
    SetTorusPhiMax(phimax);

  if (thetaslices != NULL)
    SetTorusThetaSlices(thetaslices);

  if (phislices != NULL)
    SetTorusPhiSlices(phislices);

  if (surfid != NULL)
    {
      surfid = m_pCScanner->FindIdentifier (surfid, TRUE);
      SetNodeSurfaceID(surfid);
    }

  if (shading != NULL)
  {
	DynInt *pdyniShading = TranslateShadingFlag(shading);
	SetNodeShading (pdyniShading);
  }

  if (lod != NULL)
  {
	DynInt *pdyniLOD = TranslateLODFlag(lod);
	SetNodeLOD (pdyniLOD);
  }

  if (solid != NULL)
  {
	DynInt *pdyniSolid = TranslateSolidFlag(solid);
	SetAtomNodeSolid (pdyniSolid);
  }

  if (tex_coords != NULL)
  {
    struct argList *subptr = tex_coords;
    struct argList *tc_values = NULL;
    DynVector *tex_vects[4];
    int tex_vectnum=0;

    while (subptr != NULL)
      {
	struct argList *tc_temp;
	int axis=0;
	DynFloat *tempdfloats[3];

	tc_values = TCLParseArgs_List ((CHAR *)(subptr->value), ARG_FLOAT);
	
	for (tc_temp=tc_values; tc_temp != NULL; tc_temp=tc_temp->next)
	  {
	    tempdfloats[axis] = new DynFloat;
	    
	    if (tc_temp->argType == ARG_EXPR)	      
	      {	
		tempdfloats[axis]->SetDynExpr((CHAR *)(tc_temp->value));
	      }	
	    else	
	      {
		tempdfloats[axis]->SetValue(*(FLOAT*)(tc_temp->value));
	      }
      
	    axis++;
	    // if too many things, just jump out of loop and throw err
	    if (axis > 2 && tc_temp->next != NULL)
	      {
	        axis = -1;
		break;
	      }
	  }

	if (axis == 2)
	{
	   tempdfloats[2] = new DynFloat;
	   tempdfloats[2]->SetValue((FLOAT)0.0);
	}
	else if (axis != 3)
	  {
	    TCLAjaxArgList (tex_coords);
	    HandleError(0, 0, "%s \"%s\" wrong # of arguments for Torus texture coordinates", 
			gcppcSLIDEReservedWords[sliderwtTorus],
			id);
      
	    return TCL_ERROR;
	  }

	tex_vects[tex_vectnum] = new DynVector;
	tex_vects[tex_vectnum]->SetAll (tempdfloats[0], tempdfloats[1],
				       tempdfloats[2]);	

	for (;axis >0; axis--)
	  delete tempdfloats[axis-1];
	
	tex_vectnum ++;
	subptr = subptr->next;
	// if too many things, just jump out of loop and throw err
	if (tex_vectnum > 3 && subptr != NULL)
	  {
	    tex_vectnum = -1;
	    subptr = NULL;
	  }
      }

    // if only 2 given, generate the other two.
    if (tex_vectnum == 2)
      {
	tex_vects[2] = tex_vects[1];
	tex_vects[1] = new DynVector;
	tex_vects[1]->Set ((*tex_vects[2])(0), (*tex_vects[0])(1), 0.0);
	tex_vects[1]->SetDynExprs (tex_vects[2]->GetDynExpr(0),
				   tex_vects[0]->GetDynExpr(1), NULL);
	
	tex_vects[3] = new DynVector;
	tex_vects[3]->Set ((*tex_vects[0])(0), (*tex_vects[2])(1), 0.0);
	tex_vects[3]->SetDynExprs (tex_vects[0]->GetDynExpr(0),
				   tex_vects[2]->GetDynExpr(1), NULL);
      }

    else if (tex_vectnum != 4)
      {
	TCLAjaxArgList (tex_coords);
	HandleError(0, 0, "%s \"%s\" wrong # of arguments for Torus texture points", 
		    gcppcSLIDEReservedWords[sliderwtTorus],
		    id);
      
	return TCL_ERROR;
      }

    SetTorusTexture(tex_vects[0], tex_vects[1], tex_vects[2], tex_vects[3]);
  }


  if (ribbegin != NULL)
    SetPrimitiveRIBBegin (ribbegin);

  if (ribend != NULL)
    SetPrimitiveRIBBegin (ribend);


  EndTorus();

  // Ajax
  //
  TCLAjaxArgList (tex_coords);  

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleAddCyclide(VOID* &pCreatedObject, int argc, char **argv)
{
  char *id = NULL;
  DynInt *drawellipse = NULL;
  DynInt *drawhyperbola = NULL;
  DynFloat *a = NULL;
  DynFloat *b = NULL;
  DynFloat *stringlen = NULL;
  DynFloat *thetamin = NULL;
  DynFloat *thetamax = NULL;
  DynFloat *phimin = NULL;
  DynFloat *phimax = NULL; 
  DynInt *phislices = NULL;
  DynInt *thetaslices = NULL;
  char *surfid = NULL;
  char *shading = NULL;
  char *lod = NULL;
  char *solid = NULL;
  char *ribbegin = NULL;
  char *ribend = NULL;
  struct argList *tex_coords = NULL;


  struct argFormat tclargform[] =
    { { NULL, "id",
	"Name for the Cyclide",
	ARG_SIMPLE | ARG_STRING, &id, TRUE },
      { "drawellipse", "int/intexpr",
	"Draw Ellipse",
	ARG_DYNINT, &drawellipse, FALSE },
      { "drawhyperbola", "int/intexpr",
	"Draw Hyperbola",
	ARG_DYNINT, &drawhyperbola, FALSE },
      { "a", "float/floatexpr",
	"a",
	ARG_DYNFLOAT, &a, FALSE },
      { "b", "float/floatexpr",
	"b",
	ARG_DYNFLOAT, &b, FALSE },
      { "stringlength", "float/floatexpr",
	"string length",
	ARG_DYNFLOAT, &stringlen, FALSE },
      { "thetamin", "float/floatexpr",
	"Theta-min",
	ARG_DYNFLOAT, &thetamin, FALSE },
      { "thetamax", "float/floatexpr",
	"Theta-max",
	ARG_DYNFLOAT, &thetamax, FALSE },
      { "phimin", "float/floatexpr",
	"phi-min",
	ARG_DYNFLOAT, &phimin, FALSE },
      { "phimax", "float/floatexpr",
	"phi-max",
	ARG_DYNFLOAT, &phimax, FALSE },
      { "thetaslices", "int/intexpr",
	"Theta-slices",
	ARG_DYNINT, &thetaslices, FALSE },
      { "phislices", "int/intexpr",
	"Phi-slices",
	ARG_DYNINT, &phislices, FALSE },
      { "texture", "{ texture_triple } [{ texture_triple }] { texture_triple } [{ texture_triple }]",
	"Texture coordinates",
	ARG_STRING | ARG_LIST, &tex_coords, FALSE},
      { "surface", "surfid",
	"Surface ID",
	ARG_STRING, &surfid, FALSE },
      { "shading", "shading_flag",
	"Shading type",
	ARG_STRING, &shading, FALSE },
      { "lod", "lod_flag",
	"Level of Detail flag",
	ARG_STRING, &lod, FALSE },
      { "solid", "solid_flag",
	"Object Solidity flag",
	ARG_STRING, &solid, FALSE },

      { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
      { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }
  };

  const int argCount = 19;

  
  //printf ("HandleAddCyclide: %d\n",argc);

  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create cyclide: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  id = m_pCScanner->FindIdentifier (id, TRUE);
  AddCyclide( "[TCL Generated Cyclide]", 0, 0, id);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  if (drawellipse != NULL)
    SetCyclideDrawEllipse (drawellipse);

  if (drawhyperbola != NULL)
    SetCyclideDrawHyperbola (drawhyperbola);
  
  if (a != NULL)
    SetCyclideXMajorA (a);

  if (b != NULL)
    SetCyclideYMinorB (b);

  if (stringlen != NULL)
    SetCyclideStringLength (stringlen);

  if (thetamax != NULL)
    SetCyclideThetaMax(thetamax);

  if (thetamin != NULL)
    SetCyclideThetaMax(thetamax);

  if (phimin != NULL)
    SetCyclidePhiMin(phimin);

  if (phimax != NULL)
    SetCyclidePhiMax(phimax);

  if (thetaslices != NULL)
    SetCyclideThetaSlices(thetaslices);

  if (phislices != NULL)
    SetCyclidePhiSlices(phislices);

  if (surfid != NULL)
    {
      surfid = m_pCScanner->FindIdentifier (surfid, TRUE);
      SetNodeSurfaceID(surfid);
    }


  if (shading != NULL)
  {
    DynInt *pdyniShading = TranslateShadingFlag(shading);
    SetNodeShading (pdyniShading);
  }
  
  if (lod != NULL)
  {
    DynInt *pdyniLOD = TranslateLODFlag(lod);
    SetNodeLOD (pdyniLOD);
  }  

  if (solid != NULL)
  {
	DynInt *pdyniSolid = TranslateSolidFlag(solid);
	SetAtomNodeSolid (pdyniSolid);
  }

  if (tex_coords != NULL)
  {
    struct argList *subptr = tex_coords;
    struct argList *tc_values = NULL;
    DynVector *tex_vects[4];
    int tex_vectnum=0;

    while (subptr != NULL)
      {
	struct argList *tc_temp;
	int axis=0;
	DynFloat *tempdfloats[3];

	tc_values = TCLParseArgs_List ((CHAR *)(subptr->value), ARG_FLOAT);
	
	for (tc_temp=tc_values; tc_temp != NULL; tc_temp=tc_temp->next)
	  {
	    tempdfloats[axis] = new DynFloat;
	    
	    if (tc_temp->argType == ARG_EXPR)	      
	      {	
		tempdfloats[axis]->SetDynExpr((CHAR *)(tc_temp->value));
	      }	
	    else	
	      {
		tempdfloats[axis]->SetValue(*(FLOAT*)(tc_temp->value));
	      }
      
	    axis++;
	    // if too many things, just jump out of loop and throw err
	    if (axis > 2 && tc_temp->next != NULL)
	      {
	        axis = -1;
		break;
	      }

	  }

	if (axis == 2)
	{
	   tempdfloats[2] = new DynFloat;
	   tempdfloats[2]->SetValue((FLOAT)0.0);
	}
	else if (axis != 3)
	  {
	    TCLAjaxArgList (tex_coords);
	    HandleError(0, 0, "%s \"%s\" wrong # of arguments for Cyclide texture coordinates", 
			gcppcSLIDEReservedWords[sliderwtCyclide],
			id);
      
	    return TCL_ERROR;
	  }

	tex_vects[tex_vectnum] = new DynVector;
	tex_vects[tex_vectnum]->SetAll (tempdfloats[0], tempdfloats[1],
				       tempdfloats[2]);	

	for (;axis >0; axis--)
	  delete tempdfloats[axis-1];
	
	tex_vectnum ++;
	subptr = subptr->next;
	// if too many things, just jump out of loop and throw err
	if (tex_vectnum > 3 && subptr != NULL)
	  {
	    tex_vectnum = -1;
	    subptr = NULL;
	  }
      }

    // if only 2 given, generate the other two.
    if (tex_vectnum == 2)
      {
	tex_vects[2] = tex_vects[1];
	tex_vects[1] = new DynVector;
	tex_vects[1]->Set ((*tex_vects[2])(0), (*tex_vects[0])(1), 0.0);
	tex_vects[1]->SetDynExprs (tex_vects[2]->GetDynExpr(0),
				   tex_vects[0]->GetDynExpr(1), NULL);
	
	tex_vects[3] = new DynVector;
	tex_vects[3]->Set ((*tex_vects[0])(0), (*tex_vects[2])(1), 0.0);
	tex_vects[3]->SetDynExprs (tex_vects[0]->GetDynExpr(0),
				   tex_vects[2]->GetDynExpr(1), NULL);
      }

    else if (tex_vectnum != 4)
      {
	TCLAjaxArgList (tex_coords);
	HandleError(0, 0, "%s \"%s\" wrong # of arguments for Cyclide texture points", 
		    gcppcSLIDEReservedWords[sliderwtCyclide],
		    id);
      
	return TCL_ERROR;
      }

    SetCyclideTexture(tex_vects[0], tex_vects[1], tex_vects[2], tex_vects[3]);
  }


  if (ribbegin != NULL)
    SetPrimitiveRIBBegin (ribbegin);

  if (ribend != NULL)
    SetPrimitiveRIBBegin (ribend);


  EndCyclide();

  // Ajax
  //
  TCLAjaxArgList (tex_coords);  

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleAddPolyline(VOID* &pCreatedObject, int argc, char **argv)
{
  char *id = NULL;
  DynInt *drawpoints = NULL;
  struct argList *pointlist = NULL;

  char *surfid = NULL;
  char *shading = NULL;
  char *lod = NULL;
  char *ribbegin = NULL;
  char *ribend = NULL;

  struct argFormat tclargform[] =
    { { NULL, "id",
	"Name for the Polyline",
	ARG_SIMPLE | ARG_STRING, &id, TRUE },
      { "drawpointss", "int/intexpr",
	"Draw Points",
	ARG_DYNINT, &drawpoints, FALSE },
      { "pointlist", "float/floatexpr list",
	"Point list",
	ARG_STRING | ARG_LIST, &pointlist, FALSE },
      { "surface", "surfid",
	"Surface ID",
	ARG_STRING, &surfid, FALSE },
      { "shading", "shading_flag",
	"Shading type",
	ARG_STRING, &shading, FALSE },
      { "lod", "lod_flag",
	"Level of Detail flag",
	ARG_STRING, &lod, FALSE },

      { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
      { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }
  };

  const int argCount = 8;

  //printf ("HandleAddPolyline: %d\n",argc);

  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create polyline: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  id = m_pCScanner->FindIdentifier (id, TRUE);
  AddPolyline( "[TCL Generated Polyline]", 0, 0, id);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  if (drawpoints != NULL)
    SetPolylineDrawPoints (drawpoints);

  if (pointlist != NULL)
    {
      struct argList *tempptr;
      
      CListInt* pPList = CreateList();
      for (tempptr = pointlist; tempptr != NULL; tempptr = tempptr->next)
	{
	  tempptr->value = (VOID*)(m_pCScanner->FindIdentifier( (CHAR*)(tempptr->value), 
								 TRUE) );
	  InsertList (pPList, (CHAR*)(tempptr->value));
	}
      
      SetPolylinePointList(pPList);
  }

  if (surfid != NULL)
    {
      surfid = m_pCScanner->FindIdentifier (surfid, TRUE);
      SetNodeSurfaceID(surfid);
    }
  
  if (shading != NULL)
    {
      DynInt *pdyniShading = TranslateShadingFlag(shading);
      SetNodeShading (pdyniShading);
    }
  
  if (lod != NULL)
    {
      DynInt *pdyniLOD = TranslateLODFlag(lod);
      SetNodeLOD (pdyniLOD);
    }
  
  if (ribbegin != NULL)
    SetPrimitiveRIBBegin (ribbegin);

  if (ribend != NULL)
    SetPrimitiveRIBBegin (ribend);


  EndPolyline();

  // Ajax
  //
  TCLAjaxArgList (pointlist);

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleAddBSplineCurve(VOID* &pCreatedObject, int argc, char **argv)
{
  char *id = NULL;
  DynInt *drawcontrols = NULL;
  DynInt *order = NULL;
  struct argList *knotlist = NULL;
  struct argList *controlpointlist = NULL;
  DynInt *slices = NULL;

  char *surfid = NULL;
  char *shading = NULL;
  char *lod = NULL;
  char *ribbegin = NULL;
  char *ribend = NULL;

  struct argFormat tclargform[] =
    { { NULL, "id",
	"Name for the BSplineCurve",
	ARG_SIMPLE | ARG_STRING, &id, TRUE },
      { "drawcontrols", "int/intexpr",
	"Draw Controls",
	ARG_DYNINT, &drawcontrols, FALSE },
      { "order", "int/intexpr",
	"order",
	ARG_DYNINT, &order, FALSE },
      { "knotlist", "float/floatexpr list",
	"knot list",
	ARG_FLOAT | ARG_LIST, &knotlist, FALSE },
      { "controlpointlist", "float/floatexpr list",
	"Control Point list",
	ARG_STRING | ARG_LIST, &controlpointlist, FALSE },
      { "slices", "int/intexpr",
	"Slices",
	ARG_DYNINT, &slices, FALSE },
      { "surface", "surfid",
	"Surface ID",
	ARG_STRING, &surfid, FALSE },
      { "shading", "shading_flag",
	"Shading type",
	ARG_STRING, &shading, FALSE },
      { "lod", "lod_flag",
	"Level of Detail flag",
	ARG_STRING, &lod, FALSE },
      { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
      { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }
  };

  const int argCount = 11;

  
  //printf ("HandleAddBSplineCurve: %d\n",argc);

  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create bsplinecurve: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  id = m_pCScanner->FindIdentifier (id, TRUE);
  AddBSplineCurve( "[TCL Generated BSplineCurve]", 0, 0, id);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  if (drawcontrols != NULL)
    SetBSplineCurveDrawControls (drawcontrols);

  if (order != NULL)
    SetBSplineCurveOrder (order);

  if (slices != NULL)
    SetBSplineCurveSlices (slices);

  if (knotlist != NULL)
    {
      CListInt *pKnots = CreateDynvalueList();
      struct argList *temp = knotlist;
      while (temp != NULL)
	{
	  DynFloat *tempfloat = new DynFloat;
	  if (temp->argType == ARG_EXPR)
	    tempfloat->SetDynExpr((CHAR*)(temp->value));
	  else
	    tempfloat->SetValue (*(FLOAT*)(temp->value));
	  InsertDynvalueList(pKnots, tempfloat);
	  
	  temp = temp->next;
	} 
      SetBSplineCurveKnotList (pKnots);
    }

  if (controlpointlist != NULL)
    {
      struct argList *tempptr;
      
      CListInt* pCPList = CreateList();
      for (tempptr = controlpointlist; tempptr != NULL; tempptr = tempptr->next)
	{
	  tempptr->value = (VOID*)(m_pCScanner->FindIdentifier( (CHAR*)(tempptr->value), 
								TRUE) );
	  InsertList (pCPList, (CHAR*)(tempptr->value));
	}
      SetBSplineCurveControlPointList(pCPList);
    }

  if (surfid != NULL)
    {
      surfid = m_pCScanner->FindIdentifier (surfid, TRUE);
      SetNodeSurfaceID(surfid);
    }
  
  if (shading != NULL)
    {
      DynInt *pdyniShading = TranslateShadingFlag(shading);
      SetNodeShading (pdyniShading);
    }
  
  if (lod != NULL)
    {
      DynInt *pdyniLOD = TranslateLODFlag(lod);
      SetNodeLOD (pdyniLOD);
    }
  
  if (ribbegin != NULL)
    SetPrimitiveRIBBegin (ribbegin);

  if (ribend != NULL)
    SetPrimitiveRIBBegin (ribend);


  EndBSplineCurve();

  // Ajax
  //
  TCLAjaxArgList (knotlist);
  TCLAjaxArgList (controlpointlist);

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleAddBezierCurve(VOID* &pCreatedObject, int argc, char **argv)
{
  char *id = NULL;
  DynInt *drawcontrols = NULL;
  struct argList *controlpointlist = NULL;
  DynInt *slices = NULL;

  char *surfid = NULL;
  char *shading = NULL;
  char *lod = NULL;
  char *ribbegin = NULL;
  char *ribend = NULL;

  struct argFormat tclargform[] =
    { { NULL, "id",
	"Name for the BezierCurve",
	ARG_SIMPLE | ARG_STRING, &id, TRUE },
      { "drawcontrols", "int/intexpr",
	"Draw Controls",
	ARG_DYNINT, &drawcontrols, FALSE },
      { "controlpointlist", "float/floatexpr list",
	"Control Point list",
	ARG_STRING | ARG_LIST, &controlpointlist, FALSE },
      { "slices", "int/intexpr",
	"Slices",
	ARG_DYNINT, &slices, FALSE },
      { "surface", "surfid",
	"Surface ID",
	ARG_STRING, &surfid, FALSE },
      { "shading", "shading_flag",
	"Shading type",
	ARG_STRING, &shading, FALSE },
      { "lod", "lod_flag",
	"Level of Detail flag",
	ARG_STRING, &lod, FALSE },
      { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
      { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }
  };

  const int argCount = 9;

  
  //printf ("HandleAddBezierCurve: %d\n",argc);

  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create beziercurve: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  id = m_pCScanner->FindIdentifier (id, TRUE);
  AddBSplineCurve( "[TCL Generated BezierCurve]", 0, 0, id);
  SetBSplineCurveType(1);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  if (drawcontrols != NULL)
    SetBSplineCurveDrawControls (drawcontrols);

  if (slices != NULL)
    SetBSplineCurveSlices (slices);

  if (controlpointlist != NULL)
    {
      struct argList *tempptr;
      
      CListInt* pCPList = CreateList();
      for (tempptr = controlpointlist; tempptr != NULL; tempptr = tempptr->next)
	{
	  tempptr->value = (VOID*)(m_pCScanner->FindIdentifier( (CHAR*)(tempptr->value), 
								TRUE) );
	  InsertList (pCPList, (CHAR*)(tempptr->value));
	}
      SetBSplineCurveControlPointList(pCPList);
    }

  if (surfid != NULL)
    {
      surfid = m_pCScanner->FindIdentifier (surfid, TRUE);
      SetNodeSurfaceID(surfid);
    }
  
  if (shading != NULL)
    {
      DynInt *pdyniShading = TranslateShadingFlag(shading);
      SetNodeShading (pdyniShading);
    }
  
  if (lod != NULL)
    {
      DynInt *pdyniLOD = TranslateLODFlag(lod);
      SetNodeLOD (pdyniLOD);
    }
  
  if (ribbegin != NULL)
    SetPrimitiveRIBBegin (ribbegin);

  if (ribend != NULL)
    SetPrimitiveRIBBegin (ribend);


  EndBSplineCurve();

  // Ajax
  //
  TCLAjaxArgList (controlpointlist);

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleAddSweep(VOID* &pCreatedObject, int argc, char **argv)
{
  char *id = NULL;
  DynInt *drawsweep = NULL;
  DynInt *drawpath = NULL;

  char *pathId = NULL;
  DynInt *minimizeTorsion = NULL;
  DynInt *closed = NULL;
  DynFloat *azimuth = NULL;
  DynFloat *twist = NULL;

  char *crosssectionId = NULL;
  DynInt *symmetry = NULL;
  DynInt *endcap = NULL;
  DynInt *begincap = NULL;

  struct argList *tex_coords = NULL;
  char *surfid = NULL;
  char *shading = NULL;
  char *solid = NULL;
  char *lod = NULL;
  char *ribbegin = NULL;
  char *ribend = NULL;

  struct argFormat tclargform[] =
    { { NULL, "id",
	"Name for the Sweep",
	ARG_SIMPLE | ARG_STRING, &id, TRUE },
      { "drawsweep", "int/intexpr",
	"Draw Sweep",
	ARG_DYNINT, &drawsweep, FALSE },
      { "drawpath", "int/intexpr",
	"Draw Path",
	ARG_DYNINT, &drawpath, FALSE },

      { "path", "path_name",
	"Name for the Sweep Path",
	ARG_STRING, &pathId, FALSE },
      { "minimizetorsion", "int/intexpr",
	"Minimizetorsion",
	ARG_DYNINT, &minimizeTorsion, FALSE },
      { "closed", "int/intexpr",
	"Closed",
	ARG_DYNFLOAT, &closed, FALSE },
      { "azimuth", "float/floatexpr",
	"Azimuth",
	ARG_DYNINT, &azimuth, FALSE },
      { "twist", "float/floatexpr",
	"Twist",
	ARG_DYNFLOAT, &twist, FALSE },

      { "crosssection", "crossection_name",
	"Name for the Sweep CrossSection",
	ARG_STRING, &crosssectionId, FALSE },

      { "surface", "surfid",
	"Surface ID",
	ARG_STRING, &surfid, FALSE },
      { "shading", "shading_flag",
	"Shading type",
	ARG_STRING, &shading, FALSE },
      { "solid", "solid_flag",
	"Object Solidity flag",
	ARG_STRING, &solid, FALSE },
      { "lod", "lod_flag",
	"Level of Detail flag",
	ARG_STRING, &lod, FALSE },
      { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
      { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }
  };

  const int argCount = 16;

  
  //printf ("HandleAddSweep: %d\n",argc);

  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create sweep: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  id = m_pCScanner->FindIdentifier (id, TRUE);
  AddSweep( "[TCL Generated Sweep]", 0, 0, id);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  if (drawsweep != NULL)
    SetSweepDrawSweep (drawsweep);

  if (drawpath != NULL)
    SetSweepDrawPath (drawpath);

  pathId = m_pCScanner->FindIdentifier (pathId, TRUE);
  if (pathId != NULL)
    SetSweepPathID(pathId);
  else
    {
      printf ("sweep syntax error: No path was specified\n");
      return TCL_ERROR;
    }

  if (minimizeTorsion != NULL)
    SetSweepPathMinimizeTorsion(minimizeTorsion);

  if (closed != NULL)
    SetSweepPathClosed(closed);

  if (azimuth != NULL)
    SetSweepPathAzimuth(azimuth);

  if (twist != NULL)
    SetSweepPathTwist(twist);



  crosssectionId = m_pCScanner->FindIdentifier(crosssectionId, TRUE);
  if (pathId != NULL)
    AddSweepCrossSectionID(crosssectionId);
  else
    {
      printf ("sweep syntax error: No crosssection was specified\n");
      return TCL_ERROR;
    }

  if (symmetry != NULL)
    SetSweepPathSymmetry(symmetry);

  if (endcap != NULL)
    SetSweepCrossSectionEndCap(endcap);

  if (begincap != NULL)
    SetSweepCrossSectionBeginCap(begincap);


  if (surfid != NULL)
    {
      surfid = m_pCScanner->FindIdentifier (surfid, TRUE);
      SetNodeSurfaceID(surfid);
    }
  
  if (shading != NULL)
    {
      DynInt *pdyniShading = TranslateShadingFlag(shading);
      SetNodeShading (pdyniShading);
    }
  
  if (solid != NULL)
    {
      DynInt *pdyniSolid = TranslateSolidFlag(solid);
      SetAtomNodeSolid (pdyniSolid);
    }
  
  if (lod != NULL)
    {
      DynInt *pdyniLOD = TranslateLODFlag(lod);
      SetNodeLOD (pdyniLOD);
    }

  /*  
  if (tex_coords != NULL)
    {
      struct argList *subptr = tex_coords;
      struct argList *tc_values = NULL;
      DynVector *tex_vects[4];
      int tex_vectnum=0;
      
      while (subptr != NULL)
	{
	  struct argList *tc_temp;
	  int axis=0;
	  DynFloat *tempdfloats[3];
	  
	  tc_values = TCLParseArgs_List ((CHAR *)(subptr->value), ARG_FLOAT);
	  
	  for (tc_temp=tc_values; tc_temp != NULL; tc_temp=tc_temp->next)
	    {
	      tempdfloats[axis] = new DynFloat;
	      
	      if (tc_temp->argType == ARG_EXPR)	      
		{	
		  tempdfloats[axis]->SetDynExpr((CHAR *)(tc_temp->value));
		}	
	      else	
		{
		  tempdfloats[axis]->SetValue(*(FLOAT*)(tc_temp->value));
		}
	      
	      axis++;
	      // if too many things, just jump out of loop and throw err
	      if (axis > 2 && tc_temp->next != NULL)
		{
		  axis = -1;
		  break;
		}
	    }

	if (axis == 2)
	  {
	    tempdfloats[2] = new DynFloat;
	    tempdfloats[2]->SetValue((FLOAT)0.0);
	  }
	else if (axis != 3)
	  {
	    TCLAjaxArgList (tex_coords);
	    HandleError(0, 0, "%s \"%s\" wrong # of arguments for Sweep texture coordinates", 
			gcppcSLIDEReservedWords[sliderwtSweep],
			id);
	    
	    return TCL_ERROR;
	  }
	
	tex_vects[tex_vectnum] = new DynVector;
	tex_vects[tex_vectnum]->SetAll (tempdfloats[0], tempdfloats[1],	tempdfloats[2]);	
	for (;axis >0; axis--)
	  delete tempdfloats[axis-1];
	
	tex_vectnum ++;
	subptr = subptr->next;
	// if too many things, just jump out of loop and throw err
	if (tex_vectnum > 3 && subptr != NULL)
	  {
	    tex_vectnum = -1;
	    subptr = NULL;
	  }
	}
      
      // if only 2 given, generate the other two.
      if (tex_vectnum == 2)
	{
	  tex_vects[2] = tex_vects[1];
	  tex_vects[1] = new DynVector;
	  tex_vects[1]->Set ((*tex_vects[2])(0), (*tex_vects[0])(1), 0.0);
	  tex_vects[1]->SetDynExprs (tex_vects[2]->GetDynExpr(0),
				     tex_vects[0]->GetDynExpr(1), NULL);
	  
	  tex_vects[3] = new DynVector;
	  tex_vects[3]->Set ((*tex_vects[0])(0), (*tex_vects[2])(1), 0.0);
	  tex_vects[3]->SetDynExprs (tex_vects[0]->GetDynExpr(0),
				     tex_vects[2]->GetDynExpr(1), NULL);
	}
      else if (tex_vectnum != 4)
	{
	  TCLAjaxArgList (tex_coords);
	 HandleError(0, 0, "%s \"%s\" wrong # of arguments for Sweep texture points",
		     gcppcSLIDEReservedWords[sliderwtSweep],
		     id);
      
	 return TCL_ERROR;
	}
      
      SetSweepTexture(tex_vects[0], tex_vects[1], tex_vects[2], tex_vects[3]);
    }
    */
  
  if (ribbegin != NULL)
    SetPrimitiveRIBBegin (ribbegin);

  if (ribend != NULL)
    SetPrimitiveRIBBegin (ribend);


  EndSweep();

  // Ajax
  //
  TCLAjaxArgList (tex_coords);

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleAddBSplinePatch(VOID* &pCreatedObject, int argc, char **argv)
{
  char *id = NULL;
  DynInt *drawcontrols = NULL;
  DynInt *uorder = NULL;
  DynInt *uknots = NULL;
  struct argList *uknotlist = NULL;
  DynInt *vorder = NULL;
  DynInt *vknots = NULL;
  struct argList *vknotlist = NULL;
  struct argList *controlpointlist = NULL;
  DynInt *uslices = NULL;
  DynInt *vslices = NULL;
  struct argList *tex_coords = NULL;

  char *surfid = NULL;
  char *shading = NULL;
  char *solid = NULL;
  char *lod = NULL;
  char *ribbegin = NULL;
  char *ribend = NULL;

  struct argFormat tclargform[] =
    { { NULL, "id",
	"Name for the BSplinePatch",
	ARG_SIMPLE | ARG_STRING, &id, TRUE },
      { "drawcontrols", "int/intexpr",
	"Draw Controls",
	ARG_DYNINT, &drawcontrols, FALSE },
      { "uorder", "float/floatexpr",
	"U order",
	ARG_DYNINT, &uorder, FALSE },
      { "uknots", "float/floatexpr",
	"U knots",
	ARG_DYNINT, &uknots, FALSE },
      { "uknotlist", "float/floatexpr list",
	"U knot list",
	ARG_FLOAT | ARG_LIST, &uknotlist, FALSE },
      { "vorder", "float/floatexpr",
	"V order",
	ARG_DYNINT, &vorder, FALSE },
      { "vknots", "float/floatexpr",
	"V knots",
	ARG_DYNINT, &vknots, FALSE },
      { "vknotlist", "float/floatexpr list",
	"N knot list",
	ARG_FLOAT | ARG_LIST, &vknotlist, FALSE },
      { "controlpointlist", "float/floatexpr list",
	"Control Point list",
	ARG_STRING | ARG_LIST, &controlpointlist, FALSE },
      { "uslices", "int/intexpr",
	"U Slices",
	ARG_DYNINT, &uslices, FALSE },
      { "vslices", "int/intexpr",
	"V Slices",
	ARG_DYNINT, &vslices, FALSE },
      { "texture", "{ texture_triple } [{ texture_triple }] { texture_triple } [{ texture_triple }]",
	"Texture coordinates",
	ARG_STRING | ARG_LIST, &tex_coords, FALSE},
      { "surface", "surfid",
	"Surface ID",
	ARG_STRING, &surfid, FALSE },
      { "shading", "shading_flag",
	"Shading type",
	ARG_STRING, &shading, FALSE },
      { "solid", "solid_flag",
	"Object Solidity flag",
	ARG_STRING, &solid, FALSE },
      { "lod", "lod_flag",
	"Level of Detail flag",
	ARG_STRING, &lod, FALSE },

      { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
      { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }
  };

  const int argCount = 18;

  
  //printf ("HandleAddBSplinePatch: %d\n",argc);

  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create bsplinepatch: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  id = m_pCScanner->FindIdentifier (id, TRUE);
  AddBSplinePatch( "[TCL Generated BSplinePatch]", 0, 0, id);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  if (drawcontrols != NULL)
    SetBSplinePatchDrawControls (drawcontrols);

  if (uorder != NULL)
    SetBSplinePatchUOrder (uorder);

  if (uknots != NULL)
    SetBSplinePatchUKnots (uknots);

  if (vorder != NULL)
    SetBSplinePatchVOrder (vorder);

  if (vknots != NULL)
    SetBSplinePatchVKnots (vknots);

  if (vslices != NULL)
    SetBSplinePatchVSlices (vslices);

  if (uslices != NULL)
    SetBSplinePatchUSlices (uslices);

  if (uknotlist != NULL)
  {
    CListInt *pUKnots = CreateDynvalueList();
    struct argList *temp = uknotlist;
    while (temp != NULL)
      {
	DynFloat *tempfloat = new DynFloat;
	if (temp->argType == ARG_EXPR)
	  tempfloat->SetDynExpr((CHAR*)(temp->value));
	else
	  tempfloat->SetValue (*(FLOAT*)(temp->value));
	InsertDynvalueList(pUKnots, tempfloat);
	
	temp = temp->next;
      }
    
    SetBSplinePatchUKnotList (pUKnots);
  }

  if (vknotlist != NULL)
  {
	CListInt *pVKnots = CreateDynvalueList();
	struct argList *temp = vknotlist;
	while (temp != NULL)
	{
	  DynFloat *tempfloat = new DynFloat;
	  if (temp->argType == ARG_EXPR)
		tempfloat->SetDynExpr((CHAR*)(temp->value));
	  else
		tempfloat->SetValue (*(FLOAT*)(temp->value));
	  InsertDynvalueList(pVKnots, tempfloat);

	  temp = temp->next;
	}

	SetBSplinePatchVKnotList (pVKnots);
  }

  if (controlpointlist != NULL)
    {
      struct argList *tempptr;
      
      CListInt* pCPList = CreateList();
      for (tempptr = controlpointlist; tempptr != NULL; tempptr = tempptr->next)
	{
	  tempptr->value = (VOID*)(m_pCScanner->FindIdentifier( (CHAR*)(tempptr->value), 
								 TRUE) );
	  InsertList (pCPList, (CHAR*)(tempptr->value));
	}
      
      SetBSplinePatchControlPointList(pCPList);
  }

  if (surfid != NULL)
    {
      surfid = m_pCScanner->FindIdentifier (surfid, TRUE);
      SetNodeSurfaceID(surfid);
    }
  
  if (shading != NULL)
    {
      DynInt *pdyniShading = TranslateShadingFlag(shading);
      SetNodeShading (pdyniShading);
    }
  
  if (solid != NULL)
    {
      DynInt *pdyniSolid = TranslateSolidFlag(solid);
      SetAtomNodeSolid (pdyniSolid);
    }
  
  if (lod != NULL)
    {
      DynInt *pdyniLOD = TranslateLODFlag(lod);
      SetNodeLOD (pdyniLOD);
    }
  
  if (tex_coords != NULL)
    {
      struct argList *subptr = tex_coords;
      struct argList *tc_values = NULL;
      DynVector *tex_vects[4];
      int tex_vectnum=0;
      
      while (subptr != NULL)
	{
	  struct argList *tc_temp;
	  int axis=0;
	  DynFloat *tempdfloats[3];
	  
	  tc_values = TCLParseArgs_List ((CHAR *)(subptr->value), ARG_FLOAT);
	  
	  for (tc_temp=tc_values; tc_temp != NULL; tc_temp=tc_temp->next)
	    {
	      tempdfloats[axis] = new DynFloat;
	      
	      if (tc_temp->argType == ARG_EXPR)	      
		{	
		  tempdfloats[axis]->SetDynExpr((CHAR *)(tc_temp->value));
		}	
	      else	
		{
		  tempdfloats[axis]->SetValue(*(FLOAT*)(tc_temp->value));
		}
	      
	      axis++;
	      // if too many things, just jump out of loop and throw err
	      if (axis > 2 && tc_temp->next != NULL)
		{
		  axis = -1;
		  break;
		}
	    }

	if (axis == 2)
	  {
	    tempdfloats[2] = new DynFloat;
	    tempdfloats[2]->SetValue((FLOAT)0.0);
	  }
	else if (axis != 3)
	  {
	    TCLAjaxArgList (tex_coords);
	    HandleError(0, 0, "%s \"%s\" wrong # of arguments for BSplinePatch texture coordinates", 
			gcppcSLIDEReservedWords[sliderwtBSplinePatch],
			id);
	    
	    return TCL_ERROR;
	  }
	
	tex_vects[tex_vectnum] = new DynVector;
	tex_vects[tex_vectnum]->SetAll (tempdfloats[0], tempdfloats[1],	tempdfloats[2]);	
	for (;axis >0; axis--)
	  delete tempdfloats[axis-1];
	
	tex_vectnum ++;
	subptr = subptr->next;
	// if too many things, just jump out of loop and throw err
	if (tex_vectnum > 3 && subptr != NULL)
	  {
	    tex_vectnum = -1;
	    subptr = NULL;
	  }
	}
      
      // if only 2 given, generate the other two.
      if (tex_vectnum == 2)
	{
	  tex_vects[2] = tex_vects[1];
	  tex_vects[1] = new DynVector;
	  tex_vects[1]->Set ((*tex_vects[2])(0), (*tex_vects[0])(1), 0.0);
	  tex_vects[1]->SetDynExprs (tex_vects[2]->GetDynExpr(0),
				     tex_vects[0]->GetDynExpr(1), NULL);
	  
	  tex_vects[3] = new DynVector;
	  tex_vects[3]->Set ((*tex_vects[0])(0), (*tex_vects[2])(1), 0.0);
	  tex_vects[3]->SetDynExprs (tex_vects[0]->GetDynExpr(0),
				     tex_vects[2]->GetDynExpr(1), NULL);
	}
      else if (tex_vectnum != 4)
	{
	  TCLAjaxArgList (tex_coords);
	 HandleError(0, 0, "%s \"%s\" wrong # of arguments for BSplinePatch texture points",
		     gcppcSLIDEReservedWords[sliderwtBSplinePatch],
		     id);
      
	 return TCL_ERROR;
	}
      
      SetBSplinePatchTexture(tex_vects[0], tex_vects[1], tex_vects[2], tex_vects[3]);
    }
  
  if (ribbegin != NULL)
    SetPrimitiveRIBBegin (ribbegin);

  if (ribend != NULL)
    SetPrimitiveRIBBegin (ribend);


  EndBSplinePatch();

  // Ajax
  //
  TCLAjaxArgList (tex_coords);
  TCLAjaxArgList (uknotlist);
  TCLAjaxArgList (vknotlist);
  TCLAjaxArgList (controlpointlist);

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

//unfinished, unsure if necessary
int CSLIDEParser::HandleAddHeightfield(VOID* &pCreatedObject, int argc, char **argv)
{
  char *id = NULL;
  DynInt *umin = NULL;
  DynInt *umax = NULL;
  DynInt *vmin = NULL;
  DynInt *vmax = NULL;
  DynInt *uslices = NULL;
  DynInt *vslices = NULL;
  DynVector *func = NULL;

  struct argFormat tclargform[] =
    { { NULL, "id",
	"Name for the Heightfield",
	ARG_SIMPLE | ARG_STRING, &id, TRUE },
      { "umin", "float/floatexpr",
	"U min",
	ARG_DYNFLOAT, &umin, FALSE },
      { "umax", "float/floatexpr",
	"U max",
	ARG_DYNFLOAT, &umax, FALSE },
      { "vmin", "float/floatexpr",
	"V min",
	ARG_DYNFLOAT, &vmin, FALSE },
      { "vmax", "float/floatexpr",
	"V max",
	ARG_DYNFLOAT, &vmax, FALSE },
      { "uslices", "float/floatexpr",
	"U slices",
	ARG_DYNINT, &uslices, FALSE },
      { "uslices", "float/floatexpr",
	"U slices",
	ARG_DYNINT, &uslices, FALSE }};

  return TCL_OK;
}

int CSLIDEParser::HandleAddBezierPatch(VOID* &pCreatedObject, int argc, char **argv)
{
  char *id = NULL;
  DynInt *drawcontrols = NULL;
  DynInt *uorder = NULL;
  DynInt *vorder = NULL;
  struct argList *controlpointlist = NULL;
  DynInt *uslices = NULL;
  DynInt *vslices = NULL;
  struct argList *tex_coords = NULL;

  char *surfid = NULL;
  char *shading = NULL;
  char *solid = NULL;
  char *lod = NULL;
  char *ribbegin = NULL;
  char *ribend = NULL;

  struct argFormat tclargform[] =
    { { NULL, "id",
	"Name for the BSplinePatch",
	ARG_SIMPLE | ARG_STRING, &id, TRUE },
      { "drawcontrols", "int/intexpr",
	"Draw Controls",
	ARG_DYNINT, &drawcontrols, FALSE },
      { "uorder", "float/floatexpr",
	"U order",
	ARG_DYNINT, &uorder, FALSE },
      { "vorder", "float/floatexpr",
	"V order",
	ARG_DYNINT, &vorder, FALSE },
      { "uslices", "int/intexpr",
	"U Slices",
	ARG_DYNINT, &uslices, FALSE },
      { "vslices", "int/intexpr",
	"V Slices",
	ARG_DYNINT, &vslices, FALSE },
      { "controlpointlist", "float/floatexpr list",
	"Control Point list",
	ARG_STRING | ARG_LIST, &controlpointlist, FALSE },
      { "texture", "{ texture_triple } [{ texture_triple }] { texture_triple } [{ texture_triple }]",
	"Texture coordinates",
	ARG_STRING | ARG_LIST, &tex_coords, FALSE},
      { "surface", "surfid",
	"Surface ID",
	ARG_STRING, &surfid, FALSE },
      { "shading", "shading_flag",
	"Shading type",
	ARG_STRING, &shading, FALSE },
      { "solid", "solid_flag",
	"Object Solidity flag",
	ARG_STRING, &solid, FALSE },
      { "lod", "lod_flag",
	"Level of Detail flag",
	ARG_STRING, &lod, FALSE },

      { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
      { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }
  };

  const int argCount = 14;

  
  //printf ("HandleAddBezierPatch: %d\n",argc);

  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create bezierpatch: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  id = m_pCScanner->FindIdentifier (id, TRUE);
  AddBezierPatch( "[TCL Generated BezierPatch]", 0, 0, id);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  if (drawcontrols != NULL)
    SetBezierPatchDrawControls (drawcontrols);

  if (uorder != NULL)
    SetBezierPatchUOrder (uorder);

  if (vorder != NULL)
    SetBezierPatchVOrder (vorder);

  if (vslices != NULL)
    SetBezierPatchVSlices (vslices);

  if (uslices != NULL)
    SetBezierPatchUSlices (uslices);

  if (controlpointlist != NULL)
    {
      struct argList *tempptr;
      
      CListInt* pCPList = CreateList();
      for (tempptr = controlpointlist; tempptr != NULL; tempptr = tempptr->next)
	{
	  tempptr->value = (VOID*)(m_pCScanner->FindIdentifier ((CHAR*)(tempptr->value), 
								TRUE) );
	  InsertList (pCPList, (CHAR*)(tempptr->value));
	}
      
      SetBezierPatchControlPointList(pCPList);
    }

  if (surfid != NULL)
    {
      surfid = m_pCScanner->FindIdentifier (surfid, TRUE);
      SetNodeSurfaceID(surfid);
    }
  
  if (shading != NULL)
    {
      DynInt *pdyniShading = TranslateShadingFlag(shading);
      SetNodeShading (pdyniShading);
    }
  
  if (solid != NULL)
    {
      DynInt *pdyniSolid = TranslateSolidFlag(solid);
      SetAtomNodeSolid (pdyniSolid);
    }
  
  if (lod != NULL)
    {
      DynInt *pdyniLOD = TranslateLODFlag(lod);
      SetNodeLOD (pdyniLOD);
    }
  
  if (tex_coords != NULL)
    {
      struct argList *subptr = tex_coords;
      struct argList *tc_values = NULL;
      DynVector *tex_vects[4];
      int tex_vectnum=0;
      
      while (subptr != NULL)
	{
	  struct argList *tc_temp;
	  int axis=0;
	  DynFloat *tempdfloats[3];
	  
	  tc_values = TCLParseArgs_List ((CHAR *)(subptr->value), ARG_FLOAT);
	  
	  for (tc_temp=tc_values; tc_temp != NULL; tc_temp=tc_temp->next)
	    {
	      tempdfloats[axis] = new DynFloat;
	      
	      if (tc_temp->argType == ARG_EXPR)	      
		{	
		  tempdfloats[axis]->SetDynExpr((CHAR *)(tc_temp->value));
		}	
	      else	
		{
		  tempdfloats[axis]->SetValue(*(FLOAT*)(tc_temp->value));
		}
	      
	      axis++;
	      // if too many things, just jump out of loop and throw err
	      if (axis > 2 && tc_temp->next != NULL)
		{
		  axis = -1;
		  break;
		}
	    }
	  
	  if (axis == 2)
	    {
	      tempdfloats[2] = new DynFloat;
	      tempdfloats[2]->SetValue((FLOAT)0.0);
	    }
	  else if (axis != 3)
	    {
	      TCLAjaxArgList (tex_coords);
	      HandleError(0, 0, "%s \"%s\" wrong # of arguments for BSplinePatch texture coordinates", 
			  gcppcSLIDEReservedWords[sliderwtBSplinePatch],
			  id);
	      
	      return TCL_ERROR;
	    }
	  
	  tex_vects[tex_vectnum] = new DynVector;
	  tex_vects[tex_vectnum]->SetAll (tempdfloats[0], tempdfloats[1],
				       tempdfloats[2]);	

	  for (;axis >0; axis--)
	    delete tempdfloats[axis-1];
	  
	  tex_vectnum ++;
	  subptr = subptr->next;
	  // if too many things, just jump out of loop and throw err
	  if (tex_vectnum > 3 && subptr != NULL)
	    {
	      tex_vectnum = -1;
	      subptr = NULL;
	    }
	}
      
      // if only 2 given, generate the other two.
      if (tex_vectnum == 2)
	{
	  tex_vects[2] = tex_vects[1];
	  tex_vects[1] = new DynVector;
	  tex_vects[1]->Set ((*tex_vects[2])(0), (*tex_vects[0])(1), 0.0);
	  tex_vects[1]->SetDynExprs (tex_vects[2]->GetDynExpr(0),
				     tex_vects[0]->GetDynExpr(1), NULL);
	  
	  tex_vects[3] = new DynVector;
	  tex_vects[3]->Set ((*tex_vects[0])(0), (*tex_vects[2])(1), 0.0);
	  tex_vects[3]->SetDynExprs (tex_vects[0]->GetDynExpr(0),
				     tex_vects[2]->GetDynExpr(1), NULL);
	}
      else if (tex_vectnum != 4)
	{
	  TCLAjaxArgList (tex_coords);
	  HandleError(0, 0, "%s \"%s\" wrong # of arguments for BSplinePatch texture points", 
		      gcppcSLIDEReservedWords[sliderwtBSplinePatch],
		      id);
	  
	  return TCL_ERROR;
	}

      SetBezierPatchTexture(tex_vects[0], tex_vects[1], tex_vects[2], tex_vects[3]);
    }
  
  
  if (ribbegin != NULL)
    SetPrimitiveRIBBegin (ribbegin);
  
  if (ribend != NULL)
    SetPrimitiveRIBBegin (ribend);
  
  
  EndBezierPatch();

  // Ajax
  //
  TCLAjaxArgList (tex_coords);
  TCLAjaxArgList (controlpointlist);

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleAddFog(VOID* &pCreatedObject, int argc, char **argv)
{
  char *id = NULL;
  char *fogtype = NULL;
  struct argList *color = NULL;
  DynFloat *start = NULL;
  DynFloat *end = NULL;
  DynFloat *density = NULL;

  char *ribbegin=NULL;
  char *ribend=NULL;
  
  DynFloat* pdynVectf[4];
  DynVector *pdynvect=NULL;
  int axis=0;

  struct argFormat tclargform[] =
    { { NULL, "fogID",
	"Name for new fog",
	ARG_SIMPLE | ARG_STRING, &id, TRUE },
      { "color", "R G B", 
	"R, G, and B values for new surface",
	ARG_LIST | ARG_FLOAT, &color, FALSE },
      { "start", "float/floatexp",
	"fog start distance",
	ARG_DYNFLOAT, &start, FALSE },
      { "end", "float/floatexp",
	"fog ending distance",
	ARG_DYNFLOAT, &end, FALSE },
      { "density", "float/floatexp",
	"fog density",
	ARG_DYNFLOAT, &density, FALSE },

      { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
      { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }
    };

  const int argCount = 7;
  
  //printf ("HandleAddFog: %d\n",argc);
  
  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create fog: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  id = m_pCScanner->FindIdentifier (id, TRUE);
  AddFog ("[TCL generated Fog]", 0, 0, id);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  struct argList *tempptr;

  if (color != NULL)
    {
      for (tempptr = color; tempptr != NULL; tempptr = tempptr->next)
	{
	  pdynVectf[axis] = new DynFloat;
	  
	  if (tempptr->argType == ARG_EXPR)
	      pdynVectf[axis]->SetDynExpr((CHAR *)(tempptr->value));
	  else
	      pdynVectf[axis]->SetValue(*(FLOAT*)(tempptr->value));
	  
	  axis++;
	  // if too many things, just jump out of loop and throw err
	  if (axis > 2 && tempptr->next != NULL)
	    {
	      axis = -1;
	      break;
	    }	  
	}
  
      if (axis != 3)
	{
	  TCLAjaxArgList (color);
	  HandleError(0, 0, "%s \"%s\" wrong # of arguments for fog color", 
		      gcppcSLIDEReservedWords[sliderwtFog],
		      id);
	  
	  return TCL_ERROR;
	}
      
      pdynvect = new DynVector;
      ASSERT( pdynvect != NULL );
      pdynvect->SetAll(pdynVectf[0], pdynVectf[1], pdynVectf[2]);
      
      SetFogColor(pdynvect);
    }

  if (fogtype != NULL)
  {
    DynInt *pdyniFogtype = TranslateFogFlag(fogtype);
    SetFogType (pdyniFogtype);
  }

  if (start != NULL)
    SetFogStart (start);

  if (end != NULL)
    SetFogEnd (end);

  if (density != NULL)
    SetFogDensity (density);
  
  
  //printf ("\n");

  if (ribbegin != NULL)
    SetPrimitiveRIBBegin (ribbegin);

  if (ribend != NULL)
    SetPrimitiveRIBBegin (ribend);

  
  EndFog();
  
  // Ajax
  //
  TCLAjaxArgList (color);

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleAddSurface(VOID* &pCreatedObject, int argc, char **argv)
{
  char *id = NULL;
  struct argList *color = NULL;
  struct argList *reflectivity = NULL;
  DynFloat *exponent = NULL;
  DynFloat *metallic = NULL;
  char *texture = NULL;
  char *texturetype = NULL;

  char *ribbegin=NULL;
  char *ribend=NULL;
  
  DynFloat* pdynVectf[4];
  DynVector *pdynvect=NULL;
  int axis=0;

  struct argFormat tclargform[] =
    { { NULL, "surfaceID",
	"Name for the new surface",
	ARG_SIMPLE | ARG_STRING, &id, TRUE },
      { "color", "R G B", 
	"R, G, and B values for new surface",
	ARG_LIST | ARG_FLOAT, &color, FALSE },
      { "reflectivity", "X Y Z",
	"Surface reflectivity X Y Z",
	ARG_LIST | ARG_FLOAT, &reflectivity, FALSE },
      { "exponent", "float/floatexp",
	"Surface phong exponent",
	ARG_DYNFLOAT, &exponent, FALSE },
      { "metallic", "float/floatexp",
	"metallic exponent",
	ARG_DYNFLOAT, &metallic, FALSE },
      { "texture", "textureFileName",
	"Texture image filename",
	ARG_STRING, &texture, FALSE },
      { "type", "texturetype", 
	"SLF texture type",
	ARG_STRING, &texturetype, FALSE },

      { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
	  { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }
    };

  const int argCount = 8;
  
  //printf ("HandleAddSurface: %d\n",argc);
  
  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create surface: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  id = m_pCScanner->FindIdentifier (id, TRUE);
  AddSurface ("[TCL generated Surface]", 0, 0, id);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  struct argList *tempptr;

  if (color != NULL)
    {
      for (tempptr = color; tempptr != NULL; tempptr = tempptr->next)
	{
	  pdynVectf[axis] = new DynFloat;
	  
	  if (tempptr->argType == ARG_EXPR)
	    {
	      //printf ("  %s\n", (CHAR *)(tempptr->value));
	      pdynVectf[axis]->SetDynExpr((CHAR *)(tempptr->value));
	    }
	  else
	    {
		  //printf ("  %f\n", *(FLOAT *)(tempptr->value));
	      pdynVectf[axis]->SetValue(*(FLOAT*)(tempptr->value));
	    }
	  
	  axis++;
	  // if too many things, just jump out of loop and throw err
	  if (axis > 2 && tempptr->next != NULL)
	    {
	      axis = -1;
	      break;
	    }	  
	}
  
      if (axis != 3)
	{
	  TCLAjaxArgList (color);
	  TCLAjaxArgList (reflectivity);
	  HandleError(0, 0, "%s \"%s\" wrong # of arguments for surface color", 
		      gcppcSLIDEReservedWords[sliderwtSurface],
		      id);
	  
	  return TCL_ERROR;
	}
      
      pdynvect = new DynVector;
      ASSERT( pdynvect != NULL );
      pdynvect->SetAll(pdynVectf[0], pdynVectf[1], pdynVectf[2]);
      

      SetSurfaceColor(pdynvect);
    }

  axis=0;  
  if (reflectivity != NULL)
    {
      for (tempptr = reflectivity; tempptr != NULL; tempptr = tempptr->next)
	{
	  pdynVectf[axis] = new DynFloat;
	  
	  if (tempptr->argType == ARG_EXPR)
	    {
	      //printf ("  %s\n", (CHAR *)(tempptr->value));
	      pdynVectf[axis]->SetDynExpr((CHAR *)(tempptr->value));
	    }
	  else
	    {
	      //printf ("  %f\n", *(FLOAT *)(tempptr->value));
	      pdynVectf[axis]->SetValue(*(FLOAT*)(tempptr->value));
	    }
	  
	  axis++;
	  // if too many things, just jump out of loop and throw err
	  if (axis > 2 && tempptr->next != NULL)
	    {
	      axis = -1;
	      break;
	    }	  
	}
      
      if (axis != 3)
	{
	  TCLAjaxArgList (color);
	  TCLAjaxArgList (reflectivity);
	  HandleError(0, 0, "%s \"%s\" wrong # of arguments for surface reflectivity", 
		      gcppcSLIDEReservedWords[sliderwtSurface],
		      id);
	  
	  return TCL_ERROR;
	}
      
      pdynvect = new DynVector;
      ASSERT( pdynvect != NULL );
      pdynvect->SetAll(pdynVectf[0], pdynVectf[1], pdynVectf[2]);
      
      
      SetSurfaceReflectivity(pdynvect);
    }
  
  
  
  if (exponent != NULL)
    SetSurfaceExponent(exponent);
  
  if (metallic != NULL)
    SetSurfaceMetallic(metallic);
  
  if (texture != NULL && texturetype != NULL)
    {
      texture = m_pCScanner->FindIdentifier (texture, TRUE);
      DynInt *pdyniTexture = TranslateTextureFlag(texturetype);
      SetSurfaceTexture(texture,pdyniTexture);
    }
  
  
  //printf ("\n");

  if (ribbegin != NULL)
    SetPrimitiveRIBBegin (ribbegin);

  if (ribend != NULL)
    SetPrimitiveRIBBegin (ribend);

  
  EndSurface();
  
  // Ajax
  //
  TCLAjaxArgList (color);
  TCLAjaxArgList (reflectivity);

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleAddCamera(VOID* &pCreatedObject, int argc, char **argv)
{
  char *id = NULL;
  char *projection = NULL;
  struct argList *frustum = NULL;

  char *ribbegin=NULL;
  char *ribend=NULL;
  
  struct argFormat tclargform[] =
    { { NULL, "cameraID",
	"Name for the new camera",
	ARG_SIMPLE | ARG_STRING, &id, TRUE },
	{ "projection", "projection_flag",
	"SLF projection flag",
	ARG_STRING, &projection, FALSE },
      { "frustum", "{min_triple} {max_triple}",
	"frustum",
	ARG_STRING | ARG_LIST, &frustum, FALSE },

      { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
      { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }
    };

  const int argCount = 5;
  
  //printf ("HandleAddCamera: %d\n",argc);
  
  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create camera: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  id = m_pCScanner->FindIdentifier (id, TRUE);
  AddCamera ("[TCL generated Camera]", 0, 0, id);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  if (projection != NULL)
  {
    DynInt *pdyniProj = TranslateProjectionFlag(projection);
    SetCameraProjection (pdyniProj);
  }

  if (frustum != NULL)
  {
    struct argList *subptr = frustum;
    struct argList *fr_values = NULL;
    DynVector *fr_vects[2];
    int fr_vectnum=0;

    while (subptr != NULL)
      {
	struct argList *fr_temp;
	int axis=0;
	DynFloat *tempdfloats[3];

	fr_values = TCLParseArgs_List ((CHAR *)(subptr->value), ARG_FLOAT);
	
	for (fr_temp=fr_values; fr_temp != NULL; fr_temp=fr_temp->next)
	  {
	    tempdfloats[axis] = new DynFloat;
	    
	    if (fr_temp->argType == ARG_EXPR)	      
	      {	
		tempdfloats[axis]->SetDynExpr((CHAR *)(fr_temp->value));
	      }	
	    else	
	      {
		tempdfloats[axis]->SetValue(*(FLOAT*)(fr_temp->value));
	      }

	    axis++;
	    // if too many things, just jump out of loop and throw err
	    if (axis > 2 && fr_temp->next != NULL)
	      {
		axis = -1;
		break;
	      }
	  }

	if (axis != 3)
	  {
	    TCLAjaxArgList (frustum);
	    HandleError(0, 0, "%s \"%s\" wrong # of arguments for Camera frustum triple", 
			gcppcSLIDEReservedWords[sliderwtCamera],
			id);
      
	    return TCL_ERROR;
	  }

	fr_vects[fr_vectnum] = new DynVector;
	fr_vects[fr_vectnum]->SetAll (tempdfloats[0], tempdfloats[1],
				       tempdfloats[2]);	

	for (;axis >0; axis--)
	  delete tempdfloats[axis-1];
	
	fr_vectnum ++;
	subptr = subptr->next;
	// if too many things, just jump out of loop and throw err
	if (fr_vectnum > 3 && subptr != NULL)
	  {
	    fr_vectnum = -1;
	    subptr = NULL;
	  }
      }

    if (fr_vectnum != 2)
      {
	TCLAjaxArgList (frustum);
	HandleError(0, 0, "%s \"%s\" wrong # of arguments for Camera frustum", 
		    gcppcSLIDEReservedWords[sliderwtCamera],
		    id);
      
	return TCL_ERROR;
      }

    SetCameraFrustum (fr_vects[0], fr_vects[1]);
  }
  
  
  //printf ("\n");

  if (ribbegin != NULL)
    SetPrimitiveRIBBegin (ribbegin);

  if (ribend != NULL)
    SetPrimitiveRIBBegin (ribend);

  
  EndCamera();
  
  // Ajax
  //
  TCLAjaxArgList (frustum);

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleAddLight(VOID* &pCreatedObject, int argc, char **argv)
{
  char *id = NULL;
  char *lighttype = NULL;
  struct argList *color = NULL;
  DynFloat *deaddistance = NULL;
  DynFloat *falloff = NULL;
  DynFloat *angularfalloff = NULL;

  char *ribbegin=NULL;
  char *ribend=NULL;
  
  DynFloat* pdynVectf[4];
  DynVector *pdynvect=NULL;
  int axis=0;

  struct argFormat tclargform[] =
    { { NULL, "lightID",
	"Name for the new light",
	ARG_SIMPLE | ARG_STRING, &id, TRUE },
      { "type", "lighttype", 
	"SLF light type",
	ARG_STRING, &lighttype, FALSE },
      { "color", "R G B", 
	"R, G, and B values for new light",
	ARG_LIST | ARG_FLOAT, &color, FALSE },
      { "deaddistance", "float/floatexp",
	"dead distance",
	ARG_DYNFLOAT, &deaddistance, FALSE },
      { "falloff", "float/floatexp",
	"falloff distance",
	ARG_DYNFLOAT, &falloff, FALSE },
      { "angularfalloff", "float/floatexp",
	"angular falloff",
	ARG_DYNFLOAT, &angularfalloff, FALSE },

      { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
      { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }
    };

  const int argCount = 8;
  
  //printf ("HandleAddLight: %d\n",argc);
  
  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create light: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  id = m_pCScanner->FindIdentifier (id, TRUE);
  AddLight ("[TCL generated Light]", 0, 0, id);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  struct argList *tempptr;

  if (color != NULL)
    {
      for (tempptr = color; tempptr != NULL; tempptr = tempptr->next)
	{
	  pdynVectf[axis] = new DynFloat;
	  
	  if (tempptr->argType == ARG_EXPR)
	    {
	      //printf ("  %s\n", (CHAR *)(tempptr->value));
	      pdynVectf[axis]->SetDynExpr((CHAR *)(tempptr->value));
	    }
	  else
	    {
		  //printf ("  %f\n", *(FLOAT *)(tempptr->value));
	      pdynVectf[axis]->SetValue(*(FLOAT*)(tempptr->value));
	    }
	  
	  axis++;
	  // if too many things, just jump out of loop and throw err
	  if (axis > 2 && tempptr->next != NULL)
	    {
	      axis = -1;
	      break;
	    }	  
	}
  
      if (axis != 3)
	{
	  TCLAjaxArgList (color);
	  HandleError(0, 0, "%s \"%s\" wrong # of arguments for light color", 
		      gcppcSLIDEReservedWords[sliderwtLight],
		      id);
	  
	  return TCL_ERROR;
	}
      
      pdynvect = new DynVector;
      ASSERT( pdynvect != NULL );
      pdynvect->SetAll(pdynVectf[0], pdynVectf[1], pdynVectf[2]);
      
      SetLightColor(pdynvect);
    }

  if (lighttype != NULL)
    {
      DynInt *pdyniLight = TranslateLightFlag(lighttype);
      SetLightType (pdyniLight);
    }
  
  if (deaddistance != NULL)
    SetLightDeadDistance(deaddistance);
  
  if (falloff != NULL)
    SetLightFallOff (falloff);

  if (angularfalloff != NULL)
    SetLightAngularFallOff (angularfalloff);
  
  //printf ("\n");

  if (ribbegin != NULL)
    SetPrimitiveRIBBegin (ribbegin);

  if (ribend != NULL)
    SetPrimitiveRIBBegin (ribend);

  
  EndLight();
  
  // Ajax
  //
  TCLAjaxArgList (color);

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleAddAreaLight(VOID* &pCreatedObject, int argc, char **argv)
{
  char *id = NULL;
  DynInt *samples = NULL;
  struct argList *color = NULL;
  char *surfid = NULL;
  char *shading = NULL;
  char *lod = NULL;

  char *ribbegin=NULL;
  char *ribend=NULL;
  
  DynFloat* pdynVectf[4];
  DynVector *pdynvect=NULL;
  int axis=0;

  struct argFormat tclargform[] =
    { { NULL, "lightID",
	"Name for the new light",
	ARG_SIMPLE | ARG_STRING, &id, TRUE },
      { "color", "R G B", 
	"R, G, and B values for new light",
	ARG_LIST | ARG_FLOAT, &color, FALSE },
      { "surface", "surfid",
	"Shurface ID",
	ARG_STRING, &surfid, FALSE },
      { "shading", "shading_flag",
	"Shading type",
	ARG_STRING, &shading, FALSE },
      { "lod", "lod_flag",
	"Level of Detail flag",
	ARG_STRING, &lod, FALSE },

      { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
      { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }
    };

  const int argCount = 7;
  
  //printf ("HandleAddAreaLight: %d\n",argc);
  
  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create arealight: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  id = m_pCScanner->FindIdentifier (id, TRUE);
  AddAreaLight ("[TCL generated AreaLight]", 0, 0, id);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  struct argList *tempptr;

  if (color != NULL)
    {
      for (tempptr = color; tempptr != NULL; tempptr = tempptr->next)
	{
	  pdynVectf[axis] = new DynFloat;
	  
	  if (tempptr->argType == ARG_EXPR)
	    {
	      //printf ("  %s\n", (CHAR *)(tempptr->value));
	      pdynVectf[axis]->SetDynExpr((CHAR *)(tempptr->value));
	    }
	  else
	    {
		  //printf ("  %f\n", *(FLOAT *)(tempptr->value));
	      pdynVectf[axis]->SetValue(*(FLOAT*)(tempptr->value));
	    }
	  
	  axis++;
	  // if too many things, just jump out of loop and throw err
	  if (axis > 2 && tempptr->next != NULL)
	    {
	      axis = -1;
	      break;
	    }	  
	}
  
      if (axis != 3)
	{
	  TCLAjaxArgList (color);
	  HandleError(0, 0, "%s \"%s\" wrong # of arguments for arealight color", 
		      gcppcSLIDEReservedWords[sliderwtAreaLight],
		      id);
	  
	  return TCL_ERROR;
	}
      
      pdynvect = new DynVector;
      ASSERT( pdynvect != NULL );
      pdynvect->SetAll(pdynVectf[0], pdynVectf[1], pdynVectf[2]);
      
      SetAreaLightColor(pdynvect);
    }
  
  if (samples != NULL)
    {
      SetAreaLightSamples (samples);
    }

  if (surfid != NULL)
    {
      surfid = m_pCScanner->FindIdentifier (surfid, TRUE);
      SetNodeSurfaceID(surfid);
    }

  if (shading != NULL)
  {
    DynInt *pdyniShading = TranslateShadingFlag(shading);
    SetNodeShading (pdyniShading);
  }

  if (lod != NULL)
  {
    DynInt *pdyniLOD = TranslateLODFlag(lod);
    SetNodeLOD (pdyniLOD);
  }
  
  
  //printf ("\n");

  if (ribbegin != NULL)
    SetPrimitiveRIBBegin (ribbegin);

  if (ribend != NULL)
    SetPrimitiveRIBBegin (ribend);

  
  EndAreaLight();
  
  // Ajax
  //
  TCLAjaxArgList (color);

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleAddViewport(VOID* &pCreatedObject, int argc, char **argv)
{
  char *id = NULL;
  char *window_id = NULL;
  struct argList *origin = NULL;
  struct argList *size = NULL;

  char *ribbegin=NULL;
  char *ribend=NULL;
  
  DynFloat* pdynVectf[4];
  DynVector *pdynvect=NULL;
  int axis=0;

  struct argFormat tclargform[] =
    { { NULL, "viewportID",
	"Name for the new viewport",
	ARG_SIMPLE | ARG_STRING, &id, TRUE },
      { NULL, "windowID",
	"Name for the window the viewport will be in",
	ARG_SIMPLE | ARG_STRING, &window_id, TRUE },
      { "origin", "X Y",
	"viewport origin X Y",
	ARG_LIST | ARG_FLOAT, &origin, FALSE },
      { "size", "W H",
	"viewport size W H",
	ARG_LIST | ARG_FLOAT, &size, FALSE },

      { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
      { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }
    };

  const int argCount = 6;
  
  //printf ("HandleAddViewport: %d\n",argc);
  
  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create viewport: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  id = m_pCScanner->FindIdentifier (id, TRUE);
  window_id = m_pCScanner->FindIdentifier (window_id, TRUE);
  AddViewport ("[TCL generated Viewport]", 0, 0, id, window_id);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  struct argList *tempptr;

  if (origin != NULL)
    {
      for (tempptr = origin; tempptr != NULL; tempptr = tempptr->next)
	{
	  pdynVectf[axis] = new DynFloat;
	  
	  if (tempptr->argType == ARG_EXPR)
	      pdynVectf[axis]->SetDynExpr((CHAR *)(tempptr->value));
	  else
	      pdynVectf[axis]->SetValue(*(FLOAT*)(tempptr->value));
	  
	  axis++;
	  // if too many things, just jump out of loop and throw err
	  if (axis > 1 && tempptr->next != NULL)
	    {
	      axis = -1;
	      break;
	    }	  
	}
      
      if (axis != 2)
	{
	  TCLAjaxArgList (origin);
	  TCLAjaxArgList (size);
	  HandleError(0, 0, "%s \"%s\" wrong # of arguments for viewport origin", 
		      gcppcSLIDEReservedWords[sliderwtViewport],
		      id);
	  
	  return TCL_ERROR;
	}

      pdynvect = new DynVector;
      ASSERT( pdynvect != NULL );
      pdynvect->SetAll(pdynVectf[0], pdynVectf[1]);
      
      SetViewportOrigin(pdynvect);
    }

  axis = 0;
  if (size != NULL)
    {
      for (tempptr = size; tempptr != NULL; tempptr = tempptr->next)
	{
	  pdynVectf[axis] = new DynFloat;
	  
	  if (tempptr->argType == ARG_EXPR)
	      pdynVectf[axis]->SetDynExpr((CHAR *)(tempptr->value));
	  else
	      pdynVectf[axis]->SetValue(*(FLOAT*)(tempptr->value));
	  
	  axis++;
	  // if too many things, just jump out of loop and throw err
	  if (axis > 1 && tempptr->next != NULL)
	    {
	      axis = -1;
	      break;
	    }	  
	}
      
      if (axis != 2)
	{
	  TCLAjaxArgList (origin);
	  TCLAjaxArgList (size);
	  HandleError(0, 0, "%s \"%s\" wrong # of arguments for viewport size", 
		      gcppcSLIDEReservedWords[sliderwtViewport],
		      id);
	  
	  return TCL_ERROR;
	}

      pdynvect = new DynVector;
      ASSERT( pdynvect != NULL );
      pdynvect->SetAll(pdynVectf[0], pdynVectf[1]);
      
      SetViewportSize(pdynvect);
    }


  //printf ("\n");

  if (ribbegin != NULL)
    {
      SetPrimitiveRIBBegin (ribbegin);
    }

  if (ribend != NULL)
    {
      SetPrimitiveRIBBegin (ribend);
    }
  
  EndViewport();
  
  // Ajax
  //
  TCLAjaxArgList (origin);
  TCLAjaxArgList (size);
  

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleAddRender(VOID* &pCreatedObject, int argc, char **argv)
{
  char *vp_id = NULL;
  char *cam_path = NULL;
  char *node_id = NULL;
  char *tcltag = NULL;
  char *lod = NULL;
  // TODO: stencils

  struct argList *fog = NULL;
  struct argList *lights = NULL;
  struct argList *arealights = NULL;

  char *cuitype = NULL;
  DynFloat *cuispeed = NULL;

  char *ribbegin=NULL;
  char *ribend=NULL;
  
  struct argFormat tclargform[] =
    { { NULL, "viewportID",
	"viewport to render",
	ARG_SIMPLE | ARG_STRING, &vp_id, TRUE },
      { NULL, "camera_path",
	"camera path to render",
	ARG_SIMPLE | ARG_STRING, &cam_path, TRUE },
      { NULL, "nodeID",
	"node to render",
	ARG_SIMPLE | ARG_STRING, &node_id, TRUE },
      { "tcltag", "id",
	"TCL tag",
	ARG_STRING, &tcltag, FALSE },
      { "fog", "fog_id [onoff_flag]", 
	"fog effect to render",
	ARG_LIST | ARG_STRING, &fog, FALSE },
      { "light", "{light_path [onoff_flag]} ...", 
	"light(s) to render",
	ARG_LIST | ARG_STRING, &lights, FALSE },
      { "arealight", "{arealight_path [onoff_flag]} ...", 
	"arealight(s) to render",
	ARG_LIST | ARG_STRING, &arealights, FALSE },
      { "lod", "lod_flag", 
	"LOD flag",
	ARG_STRING, &lod, FALSE },

	// Camera UI
      { "cuitype", "cuitype",
	"camera UI Type",
	ARG_STRING, &cuitype, FALSE },
      { "cuispeed", "uispeed",
	"camera UI Speed",
	ARG_DYNFLOAT, &cuispeed, FALSE },

  { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
      { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }
    };

  const int argCount = 10;

  DynInt *dyniOnOff = NULL;

  
  //printf ("HandleAddRender: %d\n",argc);
  
  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create render: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  CListInt *cam_path_list = CreateIDList();
  char *temp;

  temp = strtok (cam_path, ".");
  while (temp != NULL)
    {
	temp = m_pCScanner->FindIdentifier (temp, FALSE);
	//printf ("Adding camerapath %s\n", temp);
        InsertIDList(cam_path_list, temp);
	temp = strtok (NULL, ".");
    }

  vp_id = m_pCScanner->FindIdentifier (vp_id, TRUE);
  node_id = m_pCScanner->FindIdentifier (node_id, TRUE);
  AddRender ("[TCL generated Render]", 0, 0, vp_id, cam_path_list, node_id);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  if (tcltag != NULL)
    SetRenderTclTag ( tcltag );

  // fog
  if (fog != NULL)
    {
      char *fogid;
      dyniOnOff = NULL;

      if (fog->next != NULL)
	{
	  dyniOnOff = TranslateLODFlag ( (CHAR*)(fog->next->value) );

	  if (fog->next->next == NULL)
	    {
	      TCLAjaxArgList (fog);
	      TCLAjaxArgList (lights);
	      TCLAjaxArgList (arealights);
	      HandleError(0, 0, "%s wrong # of arguments for render fog statement", 
			  gcppcSLIDEReservedWords[sliderwtRender]);
	      
	      return TCL_ERROR;
	      
	    }
	}
      
      fogid = m_pCScanner->FindIdentifier ((CHAR*)(fog->value), FALSE);
      SetRenderFogID ( fogid, dyniOnOff );
    }

  struct argList *subptr;
  struct argList *tempptr;

  if (lights != NULL)
    {
      tempptr = lights;
      while (tempptr != NULL)
	{
	  dyniOnOff = NULL;
	  subptr = TCLParseArgs_List ((CHAR *)(tempptr->value), ARG_STRING);

	  if (subptr->next != NULL)
	    {
	      dyniOnOff = TranslateLODFlag ( (CHAR*)(subptr->next->value) );
	  
	      if (subptr->next->next == NULL)
		{
		  TCLAjaxArgList (subptr);
		  TCLAjaxArgList (fog);
		  TCLAjaxArgList (lights);
		  TCLAjaxArgList (arealights);
		  HandleError(0, 0, "%s wrong # of arguments for render statement, light arg", 
			      gcppcSLIDEReservedWords[sliderwtRender]);
	      
		  return TCL_ERROR;
	      
		}
	    }

	  CListInt *light_path_list = CreateIDList();
	  char *temp;

	  temp = strtok ((CHAR*)(subptr->value), ".");

	  while (temp != NULL)
	    {
	      temp = m_pCScanner->FindIdentifier (temp, FALSE);
	      //printf ("Adding lightpath %s\n", temp);
	      InsertIDList(light_path_list, temp);
	      temp = strtok (NULL, ".");
	    }

      
	  AddRenderLightPath ( "[TCL generated Render (light)]", 0, 0,
			       light_path_list, dyniOnOff );


	  tempptr = tempptr->next;
	}
    }


  if (arealights != NULL)
    {
      tempptr = arealights;
      while (tempptr != NULL)
	{
	  dyniOnOff = NULL;
	  subptr = TCLParseArgs_List ((CHAR *)(tempptr->value), ARG_STRING);

	  if (subptr->next != NULL)
	    {
	      dyniOnOff = TranslateLODFlag ( (CHAR*)(subptr->next->value) );
	  
	      if (subptr->next->next == NULL)
		{
		  TCLAjaxArgList (subptr);
		  TCLAjaxArgList (fog);
		  TCLAjaxArgList (lights);
		  TCLAjaxArgList (arealights);
		  HandleError(0, 0, "%s wrong # of arguments for render statement, light arg", 
			      gcppcSLIDEReservedWords[sliderwtRender]);
	      
		  return TCL_ERROR;
	      
		}
	    }

	  CListInt *arealight_path_list = CreateIDList();
	  char *temp;

	  temp = strtok ((CHAR*)(subptr->value), ".");

	  while (temp != NULL)
	    {
	      temp = m_pCScanner->FindIdentifier (temp, FALSE);
	      //printf ("Adding arealightpath %s\n", temp);
	      InsertIDList(arealight_path_list, temp);
	      temp = strtok (NULL, ".");
	    }

      
	  AddRenderAreaLightPath ( "[TCL generated Render (arealight)]", 0, 0,
			       arealight_path_list, dyniOnOff );

	  tempptr=tempptr->next;
	}
    }
  


  if (lod != NULL)
  {
    DynInt *pdyniLOD = TranslateLODFlag(lod);
    SetRenderLOD (pdyniLOD);
  }

  // Camera UI
  if ( NULL != cuitype )
  {
	DynInt *pdyniCUIType = TranslateCUIType( cuitype );
	SetRenderCUIType (pdyniCUIType);
  }

  if ( NULL != cuispeed )
	SetRenderCUISpeed (cuispeed);

  if (ribbegin != NULL)
    SetPrimitiveRIBBegin (ribbegin);

  if (ribend != NULL)
    SetPrimitiveRIBBegin (ribend);


  EndRender();
  
  // Ajax
  //
  TCLAjaxArgList (fog);
  TCLAjaxArgList (lights);
  TCLAjaxArgList (arealights);

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleAddWindow(VOID* &pCreatedObject, int argc, char **argv)
{
  char *id = NULL;
  struct argList *background = NULL;
  struct argList *origin = NULL;
  struct argList *size = NULL;
  char *resize = NULL;
  char *ratio = NULL;

  char *ribbegin=NULL;
  char *ribend=NULL;
  
  DynFloat* pdynVectf[4];
  DynVector *pdynvect=NULL;
  int axis=0;

  struct argFormat tclargform[] =
    { { NULL, "windowID",
	"Name for the new window",
	ARG_SIMPLE | ARG_STRING, &id, TRUE },
      { "background", "R G B", 
	"R, G, and B values for new window",
	ARG_LIST | ARG_FLOAT, &background, FALSE },
      { "origin", "X Y",
	"window origin X Y",
	ARG_LIST | ARG_FLOAT, &origin, FALSE },
      { "size", "W H",
	"window size W H",
	ARG_LIST | ARG_FLOAT, &size, FALSE },
      { "resize", "onoff_flag",
	"Resize on/off",
	ARG_STRING, &resize, FALSE },
      { "ratio", "onoff_flag",
	"Ratio on/off",
	ARG_STRING, &ratio, FALSE },

      { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
      { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }
    };

  const int argCount = 8;
  
  //printf ("HandleAddWindow: %d\n",argc);
  
  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide create window: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  id = m_pCScanner->FindIdentifier (id, TRUE);
  AddWindow ("[TCL generated Window]", 0, 0, id);
  pCreatedObject = m_pCSLIDEPrimitiveCurr;

  struct argList *tempptr;

  if (background != NULL)
    {
      for (tempptr = background; tempptr != NULL; tempptr = tempptr->next)
	{
	  pdynVectf[axis] = new DynFloat;
	  
	  if (tempptr->argType == ARG_EXPR)
	      pdynVectf[axis]->SetDynExpr((CHAR *)(tempptr->value));
	  else
	      pdynVectf[axis]->SetValue(*(FLOAT*)(tempptr->value));
	  
	  axis++;
	  // if too many things, just jump out of loop and throw err
	  if (axis > 2 && tempptr->next != NULL)
	    {
	      axis = -1;
	      break;
	    }	  
	}
  
      if (axis != 3)
	{
	  TCLAjaxArgList (background);
	  TCLAjaxArgList (origin);
	  TCLAjaxArgList (size);
	  HandleError(0, 0, "%s \"%s\" wrong # of arguments for window background", 
		      gcppcSLIDEReservedWords[sliderwtWindow],
		      id);
	  
	  return TCL_ERROR;
	}
      
      pdynvect = new DynVector;
      ASSERT( pdynvect != NULL );
      pdynvect->SetAll(pdynVectf[0], pdynVectf[1], pdynVectf[2]);
      

      SetWindowBackground(pdynvect);
    }
  
  axis = 0;
  if (origin != NULL)
    {
      for (tempptr = origin; tempptr != NULL; tempptr = tempptr->next)
	{
	  pdynVectf[axis] = new DynFloat;
	  
	  if (tempptr->argType == ARG_EXPR)
	      pdynVectf[axis]->SetDynExpr((CHAR *)(tempptr->value));
	  else
	      pdynVectf[axis]->SetValue(*(FLOAT*)(tempptr->value));
	  
	  axis++;
	  // if too many things, just jump out of loop and throw err
	  if (axis > 1 && tempptr->next != NULL)
	    {
	      axis = -1;
	      break;
	    }	  
	}
      
      if (axis != 2)
	{
	  TCLAjaxArgList (background);
	  TCLAjaxArgList (origin);
	  TCLAjaxArgList (size);
	  HandleError(0, 0, "%s \"%s\" wrong # of arguments for window origin", 
		      gcppcSLIDEReservedWords[sliderwtWindow],
		      id);
	  
	  return TCL_ERROR;
	}
      
      SetWindowOrigin(pdynVectf[0]->GetValue(), pdynVectf[1]->GetValue());
    }

  axis = 0;
  if (size != NULL)
    {
      for (tempptr = size; tempptr != NULL; tempptr = tempptr->next)
	{
	  pdynVectf[axis] = new DynFloat;
	  
	  if (tempptr->argType == ARG_EXPR)
	      pdynVectf[axis]->SetDynExpr((CHAR *)(tempptr->value));
	  else
	      pdynVectf[axis]->SetValue(*(FLOAT*)(tempptr->value));
	  
	  axis++;
	  // if too many things, just jump out of loop and throw err
	  if (axis > 1 && tempptr->next != NULL)
	    {
	      axis = -1;
	      break;
	    }	  
	}
      
      if (axis != 2)
	{
	  TCLAjaxArgList (background);
	  TCLAjaxArgList (origin);
	  TCLAjaxArgList (size);
	  HandleError(0, 0, "%s \"%s\" wrong # of arguments for window size", 
		      gcppcSLIDEReservedWords[sliderwtWindow],
		      id);
	  
	  return TCL_ERROR;
	}
      
      SetWindowSize(pdynVectf[0]->GetValue(), pdynVectf[1]->GetValue());
    }


  if (resize != NULL)
  {
    DynInt *pdyniResize = TranslateOnOffFlag(resize);
    SetWindowResize (pdyniResize->GetValue());
  }

  if (ratio != NULL)
  {
    DynInt *pdyniRatio = TranslateOnOffFlag(ratio);
    SetWindowAspectRatio (pdyniRatio->GetValue());
  }
  
    
  //printf ("\n");

  if (ribbegin != NULL)
    SetPrimitiveRIBBegin (ribbegin);

  if (ribend != NULL)
    SetPrimitiveRIBBegin (ribend);

  
  EndWindow();
  
  // Ajax
  //
  TCLAjaxArgList (background);
  TCLAjaxArgList (origin);
  TCLAjaxArgList (size);
  

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

///////////////////////////////////////////////////////////////////////////////////
// HandleModify

int CSLIDEParser::HandleModifyPoint(VOID* &pModifiedObject, int argc, char **argv)
{
  CHAR *pcScannerID;
  CSLIDEPoint *pCSLIDEPoint;

  char *ptname = NULL;
  struct argList *ptcoords = NULL;
  char *surfid = NULL;
  struct argList *ptnormal = NULL;
  struct argList *pttexcoords = NULL;
  char *ribbegin = NULL;
  char *ribend = NULL;

  DynFloat* pdynVectf[4];
  DynPoint *pdynpt=NULL;
  DynVector *pdynvect=NULL;
  int axis=0;

  struct argList *tempptr;

  struct argFormat tclargform[] =
    { { NULL, "name",
	"Name for the point",
	ARG_SIMPLE | ARG_STRING, &ptname, TRUE },
      { NULL, "X Y Z", 
	"X, Y, and Z values for the point",
	ARG_SIMPLE | ARG_LIST | ARG_FLOAT, &ptcoords, TRUE },
      { "normal", "X Y Z",
	"Point normal X Y Z",
	ARG_LIST | ARG_FLOAT, &ptnormal, FALSE },
      { "surface", "surf_id",
	"Point Surface ID",
	ARG_STRING, &surfid, FALSE },
      { "texture", "X Y Z",
	"texture mapping coordinates X Y Z",
	ARG_LIST | ARG_FLOAT, &pttexcoords, FALSE },

	  { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
	  { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }
    };

  const int argCount = 7;
  
  //printf ("HandleModifyPoint: %d\n",argc);
  
  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide modify point: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  pcScannerID = m_pCScanner->FindIdentifier(ptname, FALSE);
  ASSERT( pcScannerID != NULL );
  pCSLIDEPoint = NULL;
  pCSLIDEPoint = c_pCCurrParser->m_pCSLIDEWorld->PeekPoint(pcScannerID);

  if( pCSLIDEPoint == NULL )
    {
      printf ("Syntax Error: slide modify point %s: point %s does not exist\n", 
	      pcScannerID, pcScannerID);
      return TCL_ERROR;
    }

  pModifiedObject = pCSLIDEPoint;

  if( ptcoords != NULL )
    {
      for (tempptr = ptcoords; tempptr != NULL; tempptr = tempptr->next)
	{
	  pdynVectf[axis] = new DynFloat;
	  
	  if (tempptr->argType == ARG_EXPR)
	    {
	      //printf ("  %s\n", (CHAR *)(tempptr->value));
	      pdynVectf[axis]->SetDynExpr((CHAR *)(tempptr->value));
	    }
	  else
	    {
	      //printf ("  %f\n", *(FLOAT *)(tempptr->value));
	      pdynVectf[axis]->SetValue(*(FLOAT*)(tempptr->value));
	    }
	  
	  axis++;
	  
	  // if too many things, just jump out of loop and throw err
	  if (axis > 3 && tempptr->next != NULL)
	    {
	      axis = -1;
	      break;
	    }
	}
      
      if (axis != 3)
	{
	  // Ajax
	  //
	  TCLAjaxArgList (pttexcoords);
	  TCLAjaxArgList (ptcoords);
	  TCLAjaxArgList (ptnormal);
	  
	  HandleError(0, 0, "%s \"%s\" wrong # of arguments for pt coordinates", 
		      gcppcSLIDEReservedWords[sliderwtPoint], ptname);
      
	  return TCL_ERROR;
	}
      
      pdynpt = new DynPoint;
      ASSERT( pdynpt != NULL );
      pdynpt->SetAll(pdynVectf[0], pdynVectf[1], pdynVectf[2]);

      pCSLIDEPoint->SetPoint(pdynpt);
    }  

  if (surfid != NULL)
    {
      surfid = m_pCScanner->FindIdentifier(surfid, TRUE);
      pCSLIDEPoint->SetSurfaceID (surfid);
    }
  
  if (ptnormal != NULL)
    {
      axis = 0;
      for (tempptr = ptnormal; tempptr != NULL; tempptr = tempptr->next)
	{
	  pdynVectf[axis] = new DynFloat;
	  
	  if (tempptr->argType == ARG_EXPR)
	    {
	      //printf ("	 %s\n", (CHAR *)(tempptr->value));
	      pdynVectf[axis]->SetDynExpr((CHAR *)(tempptr->value));
	    }
	  else
	    {
	      //printf ("	 %f\n", *(FLOAT *)(tempptr->value));
	      pdynVectf[axis]->SetValue(*(FLOAT*)(tempptr->value));
	    }
	  
	  axis++;
	  
	  // if too many things, just jump out of loop and throw err
	  if (axis > 3 && tempptr->next != NULL)
	    {
	      axis = -1;
	      break;
	    }
	}
      
      if (axis != 3)
	{
	  // Ajax
	  //
	  TCLAjaxArgList (pttexcoords);
	  TCLAjaxArgList (ptcoords);
	  TCLAjaxArgList (ptnormal);

	  HandleError(0, 0, "%s \"%s\" wrong # of arguments for pt normal", 
		      gcppcSLIDEReservedWords[sliderwtPoint],
		      ptname);
	  
	  return TCL_ERROR;
	}
      
      pdynvect = new DynVector;
      ASSERT( pdynvect != NULL );
      pdynvect->SetAll(pdynVectf[0], pdynVectf[1], pdynVectf[2]);
      
      pCSLIDEPoint->SetNormal(pdynvect);
    }
  
  if (pttexcoords != NULL)
    {
      //printf ("pttexcoords: \n");
      
      axis = 0;
      for (tempptr = pttexcoords; tempptr != NULL; tempptr = tempptr->next)
	{
	  pdynVectf[axis] = new DynFloat;
	  
	  if (tempptr->argType == ARG_EXPR)
	    {
	      //printf ("	 %s\n", (CHAR *)(tempptr->value));
	      pdynVectf[axis]->SetDynExpr((CHAR *)(tempptr->value));
	    }
	  else
	    {
		//printf ("  %f\n", *(FLOAT *)(tempptr->value));
		pdynVectf[axis]->SetValue(*(FLOAT*)(tempptr->value));
	    }
	  
	  axis++;

	  // if too many things, just jump out of loop and throw err
	  if (axis > 3 && tempptr->next != NULL)
	    {
	      axis = -1;
	      break;
	    }
	}

      if (axis != 3)
	{
	  // Ajax
	  //
	  TCLAjaxArgList (pttexcoords);
	  TCLAjaxArgList (ptcoords);
	  TCLAjaxArgList (ptnormal);

	  HandleError(0, 0, "%s \"%s\" wrong # of arguments for pt texture coords", 
		      gcppcSLIDEReservedWords[sliderwtPoint],
		      ptname);
	 
	  return TCL_ERROR;
	}
      
      pdynpt = new DynPoint;
      ASSERT( pdynpt != NULL );
      pdynpt->SetAll(pdynVectf[0], pdynVectf[1], pdynVectf[2]);
      
      pCSLIDEPoint->SetTexture(pdynpt);
    }
  
  if (ribbegin != NULL)
    pCSLIDEPoint->SetRIBBegin(ribbegin);

  if (ribend != NULL)
    pCSLIDEPoint->SetRIBBegin(ribend);

  
  // Ajax
  //
  TCLAjaxArgList (pttexcoords);
  TCLAjaxArgList (ptcoords);
  TCLAjaxArgList (ptnormal);

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

int CSLIDEParser::HandleModifyPolyline(VOID* &pModifiedObject, int argc, char **argv)
{
  CHAR *pcScannerID;
  CSLIDEPolyline *pCSLIDEPolyline;

  char *id = NULL;
  DynInt *drawpoints = NULL;
  struct argList *pointlist = NULL;

  char *surfid = NULL;
  char *shading = NULL;
  char *lod = NULL;
  char *ribbegin = NULL;
  char *ribend = NULL;

  struct argFormat tclargform[] =
    { { NULL, "id",
	"Name for the Polyline",
	ARG_SIMPLE | ARG_STRING, &id, TRUE },
      { "drawpointss", "int/intexpr",
	"Draw Points",
	ARG_DYNINT, &drawpoints, FALSE },
      { "pointlist", "float/floatexpr list",
	"Point list",
	ARG_STRING | ARG_LIST, &pointlist, FALSE },
      { "surface", "surfid",
	"Surface ID",
	ARG_STRING, &surfid, FALSE },
      { "shading", "shading_flag",
	"Shading type",
	ARG_STRING, &shading, FALSE },
      { "lod", "lod_flag",
	"Level of Detail flag",
	ARG_STRING, &lod, FALSE },

      { "ribbegin", "rib_string",
	"Begin RIB string",
	ARG_STRING, &ribbegin, FALSE },
      { "ribend", "rib_string",
	"End RIB string",
	ARG_STRING, &ribend, FALSE }
  };

  const int argCount = 8;

  //printf ("HandleModifyPolyline: %d\n",argc);

  int err = TCLParseArgs(tclargform, argCount, argc, argv);
  if (err)
    {
      printf ("Syntax Error: slide modify polyline: arg #%d: \"%s\"\n", err, argv[err]);
      return TCL_ERROR;
    }

  pcScannerID =  m_pCScanner->FindIdentifier(id, TRUE);
  ASSERT( pcScannerID != NULL );
  pCSLIDEPolyline = NULL;
  pCSLIDEPolyline = (CSLIDEPolyline *)c_pCCurrParser->m_pCSLIDEWorld->PeekNode(pcScannerID);

  if( pCSLIDEPolyline == NULL )
    {
      printf ("Syntax Error: slide modify polyline %s: polyline %s does not exist\n", 
	      pcScannerID, pcScannerID);
      return TCL_ERROR;
    }

  pModifiedObject = pCSLIDEPolyline;

  if (drawpoints != NULL)
    pCSLIDEPolyline->SetDrawPoints(drawpoints);

  if (pointlist != NULL)
    {
      struct argList *tempptr;
      
      CListInt* pPList = CreateList();
      for (tempptr = pointlist; tempptr != NULL; tempptr = tempptr->next)
	{
	  tempptr->value = (VOID*)(m_pCScanner->FindIdentifier( (CHAR*)(tempptr->value), 
								 TRUE) );
	  InsertList (pPList, (CHAR*)(tempptr->value));
	}
      
      pCSLIDEPolyline->SetPointList(pPList);
  }

  if (surfid != NULL)
    {
      surfid = m_pCScanner->FindIdentifier (surfid, TRUE);
      pCSLIDEPolyline->SetSurfaceID(surfid);
    }
  
  if (shading != NULL)
    {
      DynInt *pdyniShading = TranslateShadingFlag(shading);
      pCSLIDEPolyline->SetShading(pdyniShading);
    }
  
  if (lod != NULL)
    {
      DynInt *pdyniLOD = TranslateLODFlag(lod);
      pCSLIDEPolyline->SetLOD(pdyniLOD);
    }
  
  if (ribbegin != NULL)
    pCSLIDEPolyline->SetRIBBegin(ribbegin);

  if (ribend != NULL)
    pCSLIDEPolyline->SetRIBBegin(ribend);


  // Ajax
  //
  TCLAjaxArgList (pointlist);

  if (m_uiErrorCount > 0)
    return TCL_ERROR;

  return TCL_OK;
}

// HandleModify
///////////////////////////////////////////////////////////////////////////////////


int CSLIDEParser::CreateOotClass(Tcl_Interp *pInterp)
{
  ASSERT( pInterp != NULL );

  m_pInterp = pInterp;
  m_pOotClass = Oot_CreateStaticClass(m_pInterp, m_cpcOotClassName, 
				      CSLIDEParser::m_pOotMethods, 
				      cuiCSLIDEParserOotMethods);

  if ( m_pOotClass == NULL )
    {
      return TCL_ERROR;
    }

  //m_hashOotSLIDEAPIFuncMap

  return TCL_OK;
}

//////////////////////////////////////////////////////////////////////////////
//
// SLIDE API Commands

int CSLIDEParser::HandleCreate(int argc, char **argv)
{
  int result;
  VOID *pCreatedObject;  
  pCreatedObject = NULL;
  
  // Check args
  if (argc <= 1)
    {
      c_pCCurrParser->HandleError(0, 0, " \"slide create\": no TYPE given");
      return TCL_ERROR;
    }
  
  int (CSLIDEParser::*cmdmethod)( VOID* &, int, char** );
  int *pos = (int*)(c_pCCurrParser->m_hashOotCreateFuncMap.Peek(argv[1]));
  if (pos == NULL)
    {
      c_pCCurrParser->HandleError(0, 0, " \"slide create %s\": %s is not a valid SLIDE type", argv[1]);
      return TCL_ERROR;
    }

  //////////////////////////////////////////////////////////////////////////////////  
  // Create objects: go through parser and create objects
  cmdmethod = (m_pCreateCmdMethods[ *pos ]);
  result = ( c_pCCurrParser->*cmdmethod )( pCreatedObject, argc-1, &(argv[1]) );

  if( result == TCL_ERROR )
    {
      return result;
    }

  //printf( "Tcl created %s\n", argv[2] );

  //////////////////////////////////////////////////////////////////////////////////
  // If objects are being added after checksemantics or preprocess then
  // we need to update the system properly
  if( (gbPreprocessDone == TRUE) || ( gbCheckSemanticsDone == TRUE ) )
    {
      ASSERT(pCreatedObject != NULL);

      UINT_32 uiNewNodeType;

      // Get the enum SLIDE type of the new node
      for( uiNewNodeType=0; uiNewNodeType < sliderwtSize; uiNewNodeType++ )
	{
	  if( !strcmp(argv[1], gcppcSLIDEReservedWords[uiNewNodeType]) )
	    {
	      break;
	    }
	}
      
      // Make sure these enum's are valid
      ASSERT( uiNewNodeType < sliderwtSize );

      switch(uiNewNodeType) 
	{
	case sliderwtInstance: 
	  {
	    CSLIDEInstance *pCSLIDEInstance;
	    pCSLIDEInstance = (CSLIDEInstance *)pCreatedObject;

	    // CheckSemantics: Initialization
	    pCSLIDEInstance->CheckSemantics(c_pCCurrParser->m_pCSLIDEWorld);

	    if( gbPreprocessDone == TRUE )
	      {
		// Preprocess: Mark the path up to root as unvisited
		pCSLIDEInstance->SetUnvisited();
		pCSLIDEInstance->Preprocess();
		c_pCCurrParser->m_pCSLIDEWorld->PreprocessGraphs();	    
		
		// Add to world dynamic list
		if( pCSLIDEInstance->GetDynamic() )
		  {
 	         c_pCCurrParser->m_pCSLIDEWorld->m_listDynamics.InsertLast(pCSLIDEInstance,
			 					  (UINT_32)pCSLIDEInstance);
		  }
		
		c_pCCurrParser->m_pCSLIDEWorld->Reprerender((CSLIDEPrimitive*)pCSLIDEInstance);
	      }
	  }
	  break;

	  // Special Geometry
	case sliderwtGroup:
	case sliderwtPoint:  
	case sliderwtFace:
	case sliderwtObject: 
	  // Geometry
	case sliderwtBezierPatch:
	case sliderwtBSplinePatch:
	case sliderwtCone:
	case sliderwtCylinder:
	case sliderwtCyclide:
	case sliderwtSphere:
	case sliderwtTorus:
	  // need to write HandleAdd for these 
	case sliderwtHeightfield:
	case sliderwtBezierCurve:
	case sliderwtBSplineCurve:
	case sliderwtPolyline:
	case sliderwtSweep:
	  // Haven't been tested
	case sliderwtConvexHull: 
	case sliderwtTerrain: 
	case sliderwtFog: 
	case sliderwtSurface:
	  {
	    CSLIDEPrimitive *pCSLIDEPrimitive;
	    pCSLIDEPrimitive = (CSLIDEPrimitive *)pCreatedObject;

	    // CheckSemantics: Initialization
	    pCSLIDEPrimitive->CheckSemantics(c_pCCurrParser->m_pCSLIDEWorld);

	    if( gbPreprocessDone == TRUE )
	      {
		// Preprocess
		pCSLIDEPrimitive->Preprocess();
		
		// Add to world dynamic list
		if( pCSLIDEPrimitive->GetDynamic() )
		  {
	        c_pCCurrParser->m_pCSLIDEWorld->m_listDynamics.InsertLast(pCSLIDEPrimitive,
		   					         (UINT_32)pCSLIDEPrimitive);
		  }
	      }
	  }
	  break;

	  // not sure how to handles these yet
	case sliderwtAreaLight: 
	case sliderwtCamera: 
	case sliderwtLight: 
	case sliderwtViewport:
	case sliderwtRender:
	case sliderwtWindow:

	default:
	  ASSERT(FALSE);
	  break;
	}
    }

  return TCL_OK;
}

int CSLIDEParser::HandleDelete(int argc, char **argv)
{
  if(argc != 3)
    {
      c_pCCurrParser->HandleError(0, 0, " \"slide delete\": expecting 2 args TYPE and ID");
      return TCL_ERROR;
    }

  UINT_32 uiType;
  CHAR *pcID, *pcType, *pcScannerID;
  CListIterInt iter;
  CSLIDEInstance *pCSLIDEInstance;
  CSLIDENode *pCSLIDENode;
  CSLIDEFace *pCSLIDEFace;
  CSLIDEPoint *pCSLIDEPoint;
  CSLIDESurface *pCSLIDESurface;
  CSLIDEFog *pCSLIDEFog;
  CSLIDEPrimitive *pCSLIDEPrimitive;

  pcType = argv[1];
  pcID   = argv[2];

  pcScannerID = NULL;
  pcScannerID = m_pCScanner->FindIdentifier(pcID, FALSE);
  ASSERT( pcScannerID != NULL );

  // Get the enum SLIDE type of the new node
  for( uiType=0; uiType < sliderwtSize; uiType++ )
    {
      if( !strcmp(pcType, gcppcSLIDEReservedWords[uiType]) )
	{
	  break;
	}
    }

  // Make sure these enum's are valid
  if( uiType == sliderwtSize )
    {
      c_pCCurrParser->HandleError(0, 0, " \"slide delete %s\": %s is not a valid SLIDE type", pcType, pcType);
      return TCL_ERROR;
    }

  switch(uiType) 
    {
    case sliderwtInstance:
      { 
	pCSLIDEInstance = c_pCCurrParser->m_pCSLIDEWorld->PeekInstance(pcScannerID);
	pCSLIDEPrimitive = (CSLIDEPrimitive*)pCSLIDEInstance;
      }
      break;
    case sliderwtGroup:
    case sliderwtBezierCurve:
    case sliderwtBezierPatch:
    case sliderwtBSplineCurve:
    case sliderwtBSplinePatch:
    case sliderwtHeightfield:
    case sliderwtCone:
    case sliderwtCylinder:
    case sliderwtCyclide:
    case sliderwtObject: 
    case sliderwtPolyline:
    case sliderwtSphere:
    case sliderwtSweep:
    case sliderwtTorus:
      {
	pCSLIDENode = c_pCCurrParser->m_pCSLIDEWorld->PeekNode(pcScannerID);
	pCSLIDEPrimitive = (CSLIDEPrimitive*)pCSLIDENode;
      }
      break;
    case sliderwtSurface:
      {
	pCSLIDESurface = c_pCCurrParser->m_pCSLIDEWorld->PeekSurface(pcScannerID);
	pCSLIDEPrimitive = (CSLIDEPrimitive*)pCSLIDESurface;
      }
      break;
    case sliderwtFog: 
      {
	pCSLIDEFog = c_pCCurrParser->m_pCSLIDEWorld->PeekFog(pcScannerID);
	pCSLIDEPrimitive = (CSLIDEPrimitive*)pCSLIDEFog;
      }
      break;
    case sliderwtFace: 
      {
	pCSLIDEFace = c_pCCurrParser->m_pCSLIDEWorld->PeekFace(pcScannerID);
	pCSLIDEPrimitive = (CSLIDEPrimitive*)pCSLIDEFace;
      }
      break;
    case sliderwtPoint: 
      {
	pCSLIDEPoint = c_pCCurrParser->m_pCSLIDEWorld->PeekPoint(pcScannerID);
	pCSLIDEPrimitive = (CSLIDEPrimitive*)pCSLIDEPoint;
      }
      break;

    case sliderwtAreaLight: 
    case sliderwtCamera: 
    case sliderwtLight: 
    default:
      ASSERT(FALSE);
      break;
    }

  if( pCSLIDEPrimitive == NULL )
    {
      c_pCCurrParser->HandleError(0, 0, " \"slide delete %s %s\": %s %s does not exist\n", pcType, pcID, pcType, pcID);
      return TCL_ERROR;
    }
  
  // If system has been initialized
  if( gbPreprocessDone == TRUE )
    {
      // Mark Graphs as unvisited
      pCSLIDEPrimitive->SetBoundUnvisited(); 
      pCSLIDEPrimitive->SetPrerenderUnvisited();
      
      // Remove this object 
      pCSLIDEPrimitive->Remove(c_pCCurrParser->m_pCSLIDEWorld);
      
      // Update geometry
      if( uiType == sliderwtPoint )
	{
	  pCSLIDEPrimitive->RecalculateParent();
	}
      
      // Update bounding boxes
      // Recalculate bboxes; BoundChanged=TRUE
      c_pCCurrParser->m_pCSLIDEWorld->UpdateGraphs();
      // unset guiChangedTransforms & BoundChange
      pCSLIDEPrimitive->SetBoundUnchanged();         
      
      // Rerender
      c_pCCurrParser->m_pCSLIDEWorld->PrerenderWindows();
    }
  else
    {
      // Remove this object 
      pCSLIDEPrimitive->Remove(c_pCCurrParser->m_pCSLIDEWorld);
    }

  //printf( "TCL deleted %s\n", pcID);
  return TCL_OK;
}

int CSLIDEParser::HandleModify(int argc, char **argv)
{
  int result;
  CSLIDEPrimitive *pCSLIDEPrimitive;
  VOID *pModifiedObject;  

  pCSLIDEPrimitive = NULL;
  pModifiedObject = NULL;

  // Check args
  if (argc <= 1)
    {
      c_pCCurrParser->HandleError(0, 0, " \"slide modify\": no TYPE given");
      return TCL_ERROR;
    }
  
  int (CSLIDEParser::*cmdmethod)( VOID* &, int, char** );
  int *pos = (int*)(c_pCCurrParser->m_hashOotModifyFuncMap.Peek(argv[1]));
  if (pos == NULL)
    {
      c_pCCurrParser->HandleError(0, 0," \"slide modify %s\": %s is not a valid SLIDE type",
				  argv[1]);
      return TCL_ERROR;
    }

  //////////////////////////////////////////////////////////////////////////////////  
  // Modify objects: go through parser and reprerender
  cmdmethod = (m_pModifyCmdMethods[ *pos ]);
  result = ( c_pCCurrParser->*cmdmethod )( pModifiedObject, argc-1, &(argv[1]) );

  if( pModifiedObject != NULL )
    {
      // CheckSemantics: Initialization
      pCSLIDEPrimitive = (CSLIDEPrimitive *)pModifiedObject;

      pCSLIDEPrimitive->CheckSemantics(c_pCCurrParser->m_pCSLIDEWorld);
      
      if( gbPreprocessDone == TRUE )
	{
	  // Preprocess
	  pCSLIDEPrimitive->SetUnvisited();
	  pCSLIDEPrimitive->Preprocess();
	  c_pCCurrParser->m_pCSLIDEWorld->PreprocessGraphs();	    

	  // Add to world dynamic list
	  if( (pCSLIDEPrimitive->GetDynamic()) &&
	      (NULL == c_pCCurrParser->m_pCSLIDEWorld->m_listDynamics.PeekUnordered((UINT_32)pCSLIDEPrimitive)) )
	    {
	      c_pCCurrParser->m_pCSLIDEWorld->m_listDynamics.InsertLast(pCSLIDEPrimitive,
									(UINT_32)pCSLIDEPrimitive);
	    }
	}

      c_pCCurrParser->m_pCSLIDEWorld->Reprerender(pCSLIDEPrimitive);
    }

  return result;
}

int CSLIDEParser::CBHandleSLIDECmd(ClientData /*clientData*/, Tcl_Interp *pInterp,
				    int argc, char **argv)
{
  ASSERT( pInterp != NULL );
  ASSERT( c_pCCurrParser != NULL );

  int result;
  
  // Check args
  if (argc <= 1)
    {
      c_pCCurrParser->HandleError(0, 0, " \"%s\": No command given to SLIDE", argv[0]);
      return TCL_ERROR;
    }
  
  int (CSLIDEParser::*cmdmethod)( int, char** );
  int *pos = (int*)(c_pCCurrParser->m_hashOotAPIFuncMap.Peek(argv[1]));
  if (pos == NULL)
    {
      c_pCCurrParser->HandleError(0, 0, " \"slide %s\": %s is not a valid SLIDE TCL command", argv[1], argv[1]);
      return TCL_ERROR;
    }
  
  // Call command
  cmdmethod = (m_pAPICmdMethods[ *pos ]);
  result = (c_pCCurrParser->*cmdmethod)(argc-1, &(argv[1]));

  return result;
}


////////////////////////////////////////////////////////////////////////////////////////
/// GUI 

int CSLIDEParser::HandleUnselect()
{
  int iResult;
  
  c_pCCurrParser->m_pCSLIDEWorld->Unselect();
  printf( "CSLIDEParser::HandleUnselect\n");

  iResult = TCL_OK;

  return iResult;
}

int CSLIDEParser::HandleGetParameters(CHAR *pcID, CHAR *pcType)
{
  UINT_32 uiType;
  int iResult;
  CHAR *pcScannerID;

  pcScannerID = NULL;
  pcScannerID = m_pCScanner->FindIdentifier(pcID, FALSE);
  ASSERT( pcScannerID != NULL );

  for( uiType=0; uiType < sliderwtSize; uiType++ )
    {
      if( !strcmp(pcType, gcppcSLIDEReservedWords[uiType]) )
	{
	  break;
	}
    }

  if( uiType == sliderwtSize )
    {
      return TCL_ERROR;
    }
  
  c_pCCurrParser->m_pCSLIDEWorld->GetParameters(pcScannerID, (SLIDEReservedWordType) uiType);

  printf( "CSLIDEParser::HandleGetParameters\n");
  
  //iResult = TCL_ERROR;
  iResult = TCL_OK;

  return iResult;
}

int CSLIDEParser::HandleSetParameters(CHAR *pcID, CHAR *pcType, BOOL bPrerender,
				      int argc, CHAR **argv)
{
  UINT_32 uiType;
  int iResult;
  CHAR *pcScannerID;

  pcScannerID = NULL;
  pcScannerID = m_pCScanner->FindIdentifier(pcID, FALSE);
  ASSERT( pcScannerID != NULL );

  // Get the enum SLIDE type
  for( uiType=0; uiType < sliderwtSize; uiType++ )
    {
      if( !strcmp(pcType, gcppcSLIDEReservedWords[uiType]) )
	{
	  break;
	}
    }

  // Make sure these enum's are valid
  if( uiType == sliderwtSize )
    {
      return TCL_ERROR;
    }

  iResult = c_pCCurrParser->m_pCSLIDEWorld->SetParameters(m_pCScanner, pcScannerID,
							  (SLIDEReservedWordType)uiType, 
							  bPrerender, argc, argv);

  printf( "CSLIDEParser::HandleSetParameters\n");

  return iResult;
}

///////////////////////////////////////////////////////////////////////////////////////
/// Call backs

int CSLIDEParser::CBHandleGUIUnselect(ClientData /*clientData*/, Tcl_Interp *pInterp,
				      int argc, char **argv)
{
  //  ASSERT( pInterp != NULL );
  //  ASSERT( c_pCCurrParser != NULL );
  if( (c_pCCurrParser == NULL) || (pInterp == NULL ))
    {
      // because this gets called when sliders are created, not just when
      // requesting a parameter change
      return TCL_OK;
    }

  int iResult;

  if (argc != 1)
    {
      c_pCCurrParser->HandleError(0, 0, "\"%s\" - expecting 0 args", argv[0]);
      return TCL_ERROR;
    }
  
  iResult = c_pCCurrParser->HandleUnselect();

  return iResult;
}

int CSLIDEParser::CBHandleGUIGetParameters(ClientData /*clientData*/, Tcl_Interp *pInterp,
				    int argc, char **argv)
{
  //  ASSERT( pInterp != NULL );
  //  ASSERT( c_pCCurrParser != NULL );
  if( (c_pCCurrParser == NULL) || (pInterp == NULL ))
    {
      // because this gets called when sliders are created, not just when
      // requesting a parameter change
      return TCL_OK;
    }

  int iResult;
  CHAR *pcID, *pcType;

  if (argc != 3)
    {
      c_pCCurrParser->HandleError(0, 0, "\"%s\" - expecting 2 args", argv[0]);
      return TCL_ERROR;
    }
  
  pcID = argv[1];
  pcType = argv[2];

  iResult = c_pCCurrParser->HandleGetParameters(pcID, pcType);

  return iResult;
}

int CSLIDEParser::CBHandleGUISetParameters(ClientData /*clientData*/, Tcl_Interp *pInterp,
					   int argc, char **argv)
{
  //  ASSERT( pInterp != NULL );
  //  ASSERT( c_pCCurrParser != NULL );
  if( (c_pCCurrParser == NULL) || (pInterp == NULL ))
    {
      // because this gets called when sliders are created, not just when
      // requesting a parameter change
      return TCL_OK;
    }

  int iResult;
  CHAR *pcID, *pcType;
  BOOL bPrerender;

  if( argc < 6)
    {
      c_pCCurrParser->HandleError(0, 0, "\"%s\" - expecting >= 5 args", argv[0]);
      return TCL_ERROR;
    }
  
  pcID = argv[1];
  pcType = argv[2];
  bPrerender = atoi(argv[3]);

  iResult = c_pCCurrParser->HandleSetParameters(pcID, pcType, bPrerender,
						argc-4, &(argv[4]));

  return iResult;
}



